#!/bin/bash -e

function read_version()
{
    local re='[[:blank:]]*=[[:blank:]]*([[:digit:]]{1,2}).*'

    local eMajor=""
    local eMinor=""
    local ePatchLevel=""
    local versionFile="src/app/gbench/gbench/gbench_version.hpp"

    if [ -e $versionFile ] ; then
        while read line
        do
            if [[ $line =~ eMajor$re ]] ; then
                eMajor=${BASH_REMATCH[1]}
            fi
            if [[ $line =~ eMinor$re ]] ; then
                eMinor=${BASH_REMATCH[1]}
            fi
            if [[ $line =~ ePatchLevel$re ]] ; then
                ePatchLevel=${BASH_REMATCH[1]}
            fi
        done < "$versionFile"
    fi

    if [ -z "eMajor" -o  -z "$eMinor" -o -z "$ePatchLevel" ] ; then
        echo "Failed to read version from \"$versionFile\"."
        exit 1
    fi

    local __resultvar=$1
    local myresult="$eMajor.$eMinor.$ePatchLevel"
    eval $__resultvar="'$myresult'"
}


read_version version
release=1

prefix=/opt/ncbi
debug=false
build_deb=false
build_rpm=false

threads=2

while [ ! -z "$1" ]
do
    case "$1" in
        --version=*)
            echo "Warning: --version parameter ignored."
            ;;

        --build=*)
            release=`echo $1 | sed 's/--build=//'`
            ;;

        --no-svn)
            ;;

        --with-debug)
            debug=true
            ;;

        --without-debug)
            debug=false
            ;;

        --build-deb)
            build_deb=true
            ;;

        --build-rpm)
            build_rpm=true
            ;;

        --threads=*)
            threads=`echo $1 | sed 's/--threads=//'`
            ;;
            
        --prefix=*)
            prefix=`echo $1 | sed 's/--prefix=//' | sed 's/\/*$//'`
            ;;

        *)
            echo "option not understood: $1"
            exit 1
            ;;
    esac
    shift
done

if [[ -z "$release" || ! $release =~ ^[0-9]+$ ]]; then
  echo "Invalid build value: $release"
  exit 1
fi

case $prefix in
    /*)
    ;;

    *)
    echo "prefix should be an absolute path: \"$prefix\""
    exit 1
    ;;
esac

GBENCH_DIR=gbench-$version
BUILD_PATH=`pwd`
ARCH=`uname -m`

if [[ $ARCH =~ i[3-6]86 ]]; then
    ARCH=i386
fi

GBENCH_TAR=gbench-linux-$ARCH-$version-$release.tgz

if [ $debug == 'false' ] ; then
    ARCH_DIR=${ARCH}Release
else
    ARCH_DIR=${ARCH}Debug
fi

RPM_ROOT=rpm-build

SPEC_FILE=ncbi-gbench.spec
DEB_CONTROL_FILE=ncbi-gbench.control

THIRD_PARTY_DIR=ThirdParty
THIRD_PARTY_PATH=$BUILD_PATH/$THIRD_PARTY_DIR


SQLITE3_VER=3.6.18
SQLITE3_TAR=sqlite-amalgamation-$SQLITE3_VER.tar.gz
SQLITE3_DIR=sqlite-$SQLITE3_VER

WXWIDGETS_VER=2.9.3
WXWIDGETS_TAR=wxWidgets-$WXWIDGETS_VER.tar.bz2
WXWIDGETS_DIR=wxWidgets-$WXWIDGETS_VER

clear_install()
{
    cd $BUILD_PATH
    rm -rf $THIRD_PARTY_DIR
}

sqlite3_install()
{
    if [ -e $THIRD_PARTY_PATH/$ARCH_DIR/lib/libsqlite3.so ]; then
        return
    fi

    cd $BUILD_PATH

    if [ ! -d $THIRD_PARTY_DIR ]; then
        mkdir $THIRD_PARTY_DIR
    fi

    cd $THIRD_PARTY_DIR

    echo "Downloading and installing SQLITE $SQLITE3_VER..."

    rm -rf $SQLITE3_DIR

    if [ ! -f $SQLITE3_TAR ]; then
        wget http://www.sqlite.org/$SQLITE3_TAR
    fi

    tar xzf $SQLITE3_TAR && cd $SQLITE3_DIR

    ./configure --prefix=$THIRD_PARTY_PATH/$ARCH_DIR && \
    make && make install
}

wxwidgets_install()
{
    if [ $debug == 'false' ] ; then
        cfg_dir=gb-release
    else
        cfg_dir=gb-debug
    fi

    if [ -e $THIRD_PARTY_PATH/$ARCH_DIR/lib/libwx_baseu-2.9.so ]; then
        return
    fi

    cd $BUILD_PATH

    if [ ! -d $THIRD_PARTY_DIR ]; then
        mkdir $THIRD_PARTY_DIR
    fi

    cd $THIRD_PARTY_DIR

    echo "Downloading and installing wxWidgets $WXWIDGETS_VER..."

    rm -rf $WXWIDGETS_DIR

    if [ ! -f $WXWIDGETS_TAR ]; then
        wget http://downloads.sourceforge.net/project/wxwindows/$WXWIDGETS_VER/$WXWIDGETS_TAR
    fi

    tar xjf $WXWIDGETS_TAR && cd $WXWIDGETS_DIR && mkdir $cfg_dir && cd $cfg_dir

    if [ $debug == 'false' ] ; then
        local d='--disable-debug'
    else
        local d='--enable-debug --enable-debug_gdb'
    fi

    local configure="../configure --prefix=$THIRD_PARTY_PATH/$ARCH_DIR --with-opengl --enable-shared --disable-monolithic --disable-sockets --enable-unicode --enable-fs_inet $d"
    echo Configuring wxWidgets at `pwd`
    echo $configure

    $configure && make -j $threads && make install
}

gbench_build()
{
    cd $BUILD_PATH

    if [ $debug == 'false' ] ; then
        local d='--without-debug'
    else
        local d='--with-debug'
    fi

    echo ./configure \
    --with-dll \
    --with-mt \
    --with-gbench \
    --with-libxml \
    --with-libxslt \
    --with-sqlite3=$THIRD_PARTY_PATH/$ARCH_DIR \
    --with-wxwidgets=$THIRD_PARTY_PATH/$ARCH_DIR \
    --with-wxwidgets-ucs \
    $d \
    --with-projects=scripts/projects/ncbi_gbench.lst \
    --with-runpath=$prefix/gbench-$version/lib \
    --without-sybase \
    --with-flat-makefile \
    --without-caution

    ./configure \
    --with-dll \
    --with-mt \
    --with-gbench \
    --with-libxml \
    --with-libxslt \
    --with-sqlite3=$THIRD_PARTY_PATH/$ARCH_DIR \
    --with-wxwidgets=$THIRD_PARTY_PATH/$ARCH_DIR \
    --with-wxwidgets-ucs \
    $d \
    --with-projects=scripts/projects/ncbi_gbench.lst \
    --with-runpath=$prefix/gbench-$version/lib \
    --without-sybase \
    --with-flat-makefile \
    --without-caution

    if [ $? -ne 0 ]; then
      echo "Error: Failed to configure GBench."
      exit 1
    fi

    if [ $debug == 'false' ] ; then
        BUILD_DIR=`ls | grep ReleaseMTDLL`
    else
        BUILD_DIR=`ls | grep DebugMTDLL`
    fi

    cd $BUILD_DIR/build && make -j $threads all_p

    if [ $? -ne 0 ]; then
      echo "Error: Failed to build GBench."
      exit 1
    fi
}

build_rpm_tree()
{
    cd $BUILD_PATH

    rm -rf $RPM_ROOT
    mkdir $RPM_ROOT && cd $RPM_ROOT
    mkdir RPMS
    mkdir SOURCES
    mkdir SPECS
    mkdir SRPMS
    mkdir -p BUILD/$GBENCH_DIR$prefix/$GBENCH_DIR
    mkdir -p BUILD/$GBENCH_DIR/usr/share/applications
    
}

run_rpm()
{
    cd $BUILD_PATH
    sed -e "s!^/opt/ncbi/gbench-%{version}\$!$prefix/gbench-%{version}!" \
        -e "s!^Version: *\$version!Version:    $version!" \
        -e "s!^Release: *\$release!Release:    $release!" \
      <src/app/gbench/gbench_install/$SPEC_FILE >$RPM_ROOT/SPECS/$SPEC_FILE

    cat <<EOF >$HOME/.rpmmacros
%_topdir $BUILD_PATH/$RPM_ROOT
%_tmpdir %{_topdir}/tmp
%packager Roman Katargin
%_signature gpg
%_gpg_name Roman Katargin
EOF

    cd $RPM_ROOT    
    rpmbuild -bb SPECS/$SPEC_FILE

    cd $BUILD_PATH
    ln -f $RPM_ROOT/RPMS/$ARCH/*.rpm gbench-linux-$ARCH-$version-$release.rpm
}

copy_fils_to_rpm()
{
    cd $BUILD_PATH

    cd $BUILD_DIR/bin/gbench
     
    echo copy_files_to_rpm $BUILD_DIR `pwd`
   
    cp -r * $BUILD_PATH/$RPM_ROOT/BUILD/$GBENCH_DIR$prefix/$GBENCH_DIR
    cp -d $THIRD_PARTY_PATH/$ARCH_DIR/lib/lib* $BUILD_PATH/$RPM_ROOT/BUILD/$GBENCH_DIR$prefix/$GBENCH_DIR/lib

    sed -e "s!\$PATH!$prefix/gbench-$version!" \
         <$BUILD_PATH/src/app/gbench/gbench_install/ncbi-gbench.desktop >$BUILD_PATH/$RPM_ROOT/BUILD/$GBENCH_DIR/usr/share/applications/ncbi-gbench.desktop

    find $BUILD_PATH/$RPM_ROOT/BUILD/$GBENCH_DIR$prefix/$GBENCH_DIR/lib/* -type f | xargs chmod 0644

    cd bin
    sed "s:^\(NCBI_GBENCH_INSTALL_DIR=\).*$:\1$prefix/gbench-$version:" <gbench >$BUILD_PATH/$RPM_ROOT/BUILD/$GBENCH_DIR$prefix/$GBENCH_DIR/bin/gbench
}

build_rpm_sources()
{
    echo "Creating tar file $GBENCH_TAR..."
    
    if [ -e $BUILD_PATH/$GBENCH_TAR ]; then
        rm $BUILD_PATH/$GBENCH_TAR
    fi

    tar --directory=$BUILD_PATH/$RPM_ROOT/BUILD/$GBENCH_DIR -czf $BUILD_PATH/$GBENCH_TAR .
    ln -f $BUILD_PATH/$GBENCH_TAR $BUILD_PATH/$RPM_ROOT/SOURCES/
}

build_deb()
{
    DEB_HOST_ARCH=`dpkg-architecture -qDEB_HOST_ARCH`

    cd $BUILD_PATH
    rm -rf $RPM_ROOT/BUILD/$GBENCH_DIR/DEBIAN

    mkdir $RPM_ROOT/BUILD/$GBENCH_DIR/DEBIAN
    sed -e "s!^Architecture: *\$ARCH!Architecture: $DEB_HOST_ARCH!" \
        -e "s!^Version: *\$version!Version: $version-$release!" \
         <src/app/gbench/gbench_install/$DEB_CONTROL_FILE >$BUILD_PATH/$RPM_ROOT/BUILD/$GBENCH_DIR/DEBIAN/control

    UBUNTU_MAJOR_VER=`lsb_release -a | grep "Release:" | sed 's/[^0-9]*\([0-9]*\).*/\1/'`
    if [ "$UBUNTU_MAJOR_VER" -gt "10" ]; then
        sed -i "s/libglewmx1.5/libglewmx1.6/" \
         $BUILD_PATH/$RPM_ROOT/BUILD/$GBENCH_DIR/DEBIAN/control
    fi

    fakeroot dpkg-deb --build $RPM_ROOT/BUILD/$GBENCH_DIR gbench-linux-$DEB_HOST_ARCH-$version-$release.deb

# cleanup
   rm -rf $BUILD_PATH/$RPM_ROOT/BUILD/$GBENCH_DIR/DEBIAN
}

#clear_install
#sqlite3_install
wxwidgets_install
gbench_build
build_rpm_tree
copy_fils_to_rpm
build_rpm_sources

if [ $build_rpm == 'true' ] ; then
    run_rpm
fi

if [ $build_deb == 'true' ] ; then
    build_deb
fi

