#ifndef GUI_CONFIG____THEMESETREG_HPP
#define GUI_CONFIG____THEMESETREG_HPP

/* $Id: theme_set.hpp 14562 2007-05-18 11:48:21Z dicuccio $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Author:  Robert G. Smith
 *
 */

 /**
 *  @class CThemeSet - track styles other types of values.
 *   CThemeSet is an adapter of the CGuiRegistry, layered on top of
 *   CSettingsSet.
 *   Besides the key hierarchy that CSettingSet expects, a CThemeSet
 *   expects a subkey 'sets' followed by keys which are the type keys
 *   of the other CSettingsSet's it tracks, whose values are the styles
 *   they will have. Example:

        GBPlugins
            MyTheme
                name = Prefs for My Plugin
                current = default_set
                default_values
                    name = *
                    sets
                        FeatColors = default_set
                        Sizes = default_set
                 ncbitheme
                    name = NCBI Theme
                    sets
                        Sizes = compact
                    ...
            FeatColors
                name = Feature Colors
                ...
            Sizes
                name = Element Sizes
                default_values
                    ...
                compact
                    ...

 */

#include <corelib/ncbistr.hpp>
#include <gui/config/settings_set.hpp>
#include <string>
#include <map>

BEGIN_NCBI_SCOPE

BEGIN_SCOPE(objects)
    class CPluginConfigCache;
END_SCOPE(objects)


class NCBI_GUICONFIG_EXPORT CThemeSet : public CSettingsSet
{
public:
    CThemeSet(const string& key,
                 CGuiRegistry* config_cache,
                 const string& delim = kDefaultDelim);
    virtual ~CThemeSet();

    /// uncache any lookup data
    void Uncache();

    /// Get a list of type keys of the settings this manages.
    vector<string>  GetTypes() const;
    /// Get a list of the names (user displayable) of the settings this manages.
    vector<string>  GetTypeDescriptions() const;
    /// for a given managed set, get the names of all its styles.
    /// @sa CSettingsSet::GetStyleNames
    list<string>    GetSetsStyles(const string& type) const;
    /// for a given managed set, get its current style.
    /// @sa CSettingsSet::GetCurrentStyleName().
    string          GetSetsCurrentStyleName(const string& type) const;
    /// for a given managed set, change its current style.
    /// @sa CSettingsSet::SetCurrentStyleName
    void            SetSetsCurrentStyleName(const string& type, const string& new_style);
    /// for a given managed set, get its name.
    /// @sa CSettingsSet::GetTypeDescription
    string          GetSetsDescription(const string& type) const;

    /// In case the 'type' SettingsSet current setting, might have gotten
    /// changed outside this theme, sync this theme back to it.
    void            SetMyCurrentStyleName(const string& type);

    /// interface functions for the member m_Settings.
    const CSettingsSet&  GetSettingsSet(const string& type) const;
    CSettingsSet&        SetSettingsSet(const string& type);

    void    DumpAll(CNcbiOstream& ostr);

private:
    string  x_GetSetsCurrentStyleKey(const string& type) const;
    void    x_SetSetsCurrentStyleKey(const string& type, const string& new_key);
    void    x_AddSettingsSet(const string& type, CSettingsSet& set);

    typedef map<string, CSettingsSet>    TSettingsReg;
    TSettingsReg        m_Settings;

private:
    /// forbidden
    CThemeSet(const CThemeSet&);
    CThemeSet& operator=(const CThemeSet&);
};


END_NCBI_SCOPE

#endif // GUI_CONFIG____THEMESETREG_HPP
