#ifndef GUI_CORE___APP_EXPLORER_ITEM__HPP
#define GUI_CORE___APP_EXPLORER_ITEM__HPP

/*  $Id: app_explorer_item.hpp 22701 2010-12-14 21:35:07Z katargir $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Andrey Yazhuk
 *
 * File Description:
 *  CExplorerItem - represents an item in Application Explorer service.
 */

#include <corelib/ncbiobj.hpp>

#include <gui/gui_export.h>

#include <wx/treebase.h>

#include <set>

class wxMenu;

BEGIN_NCBI_SCOPE

///////////////////////////////////////////////////////////////////////////////
/// CExplorerItem - an abstract item in Application Explorer
/// icons are represented as integers - indexes in the tree image list
class  NCBI_GUICORE_EXPORT  CExplorerItem : public CObject
{
    friend class CAppExplorerService;
public:
    typedef CRef<CExplorerItem> TItemRef;
    typedef vector<TItemRef>    TItemRefVector;

public:
    static set<CExplorerItem*>  sm_AliveItems;

    CExplorerItem(const wxString& label, bool is_expanded = false);
    CExplorerItem(const wxString& label, int normal, int expanded = -1, 
                  bool is_expanded = false);
    virtual ~CExplorerItem();

    virtual const wxString&  GetLabel() const;
    virtual void    SetLabel(const wxString& s);
    void    SetLabel(const string& s);

    virtual int GetIconIndex(wxTreeItemIcon which) const;
    virtual void    SetIconIndex(int normal, int expanded = -1);

    virtual bool    IsExpanded() const;
    virtual void    Expand(bool exp);

    virtual bool    IsEnabled() const;
    virtual bool    IsDisabled() const;
    virtual void    Enable(bool en);

    virtual CExplorerItem*  GetParent();
    virtual const CExplorerItem*  GetParent() const;
    virtual const TItemRefVector&   GetChildItems() const;

    virtual void    AddChildItem(CExplorerItem& ch_item);
    virtual void    EraseChildItem(CExplorerItem& ch_item);
    virtual void    EraseChildItem(TItemRefVector::iterator it);
    virtual void    EraseChildItem(size_t index);
    virtual void    EraseAllChildItems(bool recursively = false);

protected:
    virtual TItemRefVector&   x_GetChildItems();

protected:
    wxString m_Label;
    int m_NormalIcon;
    int m_ExpandedIcon;

    TItemRef    m_Parent;
    TItemRefVector  m_ChildItems;

    bool m_Enabled;
    bool m_Expanded;
};

///////////////////////////////////////////////////////////////////////////////
// CExplorerItemEx - template extension of CExplorerItem defining custom tree item
template<class TData>   class CExplorerItemEx : public    CExplorerItem
{
public:
    CExplorerItemEx(const wxString& label, bool is_open = false)
    :   CExplorerItem(label, is_open)
    {
    }
    CExplorerItemEx(const wxString& label, const TData& data)
    :   CExplorerItem(label),
        m_Data(data)
    {
    }
    CExplorerItemEx(const wxString& label,
                    const TData& data,
                    int normal, int expanded = -1)
    :   CExplorerItem(label, normal, expanded),
        m_Data(data)
    {
    }
    const TData&    GetData() const
    {
        return m_Data;
    }
    TData&          GetData()
    {
        return m_Data;
    }
    void    SetData(const TData&  data)
    {
        m_Data = data;
    }
protected:
    TData   m_Data;
};


class CAppExplorerService;
class CUICommandRegistry;

///////////////////////////////////////////////////////////////////////////////
/// IExplorerItemCmdContributor - interface representing a component that
/// contributes commands applicable to Explorer Items.
class IExplorerItemCmdContributor
{
public:
    /// Contribution consists of a Menu object and event handler
    typedef pair<wxMenu*, wxEvtHandler*>    TContribution;
    typedef vector< CRef<CExplorerItem> >    TItemRefVector;

    /// for the given set of items returns a contribution
    virtual TContribution   GetMenu(TItemRefVector& items,
                                    CAppExplorerService& app_service) = 0;

    virtual ~IExplorerItemCmdContributor()  {};
};


END_NCBI_SCOPE

#endif  /// GUI_CORE___APP_EXPLORER_ITEM__HPP
