#ifndef GUI_CORE___APP_EXPLORER_SERVICE__HPP
#define GUI_CORE___APP_EXPLORER_SERVICE__HPP

/*  $Id: app_explorer_service.hpp 24732 2011-11-16 18:44:58Z katargir $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Andrey Yazhuk
 *
 * File Description:
 *
 */

#include <corelib/ncbimtx.hpp>
#include <corelib/ncbistl.hpp>

#include <gui/core/app_explorer_item.hpp>
#include <gui/core/app_explorer_data_object.hpp>
#include <gui/core/workspace.hpp>
#include <gui/core/workspace_folder.hpp>
#include <gui/core/project.hpp>
#include <gui/core/project_item.hpp>
#include <gui/core/project_folder.hpp>
#include <gui/core/document.hpp>

#include <gui/utils/command.hpp>

#include <gui/framework/service.hpp>
#include <gui/framework/observable.hpp>
#include <gui/framework/view.hpp>
#include <gui/framework/ui_data_source.hpp>

#include <gui/widgets/wx/wm_client.hpp>

#include <gui/objutils/reg_settings.hpp>

#include <gui/objects/GBProjectHandle.hpp>
#include <gui/objects/GBWorkspace.hpp>
#include <gui/objects/WorkspaceFolder.hpp>
#include <gui/objects/ProjectFolder.hpp>
#include <gui/objects/ProjectItem.hpp>

#include <wx/imaglist.h>

BEGIN_NCBI_SCOPE

class CProjectService;


template <class TId, int Type> class TIdWrapper
{
public:
    TIdWrapper(TId id)    :   m_Id(id)    {}
    operator TId() {   return m_Id;    }
    virtual ~TIdWrapper() {}

    TId m_Id;
};


template <class TId, int TType>
    bool operator<(const TIdWrapper<TId, TType>& id1, const TIdWrapper<TId, TType>& id2)
{
    return id1.m_Id < id2.m_Id;
}


class CProjectFolderTreeItem : public CExplorerItemEx<CProjectFolderPrx::TId>
{
    typedef CExplorerItemEx<CProjectFolderPrx::TId> TParentType;
public:
    CProjectFolderTreeItem(const wxString& label, 
                           CProjectFolderPrx::TId data,
                           int normal,        ///< normal icon index
                           int expanded = -1, ///< expanded icon index
                           bool is_open = false ///< Folder open status in the tree
                           );

    //virtual bool    IsExpanded() const;
    //virtual void    Expand(bool exp);
};


////////////////////////////////////////////////////////////////////////////////
/// CAppExplorerService - TODO comments
class NCBI_GUICORE_EXPORT CAppExplorerService :
    public CObjectEx,
    public IService,
    public IServiceLocatorConsumer,
    public CObservable,
    public IObserver,
    public IRegSettings
{
public:
    typedef CGuard<CAppExplorerService>     TGuard;
    typedef CGuard<CProjectService>     TProjectServiceGuard;
    typedef CGuard<objects::CGBWorkspace>   TProjectWorkspaceGuard;
    typedef vector< CIRef<IProjectView> >   TProjectViewVector;

    typedef CRef<CExplorerItem> TItemRef;
    typedef vector<TItemRef>    TItemRefVector;

    enum EItemType {
        eWorkspace       = 0x001,
        eWorkspaceFolder = 0x002,
        eProject         = 0x004,
        eProjectFolder   = 0x008,
        eProjectItem     = 0x010,
        eDataSource      = 0x020,
        eView            = 0x040,
        eFolder          = 0x080,
        eDataLoader      = 0x100,
        eInvalid         = 0x8000000
    };
    typedef TIdWrapper<CWorkspaceConstPrx::TId, eWorkspace> TWorkspaceId;
    typedef TIdWrapper<CWorkspaceFolderPrx::TId, eWorkspaceFolder> TWsFolderId;
    typedef TIdWrapper<CProjectPrx::TId, eProject>     TProjectId;
    typedef TIdWrapper<CProjectItemPrx::TId, eProjectItem>    TPrjItemId;
    typedef TIdWrapper<string, eDataLoader>    TDataLoaderId;

    // Tree Items
    typedef CProjectFolderTreeItem          TProjectFolderTreeItem;
    typedef CExplorerItemEx<wxString>       TFolderTreeItem;
    typedef CExplorerItemEx<TWorkspaceId>   TWorkspaceTreeItem;
    typedef CExplorerItemEx<TWsFolderId>    TWsFolderTreeItem;
    typedef CExplorerItemEx<TProjectId>     TProjectTreeItem;
    typedef CExplorerItemEx<TPrjItemId>     TProjectItemTreeItem;
    typedef CExplorerItemEx<TDataLoaderId>  TDataLoaderTreeItem;
    typedef CExplorerItemEx< CIRef<IProjectView> >  TViewTreeItem;
    typedef CExplorerItemEx< CIRef<IUIDataSource> > TDataSourceTreeItem;

public:
    CAppExplorerService();
    virtual ~CAppExplorerService();

    /// @name IService interface implementation
    /// @{
    virtual void    InitService();
    virtual void    ShutDownService();
    /// @}

    /// @name IServiceLocatorConsumer interface implementation
    /// @{
    virtual void    SetServiceLocator(IServiceLocator* locator);
    /// @}

    /// @name IRegSettings interface implementation
    /// @{
    virtual void    SetRegistryPath(const string& path);
    virtual void    SaveSettings() const;
    virtual void    LoadSettings();
    /// @}

    /// @name IObserver interface implementation
    /// @{
    virtual void    OnUpdate(IObserverUpdate& update);
    /// @}

    CRef<CExplorerItem> GetRootItem();

    wxImageList&    GetIconsImageList();

    EItemType   GetItemType(const CExplorerItem& item);
    int         GetItemTypes(TItemRefVector& items);

    bool    MatchType(const CExplorerItem& item, EItemType type);
    bool    MatchTypesExactly(TItemRefVector& items, int type_mask);
    bool    ContainsAllTypes(TItemRefVector& items, int type_mask);

    template<class TData>
        void GetItemsData(TItemRefVector& items, vector<TData>& data);

    void    ItemsToViews(TItemRefVector& items, TProjectViewVector& views);

    /// remove items that do not match type mask
    void    FilterByType(TItemRefVector& items, int type_mask);

    TProjectId GetProjectIdByChild(CExplorerItem& item);

    void    OnItemActivated(CExplorerItem& item);
    void    OnItemExpandedCollapsed(CExplorerItem& item);

    bool    IsDataFolder( CExplorerItem& item );
    bool    IsViewsFolder( CExplorerItem& item );

    bool    IsFolder( const CExplorerItem& item, wxString type = wxEmptyString );

    /// @name Editing
    /// @{
    void    NewFolder(CExplorerItem& item);
    void    NewWorkspaceFolder(CExplorerItem& item);
    void    NewProjectFolder(CExplorerItem& item);

    void    OnProperties(CExplorerItem& item);
    bool    IsPropertiesCmdEnabled(CExplorerItem& item);

    void    Remove(TItemRefVector& items);

    void    EnableDisableProjectItems(TItemRefVector& items);

    bool    CanLoadUnloadProjects(TItemRefVector& items, bool load);
    /// @}

    /// @name Clipboard and D&D
    /// @{
    CAppExplorerDataObject* CreateDataObject(TItemRefVector& items, bool dnd, bool cut);
    void    CutOrCopyToClipboard(TItemRefVector& items, bool cut);
    bool    CanCutOrCopyToClipboard(TItemRefVector& items);
    void    PasteFromClipboard(CExplorerItem& target);
    bool    CanPasteTo(CExplorerItem& target);
    void    Drop(CExplorerItem& target, CAppExplorerData& data, bool move);
    bool    CanDropTo(CExplorerItem& target);
    /// @}

    static const wxString sm_DATA_FOLDER;
    static const wxString sm_VIEWS_FOLDER;
    static const wxString sm_ALL_VIEWS_FOLDER;
    static const wxString sm_LOADERS_FOLDER;
    static const wxString sm_DATA_SOURCES_FOLDER;

protected:
    int     x_AddIconToImageList(const string& alias);
    int     x_GetImageIndexByAlias(const string& alias);

    void    x_UpdateItems();

    void    x_UpdateWorkspaceTree();
    void    x_UpdateWindowsTree();
    void    x_UpdateDataSourceTree();

    void    x_UpdateWorkspaceFolder(CExplorerItem& item,
                                    CWorkspaceFolderConstPrx& folder);
    void    x_UpdateWorkspaceFolder_Folders(CExplorerItem& item,
                                            CWorkspaceFolderConstPrx& folder);
    void    x_UpdateWorkspaceFolder_Projects(CExplorerItem& item,
                                             CWorkspaceFolderConstPrx& folder);

    void    x_UpdateProject(TProjectTreeItem& item);
    void    x_UpdateProject_DataLoaders(TProjectTreeItem& item);
    void    x_UpdateProject_Views(TProjectTreeItem& item);

    static CExplorerItem*  x_FindItemByLabel(const wxString& label, TItemRefVector& items);

    void    x_UpdateProjectItem(TProjectItemTreeItem& item,
                                CProjectFolderConstPrx& parent_folder);
    void    x_UpdateProjectFolder(TProjectFolderTreeItem& item,
                                  CProjectFolderConstPrx& folder);
    void    x_UpdateProjectFolder_Folders(TProjectFolderTreeItem& item,
                                          CProjectFolderConstPrx& folder);
    void    x_UpdateProjectFolder_Items(TProjectFolderTreeItem& item,
                                        CProjectFolderConstPrx& folder);

    void    x_UpdateProjectLoadersFolder(TFolderTreeItem& item);
    void    x_UpdateProjectWindowsFolder(TFolderTreeItem& item);

    string x_GetViewLabel( IProjectView& view, bool local );
    TProjectItemTreeItem*   x_CreateProjectItemItem(CProjectItemConstPrx& prj_item);
    TViewTreeItem*          x_CreateProjectViewTreeItem(IProjectView& view, bool local);
    //TViewTreeItem*          x_CreateProjectViewerTreeItem(IProjectView& viewer, bool local);

    void x_DecorateProjectItem(TProjectItemTreeItem& tree_item,
                               CProjectItemConstPrx& item);

    /// @name Editing helpers
    /// @{
    void    x_GetTopItems(TItemRefVector& items, TItemRefVector& top_items);

    void    x_OnProjectProperties(CExplorerItem& item);
    void    x_OnWorkspaceProperties();
    void    x_OnProjectItemProperties(CExplorerItem& item);
    void    x_OnProjectFolderProperties(CExplorerItem& item);
    void    x_OnWorkspaceFolderProperties(CExplorerItem& item);

    bool    x_RemoveProjectFolder(CExplorerItem& item, bool test);
    bool    x_RemoveProjectItem(CExplorerItem& item, bool test);
    bool    x_RemoveDataLoader(CExplorerItem& item, bool test);
    bool    x_RemoveWorkspaceFolder(CExplorerItem& item, bool test);
    /// @}

    /// @name Clipboard operations helpers
    /// @{
    void x_ClearClipboardData();
    void x_Paste(TItemRefVector& items, CExplorerItem& target, bool cut);
    bool x_PasteItem(CExplorerItem& item, CProjectPrx& dst_project,
                     CProjectFolderPrx& dst_folder, bool cut);
    bool x_CanPasteProjectFolder(CProjectFolderPrx& parent,
                                 CProjectFolderPrx& child, string& error);
    void x_PasteProjectFolder(CProjectFolderPrx& parent,
                              CProjectFolderPrx& child);
    void x_MoveProjectFolder(CProjectPrx& project, CProjectFolderPrx& parent,
                             TProjectFolderTreeItem& child_item);
    void x_PasteProjectItem(CProjectFolderPrx& parent_folder,
                            CProjectItemPrx& item);
    /// @}

private:
    CAppExplorerService(const CAppExplorerService&);
    CAppExplorerService& operator= (const CAppExplorerService&);

private:
    typedef map<string, int>    TAliasToIndexMap;

    string  m_RegPath;

    IServiceLocator* m_ServiceLocator;
    CProjectService* m_ProjectService;

    // Icons
    wxImageList m_IconsImageList;

    // a map for registering user-defined icons
    TAliasToIndexMap    m_AliasToIndex;

    int m_DataSourceIcon;
    int m_FolderOpenIcon;
    int m_FolderClosedIcon;
    int m_WorkspaceClosedIcon;
    int m_WorkspaceOpenIcon;
    int m_ProjectItemIcon;
    int m_DocIconDisabled;
    int m_ProjectClosedIcon;
    int m_ProjectOpenIcon;
    int m_ViewerIcon;
    int m_ViewIcon;

    CRef<CExplorerItem>         m_RootItem;
    CRef<TWorkspaceTreeItem>    m_WorkspaceItem;
    CRef<TFolderTreeItem>   m_ViewsRootItem;
    CRef<TFolderTreeItem>   m_DSRootItem;

    CRef<CAppExplorerData> m_ClipboardData;

    /// Temporary proxy objects
    CRef<CWorkspaceConstPrx> m_Workspace;
    CRef<CProjectConstPrx>   m_CurrProject;
};


template<class TData>
    void CAppExplorerService::GetItemsData(TItemRefVector& items, vector<TData>& data)
{
    typedef CExplorerItemEx<TData> TItemEx;

    for(  size_t i = 0;  i < items.size();  i++ )   {
        CExplorerItem* item = items[i].GetPointer();
        TItemEx* ex_item = dynamic_cast<TItemEx*>(item);
        if(ex_item) {
            data.push_back(ex_item->GetData());
        }
    }
}

END_NCBI_SCOPE

#endif  /// GUI_CORE___APP_EXPLORER_SERVICE__HPP
