#ifndef GUI_OBJUTILS___ASSEMBLY_INFO__HPP
#define GUI_OBJUTILS___ASSEMBLY_INFO__HPP

/*  $Id: assm_info.hpp 27815 2013-04-08 15:10:18Z rudnev $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Author:  Liangshou Wu
 *
 * File Description:
 *   A utility class providing service for assembly information retrieval.
 * The target output is one or a list of related *full* assemblies.
 *
 */

#include <corelib/ncbiobj.hpp>
#include <gui/objects/DL_AssmInfo.hpp>
#include <objects/genomecoll/GC_Assembly.hpp>

#include <set>

BEGIN_NCBI_SCOPE


class NCBI_GUICORE_EXPORT CAssemblyInfo : public CObject
{
public:
    typedef set<string> TIds;
    /// id map from local id to gi.
    typedef map<string, int> TIdMap;
    /// named id map.
    typedef pair<string, TIdMap> TNamedIdMap;
    typedef vector<TNamedIdMap> TIdMaps;

    CRef<objects::CDL_AssmInfo> GetAssms_Gi(int gi) const;
    CRef<objects::CDL_AssmInfo> GetAssms_TaxId(const string& tax_id) const;
    CRef<objects::CDL_AssmInfo> GetAssms_Term(const string& term) const;

    static CRef<objects::CGC_Assembly> GetGCAssembly(const string& assm_acc);

    /// Create a string used in CIdMapperConfig from a assembly.
    static void GetIdMapperString(const string& assm_acc, string& mapper_str);

    /// Create a list of id maps from a assembly.
    static void GetIdMaps(const string& assm_acc, TIdMaps& id_maps);

    /// get all assembly ids associated with a gi where this gi is a chromosome.
    /// the list will be reset if there is no assembly where this GI is a chromosome
    static void GetAssmIds_GIChr(TIds& gc_ids, int gi);

private:
    /// validate if the given tax-id.
    bool x_ValidateTaxId(const string& tax_id) const;
    /// filter out the tax ids that don't have assembly on them.
    void x_FilterTaxIds(TIds& tax_ids) const;
    /// search all related links in the given db with a search term.
    void x_SearchIds(const string& term, const string& db, TIds& ids) const;
    /// get detailed information for a given tax-id.
    void x_GetTaxIdInfo(const TIds& tax_ids, objects::CDL_AssmInfo::TTax_ids& tax_info) const;
    /// get all assembly ids associated with a gi.
    void x_GetAssmIds_GI(string& gc_ids, int gi) const;
    /// get all assembly ids associated with a tax-id.
    void x_GetAssmIds_TaxId(string& gc_ids, const string& tax_id) const;
    /// get all assembly ids associated with a unknown term.
    void x_GetAssmIds_Unknown(string& gc_ids, const string& term) const;
    /// get detailed assembly information for a list of gencoll ids.
    void x_GetAssmInfo(const string& gc_ids, objects::CDL_AssmInfo::TAssemblies& assms) const;
    /// sort assemblies based on release date.
    /// apply special rules for human, mouse, cow and rat.
    void x_SortAssms(objects::CDL_AssmInfo::TAssemblies& assms) const;
    /// convert a list of ids into a str.
    string x_CreateIdStr(const TIds& ids) const;

    CRef<objects::CDL_AssmInfo> x_CreateAssmInfo(objects::CDL_AssmInfo::TAssemblies& assms) const;

};


END_NCBI_SCOPE

#endif // GUI_OBJUTILS___ASSEMBLY_INFO__HPP
