#ifndef GUI_CORE___FORM_ARG_DESC__HPP
#define GUI_CORE___FORM_ARG_DESC__HPP

/*  $Id: form_arg_desc.hpp 17878 2008-09-24 13:07:17Z dicuccio $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Yury Voronov
 *
 * File Description:
 *
 */

#include <gui/objutils/objects.hpp>

#include <wx/variant.h>
#include <wx/list.h>

BEGIN_NCBI_SCOPE

WX_DECLARE_LIST(wxVariant, CwxVarList);

class CFormArgDesc : public CObject
{
public:
    //? this list can be made extensible by using static constant members of special class
    enum EWidgetType {
        eNotSet,

        eSingleLine,
        eMultipleLine,
        eSecretLine,
        eDropDown,
        eList,
        eCheckBox,
        eRadio,
        eObjectList,
        eInputFile,
        eOutputFile,

        eMaxWidgetType
    };

    /* Value types defined in wxVariant:
    bool
    char
    long
    double
    string

    + object - CwxObject or CwxScopedObject
    */

public:
    CFormArgDesc() {}
    virtual ~CFormArgDesc() {}

    string GetType() const { return m_Type; }
    void SetType( string aType ){ m_Type = aType; }

    string GetObjectType() const { return m_ObjType; }
    void SetObjectType( string aType ){ m_ObjType = aType; }

    EWidgetType GetWidgetType() const { return m_WgtType; }
    void SetWidgetType( EWidgetType aType ){ m_WgtType = aType; }

    bool IsMultiple() const { return mf_Multiple; }
    void SetMultipleFlag( bool aFlag ){ mf_Multiple = aFlag; }

    string GetName() const { return m_Name; }
    void SetName( string aName ){ m_Name = aName; }

    string GetLabel() const { return m_Label; }
    void SetLabel( string aLabel ){ m_Label = aLabel; }

    string GetTooltip() const { return m_Tooltip; }
    void SetTooltip( string aTooltip ){ m_Tooltip = aTooltip; }

    bool IsHidden() const { return mf_Hidden; }
    void SetHiddenFlag( bool aHidden ){ mf_Hidden = aHidden; }

    bool IsOptional() const { return mf_Optional; }
    void SetOptionalFlag( bool aOptional ){ mf_Optional = aOptional; }

    wxVariant GetDefault() const { return m_Default; }
    void SetDefault( wxVariant aDefault ){ m_Default = aDefault; }

    wxVariant GetChoices() const { return m_MultipleChoices; }
    void SetChoices( wxVariant aChoices ){ m_MultipleChoices = aChoices; }

private:

    string m_Type;
    string m_ObjType;
    EWidgetType m_WgtType;
    bool mf_Multiple;

    string m_Name;
    string m_Label;
    string m_Tooltip;
    bool mf_Hidden;

    bool mf_Optional;
    wxVariant m_Default;
    wxVariant m_MultipleChoices;
};


class CwxObject : public wxObject
{
public:
    CwxObject();
    CwxObject( const CwxObject& anObj );
    explicit CwxObject( CObject& anObj );

    CwxObject& operator =( const CwxObject& anObj );
    bool operator ==( const CwxObject& anObj ) const;

    bool IsOk() const;
    const CObject& GetObject() const;
    CObject& GetObject();
    void SetObject( CObject& anObj );

private:
    CRef<CObject> m_Object;

    DECLARE_DYNAMIC_CLASS(CwxObject)
};
DECLARE_VARIANT_OBJECT(CwxObject)

class CwxScopedObject : public wxObject
{
public:
    CwxScopedObject();
    CwxScopedObject( const CwxScopedObject& aScObj );
    explicit CwxScopedObject( SScopedObject& aScObj );
    CwxScopedObject( CObject& anObj, objects::CScope& aSc );

    CwxScopedObject& operator =( const CwxScopedObject& aScObj );
    bool operator ==( const CwxScopedObject& aScObj );

    bool IsOk() const;
    const SScopedObject& GetScopedObject() const;
    SScopedObject& GetScopedObject();
    void SetScopedObject( SScopedObject& aScObj );

    const CObject& GetObject() const;
    CObject& GetObject();
    const objects::CScope& GetScope() const;
    objects::CScope& GetScope();

private:
    SScopedObject m_Scoped;

    DECLARE_DYNAMIC_CLASS(CwxScopedObject)
};
DECLARE_VARIANT_OBJECT(CwxScopedObject)



inline CwxObject::CwxObject()
{
}

inline CwxObject::CwxObject( CObject& anObj )
{
    m_Object.Reset(&anObj);
}

inline CwxObject::CwxObject( const CwxObject& anObj )
{
    m_Object.Reset( const_cast<CObject*>(&anObj.GetObject()) );
}

inline CwxObject& CwxObject::operator =( const CwxObject& anObj )
{
    m_Object.Reset( & const_cast<CwxObject&>(anObj).GetObject() ); return *this;
}

inline bool CwxObject::operator ==( const CwxObject& anObj ) const
{
    return &*m_Object == &anObj.GetObject();
}

inline bool CwxObject::IsOk() const
{
    return m_Object != NULL;
}

inline const CObject& CwxObject::GetObject() const
{
    return *m_Object;
}

inline CObject& CwxObject::GetObject()
{
    return *m_Object;
}

inline void CwxObject::SetObject( CObject& anObj )
{
    m_Object.Reset( &anObj );
}

inline CwxScopedObject::CwxScopedObject()
{
}

inline CwxScopedObject::CwxScopedObject( const CwxScopedObject& aScObj )
    : m_Scoped( const_cast<SScopedObject&>(aScObj.GetScopedObject()) )
{
}

inline CwxScopedObject::CwxScopedObject( SScopedObject& aScObj )
    : m_Scoped( aScObj )
{
}
inline CwxScopedObject::CwxScopedObject( CObject& anObj, objects::CScope& aSc )
    : m_Scoped( anObj, aSc )
{
}

inline CwxScopedObject& CwxScopedObject::operator =( const CwxScopedObject& aScObj )
{
    m_Scoped = aScObj.GetScopedObject(); return *this;
}

inline bool CwxScopedObject::operator ==( const CwxScopedObject& aScObj )
{
    return m_Scoped == aScObj.GetScopedObject();
}

inline bool CwxScopedObject::IsOk() const
{
    return m_Scoped.object != NULL && m_Scoped.scope != NULL;
}

inline const SScopedObject& CwxScopedObject::GetScopedObject() const
{
    return m_Scoped;
}

inline SScopedObject& CwxScopedObject::GetScopedObject()
{
    return m_Scoped;
}

inline void CwxScopedObject::SetScopedObject( SScopedObject& aScObj )
{
    m_Scoped = aScObj;
}

inline const CObject& CwxScopedObject::GetObject() const
{
    return *m_Scoped.object;
}

inline CObject& CwxScopedObject::GetObject()
{
    return *m_Scoped.object;
}

inline const objects::CScope& CwxScopedObject::GetScope() const
{
    return *m_Scoped.scope;
}

inline objects::CScope& CwxScopedObject::GetScope()
{
    return *m_Scoped.scope;
}

END_NCBI_SCOPE

#endif // GUI_CORE___FORM_ARG_DESC__HPP
