#ifndef GUI_CORE_DEMO___LOADING_APP_JOB__HPP
#define GUI_CORE_DEMO___LOADING_APP_JOB__HPP

/*  $Id: loading_app_job.hpp 26040 2012-07-03 16:13:24Z kuznets $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors: Andrey Yazhuk, Liangshou Wu
 *
 * File Description:
 *
 */

#include <corelib/ncbistl.hpp>
#include <corelib/ncbimtx.hpp>

#include <util/format_guess.hpp>
#include <util/compress/stream.hpp>

#include <gui/core/project_service.hpp>
#include <gui/core/project.hpp>

#include <gui/framework/app_job_task.hpp>

#include <gui/utils/app_job_impl.hpp>

#include <objmgr/scope.hpp>

#include <wx/string.h>


BEGIN_NCBI_SCOPE

BEGIN_SCOPE(objects)
    class CProjectFolder;
END_SCOPE(objects)


///////////////////////////////////////////////////////////////////////////////
/// CDataLoadingOptions - describes how new Project Items shall be added
/// to a workspace. If folder name is not empty a new Project Folder will be
/// created for the items.
class NCBI_GUICORE_EXPORT  CDataLoadingOptions
{
public:
    enum EAction {
        eInvalidAction = -1,
        eDecideLater,
        eCreateNewProject,
        eCreateSeparateProjects,
        eAddToExistingProject
    };
    typedef CProjectPrx::TId   TProjectId;

public:
    CDataLoadingOptions();

    void Set_DecideLater();
    void Set_CreateNewProject( const string& folder = kEmptyStr );
    void Set_CreateSeparateProjects();
    void Set_AddToExistingProject(
        TProjectId& project_id, const string& folder = kEmptyStr
    );

    EAction GetAction();
    TProjectId GetTargetProjectId();
    string GetFolderName();

protected:
    EAction m_Action;
    TProjectId  m_TargetProjectId;
    string  m_FolderName;
};


///////////////////////////////////////////////////////////////////////////////
/// CDataLoadingJob - a base class for Jobs loading data into projects.

class NCBI_GUICORE_EXPORT CDataLoadingJob : public CAppJob
{
public:
    typedef CRef<objects::CLoaderDescriptor>   TLoaderRef;
    typedef CRef<objects::CProjectItem>        TItemRef;
    typedef CProjectPrx::TId                   TProjectId;

    /// a loader might be tied with a specific project item.
    /// if so, we need to know such that they will go together and
    /// get added to the same project.
    typedef multimap<const objects::CProjectItem*, TLoaderRef> TLoaders;

public:
    /// constructor for derived classes
    CDataLoadingJob( CProjectService* service );
    virtual ~CDataLoadingJob();

    virtual CDataLoadingOptions& GetOptions();
    virtual void SetOptions( const CDataLoadingOptions& options );

    CProjectService* GetProjectService();

    /// @name IAppJob implementation
    /// @{
    virtual EJobState Run();
    virtual CRef<CObject> GetResult();
    /// @}

    virtual bool FailedToAddItems() const;
	virtual bool FailedToAddLoaders() const;

    const vector<TItemRef>& GetItems() const;
    objects::CScope* GetScope() const;

    void SetItems( const vector<TItemRef>& items ) { m_Items = items; }

    virtual void GetHTMLResults( wxString& /*title*/, wxString& /*html*/ ) {}

    /// @name Utility functions for data-loading
    /// @{

    /// Identify compression format by extension
    static CFormatGuess::EFormat GetCompressType(const wxString& fname);
    static CCompressionStreamProcessor* CreateCompressionProcessor(CFormatGuess::EFormat fmt);

    /// sub-class to keep compression stream pointers
    ///
    class CCompressStreamHolder
    {
    public:
        CCompressStreamHolder(const wxString& fn)
            : m_istr_p(0)
        {
            m_in.reset(new CNcbiIfstream(fn.fn_str(), ios::binary));
            CFormatGuess::EFormat fmt_cmprz = CDataLoadingJob::GetCompressType(fn);
            switch (fmt_cmprz) {
            case CFormatGuess::eGZip:
            case CFormatGuess::eBZip2:
                m_zin.reset(new CCompressionIStream(*m_in.release(), 
                    CDataLoadingJob::CreateCompressionProcessor(fmt_cmprz), CCompressionStream::fOwnAll));
                m_istr_p = m_zin.get();
                break;
            default:
                m_istr_p = m_in.get();
                break;
            } // switch
        }
        CNcbiIstream* GetIStreamP()  { return m_istr_p; }
    private:
        auto_ptr<CNcbiIfstream>       m_in;
        auto_ptr<CCompressionIStream> m_zin;
        CNcbiIstream*                 m_istr_p;
    };

    /// @}

protected:
    
    typedef CDataLoadingOptions TOptions;

    /// prepares job for execution by clearing state variables
    virtual void x_ResetState();

    virtual void x_Run();

    /// override this function in derived classes and populate m_Items.
    /// If data loaders are needed, they should be created here too.
    virtual void x_CreateProjectItems();

    /// overload this function to perform actions that require a scope
    /// the function is called before the item is added to a project
    virtual void x_PrepareProjectItem(
        objects::CProjectItem& item, objects::CScope& scope
    );

    virtual void x_AddToExistingProject(
        TProjectId id, const string& folder_name
    );
    virtual void x_CreateOneProject(
        const string& prj_text, const string& folder_name
    );
    virtual void x_CreateSeparateProjects(
        const string& prj_text, const string& folder_name
    );

    /// finds a folder with the given name or creates a new one; opens the folder
    virtual CRef<CProjectFolderPrx> x_GetFolder(
        CProjectPrx& project, const string& folder_name
    );

    void x_SetTextResult( const string& result );

protected:
    CIRef<CProjectService> m_Service;
    string m_TextResult;

    /// The objects that have been added and attached to scope.
    /// We need them when prefetching sequence descriptions
    TConstScopedObjects    m_Objects;

    /// parameters below are applicable only to tasks adding items.
    CDataLoadingOptions    m_Options;
    vector<TItemRef>       m_Items;
    TLoaders               m_Loaders;
    CRef<objects::CScope>  m_Scope;
    bool                   m_ItemsAdded;
    bool                   m_LoadersAdded;
};

typedef CDataLoadingJob CDataLoadingAppJob;


///////////////////////////////////////////////////////////////////////////////
/// CDataLoadingTask - a task that executes CDataLoadingJob.
class NCBI_GUICORE_EXPORT  CDataLoadingTask : public CAppJobTask
{
public:
    CDataLoadingTask( CDataLoadingJob& job, const string& descr = kEmptyStr );

    /// @name IAppTask interface implementation
    /// @{

    /// we override this function because our logic is a bit more complex than
    /// one implemented in CAppJobTask::Run() - under some circumstances
    /// we may need to show a dialog and then relaunch the job.
    virtual ETaskState Run();
    /// @}
};

typedef CDataLoadingTask CDataLoadingAppTask;


END_NCBI_SCOPE


#endif
// GUI_CORE_DEMO___LOADING_APP_JOB__HPP
