#ifndef GUI_CORE___PROJECT__HPP
#define GUI_CORE___PROJECT__HPP

/*  $Id: project.hpp 25589 2012-04-11 19:11:04Z katargir $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Andrey Yazhuk
 *
 * File Description:
 *
 */

#include <corelib/ncbistl.hpp>

#include <gui/gui_export.h>

#include <gui/core/project_folder.hpp>

#include <gui/core/document.hpp>
#include <gui/core/project_view.hpp>

#include <objmgr/scope.hpp>

#include <gui/objects/GBProjectHandle.hpp>
#include <gui/objects/WorkspaceFolder.hpp>
#include <gui/objects/LoaderDescriptor.hpp>

#include <corelib/ncbithr.hpp>


BEGIN_NCBI_SCOPE

class CProjectService;
class CWorkspaceConstPrx;
class CProjectPrx;
class CProjectFolderPrx;
class IDataLoaderProvider;
class CAppJob;

BEGIN_SCOPE(objects);
    class CGBProjectHandle;
END_SCOPE(objects);


////////////////////////////////////////////////////////////////////////////////
/// CProjectConstPrx
class NCBI_GUICORE_EXPORT CProjectConstPrx : public CObject
{
    friend class CWorkspaceConstPrx;
    friend class CWorkspacePrx;
    friend class CWorkspaceFolderConstPrx;
    friend class CWorkspaceFolderPrx;
    friend class CProjectFolderPrx;
    friend class CProjectItemPrx;
    friend class CProjectService;
public:
    typedef objects::CGBProjectHandle::TId TId;
    typedef vector< CIRef<IProjectView> >  TViews;
    typedef objects::IGBProject::TDataLoaders       TDataLoaders;

public:
    virtual ~CProjectConstPrx();

    TId GetId() const;

    /// returns true if the project is attached to a Workspace
    bool    IsAttached() const;

    bool    IsLoaded() const;
    bool    IsLoading() const;

    /// returns true of the project has unsaved changes
    bool    IsDirty() const;

    string  GetName() const;
    string  GetDescr() const;
    CTime   GetCreateTime() const;
    CTime   GetModifiedTime() const;

    wxString  GetFilename() const;

    objects::CScope*    GetScope();

    CRef<CProjectFolderConstPrx> GetDataFolder_C();

    CRef<CProjectItemConstPrx>
                        FindProjectItemById_C(CProjectItemConstPrx::TId id);
    CRef<CProjectFolderConstPrx>
                        FindProjectFolderById_C(CProjectFolderPrx::TId id);

    const TDataLoaders& GetDataLoaders() const;

    const TViews& GetViews(void) const;

    CRef<CGBDocument> GetDocument();

    /// Get view-specific settings.
    /// @param view_type is the view identification
    /// @param view_ins_key is the view instance identification. It is up
    /// to view to generate view-instance-id.
    string GetViewSettings(const string& view_type,
        const string& view_ins_id) const;

    /// Save view-specific settings.
    void SaveViewSettings(const string& view_type,
        const string& view_ins_id, const string& settings);

protected:
    CProjectConstPrx(CGBDocument& project,
                     CWorkspaceConstPrx* workspace,
                     bool get_lock = true);

    CRef<CGBDocument> x_GetDocument();

    void    x_ValidateCall() const;

private:
    // Prohibit copy constructor and assignment operator
    CProjectConstPrx(const CProjectConstPrx& value);
    CProjectConstPrx& operator=(const CProjectConstPrx& value);

protected:
    CRef<CGBDocument>  m_Document;

    /// reference to Workspace, if NULL - project is not attached
    CWorkspaceConstPrx*  m_WorkspaceConst; // TODO use weak ref
    bool m_HasLock;

    /// ID of the thread on which the proxy was created
    CThread::TID    m_OrigThreadID;
};


////////////////////////////////////////////////////////////////////////////////
/// CProjectPrx
class NCBI_GUICORE_EXPORT CProjectPrx : public CProjectConstPrx
{
    friend class CWorkspaceConstPrx;
    friend class CWorkspacePrx;
    friend class CWorkspaceFolderConstPrx;
    friend class CWorkspaceFolderPrx;
    friend class CProjectFolderPrx;
    friend class CProjectItemPrx;

public:
    virtual ~CProjectPrx();

    void    SetName(const string& name);
    void    SetDescr(const string& descr);
    void    SetCreateTime(const CTime& time);
    void    SetModifiedTime(const CTime& time);
    void    SetFilename(const wxString& filename);

    CRef<CProjectFolderPrx> GetDataFolder();

    CRef<CProjectItemPrx>   FindProjectItemById(CProjectItemPrx::TId id);
    CRef<CProjectFolderPrx> FindProjectFolderById(CProjectFolderPrx::TId id);
    CRef<objects::CLoaderDescriptor> FindDataLoaderByLabel(const string& label);

    /// move a folder existing in the project into a new folder
    void    MoveProjectFolder(CProjectFolderPrx& folder, CProjectFolderPrx& new_parent);

    void    MoveProjectItem(CProjectItemPrx& item, CProjectFolderPrx& new_parent);

    void    Save();

    bool    IsDataLoaderAdded(const objects::CLoaderDescriptor& loader); 
    void    AddDataLoader(objects::CLoaderDescriptor& loader);
    bool    RemoveDataLoader(objects::CLoaderDescriptor& loader);
    TDataLoaders& GetDataLoaders();
    bool    EnableDisableDataLoader(const string& label);

    void    AttachView(IProjectView* view);
    void    DetachView(IProjectView* view);
    TViews&       GetViews(void);

    /// get all added objects (items) in this projects.
    void GetScopedObjects(TConstScopedObjects& objs);

    void    CancelLoading();
    void    OnDocumentLoaded();

protected:
    CProjectPrx(CGBDocument& project, CWorkspaceConstPrx* workspace);

    bool    x_LoadData();
    void    x_UnloadData();

    void    x_AttachProjectItem(objects::CProjectItem& child_item);
    void    x_DetachProjectItem(objects::CProjectItem& child_item);

    void    x_AttachProjectItems();
    void    x_DetachProjectItems();

    IDataLoaderProvider* x_GetDataSourceByType(const string& type);
    bool    x_AttachDataLoader(const objects::CLoaderDescriptor& loader);
    bool    x_DetachDataLoader(const objects::CLoaderDescriptor& loader);

    void    x_AttachDataLoaders();
    void    x_DetachDataLoaders();

    void    x_FixProjectTitle();
    string  x_MakeUniqueItemName(const string& label) const;

    // update handler - set Dirty flags and notify Project Service
    void x_OnProjectChanged( 
        bool dirty, CProjectViewEvent::EEventSubtype sybtype = CProjectViewEvent::eNone 
    );
    void x_OnFolderChanged(
        CProjectFolderPrx& folder, CProjectViewEvent::EEventSubtype sybtype = CProjectViewEvent::eNone 
    );
    void x_OnItemChanged(
        CProjectItemPrx& item, CProjectViewEvent::EEventSubtype sybtype = CProjectViewEvent::eNone 
    );

private:
    // Prohibit copy constructor and assignment operator
    CProjectPrx(const CProjectPrx& value);
    CProjectPrx& operator=(const CProjectPrx& value);
};


END_NCBI_SCOPE

#endif  /// GUI_CORE___PROJECT__HPP
