#ifndef GUI_CORE___PROJECT_SERVICE__HPP
#define GUI_CORE___PROJECT_SERVICE__HPP

/*  $Id: project_service.hpp 27556 2013-03-05 16:04:52Z katargir $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Mike DiCuccio, Andrey Yazhuk
 *
 * File Description:
 *
 */

#include <corelib/ncbimtx.hpp>
#include <corelib/ncbistl.hpp>

#include <gui/core/project_service_types.hpp>
#include <gui/core/workspace.hpp>

#include <gui/core/dh_manager.hpp>

#include <gui/framework/service.hpp>
#include <gui/framework/observable.hpp>
#include <gui/framework/update_lock.hpp>

#include <gui/objutils/objects.hpp>
#include <gui/objutils/reg_settings.hpp>

#include <gui/utils/mru_list.hpp>
#include <gui/utils/command.hpp>

#include <gui/objects/GBProjectHandle.hpp>

#include <set>

#include <wx/string.h>

BEGIN_NCBI_SCOPE

class IViewManagerService;
class IProjectView;
class CGBDocument;

BEGIN_SCOPE(objects)
    class CGBWorkspace;
    class CProjectItem;
    class CPluginMRUEntry;
    class CWorkspaceFolder;
    class CUser_object;
END_SCOPE(objects)

////////////////////////////////////////////////////////////////////////////////
/// CProjectService - a service providing API for operations with Workspaces and
/// Projects.
/// CProjectService
///
/// CProjectService shall not have any dependencies on GUI.

class NCBI_GUICORE_EXPORT CProjectService :
    public CObjectEx,
    public IService,
    public IServiceLocatorConsumer,
    public CObservable,
    public IRegSettings
{
    friend class CProjectPrx;
    friend class CProjectFolderPrx;
    friend class CProjectItemPrx;
    friend class CWorkspaceConstPrx;
    friend class CWorkspacePrx;
public:
    typedef CTimeMRUList<wxString>    TMRUPathList;

public:
    CProjectService();
    virtual ~CProjectService();

    /// @name IService interface implementation
    /// @{
    virtual void    InitService();
    virtual void    ShutDownService();
    /// @}

    /// @name IServiceLocatorConsumer interface implementation
    /// @{
    virtual void    SetServiceLocator(IServiceLocator* locator);
    /// @}

    /// @name IRegSettings interface implementation
    /// @{
    virtual void    SetRegistryPath(const string& path);
    virtual void    SaveSettings() const;
    virtual void    LoadSettings();
    /// @}

    /// @name Workspace manipulators
    /// @{
    virtual bool    HasWorkspace();

    virtual CRef<CWorkspaceConstPrx>    GetWorkspaceConst();
    virtual CRef<CWorkspacePrx>         GetWorkspace();

    virtual void    CreateNewWorkspace();

    /// Loads a workspace from a file and optionally set as current
    virtual bool    LoadWorkspace(const wxString& filename);

    /// release the current workspace and all its contents TODO - review
    virtual void    ReleaseWorkspace();

    /// Save a workspace.  This will (potentially) ask the user for a
    /// filename to which to save the workspace
    virtual bool    SaveWorkspace();

    /// get MRU Projects and Workspaces
    virtual const TMRUPathList& GetProjectWorkspaceMRUList() const;

    virtual void    AddToProjectWorkspaceMRUList(const wxString& path);
    /// @}

    /// @name Project View Manipulators
    /// @{
    /// initialized view, connects it to a projects and adds to View Manager Service
    CIRef<IProjectView> AddProjectView(const string& view_name, SConstScopedObject& object, const objects::CUser_object* params);
    CIRef<IProjectView> AddProjectView(const string& view_name, TConstScopedObjects& objects, const objects::CUser_object* params);

    /// removes the view from View manager Service and disconnects it from the project
    void    RemoveProjectView(IProjectView& view);
    void    RemoveAllProjectViews(CProjectPrx& project, bool reset_hist_async = true);
    void    RemoveAllProjectViews();

    /// called by CDocument when IProjectView is attached/detached to it
    void    OnViewAttached(IProjectView* view);
    void    OnViewDetached(IProjectView* view);

    /// "normal" view opening recording
    void    RecordViewOpen(IProjectView* view);

    /// "normal" view closing recording
    void    RecordViewClose(IProjectView* view);

    /// launch a task for prefetching sequence descriptions for newly added items.
    void    PrefetchSeqDescr(const TConstScopedObjects& objs);

    void    GetObjProjects(TConstScopedObjects& objects, vector<objects::CGBProjectHandle::TId>& ids);

protected:
    CRef<objects::CGBWorkspace> x_LoadWorkspace(const wxString& filename);

    void x_CreateDocuments(objects::CWorkspaceFolder& folder);

    void x_ConnectDocument(CGBDocument& project);
    void x_DisconnectDocument(CProjectPrx& project, bool reset_hist_async = true);

    virtual IObserverUpdate* x_CreateStateChange();
    //virtual void x_PostUpdateEvent();

    /// marks workspace as "dirty" and updates observers
    virtual void    x_OnWorkspaceChanged(); //TODO - retire
    virtual void    x_OnWorkspaceChanged(CWorkspaceConstPrx::TId id);
    virtual void    x_OnProjectChanged(CProjectPrx::TId id);

    bool x_CanBeSavedAsLink(const CObject& obj, const CGBDocument& project);
    void x_OpenDefaultView(objects::CProjectItem& item, objects::CScope& scope);
    void x_RemoveView(IViewManagerService& view_srv, IProjectView& view, bool reset_hist_async = true);
    void x_RestoreViews(objects::CScope& scope, CGBDocument& project);

private:
    /// all constructing / copying of project managers is forbidden
    CProjectService(const CProjectService&);
    CProjectService& operator= (const CProjectService&);

    CRWLock&    x_GetWorkspaceLock();
    CRWLock&    x_GetProjectsLock();

protected:
    typedef set<CWorkspaceConstPrx::TId>    TWorkspaceIds;
    typedef set<CProjectPrx::TId>      TProjectIds;

protected:
    // CObservable base class already provides support for locking, so we do not
    // need to define any mutexes here.

    /// path in the Registry to our settings
    string  m_RegPath;

    IServiceLocator* m_ServiceLocator;

    /// the only Workspace
    CRef<objects::CGBWorkspace> m_Workspace;

    string m_Filename;  /// filename for the workspace


    CEventHandler   m_Disp; /// event handler associated with the service
    CMutex  m_DispMutex;    /// Mutex guarding m_Disp

    /// Workspace RW Lock
    CRWLock m_WorkspaceLock;

    /// RW Lock for all projects in the Workspace; we use a single lock to avoid
    /// deadlocks
    CRWLock m_ProjectsLock;

    bool m_LockViewRecord; /// lock recording of opening views

    TMRUPathList    m_ProjectWorkspaceMRUList;

    TWorkspaceIds   m_ModifiedWorkspaces;
    TProjectIds     m_ModifiedProjects;
};


END_NCBI_SCOPE

#endif  /// GUI_CORE___PROJECT_SERVICE__HPP
