#ifndef GUI_CORE___PROJECT_TASK__HPP
#define GUI_CORE___PROJECT_TASK__HPP

/*  $Id: project_task.hpp 22688 2010-12-13 22:02:43Z katargir $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Andrey Yazhuk
 *
 * File Description:
 *
 */

#include <corelib/ncbistl.hpp>
#include <corelib/ncbiobj.hpp>

#include <gui/gui_export.h>

#include <gui/core/project.hpp>

#include <gui/framework/app_task_impl.hpp>

#include <wx/string.h>

BEGIN_NCBI_SCOPE

BEGIN_SCOPE(objects);
    class CGBProjectHandle;
END_SCOPE(objects);


class IServiceLocator;
class CProjectService;

///////////////////////////////////////////////////////////////////////////////
/// CProjectTask -  a task that supports typical application operations
/// applicable to Workspaces and Projects.
/// The task relies on Project Service functions as low-level API.

class  NCBI_GUICORE_EXPORT  CProjectTask :
    public CAppTask
{
public:
    typedef CProjectPrx::TId    TProjectId;
    typedef vector<TProjectId>  TProjectIdVector;

public:
    CProjectTask(IServiceLocator* srv_locator);
    virtual ~CProjectTask();

    /// call one of the Init_ function to select the command that shall be
    /// performed by the task

    virtual void    Init_NewWorkspace();
    virtual void    Init_CloseWorkspace();
    virtual void    Init_OpenWorkspace(const wxString& filename);
    virtual void    Init_OpenProjects(vector<wxString>& filenames);
    virtual void    Init_Save(bool workspace, TProjectIdVector sel_projects);
    virtual void    Init_SaveAs(bool workspace, TProjectIdVector sel_projects);
    virtual void    Init_LoadProjects(TProjectIdVector projects);
    virtual void    Init_UnLoadProjects(TProjectIdVector projects);
    virtual void    Init_RemoveProjects(TProjectIdVector projects);

    static bool     Enable_Save(IServiceLocator* srv_locator);
    static bool     Enable_SaveAs(IServiceLocator* srv_locator);
    static bool     Enable_LoadUnloadProjects(IServiceLocator* srv_locator,
                                              TProjectIdVector projects, bool load);
protected:
    /// Project Task can be set up to execute one of the following commands
    enum    ECommand    {
        eInvalid = -1,
        eNewWorkspace,
        eCloseWorkspace,
        eOpenWorkspace,
        eOpenProjects,
        eSave,
        eSaveAs,
        eLoadProjects,
        eUnLoadProjects,
        eRemoveProjects
    };

    typedef bool (CProjectTask::*TProjectFunc)(CProjectPrx& project);

protected:
    virtual ETaskState    x_Run(); /// overriding CAppTask::x_Run()

    /// Top-level worker functions corresponding to commands
    /// @{
    virtual void    x_NewWorkspace();
    virtual void    x_OpenWorkspace();
    virtual void    x_OpenProjects();
    virtual bool    x_CloseWorkspace();
    virtual void    x_Save();
    virtual void    x_RemoveProjects(TProjectIdVector project_ids);
    virtual void    x_LoadProjects(TProjectIdVector project_ids);
    virtual void    x_UnLoadProjects(TProjectIdVector project_ids);
    /// @}

    /// @name Helper functions
    /// @{
    virtual void    x_DoOpenWorkspace(const wxString& filename);
    virtual bool    x_DoCloseWorkspace();
    virtual bool    x_DoSaveWorkspaceOnly(bool save_as);

    virtual bool    x_DoSaveUserSelected(const string& title,
                                         const string& description,
                                         bool save_as,
                                         bool en_workspace,
                                         bool sel_workspace,
                                         bool no_btn,
                                         TProjectIdVector& project_ids,
                                         const vector<int>& selected,
                                         TProjectFunc func);

    virtual bool    x_DoEnsureProjectSaved(CProjectPrx& project);
    virtual bool    x_DoSaveProjectAs(CProjectPrx& project);

    virtual wxString x_DoChooseNewWorkspaceName();

    static bool x_ProjectNotSaved(CWorkspaceConstPrx& workspace, CProjectPrx& project);

    virtual void    x_GetUnsavedProjects(CWorkspaceConstPrx& workspace,
                                         const TProjectIdVector& projects,
                                         TProjectIdVector& not_saved);

    void x_LaunchProjectLoadingTask(const TProjectIdVector& ids,
                                    const string& task_name,
                                    bool add_to_mru);
    /// @}

protected:
    IServiceLocator*  m_SrvLocator;
    CProjectService*  m_Service;

    ECommand    m_Command;
    wxString  m_WorkspaceFilename;
    vector<wxString>    m_ProjectFilenames;
    TProjectIdVector  m_ProjectIds;
    bool    m_SelectWorkspace;
};


END_NCBI_SCOPE


#endif  // GUI_CORE___PROJECT_TASK__HPP

