#ifndef GUI_CORE___PROJECT_TREE_PANEL__HPP
#define GUI_CORE___PROJECT_TREE_PANEL__HPP

/*  $Id: project_tree_panel.hpp 26121 2012-07-23 19:35:02Z katargir $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Andrey Yazhuk
 *
 * File Description:
 *
 *      Project Tree Panel class.
 */

#include <corelib/ncbistd.hpp>

#include <gui/gui_export.h>

#include <gui/core/app_explorer_item.hpp>
#include <gui/core/project.hpp>
#include <gui/core/app_explorer_service.hpp>

#include <gui/utils/event_handler.hpp>

#include <wx/panel.h>
#include <wx/treectrl.h>
#include <wx/dnd.h>


BEGIN_NCBI_SCOPE
BEGIN_SCOPE(objects)
    class CFolderInfo;
END_SCOPE(objects)

class CAppExplorerService;
class IViewManagerService;
class CAppExplorerDataObject;


//////////////////////////////////////////////////////////////////////////////
/// CProjectTreeCtrl - subclass of wxTreeCtrl to allow functions to be
///                    overridden for application-specific behavior.

class CProjectTreeCtrl: public wxTreeCtrl
{
    DECLARE_EVENT_TABLE()
public:
    CProjectTreeCtrl() : wxTreeCtrl() {}
    CProjectTreeCtrl(wxWindow *parent, 
                     wxWindowID id = wxID_ANY,
                     const wxPoint& pos = wxDefaultPosition,
                     const wxSize& size = wxDefaultSize,
                     long style = wxTR_DEFAULT_STYLE,
                     const wxValidator &validator = wxDefaultValidator,
                     const wxString& name = wxTreeCtrlNameStr) 
    : wxTreeCtrl(parent, id, pos, size, style, validator, name) {}

    virtual ~CProjectTreeCtrl() {};

    virtual wxTextCtrl *EditLabel(const wxTreeItemId& item,
                                  wxClassInfo* textCtrlClass = CLASSINFO(wxTextCtrl));

    void    OnLeftDown(wxMouseEvent& event);
    void    OnKeyUp(wxKeyEvent& event);
    void    OnMove(wxMouseEvent& event);
};

///////////////////////////////////////////////////////////////////////////////
/// CProjectTreePanel - a window that represents Project View.

class NCBI_GUICORE_EXPORT CProjectTreePanel :
    public wxPanel,
    public CEventHandler
{
    DECLARE_EVENT_TABLE()
public:
    friend class CProjectTreeViewDropTarget;

    typedef CRef<CExplorerItem> TItemRef;
    typedef vector< CRef<CExplorerItem> >  TItemRefVector;
    typedef vector<CExplorerItem*>  TItemVector;
    typedef CProjectPrx::TId TProjectId;
    typedef vector<TProjectId> TProjectIdVector;

public:
    CProjectTreePanel();
    virtual ~CProjectTreePanel();

    /// Initializes member variables
    void    Init();

    /// Creates the controls and sizers
    void    CreateControls();

    virtual void    Create(wxWindow* parent,
                           wxWindowID id = wxID_ANY,
                           const wxPoint& pos = wxDefaultPosition,
                           const wxSize& size = wxDefaultSize);

    virtual void    SetWorkbench(IWorkbench* workbench);
    virtual void    SetService(CAppExplorerService* service);
    virtual void    RefreshItems();

    virtual void    GetSelectedItems(TItemRefVector& items);
    virtual bool    IsWorkspaceSelected();
    virtual void    GetSelectedProjectIds(TProjectIdVector& ids);

    /// @name Event handlers
    /// @{
    void    OnContextMenu(wxContextMenuEvent& event);

    void    OnItemExpandedCollapsed(wxTreeEvent& event);
    void    OnItemActivated(wxTreeEvent& event);
    void    OnSelectionChanged(wxTreeEvent& event);
    void    OnBeginDrag(wxTreeEvent& event);
    void    OnEndDrag(wxTreeEvent& event);
    void    OnTreeKeyDown(wxTreeEvent& event);
    void    OnBeginLabelEdit(wxTreeEvent& event);
    void    OnEndLabelEdit(wxTreeEvent& event);

    void    OnLeftDown(wxMouseEvent& event);

    void    OnCut(wxCommandEvent& event);
    /// @}

    /// @name Command handlers
    /// @{
    /// @}

    /// @name D&D support - interface for Drop Target
    /// @{
    wxDragResult OnDragOver(wxCoord x, wxCoord y, wxDragResult def);
    wxDragResult OnDrop(wxCoord x, wxCoord y, wxDragResult def, CAppExplorerDataObject& data);
    /// @}

	/// When the data object is non-null, a D&D is in progress
    CAppExplorerDataObject* GetDataObject() { return m_DataObject; }

    /// Drag and drop support for OSX Cocoa (generic dnd path not working in OSX cocoa 2.9.3
    /// so we use dnd specific to the tree in this case). 
    void SetCopyMode(bool b) { m_CopyMode = b; }

protected:
    void    x_AddChildItems(wxTreeItemId item_id, const CExplorerItem& item);

    void    x_SetSelections(wxArrayTreeItemIds& ids);
    void    x_GetItemsFromIds(const wxArrayTreeItemIds& ids, TItemRefVector& items);

    CExplorerItem*  x_GetExplorerItem(const wxTreeItemId& id);
    CExplorerItem*  x_GetSingleSelectedItem();

    bool    x_AllViewsItemSelected(TItemRefVector& items);
    bool    x_AtLeastOneProjectLoaded();

    void    x_LoadUnloadProject(bool load);

    void    x_CutOrCopy(bool cut);
    bool    x_CanUpdateCutOrCopy();

protected:
    /// Data object used for drag and drop (this is the data being dragged)
    CAppExplorerDataObject*  m_DataObject;
    string m_RegPath;

    IWorkbench*     m_Workbench;
    CAppExplorerService*    m_ExplorerService;

    CProjectTreeCtrl* m_Tree;
    wxColour    m_DisabledColor;

    vector<wxEvtHandler*> m_ContributedHandlers;

    /// D&D state
    wxTreeItemId    m_DropItemId; // current D&D target
    CStopWatch      m_DropItemTimer;
    bool    m_DropTimerActive;

    // D&D copy (vs move) state for D&D implemented with tree-functions (osx cocoa)
    bool m_CopyMode; 
};


///////////////////////////////////////////////////////////////////////////////
/// CProjectTreeViewDropTarget - a drop target associated with CProjectTreeView.
class CProjectTreeViewDropTarget : public wxDropTarget
{
public:
    CProjectTreeViewDropTarget(CProjectTreePanel* panel);

    /// @name wxDropTarget virtual functions
    /// @{
    virtual wxDragResult OnEnter(wxCoord x, wxCoord y, wxDragResult def);
    virtual wxDragResult OnDragOver(wxCoord x, wxCoord y, wxDragResult def);
    virtual wxDragResult OnData(wxCoord x, wxCoord y, wxDragResult def);
    /// @}

protected:
    CProjectTreePanel*  m_Panel;
    bool    m_AcceptableData;
};


END_NCBI_SCOPE;

#endif // GUI_CORE___PROJECT_TREE_PANEL__HPP
