#ifndef GUI_CORE___SELECTION_SERVICE__HPP
#define GUI_CORE___SELECTION_SERVICE__HPP

/*  $Id: selection_service.hpp 20149 2009-09-28 23:53:03Z voronov $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Andrey Yazhuk
 *
 * File Description:
 *
 */

#include <corelib/ncbiobj.hpp>
#include <gui/gui_export.h>

#include <gui/objutils/objects.hpp>
#include <gui/objutils/obj_event.hpp>


BEGIN_NCBI_SCOPE

class ISelectionClient;
class CSelectionEvent;

///////////////////////////////////////////////////////////////////////////////
/// ISelectionService - an interface
class ISelectionService
{
public:
    virtual ~ISelectionService() {}

    virtual void AttachClient(ISelectionClient* client) = 0;
    virtual void DetachClient(ISelectionClient* client) = 0;
    virtual bool IsAttached(ISelectionClient* client) const = 0;

    virtual ISelectionClient* GetActiveClient() = 0;

    /// returns objects currently selected in the application, this function
    /// should be used for obtaining arguments for running tools, opening views and
    /// performing other application commands
    virtual void GetCurrentSelection(TConstScopedObjects& sel) = 0;

    /// request broadcasting of the given event, Broadcast() notifies observers
    /// about changes in selection (do not need to call OnSelectionChanged())
    virtual void Broadcast( CSelectionEvent& evt, ISelectionClient* source, bool force ) = 0;

    virtual void Broadcast( ISelectionClient* source, bool force, bool ranges = false ) = 0;

    /// this function is called by GUI components to notify server
    virtual void OnActiveClientChanged( ISelectionClient* active ) = 0;

    /// notifies Service and its observers about changes in selection
    virtual void OnSelectionChanged( ISelectionClient* source ) = 0;
};


///////////////////////////////////////////////////////////////////////////////
/// ISelectionClient - represents an object that support a notion of selection.
/// ISelectionClient provides functions for setiing and getting selections
class ISelectionClient
{
public:
    virtual ~ISelectionClient() {}

    /// shall return data objects representing selected things, return one
    /// data object for every selected thing. Implementation shall not clear
    /// the container as it can be used for accumulating objects.
    virtual void GetSelection( TConstScopedObjects& objects ) const = 0;

    virtual void SetSelectionService( ISelectionService* service ) = 0;

    /// returns a string identifying the client in UI
    virtual string GetSelClientLabel() = 0;

    /// get selection for broadcasting
    virtual void GetSelection( CSelectionEvent& evt ) const = 0;

    /// set broadcasted selection
    virtual void SetSelection( CSelectionEvent& evt ) = 0;

    /// Adds the main data objects represented by the client to "objects".
    /// Implementation shall not clear the container as it can be used for
    /// accumulating objects.
    virtual void GetMainObject( TConstScopedObjects& objects ) const = 0;

    /// if all selected elements belong to the same scope - return it,
    /// otherwise return NULL
    virtual CRef<objects::CScope> GetScope() const = 0;
};

/*
///////////////////////////////////////////////////////////////////////////////
/// CSelectionEvent
/// CSelectionEvent is used for broadcasting selection between views.
/// It containes separate "packages" for different types of selection.
///  If package
/// This class is not thread-safe.
class NCBI_GUIOBJUTILS_EXPORT CSelectionEvent : public CViewEvent
{
public:
    typedef CRangeCollection<TSeqPos>   TRangeColl;
    typedef TRangeColl::TRange          TRange;
    typedef vector< CConstRef<objects::CSeq_feat> >     TFeats;
    typedef vector< CConstRef<objects::CSeq_align> >    TAligns;
    typedef vector< CConstRef<objects::CSeq_loc> >      TSeqLocs;
    typedef vector< CConstRef<objects::CSeq_id> >       TIds;
    typedef vector< CConstRef<CIdLoc> >    TIdLocs;

    CSelectionEvent(objects::CScope& scope, CEventHandler* sender);

    enum    EObjMatchPolicy {
        eAllIds,
        eAtLeastOne
    };
    enum    EIdMatchPolicy {
        eExact,
        eAccOnly
    };

    objects::CScope& GetScope()  {   return  *m_Scope;    }

    /// Range Selection - represents a collection of selected segments
    bool    HasRangeSelection() const;
    const objects::CHandleRangeMap&  GetRangeSelection() const;
    void    AddRangeSelection(const objects::CSeq_id& id, const TRangeColl& segs);
    void    AddRangeSelection(const objects::CSeq_loc& loc);
    bool    GetRangeSelection(const objects::CSeq_id& id, objects::CScope& scope,
        TRangeColl& segs);
    void    GetRangeSelectionAsLocs(TConstObjects& objs) const;

    /// Object Selection - represents select Objects
    /// (such as features, alignments, etc)
    bool    HasObjectSelection();
    bool    AddObjectSelection(const CObject& obj);
    void    AddObjectSelection(const TConstObjects& objs); // temporary
    void    AddObjectSelection(const objects::CSeq_id& id);
    void    AddObjectSelection(const objects::CSeq_feat& feat);
    void    AddObjectSelection(const objects::CSeq_align& align);
    void    AddObjectSelection(const objects::CSeq_loc& loc);
    void    AddObjectSelection(const objects::CSeq_id& id, const CRange<TSeqPos>& range);

    const TFeats&   GetFeats()  const   {   return m_Feats;    }
    const TAligns&  GetAligns() const   {   return m_Aligns;    }
    const TSeqLocs& GetSeqLocs() const   {   return m_SeqLocs;    }
    const TIds&     GetIds() const   {   return m_Ids;    }
    const TIdLocs&     GetIdLocs() const   {   return m_IdLocs;    }

    void            GetAllObjects(TConstObjects& objs) const;

    /// @name Matching functons
    /// @{
    static bool Match(const CObject& obj1, objects::CScope& scope1,
        const CObject& obj2, objects::CScope& scope2);

    static bool MatchFeatWithFeat(const objects::CSeq_feat& feat1,
        objects::CScope& scope1,
        const objects::CSeq_feat& feat2,
        objects::CScope& scope2);

    static bool MatchFeatWithLoc(const objects::CSeq_feat& feat1,
        objects::CScope& scope1,
        const objects::CSeq_loc& loc2,
        objects::CScope& scope2);

    static bool MatchFeatWithId(const objects::CSeq_feat& feat1,
        objects::CScope& scope1,
        const objects::CSeq_id& id2,
        objects::CScope& scope2);

    static bool MatchAlignWithAlign(const objects::CSeq_align& align1,
        objects::CScope& scope1,
        const objects::CSeq_align& align2,
        objects::CScope& scope2);

    static bool MatchLocWithLoc(const objects::CSeq_loc& loc1,
        objects::CScope& scope1,
        const objects::CSeq_loc& loc2,
        objects::CScope& scope2);

    static bool MatchLocWithId(const objects::CSeq_loc& loc1,
        objects::CScope& scope1,
        const objects::CSeq_id& id2,
        objects::CScope& scope2);

    static bool MatchIdWithId(const objects::CSeq_id& id1,
        objects::CScope& scope1,
        const objects::CSeq_id& id2,
        objects::CScope& scope2);
    /// @}
protected:
    /// Matches two ids using current policy; does not consider synonyms
    static bool x_SimpleMatch(const objects::CSeq_id_Handle& h_sel_id,
        const objects::CSeq_id_Handle& h_id);
public:
    static  EObjMatchPolicy     sm_ObjMatchPolicy;
    static  EIdMatchPolicy      sm_IdMatchPolicy;
    static  bool    sm_MatchAlnLocs;
    static  bool    sm_MatchByProduct;
    static  bool    sm_AutoBroadcast;
    static  bool    sm_InterDocBroadcast; /// broadcast sel. between documents

protected:
    enum    EObjType {
        eSeq_feat,
        eSeq_align,
        eSeq_loc,
        eSeq_id
    };
    typedef map<string, EObjType> TRawNameToType;
    static  TRawNameToType  sm_RawNameToType;

    CRef<objects::CScope> m_Scope; // scope corresponding to the source of selection

    // Range Selection package
    bool    m_HasRangeSelection;
    objects::CHandleRangeMap m_RangeMap; // storage for Range Selection

    // Object Selection Package
    bool    m_HasObjectSelection;
    TFeats      m_Feats;
    TAligns     m_Aligns;
    TSeqLocs    m_SeqLocs;
    TIds        m_Ids;
    TIdLocs     m_IdLocs;
};
*/

END_NCBI_SCOPE


#endif  // GUI_CORE___SELECTION_SERVICE__HPP

