#ifndef GUI_CORE___SELECTION_SERVICE_IMPL__HPP
#define GUI_CORE___SELECTION_SERVICE_IMPL__HPP

/*  $Id: selection_service_impl.hpp 27377 2013-02-04 19:53:05Z voronov $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Andrey Yazhuk
 *
 * File Description:
 *
 */

#include <corelib/ncbistd.hpp>

#include <gui/gui_export.h>

#include <gui/core/selection_service.hpp>

#include <gui/framework/service.hpp>
#include <gui/framework/observable.hpp>

#include <gui/objutils/reg_settings.hpp>

#include <wx/panel.h>
#include <wx/sizer.h>
#include <wx/stattext.h>

BEGIN_NCBI_SCOPE


class CSelectionService;
class IView;
class IGBDocument;
class CProjectPrx;

///////////////////////////////////////////////////////////////////////////////
/// CSelectionService
class NCBI_GUICORE_EXPORT CSelectionService
:   public CObjectEx,
    public IService,
    public IServiceLocatorConsumer,
    public ISelectionService,
    public CObservable,
    public IRegSettings
{

public:
    typedef vector<ISelectionClient*> TClients;

    CSelectionService();
    virtual ~CSelectionService();

    /// @name IService interface implementation
    /// @{
    virtual void    InitService();
    virtual void    ShutDownService();
    /// @}

    virtual void GetClients(TClients& clients);
    virtual void DetachAllClients();

    /// @name ISelectionService implemenation
    /// @{
    virtual void AttachClient(ISelectionClient* client);
    virtual void DetachClient(ISelectionClient* client);
    virtual bool IsAttached(ISelectionClient* client) const;

    virtual ISelectionClient* GetActiveClient();
    virtual void GetCurrentSelection(TConstScopedObjects& sel);

    virtual void Broadcast( CSelectionEvent& evt, ISelectionClient* source, bool force );
    virtual void Broadcast( ISelectionClient* source, bool force, bool ranges = false );

    virtual void OnActiveClientChanged(ISelectionClient* active);
    virtual void OnSelectionChanged(ISelectionClient* source);
    /// @}

    /// @name IRegSettings implementation
    /// @{
    virtual void SetRegistryPath(const string& path);
    virtual void LoadSettings();
    virtual void SaveSettings() const;
    /// @}

    /// @name IServiceLocatorConsumer interface implementation
    /// @{
    virtual void    SetServiceLocator(IServiceLocator* locator);
    /// @}

    bool IsAutoBroadcast() { return mf_AutoBroadcast; }
    void SetAutoBroadcast( bool pAuto );

protected:
    void x_Broadcast(
        CSelectionEvent& evt, ISelectionClient* source, CProjectPrx* target
    );

    virtual IObserverUpdate* x_CreateStateChange();

protected:
    TClients m_Clients;
    ISelectionClient* m_ActiveClient;

    /// this flag indicates that broadcast is in progress and so all requests
    /// to brodcast will be ignored
    bool m_Broadcasting;

    IServiceLocator* m_SrvLocator;
    string  m_RegPath;

private:
    CIRef<IObserverUpdate> m_ServiceUpdate;

    /// @name various selection service options
    /// @{
    bool mf_AutoBroadcast;
    /// @}
};

class NCBI_GUICORE_EXPORT CSelectionServiceUpdate
    : public CObserverUpdate
{
public:
    enum EClientUpdateType {
        eClientAttached,
        eClientDetached,
        eActiveChanged,
        eSelectionChanged,
        eMaxClientUpdateType
    };

    CSelectionServiceUpdate( EClientUpdateType type, ISelectionClient* client )
        : CObserverUpdate() 
        , m_ClientUpdateType( type )
    {
        m_UpdatedClientWeakPtr = client;
    }

    EClientUpdateType GetClientUpdateType() const { return m_ClientUpdateType; }

    /// Returns client that was updated. Caller should check
    /// whether this client is still valid via IsAttached()
    /// as it could be detached at the moment update reaches 
    /// this caller.
    ISelectionClient* GetClientUpdated() const 
    { 
        CIRef<ISelectionClient> wic(m_UpdatedClientWeakPtr.Lock());
        return wic.GetPointerOrNull();
    };

private:
    EClientUpdateType m_ClientUpdateType;
    CWeakIRef<ISelectionClient>  m_UpdatedClientWeakPtr;
};

///////////////////////////////////////////////////////////////////////////////
/// CAppTaskServiceSlot - a Status Bar slot that displays information about
/// tasks in Application Task Service


class NCBI_GUICORE_EXPORT CSelectionBroadcastSlot : public wxPanel
{
    DECLARE_EVENT_TABLE();
public:
    CSelectionBroadcastSlot( wxEvtHandler* cmd_handler );
    ~CSelectionBroadcastSlot()
    {
        m_ObserverProxy->ReleaseObserver();
    }

    bool Create( wxWindow* parent );

    void OnPaint(wxPaintEvent& event);
    void OnMouseEnter(wxMouseEvent& event);
    void OnMouseLeave(wxMouseEvent& event);
    void OnLeftDown(wxMouseEvent& event);

    IObserver* GetIObserver() { return m_ObserverProxy.GetPointer(); }
    /// implementing IObserver
    virtual void OnUpdate( IObserverUpdate& update );

protected:
    wxSize  x_GetTextAreaSize();
    string  x_StatusText( CSelectionEvent::EObjMatchPolicy fObj, CSelectionEvent::EIdMatchPolicy fId );
    string  x_StatusTooltip( CSelectionEvent::EObjMatchPolicy fObj, CSelectionEvent::EIdMatchPolicy fId );

    int     x_GetPendingTasksCount();

    void    x_Update();

protected:
    wxEvtHandler* m_CmdHandler; // handles commands from the slot

    wxBoxSizer* m_Sizer;
    wxStaticText* m_TextItem;

    CSelectionService* m_SelectionService;

    wxString m_Text; // current status text

    bool m_Hot; // hot track (mouse inside)
    bool m_Idle; // no running tasks

    class CObserverProxy 
        : public CObjectEx
        , public IObserver
    {
    public:
        CObserverProxy( CSelectionBroadcastSlot& observer ) : m_Observer( &observer ){}
        void ReleaseObserver() { m_Observer = NULL; }

        virtual void OnUpdate( IObserverUpdate& update ){ if (m_Observer) m_Observer->OnUpdate(update); }
    private:
        CSelectionBroadcastSlot* m_Observer;
    };

    CIRef<CObserverProxy> m_ObserverProxy;

};


END_NCBI_SCOPE

#endif  // GUI_CORE___SELECTION_SERVICE_HPP
