#ifndef GUI_CORE___SELECTION_VIEW__HPP
#define GUI_CORE___SELECTION_VIEW__HPP

/*  $Id: selection_view.hpp 24317 2011-09-12 20:00:09Z falkrb $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Andrey Yazhuk, Yury Voronov
 *
 * File Description:
 *
 */

#include <corelib/ncbistd.hpp>

#include <gui/gui_export.h>

#include <gui/framework/view.hpp>
#include <gui/framework/observable.hpp>
#include <gui/core/selection_service.hpp>

#include <gui/widgets/wx/wm_client.hpp>
#include <gui/widgets/object_list/object_list_widget.hpp>
#include <gui/widgets/text_widget/text_item_panel.hpp>

#include <gui/objutils/reg_settings.hpp>
#include <gui/utils/command.hpp>
#include <gui/utils/extension.hpp>

#include <corelib/ncbitime.hpp>

#include <wx/panel.h>
#include <wx/choice.h>
#include <wx/menu.h>

BEGIN_NCBI_SCOPE

class CSelectionService;
class CSelectionView;
class CSelectionServiceUpdate;

///////////////////////////////////////////////////////////////////////////////
/// CSelectionPanel
class NCBI_GUICORE_EXPORT CSelectionPanel :
    public wxPanel,
    public IRegSettings
{
    DECLARE_EVENT_TABLE()

public:
    static const int scm_ModeCmd = 13000;

    enum EDisplayMode    {
        eInvalidDisplayMode = -1,
        eTable,
        eBriefText,
        eText
    };

public:
    CSelectionPanel( CSelectionView* view );
    virtual ~CSelectionPanel();

    virtual void Create(
        wxWindow* parent,
        wxWindowID id = wxID_ANY,
        const wxPoint& pos = wxDefaultPosition,
        const wxSize& size = wxDefaultSize
    );
    virtual bool Destroy();

    /// Initialises member variables
    void Init();

    /// Creates the controls and sizers
    void CreateControls();

    void SetSelectionService( CSelectionService* service );
    void GetSelection( TConstScopedObjects& buf ) const;

    void Update();
    void OnSelServiceStateChange( CSelectionServiceUpdate& update );

    /// @name IRegSettings interface implementation
    /// @{
    virtual void SetRegistryPath(const string& reg_path);
    virtual void LoadSettings();
    virtual void SaveSettings() const;
    /// @}

    virtual void AppendMenuItems( wxMenu& aMenu );

    /// @name Command handlers
    /// @{
    void OnContextMenu(wxContextMenuEvent& event);
    void OnTargetChanged(wxCommandEvent& event);
    void OnModeChanged(wxCommandEvent& event);
    /// @}

protected:
    EDisplayMode    x_GetModeByCommand( TCmdID cmd );
    //ISelectionClient* x_GetSelectedClient();

    void x_InitCurrentWidget();

    bool x_ResetSelObjects();
    void x_ResetTargetChoice();
    void x_ResetObjectWidget();
    void x_InitTextWidget();

protected:
    typedef map<int, ISelectionClient*> TIndexToClient;

    static bool m_RegisteredIcons;

    CSelectionView*     m_SelView;
    CSelectionService*  m_Service;
    TConstScopedObjects m_SelObjects;

    wxChoice*  m_TargetChoice;

    TIndexToClient m_IndexToClient; // m_TargetChoice item index -> ISelectionClient*
    CWeakIRef<ISelectionClient>  m_CurrentClientWeakPtr;
    //ISelectionClient* m_CurrentClient;

    EDisplayMode m_DisplayMode;

    CObjectListWidget* m_ListWidget;
    CTextItemPanel*    m_TextWidget;

    string m_RegPath; // path to the setting in CGuiRegistry
};


///////////////////////////////////////////////////////////////////////////////
/// CSelectionView
class NCBI_GUICORE_EXPORT CSelectionView
:   public CObjectEx,
    public IView,
    public IWMClient,
    public IObserver,
    public IRegSettings,
    public ISelectionClient
{
    friend class CSelectionViewFactory;

public:
    CSelectionView();
    virtual ~CSelectionView();

    /// @name IView
    /// @{
    virtual const CViewTypeDescriptor&  GetTypeDescriptor() const;
    virtual void SetWorkbench( IWorkbench* workbench );
    virtual void CreateViewWindow( wxWindow* parent );
    virtual void DestroyViewWindow();
    /// @}

    /// @name IWMClient interface implementation
    /// @{
    virtual wxWindow* GetWindow();
    virtual wxEvtHandler* GetCommandHandler() { return GetWindow()->GetEventHandler(); }
    virtual string GetClientLabel(IWMClient::ELabel ltype = IWMClient::eDefault) const;
    virtual string GetIconAlias() const;
    virtual const CRgbaColor* GetColor() const;
    virtual void SetColor( const CRgbaColor& aColor );
    virtual const wxMenu* GetMenu();
    virtual void UpdateMenu( wxMenu& aRootMenu );
    virtual CFingerprint  GetFingerprint() const;
    /// @}

    /// @name IRegSettings interface implementation
    /// @{
    virtual void SetRegistryPath(const string& reg_path);
    virtual void LoadSettings();
    virtual void SaveSettings() const;
    /// @}

    /// @name IObserver interface implementation
    /// @{
    virtual void OnUpdate( IObserverUpdate& update );
    ///  @}

    /// @name ISelectionClient interface implementation
    /// @{
    virtual void GetSelection( TConstScopedObjects& objects ) const;
    virtual void SetSelectionService( ISelectionService* service );

    virtual string GetSelClientLabel();

    virtual void GetSelection( CSelectionEvent& evt ) const;
    virtual void SetSelection( CSelectionEvent& evt );

    virtual void GetMainObject( TConstScopedObjects& objects ) const;

    virtual CRef<objects::CScope> GetScope() const;
    /// @}

protected:
    static CViewTypeDescriptor m_TypeDescr;
    CSelectionService* m_Service;

    bool m_Broadcasting; // view is broadcasting selection

    CSelectionPanel* m_Panel;
    wxEvtHandler* m_Handler;

private:
    CSelectionView(const CSelectionView&);
    CSelectionView& operator=(const CSelectionView&);
};


///////////////////////////////////////////////////////////////////////////////
/// CSelectionViewFactory
class NCBI_GUICORE_EXPORT   CSelectionViewFactory :
    public CObject,
    public IExtension,
    public IViewFactory
{
public:
    /// @name IExtension interface implementation
    /// @{
    virtual string  GetExtensionIdentifier() const;
    virtual string  GetExtensionLabel() const;
    /// @}

    /// @name IViewFactory interface implementation
    /// @{
    virtual void    RegisterIconAliases(wxFileArtProvider& provider);
    virtual const CViewTypeDescriptor&   GetViewTypeDescriptor() const;
    virtual IView*  CreateInstance() const;
    virtual IView*  CreateInstanceByFingerprint(const TFingerprint& fingerprint) const;
    /// @}
};


END_NCBI_SCOPE;

#endif // GUI_CORE___SELECTION_VIEW__HPP
