#ifndef GUI_CORE___TABLE_FILE_LOAD_MANAGER__HPP
#define GUI_CORE___TABLE_FILE_LOAD_MANAGER__HPP

/*  $Id: table_file_load_manager.hpp 26938 2012-11-28 15:46:40Z katargir $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Roman Katargin
 *
 * File Description:
 *
 */

#include <corelib/ncbistl.hpp>
#include <corelib/ncbiobj.hpp>

#include <gui/gui_export.h>

#include <gui/core/ui_file_load_manager.hpp>

#include <gui/core/table_import_data_source.hpp>
#include <gui/core/table_annot_data_source.hpp>

#include <gui/utils/extension.hpp>
#include <gui/objutils/reg_settings.hpp>

BEGIN_NCBI_SCOPE

class CTableFormatPanel;
class CTableDelimitersPanel;
class CTableFixedWidthPanel;
class CTableColumnIdPanel;
class CTableXformPanel;

///////////////////////////////////////////////////////////////////////////////
/// CTableFileLoadManager
///
/// CTableFileLoadManager - a Load Manager for table files. This class implements
/// IFileFormatLoaderManager interface in order to work with generic Format Load
/// Manager.
class  NCBI_GUICORE_EXPORT  CTableFileLoadManager :
    public CObject,
    public IFileFormatLoaderManager,
    public IExtension,
    public IRegSettings
{
public:
    CTableFileLoadManager();

    /// @name IUILoadManager interface implementation
    /// @{
    virtual void    SetServiceLocator(IServiceLocator* srv_locator);
    virtual void    SetParentWindow(wxWindow* parent);
    virtual const IUIObject&  GetDescriptor() const;
    virtual void    InitUI();
    virtual void    CleanUI();
    virtual wxPanel*    GetCurrentPanel();
    virtual bool    CanDo(EAction action);
    virtual bool    IsFinalState();
    virtual bool    IsCompletedState();
    virtual bool    DoTransition(EAction action);
    virtual IAppTask*     GetTask();
    virtual bool          SingleFileLoader() const { return true; }
    virtual IExecuteUnit* GetExecuteUnit();
    virtual IWizardPage*  GetFirstPage() { return &m_SelectFormatPage; }
    virtual void          SetPrevPage(IWizardPage* prevPage) { m_SelectFormatPage.SetPrevPage(prevPage); }
    /// @}

    /// @{ IFileFormatLoaderManager - additional members
    /// @{
    virtual string   GetFileLoaderId() const { return "file_loader_table"; }
    virtual wxString GetFormatWildcard();
    virtual bool     ValidateFilenames(const vector<wxString>& filenames);
    virtual void     SetFilenames(const vector<wxString>& filenames);
    virtual void     GetFilenames(vector<wxString>& filenames) const;
    virtual bool     IsInitialState();
    virtual bool     RecognizeFormat(const wxString& filename);
    virtual bool     RecognizeFormat(CFormatGuess::EFormat fmt);
    /// @}

    /// @name IExtension interface implementation
    /// @{
    virtual string  GetExtensionIdentifier() const;
    virtual string  GetExtensionLabel() const;
    /// @}

    /// @name IRegSettings interface implementation
    /// @{
    virtual void    SetRegistryPath(const string& path);
    virtual void    LoadSettings();
    virtual void    SaveSettings() const;
    /// @}

protected:
    CUIObject m_Descr;
    IServiceLocator* m_SrvLocator;
    wxWindow*   m_ParentWindow;
    string           m_RegPath;
    vector<wxString> m_FileNames;

    CTableFormatPanel*      x_GetSelectFormatPanel();
    CTableDelimitersPanel*  x_GetDelimitersPanel();
    CTableFixedWidthPanel*  x_GetFixedWidthPanel();
    CTableColumnIdPanel*    x_GetColumnIdPanel();
    CTableXformPanel*       x_GetColumnXForm();

    void                    x_ConvertToSeqAnnot();
    void                    x_GuessColumns();

    /// Data is loaded initially into this table for delimiter processing
    CRef<CTableImportDataSource> m_ImportedTableData;
    
    /// After delimiters and column types are identified, table is copied
    /// to this table for saving as asn and data (feature) transforms
    CRef<CTableAnnotDataSource> m_AnnotTableData;

    CTableFormatPanel*      m_TableFormatPanel;
    CTableDelimitersPanel*  m_TableDelimitersPanel;
    CTableFixedWidthPanel*  m_TableFixedWidthPanel;
    CTableColumnIdPanel*    m_TableColumnIdPanel;
    CTableXformPanel*       m_TableColumnXFormPanel;

// Wizard navigation implementation

    class CSelectFormatPage : public CWizardPage
    {
    public:
        CSelectFormatPage(CTableFileLoadManager& manager) : m_Manager(manager) {}
        virtual wxPanel* GetPanel();
        virtual bool CanLeavePage(bool forward);
        virtual IWizardPage* GetNextPage();
        CTableFileLoadManager& m_Manager;
    } m_SelectFormatPage;

    class CDelimitersPage : public CWizardPage
    {
    public:
        CDelimitersPage(CTableFileLoadManager& manager) : m_Manager(manager) {}
        virtual wxPanel* GetPanel();
        virtual bool CanLeavePage(bool forward);
        virtual IWizardPage* GetPrevPage();
        virtual IWizardPage* GetNextPage();
        CTableFileLoadManager& m_Manager;
    } m_DelimitersPage;

    class CFixedWidthPage : public CWizardPage
    {
    public:
        CFixedWidthPage(CTableFileLoadManager& manager) : m_Manager(manager) {}
        virtual wxPanel* GetPanel();
        virtual bool CanLeavePage(bool forward);
        virtual IWizardPage* GetPrevPage();
        virtual IWizardPage* GetNextPage();
        CTableFileLoadManager& m_Manager;
    } m_FixedWidthPage;

    class CColumnIdPage : public CWizardPage
    {
    public:
        CColumnIdPage(CTableFileLoadManager& manager) : m_Manager(manager) {}
        virtual wxPanel* GetPanel();
        virtual bool CanLeavePage(bool forward);
        virtual IWizardPage* GetPrevPage();
        virtual IWizardPage* GetNextPage();
        CTableFileLoadManager& m_Manager;
    } m_ColumnIdPage;

    class CColumnXFormPage : public CWizardPage
    {
    public:
        CColumnXFormPage(CTableFileLoadManager& manager) : m_Manager(manager) {}
        virtual wxPanel* GetPanel();
        virtual bool CanLeavePage(bool forward);
        virtual IWizardPage* GetPrevPage();
        virtual IWizardPage* GetNextPage();
        CTableFileLoadManager& m_Manager;
    } m_ColumnXForm;

    IWizardPage* m_CurrentPage;
};


END_NCBI_SCOPE


#endif  // GUI_CORE___TABLE_FILE_LOAD_MANAGER__HPP
