#ifndef GUI_CORE___TABLE_IMPORT_COLUMN__HPP
#define GUI_CORE___TABLE_IMPORT_COLUMN__HPP

/*  $Id: table_import_column.hpp 26187 2012-07-31 13:23:40Z falkrb $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Bob Falk
 *
 * File Description:
 *
 */

#include <corelib/ncbistl.hpp>
#include <corelib/ncbiobj.hpp>

#include <gui/core/table_import_column.hpp>

#include <objects/general/User_field.hpp>

#include <gui/gui_export.h>


BEGIN_NCBI_SCOPE

USING_SCOPE(objects);

///////////////////////////////////////////////////////////////////////////////
/// CTableImportColumn - 
///
/// This class holds meta-data for the columns in the tabular data. The size
/// and number of columns is automatically updated when the delimiter info
/// is changed. 
///
class CTableImportColumn {
public:
    enum eColumnType { eSeqIdColumn=0, eNumberColumn, eRealNumberColumn,
        eTextColumn, eSkippedColumn, eTypeUndefined };

    enum eDataType { 
        eLocationID = 0, 
        eLocationGI, 
        eLocalID,
        eChromosome,
        eUnspecifiedID,      // IDs
        eGeneName, 
        eStrand,
        eVariationName, 
        eSnpName, 
        eChromosomeNumber,
        eDataRegion,
        eUnspecifiedText,    // Text
        eUnspecifiedReal,    // Real Number 
        eStartPosition, 
        eStopPosition, 
        eLength, 
        eUnspecifiedInt,
        eUndefined,
        eLastType
    };

    static const char* m_TypeNameList[];
    static const char* m_ShortTypeNameList[];

    struct AssemblyInfo {
    public:
        AssemblyInfo() {}
        AssemblyInfo(const string& n, const string& a, const string& d)
            : m_AssemblyName(n)
            , m_Accession(a)
            , m_Description(d) {}

        /// Export assembly in ASN user-object format
        void SaveAsn(CUser_field& user_field) const;
        /// Import assembly from ASN user-object format
        void LoadAsn(CUser_field& user_field);

        string m_AssemblyName;
        string m_Accession;
        string m_Description;
    };

public:
    CTableImportColumn() 
        : m_ColumnWidth(8)
        , m_ColumnType(eTextColumn)
        , m_OneBased(true)
        , m_DataType(eUndefined)
        , m_IsCurrent(false) {}

    void SetWidth(int w) { m_ColumnWidth = w; }
    int GetWidth() const { return m_ColumnWidth; }

    void SetType(eColumnType t) { m_ColumnType = t; }
    eColumnType GetType() { return m_ColumnType; }

    void SetOneBased(bool b) { m_OneBased = b; }
    bool GetOneBased() const { return m_OneBased; }

    /// Get set bio-type info for the column
    void SetDataType(eDataType t) { m_DataType = t; }
    eDataType GetDataType() const { return m_DataType; }
    string GetDataTypeString() const { return GetStringFromDataType(m_DataType); }
    string GetShortDataTypeString() const { return GetShortStringFromDataType(m_DataType); }

    /// Return a string version of a data-type (e.g. "Length" for eLength)
    static string GetStringFromDataType(eDataType t);
    /// Return a short string of a data-type, and blank for 'undefined' types
    static string GetShortStringFromDataType(eDataType t);
    /// Return enum data-type from string
    static eDataType GetDataTypeFromString(const string& s);
    static vector<eDataType> GetMatchingDataTypes(eColumnType c);


    void SetName(const string& n) { m_ColumnName = n; }
    string GetName() const { return m_ColumnName; }
    string GetNameEx() const { return GetSkipped() ? "" : m_ColumnName; }

    void SetAssembly(const AssemblyInfo& a) { m_Assembly = a; }
    AssemblyInfo GetAssembly() const { return m_Assembly; }

    bool GetSkipped() const { return (m_ColumnType == eSkippedColumn); }

    void SetIsCurrent(bool b) { m_IsCurrent = b; }
    bool GetIsCurrent() const { return m_IsCurrent; }

    /// Log column parameters
    void LogColumnInfo() const;

    /// Export column info in ASN user-object format
    void SaveAsn(CUser_field& user_field) const;
    /// Import column info from ASN user-object format
    void LoadAsn(CUser_field& user_field);

protected:
    /// width in chracters
    int m_ColumnWidth;

    /// Column name
    string m_ColumnName;

    /// Matching Assembly for column identfier
    AssemblyInfo m_Assembly;

    /// column type
    eColumnType m_ColumnType;

    /// If integer, is it one based or 0 based?
    bool m_OneBased;

    /// The semantic meaning of the column, e.g. id, start pos, stop pos, strand
    eDataType m_DataType;

    /// If true, this is the current column for editing purposes 
    /// (for views that allow you to edit column type info)
    bool m_IsCurrent;
};


END_NCBI_SCOPE


#endif  // GUI_CORE___TABLE_IMPORT_COLUMN__HPP
