#ifndef GUI_CORE___WORKSPACE_HANDLE__HPP
#define GUI_CORE___WORKSPACE_HANDLE__HPP

/* $Id: workspace.hpp 27441 2013-02-15 21:45:13Z voronov $
* ===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data, the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties, express or implied, including
*  warranties of performance, merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================
*
* Authors:  Andrey Yazhuk
*
* File Description:
*
*
*/

#include <corelib/ncbistl.hpp>


#include <gui/gui_export.h>

#include <gui/core/workspace_folder.hpp>

#include <gui/objects/GBWorkspace.hpp>

#include <corelib/ncbithr.hpp>

#include <set>

#include <wx/string.h>


BEGIN_NCBI_SCOPE

class CProjectService;

/// A workspace proxy MUST be used only on the thread on which it is acquired.

///////////////////////////////////////////////////////////////////////////////
/// CWorkspaceConstPrx

class NCBI_GUICORE_EXPORT CWorkspaceConstPrx : public CObject
{
    friend class CWorkspaceFolderPrx;
    friend class CProjectService; // TODO later replace with a protected class?
    friend class CProjectConstPrx;
    friend class CProjectPrx;
public:
    typedef int TId;
    typedef CProjectPrx::TId TProjectId;
    typedef vector<TProjectId>  TProjectIdVector;
    typedef vector< CRef<CProjectConstPrx> > TProjectConstPrxVector;
    typedef vector< CRef<CProjectPrx> > TProjectPrxVector;
    typedef set< CRef<CSerialObject> >  TObjectsSet;

    enum EProjectState  {
        eAny,
        eLoaded,
        eUnloaded
    };

public:
    virtual ~CWorkspaceConstPrx();

    /// @name Workspace attributes manipulation
    /// @{
    TId      GetId();
    string   GetName();
    string   GetDescr();
    CTime    GetCreateTime();
    CTime    GetModifiedTime();
    bool     IsDirty() const;
    /// @}

    wxString GetFilename() const;

    CRef<CWorkspaceFolderConstPrx>
        FindWorkspaceFolderById_C(CWorkspaceFolderConstPrx::TId folder_id);

    CRef<CProjectFolderConstPrx>
        FindProjectFolderById_C(CWorkspaceFolderConstPrx::TId folder_id);

    /// @name Operations with Projects
    /// {
    /// determine if a project is owned by a workspace and is loaded
    CRef<CWorkspaceFolderConstPrx> GetRootFolder_C();

    void GetProjects_C(TProjectConstPrxVector& projects, EProjectState state = eAny);
    void GetProjects(TProjectPrxVector& projects, EProjectState state = eAny);

    void GetProjectIds(vector<TProjectId>& ids, EProjectState state = eAny);

    /// Return a project handle by project ID; starts from the root folder
    CRef<CProjectConstPrx>   FindProjectById_C(TProjectId prj_id);

    CRef<CProjectPrx>   FindProjectById(TProjectId prj_id);
    CRef<CProjectPrx>   FindProjectByFilename(const wxString& filename);

    CRef<CProjectConstPrx>   FindProjectByScope_C(objects::CScope* scope);
    CRef<CProjectPrx>   FindProjectByScope(objects::CScope* scope);
    /// @}

    /// converts an absolute path to a relative one based on current workspace
    /// directory
    wxString ToRelativePath(const wxString& abs_path);

    /// converts relative path from the current workspace directory to an
    /// absolute path
    wxString ToAbsolutePath(const wxString& rel_path);

protected:
    /// Control exception, thrown when locking construction is impossible
    /// (system cannot accure mutex)
    ///
    struct CResourceBusyEx {
    };

    /// Construction
    /// expect CResourceBusyEx exception (and re-try) 
    CWorkspaceConstPrx(objects::CGBWorkspace& workspace,
                      CProjectService& service,
                      bool get_lock);

    CProjectService*    x_GetProjectService();
    CRWLock&    x_GetProjectsLock();

    /// functions that cause Project Service to issue update
    void    x_OnProjectChanged(CProjectPrx& project);

    void    x_ValidateCall() const;

private:
    // Prohibit copy constructor and assignment operator
    CWorkspaceConstPrx(const CWorkspaceConstPrx& value);
    CWorkspaceConstPrx& operator=(const CWorkspaceConstPrx& value);

protected:
    CRef<objects::CGBWorkspace>  m_Workspace;
    CProjectService*    m_Service;

    /// ID of the thread on which the proxy was created
    CThread::TID    m_OrigThreadID;

    CStopWatch m_Sw;
};


///////////////////////////////////////////////////////////////////////////////
/// CWorkspacePrx

class NCBI_GUICORE_EXPORT CWorkspacePrx : public CWorkspaceConstPrx
{
    friend class CWorkspaceFolderPrx;
    friend class CProjectService;
    friend class CProjectLoadingTask;

public:
    virtual ~CWorkspacePrx();

    void    SetName(const string& title);
    void    SetDescr(const string& descr);
    void    SetCreateTime(const CTime& time);

    void    SetModifiedTime(const CTime& time);

    void    SetDirty( bool flag );
    /// sets new filename for the Workspace
    /// updates path to projects so that they will be correct relative paths
    void    SetFilename(const wxString& filename);

    /// @name Operations with Folders and Projects
    /// {
    CRef<CWorkspaceFolderPrx> GetRootFolder();

    CRef<CWorkspaceFolderPrx> FindWorkspaceFolderById(CWorkspaceFolderPrx::TId folder_id);
    CRef<CProjectFolderPrx>   FindProjectFolderById(CProjectFolderPrx::TId folder_id);

    CRef<CProjectPrx>   CreateNewProject();
    CRef<CProjectPrx>   CreateNewProject(TProjectId id);
    CRef<CProjectPrx>   CreateNewProject(TProjectId id, TObjectsSet& objs);

    /// create a project object for the existing project file
    CRef<CProjectPrx>   CreateNewProject(const wxString& abs_path);

    void    AddProject(CProjectPrx& project);
    void    AddProject(CGBDocument& doc); //TODO
    
    void    RemoveProject(TProjectId id, bool reset_hist_async = true);

    bool    LoadProject(CProjectPrx& project);
    void    UnloadProject(CProjectPrx& project, bool reset_hist_async = true);

    void    UnloadAllProjects();
    /// @}

protected:
    CWorkspacePrx(objects::CGBWorkspace& workspace, CProjectService& service);

    bool    x_RemoveFromFolder(objects::CWorkspaceFolder& folder, 
                               TProjectId prj_id);
    void    x_InsertProjectHandle(objects::CGBProjectHandle& project);

    /// recursive function to establish a new filename for a project
    bool x_SetProjectFilename(objects::CWorkspaceFolder& folder,
                                objects::CGBProjectHandle& project,
                                const string& filename);

    void    x_OnWorkspaceChanged();
    void    x_OnWorkspaceFolderChanged(CWorkspaceFolderPrx& folder);

private:
    // Prohibit copy constructor and assignment operator
    CWorkspacePrx(const CWorkspacePrx& value);
    CWorkspacePrx& operator=(const CWorkspacePrx& value);
};

END_NCBI_SCOPE

#endif // GUI_CORE___WORKSPACE_HANDLE__HPP
