#ifndef GUI_FRAMEWORK___OBSERVABLE__HPP
#define GUI_FRAMEWORK___OBSERVABLE__HPP

/*  $Id: observable.hpp 24490 2011-10-03 15:46:50Z kuznets $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software / database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software / database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Andrey Yazhuk
 *
 * File Description:
 *
 */

/** @addtogroup GUI_FRAMEWORK
*
* @{
*/

#include <corelib/ncbistd.hpp>

#include <gui/gui_export.h>

#include <gui/utils/event_handler.hpp>

BEGIN_NCBI_SCOPE


class IObserver;
class IObserverUpdate;


//////////////////////////////////////////////////////////////////////////////
/// IObservable interface
class IObservable
{
public:
    virtual ~IObservable()   {};

    virtual bool    AttachObserver(IObserver* observer) = 0;
    virtual bool    DetachObserver(IObserver* observer) = 0;
};


///////////////////////////////////////////////////////////////////////////////
/// IObserver
class IObserver
{
public:
    virtual ~IObserver() {}
    virtual void    OnUpdate(IObserverUpdate& update) = 0;
};


//////////////////////////////////////////////////////////////////////////////
/// IObserverUpdate
class IObserverUpdate
{
public:
    virtual ~IObserverUpdate() {}
    /// returns the observer that sends the update
    virtual IObservable*  GetObservable() = 0;
};


///////////////////////////////////////////////////////////////////////////////
/// CUpdate - update object holding information about changes in IObservable
/// derive your classes from CUpdate to supply specific information
class NCBI_GUIFRAMEWORK_EXPORT CObserverUpdate :
    public CObject,
    public IObserverUpdate
{
public:
    /// type of updates
    enum EUpdateType    {
        eInvalid = 0,
        eAttached = 1,  // indicates that Observer has been attached
        ePreDetach,     // indicates that Observer is about to be detached
        eStateChanged,
        eUpdateTypeLast // last value in the enum
    };

    CObserverUpdate() : m_Type( eInvalid ), m_Observable( NULL ) {}
    CObserverUpdate( int type, IObservable* observable )
        : m_Type( type ), m_Observable( observable )
    {}

    int GetType() { return m_Type; }
    
    virtual IObservable* GetObservable();

protected:
    friend class CObservable;
    void Init( int type, IObservable* observable );

private:
    int m_Type;
    IObservable* m_Observable;
};

inline void CObserverUpdate::Init( int type, IObservable* observable )
{
    m_Type = type;
    m_Observable = observable;
}


//////////////////////////////////////////////////////////////////////////////
/// CObservable - default implementation of IObservable, can be used as a base
/// class (mix-in).
class  NCBI_GUIFRAMEWORK_EXPORT  CObservable :
    public CEventHandler,
    public IObservable
{
public:
    typedef CWeakIRef<IObserver>     TObserverWeakPtr;
    typedef vector<TObserverWeakPtr> TObservers;
    typedef CGuard<CObservable>      TGuard;

    CObservable();
    virtual ~CObservable();

    virtual void    GetObservers(TObservers& observers);
    virtual void    DetachAllObservers();

    /// @name IObservable implementation
    /// @{
    virtual bool    AttachObserver(IObserver* observer);
    virtual bool    DetachObserver(IObserver* observer);
    /// @}

    /// @name Locking interface (for use with TGuard and TUpdateGuard)
    /// @{
    virtual void    Lock() const;     /// lock object
    virtual void    Unlock() const;   /// unlock object

    /// lock updates - increment the lock counter
    virtual void    LockUpdates();
    /// unlock updates - decrement the lock counter, flush updates if it is 0
    virtual void    UnlockUpdates();
    /// @}

protected:
    DECLARE_EVENT_MAP();

    /// event handler for CObservableUpdateEvent
    void x_OnUpdateEvent(CEvent* event);

    /// creates and update object and sends to observers, usually
    /// asynchronously
    void x_NotifyObservers();
    NCBI_DEPRECATED void x_NotifyObservers(CObserverUpdate::EUpdateType type);

    /// synchronously sends the update object to observers
    void x_NotifyObserversNow();
    void x_UpdateObserversNow(IObserverUpdate& update);

    void x_PostUpdateEvent();

    /// it is used inside x_PostUpdateEvent(); if overridden,
    /// CObserverUpdate object should be created dynamically
    /// and eStateChanged and "this" should be used in the constructor 
    virtual IObserverUpdate* x_CreateStateChange();


protected:
    TObservers  m_Observers;

    mutable CMutex  m_MainMutex;     /// guards the state of the object

    /// Update management
    CMutex  m_UpdateMutex;   /// guards update lock counter
    int     m_UpdateLockCounter;

    /// indicates whether state has changed since updates have been blocked
    bool    m_StateChanged;
};


//////////////////////////////////////////////////////////////////////////////
/// CObservableUpdateEvent
class NCBI_GUIFRAMEWORK_EXPORT CObservableUpdateEvent : public CEvent
{
public:
    CObservableUpdateEvent(IObserverUpdate& update)
        :   CEvent(eEvent_Message, 0),   m_Update(&update)
    {
    }

    CIRef<IObserverUpdate>  GetUpdate() {   return m_Update;    }
protected:
    CIRef<IObserverUpdate>  m_Update;
};


END_NCBI_SCOPE

/* @} */

#endif  /// GUI_FRAMEWORK___OBSERVABLE__HPP
