#ifndef GUI_OBJUTILS___BOOKMARK__HPP
#define GUI_OBJUTILS___BOOKMARK__HPP

/*  $Id: bookmark.hpp 14562 2007-05-18 11:48:21Z dicuccio $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Mike DiCuccio
 *
 * File Description:
 *
 */

#include <corelib/ncbiobj.hpp>
#include <gui/objutils/registry.hpp>
#include <objects/seqloc/Seq_id.hpp>
#include <objects/seqloc/Seq_loc.hpp>


BEGIN_NCBI_SCOPE


///
/// class CBookmark defines the data stored in an individual sequence-based
/// bookmark.
///

class NCBI_GUIOBJUTILS_EXPORT CBookmark : public CObject
{
public:

    /// default constructor
    CBookmark();

    /// create a bookmark at a particular position
    CBookmark(const objects::CSeq_id& id, TSeqPos pos);

    /// create a bookmark at a particular interval
    CBookmark(const objects::CSeq_id& id, TSeqPos from, TSeqPos to);

    /// save the current settings
    void Save(objects::CUser_field& field) const;

    /// restore settings from a User-field
    void Load(const objects::CUser_field& field);

    /// retrieve the ID associated with this
    const objects::CSeq_id& GetId() const;

    /// retrieve the range for this bookmark
    TSeqRange GetRange() const;

    /// retrieve the label associated with this bookmark
    const string&   GetLabel() const;

    /// retrieve a user-definable field for associated data
    const objects::CUser_field& GetData() const;

    /// set the label associated with this bookmark
    void SetLabel(const string& str);

    /// set the bookmark's location
    void SetLocation(const objects::CSeq_id& id, TSeqPos pos);
    void SetLocation(const objects::CSeq_id& id, TSeqPos from, TSeqPos to);

    /// set just the range of the bookmark
    void SetRange(TSeqPos from, TSeqPos to);

    /// retrieve a wrietable version of the data storage area
    objects::CUser_field& SetData();

private:
    /// parsed data
    TSeqRange m_Range;
    CRef<objects::CSeq_id> m_SeqId;
    string m_Label;

    /// generic data storage
    CRef<objects::CUser_field> m_Data;
};


///
/// class CBookmarkManager manages a set of bookmarks, complete with hooks
/// for serialization and efficient retrieval of bookmarks for a given seq-id
///

class NCBI_GUIOBJUTILS_EXPORT CBookmarkManager : public CObject
{
public:
    /// user-gettable list of bookmarks
    typedef list< CRef<CBookmark> > TBookmarks;

    /// load our settings from the registry
    void LoadSettings(CGuiRegistry::TReadView view);

    /// save our settings to the registry
    void SaveSettings(CGuiRegistry::TReadWriteView view) const;

    /// retrieve bookmarks for a given seq-id
    TBookmarks GetBookmarks(const objects::CSeq_id& id,
                            TSeqPos from, TSeqPos to) const;

    /// retrieve all bookmarks
    TBookmarks GetAllBookmarks() const;

    /// add a bookmark to the manager
    CRef<CBookmark> AddBookmark(const objects::CSeq_id& id,
                                const string& label,
                                TSeqPos pos);
    CRef<CBookmark> AddBookmark(const objects::CSeq_id& id,
                                const string& label,
                                TSeqPos from, TSeqPos to);

    bool RemoveBookmark(CBookmark& bookmark);

    /// retrieve the singleton instance of the bookmark manager
    static CBookmarkManager& GetInstance();

private:
    typedef map<objects::CSeq_id_Handle, TBookmarks> TBookmarkMap;
    TBookmarkMap m_Bookmarks;

};



END_NCBI_SCOPE

#endif  // GUI_OBJUTILS___BOOKMARK__HPP
