#ifndef GUI_OBJUTILS___REGISTRY__HPP
#define GUI_OBJUTILS___REGISTRY__HPP

/*  $Id: registry.hpp 26067 2012-07-16 16:13:44Z katargir $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Mike DiCuccio
 *
 * File Description:
 *
 */

#include <corelib/ncbiobj.hpp>

#include <objects/general/User_field.hpp>
#include <gui/objutils/reg_file.hpp>


BEGIN_NCBI_SCOPE

class NCBI_GUIOBJUTILS_EXPORT CGuiRegistry : public CObject
{
public:
    static const string kMetaSuffix;
    static const string kDecimalDot;

    /// enum controlling priority of added registries
    enum EPriority {
        ePriority_Temp_Inc = 4,
        ePriority_Local = 0,
        ePriority_Global = 16384

    };

    /// enum providing flags for various set/get operations
    enum EFlags {
        fUseOctetString = 0x01,
    };
    typedef int TFlags;


    /// enum controlling how a field should be handled.
    /// there are several policies that can be used:
    enum EPolicy {
        //< override this key from any lower-priority registries
        eOverride,

        //< merge this key by appending it to a list of values
        eMergeAppend,

        //< merge this key by prepending it to a list of values
        eMergePrepend,

        //< default policy: override
        eDefault = eOverride
    };

    enum EKeyAccess {
        eAccess_ReadOnly,
        eAccess_ReadWrite
    };

    struct SKeyInfo {
        string key;
        objects::CUser_field::TData::E_Choice type;
        EKeyAccess access;
    };

    /// retrieve information about all keys in the registry
    typedef list<SKeyInfo> TKeys;
    void GetKeys(TKeys& keys,
                 const string& starting_section = "",
                 const string& delim = CGuiRegistry::kDecimalDot);

    ///
    /// class CReadView provides a nested hierarchical view at a
    /// particular key.  This is a light-weight object that any observer
    /// can hold onto for the life of the application.  It provides read-only
    /// access to the registry at the key from which it was obtained; it cannot
    /// be used to access anything in the list of parent keys.  If the key
    /// that was used to create the view doesn't exist and is added subsequent
    /// to obtaining the view, that key will not show up in the list of keys
    /// for this view.
    ///
    class CReadView;
    typedef CReadView TReadView;

    class NCBI_GUIOBJUTILS_EXPORT CReadView
    {
    public:
        friend class CGuiRegistry;

        virtual ~CReadView();

        CGuiRegistry::TReadView GetReadView(const string& section,
                                            const string& delim = CGuiRegistry::kDecimalDot) const;
        CGuiRegistry::TReadView GetReadView(const string& section,
                                            const vector<string>& subsections,
                                            const string& delim = CGuiRegistry::kDecimalDot) const;

    public:
        /// access a named key at this level, with no recursion
        int    GetInt (const string& key,
                       int default_val = 0,
                       const string& delim = CGuiRegistry::kDecimalDot) const;
        double GetReal(const string& key,
                       double default_val = 0,
                       const string& delim = CGuiRegistry::kDecimalDot) const;
        bool   GetBool(const string& key,
                       bool default_val = false,
                       const string& delim = CGuiRegistry::kDecimalDot) const;
        string GetString(const string& key,
                         const string& default_val = kEmptyStr,
                         const string& delim = CGuiRegistry::kDecimalDot) const;
        string GetOctetString(const string& key,
                              const string& default_val = kEmptyStr,
                              const string& delim = CGuiRegistry::kDecimalDot) const;

        void GetIntVec(const string& key, vector<int>& val,
                       const string& delim = CGuiRegistry::kDecimalDot) const;
        void GetRealVec(const string& key, vector<double>& val,
                        const string& delim = CGuiRegistry::kDecimalDot) const;
        void GetStringVec(const string& key, vector<string>& val,
                          const string& delim = CGuiRegistry::kDecimalDot) const;
        void GetStringList(const string& key, list<string>& val,
                           const string& delim = CGuiRegistry::kDecimalDot) const;
        void GetOctetStringVec(const string& key, vector<string>& val,
                               const string& delim = CGuiRegistry::kDecimalDot) const;

        /// provide raw field access
        CConstRef<objects::CUser_field>
            GetField(const string& key,
                     const string& delim = CGuiRegistry::kDecimalDot) const;


        /// Retrieve information about the top level keys in this view.
        /// The information in the access field is invalid.
        void GetTopKeys(TKeys& keys,
                        const string& starting_section = "",
                        const string& delim = CGuiRegistry::kDecimalDot) const;

        /// Retrieve information about all keys in this view.
        /// The information in the access field is invalid.
        void GetKeys(TKeys& keys,
                     const string& starting_section = "",
                     const string& delim = CGuiRegistry::kDecimalDot) const;

        /// Does a field with this section and key exist in this view?
        bool HasField(const string& key,
                      const string& delim = CGuiRegistry::kDecimalDot) const;

        /// Dump onto the stream all the keys in this view along with their
        /// types and data values.
        void DumpAll(CNcbiOstream& ostr) const;

        /// There is nothing in this view. Probably key(s) used to get it were
        /// wrong.
        bool IsEmpty();

    protected:
        /// implementing GetTopKeys and GetKeys.
        void x_GetKeys(TKeys& keys, const string& starting_section,
                       const string& delim, bool recurse) const;

        void x_AppendToReadView(CGuiRegistry::TReadView& view,
                                const string& section,
                                const string& delim) const;

        /// reference to original registry from which we came...
        CRef<CGuiRegistry> m_Registry;

        /// the original section used to address this view
        string m_Section;

        /// ordered list of subfields to scan
        typedef list< CConstRef<objects::CUser_field> > TFields;
        TFields m_Fields;
    };

    ///
    /// class CReadWriteView provides a nested hierarchical view at a
    /// particular key.  This is a light-weight object that any observer
    /// can hold onto for the life of the application.  It provides read-write
    /// access to the registry at the key from which it was obtained; it cannot
    /// be used to access anything in the list of parent keys.
    ///
    class CReadWriteView;
    typedef CReadWriteView TReadWriteView;

    class NCBI_GUIOBJUTILS_EXPORT CReadWriteView : public CReadView
    {
    public:
        friend class CGuiRegistry;
        virtual ~CReadWriteView();

        CReadWriteView() : m_WriteSite(CGuiRegistry::ePriority_Local) {}

        CGuiRegistry::TReadWriteView
        GetReadWriteView(const string& section,
                         const string& delim = CGuiRegistry::kDecimalDot);

        /// access a named key at this level, with no recursion
        void Set(const string& key, int val,
                 const string& delim = CGuiRegistry::kDecimalDot);
        void Set(const string& key, double val,
                 const string& delim = CGuiRegistry::kDecimalDot);
        void Set(const string& key, bool val,
                 const string& delim = CGuiRegistry::kDecimalDot);
        void Set(const string& key, const string& val,
                 const string& delim = CGuiRegistry::kDecimalDot, TFlags flags = 0);
        void Set(const string& key, const char* val,
                 const string& delim = CGuiRegistry::kDecimalDot, TFlags flags = 0);
        void Set(const string& key, const vector<int>& val,
                 const string& delim = CGuiRegistry::kDecimalDot);
        void Set(const string& key, const vector<double>& val,
                 const string& delim = CGuiRegistry::kDecimalDot);
        void Set(const string& key, const vector<string>& val,
                 const string& delim = CGuiRegistry::kDecimalDot, TFlags flags = 0);
        void Set(const string& key, const list<string>& val,
                 const string& delim = CGuiRegistry::kDecimalDot, TFlags flags = 0);

        /// provide raw field access
        CRef<objects::CUser_field>
            SetField(const string& key,
                     const string& delim = CGuiRegistry::kDecimalDot);

        /// same as SetField, but complain loudly if the field has subfields in
        /// it already.
        CRef<objects::CUser_field>
            SetFieldToValue(const string& key,
                            const string& delim = CGuiRegistry::kDecimalDot);

        /// delete the specified field (and any of its subfields)
        /// from this view and from its registry.
        /// returns true on success,
        /// false if the field does not exist in the writable site.
        bool DeleteField(const string& key,
                         const string& delim = CGuiRegistry::kDecimalDot);
    protected:
        int  m_WriteSite;
    };

    CGuiRegistry();
    ~CGuiRegistry();

    /// @name Set-up and initialization functions
    /// @{

    /// establish our "local" repository.  The local repository is the
    /// highest priority repository; it is also the only repository that can
    /// be serialized directly with this interface
    void SetLocal(CNcbiIstream& istr);

    /// Set the "global" repository.  This is the lowest priority repostory
    /// and is read-only for the application.
    void SetGlobal(CNcbiIstream& istr);

    /// Add a site-specific repository.  This is a set of user-, role-, or
    /// site-wide common features that should be included.
    CRef<CRegistryFile> AddSite(CNcbiIstream& istr, int priority);
    CRef<CRegistryFile> AddSite(const objects::CUser_object& reg, int priority);


    CRef<CRegistryFile> PushWritableTempSite();

    CRef<CRegistryFile> PushTempSite(CNcbiIstream& istr, bool writable = false);
    CRef<CRegistryFile> PushTempSite(const objects::CUser_object& reg,
                                     bool writable = false);
    CRef<CRegistryFile> TopTempSite();

    void    PopTempSite();

    void    MergeTempSite();

    /// Delete a particular site from the registry.
    void DeleteSite(CRef<CRegistryFile> site);

    /// Merge site into the next lowest priority site.
    /// the keys and values look just the same afterwards, except
    /// that keys may change from begin ReadOnly to ReadWrite or vice versa.
    void MergeSite(CRef<CRegistryFile> site);

    /// @}

    /// @name Serialization functions
    /// @{

    /// Write the local policy to a specified stream.
    void Write(CNcbiOstream& ostr, int priority = ePriority_Local) const;

    /// Write the local policy to a specified stream, in INI format
    void ToConfigFile(CNcbiOstream& ostr) const;

    /// Initialize the local repository from an ini-format config file.
    void FromConfigFile(CNcbiIstream& istr);

    /// @}

    /// @name View-related functions
    /// @{

    /// get a read-only view at a particular level.  This provides limited
    /// access to a sub-key and all of its components.
    TReadView GetReadView(const string& section,
                          const string& delim = CGuiRegistry::kDecimalDot) const;

    /// get a read view that will overlay several named sub-sections in a given
    /// section.  This is useful for developing a polymorphic data structure
    /// that contains overloaded data members.  Note that the sections listed
    /// in sub_sections are presumed to be in increasing order of importance,
    /// so if you want a section named "overloaded" to replace one named
    /// "default", add them to the sections in the order "default",
    /// "overloaded"
    TReadView GetReadView(const string& section,
                          const vector<string>& sub_sections,
                          const string& delim = CGuiRegistry::kDecimalDot) const;

protected:
    /// Add a section to an existing view at a lower priority than any previously existing
    /// sections in that view. Implementation for Get*View() methods.
    void    x_AppendToView(TReadView& view,
                           const string& section,
                           const string& delim = CGuiRegistry::kDecimalDot) const;

public:
    /// get a read-write view at a particular level.  This provides limited
    /// access to a sub-key and all of its components.
    TReadWriteView GetReadWriteView(const string& section,
                                    const string& delim = CGuiRegistry::kDecimalDot);

    TReadWriteView GetReadWriteView(const string& section,
                                    int writeSite,
                                    const string& delim = CGuiRegistry::kDecimalDot);

    /// get a read-write view that will overlay several named sub-sections in a
    /// given section.  This is useful for developing a polymorphic data
    /// structure that contains overloaded data members.  Note that the
    /// sections listed in sub_sections are presumed to be in increasing order
    /// of importance, so if you want a section named "overloaded" to replace
    /// one named "default", add them to the sections in the order "default",
    /// "overloaded"
    TReadWriteView GetReadWriteView(const string& section,
                                    const vector<string>& sub_sections,
                                    const string& delim = CGuiRegistry::kDecimalDot);

    /// @}

    /// @name Const-Access Functions
    /// @{

    /// retrieve values by section and key.  Sections and keys are recursively
    /// defined and use a specified set of delimiters; the delimiters are not
    /// serialized.  A policy for agglomeration is also provided; that is, if
    /// the specified key is found in more than one place, the default
    /// behavior can be modified.
    int GetInt(const string& key,
               int default_val = 0,
               const string& delim = CGuiRegistry::kDecimalDot) const;

    double GetReal(const string& key,
                   double default_val = 0, const string& delim = CGuiRegistry::kDecimalDot) const;

    bool GetBool(const string& key,
                 bool default_val = false, const string& delim = CGuiRegistry::kDecimalDot) const;

    string GetString(const string& key,
                     const string& default_val = kEmptyStr,
                     const string& delim = CGuiRegistry::kDecimalDot) const;

    string GetOctetString(const string& key,
                          const string& default_val = kEmptyStr,
                          const string& delim = CGuiRegistry::kDecimalDot) const;

    void GetIntVec(const string& key,
                   vector<int>& val,
                   const string& delim = CGuiRegistry::kDecimalDot) const;

    void GetRealVec(const string& key,
                    vector<double>& val,
                    const string& delim = CGuiRegistry::kDecimalDot) const;

    void GetStringVec(const string& key,
                      vector<string>& val,
                      const string& delim = CGuiRegistry::kDecimalDot) const;

    void GetStringList(const string& key,
                       list<string>& val,
                       const string& delim = CGuiRegistry::kDecimalDot) const;

    void GetOctetStringVec(const string& key,
                           vector<string>& val,
                           const string& delim = CGuiRegistry::kDecimalDot) const;

    /// retrieve the best user field object for our key, taking into account
    /// our multiple stores.
    /// throw an exception if not found.
    CConstRef<objects::CUser_field>
        GetField(const string& key,
                 const string& delim = CGuiRegistry::kDecimalDot) const;

    /// Does a field with this section and key exist in this view?
    bool HasField(const string& key, const string& delim = CGuiRegistry::kDecimalDot) const;

    /// @}

    /// @name Non-Const accessors
    /// @{

    /// set specific values
    void Set(const string& key, int val,
             const string& delim = CGuiRegistry::kDecimalDot);

    void Set(const string& key, double val,
             const string& delim = CGuiRegistry::kDecimalDot);

    void Set(const string& key, bool val,
             const string& delim = CGuiRegistry::kDecimalDot);

    void Set(const string& key, const string& val,
             const string& delim = CGuiRegistry::kDecimalDot,
             TFlags flags = 0);

    void Set(const string& key, const char* val,
             const string& delim = CGuiRegistry::kDecimalDot,
             TFlags flags = 0);

    void Set(const string& key,
             const vector<int>& val,
             const string& delim = CGuiRegistry::kDecimalDot);

    void Set(const string& key,
             const vector<double>& val,
             const string& delim = CGuiRegistry::kDecimalDot);

    void Set(const string& key,
             const vector<string>& val,
             const string& delim = CGuiRegistry::kDecimalDot,
             TFlags flags = 0);

    void Set(const string& key,
             const list<string>& val,
             const string& delim = CGuiRegistry::kDecimalDot,
             TFlags flags = 0);

    void Set(int site, const string& key, int val,
             const string& delim = CGuiRegistry::kDecimalDot);

    void Set(int site, const string& key, double val,
             const string& delim = CGuiRegistry::kDecimalDot);

    void Set(int site, const string& key, bool val,
             const string& delim = CGuiRegistry::kDecimalDot);

    void Set(int site, const string& key, const string& val,
             const string& delim = CGuiRegistry::kDecimalDot,
             TFlags flags = 0);

    void Set(int site, const string& key, const char* val,
             const string& delim = CGuiRegistry::kDecimalDot,
             TFlags flags = 0);

    void Set(int site, const string& key,
             const vector<int>& val,
             const string& delim = CGuiRegistry::kDecimalDot);

    void Set(int site, const string& key,
             const vector<double>& val,
             const string& delim = CGuiRegistry::kDecimalDot);

    void Set(int site, const string& key,
             const vector<string>& val,
             const string& delim = CGuiRegistry::kDecimalDot,
             TFlags flags = 0);

    void Set(int site, const string& key,
             const list<string>& val,
             const string& delim = CGuiRegistry::kDecimalDot,
             TFlags flags = 0);

    /// retrieve the best user field object for our key, taking into account
    /// our multiple stores
    CRef<objects::CUser_field>
        SetField(const string& key,
                 const string& delim = CGuiRegistry::kDecimalDot);

    CRef<objects::CUser_field>
        SetField(int site, const string& key,
                 const string& delim = CGuiRegistry::kDecimalDot);

    /// same as SetField, but complain loudly if the field has subfields in it already.
    CRef<objects::CUser_field>
        SetFieldToValue(const string& key,
                        const string& delim = CGuiRegistry::kDecimalDot);

    CRef<objects::CUser_field>
        SetFieldToValue(int site, const string& key,
                        const string& delim = CGuiRegistry::kDecimalDot);

    /// delete the specified field (and any of its subfields).
    /// returns true on success,
    /// false if the field does not exist in the writable site,
    /// or if the field is at the top level of the registry.
    bool DeleteField(const string& key,
                     const string& delim = CGuiRegistry::kDecimalDot);

    bool DeleteField(int site, const string& key,
                     const string& delim = CGuiRegistry::kDecimalDot);

    /// Get the priority of the site that Set methods modify.
    int  GetWritableSite();

    /// Set which site Set methods modify.
    /// Creates the site if it did not exist.
    void  SetWritableSite(int priority = ePriority_Local);

    /// @}

    /// @name Singleton maintenance
    /// @{

    /// access the application-wide singleton
    static CGuiRegistry& GetInstance();

    /// @}

private:

    /// our registry stores
    /// this is a hierarchical organization

    /// the prioritized list of policies we're considering
    typedef map<int, CRef<CRegistryFile> > TRegistries;
    TRegistries m_Registries;

    /// The priority of the site that writes go to.
    int         m_WritableSite;

    /// return the registry at a particular priority.
    /// Do not create if it does not exist.
    CRef<CRegistryFile>     x_GetRegistryRef(int) const;

    /// return a registry at a particular priority, creating it if need be.
    CRegistryFile&          x_SetRegistry(int);

    CRegistryFile&          x_GetLocal();

    void        x_DeleteSite(TRegistries::iterator);
    void        x_MergeSite(TRegistries::iterator site_it);
    int         x_NextTempPriority() const;


    /// forbidden
    CGuiRegistry(const CGuiRegistry&);
    CGuiRegistry& operator=(const CGuiRegistry&);
};


class NCBI_GUIOBJUTILS_EXPORT CGuiRegistryUtil
{
public:
    /// recursively navigate an object tree, retrieving an existing named key
    static CConstRef<objects::CUser_field>
        GetFieldAt(const objects::CUser_object& obj,
                   const string& key,
                   const string& delim = CGuiRegistry::kDecimalDot);

    /// recursively navigate an object tree, setting keys where necessary to get
    /// a named level
    static CRef<objects::CUser_field>
        SetFieldAt(objects::CUser_object& obj,
                   const string& key,
                   const string& delim = CGuiRegistry::kDecimalDot);

    /// reconcile and combine a set of user fields into a single field
    /// as described by the appropriate merge policy
    static CConstRef<objects::CUser_field>
        ResolveFieldPolicy(list< CConstRef<objects::CUser_field> >& fields,
                           CGuiRegistry::EPolicy policy);

    /// create a key from a section and a subkey
    static string MakeKey(const string& section, const string& key,
                          const string& delim = CGuiRegistry::kDecimalDot);
};



END_NCBI_SCOPE

#endif  // GUI_OBJUTILS___REGISTRY__HPP
