#ifndef GUI_UTILS___TOOLTIP__HPP
#define GUI_UTILS___TOOLTIP__HPP

/*  $Id: tooltip.hpp 26683 2012-10-23 20:30:24Z rudnev $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Dmitry Rudnev
 *
 * File Description:
 *
 */

#include <corelib/ncbiobj.hpp>
#include <gui/gui.hpp>

/** @addtogroup GUI_UTILS
 *
 * @{
 */


BEGIN_NCBI_SCOPE


// primitive interface to arrange tabular data in the tooltips
// does absolutely no checking that columns and rows are added and finished correctly, so must be used with care
class ITooltipFormatter
{
public:
    enum ETooltipFormatters {
        eTooltipFormatter_CSSTable,         // CSS table based
        eTooltipFormatter_Html,             // HTML attributes (no CSS) based
        eTooltipFormatter_Text,             // plain text formatted
    };

    static CIRef<ITooltipFormatter> CreateTooltipFormatter(ETooltipFormatters);

    // table-building
    virtual void StartRow() = 0;

    // will add an empty cell if sBulletSrc is empty
    virtual void AddBulletCol(const string& sBulletSrc = "") = 0;
    virtual void AddTagCol(const string& sContents = "") = 0;
    virtual void AddValueCol(const string& sContents = "") = 0;
    virtual void FinishRow() = 0;
    virtual void AddDividerRow() = 0;

    // returns the formatted tooltip text and resets the formatter for reuse
    virtual string Render() = 0;

    // helper functions
    virtual string CreateLink(const string& sText, const string& sUrl) const = 0;
    // pmids is a comma-delimited list of numeric pubmed ids
    // will create a new row with tag "Pubmed" and value list of pubmed URLs
    virtual void AddPubmedLinksRow(const string& pmids, bool& isGoToPresent, bool isBulletColPresent) = 0;

    // conditionally add a Go To row if isGoToPresent == false; will change isGoToPresent to true
    virtual void MaybeAddGoToRow(bool& isGoToPresent, bool isBulletColPresent) = 0;

    virtual ~ITooltipFormatter() {}
};

// tooltips using html tables with css classes from style.css
class NCBI_GUIOBJUTILS_EXPORT CCSSTableTooltipFormatter : public ITooltipFormatter, public CObject
{
public:
    static CIRef<ITooltipFormatter> CreateTooltipFormatter();
    virtual void StartRow();
    virtual void AddBulletCol(const string& sBulletSrc = "");
    virtual void AddTagCol(const string& sContents = "");
    virtual void AddValueCol(const string& sContents = "");
    virtual void FinishRow();
    virtual void AddDividerRow();

    virtual string Render();

    virtual string CreateLink(const string& sText, const string& sUrl) const;
    virtual void AddPubmedLinksRow(const string& pmids, bool& isGoToPresent, bool isBulletColPresent);
    virtual void MaybeAddGoToRow(bool& isGoToPresent, bool isBulletColPresent);
private:
    string m_sTooltipText;

    CCSSTableTooltipFormatter() {}
    CCSSTableTooltipFormatter(const CCSSTableTooltipFormatter&);
    CCSSTableTooltipFormatter& operator=(const CCSSTableTooltipFormatter&);
};




// tooltips using html tables with html attibute formatting
class NCBI_GUIOBJUTILS_EXPORT CHtmlTooltipFormatter : public ITooltipFormatter, public CObject
{
public:
    static CIRef<ITooltipFormatter> CreateTooltipFormatter();
    virtual void StartRow();
    virtual void AddBulletCol(const string& sBulletSrc = "");
    virtual void AddTagCol(const string& sContents = "");
    virtual void AddValueCol(const string& sContents = "");
    virtual void FinishRow();
    virtual void AddDividerRow();

    // returns the formatted tooltip and resets the formatter for reuse
    virtual string Render();

    virtual string CreateLink(const string& sText, const string& sUrl) const;
    virtual void AddPubmedLinksRow(const string& pmids, bool& isGoToPresent, bool isBulletColPresent);
    virtual void MaybeAddGoToRow(bool& isGoToPresent, bool isBulletColPresent);
private:
    string m_sTooltipText;

    CHtmlTooltipFormatter() {}
    CHtmlTooltipFormatter(const CHtmlTooltipFormatter&);
    CHtmlTooltipFormatter& operator=(const CHtmlTooltipFormatter&);
};


// tooltips using text formatting
class NCBI_GUIOBJUTILS_EXPORT CTextTooltipFormatter : public ITooltipFormatter, public CObject
{
public:
    static CIRef<ITooltipFormatter> CreateTooltipFormatter();
    virtual void StartRow();
    virtual void AddBulletCol(const string& sBulletSrc = "");
    virtual void AddTagCol(const string& sContents = "");
    virtual void AddValueCol(const string& sContents = "");
    virtual void FinishRow();
    virtual void AddDividerRow();

    // returns the formatted tooltip and resets the formatter for reuse
    virtual string Render();

    virtual string CreateLink(const string& sText, const string& sUrl) const;
    virtual void AddPubmedLinksRow(const string& pmids, bool& isGoToPresent, bool isBulletColPresent);
    virtual void MaybeAddGoToRow(bool& isGoToPresent, bool isBulletColPresent);
private:
    string m_sTooltipText;

    CTextTooltipFormatter() {}
    CTextTooltipFormatter(const CTextTooltipFormatter&);
    CTextTooltipFormatter& operator=(const CTextTooltipFormatter&);
};






END_NCBI_SCOPE


/* @} */

#endif  // GUI_UTILS___TOOLTIP__HPP
