#ifndef GUI_OBJUTILS___UTILS__HPP
#define GUI_OBJUTILS___UTILS__HPP

/*  $Id: utils.hpp 28078 2013-05-06 22:10:55Z rafanovi $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Mike DiCuccio, Liangshou Wu
 *
 * File Description:
 *    General utility classes for GUI projects.
 */

#include <corelib/ncbistd.hpp>
#include <corelib/ncbimisc.hpp>
#include <util/range_coll.hpp>
#include <objmgr/annot_selector.hpp>
#include <objmgr/mapped_feat.hpp>
#include <gui/gui.hpp>

/** @addtogroup GUI_OBJUTILS
 *
 * @{
 */


BEGIN_NCBI_SCOPE

BEGIN_SCOPE(objects)
    class CSeq_loc;
    class CBioseq_Handle;
END_SCOPE(objects)


class NCBI_GUIOBJUTILS_EXPORT CLinkedFeature : public CObject
{
public:
    typedef list< CRef<CLinkedFeature> >  TLinkedFeats;
    CLinkedFeature(const objects::CMappedFeat& feat) : m_Feat(feat) {}

    const objects::CMappedFeat& GetMappedFeature() const { return m_Feat; }

    objects::CMappedFeat& GetMappedFeature(){ return m_Feat; }

    const objects::CSeq_feat& GetFeature() const { return m_Feat.GetOriginalFeature(); }

    const TLinkedFeats& GetChildren() const { return m_Children; }

    TLinkedFeats& GetChildren() { return m_Children; }

    void AddChild(CLinkedFeature* feat)
    { m_Children.push_back(CRef<CLinkedFeature>(feat)); }

    void AddChild(const objects::CMappedFeat& feat) 
    { m_Children.push_back(CRef<CLinkedFeature>(new CLinkedFeature(feat))); }

private:
    objects::CMappedFeat    m_Feat;
    TLinkedFeats            m_Children;
};


/// Task clients implement this callback interface.
class NCBI_GUIOBJUTILS_EXPORT ISeqTaskProgressCallback
{
public:
    virtual void SetTaskName      (const string& name) = 0;
    /// set total finished task number.
    virtual void SetTaskCompleted (int completed)      = 0;
    /// set to add newly finished task number.
    virtual void AddTaskCompleted (int delta)          = 0;
    virtual void SetTaskTotal     (int total)          = 0;
    virtual bool StopRequested    () const             = 0;
    virtual ~ISeqTaskProgressCallback() {}
};

#define __GUI_SEQ_UTILS_DEFINED__
class NCBI_GUIOBJUTILS_EXPORT CSeqUtils
{
public:
    typedef vector< CRef<objects::CSeq_loc> > TLocVec;

    /// flags for classifying annotation names.
    enum EAnnotationNameType
    {
        eAnnot_Unnamed,           ///< unnamed annotation
        eAnnot_Named,             ///< all named annotations
        eAnnot_All,               ///< all annotations
        eAnnot_Other              ///< any given named annots
    };

    typedef EAnnotationNameType     TAnnotNameType;

    static TAnnotNameType NameTypeStrToValue(const string& type);
    static const string& NameTypeValueToStr(TAnnotNameType type);

    /// flags controlling feature retrieval
    enum {
        fFeature_LinkFeatures = 0x1,
        fFeature_Default = 0
    };
    typedef int TFeatureFlags;

    /// flags controlling use of annotation selectors
    enum {
        //< don't apply any settings that affect the retrieval or presentation
        //< of named annotation sets
        fAnnot_UnsetNamed = 0x01,

        //< don't apply any settings that affect the depth of annotation retrieval
        fAnnot_UnsetDepth = 0x02
    };
    typedef int TAnnotFlags;


    /// request an annotation selector for a given type
    static objects::SAnnotSelector
        GetAnnotSelector(TAnnotFlags flags = 0);
    
    static objects::SAnnotSelector
    GetAnnotSelector(objects::SAnnotSelector::TAnnotType choice,
                     TAnnotFlags flags = 0);

    static objects::SAnnotSelector
    GetAnnotSelector(objects::SAnnotSelector::TFeatType,
                     TAnnotFlags flags = 0);

    static objects::SAnnotSelector
    GetAnnotSelector(objects::SAnnotSelector::TFeatSubtype,
                     TAnnotFlags flags = 0);

    /// create a annotation selector for collecting annotation names only.
    /// @param annots
    static objects::SAnnotSelector
        GetAnnotSelector(const vector<string>& annots);

    static objects::SAnnotSelector
        GetAnnotSelector(const vector<string>& annots,
        bool adaptive, int depth);

    /// help function for setting up an annotation.
    static void SetAnnot(objects::SAnnotSelector& sel, const string& annot);

    /// help function for setting selector resolve depth.
    /// @param sel target annotation selector
    /// @param adaptive adaptive selector if true, otherwise, exact selector
    /// @param depth resolve depth. For adaptive selector, it resolves all
    ///        if depth < 0, otherwie resolve up to the specified depth.
    ///        For exact selector, depth is the exact the resovle depth (>=0).
    static void SetResolveDepth(objects::SAnnotSelector& sel,
        bool adaptive, int depth = -1);

    typedef int TFeatLinkingMode;

    /// Link features into a hierarchical list.  This function will transform a
    /// linear, feature-order-sorted list of CLinkedFeature into a
    /// hierarchically arranged list in which genes are linked to mRNAs, mRNAs
    /// to CDSs, etc.
    /// @param mode 0:eFeatId_ignore, 1:eFeatId_by_type, 2:eFeatId_always
    static bool LinkFeatures(CLinkedFeature::TLinkedFeats& feats,
        TFeatLinkingMode mode = 1, // TFeatLinkingMode
        ISeqTaskProgressCallback* p_cb = NULL);

    /// remap a location to a parent location.  The child location is split
    /// such that all of its relative positions will map correctly to the
    /// parent location's positions
    static CRef<objects::CSeq_loc>
    RemapChildToParent(const objects::CSeq_loc& parent,
                       const objects::CSeq_loc& child,
                       objects::CScope* scope = NULL);

    /// check to see if two seq-ids are identical.  If provided with a scope,
    /// this function will use the synonyms available through a scope to check
    /// for identicality.  Note that this may require network access, and may
    /// be slow.
    static bool Match(const objects::CSeq_id& id1,
                      const objects::CSeq_id& id2,
                      objects::CScope* scope = NULL);

    static bool Match(const objects::CSeq_id_Handle& id1,
                      const objects::CSeq_id_Handle& id2,
                      objects::CScope* scope = NULL);

    // create a CSeq_loc from the given CSeq_id and a collection of ranges
    static CRef<objects::CSeq_loc>
        CreateSeq_loc(const objects::CSeq_id& id,
                      const CRangeCollection<TSeqPos>& ranges);

    // extracts from the given map all segments corresponding to the given id
    // and puts them to the "ranges" collection. Retuns false if id is not found
    // in the map.
    static bool   GetRangeCollection(const objects::CSeq_id& id,
                                     const objects::CHandleRangeMap& map,
                                     CRangeCollection<TSeqPos>& ranges);

    /// Construct a bioseq to fit a given location.  If the location is of
    /// type 'whole', then a duplicate of the original will be returned;
    /// otherwise, a new bioseq will be constructed to cover just the
    /// region indicated in the seq-loc
    static objects::CBioseq* SeqLocToBioseq(objects::CScope& scope,
                                            const objects::CSeq_loc& loc);

    /// Create a new seq-loc with a unique seq-id from a "mixed" loc.
    /// "Mixed" means mixed seq-ids, e.g. a seq-loc with multi seq-ids.
    static CRef<objects::CSeq_loc>
        MixLocToLoc(const objects::CSeq_loc& mix_loc,
                    const objects::CBioseq_Handle& handle);


    /// Tokenize a string using the delim.
    /// It skip any escaped ('\') characters.
    static void TokenizeWithEscape(const string &str,
                                   const string &delim,
                                   vector<string>& tokens,
                                   bool remove_escape = false);

    /// remove escape characters ('\') used for escaping special characters.
    static string RemoveEscape(const string &str);

    static string GetAnnotName(const objects::CSeq_annot_Handle& annot_handle);

    static string GetAnnotComment(const objects::CSeq_annot_Handle& annot_handle);

    /// check if a given annotation is a named annotation accession.
    static bool IsNAA(const string& annot);

    /// check if a given annotation is a unnamed annotation.
    static bool IsUnnamed(const string& annot);

    /// Get the commonly used symbol representing a unnnamed annotation.
    static const string& GetUnnamedAnnot();

    static bool IsPseudoFeature(const objects::CSeq_feat& feat);
    static bool IsPartialFeature(const objects::CSeq_feat& feat);
    static bool IsPartialStart(const objects::CSeq_loc& loc);
    static bool IsPartialStop(const objects::CSeq_loc& loc);
    static bool IsException(const objects::CSeq_feat& feat);
    static string GetNcbiBaseUrl();

    /// help methods for creating HTML text
    static string CreateTableStart();
    static string CreateTableEnd();
    static string CreateTableRow(const string& tag, const string& value);
    static string CreateSectionRow(const string& tag);
    static string CreateLinkRow(const string& tag,
        const string& label, const string& url);

    /// Retrieve mapped-up sequences.
    /// Help method for retrieving upper level sequences for a given gi.
    /// @param gi: the source sequence gi
    static TLocVec GetGiPlacements(int gi, int time_out_sec  = 5);

    /// Convert a range string to a range.
    /// @return flag indicating if the input range string is valid
    static bool StringToRange(const string& range_str, long& from, long& to);
};


inline
bool CSeqUtils::IsUnnamed(const string& annot)
{
    return annot == GetUnnamedAnnot();
}

inline
const string& CSeqUtils::GetUnnamedAnnot()
{
    static string kUnnamedAnnot = "Unnamed";
    return kUnnamedAnnot;
}



END_NCBI_SCOPE


/* @} */

#endif  /// GUI_OBJUTILS___UTILS__HPP
