#ifndef GUI_OPENGL___GLBACKGROUND__HPP
#define GUI_OPENGL___GLBACKGROUND__HPP

/*  $Id: glbackground.hpp 15168 2007-10-15 19:11:34Z dicuccio $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Mike DiCuccio
 *
 * File Description:
 *
 */

#include <corelib/ncbiobj.hpp>
#include <gui/opengl/glbackground.hpp>
#include <gui/utils/rgba_color.hpp>
#include <gui/opengl/glrect.hpp>

/** @addtogroup GUI_OPENGL
 *
 * @{
 */

BEGIN_NCBI_SCOPE

class CGlTexture;

//
// class CGlBackground abstracts the notion of the background for a scene.
//
// This class provides options for easily creating a number of simple
// background effects.  The options supported include:
//      - solid color
//      - color gradient from top->bottom or left->right with arbitrarily
//        defined light and dark colors
class NCBI_GUIOPENGL_EXPORT CGlBackground
{
public:
    // enum of rvarious background effects
    enum EType {
        eSolid,         // solid color
        eGradientDown,  // gradient color, light->dark from top->bottom
        eGradientRight, // gradient color, light->dark from left->right
        eTexture        // texture
    };

    CGlBackground();
    ~CGlBackground();
    explicit CGlBackground(EType type);

    // set the type of background we use
    void SetType(EType type);

    // set the colors we use for the background.  We have more than one color -
    // light color and dark.  For the solid color background, the light color
    // is used (this is exposed as SetColor()).  For the gradient colors, we
    // use two, a light and a dark.  Calling SetDark() with no parameters sets
    // the dark color as a multiple of the light color.
    // There is no requirement that the light color be brighter than the dark
    // color.
    void SetColor(const CRgbaColor& c);
    void SetLightColor(const CRgbaColor& c);
    void SetDarkColor(const CRgbaColor& c);
    void SetDarkColor(void);

    // set the texture we use.  This is only used if the background type is
    // eTexture
    void SetTexture(CGlTexture& tex);

    // draw our background.  This will explicitly preserve all matrices.
    void Draw(void);

    // draw our background in the specified viewport.  This will explicitly
    // preserve all matrices.
    void Draw(const CGlRect<int>& vp);

private:

    // the type of background we draw
    EType m_Type;

    // our colors
    CRgbaColor m_LightColor;
    CRgbaColor m_DarkColor;

    // our texture
    CRef<CGlTexture> m_Texture;

    CGlBackground(const CGlBackground&);
    CGlBackground& operator=(const CGlBackground&);
};

END_NCBI_SCOPE

/* @} */

#endif  // GUI_OPENGL___GLBACKGROUND__HPP
