#ifndef GUI_OPENGL___GL_COLOR_TABLE__HPP
#define GUI_OPENGL___GL_COLOR_TABLE__HPP

/*  $Id: glcolortable.hpp 23948 2011-06-23 17:05:49Z kuznets $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Mike DiCuccio, Andrey Yazhuk
 *
 * File Description:
 *    Incremental and mutually contrastive color table generator.
 */


#include <corelib/ncbistd.hpp>

#include <gui/utils/rgba_color.hpp>

#include <algorithm>
#include <math.h>

/** @addtogroup GUI_OPENGL
 *
 * @{
 */


BEGIN_NCBI_SCOPE

////////////////////////////////////////////////////////////////////////////////
/// CRgbaColorTable
class NCBI_GUIOPENGL_EXPORT CRgbaColorTable
{
public:

    CRgbaColorTable();
    CRgbaColorTable(size_t items, float min_color = 0.0f, float max_color = 1.0f);

    void    SetMaxItems(size_t items);
    void    SetMinColor(float f);
    void    SetMinColor(unsigned char f);
    void    SetMaxColor(float f);
    void    SetMaxColor(unsigned char f);

    CRgbaColor    GetColor(size_t i) const;

private:
    size_t m_MaxItems;
    float m_ColorStep;
    float m_MinColor;
    float m_MaxColor;
    float m_ColorDiff;

    void x_CalcColorStep();
};


////////////////////////////////////////////////////////////////////////////////
/// CRgbaGradColorTable
/// Provides a storage for colors (to eliminate color creation overhead) and
/// Functions for initializing the table.
class NCBI_GUIOPENGL_EXPORT CRgbaGradColorTable
{
public:
    CRgbaGradColorTable();
    CRgbaGradColorTable(int size);

    size_t     GetSize() const;
    void    SetSize(size_t size);

    inline CRgbaColor&   GetColor(int i);
    inline const CRgbaColor&   GetColor(int i) const;

    inline CRgbaColor&   operator[](int i);
    inline const CRgbaColor&   operator[](int i) const;

    /// returns color index by the normalized value; normalized values is a value
    /// in the [0.0, 1.0] range that is mapped to the index range of color table [0, size-1]
    inline int  GetIndex(double norm) const;

    /// initialize the whole table with gradient colors in [start_c, end_c] range
    void    FillGradient(const CRgbaColor& start_c, const CRgbaColor& end_c);

    /// initialize part of the table specified by the start index "start_i" and
    /// the number of element "n" with gradient colors in [start_c, end_c] range
    void    FillGradient(int start_i, int n,
                         const CRgbaColor& start_c,
                         const CRgbaColor& end_c);

    /// initialize part of the table specified by the start index "start_i" and
    /// the number of element "n" with gradient colors. First half of the table
    /// is initialized with [start_c, mid_c], the second part with [mid_c, end_c]
    void    FillGradient(int start_i, int n,
                         const CRgbaColor& start_c,
                         const CRgbaColor& mid_c,
                         const CRgbaColor& end_c);


protected:
    vector<CRgbaColor>    m_Colors;
};

///////////////////////////////////////////////////////////////////////////////
// inlines
inline
void CRgbaColorTable::SetMaxItems(size_t items)
{
    m_MaxItems = max((size_t)8, items);
    x_CalcColorStep();
}


inline
void CRgbaColorTable::SetMinColor(float f)
{
    m_MinColor = max(0.0f, f);
    m_MinColor = min(1.0f, m_MinColor);
    if (m_MaxColor < m_MinColor) {
        std::swap(m_MaxColor, m_MinColor);
    }
    m_ColorDiff = m_MaxColor - m_MinColor;
    x_CalcColorStep();
}


inline
void CRgbaColorTable::SetMinColor(unsigned char c)
{
    SetMinColor(float (c) / 255.0f);
}


inline
void CRgbaColorTable::SetMaxColor(float f)
{
    m_MaxColor = max(0.0f, f);
    m_MaxColor = min(1.0f, m_MaxColor);
    if (m_MaxColor < m_MinColor) {
        std::swap(m_MaxColor, m_MinColor);
    }
    m_ColorDiff = m_MaxColor - m_MinColor;
    x_CalcColorStep();
}


inline
void CRgbaColorTable::SetMaxColor(unsigned char c)
{
    SetMinColor(float (c) / 255.0f);
}


inline
void CRgbaColorTable::x_CalcColorStep()
{
    m_ColorStep = m_ColorDiff / m_MaxItems * 8;
}


inline CRgbaColor&   CRgbaGradColorTable::GetColor(int i)
{
    _ASSERT(i >=0  && i < (int) m_Colors.size());
    return m_Colors[i];
}

inline const CRgbaColor&   CRgbaGradColorTable::GetColor(int i) const
{
    _ASSERT(i >=0  && i < (int) m_Colors.size());
    return m_Colors[i];
}


inline CRgbaColor&   CRgbaGradColorTable::operator[](int i)
{
    return GetColor(i);
}


inline const CRgbaColor&   CRgbaGradColorTable::operator[](int i) const
{
    return GetColor(i);
}


inline int  CRgbaGradColorTable::GetIndex(double norm) const
{
    int index = (int) floor(norm * m_Colors.size());
    index = std::min(index, (int) m_Colors.size() - 1);
    return index;
}

END_NCBI_SCOPE

/* @} */

#endif  // GUI_OPENGL___GL_COLOR_TABLE__HPP
