#ifndef GUI_OPENGL___GL_RESOURCE_MANAGER__HPP
#define GUI_OPENGL___GL_RESOURCE_MANAGER__HPP

/*  $Id: glresmgr.hpp 27503 2013-02-26 18:16:55Z falkrb $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Bob Falk
 *
 * File Description:
 *
 */

#include <corelib/ncbiobj.hpp>
#include <gui/opengl.h>
#include <gui/gui.hpp>

#include <gui/opengl/globject.hpp>
#include <gui/opengl/glutils.hpp>
#include <gui/opengl/glbuffer.hpp>
#include <gui/opengl/glrender.hpp>
#include <gui/opengl/glstate.hpp>
#include <gui/opengl/glmodel2d.hpp>

/** @addtogroup GUI_OPENGL
 *
 * @{
 */

BEGIN_NCBI_SCOPE

class CGlVboGeom;

class NCBI_GUIOPENGL_EXPORT CGlResMgr
{
public:
    /// Needs active OpenGL context when created.  Determines best api level.
    CGlResMgr();

    /// Destroys all gl objects it created.
    ~CGlResMgr();

    /// Deletes all objects.  Since object is a singleton, don't want to call
    /// dtors during program exit as this can fail
    void Clear();

    /// Find GL object (add type option?) TBD
    //CGlObject* FindObject(const string& name);

    /// Set api level by querying opengl
    void SetApiLevel();
    /// Set the api level directly (override ctor). Good for testing.
    void SetApiLevel(ERenderTarget level);
    /// Get api level
    ERenderTarget GetApiLevel() const { return m_ApiLevel; }

    /// Gets a buffer (for vertices or attributes) that is compatible with
    /// the current OpenGL level
    /// @param target Parm for glBindBuffer() - one of GL_ARRAY_BUFFER (vertices
    ///               and attributes) or GL_ELEMENT_ARRAY_BUFFER (indices). 
    //                Ignored in OpenGL 1.1
    template<class T>
    CGlBuffer<T>* GetBuffer(const string& name, GLenum target=GL_ARRAY_BUFFER);

    /// Get geometry object for a node that renders using vertex buffers.
    CGlVboGeom* GetVboGeom(const string& name);

    ///
    /// Get/Set (buffered) 2D models
    ///


    /// Add model to set of 2D models.  m->GetName() should be unique.
    void AddModel(CIRef<CGlModel2D>& m);
    /// Find and return model with name "name" or Null if not found
    CIRef<CGlModel2D> GetModel(const string& name);
    /// Remove and return (if found) model "m"
    CIRef<CGlModel2D> RemoveModel(CIRef<CGlModel2D>& m);



    ///
    /// Get/Set Renderer
    ///

    void AddRenderer(CRef<CGlRender>& rm);
    CRef<CGlRender> RemoveRenderer(CRef<CGlRender>& rm);

    /// Returns the (first) renderer with name==rm_name or NULL
    CRef<CGlRender> GetRenderer(const string& rm_name);
    /// Returns first renderer in m_Renderers that renders to 'target'
    CRef<CGlRender> GetRenderer(ERenderTarget target);
    
    /// Set current renderer (rm must already be in m_Renderers)
    void SetCurrentRenderer(CRef<CGlRender>& rm);
    /// Returns current renderer, or NULL
    CRef<CGlRender>& GetCurrentRenderer() {return m_CurrentRenderer;}


    // Get single instance of resource manager (singleton).
    static CGlResMgr& Instance();
    /// return true if static instance has been created (via a call to Instance()
    static bool Initialized();

protected:   
    static CGlResMgr* m_StaticInstance;

    ERenderTarget m_ApiLevel;

    std::vector<CRef<CGlRender> > m_Renderers;
    CRef<CGlRender> m_CurrentRenderer;

    std::vector<CIRef<CGlModel2D> > m_Models2D;
};

template<class T>
CGlBuffer<T>* CGlResMgr::GetBuffer(const string& name, 
                                   GLenum target)
{
    if (m_ApiLevel == eOpenGL20)
        return new CGlBuffer20<T>(target);
    else
        return new CGlBuffer11<T>(target);        
}


END_NCBI_SCOPE


#endif // GUI_OPENGL___GL_RESOURCE_MANAGER__HPP
