#ifndef GUI_OPENGL___GL_VBO_GEOM__HPP
#define GUI_OPENGL___GL_VBO_GEOM__HPP

/*  $Id: glvbogeom.hpp 25176 2012-01-31 14:06:29Z falkrb $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Bob Falk
 *
 * File Description:
 * CGlVboGeom holds 1 or more vertex buffers to be used in rendering.  If there
 * are multiple buffers they are all enabled for rendering together (except for
 * buffers explicitly marked as not visible).
 *
 * Since there is a difference in how buffers are enabled and used in OpenGL 1.4 
 * and below, there are two concrete implementations of CGlVboGeom.  For CVboGeom11
 * the buffers are assumed to be arrays and the commands glVertexPointer, glColorPointer,
 * etc. provide pointers to the arrays (of vertex and vertex attribute data) to be rendered.
 *
 * For CGlVboGeom20 the buffers are OpenGL vertex buffers, so the buffers are enabled
 * prior to calling glVertexPointer, glColorPointer, etc.  Note that in the future
 * CGlVboGeom20 could use glVertexAttribPointer to specify vertex attributes in a more
 * generic fasion (and in a way that is also compatible with OpenGL ES 2.0, which does not
 * support glVertexPointer, glColorPointer etc.)
 *
 *
 *
 *                             CGlVboGeom
 *                              ^    ^
 *                              |    |
 *                        ______|    |__________
 *                        |                     |
 *                        |                     |
 *                   CGlVboGeom11          CGlVboGeom20
 */

#include <gui/opengl.h>
#include <gui/gui.hpp>

#include <gui/utils/vect2.hpp>
#include <gui/utils/vect3.hpp>
#include <gui/utils/vect4.hpp>
#include <gui/utils/matrix3.hpp>
#include <gui/utils/matrix4.hpp>

#include <gui/utils/rgba_color.hpp>
#include <gui/opengl/globject.hpp>
#include <gui/opengl/glbuffer.hpp>


/** @addtogroup GUI_OPENGL
 *
 * @{
 */

BEGIN_NCBI_SCOPE

//////////////////////////////////////////////////////////////////////////////
/// CGlVboGeom
/// This is an abstract base class to hold 1 or more buffers needed to render
/// application geometry.  Subclasses implemnent the rendering functions
/// in ways that are compatible with specific OpenGL versions (1.1 and 2.0+)
///
class NCBI_GUIOPENGL_EXPORT CGlVboGeom : public CGlObject
{
public:
    /// Buffers are initially NULL and draw mode is not set.
    CGlVboGeom(): m_DrawMode(0xFFFFFFFF), m_VertexCount(-1) {}

    /// Delete all buffers
    virtual ~CGlVboGeom() {}
    virtual void Destroy() = 0;

    /// Return number of vertices in vertex buffer.  All in-use buffers
    /// should have same size.  (number of primitives depends on draw mode,
    /// e.g. if mode==GL_LINE_STRIP and GetSize()= 4, primitives==3
    virtual GLsizei GetSize() = 0;

    /// Set/get num vertices to render.  Defaults to # of verts in vertex buffer
    void SetVertexCount(int vcount) { m_VertexCount = vcount; }
    int GetVertexCount()
    { return (m_VertexCount == -1) ? GetSize() : m_VertexCount; }

    /// Render data in buffers
    virtual void Render() = 0;

    /// Set draw mode for primitive type (GL_LINES, GL_TRIANGLES, etc)
    virtual void SetDrawMode(GLenum m) { m_DrawMode = m; }
    virtual GLenum GetDrawMode() const { return m_DrawMode; }

    /// Pre-defined buffers.  Note that these are the only buffers that are
    /// supported by an OpenGL 1.1 implementation.
    virtual CGlRenderBuffer<CRgbaColor>&  GetColorBuffer() = 0;
    virtual void SetColorBuffer(CRef<CGlBuffer<CRgbaColor> > cb) = 0;

    virtual CGlRenderBuffer<CRgbaColor>&  GetSecondaryColorBuffer() = 0;
    virtual void SetSecondaryColorBuffer(CRef<CGlBuffer<CRgbaColor> > cb) = 0;

    virtual CGlRenderBuffer<CVect3<float> >&  GetNormalBuffer() = 0;
    virtual void SetNormalBuffer(CRef<CGlBuffer<CVect3<float> > > nb) = 0;

    virtual CGlRenderBuffer<CVect2<float> >&  GetTexCoordBuffer() = 0;
    virtual void SetTexCoordBuffer(CRef<CGlBuffer<CVect2<float> > > tb) = 0;

    virtual CGlRenderBuffer<CVect2<float> >&  GetVertexBuffer2D() = 0;
    virtual void SetVertexBuffer2D(CRef<CGlBuffer<CVect2<float> > > vb) = 0;

    virtual CGlRenderBuffer<CVect3<float> >&  GetVertexBuffer3D() = 0;
    virtual void SetVertexBuffer3D(CRef<CGlBuffer<CVect3<float> > > vb) = 0;

protected:
    
    /// Enable the buffers for rendering
    virtual void x_EnableBuffers(vector<GLenum>& attrib_states, int num_verts) = 0;

    /// Drawmode based on how vertices are organized and buffer subtype:
    ///  e.g. GL_LINES, GL_LINE_STRIP, GL_LINE_LOOP, GL_TRIANGLES, GL_POINTS etc
    GLenum m_DrawMode;

    /// Number of vertices to render.  This defaults to -1, in which case the
    /// size of the vertex array is used.
    int m_VertexCount;
};

//////////////////////////////////////////////////////////////////////////////
/// CGlVboGeom20
/// This class holds one or  more OpenGL vertex buuffer objects and implements
/// the CGlVboGeom interface.
///
/// Currently this class only supports the standard buffers supported by OpenGL
/// 1.1 and does not make use of the OpenGL attribute buffers (which are 
/// designed to work only with shaders, not the fixed function pipeline)
class NCBI_GUIOPENGL_EXPORT CGlVboGeom20 : public CGlVboGeom
{
public:
    /// Buffers are initially NULL and draw mode is not set.
    CGlVboGeom20();
    /// Delete all buffers
    virtual ~CGlVboGeom20();

    /// Delete data in all buffers.  Buffers can be re-created via BufferData
    virtual void Destroy();

    /// Return number of vertices in vertex buffer.  All in-use buffers
    /// should have same size.  (number of primitives depends on draw mode,
    /// e.g. if mode==GL_LINE_STRIP and GetSize()= 4, primitives==3
    virtual GLsizei GetSize();

    /// Render using fixed-function pipeline
    virtual void Render();

    ///Pre-defined buffers supported by OpenGL 1.1 and above
    CGlRenderBuffer<CRgbaColor>&  GetColorBuffer();
    void SetColorBuffer(CRef<CGlBuffer<CRgbaColor> > cb) 
        { m_ColorBuffer.SetBuffer(cb); }

    CGlRenderBuffer<CRgbaColor>&  GetSecondaryColorBuffer();
    void SetSecondaryColorBuffer(CRef<CGlBuffer<CRgbaColor> > cb) 
        { m_SecondaryColorBuffer.SetBuffer(cb); }

    CGlRenderBuffer<CVect3<float> >&  GetNormalBuffer();
    void SetNormalBuffer(CRef<CGlBuffer<CVect3<float> > > nb) 
        { m_NormalBuffer.SetBuffer(nb); }

    CGlRenderBuffer<CVect2<float> >&  GetTexCoordBuffer();
    void SetTexCoordBuffer(CRef<CGlBuffer<CVect2<float> > > tb) 
        { m_TexCoordBuffer.SetBuffer(tb); }

    CGlRenderBuffer<CVect2<float> >&  GetVertexBuffer2D();
    void SetVertexBuffer2D(CRef<CGlBuffer<CVect2<float> > > vb) 
        { m_VertexBuffer2D.SetBuffer(vb); }

    CGlRenderBuffer<CVect3<float> >&  GetVertexBuffer3D();
    void SetVertexBuffer3D(CRef<CGlBuffer<CVect3<float> > > vb) 
        { m_VertexBuffer3D.SetBuffer(vb); }

    /// User-defined attribute buffers, supported by OpenGL 2.0 and above, 
    /// including OpenGL ES 2.0
    CGlRenderBuffer<float>&            GetFloatBuffer(const string& attrib_name);
    CGlRenderBuffer<CVect2<float> >&   GetVec2Buffer(const string& attrib_name);
    CGlRenderBuffer<CVect3<float> >&   GetVec3Buffer(const string& attrib_name);
    CGlRenderBuffer<CVect4<float> >&   GetVec4Buffer(const string& attrib_name);
    CGlRenderBuffer<CMatrix3<float> >& GetMat33Buffer(const string& attrib_name);
    CGlRenderBuffer<CMatrix4<float> >& GetMat44Buffer(const string& attrib_name);

protected:
    
    /// Call gl{...}Pointer functions (e.g. glColorPointer) for each buffer
    /// that is currently in use to set all buffers up for rendering.
    void x_EnableBuffers(vector<GLenum>& attrib_states, int num_verts);
    
    /// TBD (rendering with shaders)
    ///void x_EnableShaderBuffers();

    /// Fixed function buffers (for backward compatibility (for now...))
    CGlRenderBuffer<CRgbaColor>     m_ColorBuffer;
    CGlRenderBuffer<CRgbaColor>     m_SecondaryColorBuffer;
    CGlRenderBuffer<CVect3<float> > m_NormalBuffer;
    CGlRenderBuffer<CVect2<float> > m_TexCoordBuffer;
    CGlRenderBuffer<CVect2<float> > m_VertexBuffer2D;
    CGlRenderBuffer<CVect3<float> > m_VertexBuffer3D;

    /// Vertex attribute buffers, allows setting up arbitrary vertex attributes
    /// when rendering using shaders
    vector<CGlRenderBuffer<float> >             m_FloatBuffers;
    vector<CGlRenderBuffer<CVect2<float> > >    m_Vec2Buffers;
    vector<CGlRenderBuffer<CVect3<float> > >    m_Vec3Buffers;
    vector<CGlRenderBuffer<CVect4<float> > >    m_Vec4Buffers;
    vector<CGlRenderBuffer<CMatrix3<float> > >  m_Mat33Buffers;
    vector<CGlRenderBuffer<CMatrix4<float> > >  m_Mat44Buffers;
};

//////////////////////////////////////////////////////////////////////////////
/// CGlVboGeom11
/// This class holds one or  more OpenGL vertex buffers implemented as
/// c++ arrays.  This is fully compatible with OpenGL 1.1
class NCBI_GUIOPENGL_EXPORT CGlVboGeom11 : public CGlVboGeom
{
public:
    /// Buffers are initially NULL and draw mode is not set.
    CGlVboGeom11();
    /// Delete all buffers
    virtual ~CGlVboGeom11();

    /// Delete all buffer data. Can be re-created via their BufferData() method
    virtual void Destroy();

    /// Return number of vertices in vertex buffer.  All in-use buffers
    /// should have same size.  (number of primitives depends on draw mode,
    /// e.g. if mode==GL_LINE_STRIP and GetSize()= 4, primitives==3
    virtual GLsizei GetSize();

    /// Render using fixed-function pipeline
    virtual void Render();

    ///Pre-defined buffers (supported in immediate-mode GL)
    CGlRenderBuffer<CRgbaColor>&  GetColorBuffer();
    void SetColorBuffer(CRef<CGlBuffer<CRgbaColor> > cb) { m_ColorBuffer.SetBuffer(cb); }

    CGlRenderBuffer<CRgbaColor>&  GetSecondaryColorBuffer();
    void SetSecondaryColorBuffer(CRef<CGlBuffer<CRgbaColor> > cb) { m_SecondaryColorBuffer.SetBuffer(cb); }

    CGlRenderBuffer<CVect3<float> >&  GetNormalBuffer();
    void SetNormalBuffer(CRef<CGlBuffer<CVect3<float> > > nb) { m_NormalBuffer.SetBuffer(nb); }

    CGlRenderBuffer<CVect2<float> >&  GetTexCoordBuffer();
    void SetTexCoordBuffer(CRef<CGlBuffer<CVect2<float> > > tb) { m_TexCoordBuffer.SetBuffer(tb); }


    CGlRenderBuffer<CVect2<float> >&  GetVertexBuffer2D();
    void SetVertexBuffer2D(CRef<CGlBuffer<CVect2<float> > > vb) { m_VertexBuffer2D.SetBuffer(vb); }

    CGlRenderBuffer<CVect3<float> >&  GetVertexBuffer3D();
    void SetVertexBuffer3D(CRef<CGlBuffer<CVect3<float> > > vb) { m_VertexBuffer3D.SetBuffer(vb); }

protected:
    
    /// Call gl{...}Pointer functions (e.g. glColorPointer) for each buffer
    /// that is currently in use to set all buffers up for rendering.
    void x_EnableBuffers(vector<GLenum>& attrib_states, int num_verts);

    /// 'unmaps' buffers (so we don't have stale pointers to arrays lying around)
    void x_DisableBuffers();


    /// Fixed function buffers to work with OpenGL 1.1 and later
    CGlRenderBuffer<CRgbaColor>     m_ColorBuffer;
    CGlRenderBuffer<CRgbaColor>     m_SecondaryColorBuffer;
    CGlRenderBuffer<CVect3<float> > m_NormalBuffer;
    CGlRenderBuffer<CVect2<float> > m_TexCoordBuffer;
    CGlRenderBuffer<CVect2<float> > m_VertexBuffer2D;
    CGlRenderBuffer<CVect3<float> > m_VertexBuffer3D;
};


END_NCBI_SCOPE

/* @} */

#endif  // GUI_OPENGL___GL_VBO_GEOM__HPP
