#ifndef GUI_OPENGL___GL_VBO_NODE__HPP
#define GUI_OPENGL___GL_VBO_NODE__HPP

/*  $Id: glvbonode.hpp 25176 2012-01-31 14:06:29Z falkrb $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Bob Falk
 *
 * File Description:
 *
 */

#include <gui/opengl.h>
#include <gui/gui.hpp>

#include <gui/utils/vect2.hpp>
#include <gui/utils/vect3.hpp>
#include <gui/utils/vect4.hpp>
#include <gui/utils/matrix3.hpp>
#include <gui/utils/matrix4.hpp>

#include <gui/utils/rgba_color.hpp>
#include <gui/opengl/globject.hpp>
#include <gui/opengl/glrendernode.hpp>
#include <gui/opengl/glbuffer.hpp>
#include <gui/opengl/glvbogeom.hpp>


/** @addtogroup GUI_OPENGL
 *
 * @{
 */

BEGIN_NCBI_SCOPE

//////////////////////////////////////////////////////////////////////////////
/// CGlVboNode
/// A rendering node that holds a vertex buffer object.  The vbo object
/// includes both the vertices themselves as well as any associated attributes
/// like color, texture coords etc.
///
class NCBI_GUIOPENGL_EXPORT CGlVboNode : public CGlRenderNode
{
public:
    /// Buffers are initially NULL and draw mode is not set.
    CGlVboNode();

    // Create the buffer based on draw type (as point, line, tri or quad vbo)
    // @param draw_mode
    //   One of: GL_{POINTS, LINES, LINE_STRIP, LINE_LOOP, 
    //               TRIANGLES, TRIANGLE_STRIP TRIANGLE_FAN, QUADS, QUAD_STRIP}
    CGlVboNode(GLenum draw_mode);

    /// Delete all buffers
    virtual ~CGlVboNode();

    /// Delete data in all buffers.  Buffers can be re-created via BufferData
    virtual void Destroy();

    /// Set/get geometry object.  These can be shared between nodes.
    void SetVBOGeom(CRef<CGlVboGeom> g) { m_VBOGeom = g; }
    CRef<CGlVboGeom> GetVBOGeom() { return m_VBOGeom; }

    /// Return number of vertices in vertex buffer.  All in-use buffers
    /// should have same size.  (number of primitives depends on draw mode,
    /// e.g. if mode==GL_LINE_STRIP and GetSize()= 4, primitives==3
    GLsizei GetSize() {return (m_VBOGeom==NULL) ? 0 : m_VBOGeom->GetSize();}

    //void SetScale(const CVect3<float>& f) {m_Scale = f; }

    // set appropriate draw mode for primitive type (GL_LINES, GL_TRIS, etc)
    virtual void SetDrawMode(GLenum m);
    /// Return the current drawing mode (e.g. GL_TRIANGLES) or GL_INVALID_ENUM
    /// if the geometry is not currently defined.
    GLenum GetDrawMode() const {return (m_VBOGeom==NULL) ? GL_INVALID_ENUM : m_VBOGeom->GetDrawMode();}
    
    /// Get the vertex buffer (holds the actual position data).  May return NULL
    /// if this is not a 2D buffer.
    CGlRenderBuffer<CVect2<float> >& GetVertexBuffer2D();
    /// Get the vertex buffer (holds the actual position data).  May return NULL
    /// if this is not a 3D buffer.
    CGlRenderBuffer<CVect3<float> >& GetVertexBuffer3D();

    /*
     * Buffers for pre-defined attributes.  All attribute buffers are optional,
     * so the buffer is created on the first call to the function to get the buffer.
     */
    CGlRenderBuffer<CRgbaColor>&  GetColorBuffer();
    CGlRenderBuffer<CRgbaColor>&  GetSecondaryColorBuffer();
    CGlRenderBuffer<CVect3<float> >&  GetNormalBuffer();
    CGlRenderBuffer<CVect2<float> >&  GetTexCoordBuffer();

    /*
     * Buffers for user-defined attributes which can (only) be used with shaders.  
     * The buffer is created the first time the 'Get' function is called for
     * a buffer with a given name.
     */
    /*
    CGlRenderBuffer<float>&            GetFloatBuffer(const string& attrib_name);
    CGlRenderBuffer<CVect2<float> >&   GetVec2Buffer(const string& attrib_name);
    CGlRenderBuffer<CVect3<float> >&   GetVec3Buffer(const string& attrib_name);
    CGlRenderBuffer<CVect4<float> >&   GetVec4Buffer(const string& attrib_name);
    CGlRenderBuffer<CMatrix3<float> >& GetMat33Buffer(const string& attrib_name);
    CGlRenderBuffer<CMatrix4<float> >& GetMat44Buffer(const string& attrib_name);
    */
protected:

    /// Render VBO(s)
    virtual void x_Render();
    
    /// subclass for vboset reprerenging point, line, tri or quad geometry
    CRef<CGlVboGeom> m_VBOGeom;
};

END_NCBI_SCOPE

/* @} */

#endif  // GUI_OPENGL___GL_VBO_NODE__HPP
