/*  $Id: pubdesc_editor.hpp 27132 2012-12-26 20:23:14Z filippov $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data,  the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties,  express or implied,  including
 *  warranties of performance,  merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Igor Filippov
 */
#ifndef _PUBDESC_EDITOR_H_
#define _PUBDESC_EDITOR_H_

#include <corelib/ncbistd.hpp>
#include <objmgr/seq_entry_handle.hpp>
#include <gui/objutils/cmd_composite.hpp>
#include <objects/biblio/Auth_list.hpp>
#include <objects/biblio/Author.hpp>
#include <objects/general/Person_id.hpp>
#include <objects/general/Name_std.hpp>
#include <objects/misc/sequence_macros.hpp>
#include <objmgr/feat_ci.hpp>
#include <gui/packages/pkg_sequence_edit/capitalization_abbreviation.hpp>

BEGIN_NCBI_SCOPE
using namespace objects;

class CPubDescEditor
{
public:
    bool apply(CSeq_entry_Handle tse, ICommandProccessor* cmdProcessor, string title);
    virtual void apply_to_seq_and_feat(CSeq_entry_Handle tse, CCmdComposite* composite);
    void ApplyToCSeq_entry (const CSeq_entry& se, CCmdComposite* composite);
    virtual void ApplyToPub (const CPubdesc& pubdesc, CCmdComposite* composite, const CSeq_entry* se, const CSeqdesc* seqDesq, const CSeq_feat* seqFeat);
    virtual ~CPubDescEditor() {}
    virtual void ApplyToPubdesc (CPubdesc& pubdesc, bool& modified) = 0;
protected:
    CSeq_entry_Handle m_tse;
};

class CAuthorEditor : public CPubDescEditor
{
public:
    virtual void ApplyToPubdesc (CPubdesc& pubdesc, bool& modified);
    void ApplyToPubequiv(CPub_equiv& pe, bool& modified);
    string get_first_initial(string first);
    void fix_initials(CPerson_id& pid, bool& modified);
    virtual void ApplyToCAuth(CAuth_list& authors, bool& modified) = 0;
};


class CReverseAuthorNames : public CAuthorEditor
{
public:
    virtual void ApplyToCAuth(CAuth_list& authors, bool& modified);
    void update_initials(CPerson_id& pid, string first, string last);
};

class CFixAuthorInitials : public CAuthorEditor
{
public:
    virtual void ApplyToCAuth(CAuth_list& authors, bool& modified);
};

class CAffilEditor : public CAuthorEditor
{
public:
    virtual void ApplyToCAuth(CAuth_list& authors, bool& modified);
    virtual void ApplyToCAffil(CAffil::TStd& std, bool& modified) = 0;
};

class CFixAuthorCap : public CAuthorEditor, public CCapitalizationAbbreviations
{
public:
    virtual void ApplyToCAuth(CAuth_list& authors, bool& modified);
    string fix_name_cap(string name, bool bApostrophes, bool& modified);
};

class CFixUSA : public CAffilEditor, public CCapitalizationAbbreviations
{
public:
    virtual void ApplyToCAffil(CAffil::TStd& std, bool& modified);
  
};

class CFixTitleCap : public CPubDescEditor, public CCapitalizationAbbreviations
{
public:
    virtual void ApplyToPubdesc (CPubdesc& pubdesc, bool& modified);
    string getTitle(const CTitle& title);
    void setTitle(CTitle& title, string str);
    string fix_title(string title, bool& modified);
    void set_seh(CSeq_entry_Handle tse);
};

class CFixCountryCap : public CAffilEditor, public CCapitalizationAbbreviations
{
public:
    virtual void ApplyToCAffil(CAffil::TStd& std, bool& modified);
};

class CFixAffilCap  : public CAffilEditor, public CCapitalizationAbbreviations
{
public:
    virtual void ApplyToCAffil(CAffil::TStd& std, bool& modified);
};

class CFixAllCap : public CPubDescEditor
{
    virtual void ApplyToPub(const CPubdesc& pubdesc, CCmdComposite* composite, const CSeq_entry* se, const CSeqdesc* seqDesq, const CSeq_feat* seqFeat);
    virtual void ApplyToPubdesc (CPubdesc& pubdesc, bool& modified) {}
};


class CRemovePubs : public CPubDescEditor
{
public:
    virtual void ApplyToPubdesc (CPubdesc& pubdesc, bool& modified) = 0;
    virtual void ApplyToPub (const CPubdesc& pubdesc, CCmdComposite* composite, const CSeq_entry* se, const CSeqdesc* seqDesq, const CSeq_feat* seqFeat);
    void DeleteObject(CSeq_entry_Handle& seh, const CSeq_entry* se, const CSeqdesc& seqDesq, CCmdComposite* composite);
    void DeleteObject(CSeq_entry_Handle& seh, const CSeq_feat& seqFeat, CCmdComposite* composite);
};


class CRemoveUnpublishedPubs : public CRemovePubs
{
public:
    virtual void ApplyToPubdesc (CPubdesc& pubdesc, bool& modified);
};

class CRemovePublishedPubs : public CRemovePubs
{
public:
    virtual void ApplyToPubdesc (CPubdesc& pubdesc, bool& modified);
};

class CRemoveCollidingPubs : public CRemovePubs
{
public:
    virtual void ApplyToPubdesc (CPubdesc& pubdesc, bool& modified);
    virtual void apply_to_seq_and_feat(CSeq_entry_Handle tse, CCmdComposite* composite);
    void apply_to_seq_and_feat_single_pass(CSeq_entry_Handle tse, CCmdComposite* composite);
protected:
    bool m_SecondPass;
    map<string,int> m_collisions;
};

class CRemoveAllPubs : public CRemoveCollidingPubs
{
public:
    virtual void ApplyToPub(const CPubdesc& pubdesc, CCmdComposite* composite, const CSeq_entry* se, const CSeqdesc* seqDesq, const CSeq_feat* seqFeat);
};

class CStripAuthorSuffix : public CAuthorEditor
{
public:
    virtual void ApplyToCAuth(CAuth_list& authors, bool& modified);
};

class CRemoveAuthorConsortium : public CAuthorEditor
{
public:
    virtual void ApplyToCAuth(CAuth_list& authors, bool& modified);
};

class CConvertAllAuthorToConsortium : public CAuthorEditor
{
public:
    virtual void ApplyToCAuth(CAuth_list& authors, bool& modified);
};

class CConvertAuthorToConsortiumWhereConstraint : public CAuthorEditor
{ 
public:
  void SetConstraint(const string &input);
  virtual void ApplyToCAuth(CAuth_list& authors, bool& modified);
  virtual ~CConvertAuthorToConsortiumWhereConstraint();
private:
  CRegexp* m_Regex;
};

class CReloadPublications : public CPubDescEditor
{
public:
    virtual void ApplyToPubdesc (CPubdesc& pubdesc, bool& modified);
    void LookUpByPmid(CPub& pub, int pmid, bool& modified);
};

END_NCBI_SCOPE

#endif
    // _PUBDESC_EDITOR_H_
