/*  $Id: srcedit_util.hpp 27196 2013-01-09 15:42:41Z bollin $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data,  the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties,  express or implied,  including
 *  warranties of performance,  merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Colleen Bollin
 */
#ifndef _SRCEDIT_UTIL_H_
#define _SRCEDIT_UTIL_H_

#include <corelib/ncbistd.hpp>
#include <objmgr/seq_entry_handle.hpp>

#include <objects/seqfeat/OrgMod.hpp>
#include <objects/seqfeat/SubSource.hpp>
#include <objects/seqtable/Seq_table.hpp>
#include <objects/seqtable/SeqTable_column.hpp>
#include <objects/seqtable/SeqTable_multi_data.hpp>
#include <objects/seqtable/SeqTable_column_info.hpp>


#include <gui/widgets/wx/wx_utils.hpp>
#include <gui/widgets/grid_widget/grid_widget.hpp>
#include <gui/packages/pkg_sequence_edit/srcedit_panel.hpp>
#include <gui/objutils/table_data.hpp>


#include <gui/objutils/cmd_composite.hpp>

BEGIN_NCBI_SCOPE


class CSrcQual;
class CSourceRequirements;
typedef vector<CSrcQual *> TSrcQualList;


class CSrcQual 
{
public:
    /// Constructors
    CSrcQual();
    CSrcQual(string name, bool required, bool report_missing, string example = "");
    virtual string GetValue(const objects::CBioSource& src);
    virtual string GetFormatErrors(string value) {return ""; };
    virtual string GetExample() { return m_Example; };
    void SetExample(string example) { m_Example = example; };

    string GetName(void) { return m_Name; };
    bool IsRequired(void) { return m_Required; };
    bool ReportMissing(void) { return m_ReportMissing; };
    void SetRequired(bool required) { m_Required = required; };
    void SetReportMissing(bool report_missing) { m_ReportMissing = report_missing; };

    /// Destructor
    virtual ~CSrcQual();

protected:
    string m_Name;
    string m_Example;
    bool   m_Required;
    bool   m_ReportMissing;
};

class CTaxnameQual: public CSrcQual
{
public:
    /// Constructors
    CTaxnameQual() : CSrcQual() {};
    CTaxnameQual(string name, bool required, bool report_missing = false, string example = "") : CSrcQual(name, required, report_missing, example) {};
    virtual string GetValue(const objects::CBioSource& src);
    virtual string GetFormatErrors(string value) { return ""; };

    /// Destructor
    virtual ~CTaxnameQual() {};
};


class CUnculturedTaxnameQual: public CSrcQual
{
public:
    /// Constructors
    CUnculturedTaxnameQual() : CSrcQual() {};
    CUnculturedTaxnameQual(string name, bool required, bool report_missing = false, string example = "") : CSrcQual(name, required, report_missing, example) {};
    virtual string GetFormatErrors(string value); 

    /// Destructor
    virtual ~CUnculturedTaxnameQual() {};
};


class CGenomeQual: public CSrcQual
{
public:
    /// Constructors
    CGenomeQual() : CSrcQual() {};
    CGenomeQual(string name, bool required, bool report_missing = false, string example = "") : CSrcQual(name, required, report_missing, example) {};
    virtual string GetValue(const objects::CBioSource& src);
    virtual string GetFormatErrors(string value) { return ""; };

    /// Destructor
    virtual ~CGenomeQual() {};
};


class COrgModQual: public CSrcQual
{
public:
    /// Constructors
    COrgModQual() : CSrcQual(), m_Subtype(objects::COrgMod::eSubtype_other) {};
    COrgModQual(string name, bool required, bool report_missing = false, string example = "");
    virtual string GetValue(const objects::CBioSource& src);
    virtual string GetFormatErrors(string value);

    /// Destructor
    virtual ~COrgModQual() {};
private:
    objects::COrgMod::TSubtype m_Subtype;
};


class CSubSrcQual: public CSrcQual
{
public:
    /// Constructors
    CSubSrcQual() : CSrcQual(), m_Subtype(objects::CSubSource::eSubtype_other) {};
    CSubSrcQual(string name, bool required, bool report_missing = false, string example = "");
    virtual string GetValue(const objects::CBioSource& src);
    virtual string GetFormatErrors(string value);

    /// Destructor
    virtual ~CSubSrcQual() {};
private:
    objects::CSubSource::TSubtype m_Subtype;
    vector<string> m_BadIsolationSourceValues;
};


class CSourceRequirements
{
public:
    /// Constructors
    CSourceRequirements();
    void AddRequirement(string name, bool required, string example = "");
    void PreferentiallyAddRequirement(CRef<objects::CSeq_table> values_table, string choice1, string choice2, bool required);
    void AddUniquenessList(vector<string> list);
    void AddUniquenessList(int num, ...);
    void AddOneOfList(vector<string> list);
    void AddOneOfList(int num, ...);
    CRef<objects::CSeqTable_column> CheckSourceQuals(CRef<objects::CSeq_table> values_table);
    void AddColumnsToSeqTable(CRef<objects::CSeq_table> values_table);

    /// Destructor
    virtual ~CSourceRequirements();


    enum EWizardType {
        eWizardType_standard            =  0,
        eWizardType_viruses             =  1,
        eWizardType_uncultured_samples  =  2,
        eWizardType_rrna_its_igs        =  3, 
        eWizardType_tsa                 =  4,
        eWizardType_igs                 =  5,
        eWizardType_microsatellite      =  6,
        eWizardType_d_loop              =  7
    };

    enum EWizardSrcType {
        eWizardSrcType_any                  = 0,
        eWizardSrcType_virus_norovirus      = 1,
        eWizardSrcType_virus_foot_and_mouth = 2,
        eWizardSrcType_virus_influenza      = 3,
        eWizardSrcType_virus_rotavirus      = 4,
        eWizardSrcType_bacteria_or_archaea  = 5,
        eWizardSrcType_uncultured_fungus    = 6,
        eWizardSrcType_cultured_fungus      = 7,
        eWizardSrcType_vouchered_fungus     = 8,
        eWizardSrcType_plant                = 9,
        eWizardSrcType_animal               = 10
    };

    void SetExamples(EWizardType wizard_type, EWizardSrcType src_type);

private:
    TSrcQualList m_Requirements;
    // if any uniqueness lists are specified, at least one of the combos listed must produce a unique identifier for each sequence
    vector<vector<string> > m_UniquenessLists;
    vector<vector<string> > m_OneOfLists;

    int x_AddUniquenessProblems(CRef<objects::CSeq_table> values_table, vector<string> uniqueness_list, vector<string>& row_problems);
    void x_AddOneOfProblems(CRef<objects::CSeq_table>, vector<string> one_of_list, CRef<objects::CSeqTable_column> problems);

};

// TODO: Should these classes be just in the .cpp file?  Are they useful
// anywhere else?

class CSrcTableColumnBase : public CObject
{
public:
    // default is to do no command.  Subclasses will almost always override this
    virtual void AddToBioSource(
        objects::CBioSource & in_out_bioSource, const string & newValue ) { }
    virtual void ClearInBioSource(
        objects::CBioSource & in_out_bioSource ) { }
    virtual string GetFromBioSource(
        const objects::CBioSource & in_out_bioSource ) { return ""; }
    virtual string GetLabel() { return ""; }
};

class CSrcTableOrganismNameColumn : public CSrcTableColumnBase
{
public:
    virtual void AddToBioSource(
        objects::CBioSource & in_out_bioSource, const string & newValue );
    virtual void ClearInBioSource(
        objects::CBioSource & in_out_bioSource );
    virtual string GetFromBioSource(
        const objects::CBioSource & in_out_bioSource );
    virtual string GetLabel() { return "Organism Name"; }
};

class CSrcTableGenomeColumn : public CSrcTableColumnBase
{
public:
    virtual void AddToBioSource(
        objects::CBioSource & in_out_bioSource, const string & newValue );
    virtual void ClearInBioSource(
        objects::CBioSource & in_out_bioSource );
    virtual string GetFromBioSource(
        const objects::CBioSource & in_out_bioSource );
    virtual string GetLabel() { return "Organism Name"; }
};

class CSrcTableSubSourceColumn : public CSrcTableColumnBase
{
public:
  CSrcTableSubSourceColumn(objects::CSubSource::TSubtype subtype) { m_Subtype = subtype; };
    virtual void AddToBioSource(
        objects::CBioSource & in_out_bioSource, const string & newValue );
    virtual void ClearInBioSource(
        objects::CBioSource & in_out_bioSource );
    virtual string GetFromBioSource(
        const objects::CBioSource & in_out_bioSource );
    virtual string GetLabel() { return objects::CSubSource::GetSubtypeName(m_Subtype); }
private:
    objects::CSubSource::TSubtype m_Subtype;
};

class CSrcTableOrgModColumn : public CSrcTableColumnBase
{
public:
  CSrcTableOrgModColumn(objects::COrgMod::TSubtype subtype) { m_Subtype = subtype; };
    virtual void AddToBioSource(
        objects::CBioSource & in_out_bioSource, const string & newValue );
    virtual void ClearInBioSource(
        objects::CBioSource & in_out_bioSource );
    virtual string GetFromBioSource(
        const objects::CBioSource & in_out_bioSource );
    virtual string GetLabel() { return objects::COrgMod::GetSubtypeName(m_Subtype); }
private:
    objects::COrgMod::TSubtype m_Subtype;
};


class CSrcTableFwdPrimerSeqColumn : public CSrcTableColumnBase
{
public:
    CSrcTableFwdPrimerSeqColumn() { };
    virtual void AddToBioSource(
        objects::CBioSource & in_out_bioSource, const string & newValue );
    virtual void ClearInBioSource(
        objects::CBioSource & in_out_bioSource );
    virtual string GetFromBioSource(
        const objects::CBioSource & in_out_bioSource );
    virtual string GetLabel() { return "fwd-primer-seq"; }
};


class CSrcTableRevPrimerSeqColumn : public CSrcTableColumnBase
{
public:
    CSrcTableRevPrimerSeqColumn() { };
    virtual void AddToBioSource(
        objects::CBioSource & in_out_bioSource, const string & newValue );
    virtual void ClearInBioSource(
        objects::CBioSource & in_out_bioSource );
    virtual string GetFromBioSource(
        const objects::CBioSource & in_out_bioSource );
    virtual string GetLabel() { return "rev-primer-seq"; }
};


class CSrcTableFwdPrimerNameColumn : public CSrcTableColumnBase
{
public:
    CSrcTableFwdPrimerNameColumn() { };
    virtual void AddToBioSource(
        objects::CBioSource & in_out_bioSource, const string & newValue );
    virtual void ClearInBioSource(
        objects::CBioSource & in_out_bioSource );
    virtual string GetFromBioSource(
        const objects::CBioSource & in_out_bioSource );
    virtual string GetLabel() { return "fwd-primer-name"; }
};


class CSrcTableRevPrimerNameColumn : public CSrcTableColumnBase
{
public:
    CSrcTableRevPrimerNameColumn() { };
    virtual void AddToBioSource(
        objects::CBioSource & in_out_bioSource, const string & newValue );
    virtual void ClearInBioSource(
        objects::CBioSource & in_out_bioSource );
    virtual string GetFromBioSource(
        const objects::CBioSource & in_out_bioSource );
    virtual string GetLabel() { return "rev-primer-name"; }
};


class CSrcTableColumnBaseFactory
{
public:
    static CRef<CSrcTableColumnBase> Create(const objects::CSeqTable_column &column);
};

CRef<objects::CSeq_table> GetSeqTableFromSeqEntry (objects::CSeq_entry_Handle seh);
CRef<objects::CSeq_table> GetSeqTableForSrcQualFromSeqEntry (objects::CSeq_entry_Handle seh, vector<string> qual_names);
CRef<CCmdComposite> ApplySrcTableToSeqEntry (CRef<objects::CSeq_table>values_table, objects::CSeq_entry_Handle seh);
CSourceRequirements *GetSrcRequirements(CSourceRequirements::EWizardType wizard_type, 
                                        CSourceRequirements::EWizardSrcType source_type, 
                                        CRef<objects::CSeq_table> values_table);
bool IsFwdPrimerName (string name);
bool IsRevPrimerName (string name);
bool IsFwdPrimerSeq (string name);
bool IsRevPrimerSeq (string name);


END_NCBI_SCOPE

#endif
    // _SRCEDIT_UTIL_H_
