#ifndef UTIL__MACRO_PARSER_HPP__
#define UTIL__MACRO_PARSER_HPP__

/*  $Id: macro_parse.hpp 24335 2011-09-15 13:50:47Z kuznets $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors: Anatoliy Kuznetsov
 *
 * File Description: Macro language parser
 *
 */

/// @file macro_parse.hpp
/// Query string parsing components

#include <corelib/ncbistl.hpp>
#include <corelib/ncbistd.hpp>

#include <gui/gui_export.h>
#include <util/qparse/query_parse.hpp>

BEGIN_NCBI_SCOPE

/** @addtogroup QParser
 *
 * @{
 */


///////////////////////////////////////////////////////////////////////////////
/// CAppJobException - Exception thrown by Job Dispatcher and Job Engines.
class NCBI_GUIOBJUTILS_EXPORT CMacroException
            : EXCEPTION_VIRTUAL_BASE public CException
{
public:
    enum    EErrCode {
        eInvalidVariableType,    ///< Incorrect variable type request
        eSyntaxError             ///< Macro syntax error
    };
    NCBI_EXCEPTION_DEFAULT(CMacroException, CException);
};


/// Class describes macro language for bulk bio-editing
///
class NCBI_GUIUTILS_EXPORT CQueryMacro
{
public:
    typedef CQueryParseNode::EType EType;

    /// CHOICE descriptor (list of options to choose from
    class CChoice
    {
    public:
        CChoice(bool multi=false);
        CChoice(const CChoice& choice);
        CChoice& operator=(const CChoice& choice);
    public:
        /// Returns true if multiple choice is available
        bool IsMultiple() const { return m_IsMulti; }

        const vector<string>& GetList() const { return m_ChoiceList; }
        void AddChoiceOption(const string& str) { m_ChoiceList.push_back(str); }

    private:
        vector<string>  m_ChoiceList; ///!< list of options
        bool            m_IsMulti;    ///!< multiple choice flag
    };

    /// Class describes a macro language variable
    class CVariable
    {
    public:

    public:
        CVariable();
        CVariable(const CVariable& var);
        CVariable& operator=(const CVariable& var);

        CVariable(const string& name, const string& value, bool is_ui_ask=false);
        CVariable(const string& name, const string& value, CQueryParseNode::EType type);
        explicit CVariable(const string& name, int value);
        explicit CVariable(const string& name, double value);
        explicit CVariable(const string& name, bool value);
        /// Constructor takes ptr memory ownership of choice pointer
        CVariable(const string& name, CChoice* choice);

        ~CVariable();

        /// @name Variable value accessors
        /// @{  
        
        /// Get name of the variable
        const string& GetName() const { return m_Name; }
        /// Get variable type
        EType GetType() const { return m_Type; }
        /// Get string value
        const string& GetValue() const { return m_Value; }
        /// Get identifier
        const string& GetIdent() const { return m_Value; }
        /// Get integer
        Int4 GetInt() const;
        /// Get bool
        bool GetBool() const;
        /// get double
        double GetDouble() const;
        
        /// @}

        /// @name UI specs
        /// @{
        /// Returns true if variable specified as UI %ASK% specified
        bool IsUI_Ask() const { return m_IsAsk; }
        /// Returns true if variable is UI selectable CHOICE
        bool IsUI_Choice() const { return m_IsChoice; }
        /// Get choice list. Throws an exception if variable is not a CHOICE
        const CChoice& GetChoice() const;
        /// @}

    private:
        string                  m_Name;
        CQueryParseNode::EType  m_Type;
        union {
            Int4         m_IntConst;
            bool         m_BoolConst;
            double       m_DoubleConst;
        };    
        string                  m_Value;
        bool                    m_IsAsk;
        bool                    m_IsChoice;
        CChoice*                m_Choice;
    };

    /// Typedef represents one block of variables
    typedef vector<CVariable> TVarBlock;    
    /// Collection of blocks of variables
    typedef vector<TVarBlock> TVarBlocksCollection;

    /// WHERE clause
    typedef CQueryParseTree CWhereClause;



    /// Macro type
    ///
    enum EMacroType {
        eUndefinedMacroType,   ///< Undefined macro type 
        eForEach               ///< FOR EACH macro 
    };

    /// Selector type
    enum ESelectorType {
        eUndefinedSelector,  ///< Undefined selector
        eSeqFeat             ///< Features iterator
    };



    /// Typedef represents one block of variables
    typedef vector<CVariable> TArgList;

    /// Class describes macro function
    class CMacroFunction
    {
    public:
        CMacroFunction();
        CMacroFunction(const string& func_name, const TArgList& args);
        CMacroFunction(const CMacroFunction& func);
        CMacroFunction& operator=(const CMacroFunction& func);
        ~CMacroFunction();

        /// Returns function name
        const string& GetName() const { return m_Name; }
        /// Returns list of arguments
        const TArgList& GetArgs() const { return m_ArgList; }

    private:
        string    m_Name;
        unsigned  m_Line;
        TArgList  m_ArgList;
    };

    /// List of function calls describing the DO...DONE block
    typedef vector<CMacroFunction> TDoList;

public:
    CQueryMacro();
    CQueryMacro(const CQueryMacro& qmacro);
    CQueryMacro& operator=(const CQueryMacro& qmacro);
    ~CQueryMacro();

    /// Parse the query string
    void Parse(const char* str);

    /// @name General macro parameters
    /// @{  
    /// Get macro name
    const string& GetMacroName() const;
    /// Get macro title
    const string& GetMacroTitle() const;
    /// Get macro type
    EMacroType GetMacroType() const;
    /// Get selector-iterator type
    ESelectorType GetSelectorType() const;

    /// @}  

    /// @name Macro variables 
    /// @{  
    /// Get total count of variable blocks in the macro
    size_t GetVarBlocksCount() const;

    /// Get variable blocks
    const TVarBlocksCollection& GetVarBlocks() const;

    /// Get variable by name 
    /// (throws an exception if variable is not registered)
    const CVariable& GetVariable(const string& var_name) const;
    /// @}

protected:
    TVarBlocksCollection& SetVarBlocks();

    size_t FindVariable(const TVarBlock& vars, const string& s);

private:
    void x_SkipCommentsOrWhiteSpaces(const string* other_skip_str=0);
    void x_ParseGetVarName(string* var_name);
    void x_ParseGetString(string* var_str, char bracket = '"');
    CQueryParseNode::EType x_ParseGetNumber(string* var_str);
    bool x_ParseGetBool();
    void x_ParseGetWhere(string* where_query);
    void x_ParseGetFunctionName(string* func_name);

private:
    string         m_MacroName;
    string         m_MacroTitle;
    EMacroType     m_MacroType;
    ESelectorType  m_SelectorType;
    TVarBlocksCollection  m_VarBlocks;
    string                m_WhereQuery;
    TDoList               m_DoList;

    // parsing FSM variables
    const char*   m_Curr;
    const char*   m_Start;
    bool          m_IsEnd;
};

inline 
const string& CQueryMacro::GetMacroName() const
{
    return m_MacroName;
}

inline 
const string& CQueryMacro::GetMacroTitle() const
{
    return m_MacroTitle;
}

inline
CQueryMacro::EMacroType CQueryMacro::GetMacroType() const
{
    return m_MacroType;
}

inline
CQueryMacro::ESelectorType CQueryMacro::GetSelectorType() const
{
    return m_SelectorType;
}

inline
size_t CQueryMacro::GetVarBlocksCount() const
{
    return m_VarBlocks.size();
}

inline
const CQueryMacro::TVarBlocksCollection& CQueryMacro::GetVarBlocks() const
{
    return m_VarBlocks;
}

inline
CQueryMacro::TVarBlocksCollection& CQueryMacro::SetVarBlocks()
{
    return m_VarBlocks;
}

/* @} */

END_NCBI_SCOPE


#endif  // UTIL__MACRO_PARSER_HPP__


