#ifndef GUI_OBJUTILS___SCHEDULER_ENGINE__HPP
#define GUI_OBJUTILS___SCHEDULER_ENGINE__HPP

/*  $Id: scheduler_engine.hpp 26649 2012-10-18 20:31:07Z katargir $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Andrey Yazhuk
 *
 * File Description:
 *
 */

#include <gui/utils/app_job_engine.hpp>

#include <util/sync_queue.hpp>
#include <util/scheduler.hpp>

#include <corelib/ncbithr.hpp>

BEGIN_NCBI_SCOPE

class NCBI_GUIUTILS_EXPORT IScheduledAppJob : public IAppJob
{
public:
    /// returns a period of time after which the task shall be executed again
    virtual CTimeSpan   GetWaitPeriod() = 0;
};

///////////////////////////////////////////////////////////////////////////////
/// CSchedulerEngine
/// CSchedulerEngine is an engine that can be used with CAppJobDispatcher.
/// The idea of CSchedulerEngine is to use a single background thread for
/// periodic execution of multiple jobs. This kind of engine is useful for
/// doing quick repetitive tasks such as polling external services every once
/// in while. All jobs share the same thread so they should execute really quick.
/// To be compatible with CSchedulerEngine a job needs to implement
/// IScheduledAppJob interface.

class NCBI_GUIUTILS_EXPORT CSchedulerEngine
    :   public CObject,
        public IAppJobEngine
{
    friend class CSchedulerEngineThread;
public:
    typedef IAppJob::EJobState TJobState;

    CSchedulerEngine();
    virtual ~CSchedulerEngine();

    /// @name IAppJobEngine implementation
    /// @{
    virtual bool    IsActive();

    virtual void    SetListener(IAppJobEngineListener* listener);

    void    StartJob(IAppJob& job, IEngineParams* /*params = NULL*/);

    virtual void    CancelJob(IAppJob& job);
    virtual void    SuspendJob(IAppJob& /*job*/);
    virtual void    ResumeJob(IAppJob& /*job*/);

    virtual TJobState   GetJobState(IAppJob& job) const;

    virtual void    ShutDown();
    virtual void    RequestCancel();
    /// @}

protected:
    enum    EAction {
        eInvalidAction = -1,
        eStart,
        eCancel,
        eShutDown
    };

    /// record that holds infromation about a job
    class CJobRecord :
        public CObject,
        public IScheduler_Task
    {
    public:
        CJobRecord(IScheduledAppJob& job);
        virtual void Execute();
    public:
        CIRef<IScheduledAppJob> m_Job;
        TJobState   m_JobState;
        TScheduler_SeriesID   m_TaskID;
    };

    /// signals that Engine send to its scheduling thread
    struct SSignal
    {
        EAction     m_Action;
        CJobRecord*  m_JobRecord;
    };

protected:
    typedef CRef<CJobRecord>    TRecordRef;
    typedef map<IScheduledAppJob*, TRecordRef>  TJobToRecord;
    typedef CSyncQueue<SSignal> TSignalQueue;

    /// performs job execution on a separate scheduling thread
    void    x_WorkOnThread();

    /// these functions shall be executed on the worker thread only
    /// otherwise synchronization problems will arise
    void    x_OnSignal_OnThread(SSignal& signal);
    void    x_ExecuteReadyTasks_OnThread();
    void    x_ScheduleTask_OnThread(CJobRecord& record);
    void    x_CancelTask_OnThread(CJobRecord& record);
    void    x_CancelAllTasks_OnThread();

    IScheduledAppJob*   x_GetScheduledJob(IAppJob& job);
    TJobToRecord::iterator  x_Find(IScheduledAppJob& job);
    TJobToRecord::const_iterator    x_Find(IScheduledAppJob& job) const;

    void    x_OnJobStatusChange(CJobRecord& record);

protected:
    mutable CMutex      m_Mutex;
    bool                m_Shutdown;

    IAppJobEngineListener*  m_Listener;

    /// job registry
    TJobToRecord        m_JobToRecord;

    /// statistics about jobs executed

    size_t  m_JobsScheduled;
    size_t  m_JobsFinished;
    size_t  m_JobsCanceled;

    /// queue for sending signals to the scheduling thread
    TSignalQueue    m_Signals;

    CIRef<IScheduler>   m_Scheduler;

    CRef<CThread>    m_WorkerThread;

    /// the members below should only be accessed by the worker thread

    CTime m_Now; // current time
    CTime m_WakeUpTime; // time to wake up to execute a job
};


END_NCBI_SCOPE

#endif  // GUI_OBJUTILS___SCHEDULER_ENGINE__HPP
