#ifndef __GUI_WIDGETS_ALNMULTI___ALNMULTI_RENDERER__HPP
#define __GUI_WIDGETS_ALNMULTI___ALNMULTI_RENDERER__HPP

/*  $Id: alnmulti_renderer.hpp 26132 2012-07-25 14:58:27Z shkeda $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Andrey Yazhuk
 *
 * File Description:
 *
 */

#include <corelib/ncbistl.hpp>


#include <gui/widgets/gl/ruler.hpp>

#include <objtools/alnmgr/aln_explorer.hpp>

#include <gui/widgets/aln_multiple/alnmulti_ds.hpp>
#include <gui/widgets/aln_multiple/ialign_row.hpp>
#include <gui/widgets/aln_multiple/alnmulti_header.hpp>

#include <gui/widgets/gl/graph_container.hpp>

BEGIN_NCBI_SCOPE


////////////////////////////////////////////////////////////////////////////
/// IAlnMultiRendererContext - this interface represents context in which
/// CAlnMultiRenderer lives. It provides callbacks necessary for Renderer's
/// functioning.

class NCBI_GUIWIDGETS_ALNMULTIPLE_EXPORT IAlnMultiRendererContext
{
public:
    typedef IAlnExplorer::TNumrow    TNumrow;
    typedef IAlignRow::TColumnType  TColumnType;

public:
    virtual ~IAlnMultiRendererContext()  {};

    virtual const CGlPane& GetAlignPort() const = 0;
    virtual TNumrow     GetLinesCount() = 0;
    virtual bool    IsRendererFocused() = 0;
    virtual int     GetFocusedItemIndex() const = 0;
    virtual bool    IsItemSelected(int index) const = 0;

    virtual IAlignRow*  GetMasterRow()= 0;

    /// all Y coordinates are OpenGL Viewport coordinates (not widget coords)
    /// "index" is a line index (not row in alignment)

    virtual IAlignRow*  GetRowByLine(int index) = 0;
    virtual int     GetLineByRowNum(TNumrow row) const = 0;
    virtual int     GetLineByModelY(int y) const = 0;
    virtual int     GetLinePosY(int index) const = 0;
    virtual int     GetLineHeight(int index) const = 0;

    virtual bool    SortByColumn(TColumnType type, bool ascending) = 0;
    virtual IAlnExplorer::ESortState   GetColumnSortState(TColumnType type) = 0;
};


////////////////////////////////////////////////////////////////////////////////
/// class CAlnMultiRenderer - renders multiple alignment represented by
/// IAlnMultiRendererContext (the Model).
///
/// IAlnMultiRenderer renders a grid where rows corresponds to IAlignRow objects
/// and columns represent different types of information. Columns can be
/// configured using "Column management" API.

class NCBI_GUIWIDGETS_ALNMULTIPLE_EXPORT CAlnMultiRenderer
        :   public IAlnMultiHeaderContext
{
public:
    enum EHtmlAreaTypes {
        eHtmlArea_Link,
        eHtmlArea_CheckBox
    };
    typedef IAlignRow::TColumnType      TColumnType;
    typedef vector<CHTMLActiveArea>     TAreaVector;

public:
    /// Constructor, with column add, icons setup and resize
    ///
    CAlnMultiRenderer(const TVPRect& rc);

    /// Empty object constructor -- for two-phase construction
    /// call SetupColums() to finish the class construction process
    CAlnMultiRenderer();
    virtual ~CAlnMultiRenderer();

    /// Adds default columns for the renderer, allows custom setup
    /// (column icon override, etc)
    ///
    virtual void SetupColumns(const TVPRect& rc);

    void    SetContext(IAlnMultiRendererContext* pContext);

    void    SetBackColor(const CRgbaColor Color);
    const CRgbaColor&    GetBackColor() const;

    void    SetMasterBackColor(const CRgbaColor Color);
    const CRgbaColor&    GetMasterBackColor() const;

    void    SetFocused(bool b_set)    {   m_bFocused  = b_set;  }

    virtual void    Resize(const TVPRect& rc);
    virtual TVPRect GetRect() const;

    virtual void    Update();
    virtual void    UpdateOnDataChanged();

    /// renders OpenGL graphics
    virtual void    Render();

    /// renders OpenGL graphics and generate CHTMLActiveArea objects
    virtual void    Render(TAreaVector& areas);

    virtual bool    NeedTooltip(int vp_x, int vp_y);
    virtual string  GetTooltip(); // returns tooltip for previously tested position

    CAlnMultiHeader&    GetHeader();

    CRuler&    GetRuler();
    CGlPane&   GetRulerPane() {   return m_RulerPane;   }

    CGlPane&   GetGraphPane() {   return m_GraphPane;   }

    int     GetRulerAreaHeight()    const   {   return m_RulerRect.Height();  }
    int     GetMasterAreaHeight()    const   {   return m_MasterRect.Height();  }
    int     GetListAreaHeight() const   {   return m_ListAreaRect.Height();    }
    int     GetHeaderAreaHeight() const   {   return m_HeaderRect.Height();    }
    TVPRect GetMasterArea() const;

    int     GetListTop() const;
    int     GetColumnIndexByType(TColumnType type) const;
    TVPRect         GetColumnRect(int i_col, bool include_header)   const;
    TVPRect         GetColumnRectByType(TColumnType type, bool include_header)   const;
    TColumnType     GetColumnTypeByIndex(int i_col)   const;
    TColumnType     GetColumnTypeByX(int vp_x)   const;

    /// return OpenGL viewport coordinate of the top pixel in the list area
    int     GetVPListTop() const;

    CGraphContainer& GetGraphContainer() {
        return m_GraphCont; 
    }

    enum EHitResult {
        eNone,
        eHeader,
        eRuler,
        eGraphCont,
        eMasterRow,
        eRows
    };

    virtual int  HitTest(int vp_x, int vp_y, int& col);

    void    SetupPaneForRow(CGlPane& pane, const IAlignRow* p_row,
                            int vp_top_y) const;
    void    SetupPaneForColumn(CGlPane& pane, int i_col) const;

    /// @name IAlnMultiHeaderContext implementation
    /// @{
    virtual int     GetColumnsCount() const;
    virtual const   SColumn&  GetColumn(int index) const;
    virtual SColumn&  GetColumn(int index);
    virtual int     GetColumnIndexByX(int x) const;
    virtual int     GetResizableColumnIndex() const;
    /// @}

    /// @name Columns management
    /// @{
    /// Columns can be "fixed size" or "resizable". Fixed size columns are
    /// rendered with specifed width. Remaining screen space is divided
    /// between "resizable" columns proportionally to m_Share.

    int     AddColumn();
    int     AddColumn(int width, const string& label, int data);
    int     InsertColumn(int pos, int width, const string& label, int data);
    void    SetResizableColumn(int index);

    void    SetColumns(const vector<SColumn>& columns, int resizable_index);

    void    SetColumnWidth(int index, int width);
    void    SetColumnPos(int index, int pos);
    void    SetColumnUserData(int index, int data);
    void    SetColumnVisible(int index, bool b_visible);

    void    SetColumnsByStyle(CWidgetDisplayStyle& style);
    //void    SetHiddenColumns(const vector<string>& labels);
    void    GetVisibleColumns(vector<string>& labels);

    void    UpdateColumns(); // update positions
    /// @}

protected:
    IAlnMultiRendererContext* x_GetContext()    {   return m_Context;   }
    const IAlnMultiRendererContext* x_GetContext()  const    {   return m_Context;   }

    void    x_Layout();

    // Rendering functions
    void    x_Render(TAreaVector* p_areas);
    void    x_RenderHeader(TAreaVector* p_areas);
    void    x_RenderRuler(TAreaVector* p_areas);
    void    x_RenderGraphs(TAreaVector* p_areas);
    void    x_RenderItems(TAreaVector* p_areas);
    void    x_RenderItemsRange(int iFisrt, int iLast, TAreaVector* p_areas);
    void    x_RenderRow(IAlignRow* row, CGlPane& pane, int state, TVPUnit top_y,
                        TAreaVector* areas);
    void    x_RenderMasterRow();

    void    x_ResetRowListMap();
    void    x_InvalidateRows();
    void    x_PurgeRowListMap();

    int     x_GetLineByWindowY(int WinY) const;
    int     x_GetLineByVPY(int vp_y) const;
    TVPRect x_GetLineRect(int Index);

    inline TColumnType     x_GetColumnType(const SColumn& C)
    {
        return (IAlignRow::TColumnType) C.m_UserData;
    }

    void    x_LayoutColumns();

    string x_GetRowTooltip(const TVPPoint& pt, IAlignRow* p_row, int i_col, int vp_top_y);

protected:
    IAlnMultiRendererContext* m_Context;
    TVPRect m_rcBounds; // size of renderer in parents's coordinates

    CRgbaColor m_BackColor;
    CRgbaColor m_MasterBackColor;

    CAlnMultiHeader m_Header;

    CRuler  m_Ruler;
    CGlPane m_RulerPane;

    CGraphContainer m_GraphCont; // hosts alignment graphs
    CGlPane m_GraphPane;


    // Layout information
    TVPRect m_HeaderRect;
    TVPRect m_RulerRect;
    TVPRect m_GraphRect;
    TVPRect m_MasterRect;
    TVPRect m_ListAreaRect;

    typedef     map<IAlignRow*, bool>     TRowToList;
    typedef vector<SColumn> TColumnVector;

    TRowToList  m_RowToList; // value == "true" if key has an OpenGL Display List
                             // columns
    TColumnVector m_Columns;
    int m_ResizableColumnIndex;

    bool m_bFocused; /// indicate whether it shoulde be rendered using "focused"
    // color or not

    TVPPoint m_TooltipVPPos;
    int m_HitResult;
    //IAlignRow* m_TooltipRow;
    CWeakIRef<IAlignRow> m_TooltipRow;
    int m_TooltipColumn;
};

END_NCBI_SCOPE

#endif  // __GUI_WIDGETS_ALNMULTI___ALNMULTI_RENDERER__HPP
