#ifndef GUI_WIDGETS_DATA___QUERY_PARSE_PANEL_HPP
#define GUI_WIDGETS_DATA___QUERY_PARSE_PANEL_HPP

/*  $Id: query_parse_panel.hpp 28139 2013-05-15 17:40:08Z filippov $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Robert Falk
 *
 * File Description:
 *
 */

#include <corelib/ncbistd.hpp>

#include <gui/gui_export.h>

#include <gui/utils/command.hpp>
#include <gui/objutils/query_data_source.hpp>
#include <gui/objutils/query_widget.hpp>
#include <gui/objutils/reg_settings.hpp>

#include <gui/utils/app_job_impl.hpp>
#include <gui/utils/app_job_dispatcher.hpp>

#include <wx/panel.h>

#include <map>


class wxButton;
class wxToolBar;
class wxCheckBox;
class wxAnimationCtrl;
class wxTextCtrl;


BEGIN_NCBI_SCOPE

class IServiceLocator;
class CQueryExec;
class CQueryParseTree;

///////////////////////////////////////////////////////////////////////////////
/// CQueryParsePanel
///
/// This class represents a toolbar that allows the user to query against
/// a data source and then to visualize the query results as selected elements
/// within the attached view.  In order to support the query toolbar, the
/// developer must have a data source that is derived from IQueryDataSource
/// which supports execution of the parsed query against the data.  The 
/// developer must also provide a pointer to a widget (the view for the data
/// source) which must implement the interface IQueryWidget.
///
/// The queries themselves are run as background jobs since they could take
/// significant time in cases where network access is needed.  At job completion
/// a CQueryEvent event is sent to the IQueryWidget-derived widget.  
///
class NCBI_GUIWIDGETS_DATA_EXPORT CQueryParsePanel
:   public wxPanel
,   public CEventHandler
,   public IRegSettings
{
    DECLARE_EVENT_TABLE();
    DECLARE_EVENT_MAP();

public:
    enum ECommands   {
        eCmdStartQuery=eBaseCmdLast + 1066,        
        eCmdStopQuery,
        eCmdSelQuery,
        eCmdNext,
        eCmdPrevious,
        eCmdCaseSensitive,
        eCmdHelp,        

        eCmdLast
    };

    typedef vector<pair<string,string> >  TNamedQueries;

public:
    /// Ctor requires a data source derived from IQueryDataSource and a view
    /// of that data source derived from IQueryWidget.
    CQueryParsePanel(IQueryWidget* w,
                     IQueryDataSource* ds = NULL);
    virtual ~CQueryParsePanel();
    
    /// Create window
    virtual void Create(wxWindow* parent,
                        wxWindowID id = wxID_ANY,
                        const wxPoint& pos = wxDefaultPosition,
                        const wxSize& size = wxDefaultSize);

    /// Initialises member variables
    void    Init();

    /// Creates the controls and sizers
    void    CreateControls();

    /// Set or update data source
    void SetDataSource(IQueryDataSource* ds);

    /// Service locater needed for status bar access
    void SetServiceLocator(IServiceLocator* s);

    /// Add given string to currently displayed query text
    void AddQueryText(const string& query_mod);

    // Perform actions to update toolbar ui after query completes
    void CompleteQuery(const string& status);

    /// Return true if select all checkbox is checked
    bool IsSelectAll();


    /// Set the set of default queries for the current data source
    void SetDefaultQueries(const TNamedQueries& q);
    /// Retrieve the current defualt queries for the data source
    TNamedQueries GetDefaultQueries() const { return m_DefaultQueries; }

    /// @name IRegSettings interface implementation
    /// @{
    virtual void SetRegistryPath(const string& reg_path);
    virtual void LoadSettings();
    virtual void SaveSettings() const;
    /// @}

    /// @name Command handlers
    /// @{
    void    OnSearchctrlTextEnter(wxCommandEvent& evt);
    void    OnQuerySel( wxCommandEvent& evt );
    void    OnButtonNext( wxCommandEvent& evt );
    void    OnButtonPrevious( wxCommandEvent& evt );
    void    OnSearchctrlSearchButtonClick( wxCommandEvent& evt );
    void    OnSearchctrlSearchStop( wxCommandEvent& evt );  
    void    OnSearchHelp(wxCommandEvent& evt);
    void    OnToggleSelectAll(wxCommandEvent & evt);
    void    OnToggleHideUnselected(wxCommandEvent & evt);
    void    OnContextMenu(wxContextMenuEvent&);
    void    OnRecentQuery(wxCommandEvent& evt);

    void    OnAppJobNotification(CEvent* evt);
    /// @}

    /// Event class for notifying attached widget that query is complete
    class CQueryEvent : public CEvent
    {
    public:
        enum EEventType {
            eQueryCompletedEvent,
        };

        /// default ctor
        CQueryEvent() : CEvent() {}  

        /// create an event with all the bells and whistles
        CQueryEvent(TEventClass ecl, TEventID eid, TEventSender* tr)
        : CEvent(ecl, eid, NULL, CEvent::eDelete, tr) {}

        EEventType GetType() const { return (EEventType)GetID(); }

        void SetStatus(const string& str) { m_QueryStatus = str; }
        string GetStatus() const { return m_QueryStatus; }
    
    protected:
        /// Status string returned from query upon completion
        string m_QueryStatus;
    };

   
protected:

    /// Job class used to run query on separate thread
    class  CQueryJob : public CJobCancelable
    {
    public:
        CQueryJob(CQueryExec*  qexec,
                  CQueryParseTree* qtree)
            : m_QExec(qexec)
            , m_QueryTree(qtree)
            , m_UserCancelled(false)
            , m_NumSelected(0) {}

        ~CQueryJob();

        /// Get stats on how many elements were checked and how many were 
        /// checked but failed for some reason (such as being unable to 
        /// compare types)
        int GetQueriedCount() const;
        int GetExceptionCount() const;
        int GetNumSelected() const { return m_NumSelected; }


        /// @name IAppJob implementation
        /// @{
        virtual EJobState                   Run();
        virtual CConstIRef<IAppJobProgress> GetProgress();
        virtual CRef<CObject>               GetResult();
        virtual CConstIRef<IAppJobError>    GetError();        

        virtual string GetDescr() const  { return "Querying Tree"; }
        /// @}

        void SetUserCancelled() { m_UserCancelled = true; }
        bool GetUserCancelled() const { return m_UserCancelled; }

        CQueryExec* GetQueryExec()  { return m_QExec; }

        /// Get mutex
        CMutex& GetMutex() { return m_Mutex; }

    protected:
        /// Guard access to underlying tree
        mutable CMutex m_Mutex;

    private:
        CQueryJob();
        CQueryJob& operator=(const CQueryJob&);

        void x_Query();

        /// Query execution object
        CQueryExec*     m_QExec;
        /// Parsed query tree
        CQueryParseTree*    m_QueryTree;

        bool m_UserCancelled;
        int m_NumSelected;
    };

    /// Return true if 'query' is a single string, and strip blanks and
    /// quotes from the beginning and end of the string.
    bool x_FormatQueryString(string& query, bool casesensitive);
    /// Return true if query parses correctly, putting field names in 'fields'
    bool x_IsValidQuery(string query, 
                        bool casesensitive,
                        vector<std::pair<size_t, size_t> >& fields);
    /// Submit query job for asynch execution
    void x_StartQueryJob(CQueryParseTree* qtree, 
                         bool casesensitive);
    /// Parse query and determine how to execute
    void x_ExecuteQuery(int search_dir);
    /// If user presses cancel button, cancel running job
    void x_CancelQueryInProgress();
    /// Add a query to the list of recently used queries, if it is unique
    void x_AddRecentQuery(const std::string& q);


    /// Only need to load icons into art provider once per run
    static bool m_IconsInitialized;

    ///
    /// Widgets
    wxTextCtrl* m_pSearch;
    wxAnimationCtrl* m_Animation;
    wxToolBar* m_ToolBar;  
    wxToolBar* m_ToolBar2;
    wxButton* m_StopQuery;  
    wxCheckBox* m_CaseSensitive;
    wxCheckBox* m_pSelectAll;
    wxCheckBox* m_pHideUnselected;
    wxPanel* m_ParentPanel;

    /// Service locator for looking up status bar
    IServiceLocator* m_ServiceLocator;

    /// positions of dictionary fields in current search string
    std::vector<std::pair<size_t, size_t> > m_SearchStringFields;

    /// current search string (as of previous update to m_pSearch)
    std::string m_SearchString;

    /// Set of recent queries
    vector<pair<string,string> > m_RecentQueries;
    /// Max number of recent queries to include in list
    const static int m_MaxRecentQueries = 5;
    /// Set of default queries
    vector<pair<string,string> > m_DefaultQueries;
    const static int m_MaxDefaultQueries = 10;

    /// Interface to widget which will display query results
    IQueryWidget* m_QueryWidget;
    /// Interface to data source that is being queried
    IQueryDataSource* m_QueryDataSource;

    /// String of query last time query was executed
    string m_PrevQuery;
    /// Value of case-sensitive button last time query was executed
    bool m_PrevCaseSensitive;

    /// Current tree-query job
    CIRef<CQueryJob>           m_Job;
    /// Job ID of current tree-query job
    CAppJobDispatcher::TJobID  m_JobID;

    /// Registry path for query panel
    string m_RegPath;
    /// Registry section for specific current data source (derived from columns)
    string m_RegSection;
};




END_NCBI_SCOPE

#endif  // GUI_WIDGETS_DATA___QUERY_PARSE_PANEL_HPP
