#ifndef GUI_WIDGETS_PHYLO_TREE__PHYLO_TREE_ALGORITHM_HPP
#define GUI_WIDGETS_PHYLO_TREE__PHYLO_TREE_ALGORITHM_HPP

/*  $Id: phylo_tree_algorithm.hpp 25300 2012-02-21 20:14:48Z falkrb $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Vladimir Tereshkov
 *
 * File Description:
 *
 */


#include <corelib/ncbistd.hpp>
#include <corelib/ncbiobj.hpp>
#include <corelib/ncbi_tree.hpp>
#include <gui/opengl/glpane.hpp>
#include <algo/phy_tree/bio_tree.hpp>
#include <algo/phy_tree/bio_tree_conv.hpp>
#include <objmgr/scope.hpp>
#include <gui/widgets/phylo_tree/phylo_tree_node.hpp>
#include <gui/widgets/phylo_tree/phylo_tree_label.hpp>
#include <util/ncbi_table.hpp>


BEGIN_NCBI_SCOPE


class IPhyloTreeVisitor
{
public:
    typedef CPhyloTreeNode  TTreeType;
public:
    IPhyloTreeVisitor(){}
    virtual ~IPhyloTreeVisitor(){}

    ETreeTraverseCode operator()( TTreeType &  tree_node, int delta);
protected:
    virtual ETreeTraverseCode x_OnStepRight(TTreeType &  x_node);
    virtual ETreeTraverseCode x_OnStepLeft(TTreeType &  x_node);
    virtual ETreeTraverseCode x_OnStepDown(TTreeType &  x_node);
    virtual ETreeTraverseCode x_OnStep(TTreeType &  x_node, int delta);
};


class CPhyloTreeCalculator : public IPhyloTreeVisitor
{
public:
     typedef CPhyTreeNode::TClusterID                 TClusterID;
     typedef map< TClusterID, vector<TTreeType*> >    TClusterHash;
     typedef CNcbiTable<string, unsigned, string>     TAttrTable;

public:
    CPhyloTreeCalculator();
    virtual ~CPhyloTreeCalculator(){}
    
    unsigned GetNumNodes()  { return m_NodeCount; }
    unsigned GetVisibleNodes() { return m_VisibleNodes; }
    // as a tree number of edges has to be m_NodeCount -1
    unsigned GetNumEdges()  { return m_NodeCount-1; }
    int GetWidth()          { return m_Width; }
    int GetHeight()         { return m_Height;}
    double GetMaxDistance() { return m_MaxDistance; }
    double GetMinDistance() { return m_MinDistance; }
    int    GetMaxLabelLen() { return m_MaxLabelLen; }
    int    GetMaxChilds()   { return m_MaxChilds;}
    const  TClusterHash & GetClusters()           { return m_Clusters;        }
    void   SetClusters(const TClusterHash & clst) { m_Clusters  = clst;       }
    void   SetAttrTable(const TAttrTable  & attr);
    void   SetLabelFormat(string labelFmt)        { m_LabelFormat = labelFmt; }
    void   Init();



protected:
    virtual ETreeTraverseCode x_OnStepRight(TTreeType &  x_node);
    virtual ETreeTraverseCode x_OnStepLeft(TTreeType &  x_node);
    virtual ETreeTraverseCode x_OnStepDown(TTreeType &  x_node);
    virtual ETreeTraverseCode x_OnStep(TTreeType &  x_node, int delta);

    /// Updates node properties using values from m_AttrTable
    void x_UpdateProperties(TTreeType& x_node, int row_idx);

private:
    unsigned        m_PamlID;
    unsigned        m_ID;
    unsigned        m_NodeCount;
    unsigned        m_VisibleNodes;
    int             m_PosX;
    int             m_PosY;
    int             m_Leaves;
    int             m_Childs;
    int             m_Width;
    int             m_Height;
    int             m_MaxLabelLen;
    int             m_MaxChilds;
    double          m_MaxDistance;
    double          m_MinDistance;

    TClusterHash        m_Clusters;

    /// Attributes with seq-ids optionally provided to update tree properties
    const TAttrTable  * m_AttrTable;
    
    /// Structure allows us to more efficiently store and look up seq-id keys
    /// in m_AttrTable. 
    struct AttrKey {
        AttrKey() {}
        AttrKey(const string& seqid, unsigned int idx) 
            : m_SeqIDStr(seqid)
            , m_AttrTableIdx(idx) {}

        /// Sort keys on seq-id string value
        bool operator<(const AttrKey& rhs) const { return (m_SeqIDStr < 
                                                           rhs.m_SeqIDStr); }
        string m_SeqIDStr;       
        unsigned int m_AttrTableIdx;
    };


    /// Mapping for efficient lookup of seq-ids in m_AttrTable. 
    vector<AttrKey> m_AttrKeys;

    /// label calculation
    string              m_LabelFormat;
    CPhyloTreeLabel     m_PTL;
};


class CPhyloTreeRectCalculator : public IPhyloTreeVisitor
{
public:
    CPhyloTreeRectCalculator();
    virtual ~CPhyloTreeRectCalculator(){}
    const TModelRect & GetRect()    const {return m_Rect; }
    void               Init();
protected:
    virtual ETreeTraverseCode x_OnStep(TTreeType &  x_node, int delta);
private:
    TModelRect      m_Rect;
};


class CPhyloTreeMaxIdCalculator : public IPhyloTreeVisitor
{
public:
    CPhyloTreeMaxIdCalculator();
    virtual ~CPhyloTreeMaxIdCalculator() {}
    IPhyNode::TID GetMaxId() const {return m_Id; }

protected:
    virtual ETreeTraverseCode x_OnStep(TTreeType &  x_node, int delta);
    IPhyNode::TID      m_Id;
};

class CPhyloTreeSorter : public IPhyloTreeVisitor
{
public:
    CPhyloTreeSorter(bool x_order) : m_bOrder(x_order){}

protected:
    virtual ETreeTraverseCode x_OnStep(TTreeType &  x_node, int delta);

private:
    bool m_bOrder;
};

class IPhyloTreeFilter : public IPhyloTreeVisitor
{
public:
    IPhyloTreeFilter() : m_Tree(NULL){}
    TTreeType * GetTree() {return m_Tree;}

protected:
    virtual ETreeTraverseCode x_OnStep(TTreeType &  x_node, int delta);

    TTreeType* MakeNewTreeNode(const TTreeType& src_node)
    {
        auto_ptr<TTreeType> pnode(new TTreeType());
        unsigned int uid = src_node.GetValue().GetId();
        pnode->GetValue().SetId(uid);
        pnode->SetValue(*src_node);
        return pnode.release();
    }

    // all allowed
    virtual bool x_Allowed(TTreeType &  /*x_node*/) {return true;}

private:
    TTreeType           * m_Tree;
    vector<TTreeType*>    m_TreeStack;
};

// filter out not selected nodes
class CPhyloTreeFilter_Selector : public IPhyloTreeFilter
{
public:
    CPhyloTreeFilter_Selector(){}
protected:
    virtual bool x_Allowed(TTreeType &  x_node)
    {
        return (((*x_node).GetSelectedState()==IPhyGraphicsNode::eSelected) ||
               ((*x_node).GetSelectedState()==IPhyGraphicsNode::eShared));
    }
};

// filter out nodes, having childs number == 1
class CPhyloTreeFilter_Export : public IPhyloTreeFilter
{
public:
    CPhyloTreeFilter_Export(){}
protected:
    virtual bool x_Allowed(TTreeType &  x_node)
    {
       return (*x_node).GetChildsNmb() != 1;
    }
};


// filter out nodes, with distances falling under threshold
class CPhyloTreeFilter_Distance : public IPhyloTreeFilter
{
public:
    CPhyloTreeFilter_Distance(float x_dist) : m_Dist((float)x_dist) {}
protected:
    virtual bool x_Allowed(TTreeType &  x_node)
    {
       return !x_node.GetParent() || ((*x_node).GetDistance() > m_Dist);
    }
private:
    double m_Dist;
};

// filter out nodes, exceeding specified depth
class CPhyloTreeFilter_Depth : public IPhyloTreeFilter
{
public:
    CPhyloTreeFilter_Depth(int x_depth) : m_Depth(x_depth) {}
protected:
    virtual bool x_Allowed(TTreeType &  x_node)
    {
       return ((*x_node).IDX().first <= m_Depth);
    }
private:
    int m_Depth;
};


END_NCBI_SCOPE

#endif
