#ifndef GUI_WIDGETS_PHYLO_TREE__PHYLO_TREE_DS_HPP
#define GUI_WIDGETS_PHYLO_TREE__PHYLO_TREE_DS_HPP

/*  $Id: phylo_tree_ds.hpp 26700 2012-10-26 16:01:53Z falkrb $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Vladimir Tereshkov
 *
 * File Description:
 *
 */


#include <corelib/ncbistd.hpp>
#include <corelib/ncbiobj.hpp>
#include <corelib/ncbi_tree.hpp>
#include <gui/utils/app_job.hpp>
#include <gui/utils/app_job_dispatcher.hpp>
#include <gui/utils/event_handler.hpp>
#include <gui/utils/app_job_impl.hpp>
#include <gui/widgets/phylo_tree/phylo_tree_node.hpp>
#include <gui/widgets/phylo_tree/phylo_tree_scheme.hpp>
#include <gui/widgets/phylo_tree/phylo_tree_reader.hpp>
#include <gui/widgets/phylo_tree/phylo_tree_algorithm.hpp>
#include <gui/objutils/query_func_promote.hpp>
#include <gui/objutils/query_data_source.hpp>
#include <gui/widgets/phylo_tree/phylo_tree_query_exec.hpp>
#include <gui/widgets/phylo_tree/tree_graphics_model.hpp>
#include <gui/opengl/glpane.hpp>
#include <algo/phy_tree/bio_tree.hpp>
#include <algo/phy_tree/bio_tree_conv.hpp>
#include <gui/objutils/biotree_attr_reader.hpp>
#include <objmgr/scope.hpp>


BEGIN_NCBI_SCOPE


class  NCBI_GUIWIDGETS_PHYLO_TREE_EXPORT CPhyloTreeDataSource 
    : public CObject
    , public IQueryDataSource
{
    typedef CAppJobDispatcher::TJobID   TJobID;

public:
    typedef CPhyloTreeNode           TTreeType;
    typedef CPhyloTreeNode::TKeyType TKeyType;

private:
    // sorting visitor
    class visitor_sorter
    {
    public:
        visitor_sorter(bool x_order) : m_bOrder(x_order){}

        ETreeTraverseCode operator()( TTreeType &  tree_node, int delta)
        {
            if (delta == 0 || delta == 1) {
                tree_node.Sort(m_bOrder);
            }
            return eTreeTraverse;
        }
    private:
        bool m_bOrder;
    };

    class dynamic_to_tree
    {
        const CBioTreeFeatureDictionary * m_Dictionary;
    public:
        void SetDictionaryPtr(const CBioTreeFeatureDictionary * dict)    { m_Dictionary = dict; } 

        void operator()(CPhyloTreeNode&  dst_node,
                        const CBioTreeDynamic::TBioTreeNode& src)
        {
            (*dst_node).SetBioTreeFeatureList(src.GetValue().features);
            (*dst_node).SetDictionaryPtr(m_Dictionary);
        }

    };

    class visitor_read_properties
    {
    public:
        visitor_read_properties(){}
        ETreeTraverseCode operator()( TTreeType &  tree_node, int delta)
        {
             if (delta==1 || delta==0){
                (*tree_node).Init();
             }
             return eTreeTraverse;
         }
    };

    class tree_to_dynamic
    {
    public:
        tree_to_dynamic(CPhyloTreeDataSource * ds, CBioTreeDynamic& dynamic_tree)
        : m_DynamicTree(&dynamic_tree), m_pDS(ds)
        {
        }

        void operator()(CBioTreeDynamic::TBioTreeNode& dnode,
                        const CPhyloTreeNode&  src_node)
        {
           const CBioTreeFeatureList        & lst  = (*src_node).GetBioTreeFeatureList();
           const CBioTreeFeatureDictionary  & dict = m_pDS->GetDictionary();

           // lazy pre-init of dictionary
           // copy exact dictionary (to avoid unnecessary update of previous views
           // and any data inconsistancy
           if (m_DynamicTree->GetFeatureDict().GetFeatureDict().empty()) {
             ITERATE(CBioTreeFeatureDictionary::TFeatureDict, dc, dict.GetFeatureDict()) {
                   m_DynamicTree->GetFeatureDict().Register(dc->first, dc->second);
             }
           }

           ITERATE(CBioTreeFeatureList::TFeatureList, feat, lst.GetFeatureList()) {
               const CBioTreeFeatureDictionary::TFeatureDict & fd = dict.GetFeatureDict();
               CBioTreeFeatureDictionary::TFeatureDict::const_iterator it = fd.find((*feat).id);

               if (it!=fd.end()) {
                    m_DynamicTree->AddFeature(&dnode,
                                          it->second,
                                          (*feat).value);
               }
           }
        }

    private:
        CBioTreeDynamic*        m_DynamicTree;
        CPhyloTreeDataSource*   m_pDS;
        string             m_TmpStr;

    };
    
    class visitor_show_all
    {
    public:
         ETreeTraverseCode operator()( TTreeType &  tree_node, int delta)
         {
             if (delta==1 || delta==0){
                (*tree_node).SetChildsDisplay(IPhyGraphicsNode::eShowChilds);
                (*tree_node).Sync();
             }
             return eTreeTraverse;
         }
    };

    class visitor_set_dirty
    {
    public:
        visitor_set_dirty(bool x_dirty) : m_bDirty(x_dirty){}
        ETreeTraverseCode operator()( TTreeType &  tree_node, int delta)
        {
            if (delta==1 || delta==0){
                (*tree_node).SetDirtyState(m_bDirty?
                    IPhyGraphicsNode::eDirty : IPhyGraphicsNode::eClean);
            }
            return eTreeTraverse;
        }
    private:
        bool m_bDirty;
    };

    class visitor_node_finder
    {
    public:
        visitor_node_finder(int x_id) : m_Id(x_id), m_Node(NULL) {}
        TTreeType * GetNode() { return m_Node;}

        ETreeTraverseCode operator()( TTreeType &  tree_node, int delta)
        {
            if (delta==1 || delta==0){
                if ((*tree_node).GetId() == (unsigned) m_Id) {
                    m_Node = &tree_node;
                    return eTreeTraverseStop;
                }
            }
            return eTreeTraverse;
        }
    private:
        int         m_Id;
        TTreeType * m_Node;
    };

    class visitor_selection_reader
    {
    public:
        visitor_selection_reader(){}
        ETreeTraverseCode operator()( TTreeType &  tree_node, int delta)
        {
            if (delta==1 || delta==0){
                if ((*tree_node).GetSelectedState() == IPhyGraphicsNode::eSelected) {
                    m_SelNodes.push_back(&tree_node);
                    m_SelIDs.push_back((*tree_node).GetId());
                }
            }
            return eTreeTraverse;
        }

        const vector <int>          & GetIDs()      { return m_SelIDs;   }
        const vector<TTreeType*>    & GetNodes()    { return m_SelNodes; }
    private:
        vector <TTreeType*> m_SelNodes;
        vector <int>        m_SelIDs;
    };


    class visitor_string_query
    {
    public:
        visitor_string_query(string qstring) : m_QString(qstring){}
        ETreeTraverseCode operator()( TTreeType &  tree_node, int delta)
        {
            if (delta==1 || delta==0){
                 CBioTreeFeatureList::TFeatureList f_list = 
                     (*tree_node).GetBioTreeFeatureList().GetFeatureList();


                 ITERATE(CBioTreeFeatureList::TFeatureList, it, f_list) {
                     if (it->value.find(m_QString)!=string::npos) {
                        m_SelNodes.push_back(&tree_node);
                        m_SelIDs.push_back((*tree_node).GetId());
                        break;
                     }                     
                 }    
            }
            return eTreeTraverse;
        }

        const vector <int>          & GetIDs()      { return m_SelIDs;   }
        const vector<TTreeType*>    & GetNodes()    { return m_SelNodes; }
    private:
        vector <TTreeType*> m_SelNodes;
        vector <int>        m_SelIDs;
        string              m_QString;
    };

    class visitor_string_query_no_case
    {
    public:
        visitor_string_query_no_case(string qstring) : m_QString(qstring) 
        { 
            m_CompareString.assign(m_QString.c_str(), m_QString.size()); 
        }

        ETreeTraverseCode operator()( TTreeType &  tree_node, int delta)
        {
            if (delta==1 || delta==0){
                 CBioTreeFeatureList::TFeatureList f_list = 
                     (*tree_node).GetBioTreeFeatureList().GetFeatureList();


                 ITERATE(CBioTreeFeatureList::TFeatureList, it, f_list) {                    
                     CTempString  ts(it->value.c_str(), it->value.size());
                     if (NStr::FindNoCase(ts, m_CompareString, 0, NPOS, NStr::eFirst) != NPOS) {
                        m_SelNodes.push_back(&tree_node);
                        m_SelIDs.push_back((*tree_node).GetId());
                        break;
                     }                     
                 }    
            }
            return eTreeTraverse;
        }

        const vector <int>          & GetIDs()      { return m_SelIDs;   }
        const vector<TTreeType*>    & GetNodes()    { return m_SelNodes; }
    private:
        vector <TTreeType*> m_SelNodes;
        vector <int>        m_SelIDs;
        string              m_QString;
        CTempString         m_CompareString;
    };

    class visitor_leafs_reader
    {
    public:
        visitor_leafs_reader(){}
        ETreeTraverseCode operator()( TTreeType &  tree_node, int delta)
        {
            if (delta==1 || delta==0){
                if (tree_node->GetNodesNmb()==0) {
                    m_SelNodes.push_back(&tree_node);                  
                }
            }
            return eTreeTraverse;
        }
       
        const vector<TTreeType*>    & GetNodes()    { return m_SelNodes; }
    private:
        vector <TTreeType*> m_SelNodes;
    };
public:
    typedef map<int, CPhyloTreeNode*>                TNodeHash;
    typedef vector <CConstRef<objects::CSeq_id> >    TIds;
    typedef map<int, CConstRef<objects::CSeq_id> >   TMapIds;

    CPhyloTreeDataSource();
    CPhyloTreeDataSource(const CBioTreeDynamic & tree);
    CPhyloTreeDataSource(const CBioTreeDynamic & tree, objects::CScope& scope);
    virtual ~CPhyloTreeDataSource();

    // BioTree i/o
    void Init(const CBioTreeDynamic & tree);
    void Save(CBioTreeDynamic & tree,  const TTreeType * x_root = NULL);

    // root node
    CPhyloTreeNode* GetTree(void) { return m_Root; }

    /// Get model for rendering
    CTreeGraphicsModel& GetModel() { return  m_Model.GetObject(); }


    unsigned int GetNumNodes(void) { return m_Calc.GetNumNodes();    }
    unsigned int GetVisibleNodes(void) { return m_Calc.GetVisibleNodes();}
    unsigned int GetNumEdges(void) { return m_Calc.GetNumEdges();    }
    int GetWidth(void)             { return m_Calc.GetWidth();       }
    int GetSize(void)              { return m_Calc.GetHeight();      }
    int GetMaxLabelLen(void)       { return m_Calc.GetMaxLabelLen(); }
    int GetMaxChilds(void)         { return m_Calc.GetMaxChilds();   }

    void SetSelection(const vector<CPhyloTreeNode*>& sel_nodes);

    //TODO this should be renamed to Select()
    void SetSelection(CPhyloTreeNode * node, bool bSel,
                      bool andChilds = true, bool checkParents = true);
    void QuerySelect(CPhyloTreeNode * node, 
                     bool bSel,
                     bool select_children = false, 
                     bool select_parents = false);


    void SetCurrentNode(CPhyloTreeNode * node)      { m_Current = node; }
    CPhyloTreeNode * GetCurrentNode(void)           { return m_Current; }

    static CPhyloTreeNode * GetClipboard(void)      { return m_sClipboard;}

    vector<int>         GetSelectedIds(void) const;
    vector<TTreeType*>  GetSelectedNodes(void) const;

    CPhyloTreeNode* GetNode(int idx);

    // returning calculated bounding rectangle
    const TModelRect & GetBoundRect(void);

    // returning norm distance (max distance)
    const double GetNormDistance(void);

    // returning min distance
    const double GetMinDistance(void);

    const TMapIds   GetSeqIds() const { return m_SeqIds;    }

    // expanding all nodes
    void ShowAll(void);

    // invalidating all nodes
    void MarkAllDirty(bool bDirty = true);

    // subtree selection
    void SetSubtree(CPhyloTreeNode * newRoot);

    // re-rooting a tree
    void RerootTree(CPhyloTreeNode * newRoot);

    // childs visibility
    void ExpandCollapse(CPhyloTreeNode * node,
        CPhyTreeNode::TChildsDisplay dsp);

    //Update markers for parent nodes
    void UpdateMarkers(CPhyloTreeNode * node);

    // Write the tree to _TRACE for debugging purposes
    void DebugTrace(CPhyloTreeNode *node, int indent = 0);

    // simplification for depth, very fast
    void SimplifyForDepth(int depth);

    // refresh
    void Refresh(void);

    // biotree selection retrieval
    void ConvertSelectionToBiotree(CBioTreeDynamic & btree);

    // sorting
    void Sort(bool bAscending = true);

    // clusters check
    bool HasClusters(void);

    // clusterizing with colors
    void Clusterize(void);

    // filtering by distance
    void FilterDistances(double x_dist);

    // cleaning one-child nodes
    void Clean(void);

    // filtering selected subtree
    void Filter(void);

    // applying attributes (aka clusterization)
    void ApplyAttributes(CBioTreeAttrReader::TAttrTable & attrs);

    // relabel all nodes corresponding to format
    void Relabel(string labelFmt);

    // generate tooltip format
    string GenerateTooltipFormat();

    // cut/copy/paste operations on selected node
    void Copy(void);
    void Cut(void);
    void Paste(void);
    void NewNode(bool bAfter = true);
    void Remove(bool bSubtree = true);
    void MoveNode(bool bUp);

    const CBioTreeFeatureDictionary & GetDictionary() const    { return m_Dictionary; }
    void SetDictionary(const CBioTreeFeatureDictionary & dict) { m_Dictionary = dict; }   

    /// search for nodes via a simple string query
    vector<TTreeType*>  FindNodes(string query, bool casesensitive) const;

    // Iterate or toggle display mode of the selected set
    CPhyloTreeNode* IterateOverSelNodes( int direction, bool highlight);


    /// @name IQueryDataSource interface implementation
    /// @{
    // need indexd access to map.  Just iterate?
    virtual string GetColumnLabel(size_t col) const;
    virtual size_t GetColsCount() const;
    virtual CQueryExec* GetQueryExec(bool casesensitive);
    virtual void SetQueryResults(CQueryExec* q);
    virtual void ClearQueryResults() { m_SearchCache.clear(); }
    virtual CRef<objects::CScope> GetScope() { return m_Scope; }

    // Execute a simple string-matching query on the underlying data
    virtual string ExecuteStringQuery(string query, bool casesensitive);
    /// @}

    vector<TTreeType*>  GetQueryResults() const { return m_SearchCache; }

private:
    // search support
    vector<TTreeType*> m_SearchCache;
    TTreeType*         m_SearchCurrentNode;

    // root node
    CPhyloTreeNode* m_Root;

    /// Model for rendering (updated from current tree state)
    CRef<CTreeGraphicsModel>  m_Model;

    // visitor/calculator
    CPhyloTreeCalculator m_Calc;

    // rectangle clculator
    CPhyloTreeRectCalculator m_Rect;


    TMapIds    m_SeqIds;
    CRef<objects::CScope>    m_Scope;

    // current node, to perform operations on it
    CPhyloTreeNode * m_Current;

    // clipboard
    static CPhyloTreeNode  * m_sClipboard;

    // biotree dictionary
    CBioTreeFeatureDictionary    m_Dictionary;


    void x_MeasureTree(TTreeType& node);
    void x_MeasureTree();
    void x_UpdateMarkers(CPhyloTreeNode * node);
    void x_InitRenderModel();

    void x_DescentToChilds(CPhyloTreeNode * node, bool bSel);
    void x_AscendToRoot(CPhyloTreeNode * node, bool bSel);
    
    void x_SelectChildren(CPhyloTreeNode * node,  
                          IPhyGraphicsNode::TSelectedState sel);
    void x_SelectParents(CPhyloTreeNode * node,  
                         IPhyGraphicsNode::TSelectedState sel);

    void x_CommonNodeSearch(CPhyloTreeNode * node);
    void x_PropagateCommonState(CPhyloTreeNode * node);


    CPhyloTreeNode * x_ConvertUpstream(CPhyloTreeNode * node);
};

END_NCBI_SCOPE

#endif
