#ifndef GUI_WIDGETS_PHY_TREE___PHYLO_TREE_NODE__HPP
#define GUI_WIDGETS_PHY_TREE___PHYLO_TREE_NODE__HPP

/*  $Id: phylo_tree_node.hpp 25379 2012-03-06 16:39:19Z falkrb $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Vladimir Tereshkov
 *
 * File Description:
 *
 */

#include <corelib/ncbistd.hpp>
#include <corelib/ncbiobj.hpp>
#include <corelib/ncbi_tree.hpp>
#include <serial/serialbase.hpp>
#include <serial/serialimpl.hpp>
#include <objects/seqloc/Seq_id.hpp>
#include <algo/phy_tree/bio_tree.hpp>
#include <algo/phy_tree/bio_tree_conv.hpp>
#include <gui/utils/rgba_color.hpp>
#include <gui/utils/vect2.hpp>
#include <gui/opengl/glrect.hpp>



BEGIN_NCBI_SCOPE

class CSubtreeBoundary;

// base class for any tree node
// accessible through algorithms
class IPhyNode
{
public:
    typedef unsigned TID;

public:
    IPhyNode(){}
    IPhyNode(TID id) : m_ID(id){}
    virtual ~IPhyNode(){}

    TID GetId() const       { return m_ID; }
    void SetId(TID x_id)    { m_ID = x_id; }

    bool operator ==(IPhyNode & x_node)
    {
        return x_node.GetId() == m_ID;
    }

protected:
    TID m_ID;
};



// most commonly used newick format node
class CPhyNewickNode : public IPhyNode
{
public:
    typedef string TLabel;
    typedef float TDistance;

public:
    CPhyNewickNode() : m_Label(""), m_Distance(0) {}
    CPhyNewickNode(TID x_id, TDistance x_dist,
        const TLabel & x_label) : IPhyNode(x_id),
        m_Label(x_label), m_Distance(x_dist){};
    virtual ~CPhyNewickNode(){}

    TDistance GetDistance() const   { return fabs(m_Distance); }
    const TLabel & GetLabel() const { return m_Label;    }

    void SetDistance(TDistance x_dist)    { m_Distance = x_dist; }
    void SetLabel(const TLabel & x_label) { m_Label = x_label;   }

    void Set(TID x_id, TDistance x_dist,
        const TLabel & x_label)
    {
        m_ID       = x_id;
        m_Label    = x_label;
        m_Distance = x_dist;
    }

protected:
    TLabel      m_Label;
    TDistance   m_Distance;
};

// displayable node
class IPhyGraphicsNode
{
public:
    typedef enum {
        eNotSelected,   // node is not selected
        eSelected,      // node is selected (descending)
        eTraced,        // node is traced (up to root)
        eShared         // node is on path between selected nodes
    } TSelectedState;

    typedef enum {
        eShowChilds,  // node expanded
        eHideChilds   // node collapsed
    } TChildsDisplay;

    typedef enum {
        eClean,
        eDirty
    } TDirtyState;

    typedef enum {
        eUnBounded,
        eBounded
    } TBoundingState;

    typedef CVect2<float>        TPoint;
    typedef pair<int,int>        TIndex;
    typedef string               TColor;

   
public:
    IPhyGraphicsNode() :
          m_Angle(0.0f), m_AnnWedge(0.0f),
          m_Selected(IPhyGraphicsNode::eNotSelected),
          m_Children(IPhyGraphicsNode::eShowChilds),
          m_Dirty(IPhyGraphicsNode::eDirty),          
          m_LabelFgColor("#FFFFFF"),
          m_LabelBgColor("#6D95E0"),
          m_Boundary(NULL),
          m_Size(-1) {}          
    
    IPhyGraphicsNode(const IPhyGraphicsNode& n);

    virtual ~IPhyGraphicsNode();
    IPhyGraphicsNode& operator=(const IPhyGraphicsNode& n);

    static const TColor TNoColor;


    TPoint & XY()   { return m_XY;  }
    float & X()     { return m_XY.X(); }
    float & Y()     { return m_XY.Y(); }
    TIndex & IDX()  { return m_IDX; }

    TSelectedState GetSelectedState()             { return m_Selected;      }
    void SetSelectedState(TSelectedState x_state) { m_Selected = x_state;   }

    TChildsDisplay  GetChildsDisplay() const      { return m_Children;      }
    void SetChildsDisplay(TChildsDisplay x_chds)  { m_Children = x_chds;    }

    TBoundingState  GetBoundedDisplay() const      { return m_Bounded;      }
    void SetBoundedDisplay(TBoundingState x_bound) { m_Bounded = x_bound;   }

    TDirtyState GetDirtyState()                   { return m_Dirty;         }
    void SetDirtyState(TDirtyState x_dst)         { m_Dirty = x_dst;        }

    const TColor & GetNodeEdColor()               { return m_NodeEdColor;   }
    const TColor & GetNodeFgColor()               { return m_NodeFgColor;   }

	const TColor & GetClusterEdColor()            { return m_ClusterEdColor;   }
    const TColor & GetClusterFgColor()            { return m_ClusterFgColor;   }

    void SetLabelCoords(const TPoint& p, 
                        const CGlRect<float>& r) { m_LabelPos = p; m_LabelRect = r; }

    TPoint & GetLabelPos()                        { return m_LabelPos; }
    CGlRect<float> & GetLabelRect()               { return m_LabelRect; }

    const TColor & GetLabelFgColor()              { return m_LabelFgColor;  }
    const TColor & GetLabelBgColor()              { return m_LabelBgColor;  }    

    void SetMarkerColors(const string& str);
    string GetMarkerColorsAsString();
    vector<CRgbaColor>& GetMarkerColors() { return m_MarkerColors; }
    float GetMarkerSize() const { return m_MarkerSize; }
    void SetMarkerSize(float f) { m_MarkerSize = f; }

    /// Gets current boundary or creates a new one if m_Boundary is NULL
    CSubtreeBoundary* GetSubtreeBoundary();
    /// Deletes current boundary (if not NULL).
    void DeleteSubtreeBoundary();

    void SetNodeColor(TColor x_fg, TColor x_edge = TNoColor)
    {
        m_NodeFgColor = x_fg;
        m_NodeEdColor = x_edge;
    }

	void SetClusterColor(TColor x_fg, TColor x_edge = TNoColor)
    {
        m_ClusterFgColor = x_fg;
        m_ClusterEdColor = x_edge;
    }

    void SetLabelColor(TColor x_fg, TColor x_bg = TNoColor)
    {
        m_LabelFgColor = x_fg;
        m_LabelBgColor = x_bg;        
    }

    int GetNodeSize()                   { return m_Size;    }
    void SetNodeSize(int x_sz)          { m_Size = x_sz;    }

    float GetAnnWedge(void)            { return m_AnnWedge;  }
    void  SetAnnWedge(double wedge)   { m_AnnWedge = wedge; }

    float GetAngle(void)              { return m_Angle;  }
    void  SetAngle(float angle)       { m_Angle = angle; }
    
   
    
private:
    string x_GetParameter(const string& features, const string& parm);

    // location / coordinates
    TPoint      m_XY;
    TIndex      m_IDX;
    float       m_Angle;
    float       m_AnnWedge;
    
    // Scaled label offset (0 for non-scalable nodes, which is what they 
    // currently are)
    TPoint      m_LabelPos;
    // Pixel coordinates of label relative to node position
    CGlRect<float> m_LabelRect;


    // states
    TSelectedState  m_Selected;
    TChildsDisplay  m_Children;
    TBoundingState  m_Bounded;
    TDirtyState     m_Dirty;

    // node colors, defined by scheme
    TColor m_NodeEdColor;
    TColor m_NodeFgColor;

	// node colors, if cluster present
    TColor m_ClusterEdColor;
    TColor m_ClusterFgColor;

    // label color
    TColor m_LabelFgColor;
    TColor m_LabelBgColor;
	
    // marker colors (0 or 1 at a leaf, 0..n otherwise)
    vector<CRgbaColor> m_MarkerColors;
    // marker size, as a multiple of node size
    float m_MarkerSize;

    // Boundary for subtree at this node (if m_Bounded is eBounded)
    CSubtreeBoundary*  m_Boundary;

    // node size
    int m_Size;
};


class CPhyTreeNode : public CPhyNewickNode,
                     public IPhyGraphicsNode
{
public:
    typedef string  TClusterID;
public:
    CPhyTreeNode()
        : m_NmbChilds(0), m_NmbLeafs(0), m_Counter(0), m_PamlCounter(0),
          m_NmbChildsEx(0), m_NmbLeafsEx(0), m_NmbNodes(0), m_DistFromRoot(0.0f), m_ClusterID("")
    {}


    // cache accessors
    int     GetLeavesNmb()          { return m_NmbLeafs;  }
    void    SetLeavesNmb(int x_ln)  { m_NmbLeafs = x_ln;  }

    int     GetLeavesNmbEx()          { return m_NmbLeafsEx;  }
    void    SetLeavesNmbEx(int x_ln)  { m_NmbLeafsEx = x_ln;  }

    int     GetChildsNmb()          { return m_NmbChilds; }
    void    SetChildsNmb(int x_ch)  { m_NmbChilds = x_ch; }

    int     GetNodesNmb()           { return m_NmbNodes; }
    void    SetNodesNmb(int x_ch)   { m_NmbNodes = x_ch; }

    int     GetChildsNmbEx()          { return m_NmbChildsEx;  }
    void    SetChildsNmbEx(int x_ln)  { m_NmbChildsEx = x_ln;  }

    float   GetDistFromRoot()               { return fabs(m_DistFromRoot);    }
    void    SetDistFromRoot(float x_dist)   { m_DistFromRoot = x_dist;  }

    TClusterID GetClusterID(void)               { return m_ClusterID; }
    void       SetClusterID(TClusterID x_id)    { m_ClusterID = x_id; }

    const CRef<objects::CSeq_id> & GetSeqID()   { return m_SeqID;       }
    void SetSeqID(objects::CSeq_id * sid)       { m_SeqID.Reset(sid);   }

    const CBioTreeFeatureList & GetBioTreeFeatureList() const      { return m_Features;  }
    void  SetBioTreeFeatureList(const CBioTreeFeatureList & btcfl) { m_Features = btcfl; }

    int   GetCounter()              { return m_Counter;  }
    void  SetCounter(int x_ctr)     { m_Counter = x_ctr; }

    int   GetPamlCounter()          { return m_PamlCounter; }
    void  SetPamlCounter(int x_ctr) { m_PamlCounter = x_ctr;}
    
    const CBioTreeFeatureDictionary * GetDictionaryPtr()        { return m_Dictionary; }
    void SetDictionaryPtr(const CBioTreeFeatureDictionary * dict)    { m_Dictionary = dict; }

    bool operator ==(const CPhyTreeNode & x_node) { return m_ID == x_node.GetId(); }

    // register/update a feature
    void    SetFeature(const string & name, const string & value);

    // init from btc or feature list
    void    Init();

    // init excluding seq-id
    void    InitFeatures();

    // write to btc
    void    Sync();

private:
    // additional properties
    CRef<objects::CSeq_id>              m_SeqID;
    CBioTreeFeatureList                 m_Features;
    const CBioTreeFeatureDictionary*    m_Dictionary;
   

    // cached values
    int     m_NmbChilds;
    int     m_NmbLeafs;

    // counter, for labeling
    // it's not an ID and will be overwritten by algorithm
    int     m_Counter;

    // Paml-style counter, leaf-skipping
    int     m_PamlCounter;

    // tweaked for close/open
    int     m_NmbChildsEx;
    int     m_NmbLeafsEx;

    // number of all internal nodes
    int     m_NmbNodes;

    // distance from root
    float  m_DistFromRoot;
    
    // cluster ID
    TClusterID m_ClusterID;
};


class CPhyloTreeNode : public CTreeNode<CPhyTreeNode>
{
public:
    typedef CTreeNode<CPhyTreeNode> TTreeType;
private:
    // predicate for sorting childs by their weight
    class predicate_sort_by_leaves_count
    {
    public:
        predicate_sort_by_leaves_count(bool x_order) : m_bOrder(x_order){}
        bool operator() (TTreeType * node1, TTreeType * node2)
        {
            bool greater = node1->GetValue().GetLeavesNmb() > node2->GetValue().GetLeavesNmb();
            return m_bOrder ? greater : !greater;
        }
    private:
        bool m_bOrder;
    };

public:
    CPhyloTreeNode();
    CPhyloTreeNode(int x_id, double x_dist, const string & x_label);
   ~CPhyloTreeNode(){}

    // default sorting - by number of leaves
    void    Sort(bool bOrder);

    // sorting by state - for rendering optimization
    void    SortByState();

    bool    IsVisible();
    bool    Expanded() const { return GetValue().GetChildsDisplay() == IPhyGraphicsNode::eShowChilds; }

    void    ExpandCollapse(IPhyGraphicsNode::TChildsDisplay x_chds);
    bool    CanExpandCollapse(IPhyGraphicsNode::TChildsDisplay x_chds);
    void    SetLabel(const string & x_label);

    int NumChildren() const { return m_Nodes.size(); }

    // 'tweaked' functions for transitional period
    bool    IsLeafEx()              { return (m_Nodes.empty() || (!Expanded()));            }
    TNodeList_I SubNodeBeginEx()    { return Expanded() ? m_Nodes.begin() : m_Nodes.end();  }
    TNodeList_I SubNodeEndEx()      { return m_Nodes.end();                                 }

    /// A function to help debug/optimize overall size of tree model
    static void HowBigAmI();
protected:
    static bool x_Greater(const TNodeList::value_type& node1,
                          const TNodeList::value_type& node2)
    {
            return (**node1).GetLeavesNmbEx()
                > (**node2).GetLeavesNmbEx();
    }

    static bool x_NotGreater(const TNodeList::value_type& node1,
                             const TNodeList::value_type& node2)
    {
            return (**node1).GetLeavesNmbEx()
                <= (**node2).GetLeavesNmbEx();
    }
    // for status-based sorting needed for optimized
    // cladogram rendering
    static bool x_Brighter(const TNodeList::value_type& node1,
                           const TNodeList::value_type& node2)
    {
        // 4-way comparison
        switch ((**node1).GetSelectedState()) {
        case IPhyGraphicsNode::eNotSelected:
            {{
                //return true;
                return ((**node2).GetSelectedState()!=IPhyGraphicsNode::eNotSelected);
            }}
        case IPhyGraphicsNode::eTraced:
            {{
                return ((**node2).GetSelectedState()!=IPhyGraphicsNode::eNotSelected);
            }}
        case IPhyGraphicsNode::eShared:
            {{
                return ((**node2).GetSelectedState()==IPhyGraphicsNode::eSelected);
            }}
        case IPhyGraphicsNode::eSelected:
            {{
                return false;
            }}
        default:
            {{
                return false;
            }}
        }
    }
};


END_NCBI_SCOPE

#endif // !defined(GUI_WIDGETS_PHY_TREE___PHYLO_TREE_NODE__HPP)
