#ifndef GUI_WIDGETS_PHYLO_TREE___PHYLO_TREE_QUERY_EXEC__HPP
#define GUI_WIDGETS_PHYLO_TREE___PHYLO_TREE_QUERY_EXEC__HPP

/*  $Id: phylo_tree_query_exec.hpp 24862 2011-12-14 18:43:44Z falkrb $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors: Bob Falk
 *
 * File Description: 
 * Header file for classes needed to implement query execution.
 *
 */

#include <objmgr/scope.hpp>

#include <util/qparse/query_exec.hpp>
#include <gui/widgets/phylo_tree/phylo_tree_node.hpp>

#include <algo/phy_tree/bio_tree.hpp>

#include <stack>

BEGIN_NCBI_SCOPE

////////////////////////////////////////////////////////////////////////////////
/// class CTreeQueryExec
///
///  Subclass of CQueryExec that adds functions spcific to the phylogenetic
///  tree that allow it to extract data from the feature lists of 
///  individual nodes so that those values can be used in queries.
class NCBI_GUIWIDGETS_PHYLO_TREE_EXPORT CTreeQueryExec : public CQueryExec
{
public:
    CTreeQueryExec(const CBioTreeFeatureDictionary* d) 
        : m_EvalNode(NULL) 
        , m_RootNode(NULL)
        , m_Dictionary(d) {}

    virtual ~CTreeQueryExec() {}

    /// Set current node for query execution
    void SetRootNode(CPhyloTreeNode *node) {m_RootNode = node;}

    /// Returns id of a registred tree feature based on case-insensitive compare
    /// or -1 if feature does not exist.
    TBioTreeFeatureId 
        GetFeatureIdNoCase(const string& feature_name,
                           const CBioTreeFeatureDictionary* dict) const;

    /// @name Find and convert feature-list values in individual nodes
    /// @{    
    /// Search for bool 'identifier' in feature list and return if found
    virtual bool ResolveIdentifier(const std::string& identifier, 
                                   bool& value);    

    /// Search for integer 'identifier' in feature list and return if found
    virtual bool ResolveIdentifier(const std::string& identifier, 
                                   int& value);

    /// Search for float 'identifier' in feature list and return if found
    virtual bool ResolveIdentifier(const std::string& identifier, 
                                   double& value);

    /// Search for string 'identifier' in feature list and return if found
    virtual bool ResolveIdentifier(const std::string& identifier, 
                                   std::string& value);
    /// @}
    
    /// Search for 'identifier' in dictionary and return true if it exists. 
    /// This does not mean that the value is defined for a particular node.
    virtual bool HasIdentifier(const std::string& identifier);

    /// Some applications may know the type of an identifier.  This hook
    /// should be overriden to return an identifier's type, when available.
    /// Return one of eIntConst, eBoolConst, eFloatConst, eString, or eNotSet.
    virtual CQueryParseNode::EType IdentifierType(const std::string& /* identifier */)
        { return CQueryParseNode::eNotSet; }


    /// Move to the first row for eval, return false if table empty
    virtual void EvalStart();
    /// Move to the next row for eval, return false if table size < m_EvalRow+1
    virtual bool EvalNext(CQueryParseTree& qtree);
    virtual bool EvalComplete();

    std::vector<CPhyloTreeNode*> GetTreeSelected() const { return m_Selected; }
    
protected:
    /// Node currently being evaluated
    CPhyloTreeNode* m_EvalNode;
    CPhyloTreeNode* m_RootNode;

    /// for iterating through the node tree without recursion
    struct PhyloTreePointer {
        PhyloTreePointer(CPhyloTreeNode* n)
            : m_Node(n)
            , m_Iterator(n->SubNodeBegin()) {}
        CPhyloTreeNode* m_Node;
        CPhyloTreeNode::TNodeList_I  m_Iterator;
    };

    /// stack of visited nodes up the tree
    std::stack<PhyloTreePointer> m_NodeStack;

    /// Dictionary of 'features' for the tree
    const CBioTreeFeatureDictionary* m_Dictionary;

    /// current set of selected nodes from the query
    std::vector<CPhyloTreeNode*> m_Selected;
};



END_NCBI_SCOPE


#endif  // GUI_WIDGETS_PHYLO_TREE___PHYLO_TREE_QUERY_EXEC__HPP

