#ifndef GUI_WIDGETS_PHYLO_TREE___PHYLO_TREE_RENDER__HPP
#define GUI_WIDGETS_PHYLO_TREE___PHYLO_TREE_RENDER__HPP

/*  $Id: phylo_tree_render.hpp 28068 2013-05-06 21:52:15Z rafanovi $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Vladimir Tereshkov
 *
 * File Description:
 *
 */

#include <corelib/ncbistl.hpp>
#include <corelib/ncbitime.hpp>

#include <gui/widgets/phylo_tree/phylo_tree_scheme.hpp>
#include <gui/widgets/gl/ievent_handler.hpp>
#include <gui/widgets/wx/sticky_tooltip_handler.hpp>
#include <gui/widgets/phylo_tree/phylo_tree_lod.hpp>
#include <gui/widgets/phylo_tree/phylo_tree_label.hpp>
#include <gui/widgets/phylo_tree/phylo_tree_boundary_shapes.hpp>

#include <gui/print/pdf.hpp>

#include <gui/utils/curve_bezier.hpp>
#include <gui/utils/rgba_color.hpp>

#include <gui/opengl/glcurve.hpp>
#include <gui/opengl/gltexture.hpp>

BEGIN_NCBI_SCOPE

class CMouseZoomHandler;
class CPhyloTreeDataSource;
class CGlBitmapFont;
class CGlTextureFont;
class CGlTexture;
class CGLFrameBuffer;
class CGlVboNode;
class CGlRender;

class IPhyloTreeRendererHost
{
public:
    virtual ~IPhyloTreeRendererHost()  {};

    virtual void     HMGH_UpdateLimits(bool force_redraw = false) = 0;
    virtual TVPUnit  HMGH_GetVPPosByY(int y) const = 0;
    /// a topology or selection (color) change (requires update to buffers
    virtual void     HMGH_OnChanged(void) = 0;
    /// a chnage that requires a redraw but no GL buffer updates
    virtual void     HMGH_OnRefresh(void) = 0;
    virtual void     FireCBEvent(void) = 0;
    virtual void     FireEditEvent(void) = 0;
};

// rendering interface
class NCBI_GUIWIDGETS_PHYLO_TREE_EXPORT IPhyloTreeRenderer
    : public wxEvtHandler,
      public IGlEventHandler,
      public IPhyloTreeLOD
{
    DECLARE_EVENT_TABLE()
public:
    typedef enum {
        eDistancesRendering,
        eAdaptiveMargins,
        eSimplification,
        eSplinesRendering,
        eGenerateTextures,
        eZoomablePrimitives,
        eAutofitLabels
    } TRenderingOption;

    enum ELayoutStatus {
        eValidLayout,       // Layout is up-to-date with topology
        eNeedLayout,        // Layout is out-of-date but renderer has not changed
        eNeedLayoutAndSize, // Layout is out-of-date and renderer has changed    
    };

    typedef CVect3<float> TPoint;

    IPhyloTreeRenderer();
    IPhyloTreeRenderer(double width, double height);
    virtual ~IPhyloTreeRenderer();

    // selection set/retrieve
    const vector<int> & GetSelectedIDs(void) { return m_SelectedIDs; };
    void                SetSelectedIDs(const vector<int> & vect);

    /// @name Forwarded messages from to support IStickyTooltipHandlerHost
    /// @{
    string  TTHH_NeedTooltip(const wxPoint & pt);
    CTooltipInfo  TTHH_GetTooltip(const wxRect & rect);
   /// @}
    
    /// id is the tip id of the tip that the users mouse is currently on, or -1
    void SetActiveTooltipNode(int id);

    /// point to a particular node from a tooltip to help user spot it
    /// rectangle should be the tooltip screen rectangle in screen (not model)
    /// coordinates where the lower-left corner of the panel is (0,0)
    void PointToNode(int id, wxRect tip_rect, float sec);

    void    SetHost(IPhyloTreeRendererHost* pHost);
#ifndef TEXTURED_FONTS
    void    SetFont(CGlBitmapFont * font) {m_pLabelFont = font;}
#else
    void    SetFont(CGlTextureFont * font) {m_pLabelFont = font;}
#endif
    void    SetZoomHandler(CMouseZoomHandler* mz) { m_pMouseZoomHandler = mz; }

    void    SetDistRendering(bool flag) {m_bDistMode = flag;}
    bool    GetDistRendering(void)      {return m_bDistMode;}
    virtual bool SupportsDistanceRendering() const { return true; }

    void    SetAdaptiveMargins(bool flag)  {m_bAdaptiveMargins = flag;}
    bool    GetAdaptiveMargins(void)       {return m_bAdaptiveMargins;}

    bool    GetSimplification(void)        {return m_bSimplification;}
    void    SetSimplification(bool flag)   {m_bSimplification = flag;}

    bool    GetSplinesRendering(void)      {return m_bUseSplines;}
    void    SetSplinesRendering(bool flag) {m_bUseSplines = flag;}

    bool    GetRegenerateTexture(void)     {return m_bRegenerateTexture;}
    void    SetRegenerateTexture(bool flag){m_bRegenerateTexture = flag;}

    bool    GetZoomablePrimitives(void)         {return m_bZoomablePrimitives; }
    void    SetZoomablePrimitives(bool flag)    {m_bZoomablePrimitives = flag; }

    bool    GetHighlightCollapsed(void) { return m_bHighlightCollapsed; }
    void    SetHighlightColllapsed(bool flag) { m_bHighlightCollapsed = flag; }

    void    SetLabelViewPct(float pct) { m_LabelViewPct = pct; }
    float   GetLabelViewPct() const { return m_LabelViewPct; }

    bool    GetRenderScale() const { return m_RenderScale; }
    void    SetRenderScale(bool s) { m_RenderScale = s; }

    virtual bool    CanDragNodes(void);

    /// Allow caller to indicate layout has changed from one type to another
    /// (e.g. rect clad. to force)
    virtual void    StartRendering(bool b) { if (b) m_SwitchedLayout = true; }

    // return true if the tree has been laid out at least once
    virtual ELayoutStatus ValidLayout() { return m_ValidLayout; }

    void    OnLeftDown(wxMouseEvent& event);
    void    OnRightDown(wxMouseEvent& event);
    void    OnLeftDblClick(wxMouseEvent& event);
    void    OnLeftUp(wxMouseEvent& event);
    void    OnMotion(wxMouseEvent& event);
    void    OnTimer(wxTimerEvent& event);
    //void    OnKeyEvent(wxKeyEvent& event);
    IGenericHandlerHost* GetGenericHost();
    // IPhyloTreeLOD
    virtual TVPUnit DistanceBetweenNodes(void);
    bool    IsPrimitivesScalable(void)          {return m_bZoomablePrimitives;  }
    TVPUnit GetNodeSize(CPhyloTreeNode * node)  {return x_NodeSize(node);       }

    CRgbaColor GetNodeColor(CPhyloTreeNode * node, bool bLabel = false);
	string     GetEdgeColor(CPhyloTreeNode * node);

    CPhyloTreeNode * GetHoverNode(TVPPoint point);

    CGlTexture * GetTexture(float& xcoord_limit, float& ycoord_limit);

    virtual string  GetDescription(void) = 0;
    virtual void    BufferedRender(CGlPane& pane, CPhyloTreeDataSource& ds);
    
    /// Returns false if rendering does not occur (e.g. due to active data
    /// sychronization).
    virtual bool    Render(CGlPane& pane, CPhyloTreeDataSource& ds);
    virtual void    RenderForHardcopy(CGlPane& pane, CPhyloTreeDataSource& ds);
    virtual void    RenderPdf(CGlPane& pane, CPhyloTreeDataSource& ds, CRef<CPdf> pdf);
    virtual void    Layout(CPhyloTreeDataSource& ds, CGlPane& p);
    virtual void    Redraw(void);
    virtual void    RemoveCurrentDataSource();
    virtual void    UpdateDataSource(CPhyloTreeDataSource& ds, CGlPane& p);

    /// @name IGlEventHandler implementaion
    /// @{
    virtual void    SetPane(CGlPane* pane);
    virtual wxEvtHandler*    GetEvtHandler();
    /// @}

    const TModelRect & GetRasterRect(void);
    void UpdateRasterRectLimits(void);

    void  SetScheme(CPhyloTreeScheme & sl)
    {
        m_SL.Reset(&sl);
        m_Label.SetScheme(sl);
    }

    CPhyloTreeScheme & GetScheme(void)     {return m_SL.GetObject();     }

    void SetModelDimensions(double width, double height)
    {
        m_DimX = width;
        m_DimY = height;
    }

    void    ComputeViewingLimits(CGlPane& pane, 
                                 bool viewing_only = false);

    void SetRenderingOption(TRenderingOption opt, bool bVal);

    // For most views scale glyph should be in LL (if not LR).
    virtual bool IsDistanceBarLowerLeft() const { return true; }

    // metrics
    const double GetDimX(void) {return m_DimX;}
    const double GetDimY(void) {return m_DimY;}


    // minimum dimensions preserving full visibility
    static TVPRect GetMinDimensions(CPhyloTreeDataSource & ds,
                                    IPhyloTreeRenderer   & rr,
                                    CPhyloTreeScheme     & sc);

    virtual TVPUnit GetDefaultNodeSize(CPhyloTreeNode * node);

    virtual void  Calculate(CPhyloTreeNode * /*node*/) {}

protected:

    // data
    CPhyloTreeDataSource * m_DS;

    //style
    CRef <CPhyloTreeScheme> m_SL;

    double  m_DimX,  m_DimY;
    double  m_xStep, m_yStep;
    double  m_NormDistance;

#ifndef TEXTURED_FONTS
    CGlBitmapFont   *m_pLabelFont;
#else
    CGlTextureFont   *m_pLabelFont;
#endif

    IPhyloTreeRendererHost *m_pHost;
    CMouseZoomHandler*  m_pMouseZoomHandler;

    // scene texture
    CRef<CGlTexture> m_Texture;

    CGlCurve<CCurveBezier> m_Curve;

    CPhyloTreeLabel     m_Label;
    CPhyloTreeLabel     m_LabelExt;

    TModelRect          m_RasterRect;


    /// event handling state
    enum    EState {
        eIdle,
        eSelPoint, /// selecting a single glyph by point
        eSelRect,  /// selecting multiple glyphs by rectangle
        eDragNode
    };

    /// label alignment
    enum    ELabelAlign {
        eLeft,
        eRight
    };

    /// event handling
    CGlPane         *m_pPane;
    EState           m_State;
    wxPoint          m_StartPoint;
    wxPoint          m_LastPos;
    wxPoint          m_DragPoint;
//    Fl_Cursor        m_Cursor;
    vector<int>      m_SelectedIDs;
    CPhyloTreeNode*  m_ActiveTooltipNode;
    bool             m_bDistMode;
    bool             m_bMoveDuringSelection;
    bool             m_bAdaptiveMargins;
    bool             m_bSimplification;
    bool             m_bUseSplines;
    bool             m_bRegenerateTexture;
    bool             m_bZoomablePrimitives; // nodes, lines, etc
    bool             m_bAutosizeLabels;
    bool             m_bHighlightCollapsed;
    wxTimer          m_EffectsTimer;
    bool             m_SwitchedLayout;
    
    /// When sizing the tree to window, this is the maximum % of screen
    /// area (width) that will be allocated to labels on either side
    /// of the screen (some views, e.g. radial, have labels on both sides).
    float            m_LabelViewPct;
    
    /// If true render a scale legend at the bottom of the screen.
    bool             m_RenderScale;

    ELayoutStatus   m_ValidLayout;

    int   m_CursorId;

    /// Structure holds a 'pointer' from an active tooltip to its
    /// corresponding node
    struct NodePointer {
        NodePointer() 
            : m_NodeID(0)
            , m_Node(NULL)
            , m_TipCenter((TModelUnit)0, (TModelUnit)0)
            , m_Duration(0.0f)
        {}

        /// ID of the node to point to
        IPhyNode::TID m_NodeID;

        /// Pointer to the node with id m_NodeID
        CPhyloTreeNode* m_Node;

        /// Rectangle defining tooltip in screen coords
        wxRect m_TipRect;
        /// Center of tooltip widget, in screen coords with (0,0) in lower left
        TModelPoint  m_TipCenter;

        /// Timer to track duration of arrow display
        CStopWatch m_Timer;
        double m_Duration;
    };
    std::vector<NodePointer> m_NodePointers;   

    CGLFrameBuffer*  m_MinimapBuffer;

    virtual TModelUnit x_DistanceBetweenNodes(TModelUnit scaley);


    /// A coord (in one of +x,-x,+y,-y) and pixel offset in the same direction
    struct ProjectionElement {
        ProjectionElement() : m_ProjectionCood(0.0f) {}
        ProjectionElement(float p, float o) 
            : m_NodeCoord(p)
            , m_PixelOffset(o)
            , m_ProjectionCood(0.0f) {}

        /// Node x or y position depending on whether this extent is vert. or horiz.
        float m_NodeCoord;

        /// Pixel offset (direction can +x,-x,+y,-y) relative to the x or y position
        /// above. The owning class tracks according to offset direction.
        float m_PixelOffset;

        /// Given a particular visible area and viewport size, position to which
        /// this node projects inside the viewport
        float m_ProjectionCood;
        //std::string m_Text;
    };

    /// This is a set of positions (for nodes) and associated x or y pixel offset
    /// values.  The owning class determines whether this refers to x or y
    /// coordninates.
    class CExtentDimension
    {
    public:
        /// Compare the new coordinate and pixel offset with the contents
        /// of m_Extents. If node_coord is larger than any node coordinate
        /// in the array, or if pixel_offset is longer than any alement
        /// in the array with the same or larger node coord, add it.
        void  UpdateExtent(float node_coord,
            float pixel_offset,
            bool has_label);

        /// write out values for debugging purposes
        void DumpExtent(const string& dim);

        /// Sorted (in decreasing order) set of coordinates and pixel offsets
        vector<ProjectionElement> m_Extents;
    };

    /// This class extracts for each of the 4 directions (+x,-x,+y,-y) the maximum
    /// coordinates to which any element in the graph projects.  For node positions 
    /// this is easy since it is simply max/min x/y values.  But since text (labels)
    /// do not scale with the graph, they may extend beyond graph edges at some
    /// zoom levles but not others. We therefore also find the longest labels
    /// in each of the 4 directions. (in each of the 4 directions we add any node-
    /// label combo where the label is longer in that direction than labels
    /// of all other nodes further (left/right/up/down) in the same direction)
    class CProjectionExtents {
    public:
        /// Recursively iterate over phylo tree and find all nodes that may effect
        /// the projection size in x and y.
        void  GetTextExtents(const CPhyloTreeScheme& sl,
            CPhyloTreeLabel& labeler,
            CGlPane& pane,
            CPhyloTreeNode * node,
            TModelRect& zoomable_extent);

        /// Combine together m_MaxX and m_MinX and m_MaxY and m_MinY
        void ConsolidateExtents(vector<ProjectionElement>& xexts,
            vector<ProjectionElement>& yexts);

        /// Write output to log for debugging
        void DumpExtents();

        /// the 'rightmost' nodes and labels (includes the node with the largest
        /// x value, and any nodes to the left that have labels longer than any
        /// node to the right of them
        CExtentDimension m_MaxX;
        /// Same as for m_MaxX values, but in y dimension
        CExtentDimension m_MaxY;
        /// the 'leftmost' nodes, except that signs are flipped in order to allow
        /// uniform processing in all dimensions (ConsolidateExtents() unflips)
        CExtentDimension m_MinX;
        /// the 'bottommost' nodes, with signs flipped.
        CExtentDimension m_MinY;
    };

    void    x_FindBestSize(CGlPane& pane,
                           std::vector<ProjectionElement>& xexts,
                           std::vector<ProjectionElement>& yexts,
                           TModelRect& limits);
    void    x_RenderSelection(CGlPane& pane);
    void    x_RenderScaleMarker(CGlPane& pane);
    void    x_RenderScaleMarker(CGlPane& pane, CGlRender& gl);
    void    x_RenderTooltipHints(CGlPane& pane);

    void    x_RenderLine(double x1, double y1, double x2, double y2,
                         IPhyGraphicsNode::TSelectedState state,
                         string force_color = "");

    void    x_RenderNode(CPhyloTreeNode * node);
    void    x_DrawBoundingAreas(CPhyloTreeNode * node);

    /// Draw the bounding region for the node (if provided via a feature entry)
    void    x_RenderBoundary(CPhyloTreeNode * node);

    void    x_RenderSpline(TPoint pt1,
                           TPoint pt2,
                           TPoint pt3,
                           TPoint pt4,
                           IPhyGraphicsNode::TSelectedState state,
                           string force_color,
                           bool bPseudo = false);

    void    x_RenderLineVbo(CGlVboNode* render_lines,
                            double x1, double y1, double x2, double y2,
                            IPhyGraphicsNode::TSelectedState state,
                            string force_color = "");

    void    x_RenderNodeVbo(CPhyloTreeNode * node, CGlVboNode* render_nodes);

    void    x_DrawBoundingAreasVbo(CPhyloTreeNode * node);
    
    void    x_RenderSplineVbo(vector<CRgbaColor>& edge_colors,
                              TPoint pt1,
                              TPoint pt2,
                              TPoint pt3,
                              TPoint pt4,
                              IPhyGraphicsNode::TSelectedState state,
                              string force_color);
    void    x_RenderPseudoSplineVbo(CPhyloTreeNode* child_node,
                                    CGlVboNode* render_lines,
                                    const CVect2<float>& pt1,
                                    const CVect2<float>&  mid_point,                             
                                    const CVect2<float>&  pt2,
                                    IPhyGraphicsNode::TSelectedState state);

    void    x_DrawNodesVbo(CPhyloTreeNode * node, CGlVboNode* render_nodes);

    void    x_GenerateTexture(void);
    bool    x_KindOfHuge(void);

    void    x_DrawNodes(CPhyloTreeNode * node);

    /// user event handlers - translate event to signals
    //int x_OnMousePush(CGUIEvent& event);
    //int x_OnMouseDrag(CGUIEvent& event);
    //int x_OnMouseRelease(CGUIEvent& event);
    int x_OnMouseMove(void);
    int x_OnKeyDown(void);
    int x_OnKeyUp(void);

    void    x_OnSelectCursor(void);
    bool    x_SelectByPoint(CPhyloTreeNode * node, bool toggle);

    /// Returns true if selection status changed (or may have changed..)
    bool    x_SelectByRect(CPhyloTreeNode * node, bool toggle);
    bool    x_SelectByRectBuffered(bool toggle);
    void    x_SelectByRectRecursive(CPhyloTreeNode * node, bool toggle);

    CPhyloTreeNode * x_TestForNode(CPhyloTreeNode * node, wxPoint pos);
    CPhyloTreeNode * x_TestForNodeBuffered(wxPoint pos);
    CPhyloTreeNode * x_TestForNodeRecursive(CPhyloTreeNode * node, wxPoint pos);

    virtual void x_Layout(CPhyloTreeDataSource& ds) = 0;
    virtual void x_Render(CGlPane& pane, CPhyloTreeDataSource& ds) = 0;
    virtual void x_ComputeNodeBoundary(CPhyloTreeNode * node, 
                                       bool compute_boundary,
                                       CBoundaryPoints& boundary_pts,
                                       const string& layout_type);


    TVPUnit x_NodeSize(CPhyloTreeNode * node);


    virtual void x_RenderVbo(CPhyloTreeDataSource& /*ds*/) {}

    /// For performance analysis (time to update VBOs after topology change)
    float m_BufferdRenderTime;
};
END_NCBI_SCOPE

#endif  // GUI_WIDGETS_PHYLO_TREE___PHYLO_TREE_RENDER__HPP
