#ifndef __GUI_WIDGETS_PHYLO_TREE___GRAPHICS_MODEL__HPP
#define __GUI_WIDGETS_PHYLO_TREE___GRAPHICS_MODEL__HPP

/*  $Id: tree_graphics_model.hpp 27314 2013-01-25 16:37:25Z falkrb $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Bob Falk
 *
 * File Description:
 *
 */

#include <gui/gui_export.h>
#include <gui/opengl/glmodel2d.hpp>
#include <gui/opengl/glrect.hpp>
#include <gui/opengl/glpane.hpp>
#include <gui/widgets/phylo_tree/tree_collision_model2d.hpp>
#include <gui/widgets/phylo_tree/phylo_tree_node.hpp>
#include <gui/widgets/phylo_tree/phylo_tree_scheme.hpp>
#include <gui/widgets/phylo_tree/tree_graphics_node.hpp>
#include <gui/print/pdf.hpp>

BEGIN_NCBI_SCOPE 


////////////////////////////////////////////////////////////////////////////////
/// class CTreeBoundaryNode
/// 
class NCBI_GUIWIDGETS_PHYLO_TREE_EXPORT CTreeBoundaryNode : public  CGlRenderNode {
public:
    CTreeBoundaryNode() : CGlRenderNode() {}
    virtual ~CTreeBoundaryNode() {}

    void  SetScheme(CPhyloTreeScheme& sl) { m_SL.Reset(&sl); }

    void Clear() { m_BoundaryNodes.clear(); }
    vector<CPhyloTreeNode*>& GetNodes() { return m_BoundaryNodes; }

    void RenderPdf(CGlPane& pane, CRef<CPdf>& pdf);

protected:

    virtual void x_Render();

    /// Vector of all nodes that have a boundary defined
    vector<CPhyloTreeNode*> m_BoundaryNodes;
    CRef<CPhyloTreeScheme> m_SL;

    CRef<CGlVboNode> m_BoundaryTris;
    CRef<CGlVboNode> m_BoundaryEdges;
};


////////////////////////////////////////////////////////////////////////////////
/// class CTreeTriFanNode
/// 
class NCBI_GUIWIDGETS_PHYLO_TREE_EXPORT CTreeTriFanNode : public CGlVboNode
{
public:
    /// 
    CTreeTriFanNode() 
        : CGlVboNode(GL_TRIANGLE_FAN)
        , m_VisRect(-FLT_MAX, -FLT_MAX, FLT_MAX, FLT_MAX)
        , m_DefaultSize(1.0f) 
        , m_LabelVis(false) {}

    /// Delete all buffers
    virtual ~CTreeTriFanNode() {}

    /// Set the visible region so that we only draw visible nodes
    void SetVisibilityInfo(const CGlRect<float>& r, float default_size) 
    { m_VisRect = r; m_DefaultSize = default_size; }  

    /// Set default label visibility (if false, no labels rendered, if true, labels
    /// which are turned on and visible are rendered)
    void SetLabelVisibility(bool b) { m_LabelVis = b; }
    bool GetLabelVisibility() const { return m_LabelVis; }
    
protected:

    /// Visible region to prevent drawing nodes which are not visible
    CGlRect<float> m_VisRect;

    /// Default (unscaled) node size (used for visibility checkin
    float m_DefaultSize;

    /// If false, no labels are rendered, if true, designated labels rendered
    bool m_LabelVis;
};

////////////////////////////////////////////////////////////////////////////////
/// class CTreeGraphicsModel
/// Model for rendering tree graphics data.  Only holds data currently needed
/// to render the model.  Actual rendering is all done by objects derived from
/// CGlRenderNode.  Where there is a lot to render (nodes, edges) that is done
/// by vertex buffer objects.  Where there is less (marker nodes) that may be 
//  a simple class derived from CGlRenderNode for that purpose alone.
/// 
/// For nodes and edges in the model, they are currently first copied to arrays
/// (m_Nodes and m_Edges) prior to rendering.  This allows the data to be shared
/// with collision detection.
///
class NCBI_GUIWIDGETS_PHYLO_TREE_EXPORT  CTreeGraphicsModel : public CGlModel2D
{
public:
    CTreeGraphicsModel();
    virtual ~CTreeGraphicsModel();

    /// Allocate any default buffers etc.
    virtual void Init();

    /// Draw all the nodes
    virtual void Render(CGlPane& pane);

    /// Draw to PDF file
    virtual void RenderPDF(CGlPane& pane, CRef<CPdf>& pdf);

    /// Update buffers as needed.
    virtual void SyncBuffers();

    /// Clear arrays prior to re-synching with the model
    virtual void ClearArrays();

    /// Set rendering scheme for tree (some nodes need this to render)
    void  SetScheme(CPhyloTreeScheme& sl);

    /// Get all the nodes (entries point to nodes in phylo tree)
    vector<CTreeGraphicsNode*>& GetNodes() { return m_Nodes; }
    /// Get all the edges (entries point to nodes in phylo tree)
    vector<CTreeGraphicsEdge>& GetEdges()  { return m_Edges; }

    CTreeBoundaryNode* GetBoundaryNode() { return m_BoundaryNode; }
    CTreeTriFanNode* GetTriFanNode() { return m_TriFanNode; }
    CGlVboNode* GetTriFanNodePass2() { return m_TriFanNodePass2; }

    CTreeCollisionModel2D& GetCollisionData() { return m_CollisionData; }
    const CTreeCollisionModel2D& GetCollisionData() const { return m_CollisionData; }

protected:
    void x_RenderNodes(CGlPane& pane,
                       const CVect3<float>& scale);
    void x_RenderNodeLabelsPdf(CGlPane& pane,
                               CRef<CPdf>& pdf,
                               CVect3<float> scale);
    void x_RenderNodesPdf(CGlPane& pane,
                          CRef<CPdf>& pdf,
                          CVect3<float> scale);

    void x_RenderBoundariesPdf(CGlPane& pane,
                          CRef<CPdf>& pdf,
                          const CVect3<float>& scale);

    CVect2<float> x_Project(CVect2<float>& p);

    //spatial 2d subdivision
    CTreeCollisionModel2D m_CollisionData;

    vector<CTreeGraphicsNode*> m_Nodes;
    vector<CTreeGraphicsEdge> m_Edges;

    CTreeBoundaryNode* m_BoundaryNode;
    CTreeTriFanNode* m_TriFanNode;
    CGlVboNode* m_TriFanNodePass2;


    CRef<CPhyloTreeScheme> m_SL;

    float m_RenderTime;
    float m_LineRenderTime;
    float m_NodesRenderTime;
    int m_NodeRenderCount;
};

END_NCBI_SCOPE

#endif // __GUI_WIDGETS_PHYLO_TREE___GRAPHICS_MODEL__HPP