#ifndef __GUI_WIDGETS_PHYLO_TREE___GRAPHICS_NODE__HPP
#define __GUI_WIDGETS_PHYLO_TREE___GRAPHICS_NODE__HPP

/*  $Id: tree_graphics_node.hpp 27277 2013-01-18 20:50:36Z falkrb $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Bob Falk
 *
 * File Description:
 *
 */

#include <gui/gui_export.h>
#include <gui/opengl/glrect.hpp>
#include <gui/opengl/glpane.hpp>
#include <gui/utils/vect2.hpp>
#include <gui/utils/vect3.hpp>
#include <gui/utils/rgba_color.hpp>
#include <gui/widgets/phylo_tree/phylo_tree_node.hpp>
#include <gui/widgets/phylo_tree/phylo_tree_scheme.hpp>

BEGIN_NCBI_SCOPE 

class CGlTextureFont;

////////////////////////////////////////////////////////////////////////////////
/// class CTreeGraphicsNode
/// 
struct CTreeGraphicsNode {
    CTreeGraphicsNode() : m_Node(NULL), m_Font(NULL), m_AlreadyDrawn(false) {}
    CTreeGraphicsNode(CPhyloTreeNode* n,
                      const CRgbaColor& node_color,
                      const CRgbaColor& text_color,                     
                      const string& t,
                      float node_size,
#ifndef TEXTURED_FONTS
                      const CGlBitmapFont* f);
#else
                      const CGlTextureFont* f);
#endif
    
    /// Draw the label using scale to compute its offset from m_BasePos
    void Render(CGlPane& pane, const CVect3<float>& scale, bool truncate_labels);

    /// Return the drawing position and (possibly truncated) text
    CVect2<float> GetTextOut(CGlPane& pane, 
                             const CVect3<float>& scale,
                             bool truncate_labels,
                             string& text_out);

    /// Find the rectangle for the label given the current zoom level (scale)
    void GetLabelRect(const CVect2<float>& scale, 
                         CVect2<float>& lower_left,
                         CVect2<float>& upper_right);

    /// Find the rectangle for the node and label combined given the 
    /// current zoom level (scale)
    void GetBoundingRect(const CVect2<float>& scale, 
                         CVect2<float>& lower_left,
                         CVect2<float>& upper_right);

    /// Return labels minimal X coordinate
    float GetMinX(const CVect2<float>& scale);

    /// Return underlying node from data source
    CPhyloTreeNode* GetSourceNode() { return m_Node; }

    /// Flag to renderer to determine whether to draw label
    void SetVisible(bool b) {m_Visible = b;}
    bool GetVisible() const { return m_Visible; }

    /// set/get label text
    void SetLabel(const string& t) { m_Text = t; }
    const string& GetLabel() const { return m_Text; }

    /// set/get node size
    void SetNodeSize(float s) { m_NodeSize = s; }
    const float GetNodeSize() const { return m_NodeSize; }

    /// Set/get node color
    void SetNodeColor(const CRgbaColor& c) { m_NodeColor = c; }
    CRgbaColor GetNodeColor() const { return m_NodeColor; }

    /// Set/get text color
    void SetColor(const CRgbaColor& c) { m_FgColor = c; }
    CRgbaColor GetColor() const { return m_FgColor; }

    /// Flag set after each drawing and cleared when rendering frame done
    void SetAlreadyDrawn(bool b) { m_AlreadyDrawn = b; }
    bool GetAlreadyDrawn() const { return m_AlreadyDrawn; }

    /// Set/get pointer to font to use in drawing text
#ifndef TEXTURED_FONTS
    void SetFont(const CGlBitmapFont* f) { m_Font = f; }
    const CGlBitmapFont* GetFont() const { return m_Font; }
#else
    void SetFont(const CGlTextureFont* f) { m_Font = f; }
    const CGlTextureFont* GetFont() const { return m_Font; }
#endif

protected:
    // Pointer to node
    CPhyloTreeNode* m_Node;
    CRgbaColor m_NodeColor;
    /// Color of text to be drawn
    CRgbaColor m_FgColor;
    /// Text to render
    string m_Text;
    // Node scale factor
    float m_NodeSize;    
    /// Font to use to draw the text
#ifndef TEXTURED_FONTS
    const CGlBitmapFont* m_Font;
#else
    const CGlTextureFont* m_Font;
#endif
    /// True if node is visible
    bool m_Visible;

    /// Text is alpha-blended so we add a flag to prevent drawing multiple times
    bool m_AlreadyDrawn;
};

////////////////////////////////////////////////////////////////////////////////
/// class CTreeGraphicsEdge
/// 
struct CTreeGraphicsEdge {
    CTreeGraphicsEdge() {}
    CTreeGraphicsEdge(const CVect2<float>& p1, const CVect2<float>& p2) 
        : m_Pos1(p1)
        , m_Pos2(p2) {}

    // Position
    CVect2<float> m_Pos1;
    CVect2<float> m_Pos2; 
};


END_NCBI_SCOPE

#endif // __GUI_WIDGETS_PHYLO_TREE___GRAPHICS_NODE__HPP