#ifndef GUI_WIDGETS_SEQ_GRAPHIC___ALIGNMENT_GLYPH__HPP
#define GUI_WIDGETS_SEQ_GRAPHIC___ALIGNMENT_GLYPH__HPP

/*  $Id: alignment_glyph.hpp 26576 2012-10-10 16:27:30Z wuliangs $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Vlad Lebedev, Liangshou Wu
 *
 * File Description:
 *   CAlignGlyph -- utility class to layout alignments and hold CAlnVec objects
 */


#include <corelib/ncbiobj.hpp>
#include <gui/gui.hpp>
#include <gui/widgets/seq_graphic/seq_glyph.hpp>
#include <gui/widgets/seq_graphic/object_based_glyph.hpp>
#include <gui/widgets/seq_graphic/alignment_conf.hpp>
#include <gui/widgets/seq_graphic/alignment_score.hpp>
#include <gui/utils/rgba_color.hpp>

#include <gui/widgets/seq_graphic/alngraphic_ds.hpp>
#include <gui/objutils/label.hpp>
#include <objects/seqloc/Seq_interval.hpp>
#include <objtools/alnmgr/alnvec.hpp>
#include <objmgr/seq_align_handle.hpp>
#include <math.h>

/** @addtogroup GUI_OBJUTILS
 *
 * @{
 */


BEGIN_NCBI_SCOPE


class NCBI_GUIWIDGETS_SEQGRAPHIC_EXPORT CAlignGlyph :
    public CSeqGlyph,
    public IObjectBasedGlyph
{
public:
    typedef vector< CRef<CAlignGlyph> > TAlnList;
    typedef pair<CConstRef<CObject>, int> TAlnAnchorPair;

    enum EAlignSpacing {
        eOne2One,       // DNA to DNA or protein to protein
        eOne2Three,     // DNA to protein
        eThree2One      // Protein to DNA
    };

    // ctors
    CAlignGlyph(CConstRef<IAlnGraphicDataSource> aln_mgr,
                 const objects::CSeq_align_Handle& orig_align,
                 const objects::CSeq_align& mapped_align,
                 int orig_anchor);

    /// @name CSeqGlyph methods
    /// @{
    virtual bool NeedTooltip(const TModelPoint& p, string& tt, string& t_title) const;
    virtual void GetTooltip(const TModelPoint& p, string& tt, string& t_title) const;
    virtual void GetHTMLActiveAreas(TAreaVector* p_areas) const;
    virtual TSeqRange   GetRange(void) const;
    virtual bool IsClickable() const;
    /// @}

    /// @name IObjectBasedGlyph interface implement
    /// @{
    virtual const objects::CSeq_loc& GetLocation(void) const;
    virtual CConstRef<CObject> GetObject(TSeqPos pos) const;
    virtual void GetObjects(vector<CConstRef<CObject> >& objs) const;
    virtual bool HasObject(CConstRef<CObject> obj) const;
    virtual string GetSignature() const;
    virtual const TIntervals& GetIntervals(void) const;
    /// @}

	const IAlnGraphicDataSource&      GetAlignMgr(void) const;
    const objects::CSeq_align&        GetAlignment(void) const;
    const objects::CSeq_align_Handle& GetOrigAlignment(void) const;
    TAlnAnchorPair                    GetAlnAnchorPair(void) const;
    int                               GetOrigAnchor(void) const;

    TModelUnit GetBarHeight() const;

    void SetConfig(const CAlignmentConfig* conf);
    void SetScore(const ISGAlnScore& score);

    void DrawPWAlignElem(const TModelRect& rcm, bool selected,
        bool show_label, const CRgbaColor& c_fg,
        const CRgbaColor& c_seq, const CRgbaColor& c_mis) const;

    void GetTitle(string* title, CLabel::ELabelType type) const;
    
    bool IsNegative() const;

    void SetShowScore();
    bool IsSimplified() const;

protected:
    virtual void x_Draw() const;
    virtual void x_UpdateBoundingBox();

private:
    void x_DrawPWAlign() const;
    void x_DrawMultiAlign() const;
    void x_DrawMultiAlignLabel(int row, TModelUnit y) const;
    void x_DrawAlignRowSequence(TModelUnit center_y, int row,
        const CRgbaColor& c_fg, const CRgbaColor& c_seq,
        const CRgbaColor& c_mis) const;

    void x_RenderNonAveragableScores(TModelUnit yy, TSignedSeqRange aln_range,
        TSeqRange seq_range, TModelUnit height, int row) const;

    void x_RenderAveragableScores(TModelUnit yy, TSignedSeqRange aln_range,
        TSeqRange seq_range, TModelUnit height, int row) const;

    bool x_IsPWAlign() const;
    void x_CalcIntervals(void);
    void x_DrawBorder(float x1, float y1, float x2, float y2) const;
    bool x_IsPolyA(const objects::CBioseq_Handle& bsh,
        TSeqPos start, TSeqPos stop) const;

    bool x_ShowLabel() const;
    bool x_ShowScore() const;
    bool x_ShowUnalignedTails(TModelUnit align_len) const;

protected:
    /// pointer to the actual alignment information.
	CConstRef<IAlnGraphicDataSource> m_AlnMgr;
    CConstRef<objects::CSeq_align>   m_SeqAlign;
    objects::CSeq_align_Handle       m_OrigSeqAlign;

    /// The anchor row number in original Seq-algin, not alnmgr.
    int                              m_OrigAnchor;
    CRef<objects::CSeq_loc>          m_Location;

    /// intervals.
    TIntervals                       m_Intervals;

    CConstRef<CAlignmentConfig>      m_Config;
    CConstIRef<ISGAlnScore>          m_Score;

    // Flag indicating if the alignment is simplifed and/or colored.
    // There are two cases we could simplify an alingment:
    //   1. alignment size is small on screen ( < 4 pixels )
    //   2. std-seg alignment with different segment length
    //      across rows
    bool                             m_ShowScore;
};


///////////////////////////////////////////////////////////////////////////////
/// Inline methods

inline
const IAlnGraphicDataSource& CAlignGlyph::GetAlignMgr(void) const
{
    return *m_AlnMgr;
}

inline
const objects::CSeq_align& CAlignGlyph::GetAlignment(void) const
{
    return *m_SeqAlign;
}

inline
CAlignGlyph::TAlnAnchorPair CAlignGlyph::GetAlnAnchorPair(void) const
{
    return TAlnAnchorPair(GetObject(0), m_OrigAnchor);
}

inline
int CAlignGlyph::GetOrigAnchor(void) const
{
    return m_OrigAnchor;
}

inline
const objects::CSeq_align_Handle& CAlignGlyph::GetOrigAlignment(void) const
{
    return m_OrigSeqAlign;
}

inline
TModelUnit CAlignGlyph::GetBarHeight() const
{
    TModelUnit bar_h = m_Config->m_BarHeight;
    if (m_Config->m_CompactMode == CAlignmentConfig::eCompact) {
        bar_h = floor(bar_h * 0.4);
    } else if (m_Config->m_CompactMode == CAlignmentConfig::eExtremeCompact) {
        bar_h = floor(bar_h * 0.2);
    }
    return bar_h;
}

inline
void CAlignGlyph::SetConfig(const CAlignmentConfig* conf)
{
    m_Config.Reset(conf);
}

inline
void CAlignGlyph::SetScore(const ISGAlnScore& score)
{
    m_Score.Reset(&score);
}

inline
bool CAlignGlyph::x_IsPWAlign() const
{
    return m_AlnMgr->GetNumRows() == 2;
}

inline
bool CAlignGlyph::IsNegative() const
{
    if (x_IsPWAlign()) {
        return m_AlnMgr->IsNegativeStrand(1) !=
            m_AlnMgr->IsNegativeStrand(0);
    }
    return false;
}

inline
void CAlignGlyph::SetShowScore()
{
    m_ShowScore = true;
}

inline
bool CAlignGlyph::IsSimplified() const
{
    return m_AlnMgr->IsSimplified();
}


END_NCBI_SCOPE

/* @} */

#endif  // GUI_WIDGETS_SEQ_GRAPHIC___ALIGNMENT_GLYPH__HPP
