#ifndef GUI_WIDGETS_SEQ_GRAPHIC___ALNGRAPHIC_DS__HPP
#define GUI_WIDGETS_SEQ_GRAPHIC___ALNGRAPHIC_DS__HPP

/*  $Id: alngraphic_ds.hpp 26349 2012-08-30 15:32:11Z wuliangs $
* ===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data, the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties, express or implied, including
*  warranties of performance, merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================
*
* Authors:  Liangshou Wu
*
* File Description:
*       Borrowed from alnmulti_ds
*/

#include <corelib/ncbistd.hpp>
#include <corelib/ncbiobj.hpp>
#include <objects/seqloc/Seq_loc.hpp>
#include <objmgr/bioseq_handle.hpp>
#include <objtools/alnmgr/aln_explorer.hpp>
#include <objtools/alnmgr/alnmap.hpp>

#include <util/sequtil/sequtil_manip.hpp>
#include <util/range_coll.hpp>


BEGIN_NCBI_SCOPE


///////////////////////////////////////////////////////////////////////////////
/// IAlnGraphicDataSource - interface to a data source representing an abstract
/// alignment used in graphical view.
class NCBI_GUIWIDGETS_SEQGRAPHIC_EXPORT IAlnGraphicDataSource
    : public CObjectEx
{
public:
    /// Cetantiy of having unaligned polyA tail.
    enum EPolyATail {
        ePolyA_Yes,
        ePolyA_No,
        ePolyA_Unknown
    };

    enum ECigarOp 
    {
        eCigar_M = 'M',
        eCigar_I = 'I',
        eCigar_D = 'D',
        eCigar_N = 'N',
        eCigar_S = 'S',
        eCigar_H = 'H',
        eCigar_P = 'P',
        eCigar_Eq = '=',
        eCigar_X = 'X'
    };

    typedef pair<ECigarOp, size_t> TCigarPair;
    typedef vector<TCigarPair>     TCigar;

    typedef IAlnExplorer::TSignedRange TSignedRange;
    typedef IAlnExplorer::TNumrow    TNumrow;
    typedef IAlnExplorer::TResidue   TResidue;
    typedef IAlnExplorer::ESearchDirection   TSearchDirection;
    typedef CRangeCollection<TSeqPos>   TRangeColl;

    virtual IAlnExplorer::EAlignType  GetAlignType() const
    {
        if (GetNumRows() < 2)
            return IAlnExplorer::fInvalid;

        // not sure if this is the best solution
        IAlnExplorer::EAlignType type = IAlnExplorer::fMixed;
        switch (GetBioseqHandle(0).GetBioseqCore()->GetInst().GetMol())
        {
        case objects::CSeq_inst::eMol_dna:
        case objects::CSeq_inst::eMol_rna:
        case objects::CSeq_inst::eMol_na:
            type = IAlnExplorer::fDNA;
            break;

        case objects::CSeq_inst::eMol_aa:
            type = IAlnExplorer::fProtein;
            break;

        default:
            break;
        }

        if (type == IAlnExplorer::fMixed) {
            return type;
        }

        TNumrow row = 1;
        for (;  row < GetNumRows();  ++row) {
            IAlnExplorer::EAlignType this_type = IAlnExplorer::fMixed;
            switch (GetBioseqHandle(row).GetBioseqCore()->GetInst().GetMol())
            {
            case objects::CSeq_inst::eMol_dna:
            case objects::CSeq_inst::eMol_rna:
            case objects::CSeq_inst::eMol_na:
                this_type = IAlnExplorer::fDNA;
                break;

            case objects::CSeq_inst::eMol_aa:
                this_type = IAlnExplorer::fProtein;
                break;

            default:
                break;
            }
            if (this_type != type) {
                return IAlnExplorer::fMixed;
            }
        }

        return type;        
    }

    /// number of rows in alignment
    virtual TNumrow  GetNumRows(void) const = 0;
    virtual TSeqPos  GetAlnStart(void) const = 0;
    virtual TSeqPos  GetAlnStop(void) const = 0;
    virtual TSeqPos GetSeqStart(TNumrow row) const = 0;
    virtual TSeqPos GetSeqStop(TNumrow row) const = 0;

    virtual TSignedRange GetAlnRange(void) const = 0;
    virtual TSignedRange GetSeqAlnRange(TNumrow row) const = 0;

    virtual TSeqPos GetSeqLength(TNumrow row) const = 0;
    virtual TSeqPos GetBaseWidth(TNumrow row) const = 0;

    /// Anchoring methods
    virtual TNumrow  GetAnchor(void) const = 0;

    virtual const objects::CSeq_id&  GetSeqId(TNumrow row) const = 0;
    virtual const objects::CBioseq_Handle& GetBioseqHandle(TNumrow row) const = 0;
    virtual bool    IsPositiveStrand(TNumrow row) const = 0;
    virtual bool    IsNegativeStrand(TNumrow row) const = 0;

    virtual TSignedSeqPos  GetSeqPosFromAlnPos(TNumrow for_row,
        TSeqPos aln_pos,
        IAlnExplorer::ESearchDirection dir = IAlnExplorer::eNone,
        bool try_reverse_dir = true) const = 0;

    virtual TSignedSeqPos  GetAlnPosFromSeqPos(TNumrow row,
        TSeqPos seq_pos,
        IAlnExplorer::ESearchDirection dir = IAlnExplorer::eNone,
        bool try_reverse_dir = true) const = 0;

    /// Get sequence string for a given row in alignment coordinate.
    /// If the given row contains deletions, the deletions will be 
    /// filled with '-'. If it contains inserts, the inserted sequences
    /// will be removed from the returned string. The returns string
    /// always follows the alingment coordinate direction.
    /// Note: alignment coordinate direction may or may not follow 
    /// anchor sequence's direction.
    virtual string& GetAlnSeqString(string &buffer, TNumrow row,
        const TSignedRange& aln_range) const = 0;

    /// Get sequence string for a given row in sequence range.
    /// The method can optionally return the sequence string in 
    /// anchor sequence's direction (if anchor row and selected
    /// row have different strands.
    virtual string&  GetSeqString(string &buffer, TNumrow row,
        const TSignedRange& seq_rng,
        const TSignedRange& /*aln_rng*/,
        bool anchor_direct) const
    {
        objects::CSeqVector seq_vec =
            GetBioseqHandle(row).GetSeqVector(objects::CBioseq_Handle::eCoding_Iupac);
        seq_vec.GetSeqData(seq_rng.GetFrom(), seq_rng.GetToOpen(), buffer);
        TNumrow anchor = GetAnchor();
        if (anchor_direct  &&  row != anchor) {
            bool reversed = IsPositiveStrand(anchor) == IsPositiveStrand(row);
            if (reversed) {
                string tmp_seq;
                if (GetBaseWidth(row) == 3) {
                    // protein sequence, just reverse it
                    CSeqManip::Reverse(buffer, CSeqUtil::e_Iupacaa,
                        0, buffer.length(), tmp_seq);
                } else {
                    // nucleotide sequence, get reverse and complement
                    CSeqManip::ReverseComplement(buffer, CSeqUtil::e_Iupacna,
                        0, buffer.length(), tmp_seq);
                }
                buffer = tmp_seq;
            }
        }
        return buffer;
    }

    virtual IAlnSegmentIterator* CreateSegmentIterator(TNumrow row,
        const TSignedRange& range,
        IAlnSegmentIterator::EFlags flags) const = 0;

    //virtual void  GetAlnFromSeq(TNumrow row, const TRangeColl& seq_coll,
    //                            TRangeColl& aln_coll) const = 0;


    //virtual void  GetSeqFromAln(TNumrow row, const TRangeColl& aln_coll,
    //                            TRangeColl& seq_coll) const = 0;

    virtual TSignedSeqPos GetSeqPosFromSeqPos(TNumrow for_row,
        TNumrow row,
        TSeqPos seq_pos) const = 0;

    /// Check if there is any unaligned polyA tail.
    /// By default, it returns unknown to ask for further verification.
    virtual EPolyATail HasPolyATail() const = 0;

    /// Check if there is a CIGAR string.
    virtual bool HasCigar() const = 0;

    /// Get CIGAR vector.
    virtual const TCigar* GetCigar() const = 0;

    virtual bool IsSimplified() const 
    { return false; }

    virtual bool IsRegular() const 
    { return true; }
};

END_NCBI_SCOPE

#endif  // GUI_WIDGETS_SEQ_GRAPHIC___ALNGRAPHIC_DS__HPP
