#ifndef GUI_WIDGETS_SEQ_GRAPHIC___HISTOGRAM_GLYPH__HPP
#define GUI_WIDGETS_SEQ_GRAPHIC___HISTOGRAM_GLYPH__HPP

/*  $Id: histogram_glyph.hpp 27972 2013-05-03 20:34:20Z wuliangs $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Vlad Lebedev, Liangshou Wu
 *
 * File Description:
 *    CHistogramGlyph -- utility class for having
 *                      feature density histograms his in graphical layouts.
 *
 */


#include <corelib/ncbiobj.hpp>
#include <objects/seqloc/Seq_interval.hpp>
#include <gui/utils/rgba_color.hpp>
#include <gui/objutils/density_map.hpp>
#include <gui/widgets/seq_graphic/seq_glyph.hpp>
#include <gui/widgets/seq_graphic/histogram_conf.hpp>
#include <math.h>

BEGIN_NCBI_SCOPE


class NCBI_GUIWIDGETS_SEQGRAPHIC_EXPORT CHistogramGlyph : public CSeqGlyph
{
public:
    typedef float TDataType;
    typedef CDensityMap<TDataType> TMap;
    typedef map<string, TMap> TMaps;

    /// @name ctors
    /// @{
    CHistogramGlyph(const TMap& map, objects::CSeqFeatData::ESubtype subtype,
                    const string& title = "");
    CHistogramGlyph(const TMap& map, const string& title);
    CHistogramGlyph(const TMaps& maps, objects::CSeqFeatData::ESubtype subtype);
    CHistogramGlyph(const TMaps& maps, const string& name);
    /// @}

    /// @name CSeqGlyph virtual methods.
    /// @{
    virtual bool OnLeftDblClick(const TModelPoint& /*p*/);
    virtual bool NeedTooltip(const TModelPoint& p, string& tt, string& t_title) const;
    virtual void GetTooltip(const TModelPoint& /*p*/, string& tt, string& t_title) const;
    virtual void GetHTMLActiveAreas(TAreaVector* p_areas) const;
    virtual bool IsClickable() const;
    ///@}

    /// Access the first density map.
    const TMap& GetDensityMap(void) const;

    /// Access all the maps.
    const TMaps& GetDensityMaps(void) const;

    /// Access the number of maps.
    const int   GetMapNum() const;

    // access the position of this object.
    virtual TSeqRange GetRange(void) const;

    objects::CSeqFeatData::E_Choice GetFeatType() const;
    objects::CSeqFeatData::ESubtype GetFeatSubtype() const;

    const string& GetTitle() const;

    void SetAnnotName(const string& name);
    const string& GetAnnotName() const;

    void SetSubtype(objects::CSeqFeatData::ESubtype subtype);

    void SetDesc(const string& desc);
    const string& GetDesc() const;

    void SetConfigManager(CRef<CHistParamsManager> conf_mgr);

    //void SetConfig(CRef<CHistParams> conf);
    //CRef<CHistParams> GetConfig() const;

    /// Get the actually density max/min values.
    TDataType GetMax() const;
    TDataType GetMin() const;
    /// Get the axis limits (max/min).
    TDataType GetAxisMax() const;
    TDataType GetAxisMin() const;

    void SetAxisMax(TDataType max);
    void SetAxisMin(TDataType min);

    void SetShowTitle(bool f);
    void SetDialogHost(IGlyphDialogHost* host);

protected:
    virtual void x_Draw() const;
    virtual void x_UpdateBoundingBox();

    CRef<CHistParams> x_GetConfig() const;

protected:
    IGlyphDialogHost*               m_DlgHost;
    TMaps                           m_Maps;
    TDataType                       m_AxisMax;
    TDataType                       m_AxisMin;

    /// Key for accessing histogram settings.
    /// It is used only when m_Subtype is eSubtype_any.
    string                          m_AnnotName;
    string                          m_Desc;
    CRef<CHistParamsManager>        m_ConfigMgr;
    objects::CSeqFeatData::ESubtype m_Subtype;
    bool                            m_ShowTitle;
};

///////////////////////////////////////////////////////////////////////////////
/// CHistogramGlyph inline method implementation.

inline
const CHistogramGlyph::TMap& CHistogramGlyph::GetDensityMap(void) const
{
    TMaps::const_iterator iter = m_Maps.begin();
    return iter->second;
}

inline
const CHistogramGlyph::TMaps& CHistogramGlyph::GetDensityMaps(void) const
{
    return m_Maps;
}

inline
const int CHistogramGlyph::GetMapNum() const
{
    return (int)m_Maps.size();
}

inline
TSeqRange CHistogramGlyph::GetRange(void) const
{
    TSeqRange range;
    ITERATE (TMaps, iter, m_Maps) {
        range.CombineWith(iter->second.GetRange());
    }
    return range;
}

inline
objects::CSeqFeatData::ESubtype CHistogramGlyph::GetFeatSubtype() const
{ return m_Subtype; }

inline
const string& CHistogramGlyph::GetTitle() const
{ return m_Maps.begin()->first; }

inline
void CHistogramGlyph::SetAnnotName(const string& name)
{ m_AnnotName = name; }

inline
const string& CHistogramGlyph::GetAnnotName() const
{ return m_AnnotName; }

inline
void CHistogramGlyph::SetSubtype(objects::CSeqFeatData::ESubtype subtype)
{ m_Subtype = subtype; }

inline
void CHistogramGlyph::SetDesc(const string& desc)
{ m_Desc = desc; }

inline
const string& CHistogramGlyph::GetDesc() const
{ return m_Desc; }

inline
void CHistogramGlyph::SetConfigManager(CRef<CHistParamsManager> conf_mgr)
{ m_ConfigMgr = conf_mgr; }

inline
CHistogramGlyph::TDataType CHistogramGlyph::GetMax() const
{
    TModelUnit max_val = DBL_MIN;
    ITERATE (TMaps, map_iter, m_Maps) {
        max_val = max(max_val, (TModelUnit)map_iter->second.GetMax());
    }
    return max_val;
}

inline
CHistogramGlyph::TDataType CHistogramGlyph::GetMin() const
{
    TModelUnit min_val = DBL_MAX;
    ITERATE (TMaps, map_iter, m_Maps) {
        min_val = min(min_val, (TModelUnit)map_iter->second.GetMin());
    }
    return min_val;
}

inline
CHistogramGlyph::TDataType CHistogramGlyph::GetAxisMax() const
{ return m_AxisMax; }

inline
CHistogramGlyph::TDataType CHistogramGlyph::GetAxisMin() const
{ return m_AxisMin; }

inline
void CHistogramGlyph::SetAxisMax(TDataType max_a)
{ m_AxisMax = ceil(max(max_a, m_AxisMax)); }

inline
void CHistogramGlyph::SetAxisMin(TDataType min_a)
{ m_AxisMin = floor(min(min_a, m_AxisMin)); }

inline
void CHistogramGlyph::SetShowTitle(bool f)
{ m_ShowTitle = f; }

inline
void CHistogramGlyph::SetDialogHost(IGlyphDialogHost* host)
{ m_DlgHost = host; }

inline
CRef<CHistParams> CHistogramGlyph::x_GetConfig() const
{
    if (m_Subtype != objects::CSeqFeatData::eSubtype_any) {
        return m_ConfigMgr->GetHistParams(m_Subtype);
    }

    string name = GetAnnotName();
    NStr::ReplaceInPlace(name, ".", "_");
    return m_ConfigMgr->GetHistParams(name);
}


END_NCBI_SCOPE

/* @} */

#endif  // GUI_WIDGETS_SEQ_GRAPHIC___HISTOGRAM_GLYPH__HPP
