#ifndef GUI_WIDGETS_TEXT_WIDGET___TEXT_ITEM_PANEL__HPP
#define GUI_WIDGETS_TEXT_WIDGET___TEXT_ITEM_PANEL__HPP

/*  $Id: text_item_panel.hpp 27922 2013-04-25 16:51:42Z katargir $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Roman Katargin
 *
 * File Description:
 *
 */

#include <corelib/ncbiobj.hpp>
#include <gui/gui_export.h>

#include <vector>
#include <set>

#include <wx/scrolwin.h>
#include <wx/timer.h>

#include <gui/objutils/objects.hpp>


#include <gui/utils/event_handler.hpp>

#include <gui/widgets/text_widget/text_item.hpp>
#include <gui/widgets/text_widget/composite_text_item.hpp>
#include <gui/widgets/text_widget/calc_size_cache.hpp>


BEGIN_NCBI_SCOPE

class CTextItemPanel;

class CTextPanelState : public CObject
{
public:
    virtual ~CTextPanelState() {}

    virtual void SaveState(CTextItemPanel& textPanel) = 0;
    virtual void RestoreState(CTextItemPanel& textPanel, const vector<ITextItem*>& items) = 0;
    virtual void RestorePos(CTextItemPanel& textPanel) = 0;
};

class NCBI_GUIWIDGETS_TEXT_WIDGET_EXPORT CTextPanelContext
    : public wxEvtHandler, public CEventHandler
{
friend class CTextItemPanel;
friend class CItemStackGuard;

public:
    CTextPanelContext(int leftMatgin) : m_ItemCollapsed(false), m_Panel(0), m_MouseOverItem(0),
                          m_LeftMargin(leftMatgin), 
                          m_CharWidth(8), m_CharHeight(12), m_SpaceWidth(8),
                          m_HighlightedItem(0) {}

    CTextItemPanel* GetPanel() { return m_Panel; }
    int GetCharWidth() const { return m_CharWidth; }
    int GetIndentWidth() const { return m_CharWidth; }
    int GetCharHeight() const { return m_CharHeight; }
    int GetSpaceWidth() const { return m_SpaceWidth; }

    // in pixels
    int GetLeftMargin() const { return m_LeftMargin*m_CharWidth; }

    void SelectItem(ITextItem* item, bool controlDown, bool shiftDown);
    virtual void EditItem(ITextItem* item, bool controlDown, bool shiftDown);
    bool IsItemSelected(ITextItem* item);

    void GetSelectedObjects (vector< CConstRef<CObject> >& objs);
    void SetSelectedObjects (vector< CConstRef<CObject> >& objs);

    void GetSelectedObjects (TConstScopedObjects& objects);

// Tooltip functions    
    void    ClearMouseMove();
    void    AddHier(ITextItem* item);
    void    MouseOverItem(ITextItem* item);
    void    TriggerTooltip();

    virtual void MouseMoveInContainer(ITextItem* /*container*/) {}
    virtual wxWindow* CreateTooltipWindow() { return NULL; }

    bool    m_ItemCollapsed;

    ITextItem* GetHighlightedItem() { return m_HighlightedItem; }
    void SetHighlightedItem(ITextItem* pItem) { m_HighlightedItem = pItem; }

    CCalcSizeCache& GetTextSizeCache() { return m_TextSizeCache; }
    void ResetTextSizeCache() { m_TextSizeCache.Reset(); }

    virtual wxMenu* CreateMenu() const { return 0; }

    virtual CTextPanelState* GetViewState() const { return NULL; }

    virtual void SetPanel(CTextItemPanel* panel);

protected:
    CTextItemPanel* m_Panel;

// Tooltip data
    vector<ITextItem*> m_Hier;
    ITextItem*         m_MouseOverItem;
    set<ITextItem*>    m_SelectedItems;

private:
    int m_LeftMargin; // number of chars
    int m_CharWidth;
    int m_CharHeight;
    int m_SpaceWidth;
    ITextItem* m_HighlightedItem;

    CCalcSizeCache m_TextSizeCache;
};

class NCBI_GUIWIDGETS_TEXT_WIDGET_EXPORT CTextItemPanel : public wxScrolledWindow, public CEventHandler
{
public:
    CTextItemPanel(wxWindow* parent, wxWindowID id = wxID_ANY,
                      const wxPoint& pos = wxDefaultPosition,
                      const wxSize& size = wxDefaultSize,
                      long style = wxBORDER_NONE | wxHSCROLL | wxVSCROLL);
    ~CTextItemPanel();

    void Init();

    virtual bool Layout();
    virtual bool SetFont(const wxFont& font);

    void SetData(ITextItem* item, CTextPanelContext* context);
    CTextPanelContext* GetContext() { return m_Context.get(); }
    void RecalSize();

    void AddItems(const vector<ITextItem*>& items);
    void InsertItems(const CCompositeTextItem::TInsertBlockVec& insertBlocks);
    void RemoveItem (size_t index);

    void ScrollToBottom();

    ITextItem* GetItem() { return m_Item.get(); }
    void InitDC(wxDC& dc);

    bool FindText(const wxString& text, bool forward = true);
    void MakePosVisible(const CTextPosition& pos);

//
// Text selection functions
//
    bool IsTextSelected() const;
    wxString GetSelectedText() const;

    CTextBlock GetTextSelection() const { return m_TextSelection; }
    bool Selecting() const { return m_Selecting; }

    void SelStart(int row, int col, bool shift);
    void SelMove(int row, int col);
    void SelEnd(int row, int col);

    void SetSelection(const CTextBlock& selection);

    void SetCaretPos(int x, int y);
    void UpdateCaretPos();

    void TriggerTooltip();
    wxPoint GetMousePos() const { return m_MousePos; }

    virtual bool ProcessEvent(wxEvent& event);
    void BlockEvents(bool blockEvents) { m_BlockEvents = blockEvents; }

protected:
    void OnPaint(wxPaintEvent& event);
    void OnEraseBackground(wxEraseEvent& event);
    void OnMouseEvent(wxMouseEvent& event);
    void OnContextMenu(wxContextMenuEvent& event);
    void OnSetFocus(wxFocusEvent& event);
    void OnKillFocus(wxFocusEvent& event);
    void OnTimer(wxTimerEvent& event);

// Text selection
    void OnCopy(wxCommandEvent& event);
    void OnUpdateCopy(wxUpdateUIEvent& event);
    void OnSelectAll(wxCommandEvent& event);

    void x_TextPosChanged();

    auto_ptr<CTextPanelContext> m_Context;
    auto_ptr<ITextItem> m_Item;

    bool m_Selecting;
    CTextBlock m_TextSelection;

    wxTimer     m_Timer;
    wxWindow*   m_TooltipWindow;
    wxPoint     m_MousePos;

    bool m_BlockEvents;

    DECLARE_EVENT_TABLE()
};

END_NCBI_SCOPE

#endif  // GUI_WIDGETS_TEXT_WIDGET___TEXT_ITEM_PANEL__HPP
