#ifndef GUI_WIDGETS_WX___JOB_SYNC_CALL__HPP
#define GUI_WIDGETS_WX___JOB_SYNC_CALL__HPP

/*  $Id: async_call.hpp 26758 2012-11-02 16:27:30Z katargir $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Roman Katargin
 *
 * File Description:
 *
 */

#include <gui/gui.hpp>

#include <gui/utils/canceled_impl.hpp>
#include <gui/utils/execute_unit.hpp>

#include <wx/string.h>


BEGIN_NCBI_SCOPE

class IAppJob;

/// Class for GUI asyncronous execition. 
/// 
/// If async background call takes too much time GBench GUI automatically shows 
/// a window to inform the user to wait
///
class NCBI_GUIWIDGETS_WX_EXPORT CAsyncCall
{
public:
    class Call : public CCanceledImpl
    {
    public:
        Call(const wxString& descr) : m_Descr(descr) {}
        Call() : m_Descr(wxT("Accessing network...")) {}
        virtual ~Call() {}

        wxString GetDescr() const { return m_Descr; }

        virtual bool CanCancel() const { return false; }
        virtual void Execute() = 0;

    private:
        wxString m_Descr;
    };
  
    CAsyncCall(Call& call);
    ~CAsyncCall();

    void Execute();
    static void Execute(Call& call);

    // This version calls Yield() during processing to allow other events to be
    // processed. This user of this function should disable any interface
    // elements whose interaction might interfere with the running job.
    void ExecuteYield();

private:
    Call& m_Call;
    wxString m_Descr;
    int m_JobId;
};

/// Helper template function for easy parameter bindings to asyncronous calls
/// 
/// In the example below you can define an exec functor and easily run it
/// in background with input parameter passing and return result returned
/// across threads. Please note that exceptions should also be propagated.
/// 
/// 
/// Example:
///    struct CLocalAsyncFunc 
///    {
///        int operator()(int a) { return a + 1; }
///    };
///    int v = AsyncExec1<int, int, CLocalAsyncFunc>(CLocalAsyncFunc(), 100);
///
template<class TRetParam, class TInParam, typename TFunc>
TRetParam GUI_AsyncExec1(TFunc& func, TInParam input)
{
    struct CLocalAsyncCallBind1 : public CAsyncCall::Call
    {
        CLocalAsyncCallBind1(TFunc& func, TInParam input) : m_Func(func), m_Input(input) {} 
        virtual void Execute() { m_Ret = m_Func(m_Input); }
        TRetParam GetResult() { return m_Ret; }
    private:
        TFunc&    m_Func;
        TInParam  m_Input;
        TRetParam m_Ret;
    };

    CLocalAsyncCallBind1 f(func, input);
    CAsyncCall asyncCall(f);
    asyncCall.Execute();
    return f.GetResult();
}

class CExecuteUnitAsyncCall : public CAsyncCall::Call
{
public:
    CExecuteUnitAsyncCall(IExecuteUnit& executeUnit, const wxString& descr)
        : Call(descr), m_ExecuteUnit(executeUnit), m_Result(false) {}
    virtual void Execute() { m_Result = m_ExecuteUnit.Execute(*this); }
    bool GetResult() { return m_Result; }

    virtual bool CanCancel() const { return true; }

private:
    IExecuteUnit& m_ExecuteUnit;
    bool m_Result;
};

END_NCBI_SCOPE

#endif  // GUI_WIDGETS_WX___JOB_SYNC_CALL__HPP
