/*  $Id: gbench_app.cpp 28087 2013-05-07 19:12:03Z katargir $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Andrey Yazhuk, Mike DiCuccio
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>

#include <corelib/ncbiexec.hpp>
#include <corelib/ncbi_system.hpp>
#include <corelib/ncbi_process.hpp>
#include <corelib/ncbifile.hpp>
#include <corelib/ncbicfg.h>
#include <corelib/request_ctx.hpp>

#include "gbench_app.hpp"
#include "gbench_version.hpp"
#include "init.hpp"
#include "named_pipe_process_task.hpp"
#include "net_conn_problem_dlg.hpp"
#include "testing_ncbi_conn_dlg.hpp"
#include "new_version_dlg.hpp"

#include <gui/core/named_pipe.hpp>
#include <gui/core/project_service.hpp>
#include <gui/core/ui_data_source_service.hpp>
#include <gui/core/commands.hpp>

#include <gui/framework/pkg_manager.hpp>

#include <gui/framework/app_task_service.hpp>

#include <gui/widgets/wx/wx_app.hpp>
#include <gui/widgets/wx/log_gbench.hpp>
#include <gui/widgets/wx/wx_utils.hpp>
#include <gui/objutils/init.hpp>
#include <gui/core/init.hpp>
#include <gui/widgets/object_list/init.hpp>
#include <gui/widgets/aln_score/init_w_aln_score.hpp>

#include <gui/widgets/wx/ui_tool_registry.hpp>
#include <gui/objutils/seqfetch_queue.hpp>
#include <gui/objutils/label.hpp>

#include <gui/utils/app_job_dispatcher.hpp>
#include <gui/utils/thread_pool_engine.hpp>
#include <gui/utils/scheduler_engine.hpp>
#include <gui/objutils/object_manager_engine.hpp>
#include <gui/widgets/wx/sys_path.hpp>
#include <gui/utils/system_path.hpp>
#include <gui/utils/conn_test_thread.hpp>
#include <connect/ncbi_conn_test.hpp>

#include <gui/objects/gbench_svc_cli.hpp>
#include <gui/objects/GBenchInitRequest.hpp>
#include <gui/objects/GBenchInitReply.hpp>
#include <gui/objects/GBenchVersionRequest.hpp>
#include <gui/objects/GBenchVersionReply.hpp>

#include <objects/seqloc/Seq_id.hpp>
#include <objects/seqalign/Dense_seg.hpp>
#include <objects/seqalign/Score.hpp>

#include <util/line_reader.hpp>

#include <gui/widgets/wx/message_box.hpp>
#include <gui/utils/event_translator.hpp>
#include <gui/framework/app_task_service.hpp>
#include <gui/utils/app_job_impl.hpp>

#include <wx/apptrait.h>
#include <wx/menu.h>
#include <wx/filename.h>
#include <wx/filefn.h>
#include <wx/utils.h>
#include <wx/msgdlg.h>

#include <dbapi/driver/dbapi_svc_mapper.hpp>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

// Declare wxWidgets application class
// this will allow CNCBIwxApplication to instantiate our class
DECLARE_APP(ncbi::CGBenchApp)
IMPLEMENT_APP_NO_MAIN(ncbi::CGBenchApp)
IMPLEMENT_WX_THEME_SUPPORT

static const char* kDownloadUrl = "http://www.ncbi.nlm.nih.gov/tools/gbench/";
bool CGBenchApplication::m_MacStartInEditMode = false;

class CAppStartResult : public CObject
{
public:
	CAppStartResult() 
		: m_NewVersionAvailable( false )
	{}

	CAppStartResult( CGBenchVersionInfo& new_version, string url ) 
		: m_NewVersionAvailable( true )
		, m_NewVersion( &new_version )
		, m_NewVersionUrl( url )
	{}

    bool IsNewVersion() const { return m_NewVersionAvailable; }
    CRef<CGBenchVersionInfo> GetNewVersion() const { return m_NewVersion; }
	string GetNewVersionURL() const { return m_NewVersionUrl; }

public:
    bool m_NewVersionAvailable;
	CRef<CGBenchVersionInfo> m_NewVersion;
	string m_NewVersionUrl;
};

class CAppStartRequestJob : public CAppJob
{
public:
	CAppStartRequestJob()
		: CAppJob( "Checking for updates" )
	{}

	virtual CRef<CObject> GetResult();

	virtual EJobState Run();

private:
	CRef<CAppStartResult> m_Result;
};

CRef<CObject> CAppStartRequestJob::GetResult()
{
	CFastMutexGuard guard( m_Mutex );
	return CRef<CObject>( m_Result.GetPointer() );
}

IAppJob::EJobState CAppStartRequestJob::Run()
{
	CGBenchVersionInfo cur_version;
	GetGBenchVersionInfo( cur_version );

	CRef<CUser_object> reply = CGBenchService().SendAppStart( cur_version );

	if( reply->HasField( "current-version" ) ){
		string version_from_reply = reply->GetField( "current-version" ).GetData().GetStr();

		LOG_POST(Info 
			<< "The latest GBench version is " 
			<< version_from_reply
		);

		vector<string> tokens;
		NStr::Tokenize( version_from_reply, string("."), tokens );
		if( tokens.size() == 3 ){
			size_t ver_major = NStr::StringToNumeric( tokens[0] );
			size_t ver_minor = NStr::StringToNumeric( tokens[1] );
			size_t ver_patch = NStr::StringToNumeric( tokens[2] );

			CRef<CGBenchVersionInfo> new_version( new CGBenchVersionInfo() );
			new_version->SetVer_major( ver_major );
			new_version->SetVer_minor( ver_minor );
			new_version->SetVer_patch( ver_patch );

			if( cur_version < *new_version ){
				LOG_POST(Warning << "New GBench version " << version_from_reply << " is available" );

				string url_from_reply =
					reply->HasField( "current-url" )
					? string(reply->GetField( "current-url" ).GetData().GetStr())
					: string( kDownloadUrl )
				;

				CRef<CAppStartResult> result( new CAppStartResult( *new_version, url_from_reply ) );

				CFastMutexGuard guard( m_Mutex );
				m_Result = result;
			}
		}
	}

	return eCompleted;
}


class CVersionCheckResult : public CObject
{
public:
    CVersionCheckResult(bool newVersionAvail, CRef<CGBenchInitReply>& reply)
        : m_NewVersionAvailable(newVersionAvail), m_Reply(reply) 
    {}
    bool IsNewVersion() const { return m_NewVersionAvailable; }
    CRef<CGBenchInitReply> GetReply() const { return m_Reply; }

private:
    bool   m_NewVersionAvailable;
    CRef<CGBenchInitReply> m_Reply;
};


class CVersionCheckJob : public CJobCancelable
{
public:
    CVersionCheckJob() : m_NewVersionAvailable(false){}

    /// @name IAppJob implementation
    /// @{
    virtual EJobState                   Run();
    virtual CConstIRef<IAppJobProgress> GetProgress();
    virtual CRef<CObject>               GetResult();
    virtual CConstIRef<IAppJobError>    GetError();
    virtual string GetDescr() const;
    /// @}

protected:
    void x_SetResult(bool goodConnection, const string& descr);

    // for job results/status
    bool       m_NewVersionAvailable;
    mutable CFastMutex m_Mutex;
    CRef<CGBenchInitReply> m_Reply;
};


IAppJob::EJobState CVersionCheckJob::Run()
{
    CFastMutexGuard guard(m_Mutex);

    CGBenchVersionInfo version;
    GetGBenchVersionInfo(version);

    CGBenchService svc;

	m_Reply = svc.SendInit(version);

	const CGBenchVersionInfo& current_version = m_Reply->GetCurrent_version();
	if (version < current_version) {
		LOG_POST(Warning << "New GBench version is available");
		m_NewVersionAvailable = true;
	}

	return eCompleted;
}


CConstIRef<IAppJobProgress> CVersionCheckJob::GetProgress()
{
    return CConstIRef<IAppJobProgress>(new CAppJobProgress(1.0, "Finished"));
}

CRef<CObject> CVersionCheckJob::GetResult()
{
    CFastMutexGuard guard(m_Mutex);
    CRef<CObject> result(new CVersionCheckResult(m_NewVersionAvailable, m_Reply));
    return result;
}

CConstIRef<IAppJobError> CVersionCheckJob::GetError()
{
    return CConstIRef<IAppJobError>(new CAppJobError("No errors"));
}

string CVersionCheckJob::GetDescr() const
{
    return "Checking new version";
}

///////////////////////////////////////////////////////////////////////////////
/// CVersionCheckTask - a task to check if new version is available

class CVersionCheckTask : public CAppTask, public CEventHandler
{
    DECLARE_EVENT_MAP();
public:
    CVersionCheckTask();
    ~CVersionCheckTask();

    /// @name IAppTask interface implementation
    /// @{
    virtual ETaskState    Run();
    virtual string  GetStatusText() const;
    /// @}

    /// functions below handle notifications from the IAppJob
    void    OnAppJobNotification(CEvent* event);
    void    OnAppJobProgress(CEvent* event);

protected:
    ETaskState x_OnWakeUp();
    void       x_Finish(IAppJob::EJobState state);

    typedef CAppJobDispatcher::TJobID   TJobID;

    CIRef<IAppJob>     m_Job;
    TJobID             m_JobID;
    IAppJob::EJobState m_JobState;

    float  m_Progress;
    string m_StatusText;
    bool   m_NewVersionAvailable;
	CRef<CObject> m_Result; 
    CRef<CGBenchInitReply> m_Reply;
};

BEGIN_EVENT_MAP(CVersionCheckTask, CEventHandler)
    ON_MESSAGE(CAppJobNotification, CAppJobNotification::eStateChanged,
               &CVersionCheckTask::OnAppJobNotification)
    ON_MESSAGE(CAppJobNotification, CAppJobNotification::eProgress,
               &CVersionCheckTask::OnAppJobProgress)
END_EVENT_MAP()

CVersionCheckTask::CVersionCheckTask()
: CAppTask("Checking for new version"),
  m_Job(),
  m_JobID(CAppJobDispatcher::eInvalidJobID),
  m_JobState(IAppJob::eInvalid),
  m_Progress(0),
  m_NewVersionAvailable(false)
{
}

CVersionCheckTask::~CVersionCheckTask()
{
    if (m_JobID != CAppJobDispatcher::eInvalidJobID) {
        CAppJobDispatcher& disp = CAppJobDispatcher::GetInstance();
        disp.DeleteJob(m_JobID);
    }
}

static const bool kGenericRequest = true;

IAppTask::ETaskState CVersionCheckTask::Run()
{
    switch (m_State) {
    case eInitial:
        {{
            _ASSERT(!m_Job);
            _ASSERT(m_JobID == CAppJobDispatcher::eInvalidJobID);
            _ASSERT(m_JobState == IAppJob::eInvalid);

			if( kGenericRequest ){
				m_Job.Reset( new CAppStartRequestJob() );
			} else {
				m_Job.Reset( new CVersionCheckJob() );
			}

            m_State = eRunning;
            m_StatusText = "Starting...";

            CAppJobDispatcher& disp = CAppJobDispatcher::GetInstance();
            m_JobID = disp.StartJob(*m_Job, "ThreadPool", *this, -1, true);

            _ASSERT(m_JobID != CAppJobDispatcher::eInvalidJobID);
            m_JobState = IAppJob::eRunning;
            m_State = eBackgrounded;
            m_StatusText = "Working...";
        }}
        break;

    case eBackgrounded:
        // Job has been launched already - check whether its status has changed
        m_State = x_OnWakeUp();
        break;

    default:
        _ASSERT(false); // unexpected
        m_State = eInvalid;
        break;
    }

    if (m_State == eCompleted) {

        if( m_NewVersionAvailable ){

			string strCurVersion, strLatestVersion;
			string download_url("http://www.ncbi.lm.nih.gov/projects/gbench/");

			CGBenchVersionInfo version;
			GetGBenchVersionInfo( version );
			version.GetVersionString( &strCurVersion );

			if( kGenericRequest ){

				CAppStartResult* start_result =
					dynamic_cast<CAppStartResult*>( m_Result.GetPointer() )
				;

				start_result->GetNewVersion()->GetVersionString( &strLatestVersion );
				download_url = start_result->GetNewVersionURL();

			} else {

				try {
					CGBenchService svc;
					CGBenchVersionRequest req;
					req.SetVersion().Assign(m_Reply->GetCurrent_version());
					req.SetVersion_request(CGBenchVersionRequest::eVersion_request_download_url);

					CRef<CGBenchVersionReply> reply = svc.AskVersion_update(req);
					download_url = reply->GetDownload_path();

				} catch( CException& e ){
					LOG_POST(Error << "failed to retrieve download URL: " << e.GetMsg());
				}

				m_Reply->GetCurrent_version().GetVersionString( &strLatestVersion );
			}

            // now, notify the user

            CNewVersionDlg dlg;
            dlg.SetCurrentVersion(ToWxString(strCurVersion));
            dlg.SetLatestVersion(ToWxString(strLatestVersion));
            dlg.SetDownloadURL(ToWxString(download_url));
            dlg.Create(NULL);
            dlg.ShowModal();
            /*
            string msg;
            msg += "A new version of Genome Workbench is available.\n";

            msg += "\n";
            msg += "You are running version ";
            version.AddVersionString( &msg );
            msg += "\nThe latest version is ";
            m_Reply->GetCurrent_version().AddVersionString( &msg );
            msg += "\n\nIt can be downloaded directly from:\n";
            msg += download_url;

            wxMessageBox(ToWxString(msg), wxT("New version notification"),
                         wxOK | wxICON_WARNING);
            */
        }
    }

    return m_State;
}

string CVersionCheckTask::GetStatusText() const
{
    return m_StatusText;
}

IAppTask::ETaskState CVersionCheckTask::x_OnWakeUp()
{
    switch(m_JobState)  {
    case IAppJob::eRunning:
        return eBackgrounded;

    case IAppJob::eCompleted:
        return eCompleted;

    case IAppJob::eFailed:
        return eFailed;

    case IAppJob::eCanceled:
        return eCanceled;

    default:
        return eInvalid;
    }
}

void CVersionCheckTask::x_Finish(IAppJob::EJobState state)
{
    m_JobState = state;
    m_JobID = CAppJobDispatcher::eInvalidJobID;

    _ASSERT(m_Listener);
    if (m_Listener)  {
        CAppTaskServiceEvent::EType type = CAppTaskServiceEvent::eWakeUpSignal;
        if (m_JobState == IAppJob::eCanceled) {
            type = CAppTaskServiceEvent::eTaskCanceled;
        }

        CRef<CEvent> evt(new CAppTaskServiceEvent(*this, type));
        m_Listener->Post(evt);
    }
}

/// handles "state changed" notification from CAppJobDispatcher
void CVersionCheckTask::OnAppJobNotification(CEvent* evt)
{
    CAppJobNotification* notn = dynamic_cast<CAppJobNotification*>(evt);
    _ASSERT(notn);

    if(notn)    {
        int job_id = notn->GetJobID();
        if(m_JobID != job_id) {
            _ASSERT(false);
            ERR_POST("CVersionCheckTask::OnAppJobNotification() - unknown or invalid Job ID " << job_id);
        } else {
            switch(notn->GetState())    {
            case IAppJob::eCompleted:   {
                m_StatusText = "Finished";

				CRef<CObject> result = m_Job->GetResult();

				if( kGenericRequest ){
					if( !result.IsNull() ){
						CAppStartResult* start_result =
							dynamic_cast<CAppStartResult*>( result.GetPointer() )
						;
						m_NewVersionAvailable = start_result->IsNewVersion();
						m_Result = result;
					}
				} else {
					CVersionCheckResult* check_res =
						dynamic_cast<CVersionCheckResult*>(result.GetPointer())
					;
					if( check_res ){
						m_NewVersionAvailable = check_res->IsNewVersion();
						m_Reply = check_res->GetReply();
					}
				}

                x_Finish(IAppJob::eCompleted);
                break;
            }

            case IAppJob::eFailed:  {
                x_Finish(IAppJob::eFailed);
                break;
            }
            case IAppJob::eCanceled:    {
                m_StatusText = "";
                x_Finish(IAppJob::eCanceled);
                break;
            }
            default:
                break;
            }
        }
    }
}

void CVersionCheckTask::OnAppJobProgress(CEvent* evt)
{
    CAppJobNotification* notn =
        dynamic_cast<CAppJobNotification*>(evt);
    _ASSERT(notn);

    if(notn)    {
        int job_id = notn->GetJobID();
        if(m_JobID != job_id) {
            _ASSERT(false);
            ERR_POST("CVersionCheckTask::OnAppJobNotification() - unknown or invalid Job ID " << job_id);
        } else {
            CConstIRef<IAppJobProgress> prg = notn->GetProgress();
            const CAppJobProgress* progress =
                dynamic_cast<const CAppJobProgress*>(prg.GetPointer());
            if(progress)  {
                m_Progress = progress->GetNormDone();
                m_StatusText = progress->GetText();

                if (m_Listener) {
                    CRef<CEvent> evt(new CAppTaskServiceEvent(*this, CAppTaskServiceEvent::eStatusUpdate));
                    m_Listener->Post(evt);
                }
            }
        }
    }
}



///////////////////////////////////////////////////////////////////////////////
/// CGBenchApplication

CGBenchApplication::CGBenchApplication()
:   CNCBIwxApplication("NCBI Genome Workbench")
{
    SetDiagFilter(eDiagFilter_Trace, "!/corelib !/connect !/objmgr");
    SetDiagPostFlag(eDPF_DateTime);

    UnsetDiagTraceFlag(eDPF_LongFilename);

    m_DefaultPkgs.push_back("sequence");
    m_DefaultPkgs.push_back("alignment");
    m_DefaultPkgs.push_back("ncbi internal");
    m_DefaultPkgs.push_back("snp");
}

void CGBenchApplication::Init()
{
#   ifdef NCBI_OS_DARWIN
    // First, add run path for NCBI application on Mac

    static string s_FrameworkPath( "$ORIGIN/../Frameworks" );

    if( NCBI_GetRunpath() && *NCBI_GetRunpath() ){
        LOG_POST( Info << "Run path is " << NCBI_GetRunpath() );

        s_FrameworkPath += ":";
        s_FrameworkPath += NCBI_GetRunpath();
    }

    NCBI_SetRunpath( s_FrameworkPath.c_str() );

#   endif

	// Set up permanent GBench Session ID before any possible connection to NCBI

	string session_id;

	CNcbiRegistry& reg = GetConfig();
	if( reg.HasEntry( "GBENCH_CONN", "NCBI_SID" ) ){
		session_id = reg.Get( "GBENCH_CONN", "NCBI_SID" );
		GetDiagContext().GetRequestContext().SetSessionID( session_id );
	
	} else {
		session_id = GetDiagContext().GetRequestContext().SetSessionID();
		reg.Set( "GBENCH_CONN", "NCBI_SID", session_id, IRegistry::fPersistent );
		SaveConfig();
	}
	GetDiagContext().SetDefaultSessionID( session_id );

    CNCBIwxApplication::Init();

    // We need here our GUI registry load
    // for our services initialization

    CGBenchApp* gbench_wx_app = dynamic_cast<CGBenchApp*>(wxTheApp);
    if( gbench_wx_app ){
        /// insure that the user home path exists 
        wxString path = CSysPath::ResolvePath(wxT("<home>"));
        if( !wxFileName::DirExists( path ) ){
            ::wxMkdir( path );
        }

        // create diagnostic handler
        CwxLogDiagHandler::GetInstance();

        x_ReportVersion();

        gbench_wx_app->x_LoadGuiRegistry();
    } else {
        // something is wrong with our application 
        _ASSERT(false);
    }
}


// setup command line argument descriptions
void CGBenchApplication::x_SetupArgDescriptions()
{
    auto_ptr<CArgDescriptions> arg_desc(new CArgDescriptions);

    string basename = GetArguments().GetProgramBasename();
    arg_desc->SetUsageContext(basename, GetProgramDisplayName());

    arg_desc->AddFlag("v", "Version");
    arg_desc->AddFlag("edit", "Submission perspective");
    arg_desc->AddDefaultKey("t", "File_Type", "Type of optional data file",
                            CArgDescriptions::eString, "auto");
    arg_desc->SetConstraint("t", &(*new CArgAllow_Strings,
                        "auto", "project", "workspace", "message", "asntext",
                        "asnbin", "xml", "fasta", "textalign", "newick"));
    arg_desc->AddExtra(0, 100, "Items to open", CArgDescriptions::eString);

    SetupArgDescriptions(arg_desc.release());
}


int CGBenchApplication::x_Run()
{
    wxGetApp().SetGBenchGUIApp(this);

    CArgs args = GetArgs();

    if (args["v"]) {    // version checking
        // version is always reported in constructor.
        return 0;
    }

    x_StartCrashMonitor();

    /// start up our monitor process
    x_StartMonitor();

    int retcode = -1;



    /// if handed a file on command line, try sending it to another
    /// application instance via the named pipe if the command line
    /// switches and registry say to do so
    if (x_HandleRemote(args)) {
        retcode = 0;
    } else {

        init_gui_app_gbench();
        CGBenchPipe::Start();
        x_HandleArgs(args);

        init_gui_objutils();
        init_gui_core();
        init_w_object_list();
        init_w_aln_score();

        // Init cache for GI labels
        CLabelDescriptionCache::InitCache();


        x_TestConfig();

        // Enable Database load balancer handlers (for use in RADAR and other in-house plugins)
        DBLB_INSTALL_DEFAULT();

        // load new-style plugins
        x_InitPackages();

        x_ConfigUIToolRegistry(); // TODO

        // set global hooks  to save memory for DenSeg alignments
        //
        //CDense_seg::SetGlobalReserveHooks();

		// by Eugene Vasilchenko
		// This one is not safe because it uses shared Object-id to represent equal 
		// score.id objects in different Seq-align.
		// The benefit is the smaller memory used, but because the Object-id is shared 
		// changing score.id field in one Seq-align will affect others. You can use the 
		// hook if you are sure to not to change the score.id, or at least do it knowingly.
        CReadSharedScoreIdHook::SetGlobalHook();


        retcode = CNCBIwxApplication::x_Run();

        // we intentionally do this twice - once while the GUI is still
        // visible to allow the user to see that the app is still running
        LOG_POST(Info << "Clearing sequence prefetch queue");
        CSeqFetchQueue::GetInstance().Clear();

        // TODO - we may want to make a service out of CUIToolRegistry
        CUIToolRegistry::ShutDown();
    }

    SaveConfig();

    /// shut down the monitoring app
    /// this is explicitly *NOT GUARDED* - an exception thrown by the app will
    /// cause an unnatural shut-down. we do not do this if there was a fatal
    /// error in the message loop that was caught as an exception
    if (retcode == 0) {
        x_StopMonitor();
    }

    x_StopCrashMonitor();

    CLabelDescriptionCache::StopCache();

    return retcode;
}


void CGBenchApplication::x_ReportVersion()
{
    size_t ver_major = 0;
    size_t ver_minor = 0;
    size_t ver_patch_level = 0;
    CTime build_date;
    GetGBenchVersionInfo( ver_major, ver_minor, ver_patch_level, build_date );

    size_t ver_bits = sizeof(void*)*8;

#ifdef _DEBUG
    static const char* reldeb = "  built in debug mode on ";
#else
    static const char* reldeb = "  released on "; 
#endif        

    LOG_POST(Info << "NCBI Genome Workbench " << ver_bits << "bit v"
        << ver_major << "." << ver_minor << "." << ver_patch_level 
        << reldeb << build_date.AsString()
    );
}


// TODO - port
void CGBenchApplication::x_ReportAppInfo()
{
  /*  size_t ver_major = 0;
    size_t ver_minor = 0;
    size_t ver_patch_level = 0;
    CTime build_date;
    GetGBenchVersionInfo(ver_major, ver_minor, ver_patch_level, build_date);

    LOG_POST(Info << "NCBI Genome Workbench, version "
                << ver_major << "."
                << ver_minor << "."
                << ver_patch_level);
    LOG_POST(Info << "  build date: " << build_date.AsString());
    LOG_POST(Info << "  Genome Workbench installation path:   "
                << CSystemPath::ResolvePath("<std>", ""));
    LOG_POST(Info << "  Genome Workbench user path:   "
                << CSystemPath::ResolvePath("<home>", ""));
    LOG_POST(Info << "  Process ID: " << CProcess::GetCurrentPid());*/
}


// TODO
void CGBenchApplication::x_TestConfig()
{
    // TODO
    // CGuiRegistry& registry = CGuiRegistry::GetInstance();
    // CGuiRegistry::TReadView app_reg = registry.GetReadView("GBENCH.Application");

    if( HasLoadedConfig() ){
        LOG_POST( Info << "Config loaded from " << GetConfigPath() );
        // dump some diagnostic information about the contents of the registry

        list<string> sections;
        CNcbiRegistry& reg = GetConfig();
        reg.EnumerateSections(&sections);

        if( sections.size() ){
            LOG_POST( Info << "Config file contains:" );
            ITERATE( list<string>, iter, sections ){
                list<string> entries;
                reg.EnumerateEntries( *iter, &entries );
                ITERATE( list<string>, it, entries ){
                    if( 
                        (*it).find( "_PROXY_USER" ) == string::npos
                        && (*it).find( "_PROXY_PASS" ) == string::npos 
                    ){
                        const string& s = reg.Get(*iter, *it);
                        LOG_POST(Info << "  [" << *iter << "] " << *it << " = " << s );
                    } else { // hide proxy details including password
                        //const string& s = reg.Get(*iter, *it);
                        LOG_POST(Info << "  [" << *iter << "] " << *it << " = xxxx" );
                    }
                }
            }
        } else {
            LOG_POST(Error << "Config file is empty");
        }
    } else {
        LOG_POST(Error << "No config file found");
    }
}


bool CGBenchApplication::x_HandleRemote(CArgs& args)
{
    bool loaded = false;

    string type = args["t"].AsString();

    vector<string> fnames;
    x_GetArgs(args, fnames);
    string fname;

    try {
        ITERATE(vector<string>, it, fnames) {
            fname = *it;
            CGBenchPipe::OpenRemote(fname, type);
            loaded = true;
        }
        // Just activate application
        // We want a single instance of GBench running. JIRA: GB-1622
        if (!loaded) {
            CGBenchPipe::OpenRemote("", "app_activate");
            loaded = true;
        }
    }
    catch (const std::exception&) {
        // we don't want to report error if there is no GBench instance running
        //LOG_POST(Warning << "error opening " << fname << ": " << e.what());
    }

    return loaded;
}

void CGBenchApplication::x_HandleArgs(CArgs& args)
{
    if( args["edit"] ){
        wxGetApp().SetEditMode( true );
        m_DefaultPkgs.push_back( "sequence_edit" );
    }
#ifdef __WXMAC__
    else {
        if (m_MacStartInEditMode) {
            wxGetApp().SetEditMode(true);
            m_DefaultPkgs.push_back( "sequence_edit" );
        }
    }
#endif


    x_GetArgs(args, wxGetApp().m_Args);
}

void CGBenchApplication::x_GetArgs(CArgs& args, vector<string>& fnames)
{
    size_t items = args.GetNExtra();

    for (size_t i = 0;  i < items;  ++i) {
        string arg_name = "#" + NStr::SizetToString(i + 1);
        string arg = args[arg_name].AsString();
#if defined(WIN32)
        wxString filename = wxString::FromUTF8(arg.c_str());
#else
        wxString filename(arg.c_str(), wxConvLibc);
#endif
        wxFileName fname(filename);
        if (!fname.IsAbsolute())
            fname.MakeAbsolute();
        fnames.push_back(string(fname.GetFullPath().ToUTF8()));
    }
}

static const wxChar*  kDefUserScoreDir = wxT("<home>/align_scores");
static const wxChar*  kDefScoreDir = wxT("<std>/etc/align_scores");
static const int    kMaxMRUMethods = 7;

static const char*  kUIToolsRegKey = "GBENCH.UITools";

void CGBenchApplication::x_ConfigUIToolRegistry()
{
    list<wxString> dirs;
    dirs.push_back(kDefUserScoreDir);
    dirs.push_back(kDefScoreDir);

    CUIToolRegistry* reg = CUIToolRegistry::GetInstance();
    reg->LoadTemplateToolsInfo(dirs);

    reg->SetRegistryPath(kUIToolsRegKey);
    reg->LoadSettings();
}


wxString CGBenchApplication::GetGuardFilepath()
{
    static wxString guard_path;
    if( guard_path.IsEmpty() ){
        guard_path.Printf(wxT("<home>/gbench-pid.%u"), (int)CProcess::GetCurrentPid() );
        guard_path = CSysPath::ResolvePath( guard_path );
    }

    return guard_path;

    /*
    string pid_str = NStr::IntToString(CProcess::GetCurrentPid());
    string path = "<home>/gbench-pid." + pid_str;
    path = CSystemPath::ResolvePath(path);
    return ToWxString( path );
    */
}


void CGBenchApplication::x_StartMonitor()
{
    // enable monitor only in release builds
#ifndef NODEF //_DEBUG 
    /// establish our monitoring process; this is only done for release builds!
    try {
        /// write the PID to a guard file
        wxString guard_path = GetGuardFilepath();
        {{
            CNcbiOfstream ostr( guard_path.fn_str() );
            ostr << CTime( CTime::eCurrent ).AsString();
        }}

        // protection for spaces in filename
        if( guard_path.find_first_of(wxT(" ")) != string::npos ){
#ifdef NCBI_OS_MSWIN
            guard_path = wxT("\"") + guard_path + wxT("\"");
#else
            guard_path.Replace(wxT(" "), wxT("\\ "));
#endif
        }

#if 1
        //! We revert back NCBI Toolkit implementation of gbench monitor start
        //! as far as we cannot use wxWidgets implementation outside wxApp run cycle

        string monitor_path = CSystemPath::ResolvePath( "<bin>/gbench_monitor" );

        string pid_str = NStr::NumericToString( CProcess::GetCurrentPid() );

        LOG_POST(Info 
            << "Starting process monitor: " << monitor_path << " " << pid_str << " " << guard_path 
        );

        // Use eDetach mode on Unix to let monitor continue to work on gbench crash
        // Use eNoWait mode on Windows to hide console window 
#       ifdef NCBI_OS_MSWIN
        CExec::EMode mode = CExec::eNoWait;
#       else
        CExec::EMode mode = CExec::eDetach;
#       endif

        //wxWritableCharBuffer p = guard_path.char_str();
        
        CExec::CResult res = CExec::RunSilent(
            mode, monitor_path.c_str(), pid_str.c_str(), guard_path.ToUTF8().data(), NULL
        );
        if( ! res.GetProcessHandle() ){
            ERR_POST("Failed to start process monitor - cannot get process handle");
        }

#else
        wxString monitor_path = CSysPath::ResolvePath( wxT("<bin>/gbench_monitor") );

        //! We revert back NCBI Toolkit implementation of gbench monitor start
        //! as far as we cannot use wxWidgets implementation outside wxApp run cycle

        string pid_str = NStr::IntToString( CProcess::GetCurrentPid() );
        // wxString pid_str = wxString::Format( wxT("%u"), (int)::wxGetProcessId() );

        LOG_POST(Info 
            << "Starting process monitor: " << monitor_path << " " << pid_str << " " << guard_path 
        );

        wxString command = monitor_path + wxT(" ") + pid_str + wxT(" ") + guard_path;
        long pid = ::wxExecute( command, wxEXEC_ASYNC|wxEXEC_MAKE_GROUP_LEADER );
        if( pid == 0 ){
            ERR_POST("Failed to start process monitor - cannot get process handle");
        }
#endif

    } catch( CException& e ){
        ERR_POST( "Failed to start process monitor - " << e.GetMsg() );
    }
#endif
}


/// stops monitoring process
void CGBenchApplication::x_StopMonitor()
{
    // enable monitor only in release builds
#ifndef _DEBUG
    LOG_POST( Info << "Shutting down process monitor..." );

    /// delete our PID file - this is a sign of successful completion
    ::wxRemoveFile( GetGuardFilepath() );
    LOG_POST( Info << "Finsihed shutting down process monitor" );
#endif
}

void CGBenchApplication::x_StartCrashMonitor()
{
    CGuiRegistry& registry = CGuiRegistry::GetInstance();
#ifndef _DEBUG
    try {
        wxString path = GetCrashFilepath();
        if( wxFileName::FileExists( path ) ){
            registry.Set( "GBENCH.Application.CrashDetected", true );
            ::wxRemoveFile( path );

        } else {
            registry.Set( "GBENCH.Application.CrashDetected", false );
        }

        CNcbiOfstream ostr( path.fn_str() );
        ostr << CTime( CTime::eCurrent ).AsString();
    }
    catch( CException& e ){
        ERR_POST( "Failed to create crash monitor - " << e.GetMsg() );
    }
#else
    registry.Set( "GBENCH.Application.CrashDetected", false );
#endif

}

void CGBenchApplication::x_StopCrashMonitor()
{
#ifndef _DEBUG
    wxLogNull no_logger;
    ::wxRemoveFile( GetCrashFilepath() );
#endif
}

wxString CGBenchApplication::GetCrashFilepath()
{
    static wxString s_crash_path;
    if( s_crash_path.IsEmpty() ){
        s_crash_path = CSysPath::ResolvePath( wxT("<home>/gbench-crash") );
    }

    return s_crash_path;
}

bool CGBenchApplication::LoadConfig(
    CNcbiRegistry& reg, const string* conf, CNcbiRegistry::TFlags flags
){
    /// load the standard NCBI .ini config file

    string config_path;
    if( !conf || conf->empty() ){
        wxString wx_config_path = CSysPath::ResolvePathExisting( _("<home>/gbench.ini, <std>/etc/gbench.ini") );
        if ( !wx_config_path.empty() ){
            config_path = wx_config_path.ToUTF8().data();
            conf = &config_path;
        }
    }

    if( !conf ){
        LOG_POST( "No NCBI config file found" );

        return false;
    }


    // Let the base class load the config file.
    // If we couldn't find it above, it will try and find it, and warn otherwise.
    LOG_POST( "Loading NCBI config file from " << *conf );
    return CNcbiApplication::LoadConfig( reg, conf, flags );
}


void CGBenchApplication::SaveConfig()
{
    /// save the basic INI file registry
    if( GetConfig().Modified() ){
        wxString home_path = CSysPath::ResolvePath( wxT("<home>") );
        if ( !wxFileName::DirExists( home_path ) && !::wxMkdir( home_path ) ){
            return;
        }

        wxFileName fname( home_path, wxT("gbench.ini") );
        CNcbiOfstream ofstr( fname.GetFullPath().fn_str(), ios::out|ios::binary );
        GetConfig().Write( ofstr );
    }
}


static const char*  kUIPkgManager = "GBENCH.UIPkgManager";

void CGBenchApplication::x_InitPackages()
{
    if (!m_PkgManager.get())
        m_PkgManager.reset(new CPkgManager());

    size_t ver_major = 0;
    size_t ver_minor = 0;
    size_t ver_patch_level = 0;
    CTime build_date;
    GetGBenchVersionInfo(ver_major, ver_minor, ver_patch_level, build_date);
    m_PkgManager->SetAppPkg("GBench", ver_major, ver_minor, ver_patch_level);

    ITERATE( vector<string>, pkg_ir, m_DefaultPkgs ){
        m_PkgManager->AddDefaultPkg( *pkg_ir );
    }
    m_PkgManager->SetRegistryPath(kUIPkgManager);

    wxString packages_path =
        CSysPath::ResolvePath(wxT("<std>/packages"));
    if ( !packages_path.empty() ) {
        m_PkgManager->ScanPackages(packages_path);
    }

    m_PkgManager->ValidatePackages();
    m_PkgManager->LoadPackages();
}

///////////////////////////////////////////////////////////////////////////////
/// CGBenchApp

///////////////////////////////////////////////////////////////////////////////
/// CNCBI_GUIAppTraits

// create Traits class in order to install our own wxLog implementation
class CGBenchAppTraits : public wxGUIAppTraits
{
    virtual wxLog *CreateLogTarget()
    {
        return CwxLogDiagHandler::GetInstance();
    }
#if wxUSE_INTL
    // called during wxApp initialization to set the locale to correspond to
    // the user default (i.e. system locale under Windows, LC_ALL under Unix)
    virtual void SetLocale() {}
#endif // wxUSE_INTL
};


wxAppTraits *CGBenchApp::CreateTraits()
{
    return new CGBenchAppTraits();
}


void CGBenchApplication::Raise()
{
    CGBenchApp& app = wxGetApp();
    wxWindow* win = app.GetTopWindow();
    if(win) {
        win->Raise();
    }
}

bool CGBenchApp::x_TestNcbiConnection()
{
    // Performing short connection test.
    CRef<CConnTestThread> thread(new CConnTestThread(5));
    thread->Run();
    thread->Join();
    if (thread->GoodConnection()) {
        LOG_POST(Info << "Short connection test SUCCEEDED");
        return true;
    }

    LOG_POST(Error << "Short connection test FAILED");

    string descr = "Testing NCBI connection";
    thread.Reset(new CConnTestThread(30));
    thread->Run();

    if (CConnTest::IsNcbiInhouseClient())
        return true;

    int res = wxID_OK;

    if (!thread->Finished()) {
        CTestingNcbiConnDlg dlg;
        dlg.SetConnThread(thread.GetPointer());
        dlg.Create(NULL);
        res = dlg.ShowModal();
    }


    if (res != wxID_OK) {
        LOG_POST(Error << descr << ": CANCELLED");
        return true;
    }

    thread->Join();

    string connInfo = thread->GetConnInfo();
    if (thread->GoodConnection()) {
        LOG_POST(Info << descr << ": SUCCEEDED");
    }
    else {
        LOG_POST(Error << descr << ": FAILED");
        LOG_POST(Error << connInfo);

        CNetConnProblemDlg dlg;
        dlg.SetDescr(ToWxString(connInfo));
        dlg.Create(NULL);
        dlg.ShowModal();
    }

    return true;
}

void CGBenchApp::x_InitAppJobDispatcher()
{
    m_MaxThreads = 3;
    m_MaxJobs = 1000;

    /// this will register Thread Pool engine
    CwxNCBIApp::x_InitAppJobDispatcher();

    CAppJobDispatcher& disp = CAppJobDispatcher::GetInstance();
    disp.RegisterEngine("ObjManagerEngine", *new CObjectManagerEngine());
    disp.RegisterEngine("ObjManagerEngineAsyncCall", *new CObjectManagerEngine(2));
    disp.RegisterEngine("Scheduler", *new CSchedulerEngine());
}

bool CGBenchApp::OnInit()
{
    CwxLogDiagHandler::GetInstance();

	bool ok = CwxNCBIApp::OnInit();

	CAppTaskService* task_service = m_Workbench->GetAppTaskService();

	// Send app-start message
	CVersionCheckTask* vcheck_task = new CVersionCheckTask();
	task_service->AddTask( *vcheck_task );

	return ok;
}

int CGBenchApp::OnExit()
{
	if( kGenericRequest ){
		// Send app-end message
		CGBenchVersionInfo cur_version;
		GetGBenchVersionInfo( cur_version );
		CGBenchService().SendAppFinish( cur_version );
	}

	return CwxNCBIApp::OnExit();
}

bool CGBenchApp::x_CreateGUI()
{
    LOG_POST(Info << "Creating application GUI...");

    // create Workbench instance with m_GUI as adviser
    m_Workbench.reset(new CWorkbench(&m_GUI));

    /// set path in Registry, we need to this before calling Init()
    m_Workbench->SetRegistryPath("Workbench");

    m_GUI.SetWorkbench(m_Workbench.get());
    m_GUI.SetRegistryPath("ApplicationGUI");

    m_Workbench->Init();
   
    // On mac edit mode is indicated by pressing down the option key when starting the application
    // rather than having multipl executables.  Allow user to cancel that option here. 
#ifdef __WXMAC__
    if (wxGetApp().GetEditMode()) {
        int result = wxMessageBox(wxT("Start Genome Workbench in edit mode?"), 
                                  wxT("Genome Workbench Start Options"), wxYES_NO);
        if (result == wxNO)
            wxGetApp().SetEditMode(false);
    }
#endif

    //! Hard-coded logic for Submission Perspective. 
    //! It should be changed in future.
    if( wxGetApp().GetEditMode() ){
        m_Workbench->OpenPerspective( "edit" );
    } else {
        m_GUI.RestoreWindowLayout();
    }

    m_readPipeThread.Reset(new CReadPipeThread(m_Workbench.get()));
    m_readPipeThread->Run();

    CAppTaskService* task_service = m_Workbench->GetAppTaskService();

    // Load files provided as arguments on command line
    ITERATE(vector<string>, it, m_Args) {
        CNamedPipeProcessTask* task = new CNamedPipeProcessTask(m_Workbench.get(), *it, false);
        task_service->AddTask(*task);
    }
    m_Args.clear();

    m_WakeupThread.Reset(new CWakeupThread());
    m_WakeupThread->Run();

    string msg = wxGetApp().GetStartUpString();
    if( !msg.empty() ){
        //NcbiInfoBox(msg, "New Version of Genome Workbench Is Available!");
        NcbiInfoBox( msg, "Welcome to Genome Workbench!" );
    }

    LOG_POST(Info << "Finished creating application GUI");
    return true;
}


void CGBenchApp::x_DestroyGUI()
{
    if (m_WakeupThread) {
        m_WakeupThread->RequestStop();
        m_WakeupThread->Join();
    }
    if (m_readPipeThread) {
        m_readPipeThread->RequestStop();
        m_readPipeThread->Join();
    }

    LOG_POST(Info << "Destroying application GUI...");

    // m_Workbench already has been shut down
    m_GUI.SetWorkbench(NULL);

    m_Workbench.reset();

    LOG_POST(Info << "Finished destroying application GUI");
}

// we override this function to add idle processing for CAppTaskService
bool CGBenchApp::x_OnIdle()
{
    // we handle more urgent requests first
    bool handled = CwxNCBIApp::x_OnIdle();

    if( ! handled  &&  m_Workbench.get() != NULL)  {
        // if Task Service is available - give it a chance to execute
        CAppTaskService* task_mgr = m_Workbench->GetAppTaskService();
        if(task_mgr) {
            handled = task_mgr->IdleCallback();
        }
    }
    return handled;
}

wxString CGBenchApp::x_GetGuiRegistryPath()
{
    wxString path = CSysPath::ResolvePath( wxT("<home>/gbench.asn") );
    return path;
}

wxString CGBenchApp::x_GetLayoutRegistryPath()
{
    wxString path = CSysPath::ResolvePath( wxT("<home>/window_layout2.asn") );
    return path;
}

void CGBenchApp::x_LoadGuiRegistry()
{
    if( m_RegistryLoaded ){
        return;
    }

    /// load the standard NCBI .ini file registry

    CGuiRegistry& gui_reg = CGuiRegistry::GetInstance();

    /// Set the global repository. We have two repository files to look for
    wxString path = CSysPath::ResolvePath( wxT("<std>/etc/gbench.asn") );
    if( wxFileName::FileExists( path ) ){
        CNcbiIfstream istr( path.fn_str() );
        if( istr ){
            LOG_POST( Info << "Loading base GUI registry from " << path );
            gui_reg.AddSite( istr, CGuiRegistry::ePriority_Global );
        }
    }

    path = CSysPath::ResolvePath( wxT("<std>/etc/def_window_layout.asn") );
    if( wxFileName::FileExists( path ) ){
        CNcbiIfstream istr( path.fn_str() );
        if( istr ){
            LOG_POST( Info << "Loading window layout registry from " << path );
            gui_reg.AddSite( istr, CGuiRegistry::ePriority_Global -2);
        }
    }

    path = CSysPath::ResolvePath( wxT("<std>/etc/plugin_config.asn") );
    if( wxFileName::FileExists( path ) ){
        CNcbiIfstream istr( path.fn_str() );
        if( istr ){
            LOG_POST( Info << "Loading plugin config settings from " << path );
            gui_reg.AddSite( istr, CGuiRegistry::ePriority_Global -1 );
        }
    }

    /// set the local repository
    CwxNCBIApp::x_LoadGuiRegistry();

    /// look for any site-specific files and load these
    {{
        CGuiRegistry::TReadView view = gui_reg.GetReadView( "GBENCH.Application.Sites" );

        vector<string> site_names;
        view.GetStringVec( "ActiveSites", site_names );
        int site_prioty = CGuiRegistry::ePriority_Local;

        set<wxString> visited_sites;
        ITERATE( vector<string>, iter, site_names ){
            // TODO FromASNString
            path = wxString::FromAscii( 
                view.GetString( CGuiRegistryUtil::MakeKey("Sites", *iter) ).c_str() 
            );
            path = CSysPath::ResolvePathExisting( path );
            if( path.empty() ){
                continue;
            }

            /// don't add a site more than once
            if( visited_sites.insert( path ).second == false ){
                continue;
            }

            CNcbiIfstream istr( path.fn_str() );
            if( istr ){
                LOG_POST( Info << "Loading site registry file from " << path );
                gui_reg.AddSite( istr, ++site_prioty );
            }
        }
    }}

    {{ // x_SyncRegistryAndEnvironment()
        CNcbiApplication* app = CNcbiApplication::Instance();
        CGuiRegistry& gui_reg = CGuiRegistry::GetInstance();

        /// propagate to environment where appropriate
#ifdef NCBI_OS_MSWIN
        const char* path_sep = ";";
#else
        const char* path_sep = ":";
#endif
        CGuiRegistry::TReadView env_view =
            gui_reg.GetReadView( "GBENCH.Application.Environment" )
        ;

        CGuiRegistry::TKeys keys;
        env_view.GetTopKeys( keys );

        ITERATE( CGuiRegistry::TKeys, keyIter, keys ){
            string key = keyIter -> key;
            if( NStr::EndsWith(key, CGuiRegistry::kMetaSuffix) ){
                continue;
            }
            string policy_key = key + CGuiRegistry::kMetaSuffix;
            string policy =
                env_view.GetString(CGuiRegistryUtil::MakeKey(policy_key, "MergePolicy"))
            ;

            vector<string> val;
            env_view.GetStringVec(key, val);
    #if defined(_DEBUG)  &&  defined(NCBI_COMPILER_MSVC)
            {{
                ///
                /// HACK: on Windows, for debug builds, we require a different
                /// path for some variables
                ///
                NON_CONST_ITERATE (vector<string>, iter, val) {
                    if (iter->find("ReleaseDLL") != string::npos) {
                        *iter = NStr::Replace(*iter, "ReleaseDLL", "DebugDLL");
                    }
                }
            }}
    #endif
            if (0 == NStr::CompareNocase(policy, "append")) {
                if (val.size() > 0) {
                    string s = app->GetEnvironment().Get(key);
                    s += path_sep;
                    s += NStr::Join(val, path_sep);
                    app->SetEnvironment(keyIter -> key, s);
                }
            } else if (0 == NStr::CompareNocase(policy, "prepend")) {
                if( val.size() > 0 )    {
                    string s = NStr::Join(val, path_sep);
                    s += path_sep;
                    s += app->GetEnvironment().Get(key);
                    app->SetEnvironment(key, s);
                }
            } else {
                string value = env_view.GetString(key);
                if( ! value.empty() )   {
                    app->SetEnvironment(key, value);
                }
            }
        }

        /// propagate to some internal variables where appropriate
        // Same thing for NcbiEnvironment*
        CGuiRegistry::TReadView reg_view = gui_reg.GetReadView("GBENCH.Application.Registry");

        keys.clear();
        reg_view.GetTopKeys(keys);

        ITERATE( CGuiRegistry::TKeys, keyIter, keys ){
            string key = keyIter->key;
            gui_reg.Set( key, reg_view.GetString(key) );
        }
    }}

    /// seq-id accession mapping guide
    {{
        wxString fname = ToWxString(gui_reg.GetString("GBENCH.Application.Config.AccessionGuide"));
        if( fname.empty() ){
            // TODO this won't work with intl fnames
            fname = CSysPath::ResolvePathExisting( wxT("<std>/etc/accguide.txt") );
        }
        if( !fname.empty() ){
            try {
                CNcbiIfstream istr(fname.fn_str());
                CStreamLineReader lreader(istr);
                objects::CSeq_id::LoadAccessionGuide( lreader );
                LOG_POST( Info << "loaded sequence accession map from " << fname );
            }
            STD_CATCH_ALL( "CGBenchApplication::x_LoadGuiRegistry(): failed to load accession guide" );
        }
    }}
}


void CGBenchApp::x_SaveGuiRegistry()
{
    CwxNCBIApp::x_SaveGuiRegistry();
}

#ifdef NCBI_OS_DARWIN
void CGBenchApp::MacOpenFile(const wxString &fileName)
{
	string fname = ToStdString(fileName);

	try {
	 CGBenchPipe::OpenRemote(fname);
	}
	catch (CException& e) {
	 LOG_POST(Warning << "error opening " << fname << ": " << e.GetMsg());
	}
	catch (std::exception& e) {
	 LOG_POST(Warning << "error opening " << fname << ": " << e.what());
	}
	catch (...) {
	 LOG_POST(Warning << "unknown error opening " << fname);
	}
}
#endif

END_NCBI_SCOPE;

/////////////////////////////////////////////////////////////////////////////
//  main entry point

int main(int argc, const char* argv[])
{
    ///
    /// 5/23/2008: HACK
    /// we set a few environment variables manually here to make sure that
    /// we get a safety blanket initialization for gbench
    /// CObject's initialization changed recently.
    ///
    {{
#if defined(_DEBUG)
#ifdef NCBI_COMPILER_MSVC
        /// disable fill for MSVC - Microsoft's debug heap already does this
        ::putenv("NCBI_MEMORY_FILL=NONE");
#endif
#elif defined(NDEBUG)
        /// re-enable zero fill
        ::putenv("NCBI_MEMORY_FILL=ZERO");
#endif
    }}
    ///
    /// end HACK
    ///

    // Put mode check in as early as possible so user doesn't have
    // to hold option key down too long
#ifdef __WXMAC__
    if (ncbi::GetMacOptionKeyDown()) {
        ncbi::CGBenchApplication::m_MacStartInEditMode = true;
    }
#endif

    wxDISABLE_DEBUG_SUPPORT();

    ncbi::CGBenchApplication ncbi_app;

    // Execute main application function
    //
    // Use "Genome-Workbench" (NOT "Genome Workbench") to produce standards compliant HTTP request
    // "User-Agent:" 
    //
    return ncbi_app.AppMain(argc, argv, 0, ncbi::eDS_Default, 0, "Genome-Workbench");
}
