/*  $Id: gui.cpp 27956 2013-04-30 17:55:33Z katargir $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Andrey Yazhuk
 *
 */

#include <ncbi_pch.hpp>
#include <corelib/ncbistd.hpp>
#include <corelib/ncbitime.hpp>
#include <util/thread_pool.hpp>

#include "gui.hpp"
#include "app_services.hpp"
#include "feedback.hpp"
#include "aboutdlg.hpp"

#include <gui/core/selection_service_impl.hpp>
#include <gui/core/app_explorer_service.hpp>
#include <gui/core/app_explorer_cmd_handler.hpp>
#include <gui/core/data_mining_service.hpp>
#include <gui/core/ui_data_source_service.hpp>
#include <gui/core/project_cmd_handler.hpp>
#include <gui/core/project_service.hpp>
#include <gui/core/project_view_base_impl.hpp>
#include <gui/core/visible_range_service_impl.hpp>
#include <gui/core/download_job.hpp>
#include <gui/core/quick_launch_list.hpp>
#include <gui/core/recent_tool_list.hpp>
#include <gui/core/ui_tool_manager.hpp>
#include <gui/core/file_extensions.hpp>

#include <gui/core/project_task.hpp> //TODO
#include <gui/core/project_service_test.hpp> //TODO
#include <gui/core/app_tasks.hpp>

#include <gui/objutils/visible_range.hpp>
#include <gui/objutils/registry.hpp>

#include <gui/core/broadcast_settings_dlg.hpp>
#include <gui/core/winmask_files.hpp>
#include <gui/core/winmask_files_dlg.hpp>

#include <gui/core/commands.hpp>

#include <gui/framework/view_manager_service.hpp>
#include <gui/framework/window_manager_service.hpp>
#include <gui/framework/event_log_service_impl.hpp>
#include <gui/framework/status_bar_service.hpp>
#include <gui/framework/app_task_service.hpp>
#include <gui/framework/status_bar_service_impl.hpp>
#include <gui/framework/pkg_manager.hpp>
#include <gui/framework/gui_package.hpp>
#include <gui/framework/pkg_wb_connect.hpp>

#include <gui/widgets/wx/dock_layout.hpp>
#include <gui/widgets/wx/wx_utils.hpp>
#include <gui/widgets/wx/fileartprov.hpp>
#include <gui/widgets/wx/commands.hpp>
#include <gui/widgets/wx/ui_command.hpp>
#include <gui/widgets/wx/message_box.hpp>
#include <gui/widgets/wx/glresetdlg.hpp>
#include <gui/widgets/wx/wx_utils.hpp>

#include <gui/opengl/ftglfontmanager.hpp>
#include <gui/opengl/glresmgr.hpp>

#include <gui/objutils/label.hpp>
#include <gui/objutils/user_type.hpp>

#include <gui/widgets/wx/sys_path.hpp>
#include <gui/utils/extension_impl.hpp>

#include <objects/general/User_object.hpp>

#include <gui/framework/app_options_dlg.hpp>
#include <gui/framework/app_packages_dlg.hpp>

#include <gui/widgets/macro_edit/edit_macro_dlg.hpp>

#include <wx/menu.h>
#include <wx/app.h>
#include <wx/log.h>
#include <wx/msgdlg.h>
#include <wx/toolbar.h>
#include <wx/aui/auibar.h>
#include <wx/dnd.h>
#include <wx/filename.h>
#include <wx/filedlg.h>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

///////////////////////////////////////////////////////////////////////////////
/// CGBenchGUIDropTarget
///
class CGBenchGUIDropTarget : public wxDropTarget
{
public:
    CGBenchGUIDropTarget(CGBenchGUI& gui)
    :   m_GUI(gui)
    {
        SetDataObject(new wxFileDataObject);
    }

    virtual wxDragResult OnData(wxCoord x, wxCoord y, wxDragResult def)
    {
        if ( !GetData() )
            return wxDragNone;

        wxWindow* wnd = m_GUI.x_GetMainWindow();

        if (wnd && !wnd->IsEnabled())
            return wxDragNone;

        wxFileDataObject* dobj = (wxFileDataObject *)m_dataObject;

        if(dobj  &&  (def == wxDragMove  ||  def == wxDragCopy || def == wxDragLink)) {
            wxArrayString filenames = dobj->GetFilenames();

            vector<wxString> names;
            size_t n = filenames.GetCount();
            names.reserve(n);
            for( size_t i = 0; i < n; i++ ){
                names.push_back (filenames[i]);
            }

            m_GUI.ShowOpenDlg(names);
            return def;
        }
        return wxDragError;
    }
protected:
    CGBenchGUI& m_GUI;
};


///////////////////////////////////////////////////////////////////////////////
/// COpenViewEventRecord - event log record for "open view" event.
class COpenViewEventRecord :
    public CEventRecord,
    public IEventRecordAction
{
public:
    COpenViewEventRecord(CGBenchGUI* gui, const string& view_name, IWMClient* client)
        :   CEventRecord(eInfo),
            m_GUI(gui), m_Client(client)
    {
        _ASSERT(m_GUI  &&  m_Client);

        string s = "Created " + view_name;
        SetTitle(s);

        s = "Created a new view " + view_name + " and added to Workbench.";
        SetDescription(s);

        SetTime(time(0));
    }

    virtual IEventRecordAction* GetAction() const
    {
        const IEventRecordAction* action = this;
        return const_cast<IEventRecordAction*>(action);
    }

    /// IEventRecordAction implementation
    virtual string  GetLabel()
    {
        return "Activate view";
    }

    virtual void Do()
    {
        m_GUI->ActivateClient(m_Client);
    }

protected:
    CGBenchGUI* m_GUI;
    IWMClient* m_Client;
};


///////////////////////////////////////////////////////////////////////////////
/// CGBenchGUIMenuDisabler - disables commands not handled by other components
/// such as views. (this is a fall back update handler)

class CGBenchGUIMenuDisabler : public wxEvtHandler
{
    DECLARE_EVENT_TABLE();
public:
    ~CGBenchGUIMenuDisabler()
    {
    }
    void OnDisableCmdUpdate(wxUpdateUIEvent& event)
    {
        event.Enable(false);
    }
};


BEGIN_EVENT_TABLE(CGBenchGUIMenuDisabler, wxEvtHandler)
    EVT_UPDATE_UI( eCmdBroadcastSel, CGBenchGUIMenuDisabler::OnDisableCmdUpdate )
    EVT_UPDATE_UI_RANGE(eCmdBack, eCmdForward,
                        CGBenchGUIMenuDisabler::OnDisableCmdUpdate)
    EVT_UPDATE_UI_RANGE(eCmdZoomIn, eCmdSetEqualScale,
                        CGBenchGUIMenuDisabler::OnDisableCmdUpdate)
END_EVENT_TABLE()


///////////////////////////////////////////////////////////////////////////////
/// CGBenchGUI

static int const kDiagTimerID = 1;
static int const kStatusTimerID = 2;

BEGIN_EVENT_TABLE(CGBenchGUI, wxEvtHandler)
    EVT_MENU( wxID_OPEN, CGBenchGUI::OnFileOpen )
    EVT_MENU( wxID_ABOUT, CGBenchGUI::OnHelpAbout )
    EVT_MENU( eCmdAbout, CGBenchGUI::OnHelpAbout )
    EVT_UPDATE_UI(wxID_OPEN, CGBenchGUI::OnEnableCmdUpdate )
    EVT_MENU(eCmdNewWorkspace, CGBenchGUI::OnNewWorkspace )
    EVT_UPDATE_UI( eCmdNewWorkspace, CGBenchGUI::OnEnableCmdUpdate )
    EVT_UPDATE_UI( wxID_PRINT, CGBenchGUI::OnEnablePrintCmdUpdate )
    EVT_UPDATE_UI( eCmdSaveImages, CGBenchGUI::OnEnableSaveImagesCmdUpdate )
    EVT_UPDATE_UI( eCmdSavePdf, CGBenchGUI::OnEnableSavePdfCmdUpdate )

    EVT_MENU(eCmdCloseWorkspace, CGBenchGUI::OnCloseWorkspace )
    EVT_UPDATE_UI(eCmdCloseWorkspace, CGBenchGUI::OnHasWorskpaceUpdate )

    EVT_MENU(eCmdShowProjectTreeView, CGBenchGUI::OnShowProjectTreeView )
    EVT_MENU(eCmdShowSelectionInspector, CGBenchGUI::OnShowSelectionInspector )
    EVT_MENU(eCmdShowDiagConsole, CGBenchGUI::OnShowDiagnosticsConsole )
    EVT_MENU(eCmdShowEventView, CGBenchGUI::OnShowEventView )
    EVT_MENU(eCmdShowTaskView, CGBenchGUI::OnShowTaskView )
    EVT_MENU(eCmdShowSearchView, CGBenchGUI::OnShowSearchView )

    // always enable these commands
    EVT_UPDATE_UI_RANGE( eCmdShowProjectTreeView, eCmdShowDiagConsole, CGBenchGUI::OnEnableCmdUpdate )

    EVT_MENU( eCmdBroadcastOptions, CGBenchGUI::OnBroadcastOptions )
    EVT_UPDATE_UI( eCmdBroadcastOptions, CGBenchGUI::OnEnableCmdUpdate )

    EVT_MENU( eCmdWindowMaskerFiles, CGBenchGUI::OnWindowMaskerFiles )
    EVT_UPDATE_UI( eCmdWindowMaskerFiles, CGBenchGUI::OnEnableCmdUpdate )

    EVT_MENU(eCmdToolsOptions, CGBenchGUI::OnToolsOptions)
    EVT_MENU(eCmdToolsPackages, CGBenchGUI::OnToolsPackages)

    EVT_MENU(eCmdCloseAllProjectViews, CGBenchGUI::OnCloseAllProjectViews)
    EVT_MENU(eCmdSaveLayout, CGBenchGUI::OnSaveLayout )
    EVT_UPDATE_UI( eCmdSaveLayout, CGBenchGUI::OnSaveLayoutUpdate )
    EVT_MENU(eCmdLoadLayout, CGBenchGUI::OnLoadLayout )
    EVT_UPDATE_UI( eCmdLoadLayout, CGBenchGUI::OnLoadLayoutUpdate )
    EVT_MENU(eCmdResetLayout, CGBenchGUI::OnResetLayout )
    EVT_UPDATE_UI( eCmdResetLayout, CGBenchGUI::OnResetLayoutUpdate )

    EVT_MENU(eCmdOpenView, CGBenchGUI::OnOpenView )
    EVT_UPDATE_UI( eCmdOpenView, CGBenchGUI::OnEnableCmdUpdate )
    EVT_MENU(eCmdOpenViewDefault, CGBenchGUI::OnOpenViewDefault )

    EVT_MENU(eCmdRunTool, CGBenchGUI::OnRunTool )
    EVT_UPDATE_UI( eCmdRunTool, CGBenchGUI::OnEnableCmdUpdate )

    EVT_MENU_RANGE(eCmdQuickLaunchStart, eCmdQuickLaunchEnd,  CGBenchGUI::OnQuickLaunch)
    EVT_MENU_RANGE(eCmdRecentToolStart, eCmdRecentToolEnd,  CGBenchGUI::OnRecentTool)
    
    EVT_MENU(eCmdSearch, CGBenchGUI::OnShowSearchView )
    EVT_UPDATE_UI( eCmdSearch, CGBenchGUI::OnEnableCmdUpdate )

    EVT_MENU(eCmdSendFeedback, CGBenchGUI::OnSendFeedback )
    EVT_UPDATE_UI( eCmdSendFeedback, CGBenchGUI::OnEnableCmdUpdate ) // enable always

    EVT_MENU(eCmdTestPrjSrv, CGBenchGUI::OnTestPrjSrv )
    EVT_MENU(eCmdTestLoadBlastMask, CGBenchGUI::OnTestLoadBlastMask )
    EVT_MENU(eCmdMacroEditDlg, CGBenchGUI::OnMacroEditor)

END_EVENT_TABLE()


CGBenchGUI::CGBenchGUI() : m_Workbench(), m_MenuService(), m_GBenchGUIApp()
{
#ifdef __WXMAC__
    wxApp::s_macAboutMenuItemId = eCmdAbout;
#endif
}


CGBenchGUI::~CGBenchGUI()
{
    /*ITERATE(set<CExplorerItem*>, it, CExplorerItem::sm_AliveItems) {
        CExplorerItem* item = *it;
        string s = typeid(*item).name();
        LOG_POST(Error << s << "  " << item->GetLabel() << " is not deleted");
    }
    _ASSERT(CExplorerItem::sm_AliveItems.empty());*/
}


void CGBenchGUI::SetRegistryPath(const string& reg_path)
{
    m_RegPath = reg_path;
}


void CGBenchGUI::SetWorkbench(IWorkbench* workbench)
{
    if(workbench)   {
        m_Workbench = workbench;
    } else {
        m_Workbench = workbench;
    }
}


string CGBenchGUI::GetAppTitle()
{
    return "NCBI Genome Workbench";
}


void CGBenchGUI::x_PushEvtHandler( wxEvtHandler* new_handler )
{
    if( !new_handler ) return;

    wxEvtHandler* wk_handler = dynamic_cast<wxEvtHandler*>(m_Workbench);
    wxEvtHandler* old_handler = wk_handler->GetNextHandler();

    wk_handler->SetNextHandler(new_handler);
    new_handler->SetPreviousHandler(wk_handler);

    if( old_handler ){
        new_handler->SetNextHandler(old_handler);
        old_handler->SetPreviousHandler(new_handler);
    }
}


wxEvtHandler* CGBenchGUI::x_PopEvtHandler()
{
    wxEvtHandler* wk_handler = dynamic_cast<wxEvtHandler*>(m_Workbench);

    wxEvtHandler* handler = wk_handler->GetNextHandler();

    if(handler)  {
        wxEvtHandler* next_handler = handler->GetNextHandler();

        handler->SetPreviousHandler(NULL);
        wk_handler->SetNextHandler(next_handler);

        if(next_handler) {
            handler->SetNextHandler(NULL);
            next_handler->SetPreviousHandler(wk_handler);
        }
    }
    return handler;
}


void CGBenchGUI::PostCreateWindow()
{
    _ASSERT(m_Workbench);

    x_RegisterUserTypes();
    x_RegisterCommands();

    // Connect to Menu Service
    IMenuService* menu_srv = m_Workbench->GetMenuService();
    menu_srv->AddContributor(this);

    // Connect to ToolBar Service
    IToolBarService* tb_srv = m_Workbench->GetToolBarService();
    wxFileArtProvider& provider = m_Workbench->GetFileArtProvider();

    /// Connect Window Manager Service
    IWindowManagerService* wm_srv = m_Workbench->GetWindowManagerService();
    wm_srv->SetAdvisor(this);

    // create and register a factory for standard toolbars
    CGBenchToolBarFactory* factory = new CGBenchToolBarFactory();
    factory->RegisterImageAliases(provider);
    tb_srv->AddToolBarFactory(factory);

    CProjectViewToolBatFactory* pr_factory = new CProjectViewToolBatFactory();
    tb_srv->AddToolBarFactory(pr_factory);

    x_RegisterPackageContributions();

    menu_srv->ResetMenuBar();
    tb_srv->AddToolBarContext(this);


    // setup D&D handling (does not currently work on mac - wxWidgets issue)
    wxFrame* frame = m_Workbench->GetMainWindow();
    frame->SetDropTarget(new CGBenchGUIDropTarget(*this));

    // Initialize OpenGL font library with directory and device resolution
    wxSize ppi = wxGetDisplayPPI();
    CFtglFontManager::Instance().SetDeviceResolution(ppi.y);
    CFtglFontManager::Instance().SetFontPath(ToStdString(CSysPath::GetResourcePath()));





    // add command handlers for the application commands (lower priority - first)
    CGBenchGUIMenuDisabler* disabler = new CGBenchGUIMenuDisabler();
    x_PushEvtHandler(disabler);

    x_PushEvtHandler(this);

    CProjectCmdHandler* prj_handler = new CProjectCmdHandler(m_Workbench);
    x_PushEvtHandler(prj_handler);

    CAppExplorerCmdHandler* exp_handler = new CAppExplorerCmdHandler(m_Workbench);
    x_PushEvtHandler(exp_handler);
}


void CGBenchGUI::PostViewManagerInitialized()
{
}


//static const char* kDefLayoutTag = "DefaultLayout";

void CGBenchGUI::PostInit()
{
    // Register application-specific services
    string name = typeid(CUIDataSourceService).name();
    m_Workbench->RegisterService(name, new CUIDataSourceService());

    name = typeid(CProjectService).name();
    CProjectService* prj_srv = new CProjectService();
    m_Workbench->RegisterService(name, prj_srv);

    name = typeid(CAppExplorerService).name();
    CAppExplorerService* exp_srv = new CAppExplorerService();
    m_Workbench->RegisterService(name, exp_srv);
    prj_srv->AttachObserver(exp_srv);

    name = typeid(CSelectionService).name();
    CSelectionService* sel_srv = new CSelectionService();
    m_Workbench->RegisterService(name, sel_srv);

    name = typeid(CDataMiningService).name();
    CDataMiningService* dm_srv = new CDataMiningService();
    m_Workbench->RegisterService(name, dm_srv);

    name = typeid(IVisibleRangeService).name();
    CVisibleRangeService* vr_srv = new CVisibleRangeService();
    m_Workbench->RegisterService(name, vr_srv);

    CWinMaskerFileStorage& storage = CWinMaskerFileStorage::GetInstance();
    storage.SetRegistryPath( m_RegPath + ".WMFStorage" );
    storage.LoadSettings();

    x_SetupStatusBar();
}


void CGBenchGUI::x_SetupStatusBar()
{
    // configure popup window for the standard message slot
    CMessageSlotPopupWindow::SetExtraLink("Show Event View", eCmdShowEventView);

    // add a slot for App Task Service
    IStatusBarService* sb_srv = m_Workbench->GetStatusBarService();

    m_AppSrvSlot = new CAppTaskServiceSlot(this);
    m_AppSrvSlot->Create(x_GetMainWindow());
    wxSize size = m_AppSrvSlot->GetSize();

    sb_srv->InsertSlot(1, m_AppSrvSlot, size.x);

    m_AppSrvSlot->Show();

    CAppTaskService* task_srv = m_Workbench->GetAppTaskService();
    task_srv->AttachObserver(m_AppSrvSlot->GetIObserver());

    m_BroadcastSlot = new CSelectionBroadcastSlot( this );
    m_BroadcastSlot->Create( x_GetMainWindow() );
    sb_srv->InsertSlot( 1, m_BroadcastSlot, m_BroadcastSlot->GetSize().x );
    m_BroadcastSlot->Show();

    CSelectionService* sel_srv = m_Workbench->GetServiceByType<CSelectionService>();
    sel_srv->AttachObserver( m_BroadcastSlot->GetIObserver() );
}


void CGBenchGUI::x_ClearStatusBar()
{
    CAppTaskService* task_srv = m_Workbench->GetAppTaskService();
    task_srv->DetachObserver(m_AppSrvSlot->GetIObserver());

    m_AppSrvSlot = NULL; //will be deleted by Status Bar

    CSelectionService* sel_srv = m_Workbench->GetServiceByType<CSelectionService>();
    sel_srv->DetachObserver( m_BroadcastSlot->GetIObserver() );
    m_BroadcastSlot = NULL;
}


void CGBenchGUI::PreShutDown()
{
    CAppJobDispatcher& disp = CAppJobDispatcher::GetInstance();
    disp.CancelAllJobs();

    if( !m_Workbench->HasPerspective() ){
        x_SaveDefaultWindowLayout();
    }

    x_CloseAllViews();

    x_GetMainWindow()->Hide();

    x_ClearStatusBar();

    // Remove fonts from OpenGL texture font singleton
    CFtglFontManager::Instance().Clear();

    // Clear objects from OpenGL resource singleton
    if (CGlResMgr::Initialized())
        CGlResMgr::Instance().Clear();

    // unregister application-specific services
    CAppExplorerService* exp_srv = m_Workbench->GetServiceByType<CAppExplorerService>();
    CProjectService* prj_srv = m_Workbench->GetServiceByType<CProjectService>();
    prj_srv->DetachObserver(exp_srv);

    CWinMaskerFileStorage& storage = CWinMaskerFileStorage::GetInstance();
    storage.SaveSettings();

    m_Workbench->UnRegisterService(typeid(IVisibleRangeService).name());
    m_Workbench->UnRegisterService(typeid(CSelectionService).name());
    m_Workbench->UnRegisterService(typeid(CAppExplorerService).name());
    // This is intentionally commented to revoke DL services later
    //
    // un-registration at this point creates a problem, that some running tasks 
    // still hold objects and scopes.    
    // m_Workbench->UnRegisterService(typeid(CUIDataSourceService).name());    
}


void CGBenchGUI::PreDestroyWindow()
{
    // disconnect from event handling, delete the handlers we created
    wxEvtHandler* handler = NULL;
    while( (handler = x_PopEvtHandler()) )  {
        if(handler != this) {
            delete handler;
        }
    }

    // remove self from the list of Tool Bar contexts
    IToolBarService* tb_srv = m_Workbench->GetToolBarService();
    tb_srv->RemoveToolBarContext(this);


    // unsubscribe from menus
    IMenuService* menu_srv = m_Workbench->GetMenuService();
    menu_srv->RemoveContributor(this);

    /// Disconnect Window Manager Service
    IWindowManagerService* wm_srv = m_Workbench->GetWindowManagerService();
    wm_srv->SetAdvisor(NULL);

    IMenuContributor* contributor = dynamic_cast<IMenuContributor*>(wm_srv);
    menu_srv->RemoveContributor(contributor);

    // toolbar factory will be deleted by ToolBar Service
}


bool CGBenchGUI::CanShutDown()
{
    return x_CloseWorkspace();
}


/// reads window layout from the file and applies it to Window Manager
void CGBenchGUI::RestoreWindowLayout()
{
    CConstRef<CUser_object> layout = x_LoadWindowLayout(GetVideoId());
    if( ! layout)   {
        layout = x_LoadWindowLayout("Default");
        if( ! layout)   {
            LOG_POST(Info << "Failed to load the default window layout - installation is corrupted.");
        }
    }
    if(layout) {
        IWindowManagerService* wm_srv = m_Workbench->GetWindowManagerService();

        // apply the new layout
        IViewManagerService* view_mrg = m_Workbench->GetViewManagerService();
        IWMClientFactory* factory = dynamic_cast<IWMClientFactory*>(view_mrg);

        _ASSERT(factory);

        wm_srv->LoadLayout(*layout, *factory);
    }

    // On some platforms, (I'm looking at you, Mac..) the initial window may
    // not display until a sizing or docking event, so we send an extra size
    // notification to force display.
    m_Workbench->GetMainWindow()->SendSizeEvent();
    m_Workbench->GetMainWindow()->Refresh();
}

void CGBenchGUI::SaveWindowLayoutToFile()
{
    wxFileDialog dlg(
        NULL, wxT("Select a file for window layout"), wxT(""), wxT(""),
        CFileExtensions::GetDialogFilter(CFileExtensions::kASN) + wxT("|") 
        + CFileExtensions::GetDialogFilter(CFileExtensions::kAllFiles),
        wxFD_SAVE|wxFD_OVERWRITE_PROMPT
    );
    if( dlg.ShowModal() != wxID_OK ) return;

    IWindowManagerService* wm_srv = m_Workbench->GetWindowManagerService();
    CRef<CUser_object> layout( wm_srv->SaveLayout() );

    try {
        CNcbiOfstream file_to_save( dlg.GetPath().fn_str() );
        if( !file_to_save ){
            wxMessageBox( "Error opening selected file: " + dlg.GetPath(), "File Error" );
            return;
        }

        file_to_save << MSerial_AsnText << *layout;
    } catch( CException& ex ){
        wxMessageBox( ToWxString( ex.GetMsg() ), "Error while saving window layout" );
        return;
    }
}

void CGBenchGUI::LoadWindowLayoutFromFile()
{
    wxFileDialog dlg(
        NULL, wxT("Select a file with window layout"), wxT(""), wxT(""),
        CFileExtensions::GetDialogFilter(CFileExtensions::kASN) + wxT("|") 
        + CFileExtensions::GetDialogFilter(CFileExtensions::kAllFiles),
        wxFD_OPEN
    );
    if( dlg.ShowModal() != wxID_OK ) return;

    CRef<CUser_object> layout( new CUser_object() );

    try {
        CNcbiIfstream file_to_load( dlg.GetPath().fn_str() );
        if( !file_to_load ){
            wxMessageBox( "Error opening selected file: " + dlg.GetPath(), "File Error" );
            return;
        }

        file_to_load >> MSerial_AsnText >> *layout;
    } catch( CException& ex ){
        wxMessageBox( ToWxString( ex.GetMsg() ), "Error while loading window layout" );
        return;
    }

    if( layout ){
        wxFrame* mainWnd = m_Workbench->GetMainWindow();
        if( mainWnd ){
            int y = 0;

#           ifdef __WXMAC__
            int mbar_height, maxy, maxx;
            GetMacDragWindowBounds(mainWnd, mbar_height, maxy, maxx);
            y = mbar_height;
#           endif

            wxRect rc( 0, y, 600, 500 );
            mainWnd->SetSize( rc );
            mainWnd->Maximize();
        }

        IWindowManagerService* wm_srv = m_Workbench->GetWindowManagerService();
        wm_srv->ApplyLayout( *layout );
    }
}

CConstRef<CUser_object> CGBenchGUI::x_LoadWindowLayout(const string& reg_key)
{
    CGuiRegistry& gui_reg = CGuiRegistry::GetInstance();
    CGuiRegistry::TReadView view = gui_reg.GetReadView("WindowLayouts");
    CConstRef<CUser_field> field = view.GetField(reg_key);

    if (field && field->CanGetData() && field->GetData().IsObject()) {
        CConstRef<CUser_object> layout(&field->GetData().GetObject());
        return layout;
    }

    LOG_POST(Error << "Failed to load window layout: " << reg_key);
    return CConstRef<CUser_object>();
}


// saves current Window Manager layout to the file
void CGBenchGUI::x_SaveDefaultWindowLayout()
{
    IWindowManagerService* wm_srv = m_Workbench->GetWindowManagerService();
    CRef<CUser_object> layout(wm_srv->SaveLayout());

    CGuiRegistry& gui_reg = CGuiRegistry::GetInstance();
    CGuiRegistry::TReadWriteView view = gui_reg.GetReadWriteView("WindowLayouts", CGuiRegistry::ePriority_Local - 1);
    CRef<CUser_field> field = view.SetField(GetVideoId());
    field->SetData().SetObject(*layout);
}


void CGBenchGUI::x_RegisterUserTypes()
{
    /////////////////////////////////////////////////////////////
    // register object symbols
    CLabel::RegisterTypeIcon(CGUIUserType::sm_Tp_Sequence, "", "symbol::sequence");
    CLabel::RegisterTypeIcon(CGUIUserType::sm_Tp_Sequence, CGUIUserType::sm_SbTp_DNA, "symbol::sequence_dna");
    CLabel::RegisterTypeIcon(CGUIUserType::sm_Tp_Sequence, CGUIUserType::sm_SbTp_Protein, "symbol::sequence_protein");

    CLabel::RegisterTypeIcon(CGUIUserType::sm_Tp_Sequence_ID, "", "symbol::sequence_id");
    CLabel::RegisterTypeIcon(CGUIUserType::sm_Tp_Sequence_ID, CGUIUserType::sm_SbTp_DNA, "symbol::sequence_id_dna");
    CLabel::RegisterTypeIcon(CGUIUserType::sm_Tp_Sequence_ID, CGUIUserType::sm_SbTp_Protein, "symbol::sequence_id_protein");

    CLabel::RegisterTypeIcon(CGUIUserType::sm_Tp_Sequence_Set, "", "symbol::sequence_set");
    CLabel::RegisterTypeIcon(CGUIUserType::sm_Tp_Alignment_Set, "", "symbol::alignment_set");

    CLabel::RegisterTypeIcon(CGUIUserType::sm_Tp_Annotation, "", "symbol::annotation");
    CLabel::RegisterTypeIcon(CGUIUserType::sm_Tp_Annotation, CGUIUserType::sm_SbTp_Alignments, "symbol::annotation_alignments");
    CLabel::RegisterTypeIcon(CGUIUserType::sm_Tp_Annotation, CGUIUserType::sm_SbTp_Features, "symbol::annotation_features");
    CLabel::RegisterTypeIcon(CGUIUserType::sm_Tp_Annotation, CGUIUserType::sm_SbTp_Graphs, "symbol::annotation_graphs");
    CLabel::RegisterTypeIcon(CGUIUserType::sm_Tp_Annotation, CGUIUserType::sm_SbTp_Sequence_IDs, "symbol::annotation_sequence_ids");
    CLabel::RegisterTypeIcon(CGUIUserType::sm_Tp_Annotation, CGUIUserType::sm_SbTp_Locations, "symbol::locations");

    CLabel::RegisterTypeIcon(CGUIUserType::sm_Tp_Alignment, "", "symbol::alignment");
    CLabel::RegisterTypeIcon(CGUIUserType::sm_Tp_Feature, "", "symbol::feature");

    wxFileArtProvider& provider = m_Workbench->GetFileArtProvider();

    provider.RegisterFileAlias(wxT("symbol::sequence"), wxT("sequence_symbol.png"));
    provider.RegisterFileAlias(wxT("symbol::sequence"), wxT("sequence_symbol.png"));
    provider.RegisterFileAlias(wxT("symbol::sequence_dna"), wxT("sequence_dna_symbol.png"));
    provider.RegisterFileAlias(wxT("symbol::sequence_protein"), wxT("sequence_protein_symbol.png"));

    provider.RegisterFileAlias(wxT("symbol::sequence_id"), wxT("sequence_id_symbol.png"));
    provider.RegisterFileAlias(wxT("symbol::sequence_id_dna"), wxT("sequence_id_dna_symbol.png"));
    provider.RegisterFileAlias(wxT("symbol::sequence_id_protein"), wxT("sequence_id_protein_symbol.png"));

    provider.RegisterFileAlias(wxT("symbol::sequence_set"),   wxT("sequence_set_symbol.png"));
    provider.RegisterFileAlias(wxT("symbol::alignment_set"),   wxT("alignment_set_symbol.png"));

    provider.RegisterFileAlias(wxT("symbol::alignment"),     wxT("alignment_symbol.png"));
    provider.RegisterFileAlias(wxT("symbol::feature"),     wxT("feature_symbol.png"));

    provider.RegisterFileAlias(wxT("symbol::annotation"), wxT("annot_symbol.png"));
    provider.RegisterFileAlias(wxT("symbol::annotation_alignments"), wxT("annot_aligns_symbol.png"));
    provider.RegisterFileAlias(wxT("symbol::annotation_graphs"), wxT("annot_graphs_symbol.png"));
    provider.RegisterFileAlias(wxT("symbol::annotation_features"), wxT("annot_feats_symbol.png"));
    provider.RegisterFileAlias(wxT("symbol::annotation_locs"), wxT("annot_locs_symbol.png"));
    provider.RegisterFileAlias(wxT("symbol::annotation_ids"),  wxT("annot_ids_symbol.png"));
    provider.RegisterFileAlias(wxT("symbol::locations"), wxT("annot_locs_symbol.png"));

}


void CGBenchGUI::x_RegisterCommands()
{
    CUICommandRegistry& registry = m_Workbench->GetUICommandRegistry();
    wxFileArtProvider& provider = m_Workbench->GetFileArtProvider();

    WidgetsWx_RegisterCommands(registry, provider);
    Core_RegisterCommands(registry, provider);
    Gbench_RegisterCommands(registry, provider);

    wxFrame* frame = m_Workbench->GetMainWindow();
    registry.ApplyAccelerators(frame);
}

void CGBenchGUI::x_RegisterPackageContributions()
{
    if( !m_GBenchGUIApp ) return;

    CPkgManager* pkgManager = m_GBenchGUIApp->GetPackageManager();
    if( !pkgManager ) return;

    CUICommandRegistry& cmd_reg = m_Workbench->GetUICommandRegistry();
    IMenuService* menu_srv = m_Workbench->GetMenuService();
    IToolBarService* tb_srv = m_Workbench->GetToolBarService();


    ITERATE( CPkgManager, it, *pkgManager ){
        const CPackageInfo& pkg_info = it->second;
        if( pkg_info.m_State != CPackageInfo::eLoaded ){
            ERR_POST( "The package " << it->first << " is not loaded properly." );

            continue;
        }

        IGuiPackage* pkg = pkg_info.m_Package.get();
        if( !pkg ) continue;

        // Should be called before GetEventHandler()
        IPackageWorkbenchConnector* pkg_wb = dynamic_cast<IPackageWorkbenchConnector*>( pkg );
        if (pkg_wb) {
            pkg_wb->SetWorkbench(m_Workbench);
        }


        IUICommandContributor* cmd_ctrb = dynamic_cast<IUICommandContributor*>( pkg );
        if( cmd_ctrb ){
            vector<CUICommand*> cmds = cmd_ctrb->GetCommands();
            NON_CONST_ITERATE( vector<CUICommand*>, pix, cmds ){
                cmd_reg.RegisterCommand( *pix );
            }
            x_PushEvtHandler( cmd_ctrb->GetEventHandler() );
        }


        // Register menu with CMenuService
        IMenuContributor* menu_ctrb = dynamic_cast<IMenuContributor*>( pkg );
        if( menu_ctrb ){
            menu_srv->AddContributor( menu_ctrb );
        }

        // Register menu with CToolbarService

        IToolBarContributor* tb_ctrb = dynamic_cast<IToolBarContributor*>( pkg );
        if( tb_ctrb ){
            tb_srv->AddToolBarFactory( tb_ctrb );
        }
    }
}

static
WX_DEFINE_MENU(kMainMenu)
    WX_SUBMENU("&File")
        WX_MENU_ITEM(eCmdNewWorkspace)
        WX_MENU_ITEM(wxID_OPEN)
        WX_MENU_ITEM(eCmdCloseWorkspace)
        WX_MENU_SEPARATOR()
        WX_MENU_ITEM(wxID_SAVE)
        WX_MENU_ITEM(wxID_SAVEAS)
        WX_MENU_ITEM(eCmdExportObject)
        WX_MENU_SEPARATOR()
        WX_MENU_ITEM(eCmdRemoveProject)
        WX_MENU_SEPARATOR()
        WX_MENU_ITEM(eCmdLoadProject)
        WX_MENU_ITEM(eCmdUnLoadProject)
        WX_MENU_SEPARATOR()
        WX_MENU_ITEM(wxID_PRINT)
        WX_MENU_ITEM(eCmdSaveImages) 
        WX_MENU_ITEM(eCmdSavePdf) 
        WX_MENU_SEPARATOR()
        WX_MENU_ITEM(wxID_EXIT)
    WX_END_SUBMENU()
    WX_SUBMENU("&Edit")
        WX_MENU_ITEM(wxID_UNDO)
        WX_MENU_ITEM(wxID_REDO)
        WX_MENU_SEPARATOR()
        WX_MENU_ITEM(wxID_CUT)
        WX_MENU_ITEM(wxID_COPY)
        WX_MENU_ITEM(wxID_PASTE)
        WX_MENU_ITEM(wxID_DELETE)
        //WX_MENU_SEPARATOR()
        //WX_MENU_ITEM(wxID_FIND)
        WX_MENU_SEPARATOR()
        WX_MENU_ITEM(wxID_SELECTALL)
        WX_MENU_ITEM(wxID_CLEAR)
        WX_MENU_SEPARATOR_L("Contribs")
        WX_MENU_SEPARATOR_L("Properties")
        WX_MENU_ITEM(wxID_PROPERTIES)
    WX_END_SUBMENU()
    WX_SUBMENU("&View")
        WX_MENU_ITEM(eCmdOpenView)
        WX_MENU_SEPARATOR()
        WX_MENU_ITEM(eCmdShowProjectTreeView)
        WX_MENU_ITEM(eCmdShowSearchView)
        WX_MENU_ITEM(eCmdShowSelectionInspector)
        WX_MENU_ITEM(eCmdShowTaskView)
        WX_MENU_ITEM(eCmdShowEventView)
        WX_MENU_ITEM(eCmdShowDiagConsole)
        WX_MENU_SEPARATOR_L("Contribs")
        WX_MENU_SEPARATOR_L("Settings")
        WX_MENU_ITEM(eCmdSettings)
    WX_END_SUBMENU()
    WX_SUBMENU("&Navigate")
        WX_MENU_ITEM(eCmdBack)
        WX_MENU_ITEM(eCmdForward)
        WX_MENU_SEPARATOR()
        WX_MENU_ITEM(eCmdZoomIn)
        WX_MENU_ITEM(eCmdZoomOut)
        WX_MENU_ITEM(eCmdZoomAll)
        WX_MENU_SEPARATOR()
        WX_MENU_ITEM(eCmdZoomSeq)
        WX_MENU_ITEM(eCmdZoomSel)
        WX_MENU_ITEM(eCmdZoomSelObjects)
        WX_MENU_ITEM(eCmdZoomObjects)
        WX_MENU_SEPARATOR()
        WX_SUBMENU("Zoom &Special")
            WX_MENU_ITEM(eCmdZoomInX)
            WX_MENU_ITEM(eCmdZoomOutX)
            WX_MENU_ITEM(eCmdZoomAllX)
            WX_MENU_SEPARATOR()
            WX_MENU_ITEM(eCmdZoomInY)
            WX_MENU_ITEM(eCmdZoomOutY)
            WX_MENU_ITEM(eCmdZoomAllY)
            WX_MENU_SEPARATOR()
            WX_MENU_ITEM(eCmdSetEqualScale)
        WX_END_SUBMENU()
    WX_END_SUBMENU()
    WX_SUBMENU("&Tools")
        WX_MENU_ITEM(eCmdRunTool)
        WX_MENU_ITEM(eCmdSearch)
        WX_MENU_SEPARATOR()
        WX_MENU_ITEM(eCmdBroadcastSel)
        WX_MENU_ITEM(eCmdBroadcastOptions)
		WX_MENU_ITEM(eCmdWindowMaskerFiles)
        WX_MENU_SEPARATOR()
        WX_MENU_ITEM(eCmdToolsPackages)
        WX_MENU_ITEM(eCmdToolsOptions)
    WX_END_SUBMENU()
    WX_MENU_SEPARATOR_L("Contribs")
    WX_MENU_SEPARATOR_L("EndContribs")
    WX_SUBMENU("&Window")
        WX_MENU_SEPARATOR_L("Actions")
        WX_MENU_SEPARATOR_L("Close")
        WX_MENU_ITEM(eCmdCloseAllProjectViews)
        WX_MENU_SEPARATOR_L("Actions2")
#ifdef _DEBUG
        WX_MENU_ITEM(eCmdLoadLayout)
        WX_MENU_ITEM(eCmdSaveLayout)
#endif
        WX_MENU_ITEM(eCmdResetLayout)
        WX_MENU_SEPARATOR_L("Windows")
    WX_END_SUBMENU()
#ifdef _DEBUG
    WX_SUBMENU("&Debug")
        WX_MENU_ITEM(eCmdTestPrjSrv)
        WX_MENU_ITEM(eCmdTestLoadBlastMask)
        WX_MENU_ITEM(eCmdMacroEditDlg)
    WX_END_SUBMENU()
#endif
    // Help has to be last, to work on Mac platform
    WX_SUBMENU("&Help")
        WX_MENU_ITEM(eCmdSendFeedback)
#if defined( __WXMAC__ )
        WX_MENU_ITEM(eCmdAbout)
#else
        WX_MENU_SEPARATOR()
        WX_MENU_ITEM(wxID_ABOUT)
#endif
    WX_END_SUBMENU()
WX_END_MENU()


wxMenu* CGBenchGUI::GetMenu()
{
    CUICommandRegistry& cmd_reg = CUICommandRegistry::GetInstance();
    wxMenu* menu = cmd_reg.CreateMenu(kMainMenu);
    x_UpdateQuickLaunchMenu(*menu);
    return menu;
}


void CGBenchGUI::GetCompatibleToolBars(vector<string>& names)
{
    names.push_back("Application Toolbar");
    names.push_back(CProjectViewToolBatFactory::sm_SelTBName);
}


void CGBenchGUI::x_ShowSingletonView(const string& view_ui_name)
{
    IViewManagerService* manager = m_Workbench->GetViewManagerService();
    manager->ShowSingletonView(view_ui_name);
}


void CGBenchGUI::x_AddSystemView(const string& view_ui_name)
{
    IViewManagerService* manager = m_Workbench->GetViewManagerService();
    CIRef<IView> view = manager->CreateViewInstance(view_ui_name);
    if(view)    {
        manager->AddToWorkbench(*view);
    }
}


wxWindow* CGBenchGUI::x_GetMainWindow()
{
    return m_Workbench->GetMainWindow();
}


void CGBenchGUI::OnFileOpen(wxCommandEvent& event)
{
    COpenDlgTask* task = new COpenDlgTask(m_Workbench);
    m_Workbench->GetAppTaskService()->AddTask(*task);
}


void CGBenchGUI::ShowOpenDlg(vector<wxString>& filenames)
{
    COpenDlgTask* task = new COpenDlgTask(m_Workbench, filenames);
    m_Workbench->GetAppTaskService()->AddTask(*task);
}


void CGBenchGUI::OnRunTool(wxCommandEvent& event)
{
    CRunToolDlgTask* task = new CRunToolDlgTask(m_Workbench);
    m_Workbench->GetAppTaskService()->AddTask(*task);
}


void CGBenchGUI::OnQuickLaunch(wxCommandEvent& event)
{
    CQuickLaunchList& quickLaunch = CQuickLaunchList::GetInstance();
    if (quickLaunch.IsEmpty())
        return;

    string entryLabel;
    const list<string>& items = quickLaunch.GetItems();
    int cmd = eCmdQuickLaunchStart;
    ITERATE(list<string>, it, items) {
        if (cmd == event.GetId()) {
            entryLabel = *it;
            break;
        }
        ++cmd;
    }

    if (entryLabel.empty())
        return;

    string entriesPath = quickLaunch.GetItemPath(entryLabel);
    string toolLabel = CGuiRegistry::GetInstance().GetString(entriesPath + ".Tool");
    if (toolLabel.empty()) {
        quickLaunch.RemoveItem(entryLabel);
        wxMessageBox(wxT("Invalid entry: ") + ToWxString(entryLabel),
                     wxT("Quick Launch Error"), wxOK | wxICON_ERROR);
        return;
    }


    vector< CIRef<IUIAlgoToolManager> > managers;
    GetExtensionAsInterface("ui_algo_tool_manager", managers);

    CIRef<IUIAlgoToolManager> manager;
    for( size_t i =0; i < managers.size(); i++ ){
        if (managers[i]->GetDescriptor().GetLabel() == toolLabel) {
            manager = managers[i];
            break;
        }
    }

    if (!manager || !manager->CanQuickLaunch()) {
        quickLaunch.RemoveItem(entryLabel);
        wxMessageBox(wxT("Tool not found: ") + ToWxString(toolLabel),
                     wxT("Quick Launch Error"), wxOK | wxICON_ERROR);
        return;
    }

    IRegSettings* rgs = dynamic_cast<IRegSettings*>(manager.GetPointer());
    if(rgs) {
        rgs->SetRegistryPath(entriesPath + ".Params");
        rgs->LoadSettings();
    }

    manager->SetServiceLocator(m_Workbench->GetServiceLocator());

    TConstScopedObjects objects;
    CIRef<CSelectionService> sel_srv = m_Workbench->GetServiceByType<CSelectionService>();
    sel_srv->GetCurrentSelection(objects);

    manager->SetInputObjects(objects);

    try {
        IAppTask* task = manager->QuickLaunch();

        if( task ){
            quickLaunch.AddItem( entryLabel );
            m_Workbench->GetAppTaskService()->AddTask( *task );
        }
    } catch( CException& ex ){
        ERR_POST( ex.ReportAll() );
        NcbiErrorBox( ex.GetMsg() );

    } catch( std::exception& ex ){
        ERR_POST( ex.what() );
        NcbiErrorBox( ex.what() );
    }
}

void CGBenchGUI::OnRecentTool(wxCommandEvent& event)
{
    CRecentToolList& recentTool = CRecentToolList::GetInstance();
    if (recentTool.IsEmpty())
        return;

    string toolLabel;
    const list<string>& items = recentTool.GetItems();
    int cmd = eCmdRecentToolStart;
    ITERATE(list<string>, it, items) {
        if (cmd == event.GetId()) {
            toolLabel = *it;
            break;
        }
        ++cmd;
    }

    if (toolLabel.empty())
        return;

    vector< CIRef<IUIAlgoToolManager> > managers;
    GetExtensionAsInterface("ui_algo_tool_manager", managers);

    CIRef<IUIAlgoToolManager> manager;
    for( size_t i =0; i < managers.size(); i++ ){
        if (managers[i]->GetDescriptor().GetLabel() == toolLabel) {
            manager = managers[i];
            break;
        }
    }

    if (!manager) {
        recentTool.RemoveItem(toolLabel);
        wxMessageBox(wxT("Tool not found: ") + ToWxString(toolLabel),
                     wxT("Recent Tool Error"), wxOK | wxICON_ERROR);
        return;
    }

    CRunToolDlgTask* task = new CRunToolDlgTask(m_Workbench, toolLabel);
    m_Workbench->GetAppTaskService()->AddTask(*task);
}

void CGBenchGUI::x_UpdateQuickLaunchMenu(wxMenu& root_menu)
{
    CQuickLaunchList& quickLaunch = CQuickLaunchList::GetInstance();

    wxMenuItem* tools_item = ncbi::FindSubItem(root_menu, wxT("&Tools"));
    wxMenu* tools_menu = tools_item ? tools_item->GetSubMenu() : NULL;
    if (!tools_menu)
        return;

    wxMenuItemList& items = tools_menu->GetMenuItems();
    size_t pos = 0;
    ITERATE( wxMenuItemList, it, items ){
        wxMenuItem* item = *it;
        if ((*it)->GetId() == eCmdRunTool) {
            quickLaunch.UpdateMenu(*tools_menu, eCmdQuickLaunchStart, eCmdQuickLaunchEnd, pos + 1);
            break;
        }
        ++pos;
    }

    CRecentToolList& recentTool = CRecentToolList::GetInstance();
    recentTool.UpdateMenu(*tools_menu, eCmdRecentToolStart, eCmdRecentToolEnd, pos + 2);
}

void CGBenchGUI::OnOpenView(wxCommandEvent& event)
{
    COpenViewlDlgTask* task = new COpenViewlDlgTask(m_Workbench);
    m_Workbench->GetAppTaskService()->AddTask(*task);
}

void CGBenchGUI::OnOpenViewDefault(wxCommandEvent& event)
{
    COpenViewlDlgTask* task = new COpenViewlDlgTask(m_Workbench, true);
    m_Workbench->GetAppTaskService()->AddTask(*task);
}

void CGBenchGUI::OnNewWorkspace(wxCommandEvent& event)
{
    CProjectTask* task = new CProjectTask(m_Workbench);
    task->Init_NewWorkspace();
    m_Workbench->GetAppTaskService()->AddTask(*task);
}


void CGBenchGUI::OnCloseWorkspace(wxCommandEvent& event)
{
    CProjectTask* task = new CProjectTask(m_Workbench);
    task->Init_CloseWorkspace();
    m_Workbench->GetAppTaskService()->AddTask(*task);
}


bool CGBenchGUI::x_HasWorkspace()
{
    CProjectService* prj_srv = m_Workbench->GetServiceByType<CProjectService>();
    return prj_srv  &&  prj_srv->HasWorkspace();
}


void CGBenchGUI::OnHasWorskpaceUpdate(wxUpdateUIEvent& event)
{
    event.Enable(x_HasWorkspace());
}


void CGBenchGUI::OnEnableCmdUpdate(wxUpdateUIEvent& event)
{
    event.Enable(true);
}

void CGBenchGUI::OnEnablePrintCmdUpdate(wxUpdateUIEvent& event)
{
    // Disable print by default.  Any widget that wants to support print can
    // add this same event handler and call event.Enable(true);
    event.Enable(false);
}

void CGBenchGUI::OnEnableSaveImagesCmdUpdate(wxUpdateUIEvent& event)
{
    // Disable save images by default.  Any widget that wants to support 
    // save images can add this same event handler and call event.Enable(true);
    event.Enable(false);
}

void CGBenchGUI::OnEnableSavePdfCmdUpdate(wxUpdateUIEvent& event)
{
    // Disable save images by default.  Any widget that wants to support 
    // save images can add this same event handler and call event.Enable(true);
    event.Enable(false);
}

/// default "disabler" for unhandled commands
void CGBenchGUI::OnDisableCmdUpdate(wxUpdateUIEvent& event)
{
    event.Enable(false);
}


void CGBenchGUI::OnShowDiagnosticsConsole(wxCommandEvent& event)
{
    x_ShowSingletonView("Diagnostic Console");
}


void CGBenchGUI::OnShowEventView(wxCommandEvent& event)
{
    x_ShowSingletonView("Event View");
}


void CGBenchGUI::OnShowTaskView(wxCommandEvent& event)
{
    x_ShowSingletonView("Task View");
}


void CGBenchGUI::OnShowProjectTreeView(wxCommandEvent& event)
{
    x_ShowSingletonView("Project Tree View");
}


void CGBenchGUI::OnShowSelectionInspector(wxCommandEvent& event)
{
    x_ShowSingletonView("Selection Inspector");
}


void CGBenchGUI::OnShowSearchView(wxCommandEvent& event)
{
    x_ShowSingletonView("Search View");
}


void CGBenchGUI::OnBroadcastOptions(wxCommandEvent& event)
{
    CIRef<CSelectionService> sel_srv = m_Workbench->GetServiceByType<CSelectionService>();
    _ASSERT( sel_srv );

    SBroadcastParams params;
    params.m_BioseqMatching = CSelectionEvent::sm_ObjMatchPolicy;
    params.m_SeqIDMatching = CSelectionEvent::sm_IdMatchPolicy;
    params.m_MatchAlnLocs = CSelectionEvent::sm_MatchAlnLocs;
    params.m_MatchByProduct = CSelectionEvent::sm_MatchByProduct;
    //params.m_AutoSelBroadcast = CSelectionEvent::sm_AutoBroadcast;
    params.m_AutoSelBroadcast = sel_srv->IsAutoBroadcast();
    params.m_BroadcastBetweenProjects = CSelectionEvent::sm_InterDocBroadcast;
    params.m_AutoVRBroadcast = CVisibleRange::IsAutoBroadcast();
    params.m_VRBPolicy = CVisibleRange::GetBasicVRPolicy(
        CVisibleRange::GetDefaultVisibleRangePolicy()
    );

    CBroadcastDialog dlg(x_GetMainWindow());
    dlg.SetData( params );

    if( dlg.ShowModal() == wxID_OK ){ // apply changes
        params = dlg.GetData();

        CSelectionEvent::sm_ObjMatchPolicy
            = (CSelectionEvent::EObjMatchPolicy) params.m_BioseqMatching
        ;
        CSelectionEvent::sm_IdMatchPolicy
            = (CSelectionEvent::EIdMatchPolicy) params.m_SeqIDMatching
        ;
        CSelectionEvent::sm_MatchAlnLocs = params.m_MatchAlnLocs;
        CSelectionEvent::sm_MatchByProduct = params.m_MatchByProduct;
        CSelectionEvent::sm_InterDocBroadcast = params.m_BroadcastBetweenProjects;
        //CSelectionEvent::sm_AutoBroadcast = params.m_AutoSelBroadcast;
        sel_srv->SetAutoBroadcast( params.m_AutoSelBroadcast );

        CVisibleRange::SetAutoBroadcast( params.m_AutoVRBroadcast );
        CVisibleRange::SetDefaultVisibleRangePolicy( 
            CVisibleRange::GetVisibleRangePolicy( 
                (CVisibleRange::EBasicPolicy)params.m_VRBPolicy 
            )
        );
    }
}

void CGBenchGUI::OnWindowMaskerFiles( wxCommandEvent& event )
{
	CWinMaskFilesParams params;
	const string ftp_url = "ftp://ftp.ncbi.nlm.nih.gov/toolbox/gbench/data/window-masker2/";

    {{
        wxBusyCursor wait;
        CDownloadJob::GetFTPDirContent( &params.m_Masks, ftp_url );
    }}

    CWinMaskerFileStorage& storage = CWinMaskerFileStorage::GetInstance();

    wxString saved_path = storage.GetPath();


    params.m_LocalPath = storage.GetLocalPath();
    params.m_PredefPath = storage.GetPredefinedPath();
    params.m_EnvPreferred = storage.GetEnvPredefined();
    params.m_HelpText = ftp_url;

	//! check files that are already loaded


    CWinMaskFilesDlg dlg( x_GetMainWindow() );
    dlg.SetParams( params );

    if( dlg.ShowModal() == wxID_OK ){
        params = dlg.GetParams();

        vector<string> files_to_load;

        ITERATE( vector<int>, ix_itr, params.m_ChosenIdxs ){
            const string& fname = params.m_Masks[*ix_itr];
            if( fname == "all.tar.gz" ){
                files_to_load.clear();
                files_to_load.push_back( fname );
                break;
            }
            else {
                files_to_load.push_back( fname );
            }
        } // ITERATE

        if( !files_to_load.empty() ){
                
            params.m_LocalPath.Trim(false).Trim(true);
            
			if( params.m_LocalPath.empty() ){
                wxMessageBox( wxT("Mask location path is empty"), wxT("Error"), wxOK | wxICON_ERROR );
                return;
            } else if (!params.m_LocalPath.IsAscii()) {
                wxMessageBox( wxT("Please select download directory which doesn't containt international characters"), wxT("Error"), wxOK | wxICON_ERROR );
                return;
            } else {
                try {
                    wxFileName fname(params.m_LocalPath, wxEmptyString);
                    if (!fname.IsAbsolute()) {
                        fname.MakeAbsolute();
                        params.m_LocalPath = fname.GetPath();
                    }
                    if (!fname.DirExists() && !::wxMkdir(fname.GetPath())) {
                        wxMessageBox( wxT("Failed to create mask directory"), wxT("Error"), wxOK | wxICON_ERROR );
                        return;
                    }
                } catch( CException& ex ){
                    ERR_POST( ex.ReportAll() );
                    wxMessageBox( wxT("Failed to create mask directory"), wxT("Error"), wxOK | wxICON_ERROR );
                    return; 
                }

                CRef<CDownloadJob> djob( new CDownloadJob( "Download", string(params.m_LocalPath.ToAscii()), ftp_url, files_to_load ) );
                CRef<CAppJobTask> task( new CAppJobTask( *djob, true, "Download " + ftp_url ) );

			    m_Workbench->GetAppTaskService()->AddTask( *task );
            }

        }
        
        wxString result;
        if( saved_path != params.m_LocalPath ){
            result = wxT("Reconfigured");
        }
        if( !files_to_load.empty() ){
            result = wxT("Download");
        }
        event.SetString( result );

        saved_path = params.m_LocalPath;

        storage.SetLocalPath( saved_path );
    }
}


static const char* kOptionsDlgSection = ".OptionsDialog";

void CGBenchGUI::OnToolsOptions(wxCommandEvent& event)
{
    CAppOptionsDlg dlg(NULL, CAppOptionsDlg::ID_CAPPOPTIONSDLG, wxT("Genome Workbench Options"));
    dlg.SetRegistryPath(m_RegPath + kOptionsDlgSection);
    dlg.ShowModal();
}

void CGBenchGUI::OnToolsPackages(wxCommandEvent& event)
{
    if (!m_GBenchGUIApp) {
        wxMessageBox(wxT("GbenchGUI application not initialized."), wxT("Error"), wxOK | wxICON_ERROR);
        return;
    }
    CPkgManager* pkgManager = m_GBenchGUIApp->GetPackageManager();
    if (!pkgManager) {
        wxMessageBox(wxT("Package Manager not initialized."), wxT("Error"), wxOK | wxICON_ERROR);
        return;
    }

    CAppPackagesDlg dlg(NULL, pkgManager);
    dlg.SetRegistryPath(m_RegPath + ".PackagesDialog");
    dlg.ShowModal();
}

void CGBenchGUI::OnHelpAbout(wxCommandEvent& event)
{
    CAboutDlg dlg(NULL, -1);
    dlg.ShowModal();
}

void CGBenchGUI::OnSendFeedback(wxCommandEvent& event)
{
    ShowFeedbackDialog(false);
}


void CGBenchGUI::OnCloseAllProjectViews(wxCommandEvent& event)
{
    CProjectService* prj_srv = m_Workbench->GetServiceByType<CProjectService>();
    prj_srv->RemoveAllProjectViews();
}


void CGBenchGUI::x_CloseAllViews()
{
    // get all IWMClients currently opened and request closing
    IWindowManagerService* wm_srv = m_Workbench->GetWindowManagerService();
    vector<IWMClient*> clients;
    wm_srv->GetAllClients(clients);

    OnCloseClientsRequest(clients);
}

bool CGBenchGUI::x_CloseWorkspace()
{
    CProjectTask task( m_Workbench );
    task.Init_CloseWorkspace();
    IAppTask::ETaskState state = task.Run();

    return state == IAppTask::eCompleted;
}


// handles a signal from Window Manager indicating that a client needs to be closed
void CGBenchGUI::OnCloseClientsRequest(const TClients& clients)
{
    /// close clients unconditionally
    IViewManagerService* manager = m_Workbench->GetViewManagerService();
    for( size_t i = 0;  i < clients.size();  i++ )    {
        IWMClient* client = clients[i];
        // use CIRef to postpone destruction
        CIRef<IView> view(dynamic_cast<IView*>(client));
        if(view)    {
            IProjectView* prj_view = dynamic_cast<IProjectView*>(&*view);

            if(prj_view)    {
                // Project Service handles Project Views
                CProjectService* prj_srv = m_Workbench->GetServiceByType<CProjectService>();
                prj_srv->RemoveProjectView(*prj_view);
            } else {
                // System View are handled by View Manager Service
                manager->RemoveFromWorkbench(*view);
            }
        } else {
            _ASSERT(false); // unexpected
        }
    }
}


void CGBenchGUI::OnClientAboutToClose(IWMClient& client)
{
    //LOG_POST("CGBenchGUI::OnClientAboutToClose()");
}


void CGBenchGUI::OnClientClosed(IWMClient& client)
{
    //LOG_POST("CGBenchGUI::OnClientClosed()");

#ifdef __WXOSX_COCOA__
    // Cocoa has a problem that if a view is closed and there is only 1
    // opengl view remaining, it sometimes goes blank and stays that way even
    // after redrawing/swapping/refreshing etc (GB-1898).  The only way I
    // found to fix this is to do a context switch to another view which
    // seems to somehow re-enable the drawing ability.  
    // This code creates a minimal floating OpenGL window and hides it behind
    // the mainwindow so you never actually see it.  But when a window closes,
    // we switch to its context and any remaining gl views that then grab the 
    // context will work.
    wxFrame* f = m_Workbench->GetMainWindow();
    wxPoint p;
    p = f->GetPosition();

    static GlResetDlg* dlg = NULL;

    if (dlg == NULL) {
        dlg = new GlResetDlg(f, p);
    }
     
    ncbi::AddChildWindowBelow(f, dlg); // keeps window invisibile
    dlg->SetPosition(f->GetPosition()); // match upper-left corner of main view

    dlg->Show(true); // user will not see -behind main window always.
    dlg->SwitchContext();
    f->Refresh();
    
    // hide window when we are not using it
    ncbi::RemoveChildWindow(f, dlg); 
    dlg->Hide(); 
#endif
}


string CGBenchGUI::GetFloatingFrameTitle()
{
    return GetAppTitle();
}


void CGBenchGUI::OnActiveClientChanged(IWMClient* new_active)
{
    //_TRACE("CGBenchGUI::OnActiveClientChanged() " << new_active);

    CIRef<CSelectionService> sel_srv = m_Workbench->GetServiceByType<CSelectionService>();

    if(sel_srv) {
        ISelectionClient* sel_client = dynamic_cast<ISelectionClient*>(new_active);
        sel_srv->OnActiveClientChanged(sel_client);
    }

    // update data mining service
    CIRef<CDataMiningService>  dm_srv =  m_Workbench->GetServiceByType<CDataMiningService>();
    if(dm_srv) {
        IDataMiningContext* sel_client = dynamic_cast<IDataMiningContext*>(new_active);
        dm_srv->OnActiveContextChanged(sel_client);
    }
}


void CGBenchGUI::ActivateClient(IWMClient* client)
{
    IWindowManagerService* wm_srv = m_Workbench->GetWindowManagerService();
    wm_srv->ActivateClient(*client);
}


void CGBenchGUI::OnSaveLayout(wxCommandEvent& event)
{
/*
    IWindowManagerService* wm_srv = m_Workbench->GetWindowManagerService();
    m_Layout.Reset(wm_srv->SaveLayout());
*/
    SaveWindowLayoutToFile();
}


void CGBenchGUI::OnSaveLayoutUpdate( wxUpdateUIEvent& event )
{
}


void CGBenchGUI::OnLoadLayout(wxCommandEvent& event)
{
    IWindowManagerService* wm_srv = m_Workbench->GetWindowManagerService();
    wm_srv->CloseAllClients();

    LoadWindowLayoutFromFile();

/*
    IWindowManagerService* wm_srv = m_Workbench->GetWindowManagerService();
    IViewManagerService* view_mrg = m_Workbench->GetViewManagerService();
    IWMClientFactory* factory = dynamic_cast<IWMClientFactory*>(view_mrg);

    _ASSERT(wm_srv  &&  factory);

    wm_srv->CloseAllClients();
    wm_srv->LoadLayout(*m_Layout, *factory);
*/
}


void CGBenchGUI::OnLoadLayoutUpdate( wxUpdateUIEvent& event )
{
    //event.Enable(m_Layout.NotEmpty());
}


void CGBenchGUI::OnResetLayout(wxCommandEvent& event)
{
    CGuiRegistry& gui_reg = CGuiRegistry::GetInstance();
    gui_reg.DeleteField(CGuiRegistry::ePriority_Local - 1, "WindowPos." + GetVideoId());
    gui_reg.DeleteField(CGuiRegistry::ePriority_Local - 1, "WindowLayouts." + GetVideoId());

    CConstRef<CUser_object> layout = x_LoadWindowLayout("Default");
    if(layout)  {
        /*{{
            CGuiRegistry& gui_reg = CGuiRegistry::GetInstance();
            CGuiRegistry::TReadWriteView view = gui_reg.GetReadWriteView("WindowLayouts", CGuiRegistry::ePriority_Local - 2);
            CRef<CUser_field> field = view.SetField("Default");
            field->SetData().SetObject(const_cast<CUser_object&>(*layout));
            wxString path(wxT("C:\\TEMP\\def_window_layout.asn"));
            CNcbiOfstream ostr( path.fn_str() );
            CGuiRegistry::GetInstance().Write( ostr , CGuiRegistry::ePriority_Local - 2 );
        }}*/

        wxFrame* mainWnd = m_Workbench->GetMainWindow();
        if (mainWnd) {
            int y = 0;
#ifdef __WXMAC__
            int mbar_height, maxy, maxx;
            GetMacDragWindowBounds(mainWnd, mbar_height, maxy, maxx);
            y = mbar_height;
#endif
            wxRect rc(0, y, 600, 500);
            mainWnd->SetSize(rc);
            mainWnd->Maximize();
        }

        IWindowManagerService* wm_srv = m_Workbench->GetWindowManagerService();
        wm_srv->ApplyLayout(*layout);
    }
}


void CGBenchGUI::OnResetLayoutUpdate(wxUpdateUIEvent& event)
{
}


void CGBenchGUI::OnTestPrjSrv(wxCommandEvent& event)
{
    CProjectServiceTestTask* task = new CProjectServiceTestTask(m_Workbench);
    m_Workbench->GetAppTaskService()->AddTask(*task);
}

void CGBenchGUI::OnMacroEditor(wxCommandEvent& event)
{
    CEditMacroDlg dlg(NULL);
    dlg.ShowModal();
}

void CGBenchGUI::OnTestLoadBlastMask(wxCommandEvent& event)
{
// experimental code
/*
    vector<string> files;
    CDownloadJob::GetFTPDirContent(&files, "ftp://ftp.ncbi.nlm.nih.gov/toolbox/gbench/ver-2.1.2");

    string url = "ftp://ftp.ncbi.nlm.nih.gov/toolbox/gbench/ver-2.1.2/gbench-src-2.1.2.tgz";
    CRef<CDownloadJob> djob(new CDownloadJob("Download", "c:\\test", url));
    CRef<CAppJobTask> task(new CAppJobTask(*djob, true, "Download " + url));

    IServiceLocator* srv_loc = m_Workbench->GetServiceLocator();
    CAppTaskService* task_srv = srv_loc->GetServiceByType<CAppTaskService>();

    task_srv->AddTask(*task);
*/  
}

///////////////////////////////////////////////////////////////////////////////
/// CGBenchToolBarFactory

void CGBenchToolBarFactory::RegisterImageAliases(wxFileArtProvider& provider)
{
    // we show only standard commands, icons for them shall be registered already
}


static const string kTBApplication = "Application Toolbar";

void CGBenchToolBarFactory::GetToolBarNames(vector<string>& names)
{
    names.push_back(kTBApplication);
}


wxAuiToolBar* CGBenchToolBarFactory::CreateToolBar(const string& name, wxWindow* parent)
{
    if(name == kTBApplication) {

        long style = wxAUI_TB_HORIZONTAL | wxBORDER_NONE;

        wxAuiToolBar* toolbar = new wxAuiToolBar(parent, wxID_ANY, wxDefaultPosition,
                                           wxDefaultSize, style);
                                           
        toolbar->SetName(ToWxString(kTBApplication));

        toolbar->SetToolBitmapSize(wxSize(16, 16));

        CUICommandRegistry& cmd_reg = CUICommandRegistry::GetInstance();
        cmd_reg.AppendTool(*toolbar, wxID_OPEN);
        cmd_reg.AppendTool(*toolbar, wxID_SAVE);
        toolbar->AddSeparator();
        cmd_reg.AppendTool(*toolbar, eCmdRunTool);
        cmd_reg.AppendTool(*toolbar, eCmdShowSearchView);

        toolbar->Realize();
        return toolbar;
    }
    return NULL;
}

END_NCBI_SCOPE
