/*  $Id: gbench_feedback_agent.cpp 26584 2012-10-11 15:22:39Z katargir $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Mike DiCuccio
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>

#include <corelib/request_ctx.hpp>

#include <gui/widgets/wx/wx_app.hpp>
#include "../gbench/feedback.hpp"
#include "../gbench/gbench_version.hpp"

#include <gui/widgets/wx/sys_path.hpp>
#include <gui/objutils/registry.hpp>
#include <gui/widgets/wx/wx_utils.hpp>

#include <wx/apptrait.h>

BEGIN_NCBI_SCOPE

class CGBenchFeedbackAgentApp: public CNCBIwxApplication
{
public:
    CGBenchFeedbackAgentApp(const string& app_name);

	/// @name CNcbiApplication overridables
	/// @{
	virtual void Init();
	/// @}

protected:
    bool LoadConfig(CNcbiRegistry& reg,
                    const string* conf, CNcbiRegistry::TFlags flags);

    virtual void x_SetupArgDescriptions();
};

CGBenchFeedbackAgentApp::CGBenchFeedbackAgentApp(const string& app_name)
    : CNCBIwxApplication(app_name)
{
}

void CGBenchFeedbackAgentApp::Init()
{
	string session_id;

	CNcbiRegistry& reg = GetConfig();
	if( reg.HasEntry( "GBENCH_CONN", "NCBI_SID" ) ){
		session_id = reg.Get( "GBENCH_CONN", "NCBI_SID" );
	} else {
		session_id = "GBENCH_FEEDBACK_UNKNOWN_SID";
	}
	GetDiagContext().GetRequestContext().SetSessionID( session_id );
	GetDiagContext().SetDefaultSessionID( session_id );

	CNCBIwxApplication::Init();
}

void CGBenchFeedbackAgentApp::x_SetupArgDescriptions()
{
    // Create command-line argument descriptions class
    auto_ptr<CArgDescriptions> arg_desc(new CArgDescriptions);

    // Specify USAGE context
    arg_desc->SetUsageContext(GetArguments().GetProgramBasename(),
                              "Feedback agent for Genome Workbench");

    arg_desc->AddOptionalPositional("pid", "PID",
                                    CArgDescriptions::eInteger);

    // Setup arg.descriptions for this application
    SetupArgDescriptions(arg_desc.release());
}


bool CGBenchFeedbackAgentApp::LoadConfig(CNcbiRegistry& reg,
                                         const string* conf,
                                         CNcbiRegistry::TFlags flags)
{
    /// load the standard NCBI .ini file registry
    string config_path;
    if (conf == NULL  ||  conf->empty()) {
        // TODO this won't work with intl fnames
        config_path =
            ToStdString(CSysPath::ResolvePathExisting(wxT("<home>/gbench.ini, ")
                                                      wxT("<std>/etc/gbench.ini")));
        if (!config_path.empty()) {
            conf = &config_path;
        }
    }

    // Let the base class load the config file.
    // If we couldn't find it above, it will try and find it, and warn otherwise.
    LOG_POST(Info << "Loading NCBI registry file from " << *conf);
    bool ret_val = CNcbiApplication::LoadConfig(reg, conf, flags);

    CGuiRegistry& gui_reg = CGuiRegistry::GetInstance();

    /// set the global repository
    {{
        wxString path = CSysPath::ResolvePath(wxT("<std>/etc/gbench.asn"));
        CNcbiIfstream istr(path.fn_str());
        if (istr) {
            gui_reg.SetGlobal(istr);
        }
    }}

    /// set the local repository
    {{
        wxString path = CSysPath::ResolvePath(wxT("<home>/gbench.asn"));
        CNcbiIfstream istr(path.fn_str());
        if (istr) {
            gui_reg.SetLocal(istr);
        }
    }}

    return ret_val;
}

class CGBenchFeedbackAgent: public CwxNCBIApp
{
public:
    virtual wxAppTraits *CreateTraits();
protected:
    // we are overriding this function to create our main frame window
    virtual bool    x_CreateGUI();
    virtual void    x_LoadGuiRegistry() {}
    virtual void    x_SaveGuiRegistry() {}
};

///////////////////////////////////////////////////////////////////////////////
/// CGBenchApp

// Declare wxWidgets application class
// this will allow CNCBIwxApplication to instantiate our class
DECLARE_APP(ncbi::CGBenchFeedbackAgent)
IMPLEMENT_APP_NO_MAIN(ncbi::CGBenchFeedbackAgent)
IMPLEMENT_WX_THEME_SUPPORT

class CGBenchFeedbackAgentTraits : public wxGUIAppTraits
{
#if wxUSE_INTL
    // called during wxApp initialization to set the locale to correspond to
    // the user default (i.e. system locale under Windows, LC_ALL under Unix)
    virtual void SetLocale() {}
#endif // wxUSE_INTL
};

wxAppTraits *CGBenchFeedbackAgent::CreateTraits()
{
    return new CGBenchFeedbackAgentTraits();
}

bool CGBenchFeedbackAgent::x_CreateGUI()
{
    try {
        int pid = 0;
        CNcbiApplication* app = CNcbiApplication::Instance();
        if (app) {
            CArgs args = app->GetArgs();
            if (args["pid"].HasValue()) {
                pid = args["pid"].AsInteger();
            }
        }

        //ShowFeedbackDialog( (pid != 0) );
        ShowFeedbackDialog( false ); // User cancel opt out
                                     // and we have problems with users accidentally checked opt out: GB-2335

        /// save gui registry as ASN.1
        wxString path = CSysPath::ResolvePath(wxT("<home>/gbench.asn"));
        CNcbiOfstream ostr(path.fn_str());
        CGuiRegistry::GetInstance().Write(ostr);
    }
    catch (std::exception& e) {
        LOG_POST(Error << "error in monitor: " << e.what());
    }
    catch (...) {
        LOG_POST(Error << "unknown error in monitor");
    }

    return false;
}

END_NCBI_SCOPE


/////////////////////////////////////////////////////////////////////////////
//  MAIN

int main(int argc, char** argv)
{
    wxDISABLE_DEBUG_SUPPORT();

    // create standard NCBI wxWidgets application
    ncbi::CGBenchFeedbackAgentApp ncbi_app("NCBI Genome Workbench Feedback Agent");

    // Execute main application function
    return ncbi_app.AppMain(argc, argv, 0, ncbi::eDS_Default, 0);
}
