/* $Id: theme_set.cpp 14565 2007-05-18 12:32:01Z dicuccio $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Author:  Robert G. Smith
 *
 * File Description:
 *   CSettingsSet is an adapter of the CGuiRegistry.  It is meant to be
 *   used as a base class for individual plugins' configuration classes.
 *
 */

#include <ncbi_pch.hpp>
#include <gui/config/theme_set.hpp>

BEGIN_NCBI_SCOPE

// the part of the hierarchy we keep the other styles we are managing
// as opposed to other settings we may also have (like any other CSettingsSet).
const string kSetsKey("sets");


CThemeSet::CThemeSet(const string& key,
                 CGuiRegistry* registry,
                 const string& delim)
      : CSettingsSet( key, registry, delim)
{
    // initialize our collection of CSettingsSet-s.
    vector <string> types = GetTypes();
    if (types.empty()) {
        NCBI_THROW(CException, eUnknown, key + " not a theme set. No subkeys under '" + kSetsKey + "'");
    }
    ITERATE(vector<string>, type_it, types) {
        CSettingsSet set(*type_it, registry, delim);
        // sync the current setting.
        string style_key = x_GetSetsCurrentStyleKey(*type_it);
        set.SetCurrentStyleKey(style_key);
        x_AddSettingsSet(*type_it, set);
    }
}


CThemeSet::~CThemeSet()
{
}


void CThemeSet::Uncache()
{
    NON_CONST_ITERATE (TSettingsReg, iter, m_Settings) {
        iter->second.Uncache();
    }
}


vector<string> CThemeSet::GetTypes() const
{
    vector<string>  my_types;
    CGuiRegistry::TKeys keys;
    CGuiRegistry::CReadView subview = x_GetReadView(kSetsKey);
    subview.GetTopKeys(keys);
    ITERATE(CGuiRegistry::TKeys, key_it, keys) {
        if (key_it->type == objects::CUser_field::TData::e_Str) {
            my_types.push_back(key_it->key);
        }
    }
    return my_types;
}


vector<string> CThemeSet::GetTypeDescriptions() const
{
    vector<string>  my_descs;
    vector <string> types = GetTypes();
    ITERATE(vector<string>, type_it, types) {
        const string& this_desc = GetSettingsSet(*type_it).GetTypeDescription() ;
        my_descs.push_back(this_desc);
    }
    return my_descs;
}


list<string> CThemeSet::GetSetsStyles(const string& type) const
{
    return GetSettingsSet(type).GetStyleNames();
}



string CThemeSet::GetSetsCurrentStyleName(const string& type) const
{
    string current_key = x_GetSetsCurrentStyleKey(type);
    return GetSettingsSet(type).x_GetStyleName(current_key);
    // return GetSettingsSet(type).GetCurrentStyleName(); // must keep these two in sync.
}

void CThemeSet::SetSetsCurrentStyleName(const string& type, const string& new_style_name)
{
    string new_key = GetSettingsSet(type).x_GetStyleKey(new_style_name);
    string current_key = x_GetSetsCurrentStyleKey(type);
    x_SetSetsCurrentStyleKey(type, new_key);
    SetSettingsSet(type).SetCurrentStyleKey(new_key);
}

string CThemeSet::x_GetSetsCurrentStyleKey(const string& type) const
{
    string style_key(GetString(kSetsKey + kDefaultDelim + type));
    return style_key;
}


void CThemeSet::x_SetSetsCurrentStyleKey(const string& type, const string& new_key)
{
    Set(kSetsKey + kDefaultDelim + type, new_key);
}


void CThemeSet::SetMyCurrentStyleName(const string& type)
{
    string style = GetSettingsSet(type).GetCurrentStyleKey();
    x_SetSetsCurrentStyleKey(type, style);
}

string CThemeSet::GetSetsDescription(const string& type) const
{
    return GetSettingsSet(type).GetTypeDescription();
}

const CSettingsSet& CThemeSet::GetSettingsSet(const string& type) const
{
    TSettingsReg::const_iterator set_it = m_Settings.find(type);
    if (set_it == m_Settings.end()) {
        NCBI_THROW(CGuiConfigException, eThemeUnknownSet, "This theme doesn't know anything about '" + type + "'");
    }
    return set_it->second;
}


CSettingsSet& CThemeSet::SetSettingsSet(const string& type)
{
    TSettingsReg::iterator set_it = m_Settings.find(type);
    if (set_it == m_Settings.end()) {
        NCBI_THROW(CGuiConfigException, eThemeUnknownSet, "Don't know anything about '" + type + "'");
    }
    return set_it->second;
}


void CThemeSet::x_AddSettingsSet(const string& type, CSettingsSet& set)
{
    m_Settings.insert(make_pair(type, set));
}



void CThemeSet::DumpAll(CNcbiOstream& ostr)
{
    CSettingsSet::DumpAll(ostr);

    ostr << endl << "Managed sets:" << endl;
    vector <string> types = GetTypes();
    ITERATE(vector<string>, type_it, types) {
        CSettingsSet set(GetSettingsSet(*type_it));
        ostr << "--- key: " << set.GetType() << " name: " << set.GetTypeDescription() << endl;
        set.DumpAll(ostr);
    }
    ostr << "-- END --" << endl;
}

END_NCBI_SCOPE
