/*  $Id: app_explorer_cmd_handler.cpp 27086 2012-12-17 19:43:41Z katargir $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Andrey Yazhuk
 *
 */

#include <ncbi_pch.hpp>

#include <gui/core/app_explorer_cmd_handler.hpp>

#include <gui/core/commands.hpp>
#include <gui/core/app_explorer_service.hpp>
#include <gui/core/project_service.hpp>

#include <gui/framework/app_task_service.hpp>
#include <gui/framework/window_manager_service.hpp>

#include <gui/widgets/wx/wm_client.hpp>
#include <gui/widgets/wx/ui_command.hpp>

#include <gui/utils/extension_impl.hpp>


#include <wx/menu.h>
#include <wx/msgdlg.h>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

/// Register CAppExplorerItemCmdContributor as an Extension
static CExtensionDeclaration
    decl("project_tree_view::context_menu::item_cmd_contributor",
         new CAppExplorerItemCmdContributor());

///////////////////////////////////////////////////////////////////////////////
/// CAppExplorerItemCmdContributor

static const int kClipboardTypesMask =
    CAppExplorerService::eWorkspaceFolder |
    CAppExplorerService::eProjectFolder |
    CAppExplorerService::eProjectItem |
    CAppExplorerService::eDataLoader;

static const int kSaveTypesMask =
    CAppExplorerService::eWorkspace |
    CAppExplorerService::eProject;

static const int kPropertyTypesMask =
    CAppExplorerService::eWorkspaceFolder |
    CAppExplorerService::eProjectFolder |
    CAppExplorerService::eWorkspace |
    CAppExplorerService::eProject |
    CAppExplorerService::eProjectItem |
    CAppExplorerService::eDataLoader |
    CAppExplorerService::eDataSource;


bool sAtLeastOneProjectLoaded(vector< CRef<CExplorerItem> >& items,
                              CAppExplorerService& app_service)
{
    return app_service.CanLoadUnloadProjects(items, false);
}


IExplorerItemCmdContributor::TContribution
    CAppExplorerItemCmdContributor::GetMenu(TItemRefVector& items,
                                            CAppExplorerService& app_service)
{
    CUICommandRegistry& cmd_reg = CUICommandRegistry::GetInstance();

    int types = app_service.GetItemTypes(items);

    wxMenu* menu = new wxMenu;

    // Save and Save As commands for Workspace and Projects
    if( (types & kSaveTypesMask)  &&  (types & ~kSaveTypesMask) == 0)    {
        bool ws = types & CAppExplorerService::eWorkspace;
        if(ws  || sAtLeastOneProjectLoaded(items, app_service))    {
            menu->Append(wxID_SEPARATOR, wxT("Top Actions"));
            cmd_reg.AppendMenuItem(*menu, wxID_SAVE);
            cmd_reg.AppendMenuItem(*menu, wxID_SAVEAS);
        }
    }

    // View commands
    if(types  ==  CAppExplorerService::eView)    {
        menu->Append(wxID_SEPARATOR, wxT("Actions"));
        cmd_reg.AppendMenuItem( *menu, eCmdActivateClients );
        cmd_reg.AppendMenuItem( *menu, eCmdCloseProjectView );
        cmd_reg.AppendMenuItem( *menu, eCmdCloseAllProjectViews );

    } else if( 
        items.size() == 1 
        && app_service.IsFolder( 
            *items[0], CAppExplorerService::sm_ALL_VIEWS_FOLDER 
        )
    ){
        menu->Append(wxID_SEPARATOR, wxT("Actions"));
        cmd_reg.AppendMenuItem(*menu, eCmdCloseAllProjectViews);
    }

    // Project Item commands
    if( types == CAppExplorerService::eProjectItem ){

        menu->Append(wxID_SEPARATOR, wxT("Top Actions"));
        cmd_reg.AppendMenuItem(*menu, eCmdOpenView);
        menu->Append(wxID_SEPARATOR, wxT("Actions"));
        cmd_reg.AppendMenuItem(*menu, eCmdEnableDisable);
    }

    if( types == CAppExplorerService::eDataLoader) {
        menu->Append(wxID_SEPARATOR, wxT("Actions"));
        cmd_reg.AppendMenuItem(*menu, eCmdEnableDisable);
    }

    // Project commands
    if( types == CAppExplorerService::eProject ){
        menu->Append(wxID_SEPARATOR, wxT("Actions"));
        cmd_reg.AppendMenuItem(*menu, eCmdLoadProject);
        cmd_reg.AppendMenuItem(*menu, eCmdUnLoadProject);
        cmd_reg.AppendMenuItem(*menu, eCmdProjectTableView);
    }

    // Edit commands
    TCmdID  clp_cmds[4] = { wxID_CUT, wxID_COPY, wxID_PASTE, wxID_DELETE };

    if((types & kClipboardTypesMask) == types)  {
        menu->Append(wxID_SEPARATOR, wxT("Edit"));

        bool is_data = false;
        ITERATE( CAppExplorerService::TItemRefVector, itm, items ){
            CAppExplorerService::TItemRef item = *itm;
            if( item->GetLabel() == CAppExplorerService::sm_DATA_FOLDER ){
                is_data = true;
                break;
            }
        }
        if( is_data ){
            cmd_reg.AppendMenuItem( *menu, wxID_PASTE );

        } else {
            cmd_reg.AppendMenuItems(*menu, clp_cmds, sizeof(clp_cmds) / sizeof(TCmdID));
        }

        if(
            types == CAppExplorerService::eWorkspaceFolder  
            || types == CAppExplorerService::eProjectFolder
        ){
            cmd_reg.AppendMenuItem( *menu, eCmdNewFolder );
        }
    }

    if( types == CAppExplorerService::eProject ){
        menu->Append(wxID_SEPARATOR, wxT("Edit"));
        cmd_reg.AppendMenuItem( *menu, eCmdRemoveProject );

    } else if( types == CAppExplorerService::eWorkspace ){
        menu->Append(wxID_SEPARATOR, wxT("Edit"));
        cmd_reg.AppendMenuItem( *menu, eCmdNewFolder );
    }

    // Properties
    if( (types & kPropertyTypesMask) == types ){
        menu->Append( wxID_SEPARATOR, wxT("Properties") );
        cmd_reg.AppendMenuItem( *menu, wxID_PROPERTIES );
    }

    TContribution contrib( menu, (wxEvtHandler*)NULL );
    return contrib;
}


string CAppExplorerItemCmdContributor::GetExtensionIdentifier() const
{
    return "app_explorer::item_cmd_contributor";
}


string CAppExplorerItemCmdContributor::GetExtensionLabel() const
{
    return "Contributes commands supported by App Explorer Service and applicable to Explorer items";
}


///////////////////////////////////////////////////////////////////////////////
/// CAppExplorerCmdHandler

BEGIN_EVENT_TABLE(CAppExplorerCmdHandler, wxEvtHandler)
    EVT_MENU(wxID_CUT, CAppExplorerCmdHandler::OnCut)
    EVT_UPDATE_UI(wxID_CUT, CAppExplorerCmdHandler::OnUpdateCut)
    EVT_MENU(wxID_COPY, CAppExplorerCmdHandler::OnCopy)
    EVT_UPDATE_UI(wxID_COPY, CAppExplorerCmdHandler::OnUpdateCopy)
    EVT_MENU(wxID_PASTE, CAppExplorerCmdHandler::OnPaste)
    EVT_UPDATE_UI(wxID_PASTE, CAppExplorerCmdHandler::OnUpdatePaste)

    EVT_MENU(wxID_DELETE, CAppExplorerCmdHandler::OnRemove)
    EVT_UPDATE_UI(wxID_DELETE, CAppExplorerCmdHandler::OnUpdateRemove)

    EVT_MENU(eCmdEnableDisable, CAppExplorerCmdHandler::OnEnableDisable)
    EVT_UPDATE_UI(eCmdEnableDisable, CAppExplorerCmdHandler::OnUpdateEnableDisable)
    EVT_MENU(eCmdNewFolder, CAppExplorerCmdHandler::OnNewFolder)
    EVT_UPDATE_UI(eCmdNewFolder, CAppExplorerCmdHandler::OnUpdateNewFolder)

    EVT_MENU( eCmdCloseProjectView, CAppExplorerCmdHandler::OnCloseView )
    EVT_UPDATE_UI( eCmdCloseProjectView, CAppExplorerCmdHandler::OnUpdateCloseView )
    EVT_MENU(eCmdActivateClients, CAppExplorerCmdHandler::OnActivateClients)
    EVT_UPDATE_UI(eCmdActivateClients, CAppExplorerCmdHandler::OnUpdateActivateClients)
    EVT_MENU(wxID_PROPERTIES, CAppExplorerCmdHandler::OnProperties)
    EVT_UPDATE_UI(wxID_PROPERTIES, CAppExplorerCmdHandler::OnUpdateProperties)

    EVT_UPDATE_UI(eCmdSettings, CAppExplorerCmdHandler::OnUpdateSettings)
END_EVENT_TABLE()


CAppExplorerCmdHandler::CAppExplorerCmdHandler(IServiceLocator* srv_locator)
:   m_ServiceLocator(srv_locator)
{
}


CAppExplorerCmdHandler::~CAppExplorerCmdHandler()
{
}


void CAppExplorerCmdHandler::x_CutOrCopy(bool cut)
{
    TItemRefVector sel_items;
    x_GetSelectedItems(sel_items);

    x_GetAppExplorerService()->CutOrCopyToClipboard(sel_items, cut);
}


bool CAppExplorerCmdHandler::x_CanUpdateCutOrCopy()
{
    TItemRefVector sel_items;
    x_GetSelectedItems(sel_items);

    bool en = x_GetAppExplorerService()->CanCutOrCopyToClipboard(sel_items);
    return en;
}


void CAppExplorerCmdHandler::OnCut(wxCommandEvent& event)
{
    x_CutOrCopy(true);
}


void CAppExplorerCmdHandler::OnUpdateCut(wxUpdateUIEvent& event)
{
    bool en = x_CanUpdateCutOrCopy();
    event.Enable(en);
}


void CAppExplorerCmdHandler::OnCopy(wxCommandEvent& event)
{
    x_CutOrCopy(false);
}


void CAppExplorerCmdHandler::OnUpdateCopy(wxUpdateUIEvent& event)
{
    bool en = x_CanUpdateCutOrCopy();
    event.Enable(en);
}


void CAppExplorerCmdHandler::OnPaste(wxCommandEvent& event)
{
    TItemRefVector sel_items;
    x_GetSelectedItems(sel_items);

    if( sel_items.size() == 1 ){
        CExplorerItem& item = *sel_items[0];
        x_GetAppExplorerService()->PasteFromClipboard(item);
    } else _ASSERT(false);
}


void CAppExplorerCmdHandler::OnUpdatePaste(wxUpdateUIEvent& event)
{
    TItemRefVector sel_items;
    x_GetSelectedItems(sel_items);

    bool en = false;
    if(sel_items.size() == 1)   {
        CExplorerItem& item = *sel_items[0];
        en = x_GetAppExplorerService()->CanPasteTo(item);
    }
    event.Enable(en);
}

void CAppExplorerCmdHandler::OnRemove(wxCommandEvent& event)
{
    TItemRefVector sel_items;
    x_GetSelectedItems(sel_items);
    if (sel_items.size()) {

        wxString quest;
        if (sel_items.size() == 1) {
            quest = wxT("Delete selected item?");
        } else {
            quest.Printf(wxT("Delete %lu selected items?"), (unsigned long)sel_items.size());
        }
        int answer = 
            wxMessageBox(quest, wxT("Confirm"), 
                         wxYES_NO | wxCANCEL | wxICON_QUESTION );
        if (answer == wxYES) {
            x_GetAppExplorerService()->Remove(sel_items);
        }
    }
}


void CAppExplorerCmdHandler::OnUpdateRemove(wxUpdateUIEvent& event)
{
    // check if we can delete this selection
    TItemRefVector sel_items;
    x_GetSelectedItems(sel_items);

    bool en = false;
    for (unsigned i = 0; i < sel_items.size(); ++i) {
        CExplorerItem& item = *sel_items[i];
        CAppExplorerService* ae_service = x_GetAppExplorerService();

        // we can remove only very specific types of objects
        if (ae_service) {
            CAppExplorerService::EItemType item_type = ae_service->GetItemType(item);
            switch (item_type)
            {
            case CAppExplorerService::eProjectItem:
            case CAppExplorerService::eProjectFolder:
            case CAppExplorerService::eDataLoader:
                en = !ae_service->IsDataFolder( item ); 
                break;
            default:
                en = false; 
                break;
            } // switch
        }
        // if any element in the set is non-removable -- we exit 
        if (!en) {
            break;
        }

    } // for

    event.Enable(en);
}


void CAppExplorerCmdHandler::OnEnableDisable(wxCommandEvent& event)
{
    TItemRefVector sel_items;
    x_GetSelectedItems(sel_items);

    x_GetAppExplorerService()->EnableDisableProjectItems(sel_items);
}


void CAppExplorerCmdHandler::OnUpdateEnableDisable(wxUpdateUIEvent& event)
{
    TItemRefVector sel_items;
    x_GetSelectedItems(sel_items);

    if (x_GetAppExplorerService()->MatchTypesExactly(sel_items, CAppExplorerService::eProjectItem)  ||
        x_GetAppExplorerService()->MatchTypesExactly(sel_items, CAppExplorerService::eDataLoader)) {
        // only Project Items or Data Loaders
        bool en = false, dis = false;
        for(  size_t i = 0;  i < sel_items.size();  i++ )   {
            CExplorerItem& item = *sel_items[i];
            if(item.IsEnabled())    {
                en = true;
            } else {
                dis = true;
            }
        }
        wxString text = (en == dis) ? wxT("Enable / Disable") : (en ? wxT("Disable") : wxT("Enable"));
        event.SetText(text);
        event.Enable(true);
    } else {
        event.Enable(false);
    }
}


void CAppExplorerCmdHandler::OnNewFolder(wxCommandEvent& event)
{
    TItemRefVector sel_items;
    x_GetSelectedItems(sel_items);

    if(sel_items.size() == 1)   {
        CExplorerItem& item = *sel_items[0];
        x_GetAppExplorerService()->NewFolder(item);
    } else _ASSERT(false);
}


void CAppExplorerCmdHandler::OnUpdateNewFolder(wxUpdateUIEvent& event)
{
    bool en = false;

    TItemRefVector sel_items;
    x_GetSelectedItems(sel_items);

    if(sel_items.size() == 1)   {
        CExplorerItem& item = *sel_items[0];
        CAppExplorerService::EItemType type =
                x_GetAppExplorerService()->GetItemType(item);

        en = (type == CAppExplorerService::eProjectFolder)  ||
             (type == CAppExplorerService::eWorkspaceFolder) ||
             (type == CAppExplorerService::eWorkspace);
    }
    event.Enable(en);
}


void CAppExplorerCmdHandler::OnCloseView(wxCommandEvent& event)
{
    CAppExplorerService::TProjectViewVector views;

    TItemRefVector sel_items;
    x_GetSelectedItems(sel_items);

    x_GetAppExplorerService()->ItemsToViews(sel_items, views);

    CProjectService* prj_srv =
        m_ServiceLocator->GetServiceByType<CProjectService>();

    for(  size_t i = 0; i < views.size();  i++) {
        IProjectView& view = *views[i];
        prj_srv->RemoveProjectView(view);
    }
}


void CAppExplorerCmdHandler::OnUpdateCloseView(wxUpdateUIEvent& event)
{
    CAppExplorerService::TProjectViewVector views;

    TItemRefVector sel_items;
    x_GetSelectedItems(sel_items);

    x_GetAppExplorerService()->ItemsToViews(sel_items, views);


    event.Enable(views.size() > 0);
}


void CAppExplorerCmdHandler::OnActivateClients(wxCommandEvent& event)
{
    CAppExplorerService::TProjectViewVector views;

    TItemRefVector sel_items;
    x_GetSelectedItems(sel_items);

    x_GetAppExplorerService()->ItemsToViews(sel_items, views);

    vector<IWMClient*>  clients;
    for(  size_t i = 0; i < views.size();  i++) {
        IProjectView* view = views[i].GetPointer();
        clients.push_back(view);
    }

    IWindowManagerService* wm_srv =
        m_ServiceLocator->GetServiceByType<IWindowManagerService>();

    wm_srv->ActivateClients(clients);
}


void CAppExplorerCmdHandler::OnUpdateActivateClients(wxUpdateUIEvent& event)
{
    CAppExplorerService::TProjectViewVector views;

    TItemRefVector sel_items;
    x_GetSelectedItems(sel_items);

    x_GetAppExplorerService()->ItemsToViews(sel_items, views);

    event.Enable(views.size() > 0);
}


void CAppExplorerCmdHandler::OnProperties(wxCommandEvent& event)
{
    TItemRefVector sel_items;
    x_GetSelectedItems(sel_items);

    if(sel_items.size() == 1)   {
        CExplorerItem& item = *sel_items[0];
        x_GetAppExplorerService()->OnProperties(item);
    } else _ASSERT(false);
}


void CAppExplorerCmdHandler::OnUpdateProperties(wxUpdateUIEvent& event)
{
    bool en = false;

    TItemRefVector sel_items;
    x_GetSelectedItems(sel_items);

    if(sel_items.size() == 1)   {
        CExplorerItem& item = *sel_items[0];
        en = x_GetAppExplorerService()->IsPropertiesCmdEnabled(item);
    }
    event.Enable(en);
}

void CAppExplorerCmdHandler::OnUpdateSettings(wxUpdateUIEvent& event)
{
    event.Enable(false);
}

CAppExplorerService* CAppExplorerCmdHandler::x_GetAppExplorerService()
{
    _ASSERT(m_ServiceLocator);

    CAppExplorerService* exp_srv =
        m_ServiceLocator->GetServiceByType<CAppExplorerService>();

    _ASSERT(exp_srv);
    return exp_srv;
}


CAppTaskService* CAppExplorerCmdHandler::x_GetAppTaskService()
{
    _ASSERT(m_ServiceLocator);

    CAppTaskService* task_srv =
        m_ServiceLocator->GetServiceByType<CAppTaskService>();

    _ASSERT(task_srv);
    return task_srv;
}


void CAppExplorerCmdHandler::x_LaunchTask(IAppTask* task)
{
    CIRef<IAppTask> ref(task);
    x_GetAppTaskService()->AddTask(*ref);
}


void CAppExplorerCmdHandler::x_GetSelectedItems(TItemRefVector& items)
{
    IWindowManagerService* srv =
        m_ServiceLocator->GetServiceByType<IWindowManagerService>();
    if(srv) {
        IWMClient* client = srv->GetActiveClient();
        IExplorerItemSelectionSource* source =
                dynamic_cast<IExplorerItemSelectionSource*>(client);
        if(source)  {
            source->GetSelectedItems(items);
        }
    }
}


END_NCBI_SCOPE
