/*  $Id: app_explorer_service.cpp 26332 2012-08-28 15:58:15Z katargir $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Andrey Yazhuk, Vladimir Tereshkov
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>

#include <gui/core/app_explorer_service.hpp>

#include <gui/core/project_service.hpp>
#include <gui/core/project.hpp>
#include <gui/core/ui_data_source_service.hpp>
#include <gui/core/folder_edit_dlg.hpp>
#include <gui/core/project_edit_dlg.hpp>
#include <gui/core/commands.hpp>
#include <gui/core/fname_validator_imp.hpp>

#include <gui/framework/view_manager_service.hpp>
#include <gui/framework/menu_service.hpp>

#include <gui/widgets/wx/wx_utils.hpp>
#include <gui/widgets/wx/fileartprov.hpp>
#include <gui/widgets/wx/message_box.hpp>

#include <gui/objutils/registry.hpp>
#include <gui/objutils/label.hpp>

#include <gui/utils/ui_object.hpp>

#include <gui/objects/GBWorkspace.hpp>
#include <gui/objects/GBProjectHandle.hpp>
#include <gui/objects/FolderInfo.hpp>
#include <gui/objects/ProjectDescr.hpp>

#include <objects/seq/Annotdesc.hpp>
#include <objects/general/Date.hpp>

#include <gui/objutils/interface_registry.hpp>
#include <gui/objutils/gui_object_info.hpp>

#include <serial/iterator.hpp>

#include <corelib/ncbitime.hpp>

#include <wx/clipbrd.h>
#include <wx/sstream.h>
#include <wx/txtstrm.h>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

const wxString CAppExplorerService::sm_DATA_FOLDER      = wxT("Data");
const wxString CAppExplorerService::sm_VIEWS_FOLDER     = wxT("Views");
const wxString CAppExplorerService::sm_ALL_VIEWS_FOLDER = wxT("All Views");
const wxString CAppExplorerService::sm_LOADERS_FOLDER   = wxT("Data Loaders");
const wxString CAppExplorerService::sm_DATA_SOURCES_FOLDER = wxT("Data Sources");


////////////////////////////////////////////////////////////////////////////////
/// CProjectFolderTreeItem

CProjectFolderTreeItem::CProjectFolderTreeItem(const wxString& label,
                                               CProjectFolderPrx::TId data,
                                               int normal, int expanded,
                                               bool is_open)
: TParentType(label, is_open)
{
    SetIconIndex(normal, expanded);
    SetData(data);
}


/* @@@ TODO
bool CProjectFolderTreeItem::IsExpanded() const
{
    return info.IsSetOpen() ? info.GetOpen() : true;
}


void CProjectFolderTreeItem::Expand(bool exp)
{
    m_Data->SetInfo().SetOpen(exp);
}*/


////////////////////////////////////////////////////////////////////////////////
/// CAppExplorerService

CAppExplorerService::CAppExplorerService()
:   m_ServiceLocator(NULL),
    m_ProjectService(NULL),
    m_IconsImageList(16, 16, true, 0)
{
    //LOG_POST("CAppExplorerService constructor");
}


CAppExplorerService::~CAppExplorerService()
{
    //LOG_POST("CAppExplorerService destructor");
}


int CAppExplorerService::x_AddIconToImageList(const string& alias)
{
    wxBitmap bmp = wxArtProvider::GetBitmap(ToWxString(alias));
    if( bmp.IsOk() ){
        return m_IconsImageList.Add(bmp);
    }

    return -1;
}


int CAppExplorerService::x_GetImageIndexByAlias(const string& alias)
{
    TAliasToIndexMap::iterator it = m_AliasToIndex.find(alias);
    if(it == m_AliasToIndex.end())  {
        int index = x_AddIconToImageList(alias);
        if( index != -1 ){
            m_AliasToIndex[alias] = index;
        }
        return index;
    } else {
        return it->second;
    }
}


void CAppExplorerService::InitService()
{
    LOG_POST(Info << "Initializing Application Explorer service...");

    wxFileArtProvider* provider = GetDefaultFileArtProvider();

    provider->RegisterFileAlias(wxT("tree::data_source"),     wxT("data_source.png"));

    provider->RegisterFileAlias(wxT("tree::workspace_open_item"),     wxT("workspace_open_item.png"));
    provider->RegisterFileAlias(wxT("tree::workspace_closed_item"),   wxT("workspace_closed_item.png"));
    provider->RegisterFileAlias(wxT("tree::project_closed_item"),    wxT("project_closed_item.png"));
    provider->RegisterFileAlias(wxT("tree::project_open_item"),      wxT("project_open_item.png"));
    provider->RegisterFileAlias(wxT("tree::folder_open"),        wxT("folder_open.png"));
    provider->RegisterFileAlias(wxT("tree::folder_closed"),      wxT("folder_closed.png"));
    provider->RegisterFileAlias(wxT("tree::doc_item"),           wxT("doc_item.png"));

    provider->RegisterFileAlias(wxT("tree::view_item"),          wxT("view_item.png"));
    provider->RegisterFileAlias(wxT("tree::doc_item_disabled"),  wxT("doc_item_disabled.png"));
    provider->RegisterFileAlias(wxT("tree::viewer_item"),        wxT("viewer_item.png"));

    provider->RegisterFileAlias(wxT("tree::bio_tree_item"),        wxT("bio_tree_container.png"));
    provider->RegisterFileAlias(wxT("tree::assembly_item"),        wxT("gc_assembly.png"));

    TGuard lock(*this);

    m_DataSourceIcon     = x_AddIconToImageList("tree::data_source");
    m_WorkspaceOpenIcon     = x_AddIconToImageList("tree::workspace_open_item");
    m_WorkspaceClosedIcon   = x_AddIconToImageList("tree::workspace_closed_item");
    m_ProjectClosedIcon = x_AddIconToImageList("tree::project_closed_item");
    m_ProjectOpenIcon   = x_AddIconToImageList("tree::project_open_item");
    m_FolderOpenIcon    = x_AddIconToImageList("tree::folder_open");
    m_FolderClosedIcon  = x_AddIconToImageList("tree::folder_closed");
    m_ProjectItemIcon   = x_AddIconToImageList("tree::doc_item");

    m_ViewerIcon        = x_AddIconToImageList("tree::viewer_item");
    m_DocIconDisabled   = x_AddIconToImageList("tree::doc_item_disabled");
    m_ViewIcon          = x_AddIconToImageList("tree::view_item");

    LoadSettings();

    LOG_POST(Info << "Finished initializing Application Explorer service");
}


template<class TRef>   void sDestoryItemAndChildren(TRef& ref)
{
    if(ref) {
        ref->EraseAllChildItems(true);
    }
    ref.Reset();
}


void CAppExplorerService::ShutDownService()
{
    LOG_POST(Info << "Shutting down Application Explorer service...");

    TGuard lock(*this);

    x_ClearClipboardData();

    /// this must be coordinated correctly - order here is important!
    DetachAllObservers();

    SaveSettings();

    sDestoryItemAndChildren(m_DSRootItem);
    sDestoryItemAndChildren(m_ViewsRootItem);
    sDestoryItemAndChildren(m_WorkspaceItem);
    sDestoryItemAndChildren(m_RootItem);

    LOG_POST(Info << "Finished shutting down Application Explorer service");
}


void CAppExplorerService::SetServiceLocator(IServiceLocator* locator)
{
    m_ServiceLocator = locator;
}


void CAppExplorerService::SetRegistryPath(const string& path)
{
    m_RegPath = path;
}


static const char* kEditFolderDialogTag = ".EditFolderDialog";
static const char* kEditProjectItemDialogTag = ".EditProjectItemDialog";
static const char* kEditProjectDialogTag = ".EditProjectDialog";


void CAppExplorerService::SaveSettings() const
{
    _ASSERT(! m_RegPath.empty());

    if( ! m_RegPath.empty())   {
        CGuiRegistry& gui_reg = CGuiRegistry::GetInstance();
        CGuiRegistry::TReadWriteView view = gui_reg.GetReadWriteView(m_RegPath);

    }
}


void CAppExplorerService::LoadSettings()
{
    _ASSERT(! m_RegPath.empty());

    if( ! m_RegPath.empty())   {
        CGuiRegistry& gui_reg = CGuiRegistry::GetInstance();
        CGuiRegistry::TReadView view = gui_reg.GetReadView(m_RegPath);

    }
}


/// reacts to updates sent by CProjectService
void CAppExplorerService::OnUpdate(IObserverUpdate& update)
{
    CObserverUpdate* up = dynamic_cast<CObserverUpdate*>(&update);
    _ASSERT(up);

    IObservable* observable = up->GetObservable();
    CProjectService* srv = dynamic_cast<CProjectService*>(observable);
    _ASSERT(srv);

    switch(up->GetType())    {
    case CObserverUpdate::eAttached:
        _ASSERT(! m_ProjectService);
        m_ProjectService = srv;

        x_UpdateItems();
        break;

    case CObserverUpdate::eStateChanged:
        x_UpdateItems();
        break;

    case CObserverUpdate::ePreDetach:
        _ASSERT(m_ProjectService);

        m_ProjectService = NULL;
        x_UpdateItems();
        break;
    }
}


// rebuilds items
void CAppExplorerService::x_UpdateItems()
{
    //CStopWatch sw; sw.Start();

    x_ClearClipboardData();

    TGuard  guard(*this);

    //LOG_POST("CAppExplorerService::x_UpdateItems() got lock " << int(1000 * sw.Elapsed()) << " ms");
    //sw.Restart();

    if( ! m_RootItem)   {
        m_RootItem.Reset(new CExplorerItem(wxT("Root")));
    }

    if(m_ProjectService)    {
        x_UpdateWorkspaceTree();
        x_UpdateWindowsTree();
    }

    x_UpdateDataSourceTree();

    //LOG_POST("CAppExplorerService::x_UpdateItems() rebuilt items " << int(1000 * sw.Elapsed()) << " ms");
    //sw.Restart();

    CObservable::x_NotifyObservers();

    //LOG_POST("CAppExplorerService::x_UpdateItems() updated observers" << int(1000 * sw.Elapsed()) << " ms");
}

/// update Workspace branch that contains Workspace and Projects
void CAppExplorerService::x_UpdateWorkspaceTree()
{
    // lock project service to get access to workpace
    TProjectServiceGuard srv_lock(*m_ProjectService);

    if(m_ProjectService->HasWorkspace() == false)  { // workspace does not exist
        if(m_WorkspaceItem) {
            m_WorkspaceItem->EraseAllChildItems(true);
            m_RootItem->EraseChildItem(*m_WorkspaceItem);
            m_WorkspaceItem.Reset();
        }
    } else {
        // workspace exists but may have been changed
        m_Workspace = m_ProjectService->GetWorkspaceConst();

        TWorkspaceId id(m_Workspace->GetId());
        CRef<CWorkspaceFolderConstPrx> ws_folder = m_Workspace->GetRootFolder_C();
        string title = ws_folder->GetName();

        if(m_WorkspaceItem.GetPointer() == NULL)   {
            // create Workspace tree item
            m_WorkspaceItem.Reset(new TWorkspaceTreeItem(wxEmptyString, id,
                                      m_WorkspaceClosedIcon, m_WorkspaceOpenIcon));
            m_WorkspaceItem->Expand(true);
            m_RootItem->AddChildItem(*m_WorkspaceItem);
        } else {
            // update existing Workspace tree item
            m_WorkspaceItem->SetLabel(title);
            m_WorkspaceItem->SetData(id);
        }
        x_UpdateWorkspaceFolder(*m_WorkspaceItem, *ws_folder);

        m_Workspace.Reset();
    }
}


// Updates Workspace Folder tree item and all its subitems
void CAppExplorerService::x_UpdateWorkspaceFolder(CExplorerItem& item,
                                                  CWorkspaceFolderConstPrx& folder)

{
    string name = folder.GetName();
    bool root = (&item == m_WorkspaceItem.GetPointer());
    if(root)    {
        name = "Workspace \'" + name + "\'";
        if(m_Workspace->IsDirty())  {
            name += " (*)";
        }
    }
    item.SetLabel(name);

    x_UpdateWorkspaceFolder_Folders(item, folder);

    x_UpdateWorkspaceFolder_Projects(item, folder);
}


void CAppExplorerService::x_UpdateWorkspaceFolder_Folders(CExplorerItem& item,
                                                  CWorkspaceFolderConstPrx& folder)

{
    CWorkspaceFolderConstPrx::TFolderConstVector sub_folders;
    folder.GetChildFolders(sub_folders);

    typedef map<TWsFolderId, CWorkspaceFolderConstPrx*> TWsFolderMap;
    TWsFolderMap    new_folders;
    for( size_t i = 0;  i < sub_folders.size();  i++ )  {
        CWorkspaceFolderConstPrx* f = sub_folders[i].GetPointer();
        TWsFolderId id = f->GetId();
        new_folders[id] = f;
    }

    TItemRefVector& child_items = item.x_GetChildItems();

    // remove obsolete child items
    for(  size_t i = 0;  i < child_items.size();  )   {
        CExplorerItem& it = *child_items[i];
        const type_info& tp_info = typeid(it);

        if(tp_info == typeid(TWsFolderTreeItem)) {
            TWsFolderTreeItem* ch_item = dynamic_cast<TWsFolderTreeItem*>(&it);
            TWsFolderId folder_id = ch_item->GetData();

            TWsFolderMap::iterator it_new = new_folders.find(folder_id);
            if(it_new == new_folders.end())   { // ch_folder is obsolete
                ch_item->EraseAllChildItems(true);
                item.EraseChildItem(i);
            } else {
                x_UpdateWorkspaceFolder(*ch_item, *it_new->second); // update this item
                new_folders.erase(it_new); // exclude from futher processing
                ++i;
            }
        } else i++;
    }

    // create new tree items for the new Folders
    NON_CONST_ITERATE(TWsFolderMap, it_new, new_folders) {
        CWorkspaceFolderConstPrx* ws_folder = it_new->second;

        TWsFolderTreeItem* new_item =
            new TWsFolderTreeItem(ToWxString(ws_folder->GetName()), ws_folder->GetId(),
                                  m_FolderClosedIcon, m_FolderOpenIcon);
        item.AddChildItem(*new_item);

        x_UpdateWorkspaceFolder(*new_item, *ws_folder); // recusrsively create child items
    }
}


// updates "Project" tree items in the given Workspace Folder
void CAppExplorerService::x_UpdateWorkspaceFolder_Projects(CExplorerItem& item,
                                                           CWorkspaceFolderConstPrx& folder)

{
    CWorkspaceFolderConstPrx::TProjectVector projects;
    folder.GetProjects(projects);

    // make a list of existing Projects
    typedef set<TProjectId>  TProjectIDSet;
    TProjectIDSet   new_projects;
    for( size_t i = 0;  i < projects.size();  i++ )  {
        CRef<CProjectPrx> prj = projects[i];
        new_projects.insert(prj->GetId());
    }

    TItemRefVector& child_items = item.x_GetChildItems();

    // remove obsolete child items
    for(  size_t i = 0;  i < child_items.size();  )   {
        CExplorerItem& it = *child_items[i];
        const type_info& tp_info = typeid(it);

        if(tp_info == typeid(TProjectTreeItem)) {
            TProjectTreeItem* ch_item = dynamic_cast<TProjectTreeItem*>(&it);
            TProjectId prj_id = ch_item->GetData();

            TProjectIDSet::iterator it_new = new_projects.find(prj_id);
            if(it_new == new_projects.end())   { // ch_filename is obsolete
                ch_item->EraseAllChildItems(true);
                item.EraseChildItem(i);
            } else {
                x_UpdateProject(*ch_item); // update this item
                new_projects.erase(it_new); // exclude from futher processing
                ++i;
            }
        } else i++;
    }

    // create new tree items for the new Projects
    NON_CONST_ITERATE(TProjectIDSet, it_new_pr, new_projects) {
        TProjectId prj_id = *it_new_pr;
        wxString filename = wxT("no filename"); //### HACK
        TProjectTreeItem* new_item =
            new TProjectTreeItem(filename, prj_id, m_ProjectClosedIcon, m_ProjectOpenIcon);
        new_item->Expand(true);
        item.AddChildItem(*new_item);

        x_UpdateProject(*new_item); // recursively create child items
    }
}


// creates or update an item representing a Project
void CAppExplorerService::x_UpdateProject(TProjectTreeItem& item)
{
    // Project Service and Workspace have been locked already
    int prj_id = item.GetData();

    m_CurrProject = m_Workspace->FindProjectById(prj_id);

    _ASSERT(m_CurrProject);

    wxString label;
    if (m_CurrProject  &&  m_CurrProject->IsLoaded()) {    // project is loaded
        label = ToWxString(m_CurrProject->GetName());
        if(m_CurrProject->IsDirty())    {
            label += wxT(" (*)");
        }

        // Update "Data" folder  - must be shown only if it is not empty
        TItemRefVector& child_items = item.x_GetChildItems();
        CExplorerItem* ch_item = x_FindItemByLabel(sm_DATA_FOLDER, child_items);

        CRef<CProjectFolderConstPrx> data = m_CurrProject->GetDataFolder_C();

        TProjectFolderTreeItem* data_item = dynamic_cast<TProjectFolderTreeItem*>(ch_item);
        if (data_item == NULL) {
            bool is_open = data->GetOpen();
            data_item = new TProjectFolderTreeItem(sm_DATA_FOLDER, data->GetId(),
                                                    m_FolderClosedIcon, m_FolderOpenIcon,
                                                    is_open);
            item.AddChildItem(*data_item);
        }

        x_UpdateProjectFolder(*data_item, *data);
        
        // Update "Data Loaders" folder
        x_UpdateProject_DataLoaders(item);
        
        // Update "Views" folder
        x_UpdateProject_Views(item);

    } else if (m_CurrProject  &&  m_CurrProject->IsLoading()) {    // project is loading
        label = m_CurrProject->GetFilename() + wxT(" (Loading ...)");
        item.EraseAllChildItems(true);
    } else {
        // Project is not loaded
        label = m_CurrProject->GetFilename() + wxT(" (not loaded)");
        item.EraseAllChildItems(true);
    }

    // update the Project label
    item.SetLabel(label);

    m_CurrProject.Reset();
}


// Update "Data Loaders" folder in the given "Project" tree item
void CAppExplorerService::x_UpdateProject_DataLoaders(TProjectTreeItem& item)
{
    TItemRefVector& child_items = item.x_GetChildItems();

    // Data Loaders item must be shown only if it is not empty
    CExplorerItem* ch_item = x_FindItemByLabel(sm_LOADERS_FOLDER, child_items);
    const CProjectPrx::TDataLoaders& loaders = m_CurrProject->GetDataLoaders();
    TFolderTreeItem* loaders_item = dynamic_cast<TFolderTreeItem*>(ch_item);

    if (loaders.size()) {
        // the project has data loaders
        if(loaders_item == NULL) {
            loaders_item = new TFolderTreeItem(sm_LOADERS_FOLDER, sm_LOADERS_FOLDER, m_FolderClosedIcon, m_FolderOpenIcon);
            item.AddChildItem(*loaders_item);
            loaders_item->Expand(true);
        }
        x_UpdateProjectLoadersFolder(*loaders_item);
    } else if(ch_item) {
        x_UpdateProjectLoadersFolder(*loaders_item);

        TItemRefVector::iterator it =
            std::find(child_items.begin(), child_items.end(), TItemRef(loaders_item));

        (*it)->EraseAllChildItems(true);
        item.EraseChildItem(it);
    }
}


// Update "Views" folder in the given "Project" tree item
void CAppExplorerService::x_UpdateProject_Views(TProjectTreeItem& item)
{
    TItemRefVector& child_items = item.x_GetChildItems();

    // Views item must be shown only if it is not empty
    CExplorerItem* ch_item = x_FindItemByLabel(sm_VIEWS_FOLDER, child_items);
    const CProjectPrx::TViews& views = m_CurrProject->GetViews();
    TFolderTreeItem* views_item = dynamic_cast<TFolderTreeItem*>(ch_item);

    if (views.size()) {
        // the project has views
        if(views_item == NULL) {
            views_item = new TFolderTreeItem(sm_VIEWS_FOLDER, sm_VIEWS_FOLDER, m_FolderClosedIcon, m_FolderOpenIcon);
            item.AddChildItem(*views_item);
        }
        x_UpdateProjectWindowsFolder(*views_item);
    } else if(ch_item) {
        x_UpdateProjectWindowsFolder(*views_item);

        TItemRefVector::iterator it =
            std::find(child_items.begin(), child_items.end(), TItemRef(views_item));

        (*it)->EraseAllChildItems(true);
        item.EraseChildItem(it);
    }
}


CAppExplorerService::TProjectItemTreeItem*
     CAppExplorerService::x_CreateProjectItemItem(CProjectItemConstPrx& prj_item)
{
    string label = prj_item.GetName();
    TPrjItemId id = prj_item.GetId();

    TProjectItemTreeItem* item = new TProjectItemTreeItem(ToWxString(label), id,
                                                          m_ProjectItemIcon);
    return item;
}


void CAppExplorerService::x_DecorateProjectItem(TProjectItemTreeItem& tree_item,
                                                CProjectItemConstPrx& item )
{
    // Obtain Alias and Description from the Project Item
    string alias, descr;
    const CObject* obj = item.GetObject();

    _ASSERT(obj);

    if(obj) {
        SConstScopedObject sco(*obj, *m_CurrProject->GetScope());
        CIRef<IGuiObjectInfo> gui_info(CreateObjectInterface<IGuiObjectInfo>(sco, NULL));

        if( gui_info.IsNull() ) return;

        descr = gui_info->GetType();
        string strSubtype = gui_info->GetSubtype();
        if (!strSubtype.empty())
            descr += " (" + strSubtype + ")";

        alias = gui_info->GetIcon();
    }

    // now decorate the Tree Item
    if( !alias.empty() ){
        int index = x_GetImageIndexByAlias(alias);
        tree_item.SetIconIndex(index);
    }

    if( !descr.empty() ){
        TItemRefVector children = tree_item.GetChildItems();

        if( children.empty() ){
            TFolderTreeItem* descr_item = new TFolderTreeItem(ToWxString(descr), ToWxString(descr));
            children.push_back(TItemRef(descr_item));

        } else {
            children[0]->SetLabel(descr);
        }
    }
}


void CAppExplorerService::x_UpdateProjectFolder(TProjectFolderTreeItem& item,
                                                CProjectFolderConstPrx& folder)
{
    string name = folder.GetName();
    item.SetLabel(name);

    x_UpdateProjectFolder_Folders(item, folder);

    x_UpdateProjectFolder_Items(item, folder);
}


// Updates all "Project Item" tree items in the given "Project Folder"
void CAppExplorerService::x_UpdateProjectFolder_Items(TProjectFolderTreeItem& item,
                                                      CProjectFolderConstPrx& folder)
{
    TItemRefVector& child_items = item.x_GetChildItems();

    typedef map<TProjectId, CProjectItemConstPrx*> TProjectMap;
    TProjectMap new_prj_items;

    CProjectFolderConstPrx::TProjectItemConstVector items;
    folder.GetProjectItems_C(items);

    for(  size_t i = 0;  i < items.size();  i++ )    {
        CRef<CProjectItemConstPrx> prj_item = items[i];
        CProjectItemConstPrx::TId id = prj_item->GetId();
        new_prj_items[id] = prj_item.GetPointer();
    }

    // process existing items - delete or update
    for(  size_t i = 0;  i < child_items.size();  )   {
        CExplorerItem& it = *child_items[i];
        const type_info& tp_info = typeid(it);

        if(tp_info == typeid(TProjectItemTreeItem)) {
            TProjectItemTreeItem* ch_item = dynamic_cast<TProjectItemTreeItem*>(&it);
            CProjectPrx::TId item_id = ch_item->GetData();

            TProjectMap::iterator it_new = new_prj_items.find(item_id);
            if(it_new == new_prj_items.end())   {
                // ch_prj_item is obsolete- destroy the item and all its children
                ch_item->EraseAllChildItems(true);
                item.EraseChildItem(i);
            } else {
                x_UpdateProjectItem(*ch_item, folder); // update this item
                new_prj_items.erase(it_new); // exclude from futher processing
                ++i;
            }
        } else i++;
    }

    // create new Project Items
    NON_CONST_ITERATE(TProjectMap, it_new_pi,  new_prj_items) {
        CProjectItemConstPrx& prj_item = *it_new_pi->second;
        TProjectItemTreeItem* new_item = x_CreateProjectItemItem(prj_item);

        item.AddChildItem(*new_item);
        x_UpdateProjectItem(*new_item, folder); // recursively create child items
    }
}


// Updates all "Project Folder" tree items in the given "Project Folder" (subfolders)
void CAppExplorerService::x_UpdateProjectFolder_Folders(TProjectFolderTreeItem& item,
                                                        CProjectFolderConstPrx& folder)
{
    typedef CProjectFolderConstPrx::TId TFolderId;
    typedef map<TFolderId, CProjectFolderConstPrx*> TFolderMap;

    TFolderMap new_folders;

    CProjectFolderConstPrx::TFolderConstVector folders;
    folder.GetChildFolders_C(folders);
    for(  size_t i = 0; i < folders.size();  i++ )  {
        CRef<CProjectFolderConstPrx> f = folders[i];
        new_folders[f->GetId()] = f.GetPointer();
    }

    TItemRefVector& child_items = item.x_GetChildItems();

    // process existing items - delete or update
    for(  size_t i = 0;  i < child_items.size();  )   {
        CExplorerItem& it = *child_items[i];
        const type_info& tp_info = typeid(it);

        if(tp_info == typeid(TProjectFolderTreeItem)) {
            TProjectFolderTreeItem* ch_item = dynamic_cast<TProjectFolderTreeItem*>(&it);
            TFolderId folder_id = ch_item->GetData();

            TFolderMap::iterator it_new = new_folders.find(folder_id);
            if(it_new == new_folders.end())   {
                // ch_folder is obsolete - destroy the item and all its children
                ch_item->EraseAllChildItems(true);
                item.EraseChildItem(i);
            } else {
                CRef<CProjectFolderConstPrx> sub_folder = folder.FindChildFolderById_C(folder_id, false);
                x_UpdateProjectFolder(*ch_item, *sub_folder); // update this item
                new_folders.erase(it_new); // exclude from futher processing
                ++i;
            }
        } else i++;
    }

    // create new folders
    NON_CONST_ITERATE(TFolderMap, it_new,  new_folders) {
        CProjectFolderConstPrx& sub_folder = *it_new->second;
        wxString name = ToWxString(sub_folder.GetName());

        bool is_open = sub_folder.GetOpen();
        TProjectFolderTreeItem* new_item
            = new TProjectFolderTreeItem(name, sub_folder.GetId(),
                                         m_FolderOpenIcon, m_FolderClosedIcon,
                                         is_open);
        item.AddChildItem(*new_item);

        x_UpdateProjectFolder(*new_item, sub_folder); // recursively create child items
    }
}


CExplorerItem*  CAppExplorerService::x_FindItemByLabel(const wxString& label, TItemRefVector& items)
{
    NON_CONST_ITERATE(TItemRefVector, it, items)   {
        if((*it)->GetLabel() == label)  {
            return *it;
        }
    }
    return NULL;
}


void CAppExplorerService::x_UpdateProjectLoadersFolder(TFolderTreeItem& item)
{
    CProjectPrx::TDataLoaders loaders = m_CurrProject->GetDataLoaders();
    TItemRefVector& child_items = item.x_GetChildItems();

    typedef map<TDataLoaderId, const CLoaderDescriptor*> TLoaderMap;
    TLoaderMap new_items;

    ITERATE (CProjectPrx::TDataLoaders, iter, loaders) {
        CRef<CLoaderDescriptor> loader = *iter;
        TDataLoaderId id = loader->GetLabel();
        new_items[id] = loader.GetPointer();
    }

    // process existing items - delete or update
    for (size_t i = 0;  i < child_items.size(); ) {
        CExplorerItem& it = *child_items[i];
        TDataLoaderTreeItem* ch_item = dynamic_cast<TDataLoaderTreeItem*>(&it);

        if (ch_item) {
            const TDataLoaderId& item_id = ch_item->GetData();
            TLoaderMap::iterator it_new = new_items.find(item_id);
            if(it_new == new_items.end())   {
                // ch_prj_item is obsolete- destroy the item and all its children
                ch_item->EraseAllChildItems(true);
                item.EraseChildItem(i);
            } else {
                const CLoaderDescriptor* loader = it_new->second;
                ch_item->Enable(loader->GetEnabled());
                new_items.erase(it_new); // exclude from futher processing
                ++i;
            }
        } else i++;
    }

    // create new Project Items
    NON_CONST_ITERATE(TLoaderMap, it_new_li,  new_items) {
        const CLoaderDescriptor* loader = it_new_li->second;
        const string& label = loader->GetLabel();
        TDataLoaderTreeItem* new_item =
            new TDataLoaderTreeItem(ToWxString(label), label, m_ProjectItemIcon);

        item.AddChildItem(*new_item);
        new_item->Enable(loader->GetEnabled() );
    }
}


void CAppExplorerService::x_UpdateProjectWindowsFolder(TFolderTreeItem& item)
{
    CProjectPrx::TViews views = m_CurrProject->GetViews();

    TItemRefVector& child_items = item.x_GetChildItems();

    // remove obsolete items
    for(  size_t i = 0;  i < child_items.size();  )   {
        CExplorerItem& it = *child_items[i];
        bool found = false;

        // check whether *it represents an alive thing
        TViewTreeItem* view_item = dynamic_cast<TViewTreeItem*>( &it );
        if( view_item ){
            CProjectPrx::TViews::iterator it_new =
                std::find( views.begin(), views.end(), view_item->GetData() )
            ;
            if( it_new != views.end() ){
                IProjectView& view = *( view_item->GetData() );

                view_item->SetLabel( x_GetViewLabel( view, true ) );

                found = true;
                views.erase( it_new );
            }
        }
        if( !found ){
            // items is obsolete- destroy the item and all its children
            it.EraseAllChildItems(true);
            item.EraseChildItem(i);
        } else {
            ++i; // advance
        }
    }

    // create items for new views
    IViewManagerService* view_srv =
                    m_ServiceLocator->GetServiceByType<IViewManagerService>();

    NON_CONST_ITERATE(CProjectPrx::TViews, it_view, views)   {
        IProjectView& view = **it_view;
        if(view_srv->HasView(view))  {
            // this is an independent view, not belonging to a viewer - create item
            TViewTreeItem* view_item = x_CreateProjectViewTreeItem( view, true );
            item.AddChildItem(*view_item);
        }
    }
}


void CAppExplorerService::x_UpdateProjectItem(TProjectItemTreeItem& item,
                                              CProjectFolderConstPrx& parent_folder)
{
    TPrjItemId id = item.GetData();
    CRef<CProjectItemConstPrx> prj_item = parent_folder.FindProjectItemById_C(id, false);

    _ASSERT(prj_item);

    item.Enable(prj_item->IsEnabled());

    x_DecorateProjectItem(item, *prj_item);

    string title = prj_item->GetName();
    item.SetLabel(title);
}


string CAppExplorerService::x_GetViewLabel( IProjectView& view, bool local )
{
    if( local ){
        return
            view.GetLabel(IProjectView::eId) 
            + ": "
            + view.GetLabel(IProjectView::eTypeAndContent)
        ;
    } else {
        return view.GetClientLabel();
    }
}


CAppExplorerService::TViewTreeItem* CAppExplorerService::x_CreateProjectViewTreeItem(
    IProjectView& view, 
    bool local
)
{
    //TODO implement a mechanism for registering view type icons
    const CViewTypeDescriptor&  descr = view.GetTypeDescriptor();
    string alias = descr.GetIconAlias();
    int index = x_GetImageIndexByAlias(alias);
    TViewTreeItem* item = new TViewTreeItem( 
        ToWxString(x_GetViewLabel( view, local )), CIRef<IProjectView>(&view), index 
    );
    return item;
}

/*
CAppExplorerService::TViewTreeItem*
 CAppExplorerService::x_CreateProjectViewerTreeItem(IProjectView& viewer, bool local)
{
    string s = viewer.GetLabel(local ? IProjectView::eDefault : IProjectView::eContent);
    TViewTreeItem* item =
        new TViewTreeItem(s, CIRef<IProjectView>(&viewer), m_ViewerIcon);
    return item;
}
*/

// populates top-level WINDOWS_FOLDER folder in the Workspace Tree Browser
void CAppExplorerService::x_UpdateWindowsTree()
{
    if(m_ProjectService->HasWorkspace() == false)  { // workspace does not exist
        if(m_ViewsRootItem) {
            // delete root and all it's subitems
            m_RootItem->EraseChildItem(*m_ViewsRootItem);
            m_ViewsRootItem->EraseAllChildItems(true);
            m_ViewsRootItem.Reset();
        }
    } else {
        if(m_ViewsRootItem.GetPointer() == NULL)   {
            // create root Views item
            m_ViewsRootItem.Reset(new TFolderTreeItem(sm_ALL_VIEWS_FOLDER,
                                  sm_ALL_VIEWS_FOLDER, m_FolderClosedIcon, m_FolderOpenIcon));
            m_RootItem->AddChildItem(*m_ViewsRootItem);
            m_ViewsRootItem->Expand(true);
        }

        TItemRefVector& child_items = m_ViewsRootItem->x_GetChildItems();

        IViewManagerService* view_srv =
            m_ServiceLocator->GetServiceByType<IViewManagerService>();

        typedef IViewManagerService::TViews TViews;
        TViews views;
        view_srv->GetViews(views);

        // remove obsolete child items
        for(  size_t i = 0;  i < child_items.size();  )   {
            CExplorerItem& it_item = *child_items[i];

            CIRef<IView> view;
            // extract IWMClient from the tree item
            TViewTreeItem* view_item = dynamic_cast<TViewTreeItem*>(&it_item);
            if( view_item ){
                view = view_item->GetData();
            }
            TViews::iterator it = std::find(views.begin(), views.end(), view);
            if(it == views.end())   {
                // obsolete item - - destroy the item and all its children
                it_item.EraseAllChildItems(true);
                m_ViewsRootItem->EraseChildItem(i);

            } else {
                IProjectView& view = *(view_item->GetData()) ;

                view_item->SetLabel( x_GetViewLabel( view, false ) );

                views.erase(it);
                ++i;
            }
        }
        
        // create items for new views
        for(  size_t i = 0;  i < views.size();  i++ ){
            CExplorerItem* item  = NULL;
            IView* view = views[i];
            IProjectView* prj_view = dynamic_cast<IProjectView*>( view );
            
            if( prj_view ){
                item = x_CreateProjectViewTreeItem( *prj_view, false );
            }

            if( item ){
                m_ViewsRootItem->AddChildItem( *item );
            }
        }
    }
}


void CAppExplorerService::x_UpdateDataSourceTree()
{
    // create "Data Sources" folder if it does not exist
    if( ! m_DSRootItem) {
        m_DSRootItem.Reset(new TFolderTreeItem(sm_DATA_SOURCES_FOLDER,
                                    sm_DATA_SOURCES_FOLDER, m_DataSourceIcon));
        m_DSRootItem->Expand(true);
        m_RootItem->AddChildItem(*m_DSRootItem);
    }

    /// now update items for individual Data Sources
    _ASSERT(m_ServiceLocator);

    CUIDataSourceService* ds_srv =
        m_ServiceLocator->GetServiceByType<CUIDataSourceService>();

    _ASSERT(ds_srv);

    // obtain a list of Data Sources
    typedef CUIDataSourceService::TUIDataSourceVec TDSVec;
    TDSVec data_sources;
    ds_srv->GetDataSources(data_sources);

    TItemRefVector& child_items = m_DSRootItem->x_GetChildItems();

    // update existing items
    for(  size_t i = 0;  i < child_items.size();  )   {
        CExplorerItem& it = *child_items[i];

        TDataSourceTreeItem* ch_item = dynamic_cast<TDataSourceTreeItem*>(&it);
        CIRef<IUIDataSource> ds = ch_item->GetData();

        // find ds in data_sources
        TDSVec::iterator it_ds = std::find(data_sources.begin(), data_sources.end(), ds);
        if(it_ds == data_sources.end())   {
            // item is obsolete - destroy the item and all its children
            ch_item->EraseAllChildItems(true);
            m_DSRootItem->EraseChildItem(i);
        } else {
            //x_UpdateDataSource(*ch_item, prj); // update this item
            data_sources.erase(it_ds); // exclude from futher processing
            ++i;
        }
    }

    // create new items
    NON_CONST_ITERATE(TDSVec, it_ds, data_sources)    {
        IUIDataSource& ds = **it_ds;

        const IUIObject& descr = ds.GetDescr();
        wxString label = ToWxString(descr.GetLabel());
        string alias = descr.GetIconAlias();
        int index = x_GetImageIndexByAlias(alias);

        TDataSourceTreeItem* new_item(
            new TDataSourceTreeItem(label, CIRef<IUIDataSource>(&ds), index));
        m_DSRootItem->AddChildItem(*new_item);
    }
}


CRef<CExplorerItem> CAppExplorerService::GetRootItem(void)
{
    return m_RootItem;
}


wxImageList& CAppExplorerService::GetIconsImageList()
{
    return m_IconsImageList;
}


CAppExplorerService::EItemType
    CAppExplorerService::GetItemType(const CExplorerItem& item)
{
    const type_info& info = typeid(item);
    if(info == typeid(TWsFolderTreeItem))  {
        return eWorkspaceFolder;
    } else if(info == typeid(TWorkspaceTreeItem))  {
        return eWorkspace;
    } else if(info == typeid(TProjectTreeItem))  {
        return eProject;
    } else if(info == typeid(TProjectFolderTreeItem))  {
        return eProjectFolder;
    } else if(info == typeid(TProjectItemTreeItem))  {
        return eProjectItem;
    } else if(info == typeid(TViewTreeItem))  {
        return eView;
    } else if(info == typeid(TFolderTreeItem))  {
        return eFolder;
    } else if(info == typeid(TDataSourceTreeItem))  {
        return eDataSource;
    } else if(info == typeid(TDataLoaderTreeItem)) {
        return eDataLoader;
    }

    _ASSERT(false);
    return eInvalid;
}


int CAppExplorerService::GetItemTypes(TItemRefVector& items)
{
    int types = 0;
    for( size_t i = 0;  i < items.size(); i++)  {
        CExplorerItem& item = *items[i];
        types |= GetItemType(item);
    }
    return types;
}


bool CAppExplorerService::MatchType(const CExplorerItem& item, EItemType type)
{
    switch(type)    {
    case eWorkspaceFolder:
        return (typeid(item) == typeid(TWsFolderTreeItem)) != 0;
    case eProject:
        return (typeid(item) == typeid(TProjectTreeItem)) != 0;
    case eProjectFolder:
        return (typeid(item) == typeid(TProjectFolderTreeItem)) != 0;
    case eProjectItem:
        return (typeid(item) == typeid(TProjectItemTreeItem)) != 0;
    case eView:
        return (typeid(item) == typeid(TViewTreeItem)) != 0;
    case eFolder:
        return (typeid(item) == typeid(TFolderTreeItem)) != 0;
    case eDataSource:
        return (typeid(item) == typeid(TDataSourceTreeItem)) != 0;
    case eDataLoader:
        return (typeid(item) == typeid(TDataLoaderTreeItem)) != 0;
    default:
        _ASSERT(false); // unexpected
    }
    return false;
}


bool CAppExplorerService::MatchTypesExactly(TItemRefVector& items, int type_mask)
{
    int types = 0;
    for( size_t i = 0;  i < items.size(); i++)  {
        CExplorerItem& item = *items[i];
        types |= GetItemType(item);
    }
    return types == type_mask;
}


void CAppExplorerService::FilterByType(TItemRefVector& items, int type_mask)
{
    int filtered = 0;
    for( size_t i = 0;  i < items.size(); i++)  {
        TItemRef& item = items[i];
        int type = GetItemType(*item);
        if((type & type_mask) == 0)    {
            item.Reset();
            filtered++;
        }
    }
    if(filtered)    {
        // purge filtered items
        int shift = 0;
        for( size_t i = 0;  i < items.size(); i++)  {
            TItemRef& item = items[i];
            if(item) {
                if(shift > 0)   {
                    items[i - shift] = items[i];
                }
            } else {
                shift++; // this is "hole" increase offset
            }
        }
    }
}


bool CAppExplorerService::ContainsAllTypes(TItemRefVector& items, int type_mask)
{
    int types = 0;
    for( size_t i = 0;  i < items.size(); i++)  {
        CExplorerItem& item = *items[i];
        types |= GetItemType(item);
        if((types & type_mask) == type_mask)    {
            return true;
        }
    }
    return false;
}


void CAppExplorerService::ItemsToViews(TItemRefVector& items, TProjectViewVector& views)
{
    set<IProjectView*> st_sel_views; // to avoid dupplication

    NON_CONST_ITERATE(TItemRefVector, it, items)  {
        CExplorerItem& item = **it;
        if(typeid(item) == typeid(TViewTreeItem)) {
            TViewTreeItem* vw_item = dynamic_cast<TViewTreeItem*>(&item);
            IProjectView* view = vw_item->GetData().GetPointer();
            st_sel_views.insert(view);
        }
    }
    // copy from st_sel_views to sel_views
    NON_CONST_ITERATE(set<IProjectView*>, it, st_sel_views) {
        views.push_back(CIRef<IProjectView>(*it));
    }
}

bool CAppExplorerService::IsDataFolder( CExplorerItem& item )
{
    return IsFolder( item, sm_DATA_FOLDER );
}

bool CAppExplorerService::IsViewsFolder( CExplorerItem& item )
{
    return IsFolder( item, sm_VIEWS_FOLDER );
}

bool CAppExplorerService::IsFolder( const CExplorerItem& item, wxString type )
{
    // TFolderTreeItem == CExplorerItemEx<string>

    const TFolderTreeItem* folder_item = dynamic_cast<const TFolderTreeItem*>(&item);

    return folder_item && ( type.empty() || type == folder_item->GetData() );
}


void CAppExplorerService::NewFolder(CExplorerItem& item)
{
    TGuard lock(*this);

    EItemType type = GetItemType(item);

    switch(type)    {
    case eProjectFolder:
        NewProjectFolder(item);
        break;
    case eWorkspace:
    case eWorkspaceFolder:
        NewWorkspaceFolder(item);
        break;
    default:
        _ASSERT(false); // unexpected
    }
}

struct stFolderInfo {
    string m_Name;
    string m_Descr;

    stFolderInfo(const string& name, const string& descr)   : m_Name(name), m_Descr(descr)  {}
};


bool x_EditFolder(stFolderInfo& info, const string& title, string reg_path,
                  bool readonly, IFNameValidator& validator)
{
    CFolderEditDlg dlg(NULL, wxID_ANY, ToWxString(title));
    dlg.SetRegistryPath(reg_path);

    dlg.SetReadOnly(readonly);
    dlg.SetNameValidator(&validator);

    dlg.SetFolderName(ToWxString(info.m_Name));
    dlg.SetDescr(ToWxString(info.m_Descr));

    if(readonly)    {
        dlg.ShowModal();
        return true;
    } else {
        if(dlg.ShowModal() == wxID_OK) {
            info.m_Name = ToStdString(dlg.GetFolderName());
            info.m_Descr = ToStdString(dlg.GetDescr());
            return true;
        }
        return false;
    }
}


void CAppExplorerService::NewWorkspaceFolder(CExplorerItem& item)
{
    TProjectServiceGuard guard(*m_ProjectService);
    CRef<CWorkspacePrx> ws = m_ProjectService->GetWorkspace();

    TWorkspaceTreeItem* ws_item = dynamic_cast<TWorkspaceTreeItem*>(&item);

    CRef<CWorkspaceFolderPrx> parent_folder;
    if(ws_item) {
        parent_folder = ws->GetRootFolder();
    } else {
        TWsFolderTreeItem* folder_item = dynamic_cast<TWsFolderTreeItem*>(&item);
        _ASSERT(folder_item);
        TWsFolderId id = folder_item->GetData();
        parent_folder = ws->FindWorkspaceFolderById(id);
    }
    _ASSERT(parent_folder);

    stFolderInfo info("New Folder" , "");
    string reg_path = m_RegPath + kEditFolderDialogTag;
    CFolderNameValidator<CWorkspaceFolderPrx> validator(parent_folder, "");

    // show the dialog
    if(x_EditFolder(info, "New Workspace Folder", reg_path, false, validator))  {
        // finally add the folder to the Workspace
        parent_folder->AddChildFolder(info.m_Name, info.m_Descr);
    }
}


void CAppExplorerService::NewProjectFolder(CExplorerItem& item)
{
    TProjectServiceGuard guard(*m_ProjectService); // lock the Service

    TProjectFolderTreeItem* folder_item = dynamic_cast<TProjectFolderTreeItem*>(&item);
    _ASSERT(folder_item);

    TProjectId prj_id = GetProjectIdByChild(*folder_item);
    CRef<CWorkspaceConstPrx> ws = m_ProjectService->GetWorkspaceConst();

    //CRef<CProjectConstPrx> project_c = ws->FindProjectById_C(prj_id);
    CRef<CProjectPrx> project = ws->FindProjectById(prj_id);

    CProjectFolderPrx::TId folder_id = folder_item->GetData();
    CRef<CProjectFolderPrx> parent_folder = project->FindProjectFolderById(folder_id);

    stFolderInfo info("New Folder" , "");
    string reg_path = m_RegPath + kEditFolderDialogTag;
    CFolderNameValidator<CProjectFolderPrx> validator(&*parent_folder, "");

    // show the Editing Dialog
    if(x_EditFolder(info, "New Project Folder", reg_path, false, validator))  {
        parent_folder->AddChildFolder(info.m_Name, info.m_Descr);
    }
}


CAppExplorerService::TProjectId
    CAppExplorerService::GetProjectIdByChild(CExplorerItem& item)
{
    TGuard guard(*this);

    CExplorerItem* it = &item;
    for( ; it  &&  ! MatchType(*it, eProject);
        it = it->GetParent() )   {};

    TProjectTreeItem* prj_item = dynamic_cast<TProjectTreeItem*>(it);

    //_ASSERT(prj_item);

    return prj_item ? prj_item->GetData() : TProjectId(-1);
}


void CAppExplorerService::OnItemActivated(CExplorerItem& item)
{
    CAppExplorerService::EItemType type = GetItemType(item);

    switch(type)    {
    case CAppExplorerService::eWorkspace:
    case CAppExplorerService::eProjectFolder:
        // OnProperties(item);
        break;

    case CAppExplorerService::eProject:
        if(IsPropertiesCmdEnabled(item))    {
            // OnProperties(item);
        }
        break;

    case CAppExplorerService::eProjectItem: {{
        CIRef<IMenuService> menu_srv = m_ServiceLocator->GetServiceByType<IMenuService>();
        menu_srv->AddPendingCommand(eCmdOpenViewDefault);
        break;
    }}
    case CAppExplorerService::eView:    {{
        CIRef<IMenuService> menu_srv = m_ServiceLocator->GetServiceByType<IMenuService>();
        menu_srv->AddPendingCommand(eCmdActivateClients);
        break;
    }}
    default:
        break;
    }
}

void CAppExplorerService::OnItemExpandedCollapsed(CExplorerItem& item)
{
    CAppExplorerService::EItemType type = GetItemType(item);

    switch(type)    {
    case CAppExplorerService::eWorkspaceFolder:

        break;

    case CAppExplorerService::eProjectFolder:
        {{
            TProjectFolderTreeItem* folder_item = dynamic_cast<TProjectFolderTreeItem*>(&item);
            _ASSERT(folder_item);

            TProjectId prj_id = GetProjectIdByChild(item);

            CRef<CWorkspacePrx> ws = m_ProjectService->GetWorkspace();
            CRef<CProjectPrx> project = ws->FindProjectById(prj_id);

            CProjectFolderPrx::TId folder_id = folder_item->GetData();
            CRef<CProjectFolderPrx> folder_prx = project->FindProjectFolderById(folder_id);
            folder_prx->SetOpen(item.IsExpanded());
        }}
        break;
    default:
        break;
    }
}

void CAppExplorerService::OnProperties(CExplorerItem& item)
{
    TGuard guard(*this);

    EItemType type = GetItemType(item);
    switch(type)    {
    case eProjectFolder:
        x_OnProjectFolderProperties(item);
        break;

    case eWorkspaceFolder:
        x_OnWorkspaceFolderProperties(item);
        break;

    case eWorkspace:
        x_OnWorkspaceProperties();
        break;

    case eProject:
        x_OnProjectProperties(item);
        break;

    case eProjectItem:
        x_OnProjectItemProperties(item);
        break;

    default:
        _ASSERT(false); // not supported
        break;
    }
}


void CAppExplorerService::x_OnProjectProperties(CExplorerItem& item)
{
    TProjectTreeItem* prj_item = dynamic_cast<TProjectTreeItem*>(&item);

    _ASSERT(prj_item);

    TProjectId id = prj_item->GetData();
    CRef<CWorkspaceConstPrx> ws = m_ProjectService->GetWorkspaceConst();
    CRef<CProjectPrx> project = ws->FindProjectById(id);

    // lock project in order to modify it
    // setup Project Edit dialog
    CProjectEditDlg dlg(NULL, wxID_ANY, wxT("Project Properties"));
    dlg.SetRegistryPath(m_RegPath + kEditProjectDialogTag);

    dlg.SetProjectName(ToWxString(project->GetName()));
    dlg.SetDescr(ToWxString(project->GetDescr()));
    dlg.SetCreated(ToWxString(project->GetCreateTime().AsString()));
    dlg.SetModified(ToWxString(project->GetModifiedTime().AsString()));
    dlg.SetFilename(project->GetFilename());

    if(dlg.ShowModal() == wxID_OK)  {
        project->SetName(ToStdString(dlg.GetProjectName()));
        project->SetDescr(ToStdString(dlg.GetDescr()));
    }
}


void CAppExplorerService::x_OnWorkspaceProperties()
{
    // lock the service to get Workspace
    CRef<CWorkspacePrx> ws = m_ProjectService->GetWorkspace();

    CProjectEditDlg dlg(NULL, wxID_ANY, wxT("Workspace Properties"));
    dlg.SetRegistryPath(m_RegPath + kEditProjectDialogTag);

    dlg.SetProjectName(ToWxString(ws->GetName()));
    dlg.SetDescr(ToWxString(ws->GetDescr()));
    dlg.SetCreated(ToWxString(ws->GetCreateTime().AsString()));
    dlg.SetModified(ToWxString(ws->GetModifiedTime().AsString()));
    dlg.SetFilename(ws->GetFilename());

    if(dlg.ShowModal() == wxID_OK)  {
        ws->SetName(ToStdString(dlg.GetProjectName()));
        ws->SetDescr(ToStdString(dlg.GetDescr()));
        ws->SetModifiedTime(CTime(CTime::eCurrent));

        CRef<CWorkspaceFolderPrx> ws_folder = ws->GetRootFolder();
        ws_folder->SetName(ws->GetName());
    }
}


void CAppExplorerService::x_OnProjectItemProperties(CExplorerItem& item)
{
    CRef<CWorkspacePrx> ws = m_ProjectService->GetWorkspace();

    // get the Project and the Item proxies
    TProjectItemTreeItem* tree_item = dynamic_cast<TProjectItemTreeItem*>(&item);
    _ASSERT(tree_item);

    TProjectId prj_id = GetProjectIdByChild(item);
    CRef<CProjectPrx> project = ws->FindProjectById(prj_id);

    TPrjItemId item_id = tree_item->GetData();
    CRef<CProjectItemPrx> pr_item = project->FindProjectItemById(item_id);

    // get the parent item and check whether it is a Folder
    CExplorerItem* parent_item = item.GetParent();
    TProjectFolderTreeItem* folder_parent_item = dynamic_cast<TProjectFolderTreeItem*>(parent_item);

    _ASSERT(folder_parent_item);

    CProjectFolderPrx::TId parent_id = folder_parent_item->GetData();
    CRef<CProjectFolderPrx> pr_parent_folder = project->FindProjectFolderById(parent_id);

    string name = pr_item->GetName();
    string descr = pr_item->GetDescr();
    // TODO this needs to be redesign - we need to show different dialogs
    // for different data types
    /*CRef<CAnnotdesc> desc_date;
    CRef<CAnnotdesc> desc_comment;
    if (pr_item.CanGetDescr()) {
        for ( CProjectItem::TDescr::iterator it = pr_item.SetDescr().begin();
            it != pr_item.SetDescr().end();  ++it) {
                //if ( (*it)->IsCreate_date() ) desc_date    = *it;
                if ( (*it)->IsComment() )     desc_comment = *it;
        }
    }
    if (!desc_date.Empty()){
        param.m_CreateDate = desc_date->GetCreate_date().AsCTime();
    }
    if (!desc_comment.Empty()){
        descr = desc_comment->GetComment();
    }*/

    CFolderEditDlg dlg(NULL, wxID_ANY, wxT("Project Item Properties"));
    dlg.SetRegistryPath(m_RegPath + kEditProjectItemDialogTag);

    dlg.SetFolderName(ToWxString(name));
    dlg.SetDescr(wxString::FromUTF8(descr.c_str()));
    dlg.SetDataObject(pr_item->GetObject(), project->GetScope());

    CPrjItemNameValidator validator(*pr_parent_folder, name);
    dlg.SetNameValidator(&validator);

    if(dlg.ShowModal() == wxID_OK)  {
        pr_item->SetName(ToStdString(dlg.GetFolderName()));
        pr_item->SetDescr(ToStdString(dlg.GetDescr()));
        /*if (desc_comment.Empty()) {
            desc_comment.Reset(new CAnnotdesc());
            pr_item.SetDescr().push_back(desc_comment);
        }
        desc_comment->SetComment(descr);*/
    }
}


void CAppExplorerService::x_OnProjectFolderProperties(CExplorerItem& item)
{
    TProjectFolderTreeItem* folder_item = dynamic_cast<TProjectFolderTreeItem*>(&item);
    _ASSERT(folder_item);

    TProjectId prj_id = GetProjectIdByChild(item);

    CRef<CWorkspacePrx> ws = m_ProjectService->GetWorkspace();
    CRef<CProjectPrx> project = ws->FindProjectById(prj_id);

    CProjectFolderPrx::TId folder_id = folder_item->GetData();
    CRef<CProjectFolderPrx> folder_prx = project->FindProjectFolderById(folder_id);

    CExplorerItem* parent_item = item.GetParent();
    TProjectFolderTreeItem* folder_parent_item =
        dynamic_cast<TProjectFolderTreeItem*>(parent_item);

    CRef<CProjectFolderPrx> parent_folder;
    if( folder_parent_item)  {
        CProjectFolderPrx::TId parent_id = folder_parent_item->GetData();
        parent_folder = project->FindProjectFolderById(parent_id);
    }

    stFolderInfo info(folder_prx->GetName(), folder_prx->GetDescr());

    string reg_path = m_RegPath + kEditFolderDialogTag;
    CFolderNameValidator<CProjectFolderPrx> validator(parent_folder, info.m_Name);

    // show the dialog
    bool readonly = (parent_folder == NULL);
    if(x_EditFolder(info, "Edit Folder", reg_path, readonly, validator) && !readonly)  {
        folder_prx->SetName(info.m_Name);
        folder_prx->SetDescr(info.m_Descr);
        //m_ProjectService->OnProjectChanged(*project); //TODO
    }
}


void CAppExplorerService::x_OnWorkspaceFolderProperties(CExplorerItem& item)
{
    // get the parent item and check whether is is a Folder
    CExplorerItem* parent_item = item.GetParent();
    TWsFolderTreeItem* folder_parent_item = dynamic_cast<TWsFolderTreeItem*>(parent_item);

    CRef<CWorkspacePrx> ws = m_ProjectService->GetWorkspace();
    CRef<CWorkspaceFolderPrx> parent_folder;
    if(folder_parent_item)  {
        TWsFolderId parent_id = folder_parent_item->GetData();
        parent_folder = ws->FindWorkspaceFolderById(parent_id);
    }

    TWsFolderTreeItem* folder_item = dynamic_cast<TWsFolderTreeItem*>(&item);

    _ASSERT(folder_item);

    TWsFolderId id = folder_item->GetData();
    CRef<CWorkspaceFolderPrx> folder = ws->FindWorkspaceFolderById(id);
    stFolderInfo info(folder->GetName(), folder->GetDescr());

    string reg_path = m_RegPath + kEditFolderDialogTag;
    CFolderNameValidator<CWorkspaceFolderPrx> validator(parent_folder, info.m_Name);

    // show the dialog
    bool readonly = false;//(parent_folder == NULL);
    if(x_EditFolder(info, "Edit Folder", reg_path, readonly, validator))  {
        folder->SetName(info.m_Name);
        folder->SetDescr(info.m_Descr);
    }
}


bool CAppExplorerService::IsPropertiesCmdEnabled(CExplorerItem& item)
{
    EItemType type = GetItemType(item);

    switch(type)    {
    case eWorkspace:
    case eWorkspaceFolder:
    case eProjectItem:
    case eProjectFolder:
        return true;
    case eProject:  {{
        TProjectTreeItem* prj_item = dynamic_cast<TProjectTreeItem*>(&item);

        TProjectServiceGuard srv_loc(*m_ProjectService);

        CRef<CWorkspaceConstPrx> workspace = m_ProjectService->GetWorkspaceConst();
        CProjectPrx::TId prj_id = prj_item->GetData();
        CRef<CProjectConstPrx> project = workspace->FindProjectById_C(prj_id);

        if(project) {
            return project->IsLoaded();
        }
        return false;
    }}
    case eFolder:
    case eView:
        return false;
    case eDataSource:
    case eDataLoader:
        return false; // TODO - support the command
    default:
        _ASSERT(false);
        return false;
    }
}


void CAppExplorerService::Remove(TItemRefVector& items)
{
    x_ClearClipboardData();

     // lock itself, Project Service and Workspace
    TGuard guard(*this);
    TProjectServiceGuard pr_guard(*m_ProjectService);

    TItemRefVector top_items;
    x_GetTopItems(items, top_items);

    NON_CONST_ITERATE(TItemRefVector, it, top_items)  {
        try {
            CExplorerItem& item = **it;
            EItemType type = GetItemType(item);
            switch(type)    {
            case eProjectFolder:
                x_RemoveProjectFolder(item, false);
                break;
            case eWorkspaceFolder:
                x_RemoveWorkspaceFolder(item, false);
                break;
            case eProjectItem:
                x_RemoveProjectItem(item, false);
                break;
            case eDataLoader:
                x_RemoveDataLoader(item, false);
                break;
            default:
                _ASSERT(false); // not supported
                break;
            }
        }
        catch (CException& e) {
            LOG_POST(Error << "CAppExplorerService::OnRemove(): error removing item: " << e.GetMsg());
        }
    }
}

// builds a set that contains only items that are not children of the item
// in the given "items" container
void CAppExplorerService::x_GetTopItems(TItemRefVector& items, TItemRefVector& top_items)
{
    // build an index for fast look-up
    set<CExplorerItem*> items_set;
    for(  size_t i = 0; i < items.size();  i++ )    {
        CExplorerItem* item = items[i].GetPointer();
        items_set.insert(item);
    }

    for(  size_t i = 0; i < items.size();  i++ )    {
        bool exclude = false;
        CExplorerItem* item = items[i].GetPointer();
        CExplorerItem* parent = item->GetParent();
        while(parent)   {
            set<CExplorerItem*>::iterator it_p = items_set.find(parent);
            if(it_p != items_set.end()) {
                exclude = true;
                break;
            }
            parent = parent->GetParent();
        }
        if( ! exclude)  {
            top_items.push_back(items[i]);
        }
    }
}

bool CAppExplorerService::x_RemoveWorkspaceFolder(CExplorerItem& item, bool test)
{
    TWsFolderTreeItem* folder_item = dynamic_cast<TWsFolderTreeItem*>(&item);
    if(folder_item) {
        CExplorerItem* parent_item = folder_item->GetParent();

        CRef<CWorkspacePrx> ws = m_ProjectService->GetWorkspace();
        CRef<CWorkspaceFolderPrx> parent_folder;

        TWsFolderTreeItem* par_fl_item = dynamic_cast<TWsFolderTreeItem*>(parent_item);
        if(par_fl_item) {
            TWsFolderId parent_id = par_fl_item->GetData();
            parent_folder = ws->FindWorkspaceFolderById(parent_id);
        } else {
            _ASSERT(GetItemType(*parent_item) == eWorkspace);
            parent_folder = ws->GetRootFolder();
        }

        if(parent_folder) {
            if( ! test)  {
                // delete folder from the parent folder
                TWsFolderId id = folder_item->GetData();
                parent_folder->RemoveChildFolder(id);
            }
            return true;
        }
    }
    return false;
}


bool CAppExplorerService::x_RemoveProjectFolder(CExplorerItem& item, bool test)
{
    TProjectFolderTreeItem* folder_item = dynamic_cast<TProjectFolderTreeItem*>(&item);

    if(folder_item) {
        // get the parent item
        CExplorerItem* parent_item = folder_item->GetParent();
        TProjectFolderTreeItem* par_fl_item = dynamic_cast<TProjectFolderTreeItem*>(parent_item);

        if(par_fl_item) {
            if( ! test)  {
                CRef<CWorkspacePrx> ws = m_ProjectService->GetWorkspace();

                TProjectId prj_id = GetProjectIdByChild(*folder_item);
                CRef<CProjectPrx> project = ws->FindProjectById(prj_id);

                CProjectFolderPrx::TId parent_id = par_fl_item->GetData();
                CRef<CProjectFolderPrx> par_folder = project->FindProjectFolderById(parent_id);

                CProjectFolderPrx::TId folder_id = folder_item->GetData();
                par_folder->RemoveChildFolder(folder_id, true);
            }
            return true;
        }
    }
    return false;
}


bool CAppExplorerService::x_RemoveProjectItem(CExplorerItem& item, bool test)
{
    TProjectItemTreeItem* prj_item = dynamic_cast<TProjectItemTreeItem*>(&item);

    CRef<CWorkspacePrx> ws = m_ProjectService->GetWorkspace();
    if (prj_item) {
        if (test) {
            CExplorerItem* itm = &item;
            while ( (itm = itm->GetParent()) != NULL) {
                if (itm->GetLabel() == sm_DATA_FOLDER) {
                    return true;
                }
            }
            return false;
        }
        else {
            CExplorerItem* parent_item = item.GetParent();
            TProjectFolderTreeItem* par_fl_item = dynamic_cast<TProjectFolderTreeItem*>(parent_item);

            _ASSERT(par_fl_item);

            CProjectFolderPrx::TId par_fl_id = par_fl_item->GetData();

            TProjectId prj_id = GetProjectIdByChild(item);
            CRef<CProjectPrx> project = ws->FindProjectById(prj_id);

            CRef<CProjectFolderPrx> par_folder = project->FindProjectFolderById(par_fl_id);
            TPrjItemId item_id = prj_item->GetData();
            par_folder->RemoveChildItem(item_id);
        }
        return true;
    }
    return false;
}


bool CAppExplorerService::x_RemoveDataLoader(CExplorerItem& item, bool test)
{
    TDataLoaderTreeItem* loader_item = dynamic_cast<TDataLoaderTreeItem*>(&item);

    CRef<CWorkspacePrx> ws = m_ProjectService->GetWorkspace();
    if (loader_item) {
        TProjectId prj_id = GetProjectIdByChild(item);
        CRef<CProjectPrx> project = ws->FindProjectById(prj_id);

        if (test) {
            return project->GetViews().empty();
        } else {

            const string& label = loader_item->GetData();
            CRef<CLoaderDescriptor> loader = project->FindDataLoaderByLabel(label);
            if (loader) {
                if ( !project->RemoveDataLoader(*loader) ) {
                    NcbiMessageBoxW(wxT("Sorry, Data Loaders can't be removed!\n")
                                    wxT("Please close the corresponding views and try again!"));
                }
            }
        }
        return true;
    }
    return false;
}


CAppExplorerDataObject*
    CAppExplorerService::CreateDataObject(TItemRefVector& items, bool dnd, bool cut)
{
/*
    wxString text;
    if( !dnd ){
        // generate text for clipboard; use all items
        wxStringOutputStream strstr;
        wxTextOutputStream os(strstr);
        for( size_t  i = 0; i < items.size(); i++ ){
            if( i > 0 ){
                os << endl;
            }
            CExplorerItem& item = *items[i];
            os << item.GetLabel();

        }
        text = strstr.GetString();
    }
 */

    // select only top items
    TItemRefVector top_items;
    x_GetTopItems( items, top_items );

    // create new data descriptor that will hold actual data
    CAppExplorerData* data = new CAppExplorerData( top_items, cut );
    CAppExplorerDataObject* data_obj = new CAppExplorerDataObject( *data );

    return data_obj;
}


void CAppExplorerService::CutOrCopyToClipboard(TItemRefVector& items, bool cut)
{
    if(items.size())    {
        x_ClearClipboardData();

        CAppExplorerDataObject* data_object = CreateDataObject(items, false, cut);

        // hold a pointer to the Descr, so that we can invalidate it later
        m_ClipboardData = data_object->GetDataRef();

        wxClipboardLocker clipLocker;
        wxTheClipboard->AddData(data_object);
    }
}


void CAppExplorerService::PasteFromClipboard(CExplorerItem& target)
{
    wxClipboardLocker clipLocker;    // lock Clipboard

    // check the clipboard
    CAppExplorerDataObject data_object;
    if ( wxTheClipboard->GetData(data_object) )     {
        CRef<CAppExplorerData> data = data_object.GetDataRef();

        if( data && data->IsValid() ){
            TItemRefVector items;
            data->GetItems( items );
            bool cut = data->IsCut();

            x_Paste( items, target, cut );
        }
    }
}


// handleds D&D Drop
void CAppExplorerService::Drop(
    CExplorerItem& target, CAppExplorerData& data, bool move 
){
    _ASSERT( data.IsValid() );
    _ASSERT( CanDropTo(target) );

    CExplorerItem* real_target =
        (GetItemType(target) == eProjectFolder) ? &target : target.GetParent()
    ;

    TItemRefVector items;
    data.GetItems(items);

    x_Paste( items, *real_target, move );
}


bool CAppExplorerService::CanPasteTo(CExplorerItem& target)
{
    // we can only paste to Project Folders
    return GetItemType(target) == eProjectFolder;
}


bool CAppExplorerService::CanDropTo(CExplorerItem& target)
{
    // can drop into a Project Folder or between the Project Items (into their parent)
    const int kGoodTypes = (eProjectFolder | eProjectItem);
    return (GetItemType(target) & kGoodTypes) != 0;
}


bool CAppExplorerService::CanCutOrCopyToClipboard(TItemRefVector& items)
{
    const int kGoodTypes = (eProjectFolder | eProjectItem);
    int types = GetItemTypes(items);
    bool has_good = (types & kGoodTypes) != 0;
    bool has_others = (types & ~kGoodTypes) != 0;
    return has_good  &&  ! has_others;
}


void CAppExplorerService::x_ClearClipboardData()
{
    if(m_ClipboardData) {
        // make it unavailable for Data Objects that can refer to it
        m_ClipboardData->Invalidate();

        // disconnect
        m_ClipboardData.Reset();
    }
}


void CAppExplorerService::x_Paste(TItemRefVector& items, CExplorerItem& target, bool cut)
{
    TGuard exp_guard(*this); // lock itself
    CUpdateGuard<CProjectService> up_lock(*m_ProjectService); // accumulate all updates
    TProjectServiceGuard    prj_guard(*m_ProjectService); // lock Project Service

    // get the destination Folder and Project
    TProjectFolderTreeItem* dst_item = dynamic_cast<TProjectFolderTreeItem*>(&target);
    _ASSERT(dst_item);

    CRef<CWorkspacePrx> ws = m_ProjectService->GetWorkspace();

    TProjectId prj_id = GetProjectIdByChild(target);
    CRef<CProjectPrx> dst_project = ws->FindProjectById(prj_id);

    CProjectFolderPrx::TId folder_id = dst_item->GetData();
    CRef<CProjectFolderPrx> dst_folder = dst_project->FindProjectFolderById(folder_id);

    map< TProjectId, CUpdateGuard<CGBDocument>* > guard_map;

    guard_map[prj_id] = new CUpdateGuard<CGBDocument>( *dst_project->GetDocument() );

    // TODO here we can get a deadlock on projects
    // for all items in the Clipboard
    for(  size_t  i = 0; i < items.size();  i++)    {
        CExplorerItem& item = *items[i];

        TProjectId src_prj_id = GetProjectIdByChild(item);
        if( guard_map.find( src_prj_id ) == guard_map.end() ){
            CRef<CProjectPrx> src_project = ws->FindProjectById(src_prj_id);
            guard_map[src_prj_id] = 
                new CUpdateGuard<CGBDocument>( *src_project->GetDocument() )
            ;
        }

        x_PasteItem(item, *dst_project, *dst_folder, cut);
    }

    for(
        map< TProjectId, CUpdateGuard<CGBDocument>* >::reverse_iterator iter = guard_map.rbegin();
        iter != guard_map.rend();
        iter++
    ){
        delete iter->second;
        iter->second = NULL;
    }
    guard_map.clear();
}


bool CAppExplorerService::x_PasteItem(CExplorerItem& item,
                                      CProjectPrx& dst_project,
                                      CProjectFolderPrx& dst_folder,
                                      bool cut )
{
    TProjectFolderTreeItem* folder_item = dynamic_cast<TProjectFolderTreeItem*>(&item);
    TProjectItemTreeItem* prj_item = dynamic_cast<TProjectItemTreeItem*>(&item);

    CRef<CWorkspacePrx> ws = m_ProjectService->GetWorkspace();

    TProjectId src_prj_id = GetProjectIdByChild(item);
    CRef<CProjectPrx> src_project = ws->FindProjectById(src_prj_id);

    if( folder_item ){
        // paste Project Folder
        CProjectFolderPrx::TId folder_id = folder_item->GetData();
        CRef<CProjectFolderPrx> folder = src_project->FindProjectFolderById(folder_id);
        string err;

        if(x_CanPasteProjectFolder(dst_folder, *folder, err))  {
            if( src_project->GetId() == dst_project.GetId() && cut ){
                if( dst_folder.GetId() != folder->GetId() ){
                    // move within the project
                    x_MoveProjectFolder(dst_project, dst_folder, *folder_item);
                }
            } else {
                x_PasteProjectFolder(dst_folder, *folder);
                if (cut)  {
                    x_RemoveProjectFolder(*folder_item, false);
                }
            }
            return true;
        } else {
            NcbiErrorBox(err);
        }
    } else if (prj_item) {
        // paste Project Item
        TPrjItemId item_id = prj_item->GetData();
        CRef<CProjectItemPrx> project_item = src_project->FindProjectItemById(item_id);

        if(src_project->GetId() == dst_project.GetId()  &&  cut)  {
            // move within the project
            src_project->MoveProjectItem(*project_item, dst_folder);
        } else {
            x_PasteProjectItem(dst_folder, *project_item);
            if (cut){
                x_RemoveProjectItem(item, false);
            }
        }
        return true;
    }
    return false;
}


bool CAppExplorerService::x_CanPasteProjectFolder(CProjectFolderPrx& parent,
                                                  CProjectFolderPrx& child,
                                                  string& error)
{
    // cannot paste if parent is a sub folder of child
    // iterate on subfolder of "child"
    CProjectFolderPrx::TId parent_id = parent.GetId();
    CRef<CProjectFolderPrx> res = child.FindChildFolderById(parent_id, true);
    if(res) {
        error = "Cannot paste folder \"" + child.GetName();
        error += "\". Destination folder is a child of the source folder.";
        return false;
    }
    return true;
}


void CAppExplorerService::x_PasteProjectFolder(CProjectFolderPrx& parent, CProjectFolderPrx& child)
{
    string name = child.GetName();
    string descr = child.GetDescr();
    CRef<CProjectFolderPrx> new_folder = parent.AddChildFolder(name, descr); // TODO make unique folder label

    // paste all child items
    vector<CRef<CProjectItemPrx> > items;
    child.GetProjectItems(items);

    for(  size_t i = 0; i < items.size();  i++ )    {
        CProjectItemPrx& item = *items[i];
        x_PasteProjectItem(*new_folder, item);
    }

    // paste all child folders recursively
    vector<CRef<CProjectFolderPrx> > sub_folders;
    child.GetChildFolders(sub_folders);

    for(  size_t i = 0; i < sub_folders.size();  i++ )    {
        CProjectFolderPrx& sub_folder = *sub_folders[i];
        x_PasteProjectFolder(*new_folder, sub_folder);
    }
}


// moves "child_item" folder into "parent" folder (within the same project)
void CAppExplorerService::x_MoveProjectFolder(CProjectPrx& project,
                                              CProjectFolderPrx& parent,
                                              TProjectFolderTreeItem& child_item )
{
    CProjectFolderPrx::TId folder_id = child_item.GetData();
    CRef<CProjectFolderPrx> child_folder = project.FindProjectFolderById(folder_id);

    project.MoveProjectFolder(*child_folder, parent);
}


void CAppExplorerService::x_PasteProjectItem(CProjectFolderPrx& parent_folder,
                                             CProjectItemPrx& item)
{
    string name = item.GetName();
    string descr = item.GetDescr();
    CSerialObject* obj = const_cast<CSerialObject*>(item.GetObject());

    parent_folder.AddProjectItem(name, *obj, descr);
    //TODO modification / creation date
}


// toggle "enabled" flag for project items
void CAppExplorerService::EnableDisableProjectItems(TItemRefVector& items)
{
    CRef<CWorkspacePrx> ws = m_ProjectService->GetWorkspace();

    NON_CONST_ITERATE (TItemRefVector, it, items) {
        CExplorerItem& item = **it;

        CProjectPrx::TId prj_id = GetProjectIdByChild(item);
        CRef<CProjectPrx> project = ws->FindProjectById(prj_id);
        if ( !project ) continue;

        EItemType type = GetItemType(item);
        if (type == eProjectItem) {
            TProjectItemTreeItem* project_item = dynamic_cast<TProjectItemTreeItem*>(&item);
            _ASSERT(project_item);
            
            CProjectItemPrx::TId item_id = project_item->GetData();
            CRef<CProjectItemPrx> item_prx = project->FindProjectItemById(item_id);
            bool en = ! item_prx->IsEnabled();
            item_prx->Enable(en);
        } else if (type == eDataLoader) {
            TDataLoaderTreeItem* loader_item = dynamic_cast<TDataLoaderTreeItem*>(&item);
            _ASSERT(loader_item);
            
            const string& label = loader_item->GetData();
            if ( !project->EnableDisableDataLoader(label) ) {
                // this should happen only when the disabling data loader
                wxString msg = wxT("Sorry, Data Loaders can't be ");
                if ( !loader_item->IsEnabled() ) {
                    msg += wxT("enabled!");
                } else {
                    msg += wxT("disabled!");
                }
                msg += wxT("\nPlease close the corresponding views and try again!");
                NcbiMessageBoxW(msg);
            }
        }
    }
}


// returns true if Load / Unload command can be performed on the given set of items
// all items must represent Projects and at least one must allow the operation
bool CAppExplorerService::CanLoadUnloadProjects(TItemRefVector& items, bool load)
{
    CRef<CWorkspaceConstPrx> workspace = m_ProjectService->GetWorkspaceConst();

    NON_CONST_ITERATE (TItemRefVector, it, items) {
        CExplorerItem& item = **it;
        TProjectTreeItem* project_item = dynamic_cast<TProjectTreeItem*>(&item);

        if(project_item)    {
            CProjectPrx::TId prj_id = project_item->GetData();
            CRef<CProjectConstPrx> project = workspace->FindProjectById_C(prj_id);

            if(!project->IsLoading() && project->IsLoaded() != load)  {
                return true; // found a project on which we can perform the operation
            }
        } else {
            return false; // not a Project
        }
    }
    return false;
}


END_NCBI_SCOPE
