/*  $Id: asn_object_loader.cpp 27062 2012-12-13 16:32:50Z katargir $
* ===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data,  the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties,  express or implied,  including
*  warranties of performance,  merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================
*
* Authors:  Roman Katargin
*/


#include <ncbi_pch.hpp>

#include <gui/core/asn_object_loader.hpp>

#include <util/icanceled.hpp>

#include <gui/core/asn_load_extension.hpp>
#include <gui/utils/extension_impl.hpp>
#include <gui/objutils/label.hpp>

#include <util/format_guess.hpp>
#include <objmgr/util/obj_sniff.hpp>
#include <objmgr/object_manager.hpp>

#include <objects/general/Dbtag.hpp>
#include <objects/general/Object_id.hpp>
#include <objects/seqfeat/Imp_feat.hpp>
#include <objects/seqfeat/Gb_qual.hpp>

#include <objects/biotree/BioTreeContainer.hpp>
#include <objects/biotree/DistanceMatrix.hpp>
#include <objects/entrezgene/Entrezgene.hpp>
#include <objects/entrezgene/Entrezgene_Set.hpp>
#include <objects/pubmed/Pubmed_entry.hpp>

#include <objects/seqalign/Seq_align_set.hpp>
#include <objects/seqalign/Dense_seg.hpp>
#include <objects/seqalign/Std_seg.hpp>
#include <objects/submit/Seq_submit.hpp>

#include <objects/seq/NCBI_Sequence_module.hpp>
#include <objects/seqloc/NCBI_Seqloc_module.hpp>
#include <objects/seqalign/NCBI_Seqalign_module.hpp>
#include <objects/seqfeat/NCBI_Seqfeat_module.hpp>
#include <objects/seqset/NCBI_Seqset_module.hpp>
#include <objects/submit/NCBI_Submit_module.hpp>
#include <objects/biotree/NCBI_BioTree_module.hpp>
#include <objects/entrezgene/NCBI_Entrezgene_module.hpp>
#include <objects/pubmed/NCBI_PubMed_module.hpp>
#include <objects/seqtable/Seq_table.hpp>
#include <objects/genomecoll/GC_Assembly.hpp>
#include <gui/objects/NCBI_GUI_Project_module.hpp>
#include <gui/objects/ProjectItem.hpp>

#include <serial/serial.hpp>
#include <serial/pack_string.hpp>

#include <wx/filename.h>

#include <gui/widgets/wx/compressed_file.hpp>
#include <gui/widgets/wx/wx_utils.hpp>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

///
/// Class CSniffReader reads NCBI object files in number of different formats.
/// Uses try and fail deserialization for binary ASN.1 files
///
class CSniffReader : public CObjectsSniffer, public IAsnLoadRegister
{
public:
    CSniffReader(CAsnObjectLoader& loader);

    // Event function called when parser finds a top level object
    virtual void OnTopObjectFoundPre(const CObjectInfo& object,
        CNcbiStreampos stream_pos);

    // Event function alled after top object deserialization
    virtual void OnTopObjectFoundPost(const CObjectInfo& object);

    // Overload from CObjectsSniffer
    virtual void OnObjectFoundPre(const CObjectInfo& object,
        CNcbiStreampos stream_pos);

    // Overload from CObjectsSniffer
    virtual void OnObjectFoundPost(const CObjectInfo& object);

    virtual void Reset();

    bool ObjectLoaded() const { return m_ObjectLoaded; }

/// IAsnLoadRegister implementation
    virtual void Register(CObjectTypeInfo ti);

private:
    unsigned int                 m_ObjLevel;   // Object level counter, used to identify
    // the TSE

    CAsnObjectLoader& m_Loader;
    bool m_ObjectLoaded;
};

CAsnObjectLoader::CAsnObjectLoader(const vector<wxString>& filenames)
 : m_FileNames(filenames)
{
}

IObjectLoader::TObjects& CAsnObjectLoader::GetObjects()
{
    return m_Objects;
}

string CAsnObjectLoader::GetDescription() const
{
    return "Loading NCBI ASN.1 Files";
}

bool CAsnObjectLoader::PreExecute()
{
    return true;
}

bool CAsnObjectLoader::Execute(ICanceled& canceled)
{
    CRef<CObjectManager> obj_mgr = CObjectManager::GetInstance();
    m_Scope.Reset(new CScope(*obj_mgr));
    m_Scope->AddDefaults();

    ITERATE(vector<wxString>, it, m_FileNames) {
        if (canceled.IsCanceled())
            return false;

        try {
            m_CurrentFileName = *it;
            const wxString& fn = *it;

            CCompressedFile file(fn);
            ESerialDataFormat sfmt = eSerial_None;
            CFormatGuess::EFormat fmt = CFormatGuess(file.GetIstream()).GuessFormat();

            switch (fmt) {
            case CFormatGuess::eBinaryASN:
                sfmt = eSerial_AsnBinary;
                break;
            case CFormatGuess::eTextASN:
                sfmt = eSerial_AsnText;
                break;
            case CFormatGuess::eXml:
                sfmt = eSerial_Xml;
                break;

            default:
                {{
                    string msg("Serial format could not be determined:\n");
                    msg += it->ToUTF8();
                    LOG_POST(Info << msg);
                }}
                break;
            }

            if (sfmt == eSerial_None)
                continue;

            file.Reset();

            // use object sniffer to guess
            CSniffReader sniffer(*this);
            auto_ptr<CObjectIStream>
                sinput(CObjectIStream::Open(sfmt, file.GetIstream()));

            // memory profile optimization:
            // we should use pooled strings to reduce the memory burden
            // (from Eugene Vasilchenko)
            //
            CObjectTypeInfo info;
            info = CObjectTypeInfo(CType<CObject_id>());
            info.FindVariant("str")
                .SetLocalReadHook(*sinput, new CPackStringChoiceHook);

            info = CObjectTypeInfo(CType<CImp_feat>());
            info.FindMember("key")
                .SetLocalReadHook(*sinput, new CPackStringClassHook(32, 128));

            info = CObjectTypeInfo(CType<CDbtag>());
            info.FindMember("db")
                .SetLocalReadHook(*sinput, new CPackStringClassHook);

            info = CType<CGb_qual>();
            info.FindMember("qual")
                .SetLocalReadHook(*sinput, new CPackStringClassHook);

            sinput->UseMemoryPool();

            // now, probe!
            //

            sniffer.AddCandidate(ncbi::objects::CSeq_table::GetTypeInfo());

            sniffer.Probe(*sinput);
            if (!sniffer.ObjectLoaded()) {
                string msg("No Top objects found:\n");
                msg += ToStdString(*it);
                LOG_POST(Info << msg);
            }
        } // ITERATE
        catch (const CException&) {
        }
        catch (const exception&) {
        }
        m_CurrentFileName.clear();
    }

    return true;
}

bool CAsnObjectLoader::x_OnTopObjectFound(const CObjectInfo& objInfo)
{
    if (objInfo.GetTypeFamily() == eTypeFamilyClass ||
        objInfo.GetTypeFamily() == eTypeFamilyChoice) {

        CSerialObject* so = static_cast<CSerialObject*>(objInfo.GetObjectPtr());

        CBioTreeContainer* btrc = dynamic_cast<CBioTreeContainer*>(so);
        if (btrc) {
            // if bio-tree label is not set explicitly -- we assume the filename
            if (!btrc->IsSetLabel()) {
                wxString fbase;
                wxFileName::SplitPath(m_CurrentFileName, 0, &fbase, 0);
                btrc->SetLabel(ToStdString(fbase));
            }
        }

        string label;
        CLabel::GetLabel(*so, &label, CLabel::eDefault, m_Scope);

        if (label.empty()) {
            if (btrc && btrc->IsSetLabel()) {
                label = btrc->GetLabel();
            }
        }

        m_Objects.push_back(SObject(*so, label));
        return true;
    }

    return false;
}

bool CAsnObjectLoader::PostExecute()
{
    return true;
}

/////////////////////////////////////////////////////////////////////////////


CSniffReader::CSniffReader(CAsnObjectLoader& loader)
: m_ObjLevel(0), m_Loader(loader), m_ObjectLoaded(false)
{
    /// basic top-level data model types:
    AddCandidate(CObjectTypeInfo(CType<CSeq_entry>()));
    AddCandidate(CObjectTypeInfo(CType<CBioseq>()));
    AddCandidate(CObjectTypeInfo(CType<CBioseq_set>()));
    AddCandidate(CObjectTypeInfo(CType<CSeq_submit>()));

    /// annotations:
    AddCandidate(CObjectTypeInfo(CType<CSeq_annot>()));

    /// alignments:
    AddCandidate(CObjectTypeInfo(CType<CSeq_align>()));
    AddCandidate(CObjectTypeInfo(CType<CSeq_align_set>()));
    AddCandidate(CObjectTypeInfo(CType<CDense_seg>()));
    AddCandidate(CObjectTypeInfo(CType<CStd_seg>()));

    /// locations/identifiers:
    AddCandidate(CObjectTypeInfo(CType<CSeq_loc>()));
    AddCandidate(CObjectTypeInfo(CType<CSeq_id>()));

    /// features:
    AddCandidate(CObjectTypeInfo(CType<CSeq_feat>()));

    /// miscellaneous:
    AddCandidate(CObjectTypeInfo(CType<CBioTreeContainer>()));
    AddCandidate(CObjectTypeInfo(CType<CDistanceMatrix>()));
    AddCandidate(CObjectTypeInfo(CType<CEntrezgene>()));
    AddCandidate(CObjectTypeInfo(CType<CEntrezgene_Set>()));
    AddCandidate(CObjectTypeInfo(CType<CPubmed_entry>()));
    AddCandidate(CObjectTypeInfo(CType<CProjectItem>()));
    AddCandidate(CObjectTypeInfo(CType<CGC_Assembly>()));


    /// add types from certain modules,
    /// being sure not to duplicate what's above (that's trouble)

    /// collect types added above
    set<string> types;
    const TCandidates& cands = GetCandidates();
    ITERATE (TCandidates, cand, cands) {
        types.insert(cand->type_info.GetTypeInfo()->GetName());
    }

    /// register a bunch of modules
    vector<string> modules;

    NCBI_Sequence_RegisterModuleClasses();
    modules.push_back("NCBI-Sequence");

    NCBI_Seqloc_RegisterModuleClasses();
    modules.push_back("NCBI-Seqloc");

    NCBI_Seqalign_RegisterModuleClasses();
    modules.push_back("NCBI-Seqalign");

    NCBI_Seqfeat_RegisterModuleClasses();
    modules.push_back("NCBI-Seqfeat");

    NCBI_Seqset_RegisterModuleClasses();
    modules.push_back("NCBI-Seqset");

    NCBI_Submit_RegisterModuleClasses();
    modules.push_back("NCBI-Submit");

    NCBI_BioTree_RegisterModuleClasses();
    modules.push_back("NCBI-BioTree");

    NCBI_Entrezgene_RegisterModuleClasses();
    modules.push_back("NCBI-Entrezgene");

    NCBI_GUI_Project_RegisterModuleClasses();
    modules.push_back("NCBI-GUI-Project");

    NCBI_PubMed_RegisterModuleClasses();
    modules.push_back("NCBI-PubMed");

    /// add all types in those modules as candidates, avoiding duplication
    ITERATE (vector<string>, module, modules) {
        CClassTypeInfoBase::TRegClasses type_names;
        CClassTypeInfoBase::GetRegisteredClassNames(*module, type_names);
        ITERATE(CClassTypeInfoBase::TRegClasses, type_name, type_names) {
            if (types.find(*type_name) == types.end()) {
                const CTypeInfo* type_info =
                    CClassTypeInfoBase::GetClassInfoByName(*type_name);
                AddCandidate(type_info);
            }
        }
    }

    vector<IAsnLoadExtension*> extensions;
    GetExtensionAsInterface(EXT_POINT__ASN_LOAD_EXTENSION, extensions);

    ITERATE(vector<IAsnLoadExtension*>, it, extensions)
        (*it)->RegisterAsnTypes(*this);
}

void CSniffReader::Register(CObjectTypeInfo ti)
{
    AddCandidate(ti);
}


void CSniffReader::Reset()
{
    m_ObjLevel = 0;
}

void CSniffReader::OnObjectFoundPre(const CObjectInfo& object,
                                    CNcbiStreampos stream_pos)
{
    if (m_ObjLevel == 0) {
        OnTopObjectFoundPre(object, stream_pos);
    }
    ++m_ObjLevel;
}


void CSniffReader::OnObjectFoundPost(const CObjectInfo& object)
{
    _ASSERT(m_ObjLevel > 0);
    if (m_ObjLevel == 1) {
        OnTopObjectFoundPost(object);
    }
    --m_ObjLevel;
}


void CSniffReader::OnTopObjectFoundPre(const CObjectInfo& /*object*/,
                                       CNcbiStreampos /*stream_pos*/)
{
}


void CSniffReader::OnTopObjectFoundPost(const CObjectInfo& object)
{
    m_ObjectLoaded = (m_Loader.x_OnTopObjectFound(object) || m_ObjectLoaded);
}

END_NCBI_SCOPE
