/*  $Id: assm_info.cpp 28151 2013-05-16 19:56:13Z katargir $
* ===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data, the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties, express or implied, including
*  warranties of performance, merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================
*
* Author:  Liangshou Wu
*
* File Description:
*
*/

#include <ncbi_pch.hpp>
#include <corelib/ncbidiag.hpp>
#include <gui/core/assm_info.hpp>
#include <objmgr/util/sequence.hpp>
#include <misc/xmlwrapp/xmlwrapp.hpp>
#include <objtools/eutils/api/elink.hpp>
#include <objtools/eutils/api/esummary.hpp>
#include <objtools/eutils/esummary/DocSum.hpp>
#include <objtools/eutils/esummary/Item.hpp>
#include <objtools/eutils/api/esearch.hpp>
#include <objtools/eutils/elink/Link.hpp>
#include <objtools/eutils/elink/ELinkResult.hpp>
#include <objtools/eutils/elink/Id.hpp>
#include <objtools/eutils/elink/LinkInfo.hpp>
#include <objtools/eutils/elink/LinkSet.hpp>
#include <objtools/eutils/elink/LinkSetDbHistory.hpp>
#include <objtools/eutils/elink/IdList.hpp>
#include <objtools/eutils/elink/Id.hpp>
#include <objtools/eutils/elink/Id.hpp>
#include <objtools/eutils/esearch/ESearchResult.hpp>
#include <objtools/eutils/esearch/TranslationStack.hpp>
#include <objtools/eutils/esearch/TermSet.hpp>
#include <objtools/eutils/esearch/IdList.hpp>
#include <objects/genomecoll/genomic_collections_cli.hpp>

#include <gui/objects/DL_Assembly.hpp>
#include <gui/objects/DL_TaxId.hpp>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);
USING_SCOPE(elink);


static const string kTaxDb = "taxonomy";
static const string kNucDb = "nucleotide";
static const string kAssmDb = "assembly";
static const int kRetMax = 5000;


CRef<CDL_AssmInfo> CAssemblyInfo::GetAssms_Gi(int gi) const
{
    string gc_ids = kEmptyStr;
    x_GetAssmIds_GI(gc_ids, gi);
    if (gc_ids.empty()) {
        NCBI_THROW(CException, eUnknown, "Can't find the associated assembly for given seq-id!");
    }

    // get all assembly info for the list of GC links
    CDL_AssmInfo::TAssemblies assms;
    x_GetAssmInfo(gc_ids, assms);

    return x_CreateAssmInfo(assms);
}


CRef<CDL_AssmInfo> CAssemblyInfo::GetAssms_TaxId(const string& tax_id) const
{
    if ( !x_ValidateTaxId(tax_id) ) {
        string msg = "Invalid tax-id: " + tax_id;
        NCBI_THROW(CException, eUnknown, msg);
    }

    // search all GC links associated with the tax-id
    string gc_ids = kEmptyStr;
    x_GetAssmIds_TaxId(gc_ids, tax_id);
    if (gc_ids.empty()) {
        string msg = "Can't find the associated assembly for tax-id: " + tax_id;
        NCBI_THROW(CException, eUnknown, msg);
    }

    // get all assembly info for the list of GC links
    CDL_AssmInfo::TAssemblies assms;
    x_GetAssmInfo(gc_ids, assms);

    return x_CreateAssmInfo(assms);
}


CRef<CDL_AssmInfo> CAssemblyInfo::GetAssms_Term(const string& term) const
{
    try {
        if (NStr::StringToInt(term) > 0) {
            // maybe, the term stores a tax-id
            return GetAssms_TaxId(term);
        }
    } catch (CException&) {
        // failed! It may not be a tax-id
    }

    CRef<CDL_AssmInfo> assm_info;

    // Search the tax-id using the search term
    TIds ids;
    x_SearchIds(term, kTaxDb,  ids);
    if (ids.empty()) {
        // Let try esearch in gencoll db directly.
        // This will work for 'term=' something like hg18 or grch37
        string gc_ids = kEmptyStr;
        x_GetAssmIds_Unknown(gc_ids, term);
        if ( !gc_ids.empty() ) {
            // get all assembly info for the list of GC links
            CDL_AssmInfo::TAssemblies assms;
            x_GetAssmInfo(gc_ids, assms);
            assm_info = x_CreateAssmInfo(assms);
            return assm_info;
        }
    } else if (ids.size() > 0) {
        x_FilterTaxIds(ids);
    }

    if (ids.empty() ) {
        // modify the search term a bit to give it a second try
        x_SearchIds(term + "[orgn:exp]", kTaxDb, ids);
        if (ids.size() > 0) {
            x_FilterTaxIds(ids);
        }
    }

    CDL_AssmInfo::TAssemblies assms;
    ITERATE(TIds, it, ids) {
        string gc_ids = kEmptyStr;
        x_GetAssmIds_TaxId(gc_ids, *it);
        if (gc_ids.empty()) {
            string msg = "Can't find the associated assembly for tax-id: " + *it;
            continue;
        }
        x_GetAssmInfo(gc_ids, assms);
    }

    if (assms.empty())
    {
        string msg = "Can't find any assembly related to the given search term: " + term;
        NCBI_THROW(CException, eUnknown, msg);
    }

    assm_info = x_CreateAssmInfo(assms);
    return assm_info;
}


bool CAssemblyInfo::x_ValidateTaxId(const string& tax_id) const
{
    CRef<CEUtils_ConnContext> ctx(new CEUtils_ConnContext);

    // prepare eSummary request
    CESummary_Request sum_req(kTaxDb, ctx);
    sum_req.SetRequestMethod(CEUtils_Request::eHttp_Get);
    sum_req.GetId().SetIds(tax_id);

    CRef<esummary::CESummaryResult> result;
    try {
        result = sum_req.GetESummaryResult();
    } catch (CException& e) {
        LOG_POST(Error << "Get error when trying to get DocSum for tax id: "
                 << tax_id << ". Error: " << e.GetMsg());
    }

    if (result  &&  result->CanGet()  &&  !result->Get().empty()  &&
        result->Get().front()->IsDocSum()) {
            return true;
    }

    return false;
}


void CAssemblyInfo::x_FilterTaxIds(TIds& tax_ids) const
{
    CRef<CEUtils_ConnContext> ctx(new CEUtils_ConnContext);

    TIds new_ids;
    // prepare eSearch request
    CESearch_Request search_req(kAssmDb, ctx);
    search_req.SetRequestMethod(CEUtils_Request::eHttp_Post);
    search_req.SetTerm(x_CreateIdStr(tax_ids) + "[taxid]");
    search_req.SetRetMax(kRetMax);
    CRef<esearch::CESearchResult> result;
    try {
        result = search_req.GetESearchResult();
    } catch (CException& e) {
        LOG_POST(Error << "Get error when trying to search assmbly for tax ids: "
                 << x_CreateIdStr(tax_ids) << ". Error: " << e.GetMsg());
    }

    if (result  &&  result->CanGetData()  &&
        result->GetData().IsInfo()  &&
        result->GetData().GetInfo().CanGetContent()  &&
        result->GetData().GetInfo().GetContent().CanGetTranslationStack()) {
            const esearch::CTranslationStack::Tdata& term_sets =
                result->GetData().GetInfo().GetContent().GetTranslationStack().Get();
            ITERATE (esearch::CTranslationStack::Tdata, id_iter, term_sets) {
                if ((*id_iter)->IsTermSet()  &&  (*id_iter)->GetTermSet().IsSetTerm()) {
                    string id_str = (*id_iter)->GetTermSet().GetTerm();
                    size_t pos = id_str.find("[taxid]");
                    if (pos != string::npos) {
                        new_ids.insert(id_str.substr(0, pos));
                    }
                }
            }
    }

    tax_ids.swap(new_ids);
}


void CAssemblyInfo::x_SearchIds(const string& term,
                                const string& db,
                                TIds& ids) const
{
    CRef<CEUtils_ConnContext> ctx(new CEUtils_ConnContext);

    // prepare eSearch request
    CESearch_Request search_req(db, ctx);
    search_req.SetRequestMethod(CEUtils_Request::eHttp_Get);
    search_req.SetTerm(term);
    search_req.SetRetMax(kRetMax);
    CRef<esearch::CESearchResult> result;
    try {
        result = search_req.GetESearchResult();
    } catch (CException& e) {
        LOG_POST(Error << "Get error when trying to search records on db: "
                 << db << " with search term: "
                 << term << ". Error: " << e.GetMsg());
    }

    if (result  &&  result->CanGetData()  &&
        result->GetData().IsInfo()  &&
        result->GetData().GetInfo().CanGetContent()  &&
        result->GetData().GetInfo().GetContent().CanGetIdList()  &&
        result->GetData().GetInfo().GetContent().GetIdList().CanGetId()) {
            const list<string>& id = result->GetData().GetInfo().GetContent().GetIdList().GetId();

            ITERATE (list<string>, id_iter, id) {
                ids.insert(*id_iter);
            }
    }
}


void CAssemblyInfo::x_GetTaxIdInfo(const TIds& tax_ids,
                                   CDL_AssmInfo::TTax_ids& tax_info) const
{
    CRef<CEUtils_ConnContext> ctx(new CEUtils_ConnContext);

    // prepare eSummary request
    CESummary_Request sum_req(kTaxDb, ctx);
    sum_req.SetRequestMethod(CEUtils_Request::eHttp_Post);
    sum_req.SetRetMax(kRetMax);
    sum_req.GetId().SetIds(x_CreateIdStr(tax_ids));

    CRef<esummary::CESummaryResult> result;
    try {
        result = sum_req.GetESummaryResult();
    } catch (CException& e) {
        LOG_POST(Error << "Get error when trying to retrieve DocSum for tax ids: "
                 << x_CreateIdStr(tax_ids) << ". Error: " << e.GetMsg());
    }

    if (result  &&  result->CanGet()) {
        const esummary::CESummaryResult::Tdata& doc_sums = result->Get();
        ITERATE (esummary::CESummaryResult::Tdata, d_iter, doc_sums) {
            if ( !(*d_iter)->IsDocSum() ) continue;

            const esummary::CDocSum& doc_sum = (*d_iter)->GetDocSum();
            if ( !doc_sum.CanGetId() ) continue;

            CRef<CDL_TaxId> tax(new CDL_TaxId);
            tax->SetId(doc_sum.GetId());
            int equal_three = 0;
            ITERATE (esummary::CDocSum::TItem, i_iter, doc_sum.GetItem()) {
                if ((*i_iter)->CanGetItem()  &&
                    !(*i_iter)->GetItem().empty()  &&
                    (*i_iter)->GetItem().front()->Is_CharData() &&
                    (*i_iter)->CanGetAttlist()  &&
                    (*i_iter)->GetAttlist().IsSetName()) {
                        const string& name = (*i_iter)->GetAttlist().GetName();
                        const string& value = (*i_iter)->GetItem().front()->Get_CharData();
                        if (NStr::EqualNocase(name, "Division")) {
                            tax->SetDivision(value);
                            ++equal_three;
                        } else if (NStr::EqualNocase(name, "ScientificName")) {
                            tax->SetScientific_name(value);
                            ++equal_three;
                        } else if (NStr::EqualNocase(name, "CommonName")) {
                            tax->SetName(value);
                            ++equal_three;
                        }
                }

                if (equal_three == 3) break;
            }
            tax_info.push_back(tax);
        }
    }
}


void CAssemblyInfo::x_GetAssmIds_GI(string& gc_ids, int gi) const
{
    CRef<CEUtils_ConnContext> ctx(new CEUtils_ConnContext);

    // prepare eLink request
    CELink_Request link_req(kAssmDb, ctx);
    link_req.SetRequestMethod(CEUtils_Request::eHttp_Get);
    link_req.SetDbFrom(kNucDb);
    link_req.GetIdGroups().SetGroups(NStr::IntToString(gi));

    CRef<elink::CELinkResult> result;
    try {
        result = link_req.GetELinkResult();
    } catch (CException& e) {
        LOG_POST(Error << "Get error when trying to get assembly ids gi: "
                 << gi << ". Error: " << e.GetMsg());
    }

    if (result) {
        TIds ids;
        CTypeIterator<CLink> link_iter(*result);
        for (; link_iter; ++link_iter) {
            if (link_iter->IsSetId()  &&  link_iter->GetId().IsSetId()) {
                ids.insert(link_iter->GetId().GetId());
            }
        }
        gc_ids = x_CreateIdStr(ids);
    }
}


void CAssemblyInfo::GetAssmIds_GIChr(CAssemblyInfo::TIds& gc_ids, int gi)
{
    gc_ids.clear();

    CRef<CEUtils_ConnContext> ctx(new CEUtils_ConnContext);

    // prepare eLink request
    CELink_Request link_req(kAssmDb, ctx);
    link_req.SetRequestMethod(CEUtils_Request::eHttp_Get);
    link_req.SetDbFrom(kNucDb);
    link_req.GetIdGroups().SetGroups(NStr::IntToString(gi));
    link_req.SetCommand(CELink_Request::eCmd_neighbor_score);
    link_req.SetLinkName("nuccore_assembly");

    CRef<elink::CELinkResult> result;
    try {
        result = link_req.GetELinkResult();
    } catch (CException& e) {
        LOG_POST(Error << "Get error when trying to get assembly ids gi: "
                 << gi << ". Error: " << e.GetMsg());
    }

    if (result) {
        CTypeIterator<CLink> link_iter(*result);
        for (; link_iter; ++link_iter) {
            //!! chromosome is indicated by score of "2"
            // (magic string indicated in e-mail communication from Avi Kimchi on 04/01/2013)
            if ((link_iter->IsSetScore() && link_iter->GetScore() == "2") &&
                (link_iter->IsSetId() && link_iter->GetId().IsSetId())) {
                gc_ids.insert(link_iter->GetId().GetId());
            }
        }
    }
}


void CAssemblyInfo::x_GetAssmIds_TaxId(string& gc_ids,
                                       const string& tax_id) const
{
    TIds ids;
    string term = tax_id + "[txid]";
    x_SearchIds(term, kAssmDb, ids);
    gc_ids = x_CreateIdStr(ids);
}


void CAssemblyInfo::x_GetAssmIds_Unknown(string& gc_ids,
                                         const string& unknown_term) const
{
    TIds ids;
    x_SearchIds(unknown_term, kAssmDb, ids);
    gc_ids = x_CreateIdStr(ids);
}


void CAssemblyInfo::x_GetAssmInfo(const string& gc_ids,
                                  CDL_AssmInfo::TAssemblies& assms) const
{
    CRef<CEUtils_ConnContext> ctx(new CEUtils_ConnContext);

    // prepare eSummary request
    CESummary_Request sum_req(kAssmDb, ctx);
    sum_req.SetRequestMethod(CEUtils_Request::eHttp_Post);
    sum_req.SetRetMax(kRetMax);
    sum_req.GetId().SetIds(gc_ids);

    string result_str;
    try {
        sum_req.Read(&result_str);
    } catch (CException& e) {
        LOG_POST(Error << "Get error when trying to get DocSum for assembly ids: "
                 << gc_ids << ". Error: " << e.GetMsg());
    }

    if ( !result_str.empty() ) {
        xml::document doc(result_str.c_str(), result_str.size(), NULL);
        xml::node& root = doc.get_root_node();
        xml::node::const_iterator i = root.find("DocumentSummarySet", root.begin());
        //cout << "result: " << result_str << endl;

        if (i != root.end()) {
            ITERATE (xml::node, iter, *i) {
                xml::node::const_iterator a_acc = iter->find("AssemblyAccession", iter->begin());
                xml::node::const_iterator a_name = iter->find("AssemblyName", iter->begin());
                if (a_name != iter->end()  &&  a_acc != iter->end()) {
                    CRef<CDL_Assembly> assm(new CDL_Assembly);
                    assm->SetName(a_name->get_content());
                    assm->SetAccession(a_acc->get_content());
                    if (NStr::StartsWith(a_acc->get_content(), "GCF_", NStr::eNocase)) {
                        assm->SetRelease_type("refseq");
                    } else {
                        assm->SetRelease_type("genbank");
                    }

                    ITERATE (xml::node, sub_iter, *iter) {
                        string n_name = sub_iter->get_name();
                        string n_value = sub_iter->get_content();
                        if (n_value.empty()) continue;

                        if (NStr::EqualNocase(n_name, "Organism")) {
                            assm->SetOrganism(n_value);
                        } else if (NStr::EqualNocase(n_name, "AssemblyClass")) {
                            assm->SetClass(n_value);
                        } else if (NStr::EqualNocase(n_name, "AssemblyDescription")) {
                            assm->SetDescr(n_value);
                        } else if (NStr::EqualNocase(n_name, "NCBIReleaseDate")) {
                            assm->SetRelease_date(n_value);
                        } else if (NStr::EqualNocase(n_name, "UCSCName")) {
                            assm->SetOther_names().push_back("UCSC name: " + n_value);
                        } else if (NStr::EqualNocase(n_name, "EnsemblName")) {
                            assm->SetOther_names().push_back("Ensembl name: " + n_value);
                        }
                    }
                    assms.push_back(assm);

                    // Using "gcassembly" Entrez, you will not get the the GB assemblies
                    // for some cases (those paired to RS assemblies) as separate entries
                    // from an Entrez search. But if you are looking at the Entrez Docsums,
                    // there is a field "GbUid" that tells you the release id of the
                    // corresponding GB, and fields <Synonym>/<Genbank> and <Synonym>/<RefSeq>
                    // will tell you the accessions of both of them. (JIRA: GCOL-1493)
                    xml::node::const_iterator gbid_node = iter->find("GbUid", iter->begin());
                    if (gbid_node != iter->end()  &&  !string(gbid_node->get_content()).empty()) {
                        xml::node::const_iterator syn_node = iter->find("Synonym", iter->begin());
                        if (syn_node != iter->end()) {
                            xml::node::const_iterator gb_assm_node = syn_node->find("Genbank", syn_node->begin());
                            if (gb_assm_node != syn_node->end()) {
                                CRef<CDL_Assembly> gb_assm(new CDL_Assembly);
                                gb_assm->SetName(assm->GetName());
                                gb_assm->SetAccession(gb_assm_node->get_content());
                                gb_assm->SetRelease_type("genbank");
                                if (assm->IsSetOrganism()) {
                                    gb_assm->SetOrganism(assm->GetOrganism());
                                }
                                if (assm->IsSetClass()) {
                                    gb_assm->SetClass(assm->GetClass());
                                }
                                if (assm->IsSetDescr()) {
                                    gb_assm->SetDescr(assm->GetDescr());
                                }
                                if (assm->IsSetRelease_date()) {
                                    gb_assm->SetRelease_date(assm->GetRelease_date());
                                }
                                if (assm->IsSetOther_names()) {
                                    gb_assm->SetOther_names() = assm->GetOther_names();
                                }

                                assms.push_back(gb_assm);
                            }
                        }
                    }
                }
            }
        }
    }
}


static bool s_AssmCompare(const CRef<CDL_Assembly>& assm1, const CRef<CDL_Assembly>& assm2)
{
    if (assm2->CanGetRelease_date()) {
        if (assm1->CanGetRelease_date()) {
            return assm2->GetRelease_date() < assm1->GetRelease_date();
        }
        return false;
    }
    return true;
}


void CAssemblyInfo::x_SortAssms(CDL_AssmInfo::TAssemblies& assms) const
{
    if (assms.size() < 2) return;

    // sort assembly according to the release date
    assms.sort(s_AssmCompare);

    // for certain known organisms, we promote a specific assembly to the top
    // human: Homo sapiens: GCF_0000001405
    // mouse: Mus musculus: GCF_000001635
    // cow:   Bos taurus:   GCF_000003205
    // rat:   Rattus norvegicus: GCF_000001895
    typedef map<string, string> TAssmMap;
    TAssmMap a_map;
    a_map["Homo sapiens"] = "GCF_000001405";
    a_map["Mus musculus"] = "GCF_000001635";
    a_map["Bos taurus"] = "GCF_000003205";
    a_map["Rattus norvegicus"] = "GCF_000001895";

    CDL_AssmInfo::TAssemblies::iterator iter = assms.begin();

    if ((*iter)->CanGetOrganism()) {
        TAssmMap::const_iterator a_iter = a_map.find((*iter)->GetOrganism());
        if (a_iter != a_map.end()) {
            while (iter != assms.end()  &&
                (*iter)->GetAccession().find(a_iter->second) == string::npos) {
                    ++iter;
            }

            if (iter != assms.end()  &&  iter != assms.begin()) {
                CRef<CDL_Assembly> assm = *iter;
                assms.erase(iter);
                assms.push_front(assm);
            }
        }
    }
}


string CAssemblyInfo::x_CreateIdStr(const TIds& ids) const
{
    string id_str;
    bool first = true;
    ITERATE (TIds, iter, ids) {
        if (first) {
            id_str = *iter;
            first = false;
        } else {
            id_str += "," + *iter;
        }
    }
    return id_str;
}


CRef<CDL_AssmInfo> CAssemblyInfo::x_CreateAssmInfo(CDL_AssmInfo::TAssemblies& assms) const
{
    CRef<CDL_AssmInfo> assm_info;
    if ( !assms.empty() ) {
        x_SortAssms(assms);
        assm_info.Reset(new CDL_AssmInfo());
        assm_info->SetAssemblies() = assms;
    }

    return assm_info;
}


CRef<CGC_Assembly> CAssemblyInfo::GetGCAssembly(const string& assm_acc)
{
    CGenomicCollectionsService cli;
    return cli.GetAssembly(assm_acc);
}


void CAssemblyInfo::GetIdMapperString(const string& assm_acc, string& mapper_str)
{
    CRef<CGC_Assembly> assm = GetGCAssembly(assm_acc);
    if (assm) {
        mapper_str.clear();

        vector< CConstRef<CGC_Sequence> >  seqs;
        CGC_Assembly::TSequenceList chromosomes;
        CGC_Assembly::TSequenceList scaffold_seqs;
        CGC_Assembly::TSequenceList top_level_seqs;

        assm->GetMolecules(chromosomes, CGC_Assembly::eChromosome);
        assm->GetMolecules(scaffold_seqs, CGC_Assembly::eScaffold);
        assm->GetMolecules(top_level_seqs, CGC_Assembly::eTopLevel);

        set< CConstRef<CGC_Sequence> > unique_seqs;
        ITERATE(CGC_Assembly::TSequenceList, it, chromosomes) {
            if (unique_seqs.insert(*it).second) {
                seqs.push_back(*it);
            }
        }
        ITERATE(CGC_Assembly::TSequenceList, it, scaffold_seqs) {
            if (unique_seqs.insert(*it).second) {
                seqs.push_back(*it);
            }
        }
        ITERATE(CGC_Assembly::TSequenceList, it, top_level_seqs) {
            if (unique_seqs.insert(*it).second) {
                seqs.push_back(*it);
            }
        }
        mapper_str = "[" + assm_acc + "] \n";
        mapper_str += "map_from = UCSC \n";
        mapper_str += "map_to = NCBI  \n";
        ITERATE (vector< CConstRef<CGC_Sequence> >, iter, seqs) {
            if ( !(*iter)->GetSeq_id().IsGi() ) continue;

            mapper_str += NStr::LongToString((*iter)->GetSeq_id().GetGi());
            mapper_str += " =";
            string local_id1;
            CConstRef<CSeq_id>
                seq_id = (*iter)->GetSynonymSeq_id(CGC_TypedSeqId::e_External, CGC_SeqIdAlias::e_None);

            if (seq_id) {
                seq_id->GetLabel(&local_id1, CSeq_id::eContent);
                mapper_str += (isdigit(local_id1[0]) ? " lcl|" : " ") + local_id1;
            }

            seq_id = (*iter)->GetSynonymSeq_id(CGC_TypedSeqId::e_Private, CGC_SeqIdAlias::e_None);
            if (seq_id) {
                string local_id2;
                seq_id->GetLabel(&local_id2, CSeq_id::eContent);
                if (local_id1 != local_id2) {
                    mapper_str += (isdigit(local_id2[0]) ? " lcl|" : " ") + local_id2;
                }
            }
            mapper_str += "\n";
        }
    }
}


void CAssemblyInfo::GetIdMaps(const string& assm_acc, TIdMaps& id_maps)
{
}


END_NCBI_SCOPE
