/* $Id: document.cpp 26643 2012-10-18 17:54:05Z katargir $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Author:  Vladimir Tereshkov, Andrey Yazhuk
 *
 */

#include <ncbi_pch.hpp>


#include <gui/core/document.hpp>

#include <gui/core/project_service.hpp>
#include <gui/objutils/undo_manager_interface.hpp>

#include <gui/widgets/wx/message_box.hpp>

#include <gui/objutils/label.hpp>
#include <gui/utils/app_job_impl.hpp>

#include <objmgr/object_manager.hpp>

/// specific supported project versions
#include <gui/objects/GBProject.hpp>
#include <gui/objects/GBProject_ver2.hpp>
#include <gui/objects/GBProjectHandle.hpp>
#include <gui/objects/ProjectDescr.hpp>
#include <gui/objects/ProjectFolder.hpp>
#include <gui/objects/ProjectItem.hpp>
#include <gui/objects/FolderInfo.hpp>
#include <gui/objects/ProjectHistoryItem.hpp>
#include <gui/objects/ProjectAnnot.hpp>
#include <gui/objects/PluginMessage.hpp>

#include <objects/general/Dbtag.hpp>
#include <objects/general/Object_id.hpp>
#include <objects/seqfeat/Imp_feat.hpp>
#include <objects/seqfeat/Gb_qual.hpp>

#include <serial/objistr.hpp>
#include <serial/serial.hpp>
#include <serial/pack_string.hpp>
#include <serial/objostr.hpp>
#include <serial/iterator.hpp>

#include <corelib/rwstream.hpp>

#include <connect/ncbi_conn_stream.hpp>

#include <gui/widgets/wx/wx_utils.hpp>


BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

#define INIT_MEMBERS \
    m_UpdateLockCounter( 0 ) \
    , m_StateChanged( false ) \
    , m_AccumSubtypes( CProjectViewEvent::eNone )

/////////////////////////////////////////////////////////////////////////////
/// CGBDocument
CGBDocument::CGBDocument(CProjectService* srv)
    : m_Service(srv)
    , INIT_MEMBERS
    , m_UndoManager(this)
{
    m_Id = GetNextId();
    LOG_POST(Info << "CGBDocument empty constructor");
}


CGBDocument::CGBDocument(CProjectService* srv, TId id, CGBProject& proj_impl)
    : CGBProjectHandle(proj_impl)
    , m_Service(srv)
    , INIT_MEMBERS
    , m_UndoManager(this)
{
    m_Id = id;
    LOG_POST(Info << "CGBDocument(src, id, CGBProject) id = " << id << " " << this);
}


CGBDocument::CGBDocument(CProjectService* srv, TId id, CGBProject_ver2& proj_impl)
    : CGBProjectHandle(proj_impl)
    , m_Service(srv)
    , INIT_MEMBERS
    , m_UndoManager(this)
{
    m_Id = id;
    LOG_POST(Info << "CGBDocument(src, id, CGBProject_ver2) id = " << id << " " << this);
}


CGBDocument::CGBDocument(CProjectService* srv, const string& filename)
    : CGBProjectHandle(filename)
    , m_Service(srv)
    , INIT_MEMBERS
    , m_UndoManager(this)
{
    LOG_POST(Info << "CGBDocument(src, " << filename << ") " << this);
}


CGBDocument::~CGBDocument(void)
{
    LOG_POST(Info << "CGBDocument::~CGBDocument");
}

wxString CGBDocument::GetFileName() const
{
    return FnToWxString(CGBProjectHandle::GetFilename());
}

void CGBDocument::SetFileName (const wxString& filename)
{
    CGBProjectHandle::SetFilename(FnToStdString(filename));
}

void CGBDocument::CreateProjectScope()
{
    LOG_POST(Info << "CGBDocument::CreateProjectScope() " << this);

    _ASSERT( ! GetScope());

    CRef<CObjectManager> obj_mgr = CObjectManager::GetInstance();
    CRef<objects::CScope> scope(new objects::CScope(*obj_mgr));
    scope->AddDefaults();
    SetScope(scope.GetPointer());
}

void CGBDocument::LockUpdates()
{
    CMutexGuard mtx_guard( m_UpdateMutex ); // protect m_UpdateLockCounter

    if( m_UpdateLockCounter == 0 ){
        m_StateChanged = false; // reset state
    }
    ++m_UpdateLockCounter; // increment counter
}


void CGBDocument::UnlockUpdates()
{
    CMutexGuard mtx_guard( m_UpdateMutex ); // protect m_UpdateLockCounter

    _ASSERT( m_UpdateLockCounter >= 0 );

    --m_UpdateLockCounter;

    if( m_UpdateLockCounter == 0  &&  m_StateChanged ){
        // flush accumulated updates
        x_FireProjectChanged( m_AccumSubtypes );
        m_StateChanged = false;
        m_AccumSubtypes = CProjectViewEvent::eNone;
    }
}

class CDocLoadingJob : public CAppJob
{
public:
    CDocLoadingJob(CGBDocument& doc, const wxString& path)
        : CAppJob("Load Project Job"), m_Document(&doc), m_Path(path) {}

    /// @name IAppJob implementation
    /// @{
    virtual EJobState   Run();
    /// @}

protected:
    CRef<CGBDocument> m_Document;
    wxString  m_Path;
};

IAppJob* CGBDocument::CreateLoadJob(const wxString& fileName)
{
    if (IsLoaded() || m_LoadJob)
        return 0;

    m_LoadJob.Reset(new CDocLoadingJob(*this, fileName));

    return m_LoadJob.GetPointer();
}

bool CGBDocument::IsLoading() const
{
    return m_LoadJob;
}

void CGBDocument::CancelLoading()
{
    if (m_LoadJob) {
        m_LoadJob->RequestCancel();
        m_LoadJob.Reset();
    }
}

void CGBDocument::ResetLoading()
{
    m_LoadJob.Reset();
}

IAppJob::EJobState CDocLoadingJob::Run()
{
    bool ok = m_Document->LoadFile(m_Path, x_GetICanceled());

    if (!ok) {
        if (IsCanceled())
            return eCanceled;

        string s = "Failed to load project: \"" + ToStdString(m_Path) + "\".";
        CFastMutexGuard lock(m_Mutex);
        m_Error.Reset(new CAppJobError(s));
        return eFailed;
    }

    _ASSERT(m_Document->GetScope() == NULL);
    m_Document->CreateProjectScope();
    return eCompleted;
}

/// NULL stream writer for test serialization
class CNullWriter : public IWriter
{
public:
    ERW_Result Write(const void* buf, size_t count, size_t* bytes_written)
    {
        if (bytes_written) {
            *bytes_written = count;
        }
        return eRW_Success;
    }

    ERW_Result Flush()
    {
        return eRW_Success;
    }
};


/// Attach a view to the current document
void CGBDocument::x_AttachView(IProjectView* view)
{
    _ASSERT(view);

    m_Views.push_back(CIRef<IProjectView>(view));

    x_AssignViewIndex(view);

    // event routing
    CEventHandler* handler = dynamic_cast<CEventHandler*>(view);
    if(handler) {
        AddListener(handler);
        handler->AddListener(this);
    }

    //m_Service->OnViewAttached(view);

    x_FireViewEvent( *view, TEvent::eViewCreated );
}


/// detach a view from the current document
void CGBDocument::x_DetachView(IProjectView* view)
{
    _ASSERT(view);

    CIRef<IProjectView> ref(view);

    // unregister event handling
    CEventHandler* handler = dynamic_cast<CEventHandler*>(view);
    if (handler) {
        RemoveListener(handler);
        handler->RemoveListener(this);
    }

    m_ViewToIndex.erase(view);
    TViews::iterator it = std::find(m_Views.begin(), m_Views.end(), ref);
    m_Views.erase(it);

    //m_Service->OnViewDetached(view);

    x_FireViewEvent( *view, TEvent::eViewReleased );
}


void CGBDocument::x_AssignViewIndex(IProjectView* view)
{
    _ASSERT(view);
    const string& title = view->GetClientLabel();
    int max_index = -1;

    ITERATE (TViewToIndex, it, m_ViewToIndex) {
        IProjectView* v = it->first;
        if (v->GetClientLabel() == title)   { // view of the same type
            max_index = max(max_index, it->second);
        }
    }
    m_ViewToIndex[view] = max_index + 1;
}


/// Retrieve the existing views for this class
const CGBDocument::TViews& CGBDocument::x_GetViews(void) const
{
    return m_Views;
}


CGBDocument::TViews& CGBDocument::x_GetViews(void)
{
    return m_Views;
}


void CGBDocument::x_DetachAllViews()
{
    TViews views = m_Views; // to postpone destruction of views

    NON_CONST_ITERATE(TViews, it, m_Views)  {

        IProjectView* view = *it;

        CEventHandler* disp = dynamic_cast<CEventHandler*>(view);
        if (disp) {
            RemoveListener(disp);
            disp->RemoveListener(this);
        }

        x_FireViewEvent( *view, TEvent::eViewReleased );
        //TODO CProjectManager::OnViewDetached(view);
    }

    m_ViewToIndex.clear();
    m_Views.clear();
}


void CGBDocument::x_OnProjectChanged( CProjectViewEvent::EEventSubtype subtype )
{
    CMutexGuard mtx_guard( m_UpdateMutex );   // now it is safe to check the counter

    if( m_UpdateLockCounter == 0 ){
        x_FireProjectChanged( subtype );

    } else {
        m_StateChanged = true; // mark as changed for future
        m_AccumSubtypes = (CProjectViewEvent::EEventSubtype)((int)m_AccumSubtypes | subtype);
    }
}

void CGBDocument::x_FireProjectChanged( TEvent::EEventSubtype subtype )
{
    TEvent* ev = new TEvent( GetId(), subtype );
    if (subtype == CProjectViewEvent::eDataChanging)
        Send( CRef<CEvent>( ev ) );
    else
        Post( CRef<CEvent>( ev ) );
}

void CGBDocument::x_FireViewEvent( IProjectView& view, CViewEvent::EEventType type )
{
    TEvent ev( view, type );

    Send( &ev );
}

/*
void CGBDocument::x_PostViewEvent( IProjectView& view, CViewEvent::EEventType type )
{
    TEvent* ev = new TEvent( view, TEvent::eProjectChanged );
    //ev->m_View.Reset( &view );

    CRef<CEvent> evt( ev );
    Post( evt );
}
*/

struct STempFileLocker
{
    string filename;

    STempFileLocker(const string& tmp_name)
        : filename(tmp_name)
    {
    }

    void Lock()
    {
    }

    void Unlock()
    {
        if ( !filename.empty() ) {
            CFile(filename).Remove();
        }
    }
};

bool CGBDocument::LoadFile(const wxString& fname, ICanceled* canceledCallback)
{
    auto_ptr<CNcbiIfstream> istr;
    string tmp_name;
    if (fname.StartsWith(wxT("http://"))
        ||  fname.StartsWith(wxT("ftp://"))) {

        auto_ptr<CNcbiIstream> is;
        if (fname.StartsWith(wxT("http://"))) {
            /// we always open with the *full* path - this will pass params on
            /// the command line and log them in the weblogs.
            auto_ptr<CConn_HttpStream> http_str(new CConn_HttpStream(ToStdString(fname)));
            is.reset(http_str.release());
        } else {
            /// For FTP connections, we need to separate host name from
            /// path name

            /// separate the host to contact
            string _fname = ToStdString(fname);
            string host = _fname;
            string::size_type pos = host.find_first_of("/", 6);
            if (pos != string::npos) {
                host.erase(pos);
            }
            host.erase(0, 6);

            /// separate the path to retrieve
            string path;
            if (pos != string::npos) {
                path = _fname.substr(pos, _fname.length() - pos);
            }

            /// separate the file name
            //SOCK_SetDataLoggingAPI(eOn);
            is.reset(new CConn_FTPDownloadStream(host, path));
        }

        /// we will be trying a number of things here, so create a
        /// temporary file stream for this
        tmp_name = CFile::GetTmpName();
        {{
            CNcbiOfstream ostr(tmp_name.c_str(), ios::binary);
            NcbiStreamCopy(ostr, *is);
        }}
        istr.reset(new CNcbiIfstream(tmp_name.c_str(), ios::binary|ios::in));

    } else {
        /// default file connection
        istr.reset(new CNcbiIfstream(fname.fn_str(), ios::binary|ios::in));
    }
    _ASSERT(istr.get());

    STempFileLocker tmp_file(tmp_name);
    CGuard<STempFileLocker> LOCK(tmp_file);

    return CGBProjectHandle::Load(*istr, canceledCallback);
}

void CGBDocument::SaveFile(const wxString& fname, ESerialDataFormat fmt, bool keep_backups)
{
    LOG_POST(Info << "CGBProjectHandle::Save(" << fname.ToUTF8() << ")");

    ///
    /// step 1: write to a new file
    /// we aoid overwriting an existing file before we are sure that
    /// serialization has succeeded
    ///

    bool serialization_success = true;
    wxString serialized_fname;
    try {
        /// rename the existing file, if any
        serialized_fname = fname;
        for (int i = 1;  ;  ++i) {
            wxString f = wxString::Format(wxT("%s.%d"), fname.c_str(), i);
            if ( !::wxFileExists(f) ) {
                serialized_fname = f;
                break;
            }
        }

        CNcbiOfstream ostr(serialized_fname.fn_str(), ios::binary);
        CGBProjectHandle::Save(ostr, fmt);
    }
    catch (std::exception& e) {
        LOG_POST(Error << "CGBProjectHandle::Save(" << fname
                 << "): error in serialization: " << e.what());
        serialization_success = false;
    }

    /// scan the serialized file to make sure that nothing bad happened
    if (serialization_success) {
        try {
            CNcbiIfstream istr(serialized_fname.fn_str(), ios::binary);
            auto_ptr<CObjectIStream> is(CObjectIStream::Open(fmt, istr));

            const IGBProject& project = GetProject();
            switch (project.GetVersion()) {
            case IGBProject::eVersion1:
                is->Skip(CGBProject::GetTypeInfo());
                break;

            case IGBProject::eVersion2:
                is->Skip(CGBProject_ver2::GetTypeInfo());
                break;

            default:
                _ASSERT(false);
                NCBI_THROW(CException, eUnknown, "Unknown project version");
                break;
            }
            is->Close();
            if ( !istr ) {
                NCBI_THROW(CException, eUnknown, "stream read failure");
            }
        }
        catch (std::exception& e) {
            LOG_POST(Error << "CGBProjectHandle::Save(" << fname
                    << "): error in scanning file: " << e.what());
            serialization_success = false;
        }
    }

    if ( !serialization_success ) {
#ifndef _DEBUG
        /// in debug mode, we preserve our bread-crumbs
        ::wxRemoveFile(serialized_fname);
#endif

        string msg("An error occurred while saving '");
        msg += fname.ToUTF8();
        msg += "'.\n";
        if ( ::wxFileExists(fname) ) {
            msg += "The original file was not overwrtten and should be safe.\n";
        }
        msg += "\nYou can try to save the file again, or optionally\n";
        msg += "export any project items individually to preserve the data.";
        NCBI_THROW(CException, eUnknown, msg);
    }

    ///
    /// success!
    ///

    /// rename the existing file, if any
    if (keep_backups) {
        wxString backup_fname = fname + wxT(".bak");
        if (::wxFileExists(backup_fname))
            ::wxRemoveFile(backup_fname);
        if (::wxFileExists(fname))
            ::wxRenameFile(fname,backup_fname);
        ::wxRenameFile(serialized_fname, fname);
    }
}

END_NCBI_SCOPE

