/*  $Id: gcassembly_search_wizard.cpp 28155 2013-05-17 18:17:47Z katargir $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Liangshou Wu
 *
 */

#include <ncbi_pch.hpp>
#include <gui/core/gcassembly_search_wizard.hpp>
#include <gui/core/assm_info.hpp>
#include <gui/widgets/wx/wx_utils.hpp>
#include <objmgr/object_manager.hpp>
#include <objmgr/util/sequence.hpp>
#include <wx/msgdlg.h>
#include "assembly_list_panel.hpp"

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);


///////////////////////////////////////////////////////////////////////////////
/// CGCAssemblySearchWizard
CGCAssemblySearchWizard::CGCAssemblySearchWizard(EUsage usage)
   : m_Usage(usage)
   , m_ParentWindow(NULL)
   , m_AssmPanel(NULL)
{
}


void CGCAssemblySearchWizard::SetParentWindow(wxWindow* parent)
{
    m_ParentWindow = parent;
}


void CGCAssemblySearchWizard::InitUI()
{
}


void CGCAssemblySearchWizard::CleanUI()
{
    m_AssmPanel = NULL;
    m_Assemblies.clear();
}


wxPanel* CGCAssemblySearchWizard::GetCurrentPanel()
{
    if (m_AssmPanel == NULL) {
        x_CreateAssmPanel();
    }
    return m_AssmPanel;
}


bool CGCAssemblySearchWizard::CanDo(IUIToolManager::EAction action)
{
    return false;
}


bool CGCAssemblySearchWizard::DoTransition(IUIToolManager::EAction action)
{
    return false;
}


static const string kOrgListTag = ".OrgListPanel";
static const string kAssmListTag = ".AssmListPanel";


void CGCAssemblySearchWizard::SetRegistryPath(const string& path)
{
    m_RegPath = path; // store for later use

    if (m_AssmPanel) {
        m_AssmPanel->SetRegistryPath(path + kAssmListTag);
    }
}


void CGCAssemblySearchWizard::SaveSettings() const
{
    if ( !m_RegPath.empty() ) {
        if (m_AssmPanel) {
            m_AssmPanel->SaveSettings();
        }
    }
}


void CGCAssemblySearchWizard::LoadSettings()
{
    if ( !m_RegPath.empty() ) {
        if (m_AssmPanel) {
            m_AssmPanel->LoadSettings();
        }
    }
}


vector<string> CGCAssemblySearchWizard::GetAssmAccs() const
{
    vector<string> accs;

    if (m_AssmPanel  &&  m_AssmPanel->IsInputValid()) {
        m_AssmPanel->SaveSettings();
        accs = m_AssmPanel->GetSelectedAssemblies();
    }

    return accs;
}


bool CGCAssemblySearchWizard::IsInitialState() const
{
    return true;
}


CRef<CDL_AssmInfoResponse> CGCAssemblySearchWizard::x_GetAssmInfo(
    const string& query_str, ESearchMode mode)
{
    CRef<CDL_AssmInfoResponse> assm_resp(new CDL_AssmInfoResponse);
    assm_resp->SetSuccess(false);

    CRef<CDL_AssmInfo> assm;
    CAssemblyInfo assm_handler;

    try {
        LOG_POST(Info << "Retrieve assembly information for: " << query_str);
        if (mode == eSearch_TaxId) {
            assm = assm_handler.GetAssms_TaxId(query_str);
        } else {
            try {
                assm = assm_handler.GetAssms_Term(query_str);
            } catch (CException& e) {
                // try if the search term is a seq-id                                                             
                // get the gi for the given seq-id  
                CScope scope(*CObjectManager::GetInstance());
                scope.AddDefaults();
                
                try {
                    CRef<CSeq_id> seq_id(new CSeq_id);
                    seq_id->Set(query_str);
                    CBioseq_Handle bsh = scope.GetBioseqHandle(*seq_id);
                    if (bsh) {
                        CSeq_id_Handle idh_gi = bsh.GetSeq_id_Handle();
                        idh_gi = sequence::GetId(idh_gi, scope, sequence::eGetId_ForceGi);
                        if (idh_gi) {
                            // get all GC links associated with the gi
                            assm = assm_handler.GetAssms_Gi(idh_gi.GetGi());
                        }
                    }
                } catch (CException&) {
                    // ignore
                } 
                if ( !assm ) {
                    throw e;
                }
            }
        }

        if (assm) {
            assm_resp->SetSuccess(true);
            assm_resp->SetData(*assm);
        } else {
            throw string("Failed to get assembly information!");
        }
    } catch (CException& e){
        assm_resp->SetMsg(e.GetMsg());
    } catch (string& str) {
        assm_resp->SetMsg(str);
    }

    return assm_resp;
}

void CGCAssemblySearchWizard::x_CreateAssmPanel()
{
    m_AssmPanel = new CAssemblyListPanel(m_ParentWindow);
    if (m_Usage == eIdMapping) {
        m_AssmPanel->SetMultiSelection(false);
        m_AssmPanel->SetMainTitle(wxT("Id Mapping -- Target Assembly"));
        m_AssmPanel->SetDescr(
            wxT("The given search term results in multiple assemblies.  Please choose one."));
    }

    if ( !m_RegPath.empty() ) {
        m_AssmPanel->SetRegistryPath(m_RegPath + kAssmListTag);
        m_AssmPanel->LoadSettings();
    }
}

END_NCBI_SCOPE
