/*  $Id: loading_app_job.cpp 26643 2012-10-18 17:54:05Z katargir $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Andrey Yazhuk, Liangshou Wu
 *
 */

#include <ncbi_pch.hpp>

#include <util/compress/zlib.hpp>
#include <util/compress/bzip2.hpp>

#include <gui/core/loading_app_job.hpp>

#include <gui/core/project_selector_dlg.hpp>
#include <gui/core/job_results_dlg.hpp>

#include <gui/objects/ProjectItem.hpp>
#include <gui/objects/ProjectDescr.hpp>
#include <gui/objects/ProjectFolder.hpp>
#include <gui/objects/FolderInfo.hpp>
#include <gui/widgets/wx/wx_utils.hpp>

#include <objects/general/Date.hpp>
#include <objmgr/object_manager.hpp>
#include <corelib/ncbi_system.hpp>

#include <wx/dirdlg.h>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);


///////////////////////////////////////////////////////////////////////////////
/// CDataLoadingOptions
CDataLoadingOptions::CDataLoadingOptions()
:   m_Action(eInvalidAction)
{
}


void CDataLoadingOptions::Set_DecideLater()
{
    m_Action = eDecideLater;
}


void CDataLoadingOptions::Set_CreateNewProject(const string& folder)
{
    m_Action = eCreateNewProject;
    m_TargetProjectId = CGBDocument::GetNextId();//GenerateNewId();
    m_FolderName = folder;
}


void CDataLoadingOptions::Set_CreateSeparateProjects()
{
    m_Action = eCreateSeparateProjects;
    m_FolderName = "";
}


void CDataLoadingOptions::Set_AddToExistingProject(TProjectId& project_id,
                                                   const string& folder)
{
    m_Action = eAddToExistingProject;
    m_TargetProjectId = project_id;
    m_FolderName = folder;
}


CDataLoadingOptions::EAction CDataLoadingOptions::GetAction()
{
    return m_Action;
}


CDataLoadingOptions::TProjectId CDataLoadingOptions::GetTargetProjectId()
{
    return m_TargetProjectId;
}


string CDataLoadingOptions::GetFolderName()
{
    return m_FolderName;
}


///////////////////////////////////////////////////////////////////////////////
/// CDataLoadingJob
CDataLoadingJob::CDataLoadingJob( CProjectService* service )
    : m_Service(service)
    , m_TextResult()
    , m_ItemsAdded(false)
    , m_LoadersAdded(false)
{
}


CDataLoadingJob::~CDataLoadingJob()
{
}


CDataLoadingOptions& CDataLoadingJob::GetOptions()
{
    return m_Options;
}


void CDataLoadingJob::SetOptions(const CDataLoadingOptions& options)
{
    m_Options = options;
}


CProjectService* CDataLoadingJob::GetProjectService()
{
    return m_Service.GetPointer();
}

void CDataLoadingJob::x_SetTextResult( const string& result ){ 
    
    CFastMutexGuard lock(m_Mutex);
    m_TextResult = result; 
}



IAppJob::EJobState CDataLoadingJob::Run()
{
    static const char* kLogMsg = "CDataLoadingJob::x_Run() - exception ";

    x_ResetState();

    string err_msg;
    bool err_user_level = false;
    try {
        x_Run();

    } catch( CException& e ){
        LOG_POST(Error << kLogMsg + GetDescr() + ". " + e.GetMsg() );
        e.ReportAll();

        err_msg = e.GetMsg();
        err_user_level = true;

    } catch( std::exception& ee ){
        err_msg = kLogMsg + GetDescr() + ". " + ee.what();
        LOG_POST(Error << err_msg );

        err_user_level = false;
    } 

    CFastMutexGuard lock(m_Mutex);

    if( !err_msg.empty() ){
        m_Error.Reset( new CAppJobError( err_msg, err_user_level ) );
    }

    if( m_Error ){
        return eFailed;
    }

    LOG_POST(Info 
        << "CDataLoadingJob::Run(): Finished job " << m_Descr
    );
    return eCompleted;
}


void CDataLoadingJob::x_ResetState()
{
    CAppJob::x_ResetState();

    x_SetTextResult( m_Descr + ": no items created." );
}


void CDataLoadingJob::x_Run()
{
    m_ItemsAdded = false;
    m_LoadersAdded = false;

    // if we do not have project items - we need to create them
    // otherwise - skip this step and add existing items to a project
    if( m_Items.empty() ){
        x_SetStatusText( "Creating data items..." );
        x_CreateProjectItems(); /// populate m_Items
        if (IsCanceled()) {
            return;
        }
    }

    if( !m_Loaders.empty()  ||  !m_Items.empty() ){
        if (IsCanceled()) {
            return;
        }

        x_SetTextResult( "" );

        /// now arrange items into projects
        string prj_text = "New Project";
        string folder_name = m_Options.GetFolderName();

        if( ! m_Service->HasWorkspace())    {
            m_Service->CreateNewWorkspace();
        }

        switch(m_Options.GetAction()) {
            case TOptions::eDecideLater:
                break; // do nothing

            case TOptions::eCreateNewProject:
                x_CreateOneProject(prj_text, folder_name);
                break;

            case TOptions::eCreateSeparateProjects:
                x_CreateSeparateProjects(prj_text, folder_name);
                break;

            case TOptions::eAddToExistingProject:   {
                TProjectId prj_id = m_Options.GetTargetProjectId();
                x_AddToExistingProject(prj_id, folder_name);
                break;
            }
            default:
                _ASSERT(false);
        }

        if ( !m_Objects.empty() ) {
            m_Service->PrefetchSeqDescr(m_Objects);
            m_Objects.clear();            
        }
    }
}


void CDataLoadingJob::x_CreateProjectItems()
{
    /// override to add items to m_Items here
}


void CDataLoadingJob::x_PrepareProjectItem(CProjectItem& item, CScope& scope)
{
    // override to do something useful
}


void CDataLoadingJob::x_AddToExistingProject(TProjectId id,
                                             const string& folder_name)
{
    CScope* scope = NULL;

    {
        // get the scope and then release locks to avoid long waits
        CRef<CWorkspacePrx> ws = m_Service->GetWorkspace();
        CRef<CProjectPrx> project = ws->FindProjectById(id);
        if(project) {
            scope = project->GetScope();
        }
    }

    size_t n = m_Items.size();

    if(scope)   {
        // prepare project items without locking the project
        for( size_t i = 0;  i < n;  i++ )    {
            string s = "Preparing project items - item " + NStr::SizetToString(i + 1);
            s += " of " + NStr::SizetToString(n);
            x_SetStatusText(s);

            CProjectItem&  item = *m_Items[i];
            x_PrepareProjectItem(item, *scope);
        }
    }

    // now get the lock and attach the items
    CRef<CWorkspacePrx> ws = m_Service->GetWorkspace();
    CRef<CProjectPrx> project = ws->FindProjectById(id);
    if(project) {
        CRef<CProjectFolderPrx> new_folder = x_GetFolder(*project, folder_name);

        /// fully load and add all items
        for( size_t i = 0;  i < n;  i++ )    {
            string s = "Loading project items - item " + NStr::SizetToString(i + 1);
            s += " of " + NStr::SizetToString(n);
            x_SetStatusText(s);

            CProjectItem&  item = *m_Items[i];
            new_folder->AddProjectItem(item);
            m_Objects.push_back(SConstScopedObject(*item.GetObject(), *scope));
        }
        m_ItemsAdded = true;

        n = m_Loaders.size();
        size_t l_idx = 1;
        NON_CONST_ITERATE (TLoaders, l_iter, m_Loaders) {
            string s = "Adding data loader - " + NStr::SizetToString(l_idx);
            s += " of " + NStr::SizetToString(n);
            x_SetStatusText(s);

            if (project->IsDataLoaderAdded(*l_iter->second)) {
                LOG_POST(Warning << "The data loader (" 
                         << l_iter->second->GetLabel() 
                         << ") has already been added!");
            } else {
                project->AddDataLoader(*l_iter->second);
            }

            ++l_idx;
        }
        m_LoadersAdded = true;
    }
}


typedef set< CRef<CSerialObject> >   TObjSet;

void CDataLoadingJob::x_CreateOneProject(const string& prj_text,
                                            const string& folder_name)
{
    TObjSet empty_set;

    /// create and setup a new Project
    TProjectId id = m_Options.GetTargetProjectId();

    bool new_prj = true;
    CRef<CProjectPrx> project;

    // we create one project for all items, but this can be done in several jobs
    // so it this point the document may have been created already (by another job)
    CRef<CWorkspacePrx> ws = m_Service->GetWorkspace();
    if(ws)  {
        project = ws->FindProjectById(id);
        new_prj = project.Empty();
    }


    if(new_prj) {
        project = ws->CreateNewProject(id);
        project->SetName(prj_text);
        project->SetDescr(prj_text);

        CRef<CProjectFolderPrx> new_folder = x_GetFolder(*project, folder_name);

        /// fully load and add all m_Items
        size_t n = m_Items.size();
        for( size_t i = 0;  i < n;  i++ ) {
            string s = "Loading project items - item " + NStr::SizetToString(i + 1);
            s += " of " + NStr::SizetToString(n);
            x_SetStatusText(s);

            CProjectItem&  item = *m_Items[i];

            CScope& scope = *project->GetScope();
            x_PrepareProjectItem(item, scope);
            new_folder->AddProjectItem(item);

            m_Objects.push_back(SConstScopedObject(*item.GetObject(), scope));

        }

        /// fully load and add all loaders
        n = m_Loaders.size();
        size_t l_idx = 1;
        NON_CONST_ITERATE (TLoaders, l_iter, m_Loaders) {
            string s = "Adding data loader - " + NStr::SizetToString(l_idx);
            s += " of " + NStr::SizetToString(n);
            x_SetStatusText(s);
            project->AddDataLoader(*l_iter->second);
            ++l_idx;
        }


        ws->AddProject(*project);
        m_ItemsAdded = true;
        m_LoadersAdded = true;
    } else {
        x_AddToExistingProject(id, folder_name);
    }
}


/// for every Project Item in m_Items creates a new project,
/// adds the item to the project and loads the data
void CDataLoadingJob::x_CreateSeparateProjects(const string& prj_text,
                                               const string& folder_name)
{
    CRef<CWorkspacePrx> ws = m_Service->GetWorkspace();

    size_t n = m_Items.size();
    for( size_t i = 0;  i < n;  i++ )    {
        string s = "Creating projects and loading data - project " + NStr::SizetToString(i + 1);
        s += " of " + NStr::SizetToString(n);
        x_SetStatusText(s);

        /// create and setup a new Project
        CRef<CProjectPrx> project = ws->CreateNewProject();
        project->SetName(prj_text);
        project->SetDescr(prj_text);

        CRef<CProjectFolderPrx> new_folder = x_GetFolder(*project, folder_name);

        CScope& scope = *project->GetScope();
        CProjectItem& item = *m_Items[i];

        // add any data loaders associated with this project item
        pair<TLoaders::iterator, TLoaders::iterator> ret;
        ret = m_Loaders.equal_range(&item);
        for (TLoaders::iterator it = ret.first; it != ret.second; ++it) {
            project->AddDataLoader(*it->second);
        }
        // remove them from the loader list
        m_Loaders.erase(ret.first, ret.second);

        x_PrepareProjectItem(item, scope);

        new_folder->AddProjectItem(item);
        ws->AddProject(*project);

        m_Objects.push_back(SConstScopedObject(*item.GetObject(), scope));
    }
    m_ItemsAdded = true;

    n = m_Loaders.size();
    size_t l_idx = 1;
    NON_CONST_ITERATE (TLoaders, l_iter, m_Loaders) {
        string s = "Creating projects and adding data loader - ";
        s += NStr::SizetToString(l_idx);
        s += " of " + NStr::SizetToString(n);
        x_SetStatusText(s);

        /// create and setup a new Project
        CRef<CProjectPrx> project = ws->CreateNewProject();
        project->SetName(prj_text);
        project->SetDescr(prj_text);
        project->AddDataLoader(*l_iter->second);

        ws->AddProject(*project);
        ++l_idx;
    }

    m_LoadersAdded = true;
}


CRef<CProjectFolderPrx>
CDataLoadingJob::x_GetFolder(CProjectPrx& project,
                             const string& folder_name)
{
    CRef<CProjectFolderPrx> top_folder = project.GetDataFolder();
    //TODO @@@ top_folder.SetInfo().SetOpen(true);

    CRef<CProjectFolderPrx> new_folder = top_folder;

    if ( ! folder_name.empty() ) {
        new_folder = top_folder->FindChildFolderByName(folder_name);
        if( ! new_folder)  {
            new_folder = top_folder->AddChildFolder(folder_name, folder_name);
        }
        //TODO @@@ info.SetOpen(true);
    }

    return new_folder;
}


CRef<CObject> CDataLoadingJob::GetResult()
{
    CFastMutexGuard lock(m_Mutex);

    if( !m_TextResult.empty() ){
        return CRef<CObject>( new CAppJobTextResult( m_TextResult ) );
    }

    return CRef<CObject>();
}


bool CDataLoadingJob::FailedToAddItems() const
{
    return ! m_Items.empty()  && ! m_ItemsAdded;
}


bool CDataLoadingJob::FailedToAddLoaders() const
{
    return ! m_Loaders.empty()  && ! m_LoadersAdded;
}


const vector<CRef<CProjectItem> >& CDataLoadingJob::GetItems() const
{
    return m_Items;
}


objects::CScope* CDataLoadingJob::GetScope() const
{
    return m_Scope.GetNCPointer();
}

CFormatGuess::EFormat CDataLoadingJob::GetCompressType(const wxString& fn)
{
    wxFileName fname(fn); 
    wxString ext = fname.GetExt();
    if (ext == wxString("gz")) {
        return CFormatGuess::eGZip;
    } 
    if (ext == wxString("bz2")) {
        return CFormatGuess::eBZip2;
    } 
    return CFormatGuess::eUnknown;
}

CCompressionStreamProcessor* CDataLoadingJob::CreateCompressionProcessor(CFormatGuess::EFormat fmt)
{
    switch (fmt) {
    case CFormatGuess::eGZip:
        return new CZipStreamDecompressor(CZipCompression::fGZip);
    case CFormatGuess::eBZip2:
        return new CBZip2StreamDecompressor();
    default:
        return 0;
    } // switch
}

///////////////////////////////////////////////////////////////////////////////
/// CDataLoadingTask

CDataLoadingTask::CDataLoadingTask( CDataLoadingJob& job, const string& descr )
    : CAppJobTask( job, true, descr.empty() ? job.GetDescr() : descr )
{
}


IAppTask::ETaskState CDataLoadingTask::Run()
{
    ETaskState state = CAppJobTask::Run();

    if( state == eCompleted ){
        IAppJob* job = GetJob().GetPointer();
        CDataLoadingAppJob* data_job = dynamic_cast<CDataLoadingJob*>(job);
        if (data_job->IsCanceled()) {
            return state;
        }

        _ASSERT(data_job);

        if( data_job->FailedToAddItems() ){
            // user chose "Decide Later" option or selected a project that has
            // been closed
            CRef<CScope> default_scope( new CScope( *CObjectManager::GetInstance() ) );
            default_scope->AddDefaults();

            // we need to show a dialog to ask the user where to place the data
            const vector< CRef<CProjectItem> > items = data_job->GetItems();
            TConstScopedObjects objects;
            for(  size_t i = 0;  i < items.size();  i++ )   {
                const CProjectItem& item = *items[i];
                CScope * scope = data_job->GetScope();
                objects.push_back
                    (SConstScopedObject(*item.GetObject(), scope ? *scope : *default_scope));
            }

            string title = GetDescr();
            CProjectSelectorDlg dlg(NULL, wxID_ANY, ToWxString(title));

            dlg.SetProjectService(data_job->GetProjectService());
            dlg.SetObjects(objects);
            dlg.SetRegistryPath("Dialogs.DataLoadingTask.ProjectSelector");

            if(dlg.ShowModal() == wxID_OK)  {
                SProjectSelectorParams params;
                dlg.GetParams(params);

                CDataLoadingOptions& options = data_job->GetOptions();
                params.ToLoadingOptions(options);

                // relaunch the job
                m_State = eInitial;
                m_JobID = CAppJobDispatcher::eInvalidJobID;
                m_JobState = IAppJob::eInvalid;

                state = CAppJobTask::Run();
            }
        }

        if(data_job->FailedToAddLoaders())   {
            // user chose "Decide Later" option or selected a project that has
            // been closed

            string title = GetDescr();
            CProjectSelectorDlg dlg(NULL, wxID_ANY, ToWxString(title));
            dlg.SetProjectService(data_job->GetProjectService());
            dlg.SetRegistryPath("Dialogs.LoaderAddingTask.ProjectSelector");

            if(dlg.ShowModal() == wxID_OK)  {
                SProjectSelectorParams params;
                dlg.GetParams(params);

                CDataLoadingOptions& options = data_job->GetOptions();
                params.ToLoadingOptions(options);

                // relaunch the job
                m_State = eInitial;
                m_JobID = CAppJobDispatcher::eInvalidJobID;
                m_JobState = IAppJob::eInvalid;

                state = CAppJobTask::Run();
            }
        }

        wxString resTitle, resHTML;
        data_job->GetHTMLResults(resTitle, resHTML);

        if (!resHTML.empty()) {
            CJobResultsDlg dlg;
            dlg.SetHTML(resHTML);
            dlg.Create(NULL, wxID_ANY, resTitle);
            dlg.ShowModal();
        }
    }

    return state;
}


END_NCBI_SCOPE
