/*  $Id: object_loading_task.cpp 27061 2012-12-13 15:49:53Z katargir $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Andrey Yazhuk, Liangshou Wu
 *
 */

#include <ncbi_pch.hpp>

#include <gui/core/object_loading_task.hpp>

#include <gui/core/project_selector_dlg.hpp>
#include <gui/core/job_results_dlg.hpp>

#include <gui/objects/ProjectItem.hpp>
#include <gui/objects/ProjectDescr.hpp>
#include <gui/objects/ProjectFolder.hpp>
#include <gui/objects/FolderInfo.hpp>
#include <gui/widgets/wx/wx_utils.hpp>

#include <objmgr/object_manager.hpp>

#include <wx/dirdlg.h>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);


///////////////////////////////////////////////////////////////////////////////
/// CObjectLoadingTask

CObjectLoadingTask::CObjectLoadingTask(CDataLoadingJob& job, IObjectLoader& object_loader)
 : CAppJobTask(true, object_loader.GetDescription()),
   m_ObjectLoader(&object_loader), m_DataLoadingJob(&job), m_ObjectsLoaded(false)
{
    CIRef<IExecuteUnit> execute_unit(dynamic_cast<IExecuteUnit*>(&object_loader));
    _ASSERT(execute_unit);
    m_Job.Reset(new CAppJobExecuteUnit(*execute_unit, object_loader.GetDescription()));
}

IAppTask::ETaskState CObjectLoadingTask::Run()
{
    ETaskState state = CAppJobTask::Run();

    if (m_State == eInitial) {
        CIRef<IExecuteUnit> execute_unit(dynamic_cast<IExecuteUnit*>(m_ObjectLoader.GetPointer()));
        if (!execute_unit || !execute_unit->PreExecute())
            return eCanceled;
    }
    else if( state == eCompleted ){
        if (m_ObjectsLoaded) {
            IAppJob* job = GetJob().GetPointer();
            CDataLoadingAppJob* data_job = dynamic_cast<CDataLoadingJob*>(job);
            if (data_job->IsCanceled()) {
                return state;
            }

            _ASSERT(data_job);

            if( data_job->FailedToAddItems() ){
                // user chose "Decide Later" option or selected a project that has
                // been closed
                CRef<CScope> default_scope( new CScope( *CObjectManager::GetInstance() ) );
                default_scope->AddDefaults();

                // we need to show a dialog to ask the user where to place the data
                const vector< CRef<CProjectItem> > items = data_job->GetItems();
                TConstScopedObjects objects;
                for(  size_t i = 0;  i < items.size();  i++ )   {
                    const CProjectItem& item = *items[i];
                    CScope * scope = data_job->GetScope();
                    objects.push_back
                        (SConstScopedObject(*item.GetObject(), scope ? *scope : *default_scope));
                }

                string title = GetDescr();
                CProjectSelectorDlg dlg(NULL, wxID_ANY, ToWxString(title));

                dlg.SetProjectService(data_job->GetProjectService());
                dlg.SetObjects(objects);
                dlg.SetRegistryPath("Dialogs.DataLoadingTask.ProjectSelector");

                if(dlg.ShowModal() == wxID_OK)  {
                    SProjectSelectorParams params;
                    dlg.GetParams(params);

                    CDataLoadingOptions& options = data_job->GetOptions();
                    params.ToLoadingOptions(options);

                    // relaunch the job
                    m_State = eInitial;
                    m_JobID = CAppJobDispatcher::eInvalidJobID;
                    m_JobState = IAppJob::eInvalid;

                    state = CAppJobTask::Run();
                }
            }
        }
        else {
            m_ObjectsLoaded = true;

            CIRef<IExecuteUnit> execute_unit(dynamic_cast<IExecuteUnit*>(m_ObjectLoader.GetPointer()));
            if (!execute_unit || !execute_unit->PostExecute())
                return eCanceled;

            vector<CDataLoadingAppJob::TItemRef> items;
            NON_CONST_ITERATE (IObjectLoader::TObjects, it, m_ObjectLoader->GetObjects()) {
                CRef<CProjectItem> item(new CProjectItem());
                CSerialObject* so = dynamic_cast<CSerialObject*>(it->GetObjectPtr());
                if (so) {
                    item->SetObject(*so);
                    item->SetLabel(it->GetDescription());
                    items.push_back(item);
                }
            }

            m_Job.Reset(m_DataLoadingJob.GetPointer());

            CDataLoadingAppJob* data_job = dynamic_cast<CDataLoadingAppJob*>(GetJob().GetPointer());
            data_job->SetItems(items);

            m_State = eInitial;
            m_JobID = CAppJobDispatcher::eInvalidJobID;
            m_JobState = IAppJob::eInvalid;

            m_Descr = "Creating project items";

            state = CAppJobTask::Run();
        }
    }

    return state;
}


END_NCBI_SCOPE
