/*  $Id: project.cpp 27016 2012-12-07 21:34:14Z wuliangs $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors: Andrey Yazhuk
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>

#include <gui/core/project.hpp>

#include <gui/core/project_service.hpp>
#include <gui/core/ui_data_source_service.hpp>

#include <gui/objects/ProjectFolder.hpp>
#include <gui/objects/ProjectItem.hpp>
#include <gui/objects/FolderInfo.hpp>
#include <objects/general/Date.hpp>
#include <gui/objects/ProjectDescr.hpp>
#include <gui/utils/extension.hpp>
#include <gui/widgets/wx/async_call.hpp>

#include <serial/iterator.hpp>

#include <gui/objutils/registry.hpp>
#include <gui/utils/app_job_impl.hpp>

#include <gui/widgets/wx/async_call.hpp>
#include <gui/widgets/wx/wx_utils.hpp>

#include <util/icanceled.hpp>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);


////////////////////////////////////////////////////////////////////////////////
/// CProjectConstPrx

CProjectConstPrx::CProjectConstPrx(CGBDocument& project,
                                   CWorkspaceConstPrx* workspace,
                                   bool get_lock)
:   m_Document(&project),
    m_WorkspaceConst(workspace),
    m_HasLock(false)
{
    //LOG_POST(Info << "CProjectConstPrx constructor " << this);
    if(get_lock  &&  m_WorkspaceConst)    {
        m_WorkspaceConst->x_GetProjectsLock().ReadLock();
        m_HasLock = true;
    }

    m_OrigThreadID = CThread::GetSelf();
}


CProjectConstPrx::~CProjectConstPrx()
{
    //LOG_POST(Info << "CProjectConstPrx destructor " << this);
    // only release lock if we actually have it
    if(m_HasLock)   {
        m_WorkspaceConst->x_GetProjectsLock().Unlock();
    }
}


CRef<CGBDocument> CProjectConstPrx::x_GetDocument()
{
    x_ValidateCall();

    return m_Document;
}


CRef<CGBDocument> CProjectConstPrx::GetDocument()
{
    return x_GetDocument();
}


string CProjectConstPrx::GetViewSettings(const string& view_type,
                                         const string& view_ins_id) const
{
    return m_Document->GetViewSettings(view_type, view_ins_id);
}


void CProjectConstPrx::SaveViewSettings(const string& view_type,
                                        const string& view_ins_id,
                                        const string& settings)
{
    m_Document->SaveViewSettings(view_type, view_ins_id, settings);
}


CProjectConstPrx::TId CProjectConstPrx::GetId() const
{
    x_ValidateCall();

    return m_Document->GetId();
}


bool CProjectConstPrx::IsAttached() const
{
    x_ValidateCall();

    return m_WorkspaceConst != NULL;
}


bool CProjectConstPrx::IsLoaded() const
{
    x_ValidateCall();

    return m_Document->IsLoaded();
}


bool CProjectConstPrx::IsLoading() const
{
    x_ValidateCall();

    return m_Document->IsLoading();
}


bool CProjectConstPrx::IsDirty() const
{
    x_ValidateCall();

    bool dirty = m_Document->IsDirty();

    /// go through the views to see if we need to save anything
    if (!dirty) {
        ITERATE (TViews, iter, GetViews()) {
            const IView& view = **iter;
            const IProjectView* prj_view = dynamic_cast<const IProjectView*>(&view);
            if (prj_view) {
                if (prj_view->NeedSaveSettings()) {
                    dirty = true;
                    break;
                }
            }
        }
    }
    return dirty;
}


string CProjectConstPrx::GetName() const
{
    x_ValidateCall();

    const CProjectDescr& descr = m_Document->GetDescr();
    return descr.GetTitle();
}


string CProjectConstPrx::GetDescr() const
{
    x_ValidateCall();

    const CProjectDescr& descr = m_Document->GetDescr();
    return descr.CanGetComment() ? descr.GetComment() : "";
}


CTime CProjectConstPrx::GetCreateTime() const
{
    x_ValidateCall();

    const CProjectDescr& descr = m_Document->GetDescr();
    return descr.GetCreate_date().AsCTime();
}


CTime CProjectConstPrx::GetModifiedTime() const
{
    x_ValidateCall();

    const CProjectDescr& descr = m_Document->GetDescr();
    return descr.GetModified_date().AsCTime();
}


wxString CProjectConstPrx::GetFilename() const
{
    x_ValidateCall();

    return m_Document->GetFileName();
}


objects::CScope* CProjectConstPrx::GetScope()
{
    x_ValidateCall();

    return m_Document->GetScope();
}


CRef<CProjectFolderConstPrx> CProjectConstPrx::GetDataFolder_C()
{
    x_ValidateCall();

    CRef<CProjectFolderConstPrx>
        folder(new CProjectFolderConstPrx(m_Document->SetData(), *this));
    return folder;
}


CRef<CProjectItemConstPrx>
    CProjectConstPrx::FindProjectItemById_C(CProjectItemPrx::TId id)
{
    x_ValidateCall();

    CRef<CProjectFolderConstPrx> data = GetDataFolder_C();
    return data->FindProjectItemById_C(id, true);
}


CRef<CProjectFolderConstPrx>
    CProjectConstPrx::FindProjectFolderById_C(CProjectFolderPrx::TId id)
{
    x_ValidateCall();

    CRef<CProjectFolderConstPrx> prx = GetDataFolder_C();
    if(prx->GetId() == id)  {
        return prx;
    } else {
        return prx->FindChildFolderById_C(id, true);
    }
}


const CProjectConstPrx::TDataLoaders& CProjectConstPrx::GetDataLoaders() const
{
    x_ValidateCall();
    return m_Document->GetDataLoaders();
}


const CProjectConstPrx::TViews& CProjectConstPrx::GetViews(void) const
{
    x_ValidateCall();

    return m_Document->x_GetViews();
}


void CProjectConstPrx::x_ValidateCall() const
{
    CThread::TID id = CThread::GetSelf();

    if(id != m_OrigThreadID)    {
        _ASSERT(false);
        NCBI_THROW(CProjectServiceException, eThreadAffinityError,
            "Project Proxy can only be used on the thread on which\
            it was created.");
    }
}


////////////////////////////////////////////////////////////////////////////////
/// CProjectPrx
CProjectPrx::CProjectPrx(CGBDocument& project, CWorkspaceConstPrx* workspace)
:   CProjectConstPrx(project, workspace, false)
{
    //LOG_POST(Info << "CProjectPrx constructor " << this);
    if(m_WorkspaceConst)    {
        try {
            m_WorkspaceConst->x_GetProjectsLock().WriteLock();
            m_HasLock = true;
        } catch(CException& e)    {
            string s = e.GetMsg();
        }
    }
}


CProjectPrx::~CProjectPrx()
{
    //LOG_POST(Info << "CProjectPrx destructor " << this);
    // do not unlock here, ~CWorkspaceConstPrx will do it for us
}

void CProjectPrx::SetName( const string& name )
{
    x_ValidateCall();

    CProjectDescr& descr = m_Document->SetDescr();
    descr.SetTitle( name );
    x_OnProjectChanged( true, CProjectViewEvent::eProperties );
}

void CProjectPrx::SetDescr( const string& new_descr )
{
    x_ValidateCall();

    CProjectDescr& descr = m_Document->SetDescr();
    descr.SetComment( new_descr );
    x_OnProjectChanged( true, CProjectViewEvent::eProperties );
}

void CProjectPrx::SetCreateTime( const CTime& time )
{
    x_ValidateCall();

    CDate date( time );
    m_Document->SetCreateDate( date );
    x_OnProjectChanged( true, CProjectViewEvent::eProperties );
}

void CProjectPrx::SetModifiedTime( const CTime& time )
{
    x_ValidateCall();

    CDate date( time );
    m_Document->SetModifiedDate( date );
    x_OnProjectChanged( true, CProjectViewEvent::eProperties );
}


void CProjectPrx::SetFilename( const wxString& filename )
{
    x_ValidateCall();

    m_Document->SetFileName( filename );
    x_OnProjectChanged( true, CProjectViewEvent::eProperties );
}


CRef<CProjectFolderPrx> CProjectPrx::GetDataFolder()
{
    x_ValidateCall();

    CRef<CProjectFolderPrx>
        folder(new CProjectFolderPrx(m_Document->SetData(), *this));
    return folder;
}


CRef<CProjectItemPrx> CProjectPrx::FindProjectItemById(CProjectItemPrx::TId id)
{
    x_ValidateCall();

    CRef<CProjectFolderPrx> data = GetDataFolder();
    return data->FindProjectItemById(id, true);
}


CRef<CProjectFolderPrx> CProjectPrx::FindProjectFolderById(CProjectFolderPrx::TId id)
{
    x_ValidateCall();

    CRef<CProjectFolderPrx> prx = GetDataFolder();
    if(prx->GetId() == id)  {
        return prx;
    } else {
        return prx->FindChildFolderById(id, true);
    }
}


CRef<CLoaderDescriptor> CProjectPrx::FindDataLoaderByLabel(const string& label)
{
    x_ValidateCall();
    return m_Document->FindDataLoaderByLabel(label);
}


void CProjectPrx::MoveProjectFolder(CProjectFolderPrx& folder, CProjectFolderPrx& new_parent)
{
    x_ValidateCall();

    CRef<CProjectFolderPrx> old_parent = folder.GetParentFolder();
    old_parent->RemoveChildFolder(folder.GetId(), false);
    new_parent.x_AddChildFolder(*folder.m_Folder);
}


void CProjectPrx::MoveProjectItem(CProjectItemPrx& item, CProjectFolderPrx& new_parent)
{
    x_ValidateCall();

    CRef<CProjectFolderPrx> old_parent = item.GetParentFolder();
    // remove sub folder, but do not remove its contents
    old_parent->RemoveChildItem(item.GetId());
    new_parent.x_AddChildItem(*item.m_Item);
}

class CSaveProject_local : public CAsyncCall::Call
{
public:
    CSaveProject_local (CGBDocument& doc,
                        const wxString& path,
                        ESerialDataFormat fmt,
                        bool keep_backups)
                        : CAsyncCall::Call(wxT("Saving Project...")),
                          m_Doc(doc), m_Path(path), m_Fmt(fmt), m_KeepBackups(keep_backups) {}

    virtual void Execute() { m_Doc.SaveFile(m_Path, m_Fmt, m_KeepBackups); } 

private:
    CGBDocument& m_Doc;
    const wxString& m_Path;
    ESerialDataFormat m_Fmt;
    bool m_KeepBackups;
};

void CProjectPrx::Save()
{
    x_ValidateCall();

    CGuiRegistry& reg = CGuiRegistry::GetInstance();
    /// configuration parameters controlling serialization

    /// should the application keep back-up files?
    bool keep_backups =
        reg.GetBool("GBENCH.Application.IO.KeepBackups", true);

    /// do we serialize in text or binary mode?
    /// we have a compiled-in default and a user-override
    ESerialDataFormat fmt =
#ifdef _DEBUG
        eSerial_AsnText;
#else
        eSerial_AsnBinary;
#endif
    {{
        string fmt_str = reg.GetString("GBENCH.Application.IO.SerialFormat");
        if (NStr::CompareNocase(fmt_str, "text") == 0) {
            fmt = eSerial_AsnText;
        } else if (NStr::CompareNocase(fmt_str, "binary") == 0) {
            fmt = eSerial_AsnBinary;
        }
    }}

    // save view-specific settings
    NON_CONST_ITERATE (TViews, iter, GetViews()) {
        IView& view = **iter;
        IProjectView* prj_view = dynamic_cast<IProjectView*>(&view);
        if (prj_view) {
            if (prj_view->NeedSaveSettings()) {
                prj_view->SaveSettings();
            }
        }
    }

    wxString rel_path = m_Document->GetFileName();
    wxString abs_path = m_WorkspaceConst->ToAbsolutePath( rel_path );

    CSaveProject_local saveProject(*m_Document, abs_path, fmt, keep_backups);
    CAsyncCall::Execute(saveProject);

    //m_Document->SaveFile( abs_path, fmt, keep_backups );

    x_OnProjectChanged( false );
}


bool CProjectPrx::x_LoadData()
{
    x_ValidateCall();
    _ASSERT( ! m_Document->IsLoaded());

    // we need the fill path for reading the project
    wxString rel_path = m_Document->GetFileName();
    wxString abs_path = m_WorkspaceConst->ToAbsolutePath(rel_path);
    bool ok = m_Document->LoadFile(abs_path, 0);

    if(ok)  {
        _ASSERT(m_Document->GetScope() == NULL);

        m_Document->CreateProjectScope();
        x_AttachProjectItems();
        x_AttachDataLoaders();
    }
    return ok;
}

class CUnloadProject_local : public CAsyncCall::Call
{
public:
    CUnloadProject_local (CGBProjectHandle& projectHandle)
        : CAsyncCall::Call(wxT("Unloading project...")),
          m_ProjectHandle(projectHandle) {}

    virtual void Execute()
    {
        m_ProjectHandle.Clear();
    }
private:
    CGBProjectHandle& m_ProjectHandle;
};

void CProjectPrx::x_UnloadData()
{
    x_ValidateCall();
    _ASSERT(m_Document->IsLoaded());

    x_DetachDataLoaders();
    x_DetachProjectItems();

    CUnloadProject_local unloadProject(*m_Document);
    CAsyncCall::Execute(unloadProject);
    //m_Document->Clear(); // TODO rename
}


void CProjectPrx::x_AttachProjectItem(CProjectItem& child_item)
{
    x_ValidateCall();

    CScope* scope = GetScope();

    _ASSERT(scope);

    child_item.Attach(*scope);
    CDHManager::CreateDataHandle(child_item);
}

class CDetachProjectItem_local : public CAsyncCall::Call
{
public:
    CDetachProjectItem_local (CProjectItem& projectItem, CScope& scope)
        : CAsyncCall::Call(wxT("Unloading project item...")),
          m_ProjectItem(projectItem), m_Scope(scope) {}

    virtual void Execute()
    {
        m_ProjectItem.Detach(m_Scope);
    }
private:
    CProjectItem& m_ProjectItem;
    CScope& m_Scope;
};


void CProjectPrx::x_DetachProjectItem(CProjectItem& child_item)
{
    x_ValidateCall();

    CScope* scope = GetScope();

    _ASSERT(scope);

    CDetachProjectItem_local detachProject(child_item, *scope);
    CAsyncCall::Execute(detachProject);

    CDHManager::ReleaseDataHandle(child_item);
}


// Colelcts enaled projectItems
class CEnProjectItemCollector :
    public CProjectFolder::IProjectItemVisitor
{
public:
    virtual bool Visit(CProjectItem& item)
    {
        if(item.IsEnabled())    {
            m_Items.push_back(CRef<CProjectItem>(&item));
        }
        return true; // keep iterating until end
    }

    list< CRef<CProjectItem> >  m_Items;
};

void CProjectPrx::x_AttachProjectItems()
{
    //LOG_POST(Info << "CProjectPrx::x_AttachProjectItems() " << m_Document);
    x_ValidateCall();

    _ASSERT(m_Document->IsLoaded());

    CEnProjectItemCollector coll;
    CRef<CProjectFolderPrx> data = GetDataFolder();
    data->x_ForEachProjectItem(coll, true);

    NON_CONST_ITERATE(list< CRef<CProjectItem> >, it, coll.m_Items)    {
        CProjectItem& item = **it;
        x_AttachProjectItem(item);
    }
}


class CDetachProjectItems_local : public CAsyncCall::Call
{
public:
    CDetachProjectItems_local (list< CRef<CProjectItem> >& items, CScope& scope)
        : CAsyncCall::Call(wxT("Unloading project items...")),
          m_Items(items), m_Scope(scope) {}

    virtual void Execute()
    {
        NON_CONST_ITERATE(list< CRef<CProjectItem> >, it, m_Items) {
            CProjectItem& item = **it;
            item.Detach(m_Scope);
            CDHManager::ReleaseDataHandle(item);
        }
    }
private:
    list< CRef<CProjectItem> >& m_Items;
    CScope& m_Scope;
};

void CProjectPrx::x_DetachProjectItems()
{
    //LOG_POST(Info << "CProjectPrx::x_DetachProjectItems() " << m_Document);
    x_ValidateCall();

    _ASSERT(m_Document->IsLoaded());

    CEnProjectItemCollector coll;
    {{
        CRef<CProjectFolderPrx> data = GetDataFolder();
        data->x_ForEachProjectItem(coll, true);
    }}

    CScope* scope = GetScope();

    CDetachProjectItems_local detachItems(coll.m_Items, *scope);
    CAsyncCall::Execute(detachItems);
}


IDataLoaderProvider* CProjectPrx::x_GetDataSourceByType(const string& type)
{
    // get all UI data sources
    IServiceLocator* srv_locator = m_Document->m_Service->m_ServiceLocator;
    CIRef<CUIDataSourceService> srv =
        srv_locator->GetServiceByType<CUIDataSourceService>();
    CUIDataSourceService::TUIDataSourceVec data_sources;
    srv->GetDataSources(data_sources);
    
    // find the matched one
    IDataLoaderProvider* data_source = NULL;
    for (size_t i = 0;  i < data_sources.size();  ++i) {
        CIRef<IUIDataSource> ds = data_sources[i];
        IExtension* ext = dynamic_cast<IExtension*>(&ds->GetType());
        if (ext  &&  ext->GetExtensionIdentifier() == type) {
            data_source = dynamic_cast<IDataLoaderProvider*>(&*ds);
            break;
        }
    }

    return data_source;
}


bool CProjectPrx::x_AttachDataLoader(const CLoaderDescriptor& loader)
{
    x_ValidateCall();

    _ASSERT(m_Document->IsLoaded());

    IDataLoaderProvider* data_source =
        x_GetDataSourceByType(loader.GetLoader_type());

    if (data_source) {
        string loader_name = data_source->AddDataLoader(loader.GetData());
        if ( !loader_name.empty() ) {
            CScope::TPriority priority = CScope::kPriority_Default;
            if (loader.IsSetPriority()) {
                priority = loader.GetPriority();
            }
            GetScope()->ResetHistory();
            GetScope()->AddDataLoader(loader_name, priority);
            return true;
        }
    }

    return false;
}


bool CProjectPrx::x_DetachDataLoader(const CLoaderDescriptor& loader)
{
    x_ValidateCall();

    _ASSERT(m_Document->IsLoaded());

    IDataLoaderProvider* data_source =
        x_GetDataSourceByType(loader.GetLoader_type());

    if (data_source) {
        string loader_name = data_source->GetLoaderName(loader.GetData());
        if ( !loader_name.empty() ) {
            try {
                GetScope()->ResetHistory();
                GetScope()->RemoveDataLoader(loader_name);
                return true;
            } catch (CException& e) {
                LOG_POST(Error << "Can't detach the loader from scope: "
                         << loader.GetLabel() << ", Details: "
                         << e.GetMsg());
            }
        }
    }

    return false;
}


void CProjectPrx::x_AttachDataLoaders()
{
    //LOG_POST(Info << "CProjectPrx::x_AttachDataLoaders() " << m_Document);
    x_ValidateCall();

    _ASSERT(m_Document->IsLoaded());

    NON_CONST_ITERATE (TDataLoaders, it, m_Document->GetDataLoaders()) {
        CLoaderDescriptor& loader = **it;
        if (loader.GetEnabled()) {
            x_AttachDataLoader(loader);
        }
    }
}


void CProjectPrx::x_DetachDataLoaders()
{
    //LOG_POST(Info << "CProjectPrx::x_DetachDataLoaders() " << m_Document);
    x_ValidateCall();

    _ASSERT(m_Document->IsLoaded());

    NON_CONST_ITERATE (TDataLoaders, it, m_Document->GetDataLoaders()) {
        CLoaderDescriptor& loader = **it;
        if (loader.GetEnabled()) {
            x_DetachDataLoader(loader);
        }
    }
}


void CProjectPrx::x_FixProjectTitle()
{
    x_ValidateCall();

    CProjectDescr& descr = m_Document->SetDescr();
    string title = descr.SetTitle();
    if(title.empty())   {
        LOG_POST(Error << "CGBDocument - title not set");
        descr.SetTitle("Error - project title not set");
    }
}


string CProjectPrx::x_MakeUniqueItemName(const string& label) const
{
    x_ValidateCall();

    return m_Document->MakeUniqueItemLabel(label);
}


void CProjectPrx::x_OnProjectChanged(
    bool dirty, CProjectViewEvent::EEventSubtype subtype
){
    x_ValidateCall();

    m_Document->SetDirty( dirty );
    m_Document->x_OnProjectChanged( subtype ); //TODO should be conditional?

    if( m_WorkspaceConst ){
        m_WorkspaceConst->x_OnProjectChanged( *this );
    }
}


void  CProjectPrx::x_OnFolderChanged(
    CProjectFolderPrx& folder, CProjectViewEvent::EEventSubtype subtype
){
    x_ValidateCall();

    m_Document->SetDirty( true );
    m_Document->x_OnProjectChanged( subtype );

    if( m_WorkspaceConst ){
        m_WorkspaceConst->x_OnProjectChanged( *this ); //TODO incremental
    }
}


void CProjectPrx::x_OnItemChanged(
    CProjectItemPrx& item, CProjectViewEvent::EEventSubtype subtype
){
    x_ValidateCall();

    m_Document->SetDirty( true );
    m_Document->x_OnProjectChanged( subtype );

    if( m_WorkspaceConst ){
        m_WorkspaceConst->x_OnProjectChanged( *this ); //TODO incremental
    }
}


bool  CProjectPrx::IsDataLoaderAdded(const objects::CLoaderDescriptor& loader)
{
    x_ValidateCall();

    _ASSERT(m_Document->IsLoaded());

    const string& loader_type = loader.GetLoader_type();
    IDataLoaderProvider* data_source = x_GetDataSourceByType(loader_type);
    if ( !data_source ) return false;

    string loader_name = data_source->GetLoaderName(loader.GetData());
    
    NON_CONST_ITERATE (TDataLoaders, it, m_Document->GetDataLoaders()) {
        CLoaderDescriptor& curr_loader = **it;
        if (curr_loader.GetLoader_type() != loader_type) continue;

        string curr_name = data_source->GetLoaderName(curr_loader.GetData());
        if (curr_name == loader_name) {
            return true;
        }
    }

    return false;
}


void CProjectPrx::AddDataLoader(CLoaderDescriptor& loader)
{
    x_ValidateCall();

    string label = loader.GetLabel();
    m_Document->MakeUniqueDataLoaderLabel(label);
    loader.SetLabel(label);

    // add the loader descriptor to project
    m_Document->AddDataLoader(loader);

    // register the data loader to object manager, and
    // attach it to project scope
    if (x_AttachDataLoader(loader)) {
        // notify observers that the project data has been changed
        x_OnProjectChanged(true, CProjectViewEvent::eData);
    }
}


bool CProjectPrx::RemoveDataLoader(CLoaderDescriptor& loader)
{
    x_ValidateCall();
    if (!loader.GetEnabled()  ||  x_DetachDataLoader(loader)) {
        m_Document->RemoveDataLoader(loader);
        LOG_POST(Info << "Data loader: " << loader.GetLabel() << " is removed!");
        x_OnProjectChanged(true, CProjectViewEvent::eData);
        return true;
    }

    return false;
}


CProjectPrx::TDataLoaders& CProjectPrx::GetDataLoaders()
{
    x_ValidateCall();
    return m_Document->GetDataLoaders();
}


bool CProjectPrx::EnableDisableDataLoader(const string& label)
{
    CRef<CLoaderDescriptor> loader = m_Document->FindDataLoaderByLabel(label);
    if (loader->GetEnabled()) {
        if (x_DetachDataLoader(*loader)) {
            LOG_POST(Info << "Data loader: " << label << " is disabled!");
            loader->SetEnabled(false);
            x_OnProjectChanged(true, CProjectViewEvent::eData);
            return true;
        }
    } else {
        if (x_AttachDataLoader(*loader)) {
            LOG_POST(Info << "Data loader: " << label << " is enabled!");
            loader->SetEnabled(true);
            x_OnProjectChanged(true, CProjectViewEvent::eData);
            return true;
        }
    }

    return false;
}


void CProjectPrx::AttachView( IProjectView* view )
{
    x_ValidateCall();

    m_Document->x_AttachView(view);
}


void CProjectPrx::DetachView( IProjectView* view )
{
    x_ValidateCall();

    m_Document->x_DetachView(view);
}


CProjectPrx::TViews& CProjectPrx::GetViews(void)
{
    x_ValidateCall();

    return m_Document->x_GetViews();
}


void CProjectPrx::GetScopedObjects(TConstScopedObjects& objs)
{
    CEnProjectItemCollector coll;
    CRef<CProjectFolderPrx> data = GetDataFolder();
    data->x_ForEachProjectItem(coll, true);
    CScope* scope = GetScope();
    ITERATE(list< CRef<CProjectItem> >, it, coll.m_Items)    {
        const CProjectItem& item = **it;
        objs.push_back(SConstScopedObject(*item.GetObject(), *scope));
    }
}

void CProjectPrx::CancelLoading()
{
    x_ValidateCall();
    m_Document->CancelLoading();

    if( m_WorkspaceConst ){
        m_WorkspaceConst->x_OnProjectChanged( *this );
    }
}

void CProjectPrx::OnDocumentLoaded()
{
    x_ValidateCall();
    x_AttachProjectItems();
    x_AttachDataLoaders();
}

END_NCBI_SCOPE
