/*  $Id: project_cmd_handler.cpp 27556 2013-03-05 16:04:52Z katargir $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Andrey Yazhuk
 *
 */

#include <ncbi_pch.hpp>

#include <gui/framework/workbench.hpp>

#include <gui/core/project_cmd_handler.hpp>

#include <gui/core/commands.hpp>
#include <gui/core/project_service.hpp>
#include <gui/core/selection_service_impl.hpp>
#include <gui/core/project_task.hpp>
#include <gui/core/app_tasks.hpp>

#include <gui/framework/window_manager_service.hpp>
#include <gui/framework/app_task_service.hpp>

#include <gui/widgets/wx/wm_client.hpp>

#include <gui/objects/GBProject_ver2.hpp>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

BEGIN_EVENT_TABLE(CProjectCmdHandler, wxEvtHandler)
    EVT_MENU(wxID_SAVE, CProjectCmdHandler::OnSave )
    EVT_UPDATE_UI(wxID_SAVE, CProjectCmdHandler::OnSaveUpdate )
    EVT_MENU(wxID_SAVEAS, CProjectCmdHandler::OnSaveAs )
    EVT_UPDATE_UI(wxID_SAVEAS, CProjectCmdHandler::OnSaveAsUpdate )

    EVT_MENU(eCmdLoadProject, CProjectCmdHandler::OnLoadProject)
    EVT_UPDATE_UI(eCmdLoadProject, CProjectCmdHandler::OnUpdateLoadProject)
    EVT_MENU(eCmdUnLoadProject, CProjectCmdHandler::OnUnLoadProject)
    EVT_UPDATE_UI(eCmdUnLoadProject, CProjectCmdHandler::OnUpdateUnLoadProject)
    EVT_MENU(eCmdProjectTableView, CProjectCmdHandler::OnProjectTableView)
    EVT_UPDATE_UI(eCmdProjectTableView, CProjectCmdHandler::OnUpdateProjectTableView)

    EVT_MENU(eCmdRemoveProject, CProjectCmdHandler::OnRemoveProject)
    EVT_UPDATE_UI(eCmdRemoveProject, CProjectCmdHandler::OnUpdateRemoveProject)

    EVT_MENU(eCmdExportObject, CProjectCmdHandler::OnExport)
    EVT_UPDATE_UI(eCmdExportObject, CProjectCmdHandler::OnUpdateExport)
END_EVENT_TABLE()

CProjectCmdHandler::CProjectCmdHandler(IServiceLocator* srv_locator)
:   m_ServiceLocator(srv_locator)
{

}


CProjectCmdHandler::~CProjectCmdHandler()
{
}


void CProjectCmdHandler::OnSave(wxCommandEvent& event)
{
    bool workspace = false;
    TProjectIdVector projects;
    x_GetSelected(projects, &workspace);

    CProjectTask* task = new CProjectTask(m_ServiceLocator);
    task->Init_Save(workspace, projects);
    x_LaunchTask(task);
}


void CProjectCmdHandler::OnSaveUpdate(wxUpdateUIEvent& event)
{
    bool en = CProjectTask::Enable_Save(m_ServiceLocator);
    event.Enable(en);
}


void CProjectCmdHandler::OnSaveAs(wxCommandEvent& event)
{
    bool workspace = false;
    TProjectIdVector projects;
    x_GetSelected(projects, &workspace);

    CProjectTask* task = new CProjectTask(m_ServiceLocator);
    task->Init_SaveAs(workspace, projects);
    x_LaunchTask(task);
}


void CProjectCmdHandler::OnSaveAsUpdate(wxUpdateUIEvent& event)
{
    bool en = CProjectTask::Enable_SaveAs(m_ServiceLocator);
    event.Enable(en);
}


void CProjectCmdHandler::OnLoadProject(wxCommandEvent& event)
{
    x_LoadUnloadProject(true);
}


void CProjectCmdHandler::OnUpdateLoadProject(wxUpdateUIEvent& event)
{
    TProjectIdVector projects;
    x_GetSelected(projects);

    bool en = CProjectTask::Enable_LoadUnloadProjects(m_ServiceLocator, projects, true);
    event.Enable(en);
}


void CProjectCmdHandler::OnUnLoadProject(wxCommandEvent& event)
{
    x_LoadUnloadProject(false);
}


void CProjectCmdHandler::OnUpdateUnLoadProject(wxUpdateUIEvent& event)
{
    TProjectIdVector projects;
    x_GetSelected(projects);

    bool en = CProjectTask::Enable_LoadUnloadProjects(m_ServiceLocator, projects, false);
    event.Enable(en);
}

void CProjectCmdHandler::OnProjectTableView(wxCommandEvent& event)
{
    TProjectIdVector selected;
    x_GetSelected(selected);
    if (selected.empty())
        return;


    CIRef<CProjectService> prj_srv = m_ServiceLocator->GetServiceByType<CProjectService>();

    TConstScopedObjects objects;
    {{
        CRef<CWorkspaceConstPrx> workspace = prj_srv->GetWorkspaceConst();
        ITERATE(TProjectIdVector, it, selected)  {
            TProjectId id = *it;
            CRef<CProjectConstPrx> project = workspace->FindProjectById_C(id);
            if (project->IsLoaded()) {
                CRef<CGBDocument> doc = project->GetDocument();
                if (doc) {
                    const CGBProject_ver2* gbproj = dynamic_cast<const CGBProject_ver2*>(&doc->GetProject());
                    if (gbproj) {
                        objects.push_back(SConstScopedObject(*gbproj, *doc->GetScope()));
                    }
                }
            }
        }
    }}

    NON_CONST_ITERATE(TConstScopedObjects, it, objects)  {
        prj_srv->AddProjectView( "Project Table View", *it, 0 );
    }
}

void CProjectCmdHandler::OnUpdateProjectTableView(wxUpdateUIEvent& event)
{
    TProjectIdVector projects;
    x_GetSelected(projects);

    CIRef<CProjectService> prj_srv = m_ServiceLocator->GetServiceByType<CProjectService>();
    CRef<CWorkspaceConstPrx> workspace = prj_srv->GetWorkspaceConst();

    bool en = false;

    NON_CONST_ITERATE(TProjectIdVector, it, projects)  {
        TProjectId id = *it;
        CRef<CProjectConstPrx> project = workspace->FindProjectById_C(id);
        if (project->IsLoaded()) {
            en = true;
            break;
        }
    }
    event.Enable(en);
}

void CProjectCmdHandler::x_LoadUnloadProject(bool load)
{
    TProjectIdVector projects;
    x_GetSelected(projects);

    CProjectTask* task = new CProjectTask(m_ServiceLocator);
    if(load)    {
        task->Init_LoadProjects(projects);
    } else {
        task->Init_UnLoadProjects(projects);
    }
    x_LaunchTask(task);
}


void CProjectCmdHandler::OnRemoveProject(wxCommandEvent& event)
{
    TProjectIdVector selected, projects;
    x_GetSelected(selected);

    CIRef<CProjectService> srv = m_ServiceLocator->GetServiceByType<CProjectService>();
    CRef<CWorkspaceConstPrx> workspace = srv->GetWorkspaceConst();

    NON_CONST_ITERATE(TProjectIdVector, it, selected)  {
        TProjectId id = *it;
        CRef<CProjectConstPrx> project = workspace->FindProjectById_C(id);
        projects.push_back(id);
    }

    CProjectTask* task = new CProjectTask(m_ServiceLocator);
    task->Init_RemoveProjects(projects);
    x_LaunchTask(task);
}


void CProjectCmdHandler::OnUpdateRemoveProject(wxUpdateUIEvent& event)
{
    TProjectIdVector selected, projects;
    x_GetSelected(selected);

    CIRef<CProjectService> srv = m_ServiceLocator->GetServiceByType<CProjectService>();
    CRef<CWorkspaceConstPrx> workspace = srv->GetWorkspaceConst();

    NON_CONST_ITERATE(TProjectIdVector, it, selected)  {
        TProjectId id = *it;
        CRef<CProjectConstPrx> project = workspace->FindProjectById_C(id);
        projects.push_back(id);
    }

    event.Enable(projects.size() > 0);
}

CProjectService* CProjectCmdHandler::x_GetProjectService()
{
    _ASSERT(m_ServiceLocator);
    CProjectService* prj_srv = m_ServiceLocator->GetServiceByType<CProjectService>();
    _ASSERT(prj_srv);
    return prj_srv;
}


CAppTaskService* CProjectCmdHandler::x_GetAppTaskService()
{
    _ASSERT(m_ServiceLocator);
    CAppTaskService* task_srv = m_ServiceLocator->GetServiceByType<CAppTaskService>();
    _ASSERT(task_srv);
    return task_srv;
}


void CProjectCmdHandler::x_LaunchTask(IAppTask* task)
{
    CIRef<IAppTask> ref(task);

    _ASSERT(m_ServiceLocator);

    CAppTaskService* task_srv = m_ServiceLocator->GetServiceByType<CAppTaskService>();

    _ASSERT(task_srv);

    task_srv->AddTask(*ref);
}


void CProjectCmdHandler::x_GetSelected(TProjectIdVector& projects, bool* workspace)
{
    IWindowManagerService* srv =
        m_ServiceLocator->GetServiceByType<IWindowManagerService>();
    if(srv) {
        IWMClient* client = srv->GetActiveClient();
        IProjectSelectionSource* source =
                dynamic_cast<IProjectSelectionSource*>(client);
        if(source)  {
            if(workspace)   {
                *workspace = source->IsWorkspaceSelected();
            }
            source->GetSelectedProjectIds(projects);
        }
    }
}

void CProjectCmdHandler::OnExport(wxCommandEvent& event)
{
    /// TODO this is not a good solution, but simple
    IWorkbench* workbench = dynamic_cast<IWorkbench*>(m_ServiceLocator);

    CExportDlgTask* task = new CExportDlgTask(workbench);
    x_GetAppTaskService()->AddTask(*task);
}

void CProjectCmdHandler::OnUpdateExport(wxUpdateUIEvent& event)
{
    TConstScopedObjects objects;
    CIRef<CSelectionService> sel_srv = m_ServiceLocator->GetServiceByType<CSelectionService>();
    sel_srv->GetCurrentSelection(objects);
    event.Enable(objects.size() > 0);
}

END_NCBI_SCOPE
