/*  $Id: project_folder.cpp 24874 2011-12-15 20:07:42Z kuznets $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors: Andrey Yazhuk
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>

#include <gui/core/project_folder.hpp>

#include <gui/core/project_service.hpp>

#include <gui/objects/ProjectFolder.hpp>
#include <gui/objects/ProjectItem.hpp>
#include <gui/objects/FolderInfo.hpp>
#include <objects/general/Date.hpp>
#include <gui/objects/ProjectDescr.hpp>

#include <serial/iterator.hpp>


BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

////////////////////////////////////////////////////////////////////////////////
/// CProjectFolderConstPrx
CProjectFolderConstPrx::CProjectFolderConstPrx(CProjectFolder& folder,
                                               CProjectConstPrx& project)
:   m_Folder(&folder),
    m_ProjectConst(project)
{
}


CProjectFolderConstPrx::TId CProjectFolderConstPrx::GetId() const
{
    return m_Folder->GetId();
}


string CProjectFolderConstPrx::GetName() const
{
    const CFolderInfo& info = m_Folder->GetInfo();
    return info.GetTitle();
}


string CProjectFolderConstPrx::GetDescr() const
{
    const CFolderInfo& info = m_Folder->GetInfo();
    return info.GetComment();
}

bool CProjectFolderConstPrx::GetOpen() const
{
    const CFolderInfo& info = m_Folder->GetInfo();
    return info.IsSetOpen() ? info.GetOpen() : false;
}

CRef<CProjectItemConstPrx>
    CProjectFolderConstPrx::FindProjectItemById_C(CProjectItemConstPrx::TId id,
                                                  bool recursive)
{
    CRef<CProjectItemConstPrx> res;
    CProjectItem* item = m_Folder->FindProjectItemById(id, recursive);
    if(item)    {
        res.Reset(new CProjectItemConstPrx(*item, m_ProjectConst));
    }
    return res;
}


CRef<CProjectItemConstPrx>
    CProjectFolderConstPrx::FindProjectItemByData_C(const CSerialObject& object,
                                                  bool recursive)
{
    CRef<CProjectItemConstPrx> res;
    CProjectItem* item = m_Folder->FindProjectItemByData(object, recursive);
    if(item)    {
        res.Reset(new CProjectItemConstPrx(*item, m_ProjectConst));
    }
    return res;
}


CRef<CProjectFolderConstPrx>
    CProjectFolderConstPrx::FindChildFolderById_C(TId id, bool recursive)
{
    CRef<CProjectFolderConstPrx> res;
    CProjectFolder* folder = m_Folder->FindChildFolderById(id, recursive);
    if(folder)  {
        res.Reset(new CProjectFolderConstPrx(*folder, m_ProjectConst));
    }
    return res;
}


void CProjectFolderConstPrx::GetChildFolders_C(TFolderConstVector& folders)
{
    if(m_Folder->CanGetFolders()) {
        CProjectFolder::TFolders& sub_folders = m_Folder->SetFolders();

        NON_CONST_ITERATE(CProjectFolder::TFolders, it, sub_folders)  {
            CRef<CProjectFolderConstPrx> sub_folder(new CProjectFolderConstPrx(**it, m_ProjectConst));
            folders.push_back(sub_folder);
        }
    }
}


void CProjectFolderConstPrx::GetProjectItems_C(TProjectItemConstVector& prj_items)
{
    if(m_Folder->CanGetItems()) {
        CProjectFolder::TItems& items = m_Folder->SetItems();
        NON_CONST_ITERATE(CProjectFolder::TItems, it, items)    {
            CRef<CProjectItemConstPrx> pitem(new CProjectItemConstPrx(**it, m_ProjectConst));
            prj_items.push_back(pitem);
        }
    }
}

bool CProjectFolderConstPrx::
    x_ForEachProjectItem(CProjectFolder::IProjectItemVisitor& visitor,
                         bool recursive )
{
    return m_Folder->ForEachProjectItem(visitor, recursive);
}


////////////////////////////////////////////////////////////////////////////////
/// CProjectFolderPrx
CProjectFolderPrx::CProjectFolderPrx(objects::CProjectFolder& folder,
                                     CProjectPrx& project)
:   CProjectFolderConstPrx(folder, project),
    m_Project(&project)
{
}


void CProjectFolderPrx::SetName(const string& name)
{
    CFolderInfo& info = m_Folder->SetInfo();
    info.SetTitle(name);
    m_Project->x_OnFolderChanged( *this, CProjectViewEvent::eProperties );
}


void CProjectFolderPrx::SetDescr(const string& descr)
{
    CFolderInfo& info = m_Folder->SetInfo();
    info.SetComment(descr);
    m_Project->x_OnFolderChanged( *this, CProjectViewEvent::eProperties );
}

void CProjectFolderPrx::SetOpen(bool open)
{
    CFolderInfo& info = m_Folder->SetInfo();
    info.SetOpen(open);
}

void CProjectFolderPrx::GetChildFolders(TFolderVector& folders)
{
    if(m_Folder->CanGetFolders()) {
        CProjectFolder::TFolders& sub_folders = m_Folder->SetFolders();

        NON_CONST_ITERATE(CProjectFolder::TFolders, it, sub_folders)  {
            CRef<CProjectFolderPrx> sub_folder(new CProjectFolderPrx(**it, *m_Project));
            folders.push_back(sub_folder);
        }
    }
}


void CProjectFolderPrx::GetProjectItems(TProjectItemVector& prj_items)
{
    if(m_Folder->CanGetItems()) {
        CProjectFolder::TItems& items = m_Folder->SetItems();
        NON_CONST_ITERATE(CProjectFolder::TItems, it, items)    {
            CRef<CProjectItemPrx> pitem(new CProjectItemPrx(**it, *m_Project));
            prj_items.push_back(pitem);
        }
    }
}


CRef<CProjectFolderPrx> CProjectFolderPrx::GetParentFolder()
{
    CProjectFolder* parent = m_Folder->GetParentFolder();
    _ASSERT(parent);
    return CRef<CProjectFolderPrx>(new CProjectFolderPrx(*parent, *m_Project));
}


CRef<CProjectItemPrx> CProjectFolderPrx::AddProjectItem(const string& label,
                                                        CSerialObject& obj,
                                                        const string& descr)
{
    // Create a new Project Item
    CRef<CProjectItem> item(new CProjectItem());
    item->SetLabel(label);
    item->SetObject(obj);
    //item->SetDescr(descr); TODO

    return AddProjectItem(*item);
}


// TODO review
CRef<CProjectItemPrx>
    CProjectFolderPrx::AddProjectItem(objects::CProjectItem& item)
{
    // make sure item has a label, create it if needed
    bool has_label = item.IsSetLabel()  &&  ! item.GetLabel().empty();

    _ASSERT(has_label); // items must have labels

    string label;
    if (has_label)   {
        label = item.GetLabel();
    } else {
        // workaround for bad items
        CScope* scope = m_Project->GetScope();
        label = item.GetLabelByData(scope);
    }
    // make sure the label is unique
    CGBDocument* doc = m_Project->x_GetDocument();
    label = doc->MakeUniqueItemLabel(label);
    item.SetLabel(label);

    // correct label and add to the folder, do this via Document
    CProjectFolder& folder = *m_Folder;
    doc->AddItem(item, folder);

    m_Project->x_AttachProjectItem(item);

    m_Project->x_OnFolderChanged(*this, CProjectViewEvent::eData );

    CRef<CProjectItemPrx> res(new CProjectItemPrx(item, *m_Project));
    return res;
}


CRef<CProjectFolderPrx>
    CProjectFolderPrx::AddChildFolder(const string& name, const string& descr)
{
    // Create a new Project folder
    CRef<CProjectFolder> new_folder(new CProjectFolder());

    CFolderInfo& info = new_folder->SetInfo();
    info.SetTitle(name);
    info.SetComment(descr);
    info.SetCreate_date().SetToTime(CTime(CTime::eCurrent));

    // add the folder to the Project
    m_Folder->AddChildFolder(*new_folder);

    m_Project->x_OnFolderChanged( *this, CProjectViewEvent::eData );

    CRef<CProjectFolderPrx> res(new CProjectFolderPrx(*new_folder, *m_Project));
    return res;
}


void CProjectFolderPrx::RemoveChildFolder(TId id, bool remove_children)
{
    if(remove_children) {
        CRef<CProjectFolderPrx> sub_folder = FindChildFolderById(id, false);
        _ASSERT(sub_folder);
        sub_folder->x_RemoveChildren();
    }

    m_Folder->RemoveChildFolder(id);

    m_Project->x_OnFolderChanged( *this, CProjectViewEvent::eData );
}


void CProjectFolderPrx::RemoveChildItem(CProjectItemPrx::TId id)
{
    if(m_Folder->CanGetItems())  {
        {
            CRef<CProjectItemPrx> item = FindProjectItemById(id, true);
            m_Project->x_DetachProjectItem(*item->m_Item);
        }
        bool ok = m_Folder->RemoveChildItem(id);
        _ASSERT(ok);
        if(ok)  {
            m_Project->x_OnFolderChanged( *this, CProjectViewEvent::eData );
        }
    }
}


void CProjectFolderPrx::x_AddChildItem(CProjectItem& child_item)
{
    m_Project->x_AttachProjectItem(child_item);
    m_Folder->AddChildItem(child_item);
}


void CProjectFolderPrx::x_AddChildFolder(CProjectFolder& child_folder)
{
    m_Folder->AddChildFolder(child_folder);
}


void CProjectFolderPrx::x_RemoveChildren()
{
    // remove from the project all items contained by the folder
    m_Folder->RemoveAllChildItems();

    if(m_Folder->CanGetFolders()) {
        // remove from the project all sub folders of the current folder
        CProjectFolder::TFolders& folders = m_Folder->SetFolders();

        NON_CONST_ITERATE(CProjectFolder::TFolders, it_pf, folders)  {
            CRef<CProjectFolderPrx>
                sub_folder(new CProjectFolderPrx(**it_pf, *m_Project));
            sub_folder->x_RemoveChildren();
        }
    }
}


CRef<CProjectItemPrx>
    CProjectFolderPrx::FindProjectItemById(CProjectItemPrx::TId id,
                                                             bool recursive)
{
    CRef<CProjectItemPrx> res;
    CProjectItem* item = m_Folder->FindProjectItemById(id, recursive);
    if(item)    {
        res.Reset(new CProjectItemPrx(*item, *m_Project));
    }
    return res;
}


CRef<CProjectItemPrx>
    CProjectFolderPrx::FindProjectItemByData(const CSerialObject& object,
                                             bool recursive)
{
    CRef<CProjectItemPrx> res;
    CProjectItem* item = m_Folder->FindProjectItemByData(object, recursive);
    if(item)    {
        res.Reset(new CProjectItemPrx(*item, *m_Project));
    }
    return res;
}


CRef<CProjectItemPrx> CProjectFolderPrx::FindProjectItemByName(const string& name)
{
    CRef<CProjectItemPrx> res;
    CProjectItem* item = m_Folder->FindProjectItemByLabel(name);
    if(item)    {
        res.Reset(new CProjectItemPrx(*item, *m_Project));
    }
    return res;
}


CRef<CProjectFolderPrx>
    CProjectFolderPrx::FindChildFolderById(CProjectFolderPrx::TId id, bool recursive)
{
    CRef<CProjectFolderPrx> res;
    CProjectFolder* folder = m_Folder->FindChildFolderById(id, recursive);
    if(folder)  {
        res.Reset(new CProjectFolderPrx(*folder, *m_Project));
    }
    return res;
}


CRef<CProjectFolderPrx> CProjectFolderPrx::FindChildFolderByName(const string& name)
{
    CRef<CProjectFolderPrx> res;
    CProjectFolder* folder = m_Folder->FindChildFolderByTitle(name);
    if(folder)  {
        res.Reset(new CProjectFolderPrx(*folder, *m_Project));
    }
    return res;
}


END_NCBI_SCOPE
