/*  $Id: project_item.cpp 26413 2012-09-13 16:53:52Z katargir $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors: Andrey Yazhuk
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>

#include <gui/core/project_item.hpp>

#include <gui/core/project_service.hpp>


#include <gui/objects/ProjectFolder.hpp>
#include <gui/objects/ProjectItem.hpp>
#include <gui/objects/ProjectDescr.hpp>
#include <gui/objects/FolderInfo.hpp>
#include <objects/general/Date.hpp>
#include <objects/seq/Annotdesc.hpp>

#include <serial/iterator.hpp>


BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

////////////////////////////////////////////////////////////////////////////////
/// CProjectItemConstPrx
CProjectItemConstPrx::CProjectItemConstPrx(CProjectItem& item,
                                           CProjectConstPrx& project)
:   m_Item(&item),
    m_ProjectConst(project)
{
}

CProjectItemConstPrx::TId CProjectItemConstPrx::GetId() const
{
    return m_Item->GetId();
}


bool CProjectItemConstPrx::IsEnabled() const
{
    return m_Item->IsEnabled();
}


string CProjectItemConstPrx::GetName() const
{
    return m_Item->GetLabel();
}


string CProjectItemConstPrx::GetDescr() const
{
    CProjectItem* item = const_cast<CProjectItem*>(m_Item.GetPointer());

    list< CRef<objects::CAnnotdesc> >& descr_list = item->SetDescr();

    CRef<objects::CAnnotdesc> descr;
    ITERATE( list< CRef<objects::CAnnotdesc> >, descr_itr, descr_list ){
        if( (**descr_itr).IsComment() ){
            descr = *descr_itr;
        }
    }

    return descr.IsNull() ? "" : descr->SetComment();
}


const CSerialObject* CProjectItemConstPrx::GetObject() const
{
    return m_Item->GetObject();
}


CRef<CProjectFolderConstPrx> CProjectItemConstPrx::GetParentFolder_C()
{
    CProjectFolder* parent = m_Item->GetParentFolder();
    _ASSERT(parent);
    return CRef<CProjectFolderConstPrx>(new CProjectFolderConstPrx(*parent, m_ProjectConst));
}


////////////////////////////////////////////////////////////////////////////////
/// CProjectItemPrx

CProjectItemPrx::CProjectItemPrx(CProjectItem& item,
                                 CProjectPrx& project)
:   CProjectItemConstPrx(item, project),
    m_Project(&project)
{
}


void CProjectItemPrx::Enable(bool en)
{
    bool en_state = m_Item->CanGetDisabled() ? (! m_Item->GetDisabled()) : true;
    if (en != en_state) {
        if (en) {
            m_Project->x_AttachProjectItem(*m_Item);
        } else {
            m_Project->x_DetachProjectItem(*m_Item);
        }
        m_Item->SetDisabled( ! en);

        m_Project->x_OnItemChanged(*this, CProjectViewEvent::eData);
    }
}


void CProjectItemPrx::SetName(const string& name)
{
    if (m_Item->IsSetLabel() && m_Item->GetLabel() == name)
        return;

    string new_name = m_Project->x_MakeUniqueItemName(name);
    m_Item->SetLabel(new_name);

    m_Project->x_OnItemChanged(*this);
}


void CProjectItemPrx::SetDescr(const string& descr)
{
    list< CRef<objects::CAnnotdesc> >& descr_list = m_Item->SetDescr();

    CRef<objects::CAnnotdesc> andescr;
    ITERATE( list< CRef<objects::CAnnotdesc> >, descr_itr, descr_list ){
        if( (**descr_itr).IsComment() ){
            andescr = *descr_itr;
        }
    }

    if( andescr.IsNull() ){
        andescr.Reset( new objects::CAnnotdesc() );

        andescr->SetComment( descr );

        descr_list.push_back( andescr );
    
    } else {
        andescr->SetComment( descr );
    }

    m_Project->x_OnItemChanged(*this);
}

void CProjectItemPrx::SetObject(CSerialObject& object)
{
    m_Item->SetObject(object);
    m_Project->x_OnItemChanged(*this);
}

CRef<CProjectFolderPrx> CProjectItemPrx::GetParentFolder()
{
    CProjectFolder* parent = m_Item->GetParentFolder();
    _ASSERT(parent);
    return CRef<CProjectFolderPrx>(new CProjectFolderPrx(*parent, *m_Project));
}


void CProjectItemPrx::OnItemDataChanged()
{
    m_Project->x_OnItemChanged(*this);
}


END_NCBI_SCOPE
