/*  $Id: project_loading_task.cpp 25589 2012-04-11 19:11:04Z katargir $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Andrey Yazhuk
 *
 */

#include <ncbi_pch.hpp>

#include <gui/core/project_loading_task.hpp>
#include <gui/core/project.hpp>

#include <gui/core/project_service.hpp>
#include <gui/core/document.hpp>
#include <gui/core/loading_app_job.hpp>
#include <gui/core/save_dlg.hpp>

#include <gui/framework/workbench.hpp>
#include <gui/framework/app_task_service.hpp>
#include <gui/framework/app_job_task.hpp>

#include <gui/objects/GBWorkspace.hpp>
#include <gui/objects/ProjectDescr.hpp>
#include <gui/objects/WorkspaceFolder.hpp>

#include <gui/widgets/wx/message_box.hpp>
#include <gui/widgets/wx/wx_utils.hpp>

#include <corelib/ncbifile.hpp>

#include <wx/filedlg.h>
#include <wx/filename.h>
#include <wx/msgdlg.h>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

const wxChar* kWS_Wildcard = wxT("GBENCH Workspace (*.gbw)|*.gbw"); //TODO reuse
const wxChar* kPRJ_Wildcard = wxT("GBENCH Project (*.gbp)|*.gbp"); //TODO reuse

///////////////////////////////////////////////////////////////////////////////
/// CProjectLoadingTask

static const wxChar* kAlreadyExist =
wxT("\" already exists in the workspace. Please choose a different filename.");

SWFileDlgData
    CProjectLoadingTask::sm_SaveWorkspaceDlgData(wxT("Save Workspace As"),
                                                kWS_Wildcard,
                                                wxFD_SAVE | wxFD_OVERWRITE_PROMPT);

SWFileDlgData
    CProjectLoadingTask::sm_ProjectDlgData(wxT("Save Project As"),
                                          kPRJ_Wildcard,
                                          wxFD_SAVE | wxFD_OVERWRITE_PROMPT);

BEGIN_EVENT_MAP(CProjectLoadingTask, CEventHandler)
    ON_MESSAGE(CAppJobNotification, CAppJobNotification::eStateChanged,
               &CProjectLoadingTask::OnAppJobNotification)
    ON_MESSAGE(CAppJobNotification, CAppJobNotification::eProgress,
               &CProjectLoadingTask::OnAppJobProgress)
END_EVENT_MAP()

CProjectLoadingTask::CProjectLoadingTask(CProjectPrx::TId projId,
                                         CProjectService* srv,
                                         bool add_to_mru)
:   CAppTask("Loading project..."),
    m_TaskName("Load Project"),
    m_ProjId(projId),
    m_ProjectService(srv),
    m_AddToMRU(add_to_mru),
    m_JobID(CAppJobDispatcher::eInvalidJobID),
    m_JobState(IAppJob::eInvalid)
{
}

CProjectLoadingTask::~CProjectLoadingTask()
{
    if (m_JobID != CAppJobDispatcher::eInvalidJobID) {
        CAppJobDispatcher& disp = CAppJobDispatcher::GetInstance();
        disp.CancelJob(m_JobID);
    }
}

void CProjectLoadingTask::CancelBackgrounded()
{
    if(m_State == IAppTask::eBackgrounded)   {
        _ASSERT(m_JobID != CAppJobDispatcher::eInvalidJobID);
        CAppJobDispatcher& disp = CAppJobDispatcher::GetInstance();
        disp.CancelJob(m_JobID);
    }

    CRef<CWorkspacePrx> workspace = m_ProjectService->GetWorkspace();
    workspace->RemoveProject(m_ProjId);
}

IAppTask::ETaskState CProjectLoadingTask::Run()
{
    switch (m_State) {
    case eInitial:
        _ASSERT(m_JobID == CAppJobDispatcher::eInvalidJobID);
        _ASSERT(m_JobState == IAppJob::eInvalid);

        m_State = x_StartJob();
        m_JobState = IAppJob::eRunning;
        break;

    case eBackgrounded:
        // Job has been launched already - check whether its status has changed
        m_State = x_OnWakeUp();
        break;

    default:
        _ASSERT(false); // unexpected
        m_State = eInvalid;
        break;
    }
    return m_State;
}

IAppTask::ETaskState CProjectLoadingTask::x_StartJob()
{
    if (m_JobID != CAppJobDispatcher::eInvalidJobID)
        return eFailed;

    CRef<CWorkspacePrx> workspace = m_ProjectService->GetWorkspace();
    CRef<CProjectPrx> project = workspace->FindProjectById(m_ProjId);
    if (!project) {
        return eFailed;
    }

    CRef<CGBDocument> doc = project->GetDocument();
    if (!doc) {
        return eFailed;
    }

    wxString rel_path = doc->GetFileName();
    wxString abs_path = workspace->ToAbsolutePath(rel_path);
    CIRef<IAppJob> job(doc->CreateLoadJob(abs_path));
    if (!job) {
        return eFailed;
    }

    CAppJobDispatcher& disp = CAppJobDispatcher::GetInstance();
    m_JobID = disp.StartJob(*job, "ThreadPool", *this, 5, true);
    if (m_JobID == CAppJobDispatcher::eInvalidJobID)
        return eFailed;

    wxString prj_filename;
    if (!project.IsNull()) {
        prj_filename = workspace->ToAbsolutePath(project->GetFilename());
    }

    wxString fileName;
    if (prj_filename.StartsWith(wxT("http://")) || prj_filename.StartsWith(wxT("ftp://"))) {
        fileName = prj_filename;
    }
    else {
        fileName = wxFileName(prj_filename).GetFullName();
    }

    m_Descr = "Loading project: " + fileName;

    workspace->x_OnProjectChanged(*project);

    return eBackgrounded;
}

IAppTask::ETaskState CProjectLoadingTask::x_OnWakeUp()
{
    if (m_JobState == IAppJob::eRunning || m_JobState == IAppJob::eSuspended)
        return eBackgrounded;

    m_JobID = CAppJobDispatcher::eInvalidJobID;

    if (m_JobState == IAppJob::eCanceled)
        return eCanceled;

    CRef<CWorkspacePrx> workspace = m_ProjectService->GetWorkspace();
    CRef<CProjectPrx> project = workspace->FindProjectById(m_ProjId);
    CRef<CGBDocument> doc = project->GetDocument();
    doc->ResetLoading();

    if (m_JobState == IAppJob::eCompleted) {
        if (project) {
            project->OnDocumentLoaded();

            wxString mru_path;

            if(m_AddToMRU)  {
                //TODO synchronization needed here ?
                mru_path = project->GetFilename();
                mru_path = workspace->ToAbsolutePath(mru_path);
            }
            if (!mru_path.empty()) {
                m_ProjectService->AddToProjectWorkspaceMRUList(mru_path);
            }

            TConstScopedObjects objs;
            project->GetScopedObjects(objs);
            if (!objs.empty()) {
                /// prefetch sequence descriptions
                m_ProjectService->PrefetchSeqDescr(objs);
            }
            workspace->AddProject(*project);
        }

        return eCompleted;
    }
    else {
        // loading has not been successful - ask user what to do
        wxString prj_filename;
        if (!project.IsNull()) {
            prj_filename = workspace->ToAbsolutePath(project->GetFilename());
        }
        wxString msg = wxT("Cannot load project \"") + prj_filename;
        msg += wxT("\". Would you like to specify a different filename?");

        TDialogType type = eDialog_OkCancel | eDialog_Modal;
        EDialogReturnValue res = NcbiMessageBoxW(msg, type, eIcon_Exclamation,
                                                ToWxString(m_TaskName), eWrap);
        switch(res) {
        case eOK:
            break;
        default:
            workspace->RemoveProject(m_ProjId);
            return eFailed;
        }

        // keep asking until a correct file name is specified or operation is canceled
        while(true) {
            sm_ProjectDlgData.m_Style = wxFD_OPEN | wxFD_FILE_MUST_EXIST;
            sm_ProjectDlgData.m_Title =
                    ToWxString(m_TaskName) + wxT(" - choose a new filename for project");

            if(NcbiFileBrowser(sm_ProjectDlgData, NULL) == wxID_OK)  {
                // users specified a new filename
                wxString abs_path = sm_ProjectDlgData.GetFilename();
                wxString rel_path = workspace->ToRelativePath(abs_path);

                // look is we already have a project with this file name
                CRef<CProjectPrx> project_prx = workspace->FindProjectByFilename(rel_path);

                if(project_prx)  {
                    // a project with the selected file name already exists
                    msg = wxT("Project \"") + rel_path;
                    msg += kAlreadyExist;

                    TDialogType type = eDialog_Ok | eDialog_Modal;
                    NcbiMessageBoxW(msg, type, eIcon_Exclamation, ToWxString(m_TaskName), eWrap);
                    // remain in the loop to ask again
                } else {
                    // the filename is good - can use it to load the project
                    project->SetFilename(rel_path);
                    break; // exit the loop to load the project
                }
            } else {
                workspace->RemoveProject(m_ProjId);
                return eFailed;
            }
        } // while
    }

    return x_StartJob();
}

void CProjectLoadingTask::OnAppJobNotification( CEvent* evt )
{
    CAppJobNotification* notn = dynamic_cast<CAppJobNotification*>(evt);
    _ASSERT(notn);

    if( notn ){
        if( m_JobID !=  notn->GetJobID() ){
            _ASSERT(false);
            ERR_POST(
                "CProjectLoadingTask::OnAppJobNotification() - unknown or invalid Job ID " 
                << notn->GetJobID()
            );
            return;
        }
        m_JobState = notn->GetState();

        if (m_JobState == IAppJob::eCompleted ||
            m_JobState == IAppJob::eFailed ||
            m_JobState == IAppJob::eCanceled) {

            _ASSERT(m_Listener);
            if(m_Listener)  {
                CAppTaskServiceEvent::EType type = CAppTaskServiceEvent::eWakeUpSignal;
                if(m_JobState == IAppJob::eCanceled) {
                    type = CAppTaskServiceEvent::eTaskCanceled;
                }

                CRef<CEvent> evt(new CAppTaskServiceEvent(*this, type));
                m_Listener->Post(evt);
            }
        }
    }
}

void CProjectLoadingTask::OnAppJobProgress(CEvent* evt)
{
    CAppJobNotification* notn =
        dynamic_cast<CAppJobNotification*>(evt);
    _ASSERT(notn);

    if(notn)    {
        int job_id = notn->GetJobID();

        if(m_JobID != job_id) {
            _ASSERT(false);
            ERR_POST("CAppJobTask::OnAppJobNotification() - unknown or invalid Job ID " << job_id);
        } else {
            CConstIRef<IAppJobProgress> prg = notn->GetProgress();
        }
    }
}

END_NCBI_SCOPE

