/*  $Id: project_service.cpp 27556 2013-03-05 16:04:52Z katargir $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Mike DiCuccio, Andrey Yazhuk
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>

#include <gui/core/project_service.hpp>

#include <gui/core/project_view_base_impl.hpp>
#include <gui/core/data_store.hpp>
#include <gui/core/document.hpp>

#include <gui/framework/view_manager_service.hpp>
#include <gui/framework/app_task_service.hpp>
#include <gui/framework/app_job_task.hpp>

#include <gui/objutils/label.hpp>
#include <gui/objutils/registry.hpp>
#include <gui/objutils/obj_fingerprint.hpp>
#include <gui/objutils/utils.hpp>
#include <gui/objutils/obj_convert.hpp>
#include <gui/widgets/wx/wx_utils.hpp>

#include <gui/objects/PluginMRUEntry.hpp>
#include <gui/objects/PluginArgSet.hpp>
#include <gui/objects/GBProject.hpp>
#include <gui/objects/GBProject_ver2.hpp>
#include <gui/objects/GBWorkspace.hpp>
#include <gui/objects/ProjectDescr.hpp>
#include <gui/objects/ProjectItem.hpp>
#include <gui/objects/ProjectFolder.hpp>
#include <gui/objects/PluginRequest.hpp>
#include <gui/objects/PluginReply.hpp>
#include <gui/objects/WorkspaceFolder.hpp>
#include <gui/objects/ProjectItem.hpp>
#include <gui/objects/ProjectFolder.hpp>
#include <gui/objects/FolderInfo.hpp>
#include <gui/objects/ViewDescriptor.hpp>
#include <gui/objects/ViewDataLink.hpp>
#include <gui/utils/clipboard.hpp>
#include <gui/utils/app_job_impl.hpp>

#include <objects/biotree/BioTreeContainer.hpp>
#include <objects/submit/Seq_submit.hpp>
#include <objects/seqalign/Seq_align_set.hpp>
#include <objects/seqres/Seq_graph.hpp>
#include <objects/entrezgene/Entrezgene.hpp>
#include <objects/entrezgene/Entrezgene_Set.hpp>
#include <objects/general/Date.hpp>

#include <objmgr/util/sequence.hpp>
#include <objmgr/seq_feat_handle.hpp>
#include <objmgr/seq_graph_handle.hpp>
#include <objmgr/seq_align_handle.hpp>
#include <objmgr/seq_annot_ci.hpp>
#include <objmgr/object_manager.hpp>

#include <objtools/data_loaders/genbank/gbloader.hpp>
#include <objtools/cleanup/cleanup.hpp>

#include <serial/objistr.hpp>
#include <serial/objostr.hpp>
#include <serial/serial.hpp>
#include <serial/iterator.hpp>

#include <connect/ncbi_conn_stream.hpp>
#include <util/checksum.hpp>

#include <wx/msgdlg.h>
#include <wx/filename.h>
#include <gui/widgets/wx/message_box.hpp>


BEGIN_NCBI_SCOPE
USING_SCOPE(objects);


///////////////////////////////////////////////////////////////////////////////
/// CSeqDescrRetrievalJob - 
/// class.
class CSeqDescrRetrievalJob : public CAppJob
{
public:
    CSeqDescrRetrievalJob(const TConstScopedObjects& objs)
        : CAppJob("Retrieving sequence descriptions")
        , m_Objects(objs)
    {}

    /// @name IAppJob implementation
    /// @{
    virtual EJobState   Run();
    /// @}

private:
    TConstScopedObjects  m_Objects;
};


IAppJob::EJobState CSeqDescrRetrievalJob::Run()
{
    typedef vector< CConstRef<CSeq_id> > TSeqIds;

    TSeqIds seq_ids;
    const string target_type = "Seq-id";

    size_t total_ids = 0;
    CStopWatch sw;
    sw.Start();

    CScope scope(*CObjectManager::GetInstance());
    scope.AddDefaults();

    map<int, string> gi_descr_map;
    NON_CONST_ITERATE (TConstScopedObjects, obj_iter, m_Objects) {
        const CObject* obj = obj_iter->object.GetPointer();
        const CSeq_id* seq_id = dynamic_cast<const CSeq_id*>(obj);
        if (seq_id) {
            seq_ids.push_back(CConstRef<CSeq_id>(seq_id));
            ++total_ids;
        } else if (obj) {
            x_SetStatusText("Extracing seq-ids from the project item");
            CObjectConverter::TRelationVector relations;
            CObjectConverter::FindRelations(scope, *obj, target_type, relations);

            NON_CONST_ITERATE(CObjectConverter::TRelationVector, itr, relations) {
                if (IsCanceled()) {
                    m_Objects.clear(); // clear the array while in the background
                    return eCanceled;
                }

                /// If there are more than one relation, the first one contains
                /// all the conversions, and the rest contain categorized 
                /// conversions.
                if( relations.size() > 1 && itr == relations.begin() ){
                    continue;
                }

                CRelation& rel = **itr;
                string relName = rel.GetProperty(target_type);

                CRelation::TObjects related;
                rel.GetRelated(scope, *obj, related, CRelation::fConvert_NoExpensive, x_GetICanceled());

                ITERATE (CRelation::TObjects, rlt_iter, related) {
                    if (IsCanceled()) {
                        m_Objects.clear();
                        return eCanceled;
                    }

                    const CObject* converted_obj = &rlt_iter->GetObject();
                    const CSeq_id* seq_id = dynamic_cast<const CSeq_id*>(converted_obj);
                    if (seq_id) {
                        const string& descr = rlt_iter->GetDescription();
                        size_t pos = descr.find(":");
                        int gi = 0;
                        if (pos != string::npos) {
                            try {
                                gi = NStr::StringToInt(descr.substr(0, pos));
                                gi_descr_map[gi] = descr.substr(pos + 1);
                            } catch (CException&) {
                                // ignore;
                            }
                        }
                        if (gi == 0) {
                            seq_ids.push_back(CConstRef<CSeq_id>(seq_id));
                            ++total_ids;
                        }
                    }
                }
            }
        }
    }
    _TRACE("Get sequence title, extract seq-ids time: " << sw.Elapsed());
    sw.Restart();

    if (total_ids > 0) {
        size_t i = 1;
        ITERATE (TSeqIds, id_iter, seq_ids) {
            if (IsCanceled()) {
                m_Objects.clear();
                return eCanceled;
            }
            CLabelDescriptionCache::CacheSeqId(**id_iter, &scope);
            ++i;
            if ((i % 50) == 0) {
                string msg = "Retrieving descriptions: ";
                msg += NStr::SizetToString(i);
                msg += " of " + NStr::SizetToString(total_ids);
                x_SetStatusText(msg);
            }
            if ((i % 100) == 0) {
                scope.ResetDataAndHistory();
            }
        }
    }

    CLabelDescriptionCache::AddCaches(gi_descr_map);
    _TRACE(Info << "Get sequence title, get description total time: " << sw.Elapsed());

    // clear the array while in the background, save destructor time
    m_Objects.clear();

    return eCompleted;
}


CProjectService::CProjectService()
:   m_ServiceLocator(NULL),
    m_LockViewRecord(false)
{
    //LOG_POST("CProjectService constructor");
}


CProjectService::~CProjectService()
{
    //LOG_POST("CProjectService destructor");
}


void CProjectService::InitService()
{
    LOG_POST(Info << "Initializing Project Service...");

    TGuard lock(*this);
    LoadSettings();

    LOG_POST(Info << "Finished initializing Project Service");
}


void CProjectService::ShutDownService()
{
    LOG_POST(Info << "Shutting down Project Service...");

    TGuard lock(*this);

    /// this must be coordinated correctly - order here is important!
    DetachAllObservers();

    SaveSettings();

    /// get rid of our workspace
    ReleaseWorkspace();

    // clear the data store
    CDataStore::Clear();

    CDHManager::ShutDown(); //TODO ?

    LOG_POST(Info << "Finished shutting down Project Service");
}


void CProjectService::SetServiceLocator(IServiceLocator* locator)
{
    TGuard lock(*this);
    m_ServiceLocator = locator;
}


void CProjectService::SetRegistryPath(const string& path)
{
    TGuard lock(*this);
    m_RegPath = path;
}


static const char* kMRUTag = "MRUProjectsWorkspaces";

void CProjectService::SaveSettings() const
{
    _ASSERT(! m_RegPath.empty());

    if( ! m_RegPath.empty())   {
        // lock service to avoid races
        CProjectService& nc_this = const_cast<CProjectService&>(*this);
        TGuard lock(nc_this);

        CGuiRegistry& gui_reg = CGuiRegistry::GetInstance();
        CGuiRegistry::TReadWriteView view = gui_reg.GetReadWriteView(m_RegPath);

        // save MRU Project and Workspaces
        vector<string> values;
        // we package times and paths into the same vector (even elements - time, odd - path)
        CTimeFormat format =
            CTimeFormat::GetPredefined(CTimeFormat::eISO8601_DateTimeSec);

        const TMRUPathList::TTimeToTMap& map = m_ProjectWorkspaceMRUList.GetMap();
        ITERATE(TMRUPathList::TTimeToTMap, it, map)  {
            CTime time(it->first);
            time.ToLocalTime();
            string s_time = time.AsString(format);
            string filename = FnToStdString(it->second);
            values.push_back(s_time);
            values.push_back(filename);
        }
        view.Set(kMRUTag, values);
    }
}


void CProjectService::LoadSettings()
{
    _ASSERT(! m_RegPath.empty());

    if( ! m_RegPath.empty())   {
        // lock service to avoid races
        CProjectService& nc_this = const_cast<CProjectService&>(*this);
        TGuard lock(nc_this);

        CGuiRegistry& gui_reg = CGuiRegistry::GetInstance();
        CGuiRegistry::TReadView view = gui_reg.GetReadView(m_RegPath);

        // load MRU Project and Workspaces
        m_ProjectWorkspaceMRUList.Clear();

        vector<string> values;
        view.GetStringVec(kMRUTag, values);

        CTimeFormat format =
            CTimeFormat::GetPredefined(CTimeFormat::eISO8601_DateTimeSec);

        for(  size_t i = 0;  i + 1 < values.size() ;  )   {
            // process two elements at once
            string s_time = values[i++];
            wxString path = FnToWxString(values[i++]);
            CTime time(s_time, format);

            time_t t = time.GetTimeT();
            m_ProjectWorkspaceMRUList.Add(path, t);
        }
    }
}


void CProjectService::CreateNewWorkspace()
{
    if(m_Workspace)   {
        _ASSERT(false);
        NCBI_THROW(CProjectServiceException, eInvalidOperation,
                   "Cannot create a new Workspace - a workspace already exists");
    } else {
        {{
            TGuard  lock(*this);
            m_Workspace.Reset(new CGBWorkspace());

            /// determine a unique(ish) name for the workspace
            /// this is a generic tag followed by a counter
            /// we test this against the file system in the *current*
            /// directory (without bothering to test what current really means)

            static unsigned int counter = 1;
            string workspace_title;
            for (;;) {
                wxString str = wxString::Format(wxT("Workspace%u"), counter++);
                if (!wxFileName::FileExists(str + wxT(".gbw"))) {
                    workspace_title = str.ToAscii();
                    break;
                }
            }
            m_Workspace->SetDescr().SetTitle(workspace_title);

            CTime create_time(CTime::eCurrent);
            m_Workspace->SetDescr().SetCreateDate(create_time);
            m_Workspace->SetDescr().SetModifiedDate(create_time);
            m_Workspace->SetWorkspace().SetInfo().SetTitle(workspace_title);
            m_Workspace->SetWorkspace().SetInfo().SetCreateDate(create_time);
        }}

        x_NotifyObservers();
    }
}


bool CProjectService::HasWorkspace()
{
    TGuard  lock(*this);
    return (m_Workspace.GetPointer() != NULL);
}


CRef<CWorkspaceConstPrx> CProjectService::GetWorkspaceConst()
{
    CRef<CWorkspaceConstPrx> res;

    // spin-lock loop to acquire the CWorkspaceConstPrx
    do {
        try {
            TGuard  lock(*this); // gets unlocked if exception comes
            if(m_Workspace) {
                res.Reset(new CWorkspaceConstPrx(*m_Workspace, *this, true /*read lock*/));
            }
            break;
        } 
        catch (CWorkspaceConstPrx::CResourceBusyEx& )
        {
            SleepMilliSec(20);
        }
    } while(1);

    return res;
}


CRef<CWorkspacePrx> CProjectService::GetWorkspace()
{
    CRef<CWorkspacePrx> res;

    // spin-lock loop to acquire the CWorkspaceConstPrx
    do {
        try {
            TGuard  lock(*this); // gets unlocked if exception comes
            if(m_Workspace) {
                res.Reset(new CWorkspacePrx(*m_Workspace, *this));
            }
            break;
        }
        catch (CWorkspaceConstPrx::CResourceBusyEx& )
        {
            SleepMilliSec(20);
        }
    } while(1);

    return res;
}


void CProjectService::ReleaseWorkspace()
{
    TGuard  lock(*this);

    if (HasWorkspace()) {
        CRef<CWorkspacePrx> ws = GetWorkspace();
        ws->UnloadAllProjects();
        ws.Reset();

        m_Workspace.Reset();

        x_NotifyObservers();
    }
}

bool CProjectService::SaveWorkspace()
{
    //LOG_POST(Info << "CProjectService::SaveWorkspace()");

    static const string kErr("Error saving workspace");

    static const string INVALID_WORKSPACE =
        "Invalid workspace! The Workspace contains projects "
        "with empty filenames.";

    TGuard  lock(*this);
    string err_msg;

    try {
        if( ! HasWorkspace())   {
            _ASSERT(false);
            NCBI_THROW(CProjectServiceException, eInvalidOperation,
                       "Cannot save workspace - it does not exists.");
        }

        CRef<CWorkspacePrx> ws = GetWorkspace();
        wxString filename = ws->GetFilename();

        // check that all projects have valid filename
        CWorkspaceConstPrx::TProjectPrxVector projects;
        ws->GetProjects(projects);

        NON_CONST_ITERATE(CWorkspaceConstPrx::TProjectPrxVector, it, projects)   {
            CProjectPrx& project = **it;
            if(project.GetFilename().empty())   {
                //_ASSERT(false); // must not be empty at this point
                NCBI_THROW(CProjectServiceException, eInvalidOperation, INVALID_WORKSPACE);
            }
        }

        CNcbiOfstream ostr(filename.fn_str());
        auto_ptr<CObjectOStream> os(CObjectOStream::Open(eSerial_AsnText, ostr));
        CGBWorkspace& ws_obj = *ws->m_Workspace;
        *os << ws_obj;

        ws_obj.SetDirty(false);
        x_OnWorkspaceChanged();

        return true;
    }
    catch (CException& e) {
        err_msg = e.GetMsg();
    }
    catch (std::exception& e) {
        err_msg = e.what();
    }
    wxMessageBox(ToWxString(err_msg), wxT("Error saving workspace"),
                 wxOK|wxICON_ERROR);

    NCBI_THROW(CException, eUnknown, kErr + " : \n" + err_msg);
    return false;
}


bool CProjectService::LoadWorkspace(const wxString& filename)
{
    static const string kFailed("Failed to load workspace: ");

    string err_msg;
    try {
        CRef<CGBWorkspace> ws = x_LoadWorkspace(filename);

        if (ws) {
            TGuard  lock(*this);
            m_Workspace = ws;

            CRef<CWorkspacePrx> ws_prx = GetWorkspace();
            ws_prx->SetFilename(filename);
            ws->SetDirty(false);
            return true;
        }
    }
    catch (CException& e) {
        err_msg = kFailed + ToStdString(filename) + " - "  + e.GetMsg();
    }
    catch (std::exception& e) {
        err_msg = kFailed + ToStdString(filename) + " - " + e.what();
    }
    NCBI_THROW(CProjectServiceException, eOperationFailed, err_msg);
    return false;
}


CRef<CGBWorkspace> CProjectService::x_LoadWorkspace(const wxString& filename)
{
    /// in order, we try ASN.1 text, ASN.1 binary, and XML
    /// chances are it'll always be ASN.1 text, but we can try the
    /// others for completeness
    typedef pair<ESerialDataFormat, const char*> TSerialTypePair;
    static const TSerialTypePair sc_DataTypes[] = {
        TSerialTypePair(eSerial_AsnText,    "asn-text"),
        TSerialTypePair(eSerial_AsnBinary,  "asn-binary"),
        TSerialTypePair(eSerial_Xml,        "asn-xml")
    };

    CRef<CGBWorkspace> wks;

    size_t n = sizeof(sc_DataTypes) / sizeof(TSerialTypePair);
    for (size_t i = 0;  i < n;  ++i) {
        const char* type_name = sc_DataTypes[i].second;
        try {
            wks.Reset(new CGBWorkspace());

            ESerialDataFormat format = sc_DataTypes[i].first;
            bool binary = (format == eSerial_AsnBinary);
            CNcbiIfstream istr(filename.fn_str(), (binary ? (ios::binary|ios::in) : ios::in));
            auto_ptr<CObjectIStream> obj_istr(CObjectIStream::Open(format, istr));
            *obj_istr >> *wks;

            wks->LinkFolders();

            // HACK: promoting CGBPorjectHandles to CGBDocuments
            CGBWorkspace::TWorkspace& root_folder = wks->SetWorkspace();
            x_CreateDocuments(root_folder);

            LOG_POST(Info << "CProjectService::x_LoadWorkspace(): loaded workspace: "
                << filename << ": " << type_name);
            break;
        }
        catch (CSerialException& e) {
            wks.Reset();
            LOG_POST(Error << "CProjectService::x_LoadWorkspace(): workspace: "
                << filename << ": not " << type_name
                << ": error reading: " << e.GetMsg());
        }
    }

    return wks;
}


// recursively iterates folders and wraps Project Handles in Documents
void CProjectService::x_CreateDocuments(CWorkspaceFolder& folder)
{
    CWorkspaceFolder::TProjects& projects = folder.SetProjects();

    NON_CONST_ITERATE(CWorkspaceFolder::TProjects, it, projects)  {
        CRef<CGBProjectHandle>& handle = *it;
        // create wrapping Document
        CRef<CGBDocument> doc(new CGBDocument(this));
        doc->Assign(*handle);

        // replace handle with the document
        handle.Reset(doc);
    }

    CWorkspaceFolder::TFolders& folders = folder.SetFolders();
    NON_CONST_ITERATE(CWorkspaceFolder::TFolders, it, folders)  {
        CWorkspaceFolder& child_folder = **it;
        x_CreateDocuments(child_folder);
    }
}


void CProjectService::x_RestoreViews(CScope& scope, CGBDocument& project)
{
    /*
    // there's views stored in this project
    if (project.GetProject().IsSetViews()) {
        IGBProject::TViews & vs = project.SetProject().SetViews();

        IGBProject::TViews vset;
        vset.assign(vs.begin(), vs.end());
        vs.clear();

        NON_CONST_ITERATE(IGBProject::TViews, descr, vset) {
            if (!(*descr)->IsSetData()){
                continue;
            }

            // data for view opening
            const CViewDescriptor::TData & data = (*descr)->GetData();

            // objects required for view opening
            TConstScopedObjects         objs;

            // plugin name
            string plugin_name = (*descr)->GetView_plugin();

            // view launch info reading
            if (data.IsItem()) {
                SConstScopedObject  obj(data.GetItem(), scope);
                objs.push_back(obj);
            }
            else if (data.IsLink()){
                // retrieving link
                const CViewDataLink & link = data.GetLink();

                // ProjectItem
                const CProjectItem * prj_item = NULL;

                // project item id
                int    project_item_id  = link.CanGetProject_item_id() ?
                                          link.GetProject_item_id() : -1;

                // fingerprint
                int fingerprint = link.CanGetFingerprint()?
                                    link.GetFingerprint().GetCrc32() : 0;

                // type
                // TODO: proceed object conversion to specified type
                // const string & type = link.GetType();

                // locating project item
                if (project_item_id>0) {
                      CTypeIterator<CProjectItem> item_iter(project.SetData());
                       for ( ;  item_iter;  ++item_iter) {
                           CProjectItem& item = *item_iter;
                           if (item.GetId() == project_item_id) {
                                prj_item = &item;
                                break;
                           }
                       }
                }

                if (!fingerprint) {
                    // no fingerprint, but project item id retrieved through DataHandle
                    // adding a project item as datasource
                    if (prj_item) {
                         SConstScopedObject  obj(*prj_item, scope);
                         objs.push_back(obj);
                    }
                    else {
                        // adding nothing here - project itself will be added later
                        continue;
                    }
                }
                else {
                    if (prj_item) {
                        // comparing fingerprints
                        CBioseq_Handle handle = scope.GetBioseqHandle(prj_item->GetItem().GetId());
                        SAnnotSelector sel = CSeqUtils::GetAnnotSelector();
                        sel.SetResolveAll().SetNoMapping().SetSortOrder(sel.eSortOrder_None);
                        sel.SetAnnotType(CSeq_annot::TData::e_Ftable);
                        for (CFeat_CI iter(handle, sel);  iter;  ++iter) {
                           CChecksum cs;
                           CObjFingerprint::GetFingerprint(cs, iter->GetOriginalFeature(), &scope);
                           Uint4 crc = cs.GetChecksum();
                           if (crc == fingerprint) {
                               SConstScopedObject  obj(iter->GetOriginalFeature(), scope);
                               objs.push_back(obj);
                           }
                        }
                    }
                    else {
                        // nothing here for right now
                        continue;
                    }
                }
            }

            // if no other concrete object found to open a view with
            // try to open it on a project
            if (objs.empty()) {
                SConstScopedObject prjobj(project, scope);
                objs.push_back(prjobj);
            }

            // launch view
            if (!objs.empty()) {
                // calling plugin
                CPluginUtils::CallPlugin(plugin_name,
                                         CPluginCommand::eCommand_new_view,
                                         objs);
            }
            else {
                LOG_POST(Error << "CProjectService::x_RestoreViews: "
                    "unable to locate objects associated with view, view will not be opened");
            }
        }
    }
    */
}


const CProjectService::TMRUPathList&
    CProjectService::GetProjectWorkspaceMRUList() const
{
    CGuard<const CObservable>  lock(*this);
    return m_ProjectWorkspaceMRUList;
}


void CProjectService::AddToProjectWorkspaceMRUList(const wxString& path)
{
    TGuard lock(*this);
    m_ProjectWorkspaceMRUList.Add(path);
}


void CProjectService::x_ConnectDocument(CGBDocument& doc)
{
    {{
        /// make our project visible to the dispatcher
        CMutexGuard lock(m_DispMutex);
        m_Disp.AddListener(&doc);
        doc.AddListener(&m_Disp);
    }}
}


void CProjectService::x_DisconnectDocument(CProjectPrx& project, bool reset_hist_async)
{
    RemoveAllProjectViews(project, reset_hist_async); //TODO synchronize ??

    {{
        /// detach from our dispatcher
        CGBDocument* doc = project.x_GetDocument();

        CMutexGuard lock(m_DispMutex);
        doc->RemoveListener(&m_Disp);
        m_Disp.RemoveListener(doc);
    }}
}


void CProjectService::x_OnWorkspaceChanged() //TODO
{
    _ASSERT(HasWorkspace());

    if(HasWorkspace())  {
        CWorkspaceConstPrx::TId id;
        {
            CRef<CWorkspacePrx> workspace = GetWorkspace();
            id = workspace->GetId();
        }
        x_OnWorkspaceChanged(id);
    }
}


void CProjectService::x_OnWorkspaceChanged(CWorkspaceConstPrx::TId id)
{
    //LOG_POST("CProjectService::x_OnWorkspaceChanged() " << id);
    CMutexGuard mtx_guard(m_UpdateMutex);   // now it is safe to check the counter

    m_ModifiedWorkspaces.insert(id);

    x_NotifyObservers();
}


void CProjectService::x_OnProjectChanged(CProjectPrx::TId id)
{
    //LOG_POST("CProjectService::x_OnProjectChanged() " << id);

    CMutexGuard mtx_guard(m_UpdateMutex); // now it is safe to check the counter

    m_ModifiedProjects.insert(id); //TODO synchronize

    x_NotifyObservers();
}


IObserverUpdate* CProjectService::x_CreateStateChange()
{
    CProjectServiceUpdate* update = new CProjectServiceUpdate();
    update->m_WorkspaceIds = m_ModifiedWorkspaces;
    update->m_ProjectIds = m_ModifiedProjects;

    // clean the accumulated state
    m_ModifiedWorkspaces.clear();
    m_ModifiedProjects.clear();

    return update;
}

// overriding CObservable::x_PostUpdateEvent()
/*
void CProjectService::x_PostUpdateEvent()
{
    //LOG_POST("CProjectService::x_PostUpdateEvent()");

    CRef<CProjectServiceUpdate> update(new CProjectServiceUpdate(this));
    update->m_WorkspaceIds = m_ModifiedWorkspaces;
    update->m_ProjectIds = m_ModifiedProjects;

    // clean the accumulated state
    m_ModifiedWorkspaces.clear();
    m_ModifiedProjects.clear();

    CRef<CEvent> event(new CObservableUpdateEvent(*update));
    Post(event); // send asynchronously to myself
    //LOG_POST("CProjectService::x_PostUpdateEvent()  END");
}
*/


CRWLock& CProjectService::x_GetWorkspaceLock()
{
    return m_WorkspaceLock;
}


CRWLock& CProjectService::x_GetProjectsLock()
{
    return m_ProjectsLock;
}


bool CProjectService::x_CanBeSavedAsLink(const CObject& obj,
                                         const CGBDocument& project)
{
    // TODO: implement it
    if ( (typeid(obj) == typeid(CSeq_feat)) ||
        (typeid(obj) == typeid(CBioTreeContainer))){
            return true;
    }
    return false;
}


void sAssertMainThread(const char* err_msg)
{
    bool main_thread = (CThread::GetSelf() == 0);
    if( ! main_thread)  {
        _ASSERT(false);
        NCBI_THROW(CProjectServiceException, eThreadAffinityError, err_msg);
    }
}


CIRef<IProjectView> CProjectService::AddProjectView(
    const string& view_name,
    SConstScopedObject& object,
    const objects::CUser_object* params
){
    TConstScopedObjects objects;
    objects.push_back(object);

    return AddProjectView( view_name, objects, params );
}

CIRef<IProjectView> CProjectService::AddProjectView(
    const string& view_name, 
    TConstScopedObjects& objects,
    const objects::CUser_object* params
){
    sAssertMainThread("Views can be created only from the main thread");

    IViewManagerService* view_srv =
        m_ServiceLocator->GetServiceByType<IViewManagerService>()
    ;

    _ASSERT(view_srv);

    CIRef<IView> view( view_srv->CreateViewInstance(view_name) );
    if( !view ){
        return null;
    }

    CIRef<IProjectView> prj_view(dynamic_cast<IProjectView*>(view.GetPointer()));
    if( prj_view ){
        view_srv->AddToWorkbench(*view);

        bool ok = false;
        try {
            // initialization logging
            //
            {
                const CViewTypeDescriptor& vdescr = prj_view->GetTypeDescriptor();
                LOG_POST(Info << "InitView = " << vdescr.GetLabel());

                int ix = 0;
                int trace_limit = 10;
                string arg_label;

                ITERATE( TConstScopedObjects, it, objects ){
                    #ifdef _DEBUG
                        trace_limit = 100;
                    #endif

                    if( ix++ == trace_limit ){
                        LOG_POST(Info << "   and " << (objects.size() - trace_limit) << " objects more" );

                        break;
                    }
                    const CObject* obj = it->object.GetPointer();
                    CScope* scope = const_cast<CScope*>(it->scope.GetPointer());
                    arg_label.clear();
                    CLabel::GetLabel(*obj, &arg_label, CLabel::eDefault, scope);
                    LOG_POST(Info << "    object=" << arg_label);

                }
            }

            ok = prj_view->InitView(objects, params);

        } catch( CException& e ){
            LOG_POST( Error << e.what() );
            #ifdef _DEBUG
            NcbiErrorBox( e.what(), "Error while opening view" );
            #else
            NcbiErrorBox( e.GetMsg(), "Open View Error" );
            #endif
        }

        if( ok ){

            OnViewAttached(prj_view);

            CProjectViewBase* prj_view_base = 
                dynamic_cast<CProjectViewBase*>(prj_view.GetPointer())
            ;
            if( prj_view_base ){
                prj_view_base->RefreshViewWindow();
            }
        
        } else {
            view_srv->RemoveFromWorkbench(*view);
            prj_view.Reset();
        }
    } else {
        _ASSERT(false);
        NCBI_THROW(CProjectServiceException, eInvalidArguments,
                   "Created view is not a Project View");
    }
    return prj_view;
}


void CProjectService::RemoveProjectView(IProjectView& view)
{
    sAssertMainThread("Views can be removed only from the main thread");

    IViewManagerService* view_srv = m_ServiceLocator->GetServiceByType<IViewManagerService>();

    _ASSERT(view_srv);

    x_RemoveView(*view_srv, view);

    OnViewDetached(&view);
}


void CProjectService::RemoveAllProjectViews(CProjectPrx& project, bool reset_hist_async)
{
    sAssertMainThread("Views can be removed only from the main thread");

    CProjectPrx::TViews views = project.GetViews();

    IViewManagerService* view_srv = m_ServiceLocator->GetServiceByType<IViewManagerService>();

    for(  size_t i = 0;  i < views.size();  i++ )   {
        IProjectView& view = *views[i];
        x_RemoveView(*view_srv, view, reset_hist_async);
    }

	x_NotifyObservers();
}


void CProjectService::RemoveAllProjectViews()
{
    sAssertMainThread("Views can be removed only from the main thread");

    IViewManagerService* view_srv = m_ServiceLocator->GetServiceByType<IViewManagerService>();

    _ASSERT(view_srv);

    IViewManagerService::TViews views;
    view_srv->GetViews(views);

    for(  size_t i = 0;  i < views.size();  i++ )   {
        IView& view = *views[i];
        IProjectView* prj_view = dynamic_cast<IProjectView*>(&view);
        if(prj_view)    {
            x_RemoveView(*view_srv, *prj_view);
        }
    }

	x_NotifyObservers();
}


void CProjectService::x_RemoveView(IViewManagerService& view_srv, 
                                   IProjectView& view, 
                                   bool reset_hist_async)
{
    CIRef<IProjectView> hold(&view);

    // disconnect from the document
    view.SetAsyncDestroy(reset_hist_async);
    view.DestroyView();

    // disconnect for Workbench (removes it from Window Manager)
    view_srv.RemoveFromWorkbench(view);
}


// from plug-in message, should be dropped soon after extending IProjectView interface
void CProjectService::RecordViewOpen(IProjectView * view)
{
    /* TODO
    // creating view desctiptor
    CRef <CViewDescriptor> view_desc(new CViewDescriptor());
    view_desc->SetView_id(view->GetId());

    // getting to data
    IGBDocument* idoc = view->GetDocument();
    CGBDocument* doc = dynamic_cast<CGBDocument*>(idoc); //TODO
    TConstScopedObjects objs;
    view->GetMainObject(objs);

    ITERATE(TConstScopedObjects, it, objs) {
        const CSerialObject * obj =
            dynamic_cast<const CSerialObject*>((it->object).GetPointer());
        if (!obj) {
            continue;
        }

        if (x_CanBeSavedAsLink(*obj, *doc)) {
            // link
            CRef<CViewDataLink> data_link(new CViewDataLink());

            // adding project item id
            const CDataHandle * dh = view->GetDataHandle();
            if (dh) {
                data_link->SetProject_item_id(dh->GetItem().GetId());
            }

            // calculating checksum, or other locator here
            CChecksum checksum;
            CObjFingerprint::GetFingerprint(checksum, *obj);
            if (checksum.GetChecksum()) {
                data_link->SetFingerprint().SetCrc32(checksum.GetChecksum());
            }

            // saving type
            data_link->SetType(obj->GetThisTypeInfo()->GetName());

            // adding link
            view_desc->SetData().SetLink(*data_link);
        }
        else {
            // saving project item
            CRef<CProjectItem> proj_item(new CProjectItem());
            proj_item->SetId(0);
            const CSerialObject * ser_obj = dynamic_cast<const CSerialObject *>(obj);
            if (ser_obj) {
                proj_item->SetObject(*const_cast<CSerialObject*>(ser_obj));
            }
            view_desc->SetData().SetItem(*proj_item);
        }
    }

    // adding unique view TODO - do not access document's internals
    IGBProject::TViews & vset = doc->SetProject().SetViews();
    bool bAddView = true;
    NON_CONST_ITERATE(IGBProject::TViews, view_desc, vset) {
        if ((*view_desc)->GetView_id() == view->GetId()) {
            LOG_POST(Error << "CProjectService::RecordViewOpen: "
                "Attempt to register view, which is already exist. Action ignored.");
            bAddView = false;
            break;
        }
    }
    if (bAddView) {
        vset.push_back(view_desc);
    }*/
}


void CProjectService::RecordViewClose(IProjectView * view)
{
    /* TODO
    IGBDocument* idoc = view->GetDocument();
    CGBDocument* doc = dynamic_cast<CGBDocument*>(idoc); //TODO

    //TODO - this a violation of encapsulation !
    IGBProject::TViews& vset = doc->SetProject().SetViews();
    NON_CONST_ITERATE(IGBProject::TViews, view_desc, vset) {
        if ((*view_desc)->GetView_id() == view->GetId()) {
            vset.erase(view_desc);
            break;
        }
    }*/
}


// attach a view to the document manager
void CProjectService::OnViewAttached(IProjectView* view)
{
    if( view ){
        /*
        const CGuiRegistry& reg = CGuiRegistry::GetInstance();
        bool save_views = reg.GetBool("GBENCH.Application.SaveViews", false);
        if (save_views) {
            RecordViewOpen(view);
        }
        */

        CProjectViewEvent ev( *view, CViewEvent::eViewAttached );
        GetWorkspace()->FindProjectById( view->GetProjectId() )->GetDocument()->Send( &ev );

        x_NotifyObservers();
    }
}


void CProjectService::OnViewDetached(IProjectView* view)
{
    if (view) {
     /*   const CGuiRegistry& reg = CGuiRegistry::GetInstance();
        bool save_views = reg.GetBool("GBENCH.Application.SaveViews", false);
        if (save_views) {
            RecordViewClose(view);
        }*/
        x_NotifyObservers();
    }
}


void CProjectService::PrefetchSeqDescr(const TConstScopedObjects& objs)
{
    // prefetch sequence description for performance consideration
    CAppTaskService* task_srv = m_ServiceLocator->GetServiceByType<CAppTaskService>();
    CRef<CSeqDescrRetrievalJob> job(new CSeqDescrRetrievalJob(objs));
    CRef<CAppJobTask> task(new CAppJobTask(*job, true,
        "Retrieving sequence descriptions", 5, "ObjManagerEngine", 2 /*view display delay*/));
    task_srv->AddTask(*task);
}


void CProjectService::x_OpenDefaultView(CProjectItem& item, CScope& scope)
{
    /*
    // TODO throws exceptions - need to be fixed
    try {
        /// optionally launch a default view
        const CGuiRegistry& reg = CGuiRegistry::GetInstance();
        bool launch_default = reg.GetBool("GBENCH.Application.DefaultView", false);
        if (launch_default) {
            CPluginHandle h = CPluginRegistry::GetPlugin("CViewPlugin_ViewDefault");
            if (h) {
                CRef<CPluginMessage> msg =
                    h.CreateMessage(CPluginCommand::eCommand_new_view);
                TConstScopedObjects objs;
                objs.push_back(SConstScopedObject(item, scope));
                CPluginUtils::CallPlugin(*msg, objs);
            }
        }
    } catch(CException& e)  {
        ERR_POST("Exception in CProjectService::x_OpenDefaultView() " << e.GetMsg());
    }*/
}

void CProjectService::GetObjProjects(TConstScopedObjects& objects,
                                     vector<objects::CGBProjectHandle::TId>& ids)
{
    if (objects.empty())
        return;

    CRef<CWorkspaceConstPrx> ws = GetWorkspaceConst();
    set<objects::CGBProjectHandle::TId> idSet;

    NON_CONST_ITERATE(TConstScopedObjects, it, objects) {
        CScope* scope = it->scope;
        CRef<CProjectConstPrx> project = ws->FindProjectByScope_C(scope);
        if (project) {
            idSet.insert(project->GetId());
        }
    }

    copy(idSet.begin(), idSet.end(), back_inserter(ids));
}

END_NCBI_SCOPE
