/*  $Id: project_service_test.cpp 21798 2010-07-21 21:52:56Z kuznets $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Andrey Yazhuk
 *
 */

#include <ncbi_pch.hpp>
#include <corelib/ncbifile.hpp>

#include <gui/core/project_service_test.hpp>

#include <gui/core/project_service.hpp>
#include <gui/core/project_folder.hpp>
#include <gui/core/workspace.hpp>

#include <objects/seqloc/Seq_id.hpp>

#include <corelib/ncbi_system.hpp>


BEGIN_NCBI_SCOPE

CProjectServiceTestTask::CProjectServiceTestTask(IServiceLocator* srv_locator)
:   CAppTask("Project Service Concurrency Test", true),
    m_SrvLocator(srv_locator),
    m_Service(NULL)
{
    m_Service = m_SrvLocator->GetServiceByType<CProjectService>().GetPointer();

    _ASSERT(m_Service);
}


CProjectServiceTestTask::~CProjectServiceTestTask()
{
}


CProjectServiceTestTask::ETaskState CProjectServiceTestTask::x_Run()
{
    LOG_POST(Info << "CProjectServiceTestTask::x_Run() - BEGIN");

    CAppJobDispatcher& disp = CAppJobDispatcher::GetInstance();

    CProjectServiceTestJob* job;

    job = new CProjectServiceTestJob(m_Service, CProjectServiceTestJob::ePrintItems, 10);
    disp.StartJob(*job, "ThreadPool");

    job = new CProjectServiceTestJob(m_Service, CProjectServiceTestJob::eCreateItems, 78);
    disp.StartJob(*job, "ThreadPool");

    job = new CProjectServiceTestJob(m_Service, CProjectServiceTestJob::eDeleteItems, 301);
    disp.StartJob(*job, "ThreadPool");

    job = new CProjectServiceTestJob(m_Service, CProjectServiceTestJob::eModifyItems, 13);
    disp.StartJob(*job, "ThreadPool");

    LOG_POST(Info << "CProjectServiceTestTask::x_Run() - BEGIN(completed)");
    return eCompleted;
}


///////////////////////////////////////////////////////////////////////////////
/// CProjectServiceTestJob

CProjectServiceTestJob::CProjectServiceTestJob(CProjectService* service,
                                               EType Type, int sleep_ms)
:   CAppJob(),
    m_ProjectService(service),
    m_Type(Type),
    m_SleepMs(sleep_ms)
{
    string sType;

    switch(m_Type)  {
    case eCreateItems:
        sType = "eCreateItems";
        break;

    case eModifyItems:
        sType = "eModifyItems";
        break;

    case eDeleteItems:
        sType = "eDeleteItems";
        break;

    case ePrintItems:
        sType = "ePrintItems";
        break;
    }

    m_Descr = "Test Job - " + sType;
}


IAppJob::EJobState CProjectServiceTestJob::Run()
{
    static const char* kLogMsg = "CProjectServiceTestJob::Run() - exception in x_Run() ";
    static string kErrorMsg = "Unhandled exception in CProjectServiceTestJob::Run()";

    x_ResetState(); // reset state before execution

    string err_msg;
    try {
        while(true) {
            LOG_POST(Info << m_Descr << " - now working ");

            switch(m_Type)  {
        case eCreateItems:
            x_CreateItems();
            break;

        case eModifyItems:
            x_ModifyItems();
            break;

        case eDeleteItems:
            x_DeleteItems();
            break;

        case ePrintItems:
            x_PrintItems();
            break;

        default:
            _ASSERT(false); //unsupported
            }

            LOG_POST(Info << m_Descr << " - sleeping for " << m_SleepMs << " ms");
            SleepMilliSec(m_SleepMs);
        }
    } catch (CException& e) {
        err_msg = kLogMsg + GetDescr() + ". " + e.GetMsg();
        LOG_POST(Error << err_msg);
        LOG_POST(Error << e.ReportAll());
    }

    CFastMutexGuard lock(m_Mutex);

    if(err_msg.empty())  {
        return eCompleted;
    } else {
        string s = kErrorMsg + err_msg;
        m_Error.Reset(new CAppJobError(s));
        return eFailed;
    }
}

typedef CWorkspaceConstPrx::TProjectPrxVector TProjectPrxVector;

void CProjectServiceTestJob::x_CreateItems()
{
    static int item_counter = 0;

    LOG_POST(Info << m_Descr << " get Write lock for Workspace and Projects");
    CRef<CWorkspacePrx> ws = m_ProjectService->GetWorkspace();

    TProjectPrxVector projects;
    ws->GetProjects(projects, CWorkspacePrx::eLoaded);

    if(projects.empty())    {
        LOG_POST(Info << "Creating a new project ...");

        CRef<CProjectPrx> prj = ws->CreateNewProject();
        ws->AddProject(*prj);

        ws->GetProjects(projects, CWorkspacePrx::eLoaded);

        LOG_POST(Info << "Creating a new project ... Done");
    } else {
        for(  size_t i = 0;  i < projects.size();  i++) {
            CProjectPrx& prj = *projects[i];

            CRef<CProjectFolderPrx> root = prj.GetDataFolder();

            static int kItemsN = 2;
            LOG_POST(Info << "Creating " << kItemsN << " items in project  " << prj.GetName());

            for( int j = 0;  j < kItemsN;  j++ )  {
                CRef<objects::CSeq_id> id(new objects::CSeq_id);
                id->Set("19568015");
                string label = "Item # " + NStr::IntToString(++item_counter);
                root->AddProjectItem(label, *id);
            }

            LOG_POST(Info << "Items created");
        }
    }
    LOG_POST(Info << m_Descr << "Release locks");
}


void CProjectServiceTestJob::x_ModifyItems()
{
    static int pass = 1;

    LOG_POST(Info << m_Descr << " get READ lock for Workspace and Write for Projects");

    CRef<CWorkspaceConstPrx> ws = m_ProjectService->GetWorkspaceConst();

    TProjectPrxVector projects;
    ws->GetProjects(projects, CWorkspaceConstPrx::eLoaded);

    for(  size_t i = 0;  i < projects.size();  i++) {
        CProjectPrx& prj = *projects[i];

        LOG_POST(Info << "Renaming items " << prj.GetName());
        CRef<CProjectFolderPrx> root = prj.GetDataFolder();

        vector< CRef<CProjectItemPrx> > items;
        root->GetProjectItems(items);

        string ver = ": v " + NStr::IntToString(pass);

        for(  size_t j = 0;  j < items.size(); j++  )   {
            CProjectItemPrx& item = *items[j];

            string name = item.GetName();
            string::size_type pos = name.find_first_of(':');
            if(pos != string::npos) {
                name.resize(pos);
            }
            name += ver;

            item.SetName(name);
        }
    }

    pass++;

    LOG_POST(Info << m_Descr << "Release locks");
}


void CProjectServiceTestJob::x_DeleteItems()
{
    LOG_POST(Info << m_Descr << " get READ lock for Workspace and Write for Projects");

    CRef<CWorkspaceConstPrx> ws = m_ProjectService->GetWorkspaceConst();

    TProjectPrxVector projects;
    ws->GetProjects(projects, CWorkspaceConstPrx::eLoaded);

    for(  size_t i = 0;  i < projects.size();  i++) {
        CProjectPrx& prj = *projects[i];

        LOG_POST(Info << "Deleting items from project " << prj.GetName());
        CRef<CProjectFolderPrx> root = prj.GetDataFolder();

        vector< CRef<CProjectItemPrx> > items;
        root->GetProjectItems(items);

        for(  size_t j = 0;  j < items.size(); j++  )   {
            CProjectItemPrx& item = *items[j];
            CProjectItemPrx::TId id = item.GetId();

            LOG_POST(Info << "Deleting item  " << item.GetName());

            root->RemoveChildItem(id);
        }
    }

    LOG_POST(Info << m_Descr << "Release locks");
}


void CProjectServiceTestJob::x_PrintItems()
{
    LOG_POST(Info << m_Descr << " get Read Lock for Workspace and Projects");

    CRef<CWorkspaceConstPrx> ws = m_ProjectService->GetWorkspaceConst();

    CWorkspaceConstPrx::TProjectConstPrxVector projects;
    ws->GetProjects_C(projects);

    for(  size_t i = 0;  i < projects.size();  i++ ) {
        CProjectConstPrx& prj = *projects[i];

        LOG_POST(Info << m_Descr << " printing items in project " << prj.GetName());

        CRef<CProjectFolderConstPrx> root = prj.GetDataFolder_C();

        vector< CRef<CProjectItemConstPrx> > items;
        root->GetProjectItems_C(items);

        for(  size_t j = 0;  j < items.size(); j++  )   {
            CProjectItemConstPrx& item = *items[j];
            LOG_POST(Info << "\t- " << item.GetName());
        }
    }
    LOG_POST(Info << m_Descr << "Release locks");
}


END_NCBI_SCOPE

