/*  $Id: project_task.cpp 25614 2012-04-13 14:04:11Z kuznets $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Andrey Yazhuk
 *
 */

#include <ncbi_pch.hpp>

#include <gui/core/project_task.hpp>

#include <gui/core/project_loading_task.hpp>
#include <gui/core/project.hpp>

#include <gui/core/project_service.hpp>
#include <gui/core/document.hpp>
#include <gui/core/loading_app_job.hpp>
#include <gui/core/save_dlg.hpp>

#include <gui/framework/workbench.hpp>
#include <gui/framework/app_task_service.hpp>
#include <gui/framework/app_job_task.hpp>

#include <gui/objects/GBWorkspace.hpp>
#include <gui/objects/ProjectDescr.hpp>
#include <gui/objects/WorkspaceFolder.hpp>

#include <gui/widgets/wx/message_box.hpp>
#include <gui/widgets/wx/wx_utils.hpp>

#include <wx/filedlg.h>
#include <wx/msgdlg.h>
#include <wx/filename.h>

#include <objmgr/impl/tse_info_object.hpp>


BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

CProjectTask::CProjectTask(IServiceLocator* srv_locator)
:   CAppTask(kEmptyStr, false),
    m_SrvLocator(srv_locator),
    m_Service(NULL),
    m_Command(eInvalid),
    m_SelectWorkspace(false)
{
    m_Service = m_SrvLocator->GetServiceByType<CProjectService>().GetPointer();

    _ASSERT(m_Service);
}


CProjectTask::~CProjectTask()
{
}


void CProjectTask::Init_NewWorkspace()
{
    m_Command = eNewWorkspace;
}


void CProjectTask::Init_CloseWorkspace()
{
    m_Command = eCloseWorkspace;
}


void CProjectTask::Init_OpenWorkspace(const wxString& filename)
{
    _ASSERT( ! filename.empty());

    m_Command = eOpenWorkspace;
    m_WorkspaceFilename = filename;
}


void CProjectTask::Init_OpenProjects(vector<wxString>& filenames)
{
    _ASSERT( ! filenames.empty());

    m_Command = eOpenProjects;
    m_ProjectFilenames = filenames;
}


void CProjectTask::Init_Save(bool workspace, TProjectIdVector sel_projects)
{
    m_Command = eSave;
    m_SelectWorkspace = workspace;
    m_ProjectIds = sel_projects;  // selected projects
}


void CProjectTask::Init_SaveAs(bool workspace, TProjectIdVector sel_projects)
{
    m_Command = eSaveAs;
    m_SelectWorkspace = workspace;
    m_ProjectIds = sel_projects;  // selected projects
}


void CProjectTask::Init_LoadProjects(TProjectIdVector project_ids)
{
    _ASSERT(project_ids.size());

    m_Command = eLoadProjects;
    m_ProjectIds = project_ids;
}


void CProjectTask::Init_UnLoadProjects(TProjectIdVector project_ids)
{
    _ASSERT(project_ids.size());

    m_Command = eUnLoadProjects;
    m_ProjectIds = project_ids;
}


void CProjectTask::Init_RemoveProjects(TProjectIdVector project_ids)
{
    _ASSERT(project_ids.size());

    m_Command = eRemoveProjects;
    m_ProjectIds = project_ids;
}


bool CProjectTask::Enable_Save(IServiceLocator* srv_locator)
{
    CIRef<CProjectService> srv = srv_locator->GetServiceByType<CProjectService>();
    // we need at least have a workspace (it does not have to be dirty)
    bool en = srv  &&  srv->HasWorkspace();
    return en;
}


bool CProjectTask::Enable_SaveAs(IServiceLocator* srv_locator)
{
    CIRef<CProjectService> srv = srv_locator->GetServiceByType<CProjectService>();
    // we need at least have a workspace (it does not have to be dirty)
    bool en = srv  &&  srv->HasWorkspace();
    return en;
}


bool CProjectTask::Enable_LoadUnloadProjects(IServiceLocator* srv_locator,
                                             TProjectIdVector projects, bool load)
{
    CIRef<CProjectService> srv = srv_locator->GetServiceByType<CProjectService>();
    CRef<CWorkspaceConstPrx> workspace = srv->GetWorkspaceConst();

    NON_CONST_ITERATE(TProjectIdVector, it, projects)  {
        TProjectId id = *it;
        CRef<CProjectConstPrx> project = workspace->FindProjectById_C(id);
        if (load != (project->IsLoaded() || project->IsLoading())) {
            return true;
        }
    }
    return false;
}


CProjectTask::ETaskState CProjectTask::x_Run()
{
    ETaskState finite_state = eCompleted;

    //LOG_POST("CProjectTask::x_Run() - BEGIN");
    switch(m_Command)  {
    case eNewWorkspace:
        x_NewWorkspace();
        break;

    case eCloseWorkspace:
        if( !x_CloseWorkspace() ){
            finite_state = eCanceled;
        }
        break;

    case eOpenWorkspace:
        x_OpenWorkspace();
        break;

    case eOpenProjects:
        x_OpenProjects();
        break;

    case eSave:
    case eSaveAs:
        x_Save();
        break;

    case eLoadProjects:
        x_LoadProjects(m_ProjectIds);
        break;

    case eUnLoadProjects:
        x_UnLoadProjects(m_ProjectIds);
        break;

    case eRemoveProjects:
        x_RemoveProjects(m_ProjectIds);
        break;

    default:
        LOG_POST(Info << "CProjectTask::x_Run() - END (failed) - unsupported command");
        _ASSERT(false); // unknown command
        return eFailed;
    }
    //LOG_POST("CProjectTask::x_Run() - BEGIN(completed)");
    return finite_state;
}


#define REMOVE_PRJ_TEXT     "Remove selected projects from the workspace?"

#define CLOSE_UNSAVED_PRJ_TEXT \
    "Workspace contains new or modified projects. If not saved \
    the new projects will be removed from the workspace and \
    the modified projects will loose all changes. Would you like \
    to save the selected projects?"\

#define WS_WILL_NOT_BE_SAVED \
"Workspace will not be saved, because some of the projects are not saved."

#define CANNOT_LOAD_WS_TEXT "Cannot load workspace!"

// TODO ! all toplevel functions must handle exceptions and report errors to the app
// TODO return values from functions ?

/// any state -> new workspace
void CProjectTask::x_NewWorkspace()
{
    //LOG_POST("CProjectTask::x_NewWorkspace()");
    CProjectService::TGuard lock(*m_Service);

    if(x_CloseWorkspace())    {
        m_Service->CreateNewWorkspace();
    }
}


// UI operation - shows a dialog and load a workspace
void CProjectTask::x_OpenWorkspace()
{
    //LOG_POST("CProjectTask::x_OpenWorkspace()");
    CProjectService::TGuard lock(*m_Service);

    if(x_CloseWorkspace())    {
        if( ! m_WorkspaceFilename.empty())   {
            x_DoOpenWorkspace(m_WorkspaceFilename);

            m_Service->AddToProjectWorkspaceMRUList(m_WorkspaceFilename);
        }
    }
}


void CProjectTask::x_OpenProjects()
{
    CProjectService::TGuard lock(*m_Service);

    if( ! m_Service->HasWorkspace())    {
        m_Service->CreateNewWorkspace();
    }

    CRef<CWorkspacePrx> workspace = m_Service->GetWorkspace();

    TProjectIdVector ids;
    for(  size_t i = 0;  i < m_ProjectFilenames.size();  i++ ) {
        wxString filename = m_ProjectFilenames[i];
        wxString rel_path = workspace->ToRelativePath(filename);

        // look is we already have a project with this file name
        CRef<CProjectPrx> project = workspace->FindProjectByFilename(rel_path);
        if (project.IsNull()) {
            ERR_POST(Info << "Project: " << filename);
            project = workspace->CreateNewProject(filename);
            workspace->AddProject(*project);
            ids.push_back(project->GetId());
        }
        else {
            wxString msg = wxString::Format(
                wxT("Project \"%s\" is already opened in the workspace."), filename.c_str());
            wxMessageBox(msg, wxT("Open Project"), wxOK | wxICON_EXCLAMATION);
        }
    }

    if (!ids.empty())
        x_LaunchProjectLoadingTask(ids, "Open Projects", true);
}


/// UI operation - show dialogs if needed, save projects and workspaces
/// support Save and Save As operations
void CProjectTask::x_Save()
{
    CProjectService::TGuard lock(*m_Service);

    vector<TProjectId> ids;
    CRef<CWorkspacePrx> workspace = m_Service->GetWorkspace();
    workspace->GetProjectIds(ids, CWorkspacePrx::eLoaded);

    vector<int> selected;
    for( size_t i = 0;  i < ids.size();    i++ )   {
        TProjectId id = ids[i];

        for(  size_t j = 0;  j < m_ProjectIds.size();  j++ )  {
            TProjectId curr_id = m_ProjectIds[j];
            if(curr_id == id)  {
                selected.push_back((CTSE_Info_Object::TChunkId)i);
            }
        }
    }

    bool save_as = m_Command == eSaveAs;
    TProjectFunc func = save_as ? &CProjectTask::x_DoSaveProjectAs
                                : &CProjectTask::x_DoEnsureProjectSaved;
    string title = save_as ? "Save As" : "Save";

    // ask user about unsaved projects and save them
    x_DoSaveUserSelected(title, "", save_as, true, m_SelectWorkspace, false,
                         ids, selected, func);
}


bool CProjectTask::x_CloseWorkspace()
{
    //LOG_POST("CProjectTask::x_CloseWorkspace()");
    CProjectService::TGuard lock(*m_Service);

    if( ! m_Service->HasWorkspace() ) {
        return true; // nothing to close - we are done
    }
    try {
        TProjectIdVector loaded_prjs, unsaved_prjs;

        CRef<CWorkspacePrx> workspace = m_Service->GetWorkspace();
        workspace->GetProjectIds(loaded_prjs, CWorkspacePrx::eLoaded);

        x_GetUnsavedProjects(*workspace, loaded_prjs, unsaved_prjs);

        // ask user about unsaved projects and save them
        bool ok = true;
        if( ! unsaved_prjs.empty()  ||  workspace->IsDirty()) {
            // select all
            vector<int> selected;
            for(  size_t i = 0;  i < unsaved_prjs.size();  i++) {
                selected.push_back((CTSE_Info_Object::TChunkId)i);
            }
            ok = x_DoSaveUserSelected("Close Workspace", CLOSE_UNSAVED_PRJ_TEXT,
                                      false, true, true, true, unsaved_prjs, selected,
                                      &CProjectTask::x_DoEnsureProjectSaved);
            if(ok)  {
                // the project that has not been saved should be removed (with
                // user's approval)
                NON_CONST_ITERATE (TProjectIdVector, it, loaded_prjs) {
                    TProjectId id = *it;
                    CRef<CProjectPrx> project = workspace->FindProjectById(id);
                    if(project->GetFilename().empty())   {
                        project.Reset();
                        // remove project with (no async history clean)
                        workspace->RemoveProject(id, false);
                    }
                }
            }
        }
        if(ok)  {
            // here we are confident that all projects in the workspace are
            // saved
            x_DoCloseWorkspace();
        }
        return ok;
    }
    catch (CException& e) {
        _ASSERT(false); // how did we het here ?
        LOG_POST(Error << "Error closing Genome Workbench workspace: " << e.GetMsg());
    }
    return true;
}


#define REMOVE_PRJ_DLG_DESCR "These projects are new or modified, \
all changes will be lost if these projects are not saved. Would you \
like to save the selected projects?"

// removes selected projects from the workspace //TODO duplication
void CProjectTask::x_RemoveProjects(TProjectIdVector projects)
{
    CRef<CWorkspacePrx> workspace = m_Service->GetWorkspace();

    // collect selected projects that need to be saved
    TProjectIdVector    unsaved_prjs;
    x_GetUnsavedProjects(*workspace, projects, unsaved_prjs);

    bool ok = true;
    if( ! unsaved_prjs.empty()) {
        // select all
        vector<int> selected;
        for(  size_t i = 0;  i < unsaved_prjs.size();  i++) {
            selected.push_back((CTSE_Info_Object::TChunkId)i);
        }
        // ask user about unsaved projects and saved selected
        ok = x_DoSaveUserSelected("Remove - Save changed projects", REMOVE_PRJ_DLG_DESCR,
                                  false, false, false, true, unsaved_prjs, selected,
                                  &CProjectTask::x_DoEnsureProjectSaved);
    } else {
        // simply warn that projects will be removed
        int res = wxMessageBox(wxT(REMOVE_PRJ_TEXT), wxT("Remove Project"),
                               wxYES_NO | wxICON_QUESTION);
        ok = (res == wxYES);
    }

    // now remove all projects
    if(ok)  {
        NON_CONST_ITERATE(TProjectIdVector, it, projects) {
            workspace->RemoveProject(*it);
        }
    }
}


void CProjectTask::x_LoadProjects(TProjectIdVector project_ids)
{
    TProjectIdVector ids_toload;

    CRef<CWorkspacePrx> workspace = m_Service->GetWorkspace();

    for(  size_t i = 0;  i < project_ids.size();  i++  ) {
        TProjectId id = project_ids[i];
        CRef<CProjectPrx> project = workspace->FindProjectById(id);

        if (project  &&  ! project->IsLoaded())  { // not loaded - load
            ids_toload.push_back(id);
        }
    }

    x_LaunchProjectLoadingTask(ids_toload, "Load Project", false);
}


void CProjectTask::x_LaunchProjectLoadingTask(const TProjectIdVector& ids,
                                              const string& task_name,
                                              bool add_to_mru)
{
    CIRef<CAppTaskService> task_srv = m_SrvLocator->GetServiceByType<CAppTaskService>();

    _ASSERT(task_srv);

    ITERATE(TProjectIdVector, it, ids) {
        CRef<CProjectLoadingTask>
            task(new CProjectLoadingTask(*it, m_Service, add_to_mru));
        task_srv->AddTask(*task);
    }
}


#define UNLOAD_PRJ_DLG_DESCR "Workspace contains new or modified projects. \
If not saved the new projects will be removed from the workspace and the modified \
projects will loose all changes. Would you like to save the selected projects?"


void CProjectTask::x_UnLoadProjects(TProjectIdVector projects) //TODO duplication
{
    CRef<CWorkspacePrx> workspace = m_Service->GetWorkspace();

    TProjectIdVector unsaved_prjs;
    x_GetUnsavedProjects(*workspace, projects, unsaved_prjs);

    bool ok = true;
    if( ! unsaved_prjs.empty()) {
        // select all
        vector<int> selected;
        for(  size_t i = 0;  i < unsaved_prjs.size();  i++) {
            selected.push_back((CTSE_Info_Object::TChunkId)i);
        }
        // ask user about unsaved projects and saved selected
        ok = x_DoSaveUserSelected("Unload Project", UNLOAD_PRJ_DLG_DESCR,
                                  false, false, false, true, unsaved_prjs, selected,
                                  &CProjectTask::x_DoEnsureProjectSaved);
    }
    // now unload (remove) the projects
    if(ok)  {
        for (TProjectIdVector::iterator it = projects.begin();
                                        it != projects.end(); ++it )   {
            TProjectId id = *it;
            CRef<CProjectPrx> project = workspace->FindProjectById(id);
            if (project->IsLoading()) {
                project->CancelLoading();
            }
            else  if(project->IsLoaded()) {
                if(project->GetFilename().empty())   {
                    // the user refused to save - remove it completely
                    workspace->RemoveProject(id);
                } else {
                    // unload saved project
                    workspace->UnloadProject(*project);
                }
            }
        }
    }
}


///////////////////////////////////////////////////////////////////////////////
/// - DO functions - helper functions called from top-level protected functions

void CProjectTask::x_DoOpenWorkspace(const wxString& filename)
{
    //LOG_POST("CProjectTask::x_DoOpenWorkspace()  " << filename);

    if(m_Service->LoadWorkspace(filename))    {
        // iterate through projects and try to load them
        CRef<CWorkspacePrx> workspace = m_Service->GetWorkspace();

        // get a list of project ids to load
        TProjectIdVector ids;
        workspace->GetProjectIds(ids, CWorkspacePrx::eUnloaded);

        x_LaunchProjectLoadingTask(ids, "Open Workspace", false);

    } else {
        TDialogType type = eDialog_Ok | eDialog_Modal;
        NcbiMessageBox(CANNOT_LOAD_WS_TEXT, type, eIcon_Exclamation, "Error", eWrap);
    }
}


bool CProjectTask::x_DoCloseWorkspace()
{
    _ASSERT(m_Service->HasWorkspace());

    m_Service->ReleaseWorkspace();
    return true;
}


// saves only the workspace, assuming that the projects has been saved properly
bool CProjectTask::x_DoSaveWorkspaceOnly(bool save_as)
{
    CRef<CWorkspacePrx> workspace = m_Service->GetWorkspace();

    wxString filename = workspace->GetFilename();
    if(filename.empty() || save_as)    {
        filename = x_DoChooseNewWorkspaceName();
        if(! filename.empty())    {
            workspace->SetFilename(filename);
        }
    }

    if( ! filename.empty())    {
        // now we shall have a filename - save data
        bool ok = m_Service->SaveWorkspace();
        if(ok)  {
            m_Service->AddToProjectWorkspaceMRUList(filename);
            ERR_POST(Info << "Workspace saved: " << filename);
        }
        return ok;
    }
    return false;
}


bool CProjectTask::x_ProjectNotSaved(CWorkspaceConstPrx& workspace,
                                     CProjectPrx& project)
{
    if (project.IsDirty())
        return true;

    wxString path = workspace.ToAbsolutePath(project.GetFilename());
    return !path.empty() && !wxFileName::FileExists(path);
}


void CProjectTask::x_GetUnsavedProjects(CWorkspaceConstPrx& workspace,
                                        const TProjectIdVector& projects,
                                        TProjectIdVector& not_saved)
{
    ITERATE(TProjectIdVector, it, projects) {
        TProjectId id = *it;
        CRef<CProjectPrx> project = workspace.FindProjectById(id);

        if(x_ProjectNotSaved(workspace, *project))  {
            not_saved.push_back(id);
        }
    }
}


// shows "Save" dialog and save workspace and projects selected by the user
bool CProjectTask::x_DoSaveUserSelected(const string& title,
                                        const string& description,
                                        bool save_as,
                                        bool en_workspace,
                                        bool sel_workspace,
                                        bool no_btn,
                                        TProjectIdVector& project_ids,
                                        const vector<int>& selected,
                                        TProjectFunc func)
{
    CRef<CWorkspaceConstPrx> workspace = m_Service->GetWorkspaceConst();

    // prepare parameters for the dialog
    SSaveProjectsDlgParams params;
    params.m_Description = description;
    params.m_ShowNoButton = no_btn;

    // prepare Workspace data
    params.m_EnWorkspace = en_workspace;
    if(en_workspace)   {
        if(workspace->IsDirty())   {
            params.m_WorkspaceModified = true;
            params.m_WorkspaceNew = workspace->GetFilename().empty();
        }
        params.m_SaveWorkspace = sel_workspace;
    }

    // prepare projects data
    params.m_ProjectIds = project_ids;
    for(  size_t i = 0;  i < selected.size();  i++ )    {
        params.m_ProjectsToSave.Add(selected[i]);
    }

    // create and initialize the dialog
    CSaveDlg dlg(NULL);
    dlg.SetRegistryPath("ApplicationGUI.SaveDlg"); //TODO
    dlg.SetProjectService(m_Service);
    dlg.SetParams(params);
    dlg.SetTitle(ToWxString(title));

    int res = dlg.ShowModal();

    switch(res) {
    case wxID_SAVE:  {
        // save selected
        params = dlg.GetParams();
        bool ok = true;
        for(  size_t i = 0;  ok  && i < params.m_ProjectsToSave.size();  i++)    {
            int index = params.m_ProjectsToSave[i];
            TProjectId id = params.m_ProjectIds[index];
            CRef<CProjectPrx> project = workspace->FindProjectById(id);

            ok = (this->*func)(*project);
        }
        if(params.m_SaveWorkspace)  {
            if(ok)  {
                ok = x_DoSaveWorkspaceOnly(save_as);
            } else {
                wxMessageBox(ToWxString(WS_WILL_NOT_BE_SAVED),
                             ToWxString(title), wxOK | wxICON_EXCLAMATION);
            }
        }
        return ok;
    }
    case wxID_NO:
        return true;
    case wxID_CANCEL:
    default:
        return false;
    }
    return false;
}


wxString AskProjectSaveFilename(CProjectPrx& project, bool as = false)
{
    // generate the default filename
    wxString title = ToWxString(project.GetName());
    if (title.length() > 512) {
        title.erase(509);
        title += wxT("...");
    }
    wxString fname = title + wxT(".gbp");

    SWFileDlgData& file_data = CProjectLoadingTask::sm_ProjectDlgData;
    file_data.m_Style = wxFD_SAVE | wxFD_OVERWRITE_PROMPT;
    file_data.SetFilename(fname);
    file_data.m_Title = wxT("Save Project \"") + title + (as ? wxT("\" As") : wxT("\""));

    if(NcbiFileBrowser(CProjectLoadingTask::sm_ProjectDlgData, NULL) == wxID_OK) {
        return CProjectLoadingTask::sm_ProjectDlgData.GetFilename();
    }
    return wxEmptyString;
}



// checks if project was modified and performs saving, returns true if successful
// if filename has not been assigned to the project - asks for a filename
bool CProjectTask::x_DoEnsureProjectSaved(CProjectPrx& project)
{
    _ASSERT(project.IsLoaded());

    bool fSaved     = false;
    bool new_project = false;

    wxString prj_filename = project.GetFilename();

    /// check if the filename has not been assigned yet
    bool ask_for_file_name = (prj_filename == wxEmptyString);

    /// check if the file is a remote file
    if (prj_filename.find(wxT("http://")) == 0  || prj_filename.find(wxT("ftp://")) == 0)  {
        wxString msg = wxT("The project\n") + prj_filename;
        msg += wxT("\nmust have a local file name in order to save it.");

        NcbiMessageBoxW(msg);
        ask_for_file_name = true;
    }

    CRef<CWorkspacePrx> workspace = m_Service->GetWorkspace();

    if (ask_for_file_name) {
        /// request a file name
        new_project = true;
        wxString abs_path = AskProjectSaveFilename(project);
        if(abs_path.empty())   {
            return false; // user canceled operation
        }
        wxString rel_path = workspace->ToRelativePath(abs_path);
        project.SetFilename(rel_path);
    }

    try {
        project.Save();
        fSaved = true;
        const wxString filename = project.GetFilename();
        m_Service->AddToProjectWorkspaceMRUList(project.GetFilename());
        ERR_POST(Info << "Project saved:" << filename);
    } catch(CException& e)  {
        NcbiErrorBox(e.GetMsg(), "Save Project - Error");
        fSaved = false;
    }
    return fSaved;
}


bool CProjectTask::x_DoSaveProjectAs(CProjectPrx& project)
{
    wxString filename = AskProjectSaveFilename(project, true);
    bool fSaved = false;

    if( ! filename.empty())  {
        project.SetFilename(filename);

        project.Save();
        fSaved = true;
        m_Service->AddToProjectWorkspaceMRUList(project.GetFilename());
    }
    return fSaved;
}


// show the dialog for choosing Workspace file name
wxString CProjectTask::x_DoChooseNewWorkspaceName()
{
    _ASSERT(m_Service->HasWorkspace()); // contract for callers

    CRef<CWorkspacePrx> workspace = m_Service->GetWorkspace();

    // get filename
    wxString filename = workspace->GetFilename();
    if(filename.empty())    {
        string title = workspace->GetName();
        if ( ! title.empty()) {
            filename = ToWxString(title + string(".gbw"));
        }
    }

    wxFileDialog dlg(NULL, wxT("Save Workspace As"),
                     wxEmptyString, filename,
                     wxT("GBENCH Workspace (*.gbw)|*.gbw"),
                     wxFD_SAVE|wxFD_OVERWRITE_PROMPT);

    if (wxID_OK == dlg.ShowModal())
        return dlg.GetPath();

    return wxEmptyString;
}


END_NCBI_SCOPE
