/*  $Id: project_tree_view.cpp 25445 2012-03-19 15:23:35Z falkrb $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Andrey Yazhuk
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>

#include <gui/core/project_tree_view.hpp>

#include <gui/core/project_service.hpp>
#include <gui/core/project.hpp>
#include <gui/core/workspace_folder.hpp>
#include <gui/core/workspace.hpp>

#include <gui/core/project_tree_panel.hpp>
#include <gui/core/app_explorer_service.hpp>
#include <gui/core/selection_service_impl.hpp>
#include <gui/core/data_mining_service.hpp>

#include <gui/framework/view_manager_service.hpp> //TODO
#include <gui/framework/workbench.hpp>

#include <gui/objutils/registry.hpp>

#include <gui/utils/extension_impl.hpp>


BEGIN_NCBI_SCOPE
USING_SCOPE(objects);


static CExtensionPointDeclaration
    point1(EXT_POINT__PROJECT_TREE_VIEW__CONTEXT_MENU__ITEM_CMD_CONTRIBUTOR,
           "Project Tree View - commands applicable to Explorer Items");

/// register the factory in the Extention Point
static CExtensionDeclaration
    decl1("view_manager_service::view_factory", new CProjectTreeViewFactory());


///////////////////////////////////////////////////////////////////////////////
/// CProjectTreeView

CViewTypeDescriptor CProjectTreeView::m_TypeDescr(
    "Project Tree View", // type name
    "project_tree_view", // icon alias TODO
    "Project Tree View hint - TODO",
    "Project Tree View description - TODO",
    "PROJECT_TREE_VIEW", // help Id
    "System",     // category
    true);       // singleton

BEGIN_EVENT_MAP(CProjectTreeView, CEventHandler)
    ON_MESSAGE(CEvent, CViewEvent::eWidgetSelectionChanged,
                       &CProjectTreeView::OnWidgetSelChanged)
END_EVENT_MAP()


CProjectTreeView::CProjectTreeView()
:   m_Workbench(NULL),
    m_ExplorerService(NULL),
    m_SelectionService(NULL),
    m_Panel(NULL)
{
}


CProjectTreeView::~CProjectTreeView()
{
}


const CViewTypeDescriptor& CProjectTreeView::GetTypeDescriptor() const
{
    return m_TypeDescr;
}


void CProjectTreeView::SetWorkbench(IWorkbench* workbench)
{
    if(workbench)   {
        // connect to services
        CIRef<CAppExplorerService> srv = workbench->GetServiceByType<CAppExplorerService>();
        IObservable* observable = dynamic_cast<IObservable*>(srv.GetPointer());
        observable->AttachObserver(this);

        CIRef<CSelectionService> sel_srv = workbench->GetServiceByType<CSelectionService>();
        if(sel_srv) {
            sel_srv->AttachClient(this);
        }

        // connect
        CIRef<CDataMiningService> dm_srv = workbench->GetServiceByType<CDataMiningService>();
        if (dm_srv) {
            dm_srv->AttachContext(*this);
        }

        m_Panel->SetWorkbench(workbench);

        // to receive selection change event
        m_Panel->AddListener(this, ePool_Parent);
    } else if(m_Workbench)  {
        m_Panel->RemoveListener(this);

        m_Panel->SetWorkbench(NULL);

        // disconnect from services
        CIRef<CSelectionService> sel_srv = m_Workbench->GetServiceByType<CSelectionService>();
        if(sel_srv) {
            sel_srv->DetachClient(this);
        }

        // disconnect
        CIRef<CDataMiningService> dm_srv = m_Workbench->GetServiceByType<CDataMiningService>();
        if (dm_srv) {
            dm_srv->DetachContext(*this);
        }

        IObservable* observable = dynamic_cast<IObservable*>(m_ExplorerService);
        if(observable)  {
            observable->DetachObserver(this);
        }
    }

    m_Workbench = workbench;
}


void CProjectTreeView::CreateViewWindow(wxWindow* parent)
{
    m_Panel = new CProjectTreePanel();
    m_Panel->Create(parent, wxID_ANY);
}


void CProjectTreeView::DestroyViewWindow()
{
    if(m_Panel) {
        m_Panel->Destroy();
        m_Panel = NULL;
    }
}


void CProjectTreeView::OnUpdate(IObserverUpdate& update)
{
    CObserverUpdate* up = dynamic_cast<CObserverUpdate*>(&update);
    _ASSERT(up);

    IObservable* observable = up->GetObservable();
    CAppExplorerService* srv = dynamic_cast<CAppExplorerService*>(observable);
    _ASSERT(srv);

    switch(up->GetType())    {
    case CObserverUpdate::eAttached:
        _ASSERT(! m_ExplorerService  &&  srv  &&  m_Panel);

        m_ExplorerService = srv;
        m_Panel->SetService(m_ExplorerService);
        break;

    case CObserverUpdate::eStateChanged:
        m_Panel->RefreshItems();
        break;

    case CObserverUpdate::ePreDetach:
        _ASSERT(m_ExplorerService);

        m_ExplorerService = NULL;
        m_Panel->SetService(m_ExplorerService);

        break;
    }
}


wxWindow* CProjectTreeView::GetWindow()
{
    return m_Panel;
}


string CProjectTreeView::GetClientLabel(IWMClient::ELabel) const
{
    return m_TypeDescr.GetLabel();
}


string CProjectTreeView::GetIconAlias() const
{
    return m_TypeDescr.GetIconAlias();
}


const CRgbaColor* CProjectTreeView::GetColor() const
{
    return NULL;
}


void CProjectTreeView::SetColor(const CRgbaColor& color)
{
    // do nothing
}


const wxMenu* CProjectTreeView::GetMenu()
{
    return NULL;
}


void CProjectTreeView::UpdateMenu(wxMenu& root_menu)
{
}


IWMClient::CFingerprint CProjectTreeView::GetFingerprint() const
{
    return CFingerprint(m_TypeDescr.GetLabel(), true);
}


void CProjectTreeView::SetRegistryPath(const string& path)
{
    m_RegPath = path; // store for later use
}


//static const char* kDisplayModeTag = "DisplayMode";

void CProjectTreeView::SaveSettings() const
{
    if( ! m_RegPath.empty())   {
        CGuiRegistry& gui_reg = CGuiRegistry::GetInstance();
        CGuiRegistry::TReadWriteView view = gui_reg.GetReadWriteView(m_RegPath);
    }
}


void CProjectTreeView::LoadSettings()
{
    if( ! m_RegPath.empty())   {
        CGuiRegistry& gui_reg = CGuiRegistry::GetInstance();
        CGuiRegistry::TReadView view = gui_reg.GetReadView(m_RegPath);
    }
}


/// ISelectionClient //////////////////////////////////////////////////////////

void CProjectTreeView::OnWidgetSelChanged(CEvent* evt)
{
    _ASSERT(evt);

    if (evt->GetID() == CViewEvent::eWidgetSelectionChanged  &&  m_SelectionService) {
        m_SelectionService->OnSelectionChanged(this);
    }
}

void CProjectTreeView::GetSelection(TConstScopedObjects& objs) const
{
    typedef CAppExplorerService::TProjectTreeItem TProjectTreeItem;
    typedef CAppExplorerService::TProjectItemTreeItem TProjectItemTreeItem;

    CProjectTreeView* nc_this = const_cast<CProjectTreeView*>(this);

    CProjectTreePanel::TItemRefVector sel_items;
    nc_this->m_Panel->GetSelectedItems(sel_items);

    CIRef<CProjectService> prj_srv = m_Workbench->GetServiceByType<CProjectService>();
    CRef<CWorkspaceConstPrx> workspace = prj_srv->GetWorkspaceConst();

    NON_CONST_ITERATE (CProjectTreePanel::TItemRefVector, it, sel_items) {
        CExplorerItem& item = **it;

        //CAppExplorerService::EItemType type = m_ExplorerService->GetItemType(item);
        SConstScopedObject obj;

        switch( m_ExplorerService->GetItemType(item) ){
        case CAppExplorerService::eProject: {
            /* TODO
            TProjectTreeItem* prj_item = dynamic_cast<TProjectTreeItem*>(&item);

            CProjectPrx::TId id = prj_item->GetData();
            CRef<CProjectPrx> project = workspace->FindProjectById(id);
            if(project  &&  project->IsLoaded()  &&  project->GetScope()) {
                obj.object.Reset(project);
                obj.scope.Reset(project->GetScope());
            }*/
            break;
        }

        case CAppExplorerService::eProjectItem: {
            TProjectItemTreeItem* prj_item = dynamic_cast<TProjectItemTreeItem*>(&item);
            CProjectPrx::TId prj_id = m_ExplorerService->GetProjectIdByChild(item);
            if( prj_id == (CProjectPrx::TId)-1 ){
                continue;
            } 

            CRef<CProjectConstPrx> project = workspace->FindProjectById_C(prj_id);

            CProjectItemPrx::TId item_id = prj_item->GetData();
            CRef<CProjectItemConstPrx> project_item =
                                        project->FindProjectItemById_C(item_id);

            //_ASSERT(project_item);

            if( project_item == NULL ){
                LOG_POST( Warning << "PTV: Selected item \"" << item.GetLabel() << "\" is zombie" );
                continue;
            }

            obj.object.Reset(project_item->GetObject());
            obj.scope.Reset(project->GetScope());
            break;
        }

        default:
            break;
        }

        if (obj.scope.GetPointer()  &&  obj.object.GetPointer()) {
            objs.push_back(obj);
        }
    }
}


void CProjectTreeView::SetSelectionService(ISelectionService* service)
{
    m_SelectionService = service;
}


string CProjectTreeView::GetSelClientLabel()
{
    return GetClientLabel(IWMClient::eDefault);
}


void CProjectTreeView::GetSelection(CSelectionEvent& evt) const
{
    // do nothing, CProjectTreeView does not support selection broadcast
}


void CProjectTreeView::SetSelection(CSelectionEvent& evt)
{
    // do nothing, CProjectTreeView does not support selection broadcast
}


void CProjectTreeView::GetMainObject(TConstScopedObjects& objs) const
{
    // no main object
}


CRef<CScope> CProjectTreeView::GetScope() const
{
    CRef<CScope> sc;
    return sc; // because elements may belong to different scopes
}
/// END of ISelectionClient ///////////////////////////////////////////////////

bool CProjectTreeView::IsWorkspaceSelected()
{
    if(m_Panel) {
        return m_Panel->IsWorkspaceSelected();
    }
    return false;
}


void CProjectTreeView::GetSelectedProjectIds(TProjectIdVector& ids)
{
    if(m_Panel) {
        m_Panel->GetSelectedProjectIds(ids);
    }
}


void CProjectTreeView::GetSelectedItems(TItemRefVector& items)
{
    if(m_Panel) {
        m_Panel->GetSelectedItems(items);
    }
}


string CProjectTreeView::GetDMContextName()
{
    return GetClientLabel(IWMClient::eDefault);
}

CRef<CSeq_loc> CProjectTreeView::GetSearchLoc()
{
    CRef<CSeq_loc> loc;
    CRef<CSeq_id>  id(new CSeq_id());

    TConstScopedObjects objs;
    GetSelection(objs);
    if (objs.size()>0) {
        const CSeq_id * sid  = dynamic_cast<const CSeq_id*>(objs.begin()->object.GetPointer());
        if (sid) {
            id->Assign(*sid);
            loc.Reset(new CSeq_loc());
            loc->SetWhole(*id);
        }
    }

    return loc;
}

void CProjectTreeView::GetMultiple(TConstScopedObjects & pObjects)
{   
    TConstScopedObjects objs;
    GetSelection(objs);
    
    ITERATE (TConstScopedObjects, obj, objs) {
        CRef<CSeq_loc> loc;
        CRef<CSeq_id>  id(new CSeq_id());
       

        const CSeq_id * sid  = dynamic_cast<const CSeq_id*>(obj->object.GetPointer());

        if (sid) {
            CRef<CScope> scope = obj->scope;
            id->Assign(*sid);
            loc.Reset(new CSeq_loc());
            loc->SetWhole(*id);
            pObjects.push_back(SConstScopedObject(*loc, *scope));
        }
    }
}


CRef<CScope> CProjectTreeView::GetSearchScope()
{
    TConstScopedObjects objs;
    GetSelection(objs);
    if (objs.size()==1) {
        return objs.begin()->scope; 
    }
    return CRef<CScope>();
}


///////////////////////////////////////////////////////////////////////////////
/// CProjectTreeViewFactory
string  CProjectTreeViewFactory::GetExtensionIdentifier() const
{
    static string sid("project_tree_view_factory");
    return sid;
}


string  CProjectTreeViewFactory::GetExtensionLabel() const
{
    static string slabel("Project Tree View Factory");
    return slabel;
}


void CProjectTreeViewFactory::RegisterIconAliases(wxFileArtProvider& provider)
{
    //TODO
}


const CViewTypeDescriptor& CProjectTreeViewFactory::GetViewTypeDescriptor() const
{
    return CProjectTreeView::m_TypeDescr;
}


IView* CProjectTreeViewFactory::CreateInstance() const
{
    return new CProjectTreeView();
}


IView* CProjectTreeViewFactory::CreateInstanceByFingerprint(const TFingerprint& fingerprint) const
{
    TFingerprint print(CProjectTreeView::m_TypeDescr.GetLabel(), true);
    if(print == fingerprint)   {
        return new CProjectTreeView();
    }
    return NULL;
}


END_NCBI_SCOPE
