/*  $Id: selection_service.cpp 27457 2013-02-20 15:59:39Z voronov $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Andrey Yazhuk
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>

#include <gui/core/selection_service_impl.hpp>

#include <gui/objects/GBWorkspace.hpp>
#include <gui/core/project_service.hpp>
#include <gui/core/project_view.hpp>
#include <gui/core/document.hpp>
#include <gui/core/commands.hpp>

#include <gui/objutils/registry.hpp>
#include <gui/objutils/obj_convert.hpp>

#include <gui/widgets/wx/app_status_bar.hpp>
#include <gui/widgets/wx/wx_utils.hpp>

#include <serial/iterator.hpp>

#include <wx/dcclient.h>
#include <wx/brush.h>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);


CSelectionService::CSelectionService()
:   m_ActiveClient(NULL),
    m_Broadcasting(false),
    m_SrvLocator(NULL),
    mf_AutoBroadcast(true)
{
}


CSelectionService::~CSelectionService()
{
}


void CSelectionService::InitService()
{
    LOG_POST(Info << "Initializing Selection Service...");

    TGuard lock(*this);

    LoadSettings();
    LOG_POST(Info << "Finished initializing Selection Service");
}


void CSelectionService::ShutDownService()
{
    LOG_POST(Info << "Shutting down Selection Service...");
    TGuard lock(*this);

    DetachAllClients();

    /// this must be coordinated correctly - order here is important!
    DetachAllObservers();

    SaveSettings();

    LOG_POST(Info << "Finished shutting down Selection Service");
}


void CSelectionService::AttachClient( ISelectionClient* client )
{
    TClients::const_iterator it = std::find(m_Clients.begin(), m_Clients.end(), client);
    if( it == m_Clients.end() ){
        m_Clients.push_back(client);
        client->SetSelectionService(this);

        m_ServiceUpdate = new CSelectionServiceUpdate( 
            CSelectionServiceUpdate::eClientAttached,
            client
        );

        x_NotifyObservers();
    } else {
        ERR_POST("CSelectionService::AttachClient() - client already registered");
    }
}


void CSelectionService::DetachClient(ISelectionClient* client)
{
    TClients::iterator it = std::find(m_Clients.begin(), m_Clients.end(), client);
    if(it != m_Clients.end())   {
        m_Clients.erase(it);
        client->SetSelectionService(NULL);
        if(m_ActiveClient == client)    {
            m_ActiveClient = NULL;
        }

        m_ServiceUpdate = new CSelectionServiceUpdate( 
            CSelectionServiceUpdate::eClientDetached,
            client
        );

        x_NotifyObservers();
    } else {
        ERR_POST("CSelectionService::DetachClient() - client is not attached");
    }
}

bool CSelectionService::IsAttached(ISelectionClient* client) const
{
    TClients::const_iterator it = std::find( m_Clients.begin(), m_Clients.end(), client );
    return it != m_Clients.end();
}


void CSelectionService::DetachAllClients()
{
    LockUpdates();

    while( !m_Clients.empty() ){
        DetachClient( m_Clients.back() );
    }

    UnlockUpdates();
}

ISelectionClient*   CSelectionService::GetActiveClient()
{
    return m_ActiveClient;
}


void CSelectionService::OnActiveClientChanged(ISelectionClient* active)
{
    if( active != m_ActiveClient ){
        m_ActiveClient = active;

        m_ServiceUpdate = new CSelectionServiceUpdate( 
            CSelectionServiceUpdate::eActiveChanged,
            active
        );

        x_NotifyObservers();
    }
}


void CSelectionService::GetClients(CSelectionService::TClients& clients)
{
    clients = m_Clients;
}


void CSelectionService::Broadcast( ISelectionClient* source, bool force, bool ranges )
{
    CSelectionEvent sel_evt( *source->GetScope(), this );
    source->GetSelection( sel_evt );

    if( ranges && !sel_evt.HasRangeSelection() ){
        
        CScope& scope = sel_evt.GetScope();

        TConstObjects objs;
        sel_evt.GetAllObjects( objs );


        ITypeConverter::TObjList locs;
        size_t n = objs.size();
        for(  size_t i = 0;  i < n;  i++ )  {
            const CObject& obj = *objs[i];

            locs.clear();
            CObjectConverter::Convert( 
                scope, obj, CSeq_loc::GetTypeInfo(), 
                locs, ITypeConverter::fConvert_NoExpensive
            );


            NON_CONST_ITERATE( ITypeConverter::TObjList, it, locs ){
                ITypeConverter::SObject& s_obj = *it;
                const CObject* conv_obj = s_obj.GetObjectPtr();

                _ASSERT(typeid(*s_obj) == typeid(CSeq_loc));

                const CSeq_loc* loc = reinterpret_cast<const CSeq_loc*>(conv_obj);
                sel_evt.AddRangeSelection( *loc );
            }
        }
    }

    Broadcast( sel_evt, source, force );
}

void CSelectionService::Broadcast( 
    CSelectionEvent& evt, ISelectionClient* source, bool force
){
    typedef CWorkspaceConstPrx::TProjectPrxVector  TProjectPrxVector;

    if( !m_Broadcasting ){   // block recursion
        m_Broadcasting = true;

        try {
            if( force || IsAutoBroadcast() ){

                CIRef<CProjectService> srv =
                    m_SrvLocator->GetServiceByType<CProjectService>()
                ;
                _ASSERT(srv);

                CRef<CWorkspaceConstPrx> workspace = srv->GetWorkspaceConst();

                if( !workspace.IsNull() ){
                    IProjectView* src_view = dynamic_cast<IProjectView*>(source);
                    if( CSelectionEvent::sm_InterDocBroadcast  ||  !src_view){
                        // broadcast to all documents
                        TProjectPrxVector projects;
                        workspace->GetProjects(projects, CWorkspaceConstPrx::eLoaded);

                        NON_CONST_ITERATE(TProjectPrxVector , it, projects)  {
                            CProjectPrx* project = it->GetPointer();
                            x_Broadcast(evt, source, project);
                        }
                    } else { // src_view must be a project view
                        // broadcast to all views on the same document
                        IProjectView::TProjectId id = src_view->GetProjectId();
                        CRef<CProjectPrx> project = workspace->FindProjectById(id);
                        _ASSERT(project);

                        x_Broadcast(evt, source, project);
                    }
                }
            }

            // to observers (which view changed)
            // x_NotifyObservers( CObserverUpdate::eStateChanged );
            OnSelectionChanged( source );
        
        } catch( CException& e ){
            ERR_POST(Error << "CSelectionService::Broadcast(): " << e.GetMsg());
        
        } catch( std::exception& e ){
            ERR_POST(Error << "CSelectionService::Broadcast(): " << e.what());
        
        } 

        m_Broadcasting = false;
    }
}

IObserverUpdate* CSelectionService::x_CreateStateChange()
{
    return m_ServiceUpdate;
}

void CSelectionService::OnSelectionChanged( ISelectionClient* source )
{
    //TODO consider broadcasting options here

    //TODO this can be improved by providing more specific information
    // to observers (which view changed)
    m_ServiceUpdate = new CSelectionServiceUpdate( 
        CSelectionServiceUpdate::eSelectionChanged,
        source
    );

    x_NotifyObservers();
}


void CSelectionService::GetCurrentSelection(TConstScopedObjects& sel)
{
    ISelectionClient* client = GetActiveClient();
    if(client)  {
        client->GetSelection(sel);
        if (sel.empty()) {
            /// the main object is selected only if no other selections are available
            client->GetMainObject(sel);
        }
    }
}

void CSelectionService::SetAutoBroadcast( bool pAuto )
{ 
    mf_AutoBroadcast = pAuto; 

    x_NotifyObservers();
}

void CSelectionService::x_Broadcast(CSelectionEvent& evt, ISelectionClient* source,
                                    CProjectPrx* target)
{
    // broadcast selection to views belonging to same document
    if(target)    {
        CProjectPrx::TViews& views = target->GetViews();

        NON_CONST_ITERATE(CGBDocument::TViews, it, views) {
            IProjectView* view = &**it;
            if(source != static_cast<ISelectionClient*>(view))  {
                ISelectionClient* sel_client = dynamic_cast<ISelectionClient*>(view);
                if(sel_client)    {
                    sel_client->SetSelection(evt);
                }
            }
        }
    }
}


void CSelectionService::SetRegistryPath(const string& reg_path)
{
    m_RegPath = reg_path;
}


static const char* kObjMatchPolicy = "ObjMatchPolicy";
static const char* kIdMatchPolicy = "IdMatchPolicy";
static const char* kMatchByProduct = "MatchByProduct";
static const char* kAutoBroadcast = "AutoBroadcast";
static const char* kInterDocBroadcast = "InterDocBroadcast";
static const char* kMatchAlnLocs = "MatchAlnLocs";

void CSelectionService::SaveSettings() const
{
    if(! m_RegPath.empty())   {
        CGuiRegistry& gui_reg = CGuiRegistry::GetInstance();
        CGuiRegistry::TReadWriteView view = gui_reg.GetReadWriteView(m_RegPath);

        view.Set(kObjMatchPolicy, CSelectionEvent::sm_ObjMatchPolicy);
        view.Set(kIdMatchPolicy, CSelectionEvent::sm_IdMatchPolicy);
        view.Set(kMatchByProduct, CSelectionEvent::sm_MatchByProduct);
        //view.Set(kAutoBroadcast, CSelectionEvent::sm_AutoBroadcast);
        view.Set( kAutoBroadcast, mf_AutoBroadcast );
        view.Set(kInterDocBroadcast, CSelectionEvent::sm_InterDocBroadcast);
        view.Set(kMatchAlnLocs, CSelectionEvent::sm_MatchAlnLocs);
    }
}


void CSelectionService::LoadSettings()
{
    if(! m_RegPath.empty())   {
        CGuiRegistry& gui_reg = CGuiRegistry::GetInstance();
        CGuiRegistry::TReadWriteView view = gui_reg.GetReadWriteView(m_RegPath);

        int value = view.GetInt(kObjMatchPolicy, CSelectionEvent::sm_ObjMatchPolicy);
        value = (value < 2) ? value : 0;
        CSelectionEvent::sm_ObjMatchPolicy = (CSelectionEvent::EObjMatchPolicy) value;

        value = view.GetInt(kIdMatchPolicy, CSelectionEvent::sm_IdMatchPolicy);
        value = (value < 2) ? value : 0;
        CSelectionEvent::sm_IdMatchPolicy = (CSelectionEvent::EIdMatchPolicy) value;

        CSelectionEvent::sm_MatchByProduct =
            view.GetBool(kMatchByProduct, CSelectionEvent::sm_MatchByProduct);
        CSelectionEvent::sm_AutoBroadcast =
            view.GetBool(kAutoBroadcast, CSelectionEvent::sm_AutoBroadcast)
        ;
        mf_AutoBroadcast = view.GetBool( kAutoBroadcast, mf_AutoBroadcast );
        CSelectionEvent::sm_InterDocBroadcast =
            view.GetBool(kInterDocBroadcast, CSelectionEvent::sm_InterDocBroadcast);
        CSelectionEvent::sm_MatchAlnLocs =
            view.GetBool(kMatchAlnLocs, CSelectionEvent::sm_MatchAlnLocs);
    }
}


void CSelectionService::SetServiceLocator(IServiceLocator* srv_locator)
{
    m_SrvLocator = srv_locator;
}

///////////////////////////////////////////////////////////////////////////////
/// CSelectionBroadcastSlot


BEGIN_EVENT_TABLE(CSelectionBroadcastSlot, wxPanel)
    EVT_PAINT(CSelectionBroadcastSlot::OnPaint)
    EVT_ENTER_WINDOW(CSelectionBroadcastSlot::OnMouseEnter)
    EVT_LEAVE_WINDOW(CSelectionBroadcastSlot::OnMouseLeave)
    EVT_LEFT_DOWN(CSelectionBroadcastSlot::OnLeftDown)
    EVT_LEFT_DCLICK(CSelectionBroadcastSlot::OnLeftDown)
END_EVENT_TABLE()


CSelectionBroadcastSlot::CSelectionBroadcastSlot( wxEvtHandler* cmd_handler )
    : m_CmdHandler(cmd_handler)
    , m_Sizer(NULL)
    , m_TextItem(NULL)
    , m_SelectionService(NULL)
    , m_Hot(false)
    , m_Idle(false)
{
    m_ObserverProxy.Reset( new CObserverProxy(*this) );
}

static const int kTextOffX = 4;
static const int kTextOffY = 2;

bool CSelectionBroadcastSlot::Create( wxWindow* parent )
{
    long style = wxCLIP_CHILDREN | wxBORDER_NONE;

    if( wxPanel::Create(parent, wxID_ANY, wxDefaultPosition, wxSize(), style) ){
            Hide();
            SetBackgroundStyle(wxBG_STYLE_CUSTOM);

            wxSize text_sz = x_GetTextAreaSize();
            /*
            m_Sizer = new wxBoxSizer( wxHORIZONTAL );

            // create a Spacer for rendering text
            m_TextItem = m_Sizer->Add( text_sz.x, text_sz.y, 1, wxEXPAND | wxFIXED_MINSIZE );

            SetSizer(m_Sizer);
            */

            m_TextItem = new wxStaticText( 
                this, wxID_ANY, "XXXXXXX", 
                wxDefaultPosition, wxDefaultSize, 
                wxALIGN_CENTRE_HORIZONTAL | wxST_NO_AUTORESIZE
            );
            Fit();

            //SetSize(text_sz);
            wxSize sz = GetSize();
            SetMinSize(sz);

            m_TextItem->Bind( wxEVT_ENTER_WINDOW, &CSelectionBroadcastSlot::OnMouseEnter, this );
            m_TextItem->Bind( wxEVT_LEAVE_WINDOW, &CSelectionBroadcastSlot::OnMouseLeave, this );
            m_TextItem->Bind( wxEVT_LEFT_DOWN, &CSelectionBroadcastSlot::OnLeftDown, this );
            m_TextItem->Bind( wxEVT_LEFT_DCLICK, &CSelectionBroadcastSlot::OnLeftDown, this );

            //m_TextItem->Hide();
            return true;
    }
    return false;
}


// calculate the minimal area size sufficient for displaying text
wxSize CSelectionBroadcastSlot::x_GetTextAreaSize()
{
    string s = x_StatusText(CSelectionEvent::eAllIds, CSelectionEvent::eAccOnly );

    wxClientDC dc( this );
    dc.SetFont( GetFont() );
    wxString wx_s = ToWxString( s );

    wxSize size;
    dc.GetTextExtent( wx_s, &size.x, &size.y );

    size.x += 2 * kTextOffX;
    size.y += 2 * kTextOffY;
    return size;
}

// IObserver function that handles updates from CAppTaskService
void CSelectionBroadcastSlot::OnUpdate(IObserverUpdate& update)
{
    CObserverUpdate* up = dynamic_cast<CObserverUpdate*>(&update);
    _ASSERT(up);

    IObservable* observable = up->GetObservable();
    CSelectionService* srv = dynamic_cast<CSelectionService*>(observable);
    _ASSERT(srv);

    switch(up->GetType())    {
    case CObserverUpdate::eAttached:
        _ASSERT( !m_SelectionService );
        m_SelectionService = srv;

        x_Update();
        break;

    case CObserverUpdate::eStateChanged:
        x_Update();
        break;

    case CObserverUpdate::ePreDetach:
        _ASSERT(m_SelectionService);

        m_SelectionService = NULL;
        x_Update();
        break;
    }
}

// generates a text string representing the current App Task Service state
string CSelectionBroadcastSlot::x_StatusText( CSelectionEvent::EObjMatchPolicy fObj, CSelectionEvent::EIdMatchPolicy fId )
{
    string s;
    /*
    if( fObj == CSelectionEvent::eAllIds ){
        s += "ALL IDS";
    } else {
        s += "ONE ID";
    }

    s += ";";
    */

    if( (fId == CSelectionEvent::eExact) && (fObj == CSelectionEvent::eAllIds) ){
        s += "EXACT";
    } else {
        s += "RELAXED";
    }

    return s;
}

string CSelectionBroadcastSlot::x_StatusTooltip( CSelectionEvent::EObjMatchPolicy fObj, CSelectionEvent::EIdMatchPolicy fId )
{
    string s;

    if( fId == CSelectionEvent::eExact ){
        s += "Exact molecule id matching for selection broadcasting is set";
    } else {
        s += "Molecule versions not used for matching of selection broadcasts";
    }

    return s;
}

void CSelectionBroadcastSlot::x_Update()
{
    m_Text = wxT("");
    int running_n = 0;
    
    wxString tooltip;

    if( m_SelectionService ){
        CSelectionEvent::EObjMatchPolicy obj_match = CSelectionEvent::sm_ObjMatchPolicy;
        CSelectionEvent::EIdMatchPolicy id_match = CSelectionEvent::sm_IdMatchPolicy;

        string s = x_StatusText( obj_match, id_match );
        m_Text = ToWxString(s);

        tooltip = ToWxString( x_StatusTooltip( obj_match, id_match ) );
    }
    tooltip += wxT("\n\nClick to see Broadcast Options");


    m_TextItem->SetLabelText( m_Text );
    m_TextItem->SetToolTip( tooltip );

    const wxColour back_cl = CStatusBar::GetBackColor( m_Hot );

    wxColor text_cl = CStatusBar::GetTextColor( m_Hot );
    if( 
        CSelectionEvent::sm_IdMatchPolicy == CSelectionEvent::eExact 
        && CSelectionEvent::sm_ObjMatchPolicy == CSelectionEvent::eAllIds
    ){
        text_cl = *wxRED; 
    }

    m_TextItem->SetBackgroundColour( back_cl );
    m_TextItem->SetForegroundColour( text_cl );

    Layout();
    Refresh();
}


void CSelectionBroadcastSlot::OnPaint(wxPaintEvent& event)
{
    wxPaintDC dc(this);

    // fill background
    wxRect rc = GetRect();
    rc.x = rc.y = 0;
    wxColour back_cl = CStatusBar::GetBackColor(m_Hot);
    wxBrush brush(back_cl);
    dc.SetBrush(brush);
    dc.SetPen(*wxTRANSPARENT_PEN);

    dc.DrawRectangle(rc);

    // draw text
    //rc = m_TextItem->GetRect();
    rc = this->GetRect();

    // set text font
    wxFont font = GetFont();
    font.SetUnderlined(m_Hot);
    dc.SetFont(font);

    // set text color
    wxColor cl = CStatusBar::GetTextColor(m_Hot);
    //wxColor text_cl = m_Idle ? GetAverage(cl, back_cl, 0.5) : cl;
    const wxColor& text_cl = 
        (CSelectionEvent::sm_IdMatchPolicy == CSelectionEvent::eExact && CSelectionEvent::sm_ObjMatchPolicy == CSelectionEvent::eAllIds) ? cl : *wxRED; 
    dc.SetTextForeground(text_cl);

    // prepare and render the text
    int av_w = rc.width - kTextOffX * 2;
    wxString s = TruncateText(dc, m_Text, av_w);

    int w = 0, h = 0;
    dc.GetTextExtent(s, &w, &h);
    int pos_x = rc.x + kTextOffX;
    int pos_y = rc.y + (rc.height - h) / 2;
#if defined(__WXOSX_CARBON__) || defined(__WXMAC_CARBON__)
    pos_y-=3;
#endif

    dc.DrawText(s, pos_x, pos_y);
}


void CSelectionBroadcastSlot::OnMouseEnter(wxMouseEvent& event)
{
    m_Hot = true;

    x_Update();
}


void CSelectionBroadcastSlot::OnMouseLeave(wxMouseEvent& event)
{
    m_Hot = false;

    x_Update();
}


void CSelectionBroadcastSlot::OnLeftDown(wxMouseEvent& event)
{
    wxCommandEvent evt(wxEVT_COMMAND_MENU_SELECTED, eCmdBroadcastOptions);
    evt.SetEventObject(this);
    m_CmdHandler->ProcessEvent(evt);
}



END_NCBI_SCOPE
