/*  $Id: selection_view.cpp 26310 2012-08-24 15:38:40Z falkrb $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Andrey Yazhuk
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>

#include <gui/core/selection_view.hpp>

#include <gui/core/view_extensions.hpp> //TODO
#include <gui/core/selection_service_impl.hpp> 

#include <gui/framework/view_manager_service.hpp>
#include <gui/framework/workbench.hpp>

#include <gui/objutils/registry.hpp>

#include <gui/widgets/wx/table_utils.hpp>
#include <gui/widgets/wx/wx_utils.hpp>
#include <gui/widgets/wx/fileartprov.hpp>
#include <gui/widgets/wx/message_box.hpp>
#include <gui/widgets/wx/ui_command.hpp>

#include <gui/utils/extension_impl.hpp>

#include <gui/widgets/text_widget/plain_text_item.hpp>
#include <gui/widgets/text_widget/composite_text_item.hpp>
#include <gui/widgets/object_list/object_text_params.hpp>

#include <wx/event.h>
#include <wx/sizer.h>
#include <wx/panel.h>
#include <wx/menu.h>
#include <wx/toolbar.h>
#include <wx/statline.h>

BEGIN_NCBI_SCOPE

///////////////////////////////////////////////////////////////////////////////
/// CSelectionTable

#define ACTIVE_VIEW_TARGET_IX 0
#define ALL_VIEWS_TARGET_IX 1

#define ID_TABLE    10001
#define ID_TEXT     10002
#define ID_TOOLBAR  10003
#define ID_COMBOBOX 10004

BEGIN_EVENT_TABLE( CSelectionPanel, wxPanel )
    EVT_CONTEXT_MENU( CSelectionPanel::OnContextMenu )
    EVT_CHOICE( ID_COMBOBOX, CSelectionPanel::OnTargetChanged )
    EVT_TOOL_RANGE( scm_ModeCmd+eTable, scm_ModeCmd+eText, CSelectionPanel::OnModeChanged )
END_EVENT_TABLE()


bool CSelectionPanel::m_RegisteredIcons = false;



CSelectionPanel::CSelectionPanel(CSelectionView* view)
    : m_SelView(view)
    , m_Service(NULL)
    , m_TargetChoice(NULL)
    , m_DisplayMode(eInvalidDisplayMode)
    , m_ListWidget(NULL)
    , m_TextWidget(NULL)
{
    Init();
}

CSelectionPanel::~CSelectionPanel()
{
}

void CSelectionPanel::Create( wxWindow* parent, wxWindowID id, const wxPoint& pos, const wxSize& size )
{
    wxPanel::Create( parent, id, pos, size );
    CreateControls();

    PushEventHandler( new CCommandToFocusHandler( this ) );
}

bool CSelectionPanel::Destroy()
{
    PopEventHandler( true );

    return wxPanel::Destroy();
}

void CSelectionPanel::Init()
{
    m_Service = NULL;
    m_TargetChoice = NULL;
    m_ListWidget = NULL;
    m_TextWidget = NULL;

    // register icons only once
    if( ! m_RegisteredIcons ){
        wxFileArtProvider* provider = GetDefaultFileArtProvider();

        provider->RegisterFileAlias( wxT("menu::insp_table_mode"), wxT("insp_table_mode.png") );
        provider->RegisterFileAlias( wxT("menu::insp_brief_text_mode"), wxT("insp_brief_text_mode.png") );
        provider->RegisterFileAlias( wxT("menu::insp_text_mode"), wxT("insp_text_mode.png") );

        m_RegisteredIcons = true;
    }
}

void CSelectionPanel::CreateControls()
{
    wxSizer* sizer = new wxBoxSizer( wxVERTICAL );
    SetSizer( sizer );

    // Create tool bar
    wxToolBar* toolBar = new wxToolBar(
        this, ID_TOOLBAR, wxDefaultPosition, wxDefaultSize,
        wxTB_FLAT | wxTB_HORIZONTAL | wxTB_TEXT | wxTB_HORZ_LAYOUT
    );

    toolBar->AddRadioTool(scm_ModeCmd + eTable, wxT("Table"),
                          wxArtProvider::GetBitmap(wxT("menu::insp_table_mode")),
                          wxNullBitmap, wxT("Table mode"));
    toolBar->AddRadioTool(scm_ModeCmd + eBriefText, wxT("Brief"),
                          wxArtProvider::GetBitmap(wxT("menu::insp_brief_text_mode")),
                          wxNullBitmap, wxT("Brief Text mode"));
    toolBar->AddRadioTool(scm_ModeCmd + eText, wxT("Full"),
                          wxArtProvider::GetBitmap( wxT("menu::insp_text_mode") ),
                          wxNullBitmap, wxT("Full Text mode"));
    toolBar->AddSeparator();

    // Had used a combo box here but that has a strange error on macos carbon - it
    // becomes disabled when certain other windows (like project view) are selected
    // and can't be easily re-enabled (no standard wx functions, enable, setfocus, raise, etc
    // had any effect, but selecting certain Other views, e.g. tree view, did
    // re-enable combo-boxes on this panel...).
    wxArrayString strings;
    m_TargetChoice = new wxChoice(toolBar, ID_COMBOBOX,
                                  wxDefaultPosition, wxSize(250, -1), strings);
    toolBar->AddControl( m_TargetChoice );

    toolBar->Realize();
    sizer->Add( toolBar, 0, wxEXPAND );

    // separation line before the table
    wxStaticLine* line = new wxStaticLine(
        this, wxID_STATIC, wxDefaultPosition, wxDefaultSize, wxLI_HORIZONTAL
    );
    sizer->Add( line, 0, wxEXPAND );

    // Create list widget
    m_ListWidget = new CObjectListWidget( this, ID_TABLE, wxDefaultPosition, wxSize(100, 100), wxBORDER_NONE );
    m_ListWidget->SetContextMenuEventPropagation( true );
    sizer->Add( m_ListWidget, 1, wxEXPAND );

    // create text widget
    m_TextWidget = new CTextItemPanel( this, ID_TEXT );
    m_TextWidget->Hide();
    m_TextWidget->SetBackgroundColour( wxColor(wxT("white")) );
    m_TextWidget->SetFont(wxFont(8, wxFONTFAMILY_MODERN, wxFONTSTYLE_NORMAL, wxFONTWEIGHT_NORMAL));

    sizer->Add(m_TextWidget, 1, wxEXPAND);

    m_DisplayMode = eTable;

    Update();
}

void CSelectionPanel::SetSelectionService( CSelectionService* service )
{
    m_Service = service;
    if( !m_Service ){
         m_CurrentClientWeakPtr = NULL;
    }

    Update();
}

void CSelectionPanel::SetRegistryPath( const string& path )
{
    m_RegPath = path; // store for later use
}

void CSelectionPanel::SaveSettings() const
{
    if( ! m_RegPath.empty() ){
        CGuiRegistry& gui_reg = CGuiRegistry::GetInstance();
        CGuiRegistry::TReadWriteView view = gui_reg.GetReadWriteView(m_RegPath);

        _ASSERT(m_ListWidget);

        SaveTableSettings( *m_ListWidget, view );
    }
}

void CSelectionPanel::LoadSettings()
{
    if( ! m_RegPath.empty() ){
        CGuiRegistry& gui_reg = CGuiRegistry::GetInstance();
        CGuiRegistry::TReadView view = gui_reg.GetReadView( m_RegPath );

        _ASSERT( m_ListWidget );

        LoadTableSettings( *m_ListWidget, view );
    }
}

void CSelectionPanel::GetSelection( TConstScopedObjects& buf ) const
{
    if( m_DisplayMode == eTable ){
        m_ListWidget->GetSelection( buf );
    } else {
        CTextPanelContext* context = m_TextWidget->GetContext();
        if (context)
            context->GetSelectedObjects( buf );
    }
}

void CSelectionPanel::AppendMenuItems( wxMenu& /*aMenu*/ )
{
}

void CSelectionPanel::OnContextMenu( wxContextMenuEvent& anEvent )
{
    wxMenu* menu;

    void* data = anEvent.GetClientData();
    menu = data ? (wxMenu*)data : new wxMenu();

    AppendMenuItems( *menu );

    if( /*ShouldPropagate()*/ true ){
        anEvent.SetClientData( menu );
        anEvent.Skip();
        return;
    }

    anEvent.SetClientData( NULL );

    CleanupSeparators( *menu );
    PopupMenu( menu );

    delete menu;
}

void CSelectionPanel::OnTargetChanged( wxCommandEvent& /*event*/ )
{
    x_ResetObjectWidget();
}

void CSelectionPanel::OnModeChanged( wxCommandEvent& event )
{
    EDisplayMode new_mode = x_GetModeByCommand( event.GetId() );

    if( new_mode != m_DisplayMode ){
        m_DisplayMode = new_mode;

        if( new_mode == eTable ){
            m_TextWidget->Hide();
            m_ListWidget->Show();
        } else {
            m_ListWidget->Hide();
            m_TextWidget->Show();
            x_InitTextWidget();
        }

        x_ResetObjectWidget();
    }
}


CSelectionPanel::EDisplayMode CSelectionPanel::x_GetModeByCommand(TCmdID cmd)
{
    return EDisplayMode(cmd - scm_ModeCmd);
}

/*
ISelectionClient* CSelectionPanel::x_GetSelectedClient()
{
    int target_ix = m_TargetChoice->GetSelection();
    if( target_ix != wxNOT_FOUND ) {
        if( target_ix == ACTIVE_VIEW_TARGET_IX ){
            if(  m_Service ){
                return m_Service->GetActiveClient();
            }
        } else {
            return m_IndexToClient[target_ix];
        }
    }

    return NULL;
}
*/

void CSelectionPanel::x_InitCurrentWidget()
{
    if( m_DisplayMode == eTable ){
        m_ListWidget->SetObjects( m_SelObjects );

    } else {
        if( !( m_TextWidget->GetContext() && m_TextWidget->Selecting() ) ){
            x_InitTextWidget();
        }
    }
}

void CSelectionPanel::x_InitTextWidget()
{
    vector<ITextItem*> items;

    CObjectTextParams params((m_DisplayMode == eBriefText) ? 
        CObjectTextParams::kBrief : CObjectTextParams::kFull);

    for (size_t i = 0; i < m_SelObjects.size(); ++i) {
        ITextItem* item = CreateObjectInterface<ITextItem>(m_SelObjects[i], &params);
        if (item) {
            if (!items.empty()) {
                CPlainTextItem* itemDiv = new CPlainTextItem();
                itemDiv->AddLine( " " );
                items.push_back( itemDiv );
            }
            items.push_back(item);
        }
    }

    m_TextWidget->SetData( new CCompositeTextItem(), new CTextPanelContext(2) );
    m_TextWidget->AddItems( items );
}

void CSelectionPanel::OnSelServiceStateChange( CSelectionServiceUpdate& update )
{
    switch( update.GetClientUpdateType() ){
    case CSelectionServiceUpdate::eClientAttached:

        x_ResetTargetChoice();

        break;

    case CSelectionServiceUpdate::eClientDetached:
        {{
        CIRef<ISelectionClient> wic(m_CurrentClientWeakPtr.Lock());
        ISelectionClient* current_client = wic.GetPointerOrNull();

        if( current_client == update.GetClientUpdated() ){
            wic.ReleaseOrNull();
            m_CurrentClientWeakPtr = NULL;
            //m_CurrentClient = NULL;

            Update();
        } else {
            x_ResetTargetChoice();
        }
        }}
        break;

    case CSelectionServiceUpdate::eActiveChanged:
        {{
        CIRef<ISelectionClient> wic(m_CurrentClientWeakPtr.Lock());
        ISelectionClient* current_client = wic.GetPointerOrNull();

            ISelectionClient* new_active = update.GetClientUpdated();
            if( 
                new_active != current_client
                && new_active != m_SelView 
            ){
                if( 
                    new_active == NULL
                    ||( m_Service && m_Service->IsAttached( new_active ) )
                ){
                    wic.ReleaseOrNull();
                    m_CurrentClientWeakPtr = new_active;
                }

                Update();
            }
        }}
        break;

    case CSelectionServiceUpdate::eSelectionChanged:
        x_ResetObjectWidget();
        break;

    default:
        _ASSERT(false);
    }
}

void CSelectionPanel::Update()
{
    x_ResetTargetChoice();

    x_ResetObjectWidget();
}

void CSelectionPanel::x_ResetTargetChoice()
{
    ISelectionClient* cur_target = NULL;
    int target_ix = m_TargetChoice->GetSelection();

    if( target_ix == wxNOT_FOUND ){
        target_ix = ACTIVE_VIEW_TARGET_IX; 

    } else if( target_ix > ALL_VIEWS_TARGET_IX ){
        cur_target = m_IndexToClient[target_ix];
    }

    m_TargetChoice->Clear();
    m_IndexToClient.clear();

    string s_active = "Active View";
    {{
    CIRef<ISelectionClient> wic(m_CurrentClientWeakPtr.Lock());
    ISelectionClient* current_client = wic.GetPointerOrNull();

    // add "Active Client" line
    if( current_client ){
        s_active += " - " + current_client->GetSelClientLabel();
    }
    }}

    m_TargetChoice->Append( ToWxString(s_active) );
    m_IndexToClient[ACTIVE_VIEW_TARGET_IX] = NULL;

    /// add "All Views" line
    m_TargetChoice->Append( wxT("All Views") );
    m_IndexToClient[ALL_VIEWS_TARGET_IX] = NULL;

    CSelectionService::TClients clients;

    if( m_Service ){
        m_Service->GetClients( clients );
    }

    if( clients.empty() ){
        m_TargetChoice->SetSelection( ACTIVE_VIEW_TARGET_IX );
        return;
    }

    /// Add lines for all existing clients
    ITERATE( CSelectionService::TClients, it, clients ){
        ISelectionClient* client = *it;
        if( client == m_SelView ){   // skip myself
            continue;
        }

        string s = client->GetSelClientLabel();
        int index = m_TargetChoice->Append( ToWxString(s) );
        m_IndexToClient[index] = client;

        if( client == cur_target ){
            target_ix = index;
        }
    }
    m_TargetChoice->SetSelection( target_ix );

    m_TargetChoice->InvalidateBestSize();
    wxToolBar* toolbar = (wxToolBar*)m_TargetChoice->GetParent();
    toolbar->Realize();
}

bool CSelectionPanel::x_ResetSelObjects()
{
    int target_ix = m_TargetChoice->GetSelection();

    if( !m_Service || target_ix == wxNOT_FOUND ){
        return false;
    }

    TConstScopedObjects tmp_sel_objs;
    if( target_ix == ALL_VIEWS_TARGET_IX ){   // All Views
        CSelectionService::TClients clients;
        m_Service->GetClients( clients );

        ITERATE( CSelectionService::TClients, it, clients ){
            ISelectionClient* client = *it;
            if( client != m_SelView ){
                client->GetSelection( tmp_sel_objs );
            }
        }
    } else {
        ISelectionClient* client;
        CIRef<ISelectionClient> wic(m_CurrentClientWeakPtr.Lock());
        ISelectionClient* current_client = wic.GetPointerOrNull();

        if( target_ix == ACTIVE_VIEW_TARGET_IX ){   // Active View
            client = current_client;
        } else {
            client = m_IndexToClient[target_ix];
        }

        if( client ){
            client->GetSelection( tmp_sel_objs );
        }
    }

    bool matched = false;
    if (m_SelObjects.size() == tmp_sel_objs.size()) {
        matched = true;
        set< CConstRef<CObject> > objs;
        ITERATE (TConstScopedObjects, iter, m_SelObjects) {
            objs.insert(iter->object);
        }

        ITERATE (TConstScopedObjects, iter, tmp_sel_objs) {
            if (objs.count(iter->object) == 0) {
                matched = false;
                break;
            }
        }
    }

    if (matched) {
        // no selection changed
        return false;
    }

    m_SelObjects.swap(tmp_sel_objs);
    return true;
}

void CSelectionPanel::x_ResetObjectWidget()
{
    if (x_ResetSelObjects()) {
        x_InitCurrentWidget();
    }

    Layout();
    Refresh();
}

///////////////////////////////////////////////////////////////////////////////
/// CSelectionView

/// register the factory
static CExtensionDeclaration
    decl( "view_manager_service::view_factory", new CSelectionViewFactory() )
;

CViewTypeDescriptor CSelectionView::m_TypeDescr(
    "Selection Inspector", // type name
    "selection_inspector", // icon alias TODO
    "Selection Inspector displays information about objects selected in views",
    "Selection Inspector displays information about objects selected in views",
    "SELECTION_INSPECTOR", // help ID
    "System",     // category
    true // singleton
);

class CSelectionViewEventHandler : public wxEvtHandler
{
public:
    CSelectionViewEventHandler( CSelectionView* aView ) 
        : m_View( aView )
    {}

    void OnContextMenu( wxContextMenuEvent& anEvent );

private:
    CSelectionView* m_View;

    DECLARE_EVENT_TABLE()
};

BEGIN_EVENT_TABLE( CSelectionViewEventHandler, wxEvtHandler )
    EVT_CONTEXT_MENU( CSelectionViewEventHandler::OnContextMenu )
END_EVENT_TABLE()

void CSelectionViewEventHandler::OnContextMenu( wxContextMenuEvent& anEvent )
{
    auto_ptr<wxMenu> menu( CreateContextMenuBackbone() );

    void* data = anEvent.GetClientData();
    if( data ){
        Merge( *menu.get(), *(wxMenu*)data );

        anEvent.SetClientData( NULL );
        delete (wxMenu*)data;
    }

    CleanupSeparators( *menu );
    m_View->GetWindow()->PopupMenu( menu.get() );
}



CSelectionView::CSelectionView()
    : m_Service( NULL )
    , m_Broadcasting( false )
    , m_Panel( NULL )
    , m_Handler( new CSelectionViewEventHandler( this ) )
{
}

CSelectionView::~CSelectionView()
{
    SetWorkbench(NULL);
    delete m_Handler;
}

void CSelectionView::OnUpdate( IObserverUpdate& update )
{
    CObserverUpdate* upd = dynamic_cast<CObserverUpdate*>(&update);
    _ASSERT(upd);

    IObservable* observable = upd->GetObservable();
    CSelectionService* service = dynamic_cast<CSelectionService*>(observable);
    _ASSERT(service);

    switch( upd->GetType() ){
    case CObserverUpdate::eAttached:
        _ASSERT(! m_Service );
        m_Service = service;
        m_Panel->SetSelectionService(m_Service);
        break;
    case CObserverUpdate::eStateChanged:
        {{
            CSelectionServiceUpdate* ssu = dynamic_cast<CSelectionServiceUpdate*>(upd);

            if( ssu ){
                m_Panel->OnSelServiceStateChange( *ssu );
            } else {
                m_Panel->Update();
            }
        }}
        break;

    case CObserverUpdate::ePreDetach:
        _ASSERT(m_Service);

        m_Service = NULL;
        m_Panel->SetSelectionService(m_Service);
        break;
    }
}

/////////////////////////////////////////
/// IView interface implementation

const CViewTypeDescriptor& CSelectionView::GetTypeDescriptor() const
{
    return m_TypeDescr;
}


void CSelectionView::SetWorkbench( IWorkbench* workbench )
{
    if(workbench)  {
        // connect to services
        CIRef<CSelectionService> sel_srv  = workbench->GetServiceByType<CSelectionService>();
        if(sel_srv)  {
            sel_srv->AttachObserver(this);
        }
    } else {
        if(m_Service)   {
            // disconnect from services
            IObservable* observable = dynamic_cast<IObservable*>(m_Service);
            if(observable){
                observable->DetachObserver(this);
            }
        }
    }
}

void CSelectionView::CreateViewWindow( wxWindow* parent )
{
    m_Panel = new CSelectionPanel(this);
    m_Panel->Create( parent, wxID_ANY );

    if( m_Handler ){
        GetWindow()->PushEventHandler( m_Handler );
    }
}


void CSelectionView::DestroyViewWindow()
{
    if( m_Panel ){
        if( m_Handler ){
            GetWindow()->PopEventHandler( false );
        }

        m_Panel->Destroy();
        m_Panel = NULL;
    }
}

/////////////////////////////////////////
/// IWMClient interface implementation

wxWindow* CSelectionView::GetWindow()
{
    return m_Panel;
}


string CSelectionView::GetClientLabel(IWMClient::ELabel) const
{
    return m_TypeDescr.GetLabel();
}


string CSelectionView::GetIconAlias() const
{
    return m_TypeDescr.GetIconAlias();
}


const CRgbaColor* CSelectionView::GetColor() const
{
    return NULL;
}


void CSelectionView::SetColor( const CRgbaColor& /*color*/ )
{
    // do nothing
}


const wxMenu* CSelectionView::GetMenu()
{
    return NULL;
}


void CSelectionView::UpdateMenu( wxMenu& /*root_menu*/ )
{
}


IWMClient::CFingerprint CSelectionView::GetFingerprint() const
{
    return CFingerprint(m_TypeDescr.GetLabel(), true);
}


/////////////////////////////////////////////
/// IRegSettings interface implementation

void CSelectionView::SetRegistryPath( const string& path )
{
    m_Panel->SetRegistryPath( path ); // store for later use
}

void CSelectionView::SaveSettings() const
{
    m_Panel->SaveSettings();
}


void CSelectionView::LoadSettings()
{
    m_Panel->LoadSettings();
}


void CSelectionView::GetSelection(TConstScopedObjects& buf) const
{
    m_Panel->GetSelection(buf);
}


void CSelectionView::SetSelectionService(ISelectionService* service)
{
    CSelectionService* srv = dynamic_cast<CSelectionService*>( service );
    m_Service = srv;
}


string  CSelectionView::GetSelClientLabel()
{
    return "Selection Inspector";
}


void CSelectionView::GetSelection(CSelectionEvent& evt) const
{
    TConstScopedObjects sel_objs;
    GetSelection(sel_objs);
    if( ! sel_objs.empty())    {
        TConstObjects objs;
        ITERATE(TConstScopedObjects, it, sel_objs)   {
            objs.push_back(it->object);
        }

        // brodcast everything as Object selection, we cannot distinguish object
        // selection from range selection at this point
        evt.AddObjectSelection(objs);
    }
}


void CSelectionView::SetSelection(CSelectionEvent& evt)
{
    // not supported
}


void CSelectionView::GetMainObject(TConstScopedObjects& /*objects*/) const
{
    // there is no main object here
}


CRef<objects::CScope> CSelectionView::GetScope() const
{
    CRef<objects::CScope> sc;
    return sc; // because elements may belong to different scopes
}


///////////////////////////////////////////////////////////////////////////////
/// CSelectionViewFactory
string CSelectionViewFactory::GetExtensionIdentifier() const
{
    static string sid("selection_inspector_factory");
    return sid;
}


string CSelectionViewFactory::GetExtensionLabel() const
{
    static string slabel("Selection Inspector Factory");
    return slabel;
}


void CSelectionViewFactory::RegisterIconAliases(wxFileArtProvider& /*provider*/)
{
    //TODO
}


const CViewTypeDescriptor& CSelectionViewFactory::GetViewTypeDescriptor() const
{
    return CSelectionView::m_TypeDescr;
}


IView* CSelectionViewFactory::CreateInstance() const
{
    return new CSelectionView();
}


IView* CSelectionViewFactory::CreateInstanceByFingerprint(const TFingerprint& fingerprint) const
{
    TFingerprint print(CSelectionView::m_TypeDescr.GetLabel(), true);
    if(print == fingerprint)   {
        return CreateInstance();
    }

    return NULL;
}


END_NCBI_SCOPE
