/*  $Id: table_file_load_manager.cpp 27465 2013-02-21 14:55:55Z falkrb $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Roman Katargin
 *
 */

#include <ncbi_pch.hpp>

#include <gui/core/table_file_load_manager.hpp>

#include <gui/widgets/wx/wx_utils.hpp>
#include <gui/widgets/wx/async_call.hpp>
#include <gui/core/file_extensions.hpp>
#include <gui/core/table_object_loader.hpp>
#include <gui/core/table_column_type_guesser.hpp>

#include "table_format_panel.hpp"
#include "table_delimeters_panel.hpp"
#include "table_fixed_width_panel.hpp"
#include "table_column_id_panel.hpp"
#include "table_xform_panel.hpp"

#include <wx/filename.h>

BEGIN_NCBI_SCOPE

//
// CSelectFormatPage navigation
//

wxPanel* CTableFileLoadManager::CSelectFormatPage::GetPanel()
{
    return m_Manager.x_GetSelectFormatPanel();
}

bool CTableFileLoadManager::CSelectFormatPage::CanLeavePage(bool forward)
{
    CTableFormatPanel* panel = m_Manager.x_GetSelectFormatPanel();
    if (forward && !panel->IsInputValid())
        return false;

    panel->SaveSettings();
    return true;
}

IWizardPage* CTableFileLoadManager::CSelectFormatPage::GetNextPage()
{
    CTableFormatPanel* panel = m_Manager.x_GetSelectFormatPanel();
    if (panel->IsDelimitedFile())
        return &m_Manager.m_DelimitersPage;
    else
        return &m_Manager.m_FixedWidthPage;
}

//
// CDelimitersPage navigation
//

wxPanel* CTableFileLoadManager::CDelimitersPage::GetPanel()
{
    return m_Manager.x_GetDelimitersPanel();
}

bool CTableFileLoadManager::CDelimitersPage::CanLeavePage(bool forward)
{
    CTableDelimitersPanel* panel = m_Manager.x_GetDelimitersPanel();
    if (forward) {
        if (!panel->IsInputValid())
            return false;

        panel->SaveSettings();
        m_Manager.x_GuessColumns();
        return true;
    }

    panel->SaveSettings();
    return true;
}

IWizardPage* CTableFileLoadManager::CDelimitersPage::GetPrevPage()
{
    return &m_Manager.m_SelectFormatPage;
}

IWizardPage* CTableFileLoadManager::CDelimitersPage::GetNextPage()
{
    m_Manager.m_ColumnIdPage.SetPrevPage(this);
    return &m_Manager.m_ColumnIdPage;
}

//
// CFixedWidthPage navigation
//

wxPanel* CTableFileLoadManager::CFixedWidthPage::GetPanel()
{
    return m_Manager.x_GetFixedWidthPanel();
}

bool CTableFileLoadManager::CFixedWidthPage::CanLeavePage(bool forward)
{
    CTableFixedWidthPanel* panel = m_Manager.x_GetFixedWidthPanel();
    if (forward) {
        if (!panel->IsInputValid())
            return false;

        panel->SaveSettings();
        m_Manager.x_GuessColumns();
        return true;
    }

    panel->SaveSettings();
    return true;
}

IWizardPage* CTableFileLoadManager::CFixedWidthPage::GetPrevPage()
{
    return &m_Manager.m_SelectFormatPage;
}

IWizardPage* CTableFileLoadManager::CFixedWidthPage::GetNextPage()
{
    m_Manager.m_ColumnIdPage.SetPrevPage(this);
    return &m_Manager.m_ColumnIdPage;
}

//
// CColumnIdPage navigation
//

wxPanel* CTableFileLoadManager::CColumnIdPage::GetPanel()
{
    return m_Manager.x_GetColumnIdPanel();
}

bool CTableFileLoadManager::CColumnIdPage::CanLeavePage(bool forward)
{
    CTableColumnIdPanel* panel = m_Manager.x_GetColumnIdPanel();
    if (forward) {
        if (!panel->IsInputValid())
            return false;

        panel->SaveSettings();
        m_Manager.x_ConvertToSeqAnnot();
        return true;
    }

    panel->SaveSettings();
    return true;
}

IWizardPage* CTableFileLoadManager::CColumnIdPage::GetPrevPage()
{
    return CWizardPage::GetPrevPage();
}

IWizardPage* CTableFileLoadManager::CColumnIdPage::GetNextPage()
{
    return &m_Manager.m_ColumnXForm;
}

//
// CColumnXFormPage navigation
//

wxPanel* CTableFileLoadManager::CColumnXFormPage::GetPanel()
{
    return m_Manager.x_GetColumnXForm();
}

bool CTableFileLoadManager::CColumnXFormPage::CanLeavePage(bool forward)
{
    CTableXformPanel* panel = m_Manager.x_GetColumnXForm();
    if (forward && !panel->IsInputValid())
        return false;

    panel->SaveSettings();
    return true;
}

IWizardPage* CTableFileLoadManager::CColumnXFormPage::GetPrevPage()
{
    return &m_Manager.m_ColumnIdPage;
}

IWizardPage* CTableFileLoadManager::CColumnXFormPage::GetNextPage()
{
    return 0;
}

///////////////////////////////////////////////////////////////////////////////
//
// Asynch call for converting table data to seq-annot
//
///////////////////////////////////////////////////////////////////////////////
class CConvertTable_local : public CAsyncCall::Call
{
public:
    CConvertTable_local(CRef<CTableImportDataSource>  table,
                        CRef<CTableAnnotDataSource> seq_annot)
    : CAsyncCall::Call(wxT("Preparing Table Data..."))
    , m_Table(table)
    , m_SeqAnnot(seq_annot) {}  

    virtual void Execute() 
    { 
        m_Table->ConvertToSeqAnnot(m_SeqAnnot->GetContainer());
    } 

private:
    CRef<CTableImportDataSource> m_Table;
    CRef<CTableAnnotDataSource> m_SeqAnnot;
};

///////////////////////////////////////////////////////////////////////////////
/// CTableFileLoadManager
CTableFileLoadManager::CTableFileLoadManager()
:   m_Descr("Table files", ""),
    m_SrvLocator(NULL),
    m_ParentWindow(NULL),
    m_ImportedTableData(new CTableImportDataSource()),
    m_AnnotTableData(new CTableAnnotDataSource()),
    m_TableFormatPanel(),
    m_TableDelimitersPanel(),
    m_TableFixedWidthPanel(),
    m_TableColumnIdPanel(),
    m_TableColumnXFormPanel(),
    m_SelectFormatPage(*this),
    m_DelimitersPage(*this),
    m_FixedWidthPage(*this),
    m_ColumnIdPage(*this),
    m_ColumnXForm(*this),
    m_CurrentPage(&m_SelectFormatPage)
{
}


void CTableFileLoadManager::SetServiceLocator(IServiceLocator* srv_locator)
{
    m_SrvLocator = srv_locator;
}


void CTableFileLoadManager::SetParentWindow(wxWindow* parent)
{
    m_ParentWindow = parent;
}


const IUIObject& CTableFileLoadManager::GetDescriptor() const
{
    return m_Descr;
}


void CTableFileLoadManager::InitUI()
{
    m_CurrentPage = &m_SelectFormatPage;
}


void CTableFileLoadManager::CleanUI()
{
    m_TableFormatPanel = NULL;
    m_TableDelimitersPanel = NULL;
    m_TableFixedWidthPanel = NULL;
    m_TableColumnIdPanel = NULL;
    m_TableColumnXFormPanel = NULL;
    m_CurrentPage = &m_SelectFormatPage;
}


wxPanel* CTableFileLoadManager::GetCurrentPanel()
{
    return m_CurrentPage ? m_CurrentPage->GetPanel() : (wxPanel*)0;
}

void CTableFileLoadManager::x_ConvertToSeqAnnot()
{
    // Now convert the table to a seq-annot - could be slow if there is a lot
    // of data so use async converter to give user a message as to what's
    // going on.   
    m_AnnotTableData.Reset(new CTableAnnotDataSource()),   

    ERR_POST(Info << "Import Table - Converting SeqTable to SeqAnnot");                
    CConvertTable_local convert_to_sa(m_ImportedTableData, 
        m_AnnotTableData);
    CAsyncCall::Execute(convert_to_sa); 
}

void CTableFileLoadManager::x_GuessColumns()
{
    CGuessColumns_local guess_cols(m_ImportedTableData);
    CAsyncCall::Execute(guess_cols);  
    ERR_POST(Info << "Table Import - Guessed Columns: ");
    m_ImportedTableData->LogColumnInfo();
}

static const string kSelectFormatTag = ".SelectFormatPanel";
static const string kDelimitersTag = ".DelimitersPanel";
static const string kFixedWidthTag = ".FixedWidthPanel";
static const string kColumnIdTag = ".ColumnIdPanel";
static const string kColumnXformTag = ".ColumnXformPanel";

void CTableFileLoadManager::SetRegistryPath(const string& path)
{
    m_RegPath = path;
}

void CTableFileLoadManager::SaveSettings() const
{
}


void CTableFileLoadManager::LoadSettings()
{
}

CTableFormatPanel* CTableFileLoadManager::x_GetSelectFormatPanel()
{
    if (m_TableFormatPanel == NULL) {
        m_TableFormatPanel = new CTableFormatPanel(m_ParentWindow);
        m_TableFormatPanel->SetImportDataSource(m_ImportedTableData);
     
        if ( !m_RegPath.empty() ) {
            m_TableFormatPanel->SetRegistryPath(m_RegPath + kSelectFormatTag);
            m_TableFormatPanel->LoadSettings();
        }
    }
    return m_TableFormatPanel;
}

CTableDelimitersPanel* CTableFileLoadManager::x_GetDelimitersPanel()
{
    if (m_TableDelimitersPanel == NULL) {
        m_TableDelimitersPanel = new CTableDelimitersPanel(m_ParentWindow);
        m_TableDelimitersPanel->SetImportDataSource(m_ImportedTableData);
     
        if ( !m_RegPath.empty() ) {
            m_TableDelimitersPanel->SetRegistryPath(m_RegPath + kDelimitersTag);
            m_TableDelimitersPanel->LoadSettings();
        }
    }
    return m_TableDelimitersPanel;
}

CTableFixedWidthPanel* CTableFileLoadManager::x_GetFixedWidthPanel()
{
    if (m_TableFixedWidthPanel == NULL) {
        m_TableFixedWidthPanel = new CTableFixedWidthPanel(m_ParentWindow);
        m_TableFixedWidthPanel->SetImportDataSource(m_ImportedTableData);
     
        if ( !m_RegPath.empty() ) {
            m_TableFixedWidthPanel->SetRegistryPath(m_RegPath + kFixedWidthTag);
            m_TableFixedWidthPanel->LoadSettings();
        }
    }
    return m_TableFixedWidthPanel;
}

CTableColumnIdPanel* CTableFileLoadManager::x_GetColumnIdPanel()
{
    if (m_TableColumnIdPanel == NULL) {
        m_TableColumnIdPanel = new CTableColumnIdPanel(m_ParentWindow);
        m_TableColumnIdPanel->SetImportDataSource(m_ImportedTableData);
     
        if ( !m_RegPath.empty() ) {
            m_TableColumnIdPanel->SetRegistryPath(m_RegPath + kColumnIdTag);
            m_TableColumnIdPanel->LoadSettings();
        }
    }
    return m_TableColumnIdPanel;
}

CTableXformPanel* CTableFileLoadManager::x_GetColumnXForm()
{
    if (m_TableColumnXFormPanel == NULL) {
        m_TableColumnXFormPanel = new CTableXformPanel(m_ParentWindow);
        m_TableColumnXFormPanel->SetImportDataSource(m_ImportedTableData);
        m_TableColumnXFormPanel->SetAnnotDataSource(m_AnnotTableData);
     
        if ( !m_RegPath.empty() ) {
            m_TableColumnXFormPanel->SetRegistryPath(m_RegPath + kColumnXformTag);
            m_TableColumnXFormPanel->LoadSettings();
        }
    }
    return m_TableColumnXFormPanel;
}

bool CTableFileLoadManager::CanDo(EAction action)
{
    if (m_CurrentPage == 0)
        return false;

    if (IsInitialState()) {
        return action == eNext;
    }

    if (action == eBack || action == eNext)
        return true;

    _ASSERT(false);
    return false;
}


bool CTableFileLoadManager::IsFinalState()
{
    return (m_CurrentPage->GetNextPage() == 0);
}


bool CTableFileLoadManager::IsCompletedState()
{
    return (m_CurrentPage == 0);
}


bool CTableFileLoadManager::DoTransition(EAction action)
{
    if (eNext == action) {
        if (m_CurrentPage == 0)
            return false;

        if (!m_CurrentPage->CanLeavePage(true))
            return false;

        m_CurrentPage = m_CurrentPage->GetNextPage();
        return true;
    }
    else if (eBack == action) {
        if (m_CurrentPage == 0) {
            m_CurrentPage = &m_ColumnXForm;
        } else {
            m_CurrentPage = m_CurrentPage->GetPrevPage();
        }
        return true;
    }
    else {
        return false;
    }
    _ASSERT(false);
    return false;
}


IAppTask* CTableFileLoadManager::GetTask()
{
    return 0;
}

IExecuteUnit* CTableFileLoadManager::GetExecuteUnit()
{
    return new CTableObjectLoader(*m_AnnotTableData, m_FileNames);
}

wxString CTableFileLoadManager::GetFormatWildcard()
{
    return CFileExtensions::GetDialogFilter(CFileExtensions::kTable) + wxT("|") +
           CFileExtensions::GetDialogFilter(CFileExtensions::kAllFiles);
}


bool CTableFileLoadManager::ValidateFilenames(const vector<wxString>& filenames)
{
    if (filenames.empty())
        return false;

    CTableFormatPanel* formatPanel = x_GetSelectFormatPanel();
    formatPanel->m_PreviewFname = filenames[0];

    return formatPanel->PreviewData();
}


void CTableFileLoadManager::SetFilenames(const vector<wxString>& filenames)
{
    m_FileNames = filenames;
}

void CTableFileLoadManager::GetFilenames(vector<wxString>& filenames) const
{
    filenames = m_FileNames;
}

bool CTableFileLoadManager::IsInitialState()
{
    return (m_CurrentPage == &m_SelectFormatPage);
}


bool CTableFileLoadManager::RecognizeFormat(const wxString& filename)
{
    wxString ext;
    wxFileName::SplitPath(filename, 0, 0, &ext);
    return CFileExtensions::RecognizeExtension(CFileExtensions::kTable, ext);
}

bool CTableFileLoadManager::RecognizeFormat(CFormatGuess::EFormat fmt)
{
    return fmt == CFormatGuess::eTable;
}

string CTableFileLoadManager::GetExtensionIdentifier() const
{
    static string sid("table_file_load_manager");
    return sid;
}


string CTableFileLoadManager::GetExtensionLabel() const
{
    static string slabel("Table File Load Manager");
    return slabel;
}

END_NCBI_SCOPE
