/*  $Id: ui_data_source_service.cpp 27014 2012-12-07 20:56:18Z katargir $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Andrey Yazhuk
 *
 */

#include <ncbi_pch.hpp>

#include <gui/core/ui_data_source_service.hpp>

#include <gui/core/ui_project_load_manager.hpp>
#include <gui/core/ui_file_load_manager.hpp>
#include <gui/core/ui_gcassembly_load_manager.hpp>
#include <gui/core/data_sources_page.hpp>

#include <gui/framework/memento.hpp>
#include <gui/framework/options_dlg_extension.hpp>
#include <gui/framework/app_task_impl.hpp>
#include <gui/framework/app_task_service.hpp>

#include <gui/utils/extension_impl.hpp>

#include <objects/general/User_object.hpp>

#include <corelib/ncbi_system.hpp>


BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

///////////////////////////////////////////////////////////////////////////////
/// CUIDataSourceTask - a task for opening / closing data sources.
class CUIDataSourceTask : public CAppTask
{
public:
    enum EAction    {
        eInvalidAction = -1,
        eOpen,
        eClose
    };

    CUIDataSourceTask(EAction action, IUIDataSource& ds);

    //virtual bool CanRunOnWorkerThread() { return true; }

protected:
    virtual ETaskState    x_Run();

protected:
    EAction m_Action;
    CIRef<IUIDataSource>    m_DataSource;
};


CUIDataSourceTask::CUIDataSourceTask(EAction action, IUIDataSource& ds)
:   CAppTask("", true),
    m_Action(action),
    m_DataSource(&ds)
{
    m_Descr = "Opening data source " + m_DataSource->GetDescr().GetLabel();
}


CAppTask::ETaskState CUIDataSourceTask::x_Run()
{
    switch(m_Action)    {
    case eOpen:
        m_DataSource->Open();
        return eCompleted;

    case eClose:
        m_DataSource->Close();
        return eCompleted;

    default:
        return eFailed;
    }
}

static const char* kDataSourcesKey = "DataSources";


///////////////////////////////////////////////////////////////////////////////
/// CUIDataSourceService
CUIDataSourceService::CUIDataSourceService()
:   m_SrvLocator(NULL)
{
}


CUIDataSourceService::~CUIDataSourceService()
{
}


void CUIDataSourceService::InitService()
{
    LOG_POST(Info << "Initializing Data Source Service...");

    x_InitDataSourceTypes();

    LoadSettings();

    x_InitDefaultDataSources();

    x_OpenDataSources();

    x_InitLoadManagers();

    // contribute Data Source page to Options dialog
    CIRef<IExtensionRegistry> registry = CExtensionRegistry::GetInstance();
    m_OptionsDlgPage = new CDataSourcesOptionsDlgExtension(this);
    registry->AddExtension(EXT_POINT__OPTIONS_DLG_EXTENSION, *m_OptionsDlgPage);

    LOG_POST(Info << "Finished initializing Data Source Service");
}


void CUIDataSourceService::x_InitDataSourceTypes()
{
    /// get Data Source Types from Extension Registry
    GetExtensionAsInterface("ui_data_source_type", m_DataSourceTypes);

    for( size_t i = 0;  i < m_DataSourceTypes.size();  i++ ) {
        IUIDataSourceType& ds_type = *m_DataSourceTypes[i];
        string name = ds_type.GetDescr().GetLabel();
        LOG_POST(Info << "Registered data source type \"" <<  name << "\"");
    }
}


void CUIDataSourceService::x_InitDefaultDataSources()
{
    for( size_t i = 0;  i < m_DataSourceTypes.size();  i++ ) {
        IUIDataSourceType& ds_type = *m_DataSourceTypes[i];

        if(ds_type.AutoCreateDefaultDataSource())   {
            // look if we have a Data Source of this type already
            bool found = false;
            for( size_t j = 0;  ! found  &&  j < m_DataSources.size();  j++ ) {
                IUIDataSource& ds = *m_DataSources[j];
                found = (&ds.GetType() == &ds_type);
            }

            if( ! found)    {
                LOG_POST(Info << "Creating default Data Source for type \"" <<
                         ds_type.GetDescr().GetLabel()  << "\" ...");

                CIRef<IUIDataSource> ds(ds_type.CreateDataSource());

                /// if this source implements IRegSettings - set its registry path
                IRegSettings* rs = dynamic_cast<IRegSettings*>(&*ds);
                if( rs ){
                    string help_id = ds->GetDescr().GetHelpId();
                    if( help_id.empty() ){
                        help_id = "DefaultUIDS";
                    }
                    string reg_path = m_RegPath + "." + help_id;
                    rs->SetRegistryPath( reg_path );
                }

                Consumer_SetLocator(ds.GetPointer(), m_SrvLocator);
                AddDataSource(*ds);
            }
        }
    }
}


void CUIDataSourceService::x_InitLoadManagers()
{
    /// TODO
    // create System Load Managers
    m_LoadManagers.push_back(TUIToolManagerRef(new CProjectLoadManager()));

    CRef<CFileLoadManager> fileManager(new CFileLoadManager());
    fileManager->LoadRegisteredManagers();
    m_LoadManagers.push_back(TUIToolManagerRef(fileManager.GetPointer()));
    m_LoadManagers.push_back(TUIToolManagerRef(new CGCAssemblyLoadManager()));

    //TODO
    // obtain managers coming from Data Sources
    for( size_t i = 0;  i < m_DataSources.size();  i++ ) {
        CIRef<IUIDataSource> ds = m_DataSources[i];
        CIRef<IUIToolManager> manager(ds->GetLoadManager());
        if(manager) {
            m_LoadManagers.push_back(manager);
        }
    }

    for( size_t i = 0;  i < m_LoadManagers.size();  i++ ) {
        string name = m_LoadManagers[i]->GetDescriptor().GetLabel();
        LOG_POST(Info << "Registered data load manager \"" <<  name << "\"");
    }
}


void CUIDataSourceService::ShutDownService()
{
    SaveSettings();

    if (m_OptionsDlgPage) {
        CIRef<IExtensionRegistry> registry = CExtensionRegistry::GetInstance();
        registry->RemoveExtension(*m_OptionsDlgPage);
        m_OptionsDlgPage = 0;
    }

    x_CloseDataSources();

    m_LoadManagers.clear();
    m_DataSources.clear();
    m_DataSourceTypes.clear();
}


void CUIDataSourceService::SetServiceLocator(IServiceLocator* locator)
{
    m_SrvLocator = locator;
}


void CUIDataSourceService::SetRegistryPath(const string& path)
{
    m_RegPath = path;
}



void CUIDataSourceService::SaveSettings() const
{
    _ASSERT(! m_RegPath.empty());

    if( ! m_RegPath.empty())   {
        CGuiRegistry& gui_reg = CGuiRegistry::GetInstance();
        //CGuiRegistry::TReadWriteView view = gui_reg.GetReadWriteView(m_RegPath);

        // save DataSources
        typedef vector< CRef<CUser_object> >    TObjects;
        TObjects descriptors;

        for( size_t i = 0;  i < m_DataSources.size();  i++ )    {
            const IUIDataSource& ds = *m_DataSources[i];
            CUser_object* ds_descr = x_GetDataSourceDescr(ds);
            if(ds_descr)    {
                descriptors.push_back( CRef<CUser_object>(ds_descr) );
            }
        }

        CRef<CUser_field> ds_field =
            gui_reg.SetField(CGuiRegistryUtil::MakeKey(m_RegPath, kDataSourcesKey));
        ds_field->SetData().SetObjects() = descriptors;
    }
}


void CUIDataSourceService::LoadSettings()
{
    _ASSERT(! m_RegPath.empty());

    if( ! m_RegPath.empty())   {
        CGuiRegistry& gui_reg = CGuiRegistry::GetInstance();
        //CGuiRegistry::TReadView view = gui_reg.GetReadView(m_RegPath);

        typedef vector< CRef<CUser_object> >    TObjects;

        string key = CGuiRegistryUtil::MakeKey(m_RegPath, kDataSourcesKey);
        if(gui_reg.HasField(key))    {
            CConstRef<CUser_field> ds_field = gui_reg.GetField(key);
            const TObjects& descriptors = ds_field->GetData().GetObjects();

            for( size_t i = 0;  i < descriptors.size();  i++   )   {
                const CUser_object& descr = *descriptors[i];
                IUIDataSource* ds = x_GetDataSourceFromDescr(descr);
                if(ds)  {
                    AddDataSource(*ds);
                }
            }
        }
    }
}


static const char* kDSTypeLabel = "DataSourceTypeLabel";
static const char* kDSMemento = "Memento";


CUser_object* CUIDataSourceService::x_GetDataSourceDescr(const IUIDataSource& ds) const
{
    CUser_object* descr = NULL;

    const IMementoOriginator* originator =
        dynamic_cast<const IMementoOriginator*>(&ds);

    if(originator)  {
        auto_ptr<IMemento> memento(originator->GetMemento());
        IUserObjectConvertible* convertible =
                    dynamic_cast<IUserObjectConvertible*>(memento.get());

        if(convertible) {
            CRef<CUser_object> memento_obj(convertible->ToUserObject());
            if(memento_obj) {
                IUIDataSourceType& ds_type = ds.GetType();
                string type_label = ds_type.GetDescr().GetLabel();

                descr = new CUser_object();
                descr->AddField(kDSTypeLabel, type_label);
                descr->AddField(kDSMemento, *memento_obj);
            }
        }
    }
    return descr;
}


/// create an instance of a Data Source from the given descriptor
IUIDataSource* CUIDataSourceService::x_GetDataSourceFromDescr(const CUser_object& descr)
{
    try {
        //
        const CUser_field& f_ext_id = descr.GetField(kDSTypeLabel);
        string type_label = f_ext_id.GetData().GetStr();

        const CUser_field& f_mem = descr.GetField(kDSMemento);
        const CUser_object& mem_obj = f_mem.GetData().GetObject();

        //
        CIRef<IUIDataSourceType> ds_type = GetUIDataSourceTypeByLabel(type_label);
        if(ds_type) {
            IUIDataSource* ds = ds_type->CreateDataSource();
            _ASSERT(ds);

            // initialize the new Data Source with it's saved Memento
            IMementoOriginator* originator = dynamic_cast<IMementoOriginator*>(ds);
            _ASSERT(originator);

            if(originator)  {
                // this data source can accept a Memento
                auto_ptr<IMemento> memento(originator->GetMemento());
                IUserObjectConvertible* convertible =
                    dynamic_cast<IUserObjectConvertible*>(memento.get());

                _ASSERT(convertible);

                if(convertible) {
                    convertible->FromUserObject(mem_obj);
                    originator->SetMemento(*memento);
                    return ds;
                }
            }

            delete ds;
            LOG_POST(Error << "Cannot initialize data source of type \"" <<
                     type_label << "\" configuration invalid.");
        } else {
            LOG_POST(Error << "Data Source Type \"" << type_label<< "\" is not registred.");
        }
    } catch(CException& e)  {
        LOG_POST(Error << "Failed to create Data Source:\n" << e.what());
    }
    return NULL;
}


void CUIDataSourceService::AddDataSource(IUIDataSource& data_source)
{
    CIRef<IUIDataSource> ds_ref(&data_source);
    string ds_name = data_source.GetDescr().GetLabel();

    TUIDataSourceVec::iterator it =
        std::find(m_DataSources.begin(), m_DataSources.end(), ds_ref);

    if(it != m_DataSources.end())   {
        string msg = "Data Source " + ds_name + " is already registered";
        NCBI_THROW(CException, eUnknown, msg);
    } else {
        m_DataSources.push_back(ds_ref);
        LOG_POST(Info << "Registered data source \"" <<  ds_name << "\"");
    }
}


void CUIDataSourceService::GetDataSourceTypes(TDataSourceTypeVec& type_vec)
{
    type_vec = m_DataSourceTypes;
}


CUIDataSourceService::TUIDataSourceTypeRef
    CUIDataSourceService::GetUIDataSourceTypeByLabel(const string& label)
{
    for( size_t i = 0;  i < m_DataSourceTypes.size();  i++) {
        TUIDataSourceTypeRef& ds_type = m_DataSourceTypes[i];
        if(ds_type->GetDescr().GetLabel() == label) {
            return ds_type;
        }
    }
    return TUIDataSourceTypeRef();
}


void CUIDataSourceService::GetDataSources(TUIDataSourceVec& data_sources)
{
    data_sources = m_DataSources;
}


CUIDataSourceService::TUIDataSourceRef
    CUIDataSourceService::GetUIDataSourceByLabel(const string& label)
{
    for( size_t i = 0;  i < m_DataSources.size();  i++) {
        TUIDataSourceRef& ds = m_DataSources[i];
        if(ds->GetDescr().GetLabel() == label) {
            return ds;
        }
    }
    return TUIDataSourceRef();
}


void CUIDataSourceService::GetLoadManagers(TUIToolManagerVec& managers)
{
    managers = m_LoadManagers;
}


static void sReportDSExceptionToLog(const string& prefix,
                                        const string& ds_name, const string& what)
{
    LOG_POST(Error  << prefix << ds_name <<":\n" << what);
}

static const char* kOpenErr = "Failed to open datasource ";
static const char* kCloseErr = "Failed to close datasource ";


void CUIDataSourceService::x_OpenDataSources()
{
    LOG_POST(Info << "Closing Data Sources...");

    for( size_t i = 0;  i < m_DataSources.size();  i++ ) {
        IUIDataSource& ds = *m_DataSources[i];
        string ds_name = ds.GetDescr().GetLabel();

        try {
            _ASSERT(! ds.IsOpen());

            CAppTaskService* task_srv = m_SrvLocator->GetServiceByType<CAppTaskService>();
            CUIDataSourceTask* task = new CUIDataSourceTask(CUIDataSourceTask::eOpen, ds);
            task_srv->AddTask(*task);
        } catch(CException& e) {
            sReportDSExceptionToLog(kOpenErr, ds_name, e.GetMsg());
        }
        catch(std::exception& e) {
            sReportDSExceptionToLog(kOpenErr, ds_name, e.what());
        }
#ifndef _DEBUG
        catch (...) {
            sReportDSExceptionToLog(kOpenErr, ds_name, "Unknown error.");
        }
#endif
    }
    LOG_POST(Info << "Finished closing Data Sources");
}


void CUIDataSourceService::x_CloseDataSources()
{
    LOG_POST(Info << "Closing Data Sources...");

    for( size_t i = m_DataSources.size();  i > 0 ;  ) {
        IUIDataSource& ds = *m_DataSources[--i];
        string ds_name = ds.GetDescr().GetLabel();

        try {
            if(ds.IsOpen()) {
                ds.Close();
            }
        } catch(CException& e) {
            sReportDSExceptionToLog(kCloseErr, ds_name, e.GetMsg());
        }
        catch(std::exception& e) {
            sReportDSExceptionToLog(kCloseErr, ds_name, e.what());
        }
    }

    LOG_POST(Info << "Finished closing Data Sources");
}


END_NCBI_SCOPE
