/*  $Id: ui_gcassembly_load_manager.cpp 26938 2012-11-28 15:46:40Z katargir $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Liangshou Wu
 *
 */

#include <ncbi_pch.hpp>

#include <objmgr/object_manager.hpp>
#include <objmgr/util/sequence.hpp>
#include <objects/seq/Annotdesc.hpp>
#include <objects/genomecoll/GC_AssemblySet.hpp>
#include <objects/genomecoll/GC_AssemblyUnit.hpp>
#include <objects/genomecoll/GC_AssemblyDesc.hpp>
#include <objects/general/Dbtag.hpp>
#include <objects/general/Object_id.hpp>
#include <objects/seqfeat/BioSource.hpp>
#include <objects/seqfeat/Org_ref.hpp>
#include <objects/seq/Seq_descr.hpp>
#include <objects/seq/Seqdesc.hpp>

#include <gui/core/ui_gcassembly_load_manager.hpp>
#include <gui/core/project_service.hpp>
#include <gui/core/loading_app_job.hpp>
#include <gui/framework/service.hpp>
#include <gui/widgets/wx/wx_utils.hpp>
#include <gui/core/assm_info.hpp>
#include <wx/msgdlg.h>


BEGIN_NCBI_SCOPE
USING_SCOPE(objects);



///////////////////////////////////////////////////////////////////////////////
/// CGCAssemblyLoadingJob
class CGCAssemblyLoadingJob : public CDataLoadingAppJob
{
public:
    CGCAssemblyLoadingJob(CProjectService* service, const vector<string>& assm_accs)
        : CDataLoadingAppJob(service)
        , m_AssmAccs(assm_accs)
    {
        CFastMutexGuard lock(m_Mutex);
        m_Descr = "Retrieving Gencoll assemblies: ";
        for (size_t i = 0; i < m_AssmAccs.size(); ++i) {
            if (i > 0) {
                m_Descr += ", ";
            }
            m_Descr += m_AssmAccs[i];
        }
    }

protected:
    virtual void x_CreateProjectItems();

private:

protected:
    vector<string>  m_AssmAccs;
};


void CGCAssemblyLoadingJob::x_CreateProjectItems()
{
    // retrieve the gc-assembly
    ITERATE (vector<string>, iter, m_AssmAccs) {
        CRef<CGC_Assembly> gc_assm = CAssemblyInfo::GetGCAssembly(*iter);
        if (!gc_assm) {
            string msg = "Failed to retrieve gencoll assembly for ";
            msg += *iter;
            NCBI_THROW(CException, eUnknown, msg);
        }

        CRef<CProjectItem> item(new CProjectItem());
        item->SetObject(*gc_assm);
        string label = item->GetLabelByData(m_Scope.GetPointer());
        if (label.empty()) {
            label = *iter;
        }

        item->SetLabel(label);
        string descr = "name:\t\t " + label;
        if (gc_assm->IsAssembly_set()) {
            CGC_AssemblySet::TSet_type type = gc_assm->GetAssembly_set().GetSet_type();
            if (type == CGC_AssemblySet::eSet_type_full_assembly) {
                descr += " (full assembly)";
            } else if (type == CGC_AssemblySet::eSet_type_assembly_set) {
                descr += " (assembly set)";
            }
        }

        list< CRef<CDbtag> > ids;
        CConstRef<CGC_AssemblyDesc> desc;

        if (gc_assm->IsAssembly_set()) {
            desc.Reset(&gc_assm->GetAssembly_set().GetDesc());
            ids = gc_assm->GetAssembly_set().GetId();
        } else if (gc_assm->IsUnit()) {
            desc.Reset(&gc_assm->GetUnit().GetDesc());
            ids = gc_assm->GetUnit().GetId();
        }

        ITERATE (list< CRef<CDbtag> >, iter, ids) {
            if ((*iter)->GetDb() == "GenColl"  &&  (*iter)->GetTag().IsStr()) {
                descr += "\nAccession:\t " + (*iter)->GetTag().GetStr();
            }
            if ((*iter)->GetDb() == "UCSC_name"  &&  (*iter)->GetTag().IsStr()) {
                descr += "\nUCSC name:\t " + (*iter)->GetTag().GetStr();
            }
        }
        descr += "\nRelease type:\t ";
        descr += gc_assm->IsRefSeq() ? "RefSeq" : "Genbank";
        if (desc) {
            if (desc->IsSetLong_name()) {
                descr += "\nDescription:\t " + desc->GetLong_name();
            }
            if (desc->IsSetDescr()) {
                ITERATE (CGC_AssemblyDesc::TDescr::Tdata, it, desc->GetDescr().Get()) {
                    if ((*it)->IsSource()) {
                        descr += "\nOrganism:\t ";
                        if ((*it)->GetSource().IsSetTaxname()) {
                            descr += (*it)->GetSource().GetTaxname();
                        }
                        if ((*it)->GetSource().IsSetCommon()) {
                            descr += " (" + (*it)->GetSource().GetCommon() + ")";
                        }
                    }
                }
            }
        }

        CRef<CAnnotdesc> comment(new CAnnotdesc);
        comment->SetComment(descr);
        item->SetDescr().push_back(comment);

        m_Items.push_back(item);
    }
}


///////////////////////////////////////////////////////////////////////////////
/// CGCAssemblyLoadManager
CGCAssemblyLoadManager::CGCAssemblyLoadManager()
   : m_Descriptor("Genome", "")
   , m_SrvLocator(NULL)
   , m_ParentWindow(NULL)
   , m_State(eInvalid)
   , m_PrjPanel(NULL)
{
    m_ProjectParams.m_EnableDecideLater = false;
}


void CGCAssemblyLoadManager::SetServiceLocator(IServiceLocator* srv_locator)
{
    m_SrvLocator = srv_locator;
}


void CGCAssemblyLoadManager::SetParentWindow(wxWindow* parent)
{
    m_ParentWindow = parent;
    m_AssmWizard.SetParentWindow(parent);
}


const IUIObject& CGCAssemblyLoadManager::GetDescriptor() const
{
    return m_Descriptor;
}


void CGCAssemblyLoadManager::InitUI()
{
    m_State = eSearchAssembly;
    m_AssmWizard.InitUI();
}


void CGCAssemblyLoadManager::CleanUI()
{
    m_State = eInvalid;
    m_AssmWizard.CleanUI();
    m_PrjPanel = NULL;
    m_AssmAccs.clear();
}


wxPanel* CGCAssemblyLoadManager::GetCurrentPanel()
{
    if (m_State == eSearchAssembly) {
        return m_AssmWizard.GetCurrentPanel();
    } else if (m_State == eSelectProject) {
        if(m_PrjPanel == NULL)   {
            CIRef<CProjectService> srv = m_SrvLocator->GetServiceByType<CProjectService>();
            m_PrjPanel = new CProjectSelectorPanel(m_ParentWindow);
            m_PrjPanel->SetProjectService(srv);
            m_PrjPanel->SetParams(m_ProjectParams);
            m_PrjPanel->TransferDataToWindow();
        }
        return m_PrjPanel;
    }
    return NULL;
}


bool CGCAssemblyLoadManager::CanDo(EAction action)
{
    switch(m_State) {
    case eSearchAssembly:
        if (action == eBack) {
            return m_AssmWizard.CanDo(action);
        }
        return true;
    case eSelectProject:
        return action == eBack  ||  action == eNext;
    case eCompleted:
        return false; // nothing left to do
    default:
        _ASSERT(false);
        return false;
    }
}


bool CGCAssemblyLoadManager::IsFinalState()
{
    return m_State == eSelectProject;
}


bool CGCAssemblyLoadManager::IsCompletedState()
{
    return m_State == eCompleted;
}


bool CGCAssemblyLoadManager::DoTransition(EAction action)
{

    if (m_State == eSearchAssembly) {
        if (action == eNext  &&  !m_AssmWizard.CanDo(action)) {
            m_AssmAccs = m_AssmWizard.GetAssmAccs();
            if ( !m_AssmAccs.empty() ) {
                m_State = eSelectProject;
                return true;
            }
            return false;
        } else {
            return m_AssmWizard.DoTransition(action);
        }
    } else if (m_State == eSelectProject) {
        if (action == eBack) {
            m_State = eSearchAssembly;
            return true;
        } else if (action == eNext) {
            if(m_PrjPanel->TransferDataFromWindow()) {
                m_State = eCompleted;
                return true;
            }
            return false;
        }
    }

    _ASSERT(false);
    return false;
}


IAppTask* CGCAssemblyLoadManager::GetTask()
{
    // create loading Job
    m_PrjPanel->GetParams(m_ProjectParams);

    CIRef<CProjectService> srv = m_SrvLocator->GetServiceByType<CProjectService>();
    CGCAssemblyLoadingJob* job = new CGCAssemblyLoadingJob(srv.GetPointer(), m_AssmAccs);
    // create a wrapper task
    auto_ptr<CDataLoadingAppTask> task(new CDataLoadingAppTask(*job));
    task->SetReportErrors(false); // we handle errors ourselves

    CDataLoadingOptions& options = job->GetOptions();
    m_ProjectParams.ToLoadingOptions(options);
    return task.release();
}

static const string kOrgListTag = ".OrgListPanel";
static const string kAssmListTag = ".AssmListPanel";
static const string kProjectParamsTag = "ProjectParams";


void CGCAssemblyLoadManager::SetRegistryPath(const string& path)
{
    m_RegPath = path; // store for later use
    m_AssmWizard.SetRegistryPath(path);
}


void CGCAssemblyLoadManager::SaveSettings() const
{
    if ( !m_RegPath.empty() ) {
        CGuiRegistry& gui_reg = CGuiRegistry::GetInstance();
        CGuiRegistry::TReadWriteView view = gui_reg.GetReadWriteView(m_RegPath);
        m_ProjectParams.SaveSettings(view, kProjectParamsTag);
        m_AssmWizard.SaveSettings();
    }
}


void CGCAssemblyLoadManager::LoadSettings()
{
    if ( !m_RegPath.empty() ) {
        CGuiRegistry& gui_reg = CGuiRegistry::GetInstance();
        CGuiRegistry::TReadView view = gui_reg.GetReadView(m_RegPath);
        m_ProjectParams.LoadSettings(view, kProjectParamsTag);
        m_AssmWizard.LoadSettings();
    }
}


END_NCBI_SCOPE
