/*  $Id $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors: Yury Voronov
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>

#include <gui/core/winmask_files.hpp>

#include <gui/objutils/registry.hpp>

#include <corelib/ncbiapp.hpp>

#include <gui/widgets/wx/wx_utils.hpp>
#include <gui/widgets/wx/async_call.hpp>

#include <wx/filename.h>
#include <wx/dir.h>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

static const char* kLocalMaskStoragePath = "LocalMaskStoragePath";
//static const char* kRemoteMaskStorageURL = "RemoteMaskStorageURL";

CWinMaskerFileStorage::CWinMaskerFileStorage()
    : m_RegPath()
    , m_LocalWinMaskDirPath()
    , m_EnvVarPreferred( false )

{
}

CWinMaskerFileStorage::~CWinMaskerFileStorage()
{
}

CWinMaskerFileStorage& CWinMaskerFileStorage::GetInstance()
{
    static CRef<CWinMaskerFileStorage> s_Storage;
    DEFINE_STATIC_MUTEX(s_StorageMutex);
    if( !s_Storage ){
        CMutexGuard LOCK(s_StorageMutex);
        if( !s_Storage ){
            s_Storage.Reset( new CWinMaskerFileStorage() );
        }
    }
    return *s_Storage;
}

void CWinMaskerFileStorage::SetRegistryPath( const string& path )
{
    m_RegPath = path;
}

void CWinMaskerFileStorage::SaveSettings() const
{
    _ASSERT(!m_RegPath.empty());

    if( !m_RegPath.empty() ){
        CGuiRegistry& gui_reg = CGuiRegistry::GetInstance();
        CGuiRegistry::TReadWriteView view = gui_reg.GetReadWriteView( m_RegPath );

        // save local mask dir path
        //string local_path = CSystemPath::GetStdPath();
        view.Set( kLocalMaskStoragePath, FnToStdString(m_LocalWinMaskDirPath) );
    }
}


void CWinMaskerFileStorage::LoadSettings()
{
    _ASSERT(!m_RegPath.empty());

    if( !m_RegPath.empty() ){
        CGuiRegistry& gui_reg = CGuiRegistry::GetInstance();
        CGuiRegistry::TReadView view = gui_reg.GetReadView( m_RegPath );

        // string local_path = CSystemPath::GetStdPath();
        m_LocalWinMaskDirPath = FnToWxString(view.GetString( kLocalMaskStoragePath ));
    }
}

wxString CWinMaskerFileStorage::GetPath() const
{
    _ASSERT(!m_RegPath.empty());

    wxString dir_path = GetLocalPath();
    dir_path.Trim(false).Trim(true);

    if( dir_path.empty() || m_EnvVarPreferred ){

        dir_path = GetPredefinedPath();
    }

    if( dir_path.empty() ){
        wxString internal_path = wxT("\\\\SNOWMAN\\gbench\\data\\window-masker");
        if( wxFileName::DirExists( internal_path ) ){
            dir_path = internal_path;
        }
    }

    // wxFileName doesn't work with win UNC names wich starts with "//"
#ifdef NCBI_OS_MSWIN
    dir_path.Replace(wxT("/"), wxT("\\"));
#endif

    return dir_path;
}

wxString CWinMaskerFileStorage::GetLocalPath() const
{
    return m_LocalWinMaskDirPath;
}

wxString CWinMaskerFileStorage::GetPredefinedPath() const
{
    CNcbiApplication* app = CNcbiApplication::Instance();
    if( app ){
        CNcbiRegistry& reg = app->GetConfig();

        const string& wm_path = reg.Get( "WINDOW_MASKER", "WINDOW_MASKER_PATH" );
        if( !wm_path.empty() ){
            return ToWxString( wm_path );
        }
    }

    return wxString();
}

void CWinMaskerFileStorage::SetLocalPath( const wxString& path )
{
    //_ASSERT(!path.empty());

    //if( !path.empty() ){
        m_LocalWinMaskDirPath = path;
        m_EnvVarPreferred = false;
    //}
}

/// Functor to retrieve taxonomy labels (used for async execution)
///
struct CGetTaxonLabels_local
{
    vector<int> operator()(wxString dir_path)
    {
        vector<int> tax_ids;

        wxDir dir(dir_path);

        if ( !dir.IsOpened() ) {
            return tax_ids;
        }

        wxString dirName;
        bool cont = dir.GetFirst(&dirName, wxEmptyString, wxDIR_DIRS);
        while (cont) {
            long tax_id;
            if (dirName.ToLong(&tax_id) && tax_id)
                tax_ids.push_back((int)tax_id);

            cont = dir.GetNext(&dirName);
        }

        return tax_ids;
    }
};


void CWinMaskerFileStorage::GetWinMaskerTaxIds( vector<int>& avail_win_mask_ids ) 
{
    avail_win_mask_ids.clear();

    try {
        wxString dir_path = GetPath();
        if( dir_path.empty() )
            return;

        CGetTaxonLabels_local func;
        avail_win_mask_ids = 
            GUI_AsyncExec1<vector<int>, wxString, CGetTaxonLabels_local>(
                                        func, dir_path);
    } catch( CException& ex ){
        ERR_POST( ex.ReportAll() );
        avail_win_mask_ids.clear();
    } catch( std::exception& e ){
        ERR_POST( e.what() );
        avail_win_mask_ids.clear();
    }
}

END_NCBI_SCOPE
