/*  $Id: workspace_folder.cpp 23923 2011-06-22 14:55:46Z katargir $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors: Andrey Yazhuk
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>

#include <gui/core/workspace_folder.hpp>

#include <gui/core/project_service.hpp>

#include <gui/objects/WorkspaceFolder.hpp>
#include <gui/objects/FolderInfo.hpp>
#include <objects/general/Date.hpp>

#include <gui/widgets/wx/wx_utils.hpp>


BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

///////////////////////////////////////////////////////////////////////////////
/// CWorkspaceFolderConstPrx
CWorkspaceFolderConstPrx::CWorkspaceFolderConstPrx(CWorkspaceFolder& folder,
                                                   CWorkspaceConstPrx& workspace)
:   m_Folder(&folder),
    m_WorkspaceConst(workspace)
{
}

CWorkspaceFolderConstPrx::TId CWorkspaceFolderConstPrx::GetId() const
{
    return m_Folder->GetId();
}


std::string CWorkspaceFolderConstPrx::GetName() const
{
    const CFolderInfo& info = m_Folder->GetInfo();
    return info.GetTitle();
}



std::string CWorkspaceFolderConstPrx::GetDescr() const
{
    const CFolderInfo& info = m_Folder->GetInfo();
    return info.GetComment();
}


void CWorkspaceFolderConstPrx::GetChildFolders(TFolderConstVector& folders)
{
    if(m_Folder->CanGetFolders()) {
        CWorkspaceFolder::TFolders& ws_folders = m_Folder->SetFolders();

        NON_CONST_ITERATE(CWorkspaceFolder::TFolders, it, ws_folders)  {
            CRef<CWorkspaceFolderConstPrx> sub_folder(new CWorkspaceFolderConstPrx(**it, m_WorkspaceConst));
            folders.push_back(sub_folder);
        }
    }
}


///////////////////////////////////////////////////////////////////////////////
/// CWorkspaceFolderPrx

CWorkspaceFolderPrx::CWorkspaceFolderPrx(CWorkspaceFolder& folder,
                                         CWorkspacePrx& workspace)
:   CWorkspaceFolderConstPrx(folder, workspace),
    m_Workspace(&workspace)
{
}


void CWorkspaceFolderPrx::SetName(const string& name)
{
    CFolderInfo& info = m_Folder->SetInfo();
    info.SetTitle(name);
    m_Workspace->x_OnWorkspaceFolderChanged(*this);
}


void CWorkspaceFolderPrx::SetDescr(const string& descr)
{
    CFolderInfo& info = m_Folder->SetInfo();
    info.SetComment(descr);
    m_Workspace->x_OnWorkspaceFolderChanged(*this);
}


void CWorkspaceFolderPrx::GetChildFolders(TFolderVector& folders)
{
    if(m_Folder->CanGetFolders()) {
        CWorkspaceFolder::TFolders& ws_folders = m_Folder->SetFolders();

        NON_CONST_ITERATE(CWorkspaceFolder::TFolders, it, ws_folders)  {
            CRef<CWorkspaceFolderPrx> sub_folder(new CWorkspaceFolderPrx(**it, *m_Workspace));
            folders.push_back(sub_folder);
        }
    }
}


void CWorkspaceFolderConstPrx::GetProjects(TProjectVector& projects)
{
    if(m_Folder->CanGetProjects())  {
        CWorkspaceFolder_Base::TProjects& prj_list = m_Folder->SetProjects();

        NON_CONST_ITERATE(CWorkspaceFolder_Base::TProjects, it, prj_list)   {
            CGBDocument* doc = dynamic_cast<CGBDocument*>(it->GetPointer());
            CRef<CProjectPrx> prj(new CProjectPrx(*doc, &m_WorkspaceConst));
            projects.push_back(prj);
        }
    }
}


CRef<CWorkspaceFolderConstPrx>
    CWorkspaceFolderConstPrx::FindChildFolderById_C(TId id, bool recursive)
{
    CWorkspaceFolder* folder = m_Folder->FindChildFolderById(id, recursive);
    if(folder)  {
        return CRef<CWorkspaceFolderConstPrx>(new CWorkspaceFolderConstPrx(*folder, m_WorkspaceConst));
    } else {
        return CRef<CWorkspaceFolderConstPrx>();
    }
}


CRef<CProjectConstPrx>
    CWorkspaceFolderConstPrx::FindProjectById_C(CProjectPrx::TId id,
                                                bool recursive)
{
    CRef<CProjectConstPrx> res;
    CGBProjectHandle* project = m_Folder->FindProjectById(id, recursive);
    CGBDocument* doc = dynamic_cast<CGBDocument*>(project);
    if(doc) {
        res.Reset(new CProjectConstPrx(*doc, &m_WorkspaceConst));
    }
    return res;
}


CRef<CProjectPrx> CWorkspaceFolderConstPrx::FindProjectById(CProjectPrx::TId id,
                                                            bool recursive)
{
    CRef<CProjectPrx> res;
    CGBProjectHandle* project = m_Folder->FindProjectById(id, recursive);
    CGBDocument* doc = dynamic_cast<CGBDocument*>(project);
    if(doc) {
        res.Reset(new CProjectPrx(*doc, &m_WorkspaceConst));
    }
    return res;
}


CRef<CProjectPrx>
    CWorkspaceFolderConstPrx::FindProjectByFilename(const wxString& filename,
                                                    bool /*recursive*/)
{
    CRef<CProjectPrx> res;
    CGBProjectHandle* project = m_Folder->FindProjectByFilename(FnToStdString(filename));
    CGBDocument* doc = dynamic_cast<CGBDocument*>(project);
    if(doc) {
        res.Reset(new CProjectPrx(*doc, &m_WorkspaceConst));
    }
    return res;
}


bool CWorkspaceFolderConstPrx::x_ForEachProject(CWorkspaceFolder::IProjectVisitor& visitor,
                                                bool recursive)
{
    return m_Folder->ForEachProject(visitor, recursive);
}


CRef<CWorkspaceFolderPrx>
    CWorkspaceFolderPrx::FindChildFolderById(TId id, bool recursive)
{
    CWorkspaceFolder* folder = m_Folder->FindChildFolderById(id, recursive);
    if(folder)  {
        return CRef<CWorkspaceFolderPrx>(new CWorkspaceFolderPrx(*folder, *m_Workspace));
    } else {
        return CRef<CWorkspaceFolderPrx>();
    }
}


CRef<CWorkspaceFolderPrx>
    CWorkspaceFolderPrx::FindChildFolderByName(const string& name)
{
    CRef<CWorkspaceFolderPrx> res;
    CWorkspaceFolder* child_folder = m_Folder->FindChildFolderByTitle(name);
    if(child_folder)    {
        res.Reset(new CWorkspaceFolderPrx(*child_folder, *m_Workspace));
    }
    return res;
}


CRef<CWorkspaceFolderPrx> CWorkspaceFolderPrx::AddChildFolder(const string& name,
                                                              const string& descr)
{
    CRef<CWorkspaceFolder> new_folder(new CWorkspaceFolder());

    CFolderInfo& info = new_folder->SetInfo();
    info.SetTitle(name);
    info.SetComment(descr);
    info.SetCreate_date().SetToTime(CTime(CTime::eCurrent));

    m_Folder->AddChildFolder(*new_folder);
    m_Workspace->x_OnWorkspaceFolderChanged(*this);

    CRef<CWorkspaceFolderPrx> res(new CWorkspaceFolderPrx(*new_folder, *m_Workspace));
    return res;
}


void CWorkspaceFolderPrx::RemoveChildFolder(TId child_id)
{
    m_Folder->RemoveChildFolder(child_id);
    m_Workspace->x_OnWorkspaceFolderChanged(*this);
}


void CWorkspaceFolderPrx::x_AddProject(CGBProjectHandle& project)
{
    m_Folder->AddProject(project);
}


END_NCBI_SCOPE
