/*  $Id: diag_console_list.cpp 26026 2012-06-28 20:02:41Z kuznets $
* ===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data, the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties, express or implied, including
*  warranties of performance, merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================
*
* Authors:  Roman Katargin
*
* File Description:
*
*/

#include <ncbi_pch.hpp>

#include <wx/artprov.h>
#include <wx/imaglist.h>

#include <gui/framework/diag_console_list.hpp>

#include <gui/widgets/wx/log_gbench.hpp>

#include <gui/objutils/registry.hpp>

#include <wx/utils.h>

BEGIN_NCBI_SCOPE

BEGIN_EVENT_MAP(CDiagConsoleList, CEventHandler)
    ON_EVENT(CwxLogDiagHandler::CDiagEvent, CwxLogDiagHandler::CDiagEvent::eNewItem,
    &CDiagConsoleList::x_OnNewItem)
END_EVENT_MAP()

CDiagConsoleList::CDiagConsoleList() : m_Filter(0), m_Cache(500)
{
}

CDiagConsoleList::CDiagConsoleList(
                 wxWindow* parent,
                 wxWindowID id,
                 const wxPoint& pos,
                 const wxSize& size,
                 long style)
: m_Filter(eShowAll), m_Cache(500)
{
    Create(parent, id, pos, size, style);
}

CDiagConsoleList::~CDiagConsoleList()
{
    CwxLogDiagHandler::GetInstance()->RemoveListener(this);
}

/// Creation
bool CDiagConsoleList::Create(
            wxWindow* parent,
            wxWindowID id,
            const wxPoint& pos,
            const wxSize& size,
            long style)
{
    if (!wxListCtrl::Create(parent, id, pos, size, style|wxLC_REPORT|wxLC_VIRTUAL))
        return false;

    this->InsertColumn(0, wxT("Message"));
    this->InsertColumn(1, wxT("Time"));
    this->InsertColumn(2, wxT("Source"));

    this->SetColumnWidth(0, 320);
    this->SetColumnWidth(1, 64);
    this->SetColumnWidth(2, 400);

    CwxLogDiagHandler::GetInstance()->AddListener(this);

    m_FixGenericListCtrl.ConnectToControl(*this);

    return true;
}

void CDiagConsoleList::SetFilter(size_t filter)
{
    m_Filter = filter;
    x_ReloadMessages();
}

wxString CDiagConsoleList::GetItemRawMsg(long item) const
{
    size_t index = ((size_t)item >= m_Items.size()) ? 0 : m_Items[item];
    wxString msg = m_Cache.GetItemMessage(index);
    int image = m_Cache.GetItemImage(index);

    wxString retVal;
    switch(image) {
    case 0:
        retVal = wxT("Error: ") + msg;
        break;
    case 1:
        retVal = wxT("Warning: ") + msg;
        break;
    case 2:
        retVal = wxT("Info: ") + msg;
        break;
    default:
        retVal = wxT("Other: ") + msg;
        break;
    }

    return retVal;
}

int CDiagConsoleList::OnGetItemImage(long item) const
{
    size_t index = ((size_t)item >= m_Items.size()) ? 0 : m_Items[item];
    return m_Cache.GetItemImage(index);
}

wxString CDiagConsoleList::OnGetItemText(long item, long column) const
{
    size_t index = ((size_t)item >= m_Items.size()) ? 0 : m_Items[item];
    if (column == 0)
        return m_Cache.GetItemMessage(index);
    else if (column == 1)
        return m_Cache.GetItemTime(index);
    else if (column == 2)
        return m_Cache.GetItemSource(index);
    return wxT("");
}

void CDiagConsoleList::x_ReloadMessages()
{
    wxBusyCursor wait;
    CwxLogDiagHandler& log_handler = *CwxLogDiagHandler::GetInstance();
    size_t i, total, buffered;
    log_handler.GetMsgCount(total, buffered);

    m_Items.clear();

    for (i = total - buffered; i < total; ++i) {
        CwxLogDiagHandler::SMessage msg;
        if (!log_handler.GetMessage(i, msg, total)) {
            m_Items.clear();
            continue;
        }

        int flag = 0;
        switch(msg.severity) {
            case eDiag_Error:
                flag = eShowErrors;
                break;
            case eDiag_Warning:
                flag = eShowWarnings;
                break;
            case eDiag_Info:
                flag = eShowInfo;
                break;
            default:
                flag = eShowOther;
                break;
        }
        flag |= msg.wxMsg ? eShowWxWdidgets : eShowNCBI;

        if ((flag&m_Filter&eShowFlags) && (flag&m_Filter&eShowNCBIWxWdidgets))
            m_Items.push_back(i);
    }

    SetItemCount((long)m_Items.size());
    SetColumnWidth(1, wxLIST_AUTOSIZE);
    if (m_Items.size() > 0)
        EnsureVisible((long)(m_Items.size()-1));
    Refresh();
}

void CDiagConsoleList::x_OnNewItem(CEvent*)
{
    CwxLogDiagHandler& log_handler = *CwxLogDiagHandler::GetInstance();
    size_t i, total, buffered;
    log_handler.GetMsgCount(total, buffered);

    bool update = false;
    if (m_Items.size() > 0) {
        for (i = 0; i < m_Items.size() && m_Items[i] < total - buffered; ++i);
        if (i > 0) {
            m_Items.erase(m_Items.begin(), m_Items.begin() + i);
            update = true;
        }
    }

    i = (m_Items.size() > 0) ? m_Items.back() + 1 : total - buffered;

    for (; i < total; ++i) {
        CwxLogDiagHandler::SMessage msg;
        if (!log_handler.GetMessage(i, msg, total)) {
            m_Items.clear();
            update = true;
            continue;
        }

        int flag = 0;
        switch(msg.severity) {
            case eDiag_Error:
                flag = eShowErrors;
                break;
            case eDiag_Warning:
                flag = eShowWarnings;
                break;
            case eDiag_Info:
                flag = eShowInfo;
                break;
            default:
                flag = eShowOther;
                break;
        }
        flag |= msg.wxMsg ? eShowWxWdidgets : eShowNCBI;

        if ((flag&m_Filter&eShowFlags) && (flag&m_Filter&eShowNCBIWxWdidgets)) {
            m_Items.push_back(i);
            update = true;
        }
    }

    if (update) {
        SetItemCount((long)m_Items.size());
        SetColumnWidth(1, wxLIST_AUTOSIZE);
        if (m_Items.size() > 0)
            EnsureVisible((long)(m_Items.size()-1));
        Refresh();
    }
}

static const char* kFilterTag = "Filter";

void CDiagConsoleList::LoadSettings()
{
    if( ! m_RegPath.empty())   {
        CGuiRegistry& gui_reg = CGuiRegistry::GetInstance();
        CGuiRegistry::TReadView view = gui_reg.GetReadView(m_RegPath);

        m_Filter = view.GetInt(kFilterTag, (int)m_Filter);
        x_ReloadMessages();
    }
}

void CDiagConsoleList::SaveSettings() const
{
    if( ! m_RegPath.empty())   {
        CGuiRegistry& gui_reg = CGuiRegistry::GetInstance();
        CGuiRegistry::TReadWriteView view = gui_reg.GetReadWriteView(m_RegPath);

        view.Set(kFilterTag, (int)m_Filter);
    }
}

///////////////////////////////////////////////////////////////////////////////
/// CDiagConsoleCache

int CDiagConsoleCache::GetItemImage(size_t index)
{
    return x_GetItem(index).m_Image;
}

wxString CDiagConsoleCache::GetItemMessage(size_t index)
{
    return x_GetItem(index).m_Message;
}

wxString CDiagConsoleCache::GetItemSource(size_t index)
{
    return x_GetItem(index).m_Source;
}

wxString CDiagConsoleCache::GetItemTime(size_t index)
{
    return x_GetItem(index).m_Time;
}

const CDiagConsoleCache::CItem& CDiagConsoleCache::x_GetItem(size_t index)
{
    if (++m_AccessTime == numeric_limits<unsigned long>::max()) {
        m_Data.clear();
        m_AccessTime = 0;
    }

    TData::iterator it = m_Data.find(index);
    if (it != m_Data.end()) {
        it->second.m_AccessTime = m_AccessTime;
        return it->second;
    }

    if (m_Size == m_Data.size()) {
        TData::iterator it_erase = it = m_Data.begin();
        for (++it;it != m_Data.end(); ++it) {
            if (it->second.m_AccessTime < it_erase->second.m_AccessTime)
                it_erase = it;
        }
        m_Data.erase(it_erase);
    }

    CwxLogDiagHandler& log_handler = *CwxLogDiagHandler::GetInstance();
    CwxLogDiagHandler::SMessage msg;
    size_t total;
    CItem item;

    if (log_handler.GetMessage(index, msg, total)) {
        wxString text = wxString::FromUTF8(msg.message.c_str());
        text.Replace(wxT("\n"), wxT(" "));

        size_t i, indent = 0;
        for (i = 0; i < text.Length(); ++i) {
            if (text[i] == '\t')
                indent += 8;
            else if (text[i] == ' ')
                ++indent;
            else
                break;
        }
        if (i > 0) {
            text = text.Mid(i);
            text.Pad(indent, ' ', false);
        }

        item.m_Message = text;
        item.m_Time = ToWxString(msg.time.AsString("h:m:s"));
        item.m_Source = ToWxString(msg.source); 

        switch(msg.severity) {
        case eDiag_Error:
            item.m_Image = 0;
            break;
        case eDiag_Warning:
            item.m_Image = 1;
            break;
        case eDiag_Info:
            item.m_Image = 2;
            break;
        default:
            item.m_Image = 3;
            break;
        }
    }
    else {
        item.m_Message = wxT("Wrong item requested (problem with the console view).");
        item.m_Image = -1;
    }

    return m_Data[index] = item;
}

END_NCBI_SCOPE

