/*  $Id: observable.cpp 24494 2011-10-03 19:35:22Z kuznets $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Andrey Yazhuk
 *
 */

#include <ncbi_pch.hpp>

#include <gui/framework/observable.hpp>

BEGIN_NCBI_SCOPE

//////////////////////////////////////////////////////////////////////////////
/// IObservable interface
/*
CObserverUpdate::CObserverUpdate(int type, IObservable* observable)
:   m_Type(type),
    m_Observable(observable)
{
}


int CObserverUpdate::GetType()
{
    return m_Type;
}
*/

IObservable* CObserverUpdate::GetObservable()
{
    return m_Observable;
}


//////////////////////////////////////////////////////////////////////////////
/// CObservable

BEGIN_EVENT_MAP(CObservable, CEventHandler)
    ON_MESSAGE( CObservableUpdateEvent, 0, &CObservable::x_OnUpdateEvent )
END_EVENT_MAP();


CObservable::CObservable()
:   m_UpdateLockCounter(0),
    m_StateChanged(false)
{
}


CObservable::~CObservable()
{
}


void CObservable::GetObservers(TObservers& observers)
{
    TGuard guard(*this);
    observers = m_Observers;
}


bool CObservable::AttachObserver(IObserver* observer)
{
    TGuard guard(*this);

    _ASSERT(observer);

    if(observer)    {
        // search if this is a known observer
        //
        TObservers::const_iterator it = m_Observers.begin();
        TObservers::const_iterator it_end = m_Observers.end();
        for(; it != it_end; ++it) {
            CIRef<IObserver> oref(it->Lock());
            IObserver* obs = oref.GetNCPointerOrNull();
            if (obs == observer) 
                break;
        } // for


        if(it == it_end) {       // attaching observer
            m_Observers.push_back(TObserverWeakPtr(observer));

            CObserverUpdate up(CObserverUpdate::eAttached, this);
            observer->OnUpdate(up); // notify observer
        }
    }
    return false;
}


bool CObservable::DetachObserver(IObserver* observer)
{
    TGuard guard(*this);

    _ASSERT(observer);

    if(observer)    {
        // search if this is a known observer
        //
        TObservers::iterator it = m_Observers.begin();
        TObservers::iterator it_end = m_Observers.end();
        for(; it != it_end; ++it) {
            CIRef<IObserver> oref(it->Lock());
            IObserver* obs = oref.GetNCPointerOrNull();
            if (obs == observer) 
                break;
        } // for

        if(it != it_end) { // detaching observer
            CObserverUpdate up(CObserverUpdate::ePreDetach, this);
            observer->OnUpdate(up); // notify observer

            m_Observers.erase(it);
            return true;
        }
    }
    return false;
}


void CObservable::Lock() const
{
    m_MainMutex.Lock();
}


void CObservable::Unlock() const
{
    m_MainMutex.Unlock();
}


void CObservable::LockUpdates()
{
    //LOG_POST("");
    //LOG_POST("CObservable::LockUpdates()");

    CMutexGuard mtx_guard(m_UpdateMutex); // lock m_UpdateLockCounter

    if(m_UpdateLockCounter == 0) {
        m_StateChanged = false; // reset state
    }
    ++m_UpdateLockCounter; // increment counter
}


void CObservable::UnlockUpdates()
{
    //LOG_POST("");
    //LOG_POST("CObservable::LockUpdates()  BEGIN");

    CMutexGuard mtx_guard(m_UpdateMutex); // lock m_UpdateLockCounter

    _ASSERT(m_UpdateLockCounter >= 0);

    --m_UpdateLockCounter;

    if( m_UpdateLockCounter == 0 && m_StateChanged ){
        // flush accumulated updates
        x_PostUpdateEvent();
        m_StateChanged = false;
    }

    //LOG_POST("CObservable::LockUpdates()  END");
}


void CObservable::DetachAllObservers()
{
    CObserverUpdate up(CObserverUpdate::ePreDetach, this);

    TGuard guard(*this);

    x_UpdateObserversNow(up);
    m_Observers.clear();
}


void CObservable::x_NotifyObservers(CObserverUpdate::EUpdateType type)
{
    if( type == CObserverUpdate::eAttached || type == CObserverUpdate::ePreDetach ){
        _ASSERT(false); // should not be used for these
    }

    x_NotifyObservers();
}

void CObservable::x_NotifyObservers() 
{
    //LOG_POST("CObservable::x_NotifyObservers() BEGIN");

    // this function does not lock main mutex because it does not access
    // observers directly; locking may result in a deadlock

    // see if this update shall be postponed
    CMutexGuard mtx_guard(m_UpdateMutex);   // now it is safe to check the counter

    if( m_UpdateLockCounter == 0 ){
        x_PostUpdateEvent();
    } else {
        m_StateChanged = true; // mark as changed for future
    }
    //LOG_POST("CObservable::x_NotifyObservers() END");
}

IObserverUpdate* CObservable::x_CreateStateChange()
{
    return new CObserverUpdate();
}

void CObservable::x_PostUpdateEvent()
{
    IObserverUpdate* update = x_CreateStateChange();
    _ASSERT(update);

    CObserverUpdate* update_object = dynamic_cast<CObserverUpdate*>(update);
    if( update_object ){
        update_object->Init( CObserverUpdate::eStateChanged, this ); 
    }

    CRef<CEvent> event( new CObservableUpdateEvent(*update) );
    Post(event); // send asynchronously to myself
}

void CObservable::x_NotifyObserversNow()
{
    IObserverUpdate* update = x_CreateStateChange();
    _ASSERT(update);

    CObserverUpdate* update_object = dynamic_cast<CObserverUpdate*>(update);
    if( update_object ){
        update_object->Init( CObserverUpdate::eStateChanged, this ); 
    }

    CIRef<IObserverUpdate> update_ref( update );

    x_UpdateObserversNow( *update );
}

/// instantly delivers the update to all observers
/// x_UpdateObserversNow() should be guarded to make sure the list of observers
/// does not change
void CObservable::x_UpdateObserversNow( IObserverUpdate& update )
{
    //LOG_POST("CObservable::x_UpdateObserversNow()");
    NON_CONST_ITERATE( TObservers, it, m_Observers ){
        CIRef<IObserver> oref(it->Lock());
        if (oref.NotNull()) {
            //IObserver* observer = *it;
            oref->OnUpdate( update );
        }
    }
}


// handles CObservableUpdateEvent; sends it to itself asynchronously
void CObservable::x_OnUpdateEvent(CEvent* event)
{
    //LOG_POST("CObservable::OnUpdateEvent()");
    CObservableUpdateEvent* up_evt = dynamic_cast<CObservableUpdateEvent*>(event);
    _ASSERT(up_evt);

    if( up_evt ){
        CIRef<IObserverUpdate> update = up_evt->GetUpdate();

        _ASSERT(update);

        if(update)  {
            TGuard lock(*this);
            x_UpdateObserversNow(*update);
        }
    }
}


END_NCBI_SCOPE
