/*  $Id: bookmark.cpp 22039 2010-09-08 14:14:26Z kuznets $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Mike DiCuccio
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>
#include <gui/objutils/bookmark.hpp>
#include <objects/seqloc/Seq_point.hpp>
#include <objects/seqloc/Seq_interval.hpp>
#include <objects/seq/seq_id_handle.hpp>
#include <objects/general/Object_id.hpp>
#include <corelib/ncbimtx.hpp>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

CBookmark::CBookmark()
{
}


CBookmark::CBookmark(const CSeq_id& id, TSeqPos pos)
{
    SetLocation(id, pos);
}


CBookmark::CBookmark(const CSeq_id& id, TSeqPos from, TSeqPos to)
{
    SetLocation(id, from, to);
}


void CBookmark::Save(CUser_field& field) const
{
    field.SetLabel().SetStr("BookmarkData");
    field.AddField("from",  (int)m_Range.GetFrom());
    field.AddField("to",    (int)m_Range.GetTo());
    field.AddField("label", m_Label);

    {{
        string str;
        m_SeqId->GetLabel(&str);
        field.AddField("id", str);
    }}

    if (m_Data) {
        CRef<CUser_field> f = field.SetFieldRef("Data");
        f->Assign(*m_Data);
    }
}


void CBookmark::Load(const CUser_field& field)
{
    m_Range.SetFrom(field.GetField("from").GetData().GetInt());
    m_Range.SetTo  (field.GetField("to").GetData().GetInt());
    m_Label = field.GetField("label").GetData().GetStr();
    m_SeqId.Reset(new CSeq_id(field.GetField("id").GetData().GetStr()));

    CConstRef<CUser_field> f = field.GetFieldRef("Data");
    if (f) {
        m_Data.Reset(new CUser_field());
        m_Data->Assign(*f);
    } else {
        m_Data.Reset();
    }
}


TSeqRange CBookmark::GetRange() const
{
    return m_Range;
}


const string& CBookmark::GetLabel() const
{
    return m_Label;
}


const CSeq_id& CBookmark::GetId() const
{
    return *m_SeqId;
}


const CUser_field& CBookmark::GetData() const
{
    return *m_Data;
}


CUser_field& CBookmark::SetData()
{
    if ( !m_Data ) {
        m_Data.Reset(new CUser_field());
    }
    return *m_Data;
}


void CBookmark::SetLabel(const string& str)
{
    m_Label = str;
}


void CBookmark::SetLocation(const CSeq_id& id, TSeqPos pos)
{
    m_SeqId.Reset(new CSeq_id());
    m_SeqId->Assign(id);
    m_Range.SetFrom(pos);
    m_Range.SetTo  (pos);
}


void CBookmark::SetLocation(const CSeq_id& id, TSeqPos from, TSeqPos to)
{
    m_SeqId.Reset(new CSeq_id());
    m_SeqId->Assign(id);
    m_Range.SetFrom(from);
    m_Range.SetTo  (to);
}


void CBookmarkManager::LoadSettings(CGuiRegistry::TReadView view)
{
    CConstRef<CUser_field> field = view.GetField("Data");
    if ( !field  ||  !field->GetData().IsFields() ) {
        return;
    }

    TBookmarkMap new_map;
    ITERATE (CUser_field::TData::TFields, iter, field->GetData().GetFields()) {
        try {
            CRef<CBookmark> bm(new CBookmark());
            bm->Load(**iter);
            CSeq_id_Handle idh = CSeq_id_Handle::GetHandle(bm->GetId());
            new_map[idh].push_back(bm);
        }
        catch (std::exception&) {
        }
    }

    m_Bookmarks.swap(new_map);
}


void CBookmarkManager::SaveSettings(CGuiRegistry::TReadWriteView view) const
{
    CRef<CUser_field> field;
    ITERATE (TBookmarkMap, iter, m_Bookmarks) {
        ITERATE (TBookmarks, it, iter->second) {
            CRef<CUser_field> bm(new CUser_field);
            (*it)->Save(*bm);

            if ( !field ) {
                field = view.SetField("Data");
            }
            field->SetData().SetFields().push_back(bm);
        }
    }
}


CBookmarkManager::TBookmarks CBookmarkManager::GetAllBookmarks() const
{
    TBookmarks bookmarks;

    ITERATE(TBookmarkMap, iter, m_Bookmarks) {
        ITERATE (TBookmarks, it, iter->second) {
            bookmarks.push_back(*it);
        }
    }

    return bookmarks;
}


CBookmarkManager::TBookmarks
CBookmarkManager::GetBookmarks(const CSeq_id& id,
                               TSeqPos from, TSeqPos to) const
{
    TBookmarks bookmarks;

    CSeq_id_Handle idh = CSeq_id_Handle::GetHandle(id);

    if (from > to)
        swap(from, to);

    TSeqRange range(from, to);
    TBookmarkMap::const_iterator iter = m_Bookmarks.find(idh);
    if (iter != m_Bookmarks.end()) {
        ITERATE (TBookmarks, it, iter->second) {
            if (range.IntersectingWith((*it)->GetRange())) {
                bookmarks.push_back(*it);
            }
        }
    }

    return bookmarks;
}


/// add a bookmark to the manager
CRef<CBookmark> CBookmarkManager::AddBookmark(const CSeq_id& id,
                                              const string& label,
                                              TSeqPos pos)
{
    return AddBookmark(id, label, pos, pos);
}


CRef<CBookmark> CBookmarkManager::AddBookmark(const CSeq_id& id,
                                              const string& label,
                                              TSeqPos from, TSeqPos to)
{
    CRef<CBookmark> bookmark(new CBookmark(id, from, to));
    bookmark->SetLabel(label);
    m_Bookmarks[CSeq_id_Handle::GetHandle(id)].push_back(bookmark);
    return bookmark;
}


bool CBookmarkManager::RemoveBookmark(CBookmark& bm)
{
    CSeq_id_Handle idh = CSeq_id_Handle::GetHandle(bm.GetId());
    TBookmarkMap::iterator iter = m_Bookmarks.find(idh);
    if (iter == m_Bookmarks.end()) {
        return false;
    }

    CRef<CBookmark> ref(&bm);
    bool ret_val = false;
    TBookmarks::iterator it  = iter->second.begin();
    TBookmarks::iterator end = iter->second.end();
    for ( ;  it != end;  ) {
        if (*it == ref) {
            iter->second.erase(it++);
            ret_val = true;
        } else {
            ++it;
        }
    }
    return ret_val;
}


CBookmarkManager& CBookmarkManager::GetInstance()
{
    static CRef<CBookmarkManager> s_BookmarkMgr;
    DEFINE_STATIC_MUTEX(s_Mutex);
    if ( !s_BookmarkMgr ) {
        CMutexGuard LOCK(s_Mutex);
        if ( !s_BookmarkMgr ) {
            s_BookmarkMgr.Reset(new CBookmarkManager());
        }
    }

    return *s_BookmarkMgr;
}





END_NCBI_SCOPE
