/*  $Id: cmd_add_cds.cpp 27180 2013-01-07 15:58:11Z bollin $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data,  the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties,  express or implied,  including
 *  warranties of performance,  merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Colleen Bollin, based on a file by Roman Katargin
 */


#include <ncbi_pch.hpp>

#include <objmgr/seq_entry_ci.hpp>
#include <objmgr/seq_annot_ci.hpp>
#include <objmgr/bioseq_ci.hpp>
#include <objmgr/seqdesc_ci.hpp>
#include <objmgr/util/sequence.hpp>
#include <objects/seqloc/Seq_loc.hpp>
#include <objects/seqloc/Seq_interval.hpp>
#include <objects/seqloc/Seq_id.hpp>
#include <objects/general/Object_id.hpp>

#include <gui/objutils/cmd_add_cds.hpp>
#include <gui/objutils/cmd_add_seqentry.hpp>


BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

void CCmdAddCDS::Execute()
{
    CSeq_entry_EditHandle eh = m_seh.GetEditHandle();
    if (!eh.IsSet()) {
        eh.ConvertSeqToSet();
        eh.SetSet().SetClass(objects::CBioseq_set::eClass_nuc_prot);
        CBioseq_CI bi(eh, objects::CSeq_inst::eMol_na);
        CBioseq_EditHandle nuc_h = bi->GetEditHandle();

        // move all source and pub descriptors from Bioseq to Nuc-prot set
        CSeq_descr::Tdata& seq_descr = nuc_h.SetDescr();
        CSeq_descr::Tdata::iterator desc_it = seq_descr.begin();
        while (desc_it != seq_descr.end()) {
            if ((*desc_it)->IsSource() || (*desc_it)->IsPub()) {
                CRef<objects::CSeqdesc> cpy(new objects::CSeqdesc());
                cpy->Assign(**desc_it);
                eh.AddSeqdesc(*cpy);
                desc_it = seq_descr.erase(desc_it);
            } else {
                ++desc_it;
            }
        }
    }
              

    CSeq_annot_Handle ftable;

    CSeq_annot_CI annot_ci(m_seh, CSeq_annot_CI::eSearch_entry);
    for (; annot_ci; ++annot_ci) {
        if ((*annot_ci).IsFtable()) {
            ftable = *annot_ci;
            break;
        }
    }

    if (!ftable) {
        CRef<CSeq_annot> new_annot(new CSeq_annot());
        ftable = m_FTableCreated = eh.AttachAnnot(*new_annot);
    }

    CSeq_annot_EditHandle aeh(ftable);
    m_feh_CDS = aeh.AddFeat(*m_CDS);

    // add protein sequence that is translation
    CRef<objects::CBioseq> protein = CSeqTranslator::TranslateToProtein (*m_CDS, eh.GetScope());
    CRef<objects::CSeq_id> prot_id(new objects::CSeq_id());
    if (m_CDS->IsSetProduct()) {
        prot_id->Assign(*(m_CDS->GetProduct().GetId()));
    } else {
        // make up some new protein ID
        CBioseq_Handle nuc_h = m_seh.GetScope().GetBioseqHandle(m_CDS->GetLocation());
        prot_id->Assign(*GetNewProteinId(m_seh, nuc_h));
    }
    protein->SetId().push_back(prot_id);
    m_ProtHandle = eh.AttachBioseq(*protein, -1);

    // create molinfo descriptor
    CRef<objects::CSeqdesc> pdesc(new objects::CSeqdesc());
    pdesc->SetMolinfo().SetBiomol(objects::CMolInfo::eBiomol_peptide);
    pdesc->SetMolinfo().SetCompleteness(objects::CMolInfo::eCompleteness_complete);

    // set location for protein feature
    CRef<objects::CSeq_feat> prot_feat(new CSeq_feat());
    prot_feat->Assign(*m_Prot);
    prot_feat->SetLocation().SetInt().SetId(*prot_id);
    prot_feat->SetLocation().SetInt().SetFrom(0);
    prot_feat->SetLocation().SetInt().SetTo(protein->GetLength() - 1);
    if (m_CDS->GetLocation().IsPartialStart(objects::eExtreme_Biological)) {
        prot_feat->SetLocation().SetPartialStart(true, objects::eExtreme_Biological);
        prot_feat->SetPartial(true);
        pdesc->SetMolinfo().SetCompleteness(objects::CMolInfo::eCompleteness_no_left);
    }
    if (m_CDS->GetLocation().IsPartialStop(objects::eExtreme_Biological)) {
        prot_feat->SetLocation().SetPartialStop(true, objects::eExtreme_Biological);
        prot_feat->SetPartial(true);
        if (pdesc->GetMolinfo().GetCompleteness() == objects::CMolInfo::eCompleteness_complete) {
            pdesc->SetMolinfo().SetCompleteness(objects::CMolInfo::eCompleteness_no_right);
        } else {
            pdesc->SetMolinfo().SetCompleteness(objects::CMolInfo::eCompleteness_no_ends);
        }
    }

    CRef<CSeq_annot> new_prot_annot(new CSeq_annot());
    CSeq_annot_Handle prot_ftable = m_ProtHandle.AttachAnnot(*new_prot_annot);
    CSeq_annot_EditHandle paeh(prot_ftable);
    paeh.AddFeat(*prot_feat);

    // add molinfo descriptor
    m_ProtHandle.AddSeqdesc(*pdesc);

}

void CCmdAddCDS::Unexecute()
{
    m_ProtHandle.Remove();
    CSeq_entry_EditHandle eh = m_seh.GetEditHandle();
    if (eh.IsSet() && eh.GetSet().GetCompleteBioseq_set()->GetSeq_set().size() == 1) {
        eh.ConvertSetToSeq();    
    }      

    m_feh_CDS.Remove();

    if (m_FTableCreated) {
        m_FTableCreated.Remove();
    }
}

string CCmdAddCDS::GetLabel()
{
    return "Convert nuc to nuc-prot set";
}



END_NCBI_SCOPE
