/*  $Id: cmd_propagate_feat.cpp 28123 2013-05-13 15:43:15Z ludwigf $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data,  the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties,  express or implied,  including
 *  warranties of performance,  merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Frank Ludwig, adapting code by Roman Katargin
 */


#include <ncbi_pch.hpp>

#include <objects/seqfeat/Cdregion.hpp>
#include <objects/seqfeat/Genetic_code_table.hpp>
#include <objects/general/Object_id.hpp>

#include <objtools/readers/source_mod_parser.hpp>

#include <objmgr/seqdesc_ci.hpp>
#include <objmgr/seq_entry_ci.hpp>
#include <objmgr/feat_ci.hpp>
#include <objmgr/seq_annot_ci.hpp>
#include <objects/seq/Bioseq.hpp>
#include <objmgr/util/sequence.hpp>
#include <objmgr/seq_vector.hpp>
#include <objmgr/seq_vector_ci.hpp>

#include <gui/objutils/cmd_propagate_feat.hpp>
#include <gui/objutils/cmd_add_seqentry.hpp>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);


static CRef<CSeq_loc> sTruncateToStopCodon(const CSeq_loc& loc, size_t truncLen)
{
    CRef<CSeq_loc> pTruncated;

    size_t curLen = 0;
    for (CSeq_loc_CI it(loc); it && curLen < truncLen; ++it) {
        size_t thisLen = it.GetRange().GetLength();
        CConstRef<CSeq_loc> pThisLoc = it.GetRangeAsSeq_loc();
        if (curLen + thisLen <= truncLen) {
            if (pTruncated) {
                pTruncated->Add(*pThisLoc);
            } else {
                pTruncated.Reset(new CSeq_loc());
                pTruncated->Assign(*pThisLoc);
            }
            curLen += thisLen;
        } else {
            CRef<CSeq_loc> pPartialLoc(new CSeq_loc());
            size_t missingLen = truncLen - curLen;
            size_t start = pThisLoc->GetStart(eExtreme_Biological);
            if (missingLen == 1) {
                // make a point
                pPartialLoc->SetPnt().SetPoint(start);
            } else {
                // make an interval
                if (pThisLoc->IsSetStrand() && pThisLoc->GetStrand() == eNa_strand_minus) {
                    pPartialLoc->SetInt().SetFrom(start - missingLen + 1);
                    pPartialLoc->SetInt().SetTo(start);
                } else {
                    pPartialLoc->SetInt().SetFrom(start);
                    pPartialLoc->SetInt().SetTo(start + missingLen - 1);
                }
            }
            pPartialLoc->SetId(*pThisLoc->GetId());
            if (pThisLoc->IsSetStrand()) {
                pPartialLoc->SetStrand(pThisLoc->GetStrand());
            }
            if (pTruncated) {
                pTruncated->Add(*pPartialLoc);
            } else {
                pTruncated.Reset(new CSeq_loc());
                pTruncated->Assign(*pPartialLoc);
            }
            curLen += missingLen;  
        }
    }

    return pTruncated;
}

static CRef<CSeq_loc> sExtendToStopCodon (CRef<CSeq_feat> feat, CScope& scope)
{
    CRef<CSeq_loc> new_loc;

    if (!feat || !feat->IsSetLocation()) {
        return new_loc;
    }
    const CSeq_loc& loc = feat->GetLocation();

    CBioseq_Handle bsh = scope.GetBioseqHandle(*(loc.GetId()));
    if (!bsh) {
        return new_loc;
    }

    const CGenetic_code* code = NULL;
    if (feat->IsSetData() && feat->GetData().IsCdregion() && feat->GetData().GetCdregion().IsSetCode()) {
        code = &(feat->GetData().GetCdregion().GetCode());
    }

    size_t stop = loc.GetStop(eExtreme_Biological);
    // figure out if we have a partial codon at the end
    size_t orig_len = sequence::GetLength(loc, &scope);
    size_t len = orig_len;
    if (feat->IsSetData() && feat->GetData().IsCdregion() && feat->GetData().GetCdregion().IsSetFrame()) {
        CCdregion::EFrame frame = feat->GetData().GetCdregion().GetFrame();
        if (frame == CCdregion::eFrame_two) {
            len -= 1;
        } else if (frame == CCdregion::eFrame_three) {
            len -= 2;
        }
    }
    size_t mod = len % 3;
    CRef<CSeq_loc> vector_loc(new CSeq_loc());
    vector_loc->SetInt().SetId().Assign(*(loc.GetId()));

    if (loc.IsSetStrand() && loc.GetStrand() == eNa_strand_minus) {
        vector_loc->SetInt().SetFrom(0);
        vector_loc->SetInt().SetTo(stop + mod - 1);
        vector_loc->SetStrand(eNa_strand_minus);
    } else {
        vector_loc->SetInt().SetFrom(stop - mod + 1);
        vector_loc->SetInt().SetTo(bsh.GetInst_Length() - 1);
    }

    CSeqVector seq(*vector_loc, scope, CBioseq_Handle::eCoding_Iupac);
    // reserve our space
    const size_t usable_size = seq.size();

    // get appropriate translation table
    const CTrans_table & tbl =
        (code ? CGen_code_table::GetTransTable(*code) :
                CGen_code_table::GetTransTable(1));

    // main loop through bases
    CSeqVector::const_iterator start = seq.begin();

    size_t i;
    size_t k;
    size_t state = 0;
    size_t length = usable_size / 3;

    for (i = 0;  i < length;  ++i) {
        // loop through one codon at a time
        for (k = 0;  k < 3;  ++k, ++start) {
            state = tbl.NextCodonState(state, *start);
        }

        if (tbl.GetCodonResidue (state) == '*') {
            CSeq_loc_CI it(loc);
            CSeq_loc_CI it_next = it;
            ++it_next;
            while (it_next) {
                CConstRef<CSeq_loc> this_loc = it.GetRangeAsSeq_loc();
                if (new_loc) {
                    new_loc->Add(*this_loc);
                } else {
                    new_loc.Reset(new CSeq_loc());
                    new_loc->Assign(*this_loc);
                }
                it = it_next;
                ++it_next;
            }
            CRef<CSeq_loc> last_interval(new CSeq_loc());
            CConstRef<CSeq_loc> this_loc = it.GetRangeAsSeq_loc();
            size_t this_start = this_loc->GetStart(eExtreme_Positional);
            size_t this_stop = this_loc->GetStop(eExtreme_Positional);
            size_t extension = ((i + 1) * 3) - mod;
            last_interval->SetInt().SetId().Assign(*(this_loc->GetId()));
            if (this_loc->IsSetStrand() && this_loc->GetStrand() == eNa_strand_minus) {
                last_interval->SetStrand(eNa_strand_minus);
                last_interval->SetInt().SetFrom(this_start - extension);
                last_interval->SetInt().SetTo(this_stop);
            } else {
                last_interval->SetInt().SetFrom(this_start);
                last_interval->SetInt().SetTo(this_stop + extension);
            }
                
            if (new_loc) {
                new_loc->Add(*last_interval);
            } else {
                new_loc.Reset(new CSeq_loc());
                new_loc->Assign(*last_interval);
            }

            return new_loc;
        }
    }

    return new_loc;
}

void CCmdPropagateFeat::Execute()
{
    switch(m_feat->GetData().GetSubtype()) {
    default:
        xPropagateDefault();
        break;
    case CSeqFeatData::eSubtype_cdregion:
        xPropagateCds();
        break;
    }
}

void CCmdPropagateFeat::Unexecute()
{
    //two cases need to be handled:
    // if propagation resulted in a single new feature then that new feature
    //  needs to be squished
    // if propagation resulted in an entire nuc-prot set then the entire
    //  nuc-prot set needs to be removed
    if (m_hAddedFeat) {
        m_hAddedFeat.Remove();
    }
    if (m_hAddedTable) {
        m_hAddedTable.Remove();
    }
    if (m_hAddedProtein) {
        CSeq_entry_EditHandle nucprot = m_hAddedProtein.GetParentEntry();
        m_hAddedProtein.Remove();
        if (mCreatedNucProtSet) {
            nucprot.ConvertSetToSeq();
        }
    }
}

string CCmdPropagateFeat::GetLabel()
{
    return "Propagate Feature";
}

void CCmdPropagateFeat::xPropagateDefault()
{
    CSeq_annot_Handle th = xDefaultGetAppropriateFeatureTable();
    CSeq_annot_EditHandle aeh(th);
    CRef<CSeq_feat> pPropagatedFeature = xPropagatedFeature();
    m_hAddedFeat = aeh.AddFeat(*pPropagatedFeature);
}

CSeq_annot_Handle CCmdPropagateFeat::xDefaultGetAppropriateFeatureTable()
{
    //check is there is a feature table already
    CSeq_entry_Handle seh = m_target_h.GetSeq_entry_Handle();
    CSeq_annot_Handle th;
    CSeq_annot_CI ci(m_target_h);
    for (/**/; ci; ++ci) {
        if (ci->IsFtable()) {
            return *ci;
        }
    }
    //if not, make a new one
    CBioseq_EditHandle eh = m_target_h.GetEditHandle();
    CRef<CSeq_annot> pAnnot(new CSeq_annot());
    m_hAddedTable = eh.AttachAnnot(*pAnnot);
    return m_hAddedTable;
}

void CCmdPropagateFeat::xPropagateCds()
{
    CRef<CSeq_feat> pPropagatedFeature = xPropagatedFeature();
    if (mCdsStopAtStopCodon) {
        xCdsStopAtStopCodon(pPropagatedFeature);
    }
    if (mCdsCleanupPartials) {
        xCdsCleanupPartials(pPropagatedFeature);
    }
    xCdsAttachNucProt(pPropagatedFeature);
}

CRef<CSeq_feat> CCmdPropagateFeat::xPropagatedFeature()
{
    const CSeq_feat::TData& data = m_feat->GetData();
    const CSeq_feat::TLocation& location = m_feat->GetLocation();

    CRef<CSeq_feat> pPropagatedFeature(new CSeq_feat);
    CRef<CSeq_id> pTargetId(new CSeq_id);
    pTargetId->Assign(*m_target_h.GetSeqId());

    pPropagatedFeature->SetLocation().Assign(location);
    pPropagatedFeature->SetLocation().SetId(*pTargetId);
    pPropagatedFeature->SetData().Assign(data);
    return pPropagatedFeature;
}

void CCmdPropagateFeat::xCdsCleanupPartials(CRef<CSeq_feat> pPropagated) 
{
    if (pPropagated->GetData().GetSubtype() != CSeqFeatData::eSubtype_cdregion) {
        return;
    }

    //assume that the original is marked correctly, i.e. if there is not fuzz-from
    // then it is 5'complete and it there is no fuzz-to then it is 3' prime complete.
    // no need to recheck whther there are actual start and stop codons.

    //for the copies:
    //the copy ends in a stop codon: it is 3' complete. Nothing needs to be done.
    // the copy does not end in a stop codon. hence it is not 3' complete.
    //the original is 5' incomplete: the copy is 5' incomplete
    // the original is 5' complete, copy starts with a start codon: the copy is
    // 5' complete

    try {
        const CSeq_feat::TLocation& location = m_feat->GetLocation();
        bool origIsPartialStart = location.IsPartialStart(eExtreme_Biological);
        bool origIsPartialStop = location.IsPartialStop(eExtreme_Biological);

        const CSeq_feat& cds = *pPropagated;
        CScope& scope = m_target_h.GetScope();
        CRef<CBioseq> pTranslated = CSeqTranslator::TranslateToProtein(cds, scope);
        if (!pTranslated) {
            return;
        }
        CSeqVector seqv(*pTranslated, &scope, CBioseq_Handle::eCoding_Ncbi);
        seqv.SetCoding(CSeq_data::e_Ncbieaa);

        bool targetIsPartialStart = (*CSeqVector_CI(seqv, 0) != 'M');
        bool targetIsPartialStop = (*CSeqVector_CI(seqv, seqv.size()-1) != '*');
        if (!targetIsPartialStart  && !origIsPartialStart) {
            //propagated feature is 5' complete
            //pPropagated->SetLocation().SetPartialStart(false, eExtreme_Biological);
        }
        else {
            //propagated feature is 5' partial
            pPropagated->SetLocation().SetPartialStart(true, eExtreme_Biological);
        }
        if (!targetIsPartialStop) {
            //propagated feature ends in stop codon hence is 3' complete
            pPropagated->SetLocation().SetPartialStop(false, eExtreme_Biological);
        }
        else {
            //propagated feature is 3' partial
            pPropagated->SetLocation().SetPartialStop(true, eExtreme_Biological);
        }
    }
    catch (...) {
    }
    return;
}

void CCmdPropagateFeat::xCdsAttachNucProt(CRef<CSeq_feat> pCds) 
{
    if (pCds->GetData().GetSubtype() != CSeqFeatData::eSubtype_cdregion) {
        return;
    }
    //prepare the protein
    CRef<CSeq_entry> pProtein = ncbi::CreateTranslatedProteinSequence(
        pCds, m_target_h);
    CRef<CSeq_feat> pProtFeat = ncbi::AddProteinFeatureToProtein(
        pProtein,
        pCds->GetLocation().IsPartialStart(eExtreme_Biological),
        pCds->GetLocation().IsPartialStop(eExtreme_Biological));

    //make a new nuc-prot set
    CSeq_entry_Handle seh = m_target_h.GetSeq_entry_Handle();
    CSeq_entry_EditHandle eh = seh.GetEditHandle();
    if (!eh.IsSet()) {
        CBioseq_set_Handle nucParent = eh.GetParentBioseq_set();
        if (nucParent  &&  nucParent.IsSetClass()  &&  
                nucParent.GetClass() == CBioseq_set::eClass_nuc_prot) {
            eh = nucParent.GetParentEntry().GetEditHandle();
        }
    }
    if (!eh.IsSet()) {
        eh.ConvertSeqToSet();
        eh.SetSet().SetClass(CBioseq_set::eClass_nuc_prot);
        mCreatedNucProtSet = true;
    };
    m_hAddedProtein = eh.AttachEntry(*pProtein);

    //add cds feature at set level
    CRef<CSeq_annot> pAnnot(new CSeq_annot());
    CSeq_annot_Handle th = eh.AttachAnnot(*pAnnot);
    CSeq_annot_EditHandle teh(th);
    m_hAddedFeat = teh.AddFeat(*pCds);
    return;
}

void CCmdPropagateFeat::xCdsStopAtStopCodon(CRef<CSeq_feat> pCds)
{
    if (pCds->GetData().GetSubtype() != CSeqFeatData::eSubtype_cdregion) {
        return;
    }
    const CSeq_feat& cds = *pCds;
    CScope& scope = m_target_h.GetScope();
    CRef<CBioseq> pTranslated = CSeqTranslator::TranslateToProtein(cds, scope);
    if (!pTranslated) {
        return;
    }
    CRef<CSeq_loc> pAdjustedLoc;
    CSeqVector seqv(*pTranslated, &scope, CBioseq_Handle::eCoding_Ncbi);
    seqv.SetCoding(CSeq_data::e_Ncbieaa);
    CSeqVector_CI it(seqv);
    while (it && *it != '*') {
        it++;
    }
    if (it) {
        //found stop codon inside the given feature
        size_t newSize = 3*(it.GetPos()+1);
        if (cds.GetData().GetCdregion().IsSetFrame()) {
            switch(cds.GetData().GetCdregion().GetFrame()) {
            default:
                break;
            case CCdregion::eFrame_two:
                newSize += 1;
                break;
            case CCdregion::eFrame_three:
                newSize += 2;
                break;
            }
        }
        pAdjustedLoc = sTruncateToStopCodon(cds.GetLocation(), newSize);
    }
    else {
        pAdjustedLoc = sExtendToStopCodon(pCds, scope);
    }
    if (pAdjustedLoc) {
        pCds->SetLocation(*pAdjustedLoc);
    }
}

END_NCBI_SCOPE
