/*  $Id: entrez_conversion.cpp 26478 2012-09-21 14:50:34Z kuznets $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Andrey Yazhuk
 *
 * File Description:
 *     */

#include <ncbi_pch.hpp>

#include <gui/objutils/entrez_conversion.hpp>

#include <corelib/ncbitime.hpp>

#include <objects/entrez2/Entrez2_id_list.hpp>
#include <objects/entrez2/Entrez2_info.hpp>
#include <objects/entrez2/Entrez2_eval_boolean.hpp>
#include <objects/entrez2/Entrez2_boolean_reply.hpp>
#include <objects/entrez2/Entrez2_boolean_exp.hpp>
#include <objects/entrez2/Entrez2_boolean_element.hpp>
#include <objects/entrez2/Entrez2_docsum_data.hpp>
#include <objects/entrez2/Entrez2_limits.hpp>
#include <objects/entrez2/Entrez2_link_info.hpp>
#include <objects/entrez2/Entrez2_link_count.hpp>
#include <objects/entrezgene/Entrezgene.hpp>

#include <util/static_map.hpp>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);


typedef SStaticPair<const char*, const char*> TPair;
static const TPair sc_Dbs[] = {
    { "gene",       "Entrez Gene" },
    { "genome",     "Entrez Genome" },
    { "nucleotide", "Nucleotide" },
    { "protein",    "Protein" },
};
typedef CStaticPairArrayMap<const char*, const char*, PNocase_CStr> TEntrezDbMap;
DEFINE_STATIC_ARRAY_MAP(TEntrezDbMap, sc_EntrezDbMap, sc_Dbs);

static const TPair sc_Pairs[] = {
    { "genomeprj_genome",          "Chromosomes" },
    { "genomeprj_nucleotide",      "Contigs" },
    { "genomeprj_nucleotide_mrna", "mRNAs" },
    { "genomeprj_nucleotide_wgs",  "Whole Genome Shotgun sequences" },
    { "genomeprj_organella",       "Organelles" },
    { "genomeprj_protein",         "Proteins" },
};
typedef CStaticPairArrayMap<const char*, const char*, PNocase_CStr> TGenomeLinkMap;
DEFINE_STATIC_ARRAY_MAP(TGenomeLinkMap, sc_GenomeLinkMap, sc_Pairs);


void CEntrezDB::GetDbNames(vector<string>& names)
{
    ITERATE(TEntrezDbMap, it, sc_EntrezDbMap)   {
        names.push_back(it->first);
    }
}

void CEntrezDB::GetDbNames(vector<TNamePair>& name_pairs)
{
    ITERATE(TEntrezDbMap, it, sc_EntrezDbMap)   {
        name_pairs.push_back(TNamePair(it->first, it->second));
    }
}


string CEntrezDB::GetVisibleName(const string& db_name)
{
    TEntrezDbMap::const_iterator it = sc_EntrezDbMap.find(db_name.c_str());
    return (it == sc_EntrezDbMap.end()) ? kEmptyStr : it->second;
}


string CEntrezDB::GetDbName(const string& visible_name)
{
    ITERATE (TEntrezDbMap, iter, sc_EntrezDbMap) {
        if (iter->second == visible_name) {
            return iter->first;
        }
    }
    return kEmptyStr;
}

const string& CEntrezDB::GetDbFieldName()
{
    static const string db_field_name("_db_name_");
    return db_field_name;
}


CRef<CEntrez2_docsum_list> CEntrezDB::Query(CEntrez2Client& client, const string& db_name,
                                            const string& terms,
                                            size_t& total_uids,
                                            size_t start,
                                            size_t count,
                                            bool add_db_field)
{
    CRef<CEntrez2_docsum_list> list_ref(new CEntrez2_docsum_list());

    try {
        CEntrez2_eval_boolean req;
        CEntrez2_boolean_exp& exp = req.SetQuery();

        // set the database we're querying
        exp.SetDb().Set(db_name);

        // set the query
        CRef<CEntrez2_boolean_element> elem(new CEntrez2_boolean_element());
        elem->SetStr(terms);
        exp.SetExp().push_back(elem);

        // get UIDs first, then get the docsums set some limits - if num > 0,
        // we assume it's correct
        if (start > 0) {
            exp.SetLimits().SetOffset_UIDs((int)start);
        }
        if (count > 0) {
            exp.SetLimits().SetMax_UIDs((int)count);
        }
        req.SetReturn_UIDs(true);
        CRef<CEntrez2_boolean_reply> query_res;

        // CEntrez2Client throws exceptions even for valid queries, so we guard
        // against this ungraceful behaviour
        try {
            query_res = client.AskEval_boolean(req);
        } catch(CException& e) {
            ERR_POST("CEntrezDB::Query() - CEntrez2Client has thrown an exception ");
            ERR_POST(e.ReportAll());
        }

        // now we need to retrieve DocSums for the query result
        //

        if(query_res) {
            total_uids = query_res->GetCount();
            if (total_uids == 0)
                return list_ref;

            const CEntrez2_boolean_reply_Base::TUids& uids_list = query_res->GetUids();

            vector<int> result_uids;
            result_uids.reserve(total_uids);
            for (CEntrez2_id_list::TConstUidIterator it = uids_list.GetConstUidIterator();  
                 !it.AtEnd();  ++it) {
                result_uids.push_back(*it);
            }


            // docsum retrival is structured to get 100 items at one transaction
            // larger transactions (like 1000) are not stable and can timeout...
            //
            const size_t docsum_batch_max = 100;
            vector<int> uid_batch;
            uid_batch.reserve(docsum_batch_max);
            for (size_t i = 0; i < result_uids.size(); ) {
                uid_batch.resize(0);
                for (size_t j = 0; j < docsum_batch_max && i < result_uids.size(); ++j, ++i) {
                    int doc_uid = result_uids[i];
                    uid_batch.push_back(doc_uid);
                }
                if (uid_batch.empty()) {
                    continue;
                }
                CRef<CEntrez2_docsum_list> tmp_ds_list;
                try {
                    tmp_ds_list = client.GetDocsums(uid_batch, db_name);
                } catch(CException& e) {
                    ERR_POST("CEntrezDB::Query() - CEntrez2Client::GetDocsums has thrown an exception ");
                    ERR_POST(e.ReportAll());
                }
                
                if (!tmp_ds_list.IsNull()) {
                    CEntrez2_docsum_list::TList& src_list = tmp_ds_list->SetList();
                    CEntrez2_docsum_list::TList& dst_list = list_ref->SetList();

                    ITERATE(CEntrez2_docsum_list::TList, lit, src_list) {
                        dst_list.push_back(*lit);
                    }
                }

            } // for
            

            //list_ref = client.AskGet_docsum(uids_list);

            // add db_name field if needed
            if(add_db_field)    {
                const string& field_name = GetDbFieldName();

                CEntrez2_docsum_list::TList& list = list_ref->SetList();
                NON_CONST_ITERATE(CEntrez2_docsum_list::TList, it, list)  {
                    CEntrez2_docsum& doc_sum = **it;
                    CEntrez2_docsum::TDocsum_data& data = doc_sum.SetDocsum_data();

                    // add a new field
                    CRef<CEntrez2_docsum_data> field(new CEntrez2_docsum_data());
                    field->SetField_name(field_name);
                    field->SetField_value(db_name);
                    data.push_back(field);
                }
            }
            return list_ref;
        }
    }
    catch (CException& e) {
        LOG_POST(Error << "CEntrezDB::Query(): error executing Entrez query: "
                 << terms << ": " << e.GetMsg());
    }
    catch (std::exception& e) {
        LOG_POST(Error << "CEntrezDB::Query(): error executing Entrez query: "
                 << terms << ": " << e.what());
    }
    return list_ref;
}


CRef<CSeq_annot> CEntrezDB::CreateAnnot_Nuc_Prot(const string& query_str,
                                                 const TDocsumList& ds_list)
{
    CRef<CSeq_annot> annot(new CSeq_annot());
    ITERATE (TDocsumList, iter, ds_list) {
        CRef<CSeq_id> id(new CSeq_id());
        id->SetGi((*iter)->GetUid());
        annot->SetData().SetIds().push_back(id);
    }

    string title("Query: ");
    title += query_str;
    title += ", database: ";
    //TODO title += GetVisibleDbName();
    title += ": ";
    title += NStr::SizetToString(ds_list.size(), NStr::fWithCommas);
    title += " sequence";
    if (ds_list.size() != 1) {
        title += "s";
    }
    annot->SetNameDesc(title);
    annot->SetCreateDate(CTime(CTime::eCurrent));
    return annot;
}


CRef<CSeq_id> CEntrezDB::CreateId_Nuc_Prot(const string& query_str,
                                             const TDocsum& ds)
{
    CRef<CSeq_id> id(new CSeq_id());
    id->SetGi(ds.GetUid());
    return id;
}


/// TODO - refactor
/// made from CreateGeneSet_Gene
CRef<objects::CEntrezgene> CEntrezDB::CreateGene_Gene(const TDocsum& ds)
{
    static const string url_base
        ("http://eutils.ncbi.nlm.nih.gov"
         "/entrez/eutils/efetch.fcgi"
         "?db=gene&retmode=asn.1&id=");

    CRef<CEntrezgene> egene;

    string ids = NStr::IntToString(ds.GetUid());
    string url(url_base);
    url += ids;
    CConn_HttpStream istr(url);
    _TRACE("Retrieving url: " << url);

    // exit by exception
    try {
        CRef<CEntrezgene> ref(new CEntrezgene());
        istr >> MSerial_AsnText >> *ref;
        egene = ref;
    }
    catch (CException& e) {
        LOG_POST(Error
            << "CEntrezDB::CreateGene_Gene(): error retrieving IDs "
            << ids << ": " << e.GetMsg());
        string msg("Failed to retrieve records from Entrez Gene:\n");
        msg += e.GetMsg();
    }
    catch (std::exception& e) {
        LOG_POST(Error
            << "CEntrezDB::CreateGene_Gene(): error retrieving IDs "
            << ids << ": " << e.what());

        string msg("Failed to retrieve records from Entrez Gene:\n");
        msg += e.what();
    }

    return egene;
}


CRef<CEntrezgene_Set> CEntrezDB::CreateGeneSet_Gene(const TDocsumList& ds_list)
{
    /// for each docsum, retrieve the Entrezgene object
    static const string url_base
        ("http://eutils.ncbi.nlm.nih.gov"
         "/entrez/eutils/efetch.fcgi"
         "?db=gene&retmode=asn.1&id=");

    CRef<CEntrezgene_Set> gene_set;

    string ids;
    TDocsumList::const_iterator iter = ds_list.begin();
    TDocsumList::const_iterator end  = ds_list.end();
    for (size_t count = 0;  iter != ds_list.end();  ++count) {
        const TDocsum& ds = **iter;
        if ( !ids.empty() ) {
            ids += ",";
        }

        ids += NStr::IntToString(ds.GetUid());
        ++iter;
        if (count % 5 == 4  ||  iter == ds_list.end()) {
            /// retrieve
            string url(url_base);
            url += ids;
            CConn_HttpStream istr(url);
            _TRACE("Retrieving url: " << url);

            // exit by exception
            for (;;) {
                try {
                    CRef<CEntrezgene> egene(new CEntrezgene());
                    istr >> MSerial_AsnText >> *egene;

                    if ( !gene_set ) {
                        gene_set.Reset(new CEntrezgene_Set);
                    }
                    gene_set->Set().push_back(egene);
                }
                catch (CEofException&) {
                    /// ignore end-of-file - we expect this on completion
                    break;
                }
                catch (CException& e) {
                    LOG_POST(Error
                        << "CEntrezDB_Gene::CreateGeneSet_Gene(): error retrieving IDs "
                        << ids << ": " << e.GetMsg());
                    break;
                }
                catch (std::exception& e) {
                    LOG_POST(Error
                        << "CEntrezDB_Gene::CreateGeneSet_Gene(): error retrieving IDs "
                        << ids << ": " << e.what());
                    break;
                }
            }
            ids.erase();
        }
    }
    if ( gene_set  &&  gene_set->IsSet()  &&  !gene_set->Get().size() ) {
        gene_set.Reset();
    }

    return gene_set;
}


string
CEntrezDB::CreateEntrezLinkCountLabel(const CEntrez2_link_count& lc)
{
    string label;
    const string& name = lc.GetLink_type();

    TGenomeLinkMap::const_iterator link_it =
        sc_GenomeLinkMap.find(name.c_str());
    if (link_it != sc_GenomeLinkMap.end()) {
        label = link_it->second;
        label += " (";
        label += NStr::IntToString(lc.GetLink_count(), NStr::fWithCommas);
        label += " item";
        if (lc.GetLink_count() != 1) {
            label += "s";
        }
        label += ")";
    }
    return label;
}


CRef<CSeq_annot> CEntrezDB::CreateAnnot_Genome(const string& query_str,
                                               const TDocsumList& ds_list,
                                               const string& vis_db_name)
{
    CRef<CSeq_annot> annot(new CSeq_annot());
    ITERATE (TDocsumList, iter, ds_list) {
        try {
            CRef<CSeq_id> id(new CSeq_id((*iter)->GetValue("Extra")));
            annot->SetData().SetIds().push_back(id);
        }
        catch (CSeqIdException&) {
        }
    }
    string title("Query: ");
    title += query_str;
    title += ", database: ";
    title += vis_db_name;
    title += ": ";
    title += NStr::SizetToString(ds_list.size(), NStr::fWithCommas);
    title += " sequence";
    if (ds_list.size() != 1) {
        title += "s";
    }
    annot->SetNameDesc(title);
    return annot;
}


string CEntrezDB::CreateAnnotTitle_GenomeProject(const CEntrez2_docsum& ds,
                                                 const CEntrez2_link_count& lc)
{
    string title;
    title = ds.GetValue("Organism_Name");
    {{
        TGenomeLinkMap::const_iterator iter =
        sc_GenomeLinkMap.find(lc.GetLink_type().Get().c_str());
        title += " (";
        title += iter->second;
        title += ")";
    }}

    title += ", ";
    title += ds.GetValue("Defline");
    title += " (";
    title += ds.GetValue("Center");
    title += ")";
    return title;
}

CRef<CSeq_annot>
    CEntrezDB::CreateSeqAnnot_GenomeProject(CEntrez2Client& client,
                                            const TDocsum& ds,
                                            const CEntrez2_link_count& lc,
                                            const string& db_name)
{
    CRef<CSeq_annot> annot(new CSeq_annot);

    string name = lc.GetLink_type();
    vector<int> uids;
    if (name == "genomeprj_genome") {
        /// two levels of indirection
        client.GetNeighbors(ds.GetUid(), db_name, "genome", uids);

        CRef<CEntrez2_docsum_list> ds_list =
            client.GetDocsums(uids, "genome");

        ITERATE (CEntrez2_docsum_list::TList, ds_iter, ds_list->GetList()) {
            const CEntrez2_docsum& ds = **ds_iter;
            string id_str = ds.GetValue("Extra");
            CRef<CSeq_id> id(new CSeq_id(id_str));
            annot->SetData().SetIds().push_back(id);
        }
    } else {
        name.erase(0, db_name.size() + 1);
        client.GetNeighbors(ds.GetUid(), db_name, name, uids);

        /// UIDs should now be GIs
        ITERATE (vector<int>, it, uids) {
            CRef<CSeq_id> id(new CSeq_id());
            id->SetGi(*it);
            annot->SetData().SetIds().push_back(id);
        }
    }

    string title = CreateAnnotTitle_GenomeProject(ds, lc);
    annot->SetNameDesc(title);
    annot->SetCreateDate(CTime(CTime::eCurrent));

    return annot;
}

END_NCBI_SCOPE
