/*  $Id: gui_object_info_clone_feat.cpp 26401 2012-09-12 13:52:01Z wuliangs $
* ===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data, the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties, express or implied, including
*  warranties of performance, merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================
*
* Authors: Liangshou Wu
*
* File Description:
*     Gui Object Info class specifically for clone features.
*
*/

#include <objects/seqfeat/Clone_ref.hpp>
#include <objects/seqfeat/Clone_seq_set.hpp>
#include <objects/seqfeat/Clone_seq.hpp>


BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

class CGuiObjectInfoClone : public CGuiObjectInfoSeq_feat
{
public:
    virtual void GetToolTip(string& t_text, string& t_title,
        TSeqPos at_p = (TSeqPos)-1) const;
    virtual void GetLinks(string& links, bool no_ncbi_base) const;
};


static string s_PackStrings(const vector<string> strs)
{
    string out_str;
    bool first = true;
    ITERATE (vector<string>, iter, strs) {
        if (first) {
            first = false;
            out_str = *iter;
        } else {
            out_str += ", " + *iter;
        }
    }
    return out_str;
}


void CGuiObjectInfoClone::GetToolTip(string& t_text, string& t_title,
                                     TSeqPos /*at_p*/) const
{
    CLabel::GetLabel(*m_Feat, &t_title, CLabel::eContent, m_Scope.GetPointer());
    t_text += CSeqUtils::CreateTableStart();

    const CSeqFeatData::TClone& clone = m_Feat->GetData().GetClone();
    if (m_Location) {
        t_text += CSeqUtils::CreateTableRow("Total length",
            NStr::IntToString(m_Location->GetTotalRange().GetLength(), NStr::fWithCommas));
        string range = NStr::IntToString(
            m_Location->GetStart(eExtreme_Positional) + 1, NStr::fWithCommas);
        range += "..";
        range += NStr::IntToString(
            m_Location->GetStop(eExtreme_Positional) + 1, NStr::fWithCommas);
        t_text += CSeqUtils::CreateTableRow("Total range", range);
    }

    if (clone.IsSetPlacement_method()) {
        const CEnumeratedTypeValues* type_val =
            CClone_ref::GetTypeInfo_enum_EPlacement_method();
        if (type_val) {
            string method = type_val->FindName(clone.GetPlacement_method(), true);
            if ( !method.empty() ) {
                t_text += CSeqUtils::CreateTableRow("Placement method", method);
            }
        }
    }
    if (clone.IsSetConcordant()) {
        t_text += CSeqUtils::CreateTableRow("Clone placement",
            clone.GetConcordant() ? "concordant" : "discordant");
    }
    if (clone.IsSetUnique()) {
        t_text += CSeqUtils::CreateTableRow("Clone placement",
            clone.GetUnique() ? "unique" : "multiple");
    }
    if (clone.IsSetClone_seq()) {
        typedef map<int, vector<string> > TEndMap;
        TEndMap end_map;
        const CClone_ref::TClone_seq::Tdata& ends = clone.GetClone_seq().Get();
        ITERATE (CClone_ref::TClone_seq::Tdata, e_iter, ends) {
            const CClone_seq& seq = **e_iter;
            if (seq.CanGetSeq()  &&  seq.CanGetSupport()) {
                const CSeq_id* id = seq.GetSeq().GetId();
                if (id) {
                    string id_str;
                    CSeq_id_Handle idh = sequence::GetId(*id, *m_Scope);
                    if (idh) {
                        CSeq_id_Handle idhb =
                            sequence::GetId(idh, *m_Scope, sequence::eGetId_Best);
                        if( idhb ){
                            idh = idhb;
                        }
                    }
                    if (idh) {
                        CConstRef<CSeq_id> seq_id_ref = idh.GetSeqId();
                        seq_id_ref->GetLabel( &id_str, CSeq_id::eContent );
                    } else {
                        id->GetLabel(&id_str);
                    }
                    end_map[seq.GetSupport()].push_back(id_str);
                }
            }
        }

        if (end_map.count(CClone_seq::eSupport_prototype) > 0) {
            t_text += CSeqUtils::CreateTableRow("Prototype",
                s_PackStrings(end_map[CClone_seq::eSupport_prototype]));
        }

        if (end_map.count(CClone_seq::eSupport_supporting) > 0) {
            t_text += CSeqUtils::CreateTableRow("Supporting (" +
                NStr::NumericToString(end_map[CClone_seq::eSupport_supporting].size()) +
                ")", s_PackStrings(end_map[CClone_seq::eSupport_supporting]));
        }

        // for non-supporting ends
        size_t total_ns_num = 0;
        string ids_ns_str = kEmptyStr;
        if (end_map.count(CClone_seq::eSupport_non_supporting) > 0) {
            total_ns_num += end_map[CClone_seq::eSupport_non_supporting].size();
            ids_ns_str = s_PackStrings(end_map[CClone_seq::eSupport_non_supporting]);
        }
        if (end_map.count(CClone_seq::eSupport_supports_other) > 0) {
            total_ns_num += end_map[CClone_seq::eSupport_supports_other].size();
            if ( !ids_ns_str.empty() ) {
                ids_ns_str += ", ";
            }
            ids_ns_str += s_PackStrings(end_map[CClone_seq::eSupport_supports_other]);
        }

        if (total_ns_num > 0) {
            t_text += CSeqUtils::CreateTableRow("Non-supporting (" +
                NStr::NumericToString(total_ns_num) + ")", ids_ns_str);
        }
    }
    t_text += CSeqUtils::CreateTableEnd();
}


void CGuiObjectInfoClone::GetLinks(string& links, bool no_ncbi_base) const
{
}

END_NCBI_SCOPE