/*  $Id: gui_object_info_seq_align.cpp 27887 2013-04-18 20:51:31Z wuliangs $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors: Roman Katargin, Liangshou Wu
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>
#include <gui/objutils/label.hpp>
#include <gui/objutils/utils.hpp>
#include <objmgr/util/sequence.hpp>
#include <objmgr/seq_vector.hpp>
#include <objmgr/util/create_defline.hpp>
#include <algo/align/util/score_builder.hpp>
#include <objects/seqalign/Dense_seg.hpp>
#include <objects/seqalign/Spliced_seg.hpp>

#include <gui/objutils/gui_object_info_seq_align.hpp>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

void initCGuiObjectInfoSeq_align()
{
    CInterfaceRegistry::RegisterFactory(
            typeid(IGuiObjectInfo).name(),
            CSeq_align::GetTypeInfo(),
            new CObjectInterfaceFactory<CGuiObjectInfoSeq_align>());
}

CGuiObjectInfoSeq_align* CGuiObjectInfoSeq_align::CreateObject(
    SConstScopedObject& object, ICreateParams* /*params*/)
{
    CGuiObjectInfoSeq_align* gui_info = new CGuiObjectInfoSeq_align();
    gui_info->m_Align.Reset(dynamic_cast<const CSeq_align*>(object.object.GetPointer()));
    gui_info->m_Scope = object.scope;
    gui_info->m_AnchorRow = -1;
    return gui_info;
}

string CGuiObjectInfoSeq_align::GetLabel() const
{
    string label;
    CLabel::GetLabel(*m_Align, &label, CLabel::eDescriptionBrief, m_Scope);
    return label;
}


static size_t s_CountSegments(const CSeq_align& align)
{
    switch (align.GetSegs().Which()) {
    default:
        return 1;

    case CSeq_align::TSegs::e_Disc:
        return align.GetSegs().GetDisc().Get().size();

    case CSeq_align::TSegs::e_Spliced:
        return align.GetSegs().GetSpliced().GetExons().size();
    }
}


static bool s_IsPolyA(const objects::CBioseq_Handle& bsh,
                      TSeqPos start, TSeqPos stop)
{
    // we only apply testing on RNA sequences
    bool is_rna = false;    
    const CMolInfo* info = sequence::GetMolInfo(bsh);
    if (info) {
        if (info->GetBiomol() == CMolInfo::eBiomol_mRNA  ||
            info->GetBiomol() == CMolInfo::eBiomol_pre_RNA  ||
            info->GetBiomol() == CMolInfo::eBiomol_tRNA  ||
            info->GetBiomol() == CMolInfo::eBiomol_snRNA  ||
            info->GetBiomol() == CMolInfo::eBiomol_scRNA  ||
            info->GetBiomol() == CMolInfo::eBiomol_cRNA  ||
            info->GetBiomol() == CMolInfo::eBiomol_snoRNA  ||
            info->GetBiomol() == CMolInfo::eBiomol_ncRNA  ||
            info->GetBiomol() == CMolInfo::eBiomol_tmRNA) {
            is_rna = true;
        }
    }

    if ( !is_rna) {
        CConstRef<CSeq_id> seq_id;
        CSeq_id_Handle idh = sequence::GetId(*bsh.GetSeqId(),
            bsh.GetScope(), sequence::eGetId_Best);
        if (idh) {
            seq_id = idh.GetSeqId();
        } else {
            seq_id = bsh.GetSeqId();
        }

        CSeq_id::EAccessionInfo id_info = seq_id->IdentifyAccession();

        if ((id_info & CSeq_id::eAcc_division_mask) == CSeq_id::eAcc_est  ||
            id_info == CSeq_id::eAcc_refseq_mrna  ||
            id_info == CSeq_id::eAcc_refseq_mrna_predicted  ||
            id_info == CSeq_id::eAcc_gpipe_mrna) {
                is_rna = true;
        }

        if ( !is_rna  &&  id_info & CSeq_id::eAcc_ambig  &&
            info->GetBiomol() == CMolInfo::eBiomol_other) {
                is_rna = true;
        }
    }                                                                                                                              
   
    if (is_rna) {
        CSeqVector s_vec = bsh.GetSeqVector(CBioseq_Handle::eCoding_Iupac);
        string seq;
        s_vec.GetSeqData(start, stop + 1, seq);
        int a_cnt = 0;
        int t_cnt = 0;
        for (TSeqPos bp = 0;  bp != seq.length();  bp++) {
            if (seq[bp] == 'A') ++a_cnt;
            else if (seq[bp] == 'T') ++t_cnt;
        }
        if (a_cnt > (int)(seq.length() * 0.9)  ||
            t_cnt > (int)(seq.length() * 0.9)) {
            return true;
        }
    }

    return false;
}


static CFastMutex s_DFLock;
static sequence::CDeflineGenerator s_DefGen; ///< This class seems to be non-thread safe even as a stack var

static
inline void s_GetDefline(CBioseq_Handle bsh, string& defline )
{
    CFastMutexGuard lock(s_DFLock);
    defline = s_DefGen.GenerateDefline(bsh);
}


static string s_ShortenIdLabel(const string& label)
{
    string out_label = label;
    // make title/label shorter in case it is a cSRA alignment on panfs
    if (out_label.length() > 20  &&  out_label.find("\\panfs\\") != string::npos) {
        size_t pos = out_label.find_last_of("\\");
        if (pos != string::npos) {
            out_label = out_label.substr(pos + 1);
            if (out_label.length() > 20) {
                pos = out_label.rfind('.');
                if (pos != string::npos  &&  pos > 0) {
                    pos = out_label.rfind('.', pos - 1);
                    if (pos != string::npos) {
                        out_label = out_label.substr(pos + 1);
                    }
                }
            }
        }
    }
    return out_label;
}

static string s_RemovePanfsBasePath(const string& label)
{
    string out_label = label;
    // make title/label shorter in case it is a cSRA alignment on panfs
    size_t pos = out_label.find("\\panfs\\");
    if (pos != string::npos) {
        pos = out_label.find("\\", pos + 7);
        if (pos != string::npos) {
            out_label = out_label.substr(pos);
        }
    }
    return out_label;
}


void CGuiObjectInfoSeq_align::GetToolTip(string& t_text, string& t_title,
                                         TSeqPos at_p) const
{
    const CSeq_align::TDim num_rows_limit = 10;

    t_text += CSeqUtils::CreateTableStart();
    ///
    /// first, format our seq-ids
    ///
    vector<bool> strands;
    vector<string> ids;
    CSeq_align::TDim num_rows = m_Align->CheckNumRows();
    string curr_text;
    {{
        for(CSeq_align::TDim i = 0;  i < num_rows  &&  i < num_rows_limit;  ++i ){
            CSeq_id_Handle idh = CSeq_id_Handle::GetHandle( m_Align->GetSeq_id(i) );
            CSeq_id_Handle idh_best = sequence::GetId(idh, *m_Scope, sequence::eGetId_Best);
            if( idh_best ){
                idh = idh_best;
            }

            if ( !curr_text.empty() ) {
                curr_text += " x ";
            }
            CConstRef<CSeq_id> seq_id(idh.GetSeqIdOrNull());
            string label;
            if (!seq_id) {
                label = "UNK";
            } else {
                idh.GetSeqId()->GetLabel(&label, CSeq_id::eContent);
            }
            ids.push_back(label);
            curr_text += s_ShortenIdLabel(label);
            try {
                strands.push_back(m_Align->GetSeqStrand(i) == eNa_strand_plus);
            } catch (CException&) {
                // ignore the exception
            }
        }
    }}

    if (num_rows > ids.size()) {
        curr_text += " x ... [total ";
        curr_text += NStr::IntToString(num_rows, NStr::fWithCommas);
        curr_text += "]";
    }
    t_text += CSeqUtils::CreateTableRow("Alignment", curr_text);

    if (num_rows > ids.size()) { // early escape?
        CSeqUtils::CreateTableEnd();
        return;
    }

    bool is_protein = false;
    if (ids.size() == 2  &&  num_rows == 2  &&  m_AnchorRow >= 0  &&  m_AnchorRow < 2) {
        TSeqRange range = m_Align->GetSeqRange(m_AnchorRow);
        curr_text = ids[m_AnchorRow] + " (";
        curr_text += NStr::IntToString(range.GetFrom() + 1, NStr::fWithCommas);
        curr_text += "..";
        curr_text += NStr::IntToString(range.GetTo() + 1, NStr::fWithCommas);
        curr_text += ")";
        t_text += CSeqUtils::CreateTableRow("Anchor", curr_text);

        range = m_Align->GetSeqRange(1 - m_AnchorRow);
        t_title = s_ShortenIdLabel(ids[1 - m_AnchorRow]);
        curr_text = s_RemovePanfsBasePath(ids[1 - m_AnchorRow]) + " (";
        curr_text += NStr::IntToString(range.GetFrom() + 1, NStr::fWithCommas);
        curr_text += "..";
        curr_text += NStr::IntToString(range.GetTo() + 1, NStr::fWithCommas);
        curr_text += ")";
        t_text += CSeqUtils::CreateTableRow("Query", curr_text);

        CBioseq_Handle bsh = m_Scope->GetBioseqHandle(m_Align->GetSeq_id(1 - m_AnchorRow));
        if (bsh) {
            if (bsh.IsProtein()) {
                is_protein = true;
            }
            s_GetDefline(bsh, curr_text);
            if ( !curr_text.empty() ) {
                t_text += "<tr><td></td><td>" + curr_text  + "</td></tr>";
            }
        }
    }

    /// only report strand for pair-wise alignments
    if (!is_protein  &&  num_rows == 2  &&  strands.size() == 2) {
        t_text += CSeqUtils::CreateTableRow("Strand",
            strands[0] == strands[1] ? "forward" : "reverse");
    }

    string tag_name;
    ///
    /// next, add a remark about the total aligned range
    ///
    CScoreBuilder builder;
    // new method
    TSeqPos align_length = builder.GetAlignLength(*m_Align);
    tag_name = "Aligned ";
    tag_name += is_protein ? "residues" : "bases";

    t_text += CSeqUtils::CreateTableRow(tag_name,
        NStr::IntToString(align_length, NStr::fWithCommas));

    size_t segs = s_CountSegments(*m_Align);
    t_text += CSeqUtils::CreateTableRow("Segments",
        NStr::SizetToString(segs, NStr::fWithCommas));

    char buf[255];
    double coverage = -1.0;
    if ( !m_Align->GetNamedScore(CSeq_align::eScore_PercentCoverage, coverage) && m_Scope ){
        try {
            coverage = builder.GetPercentCoverage( *m_Scope, *m_Align );
        } catch (CException&) {
            // ignore
        }
    }
    if (coverage >= 0.0) {
        if (coverage < 100.0  &&  coverage > 99.9) {
            curr_text = "99.9+";
        } else {
            sprintf(buf, "%2.1f", coverage);
            curr_text = buf;
        }
        curr_text += "%";
        t_text += CSeqUtils::CreateTableRow("Coverage", curr_text);
    }

    if (align_length < 1000000) {
        //int identities = builder.GetIdentityCount(*m_Scope, align);
        //int mismatches = builder.GetMismatchCount(*m_Scope, align);
        int identities = -1;
        int mismatches = -1;
        if (!m_Align->GetNamedScore(CSeq_align::eScore_IdentityCount, identities)  ||
            !m_Align->GetNamedScore(CSeq_align::eScore_MismatchCount, mismatches)) {
                try {
                    builder.GetMismatchCount(*m_Scope, *m_Align, identities, mismatches);
                } catch (CException&) {
                    // ignore
                }
        }

        if (identities >= 0) {
            double identity = 0.0;
            if ( !m_Align->GetNamedScore(CSeq_align::eScore_PercentIdentity, identity) ) {
                identity = identities * 100.0 / align_length;
            }
            if (identity < 100.0  &&  identity > 99.9) {
                curr_text = "99.9+";
            } else {
                sprintf(buf, "%2.1f", identity);
                curr_text = buf;
            }
            curr_text += "%";
            t_text += CSeqUtils::CreateTableRow("Identity", curr_text);
        } 
        if (mismatches >= 0) {
            t_text += CSeqUtils::CreateTableRow("Mismatches",
            NStr::NumericToString(mismatches, NStr::fWithCommas));
        }
    }

    try {
        int gap_count  = builder.GetGapCount(*m_Align);
        t_text += CSeqUtils::CreateTableRow("Gaps",
            NStr::IntToString(gap_count, NStr::fWithCommas));
    } catch (CException&) {
        // ignore
    }

    if (m_Align->CheckNumRows() == 2  &&  m_AnchorRow > -1) {
        int aligned_seq = 1 - m_AnchorRow;
        const CSeq_id& seq_id = m_Align->GetSeq_id(aligned_seq);
        CSeq_id_Handle sih = sequence::GetId(seq_id,
            *m_Scope, sequence::eGetId_Best);
        if ( !sih ) {
            sih = sequence::GetId(seq_id, *m_Scope, sequence::eGetId_Canonical);
        }
        if (sih) {
            CBioseq_Handle bsh;
            try {
                bsh = m_Scope->GetBioseqHandle(seq_id);
            } catch (CException&) {
                // ignore
            }
            TSeqRange range = m_Align->GetSeqRange(aligned_seq);
            string unaligned_bases_5;
            string unaligned_bases_3;
            if (range.GetFrom() > 0) {
                unaligned_bases_5 = NStr::IntToString(range.GetFrom());
                if (!is_protein  &&  bsh) {
                    if (s_IsPolyA(bsh, 0, range.GetFrom() - 1)) {
                        unaligned_bases_5 += " (polyA)";
                    }
                }
            }

            if (bsh  &&  range.GetTo() < bsh.GetBioseqLength() - 1) {
                unaligned_bases_3 = NStr::IntToString(bsh.GetBioseqLength() - 1 - range.GetTo());
                if ( !is_protein ) {
                    bool is_polya = s_IsPolyA(bsh, range.GetToOpen(), bsh.GetBioseqLength() - 1);
                    if (is_polya) {
                        unaligned_bases_3 += " (polyA)";
                    }
                }
            }
            if ( !unaligned_bases_5.empty()  ||  !unaligned_bases_3.empty() ) {
                tag_name = "Unaligned ";
                tag_name += is_protein ? "residules" : "bases";
                t_text += CSeqUtils::CreateSectionRow(tag_name);
                if ( !unaligned_bases_5.empty() ) {
                    t_text += CSeqUtils::CreateTableRow(is_protein ? "N-terminal end" : "5'", unaligned_bases_5);
                } 
                if ( !unaligned_bases_3.empty() ) {
                    t_text += CSeqUtils::CreateTableRow(is_protein ? "C-terminal end" : "3'", unaligned_bases_3);
                }
            }
        }
    }

    if (at_p != (TSeqPos)-1) {
        t_text += CSeqUtils::CreateTableRow("Position",
            NStr::UIntToString(at_p + 1, NStr::fWithCommas));
    }

    t_text += CSeqUtils::CreateTableEnd();
}


END_NCBI_SCOPE
