/*  $Id: gui_object_info_seq_feat.cpp 28080 2013-05-06 22:15:03Z rafanovi $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors: Roman Katargin, Liangshou Wu
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>

#include <gui/objutils/gui_object_info_seq_feat.hpp>

// some common includes
#include <objects/general/User_object.hpp>
#include <objects/general/User_field.hpp>
#include <objects/general/Dbtag.hpp>
#include <objects/general/Object_id.hpp>
#include <objects/seqfeat/Cdregion.hpp>
#include <gui/objutils/utils.hpp>
#include <gui/objutils/label.hpp>
#include <objmgr/util/feature.hpp>
#include <objmgr/util/sequence.hpp>
#include <objmgr/seq_vector.hpp>

// include other seq-feat implmentations
#include "gui_object_info_variant.cpp"    /// structural variants
#include "gui_object_info_variation.cpp" /// SNP
#include "gui_object_info_clone_feat.cpp"
#include "gui_object_info_primer_feat.cpp"
#include "gui_object_info_other_feats.cpp"

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);


void initCGuiObjectInfoSeq_feat()
{
    CInterfaceRegistry::RegisterFactory(
            typeid(IGuiObjectInfo).name(),
            CSeq_feat::GetTypeInfo(),
            new CObjectInterfaceFactory<CGuiObjectInfoSeq_feat>());
}


CGuiObjectInfoSeq_feat* CGuiObjectInfoSeq_feat::CreateObject(
    SConstScopedObject& object, ICreateParams* /*params*/)
{
    CGuiObjectInfoSeq_feat* gui_info = NULL;
    const CSeq_feat* feat = dynamic_cast<const CSeq_feat*>(object.object.GetPointer());

    switch (feat->GetData().GetSubtype())
    {
    case CSeqFeatData::eSubtype_clone:
        {{
            if (feat->GetNamedQual("type") == "primer") {
                gui_info = new CGuiObjectInfoPrimer();
            } else {
                gui_info = new CGuiObjectInfoClone();
            }
        }}
        break;
    case CSeqFeatData::eSubtype_variation:
        gui_info = new CGuiObjectInfoVariation();
        break;
    case CSeqFeatData::eSubtype_variation_ref:
        gui_info = new CGuiObjectInfoVariant();
        break;
    case CSeqFeatData::eSubtype_imp:
        gui_info = new CGuiObjectInfoImpFeature();
        break;
    case CSeqFeatData::eSubtype_region:
        gui_info = new CGuiObjectInfoRegionFeature();
        break;
    case CSeqFeatData::eSubtype_seq:
        gui_info = new CGuiObjectInfoSeqFeature();
        break;
    case CSeqFeatData::eSubtype_site:
        gui_info = new CGuiObjectInfoSiteFeature();
        break;
    default:
        gui_info = new CGuiObjectInfoSeq_feat();
        break;
    }

    gui_info->m_Feat.Reset(feat);
    gui_info->m_Scope  = object.scope;

    /// To guard against the cases where the feature contains location
    /// with multiple seq-ids
    const CSeq_feat_Base::TLocation& loc_obj = feat->GetLocation();
    if (loc_obj.GetId()) {
        gui_info->m_Location.Reset(&loc_obj);
    }

    return gui_info;
}


string CGuiObjectInfoSeq_feat::GetSubtype() const
{
    const CFeatList& feats(*CSeqFeatData::GetFeatList());
    CFeatListItem item;
    if (feats.GetItemBySubtype(m_Feat->GetData().GetSubtype(), item)) {
        return item.GetStoragekey();
    }
    return "";
}


string CGuiObjectInfoSeq_feat::GetLabel() const
{
    string label;
    CLabel::GetLabel(*m_Feat, &label, CLabel::eDescriptionBrief, m_Scope);
    return label;
}


void CGuiObjectInfoSeq_feat::GetToolTip(string& t_text, string& t_title,
                                        TSeqPos at_p) const
{
    t_text += CSeqUtils::CreateTableStart();
    CLabel::GetLabel(*m_Feat, &t_title, CLabel::eContent, m_Scope.GetPointer());

    t_text += CSeqUtils::CreateTableRow(GetSubtype(), t_title);

    const CSeq_feat::TData& data = m_Feat->GetData();
    string curr_text;

    if (CSeqUtils::IsPseudoFeature(*m_Feat)) {
        curr_text += "Pseudo";
    }
    if (CSeqUtils::IsPartialFeature(*m_Feat)) {
        curr_text += curr_text.empty() ? "" : ", ";
        curr_text += "Partial";
    }
    if (CSeqUtils::IsPartialStart(m_Feat->GetLocation())) {
        curr_text += curr_text.empty() ? "" : ", ";
        curr_text += "Partial start";
    }
    if (CSeqUtils::IsPartialStop(m_Feat->GetLocation())) {
        curr_text += curr_text.empty() ? "" : ", ";
        curr_text += "Partial stop";
    }
    if (CSeqUtils::IsException(*m_Feat)  &&  m_Feat->CanGetExcept_text()) {
        curr_text += curr_text.empty() ? "" : ", ";
        curr_text += m_Feat->GetExcept_text();
    }
    if ( !curr_text.empty() ) {
        t_text += CSeqUtils::CreateTableRow("Qualifiers", curr_text);
        curr_text.clear();
    }

    // Brief description goes here
    if (data.IsSeq()) {
        CLabel::GetLabel(*m_Feat, &curr_text, CLabel::eDescription, m_Scope.GetPointer());
    } else if (data.IsGene()  &&  data.GetGene().IsSetDesc()) {
        curr_text += data.GetGene().GetDesc();
    } else {
        feature::GetLabel(*m_Feat, &curr_text, feature::fFGL_Content, m_Scope.GetPointer());
    }
    if (curr_text != t_title) {
        // avoid showing the duplicated information
        t_text += CSeqUtils::CreateTableRow("Title", curr_text);
    }
    curr_text.clear();

    // Truncate the title if it is too long
    CLabel::TruncateLabel(&t_title, 30, 20); 

    // Location-related information
    x_AddLocationRows(t_text);

    // add portion specific to a specific sequence postion
    // such as current mouse position
    x_AddPositionRows(t_text, at_p);

    t_text += CSeqUtils::CreateTableEnd();;
}


void CGuiObjectInfoSeq_feat::GetLinks(string& links, bool no_ncbi_base) const
{
    if (m_Feat->IsSetDbxref()) {
        bool first_time = true;
        const CSeq_feat::TDbxref& refs = m_Feat->GetDbxref();
        string ncbi_base = CSeqUtils::GetNcbiBaseUrl();
        ITERATE (CSeq_feat::TDbxref, iter, refs) {
            string name, label, label_str;
            (*iter)->GetLabel(&label_str);
            size_t pos = label_str.find(":");
            if (pos == NPOS) {
                pos = label_str.rfind(" ");
            }

            if (pos != NPOS) {
                name = label_str.substr(0, pos);
                label = label_str.substr(pos + 1);
            } else {
                label = label_str;
            }

            string url_d = (*iter)->GetUrl();
            if (label.empty()  ||  url_d.empty()) {
                break;
            }

            if (no_ncbi_base  &&  NStr::StartsWith(url_d, ncbi_base)) {
                NStr::ReplaceInPlace(url_d, ncbi_base, "");
            }

            if ((*iter)->GetType() == CDbtag::eDbtagType_GeneID) {
                // Here we get Gene ID for a feature, so we can add Gene
                // Symbol. Unfortunately, Gene Symbol is valid only for
                // a feature of subtype eSubtype_gene, so we get best
                // gene here first.
                CConstRef<CSeq_feat> gene_symbol_feat(m_Feat.GetPointer());
                // First attempt - through GetBestGeneFor*
                switch (m_Feat->GetData().GetSubtype()) {
                    case CSeqFeatData::eSubtype_gene:
                        break;
                    case CSeqFeatData::eSubtype_cdregion:
                        gene_symbol_feat =
                            sequence::GetBestGeneForCds(*m_Feat, *m_Scope);
                        break;
                    case CSeqFeatData::eSubtype_mRNA:
                        gene_symbol_feat =
                            sequence::GetBestGeneForMrna(*m_Feat, *m_Scope);
                        break;
                    default:
                        gene_symbol_feat = NULL;
                        break;
                }
                if (gene_symbol_feat.Empty()) {
                    // Second attempt, if GetBestGeneFor*
                    // did not find anything, or feature is not CDS or
                    // mRNA and not a gene - get gene with this GeneID
                    int gene_id = (*iter)->GetTag().GetId();
                    SAnnotSelector sel(CSeqUtils::GetAnnotSelector());
                    sel.IncludeFeatSubtype(CSeqFeatData::eSubtype_gene);

                    const CSeq_loc* loc = NULL;
                    if (m_Location) {
                        loc = m_Location.GetPointer();
                    } else {
                        loc = &m_Feat->GetLocation();
                    }
                    for (CFeat_CI feat_iter(*m_Scope, *loc, sel);
                        feat_iter;
                        ++feat_iter)
                    {
                        CConstRef<CDbtag> tag =
                            feat_iter->GetNamedDbxref("GeneID");
                        if (tag  &&  tag->GetTag().GetId() == gene_id) {
                            gene_symbol_feat = feat_iter->GetSeq_feat();
                            break;
                        }
                    }
                }
                if (gene_symbol_feat.NotEmpty()) {
                    string gene_symbol;
                    feature::GetLabel(gene_symbol_feat.GetObject(),
                        &gene_symbol, feature::fFGL_Content);
                    if (!gene_symbol.empty()) {
                        label += " (" + gene_symbol + ")";
                    }
                }
            }

            if (first_time) {
                links += CSeqUtils::CreateTableStart();
                links += CSeqUtils::CreateSectionRow("Go To");
                first_time = false;
            }
            links += CSeqUtils::CreateLinkRow(name, label, url_d);
        }
        if ( !links.empty()) {
            links += CSeqUtils::CreateTableEnd();
        }
    }
}


void CGuiObjectInfoSeq_feat::x_AddLocationRows(string& t_text) const
{
    if (m_Location) {
        string curr_text;
        TSeqRange total_r = m_Location->GetTotalRange();
        ENa_strand strand = sequence::GetStrand(*m_Location);

        if (strand == eNa_strand_minus) {
            curr_text = "complement(";
        }
        curr_text += NStr::IntToString(total_r.GetFrom() + 1, NStr::fWithCommas);
        if (total_r.GetLength() > 1) {
            curr_text += "..";
            curr_text += NStr::IntToString(total_r.GetTo() + 1, NStr::fWithCommas);
        }
        if (strand == eNa_strand_minus) {
            curr_text += ")";
        }
        t_text += CSeqUtils::CreateTableRow("Location", curr_text);
        
        typedef pair<string, TSeqPos> TLenPair;
        typedef vector<TLenPair> TLens;
        TLens lens;
        lens.push_back(TLens::value_type("Span", total_r.GetLength()));

        if ( !m_Location->IsInt()  &&  !m_Location->IsPnt() ) {
            try {
                TSeqPos len = sequence::GetLength(*m_Location, m_Scope.GetPointer());
                lens.push_back(TLens::value_type("Placed", len));
            } catch (CException&) {
            }
        }

        if (m_Feat->IsSetProduct()) {
            //string tag;
            try {
                CBioseq_Handle h =
                    m_Scope->GetBioseqHandle(*m_Feat->GetProduct().GetId());
                //switch (m_Feat->GetData().GetSubtype())
                //{
                //case CSeqFeatData::eSubtype_mRNA:
                //    tag += "mRNA ";
                //    break;

                //case CSeqFeatData::eSubtype_tRNA:
                //    tag += "tRNA ";
                //    break;

                //case CSeqFeatData::eSubtype_rRNA:
                //    tag += "rRNA ";
                //    break;

                //case CSeqFeatData::eSubtype_cdregion:
                //    tag += "protein ";
                //    break;

                //default:
                //    break;
                //}

                //tag = "Product";
                lens.push_back(TLens::value_type("Product", h.GetBioseqLength()));
            } catch (CException&) {
            }
        }

        if (lens.size() > 1) {
            t_text += CSeqUtils::CreateSectionRow("Length");
            ITERATE (TLens, iter, lens) {
                curr_text = NStr::IntToString(iter->second, NStr::fWithCommas);
                t_text += CSeqUtils::CreateTableRow(iter->first, curr_text);
            }
        } else if (lens.size() == 1) {
            curr_text = NStr::IntToString(lens.begin()->second, NStr::fWithCommas);
            t_text += CSeqUtils::CreateTableRow("Length", curr_text);
        }
    }
}


void CGuiObjectInfoSeq_feat::x_AddPositionRows(string& t_text, TSeqPos at_p) const
{
    if (at_p != -1) {
        string extra_info;
        x_AddExtraInfo(extra_info, at_p);
        if (!extra_info.empty()) {
            t_text += CSeqUtils::CreateSectionRow("Positional Info");
        }
        t_text += CSeqUtils::CreateTableRow("Position",
            NStr::UIntToString(at_p + 1, NStr::fWithCommas));
        t_text += extra_info;
    }
}


static const int kSeqTooltipLength = 15;

void CGuiObjectInfoSeq_feat::x_AddExtraInfo(string& t_text, TSeqPos at_p) const
{
    if ( !m_Location ) return;

    CSeqFeatData::E_Choice type = m_Feat->GetData().Which();
    if (type == CSeqFeatData::e_Rna  ||  type == CSeqFeatData::e_Cdregion) {
        SFeatSeg seg = x_GetHitSeg(at_p);
        if (seg.m_Type != SFeatSeg::eType_OutOfBound) {
            x_GetSplicedSeqTitle(seg, at_p, t_text);
            if (type == CSeqFeatData::e_Cdregion  &&  m_Feat->IsSetProduct()) {
                x_GetProductSequence(seg, at_p, t_text);
            }
        }
    }
}


/*
bool CGuiObjectInfoSeq_feat::x_GetIntron2(TSeqPos pos,
                                         TSeqRange& range) const
{
    TSeqPos total_f, total_t;
    vector<TSeqRange> introns;
    int direct = 0; // unknown
    TSeqRange pre_r(1, 0);
    for (CSeq_loc_CI iter(*m_Location);  iter;  ++iter) {
        CSeq_loc_CI::TRange range = iter.GetRange();
        if (pre_r.Empty()) {
            pre_r = range;
            total_f = range.GetFrom();
            total_t = range.GetTo();
            if (iter.GetStrand() == eNa_strand_minus) {
                direct = -1;
            } else {
                direct = 1;
            }
            continue;
        }

        if (direct > 0) {
            total_t = range.GetTo();
            if (range.GetFrom() > pre_r.GetToOpen()  ||
                pre_r.GetFrom() > range.GetToOpen()) {
                introns.push_back(TSeqRange(pre_r.GetTo(), range.GetFrom()));
            }
        } else {
            total_f = range.GetFrom();
            if (pre_r.GetFrom() > range.GetToOpen()  ||
                range.GetFrom() > pre_r.GetToOpen()) {
                introns.push_back(TSeqRange(range.GetTo(), pre_r.GetFrom()));
            }
        }
        pre_r = range;
    }

    ITERATE (vector<TSeqRange>, i_iter, introns) {
        if (i_iter->Empty()) {
            // intron crossing origin
            if (pos > i_iter->GetFrom()  ||  pos < i_iter->GetTo()) {
                range = *i_iter;
                return true;
            }
        } else {
            if (pos > i_iter->GetFrom()  &&  pos < i_iter->GetTo()) {
                range = *i_iter;
                return true;
            }
        }
    }

    // return total range
    range.Set(total_f, total_t);
    return false;
}
*/

CGuiObjectInfoSeq_feat::SFeatSeg
CGuiObjectInfoSeq_feat::x_GetHitSeg(TSeqPos pos) const
{
    SFeatSeg seg;
    int pre_dir = 0;      // Previous segment strand (0:unknow, 1:forward, -1:reverse)
    TSeqPos pre_pos = -1; // Previous segment end position close to the current segment
    TSeqPos prod_start = 0; // Current product start pos (in nucleotide bases)

    for (CSeq_loc_CI iter(*m_Location);  iter;  ++iter) {
        CSeq_loc_CI::TRange range = iter.GetRange();
        int curr_dir = 1;
        if (iter.GetStrand() == eNa_strand_minus) {
            curr_dir = -1;
        }

        // check if it hits an exon
        if (pos >= range.GetFrom()  &&  pos <= range.GetTo()) {
            seg.m_Type = SFeatSeg::eType_Exon;
            seg.m_CrossOrigin = false;
            seg.m_Range = range;
            seg.m_ProdRange.Set(prod_start, prod_start + range.GetLength() - 1);
            seg.m_ExonDirForward = curr_dir > 0;
            break;
        }

        // initialize current segment strand
        if (pre_pos != -1) {
            // check if it hits an intron
            TSeqPos curr_pos = curr_dir > 0 ? range.GetFrom() : range.GetTo();

            if (pre_dir > 0  &&  curr_pos > pre_pos) {
                if (pos > pre_pos  &&  pos < curr_pos  &&
                    (seg.m_Range.Empty()  ||  seg.m_Range.GetLength() > curr_pos - pre_pos + 1)) {
                    seg.m_Range.Set(pre_pos, curr_pos);
                    seg.m_Type = SFeatSeg::eType_Intron;
                    seg.m_CrossOrigin = false;
                    seg.m_ProdRange.Set(prod_start - 1, prod_start);
                    seg.m_ExonDirForward = pre_dir > 0;
                }
            } else if (pre_dir < 0  &&  curr_pos < pre_pos) {
                if (pos < pre_pos  &&  pos > curr_pos  &&
                    (seg.m_Range.Empty()  ||  seg.m_Range.GetLength() > pre_pos - curr_pos + 1)) {
                    seg.m_Range.Set(curr_pos, pre_pos);
                    seg.m_Type = SFeatSeg::eType_Intron;
                    seg.m_CrossOrigin = false;
                    seg.m_ProdRange.Set(prod_start - 1, prod_start);
                    seg.m_ExonDirForward = pre_dir > 0;
                }
            } else {
                // The intron must cross the origin
                // It is intentional to make the 'from' > 'to' for an intron span
                // crossing the origin.
                if (curr_pos > pre_pos  &&  (pos > curr_pos || pos < pre_pos)) {
                    seg.m_Range.Set(curr_pos, pre_pos);
                    seg.m_Type = SFeatSeg::eType_Intron;
                    seg.m_CrossOrigin = true;
                    seg.m_ProdRange.Set(prod_start - 1, prod_start);
                    seg.m_ExonDirForward = pre_dir > 0;
                } else if (curr_pos < pre_pos  &&  (pos > pre_pos || pos < curr_pos)) {
                    seg.m_Range.Set(pre_pos, curr_pos);
                    seg.m_Type = SFeatSeg::eType_Intron;
                    seg.m_CrossOrigin = true;
                    seg.m_ProdRange.Set(prod_start - 1, prod_start);
                    seg.m_ExonDirForward = pre_dir > 0;
                }
            }
        }

        // backup current segmnet attributes
        pre_pos = curr_dir > 0 ? range.GetTo() : range.GetFrom(); 
        prod_start += range.GetLength();
        pre_dir = curr_dir;
    }

    return seg;
}


// This is exactly as it done in Sequin.
TSeqPos CGuiObjectInfoSeq_feat::x_GetProtOffset() const
{
    TSeqPos offset = 1; // translation offset
    const CCdregion& cdr = m_Feat->GetData().GetCdregion();
    if (cdr.IsSetFrame ()) {
        switch (cdr.GetFrame ()) {
        case CCdregion::eFrame_two :
            offset = 2;
            break;
        case CCdregion::eFrame_three :
            offset = 3;
            break;
        default :
            break;
        }
    }
    return offset;
}


TSeqPos CGuiObjectInfoSeq_feat::x_ToProduct(TSeqPos pos) const
{
    TSeqPos offset  = x_GetProtOffset();  // initial translation offset
    return pos / 3 + offset - 1;
}


void CGuiObjectInfoSeq_feat::x_GetSplicedSeqTitle(const SFeatSeg& seg,
                                                  TSeqPos at_p,
                                                  string& t_text) const
{
    CBioseq_Handle top_handle = m_Scope->GetBioseqHandle(*m_Location->GetId());
    if ( !top_handle ) {
        return;
    }
    CSeqVector vec = top_handle.GetSeqVector(CBioseq_Handle::eCoding_Iupac);
    const TSeqPos kSeqExonLength   = 10;
    const TSeqPos kSeqIntronLength = 8;

    if (seg.m_Type == SFeatSeg::eType_Intron) {
        // add detailed title for intron
        try {
            string value;
            const TSeqRange& intr = seg.m_Range;
            TSeqPos left_stop  = intr.GetFrom() + 1;
            TSeqPos left_start  = (TSeqPos) (left_stop > kSeqExonLength ?
                left_stop - kSeqExonLength : 0);;
            string seq_left, seq_ins_l, seq_ins_r, seq_right;

            vec.GetSeqData(left_start, left_stop, seq_left);
            TSeqPos right_start = intr.GetTo();

            value = seq_left + "] ";
            if (seg.m_CrossOrigin) {
                vec.GetSeqData(left_stop, left_stop + kSeqIntronLength, seq_ins_l);
                value += seq_ins_l;
                if (vec.size() > left_stop + kSeqIntronLength) {
                    value += "...";
                }
                value += "[origin]";
                if (right_start < kSeqIntronLength) {
                    vec.GetSeqData(0, right_start, seq_ins_r);
                } else {
                    value += "...";
                    vec.GetSeqData(right_start - kSeqIntronLength, right_start, seq_ins_r);
                }
                value += seq_ins_r;
            } else if (intr.GetLength() > kSeqIntronLength * 2) {
                vec.GetSeqData(left_stop, left_stop + kSeqIntronLength, seq_ins_l);
                vec.GetSeqData(right_start - kSeqIntronLength, right_start, seq_ins_r);
                value += seq_ins_l + "..." + seq_ins_r;
            } else {
                vec.GetSeqData(left_stop, right_start + 1, seq_ins_l);
                value += seq_ins_l;
            }
            vec.GetSeqData(right_start, right_start + kSeqExonLength, seq_right);
            value += " [" + seq_right;
            t_text += CSeqUtils::CreateTableRow("Spliced sequence", value);
        }
        catch (CException&) {
            /// ignore exceptions
            t_text += CSeqUtils::CreateTableRow("Spliced sequence", "Not available");
        }
    } else if (seg.m_Type == SFeatSeg::eType_Exon) {
        string value;
        TSeqPos left  = (TSeqPos) (at_p > kSeqIntronLength ? at_p - kSeqIntronLength : 0);
        TSeqPos right = (TSeqPos) (at_p + kSeqIntronLength + 1);
        vec.GetSeqData(left, at_p, value);

        string seq_pos;
        vec.GetSeqData(at_p, (TSeqPos)(at_p + 1), seq_pos);

        string seq_r;
        vec.GetSeqData((TSeqPos)(at_p + 1), right, seq_r);

        value += "[";
        value += seq_pos;
        value += "]";
        value += seq_r;
        t_text += CSeqUtils::CreateTableRow("Sequence", value);
    }
}

void CGuiObjectInfoSeq_feat::x_GetProductSequence(const SFeatSeg& seg,
                                                  TSeqPos at_p,
                                                  string& t_text) const
{
    try {
        const CSeq_loc& product = m_Feat->GetProduct();
        CSeqVector prot_vec(product, *m_Scope, CBioseq_Handle::eCoding_Iupac);

        // are we in protein product gap?
        bool in_prot_gap = false;

        // the easiest way to find out if we are in the protein sequence
        // gap is to compare prot. sequence positions on both sides of the
        // inron. If they are diffenet, we must be in the gap.
        // Otherwise it's not the gap

        TSignedSeqPos prot_pos = 0;

        TSignedSeqPos prot_pos1 = x_ToProduct(seg.m_ProdRange.GetFrom());
        TSignedSeqPos prot_pos2 = x_ToProduct(seg.m_ProdRange.GetTo());
        const TSeqRange& intr = seg.m_Range;

        // first, if at_p is in the intron, move it to one end of the nearby exon
        if (seg.m_Type == SFeatSeg::eType_Intron) {
            if (prot_pos1 != prot_pos2) {
                in_prot_gap = true;

                bool at_from_end = false; 
                if ((seg.m_CrossOrigin  &&  at_p > intr.GetFrom())  ||
                    (!seg.m_CrossOrigin  &&  at_p - intr.GetFrom() < intr.GetTo() - at_p)) {
                        at_from_end = true;
                }

                if (at_from_end == seg.m_ExonDirForward) {
                    prot_pos = prot_pos1;
                } else {
                    prot_pos = prot_pos2;
                }
            } else {
                prot_pos = prot_pos1;
            }
        } else { // in an exon
            if (seg.m_ExonDirForward) {
                at_p = seg.m_ProdRange.GetFrom() + at_p - intr.GetFrom();
            } else {
                at_p = seg.m_ProdRange.GetFrom() + intr.GetTo() - at_p;
            }
            // the real position in protein sequence
            prot_pos = x_ToProduct(at_p);
        }

        TSeqPos from = prot_pos > kSeqTooltipLength ? prot_pos - kSeqTooltipLength : 0;
        TSeqPos to = min(prot_pos + kSeqTooltipLength, int(prot_vec.size()));

        t_text += CSeqUtils::CreateTableRow("Product Position",
            NStr::UIntToString(prot_pos + 1, NStr::fWithCommas));

        string value;
        string prot_seq;
        prot_vec.GetSeqData(from, prot_pos, prot_seq);
        value += prot_seq;
        prot_vec.GetSeqData(prot_pos, prot_pos + 1, prot_seq);
        string first_bracket = "[";
        string second_bracket = "]";
        if ( !seg.m_ExonDirForward ) {
            std::swap(first_bracket, second_bracket);
        }

        if (in_prot_gap) {
            if (prot_pos == prot_pos1) {
                value += prot_seq + first_bracket + second_bracket;
            } else {
                value += first_bracket + second_bracket + prot_seq;
            }
        } else {
            value += first_bracket + prot_seq + second_bracket;
        }

        prot_vec.GetSeqData(prot_pos + 1, to, prot_seq);
        value += prot_seq;
        if ( !seg.m_ExonDirForward ) {
            std::reverse(value.begin(), value.end());
        }
        t_text += CSeqUtils::CreateTableRow("Product sequence", value);
    } catch (CException&) {
    }
}
    

END_NCBI_SCOPE
