/*  $Id: gui_object_info_variant.cpp 27958 2013-04-30 19:33:45Z wuliangs $
* ===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data, the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties, express or implied, including
*  warranties of performance, merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================
*
* Authors: Liangshou Wu
*
* File Description:
*     Gui Object Info class specifically for structural variations from dbVar.
*
*/

#include <objects/seqfeat/Variation_ref.hpp>
#include <objects/seqfeat/Variation_inst.hpp>
#include <objects/seqfeat/Delta_item.hpp>
#include <objects/seqfeat/VariantProperties.hpp>
#include <objects/seqfeat/Phenotype.hpp>
#include <objects/seqfeat/SubSource.hpp>
#include <objects/seq/Seq_literal.hpp>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

class CGuiObjectInfoVariant : public CGuiObjectInfoSeq_feat
{
public:
    virtual string GetSubtype() const;
    virtual void GetToolTip(string& t_text, string& t_title,
        TSeqPos at_p = (TSeqPos)-1) const;
    virtual void GetLinks(string& links, bool no_ncbi_base) const;
};


string CGuiObjectInfoVariant::GetSubtype() const
{
    return "Structural variant";
}


static int s_GetChildNum(const objects::CSeq_feat& feat)
{
    string num_str = feat.GetNamedQual("Child Count");
    if ( !num_str.empty() ) {
        try {
            return NStr::StringToInt(num_str);
        } catch (CException&) {
            // ingore it
        }
    }
    return -1;
}


void CGuiObjectInfoVariant::GetToolTip(string& t_text, string& t_title,
                                        TSeqPos /*at_p*/) const
{
    CLabel::GetLabel(*m_Feat, &t_title, CLabel::eContent, m_Scope.GetPointer());
    t_text += CSeqUtils::CreateTableStart();

    const CVariation_ref& var = m_Feat->GetData().GetVariation();
    string validation = "Not Tested";
    string additional_info = kEmptyStr;
    if (var.IsSetValidated()) {
        if (var.GetValidated()) {
            validation = "Yes";
        } else {
            validation = "Fail";
        }
    } else if (var.IsSetVariant_prop()  &&  var.GetVariant_prop().IsSetOther_validation()) {
        if (var.GetVariant_prop().GetOther_validation()) {
            validation = "Yes";
        } else {
            validation = "Fail";
        }
    } else if (m_Feat->IsSetExts()) {
        const CSeq_feat::TExts& exts = m_Feat->GetExts();
        ITERATE (CSeq_feat::TExts, iter, exts) {
            if ( (*iter)->GetType().IsStr()  &&
                NStr::EqualNocase((*iter)->GetType().GetStr(), "Validation")  &&
                 (*iter)->GetFieldRef("Status")  &&
                 (*iter)->GetFieldRef("Status")->GetData().IsStr()) {
                     validation = (*iter)->GetFieldRef("Status")->GetData().GetStr();
                     CConstRef<CUser_field> a_field = (*iter)->GetFieldRef("Addition");
                     if (a_field) {
                        additional_info = a_field->GetData().GetStr();
                     }
                     break;
            }
        }
    }

    // phenotype and clinical-significance
    string phenotype = "";
    string clinical_sig = "";
    if (var.CanGetPhenotype()) {
        ITERATE (CVariation_ref::TPhenotype, pnt_iter, var.GetPhenotype()) {
            if (clinical_sig.empty()  &&  (*pnt_iter)->CanGetClinical_significance()) {
                switch ((*pnt_iter)->GetClinical_significance())
                {
                case CPhenotype::eClinical_significance_non_pathogenic:
                    clinical_sig = "Benign";
                    break;
                case CPhenotype::eClinical_significance_pathogenic:
                    clinical_sig = "Pathogenic";
                    break;
                case CPhenotype::eClinical_significance_probable_pathogenic:
                    clinical_sig = "Uncertain Significance - likely pathogenic";
                    break;
                case CPhenotype::eClinical_significance_probable_non_pathogenic:
                    clinical_sig = "Uncertain Significance - likely benign";
                    break;
                case CPhenotype::eClinical_significance_unknown:
                    clinical_sig = "Uncertain Significance";
                    break;
                case CPhenotype::eClinical_significance_untested:
                    clinical_sig = "Not tested";
                    break;
                case CPhenotype::eClinical_significance_other:
                default:
                    clinical_sig = "Other";
                    break;
                }
            }

            if (phenotype.empty()  &&  (*pnt_iter)->CanGetTerm()) {
                phenotype = (*pnt_iter)->GetTerm();
            }
        }
    }

    string sample_sex = "";
    if (var.CanGetSomatic_origin()) {
        ITERATE (CVariation_ref::TSomatic_origin, so_iter, var.GetSomatic_origin()) {
            if ((*so_iter)->CanGetSource()  &&
                (*so_iter)->GetSource().GetSubtype() == CSubSource::eSubtype_sex) {
                    sample_sex = (*so_iter)->GetSource().GetName();
                    break;
            }
        }
    }

    bool is_sv = !var.IsSetId()  ||  (var.GetId().GetTag().IsStr()  &&
            var.GetId().GetTag().GetStr().find("sv") == 1);

    // allele type
    string variant_type;
    int multiplier = -1;
    if (var.IsComplex()) {
        variant_type = "Complex";
    } else if (var.IsInsertion()) {
        variant_type = "Insertion";
    } else if (var.IsInversion()) {
        variant_type = "Inversion";
    } else if (var.IsEversion()) {
        variant_type = "Tandem duplication";
    } else if (var.IsTranslocation()) {
        variant_type = "Translocation";
    } else if (var.IsDeletion()) {
        variant_type = "Deletion";
    } else if (var.IsCNV()) {
        if (var.GetData().GetInstance().IsSetDelta()) {
            const CVariation_inst::TDelta& delta = var.GetData().GetInstance().GetDelta();
            ITERATE (CVariation_inst::TDelta, iter, delta) {
                if ((*iter)->IsSetMultiplier()) {
                    multiplier = (*iter)->GetMultiplier();
                    break;
                }
            }
        }
        if (is_sv) {
            variant_type = "Copy number variation";
        } else {
            if (var.IsGain()) {
                variant_type = "Copy number gain";
            } else if (var.IsLoss()) {
                variant_type = "Copy number loss";
            }
        }
    } else { // everything else is treated as 'unknown' regardless
        variant_type = "Unknown";
    }

    if (is_sv) {
        int child_num = s_GetChildNum(*m_Feat);
        if (child_num == 0) {
            // sv, but treated as ssv
            is_sv = false;
        } else {
            if (child_num > 0) {
                t_text += CSeqUtils::CreateTableRow(
                    "Supporting Variant Calls", NStr::IntToString(child_num));
            }
            t_text += CSeqUtils::CreateTableRow("Variant Region Type", variant_type);
            if (multiplier > -1) {
                t_text += CSeqUtils::CreateTableRow("Copy number", NStr::NumericToString(multiplier));
            }
            if ( !phenotype.empty() ) {
                t_text += CSeqUtils::CreateTableRow("Phenotype", phenotype);
            }
            if ( !clinical_sig.empty() ) {
                t_text += CSeqUtils::CreateTableRow("Clinical Interpretation", clinical_sig);
            }
            if ( !sample_sex.empty() ) {
                t_text += CSeqUtils::CreateTableRow("Gender", sample_sex);
            }
            t_text += CSeqUtils::CreateTableRow("Validation Status", validation);
            if ( !additional_info.empty() ) {
                t_text += CSeqUtils::CreateTableRow("Additional info", additional_info);
            }
            if (m_Location) {
                t_text += CSeqUtils::CreateTableRow("Total Length",
                    NStr::IntToString(m_Location->GetTotalRange().GetLength(), NStr::fWithCommas));
            }
        }
    }

    if ( !is_sv ) {
        if (var.CanGetParent_id()) {
            string parent;
            if (var.GetParent_id().GetTag().IsId()) {
                parent = NStr::IntToString(var.GetParent_id().GetTag().GetId());
            } else {
                parent = var.GetParent_id().GetTag().GetStr();
            }
            t_text += CSeqUtils::CreateTableRow("Parent Variant Region", parent);
        }

        t_text += CSeqUtils::CreateTableRow("Variant Call Type", variant_type);
        if (multiplier > -1) {
            t_text += CSeqUtils::CreateTableRow("Copy number", NStr::NumericToString(multiplier));
        }
        if ( !phenotype.empty() ) {
            t_text += CSeqUtils::CreateTableRow("Phenotype", phenotype);
        }
        if ( !clinical_sig.empty() ) {
            t_text += CSeqUtils::CreateTableRow("Clinical Interpretation", clinical_sig);
        }
        if ( !sample_sex.empty() ) {
            t_text += CSeqUtils::CreateTableRow("Gender", sample_sex);
        }

        // inheritance
        CVariation_ref::TAllele_origin org = -1;
        if (var.IsSetAllele_origin()) {
            org = var.GetAllele_origin();
        } else if (var.IsSetVariant_prop()  &&  var.GetVariant_prop().IsSetAllele_origin()) {
            org = var.GetVariant_prop().GetAllele_origin();
        }

        if (org > 0) {
            string inh;
            if (org & CVariation_ref::eAllele_origin_somatic) {
                inh += "somatic ";
            }
            if (org & CVariation_ref::eAllele_origin_inherited) {
                inh += "inherited ";
            }
            if (org & CVariation_ref::eAllele_origin_paternal) {
                inh += "paternal ";
            }
            if (org & CVariation_ref::eAllele_origin_maternal) {
                inh += "maternal ";
            }
            if (org & CVariation_ref::eAllele_origin_de_novo) {
                inh += "de_novo ";
            }
            if (org & CVariation_ref::eAllele_origin_biparental) {
                inh += "biparental ";
            }
            if (org & CVariation_ref::eAllele_origin_uniparental) {
                inh += "uniparental ";
            }
            if (org & CVariation_ref::eAllele_origin_not_tested) {
                inh += "not-tested ";
            }
            if (org & CVariation_ref::eAllele_origin_tested_inconclusive) {
                inh += "tested-inconclusive ";
            }
            if (org & CVariation_ref::eAllele_origin_other) {
                inh += "other";
            }

            if (org == CVariation_ref::eAllele_origin_unknown) {
                inh = "unknown";
            }

            t_text += CSeqUtils::CreateTableRow("Inheritance", inh);
        }

        // validation status
        t_text += CSeqUtils::CreateTableRow("Validation Status", validation);
        if ( !additional_info.empty() ) {
            t_text += CSeqUtils::CreateTableRow("Additional info", additional_info);
        }

        if (m_Feat->CanGetComment()) {
            t_text += CSeqUtils::CreateTableRow("Comment", m_Feat->GetComment());
        }
    }

    int allele_len = -1;
    if (var.GetData().IsInstance()  &&  var.GetData().GetInstance().IsSetDelta()) {
        const CVariation_inst::TDelta& delta = var.GetData().GetInstance().GetDelta();
        ITERATE (CVariation_inst::TDelta, iter, delta) {
            if ((*iter)->IsSetSeq()  &&  (*iter)->GetSeq().IsLiteral()) {
                allele_len = (*iter)->GetSeq().GetLiteral().GetLength();
                break;
            }
        }
    }
    if (allele_len > -1) {
        t_text += CSeqUtils::CreateTableRow("Allele length",
            NStr::NumericToString(allele_len, NStr::fWithCommas));
    }
    if (var.CanGetSample_id()) {
        string sample_id;
        const CVariation_ref::TSample_id& sid = var.GetSample_id();
        if (sid.IsStr()) {
            sample_id = sid.GetStr();
        } else {
            sample_id = NStr::NumericToString(sid.GetId());
        }
        if ( !sample_id.empty() ) {
            t_text += CSeqUtils::CreateTableRow("Sample id", sample_id);
        }
    }
    t_text += CSeqUtils::CreateTableEnd();
}

static const string kBaseUrl = "http://ncbi.nlm.nih.gov/dbvar/";

void CGuiObjectInfoVariant::GetLinks(string& links, bool no_ncbi_base) const
{
    const CVariation_ref& var = m_Feat->GetData().GetVariation();
    string table_rows = kEmptyStr;
    string label, name, url;
    if (var.CanGetId()  &&  var.GetId().GetTag().IsStr()  &&
        var.GetId().GetTag().GetStr().find("sv") == 1) {
            CLabel::GetLabel(*m_Feat, &label, CLabel::eContent, m_Scope.GetPointer());
    } else if (var.CanGetParent_id()) {
        if (var.GetParent_id().GetTag().IsId()) {
            label = NStr::IntToString(var.GetParent_id().GetTag().GetId());
        } else {
            label = var.GetParent_id().GetTag().GetStr();
        }
    }

    if ( !label.empty() ) {
        name = "dbVar";
        url = kBaseUrl + "variants/" + label;
        table_rows += CSeqUtils::CreateLinkRow("dbVar", label, url);
    }

    if (m_Feat->IsSetExts()) {
        ITERATE (CSeq_feat::TExts, iter, m_Feat->GetExts()) {
            if ((*iter)->GetType().IsStr()  &&
                !(*iter)->GetData().empty()) {
                    name = "Other Variant Calls from this Sample";
                    if ((*iter)->GetType().GetStr() == "related calls") {
                        const CUser_object::TData& fields = (*iter)->GetData();
                        if (fields.size() < 5) {
                            ITERATE (CUser_object::TData, f_iter, fields) {
                                label = (*f_iter)->GetLabel().GetStr();
                                url = kBaseUrl + "variants/" + (*f_iter)->GetData().GetStr();
                                table_rows += CSeqUtils::CreateLinkRow(name, label, url);
                            }
                        } else if (var.CanGetSample_id()) {
                            string term;
                            if (var.GetSample_id().IsStr()) {
                                term = var.GetSample_id().GetStr();
                            } else {
                                term = NStr::NumericToString(var.GetSample_id().GetId());
                            }
                            CConstRef<CDbtag> study_id = m_Feat->GetNamedDbxref("study_accession");
                            if (study_id) {
                                term += " and " + study_id->GetTag().GetStr();
                            }

                            label = NStr::NumericToString(fields.size());
                            url = kBaseUrl + "?term=" + term;
                            table_rows += CSeqUtils::CreateLinkRow(name, label, url);
                        }
                        break;
                    } else if ((*iter)->GetType().GetStr() == "related call count") {
                        string term;
                        ITERATE (CUser_object::TData, f_iter, (*iter)->GetData()) {
                            if ((*f_iter)->GetLabel().IsStr()  &&  (*f_iter)->GetData().IsStr()) {
                                label = (*f_iter)->GetLabel().GetStr();
                                term = (*f_iter)->GetData().GetStr();
                                break;
                            }
                        }

                        if ( !term.empty() ) {
                            CConstRef<CDbtag> study_id = m_Feat->GetNamedDbxref("study_accession");
                            if (study_id) {
                                term += " and " + study_id->GetTag().GetStr();
                            }
                            url = kBaseUrl + "?term=" + term;
                            table_rows += CSeqUtils::CreateLinkRow(name, label, url);
                        }
                        break;
                    }
            }
        }
    }

    if ( !table_rows.empty()) {
        links += CSeqUtils::CreateTableStart();
        links += CSeqUtils::CreateSectionRow("Go To");
        links += table_rows;
        links += CSeqUtils::CreateTableEnd();
    }
}


END_NCBI_SCOPE