/*  $Id: gui_object_info_variation.cpp 27866 2013-04-16 17:23:36Z rudnev $
* ===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data, the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties, express or implied, including
*  warranties of performance, merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================
*
* Authors: Liangshou Wu, Dmitry Rudnev
*
* File Description:
*     Gui Object Info class specifically for Single-nucleotide polymorphism
*  commonly known as SNP.
* 
*/

#include <objects/variation/Variation.hpp>
#include <gui/objutils/tooltip.hpp>
#include <objtools/snputil/snp_bitfield.hpp>
#include <objtools/snputil/snp_utils.hpp>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

class CGuiObjectInfoVariation : public CGuiObjectInfoSeq_feat
{
public:
    virtual string GetSubtype() const;
    virtual void GetToolTip(string& t_text, string& t_title,
        TSeqPos at_p = (TSeqPos)-1) const;
    virtual void GetLinks(string& links, bool no_ncbi_base) const;
};


string CGuiObjectInfoVariation::GetSubtype() const
{
    return "SNP";
}


void CGuiObjectInfoVariation::GetToolTip(string& t_text, string& t_title,
                                         TSeqPos /*at_p*/) const
{
    CLabel::GetLabel(*m_Feat, &t_title, CLabel::eContent, m_Scope.GetPointer());

    CConstRef<CDbtag> ref = m_Feat->GetNamedDbxref("dbSNP");
    CSnpBitfield bitfield(NSnp::GetBitfield(*m_Feat));
    CIRef<ITooltipFormatter> pFormatter(
        ITooltipFormatter::CreateTooltipFormatter(
        ITooltipFormatter::eTooltipFormatter_Html));
    bool isSNPFeature(ref);
    if (isSNPFeature) {
        string sVariantType(bitfield.GetVariationClassString());
        pFormatter->StartRow();
        pFormatter->AddTagCol("Variation ID:");
        pFormatter->AddValueCol("rs" + NStr::IntToString(ref->GetTag().GetId()));
        pFormatter->FinishRow();
        pFormatter->StartRow();
        pFormatter->AddTagCol("Variant Type:");
		string sLength(NStr::IntToString(m_Feat->GetLocation().GetTotalRange().GetLength(), NStr::fWithCommas));
        pFormatter->AddValueCol(sVariantType + ", length " + sLength);
        pFormatter->FinishRow();
    } else {
        pFormatter->StartRow();
        pFormatter->AddTagCol("Variation:");
        string sLabel;
        feature::GetLabel(*m_Feat, &sLabel,
            feature::fFGL_Type |
            feature::fFGL_Content, m_Scope.GetPointer());
        pFormatter->AddValueCol(sLabel);
        pFormatter->FinishRow();
    }

    if (m_Feat->IsSetQual()) {
        ///
        /// alleles listed in the replace string come next
        string alleles;
        ITERATE (CSeq_feat::TQual, iter, m_Feat->GetQual()) {
            const CGb_qual& qual = **iter;
            if (qual.GetQual() == "replace") {
                if ( !alleles.empty() ) {
                    alleles += '/';
                }
                if (qual.GetVal().empty()) {
                    alleles += '-';
                } else {
                    alleles += qual.GetVal();
                }
            }
        }
        if ( !alleles.empty() ) {
            pFormatter->StartRow();
            pFormatter->AddTagCol("Alleles:");
            pFormatter->AddValueCol(alleles);
            pFormatter->FinishRow();
        }
    }
    if(bitfield.IsTrue(CSnpBitfield::eIsSomatic)) {
        pFormatter->StartRow();
        pFormatter->AddTagCol("Allele origin:");
        pFormatter->AddValueCol("Somatic");
        pFormatter->FinishRow();
    }
    t_text += pFormatter->Render();
}


void CGuiObjectInfoVariation::GetLinks(string& links, bool no_ncbi_base) const
{
	CIRef<ITooltipFormatter> pFormatter(
		ITooltipFormatter::CreateTooltipFormatter(
		ITooltipFormatter::eTooltipFormatter_Html));
    bool isAnyLinks(false);
    if (m_Feat->IsSetDbxref()) {
        string base_url = no_ncbi_base ? "" : CSeqUtils::GetNcbiBaseUrl();
        const CSeq_feat::TDbxref& refs = m_Feat->GetDbxref();
        pFormatter->StartRow();
        pFormatter->AddTagCol("Go to:");
        pFormatter->FinishRow();

		ITERATE (CSeq_feat::TDbxref, iter, refs) {
            const CRef<CDbtag>& dbt = *iter;
            if (dbt->GetType() == CDbtag::eDbtagType_dbSNP) {
                int rsid = dbt->GetTag().GetId();
				isAnyLinks = true;
                ///
                /// link for SNP summary
                ///
				string name = "SNP summary:";
                string label = "rs";
                label += NStr::IntToString(rsid);
                string link = base_url + "/projects/SNP/snp_ref.cgi?rs=";
                link += NStr::IntToString(rsid);

				pFormatter->StartRow();
				pFormatter->AddTagCol(name);
				pFormatter->AddValueCol(pFormatter->CreateLink(label, link));
				pFormatter->FinishRow();

                if (m_Feat->IsSetExt()) {
                    CConstRef<CUser_field> field =
                        m_Feat->GetExt().GetFieldRef("QualityCodes");
                    if (field) {
                        CSnpBitfield bf(field->GetData().GetOs());

                        ///
                        /// link for Genotype information
                        ///
                        if (bf.IsTrue(CSnpBitfield::eHasGenotype)) {
                            name = "Genotype information for";
                            label = "rs";
                            label += NStr::IntToString(rsid);
                            link = base_url + "/projects/SNP/snp_ref.cgi?rs=";
                            link += NStr::IntToString(rsid) + "#Diversity";
							pFormatter->StartRow();
							pFormatter->AddTagCol(name);
							pFormatter->AddValueCol(pFormatter->CreateLink(label, link));
							pFormatter->FinishRow();
                        }

                        ///
                        /// link for SNP3D web page
                        ///
                        if (bf.IsTrue(CSnpBitfield::eHasSnp3D)) {
                            name = "SNP3D Page for";
                            label = "rs" + NStr::IntToString(rsid);
                            link = base_url + "/projects/SNP/snp3D.cgi?rsnum=";
                            link += NStr::IntToString(rsid);
							pFormatter->StartRow();
							pFormatter->AddTagCol(name);
							pFormatter->AddValueCol(pFormatter->CreateLink(label, link));
							pFormatter->FinishRow();
                        }

                        ///
                        /// link for OMIM page for this SNP
                        ///
                        if (bf.IsTrue(CSnpBitfield::eHasOMIM_OMIA)) {
                            name = "OMIM information for";
                            label = "rs" + NStr::IntToString(rsid);
                            link = base_url + "/projects/SNP/snp_redirect.cgi?snp_id=";
                            link += NStr::IntToString(rsid);
							pFormatter->StartRow();
							pFormatter->AddTagCol(name);
							pFormatter->AddValueCol(pFormatter->CreateLink(label, link));
							pFormatter->FinishRow();
                        }

                        // VarVu link for clinical SNPs
                        if(bf.IsTrue(CSnpBitfield::eIsClinical)) {
                            // try to match genes to the current position
                            SAnnotSelector sel_gene(CSeqUtils::GetAnnotSelector());
                            sel_gene.IncludeFeatSubtype(CSeqFeatData::eSubtype_gene);

                            if(m_Feat->CanGetLocation()) {
                                CFeat_CI feat_iter(*m_Scope, m_Feat->GetLocation(), sel_gene);

                                CConstRef<CSeq_feat> mapped_feat;
                                for ( ;  feat_iter;  ++feat_iter) {
                                    mapped_feat.Reset(&feat_iter->GetMappedFeature());
                                    string sGeneLocus;
                                    if(mapped_feat->GetData().Which() == CSeqFeatData::e_Gene  &&
                                        mapped_feat->GetData().GetGene().IsSetLocus()) {
                                            sGeneLocus = mapped_feat->GetData().GetGene().GetLocus();
                                    }

                                    if( !sGeneLocus.empty() ) {
                                        name = "Variation viewer for";
                                        label = "rs" + NStr::IntToString(rsid);
                                        label += " (" + sGeneLocus + ")",
                                            link = base_url + "/sites/varvu?rs=";
                                        link += NStr::IntToString(rsid) + "&gene=" + sGeneLocus;
										pFormatter->StartRow();
										pFormatter->AddTagCol(name);
										pFormatter->AddValueCol(pFormatter->CreateLink(label, link));
										pFormatter->FinishRow();
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
	if(isAnyLinks) {
	    links += pFormatter->Render();
    }
}


END_NCBI_SCOPE
