/*  $Id: na_utils.cpp 27815 2013-04-08 15:10:18Z rudnev $
* ===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data, the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties, express or implied, including
*  warranties of performance, merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================
*
* Authors:  Liangshou Wu
*
* File Description:
*    General utility classes for GUI projects.
*/

#include <ncbi_pch.hpp>
#include <gui/objutils/na_utils.hpp>

#include <objmgr/scope.hpp>
#include <objmgr/object_manager.hpp>
#include <objmgr/align_ci.hpp>
#include <objmgr/annot_ci.hpp>
#include <objmgr/util/sequence.hpp>
#include <objects/seqtable/Seq_table.hpp>
#include <objects/seqtable/SeqTable_column.hpp>
#include <objects/seqtable/SeqTable_column_info.hpp>

// needed for NAA meta data processing
#include <objtools/eutils/api/elink.hpp>
#include <objtools/eutils/api/esummary.hpp>
#include <objtools/eutils/api/esearch.hpp>
#include <objtools/eutils/elink/Link.hpp>
#include <objtools/eutils/elink/ELinkResult.hpp>
#include <objtools/eutils/elink/Id.hpp>
#include <objtools/eutils/elink/LinkInfo.hpp>
#include <objtools/eutils/elink/LinkSet.hpp>
#include <objtools/eutils/elink/LinkSetDbHistory.hpp>
#include <objtools/eutils/elink/IdList.hpp>
#include <objtools/eutils/elink/Id.hpp>
#include <objtools/eutils/elink/Id.hpp>
#include <objtools/eutils/esearch/ESearchResult.hpp>
#include <objtools/eutils/esearch/IdList.hpp>
#include <misc/xmlwrapp/xmlwrapp.hpp>


BEGIN_NCBI_SCOPE
USING_SCOPE(objects);
USING_SCOPE(elink);

static const int kRetMax = 5000;


enum EAlignType {
    fAlign_DNA        = 0x01,
    fAlign_Protein    = 0x02,
    fAlign_Mixed      = 0x04,
    fAlign_Invalid    = 0x80000000
};


string CDataTrackUtils::GetAlignType(const CSeq_align& align,
                                     CScope& scope)
{
    string aln_type = kEmptyStr;
    int num_row = 0;
    try {
        num_row = align.CheckNumRows();
    } catch (CException&) {
    }

    if (num_row < 2) return aln_type;

    // check align type
    EAlignType type = fAlign_Invalid;
    for (int row = 0;  row < num_row;  ++row) {
        EAlignType this_type = fAlign_Mixed;
        CBioseq_Handle handle = scope.GetBioseqHandle(align.GetSeq_id(row));
        if ( !handle ) continue;

        switch (handle.GetBioseqCore()->GetInst().GetMol())
        {
        case objects::CSeq_inst::eMol_dna:
        case objects::CSeq_inst::eMol_rna:
        case objects::CSeq_inst::eMol_na:
            this_type = fAlign_DNA;
            break;

        case objects::CSeq_inst::eMol_aa:
            this_type = fAlign_Protein;
            break;

        default:
            break;
        }

        if (this_type == fAlign_Mixed) {
            type = this_type;
            break;
        }

        if (row == 0) {
            type = this_type;
        } else if (this_type != type) {
            type = fAlign_Mixed;
            break;
        }
    }
    switch (type) {
        case fAlign_DNA:
            aln_type = "nucleotide";
            break;
        case fAlign_Protein:
            aln_type = "protein";
            break;
        case fAlign_Mixed:
            aln_type = "protein-to-nucleotide";
            break;
        default:
            break;
    }
    return aln_type;
}


void CDataTrackUtils::GetColumnHeader(set<string>& headers,
                                      const CSeq_annot& annot)
{
    if ( !annot.IsSeq_table() ) return;

    const CSeq_table::TColumns& cols =
        annot.GetData().GetSeq_table().GetColumns();
    const CEnumeratedTypeValues* type_val =
        CSeqTable_column_info::GetTypeInfo_enum_EField_id();

    ITERATE(CSeq_table::TColumns, iter, cols) {
        const CSeqTable_column::THeader& header = (*iter)->GetHeader();
        if (header.CanGetField_name()) {
            headers.insert(header.GetField_name());
        } else if (header.CanGetField_id()  &&  type_val) {
            headers.insert(type_val->FindName(header.GetField_id(), true));
        }
    }
}


void CDataTrackUtils::GetFtableSubtype(list<string>& subtypes,
                                       const set<string>& feat_subtypes)
{
    set<string> temp_types = feat_subtypes;
    set<string>::iterator type_iter = temp_types.find("gene");
    if (type_iter != temp_types.end()) {
        subtypes.push_back("gene_model");
        temp_types.erase(type_iter);
        // It contains genes, so we treat it as a gene model track.
        // All other RNA, CDS and Exon features will be shown in
        // the same track.
        const CFeatList& feats(*CSeqFeatData::GetFeatList());
        ITERATE (CFeatList, subtype_iter, feats) {
            if ((subtype_iter->GetType() == CSeqFeatData::e_Rna  &&
                subtype_iter->GetSubtype() != CSeqFeatData::eSubtype_any)  ||
                subtype_iter->GetSubtype() == CSeqFeatData::eSubtype_exon  ||
                subtype_iter->GetSubtype() == CSeqFeatData::eSubtype_cdregion) {
                    type_iter = temp_types.find(subtype_iter->GetStoragekey());
                    if (type_iter != temp_types.end()) {
                        temp_types.erase(type_iter);
                    }
            }
        }
    }
    ITERATE(set<string>, iter, temp_types) {
        if (*iter == "variation") {
            subtypes.push_back("dbSNP");
        } else if (*iter == "variation_ref") {
            subtypes.push_back("dbVar");
        } else {
            subtypes.push_back(*iter);
        }
    }
}


static bool s_MatchHeaders(const char** header_names,
                           const set<string>& headers)
{
    for ( ;  header_names  &&  *header_names;  ++header_names) {
        if (headers.count(string(*header_names)) == 0) {
            return false;
        }
    }
    return true;
}


string CDataTrackUtils::GetSeqTableSubtype(const set<string>& headers)
{
    static const char* graph_headers[] = {
        "location-from", "span", "values", NULL
    };
    if (s_MatchHeaders(graph_headers, headers)) {
        return "graph";
    }

    static const char* bins_headers[] = {
        "pos", "pvalue", "trait", "pmids", "reportedGenes",
        "mappedGenes", "snpId", "trackType", "clinSigID", NULL
    };
    if (s_MatchHeaders(bins_headers, headers)) {
        return "SNP_bins";
    }

    static const char* GWAS_headers[] = {
        "pos", "pvalue", "trackType", NULL
    };
    if (s_MatchHeaders(GWAS_headers, headers)) {
        return "GWAS";
    }

    static const char* HapMap_headers[] = {
        "pos", "value", "trackType", NULL
    };
    if (s_MatchHeaders(HapMap_headers, headers)) {
        return "HapMap";
    }

    return kEmptyStr;
}


void CDataTrackUtils::DataType2TrackType(const string& annot_type,
                                         const string& subtype,
                                         string& track_key,
                                         string& subkey)
{
    string ftable =
        CSeq_annot::TData::SelectionName(CSeq_annot::TData::e_Ftable);
    string align =
        CSeq_annot::TData::SelectionName(CSeq_annot::TData::e_Align);
    string seq_table =
        CSeq_annot::TData::SelectionName(CSeq_annot::TData::e_Seq_table);
    string graph =
        CSeq_annot::TData::SelectionName(CSeq_annot::TData::e_Graph);

    if (annot_type == align) {
        track_key = "alignment_track";
        subkey = subtype;
    } else if (annot_type == ftable) {
        if (subtype == "gene_model") {
            track_key = "gene_model_track";
        } else if (subtype == "dbVar") {
            track_key = "dbVar_track";
        } else if (subtype == "dbSNP") {
            track_key = "SNP_track";
        } else {
            track_key = "featur_track";
            subkey = subtype;
        }
    } else if (annot_type == graph) {
        if (subtype == "graph") {
            track_key = "graph_track";
        } else if (subtype == "GWAS") {
            track_key = "GWAS_track";
        }
    } else if (annot_type == seq_table) {
        if (subtype == "GWAS") {
            track_key = "GWAS_track";
        } else if (subtype == "HapMap") {
            track_key = "HapMapRR_track";
        } else if (subtype == "SNP_bins") {
            track_key = "SNP_bins_track";
        } else if (subtype == "graph") {
            track_key = "graph_track";
        } else {
            track_key = "feature_track";
            subkey = subtype;
        }
    } else {
        if (subtype == "sequence") {
            track_key = "sequence_track";
        } else if (subtype == "six_frame") {
            track_key = "six_frames_translation";
        } else if (subtype == "segment_map") {
            track_key = "segment_map_track";
        }
    }
}


/// help class for iterating through the NA DocSum one by one.
class CNADocSumIterator
{
public:
    CNADocSumIterator(const string& ids, int max_nas)
        : m_Valid(false)
    {
        if ( !ids.empty() ) {
            // prepare eSummary request
            CRef<CEUtils_ConnContext> ctx(new CEUtils_ConnContext);
            CESummary_Request sum_req("seqannot", ctx);
            sum_req.SetRequestMethod(CEUtils_Request::eHttp_Post);
            sum_req.GetId().SetIds(ids);
            sum_req.SetRetMax(max_nas);

            // Get and show the results
            string result_str;
            try {
                sum_req.Read(&result_str);
            } catch (CException& e) {
                LOG_POST(Error << "Get error when trying to retrieve NA meta-data. Error: "
                         << e.GetMsg());
            }


            if ( !result_str.empty() ) {
                m_Doc.reset(new xml::document(result_str.c_str(), result_str.size(), NULL));
                xml::node& root = m_Doc->get_root_node();
                //cout << "result: " << result_str << endl;
                m_DocSumRoot = root.find("DocumentSummarySet", root.begin());
                if (m_DocSumRoot != root.end()) {
                    m_Valid = true;
                    m_Iter = m_DocSumRoot->begin();
                }
            }
        }
    }

    bool is_valid() const
    {
        return m_Valid  &&  m_Iter != m_DocSumRoot->end();
    }

    operator bool() const
    {
        return is_valid();
    }

    void operator++()
    {
        // please do check the validity before call this method
        _ASSERT(this->is_valid());
        ++m_Iter;
    }

    const xml::node& operator*() const
    {
        // please do check the validity before call this method
        _ASSERT(this->is_valid());
        return *m_Iter;
    }

    xml::node::const_iterator operator->() const
    {
        // please do check the validity before call this method
        _ASSERT(this->is_valid());
        return m_Iter;
    }

private:
    auto_ptr<xml::document> m_Doc;
    xml::node::const_iterator m_Iter;
    xml::node::const_iterator m_DocSumRoot;
    bool m_Valid;
};


CNAUtils::CNAUtils()
{
    x_Init();
}


CNAUtils::CNAUtils(const CSeq_id& id)
    : m_TargetSeq(&id)
{
    m_Scope.Reset(new CScope(*CObjectManager::GetInstance()));
    m_Scope->AddDefaults();
    x_Init();
}


CNAUtils::CNAUtils(const CSeq_id& id, CScope& scope)
    : m_TargetSeq(&id)
    , m_Scope(&scope)
{
    x_Init();
}


void CNAUtils::GetAllNAAs(TNAAs& naas, const string& context) const
{
    naas.clear();
    string ids = x_GetNAIds(context);
    x_GetNAAs(naas, ids);
}


void CNAUtils::SearchNAAs(TNAAs& naas, const string& term) const
{
    if (term.empty()) return;

    naas.clear();
    string ids;
    x_SearchNAIds(ids, term, true);
    x_GetNAAs(naas, ids);
}


void CNAUtils::SearchNAMetaData(TNAMetaDataSet& md_set, const string& term) const
{
    if (term.empty()) return;

    md_set.clear();
    string ids;
    x_SearchNAIds(ids, term, true);
    x_GetNAMetaData(md_set, ids, false);
}


void CNAUtils::GetNAMetaData(TNAMetaDataSet& md_set,
                             const string& na,
                             bool filtering) const
{
    if (na.empty()) return;

    md_set.clear();
    string ids;
    x_SearchNAIds(ids, na, filtering);
    x_GetNAMetaData(md_set, ids, true);
}


void CNAUtils::GetNAMetaData(TNAMetaDataSet& md_set,
                             const TNAAs& naas,
                             bool filtering) const
{
    if (naas.empty()) return;

    md_set.clear();
    string term;
    ITERATE (TNAAs, iter, naas) {
        term += *iter + " OR ";
    }
    string ids;
    x_SearchNAIds(ids, term, filtering);
    x_GetNAMetaData(md_set, ids, true);
}


void CNAUtils::GetAllNAMetaData(TNAMetaDataSet& md_set,
                                const string& context) const
{
    string ids = x_GetNAIds(context);
    x_GetNAMetaData(md_set, ids, false);
}


void CNAUtils::GetNATrackInfo(TTrackInfoList& track_info,
                              const string& naa) const
{
    TNAMetaDataSet md_set;
    GetNAMetaData(md_set, naa);
    GetNATrackInfo(track_info, naa, md_set);
}


void CNAUtils::GetNATrackInfo(TTrackInfoList& track_info,
                              const string& naa,
                              const TNAMetaDataSet& md_set) const
{
    if (md_set.empty()) return;

    CRef<CAnnotMetaData> dummy(new CAnnotMetaData);
    dummy->m_Name = naa;
    TNAMetaDataSet::const_iterator iter = md_set.find(dummy);
    if (iter != md_set.end()) {
        string ftable =
            CSeq_annot::TData::SelectionName(CSeq_annot::TData::e_Ftable);
        string align =
            CSeq_annot::TData::SelectionName(CSeq_annot::TData::e_Align);
        string seq_table =
            CSeq_annot::TData::SelectionName(CSeq_annot::TData::e_Seq_table);
        string graph =
            CSeq_annot::TData::SelectionName(CSeq_annot::TData::e_Graph);

        const CAnnotMetaData& data = **iter;
        CRef<CTrackInfo> info(new CTrackInfo);
        info->m_DataId = data.m_Name;
        info->m_Title = data.m_Title;
        info->m_Descr = data.m_Descr;
        info->m_Group = data.m_xClass;
        info->m_Dbname = "SADB";
        info->m_AnnotType = data.m_AnnotType;

        if (data.m_AnnotType == ftable) {
            list<string> subtypes;
            CDataTrackUtils::GetFtableSubtype(subtypes, data.m_Subtypes);
            ITERATE (list<string>, subtype_iter, subtypes) {
                CRef<CTrackInfo> feat_info(new CTrackInfo(*info));
                feat_info->m_Subtype = *subtype_iter;
                track_info.push_back(feat_info);
            }
        } else if (data.m_AnnotType == align) {
            if (data.m_Subtypes.empty()) {
                // try to see if we could find out what type
                // of alignments it contains
                if (m_TargetSeq) {
                    if (m_Handle) {
                        objects::SAnnotSelector sel;
                        sel.IncludeAnnotType(CSeq_annot::C_Data::e_Align);
                        sel.SetMaxSize(1);
                        sel.SetCollectNames(false);
                        sel.AddNamedAnnots(data.m_Name);
                        sel.IncludeNamedAnnotAccession(data.m_Name);
                        CAlign_CI aln_iter(m_Handle, TSeqRange::GetWhole(), sel);
                        if (aln_iter) {
                            string aln_type =
                                CDataTrackUtils::GetAlignType(*aln_iter, const_cast<CScope&>(*m_Scope));
                            if ( !aln_type.empty() ) {
                                info->m_Subtype = aln_type;
                                track_info.push_back(info);
                            }
                        }
                    }
                }
            } else if (data.m_Subtypes.size() == 1) {
                info->m_Subtype = *data.m_Subtypes.begin();
                track_info.push_back(info);
            }
            // otherwise, too many subtype, give up
        } else if (data.m_AnnotType == graph  &&  data.m_Subtypes.size() < 2) {
            if (data.m_Subtypes.empty()) {
                info->m_Subtype = "graph";
            } else {
                info->m_Subtype = *data.m_Subtypes.begin();
            }
            track_info.push_back(info);
        } else if (data.m_AnnotType == seq_table) {
            if (data.m_Subtypes.size() == 1) {
                info->m_Subtype = *data.m_Subtypes.begin();
                track_info.push_back(info);
            } else if (data.m_Subtypes.empty()  &&  m_TargetSeq  &&  m_Handle) {
                objects::SAnnotSelector sel;
                sel.IncludeAnnotType(CSeq_annot::C_Data::e_Seq_table);
                sel.SetMaxSize(1);
                sel.SetCollectNames(false);
                sel.AddNamedAnnots(data.m_Name);
                sel.IncludeNamedAnnotAccession(data.m_Name);
                CAnnot_CI annot_iter(m_Handle, sel);
                if (annot_iter) {
                    const CSeq_annot& annot = *annot_iter->GetCompleteSeq_annot();
                    // check if it stores features
                    if (annot.GetData().GetSeq_table().IsSetFeat_subtype()  &&
                        annot.GetData().GetSeq_table().GetFeat_subtype() != 0) {
                        int f_subtype = annot.GetData().GetSeq_table().GetFeat_subtype();
                        const CFeatList& feats(*CSeqFeatData::GetFeatList());
                        info->m_Subtype = feats.GetStoragekey(f_subtype);
                        track_info.push_back(info);
                    } else {
                        set<string> headers;
                        CDataTrackUtils::GetColumnHeader(headers, annot);
                        info->m_Subtype = CDataTrackUtils::GetSeqTableSubtype(headers);
                        if ( !info->m_Subtype.empty() ) {
                            track_info.push_back(info);
                        }
                    }
                }
            }
        }
    }
}


void CNAUtils::GetNATrackInfo(TTrackInfoList& track_info,
                              const TNAAs& naas) const
{
    TNAMetaDataSet md_set;
    GetNAMetaData(md_set, naas);
    GetNATrackInfo(track_info, naas, md_set);
}


void CNAUtils::GetNATrackInfo(TTrackInfoList& track_info,
                              const TNAAs& naas,
                              const TNAMetaDataSet& md_set) const
{
    if (md_set.empty()) return;

    ITERATE(TNAAs, iter, naas) {
        GetNATrackInfo(track_info, *iter, md_set);
    }
}


void CNAUtils::SetMaxNAs(int size)
{
    if (size > 0) {
        m_MaxNAIds = m_MaxNAMeta = size;
    }
}

void CNAUtils::x_Init()
{
    m_GI = 0;
    m_MaxNAIds = 2 * kRetMax;
    m_MaxNAMeta = kRetMax;

    if (m_TargetSeq) {
        m_Handle = m_Scope->GetBioseqHandle(*m_TargetSeq);
        CSeq_id_Handle shdl = m_Handle.GetAccessSeq_id_Handle();
        shdl = sequence::GetId(shdl, *m_Scope, sequence::eGetId_ForceGi);
        if (shdl) {
            m_GI = shdl.GetGi();
            bool is_na = m_Handle.IsNa();
            m_DBFrom = "nucleotide";
            if ( !is_na ) {
                m_DBFrom = "protein";
            }
        }
    }
}


string CNAUtils::x_GetNAIds(const string& context) const
{
    string id_str = kEmptyStr;
    if (m_GI == 0  ||  m_DBFrom.empty()) return id_str;

    list<string> ids;
    if (context.empty()  ||  NStr::EqualNocase(context, "all")) {
        x_GetAllNAIds(ids);
    } else {
        // Get NAA ids based on a viewer context
        // Related JIRA tickets: ID-544 and SV-831
        x_GetAllNAIdsWithContext(ids, context);
    }

    ITERATE (TNAAs, iter, ids) {
        id_str += *iter + ",";
    }

    return id_str;
}


void CNAUtils::x_GetAllNAIds(list<string>& ids) const
{
    CRef<CEUtils_ConnContext> ctx(new CEUtils_ConnContext);

    // prepare eLink request
    CELink_Request link_req("seqannot", ctx);
    link_req.SetRequestMethod(CEUtils_Request::eHttp_Get);
    link_req.SetDbFrom(m_DBFrom);
    link_req.GetIdGroups().SetGroups(NStr::IntToString(m_GI));

    CRef<elink::CELinkResult> result;
    try {
        result = link_req.GetELinkResult();
    } catch (CException& e) {
        LOG_POST(Error << "Get error when trying to get NA ids for gi: "
                 << m_GI << ". Error: " << e.GetMsg());
    }

    if (result) {
        CTypeIterator<CLink> link_iter(*result);
        for (; link_iter; ++link_iter) {
            if (link_iter->IsSetId()  &&  link_iter->GetId().IsSetId()) {
                ids.push_back(link_iter->GetId().GetId());
            }
        }
    }
}


void CNAUtils::x_GetAllNAIdsWithContext(list<string>& ids,
                                        const string& context) const
{
    CRef<CEUtils_ConnContext> ctx(new CEUtils_ConnContext);

    // Get WebEnv and QueryKey for the given id using elink
    CELink_Request link_req("seqannot", ctx);
    link_req.SetRequestMethod(CEUtils_Request::eHttp_Get);
    link_req.SetDbFrom(m_DBFrom);
    link_req.GetIdGroups().SetGroups(NStr::IntToString(m_GI));
    link_req.SetCommand(CELink_Request::eCmd_neighbor_history);
    link_req.SetLinkName(m_DBFrom + "_seqannot");

    CRef<elink::CELinkResult> result;
    try {
        result = link_req.GetELinkResult();
    } catch (CException& e) {
        LOG_POST(Error << "Get error when trying to get NA list for gi: "
                 << m_GI << ", and context: " << context
                 << ". Error: " << e.GetMsg());
    }


    if (result  &&  result->IsSetLinkSet()  &&  !result->GetLinkSet().empty()) {
        const CLinkSet& link_set = *result->GetLinkSet().front();
        if (link_set.IsSetData()  &&
            link_set.GetData().IsIE()  &&
            link_set.GetData().GetIE().IsSetEL_2()  &&
            link_set.GetData().GetIE().GetEL_2().IsLW()) {

            const CLinkSet::TData::TIE::TEL_2::TLW& lw =
                link_set.GetData().GetIE().GetEL_2().GetLW();
            if (lw.CanGetLinkSetDbHistory()  &&  !lw.GetLinkSetDbHistory().empty()) {
                typedef CLinkSet::TData::TIE::TEL_2::TLW::TLinkSetDbHistory TLinkSetDbHistory;
                string query_key = kEmptyStr;
                ITERATE (TLinkSetDbHistory, h_iter, lw.GetLinkSetDbHistory()) {
                    if ((*h_iter)->IsSetQI()   &&  (*h_iter)->GetQI().IsQueryKey()) {
                        query_key = (*h_iter)->GetQI().GetQueryKey();
                        break;
                    }
                }

                if ( !query_key.empty()  &&  lw.CanGetWebEnv() ) {
                    ctx->SetQueryKey(query_key);
                    ctx->SetWebEnv(lw.GetWebEnv());

                    // Get NA ids using esearch
                    CESearch_Request search_req("seqannot", ctx);
                    search_req.SetTerm(context + "[viewer_context]");
                    CRef<esearch::CESearchResult> search_result;

                    try {
                        search_result = search_req.GetESearchResult();
                    } catch (CException& e) {
                        LOG_POST(Error << "Get error when trying to get NA id list for viewer_context: "
                                 << context << ", and query_key: " << query_key
                                 << ". Error: " << e.GetMsg());
                    }


                    if (search_result  &&  search_result->CanGetData()  &&
                        search_result->GetData().IsInfo()  &&
                        search_result->GetData().GetInfo().CanGetContent()  &&
                        search_result->GetData().GetInfo().GetContent().CanGetIdList()  &&
                        search_result->GetData().GetInfo().GetContent().GetIdList().CanGetId()) {
                        ids = search_result->GetData().GetInfo().GetContent().GetIdList().GetId();
                    }
                }
            }

        }
    }
}


void CNAUtils::x_GetNAAs(TNAAs& naas, const string& ids) const
{
    CNADocSumIterator ds_iter(ids, m_MaxNAMeta);
    for (; ds_iter; ++ds_iter) {
        xml::node::const_iterator c_i = ds_iter->find("Caption", ds_iter->begin());
        xml::node::const_iterator sum_i = ds_iter->find("ExpXml", ds_iter->begin());
        if (c_i != ds_iter->end()  &&  sum_i != ds_iter->end()) {
            string naa(c_i->get_content());
            naas.push_back(naa);
        }
    }
}


void CNAUtils::x_GetNAMetaData(TNAMetaDataSet& md_set,
                               const string& ids,
                               bool get_entrez_links) const
{
    CNADocSumIterator ds_iter(ids, m_MaxNAMeta);
    for (; ds_iter; ++ds_iter) {
        xml::node::const_iterator c_i = ds_iter->find("Caption", ds_iter->begin());
        xml::node::const_iterator sum_i = ds_iter->find("ExpXml", ds_iter->begin());
        if (c_i != ds_iter->end()  &&  sum_i != ds_iter->end()) {
            CRef<CAnnotMetaData> data(new CAnnotMetaData);
            data->m_Name = c_i->get_content();

            ITERATE (xml::attributes, a_i, ds_iter->get_attributes()) {
                if (NStr::Equal(a_i->get_name(), "uid")) {
                    data->m_Id= a_i->get_value();
                    break;
                }
            }

            x_ParseNAMetaData(*data, sum_i->get_content());
            if (get_entrez_links) {
                // we need to retrieve Entrez links
                x_GetNAEntrezLinks(*data);
            }
            md_set.insert(data);
        }
    }
}


void CNAUtils::x_SearchNAIds(string& ids, const string& term, bool filtering) const
{
    CRef<CEUtils_ConnContext> ctx(new CEUtils_ConnContext);

    // prepare eSearch request
    CESearch_Request search_req("seqannot", ctx);
    search_req.SetRequestMethod(CEUtils_Request::eHttp_Get);
    search_req.SetTerm(term);
    search_req.SetRetMax(m_MaxNAIds);

    list<string> search_ids;
    CRef<esearch::CESearchResult> result;

    try {
        result = search_req.GetESearchResult();
    } catch (CException& e) {
        LOG_POST(Error << "Get error when trying to search NAAs for term:: "
                 << term << ". Error: " << e.GetMsg());
    }

    if (result  &&  result->CanGetData()  &&
        result->GetData().IsInfo()  &&
        result->GetData().GetInfo().CanGetContent()  &&
        result->GetData().GetInfo().GetContent().CanGetIdList()  &&
        result->GetData().GetInfo().GetContent().GetIdList().CanGetId()) {
            const list<string>& id = result->GetData().GetInfo().GetContent().GetIdList().GetId();

            ITERATE (list<string>, id_iter, id) {
                search_ids.push_back(*id_iter);
            }
    }

    if (filtering  &&  m_GI != 0  &&  !m_DBFrom.empty()) {
        list<string> all_ids;
        x_GetAllNAIds(all_ids);
        all_ids.sort();
        search_ids.sort();
        list<string> new_list;
        std::set_intersection(all_ids.begin(), all_ids.end(),
            search_ids.begin(), search_ids.end(), back_inserter(new_list));
        search_ids.swap(new_list);
    }

    ITERATE (TNAAs, iter, search_ids) {
        ids += *iter + ",";
    }
}


// child node 'Descriptor'
// child node 'Provider'
// child node 'XClass'
// child node 'Content'
// child node 'Properties'

void CNAUtils::x_ParseNAMetaData(CAnnotMetaData& data, const char* xml_str) const
{
    xml::document doc(xml_str, strlen(xml_str), NULL);
    xml::node& root = doc.get_root_node();
    xml::node::const_iterator i;
    // parse descriptor
    ITERATE (xml::node, i, root) {
        if (NStr::Equal(i->get_name(), "Algorithm-List")) continue;

        string node_name = i->get_name();
        if (node_name == "Descriptor") {
            ITERATE (xml::node, e_i, *i) {
                if (NStr::Equal(e_i->get_name(), "Title")) {
                    data.m_Title = e_i->get_content();
                } else if (NStr::Equal(e_i->get_name(), "Comment")) {
                    string tmp_str = e_i->get_content();
                    tmp_str = NStr::Replace(tmp_str, "&amp;", "&");
                    tmp_str = NStr::Replace(tmp_str, "&#xA;", "\n");
                    tmp_str = NStr::Replace(tmp_str, "&#xa;", "\n");
                    data.m_Descr = tmp_str;
                }
            }
            ITERATE (xml::attributes, a_i, i->get_attributes()) {
                if (NStr::Equal(a_i->get_name(), "name")) {
                    data.m_OtherName = a_i->get_value();
                } else if (NStr::Equal(a_i->get_name(), "scope")) {
                    data.m_Scope = a_i->get_value();
                }
            }
            if (data.m_Title.empty()) {
                data.m_Title = data.m_OtherName;
            }

        } else if (node_name == "XClass") {
            ITERATE (xml::attributes, a_i, i->get_attributes()) {
                if (NStr::Equal(a_i->get_name(), "type")) {
                    data.m_xClass = a_i->get_value();
                }
            }

        } else if (node_name == "Properties") {
            string keywords = kEmptyStr;
            ITERATE (xml::node, e_i, *i) {
                string e_name = e_i->get_name();
                if (e_name == "TargetAssembly") {
                    ITERATE (xml::attributes, a_i, e_i->get_attributes()) {
                        if (NStr::Equal(a_i->get_name(), "idstr")) {
                            data.m_AssmAcc = a_i->get_value();
                        }
                    }
                } else if (e_name == "LinkList") {
                    ITERATE (xml::node, ln_i, *e_i) {
                        if ( !NStr::Equal(ln_i->get_name(), "Link") ) {
                            continue;
                        }
                        ITERATE (xml::node, url_entrez_i, *ln_i) {
                            string label, url;
                            if (NStr::Equal(url_entrez_i->get_name(), "URL_Link")) {
                                ITERATE (xml::node, lln_i, *url_entrez_i) {
                                    if (NStr::Equal(lln_i->get_name(), "Label")) {
                                        label = lln_i->get_content();
                                    } else if (NStr::Equal(lln_i->get_name(), "URL")) {
                                        url = lln_i->get_content();
                                    }
                                }
                            } else if (NStr::Equal(url_entrez_i->get_name(), "Entrez_Link")) {
                                // we shouldn't retrieve Entrez links from NAA's meta-data since
                                // they might not be up-to-date.  Instead, we will retrieve them
                                // through elink.fcgi.
                                ITERATE (xml::node, lln_i, *url_entrez_i) {
                                    if (NStr::Equal(lln_i->get_name(), "DB")) {
                                        label = lln_i->get_content();
                                    } else if (NStr::Equal(lln_i->get_name(), "ID")) {
                                        url = lln_i->get_content();
                                    }
                                }
                                if ( !label.empty()  &&  !url.empty() ) {
                                    url = "http://www.ncbi.nlm.nih.gov/" +
                                        label + "/" + url;
                                }
                            }
                            if ( !label.empty()  &&  !url.empty() ) {
                                data.m_Links[label] = url;
                            }
                        }
                    }
                }
            }

        } else if (node_name == "Content") {
            // It is possible there are more than one annotation types
            // and feature (sub)types set in meta-data. We need to
            // collect them all
            set<string> annot_types;
            ITERATE (xml::node, e_i, *i) {
                string e_name = e_i->get_name();
                if (e_name == "Annot") {
                    ITERATE (xml::attributes, a_i, e_i->get_attributes()) {
                        if (NStr::Equal(a_i->get_name(), "type")) {
                            annot_types.insert(a_i->get_value());
                        }
                    }
                    ITERATE (xml::node, t_i, *e_i) {
                        if (NStr::Equal(t_i->get_name(), "Feature")) {
                            ITERATE (xml::attributes, a_i, t_i->get_attributes()) {
                                if (NStr::Equal(a_i->get_name(), "type")  &&
                                    !NStr::Equal(a_i->get_value(), "bad")) {
                                    data.m_Subtypes.insert(string(a_i->get_value()));
                                }
                            }
                        }
                    }
                }
            }

            if ( !data.m_Subtypes.empty() ) {
                // It is feature table
                // We will use all feature subtypes
                string ftable =
                    CSeq_annot::TData::SelectionName(CSeq_annot::TData::e_Ftable);
                data.m_AnnotType = ftable;
            } else {
                vector<string> priorytized_types;
                priorytized_types.push_back(
                    CSeq_annot::TData::SelectionName(CSeq_annot::TData::e_Align));
                priorytized_types.push_back(
                    CSeq_annot::TData::SelectionName(CSeq_annot::TData::e_Seq_table));
                priorytized_types.push_back(
                    CSeq_annot::TData::SelectionName(CSeq_annot::TData::e_Graph));

                ITERATE(vector<string>, type_iter, priorytized_types) {
                    set<string>::const_iterator at_iter = annot_types.find(*type_iter);
                    if (at_iter != annot_types.end()) {
                        data.m_AnnotType = *type_iter;
                        break;
                    }
                }
            }
        }
    }
}


void CNAUtils::x_GetNAEntrezLinks(CAnnotMetaData& data) const
{
    CRef<CEUtils_ConnContext> ctx(new CEUtils_ConnContext);

    //prepare eLink request
    CELink_Request link_req("all", ctx);
    link_req.SetRequestMethod(CEUtils_Request::eHttp_Get);
    link_req.SetDbFrom("seqannot");
    link_req.GetIdGroups().SetGroups(data.m_Id);
    link_req.SetCommand(CELink_Request::eCmd_acheck);

    CRef<elink::CELinkResult> result;

    try {
        result = link_req.GetELinkResult();
    } catch (CException& e) {
        LOG_POST(Error << "Get error when trying to get Entrez links for NA id: "
                 << data.m_Id << ". Error: " << e.GetMsg());
    }

    if ( !result ) return;

    CTypeIterator<CLinkInfo> link_iter(*result);
    for (; link_iter; ++link_iter) {
        if (link_iter->CanGetDbTo()  &&  link_iter->CanGetLinkName()  &&
            link_iter->CanGetHtmlTag()) {
            const string& html_tag = link_iter->GetHtmlTag();
            const string& db_to = link_iter->GetDbTo();
            const string& link_name = link_iter->GetLinkName();

            CELink_Request e_link_req(db_to, ctx);
            e_link_req.SetRequestMethod(CEUtils_Request::eHttp_Get);
            e_link_req.SetDbFrom("seqannot");
            e_link_req.GetIdGroups().SetGroups(data.m_Id);
            e_link_req.SetLinkName(link_name);
            CRef<elink::CELinkResult> link_result;

            try {
                link_result = e_link_req.GetELinkResult();
            } catch (CException& e) {
                LOG_POST(Error << "Get error when trying to get Entrez links for link name: "
                         << link_name << ". Error: " << e.GetMsg());
            }

            if ( !link_result ) break;

            CTypeIterator<CLink> iter(*link_result);
            for (; iter; ++iter) {
                if (iter->IsSetId()  &&  iter->GetId().IsSetId()) {
                    string link = "http://www.ncbi.nlm.nih.gov/";
                    link += db_to + "/" + iter->GetId().GetId();
                    NON_CONST_ITERATE(CAnnotMetaData::TLinks, c_link, data.m_Links) {
                        if (c_link->second == link) {
                            data.m_Links.erase(c_link);
                            break;
                        }
                    }
                    data.m_Links[html_tag] = link;

                    // there should be only one link
                    break;
                }
            }
        }
    }
}


END_NCBI_SCOPE
