/*  $Id: registry.cpp 28047 2013-05-06 21:10:21Z rafanovi $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Mike DiCuccio
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>
#include <gui/objutils/registry.hpp>
#include <gui/objutils/reg_file.hpp>
#include <gui/objutils/reg_settings.hpp>
#include <algorithm>
#include <iterator>

#include <corelib/ncbimtx.hpp>
#include <objects/general/Object_id.hpp>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

const string CGuiRegistry::kMetaSuffix = "-meta";
const string CGuiRegistry::kDecimalDot = ".";

static void s_ExtractKeys(const CUser_object& obj,
                          const string& section,
                          const string& delim,
                          CGuiRegistry::TKeys& key_info);

static void s_ExtractKeys(const CUser_field& field,
                          const string& root,
                          CGuiRegistry::TKeys& key_info,
                          bool recurse = true);



CGuiRegistry::CGuiRegistry()
: m_WritableSite(ePriority_Local)
{
}


CGuiRegistry::~CGuiRegistry()
{
}

void CGuiRegistry::SetLocal(CNcbiIstream& istr)
{
    AddSite(istr, ePriority_Local);
}


void CGuiRegistry::FromConfigFile(CNcbiIstream& istr)
{
    m_Registries[ePriority_Local].Reset(new CRegistryFile());
    m_Registries[ePriority_Local]->FromConfigFile(istr);
}


void CGuiRegistry::ToConfigFile(CNcbiOstream& ostr) const
{
    TRegistries::const_iterator it = m_Registries.find(ePriority_Local);
    if (it != m_Registries.end()) {
        it->second->ToConfigFile(ostr);
    }
}


void CGuiRegistry::SetGlobal(CNcbiIstream& istr)
{
    AddSite(istr, ePriority_Global);
}


CRef<CRegistryFile> CGuiRegistry::AddSite(CNcbiIstream& istr, int priority)
{
    CRef<CRegistryFile> site;
    try {
        site.Reset(new CRegistryFile(istr));
        m_Registries[priority] = site;
    }
    catch (CException& e) {
        LOG_POST(Error << "Error reading settings: " << e.GetMsg());
    }
    catch (std::exception& e) {
        LOG_POST(Error << "Unknown error reading local settings; "
                 "the settings will be ignored " << e.what());
    }
    return site;
}


CRef<CRegistryFile> CGuiRegistry::AddSite(const objects::CUser_object& reg, int priority)
{
    CRef<CRegistryFile> site;
    site.Reset(new CRegistryFile(reg));
    m_Registries[priority] = site;
    return site;
}


int CGuiRegistry::x_NextTempPriority() const
{
    int top_priority = ePriority_Local;
    if ( ! m_Registries.empty()) {
        top_priority = m_Registries.begin()->first;
    }
    top_priority -= ePriority_Temp_Inc;
    return top_priority;
}


CRef<CRegistryFile> CGuiRegistry::PushWritableTempSite()
{
    int top_priority = x_NextTempPriority();
    CRef<CRegistryFile> site(new CRegistryFile);
    m_Registries[top_priority ] = site;
    SetWritableSite(top_priority);

    return site;
}


CRef<CRegistryFile> CGuiRegistry::PushTempSite(CNcbiIstream& istr, bool writable)
{
    int top_priority = x_NextTempPriority();
    if (writable) {
        SetWritableSite(top_priority);
    }
    return AddSite(istr, top_priority);
}


CRef<CRegistryFile> CGuiRegistry::PushTempSite(const objects::CUser_object& reg, bool writable)
{
    int top_priority = x_NextTempPriority();
    if (writable) {
        SetWritableSite(top_priority);
    }
    return AddSite(reg, top_priority);
}


CRef<CRegistryFile> CGuiRegistry::TopTempSite()
{
    if ( ! m_Registries.empty()) {
        int top_priority = m_Registries.begin()->first;
        if (top_priority < ePriority_Local) {
            return m_Registries.begin()->second;
        }
    }
    return CRef<CRegistryFile>();
}


void CGuiRegistry::PopTempSite()
{
    if ( ! m_Registries.empty()) {
        int top_priority = m_Registries.begin()->first;
        if (top_priority < ePriority_Local) {
            x_DeleteSite(m_Registries.begin());
        }
    }
}


void CGuiRegistry::MergeTempSite()
{
    if ( ! m_Registries.empty()) {
        int top_priority = m_Registries.begin()->first;
        if (top_priority < ePriority_Local) {
            x_MergeSite(m_Registries.begin());
        }
    }
}


// from "The C++ Standard Library", Josuttis, pg 211-212
template <class K, class V>
class value_equals {
private:
    V value;
public:
    value_equals(const V& v) : value(v) {}

    bool operator() (pair<const K, V> elem) {
        return elem.second == value;
    }
};

typedef value_equals<int, CRef<CRegistryFile> >  TRefRegEquals;

void CGuiRegistry::DeleteSite(CRef<CRegistryFile> site)
{
    TRegistries::iterator it = find_if(m_Registries.begin(), m_Registries.end(),
                                       TRefRegEquals(site));
    x_DeleteSite(it);
}


void CGuiRegistry::x_DeleteSite(TRegistries::iterator it)
{
    if (it != m_Registries.end()) {
        /// if deleting the writable site, switch to Local.
        if (it->first == GetWritableSite()) {
            SetWritableSite(ePriority_Local);
        }
        m_Registries.erase(it);
    }
}


void  CGuiRegistry::SetWritableSite(int priority)
{
    // this site might not exist. But if it doesn't, it will
    // be created the first time someone writes to it.
    m_WritableSite = priority;
}


int CGuiRegistry::GetWritableSite()
{
    return m_WritableSite;
}


void CGuiRegistry::MergeSite(CRef<CRegistryFile> site)
{

    TRegistries::iterator it = find_if(m_Registries.begin(), m_Registries.end(),
                                       TRefRegEquals(site));
    x_MergeSite(it);
}


void CGuiRegistry::x_MergeSite(TRegistries::iterator src_it)
{
    if (src_it == m_Registries.end())
        return;

    // use if we want to merge into the next site instead of always merging into
    // the local site. (More like a stack, but can violate read-only sites.

    // read-only sites should have their own stack, separate from write sites
    // but at a higher priority.

    TRegistries::iterator dest_it = src_it;
    ++dest_it;
    if (dest_it == m_Registries.end())
        return;
    CRegistryFile& dest_site = *dest_it->second;
    int dest_priority = dest_it->first;

    // CRegistryFile& dest_site = x_GetLocal();
    // int dest_priority = ePriority_Local;

    {{
         CRegistryFile::TGuard LOCK1(*src_it->second);
         const objects::CUser_object& src_uo = src_it->second->GetRegistry();
         int src_priority = src_it->first;

         CRegistryFile::TGuard LOCK2(dest_site);
         objects::CUser_object& dest_uo = dest_site.SetRegistry();

         CGuiRegistry::TKeys key_info;
         string delim(".");
         s_ExtractKeys(src_uo, "", delim, key_info);

         ITERATE(CGuiRegistry::TKeys, key_it, key_info) {
             const CGuiRegistry::SKeyInfo& key = *key_it;
             if (key.type == objects::CUser_field::TData::e_Fields)
                 continue;
             CConstRef<objects::CUser_field> data_field =
                 CGuiRegistryUtil::GetFieldAt(src_uo, key.key, delim);
             CRef<objects::CUser_field> dest_field =
                 CGuiRegistryUtil::SetFieldAt(dest_uo, key.key, delim);
             if (dest_field->GetData().IsFields()) {
                 NCBI_THROW(CException, eUnknown,
                            "Too few components in key to merge: \"" + key.key + "\"");
             }
             dest_field->Assign(*data_field);
         }

         // if merging the writable site (which is about to be deleted).
         // make the destination the writable.
         if (src_priority == GetWritableSite()) {
             SetWritableSite(dest_priority);
         }
     }}
    x_DeleteSite(src_it);
}


void CGuiRegistry::Write(CNcbiOstream& ostr, int priority) const
{
    CRef<CRegistryFile> reg = x_GetRegistryRef(priority);
    if (reg) {
        reg->Write(ostr);
    }
}


static void s_ExtractKeys(const CUser_field& field,
                          const string& root,
                          CGuiRegistry::TKeys& key_info,
                          bool recurse)
{
    CGuiRegistry::SKeyInfo info;
    info.key = root;
    if ( !info.key.empty() ) {
        info.key += ".";
    }
    info.key += field.GetLabel().GetStr();
    info.type = field.GetData().Which();
    key_info.push_back(info);

    if (recurse) {
        switch (field.GetData().Which()) {
        case CUser_field::TData::e_Fields:
            {{
                 string subkey = root;
                 if ( !subkey.empty() ) {
                     subkey += ".";
                 }
                 subkey += field.GetLabel().GetStr();
                 ITERATE (CUser_field::TData::TFields, iter,
                          field.GetData().GetFields()) {
                     s_ExtractKeys(**iter, subkey, key_info);
                 }
             }}
            break;

        default:
            break;
        }
    }
}


static void s_ExtractKeys(const CUser_field& field,
                          const string& root,
                          const string& delim,
                          list<string>::const_iterator key_iter,
                          list<string>::const_iterator key_end,
                          CGuiRegistry::TKeys& key_info)
{
    if (key_iter == key_end) {
        s_ExtractKeys(field, root, key_info);
    } else if (field.GetLabel().GetStr() == *key_iter) {
        string subkey = root;
        if ( !subkey.empty() ) {
            subkey += delim;
        }
        subkey += *key_iter++;

        if (field.GetData().IsFields()) {
            ITERATE (CUser_field::TData::TFields, iter,
                     field.GetData().GetFields()) {
                s_ExtractKeys(**iter, subkey, delim,
                              key_iter, key_end, key_info);
            }
        }
    }
}


static void s_ExtractKeys(const CUser_object& obj,
                          const string& section,
                          const string& delim,
                          CGuiRegistry::TKeys& key_info)
{
    key_info.clear();
    list<string> subsections;
    NStr::Split(section, delim, subsections);
    ITERATE (CUser_object::TData, iter, obj.GetData()) {
        s_ExtractKeys(**iter, "", delim,
                      subsections.begin(), subsections.end(), key_info);
    }
}


/// retrieve information about all keys in the registry
void CGuiRegistry::GetKeys(CGuiRegistry::TKeys& key_info,
                           const string& starting_section,
                           const string& delim)
{
    typedef map<string, SKeyInfo> TKeyMap;
    TKeyMap key_map;    // used to sort the final list.

    ///
    /// global registry first
    ///
    list<SKeyInfo> tmp;
    EKeyAccess access;

    TRegistries::const_reverse_iterator riter(m_Registries.end());
    TRegistries::const_reverse_iterator rend (m_Registries.begin());
    for ( ;  riter != rend;  ++riter) {

        s_ExtractKeys(riter->second->GetRegistry(),
                      starting_section, delim, tmp);

        if (riter->first == GetWritableSite()) {
            access = eAccess_ReadWrite;
        } else {
            access = eAccess_ReadOnly;
        }
        NON_CONST_ITERATE (list<SKeyInfo>, it, tmp) {
            it->access = access;
            key_map[it->key] = *it;
        }
    }

    key_info.clear();
    ITERATE (TKeyMap, iter, key_map) {
        SKeyInfo key = iter->second;
        key_info.push_back(key);
    }
}


int CGuiRegistry::GetInt(const string& key,
                         int default_val, const string& delim) const
{
    try {
        CConstRef<CUser_field> field = GetField(key, delim);
        if (field) {
            return field->GetData().GetInt();
        }
    }
    catch (std::exception&) {
    }
    return default_val;
}


double CGuiRegistry::GetReal(const string& key,
                             double default_val,
                             const string& delim) const
{
    try {
        CConstRef<CUser_field> field = GetField(key, delim);
        if (field) {
            return field->GetData().GetReal();
        }
    }
    catch (std::exception&) {
    }
    return default_val;
}


bool CGuiRegistry::GetBool(const string& key,
                           bool default_val,
                           const string& delim) const
{
    try {
        CConstRef<CUser_field> field = GetField(key, delim);
        if (field) {
            return field->GetData().GetBool();
        }
    }
    catch (std::exception&) {
    }
    return default_val;
}


string CGuiRegistry::GetString(const string& key,
                               const string& default_val,
                               const string& delim) const
{
    try {
        CConstRef<CUser_field> field = GetField(key, delim);
        if (field) {
            return field->GetData().GetStr();
        }
    }
    catch (std::exception&) {
    }
    return default_val;
}


string CGuiRegistry::GetOctetString(const string& key,
                                    const string& default_val,
                                    const string& delim) const
{
    try {
        CConstRef<CUser_field> field = GetField(key, delim);
        if (field) {
            string s;
            s.reserve(field->GetData().GetOs().size());
            std::copy(field->GetData().GetOs().begin(),
                      field->GetData().GetOs().end(),
                      back_inserter(s));
            return s;
        }
    }
    catch (std::exception&) {
    }
    return default_val;
}


void CGuiRegistry::GetIntVec(const string& key,
                             vector<int>& val,
                             const string& delim) const
{
    try {
        CConstRef<CUser_field> field = GetField(key, delim);
        if (field) {
            switch (field->GetData().Which()) {
            case CUser_field::TData::e_Fields:
                val.clear();
                ITERATE (CUser_field::TData::TFields, iter,
                         field->GetData().GetFields()) {
                    const CUser_field& f = **iter;
                    switch (f.GetData().Which()) {
                    case CUser_field::TData::e_Int:
                        val.push_back(f.GetData().GetInt());
                        break;

                    case CUser_field::TData::e_Ints:
                        val.insert(val.end(),
                                   f.GetData().GetInts().begin(),
                                   f.GetData().GetInts().end());
                        break;

                    default:
                        break;
                    }
                }
                break;

            case CUser_field::TData::e_Int:
                val.clear();
                val.push_back(field->GetData().GetInt());
                break;

            case CUser_field::TData::e_Ints:
                val = field->GetData().GetInts();
                break;

            default:
                NCBI_THROW(CException, eUnknown,
                           "invalid field type in registry request");
            }
            return;
        }
    }
    catch (CException& e) {
        LOG_POST(Error << "failed to access registry key "
                 << key << ": " << e.GetMsg());
        val.clear();
    }
    catch (std::exception&) {
        val.clear();
    }
}


void CGuiRegistry::GetRealVec(const string& key,
                              vector<double>& val,
                              const string& delim) const
{
    try {
        CConstRef<CUser_field> field = GetField(key, delim);
        if (field) {
            switch (field->GetData().Which()) {
            case CUser_field::TData::e_Fields:
                val.clear();
                ITERATE (CUser_field::TData::TFields, iter,
                         field->GetData().GetFields()) {
                    const CUser_field& f = **iter;
                    switch (f.GetData().Which()) {
                    case CUser_field::TData::e_Real:
                        val.push_back(f.GetData().GetReal());
                        break;

                    case CUser_field::TData::e_Reals:
                        val.insert(val.end(),
                                   f.GetData().GetReals().begin(),
                                   f.GetData().GetReals().end());
                        break;

                    default:
                        break;
                    }
                }
                break;

            case CUser_field::TData::e_Real:
                val.clear();
                val.push_back(field->GetData().GetReal());
                break;

            case CUser_field::TData::e_Reals:
                val = field->GetData().GetReals();
                break;

            default:
                NCBI_THROW(CException, eUnknown,
                           "invalid field type in registry request");
            }
        }
    }
    catch (CException& e) {
        LOG_POST(Error << "failed to access registry key "
                 << key << ": " << e.GetMsg());
        val.clear();
    }
    catch (std::exception&) {
        val.clear();
    }
}


void CGuiRegistry::GetStringVec(const string& key,
                                vector<string>& val,
                                const string& delim) const
{
    try {
        CConstRef<CUser_field> field = GetField(key, delim);
        if (field) {
            switch (field->GetData().Which()) {
            case CUser_field::TData::e_Fields:
                val.clear();
                ITERATE (CUser_field::TData::TFields, iter, field->GetData().GetFields()) {
                    const CUser_field& f = **iter;
                    switch (f.GetData().Which()) {
                    case CUser_field::TData::e_Str:
                        val.push_back(f.GetData().GetStr());
                        break;

                    case CUser_field::TData::e_Strs:
                        val.insert(val.end(),
                                   f.GetData().GetStrs().begin(),
                                   f.GetData().GetStrs().end());
                        break;

                    default:
                        break;
                    }
                }
                break;

            case CUser_field::TData::e_Str:
                val.clear();
                val.push_back(field->GetData().GetStr());
                break;

            case CUser_field::TData::e_Strs:
                val.clear();
                val.insert(val.end(),
                           field->GetData().GetStrs().begin(),
                           field->GetData().GetStrs().end());
                break;

            default:
                NCBI_THROW(CException, eUnknown,
                           "invalid field type in registry request");
            }
        }
    }
    catch (CException& e) {
        LOG_POST(Error << "failed to access registry key "
                 << key << ": " << e.GetMsg());
        val.clear();
    }
    catch (std::exception&) {
        val.clear();
    }
}


void CGuiRegistry::GetStringList(const string& key,
                                 list<string>& val,
                                 const string& delim) const
{
    vector<string> v;
    GetStringVec(key, v, delim);

    val.clear();
    std::copy(v.begin(), v.end(), back_inserter(val));
}


void CGuiRegistry::GetOctetStringVec(const string& key,
                                     vector<string>& val,
                                     const string& delim) const
{
    try {
        CConstRef<CUser_field> field = GetField(key, delim);
        if (field) {
            switch (field->GetData().Which()) {
            case CUser_field::TData::e_Fields:
                val.clear();
                ITERATE (CUser_field::TData::TFields, iter, field->GetData().GetFields()) {
                    const CUser_field& f = **iter;
                    switch (f.GetData().Which()) {
                    case CUser_field::TData::e_Os:
                        {{
                            val.push_back(string());
                            string& s = val.back();
                            s.reserve(f.GetData().GetOs().size());
                            std::copy(f.GetData().GetOs().begin(),
                                      f.GetData().GetOs().end(),
                                      back_inserter(s));
                        }}
                        break;

                    case CUser_field::TData::e_Oss:
                        ITERATE (CUser_field::TData::TOss, i, f.GetData().GetOss()) {
                            val.push_back(string());
                            string& s = val.back();
                            s.reserve((*i)->size());
                            std::copy((*i)->begin(),
                                      (*i)->end(),
                                      back_inserter(s));
                            s.reserve((*i)->size());
                            std::copy((*i)->begin(),
                                      (*i)->end(),
                                      back_inserter(s));
                        }
                        break;

                    default:
                        break;
                    }
                }
                break;

            case CUser_field::TData::e_Os:
                {{
                    val.clear();
                    val.push_back(string());
                    string& s = val.back();
                    s.reserve(field->GetData().GetOs().size());
                    std::copy(field->GetData().GetOs().begin(),
                              field->GetData().GetOs().end(),
                              back_inserter(s));
                }}
                break;

            case CUser_field::TData::e_Oss:
                val.clear();
                ITERATE (CUser_field::TData::TOss, i, field->GetData().GetOss()) {
                    val.push_back(string());
                    string& s = val.back();
                    s.reserve((*i)->size());
                    std::copy((*i)->begin(),
                              (*i)->end(),
                              back_inserter(s));
                }
                break;

            default:
                NCBI_THROW(CException, eUnknown,
                           "invalid field type in registry request");
            }
        }
    }
    catch (CException& e) {
        LOG_POST(Error << "failed to access registry key "
                 << key << ": " << e.GetMsg());
        val.clear();
    }
    catch (std::exception&) {
        val.clear();
    }
}


void CGuiRegistry::Set(const string& key, int val,
                       const string& delim)
{
    Set(GetWritableSite(), key, val, delim);
}


void CGuiRegistry::Set(const string& key, double val,
                       const string& delim)
{
    Set(GetWritableSite(), key, val, delim);
}


void CGuiRegistry::Set(const string& key, bool val,
                       const string& delim)
{
    Set(GetWritableSite(), key, val, delim);
}


void CGuiRegistry::Set(const string& key, const string& val,
                       const string& delim, TFlags flags)
{
    Set(GetWritableSite(), key, val, delim, flags);
}


void CGuiRegistry::Set(const string& key, const char* val,
                       const string& delim, TFlags flags)
{
    Set(key, string(val), delim, flags);
}


void CGuiRegistry::Set(const string& key,
                       const vector<int>& val,
                       const string& delim)
{
    Set(GetWritableSite(), key, val, delim);
}


void CGuiRegistry::Set(const string& key,
                       const vector<double>& val,
                       const string& delim)
{
    Set(GetWritableSite(), key, val, delim);
}


void CGuiRegistry::Set(const string& key,
                       const vector<string>& val,
                       const string& delim,
                       TFlags flags)
{
    Set(GetWritableSite(), key, val, delim, flags);
}


void CGuiRegistry::Set(const string& key,
                       const list<string>& val,
                       const string& delim,
                       TFlags flags)
{
    Set(GetWritableSite(), key, val, delim, flags);
}


void CGuiRegistry::Set(int site, const string& key, int val,
                       const string& delim)
{
    CRegistryFile& write_site = x_SetRegistry(site);
    CRegistryFile::TGuard LOCK(write_site);
    CRef<CUser_field> field = write_site.SetFieldToValue(key, delim);
    field->SetData().SetInt(val);
}


void CGuiRegistry::Set(int site, const string& key, double val,
                       const string& delim)
{
    CRegistryFile& write_site = x_SetRegistry(site);
    CRegistryFile::TGuard LOCK(write_site);
    CRef<CUser_field> field = write_site.SetFieldToValue(key, delim);
    field->SetData().SetReal(val);
}


void CGuiRegistry::Set(int site, const string& key, bool val,
                       const string& delim)
{
    CRegistryFile& write_site = x_SetRegistry(site);
    CRegistryFile::TGuard LOCK(write_site);
    CRef<CUser_field> field = write_site.SetFieldToValue(key, delim);
    field->SetData().SetBool(val);
}


void CGuiRegistry::Set(int site, const string& key, const string& val,
                       const string& delim, TFlags flags)
{
    CRegistryFile& write_site = x_SetRegistry(site);
    CRegistryFile::TGuard LOCK(write_site);
    CRef<CUser_field> field = write_site.SetFieldToValue(key, delim);
    if (flags & fUseOctetString) {
        vector<char>& v = field->SetData().SetOs();
        v.clear();
        v.insert(v.end(), val.begin(), val.end());
    } else {
        field->SetData().SetStr(val);
    }
}


void CGuiRegistry::Set(int site, const string& key, const char* val,
                       const string& delim, TFlags flags)
{
    Set(site, key, string(val), delim, flags);
}


void CGuiRegistry::Set(int site, const string& key,
                       const vector<int>& val,
                       const string& delim)
{
    CRegistryFile& write_site = x_SetRegistry(site);
    CRegistryFile::TGuard LOCK(write_site);
    CRef<CUser_field> field = write_site.SetFieldToValue(key, delim);
    field->SetData().SetInts() = val;
}


void CGuiRegistry::Set(int site, const string& key,
                       const vector<double>& val,
                       const string& delim)
{
    CRegistryFile& write_site = x_SetRegistry(site);
    CRegistryFile::TGuard LOCK(write_site);
    CRef<CUser_field> field = write_site.SetFieldToValue(key, delim);
    field->SetData().SetReals() = val;
}


void CGuiRegistry::Set(int site, const string& key,
                       const vector<string>& val,
                       const string& delim,
                       TFlags flags)
{
    CRegistryFile& write_site = x_SetRegistry(site);
    CRegistryFile::TGuard LOCK(write_site);
    CRef<CUser_field> field = write_site.SetFieldToValue(key, delim);
    if (flags & fUseOctetString) {
        field->SetData().Reset();
        CUser_field::TData::TOss& oss = field->SetData().SetOss();
        ITERATE (vector<string>, i, val) {
            auto_ptr< vector<char> > p(new vector<char>);
            p->insert(p->end(), i->begin(), i->end());
            oss.push_back(p.release());
        }
    } else {
        CUser_field_Base::C_Data::TStrs& dest = field->SetData().SetStrs();
        dest.clear();
        ITERATE( vector<string>, v, val) {
            dest.push_back(*v);
        }


    }
}


void CGuiRegistry::Set(int site, const string& key,
                       const list<string>& val,
                       const string& delim,
                       TFlags flags)
{
    vector<string> v;
    ITERATE(list<string>, it, val)  {
        v.push_back(*it);
    }
    Set(site, key, v, delim, flags);
}


static CGuiRegistry::EPolicy s_PolicyFromMeta(const CUser_field& field)
{
    CGuiRegistry::EPolicy policy =  CGuiRegistry::eOverride;
    CConstRef<CUser_field> merge_field = field.GetFieldRef("MergePolicy");
    if (merge_field) {
        string str = merge_field->GetData().GetStr();
        NStr::ToLower(str);
        if (str == "append") {
            policy = CGuiRegistry::eMergeAppend;
        } else if (str == "prepend") {
            policy = CGuiRegistry::eMergePrepend;
        }
    }
    return policy;
}


CRef<CUser_field> CGuiRegistry::SetField(const string& key,
                                         const string& delim)
{
    return SetField(GetWritableSite(), key, delim);
}


CRef<CUser_field> CGuiRegistry::SetField(int site, const string& key,
                                         const string& delim)
{
    CRegistryFile& write_site = x_SetRegistry(site);
    CRegistryFile::TGuard LOCK(write_site);
    return write_site.SetField(key, delim);
}


CRef<CUser_field> CGuiRegistry::SetFieldToValue(const string& key,
                                                const string& delim)
{
    return SetFieldToValue(GetWritableSite(), key, delim);
}


CRef<CUser_field> CGuiRegistry::SetFieldToValue(int site, const string& key,
                                                const string& delim)
{
    CRegistryFile& write_site = x_SetRegistry(site);
    CRegistryFile::TGuard LOCK(write_site);
    return write_site.SetFieldToValue(key, delim);
}

bool CGuiRegistry::DeleteField(const string& key,
                               const string& delim)
{
    return DeleteField(GetWritableSite(), key, delim);
}


bool CGuiRegistry::DeleteField(int site, const string& key,
                               const string& delim)
{
    CRegistryFile& write_site = x_SetRegistry(site);
    CRegistryFile::TGuard LOCK(write_site);
    return write_site.DeleteField(key, delim);
}


CConstRef<CUser_field> CGuiRegistry::GetField(const string& key,
                                              const string& delim) const
{
    list< CConstRef<CUser_field> > fields;

    string meta_label(key);
    EPolicy merge_policy = eOverride;

    ITERATE (TRegistries, iter, m_Registries) {
        CConstRef<CUser_field> field
            (iter->second->GetField(key, delim));
        if (field) {
            fields.push_back(field);
        }

        field = iter->second->GetField(meta_label + kMetaSuffix, delim);
        if (field) {
            merge_policy = s_PolicyFromMeta(*field);
        }
    }

    if (fields.size() == 0) {
        return CConstRef<CUser_field>();
        // NCBI_THROW(CException, eUnknown, string("failed to find field: ") + section + delim + key);
    }

    //
    // all fields should be objects, so look for sub-fields that
    // meet the criteria we want
    //
    return CGuiRegistryUtil::ResolveFieldPolicy(fields, merge_policy);
}


bool CGuiRegistry::HasField(const string& key,
                            const string& delim) const
{
    ITERATE (TRegistries, iter, m_Registries) {
        if (iter->second->HasField(key, delim)) {
            return true;
        }
    }
    return false;
}


CRegistryFile& CGuiRegistry::x_GetLocal()
{
    return x_SetRegistry(ePriority_Local);
}



/// return the registry at a particular priority.
/// Do not create if it does not exist.
CRef<CRegistryFile> CGuiRegistry::x_GetRegistryRef(int priority) const
{
    TRegistries::const_iterator it = m_Registries.find(priority);
    if (it != m_Registries.end()) {
        return it->second;
    }
    return CRef<CRegistryFile>();
}


/// return a registry at a particular priority, creating it if need be.
CRegistryFile&  CGuiRegistry::x_SetRegistry(int priority)
{
    CRef<CRegistryFile> reg_ref = m_Registries[priority];
    if (! reg_ref) {
        reg_ref.Reset(new CRegistryFile);
        m_Registries[priority] = reg_ref;
    }
    return *reg_ref;
}


CGuiRegistry& CGuiRegistry::GetInstance()
{
    static CRef<CGuiRegistry> s_Registry;
    DEFINE_STATIC_MUTEX(s_RegistryMutex);
    if ( !s_Registry ) {
        CMutexGuard LOCK(s_RegistryMutex);
        if ( !s_Registry ) {
            s_Registry.Reset(new CGuiRegistry());
        }
    }
    return *s_Registry;
}

void CGuiRegistry::x_AppendToView(CGuiRegistry::TReadView& view,
                                  const string& section,
                                  const string& delim) const
{
    if (view.m_Registry.GetPointer() != this)
        return;

    ITERATE (TRegistries, iter, m_Registries) {
        const CRegistryFile& file = *iter->second;
        CConstRef<CUser_field> section_obj = file.GetField(section, delim);
        if (section_obj) {
            view.m_Fields.push_back(section_obj);
        }
    }
}


CGuiRegistry::TReadView
CGuiRegistry::GetReadView(const string& section, const string& delim) const
{
    TReadView view;
    view.m_Section = section;
    view.m_Registry.Reset(const_cast<CGuiRegistry*>(this));

    x_AppendToView(view, section, delim);

    return view;
}


CGuiRegistry::TReadView
CGuiRegistry::GetReadView(const string& section,
                          const vector<string>& subsections,
                          const string& delim) const
{
    TReadView view;
    view.m_Registry.Reset(const_cast<CGuiRegistry*>(this));

    // the section for a view made with subsections ends with
    // the most significant subsection.
    string str(section);
    if ( !section.empty() ) {
        str += delim;
    }
    view.m_Section = str;
    if ( ! subsections.empty()) {
        view.m_Section += subsections.back();
    }

    vector<string>::const_reverse_iterator riter(subsections.end());
    vector<string>::const_reverse_iterator rend (subsections.begin());
    for ( ;  riter != rend;  ++riter) {
        string key(*riter);
        if ( !key.empty()) {
            x_AppendToView(view, str + key, delim);
        }
    }

    return view;
}


CGuiRegistry::TReadWriteView
CGuiRegistry::GetReadWriteView(const string& section, const string& delim)
{
    TReadWriteView view;
    view.m_Section = section;
    view.m_Registry.Reset(this);
    view.m_WriteSite = GetWritableSite();

    x_AppendToView(view, section, delim);

    return view;
}

CGuiRegistry::TReadWriteView
CGuiRegistry::GetReadWriteView(const string& section, int writeSite, const string& delim)
{
    TReadWriteView view;
    view.m_Section = section;
    view.m_Registry.Reset(this);
    view.m_WriteSite = writeSite;

    x_AppendToView(view, section, delim);

    return view;
}

CGuiRegistry::TReadWriteView
CGuiRegistry::GetReadWriteView(const string& section,
                               const vector<string>& subsections,
                               const string& delim)
{
    TReadWriteView view;
    view.m_Registry.Reset(const_cast<CGuiRegistry*>(this));
    view.m_WriteSite = GetWritableSite();

    // the section for a view made with subsections ends with
    // the most significant subsection.
    string str(section);
    if ( !section.empty() ) {
        str += delim;
    }
    view.m_Section = str;
    if ( ! subsections.empty()) {
        view.m_Section += subsections.back();
    }

    vector<string>::const_reverse_iterator riter(subsections.end());
    vector<string>::const_reverse_iterator rend (subsections.begin());
    for ( ;  riter != rend;  ++riter) {
        string key(*riter);
        if ( !key.empty()) {
            x_AppendToView(view, str + key, delim);
        }
    }

    return view;
}


/////////////////////////////////////////////////////////////////////////////
///
/// CReadView

CGuiRegistry::CReadView::~CReadView()
{
}

void
CGuiRegistry::CReadView::x_AppendToReadView(CGuiRegistry::TReadView& view,
                                            const string& section,
                                            const string& delim) const
{
    if (m_Registry != view.m_Registry) {
        return;
    }

    // also should assert or check that this's m_Section is the head of view's
    // m_Section.  and that 'section' has the same number of elements that as
    // the difference between the two view's sections (usually one).  Instead
    // I'm going to make this protected since it is for implementation sharing
    // anyway.

    ITERATE (TFields, iter, m_Fields) {
        const CUser_field& obj = **iter;
        CConstRef<CUser_field> section_obj =
            obj.GetFieldRef(section, delim);
        if (section_obj) {
            view.m_Fields.push_back(section_obj);
        }
    }
}


CGuiRegistry::TReadView
CGuiRegistry::CReadView::GetReadView(const string& section,
                                     const string& delim) const
{
    TReadView view;
    view.m_Section = m_Section;
    if (!section.empty()) {
        view.m_Section += delim;
        view.m_Section += section;
    }
    view.m_Registry = m_Registry;

    x_AppendToReadView(view, section, delim);
    return view;
}


CGuiRegistry::TReadView
CGuiRegistry::CReadView::GetReadView(const string& section,
                                     const vector<string>& subsections,
                                     const string& delim) const
{
    TReadView view;
    view.m_Registry = m_Registry;
    view.m_Section = m_Section;
    if (!section.empty()) {
        view.m_Section += delim;
        view.m_Section += section;
    }
    // the section for a view made with subsections ends with
    // the most significant subsection.
    if ( ! subsections.empty()) {
        view.m_Section += delim;
        view.m_Section += subsections.back();
    }

    string sec(section);
    if (! sec.empty()) {
        sec += delim;
    }

    vector<string>::const_reverse_iterator riter(subsections.end());
    vector<string>::const_reverse_iterator rend (subsections.begin());
    for ( ;  riter != rend;  ++riter) {
        string key(*riter);
        if ( !key.empty()) {
            x_AppendToReadView(view, sec + key, delim);
        }
    }

    return view;
}


CConstRef<objects::CUser_field>
CGuiRegistry::CReadView::GetField(const string& key,
                                  const string& delim) const
{
    list< CConstRef<CUser_field> > fields;

    CGuiRegistry::EPolicy merge_policy = CGuiRegistry::eDefault;

    ITERATE (TFields, iter, m_Fields) {
        const CUser_field& obj = **iter;
        CConstRef<CUser_field> field_ref = obj.GetFieldRef(key, delim);
        if (field_ref) {
            fields.push_back(field_ref);
        }

        /// access the meta-information for this key, last one takes precedence.
        CConstRef<CUser_field> meta =
            obj.GetFieldRef(key + CGuiRegistry::kMetaSuffix, delim);
        if (meta) {
            merge_policy = s_PolicyFromMeta(*meta);
        }
    }
    return CGuiRegistryUtil::ResolveFieldPolicy(fields, merge_policy);
}


/// type-specific field accessors
int CGuiRegistry::CReadView::GetInt(const string& key,
                                    int default_val,
                                    const string& delim) const
{
    CConstRef<CUser_field> field = GetField(key, delim);
    if (field.GetPointer()  &&  field->GetData().IsInt()) {
        return field->GetData().GetInt();
    } else {
        return default_val;
    }
}


double CGuiRegistry::CReadView::GetReal(const string& key,
                                        double default_val,
                                        const string& delim) const
{
    CConstRef<CUser_field> field = GetField(key, delim);
    if (field.GetPointer()  &&  field->GetData().IsReal()) {
        return field->GetData().GetReal();
    } else {
        return default_val;
    }
}


bool CGuiRegistry::CReadView::GetBool(const string& key,
                                      bool default_val,
                                      const string& delim) const
{
    CConstRef<CUser_field> field = GetField(key, delim);
    if (field.GetPointer()  &&  field->GetData().IsBool()) {
        return field->GetData().GetBool();
    } else {
        return default_val;
    }
}


string CGuiRegistry::CReadView::GetString(const string& key,
                                          const string& default_val,
                                          const string& delim) const
{
    CConstRef<CUser_field> field = GetField(key, delim);
    if (field.GetPointer()  &&  field->GetData().IsStr()) {
        return field->GetData().GetStr();
    } else {
        return default_val;
    }
}


string CGuiRegistry::CReadView::GetOctetString(const string& key,
                                               const string& default_val,
                                               const string& delim) const
{
    CConstRef<CUser_field> field = GetField(key, delim);
    if (field.GetPointer()  &&  field->GetData().IsOs()) {
        string s;
        s.reserve(field->GetData().GetOs().size());
        std::copy(field->GetData().GetOs().begin(),
                  field->GetData().GetOs().end(),
                  back_inserter(s));
        return s;
    } else {
        return default_val;
    }
}


void CGuiRegistry::CReadView::GetIntVec(const string& key,
                                        vector<int>& val,
                                        const string& delim) const
{
    val.clear();
    CConstRef<CUser_field> field = GetField(key, delim);
    if (field) {
        if (field->GetData().IsInt()) {
            val.push_back(field->GetData().GetInt());
        } else if (field->GetData().IsInts()) {
            val = field->GetData().GetInts();
        } else if (field->GetData().IsFields()) {
            ITERATE (CUser_field::TData::TFields, iter, field->GetData().GetFields()) {
                const CUser_field& f = **iter;
                if (f.GetData().IsInt()) {
                    val.push_back(f.GetData().GetInt());
                } else if (f.GetData().IsInts()) {
                    val.insert(val.end(),
                               f.GetData().GetInts().begin(),
                               f.GetData().GetInts().end());
                }
            }
        } else {
            LOG_POST(Warning << "warning: registry key "
                     << key << " is not of type int or intvec");
        }
    }
}


void CGuiRegistry::CReadView::GetRealVec(const string& key,
                                         vector<double>& val,
                                         const string& delim) const
{
    val.clear();
    CConstRef<CUser_field> field = GetField(key, delim);
    if (field) {
        if (field->GetData().IsReal()) {
            val.push_back(field->GetData().GetReal());
        } else if (field->GetData().IsReals()) {
            val = field->GetData().GetReals();
        } else if (field->GetData().IsFields()) {
            ITERATE (CUser_field::TData::TFields, iter, field->GetData().GetFields()) {
                const CUser_field& f = **iter;
                if (f.GetData().IsReal()) {
                    val.push_back(f.GetData().GetReal());
                } else if (f.GetData().IsReals()) {
                    val.insert(val.end(),
                               f.GetData().GetReals().begin(),
                               f.GetData().GetReals().end());
                }
            }
        } else {
            LOG_POST(Warning << "warning: registry key "
                     << key << " is not of type real or realvec");
        }
    }
}


void CGuiRegistry::CReadView::GetStringVec(const string& key,
                                           vector<string>& val,
                                           const string& delim) const
{
    val.clear();
    CConstRef<CUser_field> field = GetField(key, delim);
    if (field) {
        if (field->GetData().IsStr()) {
            val.push_back(field->GetData().GetStr());
        } else if (field->GetData().IsStrs()) {
            val.insert(val.end(),
                       field->GetData().GetStrs().begin(),
                       field->GetData().GetStrs().end());
        } else if (field->GetData().IsFields()) {
            ITERATE (CUser_field::TData::TFields, iter, field->GetData().GetFields()) {
                const CUser_field& f = **iter;
                if (f.GetData().IsStr()) {
                    val.push_back(f.GetData().GetStr());
                } else if (f.GetData().IsStrs()) {
                    val.insert(val.end(),
                               f.GetData().GetStrs().begin(),
                               f.GetData().GetStrs().end());
                }
            }
        } else {
            LOG_POST(Warning << "warning: registry key "
                     << key << " is not of type string or stringvec");
        }
    }
}


void CGuiRegistry::CReadView::GetOctetStringVec(const string& key,
                                                vector<string>& val,
                                                const string& delim) const
{
    val.clear();
    CConstRef<CUser_field> field = GetField(key, delim);
    if (field) {
        switch (field->GetData().Which()) {
        case CUser_field::TData::e_Fields:
            val.clear();
            ITERATE (CUser_field::TData::TFields, iter, field->GetData().GetFields()) {
                const CUser_field& f = **iter;
                switch (f.GetData().Which()) {
                case CUser_field::TData::e_Os:
                    {{
                        val.push_back(string());
                        string& s = val.back();
                        s.reserve(f.GetData().GetOs().size());
                        std::copy(f.GetData().GetOs().begin(),
                                  f.GetData().GetOs().end(),
                                  back_inserter(s));
                    }}
                    break;

                case CUser_field::TData::e_Oss:
                    ITERATE (CUser_field::TData::TOss, i, f.GetData().GetOss()) {
                        val.push_back(string());
                        string& s = val.back();
                        s.reserve((*i)->size());
                        std::copy((*i)->begin(),
                                  (*i)->end(),
                                  back_inserter(s));
                        s.reserve((*i)->size());
                        std::copy((*i)->begin(),
                                  (*i)->end(),
                                  back_inserter(s));
                    }
                    break;

                default:
                    break;
                }
            }
            break;

        case CUser_field::TData::e_Os:
            {{
                val.clear();
                val.push_back(string());
                string& s = val.back();
                s.reserve(field->GetData().GetOs().size());
                std::copy(field->GetData().GetOs().begin(),
                          field->GetData().GetOs().end(),
                          back_inserter(s));
            }}
            break;

        case CUser_field::TData::e_Oss:
            val.clear();
            ITERATE (CUser_field::TData::TOss, i, field->GetData().GetOss()) {
                val.push_back(string());
                string& s = val.back();
                s.reserve((*i)->size());
                std::copy((*i)->begin(),
                          (*i)->end(),
                          back_inserter(s));
            }
            break;

        default:
            NCBI_THROW(CException, eUnknown,
                       "invalid field type in registry request");
        }
    }
}

void CGuiRegistry::CReadView::GetStringList(const string& key,
                                            list<string>& val,
                                            const string& delim) const
{
    val.clear();

    vector<string> v;
    GetStringVec(key, v, delim);
    for( size_t i = 0;  i < v.size();  i++ )    {
        val.push_back(v[i]);
    }
}


void CGuiRegistry::CReadView::GetTopKeys(TKeys& keys,
                                         const string& starting_section,
                                         const string& delim) const
{
    x_GetKeys(keys, starting_section, delim, false);
}


void CGuiRegistry::CReadView::GetKeys(TKeys& keys,
                                      const string& starting_section,
                                      const string& delim) const
{
    x_GetKeys(keys, starting_section, delim, true);
}


void CGuiRegistry::CReadView::x_GetKeys(TKeys& keys,
                                        const string& starting_section,
                                        const string& delim,
                                        bool recurse ) const
{
    if (starting_section.empty()) {
        typedef map<string, SKeyInfo> TKeyMap;
        TKeyMap key_map;    // used to sort the final list.

        TFields::const_reverse_iterator riter(m_Fields.rbegin());
        TFields::const_reverse_iterator rend (m_Fields.rend());

        for (; riter != rend; ++riter) {
            const CUser_field& obj = **riter;
            if ( obj.GetData().IsFields() ) {
                TKeys temp;
                ITERATE (CUser_field::TData::TFields, sub_iter,
                         obj.GetData().GetFields()) {
                    s_ExtractKeys(**sub_iter, "", temp, recurse);
                }
                ITERATE(TKeys, key_it, temp) {
                    key_map[key_it->key] = *key_it;
                }
            }
        }
        keys.clear();
        ITERATE (TKeyMap, iter, key_map) {
            SKeyInfo key = iter->second;
            // access is meaningless here since we can't tell what site a given key is in.
            // set it something valid.
            key.access = CGuiRegistry::eAccess_ReadOnly;
            keys.push_back(key);
        }
    } else {
        CGuiRegistry::CReadView subview = GetReadView(starting_section, delim);
        subview.x_GetKeys(keys, kEmptyStr, delim, recurse);
        NON_CONST_ITERATE(CGuiRegistry::TKeys, key_it, keys){
            key_it->key = starting_section + delim + key_it->key;
        }
    }
}

bool CGuiRegistry::CReadView::HasField(const string& key, const string& delim) const
{
    ITERATE (TFields, iter, m_Fields) {
        const CUser_field& obj = **iter;
        if (obj.HasField(key, delim))
            return true;
    }
    return false;
}


bool CGuiRegistry::CReadView::IsEmpty()
{
    return m_Fields.empty();
}


void CGuiRegistry::CReadView::DumpAll(CNcbiOstream& ostr) const
{
    CGuiRegistry::TKeys keys;
    GetKeys(keys);

    ostr << keys.size() << " keys:" << endl;
    ITERATE (CGuiRegistry::TKeys, iter, keys) {
        const CGuiRegistry::SKeyInfo& key = *iter;

        ostr << key.key << "|" ;

        vector<string> strs;
        vector<int> ints;
        vector<double> reals;
        switch (key.type) {
        case CUser_field::TData::e_Bool:
            ostr << "bool|" << NStr::BoolToString(GetBool(key.key));
            break;
        case CUser_field::TData::e_Int:
            ostr << "int|" << GetInt(key.key);
            break;
        case CUser_field::TData::e_Ints:
            ostr << "ints|";
            GetIntVec(key.key, ints);
            copy(ints.begin(), ints.end(), ostream_iterator<int>(ostr, ";"));
            break;
        case CUser_field::TData::e_Real:
            ostr << "real|" << GetReal(key.key);
            break;
        case CUser_field::TData::e_Reals:
            ostr << "reals|";
            GetRealVec(key.key, reals);
            copy(reals.begin(), reals.end(), ostream_iterator<double>(ostr, ";"));
            break;
        case CUser_field::TData::e_Str:
            ostr << "string|" << GetString(key.key);;
            break;
        case CUser_field::TData::e_Strs:
            ostr << "strings|";
            GetStringVec(key.key, strs);
            copy(strs.begin(), strs.end(), ostream_iterator<string>(ostr, ";"));
            break;
        case CUser_field::TData::e_Fields:
            ostr << "subkey|";
            break;
        default:
            ostr << "unknown|";
            break;
        }
        ostr << endl;

    }
}


/////////////////////////////////////////////////////////////////////////////
///
/// CReadWriteView

CGuiRegistry::CReadWriteView::~CReadWriteView()
{
}


CGuiRegistry::TReadWriteView
CGuiRegistry::CReadWriteView::GetReadWriteView(const string& section,
                                               const string& delim)
{
    string sec(m_Section);
    sec += delim;
    sec += section;
    return m_Registry->GetReadWriteView(sec, m_WriteSite, delim);
}


/// retrieve a writeable field at a particular section and key
CRef<CUser_field>
CGuiRegistry::CReadWriteView::SetField(const string& key,
                                       const string& delim)
{
    string sec(m_Section);
    sec += delim;
    sec += key;
    CRef<CUser_field> section_obj = m_Registry->SetField(m_WriteSite, sec, delim);
    return section_obj;
}


CRef<CUser_field>
CGuiRegistry::CReadWriteView::SetFieldToValue(const string& key,
                                              const string& delim)
{
    CRef<CUser_field> section_obj = SetField(key, delim);
    if (section_obj->GetData().IsFields()) {
        // field is a place for subkeys, not for a value.
        NCBI_THROW(CException, eUnknown,
                   "Too few components in key \"" + key + "\"");
    }
    return section_obj;
}


bool CGuiRegistry::CReadWriteView::DeleteField(const string& key,
                                               const string& delim)
{
    string sec(m_Section);
    sec += delim;
    sec += key;
    return m_Registry->DeleteField(m_WriteSite, sec, delim);
}


/// set specific values
void CGuiRegistry::CReadWriteView::Set(const string& key,
                                       int val,
                                       const string& delim)
{
    CRef<CUser_field> field = SetFieldToValue(key, delim);
    field->SetData().SetInt(val);
}


void CGuiRegistry::CReadWriteView::Set(const string& key,
                                       double val,
                                       const string& delim)
{
    CRef<CUser_field> field = SetFieldToValue(key, delim);
    field->SetData().SetReal(val);
}


void CGuiRegistry::CReadWriteView::Set(const string& key,
                                       bool val,
                                       const string& delim)
{
    CRef<CUser_field> field = SetFieldToValue(key, delim);
    field->SetData().SetBool(val);
}


void CGuiRegistry::CReadWriteView::Set(const string& key,
                                       const string& val,
                                       const string& delim,
                                       TFlags flags)
{
    CRef<CUser_field> field = SetFieldToValue(key, delim);
    if (flags & fUseOctetString) {
        vector<char>& v = field->SetData().SetOs();
        v.clear();
        v.insert(v.end(), val.begin(), val.end());
    } else {
        field->SetData().SetStr(val);
    }
}


void CGuiRegistry::CReadWriteView::Set(const string& key,
                                       const char* val,
                                       const string& delim,
                                       TFlags flags)
{
    Set(key, string(val), delim, flags);
}


void CGuiRegistry::CReadWriteView::Set(const string& key,
                                       const vector<int>& val,
                                       const string& delim)
{
    CRef<CUser_field> field = SetFieldToValue(key, delim);
    field->SetData().SetInts() = val;
}


void CGuiRegistry::CReadWriteView::Set(const string& key,
                                       const vector<double>& val,
                                       const string& delim)
{
    CRef<CUser_field> field = SetFieldToValue(key, delim);
    field->SetData().SetReals() = val;
}


void CGuiRegistry::CReadWriteView::Set(const string& key,
                                       const vector<string>& val,
                                       const string& delim,
                                       TFlags flags)
{
    CRef<CUser_field> field = SetFieldToValue(key, delim);
    if (flags & fUseOctetString) {
        field->SetData().Reset();
        CUser_field::TData::TOss& oss = field->SetData().SetOss();
        ITERATE (vector<string>, i, val) {
            auto_ptr< vector<char> > p(new vector<char>);
            p->insert(p->end(), i->begin(), i->end());
            oss.push_back(p.release());
        }
    } else {
        CUser_field_Base::C_Data::TStrs& dest = field->SetData().SetStrs();
        dest.clear();
        ITERATE( vector<string>, v, val) {
            dest.push_back(*v);
        }
    }
}


void CGuiRegistry::CReadWriteView::Set(const string& key,
                                       const list<string>& val,
                                       const string& delim,
                                       TFlags flags)
{
    vector<string> v;
    v.reserve(val.size());
    ITERATE (list<string>, it, val) {
        v.push_back(*it);
    }
    Set(key, v, delim, flags);
}


/////////////////////////////////////////////////////////////////////////////
///
/// CGuiRegistryUtil


/// recursively navigate an object tree, retrieving an existing named key
CConstRef<CUser_field>
CGuiRegistryUtil::GetFieldAt(const CUser_object& obj,
                             const string& key,
                             const string& delim)
{
#ifdef _DEBUG
    {{
        _ASSERT(key.find(delim + delim) == string::npos);
        _ASSERT(key.find_first_of(delim) != 0);
        _ASSERT(key.size()  &&  key.find_first_of(delim) != key.size() - 1);
    }}
#endif

    string first_key(key);
    string remainder;
    string::size_type pos = key.find_first_of(delim);
    if (pos != string::npos) {
        first_key = key.substr(0, pos);
        remainder = key.substr(pos + 1, key.length() - pos - 1);
    }

    CConstRef<CUser_field>  field_ref;
    ITERATE(CUser_object::TData, field_iter, obj.GetData()) {
        const CUser_field& f = **field_iter;
        if (f.GetLabel().GetStr() == first_key) {
            if (remainder.empty()) {
                field_ref.Reset(&f);
                break;
            } else if (f.GetData().IsFields()) {
                field_ref = f.GetFieldRef(remainder, delim);
                if (field_ref) {
                    break;
                }
            }
        }
    }
    return field_ref;
}


/// recursively navigate an object tree, setting keys where necessary to get
/// a named level
CRef<CUser_field>
CGuiRegistryUtil::SetFieldAt(CUser_object& obj,
                             const string& key,
                             const string& delim)
{
#ifdef _DEBUG
    {{
        _ASSERT(key.find(delim + delim) == string::npos);
        _ASSERT(key.find_first_of(delim) != 0);
        _ASSERT(key.size()  &&  key.find_first_of(delim) != key.size() - 1);
    }}
#endif

    string first_key(key);
    string remainder;
    string::size_type pos = key.find_first_of(delim);
    if (pos != string::npos) {
        first_key = key.substr(0, pos);
        remainder = key.substr(pos + 1, key.length() - pos - 1);
    }

    CRef<CUser_field>  field_ref;
    NON_CONST_ITERATE(CUser_object::TData, field_iter, obj.SetData()) {
        CUser_field& f = **field_iter;
        if (f.GetLabel().GetStr() == first_key) {
            if (remainder.empty()) {
                field_ref.Reset(&f);
                break;
            } else {
                field_ref = f.SetFieldRef(remainder, delim);
                if (field_ref) {
                    break;
                }
            }
        }
    }

    if ( !field_ref ) {
        /// create the field for this level
        field_ref.Reset(new CUser_field());
        field_ref->SetLabel().SetStr(first_key);
        obj.SetData().push_back(field_ref);

        if ( !remainder.empty() ) {
            CRef<CUser_field> f = field_ref->SetFieldRef(remainder, delim);
            field_ref = f;
        }
    }

    return field_ref;
}


CConstRef<objects::CUser_field>
CGuiRegistryUtil::ResolveFieldPolicy
(list< CConstRef<objects::CUser_field> >& fields,
 CGuiRegistry::EPolicy merge_policy)
{
    //
    // all fields should be objects, so look for sub-fields that
    // meet the criteria we want
    //
    switch (merge_policy) {
    case CGuiRegistry::eOverride:
        /// return the first item found
        if (fields.size()) {
            return fields.front();
        }
        return CConstRef<CUser_field>();

    case CGuiRegistry::eMergeAppend:
        /// merge into a single field and return
        {{
             CRef<CUser_field> field;
             switch (fields.size()) {
             case 0:
                 return CConstRef<CUser_field>();

             case 1:
                 return fields.front();

             default:
                 field.Reset(new CUser_field());
                 ITERATE (list< CConstRef<CUser_field> >, iter, fields) {
                     CRef<CUser_field> f
                         (const_cast<CUser_field*>(iter->GetPointer()));
                     field->SetData().SetFields().push_back(f);
                 }
                 return CConstRef<CUser_field>(field.Release());
             }
         }}

    case CGuiRegistry::eMergePrepend:
        /// merge into a single field and return
        {{
             CRef<CUser_field> field;
             switch (fields.size()) {
             case 0:
                 return CConstRef<CUser_field>();

             case 1:
                 return fields.front();

             default:
                 std::reverse(fields.begin(), fields.end());
                 field.Reset(new CUser_field());
                 ITERATE (list< CConstRef<CUser_field> >, iter, fields) {
                     CRef<CUser_field> f
                         (const_cast<CUser_field*>(iter->GetPointer()));
                     field->SetData().SetFields().push_back(f);
                 }
                 return CConstRef<CUser_field>(field.Release());
             }
         }}

    default:
        NCBI_THROW(CException, eUnknown,
                   "unhandled registry policy");
    }
}


/// create a key from a section and a subkey
string CGuiRegistryUtil::MakeKey(const string& section, const string& key,
                                 const string& delim)
{
    return section + delim + key;
}


END_NCBI_SCOPE
