/*  $Id: table_data_biotreecontainer.cpp 26779 2012-11-07 14:26:11Z falkrb $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors: Roman Katargin
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>

#include <gui/objutils/interface_registry.hpp>
#include <gui/objutils/table_data_base.hpp>
#include <gui/objutils/label.hpp>

#include <objects/biotree/BioTreeContainer.hpp>
#include <objects/biotree/FeatureDictSet.hpp>
#include <objects/biotree/FeatureDescr.hpp>
#include <objects/biotree/NodeSet.hpp>
#include <objects/biotree/Node.hpp>
#include <objects/biotree/NodeFeatureSet.hpp>
#include <objects/biotree/NodeFeature.hpp>

#include <objmgr/util/sequence.hpp>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

class CTableDataBioTreeContainer : public CObject, public CTableDataBase
{
public:
    static CTableDataBioTreeContainer* CreateObject(SConstScopedObject& object, ICreateParams* params);

    virtual ColumnType GetColumnType(size_t col) const;
    virtual string     GetColumnLabel(size_t col) const;

    virtual size_t GetRowsCount() const;
    virtual size_t GetColsCount() const;

    virtual string GetStringValue(size_t row, size_t col) const;
    virtual long   GetIntValue(size_t row, size_t col) const;
    virtual double GetRealValue(size_t row, size_t col) const;
    virtual SConstScopedObject GetObjectValue(size_t row, size_t col) const;
    CRef<CScope> GetScope() { return m_Scope; }

private:
    struct ColInfo {
        ColInfo() : m_Name(""), m_Type(kNone) {}
        ColInfo(const string& n) : m_Name(n), m_Type(kNone) {}
        ColInfo(const string& n, ITableData::ColumnType t) 
            : m_Name(n), m_Type(t) {}

        string m_Name;
        ITableData::ColumnType m_Type;
    };

    void Init();

    string x_GetNodeFeature(int feature_id, const CNode& node) const;

    CConstRef<CObject> m_Object;
    mutable CRef<objects::CScope> m_Scope;
    vector<pair<int, ColInfo> > m_FeatureDescr;
    vector<CRef<CNode> > m_Nodes;
};

void initCTableDataBioTreeContainer()
{
    CInterfaceRegistry::RegisterFactory(
            typeid(ITableData).name(),
            CBioTreeContainer::GetTypeInfo(),
            new CObjectInterfaceFactory<CTableDataBioTreeContainer>());
}

CTableDataBioTreeContainer* CTableDataBioTreeContainer::CreateObject(SConstScopedObject& object, ICreateParams*)
{
    CTableDataBioTreeContainer* table_data = new CTableDataBioTreeContainer();
    table_data->m_Object = object.object;
    table_data->m_Scope  = object.scope;
    table_data->Init();
    return table_data;
}

void CTableDataBioTreeContainer::Init()
{
    const CBioTreeContainer& biotree = dynamic_cast<const CBioTreeContainer&>(*m_Object);
    const CFeatureDictSet::Tdata& fdict = biotree.GetFdict().Get();
    ITERATE(CFeatureDictSet::Tdata, it, fdict) {
        ITableData::ColumnType t = kNone;

        if (NStr::EqualNocase((*it)->GetName(), "seq-id"))
            t = kObject;

        m_FeatureDescr.push_back(pair<int, ColInfo>((*it)->GetId(), 
            ColInfo((*it)->GetName(),t)));
    }

    const CNodeSet::Tdata& nodes = biotree.GetNodes().Get();
    m_Nodes.resize(nodes.size());

    size_t i = 0;
    ITERATE(CNodeSet::Tdata, it, nodes) {
        // Guess types - look for ints and reals (with blank == 0) so that numbers
        // can have sorting and querying based on numeric rather than string semantics.
        // If a field as any non-numbers (other than blanks), call it a string. No 
        // attempt is made here to look for other IDs (only id is decided by name).
        for (size_t j=0; j<m_FeatureDescr.size(); ++j) {          
            string val = x_GetNodeFeature(m_FeatureDescr[j].first, (*it).GetObject());
            if (val == "")
                continue;

            // kObject was decided by field name and once a string always a string
            // (if we find a string value for a column, no future ints or floats
            //  will change it's type back).
            if (m_FeatureDescr[j].second.m_Type == kObject ||
                m_FeatureDescr[j].second.m_Type == kString)
                continue;

            try {
                // No exception converting to int - it's an int. Can
                // only change to int from 'none' (all values must be ints)
                NStr::StringToInt(val);
                if (m_FeatureDescr[j].second.m_Type == kNone)
                    m_FeatureDescr[j].second.m_Type = kInt;
            }
            catch (CStringException&) {
                try {
                    // No exception converting to double - it's a real number.
                    // only change to real from kInt or kNone.
                    NStr::StringToDouble(val.c_str());
                    if (m_FeatureDescr[j].second.m_Type == kNone ||
                        m_FeatureDescr[j].second.m_Type == kInt) {
                            m_FeatureDescr[j].second.m_Type = kReal;
                    }
                }
                catch (CStringException&) {
                    // It's just a string probably, but
                    // leave it as a number for range errors (overflow/underflow)
                    double d = NStr::StringToDoublePosix(val.c_str());
                    if (errno != ERANGE) {
                        m_FeatureDescr[j].second.m_Type = kString;
                    }
                }
            }
        }


        m_Nodes[i++] = *it;
    }
}

ITableData::ColumnType CTableDataBioTreeContainer::GetColumnType(size_t col) const
{    
    if (col < 2)
        return kInt;

    if (col - 2 < m_FeatureDescr.size())
        return m_FeatureDescr[col-2].second.m_Type;

    return kNone;
}

string CTableDataBioTreeContainer::GetColumnLabel(size_t col) const
{
    if (col == 0)
        return "id";
    else if (col == 1)
        return "parent";

    if (col - 2 < m_FeatureDescr.size())
        return m_FeatureDescr[col - 2].second.m_Name;

    return "";
}

size_t CTableDataBioTreeContainer::GetRowsCount() const
{
    return m_Nodes.size();
}

size_t CTableDataBioTreeContainer::GetColsCount() const
{
    return 2 + m_FeatureDescr.size();
}

string CTableDataBioTreeContainer::GetStringValue(size_t row, size_t col) const
{
    string value;

    if (row >= m_Nodes.size())
        return value;

    if (col < 2) {
        value = NStr::SizetToString(GetIntValue(row, col));
    }
    else {
        if (col - 2 < m_FeatureDescr.size()) {
            try {
                const CNode& node = *m_Nodes[row];
                value = x_GetNodeFeature(m_FeatureDescr[col - 2].first, node);
            }
            catch (const std::exception& e) {
                LOG_POST(Error << "CTableDataBioTreeContainer: " << e.what());
            }
        }
    }
    return value;
}

long CTableDataBioTreeContainer::GetIntValue(size_t row, size_t col) const
{
    long value = 0;
    if (row >= m_Nodes.size())
        return value;

    try {
        const CNode& node = *m_Nodes[row];
        if (col == 0)
            value = node.GetId();
        else if (col == 1) {
            value = node.CanGetParent() ? node.GetParent() : -1;
        }
        else {
            string strval = x_GetNodeFeature(m_FeatureDescr[col - 2].first, node);
            if (NStr::IsBlank(strval))
                return 0;
            value = NStr::StringToLong(strval);
        }
    }
    catch (const std::exception& e) {
        LOG_POST(Error << "CTableDataBioTreeContainer: " << e.what());
    }

    return value;
}

double CTableDataBioTreeContainer::GetRealValue(size_t row, size_t col) const
{
    double value = 0.0;

    if (row >= m_Nodes.size())
        return value;

    if (col < 2)
        return value;

    const CNode& node = *m_Nodes[row];
    string strval = x_GetNodeFeature(m_FeatureDescr[col - 2].first, node);

    try {
        if (NStr::IsBlank(strval))
            return 0.0;
        value = NStr::StringToDouble(strval.c_str());
    }
    catch (const std::exception& e) {
        value = NStr::StringToDoublePosix(strval.c_str());
        if (errno != ERANGE)
            LOG_POST("Value range error: " << strval);
        else
            LOG_POST(Error << "CTableDataBioTreeContainer: " << e.what());
    }

    return value;
}

SConstScopedObject CTableDataBioTreeContainer::GetObjectValue(size_t row, size_t col) const
{
    SConstScopedObject value;

    if (row >= m_Nodes.size())
        return value;

    try {
        if (col >= 2 && NStr::EqualNocase(m_FeatureDescr[col - 2].second.m_Name, "Seq-id")) {
            const CNode& node = *m_Nodes[row];
            string seq_id = x_GetNodeFeature(m_FeatureDescr[col - 2].first, node);
            if (seq_id.empty())
                return value;

            CSeq_id id( seq_id );
            CSeq_id_Handle idh = CSeq_id_Handle::GetHandle( id );
            idh = sequence::GetId( idh, *m_Scope, sequence::eGetId_Best );

            if( idh ){
                CRef<CSeq_id> id_copy( new CSeq_id() );
                id_copy->Assign( *idh.GetSeqId() );
                value.object.Reset(id_copy.GetPointer());
                value.scope = m_Scope;
            }
        }
    }
    catch (const std::exception& e) {
        LOG_POST(Error << "CTableDataBioTreeContainer: " << e.what());
    }

    return value;
}

string CTableDataBioTreeContainer::x_GetNodeFeature(int feature_id, const CNode& node) const
{
    string feature_value;

    if (node.CanGetFeatures()) {
        const CNodeFeatureSet::Tdata& features = node.GetFeatures().Get();
        ITERATE(CNodeFeatureSet::Tdata, it, features) {
            if ((*it)->GetFeatureid() == feature_id) {
                feature_value = (*it)->GetValue();
                break;
            }
        }
    }

    return feature_value;
}

END_NCBI_SCOPE

