/*  $Id: table_data_validate.cpp 28095 2013-05-08 15:01:18Z rudnev $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors: Dmitry Rudnev
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>

#include <gui/objutils/label.hpp>
#include <gui/objutils/table_data_validate.hpp>
#include <gui/objutils/utils.hpp>
#include <gui/utils/app_job.hpp>
#include <gui/utils/app_job_impl.hpp>

#include <objmgr/object_manager.hpp>
#include <objmgr/util/create_defline.hpp>
#include <objmgr/util/sequence.hpp>

#include <objects/seq/Seq_annot.hpp>
#include <objects/seq/Seq_literal.hpp>
#include <objects/seq/Seq_data.hpp>
#include <objects/general/Dbtag.hpp>
#include <objects/general/Object_id.hpp>
#include <objects/general/User_object.hpp>
#include <objects/general/User_field.hpp>
#include <objects/seqfeat/Variation_inst.hpp>
#include <objects/seqfeat/Delta_item.hpp>
#include <objects/seqfeat/Seq_feat.hpp>
#include <objects/seqfeat/BioSource.hpp>
#include <objects/submit/Seq_submit.hpp>
#include <objects/seq/Pubdesc.hpp>

#include <objtools/validator/validator.hpp>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);


class CTableDataValidateJob : public CJobCancelable
{
public:
    // objects[i].object is supposed to contain one of: Seq-entry, Seq-submit, Seq-feat, CSeq_annot_Handle, CBioSource, or CPubdesc 
	// which will be validated, anything else will be ignored
    CTableDataValidateJob(const TConstScopedObjects& objects);
    virtual ~CTableDataValidateJob() {}

    /// @name IAppJob implementation
    /// @{
    virtual EJobState                   Run();
    virtual CConstIRef<IAppJobProgress> GetProgress();
    // CObject is CObjectFor<CValidError::TErrs>
	virtual CRef<CObject>               GetResult();
    virtual CConstIRef<IAppJobError>    GetError();
    virtual string                      GetDescr() const;
    /// @}

private:
    void x_FetchAll();

private:
	TConstScopedObjects m_InputObjects;
	// stores CObjectFor<CValidError::TErrs>
	CRef<CObject> m_ResultErrs;

    size_t m_Total;
    size_t m_Completed;
    CRef<CAppJobError> m_Error;
};


CTableDataValidateJob::CTableDataValidateJob(const TConstScopedObjects& objects)
	:
	m_InputObjects(objects),
	m_Total(100),
    m_Completed(0)
{
	m_ResultErrs.Reset(new CObjectFor<CValidError::TErrs>());
}

CConstIRef<IAppJobProgress> CTableDataValidateJob::GetProgress()
{
	//!! rewrite this using callbacks from CValidator
    CRef<CAppJobProgress> p( new CAppJobProgress() );
    p->SetNormDone( float(m_Completed) / float(m_Total) );
    return CConstIRef<IAppJobProgress>( p.GetPointer() );
}

CRef<CObject> CTableDataValidateJob::GetResult()
{
    return m_ResultErrs;
}


CConstIRef<IAppJobError> CTableDataValidateJob::GetError()
{
    return CConstIRef<IAppJobError>(m_Error.GetPointer());
}


string CTableDataValidateJob::GetDescr() const
{
    return "Validator job";
}


IAppJob::EJobState CTableDataValidateJob::Run()
{
    x_FetchAll();

    return IsCanceled() ? eCanceled : eCompleted;
}


void CTableDataValidateJob::x_FetchAll()
{
    CValidError::TErrs& ValidErrs(dynamic_cast<CObjectFor<CValidError::TErrs>*>(m_ResultErrs.GetPointer())->GetData());
    CRef<CObjectManager> om(CObjectManager::GetInstance());
	validator::CValidator Validator(*om);
	Uint4 validator_opts =
		validator::CValidator::eVal_remote_fetch |
		validator::CValidator::eVal_val_align |
		validator::CValidator::eVal_far_fetch_cds_products |
		validator::CValidator::eVal_far_fetch_mrna_products |
		validator::CValidator::eVal_val_exons |
		validator::CValidator::eVal_need_isojta |
		validator::CValidator::eVal_inference_accns;
	TConstScopedObjects::size_type nInputObjects(m_InputObjects.size());
	for(TConstScopedObjects::size_type iInputObjects = 0; 
		iInputObjects < nInputObjects; 
		++iInputObjects) {
		const type_info& ObjType(typeid(*(m_InputObjects[iInputObjects].object.GetPointer())));
		CConstRef<CValidError> ValidError;
		if(ObjType == typeid(CSeq_entry)) {
			const CSeq_entry* pSeq_entry(dynamic_cast<const CSeq_entry*>(m_InputObjects[iInputObjects].object.GetPointer()));
			ValidError = Validator.Validate(*pSeq_entry,
											 m_InputObjects[iInputObjects].scope.GetPointer(),
											 validator_opts);
		}
		if(ObjType == typeid(CSeq_submit)) {
			ValidError = Validator.Validate(*dynamic_cast<const CSeq_submit*>(m_InputObjects[iInputObjects].object.GetPointer()),
											 m_InputObjects[iInputObjects].scope.GetPointer(),
											 validator_opts);
		}
		if(ObjType == typeid(CSeq_annot_Handle)) {
			ValidError = Validator.Validate(*dynamic_cast<const CSeq_annot_Handle*>(m_InputObjects[iInputObjects].object.GetPointer()),
											validator_opts);
		}
		if(ObjType == typeid(CSeq_feat)) {
			ValidError = Validator.Validate(*dynamic_cast<const CSeq_feat*>(m_InputObjects[iInputObjects].object.GetPointer()), 
											 m_InputObjects[iInputObjects].scope.GetPointer(),
											 validator_opts);
		}
		if(ObjType == typeid(CBioSource)) {
			ValidError = Validator.Validate(*dynamic_cast<const CBioSource*>(m_InputObjects[iInputObjects].object.GetPointer()),
											 m_InputObjects[iInputObjects].scope.GetPointer(),
											 validator_opts);
		}
		if(ObjType == typeid(CPubdesc)) {
			ValidError = Validator.Validate(*dynamic_cast<const CPubdesc*>(m_InputObjects[iInputObjects].object.GetPointer()),
											 m_InputObjects[iInputObjects].scope.GetPointer(),
											 validator_opts);
		}
		if(ValidError->TotalSize()) {
			const CValidError::TErrs& errs(ValidError->GetErrs());
			ValidErrs.insert(ValidErrs.end(), errs.begin(), errs.end());
		}
	}
}


BEGIN_EVENT_MAP( CTableDataValidate, CEventHandler )
    ON_MESSAGE(CAppJobNotification, CAppJobNotification::eStateChanged, &CTableDataValidate::OnJobNotification)
END_EVENT_MAP()

// maximal amount of Validate we should realistically see in a Validate table
const size_t CTableDataValidate::c_MaxRows(100000);

// widths (in chars) of eMessage, eVerboseMessage and eObjectDesc columns
// if wider, they will be wrapped
const size_t CTableDataValidate::c_MessageWidth(50);
const size_t CTableDataValidate::c_ObjectDescWidth(20);

CTableDataValidate::CTableDataValidate(TConstScopedObjects& objects)
:
m_DataLoadStatus(eNotReady)
{
    _ASSERT( !objects.empty() );
    m_Scope = objects.front().scope;

    CRef<CTableDataValidateJob> job(new CTableDataValidateJob(objects));
    CAppJobDispatcher& disp = CAppJobDispatcher::GetInstance();
    try {
        m_hJob = disp.StartJob(*job, "ObjManagerEngine", *this);
    } catch( CAppJobException& e ){
        m_DataLoadStatus = eFailed;
        LOG_POST( Error
            << "CTableDataValidate failed to start job: "
            << e.GetMsg()
        );
        LOG_POST(e.ReportAll());
    }
}

CTableDataValidate::~CTableDataValidate()
{
    if(m_DataLoadStatus == eNotReady) {
        CAppJobDispatcher& disp = CAppJobDispatcher::GetInstance();
        disp.DeleteJob(m_hJob);
    }
}



void CTableDataValidate::OnJobNotification(ncbi::CEvent *evt)
{
    CAppJobNotification* notif = dynamic_cast<CAppJobNotification*>( evt );

    _ASSERT(notif);

    int                 job_id  = notif->GetJobID();
    IAppJob::EJobState  state   = notif->GetState();

    if (job_id != m_hJob) return;

    switch (state) {
        case IAppJob::eCompleted:
        {
            CAppJobDispatcher& disp(CAppJobDispatcher::GetInstance());
            m_Errs.Reset(dynamic_cast<CObjectFor<CValidError::TErrs>*>(disp.GetJobResult(job_id).GetPointer()));
            m_DataLoadStatus = eReady;

            // send a message to the view that data has changed (i.e. ready)
            CTableDataEvent evt(CEvent::eEvent_Message, CTableDataEvent::eDataChanged);
            CTableDataEvent::TDataAttachment* attachment = new CTableDataEvent::TDataAttachment();
            CIRef<ITableData> table_data;
            table_data.Reset(this);
            attachment->SetData(table_data);
            evt.SetAttachment(attachment, CEvent::eRelease);
            Send(&evt, ePool_Parent);
            break;
        }
        case IAppJob::eCanceled:
        case IAppJob::eFailed:
            m_DataLoadStatus = eFailed;
            break;
        default:
            break;
    }
}


// number of entries and order must correspond with CTableDataValidate::EColContents
static const string s_ColNames[] = {
    "Severity",
    "Sequence",
    "Location",
	"Error title",
    "Message",
    "Message explanation",
    "Object Description"
};




ITableData::ColumnType CTableDataValidate::GetColumnType(size_t col) const
{
    if(m_DataLoadStatus != eReady || 
		(m_DataLoadStatus == eReady && m_Errs->GetData().size() == 0)) {
        return ITableData::kString;
    }
    switch(col){
        case eSeqOffset:
            return ITableData::kInt;

        default:
            return ITableData::kString;
    }
    return ITableData::kNone;
}


bool CTableDataValidate::IsHtmlColumn(size_t col) const
{
    if(m_DataLoadStatus != eReady || 
		(m_DataLoadStatus == eReady && m_Errs->GetData().size() == 0)) {
        return false;
    }
	if(col == eMessage || col == eVerboseMessage || col == eObjectDesc) {
		return true;
	}
    return false;
}

string CTableDataValidate::GetColumnLabel(size_t col) const
{
    if((m_DataLoadStatus != eReady ||
		(m_DataLoadStatus == eReady && m_Errs->GetData().size() == 0)) &&
		col == 0) {
        return "Status";
    }
    if (col < CTableDataValidate::eMaxColNum) {
        return s_ColNames[col];
	}

    return kEmptyStr;
}


size_t CTableDataValidate::GetRowsCount() const
{
    return (m_DataLoadStatus == eReady && m_Errs->GetData().size() > 0) ?
				m_Errs->GetData().size() : 1;
}


size_t CTableDataValidate::GetColsCount() const
{
    return (size_t)CTableDataValidate::eMaxColNum;
}


string CTableDataValidate::GetStringValue(size_t row, size_t col) const
{
    if(m_DataLoadStatus == eNotReady) {
        return col == 0 ? "Validating..." : "";
    }
    if(m_DataLoadStatus == eFailed) {
        return col == 0 ? "Validation failed!" : "";
    }
    if(m_DataLoadStatus == eReady && m_Errs->GetData().size() == 0) {
        return col == 0 ? "Validation complete, no errors found!" : "";
    }
    if(row+1 == c_MaxRows) {
        return "Warning: Only the first " + NStr::NumericToString(c_MaxRows) + " rows can be shown";
    }
    string value;
    if (row < GetRowsCount()) {
        CConstRef<CValidErrItem> pValidErrItem(x_GetRowData(row));

		switch(col){
            case eSeverity:
				value = CValidErrItem::ConvertSeverity(pValidErrItem->GetSeverity());
				break;

            case eMessage:
				value = pValidErrItem->CanGetMsg() ? pValidErrItem->GetMsg() : "";
                break;

            case eVerboseMessage:
				value = pValidErrItem->GetVerbose();
				break;

			case eObjectDesc:
				value = pValidErrItem->CanGetObjDesc() ? pValidErrItem->GetObjDesc() : "";
                break;

			case eErrorTitle:
				value = (pValidErrItem->CanGetErrorGroup() ? pValidErrItem->GetErrorGroup() + " " : "") + 
					    (pValidErrItem->CanGetErrorName() ? pValidErrItem->GetErrorName() : "");
                break;

			case eAccessionVersion:
				value = pValidErrItem->CanGetAccnver() ? pValidErrItem->GetAccnver() : "";
                break;

			default:
                break;
        }
    }
    return value;
}


string CTableDataValidate::GetHtmlValue(size_t row, size_t col) const
{
    list<string> values;
    string sDelimiter("<br>");
    if (row < GetRowsCount()) {
        CConstRef<CValidErrItem> pValidErrItem(x_GetRowData(row));

		switch(col){

			case eMessage:
				NStr::Wrap(pValidErrItem->CanGetMsg() ? pValidErrItem->GetMsg() : "",
							c_MessageWidth,
							values);
                break;

            case eVerboseMessage:
				NStr::Wrap(pValidErrItem->GetVerbose(), 
							c_MessageWidth, 
							values);
				break;

			case eObjectDesc:
				NStr::Wrap(pValidErrItem->CanGetObjDesc() ? pValidErrItem->GetObjDesc() : "",
							c_ObjectDescWidth,
							values);
                break;
			default:
                break;
        }
    }
    return NStr::Join(values, sDelimiter);
}


long CTableDataValidate::GetIntValue(size_t row, size_t col) const
{
    long value(0);
    if (row < GetRowsCount()) {
        CConstRef<CValidErrItem> pValidErrItem(x_GetRowData(row));

        switch(col){
            case eSeqOffset:
				value = pValidErrItem->CanGetSeqOffset() ? pValidErrItem->GetSeqOffset()+1 : 0;
				break;
        }
    }
    return value;
}


SConstScopedObject CTableDataValidate::GetObjectValue(size_t row, size_t col) const
{
    return GetRowObject(row);
}


SConstScopedObject CTableDataValidate::GetRowObject(size_t row) const
{
    SConstScopedObject value;
    if(m_DataLoadStatus == eReady && row < GetRowsCount()) {
        //!! this is incorrect since there may be several objects 
		//!! differently scoped
		value.scope = m_Scope;
        value.object = x_GetRowData(row);
    }
    return value;
}


CConstRef<CValidErrItem> CTableDataValidate::x_GetRowData(size_t row) const
{
	return row < m_Errs->GetData().size() ? m_Errs->GetData()[row] : CConstRef<CValidErrItem>();
}


END_NCBI_SCOPE

