/*  $Id: taxon_cache.cpp 27911 2013-04-24 14:04:51Z kuznets $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Roman Katargin, Anatoliy Kuznetsov
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>
#include <iostream>
#include <corelib/ncbimtx.hpp>

#include <objects/taxon1/taxon1.hpp>

#include <gui/objutils/taxon_cache.hpp>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

DEFINE_STATIC_MUTEX(s_Mutex);

CTaxonCache::CTaxonCache(const string& cache_path) 
: m_CachePath(cache_path)
{
}

CTaxonCache::~CTaxonCache()
{
}


CTaxonCache& CTaxonCache::GetInstance(const string& cache_path)
{
    static CRef<CTaxonCache> s_Cache;
    if (!s_Cache) {
        CMutexGuard LOCK(s_Mutex);
        if (!s_Cache) {
            s_Cache.Reset (new CTaxonCache(cache_path));
        }
    }
    return *s_Cache;
}

void CTaxonCache::Initialize(const vector<int>& tax_ids)
{
    CMutexGuard LOCK(s_Mutex);

    vector<int> to_initialize;
    ITERATE(vector<int>, it, tax_ids) {
        if (m_Map.find(*it) != m_Map.end())
            to_initialize.push_back(*it);
    }

    if (to_initialize.empty())
        return;

    try {
        CTaxon1 taxon;
        taxon.Init();
        string label;

        ITERATE(vector<int>, it, to_initialize) {
            label.clear();
            bool is_species = false;
            bool is_uncultured = false;
            string blast_name;
            CConstRef<COrg_ref> org_ref =
                taxon.GetOrgRef(*it, is_species, is_uncultured, blast_name);
            if (org_ref) {
                org_ref->GetLabel(&label);
                m_Map[*it] = label;
            }
        }
    }
    catch (const exception& ex) {
        LOG_POST(Error << "CTaxonCache::Initialize: " << ex.what());
    }
}

void CTaxonCache::ResetConnection()
{
    CMutexGuard LOCK(s_Mutex);
    m_TaxonClient.reset(0);
}

void CTaxonCache::x_GetLabel(int tax_id, string& label)
{
    label.clear();
    if (m_TaxonClient.get() == 0) {
        m_TaxonClient.reset(new CTaxon1);
        bool conn_res = m_TaxonClient->Init();
        if (conn_res != true) {
            m_TaxonClient.reset(0);
            return;
        }
    }

    CTaxon1& taxon = *(m_TaxonClient.get());

    bool is_species = false;
    bool is_uncultured = false;
    string blast_name;
    CConstRef<COrg_ref> org_ref =
        taxon.GetOrgRef(tax_id, is_species, is_uncultured, blast_name);
    if (!org_ref)
        return;
    org_ref->GetLabel(&label);
    m_Map.insert(TMap::value_type(tax_id, label));
}


string CTaxonCache::GetLabel(int tax_id)
{
    string label;
    {{
    CMutexGuard LOCK(s_Mutex);
    TMap::const_iterator it = m_Map.find(tax_id);
    if (it == m_Map.end()) {
        try {
            x_GetLabel(tax_id, label);
        }
        catch (const exception& ex) {
            LOG_POST(Error << "CTaxonCache::GetLabel: " << ex.what());
            m_TaxonClient.reset(0);
            // second attempt after re-initialization
            try {
                x_GetLabel(tax_id, label);
            } 
            catch (const exception& ex)
            {
                LOG_POST(Error << "Second try of CTaxonCache::GetLabel: " << ex.what());
                m_TaxonClient.reset(0);
            }
        }
        return label;
    }
    return it->second;
    }}
}

END_NCBI_SCOPE
