/*  $Id: tooltip.cpp 26476 2012-09-20 16:33:16Z rudnev $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Dmitry Rudnev
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>

#include <gui/objutils/tooltip.hpp>

BEGIN_NCBI_SCOPE



CIRef<ITooltipFormatter> ITooltipFormatter::CreateTooltipFormatter(ETooltipFormatters eFormatter)
{
    switch(eFormatter)
    {
    case eTooltipFormatter_CSSTable:
        return CCSSTableTooltipFormatter::CreateTooltipFormatter();
    case eTooltipFormatter_Html:
        return CHtmlTooltipFormatter::CreateTooltipFormatter();
    case eTooltipFormatter_Text:
        return CTextTooltipFormatter::CreateTooltipFormatter();
    }
    NCBI_ASSERT(false, "Unknown tooltip formatter type!");
	return null;
}

static void s_AddPubmedLinksRow(const string& pmids, bool& isGoToPresent, bool isBulletColPresent, ITooltipFormatter* pFormatter)
{
    const string pmURL("/pubmed/");
    string pmlinks;
    if (!pmids.empty()) {
        vector<string> tokens;
        NStr::Tokenize(pmids, ",", tokens);
        bool bFirst = true;
        NON_CONST_ITERATE(vector<string>, pmidIter, tokens) {
            string& pmid = *pmidIter;
            NStr::TruncateSpacesInPlace(pmid);
            if (bFirst) {
                bFirst = false;
                pmlinks = pFormatter->CreateLink(pmid, pmURL+pmid);
            }
            else {
                pmlinks += ", " + pFormatter->CreateLink(pmid, pmURL+pmid);
            }
        }
        pFormatter->MaybeAddGoToRow(isGoToPresent, isBulletColPresent);
        pFormatter->StartRow();
        pFormatter->AddTagCol("PubMed:");
        pFormatter->AddValueCol(pmlinks);
        pFormatter->FinishRow();
    }
}


static void s_MaybeAddGoToRow(bool& isGoToPresent, bool isBulletColPresent, ITooltipFormatter* pFormatter)
{
    if(isGoToPresent)
        return;
    pFormatter->StartRow();
    if(isBulletColPresent)
        pFormatter->AddBulletCol();
    pFormatter->AddTagCol("Go to:");
    pFormatter->FinishRow();

    isGoToPresent = true;
}



CIRef<ITooltipFormatter> CCSSTableTooltipFormatter::CreateTooltipFormatter()
{
    return CIRef<ITooltipFormatter>(new CCSSTableTooltipFormatter());
}


void CCSSTableTooltipFormatter::StartRow()
{
    m_sTooltipText += "<tr class=\"sv-bintrack\">";
}

void CCSSTableTooltipFormatter::AddBulletCol(const string& sBulletSrc)
{
    m_sTooltipText += "<td class=\"sv-bintrack-bullet\">";
    if(!sBulletSrc.empty()) {
        m_sTooltipText += "<img class=\"sv-bintrack\" src=\"" + sBulletSrc +"\">";
    }
    m_sTooltipText += "</td>";
}


void CCSSTableTooltipFormatter::AddTagCol(const string& sContents)
{
    m_sTooltipText += "<td class=\"sv-bintrack-tag\">" + sContents + "</td>";
}

void CCSSTableTooltipFormatter::AddValueCol(const string& sContents)
{
    m_sTooltipText += "<td class=\"sv-bintrack-value\">" + sContents + "</td>";
}

void CCSSTableTooltipFormatter::FinishRow()
{
    m_sTooltipText += "</tr>";
}

void CCSSTableTooltipFormatter::AddDividerRow()
{
    m_sTooltipText += "<tr class=\"sv-bintrack\"><td class=\"sv-bintrack\" colspan=\"2\"><hr></td></tr>";
}

string CCSSTableTooltipFormatter::Render()
{
    string sWholeTooltip("<table class=\"sv-bintrack\">" + m_sTooltipText + "</table>");
    m_sTooltipText.clear();
    return sWholeTooltip;
}

void CCSSTableTooltipFormatter::AddPubmedLinksRow(const string& pmids, bool& isGoToPresent, bool isBulletColPresent)
{
    s_AddPubmedLinksRow(pmids, isGoToPresent, isBulletColPresent, this);
}

void CCSSTableTooltipFormatter::MaybeAddGoToRow(bool& isGoToPresent, bool isBulletColPresent)
{
    s_MaybeAddGoToRow(isGoToPresent, isBulletColPresent, this);
}


string CCSSTableTooltipFormatter::CreateLink(const string& sText, const string& sUrl) const
{
    return "<a class=\"sv-bintrack\" href=\"" + sUrl + "\" target=\"_blank\">"+ sText + "</a>";
}

CIRef<ITooltipFormatter> CHtmlTooltipFormatter::CreateTooltipFormatter()
{
    return CIRef<ITooltipFormatter>(new CHtmlTooltipFormatter());
}


void CHtmlTooltipFormatter::StartRow()
{
    m_sTooltipText += "<tr>";
}

void CHtmlTooltipFormatter::AddBulletCol(const string& sBulletSrc)
{
    m_sTooltipText += "<td valign=\"top\" align=\"right\" nowrap=\"nowrap\">";
    if(!sBulletSrc.empty()) {
        m_sTooltipText += "<img src=\"" + sBulletSrc +"\">&nbsp;";
    }
    m_sTooltipText += "</td>";
}


void CHtmlTooltipFormatter::AddTagCol(const string& sContents)
{
    m_sTooltipText += "<td valign=\"top\" align=\"right\" nowrap=\"nowrap\"><span style=\"font-weight:bold\">" + sContents + "</span></td>";
}

void CHtmlTooltipFormatter::AddValueCol(const string& sContents)
{
    m_sTooltipText += "<td valign=\"top\">&nbsp;" + sContents + "</td>";
}

void CHtmlTooltipFormatter::FinishRow()
{
    m_sTooltipText += "</tr>";
}

void CHtmlTooltipFormatter::AddDividerRow()
{
    m_sTooltipText += "<tr><td colspan=\"2\"><hr></td></tr>";
}

string CHtmlTooltipFormatter::Render()
{
    string sWholeTooltip("<table margin=\"0\" padding=\"0\" border=\"0\" cellpadding=\"0\" cellspacing=\"0\">" + m_sTooltipText + "</table>");
    m_sTooltipText.clear();
    return sWholeTooltip;
}

void CHtmlTooltipFormatter::AddPubmedLinksRow(const string& pmids, bool& isGoToPresent, bool isBulletColPresent)
{
    s_AddPubmedLinksRow(pmids, isGoToPresent, isBulletColPresent, this);
}

void CHtmlTooltipFormatter::MaybeAddGoToRow(bool& isGoToPresent, bool isBulletColPresent)
{
    s_MaybeAddGoToRow(isGoToPresent, isBulletColPresent, this);
}

string CHtmlTooltipFormatter::CreateLink(const string& sText, const string& sUrl) const
{
    return "<a href=\"" + sUrl + "\" target=\"_blank\">"+ sText + "</a>";
}


CIRef<ITooltipFormatter> CTextTooltipFormatter::CreateTooltipFormatter()
{
    return CIRef<ITooltipFormatter>(new CTextTooltipFormatter());
}


void CTextTooltipFormatter::StartRow()
{
}

void CTextTooltipFormatter::AddBulletCol(const string& sBulletSrc)
{
    m_sTooltipText += sBulletSrc.empty() ? "   " : " * ";
}


void CTextTooltipFormatter::AddTagCol(const string& sContents)
{
    m_sTooltipText += sContents + "\t\t";
}

void CTextTooltipFormatter::AddValueCol(const string& sContents)
{
    m_sTooltipText += sContents;
}

void CTextTooltipFormatter::FinishRow()
{
    m_sTooltipText += "\n";
}

void CTextTooltipFormatter::AddDividerRow()
{
    m_sTooltipText += "-----------------------\n";
}

string CTextTooltipFormatter::Render()
{
    string sWholeTooltip(m_sTooltipText);
    m_sTooltipText.clear();
    return sWholeTooltip;
}

void CTextTooltipFormatter::AddPubmedLinksRow(const string& pmids, bool& isGoToPresent, bool isBulletColPresent)
{
    s_AddPubmedLinksRow(pmids, isGoToPresent, isBulletColPresent, this);
}

string CTextTooltipFormatter::CreateLink(const string& sText, const string& sUrl) const
{
    return "<a href=\"" + sUrl + "\" target=\"_blank\">"+ sText + "</a>";
}

void CTextTooltipFormatter::MaybeAddGoToRow(bool& isGoToPresent, bool isBulletColPresent)
{
    s_MaybeAddGoToRow(isGoToPresent, isBulletColPresent, this);
}



END_NCBI_SCOPE
