/*  $Id: util_cmds.cpp 28183 2013-05-21 19:33:28Z bollin $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data,  the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties,  express or implied,  including
 *  warranties of performance,  merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Colleen Bollin
 */


#include <ncbi_pch.hpp>

#include <objects/seq/Pubdesc.hpp>
#include <objmgr/scope.hpp>
#include <objmgr/bioseq_set_handle.hpp>
#include <objmgr/seqdesc_ci.hpp>
#include <objmgr/feat_ci.hpp>
#include <objmgr/bioseq_ci.hpp>
#include <gui/objutils/util_cmds.hpp>
#include <gui/objutils/cmd_factory.hpp>
#include <gui/objutils/cmd_del_bioseq.hpp>
#include <gui/objutils/cmd_del_bioseq_set.hpp>
#include <gui/objutils/cmd_del_seq_feat.hpp>
#include <gui/objutils/descriptor_change.hpp>
#include <gui/objutils/cmd_change_seq_feat.hpp>
#include <objects/taxon3/taxon3.hpp>

BEGIN_NCBI_SCOPE

USING_SCOPE(objects);


CRef<CCmdComposite> GetDeleteFeatureCommand(const objects::CSeq_feat_Handle& fh)
{
    CRef<CCmdComposite> cmd(new CCmdComposite("Delete Feature"));

    const CSeqFeatData& data = fh.GetData();
    if (data.IsPub()) {
        CBioseq_Handle this_seq = fh.GetScope().GetBioseqHandle(fh.GetLocationId());
        CSeq_entry_Handle top_parent = this_seq.GetSeq_entry_Handle();
        CBioseq_set_Handle parent_set = top_parent.GetParentBioseq_set();
        while (parent_set) {
            top_parent = parent_set.GetParentEntry();
            parent_set = top_parent.GetParentBioseq_set();
        }
        CIRef<IEditCommand> cit_cmd = CCmdFactory::RemoveCitations(top_parent, data.GetPub().GetPub());
        if (cit_cmd) {
            cmd->AddCommand(*cit_cmd);
        }
    }

    if (fh.IsSetProduct()) {
        CBioseq_Handle product = fh.GetScope().GetBioseqHandle(*(fh.GetProductId().GetSeqId()));
        if (product) {   
            CRef<CCmdDelBioseqInst> del_product(new CCmdDelBioseqInst(product));
            cmd->AddCommand(*del_product);
        }        
    }
    CRef<CCmdDelSeq_feat> del_feat(new CCmdDelSeq_feat(fh));
    cmd->AddCommand(*del_feat);

    return cmd;
}


CRef<CCmdComposite> GetDeleteSequenceCommand(CBioseq_Handle bsh)
{
    CRef<CCmdComposite> cmd(new CCmdComposite("Delete Sequence"));

    CBioseq_set_Handle nuc_parent = bsh.GetParentBioseq_set();
    if (nuc_parent && nuc_parent.IsSetClass() && nuc_parent.GetClass() == CBioseq_set::eClass_nuc_prot                
        && bsh.IsNa()) {
        if (nuc_parent.GetTopLevelEntry() == nuc_parent.GetParentEntry()) {
            NCBI_THROW(CException, eUnknown, "Cannot delete last sequence");
        } else {
            CRef<CCmdDelBioseqSet> delSet(new CCmdDelBioseqSet(nuc_parent));
            cmd->AddCommand(*delSet);
        }
    } else if (bsh.GetTopLevelEntry() == bsh.GetSeq_entry_Handle()) {
        NCBI_THROW(CException, eUnknown, "Cannot delete last sequence");
    } else {
        CRef<CCmdDelBioseqInst> scmd (new CCmdDelBioseqInst(bsh));
        cmd->AddCommand(*scmd);
    }
    return cmd;
}

CRef<CCmdComposite> GetDeleteAllFeaturesCommand(CSeq_entry_Handle seh, size_t& count)
{
    CRef<CCmdComposite> delete_features_cmd(new CCmdComposite("Remove All Features"));
    count = 0;
    // only delete features on nucleotide sequences, protein features are automatically removed with the coding region
    CBioseq_CI bi(seh, objects::CSeq_inst::eMol_na);
    while (bi) {
        CFeat_CI it(*bi);
        while (it) {
             CRef<CCmdComposite> cmd = GetDeleteFeatureCommand(it->GetSeq_feat_Handle());
             delete_features_cmd->AddCommand(*cmd);
             count++;
             ++it;
        }
        ++bi;
    }
    return delete_features_cmd;
}


void GetSourceDescriptors(const CSeq_entry& se, vector<CObject* >& src_descs)
{
    if (se.IsSetDescr()) {
        ITERATE(CBioseq::TDescr::Tdata, it, se.GetDescr().Get()) {
            if ((*it)->IsSource() && (*it)->GetSource().IsSetOrg()) {
                CObject* actual = (CObject*) &(**it);
                src_descs.push_back(actual);
            }
        }
    }

    if (se.IsSet() && se.GetSet().IsSetSeq_set()) {
        ITERATE(CBioseq_set::TSeq_set, it, se.GetSet().GetSeq_set()) {
            GetSourceDescriptors(**it, src_descs);
        }
    }
}
 

CRef<CCmdComposite> TaxonomyLookupCommand(CSeq_entry_Handle seh)
{
    CRef<CCmdComposite> cmd(new CCmdComposite("Tax lookup"));
    vector<CRef<COrg_ref> > rq_list;
    vector<CObject* > src_descs;
    vector<CConstRef<CSeq_feat> > src_feats;

    GetSourceDescriptors(*(seh.GetCompleteSeq_entry()), src_descs);
    vector<CObject* >::iterator desc_it = src_descs.begin();
    while (desc_it != src_descs.end()) {
        // add org ref for descriptor to request list
        CObject* actual = *desc_it;
        CSeqdesc* desc = dynamic_cast<CSeqdesc *>(actual);
        CRef<COrg_ref> org(new COrg_ref());
        org->Assign(desc->GetSource().GetOrg());
        rq_list.push_back(org);

        ++desc_it;        
    }

    CFeat_CI feat (seh, SAnnotSelector(CSeqFeatData::e_Biosrc));
    while (feat) {
        if (feat->GetData().GetBiosrc().IsSetOrg()) {
            // add org ref for feature to request list
            CRef<COrg_ref> org(new COrg_ref());
            org->Assign(feat->GetData().GetBiosrc().GetOrg());
            rq_list.push_back(org);
            // add feature to list
            src_feats.push_back(feat->GetOriginalSeq_feat());
        }
        ++feat;
    }

    if (rq_list.size() > 0) {
        CTaxon3 taxon3;
        taxon3.Init();
        CRef<CTaxon3_reply> reply = taxon3.SendOrgRefList(rq_list);
        if (reply) {
            CTaxon3_reply::TReply::const_iterator reply_it = reply->GetReply().begin();

            // process descriptor responses
            desc_it = src_descs.begin();
            
            while (reply_it != reply->GetReply().end()
                   && desc_it != src_descs.end()) {
                if ((*reply_it)->IsData()) {
                    CChangeSeqDescCommand* subcmd = new CChangeSeqDescCommand();
                    CObject* actual = *desc_it;
                    CRef<CSeqdesc> new_desc(new CSeqdesc());
                    new_desc->Assign(* (dynamic_cast<CSeqdesc *>(actual)));
                    new_desc->SetSource().SetOrg().Assign((*reply_it)->GetData().GetOrg());
                    subcmd->Add(actual, CConstRef<CObject>(new_desc));
                    cmd->AddCommand(*subcmd);
                }
                ++reply_it;
                ++desc_it;
            }

            // process feature responses
            vector<CConstRef<CSeq_feat> >::iterator feat_it = src_feats.begin();
            while (reply_it != reply->GetReply().end()
                   && feat_it != src_feats.end()) {
                if ((*reply_it)->IsData()) {
                    CRef<CSeq_feat> new_feat(new CSeq_feat());
                    new_feat->Assign(**feat_it);
                    new_feat->SetData().SetBiosrc().SetOrg().Assign((*reply_it)->GetData().GetOrg());
                    CSeq_feat_Handle fh = seh.GetScope().GetSeq_featHandle(**feat_it);
                    CIRef<IEditCommand> chgFeat(new CCmdChangeSeq_feat(fh, *new_feat));
                    cmd->AddCommand(*chgFeat);
                }
                ++reply_it;
                ++feat_it;
            }
        }
    }

    return cmd;
}


END_NCBI_SCOPE
