/*  $Id: utils.cpp 28078 2013-05-06 22:10:55Z rafanovi $
* ===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data, the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties, express or implied, including
*  warranties of performance, merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================
*
* Authors:  Mike DiCuccio, Liangshou Wu
*
* File Description:
*    General utility classes for GUI projects.
*/

#include <ncbi_pch.hpp>
#include <gui/objutils/utils.hpp>
#include <gui/objutils/registry.hpp>
#include <gui/objutils/tool_tip_info.hpp>

#include <objects/seq/Seq_descr.hpp>
#include <objects/seq/seqport_util.hpp>
#include <objects/seqfeat/Feat_id.hpp>
#include <objects/seqfeat/SeqFeatXref.hpp>
#include <objects/general/Object_id.hpp>
#include <objects/seq/Annot_descr.hpp>
#include <objects/seq/Annotdesc.hpp>
#include <objects/seqfeat/RNA_ref.hpp>
#include <objects/seqalign/Seq_align_set.hpp>
#include <objects/seqalign/Std_seg.hpp>

#include <objmgr/feat_ci.hpp>
#include <objmgr/util/sequence.hpp>
#include <objmgr/util/feature.hpp>
#include <objmgr/impl/handle_range_map.hpp>
#include <objmgr/impl/synonyms.hpp>
#include <objmgr/seq_vector.hpp>

#include <serial/iterator.hpp>
#include <connect/ncbi_types.h>
#include <connect/ncbi_conn_stream.hpp>
#include <util/line_reader.hpp>
#include <algorithm>


BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

//
// functor for sorting features based on their length
//
struct SFeatLengthPredicate
{
    bool operator()(const CMappedFeat& feat0,
        const CMappedFeat& feat1) const
    {
        TSeqRange r0 = feat0.GetLocation().GetTotalRange();
        TSeqRange r1 = feat1.GetLocation().GetTotalRange();
        return (r0.GetLength() < r1.GetLength());
    }
};


//
// functor for sorting features based on the NCBI feature sort order
//
struct SFeatSortPredicate
{
    bool operator()(const CMappedFeat& feat0,
        const CMappedFeat& feat1) const
    {
        const CSeq_feat& f0 = feat0.GetOriginalFeature();
        const CSeq_feat& f1 = feat1.GetOriginalFeature();
        return (f0.Compare(f1, feat0.GetLocation(), feat1.GetLocation()) < 0);
    }
};



typedef SStaticPair<const char*, CSeqUtils::TAnnotNameType> TNameTypeStr;
static const TNameTypeStr s_NameTypeStrs[] = {
    { "",  CSeqUtils::eAnnot_All },
    { "All",  CSeqUtils::eAnnot_All },
    { "Named",  CSeqUtils::eAnnot_Named },
    { "Unnamed",   CSeqUtils::eAnnot_Unnamed },
};

typedef CStaticArrayMap<string, CSeqUtils::TAnnotNameType> TNameTypeMap;
DEFINE_STATIC_ARRAY_MAP(TNameTypeMap, sm_NameTypeMap, s_NameTypeStrs);

CSeqUtils::TAnnotNameType CSeqUtils::NameTypeStrToValue(const string& type)
{
    TNameTypeMap::const_iterator iter = sm_NameTypeMap.find(type);
    if (iter != sm_NameTypeMap.end()) {
        return iter->second;
    } else {
        return CSeqUtils::eAnnot_Other;
    }
}


const string& 
CSeqUtils::NameTypeValueToStr(CSeqUtils::TAnnotNameType type)
{
    TNameTypeMap::const_iterator iter;
    for (iter = sm_NameTypeMap.begin();  iter != sm_NameTypeMap.end();  ++iter) {
        if (iter->second == type) {
            return iter->first;
        }
    }
    return kEmptyStr;
}


//
// retrieve an annot selector
//
SAnnotSelector CSeqUtils::GetAnnotSelector(TAnnotFlags flags)
{
    SAnnotSelector sel;
    sel
        // consider overlaps by total range...
        .SetOverlapTotalRange()
        // resolve all segments...
        .SetResolveAll()
        ;

    CGuiRegistry& reg = CGuiRegistry::GetInstance();

    CGuiRegistry::TReadView view =
        reg.GetReadView("GBENCH.Utils.NamedAnnots");
    CGuiRegistry::TKeys naa_keys;
    view.GetKeys(naa_keys);
    ITERATE (CGuiRegistry::TKeys, iter, naa_keys) {
        sel.IncludeNamedAnnotAccession(view.GetString(iter->key));
    }

    view = reg.GetReadView("GBENCH.Utils.AnnotSelector");
    if ( !(flags & fAnnot_UnsetNamed) ) {
        if (view.GetBool("ExcludeExternal")) {
            sel.SetExcludeExternal(true);
        } else {
            sel.SetExcludeExternal(false);

            ///
            /// known external annotations
            ///

            static const char* named_annots[] = {
                "SNP",  /// SNPs = variation features
                "CDD",  /// CDD  = conserved domains
                "STS",  /// STS  = sequence tagged sites
                NULL
            };

            for (const char** p = named_annots;  p  &&  *p;  ++p) {
                bool incl = view.GetBool(*p, true);
                if ( !incl ) {
                    sel.ExcludeNamedAnnots(*p);
                }
            }
        }
    }

    if ( !(flags & fAnnot_UnsetDepth) ) {
        if (view.GetBool("AdaptiveDepth", true)) {
            sel.SetAdaptiveDepth(true);
            sel.SetResolveAll();
        }
    }
    return sel;
}


//
// retrieve an annot selector for our selected annotations
//
SAnnotSelector CSeqUtils::GetAnnotSelector(SAnnotSelector::TAnnotType c,
                                           TAnnotFlags flags)
{
    SAnnotSelector sel = GetAnnotSelector(flags);
    sel
        // limit by our annotation type
        .SetAnnotType(c);
    return sel;
}


//
// retrieve an annot selector for our selected annotations
//
SAnnotSelector CSeqUtils::GetAnnotSelector(SAnnotSelector::TFeatType  feat,
                                           TAnnotFlags flags)
{
    SAnnotSelector sel = GetAnnotSelector(CSeq_annot::TData::e_Ftable, flags);
    sel
        // retrieve feature type and subtype of interest
        .SetFeatType(feat);

    return sel;
}


SAnnotSelector CSeqUtils::GetAnnotSelector(SAnnotSelector::TFeatSubtype sub,
                                           TAnnotFlags flags)
{
    SAnnotSelector sel = GetAnnotSelector(CSeq_annot::TData::e_Ftable, flags);
    sel
        // retrieve feature type and subtype of interest
        .SetFeatSubtype(sub);

    return sel;
}


SAnnotSelector CSeqUtils::GetAnnotSelector(const vector<string>& annots)
{
    SAnnotSelector sel = GetAnnotSelector();
    sel.SetCollectNames();

    //CGuiRegistry& reg = CGuiRegistry::GetInstance();
    //CGuiRegistry::TReadView view =
    //    reg.GetReadView("GBENCH.Utils.AnnotSelector");
    //bool include_naas = view.GetBool("IncludeNAAs", false);
    bool include_naas = false;

    if (include_naas  &&  annots.empty()) {
        sel.IncludeNamedAnnotAccession("NA*");
    }

    ITERATE (vector<string>, iter, annots) {
        const string& annot = *iter;
        switch (CSeqUtils::NameTypeStrToValue(annot)) {
        case CSeqUtils::eAnnot_Unnamed:
            sel.AddUnnamedAnnots();
            break;
        case CSeqUtils::eAnnot_Named:
            sel.ExcludeUnnamedAnnots();
            break;
        case CSeqUtils::eAnnot_All:
            if (include_naas) {
                sel.IncludeNamedAnnotAccession("NA*");
            }
            return sel;
        case CSeqUtils::eAnnot_Other:
        default:
            if (NStr::StartsWith(annot, "NA*")) {
                sel.IncludeNamedAnnotAccession("NA*");
            } else {
                sel.AddNamedAnnots(annot);
                if (IsNAA(annot)) {
                    sel.IncludeNamedAnnotAccession(annot);
                }
            }
            break;
        }
    }
    return sel;
}


SAnnotSelector CSeqUtils::GetAnnotSelector(const vector<string>& annots,
                                           bool adaptive, int depth)
{
    SAnnotSelector sel(GetAnnotSelector(annots));
    SetResolveDepth(sel, adaptive, depth);
    return sel;
}


void CSeqUtils::SetAnnot(objects::SAnnotSelector& sel, const string& annot)
{
    switch (CSeqUtils::NameTypeStrToValue(annot)) {
        case CSeqUtils::eAnnot_Unnamed:
            sel.AddUnnamedAnnots();
            break;
        case CSeqUtils::eAnnot_Named:
            sel.ExcludeUnnamedAnnots();
            break;
        case CSeqUtils::eAnnot_All:
            {{
                //CGuiRegistry& reg = CGuiRegistry::GetInstance();
                //CGuiRegistry::TReadView view =
                //    reg.GetReadView("GBENCH.Utils.AnnotSelector");
                //bool include_naas = view.GetBool("IncludeNAAs", false);
                bool include_naas = false;
                if (include_naas) {
                    sel.IncludeNamedAnnotAccession("NA*");
                }
            }}
            break;
        case CSeqUtils::eAnnot_Other:
        default:
            sel.AddNamedAnnots(annot);
            if (IsNAA(annot)) {
                sel.IncludeNamedAnnotAccession(annot);
            }
            break;
    }
}


void CSeqUtils::SetResolveDepth(objects::SAnnotSelector& sel,
                                bool adaptive, int depth)
{
    if (adaptive) {
        sel.SetAdaptiveDepth(true);
        sel.SetExactDepth(false);
        // TODO: watch out
        // Maybe there is bug inside selector, we have call SetResolveAll() even
        // for cases where we only want to resolve up to a given depth.
        sel.SetResolveAll();
        if (depth >=0) sel.SetResolveDepth(depth);
    } else if (depth >= 0){
        sel.SetResolveDepth(depth);
        sel.SetExactDepth(true);
        sel.SetAdaptiveDepth(false);
    }
}


//
// LinkFeatures()
// This builds explicit links between features, creating a hierarchical tree of
// features.
//

template <class T, class U>
struct SPairBy1stLess
    : public binary_function< pair<T,U>, pair<T,U>, bool>
{
    bool operator()(const pair<T,U>& p1, const pair<T,U>& p2) const
    {
        return p1.first < p2.first;
    }
};


bool CSeqUtils::LinkFeatures(CLinkedFeature::TLinkedFeats& feats,
                             CSeqUtils::TFeatLinkingMode mode,
                             ISeqTaskProgressCallback* p_cb)
{
    if (p_cb) {
        p_cb->SetTaskName("Linking features...");
        p_cb->SetTaskTotal((int)feats.size());
        p_cb->SetTaskCompleted(0);
    }

    CLinkedFeature::TLinkedFeats out_feats;

    // using CFeatTree
    feature::CFeatTree tree;
    tree.SetFeatIdMode(feature::CFeatTree::EFeatIdMode(mode));
    map<CMappedFeat, CRef<CLinkedFeature> > fmap;
    NON_CONST_ITERATE (CLinkedFeature::TLinkedFeats, iter, feats) {
        if (p_cb  &&  p_cb->StopRequested()) {
            return  false;
        }
        CRef<CLinkedFeature> curr_feat = *iter;
        tree.AddFeature(curr_feat->GetMappedFeature());
        fmap[curr_feat->GetMappedFeature()] = curr_feat;
    }
    NON_CONST_ITERATE (CLinkedFeature::TLinkedFeats, iter, feats) {
        if (p_cb  &&  p_cb->StopRequested()) {
            return  false;
        }
        CRef<CLinkedFeature> curr_feat = *iter;
        CMappedFeat parent_feat = tree.GetParent(curr_feat->GetMappedFeature());
        if ( parent_feat ) {
            fmap[parent_feat]->AddChild(curr_feat);
        }
        else {
            out_feats.push_back(curr_feat);
        }
        if (p_cb) p_cb->AddTaskCompleted(1);
    }

    out_feats.swap(feats);

    return true;
}


// remap a child location to a parent
CRef<CSeq_loc> CSeqUtils::RemapChildToParent(const CSeq_loc& parent,
                                             const CSeq_loc& child,
                                             CScope* scope)
{
    CSeq_loc dummy_parent;
    dummy_parent.SetWhole(const_cast<CSeq_id&>(sequence::GetId(parent, 0)));
    SRelLoc converter(dummy_parent, child, scope);
    converter.m_ParentLoc = &parent;
    return converter.Resolve(scope);
}

bool CSeqUtils::Match(const CSeq_id& id1, const CSeq_id& id2, CScope* scope)
{
    return Match(CSeq_id_Handle::GetHandle(id1),
        CSeq_id_Handle::GetHandle(id2),
        scope);
}


bool CSeqUtils::Match(const CSeq_id_Handle& id1,
                      const CSeq_id_Handle& id2, CScope* scope)
{
    if (id1.MatchesTo(id2))  {
        return true;
    }

    if (id1.IsGi() && id2.IsGi() )
        return false;

    if (scope) {
        CConstRef<CSynonymsSet> syns;

        syns = scope->GetSynonyms(id1);
        if (syns) {
            ITERATE (CSynonymsSet, iter, *syns) {
                if (id2.MatchesTo((*iter)->first)) {
                    return true;
                }
            }
        }

        syns = scope->GetSynonyms(id2);
        if (syns) {
            ITERATE (CSynonymsSet, iter, *syns) {
                if (id1.MatchesTo((*iter)->first)) {
                    return true;
                }
            }
        }
    }
    return false;
}

CRef<CSeq_loc>  CSeqUtils::CreateSeq_loc(const CSeq_id& id,
                                         const CRangeCollection<TSeqPos>& ranges)
{
    CRef<CSeq_loc>  seq_loc(new CSeq_loc());
    CSeq_loc::TPacked_int& p_int = seq_loc->SetPacked_int();

    ITERATE(CRangeCollection<TSeqPos>, it_r, ranges) { // for each range in mark
        if ( !it_r->Empty() ) {
            p_int.AddInterval(id, it_r->GetFrom(), it_r->GetTo());
        }
    }
    switch (p_int.Get().size()) {
    case 0:
        {{
        return CRef<CSeq_loc>();
        }}
    case 1:
        {{
            CRef<CSeq_interval> ival(p_int.Set().front());
            seq_loc->SetInt(*ival);
            /// p_int no longer valid!
        }}
        break;
    default:
        break;
    }
    return seq_loc;
}

bool   CSeqUtils::GetRangeCollection(const CSeq_id& id, const CHandleRangeMap& map,
                                     CRangeCollection<TSeqPos>& ranges)
{
    CSeq_id_Handle s_id = CSeq_id_Handle::GetHandle(id);

    // extract from the given map all segments corresponding to the given id
    const CHandleRangeMap::TLocMap& loc_map = map.GetMap();
    CHandleRangeMap::TLocMap::const_iterator it = loc_map.find(s_id);
    if(it != loc_map.end()) {
        ITERATE(CHandleRange, it_r, it->second) {
            ranges.CombineWith(it_r->first);
        }
        return true;
    } else return false;
}

CBioseq* CSeqUtils::SeqLocToBioseq(CScope& scope,
                                   const CSeq_loc& loc)
{
    // Build a Seq-entry for the query Seq-loc
    // A seq-loc may have multiple seq-ids, use the first successful seq-id.
    CBioseq_Handle handle;
    for ( CSeq_loc_CI citer (loc); citer; ++citer) {
        handle = scope.GetBioseqHandle(citer.GetSeq_id());
        if ( handle ) {
            break;
        }
    }

    if ( !handle ) {
        return CRef<CBioseq>();
    }

    /// easy out: if the bioseq is of type whole, just duplicate it
    if (loc.IsWhole()) {
        CRef<CBioseq> bioseq(new CBioseq());
        bioseq->Assign(*handle.GetCompleteBioseq());
        return bioseq.Release();
    }

    CSeqVector vec(loc, scope, CBioseq_Handle::eCoding_Iupac);
    string seq_string;
    vec.GetSeqData(0, vec.size(), seq_string);

    CRef<CBioseq> bioseq(new CBioseq());

    // curate our inst
    bioseq->SetInst().SetRepr(CSeq_inst::eRepr_raw);
    bioseq->SetInst().SetLength((int)seq_string.size());
    if (vec.IsProtein()) {
        bioseq->SetInst().SetMol(CSeq_inst::eMol_aa);
        bioseq->SetInst().SetSeq_data().SetIupacaa(*new CIUPACaa(seq_string));
    } else {
        bioseq->SetInst().SetMol(CSeq_inst::eMol_na);
        bioseq->SetInst().SetSeq_data().SetIupacna(*new CIUPACna(seq_string));
        CSeqportUtil::Pack(&bioseq->SetInst().SetSeq_data());
    }


    // add an ID for our sequence
    CRef<CSeq_id> id(new CSeq_id());
    id->Assign(*handle.GetSeqId());
    bioseq->SetId().push_back(id);

    // a title
    CRef<CSeqdesc> title(new CSeqdesc);
    string title_str;
    id->GetLabel(&title_str);
    title_str += ": ";
    loc.GetLabel(&title_str);
    title->SetTitle(title_str);
    bioseq->SetDescr().Set().push_back(title);

    return bioseq.Release();
}


CRef<CSeq_loc> CSeqUtils::MixLocToLoc(const CSeq_loc& mix_loc,
                                      const CBioseq_Handle& handle)
{
    CRef<CSeq_loc>  seq_loc(new CSeq_loc());
    CSeq_loc::TPacked_int& p_int = seq_loc->SetPacked_int();

    for (CSeq_loc_CI iter(mix_loc);  iter;  ++iter) {
        if (handle.IsSynonym(iter.GetSeq_id())) {
            CSeq_loc_CI::TRange range = iter.GetRange();
            if ( !range.Empty() ) {
                p_int.AddInterval(*handle.GetSeqId(), range.GetFrom(),
                                  range.GetTo(), iter.GetStrand());
            }
        }
    }
    switch (p_int.Get().size()) {
    case 0:
        {{
            return CRef<CSeq_loc>();
        }}
    case 1:
        {{
            CRef<CSeq_interval> ival(p_int.Set().front());
            seq_loc->SetInt(*ival);
            /// p_int no longer valid!
        }}
        break;
    default:
        break;
    }

    return seq_loc;
}


void CSeqUtils::TokenizeWithEscape(const string &str,
                                   const string &delim,
                                   vector<string>& tokens,
                                   bool remove_escape)
{
    // Special cases
    if (str.empty()) {
        return;
    } else if (delim.empty()) {
        tokens.push_back(str);
        return;
    }
 
    // Tokenization
    //
    string::size_type pos = 0;
    bool escape = false;
    string curr_str = "";
    while (pos < str.size()) {
        const char& curr_char = str[pos];
        if (curr_char == '\\') {
            if ( !escape ) {
                escape = true;
                ++pos;
                continue;
            }
        }

        bool match = delim.find(curr_char) != string::npos;
        if (match) {
            if (escape) {
                if ( !remove_escape ) {
                    curr_str.append(1, '\\');
                }
                curr_str.append(1, curr_char);
                escape = false;
            } else if ( !curr_str.empty() ) {
                tokens.push_back(curr_str);
                curr_str.clear();
            }
        } else {
            if (escape) {
                curr_str.append(1, '\\');
                curr_str.append(1, curr_char);
                escape = false;
            } else {
                curr_str.append(1, curr_char);
            }
        }
        ++pos;
    }

    if ( !curr_str.empty() ) {
        tokens.push_back(curr_str);
    }
}


string CSeqUtils::RemoveEscape(const string &str)
{
    // Special cases
    if (str.empty()) {
        return str;
    }
 
    string::size_type len = str.length();
    string out;
    out.resize(len);

    string::size_type pdst = 0;
    bool escape = false;
    for (string::size_type psrc = 0; psrc < len; ++psrc) {
        const char& curr_char = str[psrc];
        if (curr_char == '\\'  &&  !escape ) {
            escape = true;
        } else {
            out[pdst++] = curr_char;
            if (escape) {
                escape = false;
            }
        }
    }

    if (pdst < len) {
        out[pdst] = '\0';
        out.resize(pdst);
    }

    return out;
}


string CSeqUtils::GetAnnotName(const CSeq_annot_Handle& annot_handle)
{
    string name("Unnamed");
    CConstRef<CSeq_annot> annot = annot_handle.GetCompleteSeq_annot();
    if (annot  &&  annot->IsSetDesc()) {
        ITERATE (objects::CAnnot_descr::Tdata, descrIter, annot->GetDesc().Get()) {
            if ((*descrIter)->IsTitle()) {
                name = (*descrIter)->GetTitle();
                break;
            } else if ((*descrIter)->IsName()) {
                name = (*descrIter)->GetName();
            }
        }
    } else if (annot_handle.IsNamed()) {
        name = annot_handle.GetName();
    }
    return name;
}


string CSeqUtils::GetAnnotComment(const CSeq_annot_Handle& annot_handle)
{
    CConstRef<CSeq_annot> annot = annot_handle.GetCompleteSeq_annot();
    if (annot  &&  annot->IsSetDesc()) {
        ITERATE (objects::CAnnot_descr::Tdata, descrIter, annot->GetDesc().Get()) {
            if ((*descrIter)->IsComment()) {
                return (*descrIter)->GetComment();
            }        }
    }
    return kEmptyStr;
}


bool CSeqUtils::IsNAA(const string& annot)
{
    // Don't want to introduce regexp dependency to avoid 
    // affecting large scope of code since objutil is used
    // in many places.
    // CRegexp is_naa("^NA[0-9]{9}(\.[0-9]*)?$");
    // return is_naa.IsMatch(annot);

    size_t acc_len = 11;
    bool is_naa = false;
    size_t len = annot.size();
    if (NStr::StartsWith(annot, "NA")  &&  len >= acc_len) {
        size_t i = 2;
        while(i < acc_len  &&  annot[i] >= '0' && annot[i] <= '9') {
            ++i;
        }

        if (i == acc_len) {
            if (len == acc_len) {
                is_naa = true;
            } else if (annot[i++] == '.') {
                while(i < len  &&  annot[i] >= '0' && annot[i] <= '9') {
                    ++i;
                }
                if (i == len) {
                    is_naa = true; 
                }
            }
        }
    }

    return is_naa;
}


bool CSeqUtils::IsPseudoFeature(const CSeq_feat& feat)
{
    if (feat.IsSetPseudo()) {
        return feat.GetPseudo();
    } else {
        const CSeq_feat::TData& data = feat.GetData();
        if (data.IsGene()  &&  data.GetGene().IsSetPseudo()) {
            return data.GetGene().GetPseudo();
        } else if (data.IsRna()  &&  data.GetRna().IsSetPseudo()) {
            return data.GetRna().GetPseudo();
        }
    }
    return false;
}


bool CSeqUtils::IsPartialFeature(const CSeq_feat& feat)
{
    if (feat.IsSetPartial()  &&  feat.GetPartial()  &&
        !IsPartialStart(feat.GetLocation())  &&
        !IsPartialStop(feat.GetLocation())) {
        return true;
    }
    return false;
}


bool CSeqUtils::IsPartialStart(const CSeq_loc& loc)
{
    return loc.IsPartialStart(objects::eExtreme_Biological);
}


bool CSeqUtils::IsPartialStop(const CSeq_loc& loc)
{
    return loc.IsPartialStop(objects::eExtreme_Biological);
}


bool CSeqUtils::IsException(const CSeq_feat& feat)
{
    if (feat.IsSetExcept()) {
        return feat.GetExcept();
    }
    return false;
}


string CSeqUtils::GetNcbiBaseUrl()
{
    static string base_url = "http://www.ncbi.nlm.nih.gov";
    return base_url;
}


string CSeqUtils::CreateTableStart()
{
    return "<table border=\"0\" cellpadding=\"0\" cellspacing=\"0\">";
}


string CSeqUtils::CreateTableEnd()
{
    return "</table>";
}


string CSeqUtils::CreateTableRow(const string& tag, const string& value)
{
    return "<tr><td align=\"right\" valign=\"top\" nowrap><b>" + tag + ":&nbsp;" +
        "</b></td><td valign=\"top\" width=\"200\">" + value + "</td></tr>";
}


string CSeqUtils::CreateSectionRow(const string& tag)
{
    return "<tr><td align=\"right\" nowrap>[<i>" + tag + "</i>]&nbsp;&nbsp;</td><td></td></tr>";
}


string CSeqUtils::CreateLinkRow(const string& tag,
                                const string& label,
                                const string& url)
{
    return "<tr><td align=\"right\" valign=\"top\" nowrap><b>" + tag  + ":&nbsp;" +
        "</b></td><td width=\"200\"><a href=\"" + url + "\">" + label + "</a></td></tr>";
}


static const char* kGiLinkUrl =
"http://www.ncbi.nlm.nih.gov/sviewer/links.fcgi?link_name=gi_placement&report=asn&gi=";

CSeqUtils::TLocVec CSeqUtils::GetGiPlacements(int gi, int time_out_sec)
{
    TLocVec mapped_locs;
    STimeout timeout;
    timeout.sec = time_out_sec;
    timeout.usec = 0;
    
    CConn_HttpStream stream(kGiLinkUrl + NStr::NumericToString(gi),
        fHTTP_AutoReconnect, &timeout);
    auto_ptr<CObjectIStream> obj_stream(CObjectIStream::Open(eSerial_AsnText, stream));
    CSeq_align_set align_set;
    try {
        *obj_stream >> align_set;
    } catch (CException& e) {
        LOG_POST(Error << "Failed to retrieve gi placements for gi|"
                 << gi << ", error: " << e.GetMsg());
        return mapped_locs;
    }
   
    typedef map<int, TSeqRange> TRangeMap;
    TRangeMap r_map;
    CTypeConstIterator<CSeq_align> aln_iter(align_set);
    for(; aln_iter; ++aln_iter) {
        const CSeq_align& aln = *aln_iter;
        if (aln.CheckNumRows() == 2  &&  aln.GetSegs().IsStd()  &&
            aln.GetSegs().GetStd().size() == 1) {
            int target_row = 0;
            if (aln.GetSeq_id(0).IsGi()  &&  aln.GetSeq_id(0).GetGi() == gi) {
                target_row = 1;
            }
            if (aln.GetSeq_id(target_row).IsGi()) {
                int target_gi = aln.GetSeq_id(target_row).GetGi();
                TSignedSeqRange range =
                    aln.GetSegs().GetStd().front()->GetSeqRange(target_row);
                TSeqPos from = (TSeqPos)range.GetFrom();
                TSeqPos to = (TSeqPos)range.GetTo();
                if (from > to) {
                    swap(from, to);
                }
                if (r_map.count(target_gi) == 0) {
                    r_map[target_gi] = TSeqRange(from, to);
                } else {
                    r_map[target_gi].CombineWith(TSeqRange(from, to));
                }
            }
        }
    }

    ITERATE (TRangeMap, iter, r_map) {
        CRef<CSeq_id> id(new CSeq_id);
        id->SetGi(iter->first);
        CRef<CSeq_loc> loc(new CSeq_loc(*id, iter->second.GetFrom(),
            iter->second.GetTo()));
        mapped_locs.push_back(loc);
    }
    return mapped_locs;
}


bool CSeqUtils::StringToRange(const string& range_str,
                              long& from, long& to)
{
    // Any input range string that follows this pattern will be
    // consisdered as a valid input:
    // "^[ \t]*[1-9][0-9,]*[ \t]*[kKmM]?((([ \t]*([-:]|\\.\\.)[ \t]*)|([ \t]+))[1-9][0-9,]*[ \t]*[kKmM]?)?[ \t]*$"
    // Some valid range examples:
    //   - 1000
    //   - [space]1000 -[tab]2000[tab]
    //   - [space]1000[space]..[space]2000
    //   - 10,000:2,000,000
    //   - 100 k : 1m
    //   - 1000[space]2000
    //   - [space]1000[tab]2000

    from = to = 0;
    string str = NStr::TruncateSpaces(range_str);
    size_t len = str.length();

    if (len == 0) return false;

    // The loop tries to accomplish the followings:
    // - remove ','
    // - replace 'k' or 'K' with '000'
    // - replace 'm' or 'M' with '000000'
    // - remove white paces (and tabs) before ',', 'k', 'K', 'm', and 'M'
    // - remove white space (and tabs) around any separator (':', '..', and '-')
    // - replace any separator with '-'
    // - replace spaces between two numbers with '-'
    string out_str;
    bool space_before_this = false;
    bool separator_before_this = false;
    for (size_t i = 0; i < len; ++i) {
        switch (str[i])
        {
        case ' ':
        case '\t':
            // ignore space after a separator
            if ( !separator_before_this )
                space_before_this = true;
            break;
        case ',':
            if (separator_before_this)  return false; // invalid
            // ignore space before ','
            space_before_this = false;
            break;
        case 'k':
        case 'K':
            if (separator_before_this)  return false; // invalid
            // ignore space before 'k' and 'K'
            space_before_this = false;
            // replace it with "000'
            out_str.append("000");
            break;
        case 'm':
        case 'M':
            if (separator_before_this)  return false; // invalid
            // ignore space before 'm' and 'M'
            space_before_this = false;
            // replace it with "0000000'
            out_str.append("000000");
            break;
        case '.':
        case '-':
        case ':':
            // ignore space before a separator
            space_before_this = false;
            separator_before_this = true;
            break;
        default:
            if (separator_before_this) {
                out_str.append("-");
                separator_before_this = false;
            } else if (space_before_this) {
                out_str.append("-");
                space_before_this = false;
            }
            out_str.append(1, str[i]);
            break;
       }
    }

    typedef vector<string> TPositions;
    TPositions pos;
    NStr::Tokenize(out_str, "-", pos);
    if (pos.size() < 3) {
        try {
            bool is_from = true;
            NON_CONST_ITERATE (TPositions, iter, pos) {
                NStr::TruncateSpaces(*iter);
                if (iter->empty()) continue;
                if (is_from) {
                    to = from = NStr::StringToLong(*iter);
                    is_from = false;
                } else {
                    to = NStr::StringToLong(*iter);
                }            
            }
        } catch (CException&) {
            return false;
        }
    }
    return true;
}


END_NCBI_SCOPE
