/*  $Id: ftglfontmanager.cpp 28056 2013-05-06 21:14:36Z rafanovi $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Bob Falk
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>
#include <gui/opengl/ftglfontmanager.hpp>

#include <corelib/ncbistr.hpp>
#include <corelib/ncbifile.hpp>

#ifdef HAVE_LIBFTGL
#include <FTGL/ftgl.h> 
#endif


BEGIN_NCBI_SCOPE

CFtglFontManager& CFtglFontManager::Instance()
{
    static CFtglFontManager tm;
    return tm;
}

CFtglFontManager::~CFtglFontManager()
{
    // Since this object is a singleton, memory (especially OpenGL) may already
    // be cleaned up by the system by the time this dtor is called.  User should call
    // 'Clear' to delete fonts.
    /*
    Clear();
    */
}

void CFtglFontManager::Clear()
{
    FontIter font;
    for(font = fonts.begin(); font != fonts.end(); ++font)
    {
        delete (*font).second;
    }

    fonts.clear();
}

void CFtglFontManager::SetFontPath(const string& p) 
{ 
    m_FontPath = p; 
    NStr::TruncateSpacesInPlace(m_FontPath);
    
    // If it is not blank, make sure there is a termination separator, e.g. '/':
    if (!NStr::IsBlank(m_FontPath)) {
        if (!CDirEntry::IsPathSeparator(m_FontPath[m_FontPath.length()-1])) { 
            m_FontPath += CDir::GetPathSeparator();
        }
        m_FontPath += "fonts";
        m_FontPath += CDir::GetPathSeparator();      
    }
}

FTFont* CFtglFontManager::GetFont(const char *filename, 
                                  unsigned int size, 
                                  EFtglFontType ft)
{
#ifdef HAVE_LIBFTGL
    char buf[256];
    sprintf(buf, "%s %i %i", filename, size, ft);
    string fontKey = string(buf);

    FontIter result = fonts.find(fontKey);
    if(result != fonts.end())
    {
        //_TRACE("Found font " << fontKey << " in list");
        return result->second;
    }

    string fullname = m_FontPath + string(filename);

    // Look for file first in default directory.  If not found
    // try to use fiilename as the full path name. If neither
    // work, return null
    {
        CDirEntry f1(fullname);
        if (!f1.Exists()) {
            CDirEntry f2(filename);
            if (!f2.Exists()) {
                ERR_POST(Error << "Font file: " << fullname << 
                    " failed to open. Default font path: " << m_FontPath );               
                return NULL;
            }
            else {
                fullname = filename;
            }
        }
    }

    FTFont* font = NULL;
    switch (ft) {
        case eTextureFont:
            font = new FTTextureFont(fullname.c_str());
            break;
        case ePixmapFont:
            font = new FTPixmapFont(fullname.c_str());
            break;
        case eBitmapFont:
            font = new FTBitmapFont(fullname.c_str());
            break;
        case eOutlineFont:
            font = new FTOutlineFont(fullname.c_str());
            break;
        case ePolygonFont:
            font = new FTPolygonFont(fullname.c_str());
            break;
        case eExtrudedFont:
            font = new FTExtrudeFont(fullname.c_str());
            break;
    };

    //_TRACE("Created font " << fontKey);

    if(font->Error())
    {
        ERR_POST(Error << "Font file: " << fullname << 
            " failed to open. Default font path: " << m_FontPath );
        delete font;
        return NULL;
    }

    if(!font->FaceSize(size, m_DeviceResolution))
    {
        ERR_POST(Error << "Font " << filename << " failed to set size " << size);
        delete font;
        return NULL;
    }
    font->UseDisplayList(true);

    fonts[fontKey] = font;

    return font;
#else
    ERR_POST(Error << "FTGL is unavailable.");
    return NULL;
#endif
}

END_NCBI_SCOPE
