/*  $Id: glcolortable.cpp 23958 2011-06-24 15:24:59Z wuliangs $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Mike DiCuccio, Andrey Yazhuk
 *
 * File Description:
 *    Incremental and mutually contrastive color table generator.
 */

#include <ncbi_pch.hpp>
#include <gui/opengl/glcolortable.hpp>


BEGIN_NCBI_SCOPE

////////////////////////////////////////////////////////////////////////////////
/// CRgbaColorTable
CRgbaColorTable::CRgbaColorTable()
    : m_MaxItems(8),
      m_ColorStep(1.0f),
      m_MinColor(0.0f),
      m_MaxColor(1.0f),
      m_ColorDiff(1.0f)
{
}


CRgbaColorTable::CRgbaColorTable(size_t items, float min_color, float max_color)
    : m_MaxItems(items),
      m_ColorStep(1.0f),
      m_MinColor(min_color),
      m_MaxColor(max_color),
      m_ColorDiff(1.0f)
{
    if (m_MaxColor < m_MinColor) {
        std::swap(m_MaxColor, m_MinColor);
    }
    m_ColorDiff = m_MaxColor - m_MinColor;
    x_CalcColorStep();
}


CRgbaColor CRgbaColorTable::GetColor(size_t i) const
{
    i = ::abs((int)(i % m_MaxItems));

    int reps = (int)(i / 7);
    int idx  = (int)(i % 7 + 1);

    CRgbaColor color;
    float val = (reps + 1) * m_ColorStep + m_MinColor;

    color.SetRed  (val * ((idx & 0x01)     ));
    color.SetGreen(val * ((idx & 0x02) >> 1));
    color.SetBlue (val * ((idx & 0x04) >> 2));

    return color;
}


////////////////////////////////////////////////////////////////////////////////
/// CRgbaGradColorTable

CRgbaGradColorTable::CRgbaGradColorTable()
{
}


CRgbaGradColorTable::CRgbaGradColorTable(int size)
{
    SetSize(size);
}


size_t CRgbaGradColorTable::GetSize() const
{
    return m_Colors.size();
}


void CRgbaGradColorTable::SetSize(size_t size)
{
    _ASSERT(size < 0xFFFF);
    m_Colors.resize(size);
}


void CRgbaGradColorTable::FillGradient(const CRgbaColor& start_c, const CRgbaColor& end_c)
{
    if( ! m_Colors.empty())   {
        FillGradient(0, (int)m_Colors.size(), start_c, end_c);
    }
}


void CRgbaGradColorTable::FillGradient(int start_i, int n,
                                     const CRgbaColor& start_c,
                                     const CRgbaColor& end_c)
{
    _ASSERT(start_i >= 0  &&  start_i < (int) m_Colors.size());
    _ASSERT((start_i + n) <= (int) m_Colors.size());
    _ASSERT(n >= 0);

    if(n)   {
        float base_r = start_c.GetRed();
        float base_g = start_c.GetGreen();
        float base_b = start_c.GetBlue();
        float base_a = start_c.GetAlpha();

        float d_r = end_c.GetRed() - base_r;
        float d_g = end_c.GetGreen() - base_g;
        float d_b = end_c.GetBlue() - base_b;
        float d_a = end_c.GetAlpha() - base_a;

        float K = 1.0f / (n - 1);
        for( int i = 0;  i < n;  i++ ) {
            float d = i * K;
            float r = base_r + d * d_r;
            float g = base_g + d * d_g;
            float b = base_b + d * d_b;
            float a = base_a + d * d_a;
            m_Colors[start_i + i].Set(r, g, b, a);
        }
    }
}


void CRgbaGradColorTable::FillGradient(int start_i, int n,
                                     const CRgbaColor& start_c,
                                     const CRgbaColor& mid_c,
                                     const CRgbaColor& end_c)
{
    _ASSERT(start_i >= 0  &&  start_i < (int) m_Colors.size());

    if (n > 0) {
        if ((n & 1) == 0) {
            LOG_POST(Warning << "CRgbaGradColorTable::FillGradient() - \"n\""
                     << "is an even number. Gradient table will not be symmetrical!");
        }
        int half = n / 2;
        // fill the first half
        FillGradient(start_i, half + 1, start_c, mid_c);
        // fill the second half
        FillGradient(start_i + half, n - half, mid_c, end_c);
    }
}


END_NCBI_SCOPE
