/*  $Id: glrender.cpp 27888 2013-04-19 13:38:25Z falkrb $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Bob Falk
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>
#include <gui/opengl/glrender.hpp>
#include <gui/opengl/glresmgr.hpp>
#include <gui/opengl/glutils.hpp>
#include <gui/utils/matrix4.hpp>

// define this to cause OpenGL actual state to be updated immediately on state
// update calls (useful if some code is using rendermanager and some isn't, and
// state is shared between the two)
#define GL_STATE_UPDATE

BEGIN_NCBI_SCOPE

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

CGlRender& GetGl()
{
    return CGlResMgr::Instance().GetCurrentRenderer().GetObject();
}

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

CGlRender::CGlRender() 
: m_CurrentMode(GL_NONE)
, m_CurrentFont(NULL)
{
    m_MatrixMode = GL_NONE;
    m_ModelView.Identity();

    m_State.Reset(new CGlState());
    m_RenderNode.SetState(m_State.GetPointer());
    m_QuadRenderNode.SetState(m_State.GetPointer());
}

void CGlRender::Clear()
{
    m_CurrentMode = GL_NONE;
    m_MatrixMode = GL_NONE;
    m_ModelView.Identity();

    if (!m_RenderNode.GetVBOGeom().IsNull()) {
        m_RenderNode.GetVBOGeom()->Destroy();
    }

    if (!m_QuadRenderNode.GetVBOGeom().IsNull()) {
        m_QuadRenderNode.GetVBOGeom()->Destroy();
    }
}

void CGlRender::Begin(GLenum mode, const CGlState& state)
{
    m_State.GetObject() = state;   
    Begin(mode);
}

void CGlRender::Begin(GLenum mode)
{
    // If End wasn't called after last begin log an error (but continue so that subsequent
    // rendering will work)
    if (m_CurrentMode != GL_NONE) {
        LOG_POST(Error << "Begin() called without first calling CGlRender::End()");
    }

    m_CurrentMode = mode;

    if (m_CurrentMode == GL_POLYGON) {
        LOG_POST(Error << "Error - GL_POLYGON rendering not supported");
        return;
    }
    else if (m_CurrentMode == GL_QUAD_STRIP) {
        LOG_POST(Error << "Error - GL_QUAD_STRIP rendering not supported");
    }

    m_VertexBuffer.clear();
    m_ColorBuffer.clear();
    m_NormalBuffer.clear();
    m_TexCoord2DBuffer1.clear();

    if (m_RenderNode.GetVBOGeom().IsNull()) {
        m_RenderNode.SetVBOGeom(CRef<CGlVboGeom>(CGlResMgr::Instance().GetVboGeom("RenderManager Geom")));
    }
}

void CGlRender::Vertex3f(GLfloat x, GLfloat y, GLfloat z)
{
    if (m_CurrentMode == GL_NONE) {
        LOG_POST(Error << "Vertex*() called without first calling CGlRender::Begin");
        return;
    }

    m_VertexBuffer.push_back(CVect3<float>(x,y,z) );

    // If quads, break up quads into triangle pairs.  Do this whenever
    // 4 new vertices have been added (ignore groups of 6 since they
    // are the previous triangles)    (we do this because quads are deprecated
    // in OpenGL 3+ and because PDF files do not support gouraud-shaded quads)    
    if (m_CurrentMode == GL_QUADS && m_VertexBuffer.size() % 6 == 4) {
        size_t v1 = m_VertexBuffer.size()-4;
        size_t v3 = v1 + 2;

        // error condition:
        if (m_State->GetPolygonMode() == GL_LINE) {
            _TRACE("Warning: Rendering triangle pairs for quads in line mode");
        }

        if (m_ColorBuffer.size() > 0) {
            x_SyncAttribs(m_ColorBuffer);
            m_ColorBuffer.push_back(m_ColorBuffer[v1]);
            m_ColorBuffer.push_back(m_ColorBuffer[v3]);
        }

        if (m_NormalBuffer.size() > 0) {
            x_SyncAttribs(m_NormalBuffer);
            m_NormalBuffer.push_back(m_NormalBuffer[v1]);
            m_NormalBuffer.push_back(m_NormalBuffer[v3]);
        }

        if (m_TexCoord2DBuffer1.size() > 0) {
            x_SyncAttribs(m_TexCoord2DBuffer1);
            m_TexCoord2DBuffer1.push_back(m_TexCoord2DBuffer1[v1]);
            m_TexCoord2DBuffer1.push_back(m_TexCoord2DBuffer1[v3]);
        }

        // Do vertices last since x_SyncAttribs syncs size of attribute
        // buffers to current size of this (m_VertexBuffer) buffer.
        m_VertexBuffer.push_back(m_VertexBuffer[v1]);
        m_VertexBuffer.push_back(m_VertexBuffer[v3]);
    }    
}

void CGlRender::Vertex2f(GLfloat x, GLfloat y)
{
    Vertex3f(x, y, 0.0f);
}
void CGlRender::Vertex2fv(GLfloat* v)
{
    Vertex3f(v[0], v[1], 0.0f);
}
void CGlRender::Vertex3fv(GLfloat* v)
{
    Vertex3f(v[0], v[1], v[2]);
}

void CGlRender::Vertex2d(GLdouble x, GLdouble y)
{
    Vertex3f(float(x), float(y), 0.0f);
}
void CGlRender::Vertex2dv(GLdouble* v)
{
    Vertex3f(float(v[0]), float(v[1]), 0.0f);
}
void CGlRender::Vertex3d(GLdouble x, GLdouble y, GLdouble z)
{
    Vertex3f(float(x), float(y), float(z));
}
void CGlRender::Vertex3dv(GLdouble* v)
{
    Vertex3f(float(v[0]), float(v[1]), float(v[2]));
}

void CGlRender::RectC(const TModelRect& rc)
{
    Rectf(float(rc.Left()), float(rc.Bottom()), float(rc.Right()), float(rc.Top()));
}
void CGlRender::RectC(const TVPRect& rc)
{ 
    Rectf(float(rc.Left()), float(rc.Bottom()), float(rc.Right()), float(rc.Top()));
}
void CGlRender::Rectf(GLfloat x1, GLfloat y1, GLfloat x2, GLfloat y2)
{
    if (m_CurrentMode != GL_NONE) {
        LOG_POST(Error << "Rect*() called between Begin() and End()");
        return;
    }   

    CVect3<float> scale(1.0f, 1.0f, 1.0f);
    m_QuadRenderNode.SetScale(scale);

    // convert to triangles unless we are polygon mode == gl_line, in which case do lines.
    // Thi is because triangles in line mode would show an extra line across the middle.
    if (m_State->GetPolygonMode() == GL_LINE) {
        Begin(GL_LINE_LOOP);
            Vertex3f(x1, y1, 0.0f);
            Vertex3f(x2, y1, 0.0f);
            Vertex3f(x2, y2, 0.0f);
            Vertex3f(x1, y2, 0.0f);
        End();
        return;
    } else if (m_QuadRenderNode.GetVBOGeom().IsNull() || 
               m_QuadRenderNode.GetVBOGeom()->GetVertexBuffer3D().GetActiveBufferSize() == 0) {
               m_QuadRenderNode.SetVBOGeom(CRef<CGlVboGeom>(CGlResMgr::Instance().GetVboGeom("RenderManager Quad Geom")));
               m_QuadRenderNode.SetDrawMode(GL_TRIANGLES);      

        vector<CVect3<float> > verts;
        verts.push_back(CVect3<float>(0.0f, 0.0f, 0.0f));
        verts.push_back(CVect3<float>(1.0f, 0.0f, 0.0f));
        verts.push_back(CVect3<float>(1.0f, 1.0f, 0.0f));

        verts.push_back(CVect3<float>(1.0f, 1.0f, 0.0f));
        verts.push_back(CVect3<float>(0.0f, 1.0f, 0.0f));
        verts.push_back(CVect3<float>(0.0f, 0.0f, 0.0f));

        m_QuadRenderNode.GetVertexBuffer3D().BufferData(verts.size(), &verts[0], GL_STATIC_DRAW);
    }

    if (m_State->GetScaleInvarient()) {
        CVect2<TModelUnit> s = m_State->GetScaleFactor();

        CVect3<float> scale(float(s.X()), float(s.Y()), 1.0f);                          
        m_RenderNode.SetScale(scale);
    }
    
    /*
    CMatrix4<float>  mat(m_ModelView);
    // scale and translate the buffered quad ( (0,0)..(1,1) ) to
    // the size ((x1, y1)..(x2,y2))
    CMatrix4<float>  mat2(float(x2-x1), 0.0f, 0.0f, x1,
                          0.0f, float(y2-y1), 0.0f, y1,
                          0.0f, 0.0f, 1.0f, 0.0f,
                          0.0f, 0.0f, 0.0f, 1.0f);    
    mat *= mat2;

    // Using the matrix stack to compute the same result can be done like this:
    //glMatrixMode(GL_MODELVIEW);
    //glPushMatrix();
    //glTranslatef(x1, y1, 0.0f);
    //glScalef(float(x2-x1), float(y2-y1), 1.0f);
    //glGetFloatv(GL_MODELVIEW_MATRIX, m_ModelView.GetData());
    //glPopMatrix();
    */
    // Same as above code, but in 1 step for efficiency:
    float sx = float(x2-x1);
    float sy = float(y2-y1);
                         // row1
    CMatrix4<float>  mat(m_ModelView[0]*sx, m_ModelView[1]*sy, m_ModelView[2], 
                         m_ModelView[0]*x1 + m_ModelView[1]*y1 + m_ModelView[3],
                         // row2
                         m_ModelView[4]*sx, m_ModelView[5]*sy, m_ModelView[6], 
                         m_ModelView[4]*x1 + m_ModelView[5]*y1 + m_ModelView[7],
                         // row3
                         m_ModelView[8]*sx, m_ModelView[9]*sy, m_ModelView[10], 
                         m_ModelView[8]*x1 + m_ModelView[9]*y1 + m_ModelView[11],
                         // row4
                         0.0f, 0.0f, 0.0f, 1.0f);



    m_QuadRenderNode.SetPosition(mat);
    
    x_RenderBuffer(&m_QuadRenderNode);
}
void CGlRender::Rectd(GLdouble x1, GLdouble y1, GLdouble x2, GLdouble y2)
{
    Rectf(float(x1), float(y1), float(x2), float(y2));
}

void CGlRender::Color4f(GLfloat r, GLfloat g, GLfloat b, GLfloat a)
{
    ColorC(CRgbaColor(r, g, b, a));
}
void CGlRender::Color3f(GLfloat r, GLfloat g, GLfloat b)
{
    ColorC(CRgbaColor(r, g, b, 1.0f));
}
void CGlRender::Color3fv(GLfloat* v)
{
    ColorC(CRgbaColor(v[0], v[1], v[2], 1.0f));
}
void CGlRender::Color4fv(GLfloat* v)
{
    ColorC(CRgbaColor(v[0], v[1], v[2], v[3]));
}

void CGlRender::Color3d(GLdouble r, GLdouble g, GLdouble b)
{
    ColorC(CRgbaColor(float(r), float(g), float(b), 1.0f));
}
void CGlRender::Color3dv(GLdouble* v)
{
    ColorC(CRgbaColor(float(v[0]), float(v[1]), float(v[2]), 1.0f));
}
void CGlRender::Color4d(GLdouble r, GLdouble g, GLdouble b, GLdouble a)
{
    ColorC(CRgbaColor(float(r), float(g), float(b), float(a)));
}
void CGlRender::Color4dv(GLdouble* v)
{
    ColorC(CRgbaColor(float(v[0]), float(v[1]), float(v[2]), float(v[3])));
}
void CGlRender::ColorC(const CRgbaColor& c)
{
#ifdef GL_STATE_UPDATE
    glColor4f(c.GetRed(), c.GetGreen(), c.GetBlue(), c.GetAlpha());
#endif

    // Color is the one command that can be called both between Begin()/End() and
    // outside of Begin()/End().  If we are outside, update the state. If we
    // are inside, update the attribute array.
    if (m_CurrentMode == GL_NONE) {
        m_State->ColorC(c);
        return;
    }

    size_t attrib_size = m_ColorBuffer.size();
    size_t vert_size = m_VertexBuffer.size();

    // If an attribute (color, normal, texcoord) is provided for any 
    // vertex, there most be an attribute for all vertices since buffers
    // must have same number of elements when rendered.  Attributes should
    // be provided ahead of vertices (it applies to the next vertex).

    // Just right 
    if (attrib_size == vert_size) {
        m_ColorBuffer.push_back(c);
    }
    // already have correct # of attributes - replace last element
    else if (attrib_size == vert_size + 1) {
        m_ColorBuffer.back() = c;
    }
    // too few attributes (could be any number missing up to vert_size).  This
    // is normal opengl usage - e.g. one color applies to all vertices for a tri
    if (attrib_size < vert_size) {
        x_SyncAttribs(m_ColorBuffer);
        m_ColorBuffer.push_back(c);
    }
    // this should never happen
    else if (attrib_size > vert_size + 1) {
        LOG_POST(Error << "Error - color attributes exceed expected count");
        m_ColorBuffer.back() = c;
    }
}


void CGlRender::TexCoord2f(GLfloat s, GLfloat t)
{
    CVect2<float> v(s, t);

    size_t attrib_size = m_TexCoord2DBuffer1.size();
    size_t vert_size = m_VertexBuffer.size();

    // If an attribute (color, normal, texcoord) is provided for any 
    // vertex, there most be an attribute for all vertices since buffers
    // must have same number of elements when rendered.  Attributes should
    // be provided ahead of vertices (it applies to the next vertex).

    // Just right 
    if (attrib_size == vert_size) {
        m_TexCoord2DBuffer1.push_back(v);
    }
    // already have correct # of attributes - replace last element
    else if (attrib_size == vert_size + 1) {
        m_TexCoord2DBuffer1.back() = v;
    }
    // too few attributes (could be any number missing up to vert_size).  This
    // is normal opengl usage - e.g. one color applies to all vertices for a tri
    if (attrib_size < vert_size) {
        x_SyncAttribs(m_TexCoord2DBuffer1);
        m_TexCoord2DBuffer1.push_back(v);
    }
    // this should never happen
    else if (attrib_size > vert_size + 1) {
        LOG_POST(Error << "Error - texcoord2 attributes exceed expected count");
        m_TexCoord2DBuffer1.back() = v;
    }
}

void CGlRender::TexCoord2fv(GLfloat* v)
{
    TexCoord2f(v[0], v[1]);
}
void CGlRender::TexCoord2d(GLdouble s, GLdouble t)
{
    TexCoord2f(float(s), float(t));
}
void CGlRender::TexCoord2dv(GLdouble* v)
{
    TexCoord2f(float(v[0]), float(v[1]));
}

void CGlRender::Normal3f(GLfloat x, GLfloat y, GLfloat z)
{
    CVect3<float> v(x, y, z);

    size_t attrib_size = m_NormalBuffer.size();
    size_t vert_size = m_VertexBuffer.size();

    // If an attribute (color, normal, texcoord) is provided for any 
    // vertex, there most be an attribute for all vertices since buffers
    // must have same number of elements when rendered.  Attributes should
    // be provided ahead of vertices (it applies to the next vertex).

    // Just right 
    if (attrib_size == vert_size) {
        m_NormalBuffer.push_back(v);
    }
    // already have correct # of attributes - replace last element
    else if (attrib_size == vert_size + 1) {
        m_NormalBuffer.back() = v;
    }
    // too few attributes (could be any number missing up to vert_size).  This
    // is normal opengl usage - e.g. one color applies to all vertices for a tri
    if (attrib_size < vert_size) {
        x_SyncAttribs(m_TexCoord2DBuffer1);
        m_NormalBuffer.push_back(v);
    }
    // this should never happen
    else if (attrib_size > vert_size + 1) {
        LOG_POST(Error << "Error - normal attributes exceed expected count");
        m_NormalBuffer.back() = v;
    }
}
void CGlRender::Normal3fv(GLfloat* v)
{
    Normal3f(v[0], v[1], v[2]);
}
void CGlRender::Normal3d(GLdouble x, GLdouble y, GLdouble z)
{
    Normal3f(float(x), float(y), float(z));
}
void CGlRender::Normal3dv(GLdouble* v)
{
    Normal3f(float(v[0]), float(v[1]), float(v[2]));
}


void CGlRender::End()
{   
    size_t vert_size = m_VertexBuffer.size();

    if (m_CurrentMode == GL_NONE) {
        LOG_POST(Error << "End() called without first calling CGlRender::Begin");
        // Can't render anything if we don't know what user wanted (tris? lines?)
        return;
    }

    if (vert_size == 0) {
        m_CurrentMode = GL_NONE;
        return;
    }

    // All attribute vectors that are in use must be the same 
    // size as the vertex buffer
    if (m_ColorBuffer.size() > 0 && m_ColorBuffer.size() < vert_size)
        x_SyncAttribs(m_ColorBuffer);
 
    if (m_NormalBuffer.size() > 0 && m_NormalBuffer.size() < vert_size)
        x_SyncAttribs(m_NormalBuffer);

    if (m_TexCoord2DBuffer1.size() > 0 && m_TexCoord2DBuffer1.size() < vert_size)
        x_SyncAttribs(m_TexCoord2DBuffer1);

    // For quads we set the render mode passed to OpenGL to triangles since we
    // break down quads into tris.
    if (m_CurrentMode != GL_QUADS)
        m_RenderNode.SetDrawMode(m_CurrentMode);
    else
        m_RenderNode.SetDrawMode(GL_TRIANGLES);

    m_CurrentMode = GL_NONE;

    CVect3<float> scale(1.0f, 1.0f, 1.0f);
    m_RenderNode.SetScale(scale);

    if (m_State->GetScaleInvarient()) {
        CVect2<TModelUnit> s = m_State->GetScaleFactor();

        CVect3<float> scale(float(s.X()), float(s.Y()), 1.0f);                          
        m_RenderNode.SetScale(scale);
    }

    // Sync all vertex and attribute buffers with the stl vectors. It would be more efficient
    // to copy directly into the vertex buffers but we can't do that with the begin/end api
    // since we don't know in advance how many elements there will be.   
    m_RenderNode.GetVertexBuffer3D().BufferData(vert_size, &m_VertexBuffer[0], GL_DYNAMIC_DRAW);

    if (m_ColorBuffer.size() > 0) {
        m_RenderNode.GetColorBuffer().BufferData(vert_size, &m_ColorBuffer[0], GL_DYNAMIC_DRAW);
    }
    else {
        m_RenderNode.GetColorBuffer().SetActiveBufferSize(0);
    }

    if (m_NormalBuffer.size() > 0) {
        m_RenderNode.GetNormalBuffer().BufferData(vert_size, &m_NormalBuffer[0], GL_DYNAMIC_DRAW);
    }
    else {
        m_RenderNode.GetNormalBuffer().SetActiveBufferSize(0);
    }

    if (m_TexCoord2DBuffer1.size() > 0) {
        m_RenderNode.GetTexCoordBuffer().BufferData(vert_size, &m_TexCoord2DBuffer1[0], GL_DYNAMIC_DRAW);
    }
    else {
        m_RenderNode.GetTexCoordBuffer().SetActiveBufferSize(0);
    }

    /*
    CMatrix4<float>  mat;   
    glGetFloatv(GL_MODELVIEW_MATRIX, mat.GetData());
    mat.Transpose();
    m_RenderNode.SetPosition(mat);
    */
    m_RenderNode.SetPosition(m_ModelView);
    
    x_RenderBuffer(&m_RenderNode);
}

void CGlRender::x_SyncAttribs(vector<CRgbaColor>& colors)
{
    // OpenGL color default
    CRgbaColor default_color(1.0f, 1.0f, 1.0f, 1.0f);

    // But if we already have attributes, most recent one is default:
    if (colors.size() > 0) {
        default_color = colors.back();               
    }

    // Copy default value into array until length equals vertex buffer length
    colors.insert(colors.end(), 
                  m_VertexBuffer.size()-colors.size(), 
                  default_color);
}

void CGlRender::x_SyncAttribs(vector<CVect3<float> >& vecs)
{
    // default
    CVect3<float> default_value(0.0f, 0.0f, 0.0f);

    // But if we already have attributes, most recent one is default:
    if (vecs.size() > 0) {
        default_value = vecs.back();               
    }

    // Copy default value into array until length equals vertex buffer length
    vecs.insert(vecs.end(), 
                m_VertexBuffer.size()-vecs.size(), 
                default_value);
}

void CGlRender::x_SyncAttribs(vector<CVect2<float> >& vecs)
{
    // default
    CVect2<float> default_value(0.0f, 0.0f);

    // But if we already have attributes, most recent one is default:
    if (vecs.size() > 0) {
        default_value = vecs.back();               
    }

    // Copy default value into array until length equals vertex buffer length
    vecs.insert(vecs.end(), 
                m_VertexBuffer.size()-vecs.size(), 
                default_value);
}



void CGlRender::BeginText(const CGlTextureFont* font, 
                                 const CRgbaColor& color)
{
    m_CurrentFont = font;
    m_TextColor = color;

    m_CurrentFont->BeginText();
}

void CGlRender::BeginText(const CGlTextureFont* font)
{
    m_CurrentFont = font;
    m_TextColor = m_State->GetColor();

    m_CurrentFont->BeginText();
}

void CGlRender::EndText()
{
    m_CurrentFont->EndText();
    m_CurrentFont = NULL;
}

void CGlRender::WriteText(TModelUnit x, TModelUnit y, 
                          const char* text,
                          TModelUnit rotate_degrees)
{
    if (m_CurrentFont == NULL) {
        LOG_POST(Error << "Unable to write text - must call BeginText() first");
        return;
    }

    glColor4fv(m_TextColor.GetColorArray());
    m_CurrentFont->WriteText(x, y, text, rotate_degrees);
}

void CGlRender::WriteText(TModelUnit x, TModelUnit y, 
                          TModelUnit width, TModelUnit height,
                          const char* text,
                          CGlTextureFont::TAlign align,
                          CGlTextureFont::ETruncate trunc,
                          TModelUnit rotate_degrees)
{
    if (m_CurrentFont == NULL) {
        LOG_POST(Error << "Unable to write text - must call BeginText() first");
        return;
    }

    glColor4fv(m_TextColor.GetColorArray());
    m_CurrentFont->WriteText(x, y, width, height, text, align, trunc, rotate_degrees);
}

void CGlRender::Viewport(GLint x, GLint y, GLsizei width, GLsizei height)
{
    m_Viewport[0] = x;
    m_Viewport[1] = y;
    m_Viewport[2] = width;
    m_Viewport[3] = height;

    glViewport(x, y, width, height);
}

void CGlRender::GetViewport(GLint *params)
{
    // Don't get from opengl since that is limited to max buffer size but viewport, 
    // when used for other targets, e.g. pdf, may be larger.
    params[0] = m_Viewport[0];
    params[1] = m_Viewport[1];
    params[2] = m_Viewport[2];
    params[3] = m_Viewport[3];
}
    
void CGlRender::MatrixMode(GLenum mode)
{
    glMatrixMode(mode);
    m_MatrixMode = mode;
}
    
void CGlRender::PushMatrix()
{
    glPushMatrix();
}
void CGlRender::PopMatrix()
{
    glPopMatrix();
    if (m_MatrixMode == GL_MODELVIEW) {
        glGetFloatv(GL_MODELVIEW_MATRIX, m_ModelView.GetData());
        m_ModelView.Transpose();
    }
}
void CGlRender::LoadIdentity()
{
    glLoadIdentity();
    if (m_MatrixMode == GL_MODELVIEW) {
        m_ModelView.Identity();
    }
}
void CGlRender::LoadMatrixf(const GLfloat *m)
{
    glLoadMatrixf(m);
    if (m_MatrixMode == GL_MODELVIEW) {        
        glGetFloatv(GL_MODELVIEW_MATRIX, m_ModelView.GetData());
        m_ModelView.Transpose();
    }
}
void CGlRender::LoadMatrixd(const GLdouble* m)
{
    glLoadMatrixd(m);
    if (m_MatrixMode == GL_MODELVIEW) {        
        glGetFloatv(GL_MODELVIEW_MATRIX, m_ModelView.GetData());
        m_ModelView.Transpose();
    }
}
void CGlRender::MultMatrixf(const GLfloat* m)
{
    glMultMatrixf(m);
    if (m_MatrixMode == GL_MODELVIEW) {        
        glGetFloatv(GL_MODELVIEW_MATRIX, m_ModelView.GetData());
        m_ModelView.Transpose();
    }
}
void CGlRender::MultMatrixd(const GLdouble* m)
{
    glMultMatrixd(m);
    if (m_MatrixMode == GL_MODELVIEW) {        
        glGetFloatv(GL_MODELVIEW_MATRIX, m_ModelView.GetData());
        m_ModelView.Transpose();
    }
}

void CGlRender::Rotated(GLdouble angle, GLdouble x, GLdouble y, GLdouble z)
{
    glRotated(angle, x, y, z);
    if (m_MatrixMode == GL_MODELVIEW) {        
        glGetFloatv(GL_MODELVIEW_MATRIX, m_ModelView.GetData());
        m_ModelView.Transpose();
    }
}
void CGlRender::Rotatef(GLfloat angle, GLfloat x, GLfloat y, GLfloat z)
{
    glRotatef(angle, x, y, z);
    if (m_MatrixMode == GL_MODELVIEW) {        
        glGetFloatv(GL_MODELVIEW_MATRIX, m_ModelView.GetData());
        m_ModelView.Transpose();
    }
}
void CGlRender::Scaled(GLdouble x, GLdouble y, GLdouble z)
{
    glScaled(x, y, z);
    if (m_MatrixMode == GL_MODELVIEW) {        
        glGetFloatv(GL_MODELVIEW_MATRIX, m_ModelView.GetData());
        m_ModelView.Transpose();
    }
}
void CGlRender::Scalef(GLfloat x, GLfloat y, GLfloat z)
{
    glScalef(x, y, z);
    if (m_MatrixMode == GL_MODELVIEW) {        
        glGetFloatv(GL_MODELVIEW_MATRIX, m_ModelView.GetData());
        m_ModelView.Transpose();
    }
}
void CGlRender::Translated(GLdouble x, GLdouble y, GLdouble z)
{
    glTranslated(x, y, z);
    if (m_MatrixMode == GL_MODELVIEW) {        
        glGetFloatv(GL_MODELVIEW_MATRIX, m_ModelView.GetData());
        m_ModelView.Transpose();
    }
}
void CGlRender::Translatef(GLfloat x, GLfloat y, GLfloat z)
{
    glTranslatef(x, y, z);
    if (m_MatrixMode == GL_MODELVIEW) {        
        glGetFloatv(GL_MODELVIEW_MATRIX, m_ModelView.GetData());
        m_ModelView.Transpose();
    }
}

void CGlRender::Ortho(GLdouble left, GLdouble right,
                      GLdouble bottom, GLdouble top,
                      GLdouble nearVal, GLdouble farVal)
{
    glOrtho(left, right, bottom, top, nearVal, farVal);
}
void CGlRender::Perspective(GLdouble fovy, GLdouble aspect, 
                       GLdouble zNear, GLdouble zFar)
{
    gluPerspective(fovy, aspect, zNear, zFar);
}

///
/// All the OpenGL state commands set both the state object AND update
/// the opengl state.  The OpenGL state call should be unnecessary, but
/// as long as the product has both rendermanager-based rendering
/// and non-rendermanager based rendering, keeping the state as it
/// was before should help prevent visual artifacts in non-rendermanager code.
///

void CGlRender::PushAttrib(GLbitfield mask)
{
    pair<GLbitfield,CGlState> s(mask, m_State.GetNCObject());

    // The smallest maximum stacksize for attributes in OpenGL is 16, so we'll
    // give a warning when that number is exceeded (but not an error)
    if (m_PushedState.size() > 16) {
        LOG_POST(Warning << "Attribute stack size exceeded in CGlRender: " << m_PushedState.size()+1 );
        return;
    }

    m_PushedState.push_back(s);
}

void CGlRender::PopAttrib()
{
    if (m_PushedState.size() == 0) {
        LOG_POST(Error << "Attribute stack underflow - no state to pop");
        return;
    }

    pair<GLbitfield,CGlState> s = m_PushedState.back();
    m_PushedState.pop_back();

    // if the pushed attribute flags were not for all attributes (GL_ALL_ATTRIB_BITS),
    // then merge the ones which were not included in the push (and which are defined
    // in CGlState).
    m_State->MergeStates(s.second, s.first);
}

void CGlRender::Enable(GLenum glstate) 
{ 
    if (m_CurrentMode != GL_NONE) {
        LOG_POST(Error << "CGlState::Enable() called between Begin() and End()");
    }

#ifdef GL_STATE_UPDATE
    glEnable(glstate);
#endif

    m_State->Enable(glstate); 
}
void CGlRender::Disable(GLenum glstate) 
{ 
    if (m_CurrentMode != GL_NONE) {
        LOG_POST(Error << "CGlState::Disable() called between Begin() and End()");
    }

#ifdef GL_STATE_UPDATE
    glDisable(glstate);
#endif

    m_State->Disable(glstate); 
}
void CGlRender::LineWidth(GLfloat w)
{
    if (m_CurrentMode != GL_NONE) {
        LOG_POST(Error << "CGlState::LineWidth() called between Begin() and End()");
    }

#ifdef GL_STATE_UPDATE
    glLineWidth(w);
#endif

    m_State->LineWidth(w);
}
void CGlRender::PointSize(GLfloat s) 
{
    if (m_CurrentMode != GL_NONE) {
        LOG_POST(Error << "CGlState::PointSize() called between Begin() and End()");
    }
    
#ifdef GL_STATE_UPDATE
    glPointSize(s);
#endif

    m_State->PointSize(s);
}
void CGlRender::ShadeModel(GLenum mode)
{
    if (m_CurrentMode != GL_NONE) {
        LOG_POST(Error << "CGlState::ShadeModel() called between Begin() and End()");
    }
    
#ifdef GL_STATE_UPDATE
    glShadeModel(mode);
#endif

    m_State->ShadeModel(mode);
}

void CGlRender::Scissor(GLint x, GLint y,
                         GLsizei width, GLsizei height)
{
    if (m_CurrentMode != GL_NONE) {
        LOG_POST(Error << "CGlState::ShadeModel() called between Begin() and End()");
    }
    
#ifdef GL_STATE_UPDATE
    glScissor(x,y,width,height);
#endif

    m_State->Scissor(x,y,width,height);
}

void CGlRender::ColorMask(GLboolean red, 
                                 GLboolean green, 
                                 GLboolean blue, 
                                 GLboolean alpha)
{
    if (m_CurrentMode != GL_NONE) {
        LOG_POST(Error << "CGlState::ColorMask() called between Begin() and End()");
    }
    
#ifdef GL_STATE_UPDATE
    glColorMask(red, green, blue, alpha);
#endif

    m_State->ColorMask(red, green, blue, alpha);
}
void CGlRender::PolygonMode(GLenum face, GLenum mode)
{
    if (m_CurrentMode != GL_NONE) {
        LOG_POST(Error << "CGlState::PolygonMode() called between Begin() and End()");
    }
    
#ifdef GL_STATE_UPDATE
    glPolygonMode(face, mode);
#endif

    m_State->PolygonMode(face, mode);
}
void CGlRender::LineStipple(GLint factor, GLushort pattern)
{
    if (m_CurrentMode != GL_NONE) {
        LOG_POST(Error << "CGlState::LineStipple() called between Begin() and End()");
    }
    
#ifdef GL_STATE_UPDATE
    glLineStipple(factor, pattern);
#endif

    m_State->LineStipple(factor, pattern);
}
void CGlRender::PolygonStipple(GLubyte* mask)
{
    if (m_CurrentMode != GL_NONE) {
        LOG_POST(Error << "CGlState::PolygonStipple() called between Begin() and End()");
    }
    
#ifdef GL_STATE_UPDATE
    glPolygonStipple(mask);
#endif

    m_State->PolygonStipple(mask);
}
void CGlRender::BlendFunc(GLenum sfactor, GLenum dfactor)
{
    if (m_CurrentMode != GL_NONE) {
        LOG_POST(Error << "CGlState::BlendFunc() called between Begin() and End()");
    }
    
#ifdef GL_STATE_UPDATE
    glBlendFunc(sfactor, dfactor);
#endif

    m_State->BlendFunc(sfactor, dfactor);
}
void CGlRender::BlendFuncSeparate(GLenum srcRGB, GLenum dstRGB,
                                         GLenum srcAlpha, GLenum dstAlpha) 
{
    if (m_CurrentMode != GL_NONE) {
        LOG_POST(Error << "CGlState::BlendFuncSeparate() called between Begin() and End()");
    }
    
#ifdef GL_STATE_UPDATE
    // not 1.1
#endif

    m_State->BlendFuncSeparate(srcRGB, dstRGB, srcAlpha, dstAlpha);
}
void CGlRender::BlendColor(const CRgbaColor& c)
{
    if (m_CurrentMode != GL_NONE) {
        LOG_POST(Error << "CGlState::BlendColor() called between Begin() and End()");
    }
    
#ifdef GL_STATE_UPDATE
    // Not 1.1
    //glBlendColor(c.GetRed(), c.GetGreen(), c.GetBlue(), c.GetAlpha());
#endif

    m_State->BlendColor(c);
}
void CGlRender::ScaleInvarient(bool b, CVect2<TModelUnit> scale)
{
    if (m_CurrentMode != GL_NONE) {
        LOG_POST(Error << "CGlState::ScaleInvarient() called between Begin() and End()");
    }
    m_State->ScaleInvarient(b, scale);
}
void CGlRender::ScaleFactor(const CVect2<TModelUnit>& scale)
{
    if (m_CurrentMode != GL_NONE) {
        LOG_POST(Error << "CGlState::ScaleFactor() called between Begin() and End()");
    } 
    m_State->ScaleFactor(scale);
}
void CGlRender::LineJoinStyle(ELineJoinStyle s)
{
    if (m_CurrentMode != GL_NONE) {
        LOG_POST(Error << "CGlState::LineJoinStyle() called between Begin() and End()");
    }
    m_State->LineJoinStyle(s);
}
void CGlRender::LineCapStyle(ELineCapStyle c)
{
    if (m_CurrentMode != GL_NONE) {
        LOG_POST(Error << "CGlState::LineCapStyle() called between Begin() and End()");
    }
    m_State->LineCapStyle(c);
}
void CGlRender::PdfShadeStyle(EPdfShadeStyle s)
{
    if (m_CurrentMode != GL_NONE) {
        LOG_POST(Error << "CGlState::PdfShadeStyle() called between Begin() and End()");
    }
    m_State->PdfShadeStyle(s);
}

void CGlRender20::Render(CGlPane& pane, CGlModel2D* model)
{
    if (model != NULL) {
        model->Render(pane);
    }
}

void CGlRender20::x_RenderBuffer(CGlVboNode* node)
{
    node->Render();        
}

void CGlRender11::Render(CGlPane& pane, CGlModel2D* model)
{
    if (model != NULL) {
        model->Render(pane);
    }
}

void CGlRender11::x_RenderBuffer(CGlVboNode* node)
{
    node->Render();
}


















////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

void CGlRenderDebug::Begin(GLenum mode, const CGlState& state)
{
    glBegin(mode);
}

void CGlRenderDebug::Begin(GLenum mode)
{
    glBegin(mode);
}

void CGlRenderDebug::Vertex3f(GLfloat x, GLfloat y, GLfloat z)
{
    glVertex3f(x,y,z);
}

void CGlRenderDebug::RectC(const TModelRect& rc)
{
    glRectC(rc);   
}

void CGlRenderDebug::Rectf(GLfloat x1, GLfloat y1, GLfloat x2, GLfloat y2)
{
    //glRectf(x1, y1, x2, y2);
    glBegin(GL_QUADS);
        glVertex2f(x1, y1);
        glVertex2f(x2, y1);
        glVertex2f(x2, y2);
        glVertex2f(x1, y2);
    glEnd(); 
}

void CGlRenderDebug::Color4f(GLfloat r, GLfloat g, GLfloat b, GLfloat a)
{
    glColor4f(r,g,b,a);
}

void CGlRenderDebug::TexCoord2f(GLfloat s, GLfloat t)
{
    glTexCoord2f(s, t);
}

void CGlRenderDebug::Normal3f(GLfloat x, GLfloat y, GLfloat z)
{
    glNormal3f(x,y,z);
}

void CGlRenderDebug::End()
{   
    glEnd();
}

void CGlRenderDebug::BeginText(const CGlTextureFont* font, 
                               const CRgbaColor& color)
{
    CGlRenderDebug::BeginText(font);

    m_TextColor = color;
    glColor4fv(m_TextColor.GetColorArray());
}

void CGlRenderDebug::BeginText(const CGlTextureFont* font)
{
    m_CurrentFont = font;    

    if (font->GetFontFace() == CGlTextureFont::eFontFace_Courier) {
        m_BitmapFont.SetFontFace(CGlBitmapFont::eFontFace_Courier);
    }
    else if (font->GetFontFace() == CGlTextureFont::eFontFace_Helvetica) {
        m_BitmapFont.SetFontFace(CGlBitmapFont::eFontFace_Helvetica);
    }
    else if (font->GetFontFace() == CGlTextureFont::eFontFace_Lucida) {
        m_BitmapFont.SetFontFace(CGlBitmapFont::eFontFace_Lucida);
    }
    else if (font->GetFontFace() == CGlTextureFont::eFontFace_TimesRoman) {
        m_BitmapFont.SetFontFace(CGlBitmapFont::eFontFace_TimesRoman);
    }
    else if (font->GetFontFace() == CGlTextureFont::eFontFace_Fixed) {
        m_BitmapFont.SetFontFace(CGlBitmapFont::eFontFace_Fixed);
    }
    /*
    else if (font->GetFontFace() == CGlTextureFont::eFontFace_Clean) {
        m_BitmapFont.SetFontFace(CGlBitmapFont::eFontFace_Clean);
    }
    else if (font->GetFontFace() == CGlTextureFont::eFontFace_Bitmap) {
        m_BitmapFont.SetFontFace(CGlBitmapFont::eFontFace_Bitmap);
    }
    */

    m_BitmapFont.SetFontSize((CGlBitmapFont::EFontSize)font->GetFontSize());
}

void CGlRenderDebug::EndText()
{    
    m_CurrentFont = NULL;
}

void CGlRenderDebug::WriteText(TModelUnit x, TModelUnit y, 
                               const char* text,
                               TModelUnit rotate_degrees)
{
    m_BitmapFont.TextOut(x, y, text);
}

void CGlRenderDebug::WriteText(TModelUnit x, TModelUnit y, 
                               TModelUnit width, TModelUnit height,
                               const char* text,
                               CGlTextureFont::TAlign align,
                               CGlTextureFont::ETruncate trunc,
                               TModelUnit rotate_degrees)
{    
    //m_BitmapFont.TextOut(x, y, width, height, text, align, trunc);
    m_CurrentFont->BeginText();
    m_CurrentFont->WriteText(x, y, width, height, text, align, trunc, rotate_degrees);
    m_CurrentFont->EndText();
}


///
/// All the OpenGL state commands set both the state object AND update
/// the opengl state.  The OpenGL state call should be unnecessary, but
/// as long as the product has both rendermanager-based rendering
/// and non-rendermanager based rendering, keeping the state as it
/// was before should help prevent visual artifacts in non-rendermanager code.
///

void CGlRenderDebug::PushAttrib(GLbitfield mask)
{
    glPushAttrib(mask);
}

void CGlRenderDebug::PopAttrib() 
{
    glPopAttrib();
}

void CGlRenderDebug::Enable(GLenum glstate) 
{ 
    glEnable(glstate);
}

void CGlRenderDebug::Disable(GLenum glstate) 
{ 
    glDisable(glstate);
}

void CGlRenderDebug::LineWidth(GLfloat w)
{
    glLineWidth(w);
}

void CGlRenderDebug::PointSize(GLfloat s) 
{
    glPointSize(s);
}

void CGlRenderDebug::ShadeModel(GLenum mode)
{
    glShadeModel(mode);
}

void CGlRenderDebug::Scissor(GLint x, GLint y,
                            GLsizei width, GLsizei height)
{
   glScissor(x,y,width,height);
}

void CGlRenderDebug::ColorMask(GLboolean red, 
                                 GLboolean green, 
                                 GLboolean blue, 
                                 GLboolean alpha)
{
    glColorMask(red, green, blue, alpha);
}

void CGlRenderDebug::PolygonMode(GLenum face, GLenum mode)
{
    glPolygonMode(face, mode);
}

void CGlRenderDebug::LineStipple(GLint factor, GLushort pattern)
{
    glLineStipple(factor, pattern);
}

void CGlRenderDebug::PolygonStipple(GLubyte* mask)
{
    glPolygonStipple(mask);
}

void CGlRenderDebug::BlendFunc(GLenum sfactor, GLenum dfactor)
{
    glBlendFunc(sfactor, dfactor);
}

void CGlRenderDebug::BlendFuncSeparate(GLenum srcRGB, GLenum dstRGB,
                                         GLenum srcAlpha, GLenum dstAlpha) 
{
}

void CGlRenderDebug::BlendColor(const CRgbaColor& c)
{
}

void CGlRenderDebug::ScaleInvarient(bool b, CVect2<TModelUnit> scale)
{
}

void CGlRenderDebug::ScaleFactor(const CVect2<TModelUnit>& scale)
{
}

void CGlRenderDebug::LineJoinStyle(ELineJoinStyle s)
{
}

void CGlRenderDebug::LineCapStyle(ELineCapStyle c)
{
}

void CGlRenderDebug::PdfShadeStyle(EPdfShadeStyle s)
{
}


END_NCBI_SCOPE
