/*  $Id: glresmgr.cpp 27651 2013-03-20 14:00:46Z falkrb $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Bob Falk
 *
 * File Description:
 *    OpenGL resource manager implementation
 */


#include <ncbi_pch.hpp>
#include <gui/opengl/glresmgr.hpp>
#include <gui/opengl/glutils.hpp>
#include <gui/opengl/glbuffer.hpp>
#include <gui/opengl/glvbogeom.hpp>

BEGIN_NCBI_SCOPE

CGlResMgr* CGlResMgr::m_StaticInstance = NULL;

CGlResMgr& CGlResMgr::Instance()
{
    //static CGlResMgr instance;
    //return instance;
    if (m_StaticInstance == NULL) {
        // Create the instance.  This will not succeed if the OpenGL context
        // has not yet been initialized.
        try {
            m_StaticInstance = new CGlResMgr();
        }
        catch(...) {
            // Make sure instance stays NULL, and re-throw
            m_StaticInstance = NULL;
            NCBI_THROW(CException, eUnknown,
                "Attempt to access OpenGL prior to context creation");
        }
    }

    return *m_StaticInstance;
}

bool CGlResMgr::Initialized()
{
    return (m_StaticInstance!=NULL);
}

CGlResMgr::CGlResMgr()
{
    SetApiLevel();

    CRef<CGlRender>  rm;

    if (m_ApiLevel == eOpenGL20) {
        rm.Reset(new CGlRender20());
        AddRenderer(rm);
        SetCurrentRenderer(rm);
    }
    else if (m_ApiLevel == eOpenGL11) {
        rm.Reset(new CGlRender11());
        AddRenderer(rm);
        SetCurrentRenderer(rm);
    }
    else if (m_ApiLevel == eRenderDebug) {
        rm.Reset(new CGlRenderDebug());
        AddRenderer(rm);
        SetCurrentRenderer(rm);
    }
    else {
        _TRACE("Error - valid OpenGL version not detected.");
    }
}

CGlResMgr::~CGlResMgr()
{
}

void CGlResMgr::Clear()
{
}
    
void CGlResMgr::SetApiLevel()
{
    const char *v = (const char*)glGetString(GL_VERSION);

    // If OpenGL is not available, NULL will be returned.  Throw an exception
    if (v==NULL) {
        NCBI_THROW(CException, eUnknown,
            "Attempt to access OpenGL prior to context creation");
    }

    string version = v;
    vector<string> arr;

    if (version != "")
        NStr::Tokenize(version, ". \t", arr);

    bool version_two = false;

    // Check major version
    int major_version = 0;
    if (arr.size() >= 1) {       
        try {
            major_version = NStr::StringToInt(arr[0]);
            if (major_version > 1) {
                version_two = true;
            }
        }
        catch (CException& ) {
            //didn't parse as an int...
            m_ApiLevel = eApiUndefined;
        }
    }
    else  {
         m_ApiLevel = eApiUndefined;
    }

    // Can check minor version too, if needed:
    /*
    if (arr.size() > 1 && major_version == 1) {
        try {
            int minor_version = NStr::StringToInt(arr[1]);
            if (minor_version > 4) {
                version_two = true;
            }
        }
        catch (CException&) {
            //minor version didn't parse as an int
        }
    }
    */

    if (version_two)
        m_ApiLevel = eOpenGL20;
    else
        m_ApiLevel = eOpenGL11;
}
void CGlResMgr::SetApiLevel(ERenderTarget level)
{
    m_ApiLevel = level;
}

CGlVboGeom* CGlResMgr::GetVboGeom(const string& name)
{
    if (m_ApiLevel == eOpenGL20)
        return new CGlVboGeom20();
    else
        return new CGlVboGeom11();   
}

  
void CGlResMgr::AddModel(CIRef<CGlModel2D>& m)
{ 
    CIRef<CGlModel2D> current = GetModel(m->GetName());
    
    if (!current.IsNull())
        m_Models2D.push_back(m);
}

CIRef<CGlModel2D> CGlResMgr::GetModel(const string& name)
{
    vector<CIRef<CGlModel2D> >::iterator iter;

    for (iter=m_Models2D.begin(); iter!=m_Models2D.end(); ++iter) {
        if ((**iter).GetName() == name)
            return *iter;
    }

    return CIRef<CGlModel2D>();
}

CIRef<CGlModel2D> CGlResMgr::RemoveModel(CIRef<CGlModel2D>& m)
{
    vector<CIRef<CGlModel2D> >::iterator iter;
    iter = std::find(m_Models2D.begin(), m_Models2D.end(), m);
    
    if (iter != m_Models2D.end()) {
        m_Models2D.erase(iter);
        return *iter;
    }

    return CIRef<CGlModel2D>();
}


void CGlResMgr::AddRenderer(CRef<CGlRender>& rm)
{
    vector<CRef<CGlRender> >::iterator iter;
    for (iter = m_Renderers.begin(); iter != m_Renderers.end(); ++iter) {
        if ((*iter)->GetApi() == rm->GetApi() && (*iter)->GetName() == rm->GetName()) {
            break;
        }
    }

    if (iter == m_Renderers.end())
        m_Renderers.push_back(rm);
}

CRef<CGlRender> CGlResMgr::RemoveRenderer(CRef<CGlRender>& rm)
{
    vector<CRef<CGlRender> >::iterator iter;
    iter = std::find(m_Renderers.begin(), m_Renderers.end(), rm);
    
    if (iter != m_Renderers.end()) {
        m_Renderers.erase(iter);
        return *iter;
    }

    return CRef<CGlRender>();
}
    
CRef<CGlRender> CGlResMgr::GetRenderer(const string& rm_name)
{
    vector<CRef<CGlRender> >::iterator iter;

    for (iter = m_Renderers.begin(); iter != m_Renderers.end(); ++iter) {
        if ((*iter)->GetName() == rm_name) {
            return *iter;
        }
    }
    
    return CRef<CGlRender>();
}

CRef<CGlRender> CGlResMgr::GetRenderer(ERenderTarget target)
{
    vector<CRef<CGlRender> >::iterator iter;

    for (iter=m_Renderers.begin(); iter!=m_Renderers.end(); ++iter) {
        if ((**iter).GetApi() == target)
            return *iter;
    }

    return CRef<CGlRender>();
}

void CGlResMgr::SetCurrentRenderer(CRef<CGlRender>& rm)
{
    vector<CRef<CGlRender> >::iterator iter;
    iter = std::find(m_Renderers.begin(), m_Renderers.end(), rm);
    
    // Set current rendermanager to NULL if not found (so caller will know update
    // didn't work and can fix).
    if (iter != m_Renderers.end())
        m_CurrentRenderer = rm;
    else
        m_CurrentRenderer.ReleaseOrNull();
}




END_NCBI_SCOPE
