/*  $Id: gltexmanager.cpp 22736 2010-12-16 21:39:25Z katargir $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Mike DiCuccio
 *
 * File Description:
 *
 */


#include <ncbi_pch.hpp>
#include <gui/opengl/gltexture.hpp>
#include <gui/opengl/gltexmanager.hpp>
#include <util/image/image_io.hpp>


BEGIN_NCBI_SCOPE

// our list of images
CGlTextureManager::TImages CGlTextureManager::sm_TexImages;

// mutex guarding access to our images
DEFINE_STATIC_FAST_MUTEX(sm_Mutex);


//
// GetTexture()
// request an image by filename.  if this image is already loaded, return the
// previously loaded image.
//
CGlTexture* CGlTextureManager::GetTexture(const string& filename)
{
    CFastMutexGuard LOCK(sm_Mutex);

    // first, scan through our list of images to see if the
    // requested image was already loaded
    TImages::iterator iter = sm_TexImages.begin();
    for ( ; iter != sm_TexImages.end(); ++iter) {
        if ((*iter)->GetImageTag() == filename) {
            return (*iter);
        }
    }

    //
    // okay, image not found.  read from disk.
    //
    CNcbiIfstream istr(filename.c_str(), ios::in|ios::binary);
    CRef<CGlTexture> tex_image(new CGlTexture(istr, filename));
    sm_TexImages.push_back(tex_image);
    return tex_image;
}


//
// Release()
// dispose of an image, once done
//
void CGlTextureManager::Release(CGlTexture *image)
{
    CFastMutexGuard LOCK(sm_Mutex);
    if (!image) {
        return;
    }

    // scan through the list of loaded images, looking for the
    // image in question
    CRef<CGlTexture> handle(image);
    LOG_POST(Info
             << "CGlTextureManager::Release(): releasing texture "
             << image << " (" << image->GetImageTag());

    TImages::iterator iter = sm_TexImages.begin();
    while (iter != sm_TexImages.end()) {
        while (iter != sm_TexImages.end()  &&  *iter == image) {
            iter = sm_TexImages.erase(iter);
        }
        ++iter;
    }
}


//
// ReleaseAll()
// unload all images.  This is meant as a final clean-up operation
//
void CGlTextureManager::ReleaseAll(void)
{
    CFastMutexGuard LOCK(sm_Mutex);
    sm_TexImages.clear();
}


END_NCBI_SCOPE
