/*  $Id: gltexture.cpp 23968 2011-06-27 17:32:09Z kuznets $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Mike DiCuccio
 *
 * File Description:
 *
 */


#include <ncbi_pch.hpp>
#include <gui/opengl/gltexture.hpp>
#include <gui/opengl/glexception.hpp>
#include <util/image/image_io.hpp>
#include <gui/opengl.h>

BEGIN_NCBI_SCOPE


//
// static flag: can we use texture objects?
// 0 = no   1 = yes   2 = haven't tested yet
CGlTexture::ETexSupport CGlTexture::sm_TexObjFlag = CGlTexture::eNotChecked;


// default ctor
CGlTexture::CGlTexture()
    : m_TexId(0),
      m_WrapS(GL_CLAMP),
      m_WrapT(GL_CLAMP),
      m_FilterMin(GL_LINEAR),
      m_FilterMag(GL_LINEAR),
      m_TexEnv(GL_MODULATE)
{
}


//
// ctor: create an empty image
CGlTexture::CGlTexture(size_t w, size_t h, size_t depth)
    : m_TexId(0),
      m_WrapS(GL_CLAMP),
      m_WrapT(GL_CLAMP),
      m_FilterMin(GL_LINEAR),
      m_FilterMag(GL_LINEAR),
      m_TexEnv(GL_MODULATE)
{
    Init(w, h, depth);
}

// ctor:  create a texture with the texture id of a texture
// created elsewhere
CGlTexture::CGlTexture(GLuint tex_id)
    : m_TexId(tex_id),
      m_WrapS(GL_CLAMP),
      m_WrapT(GL_CLAMP),
      m_FilterMin(GL_LINEAR),
      m_FilterMag(GL_LINEAR),
      m_TexEnv(GL_MODULATE)
{
    // only works if it's really a texture
    if (!glIsTexture(tex_id))
        m_TexId = 0;
}


//
// ctor: create an image around an existing image
CGlTexture::CGlTexture(CImage* image)
    : m_TexId(0),
      m_WrapS(GL_CLAMP),
      m_WrapT(GL_CLAMP),
      m_FilterMin(GL_LINEAR),
      m_FilterMag(GL_LINEAR),
      m_TexEnv(GL_MODULATE)
{
    Swallow(image);
}


//
// get an image from a file
CGlTexture::CGlTexture(CNcbiIstream& istr, const string& tag)
    : m_ImageTag(tag),
      m_TexId(0),
      m_WrapS(GL_CLAMP),
      m_WrapT(GL_CLAMP),
      m_FilterMin(GL_LINEAR),
      m_FilterMag(GL_LINEAR),
      m_TexEnv(GL_MODULATE)
{
    m_Image.Reset(CImageIO::ReadImage(istr));
    if ( !m_Image ) {
        string msg("CGlTexture(): cannot read image");
        NCBI_THROW(COpenGLException, eTextureError, msg);
    }
}


//
// initialize an empty image
//
void CGlTexture::Init(size_t w, size_t h, size_t d)
{
    Clear();

    // initialize our image
    m_Image.Reset(new CImage(w, h, d));
}


//
// x_InitTexObj()
// test whether the system supports texture objects
//
void CGlTexture::x_InitTexObj(void)
{
    // test whether the system supports texture objects
    // the requirements are:
    //  - extensions advertize GL_EXT_texture_object
    //    -- or --
    //  - version >= 1.1
    switch (sm_TexObjFlag) {
    case eNotChecked:
        {{
             char *extens = (char *)glGetString (GL_EXTENSIONS);
             char *version = (char *)glGetString (GL_VERSION);
             if (extens  &&  version) {
                 if (strstr(extens, "GL_EXT_texture_object")) {
                     sm_TexObjFlag = eTexObj;
                     _TRACE("texture support = texture object");
                 } else {
                     string         ver(version);
                     vector<string> vvect;
                     NStr::Tokenize(ver, ". ", vvect);

                     if (vvect.size() > 1) {
                         try {
                             int major_ver  = NStr::StringToInt(vvect[0]);                             
                             int sub_ver    = NStr::StringToInt(vvect[1]);
                             if (major_ver >= 2 || (major_ver == 1 && sub_ver >= 1)) {
                                 sm_TexObjFlag = eTexObj;
                                 _TRACE("texture support = texture object");
                             }
                         }
                         catch (CException &) {
                             _TRACE("Unable to determine version and subversion of OpenGL (" + ver + ")");
                         }
                     } else {
                         sm_TexObjFlag = eTexDisplayList;
                         _TRACE("texture support = display list");
                     }
                 }
             } else {
                 _TRACE("texture support = not determinable");
             }
         }}
        break;

    default:
        break;
    }
}


//
// clear the current image
//
void CGlTexture::Clear(void)
{
    Unload();
    m_Image.Reset();
}


//
// "swallow" an image
//
void CGlTexture::Swallow(CImage *image)
{
    // clear the current image
    Clear();

    // assign the wrapped image
    m_Image.Reset(image);

    // load the texture
    Load();
}


bool CGlTexture::IsValid(void) const
{
    return glIsTexture(m_TexId) ? true : false;
}


//
// bind the image as a texture
//
void CGlTexture::MakeCurrent(void)
{
    if (sm_TexObjFlag == eNotChecked) {
        x_InitTexObj();
        if (sm_TexObjFlag == eNotChecked) {
            LOG_POST(Error << "CGlTexture::MakeCurrent(): "
                     "Failed to determine texture support");
            return;
        }
    }

    // texture is still valid?
    if (!IsValid()) {
        Load();
    }

    // Texture environment is not saved in the texture object, so we set it
    // here
    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, m_TexEnv);

    switch (sm_TexObjFlag) {
    case eTexObj:
        glBindTexture(GL_TEXTURE_2D, m_TexId);
        break;

    case eTexDisplayList:
        glCallList(m_TexId);
        break;

    default:
        /// this should not happen
        _ASSERT(false);
    }
}

//
// set (or reset) the opengl texture parameters that are saved in
// the texture object
// 
void CGlTexture::SetParams()
{
    if (!IsValid()) 
        return;

    MakeCurrent();
    glPixelStorei  (GL_UNPACK_ALIGNMENT, 1);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, m_WrapS);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, m_WrapT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, m_FilterMag);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, m_FilterMin);
}


//
// create a representation of the texture for rendering
// this is either a display list or a texture object
//
void CGlTexture::Load(void)
{
    //_TRACE("CGlTexture::Load()");
    // unload the current texture
    Unload();

    // safety first!
    if ( !m_Image ) {
        _TRACE("  no image to load!");
        return;
    }

    // make sure we know what to do with textures
    x_InitTexObj();

    switch (sm_TexObjFlag) {
    case eNotChecked:
        NCBI_THROW(COpenGLException, eTextureError,
                   "Texturing not supported");

    case eTexObj:
        glGenTextures(1, &m_TexId);
        glBindTexture(GL_TEXTURE_2D, m_TexId);
        _TRACE("created texture object = " << m_TexId);
        break;

    case eTexDisplayList:
        {{
             GLint val;
             glGetIntegerv(GL_MAX_TEXTURE_SIZE, &val);
             if ( m_Image->GetWidth() > (size_t)val) {
                 NCBI_THROW(COpenGLException, eTextureError,
                            "Texture is too large");
             }
         }}

        // all tests passed, allocate a display list
        m_TexId = glGenLists(1);
        glNewList(m_TexId, GL_COMPILE_AND_EXECUTE);
        //_TRACE("created texture display list = " << m_TexId);
        break;
    }

    // set texture params
    SetParams();

    // call glTexImage2D
    switch (m_Image->GetDepth()) {
    case 1:
        gluBuild2DMipmaps(GL_TEXTURE_2D, GL_ALPHA,
                          (GLint)m_Image->GetWidth(), (GLint)m_Image->GetHeight(),
                          GL_ALPHA, GL_UNSIGNED_BYTE,
                          m_Image->GetData());
        break;

    case 3:
        gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGB,
                          (GLint)m_Image->GetWidth(), (GLint)m_Image->GetHeight(),
                          GL_RGB, GL_UNSIGNED_BYTE,
                          m_Image->GetData());
        break;

    case 4:
        gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGBA,
                          (GLint)m_Image->GetWidth(), (GLint)m_Image->GetHeight(),
                          GL_RGBA, GL_UNSIGNED_BYTE,
                          m_Image->GetData());
        break;

    default:
        LOG_POST(Error << "CGlTexture::Load(): unhandled image depth");
        break;
    }

    // don't forget to end our display list
    if (sm_TexObjFlag == 0) {
        glEndList();
    }
}


//
// unload the image as a texture
void CGlTexture::Unload (void)
{
    // check to see if we can use texture objects
    x_InitTexObj();

    switch (sm_TexObjFlag) {
    case eNotChecked:
        break;

    case eTexObj:
        if (glIsTexture (m_TexId)) {
            glDeleteTextures (1, &m_TexId);
        }
        break;

    case eTexDisplayList:
        if (glIsList (m_TexId)) {
            glDeleteLists (m_TexId, 1);
        }
        break;
    }

    m_TexId = 0;
}


END_NCBI_SCOPE
