/*  $Id: gltexturefont.cpp 27642 2013-03-18 18:40:53Z falkrb $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Bob Falk
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>

#include <gui/opengl/gltexturefont.hpp>
#include <gui/opengl/ftglfontmanager.hpp>
#include <gui/opengl/glutils.hpp>

#include <corelib/ncbifile.hpp>

#include <math.h>
#include <stdio.h>
#include <sstream>

#include <FTGL/ftgl.h>


BEGIN_NCBI_SCOPE


string CGlTextureFont::s_FontFileNames[] = {
        "NimbusSanL-Regu.ttf",         // Helvetica
        "NimbusSanL-Bold.ttf",
        "NimbusSanL-ReguItal.ttf",
        "NimbusSanL-BoldItal.ttf",
        "LucidaBrightRegular.ttf",        // Lucida
        "LucidaBrightDemiBold.ttf",
        "LucidaBrightItalic.ttf",
        "LucidaBrightDemiItalic.ttf",
        "NimbusMonL-Regu.ttf",         //Courier
        "NimbusMonL-Bold.ttf",
        "NimbusMonL-ReguObli.ttf",
        "NimbusMonL-BoldObli.ttf",
        "dark-courier.ttf",         //Fixed
        "dark-courier-bold.ttf",
        "dark-courier-italic.ttf",
        "dark-courier-bold-italic.ttf",
        "NimbusRomanNo9-Reg.ttf",        // Times-Roman
        "NimbusRomanNo9-Med.ttf",
        "NimbusRomanNo9-Ita.ttf",
        "NimbusRomanNo9-MedIta.ttf"
};

string CGlTextureFont::s_FontNames[] = {
        "Helvetica",         // Helvetica
        "Helvetica Bold",
        "Helvetica Oblique",
        "Helvetica Bold Oblique",
        "Lucida",        // Lucida
        "Lucida Bold",
        "Lucida Italic",
        "Lucida Bold Italic",
        "Courier",         //Courier
        "Courier Bold",
        "Courier Oblique",
        "Courier Bold Oblique",
        "Fixed",         //Fixed
        "Fixed Bold",
        "Fixed Oblique",
        "Fixed Bold Oblique",
        "Times-Roman",        // Times-Roman
        "Times-Roman Bold",
        "Times-Roman Italic",
        "Times-Roman Bold Italic"
};

int CGlTextureFont::s_FontSizes[] = {
6, 7, 8, 9, 10, 11, 12, 14, 16, 18, 21, 24, 36, 48, 60, 72, 0
};


CGlTextureFont::CGlTextureFont()
: m_Font(NULL)
, m_FontFace(eFontFace_LastFont)
, m_FontSize(12)
, m_Rotate(fFontRotateBase)
, m_SnapToPixelX(true)
, m_SnapToPixelY(true)
{
    SetFontFace(eFontFace_Helvetica);
}

CGlTextureFont::CGlTextureFont(const string& font_file_name, 
                               unsigned int font_size)
: m_Font(NULL)
, m_FontFace(eFontFace_LastFont)
, m_FontSize(font_size)
, m_Rotate(fFontRotateBase)
, m_SnapToPixelX(true)
, m_SnapToPixelY(true)
{
    SetFont(font_file_name, font_size);
}

CGlTextureFont::CGlTextureFont(EFontFace face, 
                               unsigned int font_size)
: m_Font(NULL)
, m_FontFace(face)
, m_FontSize(font_size)
, m_Rotate(fFontRotateBase)
, m_SnapToPixelX(true)
, m_SnapToPixelY(true)
{
    SetFontFace(face);
}

void CGlTextureFont::SetFontFace(EFontFace face)
{
    string font_file = s_FontFileNames[(int)face];  
    m_FontFace = face;

    if (font_file != m_FontFile)
        SetFont(font_file, m_FontSize);
}

CGlTextureFont::EFontFace CGlTextureFont::GetFontFace() const
{
    return m_FontFace;
}

void CGlTextureFont::SetFontSize(unsigned int size)
{
    m_FontSize = size;

    if (m_FontFile != "")
        SetFont(m_FontFile, size);
}

unsigned int CGlTextureFont::GetFontSize() const
{
    return m_FontSize;
}


bool CGlTextureFont::SetFont(const string& font_file_name, 
                             unsigned int font_size)
{
    CFtglFontManager& fm = CFtglFontManager::Instance();
    FTFont* font = fm.GetFont(font_file_name.c_str(), font_size);

    /// Error loading - don't change current font selection
    if (font==NULL || font->Error())
        return false;
   
    m_FontFile = font_file_name;
    m_Font = font;
    m_FontSize = font_size;	

    return true;
}

void CGlTextureFont::SetSnapToPixel(bool xpix, bool ypix)
{
    m_SnapToPixelX = xpix;
    m_SnapToPixelY = ypix;
}

void CGlTextureFont::BeginText() const
{
    /// Get viewport from OpenGL
	glGetIntegerv(GL_VIEWPORT, m_Viewport);
    BeginText(TVPRect(m_Viewport[0], 
                      m_Viewport[1], 
                      m_Viewport[2], 
                      m_Viewport[3]));
}

void CGlTextureFont::BeginText(const TVPRect& viewport) const
{
    // avoid pusattrib if possible since it can be very slow
    //glPushAttrib(GL_ALL_ATTRIB_BITS);

    // Copy viewport passed in (we can't get it from OpenGL always because
    // this library also supports some pdf rendering functions and in pdf 
    // output the maximum viewing area generally exceeds the maximum 
    // supported by OpenGL)
    m_Viewport[0] = GLint(viewport.Left());
    m_Viewport[1] = GLint(viewport.Bottom());
    m_Viewport[2] = GLint(viewport.Right());
    m_Viewport[3] = GLint(viewport.Top());
    
    // Get modelview and projection matrices that will be used for 
    // unprojection (getting window coordinates) of text draw location	
    glGetDoublev(GL_MODELVIEW_MATRIX, m_ModelviewMatrix);
    glGetDoublev(GL_PROJECTION_MATRIX, m_ProjectionMatrix);

    // Set projection matrix to pixel coordinates
    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glLoadIdentity();
    glOrtho(m_Viewport[0], 
            m_Viewport[0] + m_Viewport[2],
            m_Viewport[1], 
            m_Viewport[1] + m_Viewport[3], -1.0, 1.0);
    glMatrixMode(GL_MODELVIEW);

    // set up the ogl state for text
    glDisable(GL_LIGHTING);
    glDisable(GL_DEPTH_TEST);
    glDisable(GL_CULL_FACE);
    glEnable(GL_TEXTURE_2D);
    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);  
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    CGlUtils::CheckGlError();
}
    
void CGlTextureFont::BeginText(const TVPRect& viewport, 
                               GLdouble* mview, GLdouble* proj ) const
{
    // avoid pusattrib if possible since it can be very slow
    //glPushAttrib(GL_ENABLE_BIT | GL_TRANSFORM_BIT);

    // Copy viewport passed in (we can't get it from OpenGL always because
    // this library also supports some pdf rendering functions and in pdf 
    // output the maximum viewing area generally exceeds the maximum 
    // supported by OpenGL)
    m_Viewport[0] = GLint(viewport.Left());
    m_Viewport[1] = GLint(viewport.Bottom());
    m_Viewport[2] = GLint(viewport.Right());
    m_Viewport[3] = GLint(viewport.Top());
    
    // Get modelview and projection matrices that will be used for 
    // unprojection (getting window coordinates) of text draw location	
    memcpy(m_ModelviewMatrix, mview, sizeof(GLdouble)*16);
    memcpy(m_ProjectionMatrix, proj, sizeof(GLdouble)*16);


    //glGetDoublev(GL_MODELVIEW_MATRIX, m_ModelviewMatrix);
    //glGetDoublev(GL_PROJECTION_MATRIX, m_ProjectionMatrix);
    //glGetIntegerv(GL_VIEWPORT, m_Viewport);

    // Set projection matrix to pixel coordinates
    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glLoadIdentity();
    glOrtho(m_Viewport[0], 
            m_Viewport[0] + m_Viewport[2],
            m_Viewport[1], 
            m_Viewport[1] + m_Viewport[3], -1.0, 1.0);
    glMatrixMode(GL_MODELVIEW);

    // set up the ogl state for text
    glDisable(GL_LIGHTING);
    glDisable(GL_DEPTH_TEST);
    glDisable(GL_CULL_FACE);
    glEnable(GL_TEXTURE_2D);
    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);  
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
}

void CGlTextureFont::EndText() const
{   
    // restore OpenGL state (push/pop all is expensive)
    //glPopAttrib();

    glMatrixMode(GL_PROJECTION);
    glPopMatrix();   
    glMatrixMode(GL_MODELVIEW);

    // set most reasonable options (rather than push/pop attrib)
    glDisable(GL_TEXTURE_2D);
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA); 
    glListBase(0);
}
 
void CGlTextureFont::WriteText(TModelUnit x, 
                               TModelUnit y, 
                               const char* text,
                               TModelUnit rotate_degrees) const
{
    // If rotation is between 90 and 270, the text will be backwards and upside down
    // unless we re-orient.  Check if user provided reorient flag.
    bool reorient_text = false;    
    if (rotate_degrees != TModelUnit(0)) {        
        TModelUnit rads = rotate_degrees*(TModelUnit(3.141592653589732)/TModelUnit(180));
        bool backwards = cos(double(rads)) < TModelUnit(0);
        if ((m_Rotate & fReorientText) && backwards) {
            reorient_text = true;             
        }
    } 

    // Get the window (projection) coordinates of the drawing position (x,y)
    GLdouble wx, wy, wz;
    gluProject(x, y, 0.0, 
               m_ModelviewMatrix, m_ProjectionMatrix, m_Viewport, 
               &wx, &wy, &wz);

	glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glLoadIdentity();

    // Round the drawing coordinates to nearest integer pixel coordinate
    // (drawing textured fonts on pixel boundries improves visual appearance)
    // unless user has turned option off
    double xi = wx;
    double yi = wy;

    if (m_SnapToPixelX)
        xi = (double)((int)(wx + 0.5));  
    if (m_SnapToPixelY)
        yi = (double)((int)(wy + 0.5));  

    TModelUnit rcenterx = (TModelUnit)0.0;
    TModelUnit rcentery = (TModelUnit)0.0;

    if (m_Rotate & fFontRotateMid) {
        rcentery = TextHeight()/(TModelUnit)2.0;
    }
    else if (m_Rotate & fFontRotateCap) {
        rcentery = TextHeight();
    }
    // fFontRotateBase is default

    glTranslated(xi, yi, 0.0);

    // rotate if desired
    if (rotate_degrees != TModelUnit(0.0)) {
        glTranslated((GLdouble)rcenterx, (GLdouble)rcentery, 0.0);
        glRotated(rotate_degrees, 0.0, 0.0, 1.0);
        glTranslated((GLdouble)-rcenterx, (GLdouble)-rcentery, 0.0);
        
        // If we have rotated text to be upside-down and backwards, flip it
        // across the x and y axes
        if (reorient_text) {
            glTranslatef(0.0f, float(TextHeight()), 0.0f);
            glScalef(1.0f, -1.0f, 1.0f);
            glTranslatef(float(TextWidth(text)), 0.0f, 0.0f);
            glScalef(-1.0f, 1.0f, 1.0f);
        }
    }

    TextOut(text);

    glMatrixMode(GL_MODELVIEW);
    glPopMatrix();
}

void CGlTextureFont::WriteText(TModelUnit x, TModelUnit y, 
                               TModelUnit width, TModelUnit height,
                               const char* text,
                               TAlign align,
                               ETruncate trunc,
                               TModelUnit rotate_degrees) const
{
    string output_str(text);
    x_WriteText(x, y, width, height, output_str, align, trunc, rotate_degrees, true);
}

CMatrix4<double> CGlTextureFont::GetTextXform(TModelUnit x, TModelUnit y, 
                                              TModelUnit width, TModelUnit height,
                                              string& text,
                                              TAlign align,
                                              ETruncate trunc,
                                              TModelUnit rotate_degrees) const
{
    return x_WriteText(x, y, width, height, text, align, trunc, rotate_degrees, false);
}

CMatrix4<double> CGlTextureFont::x_WriteText(TModelUnit x, TModelUnit y, 
                                             TModelUnit width, TModelUnit height,
                                             string& text,
                                             TAlign align,
                                             ETruncate trunc,
                                             TModelUnit rotate_degrees,
                                             bool write_text) const
{
    TModelUnit w = width;
    TModelUnit h = height;

    // If rotation is between 90 and 270, the text will be backwards and upside down
    // unless we re-orient.  Check if user provided reorient flag.
    bool reorient_text = false;    
    if (rotate_degrees != TModelUnit(0)) {        
        TModelUnit rads = rotate_degrees*(TModelUnit(3.141592653589732)/TModelUnit(180));
        bool backwards = cos(double(rads)) < TModelUnit(0);
        if ((m_Rotate & fReorientText) && backwards) {
            reorient_text = true;             
        }
    } 

    /// truncate as appropriate
    string str;
    switch (trunc) {
    case eTruncate_None:
        str = text;
        break;

    case eTruncate_Empty:
    case eTruncate_Ellipsis:
        {    
            if (reorient_text) {
                string rtext = text;

                // Reverse text so that it is truncated on the 
                // correct side 
                std::reverse(rtext.begin(), rtext.end());
                x_Truncate(rtext.c_str(), w, trunc, &str);
                std::reverse(str.begin(), str.end());
            }
            else {
                x_Truncate(text.c_str(), w, trunc, &str);
            }
        }
        break;
    }

    // update parameter to return (possibly truncated) text
    text = str;

    ///
    /// determine where our text starts
    ///
    TModelUnit tx = 0;
    TModelUnit ty = 0;

    // scale is the ratio of the visible width/height divided by the 
    // viewport width/height (same as in CGLPane).  Assumes of course
    // user is using orthographic projection matrix (e.g. glOrtho())
    TModelUnit scale_x = (TModelUnit(2)/TModelUnit(m_ProjectionMatrix[0]))/
        TModelUnit(m_Viewport[2]);
    TModelUnit scale_y = (TModelUnit(2)/TModelUnit(m_ProjectionMatrix[5]))/
        TModelUnit(m_Viewport[3]);

    // eAlign_Right or eAlign_Center are applied to horz position only if
    // eAlign_Left is not specifed
    if ((align & eAlign_HorizMask) == eAlign_Left)  {
        tx = x;
    } else if ((align & eAlign_HorizMask) == eAlign_Right)  {
        // right justify
        TModelUnit text_wid = TextWidth(str.c_str())*scale_x;
        tx = x + (w - text_wid);
    } else  { // center
        TModelUnit text_wid = TextWidth(str.c_str())*scale_x;
        tx = x + (w - text_wid) * TModelUnit(0.5);
    }

    if ((align & eAlign_VertMask) == eAlign_Bottom)  {
        ty = y;
    } else if ((align & eAlign_VertMask) == eAlign_Top) {
        // skip down enough to keep our text inside
        TModelUnit text_ht = TextHeight()*scale_y;
        ty = y + (h - text_ht);
    } else {
        // if nothing given or eAlign_Center is specified - center
        // vertically
        TModelUnit text_ht = TextHeight()*scale_y;
        ty = y + (h - text_ht) * TModelUnit(0.5);
    }  

    //
    // Set the matrices to pixel-coordinates and then render
    //

    // Get window projection coordnates for (x,y)
    GLdouble wx, wy, wz;
    gluProject(tx, ty, 0.0, 
               m_ModelviewMatrix, m_ProjectionMatrix, m_Viewport, 
               &wx, &wy, &wz);

	glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glLoadIdentity();

    // Round the drawing coordinates to nearest integer pixel coordinate
    // (drawing textured fonts on pixel boundries improves visual appearance)
    // unless user has turned option off
    double xi = wx;
    double yi = wy;

    if (m_SnapToPixelX)
        xi = (double)((int)(wx + 0.5));  
    if (m_SnapToPixelY)
        yi = (double)((int)(wy + 0.5));  

    // If a rotation angle is specified, we want to rotate from the base
    // of the box, not the text.  So rotate the text around the bottom
    // middle, or top of the base side of the box as appropriate.
    if (rotate_degrees != TModelUnit(0.0)) {
        TModelUnit rcenterx = (TModelUnit)0.0;
        TModelUnit rcentery = (TModelUnit)0.0;

        // Get screen coordinates to which (x,y) (box base position) projects
        GLdouble xproj, yproj, zproj;
        gluProject(x, y, 0.0, 
                   m_ModelviewMatrix, m_ProjectionMatrix, m_Viewport, 
                   &xproj, &yproj, &zproj);    

        // Get offset of box needed for rotation center
        if (m_Rotate & fFontRotateMid) {
            rcentery = height/(TModelUnit)2.0;
        }
        else if (m_Rotate & fFontRotateCap) {
            rcentery = height;
        }       

        double xoff = xi-xproj;
        double yoff = yi-yproj;

        glTranslated(xi, yi, 0.0);
        glTranslated((GLdouble)rcenterx-xoff, (GLdouble)rcentery-yoff, 0.0);
        glRotated(rotate_degrees, 0.0, 0.0, 1.0);
        glTranslated((GLdouble)-rcenterx+xoff, (GLdouble)-rcentery+yoff, 0.0);

        // If we have rotated text to be upside-down and backwards, flip it
        // across the x and y axes
        if (reorient_text) {
            glTranslatef(0.0f, float(TextHeight()), 0.0f);
            glScalef(1.0f, -1.0f, 1.0f);
            glTranslatef(float(TextWidth(str.c_str())), 0.0f, 0.0f);
            glScalef(-1.0f, 1.0f, 1.0f);
        }
    }
    else {
        glTranslated(xi, yi, 0.0);
    }

    CMatrix4<double> xform;
    xform.Identity();

    // If we write text matrix will be invalid (only do one or the other
    // since both can slow down performance)
    if (write_text) {
        TextOut(str.c_str());
    }
    else {    
        glGetDoublev(GL_MODELVIEW_MATRIX, xform.GetData());
        xform.Transpose();
    }

    glMatrixMode(GL_MODELVIEW);
    glPopMatrix();

    return xform;

    ///
    /// Debugging code to draw the bounding box for the text
    ///
    /*
	glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glLoadIdentity();

    double wx2, wy2, wz2;
    gluProject(x+width, y+height, 0.0, 
               m_ModelviewMatrix, m_ProjectionMatrix, m_Viewport, 
               &wx2, &wy2, &wz2);

    glTranslated((GLdouble)xproj+rcenterx, (GLdouble)yproj+rcentery, 0.0);
    glRotated(rotate_degrees, 0.0, 0.0, 1.0);
    glTranslated((GLdouble)-(xproj+rcenterx), (GLdouble)-(yproj+rcentery), 0.0);

    glDisable(GL_TEXTURE_2D);
    glDisable(GL_BLEND);
    glColor3f(0.0f, 0.0f, 1.0f);
    glBegin(GL_LINE_LOOP);
        glVertex3d(xproj, yproj, 0.0f);
        glVertex3d(wx2, yproj, 0.0f);
        glVertex3d(wx2, wy2, 0.0f);
        glVertex3d(xproj, wy2, 0.0f);
    glEnd();
    glEnable(GL_TEXTURE_2D);
    glEnable(GL_BLEND);
	glMatrixMode(GL_MODELVIEW);
    glPopMatrix();
    */
}

void CGlTextureFont::TextOut(const char* text) const
{
    if (m_Font != NULL && !m_Font->Error())
        m_Font->Render(text);
}

void CGlTextureFont::TextOut(TModelUnit x, 
                             TModelUnit y, 
                             const char* text) const
{
    BeginText();
    WriteText(x, y, text);
    EndText();
}

void CGlTextureFont::TextOut(TModelUnit x1, TModelUnit y1, 
                             TModelUnit x2, TModelUnit y2, 
                             const char* text,
                             TAlign align,
                             ETruncate trunc,
                             TModelUnit scale_x,
                             TModelUnit scale_y) const
{
    // Handle alignment and truncation and scaling (scaling seems be used by the ruler class - maybe not much else
    if(x2 < x1) {
        swap(x1, x2);
    }
    if(y2 < y1) {
        swap(y1, y2);
    }

    TModelUnit x = x1;
    TModelUnit y = y1;
    TModelUnit w = x2 - x1;
    TModelUnit h = y2 - y1;

    /// truncate as appropriate
    string str;
    switch (trunc) {
    case eTruncate_None:
        str = text;
        break;

    case eTruncate_Empty:
    case eTruncate_Ellipsis:
        {
            x_Truncate(text, w, trunc, &str);
        }
        break;
    }

    BeginText();

    ///
    /// determine where our text starts
    ///
    TModelUnit tx = 0;
    TModelUnit ty = 0;

    // scale is the ratio of the visible width/height divided by the 
    // viewport width/height (same as in CGLPane).  Assumes of course
    // user is using orthographic projection matrix (e.g. glOrtho())
    scale_x = (TModelUnit(2)/TModelUnit(m_ProjectionMatrix[0]))/
        TModelUnit(m_Viewport[2]);
    scale_y = (TModelUnit(2)/TModelUnit(m_ProjectionMatrix[5]))/
        TModelUnit(m_Viewport[3]);

    // eAlign_Right or eAlign_Center are applied to horz position only if
    // eAlign_Left is not specifed
    if ((align & eAlign_HorizMask) == eAlign_Left)  {
        tx = x;
    } else if ((align & eAlign_HorizMask) == eAlign_Right)  {
        // right justify
        TModelUnit text_wid = TextWidth(str.c_str()) * scale_x;
        tx = x + (w - text_wid);
    } else  { // center
        TModelUnit text_wid = TextWidth(str.c_str()) * scale_x;
        tx = x + (w - text_wid) * TModelUnit(0.5);
    }

    if ((align & eAlign_VertMask) == eAlign_Bottom)  {
        ty = y;
    } else if ((align & eAlign_VertMask) == eAlign_Top) {
        // skip down enough to keep our text inside
        TModelUnit text_ht = TextHeight() * scale_y;
        ty = y + (h - text_ht);
    } else {
        // if nothing given or eAlign_Center is specified - center
        // vertically
        ty = (y1 + y2 - GetMetric(eMetric_CharHeight))/TModelUnit(2.0);
    }  
  
    WriteText(x, y, text);
    EndText();
}

string CGlTextureFont::Truncate(const char* text, 
                                TModelUnit w, 
                                ETruncate trunc) const
{
    string res;
    x_Truncate(text, w, trunc, &res);
    return res;
}

string CGlTextureFont::Truncate(const string& text, 
                                TModelUnit w, 
                                ETruncate trunc) const
{
    string res;
    x_Truncate(text.c_str(), w, trunc, &res);
    return res;
}

TModelUnit CGlTextureFont::TextWidth(const char* text) const
{
    if (m_Font != NULL && !m_Font->Error())
        return (TModelUnit)m_Font->Advance(text);
    
    return TModelUnit(0.0);
}

TModelUnit CGlTextureFont::TextWidth(const char* text, int n) const
{
    if (m_Font != NULL && !m_Font->Error())
        return (TModelUnit)m_Font->Advance(text, n);
    
    return TModelUnit(0.0);
}

TModelUnit CGlTextureFont::TextHeight(void) const
{
    // This returns the height of the fonts without descenders - so it's a little
    // less that the full height.  This does not use any OpenGL (so it can be called
    // before window creation unlike some of the functions in GetMetric())
    if (m_Font != NULL && !m_Font->Error()) {
        TModelUnit fs = (TModelUnit)m_Font->FaceSize();
        // descender is given as a negative number
        TModelUnit desc = (TModelUnit)m_Font->Descender();
        return fs + desc;
    }

    return TModelUnit(m_FontSize);
}

TModelUnit CGlTextureFont::GetFontDescender() const
{
    if (m_Font != NULL && !m_Font->Error())
        return (TModelUnit)m_Font->Descender();

    return TModelUnit(0.0);
}

TModelUnit CGlTextureFont::GetAdvance(char c) const
{
    if (m_Font != NULL && !m_Font->Error())
        return m_Font->Advance(&c, 1);

    return TModelUnit(0.0f);
} 

 TModelUnit CGlTextureFont::GetMaxWidth(int max_num) const
{
    double max_dig_w = 0;

    for( char c = '0'; c <= '9'; c++) {
        double char_w = GetAdvance(c);
        max_dig_w = max(max_dig_w, char_w);
    }
    double comma_w = GetAdvance(',');

    double mod = abs(max_num);
    int digits_count = (int) ceil(log10(mod));
    int commas_count = (digits_count - 1) / 3;

    // length of the longest possible label in pixels (add 2 for separation)
    double w = digits_count * max_dig_w + commas_count * comma_w;
    if(max_num < 0) {
        w += GetAdvance('-');
    }
    return w; 
}

TModelUnit CGlTextureFont::GetMetric(EMetric metric, 
                                     const char* text,
                                     int len) const
{
    FTBBox bounding;


    // Standard characters (could also iterate over ascii codes).
    static const char* sc_chars = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz1234567890`~!@#$%^&*()_-+={}|[]\\:\";'<,>.?/";
    // just the caps
    static const char* sc_caps = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";

    float w = 0.0f;
    float maxw = 0.0f;

    // Note -some of these use OpenGL and should not be called before window creation
    if (m_Font != NULL && !m_Font->Error()) {
        switch(metric) {
            case eMetric_CharHeight: 
                bounding = m_Font->BBox(sc_caps);               
                return bounding.Upper().Y() - bounding.Lower().Y();                
                break;
            case eMetric_FullCharHeight:    
                return (TModelUnit)m_Font->FaceSize();
                break;
            case eMetric_AvgCharWidth:          
                w = m_Font->Advance(sc_chars);
                return (TModelUnit(w)/TModelUnit(strlen(sc_chars)));                
                break;
            case eMetric_MaxCharWidth:
                maxw = 0.0f;
                for (size_t i=0; i<strlen(sc_chars); ++i) {
                    w = m_Font->Advance(&sc_chars[i], 1);
                    if (w > maxw)
                        maxw = w;
                }
                return (TModelUnit)maxw;
                break;
                // I'm not sure why we have two of these... or if they are 
                // really used.
            case eMetric_TextWidth:
            case eMetric_FullTextWidth:
                if (text != NULL)
                    return (TModelUnit)m_Font->Advance(text);
                break;
            case eMetric_Descender:
                // This should return the max descender for the current string
                // but there is not a function to do that so just return the
                // descender for this font.
                return (TModelUnit)m_Font->Descender();
                break;
        }
    }

    return TModelUnit(0.0);
}

string CGlTextureFont::FaceToString(EFontFace face)
{
    return s_FontNames[face];
}

CGlTextureFont::EFontFace CGlTextureFont::FaceFromString(const string& str)
{
    vector<string> arr;
    NStr::Tokenize(str, " \t", arr);

    // If all tokens (such as Courier, bold, oblique occur in
    // a string return the enum for that string, otherwise return 
    // eFontFace_LastFont
    for (size_t i=0; i<(size_t)eFontFace_LastFont; ++i) {
        //string str = s_FontNames[i];
        bool match = true;

        for (size_t j=0; j<arr.size(); ++j) {
            size_t pos = NStr::Find(s_FontNames[i], arr[j], 0, NPOS, NStr::eFirst, NStr::eNocase);
            if (pos == NPOS)
                match = false;
        }

        if (match) 
            return (EFontFace)i;
    }

    return eFontFace_LastFont;
}

string CGlTextureFont::PdfBaseFontName(EFontFace face)
{
    // These are all the base-font names defined by PDF (these fonts can be
    // embedded into pdf files, but do not have to be).
    /*
        Times-Roman
        Times-Bold
        Times-Italic
        Times-BoldItalic
        Helvetica
        Helvetica-Bold
        Helvetica-Oblique
        Helvetica-BoldOblique
        Courier
        Courier-Bold
        Courier-Oblique
        Courier-BoldOblique
        Symbol
        ZapfDingbats
    */

    // This is the best mapping to built in pdf font names.  Some are only
    // approximations.  PDF does also allow fonts to be embedded, if the pdf
    // rendering were to support it.
    switch (face) {
        case eFontFace_Helvetica:
            return "Helvetica";
        case eFontFace_Helvetica_Bold:
            return "Helvetica-Bold";
        case eFontFace_Helvetica_Italic:
            return "Helvetica-Oblique";
        case eFontFace_Helvetica_BoldItalic:
            return "Helvetica-BoldOblique";
        case eFontFace_Lucida:
            return "Times-Roman";
        case eFontFace_Lucida_Bold:
            return "Times-Bold";
        case eFontFace_Lucida_Italic:
            return "Times-Italic";
        case eFontFace_Lucida_BoldItalic:
            return "Times-BoldItalic";
        case eFontFace_Courier:
            return "Courier";
        case eFontFace_Courier_Bold:
            return "Courier-Bold";
        case eFontFace_Courier_Italic:
            return "Courier-Oblique";
        case eFontFace_Courier_BoldItalic:
            return "Courier-BoldOblique";
        case eFontFace_Fixed:
            return "Courier";
        case eFontFace_Fixed_Bold:
            return "Courier-Bold";
        case eFontFace_Fixed_Italic:
            return "Courier-Oblique";
        case eFontFace_Fixed_BoldItalic:
            return "Courier-BoldOblique";
        case eFontFace_TimesRoman:
            return "Times-Roman";
        case eFontFace_TimesRoman_Bold:
            return "Times-Bold";
        case eFontFace_TimesRoman_Italic:
            return "Times-Italic";
        case eFontFace_TimesRoman_BoldItalic:
            return "Times-BoldItalic";
        default:
            return "";
    }

    return "";
}

// saves and restores font face and size in(from) a string
string CGlTextureFont::ToString() const
{
    string size_str = NStr::NumericToString(m_FontSize);
    string font_string = m_FontFile + " " + size_str;

    return font_string;
}
void CGlTextureFont::FromString(const string& value)
{
    istringstream sstr(value);

    sstr >> m_FontFile >> m_FontSize;

    if (m_FontSize >= 4 && m_FontFile != "") {
        SetFont(m_FontFile, m_FontSize);

        m_FontFace = eFontFace_LastFont;

        /// Set face name if we recognize the font file:
        for (size_t i=0; i<(size_t)eFontFace_LastFont; ++i) {
            if (s_FontFileNames[i] == m_FontFile) {
                m_FontFace = (EFontFace)i;
                break;
            }
        }
    }
}
    
void CGlTextureFont::GetAllFaces(vector<string>& faces)
{
    faces.clear();

    for (size_t i=0; i<(size_t)eFontFace_LastFont; ++i) {
        faces.push_back(s_FontNames[i]);
    }
}
void CGlTextureFont::GetAllSizes(vector<string>& sizes)
{
    sizes.clear();

    size_t idx = 0;
    while (s_FontSizes[idx] != 0) {
        sizes.push_back(NStr::NumericToString(s_FontSizes[idx++]));
    }
}


int CGlTextureFont::x_Truncate(const char* text, 
                               TModelUnit w, 
                               ETruncate trunc,
                               string* str) const
{
    if (m_Font == NULL || m_Font->Error())
        return 0;

    static const char* sc_ellipsis = "...";
    const char* ellip = NULL;

    const char* end_pos = text;
    switch (trunc) {
    case eTruncate_None:
        end_pos = text + strlen(text);
        break;

    case eTruncate_Ellipsis:    {
            // first, do a blunt cut
            int pos = x_Truncate(text, w, eTruncate_Empty);
            end_pos = text + pos;
            if ( !*end_pos ) {
                // we can fit the whole thing
                break;
            }

            // adjust for the ellipsis
            TModelUnit e_wid = TextWidth(sc_ellipsis);
            while (e_wid > 0) {
                ellip = sc_ellipsis;
                FTBBox b = m_Font->BBox(end_pos, 1);
                e_wid -= TModelUnit(b.Upper().X() - b.Lower().X());
                if (end_pos > text) {
                    --end_pos;
                } else {
                    break;
                }
            }
        }
        break;

    case eTruncate_Empty:
        for ( ;  end_pos  &&  *end_pos  && w >= 0;  ++end_pos) {
            TModelUnit wid = (TModelUnit)m_Font->Advance(end_pos, 1);
            if (w - wid < 0) {
                FTBBox b = m_Font->BBox(end_pos, 1);
                wid = TModelUnit(b.Upper().X() - b.Lower().X());
            }

            w -= wid;
        }

        if (w < 0) {
            --end_pos;
        }
        break;
    }

    if (end_pos < text) {
        end_pos = text;
    }

    if (str) {
        str->assign(text, end_pos);
        if (ellip) {
            *str += ellip;
        }
    }
    return (int)(end_pos - text);
}

END_NCBI_SCOPE
