/*  $Id: clustalw_tool_job.cpp 26643 2012-10-18 17:54:05Z katargir $
* ===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data,  the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties,  express or implied,  including
*  warranties of performance,  merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================
*
* Authors:  Roman Katargin
*/


#include <ncbi_pch.hpp>

#include <corelib/ncbifile.hpp>
#include <corelib/ncbiexec.hpp>
#include <algo/phy_tree/phy_node.hpp>
#include <objmgr/seq_vector.hpp>
#include <objmgr/util/sequence.hpp>
#include <objtools/readers/aln_reader.hpp>

#include <serial/iterator.hpp>

#include <objects/seqalign/Seq_align.hpp>
#include <objects/biotree/BioTreeContainer.hpp>

#include <algo/phy_tree/dist_methods.hpp>

#include <gui/objutils/label.hpp>
#include <gui/objects/AbstractProjectItem.hpp>

#include <gui/packages/pkg_alignment/clustalw_tool_job.hpp>
#include <gui/widgets/wx/wx_utils.hpp>

#include <wx/filename.h>
#include <wx/sstream.h>
#include <wx/txtstrm.h>
#include <wx/utils.h>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

///////////////////////////////////////////////////////////////////////////////
/// CClustalwToolJob
CClustalwToolJob::CClustalwToolJob(CProjectService* service,
                                 const CClustalwToolParams& params)
 : CDataLoadingAppJob(service), m_Params(params), m_PId(0), m_IsProtein(false)
{
    CFastMutexGuard lock(m_Mutex);
    m_Descr = "ClustalW Job";
}

typedef CClustalwToolJob::TLocPair TLocPair;
typedef CClustalwToolJob::TIdMap TIdMap;
struct STreeNodeMapper
{
    CClustalwToolJob::TIdMap* idmap;
    STreeNodeMapper()
        : idmap(NULL)
    {
    }

    ETreeTraverseCode operator()(TPhyTreeNode& node, int level)
    {
        _ASSERT(idmap);
        string& label = node.GetValue().SetLabel();
        if (label.empty()) {
            return eTreeTraverse;
        }
        TIdMap::iterator id_iter = idmap->find(label);
        if (id_iter == idmap->end()) {
            return eTreeTraverse;
        }

        TLocPair p = id_iter->second;
        CScope& scope = *p.second;
        const CSeq_id& real_id =
            sequence::GetId(*p.first, &scope);
        label.erase();
        CLabel::GetLabel(real_id, &label, CLabel::eDefault, &scope);
        return eTreeTraverse;
    }
};

static void s_DumpSequence(CScope& scope,
                           const CSeq_loc& loc,
                           TIdMap& idmap,
                           bool& is_prot,
                           CNcbiOstream& ostr)
{
    string id_str = "lcl|seq-" + NStr::SizetToString(idmap.size());
    idmap[id_str] = make_pair(CConstRef<CSeq_loc>(&loc), CRef<CScope>(&scope));

    ostr << ">" << id_str << endl;

    CSeqVector vec(loc, scope, CBioseq_Handle::eCoding_Iupac);
    if (vec.IsProtein()) {
        is_prot = true;
    }
    string data;
    vec.GetSeqData(0, vec.size(), data);
    ostr << data << endl;
}

bool CClustalwToolJob::BeforeRun()
{
    wxString clustalwExe = m_Params.GetClustalwPath();
    if (!wxFileName::FileExists(clustalwExe)) {
        NCBI_THROW(CException, eUnknown, "Path to ClustalW executable does not exist!");
    }

    /// we serialize this through a set of temporary files
    m_TmpIn      = wxFileName::CreateTempFileName(wxT("in"));
    m_TmpOut     = wxFileName::CreateTempFileName(wxT("out"));
    m_TmpTreeOut = wxFileName::CreateTempFileName(wxT("tre"));

    int seq_types = 0; // Check if we have a mix of DNA and protein seqs

    /// scoped to make sure we flush and close our file before we start!
    {{
        CNcbiOfstream ostr(m_TmpIn.fn_str());

        NON_CONST_ITERATE (TConstScopedObjects, iter, m_Params.m_Locations) {
            const CSeq_loc& loc = dynamic_cast<const CSeq_loc&>(iter->object.GetObject());
            CScope& scope = iter->scope.GetObject();
            bool is_prot = false;
            s_DumpSequence(scope, loc, m_IdMap, is_prot, ostr);
            seq_types |= is_prot ? 2 : 1;
        }
    }}

    if (seq_types == 3) { // we have mix, quit
        x_CleanUp();
        NCBI_THROW(CException, eUnknown, "ClustalW tool: all input sequences must be either DNA or protein.");
    }

    m_IsProtein = (seq_types == 2);

    ///
    /// compose the clustalw command line
    ///

#ifdef NCBI_OS_UNIX
    wxString kOptPrefix = wxT(" -");
#else
    wxString kOptPrefix = wxT(" /");
#endif

    wxStringOutputStream strstr;
    wxTextOutputStream os(strstr);

    os  << wxT("\"") << clustalwExe << wxT("\"")
        << kOptPrefix << wxT("align")
        << kOptPrefix << wxT("infile=") << m_TmpIn
        << kOptPrefix << wxT("outfile=") << m_TmpOut
        << kOptPrefix << wxT("type=") << (m_IsProtein ? wxT("PROTEIN") : wxT("DNA"))
        << kOptPrefix << wxT("newtree=") << m_TmpTreeOut;


    wxString moreParams = m_Params.GetCommandLine();
    moreParams.Trim(true);
    moreParams.Trim(false);
    if (!moreParams.empty())
        os << wxT(" ") << moreParams;

    wxString cmdline = strstr.GetString();
    LOG_POST(Info << "Launching ClustalW executabe:");
    LOG_POST(Info << cmdline);

    m_PId = ::wxExecute(cmdline);
    if (m_PId == 0)
        NCBI_THROW(CException, eUnknown, "Failed to launch ClustalW executable.");

    return true;
}

void CClustalwToolJob::x_CleanUp()
{
#ifndef _DEBUG
    if (!m_TmpIn.empty())
        ::wxRemoveFile(m_TmpIn);
    if (!m_TmpOut.empty())
        ::wxRemoveFile(m_TmpOut);
    if (!m_TmpTreeOut.empty())
        ::wxRemoveFile(m_TmpTreeOut);
#endif
}

void CClustalwToolJob::x_CreateProjectItems()
{
    if (m_PId == 0) {
        NCBI_THROW(CException, eUnknown, "MUSCLE not launched!");
    }

    while (true) {
        ::wxSleep(3);

        if (IsCanceled()) {
            ::wxKill(m_PId, wxSIGKILL);
            x_CleanUp();
            return;
        }

        ncbi::CProcess proc(m_PId, ncbi::CProcess::ePid);
        if (!proc.IsAlive()) {
            break;
        }
    }

    wxULongLong size = wxFileName::GetSize(m_TmpOut);
    if (size == 0 || size ==  wxInvalidSize) {
        x_CleanUp();
        NCBI_THROW(CException, eUnknown, "ClustalW executabale failed to produce results.");
    }

    string err_msg;
    try {
        CNcbiIfstream istr(m_TmpOut.fn_str());
        CAlnReader reader(istr);

        if (m_IsProtein) {
            reader.SetAlphabet(CAlnReader::eAlpha_Protein);
        } else {
            reader.SetAlphabet(CAlnReader::eAlpha_Nucleotide);
        }
        reader.SetAllGap("-");

        reader.Read();
        CRef<CSeq_align> align = reader.GetSeqAlign();

        CRef<CBioTreeContainer> btc;
        if (m_Params.GetGenerateTree() && wxFileName::FileExists(m_TmpTreeOut)) {
            try {
                CNcbiIfstream tree_istr(m_TmpTreeOut.fn_str());
                auto_ptr<TPhyTreeNode> tree(ReadNewickTree(tree_istr));

                /// we need to convert the labels to the correct seq-id strings
                STreeNodeMapper mapper;
                mapper.idmap = &m_IdMap;
                TreeDepthFirstTraverse(*tree, mapper);

                /// convert to the serializable form
                btc = MakeBioTreeContainer(tree.get());

                /// adding seq-id prop
                const int kLabelId = 0;
                const int kSeqIdId = 2;

                CRef<CFeatureDescr> feat_descr(new CFeatureDescr);
                feat_descr->SetId(kSeqIdId);
                feat_descr->SetName("seq-id");
                btc->SetFdict().Set().push_back(feat_descr);

                NON_CONST_ITERATE (CNodeSet::Tdata, node, btc->SetNodes().Set()) {
                    if ( !(*node)->CanGetFeatures() ) {
                        continue;
                    }
                    NON_CONST_ITERATE (CNodeFeatureSet::Tdata, node_feature,
                            (*node)->SetFeatures().Set()) {
                        if ((*node_feature)->GetFeatureid() == kLabelId) {
                            CRef<CNodeFeature> id_node_feature(new CNodeFeature);
                            id_node_feature->SetFeatureid(kSeqIdId);
                            id_node_feature->SetValue((*node_feature)->GetValue());
                            (*node)->SetFeatures().Set().push_back(id_node_feature);
                        }
                    }
                }
            }
            catch (CException& e) {
                LOG_POST(Error
                    << "Failed to read phylogenetic tree:\n"
                    << e.GetMsg()
                    << "\nPhylogenetic tree output will not be available.");
                btc.Reset();
                m_TmpTreeOut.erase();
            }
            catch (std::exception& e) {
                LOG_POST(Error
                    << "Failed to read phylogenetic tree:\n"
                    << e.what()
                    << "\nPhylogenetic tree output will not be available.");
                btc.Reset();
                m_TmpTreeOut.erase();
            }
        }

        ///
        /// perform any necessary remappings
        ///
        CDense_seg& ds = align->SetSegs().SetDenseg();
        for (size_t idx = 0;  idx < ds.GetIds().size();  ++idx) {
            CSeq_id& id = *ds.SetIds()[idx];
            string id_str = id.AsFastaString();
            TIdMap::iterator iter = m_IdMap.find(id_str);
            _ASSERT(iter != m_IdMap.end());
            if (iter == m_IdMap.end()) {
                throw runtime_error("Inconsistent identifiers in alignment");
            }

            TLocPair p = iter->second;

            const CSeq_id& real_id =
                sequence::GetId(*p.first, p.second);
            id.Assign(real_id);

            /**
            FIXME: do we need to worry about this?
            if ( !ci->second->IsWhole() ) {
                _ASSERT(false);
            }
            **/
        }

        /// create an annotation to hold our results
        CRef<CSeq_annot> annot(new CSeq_annot());
        annot->SetData().SetAlign().push_back(align);

        string title("ClustalW alignment: ");
        CLabel::GetLabel(*align, &title, CLabel::eDefault, m_Scope.GetPointer());
        annot->SetNameDesc(title);
        annot->SetCreateDate(CTime(CTime::eCurrent));

        if (annot) {
            CRef<CProjectItem> item(new CProjectItem());
            item->SetItem().SetAnnot(*annot);
            item->SetLabel(title);
            m_Items.push_back(item);
        }
        if (btc) {
            title = "Guide tree: " + title;
            btc->SetTreetype(title);

            CRef<CProjectItem> item(new CProjectItem());
            item->SetItem().SetOther().Set(*btc);
            item->SetLabel(title);
            m_Items.push_back(item);
        }
    }
    catch (CException& e) {
        err_msg = "Failed to generate alignment:\n" + e.GetMsg();
    }
    catch (std::exception& e) {
        err_msg = "Failed to generate alignment:\n" + string(e.what());
    }

    x_CleanUp();

    if (!err_msg.empty())
        NCBI_THROW(CException, eUnknown, err_msg);
}

END_NCBI_SCOPE

