/*  $Id: dot_matrix_view.cpp 26936 2012-11-28 14:35:42Z voronov@NCBI.NLM.NIH.GOV $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Andrey Yazhuk
 *
 * File Description:
 *    User-modifiable implementation file for extension of Dot Matrix
 *    viewer for GBENCH
 */

#include <ncbi_pch.hpp>

#include <gui/packages/pkg_alignment/dot_matrix_view.hpp>

#include <gui/widgets/hit_matrix/hit_matrix_widget.hpp>
#include <gui/widgets/hit_matrix/hit_matrix_ds_builder.hpp>
#include <gui/widgets/hit_matrix/dense_ds.hpp>

#include <gui/objutils/sparse_functions.hpp>

#include <gui/core/project_service.hpp>
#include <gui/objutils/visible_range.hpp>

#include <gui/framework/workbench.hpp>

#include <gui/widgets/wx/fileartprov.hpp>
#include <gui/widgets/wx/ui_command.hpp>
#include <gui/widgets/wx/table_listctrl.hpp>
#include <gui/widgets/wx/table_utils.hpp>
#include <gui/widgets/wx/async_call.hpp>

#include <gui/objutils/utils.hpp>
#include <gui/objutils/label.hpp>

#include <serial/iterator.hpp>
#include <objmgr/util/sequence.hpp>

//#include <objects/seqloc/Seq_id.hpp>
#include <objects/seqloc/Seq_interval.hpp>

#include <wx/menu.h>
#include <wx/sizer.h>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

CProjectViewTypeDescriptor CDotMatrixView::m_TypeDescr(
    "Dot Matrix View", // type name
    "dot_matrix_view", // icon alias TODO
    "Dot Matrix View",
    "The Dot Matrix View shows a dot-plot of a given alignment",
    "DOT_MATRIX_VIEW", // help ID
    "Alignment",   // category
    false,       // not a singleton
    "Seq-align",
    eSimilarObjectsAccepted
);


CDotMatrixView::CDotMatrixView()
:   m_AlnWidget(NULL)
{
}


wxWindow* CDotMatrixView::GetWindow()
{
    _ASSERT(m_AlnWidget);
    return m_AlnWidget;
}


void CDotMatrixView::x_CreateMenuBarMenu()
{
    CUICommandRegistry& cmd_reg = m_Workbench->GetUICommandRegistry();

    wxMenu* view_menu = new wxMenu();
    wxMenu* dm_menu = new wxMenu();

    cmd_reg.AppendMenuItem(*dm_menu, eCmdChooseSeq);
    cmd_reg.AppendMenuItem(*dm_menu, eCmdColorByScore);

    view_menu->Append(wxID_ANY, wxT("Dot Matrix"), dm_menu);

    wxMenu* bar_menu = new wxMenu();
    bar_menu->Append( wxID_ANY, wxT("&View"), view_menu );

    m_MenuBarMenu.reset( bar_menu );
}


void CDotMatrixView::CreateViewWindow(wxWindow* parent)
{
    _ASSERT( ! m_AlnWidget);

    m_AlnWidget = new CHitMatrixWidget(parent);
    m_AlnWidget->Create();

    AddListener(m_AlnWidget, ePool_Child);
    m_AlnWidget->AddListener(this, ePool_Parent);
}


void CDotMatrixView::DestroyViewWindow()
{
    if(m_AlnWidget) {
        m_AlnWidget->Destroy();
        m_AlnWidget = NULL;
    }
}


const CViewTypeDescriptor& CDotMatrixView::GetTypeDescriptor() const
{
    return m_TypeDescr;
}

class CCreateDataSource_local : public CAsyncCall::Call
{
public:
    CCreateDataSource_local( CHitMatrixDSBuilder& builder ) : m_Builder( builder ) { CAsyncCall( *this ).Execute(); }

    virtual void Execute();

    CIRef<IHitMatrixDataSource> GetDataSource() { return m_DataSource; }

private:
    CHitMatrixDSBuilder& m_Builder;
    CIRef<IHitMatrixDataSource> m_DataSource;
};

void CCreateDataSource_local::Execute()
{
    m_DataSource = m_Builder.CreateDataSource();
}


void CDotMatrixView::x_InitWidget()
{
    _ASSERT( !m_Aligns.empty() );

    CScope* scope = GetScope().GetPointer();
    _ASSERT(scope);

    m_AlnWidget->SetDataSource( NULL );

    CHitMatrixDSBuilder builder;
    builder.Init( *scope, m_Aligns );

    //m_DataSource = builder.CreateDataSource();
    m_DataSource = CCreateDataSource_local( builder ).GetDataSource();

    if( m_DataSource->SelectDefaultIds() ){
        m_AlnWidget->SetDataSource( m_DataSource.GetPointer() );
    }
}

objects::CScope* CDotMatrixView::x_PreAttachToProject( TConstScopedObjects& objects )
{
    TAlignVector aligns;

    if( objects.size() > 0 ){
        NON_CONST_ITERATE( TConstScopedObjects, it, objects ){
            const CObject* object = it->object.GetPointer();
            const CSeq_annot* annot = dynamic_cast<const CSeq_annot*>(object);
            if( annot ){
                ExtractSeqAligns( *annot, aligns );
            } else {
                const CSeq_align* align = dynamic_cast<const CSeq_align*>(object);
                aligns.push_back( CConstRef<CSeq_align>(align) );
            }
        }
    }

    if( aligns.size() > 0 ){

        m_Aligns = aligns;

        return objects[0].scope.GetPointer();
    }

    return NULL;
}
 
void CDotMatrixView::x_PostAttachToProject()
{
    if( x_HasProject() ){
        x_InitWidget();
    }

    x_UpdateContentLabel();
}

void CDotMatrixView::OnProjectChanged( CProjectViewEvent& evt )
{
    if( 
        evt.GetType() == CViewEvent::eProjectChanged
        && (evt.GetSubtype() & CProjectViewEvent::eProperties)
    ){
        x_UpdateContentLabel();
    }
}

void CDotMatrixView::GetSelection(TConstScopedObjects& objs) const
{
    if (m_AlnWidget  &&  x_HasProject())    {
        typedef CHitMatrixWidget::TRangeColl TRangeColl;
        CScope* scope = x_GetScope();

        // Range selection
        const TRangeColl& s_coll = m_AlnWidget->GetSubjectRangeSelection();
        if(! s_coll.empty())  {
            CRef<CSeq_loc> s_loc =
                CSeqUtils::CreateSeq_loc(*m_DataSource->GetSubjectId().GetSeqId(), s_coll);
            if (s_loc) {
                objs.push_back(SConstScopedObject(*s_loc, *scope));
            }
        }

        const TRangeColl& q_coll = m_AlnWidget->GetQueryRangeSelection();
        if(! q_coll.empty())    {
            CRef<CSeq_loc> q_loc =
                CSeqUtils::CreateSeq_loc(*m_DataSource->GetQueryId().GetSeqId(), q_coll);
            if (q_loc) {
                objs.push_back(SConstScopedObject(*q_loc, *scope));
            }
        }

        // Object Selection
        TConstObjects sel_objs;
        m_AlnWidget->GetObjectSelection(sel_objs);
        ITERATE(TConstObjects, it, sel_objs)    {
            objs.push_back(SConstScopedObject(**it, *scope));
        }
    }
}


void CDotMatrixView::GetSelection(CSelectionEvent& evt) const
{
    if (x_HasProject()) {
        // Range Selection
        evt.AddRangeSelection(*m_DataSource->GetSubjectId().GetSeqId(),
                              m_AlnWidget->GetSubjectRangeSelection());

        evt.AddRangeSelection(*m_DataSource->GetQueryId().GetSeqId(),
                              m_AlnWidget->GetQueryRangeSelection());

        // Object Selection
        TConstObjects sel_objs;
        m_AlnWidget->GetObjectSelection(sel_objs);
        evt.AddObjectSelection(sel_objs);
    }
}


// handles incoming selection broadcast (overriding CView virtual function)
void CDotMatrixView::x_OnSetSelection(CSelectionEvent& evt)
{
    CScope* scope = x_GetScope();

    _ASSERT(scope);

    if(evt.HasRangeSelection())     { // set Range Selection
        typedef CHitMatrixWidget::TRangeColl TRangeColl;
        CSelectionEvent::TRangeColl coll;

        // set Range Selection on Subject
        CConstRef<CSeq_id>  id = m_DataSource->GetSubjectId().GetSeqId();
        evt.GetRangeSelection(*id, *scope, coll);
        if (coll.size()) {
            m_AlnWidget->SetSubjectRangeSelection(coll);
        }

        // ser Range Selection on Query
        coll.clear();
        id = m_DataSource->GetQueryId().GetSeqId();
        evt.GetRangeSelection(*id, *scope, coll);
        if (coll.size()) {
            m_AlnWidget->SetQueryRangeSelection(coll);
        }
    }

    if(evt.HasObjectSelection())    { // set Object Selection
        TConstObjects sel_objs;
        evt.GetAllObjects(sel_objs);

        vector<const CSeq_align*>   sel_aligns;

        typedef CHitMatrixDataSource::THitAdapterCont THits;
        const THits& hits = m_DataSource->GetHits();
        ITERATE(THits, it_hit, hits)    {
            const IHit& hit = **it_hit;
            const CSeq_align& align = *hit.GetSeqAlign();

            ITERATE(TConstObjects, it_obj, sel_objs)    {
                if(evt.Match(align, *scope, **it_obj, evt.GetScope())) {
                    sel_aligns.push_back(&align);
                    break;
                }
            }
        }

        if (sel_aligns.size()) {
            m_AlnWidget->SetObjectSelection(sel_aligns);
        }
    }
}


void CDotMatrixView::GetVisibleRanges(CVisibleRange& vrange) const
{
    if( m_DataSource.IsNull() ){
        ERR_POST(Error << "CDotMatrixView: no data source while trying to GetVisibleRanges()" );
        return;
    }

    /// collect information
    const CGlPane& pane = m_AlnWidget->GetPort();
    TModelRect rect = pane.GetVisibleRect();

    CRef<CSeq_loc> loc(new CSeq_loc());
    loc->SetInt().SetFrom((TSeqPos)rect.Left());
    loc->SetInt().SetTo  ((TSeqPos)rect.Right());
    loc->SetId(*m_DataSource->GetSubjectId().GetSeqId());
    vrange.AddLocation(*loc);

    loc.Reset(new CSeq_loc());
    loc->SetInt().SetFrom((TSeqPos)rect.Bottom());
    loc->SetInt().SetTo  ((TSeqPos)rect.Top());
    loc->SetId(*m_DataSource->GetQueryId().GetSeqId());
    vrange.AddLocation(*loc);
}


void CDotMatrixView::OnVisibleRangeChanged(const CVisibleRange& vrange, IVisibleRangeClient* source)
{
    const CGlPane& pane = m_AlnWidget->GetPort();
    TModelRect rect = pane.GetVisibleRect();

    if (vrange.Match(*m_DataSource->GetQueryId().GetSeqId())) {
        TSeqRange range;
        range.SetFrom((TSeqPos)rect.Bottom());
        range.SetTo  ((TSeqPos)rect.Top());
        if (vrange.Clamp(m_DataSource->GetQueryHandle(), range)) {
            m_AlnWidget->SetQueryVisibleRange(range);
        }
    }

    if (vrange.Match(*m_DataSource->GetSubjectId().GetSeqId())) {
        TSeqRange range;
        range.SetFrom((TSeqPos)rect.Left());
        range.SetTo  ((TSeqPos)rect.Right());
        if (vrange.Clamp(m_DataSource->GetSubjectHandle(), range)) {
            m_AlnWidget->SetSubjectVisibleRange(range);
        }
    }
}


const CObject* CDotMatrixView::x_GetOrigObject() const
{
    if( m_Aligns.size() > 0 ){
        return m_Aligns[0].GetPointer();
    }
    return NULL;
}

///////////////////////////////////////////////////////////////////////////////
/// CDotMatrixViewManager 
///
class CDotMatrixViewManager : public COpenViewManager
{
public:
    CDotMatrixViewManager( IProjectViewFactory& factory )
        : COpenViewManager( factory )
        , mf_IsSet( false )
    {
    }
public:
    void SetObjects( const TObjectListMap& pObjMaps );
    bool AreObjectsSet() const { return mf_IsSet; }
    void ClearObjects();

    objects::CScope* GetScope() const { return m_Scope; }
    const CDotMatrixView::TAlignVector& GetAligns() const { return m_Aligns; }
    const vector<int>& GetGroups() const { return m_Groups; }
    const vector< set<CSeq_id_Handle> >& GetHandlesSet() const { return m_AlnHandles; }
    int GetNumSegments( int aGroupIx ) const {
        return m_NumSegs[aGroupIx];
    }

protected:
    virtual bool NeedParamsPanel( const TObjectListMap& objects );
    virtual CParamsPanel* CreateParamsPanel();

private:
    objects::CScope* m_Scope;
    CDotMatrixView::TAlignVector m_Aligns;
    vector<int> m_Groups;
    vector< set<CSeq_id_Handle> > m_AlnHandles;
    vector<int> m_NumSegs;
    bool mf_IsSet;
};

///////////////////////////////////////////////////////////////////////////////
/// CDotMatrixPanel 
///
class CDotMatrixPanel : public CParamsPanel
{

public:

    CDotMatrixPanel();
    CDotMatrixPanel( wxWindow* parent, wxWindowID id = SYM_CONVRESPANEL_IDNAME, const wxPoint& pos = SYM_CONVRESPANEL_POSITION, const wxSize& size = SYM_CONVRESPANEL_SIZE, long style = SYM_CONVRESPANEL_STYLE );

    virtual bool Create( wxWindow* parent, wxWindowID id = SYM_CONVRESPANEL_IDNAME, const wxPoint& pos = SYM_CONVRESPANEL_POSITION, const wxSize& size = SYM_CONVRESPANEL_SIZE, long style = SYM_CONVRESPANEL_STYLE );

    void Init();

    virtual void CreateControls();

    enum {
        ID_CCONVERSIONRESULTPANEL = 10034,
        ID_ONEVIEWPEROBJECT = 10035
    };
    ////@end CConversionResultPanel member variables

    void SetManager( CDotMatrixViewManager* aManager ){
        m_Manager = aManager;
    }

    /// add your methods here
    virtual void SetObjects( TObjectListMap& objs );

    /// Return the selected item
    virtual bool GetSelection( TConstScopedObjects& buf );

    /// Force OneViewPerObject true, and lock it
    virtual void EnforceOneViewPerObject();

    /// @name IRegSettings implementation
    /// @{
    virtual void SetRegistryPath(const string& path);
    virtual void LoadSettings();
    virtual void SaveSettings() const;
    /// @}

private:
    CDotMatrixViewManager* m_Manager;
    CwxTableListCtrl* m_ListCtrl;
};


///////////////////////////////////////////////////////////////////////////////
/// CDotMatrixPanel implementation
///
void CDotMatrixViewManager::ClearObjects() 
{
    m_Scope = NULL;
    m_Aligns.clear();
    m_Groups.clear();
    m_AlnHandles.clear();
    m_NumSegs.clear();
    
    mf_IsSet = false;
}


void CDotMatrixViewManager::SetObjects( const TObjectListMap& pObjMaps )
{
    ClearObjects();

    CObjectList obj_list;
    ITERATE( TObjectListMap, it, pObjMaps ){
        obj_list.Append(*it->second);
    }
    size_t num_objects = obj_list.GetNumRows();
    if( num_objects > 0 ){
        m_Scope = obj_list.GetScope(0);
    }
     
    for( size_t ix = 0; ix < num_objects; ix++ ){
        try {
            const CSeq_annot* annot = dynamic_cast<const CSeq_annot*>(obj_list.GetObject(ix));
            if( annot ){
                ExtractSeqAligns( *annot, m_Aligns );
            } else {
                const CSeq_align* align = dynamic_cast<const CSeq_align*>(obj_list.GetObject(ix));
                if( align ){
                    m_Aligns.push_back( CConstRef<CSeq_align>(align) );
                }
            }
        } catch( CException& ex ){
            ERR_POST( "DotMatrixVM::SetObjects(): " << ex.GetMsg() );
        }
    }
    m_Groups.resize( m_Aligns.size(), -1 );


    int ix = -1;
    ITERATE( CDotMatrixView::TAlignVector, aln_it, m_Aligns ){
        ix++;

        try {
            const CSeq_align* aln = aln_it->GetPointer();

            int num_seqs = aln->CheckNumRows();
            if( num_seqs <= 0 ){
                continue;
            }

            set<CSeq_id_Handle> idh_set;
            for( int q = 0; q < num_seqs; q++ ){
                const CSeq_id& seq_id = aln->GetSeq_id( q );
                CSeq_id_Handle idh = CSeq_id_Handle::GetHandle( seq_id );
                idh = sequence::GetId( idh, *m_Scope, sequence::eGetId_Canonical );
                idh_set.insert( idh );
            }

            vector< set<CSeq_id_Handle> >::iterator found = 
                find( m_AlnHandles.begin(), m_AlnHandles.end(), idh_set )
                ;
            if( found != m_AlnHandles.end() ){
                m_NumSegs[
                    m_Groups[ix] = found - m_AlnHandles.begin()
                ]++;
            } else {
                m_Groups[ix] = m_AlnHandles.size();
                m_AlnHandles.push_back( idh_set );
                m_NumSegs.push_back( 1 );
            }

            continue;
        } catch( CException& ex ){
            ERR_POST( "DotMatrixVM::SetObjects(): " << ex.GetMsg() );
        }
    }

    mf_IsSet = true;
}

bool CDotMatrixViewManager::NeedParamsPanel( const TObjectListMap& objects )
{
    SetObjects( objects );

    if( m_AlnHandles.size() > 1 ){
        return true;

    } else if( m_AlnHandles.size() == 1 ){
        return false;

    } else {    
        NCBI_THROW( COVMException, eGroupErr, "Input objects are not sutable" );
    }
}

CParamsPanel* CDotMatrixViewManager::CreateParamsPanel()
{
    CDotMatrixPanel* panel = new CDotMatrixPanel();

    panel->SetManager( this );

    return panel;
}


///////////////////////////////////////////////////////////////////////////////
/// CDotMatrixPanel implementation
///
CDotMatrixPanel::CDotMatrixPanel()
{
    Init();
}


CDotMatrixPanel::CDotMatrixPanel( wxWindow* parent, wxWindowID id, const wxPoint& pos, const wxSize& size, long style )
{
    Init();
    Create(parent, id, pos, size, style);
}


bool CDotMatrixPanel::Create( wxWindow* parent, wxWindowID id, const wxPoint& pos, const wxSize& size, long style )
{
    ////@begin CDotMatrixPanel creation
    CParamsPanel::Create( parent, id, pos, size, style );

    CreateControls();
    if (GetSizer())
    {
        GetSizer()->SetSizeHints(this);
    }
    Centre();
    ////@end CDotMatrixPanel creation
    return true;
}

void CDotMatrixPanel::Init()
{
    m_Manager = NULL;
    m_ListCtrl = NULL;
}


void CDotMatrixPanel::CreateControls()
{
    ////@begin CDotMatrixPanel content construction
    CDotMatrixPanel* itemPanel1 = this;

    wxBoxSizer* itemBoxSizer2 = new wxBoxSizer(wxVERTICAL);
    itemPanel1->SetSizer(itemBoxSizer2);

    m_ListCtrl = new CwxTableListCtrl( itemPanel1, wxID_ANY, wxDefaultPosition, wxSize(100, 100), 0 );
    itemBoxSizer2->Add( m_ListCtrl, 1, wxGROW|wxALL, 5 );

    /*
    m_OneViewPerObjectCheckBox = new wxCheckBox( itemPanel1, ID_ONEVIEWPEROBJECT, _("Open each object in a separate view"), wxDefaultPosition, wxDefaultSize, 0 );
    m_OneViewPerObjectCheckBox->SetValue(true);
    itemBoxSizer2->Add(m_OneViewPerObjectCheckBox, 0, wxALIGN_LEFT|wxLEFT|wxBOTTOM, 5);
    */

    // Set validators
    //m_OneViewPerObjectCheckBox->SetValidator( wxGenericValidator(& m_OneViewPerObject) );
    ////@end CDotMatrixPanel content construction

}


void CDotMatrixPanel::SetObjects( TObjectListMap& pObjMaps )
{
    _ASSERT( m_Manager );

    if( !m_Manager->AreObjectsSet() ){
        m_Manager->SetObjects( pObjMaps );
    }


    CParamsPanel::SetObjects( pObjMaps );

    vector<wxString> columns;
    columns.push_back( _("Alignment sequences") );

    const vector< set< CSeq_id_Handle > >& handlers_set = m_Manager->GetHandlesSet();
    int num_sets = handlers_set.size();

    CTextTableModel* model = new CTextTableModel();
    model->Init( columns, num_sets );

    for( int ix = 0; ix < num_sets; ix++ ){
        const set<CSeq_id_Handle>& aln_set = handlers_set[ix];

        string label;

        bool first = true;
        ITERATE( set<CSeq_id_Handle>, aln_ir, aln_set ){
            if( first ){
                first = false;
            } else {
                label += " x ";
            }
            label += objects::GetLabel( sequence::GetId( 
                *aln_ir->GetSeqId(), *m_Manager->GetScope(), sequence::eGetId_Best 
            ) );
        }
        label += " [" 
            + NStr::IntToString( m_Manager->GetNumSegments( ix ), NStr::fWithCommas ) 
            + " objects]"
        ;

        model->SetStringValueAt( ix, 0, wxString::FromUTF8( label.c_str() ) );
    }

    m_ListCtrl->SetModel( model, true );
    m_ListCtrl->SetColumnWidth( 0, 250 );
    m_ListCtrl->TransferDataToWindow();

    m_ListCtrl->SelectAll();
}

bool CDotMatrixPanel::GetSelection( TConstScopedObjects& buf )
{
    if( !m_ListCtrl->TransferDataFromWindow() ){
        return false;
    }

    buf.clear();
    if( m_ListCtrl ){

        const CDotMatrixView::TAlignVector& aligns = m_Manager->GetAligns();
        const vector<int>& groups = m_Manager->GetGroups();

        vector<int> sel_rows;
        m_ListCtrl->GetDataRowsSelected( sel_rows );

        ITERATE( vector<int>, row_ir, sel_rows ){
            for( size_t ix = 0; ix < groups.size(); ix++ ){
                if( groups[ix] == *row_ir ){
                    const objects::CSeq_align& aln = *aligns[ix];

                    SConstScopedObject obj( aln, *m_Manager->GetScope() );

                    buf.push_back( obj );
                }
            }
        }
    }
    return ( ! buf.empty());
}

void CDotMatrixPanel::EnforceOneViewPerObject()
{
    //CParamsPanel::ForceOneViewPerObject();
}


void CDotMatrixPanel::SetRegistryPath(const string& path)
{
    CParamsPanel::SetRegistryPath( path );
}

static const char* kDMGroupList = "DotMatrixGroupList";


void CDotMatrixPanel::SaveSettings() const
{
    CParamsPanel::SaveSettings();

    if( !m_RegPath.empty() ){
        CGuiRegistry& gui_reg = CGuiRegistry::GetInstance();
        CGuiRegistry::TReadWriteView view = gui_reg.GetReadWriteView(
            m_RegPath + "." + kDMGroupList 
        );
        ncbi::SaveTableSettings( *m_ListCtrl, view );
    }
}


void CDotMatrixPanel::LoadSettings()
{
    CParamsPanel::LoadSettings();
    
    // Protection from older config savings
    m_OneViewPerObject = false;

    if ( !m_RegPath.empty() ) {
        CGuiRegistry& gui_reg = CGuiRegistry::GetInstance();
        CGuiRegistry::TReadWriteView view = gui_reg.GetReadWriteView(
            m_RegPath + "." + kDMGroupList 
        );

        ncbi::LoadTableSettings( *m_ListCtrl, view );
        TransferDataToWindow();
    }
}

///////////////////////////////////////////////////////////////////////////////
/// CDotMatrixViewFactory
CDotMatrixViewFactory::CDotMatrixViewFactory()
{
    //m_ViewManager = new CDotMatrixViewManager( *this );
}

string  CDotMatrixViewFactory::GetExtensionIdentifier() const
{
    static string sid("dot_matrix_view_factory");
    return sid;
}

string CDotMatrixViewFactory::GetExtensionLabel() const
{
    static string slabel("Dot Matrix View Factory");
    return slabel;
}

void CDotMatrixViewFactory::RegisterIconAliases(wxFileArtProvider& provider)
{
    string alias = GetViewTypeDescriptor().GetIconAlias();
    provider.RegisterFileAlias(ToWxString(alias), wxT("dot_matrix_view.png"));
}

void CDotMatrixViewFactory::RegisterCommands (CUICommandRegistry& cmd_reg, wxFileArtProvider& provider)
{
    CHitMatrixWidget::RegisterCommands(cmd_reg, provider);
}

const CProjectViewTypeDescriptor& CDotMatrixViewFactory::GetProjectViewTypeDescriptor() const
{
    return CDotMatrixView::m_TypeDescr;
}

IView* CDotMatrixViewFactory::CreateInstance() const
{
    return new CDotMatrixView();
}

IView* CDotMatrixViewFactory::CreateInstanceByFingerprint(const TFingerprint& fingerprint) const
{
    /*TFingerprint print(CDotMatrixView::m_TypeDescr.GetLabel(), false);
    if(print == fingerprint)   {
        return new CDotMatrixView();
    }*/
    return NULL;
}





int CDotMatrixViewFactory::TestInputObjects( TConstScopedObjects& objects )
{
    bool found_good = false;
    bool found_bad = false;
    for( size_t i = 0; i < objects.size(); i++ ){
        const CObject* obj = objects[i].object;
        const type_info& type = typeid(*obj);

        //if( typeid(CSeq_annot) == type || typeid(CSeq_align) == type ){
        if( typeid(CSeq_align) == type ){
            found_good = true;
        } else {
            found_bad = true;
        }
    }
    if( found_good ){
        return fCanShowSeparated | (found_bad ? fCanShowSome : fCanShowAll);
    }
    return 0; // can show nothing
}


END_NCBI_SCOPE
