/*  $Id: muscle_tool_job.cpp 26643 2012-10-18 17:54:05Z katargir $
* ===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data,  the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties,  express or implied,  including
*  warranties of performance,  merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================
*
* Authors:  Roman Katargin
*/


#include <ncbi_pch.hpp>

#include <corelib/ncbifile.hpp>
#include <corelib/ncbiexec.hpp>
#include <algo/phy_tree/phy_node.hpp>
#include <objmgr/seq_vector.hpp>
#include <objmgr/util/sequence.hpp>
#include <objtools/readers/fasta.hpp>

#include <serial/iterator.hpp>

#include <objects/seqalign/Seq_align.hpp>
#include <objects/biotree/BioTreeContainer.hpp>
#include <objects/seq/Annotdesc.hpp>

#include <algo/phy_tree/dist_methods.hpp>

#include <gui/objutils/label.hpp>
#include <gui/objects/AbstractProjectItem.hpp>

#include <gui/packages/pkg_alignment/muscle_tool_job.hpp>
#include <gui/widgets/wx/wx_utils.hpp>
#include <gui/widgets/wx/async_call.hpp>

#include <wx/filename.h>
#include <wx/sstream.h>
#include <wx/txtstrm.h>
#include <wx/utils.h>
#include <wx/msgdlg.h>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

///////////////////////////////////////////////////////////////////////////////
/// CMuscleToolJob
CMuscleToolJob::CMuscleToolJob(CProjectService* service,
                                 const CMuscleToolParams& params)
                                 : CDataLoadingAppJob(service), m_Params(params), m_PId(0)
{
    CFastMutexGuard lock(m_Mutex);
    m_Descr = "MUSCLE Job";
}

/*
    TODO:
    IdMap is used now only for tree remapping. 
    It seems more effective to store labels there, not SeqLocs. 
    It seems the most effective to retrieve this mapping
    from CMappingRanges or from CSeq_loc_Mapper. Probably
    CRef<CSeq_loc> Map(const CSeq_loc& src_loc) can be used.
*/

typedef CMuscleToolJob::TLocPair TLocPair;
typedef CMuscleToolJob::TIdMap TIdMap;
struct STreeNodeMapper
{
    CMuscleToolJob::TIdMap* idmap;
    STreeNodeMapper()
        : idmap(NULL)
    {
    }

    ETreeTraverseCode operator()(TPhyTreeNode& node, int level)
    {
        _ASSERT(idmap);
        string& label = node.GetValue().SetLabel();
        if (label.empty()) {
            return eTreeTraverse;
        }
        TIdMap::iterator id_iter = idmap->find(label);
        if (id_iter == idmap->end()) {
            return eTreeTraverse;
        }

        TLocPair p = id_iter->second;
        CScope& scope = *p.second;
        const CSeq_id& real_id =
            sequence::GetId(*p.first, &scope);
        label.erase();
        CLabel::GetLabel(real_id, &label, CLabel::eDefault, &scope);
        return eTreeTraverse;
    }
};

/*
static void s_DumpSequence(CScope& scope,
                           const CSeq_loc& loc,
                           TIdMap& idmap,
                           bool& is_prot,
                           CNcbiOstream& ostr)
{
    string id_str = "lcl|seq-" + NStr::SizetToString(idmap.size());
    idmap[id_str] = make_pair(CConstRef<CSeq_loc>(&loc), CRef<CScope>(&scope));

    ostr << ">" << id_str << endl;

    CSeqVector vec(loc, scope, CBioseq_Handle::eCoding_Iupac);
    if (vec.IsProtein()) {
        is_prot = true;
    }
    string data;
    vec.GetSeqData(0, vec.size(), data);
    ostr << data << endl;
}

class CDumpSequences_local : public CAsyncCall::Call
{
public:
    CDumpSequences_local(
        CNcbiOfstream& ostream,
        TConstScopedObjects& locations,
        map<string, TLocPair>& idMap
    )
        : CAsyncCall::Call(wxT("Preparing data for MUSCLE..."))
        , m_Ostream(ostream)
        , m_Locations(locations)
        , m_IdMap(idMap)
        , m_Result(0) 
    {}

    virtual void Execute();
    int GetResult() const { return m_Result; }

private:
    CNcbiOfstream& m_Ostream;
    TConstScopedObjects& m_Locations;
    map<string, TLocPair>& m_IdMap;
    int m_Result;
};

void CDumpSequences_local::Execute()
{
    NON_CONST_ITERATE( TConstScopedObjects, iter, m_Locations ){
        const CSeq_loc& loc = dynamic_cast<const CSeq_loc&>(iter->object.GetObject());
        CScope& scope = iter->scope.GetObject();
        bool is_prot = false;
        s_DumpSequence(scope, loc, m_IdMap, is_prot, m_Ostream);
        m_Result |= is_prot ? 2 : 1;
    }
}
*/

typedef map<CSeq_id_Handle, bool> TSeqTypeMap;

static void s_DumpSequence_wRange(
    CNcbiOstream& ostr,
    const CSeq_loc* loc,
    CScope* scope,
    TSeqTypeMap& seq_types,
    CMappingRanges& ranges,
    map<string, TLocPair>& idmap
){
    static int dump_count = 0;

    string loc_id = "seq-" + NStr::SizetToString( ++dump_count );
    idmap["lcl|"+loc_id] = make_pair(CConstRef<CSeq_loc>(loc), CRef<CScope>(scope));

    ostr << ">" << "lcl|" << loc_id << endl;

    CSeqVector vec( *loc, *scope, CBioseq_Handle::eCoding_Iupac );
    string data;
    vec.GetSeqData( 0, vec.size(), data );
    ostr << data << endl;

    CBioseq_Handle handle = scope->GetBioseqHandle( *loc->GetId() );

    CRef<CSeq_id> seq_id( new CSeq_id( CSeq_id::e_Local, loc_id ) );
    CSeq_id_Handle local_sihd = CSeq_id_Handle::GetHandle( *seq_id );
    seq_types[local_sihd] = handle.IsNucleotide();

    CSeq_id_Handle base_sihd = CSeq_id_Handle::GetHandle( *loc->GetId() );
    seq_types[base_sihd] = handle.IsNucleotide();

    int trcf = handle.IsNucleotide() ? 1: 3;

    // prepare for final mapping
    ranges.AddConversion( 
        local_sihd, 
        0 *trcf, sequence::GetLength( *loc, scope ) *trcf, 
        eNa_strand_plus,
        base_sihd, 
        sequence::GetStart( *loc, scope ) *trcf,  
        sequence::GetStrand( *loc, scope )
    );
}


class CDumpSequences_wRanges : public CAsyncCall::Call
{
public:
    CDumpSequences_wRanges(
        CNcbiOfstream& ostream,
        TConstScopedObjects& locations,
        TSeqTypeMap& seq_types,
        CMappingRanges& ranges,
        map<string, TLocPair>& idMap
    )
        : CAsyncCall::Call(wxT("Preparing data for MUSCLE..."))
        , m_Ostream(ostream)
        , m_Locations(locations)
        , m_SeqTypes(seq_types)
        , m_Ranges(ranges) 
        , m_IdMap(idMap)
    {}

    virtual void Execute();

private:
    CNcbiOfstream& m_Ostream;
    TConstScopedObjects& m_Locations;
    TSeqTypeMap& m_SeqTypes;
    CMappingRanges& m_Ranges;
    map<string, TLocPair>& m_IdMap;
};

void CDumpSequences_wRanges::Execute()
{
    NON_CONST_ITERATE( TConstScopedObjects, iter, m_Locations ){
        const CSeq_loc* loc = dynamic_cast<const CSeq_loc*>( iter->object.GetPointer() );
        s_DumpSequence_wRange( m_Ostream, loc, iter->scope, m_SeqTypes, m_Ranges, m_IdMap );
    }
}

bool CMuscleToolJob::BeforeRun()
{
    wxString muscleExe = m_Params.GetMusclePath();
    if (!wxFileName::FileExists(muscleExe)) {
        NCBI_THROW(CException, eUnknown, "Path to MUSCLE executable does not exist!");
    }

    /// we serialize this through a set of temporary files
    m_TmpIn      = wxFileName::CreateTempFileName(wxT("in"));
    m_TmpOut     = wxFileName::CreateTempFileName(wxT("out"));

    bool is_nuc = true;

    /// scoped to make sure we flush and close our file before we start!
    {{
        CNcbiOfstream ostr(m_TmpIn.fn_str());
        CDumpSequences_wRanges dumpSequences( ostr, m_Params.m_Locations, m_SeqTypes, m_Ranges, m_IdMap );
        CAsyncCall::Execute( dumpSequences );

        bool ors = false, ands = true;
        ITERATE( TSeqTypeMap, iter, m_SeqTypes ){
            ors = ors || iter->second;
            ands = ands && iter->second;
        }

        if( ands != ors ){
            // we have mix, quit
            x_CleanUp();
            wxMessageBox(
                wxT("MUSCLE tool: All input sequences must be either DNA or protein."),
                wxT("Error"), wxOK | wxICON_ERROR
            );
            return false;
        }

        is_nuc = ors;
    }}

    ///
    /// compose the MUSCLE command line
    ///

    wxStringOutputStream strstr;
    wxTextOutputStream os(strstr);

    os  << wxT("\"") << muscleExe << wxT("\"")
        << wxT(" -in ") << m_TmpIn 
        << wxT(" -out ") << m_TmpOut
        << wxT(" -maxmb 2048");

    if( !is_nuc ){
        os << wxT(" -seqtype protein");
    }

    switch (m_Params.GetScoringMethod()) {
    case 1 :
        os << wxT(" -sp");
        break;
    case 2 :
        os << wxT(" -sv");
        break;
    default :
        os << wxT(" -le");
        break;
    }

    /// do we create a tree?
    if (m_Params.GetGenerateTree()) {
        m_TmpTreeOut = wxFileName::CreateTempFileName(wxT("tre"));
        os << wxT(" -tree2 ") << m_TmpTreeOut;
    }

    wxString moreParams = m_Params.GetCommandLine();
    moreParams.Trim(true);
    moreParams.Trim(false);
    if (!moreParams.empty())
        os << wxT(" ") << moreParams;

    m_CmdLine = strstr.GetString();
    LOG_POST(Info << "Launching MUSCLE executabe:");
    LOG_POST(Info << m_CmdLine);

    m_PId = ::wxExecute(m_CmdLine);
    if (m_PId == 0)
        NCBI_THROW(CException, eUnknown, "Failed to launch MUSCLE executable.");

    return true;
}

void CMuscleToolJob::x_CleanUp()
{
#ifndef _DEBUG
    if (!m_TmpIn.empty())
        ::wxRemoveFile(m_TmpIn);
    if (!m_TmpOut.empty())
        ::wxRemoveFile(m_TmpOut);
    if (!m_TmpTreeOut.empty())
        ::wxRemoveFile(m_TmpTreeOut);
#endif
}

void CMuscleToolJob::x_CreateProjectItems()
{
    if (m_PId == 0) {
        NCBI_THROW(CException, eUnknown, "MUSCLE not launched!");
    }

    while (true) {
        ::wxSleep(3);

        if (IsCanceled()) {
            ::wxKill(m_PId, wxSIGKILL);
            x_CleanUp();
            return;
        }

        ncbi::CProcess proc(m_PId, ncbi::CProcess::ePid);
        if (!proc.IsAlive()) {
            break;
        }
    }

    wxULongLong size = wxFileName::GetSize(m_TmpOut);
    if (size == 0 || size ==  wxInvalidSize) {
        x_CleanUp();
        NCBI_THROW(CException, eUnknown, "MUSCLE executabale failed to produce results.");
    }

    string err_msg;
    try {
        CNcbiIfstream istr(m_TmpOut.fn_str());
        CStreamLineReader stream_line(istr);
        CFastaReader fasta_reader(stream_line, CFastaReader::fParseGaps);

        /// -1 = multiple alignment
        CRef<CSeq_entry> entry = fasta_reader.ReadAlignedSet(-1);
        CRef<CSeq_align> align;
        CTypeIterator<CSeq_align> iter(*entry);
        if (iter) {
            align.Reset(&*iter);
            ++iter;
            if (iter) {
                LOG_POST(Error << "CAlgoPlugin_AlignMuscle::Run(): more than one alignment!");
            }
        }

        ///
        /// read our tree, if we created one
        ///
        CRef<CBioTreeContainer> btc;
        if (!m_TmpTreeOut.empty()) {
            try {
                CNcbiIfstream tree_istr(m_TmpTreeOut.fn_str());
                auto_ptr<TPhyTreeNode> tree(ReadNewickTree(tree_istr));

                /// we need to convert the labels to the correct seq-id strings
                STreeNodeMapper mapper;
                mapper.idmap = &m_IdMap;
                TreeDepthFirstTraverse(*tree, mapper);

                /// convert to the serializable form
                btc = MakeBioTreeContainer(tree.get());

                /// adding seq-id prop
                const int kLabelId = 0;
                const int kSeqIdId = 2;

                CRef<CFeatureDescr> feat_descr(new CFeatureDescr);
                feat_descr->SetId(kSeqIdId);
                feat_descr->SetName("seq-id");
                btc->SetFdict().Set().push_back(feat_descr);

                NON_CONST_ITERATE (CNodeSet::Tdata, node, btc->SetNodes().Set()) {
                    if ( !(*node)->CanGetFeatures() ) {
                        continue;
                    }
                    NON_CONST_ITERATE (CNodeFeatureSet::Tdata, node_feature,
                            (*node)->SetFeatures().Set()) {
                        if ((*node_feature)->GetFeatureid() == kLabelId) {
                            CRef<CNodeFeature> id_node_feature(new CNodeFeature);
                            id_node_feature->SetFeatureid(kSeqIdId);
                            id_node_feature->SetValue((*node_feature)->GetValue());
                            (*node)->SetFeatures().Set().push_back(id_node_feature);
                        }
                    }
                }
            }
            catch (CException& e) {
                LOG_POST(Error
                    << "Failed to read phylogenetic tree:\n"
                    << e.GetMsg()
                    << "\nPhylogenetic tree output will not be available.");
                btc.Reset();
                m_TmpTreeOut.erase();
            }
            catch (std::exception& e) {
                LOG_POST(Error
                    << "Failed to read phylogenetic tree:\n"
                    << e.what()
                    << "\nPhylogenetic tree output will not be available.");
                btc.Reset();
                m_TmpTreeOut.erase();
            }
        }

        ///
        /// perform any necessary remappings
        ///
        CSeq_loc_Mapper mapper( &m_Ranges );
        ITERATE( TSeqTypeMap, st_itr, m_SeqTypes ){
            mapper.SetSeqTypeById(
                st_itr->first, 
                st_itr->second ? CSeq_loc_Mapper::eSeq_nuc : CSeq_loc_Mapper::eSeq_prot
            );
        }

        //- translate locals  to proper ids 
        //- translate coords to proper offsets
        align = mapper.Map( *align );

        ///
        /// create an annotation to hold our results
        CRef<CSeq_annot> annot(new CSeq_annot());
        annot->SetData().SetAlign().push_back(align);

        string title("MUSCLE alignment: ");
        CLabel::GetLabel(*align, &title, CLabel::eDefault,
            m_Scope.GetPointer());
        annot->SetNameDesc(title);
        annot->SetCreateDate(CTime(CTime::eCurrent));

        string comment = "Generated by MUSCLE tool with command line:\n";
        comment += m_CmdLine.ToUTF8();

        if (annot) {
            CRef<CProjectItem> item(new CProjectItem());
            item->SetItem().SetAnnot(*annot);
            item->SetLabel(title);
            CRef<CAnnotdesc> descr(new CAnnotdesc());
            descr->SetComment(comment);
            item->SetDescr().push_back(descr);
            m_Items.push_back(item);
        }
        if (btc) {
            title = "Phylogenetic tree: " + title;
            btc->SetTreetype(title);

            CRef<CProjectItem> item(new CProjectItem());
            item->SetItem().SetOther().Set(*btc);
            item->SetLabel(title);
            CRef<CAnnotdesc> descr(new CAnnotdesc());
            descr->SetComment(comment);
            item->SetDescr().push_back(descr);
            m_Items.push_back(item);
        }
    }
    catch (CException& e) {
        err_msg = "Failed to generate alignment:\n" + e.GetMsg();
    }
    catch (std::exception& e) {
        err_msg = "Failed to generate alignment:\n" + string(e.what());
    }

    x_CleanUp();

    if (!err_msg.empty())
        NCBI_THROW(CException, eUnknown, err_msg);
}

END_NCBI_SCOPE

