/*  $Id: phy_tree_view.cpp 27650 2013-03-20 13:56:37Z falkrb $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Vladimir Tereshkov
 *
 * File Description:
 *    Phylogenetic tree viewer
 *    
 */

#include <ncbi_pch.hpp>

#include <gui/packages/pkg_alignment/phy_tree_view.hpp>

#include <gui/widgets/phylo_tree/phylo_tree_widget.hpp>
#include <gui/widgets/phylo_tree/phylo_tree_ds.hpp>
#include <gui/core/project_service.hpp>

#include <gui/framework/workbench.hpp>
#include <gui/framework/app_task_service.hpp>

#include <gui/widgets/wx/fileartprov.hpp>
#include <gui/widgets/wx/ui_command.hpp>
#include <gui/widgets/wx/wx_utils.hpp>

#include <gui/objutils/utils.hpp>
#include <gui/objutils/label.hpp>
#include <gui/objutils/object_index.hpp>
#include <gui/objutils/taxid_sel_set.hpp>

#include <serial/serialimpl.hpp>

#include <gui/widgets/phylo_tree/phylo_tree_reader.hpp>

#include <gui/objects/AbstractProjectItem.hpp>
#include <objects/biotree/FeatureDictSet.hpp>
#include <objects/biotree/FeatureDescr.hpp>
#include <objects/biotree/NodeSet.hpp>
#include <objects/biotree/Node.hpp>
#include <objects/biotree/NodeFeatureSet.hpp>
#include <objects/biotree/NodeFeature.hpp>
#include <objects/seqloc/Seq_interval.hpp>

#include <wx/menu.h>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);


class CSelNodeHandle : public ISelObjectHandle
{
public:
    CSelNodeHandle(CPhyloTreeNode* node) : m_Node(node)    {}
    CPhyloTreeNode* GetNode() const   {   return m_Node;    }

protected:
    CPhyloTreeNode* m_Node;
};

class CPhyRebuildIndexVisitor
{
public:
    CPhyRebuildIndexVisitor(CObjectIndex & idx) : m_Idx(&idx){}
    ETreeTraverseCode operator()( CPhyloTreeDataSource::TTreeType &  tree_node, int delta)
    {
        if ((delta==1 || delta==0) && !(*tree_node).GetSeqID().Empty()){
            m_Idx->Add(new CSelNodeHandle(&tree_node), (*tree_node).GetSeqID().GetObject());
        }
        return eTreeTraverse;
    }
private:
    CObjectIndex * m_Idx;
};


class CChangePhyloTreeCommand : public CObject, public IEditCommand
{
public:
    CChangePhyloTreeCommand( CPhyTreeView      * view,
                             CBioTreeDynamic & orig_tree,
                             CBioTreeDynamic & new_tree) : m_View(view)
    {
        m_Tree.orig_tree = orig_tree;
        m_Tree.new_tree  = new_tree;
        m_bUpdate        = true;
    }
    void EnableUpdates(bool bEnable)
    {
       m_bUpdate = bEnable;
    }
    virtual void Execute()
    {
        m_View->SetBiotree() = m_Tree.new_tree;
        if (m_bUpdate) {
            m_View->OnBioTreeChanged();
        }
    }
    virtual void Unexecute()
    {
        m_View->SetBiotree() = m_Tree.orig_tree;
        if (m_bUpdate) {
            m_View->OnBioTreeChanged();
        }
    }
    virtual string GetLabel() { return "Phylogenetic tree edit"; }

private:
    bool                          m_bUpdate;
    CPhyTreeView                * m_View;
   
    typedef struct SBTree {
        CBioTreeDynamic   orig_tree;
        CBioTreeDynamic   new_tree;
    } TBTree;
    TBTree m_Tree;
};




BEGIN_EVENT_MAP(CPhyTreeView, CProjectView)
    //ON_COMMAND(eCmdExportSelected,          &CViewPhyloTree::OnExportSelected)
	//ON_COMMAND(eCmdExportSelectedNew,       &CViewPhyloTree::OnExportSelectedNew)
    //ON_COMMAND(eCmdNewViewFromSelection,    &CViewPhyloTree::OnNewViewFromSelection)

    //ON_UPDATE_COMMAND_UI(eCmdExportSelected,        &CViewPhyloTree::OnUpdateExportSelected)
	//ON_UPDATE_COMMAND_UI(eCmdExportSelectedNew,     &CViewPhyloTree::OnUpdateExportSelected)
    //ON_UPDATE_COMMAND_UI(eCmdNewViewFromSelection,  &CViewPhyloTree::OnUpdateNewViewFromSelection)

    ON_MESSAGE(CEvent, eCmdSomethingEdited,  &CPhyTreeView::OnSomethingEdited)
	
	ON_MESSAGE(CEvent, eCmdExportSelected,			&CPhyTreeView::OnCreateSubtree)
	ON_MESSAGE(CEvent, eCmdExportSelectedNew,		&CPhyTreeView::OnCreateSubtree)
	ON_MESSAGE(CEvent, eCmdNewViewFromSelection,	&CPhyTreeView::OnCreateSubtree)

END_EVENT_MAP()


CProjectViewTypeDescriptor s_PhyTreeViewTypeDescr(
    "Tree View", // type name
    "phylo_tree_view", // icon alias TODO
    "Tree View",
    "View for graphical presentation of hierarchical data (phylogenetic tree, taxonomy tree, etc.) using various layout methods",
    "PHY_TREE_VIEW", // help ID
    "Alignment",   // category
    false,       // not a singleton
    "BioTreeContainer",
    eOneObjectAccepted
);

CPhyTreeView::CPhyTreeView()
:   m_PhyWidget(NULL)
{
}

CPhyTreeView::~CPhyTreeView()
{
    delete m_PhyWidget;
}


wxWindow* CPhyTreeView::GetWindow()
{
    _ASSERT(m_PhyWidget);
    return m_PhyWidget;
}


void CPhyTreeView::x_CreateMenuBarMenu()
{   
    wxMenu* view_menu = new wxMenu();

    //m_AlnWidget->AppendMenuItems( *view_menu );
    view_menu->Append( wxID_SEPARATOR, wxT("Actions") );
    view_menu->Append( eCmdHighlightCollapsed, wxT("Highlight Collapsed\tCTRL-L") );

    wxMenu* bar_menu = new wxMenu();
    bar_menu->Append( wxID_ANY, wxT("&View"), view_menu );

    m_MenuBarMenu.reset( bar_menu );
}


void CPhyTreeView::CreateViewWindow(wxWindow* parent)
{
    _ASSERT( ! m_PhyWidget);
    
    m_PhyWidget = new CPhyloTreeWidget(parent);
    m_PhyWidget->Create();

    AddListener(m_PhyWidget, ePool_Child);
    m_PhyWidget->AddListener(this, ePool_Parent);
	
}


void CPhyTreeView::DestroyViewWindow()
{
    if(m_PhyWidget) {
        m_PhyWidget->Destroy();
        m_PhyWidget = NULL;
    }    
}


const CViewTypeDescriptor& CPhyTreeView::GetTypeDescriptor() const
{
    return s_PhyTreeViewTypeDescr;
}


bool CPhyTreeView::InitView(TConstScopedObjects& objects, const objects::CUser_object*)
{
    m_PhyWidget->SetServiceLocator(m_Workbench);


    // exploring objects to find biotree object 
    if(objects.size() > 0)  {
        ITERATE(TConstScopedObjects, it, objects) {
            const CObject* object = it->object.GetPointer();
            m_BioTreeContainer = dynamic_cast<const CBioTreeContainer *>(object);

              
            // first found biotree will be shown
            // phylogenetic tree can be extended to display multiple trees
            if(m_BioTreeContainer) {
                if (m_BioTreeContainer->GetNodeCount() == 0 ) {
                    m_BioTreeContainer.Reset();
                }
                else {
                    break;
                }
            } 
        }
        if (m_BioTreeContainer) {
            CIRef<CProjectService> srv = m_Workbench->GetServiceByType<CProjectService>();

            _ASSERT(srv);

            CScope* scope = objects[0].scope.GetPointer();

            CRef<CWorkspaceConstPrx> workspace = srv->GetWorkspaceConst();
            CRef<CProjectPrx> project = workspace->FindProjectByScope(scope);
            _ASSERT(project);
            
            m_PhyWidget->LoadSettings();
            //if (m_PhyWidget->LoadSettings()) {
            //    LOG_POST(Info <<
                 //   "CPhyTreeView::InitView() - finish with loading configuration!");
          //  }
          //  else {          
          //      wxMessageBox(wxT("Something is wrong with configuration!"),
          //          wxT("Failed to load configuration for Tree view"),
          //          wxOK | wxICON_EXCLAMATION);
           //     return false;
          //  }           

            if (project) {
                // check and create scope if needed
                //project->GetDocument()->CreateProjectScope();
                //project->GetDocument()->SetScope(scope);
                x_AttachToProject(*project);
                OnProjectChanged();
                return true;
            }
            else {
                return false;
            }
        }
        else {
            x_ReportInvalidInputData(objects);
            return false;
        }        
    }
    return false;
}


void CPhyTreeView::OnProjectChanged()
{
    x_UpdateContentLabel();

    if(x_HasProject()) {
        _ASSERT( ! m_BioTreeContainer.Empty());

        // we are able to retrieve new object      
        m_BioTreeData.Clear();
        BioTreeConvertContainer2Dynamic(m_BioTreeData, *m_BioTreeContainer);
        OnBioTreeChanged();        
    }
    else {
        m_ObjectIndex.Clear(NULL);
    }
}


void CPhyTreeView::OnBioTreeChanged()
{  
    CScope* scope = x_GetScope();
    _ASSERT(scope);

    bool bInitial = m_DataSource.Empty();
    bool first_update = (m_PhyWidget->GetDS() == NULL);

    // This will cancel any background processes using the data source
    m_PhyWidget->RemoveCurrentDataSource();
    // Delete and recreate the data source
    m_DataSource.Reset(new CPhyloTreeDataSource(m_BioTreeData, *scope));

    // proceed sorting if needed
    if (m_BioTreeData.GetFeatureDict().HasFeature("do_sort") && bInitial) {
        m_DataSource->Sort();
    }

    // do color-based clustering if colors present
    /*
    if (m_BioTreeData.GetFeatureDict().HasFeature("$NODE_COLOR")) {
        m_DataSource->Clusterize();
    }
    */
    if (m_DataSource->HasClusters()) {
        m_DataSource->Clusterize();
    }

    //m_DataSource->MarkAllDirty();

    // On linux, the gui updates forced by SetDataSource() can
    // cause a crash if the window/widget is not yet fully created.
    m_PhyWidget->SetDataSourceNoUpdate(m_DataSource);

    // set default scheme
    m_Scheme.Reset(new CPhyloTreeScheme());
    if (m_PhyWidget->HasScheme())
        *m_Scheme = m_PhyWidget->GetScheme();

    m_Scheme->LoadCurrentSettings();

    // set default label style for tax tree
    if (m_BioTreeContainer->CanGetTreetype() && (m_BioTreeContainer->GetTreetype() == "CommonTaxTree")){
        m_Scheme->SetLabelFormat() = "$(name)$(seq-id)";
        m_Scheme->SetLabelVisibility(CPhyloTreeScheme::eLabelsVisible);
        //m_PhyWidget->SetCurrRenderer(3);
    }


    // apply distances rendering, if distances available
    if (bInitial) {
        m_PhyWidget->SetUseDistances(m_BioTreeData.GetFeatureDict().HasFeature("dist"));

        if (m_BioTreeData.GetFeatureDict().HasFeature("label")) {
             m_Scheme->SetLabelFormat() = "$(label)";
        }
    }

    m_Scheme->SetTooltipFormat() = m_DataSource->GenerateTooltipFormat();
    m_PhyWidget->SetScheme(*m_Scheme);

    // datasource labeling (creating labels cache)
    // TODO: Integrate this with Init() to avoid 2x iteration
    m_DataSource->Relabel(m_Scheme->SetLabelFormat());

    x_RebuildObjectIndex();

    //!! HACK: this does layout, which needs labels which are set above.  
    if (first_update)
        m_PhyWidget->SetDataSourceNoUpdate(m_DataSource); 
    else
        m_PhyWidget->SetDataSource(m_DataSource);
}


void CPhyTreeView::x_RebuildObjectIndex()
{
    _ASSERT(m_DataSource);

    CScope* scope = x_GetScope();
    m_ObjectIndex.Clear(scope);

    TreeDepthFirstTraverse(*(m_DataSource->GetTree()),
        CPhyRebuildIndexVisitor(m_ObjectIndex));
}


void CPhyTreeView::GetSelection(TConstScopedObjects& objs) const
{
    if (m_PhyWidget && m_DataSource && x_HasProject()) {
        CScope* scope = x_GetScope();
        
        CPhyloTreeDataSource * ds = 
            const_cast<CPhyloTreeDataSource *>(m_DataSource.GetPointer());
        
        if (ds->GetCurrentNode()){
            CBioTreeDynamic   btd;            
            ds->Save(btd, ds->GetCurrentNode());

            m_Cont.Reset(new CBioTreeContainer);
            BioTreeConvert2Container(*m_Cont, btd);            
            
            m_Cont->SetTreetype("Phylogenetic Tree");
            objs.push_back(SConstScopedObject(*m_Cont, *scope));    
        }

        vector<CPhyloTreeNode*>  sel_nodes = ds->GetSelectedNodes();


        
        //if (ds->GetDictionary().HasFeature("seq-id")) {
            ITERATE(vector<CPhyloTreeNode*>, it, sel_nodes) {
                if (!(***it).GetSeqID().Empty()) {
                    //  evt.AddObjectSelection((***it).GetSeqID().GetObject());
                    objs.push_back(SConstScopedObject(*(***it).GetSeqID().GetPointer(), *scope));
                }
            }
        //}
        /// Need to complete support for taxid collection class CTaxIdSelSet with
        /// CLabel first.
        /*
        else {
            TBioTreeFeatureId feature_id;
            if (ds->GetDictionary().HasFeature("tax-id")) {
                TBioTreeFeatureId feature_id = ds->GetDictionary().GetId("tax-id");

                if (feature_id != -1) {
                    CRef<CTaxIdSelSet> sel_tax_ids(new CTaxIdSelSet());
                    int sel_count  = 0;

                    ITERATE(vector<CPhyloTreeNode*>, it, sel_nodes) {
                        const string& tax_id = (***it).GetBioTreeFeatureList()[feature_id];
                        if (tax_id != "") {
                            CTaxIdSelSet::TTaxId id = (CTaxIdSelSet::TTaxId)NStr::StringToUInt(tax_id);
                            sel_tax_ids->SelectTaxId(id);
                            ++sel_count;
                        }
                    }

                    if (sel_count > 0)
                        objs.push_back(SConstScopedObject(sel_tax_ids.GetObject(), *scope));    
                }
            }
        }
        */
    } 
}


void CPhyTreeView::GetSelection(CSelectionEvent& evt) const
{
    if (m_PhyWidget && m_DataSource && x_HasProject()) {
        CPhyloTreeDataSource * ds = 
            const_cast<CPhyloTreeDataSource *>(m_DataSource.GetPointer());

        vector<CPhyloTreeNode*>  sel_nodes = ds->GetSelectedNodes();

        if (ds->GetDictionary().HasFeature("seq-id")) {
            TConstObjects objs;
            ITERATE(vector<CPhyloTreeNode*>, it, sel_nodes) {
                if (!(***it).GetSeqID().Empty()) {
                    //  evt.AddObjectSelection((***it).GetSeqID().GetObject());
                    objs.push_back(CConstRef<CObject>((***it).GetSeqID().GetPointer()));
                }
            }
            evt.AddObjectSelection(objs);        
        }
        else if (ds->GetDictionary().HasFeature("tax-id")) {
            TBioTreeFeatureId feature_id = ds->GetDictionary().GetId("tax-id");

            ITERATE(vector<CPhyloTreeNode*>, it, sel_nodes) {
                const string& tax_id = (***it).GetBioTreeFeatureList()[feature_id];
                if (tax_id != "") {
                    CTaxIdSelSet::TTaxId id = (CTaxIdSelSet::TTaxId)NStr::StringToUInt(tax_id);
                    evt.AddTaxIDSelection(id);
                }
            }
        }    
    }    
}

class visitor_taxid_query
{
public:
    visitor_taxid_query(const CSelectionEvent::TTaxIds& tids,
                        TBioTreeFeatureId tax_feat_id) 
        : m_TaxIds(tids) 
        , m_TaxFeatureId(tax_feat_id) {}
    
    ETreeTraverseCode operator()( CPhyloTreeNode &  tree_node, int delta)
    {
        if (delta==1 || delta==0){
            const string& tax_id = (*tree_node).GetBioTreeFeatureList()[m_TaxFeatureId];
            if (tax_id != "") {
                CTaxIdSelSet::TTaxId id = (CTaxIdSelSet::TTaxId)NStr::StringToUInt(tax_id);
                if (m_TaxIds.IsSelected(id)) {
                    m_SelNodes.push_back(&tree_node);
                }
            }   
        }
        return eTreeTraverse;
    }

    visitor_taxid_query& operator=(const visitor_taxid_query& rhs) 
        { m_SelNodes = rhs.m_SelNodes; return *this; }

    const vector<CPhyloTreeNode*>& GetNodes() { return m_SelNodes; }
private:
    vector <CPhyloTreeNode*> m_SelNodes;

    const CSelectionEvent::TTaxIds& m_TaxIds;
    TBioTreeFeatureId m_TaxFeatureId;
};


// handles incoming selection broadcast (overriding CView virtual function)
void CPhyTreeView::x_OnSetSelection(CSelectionEvent& evt)
{
    vector<CPhyloTreeNode*> sel_nodes;

    if(evt.HasObjectSelection())    {
        vector< CConstRef<CSeq_id> > sel_ids;

        // get Ids
        CSelectionEvent::TIds  ids = evt.GetIds();

        // get Ids from TIdLocs
        const CSelectionEvent::TIdLocs&  locs = evt.GetIdLocs();
        ITERATE(CSelectionEvent::TIdLocs, it_loc, locs)   {
            ids.push_back((*it_loc)->m_Id);
        }

        if (ids.size() != 0) {

            CObjectIndex::TResults results;

            ITERATE(CSelectionEvent::TIds, it_id, ids) {
                const CSeq_id& id = **it_id;
                CSeq_id_Descr descr(NULL, id, evt.GetScope());    // TODO - put descr in SelectionEvent

                results.clear();
                m_ObjectIndex.GetMatches(descr, results);

                ITERATE(CObjectIndex::TResults, it_r, results)  {
                    const CSelNodeHandle* handle =
                        reinterpret_cast<const CSelNodeHandle*>(*it_r);
                    sel_nodes.push_back(handle->GetNode());
                }
            }
        }
        else {            
            if (m_PhyWidget && m_DataSource && x_HasProject()) {
                CPhyloTreeDataSource * ds = 
                    const_cast<CPhyloTreeDataSource *>(m_DataSource.GetPointer());

                if (ds->GetDictionary().HasFeature("tax-id")) {

                    const CSelectionEvent::TTaxIds& tids = evt.GetTaxIDs();
                    TBioTreeFeatureId feature_id = ds->GetDictionary().GetId("tax-id");
                    if (feature_id != -1) {
                        visitor_taxid_query vtq(tids, feature_id);
                        vtq = TreeDepthFirstTraverse(*ds->GetTree(), vtq);
                        sel_nodes =  vtq.GetNodes();
                    }
                }
            }
        }
    }

    m_DataSource->SetSelection(sel_nodes);
    m_PhyWidget->OnUpdateSelChanged();    
}


void CPhyTreeView::GetVisibleRanges(CVisibleRange& vrange) const
{
    /// collect information
    /*
    const CGlPane& pane = m_PhyWidget->GetPort();
    TModelRect rect = pane.GetVisibleRect();

    CRef<CSeq_loc> loc(new CSeq_loc());
    loc->SetInt().SetFrom((TSeqPos)rect.Left());
    loc->SetInt().SetTo  ((TSeqPos)rect.Right());
    loc->SetId(*m_DataSource->GetSubjectId().GetSeqId());
    evt.AddLocation(loc);

    loc.Reset(new CSeq_loc());
    loc->SetInt().SetFrom((TSeqPos)rect.Bottom());
    loc->SetInt().SetTo  ((TSeqPos)rect.Top());
    loc->SetId(*m_DataSource->GetQueryId().GetSeqId());
    evt.AddLocation(loc);
    */
}


void CPhyTreeView::OnVisibleRangeChanged(const CVisibleRange& vrange, IVisibleRangeClient* source)
{
    /*
    const CGlPane& pane = m_PhyWidget->GetPort();
    TModelRect rect = pane.GetVisibleRect();

    if (event.Match(*m_DataSource->GetQueryId().GetSeqId())) {
        TSeqRange range;
        range.SetFrom((TSeqPos)rect.Bottom());
        range.SetTo  ((TSeqPos)rect.Top());
        if (event.Clamp(m_DataSource->GetQueryHandle(), range)) {
            m_PhyWidget->SetQueryVisibleRange(range);
        }
    }

    if (event.Match(*m_DataSource->GetSubjectId().GetSeqId())) {
        TSeqRange range;
        range.SetFrom((TSeqPos)rect.Left());
        range.SetTo  ((TSeqPos)rect.Right());
        if (event.Clamp(m_DataSource->GetSubjectHandle(), range)) {
            m_PhyWidget->SetSubjectVisibleRange(range);
        }
    }
    */
}


void CPhyTreeView::SetRegistryPath(const string& reg_path)
{
    IRegSettings* reg_set = dynamic_cast<IRegSettings*>(m_PhyWidget);
    if (reg_set)
        reg_set->SetRegistryPath(reg_path);
}

void CPhyTreeView::LoadSettings()
{
    IRegSettings* reg_set = dynamic_cast<IRegSettings*>(m_PhyWidget);
    if (reg_set)
        reg_set->LoadSettings();
}


void CPhyTreeView::SaveSettings() const
{
    IRegSettings* reg_set = dynamic_cast<IRegSettings*>(m_PhyWidget);
    if (reg_set)
        reg_set->SaveSettings();
}


const CObject* CPhyTreeView::x_GetOrigObject() const
{
    if(m_BioTreeContainer)    {
        return m_BioTreeContainer.GetPointer();
    }
    return NULL;
}

void CPhyTreeView::OnCreateSubtree(CEvent* evt)
{
	CRef <objects::CBioTreeContainer> btc(new objects::CBioTreeContainer());
    CBioTreeDynamic          btd;

    m_DataSource->ConvertSelectionToBiotree(btd);
    BioTreeConvert2Container(*btc, btd);

	CIRef<CProjectService> srv = m_Workbench->GetServiceByType<CProjectService>();
	CDataLoadingAppJob *   job = new CPhyloTreeJob(srv.GetPointer(), btc.GetPointer());
	
    CRef<CWorkspaceConstPrx> workspace = srv->GetWorkspaceConst();
	CRef<CProjectPrx> project = workspace->FindProjectByScope(x_GetScope());
	ncbi::CDataLoadingOptions::TProjectId prj_id = project->GetId();

	CDataLoadingOptions& options = job->GetOptions();
	options.Set_AddToExistingProject(prj_id);

	CDataLoadingAppTask* task = new CDataLoadingAppTask(*job);
    m_Workbench->GetAppTaskService()->AddTask(*task);

}

void CPhyTreeView::OnSomethingEdited(CEvent* evt)
{        
    CBioTreeDynamic   btd;
    m_DataSource->Save(btd);

    // Commit here since the the undomanager->execute command also calls 'project
    // updated' which propogates changes from the biotree container to the 
    // CPhyloTree.  We should change this in the future so that the edits update
    // the project (biotreecontainer) directly.
    CommitTheChanges(btd);

    
    // This uses up too much memory - we need to update this to only
    // save the deltas.
    CChangePhyloTreeCommand* phyCommand = new CChangePhyloTreeCommand(
                                          this,
                                          m_BioTreeData,
                                          btd);

    CIRef<IEditCommand> command(phyCommand);
    
     CIRef<CProjectService> srv = m_Workbench->GetServiceByType<CProjectService>();
    //_ASSERT(srv);

    CRef<CWorkspaceConstPrx> workspace = srv->GetWorkspaceConst();
    CRef<CProjectPrx> project = workspace->FindProjectByScope(x_GetScope());

    phyCommand->EnableUpdates(false);
    project->GetDocument()->GetUndoManager().Execute(command);
    phyCommand->EnableUpdates(true);
    

    // TODO: This is a bad thing to do, as commit operation should be invoked
    // only at:
    // 1) close of view
    // 2) export operations
    // 3) new tree from original creation
    // 4) saving project containing this biotree
    //CommitTheChanges(btd);    
 }

void CPhyTreeView::CommitTheChanges(CBioTreeDynamic & btd)
{
    // no alternative for this; special command to update project item may be better
    // then dropping const
    CBioTreeContainer * inProject = 
        const_cast<CBioTreeContainer *> (m_BioTreeContainer.GetPointer());

    CBioTreeContainer_Base::TTreetype tt;
    bool has_tt = inProject->IsSetTreetype();
   
    if (has_tt)
        tt = inProject->GetTreetype();
   
    inProject->Reset();

    BioTreeConvert2Container(*inProject, btd);

    if (has_tt)
        inProject->SetTreetype(tt);
}


///////////////////////////////////////////////////////////////////////////////
/// CPhyTreeViewFactory
string  CPhyTreeViewFactory::GetExtensionIdentifier() const
{
    static string sid("phy_tree_view_factory");
    return sid;
}


string CPhyTreeViewFactory::GetExtensionLabel() const
{
    static string slabel("Phylogenetic Tree View Factory");
    return slabel;
}


void CPhyTreeViewFactory::RegisterIconAliases(wxFileArtProvider& provider)
{
    string alias = GetViewTypeDescriptor().GetIconAlias();
    provider.RegisterFileAlias(ToWxString(alias), wxT("phylo_tree_view.png"));
}

void CPhyTreeViewFactory::RegisterCommands (CUICommandRegistry& cmd_reg, wxFileArtProvider& provider)
{
    CPhyloTreeWidget::RegisterCommands(cmd_reg, provider);
}

const CProjectViewTypeDescriptor& CPhyTreeViewFactory::GetProjectViewTypeDescriptor() const
{
    return s_PhyTreeViewTypeDescr;
}


IView* CPhyTreeViewFactory::CreateInstance() const
{
    return new CPhyTreeView();
}


IView* CPhyTreeViewFactory::CreateInstanceByFingerprint(const TFingerprint& fingerprint) const
{
    /*TFingerprint print(CPhyTreeView::m_TypeDescr.GetLabel(), false);
    if(print == fingerprint)   {
        return new CPhyTreeView();
    }*/
    return NULL;
}



int CPhyTreeViewFactory::TestInputObjects(TConstScopedObjects& objects)
{
    bool found_good = false;
    bool found_bad = false;
    for(  size_t i = 0;  i < objects.size();  i++)  {
        const CObject* obj = objects[i].object;
        const type_info& type = typeid(*obj);
        if(typeid(CBioTreeContainer) == type)  {
            found_good = true;
        } else {
            found_bad = true;
        }
    }
    if(found_good)   {
        return fCanShowSeparated | (found_bad ? fCanShowSome : fCanShowAll);
    }
    return 0; // can show nothing
}


CPhyloTreeJob::CPhyloTreeJob(CProjectService* service,
	const objects::CBioTreeContainer* btc): CDataLoadingAppJob(service)
{
    CFastMutexGuard lock(m_Mutex);
	m_Btc.Reset(const_cast<objects::CBioTreeContainer*>(btc)); 
	m_Descr = "Phylogenetic Tree Cutter Tool";
}


void CPhyloTreeJob::x_CreateProjectItems()
{   
	CRef<objects::CProjectItem> item(new objects::CProjectItem());

	string label = "";
    CLabel::GetLabel(*m_Btc, &label, CLabel::eDefault, m_Scope.GetPointer());
    
	item->SetItem().SetOther().Set(*m_Btc);
    item->SetLabel(label);
    m_Items.push_back(item);
}


END_NCBI_SCOPE
