/*  $Id: text_align_object_loader.cpp 27549 2013-03-04 17:18:33Z katargir $
* ===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data,  the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties,  express or implied,  including
*  warranties of performance,  merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================
*
* Authors:  Mike DiCuccio, Roman Katargin
*/


#include <ncbi_pch.hpp>

#include <util/icanceled.hpp>

#include <gui/packages/pkg_alignment/text_align_object_loader.hpp>

#include <objtools/readers/aln_reader.hpp>
#include <objtools/readers/reader_exception.hpp>

#include <util/format_guess.hpp>

#include <gui/widgets/wx/compressed_file.hpp>
#include <gui/widgets/wx/wx_utils.hpp>

#include <wx/filename.h>
#include <wx/sstream.h>
#include <wx/txtstrm.h>

#include <gui/core/job_results_dlg.hpp>

BEGIN_NCBI_SCOPE

CTextAlignObjectLoader::CTextAlignObjectLoader(const CTextAlignParams& params, const vector<wxString>& filenames)
 : m_Params(params), m_FileNames(filenames)
{
}

IObjectLoader::TObjects& CTextAlignObjectLoader::GetObjects()
{
    return m_Objects;
}

string CTextAlignObjectLoader::GetDescription() const
{
    return "Loading Text Alignment Files";
}

bool CTextAlignObjectLoader::PreExecute()
{
    return true;
}

string s_FormatErrors (const CAlnReader::TErrorList& errors)
{
    string errmsg = "";

    ITERATE (CAlnReader::TErrorList, iter2, errors) {
        const string& id = (*iter2).GetID();
        int line_num = (*iter2).GetLineNum();

        if (!NStr::IsBlank (errmsg)) {
            errmsg += "\n";
        }
        if (line_num > -1) {
            errmsg += "At line ";
            errmsg += NStr::IntToString(line_num);
        }
        if (!NStr::IsBlank (id)) {
            errmsg += "(Sequence ID ";
            errmsg += id;
            errmsg += ") ";
        }
        errmsg += (*iter2).GetMsg();
    }
    return errmsg;
}

bool CTextAlignObjectLoader::Execute(ICanceled& canceled)
{
    string unknown = ToStdString(m_Params.GetUnknown());
    string match = ToStdString(m_Params.GetMatch());
    string gapbegin = ToStdString(m_Params.GetBegin());
    string gapmiddle = ToStdString(m_Params.GetMiddle());
    string gapend = ToStdString(m_Params.GetEnd());

    ITERATE(vector<wxString>, it, m_FileNames) {
        if (canceled.IsCanceled())
            return false;

        const wxString& fn = *it;
        wxFileName fileName(fn);
        string annot_name = "File: " + ToStdString(fileName.GetFullName());

        try {
            CCompressedFile file(fn);

            CRef<objects::CSeq_entry> entry;

            string errmsg = "";

            vector<CAlnReader::EAlphabet> alphas;
            if (m_Params.GetSeqType() == 1) {
                alphas.push_back(CAlnReader::eAlpha_Nucleotide);
            } else if (m_Params.GetSeqType() == 2) {
                alphas.push_back(CAlnReader::eAlpha_Protein);
            } else {
                alphas.push_back(CAlnReader::eAlpha_Nucleotide);
                alphas.push_back(CAlnReader::eAlpha_Protein);
            }

            ITERATE (vector<CAlnReader::EAlphabet>, iter, alphas) {
                CAlnReader reader(file.GetIstream());
                try {
                    reader.SetMissing(unknown);
                    reader.SetMatch(match);
                    reader.SetBeginningGap(gapbegin);
                    reader.SetMiddleGap(gapmiddle);
                    reader.SetEndGap(gapend);
                    reader.SetAlphabet(*iter);

                    reader.Read();
                    entry = reader.GetSeqEntry();
                    errmsg = s_FormatErrors(reader.GetErrorList());
                    break;
                }
                catch (CObjReaderParseException& _DEBUG_ARG(e)) {
                    _TRACE("clustal nucleotide failed: " << e.what());
                    errmsg = s_FormatErrors (reader.GetErrorList());
                }
            }

            if (entry) {
                m_Objects.push_back(SObject(*entry, annot_name));
            } else {
                wxStringOutputStream strstrm;
                wxTextOutputStream os(strstrm);

                os << "Failed to read alignment<br>";
                os << annot_name;
                if (!NStr::IsBlank(errmsg)) {
                    os << "<br>";
                    os << errmsg;
                }
                m_HtmlErrors += strstrm.GetString();
            }
        }
        catch (const CException&) {
        }
        catch (const exception&) {
        }
    }

    return true;
}

bool CTextAlignObjectLoader::PostExecute()
{
    if (!m_HtmlErrors.empty()) {
        CJobResultsDlg dlg;
        dlg.SetHTML(m_HtmlErrors);
        dlg.Create(NULL, wxID_ANY, wxT("Text Alignment import results"));
        dlg.ShowModal();
    }

    return true;
}

END_NCBI_SCOPE
