/*  $Id: 5col_object_loader.cpp 27061 2012-12-13 15:49:53Z katargir $
* ===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data,  the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties,  express or implied,  including
*  warranties of performance,  merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================
*
* Authors:  Roman Katargin
*/


#include <ncbi_pch.hpp>

#include <corelib/ncbifile.hpp>

#include <util/icanceled.hpp>

#include <gui/packages/pkg_sequence/5col_object_loader.hpp>

#include <objects/seqset/Seq_entry.hpp>
#include <objtools/readers/readfeat.hpp>
#include <objtools/readers/error_container.hpp>

#include <objects/seq/Seq_annot.hpp>

#include <gui/widgets/wx/compressed_file.hpp>
#include <gui/widgets/wx/wx_utils.hpp>

#include <wx/filename.h>
#include <wx/sstream.h>
#include <wx/txtstrm.h>

#include <gui/core/job_results_dlg.hpp>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

namespace {
    class CErrorContainer : public CErrorContainerLenient
    {
    public:
        CErrorContainer(size_t maxErrors) : m_MaxErrors(maxErrors) {}
        virtual bool PutError(const ILineError& err)
        {
            if (Count() < m_MaxErrors) {
                CErrorContainerLenient::PutError(err);
            }
            return true;
        }
    private:
        const size_t m_MaxErrors;
    };
}

C5ColObjectLoader::C5ColObjectLoader(const vector<wxString>& filenames)
 : m_FileNames(filenames)
{
}

IObjectLoader::TObjects& C5ColObjectLoader::GetObjects()
{
    return m_Objects;
}

string C5ColObjectLoader::GetDescription() const
{
    return "Loading 5 Column Feature Files";
}

bool C5ColObjectLoader::PreExecute()
{
    return true;
}

bool C5ColObjectLoader::Execute(ICanceled& canceled)
{
    ITERATE(vector<wxString>, it, m_FileNames) {
        if (canceled.IsCanceled())
            return false;

        CRef<CErrorContainer> errCont(new CErrorContainer(100));
        const wxString& fn = *it;

        wxFileName fileName(fn);
        string annot_name = "File: " + ToStdString(fileName.GetFullName());

        try {

            CCompressedFile file(fn);
            CRef<CSeq_annot> seq_annot = CFeature_table_reader::ReadSequinFeatureTable (file.GetIstream(), 0, errCont.GetPointer());

            if (canceled.IsCanceled())
                return false;
            m_Objects.push_back(SObject(*seq_annot, annot_name));
        }
        catch (const CException&) {
        }
        catch (const exception&) {
        }

        if (errCont->Count() > 0) {
            wxStringOutputStream strstrm;
            wxTextOutputStream os(strstrm);

            os << wxT("<h3>File: ") << annot_name << wxT("</h3>");

            for (size_t i = 0; i < errCont->Count(); ++i) {
                const ILineError& lerror = errCont->GetError(i);
                os << wxT("Line ") << lerror.Line() << wxT(". ") << ToWxString(lerror.Message()) << wxT("<br />");
            }

            m_HtmlErrors += strstrm.GetString();
        }
    }

    return true;
}

bool C5ColObjectLoader::PostExecute()
{
    if (!m_HtmlErrors.empty()) {
        CJobResultsDlg dlg;
        dlg.SetHTML(m_HtmlErrors);
        dlg.Create(NULL, wxID_ANY, wxT("5 Column Feature import results"));
        dlg.ShowModal();
    }

    return true;
}

END_NCBI_SCOPE
