/*  $Id: entrez_search_tool.cpp 27924 2013-04-25 17:37:02Z voronov $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Andrey Yazhuk
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>

#include "entrez_search_tool.hpp"
#include <gui/core/data_mining_service.hpp>

#include <gui/objutils/label.hpp>

#include <gui/objutils/entrez_conversion.hpp>
#include <gui/widgets/wx/ui_tool_registry.hpp>

#include <gui/utils/extension_impl.hpp>

#include <objects/entrez2/entrez2_client.hpp>

#include <objmgr/object_manager.hpp>

#include <wx/sizer.h>
#include <wx/stattext.h>
#include <wx/choice.h>
#include <wx/srchctrl.h>

#include <gui/widgets/wx/wx_utils.hpp>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

#define ID_COMBOBOX 11003
#define ID_TEXT     11414

///////////////////////////////////////////////////////////////////////////////
/// CDocsumTableModel

typedef pair<string,string> TStringPair;

class CDocsumTableModel : public CObjectListTableModel
{
public:
    CDocsumTableModel( const string& aDbName );

    virtual int GetNumExtraColumns() const;
    virtual wxString GetExtraColumnName( int col ) const;
    virtual wxVariant GetExtraValueAt( int row, int col ) const;

    virtual wxString GetImageAlias( int row ) const;

protected:
    string m_DbName;

    /// maps doc summary field names to columns
    vector<TStringPair> m_Fields;
};


///////////////////////////////////////////////////////////////////////////////
/// IDMSearchTool
CEntrezSearchTool::CEntrezSearchTool()
{
}


IUITool* CEntrezSearchTool::Clone() const
{
    return new CEntrezSearchTool();
}


string CEntrezSearchTool::GetName() const
{
    static string s_name("Search NCBI Public Databases");
    return s_name;
}


string CEntrezSearchTool::GetDescription() const
{
    return "";
}


CIRef<IDMSearchForm> CEntrezSearchTool::CreateSearchForm()
{
    CIRef<IDMSearchForm> form(new CEntrezSearchForm(*this));
    return form;
}


bool CEntrezSearchTool::IsCompatible(IDataMiningContext* context)
{
    return false;
}


CRef<CSearchJobBase> CEntrezSearchTool::x_CreateJob(IDMSearchQuery& query)
{
    CRef<CSearchJobBase> job;
    CEntrezSearchQuery* e_query = dynamic_cast<CEntrezSearchQuery*>(&query);
    if(e_query) {
        job.Reset(new CEntrezSearchJob(*e_query));
    }
    return job;
}

string  CEntrezSearchTool::GetExtensionIdentifier() const
{
    return "search_tool::entrez_search_tool";
}

string  CEntrezSearchTool::GetExtensionLabel() const
{
    return "Datamining Tool - Entrez Search search";
}


///////////////////////////////////////////////////////////////////////////////
/// CEntrezSearchTool
CEntrezSearchQuery::CEntrezSearchQuery(const string& terms, const string& db_name)
:   m_Terms(terms),
    m_DbName(db_name)
{
}


///////////////////////////////////////////////////////////////////////////////
/// CEntrezSearchForm

CEntrezSearchForm::CEntrezSearchForm(CEntrezSearchTool& tool)
:   m_Tool(&tool),
    m_CurrDbName("Entrez Gene"),
    m_DbCombo(NULL)
{
}


void CEntrezSearchForm::Create()
{
}


void CEntrezSearchForm::Init()
{
    CSearchFormBase::Init();

    m_DbNames.clear();
    CEntrezDB::GetDbNames(m_DbNames);

    vector<string> vis_names;
    string curr_vis_name;
    ITERATE(TNamePairs, it, m_DbNames)  {
///!!!!!
        
    }
    if(curr_vis_name.empty())   {
        // invalid database name - reset
        m_CurrDbName = m_DbNames.begin()->second;
    }
}



static const char* kDatabaseTag = "Database";


void CEntrezSearchForm::x_LoadSettings(CGuiRegistry::TReadView& view)
{
    
    CSearchFormBase::x_LoadSettings(view);
    m_CurrDbName = view.GetString(kDatabaseTag);
}


void CEntrezSearchForm::x_SaveSettings(CGuiRegistry::TReadWriteView& view) const
{
    
    CSearchFormBase::x_SaveSettings(view);

    if (m_DbCombo) {        
        view.Set(kDatabaseTag, ToStdString( m_DbCombo->GetStringSelection() ));
    }
}


wxSizer* CEntrezSearchForm::GetWidget(wxWindow * parent)
{
    if ( !m_Sizer) {        
        wxFlexGridSizer * sz = new wxFlexGridSizer(1, 3, 0, 0);     
        sz->AddGrowableCol(2);
        m_Sizer = sz;
        
        m_Sizer->Add(new wxStaticText( parent, wxID_STATIC,
                                       wxT("Select NCBI Database:"),
                                       wxDefaultPosition, wxDefaultSize, 0 ),
            0, wxALIGN_CENTER_VERTICAL|wxALL, 5);   
        
        m_DbCombo = new wxChoice(parent, ID_COMBOBOX,
                                   wxDefaultPosition, wxDefaultSize,
                                   0, (const wxString*)NULL);

        m_Sizer->Add(m_DbCombo,1, wxGROW|wxALIGN_CENTER_VERTICAL|wxALL, 5);         
    
        m_Text = new CSearchControl(parent, ID_TEXT, wxT(""),
                                    wxDefaultPosition, wxDefaultSize,
                                    wxTE_PROCESS_ENTER );                       
        m_Sizer->Add(m_Text,1, wxGROW|wxALIGN_CENTER_VERTICAL|wxALL, 5);        
        //m_Text->SetAutoOff();
    }
    return m_Sizer;
}

void  CEntrezSearchForm::Update()
{
    m_DbCombo->Clear();

    m_DbNames.clear();
    CEntrezDB::GetDbNames(m_DbNames);
    ITERATE(TNamePairs, it, m_DbNames)  {
        m_DbCombo->Append(ToWxString(it->second), (void*)it->first.c_str());            
    }

    if ( !m_CurrDbName.empty() ) {
        m_DbCombo->SetStringSelection(ToWxString(m_CurrDbName));
        int sel = m_DbCombo->GetSelection();
        if (sel == wxNOT_FOUND) {
            m_DbCombo->Select(0);
        }
    } else {
        m_DbCombo->Select(0);
    }
    m_CurrDbName = ToStdString(m_DbCombo->GetStringSelection());
}

void CEntrezSearchForm::UpdateContexts()
{
    CSearchFormBase::UpdateContexts();
    // context - independent
    if (m_Controller) m_Controller->OnSearchEnabled(true);
}


CIRef<IDMSearchQuery> CEntrezSearchForm::ConstructQuery()
{
    const char * dbname =
        (const char*)m_DbCombo->GetClientData(m_DbCombo->GetSelection());

    string q_s = ToStdString(m_Text->GetValue());
    CIRef<IDMSearchQuery> ref(new CEntrezSearchQuery(q_s, dbname));
    return ref;
}


///////////////////////////////////////////////////////////////////////////////
/// CEntrezSearchJob

CEntrezSearchJob::CEntrezSearchJob(CEntrezSearchQuery& query)
:   m_Query(&query)
{
    string vis_db_name = CEntrezDB::GetVisibleName(m_Query->GetDbName());
    m_Descr = "Query: " + m_Query->GetTerms()
        + ", database = " + vis_db_name;
}


bool CEntrezSearchJob::x_ValidateParams()
{
    if(m_Query->GetTerms().empty()) {
        m_Error = new CAppJobError("Invalid input parameters - no search terms specified.");
        return false;
    }
    return true;
}


static int kMaxResults = 1000;

IAppJob::EJobState CEntrezSearchJob::x_DoSearch()
{
    // setup Entrez2 Client
    if(m_Client.GetPointer() == NULL)   {
        m_Client.Reset(new CEntrez2Client());
        string tool = "NCBI Genome Workbench";
        m_Client->SetDefaultRequest().SetTool(tool);
    }

    // prepare search params and search
    string terms = m_Query->GetTerms();
    string db_name = m_Query->GetDbName();
    size_t total_uids = 0;

    CRef<CEntrez2_docsum_list> list =
        CEntrezDB::Query(*m_Client, db_name, terms, total_uids, 0, kMaxResults, true);
    m_ResultsCount = (int)total_uids;
    m_MaxResultsCount = kMaxResults;


    CRef<CObjectManager> om(CObjectManager::GetInstance());
    CRef<CScope> scope(new CScope(*om));
    scope->AddDefaults();
    
    if(list && !IsCanceled())    {
        // process results
        CMutexGuard Guard(m_Mutex);

        CObjectList * obj_list = m_TempResult->GetObjectList();
        CEntrez2_docsum_list::TList& ls = list->SetList();
        NON_CONST_ITERATE(CEntrez2_docsum_list::TList, it, ls)    {
            CEntrez2_docsum& doc_sum = **it;
            obj_list->AddRow(&doc_sum, scope.GetPointer());
            if(IsCanceled()) {
                return eCanceled;
            }
        }
		return eCompleted;
    }
    return eCanceled;
}

CObjectListTableModel* CEntrezSearchJob::x_GetNewOLTModel() const
{
    return new CDocsumTableModel( m_Query->GetDbName() );
}


///////////////////////////////////////////////////////////////////////////////
/// CDocsumTableModel

static const TStringPair skGeneralFields[] = {
    TStringPair("Caption", "Label"),
    TStringPair("Title", "Description"),
    TStringPair("Extra", "FASTA IDs"),
    TStringPair("TaxId", "Taxonomic ID")
};

static const TStringPair skGeneFields[] = {
    TStringPair("Name", "Label"),
    TStringPair("Description", "Description"),
    TStringPair("Orgname", "Organism"),
    TStringPair("Chromosome", "Chromosome"),
    TStringPair("OtherAliases", "Aliases"),
    TStringPair("MapLocation", "Map Location")
};

static const TStringPair skGenomeFields[] = {
    TStringPair("Organism_Name", "Name"),
    TStringPair("Organism_Kingdom", "Kingdom"),
    TStringPair("Organism_Group", "Group"),
    TStringPair("Organism_Subgroup", "Subgroup"),
    TStringPair("Defline", "Defline"),
    TStringPair("Assembly_Name", "Assembly Name"),
    TStringPair("Assembly_Accession", "Assembly Accession")
};

CDocsumTableModel::CDocsumTableModel( const string& aDbName )
: m_DbName( aDbName )
{
    const TStringPair* ptr;
    int size;

    if( m_DbName == "gene" ){
        ptr = &skGeneFields[0];
        size = sizeof(skGeneFields);
    }
    else if( m_DbName == "genome" ){
        ptr = &skGenomeFields[0];
        size = sizeof(skGenomeFields);
    } else {
        ptr = &skGeneralFields[0];
        size = sizeof(skGeneralFields);
    }

    int num = size /sizeof(TStringPair);
    for( int i = 0; ptr && i < num; i++ ){
        m_Fields.push_back( *(ptr +i) );
    }
}

int CDocsumTableModel::GetNumExtraColumns() const
{
    return (int)m_Fields.size();
}

wxString CDocsumTableModel::GetExtraColumnName( int col ) const
{
    if( col < 0 || col >= GetNumExtraColumns() ){
        _ASSERT(false);
        NCBI_THROW(CException, eUnknown, "Invalid extra column index");
    }
    return ToWxString(m_Fields[col].second);
}

wxVariant CDocsumTableModel::GetExtraValueAt( int row, int col ) const
{
    if( col < 0 || col >= GetNumExtraColumns() ){
        _ASSERT(false);
        NCBI_THROW(CException, eUnknown, "Invalid extra column index");
    }

    const CObject* obj = m_ObjectList->GetObject( row );
    const CEntrez2_docsum* doc_sum = dynamic_cast<const CEntrez2_docsum*>(obj);
    if( doc_sum ){
        const string& field = m_Fields[col].first;
        return ToWxString( doc_sum->GetValue(field) );
    }

    return ToWxString( "" );
}

wxString CDocsumTableModel::GetImageAlias( int row ) const
{
    if( m_DbName == "gene" )        return wxT("symbol::feature");
    if( m_DbName == "protein" )     return wxT("symbol::sequence_protein");
    if( m_DbName == "nucleotide" )  return wxT("symbol::sequence_dna");
    if( m_DbName == "genome" )      return wxT("symbol::sequence");
    
    return wxT("");  
}


END_NCBI_SCOPE
