/*  $Id: gb_load_option_panel.cpp 27150 2012-12-31 16:07:52Z falkrb $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:
 *
 * File Description:
 *
 */


#include <ncbi_pch.hpp>

#include "gb_load_option_panel.hpp"

#include <objects/seqloc/Seq_id.hpp>
#include <gui/widgets/wx/wx_utils.hpp>
#include <gui/widgets/wx/async_call.hpp>
#include <gui/widgets/wx/message_box.hpp>
#include <objmgr/scope.hpp>
#include <objmgr/object_manager.hpp>
#include <objmgr/util/sequence.hpp>
#include <util/xregexp/regexp.hpp>

#include <wx/sizer.h>
#include <wx/stattext.h>
#include <wx/statbox.h>
#include <wx/icon.h>
#include <wx/bitmap.h>
#include <wx/msgdlg.h>

////@begin includes
////@end includes


BEGIN_NCBI_SCOPE
USING_SCOPE(objects);


IMPLEMENT_DYNAMIC_CLASS( CGenBankLoadOptionPanel, wxPanel )

BEGIN_EVENT_TABLE( CGenBankLoadOptionPanel, wxPanel )
////@begin CGenBankLoadOptionPanel event table entries
    EVT_TEXT( ID_ACC_INPUT, CGenBankLoadOptionPanel::OnAccInputUpdated )

////@end CGenBankLoadOptionPanel event table entries
END_EVENT_TABLE()


CGenBankLoadOptionPanel::CGenBankLoadOptionPanel()
{
    Init();
}

CGenBankLoadOptionPanel::CGenBankLoadOptionPanel( wxWindow* parent, wxWindowID id, const wxPoint& pos, const wxSize& size, long style )
{
    Init();
    Create(parent, id, pos, size, style);
}

bool CGenBankLoadOptionPanel::Create( wxWindow* parent, wxWindowID id, const wxPoint& pos, const wxSize& size, long style )
{
////@begin CGenBankLoadOptionPanel creation
    wxPanel::Create( parent, id, pos, size, style );

    CreateControls();
    if (GetSizer())
    {
        GetSizer()->SetSizeHints(this);
    }
    Centre();
////@end CGenBankLoadOptionPanel creation
    return true;
}


CGenBankLoadOptionPanel::~CGenBankLoadOptionPanel()
{
////@begin CGenBankLoadOptionPanel destruction
////@end CGenBankLoadOptionPanel destruction
}


void CGenBankLoadOptionPanel::Init()
{
////@begin CGenBankLoadOptionPanel member initialisation
    m_AccInput = NULL;
////@end CGenBankLoadOptionPanel member initialisation
    m_TokensNum = 0;
    m_ErrNum = 0;
}


void CGenBankLoadOptionPanel::CreateControls()
{
////@begin CGenBankLoadOptionPanel content construction
    CGenBankLoadOptionPanel* itemPanel1 = this;

    wxBoxSizer* itemBoxSizer2 = new wxBoxSizer(wxVERTICAL);
    itemPanel1->SetSizer(itemBoxSizer2);

    wxStaticBox* itemStaticBoxSizer3Static = new wxStaticBox(itemPanel1, wxID_ANY, _("Accessions to load"));
    wxStaticBoxSizer* itemStaticBoxSizer3 = new wxStaticBoxSizer(itemStaticBoxSizer3Static, wxVERTICAL);
    itemBoxSizer2->Add(itemStaticBoxSizer3, 1, wxGROW|wxALL, 5);

    wxStaticText* itemStaticText4 = new wxStaticText( itemPanel1, wxID_STATIC, _("Separate accessions with spaces, tabs, commas or semicolons.\nAdd a range to accession after colon (NT_029999:5k-10k)."), wxDefaultPosition, wxDefaultSize, 0 );
    itemStaticBoxSizer3->Add(itemStaticText4, 0, wxGROW|wxALL, 5);

    m_AccInput = new CRichTextCtrl( itemPanel1, ID_ACC_INPUT, wxEmptyString, wxDefaultPosition, wxSize(200, 50), wxWANTS_CHARS|wxTE_RICH2|wxTE_MULTILINE );
    itemStaticBoxSizer3->Add(m_AccInput, 1, wxGROW|wxALL, 5);

    itemBoxSizer2->Add(5, 5, 1, wxGROW|wxALL, 5);

////@end CGenBankLoadOptionPanel content construction
    m_DefStyle.SetBackgroundColour(m_AccInput->GetBackgroundColour());
    m_ErrStyle.SetBackgroundColour(wxColour(255, 200, 200));
    m_AccInput->SetDefaultStyle(m_DefStyle);
}


string CGenBankLoadOptionPanel::GetInput() const
{
    return ToStdString(m_AccInput->GetValue());
}


void CGenBankLoadOptionPanel::SetInput(const string& input)
{
    _ASSERT(m_AccInput);
    m_AccInput->SetValue(ToWxString(input));
}


#define GB_LOADER_ERR "GenBank - Error in input"

#define INVALID_ELEMS_MSG " elements are not valid sequence identifiers and so cannot be loaded.\
\nWould you like to proceed?"

#define INVALID_INPUT "Input does not contain valid sequence identifiers."

#define EMPTY_INPUT "Please specify one or more sequence identifiers to load!"


bool CGenBankLoadOptionPanel::IsInputValid()
{
    x_ValidateInput();
    if(m_TokensNum > 0  &&  m_ErrNum == 0)    {
        return true;
    } else {
        if(m_TokensNum == 0)    {
            wxMessageBox(wxT(EMPTY_INPUT), wxT(GB_LOADER_ERR),
                         wxOK | wxICON_ERROR, this);
            return false;
        } else if(m_TokensNum == m_ErrNum) {
            wxMessageBox(wxT(INVALID_INPUT), wxT(GB_LOADER_ERR),
                         wxOK | wxICON_ERROR, this);
            return false;
        } else {
            string text = NStr::IntToString(m_ErrNum) + INVALID_ELEMS_MSG;
            int res = wxMessageBox(ToWxString(text), wxT(GB_LOADER_ERR),
                                   wxYES_NO | wxICON_EXCLAMATION, this);
            return res == wxYES;
        }
    }
}


static string sDelim(" \t,;\n\r");
const char* sRangeRegexp = "[0-9,]+[kKmM]?([-:]|\\.\\.)[0-9,]+[kKmM]?";

static bool sParseRange(CSeq_interval& ival, const string& range)
{
    string tmp = range;
    NStr::ReplaceInPlace(tmp, ",", "");
    NStr::ReplaceInPlace(tmp, "k", "K");
    NStr::ReplaceInPlace(tmp, "m", "M");
    NStr::ReplaceInPlace(tmp, "K", "000");
    NStr::ReplaceInPlace(tmp, "M", "000000");

    list<string> pos;
    NStr::Split(tmp, "-:.", pos);
    if (pos.size() != 2)
        return false;

    long from = 0, to = 0;
    try {
        from = NStr::StringToLong(pos.front()) - 1;
        to   = NStr::StringToLong(pos.back()) - 1;
    } catch (CException&) {
        return false;
    }

    ival.SetFrom(from);
    ival.SetTo(to);

    return true;
}

class CValidateAccesions : public CAsyncCall::Call
{
public:
    class CToken
    {
    public:
        CToken() : m_Pos(0), m_Size(0), m_Valid(false) {}
        int m_Pos;
        int m_Size;
        bool m_Valid;
    };

    CValidateAccesions(const string& data) :
      Call(wxT("Validating accesssions...")), m_Data(data), m_Processed(0) {}

    const vector<CToken>& GetTokens() const { return m_Tokens; }
    const vector< CRef<CObject> >& GetIds() const { return m_Ids; }
    size_t GetProcessed() const { return m_Processed; }

    virtual bool CanCancel() const { return true; }
    virtual void Execute();

private:
    string m_Data;
    vector<CToken> m_Tokens;
    vector< CRef<CObject> > m_Ids;
    size_t m_Processed;
};

void CValidateAccesions::Execute()
{
    vector<string> tokens;
    vector<SIZE_TYPE> token_pos;
    NStr::Tokenize(m_Data, sDelim, tokens, NStr::eMergeDelims, &token_pos);
    size_t size = tokens.size();
    m_Tokens.resize(size);

    for( size_t i = 0;  i < size;  i++) {
        m_Tokens[i].m_Pos  = (int)token_pos[i];
        m_Tokens[i].m_Size = (int)tokens[i].size();
        m_Tokens[i].m_Valid = false;
    }

    CScope scope(*CObjectManager::GetInstance());
    scope.AddDefaults();

    for (m_Processed = 0;  m_Processed < size;  m_Processed++) {
        if (IsCanceled())
            break;

        string s_id = tokens[m_Processed];
        string range;
        size_t pos = s_id.find(':');

        bool seq_id_valid = true;

        if (pos != string::npos) {
            range = s_id.substr(pos + 1);
            s_id  = s_id.substr(0, pos);

            CRegexp re(sRangeRegexp);
            seq_id_valid = re.IsMatch(range);
        }

        if (seq_id_valid) {
            CSeq_id::EAccessionInfo ainfo = CSeq_id::IdentifyAccession(s_id);
            CSeq_id::E_Choice acc_type = CSeq_id::GetAccType(ainfo);
            switch( acc_type ){
            case CSeq_id::e_Local:
                seq_id_valid = false;
                break;

            default:
                try {
                    CSeq_id id( s_id );
                    CSeq_id_Handle idh = CSeq_id_Handle::GetHandle( id );
                    idh = sequence::GetId( idh, scope, sequence::eGetId_Best );

                    if( idh ){
                        CRef<CSeq_id> id_copy( new CSeq_id() );
                    
                        id_copy->Assign( *idh.GetSeqId() );

                        if (range.empty())
                            m_Ids.push_back( CRef<CObject>(id_copy.GetPointer()) );
                        else {
                            CRef<CSeq_loc> loc(new CSeq_loc());
                            CSeq_interval& ival = loc->SetInt();
                            ival.SetId().Assign(*id_copy);

                            if (sParseRange(ival, range))
                                m_Ids.push_back( CRef<CObject>(loc.GetPointer()) );
                            else
                                seq_id_valid = false;
                        }
                    } else {
                        seq_id_valid = false;
                    }

                } catch( CSeqIdException& ){
                    seq_id_valid = false;
                }
                break;
            }
        }

        m_Tokens[m_Processed].m_Valid = seq_id_valid;

        // Debug delay
        //::wxMilliSleep(100);
    }
}

void CGenBankLoadOptionPanel::x_ValidateInput()
{
    static bool fReentry = false;
    if (fReentry)
        return;
    CBoolGuard _guard(fReentry);

    string text = ToStdString(m_AccInput->GetValue());

    CValidateAccesions validator(text);
    CAsyncCall::Execute(validator);

    const vector<CValidateAccesions::CToken>& tokens = validator.GetTokens();
    m_TokensNum = (int)tokens.size();
    m_Ids = validator.GetIds();
    int processed = (int)validator.GetProcessed();
    m_ErrNum = m_TokensNum - processed;

    {{
        // gtk control loses kbd focus after Freeze/Thaw
        // restore it afterwards
        bool restoreFocus = (wxWindow::FindFocus() == m_AccInput);

        m_AccInput->Freeze();
        int valid_start = 0;
        for( size_t i = 0;  i < (size_t)processed;  i++) {
            if (!tokens[i].m_Valid) {
                ++m_ErrNum;
                int start = tokens[i].m_Pos;
                int end = start + tokens[i].m_Size;
                if (valid_start < start)
                    m_AccInput->SetStyle(valid_start, start, m_DefStyle);
                m_AccInput->SetStyle(start, end, m_ErrStyle);
                valid_start = end;
            }
        }
        if (processed < m_TokensNum) {
            int start = tokens[processed].m_Pos;
            int end = (int)text.size();
            if (valid_start < start)
                m_AccInput->SetStyle(valid_start, start, m_DefStyle);
            m_AccInput->SetStyle(start, end, m_ErrStyle);
        }
        else {
            if (valid_start < (int)text.size())
                m_AccInput->SetStyle(valid_start, (int)text.size(), m_DefStyle);
        }
        m_AccInput->Thaw();

        if (restoreFocus)
            m_AccInput->SetFocus();
    }}
}


CGenBankLoadOptionPanel::TIdsVec& CGenBankLoadOptionPanel::GetSeqIds()
{
    return m_Ids;
}

void CGenBankLoadOptionPanel::OnAccInputUpdated(wxCommandEvent& event)
{
    // Remove formatting that may be added on mac paste events (and
    // on mac wx 2.9.3 paste handler is not called)
    m_AccInput->RemoveFormatting();

    x_ValidateInput();
}


bool CGenBankLoadOptionPanel::ShowToolTips()
{
    return true;
}


wxBitmap CGenBankLoadOptionPanel::GetBitmapResource( const wxString& name )
{
    // Bitmap retrieval
////@begin CGenBankLoadOptionPanel bitmap retrieval
    wxUnusedVar(name);
    return wxNullBitmap;
////@end CGenBankLoadOptionPanel bitmap retrieval
}


wxIcon CGenBankLoadOptionPanel::GetIconResource( const wxString& name )
{
    // Icon retrieval
////@begin CGenBankLoadOptionPanel icon retrieval
    wxUnusedVar(name);
    return wxNullIcon;
////@end CGenBankLoadOptionPanel icon retrieval
}


END_NCBI_SCOPE

