/*  $Id: gb_ui_data_source.cpp 26938 2012-11-28 15:46:40Z katargir $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Andrey Yazhuk
 *
 */

#include <ncbi_pch.hpp>

#include <corelib/ncbiexec.hpp>
#include <corelib/ncbi_process.hpp>
#include <corelib/ncbi_system.hpp>

#include <gui/packages/pkg_sequence/gb_ui_data_source.hpp>

#include "gb_load_option_panel.hpp"

#include <gui/core/app_tasks.hpp>
#include <gui/core/app_explorer_service.hpp>
#include <gui/core/named_pipe.hpp>

#include <gui/framework/workbench.hpp>
#include <gui/framework/app_task_service.hpp>
#include <gui/framework/app_job_task.hpp>

#include <gui/widgets/wx/wx_utils.hpp>
#include <gui/widgets/wx/fileartprov.hpp>
#include <gui/widgets/wx/ui_command.hpp>
#include <gui/widgets/wx/message_box.hpp>
#include <gui/widgets/wx/wx_utils.hpp>
#include <gui/widgets/wx/sys_path.hpp>

#include <objmgr/object_manager.hpp>
#include <objtools/data_loaders/genbank/gbloader.hpp>
#include <connect/services/neticache_client.hpp>

#include <gui/objects/ProjectItem.hpp>
#include <gui/objutils/registry.hpp>
#include <gui/utils/extension_impl.hpp>

#include <sra/data_loaders/csra/csraloader.hpp>

#include <wx/menu.h>
#include <wx/filename.h>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

static const char* kGenBankLoadOption = "Data from GenBank";

///////////////////////////////////////////////////////////////////////////////
/// CGenBankDSEvtHandler - wxEvtHandler-derived adapter for GenBank data source.

class CGenBankDSEvtHandler : public wxEvtHandler
{
    DECLARE_EVENT_TABLE();
public:
    CGenBankDSEvtHandler(IWorkbench* workbench)
        :   m_Workbench(workbench)  {
    }
    void OnLoadFromGenBank(wxCommandEvent& event)
    {
        if(m_Workbench)    {
            COpenDlgTask* task = new COpenDlgTask(m_Workbench, kGenBankLoadOption);

            CAppTaskService* task_srv = m_Workbench->GetServiceByType<CAppTaskService>();
            task_srv->AddTask(*task);
        }
    }
protected:
    IWorkbench* m_Workbench;
};


BEGIN_EVENT_TABLE(CGenBankDSEvtHandler, wxEvtHandler)
    EVT_MENU(eCmdLoadFromGenBank, CGenBankDSEvtHandler::OnLoadFromGenBank)
END_EVENT_TABLE();


///////////////////////////////////////////////////////////////////////////////
/// CGenBankUIDataSource

static const char* kGB_DS_Icon = "icon::gb_data_source";

CGenBankUIDataSource::CGenBankUIDataSource(CGenBankUIDataSourceType& type)
:   m_Type(&type),
    m_Descr("GenBank", kGB_DS_Icon),
    m_SrvLocator(NULL),
    m_Open(false)
{
}

CGenBankUIDataSource::~CGenBankUIDataSource()
{
}


string CGenBankUIDataSource::GetExtensionIdentifier() const
{
    return "genbank_data_source";
}


string CGenBankUIDataSource::GetExtensionLabel() const
{
    return "GenBank Data Source";
}


void CGenBankUIDataSource::SetServiceLocator(IServiceLocator* locator)
{
    m_SrvLocator = locator;
}


IExplorerItemCmdContributor::TContribution
    CGenBankUIDataSource::GetMenu(TItemRefVector& items,
                                   CAppExplorerService& app_service)
{
    IExplorerItemCmdContributor::TContribution contrib;

    /// this is not a good solution, but simple
    IWorkbench* workbench = dynamic_cast<IWorkbench*>(m_SrvLocator);

    if(items.size() != 1  ||  workbench == NULL)   {
        return contrib; // return empty object - nothin to contribute
    }

    CUICommandRegistry& cmd_reg = CUICommandRegistry::GetInstance();

    CExplorerItem* item = items[0].GetPointer();
    int type = app_service.GetItemType(*item);

    if(type == CAppExplorerService::eDataSource)    {
        CAppExplorerService::TDataSourceTreeItem* ds_item =
            dynamic_cast<CAppExplorerService::TDataSourceTreeItem*>(item);

        if(ds_item) {
            CIRef<IUIDataSource> ds = ds_item->GetData();
            CGenBankUIDataSource* gb_ds = dynamic_cast<CGenBankUIDataSource*>(ds.GetPointer());

            if(gb_ds)  {
                wxMenu* menu = new wxMenu;
                menu->Append(wxID_SEPARATOR, wxT("Top Actions"));
                cmd_reg.AppendMenuItem(*menu, eCmdLoadFromGenBank);

                contrib.first = menu;
                contrib.second = new CGenBankDSEvtHandler(workbench);
            }
        }
    }
    return contrib;
}


IUIDataSourceType& CGenBankUIDataSource::GetType() const
{
    return *m_Type;
}


const IUIObject& CGenBankUIDataSource::GetDescr()
{
    return m_Descr;
}


bool CGenBankUIDataSource::IsOpen()
{
    return m_Open;
}

//static const char* kCmdExtPoint = "scoped_objects::cmd_contributor";
static const char* kAppExpCmdExtPoint = "project_tree_view::context_menu::item_cmd_contributor";

bool CGenBankUIDataSource::Open()
{
    if (m_Open) {
        LOG_POST(Error << "CGenBankUIDataSource::Open(): "
            "attempt to open already open data source");
        return false;
    }

    CStopWatch sw;
    sw.Start();

    /// register itself as menu contributor
    CIRef<IExtensionRegistry> reg = CExtensionRegistry::GetInstance();
    reg->AddExtension(kAppExpCmdExtPoint, *this);

    /// initialize the object manager
    CGuiRegistry& registry = CGuiRegistry::GetInstance();
    CGuiRegistry::TReadWriteView view =
        registry.GetReadWriteView("GBENCH.Services.ObjectManager");

    m_ObjMgr = CObjectManager::GetInstance();

    // general genbank loader options
    int priority  = view.GetInt("Priority", 99);

    // caching options
    string cache_path = FnToStdString(CSysPath::ResolvePath(wxT("<home>/cache")));
    wxString system_cache_path = FnToWxString(cache_path);
    if (sizeof(void*) == 8)
        system_cache_path.append(wxT("64"));

    {{
        /// make sure we can write in the cache path directory
        if ( !wxFileName::DirExists(system_cache_path) ) {
            if ( !wxFileName::Mkdir(system_cache_path) ) {
                LOG_POST(Error << "Error creating cache path ("
                    << system_cache_path.ToUTF8() << "); caching will be disabled");
            } else {
                LOG_POST(Info << "Created cache path: " << system_cache_path.ToUTF8());
            }
        } else {
            // clean sqlite cash if previous launch of GBench ended with crash
            if (registry.GetBool("GBENCH.Application.CrashDetected", false)) {
                ::wxRemoveFile(wxFileName(system_cache_path, wxT("cache_ids.db")).GetFullPath());
                ::wxRemoveFile(wxFileName(system_cache_path, wxT("cache_blobs.db")).GetFullPath());
            }
        }
    }}

    int cache_age = view.GetInt("CacheAge", 5);
    if (cache_age == 0) {
        cache_age = 5;   // keep objects for 5 days (default)
    }

    /// ID resolution time in hours
    int id_resolution_time = view.GetInt("IdResolutionTime", 24);
    if (id_resolution_time > 24 * 3) {
        id_resolution_time = 24 * 3; //  correct the unreasonable value
    }

    wxString objmgr_config_path =
        CSysPath::ResolvePathExisting(wxT("<home>/gbench-objmgr.ini, ")
                                      wxT("<std>/etc/gbench-objmgr.ini"));

    if (objmgr_config_path.empty()) {
        CGBDataLoader::RegisterInObjectManager(*m_ObjMgr);
    } else {
        CMemoryRegistry reg;
        CNcbiIfstream istr(objmgr_config_path.fn_str(), ios::binary|ios::in);
        reg.Read(istr);

        // disable cache in the second instance
        if (!CGBenchPipe::IsOpen()) { 
            reg.Set("genbank", "loader_method", "id2");
        } else {
            // override variables as needed
            reg.Set("genbank/cache/id_cache/sqlite3", "database",
                    cache_path + "/cache_ids.db");
            reg.Set("genbank/cache/id_cache/sqlite3", "cache_age",
                    NStr::IntToString(id_resolution_time * (60*60)));
            reg.Set("genbank/cache/blob_cache/sqlite3", "database",
                    cache_path + "/cache_blobs.db");
            reg.Set("genbank/cache/blob_cache/sqlite3", "cache_age",
                    NStr::IntToString(cache_age * (24*60*60)));
        }

        // now, create the GenBank data loader
        string msg;
        CGBDataLoader::TRegisterLoaderInfo rinfo;
        try {
            CConfig cfg(reg);
            rinfo =
                CGBDataLoader::RegisterInObjectManager
                    (*m_ObjMgr,
                     *cfg.GetTree(), CObjectManager::eDefault, priority);
        }
        catch (CException& e) {
            msg = "An error occurred while creating the GenBank connection\n(";
            msg += e.GetMsg();
            msg += ")";
        }
        catch (std::exception& e) {
            msg = "An error occurred while creating the GenBank connection\n(";
            msg += e.what();
            msg += ")";
        }

        if ( !msg.empty() ) {
            /// fall back to standard initialization
            LOG_POST(Error << msg);
            CGBDataLoader::RegisterInObjectManager(*CObjectManager::GetInstance());
        }
    }

    string t = NStr::DoubleToString(sw.Elapsed(), 3);
    LOG_POST(Info << "Registered GenBank Data Source  - " << t << " sec");

    CCSRADataLoader::RegisterInObjectManager(*CObjectManager::GetInstance(), CObjectManager::eDefault, 98);

    m_Open = true;
    return true;
}


bool CGenBankUIDataSource::Close()
{
    if (m_Open) {

        /// remove itself from menu contribution points
        CIRef<IExtensionRegistry> reg = CExtensionRegistry::GetInstance();
        reg->RemoveExtension(kAppExpCmdExtPoint, *this);

        try {
            bool revoked = m_ObjMgr->RevokeDataLoader("GBLOADER");
            if (!revoked) {
                CDataLoader* dl = m_ObjMgr->FindDataLoader("GBLOADER");
                CGBDataLoader* gbdl = dynamic_cast<CGBDataLoader*>(dl);
                if (gbdl) {
                    gbdl->CloseCache();
                }
            }
        } 
        catch(std::exception&)
        {
            LOG_POST(Error << "Cannot revoke genbank dataloader");
        }

        /// finally, drop the object manager
        CObjectManager* ptr = m_ObjMgr.Release();

        if (ptr) {
            if ( !ptr->ReferencedOnlyOnce() ) {
                LOG_POST(Error << "CGenBankUIDataSource::ShutDownService(): "
                         "object manager still referenced");
            }
        }

        m_Open = false;
        return true;
    }
    return false;
}


void CGenBankUIDataSource::EditProperties()
{
    //TODO
}


IUIToolManager* CGenBankUIDataSource::GetLoadManager()
{
    // TODO may need to link the manager to this particular datasource
    return new CGenBankUILoadManager();
}


///////////////////////////////////////////////////////////////////////////////
/// CGenBankUIDataSourceType
CGenBankUIDataSourceType::CGenBankUIDataSourceType()
:   m_Descr("GenBank Connection", "")
{
   wxFileArtProvider* provider = GetDefaultFileArtProvider();
   provider->RegisterFileAlias(ToWxString(kGB_DS_Icon),
                               wxT("gb_data_source.png"));
}


const IUIObject& CGenBankUIDataSourceType::GetDescr()
{
    return m_Descr;
}


IUIDataSource* CGenBankUIDataSourceType::CreateDataSource()
{
    return new CGenBankUIDataSource(*this);
}


bool CGenBankUIDataSourceType::AutoCreateDefaultDataSource()
{
    return true; // we want to create default "GenBank" datasource
}


string CGenBankUIDataSourceType::GetExtensionIdentifier() const
{
    static string ext_id("genbank_data_source_type");
    return ext_id;
}


string CGenBankUIDataSourceType::GetExtensionLabel() const
{
    return m_Descr.GetLabel();
}



///////////////////////////////////////////////////////////////////////////////
/// CGenBankUILoadManager
CGenBankUILoadManager::CGenBankUILoadManager()
:   m_SrvLocator(NULL),
    m_ParentWindow(NULL),
    m_Descriptor(kGenBankLoadOption, ""),
    m_State(eInvalid),
    m_OptionPanel(NULL),
    m_ProjectSelPanel(NULL)
{
}


void CGenBankUILoadManager::SetServiceLocator(IServiceLocator* srv_locator)
{
    m_SrvLocator = srv_locator;
}


void CGenBankUILoadManager::SetParentWindow(wxWindow* parent)
{
    m_ParentWindow = parent;
}


const IUIObject& CGenBankUILoadManager::GetDescriptor() const
{
    return m_Descriptor;
}


void CGenBankUILoadManager::InitUI()
{
    m_State = eSelectAcc;
}


void CGenBankUILoadManager::CleanUI()
{
    m_State = eInvalid;
    if(m_OptionPanel)   {
        m_SavedInput = m_OptionPanel->GetInput();
        m_OptionPanel = NULL; // window is destroyed by the system
    }
    m_ProjectSelPanel = NULL;
}


wxPanel* CGenBankUILoadManager::GetCurrentPanel()
{
    if(m_State == eSelectAcc)   {
        if(m_OptionPanel == NULL)   {
            m_OptionPanel = new CGenBankLoadOptionPanel(m_ParentWindow);
            m_OptionPanel->SetInput(m_SavedInput);
        }
        return m_OptionPanel;
    }  else if(m_State == eSelectProject)   {
        if(m_ProjectSelPanel == NULL)   {
            CIRef<CProjectService> srv = m_SrvLocator->GetServiceByType<CProjectService>();

            m_ProjectSelPanel = new CProjectSelectorPanel(m_ParentWindow);
            m_ProjectSelPanel->SetProjectService(srv);
            m_ProjectSelPanel->SetParams(m_ProjectParams);
            m_ProjectSelPanel->TransferDataToWindow();
        }
        return m_ProjectSelPanel;
    }
    return NULL;
}


bool CGenBankUILoadManager::CanDo(EAction action)
{
    switch(m_State) {
    case eSelectAcc:
        return action == eNext;
    case eSelectProject:
        return action == eBack  ||  action == eNext;
    case eCompleted:
        return false; // nothing left to do
    default:
        _ASSERT(false);
        return false;
    }
}


bool CGenBankUILoadManager::IsFinalState()
{
    return m_State == eSelectProject;
}


bool CGenBankUILoadManager::IsCompletedState()
{
    return m_State == eCompleted;
}


bool CGenBankUILoadManager::DoTransition(EAction action)
{
    if(m_State == eSelectAcc  &&  action == eNext)    {
        if (m_OptionPanel->IsInputValid())  {
            m_State = eSelectProject;
            return true;
        }
        return false;
    } else if( m_State == eSelectProject) {
        if(action == eBack)  {
            m_State = eSelectAcc;
            return true;
        } else if(action == eNext)  {
            if(m_ProjectSelPanel->TransferDataFromWindow()) {
                m_State = eCompleted;
                return true;
            }
            return false;
        }
    }
    _ASSERT(false);
    return false;
}


IAppTask* CGenBankUILoadManager::GetTask()
{
    // extract parameters from the dialog panels
    vector< CRef<CObject> > ids = m_OptionPanel->GetSeqIds();

    m_ProjectSelPanel->GetParams(m_ProjectParams);
    string folder_name = m_ProjectParams.m_CreateFolder ? m_ProjectParams.m_FolderName : "";

    // create loading Job
    CIRef<CProjectService> srv = m_SrvLocator->GetServiceByType<CProjectService>();
    CGenBankLoadingJob* job = new CGenBankLoadingJob(srv.GetPointer(), ids);

    m_ProjectSelPanel->GetParams(m_ProjectParams);

    CDataLoadingOptions& options = job->GetOptions();
    m_ProjectParams.ToLoadingOptions(options);

    // create a wrapper task
    CDataLoadingAppTask* task = new CDataLoadingAppTask(*job);
    return task;
}


void CGenBankUILoadManager::SetRegistryPath(const string& path)
{
    m_RegPath = path; // store for later use
}


static const char* kIdsInputTag = "IdsInput";
static const char* kProjectParamsTag = "ProjectParams";


void CGenBankUILoadManager::SaveSettings() const
{
    if( ! m_RegPath.empty())   {
        CGuiRegistry& gui_reg = CGuiRegistry::GetInstance();
        CGuiRegistry::TReadWriteView view = gui_reg.GetReadWriteView(m_RegPath);

        /// remember the selected Format (only if m_OptionPanel exists)
        if(m_OptionPanel)
            m_SavedInput = m_OptionPanel->GetInput();

        string encoded = NStr::URLEncode(m_SavedInput);
        if (encoded.size() < 10000) {
            // do not save large texts
            view.Set(kIdsInputTag, encoded);
        }

        /// save Project Panel settings
        m_ProjectParams.SaveSettings(view, kProjectParamsTag);
    }
}


void CGenBankUILoadManager::LoadSettings()
{
    if( ! m_RegPath.empty())   {
        CGuiRegistry& gui_reg = CGuiRegistry::GetInstance();
        CGuiRegistry::TReadView view = gui_reg.GetReadView(m_RegPath);

        string encoded = view.GetString(kIdsInputTag, kEmptyStr);
        m_SavedInput = NStr::URLDecode(encoded);

        /// load Project Panel settings
        m_ProjectParams.LoadSettings(view, kProjectParamsTag);
    }
}


///////////////////////////////////////////////////////////////////////////////
/// CGenBankLoadingJob

CGenBankLoadingJob::CGenBankLoadingJob(CProjectService* service, const IIdsVec& ids)
:   CDataLoadingAppJob(service),
    m_Ids(ids)
{
    CFastMutexGuard lock(m_Mutex);

    int n = m_Ids.size();
    string s = NStr::IntToString(n) + (n == 1 ? " id" : " ids");
    m_Descr = "Loading " + s + " from GenBank";
}


void CGenBankLoadingJob::x_CreateProjectItems()
{
    /// a scope for ID validation
    CRef<CScope> scope(new CScope(*CObjectManager::GetInstance()));
    scope->AddDefaults();

    // iterate by ids and create items
    for( size_t i = 0;  i < m_Ids.size();  i++  )   {
        CSeq_id*  seq_id = dynamic_cast<CSeq_id*> (m_Ids[i].GetPointer());
        CSeq_loc* seq_loc = dynamic_cast<CSeq_loc*> (m_Ids[i].GetPointer());

        if (seq_id == 0 && seq_loc == 0) {
            LOG_POST(Error << "CGenBankLoadingJob::x_CreateProjectItems(): "
                    "Invalid object type to open.");
            continue;
        }

        // validate the ID
        /*CScope::TIds ids = scope->GetIds(*id);
        if (ids.size() == 0) {
            ERR_POST("CGenBankLoadingJob::x_CreateProjectItems() - " <<
                     "Cannot resolve id '" + id->AsFastaString() + "'");
        } else */{
            // create a Project Item
            CRef<CProjectItem> item(new CProjectItem());
            if (seq_id)
                item->SetItem().SetId(*seq_id);
            else
                item->SetObject(*seq_loc);

            string label = item->GetLabelByData(scope.GetPointer());
            item->SetLabel(label);

            m_Items.push_back(item);
        }
    }
}


END_NCBI_SCOPE
