/*  $Id: gff_object_loader.cpp 28185 2013-05-21 20:19:45Z katargir $
* ===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data,  the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties,  express or implied,  including
*  warranties of performance,  merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================
*
* Authors:  Mike DiCuccio, Roman Katargin
*/


#include <ncbi_pch.hpp>

#include <util/icanceled.hpp>

#include <gui/packages/pkg_sequence/gff_object_loader.hpp>

#include <objtools/readers/gtf_reader.hpp>
#include <objtools/readers/gff3_reader.hpp>
#include <objtools/readers/gvf_reader.hpp>

#include <util/format_guess.hpp>

#include <gui/widgets/wx/compressed_file.hpp>
#include <gui/widgets/wx/wx_utils.hpp>

#include <wx/filename.h>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

CGffObjectLoader::CGffObjectLoader(const CGffLoadParams& params, const vector<wxString>& filenames)
 : m_Params(params), m_FileNames(filenames)
{
}

IObjectLoader::TObjects& CGffObjectLoader::GetObjects()
{
    return m_Objects;
}

string CGffObjectLoader::GetDescription() const
{
    return "Loading GFF/GTF Files";
}

bool CGffObjectLoader::PreExecute()
{
    return true;
}

bool CGffObjectLoader::Execute(ICanceled& canceled)
{
    static const char* errIdMapper = "GFF Importer. Creating Id mapper: ";

    CGFFReader::TFlags flags = 0;

    if (m_Params.GetNoGTF())
        flags |= CGFFReader::fNoGTF;
    if (m_Params.GetMergeExons())
        flags |= CGFFReader::fMergeExons;
    if (m_Params.GetMergeOnyCdsMrna())
        flags |= CGFFReader::fMergeOnyCdsMrna;
    if (m_Params.GetGBQuals())
        flags |= CGFFReader::fGBQuals;
    if (m_Params.GetCreateGeneFeats())
        flags |= CGFFReader::fCreateGeneFeats;
    if (m_Params.GetSetProducts())
        flags |= CGFFReader::fSetProducts;
    if (m_Params.GetSetVersion3())
        flags |= CGFFReader::fSetVersion3;

    switch(m_Params.GetParseSeqIds()) {
    case 1:
        flags |= CGFFReader::fNumericIdsAsLocal;
        break;
    case 2:
        flags |= CGFFReader::fNumericIdsAsLocal;
        break;
    }

    if (m_Params.GetUseAssemblyMapping()) {
        x_CreateMapper(m_Params.GetMapAssemblyAcc());
    }

    ITERATE(vector<wxString>, it, m_FileNames) {
        if (canceled.IsCanceled())
            return false;

        try {
            const wxString& fn = *it;
            CCompressedFile file(fn);
            CFormatGuess::EFormat fmt = CFormatGuess(file.GetIstream()).GuessFormat();
            file.Reset();

            CGff2Reader::TAnnots annots;

            /// set the name on all annotations
            string annot_name = ToStdString(m_Params.GetNameFeatSet());
            if (annot_name.empty()) {
                wxFileName fileName(fn);
                annot_name = "File: " + ToStdString(fileName.GetFullName());
            }

            if (fmt == CFormatGuess::eGff2) { 
                CGff2Reader reader(flags, annot_name, annot_name); 
                reader.ReadSeqAnnots(annots, file.GetIstream());
            } else if (fmt == CFormatGuess::eGff3) { 
                CGff3Reader reader(flags, annot_name, annot_name); 
                reader.ReadSeqAnnots(annots, file.GetIstream());
            } else if (fmt == CFormatGuess::eGtf) { 
                CGtfReader reader(flags, annot_name, annot_name); 
                reader.ReadSeqAnnots(annots, file.GetIstream());
            } else if (fmt == CFormatGuess::eGvf) {
                CGvfReader reader(flags, annot_name, annot_name);
                reader.ReadSeqAnnots(annots, file.GetIstream());
            }

            NON_CONST_ITERATE (CGff2Reader::TAnnots, annot_iter, annots) {
                if (canceled.IsCanceled())
                    return false;

                m_Objects.push_back(SObject(**annot_iter, annot_name));

                x_UpdateMap(**annot_iter);
            }
        }
        catch (const CException&) {
        }
        catch (const exception&) {
        }
    }

    return true;
}

bool CGffObjectLoader::PostExecute()
{
    return x_ShowMappingDlg(m_Objects);
}

END_NCBI_SCOPE
