/*  $Id: table_view.cpp 27834 2013-04-10 20:40:42Z rudnev $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Roman Katargin
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>

#include <gui/packages/pkg_sequence/table_view.hpp>

#include <gui/core/project_service.hpp>
#include <gui/core/document.hpp>

#include <gui/framework/workbench.hpp>
#include <gui/framework/status_bar_service.hpp>

#include <wx/grid.h>

#include <gui/objutils/interface_registry.hpp>
#include <gui/objutils/table_data_aln_summary.hpp>
#include <gui/objutils/table_data_snp.hpp>
#include <gui/objutils/table_data_validate.hpp>
#include <gui/widgets/grid_widget/grid_event.hpp>
#include <gui/widgets/grid_widget/grid.hpp>

#include <gui/widgets/wx/fileartprov.hpp>
#include <gui/widgets/wx/wx_utils.hpp>
#include <gui/widgets/wx/message_box.hpp>

#include <gui/utils/app_popup.hpp>

#include <objects/seqalign/Seq_align_set.hpp>
#include <gui/objects/GBProject_ver2.hpp>
#include <objects/general/Dbtag.hpp>
#include <objects/general/Object_id.hpp>
#include <objects/submit/Seq_submit.hpp>
#include <objects/seq/Pubdesc.hpp>
#include <objects/seqfeat/BioSource.hpp>
#include <objtools/snputil/snp_utils.hpp>

#include <objmgr/util/sequence.hpp>
#include <objects/seqalign/Seq_align.hpp>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

CProjectViewTypeDescriptor s_TableViewTypeDescr(
    "Generic Table View", // type name
    "table_view", // icon alias
    "Create an interactive table representation of data.",
    "The Generic Table View shows Seq-table object",
    "TABLE_VIEW", // help ID
    "Generic",   // category
    false,      // not a singleton
    "Seq-table",
    eOneObjectAccepted
);

CProjectViewTypeDescriptor s_AlnSummaryViewTypeDescr(
    "Alignment Summary View", // type name
    "aln_summary_view", // icon alias
    "Create an interactive table representation of alignment summary.",
    "The Table View shows summary for a list of alignments",
    "ALN_SUMMARY_VIEW", // help ID
    "Alignment",   // category
    false,      // not a singleton
    "Seq-align",
    eAnyObjectsAccepted
);

CProjectViewTypeDescriptor s_SNPViewTypeDescr(
    "SNP Table View",       // type name
    "snp_table_view",       // icon alias TODO
    "Show tabular information about SNPs",
    "The SNP Table View provides a tabular break-down of SNPs for a sequence",
    "SNP_TABLE_VIEW",       // help ID
    "Sequence",             // category
    false,                   // not a singleton
    "Seq-loc",
    eOneObjectAccepted
);

static
WX_DEFINE_MENU(kSNPMenu)
    WX_MENU_SEPARATOR_L("Contribs")
    WX_SUBMENU("&SNP")
        WX_MENU_ITEM(CSNPGridEvtExt::eCmdLinkRefSNP)
        WX_MENU_ITEM(CSNPGridEvtExt::eCmdLinkGene)
//        WX_MENU_ITEM(CSNPGridEvtExt::eCmdLinkGTY) -- dropped as asked in SNP-4726
        WX_MENU_ITEM(CSNPGridEvtExt::eCmdLinkOMIM)
        WX_MENU_ITEM(CSNPGridEvtExt::eCmdLinkSnp3D)
        WX_MENU_ITEM(CSNPGridEvtExt::eCmdLinkDisease)
    WX_END_SUBMENU()
WX_END_MENU()

CProjectViewTypeDescriptor s_ProjectTableViewTypeDescr(
    "Project Table View", // type name
    "project_table_view", // icon alias
    "Create an interactive table representation of data.",
    "The Project Table View shows Genome Workbench project",
    "PROJECT_TABLE_VIEW", // help ID
    "Generic",   // category
    false,      // not a singleton
    "GBProject-ver2",
    eOneObjectAccepted
);

CProjectViewTypeDescriptor s_ValidateViewTypeDescr(
    "Validate Table View",       // type name
    "validate_table_view",       // icon alias TODO
    "Show tabular information about features",
    "The Validate Table View provides a tabular representation for a sequence validation result",
    "VALIDATE_TABLE_VIEW",       // help ID
    "Sequence",             // category
    false,                   // not a singleton
    "Seq-entry",
    eAnyObjectsAccepted
);


CTableView::CTableView(EDataType type)
 : CProjectView()
 , m_Window(0)
 , m_DataType(type)
{
}

BEGIN_EVENT_MAP(CTableView, CProjectView)
  ON_EVENT(CGridWidgetEvent, CGridWidgetEvent::eUrlHover,
           &CTableView::x_OnUrlHover)
  ON_EVENT(CTableDataEvent, CTableDataEvent::eDataChanged,
           &CTableView::x_OnDataChanged)
END_EVENT_MAP()

wxWindow* CTableView::GetWindow()
{
    _ASSERT(m_Window);
    return m_Window;
}

wxEvtHandler* CTableView::GetCommandHandler()
{
    CGridWidget* grid = (CGridWidget*)m_Window;
    _ASSERT(grid);
    return grid->GetEventHandler();
}

const wxMenu* CTableView::GetMenu()
{
    // custom menu is created only for SNP
    if(m_DataType == eSNP) {
        //  Use CUICommandRegistry to create the menu from the static definition
        CUICommandRegistry& cmd_reg = CUICommandRegistry::GetInstance();
        wxMenu* menu = cmd_reg.CreateMenu(kSNPMenu);
        return menu;
    }
    return NULL;
}


// override function to suppress Navigation toolbar
void CTableView::GetCompatibleToolBars(vector<string>& names)
{
    names.push_back(CProjectViewToolBatFactory::sm_SelTBName);
}


bool CTableView::InitView(TConstScopedObjects& objects, const objects::CUser_object*)
{
    m_OrigObj.Reset(objects[0].object.GetPointer());

    CIRef<ITableData> table_data;
    if (m_DataType == eAlignment) {
        table_data.Reset(new CTableDataAlnSummary(objects));
    } else if (m_DataType == eSNP) {
        table_data.Reset(new CTableDataSNP(objects));
    } else if (m_DataType == eValidate) {
        table_data.Reset(new CTableDataValidate(objects));
    } else if (objects.size() == 1) {
        table_data.Reset(CreateObjectInterface<ITableData>(objects[0], NULL));
    }

    if (table_data) {
        table_data->AddListener(this, ePool_Parent);

        CGridWidget* widget = (CGridWidget*)m_Window;

        CIRef<CProjectService> srv = m_Workbench->GetServiceByType<CProjectService>();

        _ASSERT(srv);

        CRef<CWorkspaceConstPrx> workspace = srv->GetWorkspaceConst();
        CRef<CProjectPrx> project = workspace->FindProjectByScope(objects[0].scope.GetPointer());
        _ASSERT(project);

        if (project) {
            widget->Init(*table_data, &project->GetDocument()->GetUndoManager(), m_Workbench);

            if(m_DataType == eSNP) {
                widget->GetGrid()->AutoSizeRows();
            }
            x_AttachToProject(*project);
            x_UpdateContentLabel();
            return true;
        }
        return false;
    }

    // cannot represent the data
    x_ReportInvalidInputData(objects);
    return false;
}


void CTableView::OnProjectChanged(CProjectViewEvent& evt)
{
    if (m_DataType == eProject) {
        CGridWidget* widget = (CGridWidget*)m_Window;

        if (evt.GetSubtype() ==  CProjectViewEvent::eNone ||
            evt.GetSubtype() ==  CProjectViewEvent::eProperties) {
            widget->Refresh();
        }
        else if (evt.GetSubtype() == CProjectViewEvent::eData ||
                 evt.GetSubtype() == CProjectViewEvent::eBothPropsData) {
            TConstScopedObjects objects;
            GetMainObject(objects);

            CIRef<ITableData> table_data;
            if (objects.size() > 0) {
                //!! shouldn't we add other table data types here (like in InitView())?
                table_data.Reset(CreateObjectInterface<ITableData>(objects[0], NULL));
            }
            if (!table_data)
                return;
            table_data->AddListener(this, ePool_Parent);

            CIRef<CProjectService> srv = m_Workbench->GetServiceByType<CProjectService>();

            _ASSERT(srv);

            CRef<CWorkspaceConstPrx> workspace = srv->GetWorkspaceConst();
            CRef<CProjectPrx> project = workspace->FindProjectByScope(objects[0].scope.GetPointer());
            _ASSERT(project);

            if (project) {
                widget->Init(*table_data, &project->GetDocument()->GetUndoManager(), m_Workbench);
                if(m_DataType == eSNP) {
                    widget->GetGrid()->AutoSizeRows();
                }
            }
        }
    }
    else {
        if (evt.GetSubtype() == CProjectViewEvent::eDataChanging) {
            //CGridWidget* grid = (CGridWidget*)m_Window;
        }
        else if (evt.GetSubtype() == CProjectViewEvent::eData ||
                 evt.GetSubtype() == CProjectViewEvent::eBothPropsData) {
            CGridWidget* grid = (CGridWidget*)m_Window;
            grid->Refresh();
        }
    }
}

void CTableView::CreateViewWindow(wxWindow* parent)
{
    _ASSERT(!m_Window);

    CGridWidget* widget = new CGridWidget(parent, wxID_ANY, wxDefaultPosition, wxSize(0, 0),
        wxBORDER_NONE, m_DataType == eSNP ? new CSNPGridEvtExt() : NULL);
    m_Window = widget;
    widget->AddListener(this, ePool_Parent);
}


void CTableView::DestroyViewWindow()
{
    if( m_Window ){
        m_Window->Destroy();
        m_Window = 0;
    }
}


const CViewTypeDescriptor& CTableView::GetTypeDescriptor() const
{
    if (m_DataType == eAlignment) {
        return s_AlnSummaryViewTypeDescr;
    }
    else if (m_DataType == eProject) {
        return s_ProjectTableViewTypeDescr;
    }
    else if (m_DataType == eSNP) {
        return s_SNPViewTypeDescr;
    }
    else if (m_DataType == eValidate) {
        return s_ValidateViewTypeDescr;
    }

    return s_TableViewTypeDescr;
}


void CTableView::GetSelection(CSelectionEvent& evt) const
{
    if (m_Window  &&  x_HasProject()) {
        CGridWidget* grid = (CGridWidget*)m_Window;

        TConstScopedObjects sobjs;
        grid->GetSelectedObjects (sobjs);

        TConstObjects objs;
        ITERATE(TConstScopedObjects, it, sobjs) {
            objs.push_back(it->object);
        }

        evt.AddObjectSelection(objs);        
    }
}


void CTableView::GetSelection(TConstScopedObjects& objs) const
{
    if (m_Window  &&  x_HasProject()) {
        CGridWidget* grid = (CGridWidget*)m_Window;
        grid->GetSelectedObjects (objs);
    }
}


void CTableView::GetMainObject(TConstScopedObjects& objs) const
{
    if (m_Window  &&  x_HasProject()) {
        /*CGridWidget* grid = (CGridWidget*)m_Window;*/

        const CObject* object = m_OrigObj.GetPointer();
        if (object) {
            CScope* scope = x_GetScope();
            objs.push_back(SConstScopedObject(*object, *scope));
        }
    }
}


void CTableView::x_OnSetSelection (CSelectionEvent&)
{
// Doesn't work !!!!
//    if (!evt.HasObjectSelection())
//        return;

    if (m_Window  &&  x_HasProject()) {
        /*CGridWidget* grid = (CGridWidget*)m_Window;*/
    }
}

void CTableView::x_OnUrlHover(CEvent* evt)
{
    CGridWidgetEvent* gridEvt = dynamic_cast<CGridWidgetEvent*>(evt);
    if (!gridEvt) return;
    CGridWidgetEvent::TAttachment* attachment =
        dynamic_cast<CGridWidgetEvent::TAttachment*>(gridEvt->GetAttachment());
    if (!attachment) return;

    wxString text = attachment->GetData();

    CIRef<IStatusBarService> sb_srv = m_Workbench->GetServiceByType<IStatusBarService>();
    sb_srv->SetStatusMessage(ToStdString(text));
}


void CTableView::x_OnDataChanged(CEvent* evt)
{
    CTableDataEvent* tableEvt(dynamic_cast<CTableDataEvent*>(evt));
    if (!tableEvt) return;
    CTableDataEvent::TDataAttachment* attachment =
        dynamic_cast<CTableDataEvent::TDataAttachment*>(tableEvt->GetAttachment());
    if (!attachment) return;

    CIRef<ITableData> table_data(attachment->GetData());

    CGridWidget* widget = (CGridWidget*)m_Window;
    TConstScopedObjects objects;
    GetMainObject(objects);
    CIRef<CProjectService> srv = m_Workbench->GetServiceByType<CProjectService>();
    _ASSERT(srv);
    CRef<CWorkspaceConstPrx> workspace = srv->GetWorkspaceConst();
    CRef<CProjectPrx> project = workspace->FindProjectByScope(objects[0].scope.GetPointer());
    _ASSERT(project);
    widget->Init(*table_data, &project->GetDocument()->GetUndoManager(), m_Workbench);
    if(m_DataType == eSNP) {
        widget->GetGrid()->AutoSizeRows();
    }
	widget->Refresh();
}


const CObject* CTableView::x_GetOrigObject() const
{
    return m_OrigObj.GetPointer();
}

void CTableView::SetRegistryPath(const string& reg_path)
{
    IRegSettings* reg_set = dynamic_cast<IRegSettings*>(m_Window);
    if (reg_set)
        reg_set->SetRegistryPath(reg_path);
}

void CTableView::LoadSettings()
{
    IRegSettings* reg_set = dynamic_cast<IRegSettings*>(m_Window);
    if (reg_set)
        reg_set->LoadSettings();
}


void CTableView::SaveSettings() const
{
    IRegSettings* reg_set = dynamic_cast<IRegSettings*>(m_Window);
    if (reg_set)
        reg_set->SaveSettings();
}



///////////////////////////////////////////////////////////////////////////////
/// CTableViewFactory
string  CTableViewFactory::GetExtensionIdentifier() const
{
    static string sid("table_view_factory");
    return sid;
}


string CTableViewFactory::GetExtensionLabel() const
{
    static string slabel("Table View Factory");
    return slabel;
}


void CTableViewFactory::RegisterIconAliases(wxFileArtProvider& provider)
{
    string alias = GetViewTypeDescriptor().GetIconAlias();
    provider.RegisterFileAlias(ToWxString(alias), wxT("text_view.png"));
}


const CProjectViewTypeDescriptor& CTableViewFactory::GetProjectViewTypeDescriptor() const
{
    return s_TableViewTypeDescr;
}


IView* CTableViewFactory::CreateInstance() const
{
    return new CTableView(CTableView::eOthers);
}


IView* CTableViewFactory::CreateInstanceByFingerprint(const TFingerprint&) const
{
    return NULL;
}


int CTableViewFactory::TestInputObjects(TConstScopedObjects& objects)
{
    bool found_good = false;
    bool found_bad = false;

    for(  size_t i = 0;  i < objects.size();  i++)  {
        CIRef<ITableData> table_data(CreateObjectInterface<ITableData>(objects[i], NULL));

        if (table_data)
            found_good = true;
        else
            found_bad = true;
    }

    if (found_good)
        return fCanShowSeparated | (found_bad ? fCanShowSome : fCanShowAll);

    return 0; // can show nothing
}



///////////////////////////////////////////////////////////////////////////////
/// CAlnSummaryViewFactory
string  CAlnSummaryViewFactory::GetExtensionIdentifier() const
{
    static string aln_sid("aln_summary_view_factory");
    return aln_sid;
}


string CAlnSummaryViewFactory::GetExtensionLabel() const
{
    static string aln_slabel("Alignment Summary View Factory");
    return aln_slabel;
}


void CAlnSummaryViewFactory::RegisterIconAliases(wxFileArtProvider& provider)
{
    string alias = GetViewTypeDescriptor().GetIconAlias();
    provider.RegisterFileAlias(ToWxString(alias), wxT("text_view.png"));
}


const CProjectViewTypeDescriptor& CAlnSummaryViewFactory::GetProjectViewTypeDescriptor() const
{
    return s_AlnSummaryViewTypeDescr;
}


IView* CAlnSummaryViewFactory::CreateInstance() const
{
    return new CTableView(CTableView::eAlignment);
}


IView* CAlnSummaryViewFactory::CreateInstanceByFingerprint(const TFingerprint&) const
{
    return NULL;
}


bool CAlnSummaryViewFactory::IsCompatibleWith(const CObject& object, objects::CScope& scope)
{
    const type_info& type = typeid(object);
    if( typeid(CSeq_align) == type  ||
        typeid(CSeq_align_set) == type) {
            return true;
    }

    if( typeid(CSeq_annot) == type ){
        const CSeq_annot& annot = dynamic_cast<const CSeq_annot&>(object);
        if (annot.IsAlign()) {
            return true;
        }
    }

    return false;
}


int CAlnSummaryViewFactory::TestInputObjects(TConstScopedObjects& objects)
{
    bool found_good = false;
    bool found_bad = false;

    for(  size_t i = 0;  i < objects.size();  i++)  {
        if (IsCompatibleWith(*objects[i].object, *objects[i].scope)) {
            found_good = true;
        }

        found_bad = true;
    }

    if (found_good)
        return fCanShowSeparated | (found_bad ? fCanShowSome : fCanShowAll);

    return 0; // can show nothing
}


BEGIN_EVENT_TABLE(CSNPGridEvtExt, wxEvtHandler)
    EVT_MENU(CSNPGridEvtExt::eCmdLinkRefSNP, CSNPGridEvtExt::OnEventLinkout)
    EVT_MENU(CSNPGridEvtExt::eCmdLinkOMIM, CSNPGridEvtExt::OnEventLinkout)
//    EVT_MENU(CSNPGridEvtExt::eCmdLinkGTY, CSNPGridEvtExt::OnEventLinkout) -- dropped as asked in SNP-4726
    EVT_MENU(CSNPGridEvtExt::eCmdLinkGene, CSNPGridEvtExt::OnEventLinkout)
    EVT_MENU(CSNPGridEvtExt::eCmdLinkSnp3D, CSNPGridEvtExt::OnEventLinkout)
    EVT_MENU(CSNPGridEvtExt::eCmdLinkDisease, CSNPGridEvtExt::OnEventLinkout)
    EVT_UPDATE_UI(CSNPGridEvtExt::eCmdLinkRefSNP, CSNPGridEvtExt::OnEventUpdateLinkout)
    EVT_UPDATE_UI(CSNPGridEvtExt::eCmdLinkOMIM, CSNPGridEvtExt::OnEventUpdateLinkout)
//    EVT_UPDATE_UI(CSNPGridEvtExt::eCmdLinkGTY, CSNPGridEvtExt::OnEventUpdateLinkout) -- dropped as asked in SNP-4726
    EVT_UPDATE_UI(CSNPGridEvtExt::eCmdLinkGene, CSNPGridEvtExt::OnEventUpdateLinkout)
    EVT_UPDATE_UI(CSNPGridEvtExt::eCmdLinkSnp3D, CSNPGridEvtExt::OnEventUpdateLinkout)
    EVT_UPDATE_UI(CSNPGridEvtExt::eCmdLinkDisease, CSNPGridEvtExt::OnEventUpdateLinkout)
END_EVENT_TABLE()

void CSNPGridEvtExt::OnEventUpdateLinkout(wxUpdateUIEvent& event)
{    
    bool isEnabled(false);
    TConstScopedObjects SelectedObjects;
    x_GetGridWidget()->GetSelectedObjects(SelectedObjects);

    if(SelectedObjects.size() > 0) {
        CConstRef<objects::CVariation> pFirstSelected(
            dynamic_cast<const objects::CVariation*>(SelectedObjects[0].object.GetPointer()));
        if(pFirstSelected) {
            CSnpBitfield bf;
            if(pFirstSelected->CanGetExt()) {
                const CVariation::TExt& Ext(pFirstSelected->GetExt());
                ITERATE(CVariation::TExt, iExt, Ext) {
                    if((*iExt)->CanGetClass() &&
                        (*iExt)->GetClass() == SNP_VAR_EXT_CLASS &&
                        (*iExt)->GetFieldRef(SNP_VAR_EXT_BITFIELD).NotEmpty() &&
                        (*iExt)->GetFieldRef(SNP_VAR_EXT_BITFIELD)->CanGetData() &&
                        (*iExt)->GetFieldRef(SNP_VAR_EXT_BITFIELD)->GetData().IsOs()) {
                            bf = (*iExt)->GetFieldRef(SNP_VAR_EXT_BITFIELD)->GetData().GetOs();
                    }
                }
            }
            bool isBitfield(bf.GetVersion() > 0);
            switch (event.GetId()) {
                case eCmdLinkRefSNP:    
                    isEnabled = true;  
                    break;
                case eCmdLinkOMIM:      
                    isEnabled = isBitfield && bf.IsTrue(CSnpBitfield::eHasOMIM_OMIA);       
                    break; 
/*                case eCmdLinkGTY:        -- dropped as asked in SNP-4726
                    isEnabled = pFirstSelected->CanGetVariant_prop() && 
                                pFirstSelected->GetVariant_prop().CanGetGenotype() &&
                                pFirstSelected->GetVariant_prop().GetGenotype() & CVariantProperties::eGenotype_has_genotypes;
                    break; */
                case eCmdLinkDisease:   
                    isEnabled = isBitfield && bf.IsTrue(CSnpBitfield::eHasDiseaseInfo);
                    break;
                case eCmdLinkSnp3D:     
                    isEnabled = pFirstSelected->CanGetVariant_prop() && 
                                pFirstSelected->GetVariant_prop().CanGetResource_link() &&
                                pFirstSelected->GetVariant_prop().GetResource_link() & CVariantProperties::eResource_link_has3D;
                    break;
                case eCmdLinkGene:      
                    isEnabled = pFirstSelected->CanGetVariant_prop() && 
                                (pFirstSelected->GetVariant_prop().CanGetGene_location() ||
                                 pFirstSelected->GetVariant_prop().CanGetEffect());
                    break;
            }
        }
    }
    event.Enable(isEnabled);
}

void CSNPGridEvtExt::OnEventLinkout(wxCommandEvent& event)
{    
    const static string snp3dUrl  = "http://www.ncbi.nlm.nih.gov/SNP/snp3D.cgi?rsnum=";
    const static string geneUrl   = "http://www.ncbi.nlm.nih.gov/sites/entrez?"
                                    "Db=gene&DbFrom=snp&Cmd=Link&LinkName=snp_gene&LinkReadableName=Gene&"
                                    "IdsFromResult=";
//    const static string gtyUrl    = "http://www.ncbi.nlm.nih.gov/sites/entrez?cmd=Retrieve&db=Snp&dopt=GEN&list_uids=";
    const static string OmimUrl   = "http://www.ncbi.nlm.nih.gov/SNP/snp_redirect.cgi?snp_id=";
    const static string diseaseUrl= "http://www.ncbi.nlm.nih.gov/sites/varvu?gene=";

    string url;
    TConstScopedObjects SelectedObjects;
    x_GetGridWidget()->GetSelectedObjects(SelectedObjects);

    if(SelectedObjects.size() > 0) {
        CConstRef<objects::CVariation> pFirstSelected(
            dynamic_cast<const objects::CVariation*>(SelectedObjects[0].object.GetPointer()));
        if(pFirstSelected) {
            string sRsID;
            if(pFirstSelected->GetId().CanGetTag()) {
                CNcbiOstrstream ostr;
                pFirstSelected->GetId().GetTag().AsString(ostr);
                sRsID = CNcbiOstrstreamToString(ostr);
            } else {
                pFirstSelected->GetId().GetLabel(&sRsID);
            }
        
            switch (event.GetId()) {
                case eCmdLinkRefSNP:    url = pFirstSelected->GetId().GetUrl();  break;
                case eCmdLinkOMIM:      url = OmimUrl + sRsID;       break;
                // case eCmdLinkGTY:       url = gtyUrl + sRsID;        break; -- dropped as asked in SNP-4726
                case eCmdLinkDisease:   
                {
                    if(pFirstSelected->CanGetOther_ids()) {
                        const CVariation::TOther_ids& OtherIDs(pFirstSelected->GetOther_ids());
                        ITERATE(CVariation::TOther_ids, iOtherIDs, OtherIDs) {
                            if((*iOtherIDs)->CanGetDb() &&
                                ((*iOtherIDs)->GetDb() == "GeneID" ||
                                (*iOtherIDs)->GetDb() == "LocusLink") &&
                                (*iOtherIDs)->CanGetTag()) {
                                    url = diseaseUrl + NStr::NumericToString((*iOtherIDs)->GetTag().GetId());
                                    break;
                            }
                        }
                    }
                    break;
                }
                case eCmdLinkSnp3D:     url = snp3dUrl + sRsID;      break;
                case eCmdLinkGene:      
                {   
                    list<string> Effects;
                    NSNPVariationHelper::VariantPropAsStrings(Effects,
                                                                pFirstSelected->GetVariant_prop(),
                                                                NSNPVariationHelper::eSNPPropName_Effect);
                    if(!Effects.empty()) {
                        url = pFirstSelected->GetId().GetUrl() + "#locus";
                    }
                    else {
                        url = geneUrl + sRsID;
                    }
                    break;
                }
            }
            if(!url.empty())
                CAppPopup::PopupURL(url);
        }
    }
}




///////////////////////////////////////////////////////////////////////////////
/// CSnpTableViewFactory
string  CSnpTableViewFactory::GetExtensionIdentifier() const
{
    static string sid("snp_table_view_factory");
    return sid;
}

string CSnpTableViewFactory::GetExtensionLabel() const
{
    static string slabel("SNP Table View Factory");
    return slabel;
}

void CSnpTableViewFactory::RegisterIconAliases(wxFileArtProvider& provider)
{
    string alias = GetViewTypeDescriptor().GetIconAlias();
    //!! replace with another icon?
	provider.RegisterFileAlias(ToWxString(alias), wxT("feature_table_view.png"));
}


const CProjectViewTypeDescriptor& CSnpTableViewFactory::GetProjectViewTypeDescriptor() const
{
    return s_SNPViewTypeDescr;
}

vector<CUICommand*> CSnpTableViewFactory::GetCommands()
{
    vector<CUICommand*> Commands;
    CUICommand* cmd(new CUICommand(CSNPGridEvtExt::eCmdLinkRefSNP, 
                                   "&RefSNP Link", 
                                   "RefSNP Link", 
                                   ""));
    Commands.push_back(cmd);

    cmd = new CUICommand(CSNPGridEvtExt::eCmdLinkOMIM, 
                                   "&OMIM Link", 
                                   "OMIM Link", 
                                   "");
    Commands.push_back(cmd);

    cmd = new CUICommand(CSNPGridEvtExt::eCmdLinkSnp3D, 
                                   "SNP&3D Link", 
                                   "SNP3D Link", 
                                   "");
    Commands.push_back(cmd);

/*    cmd = new CUICommand(CSNPGridEvtExt::eCmdLinkGTY,   -- dropped as asked in SNP-4726
                                   "&Genotype Link", 
                                   "Genotype Link", 
                                   "");
    Commands.push_back(cmd);
*/
    cmd = new CUICommand(CSNPGridEvtExt::eCmdLinkGene, 
                                   "Gene &Function Link", 
                                   "Gene Function Link", 
                                   "");
    Commands.push_back(cmd);

    cmd = new CUICommand(CSNPGridEvtExt::eCmdLinkDisease, 
                                   "&Disease Link", 
                                   "Disease Link", 
                                   "");
    Commands.push_back(cmd);

    return Commands;
}

IView* CSnpTableViewFactory::CreateInstance() const
{
    return new CTableView(CTableView::eSNP);
}


IView* CSnpTableViewFactory::CreateInstanceByFingerprint(const TFingerprint& /*fingerprint*/) const
{
    return NULL;
}


int CSnpTableViewFactory::TestInputObjects(TConstScopedObjects& objects)
{
    bool found_good = false;
    bool found_bad = false;
    for( size_t i = 0; i < objects.size(); i++ ){
        const CObject* obj = objects[i].object;
        const type_info& type = typeid(*obj);
		bool isThisGood(typeid(CSeq_loc) == type);

		if(isThisGood && !found_good) {
			found_good = true;
		}
		if(!isThisGood && !found_bad) {
			found_bad = true;
		}
    }
    if( found_good ){
        return fCanShowSeparated | (found_bad ? fCanShowSome : fCanShowAll);
    }
    return fCanShowNone;
}

/// CProjectTableViewFactory
string  CProjectTableViewFactory::GetExtensionIdentifier() const
{
    static string sid("project_table_view_factory");
    return sid;
}


string CProjectTableViewFactory::GetExtensionLabel() const
{
    static string slabel("Project Table View Factory");
    return slabel;
}


void CProjectTableViewFactory::RegisterIconAliases(wxFileArtProvider& provider)
{
}


const CProjectViewTypeDescriptor& CProjectTableViewFactory::GetProjectViewTypeDescriptor() const
{
    return s_ProjectTableViewTypeDescr;
}


IView* CProjectTableViewFactory::CreateInstance() const
{
    return new CTableView(CTableView::eProject);
}


bool CProjectTableViewFactory::IsCompatibleWith(const CObject& object, objects::CScope& scope)
{
    const type_info& type = typeid(object);
    if( typeid(CGBProject_ver2) == type) {
        return true;
    }
    return false;
}

IView* CProjectTableViewFactory::CreateInstanceByFingerprint(const TFingerprint&) const
{
    return NULL;
}


int CProjectTableViewFactory::TestInputObjects(TConstScopedObjects& objects)
{
    bool found_good = false;
    bool found_bad = false;

    for(  size_t i = 0;  i < objects.size();  i++)  {
        if (IsCompatibleWith(*objects[i].object, *objects[i].scope)) {
            found_good = true;
        }

        found_bad = true;
    }

    if (found_good)
        return fCanShowSeparated | (found_bad ? fCanShowSome : fCanShowAll);

    return 0; // can show nothing
}


///////////////////////////////////////////////////////////////////////////////
/// CValidateTableViewFactory
string  CValidateTableViewFactory::GetExtensionIdentifier() const
{
    static string sid("validate_table_view_factory");
    return sid;
}

string CValidateTableViewFactory::GetExtensionLabel() const
{
    static string slabel("Validate Table View Factory");
    return slabel;
}

void CValidateTableViewFactory::RegisterIconAliases(wxFileArtProvider& provider)
{
    string alias = GetViewTypeDescriptor().GetIconAlias();
    //!! replace with another icon?
	provider.RegisterFileAlias(ToWxString(alias), wxT("feature_table_view.png"));
}


const CProjectViewTypeDescriptor& CValidateTableViewFactory::GetProjectViewTypeDescriptor() const
{
    return s_ValidateViewTypeDescr;
}


IView* CValidateTableViewFactory::CreateInstance() const
{
    return new CTableView(CTableView::eValidate);
}


IView* CValidateTableViewFactory::CreateInstanceByFingerprint(const TFingerprint& /*fingerprint*/) const
{
    return NULL;
}


int CValidateTableViewFactory::TestInputObjects(TConstScopedObjects& objects)
{
    bool found_good = false;
    bool found_bad = false;
    for( size_t i = 0; i < objects.size(); i++ ){
        const CObject* obj = objects[i].object;
        const type_info& type = typeid(*obj);
		bool isThisGood(typeid(CSeq_entry) == type
            || typeid(CSeq_submit) == type
            || typeid(CSeq_annot_Handle) == type
            || typeid(CSeq_feat) == type
            || typeid(CBioSource) == type
            || typeid(CPubdesc) == type);

		if(isThisGood && !found_good) {
			found_good = true;
		}
		if(!isThisGood && !found_bad) {
			found_bad = true;
		}
    }
    if( found_good ){
        return fCanCombine | (found_bad ? fCanShowSome : fCanShowAll);
    }
    return fCanShowNone; // can show nothing
}

vector<int> CValidateTableViewFactory::CombineInputObjects( const TConstScopedObjects& objs )
{
    size_t num_objects = objs.size();

    if( num_objects <= 0 ){
        return vector<int>();
    }
	// always group all validation objects into one common group
	return vector<int>( num_objects, 0 );
}

void CValidateTableViewFactory::CombineObjects( 
    const TConstScopedObjects& objects, TObjectsMap& objects_map 
){
    typedef map<vector<CSeq_id_Handle>, TConstScopedObjects> TIdsToObjectsMap;
    TIdsToObjectsMap handle_groups;
    
    ITERATE( TConstScopedObjects, obtr, objects ){

        CConstRef<CObject> object = obtr->object;
        CRef<CScope> scope = obtr->scope;

        if( object.IsNull() || scope.IsNull() ){
            continue;
        }

        vector<CSeq_id_Handle> idh_vec;

        const CSeq_id* id = dynamic_cast<const CSeq_id*>( &*object );
        if( id ){
            CSeq_id_Handle idh = sequence::GetId( *id, *scope, sequence::eGetId_Canonical );
            idh_vec.push_back( idh ? idh : CSeq_id_Handle::GetHandle( *id ) );
        }

        const CSeq_loc* loc = dynamic_cast<const CSeq_loc*>( &*object );
        if( loc ){
            //! If loc is mix, we need to iterate ids
            const CSeq_id& seq_id = *loc->GetId();
            CSeq_id_Handle idh = sequence::GetId( seq_id, *scope, sequence::eGetId_Canonical );
            idh_vec.push_back( idh ? idh : CSeq_id_Handle::GetHandle( seq_id ) );
        }

        const CSeq_align* aln = dynamic_cast<const CSeq_align*>( &*object );
        if( aln ){
            int num_seqs = aln->CheckNumRows();
            if( num_seqs <= 0 ){
                continue;
            }

            set<CSeq_id_Handle> idh_set;
            for( int seqix = 0; seqix < num_seqs; seqix++ ){
                const CSeq_id& seq_id = aln->GetSeq_id( seqix );
                CSeq_id_Handle idh = sequence::GetId( seq_id, *scope, sequence::eGetId_Canonical );
                idh_set.insert( idh ? idh : CSeq_id_Handle::GetHandle( seq_id ) );
            }

            ITERATE( set<CSeq_id_Handle>, idh_itr, idh_set ){
                idh_vec.push_back( *idh_itr );
            }

            if( idh_vec.size() == 1 ){
                idh_vec.push_back( idh_vec.front() );
            }
        }


        handle_groups[idh_vec].push_back( *obtr );
    }

    objects_map.clear();
    NON_CONST_ITERATE( TIdsToObjectsMap, sidhtr, handle_groups ){
        const vector<CSeq_id_Handle>& idh_vec = sidhtr->first;
        TConstScopedObjects& group = sidhtr->second;
        
        string label;

        bool comma = false;
        ITERATE( vector<CSeq_id_Handle>, idhtr, idh_vec ){
            if( comma ){
                label += " x ";
            } else {
                comma = true;
            }

            CSeq_id_Handle idh = sequence::GetId( *idhtr, *group[0].scope, sequence::eGetId_Best );
            label += (idh ? idh : *idhtr).GetSeqId()->GetSeqIdString( true );
        }

        if( !label.empty() ){
            objects_map[label] = group;

        } else {
            NON_CONST_ITERATE( TConstScopedObjects, scobtr, group ){
                TConstScopedObjects single;
                single.push_back( *scobtr );

                CLabel::GetLabel( *scobtr->object, &label, CLabel::eContent, scobtr->scope );

                objects_map[label].push_back( *scobtr );
                label.clear();
            }
        }
    }
}



END_NCBI_SCOPE
