/*  $Id: text_view.cpp 27567 2013-03-05 20:44:07Z katargir $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Roman Katargin
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>

#include <gui/packages/pkg_sequence/text_view.hpp>

#include <gui/core/project_service.hpp>
#include <gui/core/document.hpp>

#include <objects/general/User_object.hpp>
#include <objects/general/Object_id.hpp>

#include <gui/framework/workbench.hpp>
#include <gui/framework/status_bar_service.hpp>

#include <gui/widgets/seq/text_panel.hpp>
#include <gui/widgets/seq/text_ds.hpp>
#include <gui/widgets/edit/bioseq_editor.hpp>
#include <gui/widgets/text_widget/text_item_panel.hpp>
#include <gui/widgets/text_widget/text_event.hpp>

#include <gui/widgets/wx/fileartprov.hpp>
#include <gui/widgets/wx/wx_utils.hpp>
#include <gui/widgets/wx/message_box.hpp>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);


CProjectViewTypeDescriptor s_TextViewTypeDescr(
    "Text View", // type name
    "text_view", // icon alias
    "Create an interactive text representation of data.",
    "The Text View shows data in a variety of formats, including GenBank FlatFile, "
    "NCBI ASN.1, and FastA format.",
    "TEXT_VIEW", // help ID
    "Generic",   // category
    false,      // not a singleton
    "SerialObject",
    eOneObjectAccepted
);

CTextView::CTextView()
  : CProjectView(), m_Window(0), m_UsingStatusBar(false)
{
}

CTextView::~CTextView()
{
    if (m_UsingStatusBar && m_Workbench) {
        CIRef<IStatusBarService> sb_srv = m_Workbench->GetServiceByType<IStatusBarService>();
        if (sb_srv) {
            sb_srv->RemoveSlot(1);
        }
    }
}

BEGIN_EVENT_MAP(CTextView, CProjectView)
  ON_EVENT(CTextWidgetEvent, CTextWidgetEvent::eGotFocus,
           &CTextView::x_OnWidgetGotFocus)
  ON_EVENT(CTextWidgetEvent, CTextWidgetEvent::eLostFocus,
           &CTextView::x_OnWidgetLostFocus)
  ON_EVENT(CTextWidgetEvent, CTextWidgetEvent::ePositionChanged,
           &CTextView::x_OnWidgetPositionChanged)
END_EVENT_MAP()

wxWindow* CTextView::GetWindow()
{
    _ASSERT(m_Window);
    return m_Window;
}

wxEvtHandler* CTextView::GetCommandHandler()
{
    CTextPanel* panel = (CTextPanel*) m_Window;
    _ASSERT(panel);
    CTextItemPanel* widget = panel->GetWidget();
    return widget->GetEventHandler();
}

// override function to suppress Navigation toolbar
void CTextView::GetCompatibleToolBars(vector<string>& names)
{
    names.push_back(CProjectViewToolBatFactory::sm_SelTBName);
}


static
WX_DEFINE_MENU(kTextViewTopMenu)
    WX_SUBMENU("&Edit")
        WX_MENU_SEPARATOR_L("Contribs")
            WX_MENU_ITEM(eCmdEditObjects)
            WX_MENU_ITEM(eCmdDeleteObjects)
    WX_END_SUBMENU()
WX_END_MENU()

const wxMenu* CTextView::GetMenu()
{
    CUICommandRegistry& cmd_reg = CUICommandRegistry::GetInstance();
    wxMenu* menu = cmd_reg.CreateMenu(kTextViewTopMenu);
    return menu;
}

bool CTextView::InitView(TConstScopedObjects& objects, const objects::CUser_object* params)
{
    CScope* scope = NULL;
    const CObject* object = NULL;

    if(objects.size() == 1) {
        object = objects[0].object.GetPointer();
        scope = objects[0].scope.GetPointer();
    }

    if(object)  {
        CIRef<CProjectService> srv = m_Workbench->GetServiceByType<CProjectService>();

        _ASSERT(srv);

        CRef<CWorkspaceConstPrx> workspace = srv->GetWorkspaceConst();
        CRef<CProjectPrx> project = workspace->FindProjectByScope(scope);
        _ASSERT(project);

        if (project) {
            CTextPanel* panel = (CTextPanel*) m_Window;
            panel->SetUndoManager(&project->GetDocument()->GetUndoManager());

            if (params) {
                string viewType;
                bool expandAll = false;

                const CObject_id& type = params->GetType();
                if (type.IsStr() && type.GetStr() == "TextViewParams") {
                    ITERATE(CUser_object::TData, it, params->GetData()) {
                        const CObject_id& field_id = (*it)->GetLabel();
                        if (!field_id.IsStr())
                            continue;

                        const string& label = field_id.GetStr();
                        const CUser_field::TData& data = (*it)->GetData();

                        if (label == "TextViewType") {
                            if (data.IsStr())
                                viewType = data.GetStr();
                        }
                        else if (label == "ExpandAll") {
                            if (data.IsBool())
                                expandAll = data.GetBool();
                        }
                    }
                }

                if (!viewType.empty())
                    panel->SetInitialViewType(viewType);

                if (expandAll)
                    panel->SetInitialExpandAll();
            }

            panel->Init(objects[0]);
            m_OrigObj.Reset(objects[0].object.GetPointer());
            panel->TransferDataToWindow();

            x_AttachToProject(*project);
            x_UpdateContentLabel();
            return true;
        }
        return false;
    } else {
        // cannot represent the data
        x_ReportInvalidInputData(objects);
        return false;
    }
}

void CTextView::OnProjectChanged(CProjectViewEvent& evt)
{
    if (evt.GetSubtype() == CProjectViewEvent::eDataChanging) {
        CTextPanel* panel = (CTextPanel*) m_Window;
        CTextItemPanel* widget = panel->GetWidget();
        widget->BlockEvents(true);
    }
    else if (evt.GetSubtype() == CProjectViewEvent::eData ||
             evt.GetSubtype() == CProjectViewEvent::eBothPropsData) {
        CTextPanel* panel = (CTextPanel*) m_Window;
        CTextItemPanel* widget = panel->GetWidget();
        panel->Reload();
    }
}


void CTextView::CreateViewWindow(wxWindow* parent)
{
    _ASSERT(!m_Window);

    CTextPanel* panel = new CTextPanel(parent, wxID_ANY, wxDefaultPosition, wxSize(0, 0), wxTAB_TRAVERSAL|wxNO_BORDER);

    m_Window = panel;
    CTextItemPanel* widget = panel->GetWidget();
    widget->AddListener(this, ePool_Parent);
}


void CTextView::DestroyViewWindow()
{
    if( m_Window ){
        m_Window->Destroy();
        m_Window = 0;
    }
}


const CViewTypeDescriptor& CTextView::GetTypeDescriptor() const
{
    return s_TextViewTypeDescr;
}


void CTextView::GetSelection(CSelectionEvent& evt) const
{
    if (x_HasProject()) {
        CTextPanel* panel = (CTextPanel*)m_Window;
        CTextItemPanel* widget = panel->GetWidget();
        if (!widget)
            return;

        CTextPanelContext* context = widget->GetContext();
        if (!context)
            return;

        vector< CConstRef<CObject> > objects;
        context->GetSelectedObjects(objects);
        evt.AddObjectSelection(objects);
    }
}


void CTextView::GetSelection(TConstScopedObjects& objs) const
{
    if (m_Window  &&  x_HasProject()) {
        CScope* scope = x_GetScope();
        CTextPanel* panel = (CTextPanel*)m_Window;
        CTextItemPanel* widget = panel->GetWidget();
        if ( ! widget)
            return;

        CTextPanelContext* context = widget->GetContext();
        if ( ! context)
            return;

        vector< CConstRef<CObject> > objects;
        context->GetSelectedObjects(objects);
        ITERATE(vector< CConstRef<CObject> >, it, objects)
            objs.push_back(SConstScopedObject(**it, *scope));
    }
}


void CTextView::GetMainObject(TConstScopedObjects& objs) const
{
    if (m_Window  &&  x_HasProject()) {
        CTextPanel* panel = (CTextPanel*) m_Window;
        const CSerialObject* object = panel->GetDataObject();

        if (object) {
            CScope* scope = x_GetScope();
            objs.push_back(SConstScopedObject(*object, *scope));
        }
    }
}


void CTextView::x_OnSetSelection (CSelectionEvent& evt)
{
// Doesn't work !!!!
//    if (!evt.HasObjectSelection())
//        return;

    if (m_Window  &&  x_HasProject()) {
        CTextPanel* panel = (CTextPanel*)m_Window;
        CTextItemPanel* widget = panel->GetWidget();
        if (!widget)
            return;

        CTextPanelContext* context = widget->GetContext();
        if (!context)
            return;

        TConstObjects objs;
        evt.GetAllObjects(objs);
        context->SetSelectedObjects(objs);
    }
}

const CObject* CTextView::x_GetOrigObject() const
{
    return m_OrigObj.GetPointer();
}

void CTextView::x_OnWidgetGotFocus(CEvent*)
{
    CIRef<IStatusBarService> sb_srv = m_Workbench->GetServiceByType<IStatusBarService>();
    if (!m_UsingStatusBar) {
        sb_srv->InsertSlot(1, NULL, 120);
        m_UsingStatusBar = true;
    }
}

void CTextView::x_OnWidgetLostFocus(CEvent*)
{
    if (m_UsingStatusBar) {
        CIRef<IStatusBarService> sb_srv = m_Workbench->GetServiceByType<IStatusBarService>();
        sb_srv->RemoveSlot(1);
        m_UsingStatusBar = false;
    }
}

void CTextView::x_OnWidgetPositionChanged(CEvent* evt)
{
    CTextWidgetEvent* textEvt = dynamic_cast<CTextWidgetEvent*>(evt);
    if (!textEvt) return;
    CTextWidgetEvent::TAttachment* attachment =
        dynamic_cast<CTextWidgetEvent::TAttachment*>(textEvt->GetAttachment());
    if (!attachment) return;

    CNcbiOstrstream ostr;
    ostr << attachment->GetData().GetEnd();
    string text = CNcbiOstrstreamToString(ostr);

    CIRef<IStatusBarService> sb_srv = m_Workbench->GetServiceByType<IStatusBarService>();
    sb_srv->SetStatusText(ToWxString(text), 1);
}

///////////////////////////////////////////////////////////////////////////////
/// CTextViewFactory
string  CTextViewFactory::GetExtensionIdentifier() const
{
    static string sid("text_view_factory");
    return sid;
}


string CTextViewFactory::GetExtensionLabel() const
{
    static string slabel("Text View Factory");
    return slabel;
}


void CTextViewFactory::RegisterIconAliases(wxFileArtProvider& provider)
{
    string alias = GetViewTypeDescriptor().GetIconAlias();
    provider.RegisterFileAlias(ToWxString(alias), wxT("text_view.png"));
}

void CTextViewFactory::RegisterCommands(CUICommandRegistry& cmd_reg, wxFileArtProvider& provider)
{
    CTextDataSource::RegisterCommands(cmd_reg, provider);
}

const CProjectViewTypeDescriptor& CTextViewFactory::GetProjectViewTypeDescriptor() const
{
    return s_TextViewTypeDescr;
}


IView* CTextViewFactory::CreateInstance() const
{
    return new CTextView();
}


IView* CTextViewFactory::CreateInstanceByFingerprint(const TFingerprint&) const
{
    return NULL;
}


int CTextViewFactory::TestInputObjects(TConstScopedObjects& objects)
{
    bool found_good = false;
    bool found_bad = false;

    for(  size_t i = 0;  i < objects.size();  i++)  {
        const CSerialObject* so =
            dynamic_cast<const CSerialObject*>(objects[i].object.GetPointer());
        if (so)
            found_good = true;
        else
            found_bad = true;
    }

    if (found_good)
        return fCanShowSeparated | (found_bad ? fCanShowSome : fCanShowAll);

    return 0; // can show nothing
}


END_NCBI_SCOPE
