/*  $Id: vcf_object_loader.cpp 28185 2013-05-21 20:19:45Z katargir $
* ===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data,  the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties,  express or implied,  including
*  warranties of performance,  merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================
*
* Authors:  Josh Cherry, Roman Katargin
*/


#include <ncbi_pch.hpp>

#include <gui/packages/pkg_sequence/vcf_object_loader.hpp>

#include <util/line_reader.hpp>
#include <util/icanceled.hpp>

#include <objtools/readers/reader_base.hpp>
#include <objtools/readers/vcf_reader.hpp>

#include <objects/seq/Annot_descr.hpp>
#include <objects/seq/Annotdesc.hpp>

#include <gui/core/job_results_dlg.hpp>

#include <gui/widgets/wx/compressed_file.hpp>
#include <gui/widgets/wx/wx_utils.hpp>

#include <wx/txtstrm.h>
#include <wx/filename.h>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

CVcfObjectLoader::CVcfObjectLoader(const CBedLoadParams& params, const vector<wxString>& filenames)
 : m_Params(params), m_FileNames(filenames)
{
}

IObjectLoader::TObjects& CVcfObjectLoader::GetObjects()
{
    return m_Objects;
}

string CVcfObjectLoader::GetDescription() const
{
    return "Loading VCF (Variant Call Format) Files";
}

bool CVcfObjectLoader::PreExecute()
{
    return true;
}

namespace {
    class CErrorContainer : public CErrorContainerLenient
    {
    public:
        CErrorContainer(size_t maxErrors) : m_MaxErrors(maxErrors) {}
        virtual bool PutError(const ILineError& err)
        {
            return Count() < m_MaxErrors && CErrorContainerLenient::PutError(err);
        }
    private:
        const size_t m_MaxErrors;
    };
}

bool CVcfObjectLoader::Execute(ICanceled& canceled)
{
    static const char* errIdMapper = "VCF Importer. Creating Id mapper: ";
    typedef CRef<objects::CSeq_annot> TAnnotRef;

    if (m_Params.GetUseAssemblyMapping()) {
        x_CreateMapper(m_Params.GetMapAssemblyAcc());
    }

    ITERATE(vector<wxString>, it, m_FileNames) {
        if (canceled.IsCanceled())
            return false;

        m_CurrentFile = *it;

        try {
            ITERATE(vector<wxString>, it, m_FileNames) {
                m_CurrentFile = *it;

                const wxString& fn = *it;
                CCompressedFile file(fn);
                CStreamLineReader lineReader(file.GetIstream());

                CVcfReader reader;
                CRef<CErrorContainer> errCont(new CErrorContainer((size_t)m_Params.GetNumErrors()));
                CRef<CSeq_annot> annot(reader.ReadSeqAnnot(lineReader, errCont));
                x_UpdateHTMLResults(*errCont);

                if (!annot) continue;

                string label;

                bool has_name = false;
                if (annot->CanGetDesc()) {
                    const CAnnot_descr& desc = annot->GetDesc();
                    if (desc.CanGet()) {
                        const CAnnot_descr::Tdata& data = desc.Get();
                        CSeq_annot::TDesc::Tdata::const_iterator desc_iter = data.begin();
                        for ( ;  desc_iter != data.end();  ++desc_iter) {
                            if ((*desc_iter)->IsTitle() ) {
                                label = (*desc_iter)->GetTitle();
                                break;
                            }
                            if ((*desc_iter)->IsName() ) {
                                label = (*desc_iter)->GetName();
                                has_name = true;
                            }
                        }
                    }
                }

                {{
                wxFileName wx_fn(m_CurrentFile);
                string s = ToStdString(wx_fn.GetFullName());
                label.append(" ");
                label.append(s.c_str());
                annot->SetNameDesc(string(s.c_str()));
                }}

                m_Objects.push_back(SObject(*annot, label));

                x_UpdateMap(*annot);
            }
        }
        catch (const CException&) {
        }
        catch (const exception&) {
        }
    }

    return true;
}

void CVcfObjectLoader::x_UpdateHTMLResults(IErrorContainer& errCont)
{
    if (errCont.Count() == 0)
        return;

    wxTextOutputStream os(m_Strstrm);

    os << wxT("<h3>File: ") << m_CurrentFile << wxT("</h3>");

    for (size_t i = 0; i < errCont.Count(); ++i) {
        const ILineError& lerror = errCont.GetError(i);
        os << wxT("Line ") << lerror.Line() << wxT(". ") << ToWxString(lerror.Message()) << wxT("<br />");
    }
}

bool CVcfObjectLoader::PostExecute()
{
    wxString html = m_Strstrm.GetString();
    if (!html.IsEmpty()) {
        CJobResultsDlg dlg;
        dlg.SetHTML(html);
        dlg.Create(NULL, wxID_ANY, wxT("VCF import results"));
        dlg.ShowModal();
    }

    return x_ShowMappingDlg(m_Objects);
}

END_NCBI_SCOPE
