/*  $Id: view_graphic.cpp 27914 2013-04-24 15:07:09Z wuliangs $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software / database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software / database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Vlad Lebedev, Liangshou Wu
 *
 * File Description:
 *    User-modifiable portion of main graphical sequence viewer class
 */


#include <ncbi_pch.hpp>

#include <gui/packages/pkg_sequence/view_graphic.hpp>

#include <gui/widgets/seq_graphic/seqgraphic_widget.hpp>
#include <gui/widgets/seq_graphic/vis_range_glyph.hpp>

#include <gui/core/project_service.hpp>
#include <gui/core/data_mining_service.hpp>

#include <gui/framework/workbench.hpp>
#include <gui/widgets/wx/fileartprov.hpp>
#include <gui/widgets/wx/wx_utils.hpp>
#include <gui/widgets/wx/async_call.hpp>

#include <gui/objutils/utils.hpp>
#include <gui/objutils/label.hpp>

#include <gui/objutils/visible_range.hpp>

#include <objects/seqloc/Seq_interval.hpp>
#include <objmgr/util/sequence.hpp>

#include <serial/iterator.hpp>
#include <wx/msgdlg.h>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

BEGIN_EVENT_MAP(CViewGraphic, CProjectView)
    ON_MESSAGE(CEvent, CViewEvent::eWidgetDataChanged, &CViewGraphic::OnWidgetDataChanged)
END_EVENT_MAP()

///////////////////////////////////////////////////////////////////////////////
/// CViewGraphic

CProjectViewTypeDescriptor s_ViewGraphicTypeDescr(
    "Graphical View", // type name
    "graphical_view", // icon alias TODO
    "Graphical View",
    "The Graphical View provides an overview of annotations "
    "on sequence, with extensive drill-down capabilities.",
    "GRAPHICAL_VIEW", // help ID
    "Sequence",     // category
    false,       // singleton
    "Seq-loc",
    eSimilarObjectsAccepted
);

CViewGraphic::CViewGraphic()
:   m_SeqWidget(NULL)
{
}


CViewGraphic::~CViewGraphic()
{
}


const wxMenu* CViewGraphic::GetMenu()
{
    return NULL;// TODO
}


void CViewGraphic::CreateViewWindow(wxWindow* parent)
{
    _ASSERT( ! m_SeqWidget);

    m_SeqWidget = new CSeqGraphicWidget(parent);
    m_SeqWidget->Create();

    AddListener(m_SeqWidget, ePool_Child);
    m_SeqWidget->AddListener(this, ePool_Parent);
}


void CViewGraphic::DestroyViewWindow()
{
    //m_DataSource->DeleteAllJobs(); // Cancel all jobs before listener is deleted

    if(m_SeqWidget) {
        m_SeqWidget->Destroy();
        m_SeqWidget = NULL;
    }
}


/// Functor to get seq_id_handle (used for async execution)
///
struct CGetBioseqHandle_local
{
    CBioseq_Handle operator()(SConstScopedObject obj)
    {
        const CSeq_id* id= dynamic_cast<const CSeq_id*>(obj.object.GetPointer());
        CScope* scope = obj.scope.GetPointer();

        CBioseq_Handle handle;
        if (id  &&  scope) {
            handle = scope->GetBioseqHandle(*id);
            // resolving the seq-id in other form to avoid
            // potential future network connection
            CSeq_id_Handle idh = handle.GetSeq_id_Handle();
            idh = sequence::GetId(*idh.GetSeqId(), *scope, sequence::eGetId_Best);
        }
        return handle;
    }
};


/// Functor to initialize data source (used for async execution)
///
struct CInitDS_local
{
    struct S_InputObj {
        S_InputObj(SConstScopedObject obj, CSeqGraphicWidget* widget)
            : m_Obj(obj)
            , m_Widget(widget)
        {}

        SConstScopedObject m_Obj;
        CSeqGraphicWidget* m_Widget;
    };

    bool operator()(S_InputObj& input_obj)
    {
        input_obj.m_Widget->InitDataSource(input_obj.m_Obj);
        return true; // always return true to satisfy GUI_AsyncExec1
    }
};


//TODO this function is too long
bool CViewGraphic::InitView(TConstScopedObjects& objects, const objects::CUser_object*)
{
    CScope* scope = NULL;
    const CObject* object = NULL;

    wxBusyCursor wait;

    x_MergeObjects( objects );

    if(objects.size() == 1) {
        object = objects[0].object.GetPointer();
        scope = objects[0].scope.GetPointer();

        const CSeq_loc*  loc = dynamic_cast<const CSeq_loc*> (object);
        const CSeq_id*  sid = dynamic_cast<const CSeq_id*> (object);
        const CBioseq*  seq = dynamic_cast<const CBioseq*> (object);

        if (loc) {
            sid = &sequence::GetId(*loc, scope);
        } else if (seq) {
            sid  = seq->GetFirstId();
        }

        if (sid) {
            m_SeqId.Reset(sid);

            SConstScopedObject s_obj(*m_SeqId, *scope);
            CGetBioseqHandle_local func;
            CBioseq_Handle handle;
            try {
                handle = 
                    GUI_AsyncExec1<CBioseq_Handle,
                    SConstScopedObject,
                    CGetBioseqHandle_local>(func, s_obj);
            } catch (CException& e) {
                string str("Failed to retrieve sequence: ");
                str += e.GetMsg();
                wxMessageBox(ToWxString(str));
                return false;
            }

            if ( !handle ) {
                string str("Graphical view: failed to retrieve sequence for id ");
                m_SeqId->GetLabel(&str);
                wxMessageBox(ToWxString(str));
                return false;
            }

            CIRef<CProjectService> srv =
                m_Workbench->GetServiceByType<CProjectService>();

            _ASSERT(srv);

            CRef<CWorkspaceConstPrx> workspace = srv->GetWorkspaceConst();
            CRef<CProjectPrx> project = workspace->FindProjectByScope(scope);

            _ASSERT(project);

            // type info for retrieving view settings
            string view_type = GetTypeDescriptor().GetHelpId();
            string view_ins_id;
            m_SeqId->GetLabel(&view_ins_id, CSeq_id::eFasta);
            string settings = project->GetViewSettings(view_type, view_ins_id);
            LOG_POST(Info << "Retrieve siew settings: " << settings);

            typedef vector<string> TToken;
            TToken setting_set;
            NStr::Tokenize(settings, "&", setting_set);
            map<string, string> setting_map;
            ITERATE (TToken, iter, setting_set) {
                string curr_setting = NStr::TruncateSpaces(*iter);
                TToken tokens;
                NStr::Tokenize(curr_setting, "=", tokens);
                if (tokens.size() == 2) {
                    setting_map.insert(map<string, string>::value_type(tokens[0],
                        tokens[1]));
                }
            }

            try {
                m_SeqWidget->UpdateConfig();
                LOG_POST(Info <<
                    "CViewGraphical::InitView() - finish with loading configuration!");
            } catch (CException&) {
                wxMessageBox(wxT("Something is wrong with configuration!"),
                    wxT("Failed to load configuration for Graphical view"),
                    wxOK | wxICON_EXCLAMATION);
                return false;
            }
            
            try {
                SConstScopedObject obj(*m_SeqId, *scope);

                // initialize the data source using async call to 
                // avoid completely blocking application without
                // users' attention.
                CInitDS_local::S_InputObj input_obj(obj, m_SeqWidget);
                CInitDS_local init_ds_func;
                GUI_AsyncExec1<bool, CInitDS_local::S_InputObj&,
                    CInitDS_local>(init_ds_func, input_obj);

                // set input object to initialize the view
                m_SeqWidget->SetInputObject(obj);
                string id_desc;
                CLabel::GetLabel(*m_SeqId, &id_desc, CLabel::eContent, scope);
                LOG_POST(Info <<
                    "CViewGraphical::InitView() - load sequence: " << id_desc);
            } catch (CException&) {
                x_ReportInvalidInputData(objects);
                return false;
            }
   

            if (setting_map.count("m") != 0) {
                m_SeqWidget->SetMarkers(setting_map["m"]);
            }

            // TODO move this block into a separate function
            // set our visible range
            if (loc  &&  !loc->IsWhole()) {
                TSeqRange range = loc->GetTotalRange();
                m_SeqWidget->ZoomOnRange(range, 0);
            } else if (setting_map.count("v") != 0) {
                TToken tokens;
                NStr::Tokenize(setting_map["v"], "-", tokens);
                if ( !tokens.empty() ) {
                    try {
                        TSeqPos from =
                            NStr::StringToSizet(NStr::TruncateSpaces(tokens[0]));
                        TSeqPos to = from;
                        if (tokens.size() > 1) {
                            to = NStr::StringToSizet(NStr::TruncateSpaces(tokens[1]));
                            if (to < from) {
                                to = from;
                            }
                        }
                        m_SeqWidget->ZoomOnRange(TSeqRange(from, to), 0);
                    } catch (CException&) {
                    }
                }
            }


            //LOG_POST("CViewGraphic::InitView() finished init" << int(1000 * sw.Elapsed()) << " ms");

            x_AttachToProject(*project);

            x_UpdateContentLabel();

            return true;
        }

        // cannot represent the data
        x_ReportInvalidInputData(objects);
        return false;
    } else {
        return false;
    }
}


void CViewGraphic::OnProjectChanged(CProjectViewEvent& evt)
{
    if (evt.GetSubtype() == CProjectViewEvent::eData  ||
        evt.GetSubtype() == CProjectViewEvent::eBothPropsData) {
        m_SeqWidget->OnDataChanged();
        m_SeqWidget->Refresh();
    }
}


bool CViewGraphic::NeedSaveSettings() const
{
    return m_SeqWidget->IsDirty();
}


wxWindow* CViewGraphic::GetWindow()
{
    _ASSERT(m_SeqWidget);
    return m_SeqWidget;
}


string CViewGraphic::GetDMContextName()
{
    return GetClientLabel();
}


const CViewTypeDescriptor& CViewGraphic::GetTypeDescriptor() const
{
    return s_ViewGraphicTypeDescr;
}


void CViewGraphic::GetSelection(TConstScopedObjects& objs) const
{
    if (m_SeqWidget  &&  x_HasProject()) {
        CScope* scope = x_GetScope();

        /// object selection
        TConstObjects sel_objs;
        m_SeqWidget->GetObjectSelection(sel_objs);
        ITERATE (TConstObjects, it, sel_objs)    {
            objs.push_back(SConstScopedObject(**it, *scope));
        }

        /// range selection
        const CSeqGraphicWidget::TRangeColl& coll = m_SeqWidget->GetRangeSelection();
        if (coll.size()) {
            CRef<CSeq_loc> loc = CSeqUtils::CreateSeq_loc(*m_SeqId, coll);
            if (loc) {
                objs.push_back(SConstScopedObject(*loc, *scope));
            }
        }
    }
}


void CViewGraphic::GetSelection(CSelectionEvent& evt) const
{
    if (m_SeqWidget  &&  x_HasProject()) {
        // add range selection
        const CSeqGraphicWidget::TRangeColl& cl_sel
            = m_SeqWidget->GetRangeSelection();
        evt.AddRangeSelection(*m_SeqId, cl_sel);

        // add object selection
        TConstObjects sel_objs;
        m_SeqWidget->GetObjectSelection(sel_objs);
        evt.AddObjectSelection(sel_objs);
    }
}


// handles incoming selection broadcast (overriding CView virtual function)
void CViewGraphic::x_OnSetSelection(CSelectionEvent& evt)
{
    CScope* scope = x_GetScope();

    if (evt.HasRangeSelection()) { // set range selection
        CSeqGraphicWidget::TRangeColl coll;
        evt.GetRangeSelection(*m_SeqId, *scope, coll);
        // set selection
        if (coll.size()) {
            m_SeqWidget->ResetRangeSelection();
            m_SeqWidget->SetRangeSelection(coll);
        }
    }

    if (evt.HasObjectSelection()) {
        TConstObjects sel_objs;
        evt.GetAllObjects(sel_objs);

        //CScope::TIds ids = scope.GetIds(*m_SeqId);
        m_SeqWidget->ResetObjectSelection();
        ITERATE (TConstObjects, it_obj, sel_objs) {
            if (typeid(CSeq_id) == typeid(*it_obj)) {
                m_SeqWidget->SelectObject(*it_obj, false);
            } else if (const CSerialObject* nc_obj =
                dynamic_cast<const CSerialObject*>(it_obj->GetPointer()) ) {
                CTypeConstIterator<CSeq_id> id_iter(*nc_obj);
                for ( ;  id_iter;  ++id_iter) {
                    if(evt.MatchIdWithId(*m_SeqId, *scope, *id_iter, evt.GetScope())) {
                        m_SeqWidget->SelectObject(*it_obj, true); // scope matched
                        break;
                    }
                }
                // set selection with verified set to false
                m_SeqWidget->SelectObject(*it_obj, false);
            } else if (const CIdLoc* id_loc =
                dynamic_cast<const CIdLoc*>(it_obj->GetPointer()) ) {
                m_SeqWidget->SelectObject(id_loc->m_Id, false);
            } else { // log message
                LOG_POST(Warning << "The selected objects are not understood!");
            }
        }
    }/* else {
        m_SeqWidget->ResetObjectSelection();
    }*/
}


const CObject* CViewGraphic::x_GetOrigObject() const
{
    return m_SeqId.GetPointer();
}


void CViewGraphic::OnVisibleRangeChanged(const CVisibleRange& vrange,
                                         IVisibleRangeClient* /*source*/)
{
    IViewManagerService::TViews siblings;
    m_Workbench->GetViewManagerService()->GetViews(siblings);
    x_UpdateVisRangeLayout(siblings);

    try {
        if( vrange.Match( *m_SeqId ) ) {
            TSeqRange this_range = m_SeqWidget->GetVisibleSeqRange();
            CBioseq_Handle bsh = x_GetScope()->GetBioseqHandle(*m_SeqId);
            if (vrange.Clamp(bsh, this_range)) {
                m_SeqWidget->ZoomOnRange(this_range, CSeqGraphicWidget::fSaveRange);
            }
        }
    } catch( std::exception& e ){
        LOG_POST(Error 
            << "CViewGraphic::OnVisibleRangeChanged(): exception caught "
            << e.what() );
    }
}

void CViewGraphic::x_UpdateVisRangeLayout(const IViewManagerService::TViews& views)
{
    CSeqGlyph::TObjects tmp_objs;
    ITERATE( IViewManagerService::TViews, iter, views ){
        const IView* iview = *iter;
        const CProjectView* view_p = dynamic_cast<const CProjectView*> (iview);
        try {
            if (view_p  &&  view_p->GetId() != GetId()) {  // not self
                CRgbaColor color;
                string label;
                list<TSeqRange> vis_ranges;
                view_p->GetReflectionInfo(*m_SeqId, vis_ranges, color, label);
                if ( !vis_ranges.empty() ) {
                    CRef<CSeqGlyph> vref(new CVisRangeGlyph(view_p,
                        vis_ranges, color, label));
                    tmp_objs.push_back(vref);
                }
            }
        } catch( std::exception& e ){
            LOG_POST(Error 
                << "CViewGraphic::x_UpdateVisRangeLayout(): exception caught " 
                << e.what() );
        }
    }
    
    m_SeqWidget->SetExternalGlyphs(tmp_objs);
}


void CViewGraphic::GetVisibleRanges(CVisibleRange& vrange) const
{
    TSeqRange range = m_SeqWidget->GetVisibleSeqRange();
    CRef<CSeq_loc> loc(new CSeq_loc());
    loc->SetInt().SetFrom(range.GetFrom());

    loc->SetInt().SetTo  (range.GetTo());
    loc->SetId(*m_SeqId);
    vrange.AddLocation(*loc);
}


void CViewGraphic::OnWidgetRangeChanged()
{
    IViewManagerService::TViews siblings;
    m_Workbench->GetViewManagerService()->GetViews(siblings);
    x_UpdateVisRangeLayout(siblings);
}


void CViewGraphic::OnWidgetDataChanged(CEvent* evt)
{
    _ASSERT(evt);
    m_SeqWidget->OnMarkerChanged();
}


void CViewGraphic::OnViewAttached( IView& /*view*/ )
{
    IViewManagerService::TViews siblings;
    m_Workbench->GetViewManagerService()->GetViews( siblings );
    x_UpdateVisRangeLayout( siblings );
}

void CViewGraphic::OnViewReleased(IView& view)
{
    IViewManagerService::TViews siblings;
    m_Workbench->GetViewManagerService()->GetViews(siblings);
    const CProjectView* view_u = dynamic_cast<CProjectView*>(&view);
    NON_CONST_ITERATE(IViewManagerService::TViews, iter, siblings) {
        const IView* iview = *iter;
        const CProjectView* view_p = dynamic_cast<const CProjectView*> (iview);
        if (view_u  &&  view_p  &&  view_p->GetId() == view_u->GetId()) {
            siblings.erase(iter);
            break;
        }
    }

    x_UpdateVisRangeLayout(siblings);
}


void  CViewGraphic::SetWorkbench(IWorkbench* workbench)
{
    if(workbench) {
        // connect
        CIRef<CDataMiningService> dm_srv =
            workbench->GetServiceByType<CDataMiningService>();
        if (dm_srv) {
            dm_srv->AttachContext(*this);
        }
    }
    else {
        // disconnect
        CIRef<CDataMiningService> dm_srv =
            m_Workbench->GetServiceByType<CDataMiningService>();
        if (dm_srv) {
            dm_srv->DetachContext(*this);
        }
    }
    CProjectView::SetWorkbench(workbench);
    
}


CRef<CSeq_loc> CViewGraphic::GetSearchLoc()
{
    const CSeqGraphicWidget::TRangeColl& coll = 
        m_SeqWidget->GetRangeSelection();
    
    CRef<CSeq_loc> loc(new CSeq_loc());
    
    // if we have range selection use it for search
    if (coll.size()) {
        loc = CSeqUtils::CreateSeq_loc(*m_SeqId, coll);
    }
    else {
        CRef<CSeq_id> id(new CSeq_id());
        id->Assign(*m_SeqId);
        loc->SetWhole(*id);
    }

    return loc;
}


CRef<CScope> CViewGraphic::GetSearchScope()
{
    return m_Scope;
}


///////////////////////////////////////////////////////////////////////////////
/// CViewGraphicFactory
string  CViewGraphicFactory::GetExtensionIdentifier() const
{
    static string sid("view_graphic_factory");
    return sid;
}


string  CViewGraphicFactory::GetExtensionLabel() const
{
    static string slabel("View Graphic Factory");
    return slabel;
}


void CViewGraphicFactory::RegisterIconAliases(wxFileArtProvider& provider)
{
    string alias = GetViewTypeDescriptor().GetIconAlias();
    provider.RegisterFileAlias(ToWxString(alias), wxT("graphical_view.png"));
}


void CViewGraphicFactory::RegisterCommands (CUICommandRegistry& cmd_reg,
                                            wxFileArtProvider& provider)
{
    CSeqGraphicWidget::RegisterCommands(cmd_reg, provider);
}


const CProjectViewTypeDescriptor&
CViewGraphicFactory::GetProjectViewTypeDescriptor() const
{
    return s_ViewGraphicTypeDescr;
}


IView* CViewGraphicFactory::CreateInstance() const
{
    return new CViewGraphic();
}


IView* CViewGraphicFactory::CreateInstanceByFingerprint(const TFingerprint& /*fingerprint*/) const
{
    return NULL;
}

bool CViewGraphicFactory::IsCompatibleWith( const CObject& object, objects::CScope& )
{
    const type_info& type = typeid(object);

    return
        typeid(CSeq_loc) == type 
        || typeid(CSeq_id) == type 
        || typeid(CBioseq) == type
    ;
}


int CViewGraphicFactory::TestInputObjects( TConstScopedObjects& objects )
{
    bool found_good = false;
    bool found_bad = false;
    for( size_t i = 0; i < objects.size(); i++ ){
        if( IsCompatibleWith( *objects[i].object, *objects[i].scope ) ){
            found_good = true;
        } else {
            found_bad = true;
        }
    }

    if( found_good ){
        return fCanShowSeparated | (found_bad ? fCanShowSome : fCanShowAll);
    }
    return 0; // can show nothing
}


bool CViewGraphic::x_MergeObjects( TConstScopedObjects& objects )
{
    if( objects.size() < 2 ){
        return true;
    }
    // verify that all the objects has the same seq-id
    typedef map<CSeq_id_Handle, TSeqRange> TRanges;
    TRanges ranges;
    ITERATE( TConstScopedObjects, iter, objects ){
        const CSeq_loc* loc = dynamic_cast<const CSeq_loc*> (iter->object.GetPointer());
        if (loc) {
            CScope* scope = const_cast<CScope*>(iter->scope.GetPointer());
            CSeq_id_Handle idh = sequence::GetIdHandle(*loc, scope);
            TSeqRange range = loc->GetTotalRange();
            ranges[idh] += range;
        }
    }
    if (ranges.size() == 1) {
        CRef<objects::CScope> scope = objects[0].scope;
        CRef<CSeq_loc> loc(new CSeq_loc);
        TRanges::iterator iter = ranges.begin();
        loc->SetInt().SetFrom(iter->second.GetFrom());
        loc->SetInt().SetTo  (iter->second.GetTo());
        loc->SetId(*iter->first.GetSeqId());

        //! Should be better incapsulation of m_OutputObjects
        objects.clear();
        objects.push_back(
            SConstScopedObject(loc.GetObject(), scope.GetObject())
        );
        return true;
    } else {
        return false;
    }
}


void CViewGraphic::x_SaveViewSettings()
{
    if (NeedSaveSettings()) {
        string settings = "v=";
        settings += NStr::NumericToString(m_SeqWidget->GetVisibleSeqRange().GetFrom());
        settings += "-";
        settings += NStr::NumericToString(m_SeqWidget->GetVisibleSeqRange().GetTo());
        string markers = m_SeqWidget->GetMarkers();
        if ( !markers.empty() ) {
            settings += "&m=" + markers;
        }

        CIRef<CProjectService> srv =
            m_Workbench->GetServiceByType<CProjectService>();

        _ASSERT(srv);

        CRef<CWorkspaceConstPrx> workspace = srv->GetWorkspaceConst();
        CRef<CProjectPrx> project = workspace->FindProjectById(m_ProjectId);

        _ASSERT(project);

        string view_type = GetTypeDescriptor().GetHelpId();
        string view_ins_id;
        m_SeqId->GetLabel(&view_ins_id, CSeq_id::eFasta);

        project->SaveViewSettings(view_type, view_ins_id, settings);
        project->GetDocument()->SetDirty(true);
        LOG_POST(Info << "Save view settings: " << settings);

        m_SeqWidget->SetDirty(false);
    }
}


END_NCBI_SCOPE
