/*  $Id: wig_load_job.cpp 26141 2012-07-25 19:42:11Z wuliangs $
* ===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data,  the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties,  express or implied,  including
*  warranties of performance,  merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================
*
* Authors:  Josh Cherry, Roman Katargin
*/


#include <ncbi_pch.hpp>

#include <gui/packages/pkg_sequence/wig_load_job.hpp>

#include <objmgr/object_manager.hpp>

#include <util/line_reader.hpp>

#include <objtools/readers/reader_base.hpp>
#include <objtools/readers/wiggle_reader.hpp>
#include <objtools/readers/idmapper.hpp>

#include <objects/seq/Annot_descr.hpp>
#include <objects/seq/Annotdesc.hpp>
#include <objects/general/User_object.hpp>
#include <objects/general/User_field.hpp>
#include <objects/general/Object_id.hpp>

#include <wx/txtstrm.h>
#include <wx/filename.h>

#include <gui/widgets/wx/wx_utils.hpp>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

CWigLoadJob::CWigLoadJob(CProjectService* service,
                         const CWigLoadParams& params,
                         const vector<wxString>& filenames)
: CDataLoadingAppJob(service), m_Params(params), m_FileNames(filenames)
{
    m_Descr = "Loading WIG Files";
}

namespace {
    class CErrorContainer : public CErrorContainerLenient
    {
    public:
        CErrorContainer(size_t maxErrors) : m_MaxErrors(maxErrors) {}
        virtual bool PutError(const ILineError& err)
        {
            return Count() < m_MaxErrors && CErrorContainerLenient::PutError(err);
        }
    private:
        const size_t m_MaxErrors;
    };
}

void CWigLoadJob::x_CreateProjectItems()
{
    static string contexts[] = {
        "rn4", "mm7", "mm8", "mm9", "hg17", "hg18", "hg19" };

    auto_ptr<IIdMapper> idMapper;

    string mapContext;
    int mapIndex = m_Params.GetMappingContext();
    if (mapIndex > 0 && mapIndex <= int(sizeof(contexts)/sizeof(contexts[0])))
        idMapper.reset(new CIdMapperBuiltin(contexts[mapIndex - 1]));

    CWiggleReader reader;

    ITERATE(vector<wxString>, it, m_FileNames) {
        m_CurrentFile = *it;
        const wxString& fn = *it;
        CDataLoadingJob::CCompressStreamHolder  istream_hold(fn);
        CNcbiIstream* istr_p = istream_hold.GetIStreamP();

        CStreamLineReader lineReader(*istr_p);

        vector<TAnnotRef> annots;
        CRef<CErrorContainer> errCont(new CErrorContainer((size_t)m_Params.GetNumErrors()));
        reader.ReadSeqAnnots(annots, lineReader, errCont);
        x_UpdateHTMLResults(*errCont);

        int index = 1;
        NON_CONST_ITERATE (vector<TAnnotRef>, a_iter, annots) {
            TAnnotRef& annot = *a_iter;

            /// By default, Wiggle reader convert wig file into seq-tables.
            /// Usually, the original data are graphs, we need to mark them
            /// as 'graph'
            CRef<CUser_object> user_obj;
            if (annot->CanGetDesc()) {
                CSeq_annot::TDesc::Tdata& descs = annot->SetDesc().Set();
                CSeq_annot::TDesc::Tdata::iterator d_iter = descs.begin();
                while (d_iter != descs.end()) {
                    if ((*d_iter)->IsUser()  &&  (*d_iter)->GetUser().GetType().IsStr()  &&
                        (*d_iter)->GetUser().GetType().GetStr() == "Track Data") {
                            user_obj.Reset(&(*d_iter)->SetUser());
                            break;
                    }
                    ++d_iter;
                }
            }

            if ( !user_obj ) {
                user_obj.Reset(new CUser_object);
                user_obj->SetType().SetStr("Track Data");
                annot->AddUserObject(*user_obj);
            }
            user_obj->AddField("track type", string("graph"));

            if (idMapper.get()) {
                CRef<CSerialObject> tmp(annot);
                idMapper->MapObject(*tmp);
            }

            CRef<CProjectItem> item(new CProjectItem());
            item->SetItem().SetAnnot(*annot);

            string label;
            bool has_name = false;
            if (annot->CanGetDesc()) {
                const CAnnot_descr& desc = annot->GetDesc();
                if (desc.CanGet()) {
                    const CAnnot_descr::Tdata& data = desc.Get();
                    CSeq_annot::TDesc::Tdata::const_iterator desc_iter = data.begin();
                    for ( ;  desc_iter != data.end();  ++desc_iter) {
                        if ((*desc_iter)->IsTitle() ) {
                            label = (*desc_iter)->GetTitle();
                            break;
                        }
                        if ((*desc_iter)->IsName() ) {
                            label = (*desc_iter)->GetName();
                            has_name = true;
                        }
                    }
                }
            }

            wxFileName wx_fn(m_CurrentFile);
            string backup_name = ToStdString(wx_fn.GetFullName());
            if (label.empty()) {
                label = backup_name;
                if (annots.size() > 1) {
                    CNcbiOstrstream ostr;
                    ostr << " (WIG Seq-annot " << index << ")";
                    label += CNcbiOstrstreamToString(ostr);
                }
            }

            if ( !has_name ) {
                annot->SetNameDesc(backup_name);
            }

            item->SetLabel(label);
            m_Items.push_back(item);

            ++index;
        }
    }
}

void CWigLoadJob::x_UpdateHTMLResults(IErrorContainer& errCont)
{
    if (errCont.Count() == 0)
        return;

    wxTextOutputStream os(m_Strstrm);

    os << wxT("<h3>File: ") << m_CurrentFile << wxT("</h3>");

    for (size_t i = 0; i < errCont.Count(); ++i) {
        const ILineError& lerror = errCont.GetError(i);
        os << wxT("Line ") << lerror.Line() << wxT(". ") << ToWxString(lerror.Message()) << wxT("<br />");
    }
}

void CWigLoadJob::GetHTMLResults(wxString& title, wxString& html)
{
    title = wxT("WIG import results");
    html = m_Strstrm.GetString();
}

END_NCBI_SCOPE

