/*  $Id: append_mod_to_org.cpp 27751 2013-03-26 18:50:22Z filippov $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data,  the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties,  express or implied,  including
 *  warranties of performance,  merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Igor Filippov
 */

#include <ncbi_pch.hpp>
#include <gui/packages/pkg_sequence_edit/append_mod_to_org.hpp>

BEGIN_NCBI_SCOPE
using namespace objects;


void CAppendModToOrg::ApplyToBioSource (const CBioSource& biosource, CCmdComposite* composite, const CSeq_entry* se, const CSeqdesc* seqDesq, const CSeq_feat* seqFeat) 
{
    CRef<CSerialObject> edited_object;
    edited_object.Reset((CSerialObject*)CBioSource::GetTypeInfo()->Create());
    edited_object->Assign(biosource);
    CBioSource& edited_biosource = dynamic_cast<CBioSource&>(*edited_object);
    bool modified = false;
    ApplyToOrgMod (edited_biosource,modified);
    if (modified)
    {
        CChangeBioSourceCommand* cmd = new CChangeBioSourceCommand();
        CObject* actual = (CObject*) &biosource;
        cmd->Add(actual, CConstRef<CObject>(edited_object));
        composite->AddCommand(*cmd);
    }
}

void CAppendModToOrg::ApplyToCSeq_entry (const CSeq_entry& se, CCmdComposite* composite)  
{

    FOR_EACH_SEQDESC_ON_SEQENTRY (it, se) {
        if ((*it)->IsSource()) {
            ApplyToBioSource ((*it)->GetSource(), composite, &se, *it, NULL);
        }
    }

    if (se.IsSet()) {
        FOR_EACH_SEQENTRY_ON_SEQSET (it, se.GetSet()) {
            ApplyToCSeq_entry (**it, composite);
        }
    }
}


void  CAppendModToOrg::apply_to_seq_and_feat(objects::CSeq_entry_Handle tse, CCmdComposite* composite)
{
    ApplyToCSeq_entry (*(tse.GetCompleteSeq_entry()), composite);
    CFeat_CI feat (tse, SAnnotSelector(CSeqFeatData::e_Biosrc));
    while (feat) {
        ApplyToBioSource (feat->GetData().GetBiosrc(), composite, NULL, NULL,&(feat->GetOriginalFeature())); 
        ++feat;
    }
}

void CAppendModToOrg::ApplyToOrgMod(CBioSource& biosource, bool& modified) // biosource->org->orgname->mod->(subtype,subname)
{

    // TODO constraints

    bool ok_to_add = false;
    if (!m_Only_sp && !m_Only_cf && !m_Only_aff)
        ok_to_add = true;
  
    if (m_Only_sp && biosource.IsSetOrg() && biosource.GetOrg().IsSetTaxname() &&  biosource.GetOrg().GetTaxname().find(" sp.") != string::npos)
        ok_to_add = true;
  
    if (m_Only_cf  && biosource.IsSetOrg() && biosource.GetOrg().IsSetTaxname() &&  biosource.GetOrg().GetTaxname().find(" cf.") !=  string::npos)
        ok_to_add = true;
  
    if (m_Only_aff && biosource.IsSetOrg() && biosource.GetOrg().IsSetTaxname() &&  biosource.GetOrg().GetTaxname().find(" aff.") != string::npos)
        ok_to_add = true;

    if (m_No_taxid)
    {
        if  (biosource.IsSetOrg() && biosource.GetOrg().IsSetDb())       
            {
                COrg_ref::TDb new_db;
                FOR_EACH_DBXREF_ON_ORGREF(db,biosource.SetOrg())
                    if ((*db)->IsSetDb() && NStr::Equal((*db)->GetDb(),"taxon",NStr::eNocase))
                        ok_to_add = false;
            }
    }

    if (!ok_to_add)
        return;
  


    string str_to_add;
    if (m_IsOrgMod)
    {
        if (biosource.IsSetOrg() && biosource.GetOrg().IsSetOrgname() && biosource.GetOrg().GetOrgname().IsSetMod())
        {
            FOR_EACH_ORGMOD_ON_BIOSOURCE(orgmod,biosource)
                if ((*orgmod)->IsSetSubtype() && (*orgmod)->IsSetSubname() && (*orgmod)->GetSubtype() == m_Subtype)
                {
                    str_to_add = (*orgmod)->GetSubname();
                }
        }
    }
    else
    {
        if (biosource.IsSetSubtype())
        {
            FOR_EACH_SUBSOURCE_ON_BIOSOURCE(subsource,biosource)
                if ((*subsource)->IsSetSubtype() && (*subsource)->IsSetName() && (*subsource)->GetSubtype() == m_Subtype)
                {
                    str_to_add = (*subsource)->GetName();
                }
        }
    }
    if (!str_to_add.empty())
    {
        if (biosource.IsSetOrg() && biosource.GetOrg().IsSetTaxname())
        {
            string& taxname = biosource.SetOrg().SetTaxname();
            unsigned pos = taxname.find(" (");
            if (pos !=  string::npos)
            {
                string newtaxname = taxname.substr (0,pos);
                biosource.SetOrg().SetTaxname(newtaxname);
                modified = true;
            }
        }
        
        if (m_IsOrgMod && (m_Subtype == COrgMod::eSubtype_bio_material || m_Subtype == COrgMod::eSubtype_culture_collection || m_Subtype == COrgMod::eSubtype_specimen_voucher))
        {
            NStr::ReplaceInPlace(str_to_add,":"," ");
        }
        
        bool influenza = false;
        if (biosource.IsSetOrg() && biosource.GetOrg().IsSetTaxname())
        {
            string& taxname = biosource.SetOrg().SetTaxname();
            if (NStr::Equal(taxname,"Influenza A virus",NStr::eNocase) || NStr::Equal(taxname,"Influenza B virus",NStr::eNocase))
                influenza = true;
        }

        string str;
        if (biosource.IsSetOrg() && biosource.GetOrg().IsSetTaxname())
        {
            string& taxname = biosource.SetOrg().SetTaxname();
            str = taxname + " ";
        }
        if (!m_Abbreviation.empty() && m_Use_Abbreviation)
            str += m_Abbreviation + " ";
        if (influenza)
            str += "(";
        str += str_to_add;
        if (influenza)
        {
            str += ")";
            if (NStr::StartsWith(str, "Influenza A virus", NStr::eNocase))             //str = FixInfluenzaVirusName(str);
            {
                NStr::ReplaceInPlace (str, "  ", " ");
                NStr::ReplaceInPlace (str, "( ", "(");
                NStr::ReplaceInPlace (str, " )", ")");
                NStr::TruncateSpacesInPlace (str);
            }
        }

        // SetTaxNameAndRemoveTaxRef (biop->org, str);
        bool remove_taxref = false;
        bool is_set_taxname =  biosource.IsSetOrg() && biosource.GetOrg().IsSetTaxname();
        if (!is_set_taxname || str.empty() || str !=  biosource.GetOrg().GetTaxname())
            remove_taxref = true;
        if (biosource.IsSetOrg())
        {
            biosource.SetOrg().SetTaxname() = str;
            modified = true;
        }
        if (remove_taxref)
        {
            if (biosource.IsSetOrg() && biosource.GetOrg().IsSetCommon())
            {
                biosource.SetOrg().ResetCommon();
                modified = true;
            }

            if  (biosource.IsSetOrg() && biosource.GetOrg().IsSetDb())             //RemoveTaxRef (orp);
            {
                COrg_ref::TDb new_db;
                bool modified_here = false;
                FOR_EACH_DBXREF_ON_ORGREF(db,biosource.SetOrg())
                    if ((*db)->IsSetDb() && NStr::Equal((*db)->GetDb(),"taxon",NStr::eNocase))
                        modified_here = true;
                    else
                        new_db.push_back(*db);
                if (modified_here)
                {
                    biosource.SetOrg().SetDb() = new_db;
                    modified = true;
                }
            }
        }

        if (biosource.IsSetOrg() && biosource.GetOrg().IsSetOrgname() && biosource.GetOrg().GetOrgname().IsSetMod())  //      RemoveOldName (biop->org);
        {
            COrgName::TMod new_mod;
            bool modified_here = false;
            FOR_EACH_ORGMOD_ON_BIOSOURCE(orgmod,biosource)
                if (!((*orgmod)->IsSetSubtype()  && (*orgmod)->GetSubtype() == COrgMod::eSubtype_old_name))
                    new_mod.push_back(*orgmod);
                else
                    modified_here = true;
            if (modified_here)
            {
                biosource.SetOrg().SetOrgname().SetMod() = new_mod;
                modified = true;
            }
        }
            
    }

   
}

END_NCBI_SCOPE
