/*  $Id: dll_register.cpp 28187 2013-05-22 14:14:21Z bollin $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Andrey Yazhuk, Igor Filippov
 *
 * File Description:
 *      Defines "Sequence Editing" package for GBENCH.
 */

#include <ncbi_pch.hpp>

#include <objects/submit/Seq_submit.hpp>
#include <gui/framework/gui_package.hpp>
#include <gui/framework/pkg_wb_connect.hpp>
#include <gui/core/selection_service_impl.hpp>

#include <gui/widgets/wx/ui_command.hpp>

#include <gui/utils/extension_impl.hpp>
#include <gui/objutils/util_cmds.hpp>

#include <gui/widgets/edit/init.hpp>

#include <gui/core/file_extensions.hpp>
#include <gui/core/open_objects_dlg.hpp>
#include <gui/core/ui_file_load_manager.hpp>
#include <gui/utils/object_loader.hpp>
#include <gui/utils/execute_unit.hpp>
#include <gui/utils/extension_impl.hpp>
#include <gui/utils/app_popup.hpp>
#include <gui/widgets/wx/async_call.hpp>


#include <gui/packages/pkg_sequence_edit/autodef_tool_manager.hpp>
#include <gui/packages/pkg_sequence_edit/subprep_tool.hpp>
#include <gui/packages/pkg_sequence_edit/srcedit_tool.hpp>
#include <gui/packages/pkg_sequence_edit/subprep_view.hpp>
#include <gui/packages/pkg_sequence_edit/btns_tool.hpp>
#include <gui/packages/pkg_sequence_edit/btns_view.hpp>
#include <gui/packages/pkg_sequence_edit/seqid_fix_dlg.hpp>

#include <gui/widgets/edit/edit_object_dlg.hpp>
#include <gui/packages/pkg_sequence_edit/remove_desc_dlg.hpp>
#include <gui/packages/pkg_sequence_edit/remove_sequences_from_alignments.hpp>
#include <gui/packages/pkg_sequence_edit/remove_features_dlg.hpp>
#include <gui/packages/pkg_sequence_edit/remove_structured_comment.hpp>
#include <gui/packages/pkg_sequence_edit/remove_annot.hpp>
#include <gui/packages/pkg_sequence_edit/remove_genomeprojectdb.hpp>
#include <gui/packages/pkg_sequence_edit/add_citsubupdate.hpp>
#include <gui/packages/pkg_sequence_edit/pubdesc_editor.hpp>
#include <gui/packages/pkg_sequence_edit/modify_structured_comment.hpp>
#include <gui/packages/pkg_sequence_edit/remove_bankit_comment.hpp>
#include <gui/packages/pkg_sequence_edit/remove_structuredcomment_keyword.hpp>
#include <gui/packages/pkg_sequence_edit/reorder_sequences_dlg.hpp>
#include <gui/packages/pkg_sequence_edit/convert_comment.hpp>
#include <gui/packages/pkg_sequence_edit/modify_feature_id.hpp>
#include <gui/packages/pkg_sequence_edit/remove_sequences.hpp>
#include <gui/packages/pkg_sequence_edit/pcr-primers-edit.hpp>
#include <gui/packages/pkg_sequence_edit/fix_orgmod_institution.hpp>
#include <gui/packages/pkg_sequence_edit/append_mod_to_org_dlg.hpp>
#include <gui/packages/pkg_sequence_edit/parse_collectiondate_formats.hpp>
#include <gui/packages/pkg_sequence_edit/country_fixup.hpp>
#include <gui/packages/pkg_sequence_edit/bulkfeatureadddlg.hpp>

#include <wx/msgdlg.h>

BEGIN_NCBI_SCOPE

///////////////////////////////////////////////////////////////////////////////
/// Declare the Package Object
class CSeqEditPackage 
    : public IGuiPackage
    , public IUICommandContributor
    , public IMenuContributor
    , public IPackageWorkbenchConnector
{
public:
    CSeqEditPackage();
    virtual string  GetName() const { return "Sequence Editing"; }
    virtual void    GetVersion(size_t& verMajor, size_t& verMinor, size_t& verPatch) const
    {
        verMajor = 1;
        verMinor = 0;
        verPatch = 0;
    }
    virtual bool    Init();
    virtual void    Shut() {}

    // IUICommandContributor
    virtual vector<CUICommand*> GetCommands();
    virtual wxEvtHandler* GetEventHandler();
    
    // IPackageWorkbenchConnector
    virtual void SetWorkbench(IWorkbench* wb);

protected:
    // IMenuContributor
    virtual const SwxMenuItemRec* GetMenuDef() const;

    IWorkbench*     m_Workbench;
};

CSeqEditPackage::CSeqEditPackage()
{
    m_Workbench = NULL;
}


bool CSeqEditPackage::Init()
{
    // Declare you extension here

    // Examples:
    //CExtensionDeclaration("view_manager_service::view_factory",
    //                      new CTextViewFactory());

    //CExtensionDeclaration("ui_data_source_type",
    //                      new CGenBankUIDataSourceType());
    //CExtensionDeclaration ("file_format_loader_manager",
    //                      new CFASTAFormatLoadManager());

    //CExtensionDeclaration("data_mining_tool", new CEntrezSearchTool());

    init_w_edit();
    
    CExtensionDeclaration("ui_algo_tool_manager", new CAutodefToolManager());
    
    // Edit tools
    CExtensionDeclaration("ui_algo_tool_manager", new CSubPrepTool());
    CExtensionDeclaration("ui_algo_tool_manager", new CEditingBtnsTool());
    CExtensionDeclaration("ui_algo_tool_manager", new CSrcEditTool());

    CExtensionDeclaration("view_manager_service::view_factory",
                          new CSubPrepViewFactory());
    CExtensionDeclaration("view_manager_service::view_factory",
                          new CEditingBtnsViewFactory());

    //! Please don't register commands twice. Either you register them directly 
    //! with x_RegisterCommands(), or you use the interface IUICommandContributor.
    //! Please never do both!
    //!
    //x_RegisterCommands();
    return true;
}

void CSeqEditPackage::SetWorkbench(IWorkbench* wb)
{
    m_Workbench = wb;   
}

///////////////////////////////////////////////////////////////////////////////
/// Declare Package Entry Point
extern "C" {
    NCBI_PACKAGEENTRYPOINT_EXPORT IGuiPackage* NCBIGBenchGetPackage()
    {
        return new CSeqEditPackage();
    }
}

class CSequenceEditingEventHandler : public wxEvtHandler
{
    DECLARE_EVENT_TABLE();
public:
    CSequenceEditingEventHandler(IWorkbench* wb = NULL);
    void IsObjectSelected(wxUpdateUIEvent& event);
    // methods
    void RemoveDescriptors(wxCommandEvent& evt);
    void RemoveFeatures(wxCommandEvent& evt);
    void RemoveSeqAnnotIds( wxCommandEvent& event );
    void RemoveBankitComment( wxCommandEvent& event );
    void RemoveDuplicateStructuredComments( wxCommandEvent& event );
    void RemoveSequencesFromAlignments( wxCommandEvent& event );
    void RemoveSequences( wxCommandEvent& event );
    void RemoveSeqAnnotAlign( wxCommandEvent& event );
    void RemoveSeqAnnotGraph( wxCommandEvent& event );
    void RemoveGenomeProjectsDB( wxCommandEvent& event );
    void FixCapitalizationAll( wxCommandEvent& event );
    void FixCapitalizationAuthors( wxCommandEvent& event );
    void FixCapitalizationTitles( wxCommandEvent& event );
    void FixCapitalizationAffiliation( wxCommandEvent& event );
    void FixCapitalizationCountry( wxCommandEvent& event );
    void RemoveUnpublishedPublications( wxCommandEvent& event );
    void RemovePublishedPublications( wxCommandEvent& event );
    void RemoveCollidingPublications( wxCommandEvent& event );
    void RemoveAllPublications( wxCommandEvent& event );
    void RemoveAuthorConsortiums( wxCommandEvent& event );
    void ReverseAuthorNames( wxCommandEvent& event );
    void StripAuthorSuffixes( wxCommandEvent& event );
    void TruncateAuthorMiddleInitials( wxCommandEvent& event );
    void ConvertAuthorToConstortiumWhereLastName( wxCommandEvent& event );
    void ConvertAuthorToConsortiumAll( wxCommandEvent& event );
    void AddCitSubForUpdate( wxCommandEvent& event );
    void FixUsaAndStates( wxCommandEvent& event );
    void RemoveAllStructuredComments( wxCommandEvent& event );
    void RemoveEmptyStructuredComments( wxCommandEvent& event );
    void ModifyStructuredComment( wxCommandEvent& event );
    void RemoveStructuredCommentKeyword( wxCommandEvent& event );
    void AddStructuredCommentKeyword( wxCommandEvent& event );
    void ReorderStructuredComment( wxCommandEvent& event );
    void ConvertComment( wxCommandEvent& event );
    void ParseComment( wxCommandEvent& event );
    void AssignFeatureIds( wxCommandEvent& event );
    void ClearFeatureIds( wxCommandEvent& event );
    void ReassignFeatureIds( wxCommandEvent& event );
    void UniqifyFeatureIds( wxCommandEvent& event );
    void TrimJunkInPrimerSeqs( wxCommandEvent& event );
    void FixiInPrimerSeqs( wxCommandEvent& event );
    void FixOrgModInstitution( wxCommandEvent& event );
    void SwapPrimerSeqName( wxCommandEvent& event );
    void MergePrimerSets( wxCommandEvent& event );
    void SplitPrimerSets( wxCommandEvent& event );
    void AppendModToOrg( wxCommandEvent& event );
    void ParseCollectionDateMonthFirst( wxCommandEvent& event );
    void ParseCollectionDateDayFirst( wxCommandEvent& event );
    void CountryFixupCap( wxCommandEvent& event );
    void CountryFixupNoCap( wxCommandEvent& event );
    void AddCDS( wxCommandEvent& event );
    void AddRNA( wxCommandEvent& event );
    void AddOtherFeature( wxCommandEvent& event );
    void Validate( wxCommandEvent& event );
    void RemoveAllFeatures (wxCommandEvent& event );
    void AddSequences (wxCommandEvent& event );

private:
    ICommandProccessor* m_CmdProccessor;
    CSeq_entry_Handle m_TopSeqEntry;
    CConstRef<objects::CSeq_submit> m_SeqSubmit;
    IWorkbench*     m_Workbench;
    void x_SetUpTSEandUnDoManager(IWorkbench* wb);
};

wxEvtHandler* CSeqEditPackage::GetEventHandler()
{
    return new CSequenceEditingEventHandler(m_Workbench);
}

CSequenceEditingEventHandler::CSequenceEditingEventHandler(IWorkbench* wb)
{
    m_TopSeqEntry.Reset();
    m_CmdProccessor = NULL;
    m_Workbench = wb;
}

void CSequenceEditingEventHandler::x_SetUpTSEandUnDoManager(IWorkbench* wb)
{
    m_TopSeqEntry.Reset();
    m_SeqSubmit.Reset(NULL);
    m_CmdProccessor = NULL;
    if (!wb) return;
    
    TConstScopedObjects objects;
    CIRef<CSelectionService> sel_srv = wb->GetServiceByType<CSelectionService>();
    if (!sel_srv) return;
    sel_srv->GetCurrentSelection(objects);
    if (objects.empty()) return;
    NON_CONST_ITERATE (TConstScopedObjects, it, objects) {
        const CObject* ptr = it->object.GetPointer();
        
        /// CSeq_entry
        const objects::CSeq_entry* seqEntry = dynamic_cast<const objects::CSeq_entry*>(ptr);
        const objects::CBioseq* bioseq = dynamic_cast<const objects::CBioseq*>(ptr);
        const objects::CBioseq_set* bioseq_set = dynamic_cast<const objects::CBioseq_set*>(ptr);
        const objects::CSeq_annot* seqannot = dynamic_cast<const objects::CSeq_annot*>(ptr);
        const objects::CSeq_feat* seqfeat = dynamic_cast<const objects::CSeq_feat*>(ptr);

        objects::CSeq_entry_Handle seh;
        if (seqEntry)   seh = it->scope->GetObjectHandle (*seqEntry, CScope::eMissing_Null).GetTopLevelEntry();
        else if(bioseq) seh = it->scope->GetObjectHandle (*bioseq, CScope::eMissing_Null).GetTopLevelEntry();
        else if(bioseq_set) seh = it->scope->GetObjectHandle (*bioseq_set, CScope::eMissing_Null).GetTopLevelEntry();
        else if(seqannot) seh = it->scope->GetObjectHandle (*seqannot, CScope::eMissing_Null).GetTopLevelEntry();
        else if(seqfeat) seh = it->scope->GetObjectHandle (*seqfeat, CScope::eMissing_Null).GetAnnot().GetTopLevelEntry();
        else {
            const objects::CSeq_submit* seqsubmit = dynamic_cast<const objects::CSeq_submit*>(ptr);
            if (seqsubmit) {
                if (seqsubmit->IsEntrys()) {
                    seh = it->scope->GetSeq_entryHandle(*(seqsubmit->GetData().GetEntrys().front()));
                }
                m_SeqSubmit.Reset(seqsubmit);
            }
        } 
        if (seh) {
            m_TopSeqEntry = seh;
        }
    }
    if (!m_TopSeqEntry) return;
    CIRef<CProjectService> srv = wb->GetServiceByType<CProjectService>();
    if (srv)
    {
        CRef<CWorkspaceConstPrx> workspace = srv->GetWorkspaceConst();
        if (workspace)
        {
            CRef<CProjectPrx> project = workspace->FindProjectByScope(&m_TopSeqEntry.GetScope());
            if (project) 
                m_CmdProccessor = &project->GetDocument()->GetUndoManager(); 
        }
    }

}

void CSequenceEditingEventHandler::IsObjectSelected(wxUpdateUIEvent& event)
{
    x_SetUpTSEandUnDoManager(m_Workbench);              
    event.Enable(m_TopSeqEntry && m_CmdProccessor);
}


// Implementation

enum ESequenceEditingCommands {
    eCmdRemoveDescriptors = eBaseCmdLast + 1000,
    eCmdRemoveFeatures = eBaseCmdLast + 1001,
    eCmdRemoveSeqAnnotIds = eBaseCmdLast + 1002,
    eCmdRemoveBankitComment = eBaseCmdLast + 1003,
    eCmdRemoveDuplicateStructuredComments = eBaseCmdLast + 1004,
    eCmdRemoveSequencesFromAlignments = eBaseCmdLast + 1005,
    eCmdRemoveSequences = eBaseCmdLast + 1006,
    eCmdRemoveAlignments = eBaseCmdLast + 1007,
    eCmdRemoveGraphs = eBaseCmdLast + 1008,
    eCmdRemoveGenomeProjectId = eBaseCmdLast + 1009,
    eCmdFixCapitalizationAll = eBaseCmdLast + 1010,
    eCmdFixCapitalizationAuthors = eBaseCmdLast + 1011,
    eCmdFixCapitalizationTitles = eBaseCmdLast + 1012,
    eCmdFixCapitalizationAffiliation = eBaseCmdLast + 1013,
    eCmdFixCapitalizationCountry = eBaseCmdLast + 1014,
    eCmdRemoveUnpublishedPublications = eBaseCmdLast + 1015,
    eCmdRemovePublishedPublications = eBaseCmdLast + 1016,
    eCmdRemoveCollidingPublications = eBaseCmdLast + 1017,
    eCmdRemoveAllPublications = eBaseCmdLast + 1018,
    eCmdRemoveAuthorConsortiums = eBaseCmdLast + 1019,
    eCmdReverseAuthorNames = eBaseCmdLast + 1020,
    eCmdStripAuthorSuffixes = eBaseCmdLast + 1021,
    eCmdTruncateAuthorMiddleInitials = eBaseCmdLast + 1022,
    eCmdConvertAuthorToConstortiumWhereLastName = eBaseCmdLast + 1023,
    eCmdConvertAuthorToConsortiumAll = eBaseCmdLast + 1024,
    eCmdAddCitSubForUpdate = eBaseCmdLast + 1025,
    eCmdFixUsaAndStates = eBaseCmdLast + 1026,
    eCmdRemoveAllStructuredComments = eBaseCmdLast + 1027,
    eCmdRemoveEmptyStructuredComments = eBaseCmdLast + 1028,
    eCmdModifyStructuredComment = eBaseCmdLast + 1029,
    eCmdRemoveStructuredCommentKeyword = eBaseCmdLast + 1030,
    eCmdAddStructuredCommentKeyword = eBaseCmdLast + 1031,
    eCmdReorderStructuredComment = eBaseCmdLast + 1032,
    eCmdConvertComment = eBaseCmdLast + 1033,
    eCmdParseComment = eBaseCmdLast + 1034,
    eCmdAssignFeatureIds = eBaseCmdLast + 1035,
    eCmdClearFeatureIds = eBaseCmdLast + 1036,
    eCmdReassignFeatureIds = eBaseCmdLast + 1037,
    eCmdUniqifyFeatureIds = eBaseCmdLast + 1038,
    eCmdTrimJunkInPrimerSeqs = eBaseCmdLast + 1039,
    eCmdFixiInPrimerSeqs = eBaseCmdLast + 1040,
    eCmdFixOrgModInstitution = eBaseCmdLast + 1041,
    eCmdSwapPrimerSeqName = eBaseCmdLast + 1042,
    eCmdMergePrimerSets = eBaseCmdLast + 1043,
    eCmdSplitPrimerSets = eBaseCmdLast + 1044,
    eCmdAppendModToOrg = eBaseCmdLast + 1045,
    eCmdParseCollectionDateMonthFirst = eBaseCmdLast + 1046,
    eCmdParseCollectionDateDayFirst = eBaseCmdLast + 1047,
    eCmdCountryFixupCap = eBaseCmdLast + 1048,
    eCmdCountryFixupNoCap = eBaseCmdLast + 1049,
    eCmdApplyCDS = eBaseCmdLast + 1050,
    eCmdApplyAddRNA = eBaseCmdLast + 1051,
    eCmdApplyAddOther = eBaseCmdLast + 1052,
    eCmdValidate = eBaseCmdLast + 1053,
    eCmdRemoveAllFeatures = eBaseCmdLast + 1054,
    eCmdAddSequences,
};

vector<CUICommand*> CSeqEditPackage::GetCommands()
{
    vector<CUICommand*> commands;
    commands.push_back(new CUICommand(eCmdRemoveDescriptors, "Remove Descriptors", "Remove Descriptors from Selected Object", "RDSO"));
    commands.push_back(new CUICommand(eCmdRemoveFeatures, "Remove Features", "Remove Features from Selected Object", "RFSO"));
    commands.push_back(new CUICommand(eCmdRemoveSeqAnnotIds, "Remove Seq-annot IDs", "Remove Seq-annot IDs from Selected Object", "RSAID"));
    commands.push_back(new CUICommand(eCmdRemoveBankitComment, "Remove Bankit Comments", "Remove Bankit Comments from Selected Object", "RBC"));
    commands.push_back(new CUICommand(eCmdRemoveDuplicateStructuredComments, "Remove Duplicate Structured Comments", "Remove Duplicate Structured Comments from Selected Object", "RDSC"));
    commands.push_back(new CUICommand(eCmdRemoveSequencesFromAlignments, "Remove Sequences From Alignments", "Remove Sequences From Alignments from Selected Object", "RSFASO"));
    commands.push_back(new CUICommand(eCmdRemoveSequences, "Remove Sequences From Record", "Remove Sequences from Selected Object", "RSSO"));
    commands.push_back(new CUICommand(eCmdRemoveAlignments, "Remove Alignments", "Remove Alignments from Selected Object", "RASO"));
    commands.push_back(new CUICommand(eCmdRemoveGraphs, "Remove Graphs", "Remove Graphs from Selected Object", "RGSO"));
    commands.push_back(new CUICommand(eCmdRemoveGenomeProjectId, "Remove All Genome Project Ids", "Remove All Genome Project Ids from Selected Object", "RGPIDSO"));
    commands.push_back(new CUICommand(eCmdFixCapitalizationAll, "Fix All Capitalization", "Fix All Capitalization for Selected Object", "FACSO"));
    commands.push_back(new CUICommand(eCmdFixCapitalizationAuthors, "Fix Capitalization in Authors", "Fix Capitalization in Authors for Selected Object", "FCASO"));
    commands.push_back(new CUICommand(eCmdFixCapitalizationTitles, "Fix Capitalization in Titles", "Fix Capitalization in Titles for Selected Object", "FCTSO"));
    commands.push_back(new CUICommand(eCmdFixCapitalizationAffiliation, "Fix Capitalization in Affiliations", "Fix Capitalization in Affiliations for Selected Object", "FCAfSO"));
    commands.push_back(new CUICommand(eCmdFixCapitalizationCountry, "Fix Capitalization in Countries", "Fix Capitalization in Countries for Selected Object", "FCCSO"));
    commands.push_back(new CUICommand(eCmdRemoveUnpublishedPublications, "Remove Unpublished Publications", "Remove Unpublished Publications from Selected Object", "RUPSO"));
    commands.push_back(new CUICommand(eCmdRemovePublishedPublications, "Remove Published Publications", "Remove Published Publications from Selected Object", "RPPSO"));
    commands.push_back(new CUICommand(eCmdRemoveCollidingPublications, "Remove Colliding Publications", "Remove Colliding Publications from Selected Object", "RCPSO"));
    commands.push_back(new CUICommand(eCmdRemoveAllPublications, "Remove All Publications", "Remove All Publications from Selected Object", "RAPSO"));
    commands.push_back(new CUICommand(eCmdRemoveAuthorConsortiums, "Remove Author Consortiums", "Remove Author Consortiums from Selected Object", "RACSO"));
    commands.push_back(new CUICommand(eCmdReverseAuthorNames, "Reverse Author Names", "Reverse Author Names for Selected Object", "RANSO"));
    commands.push_back(new CUICommand(eCmdStripAuthorSuffixes, "Strip Author Suffixes", "Strip Author Suffixes from Selected Object", "SASSO"));
    commands.push_back(new CUICommand(eCmdTruncateAuthorMiddleInitials, "Truncate Author Middle Initials", "Truncate Author Middle Initials from Selected Object", "TAMISO"));
    commands.push_back(new CUICommand(eCmdConvertAuthorToConstortiumWhereLastName, "Convert Author To Consortium Where Last Name Contains consortium", "Convert Author To Consortium Where Last Name Contains consortium for Selected Object", "CATCWLSO"));
    commands.push_back(new CUICommand(eCmdConvertAuthorToConsortiumAll, "Convert All Authors To Consortium", "Convert All Authors To Consortium for Selected Object", "CAATCSO"));
    commands.push_back(new CUICommand(eCmdAddCitSubForUpdate, "Add Cit-sub for Update", "Add Cit-sub For Update for Selected Object", "ACSFUSO"));
    commands.push_back(new CUICommand(eCmdFixUsaAndStates, "Fix USA and States", "Fix USA and States  for Selected Object", "FUSSO"));
    commands.push_back(new CUICommand(eCmdRemoveAllStructuredComments,"Remove All Structured Comments","Remove All Structured Comments from Selected Object","RASCSO"));
    commands.push_back(new CUICommand(eCmdRemoveEmptyStructuredComments,"Remove Empty Structured Comments","Remove Empty Structured Comments from Selected Object","RESCSO"));
    commands.push_back(new CUICommand(eCmdModifyStructuredComment,"Remove Structured Comment Field","Remove Structured Comment Field from Selected Object","RSCFSO"));
    commands.push_back(new CUICommand(eCmdRemoveStructuredCommentKeyword,"Remove Structured Comment Keyword","Remove Structured Comment Keyword from Selected Object","RSCKSO"));
    commands.push_back(new CUICommand(eCmdAddStructuredCommentKeyword ,"Add Structured Comment Keyword","Add Structured Comment Keyword to Selected Object","ASCKSO"));
    commands.push_back(new CUICommand(eCmdReorderStructuredComment,"Reorder Structured Comment","Reorder Structured Comment in Selected Object","RSCSO"));
    commands.push_back(new CUICommand(eCmdConvertComment,"Convert Structured Comment","Convert Structured Comment in Selected Object","CSCSO"));
    commands.push_back(new CUICommand(eCmdParseComment,"Parse Structured Comment","Parse Structured Comment in Selected Object","PSCSO"));
    commands.push_back(new CUICommand(eCmdAssignFeatureIds,"Assign Feature IDs", "Assign Feature IDs for Selected Object","AFIDSO"));
    commands.push_back(new CUICommand(eCmdClearFeatureIds,"Clear Feature IDs", "Clear Feature IDs for Selected Object","CFIDSO"));
    commands.push_back(new CUICommand(eCmdReassignFeatureIds,"Reassign Feature IDs", "Reassign Feature IDs for Selected Object","RFIDSO"));
    commands.push_back(new CUICommand(eCmdUniqifyFeatureIds,"Uniqify Feature IDs", "Uniqify Feature IDs for Selected Object","UFIDSO"));
    commands.push_back(new CUICommand(eCmdTrimJunkInPrimerSeqs,"Trim Junk In Primer Seqs","Trim Junk In Primer Seqs for Selected Object","TJIPSSO"));
    commands.push_back(new CUICommand(eCmdFixiInPrimerSeqs,"Fix i in Primer Seqs","Fix i in Primer Seqs for Selected Object","FIIPSSO"));
    commands.push_back(new CUICommand(eCmdFixOrgModInstitution,"Fix OrgMod Institution Code","Fix OrgMod Institution Code for Selected Object","FOMICSO"));
    commands.push_back(new CUICommand(eCmdSwapPrimerSeqName,"Swap Primer Seq and Name","Swap Primer Seq and Name for Selected Object","SPSNSO"));
    commands.push_back(new CUICommand(eCmdMergePrimerSets,"Merge Primer Sets","Merge Primer Sets for Selected Object","MPSSO"));
    commands.push_back(new CUICommand(eCmdSplitPrimerSets,"Split Primer Sets by Position","Split Primer Sets by Position for Selected Object","SPSBPSO"));
    commands.push_back(new CUICommand(eCmdAppendModToOrg,"Append Mod To Org","Append Mod To Org for Selected Object","AMTOSO"));
    commands.push_back(new CUICommand(eCmdParseCollectionDateMonthFirst,"Parse Collection Date Month First","Parse Collection Date Month First for Selected Object","PCDMFSO"));
    commands.push_back(new CUICommand(eCmdParseCollectionDateDayFirst,"Parse Collection Date Day First","Parse Collection Date Day First for Selected Object","PCDDFSO"));
    commands.push_back(new CUICommand(eCmdCountryFixupCap,"Country Fixup Fix Capitalization After Colon","Country Fixup Fix Capitalization After Colon","CFFCAC"));
    commands.push_back(new CUICommand(eCmdCountryFixupNoCap,"Country Fixup Do Not Fix Capitalization After Colon","Country Fixup Do Not Fix Capitalization After Colon","CFDNFCAC"));
    commands.push_back(new CUICommand(eCmdApplyCDS, "Add CDS", "Add CDS", "AC"));
    commands.push_back(new CUICommand(eCmdApplyAddRNA, "Add RNA", "Add RNA", "AR"));
    commands.push_back(new CUICommand(eCmdApplyAddOther, "Add other Feature", "Add other Feature", "AOF"));
    commands.push_back(new CUICommand(eCmdValidate, "Validate", "Validate", "V"));
    commands.push_back(new CUICommand(eCmdRemoveAllFeatures, "Remove All Features", "Remove All Features", "RAF"));
    commands.push_back(new CUICommand(eCmdAddSequences, "Add Sequences", "Add Sequences", "AS"));
    return commands;
}

BEGIN_EVENT_TABLE(CSequenceEditingEventHandler, wxEvtHandler)
    EVT_MENU( eCmdRemoveDescriptors, CSequenceEditingEventHandler::RemoveDescriptors )
    EVT_MENU( eCmdRemoveFeatures, CSequenceEditingEventHandler::RemoveFeatures )
    EVT_MENU( eCmdRemoveSeqAnnotIds, CSequenceEditingEventHandler::RemoveSeqAnnotIds )
    EVT_MENU( eCmdRemoveBankitComment, CSequenceEditingEventHandler::RemoveBankitComment )
    EVT_MENU( eCmdRemoveDuplicateStructuredComments, CSequenceEditingEventHandler::RemoveDuplicateStructuredComments )
    EVT_MENU( eCmdRemoveSequencesFromAlignments, CSequenceEditingEventHandler::RemoveSequencesFromAlignments )
    EVT_MENU( eCmdRemoveSequences, CSequenceEditingEventHandler::RemoveSequences )
    EVT_MENU( eCmdRemoveAlignments, CSequenceEditingEventHandler::RemoveSeqAnnotAlign )
    EVT_MENU( eCmdRemoveGraphs, CSequenceEditingEventHandler::RemoveSeqAnnotGraph )
    EVT_MENU( eCmdRemoveGenomeProjectId, CSequenceEditingEventHandler::RemoveGenomeProjectsDB )
    EVT_MENU( eCmdFixCapitalizationAll, CSequenceEditingEventHandler::FixCapitalizationAll )
    EVT_MENU( eCmdFixCapitalizationAuthors, CSequenceEditingEventHandler::FixCapitalizationAuthors )
    EVT_MENU( eCmdFixCapitalizationTitles, CSequenceEditingEventHandler::FixCapitalizationTitles )
    EVT_MENU( eCmdFixCapitalizationAffiliation, CSequenceEditingEventHandler::FixCapitalizationAffiliation )
    EVT_MENU( eCmdFixCapitalizationCountry, CSequenceEditingEventHandler::FixCapitalizationCountry )
    EVT_MENU( eCmdRemoveUnpublishedPublications, CSequenceEditingEventHandler::RemoveUnpublishedPublications )
    EVT_MENU( eCmdRemovePublishedPublications, CSequenceEditingEventHandler::RemovePublishedPublications )
    EVT_MENU( eCmdRemoveCollidingPublications, CSequenceEditingEventHandler::RemoveCollidingPublications )
    EVT_MENU( eCmdRemoveAllPublications, CSequenceEditingEventHandler::RemoveAllPublications )
    EVT_MENU( eCmdRemoveAuthorConsortiums, CSequenceEditingEventHandler::RemoveAuthorConsortiums )
    EVT_MENU( eCmdReverseAuthorNames, CSequenceEditingEventHandler::ReverseAuthorNames )
    EVT_MENU( eCmdStripAuthorSuffixes, CSequenceEditingEventHandler::StripAuthorSuffixes )
    EVT_MENU( eCmdTruncateAuthorMiddleInitials, CSequenceEditingEventHandler::TruncateAuthorMiddleInitials )
    EVT_MENU( eCmdConvertAuthorToConstortiumWhereLastName, CSequenceEditingEventHandler::ConvertAuthorToConstortiumWhereLastName )
    EVT_MENU( eCmdConvertAuthorToConsortiumAll, CSequenceEditingEventHandler::ConvertAuthorToConsortiumAll )
    EVT_MENU( eCmdAddCitSubForUpdate, CSequenceEditingEventHandler::AddCitSubForUpdate )
    EVT_MENU( eCmdFixUsaAndStates, CSequenceEditingEventHandler::FixUsaAndStates )
    EVT_MENU( eCmdRemoveAllStructuredComments, CSequenceEditingEventHandler::RemoveAllStructuredComments )
    EVT_MENU( eCmdRemoveEmptyStructuredComments, CSequenceEditingEventHandler::RemoveEmptyStructuredComments )
    EVT_MENU( eCmdModifyStructuredComment, CSequenceEditingEventHandler::ModifyStructuredComment )
    EVT_MENU( eCmdRemoveStructuredCommentKeyword, CSequenceEditingEventHandler::RemoveStructuredCommentKeyword )
    EVT_MENU( eCmdAddStructuredCommentKeyword, CSequenceEditingEventHandler::AddStructuredCommentKeyword )
    EVT_MENU( eCmdReorderStructuredComment, CSequenceEditingEventHandler::ReorderStructuredComment )
    EVT_MENU( eCmdConvertComment, CSequenceEditingEventHandler::ConvertComment )
    EVT_MENU( eCmdParseComment, CSequenceEditingEventHandler::ParseComment )
    EVT_MENU( eCmdAssignFeatureIds, CSequenceEditingEventHandler::AssignFeatureIds )
    EVT_MENU( eCmdClearFeatureIds, CSequenceEditingEventHandler::ClearFeatureIds )
    EVT_MENU( eCmdReassignFeatureIds, CSequenceEditingEventHandler::ReassignFeatureIds )
    EVT_MENU( eCmdUniqifyFeatureIds, CSequenceEditingEventHandler::UniqifyFeatureIds )
    EVT_MENU( eCmdTrimJunkInPrimerSeqs, CSequenceEditingEventHandler::TrimJunkInPrimerSeqs )
    EVT_MENU( eCmdFixiInPrimerSeqs, CSequenceEditingEventHandler::FixiInPrimerSeqs )
    EVT_MENU( eCmdFixOrgModInstitution, CSequenceEditingEventHandler::FixOrgModInstitution )
    EVT_MENU( eCmdSwapPrimerSeqName, CSequenceEditingEventHandler::SwapPrimerSeqName )
    EVT_MENU( eCmdMergePrimerSets, CSequenceEditingEventHandler::MergePrimerSets )
    EVT_MENU( eCmdSplitPrimerSets, CSequenceEditingEventHandler::SplitPrimerSets )
    EVT_MENU( eCmdAppendModToOrg, CSequenceEditingEventHandler::AppendModToOrg )
    EVT_MENU( eCmdParseCollectionDateMonthFirst, CSequenceEditingEventHandler::ParseCollectionDateMonthFirst )
    EVT_MENU( eCmdParseCollectionDateDayFirst, CSequenceEditingEventHandler::ParseCollectionDateDayFirst )
    EVT_MENU( eCmdCountryFixupCap, CSequenceEditingEventHandler::CountryFixupCap )
    EVT_MENU( eCmdCountryFixupNoCap, CSequenceEditingEventHandler::CountryFixupNoCap )
    EVT_UPDATE_UI_RANGE(eCmdRemoveDescriptors, eCmdCountryFixupNoCap, CSequenceEditingEventHandler::IsObjectSelected )
    EVT_MENU( eCmdApplyCDS, CSequenceEditingEventHandler::AddCDS )
    EVT_MENU( eCmdApplyAddRNA, CSequenceEditingEventHandler::AddRNA )
    EVT_MENU( eCmdApplyAddOther, CSequenceEditingEventHandler::AddOtherFeature )
    EVT_MENU( eCmdValidate, CSequenceEditingEventHandler::Validate )
    EVT_MENU( eCmdRemoveAllFeatures, CSequenceEditingEventHandler::RemoveAllFeatures )
    EVT_MENU( eCmdAddSequences, CSequenceEditingEventHandler::AddSequences )

END_EVENT_TABLE()



static
WX_DEFINE_MENU(kEditingMenu)
    WX_SUBMENU("&Edit")
        WX_MENU_SEPARATOR_L("Contribs")
        WX_SUBMENU("Sequence Editing")
            WX_SUBMENU("Apply")
                WX_MENU_ITEM(eCmdApplyCDS)
                WX_MENU_ITEM(eCmdApplyAddRNA)
                WX_MENU_ITEM(eCmdApplyAddOther)
            WX_END_SUBMENU()
            WX_SUBMENU("Remove")
                WX_MENU_ITEM(eCmdRemoveDescriptors)
                WX_MENU_ITEM(eCmdRemoveFeatures)
                WX_MENU_ITEM(eCmdRemoveAllFeatures)
                WX_MENU_ITEM(eCmdRemoveSeqAnnotIds)
                WX_MENU_ITEM(eCmdRemoveBankitComment)
                WX_MENU_ITEM(eCmdRemoveDuplicateStructuredComments)
                WX_MENU_ITEM(eCmdRemoveSequencesFromAlignments)
                WX_MENU_ITEM(eCmdRemoveSequences)
                WX_MENU_ITEM(eCmdRemoveAlignments)
                WX_MENU_ITEM(eCmdRemoveGraphs)
                WX_MENU_ITEM(eCmdRemoveGenomeProjectId)
            WX_END_SUBMENU()
            WX_SUBMENU("Authors and Publications")
                WX_MENU_ITEM(eCmdFixCapitalizationAll)
                WX_MENU_ITEM(eCmdFixCapitalizationAuthors)
                WX_MENU_ITEM(eCmdFixCapitalizationTitles)
                WX_MENU_ITEM(eCmdFixCapitalizationAffiliation)
                WX_MENU_ITEM(eCmdFixCapitalizationCountry)
                WX_MENU_ITEM(eCmdRemoveUnpublishedPublications)
                WX_MENU_ITEM(eCmdRemovePublishedPublications)
                WX_MENU_ITEM(eCmdRemoveCollidingPublications)
                WX_MENU_ITEM(eCmdRemoveAllPublications)
                WX_MENU_ITEM(eCmdRemoveAuthorConsortiums)
                WX_MENU_ITEM(eCmdReverseAuthorNames)
                WX_MENU_ITEM(eCmdStripAuthorSuffixes)
                WX_MENU_ITEM(eCmdTruncateAuthorMiddleInitials)
                WX_MENU_ITEM(eCmdConvertAuthorToConstortiumWhereLastName)
                WX_MENU_ITEM(eCmdConvertAuthorToConsortiumAll)
                WX_MENU_ITEM(eCmdAddCitSubForUpdate)
                WX_MENU_ITEM(eCmdFixUsaAndStates)
            WX_END_SUBMENU()
            WX_SUBMENU("Structured Comments")
                WX_MENU_ITEM(eCmdRemoveDuplicateStructuredComments)
                WX_MENU_ITEM(eCmdRemoveAllStructuredComments)
                WX_MENU_ITEM(eCmdRemoveEmptyStructuredComments)
                WX_MENU_ITEM(eCmdModifyStructuredComment)
                WX_MENU_ITEM(eCmdRemoveStructuredCommentKeyword)
                WX_MENU_ITEM(eCmdAddStructuredCommentKeyword)
                WX_MENU_ITEM(eCmdReorderStructuredComment)
                WX_MENU_ITEM(eCmdConvertComment)
                WX_MENU_ITEM(eCmdParseComment)
            WX_END_SUBMENU()
            WX_SUBMENU("Link")
                WX_MENU_ITEM(eCmdAssignFeatureIds)
                WX_MENU_ITEM(eCmdClearFeatureIds)
                WX_MENU_ITEM(eCmdReassignFeatureIds)
                WX_MENU_ITEM(eCmdUniqifyFeatureIds)
            WX_END_SUBMENU()
            WX_SUBMENU("Organism")
                WX_MENU_ITEM(eCmdTrimJunkInPrimerSeqs)
                WX_MENU_ITEM(eCmdFixiInPrimerSeqs)
                WX_MENU_ITEM(eCmdSwapPrimerSeqName)
                WX_MENU_ITEM(eCmdMergePrimerSets)
                WX_MENU_ITEM(eCmdSplitPrimerSets)
                WX_MENU_ITEM(eCmdFixOrgModInstitution)
                WX_MENU_ITEM(eCmdAppendModToOrg)
                WX_MENU_ITEM(eCmdParseCollectionDateMonthFirst)
                WX_MENU_ITEM(eCmdParseCollectionDateDayFirst)
                WX_MENU_ITEM(eCmdCountryFixupCap)
                WX_MENU_ITEM(eCmdCountryFixupNoCap)
            WX_END_SUBMENU()
        WX_MENU_ITEM(eCmdValidate)
        WX_MENU_ITEM(eCmdAddSequences)
        WX_END_SUBMENU()
    WX_END_SUBMENU()
WX_END_MENU()

const SwxMenuItemRec* CSeqEditPackage::GetMenuDef() const
{
    return kEditingMenu;
}




void CSequenceEditingEventHandler::RemoveDescriptors(wxCommandEvent& evt)
{
    x_SetUpTSEandUnDoManager(m_Workbench);       // perhaps this is unnecessary here after IsObjectSelected has run
    if (!m_TopSeqEntry || ! m_CmdProccessor) return;
    CRemoveDescDlg dlg(NULL, m_TopSeqEntry);
    if (dlg.ShowModal() == wxID_OK) 
    {
        CRef<CCmdComposite> cmd = dlg.GetCommand();
        if (cmd)  m_CmdProccessor->Execute(cmd);                       
        dlg.Close();        
    }
}

void CSequenceEditingEventHandler::RemoveFeatures(wxCommandEvent& evt)
{
    x_SetUpTSEandUnDoManager(m_Workbench);       // perhaps this is unnecessary here after IsObjectSelected has run
    if (!m_TopSeqEntry || ! m_CmdProccessor) return;
    CRemoveFeaturesDlg dlg(NULL, m_TopSeqEntry);
    if (dlg.ShowModal() == wxID_OK) 
    {
        CRef<CCmdComposite> cmd = dlg.GetCommand();
        if (cmd)
        {
            m_CmdProccessor->Execute(cmd);                       
        }
        dlg.Close();        
    }
}

void CSequenceEditingEventHandler::RemoveAllFeatures(wxCommandEvent& evt)
{
    x_SetUpTSEandUnDoManager(m_Workbench);       // perhaps this is unnecessary here after IsObjectSelected has run
    if (!m_TopSeqEntry || ! m_CmdProccessor) return;
    size_t count = 0;
    CRef<CCmdComposite> cmd = GetDeleteAllFeaturesCommand (m_TopSeqEntry, count);
    if (count > 0) {
        if (wxMessageBox(ToWxString("Are you sure you want to remove " + NStr::NumericToString(count) + " features?"), wxT("Are you sure?"),
                         wxOK | wxCANCEL, NULL) == wxOK) {
            m_CmdProccessor->Execute(cmd);
        }
    } else {
        wxMessageBox(_("No features found!"), wxT("Error"),
                 wxOK | wxICON_ERROR, NULL);
    }
}

void CSequenceEditingEventHandler::RemoveSeqAnnotIds( wxCommandEvent& event )
{
    x_SetUpTSEandUnDoManager(m_Workbench);       // perhaps this is unnecessary here after IsObjectSelected has run
    if (!m_TopSeqEntry || ! m_CmdProccessor) return;
    CRemoveSeqAnnot worker;
    if (!worker.apply(m_TopSeqEntry,m_CmdProccessor,"Remove Seq-Annot Ids",CSeq_annot::C_Data::e_Ids))
        event.Skip();
}

void CSequenceEditingEventHandler::RemoveBankitComment( wxCommandEvent& event )
{
    x_SetUpTSEandUnDoManager(m_Workbench);       // perhaps this is unnecessary here after IsObjectSelected has run
    if (!m_TopSeqEntry || ! m_CmdProccessor) return;
    CRemoveBankitComment worker;
    if (!worker.apply(m_TopSeqEntry,m_CmdProccessor,"Remove Bankit Comment"))
        event.Skip();
}


void CSequenceEditingEventHandler::RemoveDuplicateStructuredComments( wxCommandEvent& event )
{
    x_SetUpTSEandUnDoManager(m_Workbench);       // perhaps this is unnecessary here after IsObjectSelected has run
    if (!m_TopSeqEntry || ! m_CmdProccessor) return;
    CRemoveDuplicateStructuredComment worker;
    if (!worker.apply(m_TopSeqEntry,m_CmdProccessor,"Remove Duplicate Structured Comment"))
        event.Skip();
}

void CSequenceEditingEventHandler::RemoveSequencesFromAlignments( wxCommandEvent& event )
{  
    x_SetUpTSEandUnDoManager(m_Workbench);       // perhaps this is unnecessary here after IsObjectSelected has run
    if (!m_TopSeqEntry || ! m_CmdProccessor) return;

    CRemoveSeqFromAlignDlg dlg(NULL, m_TopSeqEntry);
    if (dlg.ShowModal() == wxID_OK) 
    {
        CRef<CCmdComposite> cmd = dlg.GetCommand();
        if (cmd)  m_CmdProccessor->Execute(cmd);                       
        dlg.Close();        
    }
}  

void CSequenceEditingEventHandler::RemoveSequences( wxCommandEvent& event )
{
    x_SetUpTSEandUnDoManager(m_Workbench);       // perhaps this is unnecessary here after IsObjectSelected has run
    if (!m_TopSeqEntry || ! m_CmdProccessor) return;

    CRemoveSequencesCommandConverter* sequence_remover = new CRemoveSequencesCommandConverter();
    CTblEditDlg dlg(NULL, m_TopSeqEntry, sequence_remover, NULL, SYMBOL_CTBLEDITDLG_IDNAME, SYMBOL_CTBLEDITDLG_TITLE, SYMBOL_CTBLEDITDLG_POSITION, SYMBOL_CTBLEDITDLG_SIZE, SYMBOL_CTBLEDITDLG_STYLE,
                    wxT("Please select the sequences you would like to remove"));

    bool done = false;
    while (!done) {
        if (dlg.ShowModal() == wxID_OK) {
            sequence_remover->SetGrid(dlg.GetGrid());
            CRef<CCmdComposite> cmd = dlg.GetCommand();
            if (cmd) {
                m_CmdProccessor->Execute(cmd);                       
                dlg.Close();        
                done = true;
            } else {
                wxString error = ToWxString(dlg.GetErrorMessage());
                wxMessageBox(error, wxT("Error"),
                         wxOK | wxICON_ERROR, NULL);
            }
        } else {
            done = true;
        }
    }
}

void CSequenceEditingEventHandler::RemoveSeqAnnotAlign( wxCommandEvent& event )
{
    x_SetUpTSEandUnDoManager(m_Workbench);       // perhaps this is unnecessary here after IsObjectSelected has run
    if (!m_TopSeqEntry || ! m_CmdProccessor) return;

    CRemoveSeqAnnot worker;
    if (!worker.apply(m_TopSeqEntry,m_CmdProccessor,"Remove Alignments",CSeq_annot::C_Data::e_Align))
        event.Skip();
}

void CSequenceEditingEventHandler::RemoveSeqAnnotGraph( wxCommandEvent& event )
{
    x_SetUpTSEandUnDoManager(m_Workbench);       // perhaps this is unnecessary here after IsObjectSelected has run
    if (!m_TopSeqEntry || ! m_CmdProccessor) return;

    CRemoveSeqAnnot worker;
    if (!worker.apply(m_TopSeqEntry,m_CmdProccessor,"Remove Graphs",CSeq_annot::C_Data::e_Graph))
        event.Skip();
}

void CSequenceEditingEventHandler::RemoveGenomeProjectsDB( wxCommandEvent& event )
{
    x_SetUpTSEandUnDoManager(m_Workbench);       // perhaps this is unnecessary here after IsObjectSelected has run
    if (!m_TopSeqEntry || ! m_CmdProccessor) return;

    CRemoveGenomeProjectsDB worker;
    if (!worker.apply(m_TopSeqEntry,m_CmdProccessor,"Remove GenomeProjectsDB"))
        event.Skip();
}


void CSequenceEditingEventHandler::FixCapitalizationAll( wxCommandEvent& event )
{
    x_SetUpTSEandUnDoManager(m_Workbench);       // perhaps this is unnecessary here after IsObjectSelected has run
    if (!m_TopSeqEntry || ! m_CmdProccessor) return;

    CFixAllCap worker;
    if (!worker.apply(m_TopSeqEntry,m_CmdProccessor,"Fix All Capitalization"))
        event.Skip();
}

void CSequenceEditingEventHandler::FixCapitalizationAuthors( wxCommandEvent& event )
{
    x_SetUpTSEandUnDoManager(m_Workbench);       // perhaps this is unnecessary here after IsObjectSelected has run
    if (!m_TopSeqEntry || ! m_CmdProccessor) return;

    CFixAuthorCap worker;
    if (!worker.apply(m_TopSeqEntry,m_CmdProccessor,"Fix Author Capitalization"))
        event.Skip();
}

void CSequenceEditingEventHandler::FixCapitalizationTitles( wxCommandEvent& event )
{
    x_SetUpTSEandUnDoManager(m_Workbench);       // perhaps this is unnecessary here after IsObjectSelected has run
    if (!m_TopSeqEntry || ! m_CmdProccessor) return;

    CFixTitleCap worker;
    if (!worker.apply(m_TopSeqEntry,m_CmdProccessor,"Fix Title Capitalization"))
        event.Skip();
}

void CSequenceEditingEventHandler::FixCapitalizationAffiliation( wxCommandEvent& event )
{
    x_SetUpTSEandUnDoManager(m_Workbench);       // perhaps this is unnecessary here after IsObjectSelected has run
    if (!m_TopSeqEntry || ! m_CmdProccessor) return;

    CFixAffilCap worker;
    if (!worker.apply(m_TopSeqEntry,m_CmdProccessor,"Fix Affiliation Capitalization"))
        event.Skip();
}


void CSequenceEditingEventHandler::FixCapitalizationCountry( wxCommandEvent& event )
{
    x_SetUpTSEandUnDoManager(m_Workbench);       // perhaps this is unnecessary here after IsObjectSelected has run
    if (!m_TopSeqEntry || ! m_CmdProccessor) return;

    CFixCountryCap worker;
    if (!worker.apply(m_TopSeqEntry,m_CmdProccessor,"Fix Country Capitalization"))
        event.Skip();
}

void CSequenceEditingEventHandler::RemoveUnpublishedPublications( wxCommandEvent& event )
{
    x_SetUpTSEandUnDoManager(m_Workbench);       // perhaps this is unnecessary here after IsObjectSelected has run
    if (!m_TopSeqEntry || ! m_CmdProccessor) return;

    CRemoveUnpublishedPubs worker;
    if (!worker.apply(m_TopSeqEntry,m_CmdProccessor,"Delete Unpublished Publications"))
        event.Skip();
}

void CSequenceEditingEventHandler::RemovePublishedPublications( wxCommandEvent& event )
{
    x_SetUpTSEandUnDoManager(m_Workbench);       // perhaps this is unnecessary here after IsObjectSelected has run
    if (!m_TopSeqEntry || ! m_CmdProccessor) return;

    CRemovePublishedPubs worker;
    if (!worker.apply(m_TopSeqEntry,m_CmdProccessor,"Delete Published Publications"))
        event.Skip();
}

void CSequenceEditingEventHandler::RemoveCollidingPublications( wxCommandEvent& event )
{
    x_SetUpTSEandUnDoManager(m_Workbench);       // perhaps this is unnecessary here after IsObjectSelected has run
    if (!m_TopSeqEntry || ! m_CmdProccessor) return;

    CRemoveCollidingPubs worker;
    if (!worker.apply(m_TopSeqEntry,m_CmdProccessor,"Delete Colliding Publications"))
        event.Skip();
}


void CSequenceEditingEventHandler::RemoveAllPublications( wxCommandEvent& event )
{
    x_SetUpTSEandUnDoManager(m_Workbench);       // perhaps this is unnecessary here after IsObjectSelected has run
    if (!m_TopSeqEntry || ! m_CmdProccessor) return;

    CRemoveAllPubs worker;
    if (!worker.apply(m_TopSeqEntry,m_CmdProccessor,"Delete Colliding Publications"))
        event.Skip();
}

void CSequenceEditingEventHandler::RemoveAuthorConsortiums( wxCommandEvent& event )
{
    x_SetUpTSEandUnDoManager(m_Workbench);       // perhaps this is unnecessary here after IsObjectSelected has run
    if (!m_TopSeqEntry || ! m_CmdProccessor) return;

  CRemoveAuthorConsortium worker;
  if (!worker.apply(m_TopSeqEntry,m_CmdProccessor,"Remove Author Consortiums"))
    event.Skip();
}

void CSequenceEditingEventHandler::ReverseAuthorNames( wxCommandEvent& event )
{
    x_SetUpTSEandUnDoManager(m_Workbench);       // perhaps this is unnecessary here after IsObjectSelected has run
    if (!m_TopSeqEntry || ! m_CmdProccessor) return;

    CReverseAuthorNames worker;
    if (!worker.apply(m_TopSeqEntry,m_CmdProccessor,"Reverse Author Last and First Name"))
        event.Skip();
}

void CSequenceEditingEventHandler::StripAuthorSuffixes( wxCommandEvent& event )
{
    x_SetUpTSEandUnDoManager(m_Workbench);       // perhaps this is unnecessary here after IsObjectSelected has run
    if (!m_TopSeqEntry || ! m_CmdProccessor) return;

  CStripAuthorSuffix worker;
  if (!worker.apply(m_TopSeqEntry,m_CmdProccessor,"Strip Author Suffixes"))
    event.Skip();
}

void CSequenceEditingEventHandler::TruncateAuthorMiddleInitials( wxCommandEvent& event )
{
    x_SetUpTSEandUnDoManager(m_Workbench);       // perhaps this is unnecessary here after IsObjectSelected has run
    if (!m_TopSeqEntry || ! m_CmdProccessor) return;

    CFixAuthorInitials worker;
    if (!worker.apply(m_TopSeqEntry,m_CmdProccessor,"Truncate Author Middle Initials"))
        event.Skip();
}

void CSequenceEditingEventHandler::ConvertAuthorToConstortiumWhereLastName( wxCommandEvent& event )
{
    x_SetUpTSEandUnDoManager(m_Workbench);       // perhaps this is unnecessary here after IsObjectSelected has run
    if (!m_TopSeqEntry || ! m_CmdProccessor) return;

  CConvertAuthorToConsortiumWhereConstraint worker;
  worker.SetConstraint("consortium");
  if (!worker.apply(m_TopSeqEntry,m_CmdProccessor,"Convert Author to Consortiums Where Last Name Contains Consortium"))
    event.Skip();
}

void CSequenceEditingEventHandler::ConvertAuthorToConsortiumAll( wxCommandEvent& event )
{
    x_SetUpTSEandUnDoManager(m_Workbench);       // perhaps this is unnecessary here after IsObjectSelected has run
    if (!m_TopSeqEntry || ! m_CmdProccessor) return;
    
  CConvertAllAuthorToConsortium worker;
  if (!worker.apply(m_TopSeqEntry,m_CmdProccessor,"Convert All Authors To Consortiums"))
    event.Skip();
}

void CSequenceEditingEventHandler::AddCitSubForUpdate( wxCommandEvent& event )
{
    x_SetUpTSEandUnDoManager(m_Workbench);       // perhaps this is unnecessary here after IsObjectSelected has run
    if (!m_TopSeqEntry || ! m_CmdProccessor) return;

    CAddCitSubForUpdate worker;
    if (!worker.apply(m_TopSeqEntry,m_CmdProccessor,"Add Cit-sub for update"))
        event.Skip();
}

void CSequenceEditingEventHandler::FixUsaAndStates( wxCommandEvent& event )
{
    x_SetUpTSEandUnDoManager(m_Workbench);       // perhaps this is unnecessary here after IsObjectSelected has run
    if (!m_TopSeqEntry || ! m_CmdProccessor) return;

    CFixUSA worker;
    if (!worker.apply(m_TopSeqEntry,m_CmdProccessor,"Fix USA and States"))
        event.Skip();
}

void CSequenceEditingEventHandler::RemoveAllStructuredComments( wxCommandEvent& event )
{
    x_SetUpTSEandUnDoManager(m_Workbench);       // perhaps this is unnecessary here after IsObjectSelected has run
    if (!m_TopSeqEntry || ! m_CmdProccessor) return;

    CRemoveAllStructuredComment worker;
    if (!worker.apply(m_TopSeqEntry,m_CmdProccessor,"Remove All Structured Comments"))
        event.Skip();
}

void CSequenceEditingEventHandler::RemoveEmptyStructuredComments( wxCommandEvent& event )
{
    x_SetUpTSEandUnDoManager(m_Workbench);       // perhaps this is unnecessary here after IsObjectSelected has run
    if (!m_TopSeqEntry || ! m_CmdProccessor) return;

    CRemoveEmptyStructuredComment worker;
    if (!worker.apply(m_TopSeqEntry,m_CmdProccessor,"Remove Empty Structured Comments"))
        event.Skip();
}

void CSequenceEditingEventHandler::ModifyStructuredComment( wxCommandEvent& event )
{
    x_SetUpTSEandUnDoManager(m_Workbench);       // perhaps this is unnecessary here after IsObjectSelected has run
    if (!m_TopSeqEntry || ! m_CmdProccessor) return;

    CModifyStructuredComment worker;
    if (!worker.apply(m_TopSeqEntry,m_CmdProccessor,"Remove Structured Comment Field"))
        event.Skip();
}

void CSequenceEditingEventHandler::RemoveStructuredCommentKeyword( wxCommandEvent& event )
{
    x_SetUpTSEandUnDoManager(m_Workbench);       // perhaps this is unnecessary here after IsObjectSelected has run
    if (!m_TopSeqEntry || ! m_CmdProccessor) return;

    CRemoveStructuredCommentKeyword worker;
    if (!worker.apply(m_TopSeqEntry,m_CmdProccessor,"Remove Structured Comment Keyword"))
        event.Skip();
}

void CSequenceEditingEventHandler::AddStructuredCommentKeyword( wxCommandEvent& event )
{
    x_SetUpTSEandUnDoManager(m_Workbench);       // perhaps this is unnecessary here after IsObjectSelected has run
    if (!m_TopSeqEntry || ! m_CmdProccessor) return;

    CAddStructuredCommentKeyword worker;
    if (!worker.apply(m_TopSeqEntry,m_CmdProccessor,"Add Structured Comment Keyword"))
        event.Skip();
}

void CSequenceEditingEventHandler::ReorderStructuredComment( wxCommandEvent& event )
{
    x_SetUpTSEandUnDoManager(m_Workbench);       // perhaps this is unnecessary here after IsObjectSelected has run
    if (!m_TopSeqEntry || ! m_CmdProccessor) return;

    CReorderStructuredComment worker;
    if (!worker.apply(m_TopSeqEntry,m_CmdProccessor,"Reorder Structured Comment"))
        event.Skip();
}

void CSequenceEditingEventHandler::ConvertComment( wxCommandEvent& event )
{
    x_SetUpTSEandUnDoManager(m_Workbench);       // perhaps this is unnecessary here after IsObjectSelected has run
    if (!m_TopSeqEntry || ! m_CmdProccessor) return;

    CConvertComment worker;
    if (!worker.apply(m_TopSeqEntry,m_CmdProccessor,"Convert Comment to Structured Comment", "::","  "))
        event.Skip();
}

void CSequenceEditingEventHandler::ParseComment( wxCommandEvent& event )
{
    x_SetUpTSEandUnDoManager(m_Workbench);       // perhaps this is unnecessary here after IsObjectSelected has run
    if (!m_TopSeqEntry || ! m_CmdProccessor) return;

    CConvertComment worker;
    if (!worker.apply(m_TopSeqEntry,m_CmdProccessor,"Parse Comment to Structured Comment", "=",";"))
        event.Skip();
}

void CSequenceEditingEventHandler::AssignFeatureIds( wxCommandEvent& event )
{
    x_SetUpTSEandUnDoManager(m_Workbench);       // perhaps this is unnecessary here after IsObjectSelected has run
    if (!m_TopSeqEntry || ! m_CmdProccessor) return;

    CAssignFeatureId worker;
    if (!worker.apply(m_TopSeqEntry,m_CmdProccessor,"Assign Feature Ids"))
        event.Skip();
}

void CSequenceEditingEventHandler::ClearFeatureIds( wxCommandEvent& event )
{
    x_SetUpTSEandUnDoManager(m_Workbench);       // perhaps this is unnecessary here after IsObjectSelected has run
    if (!m_TopSeqEntry || ! m_CmdProccessor) return;

    CClearFeatureId worker;
    if (!worker.apply(m_TopSeqEntry,m_CmdProccessor,"Clear Feature Ids"))
        event.Skip();
}

void CSequenceEditingEventHandler::ReassignFeatureIds( wxCommandEvent& event )
{
    x_SetUpTSEandUnDoManager(m_Workbench);       // perhaps this is unnecessary here after IsObjectSelected has run
    if (!m_TopSeqEntry || ! m_CmdProccessor) return;
    
    CReassignFeatureId worker;
    if (!worker.apply(m_TopSeqEntry,m_CmdProccessor,"Reassign Feature Ids"))
        event.Skip();
}

void CSequenceEditingEventHandler::UniqifyFeatureIds( wxCommandEvent& event )
{
    x_SetUpTSEandUnDoManager(m_Workbench);       // perhaps this is unnecessary here after IsObjectSelected has run
    if (!m_TopSeqEntry || ! m_CmdProccessor) return;

    CUniqifyFeatureId worker;
    if (!worker.apply(m_TopSeqEntry,m_CmdProccessor,"Uniqify Feature Ids"))
        event.Skip();
}

void CSequenceEditingEventHandler::TrimJunkInPrimerSeqs( wxCommandEvent& event )
{
    x_SetUpTSEandUnDoManager(m_Workbench);       // perhaps this is unnecessary here after IsObjectSelected has run
    if (!m_TopSeqEntry || ! m_CmdProccessor) return;

    CPCRPrimerTrimJunk worker;
    if (!worker.apply(m_TopSeqEntry,m_CmdProccessor,"Trim Junk In Primer Seqs"))
        event.Skip();
}

void CSequenceEditingEventHandler::FixiInPrimerSeqs( wxCommandEvent& event )
{
    x_SetUpTSEandUnDoManager(m_Workbench);       // perhaps this is unnecessary here after IsObjectSelected has run
    if (!m_TopSeqEntry || ! m_CmdProccessor) return;

    CPCRPrimerFixI worker;
    if (!worker.apply(m_TopSeqEntry,m_CmdProccessor,"Fix i In Primer Seqs"))
        event.Skip();
}

void CSequenceEditingEventHandler::FixOrgModInstitution( wxCommandEvent& event )
{
    x_SetUpTSEandUnDoManager(m_Workbench);       // perhaps this is unnecessary here after IsObjectSelected has run
    if (!m_TopSeqEntry || ! m_CmdProccessor) return;

    CFixOrgModInst worker;
    if (!worker.apply(m_TopSeqEntry,m_CmdProccessor,"Fix OrgMod Institution"))
        event.Skip();
}

void CSequenceEditingEventHandler::SwapPrimerSeqName( wxCommandEvent& event )
{
    x_SetUpTSEandUnDoManager(m_Workbench);       // perhaps this is unnecessary here after IsObjectSelected has run
    if (!m_TopSeqEntry || ! m_CmdProccessor) return;

    CSwapPrimerNameSeq worker;
    if (!worker.apply(m_TopSeqEntry,m_CmdProccessor,"Swap Primer Seq and Name"))
        event.Skip();
}

void CSequenceEditingEventHandler::MergePrimerSets( wxCommandEvent& event )
{
    x_SetUpTSEandUnDoManager(m_Workbench);       // perhaps this is unnecessary here after IsObjectSelected has run
    if (!m_TopSeqEntry || ! m_CmdProccessor) return;

    CMergePrimerSets worker;
    if (!worker.apply(m_TopSeqEntry,m_CmdProccessor,"Merge Primer Sets"))
        event.Skip();
}

void CSequenceEditingEventHandler::SplitPrimerSets( wxCommandEvent& event )
{
    x_SetUpTSEandUnDoManager(m_Workbench);       // perhaps this is unnecessary here after IsObjectSelected has run
    if (!m_TopSeqEntry || ! m_CmdProccessor) return;

    CSplitPrimerSets worker;
    if (!worker.apply(m_TopSeqEntry,m_CmdProccessor,"Split Primer Sets By Position"))
        event.Skip();
}

void CSequenceEditingEventHandler::AppendModToOrg( wxCommandEvent& event )
{  
    x_SetUpTSEandUnDoManager(m_Workbench);       // perhaps this is unnecessary here after IsObjectSelected has run
    if (!m_TopSeqEntry || ! m_CmdProccessor) return;

    CAppendModToOrgDlg dlg(NULL, m_TopSeqEntry);
    if (dlg.ShowModal() == wxID_OK) 
    {
        CRef<CCmdComposite> cmd = dlg.GetCommand();
        if (cmd)  m_CmdProccessor->Execute(cmd);                       
        dlg.Close();        
    }
}  

void CSequenceEditingEventHandler::ParseCollectionDateMonthFirst( wxCommandEvent& event )
{
    x_SetUpTSEandUnDoManager(m_Workbench);       // perhaps this is unnecessary here after IsObjectSelected has run
    if (!m_TopSeqEntry || ! m_CmdProccessor) return;

    CParseCollectionDateFormats worker;
    if (!worker.apply(m_TopSeqEntry,m_CmdProccessor,"Parse Collection Date Month First",true))
        event.Skip();
}

void CSequenceEditingEventHandler::ParseCollectionDateDayFirst( wxCommandEvent& event )
{
    x_SetUpTSEandUnDoManager(m_Workbench);       // perhaps this is unnecessary here after IsObjectSelected has run
    if (!m_TopSeqEntry || ! m_CmdProccessor) return;

    CParseCollectionDateFormats worker;
    if (!worker.apply(m_TopSeqEntry,m_CmdProccessor,"Parse Collection Date Day First",false))
        event.Skip();
}

void CSequenceEditingEventHandler::CountryFixupCap( wxCommandEvent& event )
{
    x_SetUpTSEandUnDoManager(m_Workbench);       // perhaps this is unnecessary here after IsObjectSelected has run
    if (!m_TopSeqEntry || ! m_CmdProccessor) return;

    CCountryFixup worker;
    if (!worker.apply(m_TopSeqEntry,m_CmdProccessor,"Country Fixup Fix Capitalization After Colon",true))
        event.Skip();
}

void CSequenceEditingEventHandler::CountryFixupNoCap( wxCommandEvent& event )
{
    x_SetUpTSEandUnDoManager(m_Workbench);       // perhaps this is unnecessary here after IsObjectSelected has run
    if (!m_TopSeqEntry || ! m_CmdProccessor) return;

    CCountryFixup worker;
    if (!worker.apply(m_TopSeqEntry,m_CmdProccessor,"Country Fixup Do Not Fix Capitalization After Colon",false))
        event.Skip();
}

void CSequenceEditingEventHandler::AddCDS( wxCommandEvent& event )
{
    x_SetUpTSEandUnDoManager(m_Workbench);       // perhaps this is unnecessary here after IsObjectSelected has run
    if (!m_TopSeqEntry || ! m_CmdProccessor) return;

    CBulkFeatureAddDlg dlg(NULL, m_TopSeqEntry, 
                           CSourceRequirements::eWizardType_standard,
                           CSourceRequirements::eWizardSrcType_any,
                           m_Workbench);
    dlg.ChooseFeatureType(CSeqFeatData::eSubtype_cdregion);
    if (dlg.ShowModal() == wxID_OK) 
    {
        CRef<CCmdComposite> cmd = dlg.GetCommand();
        if (cmd) { 
            m_CmdProccessor->Execute(cmd);                       
        } else {
            string errmsg = dlg.GetErrorMessage();
            if (!NStr::IsBlank(errmsg)) {
                wxString error = ToWxString(errmsg);
                wxMessageBox(error, wxT("Error"),
                         wxOK | wxICON_ERROR, NULL);
            }
        }
        dlg.Close();        
    }
}


void CSequenceEditingEventHandler::AddRNA( wxCommandEvent& event )
{
    x_SetUpTSEandUnDoManager(m_Workbench);       // perhaps this is unnecessary here after IsObjectSelected has run
    if (!m_TopSeqEntry || ! m_CmdProccessor) return;

    CBulkFeatureAddDlg dlg(NULL, m_TopSeqEntry, 
                           CSourceRequirements::eWizardType_standard,
                           CSourceRequirements::eWizardSrcType_any,
                           m_Workbench);
    dlg.ChooseFeatureType(CSeqFeatData::eSubtype_rRNA);
    if (dlg.ShowModal() == wxID_OK) 
    {
        CRef<CCmdComposite> cmd = dlg.GetCommand();
        if (cmd) { 
            m_CmdProccessor->Execute(cmd);                       
        } else {
            string errmsg = dlg.GetErrorMessage();
            if (!NStr::IsBlank(errmsg)) {
                wxString error = ToWxString(errmsg);
                wxMessageBox(error, wxT("Error"),
                         wxOK | wxICON_ERROR, NULL);
            }
        }
        dlg.Close();        
    }
}


void CSequenceEditingEventHandler::AddOtherFeature( wxCommandEvent& event )
{
    x_SetUpTSEandUnDoManager(m_Workbench);       // perhaps this is unnecessary here after IsObjectSelected has run
    if (!m_TopSeqEntry || ! m_CmdProccessor) return;

    CBulkFeatureAddDlg dlg(NULL, m_TopSeqEntry, 
                           CSourceRequirements::eWizardType_standard,
                           CSourceRequirements::eWizardSrcType_any,
                           m_Workbench);
    dlg.ChooseFeatureType(CSeqFeatData::eSubtype_gene);
    if (dlg.ShowModal() == wxID_OK) 
    {
        CRef<CCmdComposite> cmd = dlg.GetCommand();
        if (cmd) { 
            m_CmdProccessor->Execute(cmd);                       
        } else {
            string errmsg = dlg.GetErrorMessage();
            if (!NStr::IsBlank(errmsg)) {
                wxString error = ToWxString(errmsg);
                wxMessageBox(error, wxT("Error"),
                         wxOK | wxICON_ERROR, NULL);
            }
        }
        dlg.Close();        
    }
}


void CSequenceEditingEventHandler::Validate( wxCommandEvent& event )
{
    x_SetUpTSEandUnDoManager(m_Workbench);       // perhaps this is unnecessary here after IsObjectSelected has run
    if (!m_TopSeqEntry || ! m_CmdProccessor) return;

    TConstScopedObjects objects;
    if (m_SeqSubmit) {
        objects.push_back(SConstScopedObject(*m_SeqSubmit, m_TopSeqEntry.GetScope()));
    } else {
        objects.push_back(SConstScopedObject(*(m_TopSeqEntry.GetCompleteObject()), m_TopSeqEntry.GetScope()));
    }
    CIRef<CProjectService> prj = m_Workbench->GetServiceByType<CProjectService>();
    prj->AddProjectView("Validate Table View", objects, 0);
}


void CSequenceEditingEventHandler::AddSequences( wxCommandEvent& event )
{
    x_SetUpTSEandUnDoManager(m_Workbench);       // perhaps this is unnecessary here after IsObjectSelected has run
    if (!m_TopSeqEntry || ! m_CmdProccessor) return;

    CRef<CFileLoadManager> fileManager(new CFileLoadManager());

    // get all other file format from an Extension Point
    vector< CIRef<IFileFormatLoaderManager> > fileManagers;
    GetExtensionAsInterface("file_format_loader_manager", fileManagers);

    for(  size_t i = 0;  i < fileManagers.size();  ++i  )   {
        IFileFormatLoaderManager* mgr = fileManagers[i].GetPointer();
        string loader_id = mgr->GetFileLoaderId();
        if ("file_loader_fasta" == loader_id
            || "file_loader_asn" == loader_id
            || "file_loader_text_align" == loader_id) {
            fileManager->AddFileFormat(mgr);
        }
    }

    vector<CIRef<IUIToolManager> > loadManagers;
    loadManagers.push_back(CIRef<IUIToolManager>(fileManager.GetPointer()));

    COpenObjectsDlg dlg(NULL);
    dlg.SetSize(710, 480);

    dlg.SetRegistryPath("Dialogs.Edit.OpenTable");
    dlg.SetServiceLocator(m_Workbench);
    dlg.SetManagers(loadManagers);

    if (dlg.ShowModal() == wxID_OK) {
        CIRef<IObjectLoader> object_loader(dlg.GetObjectLoader());
        if (!object_loader) {
            wxMessageBox(wxT("Failed to get object loader"), wxT("Error"),
                         wxOK | wxICON_ERROR);
        } else {
            IExecuteUnit* execute_unit = dynamic_cast<IExecuteUnit*>(object_loader.GetPointer());
            if (execute_unit) {
                if (!execute_unit->PreExecute())
                    return;

                CExecuteUnitAsyncCall call(*execute_unit, wxT("Reading file(s)..."));
                CAsyncCall::Execute(call);
                if (!call.GetResult()) // Canceled
                    return;

                if (!execute_unit->PostExecute())
                    return;
            }
            CRef<CCmdComposite> cmd (new CCmdComposite("Import File"));
            const IObjectLoader::TObjects& objects = object_loader->GetObjects();
            bool has_nuc = false;
            bool has_prot = false;
            bool success = true;
            CRef<objects::CSeq_entry> entry_to_add(new objects::CSeq_entry());
            entry_to_add->SetSet().SetClass(objects::CBioseq_set::eClass_genbank);
            ITERATE(IObjectLoader::TObjects, obj_it, objects) {
                const CObject& ptr = obj_it->GetObject();
                const objects::CSeq_entry* entry = dynamic_cast<const objects::CSeq_entry*>(&ptr);
                if (entry) {
                    CRef<objects::CSeq_entry> add(new objects::CSeq_entry());
                    add->Assign(*entry);
                    entry_to_add->SetSet().SetSeq_set().push_back(add);
                } else {
                    const objects::CSeq_submit* submit = dynamic_cast<const objects::CSeq_submit*>(&ptr);
                    if (submit) {
                        ITERATE (objects::CSeq_submit::TData::TEntrys, sit, submit->GetData().GetEntrys()) {
                            CRef<objects::CSeq_entry> add(new objects::CSeq_entry());
                            add->Assign(**sit);
                            entry_to_add->SetSet().SetSeq_set().push_back(add);
                        }
                    }
                }
                if (!success) {
                    break;
                }
            }
            if (objects.size() == 0) {
                success = false;
                wxMessageBox(wxT("Unable to read from file"), wxT("Error"),
                     wxOK | wxICON_ERROR, NULL);
            } else {
                if (entry_to_add->GetSet().IsSetSeq_set() && entry_to_add->GetSet().GetSeq_set().size() > 0) {
                    CRef<objects::CSeq_table> new_ids = GetIdsFromSeqEntry(*entry_to_add);
                    CRef<objects::CSeq_table> old_ids = GetIdsFromSeqEntry(*(m_TopSeqEntry.GetCompleteSeq_entry()));
                    CRef<objects::CSeqTable_column> problems = GetSeqIdProblems(new_ids, old_ids, 100);
                    if (problems) {
                        if (wxMessageBox(ToWxString(SummarizeIdProblems(problems) + " Do you want to fix the new sequence IDs?"), wxT("Error"),
                                         wxOK | wxCANCEL | wxICON_ERROR, NULL) == wxCANCEL) {
                            success = false;
                        } else {
                            CSeqIdFixDlg dlg(NULL, entry_to_add, m_TopSeqEntry.GetCompleteSeq_entry());
                            bool done = false;
                            while (!done) {
                                if (dlg.ShowModal() == wxID_OK) {
                                    CRef<objects::CSeq_table> repl_ids = dlg.GetReplacementTable();
                                    if (repl_ids) {
                                        ApplyReplacementIds (*entry_to_add, repl_ids);
                                        dlg.Close();        
                                        done = true;
                                    }
                                } else {
                                    done = true;
                                    success = false;
                                }
                            }
                        }
                    }
                    if (success) {
                        success = AddSeqEntryToSeqEntry(entry_to_add, m_TopSeqEntry, cmd, has_nuc, has_prot);
                    }
                }
            }
            if (success) {
                m_CmdProccessor->Execute(cmd);
            }
        }
    }
}

END_NCBI_SCOPE
