/*  $Id: fix_orgmod_institution.cpp 27558 2013-03-05 16:46:27Z filippov $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data,  the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties,  express or implied,  including
 *  warranties of performance,  merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Igor Filippov
 */

#include <ncbi_pch.hpp>
#include <gui/objutils/cmd_del_desc.hpp>
#include <gui/objutils/cmd_create_desc.hpp>
#include <gui/objutils/descriptor_change.hpp>
#include <gui/objutils/cmd_del_seq_feat.hpp>
#include <objmgr/seqdesc_ci.hpp>
#include <objmgr/seq_entry_ci.hpp>
#include <gui/packages/pkg_sequence_edit/fix_orgmod_institution.hpp>

BEGIN_NCBI_SCOPE
using namespace objects;


void CFixOrgModInst::ApplyToBioSource (const CBioSource& biosource, CCmdComposite* composite, const CSeq_entry* se, const CSeqdesc* seqDesq, const CSeq_feat* seqFeat) 
{
    CRef<CSerialObject> edited_object;
    edited_object.Reset((CSerialObject*)CBioSource::GetTypeInfo()->Create());
    edited_object->Assign(biosource);
    CBioSource& edited_biosource = dynamic_cast<CBioSource&>(*edited_object);
    bool modified = false;
    ApplyToOrgMod (edited_biosource,modified);
    if (modified)
    {
        CChangeBioSourceCommand* cmd = new CChangeBioSourceCommand();
        CObject* actual = (CObject*) &biosource;
        cmd->Add(actual, CConstRef<CObject>(edited_object));
        composite->AddCommand(*cmd);
    }
}

void CFixOrgModInst::ApplyToCSeq_entry (const CSeq_entry& se, CCmdComposite* composite)  
{

    FOR_EACH_SEQDESC_ON_SEQENTRY (it, se) {
        if ((*it)->IsSource()) {
            ApplyToBioSource ((*it)->GetSource(), composite, &se, *it, NULL);
        }
    }

    if (se.IsSet()) {
        FOR_EACH_SEQENTRY_ON_SEQSET (it, se.GetSet()) {
            ApplyToCSeq_entry (**it, composite);
        }
    }
}


void  CFixOrgModInst::apply_to_seq_and_feat(objects::CSeq_entry_Handle tse, CCmdComposite* composite)
{
    ApplyToCSeq_entry (*(tse.GetCompleteSeq_entry()), composite);
    CFeat_CI feat (tse, SAnnotSelector(CSeqFeatData::e_Biosrc));
    while (feat) {
        ApplyToBioSource (feat->GetData().GetBiosrc(), composite, NULL, NULL,&(feat->GetOriginalFeature())); 
        ++feat;
    }
}

bool CFixOrgModInst::apply(objects::CSeq_entry_Handle tse, ICommandProccessor* cmdProcessor, string title)
{
    if (tse) {
        CRef<CCmdComposite> composite(new CCmdComposite(title));    
        apply_to_seq_and_feat(tse,composite);
        cmdProcessor->Execute(composite.GetPointer());
        return true;
    }
    else
        return false;
}


void CFixOrgModInst::ApplyToOrgMod(CBioSource& biosource, bool& modified) // biosource->org->orgname->mod->(subtype,subname)
{
    if (biosource.IsSetOrg() && biosource.GetOrg().IsSetOrgname() && biosource.GetOrg().GetOrgname().IsSetMod())
        EDIT_EACH_ORGMOD_ON_BIOSOURCE(orgmod,biosource)
            if ((*orgmod)->IsSetSubtype() && (*orgmod)->IsSetSubname() &&
                ((*orgmod)->GetSubtype() == COrgMod::eSubtype_specimen_voucher || (*orgmod)->GetSubtype() == COrgMod::eSubtype_culture_collection || (*orgmod)->GetSubtype() == COrgMod::eSubtype_bio_material))
            {
                string inst_code, coll_code, id, inst_coll;      // See CValidError_imp::ValidateOrgModVoucher
                if (! COrgMod::ParseStructuredVoucher ((*orgmod)->GetSubname(), inst_code, coll_code, id)) continue;
                if (NStr::IsBlank (coll_code))   inst_coll = inst_code;
                else inst_coll = inst_code + ":" + coll_code;
                string voucher_type;
                bool is_miscapitalized;
                bool needs_country;
                bool erroneous_country;
                string correct_cap;
                if (COrgMod::IsInstitutionCodeValid(inst_coll, voucher_type, is_miscapitalized, correct_cap, needs_country, erroneous_country)) continue; // Returns true if invalid, actually
                if (NStr::StartsWith(inst_coll, "personal", NStr::eNocase)) continue;
                if (COrgMod::IsInstitutionCodeValid(inst_code, voucher_type, is_miscapitalized, correct_cap, needs_country, erroneous_country) && NStr::Equal (coll_code, "DNA")) continue;
                string abbr = COrgMod::GetInstitutionShortName(inst_code);
                if (!NStr::IsBlank(abbr))
                {
                    if (NStr::IsBlank (id))
                        (*orgmod)->SetSubname(abbr);
                    else
                        (*orgmod)->SetSubname(abbr+ ":" + id);
                    modified = true;
                }
            }
}

END_NCBI_SCOPE
