/*  $Id: modify_feature_id.cpp 27406 2013-02-08 18:35:12Z filippov $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data,  the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties,  express or implied,  including
 *  warranties of performance,  merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Igor Filippov
 */


#include <ncbi_pch.hpp>
#include <objmgr/scope.hpp>
#include <objects/seqfeat/Seq_feat.hpp>
#include <objects/seqfeat/Feat_id.hpp>
#include <gui/objutils/cmd_change_seq_feat.hpp>
#include <gui/packages/pkg_sequence_edit/modify_feature_id.hpp>

BEGIN_NCBI_SCOPE
using namespace objects;


static int s_FindHighestFeatureId(const CSeq_entry_EditHandle& entry)
{
    int feat_id = 0;
    for (CFeat_CI feat_it(entry); feat_it; ++feat_it)
    {
        CSeq_feat_EditHandle efeat(*feat_it);
        if (efeat.IsSetId())
        {
            const CFeat_id &id = efeat.GetId();
            if (id.IsLocal() && id.GetLocal().IsId() && id.GetLocal().GetId() > feat_id) feat_id = id.GetLocal().GetId();
        }
    }
    return feat_id;
}


bool CAssignFeatureId::AssignHighestFeatureId(const CSeq_entry_EditHandle& entry, int &feat_id, CCmdComposite* composite)
{
    bool modified = false;
    for (CFeat_CI feat_it(entry); feat_it; ++feat_it)
    {
        CSeq_feat_Handle feat = feat_it->GetSeq_feat_Handle();
        if (feat.IsSetId() && feat.GetId().IsLocal()) continue;
        CRef<CSerialObject> edited_object;
        edited_object.Reset((CSerialObject*)CSeq_feat::GetTypeInfo()->Create());
        edited_object->Assign(feat_it->GetOriginalFeature());
        CSeq_feat& edited = dynamic_cast<CSeq_feat&>(*edited_object);
        feat_id++;
        edited.SetId().SetLocal().SetId(feat_id);
        CIRef<IEditCommand> cmd(new CCmdChangeSeq_feat(feat,edited));
        composite->AddCommand(*cmd);
        modified = true;
    }
    return modified;
}

bool CAssignFeatureId::apply(objects::CSeq_entry_Handle tse, ICommandProccessor* cmdProcessor, string title)
{
    if (tse) 
    {
        int top_id = s_FindHighestFeatureId(tse.GetEditHandle());
        CRef<CCmdComposite> composite(new CCmdComposite(title)); 
        bool modified = AssignHighestFeatureId(tse.GetEditHandle(), top_id, composite);
       if (modified)
       {
           cmdProcessor->Execute(composite.GetPointer());
           return true;
       }
    }
    return false;
}

void CClearFeatureId::ClearFeatureIds(const CSeq_entry_EditHandle& entry, CCmdComposite* composite)
{
    for ( CFeat_CI feat_it(entry); feat_it; ++feat_it ) 
    {
        bool modified = false;
        CSeq_feat_Handle feat = feat_it->GetSeq_feat_Handle();
        CRef<CSerialObject> edited_object;
        edited_object.Reset((CSerialObject*)CSeq_feat::GetTypeInfo()->Create());
        edited_object->Assign(feat_it->GetOriginalFeature());
        CSeq_feat& edited = dynamic_cast<CSeq_feat&>(*edited_object);
        if (edited.IsSetId())    
        {    
            edited.ResetId();  
            modified = true;
        }
       if (edited.CanGetXref())
        {
            CSeq_feat::TXref::iterator xref_it = edited.SetXref().begin();
            while ( xref_it != edited.SetXref().end() )
            {
                if ((*xref_it)-> IsSetId())
                {
                    (*xref_it)->ResetId();
                    modified = true;
                }
                if (!(*xref_it)-> IsSetData()) 
                {
                    xref_it = edited.SetXref().erase(xref_it);
                    modified = true;
                }
                else
                    ++xref_it;
            }
            if (edited.SetXref().empty())
            {
                edited.ResetXref();
                modified = true;
            }
        }
       if (modified)
       {
           CIRef<IEditCommand> cmd(new CCmdChangeSeq_feat(feat,edited));
           composite->AddCommand(*cmd);
       }
    }
}


bool CClearFeatureId::apply(objects::CSeq_entry_Handle tse, ICommandProccessor* cmdProcessor, string title)
{
    if (tse) 
    {
        CRef<CCmdComposite> composite(new CCmdComposite(title)); 
        ClearFeatureIds(tse.GetEditHandle(),composite);
        cmdProcessor->Execute(composite.GetPointer());
        return true;
    }
    return false;
}


void CReassignFeatureId::MakeIDPairs(const CSeq_entry_EditHandle& entry, map<int,int> &id_pairs)
{
    int feat_id = 0;
    for (CFeat_CI feat_it(entry); feat_it; ++feat_it)
    {
        CSeq_feat_EditHandle efeat(*feat_it);
        if (efeat.IsSetId())
        {
            const CFeat_id &id = efeat.GetId();
            if (id.IsLocal() && id.GetLocal().IsId()) id_pairs[id.GetLocal().GetId()] = ++feat_id;
        }
    }
}

void CReassignFeatureId::ReassignFeatureIds(const CSeq_entry_EditHandle& entry, CCmdComposite* composite, map<int,int> &id_pairs)
{
    for ( CFeat_CI feat_it(entry); feat_it; ++feat_it ) 
    {
        bool modified = false;
        CSeq_feat_Handle feat = feat_it->GetSeq_feat_Handle();
        CRef<CSerialObject> edited_object;
        edited_object.Reset((CSerialObject*)CSeq_feat::GetTypeInfo()->Create());
        edited_object->Assign(feat_it->GetOriginalFeature());
        CSeq_feat& edited = dynamic_cast<CSeq_feat&>(*edited_object);
        if (edited.IsSetId() && edited.GetId().IsLocal() && edited.GetId().GetLocal().IsId())        
        {
            int id = id_pairs[edited.GetId().GetLocal().GetId()];
            edited.SetId().SetLocal().SetId(id);  
            modified = true;
        }
       if (edited.IsSetXref())
        {
            CSeq_feat::TXref::iterator xref_it = edited.SetXref().begin();
            while ( xref_it != edited.SetXref().end() )
            {
                if ((*xref_it)-> IsSetId() && (*xref_it)->GetId().IsLocal() && (*xref_it)->GetId().GetLocal().IsId())
                {
                    int id = id_pairs[(*xref_it)->GetId().GetLocal().GetId()];
                    (*xref_it)->SetId().SetLocal().SetId(id);
                    modified = true;
                }
                ++xref_it;
            }
        }
       if (modified)
       {
           CIRef<IEditCommand> cmd(new CCmdChangeSeq_feat(feat,edited));
           composite->AddCommand(*cmd);
       }
    }
}

bool CReassignFeatureId::apply(objects::CSeq_entry_Handle tse, ICommandProccessor* cmdProcessor, string title)
{
    if (tse) 
    {
        map<int,int> id_pairs;
        MakeIDPairs(tse.GetEditHandle(),id_pairs);
        CRef<CCmdComposite> composite(new CCmdComposite(title)); 
        ReassignFeatureIds(tse.GetEditHandle(),composite, id_pairs);
        cmdProcessor->Execute(composite.GetPointer());
        return true;
    }
    return false;
}


void CUniqifyFeatureId::UpdateFeatureIds(const CSeq_entry_EditHandle& entry, CCmdComposite* composite, int offset)
{
    for ( CFeat_CI feat_it(entry); feat_it; ++feat_it ) 
    {
        bool modified = false;
        CSeq_feat_Handle feat = feat_it->GetSeq_feat_Handle();
        CRef<CSerialObject> edited_object;
        edited_object.Reset((CSerialObject*)CSeq_feat::GetTypeInfo()->Create());
        edited_object->Assign(feat_it->GetOriginalFeature());
        CSeq_feat& edited = dynamic_cast<CSeq_feat&>(*edited_object);
        if (edited.IsSetId() && edited.GetId().IsLocal() && edited.GetId().GetLocal().IsId())        
        {
            int id = edited.GetId().GetLocal().GetId();
            edited.SetId().SetLocal().SetId(id+offset);  
            modified = true;
        }
        if (edited.IsSetXref())
        {
            CSeq_feat::TXref::iterator xref_it = edited.SetXref().begin();
            while ( xref_it != edited.SetXref().end() )
            {
                if ((*xref_it)-> IsSetId() && (*xref_it)->GetId().IsLocal() && (*xref_it)->GetId().GetLocal().IsId())
                {
                    int id = (*xref_it)->GetId().GetLocal().GetId();
                    (*xref_it)->SetId().SetLocal().SetId(id+offset);
                    modified = true;
                }
                ++xref_it;
            }
        }
        if (modified)
        {
            CIRef<IEditCommand> cmd(new CCmdChangeSeq_feat(feat,edited));
           composite->AddCommand(*cmd);
        }
    }
}


void CUniqifyFeatureId::ApplyToSeqInSet(objects::CSeq_entry_Handle tse, const CBioseq_set_Handle& bioset, CCmdComposite* composite)
{
    int offset = 0;
    FOR_EACH_SEQENTRY_ON_SEQSET (it, *bioset.GetCompleteBioseq_set()) 
    {
        CSeq_entry_EditHandle entry = tse.GetScope().GetSeq_entryEditHandle(**it);
        int top_id = s_FindHighestFeatureId(entry);
        UpdateFeatureIds(entry,composite,offset);
        offset += top_id;
    }
}

bool CUniqifyFeatureId::apply(objects::CSeq_entry_Handle tse, ICommandProccessor* cmdProcessor, string title)
{
    if (tse && tse.IsSet() && tse.GetSet().IsSetClass() && tse.GetSet().GetClass() == CBioseq_set::eClass_genbank) 
    {
        
        CRef<CCmdComposite> composite(new CCmdComposite(title)); 
        ApplyToSeqInSet(tse,tse.GetSet(),composite);
        cmdProcessor->Execute(composite.GetPointer());
        return true;
    }
    return false;
}


END_NCBI_SCOPE
