/*  $Id: pcr-primers-edit.cpp 27536 2013-02-28 20:40:27Z filippov $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data,  the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties,  express or implied,  including
 *  warranties of performance,  merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Igor Filippov
 */

#include <ncbi_pch.hpp>
#include <gui/objutils/cmd_del_desc.hpp>
#include <gui/objutils/cmd_create_desc.hpp>
#include <gui/objutils/descriptor_change.hpp>
#include <gui/objutils/cmd_del_seq_feat.hpp>
#include <objmgr/seqdesc_ci.hpp>
#include <objmgr/seq_entry_ci.hpp>
#include <gui/packages/pkg_sequence_edit/pcr-primers-edit.hpp>

BEGIN_NCBI_SCOPE
using namespace objects;


void CPCRPrimerEdit::ApplyToBioSource (const CBioSource& biosource, CCmdComposite* composite, const CSeq_entry* se, const CSeqdesc* seqDesq, const CSeq_feat* seqFeat) 
{
    CRef<CSerialObject> edited_object;
    edited_object.Reset((CSerialObject*)CBioSource::GetTypeInfo()->Create());
    edited_object->Assign(biosource);
    CBioSource& edited_biosource = dynamic_cast<CBioSource&>(*edited_object);
    bool modified = false;
    ApplyToPCRPrimer (edited_biosource,modified);
    if (modified)
    {
        CChangeBioSourceCommand* cmd = new CChangeBioSourceCommand();
        CObject* actual = (CObject*) &biosource;
        cmd->Add(actual, CConstRef<CObject>(edited_object));
        composite->AddCommand(*cmd);
    }
}

void CPCRPrimerEdit::ApplyToCSeq_entry (const CSeq_entry& se, CCmdComposite* composite)  
{

    FOR_EACH_SEQDESC_ON_SEQENTRY (it, se) {
        if ((*it)->IsSource()) {
            ApplyToBioSource ((*it)->GetSource(), composite, &se, *it, NULL);
        }
    }

    if (se.IsSet()) {
        FOR_EACH_SEQENTRY_ON_SEQSET (it, se.GetSet()) {
            ApplyToCSeq_entry (**it, composite);
        }
    }
}


void  CPCRPrimerEdit::apply_to_seq_and_feat(objects::CSeq_entry_Handle tse, CCmdComposite* composite)
{
    ApplyToCSeq_entry (*(tse.GetCompleteSeq_entry()), composite);
    CFeat_CI feat (tse, SAnnotSelector(CSeqFeatData::e_Biosrc));
    while (feat) {
        ApplyToBioSource (feat->GetData().GetBiosrc(), composite, NULL, NULL,&(feat->GetOriginalFeature())); 
        ++feat;
    }
}

bool CPCRPrimerEdit::apply(objects::CSeq_entry_Handle tse, ICommandProccessor* cmdProcessor, string title)
{
    if (tse) {
        CRef<CCmdComposite> composite(new CCmdComposite(title));    
        apply_to_seq_and_feat(tse,composite);
        cmdProcessor->Execute(composite.GetPointer());
        return true;
    }
    else
        return false;
}


void CPCRPrimerEdit::ApplyToPCRPrimer(CBioSource& biosource, bool& modified)
{
    if (biosource.IsSetPcr_primers())
    {
        EDIT_EACH_PCRREACTION_IN_PCRREACTIONSET(reaction,biosource.SetPcr_primers())
        {
            EDIT_EACH_PCRPRIMER_IN_PCRPRIMERSET(primer,(*reaction)->SetForward())
            {
                if ((*primer)->IsSetSeq())
                    EditPrimerSet((*primer)->SetSeq().Set(),modified);
            }
            EDIT_EACH_PCRPRIMER_IN_PCRPRIMERSET(primer,(*reaction)->SetReverse())
            {
                if ((*primer)->IsSetSeq())
                    EditPrimerSet((*primer)->SetSeq().Set(),modified);
            }
        }
    }

}

void CPCRPrimerTrimJunk::EditPrimerSet(string &seq, bool& modified)
{
    const char* starts[] = {"5'-","5`-","5-","5'","5`","-",NULL};
    const char* ends[] = {"-3'","-3`","-3","3'","3`","-",NULL};

    string orig(seq);

    for(int i=0; starts[i] != NULL; i++)
    {
        int len = strlen(starts[i]);
        int pos = seq.find(starts[i]);
        if (pos == 0 && len < seq.length())
            seq = seq.substr(len);
    }
    for(int i=0; ends[i] != NULL; i++)
    {
        int len = strlen(ends[i]);
        int pos = seq.rfind(ends[i]);
        if (len < seq.length() && pos == seq.length()-len)
            seq = seq.substr(0,seq.length()-len);
    }
    if (orig != seq) modified = true;
}

void CPCRPrimerFixI::EditPrimerSet(string &seq, bool& modified)
{
    string orig(seq);
    int pos = 0;
    while (pos != string::npos && pos<seq.length())
    {
        pos = seq.find("I",pos);
        if (pos != string::npos)
        {
            seq.replace(pos,1,"i");
            pos++;
        }
    }
    pos = 0;
    while (pos != string::npos && pos<seq.length())
    {
        pos = seq.find("i",pos);
        if (pos != string::npos)
        {
            string repl;
            if(pos==0 || seq[pos-1]!='<')
                repl = "<";
            repl += "i";
            if (pos==seq.length()-1 || seq[pos+1]!='>')
                repl += ">";
            seq.replace(pos,1,repl);
            pos += repl.length();
        }
    }
    
    if (orig != seq) modified = true;
}


void CSwapPrimerNameSeq::ApplyToPCRPrimer(CBioSource& biosource, bool& modified)
{
    if (biosource.IsSetPcr_primers())
    {
        EDIT_EACH_PCRREACTION_IN_PCRREACTIONSET(reaction,biosource.SetPcr_primers())
        {
            EDIT_EACH_PCRPRIMER_IN_PCRPRIMERSET(primer,(*reaction)->SetForward())
            {
                if ((*primer)->IsSetSeq() && (*primer)->IsSetName())
                {
                    string seq = (*primer)->SetSeq().Set();
                    string name = (*primer)->SetName().Set();
                    (*primer)->SetSeq().Set() = name;
                    (*primer)->SetName().Set() = seq;
                    modified = true;
                }
            }
            EDIT_EACH_PCRPRIMER_IN_PCRPRIMERSET(primer,(*reaction)->SetReverse())
            {
                if ((*primer)->IsSetSeq() && (*primer)->IsSetName())
                {
                    string seq = (*primer)->SetSeq().Set();
                    string name = (*primer)->SetName().Set();
                    (*primer)->SetSeq().Set() = name;
                    (*primer)->SetName().Set() = seq;
                    modified = true;
                }
            }
        }
    }

}

void CMergePrimerSets::ApplyToPCRPrimer(CBioSource& biosource, bool& modified)
{
    if (biosource.IsSetPcr_primers())
    {
        int num_reaction = 0;
        vector< CRef<CPCRPrimer> > forward, reverse;
        EDIT_EACH_PCRREACTION_IN_PCRREACTIONSET(reaction,biosource.SetPcr_primers())
        {
            if (num_reaction > 0)
            {
                EDIT_EACH_PCRPRIMER_IN_PCRPRIMERSET(primer,(*reaction)->SetForward())
                {
                    forward.push_back(*primer);
                }
                EDIT_EACH_PCRPRIMER_IN_PCRPRIMERSET(primer,(*reaction)->SetReverse())
                {
                    reverse.push_back(*primer);
                }
                ERASE_PCRREACTION_IN_PCRREACTIONSET(reaction,biosource.SetPcr_primers());
                modified = true;
            }
            
            num_reaction++;
        }

        for (vector< CRef<CPCRPrimer> >::iterator p = forward.begin(); p != forward.end(); ++p)
            biosource.SetPcr_primers().Set().front()->SetForward().Set().push_back(*p);
        for (vector< CRef< CPCRPrimer> >::iterator p = reverse.begin(); p != reverse.end(); ++p)
            biosource.SetPcr_primers().Set().front()->SetReverse().Set().push_back(*p);
    }

}

void CSplitPrimerSets::ApplyToPCRPrimer(CBioSource& biosource, bool& modified)
{
    if (biosource.IsSetPcr_primers())
    {
        vector< CRef<CPCRPrimer> > forward, reverse;
        EDIT_EACH_PCRREACTION_IN_PCRREACTIONSET(reaction,biosource.SetPcr_primers())
        {
            int num_primer = 0;   
            EDIT_EACH_PCRPRIMER_IN_PCRPRIMERSET(primer,(*reaction)->SetForward())
            {
                if (num_primer > 0)
                {
                    forward.push_back(*primer);
                    ERASE_PCRPRIMER_IN_PCRPRIMERSET(primer, (*reaction)->SetForward());
                    modified = true;
                }
                num_primer++;
            }
            num_primer = 0;
            EDIT_EACH_PCRPRIMER_IN_PCRPRIMERSET(primer,(*reaction)->SetReverse())
            {
                if (num_primer > 0)
                {
                    reverse.push_back(*primer);
                    ERASE_PCRPRIMER_IN_PCRPRIMERSET(primer, (*reaction)->SetReverse());
                    modified = true;
                }
                num_primer++;
            }
        }
        if (forward.size() == reverse.size())
            for (int i = 0; i < forward.size(); i++)
            {
                CRef< CPCRReaction > reaction(new CPCRReaction);
                reaction->SetForward().Set().push_back(forward[i]);
                reaction->SetReverse().Set().push_back(reverse[i]);
                biosource.SetPcr_primers().Set().push_back(reaction);
            }
    }

}


END_NCBI_SCOPE
