/*  $Id: pubdesc_editor.cpp 27132 2012-12-26 20:23:14Z filippov $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data,  the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties,  express or implied,  including
 *  warranties of performance,  merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Igor Filippov
 */

#include <ncbi_pch.hpp>
#include <gui/objutils/cmd_del_desc.hpp>
#include <gui/objutils/cmd_create_desc.hpp>
#include <gui/objutils/descriptor_change.hpp>
#include <gui/objutils/cmd_del_seq_feat.hpp>
#include <objmgr/seqdesc_ci.hpp>
#include <objmgr/seq_entry_ci.hpp>
#include <objtools/eutils/api/efetch.hpp>
#include <objects/pubmed/Pubmed_entry.hpp>
#include <objects/medline/Medline_entry.hpp>
#include <objects/biblio/Cit_art.hpp>
#include <gui/packages/pkg_sequence_edit/pubdesc_editor.hpp>

BEGIN_NCBI_SCOPE
using namespace objects;


void CPubDescEditor::ApplyToPub (const CPubdesc& pubdesc, CCmdComposite* composite, const CSeq_entry* se, const CSeqdesc* seqDesq, const CSeq_feat* seqFeat) // CEditObjectPubdesc::GetEditCommand from edit_object_seq_desc
{
    CRef<CSerialObject> edited_object;
    edited_object.Reset((CSerialObject*)CPubdesc::GetTypeInfo()->Create());
    edited_object->Assign(pubdesc);
    CPubdesc& edited_pubdesc = dynamic_cast<CPubdesc&>(*edited_object);
    bool modified = false;
    ApplyToPubdesc (edited_pubdesc,modified);
    if (modified)
    {
        CChangePubdescCommand* cmd = new CChangePubdescCommand();
        CObject* actual = (CObject*) &pubdesc;
        cmd->Add(actual, CConstRef<CObject>(edited_object));
        composite->AddCommand(*cmd);
    }
}

void CPubDescEditor::ApplyToCSeq_entry (const CSeq_entry& se, CCmdComposite* composite)  // s_CollectPubDescriptorLabels in validatorp.cpp
{

    FOR_EACH_SEQDESC_ON_SEQENTRY (it, se) {
        if ((*it)->IsPub()) {
            ApplyToPub ((*it)->GetPub(), composite, &se, *it, NULL);
        }
    }

    if (se.IsSet()) {
        FOR_EACH_SEQENTRY_ON_SEQSET (it, se.GetSet()) {
            ApplyToCSeq_entry (**it, composite);
        }
    }
}


void CPubDescEditor::apply_to_seq_and_feat(objects::CSeq_entry_Handle tse, CCmdComposite* composite)
{
    ApplyToCSeq_entry (*(tse.GetCompleteSeq_entry()), composite);
    CFeat_CI feat (tse, SAnnotSelector(CSeqFeatData::e_Pub));
    while (feat) {
        ApplyToPub (feat->GetData().GetPub(), composite, NULL, NULL,&(feat->GetOriginalFeature())); 
        ++feat;
    }
}

bool CPubDescEditor::apply(objects::CSeq_entry_Handle tse, ICommandProccessor* cmdProcessor, string title)
{
    if (tse) {
        m_tse = tse;
        CRef<CCmdComposite> composite(new CCmdComposite(title));    // See CEditObjectPubdesc::GetEditCommand from edit_object_seq_desc.cpp
        apply_to_seq_and_feat(tse,composite);
        cmdProcessor->Execute(composite.GetPointer());
        return true;
    }
    else
        return false;
}


void CAuthorEditor::ApplyToPubdesc(CPubdesc& pubdesc, bool& modified)
{
    if (pubdesc.IsSetPub())
    {
        CPub_equiv& pe = pubdesc.SetPub();
        ApplyToPubequiv(pe,modified);    
    }

}

void CAuthorEditor::ApplyToPubequiv(CPub_equiv& pe, bool& modified)
{
        EDIT_EACH_PUB_ON_PUBEQUIV (pub_iter, pe) {
            CPub& pub = **pub_iter;
            CAuth_list* authors = 0;
            switch ( pub.Which() ) {
            case CPub::e_Gen:
                if ( pub.GetGen().IsSetAuthors() ) {
                    authors = &(pub.SetGen().SetAuthors());
                }
                break;
            case CPub::e_Sub:
                authors = &(pub.SetSub().SetAuthors());
                break;
            case CPub::e_Article:
                if ( pub.GetArticle().IsSetAuthors() ) {
                    authors = &(pub.SetArticle().SetAuthors());
                }
                break;
            case CPub::e_Book:
                authors = &(pub.SetBook().SetAuthors());
                break;
            case CPub::e_Proc:
                authors = &(pub.SetProc().SetBook().SetAuthors());
                break;
            case CPub::e_Man:
                authors = &(pub.SetMan().SetCit().SetAuthors());
                break;
            case CPub::e_Patent:
                authors = &(pub.SetPatent().SetAuthors());
                break;
            case CPub::e_Equiv:
                ApplyToPubequiv(pub.SetEquiv(),modified);
                break;
            default:
                break;
            }
            
            if ( !authors )  continue;
            
            ApplyToCAuth(*authors,modified);
        }
    
}

string CAuthorEditor::get_first_initial(string first)
{
    char letter;
    string inits;
    string::iterator p = first.begin();
    while (p != first.end()) {
        // skip leading punct
        while (p != first.end() && ! isalpha (*p)) p++;

        if (p != first.end() && isalpha (*p)) {
            letter = *p;
            inits += letter;
            p++;
        }
        // skip rest of name
        while (p != first.end() && isalpha (*p)) p++;

        if (p != first.end() && *p == '-') {
            inits += "-";
            p++;
        }
    }
    return NStr::ToUpper(inits);
}

void CAuthorEditor::fix_initials(CPerson_id& pid, bool &modified) // TransferDataFromWindow in singleauthor_panel.cpp
{

    if (pid.SetName().IsSetInitials() && !NStr::IsBlank(pid.GetName().GetInitials())) {
        string first_init;
        if (pid.SetName().IsSetFirst()) {
            string first = pid.GetName().GetFirst();
            first_init = get_first_initial (first);
            first_init += ".";
        }
        string original_init = pid.GetName().GetInitials();
        string middle_init = NStr::ToUpper(original_init);
        if (! NStr::IsBlank(first_init) && NStr::StartsWith(middle_init, first_init, NStr::eNocase)) 
            middle_init = middle_init.substr(first_init.length());
        if (NStr::EndsWith(middle_init, ".")) 
            middle_init = middle_init.substr(0, middle_init.length() - 1);
        
        string new_middle_init = get_first_initial(middle_init);
        string init = first_init;
        if (!NStr::IsBlank(new_middle_init)) init += new_middle_init + ".";
        if (!NStr::IsBlank(init) && init != original_init) {
            pid.SetName().SetInitials(init);
            modified = true;
        }
    }
}


void CReverseAuthorNames::update_initials(CPerson_id& pid, string first, string last) // TransferDataFromWindow in singleauthor_panel.cpp
{

    if (pid.SetName().IsSetInitials()) {
        string first_init = get_first_initial (first);
        string middle_init = pid.GetName().GetInitials();
        first_init += ".";
        if (NStr::StartsWith(middle_init, first_init, NStr::eNocase)) 
            middle_init = middle_init.substr(first_init.length());
        if (NStr::EndsWith(middle_init, ".")) 
            middle_init = middle_init.substr(0, middle_init.length() - 1);
        
        string new_middle_init = get_first_initial(middle_init);
        string new_first_init = get_first_initial(last);
        if (!NStr::IsBlank(new_first_init))
        {
            string init = new_first_init + ".";
            if (!NStr::IsBlank(new_middle_init))
                init += new_middle_init + ".";
            pid.SetName().SetInitials(init);
        }
    }
}

void CReverseAuthorNames::ApplyToCAuth(CAuth_list& authors, bool& modified) // HasName in valid_pub.cpp - to get to last,first names
{
    if ( authors.CanGetNames() ) {
        CAuth_list::TNames &author_names = authors.SetNames();
        switch ( author_names.Which() ) {
            case CAuth_list::TNames::e_Std:
                for ( list< CRef< CAuthor > >::iterator auth = author_names.SetStd().begin();  auth != author_names.SetStd().end(); auth++  ) {
                    CPerson_id &pid = (*auth)->SetName();
                    if ( pid.IsName() ) {
                        if ( ! NStr::IsBlank(pid.GetName().GetLast()) && ! NStr::IsBlank(pid.GetName().GetFirst())) {
                            string last = pid.GetName().GetLast();
                            string first = pid.GetName().GetFirst();
                            pid.SetName().SetLast(first);
                            pid.SetName().SetFirst(last);
                            update_initials(pid,first,last);
                            modified = true;
                        }
                    }
                }
                break;
            default:
                break;
        }
    }
}




void CFixAuthorInitials::ApplyToCAuth(CAuth_list& authors, bool& modified) // HasName in valid_pub.cpp - to get to last,first names
{
    if ( authors.CanGetNames() ) {
        CAuth_list::TNames &author_names = authors.SetNames();
        switch ( author_names.Which() ) {
        case CAuth_list::TNames::e_Std:
            for ( list< CRef< CAuthor > >::iterator auth = author_names.SetStd().begin();  auth != author_names.SetStd().end(); auth++  ) {
                CPerson_id &pid = (*auth)->SetName();
                if ( pid.IsName() ) 
                    fix_initials(pid,modified);
                
            }
            break;
        default:
            break;
        }
    }
}

string CFixAuthorCap::fix_name_cap(string name, bool bApostrophes, bool& modified) // FixCapitalizationInElement in sqnutil2.c
{
    string result = name;
    FixCapitalizationInElement(result);
    if (bApostrophes) CapitalizeAfterApostrophe(result);

    if (result != name) modified = true;
    return result;
}


void CFixAuthorCap::ApplyToCAuth(CAuth_list& authors, bool& modified) // HasName in valid_pub.cpp - to get to last,first names
{
    if ( authors.CanGetNames() ) {
        CAuth_list::TNames &author_names = authors.SetNames();
        switch ( author_names.Which() ) {
        case CAuth_list::TNames::e_Std:
            for ( list< CRef< CAuthor > >::iterator auth = author_names.SetStd().begin();  auth != author_names.SetStd().end(); auth++  ) {
                CPerson_id &pid = (*auth)->SetName();
                if ( pid.IsName() ) 
                {
                    string last = pid.GetName().GetLast();
                    string first = pid.GetName().GetFirst();
                    last = fix_name_cap(last,true,modified);
                    first = fix_name_cap(first,false,modified);
                    pid.SetName().SetLast(last);
                    pid.SetName().SetFirst(first);
                    fix_initials(pid,modified);
                }
            }
            break;
        default:
            break;
        }
    }
}

void CAffilEditor::ApplyToCAuth(CAuth_list& authors, bool& modified)
{
    if ( authors.CanGetAffil() ) {
        CAffil& affil = authors.SetAffil();
        switch ( affil.Which() ) {
        case CAffil::e_Std: {
            CAffil::TStd& std = affil.SetStd();
            ApplyToCAffil(std,modified);
        }
            break;
        default:
                break;
        }
    }
    if ( authors.CanGetNames() ) {
        CAuth_list::TNames& author_names = authors.SetNames();
        switch ( author_names.Which() ) {
        case CAuth_list::TNames::e_Std:
            for ( list< CRef< CAuthor > >::iterator auth = author_names.SetStd().begin();  auth != author_names.SetStd().end(); auth++  ) 
                if ((*auth)->CanGetAffil())
                {
                    CAffil::TStd& std = (*auth)->SetAffil().SetStd();
                    ApplyToCAffil(std,modified);
                }
            break;
        default:
            break;
        }
    }

}

void CFixUSA::ApplyToCAffil(CAffil::TStd& std, bool& modified)
{
    if (std.IsSetCountry() && !NStr::IsBlank(std.GetCountry()))    {

        string country = std.GetCountry();
        NStr::ReplaceInPlace (country, "  ", " ");
        NStr::TruncateSpacesInPlace (country);

        if ( NStr::CompareNocase(country, "United States of America") == 0 ||
             NStr::CompareNocase(country, "United States") == 0 ||
             NStr::CompareNocase(country, "U.S.A.") == 0 ||
             NStr::CompareNocase(country, "U S A") == 0 )
        {
            std.SetCountry("USA");
            modified = true;
        }
            
        if (NStr::EqualCase (std.GetCountry(), "USA") && std.IsSetSub() && !NStr::IsBlank (std.GetSub())) {
            string state = std.GetSub();
            string result = state;
            FixStateAbbreviationsInAffil(result);
            if (result != state)
            {
                std.SetSub(result);
                modified = true;
            }
        }
    }
}


void CFixCountryCap::ApplyToCAffil(CAffil::TStd& std, bool& modified)  // FixCapitalizationInCountryString
{
    if (std.IsSetCountry() && !NStr::IsBlank(std.GetCountry()))    {

        string country = std.GetCountry();
        string result = country;
        NStr::ReplaceInPlace (result, "  ", " ");
        NStr::TruncateSpacesInPlace (result);

        InsertMissingSpacesAfterCommas(result);
        InsertMissingSpacesAfterNo(result);
        FixCapitalizationInElement(result);
        FixShortWordsInElement(result);
        FixAbbreviationsInElement(result);
        FindReplaceString_CountryFixes(result);
     
        if (result != country) 
        {
            modified = true;
            std.SetCountry(result);
        }
    }
}


void CFixAffilCap::ApplyToCAffil(CAffil::TStd& std, bool& modified) // FixCapsInPubAffilEx
{
    if (std.IsSetAffil() && !NStr::IsBlank(std.GetAffil())) 
    {
        string affil = std.GetAffil();
        NStr::ReplaceInPlace (affil, "  ", " ");
        NStr::TruncateSpacesInPlace (affil);
        FixCapitalizationInElement(affil);
        FixShortWordsInElement(affil);
        FixAbbreviationsInElement(affil);
        FixAffiliationShortWordsInElement(affil);
        FixKnownAbbreviationsInElement(affil);
        InsertMissingSpacesAfterCommas (affil);
        InsertMissingSpacesAfterNo(affil);
        FixOrdinalNumbers (affil);
        if (affil != std.GetAffil()) 
        {
            modified = true;
            std.SetAffil(affil);
        }
    }

    if (std.IsSetDiv() && !NStr::IsBlank(std.GetDiv())) 
    {
        string div = std.GetDiv();
        NStr::ReplaceInPlace (div, "  ", " ");
        NStr::TruncateSpacesInPlace (div);
        FixCapitalizationInElement(div);
        FixShortWordsInElement(div);
        FixAbbreviationsInElement(div);
        FixAffiliationShortWordsInElement(div);
        FixKnownAbbreviationsInElement(div);
        InsertMissingSpacesAfterCommas(div);
        InsertMissingSpacesAfterNo(div);
        FixOrdinalNumbers (div);
        if (div != std.GetDiv()) 
        {
            modified = true;
            std.SetDiv(div);
        }
    }
    
    if (std.IsSetCity() && !NStr::IsBlank(std.GetCity())) 
    {
        string city = std.GetCity();
        NStr::ReplaceInPlace (city, "  ", " ");
        NStr::TruncateSpacesInPlace (city);
        FixCapitalizationInElement(city);
        FixShortWordsInElement(city);
        FixAffiliationShortWordsInElement(city);
        FixKnownAbbreviationsInElement(city);        
        InsertMissingSpacesAfterCommas(city);
        InsertMissingSpacesAfterNo(city);
        FixOrdinalNumbers (city);
        if (city != std.GetCity()) 
        {
            modified = true;
            std.SetCity(city);
        }
    }

    if (std.IsSetStreet() && !NStr::IsBlank(std.GetStreet())) 
    {
        string street = std.GetStreet();
        NStr::ReplaceInPlace (street, "  ", " ");
        NStr::TruncateSpacesInPlace (street);
        FixKnownAbbreviationsInElement(street);
        FixCapitalizationInElement(street);
        FixShortWordsInElement(street);
        FixAffiliationShortWordsInElement(street);
        InsertMissingSpacesAfterCommas(street);
        InsertMissingSpacesAfterNo(street);
        FixOrdinalNumbers(street);
        if (street != std.GetStreet()) 
        {
            modified = true;
            std.SetStreet(street);
        }
    }



    if (std.IsSetCountry() && !NStr::IsBlank(std.GetCountry())) 
    {
        string country = std.GetCountry();
        NStr::ReplaceInPlace (country, "  ", " ");
        NStr::TruncateSpacesInPlace (country);
        InsertMissingSpacesAfterCommas(country);
        InsertMissingSpacesAfterNo(country);
        FixCapitalizationInElement(country);
        FixShortWordsInElement(country);
        FixAbbreviationsInElement(country);
        FindReplaceString_CountryFixes(country);
        if (country != std.GetCountry()) 
        {
            modified = true;
            std.SetCountry(country);
        }
    }
    
    if (std.IsSetSub() && !NStr::IsBlank (std.GetSub()))
    {
        if (NStr::EqualCase (std.GetCountry(), "USA")) 
        {
            string state = std.GetSub();
            string result = state;
            NStr::ReplaceInPlace (result, "  ", " ");
            NStr::TruncateSpacesInPlace (result);
            FixStateAbbreviationsInAffil(result);
            if (result.length() == 2) result = NStr::ToUpper(result);
            if (result != state)
            {
                std.SetSub(result);
                modified = true;
            }
        }
        else
        {
            string sub = std.GetSub();
            NStr::ReplaceInPlace (sub, "  ", " ");
            NStr::TruncateSpacesInPlace (sub);
            FixCapitalizationInElement(sub);
            FixShortWordsInElement(sub);
            FixAffiliationShortWordsInElement(sub);
            InsertMissingSpacesAfterCommas(sub);
            if (sub != std.GetSub()) 
            {
                modified = true;
                std.SetSub(sub);
            }
        }  
    }

}


void CFixTitleCap::ApplyToPubdesc (objects::CPubdesc& pubdesc, bool& modified)
{
    EDIT_EACH_PUB_ON_PUBDESC (pub_iter, pubdesc) {
        CPub& pub = **pub_iter;
        string title;
        switch( pub.Which() ) {
        case CPub::e_Article: {
            CCit_art& art = pub.SetArticle();
            if ( art.IsSetTitle() ) 
            {
                title = getTitle(art.GetTitle());
                if (!NStr::IsBlank(title))
                {
                    title = fix_title(title,modified);
                    if (modified)
                        setTitle(art.SetTitle(),title);
                }
            }
        }
            break;
        case CPub::e_Gen:
            if ( pub.GetGen().IsSetTitle() ) {
                title = pub.GetGen().GetTitle();
                if (!NStr::IsBlank(title)) 
                {
                    title = fix_title(title,modified);
                    if (modified)
                        pub.SetGen().SetTitle(title);
                }
            }
            break;
        case CPub::e_Sub:
            if ( pub.GetSub().IsSetDescr() ) {
                title = pub.GetSub().GetDescr();
                if (!NStr::IsBlank(title)) 
                {
                    title = fix_title(title,modified);
                    if (modified)
                        pub.SetSub().SetDescr(title);
                }
            }
            break;
        case CPub::e_Patent:
            if ( pub.GetPatent().IsSetTitle() ) {
                title = pub.GetPatent().GetTitle();
                if (!NStr::IsBlank(title)) 
                {
                    title = fix_title(title,modified);
                    if (modified)
                        pub.SetPatent().SetTitle(title);
                }
            }
            break;
        default:
            break;
        }
    }
}

string CFixTitleCap::getTitle(const CTitle& title) // CValidError_imp::HasTitle(const CTitle& title) in valid_pub.cpp
{
    ITERATE (CTitle::Tdata, item, title.Get() ) {
        const string *str = 0;
        switch ( (*item)->Which() ) {
        case CTitle::C_E::e_Name:
            str = &(*item)->GetName();
            break;

        case CTitle::C_E::e_Tsub:
            str = &(*item)->GetTsub();
            break;

        case CTitle::C_E::e_Trans:
            str = &(*item)->GetTrans();
            break;

        case CTitle::C_E::e_Jta:
            str = &(*item)->GetJta();
            break;

        case CTitle::C_E::e_Iso_jta:
            str = &(*item)->GetIso_jta();
            break;

        case CTitle::C_E::e_Ml_jta:
            str = &(*item)->GetMl_jta();
            break;

        case CTitle::C_E::e_Coden:
            str = &(*item)->GetCoden();
            break;

        case CTitle::C_E::e_Issn:
            str = &(*item)->GetIssn();
            break;

        case CTitle::C_E::e_Abr:
            str = &(*item)->GetAbr();
            break;

        case CTitle::C_E::e_Isbn:
            str = &(*item)->GetIsbn();
            break;

        default:
            break;
        };
        if ( !NStr::IsBlank(*str) )  return *str;
        
    }
    return "";
}

void CFixTitleCap::setTitle(CTitle& title, string str) // CValidError_imp::HasTitle(const CTitle& title) in valid_pub.cpp
{
     NON_CONST_ITERATE (CTitle::Tdata, item, title.Set() ) {
        switch ( (*item)->Which() ) {
        case CTitle::C_E::e_Name:
            (*item)->SetName(str);
            break;

        case CTitle::C_E::e_Tsub:
            (*item)->SetTsub(str);
            break;

        case CTitle::C_E::e_Trans:
            (*item)->SetTrans(str);
            break;

        case CTitle::C_E::e_Jta:
            (*item)->SetJta(str);
            break;

        case CTitle::C_E::e_Iso_jta:
            (*item)->SetIso_jta(str);
            break;

        case CTitle::C_E::e_Ml_jta:
            (*item)->SetMl_jta(str);
            break;

        case CTitle::C_E::e_Coden:
            (*item)->SetCoden(str);
            break;

        case CTitle::C_E::e_Issn:
            (*item)->SetIssn(str);
            break;

        case CTitle::C_E::e_Abr:
            (*item)->SetAbr(str);
            break;

        case CTitle::C_E::e_Isbn:
            (*item)->SetIsbn(str);
            break;

        default:
            break;
        };
    }
}

string CFixTitleCap::fix_title(string title, bool& modified)
{
    if (title.empty()) return title;
   
    string result = title;
    FixCapitalizationInElement(result);
    CapitalizeSAfterNumber(result);
    FixAbbreviationsInElement(result);
    FixOrgNames(m_tse,result); 

    if (result != title) modified = true;
    return result;
}

void CFixTitleCap::set_seh(objects::CSeq_entry_Handle tse)
{
    m_tse = tse;
}

void CFixAllCap::ApplyToPub(const CPubdesc& pubdesc, CCmdComposite* composite, const CSeq_entry* se, const CSeqdesc* seqDesq, const CSeq_feat* seqFeat) // CEditObjectPubdesc::GetEditCommand from edit_object_seq_desc
{
    CRef<CSerialObject> edited_object;
    edited_object.Reset((CSerialObject*)CPubdesc::GetTypeInfo()->Create());
    edited_object->Assign(pubdesc);
    CPubdesc& edited_pubdesc = dynamic_cast<CPubdesc&>(*edited_object);
    bool modified = false;

    CFixAffilCap affil_cap;
    affil_cap.ApplyToPubdesc (edited_pubdesc,modified);
    CFixAuthorCap author_cap;
    author_cap.ApplyToPubdesc (edited_pubdesc,modified);
    CFixTitleCap title_cap;
    title_cap.set_seh(m_tse);
    title_cap.ApplyToPubdesc (edited_pubdesc,modified);

    if (modified)
    {
        CChangePubdescCommand* cmd = new CChangePubdescCommand();
        CObject* actual = (CObject*) &pubdesc;
        cmd->Add(actual, CConstRef<CObject>(edited_object));
        composite->AddCommand(*cmd);
    }
}

void CRemovePubs::ApplyToPub (const CPubdesc& pubdesc, CCmdComposite* composite, const CSeq_entry* se, const CSeqdesc* seqDesq, const CSeq_feat* seqFeat) 
{

    CRef<CSerialObject> edited_object;
    edited_object.Reset((CSerialObject*)CPubdesc::GetTypeInfo()->Create());
    edited_object->Assign(pubdesc);
    CPubdesc& edited_pubdesc = dynamic_cast<CPubdesc&>(*edited_object);
    bool modified = false;
    ApplyToPubdesc (edited_pubdesc,modified);
    if (modified)
    {
        CChangePubdescCommand* cmd = new CChangePubdescCommand();
        CObject* actual = (CObject*) &pubdesc;
        cmd->Add(actual, CConstRef<CObject>(edited_object));
        composite->AddCommand(*cmd);

        CSeq_entry_Handle& tse = m_tse;

        if (seqDesq) 
        {
            DeleteObject(tse, se, *seqDesq, composite);
            return;
        }
        
        if (seqFeat) 
            DeleteObject(tse, *seqFeat, composite);
    }
}


void CRemovePubs::DeleteObject(objects::CSeq_entry_Handle& tse, const CSeq_entry* se, const objects::CSeqdesc& seqDesq, CCmdComposite* composite) // see x_DeleteObject in bioseq_editor.cpp
{
    CSeqdesc::E_Choice choice = seqDesq.Which();

    CSeq_entry_Handle sehDesc;
    CSeq_entry_Handle seh = tse.GetScope().GetSeq_entryHandle(*se);

    CSeqdesc_CI it(seh, choice);
    for (; it; ++it) {
        if (&seqDesq == &*it) {
            sehDesc = it.GetSeq_entry_Handle();
            break;
        }
    }

    if (!sehDesc)
        return;
    
    CIRef<IEditCommand> cmdDelDesc(new CCmdDelDesc(sehDesc, seqDesq));
    composite->AddCommand(*cmdDelDesc);
    
    for (CSeq_entry_CI it(sehDesc.GetSet()); it; ++it) {
        if (*it != seh) {
            CIRef<IEditCommand> cmdAddDesc(new CCmdCreateDesc(*it, seqDesq));
            composite->AddCommand(*cmdAddDesc);
        }
    }
}

void CRemovePubs::DeleteObject(objects::CSeq_entry_Handle& seh, const objects::CSeq_feat& seqFeat, CCmdComposite* composite) // see x_DeleteObject in bioseq_editor.cpp
{
    if (!seqFeat.IsSetData())  return;

    const CSeqFeatData& data = seqFeat.GetData();

    CSeq_feat_Handle fh;
    SAnnotSelector sel(seqFeat.GetData().GetSubtype());
    for (CFeat_CI mf(seh.GetScope(), seqFeat.GetLocation(), sel); mf; ++mf) {
        if (mf->GetOriginalFeature().Equals(seqFeat)) {
            fh = mf->GetSeq_feat_Handle();
            break;
        }
    }

    if (!fh)  return;

    CIRef<IEditCommand> mainCmd(new CCmdDelSeq_feat(fh));
    composite->AddCommand(*mainCmd);
}

void CRemoveUnpublishedPubs::ApplyToPubdesc (objects::CPubdesc& pubdesc, bool& modified)
{
    if (pubdesc.IsSetPub())
    {
        CPub_set::TPub& pe = pubdesc.SetPub();
        for (CPub_set::TPub::iterator pub_iter = pe.begin(); pub_iter != pe.end(); ) 
        {
            CPub& pub = **pub_iter;
            bool to_delete = false;
            switch( pub.Which() ) {
            case CPub::e_Gen:
            {
                const CCit_gen& gen = pub.GetGen();
                if ( gen.IsSetCit()  &&  !gen.GetCit().empty() ) {
                    const string& cit = gen.GetCit();
                    if (NStr::StartsWith (cit, "unpublished", NStr::eNocase))
                    {
                        // Cit-gen.cit = "unpublished"
                        to_delete = true;
                    }
                }
            }
            break;
            case CPub::e_Article:
            {
                const CCit_art& art = pub.GetArticle();
                if ( art.GetFrom().IsJournal() ) 
                {
                    const CCit_jour& jour = art.GetFrom().GetJournal();
                    if ( jour.CanGetImp() ) 
                    {
                        const CImprint& imp = jour.GetImp();
                        if ( imp.IsSetPrepub() && imp.GetPrepub() != CImprint::ePrepub_in_press)
                        {
                            // Cit-art.from.jour.imp.prepub is not 0 or 2
                            to_delete = true;
                        }
                    }
                }
                if ( art.GetFrom().IsBook() ) 
                {
                    const CCit_book& book = art.GetFrom().GetBook();
                    if ( book.CanGetImp() ) 
                    {
                        const CImprint& imp = book.GetImp();
                        if ( imp.IsSetPrepub() && imp.GetPrepub() != CImprint::ePrepub_in_press)
                        {
                            // Cit-art.from.book.imp.prepub is not 0 or 2
                            to_delete = true;
                        }
                    }
                }
            }
            break;
            case CPub::e_Book:
            {
                const CCit_book& book = pub.GetBook();
                if ( book.CanGetImp() )
                { 
                    const CImprint& imp = book.GetImp();
                
                    if ( imp.IsSetPrepub() && imp.GetPrepub() != CImprint::ePrepub_in_press)
                    {
                        // Cit-book where Cit-book.imp.prepub is not 0 or 2 
                        to_delete = true;
                    }
                }
            }
            break;
            case CPub::e_Journal:
            {
                const CCit_jour& jour = pub.GetJournal();
                if ( jour.CanGetImp() ) 
                {
                    const CImprint& imp = jour.GetImp();
                    
                    if ( imp.IsSetPrepub() && imp.GetPrepub() != CImprint::ePrepub_in_press)
                    {
                        // Cit-jour where Cit-jour.imp.prepub is not 0 or 2 
                        to_delete = true;
                    }
                }
            }
            break;
            case CPub::e_Man:
            {
                const CCit_book& book = pub.GetMan().GetCit();
                if ( book.CanGetImp() ) 
                {
                    const CImprint& imp = book.GetImp();
                    
                    if ( imp.IsSetPrepub() && imp.GetPrepub() != CImprint::ePrepub_in_press)
                    {
                        // Cit-let where Cit-let.imp.prepub is not 0 or 2
                        to_delete = true;
                    }
                    
                }
            }
            break;
            default:
                break;
            }
            if (to_delete)
            {
                modified = true;
                pub_iter = pe.erase(pub_iter);
            }
            else
                ++pub_iter;
            
        }
    }

}

void CRemovePublishedPubs::ApplyToPubdesc (objects::CPubdesc& pubdesc, bool& modified)
{
    if (pubdesc.IsSetPub())
    {
        CPub_set::TPub& pe = pubdesc.SetPub();
        for (CPub_set::TPub::iterator pub_iter = pe.begin(); pub_iter != pe.end(); ) 
        {
            CPub& pub = **pub_iter;
            bool to_delete = false;
            switch( pub.Which() ) {
            case CPub::e_Gen:
            {
                const CCit_gen& gen = pub.GetGen();
                if ( gen.IsSetCit()  &&  !gen.GetCit().empty() ) {
                    const string& cit = gen.GetCit();
                    if (!NStr::StartsWith (cit, "unpublished", NStr::eNocase))
                    {
                        // Cit-gen.cit != "unpublished"
                        to_delete = true;
                    }
                }
            }
            break;
            case CPub::e_Patent:
                to_delete = true;
                break;
            case CPub::e_Article:
            {
                const CCit_art& art = pub.GetArticle();
                if ( art.GetFrom().IsJournal() ) 
                {
                    const CCit_jour& jour = art.GetFrom().GetJournal();
                    if ( jour.CanGetImp() ) 
                    {
                        const CImprint& imp = jour.GetImp();
                        if ( !imp.IsSetPrepub() )
                        {
                            // Cit-art.from.jour.imp.prepub is 0
                            to_delete = true;
                        }
                    }
                }
                if ( art.GetFrom().IsBook() ) 
                {
                    const CCit_book& book = art.GetFrom().GetBook();
                    if ( book.CanGetImp() ) 
                    {
                        const CImprint& imp = book.GetImp();
                        if ( !imp.IsSetPrepub() )
                        {
                            // Cit-art.from.book.imp.prepub is 0 
                            to_delete = true;
                        }
                    }
                }
            }
            break;
            case CPub::e_Book:
            {
                const CCit_book& book = pub.GetBook();
                if ( book.CanGetImp() )
                { 
                    const CImprint& imp = book.GetImp();
                
                    if ( !imp.IsSetPrepub() )
                    {
                        // Cit-book where Cit-book.imp.prepub is 0
                        to_delete = true;
                    }
                }
            }
            break;
            case CPub::e_Journal:
            {
                const CCit_jour& jour = pub.GetJournal();
                if ( jour.CanGetImp() ) 
                {
                    const CImprint& imp = jour.GetImp();
                    
                    if ( !imp.IsSetPrepub() )
                    {
                        // Cit-jour where Cit-jour.imp.prepub is  0
                        to_delete = true;
                    }
                }
            }
            break;
            case CPub::e_Man:
            {
                const CCit_book& book = pub.GetMan().GetCit();
                if ( book.CanGetImp() ) 
                {
                    const CImprint& imp = book.GetImp();
                    
                    if ( !imp.IsSetPrepub() )
                    {
                        // Cit-let where Cit-let.imp.prepub is 0
                        to_delete = true;
                    }
                    
                }
            }
            break;
            default:
                break;
            }
            if (to_delete)
            {
                modified = true;
                pub_iter = pe.erase(pub_iter);
            }
            else
                ++pub_iter;
            
        }
    }

}


void CRemoveCollidingPubs::ApplyToPubdesc (objects::CPubdesc& pubdesc, bool& modified)
{
    FOR_EACH_PUB_ON_PUBDESC (pub_iter, pubdesc) {
        const CPub& pub = **pub_iter;
        string label;
        pub.GetLabel(&label, CPub::eContent, CPub::fLabel_Unique, CPub::eLabel_V1 );
        if (!m_SecondPass)
            m_collisions[label]++;
        else if (m_collisions[label] > 1)
            modified = true;
    }
}

void CRemoveCollidingPubs::apply_to_seq_and_feat_single_pass(objects::CSeq_entry_Handle tse, CCmdComposite* composite)
{
    ApplyToCSeq_entry (*(tse.GetCompleteSeq_entry()), composite);
    CFeat_CI feat (tse, SAnnotSelector(CSeqFeatData::e_Pub));
    while (feat) {
        ApplyToPub (feat->GetData().GetPub(), composite, NULL, NULL,&(feat->GetOriginalFeature())); 
        ++feat;
    }
}

void CRemoveCollidingPubs::apply_to_seq_and_feat(objects::CSeq_entry_Handle tse, CCmdComposite* composite)
{
    m_SecondPass = false;
    apply_to_seq_and_feat_single_pass(tse, composite);

    m_SecondPass = true;
    apply_to_seq_and_feat_single_pass(tse, composite);
}


void CRemoveAllPubs::ApplyToPub(const CPubdesc& pubdesc, CCmdComposite* composite, const CSeq_entry* se, const CSeqdesc* seqDesq, const CSeq_feat* seqFeat) // CEditObjectPubdesc::GetEditCommand from edit_object_seq_desc
{
    CRef<CSerialObject> edited_object;
    edited_object.Reset((CSerialObject*)CPubdesc::GetTypeInfo()->Create());
    edited_object->Assign(pubdesc);
    CPubdesc& edited_pubdesc = dynamic_cast<CPubdesc&>(*edited_object);
    bool modified = false;
    if (m_SecondPass)
    {
        CRemoveUnpublishedPubs a;
        a.ApplyToPubdesc (edited_pubdesc,modified);
        CRemovePublishedPubs b;
        b.ApplyToPubdesc (edited_pubdesc,modified);

        ApplyToPubdesc (edited_pubdesc,modified);

        if (modified)
        {
            CChangePubdescCommand* cmd = new CChangePubdescCommand();
            CObject* actual = (CObject*) &pubdesc;
            cmd->Add(actual, CConstRef<CObject>(edited_object));
            composite->AddCommand(*cmd);
            
            CSeq_entry_Handle& tse = m_tse;
            
            if (seqDesq) 
            {
                DeleteObject(tse, se, *seqDesq, composite);
                return;
            }
            
            if (seqFeat) 
                DeleteObject(tse, *seqFeat, composite);
        }
    }
    else
       ApplyToPubdesc (edited_pubdesc,modified);
    
}

void CStripAuthorSuffix::ApplyToCAuth(CAuth_list& authors, bool& modified) // HasName in valid_pub.cpp - to get to last,first names
{
    if ( authors.CanGetNames() ) {
        CAuth_list::TNames &author_names = authors.SetNames();
        switch ( author_names.Which() ) {
            case CAuth_list::TNames::e_Std:
                for ( list< CRef< CAuthor > >::iterator auth = author_names.SetStd().begin();  auth != author_names.SetStd().end(); auth++  ) 
		  {
                    CPerson_id &pid = (*auth)->SetName();
                    if ( pid.IsName() && pid.GetName().CanGetSuffix() && !NStr::IsBlank(pid.GetName().GetSuffix()) )
		      {
			pid.SetName().ResetSuffix();
			modified = true;
		      }
		  }
	
		break;
	default:
	  break;
	}
    }
}

void CRemoveAuthorConsortium::ApplyToCAuth(CAuth_list& authors, bool& modified) // HasName in valid_pub.cpp - to get to last,first names
{
    if ( authors.CanGetNames() ) {
        CAuth_list::TNames &author_names = authors.SetNames();
        switch ( author_names.Which() ) {
            case CAuth_list::TNames::e_Std:
                for ( list< CRef< CAuthor > >::iterator auth = author_names.SetStd().begin();  auth != author_names.SetStd().end();   ) 
		  {
                    CPerson_id &pid = (*auth)->SetName();
                    if ( pid.IsConsortium() )
		      {
			auth = author_names.SetStd().erase(auth);
			modified = true;
		      }
		    else
		      ++auth;
		  }
	
		break;
	default:
	  break;
	}
    }
}

void CConvertAllAuthorToConsortium::ApplyToCAuth(CAuth_list& authors, bool& modified) // HasName in valid_pub.cpp - to get to last,first names
{
    if ( authors.CanGetNames() ) {
        CAuth_list::TNames &author_names = authors.SetNames();
        switch ( author_names.Which() ) {
            case CAuth_list::TNames::e_Std:
                for ( list< CRef< CAuthor > >::iterator auth = author_names.SetStd().begin();  auth != author_names.SetStd().end(); ++auth  ) 
		  {
                    CPerson_id &pid = (*auth)->SetName();
                    if ( pid.IsName() && pid.GetName().CanGetLast())
		      {
			string last = pid.GetName().GetLast();
			modified = true;
			pid.Reset();
			pid.Select(CPerson_id::e_Consortium);
			pid.SetConsortium(last);
		      }
		  }
	
		break;
	default:
	  break;
	}
    }
}


void CConvertAuthorToConsortiumWhereConstraint::SetConstraint(const string &input)
{
  m_Regex = new CRegexp("\\b"+input+"\\b",CRegexp::fCompile_ignore_case);
}

CConvertAuthorToConsortiumWhereConstraint::~CConvertAuthorToConsortiumWhereConstraint()
{
  if (m_Regex) delete m_Regex;
}

void CConvertAuthorToConsortiumWhereConstraint::ApplyToCAuth(CAuth_list& authors, bool& modified) // HasName in valid_pub.cpp - to get to last,first names
{
    if ( authors.CanGetNames() ) {
        CAuth_list::TNames &author_names = authors.SetNames();
        switch ( author_names.Which() ) {
            case CAuth_list::TNames::e_Std:
                for ( list< CRef< CAuthor > >::iterator auth = author_names.SetStd().begin();  auth != author_names.SetStd().end(); ++auth  ) 
		  {
                    CPerson_id &pid = (*auth)->SetName();
                    if ( pid.IsName() && pid.GetName().CanGetLast() && m_Regex)
		      {
			string last = pid.GetName().GetLast();
			m_Regex->GetMatch(last, 0, 0, CRegexp::fMatch_default, true);
			if (m_Regex->NumFound() > 0) 
			  {
			    modified = true;
			    pid.Reset();
			    pid.Select(CPerson_id::e_Consortium);
			    pid.SetConsortium(last);
			  }
		      }
		  }
	
		break;
	default:
	  break;
	}
    }
}


void CReloadPublications::LookUpByPmid(CPub& pub, int pmid, bool& modified)
{
    string pmid_str = NStr::IntToString(pmid);
    if (!pmid_str.empty())
    {
        CRef<CEUtils_ConnContext> ctx(new CEUtils_ConnContext);
        CEFetch_Request req(ctx);
        req.SetRetMode(CEFetch_Request::eRetMode_asn);
        req.SetDatabase("pubmed");
        req.SetRetMax(1);
        req.GetId().SetIds(pmid_str);
        CObjectIStream* istr = req.GetObjectIStream();
        if (istr->InGoodState() && !istr->EndOfData())
        {
            string obj_type;
            try {
                obj_type = istr->ReadFileHeader();   // If nothing returned it's crashing here
                } catch(...) {} // truly ugly but there doesn't seem to be any other way
            CRef<CSerialObject> obj;
            if (obj_type == "Pubmed-entry") 
            {
                obj.Reset(new CPubmed_entry());
                istr->ReadObject(obj.GetPointer(), CPubmed_entry::GetTypeInfo());
                CPubmed_entry& pubmed = dynamic_cast<CPubmed_entry &>(*obj);
                if (pubmed.CanGetMedent() && pubmed.GetMedent().CanGetCit())
                {
                    CCit_art& article = pubmed.SetMedent().SetCit();
                    pub.SetArticle(article);
                    modified = true;
                }
            }
        }
    }
}

void CReloadPublications::ApplyToPubdesc (objects::CPubdesc& pubdesc, bool& modified)
{
    int pmid = 0;
    EDIT_EACH_PUB_ON_PUBDESC (pub_iter, pubdesc) {
        CPub& pub = **pub_iter;
        switch( pub.Which() ) {
        case CPub::e_Pmid:
            pmid = pub.GetPmid();
            break;
        default:
            break;
        }
    }
    if (pmid !=0 )
    {
        EDIT_EACH_PUB_ON_PUBDESC (pub_iter, pubdesc) {
            CPub& pub = **pub_iter;
            switch( pub.Which() ) {
            case CPub::e_Article: 
                LookUpByPmid(pub,pmid, modified);
                break;
            default:
                break;
            }
        }
    }
}

END_NCBI_SCOPE
