/*  $Id: srcedit_util.cpp 27966 2013-05-01 19:06:59Z katargir $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data,  the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties,  express or implied,  including
 *  warranties of performance,  merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Colleen Bollin
 */


#include <ncbi_pch.hpp>

#include <objects/seqset/Seq_entry.hpp>
#include <objects/submit/Seq_submit.hpp>
#include <objects/submit/Submit_block.hpp>
#include <objects/submit/Contact_info.hpp>
#include <objects/biblio/Author.hpp>
#include <objects/biblio/Affil.hpp>
#include <objects/general/Person_id.hpp>
#include <objects/general/Name_std.hpp>
#include <objects/general/User_object.hpp>
#include <objects/general/User_field.hpp>
#include <objects/general/Object_id.hpp>
#include <objects/seqfeat/BioSource.hpp>
#include <objects/seqfeat/Org_ref.hpp>
#include <objects/seqfeat/OrgName.hpp>
#include <objects/seqfeat/PCRReactionSet.hpp>
#include <objects/seqfeat/PCRReaction.hpp>
#include <objects/seqfeat/PCRPrimerSet.hpp>
#include <objects/seqfeat/PCRPrimer.hpp>
#include <objects/seqfeat/PCRPrimerSeq.hpp>
#include <objects/seqset/Bioseq_set.hpp>
#include <objects/seq/Bioseq.hpp>
#include <objects/seqloc/Seq_id.hpp>
#include <objmgr/seqdesc_ci.hpp>
#include <objmgr/bioseq_ci.hpp>
#include <objmgr/feat_ci.hpp>

#include <gui/widgets/wx/table_utils.hpp>
#include <gui/widgets/wx/wx_utils.hpp>
#include <gui/packages/pkg_sequence_edit/srcedit_util.hpp>
#include <gui/widgets/object_list/object_list_widget.hpp>
#include <gui/objutils/cmd_composite.hpp>
#include <gui/objutils/cmd_create_desc.hpp>
#include <gui/objutils/cmd_del_desc.hpp>
#include <gui/objutils/descriptor_change.hpp>
#include <gui/core/file_extensions.hpp>
#include <gui/widgets/wx/csv_exporter.hpp>
#include <gui/widgets/edit/biosource_autocomplete.hpp>

#include <objects/seqtable/SeqTable_multi_data.hpp>
#include <objects/seqtable/SeqTable_column_info.hpp>
#include <objects/seqloc/Seq_id.hpp>

#include <gui/packages/pkg_sequence_edit/subprep_util.hpp>

BEGIN_NCBI_SCOPE


CSrcQual::CSrcQual()
{
    m_Name = "";
    m_Required = false;
    m_ReportMissing = false;
    m_Example = "";
}


CSrcQual::CSrcQual(string name, bool required, bool report_missing, string example)
: m_Name (name), m_Required (required), m_ReportMissing (report_missing), m_Example (example)
{
}

CSrcQual::~CSrcQual()
{
}


string CSrcQual::GetValue(const objects::CBioSource& src)
{
    return "";
}

string CTaxnameQual::GetValue(const objects::CBioSource& src)
{
    if (src.IsSetOrg() && src.GetOrg().IsSetTaxname()) {
        return src.GetOrg().GetTaxname();
    } else {
        return "";
    }
}


string CUnculturedTaxnameQual::GetFormatErrors(string value)
{
    vector<string> unwise_names;
    unwise_names.push_back("unknown");
    unwise_names.push_back("uncultured organism");
    unwise_names.push_back("uncultured sample");
    unwise_names.push_back("uncultured");
    unwise_names.push_back("unidentified");
    
    vector<string>::iterator s = unwise_names.begin();
    while (s != unwise_names.end()) {
        if (NStr::EqualNocase(value, *s)) {
            return "Ambiguous organism name";
        }
        ++s;
    }
    if (NStr::StartsWith (value, "uncultured ")) {
        return "";
    } else {
        return "For uncultured samples, organism name should start with 'uncultured'";
    }
}


string CGenomeQual::GetValue(const objects::CBioSource& src)
{
    string val = "";

    if (src.IsSetGenome()) {
        int genome = src.GetGenome();
        val = objects::CBioSource::GetOrganelleByGenome (genome);
    }

    return val;
}


COrgModQual::COrgModQual(string name, bool required, bool report_missing, string example)
: CSrcQual(name, required, report_missing, example)
{
    m_Subtype = objects::COrgMod::GetSubtypeValue(name, objects::COrgMod::eVocabulary_insdc);
}


string COrgModQual::GetValue(const objects::CBioSource& src)
{
    if (src.IsSetOrg() && src.GetOrg().IsSetOrgname() && src.GetOrg().GetOrgname().IsSetMod()) {
        ITERATE(objects::COrgName::TMod, mit, src.GetOrg().GetOrgname().GetMod()) {
            if ((*mit)->GetSubtype() == m_Subtype && (*mit)->IsSetSubname()) {
                return (*mit)->GetSubname();
            }
        }
    }
    return "";
}


static bool s_IsAllNumbers (string val)
{
    bool all_numbers = true;
    if (NStr::IsBlank (val)) {
        return false;
    }
    string::iterator it = val.begin();
    while (it != val.end() && all_numbers) {
        if (!isdigit (*it)) {
            all_numbers = false;
        }
        ++it;
    }
    return all_numbers;
}


string COrgModQual::GetFormatErrors(string value)
{
    string rval = "";

    switch (m_Subtype) {
        case objects::COrgMod::eSubtype_nat_host:
            if (s_IsAllNumbers(value)) {
                rval = "Invalid host";
            }
            break;
        default:
            break;
    }
    return rval;           
}


CSubSrcQual::CSubSrcQual(string name, bool required, bool report_missing, string example)
: CSrcQual(name, required, report_missing, example)
{
    m_Subtype = objects::CSubSource::GetSubtypeValue(name, objects::CSubSource::eVocabulary_insdc);

    m_BadIsolationSourceValues.push_back("uncultured bacterium");
    m_BadIsolationSourceValues.push_back("uncultured bacteria");
    m_BadIsolationSourceValues.push_back("uncultured fungus");
    m_BadIsolationSourceValues.push_back("uncultured fungi");
    m_BadIsolationSourceValues.push_back("uncultured bacterium clone");
    m_BadIsolationSourceValues.push_back("uncultured bacteria clone");
    m_BadIsolationSourceValues.push_back("uncultured fungus clone");
    m_BadIsolationSourceValues.push_back("uncultured fungi clone");
    m_BadIsolationSourceValues.push_back("isolation source");

}


string CSubSrcQual::GetValue(const objects::CBioSource& src)
{
    if (src.IsSetSubtype()) {
        ITERATE(objects::CBioSource::TSubtype, mit, src.GetSubtype()) {
            if ((*mit)->GetSubtype() == m_Subtype && (*mit)->IsSetName()) {
                return (*mit)->GetName();
            }
        }
    }
    return "";
}


string CSubSrcQual::GetFormatErrors(string value)
{
    string rval = "";

    switch (m_Subtype) {
        case objects::CSubSource::eSubtype_isolation_source:
            if (value.length() < 3) {
                rval = "Suspiciously short isolation source";
            } else {
                ITERATE (vector<string>, it, m_BadIsolationSourceValues) {
                    if (NStr::EqualNocase(value, *it)) {
                        rval = "Invalid isolation source";
                        break;
                    }
                }
            }
            break;
        case objects::CSubSource::eSubtype_country:
            if (s_IsAllNumbers(value)) {
                rval = "Invalid country";
            }
            break;
        default:
            break;
    }
    return rval;           
}


CSourceRequirements::CSourceRequirements()
{
    m_Requirements.clear();
}


CSourceRequirements::~CSourceRequirements()
{
    for (size_t col = 0; col < m_Requirements.size(); col++) {
        delete(m_Requirements[col]);
    }
    m_Requirements.clear();
}


void CSourceRequirements::AddRequirement(string name, bool required, string example)
{
    // first, check to see if we already have requirement, increase requirement if necessary   
    NON_CONST_ITERATE (TSrcQualList, rit, m_Requirements) {
        if (MatchColumnName((*rit)->GetName(), name)) {
            if (required) {
                (*rit)->SetRequired(true);
            }
            if (!NStr::IsBlank(example)) {
                (*rit)->SetExample(example);
            }
            return;
        }
    }

    if (NStr::EqualNocase(name, "uncultured taxname")) {
        m_Requirements.push_back(new CUnculturedTaxnameQual("taxname", required, false, example));
    } else if (IsOrgColumnName(name)) {
        m_Requirements.push_back(new CTaxnameQual(name, required, false, example));
    } else {
        try {
            m_Requirements.push_back(new CSubSrcQual(name, required, false, example));
        } catch (CException& ) {
            m_Requirements.push_back(new COrgModQual(name, required, false, example));
        }        
    }
}


void CSourceRequirements::AddUniquenessList(vector<string> list)
{
    ITERATE(vector<string>, it, list) {
        AddRequirement(*it, false);
    }
    m_UniquenessLists.push_back(list);
}


void  CSourceRequirements::AddUniquenessList(int num, ... )
{
    va_list arguments;
    vector<string> list;

    va_start ( arguments, num );           
    for ( int x = 0; x < num; x++ ) {  
        string item = va_arg ( arguments, char * );
        AddRequirement(item, false);
        list.push_back(item ); 
    }
    va_end ( arguments ); 

    m_UniquenessLists.push_back(list);
}


void CSourceRequirements::AddOneOfList(vector<string> list)
{
    ITERATE(vector<string>, it, list) {
        AddRequirement(*it, false);
    }
    m_OneOfLists.push_back(list);
}


void  CSourceRequirements::AddOneOfList(int num, ... )
{
    va_list arguments;
    vector<string> list;

    va_start ( arguments, num );           
    for ( int x = 0; x < num; x++ ) {  
        string item = va_arg ( arguments, char * );
        AddRequirement(item, false);
        list.push_back(item ); 
    }
    va_end ( arguments ); 

    m_OneOfLists.push_back(list);
}


static string s_MakeUniquenessVal (vector<CRef<objects::CSeqTable_column> > cols, size_t row)
{
    size_t i;
    string this_val = "";

    for (i = 0; i < cols.size(); i++) {
        // later, remove non-uniqueing characters
        if (row < cols[i]->GetData().GetSize()) {
            this_val += cols[i]->GetData().GetString()[row] + ";";
        }
    }
    NStr::ToUpper(this_val);
    return this_val;
}


int CSourceRequirements::x_AddUniquenessProblems(CRef<objects::CSeq_table> values_table, vector<string> uniqueness_list, vector<string>& row_problems)
{
    int num_duplicates = 0;

    if (values_table->GetNum_rows() < 2 || uniqueness_list.size() < 1) {
        return 0;
    }

    size_t j;
    string description = uniqueness_list[0];
    for (j = 1; j < uniqueness_list.size(); j++) {
       description += "/" + uniqueness_list[j];
    }

    vector<CRef<objects::CSeqTable_column> > cols;
    for (j = 0; j < uniqueness_list.size(); j++) {
        CRef<objects::CSeqTable_column> col = FindSeqTableColumnByName(values_table, uniqueness_list[j]);
        if (col) {
            cols.push_back(col);
        }
    }

    CRef<objects::CSeqTable_column> id_col = FindSeqTableColumnByName(values_table, kSequenceIdColLabel);

    vector<string> u_list;
    int row;
    for (row = 0; row < values_table->GetNum_rows(); row++) {
        u_list.push_back(s_MakeUniquenessVal (cols, row));
    }
    sort(u_list.begin(), u_list.end());
    vector<string>::iterator sit = u_list.begin();
    vector<string>::iterator sit2 = sit;
    sit2++;
    bool is_dup = false;
    while (sit2 != u_list.end()) {
        if (NStr::Equal(*sit, *sit2)) {
            // if the next value matches this one, delete it
            sit2 = u_list.erase(sit2);
            is_dup = true;
        } else {
            if (!is_dup) {
                // if the next value does not match this one,
                // and there were no matches, delete from the list
                sit = u_list.erase(sit);
            } else {
                sit++;
            }
            sit2 = sit;
            if (sit2 != u_list.end()) {
                sit2++;
            }
            is_dup = false;
        }
    }
    if (!is_dup && sit != u_list.end()) {
        sit = u_list.erase(sit);
    }
    sit = u_list.begin();
    while (sit != u_list.end()) {
        bool is_first = true;
        int first_row = 0;
        for (row = 0; row < values_table->GetNum_rows(); row++) {
            if (NStr::Equal(*sit, s_MakeUniquenessVal (cols, row))) {
                if (is_first) {
                    first_row = row;
                    row_problems[row] += description + " is duplicated";
                    is_first = false;
                } else {
                    string id_label;
                    if (id_col) {
                        id_col->GetData().GetId()[first_row]->GetLabel(&id_label);
                    } else {
                        id_label = NStr::NumericToString(first_row + 1);
                    }
                    row_problems[row] += description + " duplicates row " + id_label;
                }
                num_duplicates++;
            }
        }
        sit++;
    }
    return num_duplicates;
}


void CSourceRequirements::x_AddOneOfProblems(CRef<objects::CSeq_table> values_table, vector<string> one_of_list, CRef<objects::CSeqTable_column> problems)
{
    if (one_of_list.size() < 1) {
        return;
    }
    size_t j;
    string description = "Missing " + one_of_list[0];
    for (j = 1; j < one_of_list.size(); j++) {
       description += " or " + one_of_list[j];
    }

    vector<CRef<objects::CSeqTable_column> > cols;
    for (j = 0; j < one_of_list.size(); j++) {
        CRef<objects::CSeqTable_column> col = FindSeqTableColumnByName(values_table, one_of_list[j]);
        if (col) {
            cols.push_back(col);
        }
    }

    int row;
    for (row = 0; row < values_table->GetNum_rows(); row++) {
        bool any = false;
        for (j = 0; j < cols.size() && !any; j++) {
            if (!NStr::IsBlank(cols[j]->GetData().GetString()[row])) {
                any = true;
            }
        }
        if (!any) {
            if (!NStr::IsBlank(problems->SetData().SetString()[row])) {
                problems->SetData().SetString()[row] += ", ";
            }
            problems->SetData().SetString()[row] += description;
        }
    }
}


CRef<objects::CSeqTable_column> CSourceRequirements::CheckSourceQuals(CRef<objects::CSeq_table> values_table)
{
    string qual_report = "";
    int row;
    CRef<objects::CSeqTable_column> problems (new objects::CSeqTable_column());
    problems->SetHeader().SetTitle(kProblems);
    for (row = 0; row < values_table->GetNum_rows(); row++) {
        problems->SetData().SetString().push_back("");
    }

    // check for missing or invalid values
    ITERATE (TSrcQualList, rit, m_Requirements) {
        CRef<objects::CSeqTable_column> col = FindSeqTableColumnByName (values_table, (*rit)->GetName());
        if (col) {
            // found value column, now validate
            for (row = 0; row < values_table->GetNum_rows(); row++) {
                string val = "";
                if (col->GetData().GetSize() > row) {
                    val = col->GetData().GetString()[row];
                }
                if ((*rit)->IsRequired() && NStr::IsBlank(val)) {
                    if (!NStr::IsBlank(problems->SetData().SetString()[row])) {
                        problems->SetData().SetString()[row] += ", ";
                    }
                    problems->SetData().SetString()[row] += (*rit)->GetName() + " is missing";
                } else {
                    string tmp = (*rit)->GetFormatErrors(val);
                    if (!NStr::IsBlank(tmp)) {
                        if (!NStr::IsBlank(problems->SetData().SetString()[row])) {
                            problems->SetData().SetString()[row] += ", ";
                        }
                        problems->SetData().SetString()[row] += tmp;
                    }
                }
            }
        } else {
            for (row = 0; row < values_table->GetNum_rows(); row++) {
                if (!NStr::IsBlank(problems->SetData().SetString()[row])) {
                    problems->SetData().SetString()[row] += ", ";
                }
                problems->SetData().SetString()[row] += (*rit)->GetName() + " is missing";
            }
        }
    }

    size_t i;

    // check lists where at least one qualifier in the list is required
    for (i = 0; i < m_OneOfLists.size(); i++) {
        x_AddOneOfProblems(values_table, m_OneOfLists[i], problems);
    }

    // if any uniqueness lists exist, at least one must be satisfied
    if (values_table->GetNum_rows() > 1) {
        bool found_good_combo = false;
        int best_num = 0;
        vector<string > uniqueness_problems;
        vector<vector<int> > num_problems;
        for (i = 0; i < m_UniquenessLists.size() && !found_good_combo; i++) {
            vector<string> this_list;
            for (row = 0; row < values_table->GetNum_rows(); row++) {
                this_list.push_back("");
            }
            int num_problems = x_AddUniquenessProblems(values_table, m_UniquenessLists[i], this_list);
            if (num_problems > 0) {
                if (best_num == 0 || best_num > num_problems) {
                    uniqueness_problems = this_list;
                    best_num = num_problems;
                }
            } else {
                found_good_combo = true;
            }
        }
        if (!found_good_combo) {
            for (i = 0; i < uniqueness_problems.size(); i++) {
                if (!NStr::IsBlank(uniqueness_problems[i])) {
                    if (!NStr::IsBlank(problems->GetData().GetString()[i])) {
                        problems->SetData().SetString()[i] += ", ";
                    }
                    problems->SetData().SetString()[i] += uniqueness_problems[i];
                }
            }
        }                
    }

    return problems;
}


void CSourceRequirements::PreferentiallyAddRequirement(CRef<objects::CSeq_table> values_table, string choice1, string choice2, bool required)
{
    CRef<objects::CSeqTable_column> col1 = FindSeqTableColumnByName (values_table, choice1);
    CRef<objects::CSeqTable_column> col2 = FindSeqTableColumnByName (values_table, choice2);

    if (!col2) {
        AddRequirement (choice1, required);
    } else if (!col1) {
        AddRequirement (choice2, required);
    } else {
        bool any1 = false, any2 = false;
        for (int row = 0; row < values_table->GetNum_rows() && !any1; row++) {
            if (!NStr::IsBlank(col1->GetData().GetString()[row])) {
                any1 = true;
            } else if (!NStr::IsBlank(col2->GetData().GetString()[row])) {
                any2 = true;
            }
        }
        if (any1 || !any2) {
            AddRequirement(choice1, required);
        } else {
            AddRequirement(choice2, required);
        }
    }
}


void CSourceRequirements::AddColumnsToSeqTable(CRef<objects::CSeq_table> values_table)
{
    if (!values_table) {
        return;
    }

    NON_CONST_ITERATE (TSrcQualList, rit, m_Requirements) {
        string val_name = (*rit)->GetName();
        string example = (*rit)->GetExample();
        if (NStr::IsBlank(example)) {
            val_name += "\n";
        } else {
            val_name += "\n[" + example + "]";
        }
        CRef<objects::CSeqTable_column> col = FindSeqTableColumnByName (values_table, val_name);
        if (col) {
            col->SetHeader().SetTitle(val_name);
        } else {
            CRef<objects::CSeqTable_column> new_col(new objects::CSeqTable_column());
            new_col->SetHeader().SetTitle(val_name);
            while (new_col->SetData().SetString().size() < values_table->GetNum_rows()) {
                new_col->SetData().SetString().push_back ("");
            }
            values_table->SetColumns().push_back(new_col);
        }
    }
}


typedef struct exampletable {
  string field_name;
  int wizard_type;
  int src_type;
  string example;
} ExampleTableData;


static const char* kHost = "Nat-host";

static const ExampleTableData s_ExampleTable[] = {
  { "organism name", CSourceRequirements::eWizardType_viruses,  CSourceRequirements::eWizardSrcType_virus_influenza, "Influenza A virus" } ,
  { "organism name", CSourceRequirements::eWizardType_viruses,  CSourceRequirements::eWizardSrcType_virus_norovirus, "Norovirus" } ,
  { "organism name", CSourceRequirements::eWizardType_viruses,  CSourceRequirements::eWizardSrcType_virus_rotavirus, "Rotavirus A" } ,
  { "organism name", CSourceRequirements::eWizardType_viruses,  CSourceRequirements::eWizardSrcType_virus_foot_and_mouth, "Foot-and-mouth disease virus - type O" } ,
  { "organism name", CSourceRequirements::eWizardType_viruses,  -1, "Tula virus" } ,
  { "organism name", CSourceRequirements::eWizardType_rrna_its_igs, CSourceRequirements::eWizardSrcType_bacteria_or_archaea, "Bacillus cereus" } ,
  { "organism name", -1, CSourceRequirements::eWizardSrcType_cultured_fungus, "Morchella esculenta" } ,
  { "organism name", -1, CSourceRequirements::eWizardSrcType_vouchered_fungus, "Morchella esculenta" } ,
  { "organism name", CSourceRequirements::eWizardType_rrna_its_igs, -1, "Taxodium distichum" } ,
  { "organism name", CSourceRequirements::eWizardType_igs, -1, "Taxodium distichum" } ,
  { "organism name", CSourceRequirements::eWizardType_tsa, -1, "Homo sapiens" } ,
  { "organism name", CSourceRequirements::eWizardType_d_loop, -1, "Coffea arabica" } ,
  { "organism name", CSourceRequirements::eWizardType_microsatellite, -1, "Coffea arabica" } ,
  { "Collection-date", -1, -1, "05-Feb-2005" } ,
  { "Collection-date", CSourceRequirements::eWizardType_viruses, -1, "01-Jan-2011" } ,
  { kHost, CSourceRequirements::eWizardType_viruses, -1, "Microtus arvalis" } ,
  { "Serotype", CSourceRequirements::eWizardType_viruses, CSourceRequirements::eWizardSrcType_virus_foot_and_mouth, "O" } ,
  { "Collection-date", CSourceRequirements::eWizardType_viruses, CSourceRequirements::eWizardSrcType_virus_foot_and_mouth, "05-Nov-2007" } ,
  { kHost, CSourceRequirements::eWizardType_viruses, CSourceRequirements::eWizardSrcType_virus_foot_and_mouth, "Bos taurus" } ,
  { "Country", CSourceRequirements::eWizardType_viruses, CSourceRequirements::eWizardSrcType_virus_foot_and_mouth, "Iran" } ,
  { "Serotype", CSourceRequirements::eWizardType_viruses, CSourceRequirements::eWizardSrcType_virus_influenza, "H1N1" } ,
  { "Collection-date", CSourceRequirements::eWizardType_viruses, CSourceRequirements::eWizardSrcType_virus_influenza, "02-Feb-2011" } ,
  { kHost, CSourceRequirements::eWizardType_viruses, CSourceRequirements::eWizardSrcType_virus_influenza, "Homo sapiens; male" } ,
  { "Country", CSourceRequirements::eWizardType_viruses, CSourceRequirements::eWizardSrcType_virus_influenza, "Italy" } ,
  { "segment", CSourceRequirements::eWizardType_viruses, CSourceRequirements::eWizardSrcType_virus_influenza, "4" } ,
  { "Passage History", CSourceRequirements::eWizardType_viruses, CSourceRequirements::eWizardSrcType_virus_influenza, "E1" } ,
  { "organism", CSourceRequirements::eWizardType_viruses, CSourceRequirements::eWizardSrcType_virus_influenza, "Influenza A virus" } ,
  { "strain", CSourceRequirements::eWizardType_viruses, CSourceRequirements::eWizardSrcType_virus_influenza, "A/Milan/2a18/2011" } ,
  { "isolate", CSourceRequirements::eWizardType_viruses, CSourceRequirements::eWizardSrcType_virus_influenza, "A/Milan/2a18/2011" } ,
  { "Collection-date", CSourceRequirements::eWizardType_viruses, CSourceRequirements::eWizardSrcType_virus_norovirus, "09-Jan-2003" } ,
  { "Country", CSourceRequirements::eWizardType_viruses, CSourceRequirements::eWizardSrcType_virus_norovirus, "Brazil" } ,
  { kHost, CSourceRequirements::eWizardType_viruses, CSourceRequirements::eWizardSrcType_virus_norovirus, "Homo sapiens" } ,
  { "Genotype", CSourceRequirements::eWizardType_viruses, CSourceRequirements::eWizardSrcType_virus_norovirus, "G1" } ,
  { "Collection-date", CSourceRequirements::eWizardType_viruses, CSourceRequirements::eWizardSrcType_virus_rotavirus, "15-Jul-2008" } ,
  { "Country", CSourceRequirements::eWizardType_viruses, CSourceRequirements::eWizardSrcType_virus_rotavirus, "USA: Idaho" } ,
  { kHost, CSourceRequirements::eWizardType_viruses, CSourceRequirements::eWizardSrcType_virus_rotavirus, "Bos taurus" } ,
  { "Genotype", CSourceRequirements::eWizardType_viruses, CSourceRequirements::eWizardSrcType_virus_rotavirus, "G1" } ,
  { "Specimen-voucher", CSourceRequirements::eWizardType_rrna_its_igs, CSourceRequirements::eWizardSrcType_vouchered_fungus, "AMNH 000000" } ,
  { "Specimen-voucher", CSourceRequirements::eWizardType_igs, CSourceRequirements::eWizardSrcType_vouchered_fungus, "AMNH 000000" } ,
  { kHost, CSourceRequirements::eWizardType_uncultured_samples, -1, "Cocos nucifera" } ,
  { "serotype", CSourceRequirements::eWizardType_viruses, -1, "Ex1" } , 
  { "genotype", CSourceRequirements::eWizardType_viruses, -1, "D9" } ,
  { "segment", CSourceRequirements::eWizardType_viruses, -1, "10" } ,
  { "genotype", CSourceRequirements::eWizardType_viruses, CSourceRequirements::eWizardSrcType_virus_foot_and_mouth, "VII" } ,
  { "serotype", CSourceRequirements::eWizardType_viruses, CSourceRequirements::eWizardSrcType_virus_norovirus, "1" } ,
  { "serotype", CSourceRequirements::eWizardType_viruses, CSourceRequirements::eWizardSrcType_virus_rotavirus, "Ex2a" } ,
  { "segment", CSourceRequirements::eWizardType_viruses, CSourceRequirements::eWizardSrcType_virus_rotavirus, "10" } ,
  { kHost, CSourceRequirements::eWizardType_igs, CSourceRequirements::eWizardSrcType_bacteria_or_archaea, "Nepenthes sp." } ,
  { kHost, CSourceRequirements::eWizardType_igs, CSourceRequirements::eWizardSrcType_cultured_fungus, "Ulmus sp." } ,
  { kHost, CSourceRequirements::eWizardType_igs, CSourceRequirements::eWizardSrcType_vouchered_fungus, "Ulmus sp." } ,
  { kHost, CSourceRequirements::eWizardType_rrna_its_igs, CSourceRequirements::eWizardSrcType_cultured_fungus, "Ulmus sp." } ,
  { kHost, CSourceRequirements::eWizardType_rrna_its_igs, CSourceRequirements::eWizardSrcType_vouchered_fungus, "Ulmus sp." } ,
  { "dev-stage", CSourceRequirements::eWizardType_tsa, -1, "seed" } ,
  { "cell-line", CSourceRequirements::eWizardType_tsa, -1, "HK234" } ,
  { "cell-type", CSourceRequirements::eWizardType_tsa, -1, "leukocyte" } ,
  { "cultivar", CSourceRequirements::eWizardType_tsa, -1, "Microtom" } ,
  { "tissue-type", CSourceRequirements::eWizardType_tsa, -1, "liver" } ,
  { "haplotype", CSourceRequirements::eWizardType_d_loop, -1, "A1" } ,
  { "specimen-voucher", CSourceRequirements::eWizardType_d_loop, -1, "A1" } ,
  { kHost, -1, -1, "Homo sapiens" } ,
  { "lat-lon", -1, -1, "39.00 N 77.10 W" } ,
  { "strain", -1, -1, "ABC123" } ,
  { "country", -1, -1, "USA: Ann Arbor, MI" } ,
  { "country", CSourceRequirements::eWizardType_viruses, -1, "Finland" } ,
  { "country", CSourceRequirements::eWizardType_viruses, CSourceRequirements::eWizardSrcType_virus_foot_and_mouth, "Iran" } ,
  { "country", CSourceRequirements::eWizardType_viruses, -1, "Finland" } ,
  { "Isolate", -1, -1, "SDZ123" } ,
  { "Isolate", CSourceRequirements::eWizardType_viruses, CSourceRequirements::eWizardSrcType_virus_foot_and_mouth, "IND19" } ,
  { "Isolate", CSourceRequirements::eWizardType_viruses, CSourceRequirements::eWizardSrcType_virus_influenza, "A/Milan/2a18/2011" } ,
  { "Isolate", CSourceRequirements::eWizardType_viruses, CSourceRequirements::eWizardSrcType_virus_norovirus, "Hu/G1/T65/BRA/2003" } ,
  { "Isolate", CSourceRequirements::eWizardType_viruses, CSourceRequirements::eWizardSrcType_virus_rotavirus, "cow/C1/USA/2008/G1" } ,
  { "Isolate", CSourceRequirements::eWizardType_rrna_its_igs, CSourceRequirements::eWizardSrcType_vouchered_fungus, "xyz1a" } ,
  { "Isolate", CSourceRequirements::eWizardType_rrna_its_igs, -1, "EX-A" } ,
  { "Isolate", CSourceRequirements::eWizardType_igs, CSourceRequirements::eWizardSrcType_vouchered_fungus, "xyz1a" } ,
  { "Isolate", CSourceRequirements::eWizardType_igs, -1, "EX-A" } ,
  { "Isolate", CSourceRequirements::eWizardType_d_loop, -1, "xyz1a" } ,
  { "Specimen-voucher", -1, -1, "USNM:12345" } ,
  { "sex", -1, -1, "female" } ,
  { "breed", -1, -1, "Holstein" } ,
  { "cultivar", -1, -1, "Granny Smith" } ,
  { "isolation-source", CSourceRequirements::eWizardType_uncultured_samples, -1, "diseased leaf" } ,
  { "isolation-source", CSourceRequirements::eWizardType_rrna_its_igs, CSourceRequirements::eWizardSrcType_bacteria_or_archaea, "leaf surface" } ,
  { "isolation-source", CSourceRequirements::eWizardType_rrna_its_igs, CSourceRequirements::eWizardSrcType_cultured_fungus, "soil under elm tree" } ,
  { "isolation-source", CSourceRequirements::eWizardType_rrna_its_igs, CSourceRequirements::eWizardSrcType_vouchered_fungus, "soil under elm tree" } ,
  { "isolation-source", CSourceRequirements::eWizardType_rrna_its_igs, -1, "lake shoreline" } ,
  { "isolation-source", CSourceRequirements::eWizardType_igs, CSourceRequirements::eWizardSrcType_bacteria_or_archaea, "leaf surface" } ,
  { "isolation-source", CSourceRequirements::eWizardType_igs, CSourceRequirements::eWizardSrcType_cultured_fungus, "soil under elm tree" } ,
  { "isolation-source", CSourceRequirements::eWizardType_igs, -1, "lake shoreline" } ,
  { "isolation-source", CSourceRequirements::eWizardType_viruses, CSourceRequirements::eWizardSrcType_virus_foot_and_mouth, "skin" } ,
  { "isolation-source", CSourceRequirements::eWizardType_viruses, CSourceRequirements::eWizardSrcType_virus_influenza, "nasal swab" } ,
  { "isolation-source", CSourceRequirements::eWizardType_viruses, -1, "feces" } ,
  { "isolation-source", -1, -1, "soil" } ,
  { "clone", CSourceRequirements::eWizardType_uncultured_samples, -1, "abc-1" } ,
  { "clone", CSourceRequirements::eWizardType_microsatellite, -1, "Ca-789" } ,
};


static const int k_NumExampleTableRows = sizeof (s_ExampleTable) / sizeof (ExampleTableData);


static bool IsExampleTableRowAcceptable (const ExampleTableData * e1, string field_name, CSourceRequirements::EWizardType wizard_type, CSourceRequirements::EWizardSrcType src_type)
{
    if (e1 == NULL) {
        return false;
    } 
    if (!NStr::EqualNocase(field_name, e1->field_name)) {
        return false;
    }
    if (e1->wizard_type != -1 && e1->wizard_type != wizard_type) {
        return false;
    }
    if (e1->src_type != -1 && e1->src_type != src_type) {
        return false;
    }
    return true;
}


static int CompareExampleTableRows (const ExampleTableData * e1, const ExampleTableData * e2)
{
    if (e1->wizard_type == -1 && e2->wizard_type != -1) {
        return -1;
    } else if (e1->wizard_type != -1 && e2->wizard_type == -1) {
        return 1;
    } else if (e1->src_type == -1 && e2->wizard_type != -1) {
        return -1;
    } else if (e1->src_type != -1 && e2->wizard_type == -1) {
        return 1;
    } else {
        return 0;
    }
}


void CSourceRequirements::SetExamples(CSourceRequirements::EWizardType wizard_type, CSourceRequirements::EWizardSrcType src_type)
{
    NON_CONST_ITERATE (TSrcQualList, rit, m_Requirements) {
        if (NStr::IsBlank((*rit)->GetExample())) {
            int best_k = -1;
            for (int k = 0; k < k_NumExampleTableRows; k++) {
                if (IsExampleTableRowAcceptable(s_ExampleTable + k, (*rit)->GetName(), wizard_type, src_type)
                    && (best_k == -1 || CompareExampleTableRows(s_ExampleTable + best_k, s_ExampleTable + k) < 0)) {
                    best_k = k;
                }
            }
            if (best_k > -1) {
                (*rit)->SetExample(s_ExampleTable[best_k].example);
            }
        }
    }

}


CSourceRequirements *GetSrcRequirements(CSourceRequirements::EWizardType wizard_type, 
                                        CSourceRequirements::EWizardSrcType source_type, 
                                        CRef<objects::CSeq_table> values_table)
{
    CSourceRequirements *requirements = new CSourceRequirements();
    if (wizard_type == CSourceRequirements::eWizardType_uncultured_samples) {
        requirements->AddRequirement("uncultured taxname", true, "uncultured bacterium");
    } else {
        requirements->AddRequirement("organism name", true);
    }

    switch (wizard_type) {
        case CSourceRequirements::eWizardType_uncultured_samples:
          {
              bool has_gelband_isolate = false;
              CRef<objects::CSeqTable_column> isolate_col = FindSeqTableColumnByName (values_table, "isolate");
              if (isolate_col) {
                  for (size_t row = 0; row < isolate_col->GetData().GetSize() && !has_gelband_isolate; row++) {
                      string isolate_val = isolate_col->GetData().GetString()[row];
                      if (NStr::StartsWith(isolate_val, "DGGE") || NStr::StartsWith(isolate_val, "TGGE")) {
                          has_gelband_isolate = true;
                      }
                  }
              }
              if (has_gelband_isolate) {
                  requirements->AddRequirement("isolate", true);
              } else {
                  requirements->AddRequirement("clone", true);
              }
              requirements->AddRequirement("isolation-source", true);
              break;
          }
        case CSourceRequirements::eWizardType_viruses:
            if (source_type == CSourceRequirements::eWizardSrcType_virus_influenza) {
                requirements->PreferentiallyAddRequirement (values_table, "strain", "isolate", true);
            } else {
                requirements->PreferentiallyAddRequirement (values_table, "isolate", "strain", true);
            }
            switch (source_type) {
                case CSourceRequirements::eWizardSrcType_virus_norovirus:
                  requirements->AddRequirement("collection-date", true);
                  requirements->AddRequirement("country", true);
                  requirements->AddRequirement("genotype", true);
                  requirements->AddOneOfList(2, kHost, "isolation-source");
                  break;
                case CSourceRequirements::eWizardSrcType_virus_foot_and_mouth:
                  requirements->AddRequirement("serotype", true);
                  requirements->AddRequirement("collection-date", true);
                  requirements->AddRequirement(kHost, true);
                  requirements->AddRequirement("country", true);
                  break;
                case CSourceRequirements::eWizardSrcType_virus_influenza:
                  requirements->AddRequirement("serotype", true);
                  requirements->AddRequirement("collection-date", true);
                  requirements->AddRequirement("country", true);
                  requirements->AddRequirement("segment", true);
                  requirements->AddOneOfList(2, kHost, "isolation-source");
                  break;
                case CSourceRequirements::eWizardSrcType_virus_rotavirus:
                  requirements->AddRequirement("collection-date", true);
                  requirements->AddRequirement("country", true);
                  requirements->AddRequirement("genotype", true);
                  requirements->AddOneOfList(2, kHost, "isolation-source");
                  break;
                default:
                  requirements->AddRequirement("country", true);
                  requirements->AddRequirement("collection-date", true);
                  requirements->AddRequirement(kHost, true);
                  break;
            }
            requirements->AddUniquenessList(1, "segment");
            requirements->AddUniquenessList(2, "segment", "isolate");
            requirements->AddUniquenessList(2, "segment", "strain");
            break;
        case CSourceRequirements::eWizardType_rrna_its_igs:
            switch (source_type) {
                case CSourceRequirements::eWizardSrcType_bacteria_or_archaea:
                    requirements->PreferentiallyAddRequirement (values_table, "strain", "isolate", true);
                    requirements->AddRequirement("isolation-source", true);
                    break;
                case CSourceRequirements::eWizardSrcType_cultured_fungus:
                    requirements->PreferentiallyAddRequirement (values_table, "strain", "isolate", true);
                    requirements->AddRequirement("isolation-source", true);
                    break;
                case CSourceRequirements::eWizardSrcType_vouchered_fungus:
                    requirements->AddRequirement("specimen-voucher", true);
                    break;
                default:
                    requirements->AddRequirement("isolate", true);
                    requirements->AddRequirement("isolation-source", true);
                    break;
            }
            break;
        case CSourceRequirements::eWizardType_igs:
            switch (source_type) {
                case CSourceRequirements::eWizardSrcType_cultured_fungus:
                    requirements->PreferentiallyAddRequirement (values_table, "strain", "isolate", true);
                    requirements->AddRequirement("isolation-source", true);
                    break;
                case CSourceRequirements::eWizardSrcType_vouchered_fungus:
                    requirements->AddRequirement("specimen-voucher", true);
                    requirements->AddUniquenessList(5, "organism", "specimen-voucher", "isolate", "bio-material", "culture-collection");
                    break;
                case CSourceRequirements::eWizardSrcType_plant:
                case CSourceRequirements::eWizardSrcType_animal:
                default:
                    requirements->AddRequirement("isolate", true);
                    requirements->AddRequirement("isolation-source", true);
                    requirements->AddRequirement("specimen-voucher", false);
                    requirements->AddRequirement("cultivar", false);
                    requirements->AddRequirement("bio-material", false);
                    requirements->AddRequirement("culture-collection", false);
                    requirements->AddUniquenessList(1, "organism");
                    requirements->AddUniquenessList(2, "organism", "isolate");
                    requirements->AddUniquenessList(2, "organism", "specimen-voucher");
                    requirements->AddUniquenessList(2, "organism", "cultivar");
                    requirements->AddUniquenessList(2, "organism", "bio-material");
                    requirements->AddUniquenessList(2, "organism", "culture-collection"); 
                    break;
            }
            break;
        case CSourceRequirements::eWizardType_d_loop:
            requirements->AddUniquenessList(1, "organism");
            requirements->AddUniquenessList(2, "organism", "isolate");
            requirements->AddUniquenessList(2, "organism", "specimen-voucher");
            requirements->AddUniquenessList(2, "organism", "haplotype");            
            break;
        default:
            break;
    }

    requirements->SetExamples(wizard_type, source_type);
    return requirements;
}


static string GetPrimerSetNameValues(const objects::CPCRPrimerSet& primer_set)
{
    string this_name = "";

    ITERATE (objects::CPCRPrimerSet::Tdata, sit, primer_set.Get()) {
        string new_name = "";
        if ((*sit)->IsSetName()) {
            new_name = (*sit)->GetName();
        }
        if (!NStr::IsBlank(this_name)) {
           this_name += ",";
        }
        this_name += new_name;
    }
    return this_name;
}


static string GetPrimerSetSeqValues(const objects::CPCRPrimerSet& primer_set)
{
    string this_seq = "";

    ITERATE (objects::CPCRPrimerSet::Tdata, sit, primer_set.Get()) {
        string new_seq = "";
        if ((*sit)->IsSetSeq()) {
            new_seq = (*sit)->GetSeq();
        }
        if (!NStr::IsBlank(this_seq)) {
           this_seq += ",";
        }
        this_seq += new_seq;
    }
    return this_seq;
}


static void RemoveLastCharacter(string& str)
{
    size_t len = str.length();
    if (len > 0) {
        str = str.substr(0, len - 1);
    }
}


CRef<objects::CSeq_table> GetSeqTableFromSeqEntry (objects::CSeq_entry_Handle seh)
{
    CRef<objects::CSeq_table> table(new objects::CSeq_table());
    CRef<objects::CSeqTable_column> id_col(new objects::CSeqTable_column());
    id_col->SetHeader().SetField_id(objects::CSeqTable_column_info::eField_id_location_id);
    id_col->SetHeader().SetTitle(kSequenceIdColLabel);
    table->SetColumns().push_back(id_col);
    CRef<objects::CSeqTable_column> taxname_col(new objects::CSeqTable_column());
    taxname_col->SetHeader().SetTitle("Organism Name");
    taxname_col->SetHeader().SetField_name("org.taxname");
    taxname_col->SetData().SetString();
    table->SetColumns().push_back(taxname_col);

    size_t row = 0;
    objects::CBioseq_CI b_iter(seh, objects::CSeq_inst::eMol_na);
    for ( ; b_iter ; ++b_iter ) {
        objects::CSeqdesc_CI it (*b_iter, objects::CSeqdesc::e_Source);
        CRef<objects::CSeq_id> id(new objects::CSeq_id());
        id->Assign (*(b_iter->GetSeqId()));
        id_col->SetData().SetId().push_back(id);
        if (it) {
            // populate taxname
            if (it->GetSource().IsSetTaxname()) {
                while (taxname_col->GetData().GetString().size() < row + 1) {
                    taxname_col->SetData().SetString().push_back("");
                }
                taxname_col->SetData().SetString()[row] = it->GetSource().GetTaxname();
            }
            // populate genome
            if (it->GetSource().IsSetGenome() 
                && it->GetSource().GetGenome() != objects::CBioSource::eGenome_unknown) {
                AddValueToTable (table, "genome", 
                                 objects::CBioSource::GetOrganelleByGenome(it->GetSource().GetGenome()), row);                
            }
                
            if (it->GetSource().IsSetSubtype()) {
                ITERATE (objects::CBioSource::TSubtype, sit, it->GetSource().GetSubtype()) {
                    string subtype_name = objects::CSubSource::GetSubtypeName((*sit)->GetSubtype());
                    if (NStr::EqualNocase(subtype_name, "note")) {
                        subtype_name = "SubSource Note";
                    }
                    string val = "";
                    if ((*sit)->IsSetName()) {
                        val = (*sit)->GetName();
                    }
                    if (NStr::IsBlank(val) && objects::CSubSource::NeedsNoText((*sit)->GetSubtype())) {
                        val = "true";
                    }
                    AddValueToTable(table, subtype_name, val, row);
                }
            }
            if (it->GetSource().IsSetOrgMod()) {
                ITERATE (objects::COrgName::TMod, sit, it->GetSource().GetOrgname().GetMod()) {
                    string subtype_name = objects::COrgMod::GetSubtypeName((*sit)->GetSubtype());
                    if (NStr::EqualNocase(subtype_name, "note")) {
                        subtype_name = "OrgMod Note";
                    }
                    AddValueToTable(table, subtype_name, (*sit)->GetSubname(), row);
                }
            }
            if (it->GetSource().IsSetPcr_primers()) {
                // list them all
                string fwd_name = "";
                string rev_name = "";
                string fwd_seq = "";
                string rev_seq = "";
                ITERATE (objects::CPCRReactionSet::Tdata, pit, it->GetSource().GetPcr_primers().Get()) {
                    string this_fwd_name = "";
                    string this_fwd_seq = "";
                    string this_rev_name = "";
                    string this_rev_seq = "";
                    if ((*pit)->IsSetForward()) {
                        this_fwd_name = GetPrimerSetNameValues((*pit)->GetForward());
                        this_fwd_seq = GetPrimerSetSeqValues((*pit)->GetForward());
                    }
                    if ((*pit)->IsSetReverse()) {
                        this_rev_name = GetPrimerSetNameValues((*pit)->GetReverse());
                        this_rev_seq = GetPrimerSetSeqValues((*pit)->GetReverse());
                    }
                    fwd_name += this_fwd_name + ";";
                    fwd_seq += this_fwd_seq + ";";
                    rev_name += this_rev_name + ";";
                    rev_seq += this_rev_seq + ";";
                }
                RemoveLastCharacter(fwd_name);
                RemoveLastCharacter(fwd_seq);
                RemoveLastCharacter(rev_name);
                RemoveLastCharacter(rev_seq);
                if (!NStr::IsBlank(fwd_name) || !NStr::IsBlank(fwd_seq) || !NStr::IsBlank(rev_name) || !NStr::IsBlank(rev_seq)) {
                    AddValueToTable(table, "fwd-primer-name", fwd_name, row);
                    AddValueToTable(table, "fwd-primer-seq", fwd_seq, row);
                    AddValueToTable(table, "rev-primer-name", rev_name, row);
                    AddValueToTable(table, "rev-primer-seq", rev_seq, row);
                }
            }
                
        }
        row++;

    }            

    table->SetNum_rows(row);
    return table;
}


CRef<objects::CSeq_table> GetSeqTableForSrcQualFromSeqEntry (objects::CSeq_entry_Handle seh, vector<string> qual_names)
{
    CRef<objects::CSeq_table> table(new objects::CSeq_table());
    CRef<objects::CSeqTable_column> id_col(new objects::CSeqTable_column());
    id_col->SetHeader().SetField_id(objects::CSeqTable_column_info::eField_id_location_id);
    id_col->SetHeader().SetTitle(kSequenceIdColLabel);
    table->SetColumns().push_back(id_col);

    vector< CRef<CSrcTableColumnBase> > handlers;

    ITERATE(vector<string>, q, qual_names) {
        CRef<objects::CSeqTable_column> col(new objects::CSeqTable_column());
        col->SetHeader().SetTitle(*q);
        handlers.push_back(CSrcTableColumnBaseFactory::Create(*col));
        table->SetColumns().push_back(col);
    }

    size_t row = 0;
    objects::CBioseq_CI b_iter(seh, objects::CSeq_inst::eMol_na);
    for ( ; b_iter ; ++b_iter ) {
        objects::CSeqdesc_CI it (*b_iter, objects::CSeqdesc::e_Source);
        CRef<objects::CSeq_id> id(new objects::CSeq_id());
        id->Assign (*(b_iter->GetSeqId()));
        id_col->SetData().SetId().push_back(id);
        for (size_t i = 0; i < handlers.size(); i++) {
            if (it) {
                table->SetColumns()[i + 1]->SetData().SetString().push_back(handlers[i]->GetFromBioSource(it->GetSource()));
            } else {
                table->SetColumns()[i + 1]->SetData().SetString().push_back("");
            }
        }
        row++;
    }            

    table->SetNum_rows(row);
    return table;
}


void CSrcTableOrganismNameColumn::AddToBioSource(
        objects::CBioSource & in_out_bioSource, const string & newValue )
{
    in_out_bioSource.SetOrg().SetTaxname(newValue);
}

void CSrcTableOrganismNameColumn::ClearInBioSource(
      objects::CBioSource & in_out_bioSource )
{
    in_out_bioSource.SetOrg().ResetTaxname();
}


string CSrcTableOrganismNameColumn::GetFromBioSource(
      const objects::CBioSource & in_out_bioSource )
{
    string val = "";
    if (in_out_bioSource.IsSetOrg() && in_out_bioSource.GetOrg().IsSetTaxname()) {
        val = in_out_bioSource.GetOrg().GetTaxname();
    }
    return val;
}


void CSrcTableGenomeColumn::AddToBioSource(
        objects::CBioSource & in_out_bioSource, const string & newValue )
{
    in_out_bioSource.SetGenome(objects::CBioSource::GetGenomeByOrganelle(newValue));
}

void CSrcTableGenomeColumn::ClearInBioSource(
      objects::CBioSource & in_out_bioSource )
{
    in_out_bioSource.ResetGenome();
}


string CSrcTableGenomeColumn::GetFromBioSource(
      const objects::CBioSource & in_out_bioSource )
{
    string val = "";
    if (in_out_bioSource.IsSetGenome()) {
        val = in_out_bioSource.GetOrganelleByGenome( in_out_bioSource.GetGenome());
    }
    return val;
}



void CSrcTableSubSourceColumn::AddToBioSource(
        objects::CBioSource & in_out_bioSource, const string & newValue )
{
    if (!NStr::IsBlank(newValue)) {
        if (objects::CSubSource::NeedsNoText(m_Subtype)) {
            CRef<objects::CSubSource> s(new objects::CSubSource(m_Subtype, " "));        
            in_out_bioSource.SetSubtype().push_back(s);
        } else {
            CRef<objects::CSubSource> s(new objects::CSubSource(m_Subtype, newValue));        
            in_out_bioSource.SetSubtype().push_back(s);
        }
    }
}

void CSrcTableSubSourceColumn::ClearInBioSource(
      objects::CBioSource & in_out_bioSource )
{
    if (in_out_bioSource.IsSetSubtype()) {
        objects::CBioSource::TSubtype::iterator it = in_out_bioSource.SetSubtype().begin();
        while (it != in_out_bioSource.SetSubtype().end()) {
            if ((*it)->GetSubtype() == m_Subtype) {
                it = in_out_bioSource.SetSubtype().erase(it);
            } else {
                ++it;
            }
        }
        if (in_out_bioSource.SetSubtype().empty()) {
            in_out_bioSource.ResetSubtype();
        }
    }
}


string CSrcTableSubSourceColumn::GetFromBioSource(
      const objects::CBioSource & in_out_bioSource )
{
    string val = "";
    if (in_out_bioSource.IsSetSubtype()) {
        objects::CBioSource::TSubtype::const_iterator it = in_out_bioSource.GetSubtype().begin();
        while (it != in_out_bioSource.GetSubtype().end()) {
            if ((*it)->GetSubtype() == m_Subtype && (*it)->IsSetName()) {
                val = (*it)->GetName();
                break;
            }
            ++it;
        }
    }
    return val;
}



void CSrcTableOrgModColumn::AddToBioSource(
        objects::CBioSource & in_out_bioSource, const string & newValue )
{
    if (!NStr::IsBlank(newValue)) {
        CRef<objects::COrgMod> s(new objects::COrgMod(m_Subtype, newValue));        
        in_out_bioSource.SetOrg().SetOrgname().SetMod().push_back(s);
    }
}

void CSrcTableOrgModColumn::ClearInBioSource(
      objects::CBioSource & in_out_bioSource )
{
    if (in_out_bioSource.IsSetOrg() && in_out_bioSource.GetOrg().IsSetOrgname() && in_out_bioSource.GetOrg().GetOrgname().IsSetMod()) {
        objects::COrgName::TMod::iterator it = in_out_bioSource.SetOrg().SetOrgname().SetMod().begin();
        while (it != in_out_bioSource.SetOrg().SetOrgname().SetMod().end()) {
            if ((*it)->GetSubtype() == m_Subtype) {
                it = in_out_bioSource.SetOrg().SetOrgname().SetMod().erase(it);
            } else {
                ++it;
            }
        }
    }
}


string CSrcTableOrgModColumn::GetFromBioSource(
      const objects::CBioSource & in_out_bioSource )
{
    string val = "";
    if (in_out_bioSource.IsSetOrg() && in_out_bioSource.GetOrg().IsSetOrgname() && in_out_bioSource.GetOrg().GetOrgname().IsSetMod()) {
        objects::COrgName::TMod::const_iterator it = in_out_bioSource.GetOrg().GetOrgname().GetMod().begin();
        while (it != in_out_bioSource.GetOrg().GetOrgname().GetMod().end()) {
            if ((*it)->GetSubtype() == m_Subtype && (*it)->IsSetSubname()) {
                val = (*it)->GetSubname();
                break;
            }
            ++it;
        }
    }
    return val;
}


void CSrcTableFwdPrimerSeqColumn::AddToBioSource(
        objects::CBioSource & in_out_bioSource, const string & newValue )
{
    objects::CPCRPrimerSeq seq(newValue);
    bool add_new_set = true;
    if (in_out_bioSource.IsSetPcr_primers() && in_out_bioSource.GetPcr_primers().IsSet()
        && in_out_bioSource.GetPcr_primers().Get().size() > 0) {
        // some sets exist, can we add to the last one?
        if (in_out_bioSource.GetPcr_primers().Get().back()->IsSetForward()
            && in_out_bioSource.GetPcr_primers().Get().back()->GetForward().IsSet()) {
            if (!in_out_bioSource.GetPcr_primers().Get().back()->GetForward().Get().front()->IsSetSeq()) {
                in_out_bioSource.SetPcr_primers().Set().back()->SetForward().Set().front()->SetSeq(seq);
                add_new_set = false;
            }
        } else {
            CRef<objects::CPCRPrimer> primer (new objects::CPCRPrimer());
            primer->SetSeq(seq);
            in_out_bioSource.SetPcr_primers().Set().back()->SetForward().Set().push_back(primer);
            add_new_set = false;
        }
    }

    if (add_new_set) {
        CRef<objects::CPCRPrimer> primer (new objects::CPCRPrimer());
        primer->SetSeq(seq);
        CRef<objects::CPCRReaction> reaction (new objects::CPCRReaction());
        reaction->SetForward().Set().push_back(primer);
        in_out_bioSource.SetPcr_primers().Set().push_back(reaction);
    }
}

void CSrcTableFwdPrimerSeqColumn::ClearInBioSource(
      objects::CBioSource & in_out_bioSource )
{
    in_out_bioSource.ResetPcr_primers();
}


string CSrcTableFwdPrimerSeqColumn::GetFromBioSource(
      const objects::CBioSource & in_out_bioSource )
{
    string val = "";
    if (in_out_bioSource.IsSetPcr_primers()
        && in_out_bioSource.GetPcr_primers().IsSet()
        && in_out_bioSource.GetPcr_primers().Get().size() > 0
        && in_out_bioSource.GetPcr_primers().Get().front()->IsSetForward()
        && in_out_bioSource.GetPcr_primers().Get().front()->GetForward().IsSet()
        && in_out_bioSource.GetPcr_primers().Get().front()->GetForward().Get().front()->IsSetSeq()) {
        val = in_out_bioSource.GetPcr_primers().Get().front()->GetForward().Get().front()->GetSeq();
    }
    return val;
}


void CSrcTableRevPrimerSeqColumn::AddToBioSource(
        objects::CBioSource & in_out_bioSource, const string & newValue )
{
    objects::CPCRPrimerSeq seq(newValue);
    bool add_new_set = true;
    if (in_out_bioSource.IsSetPcr_primers() && in_out_bioSource.GetPcr_primers().IsSet()
        && in_out_bioSource.GetPcr_primers().Get().size() > 0) {
        // some sets exist, can we add to the last one?
        if (in_out_bioSource.GetPcr_primers().Get().back()->IsSetReverse()
            && in_out_bioSource.GetPcr_primers().Get().back()->GetReverse().IsSet()) {
            if (!in_out_bioSource.GetPcr_primers().Get().back()->GetReverse().Get().front()->IsSetSeq()) {
                in_out_bioSource.SetPcr_primers().Set().back()->SetReverse().Set().front()->SetSeq(seq);
                add_new_set = false;
            }
        } else {
            CRef<objects::CPCRPrimer> primer (new objects::CPCRPrimer());
            primer->SetSeq(seq);
            in_out_bioSource.SetPcr_primers().Set().back()->SetReverse().Set().push_back(primer);
            add_new_set = false;
        }
    }

    if (add_new_set) {
        CRef<objects::CPCRPrimer> primer (new objects::CPCRPrimer());
        primer->SetSeq(seq);
        CRef<objects::CPCRReaction> reaction (new objects::CPCRReaction());
        reaction->SetReverse().Set().push_back(primer);
        in_out_bioSource.SetPcr_primers().Set().push_back(reaction);
    }
}


void CSrcTableRevPrimerSeqColumn::ClearInBioSource(
      objects::CBioSource & in_out_bioSource )
{
    in_out_bioSource.ResetPcr_primers();
}


string CSrcTableRevPrimerSeqColumn::GetFromBioSource(
      const objects::CBioSource & in_out_bioSource )
{
    string val = "";
    if (in_out_bioSource.IsSetPcr_primers()
        && in_out_bioSource.GetPcr_primers().IsSet()
        && in_out_bioSource.GetPcr_primers().Get().size() > 0
        && in_out_bioSource.GetPcr_primers().Get().front()->IsSetReverse()
        && in_out_bioSource.GetPcr_primers().Get().front()->GetReverse().IsSet()
        && in_out_bioSource.GetPcr_primers().Get().front()->GetReverse().Get().front()->IsSetSeq()) {
        val = in_out_bioSource.GetPcr_primers().Get().front()->GetReverse().Get().front()->GetSeq();
    }
    return val;
}


void CSrcTableFwdPrimerNameColumn::AddToBioSource(
        objects::CBioSource & in_out_bioSource, const string & newValue )
{
    objects::CPCRPrimerName name(newValue);
    bool add_new_set = true;

    if (in_out_bioSource.IsSetPcr_primers() && in_out_bioSource.GetPcr_primers().IsSet()
        && in_out_bioSource.GetPcr_primers().Get().size() > 0) {
        // some sets exist, can we add to the last one?
        if (in_out_bioSource.GetPcr_primers().Get().back()->IsSetForward()
            && in_out_bioSource.GetPcr_primers().Get().back()->GetForward().IsSet()) {
            if (!in_out_bioSource.GetPcr_primers().Get().back()->GetForward().Get().front()->IsSetName()) {
                in_out_bioSource.SetPcr_primers().Set().back()->SetForward().Set().front()->SetName(name);
                add_new_set = false;
            }
        } else {
            CRef<objects::CPCRPrimer> primer (new objects::CPCRPrimer());
            primer->SetName(name);
            in_out_bioSource.SetPcr_primers().Set().back()->SetForward().Set().push_back(primer);
            add_new_set = false;
        }
    }

    if (add_new_set) {
        CRef<objects::CPCRPrimer> primer (new objects::CPCRPrimer());
        primer->SetName(name);
        CRef<objects::CPCRReaction> reaction (new objects::CPCRReaction());
        reaction->SetForward().Set().push_back(primer);
        in_out_bioSource.SetPcr_primers().Set().push_back(reaction);
    }
}

void CSrcTableFwdPrimerNameColumn::ClearInBioSource(
      objects::CBioSource & in_out_bioSource )
{
    in_out_bioSource.ResetPcr_primers();
}


string CSrcTableFwdPrimerNameColumn::GetFromBioSource(
      const objects::CBioSource & in_out_bioSource )
{
    string val = "";
    if (in_out_bioSource.IsSetPcr_primers()
        && in_out_bioSource.GetPcr_primers().IsSet()
        && in_out_bioSource.GetPcr_primers().Get().size() > 0
        && in_out_bioSource.GetPcr_primers().Get().front()->IsSetForward()
        && in_out_bioSource.GetPcr_primers().Get().front()->GetForward().IsSet()
        && in_out_bioSource.GetPcr_primers().Get().front()->GetForward().Get().front()->IsSetName()) {
        val = in_out_bioSource.GetPcr_primers().Get().front()->GetForward().Get().front()->GetName();
    }
    return val;
}


void CSrcTableRevPrimerNameColumn::AddToBioSource(
        objects::CBioSource & in_out_bioSource, const string & newValue )
{
    objects::CPCRPrimerName name(newValue);
    bool add_new_set = true;

    if (in_out_bioSource.IsSetPcr_primers() && in_out_bioSource.GetPcr_primers().IsSet()
        && in_out_bioSource.GetPcr_primers().Get().size() > 0) {
        // some sets exist, can we add to the last one?
        if (in_out_bioSource.GetPcr_primers().Get().back()->IsSetReverse()
            && in_out_bioSource.GetPcr_primers().Get().back()->GetReverse().IsSet()) {
            if (!in_out_bioSource.GetPcr_primers().Get().back()->GetReverse().Get().front()->IsSetName()) {
                in_out_bioSource.SetPcr_primers().Set().back()->SetReverse().Set().front()->SetName(name);
                add_new_set = false;
            }
        } else {
            CRef<objects::CPCRPrimer> primer (new objects::CPCRPrimer());
            primer->SetName(name);
            in_out_bioSource.SetPcr_primers().Set().back()->SetReverse().Set().push_back(primer);
            add_new_set = false;
        }
    }

    if (add_new_set) {
        CRef<objects::CPCRPrimer> primer (new objects::CPCRPrimer());
        primer->SetName(name);
        CRef<objects::CPCRReaction> reaction (new objects::CPCRReaction());
        reaction->SetReverse().Set().push_back(primer);
        in_out_bioSource.SetPcr_primers().Set().push_back(reaction);
    }
}


void CSrcTableRevPrimerNameColumn::ClearInBioSource(
      objects::CBioSource & in_out_bioSource )
{
    in_out_bioSource.ResetPcr_primers();
}


string CSrcTableRevPrimerNameColumn::GetFromBioSource(
      const objects::CBioSource & in_out_bioSource )
{
    string val = "";
    if (in_out_bioSource.IsSetPcr_primers()
        && in_out_bioSource.GetPcr_primers().IsSet()
        && in_out_bioSource.GetPcr_primers().Get().size() > 0
        && in_out_bioSource.GetPcr_primers().Get().front()->IsSetReverse()
        && in_out_bioSource.GetPcr_primers().Get().front()->GetReverse().IsSet()
        && in_out_bioSource.GetPcr_primers().Get().front()->GetReverse().Get().front()->IsSetName()) {
        val = in_out_bioSource.GetPcr_primers().Get().front()->GetReverse().Get().front()->GetName();
    }
    return val;
}


bool IsFwdPrimerName (string name)
{
    return (NStr::EqualNocase(name, "fwd-primer-name") || NStr::EqualNocase(name, "fwd primer name")
        || NStr::EqualNocase(name, "fwd-name") || NStr::EqualNocase(name, "fwd name"));
}


bool IsRevPrimerName (string name)
{
    return (NStr::EqualNocase(name, "rev-primer-name") || NStr::EqualNocase(name, "rev primer name")
        || NStr::EqualNocase(name, "rev-name") || NStr::EqualNocase(name, "rev name"));
}


bool IsFwdPrimerSeq (string name)
{
    return (NStr::EqualNocase(name, "fwd-primer-seq") || NStr::EqualNocase(name, "fwd primer seq")
        || NStr::EqualNocase(name, "fwd-seq") || NStr::EqualNocase(name, "fwd seq"));
}


bool IsRevPrimerSeq (string name)
{
    return (NStr::EqualNocase(name, "rev-primer-seq") || NStr::EqualNocase(name, "rev primer seq")
        || NStr::EqualNocase(name, "rev-seq") || NStr::EqualNocase(name, "rev seq"));
}


CRef<CSrcTableColumnBase>
CSrcTableColumnBaseFactory::Create(const objects::CSeqTable_column &column)
{
    // the default is a CSrcTableEditCommandFactory that does nothing

    // extract the title of the column
    if( ! column.IsSetHeader() || ! column.GetHeader().IsSetTitle() ) {
        return CRef<CSrcTableColumnBase>( new CSrcTableColumnBase );
    }

    string sTitle = column.GetHeader().GetTitle();
    size_t pos = NStr::Find(sTitle, "\n");
    if (pos != string::npos) {
        sTitle = sTitle.substr(0, pos);
    }
    if( sTitle.empty() ) {
        return CRef<CSrcTableColumnBase>( new CSrcTableColumnBase );
    }

    if( sTitle == kSequenceIdColLabel ) {
        return CRef<CSrcTableColumnBase>( new CSrcTableColumnBase );
    }
    
    if( NStr::EqualNocase(sTitle, "Organism Name") || NStr::EqualNocase(sTitle, "org") || NStr::EqualNocase(sTitle, "taxname") ) {
        return CRef<CSrcTableColumnBase>( new CSrcTableOrganismNameColumn );
    }

    if (NStr::EqualNocase( sTitle, "genome" )) {
        return CRef<CSrcTableColumnBase>( new CSrcTableGenomeColumn );
    }

    if (NStr::EqualNocase(sTitle, "SubSource Note") ) {
        return CRef<CSrcTableColumnBase>( new CSrcTableSubSourceColumn (objects::CSubSource::eSubtype_other));
    }

    if (NStr::EqualNocase(sTitle, "OrgMod Note") ) {
        return CRef<CSrcTableColumnBase>( new CSrcTableOrgModColumn (objects::COrgMod::eSubtype_other));
    }

    if (IsFwdPrimerName(sTitle)) {
        return CRef<CSrcTableColumnBase>( new CSrcTableFwdPrimerNameColumn ());
    }

    if (IsFwdPrimerSeq(sTitle)) {
        return CRef<CSrcTableColumnBase>( new CSrcTableFwdPrimerSeqColumn ());
    }

    if (IsRevPrimerName(sTitle)) {
        return CRef<CSrcTableColumnBase>( new CSrcTableRevPrimerNameColumn ());
    }

    if (IsRevPrimerSeq(sTitle)) {
        return CRef<CSrcTableColumnBase>( new CSrcTableRevPrimerSeqColumn ());
    }

    // see if it's a SubSource subtype:
    try {
        objects::CSubSource::TSubtype st = objects::CSubSource::GetSubtypeValue (sTitle, objects::CSubSource::eVocabulary_insdc);
        return CRef<CSrcTableColumnBase>( new CSrcTableSubSourceColumn(st) );
    } catch (...) {
        try {
            objects::COrgMod::TSubtype st = objects::COrgMod::GetSubtypeValue (sTitle, objects::COrgMod::eVocabulary_insdc);
            return CRef<CSrcTableColumnBase>( new CSrcTableOrgModColumn(st) );
        } catch (...) {
        }
    }

    return CRef<CSrcTableColumnBase>( NULL );
}

CRef<CCmdComposite> ApplySrcTableToSeqEntry (CRef<objects::CSeq_table>values_table, objects::CSeq_entry_Handle seh)
{
    CRef<CCmdComposite> cmd( new CCmdComposite("Bulk Source Edit") );

    CRef<objects::CSeqTable_column> id_col = FindSeqTableColumnByName (values_table, kSequenceIdColLabel);

    if (!id_col) {
        return cmd;
    }

    if( ! values_table->IsSetColumns() || values_table->GetColumns().empty() ) {
        return cmd;
    }
    const objects::CSeq_table::TColumns & columns = values_table->GetColumns();
    size_t num_cols = columns.size();

    // create an edit-command maker for every column.  This somewhat byzantine
    // system is here to speed things up to avoid re-parsing the meaning
    // of each column every time.
    vector< CRef<CSrcTableColumnBase> > vecColEditFactories;

    ITERATE( objects::CSeq_table::TColumns, col_iter, columns ) {
        if (NStr::EqualNocase((*col_iter)->GetHeader().GetTitle(), kProblems)) {
            // don't create edit-command maker for problems column
            num_cols--;
        } else {
            vecColEditFactories.push_back( 
                CSrcTableColumnBaseFactory::Create(**col_iter) );
        }
    }

    for (int row = 0; row < values_table->GetNum_rows() && row < id_col->GetData().GetSize(); row++) {
        objects::CBioseq_Handle bsh = seh.GetBioseqHandle(*(id_col->GetData().GetId()[row]));

        CRef<objects::CSeqdesc> new_source_desc( new objects::CSeqdesc );
        objects::CBioSource & bioSource = new_source_desc->SetSource();

        // iterate through the columns, skipping Seq-id column, to erase current values
        // need to do this in separate steps, in case of multiple qualifier columns in table
        for( size_t col = 1; col < num_cols; ++col ) {
            vecColEditFactories[col]->ClearInBioSource(bioSource);
            // TODO: detect if change occurred to avoid excessive changes
        }

        // iterate through the columns, skipping Seq-id column
        for( size_t col = 1; col < num_cols; ++col ) {
            if (columns[col]->GetData().GetSize() > row) {
                vecColEditFactories[col]->AddToBioSource(
                    bioSource, *columns[col]->GetStringPtr(row) );
            }
            // TODO: detect if change occurred to avoid excessive changes
        }
        // automatically populate taxon ID, div, genetic codes if available
        CBioSourceAutoComplete::AutoFill(bioSource.SetOrg());

        objects::CSeqdesc_CI desc_ci( bsh, objects::CSeqdesc::e_Source);
        if (desc_ci) {
            CRef<CChangeSeqDescCommand> ecmd(new CChangeSeqDescCommand());
            CObject* actual = (CObject*) &(*desc_ci);
            ecmd->Add(actual, CConstRef<CObject>(new_source_desc));
            cmd->AddCommand (*ecmd);
        } else {
            objects::CBioseq_set_Handle bssh = bsh.GetParentBioseq_set();
            if (bssh && bssh.IsSetClass() && bssh.GetClass() == objects::CBioseq_set::eClass_nuc_prot) {
                cmd->AddCommand( *CRef<CCmdCreateDesc>(new CCmdCreateDesc(bssh.GetParentEntry(), *new_source_desc)) );
            } else {
                cmd->AddCommand( *CRef<CCmdCreateDesc>(new CCmdCreateDesc(bsh.GetParentEntry(), *new_source_desc)) );
            }
        }
    }

    // send composite command
    return cmd;
}


static bool RelaxedMatch (CRef<objects::CSeq_id> id1, CRef<objects::CSeq_id> id2)
{
    bool found = false;
    if (id1->IsLocal()) {
        string id1_label = "";
        id1->GetLabel(&id1_label, objects::CSeq_id::eContent);
        string id2_label = "";
        id2->GetLabel(&id2_label, objects::CSeq_id::eContent);
        id2->GetLabel(&id2_label, objects::CSeq_id::eContent);
        size_t pos = NStr::Find (id2_label, id1_label);
        if (pos == 0) {
            if (NStr::Equal(id2_label.substr(id1_label.length(), 1), ".")) {
                found = true;
            } 
        } else if (pos != string::npos && NStr::EndsWith(id2_label, id1_label)) {
            string delim = id2_label.substr(pos - 1, 1);
            if (NStr::Equal(delim, "|") || NStr::Equal(delim, "/") || NStr::Equal(delim, ":")) {
                found = true;
            }
        }
    }
    return found;
}


static bool OneRowOk (CRef<objects::CSeq_id> id, CRef<objects::CSeqTable_column> id_col, objects::CSeq_entry_Handle seh) 
{
    if (!id || !id_col) {
        return false;
    }

    size_t row = 0;
    bool found = false;
    while (row < id_col->GetData().GetSize() && !found) {
        CRef<objects::CSeq_id> row_id = id_col->GetData().GetId()[row];
        objects::CSeq_id::E_SIC compare = id->Compare(*row_id);
        if (compare == objects::CSeq_id::e_YES) {
            found = true;
        } else if (compare == objects::CSeq_id::e_DIFF) {
            if (RelaxedMatch(id, row_id)) {
                found = true;
                id->Assign(*row_id);
            }
        }
        row++;
    }
    row = 0;    
    while (row < id_col->GetData().GetSize() && !found) {
        CRef<objects::CSeq_id> row_id = id_col->GetData().GetId()[row];
        objects::CBioseq_Handle bsh = seh.GetBioseqHandle (*row_id);
        if (bsh) {
            CConstRef<objects::CBioseq> b = bsh.GetCompleteBioseq();
            ITERATE (objects::CBioseq::TId, id_it, b->GetId()) {
                objects::CSeq_id::E_SIC compare = id->Compare(**id_it);
                if (compare == objects::CSeq_id::e_YES) {
                    found = true;
                    id->Assign(*row_id);
                    break;
                } else if (compare == objects::CSeq_id::e_DIFF) {
                    if (RelaxedMatch(id, *id_it)) {
                        found = true;
                        id->Assign(*row_id);
                        break;
                    }
                }
            }
        }
        row++;
    }

    return found;           
}



END_NCBI_SCOPE
