/*  $Id: subprep_view.cpp 27556 2013-03-05 16:04:52Z katargir $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Roman Katargin
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>

#include <gui/packages/pkg_sequence_edit/subprep_view.hpp>

#include <gui/core/project_service.hpp>
#include <gui/core/document.hpp>

#include <gui/framework/workbench.hpp>

#include <gui/packages/pkg_sequence_edit/subprep_panel.hpp>

#include <gui/widgets/wx/fileartprov.hpp>
#include <gui/widgets/wx/wx_utils.hpp>
#include <gui/widgets/wx/message_box.hpp>

#include <objects/submit/Seq_submit.hpp>

#include "subprep_event.hpp"

#include <wx/msgdlg.h>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);


CProjectViewTypeDescriptor s_SubPrepViewTypeDescr(
    "Submission Preparation View", // type name
    "subprep_view", // icon alias
    "Assistant for preparing sequences for submission to GenBank",
    "The Text View shows data in a variety of formats, including GenBank FlatFile, "
    "NCBI ASN.1, and FastA format.",
    "SUBPREP_VIEW", // help ID
    "Generic",   // category
    false,      // not a singleton
    "SerialObject",
    eOneObjectAccepted
);

CSubPrepView::CSubPrepView()
  : CProjectView(), m_Window(0)
{
}

CSubPrepView::~CSubPrepView()
{
}

BEGIN_EVENT_MAP(CSubPrepView, CProjectView)
  ON_EVENT(CSubPrepEvent, CSubPrepEvent::eWrapSeq_submit,
           &CSubPrepView::x_OnWrapSeq_submit)
END_EVENT_MAP()

wxWindow* CSubPrepView::GetWindow()
{
    _ASSERT(m_Window);
    return m_Window;
}

wxEvtHandler* CSubPrepView::GetCommandHandler()
{
    CSubPrep_panel* panel = (CSubPrep_panel*) m_Window;
    _ASSERT(panel);
    return panel->GetEventHandler();
}

// override function to suppress Navigation toolbar
void CSubPrepView::GetCompatibleToolBars(vector<string>& names)
{
    names.push_back(CProjectViewToolBatFactory::sm_SelTBName);
}


bool CSubPrepView::InitView(TConstScopedObjects& objects, const objects::CUser_object*)
{
    CScope* scope = NULL;
    const CObject* object = NULL;

    if(objects.size() == 1) {
        object = objects[0].object.GetPointer();
        scope = objects[0].scope.GetPointer();
    }

    if(object)  {
        CIRef<CProjectService> srv = m_Workbench->GetServiceByType<CProjectService>();

        _ASSERT(srv);

        CRef<CWorkspaceConstPrx> workspace = srv->GetWorkspaceConst();
        CRef<CProjectPrx> project = workspace->FindProjectByScope(scope);
        _ASSERT(project);

        if (project) {
            CSubPrep_panel* panel = (CSubPrep_panel*) m_Window;
            panel->SetUndoManager(&project->GetDocument()->GetUndoManager());
            panel->SetWorkbench(m_Workbench);
            panel->SetObjects(&objects);
            m_OrigObj.Reset(objects[0].object.GetPointer());
            panel->TransferDataToWindow();

            x_AttachToProject(*project);
            x_UpdateContentLabel();
            return true;
        }
        return false;
    } else {
        // cannot represent the data
        x_ReportInvalidInputData(objects);
        return false;
    }
}

void CSubPrepView::OnProjectChanged(CProjectViewEvent& evt)
{
    if (evt.GetSubtype() == CProjectViewEvent::eData) {
        CSubPrep_panel* panel = (CSubPrep_panel*) m_Window;
        panel->OnDataChange();
    }
}


void CSubPrepView::CreateViewWindow(wxWindow* parent)
{
    _ASSERT(!m_Window);

    CSubPrep_panel* panel = new CSubPrep_panel(parent);
    m_Window = panel;
    panel->AddListener(this, ePool_Parent);
}


void CSubPrepView::DestroyViewWindow()
{
    if( m_Window ){
        m_Window->Destroy();
        m_Window = 0;
    }
}


const CViewTypeDescriptor& CSubPrepView::GetTypeDescriptor() const
{
    return s_SubPrepViewTypeDescr;
}


void CSubPrepView::GetSelection(CSelectionEvent& evt) const
{
    if (x_HasProject()) {
        CSubPrep_panel* panel = (CSubPrep_panel*)m_Window;
    }
}


void CSubPrepView::GetSelection(TConstScopedObjects& objs) const
{
    if (m_Window  &&  x_HasProject()) {
        CScope* scope = x_GetScope();
        CSubPrep_panel* panel = (CSubPrep_panel*)m_Window;
    }
}


void CSubPrepView::GetMainObject(TConstScopedObjects& objs) const
{
    if (m_Window  &&  x_HasProject()) {
        const CObject* object = m_OrigObj.GetPointer();
        if (object) {
            CScope* scope = x_GetScope();
            objs.push_back(SConstScopedObject(*object, *scope));
        }
    }
}


void CSubPrepView::x_OnSetSelection (CSelectionEvent& evt)
{
// Doesn't work !!!!
//    if (!evt.HasObjectSelection())
//        return;

    if (m_Window  &&  x_HasProject()) {
        CSubPrep_panel* panel = (CSubPrep_panel*)m_Window;
    }
}

const CObject* CSubPrepView::x_GetOrigObject() const
{
    return m_OrigObj.GetPointer();
}

void CSubPrepView::x_OnWrapSeq_submit(CEvent* evt)
{
    CSubPrepEvent* subPrepEvt = dynamic_cast<CSubPrepEvent*>(evt);
    if (!subPrepEvt)
        return;

    CSubPrepEvent::TSeq_submitAttachment* attachment =
        dynamic_cast<CSubPrepEvent::TSeq_submitAttachment*>(subPrepEvt->GetAttachment());
    if (!attachment) return;

    CRef<CSeq_submit> seq_submit = attachment->GetData();
    if (!seq_submit) {
        _ASSERT(false);
        return;
    }

    const CSerialObject* so = dynamic_cast<const CSerialObject*>(m_OrigObj.GetPointer());
    if (dynamic_cast<const CSeq_submit*>(so)) {
        wxMessageBox(wxT("The object is already Seq-submit!"),
                     wxT("Submission Preparation Error"), wxOK | wxICON_ERROR);
        return;
    }

    CIRef<CProjectService> proj_srv = m_Workbench->GetServiceByType<CProjectService>();
    if (!proj_srv) {
        _ASSERT(false); // This shouldn't happen
        return;
    }
    CRef<CWorkspaceConstPrx> workspace = proj_srv->GetWorkspaceConst();
    CRef<CProjectPrx> project = workspace->FindProjectById(m_ProjectId);
    if (!project) {
        _ASSERT(false); // This shouldn't happen
        return;
    }

    m_OrigObj.Reset(seq_submit.GetPointer());

    CRef<CProjectFolderPrx> data_folder = project->GetDataFolder();
    CRef<CProjectItemPrx> project_item = data_folder->FindProjectItemByData(*so, true);
    if (project_item) {
        project_item->SetObject(*seq_submit);
    }

}

///////////////////////////////////////////////////////////////////////////////
/// CSubPrepViewFactory
string  CSubPrepViewFactory::GetExtensionIdentifier() const
{
    static string sid("subprep_view_factory");
    return sid;
}


string CSubPrepViewFactory::GetExtensionLabel() const
{
    static string slabel("SubPrep View Factory");
    return slabel;
}


void CSubPrepViewFactory::RegisterIconAliases(wxFileArtProvider& provider)
{
    string alias = GetViewTypeDescriptor().GetIconAlias();
    provider.RegisterFileAlias(ToWxString(alias), wxT("subprep_view.png"));
}

void CSubPrepViewFactory::RegisterCommands(CUICommandRegistry& cmd_reg, wxFileArtProvider& provider)
{
}

const CProjectViewTypeDescriptor& CSubPrepViewFactory::GetProjectViewTypeDescriptor() const
{
    return s_SubPrepViewTypeDescr;
}


IView* CSubPrepViewFactory::CreateInstance() const
{
    return new CSubPrepView();
}


IView* CSubPrepViewFactory::CreateInstanceByFingerprint(const TFingerprint&) const
{
    return NULL;
}

bool CSubPrepViewFactory::IsCompatibleWith(const CObject& object, objects::CScope& scope)
{
    const type_info& type = typeid(object);

    if( typeid(CSeq_entry) == type  || typeid(CSeq_annot) == type  ||
        typeid(CSeq_submit) == type || typeid(CSeq_id) == type) {
            return true;
    }

    return false;
}

int CSubPrepViewFactory::TestInputObjects(TConstScopedObjects& objects)
{
    bool found_good = false;
    bool found_bad = false;

    for(  size_t i = 0;  i < objects.size();  i++)  {
        if (IsCompatibleWith(*objects[i].object, *objects[i].scope)) {
            found_good = true;
        }

        found_bad = true;
    }

    if (found_good)
        return fCanShowSeparated | (found_bad ? fCanShowSome : fCanShowAll);

    return 0; // can show nothing
}


END_NCBI_SCOPE
