/*  $Id: glrenderpdf.cpp 27633 2013-03-15 13:42:04Z wuliangs $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Bob Falk
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>
#include <gui/print/glrenderpdf.hpp>
#include <gui/opengl/glutils.hpp>
#include <gui/opengl/glresmgr.hpp>

#include <gui/utils/matrix4.hpp>
#include <gui/utils/matrix3.hpp>

#include <gui/print/pdf.hpp>
#include "page_buffers.hpp"

BEGIN_NCBI_SCOPE

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

void CGlRenderPdf::Render(CGlPane& pane, CGlModel2D* model)
{
    if (model != NULL) {       
        model->RenderPDF(pane, m_Pdf);        
    }
}

void CGlRenderPdf::Vertex3f(GLfloat x, GLfloat y, GLfloat z)
{
    if (m_CurrentMode == GL_NONE) {
        LOG_POST(Error << "Vertex*() called without first calling CGlRender::Begin");
        return;
    }

    m_VertexBuffer.push_back(CVect3<float>(x,y,z) );

    // In pdf mode, break quads into triangles if smooth shading is indicated
    // in opengl, OR if smooth shading is requested for pdf (the default).  
    // In PDF, smooth shading also lets polygon vertices merge, which they don't
    // for the flat shading model
    if ((m_State->GetShadeModel() == GL_SMOOTH || 
         m_State->GetPdfShadeStyle() == CGlState::eGouraud) &&
         m_State->GetPolygonMode() != GL_LINE) {

            // Break after 4 new vertices have been added (ignore groups of 6 since
            // they are the previous triangles)
            if (m_CurrentMode == GL_QUADS && m_VertexBuffer.size() % 6 == 4) {
                size_t v1 = m_VertexBuffer.size()-4;
                size_t v3 = v1 + 2;

                if (m_ColorBuffer.size() > 0) {
                    x_SyncAttribs(m_ColorBuffer);
                    m_ColorBuffer.push_back(m_ColorBuffer[v1]);
                    m_ColorBuffer.push_back(m_ColorBuffer[v3]);
                }

                if (m_NormalBuffer.size() > 0) {
                    x_SyncAttribs(m_NormalBuffer);
                    m_NormalBuffer.push_back(m_NormalBuffer[v1]);
                    m_NormalBuffer.push_back(m_NormalBuffer[v3]);
                }

                if (m_TexCoord2DBuffer1.size() > 0) {
                    x_SyncAttribs(m_TexCoord2DBuffer1);
                    m_TexCoord2DBuffer1.push_back(m_TexCoord2DBuffer1[v1]);
                    m_TexCoord2DBuffer1.push_back(m_TexCoord2DBuffer1[v3]);
                }

                // Do vertices last since x_SyncAttribs syncs size of attribute
                // buffers to current size of this (m_VertexBuffer) buffer.
                m_VertexBuffer.push_back(m_VertexBuffer[v1]);
                m_VertexBuffer.push_back(m_VertexBuffer[v3]);
            }
    }
}


void CGlRenderPdf::Rectf(GLfloat x1, GLfloat y1, GLfloat x2, GLfloat y2)
{
    if (m_CurrentMode != GL_NONE) {
        LOG_POST(Error << "Rect*() called between Begin() and End()");
        return;
    }   

    CVect3<float> scale(1.0f, 1.0f, 1.0f);
    m_QuadRenderNode.SetScale(scale);

    // convert to triangles unless we are polygon mode == gl_line, in which case do lines.
    // Thi is because triangles in line mode would show an extra line across the middle.
    if (m_State->GetPolygonMode() == GL_LINE) {
        Begin(GL_LINE_LOOP);
            Vertex3f(x1, y1, 0.0f);
            Vertex3f(x2, y1, 0.0f);
            Vertex3f(x2, y2, 0.0f);
            Vertex3f(x1, y2, 0.0f);
        End();
        return;
    } else if (m_QuadRenderNode.GetVBOGeom().IsNull() || 
               m_QuadRenderNode.GetVBOGeom()->GetVertexBuffer3D().GetActiveBufferSize() == 0) {
               m_QuadRenderNode.SetVBOGeom(CRef<CGlVboGeom>(CGlResMgr::Instance().GetVboGeom("RenderManager Quad Geom")));
               m_QuadRenderNode.SetDrawMode(GL_QUADS);      

        vector<CVect3<float> > verts;
        verts.push_back(CVect3<float>(0.0f, 0.0f, 0.0f));
        verts.push_back(CVect3<float>(1.0f, 0.0f, 0.0f));
        verts.push_back(CVect3<float>(1.0f, 1.0f, 0.0f));
        verts.push_back(CVect3<float>(0.0f, 1.0f, 0.0f));

        m_QuadRenderNode.GetVertexBuffer3D().BufferData(verts.size(), &verts[0], GL_STATIC_DRAW);
    }

    if (m_State->GetScaleInvarient()) {
        CVect2<TModelUnit> s = m_State->GetScaleFactor();

        CVect3<float> scale(float(s.X()), float(s.Y()), 1.0f);                          
        m_RenderNode.SetScale(scale);
    }
    
    /*
    CMatrix4<float>  mat(m_ModelView);
    // scale and translate the buffered quad ( (0,0)..(1,1) ) to
    // the size ((x1, y1)..(x2,y2))
    CMatrix4<float>  mat2(float(x2-x1), 0.0f, 0.0f, x1,
                          0.0f, float(y2-y1), 0.0f, y1,
                          0.0f, 0.0f, 1.0f, 0.0f,
                          0.0f, 0.0f, 0.0f, 1.0f);    
    mat *= mat2;

    // Using the matrix stack to compute the same result can be done like this:
    //glMatrixMode(GL_MODELVIEW);
    //glPushMatrix();
    //glTranslatef(x1, y1, 0.0f);
    //glScalef(float(x2-x1), float(y2-y1), 1.0f);
    //glGetFloatv(GL_MODELVIEW_MATRIX, m_ModelView.GetData());
    //glPopMatrix();
    */
    // Same as above code, but in 1 step for efficiency:
    float sx = float(x2-x1);
    float sy = float(y2-y1);
                         // row1
    CMatrix4<float>  mat(m_ModelView[0]*sx, m_ModelView[1]*sy, m_ModelView[2], 
                         m_ModelView[0]*x1 + m_ModelView[1]*y1 + m_ModelView[3],
                         // row2
                         m_ModelView[4]*sx, m_ModelView[5]*sy, m_ModelView[6], 
                         m_ModelView[4]*x1 + m_ModelView[5]*y1 + m_ModelView[7],
                         // row3
                         m_ModelView[8]*sx, m_ModelView[9]*sy, m_ModelView[10], 
                         m_ModelView[8]*x1 + m_ModelView[9]*y1 + m_ModelView[11],
                         // row4
                         0.0f, 0.0f, 0.0f, 1.0f);



    m_QuadRenderNode.SetPosition(mat);
    
    x_RenderBuffer(&m_QuadRenderNode);
}

void CGlRenderPdf::x_RenderBuffer(CGlVboNode* node)
{
    TVPRect viewport(m_Viewport[0], m_Viewport[1], m_Viewport[2], m_Viewport[3]);
    m_Pdf->PrintBuffer(node, viewport);
}
 
void CGlRenderPdf::BeginText(const CGlTextureFont* font, 
                             const CRgbaColor& color)
{
    if (m_CurrentFont != NULL) {
        LOG_POST("BeginText() called without first calling CGlRenderPdf::EndText()");
        return;
    }

    m_CurrentFont = font;
    m_TextColor = color;

    CRef<CPdfObject> content = m_Pdf->BeginContent(CPdf::ePdfText); 
}
void CGlRenderPdf::BeginText(const CGlTextureFont* font)
{
    if (m_CurrentFont != NULL) {
        LOG_POST("BeginText() called without first calling CGlRenderPdf::EndText()");
        return;
    }

    m_CurrentFont = font;
    m_TextColor = m_RenderNode.GetState().GetColor();

    CRef<CPdfObject> content = m_Pdf->BeginContent(CPdf::ePdfText); 
}

void CGlRenderPdf::EndText()
{
    if (m_CurrentFont == NULL) {
        LOG_POST("EndText() called without first calling CGlRenderPdf::BeginText()");
        return;
    }

    m_CurrentFont = NULL;
    m_Pdf->EndContent();
}

void CGlRenderPdf::WriteText(TModelUnit x, TModelUnit y, 
                                    const char* text,
                                    TModelUnit rotate_degrees)
{
    if (m_CurrentFont == NULL) {
        LOG_POST("Unable to write text - must call BeginText() first");
        return;
    }
 
    CVect2<float> posp;                    
    GLdouble px, py;
    GLdouble dummyz;

    GLdouble modelview_matrix[16];
    GLdouble projection_matrix[16];

    glGetDoublev(GL_MODELVIEW_MATRIX, modelview_matrix);
    glGetDoublev(GL_PROJECTION_MATRIX, projection_matrix);

    // Use viewport from the rendermanager object since OpenGL viewport maximum
    // size is below the max viewable area for a pdf page (and so
    // the OpenGL value will in that case be truncated).
    gluProject(GLdouble(x), GLdouble(y), GLdouble(0.0), 
        modelview_matrix, projection_matrix, m_Viewport, 
        &px, &py, &dummyz);
    posp.X() = float(px);
    posp.Y() = float(py);

    CGlTextureFont::EFontFace face = m_CurrentFont->GetFontFace();
    unsigned int font_size = m_CurrentFont->GetFontSize();

    m_Pdf->PrintText(m_Pdf->GetFontHandler(), face, float(font_size), posp, text, m_TextColor);
}

void CGlRenderPdf::WriteText(TModelUnit x, TModelUnit y, 
                                    TModelUnit width, TModelUnit height,
                                    const char* text,
                                    CGlTextureFont::TAlign align,
                                    CGlTextureFont::ETruncate trunc,
                                    TModelUnit rotate_degrees)
{
    if (m_CurrentFont == NULL) {
        LOG_POST("Unable to write text - must call BeginText() first");
        return;
    }



    CGlTextureFont::EFontFace face = m_CurrentFont->GetFontFace();
    unsigned int font_size = m_CurrentFont->GetFontSize();
   
    m_Pdf->GetCurrentContent()->SetFillColor(m_TextColor);    
    string truncated_text = text;
    CMatrix4<double> mat;
    
    // Get viewport from the rendermanager object since OpenGL viewport maximum
    // size is below the max viewable area for a pdf page (and so
    // the OpenGL value will in that case be truncated).
    TVPRect viewport(m_Viewport[0], m_Viewport[1], m_Viewport[2], m_Viewport[3]);  
    m_CurrentFont->BeginText(viewport);
    // Get transformation matrix needed for text at the desired location and rotation
    mat = m_CurrentFont->GetTextXform(x, y, 
                                      width, height, 
                                      truncated_text, align, trunc, rotate_degrees);
    m_CurrentFont->EndText();

    // returned mat is 4x4 row-order matrix.  Create a 3x3 row matrix (homgeneous x,y
    // coordinates) for transforming text
    CMatrix3<double> mat33(mat(0,0), mat(0,1), mat(0,3),
                           mat(1,0), mat(1,1), mat(1,3),
                           mat(2,0), mat(2,1), mat(2,3));

    m_Pdf->PrintText(m_Pdf->GetFontHandler(), face, float(font_size), mat33, truncated_text.c_str(), m_TextColor);  
}



END_NCBI_SCOPE
