/*  $Id: macro_parse.cpp 24331 2011-09-14 19:39:12Z kuznets $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Anatoliy Kuznetsov
 *
 * File Description:  Parser for macro language
 *
 */


#include <ncbi_pch.hpp>
#include <gui/utils/macro_parse.hpp>


BEGIN_NCBI_SCOPE

CQueryMacro::CQueryMacro()
: m_MacroType(eUndefinedMacroType),
  m_SelectorType(eUndefinedSelector)
{
}

CQueryMacro::CQueryMacro(const CQueryMacro& qmacro)
: m_MacroName(qmacro.m_MacroName),
  m_MacroTitle(qmacro.m_MacroTitle),
  m_MacroType(qmacro.m_MacroType),
  m_SelectorType(qmacro.m_SelectorType),
  m_VarBlocks(qmacro.m_VarBlocks)
{
}

CQueryMacro& CQueryMacro::operator=(const CQueryMacro& qmacro)
{
    m_MacroName = qmacro.GetMacroName();
    m_MacroTitle = qmacro.GetMacroTitle();
    m_MacroType = qmacro.GetMacroType();
    m_SelectorType = qmacro.GetSelectorType();
    m_VarBlocks = qmacro.GetVarBlocks();

    return *this;
}

CQueryMacro::~CQueryMacro()
{
}

/////////////////////////////////////////////////////////////////////////////
//
//  CQueryMacro::CChoice
//

CQueryMacro::CChoice::CChoice(bool multi)
: m_IsMulti(multi)
{
}

CQueryMacro::CChoice::CChoice(const CChoice& choice)
: m_ChoiceList(choice.m_ChoiceList),
  m_IsMulti(choice.m_IsMulti)
{
}

CQueryMacro::CChoice& CQueryMacro::CChoice::operator=(const CChoice& choice)
{
    m_ChoiceList = choice.m_ChoiceList;
    m_IsMulti = choice.m_IsMulti;
    return *this;
}


/////////////////////////////////////////////////////////////////////////////
//
//  CQueryMacro::CVariable
//


#ifdef _MSC_VER
#pragma warning( disable : 4062 )
#endif


CQueryMacro::CVariable::CVariable()
: m_Type(CQueryParseNode::eNotSet),
  m_IsAsk(false),
  m_IsChoice(false),
  m_Choice(0)
{
}

CQueryMacro::CVariable::CVariable(const string& name, const string& value, bool is_ui_ask)
: m_Name(name),
  m_Type(CQueryParseNode::eString),
  m_Value(value),
  m_IsAsk(is_ui_ask),
  m_IsChoice(false),
  m_Choice(0)
{
}

CQueryMacro::CVariable::CVariable(const string& name, const string& value, CQueryParseNode::EType type)
: m_Name(name),
  m_Type(type),
  m_Value(value),
  m_IsAsk(false),
  m_IsChoice(false),
  m_Choice(0)
{
}


CQueryMacro::CVariable::CVariable(const string& name, int value)
: m_Name(name),
  m_Type(CQueryParseNode::eIntConst),
  m_IsAsk(false),
  m_IsChoice(false),
  m_Choice(0)
{
    m_IntConst = value;
}

CQueryMacro::CVariable::CVariable(const string& name, double value)
: m_Name(name),
  m_Type(CQueryParseNode::eFloatConst),
  m_IsAsk(false),
  m_IsChoice(false),
  m_Choice(0)
{
    m_DoubleConst = value;
}


CQueryMacro::CVariable::CVariable(const string& name, bool value)
: m_Name(name),
  m_Type(CQueryParseNode::eBoolConst),
  m_IsAsk(false),
  m_IsChoice(false),
  m_Choice(0)
{
    m_BoolConst = value;
}

CQueryMacro::CVariable::CVariable(const string& name, CChoice* choice)
: m_Name(name),
  m_Type(CQueryParseNode::eString),
  m_IsAsk(true),
  m_IsChoice(true),
  m_Choice(choice)
{
}


CQueryMacro::CVariable::CVariable(const CVariable& var)
: m_Name(var.m_Name),
  m_Type(var.m_Type),
  m_Value(var.m_Value),
  m_IsAsk(var.m_IsAsk),
  m_IsChoice(var.m_IsChoice),
  m_Choice(0)
{
    switch(m_Type)
    {
    case CQueryParseNode::eIntConst:
        m_IntConst = var.m_IntConst;
        break;
    case CQueryParseNode::eBoolConst:
        m_BoolConst = var.m_BoolConst;
        break;
    case CQueryParseNode::eFloatConst:
        m_DoubleConst = var.m_DoubleConst;
        break;
    default:
        break;
    }
    if (var.m_IsChoice && var.m_Choice) {
        m_Choice = new CChoice(*(var.m_Choice));
    }
}

CQueryMacro::CVariable& CQueryMacro::CVariable::operator=(const CVariable& var)
{
    m_Name = var.GetName();
    m_Type = var.GetType();
    m_Value = var.GetValue();
    m_IsAsk = var.IsUI_Ask();
    m_IsChoice = var.IsUI_Choice();

    switch(m_Type)
    {
    case CQueryParseNode::eIntConst:
        m_IntConst = var.m_IntConst;
        break;
    case CQueryParseNode::eBoolConst:
        m_BoolConst = var.m_BoolConst;
        break;
    case CQueryParseNode::eFloatConst:
        m_DoubleConst = var.m_DoubleConst;
        break;
    }
    delete m_Choice; m_Choice = 0;
    if (var.m_Choice && var.m_IsChoice) {
        m_Choice = new CChoice(*(var.m_Choice));
    }

    return *this;
}

CQueryMacro::CVariable::~CVariable()
{
    delete m_Choice;
}


Int4 CQueryMacro::CVariable::GetInt() const
{
    switch(m_Type)
    {
    case CQueryParseNode::eIntConst:
        return m_IntConst;
    case CQueryParseNode::eBoolConst:
        return (int)m_BoolConst;
    case CQueryParseNode::eFloatConst:
        return (int) m_DoubleConst;
    }

    string s = string("Variable ") + m_Name;
    s += " cannot be promoted to Int";
    NCBI_THROW(CMacroException, eInvalidVariableType, s);

    return 0;
}

bool CQueryMacro::CVariable::GetBool() const
{
    switch(m_Type)
    {
    case CQueryParseNode::eIntConst:
        return m_IntConst!=0;
    case CQueryParseNode::eBoolConst:
        return m_BoolConst;
    }

    string s = string("Variable ") + m_Name;
    s += " cannot be promoted to Bool";
    NCBI_THROW(CMacroException, eInvalidVariableType, s);

    return false;
}

double CQueryMacro::CVariable::GetDouble() const
{
    switch(m_Type)
    {
    case CQueryParseNode::eIntConst:
        return m_IntConst;
    case CQueryParseNode::eBoolConst:
        return (double)m_BoolConst;
    case CQueryParseNode::eFloatConst:
        return m_DoubleConst;
    }

    string s = string("Variable ") + m_Name;
    s += " cannot be promoted to Int";
    NCBI_THROW(CMacroException, eInvalidVariableType, s);

    return 0;
}

const CQueryMacro::CChoice& CQueryMacro::CVariable::GetChoice() const
{
    if (IsUI_Choice()) {
        return *m_Choice;
    }

    string s = string("Variable ") + m_Name;
    s += " is not a CHOICE type";
    NCBI_THROW(CMacroException, eInvalidVariableType, s);
}

/////////////////////////////////////////////////////////////////////////////
//
//  CQueryMacro::CMacroFunction
//


CQueryMacro::CMacroFunction::CMacroFunction()
:m_Line(0)
{}

CQueryMacro::CMacroFunction::CMacroFunction(const string& func_name, const TArgList& args)
: m_Name(func_name), 
  m_Line(0),
  m_ArgList(args)
{}

CQueryMacro::CMacroFunction::CMacroFunction(const CQueryMacro::CMacroFunction& func)
: m_Name(func.m_Name), 
  m_Line(func.m_Line),
  m_ArgList(func.m_ArgList)
{}

CQueryMacro::CMacroFunction& 
CQueryMacro::CMacroFunction::operator=(const CQueryMacro::CMacroFunction& func)
{
  m_Name = func.m_Name;
  m_Line = func.m_Line;
  m_ArgList = func.m_ArgList;
  return *this;
}

CQueryMacro::CMacroFunction::~CMacroFunction()
{
}

/////////////////////////////////////////////////////////////////////////////

size_t CQueryMacro::FindVariable(const TVarBlock& vars, const string& s)
{
    for (size_t i = 0; i < vars.size(); ++i) {
        const CVariable& var = vars[i];
        if (NStr::CompareNocase(var.GetName(), s) == 0) {
            return i;
        }
    }
    return (size_t)(-1);
}

///////////////////////////////////////////////////////////////////////
// Parser
//

static inline 
bool s_IsSpaceTerm(const char *c)
{
    return isspace(*c) || (*c == '\n') || (*c == '\r' && *(c+1) == '\n');
}

static 
bool s_Compare(const char* p, const char* t)
{
    for ( ;  *p  &&  t  &&  *t;  ++p, ++t) {
        if (toupper(*p) != *t) {
            return false;
        }
    }
    return !*t  &&  ( !*p  ||  isspace(*p));
}

void CQueryMacro::Parse(const char* str)
{
    // init the FSM to start parsing
    m_IsEnd = false;
    m_Curr = m_Start = str;

    // 1. Header comments
    x_SkipCommentsOrWhiteSpaces();
    if (m_IsEnd) { 
        string s = "Macro parsing error: Macro not found.";
        NCBI_THROW(CMacroException, eSyntaxError, s);
    }

    // 2. Macro name and title parsing
    bool is_macro = s_Compare(m_Curr, "MACRO");
    if (!is_macro) {
        string s = "Macro parsing error: Macro not found.";
        NCBI_THROW(CMacroException, eSyntaxError, s);
    }
    m_Curr += 5;
    x_SkipCommentsOrWhiteSpaces();
    if (m_IsEnd) { 
        string s = "Macro parsing error: Macro name not found.";
        NCBI_THROW(CMacroException, eSyntaxError, s);
    }
    x_ParseGetVarName(&m_MacroName);
    x_SkipCommentsOrWhiteSpaces();
    if (m_IsEnd) { 
        string s = "Macro parsing error: Macro title not found";
        NCBI_THROW(CMacroException, eSyntaxError, s);
    }
    x_ParseGetString(&m_MacroTitle);



    // 3. Parse variables
    string assignment_chr = "=";
    string var_name, var_str; 
    bool vars_parsing = true;
    for (size_t var_block_idx = 0; vars_parsing; ++var_block_idx) {
        TVarBlocksCollection& vars_coll = SetVarBlocks();
        if (vars_coll.size() <= var_block_idx) {
            vars_coll.push_back(TVarBlock());
        }
        TVarBlock& vars = vars_coll[var_block_idx];

        x_SkipCommentsOrWhiteSpaces();
        if (m_IsEnd) { 
            string s = "Macro parsing error: Macro vars or body not found";
            NCBI_THROW(CMacroException, eSyntaxError, s);
        }
        bool is_var = s_Compare(m_Curr, "VARS");
        if (is_var) {
            m_Curr += 4;
        } else {
            is_var = s_Compare(m_Curr, "VAR");
            if (is_var) {
                m_Curr += 3;
            }
        }
        if (!is_var) break;

        while (is_var) {
            x_SkipCommentsOrWhiteSpaces();
            if (m_IsEnd) { 
                string s = "Macro parsing error: Macro vars or body not found";
                NCBI_THROW(CMacroException, eSyntaxError, s);
            }


            // parse var clause: (variable = value);
            x_ParseGetVarName(&var_name);
            if (m_IsEnd || var_name.empty()) { 
                string s = "Macro parsing error: Macro vars or body not found";
                NCBI_THROW(CMacroException, eSyntaxError, s);
            }
            x_SkipCommentsOrWhiteSpaces(&assignment_chr);
            if (m_IsEnd) { 
                string s = "Macro parsing error: variable value not found (";
                s.append(var_name); s.append(")");
                NCBI_THROW(CMacroException, eSyntaxError, s);
            }
            // check if var name is unique
            size_t var_idx = FindVariable(vars, var_name);
            if (var_idx != -1) {
                string s = "Macro parsing error: variable already exists (";
                s.append(var_name); s.append(")");
                NCBI_THROW(CMacroException, eSyntaxError, s);
            }


            // parse CHOICE statement
            //
            bool is_multi = false;
            bool is_choice = s_Compare(m_Curr, "CHOICE");
            if (is_choice) {
                m_Curr += 6;
            } else {
                is_choice = s_Compare(m_Curr, "MCHOICE");
                if (is_choice) {
                    m_Curr += 7;
                    is_multi = true;
                }
            }

            if (is_choice) {
                x_SkipCommentsOrWhiteSpaces();
                if (m_IsEnd || *m_Curr != '{') { 
                    string s = "Macro parsing error: CHOICE list not found";
                    NCBI_THROW(CMacroException, eSyntaxError, s);
                }              
                ++m_Curr;

                // choice list parsing
                auto_ptr<CChoice> pchoice(new CChoice(is_multi));
                while (1) {
                    x_SkipCommentsOrWhiteSpaces();
                    if (m_IsEnd) { 
                        string s = "Macro parsing error: CHOICE list error";
                        NCBI_THROW(CMacroException, eSyntaxError, s);
                    }

                    bool is_choice_end = s_Compare(m_Curr, "}");
                    if (is_choice_end) {
                        if (pchoice->GetList().size() == 0) {
                            string s = "Macro parsing error: CHOICE list error;";
                            NCBI_THROW(CMacroException, eSyntaxError, s);
                        }
                        vars.push_back(CVariable(var_name, pchoice.release()));
                        break;
                    }

                    char ch = toupper(*m_Curr);
                    if (ch == '"') {             // eString
                        string var_str;
                        x_ParseGetString(&var_str);
                        pchoice->AddChoiceOption(var_str);
                        if (m_IsEnd) { 
                            string s = "Macro parsing error: CHOICE end not found";
                            NCBI_THROW(CMacroException, eSyntaxError, s);
                        }
                    }
                    else
                    if (isdigit(ch)) {      // number (eIntConst or eFloatConst)
                        string var_str;
                        x_ParseGetNumber(&var_str);
                        pchoice->AddChoiceOption(var_str);
                    }
                    else
                    if (ch == ',' || isspace(ch)) {
                        m_Curr++;                
                    }
                    else 
                    if (ch == '}') {
                        m_Curr++;
                        vars.push_back(CVariable(var_name, pchoice.release()));
                        break;
                    }
                    else // variable name (identifier)
                    {
                        string vname;
                        x_ParseGetVarName(&vname);
                        pchoice->AddChoiceOption(vname);
                    }
                } // while (CHOICE list)

                continue;
            } // if CHOICE

            // check what initializer does come next:
            char ch = toupper(*m_Curr);
            if (ch == '"') {             // eString
                x_ParseGetString(&var_str);
                if (m_IsEnd) { 
                    string s = "Macro parsing error: Macro body not found";
                    NCBI_THROW(CMacroException, eSyntaxError, s);
                }
                vars.push_back(CVariable(var_name, var_str));
            }
            else
            if (isdigit(*m_Curr)) {      // number (eIntConst or eFloatConst)
                CQueryParseNode::EType vtype = x_ParseGetNumber(&var_str);
                if (vtype == CQueryParseNode::eIntConst) {
                    int val = NStr::StringToInt(var_str);
                    vars.push_back(CVariable(var_name, val));
                } else
                if (vtype == CQueryParseNode::eFloatConst) {
                    double val = NStr::StringToDouble(var_str);
                    vars.push_back(CVariable(var_name, val));
                }
            }
            else
            if (ch == 'T' || ch == 'F') { // eBoolConst
                bool val = x_ParseGetBool();
                vars.push_back(CVariable(var_name, val));
            }
            else
            if (ch == '%') {              // interactive substitution
                x_ParseGetString(&var_str, '%');
                if (m_IsEnd) { 
                    string s = "Macro parsing error: Macro body not found";
                    NCBI_THROW(CMacroException, eSyntaxError, s);
                }
                vars.push_back(CVariable(var_name, var_str, true/*UI ASK*/));
            }
            x_SkipCommentsOrWhiteSpaces();
            bool b = s_Compare(m_Curr, "VAR");
            if (b) {
                break;
            }
            // check if we need to get out of
            b = s_Compare(m_Curr, "FOR");
            if (b) {
                vars_parsing = false;
                break;
            }
        } // while is_var (single var block parsing)
    } // for (var blocks parsing)

    // 4. FOR EACH 
    //

    x_SkipCommentsOrWhiteSpaces();
    if (m_IsEnd) { 
        string s = "Macro parsing error: Macro not found.";
        NCBI_THROW(CMacroException, eSyntaxError, s);
    }
    {{
        bool for_each = s_Compare(m_Curr, "FOR");
        if (for_each) {
            m_Curr+=3;
            x_SkipCommentsOrWhiteSpaces();
            if (m_IsEnd) { 
                string s = "Macro parsing error: Incomplete FOR EACH.";
                NCBI_THROW(CMacroException, eSyntaxError, s);
            }
            for_each = s_Compare(m_Curr, "EACH");
            if (!for_each) {
                string s = "Macro parsing error: Incomplete FOR EACH.";
                NCBI_THROW(CMacroException, eSyntaxError, s);
            }
            m_Curr += 4;
            m_MacroType = eForEach;

            x_SkipCommentsOrWhiteSpaces();
            if (m_IsEnd) { 
                string s = "Macro parsing error: Incomplete FOR EACH, undefined selector(iterator) type";
                NCBI_THROW(CMacroException, eSyntaxError, s);
            }

            // selector type
            bool seq_feat = s_Compare(m_Curr, "SEQFEAT");
            if (seq_feat) {
                m_Curr += 7;
            } else {
                seq_feat = s_Compare(m_Curr, "SEQ-FEAT");
                if (seq_feat) {
                    m_Curr += 8;
                } else {
                    seq_feat = s_Compare(m_Curr, "FEATURE");
                    if (seq_feat) {
                        m_Curr += 7;
                    }
                }
            }
            if (seq_feat) {
                m_SelectorType = eSeqFeat;
            }
            else {
                string s = "Macro parsing error: Incomplete FOR EACH, unrecognized selector(iterator) type";
                NCBI_THROW(CMacroException, eSyntaxError, s);
            }
        } 
    }} // FOR EACH <selector>

    // 5. WHERE
    x_SkipCommentsOrWhiteSpaces();
    if (m_IsEnd) { 
        string s = "Macro parsing error: Macro not found.";
        NCBI_THROW(CMacroException, eSyntaxError, s);
    }

    {{
    m_WhereQuery.clear();
    bool is_where = s_Compare(m_Curr, "WHERE");
    if (is_where) {
        m_Curr += 5;
        x_ParseGetWhere(&m_WhereQuery);
    }
    }}

    // 6. DO ... DONE
    //
    {{
    x_SkipCommentsOrWhiteSpaces();
    if (m_IsEnd) { 
        string s = "Macro parsing error: Macro body not found.";
        NCBI_THROW(CMacroException, eSyntaxError, s);
    }
    bool is_do = s_Compare(m_Curr, "DO");
    if (!is_do) {
        string s = "Macro parsing error: Macro body not found.";
        NCBI_THROW(CMacroException, eSyntaxError, s);
    }
    m_Curr+=2;

    // 6.1 parse all functions
    //
    string func_name;
    while(!m_IsEnd) {
        x_SkipCommentsOrWhiteSpaces();
        if (m_IsEnd) { 
            string s = "Macro parsing error: Macro end (DONE) not found.";
            NCBI_THROW(CMacroException, eSyntaxError, s);
        }
        bool is_done = s_Compare(m_Curr, "DONE");
        if (is_done) {
            break;
        }
        // get function name
        x_ParseGetFunctionName(&func_name);
        if (m_IsEnd) { 
            string s = "Macro parsing error: Macro end (DONE) not found.";
            NCBI_THROW(CMacroException, eSyntaxError, s);
        }
        if (func_name.empty()) {
            string s = "Macro parsing error: incorrect function name.";
            NCBI_THROW(CMacroException, eSyntaxError, s);
        }

        // parse functions arg list
        //
        TArgList arg_list;
        string var_name;
        while (1) {
            x_SkipCommentsOrWhiteSpaces();
            if (m_IsEnd) { 
                string s = "Macro parsing error: function argument list error (";
                s.append(func_name);
                s.append(')', 1);
                NCBI_THROW(CMacroException, eSyntaxError, s);
            }

            bool is_done = s_Compare(m_Curr, "DONE");
            if (is_done) {
                if (!arg_list.empty()) {
                    string s = "Macro parsing error: function argument list error (";
                    s.append(func_name);
                    s.append(')', 1);
                    NCBI_THROW(CMacroException, eSyntaxError, s);
                }
                break;
            }

            char ch = toupper(*m_Curr);
            if (ch == '"') {             // eString
                string var_str;
                x_ParseGetString(&var_str);
                if (m_IsEnd) { 
                    string s = "Macro parsing error: Macro body not found";
                    NCBI_THROW(CMacroException, eSyntaxError, s);
                }
                CVariable var(var_name, var_str);
                arg_list.push_back(var);
            }
            else
            if (isdigit(ch)) {      // number (eIntConst or eFloatConst)
                string var_str;
                CQueryParseNode::EType vtype = x_ParseGetNumber(&var_str);
                if (vtype == CQueryParseNode::eIntConst) {
                    int val = NStr::StringToInt(var_str);
                    CVariable var(var_name, val);
                    arg_list.push_back(var);
                } else
                if (vtype == CQueryParseNode::eFloatConst) {
                    double val = NStr::StringToDouble(var_str);
                    CVariable var(var_name, val);
                    arg_list.push_back(var);
                }
            }
            else
            if (ch == ',' || isspace(ch)) {
                m_Curr++;                
            }
            else 
            if (ch == ')') {
                m_Curr++;
                CMacroFunction func(func_name, arg_list);
                m_DoList.push_back(func);
                arg_list.clear();
                break;
            }
            else // variable name (identifier)
            {
                string vname;
                x_ParseGetVarName(&vname);

                NStr::ToUpper(vname);
                if (vname == "TRUE") {
                    arg_list.push_back(CVariable(var_name, true));
                }
                else
                if (vname == "FALSE") {
                    arg_list.push_back(CVariable(var_name, false));
                }
                else {
                    CVariable var(var_name, vname, CQueryParseNode::eIdentifier);
                    arg_list.push_back(var);
                }
            }

        } // while (function args)


    } // while (functions)



    }}


}


void CQueryMacro::x_ParseGetVarName(string* var_name)
{
    _ASSERT(var_name);

    var_name->clear();
    while (1)
    {
        char ch = *m_Curr;
        if (ch == 0) {
            m_IsEnd = true;
            break;
        }
        if (isspace(ch) || (ch =='\'') || ch == '"' || ch == '\n' || ch == ',' || ch == ')') {
            break;
        }
        var_name->append(1, ch);
        ++m_Curr;
    } // while
}

void CQueryMacro::x_ParseGetFunctionName(string* func_name)
{
    _ASSERT(func_name);

    func_name->clear();
    while (1)
    {
        char ch = *m_Curr;
        if (ch == 0) {
            m_IsEnd = true;
            break;
        }
        if (isspace(ch) || ch == '\n') {
            // nothing to do, just eat spaces
        } else {
            if ((ch =='(')) {
                ++m_Curr;
                break;
            }
            func_name->append(1, ch);
        }
        ++m_Curr;
    } // while    
}

void CQueryMacro::x_ParseGetWhere(string* where_query)
{
    _ASSERT(where_query);

    where_query->clear();
    while (1)
    {
        if (s_Compare(m_Curr, "DO")) {
            break;
        }
        char ch = *m_Curr;
        if (ch == 0) {
            m_IsEnd = true;
            break;
        }
        where_query->append(1, ch);
        ++m_Curr;
    } // while
}

void CQueryMacro::x_ParseGetString(string* var_str, char bracket)
{
    _ASSERT(var_str);
    var_str->clear();

    char ch = *m_Curr;
    if (ch != bracket) {
        string s = "Macro parsing error: string expected";
        NCBI_THROW(CMacroException, eSyntaxError, s);
    }
    ++m_Curr;

    while (1)
    {
        ch = *m_Curr;
        if (ch == 0) {
            m_IsEnd = true;
            string s = "Macro parsing error: end of string not found ('";
            s.append(*var_str);
            s.append("')");
            NCBI_THROW(CMacroException, eSyntaxError, s);
        }
        if (ch == bracket) {
            ++m_Curr;
            break;
        }

        var_str->append(1, ch);
        ++m_Curr;
    } // while
}

CQueryParseNode::EType CQueryMacro::x_ParseGetNumber(string* var_str)
{
    _ASSERT(var_str);
    var_str->clear();

    CQueryParseNode::EType vtype = CQueryParseNode::eIntConst;

    while (1)
    {
        char ch = *m_Curr;
        if (ch == 0) {
            m_IsEnd = true;
            string s = "Macro parsing error: unexpected end of number";
            NCBI_THROW(CMacroException, eSyntaxError, s);
        }
        if (ch == '.') { // floating point
            vtype = CQueryParseNode::eFloatConst;
        } else
        if (!isdigit(ch)) {
            break;
        }

        var_str->append(1, ch);
        ++m_Curr;
    } // while
    return vtype;
}

bool CQueryMacro::x_ParseGetBool()
{
    bool is_true = s_Compare(m_Curr, "TRUE");
    if (is_true) {
        m_Curr += 4;
        return true;
    }
    bool is_false = s_Compare(m_Curr, "FALSE");
    if (is_false) {
        m_Curr += 5;
        return false;
    }
    string s = "Macro parsing error: boolean expected";
    NCBI_THROW(CMacroException, eSyntaxError, s);
}

void CQueryMacro::x_SkipCommentsOrWhiteSpaces(const string* other_skip_str)
{
    if (m_IsEnd)
        return;

    while (1)
    {
        if (*m_Curr == ';') {
            for (++m_Curr; *m_Curr != '\n'; ++m_Curr) {
                if (*m_Curr == 0) {
                    m_IsEnd = true;
                    return;
                }
            } // for
        }
        if (s_IsSpaceTerm(m_Curr)) {
            ++m_Curr;
        } else {
            if (other_skip_str) {
                basic_string <char>::size_type idx = other_skip_str->find_first_of(*m_Curr);
                if (idx != string::npos) {
                    ++m_Curr;
                    continue;
                }
            }
            break;
        }
    } // while
}


END_NCBI_SCOPE
