/*  $Id: scheduler_engine.cpp 24876 2011-12-16 16:03:41Z kuznets $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Andrey Yazhuk
 *
 * File Description:
 *
 *
 */

#include <ncbi_pch.hpp>


#include <gui/utils/scheduler_engine.hpp>

#include <corelib/ncbi_system.hpp>


BEGIN_NCBI_SCOPE


static const TScheduler_SeriesID kInvalidSchedulerId = (TScheduler_SeriesID)-1;

///////////////////////////////////////////////////////////////////////////////
/// CSchedulerEngineThread
class CSchedulerEngineThread : public CThread
{
public:
    CSchedulerEngineThread(CSchedulerEngine* engine)
        :   m_Engine(engine)    {}

    virtual void* Main(void)
    {
        m_Engine->x_WorkOnThread();
        return NULL;
    }

protected:
    CSchedulerEngine* m_Engine;
};


///////////////////////////////////////////////////////////////////////////////
/// CSchedulerEngine
CSchedulerEngine::CJobRecord::CJobRecord(IScheduledAppJob& job)
:   m_Job(&job),
    m_JobState(IAppJob::eRunning)
{
}


void CSchedulerEngine::CJobRecord::Execute()
{
    m_JobState = m_Job->Run();
}


///////////////////////////////////////////////////////////////////////////////
/// CSchedulerEngine
CSchedulerEngine::CSchedulerEngine()
:   m_Shutdown(false),
    m_Listener(NULL),
    m_JobsScheduled(0),
    m_JobsFinished(0),
    m_JobsCanceled(0)
{
    CMutexGuard guard(m_Mutex);

    m_Scheduler = IScheduler::Create();

    m_WorkerThread.Reset(new CSchedulerEngineThread(this));
    m_WorkerThread->Run(CThread::fRunNice);
}


CSchedulerEngine::~CSchedulerEngine()
{
    LOG_POST(Info << "CSchedulerEngine::~CSchedulerEngine(" << this << "): "
             << m_JobsScheduled << " scheduled / "
             << m_JobsFinished << " finished / "
             << m_JobsCanceled << " canceled");
}


void CSchedulerEngine::ShutDown()
{
    LOG_POST(Info << "CSchedulerEngine::ShutDown()");

    SSignal signal = { eShutDown, NULL  };
    m_Shutdown = true;
    m_Signals.Push(signal);

    m_WorkerThread->Join();

    LOG_POST(Info << "CSchedulerEngine::ShutDown(): completed");
}

void CSchedulerEngine::RequestCancel()
{
    // TODO: implement
}

bool CSchedulerEngine::IsActive()
{
    /// this engine is active - it send notifications when a job changes state
    return true;
}


void CSchedulerEngine::SetListener(IAppJobEngineListener* listener)
{
    m_Listener = listener;
}


void CSchedulerEngine::StartJob(IAppJob& job, IEngineParams* /*params*/)
{
    //_ASSERT(params == NULL); // do not support any

    // we only accept jobs that implement IScheduledAppJob - verify
    IScheduledAppJob* sc_job = x_GetScheduledJob(job);

    CMutexGuard guard(m_Mutex);

    /// check that job is new
    TJobToRecord::const_iterator it = m_JobToRecord.find(sc_job);
    if(it != m_JobToRecord.end())  {
        string s = string("Job ") + job.GetDescr();
        s += " has been already started on this Engine";
        NCBI_THROW(CAppJobException, eEngine_InvalidOperation, s);
    } else {
        /// create a new Record for the job
        TRecordRef rec(new CJobRecord(*sc_job));

        // register this job's
        m_JobToRecord[sc_job] = rec;

        // create a signal for worker thread to start this job
        SSignal signal = { eStart, rec};
        m_Signals.Push(signal);

        x_OnJobStatusChange(*rec);
    }
    ++m_JobsScheduled;
}


void CSchedulerEngine::CancelJob(IAppJob& job)
{
    //LOG_POST(Info << "CSchedulerEngine::CancelJob() " << job.GetDescr());

    CMutexGuard guard(m_Mutex);

    // we only accept jobs that implement IScheduledAppJob - verify
    IScheduledAppJob* sc_job = x_GetScheduledJob(job);

    _ASSERT(sc_job);

    TJobToRecord::iterator it = x_Find(*sc_job);

    CJobRecord& rec = *it->second;

    switch(rec.m_JobState)  {
    case IAppJob::eRunning:  {{
        // notify the job in case it already running
        sc_job->RequestCancel();

        /// signal worker thread to stop scheduling the job
        SSignal signal = { eCancel, &rec};
        m_Signals.Push(signal);
        ++m_JobsCanceled;
        break;
    }}
    default:
        break;
    }
}


void CSchedulerEngine::SuspendJob(IAppJob& /*job*/)
{
    static const char* kMsg = "CSchedulerEngine::SuspendJob()\
                              - operation is not supported";
    NCBI_THROW(CAppJobException, eEngine_InvalidOperation, kMsg);
}


void CSchedulerEngine::ResumeJob(IAppJob& /*job*/)
{
    static const char* kMsg = "CSchedulerEngine::ResumeJob()\
                              - operation is not supported";
    NCBI_THROW(CAppJobException, eEngine_InvalidOperation, kMsg);
}


IAppJob::EJobState CSchedulerEngine::GetJobState(IAppJob& job) const
{
    CMutexGuard guard(m_Mutex);

    CSchedulerEngine* nc_this = const_cast<CSchedulerEngine*>(this);
    IScheduledAppJob* sc_job = nc_this->x_GetScheduledJob(job);

    _ASSERT(sc_job);

    TJobToRecord::const_iterator it = x_Find(*sc_job);

    const CJobRecord& rec = *it->second;
    return rec.m_JobState;
}


void CSchedulerEngine::x_WorkOnThread()
{
    LOG_POST(Info << "Scheduler Engine -  working thread started");

    CTimeSpan timeout(1.0);

    // initially we do not specify a timeout so that this function will wait on
    // the queue indefinitely
    bool has_tasks = false;

    while ( !m_Shutdown ) {
        try {
            // wait until the next task or until a signal arrives
            SSignal signal = m_Signals.Pop(has_tasks ? &timeout : NULL);

            // we have a single to process
            x_OnSignal_OnThread(signal);
        } catch(CSyncQueueException&) {
            // timeout - time to execute tasks
            x_ExecuteReadyTasks_OnThread();

            m_WakeUpTime = m_Scheduler->GetNextExecutionTime();
        }
        // TODO - a workaround for CScheduler bugs

        has_tasks = ! m_Scheduler->IsEmpty();
        //vector<SScheduler_SeriesInfo> series;
        //m_Scheduler->GetScheduledSeries(&series);
        //has_tasks = ! series.empty();

        if(has_tasks)   {
            // we know when to wake-up - calculate time out
            m_Now.SetCurrent();
            timeout = m_WakeUpTime.DiffTimeSpan(m_Now);
        }
        /// now go to sleep until something happens
    }
    LOG_POST(Info << "Scheduler Engine -  working thread finished");
}


void CSchedulerEngine::x_OnSignal_OnThread(SSignal& signal)
{
    //LOG_POST(Info << "CSchedulerEngine::x_OnSignal_OnThread " << signal.m_Action);

    CJobRecord* record = signal.m_JobRecord;

    //LOG_POST(Info << "CSchedulerEngine::x_OnSignal_OnThread() " << record);

    // we here - that means the signal has arrived
    switch(signal.m_Action) {
    case eStart:
        _ASSERT(record);
        m_Now.SetCurrent();
        x_ScheduleTask_OnThread(*record);
        break;

    case eCancel:
        _ASSERT(record);
        x_CancelTask_OnThread(*record);
        break;

    case eShutDown:
        x_CancelAllTasks_OnThread();
        break;

    default:
        _ASSERT(false); // unknown signal
        break;
    };
}


/// execute tasks that are ready to be executed
void CSchedulerEngine::x_ExecuteReadyTasks_OnThread()
{
    while (m_Scheduler->HasTasksToExecute(m_Now)) {
        m_Now.SetCurrent();
        SScheduler_SeriesInfo task_info = m_Scheduler->GetNextTaskToExecute(m_Now);
        CJobRecord* record = dynamic_cast<CJobRecord*>(task_info.task.GetPointer());
        _ASSERT(record);

        IAppJob::EJobState old_state = record->m_JobState;

        record->Execute();

        m_Now.SetCurrent();
        m_Scheduler->TaskExecuted(task_info.id, m_Now);

        // now the task needs to be rescheduled
        switch(record->m_JobState)  {
        case IAppJob::eRunning:
            // now the task needs to be rescheduled
            x_ScheduleTask_OnThread(*record); // create a new one
            break;

        case IAppJob::eCompleted:
        case IAppJob::eCanceled:
        case IAppJob::eFailed:
            record->m_TaskID = kInvalidSchedulerId;
            ++m_JobsFinished;
            break;

        default:
            _ASSERT(false); // unexpected
            break;
        }

        if(old_state != record->m_JobState) {
            x_OnJobStatusChange(*record);
        }
    }
}


// adds a task to the schedule
void CSchedulerEngine::x_ScheduleTask_OnThread(CJobRecord& record)
{
    //LOG_POST("CSchedulerEngine::x_ScheduleTask_OnThread " << &record);

    CTimeSpan period = record.m_Job->GetWaitPeriod();
    CTime new_time = m_Now + period;
    record.m_TaskID = m_Scheduler->AddTask(&record, new_time);

    m_WakeUpTime = m_Scheduler->GetNextExecutionTime();
}


void CSchedulerEngine::x_CancelTask_OnThread(CJobRecord& record)
{
    //LOG_POST(Info << "CSchedulerEngine::x_CancelTask_OnThread " << &record);

    m_Scheduler->RemoveSeries(record.m_TaskID);
    record.m_TaskID = kInvalidSchedulerId;
    record.m_JobState = IAppJob::eCanceled;

    x_OnJobStatusChange(record);

    m_WakeUpTime = m_Scheduler->GetNextExecutionTime();
}


void CSchedulerEngine::x_CancelAllTasks_OnThread()
{
    //TODO - temporary needs to be replaced with better code
    vector<SScheduler_SeriesInfo> infos;
    m_Scheduler->GetScheduledSeries(&infos);

    for( size_t i = 0;  i < infos.size();  i++ )    {
        SScheduler_SeriesInfo& task_info = infos[i];
        CJobRecord* record = dynamic_cast<CJobRecord*>(task_info.task.GetPointer());

        m_Scheduler->RemoveSeries(record->m_TaskID);
        record->m_TaskID = kInvalidSchedulerId;
        record->m_JobState = IAppJob::eCanceled;
    }

    /// now, mark that we are done
    m_Shutdown = true;
}


IScheduledAppJob* CSchedulerEngine::x_GetScheduledJob(IAppJob& job)
{
    IScheduledAppJob* sc_job = dynamic_cast<IScheduledAppJob*>(&job);
    if(sc_job == NULL)  {
        string s = string("Job ") + job.GetDescr();
        s += " does not implement IScheduledAppJob interface";
        NCBI_THROW(CAppJobException, eEngine_IncompatibleJob, s);
    }
    return sc_job;
}


void CSchedulerEngine::x_OnJobStatusChange(CJobRecord& record)
{
    IAppJob& job = *record.m_Job;

    if(m_Listener)  {
        m_Listener->OnEngineJobStateChanged(job, record.m_JobState);
    }
}


/// try to find a job, if not found - throws an exception
CSchedulerEngine::TJobToRecord::iterator
    CSchedulerEngine::x_Find(IScheduledAppJob& job)
{
    TJobToRecord::iterator it = m_JobToRecord.find(&job);
    if(it == m_JobToRecord.end())  {
        string s = string("Job ") + job.GetDescr();
        s += " is not registered with this Engine";
        NCBI_THROW(CAppJobException, eEngine_UnknownJob, s);
    }
    return it;
}


/// try to find a job, if not found - throws an exception
CSchedulerEngine::TJobToRecord::const_iterator
    CSchedulerEngine::x_Find(IScheduledAppJob& job) const
{
    TJobToRecord::const_iterator it = m_JobToRecord.find(&job);
    if(it == m_JobToRecord.end())  {
        string s = string("Job ") + job.GetDescr();
        s += " is not registred with this Engine";
        NCBI_THROW(CAppJobException, eEngine_UnknownJob, s);
    }
    return it;
}


END_NCBI_SCOPE
