/*  $Id: table_names.cpp 25930 2012-06-07 17:54:36Z kuznets $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data,  the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties,  express or implied,  including
 *  warranties of performance,  merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors: Robert Smith
 *  Associate names with the standard score matrixes.
 */

#include <ncbi_pch.hpp>

#include <corelib/ncbistd.hpp>
#include <corelib/ncbi_safe_static.hpp>
#include <corelib/ncbimtx.hpp>
#include <gui/utils/table_names.hpp>
#include <util/static_map.hpp>

BEGIN_NCBI_SCOPE

namespace {
    typedef CStaticArrayMap<const char*, const SNCBIPackedScoreMatrix *, PNocase_CStr> TNameTableMap;
    auto_ptr<TNameTableMap> s_NameTableMap;
}

static CAtomicCounter s_InitGuard;
static CFastMutex     s_Guard;


/// function to check if static array is initialized in MT-safe manner
///
static 
void s_CheckInit_NameTable()
{
    if (s_InitGuard.Get() != 0) {
        return;
    }
    CFastMutexGuard LOCK(s_Guard);

    typedef pair<const char*, const SNCBIPackedScoreMatrix *> TNameTablePair;
    const TNameTablePair s_TableNames[] = {
        TNameTablePair( "Blosum 45", &NCBISM_Blosum45 ),
        TNameTablePair( "Blosum 62", &NCBISM_Blosum62 ),
        TNameTablePair( "Blosum 80", &NCBISM_Blosum80 ),
        TNameTablePair( "PAM 30", &NCBISM_Pam30 ),
        TNameTablePair( "PAM 70", &NCBISM_Pam70 ),
        TNameTablePair( "PAM250", &NCBISM_Pam250 )
   };

   s_NameTableMap.reset(new TNameTableMap(s_TableNames, sizeof(s_TableNames), __FILE__, __LINE__));
}

string  CTableNames::TableToName(const SNCBIPackedScoreMatrix *mp)
{
    s_CheckInit_NameTable();

    const TNameTableMap& tnmap = *(s_NameTableMap.get());
    ITERATE(TNameTableMap, iter, tnmap) {
        if (mp == iter->second) {
            return iter->first;
        }
    }
    return kEmptyStr;
}


const SNCBIPackedScoreMatrix *  CTableNames::NameToTable(const string& name)
{
    s_CheckInit_NameTable();
    const TNameTableMap& tnmap = *(s_NameTableMap.get());

    TNameTableMap::const_iterator iter = tnmap.find(name.c_str());
    if (iter == tnmap.end())
        return NULL;
    return iter->second;
}


vector<string>  CTableNames::AllTableNames()
{
    s_CheckInit_NameTable();
    const TNameTableMap& tnmap = *(s_NameTableMap.get());

    vector<string> names;
    ITERATE(TNameTableMap, iter, tnmap) {
        names.push_back(iter->first);
    }
    return names;
}

END_NCBI_SCOPE
