/*  $Id: align_row.cpp 27581 2013-03-08 17:16:25Z falkrb $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Andrey Yazhuk, Anatoliy Kuznetsov
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>
#include <corelib/ncbistd.hpp>
#include <corelib/ncbitime.hpp>

#include <math.h>
#include <gui/types.hpp>
#include <gui/objutils/label.hpp>
#include <gui/opengl/glhelpers.hpp>
#include <gui/opengl/glbitmapfont.hpp>
#include <gui/opengl/rasterizer.hpp>
#include <gui/opengl/glrender.hpp>
#include <gui/objutils/label.hpp>

#include <objmgr/graph_ci.hpp>
#include <objects/general/Dbtag.hpp>
#include <objects/general/Object_id.hpp>
#include <objects/seqfeat/Org_ref.hpp>

#include <objtools/alnmgr/aln_explorer.hpp>

#include <gui/widgets/aln_multiple/align_row.hpp>
#include <gui/widgets/aln_multiple/alnmulti_renderer.hpp>
#include <gui/widgets/aln_multiple/alnvec_row_graph.hpp>
#include <gui/objutils/taxon_cache.hpp>

#include <objmgr/util/sequence.hpp>


BEGIN_NCBI_SCOPE
USING_SCOPE(ncbi::objects);

////////////////////////////////////////////////////////////////////////////////
/// CAlnRowState

CAlnRowState::CAlnRowState()
: m_RowState(0)
{}

int CAlnRowState::GetRowState() const
{
    return m_RowState;
}

void CAlnRowState::SetRowState(int mask, bool b_set)
{
    if(b_set)   {
        m_RowState |= mask;
    } else {
        m_RowState &= ~mask;
    }
}


////////////////////////////////////////////////////////////////////////////////
/// CAlnVecRow
CAlnVecRow::CAlnVecRow(const IAlignRowHandle& handle)
:   m_Handle(handle),
    m_Host(NULL),
    m_Height(0),
    m_BaseHeight(0),
    m_TaxId(0),
    m_bExpanded(false),
    m_Style(NULL),
    m_Cache(NULL),
    m_PixLeft(-1), m_PixRight(-1), m_ModelLeft(-1), m_ModelRight(-1)
{
    const objects::CBioseq_Handle& bio_handle =  m_Handle.GetBioseqHandle();
    CLabel::GetLabel(*bio_handle.GetSeqId(), &m_Text,
                     CLabel::eDefault, &m_Handle.GetBioseqHandle().GetScope());

    m_TaxId = sequence::GetTaxId(bio_handle);
    CTaxonCache& tc =  CTaxonCache::GetInstance("");
    m_TaxLabel = tc.GetLabel(m_TaxId);
}


CAlnVecRow::CAlnVecRow(const IAlignRowHandle& handle, vector<IAlnVecRowGraph*>& graphs)
:   m_Handle(handle),
    m_Host(NULL),
    m_Height(0),
    m_BaseHeight(0),
    m_TaxId(0),
    m_bExpanded(false),
    m_Style(NULL),
    m_Cache(NULL),
    m_PixLeft(-1), m_PixRight(-1), m_ModelLeft(-1), m_ModelRight(-1)
{
    const objects::CBioseq_Handle& bio_handle =  m_Handle.GetBioseqHandle();
    CLabel::GetLabel(*bio_handle.GetSeqId(), &m_Text,
                     CLabel::eDefault, &(bio_handle.GetScope()));
    m_TaxId = sequence::GetTaxId(bio_handle);
    CTaxonCache& tc =  CTaxonCache::GetInstance("");
    m_TaxLabel = tc.GetLabel(m_TaxId);

 
    m_GraphCont.Reset(new CAlignRowGraphCont());

    NON_CONST_ITERATE(vector<IAlnVecRowGraph*>, it,  graphs) {
        IRenderable* rend = dynamic_cast<IRenderable*>(*it);
        _ASSERT(rend);
        if(rend)    {
            m_GraphCont->AddGraph(rend);
        }
    }

    AddListener(m_GraphCont.GetPointer(), ePool_Child);
    m_GraphCont->AddListener(this, ePool_Parent);
}


BEGIN_EVENT_MAP(CAlnVecRow, CEventHandler)
    ON_EVENT(CGraphNotification, CGraphNotification::eContentChanged,
               &CAlnVecRow::x_OnGraphContentChanged)
    ON_EVENT(CGraphNotification, CGraphNotification::eSizeChanged,
               &CAlnVecRow::x_OnGraphSizeChanged)
END_EVENT_MAP()


CAlnVecRow::~CAlnVecRow()
{
    m_DList.Delete();
    m_Style = NULL; //do not delete
}


void CAlnVecRow::SetScoreCache(CScoreCache* cache)
{
    m_Cache = cache;
}


void CAlnVecRow::GraphicsCacheCmd(EGraphCacheCmd cmd)
{
    switch(cmd) {
    case eInvalidate: m_DList.Invalidate(); break;
    case eDelete: m_DList.Delete(); break;
    }
}


void CAlnVecRow::GetHTMLActiveAreas(TColumnType col_type, CGlPane& pane, TAreaVector& areas)
{
    string s_row = NStr::IntToString(m_Handle.GetRowNum());

    switch(col_type)    {
    case eDescr: {
        TVPRect rc = pane.GetViewport();
        CHTMLActiveArea area(CAlnMultiRenderer::eHtmlArea_Link, rc, s_row, m_Text, "row description");
        areas.push_back(area);
        break;
    }
    case eIcons:    {       // generate click-able button for expanding/collapsing row
        if (!IsExpandable())
            break;

        string s_row = NStr::IntToString(m_Handle.GetRowNum());

        CHTMLActiveArea area(CAlnMultiRenderer::eHtmlArea_CheckBox,
                             x_GetButtonRect(pane, eExpand),
                            s_row, "Click to expand/collapse", "toggle_expand");
        areas.push_back(area);
        break;
    };
    case eAlignment:    {   // create area for alignment
        const TModelRect& rc_vis = pane.GetVisibleRect();
        TSignedSeqPos vis_start = (TSignedSeqPos) floor(rc_vis.Left());
        TSignedSeqPos vis_stop = (TSignedSeqPos) ceil(rc_vis.Right());

        TSignedSeqPos start = max(m_Handle.GetSeqAlnStart(), vis_start);
        TSignedSeqPos stop = min(m_Handle.GetSeqAlnStop(), vis_stop);

        pane.OpenOrtho();
        pane.Close();

        TVPUnit vp_start = pane.ProjectX(start);
        TVPUnit vp_stop = pane.ProjectX(stop);
        TVPRect rc = pane.GetViewport();
        rc.SetHorz(vp_start, vp_stop);

        CHTMLActiveArea area(CAlnMultiRenderer::eHtmlArea_Link, rc,
                             s_row, m_Text, "whole alignment");
        areas.push_back(area);

        // generate HTML areas representing segments
        TSignedRange   range(start, stop);
        IAlnSegmentIterator::EFlags flags = IAlnSegmentIterator::eSkipInserts;
        auto_ptr<IAlnSegmentIterator> p_it(m_Handle.CreateSegmentIterator(range, flags));

        for ( IAlnSegmentIterator& it(*p_it);  it;  ++it )  {
            const IAlnSegment& seg = *it;
            const TSignedSeqRange& R = seg.GetAlnRange();
            start = max(R.GetFrom(), vis_start);
            stop = min(R.GetToOpen(), vis_stop);
            vp_start = pane.ProjectX(start);
            vp_stop = pane.ProjectX(stop);
            rc.SetHorz(vp_start, vp_stop);

            CHTMLActiveArea sg_area(CAlnMultiRenderer::eHtmlArea_Link, rc, s_row,
                                 m_Text, "alignment section");
            areas.push_back(sg_area);
        }
        break;
    };
    default:
        break;
    };
}


/// height of "white-space" between row borders and alignment band
static const int kAlignSpace = 1;

/// vertical offset from the border of the alignment band to the border of gap band
static const int kGapOffset = 3;

static const int kTextVertSpace = 2;
static const int kSeqVertSpace = kAlignSpace + 2;


void CAlnVecRow::SetDisplayStyle(const CRowDisplayStyle* style)
{
    m_Style = style;

    UpdateOnStyleChanged();
}


void CAlnVecRow::UpdateOnStyleChanged()
{
    m_Height = m_BaseHeight = x_GetBaseHeight();

    if(m_bExpanded) {
        if(m_GraphCont)   {
            m_Height += m_GraphCont->PreferredSize().Y();
        }
    }
}

void CAlnVecRow::UpdateOnAnchorChanged()
{
    if(m_bExpanded  &&  m_GraphCont->IsCreated()) {
        m_GraphCont->Destroy();
        m_GraphCont->Create();
    }
}


int CAlnVecRow::x_GetBaseHeight() const
{
    TModelUnit h1 = kTextVertSpace * 2;
    TModelUnit h2 = kSeqVertSpace * 2;

    if(m_Style)
    {
        h1 += m_Style->GetTextFont().TextHeight();
        h2 += m_Style->GetSeqFont().TextHeight();
    }
    return (int) max(h1, h2);
}


const CRowDisplayStyle*    CAlnVecRow::GetDisplayStyle()
{
    return m_Style;
}


void CAlnVecRow::SetHost(IAlignRowHost* host)
{
    m_Host = host;
}


int CAlnVecRow::GetRowNum()  const
{
    return m_Handle.GetRowNum();
}


int CAlnVecRow::GetHeightPixels()    const
{
    return m_Height;
}


void CAlnVecRow::RenderColumn(TColumnType col_type, CGlPane& pane, int state)
{
    string error;
    try {
    CGlAttrGuard AttrGuard(GL_POLYGON_MODE | GL_LINE_BIT);

    if(m_Style)    {
        switch(col_type) {
        case eDescr:     x_RenderDescr(pane, state);       break;
        case eIcons:     x_RenderIcons(pane, state);       break;
        case eIconStrand:x_RenderIconStrand(pane, state);  break;
        case eIconExpand:x_RenderIconExpand(pane, state);  break;
        case eSeqStart:  x_RenderSeqStart(pane, state);    break;
        case eStart:     x_RenderStartPos(pane, state);    break;
        case eAlignment: x_RenderAlign(pane);              break;
        case eEnd:       x_RenderEndPos(pane, state);      break;
        case eSeqEnd:    x_RenderSeqEnd(pane, state);      break;
        case eSeqLength: x_RenderSeqLength(pane, state);   break;
        case eTaxLabel:  x_RenderTaxLabel(pane, state);    break;
        default: 
            NCBI_ASSERT(0, "Unknown column type");
            break; // unknown column
        } // switch
    }
    } catch(CException& e)  {
        error = e.GetMsg();
    } catch(std::exception& e)  {
        error = e.what();
    } 
    if( ! error.empty())    {
        ERR_POST("CAlnVecRow::RenderColumn() " << error);
    }
}


void CAlnVecRow::OnMouseEvent(wxMouseEvent& event, TColumnType col_type, CGlPane& pane)
{
    wxEventType type = event.GetEventType();
    if(type == wxEVT_LEFT_DOWN)   {
        wxPoint pos = event.GetPosition();
        TVPPoint pt = x_GetHost()->ARH_GetVPPosByWindowPos(pos);
        int shift = ((col_type == IAlignRow::eIconExpand) ? 0 : 1);
        TVPRect rc_btn = x_GetButtonRect(pane, eExpand, shift);
        if(rc_btn.PtInRect(pt)) {
            x_OnClickExpandBtn();
            return;
        }
    }
    event.Skip();
}


static const int kTooltipMaxLines = 5;
static const int kMaxSeqInTooltip = 20;


string  CAlnVecRow::GetTooltip(const TVPPoint& vp_pt, TColumnType col_type,
                               CGlPane& pane)
{
    switch(col_type) {
    case eDescr:
        return x_GetDescrTooltip(vp_pt, pane);
    case eIcons:
        return x_GetIconsTooltip(vp_pt, pane);
    case eSeqStart:
        return "First position in sequence coordinates";
    case eStart:
        return "First visible position in alignment coordinates";
    case eAlignment:
        return x_GetAlignmentTooltip(vp_pt, pane);
    case eEnd:
        return "Last visible position in alignment coordinates";
    case eSeqEnd:
        return "Last position in sequence";
    case eSeqLength:
        return "Sequence length";
    case eTaxLabel:
        return "Organism name";
    default: break;
    }
    return "";
}


string  CAlnVecRow::x_GetDescrTooltip(const TVPPoint& /*vp_pt*/, CGlPane& /*pane*/)
{
    CBioseq_Handle handle = m_Handle.GetBioseqHandle();

    string str;

    /// first, the text for our row
    str += m_Text + "\n";

    /// next, title of the bioseq
    str += sequence::GetTitle(handle);
    str += "\n";

    /// org info
    try {
        const COrg_ref& ref = sequence::GetOrg_ref(handle);
        str += "Organism: ";
        ref.GetLabel(&str);
    }
    catch (CException&) {
    }

    /// sequence length information
    string type;
    if (handle.IsNa()) {
        type = " bases";
    } else {
        type = " residues";
    }

    str += "\nSequence length: " + NStr::IntToString(handle.GetBioseqLength(), NStr::fWithCommas);
    str += type;
    str += "\nAligned length: " +
        NStr::IntToString(m_Handle.GetSeqStop() - m_Handle.GetSeqStart() + 1, NStr::fWithCommas);
    str += type;

    return str;
}

string  CAlnVecRow::x_GetIconsTooltip(const TVPPoint& vp_pt, CGlPane& pane)
{
    _ASSERT(x_GetHost());
    string s_tip;

    EButtons btn = x_HitTestButton(pane, vp_pt);
    switch(btn) {
    case eStrandMarker: {
        s_tip = "Strand - ";
        s_tip += m_Handle.IsNegativeStrand() ? "Negative" : "Positive";
    }; break;
    case eExpand:   {
        if(IsExpandable())    {
            s_tip = "Graphs: Click to ";
            s_tip += m_bExpanded ? "collapse" : "expand";
        } else {
            s_tip = "No graphs available - nothing to expand";
        }
    }; break;
    //case eSetupGraphs:  s_tip = "Setup Graphs"; break;
    default: break;
    }
    return s_tip;
}


#define TOOLTIP_RANGE_PIX 3

string  FormatRange(TSignedSeqPos from, TSignedSeqPos to)
{
    string s =  NStr::IntToString(from + 1, NStr::fWithCommas);
    s += "-";
    s += NStr::IntToString(to + 1, NStr::fWithCommas);
    return s;
}


string  CAlnVecRow::x_GetAlignmentTooltip(const TVPPoint& vp_pt, CGlPane& pane)
{
    string s_tip;

    if(m_bExpanded  &&  m_GraphCont) {
        // if vp_pt is in graph rect
        TVPRect graph_rc_vp = m_GraphCont->GetVPRect();
        if(graph_rc_vp.PtInRect(vp_pt)) {
            // try to get graph tooltip
            if(m_GraphCont->NeedTooltip(pane, vp_pt.X(), vp_pt.Y()))   {
                s_tip = m_GraphCont->GetTooltip();
                if( ! s_tip.empty())    {
                    return s_tip;
                }
            }
        }
    }


    // if got - return
    pane.OpenOrtho();
    pane.Close();

    double w_x = pane.UnProjectX(vp_pt.X());
    double pix_w = pane.UnProjectWidth(TOOLTIP_RANGE_PIX);
    TSeqPos pos = (TSeqPos) w_x;

    s_tip = x_GetAlignmentTooltip_General(pos);

    // Append information on gaps and inserts
    TSeqPos start = (TSeqPos) floor(w_x - pix_w / 2);
    TSeqPos end = (TSeqPos) ceil(w_x + pix_w / 2);
    TSignedRange range(start, end);

    // obtain segments in the range [start, end]
    IAlnSegmentIterator::EFlags flags = IAlnSegmentIterator::eAllSegments;
    auto_ptr<IAlnSegmentIterator> p_it(m_Handle.CreateSegmentIterator(range, flags));
    IAlnSegmentIterator& it = *p_it;

    if(it)   {
        // last positions on the last aligned segment
        TSignedSeqPos prev_to = -2, prev_aln_to = -2;
        const int kMaxRegions = 5; // max inserts or unaligned regions reported
        int n_regions = 0;

        for (int i = 0;  it  &&  i < kTooltipMaxLines;  i++, ++it  )    {
            CAlnMap::TSegTypeFlags type  = it->GetType();
            const TSignedSeqRange& aln_r = it->GetAlnRange();
            const TSignedSeqRange& seq_r = it->GetRange();

            bool aligned = (type == IAlnSegment::fAligned)  &&  ! aln_r.Empty();
            bool insert = (type == IAlnSegment::fAligned)  &&  aln_r.Empty();
            bool gap = (type == IAlnSegment::fGap)  &&  aln_r.GetLength() > 0;
            // TODO there is a problem, we do not distinguish between real gaps
            // and the regions beyond the sequence ends

            if(insert)  {
                if(n_regions < kMaxRegions) {
                    string s_insert = x_GetAlignmentTooltip_Insert(aln_r, seq_r);
                    s_tip += s_insert;
                }
                n_regions++;
            }   else if(gap)   {
                if(n_regions < kMaxRegions) {
                    string s_gap = x_GetAlignmentTooltip_Gap(aln_r, seq_r);
                    s_tip += s_gap;
                }
                n_regions++;
            } else if(aligned)  {
                if(aln_r.GetFrom() == prev_aln_to + 1  &&  seq_r.GetFrom() > prev_to + 1)  {
                    // there is an unaligned region
                    if(n_regions < kMaxRegions) {
                        string s_unaln =
                            x_GetAlignmentTooltip_Unaligned(prev_aln_to, aln_r.GetFrom(),
                                                            prev_to, seq_r.GetFrom());
                        s_tip += s_unaln;
                    }
                    n_regions++;
                }

                prev_aln_to = aln_r.GetTo();
                prev_to = seq_r.GetTo();
            }
        }

        if(n_regions > kMaxRegions) {
            s_tip += "\n... total " + NStr::IntToString(n_regions);
        }
    }

    return s_tip;
}


string CAlnVecRow::x_GetAlignmentTooltip_General(TSeqPos pos)
{
    // append Name and Position
    string s_tip = m_Text;
    if(m_Handle.IsNegativeStrand())
        s_tip += ", Negative Strand";

    s_tip += string("\nAlignment Pos ") + NStr::IntToString(pos + 1, NStr::fWithCommas);

    TSignedSeqPos seq_pos = m_Handle.GetSeqPosFromAlnPos(pos);
    if(seq_pos >= 0)    {
        s_tip += ", on Sequence ";
        s_tip += NStr::IntToString(seq_pos + 1, NStr::fWithCommas);
    }
	return s_tip;
}


string CAlnVecRow::x_GetAlignmentTooltip_Insert(const TSignedSeqRange& aln_r,
                                                const TSignedSeqRange& seq_r)
{
    TSeqPos aln_left = aln_r.GetTo(); // a bit of hack

    string s_tip = "\nInsert - ";

    s_tip += NStr::IntToString(aln_left + 1, NStr::fWithCommas);
    s_tip += "][";
    s_tip += NStr::IntToString(aln_left + 2, NStr::fWithCommas);

    s_tip += ", on Sequence [";
    s_tip += FormatRange(seq_r.GetFrom(), seq_r.GetTo() + 1);
    s_tip += "] : ";

    s_tip += "\n  " + x_GetTooltipSequence(seq_r.GetFrom(), seq_r.GetTo());

    return s_tip;
}


string CAlnVecRow::x_GetAlignmentTooltip_Gap(const TSignedSeqRange& aln_r,
                                                const TSignedSeqRange& seq_r)
{
    string s_tip = "\nGap ";
    s_tip += FormatRange(aln_r.GetFrom() + 1, aln_r.GetTo() + 1);
    s_tip += ", on Sequence ";
    s_tip += FormatRange(seq_r.GetFrom() + 1, seq_r.GetTo() + 1);
    return s_tip;
}


string CAlnVecRow::x_GetAlignmentTooltip_Unaligned(TSignedSeqPos aln_prev_to,
                                                   TSignedSeqPos aln_from,
                                                   TSignedSeqPos prev_to,
                                                   TSignedSeqPos from)
{
    string s_tip = "\nUnaligned region - ";

    s_tip += NStr::IntToString(aln_prev_to + 1, NStr::fWithCommas);
    s_tip += "][";
    s_tip += NStr::IntToString(aln_from + 1, NStr::fWithCommas);

    s_tip += ", on Sequence [";
    s_tip += FormatRange(prev_to + 1, from - 1);
    s_tip += "] : ";

    s_tip += "\n  " + x_GetTooltipSequence(prev_to + 1, from - 1);

    return s_tip;
}


string CAlnVecRow::x_GetTooltipSequence(TSeqPos from, TSeqPos to)
{
    string seq;
    TSeqRange seq_text_r(from, to);

    if((int) seq_text_r.GetLength() > kMaxSeqInTooltip)    { // to long to be displayed
        // get kMaxSeqInTooltip / 2 bases on the left and on the right
        TSeqPos left = seq_text_r.GetFrom();
        TSeqRange part_r(left, left + kMaxSeqInTooltip / 2);

        m_Handle.GetSeqString(seq, part_r);

        TSeqPos to_open = to + 1;
        part_r.Set(to_open - kMaxSeqInTooltip / 2, to_open -1);
        string right_seq;
        m_Handle.GetSeqString(right_seq, part_r);

        seq += "..." + right_seq;
    } else {
        m_Handle.GetSeqString(seq, seq_text_r);
    }
    return seq;
}


void CAlnVecRow::x_OnClickExpandBtn()
{
    wxBusyCursor wait;

    Expand(! IsExpanded());
}


inline  void    glLined(TModelUnit x1, TModelUnit y1, TModelUnit x2, TModelUnit y2)
{
    CGlRender& gl = GetGl();
    gl.Vertex2d(x1, y1);
    gl.Vertex2d(x2, y2);
}


const static TModelUnit kScoreMinScale = 0.5;

/// Renders graphical part of the row
void CAlnVecRow::x_RenderAlign(CGlPane& pane)
{
    _ASSERT(m_Style);
    pane.OpenOrtho();

    CGlRender& gl = GetGl();

    TModelRect rc_lim = pane.GetModelLimitsRect();
    TModelRect rc_model = pane.GetVisibleRect();
    TModelUnit offset_y = pane.GetOffsetY();

    TModelUnit y2 = rc_lim.Top() + kAlignSpace - offset_y;
    TModelUnit y1 = rc_lim.Top() + m_BaseHeight -1 - kAlignSpace - offset_y;

    gl.LineWidth(1.0f);
    gl.ColorC(m_Style->GetColor(CWidgetDisplayStyle::eAlignSegs));

    // now render segments actually present
    int left = (int) floor(rc_model.Left());
    int right = int(ceil(rc_model.Right())) - 1;

    TSignedRange range(left, right);
    // render aligned segments
    IAlnSegmentIterator::EFlags flags = IAlnSegmentIterator::eAllSegments;
    auto_ptr<IAlnSegmentIterator> p_it(m_Handle.CreateSegmentIterator(range, flags));
    IAlnSegmentIterator& it = *p_it;

    bool averageable = false;
    IScoringMethod* method = NULL;
    if(m_Cache)    {
        method = m_Cache->GetScoringMethod();
        if(method)  {
            averageable = method->IsAverageable();
        }
    }
    TModelUnit scale_x = pane.GetScaleX();
    pane.Close();

    /// Render Alignment segments & Gaps
    bool has_scores = m_Cache  &&  m_Cache->HasScores();
    bool can_render_scores = averageable  ||  scale_x < kScoreMinScale; // resolution permits rendering
    if(has_scores  &&  can_render_scores)   {
        // render parts of segments accordingly to scores
        x_RenderAlignScores(pane, it, (int) y1, (int) y2, averageable);
    } else {
        // scores cannot be rendered
        CWidgetDisplayStyle::EColorType type =
            has_scores ? CWidgetDisplayStyle::eAlignSegsNoScores
                       : CWidgetDisplayStyle::eAlignSegs;
        const CRgbaColor& color = m_Style->GetColor(type);
        x_RenderAlignSegments(pane, it, y1, y2, color);
    }

    pane.OpenOrtho();

    /// Render Inserts
    x_RenderInserts(pane, left, right, y1, y2);

    x_RenderUnaligned(pane, left, right, y1, y2);

    /// Render Sequence
    if(x_CanRenderSeq(pane))    {
        int start = (left < rc_model.Left()) ? left + 1 : left;
        int stop = (int) floor(rc_model.Right());
        if(stop + 1 > rc_model.Right()) {
            stop--;
        }
        x_RenderSequence(pane, y1, y2, start, stop);
    }

    pane.Close();

    /// Render Graph
    x_RenderGraph(pane, range);
}


inline double CompressLength(double len)
{
    if(len <= 10.0) {
        return len;
    } else {
        double log = log10(len) * 10;
        return log;
    }
}


void CAlnVecRow::x_RenderInserts(CGlPane& pane, int from, int to,
                                 TModelUnit top_y, TModelUnit bottom_y)
{
    CGlRender& gl = GetGl();

    TSignedRange range(from, to);
    TModelUnit offset_x = pane.GetOffsetX();

    CRgbaColor cl_insert(0.0f, 0.0f, 1.0f);
    gl.ColorC(cl_insert);
    gl.Begin(GL_LINES);

    IAlnSegmentIterator::EFlags flags = IAlnSegmentIterator::eInsertsOnly;
    auto_ptr<IAlnSegmentIterator> p_it_2(m_Handle.CreateSegmentIterator(range, flags));

    const TModelUnit kMinLen = pane.GetScaleX() * 3; // 3 pixels in model units

    for(  IAlnSegmentIterator& it_2 = *p_it_2;  it_2;  ++it_2 )  {
        TModelUnit ins_pos = it_2->GetAlnRange().GetFrom() - offset_x;
        TModelUnit len = it_2->GetRange().GetLength();
        TModelUnit log_len = CompressLength(len);
        log_len = max(log_len, kMinLen);
        TModelUnit x1 = ins_pos - log_len / 2;
        TModelUnit x2 = x1 + log_len;

        // assuming glBegin(GL_LINES)
        glLined(ins_pos, top_y, ins_pos, bottom_y);
        glLined(x1, top_y, x2, top_y);
        glLined(x1, bottom_y, x2, bottom_y);
    }
    gl.End();
}


void CAlnVecRow::x_RenderUnaligned(CGlPane& pane, int from, int to,
                                 TModelUnit top_y, TModelUnit bottom_y)
{
    CGlRender& gl = GetGl();

    TSignedRange range(from, to);
    TModelUnit offset_x = pane.GetOffsetX();

    CRgbaColor cl_unaligned(0.0f, 0.5f, 0.0f);
    gl.ColorC(cl_unaligned);
    gl.Begin(GL_LINES);

    IAlnSegmentIterator::EFlags flags = IAlnSegmentIterator::eSkipGaps;
    auto_ptr<IAlnSegmentIterator> p_it_2(m_Handle.CreateSegmentIterator(range, flags));

    const TModelUnit kMinLen = pane.GetScaleX() * 3; // 3 pixels in model units

    TSignedSeqPos prev_pos = -2, prev_aln_pos = -2;
    for(  IAlnSegmentIterator& it_2 = *p_it_2;  it_2;  ++it_2 )  {
        const TSignedRange& aln_r = it_2->GetAlnRange();
        const TSignedRange& r = it_2->GetRange();

        if(aln_r.GetFrom() == prev_aln_pos + 1  &&  r.GetFrom() > prev_pos + 1)  {
            TModelUnit ins_pos = aln_r.GetFrom() - offset_x;
            TModelUnit len = it_2->GetRange().GetLength();
            TModelUnit log_len = CompressLength(len);
            log_len = max(log_len, kMinLen);
            TModelUnit x1 = ins_pos - log_len / 2;
            TModelUnit x2 = x1 + log_len;

            // assuming gl.Begin(GL_LINES)
            glLined(ins_pos, top_y, ins_pos, bottom_y);
            glLined(x1, top_y, x2, top_y);
            glLined(x1, bottom_y, x2, bottom_y);
        }

        prev_aln_pos = aln_r.GetTo();
        prev_pos = r.GetTo();
    }
    gl.End();
}


struct FWeightedSum
{
    void    operator()(float& dst, const float& src, float fraction, bool long_seg)
    {
        dst += src * fraction;
    }
};


void CAlnVecRow::x_RenderAlignSegments(CGlPane& pane, IAlnSegmentIterator& it,
                                       TModelUnit top_y, TModelUnit bottom_y,
                                       const CRgbaColor& color)
{
    CGlRender& gl = GetGl();

    const TVPRect& rc_vp = pane.GetViewport();
    TModelRect rc_vis = pane.GetVisibleRect();

    // render a single line for all possible Gaps
    pane.OpenOrtho();

    TSeqPos row_start = m_Handle.GetSeqAlnStart();
    TSeqPos row_stop = m_Handle.GetSeqAlnStop();

    TModelUnit offset_x = pane.GetOffsetX();
    TModelUnit y_c = (top_y + bottom_y) / 2;

    gl.Begin(GL_LINES);
    gl.ColorC(color);
    glLined(row_start - offset_x, y_c, row_stop - offset_x, y_c);
    gl.End();

    pane.Close();

    /// Render Segments
    pane.OpenPixels();

    /// Rasterize segments - raster represents density of pixel coverage by segments
    typedef CRasterizer<float>    TR;
    TR rasterizer(rc_vp.Left(), rc_vp.Right(), rc_vis.Left(), rc_vis.Right());

    FWeightedSum f_sum;
    for( ;  it;  ++it )  {
        const IAlnSegment& seg = *it;
        if(seg.IsAligned())  {
            const TSignedSeqRange& r = seg.GetAlnRange();
            rasterizer.AddInterval(r.GetFrom(), r.GetToOpen(), 1.0, f_sum); /* NCBI_FAKE_WARNING: WorkShop */
        }
    }

    TVPUnit vp_top_y = rc_vp.Top() - kAlignSpace;
    TVPUnit vp_bottom_y = rc_vp.Top() - (m_BaseHeight-1 - kAlignSpace);

    const TR::TRaster& raster = rasterizer.GetRaster();

    CGlAttrGuard guard(GL_LINE_BIT |  GL_ENABLE_BIT  | GL_COLOR_BUFFER_BIT );
    gl.Enable(GL_BLEND);
    gl.BlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    gl.PolygonMode(GL_FRONT_AND_BACK, GL_FILL);

    int off_x = rc_vp.Left();
    static const float eps = 0.001f;

    CRgbaColor cl(color);
    for( size_t i = 0;  i < raster.size(); )    {
        size_t start = i;
        float v = raster[i];
        while(i < raster.size()  &&  raster[i] == v) {
            i++;
        }

        if(v > eps)   {    /// otherwise consider it to be transparent
            /// use density as a measure of transparency
            // v is distributed in [0, 1], however Aplha channel needs to be in
            // [0.25, 1.0] range
            double alpha = 0.25 + v * 0.75;
            cl.SetAlpha((float) alpha);
            gl.ColorC(cl);
            gl.Rectd(start + off_x, vp_top_y, i + off_x, vp_bottom_y + 1);
        }
    }

    pane.Close();
}


/// represents a pixel with a score, if several intervals cover a pixel, the interval
/// with the maximum coverage will "own" the pixel (i.e. the pixel will be colored
/// according to the score of this interval)
/// This is used for non-averageable scores.
struct SScPix   {
    float   m_Score;    /// score assigned to the pixel
    float   m_MaxCoverage; /// fraction of the pixel corresponding to the m_Score
    bool    m_Gap; /// true if this is a gap interval

    bool operator==(const SScPix& other) const
    {
        return m_Score == other.m_Score  &&  m_MaxCoverage == other.m_MaxCoverage
               &&  m_Gap == other.m_Gap;
    }
};


/// a pixel with the Max coverage (fraction) overrides all attributes
static void FSetPixMaxCoverage(SScPix& dst, const SScPix& src, float fraction, bool long_seg)
{
    if(fraction > dst.m_MaxCoverage)   {
        dst.m_Score = src.m_Score;
        dst.m_Gap = src.m_Gap;
        dst.m_MaxCoverage = fraction;
    }
}


void CAlnVecRow::x_RenderAlignScores(CGlPane& pane, IAlnSegmentIterator& it_seg,
                                      int /*top_y*/, int /*bottom_y*/, bool averageable)
{
    CGlRender& gl = GetGl();

    typedef CScoreCache::TScore TScore;
    pane.OpenPixels();

    const TVPRect& rc_vp = pane.GetViewport();
    const TModelRect rc_vis = pane.GetVisibleRect();

    gl.Translated(0.0, rc_vp.Top(), 0.0); // to make the display list "scrollable" vertically

    //if horz projection has changed - invalidate display list
    if(m_PixLeft != rc_vp.Left()  ||  m_PixRight != rc_vp.Right()
        ||  m_ModelLeft != rc_vis.Left()  ||  m_ModelRight != rc_vis.Right())    {

        m_DList.Invalidate();

        m_PixLeft = rc_vp.Left();
        m_PixRight = rc_vp.Right();
        m_ModelLeft = rc_vis.Left();
        m_ModelRight = rc_vis.Right();
    }

    if(m_DList.IsValid())   {
        m_DList.Call();
    } else { // need to create a new one
        m_DList.Begin(GL_COMPILE_AND_EXECUTE);

        CGlAttrGuard guard(GL_ENABLE_BIT  | GL_COLOR_BUFFER_BIT );
        gl.Enable(GL_BLEND);
        gl.BlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

        TVPUnit vp_top_y = -kAlignSpace;
        TVPUnit vp_bottom_y = - (m_BaseHeight -1 - kAlignSpace);

        // obtain Score Collection and setup iterators
        int range_from = (int) (floor(rc_vis.Left()));
        const CScoreCache::TScoreColl& ScoreColl = m_Cache->GetScores(m_Handle.GetRowNum());
        CScoreCache::TScoreColl::const_iterator it = ScoreColl.begin();
        const CScoreCache::TScoreColl::const_iterator it_end = ScoreColl.end();
        it = ScoreColl.find(it, it_end, range_from);

        if(averageable) {
            x_RenderAveragableScores(pane, it_seg, it, it_end, vp_top_y, vp_bottom_y);
        } else {
            x_RenderNonAveragableScores(pane, it_seg, it, it_end, vp_top_y, vp_bottom_y);
        }

        m_DList.End();
    }
    gl.Translated(0.0, -rc_vp.Top(), 0.0);

    pane.Close();
}


/// represents a pixel with two scores, one corresponds to a gap interval, another -
/// to an aligned segment.
struct SScPix2   {
    float m_IntScore;
    float m_IntLen;
    float m_GapScore;
    float m_GapLen;
    float m_MinScore;
    bool  m_Init;

    bool    operator==(const SScPix2& other) const
    {
        return m_IntScore == other.m_IntScore  &&  m_IntLen == other.m_IntLen &&
                m_GapScore == other.m_GapScore  &&  m_GapLen == other.m_GapLen &&
                m_MinScore == other.m_MinScore;
    }
};


/// accumulate weighted scores
static void FSetPix2(SScPix2& dst, const SScPix2& src, float fraction, bool long_seg)
{
    dst.m_IntScore += src.m_IntScore * fraction;
    dst.m_IntLen += src.m_IntLen * fraction;
    dst.m_GapScore += src.m_GapScore * fraction;
    dst.m_GapLen += src.m_GapLen * fraction;

    if( ! long_seg)  {
        // only for small segments
        static const float eps = 0.00001f;
        if( ! dst.m_Init)   {
            if(src.m_IntLen > eps)  {
                dst.m_MinScore = src.m_IntScore;
                dst.m_Init = true;
            } else if(src.m_GapLen > eps)  {
                dst.m_MinScore = src.m_GapScore;
                dst.m_Init = true;
            }
        } else {
            if(src.m_IntLen > eps)  {
                dst.m_MinScore = min(dst.m_MinScore, src.m_IntScore);
            }
            if(src.m_GapLen > eps)  {
                dst.m_MinScore = min(dst.m_MinScore, src.m_GapScore);
            }
        }
    }
}


void CAlnVecRow::x_RenderAveragableScores(CGlPane& pane, IAlnSegmentIterator& it_seg,
                                             CScoreCache::TScoreColl::const_iterator it,
                                             const CScoreCache::TScoreColl::const_iterator it_end,
                                             int vp_top_y, int vp_bottom_y)
{
    CGlRender& gl = GetGl();

    const TVPRect& rc_vp = pane.GetViewport();
    const TModelRect rc_vis = pane.GetVisibleRect();

    int row_start = m_Handle.GetSeqAlnStart();
    int row_stop = m_Handle.GetSeqAlnStop();

    typedef CRasterizer<SScPix2>    TR;
    TR rasterizer(rc_vp.Left(), rc_vp.Right(), rc_vis.Left(), rc_vis.Right());

    while(it_seg  &&  it != it_end  &&  it->GetFrom() <= rc_vis.Right()) {
        // get the intersection of the score and align intervals
        const IAlnSegment& seg = *it_seg;
        const TSignedSeqRange& r = seg.GetAlnRange();
        int from = std::max<int>(r.GetFrom(), it->GetFrom());
        int to = std::min<int>(r.GetTo(), it->GetTo());

        /// apply attributes to the intersection
        SScPix2 pix = { 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, false };
        bool empty_seg = (from < row_start  ||  to > row_stop);
        if(seg.IsAligned())  {
            pix.m_IntLen = 1.0;
            pix.m_IntScore = it->GetAttr();
        } else if( ! empty_seg) {
            pix.m_GapLen = 1.0;
            pix.m_GapScore = it->GetAttr();
        }
        rasterizer.AddInterval(from, to + 1, pix, FSetPix2);

        /// advance iterators
        if(to == r.GetTo()) {
            ++it_seg;
        }
        if(to == (int) it->GetTo())   {
            ++it;
        }
    }

    // Rendering
    gl.PolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    const TR::TRaster& raster = rasterizer.GetRaster();

    int shift_x = rc_vp.Left();
    static const float eps = 0.001f;

    for( size_t i = 0;  i < raster.size(); )    {
        size_t start = i;
        const SScPix2& sc_pix = raster[i];
        while(i < raster.size()  &&  raster[i] == sc_pix) {
            i++;
        }

        if(sc_pix.m_IntLen > eps) {
            float sc = sc_pix.m_IntScore / sc_pix.m_IntLen;
            const CRgbaColor& color =
                m_Cache->GetColorForScore(sc, IScoringMethod::fBackground);
            double alpha = 0.25 + sc_pix.m_IntLen * 0.75;
            gl.Color4d(color.GetRed(), color.GetGreen(), color.GetBlue(), alpha);

            gl.Rectd(start + shift_x, vp_top_y, i + shift_x, vp_bottom_y + 1);

            if(sc_pix.m_Init)   {
                CRgbaColor color =
                    m_Cache->GetColorForScore(sc_pix.m_MinScore, IScoringMethod::fBackground);
                gl.Color4d(color.GetRed(), color.GetGreen(), color.GetBlue(), 1.0);

                TVPUnit y_c = (vp_top_y + vp_bottom_y ) / 2;
                gl.Rectd(start + shift_x, vp_top_y, i + shift_x, y_c);
            }
        }
        if(sc_pix.m_GapLen > eps) {
            float sc = sc_pix.m_GapScore / sc_pix.m_GapLen;
            const CRgbaColor& color =
                m_Cache->GetColorForScore(sc, IScoringMethod::fBackground);
            double alpha = 0.25 + sc_pix.m_GapLen * 0.75;
            gl.Color4d(color.GetRed(), color.GetGreen(), color.GetBlue(), alpha);

            gl.Rectd(start + shift_x, vp_top_y - kGapOffset,
                    i + shift_x, vp_bottom_y + 1 + kGapOffset);


            if(sc_pix.m_Init)   {
                CRgbaColor color =
                    m_Cache->GetColorForScore(sc_pix.m_MinScore, IScoringMethod::fBackground);
                gl.Color4d(color.GetRed(), color.GetGreen(), color.GetBlue(), 1.0);

                TVPUnit y_c = (vp_top_y + vp_bottom_y ) / 2;
                gl.Rectd(start + shift_x, vp_top_y - kGapOffset, i + shift_x, y_c);
            }

        }
    }
}


void CAlnVecRow::x_RenderNonAveragableScores(CGlPane& pane, IAlnSegmentIterator& it_seg,
                                             CScoreCache::TScoreColl::const_iterator it,
                                             const CScoreCache::TScoreColl::const_iterator it_end,
                                             int vp_top_y, int vp_bottom_y)
{
    CGlRender& gl = GetGl();

    const TVPRect& rc_vp = pane.GetViewport();
    const TModelRect rc_vis = pane.GetVisibleRect();

    //LOG_POST("CAlnVecRow::x_RenderNonAveragableScores()  "  << rc_vis.Left() << " " << rc_vis.Right());
    int row_start = m_Handle.GetSeqAlnStart();
    int row_stop = m_Handle.GetSeqAlnStop();

    typedef CRasterizer<SScPix>    TR;
    TR rasterizer(rc_vp.Left(), rc_vp.Right(), rc_vis.Left(), rc_vis.Right());
    SScPix pix = { 0.0f, 0.0f, false };

    while(it_seg  &&  it != it_end  &&  it->GetFrom() <= rc_vis.Right()) {
        // get the intersection of the score and align intervals
        const IAlnSegment& seg = *it_seg;
        const TSignedSeqRange& r = seg.GetAlnRange();
        int from = std::max<int>(r.GetFrom(), it->GetFrom());
        int to = std::min<int>(r.GetTo(), it->GetTo());

        /// apply attributes to the intersection
        bool empty_seg = (from < row_start  ||  to > row_stop);
        if(! empty_seg) {
            pix.m_Score = it->GetAttr();
            pix.m_Gap = ! seg.IsAligned();

            rasterizer.AddInterval(from, to + 1, pix, FSetPixMaxCoverage);
        }

        /// advance iterators
        if(to == r.GetTo()) {
            ++it_seg;
        }
        if(to == (int) it->GetTo())   {
            ++it;
        }
    }

    // Rendering
    gl.PolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    const TR::TRaster& raster = rasterizer.GetRaster();

    static float eps = 0.0001f;

    int shift_x = rc_vp.Left();
    for( size_t i = 0;  i < raster.size(); )    {
        size_t start = i;
        const SScPix& sc_pix = raster[i];
        while(++i < raster.size()  &&  raster[i] == sc_pix) ;

        if(sc_pix.m_MaxCoverage > eps)    {
            const CRgbaColor& color =
                m_Cache->GetColorForScore(sc_pix.m_Score, IScoringMethod::fBackground);
            gl.ColorC(color);

            int off_y = sc_pix.m_Gap ? kGapOffset : 0;
            gl.Rectd(start + shift_x, vp_top_y - off_y, i + shift_x, vp_bottom_y + 1 + off_y);
            //LOG_POST("Start " << start << ", end " << i << ", color " << color.ToString()
            //         << ",  c_pix.m_MaxCoverage " << sc_pix.m_MaxCoverage
            //         << ", sc_pix.m_Score " << sc_pix.m_Score);
        }
    }
}

#define SEQ_OFF_VERT 0
#define SEQ_OFF_HORZ 0


/// Checks if sequence can be drawn.

bool CAlnVecRow::x_CanRenderSeq(CGlPane& pane)
{
    if(m_Style)    {
        double min_w = 1.0 / m_Style->GetSeqFont().GetMetric(CGlBitmapFont::eMetric_MaxCharWidth);
        double scale_x = pane.GetScaleX();
        return min_w >= scale_x;
    } else return false;
}


bool IfDash(char ch) {   return ch == '-';  }


// renders a row of seq symbols in the range [Start, Stop]
void CAlnVecRow::x_RenderSequence(CGlPane& pane, TModelUnit top_y, TModelUnit bottom_y,
                                  int start, int stop)
{
    _ASSERT(m_Style);
    CGlRender& gl = GetGl();

    string seq;
    m_Handle.GetAlnSeqString(seq, TSignedRange(start, stop));

    // replace gap symbols with spaces
    replace_if(seq.begin(), seq.end(), IfDash, ' ');

    const CGlBitmapFont& font = m_Style->GetSeqFont();

    TModelUnit x = start + 0.5 - pane.GetOffsetX();
    TModelUnit sc_x = pane.GetScaleX();
    TModelUnit H = m_Style->GetSeqFont().TextHeight();
    TModelUnit y = floor((top_y + bottom_y + H) / 2);

    bool render_scores = false;
    if(m_Cache  &&  m_Cache->HasScores())   {
        IScoringMethod* method = m_Cache->GetScoringMethod();
        render_scores = method  &&
            (method->GetSupportedColorTypes() & IScoringMethod::fForeground);
    }

    if(render_scores)    {
        /// need to draw colored sequence on colored background
        const CScoreCache::TScoreColl& ScoreColl = m_Cache->GetScores(m_Handle.GetRowNum());
        CScoreCache::TScoreColl::const_iterator it = ScoreColl.begin();
        const CScoreCache::TScoreColl::const_iterator it_end = ScoreColl.end();
        it = ScoreColl.find(it, it_end, start);

        int len = stop - start + 1;
        _ASSERT(len >= 0);
        vector<CRgbaColor*> colors(len); // create vector of the required length

        for( ; it != it_end; ++it)    {
            TSeqPos from = max(it->GetFrom(), (TSeqPos) start);
            TSeqPos to = min(it->GetTo(), (TSeqPos) stop);
            if(to >= from)   {
                IScoringMethod::TScore score = it->GetAttr();
                CRgbaColor& cl = const_cast<CRgbaColor&>(
                    m_Cache->GetColorForScore(score, IScoringMethod::fForeground));

                std::fill_n(colors.begin() + (from - start), to - from + 1, &cl);
            }
        }

        font.ArrayTextOut(x, y, 1.0, 0.0, seq.c_str(), &colors, sc_x, 0.0);
    } else {
        /// simply render the sequence
        gl.ColorC(m_Style->GetColor(CWidgetDisplayStyle::eSequence));
        font.ArrayTextOut(x, y, 1.0f, 0.0f, seq.c_str(), NULL, sc_x, 0.0);
    }
}

static const int kTextOffsetY = 0;
static const int kTextOffsetX = 2;
static const int kPosOffsetX = 4;


void CAlnVecRow::x_RenderDescr(CGlPane& pane, int state)
{
    _ASSERT(m_Style);
    x_RenderTextCell(pane, m_Text, state, IGlFont::eAlign_Left);
}

void CAlnVecRow::x_RenderTaxLabel(CGlPane& pane, int state)
{
    _ASSERT(m_Style);
    x_RenderTextCell(pane, m_TaxLabel, state, IGlFont::eAlign_Left);
}


void CAlnVecRow::x_RenderGraph(CGlPane& pane, const TSignedRange& range)
{
    if(IsExpanded()) {
        if(m_GraphCont  &&  m_GraphCont->IsCreated())   {
            //LOG_POST(Info << "CAlnVecRow::x_RenderGraph");
            //LOG_POST(Info << "    Viewport " << pane.GetViewport().ToString() <<
            //         "\n    Visible " << pane.GetVisibleRect().ToString());

            IAlnSegmentIterator::EFlags flags = IAlnSegmentIterator::eSkipInserts;
            auto_ptr<IAlnSegmentIterator> p_it_3(m_Handle.CreateSegmentIterator(range, flags));

            // setup CGlPane for the graph
            CGlPane gr_pane(pane);
            TVPRect rc_vp = gr_pane.GetViewport();

            // set clipping
            TVPRect rc_clip_gr = rc_vp;
            const TVPRect* rc_clip  = pane.GetClipRect();
            if(rc_clip) {
                rc_clip_gr.IntersectWith(*rc_clip);
            }
            gr_pane.SetClipRect(&rc_clip_gr);

            // setup Graph Container
            rc_vp.SetTop(rc_vp.Top() - m_BaseHeight);

            TModelRect rc_vis = gr_pane.GetVisibleRect();
            rc_vis.SetTop(rc_vis.Top() + m_BaseHeight);

            m_GraphCont->SetVPRect(rc_vp);
            m_GraphCont->SetModelRect(rc_vis);

            m_GraphCont->Render(pane, *p_it_3);
        }
    }
}


const static int kButtonOffset = 2;
const static int kButtonSize = 11;
const static int kButtonSignSize = 7;


// Returns buttons bounds in screen coordinates, "pane" parameter specifies
// CGlPane corresponding to "Icons" column
TVPRect CAlnVecRow::x_GetButtonRect(CGlPane& pane, EButtons btn, int shift)
{
    const TVPRect& rc_vp = pane.GetViewport();

    int btn_space = kButtonSize + kButtonOffset;
    int x = rc_vp.Left() + kButtonOffset + ((int)btn * btn_space * shift);
    int y_top = rc_vp.Top() - (m_BaseHeight - kButtonSize) / 2;

    TVPRect rc(x, y_top - kButtonSize + 1, x + kButtonSize - 1, y_top);
    return rc;
}


CAlnVecRow::EButtons
    CAlnVecRow::x_HitTestButton(CGlPane& pane, const TVPPoint& vp_pt)
{
    const TVPRect& rc_vp = pane.GetViewport();
    //int y_top = rc_vp.Top() - (m_BaseHeight - kButtonSize) / 2;

    //if(vp_pt.Y() <= y_top  &&  vp_pt.Y() >= y_top - kButtonSize + 1)    { // vertical hit
        int left = rc_vp.Left() + kButtonOffset;
        for( int i = 0; i < 3; i++ )    {
            if(vp_pt.X() >= left  &&  vp_pt.X() < left + kButtonSize) {
                return (EButtons) i;
            }
            left += kButtonSize + kButtonOffset;
        }
    //}
    return eInvalid;
}


void CAlnVecRow::x_RenderIcons(CGlPane& pane, int /*state*/)
{
    _ASSERT(m_Style);

    CGlRender& gl = GetGl();

    pane.OpenPixels();

    // draw strand marker
    TVPRect rc_btn = x_GetButtonRect(pane, eStrandMarker);
    int y_c = rc_btn.CenterPoint().Y();
    int half = rc_btn.Height() / 2;
    int x1 = rc_btn.Left();
    int x2 = rc_btn.Right();

    if(m_Handle.IsNegativeStrand()) {
        swap(x1, x2);
    }

    gl.Disable(GL_LINE_SMOOTH);
    gl.PolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    gl.Color3d(0.0, 0.5, 0.0);
    gl.Begin(GL_TRIANGLES); // render triangle
        gl.Vertex2d(x1, y_c - half);
        gl.Vertex2d(x2, y_c);
        gl.Vertex2d(x1, y_c + half);
    gl.End();

    // draw "Expand" button

    gl.LineWidth(1.0f);
    gl.ColorC(m_Style->GetColor(CWidgetDisplayStyle::eText));
    gl.PolygonMode(GL_FRONT_AND_BACK, GL_LINE);

    rc_btn = x_GetButtonRect(pane, eExpand);
    gl.Rectd(rc_btn.Left(), rc_btn.Bottom(), rc_btn.Right(), rc_btn.Top());

    if(IsExpandable())    {
        int offset = (kButtonSize - kButtonSignSize) / 2;
        TVPPoint pt_c = rc_btn.CenterPoint();

        // render "+" or "-"
        gl.Begin(GL_LINES);
            gl.Vertex2d(rc_btn.Left() + offset, pt_c.Y());
            gl.Vertex2d(rc_btn.Left() + offset + kButtonSignSize, pt_c.Y());
            if(! m_bExpanded)   {
                gl.Vertex2d(pt_c.X(), rc_btn.Bottom() + offset);
                gl.Vertex2d(pt_c.X(), rc_btn.Bottom() + offset + kButtonSignSize);
            }
        gl.End();
    }
    /**
    if(m_GraphCont  &&  m_GraphCont->HasData())   {   // draw "Graph" button
        rc_btn = x_GetButtonRect(pane, eSetupGraphs);
        gl.Rectd(rc_btn.Left(), rc_btn.Bottom(), rc_btn.Right(), rc_btn.Top());

        m_Style->GetTextFont().TextOut(rc_btn.Left() + 1, rc_btn.Bottom() + 1,
                                        kButtonSize - 2, kButtonSize - 2, "...", IGlFont::eAlign_Center);
    }
    **/

    pane.Close();
}


void CAlnVecRow::x_RenderIconStrand(CGlPane& pane, int /*state*/)
{
    _ASSERT(m_Style);

    CGlRender& gl = GetGl();

    pane.OpenPixels();

    // draw strand marker
    TVPRect rc_btn = x_GetButtonRect(pane, eExpand);
    int y_c = rc_btn.CenterPoint().Y();
    int half = rc_btn.Height() / 2;
    int x1 = rc_btn.Left();
    int x2 = rc_btn.Right();

    if(m_Handle.IsNegativeStrand()) {
        swap(x1, x2);
    }

    gl.PolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    gl.Color3d(0.0, 0.5, 0.0);
    gl.Begin(GL_TRIANGLES); // render triangle
        gl.Vertex2d(x1, y_c - half);
        gl.Vertex2d(x2, y_c);
        gl.Vertex2d(x1, y_c + half);
    gl.End();


    pane.Close();
}

void CAlnVecRow::x_RenderIconExpand(CGlPane& pane, int state)
{
    _ASSERT(m_Style);

    CGlRender& gl = GetGl();

    pane.OpenPixels();

    // draw "Expand" button
    TVPRect rc_btn = x_GetButtonRect(pane, eStrandMarker);

    gl.ColorC(m_Style->GetColor(CWidgetDisplayStyle::eText));
    gl.PolygonMode(GL_FRONT_AND_BACK, GL_LINE);
    gl.LineWidth(1.0f);     
    gl.Disable(GL_LINE_SMOOTH);

    rc_btn = x_GetButtonRect(pane, eExpand, 0);
    gl.Rectd(rc_btn.Left(), rc_btn.Bottom(), rc_btn.Right(), rc_btn.Top());

    if(IsExpandable())    {
        int offset =(kButtonSize - kButtonSignSize) / 2;
        TVPPoint pt_c = rc_btn.CenterPoint();

        // render "+" or "-"
        gl.Begin(GL_LINES);
            gl.Vertex2d(rc_btn.Left() + offset, pt_c.Y());
            gl.Vertex2d(rc_btn.Left() + offset + kButtonSignSize, pt_c.Y());
            if(! m_bExpanded)   {
                gl.Vertex2d(pt_c.X(), rc_btn.Bottom() + offset);
                gl.Vertex2d(pt_c.X(), rc_btn.Bottom() + offset + kButtonSignSize);
            }
        gl.End();
    }
    /**
    if(m_GraphCont  &&  m_GraphCont->HasData())   {   // draw "Graph" button
        rc_btn = x_GetButtonRect(pane, eSetupGraphs);
        gl.Rectd(rc_btn.Left(), rc_btn.Bottom(), rc_btn.Right(), rc_btn.Top());

        m_Style->GetTextFont().TextOut(rc_btn.Left() + 1, rc_btn.Bottom() + 1,
                                        kButtonSize - 2, kButtonSize - 2, "...", IGlFont::eAlign_Center);
    }
    **/

    pane.Close();

}


void CAlnVecRow::x_RenderStartPos(CGlPane& pane, int state)
{
    const TModelRect& rc_v = pane.GetVisibleRect();
    int aln_pos = (int) floor(rc_v.Left());

    x_RenderAlnPos(pane, aln_pos, IAlnExplorer::eRight, state);
}


void CAlnVecRow::x_RenderEndPos(CGlPane& pane, int state)
{
    const TModelRect& rc_v = pane.GetVisibleRect();
    int aln_pos = -1 + (int) ceil(rc_v.Right());

    x_RenderAlnPos(pane, aln_pos, IAlnExplorer::eLeft, state);
}


void CAlnVecRow::x_RenderSeqStart(CGlPane& pane, int state)
{
    TSeqPos pos = m_Handle.IsNegativeStrand() ?
                  m_Handle.GetSeqStop() : m_Handle.GetSeqStart();

    string s_num = CTextUtils::FormatSeparatedNumber(pos + 1);

    x_RenderTextCell(pane, s_num, state, IGlFont::eAlign_Right);
}


void CAlnVecRow::x_RenderSeqEnd(CGlPane& pane, int state)
{
    TSeqPos pos = m_Handle.IsNegativeStrand() ?
                  m_Handle.GetSeqStart() : m_Handle.GetSeqStop();

    string s_num = CTextUtils::FormatSeparatedNumber(pos + 1);

    x_RenderTextCell(pane, s_num, state, IGlFont::eAlign_Right);
}


void CAlnVecRow::x_RenderSeqLength(CGlPane& pane, int state)
{
    TSeqPos pos = m_Handle.GetBioseqHandle().GetBioseqLength() -1;
    string s_num = CTextUtils::FormatSeparatedNumber(pos + 1);

    x_RenderTextCell(pane, s_num, state, IGlFont::eAlign_Right);
}


// generates text label for given position in alignment
void CAlnVecRow::x_RenderAlnPos(CGlPane& pane, TSeqPos aln_pos,
                                IAlnExplorer::ESearchDirection dir, int state)
{
    const char *s_marker = "";

    int seq_pos = m_Handle.GetSeqPosFromAlnPos(aln_pos);
    if(seq_pos == -1)   {
        seq_pos = m_Handle.GetSeqPosFromAlnPos(aln_pos, dir);
        switch(dir) {
        case IAlnExplorer::eRight:  s_marker = ">>  "; break;
        case IAlnExplorer::eLeft:  s_marker = "<<  "; break;
        default: break;
        }
    }

    string s_num = CTextUtils::FormatSeparatedNumber(seq_pos + 1);
    string s_label(s_marker);
    s_label += s_num;

    x_RenderTextCell(pane, s_num, state, IGlFont::eAlign_Right);
}


void CAlnVecRow::x_RenderTextCell(CGlPane& pane, const string& text, int state,
                                  IGlFont::TAlign align)
{
    _ASSERT(m_Style);

    CGlRender& gl = GetGl();

    pane.OpenPixels();
    TVPRect rc_vp = pane.GetViewport();

    bool selected = (state & fItemSelected) != 0;
    bool wid_focused = (state & fWidgetFocused) != 0;

    if(selected) {         //draw background
        gl.ColorC(m_Style->GetColor(wid_focused ? CWidgetDisplayStyle::eFocusedBack
                                                : CWidgetDisplayStyle::eSelectedBack));
        gl.PolygonMode(GL_FRONT_AND_BACK, GL_FILL);
        gl.Rectd(rc_vp.Left(), rc_vp.Bottom() - 1, rc_vp.Right(), rc_vp.Top() + 1);
    } else {
        x_RenderEmptyBox(pane);
    }

    // draw text label
    gl.ColorC(m_Style->GetColor(selected ? CWidgetDisplayStyle::eSelectedText
                                         : CWidgetDisplayStyle::eText));

    TModelUnit x = rc_vp.Left() + kTextOffsetX * 2;
    TModelUnit y = rc_vp.Top() - m_BaseHeight + 1;
    TModelUnit w = rc_vp.Width() - kTextOffsetX * 4;
    const CGlBitmapFont& font = m_Style->GetTextFont();
    font.TextOut(x, y, x + w -1, y + m_BaseHeight - 1, text.c_str(), align);

    // draw focus frame
    if(wid_focused  &&  (state & fItemFocused)) {
        gl.PolygonMode(GL_FRONT_AND_BACK, GL_LINE);
        gl.LineWidth(1.0f);

        gl.Begin(GL_LINE_LOOP);
            gl.Vertex2d(rc_vp.Left(), rc_vp.Bottom());
            gl.Vertex2d(rc_vp.Right(), rc_vp.Bottom());
            gl.Vertex2d(rc_vp.Right(), rc_vp.Top());
            gl.Vertex2d(rc_vp.Left(), rc_vp.Top());
        gl.End();
    }

    pane.Close();
}


void CAlnVecRow::x_RenderEmptyBox(CGlPane& pane)
{
    CGlRender& gl = GetGl();

    gl.ColorC(m_Style->GetColor(CWidgetDisplayStyle::eFrame));
    gl.PolygonMode(GL_FRONT_AND_BACK, GL_LINE);
    TVPRect rc_vp = pane.GetViewport();
    gl.LineWidth(1.0);
    gl.Rectd(rc_vp.Left(), rc_vp.Bottom(), rc_vp.Right(), rc_vp.Top());
}


bool    CAlnVecRow::IsExpandable() const
{
    //return m_GraphCont->HasData();
    return true; //TODO HACK
}


bool    CAlnVecRow::IsExpanded() const
{
    return m_bExpanded;
}


void CAlnVecRow::Expand(bool b_exp)
{
    if(IsExpandable()  &&  IsExpanded() != b_exp)   {
        if(b_exp) {
            m_GraphCont->Create();
        } else {
            m_GraphCont->Destroy();
        }
        m_bExpanded = b_exp;

        x_UpdateRowHeight();
        if (x_GetHost()) {
            x_GetHost()->ARH_OnRowChanged(this);
        }
    }
}


void CAlnVecRow::x_UpdateRowHeight()
{
    m_Height = m_BaseHeight;

    if(m_GraphCont  &&  m_GraphCont->IsCreated())  {
        m_Height += m_GraphCont->PreferredSize().Y();
    }
}

const IAlnRowGraphProperties*     CAlnVecRow::GetProperties() const
{
    return m_GraphCont ? m_GraphCont->GetProperties() : NULL;
}


void CAlnVecRow::SetProperties(IAlnRowGraphProperties* props)
{
    if(m_GraphCont)   {
        m_GraphCont->SetProperties(props);

        m_Height = m_BaseHeight;
        if(m_GraphCont->IsCreated())  {
            m_Height += m_GraphCont->PreferredSize().Y();
        }
    }
}


TSeqPos CAlnVecRow::GetColumnAsSeqPos(TColumnType type, const TModelRect& rc_vis)
{
    TSeqPos pos = (TSeqPos) -1;
    switch(type)    {
    case eStart:    {
        int aln_pos = (int) floor(rc_vis.Left());
        pos = m_Handle.GetSeqPosFromAlnPos(aln_pos, IAlnExplorer::eRight);
        break;
    }
    case eEnd:  {
        int aln_pos = -1 + (int) ceil(rc_vis.Right());
        pos = m_Handle.GetSeqPosFromAlnPos(aln_pos, IAlnExplorer::eLeft);
        break;
    }
    case eSeqStart:
        pos = m_Handle.IsNegativeStrand() ?
                m_Handle.GetSeqStop() : m_Handle.GetSeqStart();
        break;
    case eSeqEnd:
        pos = m_Handle.IsNegativeStrand() ?
                m_Handle.GetSeqStart() : m_Handle.GetSeqStop();
        break;
    case eSeqLength:
        pos = m_Handle.GetBioseqHandle().GetBioseqLength() - 1;
        break;
    default:
        break;
    }
    return pos;
}


string CAlnVecRow::GetColumnAsString(TColumnType type)
{
    switch(type)    {
    case eDescr:
        return m_Text;
    case eTaxLabel:
        return m_TaxLabel;
    default:
        return "";
    }
}

string& CAlnVecRow::GetAlnSeqString( string &buffer, const IAlnExplorer::TSignedRange &aln_rng ) const
{
	return m_Handle.GetAlnSeqString( buffer, aln_rng );
}

string& CAlnVecRow::GetRowStatistics( string &buffer ) const
{
    TSignedSeqPos alstart = m_Handle.GetSeqAlnStart();
    TSignedSeqPos alstop = m_Handle.GetSeqAlnStop();

    TSignedRange alrange;
    if( alstart < alstop ){
        alrange.Set( alstart, alstop );
    } else {
        alrange.Set( alstop, alstart );
    }
    IAlnSegmentIterator::EFlags flags = IAlnSegmentIterator::eAllSegments;
    auto_ptr<IAlnSegmentIterator> p_it( m_Handle.CreateSegmentIterator( alrange, flags ) );

    int aligned = 0;
    int gap = 0;
    int reversed = 0;
    int indel = 0;
    int unaligned = 0;

    for( IAlnSegmentIterator& it(*p_it); it; ++it ){
        const IAlnSegment& seg = *it;
        const TSignedSeqRange& segran = seg.GetAlnRange();

        int type = seg.GetType();

        if( type & IAlnSegment::fAligned ){
            aligned += segran.GetLength();
        }
        if( type & IAlnSegment::fGap ){
            gap += segran.GetLength();
        }
        if( type & IAlnSegment::fReversed ){
            reversed += segran.GetLength();
        }
        if( type & IAlnSegment::fIndel ){
            indel += segran.GetLength();
        }
        if( type & IAlnSegment::fUnaligned ){
            unaligned += segran.GetLength();
        }
    }

    int start = m_Handle.GetSeqStart();
    int stop = m_Handle.GetSeqStop();

    buffer.clear();

    buffer += std::string( "Sequence is aligned from " ) + NStr::IntToString( start ) + " to " + NStr::IntToString( stop ) + " positions";
    buffer += string( "\nBases aligned: " ) + NStr::IntToString( aligned );
    buffer += string( "\nBases through gaps: " ) + NStr::IntToString( gap );
    buffer += string( "\nBases in indels: " ) + NStr::IntToString( indel );
    buffer += string( "\nBases unaligned: " ) + NStr::IntToString( unaligned );

    return buffer;
}

void CAlnVecRow::x_OnGraphContentChanged(CEvent* evt)
{
    //LOG_POST("CAlnVecRow::x_OnGraphContentChanged");
    x_UpdateRowHeight();
    m_Host->ARH_OnRowChanged(this);
}


void CAlnVecRow::x_OnGraphSizeChanged(CEvent* evt)
{
    //LOG_POST("CAlnVecRow::x_OnGraphSizeChanged");
    x_UpdateRowHeight();
    m_Host->ARH_OnRowChanged(this);
}


END_NCBI_SCOPE