/*  $Id: alnvec_multi_ds.cpp 19605 2009-07-14 17:41:39Z wuliangs $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Andrey Yazhuk
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>

#include <gui/widgets/aln_multiple/alnvec_multi_ds.hpp>

#include <gui/objutils/utils.hpp>
#include <gui/utils/event_translator.hpp>

#include <objects/seqloc/Seq_loc.hpp>
#include <objects/seqloc/Seq_interval.hpp>
#include <objects/seq/Bioseq.hpp>

#include <objmgr/align_ci.hpp>




BEGIN_NCBI_SCOPE
USING_SCOPE(ncbi::objects);


CAlnVecMultiDataSource::CAlnVecMultiDataSource(CScope& scope)
:   m_Scope(&scope), m_ConsRowIndex(-1),
    m_isDataReadSync(false)
{
}


CAlnVecMultiDataSource::~CAlnVecMultiDataSource()
{
    x_ClearHandles();

    try {
        if(x_IsJobRunning())   {
            x_DeleteJob();
        }
    }
    catch (CException& e) {
        LOG_POST(Error << "CAlnVecMultiDataSource::~CAlnVecMultiDataSource(): "
            << "failed to delete job: " << e.GetMsg());
    }
}


BEGIN_EVENT_MAP(CAlnVecMultiDataSource, CEventHandler)
    ON_MESSAGE(CAppJobNotification, CAppJobNotification::eStateChanged,
               &CAlnVecMultiDataSource::OnAppJobNotification)
    ON_MESSAGE(CAppJobNotification, CAppJobNotification::eProgress,
               &CAlnVecMultiDataSource::OnAppJobNotification)
END_EVENT_MAP()


int CAlnVecMultiDataSource::GetConsensusRow() const
{
    return m_ConsRowIndex;
}


void CAlnVecMultiDataSource::Init(const objects::CSeq_annot& annot, bool sync)
{
    vector< CConstRef<CSeq_align> > aligns;

    if(annot.GetData().IsAlign() ) {
        ITERATE (CSeq_annot::TData::TAlign, it, annot.GetData().GetAlign()) {
            aligns.push_back(CConstRef<CSeq_align>(*it));
        }
    }
    Init(aligns, sync);
}


void CAlnVecMultiDataSource::Init(const objects::CBioseq_Handle& handle, bool sync)
{
    vector< CConstRef<CSeq_align> > aligns;

    SAnnotSelector sel =
        CSeqUtils::GetAnnotSelector(CSeq_annot::TData::e_Align);
    CAlign_CI it(handle, sel);
    for ( ;  it;  ++it) {
        aligns.push_back(CConstRef<CSeq_align>(&*it));
    }
    Init(aligns, sync);
}


void CAlnVecMultiDataSource::Init(const vector< CConstRef<CSeq_align> >& aligns,
                                  bool sync)
{
    _ASSERT( ! m_Job);
    CAppJobDispatcher& disp = CAppJobDispatcher::GetInstance();

    m_Job.Reset(new CBuildAlnVecJob(aligns, *m_Scope));
    m_isDataReadSync = sync;

    if(sync)    {
        // do everything synchronously
        m_JobID = -1;
        int type = CDataChangeNotifier::eError;

        if(m_Job->Run())    {
            type = CDataChangeNotifier::eChanged;
            CBuildAlnVecResult* res = dynamic_cast<CBuildAlnVecResult*>(m_Job->GetResult().GetPointer());
            if (!res) {
                NCBI_THROW(CCoreException, eNullPtr, "Cannot get merge results: AlnVec is empty");
            }
            _ASSERT(res);
            x_Assign(*res->m_AlnVec);
        }

        //m_HasScores = x_TransferResults(m_Job->GetResult().GetPointer());
        m_Job.Reset();

        CUpdate up(type, "");
        NotifyListener(up);
    } else {
        // use CAppJobDispatcher to execute background jobs
        m_JobID = disp.StartJob(*m_Job, "ThreadPool", *this, 1, true);
    }
}


bool CAlnVecMultiDataSource::IsEmpty() const
{
    return m_AlnVec.GetPointer() == NULL;
}


void CAlnVecMultiDataSource::OnAppJobNotification(CEvent* evt)
{
    CAppJobNotification* notn = dynamic_cast<CAppJobNotification*>(evt);
    _ASSERT(notn);

    if(notn)    {
        int job_id = notn->GetJobID();
        if(m_JobID != job_id) {
            ERR_POST("CAlnVecMultiDataSource::OnAppJobNotification() - unknown Job ID " << job_id);
        } else {
            switch(notn->GetState())    {
            case IAppJob::eCompleted:   {
                CBuildAlnVecResult* res = dynamic_cast<CBuildAlnVecResult*>(notn->GetResult().GetPointer());
                _ASSERT(res);
                x_Assign(*res->m_AlnVec);
                CUpdate up(CDataChangeNotifier::eChanged, "");
                NotifyListener(up);
                m_Job.Reset();
                break;
            }
            case IAppJob::eFailed:  {
                CConstIRef<IAppJobError> error = notn->GetError();
                CUpdate up(CDataChangeNotifier::eError, error->GetText());
                NotifyListener(up);
            }
            case IAppJob::eCanceled:    {
                CUpdate up(CDataChangeNotifier::eError, "Canceled");
                NotifyListener(up);
                break;
            }
            case IAppJob::eRunning: {
                CConstIRef<IAppJobProgress> progressSoFar = notn->GetProgress();
                if (progressSoFar) {
                    CUpdate up( CDataChangeNotifier::eProgressReport,
                        progressSoFar->GetText(), progressSoFar->GetNormDone());
                    NotifyListener(up);
                }
                break;
            }
            default:
                _ASSERT(false);
            }
        }
    }
}


void CAlnVecMultiDataSource::x_Assign(objects::CAlnVec& aln_vec)
{
    x_ClearHandles();

    m_AlnVec.Reset(&aln_vec);
    SetGapChar('-');

    x_CreateHandles();
}


bool CAlnVecMultiDataSource::x_IsJobRunning()
{
    return m_Job.GetPointer() != NULL;
}


void CAlnVecMultiDataSource::x_DeleteJob()
{
    _ASSERT(m_Job  &&  m_JobID != -1);

    CAppJobDispatcher& disp = CAppJobDispatcher::GetInstance();
    disp.DeleteJob(m_JobID);

    m_JobID = -1;
    m_Job.Reset();

    //TODO update listener ?
}


void CAlnVecMultiDataSource::x_OnJobProgress(CAppJobNotification& notn)
{
    CConstIRef<IAppJobProgress> prg = notn.GetProgress();
    if(prg) {
        string text = prg->GetText();

        if(x_IsJobRunning())    {
            CUpdate up(eProgressReport, text);
            NotifyListener(up);
        }
    }
}


void CAlnVecMultiDataSource::SetGapChar(TResidue gap_char)
{
    m_AlnVec->SetGapChar(gap_char);
}


IAlnExplorer::EAlignType CAlnVecMultiDataSource::GetAlignType() const
{
    if( ! m_AlnVec) {
        return IAlnExplorer::fInvalid;
    }
    if(m_AlnVec->GetDenseg().IsSetWidths())   {
        return IAlnExplorer::fMixed;
    } else {
        const CBioseq& bioseq = *GetBioseqHandle(0).GetBioseqCore();
        if(bioseq.GetInst().GetMol() == CSeq_inst::eMol_aa) {
            return IAlnExplorer::fProtein;
        } else {
            return IAlnExplorer::fDNA;
        }
    }
}


CAlnVecMultiDataSource::TNumrow  CAlnVecMultiDataSource::GetNumRows (void) const
{
    return (TNumrow) m_AlnVec->GetNumRows();
}


TSeqPos  CAlnVecMultiDataSource::GetAlnStart (void) const
{
    return m_AlnVec->GetAlnStart();
}


TSeqPos  CAlnVecMultiDataSource::GetAlnStop (void) const
{
    return m_AlnVec->GetAlnStop();
}


TSeqPos  CAlnVecMultiDataSource::GetSeqStart(TNumrow row) const
{
    return m_AlnVec->GetSeqStart(row);
}


TSeqPos  CAlnVecMultiDataSource::GetSeqStop(TNumrow row) const
{
    return m_AlnVec->GetSeqStop(row);
}


bool  CAlnVecMultiDataSource::IsSetAnchor (void) const
{
    return m_AlnVec->IsSetAnchor();
}


CAlnVecMultiDataSource::TNumrow CAlnVecMultiDataSource::GetAnchor(void) const
{
    return m_AlnVec->GetAnchor();
}


bool CAlnVecMultiDataSource::CanChangeAnchor(void) const
{
    return true;
}


bool CAlnVecMultiDataSource::SetAnchor(TNumrow anchor)
{
    m_AlnVec->SetAnchor(anchor);
    return true;
}


bool CAlnVecMultiDataSource::UnsetAnchor (void)
{
    m_AlnVec->UnsetAnchor();
    return true;
}


bool CAlnVecMultiDataSource::CanGetId(TNumrow row) const
{
    return (row >= 0  &&  row < m_AlnVec->GetNumRows());
}


const CSeq_id&  CAlnVecMultiDataSource::GetSeqId(TNumrow row) const
{
    return m_AlnVec->GetSeqId(row);
}


const CBioseq_Handle &  CAlnVecMultiDataSource::GetBioseqHandle(TNumrow row) const
{
    return m_AlnVec->GetBioseqHandle(row);
}

bool CAlnVecMultiDataSource::IsPositiveStrand(TNumrow row) const
{
    return m_AlnVec->IsPositiveStrand(row);
}


bool CAlnVecMultiDataSource::IsNegativeStrand(TNumrow row) const
{
    return m_AlnVec->IsNegativeStrand(row);
}


TSignedSeqPos  CAlnVecMultiDataSource::GetAlnPosFromSeqPos(TNumrow for_row, TSeqPos aln_pos,
                                                        IAlnExplorer::ESearchDirection dir,
                                                        bool try_reverse_dir) const
{
    return m_AlnVec->GetAlnPosFromSeqPos(for_row, aln_pos,
                (CAlnVec::ESearchDirection) dir, try_reverse_dir);
}


TSignedSeqPos  CAlnVecMultiDataSource::GetSeqPosFromAlnPos(TNumrow row, TSeqPos seq_pos,
                                            IAlnExplorer::ESearchDirection dir,
                                            bool try_reverse_dir) const
{
    return m_AlnVec->GetSeqPosFromAlnPos(row, seq_pos, (CAlnVec::ESearchDirection) dir, try_reverse_dir);
}



string& CAlnVecMultiDataSource::GetAlnSeqString(TNumrow row, string &buffer,
                                                const IAlnExplorer::TSignedRange& aln_range) const
{
    return m_AlnVec->GetAlnSeqString(buffer, row, aln_range);
}


void  CAlnVecMultiDataSource::GetAlnFromSeq(TNumrow row,
                                            const TRangeColl& seq_coll,
                                            TRangeColl& aln_coll) const
{
    aln_coll.empty();

    TSeqPos seq_start = m_AlnVec->GetSeqStart(row);
    TSeqPos seq_stop = m_AlnVec->GetSeqStop(row);
    if(seq_start > seq_stop)    {
        swap(seq_start, seq_stop);
    }

    ITERATE(TRangeColl, it, seq_coll)  {
        // clip collection by alignment
        TSeqPos from = max(seq_start, it->GetFrom());
        TSeqPos to = min(seq_stop, it->GetTo());
        // translate
        TSeqPos aln_from = m_AlnVec->GetAlnPosFromSeqPos(row, from, CAlnVec::eRight);
        TSeqPos aln_to = m_AlnVec->GetAlnPosFromSeqPos(row, to, CAlnVec::eLeft);
        if(aln_from > aln_to)    {
            swap(aln_from, aln_to);
        }
        aln_coll.CombineWith(TSeqRange(aln_from, aln_to));
    }
}


void  CAlnVecMultiDataSource::GetSeqFromAln(TNumrow row,
                                            const TRangeColl& aln_coll,
                                            TRangeColl& seq_coll) const
{
    seq_coll.empty();

    TSeqPos aln_start = m_AlnVec->GetSeqAlnStart(row);
    TSeqPos aln_stop = m_AlnVec->GetSeqAlnStop(row);

    ITERATE(TRangeColl, it, aln_coll)  {
        // clip collection by alignment
        TSeqPos aln_from = max(aln_start, it->GetFrom());
        TSeqPos aln_to = min(aln_stop, it->GetTo());

        // translate
        TSeqPos from = m_AlnVec->GetSeqPosFromAlnPos(row, aln_from, CAlnVec::eRight);
        TSeqPos to = m_AlnVec->GetSeqPosFromAlnPos(row, aln_to, CAlnVec::eLeft);

        if(from > to)    {
            swap(from, to);
        }
        seq_coll.CombineWith(TSeqRange(from, to));
    }
}


const IAlignRowHandle*  CAlnVecMultiDataSource::GetRowHandle(TNumrow row) const
{
    return static_cast<const IAlignRowHandle*>(m_Handles[row]);
}


IAlnSegmentIterator*
    CAlnVecMultiDataSource::CreateSegmentIterator(TNumrow row, const IAlnExplorer::TSignedRange& range,
                                     IAlnSegmentIterator::EFlags flags) const
{
    const IAlignRowHandle* handle = GetRowHandle(row);
    return handle->CreateSegmentIterator(range, flags);
}


void CAlnVecMultiDataSource::x_ClearHandles()
{
    NON_CONST_ITERATE(THandleVector, it, m_Handles) {
        delete *it;
    }
    m_Handles.clear();
}


void CAlnVecMultiDataSource::x_CreateHandles()
{
    TNumrow row_n = m_AlnVec->GetNumRows();
    m_Handles.resize(row_n);
    for( TNumrow r = 0; r < row_n;  r++ )   {
        m_Handles[r] = new CAlnVecRowHandle(m_AlnVec, r);
    }
}


bool CAlnVecMultiDataSource::CanCreateConsensus()
{
    return true;
}


void CAlnVecMultiDataSource::CreateConsensus()
{
    if(m_AlnVec.NotEmpty()  &&  m_ConsRowIndex == -1)    {
        x_ClearHandles();

        CRef<CDense_seg> ds = m_AlnVec->CreateConsensus(m_ConsRowIndex);
        m_AlnVec.Reset(new CAlnVec(*ds, m_AlnVec->GetScope()));

        x_CreateHandles();
    }
}


END_NCBI_SCOPE
