/*  $Id: quality_methods.cpp 27622 2013-03-13 19:34:06Z katargir $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data,  the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties,  express or implied,  including
 *  warranties of performance,  merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Andrey Yazhuk
 */

#include <ncbi_pch.hpp>

#include <gui/widgets/aln_score/quality_methods.hpp>

#include <gui/widgets/aln_score/gradient_color_panel.hpp>

#include <gui/utils/table_names.hpp>

#include <gui/widgets/wx/wx_utils.hpp>

#include "wx/valgen.h"
#include "wx/choice.h"
#include "wx/checkbox.h"

BEGIN_NCBI_SCOPE

///////////////////////////////////////////////////////////////////////////////
///  CQualityScoringMethod
///
CQualityScoringMethod::CQualityScoringMethod()
:   m_Space(0), m_Gap(0)
{
    x_Init(sm_DefGradientSize);
}


CQualityScoringMethod::~CQualityScoringMethod()
{
}


CQualityScoringMethod::CQualityScoringMethod(int colors)
:    m_Options(0), m_Space(0), m_Gap(0)
{
    x_Init(colors);
}


void CQualityScoringMethod::x_Init(int colors)
{
    m_Worst.Set(1.0f, 0.0f, 0.0f);
    m_Best.Set(0.8f, 0.8f, 0.8f);
    SetOptions(fIgnoreEmptySpace);

    CreateColorTable(colors);
}


CQualityScoringMethod::CQualityScoringMethod(const CQualityScoringMethod& orig)
:   m_Worst(orig.m_Worst),
    m_Best(orig.m_Best),
    m_Options(orig.m_Options),
    m_Space(orig.m_Space),
    m_Gap(orig.m_Gap)
{
    m_ColorTable = orig.m_ColorTable;
}


void    CQualityScoringMethod::SetOptions(int options)
{
    m_Options = options;
    // for now, always ignore empty spaces.
    m_Space = ' '; // (m_Options & fIgnoreEmptySpace) ? ' ' : 0;
    m_Gap = (m_Options & fIgnoreGaps) ? '-' : 0;
}


int     CQualityScoringMethod::GetOptions() const
{
    return m_Options;
}


bool CQualityScoringMethod::IsScoreWholeColumn() const
{
    return (m_Options & fScoreWholeColumn) ? true : false;
}


void    CQualityScoringMethod::CreateColorTable(int size)
{
    _ASSERT(size >= 0);

    m_ColorTable.SetSize(size);
    m_ColorTable.FillGradient(0, size, m_Worst, m_Best);
}


#define QUALITY_DESCR "This scheme assigns scores to residues" \
" based on how well a particular residue agrees with the others in a column." \
" It is an implementation of ClustalX's alignment quality score described here: \n" \
" http://bips.u-strasbg.fr/en/Documentation/ClustalX/#Q"


string  CQualityScoringMethod::GetDescription() const
{
    return QUALITY_DESCR;
}


bool CQualityScoringMethod::IsAverageable() const
{
    return true;
}


void CQualityScoringMethod::CalculateScores(char /*cons*/,
                                            const string& column,
                                            TScore& col_score,
                                            TScoreVector& scores)
{
    _ASSERT(scores.size() == column.size());

    const string symbols(GetAllowedSymbols());

    size_t n_space = 0;
    vector<int> base_count(NCBI_FSM_DIM, 0);
    ITERATE(string, c_it, column) {
        char c = *c_it;
        if (c != m_Space  && c != m_Gap) {
            ++base_count[*c_it];
        } else {
            ++n_space;
        }
    }

    // size of the column w/o spaces we are not counting.
    size_t column_size = column.size() - n_space;
    vector<double> X(symbols.size(), 0);
    for (size_t r = 0; r <symbols.size(); ++r) {
        ITERATE(string, c_it, symbols) {
            X[r] += base_count[*c_it] * GetSubstitutionScore(*c_it, symbols[r]);
        }
        X[r] /= column_size;
    }

    double quality_score = 0;
    for (size_t i = 0; i < column.size(); ++i) {
        char c = column[i];
        if (c != m_Space  && c != m_Gap) {
            double score_sq_sum = 0;
            for (size_t r = 0; r < symbols.size(); ++r) {
                double score_diff = X[r] - GetSubstitutionScore(symbols[r], c);
                score_sq_sum += score_diff * score_diff;
            }
            TScore resid_score = (TScore) sqrt(score_sq_sum);
            if ( ! IsScoreWholeColumn()) {
                scores[i] = resid_score;
            }
            quality_score += resid_score;
        }
    }

    col_score = (TScore) (quality_score / column_size);

    if (IsScoreWholeColumn()) {
        for (size_t i = 0; i < column.size(); ++i) {
            char c = column[i];
            if (c != m_Space  && c != m_Gap) {
                scores[i] = col_score;
            }
        }
    }
}


bool CQualityScoringMethod::CanCalculateScores(const IScoringAlignment&)
{
    return true;
}


const CRgbaColor& CQualityScoringMethod::GetColorForNoScore(EColorType /*type*/) const
{
    static CRgbaColor not_used;
    _ASSERT(false);
    return not_used;
}


void CQualityScoringMethod::SetColorForNoScore(EColorType /*type*/,
                                               const CRgbaColor& /*color*/)
{
}


int     CQualityScoringMethod::GetSupportedColorTypes() const
{
    return fBackground;
}


const CRgbaColor& CQualityScoringMethod::GetColorForScore(TScore score,
                                                       EColorType type) const
{
    int size = (int)m_ColorTable.GetSize();
    _ASSERT(size >= 0);
    static const TScore kMaxScore = 15.0;
    score = (kMaxScore - score) / kMaxScore;
    if (score < 0.0) score = 0.0;

    if(type == fBackground) {
        int index = (int) floor(score * size);
        index = min(index, size - 1);
        return m_ColorTable[index];
    } else {
        _ASSERT(false); // not supported
        return m_ColorTable[0];
    }
}


const CMenuItem*    CQualityScoringMethod::GetMenu()
{
    return 0;
}


bool        CQualityScoringMethod::HasPropertiesPanel() const
{
    return true;
}

///////////////////////////////////////////////////////////////////////////////
/// CQualityScoringPanel

class CQualityScoringPanel : public wxPanel
{
    DECLARE_EVENT_TABLE()
public:
    CQualityScoringPanel(CQualityScoringMethod& method, wxWindow* parent, wxWindowID id = wxID_ANY);
protected:
    void CreateControls();
    void Init();
    void OnApply ( wxCommandEvent& event );
    CQualityScoringMethod& m_Method;
    wxChoice* m_MatrixChoice;
    CGradientColorPanel* m_GradPanel;
// data members
    bool m_ColScore;
    bool m_Gap;
};

BEGIN_EVENT_TABLE( CQualityScoringPanel, wxPanel )
    EVT_BUTTON( wxID_APPLY, CQualityScoringPanel::OnApply )
END_EVENT_TABLE()

CQualityScoringPanel::CQualityScoringPanel(
    CQualityScoringMethod& method, wxWindow* parent, wxWindowID id) :
        m_Method(method),
        m_MatrixChoice(), m_GradPanel()
{
    Init();
    Create(parent, id);
    CreateControls();
}

void CQualityScoringPanel::Init()
{
    int options = m_Method.GetOptions();
    m_ColScore = (options & CQualityScoringMethod::fScoreWholeColumn) != 0;
    m_Gap = (options & CQualityScoringMethod::fIgnoreGaps) != 0;
}

void CQualityScoringPanel::CreateControls()
{
    wxBoxSizer* itemBoxSizer1 = new wxBoxSizer(wxVERTICAL);
    this->SetSizer(itemBoxSizer1);

    CGradientColorPanel::stParams params;
    params.m_FirstColor = m_Method.m_Worst;
    params.m_LastColor = m_Method.m_Best;
    params.m_Steps = (int)m_Method.m_ColorTable.GetSize();
    params.m_ThreeColors = false;
    params.m_Reversable = true;

    m_GradPanel = new CGradientColorPanel(params, this);
    itemBoxSizer1->Add(m_GradPanel, 1, wxGROW|wxALL, 5);

    if (m_Method.GetType() & IAlnExplorer::fProtein) {
        CQualityScoringMethodAA* aa_method = dynamic_cast<CQualityScoringMethodAA*>(&m_Method);
        _ASSERT(aa_method);

        wxBoxSizer* itemBoxSizer2 = new wxBoxSizer(wxHORIZONTAL);
        itemBoxSizer1->Add(itemBoxSizer2, 0, wxALIGN_RIGHT|wxALL, 5);

        itemBoxSizer2->Add(new wxStaticText(this, wxID_STATIC, _("Substitution matrix:")),
                           0, wxALIGN_CENTER_VERTICAL|wxALL, 5);

        m_MatrixChoice = new wxChoice(this, wxID_ANY);
        itemBoxSizer2->Add(m_MatrixChoice, 0, wxALIGN_CENTER_VERTICAL|wxALL, 5);

        // add names to popup.
        vector<string> table_names = CTableNames::AllTableNames();
        ITERATE(vector<string>, iter, table_names) {
            m_MatrixChoice->Append(ToWxString(*iter));
        }

        // set the current name.
        string matrix_name = CTableNames::TableToName(aa_method->GetSubstitutionMatrix()).c_str();
        if (!matrix_name.empty()) {
            m_MatrixChoice->SetStringSelection(ToWxString(matrix_name));
        }
    }

    wxCheckBox* itemCheckBox1 = new wxCheckBox(this, wxID_ANY, _("Score column as a whole"));
    itemCheckBox1->SetValidator(wxGenericValidator(&m_ColScore));
    itemBoxSizer1->Add(itemCheckBox1, 0, wxALIGN_LEFT|wxALL, 5);

    wxCheckBox* itemCheckBox2 = new wxCheckBox(this, wxID_ANY, _("Ignore Gaps"));
    itemCheckBox2->SetValidator(wxGenericValidator(&m_Gap));
    itemBoxSizer1->Add(itemCheckBox2, 0, wxALIGN_LEFT|wxALL, 5);
}

void CQualityScoringPanel::OnApply ( wxCommandEvent& WXUNUSED(event) )
{
    CGradientColorPanel::stParams params;
    m_GradPanel->GetParams(params);
    m_Method.m_Worst = params.m_FirstColor;
    m_Method.m_Best = params.m_LastColor;

    m_Method.CreateColorTable(params.m_Steps);

    int options = 0;
    if(m_Gap) options |= CQualityScoringMethod::fIgnoreGaps;
    if (m_ColScore) options |= CQualityScoringMethod::fScoreWholeColumn;
    m_Method.SetOptions(options);

    if (m_Method.GetType() & IAlnExplorer::fProtein ) {
        CQualityScoringMethodAA* aa_method = dynamic_cast<CQualityScoringMethodAA*>(&m_Method);
        _ASSERT(aa_method);
        wxString matrix_name = m_MatrixChoice->GetStringSelection();
        if (!matrix_name.empty()) {
            const SNCBIPackedScoreMatrix* m = CTableNames::NameToTable(ToStdString(matrix_name));
            if (m != NULL)
                aa_method->SetSubstitutionMatrix(m);
        }
    }
}

wxWindow* CQualityScoringMethod::CreatePropertiesPanel(wxWindow* parent)
{
    return new CQualityScoringPanel(*this, parent);
}

///////////////////////////////////////////////////////////////////////////////
///  CQualityScoringMethodNA
///
CQualityScoringMethodNA::CQualityScoringMethodNA()
    : CQualityScoringMethod()
{
}


CQualityScoringMethodNA::CQualityScoringMethodNA(int colors)
    : CQualityScoringMethod(colors)
{
}


CQualityScoringMethodNA::~CQualityScoringMethodNA()
{
}


string  CQualityScoringMethodNA::GetName() const
{
    return "Column Quality score - DNA";
}


IScoringMethod* CQualityScoringMethodNA::Clone() const
{
    return new CQualityScoringMethodNA(*this);
}


int CQualityScoringMethodNA::GetType() const
{
    return IAlnExplorer::fDNA;
}


string CQualityScoringMethodNA::GetAllowedSymbols() const
{
    return string("ACGTU");
}


TNCBIScore CQualityScoringMethodNA::GetSubstitutionScore(int aa1, int aa2) const
{
    if (aa1 == aa2)
        return 15;
/*
    if ((aa1 == 'T'  &&  aa2 == 'U')  ||  (aa1 == 'U'  &&  aa2 == 'T'))
        return 1;
*/
    return 0;
}


///////////////////////////////////////////////////////////////////////////////
///  CQualityScoringMethodAA
///
CQualityScoringMethodAA::CQualityScoringMethodAA()
    : CQualityScoringMethod()
{
    SetSubstitutionMatrix(&NCBISM_Blosum62);
    // or NCBISM_Blosum45 NCBISM_Blosum80 NCBISM_Pam30 NCBISM_Pam70 NCBISM_Pam250
}


CQualityScoringMethodAA::CQualityScoringMethodAA(int colors)
    : CQualityScoringMethod(colors)
{
    SetSubstitutionMatrix(&NCBISM_Blosum62);
    // or NCBISM_Blosum45 NCBISM_Blosum80 NCBISM_Pam30 NCBISM_Pam70 NCBISM_Pam250
}


CQualityScoringMethodAA::CQualityScoringMethodAA(const CQualityScoringMethodAA& orig)
    : CQualityScoringMethod(orig)
{
    SetSubstitutionMatrix(orig.m_SubstMatrix);
}


CQualityScoringMethodAA::~CQualityScoringMethodAA()
{
}


string  CQualityScoringMethodAA::GetName() const
{
    return "Column Quality score - Protein";
}


IScoringMethod* CQualityScoringMethodAA::Clone() const
{
    return new CQualityScoringMethodAA(*this);
}


int CQualityScoringMethodAA::GetType() const
{
    return IAlnExplorer::fProtein;
}


TNCBIScore CQualityScoringMethodAA::GetSubstitutionScore(int aa1, int aa2) const
{
    return m_FullSubstMatrix.s[aa1][aa2];
}


string CQualityScoringMethodAA::GetAllowedSymbols() const
{
    return m_SubstMatrix->symbols;
}


void  CQualityScoringMethodAA::SetSubstitutionMatrix(const SNCBIPackedScoreMatrix* m)
{
    m_SubstMatrix = m;
    NCBISM_Unpack(m_SubstMatrix, &m_FullSubstMatrix);
}


const SNCBIPackedScoreMatrix* CQualityScoringMethodAA::GetSubstitutionMatrix()
{
    return m_SubstMatrix;
}

END_NCBI_SCOPE
