/*  $Id $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Robert G. Smith
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>
#include <gui/widgets/config/theme_config_panel2.hpp>

#include <wx/panel.h>
#include <wx/stattext.h>
#include <wx/choice.h>
#include <wx/button.h>
#include <wx/sizer.h>

BEGIN_NCBI_SCOPE

class CDataItemValidator : public wxValidator
{
public:
    CDataItemValidator(CThemeConfigPanel2::SItemData* val_ptr);
    CDataItemValidator(const CDataItemValidator& val);

    // Make a clone of this validator (or return NULL) - currently necessary
    // if you're passing a reference to a validator.
    // Another possibility is to always pass a pointer to a new validator
    // (so the calling code can use a copy constructor of the relevant class).
    virtual wxObject *Clone() const { return new CDataItemValidator(*this); }
    bool Copy(const CDataItemValidator& val);

    // Called when the value in the window must be validated.
    // This function can pop up an error message.
    virtual bool Validate(wxWindow *parent) { return true; }

    // Called to transfer data to the window
    virtual bool TransferToWindow();

    // Called to transfer data from the window
    virtual bool TransferFromWindow();

protected:
    bool CheckValidator() const;

    CThemeConfigPanel2::SItemData* m_pDataItem;

private:
    CDataItemValidator& operator=(const CDataItemValidator&);
};

CDataItemValidator::CDataItemValidator(CThemeConfigPanel2::SItemData* val_ptr) :
    m_pDataItem(val_ptr)
{
}

CDataItemValidator::CDataItemValidator(const CDataItemValidator& val)
    : wxValidator()
{
    Copy(val);
}

bool CDataItemValidator::Copy(const CDataItemValidator& val)
{
    wxValidator::Copy(val);
    m_pDataItem = val.m_pDataItem;
    return true;
}

// Called to transfer data to the window
bool CDataItemValidator::TransferToWindow()
{
    if( !CheckValidator() )
        return false;

    wxControlWithItems *control = (wxControlWithItems *) m_validatorWindow;

    if ( m_pDataItem ) {
        control->Clear();

        ITERATE(CThemeConfigPanel2::TStyleList, it, m_pDataItem->styles)
            control->Append(wxString(it->c_str()));

        control->SetStringSelection(wxString(m_pDataItem->current_style.c_str()));
    }

    return true;
}

// Called to transfer data to the window
bool CDataItemValidator::TransferFromWindow()
{
    if( !CheckValidator() )
        return false;

    wxControlWithItems *control = (wxControlWithItems *) m_validatorWindow;

    if ( m_pDataItem )
        m_pDataItem->current_style = control->GetStringSelection().c_str();

    return true;
}

bool CDataItemValidator::CheckValidator() const
{
    wxCHECK_MSG( m_validatorWindow, false,
                 _("No window associated with validator") );
    wxCHECK_MSG( m_validatorWindow->IsKindOf(CLASSINFO(wxChoice)), false,
                 _("CDataItemValidator is only for wxChoice's") );

    return true;
}


class CwxThemeConfigPanel : public wxPanel

{
    DECLARE_EVENT_TABLE()

    enum {
        kIdBase = 10000
    };

public:
    CwxThemeConfigPanel(CThemeConfigPanel2& configPanel,
                        wxWindow* parent, wxWindowID id = wxID_ANY,
                        const wxPoint& pos = wxDefaultPosition,
                        const wxSize& size = wxDefaultSize);

    wxPanel* GetPreviewPanel() { return m_PreviewPanel; }

protected:
    void CreateControls();
    void Init();

    void OnStyleSelected( wxCommandEvent& event );
    void OnAdvClick( wxCommandEvent& event );

    CThemeConfigPanel2& m_ConfigPanel;
    vector<IClickAction*> m_Actions;
    wxPanel* m_PreviewPanel;
};

BEGIN_EVENT_TABLE( CwxThemeConfigPanel, wxPanel)
    EVT_CHOICE( wxID_ANY, CwxThemeConfigPanel::OnStyleSelected )
    EVT_BUTTON( wxID_ANY, CwxThemeConfigPanel::OnAdvClick )
END_EVENT_TABLE()

CwxThemeConfigPanel::CwxThemeConfigPanel(
            CThemeConfigPanel2& configPanel,
            wxWindow* parent, wxWindowID id, const wxPoint& pos, const wxSize& size) :
    m_ConfigPanel(configPanel), m_PreviewPanel(0)
{
    Create(parent, id, pos, size);
    CreateControls();
}

void CwxThemeConfigPanel::CreateControls()
{
    bool horizontal =
        (m_ConfigPanel.m_LayoutDirection == CThemeConfigPanel2::eLayoutDirection_Horizontal);


    wxBoxSizer* itemBoxSizer1 = new wxBoxSizer(horizontal ? wxVERTICAL : wxHORIZONTAL);
    wxBoxSizer* itemBoxSizer2 = new wxBoxSizer(horizontal ? wxHORIZONTAL : wxVERTICAL);

    SetSizer(itemBoxSizer1);
    if (horizontal)
        itemBoxSizer1->Add(itemBoxSizer2, 0, wxALIGN_LEFT|wxALL, 5);
    else
        itemBoxSizer1->Add(itemBoxSizer2, 0, wxALIGN_TOP|wxALL, 5);

    m_Actions.resize(m_ConfigPanel.m_ItemData.size(), 0);

    wxWindowID id = kIdBase;
    for (size_t i = 0; i < m_ConfigPanel.m_ItemData.size(); ++i, id += 2) {
        wxFlexGridSizer* itemBoxSizer3 = new wxFlexGridSizer(2, 2);
        if (!horizontal)
            itemBoxSizer3->AddGrowableCol(0);

        itemBoxSizer3->Add(
            new wxStaticText(this, wxID_STATIC, m_ConfigPanel.m_ItemData[i].desc.c_str()),
                            0, wxALIGN_LEFT|wxALL, 2);
        itemBoxSizer3->AddSpacer(1);

        wxChoice* itemChoice = new wxChoice(this, id);
        itemChoice->SetValidator(CDataItemValidator(&m_ConfigPanel.m_ItemData[i]));
        if (horizontal)
            itemBoxSizer3->Add(itemChoice, 0, wxALIGN_LEFT|wxALL, 0);
        else
            itemBoxSizer3->Add(itemChoice, 1, wxGROW|wxALL, 0);

        CRef<IClickAction> click_act =
                m_ConfigPanel.GetClickAction(m_ConfigPanel.m_ItemData[i].type);

        if (click_act) {
            wxButton* itemButton = new wxButton(this, id + 1, wxT("..."),
                        wxDefaultPosition, wxSize(this->ConvertDialogToPixels(wxSize(16, -1)).x, -1), 0 );

            itemBoxSizer3->Add(itemButton, 0, wxALIGN_CENTER_VERTICAL|wxLEFT, 4);
            m_Actions[i] = click_act.GetPointer();
        }
        else
            itemBoxSizer3->AddSpacer(1);

        if (!horizontal)
            itemBoxSizer2->Add(itemBoxSizer3, 0, wxGROW|wxALL, 2);
        else
            itemBoxSizer2->Add(itemBoxSizer3, 0, wxALIGN_CENTER_VERTICAL|wxALIGN_LEFT|wxALL, 2);
    }

    m_PreviewPanel = new wxPanel(this, wxID_ANY, wxDefaultPosition, wxSize(315, 160),
                                 wxBORDER_SUNKEN);

    if (horizontal)
        itemBoxSizer1->Add(m_PreviewPanel, 1, wxGROW|wxALL, 5);
    else
        itemBoxSizer1->Add(m_PreviewPanel, 1, wxGROW|wxALL, 5);

    TransferDataToWindow();

    itemBoxSizer1->SetSizeHints(this);
    Fit();
}

void CwxThemeConfigPanel::OnStyleSelected( wxCommandEvent& event )
{
    TransferDataFromWindow();
    m_ConfigPanel.UpdatePreview();
}

void CwxThemeConfigPanel::OnAdvClick( wxCommandEvent& event )
{
    int index = (event.GetId()- kIdBase - 1)/2;
    if (m_Actions[index]) {
        m_Actions[index]->DoClick();
        m_ConfigPanel.UpdateWidgets();
    }
}

CThemeConfigPanel2::CThemeConfigPanel2() :
    m_LayoutDirection(CThemeConfigPanel2::eLayoutDirection_Horizontal)
{
}


CThemeConfigPanel2::~CThemeConfigPanel2()
{
}

wxWindow* CThemeConfigPanel2::CreateConfigPanel(wxWindow* parent)
{
    //SetLayoutDirection(CThemeConfigPanel2::eLayoutDirection_Horizontal);
    m_ConfigPanel = new CwxThemeConfigPanel(*this, parent);
    return m_ConfigPanel;
}

void CThemeConfigPanel2::MakeWidgets()
{
    // UpdateWidgets();
}

wxPanel* CThemeConfigPanel2::GetPreviewLocation()
{
    return ((CwxThemeConfigPanel*)m_ConfigPanel)->GetPreviewPanel();
}

void  CThemeConfigPanel2::UpdateWidgets()
{
    m_ConfigPanel->TransferDataToWindow();
    UpdatePreview();
    m_ConfigPanel->Refresh();
}


void CThemeConfigPanel2::UpdatePreview()
{ // override to do something.
}

void CThemeConfigPanel2::SetLayoutDirection(CThemeConfigPanel2::ELayoutDirection d)
{
    m_LayoutDirection = d;
}

CThemeConfigPanel2::ELayoutDirection  CThemeConfigPanel2::GetLayoutDirection()
{
    return m_LayoutDirection;
}

END_NCBI_SCOPE
