/*  $Id: query_parse_panel.cpp 28067 2013-05-06 21:51:49Z rafanovi $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and m_Data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Robert Falk
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>

#include <gui/widgets/data/query_parse_panel.hpp>

#include <gui/framework/status_bar_service.hpp>
#include <gui/framework/event_log_service_impl.hpp>
#include <gui/framework/service.hpp>

#include <gui/widgets/wx/wx_utils.hpp>
#include <gui/widgets/wx/fileartprov.hpp>
#include <gui/widgets/wx/sys_path.hpp>

#include <gui/objutils/query_data_source.hpp>
#include <gui/objutils/query_func_promote.hpp>

#include <util/qparse/query_parse.hpp>
#include <util/checksum.hpp>

#include <wx/app.h>
#include <wx/sizer.h>
#include <wx/menu.h>
#include <wx/button.h>
#include <wx/animate.h>
#include <wx/statline.h>
#include <wx/stattext.h>
#include <wx/toolbar.h>
#include <wx/choice.h>
#include <wx/panel.h>
#include <wx/timer.h>
#include <wx/textctrl.h>
#include <wx/checkbox.h>



#include <algorithm>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);


#define ID_SEARCHCTRL       13001
#define ID_SELECT_ALL       13005
#define ID_HIDE_UNSELECTED  13006
#define ID_QUERY_ANIMATIONCTRL 13007
#define ID_TOOLBAR          13010
#define ID_TOOLBAR2          13011
#define ID_CASE_SENSITIVE   13012
#define ID_MRU1             13013

static const char* kRecentQueries = "Recent Queries";
static const char* kQueryPanel = ".QueryPanelGrid";

///////////////////////////////////////////////////////////////////////////////
//
// Background job for executing query on data source
//
///////////////////////////////////////////////////////////////////////////////
inline void CQueryParsePanel::CQueryJob::x_Query()
{
    m_QExec->EvalStart();
    
    while (!m_QExec->EvalComplete()) {
        if (IsCanceled())
            return;

        if (m_QExec->EvalNext(*m_QueryTree)) 
            ++m_NumSelected;
    }
}

CQueryParsePanel::CQueryJob::~CQueryJob() 
{
    delete m_QueryTree;
}

IAppJob::EJobState CQueryParsePanel::CQueryJob::Run()
{
    // execute query on all elements in data source
    x_Query();

    return eCompleted;
}

CConstIRef<IAppJobProgress> CQueryParsePanel::CQueryJob::GetProgress()
{
    return CConstIRef<IAppJobProgress>(new CAppJobProgress(1.0, "Finished"));
}

CRef<CObject> CQueryParsePanel::CQueryJob::GetResult()
{
    CMutexGuard guard(m_Mutex);
    CRef<CObject> result(new CObject());
    return result;
}

CConstIRef<IAppJobError> CQueryParsePanel::CQueryJob::GetError()
{
    return CConstIRef<IAppJobError>(new CAppJobError("No errors"));
}

int CQueryParsePanel::CQueryJob::GetQueriedCount() const 
{ 
    return m_QExec->GetQueriedCount(); 
}

int CQueryParsePanel::CQueryJob::GetExceptionCount() const 
{ 
    return m_QExec->GetExceptionCount();
}



///////////////////////////////////////////////////////////////////////////////
//
// CQueryParsePanel
// 
///////////////////////////////////////////////////////////////////////////////

bool CQueryParsePanel::m_IconsInitialized = false;
const int CQueryParsePanel::m_MaxRecentQueries;   
const int CQueryParsePanel::m_MaxDefaultQueries;

BEGIN_EVENT_MAP(CQueryParsePanel, CEventHandler)
    ON_MESSAGE(CAppJobNotification, CAppJobNotification::eStateChanged,
               &CQueryParsePanel::OnAppJobNotification)
END_EVENT_MAP()

BEGIN_EVENT_TABLE( CQueryParsePanel, wxPanel )
    EVT_CONTEXT_MENU( CQueryParsePanel::OnContextMenu )
    EVT_TEXT(ID_SEARCHCTRL, CQueryParsePanel::OnSearchctrlTextEnter)
    EVT_TEXT_ENTER(ID_SEARCHCTRL, CQueryParsePanel::OnSearchctrlSearchButtonClick)    
    EVT_TOOL(eCmdSelQuery, CQueryParsePanel::OnQuerySel)
    EVT_TOOL(eCmdNext, CQueryParsePanel::OnButtonNext)
    EVT_TOOL(eCmdPrevious, CQueryParsePanel::OnButtonPrevious)   
    EVT_TOOL(eCmdStartQuery, CQueryParsePanel::OnSearchctrlSearchButtonClick)
    EVT_TOOL(eCmdStopQuery, CQueryParsePanel::OnSearchctrlSearchStop)
    EVT_TOOL(eCmdHelp, CQueryParsePanel::OnSearchHelp)
    EVT_CHECKBOX(ID_SELECT_ALL, CQueryParsePanel::OnToggleSelectAll)
    EVT_CHECKBOX(ID_HIDE_UNSELECTED, CQueryParsePanel::OnToggleHideUnselected)
    EVT_COMMAND_RANGE(ID_MRU1, ID_MRU1 + 
                      CQueryParsePanel::m_MaxRecentQueries + 
                      CQueryParsePanel::m_MaxDefaultQueries - 1, 
                      wxEVT_COMMAND_MENU_SELECTED, 
                      CQueryParsePanel::OnRecentQuery)
END_EVENT_TABLE()



CQueryParsePanel::CQueryParsePanel(IQueryWidget* w, IQueryDataSource* ds)
: m_pSearch(NULL)
, m_Animation(NULL)
, m_ToolBar(NULL)
, m_ToolBar2(NULL)
, m_CaseSensitive(NULL)
, m_pSelectAll(NULL)
, m_pHideUnselected(NULL)
, m_ParentPanel(NULL)
, m_ServiceLocator(NULL)
, m_QueryWidget(w)
, m_QueryDataSource(NULL)
, m_PrevCaseSensitive(false)
, m_JobID(CAppJobDispatcher::eInvalidJobID)
{
    SetDataSource(ds);
    Init();
}

CQueryParsePanel::~CQueryParsePanel()
{
    #ifndef __WXMAC__
        PopEventHandler( true );
    #endif

    // If a query job is running take care in cancelling it since it uses
    // data from the datasource object.  If the query is running, cancel
    // and give it a little time to finish up.  (testing shows the delay 
    // to be normally very short between cancel and completion)
	if (m_JobID != CAppJobDispatcher::eInvalidJobID) {
        CAppJobDispatcher& disp = CAppJobDispatcher::GetInstance();
        if (disp.GetJobState(m_JobID) != IAppJob::eInvalid) {
            disp.CancelJob(m_JobID);

            IAppJob::EJobState state = disp.GetJobState(m_JobID);

            int delay_count = 0;
            while (state == IAppJob::eRunning && ++delay_count < 1000){
                wxMilliSleep(10);
                state = disp.GetJobState(m_JobID);
            }
        }
    }
}

void CQueryParsePanel::Create( wxWindow* parent, wxWindowID id, const wxPoint& pos, const wxSize& size )
{
    // these are necessary for MacOSX proper background erasure
    #ifdef __WXMAC__    
        SetBackgroundStyle(wxBG_STYLE_COLOUR);
        SetBackgroundColour(*wxWHITE);
    #endif

    wxPanel::Create(parent, id, pos, size);
    CreateControls();

    #ifndef __WXMAC__
        PushEventHandler( new CCommandToFocusHandler( this ) );
    #endif
}

void CQueryParsePanel::Init()
{
    m_pSearch          = NULL;
    m_Animation        = NULL;
    m_ToolBar          = NULL; 
    m_ToolBar2         = NULL; 
    m_CaseSensitive    = NULL;
    m_pSelectAll       = NULL;
    m_pHideUnselected  = NULL;
    m_ParentPanel      = NULL;


    if( !m_IconsInitialized ){
        wxFileArtProvider* provider = GetDefaultFileArtProvider();

        provider->RegisterFileAlias(wxT("menu::dm_start"), wxT("playhs.png"));
        provider->RegisterFileAlias(wxT("menu::dm_stop"),  wxT("stophs.png"));
        provider->RegisterFileAlias(wxT("menu::back"), wxT("back.png"));
        provider->RegisterFileAlias(wxT("menu::forward"), wxT("forward.png"));
        provider->RegisterFileAlias(wxT("menu::help"), wxT("help.png"));
        provider->RegisterFileAlias(wxT("menu::query_sel"), wxT("dropdown_sel.png"));
        
        m_IconsInitialized = true;
    }
}

void CQueryParsePanel::CreateControls()
{
    wxBoxSizer* panel_sizer = new wxBoxSizer(wxHORIZONTAL);

    long styles = wxTB_FLAT | wxTB_HORIZONTAL;
    m_ToolBar = new wxToolBar(this, ID_TOOLBAR,
                              wxDefaultPosition, wxSize(-1, -1), styles);

    // Two tool bars are used because wxTB_HORZ_LAYOUT style is needed for
    // icons that have text next to them, but when wxTB_HORZ_LAYOUT is used
    // ALL icons in that toolbar will have text next to them, and we 
    // can't have text next to the 'query-select' (eCmdSelQuery) buttion.
    // Except on mac, which can have visual errors (mismatched height) 
    // between the adjacent toolbars. Don't use text on mac icons.
#ifdef __WXMAC__
    m_ToolBar2 = m_ToolBar;
#else
    long styles2 =  wxTB_FLAT | wxTB_HORIZONTAL | wxTB_TEXT | wxTB_HORZ_LAYOUT;
    m_ToolBar2 = new wxToolBar(this, ID_TOOLBAR2,
                              wxDefaultPosition, wxSize(-1, -1), styles2);
#endif

    // label    
    wxStaticText* tool_text = new wxStaticText( m_ToolBar, wxID_STATIC,
                                                wxT(" Search: "));
    m_ToolBar->AddControl(tool_text);

    // Must supply wxTE_RICH otherwise the context menu (cut, copy, past...) for the search box will not work.
    // Without wxTE_RICH, context menus from the parent widgets (this one, and CDockPanel) will both
    // take priority over the search controls context menu meaning it never shows up.  This is probably a wx
    // bug coming from the fact wx uses the built-in context menu when wxTE_RICH is not used, and a wx-created
    // context menu when it is. (and search control is derived from text control, so wxTE_RICH is recognized)

    // Update - The other reason to use TE_RICH is so that colored text (for fieldnames ) will work.
    // unfortunately TE_RICH on cocoa forces creation of a multi-line control which is unusuable :-(.  
    // So I've patched wxWidgets so that when you do not provide wxTE_MULTILINE with wxTE_RICH it will
    // create a single-line rich text control.  Updates are in src/osx/cocoa/textctrl.mm (wxNSTextViewControl ctor)
    // and include/wx/osx/cocoa/private/textimpl.h (ctor declaration). 
/*  Actual wxWidgets patch.  This was a version of wxWidgets between 2.9.2 and 2.9.3.
Index: src/osx/cocoa/textctrl.mm
===================================================================
--- src/osx/cocoa/textctrl.mm   (revision 69613)
+++ src/osx/cocoa/textctrl.mm   (working copy)
@@ -424,24 +424,49 @@
 
 // wxNSTextViewControl
 
-wxNSTextViewControl::wxNSTextViewControl( wxTextCtrl *wxPeer, WXWidget w )
+wxNSTextViewControl::wxNSTextViewControl( wxTextCtrl *wxPeer, WXWidget w, long style )
     : wxWidgetCocoaImpl(wxPeer, w),
       wxTextWidgetImpl(wxPeer)
 {
     wxNSTextScrollView* sv = (wxNSTextScrollView*) w;
     m_scrollView = sv;
 
-    [m_scrollView setHasVerticalScroller:YES];
+    bool single_line = !(style & wxTE_MULTILINE); 
+
+    [m_scrollView setHasVerticalScroller:NO];
     [m_scrollView setHasHorizontalScroller:NO];
-    [m_scrollView setAutoresizingMask:NSViewWidthSizable | NSViewHeightSizable];
+
+    if (single_line) {
+        // Removes vertical scroll bar from single-line text
+        [m_scrollView setAutoresizingMask:NSViewNotSizable];
+    }
+    else {
+        [m_scrollView setHasVerticalScroller:YES];
+        [m_scrollView setAutoresizingMask:NSViewWidthSizable | NSViewHeightSizable];
+    }
+
     NSSize contentSize = [m_scrollView contentSize];
 
     wxNSTextView* tv = [[wxNSTextView alloc] initWithFrame: NSMakeRect(0, 0,
             contentSize.width, contentSize.height)];
     m_textView = tv;
+
+    const CGFloat LargeNumberForText = 1.0e7;
+
+    if (single_line) {
+        NSTextContainer *textContainer = [tv textContainer];
+        [textContainer setContainerSize:NSMakeSize(LargeNumberForText, LargeNumberForText)];
+        [textContainer setWidthTracksTextView:NO];
+        [tv setMaxSize:NSMakeSize(LargeNumberForText, LargeNumberForText)];
+        [tv setHorizontallyResizable:YES];
+        [tv setAutoresizingMask:NSViewNotSizable];
+    }
+    else {
+        [tv setHorizontallyResizable:NO];
+        [tv setAutoresizingMask:NSViewWidthSizable];
+    }
+
     [tv setVerticallyResizable:YES];
-    [tv setHorizontallyResizable:NO];
-    [tv setAutoresizingMask:NSViewWidthSizable];
 
     [m_scrollView setDocumentView: tv];
 
@@ -833,7 +858,7 @@
     {
         wxNSTextScrollView* v = nil;
         v = [[wxNSTextScrollView alloc] initWithFrame:r];
-        c = new wxNSTextViewControl( wxpeer, v );
+        c = new wxNSTextViewControl( wxpeer, v, style);
     }
     else
     {
Index: include/wx/osx/cocoa/private/textimpl.h
===================================================================
--- include/wx/osx/cocoa/private/textimpl.h     (revision 69613)
+++ include/wx/osx/cocoa/private/textimpl.h     (working copy)
@@ -56,7 +56,7 @@
 class wxNSTextViewControl : public wxWidgetCocoaImpl, public wxTextWidgetImpl
 {
 public:
-    wxNSTextViewControl( wxTextCtrl *wxPeer, WXWidget w );
+    wxNSTextViewControl( wxTextCtrl *wxPeer, WXWidget w, long style );
     virtual ~wxNSTextViewControl();
 
     virtual wxString GetStringValue() const ;
*/

    
    int search_width = 180;

    // Search box on mac is a little too short (vertically) so set size here:
#ifdef __WXOSX_COCOA__    
    int search_height = tool_text->GetFont().GetPointSize()+6;
#else
    int search_height = -1;
#endif

    // If you do not have a pathed version of wxWidgets (on mac cocoa), then 
    // do not include wxTE_RICH (loses font highlight, but toolbar still works)
    m_pSearch = new wxTextCtrl(m_ToolBar,
                               ID_SEARCHCTRL, wxEmptyString, 
                               wxDefaultPosition, wxSize(search_width, search_height),  
                               wxTE_PROCESS_ENTER | wxTE_RICH);
   
    wxAcceleratorEntry entries[6];
            
    entries[0].Set(wxACCEL_CMD, (int) 'C', wxID_COPY);
    entries[1].Set(wxACCEL_CMD, (int) 'X', wxID_CUT);
    entries[2].Set(wxACCEL_CMD, (int) 'V', wxID_PASTE);
    entries[3].Set(wxACCEL_CMD, (int) 'A', wxID_SELECTALL);
    entries[4].Set(wxACCEL_CMD,       384, wxID_COPY);
    entries[5].Set(wxACCEL_SHIFT,     384, wxID_PASTE);
    
    wxAcceleratorTable accel(6, entries);
    m_pSearch->SetAcceleratorTable(accel);
    
    m_pSearch->SetToolTip(wxT("Enter search text as:\n\
1. A string to match against all fields\n\
2. A query based on field names, e.g. :\n\
   dist > 0.2 and seq-id == \"NP_110296\""));

    m_ToolBar->AddControl(m_pSearch);

    wxBitmap bmp_selquery = wxArtProvider::GetBitmap(wxT("menu::query_sel"));
    m_ToolBar->AddTool(eCmdSelQuery, wxEmptyString, bmp_selquery, wxNullBitmap, wxITEM_NORMAL, wxT("Select Query")); 
    m_ToolBar->EnableTool(eCmdSelQuery, true);

    // I really just want empty space before the checkbox, but this is the only thing
    // that seems to work.
    m_ToolBar->AddSeparator();

    m_CaseSensitive = new wxCheckBox(m_ToolBar2, ID_CASE_SENSITIVE, _("Case Sensitive"), 
        wxDefaultPosition, wxDefaultSize, wxNO_BORDER );
    m_CaseSensitive->SetValue(false);
    m_CaseSensitive->SetToolTip(wxT("Check to do case-sensitive string comparisons"));
    m_ToolBar2->AddControl(m_CaseSensitive);

    m_ToolBar2->AddSeparator();

    // buttons
    wxBitmap bmp_start = wxArtProvider::GetBitmap(wxT("menu::dm_start"));
    m_ToolBar2->AddTool(eCmdStartQuery, wxT("Start"), bmp_start, wxNullBitmap, wxITEM_NORMAL, wxT("Start Search")); 
    m_ToolBar2->EnableTool(eCmdStartQuery, false);

    m_Animation = new wxAnimationCtrl(m_ToolBar2, ID_QUERY_ANIMATIONCTRL, wxAnimation(),
                                      wxDefaultPosition, wxSize(22, 22));
    m_Animation->SetToolTip(wxT("Light moves while query is running"));

    wxString path = CSysPath::ResolvePath(wxT("<res>\\status_anim.gif"));
    m_Animation->LoadFile(path);
    m_ToolBar2->AddControl(m_Animation); 

    wxBitmap bmp_stop = wxArtProvider::GetBitmap(wxT("menu::dm_stop"));
    m_ToolBar2->AddTool(eCmdStopQuery, wxT("Stop"), bmp_stop, wxNullBitmap, wxITEM_NORMAL, wxT("Stop Search"));
    m_ToolBar2->EnableTool(eCmdStopQuery, false);
       
    m_ToolBar2->AddSeparator();

    m_pHideUnselected = new wxCheckBox(m_ToolBar2, ID_HIDE_UNSELECTED, _("Filter"), 
        wxDefaultPosition, wxDefaultSize, wxNO_BORDER );
    m_pHideUnselected->SetValue(false);
    m_pHideUnselected->SetToolTip(wxT("Only show query results"));
    m_ToolBar2->AddControl(m_pHideUnselected);
    m_ToolBar2->AddSeparator();

    wxBitmap bmp_prev = wxArtProvider::GetBitmap(wxT("menu::back"));
    m_ToolBar2->AddTool(eCmdPrevious, wxT("Prev"), bmp_prev, wxT("Previous selection"));
    // Initially disabled because 'All' is true.  Can only use previous
    // and next when 'All' is false (so you are only selecting one at a
    // time)
    m_ToolBar2->EnableTool(eCmdPrevious, false);
 
    wxBitmap bmp_next = wxArtProvider::GetBitmap(wxT("menu::forward"));
    m_ToolBar2->AddTool(eCmdNext, wxT("Next"), bmp_next, wxT("Next selection"));
    m_ToolBar2->EnableTool(eCmdNext, false);

    m_pSelectAll = new wxCheckBox(m_ToolBar2, ID_SELECT_ALL, _("All"), 
        wxDefaultPosition, wxDefaultSize, wxNO_BORDER );
    m_pSelectAll->SetValue(true);
    m_pSelectAll->SetToolTip(wxT("Check to show all matching nodes as selected\n\
Uncheck to show one matching node at a time\n"));
    m_ToolBar2->AddControl(m_pSelectAll);

    if (m_QueryWidget != NULL &&
        m_QueryWidget->GetSearchHelpAddr() != string("")) {            
            string help_addr = m_QueryWidget->GetSearchHelpAddr();
            m_ToolBar2->AddSeparator();
            wxBitmap bmp_help = wxArtProvider::GetBitmap(wxT("menu::help"));
            m_ToolBar2->AddTool(eCmdHelp, wxT("Help"), bmp_help, wxT("Online query help"));
    }
    
    // Just use one toolbar on the mac (m_ToolBar==m_ToolBar2) without any text.  Using two, one with
    // and one without text, causes visual errors.
#ifndef __WXMAC__
    m_ToolBar->Realize();
#endif
    m_ToolBar2->Realize();

#ifndef __WXMAC__
    panel_sizer->Add(m_ToolBar, 0, wxALIGN_LEFT, 0);
#endif
    panel_sizer->Add(m_ToolBar2, 1, wxALL|wxGROW|wxALIGN_LEFT, 0);

    SetSizerAndFit(panel_sizer);
}

void CQueryParsePanel::SetDataSource(IQueryDataSource* ds)
{
    m_QueryDataSource = ds;

    string prev_section = m_RegSection;

    if (ds != NULL) {
        CChecksum cs(CChecksum::eMD5);
        size_t i;
        for (i = 0; i < m_QueryDataSource->GetColsCount(); ++i) {
            cs.AddLine(m_QueryDataSource->GetColumnLabel(i));
        }

        m_RegSection = GetMD5Digest(cs);

        if (m_RegSection != prev_section)
            LoadSettings();
    }
    else {
        m_RegSection = "";
    }

    // Update selection so that if data source changed, we are not pointing to 
    // selected elements from previous source. Don't do this during widget 
    // initialization (when m_pSelectAll is NULL still)
    if (m_QueryWidget != NULL && m_pSelectAll != NULL)
        m_QueryWidget->SetSelectAll(IsSelectAll());
}


void CQueryParsePanel::SetServiceLocator(IServiceLocator* s)
{
    m_ServiceLocator = s;
}

void CQueryParsePanel::SetDefaultQueries(const TNamedQueries& q)
{
    m_DefaultQueries = q;

    if (m_DefaultQueries.size() > (unsigned int)m_MaxDefaultQueries) {
        m_DefaultQueries.erase(m_DefaultQueries.begin() + m_MaxDefaultQueries, 
                               m_DefaultQueries.end());
    }
}

void CQueryParsePanel::SetRegistryPath(const string& reg_path) 
{ 
    m_RegPath = reg_path + kQueryPanel;
}

void CQueryParsePanel::LoadSettings()
{
    if (!m_RegPath.empty()) {
        CGuiRegistry& gui_reg = CGuiRegistry::GetInstance();
        CGuiRegistry::TReadView view = gui_reg.GetReadView(m_RegPath + "." + m_RegSection);

        vector<string> queries;
        view.GetStringVec(kRecentQueries, queries);

        for (size_t i=0; i<queries.size(); ++i) {
            if (m_RecentQueries.size() < (unsigned int)m_MaxRecentQueries)      
                m_RecentQueries.push_back(pair<string,string>(string(), queries[i]));
        }
    }
}

void CQueryParsePanel::SaveSettings() const
{
    if (m_RegPath.empty())
        return;

    CGuiRegistry& gui_reg = CGuiRegistry::GetInstance();
    CGuiRegistry::TReadWriteView view = gui_reg.GetReadWriteView(m_RegPath + "." + m_RegSection);

    vector<string> queries;

    vector<pair<string,string> >::const_iterator iter;
    for (iter = m_RecentQueries.begin(); iter != m_RecentQueries.end(); ++iter) {
        if (queries.size() < (unsigned int)m_MaxRecentQueries)
            queries.push_back((*iter).second);      
    }

    if (queries.size() > 0)
        view.Set(kRecentQueries, queries);
}


void CQueryParsePanel::AddQueryText(const string& query_mod)
{
    wxString str = ToWxString(query_mod);
    
    if (str != "") {
        wxString new_queryw = m_pSearch->GetValue();        
        new_queryw.Trim();
        string new_query = ToStdString(new_queryw);

        if (new_query != "")  {

            // If there is already query text, add an "AND" before
            // the new subclause UNLESS the last term in the query is
            // already AND, OR, NOT or XOR            
            vector<string> arr;
            NStr::Tokenize(new_query, " ", arr);

            if (arr.size() > 0) {
                string last_token = arr[arr.size()-1];
                if ((NStr::CompareNocase(last_token, "and") != 0) &&
                    (NStr::CompareNocase(last_token, "or") != 0) &&
                    (NStr::CompareNocase(last_token, "xor") != 0) &&
                    (NStr::CompareNocase(last_token, "not") != 0)) {
                        new_query += " AND ";
                }
               
                else {
                }
            }
        }

        // We trimmed the existing query above.  Now add one space to keep things
        // from running together.
        new_query += " " + str;

        m_pSearch->SetValue(new_query);
        // This should automatically call OnSearchctrlTextEnter.
    }
}

void CQueryParsePanel::OnContextMenu( wxContextMenuEvent& evt )
{
}

void CQueryParsePanel::OnRecentQuery(wxCommandEvent& evt)
{
    unsigned int query_idx = evt.GetId() - ID_MRU1;
    string query;

    if (query_idx < m_DefaultQueries.size())
        query = m_DefaultQueries[query_idx].second;
    else 
        query = m_RecentQueries[query_idx - m_DefaultQueries.size()].second;

    m_pSearch->SetValue(ToWxString(query));

    wxCommandEvent evt_dummy;    
    OnSearchctrlTextEnter(evt_dummy);
    x_ExecuteQuery(0);
}

void CQueryParsePanel::OnSearchctrlSearchButtonClick( wxCommandEvent& evt )
{
    x_ExecuteQuery(0);
}

void CQueryParsePanel::OnSearchctrlSearchStop( wxCommandEvent& evt )
{
    x_CancelQueryInProgress();
}

void CQueryParsePanel::OnSearchHelp(wxCommandEvent& evt)
{
    string help_addr = m_QueryWidget->GetSearchHelpAddr();

    ::wxLaunchDefaultBrowser(ToWxString(help_addr));
}

void CQueryParsePanel::OnSearchctrlTextEnter(wxCommandEvent& evt)
{
    // Check if query is currently valid and give user some feedback if 
    // it is (or isn't)
    string str = ToStdString(m_pSearch->GetValue());
    vector<std::pair<size_t, size_t> > fields;

    bool valid = x_IsValidQuery(str, m_CaseSensitive->IsChecked(), fields);

    wxFont f = m_pSearch->GetFont();

    // Remove any formatting from previous calls to this function:
    wxTextAttr attr(wxColour(wxString(wxT("black"))));
    attr.SetFlags(wxTEXT_ATTR_TEXT_COLOUR|wxTEXT_ATTR_FONT_WEIGHT);
    attr.SetFont(f);
    m_pSearch->SetStyle(0, str.length(), attr);

    // Find out how many leading blanks there are.  Since those are truncated
    // before query parsing, we need to add them back in when we highlight
    // data fields.
    string truncated_str = NStr::TruncateSpaces(str, NStr::eTrunc_Begin);
    int tdelta = str.length() - truncated_str.length();

    // If query is valid, enable the search button:
    if (valid)
        m_ToolBar2->EnableTool(eCmdStartQuery, true);
    else
        m_ToolBar2->EnableTool(eCmdStartQuery, false);

    // If query is valid,  highlight any substrings that were found during query 
    // parsing to be fields from the data dictionary.  An invalid query may
    // also return valid fields so highlight found-fields for invalid
    // queries too.
    if (valid || fields.size() > 0) {       
        wxTextAttr attr(wxColour(wxString(wxT("blue"))));
        attr.SetFlags(wxTEXT_ATTR_TEXT_COLOUR|wxTEXT_ATTR_FONT_WEIGHT);
        attr.SetFont(f);

        for (size_t i=0; i<fields.size(); ++i) {
            m_pSearch->SetStyle((long)(fields[i].first + tdelta),
                                (long)(fields[i].second + tdelta), 
                                attr);
        }

        // Remember 'good' string and fields so that we can continue
        // to highlight fields in parts of the query that have not been
        // modified when the query doesn't parse.
        m_SearchString = str;
        m_SearchStringFields = fields;
    }
    // If the query is not valid and no fields were returned, highlight only
    // those fields which appear in the part of the query string prior to
    // any part of the string that has not been altered (this highlights
    // fields found from the most recent valid query).
    else {        
        // Keep highlighted dictionary fields that are in the text
        // prior to any changes (which presumably made the text invalid)
        if (m_SearchStringFields.size() > 0) {
            wxTextAttr attr(wxColour(wxString(wxT("blue"))));
            attr.SetFlags(wxTEXT_ATTR_TEXT_COLOUR|wxTEXT_ATTR_FONT_WEIGHT);
            attr.SetFont(f);

            string::size_type idx;
            for (idx = 0; idx<str.length() && idx<m_SearchString.length(); ++idx) {
                if (str[idx] != m_SearchString[idx])
                    break;
            }

            for (size_t i=0; i<m_SearchStringFields.size(); ++i) {
                if (m_SearchStringFields[i].second < idx) {
                    m_pSearch->SetStyle((long)(m_SearchStringFields[i].first + tdelta),
                                        (long)(m_SearchStringFields[i].second + tdelta),
                                        attr);
                }
            }
        }
    }

    m_pSearch->Refresh();

    evt.Skip();
}

void CQueryParsePanel::OnQuerySel( wxCommandEvent& evt )
{
    wxMenu* m = new wxMenu();
    unsigned int i;
    unsigned int max_id = ID_MRU1;

    if (m_DefaultQueries.size() > 0) {
        wxMenu* default_menu = new wxMenu();

        for (i=0; i<m_DefaultQueries.size(); ++i) {
            default_menu->Append(max_id++, ToWxString(m_DefaultQueries[i].first));
        }

        m->AppendSubMenu(default_menu, wxT("Saved Queries"));
    }

    for (i=0; i<m_RecentQueries.size(); ++i) {
        m->Append(max_id++, ToWxString(m_RecentQueries[i].second));
    }

    m_ToolBar->PopupMenu(m);

    delete m;
}

void CQueryParsePanel::OnButtonNext( wxCommandEvent& evt )
{
    m_QueryWidget->IterateSelection(1);
}

void CQueryParsePanel::OnButtonPrevious( wxCommandEvent& evt )
{
    m_QueryWidget->IterateSelection(-1);
}

bool CQueryParsePanel::IsSelectAll() 
{ 
    return m_pSelectAll->IsChecked();
}

void CQueryParsePanel::OnToggleSelectAll(wxCommandEvent & evt)
{
    // Enable or disable previous/next selected element based on whether
    // select all is checked.  In 'select all' mode, previous and next
    // do not do anything.
    bool select_all = //m_QueryDataSource->GetQueryResults().size() > 0 &&
                          m_pSelectAll->IsChecked();
    m_ToolBar2->EnableTool(eCmdNext, !select_all);
    m_ToolBar2->EnableTool(eCmdPrevious, !select_all);

    m_QueryWidget->SetSelectAll(select_all);
}

void CQueryParsePanel::OnToggleHideUnselected(wxCommandEvent & evt)
{  
    // Quick check to see if the query has already been run - if not,
    // clicking 'filter' will run the query too.
    string query = ToStdString(m_pSearch->GetValue());
    bool  casesensitive = m_CaseSensitive->IsChecked();

    // Re-run the query if options that can change query results have changed
    if ( query!=m_PrevQuery || casesensitive!=m_PrevCaseSensitive) {  
        x_ExecuteQuery(0);
    }

    m_QueryWidget->SetHideUnselected(m_pHideUnselected->IsChecked());
}

///////////////////////////////////////////////////////////////////////////////
/// Class for extracting positions of data fields from the query.  A data field
/// is a query string token that matches a name in the dictionary.
///
class CQueryHighlightFunction
{
public: 
    CQueryHighlightFunction() {}

    std::vector<std::pair<size_t, size_t> > 
        GetFieldPositions() const { return m_FieldPositions; }
          
    ETreeTraverseCode 
    operator()(const CTreeNode<CQueryParseNode>& tr, int delta) 
    {
        if (delta < 0)
            return eTreeTraverse;

        const CQueryParseNode& qnode = tr.GetValue();

        const IQueryParseUserObject* uo = qnode.GetUserObject();
        const CQueryNodeValue* v = dynamic_cast<const CQueryNodeValue*>(uo);

        if (v != NULL && v->IsDataField()) {
            std::pair<size_t, size_t> field_position;

            // The length in SSrcLoc is left as 0
            field_position.first = qnode.GetLoc().pos;
            field_position.second = qnode.GetLoc().pos + qnode.GetStrValue().length();

            m_FieldPositions.push_back(field_position);
        }

        return eTreeTraverse;
    }


protected:
    std::vector<std::pair<size_t, size_t> > m_FieldPositions;
};


bool CQueryParsePanel::x_IsValidQuery(string query, 
                                      bool casesensitive,  
                                      vector<std::pair<size_t, size_t> >& fields)
{
    // Get query string with outer-most quotes stripped. 
    bool simple_string = x_FormatQueryString(query, casesensitive);
    bool valid = (query.length() > 0);

    // A simple string is always a valid query
    if (!simple_string) {
        CQueryParseTree qtree;

        try {
            // parse the query.  This will throw exceptions for syntax errors
            // like 'dist > 0.2 AND' but since it does not look at the tree's 
            // dictionary, strings/phrases that can't be promoted to a reasonable
            // value still pass, becuase they *could* be data fields, e.g.:
            // 'dist > 0.7 AND my dog likes to bark' will parse without error.            
            qtree.Parse(query.c_str(), 
                        casesensitive ? CQueryParseTree::eCaseSensitiveUpper : 
                        CQueryParseTree::eCaseInsensitive,
                        CQueryParseTree::eSyntaxCheck);   

            CQueryExec* qexec = m_QueryDataSource->GetQueryExec(casesensitive);

            // Pre-process the query.  This step determines, to the extent
            // possible, the types to be used in comaparison and whether a valid
            // number of arguments are present. Where possible, static types
            // (as opposed to identifers for field-values) are converted into
            // their required types for comparisons.
            try {
                CRef<objects::CScope> sc = m_QueryDataSource->GetScope();               

                CQueryExecPreProcessFunc  pre_process_exec(sc.GetPointerOrNull(),
                                                           *qexec);
                TreeDepthFirstTraverse(*qtree.GetQueryTree(), 
                                       pre_process_exec);
            }
            catch(CQueryExecException &e) {
                LOG_POST(Info << "Error pre-processing query: " << e.GetMsg());
                valid = false;
            }

            CQueryHighlightFunction field_finder;
            field_finder = TreeDepthFirstTraverse(*qtree.GetQueryTree(), 
                                                  field_finder);

            fields = field_finder.GetFieldPositions();
        }
        // Catch exceptions from parsing the query or from executing it on the
        // data source.
        catch(CQueryParseException &ex) {
            LOG_POST(Info << "Query Parse Error: " << ex.GetMsg());
            valid = false;
        }
    }

    // A simple string query.  Return true unless string is empty:
    return (valid);
}

void CQueryParsePanel::CompleteQuery(const string& status)
{
    // Let widget make any needed changes with query completed,
    // e.g. enable widget interaction, update selection set, etc.
    m_QueryWidget->QueryEnd();

    // Stop animation in query tool bar
    m_Animation->Stop();

    // Re-enable all input to window (and disable cancel button)
    this->SetCursor(*wxSTANDARD_CURSOR);
    m_ToolBar2->EnableTool(eCmdStopQuery, false);
    m_ToolBar2->EnableTool(eCmdStartQuery, true);
    m_ToolBar2->EnableTool(eCmdNext, !m_pSelectAll->IsChecked());
    m_ToolBar2->EnableTool(eCmdPrevious, !m_pSelectAll->IsChecked());
    m_pSearch->Enable();
    m_CaseSensitive->Enable();
    m_pSelectAll->Enable();
    m_pHideUnselected->Enable();   

    // Because we disable windows during the query we lose focus in the text 
    // box. This brings it back.  Now since the user can use other windows 
    // while search is running (if the search turns out to be long...), we 
    // only put focus back if it hasn't been assigned somewhere else.
    wxWindow* fwin = wxWindow::FindFocus();

    if (fwin == NULL)
        m_pSearch->SetFocus();

    /// Update staus bar with status message from query
    if (m_ServiceLocator != NULL) {
        IStatusBarService* sb_srv = m_ServiceLocator->GetServiceByType<IStatusBarService>();
        if (sb_srv != NULL && status != "") {
            sb_srv->SetStatusMessage(status);
        }
    }   

    // On mac (osx Cocoa) formatting can be lost after enter (when the wxTextCtrl
    // m_pSearch was disabled then re-enabled).  This re-computes the highlighting.
    // (note: formatting is not lost on other platforms when we do this, but no
    //  harm in re-setting it here)
    wxCommandEvent evt_dummy;
    OnSearchctrlTextEnter(evt_dummy);
}

bool CQueryParsePanel::x_FormatQueryString(string& query, bool casesensitive)
{
    // Strip beginning and ending blanks from beginning the string, 
    query = NStr::TruncateSpaces(query);

    // Return true if 'query' is a simple string so that the query engine
    // has the option of looking for that string in all the text instead
    // of doing a full query.
    // A simple string can take these forms:    
    // a.  asimplestringwithnoblanks
    // b.  "a simple string with no embedded ' double quotes"
    // c.  'another simple string with no embedded " single quotes'
    // d.  not_a>=simple_string_because_of_comparison_operator
    bool simple_string = false;
    vector<string> token_array1;
    vector<string> token_array2;
    vector<string> token_array3;

    // really short == simple string
    if (query.length() <= 1) {
        simple_string = true;
    }      
    // Check for (and remove if found) double quotes:
    else if (query[0] == '\"' && 
        query[query.length()-1] == '\"' &&
        query.find_first_of('\"', 1) == query.length()-1) {
            simple_string = true;
            query = query.substr(1, query.length()-2);
    }
    // Check for (and remove if found) single quotes:
    else if (query[0] == '\'' && 
        query[query.length()-1] == '\'' &&
        query.find_first_of('\'', 1) == query.length()-1) {
            simple_string = true;
            query = query.substr(1, query.length()-2);
    }
    // No embedded blanks or embedded comparison operators =>simple string
    else if (NStr::Tokenize(query, " <>=", token_array1).size() == 1 &&
             NStr::TokenizePattern(query, "&&", token_array2).size() == 1 &&
             NStr::TokenizePattern(query, "||", token_array3).size() == 1) {
        simple_string = true;
    }  

    return simple_string;
}

void CQueryParsePanel::x_StartQueryJob(CQueryParseTree* qtree, 
                                       bool casesensitive)
{
    CQueryExec* qexec = m_QueryDataSource->GetQueryExec(casesensitive);

    string result_status;

    // Pre-process the query.  This step determines, to the extent
    // possible, the types to be used in comaparison and whether a valid
    // number of arguments are present. Where possible, static types
    // (as opposed to identifers for field-values) are converted into
    // their required types for comparisons.
    try {
        CRef<objects::CScope> sc = m_QueryDataSource->GetScope();               
        CQueryExecPreProcessFunc  pre_process_exec(sc.GetPointerOrNull(),
                                                   *qexec);
        TreeDepthFirstTraverse(*(qtree->GetQueryTree()), 
                               pre_process_exec);
    }
    catch(CQueryExecException &e) {
        LOG_POST(Info << "Error pre-processing query: " << e.GetMsg());
    }

    // Create a job to run in the background so that the user can still access
    // the cancel button or interact with other views while the job is
    // running
    try {
        CQueryJob* j = m_Job.GetPointerOrNull();

        if (m_JobID != CAppJobDispatcher::eInvalidJobID) {
            CAppJobDispatcher& disp = CAppJobDispatcher::GetInstance();
            IAppJob::EJobState state = disp.GetJobState(m_JobID);
            if (state == IAppJob::eRunning)
                disp.CancelJob(m_JobID);
            else
                disp.DeleteJob(m_JobID);

            m_Job.Reset();
        }

        j = m_Job.GetPointerOrNull();

        if (j == NULL) {
            CQueryJob *qj = new CQueryJob(qexec, qtree);

            m_Job.Reset(qj);        
            m_JobID = CAppJobDispatcher::GetInstance().StartJob(*m_Job, "ThreadPool", *this, 1, true);
        }
    }
    // Unable to start a job - try to do physics in foreground
    catch (CAppJobException&) {
        _TRACE("Query job start failed");
    }
}

void CQueryParsePanel::x_ExecuteQuery(int search_dir)
{
    string status_msg;

    // Execute the query.  If it is a simple string query, do it synchronously,
    // otherwise start a asynchronous job to complete the query.  If the query
    // is unchanged from the lasttime this funcion was called, do not re-execute
    // the query.
    if (m_ToolBar2->GetToolEnabled(eCmdStartQuery)) {       
        m_Animation->Play();

        // Disable all input to window except the cancel button
        this->SetCursor(*wxHOURGLASS_CURSOR);
        this->Update();
        m_ToolBar2->EnableTool(eCmdStopQuery, true);
        m_ToolBar2->EnableTool(eCmdStartQuery, false);
        m_ToolBar2->EnableTool(eCmdNext, false);
        m_ToolBar2->EnableTool(eCmdPrevious, false);
        
        // Let host widget perform query-start actions (disable widget, set cursor...)
        m_QueryWidget->QueryStart();

        m_pSearch->Disable();
        m_CaseSensitive->Disable();
        m_pSelectAll->Disable();
        m_pHideUnselected->Disable();

        // Setting host widget cursor my also set it for toolbar but we don't want
        // that since the cancel button is still active.
        m_ToolBar->SetCursor(*wxSTANDARD_CURSOR);       
        m_ToolBar2->SetCursor(*wxSTANDARD_CURSOR);       

        try {
            string query = ToStdString(m_pSearch->GetValue());
            bool  casesensitive = m_CaseSensitive->IsChecked();

            bool simple_string = x_FormatQueryString(query, casesensitive);

            // Re-run the query if options that can change query results have changed
            if ( query!=m_PrevQuery || casesensitive!=m_PrevCaseSensitive) {  
                m_PrevQuery = query;        
                m_PrevCaseSensitive = casesensitive;    

                m_QueryDataSource->ClearQueryResults();

                if (simple_string) {
                    // simple string queries are completed synchronously since
                    // they can never get hung up in network access.
                    status_msg = m_QueryDataSource->ExecuteStringQuery(query, casesensitive);                      
                    LOG_POST(Info << "String Query: " << query);

                    x_AddRecentQuery(query);

                    // Notify widget query is completed
                    CompleteQuery(status_msg);
                }
                else {            
                    CQueryParseTree* qtree = new CQueryParseTree();

                    try {
                        // If query tree parsing is case-sensitive, keywords like
                        // AND and LIKE have to be uppercase, so we set that to
                        // be allows insensitive.  The casesensitive we use 
                        // refers to string comparisons during execution and 
                        // does not effect parsing.
                        qtree->Parse(query.c_str(), 
                                     CQueryParseTree::eCaseInsensitive,
                                     CQueryParseTree::eSyntaxCheck);

                        CNcbiOstrstream  strstrm;                
                        qtree->Print(strstrm);

                        // Logging merges lines (even if i call 
                        // UnsetDiagPostFlag(eDPF_PreMergeLines/eDPF_MergeLines); 
                        LOG_POST(Info << "Parsed Query: " << query);
                        vector<string> arr;
                        string s = (string)CNcbiOstrstreamToString(strstrm);
                        NStr::Tokenize(s, "\n", arr);
                        for (size_t i=0; i<arr.size(); ++i)
                            LOG_POST(Info << arr[i]);                
                    }
                    catch (CQueryParseException& e) {
                        LOG_POST(Info << "Error parsing query: " << e.GetMsg());
                        m_QueryDataSource->ClearQueryResults();
                        delete qtree;

                        // Return completion status to caller:                      
                        status_msg =  "Query execution failed: " + e.GetMsg();
                        CompleteQuery(status_msg);
                    }

                    // Regular queries are handled asynchronously since they
                    // may require network access, which could get tied up.
                    x_StartQueryJob(qtree, casesensitive);

                    x_AddRecentQuery(query);
                }
            }
            // This query was a duplicate of previous query, so just 
            // return without changing the selection set:
            else {
                CompleteQuery("Duplicate Query");

                // This re-selects the previous selection set from the
                // previous (duplicated) query.
                m_QueryWidget->SetSelectAll(IsSelectAll());
            }
        }
        // If this widget has been marked for deletion and has an exception,
        // return For a normal (query exectution) exception without deletion,
        // continue normal processing.
        catch (CException&) {
        }
    }
}

/// handles "state changed" notification from CAppJobDispatcher
void CQueryParsePanel::OnAppJobNotification(CEvent* evt)
{
    string result_status;

    CAppJobNotification* notn = dynamic_cast<CAppJobNotification*>(evt);
    _ASSERT(notn);

     CQueryParsePanel::CQueryJob* j = m_Job.GetPointerOrNull();

     // Make sure there is an active job and it is a job for this widget
     if (j == NULL || m_JobID != notn->GetJobID()) {
         return;
     }

     // If the job is finished, collect results and send completion message
     // to widget
     if (notn->GetState() == IAppJob::eCompleted && 
         !j->GetUserCancelled()) {
         LOG_POST(Info << "Number evaluated: " << 
             j->GetQueriedCount() << 
             "  Number of query execution errors: " << 
             j->GetExceptionCount());

         result_status = string("Number evaluated: ") + 
             NStr::IntToString(j->GetQueriedCount()) +
             string(". Number Selected: ") +
             NStr::IntToString(j->GetNumSelected()) + string(".");
         if (j->GetExceptionCount() > 0) {
             result_status += string(" Number with query errors: ") + 
                 NStr::IntToString(j->GetExceptionCount());
         }

         // Set the selection set to the returned entries 
         m_QueryDataSource->SetQueryResults(j->GetQueryExec());
     }
     // Done with job, but we don't want the results:
     else {
         LOG_POST(Info << "Number evaluated (for cancelled search): " << 
             j->GetQueriedCount() << 
             " Number with query execution errors: " << 
             j->GetExceptionCount());

         result_status = string("Number evaluated (for cancelled search): ") + 
             NStr::IntToString(j->GetQueriedCount()) +
             string(". Number Selected: ") +
             NStr::IntToString(j->GetNumSelected()) + string(".");
         if (j->GetExceptionCount() > 0) {
             result_status += string(" Number with query errors: ") + 
                 NStr::IntToString(j->GetExceptionCount());
         }


     }

     // Could send an event to allow listeners to know query is done
     // (but widget gets directly updated so this is not currently
     // needed)
     /*
     CQueryEvent *e = new CQueryEvent(CEvent::eEvent_Message, 
                                      CQueryEvent::eQueryCompletedEvent,
                                      this);
     e->SetStatus(result_status);
     Send(e);
     */
     CompleteQuery(result_status);

     // Delete the completed job object
     m_Job.Reset();
}


void CQueryParsePanel::x_CancelQueryInProgress()
{
    CQueryJob* j = m_Job.GetPointerOrNull();

    if (m_JobID != CAppJobDispatcher::eInvalidJobID && j != NULL) {
        CAppJobDispatcher& disp = CAppJobDispatcher::GetInstance();
        IAppJob::EJobState state = disp.GetJobState(m_JobID);
        if (state == IAppJob::eRunning) {
            disp.CancelJob(m_JobID);  
            j->SetUserCancelled();

            // by default, we do not run the same query twice (since the results
            // are already there) but if the job is cancelled, this allows the
            // query to be re-executed.
            m_PrevQuery = "";
        }
    }
}

void CQueryParsePanel::x_AddRecentQuery(const std::string& q)
{
    if (q == "")
        return;

    vector<pair<string,string> >::iterator iter;
    for (iter = m_RecentQueries.begin(); iter != m_RecentQueries.end(); ++iter) {
        if ((*iter).second == q) {
            m_RecentQueries.erase(iter);
            break;
        }
    }

    m_RecentQueries.insert(m_RecentQueries.begin(), pair<string,string>(string(), q));

    if (m_RecentQueries.size() > (unsigned int)m_MaxRecentQueries)
        m_RecentQueries.erase(m_RecentQueries.begin() + m_RecentQueries.size()-1);
}



END_NCBI_SCOPE
