/*  $Id: bioseq_editor.cpp 27922 2013-04-25 16:51:42Z katargir $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data,  the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties,  express or implied,  including
 *  warranties of performance,  merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Roman Katargin
 */

#include <ncbi_pch.hpp>

#include <gui/widgets/edit/edit_object.hpp>
#include <gui/widgets/edit/bioseq_editor.hpp>
#include <gui/widgets/edit/edit_object_dlg.hpp>

#include <gui/objutils/cmd_factory.hpp>

#include <objmgr/bioseq_ci.hpp>
#include <objmgr/seqdesc_ci.hpp>
#include <objmgr/feat_ci.hpp>
#include <objmgr/seq_entry_ci.hpp>

#include <objects/submit/Submit_block.hpp>

#include <objects/seqfeat/BioSource.hpp>
#include <objects/seqfeat/Cdregion.hpp>
#include <objects/seqfeat/Imp_feat.hpp>
#include <objects/seqfeat/RNA_ref.hpp>
#include <objects/seq/Pubdesc.hpp>
#include <objects/general/User_object.hpp>
#include <objects/general/Object_id.hpp>
#include <objects/seq/MolInfo.hpp>

#include <gui/objutils/seqdesc_title_edit.hpp>
#include <gui/objutils/cmd_del_desc.hpp>
#include <gui/objutils/cmd_create_desc.hpp>
#include <gui/objutils/cmd_composite.hpp>
#include <gui/objutils/cmd_del_seq_feat.hpp>
#include <gui/objutils/util_cmds.hpp>

#include <objmgr/seq_feat_handle.hpp>

#include <objmgr/util/sequence.hpp>

#include "deletefrombioseq_setdlg.hpp"

#include "edit_object_seq_feat.hpp"
#include "edit_object_seq_desc.hpp"
#include "edit_object_submit_block.hpp"
#include "edit_object_feature_propagate.hpp"

#include <gui/widgets/wx/wx_utils.hpp>
#include <gui/widgets/wx/ui_command.hpp>

#include <wx/menu.h>
#include <wx/msgdlg.h>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

string CBioseqEditorFactory::GetExtensionIdentifier() const
{
    static string sid("bioseq_editor_factory");
    return sid;
}

string CBioseqEditorFactory::GetExtensionLabel() const
{
    static string slabel("Bioseq Editor Factory");
    return slabel;
}

typedef struct featcmd {
  EBioseqEditCommands cmd;
  string feature_name;
} FeatCmd;

static const FeatCmd feat_cmds[] = 
{
  { eCmdCreateCodingRegion, "Coding Region" },
  { eCmdCreateGene, "gene" },
  { eCmdCreateProt, "prot" },
  { eCmdCreateLTR, "LTR" },
  { eCmdCreateRepeatRegion, "repeat_region" },
  { eCmdCreateStemLoop, "stem_loop" },
  { eCmdCreateSTS, "STS" },
  { eCmdCreateRegion, "Region" },
  { eCmdCreatePreRNA, "preRNA" },
  { eCmdCreateMRNA, "mRNA" },
  { eCmdCreateExon, "exon" },
  { eCmdCreateIntron, "intron" },
  { eCmdCreate3UTR, "3'UTR" },
  { eCmdCreate5UTR, "5'UTR" },
  { eCmdCreatetRNA, "tRNA" },
  { eCmdCreaterRNA, "rRNA" },
  { eCmdCreateRNA, "RNA" },
  { eCmdCreatencRNA, "ncRNA" },
  { eCmdCreatetmRNA, "tmRNA" },
  { eCmdCreatePolyASignal, "polyA_signal" },
  { eCmdCreatePolyASite, "polyA_site" },
  { eCmdCreatePrimerBind, "primer_bind" } ,
  { eCmdCreateProteinBind, "protein_bind" },
  { eCmdCreateRBS, "RBS" },
  { eCmdCreateRepOrigin, "rep_origin" },
  { eCmdCreateTATASignal, "TATA_signal" },
  { eCmdCreateTerminator, "terminator" },
  { eCmdCreate10Signal, "~10_signal" },
  { eCmdCreate35Signal, "~35_signal" },
  { eCmdCreateBond, "Bond" },
  { eCmdCreateSite, "Site" },
  { eCmdCreateOriT, "oriT" },
  { eCmdCreateAttenuator, "attenuator" },
  { eCmdCreateCRegion, "C_region" },
  { eCmdCreateCAATSignal, "CAAT_signal" },
  { eCmdCreateDLoop, "D-loop" },
  { eCmdCreateDSegment, "D_segment" },
  { eCmdCreateEnhancer, "enhancer" },
  { eCmdCreateGCSignal, "GC_signal" },
  { eCmdCreateJSegment, "J_segment" },
  { eCmdCreateMiscBinding, "misc_binding" },
  { eCmdCreateMiscDifference,  "misc_difference" },
  { eCmdCreateMiscFeature, "misc_feature" },
  { eCmdCreateMiscRecomb, "misc_recomb" },
  { eCmdCreateMiscSignal,   "misc_signal" },
  { eCmdCreateMiscStructure, "misc_structure" },
  { eCmdCreateModifiedBase, "modified_base" },
  { eCmdCreateNRegion, "N_region" } ,
  { eCmdCreatePrimTranscript, "prim_transcript" },
  { eCmdCreatePromoter, "promoter" },
  { eCmdCreateSRegion,  "S_REGION" },
  { eCmdCreateSatellite, "satellite" },
  { eCmdCreateUnsure, "unsure" },
  { eCmdCreateVRegion, "V_region" },
  { eCmdCreateVSegment, "V_segment" },
  { eCmdCreateVariation, "variation" },
  { eCmdCreateAssemblyGap, "assembly_gap" },
  { eCmdCreateStructuredComment, "StructuredComment" },
  { eCmdCreateDBLink, "DBLink" },
  { eCmdCreateRegionDescriptor, "Region" },
  { eCmdCreateCommentDescriptor, "Comment" },
  { eCmdCreateTitle, "Title" },
  { eCmdCreateMolInfo, "Molecule Description" },
  { eCmdCreatePubDescriptor, "Publication" }

};

static const int num_feat_cmds = sizeof (feat_cmds) / sizeof (FeatCmd);


string GetKeyFromCmdID (int cmd_id)
{
    for (int i = 0; i < num_feat_cmds; i++) {
        if (feat_cmds[i].cmd == cmd_id) {
            return feat_cmds[i].feature_name;
        }
    }
    return "";
}


void CBioseqEditorFactory::RegisterCommands(CUICommandRegistry& cmd_reg, wxFileArtProvider&)
{
    static bool registered = false;
    if (!registered) {
        registered = true;

        cmd_reg.RegisterCommand(eCmdEditObjects,
            "Edit Selected Object(s)",
            "Edit Selected Object(s)",
            NcbiEmptyString,
            NcbiEmptyString,
            "Command opens Dialog to Edit Selected Object(s)");
        cmd_reg.RegisterCommand(eCmdDeleteObjects,
            "Delete Selected Object(s)",
            "Delete Selected Object(s)",
            NcbiEmptyString,
            NcbiEmptyString,
            "Command deletes Selected Object(s)");
        cmd_reg.RegisterCommand(eCmdEditBiosourceDesc,
            "Edit BioSource Descriptor",
            "Edit BioSource Descriptor",
            NcbiEmptyString,
            NcbiEmptyString,
            "Command opens Dialog to Edit Biosource Descriptor of the Current Bioseq");
        cmd_reg.RegisterCommand(eCmdDeleteBiosourceDesc,
            "Delete BioSource Descriptor",
            "Delete BioSource Descriptor",
            NcbiEmptyString,
            NcbiEmptyString,
            "Command deletes Biosource Descriptor of the Current Bioseq");
        cmd_reg.RegisterCommand(eCmdCreateBiosourceDesc,
            "Create BioSource Descriptor",
            "Create BioSource Descriptor",
            NcbiEmptyString,
            NcbiEmptyString,
            "Command opens Dialog to Create Biosource Descriptor of the Current Bioseq");
        cmd_reg.RegisterCommand(eCmdEditBiosourceFeat,
            "Edit BioSource Feature",
            "Edit BioSource Feature",
            NcbiEmptyString,
            NcbiEmptyString,
            "Command opens Dialog to Edit Biosource Feature of the Current Bioseq");
        cmd_reg.RegisterCommand(eCmdDeleteBiosourceFeat,
            "Delete BioSource Feature",
            "Delete BioSource Feature",
            NcbiEmptyString,
            NcbiEmptyString,
            "Command deletes Biosource Feature of the Current Bioseq");
        cmd_reg.RegisterCommand(eCmdCreateBiosourceFeat,
            "Create BioSource Feature",
            "Create BioSource Feature",
            NcbiEmptyString,
            NcbiEmptyString,
            "Command opens Dialog to Create Feature Descriptor of the Current Bioseq");

        cmd_reg.RegisterCommand(eCmdPropagateAllFeatures,
            "Propagate All Features",
            "Propagate All Features",
            NcbiEmptyString,
            NcbiEmptyString,
            "Command opens Dialog to Propagate All Features from the Current Bioseq");

        cmd_reg.RegisterCommand(eCmdPropagateSelectedFeatures,
            "Propagate Selected Feature(s)",
            "Propagate Selected Feature(s)",
            NcbiEmptyString,
            NcbiEmptyString,
            "Command opens Dialog to Propagate Selected Feature from the Current Bioseq");

        for (int i = 0; i < num_feat_cmds; i++) {
            cmd_reg.RegisterCommand(feat_cmds[i].cmd,
                feat_cmds[i].feature_name,
                feat_cmds[i].feature_name,
                NcbiEmptyString,
                NcbiEmptyString,
                "Command opens Dialog to Create " + feat_cmds[i].feature_name + " on the Current Bioseq");
        }
    }
}

CObject* CBioseqEditorFactory::CreateEditor(const string& interface_name, ICommandProccessor& cmdProccessor)
{
    if (typeid(IBioseqEditor).name() == interface_name) {
        return new CBioseqEditor(cmdProccessor);
    }
    return 0;
}

static
WX_DEFINE_MENU(kBioseqEditorContextMenu)
    WX_MENU_SEPARATOR_L("Edit Selection")
    WX_MENU_ITEM(eCmdEditObjects)
    WX_MENU_ITEM(eCmdDeleteObjects)
    WX_MENU_ITEM(eCmdPropagateSelectedFeatures)
WX_END_MENU()

wxMenu* CBioseqEditor::CreateContextMenu() const
{
    return CUICommandRegistry::GetInstance().CreateMenu(kBioseqEditorContextMenu);
}

static
WX_DEFINE_MENU(kBioseqEditorBioseqMenu)
    WX_SUBMENU("BioSource")
        WX_MENU_SEPARATOR_L("Edit Bioseq")
        WX_MENU_ITEM(eCmdEditBiosourceDesc)
        WX_MENU_ITEM(eCmdDeleteBiosourceDesc)
        WX_MENU_ITEM(eCmdCreateBiosourceDesc)
        WX_MENU_ITEM(eCmdEditBiosourceFeat)
        WX_MENU_ITEM(eCmdDeleteBiosourceFeat)
        WX_MENU_ITEM(eCmdCreateBiosourceFeat)
    WX_END_SUBMENU()
    WX_SUBMENU("Create Features")
        WX_MENU_SEPARATOR_L("Edit Bioseq")
        WX_MENU_ITEM(eCmdCreateBiosourceFeat)
        WX_SUBMENU("Genes and Named Regions")
          WX_MENU_ITEM(eCmdCreateGene)
          WX_MENU_ITEM(eCmdCreateProt)
          WX_MENU_ITEM(eCmdCreateLTR)
          WX_MENU_ITEM(eCmdCreateRepeatRegion)
          WX_MENU_ITEM(eCmdCreateStemLoop)
          WX_MENU_ITEM(eCmdCreateSTS)
          WX_MENU_ITEM(eCmdCreateRegion)
        WX_END_SUBMENU()
        WX_SUBMENU("Coding Regions and Transcripts")
          WX_MENU_ITEM(eCmdCreateCodingRegion)
          WX_MENU_ITEM(eCmdCreatePreRNA)
          WX_MENU_ITEM(eCmdCreateMRNA)
          WX_MENU_ITEM(eCmdCreateExon)
          WX_MENU_ITEM(eCmdCreateIntron)
          WX_MENU_ITEM(eCmdCreate3UTR)
          WX_MENU_ITEM(eCmdCreate5UTR)
        WX_END_SUBMENU()
        WX_SUBMENU("Structural RNAs")
          WX_MENU_ITEM(eCmdCreatetRNA)
          WX_MENU_ITEM(eCmdCreaterRNA)
          WX_MENU_ITEM(eCmdCreateRNA)
          WX_MENU_ITEM(eCmdCreatencRNA)
          WX_MENU_ITEM(eCmdCreatetmRNA)
        WX_END_SUBMENU()
        WX_SUBMENU("Sites and Bonds")
          WX_MENU_ITEM(eCmdCreatePolyASignal)
          WX_MENU_ITEM(eCmdCreatePolyASite)
          WX_MENU_ITEM(eCmdCreatePrimerBind)
          WX_MENU_ITEM(eCmdCreateProteinBind)
          WX_MENU_ITEM(eCmdCreateRBS)
          WX_MENU_ITEM(eCmdCreateRepOrigin)
          WX_MENU_ITEM(eCmdCreateTATASignal)
          WX_MENU_ITEM(eCmdCreateTerminator)
          WX_MENU_ITEM(eCmdCreate10Signal)
          WX_MENU_ITEM(eCmdCreate35Signal)
          WX_MENU_ITEM(eCmdCreateBond)
          WX_MENU_ITEM(eCmdCreateSite)
          WX_MENU_ITEM(eCmdCreateOriT)
        WX_END_SUBMENU()
        WX_SUBMENU("Remaining Features")
          WX_MENU_ITEM(eCmdCreateAttenuator)
          WX_MENU_ITEM(eCmdCreateCRegion)
          WX_MENU_ITEM(eCmdCreateCAATSignal)
          WX_MENU_ITEM(eCmdCreateDLoop)
          WX_MENU_ITEM(eCmdCreateDSegment)
          WX_MENU_ITEM(eCmdCreateEnhancer)
          WX_MENU_ITEM(eCmdCreateGCSignal)
          WX_MENU_ITEM(eCmdCreateJSegment)
          WX_MENU_ITEM(eCmdCreateMiscBinding)
          WX_MENU_ITEM(eCmdCreateMiscDifference)
          WX_MENU_ITEM(eCmdCreateMiscFeature)
          WX_MENU_ITEM(eCmdCreateMiscRecomb)
          WX_MENU_ITEM(eCmdCreateMiscSignal) 
          WX_MENU_ITEM(eCmdCreateMiscStructure)
          WX_MENU_ITEM(eCmdCreateModifiedBase)
          WX_MENU_ITEM(eCmdCreateNRegion)
          WX_MENU_ITEM(eCmdCreatePrimTranscript)
          WX_MENU_ITEM(eCmdCreatePromoter)
          WX_MENU_ITEM(eCmdCreateSRegion)
          WX_MENU_ITEM(eCmdCreateSatellite)
          WX_MENU_ITEM(eCmdCreateUnsure)
          WX_MENU_ITEM(eCmdCreateVRegion)
          WX_MENU_ITEM(eCmdCreateVSegment)
          WX_MENU_ITEM(eCmdCreateVariation)
          WX_MENU_ITEM(eCmdCreateAssemblyGap)
        WX_END_SUBMENU()
        WX_MENU_ITEM(eCmdPropagateAllFeatures)
    WX_END_SUBMENU()        
    WX_SUBMENU("Create Descriptors")
        WX_MENU_SEPARATOR_L("Edit Bioseq")
        WX_MENU_ITEM(eCmdCreateStructuredComment)
        WX_MENU_ITEM(eCmdCreateDBLink)
        WX_MENU_ITEM(eCmdCreateRegionDescriptor)
        WX_MENU_ITEM(eCmdCreateCommentDescriptor)
        WX_MENU_ITEM(eCmdCreateTitle)
        WX_MENU_ITEM(eCmdCreateMolInfo)
        WX_MENU_ITEM(eCmdCreatePubDescriptor)
    WX_END_SUBMENU()        
WX_END_MENU()

wxMenu* CBioseqEditor::CreateBioseqMenu() const
{
    return CUICommandRegistry::GetInstance().CreateMenu(kBioseqEditorBioseqMenu);
}

BEGIN_EVENT_TABLE( CBioseqEditor, wxEvtHandler )
    EVT_MENU(eCmdEditBiosourceDesc, CBioseqEditor::OnEditBiosourceDesc)
    EVT_UPDATE_UI(eCmdEditBiosourceDesc, CBioseqEditor::OnUpdateEditBiosourceDesc)
    EVT_MENU(eCmdCreateBiosourceDesc, CBioseqEditor::OnCreateBiosourceDesc)
    EVT_UPDATE_UI(eCmdCreateBiosourceDesc, CBioseqEditor::OnUpdateCreateBiosourceDesc)
    EVT_MENU(eCmdDeleteBiosourceDesc, CBioseqEditor::OnDeleteBiosourceDesc)
    EVT_UPDATE_UI(eCmdDeleteBiosourceDesc, CBioseqEditor::OnUpdateDeleteBiosourceDesc)

    EVT_MENU(eCmdEditBiosourceFeat, CBioseqEditor::OnEditBiosourceFeat)
    EVT_UPDATE_UI(eCmdEditBiosourceFeat, CBioseqEditor::OnUpdateEditBiosourceFeat)
    EVT_MENU(eCmdCreateBiosourceFeat, CBioseqEditor::OnCreateBiosourceFeat)
    EVT_UPDATE_UI(eCmdCreateBiosourceFeat, CBioseqEditor::OnUpdateCreateBiosourceFeat)
    EVT_MENU(eCmdDeleteBiosourceFeat, CBioseqEditor::OnDeleteBiosourceFeat)
    EVT_UPDATE_UI(eCmdDeleteBiosourceFeat, CBioseqEditor::OnUpdateDeleteBiosourceFeat)
    EVT_MENU(eCmdPropagateAllFeatures, CBioseqEditor::OnPropagateAllFeatures)
    EVT_UPDATE_UI(eCmdPropagateAllFeatures, CBioseqEditor::OnUpdatePropagateAllFeatures)

    EVT_MENU(eCmdCreateCodingRegion, CBioseqEditor::OnCreateCodingRegion)
    EVT_UPDATE_UI(eCmdCreateCodingRegion, CBioseqEditor::OnUpdateNucleotideFeat)
    EVT_MENU(eCmdCreateGene, CBioseqEditor::OnCreateGene)
    EVT_UPDATE_UI(eCmdCreateGene, CBioseqEditor::OnUpdateNucleotideFeat)
    EVT_MENU(eCmdCreateProt, CBioseqEditor::OnCreateProt)
    EVT_UPDATE_UI(eCmdCreateProt, CBioseqEditor::OnUpdateProteinFeat)
    EVT_MENU(eCmdCreateLTR, CBioseqEditor::OnCreateImport)
    EVT_UPDATE_UI(eCmdCreateLTR, CBioseqEditor::OnUpdateNucleotideFeat)
    EVT_MENU(eCmdCreateRepeatRegion, CBioseqEditor::OnCreateImport)
    EVT_UPDATE_UI(eCmdCreateRepeatRegion, CBioseqEditor::OnUpdateNucleotideFeat)
    EVT_MENU(eCmdCreateStemLoop, CBioseqEditor::OnCreateImport)
    EVT_UPDATE_UI(eCmdCreateStemLoop, CBioseqEditor::OnUpdateNucleotideFeat)
    EVT_MENU(eCmdCreateSTS, CBioseqEditor::OnCreateImport)
    EVT_UPDATE_UI(eCmdCreateSTS, CBioseqEditor::OnUpdateNucleotideFeat)
    EVT_MENU(eCmdCreateRegion, CBioseqEditor::OnCreateRegion)

    EVT_MENU(eCmdCreatePreRNA, CBioseqEditor::OnCreateRNA)
    EVT_UPDATE_UI(eCmdCreatePreRNA, CBioseqEditor::OnUpdateNucleotideFeat)
    EVT_MENU(eCmdCreateMRNA, CBioseqEditor::OnCreateRNA)
    EVT_UPDATE_UI(eCmdCreateMRNA, CBioseqEditor::OnUpdateNucleotideFeat)
    EVT_MENU(eCmdCreateExon, CBioseqEditor::OnCreateImport)
    EVT_UPDATE_UI(eCmdCreateExon, CBioseqEditor::OnUpdateNucleotideFeat)
    EVT_MENU(eCmdCreateIntron, CBioseqEditor::OnCreateImport)
    EVT_UPDATE_UI(eCmdCreateIntron, CBioseqEditor::OnUpdateNucleotideFeat)
    EVT_MENU(eCmdCreate3UTR, CBioseqEditor::OnCreateImport)
    EVT_UPDATE_UI(eCmdCreate3UTR, CBioseqEditor::OnUpdateNucleotideFeat)
    EVT_MENU(eCmdCreate5UTR, CBioseqEditor::OnCreateImport)
    EVT_UPDATE_UI(eCmdCreate5UTR, CBioseqEditor::OnUpdateNucleotideFeat)

    EVT_MENU(eCmdCreatetRNA, CBioseqEditor::OnCreateRNA)
    EVT_UPDATE_UI(eCmdCreatetRNA, CBioseqEditor::OnUpdateNucleotideFeat)
    EVT_MENU(eCmdCreaterRNA, CBioseqEditor::OnCreateRNA)
    EVT_UPDATE_UI(eCmdCreaterRNA, CBioseqEditor::OnUpdateNucleotideFeat)
    EVT_MENU(eCmdCreateRNA, CBioseqEditor::OnCreateRNA)
    EVT_UPDATE_UI(eCmdCreateRNA, CBioseqEditor::OnUpdateNucleotideFeat)
    EVT_MENU(eCmdCreatencRNA, CBioseqEditor::OnCreateRNA)
    EVT_UPDATE_UI(eCmdCreatencRNA, CBioseqEditor::OnUpdateNucleotideFeat)
    EVT_MENU(eCmdCreatetmRNA, CBioseqEditor::OnCreateRNA)
    EVT_UPDATE_UI(eCmdCreatetmRNA, CBioseqEditor::OnUpdateNucleotideFeat)

    EVT_MENU(eCmdCreatePolyASignal, CBioseqEditor::OnCreateImport)
    EVT_UPDATE_UI(eCmdCreatePolyASignal, CBioseqEditor::OnUpdateNucleotideFeat)
    EVT_MENU(eCmdCreatePolyASite, CBioseqEditor::OnCreateImport)
    EVT_UPDATE_UI(eCmdCreatePolyASite, CBioseqEditor::OnUpdateNucleotideFeat)
    EVT_MENU(eCmdCreatePrimerBind, CBioseqEditor::OnCreateImport)
    EVT_UPDATE_UI(eCmdCreatePrimerBind, CBioseqEditor::OnUpdateNucleotideFeat)
    EVT_MENU(eCmdCreateProteinBind, CBioseqEditor::OnCreateImport)
    EVT_UPDATE_UI(eCmdCreateProteinBind, CBioseqEditor::OnUpdateNucleotideFeat)
    EVT_MENU(eCmdCreateRBS, CBioseqEditor::OnCreateImport)
    EVT_UPDATE_UI(eCmdCreateRBS, CBioseqEditor::OnUpdateNucleotideFeat)
    EVT_MENU(eCmdCreateRepOrigin, CBioseqEditor::OnCreateImport)
    EVT_UPDATE_UI(eCmdCreateRepOrigin, CBioseqEditor::OnUpdateNucleotideFeat)
    EVT_MENU(eCmdCreateTATASignal, CBioseqEditor::OnCreateImport)
    EVT_UPDATE_UI(eCmdCreateTATASignal, CBioseqEditor::OnUpdateNucleotideFeat)
    EVT_MENU(eCmdCreateTerminator, CBioseqEditor::OnCreateImport)
    EVT_UPDATE_UI(eCmdCreateTerminator, CBioseqEditor::OnUpdateNucleotideFeat)
    EVT_MENU(eCmdCreate10Signal, CBioseqEditor::OnCreateImport)
    EVT_UPDATE_UI(eCmdCreate10Signal, CBioseqEditor::OnUpdateNucleotideFeat)
    EVT_MENU(eCmdCreate35Signal, CBioseqEditor::OnCreateImport)
    EVT_UPDATE_UI(eCmdCreate35Signal, CBioseqEditor::OnUpdateNucleotideFeat)
    EVT_MENU(eCmdCreateBond, CBioseqEditor::OnCreateBond)
    EVT_UPDATE_UI(eCmdCreateBond, CBioseqEditor::OnUpdateProteinFeat)
    EVT_MENU(eCmdCreateSite, CBioseqEditor::OnCreateSite)
    EVT_MENU(eCmdCreateOriT, CBioseqEditor::OnCreateImport)
    EVT_UPDATE_UI(eCmdCreateOriT, CBioseqEditor::OnUpdateNucleotideFeat)

    EVT_MENU(eCmdCreateAttenuator, CBioseqEditor::OnCreateImport)
    EVT_UPDATE_UI(eCmdCreateAttenuator, CBioseqEditor::OnUpdateNucleotideFeat)
    EVT_MENU(eCmdCreateCRegion, CBioseqEditor::OnCreateImport)
    EVT_UPDATE_UI(eCmdCreateCRegion, CBioseqEditor::OnUpdateNucleotideFeat)
    EVT_MENU(eCmdCreateCAATSignal, CBioseqEditor::OnCreateImport)
    EVT_UPDATE_UI(eCmdCreateCAATSignal, CBioseqEditor::OnUpdateNucleotideFeat)
    EVT_MENU(eCmdCreateDLoop, CBioseqEditor::OnCreateImport)
    EVT_UPDATE_UI(eCmdCreateDLoop, CBioseqEditor::OnUpdateNucleotideFeat)
    EVT_MENU(eCmdCreateDSegment, CBioseqEditor::OnCreateImport)
    EVT_UPDATE_UI(eCmdCreateDSegment, CBioseqEditor::OnUpdateNucleotideFeat)
    EVT_MENU(eCmdCreateEnhancer, CBioseqEditor::OnCreateImport)
    EVT_UPDATE_UI(eCmdCreateEnhancer, CBioseqEditor::OnUpdateNucleotideFeat)
    EVT_MENU(eCmdCreateGCSignal , CBioseqEditor::OnCreateImport)
    EVT_UPDATE_UI(eCmdCreateGCSignal , CBioseqEditor::OnUpdateNucleotideFeat)
    EVT_MENU(eCmdCreateJSegment, CBioseqEditor::OnCreateImport)
    EVT_UPDATE_UI(eCmdCreateJSegment, CBioseqEditor::OnUpdateNucleotideFeat)
    EVT_MENU(eCmdCreateMiscBinding, CBioseqEditor::OnCreateImport)
    EVT_UPDATE_UI(eCmdCreateMiscBinding, CBioseqEditor::OnUpdateNucleotideFeat)
    EVT_MENU(eCmdCreateMiscDifference , CBioseqEditor::OnCreateImport)
    EVT_UPDATE_UI(eCmdCreateMiscDifference , CBioseqEditor::OnUpdateNucleotideFeat)
    EVT_MENU(eCmdCreateMiscFeature, CBioseqEditor::OnCreateImport)
    EVT_UPDATE_UI(eCmdCreateMiscFeature, CBioseqEditor::OnUpdateNucleotideFeat)
    EVT_MENU(eCmdCreateMiscRecomb, CBioseqEditor::OnCreateImport)
    EVT_UPDATE_UI(eCmdCreateMiscRecomb, CBioseqEditor::OnUpdateNucleotideFeat)
    EVT_MENU(eCmdCreateMiscSignal, CBioseqEditor::OnCreateImport)
    EVT_UPDATE_UI(eCmdCreateMiscSignal, CBioseqEditor::OnUpdateNucleotideFeat)
    EVT_MENU(eCmdCreateMiscStructure, CBioseqEditor::OnCreateImport)
    EVT_UPDATE_UI(eCmdCreateMiscStructure, CBioseqEditor::OnUpdateNucleotideFeat)
    EVT_MENU(eCmdCreateModifiedBase, CBioseqEditor::OnCreateImport)
    EVT_UPDATE_UI(eCmdCreateModifiedBase, CBioseqEditor::OnUpdateNucleotideFeat)
    EVT_MENU(eCmdCreateNRegion, CBioseqEditor::OnCreateImport)
    EVT_UPDATE_UI(eCmdCreateNRegion, CBioseqEditor::OnUpdateNucleotideFeat)
    EVT_MENU(eCmdCreatePrimTranscript, CBioseqEditor::OnCreateImport)
    EVT_UPDATE_UI(eCmdCreatePrimTranscript, CBioseqEditor::OnUpdateNucleotideFeat)
    EVT_MENU(eCmdCreatePromoter, CBioseqEditor::OnCreateImport)
    EVT_UPDATE_UI(eCmdCreatePromoter, CBioseqEditor::OnUpdateNucleotideFeat)
    EVT_MENU(eCmdCreateSRegion, CBioseqEditor::OnCreateImport)
    EVT_UPDATE_UI(eCmdCreateSRegion, CBioseqEditor::OnUpdateNucleotideFeat)
    EVT_MENU(eCmdCreateSatellite, CBioseqEditor::OnCreateImport)
    EVT_UPDATE_UI(eCmdCreateSatellite, CBioseqEditor::OnUpdateNucleotideFeat)
    EVT_MENU(eCmdCreateUnsure, CBioseqEditor::OnCreateImport)
    EVT_UPDATE_UI(eCmdCreateUnsure, CBioseqEditor::OnUpdateNucleotideFeat)
    EVT_MENU(eCmdCreateVRegion, CBioseqEditor::OnCreateImport)
    EVT_UPDATE_UI(eCmdCreateVRegion, CBioseqEditor::OnUpdateNucleotideFeat)
    EVT_MENU(eCmdCreateVSegment, CBioseqEditor::OnCreateImport)
    EVT_UPDATE_UI(eCmdCreateVSegment, CBioseqEditor::OnUpdateNucleotideFeat)
    EVT_MENU(eCmdCreateVariation, CBioseqEditor::OnCreateImport)
    EVT_UPDATE_UI(eCmdCreateVariation, CBioseqEditor::OnUpdateNucleotideFeat)
    EVT_MENU(eCmdCreateAssemblyGap, CBioseqEditor::OnCreateImport)
    EVT_UPDATE_UI(eCmdCreateAssemblyGap, CBioseqEditor::OnUpdateNucleotideFeat)

    EVT_MENU(eCmdCreateStructuredComment, CBioseqEditor::OnCreateDescriptor)
    EVT_MENU(eCmdCreateDBLink, CBioseqEditor::OnCreateDescriptor)
    EVT_MENU(eCmdCreateRegionDescriptor, CBioseqEditor::OnCreateDescriptor)
    EVT_MENU(eCmdCreateCommentDescriptor, CBioseqEditor::OnCreateDescriptor)
    EVT_MENU(eCmdCreateTitle, CBioseqEditor::OnCreateDescriptor)
    EVT_MENU(eCmdCreateMolInfo, CBioseqEditor::OnCreateDescriptor)
    EVT_MENU(eCmdCreatePubDescriptor, CBioseqEditor::OnCreateDescriptor)

    EVT_MENU(eCmdEditObjects, CBioseqEditor::OnEditSelection)
    EVT_UPDATE_UI(eCmdEditObjects, CBioseqEditor::OnUpdateEditSelection)

    EVT_MENU(eCmdDeleteObjects, CBioseqEditor::OnDeleteSelection)
    EVT_UPDATE_UI(eCmdDeleteObjects, CBioseqEditor::OnUpdateDeleteSelection)

    EVT_MENU(eCmdPropagateSelectedFeatures, CBioseqEditor::OnPropagateSelectedFeatures)
    EVT_UPDATE_UI(eCmdPropagateSelectedFeatures, CBioseqEditor::OnUpdatePropagateSelectedFeatures)
END_EVENT_TABLE()

bool CBioseqEditor::ProcessEvent(wxEvent& event)
{
    if (!m_CB)
        return false;

    return wxEvtHandler::ProcessEvent(event);
}

void CBioseqEditor::OnUpdateEditBiosourceDesc(wxUpdateUIEvent& event)
{
    event.Enable(x_HaveBiosourceDesc());
}

void CBioseqEditor::OnUpdateCreateBiosourceDesc(wxUpdateUIEvent& event)
{
    event.Enable(!x_HaveBiosourceDesc());
}

void CBioseqEditor::OnUpdateDeleteBiosourceDesc(wxUpdateUIEvent& event)
{
    event.Enable(x_HaveBiosourceDesc());
}

void CBioseqEditor::OnEditBiosourceDesc(wxCommandEvent& /*event*/)
{
    const char* szCommand = "Edit Biosource Descriptor";

    try {
        CBioseq_Handle bh = m_CB->GetCurrentBioseq();
        if (!bh)
            return;

        CSeqdesc_CI desc_it(bh, CSeqdesc::e_Source);
        if (!desc_it)
            return;

        CSeq_entry_Handle seh = bh.GetSeq_entry_Handle();
        CIRef<IEditObject> editor(new CEditObjectSeq_desc(*desc_it, seh, bh.GetScope(), false));
        ERR_POST(Info << "Start " << szCommand);
        x_EditObject(*editor);
        ERR_POST(Info << "End " << szCommand);
    } catch (const std::exception& e) {
        x_ReportError(szCommand, e);
    }
}

void CBioseqEditor::OnCreateBiosourceDesc(wxCommandEvent& /*event*/)
{
    const char* szCommand = "Create Biosource Descriptor";

    try {
        CBioseq_Handle bh = m_CB->GetCurrentBioseq();
        if (!bh)
            return;

        CSeqdesc_CI desc_it(bh, CSeqdesc::e_Source);
        if (desc_it) {
            ERR_POST(Info << "Error: " << szCommand << ": Biosource descriptor already exists");
            return;
        }

        CRef<CSeqdesc> secdesc(new CSeqdesc());
        CRef<CBioSource> bioSource(new CBioSource());
        secdesc->SetSource(*bioSource);

        CSeq_entry_Handle seh = bh.GetSeq_entry_Handle();
        CIRef<IEditObject> editor(new CEditObjectSeq_desc(*secdesc, seh, bh.GetScope(), true));
        ERR_POST(Info << "Start " << szCommand);
        x_EditObject(*editor);
        ERR_POST(Info << "End " << szCommand);

    } catch (const std::exception& e) {
        x_ReportError(szCommand, e);
    }
}

void CBioseqEditor::OnDeleteBiosourceDesc(wxCommandEvent& event)
{
    const char* szCommand = "Delete Biosource Descriptor";

    try {
        CBioseq_Handle bh = m_CB->GetCurrentBioseq();
        if (!bh)
            return;

        CSeqdesc_CI desc_it(bh, CSeqdesc::e_Source);
        if (!desc_it)
            return;

        CSeq_entry_Handle seh = bh.GetSeq_entry_Handle();
        x_DeleteObject(seh, *desc_it);
    } catch (const std::exception& e) {
        x_ReportError(szCommand, e);
    }
}

void CBioseqEditor::OnUpdateEditBiosourceFeat(wxUpdateUIEvent& event)
{
    event.Enable(x_HaveBiosourceFeat());
}

void CBioseqEditor::OnUpdateCreateBiosourceFeat(wxUpdateUIEvent& event)
{
    event.Enable(!x_HaveBiosourceFeat());
}

void CBioseqEditor::OnUpdateDeleteBiosourceFeat(wxUpdateUIEvent& event)
{
    event.Enable(x_HaveBiosourceFeat());
}

void CBioseqEditor::OnEditBiosourceFeat(wxCommandEvent& /*event*/)
{
    const char* szCommand = "Edit Biosource Feature";

    try {
        CBioseq_Handle bh = m_CB->GetCurrentBioseq();
        if (!bh)
            return;

        CFeat_CI fsrc_it(bh, CSeqFeatData::e_Biosrc);
        if (!fsrc_it)
            return;

        CSeq_feat_Handle fh = *fsrc_it;
        CSeq_entry_Handle seh = bh.GetSeq_entry_Handle();

        CIRef<IEditObject> editor(new CEditObjectSeq_feat(*fh.GetOriginalSeq_feat(), seh, bh.GetScope(), false));
        ERR_POST(Info << "Start " << szCommand);
        x_EditObject(*editor);
        ERR_POST(Info << "End " << szCommand);
    } catch (const std::exception& e) {
        x_ReportError(szCommand, e);
    }
}

/*fl*/
void CBioseqEditor::OnPropagateAllFeatures(wxCommandEvent& /*event*/)
{
    //Context:
    // The user selected "Propagate All Features" from the edit link next to a
    // bioseq. The intent is to propagate every feature of that bioseq to one
    // or all other bioseqs in the set.
    // Where to and how to propagate still needs to be settled but the features to be
    // propagated are already fixed. It needs therefore be passed into all the actions
    // that will be triggered from here.
    //
    const char* szCommand = "Feature Propagate";
    try {
        //Generate list of features to be propagated. That would all features of
        // the current bioseq.
        CBioseq_Handle bh = m_CB->GetCurrentBioseq();
        if (!bh) {
            return;
        }
        vector<const CSeq_feat*> propagatedFeats;
        for (CFeat_CI ci(bh); ci; ++ci) {
            const CMappedFeat mf = *ci;
            propagatedFeats.push_back(mf.GetSeq_feat().GetPointer());
        }

        //Create and trigger the magic object editor that will take care of the 
        // rest 
        CIRef<IEditObject> editor(
            new CEditObjectFeaturePropagate(bh, propagatedFeats));
        ERR_POST(Info << "Start " << szCommand);
        x_EditObject(*editor);
        ERR_POST(Info << "End " << szCommand);

    } catch (const std::exception& e) {
        x_ReportError(szCommand, e);
    }
}

void CBioseqEditor::OnUpdatePropagateAllFeatures(wxUpdateUIEvent& /*event*/)
{
}

/*fl*/
void CBioseqEditor::OnPropagateSelectedFeatures(wxCommandEvent& /*event*/)
{
    //Context
    // The user highlighted one or more features of one of the bioseqs displayed. The
    // intent is to propagate all of the selected features to one or all of the other
    // sequences.
    // Where to and how to propagate still needs to be settled but the features to be
    // propagated are already fixed. It needs therefore be passed into all the actions
    // that will be triggered from here.
    //
    const char* szCommand = "Feature Propagate";

    try {
        //Generate list of features to be propagated. That would be the currently selected
        // ones
        TEditObjectList objList;
        m_CB->GetCurrentSelection(objList);
        if (objList.empty()) {
            return;
        }
        vector<const CSeq_feat*> propagatedFeats;
        for (TEditObjectList::iterator it = objList.begin(); it != objList.end(); ++it) {
            const CSeq_feat* pSeqFeat = 
                dynamic_cast<const CSeq_feat*>(it->m_Obj.GetPointer());
            if (pSeqFeat) {
                propagatedFeats.push_back(pSeqFeat);
            }
        }

        CSeq_entry_Handle seh = objList.front().m_SEH;
        CBioseq_Handle bh = seh.GetSeq();
        if (!bh) {
            return;
        }
        //Create and trigger the magic object editor that will take care of the 
        // rest 
        CIRef<IEditObject> editor(
            new CEditObjectFeaturePropagate(bh, propagatedFeats));
        ERR_POST(Info << "Start " << szCommand);
        x_EditObject(*editor);
        ERR_POST(Info << "End " << szCommand);

    } catch (const std::exception& e) {
        x_ReportError(szCommand, e);
    }
}

void CBioseqEditor::OnUpdatePropagateSelectedFeatures(wxUpdateUIEvent& /*event*/)
{
}

void CBioseqEditor::OnCreateBiosourceFeat(wxCommandEvent& /*event*/)
{
    const char* szCommand = "Create Biosource Feature";

    try {
        CBioseq_Handle bh = m_CB->GetCurrentBioseq();
        if (!bh)
            return;

        CFeat_CI fsrc_it(bh, CSeqFeatData::e_Biosrc);
        if (fsrc_it) {
            ERR_POST(Info << "Error: " << szCommand << ": Biosource feature already exists");
            return;
        }

        CRef<CSeq_feat> seqFeat(new CSeq_feat());
        CRef<CSeqFeatData> featData(new CSeqFeatData());
        CRef<CBioSource> bioSource(new CBioSource());
        seqFeat->SetData(*featData);
        featData->SetBiosrc(*bioSource);

        CRef<CSeq_id> seq_id(new CSeq_id);
        CConstRef<CSeq_id> bseq_seq_id = bh.GetSeqId();
        TSeqPos length = bh.GetBioseqLength();
        seq_id->Assign(*bseq_seq_id);
        CRef<CSeq_loc> seqloc(new CSeq_loc(*seq_id, 0, length - 1));
        seqFeat->SetLocation(*seqloc);

        CSeq_entry_Handle seh = bh.GetSeq_entry_Handle();
        CIRef<IEditObject> editor(new CEditObjectSeq_feat(*seqFeat, seh, seh.GetScope(), true));
        ERR_POST(Info << "Start " << szCommand);
        x_EditObject(*editor);
        ERR_POST(Info << "End " << szCommand);

    } catch (const std::exception& e) {
        x_ReportError(szCommand, e);
    }

}

void CBioseqEditor::OnDeleteBiosourceFeat(wxCommandEvent& event)
{
    const char* szCommand = "Delete Biosource Feature";

    try {
        CBioseq_Handle bh = m_CB->GetCurrentBioseq();
        if (!bh)
            return;

        CFeat_CI fsrc_it(bh, CSeqFeatData::e_Biosrc);
        if (!fsrc_it)
            return;

        CSeq_entry_Handle seh = bh.GetSeq_entry_Handle();
        CIRef<IEditCommand> command(new CCmdDelSeq_feat(fsrc_it->GetSeq_feat_Handle()));
        seh.GetEditHandle();
        m_CmdProccessor.Execute(command);
    } catch (const std::exception& e) {
        x_ReportError(szCommand, e);
    }
}



// feature creators
void CBioseqEditor::OnCreateFeature(CRef<CSeq_feat> feature)
{
    string feat_name = "Create " + feature->GetData().GetKey();    
    const char* szCommand = feat_name.c_str();

    try {
        CBioseq_Handle bh = m_CB->GetCurrentBioseq();
        if (!bh)
            return;


        CRef<CSeq_id> seq_id(new CSeq_id);
        CConstRef<CSeq_id> bseq_seq_id = bh.GetSeqId();
        TSeqPos length = bh.GetBioseqLength();
        seq_id->Assign(*bseq_seq_id);
        CRef<CSeq_loc> seqloc(new CSeq_loc(*seq_id, 0, length - 1));
        feature->SetLocation(*seqloc);

        CSeq_entry_Handle seh = bh.GetSeq_entry_Handle();
        CIRef<IEditObject> editor(new CEditObjectSeq_feat(*feature, seh, seh.GetScope(), true));
        ERR_POST(Info << "Start " << szCommand);
        x_EditObject(*editor);
        ERR_POST(Info << "End " << szCommand);

    } catch (const std::exception& e) {
        x_ReportError(szCommand, e);
    }
}


void CBioseqEditor::OnUpdateNucleotideFeat(wxUpdateUIEvent& event)
{
    CBioseq_Handle bh = m_CB->GetCurrentBioseq();
    if (!bh)
        return;
    event.Enable(!bh.IsAa());
}


void CBioseqEditor::OnUpdateProteinFeat(wxUpdateUIEvent& event)
{
    CBioseq_Handle bh = m_CB->GetCurrentBioseq();
    if (!bh)
        return;
    event.Enable(bh.IsAa());
}


void CBioseqEditor::OnCreateCodingRegion(wxCommandEvent& /*event*/)
{
    CRef<CSeq_feat> seqFeat(new CSeq_feat());
    seqFeat->SetData().SetCdregion();
    OnCreateFeature (seqFeat);
}


void CBioseqEditor::OnCreateGene(wxCommandEvent& /*event*/)
{
    CRef<CSeq_feat> seqFeat(new CSeq_feat());
    seqFeat->SetData().SetGene();
    OnCreateFeature (seqFeat);
}


void CBioseqEditor::OnCreateProt(wxCommandEvent& /*event*/)
{
    CRef<CSeq_feat> seqFeat(new CSeq_feat());
    seqFeat->SetData().SetProt();
    OnCreateFeature (seqFeat);
}


void CBioseqEditor::OnCreateImport(wxCommandEvent& evt)
{
    CRef<CSeq_feat> seqFeat(new CSeq_feat());
    seqFeat->SetData().SetImp().SetKey(GetKeyFromCmdID(evt.GetId()));
    OnCreateFeature (seqFeat);
}


void CBioseqEditor::OnCreateRegion(wxCommandEvent& /*event*/)
{
    CRef<CSeq_feat> seqFeat(new CSeq_feat());
    seqFeat->SetData().SetRegion();
    OnCreateFeature (seqFeat);
}


void CBioseqEditor::OnCreateSite(wxCommandEvent& /*event*/)
{
    CRef<CSeq_feat> seqFeat(new CSeq_feat());
    seqFeat->SetData().SetSite();
    OnCreateFeature (seqFeat);
}


void CBioseqEditor::OnCreateBond(wxCommandEvent& /*event*/)
{
    CRef<CSeq_feat> seqFeat(new CSeq_feat());
    seqFeat->SetData().SetBond();
    OnCreateFeature (seqFeat);
}


void CBioseqEditor::OnCreateRNA(wxCommandEvent& evt)
{
    CRef<CSeq_feat> seqFeat(new CSeq_feat());

    string rna_type = GetKeyFromCmdID(evt.GetId());
    if (NStr::EqualNocase(rna_type, "preRNA")) {
        seqFeat->SetData().SetRna().SetType(CRNA_ref::eType_premsg);
    } else if (NStr::EqualNocase(rna_type, "mRNA")) {
        seqFeat->SetData().SetRna().SetType(CRNA_ref::eType_mRNA);
    } else if (NStr::EqualNocase(rna_type, "tRNA")) {
        seqFeat->SetData().SetRna().SetType(CRNA_ref::eType_tRNA);
    } else if (NStr::EqualNocase(rna_type, "rRNA")) {
        seqFeat->SetData().SetRna().SetType(CRNA_ref::eType_rRNA);
    } else if (NStr::EqualNocase(rna_type, "ncRNA")) {
        seqFeat->SetData().SetRna().SetType(CRNA_ref::eType_ncRNA);
    } else if (NStr::EqualNocase(rna_type, "tmRNA")) {
        seqFeat->SetData().SetRna().SetType(CRNA_ref::eType_tmRNA);
    } else {
        seqFeat->SetData().SetRna().SetType(CRNA_ref::eType_miscRNA);
    }
        
    OnCreateFeature (seqFeat);
}


void CBioseqEditor::OnCreateDescriptor(wxCommandEvent& evt)
{
    string descriptor_type = GetKeyFromCmdID(evt.GetId());
    string cmd_name = "Create " + descriptor_type + " Descriptor";
    const char* szCommand = cmd_name.c_str();

    try {
        CBioseq_Handle bh = m_CB->GetCurrentBioseq();
        if (!bh)
            return;

        // pre-checks
        if (NStr::EqualNocase(descriptor_type, "BioSource")) {
            CSeqdesc_CI desc_it(bh, CSeqdesc::e_Source);
            if (desc_it) {
                ERR_POST(Info << "Error: " << szCommand << ": Biosource descriptor already exists");
                return;
            }
        } else if (NStr::EqualNocase(descriptor_type, "Molecule Description")) {
            CSeqdesc_CI desc_it(bh, CSeqdesc::e_Molinfo);
            if (desc_it) {
                CSeq_entry_Handle seh = bh.GetSeq_entry_Handle();
                CIRef<IEditObject> editor(new CEditObjectSeq_desc(*desc_it, seh, bh.GetScope(), false));
                ERR_POST(Info << "Start " << szCommand);
                x_EditObject(*editor);
                ERR_POST(Info << "End " << szCommand);
                return;
            }
        }

        CRef<CSeqdesc> seqdesc(new CSeqdesc());
        if (NStr::EqualNocase(descriptor_type, "BioSource")) {
            CRef<CBioSource> bioSource(new CBioSource());
            seqdesc->SetSource(*bioSource);
        } else if (NStr::EqualNocase(descriptor_type, "StructuredComment")) {
            seqdesc->SetUser().SetType().SetStr("StructuredComment");
        } else if (NStr::EqualNocase(descriptor_type, "DBLink")) {
            seqdesc->SetUser().SetType().SetStr("DBLink");
        } else if (NStr::EqualNocase(descriptor_type, "Region")) {
            seqdesc->SetRegion();
        } else if (NStr::EqualNocase(descriptor_type, "Name")) {
            seqdesc->SetName();
        } else if (NStr::EqualNocase(descriptor_type, "Comment")) {
            seqdesc->SetComment();
        } else if (NStr::EqualNocase(descriptor_type, "Title")) {
            seqdesc->SetTitle();
        } else if (NStr::EqualNocase(descriptor_type, "Molecule Description")) {
            seqdesc->SetMolinfo();
        } else if (NStr::EqualNocase(descriptor_type, "Publication")) {
            seqdesc->SetPub();
        }

        CSeq_entry_Handle seh = bh.GetSeq_entry_Handle();
        CIRef<IEditObject> editor(new CEditObjectSeq_desc(*seqdesc, seh, bh.GetScope(), true));
        ERR_POST(Info << "Start " << szCommand);
        x_EditObject(*editor);
        ERR_POST(Info << "End " << szCommand);

    } catch (const std::exception& e) {
        x_ReportError(szCommand, e);
    }

}


void CBioseqEditor::EditSelection()
{
    string command = "Edit Object: ";

    try {
        TEditObjectList objList;
        m_CB->GetCurrentSelection(objList);
        if (objList.size() != 1)
            return;

        CEditObject& editObj = objList.front();
        CSeq_entry_Handle seh = editObj.m_SEH;

        CIRef<IEditObject> editor;

        const CSeq_feat* seqFeat = dynamic_cast<const CSeq_feat*>(editObj.m_Obj.GetPointer());
        const CTitleEdit* titleEdit = dynamic_cast<const CTitleEdit*>(editObj.m_Obj.GetPointer());
        const CSeqdesc* seqDesc = dynamic_cast<const CSeqdesc*>(editObj.m_Obj.GetPointer());
        const CSubmit_block* submitBlock = dynamic_cast<const CSubmit_block*>(editObj.m_Obj.GetPointer());

        if (seqFeat) {
            editor.Reset(new CEditObjectSeq_feat(*seqFeat, seh, seh.GetScope(), false));
        }
        else if (seqDesc) {
            editor.Reset(new CEditObjectSeq_desc(*seqDesc, seh, seh.GetScope(), false));
        }
        else if (titleEdit) {
            CBioseq_Handle bh = seh.GetSeq();
            CConstRef<CObject> dummyDesc(new CSeqdescTitleEdit(bh, titleEdit->GetTitle()));
            editor.Reset(new CEditObjectSeq_desc(*dummyDesc, seh, seh.GetScope(), false));
        }
        else if (submitBlock) {
            // Create IEditObject object for Submit-block here
            CScope& scope = *editObj.m_Scope;
            editor.Reset(new CEditObjectSubmit_block(*submitBlock, scope, false));
        }

        const CSerialObject* serial =
            dynamic_cast<const CSerialObject*>(editObj.m_Obj.GetPointer());
        string type_name =
            serial ? serial->GetThisTypeInfo()->GetName() : string("CObject");

        if (!editor) {
            wxMessageBox(wxT("No editor found for object of type \'") +
                         ToWxString(type_name) +
                         wxT("\'."),
                         wxT("Error"), wxOK|wxICON_ERROR);
            return;
        }

        command += type_name;

        ERR_POST(Info << "Start " << command << type_name);
        x_EditObject(*editor);
        ERR_POST(Info << "End " << command<< type_name);
    } catch (const std::exception& e) {
        x_ReportError(command, e);
    }
}

void CBioseqEditor::OnEditSelection(wxCommandEvent& WXUNUSED(event))
{
    EditSelection();
}

void CBioseqEditor::OnUpdateEditSelection(wxUpdateUIEvent& event)
{
    event.Enable(false);

    try {
        TEditObjectList objList;
        m_CB->GetCurrentSelection(objList);
        if (objList.size() != 1)
            return;

        const CObject* obj = objList.front().m_Obj.GetPointer();
        const CSeqdesc* seqDesq = dynamic_cast<const CSeqdesc*>(obj);
        const CSeq_feat* seqFeat = dynamic_cast<const CSeq_feat*>(obj);
        const CSubmit_block* submitBlock = dynamic_cast<const CSubmit_block*>(obj);
        const CTitleEdit* titleEdit = dynamic_cast<const CTitleEdit*>(obj);

        event.Enable(seqDesq || seqFeat || titleEdit || submitBlock);
    }
    catch (const std::exception&) {
    }
}

void CBioseqEditor::OnDeleteSelection(wxCommandEvent& event)
{
    const char* szCommand = "Delete Object";

    try {
        TEditObjectList objList;
        m_CB->GetCurrentSelection(objList);
        if (objList.size() != 1)
            return;

        CEditObject& editObj = objList.front();

        CSeq_entry_Handle seh = editObj.m_SEH;
        const CSeqdesc* seqDesq = dynamic_cast<const CSeqdesc*>(editObj.m_Obj.GetPointer());
        if (seqDesq) {
            if (seqDesq->IsSource()) {
                wxMessageDialog dlg(NULL, wxT("Are you sure you want to delete the BioSource?"), wxT("Delete BioSource"), wxOK|wxCANCEL);
                if (dlg.ShowModal() == wxID_CANCEL) {
                    return;
                }
            }
            x_DeleteObject(seh, *seqDesq);
            return;
        }
        const CSeq_feat* seqFeat = dynamic_cast<const CSeq_feat*>(editObj.m_Obj.GetPointer());
        if (seqFeat) {
            x_DeleteObject(seh, *seqFeat);
        }
    } catch (const std::exception& e) {
        x_ReportError(szCommand, e);
    }
}

void CBioseqEditor::OnUpdateDeleteSelection(wxUpdateUIEvent& event)
{
    event.Enable(false);

    try {
        TEditObjectList objList;
        m_CB->GetCurrentSelection(objList);
        if (objList.size() != 1)
            return;

        const CObject* obj = objList.front().m_Obj.GetPointer();
        const CSeqdesc* seqDesq = dynamic_cast<const CSeqdesc*>(obj);
        const CSeq_feat* seqFeat = dynamic_cast<const CSeq_feat*>(obj);

        event.Enable(seqDesq || seqFeat);
    } catch (const std::exception&) {
    }
}

void CBioseqEditor::x_EditObject(IEditObject& edit)
{
    CEditObjectDlg edit_dlg(NULL);
    wxWindow* editorWindow = edit.CreateWindow(&edit_dlg);
    edit_dlg.SetEditorWindow(editorWindow);
    if (edit_dlg.ShowModal() != wxID_OK) {
        ERR_POST(Info << "Edit canceled");
        return;
    }
    m_CmdProccessor.Execute(edit.GetEditCommand());
}

void CBioseqEditor::x_DeleteObject(objects::CSeq_entry_Handle& seh, const objects::CSeqdesc& seqDesq)
{
    CSeqdesc::E_Choice choice = seqDesq.Which();

    CSeq_entry_Handle sehDesc;

    CSeqdesc_CI it(seh, choice);
    for (; it; ++it) {
        if (&seqDesq == &*it) {
            sehDesc = it.GetSeq_entry_Handle();
            break;
        }
    }

    if (!sehDesc)
        return;

    bool askSeqSet = false, deleteFromSeqSet = false;

    switch(choice) {
    case CSeqdesc::e_Name :
    case CSeqdesc::e_Title :
    case CSeqdesc::e_Comment :
    case CSeqdesc::e_Genbank :
    case CSeqdesc::e_Region :
    case CSeqdesc::e_User :
    case CSeqdesc::e_Embl :
    case CSeqdesc::e_Pub :
    case CSeqdesc::e_Source :
        askSeqSet = (seh.IsSeq() && sehDesc != seh);
        break;
    default:
        break;
    }

    CIRef<IEditCommand> cmd;

    if (askSeqSet) {
        CDeleteFromBioseq_setDlg dlg(NULL);
        if (dlg.ShowModal() != wxID_OK)
            return;

        if (dlg.GetChoice() == 0) {
            cmd.Reset(new CCmdDelDesc(sehDesc, seqDesq));
        }
        else {
            CRef<CCmdComposite> composite(new CCmdComposite("Delete Descriptor(s)"));
            CIRef<IEditCommand> cmdDelDesc(new CCmdDelDesc(sehDesc, seqDesq));
            composite->AddCommand(*cmdDelDesc);

            for (CSeq_entry_CI it(sehDesc.GetSet()); it; ++it) {
                if (*it != seh) {
                    CIRef<IEditCommand> cmdAddDesc(new CCmdCreateDesc(*it, seqDesq));
                    composite->AddCommand(*cmdAddDesc);
                }
            }
            cmd.Reset(composite.GetPointer());
        }
    } else {
        cmd.Reset(new CCmdDelDesc(sehDesc, seqDesq));
    }

    m_CmdProccessor.Execute(cmd);
}

void CBioseqEditor::x_DeleteObject(objects::CSeq_entry_Handle& seh, const objects::CSeq_feat& seqFeat)
{
    if (!seqFeat.IsSetData()) {
        LOG_POST(Error << "CBioseqEditor.Delete.Feature data not set");
        return;
    }

    CSeq_feat_Handle fh;
    SAnnotSelector sel(seqFeat.GetData().GetSubtype());
    for (CFeat_CI mf(seh.GetScope(), seqFeat.GetLocation(), sel); mf; ++mf) {
        if (mf->GetOriginalFeature().Equals(seqFeat)) {
            fh = mf->GetSeq_feat_Handle();
            break;
        }
    }

    if (!fh) {
        LOG_POST(Error << "CBioseqEditor.Delete.Feature Feature not found");
        return;
    }

    CIRef<IEditCommand> cmd, addCommand;

    const CSeqFeatData& data = seqFeat.GetData();
    if (data.IsGene()) {
        addCommand = CCmdFactory::DeleteGeneXRefs(seh, data.GetGene());
        if (addCommand) {
            int res = wxMessageBox(wxT("Delete features which reference this gene?"),
                                   wxT("Delete Gene"), wxYES_NO | wxICON_QUESTION );
            if (res != wxYES)
                addCommand.Reset();
        }
    }

    if (addCommand) {
        CRef<CCmdComposite> composite(new CCmdComposite("Delete Feature"));
        composite->AddCommand(*addCommand);
        composite->AddCommand(*GetDeleteFeatureCommand(fh));
        cmd.Reset(&*composite);
    }
    else {
        cmd.Reset(&*GetDeleteFeatureCommand(fh));
    }

    if (!cmd) {
        LOG_POST(Error << "CBioseqEditor.Delete.Feature failed to create command");
        return;
    }

    seh.GetEditHandle();
    m_CmdProccessor.Execute(cmd);
}

bool CBioseqEditor::x_HaveBiosourceDesc()
{
    try {
        CBioseq_Handle bh = m_CB->GetCurrentBioseq();
        if (!bh)
            return false;

        CSeqdesc_CI desc_it(bh, CSeqdesc::e_Source);
        if (desc_it) {
            return true;
        }
    }
    catch (const std::exception&) {
    }

    return false;
}

bool CBioseqEditor::x_HaveBiosourceFeat()
{
    try {
        CBioseq_Handle bh = m_CB->GetCurrentBioseq();
        if (!bh)
            return false;

        CFeat_CI fsrc_it(bh, CSeqFeatData::e_Biosrc);
        if (fsrc_it) {
            return true;
        }
    }
    catch (const std::exception&) {
    }

    return false;
}

void CBioseqEditor::x_ReportError(const string& cmd, const std::exception& e)
{
    const CException* pex = dynamic_cast<const CException*>(&e);
    if (pex) {
        ERR_POST(Error << cmd << " Failed: " << pex->GetMsg());
    }
    else {
        ERR_POST(Error << cmd << " Failed: " << e.what());
    }
}

END_NCBI_SCOPE
