/*  $Id: biosource_autocomplete.cpp 27370 2013-02-04 18:09:51Z bollin $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors: Colleen Bollin
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>
#include <util/line_reader.hpp>

#include <objects/seqfeat/OrgName.hpp>
#include <objects/general/Dbtag.hpp>
#include <objects/general/Object_id.hpp>

#include <gui/widgets/edit/biosource_autocomplete.hpp>
#include <gui/widgets/wx/wx_utils.hpp>
#include <gui/widgets/wx/sys_path.hpp>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

/*
 * CBioSourceAutoComplete
 */


CBioSourceAutoComplete::CBioSourceAutoComplete()
{
    x_Init();
}


CBioSourceAutoComplete::~CBioSourceAutoComplete()
{
}


vector<CRef<objects::COrg_ref> > CBioSourceAutoComplete::m_OrgList;

bool CBioSourceAutoComplete::x_Init()
{
    if (!m_OrgList.empty())
        return true;

    wxString path = CSysPath::ResolvePathExisting( wxT("<std>/etc/taxlist.txt") );
    if (path.IsEmpty())
        return false;

    try {
        CNcbiIfstream istr(path.fn_str());
        CStreamLineReader line_reader(istr);

        ++line_reader; // First line of the file seems to be garbage
        for (++line_reader; !line_reader.AtEOF();++line_reader) {
            string str = *line_reader;
            vector<string> row_values;
            NStr::Tokenize(str, "\t", row_values, NStr::eNoMergeDelims);
            if (!row_values.empty()) {
                CRef<COrg_ref> org(new COrg_ref());
                org->SetTaxname(row_values[0]);
                org->SetCommon(row_values[1]);
                org->SetOrgname().SetGcode(NStr::StringToInt(row_values[2]));
                org->SetOrgname().SetMgcode(NStr::StringToInt(row_values[3]));
                org->SetOrgname().SetDiv(row_values[4]);
                CRef<CDbtag> tag(new CDbtag());
                tag->SetDb("taxon");
                tag->SetTag().SetId(NStr::StringToInt(row_values[5]));
                org->SetDb().push_back(tag);
                m_OrgList.push_back(org);
            }
        }
    }
    catch (const CException& ex) {
        ERR_POST(ex.ReportAll());
        return false;
    }

    return true;
}


bool CBioSourceAutoComplete::AutoFill (objects::COrg_ref& org)
{
    if (!x_Init()  || !org.IsSetTaxname() || NStr::IsBlank (org.GetTaxname())) {
        return false;
    }
 
    bool rval = false;
    size_t i = 0;

    while (i < m_OrgList.size() && !rval) {
        if (NStr::EqualNocase(org.GetTaxname(), m_OrgList[i]->GetTaxname())) {
            org.SetCommon(m_OrgList[i]->GetCommon());
            org.SetOrgname().SetGcode(m_OrgList[i]->GetOrgname().GetGcode());
            org.SetOrgname().SetMgcode(m_OrgList[i]->GetOrgname().GetMgcode());
            org.SetOrgname().SetDiv(m_OrgList[i]->GetOrgname().GetDiv());
            bool has_taxon = false;
            if (org.IsSetDb()){
                NON_CONST_ITERATE (COrg_ref::TDb, it, org.SetDb()) {
                    if ((*it)->IsSetDb() && NStr::Equal((*it)->GetDb(), "taxon")) {
                        (*it)->SetTag().Assign(m_OrgList[i]->GetDb().front()->GetTag());
                        has_taxon = true;
                        break;
                    }
                }
            }
            if (!has_taxon) {
                CRef<CDbtag> tag(new CDbtag());
                tag->Assign(*(m_OrgList[i]->GetDb().front()));
                org.SetDb().push_back(tag);
            }
            rval = true;
        }
        i++;
    }

    return rval;
}


END_NCBI_SCOPE
