/*  $Id: cds_translation_panel.cpp 28164 2013-05-20 19:11:55Z katargir $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data,  the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties,  express or implied,  including
 *  warranties of performance,  merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Colleen Bollin
 */


#include <ncbi_pch.hpp>
#include <gui/widgets/wx/wx_utils.hpp>
#include <gui/objutils/cmd_add_seqentry.hpp>
#include <objects/seqloc/Seq_id.hpp>
#include <objects/seq/Seq_inst.hpp>
#include <objects/seq/Seq_data.hpp>
#include <objects/seqloc/Seq_loc.hpp>
#include <objects/seqfeat/Genetic_code.hpp>
#include <objects/seqfeat/Genetic_code_table.hpp>
#include <objmgr/bioseq_ci.hpp>
#include <objmgr/scope.hpp>
#include <objmgr/seq_vector.hpp>
#include <objmgr/seq_vector_ci.hpp>
#include <objmgr/util/sequence.hpp>
#include "generic_props_panel.hpp"

////@begin includes
////@end includes

#include <wx/treebook.h>
#include <wx/panel.h>
#include <wx/sizer.h>
#include <wx/choice.h>
#include <wx/textctrl.h>
#include <wx/stattext.h>
#include <wx/hyperlink.h>
#include <wx/checkbox.h>
#include <wx/msgdlg.h>


#include "cds_translation_panel.hpp"
#include "location_panel.hpp"
#include "cds_exception_panel.hpp"

////@begin XPM images
////@end XPM images

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

/*!
 * CCDSTranslationPanel type definition
 */

IMPLEMENT_DYNAMIC_CLASS( CCDSTranslationPanel, wxPanel )


/*!
 * CCDSTranslationPanel event table definition
 */

BEGIN_EVENT_TABLE( CCDSTranslationPanel, wxPanel )

////@begin CCDSTranslationPanel event table entries
    EVT_HYPERLINK( ID_HYPERLINKCTRL1, CCDSTranslationPanel::OnHyperlinkctrl1HyperlinkClicked )

    EVT_HYPERLINK( ID_HYPERLINKCTRL2, CCDSTranslationPanel::AdjustForStopCodonHyperlinkClicked )

////@end CCDSTranslationPanel event table entries

END_EVENT_TABLE()


/*!
 * CCDSTranslationPanel constructors
 */

CCDSTranslationPanel::CCDSTranslationPanel()
{
    Init();
}

CCDSTranslationPanel::CCDSTranslationPanel( wxWindow* parent, CSeq_feat& feat, CScope& scope, wxWindowID id, 
                                           const wxString& caption, const wxPoint& pos, const wxSize& size, long style ) :
  m_EditedFeat(new CSeq_feat()), m_Scope(&scope), m_EditedBioseq(0)
{
    Init();
    m_EditedFeat->Assign (feat);
    if (feat.IsSetProduct()) {
        const CSeq_id *id = feat.GetProduct().GetId();
        if (id) {
            CBioseq_Handle bsh = m_Scope->GetBioseqHandle(*id);
            if (bsh) {
                m_EditedBioseq = new CBioseq();
                m_EditedBioseq->Assign(*(bsh.GetCompleteBioseq()));
            }
        }
    }
    Create(parent, id, caption, pos, size, style);
}


/*!
 * CCDSTranslationPanel creator
 */

bool CCDSTranslationPanel::Create( wxWindow* parent, wxWindowID id, const wxString& caption, const wxPoint& pos, const wxSize& size, long style )
{
////@begin CCDSTranslationPanel creation
    SetExtraStyle(wxWS_EX_BLOCK_EVENTS);
    wxPanel::Create( parent, id, pos, size, style );

    CreateControls();
    if (GetSizer())
    {
        GetSizer()->SetSizeHints(this);
    }
    Centre();
////@end CCDSTranslationPanel creation
    return true;
}


/*!
 * CCDSTranslationPanel destructor
 */

CCDSTranslationPanel::~CCDSTranslationPanel()
{
////@begin CCDSTranslationPanel destruction
////@end CCDSTranslationPanel destruction
}


/*!
 * Member initialisation
 */

void CCDSTranslationPanel::Init()
{
////@begin CCDSTranslationPanel member initialisation
    m_GeneticCode = NULL;
    m_Frame = NULL;
    m_ConflictCtrl = NULL;
    m_UpdatemRNASpan = NULL;
    m_ProductCtrl = NULL;
    m_TranslationCtrl = NULL;
    m_RetranslateLink = NULL;
////@end CCDSTranslationPanel member initialisation
}


/*!
 * Control creation for CCDSTranslationPanel
 */

void CCDSTranslationPanel::CreateControls()
{    
////@begin CCDSTranslationPanel content construction
    CCDSTranslationPanel* itemPanel1 = this;

    wxBoxSizer* itemBoxSizer2 = new wxBoxSizer(wxVERTICAL);
    itemPanel1->SetSizer(itemBoxSizer2);

    wxFlexGridSizer* itemFlexGridSizer3 = new wxFlexGridSizer(0, 2, 0, 0);
    itemBoxSizer2->Add(itemFlexGridSizer3, 0, wxALIGN_CENTER_HORIZONTAL|wxALL, 5);

    wxStaticText* itemStaticText4 = new wxStaticText( itemPanel1, wxID_STATIC, _("Genetic Code"), wxDefaultPosition, wxDefaultSize, 0 );
    itemFlexGridSizer3->Add(itemStaticText4, 0, wxALIGN_RIGHT|wxALIGN_CENTER_VERTICAL|wxALL, 5);

    wxArrayString m_GeneticCodeStrings;
    m_GeneticCode = new wxChoice( itemPanel1, ID_CHOICE1, wxDefaultPosition, wxDefaultSize, m_GeneticCodeStrings, 0 );
    if (CCDSTranslationPanel::ShowToolTips())
        m_GeneticCode->SetToolTip(_("Choose the genetic code that should be used to translate the nucleotide sequence.  For more information, and for the translation tables themselves, see the NCBI Taxonomy page.  If the genetic code is already populated from the taxonomy database, do not change this selection."));
    itemFlexGridSizer3->Add(m_GeneticCode, 0, wxALIGN_CENTER_HORIZONTAL|wxALIGN_CENTER_VERTICAL|wxALL, 5);

    wxStaticText* itemStaticText6 = new wxStaticText( itemPanel1, wxID_STATIC, _("Reading Frame"), wxDefaultPosition, wxDefaultSize, 0 );
    itemFlexGridSizer3->Add(itemStaticText6, 0, wxALIGN_RIGHT|wxALIGN_CENTER_VERTICAL|wxALL, 5);

    wxArrayString m_FrameStrings;
    m_FrameStrings.Add(_("One"));
    m_FrameStrings.Add(_("Two"));
    m_FrameStrings.Add(_("Three"));
    m_Frame = new wxChoice( itemPanel1, ID_CHOICE2, wxDefaultPosition, wxDefaultSize, m_FrameStrings, 0 );
    if (CCDSTranslationPanel::ShowToolTips())
        m_Frame->SetToolTip(_("Choose the reading frame in which to translate the sequence."));
    itemFlexGridSizer3->Add(m_Frame, 0, wxALIGN_LEFT|wxALIGN_CENTER_VERTICAL|wxALL, 5);

    wxStaticText* itemStaticText8 = new wxStaticText( itemPanel1, wxID_STATIC, _("Conflict"), wxDefaultPosition, wxDefaultSize, 0 );
    itemFlexGridSizer3->Add(itemStaticText8, 0, wxALIGN_RIGHT|wxALIGN_CENTER_VERTICAL|wxALL, 5);

    m_ConflictCtrl = new wxCheckBox( itemPanel1, ID_CHECKBOX, wxEmptyString, wxDefaultPosition, wxDefaultSize, 0 );
    m_ConflictCtrl->SetValue(false);
    itemFlexGridSizer3->Add(m_ConflictCtrl, 0, wxALIGN_LEFT|wxALIGN_CENTER_VERTICAL|wxALL, 5);

    wxStaticLine* itemStaticLine10 = new wxStaticLine( itemPanel1, wxID_STATIC, wxDefaultPosition, wxDefaultSize, wxLI_HORIZONTAL );
    itemBoxSizer2->Add(itemStaticLine10, 0, wxGROW|wxALL, 5);

    m_UpdatemRNASpan = new wxCheckBox( itemPanel1, ID_CHECKBOX3, _("Update mRNA span"), wxDefaultPosition, wxDefaultSize, 0 );
    m_UpdatemRNASpan->SetValue(false);
    itemBoxSizer2->Add(m_UpdatemRNASpan, 0, wxALIGN_CENTER_HORIZONTAL|wxALL, 5);

    wxBoxSizer* itemBoxSizer12 = new wxBoxSizer(wxHORIZONTAL);
    itemBoxSizer2->Add(itemBoxSizer12, 0, wxALIGN_CENTER_HORIZONTAL|wxALL, 5);

    wxStaticText* itemStaticText13 = new wxStaticText( itemPanel1, wxID_STATIC, _("Protein Product"), wxDefaultPosition, wxDefaultSize, 0 );
    itemBoxSizer12->Add(itemStaticText13, 0, wxALIGN_CENTER_VERTICAL|wxALL, 5);

    wxArrayString m_ProductCtrlStrings;
    m_ProductCtrl = new wxComboBox( itemPanel1, ID_COMBOBOX, wxEmptyString, wxDefaultPosition, wxDefaultSize, m_ProductCtrlStrings, wxCB_DROPDOWN );
    itemBoxSizer12->Add(m_ProductCtrl, 0, wxALIGN_CENTER_VERTICAL|wxALL, 5);

    m_TranslationCtrl = new wxTextCtrl( itemPanel1, ID_TEXTCTRL27, wxEmptyString, wxDefaultPosition, wxSize(400, 100), wxTE_MULTILINE|wxTE_READONLY );
    m_TranslationCtrl->SetBackgroundColour(wxColour(255, 255, 255));
    m_TranslationCtrl->Enable(false);
    itemBoxSizer2->Add(m_TranslationCtrl, 0, wxGROW|wxALL, 5);

    wxBoxSizer* itemBoxSizer16 = new wxBoxSizer(wxHORIZONTAL);
    itemBoxSizer2->Add(itemBoxSizer16, 0, wxALIGN_CENTER_HORIZONTAL|wxALL, 5);

    m_RetranslateLink = new wxHyperlinkCtrl( itemPanel1, ID_HYPERLINKCTRL1, _("Retranslate"), wxEmptyString, wxDefaultPosition, wxDefaultSize, wxHL_DEFAULT_STYLE );
    m_RetranslateLink->SetNormalColour(wxColour(0, 0, 255));
    m_RetranslateLink->SetVisitedColour(wxColour(0, 0, 255));
    itemBoxSizer16->Add(m_RetranslateLink, 0, wxALIGN_CENTER_VERTICAL|wxALL, 5);

    wxHyperlinkCtrl* itemHyperlinkCtrl18 = new wxHyperlinkCtrl( itemPanel1, ID_HYPERLINKCTRL2, _("Adjust for Stop Codon"), wxEmptyString, wxDefaultPosition, wxDefaultSize, wxHL_DEFAULT_STYLE );
    itemHyperlinkCtrl18->SetNormalColour(wxColour(0, 0, 255));
    itemHyperlinkCtrl18->SetVisitedColour(wxColour(0, 0, 255));
    itemBoxSizer16->Add(itemHyperlinkCtrl18, 0, wxALIGN_CENTER_VERTICAL|wxALL, 5);

////@end CCDSTranslationPanel content construction

    CScope::TTSE_Handles tses;
    m_Scope->GetAllTSEs(tses, CScope::eAllTSEs);
    ITERATE (CScope::TTSE_Handles, handle, tses) {
        for (CBioseq_CI bioseq_it(*handle, CSeq_inst::eMol_aa);  bioseq_it;  ++bioseq_it) {
            string label;
            bioseq_it->GetId().front().GetSeqId()->GetLabel(&label);
            m_ProductCtrlStrings.Add(ToWxString(label));          
        }
    }
    m_ProductCtrl->Append (m_ProductCtrlStrings);

    const CGenetic_code_table& code_table = CGen_code_table::GetCodeTable();
    const CGenetic_code_table::Tdata& codes = code_table.Get();

    ITERATE (CGenetic_code_table::Tdata, it, codes) {
        string str = (*it)->GetName();
        int id = (*it)->GetId();

        string::size_type pos = str.find(';');
        string tmp = (pos != string::npos) ? str.substr(0, pos) + ", etc." : str;
        m_GeneticCode->Append(ToWxString(tmp), (void*)id);
    }

}

bool CCDSTranslationPanel::TransferDataToWindow()
{
    if (!wxPanel::TransferDataToWindow()) {
        return false;
    }

    string label = "";
    if (m_EditedFeat->IsSetProduct()) {
        const CSeq_id *id = m_EditedFeat->GetProduct().GetId();
        if (id) {
            CBioseq_Handle bsh = m_Scope->GetBioseqHandle(*id);
            if (bsh) {
                bsh.GetSeqId()->GetLabel(&label);
            } else {
                m_EditedFeat->GetProduct().GetId()->GetLabel(&label);
            }
        }
    }
    m_ProductCtrl->SetValue (ToWxString(label));

    if (m_EditedBioseq) {
        CSeqVector prot_vec(*m_EditedBioseq, m_Scope);
        prot_vec.SetCoding(CSeq_data::e_Iupacaa);

        string prot_seq;
        prot_vec.GetSeqData(0, prot_vec.size(), prot_seq);
        m_TranslationCtrl->SetValue(ToWxString(prot_seq));
        m_TranslationCtrl->Enable(true);
    } else {
        m_TranslationCtrl->SetValue(wxEmptyString);
    }

    CCdregion& cds = m_EditedFeat->SetData().SetCdregion();

    string codeName = "";
    int codeId = 1;

    if (cds.IsSetCode()) {
        const CCdregion::TCode& code = cds.GetCode();
        codeId = code.GetId();
        codeName = code.GetName();
    }

    if (NStr::IsBlank(codeName)) {
        const CGenetic_code_table& code_table = CGen_code_table::GetCodeTable();
        const CGenetic_code_table::Tdata& codes = code_table.Get();

        ITERATE (CGenetic_code_table::Tdata, it, codes) {
            string str = (*it)->GetName();
            int id = (*it)->GetId();
            if (id == codeId) {
                codeName = str;
                break;
            }
        }
    }            

    m_GeneticCode->SetStringSelection(ToWxString (codeName));

    if (cds.IsSetConflict() && cds.GetConflict()) {
        m_ConflictCtrl->SetValue (true);
    }

    if (cds.IsSetFrame()) {
        switch (cds.GetFrame()) {
            case CCdregion::eFrame_two:
                m_Frame->SetSelection(1);
                break;
            case CCdregion::eFrame_three:
                m_Frame->SetSelection(2);
                break;
            default:
                m_Frame->SetSelection(0);
                break;
        }
    } else {
        m_Frame->SetSelection(0);
    }

    return true;
}


wxTreebook* CCDSTranslationPanel::x_GetTree()
{
    wxWindow* w = this->GetParent();
    while (w != NULL) {
        wxTreebook* parent = dynamic_cast<wxTreebook*>(w);
        if (parent) {
            return parent;
        } else {
            w = w->GetParent();
        }
    }
    return NULL;    
}


bool CCDSTranslationPanel::x_IsPseudo()
{
    wxTreebook* parent = x_GetTree();
    if (!parent) {
        return false;
    }
    CGenericPropsPanel* gen_panel = 0;
    for (size_t i = 0; i < parent->GetPageCount(); ++i) {
        gen_panel = dynamic_cast<CGenericPropsPanel*>(parent->GetPage(i));
        if (gen_panel)
            break;
    }
    if (!gen_panel) {
        return false;
    }
    if (gen_panel->m_Pseudo->GetValue()) {
        return true;
    }
}


bool CCDSTranslationPanel::TransferDataFromWindow()
{
    if (!wxPanel::TransferDataFromWindow()) {
        return false;
    }

    // change product ID
    string id_string = ToStdString (m_ProductCtrl->GetValue());
    if (NStr::IsBlank(id_string)) {
        CBioseq_Handle bsh = m_Scope->GetBioseqHandle(m_EditedFeat->GetLocation());
        CSeq_entry_Handle seh = bsh.GetTopLevelEntry();
        CRef<CSeq_id> new_id = GetNewProteinId(seh, bsh);
        m_EditedFeat->SetProduct().SetWhole().Assign(*new_id);
        new_id->GetLabel(&id_string, objects::CSeq_id::eContent);
        m_ProductCtrl->SetValue(ToWxString(id_string));
    } else {
        m_EditedFeat->SetProduct().SetWhole().Set(id_string);
    }

    int index = m_GeneticCode->GetSelection();
    int id = 1;
    if (index != wxNOT_FOUND)
        id = (long)m_GeneticCode->GetClientData(index);

    CRef< CGenetic_code::C_E > ce(new CGenetic_code::C_E);
    ce->SetId(id);
    CRef<CGenetic_code> code(new CGenetic_code());
    code->Set().push_back(ce);

    CCdregion& cds = m_EditedFeat->SetData().SetCdregion();
    cds.SetCode(*code);

    if (m_ConflictCtrl->GetValue ()) {
        cds.SetConflict(true);
    } else {
        cds.ResetConflict();
    }

    switch (m_Frame->GetSelection ()) {
        case 0:
            cds.SetFrame(CCdregion::eFrame_one);
            break;
        case 1:
            cds.SetFrame(CCdregion::eFrame_two);
            break;
        case 2:
            cds.SetFrame(CCdregion::eFrame_three);
            break;
        default:
            cds.ResetFrame();
            break;
    }

    if (!x_IsPseudo() && NStr::IsBlank(ToStdString(m_TranslationCtrl->GetValue()))) {
        wxMessageBox(wxT("A coding region must have a protein translation unless it is pseudo"), wxT("Error"),
                         wxOK | wxICON_ERROR, this);
        return false;
    }
    return true;
}


void CCDSTranslationPanel::SetProduct(CSeq_feat& feat)
{
    if (m_EditedFeat->IsSetProduct()) {
        CRef<CSeq_loc> product(new CSeq_loc());
        product->Assign(m_EditedFeat->GetProduct());
        feat.SetProduct(*product);
    } else {
        feat.ResetProduct();
    }

    const CCdregion& cds = m_EditedFeat->GetData().GetCdregion();
    CCdregion& edit_cds = feat.SetData().SetCdregion();
    if (cds.IsSetCode()) {
        CRef<CGenetic_code> new_code(new CGenetic_code());
        new_code->Assign(cds.GetCode());
        edit_cds.SetCode(*new_code);
    } else {
        edit_cds.ResetCode();
    }

    if (cds.IsSetFrame()) {
        edit_cds.SetFrame(cds.GetFrame());
    } else {
        edit_cds.ResetFrame();
    }
    
    if (cds.IsSetConflict() && cds.GetConflict()) {
        edit_cds.SetConflict(true);
    } else {
        edit_cds.ResetConflict();
    }
        
        
}


void CCDSTranslationPanel::SetProtein(CBioseq& bioseq)
{
    if (m_EditedBioseq) {
        // only copy data and length and ID
        CRef<CSeq_inst> inst(new CSeq_inst());
        inst->Assign(m_EditedBioseq->GetInst());
        bioseq.SetInst(*inst);
        if (m_EditedBioseq->IsSetId()) {
            CRef<CSeq_id> new_id(new CSeq_id());
            new_id->Assign(*(m_EditedBioseq->GetId().front()));
            bioseq.SetId().push_back(new_id);
        }
    } else {
        bioseq.ResetInst();
    }
}


/*!
 * Should we show tooltips?
 */

bool CCDSTranslationPanel::ShowToolTips()
{
    return true;
}

/*!
 * Get bitmap resources
 */

wxBitmap CCDSTranslationPanel::GetBitmapResource( const wxString& name )
{
    // Bitmap retrieval
////@begin CCDSTranslationPanel bitmap retrieval
    wxUnusedVar(name);
    return wxNullBitmap;
////@end CCDSTranslationPanel bitmap retrieval
}

/*!
 * Get icon resources
 */

wxIcon CCDSTranslationPanel::GetIconResource( const wxString& name )
{
    // Icon retrieval
////@begin CCDSTranslationPanel icon retrieval
    wxUnusedVar(name);
    return wxNullIcon;
////@end CCDSTranslationPanel icon retrieval
}


/*!
 * wxEVT_COMMAND_HYPERLINK event handler for ID_HYPERLINKCTRL1
 */

void CCDSTranslationPanel::OnHyperlinkctrl1HyperlinkClicked( wxHyperlinkEvent& event )
{
    TransferDataFromWindow();

    // get new location and new frame/code, if it has been edited
    wxWindow* parent = this->GetParent();
    wxTreebook* tbook = dynamic_cast<wxTreebook*>(parent);
    if (tbook) {
        CLocationPanel* locPanel = 0;
        CCDSExceptionPanel* exceptionPanel = 0;

        for (size_t i = 0; i < tbook->GetPageCount(); ++i) {
            if (!locPanel) {
                locPanel = dynamic_cast<CLocationPanel*>(tbook->GetPage(i));
            }
            if (!exceptionPanel) {
                exceptionPanel = dynamic_cast<CCDSExceptionPanel*>(tbook->GetPage(i));
            }

            if (locPanel && exceptionPanel) {
                break;
            }
        }
        if (locPanel) {
            locPanel->TransferDataFromWindow();
            CRef<objects::CSeq_loc> loc = locPanel->GetSeq_loc();
            m_EditedFeat->SetLocation(*loc);
        }

        if (exceptionPanel) {
            exceptionPanel->TransferDataFromWindow();
            exceptionPanel->SetExceptions(m_EditedFeat->SetData().SetCdregion());
        }

    }        

    CRef<CBioseq> bioseq = CSeqTranslator::TranslateToProtein (*m_EditedFeat, *m_Scope);
    if (bioseq) {
        objects::CSeqTranslator::ChangeDeltaProteinToRawProtein(bioseq);
        if (!m_EditedBioseq) {
            m_EditedBioseq = new CBioseq();
        }
        m_EditedBioseq->Assign(*bioseq);
    } else {
        m_EditedBioseq->Reset();
    }
    TransferDataToWindow();
}


static CRef<CSeq_loc> TruncateSeqLoc (CRef<CSeq_loc> orig_loc, size_t new_len)
{
    CRef<CSeq_loc> new_loc;
    if (!orig_loc) {
        return new_loc;
    }
    size_t len = 0;
    for (CSeq_loc_CI it(*orig_loc); it && len < new_len; ++it) {
        size_t this_len = it.GetRange().GetLength();
        CConstRef<CSeq_loc> this_loc = it.GetRangeAsSeq_loc();
        if (len + this_len <= new_len) {
            if (new_loc) {
                new_loc->Add(*this_loc);
            } else {
                new_loc.Reset(new CSeq_loc());
                new_loc->Assign(*this_loc);
            }
            len += this_len;
        } else {
            CRef<CSeq_loc> partial_loc(new CSeq_loc());
            size_t len_wanted = new_len - len;
            size_t start = this_loc->GetStart(eExtreme_Biological);
            if (len_wanted == 1) {
                // make a point
                partial_loc->SetPnt().SetPoint(start);
            } else {
                // make an interval
                if (this_loc->IsSetStrand() && this_loc->GetStrand() == eNa_strand_minus) {
                    partial_loc->SetInt().SetFrom(start - len_wanted + 1);
                    partial_loc->SetInt().SetTo(start);
                } else {
                    partial_loc->SetInt().SetFrom(start);
                    partial_loc->SetInt().SetTo(start + len_wanted - 1);
                }
            }
            partial_loc->SetId(*this_loc->GetId());
            if (this_loc->IsSetStrand()) {
                partial_loc->SetStrand(this_loc->GetStrand());
            }
            if (new_loc) {
                new_loc->Add(*partial_loc);
            } else {
                new_loc.Reset(new CSeq_loc());
                new_loc->Assign(*partial_loc);
            }
            len += len_wanted;  
        }
    }

    return new_loc;
}


static CRef<CSeq_loc> ExtendToStopCodon (CRef<CSeq_feat> feat, CScope* scope)
{
    CRef<CSeq_loc> new_loc;

    if (!feat || !feat->IsSetLocation()) {
        return new_loc;
    }
    const CSeq_loc& loc = feat->GetLocation();

    CBioseq_Handle bsh = scope->GetBioseqHandle(*(loc.GetId()));
    if (!bsh) {
        return new_loc;
    }

    const CGenetic_code* code = NULL;
    if (feat->IsSetData() && feat->GetData().IsCdregion() && feat->GetData().GetCdregion().IsSetCode()) {
        code = &(feat->GetData().GetCdregion().GetCode());
    }

    size_t stop = loc.GetStop(eExtreme_Biological);
    // figure out if we have a partial codon at the end
    size_t orig_len = sequence::GetLength(loc, scope);
    size_t len = orig_len;
    if (feat->IsSetData() && feat->GetData().IsCdregion() && feat->GetData().GetCdregion().IsSetFrame()) {
        CCdregion::EFrame frame = feat->GetData().GetCdregion().GetFrame();
        if (frame == CCdregion::eFrame_two) {
            len -= 1;
        } else if (frame == CCdregion::eFrame_three) {
            len -= 2;
        }
    }
    size_t mod = len % 3;
    CRef<CSeq_loc> vector_loc(new CSeq_loc());
    vector_loc->SetInt().SetId().Assign(*(loc.GetId()));

    if (loc.IsSetStrand() && loc.GetStrand() == eNa_strand_minus) {
        vector_loc->SetInt().SetFrom(0);
        vector_loc->SetInt().SetTo(stop + mod - 1);
        vector_loc->SetStrand(eNa_strand_minus);
    } else {
        vector_loc->SetInt().SetFrom(stop - mod + 1);
        vector_loc->SetInt().SetTo(bsh.GetInst_Length() - 1);
    }

    CSeqVector seq(*vector_loc, *scope, CBioseq_Handle::eCoding_Iupac);
    // reserve our space
    const size_t usable_size = seq.size();

    // get appropriate translation table
    const CTrans_table & tbl =
        (code ? CGen_code_table::GetTransTable(*code) :
                CGen_code_table::GetTransTable(1));

    // main loop through bases
    CSeqVector::const_iterator start = seq.begin();

    size_t i;
    size_t k;
    size_t state = 0;
    size_t length = usable_size / 3;

    for (i = 0;  i < length;  ++i) {
        // loop through one codon at a time
        for (k = 0;  k < 3;  ++k, ++start) {
            state = tbl.NextCodonState(state, *start);
        }

        if (tbl.GetCodonResidue (state) == '*') {
            CSeq_loc_CI it(loc);
            CSeq_loc_CI it_next = it;
            ++it_next;
            while (it_next) {
                CConstRef<CSeq_loc> this_loc = it.GetRangeAsSeq_loc();
                if (new_loc) {
                    new_loc->Add(*this_loc);
                } else {
                    new_loc.Reset(new CSeq_loc());
                    new_loc->Assign(*this_loc);
                }
                it = it_next;
                ++it_next;
            }
            CRef<CSeq_loc> last_interval(new CSeq_loc());
            CConstRef<CSeq_loc> this_loc = it.GetRangeAsSeq_loc();
            size_t this_start = this_loc->GetStart(eExtreme_Positional);
            size_t this_stop = this_loc->GetStop(eExtreme_Positional);
            size_t extension = ((i + 1) * 3) - mod;
            last_interval->SetInt().SetId().Assign(*(this_loc->GetId()));
            if (this_loc->IsSetStrand() && this_loc->GetStrand() == eNa_strand_minus) {
                last_interval->SetStrand(eNa_strand_minus);
                last_interval->SetInt().SetFrom(this_start - extension);
                last_interval->SetInt().SetTo(this_stop);
            } else {
                last_interval->SetInt().SetFrom(this_start);
                last_interval->SetInt().SetTo(this_stop + extension);
            }
                
            if (new_loc) {
                new_loc->Add(*last_interval);
            } else {
                new_loc.Reset(new CSeq_loc());
                new_loc->Assign(*last_interval);
            }

            return new_loc;
        }
    }

    return new_loc;
}


/*!
 * wxEVT_COMMAND_HYPERLINK event handler for ID_HYPERLINKCTRL2
 */

void CCDSTranslationPanel::AdjustForStopCodonHyperlinkClicked( wxHyperlinkEvent& event )
{
    TransferDataFromWindow();

    // get new location and new frame/code, if it has been edited
    wxWindow* parent = this->GetParent();
    wxTreebook* tbook = dynamic_cast<wxTreebook*>(parent);
    if (!tbook) {
        return;
    }

    CLocationPanel* locPanel = 0;
    CCDSExceptionPanel* exceptionPanel = 0;

    for (size_t i = 0; i < tbook->GetPageCount(); ++i) {
        if (!locPanel) {
            locPanel = dynamic_cast<CLocationPanel*>(tbook->GetPage(i));
        }
        if (!exceptionPanel) {
            exceptionPanel = dynamic_cast<CCDSExceptionPanel*>(tbook->GetPage(i));
        }

        if (locPanel && exceptionPanel) {
            break;
        }
    }
    if (!locPanel) {
        return;
    }

    locPanel->TransferDataFromWindow();
    CRef<objects::CSeq_loc> loc = locPanel->GetSeq_loc();
    m_EditedFeat->SetLocation(*loc);

    if (exceptionPanel) {
        exceptionPanel->TransferDataFromWindow();
        exceptionPanel->SetExceptions(m_EditedFeat->SetData().SetCdregion());
    }

    CRef<CBioseq> bioseq = CSeqTranslator::TranslateToProtein (*m_EditedFeat, *m_Scope);
    if (bioseq) {
        // adjust location until protein ends at stop codon
        CSeqVector seq(*bioseq, m_Scope, CBioseq_Handle::eCoding_Ncbi);
        seq.SetCoding(CSeq_data::e_Ncbieaa);
        CSeqVector_CI it(seq);
        while (it && *it != '*') {
            it++;
        }
        size_t len_wanted = 0;
        CRef<CSeq_loc> new_loc;
        if (it) {
            // want to truncate the location and retranslate
            len_wanted =  3 * (it.GetPos() + 1);
            if (m_EditedFeat->GetData().GetCdregion().IsSetFrame()) {
                CCdregion::EFrame frame = m_EditedFeat->GetData().GetCdregion().GetFrame();
                if (frame == CCdregion::eFrame_two) {
                    len_wanted += 1;
                } else if (frame == CCdregion::eFrame_three) {
                    len_wanted += 2;
                }
            }
            if (len_wanted > 0) {
                new_loc = TruncateSeqLoc (loc, len_wanted);
            }
        } else {
            // want to extend the location and retranslate
            new_loc = ExtendToStopCodon(m_EditedFeat, m_Scope);
        }

        if (new_loc) {
            m_EditedFeat->SetLocation(*new_loc);
            locPanel->SetSeq_loc(m_EditedFeat->SetLocation());
            bioseq = CSeqTranslator::TranslateToProtein (*m_EditedFeat, *m_Scope);
            if (bioseq) {
                m_EditedBioseq->Assign(*bioseq);
            } else {
                m_EditedBioseq->Reset();
            }
            TransferDataToWindow();
        }
    }
}


END_NCBI_SCOPE
