/*  $Id: edit_object_feature_propagate.cpp 28123 2013-05-13 15:43:15Z ludwigf $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors: Colleen Bollin, based on a file by Roman Katargin
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>

#include "edit_object_feature_propagate.hpp"
#include <gui/objutils/undo_manager_interface.hpp>
#include <gui/objutils/cmd_propagate_feat.hpp>
#include <gui/objutils/cmd_composite.hpp>
#include <gui/objutils/descriptor_change.hpp>
#include <gui/objutils/cmd_create_desc.hpp>
#include <gui/objutils/cmd_add_seqentry.hpp>

#include <gui/objutils/seqdesc_title_edit.hpp>
#include <gui/objutils/seqdesc_title_change.hpp>

#include <gui/widgets/wx/wx_utils.hpp>

#include <objects/seq/Seqdesc.hpp>
#include <objects/seq/Pubdesc.hpp>
#include <objects/general/Object_id.hpp>
#include <objects/general/User_field.hpp>
#include <objects/misc/sequence_macros.hpp>

#include <objmgr/util/sequence.hpp>
#include <objmgr/seqdesc_ci.hpp>
#include <objmgr/bioseq_ci.hpp>

#include <wx/panel.h>
#include <wx/treebook.h>

#include "descriptortext_panel.hpp"
#include "publicationtype_panel.hpp"
#include "molinfo_panel.hpp"
#include "structuredcomment_panel.hpp"
#include <gui/widgets/edit/sequencingtechnologypanel.hpp>
#include <gui/widgets/edit/dblinkpanel.hpp>
#include <gui/widgets/edit/biosource_autocomplete.hpp>

#include "loc_gcode_panel.hpp"
#include "biosourcepanel.hpp"
#include "srcmod_list_panel.hpp"
#include "dbxref_panel.hpp"

#ifdef _DEBUG
#include <serial/objostr.hpp>
#include <serial/serial.hpp>
#endif

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

wxWindow* CEditObjectFeaturePropagate::CreateWindow(wxWindow* parent)
{
    if (m_Window) {
        NCBI_THROW(CException, eUnknown, 
           "CEditObjectFeaturePropagate::CreateWindow - Window already created!");
    }

    //Populate GUI:
    m_Window = new wxPanel(parent);
    wxSizer* pTopSizer = new wxBoxSizer(wxVERTICAL);
    m_Window->SetSizer(pTopSizer);

    pTopSizer->Add(xTargetSizer(), 0, wxEXPAND);
    //pTopSizer->Add(xFeatureSelectSizer(), 0, wxEXPAND);
    pTopSizer->Add(xGapHandlingSizer(), 0, wxEXPAND);
    pTopSizer->Add(xCdsHandlingSizer(), 0, wxEXPAND);
    pTopSizer->Fit(m_Window);

    return m_Window;
}

IEditCommand* CEditObjectFeaturePropagate::GetEditCommand()
{
    if (!m_Window) {
        return 0;
    }
    if (m_EditAction) {
        return m_EditAction.GetPointer();
    }

    string key = ToStdString(m_pTargetCtrl->GetStringSelection());
    CBioseq_Handle bsh = m_SeqMap[key];
    if (bsh) {
        return xGetEditCommandSingleSequence(bsh);
    }
    else {//m_SeqMap key was the dummy for "All Sequences"
        return xGetEditCommandAllSequences();
    }
}

IEditCommand* CEditObjectFeaturePropagate::xGetEditCommandAllSequences()
{
    map<string, objects::CBioseq_Handle>::iterator it = m_SeqMap.begin();
    CRef<CCmdComposite> composite(new CCmdComposite("Propagate Feature"));
    CConstRef<CSeq_id> pBase = m_Bsh.GetSeqId();
    for (/**/; it != m_SeqMap.end(); ++it) {
        CBioseq_Handle bsh = it->second;
        if (!bsh  ||  bsh == m_Bsh) {
            continue;
        }
        CConstRef<CSeq_id> pCurrent = bsh.GetSeqId();
        vector<const CSeq_feat*>::const_iterator cit = m_PropagatedFeats.begin();
        for (/**/; cit != m_PropagatedFeats.end(); ++cit) {
            const CSeq_feat* pPropagateFeat = *cit;
            CCmdPropagateFeat* pCommand = new CCmdPropagateFeat(*pPropagateFeat, bsh);
            if (xStopCdsAtStopCodon()) {
                pCommand->SetStopCdsAtStopCodon();
            }
            CIRef<IEditCommand> pPropagateCmd(pCommand);
            composite->AddCommand(*pPropagateCmd);
        }
    }
    m_EditAction.Reset(composite);
    return m_EditAction.GetPointer();
}

IEditCommand* CEditObjectFeaturePropagate::xGetEditCommandSingleSequence(
    CBioseq_Handle bsh)
{
    CRef<CCmdComposite> composite(new CCmdComposite("Propagate Feature"));
    vector<const CSeq_feat*>::const_iterator cit;
    for (cit = m_PropagatedFeats.begin(); cit != m_PropagatedFeats.end(); ++cit) {
        const CSeq_feat* pPropagateFeat = *cit;
        CCmdPropagateFeat* pCommand = new CCmdPropagateFeat(*pPropagateFeat, bsh);
        if (xStopCdsAtStopCodon()) {
            pCommand->SetStopCdsAtStopCodon();
        }
        CIRef<IEditCommand> pPropagateCmd(pCommand);
        composite->AddCommand(*pPropagateCmd);
    }
    m_EditAction.Reset(composite);
    return m_EditAction.GetPointer();
}

wxSizer* CEditObjectFeaturePropagate::xTargetSizer()
{
    wxSizer* pSizer = new wxBoxSizer(wxHORIZONTAL);
    wxStaticText* pLabel = new wxStaticText(m_Window, -1, 
        "Propagation target:");
    pSizer->Add(pLabel, 0, wxALL, 2);
    m_pTargetCtrl = new wxComboBox(m_Window, -1);
    m_pTargetCtrl->Append("All Sequences");
    m_pTargetCtrl->Append(m_SeqStrings);
    m_pTargetCtrl->SetSelection(0);
    pSizer->Add(m_pTargetCtrl, 1, wxEXPAND|wxALL, 4);
    return pSizer;
}

wxSizer* CEditObjectFeaturePropagate::xFeatureSelectSizer()
{
    wxArrayString labels;
    labels.Add("All features");
    labels.Add("Selected feature");
    m_pFeatureSelectCtrl = new wxRadioBox(m_Window, -1, "Feature Selection:", 
        wxDefaultPosition, wxDefaultSize, labels, 2);
    m_pFeatureSelectCtrl->SetSelection(0);
    wxSizer* pSizer = new wxBoxSizer(wxHORIZONTAL);
    pSizer->Add(m_pFeatureSelectCtrl, 1, wxALL|wxEXPAND, 4);
    return pSizer;
}

wxSizer* CEditObjectFeaturePropagate::xGapHandlingSizer()
{
    wxArrayString labels;
    labels.Add("Extend over gaps");
    labels.Add("Split at gaps");
    m_pGapHandlingCtrl = new wxRadioBox(m_Window, -1, "Gap Handling:", 
        wxDefaultPosition, wxDefaultSize, labels, 2);
    m_pGapHandlingCtrl->SetSelection(0);
    wxSizer* pSizer = new wxBoxSizer(wxVERTICAL);
    pSizer->Add(m_pGapHandlingCtrl, 1, wxALL|wxEXPAND, 4);
    return pSizer;
}

wxSizer* CEditObjectFeaturePropagate::xCdsHandlingSizer()
{
    wxSizer* pSizer = new wxBoxSizer(wxVERTICAL);
    //pSizer->AddSpacer(8);
    m_pStopCdsAtStopCodon = new wxCheckBox(m_Window, -1,
        "Stop CDS translation at internal stop codon");
    m_pStopCdsAtStopCodon->SetValue(false);
    pSizer->Add(m_pStopCdsAtStopCodon, 1, wxALL|wxEXPAND, 4);
    m_pExtendCdsPastPartialThreePrime = new wxCheckBox(m_Window, -1,
        "Translate CDS after partial 3\' boundary");
    m_pExtendCdsPastPartialThreePrime->SetValue(false);
    pSizer->Add(m_pExtendCdsPastPartialThreePrime, 1, wxALL|wxEXPAND, 4);
    m_pCleanupCdsPartials = new wxCheckBox(m_Window, -1,
        "Cleanup CDS partials after propagation");
    m_pCleanupCdsPartials->SetValue(true);
    m_pCleanupCdsPartials->Enable(false);
    pSizer->Add(m_pCleanupCdsPartials, 1, wxALL|wxEXPAND, 4);
    m_pFuseAdjacentIntervals = new wxCheckBox(m_Window, -1,
        "Fuse adjacent propagated intervals");
    m_pFuseAdjacentIntervals->SetValue(false);
    pSizer->Add(m_pFuseAdjacentIntervals, 1, wxALL|wxEXPAND, 4);
    //pSizer->AddSpacer(8);
    //m_pStopCdsAtStopCodon->Enable(false);
    m_pExtendCdsPastPartialThreePrime->Enable(false);
    m_pCleanupCdsPartials->Enable(false);
    m_pFuseAdjacentIntervals->Enable(false);
    return pSizer;
}

bool CEditObjectFeaturePropagate::xStopCdsAtStopCodon() const
{
    if (!m_pStopCdsAtStopCodon) {
        return false;
    }
    return m_pStopCdsAtStopCodon->GetValue();
}


void CEditObjectFeaturePropagate::xInitSequenceMap()
{
    //Populate m_SeqStrings with the ID strings of all sibling sequences
    // of m_Bsh
    //Populate m_SeqMap with all sibling sequences, indexed by their
    // ID strings
    //
    CSeq_entry_Handle seh = m_Bsh.GetTopLevelEntry();
    for (CBioseq_CI bsit = CBioseq_CI(seh); bsit; ++bsit) {
        CBioseq_Handle bsh = *bsit;
        if (bsh == m_Bsh) {
            continue;
        }
        if (bsh.IsAa()) {
            continue;
        }
        CSeq_id_Handle idh = CSeq_id_Handle::GetHandle(*bsh.GetSeqId());
        CSeq_id_Handle best_idh = sequence::GetId(
            idh, bsh.GetScope(), sequence::eGetId_Best);
        if (best_idh) {
            idh = best_idh;
        }
        string label = idh.GetSeqId()->GetSeqIdString(true);
        m_SeqStrings.Add(label);
        m_SeqMap[label] = bsh;
    }
}

END_NCBI_SCOPE
