/*  $Id: edit_object_seq_desc.cpp 27443 2013-02-15 22:27:12Z filippov $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors: Colleen Bollin, based on a file by Roman Katargin
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>

#include "edit_object_seq_desc.hpp"

#include <gui/objutils/undo_manager_interface.hpp>
#include <gui/objutils/descriptor_change.hpp>
#include <gui/objutils/cmd_create_desc.hpp>

#include <gui/objutils/seqdesc_title_edit.hpp>
#include <gui/objutils/seqdesc_title_change.hpp>

#include <objects/seq/Seqdesc.hpp>
#include <objects/seq/Pubdesc.hpp>
#include <objects/general/Object_id.hpp>
#include <objects/general/User_field.hpp>
#include <objects/misc/sequence_macros.hpp>

#include <objmgr/util/sequence.hpp>
#include <objmgr/seqdesc_ci.hpp>
#include <objmgr/bioseq_ci.hpp>

#include <wx/panel.h>
#include <wx/treebook.h>

#include "descriptortext_panel.hpp"
#include "publicationtype_panel.hpp"
#include "molinfo_panel.hpp"
#include "structuredcomment_panel.hpp"
#include <gui/widgets/edit/sequencingtechnologypanel.hpp>
#include <gui/widgets/edit/dblinkpanel.hpp>
#include <gui/widgets/edit/biosource_autocomplete.hpp>

#include "loc_gcode_panel.hpp"
#include "biosourcepanel.hpp"
#include "srcmod_list_panel.hpp"
#include "dbxref_panel.hpp"

#ifdef _DEBUG
#include <serial/objostr.hpp>
#include <serial/serial.hpp>
#endif

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

static CConstRef<CBioseq> s_GetBioseqForSeqdesc (CRef<CScope> scope, const CSeqdesc& seq_desc)
{
    CConstRef<CBioseq> bioseq;

    if (!scope) {
        return bioseq;
    }

    CScope::TTSE_Handles tses;
    scope->GetAllTSEs(tses, CScope::eAllTSEs);
    ITERATE (CScope::TTSE_Handles, handle, tses) {
        for (CBioseq_CI bioseq_it(*handle);  bioseq_it;  ++bioseq_it) {
            // Is Seqdesc on this Bioseq?
            if (bioseq_it->IsSetDescr()) {
                CConstRef<CBioseq> r_bioseq = bioseq_it->GetCompleteBioseq();
                ITERATE (CBioseq::TDescr::Tdata, dit, r_bioseq->GetDescr().Get()) {
                    if (dit->GetPointer() == &seq_desc) {
                        return r_bioseq;
                    }
                }
            }
        }
    }
    return bioseq;
}


wxWindow* CEditObjectSeq_desc::CreateWindow(wxWindow* parent)
{
    if (m_Window) {
        NCBI_THROW(CException, eUnknown, 
                   "CEditObjectSeq_desc::CreateWindow - Window already created!");
    }

    const CSeqdesc& seq_desc = dynamic_cast<const CSeqdesc&>(*m_Object);

    m_EditedDescriptor.Reset((CSerialObject*)CSeqdesc::GetTypeInfo()->Create());
    m_EditedDescriptor->Assign(seq_desc);
    m_EditedBioseq.Reset();
    // need to get bioseq for descriptor if editing molinfo
    if (seq_desc.IsMolinfo()) {
        CConstRef<CBioseq> b = s_GetBioseqForSeqdesc(m_Scope, seq_desc);
        if (!b) {
            if (m_SEH.IsSeq()) {
                b = m_SEH.GetSeq().GetCompleteBioseq();
            }
        }
        if (b) {
            CBioseq* bioseq = (CBioseq*)(b.GetNonNullPointer());
            m_Bioseq.Reset((CSerialObject*)bioseq);
            CRef<CBioseq> new_bioseq (new CBioseq());
            new_bioseq->Assign (*bioseq);
            m_EditedBioseq.Reset((CSerialObject*)new_bioseq.GetPointer());
        }
    }

    CSeqdesc& edited_seq_desc = dynamic_cast<CSeqdesc&>(*m_EditedDescriptor);

    string subtype;
    edited_seq_desc.GetLabel(&subtype, CSeqdesc::eType);

    CDescriptorTextPanel * text_panel = 0;

    switch (edited_seq_desc.Which()) {
        case CSeqdesc::e_Comment:
            text_panel = new CDescriptorTextPanel( parent );
            text_panel->SetText (ToWxString(edited_seq_desc.GetComment()));
            m_Window = text_panel;
            break;
        case CSeqdesc::e_Title:
            text_panel = new CDescriptorTextPanel( parent );
            text_panel->SetText (ToWxString(edited_seq_desc.GetTitle()));
            m_Window = text_panel;
            break;
        case CSeqdesc::e_Name:
            text_panel = new CDescriptorTextPanel( parent );
            text_panel->SetText (ToWxString(edited_seq_desc.GetName()));
            m_Window = text_panel;
            break;
        case CSeqdesc::e_Region:
            text_panel = new CDescriptorTextPanel( parent );
            text_panel->SetText (ToWxString(edited_seq_desc.GetRegion()));
            m_Window = text_panel;
            break;
        case CSeqdesc::e_Pub:
            {
            CRef<objects::CPubdesc> pubdesc(&(edited_seq_desc.SetPub()));
            m_Window = new CPublicationTypePanel ( parent, pubdesc );
            }
            break;
        case CSeqdesc::e_Molinfo:
            {
            CRef<objects::CMolInfo> molinfo(&(edited_seq_desc.SetMolinfo()));
            CRef<objects::CBioseq> bioseq;
            if (m_EditedBioseq) {
                CBioseq&  edited_bioseq   = dynamic_cast<CBioseq&>(*m_EditedBioseq);
                bioseq = &(edited_bioseq);
            }
            m_Window = new CMolInfoPanel ( parent, molinfo, bioseq);
            }
            break;
        case CSeqdesc::e_User:
            {
            CRef<objects::CUser_object> user(&(edited_seq_desc.SetUser()));
            if (user->IsSetType() 
                && user->GetType().IsStr()) {
                if (NStr::EqualNocase(user->GetType().GetStr(), "StructuredComment")) {
                    bool is_assembly = false;
                    ITERATE (CUser_object::TData, it, user->GetData()) {
                        if ((*it)->IsSetLabel()
                            && (*it)->GetLabel().IsStr()
                            && NStr::Equal((*it)->GetLabel().GetStr(), "StructuredCommentPrefix")
                            && (*it)->IsSetData()
                            && (*it)->GetData().IsStr()
                            && NStr::Equal((*it)->GetData().GetStr(), "##Assembly-Data-START##")) {
                            is_assembly = true;
                            break;
                        }
                    }
                    if (is_assembly) {
                        m_Window = new CSequencingTechnologyPanel ( parent, user );
                    } else {
                        m_Window = new CStructuredCommentPanel ( parent, user );
                    }
                } else if (NStr::EqualNocase(user->GetType().GetStr(), "DBLink")) {
                    m_Window = new CDBLinkPanel ( parent, user );
                } else {
                    m_Window = new wxPanel( parent );
                }
            } else {
                m_Window = new wxPanel( parent );
            }
            }
            break;
		    case CSeqdesc::e_Source:
			      {
				        m_Window = new wxPanel(parent);
        				
				        wxBoxSizer* source_sizer = new wxBoxSizer(wxHORIZONTAL);
				        m_Window->SetSizer(source_sizer);

				        m_SourceTreebook = new wxTreebook(m_Window, ID_TREEBOOK, wxDefaultPosition, wxSize(-1, 420), wxBK_DEFAULT );
                                        source_sizer->Add(m_SourceTreebook, 0, wxALIGN_CENTER_VERTICAL|wxALL, 5);

				        wxWindow* panel = new CBioSourcePanel(m_SourceTreebook, edited_seq_desc.SetSource());
				        m_SourceTreebook->AddPage(panel, wxT("Source"));

				        panel = new CLocAndGCodePanel(m_SourceTreebook, edited_seq_desc.SetSource());
				        m_SourceTreebook->AddPage(panel, wxT("Genetic Code"));

				        panel = new CSrcModListPanel(m_SourceTreebook, edited_seq_desc.SetSource());
				        m_SourceTreebook->AddPage(panel, wxT("Source Modifiers"));
        				
                                        panel = new CDbxrefPanel(m_SourceTreebook, edited_seq_desc.SetSource().SetOrg());
				        m_SourceTreebook->AddPage(panel, wxT("Dbxrefs"));
			      }
			      break;
        default:
            m_Window = new wxPanel( parent );
            break;
    }

    return m_Window;
}


IEditCommand* CEditObjectSeq_desc::GetEditCommand()
{
    if (!m_Window)
        return 0;

    if (!m_EditAction) {
        CSeqdesc& edited_seq_desc = dynamic_cast<CSeqdesc&>(*m_EditedDescriptor);

        CPublicationTypePanel* pubPanel = 0;
        CMolInfoPanel* molPanel = 0;
        CStructuredCommentPanel* strucPanel = 0;
        CDBLinkPanel* dbLinkPanel = 0;
        CSequencingTechnologyPanel* assemPanel = 0;
		    CBioSourcePanel* srcPanel = 0;

        string text;
        CDescriptorTextPanel* textPanel = dynamic_cast<CDescriptorTextPanel*>(m_Window);
        if (textPanel) {
            text = ToStdString (textPanel->GetText());            
            switch (edited_seq_desc.Which()) {
                case CSeqdesc::e_Comment:
                    edited_seq_desc.SetComment(text);
                    break;
                case CSeqdesc::e_Title:
                    edited_seq_desc.SetTitle(text);
                    break;
                case CSeqdesc::e_Name:
                    edited_seq_desc.SetName(text);
                    break;
                case CSeqdesc::e_Region:
                    edited_seq_desc.SetRegion(text);
                    break;
                default:
                    break;
            }
        } else if ((pubPanel = dynamic_cast<CPublicationTypePanel*>(m_Window)) != 0) {
            CRef<objects::CPubdesc> pub = pubPanel->GetPubdesc();
            edited_seq_desc.SetPub(*pub);
        } else if ((molPanel = dynamic_cast<CMolInfoPanel*>(m_Window)) != 0) {
            CRef<objects::CMolInfo> molinfo = molPanel->GetMolinfo();
            edited_seq_desc.SetMolinfo(*molinfo);
            CBioseq&  edited_bioseq   = dynamic_cast<CBioseq&>(*m_EditedBioseq);
            CRef<objects::CBioseq> bioseq(&(edited_bioseq));
            molPanel->SetBioseqValues(bioseq);
        } else if ((strucPanel = dynamic_cast<CStructuredCommentPanel*>(m_Window)) != 0) {
            CRef<CUser_object> user = strucPanel->GetUser_object();
            edited_seq_desc.SetUser(*user);
        } else if ((dbLinkPanel = dynamic_cast<CDBLinkPanel*>(m_Window)) != 0) {
            CRef<CUser_object> user = dbLinkPanel->GetUser_object();
            edited_seq_desc.SetUser(*user);
        } else if ((assemPanel = dynamic_cast<CSequencingTechnologyPanel*>(m_Window)) != 0) {
            CRef<CUser_object> user = assemPanel->GetUser_object();
            edited_seq_desc.SetUser(*user);
		    } else if (m_SourceTreebook != 0) {
		        CBioSourcePanel* biosource_panel = 0;
            CLocAndGCodePanel* loc_gc_panel = 0;
		        CSrcModListPanel* src_mod_panel = 0;
            CDbxrefPanel* xrefPanel = 0;
		        for (size_t i = 0; i < m_SourceTreebook->GetPageCount(); ++i) {
                if ((biosource_panel = dynamic_cast<CBioSourcePanel*>(m_SourceTreebook->GetPage(i))) != 0) {
                    biosource_panel->PopulateBioSource(edited_seq_desc.SetSource());
                } else if ((loc_gc_panel = dynamic_cast<CLocAndGCodePanel*>(m_SourceTreebook->GetPage(i))) != 0) {
                    loc_gc_panel->PopulateLocAndGC(edited_seq_desc.SetSource());
                } else if ((src_mod_panel = dynamic_cast<CSrcModListPanel*>(m_SourceTreebook->GetPage(i))) != 0) {
                    src_mod_panel->PopulateModifiers(edited_seq_desc.SetSource());
                } else if ((xrefPanel = dynamic_cast<CDbxrefPanel*>(m_SourceTreebook->GetPage(i))) != 0) {
                    xrefPanel->PopulateDbxrefs(edited_seq_desc.SetSource().SetOrg());
                }
            }
            CBioSourceAutoComplete::AutoFill(edited_seq_desc.SetSource().SetOrg());
		    }

		    const CSeqdescTitleEdit* seq_desc_edit = dynamic_cast<const CSeqdescTitleEdit*>(m_Object.GetPointer());
        if (seq_desc_edit) {
            CBioseq_Handle bh = seq_desc_edit->GetBioseq_Handle();
            m_EditAction.Reset(new CChangeSeqdescTitleCommand(bh, text));
        }
        else {
            try {
                if (m_CreateMode) {
                    m_EditAction.Reset(new CCmdCreateDesc(m_SEH,
                        dynamic_cast<const CSeqdesc&>(m_EditedDescriptor.GetObject())));
                }
                else {
                    CChangeSeqDescCommand* cmd = new CChangeSeqDescCommand();
                    CObject* actual = (CObject*) m_Object.GetPointer();
                    cmd->Add(actual, CConstRef<CObject>(m_EditedDescriptor));
#ifdef _DEBUG
    CNcbiOstrstream ostr;
    {{
        auto_ptr<CObjectOStream>
            out(CObjectOStream::Open(eSerial_AsnText, ostr));
        *out << dynamic_cast<CSeqdesc&>(*m_EditedDescriptor);
		*out << dynamic_cast<CSeqdesc&>(*actual);
    }}
    _TRACE(string(CNcbiOstrstreamToString(ostr)));
#endif
                    if (edited_seq_desc.IsMolinfo() && m_Bioseq) {
                        SConstScopedObject sobject(*m_Bioseq, *m_Scope);
                        cmd->AddBioseq(sobject, m_EditedBioseq);
                    }
                    m_EditAction.Reset(cmd);
                }
            }
            catch (CException& e) {
                LOG_POST(Error << "CEditObjectSeq_desc::GetEditAction(): " << e.GetMsg());
            }
        }
#ifdef _DEBUG
    CNcbiOstrstream ostr;
    {{
        auto_ptr<CObjectOStream>
            out(CObjectOStream::Open(eSerial_AsnText, ostr));
        *out << edited_seq_desc;
    }}
    _TRACE(string(CNcbiOstrstreamToString(ostr)));
	//LOG_POST(Info << string(CNcbiOstrstreamToString(ostr)));
#endif
    }

    return m_EditAction.GetPointer();
}

wxWindow* CEditObjectPubdesc::CreateWindow(wxWindow* parent)
{
    if (m_Window) {
        NCBI_THROW(CException, eUnknown, 
                   "CEditObjectSeq_desc::CreateWindow - Window already created!");
    }

    const CPubdesc& pubdesc = dynamic_cast<const CPubdesc&>(*m_Object);

    m_EditedPubdesc.Reset((CSerialObject*)CPubdesc::GetTypeInfo()->Create());
    m_EditedPubdesc->Assign(pubdesc);

    CPubdesc& edited_pubdesc = dynamic_cast<CPubdesc&>(*m_EditedPubdesc);

    CRef<objects::CPubdesc> r_pubdesc(&edited_pubdesc);
    m_Window = new CPublicationTypePanel ( parent, r_pubdesc );

    return m_Window;
}


IEditCommand* CEditObjectPubdesc::GetEditCommand()
{
    if (!m_Window)
        return 0;

    if (!m_EditAction) {
        CPubdesc& pubdesc = dynamic_cast<CPubdesc&>(*m_EditedPubdesc);

        CPublicationTypePanel* pubPanel = dynamic_cast<CPublicationTypePanel*>(m_Window);        
        CRef<objects::CPubdesc> pub = pubPanel->GetPubdesc();
        pubdesc.Assign(*pub);
        try {
            CChangePubdescCommand* cmd = new CChangePubdescCommand();
            CObject* actual = (CObject*) m_Object.GetPointer();
            cmd->Add(actual, CConstRef<CObject>(m_EditedPubdesc));
            m_EditAction.Reset(cmd);
        }
        catch (CException& e) {
            LOG_POST(Error << "CEditObjectPubdesc::GetEditAction(): " << e.GetMsg());
        }
    }

    return m_EditAction.GetPointer();
}

static CConstRef<CBioseq> s_GetBioseqForMolInfo (CRef<CScope> scope, const CMolInfo& molinfo)
{
    CConstRef<CBioseq> bioseq;

    if (!scope) {
        return bioseq;
    }

    CScope::TTSE_Handles tses;
    scope->GetAllTSEs(tses, CScope::eAllTSEs);
    ITERATE (CScope::TTSE_Handles, handle, tses) {
        for (CBioseq_CI bioseq_it(*handle);  bioseq_it;  ++bioseq_it) {
            // Is Seqdesc on this Bioseq?
            if (bioseq_it->IsSetDescr()) {
                CConstRef<CBioseq> r_bioseq = bioseq_it->GetCompleteBioseq();
                ITERATE (CBioseq::TDescr::Tdata, dit, r_bioseq->GetDescr().Get()) {
                    if ((*dit)->IsMolinfo() && &((*dit)->GetMolinfo()) == &molinfo) {
                        return r_bioseq;
                    }
                }
            }
        }
    }
    return bioseq;
}


wxWindow* CEditObjectMolInfo::CreateWindow(wxWindow* parent)
{
    if (m_Window) {
        NCBI_THROW(CException, eUnknown, 
                   "CEditObjectSeq_desc::CreateWindow - Window already created!");
    }

    const CMolInfo& molinfo = dynamic_cast<const CMolInfo&>(*m_Object);

    m_EditedMolInfo.Reset((CSerialObject*)CMolInfo::GetTypeInfo()->Create());
    m_EditedMolInfo->Assign(molinfo);

    m_EditedBioseq.Reset();
    CConstRef<CBioseq> b = s_GetBioseqForMolInfo(m_Scope, molinfo);
    if (b) {
        CBioseq* bioseq = (CBioseq*)(b.GetNonNullPointer());
        m_Bioseq.Reset((CSerialObject*)bioseq);
        CRef<CBioseq> new_bioseq (new CBioseq());
        new_bioseq->Assign (*bioseq);
        m_EditedBioseq.Reset((CSerialObject*)new_bioseq.GetPointer());
    }

    CMolInfo& edited_molinfo = dynamic_cast<CMolInfo&>(*m_EditedMolInfo);

    CRef<objects::CMolInfo> r_molinfo(&edited_molinfo);
    CRef<objects::CBioseq> bioseq;
    if (m_EditedBioseq) {
        CBioseq&  edited_bioseq   = dynamic_cast<CBioseq&>(*m_EditedBioseq);
        bioseq = &(edited_bioseq);
    }
    m_Window = new CMolInfoPanel ( parent, r_molinfo, bioseq);

    return m_Window;
}


IEditCommand* CEditObjectMolInfo::GetEditCommand()
{
    if (!m_Window)
        return 0;

    if (!m_EditAction) {
        CMolInfo& edited_molinfo = dynamic_cast<CMolInfo&>(*m_EditedMolInfo);

        CMolInfoPanel* molPanel = dynamic_cast<CMolInfoPanel*>(m_Window);
        CRef<objects::CMolInfo> molinfo = molPanel->GetMolinfo();
        edited_molinfo.Assign(*molinfo);
        CBioseq&  edited_bioseq   = dynamic_cast<CBioseq&>(*m_EditedBioseq);
        CRef<objects::CBioseq> bioseq(&(edited_bioseq));
        molPanel->SetBioseqValues(bioseq);

        try {
            CChangeMolInfoCommand* cmd = new CChangeMolInfoCommand();
            CObject* actual = (CObject*) m_Object.GetPointer();
            cmd->Add(actual, CConstRef<CObject>(m_EditedMolInfo));
            if (m_Bioseq) {
                SConstScopedObject sobject(*m_Bioseq, *m_Scope);
                cmd->AddBioseq(sobject, m_EditedBioseq);
            }
            m_EditAction.Reset(cmd);
        }
        catch (CException& e) {
            LOG_POST(Error << "CEditObjectMolInfo::GetEditAction(): " << e.GetMsg());
        }
    }

    return m_EditAction.GetPointer();
}

END_NCBI_SCOPE
