/*  $Id: edit_object_seq_feat.cpp 28135 2013-05-14 19:41:26Z bollin $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors: Roman Katargin
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>

#include "edit_object_seq_feat.hpp"

#include <gui/objutils/undo_manager_interface.hpp>
#include <gui/objutils/cmd_create_feat.hpp>
#include <gui/objutils/cmd_change_seq_feat.hpp>
#include <gui/objutils/cmd_change_bioseq_inst.hpp>
#include <gui/objutils/cmd_composite.hpp>
#include <gui/objutils/cmd_add_seqentry.hpp>

#include <objects/seqfeat/Seq_feat.hpp>
#include <objects/seqfeat/SeqFeatXref.hpp>
#include <objects/seqfeat/RNA_ref.hpp>
#include <objects/seqfeat/Cdregion.hpp>
#include <objects/seq/Pubdesc.hpp>
#include <objects/misc/sequence_macros.hpp>

#include <objmgr/util/sequence.hpp>
#include <objmgr/feat_ci.hpp>

#include <wx/treebook.h>
#include <wx/panel.h>
//#include <wx/sizer.h>
//#include <wx/bitmap.h>
//#include <wx/icon.h>
//#include <wx/choice.h>
//#include <wx/textctrl.h>
//#include <wx/stattext.h>
//#include <wx/toplevel.h>
#include <wx/hyperlink.h>


#include "generic_props_panel.hpp"
#include "comment_panel.hpp"
#include "location_panel.hpp"
#include "gene_panel.hpp"
#include "synonyms_panel.hpp"
#include "rna_panel.hpp"
#include "cds_exception_panel.hpp"
#include "cds_protein_panel.hpp"
#include "cds_translation_panel.hpp"
#include "protein_panel.hpp"
#include "ec_numbers_panel.hpp"
#include "protein_activities_panel.hpp"
#include "importfeature_panel.hpp"
#include "gbqual_panel.hpp"
#include "dbxref_panel.hpp"
#include "identifiers_panel.hpp"
#include "experiments_panel.hpp"
#include "inference_list_panel.hpp"
#include "citations_panel.hpp"
#include "loc_gcode_panel.hpp"
#include "biosourcepanel.hpp"
#include "srcmod_list_panel.hpp"
#include "featuregenepanel.hpp"
#include "region_panel.hpp"
#include "bond_panel.hpp"
#include "site_panel.hpp"

// publication editor includes
#include "publicationtype_panel.hpp"

#ifdef _DEBUG
#include <serial/objostr.hpp>
#include <serial/serial.hpp>
#endif

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

static const CBioseq* GetCDSProductSeq(const CSeq_feat* feat, CScope& scope)
{
    const CBioseq* protein = NULL;

    if (feat->IsSetProduct() && feat->GetProduct().GetId()) {
        CBioseq_Handle bsh = scope.GetBioseqHandle(*(feat->GetProduct().GetId()));
        if (bsh) {
            protein = bsh.GetCompleteBioseq().GetPointer();
        }
    }
    return protein;
}

    
static const CSeq_feat* GetFeatureProtein(const CSeq_feat* feat, CScope& scope)
{
    //const CProt_ref* protRef = feat->GetProtXref();
    const CSeq_feat* protein = NULL;

    if (feat->IsSetProduct()) {
        const CSeq_loc& product = feat->GetProduct();
        int dist = numeric_limits<int>::max();
        CFeat_CI it(scope, product, SAnnotSelector(CSeqFeatData::e_Prot));
        for (; it; ++it) {
            const CSeq_loc& loc = it->GetLocation();
            sequence::ECompare comp = sequence::Compare(product, loc, &scope);
            if (comp == sequence::eSame) {
                protein = &it->GetOriginalFeature();
                break;
            }
            else if (comp == sequence::eContains) {
                int tmp = sequence::GetLength(product, &scope) -
                    sequence::GetLength(loc, &scope);
                if (tmp < dist) {
                    dist = tmp;
                    protein = &it->GetOriginalFeature();
                }
            }
        }
    }

    return protein;
}


void CEditObjectSeq_feat::x_AddCodingRegionProteinNamePanel(objects::CProt_ref& prot)
{
    wxWindow* panel;

    panel = new CCDSProteinPanel(m_Window, prot, *m_Scope);
    m_Window->AddPage(panel, wxT("Protein Name and Description"));
}


void CEditObjectSeq_feat::x_AddCodingRegionProteinPanels(objects::CProt_ref& prot)
{
    wxWindow* panel;

    panel = new CECNumbersPanel(m_Window, prot);
    m_Window->AddPage(panel, wxT("Enzyme Commission Numbers"));
    panel = new CProteinActivitiesPanel(m_Window, prot);
    m_Window->AddPage(panel, wxT("Activity"));
}


wxWindow* CEditObjectSeq_feat::CreateWindow(wxWindow* parent)
{
    if (m_Window) {
        NCBI_THROW(CException, eUnknown, 
                   "CEditObjectSeq_feat::CreateWindow - Window already created!");
    }

    m_Window = new wxTreebook( parent, ID_NOTEBOOK, wxDefaultPosition, wxDefaultSize, wxBK_DEFAULT );

    const CSeq_feat& seq_feat = dynamic_cast<const CSeq_feat&>(*m_Object);

    m_EditedFeature.Reset((CSerialObject*)CSeq_feat::GetTypeInfo()->Create());
    m_EditedFeature->Assign(seq_feat);

    CSeq_feat& edited_seq_feat = dynamic_cast<CSeq_feat&>(*m_EditedFeature);
    CSeqFeatData& seq_feat_data = edited_seq_feat.SetData();

    string sub_type = seq_feat_data.GetKey();
    if (sub_type.find("RNA") == sub_type.length()-3)
        sub_type = "RNA";

    wxWindow* panel;

    if (seq_feat_data.IsGene()) {
        CGene_ref& gene_ref = seq_feat_data.SetGene();

        panel = new CGenePanel(m_Window, gene_ref, *m_Scope);
        m_Window->AddPage(panel, wxT("Gene"));

        panel = new CLocationPanel(m_Window, edited_seq_feat.SetLocation(), *m_Scope, true, false);
        m_Window->AddPage(panel, wxT("Location"));

        panel = new CSynonymsPanel(m_Window, gene_ref, *m_Scope);
        m_Window->AddPage(panel, wxT("Synonyms"));

    }
    else if (seq_feat_data.IsRna()) {
        panel = new CRNAPanel(m_Window, edited_seq_feat, *m_Scope);
        m_Window->AddPage(panel, wxT("RNA"));

        panel = new CLocationPanel(m_Window, edited_seq_feat.SetLocation(), *m_Scope, true, false);
        m_Window->AddPage(panel, wxT("Location"));

    }
    else if (seq_feat_data.IsCdregion()) {
        CCdregion& cdregion = seq_feat_data.SetCdregion();
        const CSeq_feat* protein = GetFeatureProtein(&seq_feat, *m_Scope);
        CRef<CProt_ref> prot_feat;

        if (protein && protein->GetData().IsProt()) {
            m_EditedProtein.Reset(
                static_cast<CSeq_feat*>(protein->GetThisTypeInfo()->Create()));
            m_EditedProtein->Assign(*protein);

            CSeq_feat& edited_protein = dynamic_cast<CSeq_feat&>(*m_EditedProtein);
            prot_feat = &(edited_protein.SetData().SetProt());
        } else {
            bool has_xref = false;
            if (edited_seq_feat.IsSetXref()) {
                NON_CONST_ITERATE (CSeq_feat::TXref, it, edited_seq_feat.SetXref()) {
                    if ((*it)->IsSetData() && (*it)->GetData().IsProt()) {
                        prot_feat = &((*it)->SetData().SetProt());
                        has_xref = true;
                        break;
                    }
                }
            }
            if (!has_xref) {
                CRef<CSeqFeatXref> xref(new CSeqFeatXref());
                xref->SetData().SetProt();
                edited_seq_feat.SetXref().push_back(xref);
                prot_feat = &(xref->SetData().SetProt());
            }
        }
        panel = new CCDSTranslationPanel (m_Window, edited_seq_feat, *m_Scope);
        m_Window->AddPage(panel, wxT("Coding Region"));
        x_AddCodingRegionProteinNamePanel(*prot_feat);
       

        panel = new CLocationPanel(m_Window, edited_seq_feat.SetLocation(), *m_Scope, true, false);
        m_Window->AddPage(panel, wxT("Location"));

        x_AddCodingRegionProteinPanels(*prot_feat);
        panel = new CCDSExceptionPanel (m_Window, cdregion, edited_seq_feat.SetLocation(), *m_Scope);
        m_Window->AddPage(panel, wxT("Translation Exceptions"));
    } else if (seq_feat_data.IsPub()) {
        CRef<objects::CPubdesc> pubdesc(&(seq_feat_data.SetPub()));
        panel = new CPublicationTypePanel (m_Window, pubdesc);
        m_Window->AddPage (panel, wxT("Publication"));

        panel = new CLocationPanel(m_Window, edited_seq_feat.SetLocation(), *m_Scope, true, false);
        m_Window->AddPage(panel, wxT("Location"));

    }
    else if (seq_feat_data.IsImp()) {
        panel = new CImportFeaturePanel(m_Window, edited_seq_feat);
        m_Window->AddPage(panel, wxT("Import Feature"));

        panel = new CLocationPanel(m_Window, edited_seq_feat.SetLocation(), *m_Scope, true, false);
        m_Window->AddPage(panel, wxT("Location"));

    }
    else if (seq_feat_data.IsProt()) {
        panel = new CProteinPanel(m_Window, edited_seq_feat.SetData().SetProt());
        m_Window->AddPage(panel, wxT("Protein Name and Description"));
        panel = new CECNumbersPanel(m_Window, edited_seq_feat.SetData().SetProt());
        m_Window->AddPage(panel, wxT("Enzyme Commission Numbers"));
        panel = new CProteinActivitiesPanel(m_Window, edited_seq_feat.SetData().SetProt());
        m_Window->AddPage(panel, wxT("Activity"));
        panel = new CLocationPanel(m_Window, edited_seq_feat.SetLocation(), *m_Scope, false, true);
        m_Window->AddPage(panel, wxT("Location"));
    }
    else if (seq_feat_data.IsBiosrc()) {
        panel = new CBioSourcePanel(m_Window, edited_seq_feat.SetData().SetBiosrc());
        m_Window->AddPage(panel, wxT("Source"));
        
        panel = new CLocAndGCodePanel(m_Window, edited_seq_feat.SetData().SetBiosrc());
        m_Window->AddSubPage(panel, wxT("Genetic Code"));
        
        panel = new CSrcModListPanel(m_Window, edited_seq_feat.SetData().SetBiosrc());
        m_Window->AddSubPage(panel, wxT("Source Modifiers"));
        
        panel = new CDbxrefPanel(m_Window, edited_seq_feat.SetData().SetBiosrc().SetOrg());
        m_Window->AddSubPage(panel, wxT("Dbxrefs"));
        
        panel = new CLocationPanel(m_Window, edited_seq_feat.SetLocation(), *m_Scope, true, false);
        m_Window->AddPage(panel, wxT("Location"));
    }
    else if (seq_feat_data.IsRegion()) {
        panel = new CRegionPanel(m_Window, edited_seq_feat.SetData().SetRegion()); 
        m_Window->AddPage(panel, wxT("Region"));

        panel = new CLocationPanel(m_Window, edited_seq_feat.SetLocation(), *m_Scope, true, false);
        m_Window->AddPage(panel, wxT("Location"));
    }
    else if (seq_feat_data.IsBond()) {
        panel = new CBondPanel(m_Window, edited_seq_feat.SetData().SetBond()); 
        m_Window->AddPage(panel, wxT("Bond"));
        
        //panel = new CLocationPanel(m_Window, edited_seq_feat.SetLocation(), *m_Scope, true, false); // TODO - location can't seem to handle bond seq-loc
        //m_Window->AddPage(panel, wxT("Location"));
        }
    else if (seq_feat_data.IsSite()) {
        panel = new CSitePanel(m_Window, edited_seq_feat.SetData().SetSite()); 
        m_Window->AddPage(panel, wxT("Site"));
        
        panel = new CLocationPanel(m_Window, edited_seq_feat.SetLocation(), *m_Scope, true, false); 
        m_Window->AddPage(panel, wxT("Location"));
        }
    else {
        panel = new CLocationPanel(m_Window, edited_seq_feat.SetLocation(), *m_Scope);
        m_Window->AddPage(panel, wxT("Location"));
    }

    if (!seq_feat_data.IsGene()) {
        panel = new CFeatureGenePanel(m_Window, edited_seq_feat, *m_Scope);
        m_Window->AddPage(panel, wxT("Gene"));
    }

    panel = new CCommentPanel(m_Window, edited_seq_feat, *m_Scope);
    m_Window->AddPage(panel, wxT("Comment"));

    if (CGBQualPanel::PanelNeeded(edited_seq_feat)) {
        panel = new CGBQualPanel(m_Window, edited_seq_feat);
        m_Window->AddPage (panel, wxT("GenBank Qualifiers"));
    }

    panel = new CGenericPropsPanel(m_Window, edited_seq_feat, *m_Scope);
    m_Window->AddPage(panel, wxT("General                         "));

    panel = new CDbxrefPanel (m_Window, edited_seq_feat);
    m_Window->AddSubPage (panel, wxT("Dbxrefs"));

    panel = new CIdentifiersPanel (m_Window, edited_seq_feat);
    m_Window->AddSubPage (panel, wxT("Feature Identifiers"));

    panel = new CExperimentsPanel (m_Window, edited_seq_feat);
    m_Window->AddSubPage (panel, wxT("Experiments"));

    panel = new CInferenceListPanel (m_Window, edited_seq_feat);
    m_Window->AddSubPage (panel, wxT("Inferences"));

    panel = new CCitationsPanel (m_Window, edited_seq_feat, *m_Scope);
    m_Window->AddSubPage (panel, wxT("Citations"));

    return m_Window;
}


bool s_ShouldSetFeaturePartial(const CSeq_loc& loc, CRef<CScope> scope)
{
    bool rval = false;
    unsigned int partial_loc  = sequence::SeqLocPartialCheck(loc, scope );
    if (( partial_loc & sequence::eSeqlocPartial_Start )
        || ( partial_loc & sequence::eSeqlocPartial_Stop )
        || ( partial_loc & sequence::eSeqlocPartial_Internal )
        || ( partial_loc & sequence::eSeqlocPartial_Other )
        || ( partial_loc & sequence::eSeqlocPartial_Nostart )
        || ( partial_loc & sequence::eSeqlocPartial_Nostop )
        || ( partial_loc & sequence::eSeqlocPartial_Nointernal)) {
        rval = true;
    }
    return rval;
}


IEditCommand* CEditObjectSeq_feat::GetEditCommand()
{
    if (!m_Window)
        return 0;

    if (!m_EditAction) {
        CRef<CCmdComposite> composite(new CCmdComposite("Edit Feature"));

        const CSeq_feat& seq_feat = dynamic_cast<const CSeq_feat&>(*m_Object);
        CSeq_feat& edited_seq_feat = dynamic_cast<CSeq_feat&>(*m_EditedFeature);

        CLocationPanel* locPanel = 0;
        for (size_t i = 0; i < m_Window->GetPageCount(); ++i) {
            locPanel = dynamic_cast<CLocationPanel*>(m_Window->GetPage(i));
            if (locPanel)
                break;
        }
        if (locPanel) {
            CRef<objects::CSeq_loc> loc = locPanel->GetSeq_loc();
            edited_seq_feat.SetLocation(*loc);
            // note - fields from generic properties panel have already been set
            // if feature was deliberately flagged as partial, will have already been set, otherwise will have been reset
            if (s_ShouldSetFeaturePartial(*loc, m_Scope)) {
                edited_seq_feat.SetPartial(true);
            }
        }

        CConstRef<CBioseq> orig_product;
        CRef<CBioseq> edited_product;
        CCDSTranslationPanel* transPanel = 0;
        for (size_t i = 0; i < m_Window->GetPageCount(); ++i) {
            transPanel = dynamic_cast<CCDSTranslationPanel*>(m_Window->GetPage(i));
            if (transPanel)
                break;
        }
        if (transPanel) {            
            const CBioseq * product = GetCDSProductSeq(&seq_feat, *m_Scope);
            if (product) {
                // adjust existing product
                CBioseq_Handle bh = m_Scope->GetBioseqHandle(*product);

                edited_product.Reset(new CBioseq());
                edited_product->Assign(*product);
                transPanel->SetProtein(*edited_product);

                CIRef<IEditCommand> chgInst(new CCmdChangeBioseqInst(bh, edited_product->GetInst()));
                composite->AddCommand(*chgInst);                
            } else {
                // create new product sequence
                CRef<CBioseq> new_product(new CBioseq());
                transPanel->SetProtein(*new_product);
                if (new_product->IsSetInst() && new_product->GetInst().IsSetRepr()
                    && new_product->GetInst().GetRepr() != CSeq_inst::eRepr_not_set) {
                    if (!new_product->IsSetId()) {
                        CBioseq_Handle bsh = m_Scope->GetBioseqHandle(edited_seq_feat.GetLocation());
                        CSeq_entry_Handle seh = bsh.GetTopLevelEntry();
                        CRef<CSeq_id> product_id = GetNewProteinId(seh, bsh);
                        new_product->SetId().push_back(product_id);
                        edited_seq_feat.SetProduct().SetWhole().Assign(*product_id);
                    }

                    CRef<CSeq_entry> prot_entry (new CSeq_entry());
                    prot_entry->SetSeq(*new_product);
                    bool partial5 = edited_seq_feat.GetLocation().IsPartialStart(eExtreme_Biological);
                    bool partial3 = edited_seq_feat.GetLocation().IsPartialStop(eExtreme_Biological);
                    SetMolinfoForProtein (prot_entry, partial5, partial3);
                    CRef<CSeq_feat> prot = AddEmptyProteinFeatureToProtein(prot_entry, partial5, partial3);
                    if (edited_seq_feat.IsSetXref()) {
                        NON_CONST_ITERATE (CSeq_feat::TXref, it, edited_seq_feat.SetXref()) {
                            if ((*it)->IsSetData() && (*it)->GetData().IsProt()) {
                                prot->SetData().SetProt().Assign((*it)->GetData().GetProt());
                                edited_seq_feat.SetXref().erase(it);
                                break;
                            }
                        }
                        if (edited_seq_feat.GetXref().size() == 0) {
                            edited_seq_feat.ResetXref();
                        }
                    }
                    

                    CRef<objects::CCmdAddSeqEntry> subcmd(new objects::CCmdAddSeqEntry(prot_entry, 
                                                           m_Scope->GetBioseqHandle(edited_seq_feat.GetLocation()).GetParentEntry()));
                    composite->AddCommand(*subcmd);
                }                
            }
            transPanel->SetProduct(edited_seq_feat);
        }

        const CSeq_feat* orig_protein = 0;
        if (m_EditedProtein && edited_product) {
            /* automatic adjustment of protein feature */
            CSeq_feat& edited_protein = dynamic_cast<CSeq_feat&>(*m_EditedProtein);
            edited_protein.SetLocation().SetInt().SetId().Assign(*(edited_product->GetId().front()));
            edited_protein.SetLocation().SetInt().SetFrom(0);
            edited_protein.SetLocation().SetInt().SetTo(edited_product->GetLength() - 1);
            edited_protein.SetLocation().SetPartialStart(edited_seq_feat.GetLocation().IsPartialStart(eExtreme_Biological), eExtreme_Biological);
            edited_protein.SetLocation().SetPartialStop(edited_seq_feat.GetLocation().IsPartialStop(eExtreme_Biological), eExtreme_Biological);
            orig_protein = GetFeatureProtein(&seq_feat, *m_Scope);

            CSeq_feat_Handle fh = m_Scope->GetSeq_featHandle(*orig_protein);
            CIRef<IEditCommand> chgFeat(new CCmdChangeSeq_feat(fh, edited_protein));
            composite->AddCommand(*chgFeat);
        }

        if (edited_seq_feat.GetData().IsCdregion()) {
            CCDSExceptionPanel *transl_except_panel = 0;
            for (size_t i = 0; i < m_Window->GetPageCount(); ++i) {
                transl_except_panel = dynamic_cast<CCDSExceptionPanel*>(m_Window->GetPage(i));
                if (transl_except_panel)
                    break;
            }
            if (transl_except_panel) {
                transl_except_panel->SetExceptions(edited_seq_feat.SetData().SetCdregion());
            }
        }
        

        if (m_CreateMode) {
            CIRef<IEditCommand> addFeat(new CCmdCreateFeat(m_SEH, edited_seq_feat));
            composite->AddCommand(*addFeat);
        } else {
            CSeq_feat_Handle fh = m_Scope->GetSeq_featHandle(seq_feat);
            CIRef<IEditCommand> chgFeat(new CCmdChangeSeq_feat(fh, edited_seq_feat));
            composite->AddCommand(*chgFeat);
        }


        // set data for publication feature
        CPublicationTypePanel* pubPanel = 0;
        for (size_t i = 0; i < m_Window->GetPageCount(); ++i) {
            pubPanel = dynamic_cast<CPublicationTypePanel*>(m_Window->GetPage(i));
            if (pubPanel)
                break;
        }

        if (pubPanel) {
            // copy to data
            CRef<objects::CPubdesc> pub = pubPanel->GetPubdesc();
            edited_seq_feat.SetData().SetPub(*pub);
        }

        // import feature
        CImportFeaturePanel* impPanel = 0;
        for (size_t i = 0; i < m_Window->GetPageCount(); ++i) {
            impPanel = dynamic_cast<CImportFeaturePanel*>(m_Window->GetPage(i));
            if (impPanel)
                break;
        }

        if (impPanel) {
            impPanel->PopulateImpFeat (edited_seq_feat);
        }

        //biosource
        bool is_biosrc = false;
		    CBioSourcePanel* biosource_panel = 0;
		    for (size_t i = 0; i < m_Window->GetPageCount(); ++i) {
            biosource_panel = dynamic_cast<CBioSourcePanel*>(m_Window->GetPage(i));
            if (biosource_panel)
                break;
        }
        if (biosource_panel) {
			      biosource_panel->PopulateBioSource(edited_seq_feat.SetData().SetBiosrc());
            is_biosrc = true;
        }

		    CLocAndGCodePanel* loc_gc_panel = 0;
		    for (size_t i = 0; i < m_Window->GetPageCount(); ++i) {
            loc_gc_panel = dynamic_cast<CLocAndGCodePanel*>(m_Window->GetPage(i));
            if (loc_gc_panel)
                break;
        }
        if (loc_gc_panel) {
			      loc_gc_panel->PopulateLocAndGC(edited_seq_feat.SetData().SetBiosrc());
        }

		    CSrcModListPanel* src_mod_panel = 0;
		    for (size_t i = 0; i < m_Window->GetPageCount(); ++i) {
            src_mod_panel = dynamic_cast<CSrcModListPanel*>(m_Window->GetPage(i));
            if (src_mod_panel)
                break;
        }
		    if (src_mod_panel) {
			      src_mod_panel->PopulateModifiers(edited_seq_feat.SetData().SetBiosrc());
		    }

        // GenBank quals
        CGBQualPanel* qualPanel = 0;
        for (size_t i = 0; i < m_Window->GetPageCount(); ++i) {
            qualPanel = dynamic_cast<CGBQualPanel*>(m_Window->GetPage(i));
            if (qualPanel)
                break;
        }
        if (qualPanel) {
            qualPanel->PopulateGBQuals(edited_seq_feat);
        }

        // Cross-refs
        CDbxrefPanel* xrefPanel = 0;
        bool added_biosrc_xrefs = false;
        for (size_t i = 0; i < m_Window->GetPageCount(); ++i) {
            xrefPanel = dynamic_cast<CDbxrefPanel*>(m_Window->GetPage(i));
            if (xrefPanel) {
                if (is_biosrc && !added_biosrc_xrefs) {
                    xrefPanel->PopulateDbxrefs(edited_seq_feat.SetData().SetBiosrc().SetOrg());
                    added_biosrc_xrefs = true;
                } else {
                    xrefPanel->PopulateDbxrefs(edited_seq_feat);
                    break;
                }
            }
        }

        // identifiers
        CIdentifiersPanel* idPanel = 0;
        for (size_t i = 0; i < m_Window->GetPageCount(); ++i) {
            idPanel = dynamic_cast<CIdentifiersPanel*>(m_Window->GetPage(i));
            if (idPanel)
                break;
        }
        if (idPanel) {
            idPanel->PopulateIdentifiers(edited_seq_feat);
        }

        // inferences
        CInferenceListPanel* inferencePanel = 0;
        for (size_t i = 0; i < m_Window->GetPageCount(); ++i) {
            inferencePanel = dynamic_cast<CInferenceListPanel*>(m_Window->GetPage(i));
            if (inferencePanel)
                break;
        }
        if (inferencePanel) {
            inferencePanel->PopulateGBQuals(edited_seq_feat);
        }

        // inferences
        CCitationsPanel* citationsPanel = 0;
        for (size_t i = 0; i < m_Window->GetPageCount(); ++i) {
            citationsPanel = dynamic_cast<CCitationsPanel*>(m_Window->GetPage(i));
            if (citationsPanel)
                break;
        }
        if (citationsPanel) {
            citationsPanel->SetCitations(edited_seq_feat);
        }

        CRegionPanel* regionPanel = 0;
        for (size_t i = 0; i < m_Window->GetPageCount(); ++i) {
            regionPanel = dynamic_cast<CRegionPanel*>(m_Window->GetPage(i));
            if (regionPanel)
                break;
        }

        if (regionPanel) {
            // copy to data
            string region = regionPanel->GetRegion();
            edited_seq_feat.SetData().SetRegion(region);
        }

        CBondPanel* bondPanel = 0;
        for (size_t i = 0; i < m_Window->GetPageCount(); ++i) {
            bondPanel = dynamic_cast<CBondPanel*>(m_Window->GetPage(i));
            if (bondPanel)
                break;
        }

        if (bondPanel) {
            // copy to data
            CSeqFeatData::EBond bond = bondPanel->GetBond();
            edited_seq_feat.SetData().SetBond(bond);
            }
        
        CSitePanel* sitePanel = 0;
        for (size_t i = 0; i < m_Window->GetPageCount(); ++i) {
            sitePanel = dynamic_cast<CSitePanel*>(m_Window->GetPage(i));
            if (sitePanel)
                break;
        }
        
        if (sitePanel) {
            // copy to data
            CSeqFeatData::ESite site = sitePanel->GetSite();
            edited_seq_feat.SetData().SetSite(site);
            }
        
		// some generic cleanup
        if (edited_seq_feat.IsSetQual() && edited_seq_feat.GetQual().empty()) {
            edited_seq_feat.ResetQual();
        }

        // find original overlapping mRNA, if editing a coding region
        if (edited_seq_feat.IsSetData() && edited_seq_feat.GetData().IsCdregion()
            && edited_seq_feat.IsSetLocation()) {
            bool update_mrna_span = false;
            bool update_mrna_product = false;

            CCDSProteinPanel* cdsProteinPanel = 0;
            CCDSTranslationPanel* cdsProductPanel = 0;
            for (size_t i = 0; i < m_Window->GetPageCount(); ++i) {
                if (!cdsProteinPanel) {
                    cdsProteinPanel = dynamic_cast<CCDSProteinPanel*>(m_Window->GetPage(i));
                }
                if (!cdsProductPanel) {
                    cdsProductPanel = dynamic_cast<CCDSTranslationPanel*>(m_Window->GetPage(i));
                }

                if (cdsProteinPanel && cdsProductPanel) {
                    break;
                }
            }
            if (cdsProteinPanel) {
                update_mrna_product = cdsProteinPanel->ShouldUpdatemRNAProduct();
            }
            if (cdsProductPanel) {
                update_mrna_span = cdsProductPanel->ShouldUpdatemRNALocation();
            }

            if (update_mrna_product || update_mrna_span) {
                // note - if the edited feature refers to an mRNA by feature ID, use that feature
                bool has_xref = false;
                CConstRef<CSeq_feat> mrna;
                if (edited_seq_feat.IsSetXref()) {
                    CBioseq_Handle bsh = m_Scope->GetBioseqHandle(*(edited_seq_feat.GetLocation().GetId()));
                    CTSE_Handle tse = bsh.GetTSE_Handle();
                    FOR_EACH_SEQFEATXREF_ON_SEQFEAT (it, edited_seq_feat) {
                        if ((*it)->IsSetId() && (*it)->GetId().IsLocal() && (*it)->GetId().GetLocal().IsId()) {
                            CSeq_feat_Handle mrna_h = tse.GetFeatureWithId(CSeqFeatData::eSubtype_mRNA, (*it)->GetId().GetLocal().GetId());
                            if (mrna_h) {
                                mrna = mrna_h.GetSeq_feat();
                            }
                            has_xref = true;
                        }
                    }
                }

                if (!has_xref) {
                    mrna = GetBestOverlappingFeat(
                    seq_feat.GetLocation(), // note - using original location to find mRNA
                    CSeqFeatData::eSubtype_mRNA,
                    sequence::eOverlap_CheckIntRev,
                    *m_Scope);
                }
                if ( mrna ) {
                    CRef<CSeq_feat> new_mrna(new CSeq_feat());
                    new_mrna->Assign(*mrna);
                    bool any_change = false;

                    if (update_mrna_product) {
                        // update product
                        string new_prot_product = "";
                        if (m_EditedProtein) {
                            CSeq_feat& edited_protein = dynamic_cast<CSeq_feat&>(*m_EditedProtein);
                            if ( edited_protein.GetData().GetProt().IsSetName()
                                 && edited_protein.GetData().GetProt().GetName().size() > 0) {
                                new_prot_product = edited_protein.GetData().GetProt().GetName().front();
                            }
                        }              
                        string orig_mrna_product = "";
                        if (mrna->GetData().GetRna().IsSetExt() && mrna->GetData().GetRna().GetExt().IsName()) {
                            orig_mrna_product = mrna->GetData().GetRna().GetExt().GetName();
                        }
                        if (!NStr::Equal(new_prot_product, orig_mrna_product) 
                            || !mrna->GetData().GetRna().IsSetExt()
                            || !mrna->GetData().GetRna().GetExt().IsName()) {
                            new_mrna->SetData().SetRna().SetExt().SetName(new_prot_product);
                            any_change = true;
                        }
                    }

                    if (update_mrna_span) {
                        // adjust location
                        sequence::ECompare orig_comp = sequence::Compare(seq_feat.GetLocation(), mrna->GetLocation(), m_Scope);
                        sequence::ECompare new_comp = sequence::Compare(edited_seq_feat.GetLocation(), new_mrna->GetLocation(), m_Scope);
                        if (orig_comp == sequence::eSame || new_comp != sequence::eContained) {
                            new_mrna->SetLocation().Assign(edited_seq_feat.GetLocation());
                            bool make_partial = s_ShouldSetFeaturePartial(new_mrna->GetLocation(), m_Scope);
                            if (make_partial) {
                                new_mrna->SetPartial(true);
                            } else {
                                new_mrna->ResetPartial();
                            }
                            any_change = true;
                        }
                    }

                    if (any_change) {
                        CSeq_feat_Handle fh = m_Scope->GetSeq_featHandle(*mrna);
                        CIRef<IEditCommand> chgFeat(new CCmdChangeSeq_feat(fh, *new_mrna));
                        composite->AddCommand(*chgFeat);
                    }
                }
            }
            
        }

        CFeatureGenePanel* addedGenePanel = 0;
        CRef<CSeq_feat> new_gene;
        CConstRef<CSeq_feat> old_gene;
        for (size_t i = 0; i < m_Window->GetPageCount(); ++i) {
            addedGenePanel = dynamic_cast<CFeatureGenePanel*>(m_Window->GetPage(i));
            if (addedGenePanel) {
                new_gene = addedGenePanel->GetNewGene();
                old_gene = addedGenePanel->GetGeneToEdit();
            }
        }
 
        try {
            if (!m_CreateMode) {
                if (!new_gene && old_gene) {
                    // may need to adjust location of overlapping gene
                    sequence::ECompare before_comp = sequence::Compare(old_gene->GetLocation(), seq_feat.GetLocation(), m_Scope);
                    sequence::ECompare after_comp = sequence::Compare(old_gene->GetLocation(), edited_seq_feat.GetLocation(), m_Scope);
                    if ((before_comp == sequence::eSame && after_comp != sequence::eSame)
                        || (after_comp != sequence::eSame && after_comp != sequence::eContains)) {
                        CRef<CSeq_feat> adjusted_gene(new CSeq_feat());
                        adjusted_gene->Assign(*old_gene);
                        adjusted_gene->SetLocation().SetInt().SetFrom(edited_seq_feat.GetLocation().GetStart(eExtreme_Positional));
                        adjusted_gene->SetLocation().SetInt().SetTo(edited_seq_feat.GetLocation().GetStop(eExtreme_Positional));
                        if (sequence::Compare(old_gene->GetLocation(), adjusted_gene->GetLocation(), m_Scope) != sequence::eSame) {
                            CSeq_feat_Handle fh = m_Scope->GetSeq_featHandle(*old_gene);
                            CIRef<IEditCommand> chgFeat(new CCmdChangeSeq_feat(fh, *adjusted_gene));
                            composite->AddCommand(*chgFeat);
                        }
                    }
                }

                if (new_gene) {
                    CIRef<IEditCommand> addFeat(new CCmdCreateFeat(m_SEH, *new_gene));
                    composite->AddCommand(*addFeat);
                }
                
            }
            m_EditAction.Reset(composite);
        }
        catch (CException& e) {
            LOG_POST(Error << "CEditObjectSeq_feat::GetEditAction(): " << e.GetMsg());
        }

#if 0
#ifdef _DEBUG
    CNcbiOstrstream ostr;
    {{
        auto_ptr<CObjectOStream>
            out(CObjectOStream::Open(eSerial_AsnText, ostr));
        *out << edited_seq_feat;
    }}
    _TRACE(string(CNcbiOstrstreamToString(ostr)));
	//LOG_POST(Info << string(CNcbiOstrstreamToString(ostr)));
#endif
#endif
    }

    return m_EditAction.GetPointer();
}

END_NCBI_SCOPE
