/*  $Id: journal_panel.cpp 26686 2012-10-24 18:22:06Z filippov $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data,  the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties,  express or implied,  including
 *  warranties of performance,  merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Roman Katargin
 */


#include <ncbi_pch.hpp>
#include <gui/objutils/undo_manager_interface.hpp>
#include <gui/widgets/edit/edit_object.hpp>

#include <objects/biblio/Cit_jour.hpp>
#include <objects/biblio/Title.hpp>
#include <objects/biblio/Imprint.hpp>
#include <objects/biblio/CitRetract.hpp>
#include <objects/general/Date.hpp>
#include <objects/general/Date_std.hpp>
#include <objects/seq/Pubdesc.hpp>
#include <objects/pub/Pub.hpp>
#include <objects/pub/Pub_equiv.hpp>
#include <objects/pub/Pub_set.hpp>
#include <objects/pubmed/Pubmed_entry.hpp>
#include <objects/medline/Medline_entry.hpp>
#include <objects/biblio/Cit_art.hpp>

#include <gui/widgets/edit/serial_member_primitive_validators.hpp>
#include <gui/widgets/wx/number_validator.hpp>
#include <gui/widgets/wx/wx_utils.hpp>

////@begin includes
////@end includes
////@begin includes
////@end includes

#include "journal_panel.hpp"


#include <wx/sizer.h>
#include <wx/bitmap.h>
#include <wx/icon.h>
#include <wx/stattext.h>
#include <wx/textctrl.h>
#include <wx/choice.h>
#include <wx/button.h>

#include <objtools/eutils/api/efetch.hpp>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

////@begin XPM images
////@end XPM images


/*
 * CJournalPanel type definition
 */

IMPLEMENT_DYNAMIC_CLASS( CJournalPanel, wxPanel )


/*
 * CJournalPanel event table definition
 */

BEGIN_EVENT_TABLE( CJournalPanel, wxPanel )

////@begin CJournalPanel event table entries
    EVT_CHOICE( ID_JOURNAL_ERRATUM, CJournalPanel::OnJournalErratumSelected )

    EVT_BUTTON( ID_JOURNAL_LOOKUP_RELAXED, CJournalPanel::OnJournalLookupRelaxedClick )

    EVT_BUTTON( ID_JOURNAL_LOOKUP_ISOJTA, CJournalPanel::OnJournalLookupIsojtaClick )

    EVT_BUTTON( ID_PMID_LOOKUP, CJournalPanel::OnPmidLookupClick )

////@end CJournalPanel event table entries

END_EVENT_TABLE()


/*
 * CJournalPanel constructors
 */

 CJournalPanel::CJournalPanel() : m_Object(0), m_pPmid(0), m_pMuid(0)
{
    Init();
}

CJournalPanel::CJournalPanel( wxWindow* parent, CSerialObject& object, long * pmid, long * muid,
         wxWindowID id, const wxPoint& pos, const wxSize& size, long style ) :
     m_Object(0), m_pPmid(pmid), m_pMuid(muid)
{
    Init();
    m_Object = dynamic_cast<CCit_jour*>(&object);

    const CCit_jour& journal = dynamic_cast<const CCit_jour&>(*m_Object);
    m_EditedJournal.Reset((CSerialObject*)CCit_jour::GetTypeInfo()->Create());
    m_EditedJournal->Assign(journal);

    Create(parent, id, pos, size, style);
}


/*
 * CJournalPanel creator
 */

bool CJournalPanel::Create( wxWindow* parent, wxWindowID id, const wxPoint& pos, const wxSize& size, long style )
{
////@begin CJournalPanel creation
    SetExtraStyle(wxWS_EX_BLOCK_EVENTS);
    wxPanel::Create( parent, id, pos, size, style );

    CreateControls();
    if (GetSizer())
    {
        GetSizer()->SetSizeHints(this);
    }
    Centre();
////@end CJournalPanel creation
    return true;
}


/*
 * CJournalPanel destructor
 */

CJournalPanel::~CJournalPanel()
{
////@begin CJournalPanel destruction
////@end CJournalPanel destruction
}


/*
 * Member initialisation
 */

void CJournalPanel::Init()
{
////@begin CJournalPanel member initialisation
    m_JournalTitle = NULL;
    m_Volume = NULL;
    m_Issue = NULL;
    m_Pages = NULL;
    m_PmidCtrl = NULL;
    m_MuidCtrl = NULL;
    m_ErratumCtrl = NULL;
    m_ErratumExplanation = NULL;
    m_PubstatusCtrl = NULL;
    m_PubDateSizer = NULL;
    m_PubDatePanel = NULL;
////@end CJournalPanel member initialisation
}


/*
 * Control creation for CJournalPanel
 */

void CJournalPanel::CreateControls()
{    
////@begin CJournalPanel content construction
    CJournalPanel* itemPanel1 = this;

    wxBoxSizer* itemBoxSizer2 = new wxBoxSizer(wxVERTICAL);
    itemPanel1->SetSizer(itemBoxSizer2);

    wxFlexGridSizer* itemFlexGridSizer3 = new wxFlexGridSizer(0, 2, 0, 0);
    itemBoxSizer2->Add(itemFlexGridSizer3, 0, wxALIGN_CENTER_HORIZONTAL|wxALL, 5);

    wxStaticText* itemStaticText4 = new wxStaticText( itemPanel1, wxID_STATIC, _("Journal"), wxDefaultPosition, wxSize(85, -1), wxALIGN_RIGHT );
    itemFlexGridSizer3->Add(itemStaticText4, 0, wxALIGN_RIGHT|wxALIGN_CENTER_VERTICAL|wxALL, 5);

    m_JournalTitle = new wxTextCtrl( itemPanel1, ID_JOURNAL_TITLE, wxEmptyString, wxDefaultPosition, wxDefaultSize, 0 );
    itemFlexGridSizer3->Add(m_JournalTitle, 1, wxGROW|wxALIGN_CENTER_VERTICAL|wxALL, 5);

    wxStaticText* itemStaticText6 = new wxStaticText( itemPanel1, wxID_STATIC, _("Volume"), wxDefaultPosition, wxSize(85, -1), wxALIGN_RIGHT );
    itemFlexGridSizer3->Add(itemStaticText6, 0, wxALIGN_LEFT|wxALIGN_CENTER_VERTICAL|wxALL, 5);

    m_Volume = new wxTextCtrl( itemPanel1, ID_JOURNAL_VOLUME, wxEmptyString, wxDefaultPosition, wxDefaultSize, 0 );
    itemFlexGridSizer3->Add(m_Volume, 0, wxALIGN_LEFT|wxALIGN_CENTER_VERTICAL|wxALL, 5);

    wxStaticText* itemStaticText8 = new wxStaticText( itemPanel1, wxID_STATIC, _("Issue"), wxDefaultPosition, wxDefaultSize, wxALIGN_RIGHT );
    itemFlexGridSizer3->Add(itemStaticText8, 0, wxALIGN_RIGHT|wxALIGN_CENTER_VERTICAL|wxALL, 5);

    m_Issue = new wxTextCtrl( itemPanel1, ID_JOURNAL_ISSUE, wxEmptyString, wxDefaultPosition, wxDefaultSize, 0 );
    itemFlexGridSizer3->Add(m_Issue, 0, wxALIGN_LEFT|wxALIGN_CENTER_VERTICAL|wxALL, 5);

    wxStaticText* itemStaticText10 = new wxStaticText( itemPanel1, wxID_STATIC, _("Pages"), wxDefaultPosition, wxDefaultSize, 0 );
    itemFlexGridSizer3->Add(itemStaticText10, 0, wxALIGN_RIGHT|wxALIGN_CENTER_VERTICAL|wxALL, 5);

    m_Pages = new wxTextCtrl( itemPanel1, ID_JOURNAL_PAGES, wxEmptyString, wxDefaultPosition, wxDefaultSize, 0 );
    itemFlexGridSizer3->Add(m_Pages, 0, wxALIGN_LEFT|wxALIGN_CENTER_VERTICAL|wxALL, 5);

    wxStaticText* itemStaticText12 = new wxStaticText( itemPanel1, wxID_STATIC, _("PubMedId"), wxDefaultPosition, wxSize(85, -1), wxALIGN_RIGHT );
    itemFlexGridSizer3->Add(itemStaticText12, 0, wxALIGN_RIGHT|wxALIGN_CENTER_VERTICAL|wxALL, 5);

    m_PmidCtrl = new wxTextCtrl( itemPanel1, ID_JOURNAL_PMID, wxEmptyString, wxDefaultPosition, wxDefaultSize, 0 );
    itemFlexGridSizer3->Add(m_PmidCtrl, 0, wxALIGN_LEFT|wxALIGN_CENTER_VERTICAL|wxALL, 5);

    wxStaticText* itemStaticText14 = new wxStaticText( itemPanel1, wxID_STATIC, _("muid"), wxDefaultPosition, wxDefaultSize, wxALIGN_RIGHT );
    itemFlexGridSizer3->Add(itemStaticText14, 0, wxALIGN_RIGHT|wxALIGN_CENTER_VERTICAL|wxALL, 5);

    m_MuidCtrl = new wxTextCtrl( itemPanel1, ID_JOURNAL_MUID, wxEmptyString, wxDefaultPosition, wxDefaultSize, 0 );
    itemFlexGridSizer3->Add(m_MuidCtrl, 0, wxALIGN_LEFT|wxALIGN_CENTER_VERTICAL|wxALL, 5);

    wxStaticText* itemStaticText16 = new wxStaticText( itemPanel1, wxID_STATIC, _("Erratum"), wxDefaultPosition, wxSize(85, -1), wxALIGN_RIGHT );
    itemFlexGridSizer3->Add(itemStaticText16, 0, wxALIGN_RIGHT|wxALIGN_CENTER_VERTICAL|wxALL, 5);

    wxArrayString m_ErratumCtrlStrings;
    m_ErratumCtrlStrings.Add(wxEmptyString);
    m_ErratumCtrlStrings.Add(_("Retracted"));
    m_ErratumCtrlStrings.Add(_("Notice"));
    m_ErratumCtrlStrings.Add(_("In-error"));
    m_ErratumCtrlStrings.Add(_("Erratum"));
    m_ErratumCtrl = new wxChoice( itemPanel1, ID_JOURNAL_ERRATUM, wxDefaultPosition, wxDefaultSize, m_ErratumCtrlStrings, 0 );
    itemFlexGridSizer3->Add(m_ErratumCtrl, 0, wxALIGN_LEFT|wxALIGN_CENTER_VERTICAL|wxALL, 5);

    wxStaticText* itemStaticText18 = new wxStaticText( itemPanel1, wxID_STATIC, _("Explanation"), wxDefaultPosition, wxDefaultSize, wxALIGN_RIGHT );
    itemFlexGridSizer3->Add(itemStaticText18, 0, wxALIGN_RIGHT|wxALIGN_CENTER_VERTICAL|wxALL, 5);

    m_ErratumExplanation = new wxTextCtrl( itemPanel1, ID_JOURNAL_EXPLANATION, wxEmptyString, wxDefaultPosition, wxSize(200, -1), 0 );
    itemFlexGridSizer3->Add(m_ErratumExplanation, 0, wxALIGN_LEFT|wxGROW|wxALL, 5);

    wxStaticText* itemStaticText20 = new wxStaticText( itemPanel1, wxID_STATIC, _("Publication Status"), wxDefaultPosition, wxSize(85, -1), wxALIGN_RIGHT );
    itemFlexGridSizer3->Add(itemStaticText20, 0, wxALIGN_RIGHT|wxALIGN_CENTER_VERTICAL|wxALL, 5);

    wxArrayString m_PubstatusCtrlStrings;
    m_PubstatusCtrlStrings.Add(_("Received"));
    m_PubstatusCtrlStrings.Add(_("Accepted"));
    m_PubstatusCtrlStrings.Add(_("Electronic Publication"));
    m_PubstatusCtrlStrings.Add(_("Print Publication"));
    m_PubstatusCtrlStrings.Add(_("Revised"));
    m_PubstatusCtrlStrings.Add(_("PMC Publication"));
    m_PubstatusCtrlStrings.Add(_("PMC Revision"));
    m_PubstatusCtrlStrings.Add(_("Entrez Date"));
    m_PubstatusCtrlStrings.Add(_("PubMed Revision"));
    m_PubstatusCtrlStrings.Add(_("Ahead of Print"));
    m_PubstatusCtrlStrings.Add(_("Pre-Medline (Obsolete)"));
    m_PubstatusCtrlStrings.Add(_("MeSH Date"));
    m_PubstatusCtrlStrings.Add(_("Other"));
    m_PubstatusCtrl = new wxChoice( itemPanel1, ID_JOURNAL_PUBSTATUS, wxDefaultPosition, wxDefaultSize, m_PubstatusCtrlStrings, 0 );
    itemFlexGridSizer3->Add(m_PubstatusCtrl, 0, wxALIGN_LEFT|wxALIGN_CENTER_VERTICAL|wxALL, 5);

    wxBoxSizer* itemBoxSizer22 = new wxBoxSizer(wxHORIZONTAL);
    itemBoxSizer2->Add(itemBoxSizer22, 0, wxALIGN_CENTER_HORIZONTAL|wxALL, 5);

    wxStaticText* itemStaticText23 = new wxStaticText( itemPanel1, wxID_STATIC, _("Publication Date"), wxDefaultPosition, wxSize(85, -1), wxALIGN_RIGHT );
    itemBoxSizer22->Add(itemStaticText23, 0, wxALIGN_CENTER_VERTICAL|wxALL, 5);

    m_PubDateSizer = new wxBoxSizer(wxHORIZONTAL);
    itemBoxSizer22->Add(m_PubDateSizer, 0, wxALIGN_CENTER_VERTICAL|wxALL, 5);

    wxBoxSizer* itemBoxSizer25 = new wxBoxSizer(wxHORIZONTAL);
    itemBoxSizer2->Add(itemBoxSizer25, 0, wxALIGN_CENTER_HORIZONTAL|wxALL, 5);

    wxButton* itemButton26 = new wxButton( itemPanel1, ID_JOURNAL_LOOKUP_RELAXED, _("Lookup Relaxed"), wxDefaultPosition, wxDefaultSize, 0 );
    itemBoxSizer25->Add(itemButton26, 0, wxALIGN_CENTER_VERTICAL|wxALL, 5);

    wxButton* itemButton27 = new wxButton( itemPanel1, ID_JOURNAL_LOOKUP_ISOJTA, _("Lookup ISOJTA"), wxDefaultPosition, wxDefaultSize, 0 );
    itemBoxSizer25->Add(itemButton27, 0, wxALIGN_CENTER_VERTICAL|wxALL, 5);

    wxButton* itemButton28 = new wxButton( itemPanel1, ID_PMID_LOOKUP, _("Lookup by PubMedId"), wxDefaultPosition, wxDefaultSize, 0 );
    itemBoxSizer25->Add(itemButton28, 0, wxALIGN_CENTER_VERTICAL|wxALL, 5);

////@end CJournalPanel content construction

    CCit_jour& journal = dynamic_cast<CCit_jour&>(*m_EditedJournal);

    // add control for publication date
   
    CRef<objects::CDate> p_date(&(journal.SetImp().SetDate()));
//    wxWindow* pubdate = new CFlexibleDatePanel(itemPanel1, p_date);
    m_PubDatePanel = new CFlexibleDatePanel(itemPanel1, p_date);
    m_PubDateSizer->Add ( m_PubDatePanel );

    CSerialObject& title_object = journal.SetTitle();
    m_JournalTitle->SetValidator( CSerialTitleValidator(title_object, CTitle::C_E::e_Iso_jta) );

    CSerialObject& imp_object = journal.SetImp();
    m_Volume->SetValidator( CSerialTextValidator(imp_object, "volume") );
    m_Issue->SetValidator( CSerialTextValidator(imp_object, "issue") );
    m_Pages->SetValidator( CSerialTextValidator(imp_object, "pages") );
    m_PmidCtrl->SetValidator( CNumberOrBlankValidator(&m_Pmid));
    m_MuidCtrl->SetValidator( CNumberOrBlankValidator(&m_Muid));
    
}


bool CJournalPanel::TransferDataToWindow()
{
    if (m_pPmid != 0) {
        m_Pmid = *m_pPmid;
    }
    if (m_pMuid != 0) {
        m_Muid = *m_pMuid;
    }

    


    if (!wxPanel::TransferDataToWindow())
        return false;

    CCit_jour& journal = dynamic_cast<CCit_jour&>(*m_EditedJournal);
    if (journal.IsSetImp() && journal.GetImp().IsSetPubstatus()) {
        int pubstatus = journal.GetImp().GetPubstatus();
        switch (pubstatus) {
        case 1:
            case 2:
            case 3:
            case 4:
            case 5:
            case 6:
            case 7:
            case 8:
            case 9:
            case 10:
            case 11:
            case 12:
                m_PubstatusCtrl->SetSelection (pubstatus - 1);
                break;
            case 255:
                m_PubstatusCtrl->SetSelection (13);
                break;
            default:
                m_PubstatusCtrl->SetSelection (14);
                break;
        }
    } else {
        m_PubstatusCtrl->SetSelection (14);
    }

    if (journal.IsSetImp() && journal.GetImp().IsSetRetract() && journal.GetImp().GetRetract().IsSetType()) {
        int erratum_type = journal.GetImp().GetRetract().GetType();
        switch (erratum_type) {
            case 1:
            case 2:
            case 3:
            case 4:
                m_ErratumCtrl->SetSelection (erratum_type);
                m_ErratumExplanation->Enable(true);
                if (journal.GetImp().GetRetract().IsSetExp()) {
                    m_ErratumExplanation->SetValue (ToWxString(journal.GetImp().GetRetract().GetExp()));
                } else {
                    m_ErratumExplanation->SetValue(wxEmptyString);
                }
                break;
            default:
                m_ErratumCtrl->SetSelection(0);
                m_ErratumExplanation->SetValue(wxEmptyString);
                m_ErratumExplanation->Enable(false);
                break;
        }
    } else {
        m_ErratumCtrl->SetSelection(0);
        m_ErratumExplanation->SetValue(wxEmptyString);
        m_ErratumExplanation->Enable(false);
    }
    return true;
}


bool CJournalPanel::TransferDataFromWindow()
{
    if (!wxPanel::TransferDataFromWindow())
        return false;
    if (m_pPmid != 0) {
        *m_pPmid = m_Pmid;
    }
    if (m_pMuid != 0) {
        *m_pMuid = m_Muid;
    }
    CCit_jour& journal = dynamic_cast<CCit_jour&>(*m_EditedJournal);
    int pubstatus = m_PubstatusCtrl->GetCurrentSelection();
    if (pubstatus > -1 && pubstatus < 12) {
        journal.SetImp().SetPubstatus(pubstatus + 1);
    } else if (pubstatus == 12) {
        journal.SetImp().SetPubstatus(255);
    } else {
        journal.SetImp().ResetPubstatus();
    } 

    int erratum = m_ErratumCtrl->GetCurrentSelection();
    if (erratum > 0 && erratum < 5) {
        journal.SetImp().SetRetract().SetType((objects::CCitRetract_Base::TType)(erratum));
        string exp = ToStdString (m_ErratumExplanation->GetValue());
        if (!NStr::IsBlank(exp)) {
            journal.SetImp().SetRetract().SetExp(exp);
        } else {
            journal.SetImp().SetRetract().ResetExp();
        }
    } else {
        journal.SetImp().ResetRetract();
    }
    return true;
}


CRef<CCit_jour> CJournalPanel::GetCit_jour() const
{
    const CCit_jour& journal = dynamic_cast<const CCit_jour&>(*m_EditedJournal);

    CRef<CCit_jour> ret(new CCit_jour());
    ret->Assign(journal);
    return ret;
}


/*
 * Should we show tooltips?
 */

bool CJournalPanel::ShowToolTips()
{
    return true;
}

/*
 * Get bitmap resources
 */

wxBitmap CJournalPanel::GetBitmapResource( const wxString& name )
{
    // Bitmap retrieval
////@begin CJournalPanel bitmap retrieval
    wxUnusedVar(name);
    return wxNullBitmap;
////@end CJournalPanel bitmap retrieval
}

/*
 * Get icon resources
 */

wxIcon CJournalPanel::GetIconResource( const wxString& name )
{
    // Icon retrieval
////@begin CJournalPanel icon retrieval
    wxUnusedVar(name);
    return wxNullIcon;
////@end CJournalPanel icon retrieval
}


/*
 * wxEVT_COMMAND_CHOICE_SELECTED event handler for ID_JOURNAL_ERRATUM
 */

void CJournalPanel::OnJournalErratumSelected( wxCommandEvent& event )
{
    int erratum = m_ErratumCtrl->GetSelection();

    if (erratum > -1 && erratum < 4) {
        m_ErratumExplanation->Enable(true);
    } else {
        m_ErratumExplanation->Enable(false);
    }
}


/*
 * wxEVT_COMMAND_BUTTON_CLICKED event handler for ID_JOURNAL_LOOKUP_RELAXED
 */

void CJournalPanel::OnJournalLookupRelaxedClick( wxCommandEvent& event )
{
////@begin wxEVT_COMMAND_BUTTON_CLICKED event handler for ID_JOURNAL_LOOKUP_RELAXED in CJournalPanel.
    // Before editing this code, remove the block markers.
    event.Skip();
////@end wxEVT_COMMAND_BUTTON_CLICKED event handler for ID_JOURNAL_LOOKUP_RELAXED in CJournalPanel. 
}


/*
 * wxEVT_COMMAND_BUTTON_CLICKED event handler for ID_JOURNAL_LOOKUP_ISOJTA
 */

void CJournalPanel::OnJournalLookupIsojtaClick( wxCommandEvent& event )
{
////@begin wxEVT_COMMAND_BUTTON_CLICKED event handler for ID_JOURNAL_LOOKUP_ISOJTA in CJournalPanel.
    // Before editing this code, remove the block markers.
    event.Skip();
////@end wxEVT_COMMAND_BUTTON_CLICKED event handler for ID_JOURNAL_LOOKUP_ISOJTA in CJournalPanel. 
}


/*
 * wxEVT_COMMAND_BUTTON_CLICKED event handler for ID_PMID_LOOKUP
 */

void CJournalPanel::OnPmidLookupClick( wxCommandEvent& event )
{
    wxString pmid_wx = m_PmidCtrl->GetValue();
    string pmid_str(pmid_wx.c_str());
    bool modified = false;
    if (!pmid_str.empty())
    {
        CRef<CEUtils_ConnContext> ctx(new CEUtils_ConnContext);
        CEFetch_Request req(ctx);
        req.SetRetMode(CEFetch_Request::eRetMode_asn);
        req.SetDatabase("pubmed");
        req.SetRetMax(1);
        req.GetId().SetIds(pmid_str);
        CObjectIStream* istr = req.GetObjectIStream();
        if (istr->InGoodState() && !istr->EndOfData())
        {
            string obj_type;
            try {
                    obj_type = istr->ReadFileHeader();   // If nothing returned it's crashing here
                } catch(...) {} // truly ugly but there doesn't seem to be any other way
                CRef<CSerialObject> obj;
                if (obj_type == "Pubmed-entry") 
                {
                    obj.Reset(new CPubmed_entry());
                    istr->ReadObject(obj.GetPointer(), CPubmed_entry::GetTypeInfo());
                    CPubmed_entry& pubmed = dynamic_cast<CPubmed_entry &>(*obj);
                    if (pubmed.CanGetMedent() && pubmed.GetMedent().CanGetCit() && pubmed.GetMedent().GetCit().GetFrom().IsJournal())
                    {
                        const CCit_jour& journal = pubmed.GetMedent().GetCit().GetFrom().GetJournal();
                        PushToWindow(journal);
                        modified = true;
                    }
                }
        }
    }
    if (!modified) event.Skip();
}


void CJournalPanel::PushToWindow(const CCit_jour& journal)
{
        if ( journal.CanGetTitle() && journal.GetTitle().CanGet() && !journal.GetTitle().Get().empty() && journal.GetTitle().Get().front()->IsIso_jta() )
            m_JournalTitle->SetValue( journal.GetTitle().Get().front()->GetIso_jta() );

        if ( journal.CanGetImp() )
        {
            const CImprint& imp = journal.GetImp();
            if ( imp.CanGetVolume() )
                m_Volume->SetValue( imp.GetVolume() );
            if ( imp.CanGetIssue() )
                m_Issue->SetValue( imp.GetIssue() );
            if ( imp.CanGetPages() )
                m_Pages->SetValue( imp.GetPages() );


            if (imp.IsSetPubstatus()) 
            {
                int pubstatus = imp.GetPubstatus();
                switch (pubstatus) {
                case 1:
                case 2:
                case 3:
                case 4:
                case 5:
                case 6:
                case 7:
                case 8:
                case 9:
                case 10:
                case 11:
                case 12:
                    m_PubstatusCtrl->SetSelection (pubstatus - 1);
                    break;
                case 255:
                    m_PubstatusCtrl->SetSelection (13);
                    break;
                default:
                    m_PubstatusCtrl->SetSelection (14);
                    break;
                }
            } 
            else 
                m_PubstatusCtrl->SetSelection (14);
            

            if (imp.IsSetRetract() && imp.GetRetract().IsSetType()) 
            {
                int erratum_type = imp.GetRetract().GetType();
                switch (erratum_type) {
                case 1:
                case 2:
                case 3:
                case 4:
                    m_ErratumCtrl->SetSelection (erratum_type);
                    m_ErratumExplanation->Enable(true);
                    if (imp.GetRetract().IsSetExp()) 
                        m_ErratumExplanation->SetValue (ToWxString(imp.GetRetract().GetExp()));
                    else 
                        m_ErratumExplanation->SetValue(wxEmptyString);
                    break;
                default:
                    m_ErratumCtrl->SetSelection(0);
                    m_ErratumExplanation->SetValue(wxEmptyString);
                    m_ErratumExplanation->Enable(false);
                    break;
                }
            } 
            else 
            {
                m_ErratumCtrl->SetSelection(0);
                m_ErratumExplanation->SetValue(wxEmptyString);
                m_ErratumExplanation->Enable(false);
            }
        
            if (imp.CanGetDate())
            {
                const CDate& date = imp.GetDate();
                if (m_PubDatePanel)
                    m_PubDatePanel->SetDate(date);
            }

        }

}


END_NCBI_SCOPE
