/*  $Id: legend_handler.cpp 23691 2011-05-18 18:50:11Z wuliangs $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Vlad Lebedev, Andrey Yazhuk
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>
#include <corelib/ncbistd.hpp>

#include <gui/widgets/gl/legend_handler.hpp>
#include <gui/opengl/glhelpers.hpp>
#include <gui/utils/system_path.hpp>

#include <math.h>

BEGIN_NCBI_SCOPE

static float kOffset = 20.0f;

////////////////////////////////////////////////////////////////////////////////
/// class CLegendHandler

BEGIN_EVENT_TABLE(CLegendHandler, wxEvtHandler)
    EVT_LEFT_DOWN(CLegendHandler::OnLeftDown)
    EVT_LEFT_UP(CLegendHandler::OnLeftUp)
    EVT_MOTION(CLegendHandler::OnMotion)
    EVT_KEY_DOWN(CLegendHandler::OnKeyDown)
    EVT_KEY_UP(CLegendHandler::OnKeyUp)
END_EVENT_TABLE()


CLegendHandler::CLegendHandler()
    : m_Pane(NULL)
    , m_Host(NULL)
    , m_Font_Helv14(CGlBitmapFont::eHelvetica14)
    , m_Font_Helv20(CGlBitmapFont::eHelvetica20)
    , m_NeedLegend(false)
    , m_ShowCloseButton(false)
    , m_Highlighted(false)
    , m_ReadyToHideLegend(false)
{
}


CLegendHandler::~CLegendHandler()
{
}


void CLegendHandler::SetPane(CGlPane* pane)
{
    m_Pane = pane;
}


void CLegendHandler::SetHost(ILegendHandlerHost* pHost)
{
    m_Host = pHost;
}


wxEvtHandler* CLegendHandler::GetEvtHandler()
{
    return this;
}


IGenericHandlerHost* CLegendHandler::GetGenericHost()
{
    return dynamic_cast<IGenericHandlerHost*>(m_Host);
}

////////////////////////////////////////////////////////////////////////////////
// event handlers
void CLegendHandler::OnLeftDown(wxMouseEvent& event)
{
    if (m_NeedLegend  &&  m_ShowCloseButton  &&  m_Pane) {
        wxPoint ms_pos = event.GetPosition();
        if (x_PntOnCloseButton(ms_pos.x, ms_pos.y)) {
            m_ReadyToHideLegend = true;
            return;
        }
        m_ReadyToHideLegend = false;
    }

    event.Skip();
}


void CLegendHandler::OnLeftUp(wxMouseEvent& event)
{
    if (m_ReadyToHideLegend  &&  m_NeedLegend  &&
        m_ShowCloseButton  &&  m_Pane) {
        m_ReadyToHideLegend = false;
        wxPoint ms_pos = event.GetPosition();
        if (x_PntOnCloseButton(ms_pos.x, ms_pos.y)) {
            m_NeedLegend = false;
            m_ShowCloseButton = false;
            m_Highlighted = false;
            GetGenericHost()->GHH_Redraw();
            return;
        }
    }

    event.Skip();
}


void CLegendHandler::OnMotion(wxMouseEvent& event)
{
    if (m_NeedLegend  &&  m_ShowCloseButton  &&  m_Pane  &&
        (!event.ButtonIsDown(wxMOUSE_BTN_LEFT) || m_ReadyToHideLegend)) {
        wxPoint ms_pos = event.GetPosition();
        bool highlighted = false;
        if (x_PntOnCloseButton(ms_pos.x, ms_pos.y)) {
            highlighted = true;
        }

        if (m_Highlighted != highlighted) {
            m_Highlighted = highlighted;
            GetGenericHost()->GHH_Redraw();
        }

        if (m_Highlighted) {
            return;
        }
    }

    event.Skip();
}

void CLegendHandler::OnKeyDown(wxKeyEvent& event)
{
    if (wxGetKeyState(wxKeyCode('L'))) {
        if ( ! m_NeedLegend)
            GetGenericHost()->GHH_Redraw();

        m_NeedLegend = true;
    } else {
        event.Skip();
    }
}


void CLegendHandler::OnKeyUp(wxKeyEvent& event)
{
    if (m_NeedLegend  &&  event.GetKeyCode() == wxKeyCode('L')) {
        m_NeedLegend = false;
        m_ShowCloseButton = false;
        m_Highlighted = false;
        GetGenericHost()->GHH_Redraw();
    } else {
        event.Skip();
    }
}


////////////////////////////////////////////////////////////////////////////////
/// helper functions

void CLegendHandler::Render(CGlPane& pane)
{
    if (!m_NeedLegend)
        return;

    pane.OpenPixels();

    const TVPRect& rc = pane.GetViewport();
    float spacing = 8;
    float size = 16;

    x_DrawColorLegendBackGround(rc);

    float x = rc.Left() + kOffset * 2;
    float y = rc.Top()  - kOffset * 2;
    float m = rc.Left() + rc.Width() / 2 + kOffset;

    ILegendHandlerHost::TColorLegend data = m_Host->LHH_GetLegend();
    ITERATE(ILegendHandlerHost::TColorLegend, it, data) {
        const string& label =  it->first;
        TModelUnit width    = m_Font_Helv14.TextWidth(label.c_str());
        ILegendHandlerHost::TLegendValue val = it->second;

        switch (val.first) {
        case ILegendHandlerHost::eColor:
            glColorC(CRgbaColor(val.second));
            glBegin(GL_QUADS);  // filled rect
                glVertex2d(x, y);
                glVertex2d(x+size, y);
                glVertex2d(x+size, y-size);
                glVertex2d(x, y-size);
            glEnd();
            glColor3f(1.0, 1.0, 1.0);
            glBegin(GL_LINE_STRIP);  // outline
                glVertex2d(x, y);
                glVertex2d(x+size, y);
                glVertex2d(x+size, y-size);
                glVertex2d(x, y-size);
                glVertex2d(x, y);
            glEnd();
            m_Font_Helv14.TextOut(x + size * 2, y - 13, label.c_str());
            break;

        case ILegendHandlerHost::eKey:
            glColor3f(1.0, 1.0, 1.0);  // legend color
            m_Font_Helv14.TextOut(x, y - 13, label.c_str());
            glColor3f(0.0, 1.0, 1.0);
            m_Font_Helv20.TextOut(x + width + spacing * 2, y - 15, val.second.c_str());
            break;

        case ILegendHandlerHost::eSeparator:
            break;
        }

        // next line
        y -= size * 1.5f;

        if (y < rc.Bottom() + size * 3) {  // start a second col.
            x = m;
            y = rc.Top()  - kOffset * 2;
        }
    }

    if (m_ShowCloseButton) {
        x_DrawCloseButton(rc);
    }

    pane.Close();
}


void CLegendHandler::SetNeedLegend(bool f)
{
    m_NeedLegend = f;
}


void CLegendHandler::SetShowButton(bool f)
{
    m_ShowCloseButton = f;
}


void CLegendHandler::ToggleLegend()
{
    m_NeedLegend = !m_NeedLegend;
    if (m_NeedLegend) {
        m_ShowCloseButton = true;
    } else {
        m_ShowCloseButton = false;
    }
}


bool CLegendHandler::NeedLegend() const
{
    return m_NeedLegend;
}


void CLegendHandler::x_DrawColorLegendBackGround(const TVPRect& rc) const
{
    // dim display
    glColor4d(1.0, 1.0, 1.0, 0.4);
    glRecti(rc.Left(), rc.Bottom(), rc.Right() + 1, rc.Top());

    glColor4d(0.0, 0.0, 0.0, 0.65);
    glBegin(GL_QUADS);
        glVertex2d(rc.Left()  + kOffset * 2, rc.Top()    - kOffset);
        glVertex2d(rc.Left()  + kOffset * 2, rc.Bottom() + kOffset);
        glVertex2d(rc.Right() - kOffset * 2, rc.Bottom() + kOffset);
        glVertex2d(rc.Right() - kOffset * 2, rc.Top()    - kOffset);

        glVertex2d(rc.Left() + kOffset,     rc.Top()    - kOffset * 2);
        glVertex2d(rc.Left() + kOffset,     rc.Bottom() + kOffset * 2);
        glVertex2d(rc.Left() + kOffset * 2, rc.Bottom() + kOffset * 2);
        glVertex2d(rc.Left() + kOffset * 2, rc.Top()    - kOffset * 2);

        glVertex2d(rc.Right() - kOffset * 2, rc.Top()    - kOffset * 2);
        glVertex2d(rc.Right() - kOffset * 2, rc.Bottom() + kOffset * 2);
        glVertex2d(rc.Right() - kOffset,     rc.Bottom() + kOffset * 2);
        glVertex2d(rc.Right() - kOffset,     rc.Top()    - kOffset * 2);
    glEnd();


    glPushMatrix();
    GLUquadricObj *pQObj = gluNewQuadric();
    float radius = kOffset;
    glTranslated(rc.Left() + kOffset * 2, rc.Top() - kOffset * 2, 0.0);
    gluPartialDisk(pQObj, 0., radius + 1, 32, 1, -90, 90);
    glPopMatrix(); glPushMatrix();

    glTranslated(rc.Right() - kOffset * 2, rc.Top() - kOffset * 2, 0.0);
    gluPartialDisk(pQObj, 0., radius, 32, 1, 0, 90);
    glPopMatrix(); glPushMatrix();

    glTranslated(rc.Left() + kOffset * 2, rc.Bottom() + kOffset * 2, 0.0);
    gluPartialDisk(pQObj, 0., radius, 32, 1, 180, 90);
    glPopMatrix(); glPushMatrix();

    glTranslated(rc.Right() - kOffset * 2, rc.Bottom() + kOffset * 2, 0.0);
    gluPartialDisk(pQObj, 0., radius, 32, 1, 90, 90);
    glPopMatrix();

    gluDeleteQuadric(pQObj);
}


void CLegendHandler::x_DrawCloseButton(const TVPRect& rc) const
{
    TModelRect b_rc;
    x_GetCloseButtonRect(rc, b_rc);

    double PI = 3.1415926543;
    double radius1 = b_rc.Width() * 0.5;
    double radius2 = radius1 - 5.0;
    TModelPoint c_pnt = b_rc.CenterPoint();

    int num = 40;
    double p = 0.0;
    double step = 2 * PI / num;

    glColor3f(0.5f, 0.5f, 0.5f);

    glBegin(GL_QUAD_STRIP);
    for (int i = 0; i <= num; ++i) {
        if (i == num) {
            p = 0.0;
        }
        double x1 = c_pnt.X() + radius2 * cos(p);
        double x2 = c_pnt.X() + radius1 * cos(p);
        double y1 = c_pnt.Y() + radius2 * sin(p);
        double y2 = c_pnt.Y() + radius1 * sin(p);
        glVertex2d(x1, y1);
        glVertex2d(x2, y2);
        p += step;
    }
    glEnd();

    if (m_Highlighted) {
        glColor3f(1.0f, 0.0f, 0.0f);
    } else {
        glColor3f(1.0f, 1.0f, 1.0f);
    }

    double l = radius2 * 0.8 * sin(PI * 0.25);
    double w = l / 4.0;
    glBegin(GL_QUADS);
        glVertex2d(c_pnt.X() + l - w, c_pnt.Y() - l - w);
        glVertex2d(c_pnt.X() + l + w, c_pnt.Y() - l + w);
        glVertex2d(c_pnt.X() - l + w, c_pnt.Y() + l + w);
        glVertex2d(c_pnt.X() - l - w, c_pnt.Y() + l - w);

        glVertex2d(c_pnt.X() + l + w, c_pnt.Y() + l - w);
        glVertex2d(c_pnt.X() + l - w, c_pnt.Y() + l + w);
        glVertex2d(c_pnt.X() - l - w, c_pnt.Y() - l + w);
        glVertex2d(c_pnt.X() - l + w, c_pnt.Y() - l - w);
    glEnd();
}


void CLegendHandler::x_GetCloseButtonRect(const TVPRect& rc,
                                          TModelRect& rect) const
{
    double half_size = 24.0;
    double center_x = rc.Right() - kOffset - half_size - 10.0;
    double center_y = rc.Bottom() + kOffset + half_size + 10.0;
    rect.Init(center_x, center_y);
    rect.Inflate(half_size, half_size);
}


bool CLegendHandler::x_PntOnCloseButton(int x, int y) const
{
    if (m_NeedLegend  &&  m_ShowCloseButton  &&  m_Pane) {
        const TVPRect& rc = m_Pane->GetViewport();
        TModelRect rect;
        x_GetCloseButtonRect(rc, rect);
        if (rect.PtInRect(x, m_Host->LHH_GetVPPosByY(y))) {
            return true;
        }
    }
    return false;
}

END_NCBI_SCOPE
