/*  $Id: seqmark_handler.cpp 27886 2013-04-18 20:28:32Z wuliangs $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Vlad Lebedev
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>


#include <gui/widgets/gl/seqmark_handler.hpp>
#include <gui/widgets/wx/gui_event.hpp>
#include <gui/opengl/glhelpers.hpp>
#include <gui/opengl/glrender.hpp>
#include <gui/utils/system_path.hpp>

#include <math.h>


BEGIN_NCBI_SCOPE


const int kMarkerSize = 6;
////////////////////////////////////////////////////////////////////////////////
/// class CSeqMarkHandler
BEGIN_EVENT_TABLE(CSeqMarkHandler, wxEvtHandler)
    EVT_LEFT_DOWN(CSeqMarkHandler::OnLeftDown)
    EVT_LEFT_DCLICK(CSeqMarkHandler::OnLeftDoubleClick)
    EVT_LEFT_UP(CSeqMarkHandler::OnLeftUp)
    EVT_MOTION(CSeqMarkHandler::OnMotion)
    EVT_MOUSE_CAPTURE_LOST(CSeqMarkHandler::OnMouseCaptureLost)
END_EVENT_TABLE()


CSeqMarkHandler::CSeqMarkHandler(EOrientation orient)
    : m_Orientation(orient)
    , m_Host(NULL)
    , m_MarkDrag(false)
    , m_DragPos(0.0)
    , m_PosOff(0.0)
    , m_Font_Helv10(CGlTextureFont::eFontFace_Helvetica, 10)
    , m_LabelFont(CGlTextureFont::eFontFace_Helvetica, 10)
    , m_Color(1.0f, 0.0f, 0.0f, 0.3f)
    , m_HitMe(false)
{
}


CSeqMarkHandler::~CSeqMarkHandler()
{
}


void CSeqMarkHandler::SetPane(CGlPane* pane)
{
    m_Pane = pane;
}


wxEvtHandler* CSeqMarkHandler::GetEvtHandler()
{
    return this;
}


void CSeqMarkHandler::SetHost(ISeqMarkHandlerHost* host)
{
    m_Host = host;
}


IGenericHandlerHost* CSeqMarkHandler::GetGenericHost()
{
    return dynamic_cast<IGenericHandlerHost*>(m_Host);
}


void CSeqMarkHandler::SetOrientation(EOrientation orient)
{
    m_Orientation = orient;
    if (m_Orientation == eVert) { // top & bottom
        //m_LabelFont.SetFontRotate(CGlTextureFont::eFontRotate_90_Base);
    }
}


////////////////////////////////////////////////////////////////////////////////
// event handlers

void CSeqMarkHandler::OnLeftDown(wxMouseEvent& event)
{
    _ASSERT(m_Pane);

    CGUIEvent::EGUIState state = CGUIEvent::wxGetSelectState(event);
    m_PosOff = 0.0;
    if (state == CGUIEvent::eSelectState)  {
        wxPoint pos = event.GetPosition();
        TModelUnit x_pos = m_Host->SMHH_GetModelByWindow(pos.x, eHorz);
        TModelUnit y_pos = m_Host->SMHH_GetModelByWindow(pos.y, eVert);

        TModelPoint click_point = TModelPoint(x_pos, y_pos);
        TModelRect rc = x_GetMarkRect(*m_Pane);

        if (rc.PtInRect(click_point) ) {
            m_MarkDrag = true;
            TSeqPos pos = m_Host->SMHH_GetSeqMark();
            if (pos != (TSeqPos)(-1)) {
                m_DragPos = (TModelUnit)pos;
            }
            m_PosOff = m_DragPos - (m_Orientation == eHorz ? x_pos : y_pos);
            GetGenericHost()->GHH_CaptureMouse();
            GetGenericHost()->GHH_Redraw();
            return;
        }
    }
    event.Skip();
}


void CSeqMarkHandler::OnLeftDoubleClick(wxMouseEvent& event)
{
    CGUIEvent::EGUIState state = CGUIEvent::wxGetSelectState(event);

    if (state == CGUIEvent::eSelectState)  {
        wxPoint pos = event.GetPosition();
        TModelUnit x_pos = m_Host->SMHH_GetModelByWindow(pos.x, eHorz);
        TModelUnit y_pos = m_Host->SMHH_GetModelByWindow(pos.y, eVert);

        TModelPoint click_point = TModelPoint(x_pos, y_pos);
        TModelRect rc = x_GetMarkRect(*m_Pane);

        if (rc.PtInRect(click_point) ) {
            m_MarkDrag = false;
            m_DragPos  = 0;
            m_Host->SMHH_SetSeqMark(0); // reset to 0
            m_Host->SMHH_OnReset(GetId());
            GetGenericHost()->GHH_Redraw();
            return;
        }
    }
    event.Skip();
}

TModelRect CSeqMarkHandler::x_GetMarkRect(CGlPane& pane) const
{
    TModelRect rc     = pane.GetVisibleRect();
    TModelUnit half_x = pane.UnProjectWidth(kMarkerSize);
    TModelUnit half_y = pane.UnProjectHeight(kMarkerSize);
    TModelUnit pos = m_DragPos;
    TSeqPos marker_pos = m_Host->SMHH_GetSeqMark();
    if (marker_pos != (TSeqPos)(-1)) {
        pos = (TModelUnit)marker_pos;
    }
    pos += 0.5;

    TModelUnit l_w = m_LabelFont.TextWidth(m_Label.c_str());
    TModelRect res;
    if (m_Orientation == eHorz) {
        l_w = pane.UnProjectWidth(l_w);
        res = TModelRect(pos - half_x, rc.Top() - half_y * 2,
                         pos + half_x + l_w, rc.Top());
    } else {
        l_w = pane.UnProjectHeight(l_w);
        res = TModelRect(rc.Right() - half_x * 2, pos - half_y,
                         rc.Right(), pos + half_y + l_w);
    }
    return res;
}


void CSeqMarkHandler::OnMotion(wxMouseEvent& event)
{
    m_HitMe = false;
    if(event.Dragging())    {
        if (m_MarkDrag) {
            wxPoint ms_pos = event.GetPosition();
            if (m_Orientation == eHorz) {
                m_DragPos = m_PosOff + m_Host->SMHH_GetModelByWindow(ms_pos.x, eHorz); // + 0.5;
            } else {
                m_DragPos = m_PosOff + m_Host->SMHH_GetModelByWindow(ms_pos.y, eVert); // + 0.5;
            }
            GetGenericHost()->GHH_Redraw();
        }
        // always handle drags
    } else {
        wxPoint pos = event.GetPosition();
        TModelUnit x_pos = m_Host->SMHH_GetModelByWindow(pos.x, eHorz);
        TModelUnit y_pos = m_Host->SMHH_GetModelByWindow(pos.y, eVert);

        TModelPoint click_point = TModelPoint(x_pos, y_pos);
        TModelRect rc = x_GetMarkRect(*m_Pane);

        if ( rc.PtInRect(click_point) ) {
            m_HitMe = true;
        }

        event.Skip();
    }
}


void CSeqMarkHandler::OnLeftUp(wxMouseEvent& event)
{
    if (m_MarkDrag) {
        x_OnEndAction();
    }
}

void CSeqMarkHandler::OnMouseCaptureLost(wxMouseCaptureLostEvent& evt)
{
    if (m_MarkDrag) {
        x_OnEndAction();
    }
}

void CSeqMarkHandler::x_OnEndAction()
{
    m_MarkDrag = false;
    m_DragPos = int(m_DragPos + 0.5);
    m_Host->SMHH_SetSeqMark( x_SeqDragPos() );
    GetGenericHost()->GHH_ReleaseMouse();
    GetGenericHost()->GHH_Redraw();
}

////////////////////////////////////////////////////////////////////////////////
/// helper functions

TSeqPos CSeqMarkHandler::x_SeqDragPos() const
{
    return (TSeqPos) floor(m_DragPos < 0 ? 0 : m_DragPos);
}


void CSeqMarkHandler::x_DrawMarker(TModelUnit x_c, TModelUnit y_c, int half) const
{
    CGlRender& gl = GetGl();
    gl.Begin(GL_TRIANGLE_FAN);
    if (m_Orientation == eHorz) { // top & bottom
        gl.Vertex2d(x_c, y_c);
        gl.Vertex2d(x_c, y_c + half);
        gl.Vertex2d(x_c + half, y_c);
        gl.Vertex2d(x_c + half, y_c - half);
        gl.Vertex2d(x_c - half, y_c - half);
        gl.Vertex2d(x_c - half, y_c);
        gl.Vertex2d(x_c, y_c + half);
    } else { // left & right
        gl.Vertex2d(x_c, y_c);
        gl.Vertex2d(x_c + half, y_c);
        gl.Vertex2d(x_c, y_c - half);
        gl.Vertex2d(x_c - half, y_c - half);
        gl.Vertex2d(x_c - half, y_c + half);
        gl.Vertex2d(x_c, y_c + half);
        gl.Vertex2d(x_c + half, y_c);
    }
    gl.End();
}


void CSeqMarkHandler::x_DrawMarkerBorder(TModelUnit x_c, TModelUnit y_c, int half) const
{
    CGlRender& gl = GetGl();
    gl.Begin(GL_LINE_LOOP);
    if (m_Orientation == eHorz) { // top & bottom
        gl.Vertex2d(x_c, y_c + half);
        gl.Vertex2d(x_c + half, y_c);
        gl.Vertex2d(x_c + half, y_c - half);
        gl.Vertex2d(x_c - half, y_c - half);
        gl.Vertex2d(x_c - half, y_c);
    } else { // left & right
        gl.Vertex2d(x_c + half, y_c);
        gl.Vertex2d(x_c, y_c - half);
        gl.Vertex2d(x_c - half, y_c - half);
        gl.Vertex2d(x_c - half, y_c + half);
        gl.Vertex2d(x_c, y_c + half);
    }
    gl.End();
}


void CSeqMarkHandler::x_DrawLabel(TModelUnit x, TModelUnit y, int half) const
{
    TModelUnit w = m_LabelFont.TextWidth(m_Label.c_str()) + 4.0;
    TModelUnit h = m_LabelFont.TextHeight() + 1.0;

    CGlRender& gl = GetGl();
    gl.Begin(GL_TRIANGLE_FAN);
    if (m_Orientation == eHorz) { // top & bottom
        x += half;
        y = y + half + 1.0;
        gl.Vertex2d(x + floor(w * 0.5), y - floor(h * 0.5 + 1.5));
        gl.Vertex2d(x, y);
        gl.Vertex2d(x, y - h);
        gl.Vertex2d(x + 3.0, y - h - 3.0);
        gl.Vertex2d(x + w, y - h - 3.0);
        gl.Vertex2d(x + w, y);
        gl.Vertex2d(x, y);
        y = y - h;
        x += 2.0;
    } else { // left & right
        y += half;
        x = x - half - 1.0;
        gl.Vertex2d(x + floor(h * 0.5 + 1.5), y + floor(w * 0.5));
        gl.Vertex2d(x, y);
        gl.Vertex2d(x + h, y);
        gl.Vertex2d(x + h + 3.0, y + 3.0);
        gl.Vertex2d(x + h + 3.0, y + w);
        gl.Vertex2d(x, y + w);
        gl.Vertex2d(x, y);
        y += 2.0;
        x += h;
    }
    gl.End();

    gl.Color4f(1.0f, 1.0f, 1.0f, 1.0f);
    gl.BeginText(&m_LabelFont);
    gl.WriteText(x, y, m_Label.c_str());  
    gl.EndText();
}


void CSeqMarkHandler::Render(CGlPane& pane, ERenderingOption option)
{
    CGlAttrGuard AttrGuard(GL_ENABLE_BIT | GL_POLYGON_BIT | GL_HINT_BIT);
    CGlPaneGuard GUARD(pane, CGlPane::ePixels);

    const TVPRect& rcV = pane.GetViewport();

    TModelUnit pos = m_DragPos;
    if ( !m_DragPos ) {
        TSeqPos marker_pos = m_Host->SMHH_GetSeqMark();
        if (marker_pos != (TSeqPos)(-1)) {
            pos = (TModelUnit)marker_pos;
        }
    }

    pos += 0.5;
    TModelPoint p;
    string text_out = x_GetText(TSeqPos(pos + 1.0));
    TModelUnit width = m_Font_Helv10.TextWidth(text_out.c_str());
    TModelUnit off = option == ePassiveState ? 0 : kMarkerSize * 1.6;

    CGlRender& gl = GetGl();
    m_Color.SetAlpha(0.3f);
    gl.ColorC(m_Color);
    if (m_Orientation == eHorz) {
        TModelUnit cur_pos = pane.ProjectX(pos);

        int add_on = (int) (option == ePassiveState ? m_Font_Helv10.TextHeight() + 3 : 0);
        gl.Begin(GL_LINES); // draw marker line
        gl.Vertex2d(cur_pos, rcV.Bottom() + add_on);
        gl.Vertex2d(cur_pos, rcV.Top() - off);
        gl.End();

        if (option == ePassiveState) { // show currect position
            m_Color.SetAlpha(1.0f);
            gl.ColorC(m_Color);
            gl.BeginText(&m_Font_Helv10);
            gl.WriteText(cur_pos - width / 2.0,
                rcV.Bottom() + 2, text_out.c_str());  
            gl.EndText();
        }
        p = TModelPoint(cur_pos, rcV.Top() - kMarkerSize - 1);
    } else {
        TModelUnit cur_pos = pane.ProjectY(pos - 0.5);

        gl.Begin(GL_LINES);  // draw marker line
        gl.Vertex2d(rcV.Right() - off, cur_pos);
        gl.Vertex2d(rcV.Left(), cur_pos);
        gl.End();

        if (option == ePassiveState) { // show currect position
            m_Color.SetAlpha(1.0f);
            gl.ColorC(m_Color);
            TModelUnit text_y;  // above or below marker line
            if (cur_pos < rcV.Height() / 2) {
                text_y = cur_pos + 2;
            } else {
                text_y = cur_pos - m_Font_Helv10.TextHeight() - 2;
            }
            gl.BeginText(&m_Font_Helv10);
            gl.WriteText(rcV.Left() + 2, text_y, text_out.c_str());  
            gl.EndText();
        }
        p = TModelPoint(rcV.Right() - kMarkerSize, cur_pos);
    }

    // Active State (pane)
    if(option == eActiveState) {  // Draw mark texture
        m_Color.SetAlpha(0.3f);
        gl.ColorC(m_Color);
        //gl.PolygonMode(GL_FRONT_AND_BACK, GL_FILL);
        x_DrawMarker(p.X(), p.Y(), -kMarkerSize);

        if (m_MarkDrag) {
            gl.Color4f(0.0f, 0.0f, 0.0f, 0.9f);  // dark border
        } else {
            m_Color.SetAlpha(0.9f);
            gl.ColorC(m_Color);  // bright border
        }
        x_DrawMarkerBorder(p.X(), p.Y(), -kMarkerSize);

        if ( !m_Label.empty() ) {
            //gl.PolygonMode(GL_FRONT_AND_BACK, GL_FILL);
            m_Color.SetAlpha(0.9f);
            gl.ColorC(m_Color);
            x_DrawLabel(p.X(), p.Y(), kMarkerSize);
        }
    }  // option == eActiveState
}

string CSeqMarkHandler::x_GetText(TSeqPos pos) 
{
    return NStr::SizetToString( (unsigned long) pos, NStr::fWithCommas);
}



END_NCBI_SCOPE
