/*  $Id: phylo_tree_cgi.cpp 26701 2012-10-26 18:06:01Z falkrb $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Author:  Vladimir Tereshkov
 *
 * File Description: Phylogenetic tree CGI, using NetCache
 */

#include <ncbi_pch.hpp>
#include <corelib/ncbistl.hpp>
#include <gui/opengl.h>
#include <cgi/cgictx.hpp>
#include <gui/opengl/mesa/glcgi_image.hpp>
#include <gui/opengl/mesa/gloscontext.hpp>
#include <gui/opengl/glutils.hpp>
#include <gui/opengl/glpane.hpp>
#include <gui/opengl/ftglfontmanager.hpp>
#include <gui/opengl/gltexturefont.hpp>
#include <gui/widgets/wx/sys_path.hpp>
#include <gui/widgets/wx/wx_utils.hpp>
#include <serial/serial.hpp>


#include <objmgr/object_manager.hpp>
#include <objtools/data_loaders/genbank/gbloader.hpp>

#include <gui/widgets/phylo_tree/phylo_tree_rect_cladogram.hpp>
#include <gui/widgets/phylo_tree/phylo_tree_slanted_cladogram.hpp>
#include <gui/widgets/phylo_tree/phylo_tree_radial.hpp>
#include <gui/widgets/phylo_tree/phylo_tree_force.hpp>
#include <gui/widgets/phylo_tree/phylo_tree_node.hpp>
#include <gui/widgets/phylo_tree/phylo_tree_scheme.hpp>

#include <gui/widgets/phylo_tree/phylo_tree_ds.hpp>
#include <gui/widgets/phylo_tree/phylo_tree_render.hpp>

#include <html/html.hpp>
#include <html/page.hpp>

#include <util/image/image.hpp>
#include <util/image/image_util.hpp>
#include <util/image/image_io.hpp>

#include <corelib/ncbi_tree.hpp>
#include <algo/phy_tree/bio_tree.hpp>
#include <algo/phy_tree/bio_tree_conv.hpp>


#include <connect/services/netcache_api.hpp>
//#include <connect/ncbi_socket.hpp>

// output template
#define PHYLO_TREE_TEMPLATE "phylo_tree.html"

// meta-tags inside template ^^^
#define NETCACHE_ID         "NETCACHE_ID"
#define IMAGE_MAP           "IMAGE_MAP"
#define IMAGE_WIDTH         "IMAGE_WIDTH"
#define IMAGE_HEIGHT        "IMAGE_HEIGHT"
#define SELF_URL            "SELF_URL"

// netcache server params
#define NETCACHE_HOST       "graceland"
#define NETCACHE_PORT       9001


BEGIN_NCBI_SCOPE

USING_SCOPE(objects);


class CPhyloTreeCGIMap
{
public:
    CPhyloTreeCGIMap(CGlPane * x_pane) : m_Pane(x_pane), m_Map("") {}
    const string & GetMap() { return m_Map;}

    ETreeTraverseCode operator()(CPhyloTreeNode &  tree_node, int delta)
    {
            if (delta==1 || delta==0){
                int x  = m_Pane->ProjectX((*tree_node).X());
                int y  = m_Pane->GetViewport().Height() - m_Pane->ProjectY((*tree_node).Y());
                int ps = 5;

                string strTooltip = (*tree_node).GetLabel().size()?(*tree_node).GetLabel():"No Label";

                m_Map += "<area alt="+strTooltip+" title="+strTooltip+" href=\'javascript:MouseHit("+NStr::IntToString((*tree_node).GetId())+");\' coords=\'";
                m_Map += NStr::IntToString(x-ps); m_Map+=",";
                m_Map += NStr::IntToString(y-ps); m_Map+=",";
                m_Map += NStr::IntToString(x+ps); m_Map+=",";
                m_Map += NStr::IntToString(y+ps);
                m_Map+="\'>";
            }
            return eTreeTraverse;
    }
private:
    CGlPane * m_Pane;
    string    m_Map;
};



class CPhyloTreeCGIException : public CException
{
public:
    enum EErrCode   {
        eNoRequest,
        eInvalidRequest,
        eInvalidBiotree,
        eCannotLoadBioseq,
        eInvalidVisibleRange,
        eInvalidAnchor
    };
    virtual const char* GetErrCodeString(void) const
    {
        switch(GetErrCode())    {
            case eNoRequest: return "Mandatory parameter request is not specified";
            case eInvalidRequest: return "Invalid request value during";
            case eInvalidBiotree: return "eInvalidBiotree";
            case eCannotLoadBioseq: return "eCannotLoadBioseq";
            case eInvalidVisibleRange: return "eInvalidVisibleRange";
            case eInvalidAnchor: return "eInvalidAnchor";
            default:    return CException::GetErrCodeString();
        };
    }
    NCBI_EXCEPTION_DEFAULT(CPhyloTreeCGIException, CException);
};


/////////////////////////////////////////////////////////////////////////////
//  CPhyloTreeCGIApplication::
//

class CPhyloTreeCGIApplication : public CGlCgiImageApplication
{
public:
    CPhyloTreeCGIApplication();

    virtual int  ProcessRequest(CCgiContext &ctx);
    virtual void Render(CCgiContext& ctx);
protected:
    virtual void x_PreProcess(CCgiContext& ctx);
    virtual void x_PostProcess(CCgiContext& ctx);

    void         x_RenderPage(CCgiContext& ctx);
    void         x_RenderImage(CCgiContext& ctx, bool bToNetcache=false);
    CNetCacheAPI::EReadResult  x_RenderImageFromCache(CCgiContext& ctx);

    void         x_InitWithDummyData(void);
    void         x_InitWithRealData(void);

    string       x_GetMap(void);
    void         x_CreateLayout(CCgiContext &ctx);

    void x_TestHits(void);
private:
    auto_ptr<CGlPane>            m_Pane;
    auto_ptr<IPhyloTreeRenderer> m_Render;
    CRef<CPhyloTreeDataSource>   m_DataSource;
    string                       m_Key;
    int                          m_Hit;

    CCgiRequest               *  m_pRequest;
    TCgiEntries               *  m_pEntries;

    string m_NcService;
    string m_NcClient;
};


CPhyloTreeCGIApplication::CPhyloTreeCGIApplication():
    m_pRequest(NULL), m_pEntries(NULL),
        m_NcService("NC_SeqGraphView"), m_NcClient("demo_phylo_tree.cgi")
{
        m_Key = "";
        m_Hit =-2;
}

int CPhyloTreeCGIApplication::ProcessRequest(CCgiContext &ctx)
{
    // retrieve our CGI rendering params
    m_pRequest = &ctx.GetRequest();
    m_pEntries = &m_pRequest->GetEntries();

    // if key "dummydata" found - init with fake
    TCgiEntries::const_iterator it_dummy = m_pEntries->find("dummydata");
    if (it_dummy != m_pEntries->end()) {
        x_InitWithDummyData();
    } else {
        x_InitWithRealData();
    }

    // trying to obtain a netcache ID
    TCgiEntries::const_iterator it_netcache = m_pEntries->find("netcache_id");
    if (it_netcache != m_pEntries->end()) {
        m_Key = it_netcache->second;
    }

    // reading sizes if possible
    TCgiEntries::const_iterator it_width  = m_pEntries->find("width");
    TCgiEntries::const_iterator it_height = m_pEntries->find("height");

    if (it_width!=m_pEntries->end())  m_Width   =  NStr::StringToInt(it_width->second);;
    if (it_height!=m_pEntries->end()) m_Height  =  NStr::StringToInt(it_height->second);

    // reading mouse hit
    TCgiEntries::const_iterator it_hitid  = m_pEntries->find("hit_id");
    if (it_hitid!=m_pEntries->end()){
        m_Hit  =  NStr::StringToInt(it_hitid->second);
    }


    // looking for request and executing it
    TCgiEntries::const_iterator it_request   = m_pEntries->find("request");
    if (it_request!=m_pEntries->end()) {
        if  (it_request->second == "image") {
            // no way to read image from netcache - recalculate it
            if (m_Key.empty() || (x_RenderImageFromCache(ctx)!=CNetCacheAPI::eReadComplete)) {
                x_CreateLayout(ctx);
                x_RenderImage(ctx); // rendering image
            }
        }
        else  if (it_request->second == "page") {
             x_CreateLayout(ctx);
             x_TestHits();
             x_RenderImage(ctx, true); //  render image to netcache
             x_RenderPage(ctx);  //  rendering page with netcache ID (if possible)
        }
        else {
            NCBI_THROW(CPhyloTreeCGIException, eInvalidRequest, "Invalid Request");
        }
    }
    else {
        NCBI_THROW(CPhyloTreeCGIException, eNoRequest, "Mandatory parameter request is not specified");
    }
    return 0;
}

void CPhyloTreeCGIApplication::x_CreateLayout(CCgiContext& ctx)
{
    // creating pane first
    m_Pane.reset(new CGlPane());

    // creating renderer
    TCgiEntries::const_iterator it_renderer   = m_pEntries->find("renderer");
    if  (it_renderer!=m_pEntries->end()){
        if (it_renderer->second == "rectangle")   m_Render.reset(new CPhyloRectCladogram());
        else if (it_renderer->second == "slanted") m_Render.reset(new CPhyloSlantedCladogram());
        else if (it_renderer->second == "radial") m_Render.reset(new CPhyloRadial());
        else if (it_renderer->second == "force")  m_Render.reset(new CPhyloForce());
    }
    else {
        m_Render.reset(new CPhyloRectCladogram());
    }
    m_Render->SetModelDimensions(m_Width, m_Height);

    // distance mode
    TCgiEntries::const_iterator it_distmode = m_pEntries->find("distmode");
    if (it_distmode!=m_pEntries->end()) {
        m_Render->SetDistRendering(true);
    }

    // visible rect
    TCgiEntries::const_iterator it_vx = m_pEntries->find("vx");
    TCgiEntries::const_iterator it_vy = m_pEntries->find("vy");
    TCgiEntries::const_iterator it_vw = m_pEntries->find("vw");
    TCgiEntries::const_iterator it_vh = m_pEntries->find("vh");

    // margins
    TCgiEntries::const_iterator it_margin_l = m_pEntries->find("marginl");
    TCgiEntries::const_iterator it_margin_t = m_pEntries->find("margint");
    TCgiEntries::const_iterator it_margin_r = m_pEntries->find("marginr");
    TCgiEntries::const_iterator it_margin_b = m_pEntries->find("marginb");

    // creating layout
#ifndef TEXTURED_FONTS
    m_Render->SetFont(new CGlBitmapFont(CGlBitmapFont::eHelvetica10));
#else    
    m_Render->SetFont(new CGlTextureFont(CGlTextureFont::eFontFace_Helvetica, 10));
#endif

    CPhyloTreeScheme * myScheme = new CPhyloTreeScheme();
    myScheme->SetMargins(10, 10, 10, 10);

    // setting margins
    GLdouble ml, mt, mr, mb;
    myScheme->GetMargins(ml, mt, mr, mb);
    if (it_margin_l!=m_pEntries->end()) ml = NStr::StringToInt(it_margin_l->second);
    if (it_margin_t!=m_pEntries->end()) mt = NStr::StringToInt(it_margin_t->second);
    if (it_margin_r!=m_pEntries->end()) mr = NStr::StringToInt(it_margin_r->second);
    if (it_margin_b!=m_pEntries->end()) mb = NStr::StringToInt(it_margin_b->second);
    myScheme->SetMargins(ml, mt, mr, mb);

    // setting node size
    TCgiEntries::const_iterator it_nodesize = m_pEntries->find("nodesize");
    if (it_nodesize!=m_pEntries->end()) {
        myScheme->SetSize(CPhyloTreeScheme::eNodeSize) = NStr::StringToInt(it_nodesize->second);
    }

    // setting colors
    TCgiEntries::const_iterator bgc = m_pEntries->find("bgcolor");
    TCgiEntries::const_iterator nc = m_pEntries->find("node_color");
    TCgiEntries::const_iterator ncs = m_pEntries->find("node_color_sel");
    TCgiEntries::const_iterator nct = m_pEntries->find("node_color_trace");
    TCgiEntries::const_iterator lc = m_pEntries->find("line_color");
    TCgiEntries::const_iterator lcs = m_pEntries->find("line_color_sel");
    TCgiEntries::const_iterator lct = m_pEntries->find("line_color_trace");
    TCgiEntries::const_iterator bc = m_pEntries->find("label_color");
    TCgiEntries::const_iterator bcs = m_pEntries->find("label_color_sel");

    if (bgc!=m_pEntries->end()) myScheme->SetColor(CPhyloTreeScheme::eTree, CPhyloTreeScheme::eBgColor) = CRgbaColor(bgc->second);
    if (nc!=m_pEntries->end())  myScheme->SetColor(CPhyloTreeScheme::eNode,  CPhyloTreeScheme::eColor) = CRgbaColor(nc->second);
    if (ncs!=m_pEntries->end()) myScheme->SetColor(CPhyloTreeScheme::eNode, CPhyloTreeScheme::eSelColor) = CRgbaColor(ncs->second);
    if (nct!=m_pEntries->end()) myScheme->SetColor(CPhyloTreeScheme::eNode, CPhyloTreeScheme::eTraceColor) = CRgbaColor(nct->second);
    if (lc!=m_pEntries->end())  myScheme->SetColor(CPhyloTreeScheme::eLine, CPhyloTreeScheme::eColor) = CRgbaColor(lc->second);
    if (lcs!=m_pEntries->end()) myScheme->SetColor(CPhyloTreeScheme::eLine, CPhyloTreeScheme::eSelColor) = CRgbaColor(lcs->second);
    if (lct!=m_pEntries->end()) myScheme->SetColor(CPhyloTreeScheme::eLine, CPhyloTreeScheme::eTraceColor) = CRgbaColor(lct->second);
    if (bc!=m_pEntries->end())  myScheme->SetColor(CPhyloTreeScheme::eLabel, CPhyloTreeScheme::eColor) = CRgbaColor(bc->second);
    if (bcs!=m_pEntries->end()) myScheme->SetColor(CPhyloTreeScheme::eLabel, CPhyloTreeScheme::eSelColor) = CRgbaColor(bcs->second);


    // subtree display
    TCgiEntries::const_iterator it_subid = m_pEntries->find("sub_id");
    if (it_subid!=m_pEntries->end()){
        int subID = 0;
        try {
            subID = NStr::StringToInt(it_subid->second);
        }
        catch(std::exception&)
        {
        }

        CPhyloTreeNode * nnn = m_DataSource->GetNode(subID);

        if (nnn) {
            m_DataSource->SetSubtree(nnn);
        }

    }


    //myScheme->SetSize(CPhyloTreeScheme::eNodeSize) = 10;
    m_Render->SetScheme(*myScheme);
    m_Render->SetZoomablePrimitives(false);


    m_Render->Layout(m_DataSource.GetObject(), *m_Pane);


    // modifying pane to reflect layout changes
    m_Pane->SetAdjustmentPolicy(CGlPane::fAdjustAll, CGlPane::fAdjustAll);
    m_Pane->SetMinScaleX(1 / 30.0);  m_Pane->SetMinScaleY(1 / 30.0);
    m_Pane->SetOriginType(CGlPane::eOriginLeft, CGlPane::eOriginBottom);
    m_Pane->EnableZoom(true, true);
    m_Pane->SetViewport(TVPRect(0, 0, m_Width, m_Height));

    //m_Pane->SetVisibleRect(TModelRect(0,0,m_Width, m_Height));
    m_Pane->SetModelLimitsRect(m_Render->GetRasterRect());
    m_Pane->SetVisibleRect(m_Render->GetRasterRect());

    if (it_vx!=m_pEntries->end() && it_vy!=m_pEntries->end()&&
        it_vw!=m_pEntries->end() && it_vh!=m_pEntries->end()) {
        int x = NStr::StringToInt(it_vx->second);
        int y = NStr::StringToInt(it_vy->second);
        int w = NStr::StringToInt(it_vw->second);
        int h = NStr::StringToInt(it_vh->second);

        TModelUnit s_x = m_Pane->GetScaleX();
        TModelUnit s_y = m_Pane->GetScaleY();

        m_Pane->ZoomRect(TModelRect(x*s_x, (m_Height-y-h)*s_y, x*s_x+w*s_x, (m_Height-y)*s_y));
    }
    else {
        m_Pane->ZoomAll();
    }
}



void CPhyloTreeCGIApplication::x_InitWithDummyData(void)
{
    CPhyloPhylipReader * reader = new CPhyloPhylipReader();
    CScope * scope = NULL;
    m_DataSource.Reset(new CPhyloTreeDataSource(reader->GetTree(), *scope));
}

void CPhyloTreeCGIApplication::x_InitWithRealData(void)
{
    // TODO: here goes datasource initialization with real data
}

// image render
void CPhyloTreeCGIApplication::Render(CCgiContext& ctx)
{
   glClearColor(0.95, 1.0, 0.95, 1.0);
   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
   m_Render->Render(*m_Pane.get(), m_DataSource.GetObject());
}

void CPhyloTreeCGIApplication::x_RenderPage(CCgiContext& ctx)
{
    CCgiResponse & response = ctx.GetResponse();

    auto_ptr<CHTMLPage> page;
    page.reset(new CHTMLPage("Phylogenetic Tree", PHYLO_TREE_TEMPLATE));
    page->AddTagMap(IMAGE_MAP,  new CHTMLText(x_GetMap()));
    page->AddTagMap(NETCACHE_ID,  new CHTMLText(m_Key));
    page->AddTagMap(IMAGE_WIDTH,  new CHTMLText(NStr::IntToString(m_Width)));
    page->AddTagMap(IMAGE_HEIGHT, new CHTMLText(NStr::IntToString(m_Height)));
    page->AddTagMap(SELF_URL,     new CHTMLText(ctx.GetSelfURL()));


    response.WriteHeader();
    page->Print(response.out(), CNCBINode::eHTML);
}

string CPhyloTreeCGIApplication::x_GetMap(void)
{
    CPhyloTreeCGIMap map_visitor(m_Pane.get());
    map_visitor = TreeDepthFirstTraverse(*m_DataSource->GetTree(), map_visitor);
    return map_visitor.GetMap();
}

void CPhyloTreeCGIApplication::x_TestHits(void)
{
    if ((m_Hit+1) >=0){
        CPhyloTreeNode * nnn = m_DataSource->GetNode(m_Hit);
        if (nnn) {
            m_DataSource->SetSelection(nnn, true, true, true);
        }
    }

}

CNetCacheAPI::EReadResult CPhyloTreeCGIApplication::x_RenderImageFromCache(CCgiContext& ctx)
{
    CNetCacheAPI cli(m_NcService, m_NcClient);

    CImage cached_image(m_Width, m_Height);
    CNetCacheAPI::EReadResult res = cli.GetData(m_Key, cached_image.SetData(), m_Width*m_Height*3);

    if (res==CNetCacheAPI::eReadComplete){
        string encoding("image/png");
        CCgiResponse& response = ctx.GetResponse();
        response.SetContentType(encoding);
        response.WriteHeader();
        CImageIO::WriteImage(cached_image, response.out(), m_Format);
    }

    return res;
}

void CPhyloTreeCGIApplication::x_RenderImage(CCgiContext& ctx, bool bToNetcache)
{
   try {
         if ( !x_GetContext().MakeCurrent() ) {
             LOG_POST(Error << "CGlCgiImageApplication::ProcessRequest(): "
                     "Failed to make off-screen renderer current");
             return;
         }

         x_PreProcess(ctx);

         Render(ctx);
         glFinish();
         m_Context->SetBuffer().SetDepth(3);
         CImageUtil::FlipY(m_Context->SetBuffer());

         x_PostProcess(ctx);
    }
    catch (CException& e) {
        x_HandleError(ctx, e.GetMsg());
        return;
    }
    catch (std::exception& e) {
        x_HandleError(ctx, e.what());
        return;
    }

    if (!bToNetcache) { // rendering to browser
        string encoding("image/png");
        CCgiResponse& response = ctx.GetResponse();
        response.SetContentType(encoding);
        response.WriteHeader();
        CImageIO::WriteImage(m_Context->GetBuffer(), response.out(), m_Format);
    }
    else {  // rendering to netcache
        CNetCacheAPI cli(m_NcService, m_NcClient);
        m_Key = cli.PutData(m_Context->GetBuffer().GetData(),  m_Width*m_Height*3);
    }
}


// creating layout
void    CPhyloTreeCGIApplication::x_PreProcess(CCgiContext& ctx)
{
   CGlCgiImageApplication::x_PreProcess(ctx);
   CFtglFontManager::Instance().SetFontPath(ToStdString(CSysPath::GetResourcePath())); 
}


// creating layout
void    CPhyloTreeCGIApplication::x_PostProcess(CCgiContext& ctx)
{

}

END_NCBI_SCOPE
USING_NCBI_SCOPE;


/////////////////////////////////////////////////////////////////////////////
//  MAIN
//
int main(int argc, const char* argv[])
{
    int result = CPhyloTreeCGIApplication().AppMain(argc, argv, 0, eDS_Default, 0);
    return result;
}

