/*  $Id: phylo_pdf_grabber.cpp 27936 2013-04-26 13:19:38Z falkrb $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Bob Falk
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>
#include <gui/widgets/phylo_tree/phylo_pdf_grabber.hpp>
#include <gui/widgets/phylo_tree/phylo_tree_pane.hpp>
#include <gui/opengl/glresmgr.hpp>
#include <gui/print/glrenderpdf.hpp>
#include <gui/print/pdf.hpp>
#include <gui/print/print_media.hpp>

#include <util/image/image.hpp>

#include <corelib/ncbifile.hpp>



BEGIN_NCBI_SCOPE

CPhyloPdfGrabber::CPhyloPdfGrabber(CPhyloTreePane& pane) 
: m_Pane(pane) 
{
    m_PreviousGlPane = m_Pane.GetPane();
}


IImageGrabber::eCaptureResult 
CPhyloPdfGrabber::GrabImages(int tex_size, IImageGrabberProgress* progress)
{
    if (progress != NULL)
        progress->SetGLContext();

    x_BeginCaptures(tex_size);

    string pdfname = CDir::ConcatPath(m_Directory, m_BaseImageName + "." + m_ImageFormat);
    CNcbiOfstream  pstrm(pdfname.c_str(), ios::out | ios::binary);    
    m_Pdf->SetOutputStream(&pstrm);    


    //***************************************************

    m_Pdf->BeginDocument();
    m_Pdf->BeginPage();
    m_Pane.RenderPdf(m_Pdf);
    m_Pdf->EndPage();

    m_Pdf->EndDocument();
    x_EndCaptures();

    // Show on dialog widget that the image has been saved
    if (progress != NULL)
        progress->ImageSaved(0,0);

    m_Pdf.ReleaseOrNull();

    return eSuccess;
}

void CPhyloPdfGrabber::x_BeginCaptures(int /*buffer_size*/)
{
    // Save some current rendering values we want to change
    CPhyloTreeScheme& s = m_Pane.GetCurrRenderer()->GetScheme();
    m_LabelsTruncated = s.GetLabelTruncation();
    s.SetLabelTruncation(CPhyloTreeScheme::eLabelsFull);
    s.GetMargins(m_MarginLeft, m_MarginTop, m_MarginRight, m_MarginBottom);

    // Update rendering options. It would be better if these margins explicitly
    // included any boundary widths from subtree boundaries that are included
    // in the graph.  But normally those boundaries will be pretty narrow and
    // this should handle it.
    s.SetMargins(20.0, 20.0, 20.0, 20.0);
    
    int viewport_x, viewport_y;

    m_Pdf.Reset(new CPdf());   

    CPrintOptions po;
    po.SetPageOrientation(CPrintOptions::ePortrait);
    string pdfname = CDir::ConcatPath(m_Directory, m_BaseImageName + "." + m_ImageFormat);
    po.SetFilename(pdfname.c_str());
    po.SetMarginBottom(20);
    po.SetMarginLeft(20);
    po.SetMarginRight(20);
    po.SetMarginTop(20);
    po.SetPagesWide(1);
    po.SetPagesTall(1);

    CGlRect< TModelUnit> full_rect = m_Pane.GetPane().GetModelLimitsRect();
    CGlRect< TModelUnit> vis_rect = m_Pane.GetPane().GetVisibleRect();
    TVPRect viewport = m_Pane.GetPane().GetViewport();

    // Use the ratio of visible area to total area to determine
    // the level of zoom, then expand the viewport to accomodate
    // the total view
    double ratiox = vis_rect.Width()/full_rect.Width();
    double ratioy = vis_rect.Height()/full_rect.Height();
    viewport_x = (1.0/ratiox)*double(viewport.Width());
    viewport_y = (1.0/ratioy)*double(viewport.Height());

    float maxdim = std::max(viewport_x, viewport_y);
    if (maxdim > 14000.0f) {
        float scaler = 14000.0f/maxdim;
        float user_unit = 1.0f/scaler;
        po.SetUserUnit(user_unit);
        m_Pane.GetPane().SetOutputUnitScaler(user_unit);

        viewport_x *= scaler;
        viewport_y *= scaler;
    }

    CMedia m(m_OutputFormat, viewport_x, viewport_y, CUnit::ePdfUnit);
    po.SetMedia(m);

    CGlRect<int> vp(0, 0, viewport_x, viewport_y);
    m_Pane.GetPane().SetViewport(vp);

    po.SetOutputFormat(CPrintOptions::ePdf);
    po.SetPagesTall(1);
    po.SetPagesWide(1);
    m_Pdf->SetOptions(po);

    m_Pane.SetupHardcopyRender();
}

void CPhyloPdfGrabber::x_EndCaptures()
{
    m_Pane.GetPane() = m_PreviousGlPane;

    /// Reset rendering options we may have updated.
    CPhyloTreeScheme& s = m_Pane.GetCurrRenderer()->GetScheme();
    s.SetLabelTruncation(m_LabelsTruncated);
    s.SetMargins(m_MarginLeft, m_MarginTop, m_MarginRight, m_MarginBottom);
}



END_NCBI_SCOPE
