/*  $Id: phylo_tree_algorithm.cpp 25300 2012-02-21 20:14:48Z falkrb $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Vladimir Tereshkov
 *
 * File Description:
 *
 */


#include <ncbi_pch.hpp>
#include <corelib/ncbiobj.hpp>
#include <corelib/ncbistd.hpp>

#include <algorithm>
#include <functional>
#include <gui/objutils/utils.hpp>
#include <gui/widgets/phylo_tree/phylo_tree_algorithm.hpp>


#include <algo/phy_tree/bio_tree.hpp>
#include <algo/phy_tree/bio_tree_conv.hpp>
#include <util/bitset/bitset_debug.hpp>
#include <util/bitset/ncbi_bitset.hpp>

#include <objects/biotree/BioTreeContainer.hpp>
#include <objects/biotree/FeatureDescr.hpp>
#include <objects/biotree/FeatureDictSet.hpp>
#include <objects/biotree/NodeFeatureSet.hpp>
#include <objects/biotree/NodeFeature.hpp>
#include <objects/biotree/NodeSet.hpp>
#include <objects/biotree/Node.hpp>

BEGIN_NCBI_SCOPE

// IPhyloTreeVisitor
ETreeTraverseCode IPhyloTreeVisitor::operator()( TTreeType &  tree_node, int delta)
{
    return x_OnStep(tree_node, delta);
}

ETreeTraverseCode IPhyloTreeVisitor::x_OnStep(TTreeType &  x_node, int delta)
{
    switch (delta) {
    case 0:  return x_OnStepDown(x_node);
    case 1:  return x_OnStepRight(x_node);
    case -1: return x_OnStepLeft(x_node);
    default: break;
    }
    return eTreeTraverse;
}

ETreeTraverseCode IPhyloTreeVisitor::x_OnStepDown(TTreeType & /* x_node */)
{
    return eTreeTraverse;
}

ETreeTraverseCode IPhyloTreeVisitor::x_OnStepRight(TTreeType & /* x_node */)
{
    return eTreeTraverse;
}

ETreeTraverseCode IPhyloTreeVisitor::x_OnStepLeft(TTreeType & /* x_node */)
{
    return eTreeTraverse;
}

// CPhyloTreeCalculator
CPhyloTreeCalculator::CPhyloTreeCalculator()
{
    Init();
}

void CPhyloTreeCalculator::Init()
{
    m_PamlID = 0;
    m_ID   = 0;
    m_NodeCount = 0;
    m_VisibleNodes = 0;
    m_PosX = 0;
    m_PosY = 0;
    m_Leaves = 0;
    m_Childs = 0;
    m_Width = 0;
    m_Height = 0;
    m_MaxDistance = 0;
    m_MinDistance = 0;
    m_MaxLabelLen = 0;
    m_MaxChilds   = 0;
    m_Clusters.clear();
    m_AttrTable = NULL;
    m_LabelFormat = "";
}

void CPhyloTreeCalculator::SetAttrTable(const TAttrTable  & attr) 
{ 
    m_AttrTable = &attr;

    m_AttrKeys.clear();
    m_AttrKeys.reserve(attr.Rows());

    string key_id = "seq-id";
    unsigned int key_x = m_AttrTable->ColumnIdx(key_id);

    // Initialize a separate vector of sorted seq-id strings
    // from the table.  This can be used to look up table entries
    // much faster when node seq-ids and table seq-ids match via
    // string compare, or when the table entry is a shortened
    // version of the node entry, e.g. AAT66223 vs AAT66223.1
    for (unsigned int i=0; i<m_AttrTable->Rows(); i++) {
        string seqid_str = m_AttrTable->GetCell(i, key_x);

        // Do not create an index for the entry if theris no seq-id
        if (seqid_str.empty()) continue;

        AttrKey akey(seqid_str, i);
        m_AttrKeys.push_back(akey);            
    }

    std::sort(m_AttrKeys.begin(), m_AttrKeys.end());
}

void CPhyloTreeCalculator::x_UpdateProperties(TTreeType& x_node, int row_idx)
{
    /// Go through all the properties in 'row_idx' of the attribute table
    /// and either update the corresponding value in the node, or add 
    /// a new value if one doesn't already exist.
    for (unsigned int j=0; j<m_AttrTable->Cols(); ++j) {
        string attr_name = m_AttrTable->Column(j);
        string attr_value = m_AttrTable->GetCell(row_idx, j);

        (*x_node).SetFeature(attr_name, attr_value);

        if (attr_name == "label") {
            (*x_node).SetLabel(attr_value);
        }
    }
    (*x_node).InitFeatures();
}

ETreeTraverseCode CPhyloTreeCalculator::x_OnStep(TTreeType &  x_node, int delta)
{
    ETreeTraverseCode traverse_code =
        IPhyloTreeVisitor::x_OnStep(x_node, delta);

    if (delta==1 || delta==0) {                
        // If the optional attribute table has been provided, see if it has
        // an entry with a seq-id matching the current node and, if so, 
        // update/add current node properties with properties from the table.
        if (m_AttrTable && 
            m_AttrTable->Rows()>0 && 
            !(*x_node).GetSeqID().IsNull()) {

            string key_id = "seq-id";
            unsigned key_x = m_AttrTable->ColumnIdx(key_id);

            bool found = false;

            // First try to look up the matching entry based on simple string 
            // comparison using the sorted array of table seq-ids placed in
            // m_AttrKeys.  If that fails, we do a full (inefficient) scan
            // of all table entries to try to find a match.
            std::vector<AttrKey>::iterator iter;
            AttrKey comp_key((*x_node).GetSeqID()->GetSeqIdString(), 0);

            iter = std::lower_bound(m_AttrKeys.begin(), m_AttrKeys.end(), 
                                    comp_key);

            // iter is closet string match, not necssarily excat (but
            // we check it either way).
            if (iter != m_AttrKeys.end()) {
                try {
                    // put more efficient check first...
                    if (comp_key.m_SeqIDStr == (*iter).m_SeqIDStr) {
                        found = true;                            
                    }
                    else {
                        objects::CSeq_id  sid((*iter).m_SeqIDStr);
                        if ((*x_node).GetSeqID()->Match(sid))
                            found = true;                            
                    }

                    if (found)
                        x_UpdateProperties(x_node, (*iter).m_AttrTableIdx);                   
                }
                // May have errors from mal-formed seq-ids in file - ignore those.
                catch(...)
                {
                    found = false;
                }
            }

            // Not found.  Check all table entries one at a time.
            if (!found) {
                for (unsigned i=0; i<m_AttrTable->Rows(); i++) {
                    try {
                        string seqid_str = m_AttrTable->GetCell(i, key_x);

                        // Ignore attribute rows without a key (seq-id).
                        // (could also use node-id potentially as a lookup - but 
                        //  it would have to be exported from the tree first)
                        if (seqid_str.empty()) continue;

                        objects::CSeq_id sid(seqid_str);

                        if ((*x_node).GetSeqID()->Match(sid)) {
                            x_UpdateProperties(x_node, i);
                            break;
                        }
                    }
                    // May have errors from mal-formed seq-ids in file - ignore those.
                    catch (...){break;}
                }
            }
        }

        /// Perform all other node calculations
        TTreeType * parent = static_cast<TTreeType*>(x_node.GetParent());
        if (parent) {
            // distances recalculation
            double dist = (double)((**parent).GetDistFromRoot() + (*x_node).GetDistance());
            (*x_node).SetDistFromRoot(dist);
            if (m_MaxDistance < dist) {
                m_MaxDistance = dist;
            }
            if (m_MinDistance==0 || (m_MinDistance > dist)){
                m_MinDistance = dist;
            }
        }
        (*x_node).SetNodesNmb(0);
        (*x_node).SetLeavesNmb(0);
        (*x_node).SetChildsNmb(0);
        (*x_node).SetChildsNmbEx(0);
        (*x_node).SetLeavesNmbEx(0);
        (*x_node).SetCounter(m_ID++);
        
        // Under current algorithm, this will be same as ID
        ++m_NodeCount;

        if (x_node.IsVisible())
            ++m_VisibleNodes;

        if (!(x_node).IsLeaf()) {
            (*x_node).SetPamlCounter(m_PamlID++);
        }

        // update label corresponding to format
        if (!m_LabelFormat.empty()) {
            (*x_node).SetLabel(m_PTL.GetLabelForNode(&x_node, m_LabelFormat));
        }


        if (x_node.IsLeaf() && ((*x_node).GetLabel().length() > (unsigned) m_MaxLabelLen)) {
           m_MaxLabelLen = (*x_node).GetLabel().length();
        }       

        string clusterID = (*x_node).GetClusterID();

        // cluster ID
        if (!clusterID.empty()) {
            m_Clusters[clusterID].push_back(&x_node);
        }
    }
    return traverse_code;
}


ETreeTraverseCode CPhyloTreeCalculator::x_OnStepDown(TTreeType & x_node)
{
    m_PosY += (x_node.IsVisible()?1:0);

    (*x_node).IDX().first  = m_PosX;
    (*x_node).IDX().second = m_PosY;



    // adjusting height
    if (m_Height < m_PosY) { m_Height = m_PosY; }

    return eTreeTraverse;
}

ETreeTraverseCode CPhyloTreeCalculator::x_OnStepRight(TTreeType & x_node)
{
    m_PosX += (x_node.IsVisible()?1:0);

    // position
    (*x_node).IDX().first  = m_PosX;
    (*x_node).IDX().second = m_PosY;

    
    // adjusting width
    if (m_Width < m_PosX) { m_Width = m_PosX; }

    return eTreeTraverse;
}

ETreeTraverseCode CPhyloTreeCalculator::x_OnStepLeft(TTreeType & x_node)
{
    m_PosX -= ((x_node.IsVisible() && x_node.Expanded())?1:0);

    // number of leaves/childs calculation
    for (TTreeType::TNodeList_I it = x_node.SubNodeBegin();
            it != x_node.SubNodeEnd();
            it++){
        CPhyloTreeNode * nn = static_cast<CPhyloTreeNode*>(*it);

        // our leafs
        if (x_node.Expanded()){
            if (nn->IsLeafEx()){
                (*x_node).SetLeavesNmbEx((*x_node).GetLeavesNmbEx()+1);
            }
            else {
                (*x_node).SetLeavesNmbEx((*x_node).GetLeavesNmbEx() + (***it).GetLeavesNmbEx());
            }
            (*x_node).SetChildsNmbEx((*x_node).GetChildsNmbEx()+1);
        }

        // 'real leafs'
        if (nn->IsLeaf()){
            (*x_node).SetLeavesNmb((*x_node).GetLeavesNmb()+1);
        }
        else {
            (*x_node).SetLeavesNmb((*x_node).GetLeavesNmb() + (***it).GetLeavesNmb());
            (*x_node).SetNodesNmb((*x_node).GetNodesNmb() + (***it).GetNodesNmb());
        }
        (*x_node).SetChildsNmb((*x_node).GetChildsNmb() + 1);
        (*x_node).SetNodesNmb((*x_node).GetNodesNmb()   + 1);

        if ((*x_node).GetChildsNmb() > m_MaxChilds) {
           m_MaxChilds = (*x_node).GetChildsNmb();
        }
    }

    return eTreeTraverse;
}

// CPhyloTreeSorter
ETreeTraverseCode CPhyloTreeSorter::x_OnStep(TTreeType &  tree_node, int delta)
{
    if (delta == 0 || delta == 1) {
        tree_node.Sort(m_bOrder);
    }
    return eTreeTraverse;
}

// CPhyloTreeRectCalculator
CPhyloTreeRectCalculator::CPhyloTreeRectCalculator()
{
    Init();
}

ETreeTraverseCode CPhyloTreeRectCalculator::x_OnStep(TTreeType &  x_node, int delta)
{
    if (!x_node.IsVisible()) return eTreeTraverseStepOver;

    else if (delta == 0 || delta == 1) {
        const IPhyGraphicsNode::TPoint & xy = (*x_node).XY();
        /// m_Rect.IsEmpty() returns true even when rect is line
        if ((m_Rect.Left()==m_Rect.Right()) && (m_Rect.Top()==m_Rect.Bottom())){
            m_Rect.Init(xy.X()-1e-04, xy.Y()-1e-04, xy.X()+1e-04, xy.Y()+1e-04);
        }
        else {
            if (m_Rect.Left() > xy.X())    m_Rect.SetLeft(xy.X());
            if (m_Rect.Right() < xy.X())   m_Rect.SetRight(xy.X());
            if (m_Rect.Bottom() > xy.Y()) m_Rect.SetBottom(xy.Y());
            if (m_Rect.Top() < xy.Y())    m_Rect.SetTop(xy.Y());  
        }
    }
    return eTreeTraverse;
}


void CPhyloTreeRectCalculator::Init()
{
    m_Rect.Init();
}

CPhyloTreeMaxIdCalculator::CPhyloTreeMaxIdCalculator() 
: m_Id(0) 
{
}

ETreeTraverseCode CPhyloTreeMaxIdCalculator::x_OnStep(TTreeType &  x_node, int delta)
{
    if (delta == 0 || delta == 1) {
        m_Id = std::max(m_Id, (*x_node).GetId());
    }
    return eTreeTraverse;
}

ETreeTraverseCode IPhyloTreeFilter::x_OnStep(TTreeType &  x_node, int delta)
{
    bool bAllowed = x_Allowed(x_node);

    if (m_TreeStack.size() == 0) {
        // first node to begin conversion with
        if (bAllowed) {
            auto_ptr<TTreeType> pnode(MakeNewTreeNode(x_node));
            m_TreeStack.push_back(pnode.get());
            m_Tree = pnode.release();
        }
        return eTreeTraverse;
    }

    if (delta == 0) {
        
        TTreeType* back_node = NULL;

        if (m_TreeStack.size() > 0) {
            back_node = m_TreeStack.back();
            m_TreeStack.pop_back();           
        }

        if (!bAllowed) {         
            if (m_TreeStack.size() == 0)
                m_TreeStack.push_back(back_node);
            else
                m_TreeStack.push_back(m_TreeStack.back());
            return eTreeTraverse;
        }


        TTreeType* parent_node = m_TreeStack.back();
        TTreeType* pnode= MakeNewTreeNode(x_node);

        parent_node->AddNode(pnode);
        m_TreeStack.push_back(pnode);
        return eTreeTraverse;
    }

    if (delta == 1) {
        // adjusting stack
        if (!bAllowed) {
            m_TreeStack.push_back(m_TreeStack.back());
            return eTreeTraverse;
        }

        TTreeType* parent_node = m_TreeStack.back();
        TTreeType* pnode= MakeNewTreeNode(x_node);

        parent_node->AddNode(pnode);
        m_TreeStack.push_back(pnode);
        return eTreeTraverse;
    }
    if (delta == -1) {
        // returned to first node = stop traversal
        if (m_TreeStack.empty()) {
            return eTreeTraverseStop;
        }
        m_TreeStack.pop_back();
    }

    return eTreeTraverse;
}


END_NCBI_SCOPE
