/*  $Id: phylo_tree_ds.cpp 26700 2012-10-26 16:01:53Z falkrb $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Vladimir Tereshkov
 *
 * File Description:
 *
 */


#include <ncbi_pch.hpp>

#include <gui/widgets/phylo_tree/phylo_tree_ds.hpp>

#include <corelib/ncbiobj.hpp>
#include <corelib/ncbistd.hpp>

#include <algorithm>
#include <functional>
#include <gui/objutils/utils.hpp>
#include <gui/utils/rgba_color.hpp>
#include <gui/objutils/query_func_promote.hpp>

#include <algo/phy_tree/bio_tree.hpp>
#include <algo/phy_tree/bio_tree_conv.hpp>
#include <util/xregexp/regexp.hpp>

#include <gui/utils/event_translator.hpp>
#include <gui/framework/app_task_service.hpp>
#include <gui/utils/app_job_impl.hpp>
#include <gui/utils/view_event.hpp>

#include <objects/biotree/BioTreeContainer.hpp>
#include <wx/richtext/richtextctrl.h>


BEGIN_NCBI_SCOPE

CPhyloTreeNode * CPhyloTreeDataSource::m_sClipboard;

CPhyloTreeDataSource::CPhyloTreeDataSource() 
: m_SearchCurrentNode(NULL)
, m_Current(NULL)
{
    m_Root = new CPhyloTreeNode();
    x_InitRenderModel();
}


CPhyloTreeDataSource::CPhyloTreeDataSource(const CBioTreeDynamic & tree)
 : m_SearchCurrentNode(NULL)
 , m_Current(NULL)
{
    Init(tree);
}


CPhyloTreeDataSource::CPhyloTreeDataSource(const CBioTreeDynamic& tree,
                                           objects::CScope& scope) 
: m_SearchCurrentNode(NULL)
, m_Scope(&scope)
, m_Current(NULL)
{
   Init(tree);
}


CPhyloTreeDataSource::~CPhyloTreeDataSource()
{    
    delete m_Root;
    m_Root = NULL;
}

void CPhyloTreeDataSource::Init(const CBioTreeDynamic & tree)
{
    m_SearchCache.clear();
    m_SearchCurrentNode = NULL;

    m_Root = new CPhyloTreeNode();
    m_Dictionary = tree.GetFeatureDict();
   
    dynamic_to_tree dttFunc;
    dttFunc.SetDictionaryPtr(&m_Dictionary);

    DynamicConvert2Tree(tree, dttFunc, m_Root);

    TreeDepthFirstTraverse(*m_Root, visitor_read_properties());
    x_MeasureTree();

    x_InitRenderModel();
}

void CPhyloTreeDataSource::Relabel(string labelFmt)
{
    m_Calc.Init();
    m_Calc.SetLabelFormat(labelFmt);
    m_Calc = TreeDepthFirstTraverse(*m_Root, m_Calc);
}


void CPhyloTreeDataSource::x_MeasureTree(TTreeType & node)
{
    m_Calc.Init();
    m_Calc = TreeDepthFirstTraverse(node, m_Calc);
}

void CPhyloTreeDataSource::x_MeasureTree()
{
    x_MeasureTree(*m_Root);
}


void CPhyloTreeDataSource::Save(CBioTreeDynamic & btree, const TTreeType * x_root)
{
    TreeConvert2Dynamic(btree, x_root ? x_root : m_Root, tree_to_dynamic(this, btree));
}


void CPhyloTreeDataSource::Sort(bool bAscending)
{
    TreeDepthFirstTraverse(*m_Root, CPhyloTreeSorter(bAscending));
    x_MeasureTree();
}

string CPhyloTreeDataSource::GenerateTooltipFormat()
{
    string ttf = "";
    
    if (m_Root) {
        CBioTreeFeatureDictionary dict = *(**m_Root).GetDictionaryPtr();
        
        ITERATE(CBioTreeFeatureDictionary::TFeatureDict, it, dict.GetFeatureDict()) {
            string sName = it->second;
            string sKey  = "$(" + it->second + ")";
           
            ttf += ttf.empty()?"" : "\n";
            ttf += (sName + ": " + sKey);            
        }
    }

    return ttf.empty()?"Tree Node":ttf;
}


void CPhyloTreeDataSource::Clusterize(void)
{
    // seed color
   // CRgbaColor          color("red2");
   // CRgbaColor          color(51, 102, 255);
    CRgbaColor          color(204, 153, 102);
    CRgbaColor          white(255,255,255,255);
    
    // COLOR 'STEP'
    int   nmbClusters = m_Calc.GetClusters().size();
    float colorStep = nmbClusters? (180.0 / nmbClusters): 0;
    //unsigned contrast = 1;
    // for narrow color angle we will try an alternative color assignment 
    bool narrow_angle = colorStep < 5.0; 

    // order clusters numerically (if possible)
    typedef map<unsigned, CPhyloTreeCalculator::TClusterHash::key_type> TNumClusterIdx;
    TNumClusterIdx clust_idx;
    ITERATE (CPhyloTreeCalculator::TClusterHash, cl, m_Calc.GetClusters()) {
        unsigned cluster_id = NStr::StringToUInt(cl->first, NStr::fConvErr_NoThrow);
        if (cluster_id) {
            clust_idx[cluster_id] = cl->first;
        }
    } // ITERATE


    // propagate colors
    const CPhyloTreeCalculator::TClusterHash& cluster_map = m_Calc.GetClusters();
    ITERATE(TNumClusterIdx, clid_key, clust_idx) {

        const string& clid_str = clid_key->second;
        CPhyloTreeCalculator::TClusterHash::const_iterator cl = cluster_map.find(clid_str);
        if (cl == cluster_map.end())
            continue;
        
        // getting min x of each cluster
        int minX = m_Calc.GetWidth();
        ITERATE (vector<CPhyloTreeNode*>, node, cl->second) {
            CPhyloTreeNode * par =  static_cast<CPhyloTreeNode*>((*node)->GetParent());
            if (par && (**par).IDX().first < minX) {
                minX = (**par).IDX().first;    
            }
        }
        
        ITERATE (vector<CPhyloTreeNode*>, node, cl->second) {
            //(***node).SetNodeColor(color.ToString(), color.ToString());

            // if it is not lowest common node, mark parents       
            for (CPhyloTreeNode * cursor = *node; cursor; 
                cursor = static_cast<CPhyloTreeNode*>(cursor->GetParent())){
                    if ((**cursor).IDX().first > minX) {
                         (**cursor).SetClusterColor(color.ToString(), color.ToString());
                    }
                    else {
                        //(**cursor).SetNodeColor(IPhyGraphicsNode::TNoColor, white.ToString());
                        break;
                    }
            }
        }

       float rotate_angle = colorStep + (narrow_angle ? 90.0 : 180.0);
       color = CRgbaColor::RotateColor(color, rotate_angle); 

       // This code is to protect us from colors too bright (almost white)
       //
       float c_dist = CRgbaColor::ColorDistance(white, color);
       while (c_dist < 0.2f) {
           color.Darken(0.15f);
           c_dist = CRgbaColor::ColorDistance(white, color);
       }

    }
}


void CPhyloTreeDataSource::Refresh(void)
{
    TreeDepthFirstTraverse(*m_Root, visitor_read_properties());
    x_MeasureTree();
}

void CPhyloTreeDataSource::ShowAll(void)
{
    TreeDepthFirstTraverse(*m_Root, visitor_show_all());
    x_MeasureTree();
}

void CPhyloTreeDataSource::MarkAllDirty(bool bDirty)
{
    TreeDepthFirstTraverse(*m_Root, visitor_set_dirty(bDirty));
    x_MeasureTree();
}


const TModelRect & CPhyloTreeDataSource::GetBoundRect(void)
{
    m_Rect.Init();
    m_Rect = TreeDepthFirstTraverse(*m_Root, m_Rect);
    return m_Rect.GetRect();
}


void CPhyloTreeDataSource::SetSelection(const vector<CPhyloTreeNode*>& sel_nodes)
{
    SetSelection(GetTree(), false, true);

    ITERATE(vector<CPhyloTreeNode*>, it, sel_nodes) {
        SetSelection(*it, true, true);
    }
}


void  CPhyloTreeDataSource::SetSelection(CPhyloTreeNode * node,
                                         bool bSel,
                                         bool andChilds,
                                         bool checkParents)
{
    (**node).SetSelectedState(bSel ?    IPhyGraphicsNode::eSelected :
                                        IPhyGraphicsNode::eNotSelected);

    if (andChilds) {
        x_DescentToChilds(node, bSel);
    }

    if (checkParents) {
        x_AscendToRoot(node, bSel);
        x_CommonNodeSearch(node);
    }
}

void  CPhyloTreeDataSource::QuerySelect(CPhyloTreeNode * node, 
                                        bool bSel,
                                        bool select_children,
                                        bool select_parents)
{
    (**node).SetSelectedState(bSel ?    IPhyGraphicsNode::eSelected :
                                        IPhyGraphicsNode::eNotSelected);

    if (select_children) {
        x_SelectChildren(node, IPhyGraphicsNode::eSelected);
    }

    if (select_parents) {
        x_SelectParents(node, IPhyGraphicsNode::eSelected);
    }

}


void  CPhyloTreeDataSource::x_DescentToChilds(CPhyloTreeNode * node,
                                              bool bSel)
{
    (**node).SetSelectedState(bSel ?    IPhyGraphicsNode::eSelected :
                                        IPhyGraphicsNode::eNotSelected);
    if (!node->IsLeaf()) {
        for(CPhyloTreeNode::TNodeList_I  it = node->SubNodeBegin();
                                         it != node->SubNodeEnd(); it++ )  {
             x_DescentToChilds(static_cast<CPhyloTreeNode*>(*it), bSel);
        }
    }
}

void  CPhyloTreeDataSource::x_AscendToRoot(CPhyloTreeNode * node, bool bSel)
{
    CPhyloTreeNode * parent = static_cast<CPhyloTreeNode*>(node->GetParent());

    if (parent){
        bool bAllSelected  = true;
        bool bSomething    = false;
        for (CPhyloTreeNode::TNodeList_I  it = parent->SubNodeBegin();
                                          it != parent->SubNodeEnd(); it++ ) {
            bool bSelState   = ((***it).GetSelectedState() == IPhyGraphicsNode::eSelected);
            bool bTraceState = ((***it).GetSelectedState() != IPhyGraphicsNode::eNotSelected);

            bAllSelected &= bSelState;
            bSomething   |= bTraceState;
        }
        if (bAllSelected) {
            (**parent).SetSelectedState(bSel ? IPhyGraphicsNode::eSelected :
                                               IPhyGraphicsNode::eNotSelected);
        }
        else {
            (**parent).SetSelectedState(bSomething ? IPhyGraphicsNode::eTraced :
                                               IPhyGraphicsNode::eNotSelected);
        }
        x_AscendToRoot(parent, bSel);
    }
}

void  CPhyloTreeDataSource::x_SelectChildren(CPhyloTreeNode * node,
                                             IPhyGraphicsNode::TSelectedState sel)                                          
{
    (**node).SetSelectedState(sel);  
    for(CPhyloTreeNode::TNodeList_I it = node->SubNodeBegin();
                                    it != node->SubNodeEnd(); it++ )  {
        x_SelectChildren(static_cast<CPhyloTreeNode*>(*it), sel);
    }
}

void  CPhyloTreeDataSource::x_SelectParents(CPhyloTreeNode * node, 
                                            IPhyGraphicsNode::TSelectedState sel)
{
    CPhyloTreeNode * parent = static_cast<CPhyloTreeNode*>(node->GetParent());

    if (parent){
        bool bAllSelected  = true;
        bool bSomething    = false;
        for (CPhyloTreeNode::TNodeList_I  it = parent->SubNodeBegin();
                                          it != parent->SubNodeEnd(); it++ ) {
            bool bSelState   = 
                (***it).GetSelectedState() == IPhyGraphicsNode::eSelected;
            bool bTraceState = ((***it).GetSelectedState() != 
                                IPhyGraphicsNode::eNotSelected);

            bAllSelected &= bSelState;
            bSomething   |= bTraceState;
        }
        if (bAllSelected) {
            (**parent).SetSelectedState(sel);
        }
        else {
            (**parent).SetSelectedState(bSomething ? IPhyGraphicsNode::eTraced :
                                               IPhyGraphicsNode::eNotSelected);
        }
        x_AscendToRoot(parent, sel);
    }
}


void CPhyloTreeDataSource::x_CommonNodeSearch(CPhyloTreeNode * node)
{
    // finding common node
    CPhyloTreeNode * common = NULL;
    CPhyloTreeNode * cursor = node;

    while (cursor){

        int iMarkedNodes=0;

        for (CPhyloTreeNode::TNodeList_I  it = cursor->SubNodeBegin();
                it != cursor->SubNodeEnd(); it++ ) {
            if ((***it).GetSelectedState()!=
                IPhyGraphicsNode::eNotSelected) {
               iMarkedNodes++;
            }
        }
        if (iMarkedNodes > 1){
            common = cursor;
        }
        cursor = static_cast<CPhyloTreeNode*>(cursor->GetParent());
    }

    // common found, descending
    if (common) {
        x_PropagateCommonState(common);
    }
}


void  CPhyloTreeDataSource::x_PropagateCommonState(CPhyloTreeNode * node)
{
    if ((**node).GetSelectedState()==IPhyGraphicsNode::eTraced) {
        (**node).SetSelectedState(IPhyGraphicsNode::eShared);

        if (!node->IsLeaf()) {
            for(CPhyloTreeNode::TNodeList_I  it = node->SubNodeBegin();
                it != node->SubNodeEnd(); it++ )  {
                x_PropagateCommonState(static_cast<CPhyloTreeNode*>(*it));
            }
        }
    }
}


vector<int> CPhyloTreeDataSource::GetSelectedIds(void) const
{
    visitor_selection_reader sel;
    sel = TreeDepthFirstTraverse(*m_Root, sel);
    return sel.GetIDs();
}


vector<CPhyloTreeDataSource::TTreeType*> CPhyloTreeDataSource::GetSelectedNodes(void) const
{
    visitor_selection_reader sel;
    sel = TreeDepthFirstTraverse(*m_Root, sel);
    return sel.GetNodes();
}

class visitor_regex_query
{
public:
    visitor_regex_query(CRegexp* regex)  : m_QueryEx(regex) {}
    visitor_regex_query()  : m_QueryEx(NULL) {}

    ETreeTraverseCode operator()( CPhyloTreeDataSource::TTreeType &  tree_node, int delta)
    {
        if (delta==1 || delta==0){
            CBioTreeFeatureList::TFeatureList f_list = 
                (*tree_node).GetBioTreeFeatureList().GetFeatureList();


            ITERATE(CBioTreeFeatureList::TFeatureList, it, f_list) {
                ///if (it->value.find(m_QString)!=string::npos) {
                if (m_QueryEx->IsMatch(it->value)) {
                    m_SelNodes.push_back(&tree_node);
                    m_SelIDs.push_back((*tree_node).GetId());
                    break;
                }                     
            }    
        }
        return eTreeTraverse;
    }

    const vector <int>          & GetIDs()      { return m_SelIDs;   }
    const vector<CPhyloTreeDataSource::TTreeType*>    & GetNodes()    { return m_SelNodes; }
private:
    vector <CPhyloTreeDataSource::TTreeType*> m_SelNodes;
    vector <int>        m_SelIDs;
    CRegexp*            m_QueryEx;
};

vector<CPhyloTreeDataSource::TTreeType*> 
CPhyloTreeDataSource::FindNodes(string query, bool casesensitive) const
{
    // If search string has a '*' or a '?', convert it to a regular expression
    // to support a wildcard search, otherwise do a (probably faster) simple
    // string search.
    if (query.find_first_of('*') != std::string::npos ||
        query.find_first_of('?') != std::string::npos) {
        
            try {
                string pattern = CRegexp::WildcardToRegexp(query);
                CRegexp regex(pattern, casesensitive ? 
                    CRegexp::fCompile_default : CRegexp::fCompile_ignore_case);

                visitor_regex_query finder(&regex);
                finder = TreeDepthFirstTraverse(*m_Root, finder);
                return finder.GetNodes();
            }
            catch(CRegexpException& err) {
                LOG_POST("Error creating regular expression from wildcard pattern : " << err.GetMsg());
                vector<CPhyloTreeDataSource::TTreeType*> results_error;
                return results_error;
            }
    }
    else {
        if (casesensitive) {
            visitor_string_query finder(query);
            finder = TreeDepthFirstTraverse(*m_Root, finder);
            return finder.GetNodes();
    }
        else {
            visitor_string_query_no_case finder(query);
            finder = TreeDepthFirstTraverse(*m_Root, finder);
            return finder.GetNodes();
        }
    }
}


const double CPhyloTreeDataSource::GetNormDistance(void)
{
     return m_Calc.GetMaxDistance();
}


const double CPhyloTreeDataSource::GetMinDistance(void)
{
     return m_Calc.GetMinDistance();
}


void CPhyloTreeDataSource::RerootTree(CPhyloTreeNode * newRoot)
{
    if (newRoot->GetParent() && newRoot->Expanded()){
        CPhyloTreeNode * upStream   =
            static_cast<CPhyloTreeNode*>(newRoot->GetParent());

        CPhyloTreeNode * downStream =
            static_cast<CPhyloTreeNode*>(newRoot->GetParent()->DetachNode(newRoot));

        CPhyloTreeNode * convertedStream = x_ConvertUpstream(upStream);

        downStream->InsertNode(downStream->SubNodeBegin(), convertedStream);

        m_Root = downStream;

        x_MeasureTree();
    }
}


void CPhyloTreeDataSource::SetSubtree(CPhyloTreeNode * newRoot)
{
    if ((newRoot!=m_Root) && newRoot->GetParent()) {
        m_Root =
            static_cast<CPhyloTreeNode*>(newRoot->GetParent()->DetachNode(newRoot));
        x_MeasureTree();
    }
}


CPhyloTreeNode * CPhyloTreeDataSource::x_ConvertUpstream(CPhyloTreeNode * node)
{
    if (node->GetParent()){
        CPhyloTreeNode * parent = static_cast<CPhyloTreeNode*>(node->GetParent());
        node->GetParent()->DetachNode(node);
        x_ConvertUpstream(parent);
        node->InsertNode(node->SubNodeBegin(), parent);
    }
    return node;
}


void CPhyloTreeDataSource::ExpandCollapse(CPhyloTreeNode * node,
                                          IPhyGraphicsNode::TChildsDisplay dsp)
{
    if (node->CanExpandCollapse(dsp)){
        node->ExpandCollapse(dsp);
        x_MeasureTree();
    }
}

void CPhyloTreeDataSource::x_UpdateMarkers(CPhyloTreeNode * node)
{
    (*node)->GetMarkerColors().clear();

    std::string feature_str;
    float marker_size = 0.0f;
    float marker_count = 0.0f;

    for(CPhyloTreeNode::TNodeList_I  it = node->SubNodeBegin();
                                     it != node->SubNodeEnd(); it++ )  {      
        CPhyloTreeNode* n = static_cast<CPhyloTreeNode*>(*it);
        if ( (*n)->GetMarkerColors().size() > 0) {
            (*node)->GetMarkerColors().insert((*node)->GetMarkerColors().end(),
                                               (*n)->GetMarkerColors().begin(),
                                               (*n)->GetMarkerColors().end());

            ++marker_count;
            marker_size += (*n)->GetMarkerSize();
        }
    }

    // Take average marker size of children (in case they vary in size).
    (*node)->SetMarkerSize(marker_size * 1.0f/marker_count);

    (*node)->SetFeature("marker", (*node)->GetMarkerColorsAsString());

    if (node->GetParent() != NULL) {     
        x_UpdateMarkers(static_cast<CPhyloTreeNode*>(node->GetParent()));
    }
}

void CPhyloTreeDataSource::x_InitRenderModel()
{
#ifdef BUFFERED_RENDER
    if (m_Model.IsNull()) {
        m_Model.Reset(new CTreeGraphicsModel());
    }
    else {
        m_Model->ClearAll();
    }
    m_Model->Init();
#endif
}

void CPhyloTreeDataSource::DebugTrace(CPhyloTreeNode *node, int indent)
{
    std::string indent_str(indent, ' ');

    std::string parent_str = (node->GetParent() == NULL) ? "" : (*node->GetParent())->GetLabel();
    _TRACE(indent_str << (*node)->GetLabel() << " Parent: " << parent_str << " ID: " << (*node)->GetId());
    _TRACE(indent_str << "Children: ");

    CPhyloTreeNode::TNodeList_I  it;

    for(CPhyloTreeNode::TNodeList_I  it = node->SubNodeBegin();
                                     it != node->SubNodeEnd(); it++ )  {      
        CPhyloTreeNode* c = static_cast<CPhyloTreeNode*>(*it);
        DebugTrace(c, indent + 4);
    }
}

void CPhyloTreeDataSource::UpdateMarkers(CPhyloTreeNode * node)
{
    // Updates parent markers to share marker values of children.  Not currently
    // being used (may not be a required feature).
    if (node->GetParent() != NULL) {     
        x_UpdateMarkers(static_cast<CPhyloTreeNode*>(node->GetParent()));
    }
}

void CPhyloTreeDataSource::SimplifyForDepth(int depth)
{
    CPhyloTreeFilter_Depth selector(depth);
    selector = TreeDepthFirstTraverse(*m_Root, selector);
    m_Root = selector.GetTree();
    x_MeasureTree();
}


CPhyloTreeNode* CPhyloTreeDataSource::GetNode(int idx)
{
    visitor_node_finder finder(idx);
    finder = TreeDepthFirstTraverse(*m_Root, finder);
    return finder.GetNode();

}



void CPhyloTreeDataSource::ConvertSelectionToBiotree(CBioTreeDynamic & btree)
{
    CPhyloTreeFilter_Selector selector;
    selector = TreeDepthFirstTraverse(*m_Root, selector);

    // reindex tree
    TreeDepthFirstTraverse(*selector.GetTree(), CPhyloTreeCalculator());

    CPhyloTreeFilter_Export fexport;
    fexport = TreeDepthFirstTraverse(*selector.GetTree(), fexport);
    Save(btree, fexport.GetTree());
}


void CPhyloTreeDataSource::Clean(void)
{
    CPhyloTreeFilter_Export fexport;
    fexport = TreeDepthFirstTraverse(*m_Root, fexport);
    m_Root = fexport.GetTree();
    x_MeasureTree();
}

void CPhyloTreeDataSource::Filter(void)
{
    CPhyloTreeFilter_Selector selector;
    selector = TreeDepthFirstTraverse(*m_Root, selector);
    m_Root = selector.GetTree();
    x_MeasureTree();
}


void CPhyloTreeDataSource::FilterDistances(double x_dist)
{
    CPhyloTreeFilter_Distance dfilter(x_dist);
    dfilter = TreeDepthFirstTraverse(*m_Root, dfilter);
    m_Root = dfilter.GetTree();
    x_MeasureTree();
    Clean();
}


void CPhyloTreeDataSource::Copy(void)
{
    if (!m_Current) {
        _TRACE("CPhyloTreeDataSource::Copy - Nothing is selected");
        return;
    }

    _TRACE("CPhyloTreeDataSource::Copy - Not implemented");
    //m_Clipboard = m_Current;
}

void CPhyloTreeDataSource::Cut(void)
{
    if (!m_Current) {
        _TRACE("CPhyloTreeDataSource::Cut - Nothing is selected");
        return;
    }
    if (m_Current == m_Root) {
        _TRACE("CPhyloTreeDataSource::Cut - Removing root item is illegal");
        return;
    }
    if (m_sClipboard) {
        delete m_sClipboard;
        m_sClipboard = NULL;
    }


    if (m_Current->GetParent()){
        CPhyloTreeNode * upStream   =
            static_cast<CPhyloTreeNode*>(m_Current->GetParent());

        m_sClipboard =
            static_cast<CPhyloTreeNode*>(m_Current->GetParent()->DetachNode(m_Current));

        SetSelection(upStream, true,  true);


        /*
        if (!m_Current->IsLeaf()) {
            upStream->AddNode(**m_Clipboard);
        }
        */

        x_MeasureTree();
    }
}

void CPhyloTreeDataSource::Paste(void)
{
    if (!m_Current) {
        _TRACE("CPhyloTreeDataSource::Paste - Nothing is selected");
        return;
    }
    if (!m_sClipboard) {
        _TRACE("CPhyloTreeDataSource::Paste - Clipboard is empty");
        return;
    }

    int nmbInserted = 0;
    /*
    for(CPhyloTreeNode::TNodeList_I  it = m_Clipboard->SubNodeBegin();
                                         it != m_Clipboard->SubNodeEnd(); it++ )  {
        m_Current->InsertNode(m_Current->SubNodeBegin(), *it);
        nmbInserted++;
    }
    */
    if (nmbInserted == 0) {
        m_Current->InsertNode(m_Current->SubNodeBegin(), m_sClipboard);
    }

    m_sClipboard = NULL;
    x_MeasureTree();
}

void CPhyloTreeDataSource::NewNode(bool bAfter)
{
    if (!m_Current) {
        _TRACE("CPhyloTreeDataSource::NewNode - Nothing is selected");
        return;
    }

    if (!bAfter && !m_Current->GetParent()) {
        _TRACE("CPhyloTreeDataSource::NewNode - New root cannot be added");
        return;
    }

	CPhyloTreeNode* newnode =  new CPhyloTreeNode();

    // Set node id to current maximim node id value +1.  IDs needed during export.
    CPhyloTreeMaxIdCalculator next_id;
    next_id = TreeDepthFirstTraverse(*m_Root, next_id);
    (*newnode)->SetId(next_id.GetMaxId() + 1);    

    if (bAfter) { // insert as a child
        m_Current->InsertNode(m_Current->SubNodeBegin(), newnode);
    }
    else {
        CPhyloTreeNode * parent =
            static_cast<CPhyloTreeNode*>(m_Current->GetParent());

        CPhyloTreeNode* current =
            static_cast<CPhyloTreeNode*>(parent->DetachNode(m_Current));

        parent->InsertNode(parent->SubNodeBegin(), newnode);
        newnode->InsertNode(newnode->SubNodeBegin(), current);
    }
    
	// Set dictionary and a blank feature table with all values from dictionary.
    newnode->GetValue().SetDictionaryPtr(&m_Dictionary);
    ITERATE(CBioTreeFeatureDictionary::TFeatureDict, it, m_Dictionary.GetFeatureDict()) {
        newnode->GetValue().SetFeature(it->second, "");
    }

    (*newnode)->Init();
    x_MeasureTree();
}


void CPhyloTreeDataSource::Remove(bool bSubtree)
{
    if (!m_Current) {
        _TRACE("CPhyloTreeDataSource::Remove - Nothing is selected");
        return;
    }

    if (m_Current == m_Root) {
        _TRACE("CPhyloTreeDataSource::Remove - Removing root item is illegal");
        return;
    }

    if (bSubtree) {
        CPhyloTreeNode * junk =
            static_cast<CPhyloTreeNode*>(m_Current->GetParent()->DetachNode(m_Current));

        m_Current = NULL;

        delete junk;


    }
    else {
        CPhyloTreeNode * parent =
            static_cast<CPhyloTreeNode*>(m_Current->GetParent());



        for(CPhyloTreeNode::TNodeList_I  it = m_Current->SubNodeBegin();
                                         it != m_Current->SubNodeEnd(); it++ )  {
            parent->InsertNode(parent->SubNodeBegin(), *it);
        }
        m_Current->RemoveAllSubNodes(CPhyloTreeNode::eNoDelete);
        parent->RemoveNode(m_Current);
        m_Current = parent;
    }

    x_MeasureTree();
}

void CPhyloTreeDataSource::MoveNode(bool bUp)
{
    if (!m_Current) {
        _TRACE("CPhyloTreeDataSource::MoveNode - Nothing is selected");
        return;
    }

    if (m_Current == m_Root) {
        _TRACE("CPhyloTreeDataSource::MoveNode - Nowhere to move root");
        return;
    }

    CPhyloTreeNode * parent =
            static_cast<CPhyloTreeNode*>(m_Current->GetParent());


    for(CPhyloTreeNode::TNodeList_I  it = parent->SubNodeBegin();
                                     it != parent->SubNodeEnd(); it++ )  {
        if (*it == m_Current) {
            if (bUp && (it!=parent->SubNodeBegin())) {
                CPhyloTreeNode::TNodeList_I pre = it--;
                swap(*pre, *it);
                break;
            }
            else {
                CPhyloTreeNode::TNodeList_I last = parent->SubNodeEnd(); last--;
                if (it != last) {
                    CPhyloTreeNode::TNodeList_I pst = it++;
                    swap(*pst, *it);
                    break;
                }
            }
        }
    }
}

bool CPhyloTreeDataSource::HasClusters(void)
{
    return !m_Calc.GetClusters().empty();
}

void CPhyloTreeDataSource::ApplyAttributes(CBioTreeAttrReader::TAttrTable & attrs)
{
    m_Calc.Init();
    m_Calc.SetAttrTable(attrs);
    m_Calc = TreeDepthFirstTraverse(*m_Root, m_Calc);
    Clusterize();
}

string CPhyloTreeDataSource::GetColumnLabel(size_t col) const
{ 
    const CBioTreeFeatureDictionary:: TFeatureDict& fdict = 
        m_Dictionary.GetFeatureDict();
    CBioTreeFeatureDictionary::TFeatureDict::const_iterator fiter;

    size_t count = 0;
    for (fiter=fdict.begin(); fiter != fdict.end(); ++fiter, ++count) {
        if (count == col) {
            return (*fiter).second; 
        }
    }

    return "";
}

size_t CPhyloTreeDataSource::GetColsCount() const  
{ 
    return m_Dictionary.GetFeatureDict().size(); 
}

CQueryExec* CPhyloTreeDataSource::GetQueryExec(bool casesensitive)
{
    CTreeQueryExec* qexec = new CTreeQueryExec(&(this->GetDictionary()));

    qexec->SetRootNode(this->GetTree());

    // Logical operators:
    qexec->AddFunc(CQueryParseNode::eAnd,
            new CQueryFuncPromoteLogic());
    qexec->AddFunc(CQueryParseNode::eOr,
            new CQueryFuncPromoteLogic());
    qexec->AddFunc(CQueryParseNode::eSub,
            new CQueryFuncPromoteLogic());
    qexec->AddFunc(CQueryParseNode::eXor,
            new CQueryFuncPromoteLogic());
    qexec->AddFunc(CQueryParseNode::eNot,
            new CQueryFuncPromoteLogic());

    // Constants:
    qexec->AddFunc(CQueryParseNode::eIntConst,
            new CQueryFuncPromoteValue());
    qexec->AddFunc(CQueryParseNode::eFloatConst,
            new CQueryFuncPromoteValue());
    qexec->AddFunc(CQueryParseNode::eBoolConst,
            new CQueryFuncPromoteValue());
    qexec->AddFunc(CQueryParseNode::eString,
            new CQueryFuncPromoteValue());

    NStr::ECase cs = casesensitive ? NStr::eCase  : NStr::eNocase;

    // Comparison operators:
    qexec->AddFunc(CQueryParseNode::eEQ,
            new CQueryFuncPromoteEq(cs));
    qexec->AddFunc(CQueryParseNode::eGT,
            new CQueryFuncPromoteGtLt(CQueryParseNode::eGT, cs));
    qexec->AddFunc(CQueryParseNode::eGE,
            new CQueryFuncPromoteGtLt(CQueryParseNode::eGE, cs));
    qexec->AddFunc(CQueryParseNode::eLT,
            new CQueryFuncPromoteGtLt(CQueryParseNode::eLT, cs));
    qexec->AddFunc(CQueryParseNode::eLE,
            new CQueryFuncPromoteGtLt(CQueryParseNode::eLE, cs));
    qexec->AddFunc(CQueryParseNode::eIn,
            new CQueryFuncPromoteIn(cs));
    qexec->AddFunc(CQueryParseNode::eBetween,
            new CQueryFuncPromoteBetween(cs));
    qexec->AddFunc(CQueryParseNode::eLike,
            new CQueryFuncLike(cs));

    return qexec;
}

void CPhyloTreeDataSource::SetQueryResults(CQueryExec* q)
{
    CTreeQueryExec *e = dynamic_cast<CTreeQueryExec*>(q);

    m_SearchCache =  e->GetTreeSelected();
    m_SearchCurrentNode = NULL;
}

string CPhyloTreeDataSource::ExecuteStringQuery(string query, 
                                                bool casesensitive)
{
    // Re-run the query if options that can change query results have changed
    m_SearchCache.clear();
    m_SearchCurrentNode = NULL;

    // simple string queries are completed synchronously since
    // they can never get hung up in network access.
    if (query != "")
        m_SearchCache = FindNodes(query, casesensitive);

    LOG_POST(Info << "String Query: " << query);

    return "";
}

CPhyloTreeNode* CPhyloTreeDataSource::IterateOverSelNodes(int direction, 
                                                          bool highlight)
{
    // highlight
    SetSelection(GetTree(), false, true, false);

    if (m_SearchCache.empty())
        return NULL;

    // cursor
    if (direction==0) {
        if (!m_SearchCurrentNode)
            m_SearchCurrentNode = (*m_SearchCache.begin());
    }
    else if (direction>0) {
        if (!m_SearchCurrentNode) {
            m_SearchCurrentNode = (*m_SearchCache.begin());
        }
        else {
            vector<TTreeType*>::iterator itt = 
                find(m_SearchCache.begin(), m_SearchCache.end(), m_SearchCurrentNode);

            // If not found (unexpected) set cursor to beginning.
            if (itt==m_SearchCache.end()) {
                m_SearchCurrentNode =  (*m_SearchCache.begin());
            }
            // else increment unless we are already at the end.
            else if (++itt!=m_SearchCache.end()) {
                m_SearchCurrentNode = *itt;
            }
        }
    }
    else {
        if (!m_SearchCurrentNode) {
            m_SearchCurrentNode = *(m_SearchCache.begin() + (m_SearchCache.size()-1));
        }
        else {
            vector<TTreeType*>::iterator itt = 
                find(m_SearchCache.begin(), m_SearchCache.end(), m_SearchCurrentNode);

            // If not found (unexpected) set cursor to the end
            if (itt==m_SearchCache.end()) {
                m_SearchCurrentNode = *(m_SearchCache.begin() + (m_SearchCache.size()-1));
            }
            // else decrement cursor if not already at the beginning
            else if (itt!=m_SearchCache.begin()) {
                m_SearchCurrentNode = *(--itt);
            }
        }
    }

    // highlight
    if (highlight) {
        ITERATE(vector<CPhyloTreeNode*>, it, m_SearchCache) {
            SetSelection(*it, true, false, true);
        }
    }
    else {
        if (m_SearchCurrentNode) {
            SetSelection(m_SearchCurrentNode, true, false, true);
        }
    }

    return m_SearchCurrentNode;
}


END_NCBI_SCOPE
