/*  $Id: phylo_tree_label.cpp 27010 2012-12-07 16:37:16Z falkrb $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Vladimir Tereshkov
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>
#include <gui/widgets/phylo_tree/phylo_tree_label.hpp>
#include <gui/widgets/phylo_tree/phylo_tree_node.hpp>

#include <gui/opengl/glutils.hpp>

BEGIN_NCBI_SCOPE

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CPhyloTreeLabel::CPhyloTreeLabel()
{
    m_Format = "";
}

CPhyloTreeLabel::~CPhyloTreeLabel()
{
}

#ifndef TEXTURED_FONTS
const CGlBitmapFont *  CPhyloTreeLabel::GetCurrFont(void)
#else
const CGlTextureFont *  CPhyloTreeLabel::GetCurrFont(void)
#endif
{
    return &m_SL->GetFont(); 
}


void CPhyloTreeLabel::GetNodeLabelOffset(CPhyloTreeNode* node,
                                         IPhyloTreeLOD* lod,
                                         CVect2<float>& scalable_offset,
                                         CVect2<float>& pixel_offset)
{
    scalable_offset.X() = 0.0;
    scalable_offset.Y() = 0.0;

    pixel_offset.X() = 0.0f;
    pixel_offset.Y() = 0.0f;

    float defNodeSize = (float)lod->GetDefaultNodeSize(node);

#ifndef TEXTURED_FONTS
    const CGlBitmapFont* bestFont = &m_SL->GetFont(); 
#else
    const CGlTextureFont* bestFont = &m_SL->GetFont(); 
#endif

    
    // we have place to show it
    if (bestFont) {
        string text   = x_GetLabel(node);
        int text_h = (int)bestFont->GetMetric(IGlFont::eMetric_CharHeight);
        int text_w = (int)bestFont->TextWidth(text.c_str());

        bool bLeft = ((**node).GetAngle() < -1.57f ||  (**node).GetAngle()>1.57f);

        if (!bLeft) {
            // 2 pixels visible distance from node          
            if (lod->IsPrimitivesScalable()){
                scalable_offset.X() = defNodeSize;
                pixel_offset.X() = 2.0f;
            }
            else {
                pixel_offset.X() = defNodeSize + 2.0f;
            }

            pixel_offset.Y() = (float)(-text_h/2);
        }
        else {           
            if (lod->IsPrimitivesScalable()){
                scalable_offset.X() = -defNodeSize;
                pixel_offset.X() = (float)(-(2 + text_w));
            }
            else {              
                pixel_offset.X() = (float)(-(defNodeSize + 5 + text_w)); 
            }

            pixel_offset.Y() = (float)(-text_h/2);
        }
    }
}

CGlRect<double> CPhyloTreeLabel::GetRect(CGlPane& pane,
                                         CPhyloTreeNode * node,
                                         bool bLeft,
                                         IPhyloTreeLOD * lod)
{
    CGlRect<double>  boundBox(0,0,0,0);

    TVPUnit defNodeSize = lod->GetNodeSize(node);

//    CGlBitmapFont *  bestFont =
//        GetBestFont(pane.ProjectX(pane.GetVisibleRect().Left() +
//                                 defNodeSize), lod);    
#ifndef TEXTURED_FONTS
    const CGlBitmapFont* bestFont = &m_SL->GetFont(); 
#else
    const CGlTextureFont* bestFont = &m_SL->GetFont(); 
#endif

    
    // we have place to show it
    if (bestFont) {
        string text   = x_GetLabel(node);
        double text_h = bestFont->GetMetric(IGlFont::eMetric_CharHeight);
        double text_w = bestFont->TextWidth(text.c_str());

        if (!bLeft) {
            // space between node and right edge
            double space_x = pane.ProjectX(pane.GetVisibleRect().Right() - (**node).X() +
                                            pane.GetVisibleRect().Left());

            // 2 pixels visible distance from node
            double pos_x;
            if (lod->IsPrimitivesScalable()){
                pos_x= (**node).X() +
                            defNodeSize +
                            2 * pane.GetScaleX()/pane.GetScaleY();;
            }
            else {
                pos_x = (**node).X() + (defNodeSize + 2)*pane.GetScaleX();
            }

            double pos_y  = (**node).Y() - text_h*pane.GetScaleY()/2;


            if (space_x > 0) {
                boundBox.Init(pos_x,
                              pos_y,
                              pos_x + text_w * pane.GetScaleX(),
                              pos_y + text_h * pane.GetScaleY());
            }
        }
        else {
            double pos_x;
            if (lod->IsPrimitivesScalable()){
                pos_x =  (**node).X() -
                            defNodeSize -
                            text_w * pane.GetScaleX() -
                            2 * pane.GetScaleX()/pane.GetScaleY();
            }
            else {
                pos_x = (**node).X() -
                        (defNodeSize + 5)* pane.GetScaleX() -
                          text_w*pane.GetScaleX();
            }

            double pos_y  = (**node).Y() - text_h*pane.GetScaleY()/2;

            boundBox.Init(pos_x,
                          pos_y,
                          pos_x + text_w * pane.GetScaleX(),
                          pos_y + text_h * pane.GetScaleY());
        }
    }

    return boundBox;
}

CVect2<float> CPhyloTreeLabel::GetSize(CPhyloTreeNode * node)
{
#ifndef TEXTURED_FONTS
    const CGlBitmapFont* bestFont = &m_SL->GetFont(); 
#else
    const CGlTextureFont* bestFont = &m_SL->GetFont(); 
#endif

    string label = x_GetLabel(node);

    CVect2<float> labelSize((float)bestFont->TextWidth(label.c_str()),
                            (float)bestFont->GetMetric(IGlFont::eMetric_CharHeight));

    return labelSize;
}


bool CPhyloTreeLabel::Render(CGlPane& pane,
                             CPhyloTreeNode * node,
                             bool bLeft,
                             IPhyloTreeLOD * lod)
{
    TVPUnit defNodeSize = lod->GetNodeSize(node);

#ifndef TEXTURED_FONTS
    const CGlBitmapFont* bestFont = &m_SL->GetFont(); 
#else
    const CGlTextureFont* bestFont = &m_SL->GetFont(); 
#endif

    // we have place to show it
    if (bestFont) {
        string text   = x_GetLabel(node);
        double text_h = bestFont->GetMetric(IGlFont::eMetric_CharHeight);
        double text_w = bestFont->TextWidth(text.c_str());

        //double vis_left = pane.GetVisibleRect().Left();

        CGlRect<double> rect = GetRect(pane, node, bLeft, lod);

        IGlFont::ETruncate  truncate;
        if (m_SL->GetLabelTruncation() == CPhyloTreeScheme::eLabelsTruncated)
            truncate = IGlFont::eTruncate_Ellipsis;
        else 
            truncate = IGlFont::eTruncate_None;


        if (!bLeft) {
		    // 2 pixels visible distance from node
            double pos_x;
            if (lod->IsPrimitivesScalable()){
                pos_x= (**node).X() +
                            defNodeSize +
                            2 * pane.GetScaleX()/pane.GetScaleY();
            }
            else {
                pos_x = (**node).X() + (defNodeSize + 2)*pane.GetScaleX();
            }

            // space between node and right edge
            double space_x = pane.ProjectX(pane.GetVisibleRect().Right() - pos_x +
                                            pane.GetVisibleRect().Left());

            // space between left edge and label
            // space_x is the space between right edge and node 
            double delta_x = defNodeSize + (rect.Right() - pos_x) / pane.GetScaleX();

            space_x = delta_x < space_x ?  delta_x : space_x;

            // for internal nodes
            if (!node->IsLeafEx() && !lod->CanDragNodes()) {
                // finding child to get distance, nonlinear also supported this way
                CPhyloTreeNode::TNodeList_I fst = node->SubNodeBegin();
                double hole_x = ((***fst).X() - (**node).X()) / pane.GetScaleX();
            
                space_x = hole_x < space_x ?  hole_x : space_x;
            }
           
            double pos_y  = (**node).Y() - text_h*pane.GetScaleY()/2;

            // Don't draw text outside of window area (slows things down)
            bool outside = false;
            if ( (pos_y + text_h*pane.GetScaleY()) < pane.GetVisibleRect().Bottom() ||
                 (pos_y - text_h*pane.GetScaleY()) > pane.GetVisibleRect().Top() ||
                 (pos_x + text_w*pane.GetScaleX()) < pane.GetVisibleRect().Left() ||
                 (pos_x > pane.GetVisibleRect().Right()) )
                outside = true;           

            if (space_x > 0 && !outside) {                                
                bestFont->TextOut(pos_x, pos_y, pos_x + space_x, pos_y + text_h, text.c_str(),
                                  IGlFont::eAlign_Left, truncate);                
            }
        }
        else {
            double pos_x;
            if (lod->IsPrimitivesScalable()){
                pos_x =  (**node).X() -
                            defNodeSize -
                            text_w * pane.GetScaleX() -
                            2 * pane.GetScaleX()/pane.GetScaleY();
            }
            else {
                pos_x = (**node).X() -
                        (defNodeSize + 5 )* pane.GetScaleX() -
                          text_w*pane.GetScaleX();
            }

            double pos_y  = (**node).Y() - text_h*pane.GetScaleY()/2;

            // Don't draw text outside of window area (slows things down)
            bool outside = false;
            if ( (pos_y + text_h*pane.GetScaleY()) < pane.GetVisibleRect().Bottom() ||
                 (pos_y - text_h*pane.GetScaleY()) > pane.GetVisibleRect().Top() ||
                 (pos_x + text_w*pane.GetScaleX()) < pane.GetVisibleRect().Left() ||
                 (pos_x > pane.GetVisibleRect().Right()) )
                outside = true;  
                 
            if (!outside) {
                bestFont->TextOut(pos_x, pos_y, pos_x + text_w + 1, pos_y + text_h,
                                  text.c_str(), IGlFont::eAlign_Left, truncate);
            }
        }
    }
    return (bestFont!=NULL);

}

string CPhyloTreeLabel::x_GenerateLabel(CPhyloTreeNode * node)
{
    string  retVal = m_Format;
    const   CBioTreeFeatureDictionary & dict = *(**node).GetDictionaryPtr();


    ITERATE(CBioTreeFeatureDictionary::TFeatureDict, it, dict.GetFeatureDict()) {
        string sKey = "$(" + it->second + ")";
        string sVal =  (**node).GetBioTreeFeatureList().GetFeatureValue(it->first);

        string::size_type keyPos = string::npos;
        while ((keyPos = retVal.find(sKey))!=string::npos){
            retVal.replace(keyPos, sKey.size(), sVal);
        }
    }

    // clean up any non-substituted variables
    size_t srt_var = string::npos;
    size_t end_var = string::npos;
    do {
        srt_var = retVal.find("$(");
        if (srt_var!=string::npos) {
            end_var = retVal.find(")", srt_var);
            if (end_var!=string::npos) {
                retVal = retVal.erase(srt_var, end_var);
            }
        }
    } while (srt_var!=string::npos && end_var!=string::npos);

    return retVal!=m_Format ? retVal : "";
}


string CPhyloTreeLabel::x_GenerateAutoLabel(CPhyloTreeNode * node,
                                            CPhyloTreeScheme::TAutoLabels lbl)
{
    string  autoLabel;

    switch (lbl){
    case CPhyloTreeScheme::eAlNone:
    {
        autoLabel = "";
        break;
    }
    case CPhyloTreeScheme::eAlPhylip:
    {
        int ilabel;

        if (!node->IsLeaf()) {
            ilabel = (**node).GetPamlCounter() + 1;
        }
        else {
            int ibase = node->GetRoot()->GetValue().GetNodesNmb() -
                        node->GetRoot()->GetValue().GetLeavesNmb();
            ilabel = ibase + node->GetValue().IDX().second + 1;
        }

        autoLabel = NStr::IntToString(ilabel);
        break;
    }
    case CPhyloTreeScheme::eAlPaml:
    {
        /*
        autoLabel = NStr::IntToString(
            node->GetRoot()->GetValue().GetLeavesNmb() + (**node).GetCounter() + 1);
            */
        int ilabel;

        if (node->IsLeaf()) {
            ilabel = node->GetValue().IDX().second;
        }
        else if (node->GetParent()==NULL) {
            ilabel =  node->GetValue().GetLeavesNmb() + 1;
        }
        else {
            ilabel = node->GetRoot()->GetValue().GetLeavesNmb() + (**node).GetPamlCounter() + 1;
        }
        autoLabel = NStr::IntToString(ilabel);
        break;
    }
    default:
        autoLabel = "";
        break;
    }

    return autoLabel;
}


const string & CPhyloTreeLabel::x_GetLabel(CPhyloTreeNode * node)
{
    m_CurrLabel = (**node).GetLabel();

    // adding auto-label
    string autoLabel = x_GenerateAutoLabel(node, m_SL->GetAutoLabels());
    if (autoLabel.length()) {
        if (m_CurrLabel.length()) {
            m_CurrLabel = " | " + m_CurrLabel;
        }
        m_CurrLabel = autoLabel + m_CurrLabel;
    }

    if (!(*node).Expanded()) {
        //label += (" (" + NStr::IntToString((**node).CountSubnodes()) + " nodes)");
        if (m_CurrLabel.length()) {
            m_CurrLabel += " | ";
        }
        m_CurrLabel += (NStr::IntToString((**node).GetLeavesNmb()) + " leaves");
    }

    return m_CurrLabel;
}

void CPhyloTreeLabel::SetScheme(CPhyloTreeScheme & sl)
{
    m_SL.Reset(&sl);
    SetLabelFormat(sl.SetLabelFormat());
}

void CPhyloTreeLabel::SetLabelFormat(const string fmt)
{
    if (!fmt.empty() && (m_Format!=fmt)){
        m_Format = fmt;
    }
}

const string CPhyloTreeLabel::GetLabelForNode(CPhyloTreeNode * node,
                                              const string & format)
{
    string saved_format = m_Format;
    m_Format = format;
    string ready_label  = x_GenerateLabel(node);
    m_Format = saved_format;
    return ready_label;
}

TVPUnit CPhyloTreeLabel::GetMinVerticalSeparation() const
{
    TVPUnit h = -1;

#ifndef TEXTURED_FONTS
    const CGlBitmapFont* font = &m_SL->GetFont(); 
#else
    const CGlTextureFont* font = &m_SL->GetFont(); 
#endif

    if (font != NULL) {
        h = ((TVPUnit) font->TextHeight()) + 1;
    }

    return h;
}

bool CPhyloTreeLabel::IsVisible(IPhyloTreeLOD * lod)
{
    TModelUnit h = GetMinVerticalSeparation();

    if (h > 0) {
        TVPUnit visibleDistance = lod ? lod->DistanceBetweenNodes() :
                                 (int) lod->GetScheme().SetSize(CPhyloTreeScheme::eNodeSize);
        return visibleDistance >= h;
    }
    
    return false;
}

END_NCBI_SCOPE
