/*  $Id: phylo_tree_node.cpp 27010 2012-12-07 16:37:16Z falkrb $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Vladimir Tereshkov
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>
#include <gui/widgets/phylo_tree/phylo_tree_node.hpp>
#include <gui/widgets/phylo_tree/phylo_tree_boundary_shapes.hpp>

BEGIN_NCBI_SCOPE

#define PHYLO_FEAT_LABEL     "label"
#define PHYLO_FEAT_DIST      "dist"
#define PHYLO_FEAT_SEQID     "seq-id"
#define PHYLO_FEAT_SEQTITLE  "seq-title"
#define PHYLO_FEAT_TAXID     "tax-id"
#define PHYLO_FEAT_CLUSTERID "cluster-id"
#define PHYLO_FEAT_EVISIBLE  "branch-visible"
#define PHYLO_NODE_MARKER    "marker"


#define PHYLO_NODE_COLOR     "$NODE_COLOR"
#define PHYLO_EDGE_COLOR     "$EDGE_COLOR"
#define PHYLO_NODE_COLLAPSED "$NODE_COLLAPSED"
#define PHYLO_NODE_BOUNDED   "$NODE_BOUNDED"

#define PHYLO_LABEL_COLOR    "$LABEL_COLOR"
#define PHYLO_LABEL_BGCOLOR  "$LABEL_BG_COLOR"

#define PHYLO_NODE_SIZE      "$NODE_SIZE"
#define PHYLO_NODE_SHAPE     "$NODE_SHAPE"

#define MAX_PROPERTIES       255

const IPhyGraphicsNode::TColor IPhyGraphicsNode::TNoColor = "";

 
IPhyGraphicsNode::IPhyGraphicsNode(const IPhyGraphicsNode& n)
: m_XY(n.m_XY)
, m_IDX(n.m_IDX)
, m_Angle(n.m_Angle)
, m_AnnWedge(n.m_AnnWedge)
, m_LabelPos(n.m_LabelPos)
, m_LabelRect(n.m_LabelRect)
, m_Selected(n.m_Selected)
, m_Children(n.m_Children)
, m_Bounded(n.m_Bounded)
, m_Dirty(n.m_Dirty)
, m_NodeEdColor(n.m_NodeEdColor)
, m_NodeFgColor(n.m_NodeFgColor)
, m_ClusterEdColor(n.m_ClusterEdColor)
, m_ClusterFgColor(n.m_ClusterFgColor)
, m_LabelFgColor(n.m_LabelFgColor)
, m_LabelBgColor(n.m_LabelBgColor)
, m_MarkerColors(n.m_MarkerColors)
, m_MarkerSize(n.m_MarkerSize)
, m_Boundary(NULL)
, m_Size(n.m_Size)
{
    if (n.m_Boundary != NULL) {
        m_Boundary = new CSubtreeBoundary(*n.m_Boundary);
    }
}

IPhyGraphicsNode::~IPhyGraphicsNode()
{
    delete m_Boundary;
    m_Boundary = NULL;
}

   
IPhyGraphicsNode& IPhyGraphicsNode::operator=(const IPhyGraphicsNode& n)
{
    m_XY = n.m_XY;
    m_IDX = n.m_IDX;
    m_Angle = n.m_Angle;
    m_AnnWedge = n.m_AnnWedge;
    m_LabelPos = n.m_LabelPos;
    m_LabelRect = n.m_LabelRect;
    m_Selected = n.m_Selected;
    m_Children = n.m_Children;
    m_Bounded = n.m_Bounded;
    m_Dirty = n.m_Dirty;
    m_NodeEdColor = n.m_NodeEdColor;
    m_NodeFgColor = n.m_NodeFgColor;
    m_ClusterEdColor = n.m_ClusterEdColor;
    m_ClusterFgColor = n.m_ClusterFgColor;
    m_LabelFgColor = n.m_LabelFgColor;
    m_LabelBgColor = n.m_LabelBgColor;
    m_MarkerColors = n.m_MarkerColors;
    m_MarkerSize = n.m_MarkerSize;
    m_Size = n.m_Size;

    delete m_Boundary;
    m_Boundary = NULL;

    if (n.m_Boundary != NULL) {
        m_Boundary = new CSubtreeBoundary(*n.m_Boundary);
    }

    return *this;
}


string IPhyGraphicsNode::x_GetParameter(const string& features, const string& parm)
{
    string parm_lower = parm;
    parm_lower = NStr::ToLower(parm_lower) + "=";

    string features_lower = features;
    features_lower = NStr::ToLower(features_lower);

    size_t idx = features_lower.find(parm_lower);
    if (idx != string::npos) {
        size_t end_idx = features.find_first_of(" ,\t", idx);
        if (end_idx == string::npos)
            end_idx = features.size();

        return features.substr(idx + parm_lower.length(), end_idx-(idx+parm_lower.length()));
    }

    return "";
}

void IPhyGraphicsNode::SetMarkerColors(const string& str) 
{
    // marker colors should be in format:
    // [r1,g1,b1,a1][r2,g2,b2,a2]....[rn,gn,bn,an], n = (1..n)

    string remainder = str;

    string marker_size_str = x_GetParameter(str, "size");
    float marker_size = 2.0f;
    if (marker_size_str != "") {
        marker_size = NStr::StringToDouble(marker_size_str, NStr::fConvErr_NoThrow);
        if (marker_size <= 0.0f) marker_size = 2.0f;
    }
    SetMarkerSize(marker_size);
    
    string::size_type idx = remainder.find_first_of(']');

    while (idx != string::npos) {
        string color = remainder.substr(0,idx+1);
        size_t start_idx = remainder.find_first_of('[');
        if (start_idx == string::npos)
            return;
        color = color.substr(start_idx, color.length()-start_idx);

        m_MarkerColors.push_back(CRgbaColor(color));

        remainder = remainder.substr(idx+1, remainder.size()-(idx+1));
        idx = remainder.find_first_of(']');
    }        
}

string IPhyGraphicsNode::GetMarkerColorsAsString()
{
    string c;

    for (size_t i=0; i<m_MarkerColors.size(); ++i) {
        // Convert to string format [r,g,b] where r,g,b are 0..255.
        c += string("[") + m_MarkerColors[i].ToString(false,true) +  std::string("] ");
    }

    return c;
}
   
CSubtreeBoundary* IPhyGraphicsNode::GetSubtreeBoundary() 
{ 
    if (m_Boundary == NULL)
        m_Boundary = new CSubtreeBoundary();

    return m_Boundary; 
}

void IPhyGraphicsNode::DeleteSubtreeBoundary() 
{ 
    delete m_Boundary; 
    m_Boundary = NULL;
}

void CPhyloTreeNode::Sort(bool bOrder)
{
    //m_Nodes.sort(predicate_sort_by_leaves_count(bOrder));
    m_Nodes.sort(bOrder ? x_Greater : x_NotGreater);
}

void CPhyloTreeNode::SortByState()
{
    m_Nodes.sort(x_Brighter);
}


CPhyloTreeNode::CPhyloTreeNode()
{       
  
}


CPhyloTreeNode::CPhyloTreeNode(int x_id, double x_dist, const string & x_label)
{
   GetValue().Set(x_id, x_dist, x_label);
}

bool CPhyloTreeNode::IsVisible()
{
    TTreeType * node = this;
    TTreeType * parent;

    while ((parent=node->GetParent())!=NULL) {
        if ((**parent).GetChildsDisplay()==IPhyGraphicsNode::eHideChilds){
            return false;
        }
        node = parent;
    }
    return true;
}

void CPhyTreeNode::SetFeature(const string & name, const string & value)
{
    TBioTreeFeatureId   id;

    if (m_Dictionary->HasFeature(name)){
        id = m_Dictionary->GetId(name);
    }
    else { // find smallest available id
        for (id=0; id < MAX_PROPERTIES; id++){
            if (!m_Dictionary->HasFeature(id)) {

                // intentionally only single write access point to dictionary
                CBioTreeFeatureDictionary * rwDict = 
                    const_cast<CBioTreeFeatureDictionary *>(m_Dictionary);
                rwDict->Register(id, name);
                break;
            }
        }
        if (id==MAX_PROPERTIES) {
             NCBI_THROW(CException, eUnknown,
                        "CPhyloTreeNode::SetFeature - maximum number of features exceeded");
        }
    }
    m_Features.SetFeature(id, value);
}


void CPhyTreeNode::InitFeatures()
{
    const CBioTreeFeatureList & bfl = m_Features;

    const string& dist = bfl.GetFeatureValue(m_Dictionary->GetId(PHYLO_FEAT_DIST));
    try {
        if (!dist.empty()) {
            double ddist = NStr::StringToDouble(dist);
            SetDistance(ddist);
        }
    }
    catch(std::exception&) {}

    const string & size = bfl.GetFeatureValue(m_Dictionary->GetId(PHYLO_NODE_SIZE));
    try {
        if (!size.empty()) {
            int isize = NStr::StringToInt(size);
            SetNodeSize(isize);
        }
    }
    catch(std::exception&) {}

    GetMarkerColors().clear();
    const string & marker = bfl.GetFeatureValue(m_Dictionary->GetId(PHYLO_NODE_MARKER));
    try {
        if (!marker.empty()) {
            SetMarkerColors(marker);
        }
    }
    catch(std::exception&){}

    // cluster ID if any
    SetClusterID( bfl.GetFeatureValue(m_Dictionary->GetId(PHYLO_FEAT_CLUSTERID)));
    
    // collapsed/expand node
    const string & ce = bfl.GetFeatureValue(m_Dictionary->GetId(PHYLO_NODE_COLLAPSED));
    if (ce == "1") SetChildsDisplay(IPhyGraphicsNode::eHideChilds);
    if (ce == "0") SetChildsDisplay(IPhyGraphicsNode::eShowChilds);
    // don't touch node expand/collapse status otherwise!

    // node (and subtree) has bounding region (or not)
    const string & bounded = bfl.GetFeatureValue(m_Dictionary->GetId(PHYLO_NODE_BOUNDED));
    if (bounded == "") {
        SetBoundedDisplay(IPhyGraphicsNode::eUnBounded);
    }
    else {
        SetBoundedDisplay(IPhyGraphicsNode::eBounded);        
        GetSubtreeBoundary()->CreateShapes(bounded);
    }


    // lets keep label to manual entry
    //SetLabel(       bfl.GetFeatureValue(m_Dictionary->GetId(PHYLO_FEAT_LABEL)));
    SetNodeColor(   bfl.GetFeatureValue(m_Dictionary->GetId(PHYLO_NODE_COLOR)),
                    bfl.GetFeatureValue(m_Dictionary->GetId(PHYLO_EDGE_COLOR)));

    SetLabelColor(  bfl.GetFeatureValue(m_Dictionary->GetId(PHYLO_LABEL_COLOR)),
                    bfl.GetFeatureValue(m_Dictionary->GetId(PHYLO_LABEL_BGCOLOR)));

}

void CPhyTreeNode::Init()
{
    const CBioTreeFeatureList &bfl = m_Features;

    const string& sid = bfl.GetFeatureValue(m_Dictionary->GetId(PHYLO_FEAT_SEQID));
    try {
        if (!sid.empty()) {
            SetSeqID(new objects::CSeq_id(sid));
        }
    }

    catch(CException& )
    {
        // try ad-hoc GI extraction (for misformed seq-id strings like: "gi|55823257|ref|YP_141698.1"
        // (data coming from some other ad-hoc scripts)
        //
        try {
            SIZE_TYPE pos = NStr::FindNoCase(sid, "gi|");
            if (pos != NPOS) {
                string gi_str = "gi|";
                for (unsigned i = pos + 3; i < sid.length(); ++i) {
                    char c = sid.at(i);
                    if (isdigit(c)) {
                        gi_str.append(1, c);
                    } else {
                        break;
                    }
                }
                SetSeqID(new objects::CSeq_id(gi_str));
            }
            else {
                string lcl_sid = "lcl|";
                lcl_sid.append(sid);
                SetSeqID(new objects::CSeq_id(lcl_sid));
            }

        }
        catch(std::exception&)
        {
        }

    }
    catch(std::exception&)
    {
    }

    InitFeatures();
}

void  CPhyTreeNode::Sync()
{
    string tmp;
    NStr::DoubleToString(tmp, GetDistance());

    SetFeature(PHYLO_FEAT_DIST,     tmp);
    SetFeature(PHYLO_FEAT_LABEL,    GetLabel());
    SetFeature(PHYLO_NODE_COLLAPSED,
               GetChildsDisplay()==IPhyGraphicsNode::eShowChilds?"0":"1");
    SetFeature(PHYLO_FEAT_CLUSTERID, GetClusterID());
}

void  CPhyloTreeNode::ExpandCollapse(IPhyGraphicsNode::TChildsDisplay x_chds)
{
    GetValue().SetChildsDisplay(x_chds);
    GetValue().Sync();
}

void  CPhyloTreeNode::SetLabel(const string & x_label)
{
    GetValue().SetLabel(x_label);
    GetValue().Sync();
}

bool CPhyloTreeNode::CanExpandCollapse(IPhyGraphicsNode::TChildsDisplay x_chds)
{
    // don't do expand/cpllapse on root
    if (GetRoot() == this) {
        return false;
    }
    else if (x_chds == IPhyGraphicsNode::eShowChilds) {
        return !Expanded();
    }
    else if (x_chds == IPhyGraphicsNode::eHideChilds) {
        return !IsLeafEx();
    }
    return true;
}

void CPhyloTreeNode::HowBigAmI()
{
    size_t nsize1 = sizeof(CPhyloTreeNode);
    size_t nsize2 = sizeof(CTreeNode<CPhyTreeNode>);
    size_t nsize3 = sizeof(CPhyTreeNode);
    size_t nsize4 = sizeof(IPhyGraphicsNode);
    size_t nsize5 = sizeof(CPhyNewickNode);
    size_t nsize6 = sizeof(IPhyNode);
    size_t nsize7 = sizeof(float);
    size_t nsize8 = sizeof(pair<int,int>);
    size_t nsize9 = sizeof(IPhyGraphicsNode::TSelectedState);
    size_t nsize10 = sizeof(string); // 32 debug, 28 in release
    size_t nsize11 = sizeof(CRef<objects::CSeq_id>); //(4 (ref only..)
    size_t nsize12 = sizeof(CBioTreeFeatureList);  //(16)

    LOG_POST("Node Sizes: " <<
        "\nsizeof(CPhyloTreeNode):          " << nsize1 << 
        "\nsizeof(CTreeNode<CPhyTreeNode>): " << nsize2 <<
        "\nsizeof(CPhyTreeNode):            " << nsize3 <<
        "\nsizeof(IPhyGraphicsNode):        " << nsize4 <<
        "\nsizeof(CPhyNewickNode):          " << nsize5 <<
        "\nsizeof(IPhyNode):                " << nsize6 <<
        "\nsizeof(float):                   " << nsize7 <<
        "\nsizeof(pair<int,int>):           " << nsize8 <<
        "\nsizeof(IPhyGraphicsNode::TSelectedState): " << nsize9 <<
        "\nsizeof(string):                  " << nsize10 <<
        "\nsizeof(CRef<objects::CSeq_id>):  " << nsize11 <<
        "\nsizeof(CBioTreeFeatureList):     " << nsize12);
}

END_NCBI_SCOPE
