/*  $Id: phylo_tree_query_exec.cpp 24336 2011-09-15 17:23:58Z falkrb $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Bob Falk
 *
 * File Description:
 *      Implementation for tree query execution
 */

#include <ncbi_pch.hpp>

#include <gui/widgets/phylo_tree/phylo_tree_query_exec.hpp>
#include <gui/widgets/phylo_tree/phylo_tree_node.hpp>

#include <gui/objutils/query_func_promote.hpp>

#include <objects/seq/seq_id_handle.hpp>

#include <corelib/ncbiobj.hpp>
#include <corelib/ncbistd.hpp>

BEGIN_NCBI_SCOPE


/******************************************************************************
 *
 * CPhylQueryExec Implementation
 *
 *****************************************************************************/ 

TBioTreeFeatureId 
CTreeQueryExec::GetFeatureIdNoCase(const string& feature_name,
                                   const CBioTreeFeatureDictionary* dict) const
{
    const CBioTreeFeatureDictionary::TFeatureDict& fd = dict->GetFeatureDict();
    CBioTreeFeatureDictionary::TFeatureDict::const_iterator it;

    for(it = fd.begin(); it != fd.end(); ++it) {
        int cmp = NStr::CompareNocase(feature_name.c_str(), 
                                      (*it).second.c_str());
        if (cmp==0)
            return it->first;
    }

    return (TBioTreeFeatureId)-1;
}

bool CTreeQueryExec::ResolveIdentifier(const std::string& identifier, 
                                        bool& value) 
{ 
    const CBioTreeFeatureDictionary* dict = 
        m_EvalNode->GetValue().GetDictionaryPtr();
    TBioTreeFeatureId id = GetFeatureIdNoCase(identifier, dict);

    if (id !=(TBioTreeFeatureId)-1) {
        const CBioTreeFeatureList& f_list = 
            m_EvalNode->GetValue().GetBioTreeFeatureList();

        std::string feature_value = f_list[id];
        try {
            value = NStr::StringToBool(feature_value);
            return true;
        }
        catch (CStringException&) {
            return false;
        }
    }

    return false; 
}

bool CTreeQueryExec::ResolveIdentifier(const std::string& identifier, 
                                        int& value) 
{ 
    const CBioTreeFeatureDictionary* dict = 
        m_EvalNode->GetValue().GetDictionaryPtr();
    TBioTreeFeatureId id = GetFeatureIdNoCase(identifier, dict);

    if (id != (TBioTreeFeatureId)-1) {
        const CBioTreeFeatureList& f_list = 
            m_EvalNode->GetValue().GetBioTreeFeatureList();
        std::string feature_value = f_list[id];

        try {
            value = NStr::StringToInt(feature_value);
            return true;
        }
        catch (CStringException&) {
            return false;
        }
    }

    return false; 
}

bool CTreeQueryExec::ResolveIdentifier(const std::string& identifier, 
                                        double& value) 
{ 
    const CBioTreeFeatureDictionary* dict = 
        m_EvalNode->GetValue().GetDictionaryPtr();
    TBioTreeFeatureId id = GetFeatureIdNoCase(identifier, dict);

    if (id != (TBioTreeFeatureId)-1) {
        const CBioTreeFeatureList& f_list = 
            m_EvalNode->GetValue().GetBioTreeFeatureList();
        std::string feature_value = f_list[id];

        try {
            value = NStr::StringToDouble(feature_value);
            return true;
        }
        catch (CStringException&) {
            return false;
        }
    }

    return false; 
}

bool CTreeQueryExec::ResolveIdentifier(const std::string& identifier, 
                                        std::string& value) 
{ 
    const CBioTreeFeatureDictionary* dict = 
        m_EvalNode->GetValue().GetDictionaryPtr();
    TBioTreeFeatureId id = GetFeatureIdNoCase(identifier, dict);

    if (id != (TBioTreeFeatureId)-1) {
        const CBioTreeFeatureList& f_list = 
            m_EvalNode->GetValue().GetBioTreeFeatureList();

        value = f_list[id];

        return true;
    }

    return false;
}

bool CTreeQueryExec::HasIdentifier(const std::string& identifier) 
{ 
    TBioTreeFeatureId id = GetFeatureIdNoCase(identifier, m_Dictionary);

    return (id != (TBioTreeFeatureId)-1);
}


void CTreeQueryExec::EvalStart()
{
    m_ExceptionCount = 0;
    m_QueriedCount = 0;
    m_EvalNode = m_RootNode;
    m_NodeStack.push(PhyloTreePointer(m_RootNode));
}

bool CTreeQueryExec::EvalComplete()
{
    return m_NodeStack.empty();
}

bool CTreeQueryExec::EvalNext(CQueryParseTree& qtree) {
    bool selected = false;

    try {
        ++m_QueriedCount;
        Evaluate(qtree);

        CQueryParseTree::TNode* top_node = qtree.GetQueryTree();
        CQueryNodeValue*  v = dynamic_cast<CQueryNodeValue*>(
            top_node->GetValue().GetUserObject());

        if (v != NULL) {
            if (v->GetValue()) {
                m_Selected.push_back(m_EvalNode);
                selected = true;
            }
        }
    }
    catch(CQueryParseException &pe) {
        // Could avoid flooding log since many may be identical, but
        // maybe the last one is the one you need to see (for debugging...)
        LOG_POST("Query parse error during execution: " << pe.GetMsg());
        ++m_ExceptionCount;
    }
    catch(CQueryExecException &ex) {
        // Could avoid flooding log since many may be identical, but
        // maybe the last one is the one you need to see (for debugging...)
        LOG_POST(Info << "Query execution error: " << ex.GetMsg());
        ++m_ExceptionCount;
    }

    PhyloTreePointer p = m_NodeStack.top();

    while (p.m_Iterator == p.m_Node->SubNodeEnd() && !m_NodeStack.empty()) {
        m_NodeStack.pop();
        if (!m_NodeStack.empty())
            p = m_NodeStack.top();
    }

    if (!m_NodeStack.empty()) {
        m_EvalNode = static_cast<CPhyloTreeNode*>(*(p.m_Iterator));
        ++(m_NodeStack.top().m_Iterator);
        m_NodeStack.push(PhyloTreePointer(m_EvalNode));
    }

    return selected;
}


END_NCBI_SCOPE
