/*  $Id: phylo_tree_radial.cpp 27503 2013-02-26 18:16:55Z falkrb $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Vladimir Tereshkov
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>
#include <corelib/ncbistl.hpp>

#include <gui/widgets/phylo_tree/phylo_tree_radial.hpp>

#include <gui/widgets/phylo_tree/phylo_tree_render.hpp>
#include <gui/widgets/phylo_tree/phylo_tree_ds.hpp>
#include <gui/opengl/glhelpers.hpp>
#include <gui/graph/igraph_utils.hpp>


#include <math.h>

BEGIN_NCBI_SCOPE


CPhyloRadial::CPhyloRadial()
{
}

CPhyloRadial::CPhyloRadial(double w, double h):
    IPhyloTreeRenderer(w, h)
{
}


CPhyloRadial::~CPhyloRadial()
{
}

void  CPhyloRadial::x_Layout(CPhyloTreeDataSource& ds)
{
    Int4 leafs = ds.GetSize();
    Int4 width = ds.GetWidth();

    m_xStep = m_DimX /  width;
    m_yStep = m_DimY / ((leafs>1)?(leafs-1):2);


    m_Radius = sqrt(m_xStep*m_xStep + m_yStep*m_yStep)/2;
    m_StartDegree = 0.0;
    m_ParentDegree = 0.0;

    if (ds.GetNormDistance() > 0){
        //m_NormDistance = m_Radius / ds.GetNormDistance(false);
         m_NormDistance = m_DimX / ds.GetNormDistance();
    }

    CBoundaryPoints boundary_pts;
    x_Calculate(ds.GetTree(), boundary_pts);

    // changing raster - x_Calculate for shure changed size in unpredictable way
    TModelRect newRect = ds.GetBoundRect();

    m_RasterRect.Init(newRect.Left(),
                      newRect.Bottom(),
                      newRect.Right(),
                      newRect.Top());

    m_pPane->SetModelLimitsRect(m_RasterRect);
    m_pPane->SetVisibleRect(m_RasterRect);

    // Set drawing options that are specific to radial layout
#ifdef BUFFERED_RENDER
    CGlVboNode* edge_node = ds.GetModel().FindGeomNode("TreeEdges");
    CGlVboNode* narrow_edge_node = ds.GetModel().FindGeomNode("NarrowTreeEdges");

    if (edge_node != NULL && narrow_edge_node != NULL) {   
        edge_node->GetState().Enable(GL_LINE_SMOOTH);
        edge_node->GetState().Enable(GL_BLEND);
        edge_node->GetState().BlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);        

        narrow_edge_node->GetState().Enable(GL_LINE_SMOOTH);
        narrow_edge_node->GetState().Enable(GL_BLEND);
        narrow_edge_node->GetState().BlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
        narrow_edge_node->GetState().ColorC(CRgbaColor(1.0f, 1.0f, 1.0f, 0.7f));
    }

    CGlVboNode* filler_points_node = m_DS->GetModel().FindGeomNode("FillerPoints");
    if (filler_points_node != NULL)
        filler_points_node->SetVisible(false);
#endif

    m_ValidLayout = eValidLayout;
    ComputeViewingLimits(*m_pPane, false);
    //m_pPane->SetVisibleRect(m_pPane->GetModelLimitsRect());
}


void  CPhyloRadial::x_Render(CGlPane& pane, CPhyloTreeDataSource& ds)
{
    pane.OpenOrtho();
    x_DrawBoundingAreas(ds.GetTree());
    x_DrawTree( ds.GetTree() );
    x_DrawNodes( ds.GetTree() );
    pane.Close();
}

void  CPhyloRadial::x_RenderVbo(CPhyloTreeDataSource& ds)
{   
    CGlVboNode* edge_node = m_DS->GetModel().FindGeomNode("TreeEdges");
    CGlVboNode* narrow_edge_node = m_DS->GetModel().FindGeomNode("NarrowTreeEdges");
    CGlVboNode* point_node = m_DS->GetModel().FindGeomNode("NodePoints");

    _ASSERT(edge_node != NULL && narrow_edge_node != NULL);

    GLsizei edge_count = (GLsizei)ds.GetNumEdges();
    edge_node->GetColorBuffer().BufferDataResize(edge_count*2, NULL, GL_DYNAMIC_DRAW);

    if (edge_count > 0) {    
        edge_node->GetColorBuffer().MapBuffer(GL_WRITE_ONLY);
        x_DrawTreeVbo(ds.GetTree());
        edge_node->GetColorBuffer().UnmapBuffer();

        narrow_edge_node->SetDefaultPosition();
        narrow_edge_node->SetVisible(true);

        edge_node->SetDefaultPosition();
        edge_node->SetVisible(true);
    }
    else {
        edge_node->SetVisible(false);
    }

    GLsizei node_count = (GLsizei)ds.GetNumNodes();
    point_node->GetColorBuffer().BufferDataResize(node_count, NULL, GL_DYNAMIC_DRAW);
   
    if (node_count > 0) {    
        point_node->GetColorBuffer().MapBuffer(GL_WRITE_ONLY);

        x_DrawNodesVbo(ds.GetTree(), point_node);

        point_node->GetColorBuffer().UnmapBuffer();   
        point_node->ClearPositions();
        point_node->SetVisible(true);
    }
    else {
        point_node->SetVisible(false);
    }
}

void CPhyloRadial::x_Calculate(CPhyloTreeNode* node,                                           
                               CBoundaryPoints& parent_boundary_pts,                     
                               bool compute_boundary)
{
    if ((**node).GetBoundedDisplay() == IPhyGraphicsNode::eBounded)
        compute_boundary = true;

    CBoundaryPoints node_boundary_pts;

    if (!node->GetParent()){ // root
        (**node).X() = m_DimX / 2;
        (**node).Y() = m_DimY / 2;
        (**node).SetAngle(0.);
        (**node).SetAnnWedge(6.28);
    }
    else {
        (**node).X()  = node->GetParent()->GetValue().X();
        (**node).Y() = node->GetParent()->GetValue().Y();

        (**node).SetAnnWedge( node->GetParent()->GetValue().GetAnnWedge() *
                              ((**node).GetLeavesNmbEx()+1) /
                              (node->GetParent()->GetValue().GetLeavesNmbEx()+1));


        // annulus wedge degree
        // double awd   =  min(node->GetAnnWedge(),
        //               acos((m_Radius *  (node->IDX().first-1)) / (m_Radius * node->IDX().first)) * 2 );
        // that is for sure good place for thinking
        double awd   =  (**node).GetAnnWedge();

        // start degree
        CPhyloTreeNode * pp = static_cast<CPhyloTreeNode*>(node->GetParent());
        if (node == *pp->SubNodeBeginEx()){
            m_StartDegree =
                node->GetParent()->GetValue().GetAngle() -
                    node->GetParent()->GetValue().GetAnnWedge()/2.0;
        }

        // setting angle and position for this vertex
        double angle = m_StartDegree + (awd / 2.0);

        (**node).SetAngle(angle);

        if (m_bDistMode){
            m_Radius = (**node).GetDistance() * m_NormDistance;
        }
        // When adjacent leaves have dist==0, then they are written on top
        // of each other in distance-mode.  This forces them to have a mininma
        // distnace (1 node diameter which is 2*node size).  But this is also
        // not correct, so we'll leave it out for now.  User can always choose
        // not to use distances.
        //m_Radius = std::max(m_Radius, 
        //    2.0*(double)GetDefaultNodeSize(node)*(double)m_pPane->GetScaleX() );

        (**node).X()  += m_Radius * cos(angle);
        (**node).Y() += m_Radius * sin(angle);

        m_StartDegree += awd;
    }

    if (!node->IsLeaf())  {
        for(CPhyloTreeNode::TNodeList_I  it = node->SubNodeBeginEx();
            it != node->SubNodeEnd(); it++ )  {
            CPhyloTreeNode* subnode = static_cast<CPhyloTreeNode*>(*it);
            x_Calculate(subnode, node_boundary_pts, compute_boundary);
        }
    }

    x_ComputeNodeBoundary(node, compute_boundary, node_boundary_pts, "Radial");

    // Each node that computes a boundary has to add in the points (and text rectangles)
    // for all nodes below it in the tree
    if (compute_boundary) {
        parent_boundary_pts.AddBoundedPoints(node_boundary_pts);
    }
}

void CPhyloRadial :: x_DrawTree(CPhyloTreeNode * node)
{
    // drawing leaf node
    if (node->GetParent()){
        x_RenderLine(node->GetParent()->GetValue().X(),
                     node->GetParent()->GetValue().Y(),
                     node->GetValue().X(),
                     node->GetValue().Y(),
                     (**node).GetSelectedState(),
                     GetEdgeColor(node));
    }

    for(CPhyloTreeNode::TNodeList_I  it = node->SubNodeBeginEx();
        it != node->SubNodeEnd(); it++ )  x_DrawTree(static_cast<CPhyloTreeNode*>(*it));

    //x_RenderNode(node);
}

void CPhyloRadial::x_DrawTreeVbo(CPhyloTreeNode * node)
{
    // drawing leaf node
    CGlVboNode* edge_node = m_DS->GetModel().FindGeomNode("TreeEdges");
    _ASSERT(edge_node != NULL);

    if (node->GetParent()){
        x_RenderLineVbo(edge_node,
                        node->GetParent()->GetValue().X(),
                        node->GetParent()->GetValue().Y(),
                        node->GetValue().X(),
                        node->GetValue().Y(),
                        (**node).GetSelectedState(),
                        GetEdgeColor(node));
    }

    for(CPhyloTreeNode::TNodeList_I  it = node->SubNodeBeginEx();
        it != node->SubNodeEnd(); it++ )  x_DrawTreeVbo(static_cast<CPhyloTreeNode*>(*it));
}

bool CPhyloRadial::CanDragNodes(void)
{
    return true;
}

string CPhyloRadial::GetDescription(void)
{
    return "Radial Tree";
}

END_NCBI_SCOPE
