/*  $Id: phylo_tree_rect_cladogram.cpp 27503 2013-02-26 18:16:55Z falkrb $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Vladimir Tereshkov
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>
#include <corelib/ncbistl.hpp>

#include <gui/widgets/phylo_tree/phylo_tree_rect_cladogram.hpp>
#include <gui/widgets/phylo_tree/phylo_tree_ds.hpp>
#include <gui/opengl/glhelpers.hpp>
#include <gui/graph/igraph_utils.hpp>
#include <gui/widgets/gl/attrib_menu.hpp>

#include <math.h>

BEGIN_NCBI_SCOPE


CPhyloRectCladogram::CPhyloRectCladogram()
{
}

CPhyloRectCladogram::CPhyloRectCladogram(double w, double h):
    IPhyloTreeRenderer(w, h)
{
}


CPhyloRectCladogram::~CPhyloRectCladogram()
{
}


void  CPhyloRectCladogram::x_Layout(CPhyloTreeDataSource& ds)
{
    Int4 leafs = ds.GetSize();
    Int4 width = ds.GetWidth();

    if (width>0){
        m_xStep = m_DimX /  width;
    }
    else {
        m_xStep = m_DimX /  2;
    }

    if (leafs>1){
        m_yStep = m_DimY / leafs;
    }
    else {
        m_yStep = m_DimY / 2;
    }

    if (ds.GetNormDistance() > 0){
        m_NormDistance = m_DimX /  ds.GetNormDistance();
    }

    CBoundaryPoints boundary_pts;
    x_Calculate(ds.GetTree(), boundary_pts);

    TModelRect newRect = ds.GetBoundRect();

    // single leaf
    if (newRect.Top() == newRect.Bottom()) {
        newRect.SetTop(m_DimY);
        newRect.SetBottom(0);
    }

    m_RasterRect.Init(newRect.Left(),  newRect.Bottom(),
                      newRect.Right(), newRect.Top());

    // !!Hack for now to handle shapes since labels only can be computed after
    // !!first visible area computed.
    m_pPane->SetModelLimitsRect(m_RasterRect);
    m_pPane->SetVisibleRect(m_RasterRect);

    // Set drawing options that are specific to rectcladogram
#ifdef BUFFERED_RENDER
    CGlVboNode* edge_node = ds.GetModel().FindGeomNode("TreeEdges");
    CGlVboNode* narrow_edge_node = ds.GetModel().FindGeomNode("NarrowTreeEdges");
    CGlVboNode* filler_points_node = m_DS->GetModel().FindGeomNode("FillerPoints");

    if (edge_node != NULL && 
        narrow_edge_node != NULL &&
        filler_points_node != NULL) { 
        if (!m_bUseSplines){ 
            // Lines will be aligned to pixels so no smoothing:        
            edge_node->GetState().Disable(GL_LINE_SMOOTH);
            edge_node->GetState().Enable(GL_BLEND);
            edge_node->GetState().BlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
            filler_points_node->SetVisible(true);

            narrow_edge_node->GetState().Disable(GL_LINE_SMOOTH);
            narrow_edge_node->GetState().Enable(GL_BLEND);
            narrow_edge_node->GetState().BlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
            //narrow_edge_node->GetState().Disable(GL_BLEND); 
        }
        else {
            // Curved lines - need to smooth them
            edge_node->GetState().Enable(GL_LINE_SMOOTH);
            edge_node->GetState().Enable(GL_BLEND);
            edge_node->GetState().BlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);    

            filler_points_node->SetVisible(false);

            narrow_edge_node->GetState().Enable(GL_LINE_SMOOTH);
            narrow_edge_node->GetState().Enable(GL_BLEND);
            narrow_edge_node->GetState().BlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
            narrow_edge_node->GetState().ColorC(CRgbaColor(1.0f, 1.0f, 1.0f, 0.7f));
        }
    }
#endif

    m_ValidLayout = eValidLayout;
    ComputeViewingLimits(*m_pPane, false);
}

void  CPhyloRectCladogram::x_Render(CGlPane& pane, CPhyloTreeDataSource& ds)
{   
    pane.OpenOrtho();
    x_DrawBoundingAreas(ds.GetTree());
    x_DrawTree(ds.GetTree());
    x_DrawNodes(ds.GetTree());
    pane.Close();
}

void  CPhyloRectCladogram::x_RenderVbo(CPhyloTreeDataSource& ds)
{      
    CGlVboNode* edge_node = m_DS->GetModel().FindGeomNode("TreeEdges");
    CGlVboNode* narrow_edge_node = m_DS->GetModel().FindGeomNode("NarrowTreeEdges");
    CGlVboNode* point_node = m_DS->GetModel().FindGeomNode("NodePoints");
    _ASSERT(edge_node != NULL && narrow_edge_node != NULL);

    GLsizei node_count = (GLsizei)ds.GetNumNodes();
    point_node->GetColorBuffer().BufferDataResize(node_count, NULL, GL_DYNAMIC_DRAW);

    if (node_count > 0) {        
        point_node->GetColorBuffer().MapBuffer(GL_WRITE_ONLY);    

        if (m_bUseSplines)
            x_RenderVboSplines(ds);
        else
            x_RenderVboPseudoSplines(ds);

        narrow_edge_node->SetDefaultPosition();
        narrow_edge_node->SetVisible(true);
   
        edge_node->SetDefaultPosition();
        edge_node->SetVisible(true);

        point_node->GetColorBuffer().UnmapBuffer();
        point_node->GetColorBuffer().SetActiveBufferSize(m_DS->GetModel().GetNodes().size());
        point_node->ClearPositions();
        point_node->SetVisible(true);
    }
    else {
        point_node->SetVisible(false);
    }
}

void  CPhyloRectCladogram::x_RenderVboSplines(CPhyloTreeDataSource& ds)
{
    CGlVboNode* edge_node = m_DS->GetModel().FindGeomNode("TreeEdges");
    _ASSERT(edge_node != NULL);

    CGlVboNode* point_node = m_DS->GetModel().FindGeomNode("NodePoints");
    _ASSERT(point_node != NULL);

    if (ds.GetNumEdges() > 0) {
        vector<CRgbaColor> edge_colors;
        x_DrawSplineTree(ds.GetTree(), edge_colors, point_node);

        edge_node->GetColorBuffer().BufferData(edge_colors.size(), &edge_colors[0], GL_DYNAMIC_DRAW);
        edge_node->SetVisible(true);
    }
    else {
        edge_node->SetVisible(false);
        // Render single node:
        if (ds.GetTree() != NULL)
            x_RenderNodeVbo(ds.GetTree(), point_node);
    }
}

void  CPhyloRectCladogram::x_RenderVboPseudoSplines(CPhyloTreeDataSource& ds)
{
    CGlVboNode* edge_node = m_DS->GetModel().FindGeomNode("TreeEdges");
    _ASSERT(edge_node != NULL);

    CGlVboNode* point_node = m_DS->GetModel().FindGeomNode("NodePoints");
    _ASSERT(point_node != NULL);

    // The pseudo spline is drawn as 2 lines per edge
    GLsizei edge_count = (GLsizei)ds.GetNumEdges()*2;
    edge_node->GetColorBuffer().BufferDataResize(edge_count*2, NULL, GL_DYNAMIC_DRAW);    
    //edge_node->GetColorBuffer()->SetActiveBufferSize(edge_count*2);

    if (edge_count > 0) {        
        edge_node->GetColorBuffer().MapBuffer(GL_WRITE_ONLY);

        x_DrawPseudoSplineTree(ds.GetTree(), point_node, edge_node);

        edge_node->GetColorBuffer().UnmapBuffer();
        edge_node->SetVisible(true);
    }
    else {
        edge_node->SetVisible(false);

        // Render single node:
        if (ds.GetTree() != NULL)
            x_RenderNodeVbo(ds.GetTree(), point_node);
    }
}


// Could probably pass bool down so that points only handled on
// branches that neeed them.
void CPhyloRectCladogram::x_Calculate(CPhyloTreeNode * node,                                           
                                      CBoundaryPoints& parent_boundary_pts,                     
                                      bool compute_boundary)
{    
    (**node).SetAngle(0.0);

    if ((**node).GetBoundedDisplay() == IPhyGraphicsNode::eBounded)
        compute_boundary = true;

    CBoundaryPoints node_boundary_pts;


//    bool bLeft = ((**node).GetAngle() < -1.57 ||  (**node).GetAngle()>1.57);

    if (node->IsLeafEx()) {
        if (m_bDistMode){
            (**node).X()  =
                m_NormDistance * (**node).GetDistFromRoot();
        }
        else {
            (**node).X()  = m_DimX;   // all leafs on the right
        }

        (**node).Y() = (m_DimY - (**node).IDX().second * m_yStep);     // y coordinate
        (**node).SetAngle(0); // for correct labeling
    }
    else {       
        for(CPhyloTreeNode::TNodeList_I  it = node->SubNodeBeginEx();
            it != node->SubNodeEndEx(); it++)  {
            CPhyloTreeNode* subnode = static_cast<CPhyloTreeNode*>(*it);
            x_Calculate(subnode, node_boundary_pts, compute_boundary);
        }

        CPhyloTreeNode::TNodeList_I fst = node->SubNodeBegin();
        CPhyloTreeNode::TNodeList_I lst = node->SubNodeEnd();
        lst--;

        double yf = (***fst).XY().Y();
        double yl = (***lst).XY().Y();

        if (m_bDistMode){
            (**node).XY().X()  =
                m_NormDistance * (**node).GetDistFromRoot();
        }
        else {
            (**node).XY().X()  = ((**node).IDX().first * m_xStep);
        }

        (**node).XY().Y() = (yl + (yf - yl)/2);
    }
    
    x_ComputeNodeBoundary(node, compute_boundary, node_boundary_pts, "RectCladogram");

    // Each node that computes a boundary has to add in the points (and text rectangles)
    // for all nodes below it in the tree
    if (compute_boundary) {
        parent_boundary_pts.AddBoundedPoints(node_boundary_pts);
    }

    return;
}

void CPhyloRectCladogram:: x_DrawTree(CPhyloTreeNode * node)
{
    CPhyloTreeNode sorted_node;

    for (CPhyloTreeNode::TNodeList_I it = node->SubNodeBeginEx();        
         it!=node->SubNodeEndEx(); it++){

        sorted_node.AddNode(***it);
    }
    sorted_node.SortByState();

    for (CPhyloTreeNode::TNodeList_I it = sorted_node.SubNodeBegin();
            it!=sorted_node.SubNodeEnd(); it++){

        TPoint pt1((float)(**node).X(),
                    (float)(**node).Y(), 0);
        TPoint pt2((float)(**node).X(),
                    (float)(***it).Y(), 0);
        TPoint pt3((float)(***it).X(),
                    (float)(***it).Y(), 0);

        IPhyGraphicsNode::TSelectedState beg_state =
            (**node).GetSelectedState();

        IPhyGraphicsNode::TSelectedState end_state =
            (***it).GetSelectedState();

        IPhyGraphicsNode::TSelectedState state = end_state;

        if (beg_state != end_state &&
            (state==IPhyGraphicsNode::eSelected || state==IPhyGraphicsNode::eShared)){
            state = beg_state;
        }

        x_RenderSpline(pt1, pt2, pt2, pt3, state,
            GetEdgeColor((CPhyloTreeNode * )*it), !m_bUseSplines);
    }

    // detaching all subnodes from sorted node, so they will not be destroyed
    //sorted_node.RemoveAllSubNodes(CPhyloTreeNode::eNoDelete);

    // draw childs
    for(CPhyloTreeNode::TNodeList_I  it = node->SubNodeBeginEx();
        it != node->SubNodeEndEx();  it++ )  {
        x_DrawTree(static_cast<CPhyloTreeNode*>(*it));
    }
}

void CPhyloRectCladogram::x_DrawSplineTree(CPhyloTreeNode* node,
                                           vector<CRgbaColor>& edge_colors, 
                                           CGlVboNode* point_node)
{
    CPhyloTreeNode sorted_node;

    //??
    for (CPhyloTreeNode::TNodeList_I it = node->SubNodeBeginEx();        
         it!=node->SubNodeEndEx(); it++){

        sorted_node.AddNode(***it);
    }
    sorted_node.SortByState();    

    for (CPhyloTreeNode::TNodeList_I it = sorted_node.SubNodeBegin();
            it!=sorted_node.SubNodeEnd(); it++){

        TPoint pt1((float)(**node).X(),
                    (float)(**node).Y(), 0);
        TPoint pt2((float)(**node).X(),
                    (float)(***it).Y(), 0);
        TPoint pt3((float)(***it).X(),
                    (float)(***it).Y(), 0);

        IPhyGraphicsNode::TSelectedState beg_state =
            (**node).GetSelectedState();

        IPhyGraphicsNode::TSelectedState end_state =
            (***it).GetSelectedState();

        IPhyGraphicsNode::TSelectedState state = end_state;

        if (beg_state != end_state &&
            (state==IPhyGraphicsNode::eSelected || state==IPhyGraphicsNode::eShared)){
            state = beg_state;
        }

        x_RenderSplineVbo(edge_colors, 
                          pt1, pt2, pt2, pt3, 
                          state,
                          GetEdgeColor((CPhyloTreeNode * )*it));
    }

    // detaching all subnodes from sorted node, so they will not be destroyed
    //sorted_node.RemoveAllSubNodes(CPhyloTreeNode::eNoDelete);

    // draw children
    for(CPhyloTreeNode::TNodeList_I  it = node->SubNodeBeginEx();
        it != node->SubNodeEndEx();  it++ )  {
        x_DrawSplineTree(static_cast<CPhyloTreeNode*>(*it), edge_colors, point_node);
    }

    x_RenderNodeVbo(node, point_node);
}

void CPhyloRectCladogram::x_DrawPseudoSplineTree(CPhyloTreeNode* node, 
                                                 CGlVboNode* point_node, 
                                                 CGlVboNode* edge_node)
{
    CPhyloTreeNode sorted_node;

    // Maybe this could be replaced with a blending option that prioritizes based on
    // alpha value... (this may be 20% of buffered render)
    for (CPhyloTreeNode::TNodeList_I it = node->SubNodeBeginEx();        
         it!=node->SubNodeEndEx(); it++){

        sorted_node.AddNode(***it);
    }
    sorted_node.SortByState();

    for (CPhyloTreeNode::TNodeList_I it = sorted_node.SubNodeBegin();
            it!=sorted_node.SubNodeEnd(); it++){

        CVect2<float> mid_point((**node).X(), (***it).Y());

        IPhyGraphicsNode::TSelectedState beg_state =
            (**node).GetSelectedState();

        IPhyGraphicsNode::TSelectedState end_state =
            (***it).GetSelectedState();

        IPhyGraphicsNode::TSelectedState state = end_state;

        if (beg_state != end_state &&
            (state==IPhyGraphicsNode::eSelected || state==IPhyGraphicsNode::eShared)){
            state = beg_state;
        }

        x_RenderPseudoSplineVbo((CPhyloTreeNode*)*it, edge_node, (*node)->XY(), mid_point, (***it).XY(), state);
    }    

    // detaching all subnodes from sorted node, so they will not be destroyed
    //sorted_node.RemoveAllSubNodes(CPhyloTreeNode::eNoDelete);

    // draw children
    for(CPhyloTreeNode::TNodeList_I  it = node->SubNodeBeginEx();
        it != node->SubNodeEndEx();  it++ )  {
        x_DrawPseudoSplineTree(static_cast<CPhyloTreeNode*>(*it), point_node, edge_node);
    }

    x_RenderNodeVbo(node, point_node);
}

string CPhyloRectCladogram::GetDescription(void)
{
    return "Rectangle Cladogram";
}

TVPUnit CPhyloRectCladogram::DistanceBetweenNodes(void)
{
    return (TVPUnit)x_DistanceBetweenNodes(m_pPane->GetScaleY());
}

TModelUnit CPhyloRectCladogram::x_DistanceBetweenNodes(TModelUnit scaley)
{
    return  (GetDimY() / (((TModelUnit)m_DS->GetSize()) * scaley));
}

END_NCBI_SCOPE
