/*  $Id: phylo_tree_scheme.cpp 26700 2012-10-26 16:01:53Z falkrb $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Vladimir Tereshkov
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>

#include <gui/widgets/phylo_tree/phylo_tree_scheme.hpp>
#include <gui/widgets/phylo_tree/phylo_tree_exception.hpp>
#include <gui/widgets/wx/sys_path.hpp>
#include <gui/widgets/wx/wx_utils.hpp>
#include <gui/config/settings_set.hpp>
#include <gui/opengl/glfont.hpp>
#include <gui/types.hpp>

#include <objects/general/User_field.hpp>


BEGIN_NCBI_SCOPE

CPhyloTreeScheme::CPhyloTreeScheme(CGuiRegistry* registry)
 : m_Registry( registry ? registry : &CGuiRegistry::GetInstance())
 , m_CurrTheme("")
 , m_ThemeKey("Theme")
 , m_DefaultTheme("Default")
{
    x_DefaultScheme();
}

void CPhyloTreeScheme :: x_DefaultScheme(void)
{
    m_LabelFormat   = "$(label)";
    m_TooltipFormat = "Label: $(label) \nDistance: $(dist)";
    m_TopMargin = m_BottomMargin = m_LeftMargin =  m_RightMargin = 10;
    
    m_LineWidth = 3;
    m_NodeSize  = 3;
    m_MinNodeSize = m_NodeSize;
    m_MaxNodeSize = 8;
    m_MinChildsNmb = 1;
    m_MaxChildsNmb = 30;
    m_bBoaNodes = false;

    m_LabelVisibility = eLabelsForLeavesOnly;
    m_LabelTruncation = eLabelsTruncated;
    m_LabelStyle      = eFancyLabels;
    m_ZoomBehavior    = eZoomXY;
    m_AutoLabels      = eAlNone;
	m_Coloration	  = eClusters;
    m_SelectionVisibility = eNormalSelection;

    m_NonSelectedAlpha = 0.1f;

    m_BgColor.Set(1.0f, 1.0f, 1.0f, 1.0f);

    m_LineColor.Set(0.7f, 0.7f, 0.7f, 1.0f);
    m_LineSelColor.Set(0.0f, 0.0f, 1.0f, 1.0f);
    m_LineSharedColor.Set(0.3f, 0.0f, 0.9f, 1.0f);
    m_LineTraceColor.Set(0.0f, 0.8f, 0.0f, 1.0f);

    m_NodeColor.Set (0.4f, 0.4f, 0.4f, 1.0f);
    m_NodeSelColor.Set(1.0f, 0.0f, 0.0f, 1.0f);
    m_NodeSharedColor.Set(0.1f, 0.0f, 0.6f, 1.0f);
    m_NodeTraceColor.Set(0.0f, 0.4f, 0.0f, 1.0f);
    m_NodeTipSelColor.Set(1.0f, 0.0f, 0.0f, 0.25f);

    m_LabelColor.Set(0.0f, 0.0f, 0.0f, 1.0f);
    m_LabelSharedColor.Set(0.0f, 0.0f, 0.0f, 1.0f);
    m_LabelTraceColor.Set(0.0f, 0.0f, 0.0f, 1.0f);
    m_LabelSelColor.Set(1.0f, 0.0f, 0.0f, 1.0f);


    // monochrome colors
    m_BgColorMono.Set(1.0f, 1.0f, 1.0f, 1.0f);

    m_LineColorMono.Set(0.7f, 0.7f, 0.7f, 1.0f);
    m_LineSelColorMono.Set(0.8f, 0.8f, 0.8f, 1.0f);
    m_LineSharedColorMono.Set(0.3f, 0.3f, 0.3f, 1.0f);
    m_LineTraceColorMono.Set(0.3f, 0.3f, 0.3f, 1.0f);

    m_NodeColorMono.Set (0.3f, 0.3f, 0.3f, 1.0f);
    m_NodeSelColorMono.Set(0.5f, 0.5f, 0.5f, 1.0f);
    m_NodeSharedColorMono.Set(0.8f, 0.8f, 0.8f, 1.0f);
    m_NodeTraceColorMono.Set(0.1f, 0.1f, 0.1f, 1.0f);
    m_NodeTipSelColorMono.Set(0.4f, 0.4f, 0.4f, 0.25f);

    m_LabelColorMono.Set(0.0f, 0.0f, 0.0f, 1.0f);
    m_LabelSharedColorMono.Set(0.0f, 0.0f, 0.0f, 1.0f);
    m_LabelTraceColorMono.Set(0.0f, 0.0f, 0.0f, 1.0f);
    m_LabelSelColorMono.Set(0.2f, 0.2f, 0.2f, 1.0f);

#ifndef TEXTURED_FONTS
	m_Font.SetFont(CGlBitmapFont::eHelvetica8);
#else
    m_Font.SetFontSize(8);
    m_Font.SetFontFace(CGlTextureFont::eFontFace_Helvetica);
#endif
}

CRgbaColor& CPhyloTreeScheme::SetColor(TPhyloTreeObject obj,
                                       TPhyloTreeColors col)
{
    if (GetColoration() == eMonochrome)
        return x_SetColorMono(obj, col);


    switch (obj) {
    case eNode:
        {{
             switch (col) {
             case eColor:        return m_NodeColor;
             case eSelColor:     return m_NodeSelColor;
             case eTraceColor:   return m_NodeTraceColor;
             case eSharedColor:  return m_NodeSharedColor;
             case eTipSelColor:  return m_NodeTipSelColor;
             default: break;
             }
             break;
         }}

    case eLine:
        {{
             switch (col) {
             case eColor:        return m_LineColor;
             case eSelColor:     return m_LineSelColor;
             case eTraceColor:   return m_LineTraceColor;
             case eSharedColor:  return m_LineSharedColor;
             default: break;
             }
             break;
         }}
    case eLabel:
        {{
             switch (col) {
             case eColor:        return m_LabelColor;
             case eSelColor:     return m_LabelSelColor;
             case eTraceColor:   return m_LabelTraceColor;
             case eSharedColor:  return m_LabelSharedColor;
             default: break;
             }
             break;
         }}
    default:
        break;
    }

    // bgColor is one for all elements
    if (col == eBgColor) {
        return m_BgColor;
    }

    // throw error
    NCBI_THROW(CPhyloTreeException, eNoValue,
               "Unable to get/set specified value");
}

CRgbaColor& CPhyloTreeScheme::x_SetColorMono(TPhyloTreeObject obj,
                                           TPhyloTreeColors col)
{

    switch (obj) {
    case eNode:
        {{
             switch (col) {
             case eColor:        return m_NodeColorMono;
             case eSelColor:     return m_NodeSelColorMono;
             case eTraceColor:   return m_NodeTraceColorMono;
             case eSharedColor:  return m_NodeSharedColorMono;
             case eTipSelColor:  return m_NodeTipSelColorMono;
             default: break;
             }
             break;
         }}

    case eLine:
        {{
             switch (col) {
             case eColor:        return m_LineColorMono;
             case eSelColor:     return m_LineSelColorMono;
             case eTraceColor:   return m_LineTraceColorMono;
             case eSharedColor:  return m_LineSharedColorMono;
             default: break;
             }
             break;
         }}
    case eLabel:
        {{
             switch (col) {
             case eColor:        return m_LabelColorMono;
             case eSelColor:     return m_LabelSelColorMono;
             case eTraceColor:   return m_LabelTraceColorMono;
             case eSharedColor:  return m_LabelSharedColorMono;
             default: break;
             }
             break;
         }}
    default:
        break;
    }

    // bgColor is one for all elements
    if (col == eBgColor) {
        return m_BgColorMono;
    }

    // throw error
    NCBI_THROW(CPhyloTreeException, eNoValue,
               "Unable to get/set specified value");
}

GLdouble & CPhyloTreeScheme::SetSize(TPhyloTreeSizes obj)
{
    switch (obj) {
    case eLeftMargin:   return m_LeftMargin;
    case eTopMargin:    return m_TopMargin;
    case eRightMargin:  return m_RightMargin;
    case eBottomMargin: return m_BottomMargin;
    case eNodeSize:     return m_NodeSize;
    case eMinNodeSize:  return m_MinNodeSize;
    case eMaxNodeSize:  return m_MaxNodeSize;
    case eLineWidth:    return m_LineWidth;
    case eMinChildsNmb: return m_MinChildsNmb;
    case eMaxChildsNmb: return m_MaxChildsNmb;
    default:            break;
    };

    // throw error
    NCBI_THROW(CPhyloTreeException, eNoValue,
        "Unable to get/set specified value");
}

CGuiRegistry::CReadView 
CPhyloTreeScheme::x_GetReadView(const string& base_key,
                                const string& curr_key,
                                const string& def_key1)
{
    vector <string> keys;
    if ( !def_key1.empty()) {
        keys.push_back(def_key1);
    }
    if (curr_key != def_key1) {
        keys.push_back(curr_key);
    }
    return m_Registry->GetReadView(base_key, keys);
}

CGuiRegistry::CReadWriteView 
CPhyloTreeScheme::x_GetReadWriteView(const string& base_key,
                                     const string& curr_key,
                                     const string& def_key1)
{
    vector <string> keys;
    if ( !def_key1.empty()) {
        keys.push_back(def_key1);
    }
    if (curr_key != def_key1) {
        keys.push_back(curr_key);
    }
    return m_Registry->GetReadWriteView(base_key, keys);
}

bool CPhyloTreeScheme::x_UserFieldToColor(const CGuiRegistry::TReadView& view,
                                          const string& key, 
                                          CRgbaColor& c)
{
    CConstRef<objects::CUser_field> ref(view.GetField(key));
    if (!ref) {
        return false;
    }

    const objects::CUser_field& f = *ref;

    switch(f.GetData().Which()) {
    case objects::CUser_field::TData::e_Ints:
        {{
             vector<int> cvi = f.GetData().GetInts();
             if (cvi.size() > 2 ) {
                 vector<int>::const_iterator cv_it = cvi.begin();
                 c.SetRed  (static_cast<unsigned char>(*cv_it++));
                 c.SetGreen(static_cast<unsigned char>(*cv_it++));
                 c.SetBlue (static_cast<unsigned char>(*cv_it++));
                 if (cv_it != cvi.end()) {
                     c.SetAlpha(static_cast<unsigned char>(*cv_it));
                 }
             } else {
                 return false;
             }
             break;
         }}
    case objects::CUser_field::TData::e_Reals:   {{
         vector<double> cvd = f.GetData().GetReals();
         if (cvd.size() > 2 ) {
             vector<double>::const_iterator cv_it = cvd.begin();
             c.SetRed  (static_cast<float>(*cv_it++));
             c.SetGreen(static_cast<float>(*cv_it++));
             c.SetBlue (static_cast<float>(*cv_it++));
             if (cv_it != cvd.end()) {
                 c.SetAlpha(static_cast<float>(*cv_it));
             }
         } else {
             return false;
         }
         break;
     }}
    case objects::CUser_field::TData::e_Str: {{
         string cs = f.GetData().GetStr();
         c.FromString(cs);
         break;
     }}
    default:
        return false;
    }

    return true;
}


void CPhyloTreeScheme::x_ColorToUserField(CGuiRegistry::TReadWriteView& view,
                                          const string& key,
                                          const CRgbaColor& c)
{
    objects::CUser_field& f = *view.SetField(key);

    // store the color as a vector of ints.
    vector<int> val;
    val.push_back(c.GetRedUC());
    val.push_back(c.GetGreenUC());
    val.push_back(c.GetBlueUC());
    val.push_back(c.GetAlphaUC());

    f.SetData().SetInts() = val;
}

bool CPhyloTreeScheme::LoadCurrentSettings()
{
    try {

    CGuiRegistry::TReadView view = m_Registry->GetReadView(kPhyloBaseKey);

    // check  for a preset theme
    if ( m_CurrTheme.empty()  ||
        !view.HasField(m_ThemeKey + "." + m_CurrTheme) ) {
        m_CurrTheme = view.GetString("CurrentTheme", m_DefaultTheme);
    }

    // load theme settings
    view = x_GetReadView(kPhyloBaseKey + "." + m_ThemeKey,
        m_CurrTheme, m_DefaultTheme);
    if (m_CurrColorTheme.empty()) {
        m_CurrColorTheme = view.GetString("ColorTheme", "Color");
    }
    if (m_CurrSizeTheme.empty()) {
        m_CurrSizeTheme = view.GetString("SizeLevel", "Normal");
    }
    if (m_CurrGlobalTheme.empty()) {
        m_CurrGlobalTheme = view.GetString("GlobalSettings", "Default");
    }

    // Get key for sizes and font
    view = x_GetReadView(kPhyloBaseKey, 
                         "Size." + m_CurrSizeTheme, 
                         "Size.Normal");

    // load sizes
    m_LeftMargin = view.GetReal("LeftMargin", m_LeftMargin);
    m_RightMargin = view.GetReal("RightMargin", m_RightMargin);
    m_TopMargin = view.GetReal("TopMargin", m_TopMargin);
    m_BottomMargin = view.GetReal("BottomMargin", m_BottomMargin);
    m_NodeSize = view.GetReal("NodeSize", m_NodeSize);
    m_LineWidth = view.GetReal("LineWidth", m_LineWidth);
     
    // load font
    m_LabelFormat = view.GetString("LabelFormat", m_LabelFormat);
#ifndef TEXTURED_FONTS
    CGlBitmapFont::EFontFace font_face = CGlBitmapFont::FaceFromString(
                                    view.GetString("LabelFontFace", "Helvetica"));
                                    CGlBitmapFont::EFontSize font_size = CGlBitmapFont::SizeFromInt(
                                    view.GetInt("LabelFontSize", 10));

    m_Font.SetFontSize(font_size);
    m_Font.SetFontFace(font_face);
#else
    CGlTextureFont::EFontFace font_face = CGlTextureFont::FaceFromString(
                                    view.GetString("LabelFontFace", "Helvetica"));
    int font_size = view.GetInt("LabelFontSize", 10);

    if (font_face != CGlTextureFont::eFontFace_LastFont) {                                 
        m_Font.SetFontSize(font_size);
        m_Font.SetFontFace(font_face);
    }
#endif


    
    // Get key for colors
    // could make black/white a theme instead of separate...
    view = x_GetReadView(kPhyloBaseKey, 
                         "Color." + m_CurrColorTheme, 
                         "Color.Color");

    // load colors
    x_UserFieldToColor(view, "LineColor", m_LineColor);
    x_UserFieldToColor(view,"LineSelColor",  m_LineSelColor);
    x_UserFieldToColor(view, "LineTraceColor",  m_LineTraceColor);
    x_UserFieldToColor(view, "LineSharedColor",  m_LineSharedColor);

    x_UserFieldToColor(view, "NodeColor", m_NodeColor);
    x_UserFieldToColor(view, "NodeSelColor", m_NodeSelColor);
    x_UserFieldToColor(view, "NodeTraceColor", m_NodeTraceColor);
    x_UserFieldToColor(view, "NodeSharedColor", m_NodeSharedColor);
    x_UserFieldToColor(view, "NodeTipSelColor", m_NodeTipSelColor);

    x_UserFieldToColor(view, "LabelColor", m_LabelColor);
    x_UserFieldToColor(view, "LabelSelColor", m_LabelSelColor);
    x_UserFieldToColor(view, "LabelTraceColor", m_LabelTraceColor);
    x_UserFieldToColor(view, "LabelSharedColor", m_LabelSharedColor);

    x_UserFieldToColor(view, "LineColorMono", m_LineColorMono);
    x_UserFieldToColor(view, "LineSelColorMono", m_LineSelColorMono);
    x_UserFieldToColor(view, "LineTraceColorMono", m_LineTraceColorMono);
    x_UserFieldToColor(view, "LineSharedColorMono", m_LineSharedColorMono);

    x_UserFieldToColor(view, "NodeColorMono", m_NodeColorMono);
    x_UserFieldToColor(view, "NodeSelColorMono", m_NodeSelColorMono);
    x_UserFieldToColor(view, "NodeTraceColorMono", m_NodeTraceColorMono);
    x_UserFieldToColor(view, "NodeSharedColorMono", m_NodeSharedColorMono);
    x_UserFieldToColor(view, "NodeTipSelColorMono", m_NodeTipSelColorMono);

    x_UserFieldToColor(view, "LabelColorMono", m_LabelColorMono);
    x_UserFieldToColor(view, "LabelSelColorMono", m_LabelSelColorMono);
    x_UserFieldToColor(view, "LabelTraceColorMono", m_LabelTraceColorMono);
    x_UserFieldToColor(view, "LabelSharedColorMono", m_LabelSharedColorMono);
       
    m_NonSelectedAlpha = (float)view.GetReal("NonSelectedAlpha", m_NonSelectedAlpha);
    // loading any global settings
    //  view = x_GetReadView(kPhyloBaseKey, 
    //                       "GlobalSettings." + m_CurrGlobalTheme, 
    //                       "GlobalSettings.Default");

    }
    catch (std::exception&)
    {
        return false;
    }

    return true;
}

bool CPhyloTreeScheme::SaveCurrentSettings()
{
 
    try {
        CGuiRegistry::TReadWriteView view = m_Registry->GetReadWriteView(kPhyloBaseKey);
        view.Set("CurrentTheme", m_CurrTheme);

        // save theme settings
        view = m_Registry->GetReadWriteView(kPhyloBaseKey + "." + m_ThemeKey, m_CurrTheme);
        view.Set("ColorTheme", m_CurrColorTheme);
        view.Set("SizeLevel", m_CurrSizeTheme);
        view.Set("GlobalSettings", m_CurrGlobalTheme);

        // Get key for sizes and font
        view = x_GetReadWriteView(kPhyloBaseKey, 
                                  "Size." + m_CurrSizeTheme,
                                  "Size.Normal");

        // save size settings
        view.Set("LeftMargin", m_LeftMargin);
        view.Set("RightMargin", m_RightMargin);
        view.Set("TopMargin", m_TopMargin);
        view.Set("BottomMargin", m_BottomMargin);
        view.Set("NodeSize", m_NodeSize);
        view.Set("LineWidth", m_LineWidth);

        // save font settings
        view.Set("LabelFormat", m_LabelFormat);

#ifndef TEXTURED_FONTS
        view.Set("LabelFontFace", 
                 CGlBitmapFont::FaceToString(m_Font.GetFontFace()));
        view.Set("LabelFontSize", 
                 NStr::StringToInt(CGlBitmapFont::SizeToString(m_Font.GetFontSize())));
#else
        view.Set("LabelFontFace", 
                 CGlTextureFont::FaceToString(m_Font.GetFontFace()));
        view.Set("LabelFontSize", (int)m_Font.GetFontSize());               
#endif

        // save color settings
        // could make black/white a theme instead of separate...
        view = x_GetReadWriteView(kPhyloBaseKey, 
                                  "Color." + m_CurrColorTheme, 
                                  "Color.Color");     
        x_ColorToUserField(view, "LineColor", m_LineColor);
        x_ColorToUserField(view, "LineSelColor", m_LineSelColor);
        x_ColorToUserField(view, "LineTraceColor", m_LineTraceColor);
        x_ColorToUserField(view, "LineSharedColor", m_LineSharedColor);

        x_ColorToUserField(view, "NodeColor", m_NodeColor);
        x_ColorToUserField(view, "NodeSelColor", m_NodeSelColor);
        x_ColorToUserField(view, "NodeTraceColor", m_NodeTraceColor);
        x_ColorToUserField(view, "NodeSharedColor", m_NodeSharedColor);
        x_ColorToUserField(view, "NodeTipSelColor", m_NodeTipSelColor);

        x_ColorToUserField(view, "LabelColor", m_LabelColor);
        x_ColorToUserField(view, "LabelSelColor", m_LabelSelColor);
        x_ColorToUserField(view, "LabelTraceColor", m_LabelTraceColor);
        x_ColorToUserField(view, "LabelSharedColor", m_LabelSharedColor);

        x_ColorToUserField(view, "LineColorMono", m_LineColorMono);
        x_ColorToUserField(view, "LineSelColorMono", m_LineSelColorMono);
        x_ColorToUserField(view, "LineTraceColorMono", m_LineTraceColorMono);
        x_ColorToUserField(view, "LineSharedColorMono", m_LineSharedColorMono);

        x_ColorToUserField(view, "NodeColorMono", m_NodeColorMono);
        x_ColorToUserField(view, "NodeSelColorMono", m_NodeSelColorMono);
        x_ColorToUserField(view, "NodeTraceColorMono", m_NodeTraceColorMono);
        x_ColorToUserField(view, "NodeSharedColorMono", m_NodeSharedColorMono);
        x_ColorToUserField(view, "NodeTipSelColorMono", m_NodeTipSelColorMono);

        x_ColorToUserField(view, "LabelColorMono", m_LabelColorMono);
        x_ColorToUserField(view, "LabelSelColorMono", m_LabelSelColorMono);
        x_ColorToUserField(view, "LabelTraceColorMono", m_LabelTraceColorMono);
        x_ColorToUserField(view, "LabelSharedColorMono", m_LabelSharedColorMono);    

        view.Set("NonSelectedAlpha", m_NonSelectedAlpha);

        // saving the global settings
        // view = x_GetReadWriteView(kPhyloBaseKey, 
        //                           "GlobalSettings." + m_CurrGlobalTheme, 
        //                           "GlobalSettings.Default");
    }
    catch (std::exception&)
    {
        return false;
    }

    return true;    
}



END_NCBI_SCOPE
