/*  $Id: phylo_tree_slanted_cladogram.cpp 27503 2013-02-26 18:16:55Z falkrb $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Vladimir Tereshkov
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>
#include <corelib/ncbistl.hpp>

#include <gui/widgets/phylo_tree/phylo_tree_slanted_cladogram.hpp>
#include <gui/widgets/phylo_tree/phylo_tree_ds.hpp>

#include <math.h>

BEGIN_NCBI_SCOPE


CPhyloSlantedCladogram::CPhyloSlantedCladogram()
{
    m_bDistMode = false;
}

CPhyloSlantedCladogram::CPhyloSlantedCladogram(double w, double h):
    IPhyloTreeRenderer(w, h)
{
    m_bDistMode = false;
}


CPhyloSlantedCladogram::~CPhyloSlantedCladogram()
{
}


void  CPhyloSlantedCladogram::x_Layout(CPhyloTreeDataSource& ds)
{
    m_Leafs = ds.GetSize();
    m_Width = ds.GetWidth();

    m_yStep = m_DimY / ((m_Leafs>1)?(m_Leafs-1):2);
    m_xStep = m_yStep;

    CBoundaryPoints boundary_pts;
    m_bDistMode = false;
    x_Calculate(ds.GetTree(), boundary_pts);

    if (ds.GetNormDistance() > 0){
        m_NormDistance = m_DimX /  ds.GetNormDistance();
    }

    // changing raster - x_Calculate for sure changed size in unpredictable way
    TModelRect newRect = ds.GetBoundRect();

    // single leaf
    if (newRect.Top() == newRect.Bottom()) {
        newRect.SetTop(m_DimY);
        newRect.SetBottom(0);
    }        

    m_RasterRect.Init(newRect.Left(),
                      newRect.Bottom(),
                      newRect.Right(),
                      newRect.Top());

    m_pPane->SetModelLimitsRect(m_RasterRect);
    m_pPane->SetVisibleRect(m_RasterRect);

    // Set drawing options that are specific to slanted cladogram
#ifdef BUFFERED_RENDER
    CGlVboNode* edge_node = ds.GetModel().FindGeomNode("TreeEdges");
    CGlVboNode* narrow_edge_node = ds.GetModel().FindGeomNode("NarrowTreeEdges");

    if (edge_node != NULL && narrow_edge_node != NULL) {   
        edge_node->GetState().Enable(GL_LINE_SMOOTH);
        edge_node->GetState().Enable(GL_BLEND);
        edge_node->GetState().BlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);        

        narrow_edge_node->GetState().Enable(GL_LINE_SMOOTH);
        narrow_edge_node->GetState().Enable(GL_BLEND);
        narrow_edge_node->GetState().BlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
        narrow_edge_node->GetState().ColorC(CRgbaColor(1.0f, 1.0f, 1.0f, 0.7f));
    }
#endif

    m_ValidLayout = eValidLayout;
    ComputeViewingLimits(*m_pPane, false);
    //m_pPane->SetVisibleRect(m_pPane->GetModelLimitsRect());
}


void  CPhyloSlantedCladogram::x_Render(CGlPane& pane, CPhyloTreeDataSource& ds)
{
    pane.OpenOrtho();
    x_DrawBoundingAreas(ds.GetTree());
    x_DrawTree(ds.GetTree());
    x_DrawNodes(ds.GetTree());
    pane.Close();
}

void  CPhyloSlantedCladogram::x_RenderVbo(CPhyloTreeDataSource& ds)
{    
    CGlVboNode* edge_node = m_DS->GetModel().FindGeomNode("TreeEdges");
    CGlVboNode* narrow_edge_node = m_DS->GetModel().FindGeomNode("NarrowTreeEdges");
    CGlVboNode* point_node = m_DS->GetModel().FindGeomNode("NodePoints");
    _ASSERT(edge_node != NULL && narrow_edge_node != NULL);

    //
    GLsizei edge_count = (GLsizei)ds.GetNumEdges();    
    edge_node->GetColorBuffer().BufferDataResize(edge_count*2, NULL, GL_DYNAMIC_DRAW);

    if (edge_count > 0) {        
        edge_node->GetColorBuffer().MapBuffer(GL_WRITE_ONLY);

        x_DrawTreeVbo(ds.GetTree());

        edge_node->GetColorBuffer().UnmapBuffer();
        edge_node->SetVisible(true);
    }
    else {
        edge_node->SetVisible(false);
    }

    narrow_edge_node->SetDefaultPosition();
    narrow_edge_node->SetVisible(true);
   
    edge_node->SetDefaultPosition();
    edge_node->SetVisible(true);

    GLsizei node_count = (GLsizei)ds.GetNumNodes();
    if (node_count > 0) {
        point_node->GetColorBuffer().BufferDataResize(node_count, NULL, GL_DYNAMIC_DRAW);
        point_node->GetColorBuffer().MapBuffer(GL_WRITE_ONLY);        

        x_DrawNodesVbo(ds.GetTree(), point_node);

        point_node->GetColorBuffer().UnmapBuffer();   
        point_node->ClearPositions();
        point_node->SetVisible(true);
    }
    else {
        point_node->SetVisible(false);
    }

}


void CPhyloSlantedCladogram:: x_Calculate(CPhyloTreeNode * node,                                           
                                          CBoundaryPoints& parent_boundary_pts,
                                          bool compute_boundary)
{
    if ((**node).GetBoundedDisplay() == IPhyGraphicsNode::eBounded)
        compute_boundary = true;
    (**node).SetAngle(0); // for correct labeling

    CBoundaryPoints node_boundary_pts;

    if (node->IsLeafEx()) {
        (**node).X()  = m_DimX;                                  // all leafs on the right
        (**node).Y() = (m_DimY - (**node).IDX().second * m_yStep);     // y coordinate        
    }
    else {
        Int4 childLeafs = (**node).GetLeavesNmbEx();

        (**node).X() = (m_DimX - (childLeafs-1) * m_xStep);
        (**node).Y() = (m_DimY - (**node).IDX().second * m_yStep -
                            ((childLeafs-1)*m_yStep)/2.0);

        for(CPhyloTreeNode::TNodeList_I  it = node->SubNodeBeginEx();
            it != node->SubNodeEnd(); it++)  {
            CPhyloTreeNode* subnode = static_cast<CPhyloTreeNode*>(*it);
            x_Calculate(subnode, node_boundary_pts, compute_boundary);
        }
    }

    x_ComputeNodeBoundary(node, compute_boundary, node_boundary_pts, "SlantedCladogram");

    // Each node that computes a boundary has to add in the points (and text rectangles)
    // for all nodes below it in the tree
    if (compute_boundary) {
        parent_boundary_pts.AddBoundedPoints(node_boundary_pts);
    }
}

void CPhyloSlantedCladogram::x_DrawTree(CPhyloTreeNode * node)
{
    // drawing leaf node
    if (node->GetParent()){
        x_RenderLine(node->GetParent()->GetValue().X(),
                     node->GetParent()->GetValue().Y(),
                     node->GetValue().X(),
                     node->GetValue().Y(),
                     (**node).GetSelectedState(),
                     GetEdgeColor(node));

    }
    for(CPhyloTreeNode::TNodeList_I  it = node->SubNodeBeginEx();
        it != node->SubNodeEnd(); it++ )  x_DrawTree(static_cast<CPhyloTreeNode*>(*it));

    // x_RenderNode(node);
}

void CPhyloSlantedCladogram::x_DrawTreeVbo(CPhyloTreeNode * node)
{
    // drawing leaf node
    if (node->GetParent()){
        CGlVboNode* edge_node = m_DS->GetModel().FindGeomNode("TreeEdges");
        _ASSERT(edge_node != NULL);

        x_RenderLineVbo(edge_node,
                        node->GetParent()->GetValue().X(),
                        node->GetParent()->GetValue().Y(),
                        node->GetValue().X(),
                        node->GetValue().Y(),
                        (**node).GetSelectedState(),
                        GetEdgeColor(node));

    }
    for(CPhyloTreeNode::TNodeList_I  it = node->SubNodeBeginEx();
        it != node->SubNodeEnd(); it++ )  x_DrawTreeVbo(static_cast<CPhyloTreeNode*>(*it));
}


string CPhyloSlantedCladogram::GetDescription(void)
{
    return "Slanted Cladogram";
}

TVPUnit CPhyloSlantedCladogram::DistanceBetweenNodes(void)
{
    return x_DistanceBetweenNodes(m_pPane->GetScaleY());
}

TModelUnit CPhyloSlantedCladogram::x_DistanceBetweenNodes(TModelUnit scaley)
{
    return  (TModelUnit) (GetDimY() / (m_DS->GetSize() * scaley));
}

END_NCBI_SCOPE
