/*  $Id: tree_collision_model2d.cpp 27277 2013-01-18 20:50:36Z falkrb $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Bob Falk
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>
#include <corelib/ncbistl.hpp>

#include <gui/widgets/phylo_tree/tree_collision_model2d.hpp>
#include <gui/widgets/phylo_tree/tree_graphics_model.hpp>
#include <gui/widgets/phylo_tree/phylo_tree_ds.hpp>
#include <gui/widgets/phylo_tree/phylo_tree_node.hpp>

#include <gui/widgets/gl/attrib_menu.hpp> 
#include <corelib/ncbitime.hpp>


BEGIN_NCBI_SCOPE

CTreeCollisionModel2D::CTreeCollisionModel2D()
: m_Model(NULL)
{
#ifdef ATTRIB_MENU_SUPPORT
    m_VisUpdateTime = 0.0f;
    m_ScaleUpdateTime = 0.0f;

    CAttribMenu& m = CAttribMenuInstance::GetInstance();

    CAttribMenu* sub_menu = m.AddSubMenuUnique("Collision", this);

    sub_menu->AddIntReadOnly("CD Width", &m_Width);
    sub_menu->AddIntReadOnly("CD Height", &m_Height);

    sub_menu->AddFloatReadOnly("UpdateVis T", &m_VisUpdateTime);
    sub_menu->AddFloatReadOnly("UpdateScale T", &m_ScaleUpdateTime);

    sub_menu->AddFloat("Resolution Scaler", &m_ResolutionScaler, 1.0f, 0.1f, 10.0f, 0.05f);
#endif 
}

CTreeCollisionModel2D::~CTreeCollisionModel2D()
{
#ifdef ATTRIB_MENU_SUPPORT
    CAttribMenuInstance::GetInstance().RemoveMenuR("Collision", this);
#endif
}

void CTreeCollisionModel2D::Update(const CVect2<float>& scale)
{
    Clear();

    m_Scale = scale;

    UpdateScaled(m_Scale);    
}


void CTreeCollisionModel2D::UpdateScaled(const CVect2<float>& scale)
{
    if (m_Model == NULL)
        return;

    CStopWatch  t;
    t.Start();

    ClearScaled();

    m_Scale = scale;
   
    vector<CTreeGraphicsNode*>& node_vec = m_Model->GetNodes();
    for (size_t idx=0; idx<node_vec.size(); ++idx) {
        CVect2<float> ll, ur;

        node_vec[idx]->GetBoundingRect(m_Scale, ll, ur);

        
        CVect2<int> posi_ll(int(ll[0]/m_Resolution), 
                            int(ll[1]/m_Resolution));
        CVect2<int> posi_ur(int(ur[0]/m_Resolution), 
                            int(ur[1]/m_Resolution));

        // Make sure we only look at sections inside our spatial subdivision (which
        // may not include regions when geometry is larger than the un-zoomed viewing
        // area)
        posi_ll.X() = std::max(posi_ll.X(), m_MinIdx.X());
        posi_ll.Y() = std::max(posi_ll.Y(), m_MinIdx.Y());
        posi_ur.X() = std::min(posi_ur.X(), m_MaxIdx.X());
        posi_ur.Y() = std::min(posi_ur.Y(), m_MaxIdx.Y());

        posi_ll -= m_MinIdx;
        posi_ur -= m_MinIdx;

        // Guarantee no out-of-bounds errors in grid:
        posi_ll.X() = std::min(posi_ll.X(), m_Width);
        posi_ll.Y() = std::min(posi_ll.Y(), m_Height);

        posi_ur.X() = std::min(posi_ur.X(), m_Width);
        posi_ur.Y() = std::min(posi_ur.Y(), m_Height);

        CSpatialHash2D<CollisionEntry>::TVeci grid_pos;
        for (grid_pos.Y() = posi_ll.Y(); grid_pos.Y()<=posi_ur.Y(); ++grid_pos.Y()) {
            for (grid_pos.X() = posi_ll.X(); grid_pos.X()<=posi_ur.X(); ++grid_pos.X()) {
                Get(grid_pos).m_Labels.push_back(idx);
            }
        }
    }    

    m_ScaleUpdateTime = (float)t.Elapsed();
}

CPhyloTreeNode* CTreeCollisionModel2D::TestForNode(float x, float y)
{
    if (m_Model == NULL)
        return NULL;

    /// Get index for cell that holds (x,y)
    TVeci idx(int(x/m_Resolution), 
              int(y/m_Resolution));  

    CPhyloTreeNode* sel_node = NULL;

    idx -= m_MinIdx;

    /// Get the set of nodes (indices) in the cell that (x,y) falls into
    std::vector<size_t> cell_nodes = GetSafe(idx).m_Labels;
    
    int best_vis_idx = -1;
    float best_vis_label_dist = FLT_MAX;

    int best_hidden_idx = -1;
    float best_hidden_label_dist = FLT_MAX;

    // iterate over all the nodes in the cell for (x,y) and check
    // to see if one is close enough to (x,y) to call select it
    vector<CTreeGraphicsNode*>& node_vec = m_Model->GetNodes();
    for (size_t i=0; i<cell_nodes.size(); ++i) {
        CTreeGraphicsNode* node = node_vec[cell_nodes[i]];
        CVect2<float> p = (*node->GetSourceNode())->XY();
         
        // The node size is in pixels.  add a couple to make it selection easier...      
        float r = node->GetNodeSize() + 3.0f;     

        // A 'hit' should be within a couple of pixels...
        float xdelta = (x-p.X()) * 1.0f/m_Scale.X();
        float ydelta = (y-p.Y()) * 1.0f/m_Scale.Y();

        float dist = sqrtf( xdelta*xdelta + ydelta*ydelta );

        if (dist < r) {
            bool vis_label = node->GetVisible();

            // Favor nodes with visible labels over ones without:
            if (vis_label && dist < best_vis_label_dist) {
                best_vis_idx = cell_nodes[i];
                best_vis_label_dist = dist;
            }
            else if (!vis_label && dist < best_hidden_label_dist) {
                best_hidden_idx = cell_nodes[i];
                best_hidden_label_dist = dist;
            }
        }        
    }
  
    if (best_vis_idx > -1) {
        sel_node = node_vec[best_vis_idx]->GetSourceNode();
    }
    else if (best_hidden_idx != -1) {
        sel_node = node_vec[best_hidden_idx]->GetSourceNode();
    }

    return sel_node;
}

vector<CPhyloTreeNode*> CTreeCollisionModel2D::SelectNodes(float x1, float y1, float x2, float y2)
{
    vector<CPhyloTreeNode*> sel_nodes;

    if (m_Model == NULL)
        return sel_nodes;

    // Get index for lower-left corner cell in which (x1,y1) falls.
    TVeci idx1(int(x1/m_Resolution), 
               int(y1/m_Resolution));  
    idx1 -= m_MinIdx;

    // Get index for upper-right corner cell in which (x2,y2) falls
    TVeci idx2(int(x2/m_Resolution), 
               int(y2/m_Resolution));     
    idx2 -= m_MinIdx;

    // Iterate over all the cells (idx1-idx2)
    // and for each cell iterate over all the nodes in it too
    // see if they fall within (x1,x2)-(y1,y2)
    vector<CTreeGraphicsNode*>& node_vec = m_Model->GetNodes();
    for (int x = idx1.X(); x <= idx2.X(); ++x) {
        for (int y = idx1.Y(); y <= idx2.Y(); ++y) {
            TVeci idx(x,y);

            std::vector<size_t> cell_nodes = GetSafe(idx).m_Labels;

            // Find collisions between nodes hashed in this entry
            for (size_t i=0; i<cell_nodes.size(); ++i) {
                CTreeGraphicsNode* node = node_vec[cell_nodes[i]];
                CVect2<float> p = (*node->GetSourceNode())->XY();

                // A 'hit' will include the center of the node in the bounding box
                if (p.X() >= x1 && p.X() <= x2 &&
                    p.Y() >= y1 && p.Y() <= y2) {

                    sel_nodes.push_back(node->GetSourceNode());
                }
            }
        }
    }

    // A node may overalp more than one cell in the sptial data structure, so
    // we eliminate duplicates here 
    sel_nodes.erase(std::unique(sel_nodes.begin(), sel_nodes.end()), sel_nodes.end());

    return sel_nodes;   
}


void CTreeCollisionModel2D::Sync(CTreeGraphicsModel* m)
{
    m_Model = m;

    Update(m_Scale);
}

bool CTreeCollisionModel2D::RectIntersect(const CVect2<float>& ll1,
                                          const CVect2<float>& ur1,
                                          const CVect2<float>& ll2,
                                          const CVect2<float>& ur2)
{    
    bool xoverlap = !( (ll2.X() > ur1.X() || ur2.X() < ll1.X()) );
    bool yoverlap = !( (ll2.Y() > ur1.Y() || ur2.Y() < ll1.Y()) );

    return (xoverlap && yoverlap);
}

struct CompareLabelX {
    CompareLabelX(vector<CTreeGraphicsNode*>& nodes, 
                  const CVect2<float>& s) 
    : m_NodeVec(nodes)
    , m_Scale(s) {}

    bool operator()(size_t lhs, size_t rhs) 
    { 
        return (m_NodeVec[lhs]->GetMinX(m_Scale) <
                m_NodeVec[rhs]->GetMinX(m_Scale)); 
    }

    vector<CTreeGraphicsNode*>& m_NodeVec;
    CVect2<float> m_Scale;
};


void CTreeCollisionModel2D::UpdateVisibility(const TModelRect& r)
{
    if (m_Model == NULL)
        return;

    CStopWatch  t;
    t.Start();

    // Lower left and upper right corners of currently visible rectangle
    CVect2<float> ll((float)r.Left(), (float)r.Bottom());
    CVect2<float> ur((float)r.Right(), (float)r.Top());

    // Indices in the datastructure for lower-left and upper right corners of
    // current rectangle
    CVect2<int> posi_ll(int(ll[0]/m_Resolution) - m_MinIdx.X(), 
                        int(ll[1]/m_Resolution) - m_MinIdx.Y());
    CVect2<int> posi_ur(int(ur[0]/m_Resolution) - m_MinIdx.X(), 
                        int(ur[1]/m_Resolution) - m_MinIdx.Y());

    m_VisMinIdx = posi_ll;
    m_VisMaxIdx = posi_ur;

    // Guarantee no out-of-bounds errors:
    posi_ll.X() = std::min(posi_ll.X(), m_Width-1);
    posi_ll.X() = std::max(posi_ll.X(), 0);

    posi_ll.Y() = std::min(posi_ll.Y(), m_Height-1);
    posi_ll.Y() = std::max(posi_ll.Y(), 0);

    posi_ur.X() = std::min(posi_ur.X(), m_Width-1);
    posi_ur.X() = std::max(posi_ur.X(), 0);

    posi_ur.Y() = std::min(posi_ur.Y(), m_Height-1);
    posi_ur.Y() = std::max(posi_ur.Y(), 0);

    // Indices in the datastructure for lower-left and upper right corners of
    // rectangle visible at previous update
    CVect2<int> prev_posi_ll(int(m_PrevVisMin[0]/m_Resolution) - m_MinIdx.X(), 
                             int(m_PrevVisMin[1]/m_Resolution) - m_MinIdx.Y());
    CVect2<int> prev_posi_ur(int(m_PrevVisMax[0]/m_Resolution) - m_MinIdx.X(), 
                             int(m_PrevVisMax[1]/m_Resolution) - m_MinIdx.Y());

    // Guarantee no out-of-bounds errors:
    prev_posi_ll.X() = std::min(prev_posi_ll.X(), m_Width-1);
    prev_posi_ll.X() = std::max(prev_posi_ll.X(), 0);

    prev_posi_ll.Y() = std::min(prev_posi_ll.Y(), m_Height-1);
    prev_posi_ll.Y() = std::max(prev_posi_ll.Y(), 0);

    prev_posi_ur.X() = std::min(prev_posi_ur.X(), m_Width-1);
    prev_posi_ur.X() = std::max(prev_posi_ur.X(), 0);

    prev_posi_ur.Y() = std::min(prev_posi_ur.Y(), m_Height-1);
    prev_posi_ur.Y() = std::max(prev_posi_ur.Y(), 0);

    vector<CTreeGraphicsNode*>& node_vec = m_Model->GetNodes();

    // Since an individual element may be in multiple grid cells, we 
    // iterate twice - once to turn off visibility for any element that
    // overlaps a grid cell visible on the previous update, then again
    // to turn on visibility for all overlapped cells.

    // A last interation finds overlaps between lables and picks labels to
    // not draw.

    // Iterate over the whole grid subregion that was previously displayed
    // and update all labels to be 'not visible' in that region.
    for (int x = prev_posi_ll.X(); x <= prev_posi_ur.X(); ++x) {
        for (int y = prev_posi_ll.Y(); y <= prev_posi_ur.Y(); ++y) {
            TVeci idx(x,y);

            std::vector<size_t> cell_nodes = Get(idx).m_Labels;
            // Find collisions between nodes hashed in this entry
            for (size_t i=0; i<cell_nodes.size(); ++i) {
                node_vec[cell_nodes[i]]->SetVisible(false);
            }
        }
    }

    // Iterate over the currently-visible grid which subdivides 
    // the space occupied by the tree into individual cells, 
    // turning visibility on for all non-empty labels in the region
    for (int x = posi_ll.X(); x <= posi_ur.X(); ++x) {
        for (int y = posi_ll.Y(); y <= posi_ur.Y(); ++y) {
            TVeci idx(x,y);
            std::vector<size_t>& cell_nodes = Get(idx).m_Labels;

            // Maintain sorted order here (could be moved to sync, but this will
            // be fast for subsequent (already sorted..) updates...
            // Sorting is desired so that we favor in overlaps the node starting
            // further to the left (picking a standard makes display more consistent).
            CompareLabelX  cmp(node_vec, m_Scale);
            std::sort(cell_nodes.begin(), cell_nodes.end(), cmp);

            /// Could also be more efficient here to update bounding rect data
            /// now rather than do it in inner AND outer loops below. (but we
            /// have to add data to label class, making array larger...

            // Find collisions between nodes hashed in this entry   
            for (size_t i=0; i<cell_nodes.size(); ++i) {
                if (node_vec[cell_nodes[i]]->GetLabel() != "")
                    node_vec[cell_nodes[i]]->SetVisible(true);
            }
        }
    }

    // Find overlaps between visible labels and when that happens,
    // set the second label to not be visible (second label, due to sorting,
    // should be to the right of the first label).
    for (int x = posi_ll.X(); x <= posi_ur.X(); ++x) {
        for (int y = posi_ll.Y(); y <= posi_ur.Y(); ++y) {
            TVeci idx(x,y);

            std::vector<size_t> cell_nodes = Get(idx).m_Labels;

            // Find collisions between nodes hashed in this entry   
            for (size_t i=0; i<cell_nodes.size(); ++i) {
                CTreeGraphicsNode* l1 = node_vec[cell_nodes[i]];

                CVect2<float> ll1, ur1;
                if (l1->GetVisible()) {
                    l1->GetLabelRect(m_Scale, ll1, ur1);

                    for (size_t j=i+1; j<cell_nodes.size(); ++j) {
                        CTreeGraphicsNode* l2 = node_vec[cell_nodes[j]];

                        /// Need to consider possibility of collisions in multiple nodes and identical
                        /// xcoords
                        if (l2->GetVisible()) {
                            CVect2<float> ll2, ur2;

                            l2->GetLabelRect(m_Scale, ll2, ur2);
                            if (RectIntersect(ll1, ur1, ll2, ur2)) {
                                l2->SetVisible(false);
                            }
                        }
                    }
                }
            }
        }
    }

    m_PrevVisMin = ll;
    m_PrevVisMax = ur;

    m_VisUpdateTime = t.Elapsed();
}



END_NCBI_SCOPE
