/*  $Id: tree_graphics_model.cpp 27961 2013-04-30 19:58:53Z falkrb $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Bob Falk
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>
#include <corelib/ncbistl.hpp>

#include <gui/widgets/phylo_tree/tree_graphics_model.hpp>
#include <gui/widgets/phylo_tree/phylo_tree_node.hpp>
#include <gui/widgets/phylo_tree/phylo_tree_boundary_shapes.hpp>

#include <gui/opengl/glresmgr.hpp>

#include <gui/print/pdf.hpp>
#include <gui/print/print_media.hpp>
#include <gui/print/pdf_object.hpp>

#include <gui/widgets/gl/attrib_menu.hpp>

#include <gui/utils/matrix3.hpp>
#include <gui/utils/matrix4.hpp>

#include <corelib/ncbitime.hpp>

BEGIN_NCBI_SCOPE

void CTreeBoundaryNode::x_Render()
{
    if (!IsVisible())
        return;

    if (m_BoundaryTris.IsNull()) {
        m_BoundaryTris.Reset(new CGlVboNode(GL_TRIANGLES));
        m_BoundaryTris->GetState().Enable(GL_BLEND);
        m_BoundaryTris->GetState().BlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    }
    if (m_BoundaryEdges.IsNull()) {
        m_BoundaryEdges.Reset(new CGlVboNode(GL_LINES));
        m_BoundaryEdges->GetState().Enable(GL_BLEND);
        m_BoundaryEdges->GetState().BlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    }

    for (size_t i=0; i<m_BoundaryNodes.size(); ++i) {
        CPhyloTreeNode* node = m_BoundaryNodes[i];

        if ((*node)->GetChildsDisplay() == IPhyGraphicsNode::eHideChilds)
            continue;
      
        //glTranslated((**n).X(), (**n).Y(), 0.0);
        float alpha_mod = 1.0f;

        if ((**node).GetSelectedState()!=IPhyGraphicsNode::eSelected) {
            if (m_SL->GetSelectionVisibility() == 
                CPhyloTreeScheme::eHighlightSelection) {
                    alpha_mod = m_SL->GetNonSelectedAlpha();
            }
	    } 

        CVect2<float> scale2d(m_Scale.X(), m_Scale.Y());

        m_BoundaryEdges->GetVertexBuffer2D().SetActiveBufferSize(0);
        m_BoundaryTris->GetVertexBuffer2D().SetActiveBufferSize(0);

        //(**node).GetSubtreeBoundary()->RenderBoundary(scale2d, alpha_mod); 
        (**node).GetSubtreeBoundary()->RenderBoundaryVbo(m_BoundaryTris, m_BoundaryEdges, scale2d, alpha_mod);
        m_BoundaryTris->Render();
        m_BoundaryEdges->Render();
    }
}

void CTreeBoundaryNode::RenderPdf(CGlPane& pane, CRef<CPdf>& pdf)
{
    if (!IsVisible())
        return;

	GLint m_Viewport[4];
    GLdouble m_ModelviewMatrix[16];
    GLdouble m_ProjectionMatrix[16];

    // Get viewport from pane object, which may have a larger viewport than     
    // OpenGL maximum (for pdf)
    //glGetIntegerv(GL_VIEWPORT, m_Viewport);
    m_Viewport[0] = pane.GetViewport().Left();
    m_Viewport[1] = pane.GetViewport().Bottom();
    m_Viewport[2] = pane.GetViewport().Right();
    m_Viewport[3] = pane.GetViewport().Top();

    glGetDoublev(GL_MODELVIEW_MATRIX, m_ModelviewMatrix);
    glGetDoublev(GL_PROJECTION_MATRIX, m_ProjectionMatrix);

    if (m_BoundaryTris.IsNull())
        m_BoundaryTris.Reset(new CGlVboNode(GL_TRIANGLES));
    if (m_BoundaryEdges.IsNull())
        m_BoundaryEdges.Reset(new CGlVboNode(GL_LINES));

    for (size_t i=0; i<m_BoundaryNodes.size(); ++i) {
        CPhyloTreeNode* node = m_BoundaryNodes[i];

        if ((*node)->GetChildsDisplay() == IPhyGraphicsNode::eHideChilds)
            continue;
      
        //glTranslated((**n).X(), (**n).Y(), 0.0);
        float alpha_mod = 1.0f;

        if ((**node).GetSelectedState()!=IPhyGraphicsNode::eSelected) {
            if (m_SL->GetSelectionVisibility() == 
                CPhyloTreeScheme::eHighlightSelection) {
                    alpha_mod = m_SL->GetNonSelectedAlpha();
            }
	    } 

        CVect2<float> scale2d(m_Scale.X(), m_Scale.Y());

        m_BoundaryEdges->GetVertexBuffer2D().SetActiveBufferSize(0);
        m_BoundaryTris->GetVertexBuffer2D().SetActiveBufferSize(0);
      
        (**node).GetSubtreeBoundary()->RenderBoundaryVbo(m_BoundaryTris, m_BoundaryEdges, scale2d, alpha_mod);

        pdf->PrintTriBuffer(m_BoundaryTris, pane.GetViewport());
       

        if (!m_BoundaryEdges.IsNull() && 
            m_BoundaryEdges->GetVertexBuffer2D().GetActiveBufferSize() > 0) {
                pdf->PrintLineBuffer(m_BoundaryEdges, pane.GetViewport());
        }
    }
}


CTreeGraphicsModel::CTreeGraphicsModel()
: m_BoundaryNode(NULL)
, m_TriFanNode(NULL)
, m_TriFanNodePass2(NULL)
, m_RenderTime(0.0f)
, m_LineRenderTime(0.0f)
, m_NodesRenderTime(0.0f)
, m_NodeRenderCount(0)
{
#ifdef ATTRIB_MENU_SUPPORT
    CAttribMenu& m = CAttribMenuInstance::GetInstance();
    CAttribMenu* sub_menu = m.AddSubMenuUnique("Model", this);

    sub_menu->AddFloatReadOnly("Render Time", &m_RenderTime);
    sub_menu->AddFloatReadOnly("Line Render Time", &m_LineRenderTime);
    sub_menu->AddFloatReadOnly("Nodes Render Time", &m_NodesRenderTime);
    sub_menu->AddIntReadOnly("Nodes Render Count", &m_NodeRenderCount);
#endif
}

CTreeGraphicsModel::~CTreeGraphicsModel()
{
    for (size_t i=0; i<m_Nodes.size(); ++i)
        delete m_Nodes[i];
    m_Nodes.clear();

    delete m_BoundaryNode;
    m_BoundaryNode = NULL;

    delete m_TriFanNode;
    m_TriFanNode = NULL;

    delete m_TriFanNodePass2;
    m_TriFanNodePass2 = NULL;

#ifdef ATTRIB_MENU_SUPPORT
    CAttribMenuInstance::GetInstance().RemoveMenuR("Model", this);
#endif
}


void CTreeGraphicsModel::Init()
{    
    // Order that the CGlVboNode's are created is significant, as it 
    // determines their drawing order (and thus what is drawn on top of what)

    // Used by some renderings to fill in joint where two lines with 
    // width > 1 meet (otherwise you get a gap at the corner,as shown
    // below where two lines of width 3 pixels meet but leave the position
    // marked by ? blank.
    //  ?xxxxxxx
    //  xxxxxxxx  <- line 2
    //  xxxxxxxx
    //  xxx
    //  xxx <-line 1
    //  xxx
    // (faster)?:
    CGlVboNode* filler_point_node = FindGeomNode("FillerPoints");
    if (!filler_point_node) {
        filler_point_node = AddGeomNode(GL_POINTS, "FillerPoints", true);    
        filler_point_node->GetState().Enable(GL_BLEND);
        filler_point_node->GetState().BlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

        // Don't draw points in pdf mode (pdf fills line elbows differently)
        filler_point_node->SkipTarget(eRenderPDF, true);
    }

    CGlVboNode* default_edge_node = FindGeomNode("TreeEdges");
    if (!default_edge_node) {
        default_edge_node = AddGeomNode(GL_LINES, "TreeEdges", true);
        default_edge_node->GetState().Disable(GL_BLEND);
        default_edge_node->GetState().Disable(GL_TEXTURE_2D);
        default_edge_node->GetState().LineJoinStyle(CGlState::eRoundedJoin);
        default_edge_node->GetState().LineCapStyle(CGlState::eRoundCap);
    } 

    CGlVboNode* boundary_node = FindGeomNode("BoundaryNodes");
    if (!boundary_node) {
        boundary_node = AddGeomNode(GL_TRIANGLES, "BoundaryNodes", true);
        boundary_node->GetState().Disable(GL_BLEND);
        boundary_node->GetState().Disable(GL_TEXTURE_2D);
    }

    // These edges are inside wider edges (same geometry buffer as 'TreeEdges')
    CGlVboNode* narrow_edge_node = FindGeomNode("NarrowTreeEdges");
    if (!narrow_edge_node) {        
        narrow_edge_node = AddGeomNode(GL_LINES, "NarrowTreeEdges", true);
        CRef<CGlVboGeom> geom = narrow_edge_node->GetVBOGeom();
        CGlVboGeom* vbo_geom = dynamic_cast<CGlVboGeom*>(geom.GetPointerOrNull());
        if (vbo_geom != NULL) {
            vbo_geom->SetVertexBuffer2D(default_edge_node->GetVertexBuffer2D().GetBuffer());
        } 
        
        narrow_edge_node->GetState().LineWidth(1.0f);
        narrow_edge_node->GetState().Enable(GL_BLEND);
        narrow_edge_node->GetState().BlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
        // color to be used for lightinging in blending equation
        narrow_edge_node->GetState().ColorC(CRgbaColor(1.0f, 1.0f, 1.0f, 0.7f));
        narrow_edge_node->GetState().Disable(GL_TEXTURE_2D);
        narrow_edge_node->GetState().LineJoinStyle(CGlState::eMiteredJoin);
        narrow_edge_node->GetState().LineCapStyle(CGlState::eButtCap);
    }

    // Used by some renderings to fill in joint where two lines with 
    // width > 1 meet (otherwise you get a gap at the corner,as shown
    // below where two lines of width 3 pixels meet but leave the position
    // marked by ? blank.
    //  ?xxxxxxx
    //  xxxxxxxx  <- line 2
    //  xxxxxxxx
    //  xxx
    //  xxx <-line 1
    //  xxx
    // (faster)?:
    // Instead, draw the buffer twice, but second time call it a 
    // point buffer and set the offset to only draw the corner points
    // (preferably without using an index buffer. (Can't draw end points
    // since that would extend the line)
    if (filler_point_node != NULL) {
        CRef<CGlVboGeom> geom = filler_point_node->GetVBOGeom();
        CGlVboGeom* vbo_geom = dynamic_cast<CGlVboGeom*>(geom.GetPointerOrNull());
        if (vbo_geom != NULL) {
            vbo_geom->SetVertexBuffer2D(default_edge_node->GetVertexBuffer2D().GetBuffer());
            vbo_geom->GetVertexBuffer2D().SetStride(32);
            vbo_geom->GetVertexBuffer2D().SetOffset(16);

            vbo_geom->SetColorBuffer(default_edge_node->GetColorBuffer().GetBuffer());
            vbo_geom->GetColorBuffer().SetStride(64);
            vbo_geom->GetColorBuffer().SetOffset(32);
        }         
    }

    CGlVboNode* point_node = FindGeomNode("NodePoints");
    if (!point_node) {
        point_node = AddGeomNode(GL_POINTS, "NodePoints", true);
    }

    if (m_BoundaryNode == NULL) {
        m_BoundaryNode = new CTreeBoundaryNode();
        if (m_SL.GetPointerOrNull() != NULL)
            m_BoundaryNode->SetScheme(m_SL.GetObject());
    }

    if (m_TriFanNode == NULL) {
        m_TriFanNode = new CTreeTriFanNode();
        m_TriFanNode->GetState().Disable(GL_TEXTURE_2D);
        m_TriFanNode->GetState().Disable(GL_BLEND);
        m_TriFanNode->GetState().ShadeModel(GL_SMOOTH);
        m_TriFanNode->GetState().PolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    }

    if (m_TriFanNodePass2 == NULL) {
        m_TriFanNodePass2 = new CGlVboNode(GL_TRIANGLE_FAN);
        m_TriFanNodePass2->GetState().Enable(GL_BLEND);
        m_TriFanNodePass2->GetState().Disable(GL_TEXTURE_2D);
        m_TriFanNodePass2->GetState().BlendFunc(GL_ONE_MINUS_DST_ALPHA, GL_DST_ALPHA);
        m_TriFanNodePass2->GetState().ShadeModel(GL_SMOOTH);
        m_TriFanNode->GetState().PolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    }
}

void CTreeGraphicsModel::SyncBuffers()
{
    // Make sure color class is right size (nobody has added virtual functions
    // or other data).
    _ASSERT(sizeof(CRgbaColor) == 16);

    CGlVboNode* point_node = FindGeomNode("NodePoints");
    CGlVboNode* filler_point_node = FindGeomNode("FillerPoints");
    CGlVboNode* edge_node = FindGeomNode("TreeEdges");
    
    if (m_Nodes.size() > 0) {
        size_t cb_size = point_node->GetColorBuffer().GetActiveBufferSize();
        //_ASSERT(cb_size >= m_Nodes.size());
        // color buffer could be bigger since it is sized to hold all nodes, even
        // ones not currently visible. But the contents only reflect visible nodes.
        point_node->GetColorBuffer().SetActiveBufferSize( cb_size);

        point_node->SetVisible(true);
        // Gets cleared in clear arrays.  Seems like a bit of a hack...
        point_node->SetDefaultPosition();

        point_node->GetVertexBuffer2D().BufferDataResize(m_Nodes.size(), NULL, GL_DYNAMIC_DRAW);
        point_node->GetVertexBuffer2D().MapBuffer(GL_WRITE_ONLY);

        for (size_t i=0; i<m_Nodes.size(); ++i) {
            if ((*m_Nodes[i]->GetSourceNode())->GetChildsDisplay() != IPhyGraphicsNode::eHideChilds)
                point_node->GetVertexBuffer2D().PushBack((*m_Nodes[i]->GetSourceNode())->XY());
        }

        point_node->GetVertexBuffer2D().UnmapBuffer();
    }
    else if (point_node != NULL) {
        point_node->SetVisible(false);
    }   

    // The filler points are set to visible only for pseudo-spline (right-angle
    // edges) rendering.  Otherwise visible will be set to false.
    if (filler_point_node->IsVisible()) {
        filler_point_node->SetDefaultPosition();
        filler_point_node->GetVBOGeom()->SetVertexCount(m_Edges.size()/2);
    }
    

    if (m_Edges.size() > 0) {
        size_t cb_size = edge_node->GetColorBuffer().GetActiveBufferSize();
        _ASSERT(cb_size >= m_Edges.size()*2);
        // color buffer could be bigger since it is sized to hold all edges, even
        // ones not currently visible. But the contents only reflect visible edges.
        edge_node->GetColorBuffer().SetActiveBufferSize(m_Edges.size()*2);

        edge_node->SetVisible(true);
        // Gets cleared in clear arrays.  Seems like a bit of a hack...
        edge_node->SetDefaultPosition();


        // edge_node size is number of vertices, but m_Edges size is number of
        // line segments (so num line-segments = num-vertices * 2)
        if (edge_node->GetSize() < (int)m_Edges.size()*2) {
            edge_node->GetVertexBuffer2D().BufferData(m_Edges.size()*2, 
                                                       ((CVect2<float>*)&m_Edges[0]),
                                                       GL_DYNAMIC_DRAW);
        }
        else {  
            edge_node->GetVertexBuffer2D().BufferSubData(0, m_Edges.size()*2, 
                                                          ((CVect2<float>*)&m_Edges[0]));
            edge_node->GetVertexBuffer2D().SetActiveBufferSize(m_Edges.size()*2);
        }
    }
    else if (edge_node != NULL) {
        edge_node->SetVisible(false);
    }        
}

void CTreeGraphicsModel::ClearArrays()
{
    CGlModel2D::ClearArrays();

    for (size_t i=0; i<m_Nodes.size(); ++i)
        delete m_Nodes[i];
    m_Nodes.clear();

    m_Edges.clear();
    m_CollisionData.Clear();

    if (m_BoundaryNode != NULL)
        m_BoundaryNode->Clear();

    if (m_TriFanNode != NULL) {
       m_TriFanNode->ClearPositions();
       m_TriFanNode->GetVertexBuffer2D().SetActiveBufferSize(0);
    }

    if (m_TriFanNodePass2 != NULL) {
       m_TriFanNodePass2->ClearPositions();       
    }
}

void  CTreeGraphicsModel::SetScheme(CPhyloTreeScheme& sl) 
{ 
    m_SL.Reset(&sl); 

    if (m_BoundaryNode != NULL)
        m_BoundaryNode->SetScheme(sl);
}

void CTreeGraphicsModel::x_RenderNodes(CGlPane& pane,
                                       const CVect3<float>& scale)
{
    /// Visibility for the node is based on visibility of the geometry (triangle fan)
    /// not the associated label, but it would look strange to have the label (which
    /// is larger than the node) without the node.
    if (!m_TriFanNode->IsVisible())
        return;

    // Indices in the datastructure for lower-left and upper right corners of
    // current rectangle
    CTreeCollisionModel2D::TVeci posi_ll = m_CollisionData.GetMinVisIndex();
    CTreeCollisionModel2D::TVeci posi_ur = m_CollisionData.GetMaxVisIndex();

    m_NodeRenderCount = 0;

    CMatrix4<float> mat;
    mat.Identity();  

    // Set flag for all nodes is visible area that they have not been drawn
    // yet.  Must do this since node labels are alpha-blended and so look
    // different if they are drawn multiple tims.
    for (int x = posi_ll.X(); x <= posi_ur.X(); ++x) {
        for (int y = posi_ll.Y(); y <= posi_ur.Y(); ++y) {
            CTreeCollisionModel2D::TVeci idx(x,y);

            // Should always be true unless vis rectangle not intialized
            if (!m_CollisionData.InBounds(idx))
                continue;

            std::vector<size_t>& cell_nodes = m_CollisionData.Get(idx).m_Labels;

            // Find collisions between nodes hashed in this entry   
            for (size_t i=0; i<cell_nodes.size(); ++i) {
                m_Nodes[cell_nodes[i]]->SetAlreadyDrawn(false);
            }
        }
    }

    bool truncate_labels = (m_SL->GetLabelTruncation() == 
                            CPhyloTreeScheme::eLabelsTruncated);

    // Iterate over the currently-visible grid which subdivides 
    // the space occupied by the tree into individual cells, 
    // turning visibility on for all labels in this region
    for (int x = posi_ll.X(); x <= posi_ur.X(); ++x) {
        for (int y = posi_ll.Y(); y <= posi_ur.Y(); ++y) {
            CTreeCollisionModel2D::TVeci idx(x,y);

            // Should always be true unless vis rectangle not intialized
            if (!m_CollisionData.InBounds(idx))
                continue;

            std::vector<size_t>& cell_nodes = m_CollisionData.Get(idx).m_Labels;

            // Find collisions between nodes hashed in this entry   
            for (size_t i=0; i<cell_nodes.size(); ++i) {
                // Label draw checks visibility (updated in UpdateVisibility)
                CTreeGraphicsNode* n = m_Nodes[cell_nodes[i]];      

                if (m_TriFanNode->GetLabelVisibility()) {
                    if (!n->GetAlreadyDrawn()) {

                        n->Render(pane, scale, truncate_labels);
                        n->SetAlreadyDrawn(true);
                    }
                }
                ++m_NodeRenderCount;

                if ((*n->GetSourceNode())->GetChildsDisplay() != IPhyGraphicsNode::eHideChilds &&
                     (*n->GetSourceNode())->GetMarkerColors().size() == 0) {
                    mat[3] = (*n->GetSourceNode())->X();  // mat[3] = mat(0,3)
                    mat[7] = (*n->GetSourceNode())->Y();  // mat[7] = mat(1,3)       
                    mat[0] = n->GetNodeSize();    // mat[0] == mat(0,0)
                    mat[5] = n->GetNodeSize();    // mat[5] == mat(1,1)  

                    CRgbaColor c = n->GetNodeColor();   
                    m_TriFanNode->SetPosition(mat);
                    
                    // For opengl 1.1 set the color for the triangle fan
                    // directly in the buffer since MS OpenGL 1.1 does not
                    // support blending with destination alpha.  Also some linux
                    // gl drivers seem not to support destination alpha.
                    bool onlinux = false;
#ifdef NCBI_OS_LINUX
                    onlinux = true;
#endif

                    if (onlinux || CGlResMgr::Instance().GetApiLevel()==eOpenGL11) {                        
                         CRgbaColor w(1.0f, 1.0f, 1.0f, 1.0f);

                         CGlRenderBuffer<CRgbaColor>& cbuf = m_TriFanNode->GetColorBuffer();
                         CRgbaColor* cptr = cbuf.MapBuffer(GL_READ_WRITE);

                         for (GLsizei cnt=0; cnt<cbuf.GetActiveBufferSize(); ++cnt) {
                             CRgbaColor mix_color = CRgbaColor::Interpolate(w, c, cptr->GetAlpha());
                             mix_color.SetAlpha(cptr->GetAlpha());

                             (*cptr++) = mix_color;
                         }

                         cbuf.UnmapBuffer();
                         m_TriFanNode->Render();
                    }
                    else {                                      
                        m_TriFanNode->Render();
                                       
                        m_TriFanNodePass2->SetPosition(mat);                    
                        m_TriFanNodePass2->GetState().ColorC(c);
                        m_TriFanNodePass2->Render();
                    }
                }
            }
        }
    }
}

void CTreeGraphicsModel::Render(CGlPane& pane)
{
    CStopWatch  render_total;
    render_total.Start();

    CVect3<float> scale(pane.GetScaleX(), pane.GetScaleY(), 1.0f);

    pane.OpenOrtho();
    CGlUtils::CheckGlError();

    if (m_BoundaryNode != NULL) {
        m_BoundaryNode->SetScale(scale);
        m_BoundaryNode->Render();
    }

    CStopWatch  t;
    t.Start();
    CGlModel2D::Render(pane);
    m_LineRenderTime = (float)t.Restart();

    if (m_TriFanNode != NULL) {
        x_RenderNodes(pane, scale);
    }

    m_NodesRenderTime = (float)t.Restart();    

    m_RenderTime = render_total.Elapsed();

    CGlUtils::CheckGlError();
    pane.Close();
}


void CTreeGraphicsModel::x_RenderNodeLabelsPdf(CGlPane& pane,
                                               CRef<CPdf>& pdf,
                                               CVect3<float> scale)
{
    if (!m_TriFanNode->GetLabelVisibility())
        return;

    CGlRender& gl = GetGl();

    bool truncate_labels = (m_SL->GetLabelTruncation() == 
                            CPhyloTreeScheme::eLabelsTruncated);

    float uu_inv = 1.0f/pdf->GetOptions().GetUserUnit();
    scale *= uu_inv;

    // Do all backgrounds first since they are separate objects (in pdf)
    // and we're putting all the text into a single object.
    size_t i;
    for (i=0; i<m_Nodes.size(); ++i) {
        CTreeGraphicsNode* n = m_Nodes[i];      

        if (!n->GetAlreadyDrawn()) {
            string txt;

            if (!(*n->GetSourceNode())->GetLabelBgColor().empty()) {
                CVect2<float> pos = n->GetTextOut(pane, scale, truncate_labels, txt);   
                CVect2<float> posp;                    
                double px, py;
                double dummyz;

                gluProject(pos.X(), pos.Y(), 0.0, 
                   m_ModelviewMatrix, m_ProjectionMatrix, m_Viewport, 
                  &px, &py, &dummyz);
                posp.X() = px;
                posp.Y() = py;

                CGlTextureFont::EFontFace face = n->GetFont()->GetFontFace();
                float font_size = (float)n->GetFont()->GetFontSize();
            
                gl.ColorC(CRgbaColor((*n->GetSourceNode())->GetLabelBgColor()));

                TModelUnit text_h = n->GetFont()->GetMetric(IGlFont::eMetric_CharHeight);
                TModelUnit text_w = n->GetFont()->TextWidth(txt.c_str());

                gl.PolygonMode(GL_FRONT_AND_BACK, GL_FILL);
                gl.RectC(TModelRect(pos.X() - 1.0 * pane.GetScaleX(),
                                    pos.Y() - 1.0 * pane.GetScaleY(),
                                    pos.X() + (text_w+1.0) * pane.GetScaleX(),
                                    pos.Y() + (text_h+1.0) * pane.GetScaleY()));
            } 
        }
    }


    // Now print out all the text
    CRef<CPdfObject> content = pdf->BeginContent(CPdf::ePdfText); 

    for (i=0; i<m_Nodes.size(); ++i) {
        CTreeGraphicsNode* n = m_Nodes[i];      

        if (!n->GetAlreadyDrawn()) {
            string txt;

            CVect2<float> pos = n->GetTextOut(pane, scale, truncate_labels, txt);   
            CVect2<float> posp;                    
            double px, py;
            double dummyz;

            gluProject(pos.X(), pos.Y(), 0.0, 
                m_ModelviewMatrix, m_ProjectionMatrix, m_Viewport, 
                &px, &py, &dummyz);
            posp.X() = px;
            posp.Y() = py;

            CGlTextureFont::EFontFace face = n->GetFont()->GetFontFace();
            float font_size = (float)n->GetFont()->GetFontSize();

            pdf->PrintText(pdf->GetFontHandler(), face, font_size, posp, 
                txt.c_str(), n->GetColor());

            n->SetAlreadyDrawn(true);
        }
    }
    pdf->EndContent();
}

CVect2<float> CTreeGraphicsModel::x_Project(CVect2<float>& p)
{
    double px, py;
    double dummyz;    

    gluProject(p.X(), p.Y(), 0.0, 
        m_ModelviewMatrix, m_ProjectionMatrix, m_Viewport, 
        &px, &py, &dummyz);

    return CVect2<float>((float)px, (float)py);  
}

    
void CTreeGraphicsModel::x_RenderNodesPdf(CGlPane& pane,
                                          CRef<CPdf>& pdf,
                                          CVect3<float> scale)
{
    if (!m_TriFanNode->IsVisible())
        return;

    map<int, string> shader_ids;

    // For rendering to the screen, I do not use a color buffer since there are only two colors
    // (I use a blending trick instead).  Here create the needed color buffer for pdf rendering.
    vector<CRgbaColor> colors, prev_colors;
    CGlRenderBuffer<CRgbaColor>& color_buf = m_TriFanNode->GetColorBuffer();
    color_buf.GetData(prev_colors);

    CRgbaColor cback(1.0f, 1.0f, 1.0f);
    float uu_inv = 1.0f/pdf->GetOptions().GetUserUnit();
    scale *= uu_inv;

    // for each node, check its color and see if we have alread drawn a node
    // of the same color to the pdef file.  If so, just add a new shader 
    // reference to the existing pdf triangle shader with an appropriate 
    // transformation.  If not, create a new pdf shader and add it to the
    // shading dictionary.
    for (size_t i=0; i<m_Nodes.size(); ++i) {
        CTreeGraphicsNode* n = m_Nodes[i];      

        if ((*n->GetSourceNode())->GetChildsDisplay() != IPhyGraphicsNode::eHideChilds &&
            (*n->GetSourceNode())->GetMarkerColors().size() == 0) {

                // create a unique identifer (unsigned int) for the color
                CRgbaColor c = n->GetNodeColor();
                unsigned int id = c.GetRedUC()<<16;
                id |= c.GetGreenUC()<<8;
                id |= c.GetBlueUC();

                // see if we've already created a node of this color
                string shader = shader_ids[id];

                // If not, create a new triangle shader node and map its identifer
                // to the colors unique id value
                if (shader == "") {
                    // Fill color buffer before calling pdf to render tris:
                    colors.clear();
                    colors.push_back(cback);
                    colors.resize(m_TriFanNode->GetVertexBuffer2D().GetActiveBufferSize(), c);
                    colors[0] = cback;
                    color_buf.BufferData(colors.size(), &colors[0], GL_DYNAMIC_DRAW);

                    //shader = pdf->AddShadedTris(m_TriFanNode);
                    shader = pdf->AddShadedTris(m_TriFanNode, CPdfObject::e16Bit, -200, 200, -200, 200);

                    shader_ids[id] = shader;
                }

                // Pdf render function gets transformation values from opengl               
                CMatrix3<double> mat;
                mat.Identity();

                glMatrixMode(GL_MODELVIEW);
                glPushMatrix();
                glTranslatef((*n->GetSourceNode())->X(), (*n->GetSourceNode())->Y(), 0.0f);
                glScalef(n->GetNodeSize() * scale.X(), n->GetNodeSize() * scale.Y(), 1.0f);

                // create a reference in the pdf file to the shader (tris) at the nodes position
                pdf->RenderShaderInstance(m_TriFanNode, shader, pane.GetViewport());
                glPopMatrix();
        }
    }

    // Restore previous colors for node (so that it will work correctly
    // in non-pdf drawing)
    color_buf.BufferData(prev_colors.size(), &prev_colors[0], GL_DYNAMIC_DRAW);
}

void CTreeGraphicsModel::RenderPDF(CGlPane& pane, CRef<CPdf>& pdf)
{  
    pane.OpenOrtho();

    // get viewport directly from pane since size may exceed
    // opengl max viewport size (but not max pdf viewable page size)
    //glGetIntegerv(GL_VIEWPORT, m_Viewport);
    m_Viewport[0] = pane.GetViewport().Left();
    m_Viewport[1] = pane.GetViewport().Bottom();
    m_Viewport[2] = pane.GetViewport().Right();
    m_Viewport[3] = pane.GetViewport().Top();

    glGetDoublev(GL_MODELVIEW_MATRIX, m_ModelviewMatrix);
    glGetDoublev(GL_PROJECTION_MATRIX, m_ProjectionMatrix);

    CVect3<float> scale((float)pane.GetScaleX(), (float)pane.GetScaleY(), 1.0f);

    if (m_BoundaryNode != NULL) {
        m_BoundaryNode->SetScale(scale);
        m_BoundaryNode->RenderPdf(pane, pdf);
    }
    
    pdf->PrintModel(pane, *this);
    x_RenderNodesPdf(pane, pdf, scale);
    x_RenderNodeLabelsPdf(pane, pdf, scale);  

    pane.Close();
}


END_NCBI_SCOPE
