/*  $Id: tree_graphics_node.cpp 28019 2013-05-06 20:50:50Z rafanovi $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Bob Falk
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>
#include <gui/widgets/phylo_tree/tree_graphics_node.hpp>
#include <gui/opengl/glutils.hpp>
#include <gui/opengl/glbitmapfont.hpp>
#include <gui/opengl/gltexturefont.hpp>

BEGIN_NCBI_SCOPE

CTreeGraphicsNode::CTreeGraphicsNode(CPhyloTreeNode* n,
                                     const CRgbaColor& node_color,
                                     const CRgbaColor& text_color,
                                     const string& t,
                                     float node_size,              
#ifndef TEXTURED_FONTS
                                     const CGlBitmapFont* f)
#else
                                     const CGlTextureFont* f)
#endif
: m_Node(n)
, m_NodeColor(node_color)
, m_FgColor(text_color)
, m_Text(t)
, m_NodeSize(node_size)
, m_Font(f)
, m_Visible(true)
, m_AlreadyDrawn(false)
{
}

void CTreeGraphicsNode::Render(CGlPane& pane, 
                               const CVect3<float>& scale,
                               bool truncate_labels) 
{
    if (!m_Visible || m_Font == NULL || m_AlreadyDrawn)
        return;

    CGlRect<float> r = (*m_Node)->GetLabelRect();

    TModelUnit pos_x = (TModelUnit)((*m_Node)->X() +
        (r.Left() * scale.X()));
    TModelUnit pos_y = (TModelUnit)((*m_Node)->Y() + 
        (r.Bottom() * scale.Y()));

    string text_out;

    if (truncate_labels) {
        TModelUnit screen_x = pane.ProjectX(pos_x); 

        // Label may go off right side of screen so see if it has
        // to be truncated on the right;
        if (screen_x > 0.0) {
            TModelUnit w = pane.GetViewport().Width()-screen_x;
            text_out = m_Font->Truncate(m_Text.c_str(), w);
        }
        // Label starts before left side of screen (definitly clipped)
        // so truncate it on the left
        else {
            text_out = m_Text;
            // Reverse the text so that the beginning (left side) is truncated
            // then set width to total width - width of text to left of viewport
            std::reverse(text_out.begin(), text_out.end());
            text_out = m_Font->Truncate(text_out.c_str(), r.Width()+screen_x);
            std::reverse(text_out.begin(), text_out.end());
            // After truncation, draw text from left edge of viewport
            pos_x = pane.UnProjectX(0);
        }           
    }
    else {
        text_out = m_Text;
    }

    // Draw label background, if it has one specified
    if ((**m_Node).GetLabelBgColor().size() > 0) {
        glColorC(CRgbaColor((**m_Node).GetLabelBgColor()));

        TModelUnit text_h = m_Font->GetMetric(IGlFont::eMetric_CharHeight);
        TModelUnit text_w = m_Font->TextWidth(text_out.c_str());

        glEnable(GL_BLEND);
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
        glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

        glBegin(GL_TRIANGLE_STRIP);
            glVertex2d(pos_x - 1.0 * pane.GetScaleX(),          pos_y + (text_h+1.0) * pane.GetScaleY());
            glVertex2d(pos_x - 1.0 * pane.GetScaleX(),          pos_y - 1.0*pane.GetScaleY());
            glVertex2d(pos_x + (text_w+1.0) * pane.GetScaleX(), pos_y + (text_h+1.0) * pane.GetScaleY());
            glVertex2d(pos_x + (text_w+1.0) * pane.GetScaleX(), pos_y - 1.0*pane.GetScaleY());        
        glEnd();
    } 

    glColor4fv(m_FgColor.GetColorArray());
    m_Font->TextOut(pos_x, pos_y, text_out.c_str());
}

CVect2<float> CTreeGraphicsNode::GetTextOut(CGlPane& pane, 
                                            const CVect3<float>& scale,
                                            bool truncate_labels,
                                            string& text_out)
{
    CGlRect<float> r = (*m_Node)->GetLabelRect();

    CVect2<float> pos;

    pos.X() = (float)((*m_Node)->X() +
        (r.Left() * scale.X()));
    pos.Y() = (float)((*m_Node)->Y() + 
        (r.Bottom() * scale.Y()));

    if (truncate_labels) {
        float screen_x = (float)(pane.ProjectX(pos.X()));

        // Label may go off right side of screen so see if it has
        // to be truncated on the right;
        if (screen_x > 0.0f) {
            float w = ((float)pane.GetViewport().Width()-screen_x);
            text_out = m_Font->Truncate(m_Text.c_str(), w);
        }
        // Label starts before left side of screen (definitly clipped)
        // so truncate it on the left
        else {
            text_out = m_Text;
            // Reverse the text so that the beginning (left side) is truncated
            // then set width to total width - width of text to left of viewport
            std::reverse(text_out.begin(), text_out.end());
            text_out = m_Font->Truncate(text_out.c_str(), r.Width()+screen_x);
            std::reverse(text_out.begin(), text_out.end());
            // After truncation, draw text from left edge of viewport
            pos.X() = (float)pane.UnProjectX(0.0);        
        }
    }
    else {
        text_out = m_Text;
    }

    return pos;
}


void CTreeGraphicsNode::GetLabelRect(const CVect2<float>& scale, 
                                     CVect2<float>& lower_left,
                                     CVect2<float>& upper_right)
{
    CGlRect<float> r = (*m_Node)->GetLabelRect();
    lower_left.X() = (*m_Node)->X() + r.Left()*scale.X();
    lower_left.Y() = (*m_Node)->Y() + r.Bottom()*scale.Y();

    upper_right.X() = (*m_Node)->X() + r.Right()*scale.X();
    upper_right.Y() = (*m_Node)->Y() + r.Top()*scale.Y();
}

void CTreeGraphicsNode::GetBoundingRect(const CVect2<float>& scale, 
                                        CVect2<float>& lower_left,
                                        CVect2<float>& upper_right)
{
    GetLabelRect(scale, lower_left, upper_right);

    float deltax = m_NodeSize*scale.X();
    float deltay = m_NodeSize*scale.Y();

    lower_left.X() = std::min(lower_left.X(), (*m_Node)->X()-deltax);
    lower_left.Y() = std::min(lower_left.Y(), (*m_Node)->Y()-deltay);
    upper_right.X() = std::max(upper_right.X(), (*m_Node)->X()+deltax);
    upper_right.Y() = std::max(upper_right.Y(), (*m_Node)->Y()+deltay);
}

float CTreeGraphicsNode::GetMinX(const CVect2<float>& scale)
{
    return ((*m_Node)->X() + ((*m_Node)->GetLabelRect().Left() * scale.X()));
}


END_NCBI_SCOPE
