/*  $Id: asn_view_context.cpp 25887 2012-05-30 15:35:23Z katargir $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Roman Katargin
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>

#include <gui/objutils/interface_registry.hpp>
#include <gui/objutils/undo_manager_interface.hpp>

#include <gui/widgets/seq/asn_view_context.hpp>
#include <gui/widgets/text_widget/composite_text_item.hpp>
#include <gui/widgets/text_widget/expand_item.hpp>
#include <gui/widgets/seq/asn_header_item.hpp>
#include <gui/widgets/seq/asn_element_block_start.hpp>
#include <gui/widgets/seq/asn_element_item.hpp>

#include <gui/widgets/text_widget/plain_text_item.hpp>

#include <gui/objutils/registry.hpp>

#include <wx/menu.h>
#include <wx/settings.h>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

static const char
*kShowAsnTypes = "GBENCH.TextView.Asn.ShowAsnTypes",
*kShowPathToRoot = "GBENCH.TextView.Asn.ShowPathToRoot";

CAsnViewContext::CAsnViewContext(objects::CScope& scope, IBioseqEditor* bioseqEditor, bool tooltpContext)
    : CTextPanelContext(1),
      m_Scope(&scope),
      m_BioseqEditor(bioseqEditor),
      m_TooltipContext(tooltpContext),
      m_ShowAsnTypes(false),
      m_ShowPathToRoot(false)
{
    CGuiRegistry& gui_reg = CGuiRegistry::GetInstance();
    m_ShowAsnTypes = gui_reg.GetBool(kShowAsnTypes, m_ShowAsnTypes);
    m_ShowPathToRoot = gui_reg.GetBool(kShowPathToRoot, m_ShowPathToRoot);

    if (m_BioseqEditor)
        m_BioseqEditor->SetCallBack(this);
}

void CAsnViewContext::SetPanel(CTextItemPanel* panel)
{
    if (m_Panel && m_BioseqEditor) {
        m_Panel->PopEventHandler();
    }

    CTextPanelContext::SetPanel(panel);

    if (m_Panel && m_BioseqEditor) {
        m_Panel->PushEventHandler(m_BioseqEditor);
    }
}

CAsnViewContext::~CAsnViewContext()
{
    ITERATE(vector<CAsnBioContext*>, it, m_Contexts) {
        delete *it;
    }
}

void CAsnViewContext::SetShowAsnTypes(bool show)
{
    if (m_ShowAsnTypes == show)
        return;

    m_ShowAsnTypes = show;

    CGuiRegistry& gui_reg = CGuiRegistry::GetInstance();
    gui_reg.Set(kShowAsnTypes, m_ShowAsnTypes);

    if (m_Panel)
        m_Panel->Refresh();
}

void CAsnViewContext::SetShowPathToRoot(bool show)
{
    if (m_ShowPathToRoot == show)
        return;

    m_ShowPathToRoot = show;

    CGuiRegistry& gui_reg = CGuiRegistry::GetInstance();
    gui_reg.Set(kShowPathToRoot, m_ShowPathToRoot);
}

void CAsnViewContext::MouseMoveInContainer(ITextItem* container)
{
    CCompositeTextItem* composite = dynamic_cast<CCompositeTextItem*>(container);
    if (composite == 0 || composite->GetItemCount() == 0)
        return;

    CAsnHeaderItem* headerItem = dynamic_cast<CAsnHeaderItem*>(composite->GetItem(0));
    if (headerItem) {
        AddHier(headerItem);
        return;
    }

    CAsnElementBlockStart* blockStart = dynamic_cast<CAsnElementBlockStart*>(composite->GetItem(0));
    if (blockStart) {
        AddHier(blockStart);
        return;
    }
}

wxWindow* CAsnViewContext::CreateTooltipWindow()
{
    if (!m_ShowPathToRoot || m_MouseOverItem == 0 || m_Hier.empty())
        return NULL;

    CTextItemPanel* tooltipWnd = new CTextItemPanel( m_Panel, wxID_ANY, wxDefaultPosition, wxSize(0,0), wxBORDER_SIMPLE|wxHSCROLL|wxVSCROLL );
    tooltipWnd->Hide();
    tooltipWnd->SetBackgroundColour(wxSystemSettings::GetColour(wxSYS_COLOUR_INFOBK));

    CCompositeTextItem* mainItem = new CCompositeTextItem();

    CPlainTextItem* info = new CPlainTextItem();
    info->AddLine("Path to Root:");
    mainItem->AddItem(info);

    for (size_t i = 0; i < m_Hier.size(); ++i) {
        ITextItem* item = m_Hier[i]->Clone();
        if (item) {
            mainItem->AddItem(item);
        }
    }

    tooltipWnd->SetData(mainItem, new CAsnViewContext(*m_Scope, 0, true));
    tooltipWnd->RecalSize();
    wxSize size = mainItem->GetSize();
    tooltipWnd->SetClientSize(size);
    tooltipWnd->Layout();

    return tooltipWnd;
}

wxMenu* CAsnViewContext::CreateMenu() const
{
    if (m_BioseqEditor) {
        return m_BioseqEditor->CreateContextMenu();
    }
    return 0;
}

void CAsnViewContext::GetCurrentSelection(TEditObjectList& objects)
{
    ITERATE (set<ITextItem*>, it, m_SelectedItems) {
        CAsnElementItem* asnItem = 0;
        CExpandItem* expandItem = dynamic_cast<CExpandItem*>(*it);
        if (expandItem) {
            asnItem = dynamic_cast<CAsnElementItem*>(expandItem->GetCollapsedItem());
        }
        else
            asnItem = dynamic_cast<CAsnElementItem*>(*it);

        if (!asnItem)
            continue;

        CAsnBioContext* bioContext = asnItem->GetBioContext();

        TTypeInfo type = asnItem->m_Type;
        TConstObjectPtr ptr = asnItem->m_Ptr;

        ETypeFamily family = type->GetTypeFamily();
        while (family == eTypeFamilyPointer) {
            const CPointerTypeInfo* pointerType =
                CTypeConverter<CPointerTypeInfo>::SafeCast(type);
            ptr = pointerType->GetObjectPointer(ptr);
            type = pointerType->GetPointedType();
            family = type->GetTypeFamily();
        }

        if (family == eTypeFamilyClass) {
            const CClassTypeInfo* classType =
                CTypeConverter<CClassTypeInfo>::SafeCast(type);
            if (type->GetName() == "Seq-feat" && bioContext) {
                objects.push_back(
                    CEditObject(*classType->GetCObjectPtr(ptr),
                                bioContext->GetSeq_entry_Handle()));
            }
            if (type->GetName() == "Submit-block") {
                objects.push_back(
                    CEditObject(*classType->GetCObjectPtr(ptr),
                                *m_Scope));
            }
        }
        else if (family == eTypeFamilyChoice) {
            if (type->GetName() == "Seqdesc" && bioContext) {
                const CChoiceTypeInfo* choiceType =
                    CTypeConverter<CChoiceTypeInfo>::SafeCast(type);
                objects.push_back(
                    CEditObject(*choiceType->GetCObjectPtr(ptr),
                                bioContext->GetSeq_entry_Handle()));
            }
        }
    }
}

CBioseq_Handle CAsnViewContext::GetCurrentBioseq()
{
    return CBioseq_Handle();
}

END_NCBI_SCOPE
