/*  $I$
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Roman Katargin
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>

#include <objmgr/bioseq_handle.hpp>
#include <objmgr/bioseq_ci.hpp>
#include <objmgr/util/sequence.hpp>

#include <objtools/format/flat_file_config.hpp>
#include <objtools/format/context.hpp>
#include <objtools/format/item_ostream.hpp>
#include <objtools/format/item_formatter.hpp>
#include <objtools/format/gather_items.hpp>

#include <objtools/format/items/item.hpp>
#include <objtools/format/items/locus_item.hpp>
#include <objtools/format/items/defline_item.hpp>
#include <objtools/format/items/accession_item.hpp>
#include <objtools/format/items/version_item.hpp>
#include <objtools/format/items/dbsource_item.hpp>
#include <objtools/format/items/keywords_item.hpp>
#include <objtools/format/items/source_item.hpp>
#include <objtools/format/items/reference_item.hpp>
#include <objtools/format/items/dbsource_item.hpp>
#include <objtools/format/items/comment_item.hpp>
#include <objtools/format/items/feature_item.hpp>
#include <objtools/format/items/basecount_item.hpp>
#include <objtools/format/items/origin_item.hpp>
#include <objtools/format/items/locus_item.hpp>
#include <objtools/format/items/sequence_item.hpp>
#include <objtools/format/items/contig_item.hpp>
#include <objtools/format/items/ctrl_items.hpp>
#include <objtools/format/items/segment_item.hpp>
#include <objtools/format/items/genome_project_item.hpp>

#include <gui/widgets/seq/flat_file_context.hpp>
#include <gui/widgets/seq/flat_file_text_item.hpp>
#include <gui/widgets/seq/flat_file_header_item.hpp>
#include <gui/widgets/seq/flat_file_feature_collapsed.hpp>
#include <gui/widgets/seq/flat_file_seqblock_collapsed.hpp>
#include <gui/widgets/seq/flat_file_view_params.hpp>

#include <gui/widgets/text_widget/expand_item.hpp>
#include <gui/widgets/text_widget/composite_text_item.hpp>
#include <gui/widgets/text_widget/draw_styled_text.hpp>
#include <gui/widgets/text_widget/text_item_panel.hpp>
#include <gui/widgets/text_widget/plain_text_item.hpp>
#include <gui/widgets/seq/flat_file_locus_item.hpp>

#include "features_group.hpp"
#include "sequence_group.hpp"

#include <gui/widgets/seq/flat_file_retrieve_job.hpp>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

namespace // anonymous
{

enum EFlatfileRowType {
    eRow_unknown,

    eRow_locus,
    eRow_defline,
    eRow_accession,
    eRow_genom_project,
    eRow_version,
    eRow_dbsource,
    eRow_keywords,
    eRow_source,
    eRow_reference,
    eRow_comment,
    eRow_feature_header,
    eRow_source_feature,
    eRow_feature,
    eRow_base_count,
    eRow_origin,
    eRow_contig,
    eRow_sequence,
    // XXX - add the rest

    // control items
    eRow_begin,
    eRow_begin_section,
    eRow_end_section,
    eRow_end
};

enum EChapterType
{
    eChapter_unknown,

    eChapter_header,
    eChapter_references,
    eChapter_comment,
    eChapter_features,
    eChapter_sequence
};

static EFlatfileRowType s_GetFlatRowType(const IFlatItem* itemp)
{
    if (itemp == NULL) {
        return eRow_unknown;
    } else if (dynamic_cast<const CLocusItem*>(itemp) != NULL) {
        return eRow_locus;
    } else if (dynamic_cast<const CDeflineItem*>(itemp) != NULL) {
        return eRow_defline;
    } else if (dynamic_cast<const CAccessionItem*>(itemp) != NULL) {
        return eRow_accession;
    } else if (dynamic_cast<const CGenomeProjectItem*>(itemp) != NULL) {
        return eRow_genom_project;
    } else if (dynamic_cast<const CVersionItem*>(itemp) != NULL) {
        return eRow_version;
    } else if (dynamic_cast<const CDBSourceItem*>(itemp) != NULL) {
        return eRow_dbsource;
    } else if (dynamic_cast<const CKeywordsItem*>(itemp) != NULL) {
        return eRow_keywords;
    } else if (dynamic_cast<const CSourceItem*>(itemp) != NULL) {
        return eRow_source;
    } else if (dynamic_cast<const CReferenceItem*>(itemp) != NULL) {
        return eRow_reference;
    } else if (dynamic_cast<const CCommentItem*>(itemp) != NULL) {
        return eRow_comment;
    } else if (dynamic_cast<const CFeatHeaderItem*>(itemp) != NULL) {
        return eRow_feature_header;
    } else if (dynamic_cast<const CSourceFeatureItem*>(itemp) != NULL) {
        return eRow_source_feature;
    } else if (dynamic_cast<const CFeatureItem*>(itemp) != NULL) {
        return eRow_feature;
    } else if (dynamic_cast<const CBaseCountItem*>(itemp) != NULL) {
        return eRow_base_count;
    } else if (dynamic_cast<const COriginItem*>(itemp) != NULL) {
        return eRow_origin;
    } else if (dynamic_cast<const CContigItem*>(itemp) != NULL) {
        return eRow_contig;
    } else if (dynamic_cast<const CSequenceItem*>(itemp) != NULL) {
        return eRow_sequence;
    } else if (dynamic_cast<const CStartItem*>(itemp) != NULL) {
        return eRow_begin;
    } else if (dynamic_cast<const CStartSectionItem*>(itemp) != NULL) {
        return eRow_begin_section;
    } else if (dynamic_cast<const CEndSectionItem*>(itemp) != NULL) {
        return eRow_end_section;
    } else if (dynamic_cast<const CEndItem*>(itemp) != NULL) {
        return eRow_end;
    }

    return eRow_unknown;
}

EChapterType s_GetChapterForItem(EFlatfileRowType i_type)
{
    switch (i_type) {
    case eRow_locus:
    case eRow_defline:
    case eRow_accession:
    case eRow_genom_project:
    case eRow_version:
    case eRow_dbsource:
    case eRow_keywords:
    case eRow_source:
        return eChapter_header;

    case eRow_reference:
        return eChapter_references;

    case eRow_comment:
        return eChapter_comment;

    case eRow_feature_header:
    case eRow_source_feature:
    case eRow_feature:
        return eChapter_features;

    case eRow_base_count:
    case eRow_origin:
    case eRow_contig:
    case eRow_sequence:
        return eChapter_sequence;

    case eRow_begin:
    case eRow_begin_section:
    case eRow_end_section:
    case eRow_end:
    case eRow_unknown:
    default:
        break;
    }

    return eChapter_unknown;
}

} // anonymous namespace

const size_t kMinSeqBlockDigits = 4;

class CFlatFileFactoryBuilder : public CFlatItemOStream
{
public:
    CFlatFileFactoryBuilder(CFlatFileRetrieveJob& job)
        : m_Job(job), m_ChapterType(eChapter_unknown), m_CurrentSection(0), m_CurrentChapter(0), m_SeqBlockDigits(kMinSeqBlockDigits) {}

private:
    virtual void AddItem(CConstRef<IFlatItem> item);

    CFlatFileRetrieveJob& m_Job;
    EChapterType m_ChapterType;
    CCompositeTextItem* m_CurrentSection;
    CCompositeTextItem* m_CurrentChapter;

    size_t m_SeqBlockDigits;
    string m_SeqName;
};

static size_t s_CalcSeqBlockDigits(size_t length)
{
    const size_t kLineSize = 60;

    if (length == 0)
        return kMinSeqBlockDigits;

    size_t numLines = (length + kLineSize - 1)/kLineSize;
    size_t maxNumber = (numLines - 1)*kLineSize + 1;
    size_t digits = 0;
    do {
        maxNumber /= 10;
        ++digits;
    } while (maxNumber > 0);

    return (kMinSeqBlockDigits < digits) ? digits : kMinSeqBlockDigits;
}

void CFlatFileFactoryBuilder::AddItem(CConstRef<IFlatItem> item)
{
    if (m_Job.IsCanceled())
        throw CTextRetrieveJob::CCancelException();

    const IFlatItem* itemp = item.GetPointerOrNull();
    if (itemp == NULL)
        return;

    EFlatfileRowType row_type = s_GetFlatRowType(itemp);
    EChapterType chap_type = s_GetChapterForItem(row_type);

    if (row_type == eRow_begin)
        return;

    if (row_type == eRow_begin_section) {
        m_SeqBlockDigits = kMinSeqBlockDigits;
        m_CurrentSection = new CCompositeTextItem();
        //m_Job.x_AddItem(m_CurrentSection, 0);
        return;
    }
    else if (row_type == eRow_end_section) {
        m_CurrentChapter = 0;

        CPlainTextItem* textItem = new CPlainTextItem();
        textItem->AddLine("//");
        textItem->AddLine("");
        m_Job.x_AddItem(textItem, m_CurrentSection);

        CExpandItem* expandItem =
            new CExpandItem(new CFlatFileHeader(m_SeqName), m_CurrentSection, true);
        expandItem->SetSelectable(false);
        m_Job.x_AddItem(expandItem, 0);

        m_SeqName.clear();

        return;
    }
    else if (row_type == eRow_locus) {
        const CLocusItem* locusItem = dynamic_cast<const CLocusItem*>(itemp);
        m_SeqName = locusItem->GetName();
        CBioseqContext* ctx = locusItem->GetContext();
        if (ctx) {
            CBioseq_Handle& handle = ctx->GetHandle();
            if (handle) {
                m_SeqBlockDigits = s_CalcSeqBlockDigits(handle.GetBioseqLength());
            }
        }
    }

    if (!m_CurrentSection) {
        return;
    }

    if (row_type == eRow_unknown)
        return;

    if (row_type == eRow_end || row_type == eRow_end_section || chap_type != m_ChapterType)
        m_CurrentChapter = 0;

    if (row_type == eRow_end)
        return;

    if (m_CurrentChapter == 0) {

        m_CurrentChapter = new CCompositeTextItem();
        m_ChapterType = chap_type;

        switch (m_ChapterType) {
        case eChapter_header:
            {{
                //CExpandItem* expandItem =
                    //new CExpandItem(new CFlatFileHeader("HEADER ..."), m_CurrentChapter, true);
                //expandItem->SetSelectable(false);
                //m_Job.x_AddItem(expandItem, m_CurrentSection);
                m_Job.x_AddItem(m_CurrentChapter, m_CurrentSection);
            }}
            break;
        case eChapter_references:
            {{
                CExpandItem* expandItem =
                    new CExpandItem(new CFlatFileHeader("PUBLICATIONS ..."), m_CurrentChapter, true);
                expandItem->SetSelectable(false);
                m_Job.x_AddItem(expandItem, m_CurrentSection);
            }}
            break;
        case eChapter_comment:
            {{
                CExpandItem* expandItem =
                    new CExpandItem(new CFlatFileHeader("COMMENT ..."), m_CurrentChapter, true);
                expandItem->SetSelectable(false);
                m_Job.x_AddItem(expandItem, m_CurrentSection);
            }}
            break;
        case eChapter_features:
            {{
                CExpandItem* expandItem =
                    new CFeaturesGroup(new CFlatFileHeader("FEATURES ..."), m_CurrentChapter, true);
                m_Job.x_AddItem(expandItem, m_CurrentSection);
            }}
            break;
        case eChapter_sequence:
            m_Job.x_AddItem(
                new CSequenceGroup(new CFlatFileHeader("SEQUENCE ..."), m_CurrentChapter, true), m_CurrentSection);
            break;
        default:
            m_Job.x_AddItem(new CFlatFileTextItem(item), m_CurrentSection);
            return;
        }
    }

    if (eRow_source_feature == row_type || eRow_feature == row_type) {
        CFlatFileTextItem* flatFileItem = new CFlatFileTextItem(item);
        flatFileItem->SetSelectable(false);
        m_Job.x_AddItem(new CExpandItem(new CFlatFileFeatureCollapsed(item),
                            flatFileItem, false), m_CurrentChapter);
    }
    else if (eRow_sequence == row_type) {
        CFlatFileTextItem* flatFileItem = new CFlatFileTextItem(item);
        flatFileItem->SetSelectable(false);
        m_Job.x_AddItem(new CExpandItem(new CFlatFileSeqBlockCollapsed(item, m_SeqBlockDigits),
                            flatFileItem, false), m_CurrentChapter);
    }
    else if (eRow_locus == row_type) {
        m_Job.x_AddItem(new CFlatFileLocusItem(item), m_CurrentChapter);
    }
    else
        m_Job.x_AddItem(new CFlatFileTextItem(item), m_CurrentChapter);
}

CFlatFileRetrieveJob::CFlatFileRetrieveJob(
    const CFlatFileConfig& config,
    CFlatFileContext& context)
: m_Config(config), m_Context(&context)
{
}

CFlatFileRetrieveJob::~CFlatFileRetrieveJob()
{
}

void CFlatFileRetrieveJob::x_AddItem(ITextItem* item, CCompositeTextItem* container)
{
    CFastMutexGuard guard(m_Mutex);
    NON_CONST_ITERATE(CCompositeTextItem::TInsertBlockVec, it, m_Items) {
        if (it->m_Container == container) {
            it->m_Items.push_back(item);
            return;
        }
    }
    if (container == 0)
        m_Items.push_back(CCompositeTextItem::TInsertBlock(item, -2));
    else
        m_Items.push_back(CCompositeTextItem::TInsertBlock(item, container, -1));
}

IAppJob::EJobState CFlatFileRetrieveJob::x_Run()
{
    CRef<CFlatFileFactoryBuilder> builder(new CFlatFileFactoryBuilder(*this));

    CRef<CFlatGatherer> gatherer(CFlatGatherer::New(m_Config.GetFormat()));
    gatherer->Gather(*m_Context, *builder);

    CFastMutexGuard guard(m_Mutex);
    if (!m_Items.empty()) {
        m_Result.Reset(new CTextViewRetrieveResult());
        m_Result->SetItems(m_Items);
        m_Items.clear();
    }

    return eCompleted;
}

CConstIRef<IAppJobProgress> CFlatFileRetrieveJob::GetProgress()
{
    // Dynamic FlatFile view creation causes race condition
    // in FlatFile generator code - crash on Linuix/Mac
/*
    CFastMutexGuard guard(m_Mutex);
    CTextViewRetrieveResult* result = new CTextViewRetrieveResult();

    result->SetItems(m_Items);
    m_Items.clear();
 */
    CTextViewRetrieveProgress* progress = new CTextViewRetrieveProgress(NULL);
    return CConstIRef<IAppJobProgress>(progress);
}

END_NCBI_SCOPE
