/*  $Id: flat_file_text_item.cpp 27074 2012-12-14 19:37:37Z katargir $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Roman Katargin
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>

#include <corelib/ncbistre.hpp>

#include <objmgr/seqdesc_ci.hpp>

#include <objtools/format/text_ostream.hpp>
#include <objtools/format/items/feature_item.hpp>
#include <objtools/format/items/sequence_item.hpp>

#include <gui/widgets/text_widget/draw_styled_text.hpp>
#include <gui/widgets/text_widget/selection_helper.hpp>

#include <gui/widgets/seq/flat_file_text_item.hpp>
#include <gui/widgets/seq/flat_file_context.hpp>
#include <gui/widgets/seq/flat_file_view_params.hpp>

#include <gui/objutils/seqdesc_title_edit.hpp>

#include <objtools/format/items/defline_item.hpp>
#include <objtools/format/items/source_item.hpp>
#include <objtools/format/items/comment_item.hpp>
#include <objtools/format/items/feature_item.hpp>
#include <objtools/format/items/genome_project_item.hpp>

#include <wx/dcclient.h>
#include <wx/txtstrm.h>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

namespace // anonymous
{

class CCountLines : public IFlatTextOStream
{
public:
    CCountLines() : m_Lines(0) {}

    virtual void AddParagraph(const list<string>&  text, const CSerialObject*)
    {
        m_Lines += text.size();
    }
    virtual void AddLine(const CTempString& line,
                         const CSerialObject* obj,
                         EAddNewline add_newline)
    {
        if (eAddNewline_Yes == add_newline)
            ++m_Lines;
    }

    size_t GetLines() const { return m_Lines; }

private:
    size_t m_Lines;
};

class CGetText : public IFlatTextOStream
{
public:
    CGetText() {}

    virtual void AddParagraph(const list<string>&  text, const CSerialObject*)
    {
        ITERATE(list<string>, it, text)
            m_OS << *it << "\n";
    }
    virtual void AddLine(const CTempString& line,
                         const CSerialObject* obj,
                         EAddNewline add_newline)
    {
        m_OS << line;
        if (eAddNewline_Yes == add_newline)
            m_OS << "\n";
    }

    string Text() { return CNcbiOstrstreamToString(m_OS); }

private:
    CNcbiOstrstream m_OS;
};

class CDrawLines : public IFlatTextOStream
{
public:
    CDrawLines(CStyledTextOStream& drawText) : m_DrawText(drawText) {}

    virtual void AddParagraph(const list<string>&  text, const CSerialObject*)
    {
        ITERATE(list<string>, it, text)
            m_DrawText << *it << NewLine();
    }
    virtual void AddLine(const CTempString& line,
                         const CSerialObject* obj,
                         EAddNewline add_newline)
    {
        m_DrawText << line;
        if (eAddNewline_Yes == add_newline)
            m_DrawText << NewLine();
    }

private:
    CStyledTextOStream& m_DrawText;
};

} // anonymous namespace


const CConstRef<CObject> CFlatFileTextItem::GetAssosiatedObject() const
{
    if (m_Item) {
        const CDeflineItem* defLineItem = dynamic_cast<const CDeflineItem*>(m_Item.GetPointer());
        if (defLineItem) {
            CBioseqContext* ctx = defLineItem->GetContext();
            if (!ctx)
                return null;

            return CConstRef<CObject>(
                new CSeqdescTitleEdit(ctx->GetHandle(), defLineItem->GetDefline()));
        }
        const CSourceItem* sourceItem = dynamic_cast<const CSourceItem*>(m_Item.GetPointer());
        if (sourceItem) {
            CBioseqContext* ctx = sourceItem->GetContext();
            if (!ctx)
                return null;

            CSeqdesc_CI dsrc_it(ctx->GetHandle(), CSeqdesc::e_Source);
            if ( dsrc_it ) {
                const CSeqdesc& seqDesc = *dsrc_it;
                return CConstRef<CObject>(&seqDesc);
            }
            const CSeq_feat* seq_feat = dynamic_cast<const CSeq_feat*>(m_Item->GetObject());
            if (seq_feat && seq_feat->GetData().IsBiosrc()) {
                return CConstRef<CObject>(seq_feat);
            }
        }

        const CSourceFeatureItem* sourceFeatureItem 
            = dynamic_cast<const CSourceFeatureItem*>(m_Item.GetPointer());
        if (sourceFeatureItem) {
            CBioseqContext* ctx = sourceFeatureItem->GetContext();
            if (!ctx)
                return null;

            const CSeq_feat* seq_feat = dynamic_cast<const CSeq_feat*>(m_Item->GetObject());
            if (!seq_feat)
                return null;

            const CSeq_feat::TData& data = seq_feat->GetData();
            if (data.IsBiosrc()) {
                const CBioSource& bioSource = data.GetBiosrc();
                CSeqdesc_CI dsrc_it(ctx->GetHandle(), CSeqdesc::e_Source);
                if ( dsrc_it ) {
                    const CSeqdesc& seqDesc = *dsrc_it;
                    if (seqDesc.IsSource()) {
                        const CBioSource& bioSource2 = seqDesc.GetSource();
                        if (&bioSource2 == &bioSource) {
                            return CConstRef<CObject>(&seqDesc);
                        }
                    }
                }
            }

            CFeat_CI fsrc_it(ctx->GetHandle(), CSeqFeatData::e_Biosrc);
            if (fsrc_it) {
                return CConstRef<CObject>(fsrc_it->GetOriginalSeq_feat().GetPointer());
            }
            return null;
        }

        const CGenomeProjectItem* genomeProjectItem 
            = dynamic_cast<const CGenomeProjectItem*>(m_Item.GetPointer());
        if (genomeProjectItem ) {
            CBioseqContext* ctx = genomeProjectItem ->GetContext();
            if (!ctx)
                return null;

            // extract all the useful user objects
            for (CSeqdesc_CI desc(ctx->GetHandle(), CSeqdesc::e_User);  desc;  ++desc) {
                const CUser_object& uo = desc->GetUser();

                if ( !uo.GetType().IsStr() ) {
                    continue;
                }
                if( NStr::EqualNocase( uo.GetType().GetStr(), "DBLink" ) ) {
                    const CSeqdesc& seqDesc = *desc;
                    return CConstRef<CObject>(&seqDesc);
                }
            }
            return  null;
        }

        const CSerialObject* itemObject = m_Item->GetObject();
        if (!itemObject)
            return null;

        const CBioseq* bioseq = dynamic_cast<const CBioseq*>(itemObject);
        const CSequenceItem* seqItem =
            dynamic_cast<const CSequenceItem*>(m_Item.GetPointer());
        if (seqItem && bioseq) {
            CRef<CSeq_loc> loc(new CSeq_loc());
            loc->SetInt().SetFrom(seqItem->GetFrom()-1);
            loc->SetInt().SetTo(seqItem->GetTo()-1);
            try {
                loc->SetId(*(bioseq->GetFirstId()));
            }
            catch (const CObjMgrException&) {
                return null;
            }
            return CConstRef<CObject>(loc.GetPointer());
        }
        else 
            return CConstRef<CObject>(itemObject);
    }
    return null;
}

void CFlatFileTextItem::CalcSize(wxDC& dc, CTextPanelContext* context)
{
    CFlatFileViewContext* pcntx = static_cast<CFlatFileViewContext*>(context);
    CCountLines os;

    if (!m_Item->Skip()) {
        m_Item->Format(pcntx->GetFormatter(), os);
        m_LineCount = os.GetLines();
    }
    else
        m_LineCount = 1;

    wxSize size = dc.GetTextExtent(wxString(80, wxChar('W')));
    size.SetHeight(size.GetHeight()*m_LineCount);
    size.IncBy(context->GetLeftMargin(), 0);
    m_Size = size;
}

int CFlatFileTextItem::GetTextOffset(CTextPanelContext* context) const
{
    return context->GetLeftMargin();
}

void CFlatFileTextItem::x_RenderText(CStyledTextOStream& ostream, CTextPanelContext* context) const
{
    CFlatFileViewContext* pcntx = static_cast<CFlatFileViewContext*>(context);
    CDrawLines draw(ostream);
    if (!m_Item->Skip())
        m_Item->Format(pcntx->GetFormatter(), draw);
}

string CFlatFileTextItem::GetToolTipText(CTextPanelContext* context) const
{
    if (dynamic_cast<const CFeatureItemBase*>(m_Item.GetPointer()) == NULL)
        return "";

    CFlatFileViewContext* pcntx = static_cast<CFlatFileViewContext*>(context);
    CGetText os;
    if (!m_Item->Skip())
        m_Item->Format(pcntx->GetFormatter(), os);
    return os.Text();
}

END_NCBI_SCOPE
