/*  $Id: text_ds.cpp 26031 2012-06-29 15:00:37Z katargir $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Roman Katargin
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>

#include <gui/utils/event_translator.hpp>

#include <gui/utils/extension_impl.hpp>
#include <gui/widgets/wx/ieditor_factory.hpp>

#include <gui/widgets/text_widget/text_item_panel.hpp>
#include <gui/widgets/text_widget/composite_text_item.hpp>
#include <gui/widgets/text_widget/plain_text_item.hpp>
#include <gui/widgets/text_widget/draw_styled_text.hpp>

#include <gui/widgets/seq/fasta_retrieve_job.hpp>
#include <gui/widgets/seq/flat_file_retrieve_job.hpp>
#include <gui/widgets/seq/flat_feature_retrieve_job.hpp>
#include <gui/widgets/seq/asn_retrieve_job.hpp>
#include <gui/widgets/seq/xml_retrieve_job.hpp>

#include <gui/widgets/seq/flat_file_context.hpp>
#include <gui/widgets/seq/asn_view_context.hpp>
#include <gui/widgets/seq/text_ds.hpp>

#include <objtools/format/item_formatter.hpp>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

BEGIN_EVENT_MAP(CTextDataSource, CEventHandler)
    ON_MESSAGE(CAppJobNotification, CAppJobNotification::eStateChanged,
                &CTextDataSource::OnJobNotification)
    ON_MESSAGE(CAppJobNotification, CAppJobNotification::eProgress,
                &CTextDataSource::OnProgress)
END_EVENT_MAP()

////////////////////////////////////////////////////////////////////////////////

CTextViewRetrieveResult::CTextViewRetrieveResult() 
{
}

CTextViewRetrieveResult::~CTextViewRetrieveResult()
{
    x_FreeItems();
}

void CTextViewRetrieveResult::x_FreeItems()
{
    ITERATE(CCompositeTextItem::TInsertBlockVec, it, m_NewItems) {
        ITERATE(vector<ITextItem*>, it2, it->m_Items) {
            delete *it2;
        }
    }
    m_NewItems.clear();
}

void CTextViewRetrieveResult::ForgetItems() const
{
    m_NewItems.clear();
}

void CTextViewRetrieveResult::AddItem(ITextItem* item, CCompositeTextItem* container)
{
    NON_CONST_ITERATE(CCompositeTextItem::TInsertBlockVec, it, m_NewItems) {
        if (it->m_Container == container) {
            it->m_Items.push_back(item);
            return;
        }
    }

    m_NewItems.push_back(CCompositeTextItem::TInsertBlock(item, -2));
}

void CTextViewRetrieveResult::SetItems(const CCompositeTextItem::TInsertBlockVec& newItems)
{
    x_FreeItems(); // free the existing items
    m_NewItems = newItems;
}

////////////////////////////////////////////////////////////////////////////////

void CTextDataSource::RegisterCommands(CUICommandRegistry& cmd_reg, wxFileArtProvider& provider)
{
    CFlatFileViewContext::RegisterCommands(cmd_reg, provider);

    vector<IEditorFactory*> factories;
    GetExtensionAsInterface(EXT_POINT__EDITOR_FACTORY, factories);

    ITERATE(vector<IEditorFactory*>, it, factories) {
        (*it)->RegisterCommands(cmd_reg, provider);
    }
}

CTextDataSource::CTextDataSource() : m_CmdProccessor(0)
{
}

CTextDataSource::~CTextDataSource()
{
    CAppJobDispatcher& disp = CAppJobDispatcher::GetInstance();
    ITERATE(TJobs, it, m_Jobs)
        disp.DeleteJob(it->second);
}

void CTextDataSource::RemoveJob(CTextItemPanel* panel)
{
    NON_CONST_ITERATE(TJobs, it, m_Jobs) {
        if (it->first == panel) {
            CAppJobDispatcher::GetInstance().DeleteJob(it->second);
            m_Jobs.erase(it);

            TPanelStates::iterator it = m_PanelStates.find(panel);
            if (it != m_PanelStates.end())
                m_PanelStates.erase(it);

            break;
        }
    }
}

void CTextDataSource::x_RemoveJob(CAppJobDispatcher::TJobID jobId)
{
    NON_CONST_ITERATE(TJobs, it, m_Jobs) {
        if (it->second == jobId) {
            m_Jobs.erase(it);
            break;
        }
    }
}

CTextItemPanel* CTextDataSource::x_GetPanel(CAppJobDispatcher::TJobID jobId) const
{
    ITERATE(TJobs, it, m_Jobs) {
        if (jobId == it->second)
            return it->first;
    }

    return 0;
}

void CTextDataSource::LoadFasta(CTextItemPanel* panel, objects::CSeq_entry_Handle& h)
{
    _ASSERT(panel);
    _ASSERT(h);
    RemoveJob(panel);

    CIRef<IAppJob> job(new CFastaRetrieveJob(h));
    CAppJobDispatcher& disp = CAppJobDispatcher::GetInstance();

    CCompositeTextItem* group = new CCompositeTextItem();
    CPlainTextItem* textItem = new CPlainTextItem();
    textItem->SetStyle(CTextStyle(255,0,0,255,255,255,wxTRANSPARENT,false,false));
    textItem->AddLine("Loading...");
    group->AddItem(textItem);
    panel->SetData(group, new CTextPanelContext(3));

    try {
        CAppJobDispatcher::TJobID jobId = disp.StartJob(*job, "ObjManagerEngine", *this, 1, true);
        m_Jobs.push_back(make_pair(panel, jobId));
    } catch(CAppJobException& e)  {
        ERR_POST("CTextDataSource::LoadFasta() - Failed to start job");
        e.ReportAll();
    }
}

void CTextDataSource::LoadFasta(CTextItemPanel* panel, const objects::CSeq_loc& loc, CScope& scope)
{
    _ASSERT(panel);
    RemoveJob(panel);

    CIRef<IAppJob> job(new CFastaRetrieveLocJob(loc, scope));
    CAppJobDispatcher& disp = CAppJobDispatcher::GetInstance();

    CCompositeTextItem* group = new CCompositeTextItem();
    CPlainTextItem* textItem = new CPlainTextItem();
    textItem->SetStyle(CTextStyle(255,0,0,255,255,255,wxTRANSPARENT,false,false));
    textItem->AddLine("Loading...");
    group->AddItem(textItem);
    panel->SetData(group, new CTextPanelContext(3));

    try {
        CAppJobDispatcher::TJobID jobId = disp.StartJob(*job, "ObjManagerEngine", *this, 1, true);
        m_Jobs.push_back(make_pair(panel, jobId));
    } catch(CAppJobException& e)  {
        ERR_POST("CTextDataSource::LoadFasta() - Failed to start job");
        e.ReportAll();
    }
}

void CTextDataSource::LoadFlatFile(CTextItemPanel* panel, objects::CSeq_entry_Handle& h,
                                   const objects::CSeq_loc* seq_loc,
                                   const objects::CFlatFileConfig& config,
                                   CTextPanelState* state)
{
    _ASSERT(panel);
    _ASSERT(h);
    RemoveJob(panel);

    CRef<CFlatFileContext> ctx(new CFlatFileContext(config));
    ctx->SetEntry(h);
    ctx->SetLocation(seq_loc);

    CRef<CFlatItemFormatter> formatter(CFlatItemFormatter::New(config.GetFormat()));
    formatter->SetContext(*ctx);

    CIRef<IAppJob> job(new CFlatFileRetrieveJob(config, *ctx));
    CAppJobDispatcher& disp = CAppJobDispatcher::GetInstance();

    CCompositeTextItem* group = new CCompositeTextItem();
    CPlainTextItem* textItem = new CPlainTextItem();
    textItem->SetStyle(CTextStyle(255,0,0,255,255,255,wxTRANSPARENT,false,false));
    textItem->AddLine("Loading...");
    group->AddItem(textItem);

    CIRef<IBioseqEditor> editor;

    if (m_CmdProccessor) {
        vector<IEditorFactory*> factories;
        GetExtensionAsInterface(EXT_POINT__EDITOR_FACTORY, factories);

        ITERATE(vector<IEditorFactory*>, it, factories) {
            editor.Reset(dynamic_cast<IBioseqEditor*>(
                (*it)->CreateEditor(typeid(IBioseqEditor).name(), *m_CmdProccessor)));

            if (editor)
                break;
        }
    }

    panel->SetData(group, new CFlatFileViewContext(*formatter, h.GetScope(), editor));

    try {
        CAppJobDispatcher::TJobID jobId = disp.StartJob(*job, "ObjManagerEngine", *this, 1, true);
        m_Jobs.push_back(make_pair(panel, jobId));
        if (state)
            m_PanelStates[panel] = CRef<CTextPanelState>(state);
    } catch(CAppJobException& e)  {
        ERR_POST("CTextDataSource::LoadSequence() - Failed to start job");
        e.ReportAll();
    }
}

void CTextDataSource::LoadFlatFile(CTextItemPanel* panel,
                                   objects::CScope& scope,
                                   const objects::CSeq_feat& seq_feat,
                                   bool getGeneModel,
                                   const objects::CFlatFileConfig& config,
                                   CTextPanelState* state)
{
    _ASSERT(panel);
    RemoveJob(panel);

    CRef<CFlatFileContext> ctx(new CFlatFileContext(config));

    CRef<CFlatItemFormatter> formatter(CFlatItemFormatter::New(config.GetFormat()));
    formatter->SetContext(*ctx);

    CIRef<IAppJob> job(new CFlatFeatureRetrieveJob(config, *ctx, scope, seq_feat, getGeneModel));
    CAppJobDispatcher& disp = CAppJobDispatcher::GetInstance();

    CCompositeTextItem* group = new CCompositeTextItem();
    CPlainTextItem* textItem = new CPlainTextItem();
    textItem->SetStyle(CTextStyle(255,0,0,255,255,255,wxTRANSPARENT,false,false));
    textItem->AddLine("Loading...");
    group->AddItem(textItem);
    panel->SetData(group, new CFlatFileViewContext(*formatter, scope, 0));

    try {
        CAppJobDispatcher::TJobID jobId = disp.StartJob(*job, "ObjManagerEngine", *this, 1, true);
        m_Jobs.push_back(make_pair(panel, jobId));
        if (state)
            m_PanelStates[panel] = CRef<CTextPanelState>(state);
    } catch(CAppJobException& e)  {
        ERR_POST("CTextDataSource::LoadSequence() - Failed to start job");
        e.ReportAll();
    }
}

void CTextDataSource::LoadAsn(CTextItemPanel* panel, const CSerialObject& so, objects::CScope& scope)
{
    _ASSERT(panel);
    RemoveJob(panel);

    CIRef<IBioseqEditor> editor;

    if (m_CmdProccessor) {
        vector<IEditorFactory*> factories;
        GetExtensionAsInterface(EXT_POINT__EDITOR_FACTORY, factories);

        ITERATE(vector<IEditorFactory*>, it, factories) {
            editor.Reset(dynamic_cast<IBioseqEditor*>(
                (*it)->CreateEditor(typeid(IBioseqEditor).name(), *m_CmdProccessor)));

            if (editor)
                break;
        }
    }

    CAsnViewContext* asnContext = new CAsnViewContext(scope, editor);

    CIRef<IAppJob> job(new CAsnRetrieveJob(so, *asnContext));
    CAppJobDispatcher& disp = CAppJobDispatcher::GetInstance();

    CCompositeTextItem* group = new CCompositeTextItem();
    CPlainTextItem* textItem = new CPlainTextItem();
    textItem->SetStyle(CTextStyle(255,0,0,255,255,255,wxTRANSPARENT,false,false));
    textItem->AddLine("Loading...");
    group->AddItem(textItem);
    panel->SetData(group, asnContext);

    try {
        CAppJobDispatcher::TJobID jobId = disp.StartJob(*job, "ObjManagerEngine", *this, 1, true);
        m_Jobs.push_back(make_pair(panel, jobId));
    } catch(CAppJobException& e)  {
        ERR_POST("CTextDataSource::LoadSequence() - Failed to start job");
        e.ReportAll();
    }
}

void CTextDataSource::LoadXml(CTextItemPanel* panel, const CSerialObject& so)
{
    _ASSERT(panel);
    RemoveJob(panel);

    CIRef<IAppJob> job(new CXmlRetrieveJob(so));
    CAppJobDispatcher& disp = CAppJobDispatcher::GetInstance();

    CCompositeTextItem* group = new CCompositeTextItem();
    CPlainTextItem* textItem = new CPlainTextItem();
    textItem->SetStyle(CTextStyle(255,0,0,255,255,255,wxTRANSPARENT,false,false));
    textItem->AddLine("Loading...");
    group->AddItem(textItem);
    panel->SetData(group, new CTextPanelContext(1));

    try {
        CAppJobDispatcher::TJobID jobId = disp.StartJob(*job, "ObjManagerEngine", *this, 1, true);
        m_Jobs.push_back(make_pair(panel, jobId));
    } catch(CAppJobException& e)  {
        ERR_POST("CTextDataSource::LoadSequence() - Failed to start job");
        e.ReportAll();
    }
}

void CTextDataSource::OnJobNotification(CEvent* evt)
{
    CAppJobNotification* notn = dynamic_cast<CAppJobNotification*>(evt);
    _ASSERT(notn);
    if (!notn) return;

    CAppJobDispatcher::TJobID job_id = notn->GetJobID();
    CTextItemPanel* panel = x_GetPanel(job_id);

    if (!panel) {
        ERR_POST("CTextDataSource::OnJobNotification() - unknown Job ID " << job_id);
        return;
    }

    switch (notn->GetState())    {
    case IAppJob::eCompleted :
        {{
            CRef<CObject> result = notn->GetResult();
            x_ProcessResult(panel, dynamic_cast<CTextViewRetrieveResult*>(result.GetPointer()));

            CCompositeTextItem* group =
                dynamic_cast<CCompositeTextItem*>(panel->GetItem());
            panel->RemoveItem(group->GetItemCount()-1);
        }}

        x_RemoveJob(job_id);
        break;
    case IAppJob::eFailed:
        {{
            CCompositeTextItem* group =
                dynamic_cast<CCompositeTextItem*>(panel->GetItem());
            panel->RemoveItem(group->GetItemCount()-1);

            CConstIRef<IAppJobError> error = notn->GetError();

            string err_msg = "Failed: ";
            if (error)
                err_msg += error->GetText();
            else
                err_msg += "Unknown fatal error";

            CPlainTextItem* textItem = new CPlainTextItem();
            textItem->SetStyle(CTextStyle(255,0,0,255,255,255,wxTRANSPARENT,false,false));
            textItem->AddLine(err_msg);

            CCompositeTextItem::TInsertBlockVec insertItems;
            insertItems.push_back(CCompositeTextItem::TInsertBlock(textItem, -1));
            panel->InsertItems(insertItems);
        }}

        x_RemoveJob(job_id);
        break;
    case IAppJob::eCanceled:
        x_RemoveJob(job_id);
        break;
    default:
        break;
    }
}

void CTextDataSource::OnProgress(CEvent* evt)
{
    CAppJobNotification* notn = dynamic_cast<CAppJobNotification*>(evt);
    _ASSERT(notn);
    CConstIRef<IAppJobProgress> progress = notn->GetProgress();
    if (!progress)
        return;

    CAppJobDispatcher::TJobID job_id = notn->GetJobID();
    CTextItemPanel* panel = x_GetPanel(job_id);

    if (!panel) {
        ERR_POST("CTextDataSource::OnJobNotification() - unknown Job ID " << job_id);
        return;
    }

    const CTextViewRetrieveProgress* text_progress =
        dynamic_cast<const CTextViewRetrieveProgress*>(progress.GetPointer());
    _ASSERT(text_progress);
    if (text_progress->m_Result) {
        x_ProcessResult(panel, text_progress->m_Result);
    }
}

void CTextDataSource::x_ProcessResult(CTextItemPanel* panel, const CTextViewRetrieveResult* result)
{
    if (!result) return;

    _ASSERT(panel);

    CRef<CTextPanelState> state;
    TPanelStates::const_iterator it = m_PanelStates.find(panel);
    if (it != m_PanelStates.end())
        state = it->second;

    const CCompositeTextItem::TInsertBlockVec& newItems = result->GetItems();
    if (!newItems.empty()) {
        if (state) {
            ITERATE(CCompositeTextItem::TInsertBlockVec, it, newItems) {
                state->RestoreState(*panel, it->m_Items);
            }
        }
        panel->InsertItems(newItems);
        result->ForgetItems();
        if (state)
            state->RestorePos(*panel);
    }
}

END_NCBI_SCOPE

