/*  $Id: text_panel.cpp 27575 2013-03-07 18:41:05Z katargir $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data,  the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties,  express or implied,  including
 *  warranties of performance,  merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Roman Katargin
 */


#include <ncbi_pch.hpp>

#include <wx/sizer.h>
#include <wx/choice.h>
#include <wx/stattext.h>
#include <wx/checkbox.h>
#include <wx/textctrl.h>
#include <wx/button.h>
#include <wx/bitmap.h>
#include <wx/icon.h>
#include <wx/msgdlg.h>

////@begin includes
////@end includes

#include <gui/objutils/objects.hpp>
#include <objmgr/bioseq_ci.hpp>

#include <objtools/format/flat_file_config.hpp>

#include <gui/widgets/text_widget/text_item_panel.hpp>
#include <gui/widgets/text_widget/expand_all_state.hpp>

#include <gui/widgets/seq/text_panel.hpp>

#include <gui/widgets/seq/text_ds.hpp>
#include <gui/widgets/seq/asn_view_context.hpp>

#include <objects/seqloc/Seq_loc.hpp>
#include <objects/seqfeat/Seq_feat.hpp>
#include <objects/submit/Seq_submit.hpp>

#include <gui/objutils/registry.hpp>

#include <gui/widgets/wx/wx_utils.hpp>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

////@begin XPM images
////@end XPM images


/*!
 * CTextPanel type definition
 */

IMPLEMENT_DYNAMIC_CLASS( CTextPanel, wxPanel )


/*!
 * CTextPanel event table definition
 */

BEGIN_EVENT_TABLE( CTextPanel, wxPanel )

////@begin CTextPanel event table entries
    EVT_CHOICE( ID_CHOICE1, CTextPanel::OnViewTypeSelected )

    EVT_CHOICE( ID_CHOICE_MODE, CTextPanel::OnFlatFileModeSelected )

    EVT_CHOICE( ID_CHOICE_STYLE, CTextPanel::OnFlatFileStyleSelected )

    EVT_CHOICE( ID_CHOICE_FONT, CTextPanel::OnFontFamilySelected )

    EVT_CHOICE( ID_CHOICE_SIZE, CTextPanel::OnFontSizeSelected )

    EVT_CHOICE( ID_CHOICE_FEAT_MODE, CTextPanel::OnChoiceFeatModeSelected )

    EVT_CHECKBOX( ID_SHOW_ASN_TYPES, CTextPanel::OnShowAsnTypesClick )
    EVT_UPDATE_UI( ID_SHOW_ASN_TYPES, CTextPanel::OnShowAsnTypesUpdate )

    EVT_CHECKBOX( ID_SHOW_PATH_TO_ROOT, CTextPanel::OnShowPathToRootClick )
    EVT_UPDATE_UI( ID_SHOW_PATH_TO_ROOT, CTextPanel::OnShowPathToRootUpdate )

    EVT_TEXT_ENTER( ID_TEXTCTRL, CTextPanel::OnTextctrlEnter )

    EVT_BUTTON( ID_BUTTON, CTextPanel::OnFindNextClick )

////@end CTextPanel event table entries

END_EVENT_TABLE()

const wxChar* pszFlatFile = wxT("Flat File");
const wxChar* pszFastA  = wxT("FastA");
const wxChar* pszASN  = wxT("ASN");
const wxChar* pszXML  = wxT("XML");

/*!
 * CTextPanel constructors
 */

CTextPanel::CTextPanel()
{
    Init();
}

CTextPanel::CTextPanel( wxWindow* parent, wxWindowID id, const wxPoint& pos, const wxSize& size, long style)
{
    Init();
    Create(parent, id, pos, size, style);
    CreateControls();
}

/*!
 * CTextPanel destructor
 */

CTextPanel::~CTextPanel()
{
////@begin CTextPanel destruction
////@end CTextPanel destruction
}

/*!
 * Member initialisation
 */

static const char
*kTextViewPath = "GBENCH.TextView",
*kViewTypeSeq = "ViewTypeSeq",
*kViewTypeFeat = "ViewTypeFeat",
*kViewTypeGeneric = "ViewTypeGeneric",
*kFontFamily = "FontFamily",
*kFontSize = "FontSize";

void CTextPanel::Init()
{
////@begin CTextPanel member initialisation
    m_ViewType = kUninitialized;
    m_FontFamily = 0;
    m_FontSize = 1;
    m_FlatFileMode = 0;
    m_FlatFileStyle = 0;
    m_FeatMode = 0;
    m_ExpandAll = false;
    m_FindTextCtrl = NULL;
    m_SearchUpCtrl = NULL;
    m_TextWindow = NULL;
////@end CTextPanel member initialisation

    CGuiRegistry& gui_reg = CGuiRegistry::GetInstance();
    CGuiRegistry::TReadView view = gui_reg.GetReadView(kTextViewPath);

    m_FontFamily = view.GetInt(kFontFamily, m_FontFamily);
    m_FontSize = view.GetInt(kFontSize, m_FontSize);
}

void CTextPanel::SetUndoManager(ICommandProccessor* cmdProccessor)
{
    if (!m_DataSource)
        m_DataSource.Reset(new CTextDataSource());

    m_DataSource->SetUndoManager(cmdProccessor);
}

/*!
 * Control creation for CTextPanel
 */

void CTextPanel::CreateControls()
{
    SetExtraStyle(wxWS_EX_VALIDATE_RECURSIVELY);

    #ifdef __WXMAC__    
		SetBackgroundStyle(wxBG_STYLE_COLOUR);
		SetBackgroundColour(*wxWHITE);
	#else
		SetBackgroundStyle(wxBG_STYLE_SYSTEM);
	#endif

    Hide();

    m_DataSource.Reset(new CTextDataSource());

////@begin CTextPanel content construction
    CTextPanel* itemPanel1 = this;

    wxBoxSizer* itemBoxSizer2 = new wxBoxSizer(wxVERTICAL);
    itemPanel1->SetSizer(itemBoxSizer2);

    wxBoxSizer* itemBoxSizer3 = new wxBoxSizer(wxHORIZONTAL);
    itemBoxSizer2->Add(itemBoxSizer3, 0, wxGROW|wxALL, 0);

    wxStaticText* itemStaticText4 = new wxStaticText( itemPanel1, wxID_STATIC, _("Format"), wxDefaultPosition, wxDefaultSize, 0 );
    itemBoxSizer3->Add(itemStaticText4, 0, wxALIGN_CENTER_VERTICAL|wxALL, 5);

    wxArrayString itemChoice5Strings;
    itemChoice5Strings.Add(_("Flat File"));
    itemChoice5Strings.Add(_("ASN"));
    itemChoice5Strings.Add(_("FastA"));
    wxChoice* itemChoice5 = new wxChoice( itemPanel1, ID_CHOICE1, wxDefaultPosition, wxDefaultSize, itemChoice5Strings, 0 );
    itemChoice5->SetStringSelection(_("Flat File"));
    itemBoxSizer3->Add(itemChoice5, 0, wxALIGN_CENTER_VERTICAL|wxALL, 5);

    itemBoxSizer3->Add(7, 8, 0, wxALIGN_CENTER_VERTICAL|wxALL, 5);

    wxStaticText* itemStaticText7 = new wxStaticText( itemPanel1, ID_STATIC_MODE, _("Mode"), wxDefaultPosition, wxDefaultSize, 0 );
    itemBoxSizer3->Add(itemStaticText7, 0, wxALIGN_CENTER_VERTICAL|wxALL, 5);

    wxArrayString itemChoice8Strings;
    itemChoice8Strings.Add(_("GenBank"));
    itemChoice8Strings.Add(_("Release"));
    itemChoice8Strings.Add(_("Sequin"));
    itemChoice8Strings.Add(_("Dump"));
    wxChoice* itemChoice8 = new wxChoice( itemPanel1, ID_CHOICE_MODE, wxDefaultPosition, wxDefaultSize, itemChoice8Strings, 0 );
    itemBoxSizer3->Add(itemChoice8, 0, wxALIGN_CENTER_VERTICAL|wxALL, 5);

    wxStaticText* itemStaticText9 = new wxStaticText( itemPanel1, ID_STATIC_STYLE, _("Style"), wxDefaultPosition, wxDefaultSize, 0 );
    itemBoxSizer3->Add(itemStaticText9, 0, wxALIGN_CENTER_VERTICAL|wxALL, 5);

    wxArrayString itemChoice10Strings;
    itemChoice10Strings.Add(_("Normal"));
    itemChoice10Strings.Add(_("Segment"));
    itemChoice10Strings.Add(_("Master"));
    itemChoice10Strings.Add(_("Contig"));
    wxChoice* itemChoice10 = new wxChoice( itemPanel1, ID_CHOICE_STYLE, wxDefaultPosition, wxDefaultSize, itemChoice10Strings, 0 );
    itemBoxSizer3->Add(itemChoice10, 0, wxALIGN_CENTER_VERTICAL|wxALL, 5);

    wxStaticText* itemStaticText11 = new wxStaticText( itemPanel1, ID_STATIC_FONT, _("Font"), wxDefaultPosition, wxDefaultSize, 0 );
    itemBoxSizer3->Add(itemStaticText11, 0, wxALIGN_CENTER_VERTICAL|wxALL, 5);

    wxArrayString itemChoice12Strings;
    itemChoice12Strings.Add(_("fixed"));
    itemChoice12Strings.Add(_("sans-serif"));
    itemChoice12Strings.Add(_("serif"));
    wxChoice* itemChoice12 = new wxChoice( itemPanel1, ID_CHOICE_FONT, wxDefaultPosition, wxDefaultSize, itemChoice12Strings, 0 );
    itemBoxSizer3->Add(itemChoice12, 0, wxALIGN_CENTER_VERTICAL|wxALL, 5);

    wxStaticText* itemStaticText13 = new wxStaticText( itemPanel1, ID_STATIC_SIZE, _("Size"), wxDefaultPosition, wxDefaultSize, 0 );
    itemBoxSizer3->Add(itemStaticText13, 0, wxALIGN_CENTER_VERTICAL|wxALL, 5);

    wxArrayString itemChoice14Strings;
    itemChoice14Strings.Add(_("8"));
    itemChoice14Strings.Add(_("10"));
    itemChoice14Strings.Add(_("12"));
    itemChoice14Strings.Add(_("14"));
    wxChoice* itemChoice14 = new wxChoice( itemPanel1, ID_CHOICE_SIZE, wxDefaultPosition, wxDefaultSize, itemChoice14Strings, 0 );
    itemBoxSizer3->Add(itemChoice14, 0, wxALIGN_CENTER_VERTICAL|wxALL, 5);

    wxStaticText* itemStaticText15 = new wxStaticText( itemPanel1, ID_STATIC_FEAT_MODE, _("Feature Mode"), wxDefaultPosition, wxDefaultSize, 0 );
    itemBoxSizer3->Add(itemStaticText15, 0, wxALIGN_CENTER_VERTICAL|wxALL, 5);

    wxArrayString itemChoice16Strings;
    itemChoice16Strings.Add(_("Show Gene Model"));
    itemChoice16Strings.Add(_("All features in range"));
    wxChoice* itemChoice16 = new wxChoice( itemPanel1, ID_CHOICE_FEAT_MODE, wxDefaultPosition, wxDefaultSize, itemChoice16Strings, 0 );
    itemBoxSizer3->Add(itemChoice16, 0, wxALIGN_CENTER_VERTICAL|wxALL, 5);

    wxCheckBox* itemCheckBox17 = new wxCheckBox( itemPanel1, ID_SHOW_ASN_TYPES, _("Show ASN.1 types"), wxDefaultPosition, wxDefaultSize, 0 );
    itemCheckBox17->SetValue(false);
    itemBoxSizer3->Add(itemCheckBox17, 0, wxALIGN_CENTER_VERTICAL|wxALL, 5);

    wxCheckBox* itemCheckBox18 = new wxCheckBox( itemPanel1, ID_SHOW_PATH_TO_ROOT, _("Show Path to Root"), wxDefaultPosition, wxDefaultSize, 0 );
    itemCheckBox18->SetValue(false);
    itemBoxSizer3->Add(itemCheckBox18, 0, wxALIGN_CENTER_VERTICAL|wxALL, 5);

    wxStaticText* itemStaticText19 = new wxStaticText( itemPanel1, wxID_STATIC, _("Find what:"), wxDefaultPosition, wxDefaultSize, 0 );
    itemBoxSizer3->Add(itemStaticText19, 0, wxALIGN_CENTER_VERTICAL|wxALL, 5);

    m_FindTextCtrl = new wxTextCtrl( itemPanel1, ID_TEXTCTRL, wxEmptyString, wxDefaultPosition, wxDefaultSize, wxTE_PROCESS_ENTER );
    itemBoxSizer3->Add(m_FindTextCtrl, 0, wxALIGN_CENTER_VERTICAL|wxALL, 5);

    wxButton* itemButton21 = new wxButton( itemPanel1, ID_BUTTON, _("Find Next"), wxDefaultPosition, wxDefaultSize, 0 );
    itemBoxSizer3->Add(itemButton21, 0, wxALIGN_CENTER_VERTICAL|wxALL, 5);

    m_SearchUpCtrl = new wxCheckBox( itemPanel1, ID_CHECKBOX, _("Search up"), wxDefaultPosition, wxDefaultSize, 0 );
    m_SearchUpCtrl->SetValue(false);
    itemBoxSizer3->Add(m_SearchUpCtrl, 0, wxALIGN_CENTER_VERTICAL|wxALL, 5);

    m_TextWindow = new CTextItemPanel( itemPanel1, ID_WIDGET, wxDefaultPosition, wxDefaultSize, wxNO_BORDER|wxHSCROLL|wxVSCROLL );
    itemBoxSizer2->Add(m_TextWindow, 1, wxGROW, 0);

    // Set validators
    itemChoice8->SetValidator( wxGenericValidator(& m_FlatFileMode) );
    itemChoice10->SetValidator( wxGenericValidator(& m_FlatFileStyle) );
    itemChoice12->SetValidator( wxGenericValidator(& m_FontFamily) );
    itemChoice14->SetValidator( wxGenericValidator(& m_FontSize) );
    itemChoice16->SetValidator( wxGenericValidator(& m_FeatMode) );
////@end CTextPanel content construction

    m_TextWindow->SetBackgroundColour(wxColor(wxT("white")));
/*
    wxChoice* textType = (wxChoice*)FindWindow(ID_CHOICE1);
    x_UpdateToolbar();
    x_UpdateViewType(textType->GetStringSelection());
    x_UpdatePanelFont();
 */
}

void CTextPanel::Reload()
{
    CRef<CTextPanelState> state;
    CTextPanelContext* context = m_TextWindow->GetContext();

    
    if (m_ExpandAll) {
        state.Reset(new CExpandAllState());
    }
    else if (context) {
        state.Reset(context->GetViewState());
    }

    x_Load(state);
}

void CTextPanel::x_Load(CTextPanelState* state)
{
    m_TextWindow->SetData(NULL, NULL);
    x_UpdatePanelFont();

    switch(m_ViewType) {
    case kFlatFile:
        x_LoadFlatFile(state);
        break;
    case kFastA:
        x_LoadFastA();
        break;
    case kXML:
        x_LoadXML();
        break;
    default:
        x_LoadASN();
        break;
    }
}

static const CSeq_entry* GetSeqEntryFromSeqSubmit(const CSerialObject* so) {
    const CSeq_entry* seq_entry = 0;
    const CSeq_submit* seq_submit = dynamic_cast<const CSeq_submit*>(so);
    if (seq_submit) {
        if (seq_submit->IsSetData()) {
            const CSeq_submit::TData& data = seq_submit->GetData();
            if (data.IsEntrys()) {
                const CSeq_submit::TData::TEntrys& entries = data.GetEntrys();
                if (!entries.empty()) {
                    seq_entry = *entries.begin();
                }
            }
        }
    }
    return seq_entry;
}

static CSeq_entry_Handle GetSeqEntryHandle(CScope& scope, const CSerialObject& so)
{
    const CSeq_id* seq_id = dynamic_cast<const CSeq_id*>(&so);
    const CSeq_loc* seq_loc = dynamic_cast<const CSeq_loc*>(&so);
    const CSeq_entry* seq_entry = dynamic_cast<const CSeq_entry*>(&so);
    const CBioseq* bioseq = dynamic_cast<const CBioseq*>(&so);

    if (!seq_entry) {
        seq_entry = GetSeqEntryFromSeqSubmit(&so);
    }

    CSeq_entry_Handle seh;

    if (seq_entry)
        seh = scope.GetSeq_entryHandle(*seq_entry);
    else if (bioseq) {
        CBioseq_Handle handle = scope.GetBioseqHandle(*bioseq);
        if (handle)
            seh = handle.GetTopLevelEntry();
    }
    else if (seq_id) {
        CBioseq_Handle handle = scope.GetBioseqHandle(*seq_id);
        if (handle)
            seh = handle.GetTopLevelEntry();
    }
    else if (seq_loc) {
        CBioseq_Handle handle = scope.GetBioseqHandle(*seq_loc);
        if (handle)
            seh = handle.GetTopLevelEntry();
    }

    return seh;
}

void CTextPanel::x_LoadFlatFile(CTextPanelState* state)
{
    if (!m_SO) return;

    const CSeq_feat* seq_feat = dynamic_cast<const CSeq_feat*>(m_SO.GetPointer());
    const CSeq_loc* seq_loc = dynamic_cast<const CSeq_loc*>(m_SO.GetPointer());

    CSeq_entry_Handle seh;
    if (!seq_feat) seh = GetSeqEntryHandle(*m_Scope, *m_SO);
    if (!seq_feat && !seh)
        return;

    CFlatFileConfig::EMode mode;
    CFlatFileConfig::EStyle style;

    switch(m_FlatFileMode) {
        default: mode = CFlatFileConfig::eMode_Entrez; break;
        case 1:  mode = CFlatFileConfig::eMode_Release; break;
        case 2:  mode = CFlatFileConfig::eMode_GBench; break;
        case 3:  mode = CFlatFileConfig::eMode_Dump; break;
    }
    switch(m_FlatFileStyle) {
        default: style = CFlatFileConfig::eStyle_Normal; break;
        case 1:  style = CFlatFileConfig::eStyle_Segment; break;
        case 2:  style = CFlatFileConfig::eStyle_Master; break;
        case 3:  style = CFlatFileConfig::eStyle_Contig; break;
    }

    wxBusyCursor wait;

    CFlatFileConfig cfg(CFlatFileConfig::eFormat_GenBank, mode, style, 0, CFlatFileConfig::fViewAll);

    m_DataSource->RemoveJob(m_TextWindow);
    
    if (seq_feat)
        m_DataSource->LoadFlatFile(m_TextWindow, *m_Scope, *seq_feat, m_FeatMode == 0, cfg, state);
    else
        m_DataSource->LoadFlatFile(m_TextWindow, seh, seq_loc, cfg, state);
}


void CTextPanel::x_LoadASN()
{
    if (!m_SO) return;

    CConstRef<CSerialObject> so = m_SO;
    if (!GetSeqEntryFromSeqSubmit(so)) {
        CSeq_entry_Handle seh = GetSeqEntryHandle(*m_Scope, *m_SO);
        if (seh) {
            CConstRef<CSeq_entry> seq_entry = seh.GetSeq_entryCore();
            if (seq_entry)
                so = seq_entry;
        }
    }

    wxBusyCursor wait;
    m_DataSource->RemoveJob(m_TextWindow);
    m_DataSource->LoadAsn(m_TextWindow, *so, *m_Scope);
}

void CTextPanel::x_LoadXML()
{
    if (!m_SO) return;

    CConstRef<CSerialObject> so = m_SO;
    if (!GetSeqEntryFromSeqSubmit(so)) {
        CSeq_entry_Handle seh = GetSeqEntryHandle(*m_Scope, *m_SO);
        if (seh) {
            CConstRef<CSeq_entry> seq_entry = seh.GetSeq_entryCore();
            if (seq_entry)
                so = seq_entry;
        }
    }

    wxBusyCursor wait;
    m_DataSource->RemoveJob(m_TextWindow);
    m_DataSource->LoadXml(m_TextWindow, *so);
}

void CTextPanel::x_LoadFastA()
{
    if (!m_SO) return;

    wxBusyCursor wait;
    m_DataSource->RemoveJob(m_TextWindow);

    CSeq_entry_Handle seh;
    const CSeq_loc* seq_loc = dynamic_cast<const CSeq_loc*>(m_SO.GetPointer());
    if (seq_loc) {
        m_DataSource->LoadFasta(m_TextWindow, *seq_loc, *m_Scope);
    }
    else {
        CSeq_entry_Handle seh = GetSeqEntryHandle(*m_Scope, *m_SO);
        if (!seh) return;
        m_DataSource->LoadFasta(m_TextWindow, seh);
    }
}


bool CTextPanel::ShowToolTips()
{
    return true;
}


wxBitmap CTextPanel::GetBitmapResource( const wxString& name )
{
    // Bitmap retrieval
////@begin CTextPanel bitmap retrieval
    wxUnusedVar(name);
    return wxNullBitmap;
////@end CTextPanel bitmap retrieval
}


wxIcon CTextPanel::GetIconResource( const wxString& name )
{
    // Icon retrieval
////@begin CTextPanel icon retrieval
    wxUnusedVar(name);
    return wxNullIcon;
////@end CTextPanel icon retrieval
}

void CTextPanel::SetInitialViewType(const string& viewType)
{
    wxString type = ToWxString(viewType);

    if (type == pszFlatFile) {
        m_ViewType = kFlatFile;
    }
    else if (type == pszASN) {
        m_ViewType = kASN;
    }
    else if (type == pszFastA) {
        m_ViewType = kFastA;
    }
    else if (type == pszXML) {
        m_ViewType = kXML;
    }
}

void CTextPanel::Init(SConstScopedObject& object)
{
    wxBusyCursor wait;

    m_Scope.Reset(object.scope.GetPointer());
    m_SO.Reset(dynamic_cast<const CSerialObject*>(object.object.GetPointer()));

    const CSeq_id* seq_id = dynamic_cast<const CSeq_id*>(m_SO.GetPointer());
    const CSeq_loc* seq_loc = dynamic_cast<const CSeq_loc*>(m_SO.GetPointer());
    const CSeq_feat* seq_feat = dynamic_cast<const CSeq_feat*>(m_SO.GetPointer());
    const CSeq_entry* seq_entry = dynamic_cast<const CSeq_entry*>(m_SO.GetPointer());
    const CBioseq* bioseq = dynamic_cast<const CBioseq*>(m_SO.GetPointer());

    if (!seq_entry) {
        seq_entry = GetSeqEntryFromSeqSubmit(m_SO);
    }

    if (seq_feat) {
        CSeqFeatData::E_Choice type = seq_feat->GetData().Which();
        if (type != CSeqFeatData::e_Gene &&
            type != CSeqFeatData::e_Cdregion &&
            type != CSeqFeatData::e_Rna)
            seq_feat = NULL;
    }

    if (seq_id) {
        CBioseq_Handle handle = object.scope->GetBioseqHandle(*seq_id);
        if (!handle) {
            string str("failed to retrieve sequence for id ");
            seq_id->GetLabel(&str);
            NCBI_THROW(CException, eUnknown, str);
        }
    }

    wxChoice* textType = (wxChoice*)FindWindow(ID_CHOICE1);
    textType->Clear();

    int viewType = m_ViewType;
    wxString viewTypeString;

    CGuiRegistry& gui_reg = CGuiRegistry::GetInstance();
    CGuiRegistry::TReadView view = gui_reg.GetReadView(kTextViewPath);

    if (seq_id || seq_loc || seq_feat || seq_entry || bioseq) {
        textType->Append(pszFlatFile);
        if (!seq_feat) textType->Append(pszFastA);
        textType->Append(pszASN);
        textType->Append(pszXML);

        if (viewType == kUninitialized) {
            if (seq_feat != NULL)
                viewType = view.GetInt(kViewTypeFeat, m_ViewType);
            else
                viewType = view.GetInt(kViewTypeSeq, m_ViewType);
        }

        switch(viewType) {
        case kFastA:
            viewTypeString = (seq_feat != NULL) ? pszFlatFile : pszFastA;
            break;
        case kASN:
            viewTypeString = pszASN;
            break;
        case kXML:
            viewTypeString = pszXML;
            break;
        case kFlatFile:
        case kUninitialized:
        default:
            viewTypeString = pszFlatFile;
            break;
        }
    }
    else {
        textType->Append(pszASN);
        textType->Append(pszXML);

        if (viewType == kUninitialized) {
            viewType = view.GetInt(kViewTypeGeneric, viewType);
        }
        viewTypeString = (viewType == kXML) ? pszXML : pszASN;
    }

    textType->SetStringSelection(viewTypeString);
    m_ViewType = kUninitialized;
    x_UpdateViewType(viewTypeString);
}

void CTextPanel::x_SaveViewType()
{
    const CSeq_id* seq_id = dynamic_cast<const CSeq_id*>(m_SO.GetPointer());
    const CSeq_loc* seq_loc = dynamic_cast<const CSeq_loc*>(m_SO.GetPointer());
    const CSeq_feat* seq_feat = dynamic_cast<const CSeq_feat*>(m_SO.GetPointer());
    const CSeq_entry* seq_entry = dynamic_cast<const CSeq_entry*>(m_SO.GetPointer());
    const CBioseq* bioseq = dynamic_cast<const CBioseq*>(m_SO.GetPointer());

    if (!seq_entry) {
        seq_entry = GetSeqEntryFromSeqSubmit(m_SO);
    }

    if (seq_feat) {
        CSeqFeatData::E_Choice type = seq_feat->GetData().Which();
        if (type != CSeqFeatData::e_Gene &&
            type != CSeqFeatData::e_Cdregion &&
            type != CSeqFeatData::e_Rna)
            seq_feat = NULL;
    }

    const char* strType = kViewTypeGeneric;

    if (seq_feat)
        strType = kViewTypeFeat;
    else if (seq_id || seq_loc || seq_entry || bioseq)
        strType = kViewTypeSeq;

    CGuiRegistry& gui_reg = CGuiRegistry::GetInstance();
    CGuiRegistry::TReadWriteView view = gui_reg.GetReadWriteView(kTextViewPath);
    view.Set(strType, (int)m_ViewType);
}

void CTextPanel::OnViewTypeSelected( wxCommandEvent& event )
{
    x_UpdateViewType(event.GetString());
}

void CTextPanel::x_UpdateViewType(const wxString& type)
{
    TViewType viewType = kASN;

    if (type == pszFlatFile)
        viewType = kFlatFile;
    else if (type == pszFastA)
        viewType = kFastA;
    else if (type == pszXML)
        viewType = kXML;

    if (viewType != m_ViewType) {
        m_ViewType = viewType;
        x_SaveViewType();
        
        x_UpdateToolbar();
        GetSizer()->Layout();

        CRef<CTextPanelState> state;
        if (m_ExpandAll) {
            state.Reset(new CExpandAllState());
        }

        x_Load(state);
    }
}

void CTextPanel::x_UpdateToolbar()
{
    bool flatFile = (m_ViewType == kFlatFile);
    const CSeq_feat* seq_feat = dynamic_cast<const CSeq_feat*>(m_SO.GetPointer());

    FindWindow(ID_STATIC_MODE)->Show(flatFile && seq_feat == 0);
    FindWindow(ID_CHOICE_MODE)->Show(flatFile && seq_feat == 0);
    FindWindow(ID_STATIC_STYLE)->Show(flatFile && seq_feat == 0);
    FindWindow(ID_CHOICE_STYLE)->Show(flatFile && seq_feat == 0);

    FindWindow(ID_STATIC_FEAT_MODE)->Show(flatFile && seq_feat != 0);
    FindWindow(ID_CHOICE_FEAT_MODE)->Show(flatFile && seq_feat != 0);

    FindWindow(ID_STATIC_FONT)->Show(!flatFile);
    FindWindow(ID_CHOICE_FONT)->Show(!flatFile);
    FindWindow(ID_STATIC_SIZE)->Show(!flatFile);
    FindWindow(ID_CHOICE_SIZE)->Show(!flatFile);

    FindWindow(ID_SHOW_ASN_TYPES)->Show(m_ViewType == kASN);
    FindWindow(ID_SHOW_PATH_TO_ROOT)->Show(m_ViewType == kASN);
}


void CTextPanel::x_UpdatePanelFont()
{
    int pointSize = 10;
    wxFontFamily fontFamily = wxFONTFAMILY_MODERN;
    if (m_ViewType != kFlatFile) {
        switch (m_FontFamily) {
        case 1 : fontFamily = wxFONTFAMILY_SWISS; break;
        case 2 : fontFamily = wxFONTFAMILY_ROMAN; break;
        }

        switch(m_FontSize) {
        case 0 : pointSize = 8; break;
        case 2 : pointSize = 12; break;
        case 3 : pointSize = 14; break;
        }
    }
    m_TextWindow->SetFont(wxFont(pointSize, fontFamily, wxFONTSTYLE_NORMAL, wxFONTWEIGHT_NORMAL));
}

void CTextPanel::OnFontFamilySelected( wxCommandEvent& event )
{
    if (event.GetInt() != m_FontFamily) {
        m_FontFamily = event.GetInt();

        CGuiRegistry& gui_reg = CGuiRegistry::GetInstance();
        CGuiRegistry::TReadWriteView view = gui_reg.GetReadWriteView(kTextViewPath);
        view.Set(kFontFamily, m_FontFamily);

        if (m_ViewType != kFlatFile)
            x_UpdatePanelFont();
    }
}


void CTextPanel::OnFontSizeSelected( wxCommandEvent& event )
{
    if (event.GetInt() != m_FontSize) {
        m_FontSize = event.GetInt();

        CGuiRegistry& gui_reg = CGuiRegistry::GetInstance();
        CGuiRegistry::TReadWriteView view = gui_reg.GetReadWriteView(kTextViewPath);
        view.Set(kFontSize, m_FontSize);

        if (m_ViewType != kFlatFile)
            x_UpdatePanelFont();
    }
}


void CTextPanel::OnFlatFileModeSelected( wxCommandEvent& event )
{
    if (event.GetInt() != m_FlatFileMode)
        m_FlatFileMode = event.GetInt();

    if (m_ViewType == kFlatFile)
        x_LoadFlatFile();
}


void CTextPanel::OnFlatFileStyleSelected( wxCommandEvent& event )
{
    if (event.GetInt() != m_FlatFileStyle)
        m_FlatFileStyle = event.GetInt();

    if (m_ViewType == kFlatFile)
        x_LoadFlatFile();
}

void CTextPanel::OnChoiceFeatModeSelected( wxCommandEvent& event )
{
    if (event.GetInt() != m_FeatMode)
        m_FeatMode = event.GetInt();

    if (m_ViewType == kFlatFile)
        x_LoadFlatFile();
}

void CTextPanel::OnShowAsnTypesClick( wxCommandEvent& event )
{
    if (m_ViewType != kASN)
        return;

    CTextItemPanel* widget = GetWidget();
    if (widget == NULL)
        return;

    CAsnViewContext* context = dynamic_cast<CAsnViewContext*>(widget->GetContext());
    if (context == NULL)
        return;

    context->SetShowAsnTypes(event.IsChecked());
}

void CTextPanel::OnShowAsnTypesUpdate( wxUpdateUIEvent& event )
{
    if (m_ViewType != kASN)
        return;

    CTextItemPanel* widget = GetWidget();
    if (widget == NULL)
        return;

    CAsnViewContext* context = dynamic_cast<CAsnViewContext*>(widget->GetContext());
    if (context == NULL)
        return;

    event.Check(context->GetShowAsnTypes());
}

void CTextPanel::OnShowPathToRootClick( wxCommandEvent& event )
{
    if (m_ViewType != kASN)
        return;

    CTextItemPanel* widget = GetWidget();
    if (widget == NULL)
        return;

    CAsnViewContext* context = dynamic_cast<CAsnViewContext*>(widget->GetContext());
    if (context == NULL)
        return;

    context->SetShowPathToRoot(event.IsChecked());
}

void CTextPanel::OnShowPathToRootUpdate( wxUpdateUIEvent& event )
{
    if (m_ViewType != kASN)
        return;

    CTextItemPanel* widget = GetWidget();
    if (widget == NULL)
        return;

    CAsnViewContext* context = dynamic_cast<CAsnViewContext*>(widget->GetContext());
    if (context == NULL)
        return;

    event.Check(context->GetShowPathToRoot());
}

void CTextPanel::x_FindText()
{
    CTextItemPanel* widget = GetWidget();
    if (widget == NULL)
        return;

    const wxString caption = wxT("Find Text");

    wxString text = m_FindTextCtrl->GetValue();
    if (text.empty()) {
        wxMessageBox(wxT("Search string is empty."), caption, wxOK | wxICON_EXCLAMATION);
        return;
    }

    bool searchUp = m_SearchUpCtrl->IsChecked();

    if (!widget->FindText(text, !searchUp)) {
        wxMessageBox(wxT("Search string not found."), caption, wxOK | wxICON_EXCLAMATION);
    }
}


void CTextPanel::OnFindNextClick( wxCommandEvent& event )
{
    x_FindText();
}

void CTextPanel::OnTextctrlEnter( wxCommandEvent& event )
{
    x_FindText();
}

END_NCBI_SCOPE
