/*  $Id: align_statistics_glyph.cpp 28152 2013-05-16 19:59:26Z wuliangs $
* ===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data, the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties, express or implied, including
*  warranties of performance, merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================
*
* Authors:  Liangshou Wu
*
*
*/

#include <ncbi_pch.hpp>
#include <gui/widgets/seq_graphic/align_statistics_glyph.hpp>
#include <gui/widgets/seq_graphic/rendering_ctx.hpp>
#include <gui/opengl/glrender.hpp>
#include <gui/opengl/glutils.hpp>
#include "wx_aln_stat_dlg.hpp"
//#include <gui/widgets/wx/message_box.hpp>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

static int kSmearRowHeight = 10;

///////////////////////////////////////////////////////////////////////////////
///   CAlnStatGlyph

CAlnStatGlyph::CAlnStatGlyph(TSeqPos start, TModelUnit scale)
    : m_DlgHost(NULL)
    , m_StartPos(start)
    , m_ZoomScale(scale)
{
}


bool CAlnStatGlyph::OnLeftDblClick(const TModelPoint& /*p*/)
{
    CwxAlnStatOptionsDlg dlg(NULL);
    dlg.SetContent(m_Config->ShowAGTC());
    dlg.SetDisplay(m_Config->IsBarGraph());
    dlg.SetValueType(m_Config->ShowCount());
    dlg.SetGraphHeight(m_Config->m_GraphHeight);
    dlg.SetZoomLevel(m_Config->m_StatZoomLevel);
    if (m_DlgHost) {
        m_DlgHost->PreDialogShow();
    }
    if(dlg.ShowModal() == wxID_OK) {
        m_Config->SetDisplayFlag(CAlnStatConfig::fBarGraph, dlg.IsBarGraph());
        m_Config->SetDisplayFlag(CAlnStatConfig::fShowMismatch, !dlg.IsAGTC());
        m_Config->SetDisplayFlag(CAlnStatConfig::fShowCount, dlg.IsCount());
        m_Config->m_GraphHeight = dlg.GetGraphHeight();
        m_Config->m_StatZoomLevel = dlg.GetZoomLevel();
        if (m_Config->m_GraphHeight < 10) {
            // set a minimal height
            m_Config->m_GraphHeight = 10;
        }
        x_OnLayoutChanged();
    }
    if (m_DlgHost) {
        m_DlgHost->PostDialogShow();
    }
    return true;
}


bool CAlnStatGlyph::NeedTooltip(const TModelPoint& p, string& tt, string& t_title) const
{
    GetTooltip(p, tt, t_title);
    return true;
}


void CAlnStatGlyph::GetTooltip(const TModelPoint& p, string& tt, string& /*t_title*/) const
{
    if (p.X() >= 0) {
        TSeqPos seq_pos = (TSeqPos)p.X();
        tt += "Alignment statistics\n";
        tt += "Base position: " + NStr::UIntToString(seq_pos + 1, NStr::fWithCommas);
        int idx = seq_pos - m_StartPos;
        if (m_Context->GetScale() > 1.0) {
            idx = (int)(idx / m_Context->GetScale());
        }

        char buf[255];
        if (idx >= 0  &&  idx < (int)m_StatVec.size()) {
            const SStatStruct& stat = m_StatVec[idx];
            int total = stat.m_Data[CAlnStatConfig::eStat_Total];
            tt += "\nTotal count\t: " + NStr::Int8ToString(total);
            int num = stat.m_Data[CAlnStatConfig::eStat_Match];
            tt += "\nMatches\t: " + NStr::Int8ToString(num);
            if (total > 0) {
                sprintf(buf, "%2.1f", 100.0 * num / total);
                tt += " (";
                tt += buf;
                tt += "%)";
            }
            num = stat.m_Data[CAlnStatConfig::eStat_Mismatch];
            tt += "\nMismatches\t: " + NStr::Int8ToString(num);
            if (total > 0) {
                sprintf(buf, "%2.1f", 100.0 * num / total);
                tt += " (";
                tt += buf;
                tt += "%)";
            }
            num = stat.m_Data[CAlnStatConfig::eStat_Gap];
            tt += "\nGaps\t\t: " + NStr::Int8ToString(num);
            if (total > 0) {
                sprintf(buf, "%2.1f", 100.0 * num / total);
                tt += " (";
                tt += buf;
                tt += "%)";
            }
            num = stat.m_Data[CAlnStatConfig::eStat_A];
            tt += "\nA\t\t: " + NStr::Int8ToString(num);
            if (total > 0) {
                sprintf(buf, "%2.1f", 100.0 * num / total);
                tt += " (";
                tt += buf;
                tt += "%)";
            }
            num = stat.m_Data[CAlnStatConfig::eStat_T];
            tt += "\nT\t\t: " + NStr::Int8ToString(num);
            if (total > 0) {
                sprintf(buf, "%2.1f", 100.0 * num / total);
                tt += " (";
                tt += buf;
                tt += "%)";
            }
            num = stat.m_Data[CAlnStatConfig::eStat_G];
            tt += "\nG\t\t: " + NStr::Int8ToString(num);
            if (total > 0) {
                sprintf(buf, "%2.1f", 100.0 * num / total);
                tt += " (";
                tt += buf;
                tt += "%)";
            }
            num = stat.m_Data[CAlnStatConfig::eStat_C];
            tt += "\nC\t\t: " + NStr::Int8ToString(num);
            if (total > 0) {
                sprintf(buf, "%2.1f", 100.0 * num / total);
                tt += " (";
                tt += buf;
                tt += "%)";
            }
        }
        tt += "\n\nDouble click on the statistics graph to change settings...";
    }
}


void CAlnStatGlyph::GetHTMLActiveAreas(TAreaVector* p_areas) const
{
    if (m_Config->IsBarGraph()) return;

    vector<int> rows = x_GetShowList();
    TModelUnit left = 0;
    TModelUnit top = 0;
    x_Local2World(left, top);
    top -= 3.0;
    vector<int>::reverse_iterator iter = rows.rbegin();
    for (; iter != rows.rend(); ++iter) {
        CHTMLActiveArea area;
        area.m_Bounds.SetTop(top);
        top += kSmearRowHeight;
        area.m_Bounds.SetBottom(top);
        area.m_Bounds.SetLeft(-1);
        area.m_Bounds.SetRight(0);
        area.m_Flags = CHTMLActiveArea::fComment
            | CHTMLActiveArea::fDrawBackground
            | CHTMLActiveArea::fNoHighlight
            | CHTMLActiveArea::fNoSelection
            | CHTMLActiveArea::fNoTooltip;

        area.m_ID = CAlnStatConfig::GetStatLabel(*iter);
        if ( !m_Config->ShowCount()  &&  *iter != CAlnStatConfig::eStat_Total) {
            area.m_ID += "(%)";
        }

        // required, but nonsense fields
        area.m_PositiveStrand = true;
        area.m_SeqRange.SetFrom(0);
        area.m_SeqRange.SetTo(0);

        p_areas->push_back(area);
    }
}


bool CAlnStatGlyph::IsClickable() const
{
    return true;
}


void CAlnStatGlyph::x_Draw() const
{
    if (m_Config->IsBarGraph()) {
        x_DrawBarGraph();
    } else {
        x_DrawDensityTable();
    }
}


void CAlnStatGlyph::x_UpdateBoundingBox()
{
    SetLeft(m_Context->GetVisibleFrom());
    SetWidth(m_Context->GetVisibleRange().GetLength());
    if (m_Config->IsBarGraph()) {
        SetHeight(m_Config->m_GraphHeight);
    } else {
        int row = (int)(x_GetShowList().size());
        SetHeight(kSmearRowHeight * row);
    }
}


void CAlnStatGlyph::x_DrawBarGraph() const
{
    CGlRender& gl = GetGl();

    TModelUnit bottom = GetBottom();
    TModelUnit top = GetTop();
    TModelUnit left = GetLeft();
    TModelUnit right = GetRight();
    TModelUnit bar_h = m_Config->m_GraphHeight;
    int max_t = x_GetMaxTotal();

    // draw background scales.
    // we do this only when we show the count for each individual statistics,
    // not when we show the percentage since it will be 100% always.
    if (m_Config->ShowCount()) {
        double tick_dist = 30.0;
        int tick_n = (int)(floor(bar_h / tick_dist));
        tick_dist = bar_h / tick_n;

        CGlAttrGuard AttrGuard(GL_LINE_BIT);
        gl.Disable(GL_LINE_SMOOTH);
        gl.LineWidth(1.0f);
        CRgbaColor line_color(0.7f, 0.7f, 0.7f, 0.5f);
        gl.ColorC(line_color);
        m_Context->DrawLine(left, bottom, right, bottom);
        if (max_t == 0) return;

        for (int i = 0; i < tick_n; ++i) {
            m_Context->DrawLine(left, bottom - tick_dist * (i + 1), right, bottom - tick_dist * (i + 1));
        }

        int ruler_n = 4;
        TModelUnit dist_x = GetWidth() / ruler_n;
        TModelUnit min_dist = m_Context->ScreenToSeq(400.0);
        if (dist_x < min_dist) {
            dist_x = min_dist;
            ruler_n = (int)(GetWidth() / dist_x);
        }
        TModelUnit label_off = m_Context->ScreenToSeq(1.0);
        //CGlBitmapFont font(CGlBitmapFont::eHelvetica8);
        CGlTextureFont font(CGlTextureFont::eFontFace_Helvetica_Bold, 8);
        TModelUnit x = left + dist_x * 0.1;
        string max_str = NStr::IntToString(max_t);
        for (;  x < right;  x += dist_x) {
            gl.ColorC(line_color);
            m_Context->DrawLine(x, top, x, bottom);
            gl.Color3f(0.7f, 0.7f, 0.9f);
            m_Context->TextOut(&font, max_str.c_str(), x + label_off, top + font.TextHeight() + 1, false);
        }
    }

    TModelUnit h_factor = bar_h;
    bool show_count = m_Config->ShowCount();
    if (show_count) {
        h_factor /= max_t;
    }

    vector<int> show_list = x_GetShowList();
    TModelUnit x = (TModelUnit)m_StartPos;

    // When the zoom level is greater than 1.0, the statistics data are computed
    // based on screen pixel (one set of data per pixel), not the nucleotide base.
    // So we need to use the correct offset (original scale when data are calculated),
    // not the the current zoom scale (m_Context->GetScale()).
    // When zoom level is less than or equal to 1.0, the statistics data are
    // computed based on nucleotide base (on set of data per nucleotide base).
    // So in that case, the offset will always be 1.
    TModelUnit off = m_ZoomScale > 1.0 ? m_ZoomScale : 1.0;

    ITERATE (TStatVec, iter, m_StatVec) {
        x += off;
        int total = iter->m_Data[CAlnStatConfig::eStat_Total];
        if (total == 0) continue;

        double total_factor = 1.0;
        if ( !show_count ) {
            total_factor /= total;
        }
        TModelUnit y = bottom;
        ITERATE (vector<int>, v_iter, show_list) {
            if (iter->m_Data[*v_iter] == 0) continue;

            TModelUnit h = iter->m_Data[*v_iter] * total_factor * h_factor;
            gl.ColorC(m_Config->m_Colors[*v_iter]);
            m_Context->DrawQuad(x - off, y, x, y - h);
            y -= h;
        }
        //if (m_Context->WillSeqLetterFit()) {
        //    // show border
        //    gl.Color4f(0.2f, 0.2f, 0.2f, 0.8f);
        //    m_Context->DrawLine(x - 1.0, bottom, x - 1.0, y);
        //    m_Context->DrawLine(x, bottom, x, y);
        //}
    }
}


void CAlnStatGlyph::x_DrawDensityTable() const
{
    CGlRender& gl = GetGl();

    TModelUnit bottom = GetBottom();
    TModelUnit top = GetTop();
    TModelUnit left = GetLeft();
    TModelUnit right = GetRight();
    
    // draw boundary lines
    {
        CGlAttrGuard AttrGuard(GL_LINE_BIT);
        gl.Disable(GL_LINE_SMOOTH);
        gl.LineWidth(1.0f);
        gl.Color4f(0.6f, 0.6f, 0.8f, 0.5f);
        m_Context->DrawLine(left, bottom, right, bottom);
        m_Context->DrawLine(left, top, right, top);
    }

    int max_t = x_GetMaxTotal();
    if (max_t == 0) return;

    TModelUnit d_factor = 1.0;
    bool show_count = m_Config->ShowCount();
    if (show_count) {
        d_factor /= max_t;
    }

    //CGlBitmapFont font(CGlBitmapFont::eHelvetica8);
    CGlTextureFont font(CGlTextureFont::eFontFace_Helvetica_Bold, 8);

    vector<int> show_list = x_GetShowList();
    TModelUnit x = (TModelUnit)m_StartPos;

    // See comment in CAlnStatGlyph::x_DrawBarGraph() on why offsets are different
    // for different zoom scales.
    TModelUnit off = m_ZoomScale > 1.0 ? m_ZoomScale : 1.0;

    ITERATE (TStatVec, iter, m_StatVec) {
        x += off;
        int total = iter->m_Data[CAlnStatConfig::eStat_Total];
        if (total == 0) continue;

        double total_factor = 1.0;
        if ( !show_count ) {
            total_factor /= total;
        }
        TModelUnit y = bottom;
        ITERATE (vector<int>, v_iter, show_list) {
            y -= kSmearRowHeight;
            if (iter->m_Data[*v_iter] == 0) continue;

            TModelUnit num = iter->m_Data[*v_iter];
            TModelUnit c_factor = num;
            if (*v_iter == CAlnStatConfig::eStat_Total) {
                c_factor /= max_t;
            } else {
                c_factor *= total_factor * d_factor;
                num *= total_factor;
                if ( !show_count ) num *= 100;
            }
            CRgbaColor color_max = m_Config->m_Colors[*v_iter];
            CRgbaColor color_min = color_max;
            color_min.SetAlpha(0.3f);
            CRgbaColor color(CRgbaColor::Interpolate(color_max, color_min, c_factor));
            gl.ColorC(color);
            m_Context->DrawQuad(x - off, y + kSmearRowHeight, x, y);

            // show count
            if (m_Context->GetScale() <= 1.0f / 16.0) {
                string out_text;
                if (num > 1000.0) {
                    int num_digit = 0;
                    while (num > 10.0) {
                        ++num_digit;
                        num *= 0.1;
                    }
                    out_text = NStr::IntToString((int)num) + "e" + NStr::IntToString(num_digit);
                } else if (num < 1.0) {
                    out_text = "<1";
                } else {
                    out_text = NStr::IntToString((int)num);
                }
                color = color.ContrastingColor();
                gl.ColorC(color);
                m_Context->TextOut(&font, out_text.c_str(),
                    x - 0.5, y + (kSmearRowHeight + font.TextHeight()) * 0.5, true);

            }
        }
    }


    if (m_Config->ShowLabel()) {
        // draw labels (left-most)
        vector<string> labels;
        TModelUnit max_label_w = 0.0;
        ITERATE (vector<int>, v_iter, show_list) {
            string label = CAlnStatConfig::GetStatLabel(*v_iter);
            if ( !show_count  &&  *v_iter != CAlnStatConfig::eStat_Total) {
                label += "(%)";
            }
            labels.push_back(label);
            max_label_w = max(max_label_w, font.TextWidth(label.c_str()));
        }

        x = left;
        TModelUnit y = bottom;

        gl.Color4f(0.8f, 1.0f, 0.8f, 1.0f);
        max_label_w += 4.0;
        max_label_w = m_Context->ScreenToSeq(max_label_w);
        m_Context->DrawQuad(x, y, x + max_label_w, top);

        gl.Color3f(0.0f, 0.0f, 0.0f);
        ITERATE (vector<string>, l_iter, labels) {
            const string& label = *l_iter;
            TModelUnit label_w = font.TextWidth(label.c_str()) + 2.0;
            label_w = m_Context->ScreenToSeq(label_w);
            m_Context->TextOut(&font, label.c_str(),
                               x + max_label_w - label_w, y - (kSmearRowHeight - font.TextHeight()) * 0.5, false);
            y -= kSmearRowHeight;
        }
    }
}


int CAlnStatGlyph::x_GetMaxTotal() const
{
    int max_t = 0;
    ITERATE (TStatVec, iter, m_StatVec) {
        if (max_t < iter->m_Data[CAlnStatConfig::eStat_Total]) {
            max_t = iter->m_Data[CAlnStatConfig::eStat_Total];
        }
    }
    return max_t;
}


vector<int> CAlnStatGlyph::x_GetShowList() const
{
    vector<int> show_list;
    if ( !m_Config->IsBarGraph()  &&  m_Config->ShowTotal() ) {
        show_list.push_back(CAlnStatConfig::eStat_Total);
    }

    if (m_Config->ShowAGTC()) {
        show_list.push_back(CAlnStatConfig::eStat_A);
        show_list.push_back(CAlnStatConfig::eStat_T);
        show_list.push_back(CAlnStatConfig::eStat_G);
        show_list.push_back(CAlnStatConfig::eStat_C);
    } else {
        show_list.push_back(CAlnStatConfig::eStat_Match);
        show_list.push_back(CAlnStatConfig::eStat_Mismatch);
    }
    show_list.push_back(CAlnStatConfig::eStat_Gap);
    return show_list;
}


END_NCBI_SCOPE
