/*  $Id: alignment_glyph.cpp 27850 2013-04-12 19:54:30Z wuliangs $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Vlad Lebedev, Liangshou Wu
 *
 * File Description:
 *   CAlignGlyph -- utility class to layout alignments
 */

#include <ncbi_pch.hpp>
#include <corelib/ncbistd.hpp>
#include <objmgr/util/sequence.hpp>
#include <algo/sequence/consensus_splice.hpp>
#include <objects/seq/MolInfo.hpp>
#include <objects/seqalign/Spliced_seg.hpp>
#include <objects/seqalign/Spliced_exon.hpp>
#include <objects/seqalign/Splice_site.hpp>

#include <gui/widgets/seq_graphic/seqgraphic_utils.hpp>
#include <gui/widgets/seq_graphic/alignment_glyph.hpp>
#include <gui/widgets/seq_graphic/rendering_ctx.hpp>
#include <gui/objutils/obj_fingerprint.hpp>
#include <gui/objutils/gui_object_info_seq_align.hpp>

#include <gui/opengl/glrender.hpp>
#include <gui/opengl/glutils.hpp>
#include <gui/opengl/rasterizer.hpp>
#include <gui/opengl/gltexturefont.hpp>
#include <util/sequtil/sequtil_manip.hpp>
#include <math.h>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

// How closely pack align. sequence when showing DNA alignments for proteins
// (pack 3 alignment bases into 1 sequence base)
static const TModelUnit kPWAlignPackLevel = 0.333;
static const TSeqPos kTailLen = 12;
static const int kTailHeightExtra = 0;
static const int kAlignRowSpace = 3;
static const int kLabelSpace_v = 2;
static const int kLabelSpace_h = 5;

CAlignGlyph::CAlignGlyph(CConstRef<IAlnGraphicDataSource> aln_mgr,
                           const CSeq_align_Handle& orig_align,
                           const CSeq_align& mapped_align,
                           int orig_anchor)
    : m_AlnMgr(aln_mgr)
    , m_SeqAlign(&mapped_align)
    , m_OrigSeqAlign(orig_align)
    , m_OrigAnchor(orig_anchor)
    , m_ShowScore(false)
{
    IAlnGraphicDataSource::TNumrow anchor = m_AlnMgr->GetAnchor();
    TSeqPos base_width_anchor = m_AlnMgr->GetBaseWidth(anchor);

    TSignedSeqPos start = m_AlnMgr->GetSeqStart(anchor) / base_width_anchor;
    TSignedSeqPos stop  = m_AlnMgr->GetSeqStop (anchor) / base_width_anchor;

    m_Location.Reset(new CSeq_loc());
    m_Location->SetInt().SetFrom(start);
    m_Location->SetInt().SetTo  (stop);

    CConstRef<CSeq_id> id(&m_AlnMgr->GetSeqId(anchor));
    m_Location->SetId(*id);
}


bool CAlignGlyph::NeedTooltip(const TModelPoint& /*p*/, string& /*tt*/, string& /*t_title*/) const
{
    return true;
}


void CAlignGlyph::GetTooltip(const TModelPoint& p, string& tt, string& t_title) const
{
    CScope& scope = m_Context->GetScope();
    SConstScopedObject scoped_obj(GetAlignment(), scope);
    CIRef<IGuiObjectInfo> gui_info(
        CreateObjectInterface<IGuiObjectInfo>(scoped_obj, NULL));

    if ( !gui_info ) return;

    CGuiObjectInfoSeq_align* gui_info_aln =
        dynamic_cast<CGuiObjectInfoSeq_align*>(gui_info.GetPointer());
    if (gui_info_aln) {
        gui_info_aln->SetAnchorRow(m_OrigAnchor);
    }

    TSeqPos at_p = (TSeqPos)-1;
    if (p.X() >= 0) {
        at_p = (TSeqPos)p.X();
    }

    gui_info->GetToolTip(tt, t_title, at_p);
}


void CAlignGlyph::GetHTMLActiveAreas(TAreaVector* p_areas) const
{
    CHTMLActiveArea area;
    CSeqGlyph::x_InitHTMLActiveArea(area);
    if (m_AlnMgr->GetNumRows() == 2) {
        area.m_PositiveStrand =
            m_AlnMgr->IsNegativeStrand(0) == m_AlnMgr->IsNegativeStrand(1);
    }
    area.m_Signature = GetSignature();
    p_areas->push_back(area);
}


TSeqRange CAlignGlyph::GetRange(void) const
{
    if (m_Intervals.size() == 1) {
        return m_Intervals.front();
    }

    return m_Location->GetTotalRange();
}


bool CAlignGlyph::IsClickable() const
{
    return true;
}


CConstRef<CObject> CAlignGlyph::GetObject(TSeqPos) const
{
    return CConstRef<CObject>(m_OrigSeqAlign.GetSeq_align());
}


void CAlignGlyph::GetObjects(vector<CConstRef<CObject> >& objs)  const
{
    objs.push_back( CConstRef<CObject>(m_OrigSeqAlign.GetSeq_align()));
}


bool CAlignGlyph::HasObject(CConstRef<CObject> obj) const
{
    return m_OrigSeqAlign.GetSeq_align() == obj.GetPointer();
}


const objects::CSeq_loc& CAlignGlyph::GetLocation(void) const
{
    return *m_Location;
}


string CAlignGlyph::GetSignature() const
{
    return CObjFingerprint::GetAlignSignature(
        GetOrigAlignment().GetSeq_align().GetObject(),
        GetLocation(), &m_Context->GetScope(), m_OrigSeqAlign.GetAnnot());
}


const CAlignGlyph::TIntervals& CAlignGlyph::GetIntervals(void) const
{
    return m_Intervals;
}


inline double CompressLength(double len)
{
    if(len <= 10.0) {
        return len;
    } else {
        double log = log10(len) * 10;
        return log;
    }
}


void CAlignGlyph::DrawPWAlignElem(const TModelRect& rcm,
                                  bool selected,
                                  bool show_label,
                                  const CRgbaColor& c_fg,
                                  const CRgbaColor& c_seq,
                                  const CRgbaColor& c_mis) const
{
    CGlRender& gl = GetGl();

    int anchor = m_AlnMgr->GetAnchor();
    int aligned_seq = 1 - anchor;
    if (aligned_seq < 0) {
        aligned_seq = 0;
    }
    TSeqPos base_width_anchor = m_AlnMgr->GetBaseWidth(anchor);
    TSeqPos base_width_curr = m_AlnMgr->GetBaseWidth(aligned_seq);

    TSeqPos f = m_AlnMgr->GetSeqStart(anchor);
    TSeqPos t = m_AlnMgr->GetSeqStop(anchor);
    f /= base_width_anchor;
    t /= base_width_anchor;

    TSeqRange seq_range(f, t);
    if (seq_range.Empty()) {
        LOG_POST(Error
            << "CAlignGlyph::DrawPWAlignElem: zero alignment range!");
        return;
    }

    TModelRange inrc = m_Context->IntersectVisible(seq_range);
    seq_range.Set((TSeqPos)inrc.GetFrom(), (TSeqPos)inrc.GetToOpen());
    if (seq_range.Empty()) {
        return;
    }

    TModelUnit bar_h = GetBarHeight();
    TModelUnit half = bar_h * 0.5;
    TModelUnit center_y = rcm.Bottom() - half - 1.0;
    TModelUnit yy = center_y - half;

    bool neg_strand = m_AlnMgr->IsNegativeStrand(aligned_seq) !=
        m_AlnMgr->IsNegativeStrand(anchor);

    if (IsSimplified()) {
        bool is_uneven_seg = false;
        if (m_ShowScore  &&  m_AlnMgr->IsRegular()) {
            double identity = 100.0;
                if ( !m_SeqAlign->GetNamedScore(CSeq_align::eScore_PercentIdentity, identity) ) {
                TSeqPos align_length = m_SeqAlign->GetAlignLength();
                int identities = (int)align_length;
                if ( !m_SeqAlign->GetNamedScore(CSeq_align::eScore_IdentityCount, identities) ) {
                    //    CScoreBuilder builder;
                    //    identities = builder.GetIdentityCount(*scope, align);
                } 
                identity = identities * 100 / align_length;
            }
            identity *= 0.01;
            if (identity > 0.6) {
                identity = 1.75 * identity - 0.75;
            } else {
                identity /= 2.0;
            }
            CRgbaColor color_max(0.78f, 0.78f, 0.78f);
            CRgbaColor color_min(1.0f, 0.0f, 0.0f);
            CRgbaColor color(CRgbaColor::Interpolate(color_max, color_min, identity));
            gl.ColorC(color);
            m_Context->DrawQuad(inrc.GetFrom(), yy, inrc.GetTo(), yy + bar_h, false);
        } else {
            m_Context->Draw3DQuad(inrc.GetFrom(), yy, inrc.GetTo(), yy + bar_h, c_fg, true);
        }

        // draw strand indicators
        if (m_Config->m_CompactMode != CAlignmentConfig::eExtremeCompact  &&
            m_Context->SeqToScreen(inrc.GetLength()) > 8.0) {
                TModelUnit xM = (inrc.GetFrom() + inrc.GetTo()) * 0.5;
                if (m_Context->WillSeqLetterFit()) {
                    // make sure we don't step on the sequence letter
                    // when sequence is shown
                    xM = ceil(xM);
                }
                TModelUnit yM = yy + bar_h * 0.5;
                gl.Color4f(0.0f, 0.0f, 0.0f, 1.0f);  // Black
                m_Context->DrawGreaterLessSign(xM, yM, bar_h, neg_strand);
        }

        if (IsSelected()) {
            m_Context->DrawSelection(rcm);
        }
        return;
    }

    TSignedSeqPos aln_r_s = m_AlnMgr->GetAlnPosFromSeqPos(
        anchor, seq_range.GetFrom() * base_width_anchor, IAlnExplorer::eRight);
    TSignedSeqPos aln_r_e = m_AlnMgr->GetAlnPosFromSeqPos(
        anchor, seq_range.GetTo() * base_width_anchor, IAlnExplorer::eLeft);
    if (aln_r_s > aln_r_e) {
        swap(aln_r_s, aln_r_e);
    }

    TSignedSeqRange aln_range(aln_r_s, aln_r_e);
    const CGlTextureFont& l_font = m_Config->m_LabelFont;
    bool show_score = x_ShowScore();
    int row_num = m_AlnMgr->GetNumRows();
    gl.ColorC(c_fg);

    TSeqPos tail_3 = 0;
    TModelUnit align_len = m_Context->SeqToScreen(seq_range.GetLength());
    bool show_unaligned_tails = x_ShowUnalignedTails(align_len);

    // get tail sizes
    if (show_unaligned_tails) {
        TSeqPos seq_len = m_AlnMgr->GetSeqLength(aligned_seq);
        tail_3 = seq_len - m_AlnMgr->GetSeqStop(aligned_seq) / base_width_curr - 1;
    }

    if ((m_Context->GetScale() < 1  &&  !show_score)  ||
        (show_score  &&  m_Context->WillSeqLetterFit()  &&  tail_3 == 0)) {
        TModelUnit over7 = m_Context->ScreenToSeq(7);
        TModelUnit over2 = m_Context->ScreenToSeq(2);

        if (neg_strand) {
            m_Context->DrawLine_NoAA(f - over2, yy + half, t + 1, yy + half);
        } else {
            m_Context->DrawLine_NoAA(f, yy + half, t + over2 + 1, yy + half);
        }

        if (show_score) {
            // make the color for the strand indicator matches well with
            // the alignment bar color
            gl.Color3f(0.5f, 0.5f, 0.5f);
        }
        if (neg_strand) {
            m_Context->DrawTriangle(f - over7, yy + half,
                f - over2, yy - 1, f - over2, yy + bar_h + 1);
        } else {
            m_Context->DrawTriangle(t + 1 + over7, yy + half,
                t + 1 + over2, yy - 1, t + 1 + over2, yy + bar_h + 1);
        }
    } else {
        m_Context->DrawLine_NoAA(f, yy + half, t, yy + half);
    }  // arrows


    // Draw Sequence and quads for segments
    if (show_score) {
        // do score coloration
        for (int row = 0;  row < row_num;  ++row) {
            if (row == anchor) {
                continue;  // skip master seq
            }

            if (m_Score->IsAverageable()) {
                x_RenderAveragableScores(yy, aln_range, seq_range, bar_h, row);
            } else {
                x_RenderNonAveragableScores(yy, aln_range, seq_range, bar_h, row);
            }
        }

        {{
            // color the gap differently
            CGlAttrGuard AttrGuard(GL_LINE_BIT);
            gl.Disable(GL_LINE_SMOOTH);

            const TModelUnit y_center = yy + half;
            CRgbaColor gap_color_min(0.78f, 0.5f, 0.5f);
            CRgbaColor gap_color_max(0.6f, 0.0f, 0.0f);

            for (int row = 0;  row < row_num;  ++row) {
                if (row == anchor) {
                    continue;  // skip master seq
                }
                auto_ptr<IAlnSegmentIterator> p_it(
                    m_AlnMgr->CreateSegmentIterator(row,
                    aln_range, IAlnSegmentIterator::eAllSegments));

                for (IAlnSegmentIterator& it = *p_it;  it;  ++it) {
                    const IAlnSegment& seg = *it;
                    if ( !(seg.GetType() & IAlnSegment::fIndel)  ||
                         !seg.GetRange().Empty() ) {
                             continue;
                    }

                    const IAlnSegment::TSignedRange& curr_aln_r = seg.GetAlnRange();
                    TModelUnit curr_seq_from =
                        m_AlnMgr->GetSeqPosFromAlnPos(anchor, curr_aln_r.GetFrom()) /
                        base_width_anchor;
                    TModelUnit curr_seq_to =
                        m_AlnMgr->GetSeqPosFromAlnPos(anchor, curr_aln_r.GetTo()) /
                        base_width_anchor;
                    if (curr_seq_from > curr_seq_to) {
                        swap (curr_seq_from, curr_seq_to);
                    }
                    curr_seq_to += 1.0 / base_width_anchor;
                    TModelUnit gap_size = m_Context->SeqToScreen(curr_seq_to - curr_seq_from);
                    if (gap_size > 1.5) {
                        gl.LineWidth(2.0);
                        gl.ColorC(gap_color_max);
                        m_Context->DrawLine(curr_seq_from, y_center, curr_seq_to, y_center);
                    } else {
                        if (gap_size > 1.0) {
                            gap_size = 1.0;
                        } else if (gap_size < 0.1) {
                            gap_size = 0.1;
                        }

                        CRgbaColor gap_color_ins(CRgbaColor::Interpolate(gap_color_max,
                            gap_color_min, gap_size));
                        gl.ColorC(gap_color_ins);
                        gl.LineWidth(1.0);
                        TModelUnit c_x = (curr_seq_from + curr_seq_to) * 0.5;
                        m_Context->DrawLine_NoAA(c_x, yy, c_x, yy + bar_h);
                    }
                }
            }
            gl.LineWidth(1.0f);
        }}

        // draw inserts

        TModelUnit over = 2.0;
        TModelUnit down = 3.0;
        const TModelUnit y1 = yy + down;
        const TModelUnit y2 = yy + bar_h - down;

        CRgbaColor color_max(0.3f, 0.3f, 0.8f);
        CRgbaColor color_min(0.7f, 0.7f, 0.8f);

        for (int row = 0;  row < row_num;  ++row) {
            if (row == anchor) {
                continue;  // skip master seq
            }
            auto_ptr<IAlnSegmentIterator> p_it(
                m_AlnMgr->CreateSegmentIterator(row,
                aln_range, IAlnSegmentIterator::eAllSegments));

            int last_aligned_seg_to_anchor = -2;
            int last_aligned_seg_to_curr = -2;
            int last_aligned_seg_from_anchor = -2;
            int last_aligned_seg_from_curr = -2;

            int last_large_ins = -100000;  // >= 5 bases
            TSeqPos base_width_curr = m_AlnMgr->GetBaseWidth(row);

            int seg_iter_dir = 0;  // 0: unknown, 1: forward (anchor), -1: reverse (anchor) 
            for (IAlnSegmentIterator& it = *p_it;  it;  ++it)  {
                const IAlnSegment& seg = *it;
                TSeqPos start = 0;
                int ins_len = 0;
                if ( !(seg.GetType() & IAlnSegment::fIndel)  ||
                    seg.GetRange().Empty() ) {
                    // it is not marked as insertion
                    // check if it is intron
                    if (seg.GetType() & IAlnSegment::fAligned) {
                        if (seg_iter_dir == 0) {
                            int anchor_to =
                                m_AlnMgr->GetSeqPosFromAlnPos(anchor, seg.GetAlnRange().GetTo(), IAlnExplorer::eRight);
                            int anchor_from =
                                m_AlnMgr->GetSeqPosFromAlnPos(anchor, seg.GetAlnRange().GetFrom(), IAlnExplorer::eRight);
                            if (last_aligned_seg_to_anchor < 0) {
                                // It must be the first segment (unknow iterator direction)
                                last_aligned_seg_to_anchor = anchor_to;
                                last_aligned_seg_from_anchor = anchor_from;
                                last_aligned_seg_to_curr = seg.GetRange().GetTo();
                                last_aligned_seg_from_curr = seg.GetRange().GetFrom();
                            } else {
                                if (anchor_to < last_aligned_seg_to_anchor || anchor_from < last_aligned_seg_to_anchor) {
                                    seg_iter_dir = -1;
                                } else {
                                    seg_iter_dir = 1;
                                }
                            }
                        }
                        
                        if (seg_iter_dir == -1) {
                            TSeqPos t_pos = m_AlnMgr->GetSeqPosFromAlnPos(anchor,
                                seg.GetAlnRange().GetTo(), IAlnExplorer::eRight);
                            if ((int)t_pos == last_aligned_seg_from_anchor - 1) {
                                start = t_pos / base_width_anchor;
                                if (neg_strand) {
                                    ins_len = seg.GetRange().GetFrom() - last_aligned_seg_to_curr;
                                } else {
                                    ins_len = last_aligned_seg_from_curr - seg.GetRange().GetTo();
                                }
                                ins_len -= 1;
                            }
                            last_aligned_seg_to_anchor = t_pos;
                            last_aligned_seg_from_anchor = m_AlnMgr->GetSeqPosFromAlnPos(anchor,
                                seg.GetAlnRange().GetFrom(), IAlnExplorer::eRight);
                        } else if (seg_iter_dir == 1) {
                            TSeqPos f_pos = m_AlnMgr->GetSeqPosFromAlnPos(anchor,
                                seg.GetAlnRange().GetFrom(), IAlnExplorer::eRight);
                            if ((int)f_pos == last_aligned_seg_to_anchor + 1) {
                                start = f_pos / base_width_anchor;
                                if (neg_strand) {
                                    ins_len = last_aligned_seg_from_curr - seg.GetRange().GetTo();
                                } else {
                                    ins_len = seg.GetRange().GetFrom() - last_aligned_seg_to_curr;
                                }
                                ins_len -= 1;
                            }
                            last_aligned_seg_from_anchor = f_pos;
                            last_aligned_seg_to_anchor = m_AlnMgr->GetSeqPosFromAlnPos(anchor,
                                seg.GetAlnRange().GetTo(), IAlnExplorer::eRight);
                        }
                        last_aligned_seg_from_curr = seg.GetRange().GetFrom();
                        last_aligned_seg_to_curr = seg.GetRange().GetTo();
                    }
                } else {
                    start = m_AlnMgr->GetSeqPosFromAlnPos(anchor,
                        seg.GetAlnRange().GetFrom(),
                        IAlnExplorer::eRight);
                    start /= base_width_anchor;
                    ins_len = seg.GetRange().GetLength();
                    // reset to avoid drawing insertion twice
                    last_aligned_seg_to_anchor = -2;
                    last_aligned_seg_from_anchor = -2;
                }
                if (ins_len <= 0) continue;

                TModelUnit ins_len_pix = m_Context->SeqToScreen(ins_len);
                // adjust it to the anchor sequence base length
                ins_len_pix /= base_width_anchor; 

                if (ins_len < 5) {
                    if (((int)start - last_large_ins) < m_Context->ScreenToSeq(1.0)) {
                        // skip the minor ones that overlap with the last major one
                        continue;
                    }
                } else if (ins_len_pix > 2.0) {
                    // store the last major insert ( > 5 bases) location
                    last_large_ins = (int)start;
                }


                TModelUnit scale = min(10.0 * ins_len/ m_Context->GetScale(), 1.0);
                CRgbaColor color_ins(CRgbaColor::Interpolate(color_max, color_min, scale));

                gl.ColorC(color_ins);
                m_Context->DrawLine_NoAA(start, yy, start, yy + bar_h);
                if (ins_len_pix > 2.0  &&  m_Config->m_CompactMode == CAlignmentConfig::eNormal) {
                    TModelUnit size_in_model
                        = min(m_Context->ScreenToSeq(4), (TModelUnit)ins_len);
                    m_Context->DrawTriangle(start, y1,
                        start + size_in_model, yy - over,
                        start - size_in_model, yy - over);
                    m_Context->DrawTriangle(start, y2,
                        start - size_in_model, yy + bar_h + over,
                        start + size_in_model, yy + bar_h + over);
                    TModelUnit len = max(CompressLength(ins_len), size_in_model * 2.0);
                    TModelUnit x1 = start - len * 0.5;
                    TModelUnit x2 = x1 + len;
                    m_Context->DrawLine_NoAA(x1, yy - over, x2, yy - over);
                    m_Context->DrawLine_NoAA(x1, yy + bar_h + over, x2, yy + bar_h + over);
                }
            }
        }

        // draw unaligned tails
        if (show_unaligned_tails) {
            int aligned_seq = anchor == 0 ? 1 : 0;
            TSeqPos base_width_curr = m_AlnMgr->GetBaseWidth(aligned_seq);
            TSeqPos seq_len = m_AlnMgr->GetSeqLength(aligned_seq);
            TSeqPos seq_start = m_AlnMgr->GetSeqStart(aligned_seq) / base_width_curr;
            TSeqPos seq_end = m_AlnMgr->GetSeqStop(aligned_seq) / base_width_curr;
            TModelUnit size = m_Context->ScreenToSeq(kTailLen);
            TModelUnit gap = m_Context->ScreenToSeq(kLabelSpace_h);
            TModelUnit y_base = yy;

            // starting tail
            TSeqPos unaligned_len =
                neg_strand ? seq_len - seq_end - 1 : seq_start;
            bool top_label =
                m_Config->m_LabelPos == CAlignmentConfig::ePos_Above;

            if (unaligned_len > 0) {
                TModelUnit x_base = f - gap;
                gl.ColorC(c_fg);
                bool is_polya = x_IsPolyA(
                    m_AlnMgr->GetBioseqHandle(aligned_seq),
                    neg_strand ? seq_end + 1 : 0,
                    neg_strand ? seq_len - 1 : seq_start - 1);
                m_Context->DrawUnalignedTail(x_base - size, x_base, y_base,
                    y_base + bar_h, is_polya);
                if (show_label) {
                    TModelUnit xM = x_base - size * 0.5 ;
                    TModelUnit yM = y_base - kLabelSpace_v;
                    string len_str =
                        NStr::IntToString(unaligned_len, NStr::fWithCommas);
                    if (top_label) {
                        TModelUnit num_len = l_font.TextWidth(len_str.c_str());
                        xM = x_base - size - gap - m_Context->ScreenToSeq(num_len);
                        yM = GetBottom() - 2.0;
                    }
                    gl.ColorC(m_Config->m_Label);
                    m_Context->TextOut(&l_font, len_str.c_str(), xM, yM, !top_label);
                }
            }

            // ending tail
            unaligned_len = neg_strand ? seq_start : seq_len - seq_end - 1;
            if (unaligned_len > 0) {
                TModelUnit x_base = t + gap + 1;
                gl.ColorC(c_fg);
                bool is_polya = x_IsPolyA(
                    m_AlnMgr->GetBioseqHandle(aligned_seq),
                    neg_strand ? 0 : seq_end + 1,
                    neg_strand ? seq_start - 1 : seq_len - 1);
                m_Context->DrawUnalignedTail(x_base, x_base + size, y_base,
                    y_base + bar_h, is_polya);
                if (show_label) {
                    TModelUnit xM = x_base + size * 0.5;
                    TModelUnit yM = y_base - kLabelSpace_v;
                    string len_str =
                        NStr::IntToString(unaligned_len, NStr::fWithCommas);
                    if (top_label) {
                        xM = x_base + size + gap;
                        yM = GetBottom() - 2.0;
                    }
                    gl.ColorC(m_Config->m_Label);
                    m_Context->TextOut(&l_font, len_str.c_str(), xM, yM, !top_label);
                }
            }
        }
    } else {
        for (int row = 0;  row < row_num;  ++row) {
            if (row == anchor) {
                continue;  // skip master seq
            }

            auto_ptr<IAlnSegmentIterator> p_it(
                m_AlnMgr->CreateSegmentIterator(
                row, aln_range, IAlnSegmentIterator::eSkipGaps));

            for (IAlnSegmentIterator& it(*p_it);  it;  ++it) {
                const IAlnSegment& seg = *it;
                TModelUnit sStart = m_AlnMgr->GetSeqPosFromAlnPos(anchor,
                    seg.GetAlnRange().GetFrom());
                sStart /= base_width_anchor;
                TModelUnit sStop  = m_AlnMgr->GetSeqPosFromAlnPos(anchor,
                    seg.GetAlnRange().GetTo());
                sStop /= base_width_anchor;

                if (sStart > sStop) {
                    swap(sStart, sStop);
                }
                m_Context->Draw3DQuad(sStart, yy, sStop, yy + bar_h, c_fg, true);
            }
        } 
    }

    // special case
    // reflect non-consensus splice sites for spliced-seg alignment
    // we only show non-consensus site when showing in genomic sequence
    // context, not product sequence context.
    if (m_SeqAlign->GetSegs().Which() == CSeq_align::TSegs::e_Spliced  &&
        m_OrigAnchor == 1) {
        const CSpliced_seg& spliced_seg = m_SeqAlign->GetSegs().GetSpliced();
        string splice5;
        TSeqPos pre_stop = 0;
        bool neg_genomic = (spliced_seg.GetSeqStrand(1) == eNa_strand_minus);
        ITERATE (CSpliced_seg::TExons, exon_it, spliced_seg.GetExons()) {
            const CSpliced_exon& exon = **exon_it;
            string splice3;
            if (exon.IsSetAcceptor_before_exon()) {
                splice3 = exon.GetAcceptor_before_exon().GetBases();
            }

            // render a bold red bar for non-consensus splice site
            if ( !splice5.empty()  &&  !splice3.empty()  &&
                !IsConsensusSplice(splice5, splice3) ) {
                gl.LineWidth(4.0f);
                gl.Color3f(0.91f, 0.51f, 0.07f);
                m_Context->DrawLine(pre_stop, center_y,
                    neg_genomic ? exon.GetGenomic_end() + 1 :
                    exon.GetGenomic_start(), center_y);
                gl.LineWidth(1.0f);
            }
            if (exon.IsSetDonor_after_exon()) {
                splice5 = exon.GetDonor_after_exon().GetBases();
            } else {
                splice5 = kEmptyStr;
            }
            pre_stop = neg_genomic ?
                exon.GetGenomic_start() : exon.GetGenomic_end() + 1;
        }            
    }

    if ((m_Context->WillSeqLetterFit()  &&  base_width_anchor <= base_width_curr)
        || (m_Context->GetScale() <= 0.05  &&  base_width_anchor > base_width_curr)) {
        try {
            x_DrawAlignRowSequence(center_y, 1 - anchor, c_fg, c_seq, c_mis);
        } catch (CException& e) {
            LOG_POST(Warning << "Can't show alignment sequence due to an error: "
                << e.GetMsg());
        }
    }
    
    // draw strand indicators
    if (m_Config->m_CompactMode != CAlignmentConfig::eExtremeCompact  &&
        m_Context->SeqToScreen(inrc.GetLength()) > 8.0) {
        TModelUnit xM = (inrc.GetFrom() + inrc.GetTo()) * 0.5;
        if (m_Context->WillSeqLetterFit()) {
            // make sure we don't step on the sequence letter
            // when sequence is shown
            xM = ceil(xM);
        }
        TModelUnit yM = yy + bar_h * 0.5;
        gl.Color4f(0.0f, 0.0f, 0.0f, 1.0f);  // Black
        m_Context->DrawGreaterLessSign(xM, yM, bar_h, neg_strand);
    }

    // Draw selection
    if (selected) {
        m_Context->DrawSelection(rcm);
    }
}


void CAlignGlyph::GetTitle(string* title, CLabel::ELabelType type) const
{
    title->erase();

    if ( !x_IsPWAlign()) {
        *title = "[Alignment] " +
            NStr::UIntToString(m_AlnMgr->GetNumRows()) + " sequences";
        return;
    }

    // pairwise align
    switch (type) {
    case CLabel::eType:
    case CLabel::eContent:
    case CLabel::eUserTypeAndContent:
        {
            // display the name of the aligned sequence.
            int row = 1 - m_OrigAnchor;
            CLabel::GetLabel(m_OrigSeqAlign.GetSeq_align()->GetSeq_id(row),
                title, type, &m_Context->GetScope());

            // make title/label shorter incase it is a cSRA alignment on panfs
            if (title->length() > 20  &&  title->find("\\panfs\\") != string::npos) {
                size_t pos = title->find_last_of("\\");
                if (pos != string::npos) {
                    *title = title->substr(pos + 1);
                    if (title->length() > 20) {
                        pos = title->rfind('.');
                        if (pos != string::npos  &&  pos > 0) {
                            pos = title->rfind('.', pos - 1);
                            if (pos != string::npos) {
                                *title = title->substr(pos + 1);
                            }
                        }
                    }
                }
            }
        };
        break;
    case CLabel::eDescriptionBrief:
    case CLabel::eDescription:
        CLabel::GetLabel(*GetOrigAlignment().GetSeq_align(),
            title, type, &m_Context->GetScope());
        break;
    default:
        break;
    }
}


void CAlignGlyph::x_Draw() const
{
    if (x_IsPWAlign()) {
        x_DrawPWAlign();
    } else {
        x_DrawMultiAlign();
    }
}


void CAlignGlyph::x_UpdateBoundingBox()
{
    if (m_Intervals.empty()) {
        x_CalcIntervals();
    }

    TSeqRange range = GetRange();
    m_Width = range.GetLength();
    m_Pos.m_X = range.GetFrom();

    TModelUnit bar_h = GetBarHeight();
    if (x_IsPWAlign()) {
        m_Height = bar_h + (m_Config->m_CompactMode == CAlignmentConfig::eNormal ? 2 : 0);
    } else {
        m_Height = (bar_h + kAlignRowSpace) * (m_AlnMgr->GetNumRows() - 1);
    }

    bool show_label = x_ShowLabel();
    // don't show labels and unaligned tails for simplified rendering
    if (IsSimplified()  &&  !show_label) {
        return;
    }

    bool top_label = m_Config->m_LabelPos == CAlignmentConfig::ePos_Above;
    const CGlTextureFont& font = m_Config->m_LabelFont;
    int anchor = m_AlnMgr->GetAnchor();
    bool neg_strand = m_AlnMgr->IsNegativeStrand(0) !=
        m_AlnMgr->IsNegativeStrand(1);
    TSeqPos tail_5 = 0;
    TSeqPos tail_3 = 0;
    TModelUnit align_len = m_Context->SeqToScreen(GetRange().GetLength());
    
    // get tail sizes
    if (x_ShowUnalignedTails(align_len)) {
        int aligned_seq = 1 - anchor;
        TSeqPos base_width_curr = m_AlnMgr->GetBaseWidth(aligned_seq);
        TSeqPos seq_len = m_AlnMgr->GetSeqLength(aligned_seq);
        TSeqPos seq_start = m_AlnMgr->GetSeqStart(aligned_seq) / base_width_curr;
        TSeqPos seq_end = m_AlnMgr->GetSeqStop(aligned_seq) / base_width_curr;

        // starting tail
        tail_5 = seq_start;

        // ending tail
        tail_3 = seq_len - seq_end - 1;

        // preserve horizontal space for unaligned tail glyph
        TModelUnit tail_len = m_Context->ScreenToSeq(kTailLen + kLabelSpace_h);
        if (tail_5 > 0) {
            m_Width += tail_len;
            if ( !neg_strand ) {
                m_Pos.m_X -= tail_len;
            }
        }
        if (tail_3 > 0) {
            m_Width += tail_len;
            if (neg_strand) {
                m_Pos.m_X -= tail_len;
            }
        }
    }

    if (show_label) {
        if (top_label) {
            if (x_IsPWAlign()) {
                // vertical space for alignment label
                m_Height += font.TextHeight() + 3;
                // horizontal space for unaligned tail labels
                // For top label, we put the unaligned tail numbers on the side
                if (tail_5 > 0) {
                    string len_str = NStr::IntToString(tail_5, NStr::fWithCommas);
                    TModelUnit num_w = font.TextWidth(len_str.c_str());
                    num_w = m_Context->ScreenToSeq(num_w + kLabelSpace_h);
                    m_Width += num_w;
                    if ( !neg_strand ) {
                        m_Pos.m_X -= num_w;
                    }
                }
                if (tail_3 > 0) {
                    string len_str = NStr::IntToString(tail_3, NStr::fWithCommas);
                    TModelUnit num_w = font.TextWidth(len_str.c_str());
                    num_w = m_Context->ScreenToSeq(num_w + kLabelSpace_h);
                    m_Width += num_w;
                    if (neg_strand) {
                        m_Pos.m_X -= num_w;
                    }
                }
            } else {
                m_Height += (font.TextHeight() + kLabelSpace_v) *
                    (m_AlnMgr->GetNumRows() - 1);
            }
        } else { // not top label (side label)
            if (x_IsPWAlign()) {
                // horizontal space for alignment label
                string label;
                GetTitle(&label, CLabel::eDefault);
                TModelUnit label_w = font.TextWidth(label.c_str());
                label_w = m_Context->ScreenToSeq(label_w + kLabelSpace_h);
                m_Width += label_w;
                if ( !neg_strand ) {
                    m_Pos.m_X -= label_w;
                }

                // vertical space for unaligned tail labels
                // For side label, we put the unaligned tail numbers on the top
                if (tail_5 > 0  ||  tail_3 > 0) {
                    m_Height += font.TextHeight() + 3;
                }

                // horizontal space for unaligned tail labels.
                // It is possible the 3' side unaligned tail number is too big
                // to fit within the tail glyph size. We need to increase the 
                // bounding box size to make sure the tail label is included.
                // 5' side should be fine since we have already preserve space
                // for alignment's side label.
                if (tail_3 > 0) {
                    string len_str = NStr::IntToString(tail_3, NStr::fWithCommas);
                    TModelUnit num_w = font.TextWidth(len_str.c_str());
                    if (num_w > kTailLen) {
                        TModelUnit extra_space =
                            m_Context->ScreenToSeq(num_w - kTailLen) * 0.5;
                        m_Width += extra_space;
                        if (neg_strand) {
                            m_Pos.m_X -= extra_space;
                        }
                    }
                }
            } else {
                TModelUnit max_label_w = 0.0;
                for (int row = 0;  row < m_AlnMgr->GetNumRows();  ++row) {
                    if (row == anchor) {
                        continue;  // skip master seq
                    }
                    string label;
                    CLabel::GetLabel(m_AlnMgr->GetSeqId(row), &label,
                        CLabel::eContent, &m_Context->GetScope());
                    if (m_AlnMgr->IsNegativeStrand(anchor) !=
                        m_AlnMgr->IsNegativeStrand(row)) {
                        label += " (Neg)";
                    }
                    TModelUnit label_w = font.TextWidth(label.c_str());
                    max_label_w = max(max_label_w, label_w);
                }
                max_label_w += kLabelSpace_h;
                max_label_w = m_Context->ScreenToSeq(max_label_w);
                m_Width += max_label_w;
                m_Pos.m_X -= max_label_w;
            }
        }
    }
}


void CAlignGlyph::x_DrawPWAlign() const
{
    CGlRender& gl = GetGl();

    TModelRect rcm = GetModelRect();
    bool show_label = x_ShowLabel();

    DrawPWAlignElem(rcm, IsSelected(), show_label, m_Config->m_FG,
        m_Config->m_Sequence, m_Config->m_SeqMismatch);

    if (show_label) {
        string label;
        GetTitle(&label, CLabel::eDefault);
        const CGlTextureFont& l_font = m_Config->m_LabelFont;
        bool top_label = m_Config->m_LabelPos == CAlignmentConfig::ePos_Above;
        TModelUnit label_h = l_font.TextHeight();
        TModelRange inrc = m_Context->IntersectVisible(GetRange());
        TModelUnit xM = (inrc.GetFrom() + inrc.GetTo()) * 0.5;
        TModelUnit yM = rcm.Top() + label_h + 1.0;
        TModelUnit max_label_w = inrc.GetLength();
        inrc = m_Context->IntersectVisible(this);
        if (!top_label) {
            max_label_w = inrc.GetLength();
        }
        max_label_w = m_Context->SeqToScreen(max_label_w);
        if (max_label_w > 2.0) max_label_w -= 2.0;
        TModelUnit label_w = l_font.TextWidth(label.c_str());
        if (label_w > max_label_w) {
            label_w = max_label_w;
            label = l_font.Truncate(label.c_str(), label_w);
        }

        if ( !top_label ) {
            TModelUnit t_w = m_Context->ScreenToSeq(label_w);
            bool neg_strand = m_AlnMgr->IsNegativeStrand(0) !=
                m_AlnMgr->IsNegativeStrand(1);
            yM = GetBottom() - 2.0;
            if (neg_strand) {
                xM = inrc.GetTo() - t_w;
                if (inrc.GetTo() < GetRight()) {
                    gl.ColorC(m_Config->m_BG);
                    m_Context->DrawBackground( TModelRect(xM, yM + 1.0,
                        inrc.GetTo(), yM - label_h - 1.0), 0);
                }
            } else {
                xM = inrc.GetFrom();
                if (inrc.GetFrom() > GetLeft()) {
                    gl.ColorC(m_Config->m_BG);
                    m_Context->DrawBackground( TModelRect(xM, yM + 1.0,
                        xM + t_w, yM - label_h - 1.0), 0);
                }
            }
        }

        if (IsSelected()) {
            gl.ColorC(m_Context->GetSelLabelColor());
        } else {
            gl.ColorC(m_Config->m_Label);
        }
        m_Context->TextOut(&l_font, label.c_str(), xM, yM, top_label);
    }
}


void CAlignGlyph::x_DrawMultiAlign() const
{
    CGlRender& gl = GetGl();

    int anchor = m_AlnMgr->GetAnchor();
    TSeqPos base_width_anchor = m_AlnMgr->GetBaseWidth(anchor);
    TSeqRange anchor_range(
        m_AlnMgr->GetSeqStart(anchor) / base_width_anchor,
        m_AlnMgr->GetSeqStop(anchor) / base_width_anchor);
    if (anchor_range.Empty()) {
        LOG_POST(Error 
            << "CAlignGlyph::x_DrawMultiAlign: zero alignment range!");
        return;
    }

    TModelRange inrc = m_Context->IntersectVisible(anchor_range);
    anchor_range.Set((TSeqPos)inrc.GetFrom(), (TSeqPos)inrc.GetTo());
    if (anchor_range.Empty()) {
        return;
    }

    TSignedSeqPos aln_r_s = m_AlnMgr->GetAlnPosFromSeqPos(anchor,
        anchor_range.GetFrom() * base_width_anchor, IAlnExplorer::eRight);
    TSignedSeqPos aln_r_e = m_AlnMgr->GetAlnPosFromSeqPos(anchor,
        anchor_range.GetTo() * base_width_anchor, IAlnExplorer::eLeft);
    if (aln_r_s > aln_r_e) {
        swap(aln_r_s, aln_r_e);
    }
    TSignedSeqRange aln_total_range(aln_r_s, aln_r_e);


    TModelRect rcm = GetModelRect();
    TModelUnit base = rcm.Top();
    TModelUnit bar_h = GetBarHeight();

    bool show_label = x_ShowLabel();
    TModelUnit yy = base + 1;
    for (int row = 0;  row < m_AlnMgr->GetNumRows();  ++row) {
        if (row == anchor) {
            continue;  // skip master seq
        }

        TSignedSeqRange aln_range =
            aln_total_range.IntersectionWith(m_AlnMgr->GetSeqAlnRange(row));
        TModelUnit xM = (inrc.GetFrom() + inrc.GetTo()) * 0.5;
        bool neg_strand =
            m_AlnMgr->IsNegativeStrand(row) !=
            m_AlnMgr->IsNegativeStrand(anchor);

        // Draw label
        if (show_label  &&
            m_Config->m_LabelPos == CAlignmentConfig::ePos_Above) {
            const CGlTextureFont& l_font = m_Config->m_LabelFont;
            yy += l_font.TextHeight();
            x_DrawMultiAlignLabel(row, yy);
            yy += kLabelSpace_v;
        }

        TModelUnit mid = yy + bar_h * 0.5;
        if (x_ShowScore()) {
            // do score coloration
            if (m_Score->IsAverageable()) {
                x_RenderAveragableScores(yy, aln_range, anchor_range, bar_h, row);
            } else {
                x_RenderNonAveragableScores(yy, aln_range, anchor_range, bar_h, row);
            }

            // color the gap differently
            gl.Color4f(1.0f, 0.0f, 0.0f, 1.0f);
            CGlAttrGuard AttrGuard(GL_LINE_BIT);
            gl.Disable(GL_LINE_SMOOTH);

            auto_ptr<IAlnSegmentIterator> p_it(
                m_AlnMgr->CreateSegmentIterator(row, aln_range,
                IAlnSegmentIterator::eAllSegments));

            gl.LineWidth(2.0f);
            for (IAlnSegmentIterator& it = *p_it;  it;  ++it)  {
                const IAlnSegment& seg = *it;
                if ( !(seg.GetType() & IAlnSegment::fIndel)  ||
                    !seg.GetRange().Empty() ) {
                        continue;
                }

                TSeqPos start = m_AlnMgr->GetSeqPosFromAlnPos(anchor, seg.GetAlnRange().GetFrom());
                start /= base_width_anchor;
                TSeqPos stop = start + seg.GetAlnRange().GetLength() / base_width_anchor;
                m_Context->DrawLine(start, mid, stop, mid);
            }
            gl.LineWidth(1.0f);

            // draw inserts
            gl.Color4f(0.0f, 0.0f, 1.0f, 1.0f);
            const TModelUnit barOff = 
                m_AlnMgr->IsNegativeStrand(anchor) ? 1.0 : 0.0;
            const TModelUnit kMinLen = m_Context->ScreenToSeq(4); // 4 pixels in model units

            TModelUnit y1 = yy;
            TModelUnit y2 = yy + bar_h;
            p_it.reset(m_AlnMgr->CreateSegmentIterator(
                row, aln_range, IAlnSegmentIterator::eAllSegments));
           
            for (IAlnSegmentIterator& it = *p_it;  it;  ++it)  {
                const IAlnSegment& seg = *it;

                if ( !(seg.GetType() & IAlnSegment::fIndel)  ||
                    seg.GetRange().Empty() ) {
                        continue;
                }

                TSeqPos start = m_AlnMgr->GetSeqPosFromAlnPos(anchor,
                    seg.GetAlnRange().GetTo(),
                    IAlnExplorer::eRight);
                start /= base_width_anchor;
                TModelUnit len = CompressLength(seg.GetRange().GetLength());
                len = max(len, kMinLen);
                TModelUnit x1 = start + barOff - len * 0.5;
                TModelUnit x2 = x1 + len;

                m_Context->DrawLine_NoAA(start + barOff, y1, start + barOff, y2);
                m_Context->DrawLine_NoAA(x1, y1, x2, y1);
                m_Context->DrawLine_NoAA(x1, y2, x2, y2);
            }
        } else {  // use default scheme
            gl.ColorC(m_Config->m_FG);

            // draw connection line in the middle
            m_Context->DrawLine_NoAA(anchor_range.GetFrom(), mid, anchor_range.GetTo(),  mid);
            auto_ptr<IAlnSegmentIterator> p_it(
                m_AlnMgr->CreateSegmentIterator(row,
                aln_range, IAlnSegmentIterator::eSkipGaps));

            // draw elements as quads
            for ( IAlnSegmentIterator& it(*p_it);  it;  ++it )  {
                const IAlnSegment& seg = *it;
                const TSignedSeqRange& R = seg.GetAlnRange();
                TSeqPos start = m_AlnMgr->GetSeqPosFromAlnPos(anchor, R.GetFrom());
                start /= base_width_anchor;
                TSeqPos stop = m_AlnMgr->GetSeqPosFromAlnPos(anchor, R.GetTo());
                stop /= base_width_anchor;
                if (start > stop) {
                    swap(start, stop);
                }
                m_Context->Draw3DQuad(start, yy, stop, yy + bar_h, m_Config->m_FG, true);
            }
        }

        if (m_Context->WillSeqLetterFit()) {
            try {
                x_DrawAlignRowSequence(mid, row, m_Config->m_FG,
                    m_Config->m_Sequence, m_Config->m_SeqMismatch);
            } catch (CException& e) {
                LOG_POST(Warning << "Can't show alignment sequence due to an error: "
                    << e.GetMsg());
            }

        } else if (m_Context->WillLabelFit(inrc)) {
            // draw strand indicators
            // do not draw strand when sequence is shown
            gl.Color4f(0.0f, 0.0f, 0.0f, 1.0f);  // Black
            gl.Color4f(0.0f, 0.0f, 0.0f, 1.0f);  // Black
            TModelUnit fs = m_Config->m_SeqFont.TextHeight() - 1;
            TModelUnit yM = yy + fs + 1;
            char strandLabel[2] = {'>', '\0'};
            if (m_Context->IsHorizontal()){
                if (m_Context->IsFlippedStrand()){
                    strandLabel[0] = neg_strand ? '>' : '<';
                } else {
                    strandLabel[0] = neg_strand ? '<' : '>';
                }
            } else {
                if (m_Context->IsFlippedStrand()){
                    strandLabel[0] = neg_strand ? '<' : '>';
                } else {
                    strandLabel[0] = neg_strand ? '>' : '<';
                }
            }
            m_Context->TextOut(&m_Config->m_SeqFont, strandLabel, xM, yM, true);
        }

        if (show_label  &&
            m_Config->m_LabelPos == CAlignmentConfig::ePos_Side) {
            const CGlTextureFont& l_font = m_Config->m_LabelFont;
            TModelUnit yM = yy + l_font.TextHeight();
            x_DrawMultiAlignLabel(row, yM);
        }
        // advance to next row: 1 pixel between
        yy += bar_h + kAlignRowSpace;
    }

    // Draw selection
    if (IsSelected()) {
        m_Context->DrawSelection(rcm);
    }
}


void CAlignGlyph::x_DrawMultiAlignLabel(int row, TModelUnit y) const
{
    CGlRender& gl = GetGl();

    string label;
    CLabel::GetLabel(m_AlnMgr->GetSeqId(row), &label,
        CLabel::eContent, &m_Context->GetScope());
    if (m_AlnMgr->IsNegativeStrand(m_AlnMgr->GetAnchor()) !=
        m_AlnMgr->IsNegativeStrand(row)) {
        label += " (Neg)";
    }
    const CGlTextureFont& l_font = m_Config->m_LabelFont;

    TModelRange inrc = m_Context->IntersectVisible(this);
    TModelUnit max_label_w = m_Context->SeqToScreen(inrc.GetLength());
    TModelUnit label_w = l_font.TextWidth(label.c_str());
    if (label_w > max_label_w) {
        label_w = max_label_w;
        label = l_font.Truncate(label.c_str(), max_label_w);
    }
    bool top_label = m_Config->m_LabelPos == CAlignmentConfig::ePos_Above;
    TModelUnit xM = (inrc.GetFrom() + inrc.GetTo()) * 0.5;
    if ( !top_label ) {
        xM = inrc.GetFrom();
        if (inrc.GetFrom() > GetLeft()) {
            gl.ColorC(m_Config->m_BG);
            label_w = m_Context->ScreenToSeq(label_w);
            m_Context->DrawBackground( TModelRect(xM, y + 1.0,
                xM + label_w, y - l_font.TextHeight() - 1.0), 0);
        }
    }

    if (IsSelected()) {
        gl.ColorC(m_Context->GetSelLabelColor());
    } else {
        gl.ColorC(m_Config->m_Label);
    }
    m_Context->TextOut(&l_font, label.c_str(), xM, y, top_label);
}


void CAlignGlyph::x_DrawAlignRowSequence(TModelUnit center_y,
                                         int row,
                                         const CRgbaColor& c_fg,
                                         const CRgbaColor& c_seq,
                                         const CRgbaColor& c_mis) const
{
    CGlRender& gl = GetGl();

    int anchor = m_AlnMgr->GetAnchor();
    TSeqPos base_width_anchor = m_AlnMgr->GetBaseWidth(anchor);
    TSeqPos base_width_curr = m_AlnMgr->GetBaseWidth(row);
    TSeqRange anchor_range(
        m_AlnMgr->GetSeqStart(anchor) / base_width_anchor,
        m_AlnMgr->GetSeqStop(anchor) / base_width_anchor);
    TModelRange inrc = m_Context->IntersectVisible(anchor_range);
    anchor_range.Set((TSeqPos)inrc.GetFrom(), (TSeqPos)inrc.GetTo());
    if (anchor_range.Empty()) {
        return;
    }


    TSignedSeqPos aln_r_s = m_AlnMgr->GetAlnPosFromSeqPos(anchor,
            anchor_range.GetFrom() * base_width_anchor, IAlnExplorer::eRight);
    TSignedSeqPos aln_r_e = m_AlnMgr->GetAlnPosFromSeqPos(anchor,
            anchor_range.GetTo() * base_width_anchor, IAlnExplorer::eLeft);
    if (aln_r_s > aln_r_e) {
        swap(aln_r_s, aln_r_e);
    }
    TSignedSeqRange aln_range(aln_r_s, aln_r_e);

    if (aln_range.GetFrom() > 2) {
        aln_range.SetFrom(aln_range.GetFrom() - 3);
    }

    gl.ColorC(c_fg);

    bool neg_strand =
        m_AlnMgr->IsNegativeStrand(row) !=
        m_AlnMgr->IsNegativeStrand(anchor);

    auto_ptr<IAlnSegmentIterator> p_it(
        m_AlnMgr->CreateSegmentIterator(row,
        aln_range, IAlnSegmentIterator::eSkipGaps));

    const CBioseq_Handle& mhndl = m_AlnMgr->GetBioseqHandle(anchor);
    const CBioseq_Handle& hndl  = m_AlnMgr->GetBioseqHandle(row);

    bool ab = mhndl.IsAa();
    bool rb = hndl.IsAa();

    EAlignSpacing spacing(eOne2One);
    if (!ab && rb) spacing = eOne2Three;
    if (ab && !rb) spacing = eThree2One;

    for ( IAlnSegmentIterator& it(*p_it);  it;  ++it )  {
        const IAlnSegment& seg = *it;

        // We need to deal with some tricky genomic-to-protein alignments
        // described in both GB-2336 and GB-2324. Sometimes, the insertions
        // on the protein sequence row are one-(nucleotide)-base insertions.
        // That means two nucleiotide bases on genomic sewquence correspond
        // to one residue on the protein sequence at that insertion postion. 
        // Normally, it one-to-three or three-to-one relation. That introduces
        // all sort of troubles including:
        //    - showing correct sequence
        //    - positioning the sequence correctly,
        //    - retrieving the correct chunk of sequences for both protein
        //      translation and protein-to-genomic comparison 

        TSeqPos start = seg.GetRange().GetFrom();
        TSeqPos stop  = seg.GetRange().GetTo();
        TSeqPos sStart = (TModelUnit)m_AlnMgr->GetSeqPosFromAlnPos(anchor, seg.GetAlnRange().GetFrom());
        TSeqPos sStop  = (TModelUnit)m_AlnMgr->GetSeqPosFromAlnPos(anchor, seg.GetAlnRange().GetTo());

        TSeqPos frame_shift_f = start % base_width_curr;
        TSeqPos frame_shift_t = 2 - stop % base_width_curr;
        if (base_width_anchor == 3) {
            frame_shift_f = sStart % base_width_anchor;
            frame_shift_t = 2 - sStop % base_width_anchor;
        }
        if (sStart > sStop) {
            swap(sStart, sStop);
            swap(frame_shift_f, frame_shift_t);
        }

        stop /= base_width_curr;
        start /= base_width_curr;
        sStart /= base_width_anchor;
        sStop /= base_width_anchor;

        // Make sure nucleotide coordinate aligns protein codon's
        // starting/ending base. We need to do protein 
        if (spacing == eOne2Three) {
            // top/anchor sequence is nucleotide, and and aligned sequence
            // is protein. 
            sStop += frame_shift_t;
            sStart -= frame_shift_f;
        } else if (spacing == eThree2One) {
            // make sure it aligns codon's starting base
            stop += frame_shift_t;
            start -= frame_shift_f;
        }

        // check if sStop is larger than the sequence lenght
        sStop = min(sStop, mhndl.GetSeqVector().size());

        try {

            CSeqVector ms_vec =
                mhndl.GetSeqVector(CBioseq_Handle::eCoding_Iupac);
            CSeqVector s_vec =
                hndl.GetSeqVector(CBioseq_Handle::eCoding_Iupac);

            string seq, mseq, tmp_seq;
            s_vec.GetSeqData (start, stop + 1, tmp_seq);
            ms_vec.GetSeqData(sStart, sStop + 1, mseq);

            if (neg_strand) {
                if (rb) {
                    // protein sequence, just reverse it
                    CSeqManip::Reverse(tmp_seq, CSeqUtil::e_Iupacaa,
                        0, tmp_seq.length(), seq);
                } else {
                    // nucleotide sequence, get reverse and complement
                    CSeqManip::ReverseComplement(tmp_seq, CSeqUtil::e_Iupacna,
                        0, tmp_seq.length(), seq);
                }
            } else {
                seq = tmp_seq;
            }

            string translated_seq;
            switch (spacing) {
            case eOne2Three:
                {
                    string seq_candidate = mseq;
                    if (neg_strand) {
                        CSeqManip::ReverseComplement(mseq, CSeqUtil::e_Iupacna,
                            0, mseq.length(), seq_candidate);
                    }
                    CSeqTranslator::Translate(seq_candidate, translated_seq);
                    if (neg_strand) {
                        // don't use the same string as the input and output string.
                        CSeqManip::Reverse(translated_seq, CSeqUtil::e_Iupacaa,
                            0, translated_seq.length(), tmp_seq);
                        translated_seq = tmp_seq;
                    }
                }
                break;
            case eThree2One:
                {
                    CSeqTranslator::Translate(seq, translated_seq);
                }
                break;
            default:
                break;
            }

            char bases[2];
            bases[1] = '\0';
            TVPUnit font_w =
                (TVPUnit)(m_Config->m_SeqFont.TextWidth("G") * 0.5f);
            TModelUnit addon = 0.5 - m_Context->ScreenToSeq(font_w);

            TSeqPos start_base = frame_shift_f % base_width_anchor;
            TSeqPos stop_base = min(stop - start + 1 - frame_shift_t % base_width_anchor,
                (TSeqPos)seq.length());

            for (TSeqPos bp = start_base;  bp < stop_base;  bp++) {
                TModelUnit pos = 0;
                bool match = true;
                switch (spacing) {
                    case eOne2One:
                        pos = sStart + bp + addon;
                        match = seq[bp] == mseq[bp];
                        break;
                    case eOne2Three:
                        pos =  sStart + bp * 3 + addon + 1;
                        match = seq[bp] == translated_seq[bp];
                        break;
                    case eThree2One:
                        pos =  sStart + bp * kPWAlignPackLevel + addon - kPWAlignPackLevel;
                        match = mseq[bp/3] == translated_seq[bp/3];
                        break;
                    default:
                        break;
                }

                // Highlight mismatch?
                match ? gl.ColorC(c_seq) : gl.ColorC(c_mis);

                // se scoring if avaliable
                if (m_Score  &&  m_Score->HasScores()) {
                    if (m_Score->GetSupportedColorTypes() & IScoringMethod::fForeground) {
                        auto_ptr<IAlnScoreIterator> score_iter(m_Score->GetScoreIterator(row));
                        score_iter->MoveTo(bp);
                        IScoringMethod::TScore score = score_iter->GetAttr();
                        const CRgbaColor& cl = m_Score->GetColorForScore(score,
                            IScoringMethod::fForeground);
                        gl.ColorC(cl);  // use scoreing color
                    } else {
                        gl.ColorC(c_seq);  // use default color
                    }
                }

                bases[0] = seq[bp];

                TModelUnit fs = m_Config->m_SeqFont.TextHeight();
                m_Context->TextOut(&m_Config->m_SeqFont, bases,
                    pos, center_y + fs * 0.5, false);
            } // for TSeqPos
        }
        catch (CException&) {
            /// ignore exceptions - these are all objmgr related
        }
    } // for i = 0
}


// Scoring structures/functions start
/// represents a pixel with a score, if several intervals cover a pixel, the interval
/// with the maximum coverage will "own" the pixel (i.e. the pixel will be colored
/// according to the score of this interval)
/// This is used for non-averageable scores.
struct SScPix   {
    float   m_Score;    /// score assigned to the pixel
    float   m_MaxCoverage; /// fraction of the pixel corresponding to the m_Score
    bool    m_Gap; /// true if this is a gap interval

    bool operator==(const SScPix& other) const
    {
        return m_Score == other.m_Score  &&
            m_MaxCoverage == other.m_MaxCoverage  &&
            m_Gap == other.m_Gap;
    }
};


/// a pixel with the Max coverage (fraction) overrides all attributes
static void FSetPixMaxCoverage(SScPix& dst, const SScPix& src,
                               float fraction, bool)
{
    if(fraction > dst.m_MaxCoverage)   {
        dst.m_Score = src.m_Score;
        dst.m_Gap = src.m_Gap;
        dst.m_MaxCoverage = fraction;
    }
}


void CAlignGlyph::x_RenderNonAveragableScores(TModelUnit yy,
                                              TSignedSeqRange aln_range,
                                              TSeqRange seq_range,
                                              TModelUnit height,
                                              int row) const
{
    CGlRender& gl = GetGl();

    TModelUnit fromM = (TModelUnit)seq_range.GetFrom();
    TModelUnit toM   = (TModelUnit)seq_range.GetToOpen();
    int anchor = m_AlnMgr->GetAnchor();
    TSeqPos base_width_anchor = m_AlnMgr->GetBaseWidth(anchor);

    // coloration range in the viewport coordinate system
    TVPUnit scoreStart = m_Context->SeqToScreenX(fromM);
    TVPUnit scoreEnd = m_Context->SeqToScreenX(toM);

    typedef CRasterizer<SScPix> TR;
    TR rasterizer(scoreStart, scoreEnd, fromM, toM);
    SScPix pix = { 0.0f, 0.0f, false };

    auto_ptr<IAlnScoreIterator> score_iter(m_Score->GetScoreIterator(row));
    IAlnScoreIterator& it = *score_iter;
    auto_ptr<IAlnSegmentIterator> p_it(m_AlnMgr->CreateSegmentIterator(row,
        aln_range, IAlnSegmentIterator::eSkipGaps));

    IAlnSegmentIterator& aln_seg_it(*p_it);
    while (aln_seg_it  &&  it) {
        const IAlnSegment& seg = *aln_seg_it;
        const TSignedSeqRange& r = seg.GetAlnRange();

        TSignedSeqPos start1 = r.GetFrom();
        TSignedSeqPos stop1 = r.GetTo();
        TSignedSeqPos start2 = it.GetFrom();
        TSignedSeqPos stop2 = it.GetTo();

        while (it  &&  start1 > stop2) {
            ++it;
            start2 = it.GetFrom();
            stop2 = it.GetTo();
        }

        if (start1 > stop2) {
            break;
        }

        int from, to;
        from = std::max<int>(start1, start2);
        to   = std::min<int>(stop1,  stop2);

        TModelUnit act_from = m_AlnMgr->GetSeqPosFromAlnPos(anchor, from);
        act_from /= base_width_anchor;
        TModelUnit act_to = m_AlnMgr->GetSeqPosFromAlnPos(anchor, to);
        act_to /= base_width_anchor;
        if (act_from > act_to) {
            swap(act_from, act_to);
        }

        if (act_from <= act_to) {//! seg.IsGap()) {
            pix.m_Score = it.GetAttr();
            pix.m_Gap = seg.IsGap();
            rasterizer.AddInterval(act_from, act_to + 1.0 / base_width_anchor, pix, FSetPixMaxCoverage);
        }

        /// advance iterators
        if (to == stop1) {
            ++aln_seg_it;
        }
        if (to == stop2) {
            ++it;
        }
    }

    // Rendering
    const TR::TRaster& raster = rasterizer.GetRaster();
    static float eps = 0.0001f;
    double shift_x = scoreStart;

    gl.PushMatrix();
    gl.Scalef(m_Context->GetScale(), 1.0f, 1.0f);
    for (size_t i = 0;  i < raster.size(); ) {
        double start = i + shift_x;
        const SScPix& sc_pix = raster[i];
        while (++i < raster.size()  &&  raster[i] == sc_pix);

        if (sc_pix.m_MaxCoverage > eps)    {
            const CRgbaColor& color = m_Score->GetColorForScore(
                sc_pix.m_Score, IScoringMethod::fBackground);
            gl.ColorC(color);

            // vertical offset from the border of the alignment band
            // to the border of gap band
            int GapOffset = 2;
            int off_y = sc_pix.m_Gap ? GapOffset : 0;
            gl.Rectd(start, yy + off_y, i + shift_x, yy + height - off_y);
            x_DrawBorder(start, yy + off_y, i + shift_x, yy + height - off_y);
        }
    }
    gl.PopMatrix();
}


struct SScPix2   {
    float m_IntScore;
    float m_IntLen;
    float m_GapScore;
    float m_GapLen;
    float m_MinScore;
    bool  m_Init;

    bool    operator==(const SScPix2& other) const
    {
        return m_IntScore == other.m_IntScore  &&  m_IntLen == other.m_IntLen &&
        m_GapScore == other.m_GapScore  &&  m_GapLen == other.m_GapLen &&
        m_MinScore == other.m_MinScore;
    }
};


/// accumulate weighted scores
static void FSetPix2(SScPix2& dst, const SScPix2& src, float fraction, bool long_seg)
{
    dst.m_IntScore += src.m_IntScore * fraction;
    dst.m_IntLen += src.m_IntLen * fraction;
    dst.m_GapScore += src.m_GapScore * fraction;
    dst.m_GapLen += src.m_GapLen * fraction;

    if( ! long_seg)  {
        // only for small segments
        static const float eps = 0.00001f;
        if( ! dst.m_Init)   {
            if(src.m_IntLen > eps)  {
                dst.m_MinScore = src.m_IntScore;
                dst.m_Init = true;
            } else if(src.m_GapLen > eps)  {
                dst.m_MinScore = src.m_GapScore;
                dst.m_Init = true;
            }
        } else {
            if(src.m_IntLen > eps)  {
                dst.m_MinScore = min(dst.m_MinScore, src.m_IntScore);
            }
            if(src.m_GapLen > eps)  {
                dst.m_MinScore = min(dst.m_MinScore, src.m_GapScore);
            }
        }
    }
}


void CAlignGlyph::x_RenderAveragableScores(TModelUnit yy,
                                           TSignedSeqRange aln_range,
                                           TSeqRange seq_range,
                                           TModelUnit height,
                                           int row) const
{
    CGlRender& gl = GetGl();

    TModelUnit fromM = (TModelUnit)seq_range.GetFrom();
    TModelUnit toM   = (TModelUnit)seq_range.GetToOpen();
    int anchor = m_AlnMgr->GetAnchor();
    TSeqPos base_width_anchor = m_AlnMgr->GetBaseWidth(anchor);

    // coloration range in the viewport coordinate system
    TVPUnit scoreStart = m_Context->SeqToScreenX(fromM);
    TVPUnit scoreEnd = TVPUnit(ceil(m_Context->SeqToScreenXInModelUnit(toM)));

    typedef CRasterizer<SScPix2>  TR;
    TR rasterizer(scoreStart, scoreEnd, fromM, toM);

    auto_ptr<IAlnScoreIterator> score_it(m_Score->GetScoreIterator(row));
    IAlnScoreIterator& it = *score_it;
    auto_ptr<IAlnSegmentIterator> p_it(m_AlnMgr->CreateSegmentIterator(row,
        aln_range, IAlnSegmentIterator::eSkipGaps));

    IAlnSegmentIterator& aln_seg_it(*p_it);
    while (aln_seg_it  &&  it) {
        const IAlnSegment& seg = *aln_seg_it;
        const TSignedSeqRange& r = seg.GetAlnRange();

        TSignedSeqPos start1 = r.GetFrom();
        TSignedSeqPos stop1 = r.GetTo();
        TSignedSeqPos start2 = it.GetFrom();
        TSignedSeqPos stop2 = it.GetTo();

        while (it  &&  start1 > stop2) {
            ++it;
            if (it) {
                start2 = it.GetFrom();
                stop2 = it.GetTo();
            }
        }
        
        if (start1 > stop2) {
            break;
        }

        TSignedSeqPos from, to;
        from = std::max<int>(start1, start2);
        to   = std::min<int>(stop1,  stop2);

        SScPix2 pix = { 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, false };
        TModelUnit act_from = m_AlnMgr->GetSeqPosFromAlnPos(anchor, from);
        act_from /= base_width_anchor;
        TModelUnit act_to = m_AlnMgr->GetSeqPosFromAlnPos(anchor, to);
        act_to /= base_width_anchor;
        if (act_from > act_to) {
            swap(act_from, act_to);
        }

        //bool empty_seg = (seg.IsGap()  ||  act_from > (int)seq_range.GetTo()  ||
        //    act_to < (int)seq_range.GetFrom());
        //if (seg.GetType() & IAlnSegment::fAligned) {
        //    pix.m_IntLen = 1.0;
        //    pix.m_IntScore = it->GetAttr();
        //} else if (!empty_seg) {
        //    pix.m_GapLen = 1.0;
        //    pix.m_GapScore = it->GetAttr();
        //}

        if (act_from <= act_to) {
            pix.m_IntLen = 1.0;
            pix.m_IntScore = it.GetAttr();
            rasterizer.AddInterval(act_from, act_to + 1.0 / base_width_anchor, pix, FSetPix2);
        }

        /// advance iterators
        if (to == stop1) {
            ++aln_seg_it;
        }
        if (to == stop2) {
            ++it;
        }
    }

    // Rendering
    //gl.PolygonMode(GL_FRONT_AND_BACK, GL_FILL);

    const TR::TRaster& raster = rasterizer.GetRaster();

    double shift_x = scoreStart;
    static const float eps = 0.001f;

    gl.PushMatrix();
    gl.Scalef(m_Context->GetScale(), 1.0f, 1.0f);
    for (size_t i = 0;  i < raster.size(); )    {
        double start = i + shift_x;
        const SScPix2& sc_pix = raster[i];
        while (++i < raster.size()  &&  raster[i] == sc_pix);

        if (sc_pix.m_IntLen > eps) {
            float sc = sc_pix.m_IntScore / sc_pix.m_IntLen;
            const CRgbaColor& color =
                m_Score->GetColorForScore(sc, IScoringMethod::fBackground);
            gl.Color4d(color.GetRed(), color.GetGreen(), color.GetBlue(), 1.0);
            gl.Rectd(start, yy, i + shift_x, yy + height);
            x_DrawBorder(start, yy, i + shift_x, yy + height);
            if(sc_pix.m_Init)   {
                CRgbaColor color2 = m_Score->GetColorForScore(sc_pix.m_MinScore,
                    IScoringMethod::fBackground);
                double alpha = 1.0 - sc * 0.7;
                gl.Color4d(color2.GetRed(), color2.GetGreen(), color2.GetBlue(), alpha);
                gl.Rectd(start, yy, i + shift_x, yy + height);
                x_DrawBorder(start, yy, i + shift_x, yy + height);
            }
        }
        if (sc_pix.m_GapLen > eps) {
            float sc = sc_pix.m_GapScore / sc_pix.m_GapLen;
            const CRgbaColor& color =
                m_Score->GetColorForScore(sc, IScoringMethod::fBackground);
            double alpha = 0.25 + sc_pix.m_GapLen * 0.75;
            gl.Color4d(color.GetRed(), color.GetGreen(), color.GetBlue(), alpha);
            int GapOffset = 0;//height / 3;
            gl.Rectd(start, yy + GapOffset, i + shift_x + 1, yy + height - GapOffset);
            x_DrawBorder(start, yy + GapOffset, i + shift_x + 1, yy + height - GapOffset);

            if (sc_pix.m_Init)   {
                CRgbaColor color_m = m_Score->GetColorForScore(sc_pix.m_MinScore,
                    IScoringMethod::fBackground);
                color_m.SetAlpha(1.0f);
                gl.ColorC(color_m);
                gl.Rectd(start, yy, i + shift_x + 1, yy + height);
                x_DrawBorder(start, yy, i + shift_x + 1, yy + height);
            }
        }
    }
    gl.PopMatrix();
}


void CAlignGlyph::x_CalcIntervals(void)
{
    if (GetAlignMgr().GetNumRows() != 2) {
        return;
    }

    m_Intervals.clear();

    int anchor = m_AlnMgr->GetAnchor();
    TSeqPos base_width_anchor = m_AlnMgr->GetBaseWidth(anchor);
    TSeqRange r(
        m_AlnMgr->GetSeqStart(anchor) / base_width_anchor,
        m_AlnMgr->GetSeqStop(anchor) / base_width_anchor);

    if (r.Empty()) {
        LOG_POST(Error
            << "CAlignGlyph::x_CalcIntervals: zero alignment range!");
        return;
    }

    // optimization
    // Only use the total range if the size is small
    if (IsSimplified()) {
        m_Intervals.push_back(r);
        return;
    }

    for (int row = 0 ;  row < m_AlnMgr->GetNumRows();  ++row) {
        if (row == anchor) {
            continue;  // skip master seq
        }

        auto_ptr<IAlnSegmentIterator> p_it(m_AlnMgr->CreateSegmentIterator(row,
            TSignedSeqRange(m_AlnMgr->GetAlnStart(), m_AlnMgr->GetAlnStop()),
            IAlnSegmentIterator::eSkipGaps));

        for ( IAlnSegmentIterator& it(*p_it);  it;  ++it )  {
            const IAlnSegment& seg = *it;
            const TSignedSeqRange& aln_r = seg.GetAlnRange();
            TSeqPos start = m_AlnMgr->GetSeqPosFromAlnPos(anchor, aln_r.GetFrom());
            start /= base_width_anchor;
            TSeqPos stop = m_AlnMgr->GetSeqPosFromAlnPos(anchor, aln_r.GetTo());
            stop /= base_width_anchor;
            if (start > stop) {
                swap(start, stop);
            }
            m_Intervals.push_back( TSeqRange(start, stop) );
        }
    }
}


void CAlignGlyph::x_DrawBorder(float x1, float y1, float x2, float y2) const
{
    CGlRender& gl = GetGl();

    gl.PolygonMode(GL_FRONT_AND_BACK, GL_LINE);
    gl.Rectd(x2, y1, x2, y2);
    gl.PolygonMode(GL_FRONT_AND_BACK, GL_FILL);
}


bool CAlignGlyph::x_IsPolyA(const objects::CBioseq_Handle& bsh,
                            TSeqPos start, TSeqPos stop) const
{
    // we only apply testing on RNA sequences
    bool is_rna = false;    
    const CMolInfo* info = sequence::GetMolInfo(bsh);
    if (info) {
        if (info->GetBiomol() == CMolInfo::eBiomol_mRNA  ||
            info->GetBiomol() == CMolInfo::eBiomol_pre_RNA  ||
            info->GetBiomol() == CMolInfo::eBiomol_tRNA  ||
            info->GetBiomol() == CMolInfo::eBiomol_snRNA  ||
            info->GetBiomol() == CMolInfo::eBiomol_scRNA  ||
            info->GetBiomol() == CMolInfo::eBiomol_cRNA  ||
            info->GetBiomol() == CMolInfo::eBiomol_snoRNA  ||
            info->GetBiomol() == CMolInfo::eBiomol_ncRNA  ||
            info->GetBiomol() == CMolInfo::eBiomol_tmRNA) {
            is_rna = true;
        }
    }

    if ( !is_rna) {
        CConstRef<CSeq_id> seq_id;
        CSeq_id_Handle idh = sequence::GetId(*bsh.GetSeqId(),
            bsh.GetScope(), sequence::eGetId_Best);
        if (idh) {
            seq_id = idh.GetSeqId();
        } else {
            seq_id = bsh.GetSeqId();
        }

        CSeq_id::EAccessionInfo id_info = seq_id->IdentifyAccession();

        if ((id_info & CSeq_id::eAcc_division_mask) == CSeq_id::eAcc_est  ||
            id_info == CSeq_id::eAcc_refseq_mrna  ||
            id_info == CSeq_id::eAcc_refseq_mrna_predicted  ||
            id_info == CSeq_id::eAcc_gpipe_mrna) {
                is_rna = true;
        }

        if ( !is_rna  &&  id_info & CSeq_id::eAcc_ambig  &&
            info->GetBiomol() == CMolInfo::eBiomol_other) {
                is_rna = true;
        }
    }                                                                                                                              
   
    if (is_rna) {
        CSeqVector s_vec = bsh.GetSeqVector(CBioseq_Handle::eCoding_Iupac);
        string seq;
        s_vec.GetSeqData(start, stop + 1, seq);
        int a_cnt = 0;
        int t_cnt = 0;
        for (TSeqPos bp = 0;  bp != seq.length();  bp++) {
            if (seq[bp] == 'A') ++a_cnt;
            else if (seq[bp] == 'T') ++t_cnt;
        }
        if (a_cnt > (int)(seq.length() * 0.9)  ||
            t_cnt > (int)(seq.length() * 0.9)) {
            return true;
        }
    }

    return false;
}


bool CAlignGlyph::x_ShowLabel() const
{
    bool shown = false;
    if (m_Config->m_ShowLabel  &&
        m_Config->m_LabelPos != CAlignmentConfig::ePos_NoLabel) {
        if (m_Config->m_LabelPos == CAlignmentConfig::ePos_Above) {
            TModelRange r = m_Context->IntersectVisible(GetRange());
            if (m_Context->WillLabelFit(r)) {
                shown = true;
            }
        } else {
            shown = true;
        }
    }
    return shown;
}


bool CAlignGlyph::x_ShowScore() const
{
    return !m_Context->IsOverviewMode()  &&  m_Score  &&  m_Score->HasScores();
}


bool CAlignGlyph::x_ShowUnalignedTails(TModelUnit align_len) const
{
    // Only for pairwise alignments with a minimal screen size
    return m_AlnMgr->HasPolyATail() != IAlnGraphicDataSource::ePolyA_No  &&
        x_ShowScore()  &&  align_len > kTailLen  &&
        (m_Config->m_ShowUnalignedTails  ||
        (m_Config->m_ShowUnalignedTailsForTrans2GenomicAln  &&  m_SeqAlign->GetSegs().IsSpliced()))  &&
        x_IsPWAlign();
}



END_NCBI_SCOPE
