/*  $Id: alignment_job.cpp 27972 2013-05-03 20:34:20Z wuliangs $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Liangshou Wu
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>

#include <gui/widgets/seq_graphic/alignment_job.hpp>
#include <gui/widgets/seq_graphic/alignment_glyph.hpp>
#include <gui/widgets/seq_graphic/align_statistics_glyph.hpp>
#include <gui/widgets/seq_graphic/gene_model_group.hpp>
#include <gui/widgets/seq_graphic/mate_pair_glyph.hpp>
#include <gui/widgets/seq_graphic/alignment_smear_glyph.hpp>
#include <gui/widgets/seq_graphic/seqgraphic_utils.hpp>
#include <gui/widgets/seq_graphic/alnvec_graphic_ds.hpp>
#include <gui/widgets/seq_graphic/sparsealn_graphic_ds.hpp>
#include <gui/widgets/seq_graphic/denseg_graphic_ds.hpp>
#include <gui/widgets/seq_graphic/simple_graphic_ds.hpp>
#include <gui/widgets/seq_graphic/alignment_sorter.hpp>
#include <gui/widgets/seq_graphic/named_group.hpp>

#include <objtools/alnmgr/aln_converters.hpp>
#include <objtools/alnmgr/aln_builders.hpp>
#include <objtools/alnmgr/aln_container.hpp>
#include <objmgr/align_ci.hpp>
#include <objmgr/util/sequence.hpp>
#include <objmgr/util/feature.hpp>
#include <objmgr/seq_loc_mapper.hpp>
#include <objmgr/graph_ci.hpp>

#include <objects/general/Dbtag.hpp>
#include <objects/general/Object_id.hpp>
#include <objects/general/User_object.hpp>
#include <objects/seq/Annot_descr.hpp>
#include <objects/seq/Annotdesc.hpp>
#include <objects/seqres/Byte_graph.hpp>
#include <objects/seqres/Int_graph.hpp>

#include <math.h>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

/// always show individual alignments when zoom level is below this threshold.
/// Unit: bases per screen pixel (BPP)
static const double kSequenceZoomLevel = 1.0;

/// maximal number of alignments allowed to load to avoid potential memery 
/// and performance problem.
/// Unit: number of alignments
static const int kMaxAlignmentLoaded = 500000;

///////////////////////////////////////////////////////////////////////////////
/// CSGAlignmentJob
///////////////////////////////////////////////////////////////////////////////
typedef pair<CConstRef<IAlnGraphicDataSource>, int> TAlnDS_Anchor;
typedef vector<TAlnDS_Anchor> TAlnDataSources;

static
TAlnDataSources s_BuildAlignDataSource(CBioseq_Handle handle,
                                       const CSeq_align& align,
                                       bool sparseAln,
                                       TModelUnit window);


CSGAlignmentJob::CSGAlignmentJob(const string& desc,
                                 objects::CBioseq_Handle handle,
                                 const objects::SAnnotSelector& sel,
                                 const TSeqRange& range, TModelUnit window)
    : CSGAnnotJob(desc, handle, sel, range)
    , m_Window(window)
    , m_AlignLimit(-1)
    , m_SmearOverLimit(false)
    , m_LinkPair(false)
    , m_LoadCoverageGraph(true)
{
}


CSGAlignmentJob::CSGAlignmentJob(const string& desc,
                                 objects::CBioseq_Handle handle,
                                 const objects::SAnnotSelector& sel,
                                 const TSeqRange& range, TModelUnit window,
                                 int align_limit, bool smear_if_overlimit,
                                 bool link_pair)
    : CSGAnnotJob(desc, handle, sel, range)
    , m_Window(window)
    , m_AlignLimit(align_limit)
    , m_SmearOverLimit(smear_if_overlimit)
    , m_LinkPair(link_pair)
    , m_LoadCoverageGraph(false)
{}


CSGAlignmentJob::CSGAlignmentJob(const string& desc,
                                 objects::CBioseq_Handle handle,
                                 const objects::SAnnotSelector& sel,
                                 const TSeqRange& range,
                                 TModelUnit window,
                                 vector< CRef<CAlignGlyph> > aligns)
    : CSGAnnotJob(desc, handle, sel, range)
    , m_Aligns(aligns)
    , m_Window(window)
    , m_AlignLimit(-1)
    , m_SmearOverLimit(false)
    , m_LinkPair(false)
    , m_LoadCoverageGraph(false)
{}


void CSGAlignmentJob::GetAnnotNames(const objects::CBioseq_Handle& handle,
                                    const TSeqRange& range,
                                    objects::SAnnotSelector& sel,
                                    TAnnotNameTitleMap& names)
{
    CAlign_CI aln_iter(handle, range, sel);
    ITERATE (CAlign_CI::TAnnotNames, iter, aln_iter.GetAnnotNames()) {
        if (iter->IsNamed()) {
            if (iter->GetName().find("@@") == string::npos) {
                names.insert(TAnnotNameTitleMap::value_type(iter->GetName(), ""));
            }
        } else {
            names.insert(TAnnotNameTitleMap::value_type(
                CSeqUtils::GetUnnamedAnnot(), ""));
        }
    }
}


void CSGAlignmentJob::SetSortBy(const string& sort_by)
{
    if (sort_by.empty()) return;

    size_t found = sort_by.find_first_of('|');
    string sorter_type = NStr::TruncateSpaces(sort_by.substr(0, found));
    string sort_str = kEmptyStr;
    if (found != string::npos) {
        sort_str = sort_by.substr(found + 1);
    }

    if (NStr::EqualNocase(sorter_type, CAlignStrandSorter::GetID())) {
        m_Sorter.Reset(new CAlignStrandSorter(sort_str));
    }
}


IAlnExplorer::EAlignType
CSGAlignmentJob::GetAlignType(const objects::CBioseq_Handle& handle,
                              objects::SAnnotSelector& sel,
                              bool& has_quality_map)
{
    sel.SetMaxSize(1);
    sel.SetCollectNames(false);
    CAlign_CI aln_iter(handle, TSeqRange::GetWhole(), sel);

    IAlnExplorer::EAlignType type = IAlnExplorer::fInvalid;
    if (aln_iter) {
        const CSeq_align& align = *aln_iter;
        int num_row = 0;
        try {
            num_row = align.CheckNumRows();
        } catch (CException&) {
        }

        if (num_row < 2) return type;

        CScope& scope = handle.GetScope();

        // check align type
        for (int row = 0;  row < num_row;  ++row) {
            IAlnExplorer::EAlignType this_type = IAlnExplorer::fMixed;
            CBioseq_Handle row_handle = scope.GetBioseqHandle(align.GetSeq_id(row));
            if ( !row_handle ) continue;

            switch (row_handle.GetBioseqCore()->GetInst().GetMol())
            {
            case objects::CSeq_inst::eMol_dna:
            case objects::CSeq_inst::eMol_rna:
            case objects::CSeq_inst::eMol_na:
                this_type = IAlnExplorer::fDNA;
                break;

            case objects::CSeq_inst::eMol_aa:
                this_type = IAlnExplorer::fProtein;
                break;

            default:
                break;
            }

            if (this_type == IAlnExplorer::fMixed) {
                type = this_type;
                break;
            }

            if (row == 0) {
                type = this_type;
            } else if (this_type != type) {
                type = IAlnExplorer::fMixed;
                break;
            }
        }

        // check if there is quality graph
        for (int row = num_row - 1;  row >= 0;  --row) {
            SAnnotSelector g_sel(CSeqUtils::GetAnnotSelector(CSeq_annot::TData::e_Graph));
            g_sel.AddUnnamedAnnots();
            g_sel.SetCollectNames();
            sel.SetMaxSize(1);
            CBioseq_Handle row_handle = scope.GetBioseqHandle(align.GetSeq_id(row));
            if (row_handle) {
                CGraph_CI graph_iter(row_handle, align.GetSeqRange(row), sel);
                if (graph_iter  &&  graph_iter.GetSize() == 1) {
                    has_quality_map = true;
                    break;
                }
            }
        }
    }

    return type;
}


static void s_ObjectIdToStr(const CObject_id& id, string& id_str)
{
    if (id.IsStr()) {
        id_str = id.GetStr();
    } else {
        id_str = NStr::NumericToString(id.GetId());
    }
}


static bool
s_CheckTraceMateAlign(bool match_id,
                      bool is_bam_align,
                      const CSeq_align& align,
                      string& ti,
                      string& mate_ti,
                      const string& expect_mate_ti = "")
{
    //
    // specific requirements for mate pair alignments:
    //

    // must be a pairwise dense-seg
    if ( !align.GetSegs().IsDenseg()  ||
        align.GetSegs().GetDenseg().GetIds().size() != 2) {
            return false;
    }
    ti = "";
    mate_ti = "";

    // there are two possible cases:
    //  1. trace assemblies that store mate pair info in seq-align::score
    //  2. cSRA/Bam short reads that store mate pair info in seq-align::ext

    // first, let's check short reads case
    if (is_bam_align) {
        if (align.IsSetExt()) {
            // exclude secondary alignments
            ITERATE (CSeq_align::TExt, iter, align.GetExt()) {
                if ((*iter)->GetType().IsStr()  &&
                    (*iter)->GetType().GetStr() == "Secondary") {
                    return false;
                }
            }
        }
        
        int i = 1;
        // assumming the second row is the short read
        // try second row first.
        while (ti.empty()  &&  i >=0) {
            const CSeq_id& seq_id = align.GetSeq_id(i);
            if (seq_id.IsLocal()) {
                s_ObjectIdToStr(seq_id.GetLocal(), ti);
            } else if (seq_id.IsGeneral()) {
                s_ObjectIdToStr(seq_id.GetGeneral().GetTag(), ti);
            }
            --i;
        }
        
        if (ti.empty()) {
            return false;
        }
    }

    if (match_id) {
        size_t len = ti.size();
        mate_ti = ti.substr(0, len - 1) + (ti[len-1] == '1' ? "2" : "1");
    } else if (align.IsSetExt()  &&  !ti.empty()) {
        ITERATE (CSeq_align::TExt, iter, align.GetExt()) {
            if ((*iter)->GetType().IsStr()  &&
                (*iter)->GetType().GetStr() == "Secondary") {
                return false;
            }
            if ((*iter)->GetType().IsStr()  &&
                (*iter)->GetType().GetStr() == "Mate read") {
                    CConstRef<CUser_field> user_field = (*iter)->GetFieldRef("lcl|");
                    if (user_field) {
                        if (user_field->GetData().IsStr()) {
                            mate_ti = user_field->GetData().GetStr();
                        } else if (user_field->GetData().IsInt()) {
                            mate_ti = NStr::NumericToString(user_field->GetData().GetInt());
                        }
                    }
            }
        }
    }
     
    if (is_bam_align) {
        if (mate_ti.empty()) {
            // check if it is possible to generate mate_ti based on ti
            // The ti naming pattern should be: xxxxx.[12]
            size_t len = ti.size();
            const char last_second_c = ti[len-2];
            const char last_c = ti[len-1];
            if (last_second_c == '.'  &&  (last_c == '1' ||  last_c == '2')) {
                mate_ti = ti.substr(0, len - 1) + (last_c == '1' ? "2" : "1");
            }
        }

        if (mate_ti.empty()) {
            return false;
        } else {
            return true;
        }
    }

    // then, check the IDs for a trace ID
    ITERATE (CDense_seg::TIds, iter,
        align.GetSegs().GetDenseg().GetIds()) {
            const CSeq_id& id = **iter;
            if (id.IsGeneral()  &&
                (id.GetGeneral().GetDb() == "ti"  ||
                id.GetGeneral().GetDb() == "TRACE")) {
                    if (id.GetGeneral().GetTag().IsId()) {
                        ti = NStr::IntToString(id.GetGeneral().GetTag().GetId());
                    } else {
                        ti = id.GetGeneral().GetTag().GetStr();
                    }
                    break;
            }
    }

    if (ti.empty()) {
        return false;
    }

    // must have a score field named 'matepair_ti'
    int mate_ti_int = 0;
    if ( !align.GetNamedScore("matepair ti", mate_ti_int)  &&
         !align.GetNamedScore("bad matepair ti", mate_ti_int)) {
            return false;
    }

    mate_ti = NStr::SizetToString((unsigned int)mate_ti_int);
    if (!expect_mate_ti.empty()  &&  mate_ti != expect_mate_ti) {
        return false;
    }

    return true;
}


IAppJob::EJobState CSGAlignmentJob::x_Execute()
{
    if ( !m_Aligns.empty() ) {
        CBatchJobResult* result = new CBatchJobResult();
        m_Result.Reset(result);
        result->m_Token = m_Token;
        NON_CONST_ITERATE (vector< CRef<CAlignGlyph> >, iter, m_Aligns) {
            if (IsCanceled()) {
                return IAppJob::eCanceled;
            }
            CRef<CSGJobResult> single_res = x_LoadAlignmentFeats(**iter);
            if (single_res) {
                result->m_Results.push_back(single_res);
            }
        }

        return eCompleted;
    } else if (m_LoadCoverageGraph) {
        return x_GetCoverageGraph();
    }
    
    return x_LoadAlignments();
}


IAppJob::EJobState CSGAlignmentJob::x_LoadAlignments()
{
    CSGJobResult* result = new CSGJobResult();
    m_Result.Reset(result);
    result->m_Token = m_Token;

    // maximal number alignment allowed to load
    if (m_AlignLimit > kMaxAlignmentLoaded) {
        m_AlignLimit = kMaxAlignmentLoaded;
    }

    // upper limit used to control maximal number of alignments to load
    int upper_limit = m_AlignLimit;

    if (m_SmearOverLimit) {
        // In this case, we need to smear alignments if total number
        // is over the limit.  So we have to load all alignment. This
        // is for regular alignments only.
        upper_limit = kMaxAlignmentLoaded;
    } else if (m_AlignLimit > -1  &&  m_Window <= kSequenceZoomLevel) {
        // If alignments won't be smeared even the number of alignments
        // is over the limit (!m_SmearOverLimit), and it is not in the 
        // smear mode (m_AlignLimit != -1), and, and the zoom level
        // is at sequence level (m_Window <= kSequenceZoomLevel), we
        // shall load all alignments and show them. 
        upper_limit = kMaxAlignmentLoaded;

        // increase the limit to accept any number of alignment
        // available in the given range.
        m_AlignLimit = upper_limit + 1;
    } 

    if (upper_limit <= 0) {
        return eCompleted;
    }

    m_Sel.SetMaxSize(upper_limit + 1);
    CAlign_CI align_iter(m_Handle, m_Range, m_Sel);

    int obj_size = (int)align_iter.GetSize();
    SetTaskTotal(obj_size);
    SetTaskCompleted(0);

    CSeqGlyph::TObjects& objs = result->m_ObjectList;
    if (obj_size > m_AlignLimit) {
        if (m_SmearOverLimit) {
            // smear bar is requested when over limit
            return x_GetAlignSmear(objs, align_iter);
        } 
        // Otherwise, do thing. Maybe, pileup display is required in
        // this case which will be requested and done in other place
        // ELSE { RETURN NOTHING }
    } else {
        // return all alignments, this includes the case when it
        // is at sequence level (m_AlignLimit has been adjusted)
        return x_GetAlignments(objs, align_iter, true);
    }
    SetTaskCompleted(obj_size);

    return eCompleted;
}


CRef<CSGJobResult>
CSGAlignmentJob::x_LoadAlignmentFeats(CAlignGlyph& align)
{
    CRef<CSGJobResult> result;
    CSeqGlyph::TObjects tmp;

    const IAlnGraphicDataSource& aln_mgr = align.GetAlignMgr();
    // aligned seq-id
    const CSeq_id& aligned_seq =
        aln_mgr.GetSeqId(aln_mgr.GetAnchor() == 0 ? 1 : 0);
    // anchored seq-loc
    const CSeq_loc& loc     = align.GetLocation();

    try {
        // map visible range to product feature
        CSeq_loc_Mapper seq_range_mapper(align.GetAlignment(),
            (size_t)(1 - align.GetOrigAnchor()), &GetScope());
        CRef<CSeq_loc> tmp_loc(new CSeq_loc());
        tmp_loc->SetInt().SetFrom(m_Range.GetFrom());
        tmp_loc->SetInt().SetTo  (m_Range.GetTo());
        tmp_loc->SetId(*loc.GetId());

        // our new location for feature iterator
        CRef<CSeq_loc> seq_loc(new CSeq_loc());
        CRef<CSeq_loc> mapped_loc = seq_range_mapper.Map(tmp_loc.GetObject());
        
        CSeq_loc::TRange mapped_range = mapped_loc->GetTotalRange();
        seq_loc->SetInt().SetFrom(mapped_range.GetFrom());
        seq_loc->SetInt().SetTo(mapped_range.GetTo());
        seq_loc->SetId(aligned_seq);

        // use newly created location to get the features
        CFeat_CI feat_iter(GetScope(), seq_loc.GetObject(), m_Sel);

        if (feat_iter.GetSize() > 0) {
            //::wxWakeUpIdle();
            CSeq_loc_Mapper mapper(
                align.GetAlignment(), (size_t)align.GetOrigAnchor(), &GetScope());

            CLinkedFeature::TLinkedFeats main_features;
            CSeqGlyph::TObjects other_features;
            for ( ;  feat_iter;  ++feat_iter) {
                if (IsCanceled()) {
                    return result;
                }
                const CMappedFeat& feat = *feat_iter;
                int subtype = feat.GetFeatSubtype();
                int type = feat.GetFeatType();
                if (CSGUtils::IsMainFeature(type, subtype)) {
                    CRef<CLinkedFeature> fref( new CLinkedFeature(feat) );
                    main_features.push_back(fref);
                } else {
                    CRef<CSeqGlyph> g_glyph = x_CreateFeatGlyph(mapper, feat);
                    if (g_glyph) {
                        other_features.push_back(g_glyph);
                    }
                }
            }

            // linking the features
            if ( !CSeqUtils::LinkFeatures(main_features,
                feature::CFeatTree::eFeatId_by_type, this)  ||
                !x_CreateGeneModels(mapper, main_features, tmp) ) {
                return result;
            }
            
            std::copy(other_features.begin(), other_features.end(),
                back_inserter(tmp));
        }
    }
    catch (CAnnotMapperException&) {
        /// ignore errors from location mapping
        return result;
    }

    if ( !tmp.empty() ) {
        result.Reset(new CSGJobResult());
        result->m_ObjectList.swap(tmp);
        result->m_Token = m_Token;
        result->m_Owner = CRef<CSeqGlyph>(&align);
    }
    return result;
}


CRef<CSeqGlyph>
CSGAlignmentJob::x_CreateFeatGlyph(CSeq_loc_Mapper& mapper,
                                   const CMappedFeat& mapped_feat) const
{
    CRef<CSeqGlyph> glyph;
    CConstRef<CSeq_loc> mapped_loc = mapper.Map(mapped_feat.GetLocation());
    if (mapped_loc->Which() != CSeq_loc::e_Null  &&
        mapped_loc->Which() != CSeq_loc::e_Empty) {
        CFeatGlyph* feat = new CFeatGlyph(mapped_feat, *mapped_loc);
        glyph.Reset(feat);
        CFeatGlyph::TIntervals& intervals = feat->SetIntervals();
        const CSeq_loc& loc = feat->GetLocation(); 
        CSeq_loc_CI iter(loc);
        for ( ;  iter;  ++iter) {
            intervals.push_back(iter.GetRange());
        }
    }
    return glyph;
}


bool 
CSGAlignmentJob::x_CreateGeneModels(CSeq_loc_Mapper& mapper,
                                    CLinkedFeature::TLinkedFeats& feats,
                                    CSeqGlyph::TObjects& objs) const
{
    // Create glyph for each feature, and put the linked features
    // into the right group

    CRef<CLayoutGroup> global_exon_group;
    NON_CONST_ITERATE (CLinkedFeature::TLinkedFeats, iter, feats) {
        if (IsCanceled()) return false;

        const CMappedFeat& mapped_feat = (*iter)->GetMappedFeature();
        CRef<CSeqGlyph> glyph = x_CreateFeatGlyph(mapper, mapped_feat);
        if ( !glyph ) continue;

        CSeqFeatData::ESubtype subtype =
            (*iter)->GetFeature().GetData().GetSubtype();
        if ((*iter)->GetChildren().empty()) {
            if (subtype == CSeqFeatData::eSubtype_exon) {
                if ( !global_exon_group ) {
                    global_exon_group.Reset(new CLayoutGroup);
                }
                global_exon_group->PushBack(glyph);
            } else {
                objs.push_back(glyph);
            }
            continue;
        }

        CRef<CLayoutGroup> exon_group;
        CRef<CLayoutGroup> gene_group;
        gene_group.Reset(new CGeneGroup());
        objs.push_back(CRef<CSeqGlyph>(gene_group.GetPointer()));
        gene_group->PushBack(glyph);
        NON_CONST_ITERATE (CLinkedFeature::TLinkedFeats, c_iter, (*iter)->GetChildren()) {
            const CMappedFeat& c_mapped_feat = (*c_iter)->GetMappedFeature();
            CRef<CSeqGlyph> c_glyph = x_CreateFeatGlyph(mapper, c_mapped_feat);
            if ( !c_glyph ) continue;

            subtype = (*c_iter)->GetFeature().GetData().GetSubtype();
            if (subtype == CSeqFeatData::eSubtype_exon) {
                if ( !exon_group ) {
                    exon_group.Reset(new CLayoutGroup);
                }
                exon_group->PushBack(c_glyph);
            } else {
                gene_group->PushBack(c_glyph);
                NON_CONST_ITERATE (CLinkedFeature::TLinkedFeats, g_iter, (*c_iter)->GetChildren()) {
                    const CMappedFeat& g_mapped_feat = (*g_iter)->GetMappedFeature();
                    CRef<CSeqGlyph> g_glyph = x_CreateFeatGlyph(mapper, g_mapped_feat);
                    if (g_glyph) {
                        gene_group->PushBack(g_glyph);
                    }
                }
            }
        }
        if (exon_group) {
            if (exon_group->GetChildren().size() == 1) {
                gene_group->Append(exon_group->SetChildren());
            } else {
                gene_group->PushBack(exon_group.GetPointer());
            }
        }
    }
    if (global_exon_group) {
        if (objs.size() == 1  &&  dynamic_cast<CGeneGroup*>(objs.front().GetPointer())) {
            CGeneGroup* group = dynamic_cast<CGeneGroup*>(objs.front().GetPointer());
            if (global_exon_group->GetChildren().size() == 1) {
                group->Append(global_exon_group->SetChildren());
            } else {
                group->PushBack(global_exon_group.GetPointer());
            }
        } else {
            if (global_exon_group->GetChildren().size() == 1) {
                objs.push_back(global_exon_group->SetChildren().front());
            } else {
                objs.push_back(CRef<CSeqGlyph>(global_exon_group.GetPointer()));
            }
        }
    }

    return true;
}


IAppJob::EJobState
CSGAlignmentJob::x_GetAlignSmear(CSeqGlyph::TObjects& objs,
                                 CAlign_CI& align_iter)
{
    if (align_iter.GetSize() == 0) return eCompleted;

    CRef<CAlignSmearGlyph> smear_glyph;

    const CSeq_annot& seq_annot_first =
        *align_iter.GetSeq_align_Handle().GetAnnot().GetCompleteSeq_annot();
    if (CAlignmentSmear::SeparateStrands(seq_annot_first)) {
        // positive strand
        smear_glyph.Reset(new CAlignSmearGlyph(m_Handle, m_Range.GetFrom(),
        m_Range.GetTo(), m_Window, CAlignmentSmear::eSmearStrand_Pos));
        if (smear_glyph) {
            CAlignmentSmear& smear = smear_glyph->GetAlignSmear();
            if ( !smear.AddAlignments(align_iter, this) ) {
                return eCanceled;
            }
            objs.push_back(CRef<CSeqGlyph>(smear_glyph.GetPointer()));
        }

        // negative strand
        smear_glyph.Reset(new CAlignSmearGlyph(m_Handle, m_Range.GetFrom(),
        m_Range.GetTo(), m_Window, CAlignmentSmear::eSmearStrand_Neg));
        if (smear_glyph) {
            CAlignmentSmear& smear = smear_glyph->GetAlignSmear();
            if ( !smear.AddAlignments(align_iter, this)) {
                return eCanceled;
            }
            objs.push_back(CRef<CSeqGlyph>(smear_glyph.GetPointer()));
        }
    } else {
        smear_glyph.Reset(new CAlignSmearGlyph(m_Handle, m_Range.GetFrom(),
        m_Range.GetTo(), m_Window, CAlignmentSmear::eSmearStrand_Both));
        if (smear_glyph) {
            CAlignmentSmear& smear = smear_glyph->GetAlignSmear();
            if ( !smear.AddAlignments(align_iter, this) ) {
                return eCanceled;
            }
            objs.push_back(CRef<CSeqGlyph>(smear_glyph.GetPointer()));
        }
    }

    return eCompleted;
}


IAppJob::EJobState
CSGAlignmentJob::x_GetAlignments(CSeqGlyph::TObjects& objs,
                                 CAlign_CI& align_iter,
                                 bool sparseAln)
{
    if (align_iter.GetSize() == 0) return eCompleted;

    SetTaskName("Creating alignment glyphs...");
    typedef pair<CRef<CAlignGlyph> , string> TMatedAlign;
    typedef map<string, TMatedAlign> TMatedAlignments;
    TMatedAlignments mated_aligns;

    bool match_id = false;
    CConstRef<CSeq_annot> aln_annot =
        align_iter.GetSeq_align_Handle().GetAnnot().GetCompleteSeq_annot();
    if (aln_annot  &&  aln_annot->CanGetDesc()) {
        ITERATE (objects::CAnnot_descr::Tdata, descrIter, aln_annot->GetDesc().Get()) {
            if ((*descrIter)->IsUser()  &&
                (*descrIter)->GetUser().GetType().IsStr()  &&
                (*descrIter)->GetUser().GetType().GetStr() == "Mate read") {
                    CConstRef<CUser_field> user_field =
                        (*descrIter)->GetUser().GetFieldRef("Match by local Seq-id");
                    if (user_field  &&  user_field->GetData().IsBool()  &&
                        user_field->GetData().GetBool()) {
                            match_id = true;
                            break;
                    }
            }
        }
    }

    CSGAlignmentJob::EAlignFlags flag = CSGAlignmentJob::fAlign_Default;
    if (m_LinkPair) {
        flag = CSGAlignmentJob::fAlign_LinkMatePairs;
    }
    for ( ;  align_iter;  ++align_iter) {
        if (IsCanceled()) {
            return eCanceled;
        }

        const CSeq_align_Handle& align_handle =
            align_iter.GetSeq_align_Handle();
        const CSeq_align& align = *align_iter;

        TAlnDataSources aln_datasources;
        try {
            // try using both direction
            aln_datasources = s_BuildAlignDataSource(m_Handle, align, sparseAln, m_Window);
        } catch (CException& e) {
            // log errors
            LOG_POST(Warning << "error in GetAlignments(): " << e.GetMsg());
        }

        for (size_t i = 0; i < aln_datasources.size(); ++i) {
            int anchor = aln_datasources[i].second;
            CConstRef<IAlnGraphicDataSource>& aln_datasource =
                aln_datasources[i].first;

            //
            // check to see if this is a mate pair alignment
            //
            string ti = "";
            string mate_ti = "";
            if ( s_CheckTraceMateAlign(match_id, match_id, align, ti, mate_ti) ) {
                CRef<CAlignGlyph> pwal
                    (new CAlignGlyph(aln_datasource, align_handle, align, anchor));
                if (flag & fAlign_LinkMatePairs) {
                    // we will process mate mair alignments in a separate pass
                    // after we complete all other alignments
                    TMatedAlign mp(pwal, mate_ti);
                    mated_aligns[ti] = mp;
                } else {
                    objs.push_back(CRef<CSeqGlyph>(pwal.GetPointer()));
                }
            } else {
                // "normal" alignment
                CAlignGlyph* aln_glyph =
                    new CAlignGlyph(aln_datasource, align_handle, align, anchor);
                CRef<CSeqGlyph> ref(aln_glyph);
                objs.push_back(ref);
            }
        }
        AddTaskCompleted(1);
    }

    //
    // final pass - connect our mate pair alignments, if we can
    //
    NON_CONST_ITERATE (TMatedAlignments, iter, mated_aligns) {
        if (IsCanceled()) {
            return eCanceled;
        }

        if ( !iter->second.first ) {
            continue;
        }

        CAlignGlyph& first_mate = *iter->second.first;
        const CSeq_align& align = first_mate.GetAlignment();
        const CSeq_align_Handle& orig_align = first_mate.GetOrigAlignment();
        const string& ti = iter->first;
        string mate_ti = iter->second.second;

        //
        // verify that we have a mated alignment for this trace alignment
        //
        CRef<CAlignGlyph> second_mate;
        TMatedAlignments::iterator mate_iter = mated_aligns.find(mate_ti);
        if (mate_iter == mated_aligns.end()) {
            //
            // NOT FOUND
            // try iterating. We do this only for trace assembly, not for
            // other cases like cSRA/BAM short reads.
            //
            CSeq_id id("gnl|ti|" + mate_ti);
            CBioseq_Handle h = m_Handle.GetScope().GetBioseqHandle(id);
            if (h) {
                SAnnotSelector sel =
                    CSeqUtils::GetAnnotSelector(CSeq_annot::TData::e_Align);
                CAlign_CI second_mate_iter(h, sel);
                for ( ;  second_mate_iter;  ++second_mate_iter) {
                    string temp = "";
                    if ( !s_CheckTraceMateAlign(false, false, *second_mate_iter,
                        temp, mate_ti, ti) ) {
                        continue;
                    }

                    try {
                        // create CAlnGraphicDataSource
                        TAlnDataSources aln_datasources =
                            s_BuildAlignDataSource(m_Handle, align, sparseAln, m_Window);
                        if ( !aln_datasources.empty() ) {
                            second_mate.Reset(new CAlignGlyph(
                                aln_datasources.front().first, orig_align, align,
                                first_mate.GetOrigAnchor()));
                        }
                    } catch (CException& e){
                        LOG_POST(Warning << "error in GetAlignments(): " << e.GetMsg());
                    }
                    break;
                }
            }

            if ( !second_mate ) {
                //LOG_POST(Warning << "failed to find mate pair for ti = " << ti);
                // no mate available, so treat it as a normal alignment
                objs.push_back(CRef<CSeqGlyph>(&first_mate));
                continue;
            }
        } else if ( !mate_iter->second.first ) {
            // already processed
            continue;
        } else {
            second_mate = mate_iter->second.first;
            mate_iter->second.first.Reset();
        }

        CMatePairGlyph::TAlignList als;
        als.push_back(CRef<CAlignGlyph>(&first_mate));
        als.push_back(second_mate);
        CRef<CSeqGlyph> obj(new CMatePairGlyph(als));
        objs.push_back(obj);
    }

    if (m_Sorter  &&  mated_aligns.empty()) {
        // we don't sort alignments if they are paired and linked together
        typedef map<int, CSeqGlyph::TObjects> TAlignGroupMap;
        TAlignGroupMap sorted_aligns;
        NON_CONST_ITERATE (CSeqGlyph::TObjects, iter, objs) {
            CAlignGlyph* align_glyph = dynamic_cast<CAlignGlyph*>(iter->GetPointer());
            _ASSERT(align_glyph);
            int idx = m_Sorter->GetGroupIdx(align_glyph->GetAlignment());
            if (idx > -1) {
                sorted_aligns[idx].push_back(*iter);
            }
        }
        if (sorted_aligns.size() > 1) {
            objs.clear();
            NON_CONST_ITERATE (TAlignGroupMap, g_iter, sorted_aligns) {
                CRef<CNamedGroup> named_group(
                    new CNamedGroup(m_Sorter->GroupIdxToName(g_iter->first)));
                named_group->Set(g_iter->second);
                objs.push_back(CRef<CSeqGlyph>(named_group.GetPointer()));
            }
        }
    }

    SetTaskCompleted((int)align_iter.GetSize());
    return eCompleted;
}


IAppJob::EJobState
CSGAlignmentJob::x_GetCoverageGraph()
{
    CSGJobResult* result = new CSGJobResult();
    m_Result.Reset(result);
    SetTaskName("Loading coverate graph...");

    CSeqGlyph::TObjects glyphs;
    EJobState state = x_CreateHistFromGraph(glyphs, m_Sel, m_Window, false);
    if (state == eCompleted) {
        result->m_ObjectList.swap(glyphs);
    }

    result->m_Token = m_Token;

    return state;
}


///////////////////////////////////////////////////////////////////////////////
/// CSGAlignScoringJob
///////////////////////////////////////////////////////////////////////////////
CSGAlignScoringJob::CSGAlignScoringJob(const string& desc,
                                       const TAlnScoreMap& scores)
    : CSeqGraphicJob(desc)
    , m_AlnScoreMap(scores)
{
    SetTaskName("Calculating alignment scores...");
}


IAppJob::EJobState CSGAlignScoringJob::x_Execute()
{
    CJobResultBase* result = new CJobResultBase();
    m_Result.Reset(result);

    SetTaskTotal((int)m_AlnScoreMap.size());
    SetTaskCompleted(0);

    NON_CONST_ITERATE (TAlnScoreMap, iter, m_AlnScoreMap) {
        if (IsCanceled()) {
            return IAppJob::eCanceled;
        }
        CIRef<ISGAlnScore> score = iter->second;
        if ( !score->HasScores() ) {
            score->CalculateScores();
        }
        AddTaskCompleted(1); 
    }
    result->m_Token = m_Token;
    return IAppJob::eCompleted;
}


///////////////////////////////////////////////////////////////////////////////
/// CSGAlignStatJob
CSGAlignStatJob::CSGAlignStatJob(const string& desc,
                                 objects::CBioseq_Handle handle,
                                 const objects::SAnnotSelector& aln_sel,
                                 const objects::SAnnotSelector& graph_sel,
                                 const TSeqRange& range, 
                                 const TAlnMgrVec& aligns,
                                 TModelUnit scale)
    : CSGAnnotJob(desc, handle, aln_sel, range)
    , m_GraphSel(graph_sel)
    , m_Aligns(aligns)
    , m_Window(scale)
{
    SetTaskName("Loading alignment pileup ...");
}


/// layout style to layout display name
typedef pair<string, CAlnStatConfig::EStatType> TGraphTitle;
static const TGraphTitle s_GraphTitles[] = {
    TGraphTitle("Number of A bases",  CAlnStatConfig::eStat_A),
    TGraphTitle("Number of C bases",  CAlnStatConfig::eStat_C),
    TGraphTitle("Number of G bases",  CAlnStatConfig::eStat_G),
    TGraphTitle("Number of T bases",  CAlnStatConfig::eStat_T),
    TGraphTitle("Number of inserts",  CAlnStatConfig::eStat_Gap),
    TGraphTitle("Number of matches",  CAlnStatConfig::eStat_Match),
};

typedef CStaticArrayMap<string, CAlnStatConfig::EStatType> TGraphTitleMap;
DEFINE_STATIC_ARRAY_MAP(TGraphTitleMap, sm_GraphTitleMap, s_GraphTitles);

size_t s_GraphTitleToIndex(const string& title)
{
    TGraphTitleMap::const_iterator iter = sm_GraphTitleMap.find(title);
    if (iter != sm_GraphTitleMap.end()) {
        return (size_t)iter->second;
    }
    return -1;
}


const string& s_GraphIndexToTitle(size_t idx)
{
    TGraphTitleMap::const_iterator iter;
    for (iter = sm_GraphTitleMap.begin();  iter != sm_GraphTitleMap.end();  ++iter) {
        if ((size_t)iter->second == idx) {
            return iter->first;
        }
    }
    _ASSERT(false); // shouldn't get to here
    return kEmptyStr;
}


IAppJob::EJobState CSGAlignStatJob::x_Execute()
{
    typedef IAlnGraphicDataSource::TNumrow TNumrow;
    TSeqPos start = m_Range.GetFrom();
    TSeqPos stop = m_Range.GetToOpen();

    // prepare the job result object
    CSGJobResult* result = new CSGJobResult();
    m_Result.Reset(result);
    result->m_Token = m_Token;

    // Original pileup graphs
    // Graphs could be from CGraph_CI that can't be stored in CRef (maybe)
    typedef vector<const CSeq_graph*> TGraphs;
    typedef vector<TGraphs> TGraphVec;
    // The graphs are stored in the order of CAlnStatConfig::EStatType
    // which is A, G, T, C, Gap, and Match
    TGraphVec graphs;

    bool sra_pileup = false;
    bool bam_pileup = false;
    // try to see if there is pileup graphs available
    SetTaskName("Loading pileup graphs ...");
    CGraph_CI graph_iter(m_Handle, m_Range, m_GraphSel);

    if (graph_iter.GetSize() >= 5) {
        graphs.resize(6);
        set<size_t> graph_tracker;
        for (;  graph_iter;  ++graph_iter) {
            size_t g_idx = s_GraphTitleToIndex(graph_iter->GetTitle());
            if (g_idx != (size_t)-1) {
                graphs[g_idx].push_back(&graph_iter->GetMappedGraph());
                graph_tracker.insert(g_idx);
            }
        }

        if (graph_tracker.size() == 6) {
            sra_pileup = true;
        } else if (graph_tracker.size() == 5) {
            // bam pileup doesn't contain match graph
            _ASSERT(graph_tracker.count(6) == 0);
            bam_pileup = true;
        } else {
            graphs.clear();
        }
    }

    // graphs holder created from the loaded seq-aligns.
    vector< CRef<CSeq_graph> > graph_holder;

    if (graphs.empty()) {
        // no pileup graph is available, try computing pileup on the fly

        if (m_Aligns.empty()) {
            // make sure the alignments are loaded
            SetTaskName("Loading alignments ...");
            EJobState status = x_LoadAlignments();
            if ( status != eCompleted) {
                return status;
            }
        }
        if (m_Aligns.empty()) {
            return eCompleted;
        }
   
        graphs.resize(6);

        SetTaskName("Creating pileup graphs ...");
        SetTaskTotal((int)m_Aligns.size());
        SetTaskCompleted(0);

        // initialize the top sequence to generate statistics for every base
        vector<unsigned char> ref_bases;
        size_t ref_len = 0;
        {{
            string seq_str;
            CSeqVector seq_vector =
                m_Handle.GetSeqVector(CBioseq_Handle::eCoding_Iupac);
            seq_vector.GetSeqData(start, stop, seq_str);
            ref_len = seq_str.length();
            if (ref_len < stop - start) {
                stop = start + ref_len;
            }
            ref_bases.resize(ref_len);
            for (size_t i = 0; i < ref_len; ++i) {
                ref_bases[i] = (unsigned char)CAlnStatConfig::GetRefSeqIdex(seq_str[i]);
            }
        }}

        CRef<CSeq_loc> loc(new CSeq_loc());
        loc->SetInt().SetFrom(start);
        loc->SetInt().SetTo(stop - 1);
        CConstRef<CSeq_id> seq_id = m_Handle.GetSeqId();
        loc->SetId(*seq_id);
        for (int i = 0; i <= CAlnStatConfig::eStat_Match; ++i) {
            CRef<CSeq_graph> g(new CSeq_graph);
            g->SetTitle(s_GraphIndexToTitle((size_t)i));
            g->SetLoc(*loc);
            g->SetGraph().SetInt().SetValues().resize(ref_bases.size(), 0);
            graph_holder.push_back(g);
            graphs[i].push_back(g.GetPointer());
        }

        string aln_seq; // variable used in nested loops down here, brought up decl. for better heap alloc.performance

        ITERATE (TAlnMgrVec, iter, m_Aligns) {
            if (IsCanceled()) {
                return eCanceled;
            }
            AddTaskCompleted(1); 

            const IAlnGraphicDataSource& align = **iter;

            if (align.GetAlignType() != IAlnExplorer::fDNA) {
                // If it is not a DNA-to-DNA alignment, skip it
                continue;
            }

            TNumrow row_n = align.GetNumRows();
            TNumrow anchor = align.GetAnchor();
            TSeqPos anchor_start = align.GetSeqStart(anchor);
            TSeqPos anchor_stop = align.GetSeqStop(anchor);
            anchor_start = max(anchor_start, start);
            anchor_stop = min(anchor_stop, stop - 1);
            if ((*iter)->IsSimplified()) {
                // No need to look at match/mismatch.
                // Only count overall coverage.
                vector<int>& match_vals =
                    graph_holder[CAlnStatConfig::eStat_Match]->SetGraph().SetInt().SetValues();

                if (anchor_start < start)  anchor_start = start;
                if (anchor_stop > stop) anchor_stop = stop;
                while (anchor_start <= anchor_stop) {
                    size_t curr_pos = anchor_start - start;
                    vector<int>& ref_vals =
                        graph_holder[ref_bases[curr_pos]]->SetGraph().SetInt().SetValues();
                    ref_vals[curr_pos]++;
                    match_vals[curr_pos]++;
                    anchor_start++;
                }
                continue;
            }

            TSignedSeqRange aln_range(
                align.GetAlnPosFromSeqPos(anchor, anchor_start, IAlnExplorer::eRight),
                align.GetAlnPosFromSeqPos(anchor, anchor_stop, IAlnExplorer::eLeft));

            for (TNumrow r = 0;  r < row_n; r++)  {
                if (r == anchor) continue;

                auto_ptr<IAlnSegmentIterator> p_it(
                    align.CreateSegmentIterator(r,
                    aln_range, IAlnSegmentIterator::eAllSegments));

                for (IAlnSegmentIterator& it = *p_it;  it;  ++it)  {
                    const IAlnSegment& seg = *it;
        
                    if ((seg.GetType() & IAlnSegment::fIndel  &&
                        !seg.GetRange().Empty())  ||
                        seg.GetType() & IAlnSegment::fUnaligned  ||
                        seg.GetType() & IAlnSegment::fGap) {
                            // ignore the inserts
                            continue;
                    }

                    const IAlnSegment::TSignedRange& curr_aln_r = seg.GetAlnRange();
                    TSignedSeqPos curr_aln_start = curr_aln_r.GetFrom();
                    TSeqPos seg_len = (TSeqPos)curr_aln_r.GetLength();
                    TSeqPos curr_seq_from =
                        (TSeqPos)align.GetSeqPosFromAlnPos(anchor, curr_aln_start);
                    TSeqPos off = 0;
                    size_t pos = 0;
                    if (curr_seq_from < start) {
                        pos = start - curr_seq_from;
                    } else {
                        off = curr_seq_from - start;
                    }
                    size_t curr_pos = off;
                    if (seg.GetType() & IAlnSegment::fAligned) {
                        aln_seq.clear();
                        align.GetAlnSeqString(aln_seq, r, curr_aln_r);
                    }
                    while (pos < seg_len  &&  curr_pos < ref_len) {
                        if (IsCanceled()) {
                            return eCanceled;
                        }

                        if (seg.GetType() & IAlnSegment::fAligned) {
                            if (pos < aln_seq.size()) {
                                int base_type = CAlnStatConfig::GetRefSeqIdex(aln_seq[pos]);
                                vector<int>& vals =
                                    graph_holder[base_type]->SetGraph().SetInt().SetValues();
                                vals[curr_pos] += 1;
                                if (base_type != CAlnStatConfig::eStat_Gap  &&
                                    base_type == ref_bases[curr_pos]) {
                                        vector<int>& m_vals =
                                            graph_holder[CAlnStatConfig::eStat_Match]->SetGraph().SetInt().SetValues();
                                        m_vals[curr_pos] += 1;
                                }
                            } else {
                                // It shouldn't get to here
                                _ASSERT(false);
                            }
                        } else {
                            // gap
                            vector<int>& vals =
                                graph_holder[CAlnStatConfig::eStat_Gap]->SetGraph().SetInt().SetValues();
                            vals[curr_pos] += 1;
                        }
                        ++pos;
                        ++curr_pos;
                    }
                }
            }
        }
    }


    // We don't store more than one set of statistics for each screen pixel.
    // If the zoom scale is larger than one nucleotide per pixel, then we 
    // need to squeeze several sets of statistics (one per base) into one (pixel)

    size_t val_num = (size_t)(stop - start);

    // crate aln stat glyph based the pileup graphs
    CRef<CAlnStatGlyph> stat_glyph(new CAlnStatGlyph(start, m_Window));
    CAlnStatGlyph::TStatVec& stats = stat_glyph->GetStatVec();

    // resize the stats to the actual size needed
    if (m_Window < 1.0) m_Window = 1.0;
    size_t pix_num = (size_t)ceil(val_num / m_Window);
    stats.resize(pix_num);

    // reference sequence
    string ref_seq_str;

    size_t num_grp = graphs[0].size();
    for (size_t grp_id = 0; grp_id < num_grp; ++grp_id) {

        // get the real graph range
        TSeqRange range = graphs[0][grp_id]->GetLoc().GetTotalRange();
        size_t curr_start = start; 
        if (range.GetFrom() > curr_start) {
            curr_start = range.GetFrom();
        }

        size_t curr_idx = curr_start - range.GetFrom();
        size_t num_vals = range.GetLength();
        int graph_off = (int)range.GetFrom() - (int)start;

        double curr_pos = (double)(curr_start - start);
        size_t stat_idx = (size_t)floor(curr_pos / m_Window);

        for (; stat_idx < pix_num  &&  curr_idx < num_vals; ++stat_idx) {
            curr_pos += m_Window;
            size_t end_idx = (size_t)(floor(curr_pos + 0.5) - graph_off);
            if (end_idx >= num_vals) {
                end_idx = num_vals - 1;
            }
            // choose one set of stats from [start_idx, end_idx) to 
            // represent current pixel (i).  The criteria are
            // like this: max(max(gap(n), mismatch(n))), where n belongs
            // to one of [start_idx, end_idx). In case all of them have
            // the same max(gap, mismatch), then the first one is used.

            // -1 can make sure it copies the first one always
            int max_count = -1;
            ref_seq_str.clear();
            if (bam_pileup) {
                CSeqVector seq_vector =
                    m_Handle.GetSeqVector(CBioseq_Handle::eCoding_Iupac);
                seq_vector.GetSeqData(range.GetFrom() + curr_idx, range.GetFrom() + end_idx, ref_seq_str);
            }

            for (size_t ref_idx = 0; curr_idx < end_idx; ++curr_idx, ++ref_idx) {
                // collect pileup at base start_idx
                CAlnStatGlyph::SStatStruct tmp_stat;
                for (size_t g_idx = 0; g_idx < 6; ++g_idx) {
                    if (bam_pileup  &&  g_idx == (size_t)CAlnStatConfig::eStat_Match) {
                        int matches = 0;
                        if (ref_idx < ref_seq_str.size()) {
                            int base_type = CAlnStatConfig::GetRefSeqIdex(ref_seq_str[ref_idx]);
                            if (base_type != CAlnStatConfig::eStat_Gap) {
                                matches = tmp_stat.m_Data[base_type];
                            }
                        }
                        tmp_stat.m_Data[CAlnStatConfig::eStat_Match] = matches;
                    } else {
                        const CSeq_graph::TGraph& graph = graphs[g_idx][grp_id]->GetGraph();
                        if (graph.IsInt()) {
                            const vector<int>& vals = graph.GetInt().GetValues();
                            tmp_stat.m_Data[g_idx] = vals[curr_idx];
                        } else { // must be byte
                            const vector<char>& vals = graph.GetByte().GetValues();
                            tmp_stat.m_Data[g_idx] = (int)(unsigned char)(vals[curr_idx]);
                        }
                    }
                }
                int agtc = tmp_stat.m_Data[CAlnStatConfig::eStat_A] +
                    tmp_stat.m_Data[CAlnStatConfig::eStat_G] +
                    tmp_stat.m_Data[CAlnStatConfig::eStat_T] +
                    tmp_stat.m_Data[CAlnStatConfig::eStat_C];
                if (sra_pileup) {
                    tmp_stat.m_Data[CAlnStatConfig::eStat_Mismatch] = agtc;
                    tmp_stat.m_Data[CAlnStatConfig::eStat_Total] = agtc +
                        tmp_stat.m_Data[CAlnStatConfig::eStat_Gap] +
                        tmp_stat.m_Data[CAlnStatConfig::eStat_Match];
                } else {
                    tmp_stat.m_Data[CAlnStatConfig::eStat_Mismatch] =
                        agtc - tmp_stat.m_Data[CAlnStatConfig::eStat_Match];
                    tmp_stat.m_Data[CAlnStatConfig::eStat_Total] =
                        agtc + tmp_stat.m_Data[CAlnStatConfig::eStat_Gap];
                }

                int curr_max = max(tmp_stat.m_Data[CAlnStatConfig::eStat_Gap],
                    tmp_stat.m_Data[CAlnStatConfig::eStat_Mismatch]);
                if (max_count < curr_max) {
                    stats[stat_idx] = tmp_stat;
                    max_count = curr_max;
                }
            }
        }
    }

    // store stat glyph in the job result
    result->m_ObjectList.push_back(CRef<CSeqGlyph>(stat_glyph.GetPointer()));

    return eCompleted;
}


IAppJob::EJobState CSGAlignStatJob::x_LoadAlignments()
{
    SetTaskName("Load alignments ...");
    m_Sel.SetMaxSize(kMaxAlignmentLoaded);
    CAlign_CI align_iter(m_Handle, m_Range, m_Sel);

    if (align_iter.GetSize() == 0) return eCompleted;

    SetTaskName("Create alignment data source ...");
    SetTaskTotal((int)align_iter.GetSize());
    SetTaskCompleted(0);

    for ( ;  align_iter;  ++align_iter) {
        if (IsCanceled()) {
            return eCanceled;
        }

        TAlnDataSources aln_datasources;
        try {
            // try using both direction
            aln_datasources = s_BuildAlignDataSource(m_Handle, *align_iter, true, 0.0);
        } catch (CException& e) {
            // log errors
            LOG_POST(Warning << "error in GetAlignments(): " << e.GetMsg());
        }

        for (size_t i = 0; i < aln_datasources.size(); ++i) {
            CConstRef<IAlnGraphicDataSource>& aln_datasource =
                aln_datasources[i].first;
            m_Aligns.push_back(aln_datasource);
        }
        AddTaskCompleted(1);
    }

    SetTaskCompleted((int)align_iter.GetSize());
    return eCompleted;

}

static
TAlnDataSources s_BuildAlignDataSource(CBioseq_Handle handle,
                                       const CSeq_align& align,
                                       bool sparseAln,
                                       TModelUnit window)
{
    TAlnDataSources data_sources;
    // get the anchor rows which are the referent sequence
    vector<CSeq_align::TDim> anchors;
    CSeq_align::TDim num_row = align.CheckNumRows();
    if (num_row == 0) {
        // empty alignment
        NCBI_THROW(CException, eUnknown, "Get empty alignment!");
    }
    CSeq_align::TDim row = 0;
    for (row = 0;  row < num_row;  ++row) {
        CSeq_id_Handle idh = sequence::GetId(align.GetSeq_id(row),
            handle.GetScope(), sequence::eGetId_Best);
        if ( !idh ) {
            idh = sequence::GetId(align.GetSeq_id(row),
                handle.GetScope(), sequence::eGetId_Canonical);
        }
        if ( handle.IsSynonym(idh) ) {
            anchors.push_back(row);
        }
    }


    if (anchors.empty()) {
        /// try a more aggressive matching approach
        for (size_t level = 0;  level <= 5  &&  anchors.empty();  ++level) {
            for (row = 0;  row < num_row;  ++row) {
                CSeq_id_Handle idh = sequence::GetId(align.GetSeq_id(row),
                    handle.GetScope(), sequence::eGetId_Best);
                if ( !idh ) {
                    idh = sequence::GetId(align.GetSeq_id(row),
                        handle.GetScope(), sequence::eGetId_Canonical);
                }
                if ( handle.ContainsSegment(idh, level) ) {
                    anchors.push_back(row);
                }
            }
        }
    }

    _ASSERT( !anchors.empty() );
    if (anchors.empty()) {
        NCBI_THROW(CException, eUnknown,
            "Can find the anchor sequence in the alignment!");
    }

    bool self_alignment = false;
    if (num_row  ==  anchors.size()) {
        // This is a self-aligned alignment (same sequence for all rows).
        // Request no merge (ePreserveRows) when creating CAnchoredAln 
        self_alignment = true;

        // check if all aligned rows match with the exactly same range
        bool exact_self_aligned = true;
        CRange<TSeqPos> aligned_r = align.GetSeqRange(0);
        for (CSeq_align::TDim row = 1; row < num_row; ++row) {
            if (aligned_r != align.GetSeqRange(row)) {
                exact_self_aligned = false;
                break;
            }
        }

        // Use the first row only to avoid duplication if it is an exact
        // self-aligned alignment (JIRA GB-1289). Otherwise show the alignment
        // multiple times, once for each row as the anchor sequence range
        // (JIRA GB-2349)
        if (exact_self_aligned) {
            anchors.clear();
            anchors.push_back(0);
        }
    }

    if (anchors.size() == 1  ||  self_alignment) {
        ITERATE (vector<CSeq_align::TDim>, row_iter, anchors) {
            CSeq_align::TDim anchor_row = *row_iter;
            try {
                if (window > 0.0  &&  align.GetSeqRange(anchor_row).GetLength() / window < 5.0) {
                    CConstRef<IAlnGraphicDataSource>
                        ds(new CSimpleGraphicDataSource(align, handle.GetScope(), anchor_row));
                    data_sources.push_back(TAlnDS_Anchor(ds, anchor_row));
                } else if (align.CheckNumRows() == 2) {
                    if (align.GetSegs().IsDenseg()  &&
                        align.GetSegs().GetDenseg().CheckNumSegs() > 0  &&
                        !align.GetSegs().GetDenseg().IsSetWidths()) { // we don't deal with mixed denseg

                            CConstRef<IAlnGraphicDataSource>
                                ds(new CDensegGraphicDataSource(align, handle.GetScope(), anchor_row));
                            data_sources.push_back(TAlnDS_Anchor(ds, anchor_row));

                    } else if (align.GetSegs().IsStd()  &&  align.GetSegs().GetStd().size() == 1) {
                        const vector< CRef< CSeq_loc > >& locs =
                            align.GetSegs().GetStd().front()->GetLoc();
                        _ASSERT(locs.size() == 2);
                        if (locs[0]->IsInt()  &&  locs[1]->IsInt()) {
                            TSeqPos len1 = locs[0]->GetTotalRange().GetLength();
                            TSeqPos len2 =  locs[1]->GetTotalRange().GetLength();
                            if (len1 != len2  &&  len1 != 3 * len2  &&  len2 != 3 * len1) {
                                // For std-seg with two uneven aligned segments, if it is not
                                // a protein-to-genomic alignment, use a simplified alignemnt
                                // manager to handle this special case.
                                CSimpleGraphicDataSource* simple_ds =
                                    new CSimpleGraphicDataSource(align, handle.GetScope(), anchor_row);
                                CConstRef<IAlnGraphicDataSource> ds(simple_ds);
                                simple_ds->SetRegularity(false);
                                data_sources.push_back(TAlnDS_Anchor(ds, anchor_row));
                            }
                        }
                    }
                }
            } catch (CException&) {
                // ignore
            }
        }
    }

    if ( !data_sources.empty() ) {
        return data_sources;
    }

    ITERATE (vector<CSeq_align::TDim>, row_iter, anchors) {
        CSeq_align::TDim anchor = *row_iter;
        // generate an alignment manager for this alignment (either CAlnVec or CSparseAln)
        if (sparseAln) {
            typedef vector<const objects::CSeq_align*> TAlnVector;
            typedef CAlnIdMap<TAlnVector, TIdExtract> TAlnIdMap;
            typedef CAlnStats<TAlnIdMap> TAlnStats;

            CAlnContainer aln_container;
            aln_container.insert(align);
            CAlnSeqIdsExtract<CAlnSeqId> id_extract;
            TAlnIdMap aln_id_map(id_extract, aln_container.size());
            int num_align = 0;
            ITERATE(CAlnContainer, aln_it, aln_container) {
                try {
                    // Create a vector of seq-ids for the seq-align
                    aln_id_map.push_back(**aln_it);
                    ++num_align;
                } catch (CAlnException e) {
                    // Skipping this alignment
                }
            }
            // Create align statistics object
            TAlnStats aln_stats(aln_id_map);
            // Create user options
            CAlnUserOptions aln_user_options;
            aln_user_options.m_Direction = CAlnUserOptions::eBothDirections;
            TAnchoredAlnVec anchored_aln_vec;

            // Explicitly specify anchor sequence seq_id 
            //aln_user_options.SetAnchorId(aln_id_map[0][anchor]);
            //CreateAnchoredAlnVec(aln_stats, anchored_aln_vec, aln_user_options);

            // Use anchor sequence row number stored in seq-align
            for (size_t aln_idx = 0; aln_idx < aln_stats.GetAlnCount(); ++aln_idx) {
                CRef<CAnchoredAln> anchored_aln =
                    CreateAnchoredAlnFromAln(aln_stats, aln_idx, aln_user_options, anchor);
                if ( anchored_aln ) {
                    anchored_aln_vec.push_back(anchored_aln);
                    // Calc scores
                    for (CSeq_align::TDim row = 0; row < anchored_aln->GetDim(); ++row) {
                        ITERATE(CPairwiseAln, rng_it, *anchored_aln->GetPairwiseAlns()[row]) {
                            anchored_aln->SetScore() += rng_it->GetLength();
                        }
                    }
                    anchored_aln->SetScore() /= anchored_aln->GetDim();
                }
            }

            if (self_alignment) {
                aln_user_options.m_MergeAlgo = CAlnUserOptions::ePreserveRows;
            } else {
                aln_user_options.m_MergeAlgo = CAlnUserOptions::eDefaultMergeAlgo;
            }
            CRef<CAnchoredAln> out_aln(new CAnchoredAln);
            BuildAln(anchored_aln_vec, *out_aln, aln_user_options);
            anchored_aln_vec.clear();
            anchored_aln_vec.push_back(out_aln);

            NON_CONST_ITERATE (TAnchoredAlnVec, a_iter, anchored_aln_vec) {
                // Build a sparse align
                CConstRef<CSparseAln> sparse_aln(
                    new CSparseAln(**a_iter, handle.GetScope()));

                _ASSERT(sparse_aln->GetDim() > 1);
                if (sparse_aln->GetDim() < 2) {
                    // something is wrong, skip this one
                    LOG_POST(Warning << "CSGAlignmentJob::s_BuildAlignDataSource(): "
                        << "Something is wrong with the alignment. The created CSparseAln has only one row.");
                    continue;
                }
                if (sparse_aln->GetBaseWidth(0) == -1) {
                    // use a simplified alignemnt manager to handle this special case.
                    CSimpleGraphicDataSource* simple_ds =
                        new CSimpleGraphicDataSource(align, handle.GetScope(), anchor);
                    CConstRef<IAlnGraphicDataSource> ds(simple_ds);
                    simple_ds->SetRegularity(false);
                    data_sources.push_back(TAlnDS_Anchor(ds, anchor));
                } else {
                    // create CSparseGraphicDataSource
                    CConstRef<IAlnGraphicDataSource>
                        ds(new CSparseAlnGraphicDataSource(sparse_aln));
                    data_sources.push_back(TAlnDS_Anchor(ds, anchor));
                }
            }
        } else {
            CRef<CAlnVec> aln_mgr;
            if (align.GetSegs().IsDenseg()) {
                aln_mgr.Reset(new CAlnVec(align.GetSegs().GetDenseg(),
                    handle.GetScope()));
            } else {
                CAlnMix mix(handle.GetScope());
                mix.Add(align);
                mix.Merge(CAlnMix::fGapJoin);

                aln_mgr.Reset(new CAlnVec(mix.GetDenseg(), handle.GetScope()));
            }
            // anchor the alignment on the referent sequence
            aln_mgr->SetAnchor(anchor);

            // create CAlnVecGraphicDataSource
            CConstRef<IAlnGraphicDataSource>
                ds(new CAlnVecGraphicDataSource(*aln_mgr));
            data_sources.push_back(TAlnDS_Anchor(ds, anchor));
        }
    }

    //// preload aligned sequences
    //try {
    //    for (int row = 0; row < aln_datasource->GetNumRows(); ++row) {
    //        aln_datasource->GetBioseqHandle(row);
    //    }
    //} catch (CException& e) {
    //    // ignore the error on resolving the seq-id
    //    LOG_POST(Error << e.GetMsg());
    //}

    return data_sources;
}


END_NCBI_SCOPE
