/*  $Id: alignment_track.cpp 27877 2013-04-17 18:49:16Z wuliangs $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Liangshou Wu
 *
 * File Description:
 *  There are some fairly complicated logics on determining how to display
 *  the alignments based on several different factors:
 *    - Incoming data source: regular alignments vs cSRA/BAM alignments
 *    - Alignment display (layout): Smear (packed), AdaptiveSeq, and so on,
 *      see the comment on CAlignmentTrack::ELayout for more details.
 *    - Zoom level: high level vs pileup display level vs low level
 *    - Pileup display setting: enabled vs disabled
 *    - Alignment type: DNA vs protein vs mixed
 *
 *  Here are preknowledge:
 *    - cSRA/BAM files have precomputed coverage graphs
 *    - cSRA/Bam loader is able to deliver alignment pileup quickly
 *    - For regular alignments, pileup calculation is expensive and slow.
 *    - There is a (upper) limit determining what zoom level is appropriate
 *      to show alignment pileup (for performance consideration).
 *    - Alignment display setting determines when (what zoom level) to show
 *      individual alignments.
 * 
 *  And here are the requirements:
 *    1. At high level, the alignments shall be shown in a compact form
 *       that is either coverage graph (for cSRA data) or smear bar (for
 *       regular alignemnts). The high level is defined when the zoom level
 *       reaches to a zoom level,
 *       a. that is over the predefined pileup display level for cSRA data, or
 *       b. at which there are more alignments than a predefined number 
 *          determined by the alignment display option within the visible
 *          rnage for regular alignments.
 *    2. For regular alignments with piplup display enabled and for all cSRA
 *       data regardless of piplup display enabled or not, if the zoom level
 *       is between the predefined pileup zoom level and the zoom level
 *       determined by alignment display setting, the pileup display, instead
 *       of smear bar or coverage graph, shall be shown.
 *    3. At low level, all individual alignments shall be shown except for
 *       one alignment display setting which is 'Smear'. With this setting,
 *       alignments shall be shown in a packed form similar to high level mode.
 *       The low level is defined when zoom level reaches to either:
 *         a. sequence level, or
 *         b. any zoom level at which there are less alignments than the
 *            predefined number determined by the alignment display setting.
 *       At this level, if pipleup display is enabled, both individual
 *       alignments and pileup shall be shown.
 *
 */

#include <ncbi_pch.hpp>

#include <gui/widgets/seq_graphic/alignment_track.hpp>
#include <gui/widgets/seq_graphic/seqgraphic_ds_manager.hpp>
#include <gui/widgets/seq_graphic/seqgraphic_graph_ds.hpp>
#include <gui/widgets/seq_graphic/gene_model_track.hpp>
#include <gui/widgets/seq_graphic/gene_model_group.hpp>
#include <gui/widgets/seq_graphic/feature_track.hpp>
#include <gui/widgets/seq_graphic/seqgraphic_utils.hpp>
#include <gui/widgets/seq_graphic/config_utils.hpp>
#include <gui/widgets/seq_graphic/layout_track_proxy.hpp>
#include <gui/widgets/seq_graphic/rendering_ctx.hpp>
#include <gui/widgets/seq_graphic/alignment_glyph.hpp>
#include <gui/widgets/seq_graphic/mate_pair_glyph.hpp>
#include <gui/widgets/seq_graphic/alignment_smear_glyph.hpp>
#include <gui/widgets/seq_graphic/histogram_glyph.hpp>
#include <gui/widgets/seq_graphic/coord_mapper.hpp>
#include <gui/widgets/seq_graphic/named_group.hpp>
#include <gui/widgets/seq_graphic/alignment_sorter.hpp>
#include <gui/widgets/seq_graphic/alignment_quality_score.hpp>
#include <gui/widgets/wx/wx_utils.hpp>

#include <gui/widgets/seq/feature_check_dialog.hpp>

#include <gui/utils/system_path.hpp>
#include <gui/objutils/utils.hpp>
#include <gui/objutils/label.hpp>
#include <gui/objutils/registry.hpp>

#include <objmgr/util/sequence.hpp>

#include <wx/event.h>
#include <wx/menu.h>

#include "wx_aln_stat_dlg.hpp"

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

// the maximal number of object allowed to show without smearing
//static const int kMaxAlnShown = 250000;

// menu item id for projected features
static const size_t kProjectedFeatID = 10000;
// content menu item base id
static const size_t kContentBaseID = 10001;
// potential maximal menu item id range
// assumming there will be less than 100 menu items
static const int kMaxID = 100;

class CAlnContentEvtHandler : public wxEvtHandler
{
public:
    CAlnContentEvtHandler(CAlignmentTrack* track)
        : m_Track( track )
    {}

    void OnContextMenu( wxContextMenuEvent& anEvent );

private:
    /// @name event handlers.
    /// @{
    /// on toggle a conent menu item.
    void x_OnToggleContent(wxCommandEvent& event)
    {
        int track_order = event.GetId() - kContentBaseID;
        m_Track->OnToggleContent(track_order);
    }

    /// on change the projected feature list.
    void x_OnChangeProjectedFeats(wxCommandEvent& /*event*/)
    { m_Track->OnChangeProjectedFeats(); }

    /// @}

private:
    CAlignmentTrack* m_Track;

    DECLARE_EVENT_TABLE()
};


BEGIN_EVENT_TABLE(CAlnContentEvtHandler, wxEvtHandler)
    EVT_MENU(kProjectedFeatID, CAlnContentEvtHandler::x_OnChangeProjectedFeats)
    EVT_MENU_RANGE(kContentBaseID, kContentBaseID + kMaxID, CAlnContentEvtHandler::x_OnToggleContent)
END_EVENT_TABLE()


///////////////////////////////////////////////////////////////////////////////
///   CAlignmentTrack

// static method for converting layout setting to smear bar threshold
static int s_LayoutToCutoff(CAlignmentTrack::ELayout layout)
{
    int cut_off = 1000;
    switch (layout) {
        case CAlignmentTrack::eLayout_Smear:
            cut_off = -1;
            break;
        case CAlignmentTrack::eLayout_AdaptiveSeq:
            cut_off = 0;
            break;
        case CAlignmentTrack::eLayout_Adaptive200:
            cut_off = 200;
            break;
        case CAlignmentTrack::eLayout_Adaptive5000:
            cut_off = 5000;
            break;
        case CAlignmentTrack::eLayout_Adaptive20000:
            cut_off = 20000;
            break;
        case CAlignmentTrack::eLayout_Adaptive100000:
            cut_off = 100000;
            break;
        case CAlignmentTrack::eLayout_Adaptive250000:
        case CAlignmentTrack::eLayout_ExpandedByPos:
            cut_off = 250000;
            break;
        case CAlignmentTrack::eLayout_Adaptive1000:
        default:
            cut_off = 1000;
            break;
    }
    return cut_off;
}


/// layout style to layout display name
typedef SStaticPair<CAlignmentTrack::ELayout, const char*> TLayoutDispStr;
static const TLayoutDispStr s_LayoutDispStrs[] = {
    { CAlignmentTrack::eLayout_Smear,          "Packed" },
    { CAlignmentTrack::eLayout_AdaptiveSeq,    "Adaptive (packed if not at sequence level)" },
    { CAlignmentTrack::eLayout_Adaptive200,    "Adaptive (packed if > 200)" },
    { CAlignmentTrack::eLayout_Adaptive1000,   "Adaptive (packed if > 1000)" },
    { CAlignmentTrack::eLayout_Adaptive5000,   "Adaptive (packed if > 5000)" },
    { CAlignmentTrack::eLayout_Adaptive20000,  "Adaptive (packed if > 20000)" },
    { CAlignmentTrack::eLayout_Adaptive100000, "Adaptive (packed if > 100000)" },
    { CAlignmentTrack::eLayout_Adaptive250000, "Adaptive (packed if > 250000)" },
    { CAlignmentTrack::eLayout_ExpandedByPos,  "Ladder (one alignment per row)" },
};


typedef CStaticArrayMap<CAlignmentTrack::ELayout, string> TLayoutDispMap;
DEFINE_STATIC_ARRAY_MAP(TLayoutDispMap, sm_LayoutDispMap, s_LayoutDispStrs);

const string& s_LayoutToDisplayName(CAlignmentTrack::ELayout layout)
{
    TLayoutDispMap::const_iterator iter = sm_LayoutDispMap.find(layout);
    if (iter != sm_LayoutDispMap.end()) {
        return iter->second;
    }
    return kEmptyStr;
}

// layout style to/from layout string/name
typedef SStaticPair<const char*, CAlignmentTrack::ELayout> TLayoutStr;
static const TLayoutStr s_LayoutStrs[] = {
    { "Adaptive1000",       CAlignmentTrack::eLayout_Adaptive1000 },
    { "Adaptive100000",     CAlignmentTrack::eLayout_Adaptive100000 },
    { "Adaptive200",        CAlignmentTrack::eLayout_Adaptive200 },
    { "Adaptive20000",      CAlignmentTrack::eLayout_Adaptive20000 },
    { "Adaptive250000",     CAlignmentTrack::eLayout_Adaptive250000 },
    { "Adaptive5000",       CAlignmentTrack::eLayout_Adaptive5000 },
    { "AdaptiveSeq",        CAlignmentTrack::eLayout_AdaptiveSeq },
    { "Expanded",           CAlignmentTrack::eLayout_ExpandedByPos },
    { "Smear",              CAlignmentTrack::eLayout_Smear },
};

typedef CStaticArrayMap<string, CAlignmentTrack::ELayout> TLayoutMap;
DEFINE_STATIC_ARRAY_MAP(TLayoutMap, sm_LayoutMap, s_LayoutStrs);


CAlignmentTrack::ELayout CAlignmentTrack::LayoutStrToValue(const string& layout)
{
    TLayoutMap::const_iterator iter = sm_LayoutMap.find(layout);
    if (iter != sm_LayoutMap.end()) {
        return iter->second;
    }
    NCBI_THROW(CException, eInvalid, "Invalid layout string: " + layout);
}


const string& CAlignmentTrack::LayoutValueToStr(CAlignmentTrack::ELayout layout)
{
    TLayoutMap::const_iterator iter;
    for (iter = sm_LayoutMap.begin();  iter != sm_LayoutMap.end();  ++iter) {
        if (iter->second == layout) {
            return iter->first;
        }
    }
    return kEmptyStr;
}


typedef SStaticPair<const char*, CAlignmentConfig::ELabelPosition> TLabelPosStr;
static const TLabelPosStr s_LabelPosStrs[] = {
    { "above",    CAlignmentConfig::ePos_Above },
    { "no label", CAlignmentConfig::ePos_NoLabel },
    { "side",     CAlignmentConfig::ePos_Side },
};

typedef CStaticArrayMap<string, CAlignmentConfig::ELabelPosition> TLabelPosMap;
DEFINE_STATIC_ARRAY_MAP(TLabelPosMap, sm_LabelPosMap, s_LabelPosStrs);


CAlignmentConfig::ELabelPosition
CAlignmentTrack::LabelPosStrToValue(const string& pos)
{
    TLabelPosMap::const_iterator iter = sm_LabelPosMap.find(pos);
    if (iter != sm_LabelPosMap.end()) {
        return iter->second;
    }
    NCBI_THROW(CException, eInvalid, "Invalid label position string: " + pos);
}


const string&
CAlignmentTrack::LabelPosValueToStr(CAlignmentConfig::ELabelPosition pos)
{
    TLabelPosMap::const_iterator iter;
    for (iter = sm_LabelPosMap.begin();  iter != sm_LabelPosMap.end();  ++iter) {
        if (iter->second == pos) {
            return iter->first;
        }
    }
    return kEmptyStr;
}



static const string kAlignGlyphKey = "GBPlugins.SeqGraphicAlignGlyphs";
static const string kMatePairKey = "GBPlugins.SeqGraphicMatePair";
static const string kBaseKey = "GBPlugins.SeqGraphicAlignment";

static const string kTrackName = "Alignments";
static const string kDefProfile = "Default";

CTrackTypeInfo CAlignmentTrack::m_TypeInfo("alignment_track",
                                           "Graphical View Alignment Track");

CAlignmentTrack::CAlignmentTrack(CSGAlignmentDS* ds, CRenderingContext* r_cntx)
    : CDataTrack(r_cntx)
    , m_ContentHandler(new CAlnContentEvtHandler(this))
    , m_DS(ds)
    , m_Layout(eLayout_Default)
    , m_ShowAlignedSeqFeats(false)
    , m_ShowLabel(true)
    , m_ShowUnalignedTails(true)
    , m_ShowUnalignedTailsForTrans2GenomicAln(true)
    , m_ShowAlnStat(false)
    , m_CompactEnabled(true)
    , m_CompactThreshold(100)
    , m_ExtremeCompactThreshold(1000)
    , m_AlnType(IAlnExplorer::fInvalid)
    , m_ObjNum(0)
    , m_Column(new CColumnLayout)
    , m_GeneModelLayout(new CSimpleLayout)
{
    m_DS->SetJobListener(this);

    // initialize annotation selector
    SetLayoutPolicy(m_Column);
    x_RegisterIcon(SIconInfo(eIcon_Content, "Content", true, "track_content"));
    x_RegisterIcon(SIconInfo(eIcon_Layout, "Layout style",
        true, "track_layout"));
    x_RegisterIcon(SIconInfo(eIcon_Score, "Alignment score coloration",
        true, "track_align_score"));
    x_RegisterIcon(SIconInfo(eIcon_Stat, "Alignment statistics settings",
        true, "track_stat"));

    
    m_Column->SetVertSpace(2);
    m_GeneModelLayout->SetVertSpace(2);
    m_Simple->SetVertSpace(2);
}


CAlignmentTrack::~CAlignmentTrack()
{
    if ( !m_DS->AllJobsFinished() ) {
        m_DS->DeleteAllJobs();
    }
}


const CTrackTypeInfo& CAlignmentTrack::GetTypeInfo() const
{
    return m_TypeInfo;
}


string CAlignmentTrack::GetFullTitle() const
{
    string title = GetTitle();
    if (title.empty()) {
        if (CSeqUtils::IsUnnamed(m_AnnotName)) {
            title = "Other alignments";
        } else {
            title = m_AnnotName;
        }
    } else {
        if (m_AnnotName.find(title) != string::npos) {
            title = m_AnnotName;
        } else if (title.find(m_AnnotName) == string::npos  &&
                   !CSeqUtils::IsUnnamed(m_AnnotName)) {
            title = title + " - " + m_AnnotName;
        }
    }
    return title;
}


void CAlignmentTrack::SetAnnot(const string& annot)
{
    m_AnnotName = annot;
}


void CAlignmentTrack::OnToggleContent(int id)
{
    // the menu item order:
    // 0 - "Show labels"
    // 1 - "Show unaligned tails for all alignments"
    // 2 - "Show unaligned tails for transcript-to-genomic alignments"
    // 3 - "Project features for aligned sequences"
    // 4 - "Link mate pairs"
    switch (id) {
    case 0:
        {{
            m_ShowLabel = !m_ShowLabel;
            if (m_MultiAlignConf) {
                m_MultiAlignConf->m_ShowLabel = m_ShowLabel;
            }
            if (m_PWAlignConf) {
                m_PWAlignConf->m_ShowLabel = m_ShowLabel;
            }
            if (m_AlignSmearConf) {
                m_AlignSmearConf->m_ShowLabel = m_ShowLabel;
            }
            if (m_MatePairConf) {
                m_MatePairConf->m_ShowLabel = m_ShowLabel;
            }
            break;
        }}

    case 1:
        {{
            m_ShowAlnStat = !m_ShowAlnStat;
            break;
        }}
    case 2:
        {{
            m_ShowUnalignedTails = !m_ShowUnalignedTails;
            if (m_PWAlignConf) {
                m_PWAlignConf->m_ShowUnalignedTails = m_ShowUnalignedTails;
            }
            if (m_AlignSmearConf) {
                m_AlignSmearConf->m_ShowUnalignedTails = m_ShowUnalignedTails;
            }
            break;
        }}
    case 3:
        {{
            m_ShowUnalignedTailsForTrans2GenomicAln =
                !m_ShowUnalignedTailsForTrans2GenomicAln;
            if (m_PWAlignConf) {
                m_PWAlignConf->m_ShowUnalignedTailsForTrans2GenomicAln =
                    m_ShowUnalignedTailsForTrans2GenomicAln;
            }
            if (m_AlignSmearConf) {
                m_AlignSmearConf->m_ShowUnalignedTailsForTrans2GenomicAln =
                    m_ShowUnalignedTailsForTrans2GenomicAln;
            }
            break;
        }}
    case 4:
        m_ShowAlignedSeqFeats = !m_ShowAlignedSeqFeats;
        break;
    case 5:
        m_DS->SetLinkMatePairs( !m_DS->GetLinkMatePairs() );
        break;
    case 6:
        {
            string sort_by = m_DS->GetSortBy().empty() ?
                CAlignStrandSorter::GetID() : "";
            m_DS->SetSortBy(sort_by);
        }
        break;
    default:
        break;
    }
    x_UpdateData();
}


void CAlignmentTrack::OnChangeProjectedFeats()
{
    CFeatureCheckDialog dlg(NULL);
    CFeatureCheckDialog::TFeatTypeItemSet items_set;
	const CFeatList* feat_list = CSeqFeatData::GetFeatList();
    CFeatListItem item;
    ITERATE(TSubtypeVec, subtype, m_ProjectedFeats) {
        feat_list->GetItemBySubtype(*subtype, item);
        items_set.insert(item);
    }
    dlg.SetSelected(items_set);
    IGlyphDialogHost* dlg_host = dynamic_cast<IGlyphDialogHost*>(m_LTHost);
    if (dlg_host) {
        dlg_host->PreDialogShow();
    }
    if(dlg.ShowModal() == wxID_OK) {
        dlg.GetSelected(items_set);
        m_ProjectedFeats.clear();
        ITERATE(CFeatureCheckDialog::TFeatTypeItemSet, iter, items_set) {
            m_ProjectedFeats.push_back(iter->GetSubtype());		
        }
        x_UpdateData();
    }
    if (dlg_host) {
        dlg_host->PostDialogShow();
    }
}


void CAlignmentTrack::x_LoadSettings(const string& preset_style,
                                     const TKeyValuePairs& settings)
{
    if (preset_style.empty()) {
        SetProfile(kDefProfile);
    } else {
        SetProfile(preset_style);
    }

    if ( !m_StatConf ) {
        m_StatConf.Reset(new CAlnStatConfig);
    }

    CGuiRegistry& registry = CGuiRegistry::GetInstance();
    CGuiRegistry::TReadView view =
        CSGConfigUtils::GetReadView(registry, kBaseKey, GetProfile(), kDefProfile);

    m_DS->SetLinkMatePairs(view.GetBool("LinkMatePairAligns", true));
    m_DS->SetDNAScoringMethod(view.GetString("DNAScoringMethod"));
    m_DS->SetProteinScoringMethod(view.GetString("ProteinScoringMethod"));
    m_DS->SetEnableColoration(view.GetBool("EnableColoration", true));
    m_ShowAlignedSeqFeats = view.GetBool("ShowAlignedSeqFeats", false);
    m_ShowLabel = view.GetBool("ShowLabel", false);
    m_ShowUnalignedTails = view.GetBool("ShowUnalignedTails", false);
    m_ShowUnalignedTailsForTrans2GenomicAln = view.GetBool("ShowUnalignedTailsForTrans2GenomicAln", true);
    m_ShowAlnStat = view.GetBool("ShowAlnStat", false);
    m_CompactEnabled = view.GetBool("SupportCompactDisplay");
    m_CompactThreshold = (size_t)view.GetInt("CompactThreshold", 100);
    m_ExtremeCompactThreshold = (size_t)view.GetInt("ExtremeCompactThreshold", 1000);
    m_StatConf->m_Display = view.GetInt("StatDisplay", 15);
    m_StatConf->m_StatZoomLevel = view.GetInt("StatZoomLevel", 100);
    try {
        m_Layout = LayoutStrToValue(view.GetString("Layout"));
    } catch (CException& e) {
        LOG_POST(Warning << "CAlignmentTrack::x_LoadSettings() " << e.GetMsg());
    }

    CConstRef<CUser_field> f = view.GetField("ProjectedFeats");
    if (f  &&  f->GetData().Which() == CUser_field::TData::e_Str) {
        string str = f->GetData().GetStr();
        NStr::TruncateSpacesInPlace(str);
        list<string> toks;
        NStr::Split(str, ", |", toks);
        try {
            ITERATE (list<string>, iter, toks) {
                m_ProjectedFeats.push_back(NStr::StringToInt(*iter));
            }
        } catch (CException& e) {
            LOG_POST(Warning << "CAlignmentTrack::x_LoadSettings error on "
                <<"parsing projected feature list: " << e.GetMsg());
        }
    }

    m_MultiAlignProfile = view.GetString("MultiAlignProfile", kDefProfile);
    m_PWAlignProfile = view.GetString("PWAlignProfile", kDefProfile);
    m_AlignSmearProfile = view.GetString("AlignSmearProfile", kDefProfile);
    m_MatePairProfile = view.GetString("MatePairProfile", kDefProfile);

    view = CSGConfigUtils::GetColorReadView(registry, kBaseKey, GetProfile(), kDefProfile);
    CSGConfigUtils::GetColor(view, "A", m_StatConf->m_Colors[CAlnStatConfig::eStat_A]);
    CSGConfigUtils::GetColor(view, "T", m_StatConf->m_Colors[CAlnStatConfig::eStat_T]);
    CSGConfigUtils::GetColor(view, "G", m_StatConf->m_Colors[CAlnStatConfig::eStat_G]);
    CSGConfigUtils::GetColor(view, "C", m_StatConf->m_Colors[CAlnStatConfig::eStat_C]);
    CSGConfigUtils::GetColor(view, "Gap", m_StatConf->m_Colors[CAlnStatConfig::eStat_Gap]);
    CSGConfigUtils::GetColor(view, "Match", m_StatConf->m_Colors[CAlnStatConfig::eStat_Match]);
    CSGConfigUtils::GetColor(view, "Mismatch", m_StatConf->m_Colors[CAlnStatConfig::eStat_Mismatch]);
    CSGConfigUtils::GetColor(view, "Total", m_StatConf->m_Colors[CAlnStatConfig::eStat_Total]);
    
    view = CSGConfigUtils::GetSizeReadView(registry, kBaseKey, GetProfile(), kDefProfile);
    m_StatConf->m_GraphHeight = view.GetInt("StatGraphHeight", 40);

    // override those default settings passed in through profile settings
    ITERATE (TKeyValuePairs, iter, settings) {
        try {
            if (NStr::EqualNocase(iter->first, "Color")) {
                if (NStr::EqualNocase(iter->second, "false")) {
                    m_DS->SetEnableColoration(false);
                } else {
                    m_DS->SetEnableColoration(true);
                    if ( !NStr::EqualNocase(iter->second, "true") ) {
                        if ((m_AlnType == IAlnExplorer::fProtein  ||
                            m_AlnType == IAlnExplorer::fMixed)  &&
                            m_DS->IsValidProteinScoringMethod(iter->second)) {
                            m_DS->SetProteinScoringMethod(iter->second);
                        } else if (m_DS->IsValidDNAScoringMethod(iter->second)) {
                            // treat it as DNA
                            m_DS->SetDNAScoringMethod(iter->second);
                        }
                    }
                }
            } else if (NStr::EqualNocase(iter->first, "DNAScoringMethod")) {
                m_DS->SetDNAScoringMethod(iter->second);
            } else if (NStr::EqualNocase(iter->first, "ProteinScoringMethod")) {
                m_DS->SetProteinScoringMethod(iter->second);
            } else if (NStr::EqualNocase(iter->first, "AlignedSeqFeats"))
                m_ShowAlignedSeqFeats = NStr::StringToBool(iter->second);
            else if (NStr::EqualNocase(iter->first, "Label"))
                m_ShowLabel = NStr::StringToBool(iter->second);
            else if (NStr::EqualNocase(iter->first, "UnalignedTails"))
                m_ShowUnalignedTails = NStr::StringToBool(iter->second);
            else if (NStr::EqualNocase(iter->first, "ShowUnalignedTailsForTrans2GenomicAln"))
                m_ShowUnalignedTailsForTrans2GenomicAln = NStr::StringToBool(iter->second);
            else if (NStr::EqualNocase(iter->first, "LinkMatePairAligns"))
                m_DS->SetLinkMatePairs(NStr::StringToBool(iter->second));
            else if (NStr::EqualNocase(iter->first, "Layout"))
                m_Layout = LayoutStrToValue(iter->second);
            else if (NStr::EqualNocase(iter->first, "StatDisplay"))
                m_StatConf->m_Display = NStr::StringToInt(iter->second);
            else if (NStr::EqualNocase(iter->first, "StatZoomLevel")) {
                m_StatConf->m_StatZoomLevel = NStr::StringToInt(iter->second);
            } else if (NStr::EqualNocase(iter->first, "ShowAlnStat")) {
                m_ShowAlnStat = NStr::StringToBool(iter->second);
            } else if (NStr::EqualNocase(iter->first, "SupportCompactDisplay")) {
                m_CompactEnabled = NStr::StringToBool(iter->second);
            } else if (NStr::EqualNocase(iter->first, "SortBy")) {
                m_DS->SetSortBy(iter->second);
            } else if (NStr::EqualNocase(iter->first, "StatGraphHeight")) {
                m_StatConf->m_GraphHeight = NStr::StringToInt(iter->second);
            } else if (NStr::EqualNocase(iter->first, "ProjectedFeats")) {
                list<string> toks;
                NStr::Split(iter->second, ", |", toks);
                if ( !toks.empty() ) {
                    m_ProjectedFeats.clear();
                }
                try {
                    ITERATE (list<string>, f_iter, toks) {
                        m_ProjectedFeats.push_back(NStr::StringToInt(*f_iter));
                    }
                } catch (CException& e) {
                    LOG_POST(Warning << "CAlignmentTrack::x_LoadSettings() error on "
                        <<"parsing projected feature list: " << e.GetMsg());
                }
            } else
                LOG_POST(Warning << "CAlignmentTrack::x_LoadSettings(): the setting is not supported - " 
                         << iter->first << ":" << iter->second);
        } catch (CException&) {
            LOG_POST(Warning << "CAlignmentTrack::x_LoadSettings() invalid settings: " 
                     << iter->first << ":" << iter->second);
        }
    }

    x_LoadPWAlignSettings();
    if (m_MultiAlignConf) {
        x_LoadMultiAlignSettings();
    }
    if (m_AlignSmearConf) {
        x_LoadAlignSmearSettings();
    }
    if (m_MPPWAlignConf) {
        x_LoadMPPWAlignSettings();
    }
    if (m_MatePairConf) {
        x_LoadMatePairSettings();
    }

    m_StatConf->SetShowLabel(m_gConfig->GetShowComments());

    m_Layered->SetVertSpace(2);
    m_Column->SetVertSpace(2);
    x_OnLayoutPolicyChanged();
}


void CAlignmentTrack::x_SaveSettings(const string& preset_style)
{
    TKeyValuePairs settings;

    if ( !preset_style.empty() ) {
        settings["profile"] = preset_style;
    }

    settings["Color"] = NStr::BoolToString(m_DS->GetEnableColoration());
    settings["DNAScoringMethod"] = m_DS->GetDNAScoringMethod();
    settings["ProteinScoringMethod"] = m_DS->GetProteinScoringMethod();
    settings["AlignedSeqFeats"] = NStr::BoolToString(m_ShowAlignedSeqFeats);
    settings["Label"] = NStr::BoolToString(m_ShowLabel);
    settings["UnalignedTails"] = NStr::BoolToString(m_ShowUnalignedTails);
    settings["LinkMatePairAligns"] = NStr::BoolToString(m_DS->GetLinkMatePairs());
    settings["ShowUnalignedTailsForTrans2GenomicAln"] =
        NStr::BoolToString(m_ShowUnalignedTailsForTrans2GenomicAln);
    settings["Layout"] = LayoutValueToStr(m_Layout);

    string projected_feats = kEmptyStr;
    ITERATE (TSubtypeVec, iter, m_ProjectedFeats) {
        projected_feats += NStr::IntToString(*iter) + " ";
    }
    if ( !projected_feats.empty() ) {
        settings["ProjectedFeats"] = projected_feats;
    }
    settings["StatDisplay"] = NStr::IntToString(m_StatConf->m_Display);
    settings["ShowAlnStat"] = NStr::BoolToString(m_ShowAlnStat);
    settings["SupportCompactDisplay"] = NStr::BoolToString(m_CompactEnabled);
    settings["StatGraphHeight"] = NStr::IntToString(m_StatConf->m_GraphHeight);
    settings["StatZoomLevel"] = NStr::IntToString(m_StatConf->m_StatZoomLevel);
    settings["SortBy"] = m_DS->GetSortBy();
    SetProfile(CSGConfigUtils::ComposeProfileString(settings));
}


void CAlignmentTrack::x_OnIconClicked(TIconID id)
{
    switch (id) {
    case eIcon_Content:
        x_OnContentIconClicked();
        break;
    case eIcon_Layout:
        x_OnLayoutIconClicked();
        break;
    case eIcon_Score:
        x_OnScoreIconClicked();
        break;
    case eIcon_Stat:
        x_OnStatIconClicked();
        break;
    default:
        // use default handlers
        CLayoutTrack::x_OnIconClicked(id);
    }
}


void CAlignmentTrack::x_UpdateData()
{
    m_DS->DeleteAllJobs();
    x_SetStartStatus();

    int align_limit = s_LayoutToCutoff(m_Layout);
    const TSeqRange& vis_range = m_Context->GetVisSeqRange();
    const TModelUnit& zoom_level = m_Context->GetScale();

    if (m_DS->HasCoverageGraph()) {
        // bam alignments
        if (m_Context->GetScale() > m_StatConf->m_StatZoomLevel) {
            m_DS->LoadCoverageGraph(vis_range, zoom_level, eJob_CoverageGraph);
        } else {
            m_DS->LoadAlignments(vis_range, zoom_level, align_limit,
                false, eJob_Align);
            // After alignment loading job is completed, align pileup
            // graph will be loaded.
        }
    } else {
        // regular alignment annotation
        bool show_stat = m_ShowAlnStat &&
            m_Context->GetScale() <= m_StatConf->m_StatZoomLevel  &&
            m_AlnType == IAlnExplorer::fDNA;

        // Alignment statistics job, if enabled, will be
        // launched right after alignments get loaded.
        m_DS->LoadAlignments(vis_range, zoom_level, align_limit,
            !show_stat, eJob_Align);
    }
}


void CAlignmentTrack::x_OnJobCompleted(CAppJobNotification& notify)
{
    m_DS->ClearJobID(notify.GetJobID());
    CRef<CObject> res_obj = notify.GetResult();
    CJobResultBase* result = dynamic_cast<CJobResultBase*>(&*res_obj);
    if (!result) {
        LOG_POST(Error << "CAlignmentTrack::x_OnJobCompleted() notification for job \
                does not contain results.");
        return;
    }

    switch (result->m_Token) {
        case eJob_Align:
            x_AddAlignmentLayout(dynamic_cast<CSGJobResult&>(*res_obj));
            break;
        case eJob_AlignFeats:
            x_AddAlignFeatLayout(dynamic_cast<CBatchJobResult&>(*res_obj));
            break;
        case eJob_AlignScore:
            x_SetFinishStatus();
            x_SetMsg();

            // make sure we don't update layout twice for synchronized jobs
            if (m_DS->IsBackgroundJob()) {
                x_UpdateLayout();
            }
            break;
        case eJob_CoverageGraph:
            x_AddGraphLayout(dynamic_cast<CSGJobResult&>(*res_obj));
            break;
        case eJob_AlignStat:
            x_AddAlignStatLayout(dynamic_cast<CSGJobResult&>(*res_obj));
            break;
        default:
            break;
    }
}


void CAlignmentTrack::x_OnContentIconClicked()
{
    wxMenu menu;
    int curr_id = 0;
    wxMenuItem* item;
    item = menu.AppendCheckItem(kContentBaseID + curr_id++, wxT("Show labels"));
    if (m_ShowLabel) {
        item->Check();
    }
    item = menu.AppendCheckItem(kContentBaseID + curr_id++,
        wxT("Show alignment statistics"));
    if (m_ShowAlnStat) {
        item->Check();
    }
    item = menu.AppendCheckItem(kContentBaseID + curr_id++,
        wxT("Show unaligned tails for all alignments"));
    if (m_ShowUnalignedTails) {
        item->Check();
    }
    item = menu.AppendCheckItem(kContentBaseID + curr_id++,
        wxT("Show unaligned tails for transcript-to-genomic alignments"));
    if (m_ShowUnalignedTailsForTrans2GenomicAln) {
        item->Check();
    }
    item = menu.AppendCheckItem(kContentBaseID + curr_id++,
        wxT("Project features for aligned sequences"));
    if (m_ShowAlignedSeqFeats) {
        item->Check();
    }
    item = menu.AppendCheckItem(kContentBaseID + curr_id++, wxT("Link mate pairs"));
    if (m_DS->GetLinkMatePairs()) {
        item->Check();
    }
    item = menu.AppendCheckItem(kContentBaseID + curr_id++, wxT("Sort alignments by strand"));
    if ( !m_DS->GetSortBy().empty() ) {
        item->Check();
    }

    menu.AppendSeparator();
    item = menu.Append(kProjectedFeatID, wxT("Change projected feature list..."));

    m_LTHost->LTH_PushEventHandler(m_ContentHandler);
    m_LTHost->LTH_PopupMenu(&menu);
    m_LTHost->LTH_PopEventHandler();
}


void CAlignmentTrack::x_OnLayoutIconClicked()
{
    wxMenu menu;
    UseDefaultMarginWidth(menu);
    int id_base = 10000;

    ITERATE (TLayoutDispMap, iter, sm_LayoutDispMap) {
        bool l_default = iter->first == eLayout_Default;
        wxMenuItem* item = menu.AppendRadioItem(id_base + iter->first,
            ToWxString(iter->second + (l_default ? " (Default)" : "")));
        if (m_Layout == iter->first) {
            item->Check();
        }
    }

    int compact_item_idx = id_base + max(100, (int)sm_LayoutDispMap.size());
    {
        menu.AppendSeparator();
        wxMenuItem* item = menu.AppendCheckItem(compact_item_idx,
            wxT("Enable compact display"));
        item->Check(m_CompactEnabled);
    }

    m_LTHost->LTH_PopupMenu(&menu);
    wxMenuItemList& item_list = menu.GetMenuItems();
    ITERATE (wxMenuItemList, iter, item_list) {
        if ((*iter)->GetId() == compact_item_idx) {
            if (m_CompactEnabled != (*iter)->IsChecked()) {
                m_CompactEnabled = (*iter)->IsChecked();
                x_UpdateData();
                break;
            }
        } else {
            ELayout id = (ELayout)((*iter)->GetId() - id_base);
            if ((*iter)->IsChecked()  &&  id != m_Layout) {
                m_Layout = id;
                x_OnLayoutPolicyChanged();
                x_UpdateData();
                break;
            }
        }
    }
}


void CAlignmentTrack::x_OnScoreIconClicked()
{
    if (m_AlnType ==  IAlnExplorer::fInvalid) {
        return;
    }

    wxMenu menu;
    UseDefaultMarginWidth(menu);
    int id_base = 10000;
    int score_id = 0;
    int curr_id = 0;


    const string curr_method = m_AlnType == IAlnExplorer::fDNA ?
        m_DS->GetDNAScoringMethod() : m_DS->GetProteinScoringMethod();

    const CSGAlignmentDS::TMethods& methods = m_DS->GetScoringMethods(m_AlnType);
    ITERATE (CSGAlignmentDS::TMethods, iter, methods) {
        wxMenuItem* item = menu.AppendRadioItem(id_base + curr_id,
            ToWxString((*iter)->GetName()),
            ToWxString((*iter)->GetDescription()));

        if (curr_method == (*iter)->GetName()) {
            item->Check();
            score_id = curr_id;
        }
        curr_id++;
    }
    if (m_DS->HasQualityMap()) {
        const string& score_name = CSGAlnQualityScore::GetScoreName();
        wxMenuItem* item = menu.AppendRadioItem(id_base + curr_id,
            ToWxString(score_name), ToWxString(score_name));

        if (curr_method == score_name) {
            item->Check();
            score_id = curr_id;
        }
        curr_id++;
    }

    wxMenuItem* item = menu.AppendRadioItem(id_base + curr_id, wxT("Disabled"));
    if ( !m_DS->GetEnableColoration() ) {
       item->Check();
       score_id = curr_id;
    }

    m_LTHost->LTH_PopupMenu(&menu);
    wxMenuItemList& item_list = menu.GetMenuItems();
    ITERATE (wxMenuItemList, iter, item_list) {
        int checked_id = (*iter)->GetId() - id_base;
        if ((*iter)->IsChecked()  &&  score_id != checked_id) {
            if (checked_id == curr_id) {
                m_DS->SetEnableColoration(false);
            } else {
                m_DS->SetEnableColoration(true);
                if (m_AlnType == IAlnExplorer::fDNA) {
                    m_DS->SetDNAScoringMethod(ToStdString((*iter)->GetItemLabel()));
                } else if (m_AlnType == IAlnExplorer::fProtein  ||
                    m_AlnType == IAlnExplorer::fMixed) {
                    m_DS->SetProteinScoringMethod(ToStdString((*iter)->GetItemLabel()));
                }
            }
            x_UpdateData();
            break;
        }
    }
}


void CAlignmentTrack::x_OnStatIconClicked()
{
    if ( !m_StatConf ) return;

    CwxAlnStatOptionsDlg dlg(NULL);
    dlg.SetContent(m_StatConf->ShowAGTC());
    dlg.SetDisplay(m_StatConf->IsBarGraph());
    dlg.SetValueType(m_StatConf->ShowCount());
    dlg.SetGraphHeight(m_StatConf->m_GraphHeight);
    dlg.SetZoomLevel(m_StatConf->m_StatZoomLevel);
    IGlyphDialogHost* dlg_host = dynamic_cast<IGlyphDialogHost*>(m_LTHost);
    if (dlg_host) {
        dlg_host->PreDialogShow();
    }
    if(dlg.ShowModal() == wxID_OK) {
        m_StatConf->SetDisplayFlag(CAlnStatConfig::fBarGraph, dlg.IsBarGraph());
        m_StatConf->SetDisplayFlag(CAlnStatConfig::fShowMismatch, !dlg.IsAGTC());
        m_StatConf->SetDisplayFlag(CAlnStatConfig::fShowCount, dlg.IsCount());
        m_StatConf->m_GraphHeight = dlg.GetGraphHeight();
        m_StatConf->m_StatZoomLevel = dlg.GetZoomLevel();
        if (m_StatConf->m_GraphHeight < 10) {
            // set a minimal height
            m_StatConf->m_GraphHeight = 10;
        }
        m_ShowAlnStat = true;
        x_UpdateData();
    }
    if (dlg_host) {
        dlg_host->PostDialogShow();
    }
}


void CAlignmentTrack::x_OnLayoutPolicyChanged()
{
    CLayoutGroup* p_group = NULL;
    if (m_StatGlyph) {
        SetLayoutPolicy(m_GeneModelLayout);
        CRef<CSeqGlyph> p_glyph = SetGroup().GetChild(1);
        if (p_glyph) {
            p_group = dynamic_cast<CLayoutGroup*>(p_glyph.GetPointer());
        }
    }
    if ( !p_group ) {
        p_group = &SetGroup();
    }

    const CNamedGroup* first_group = NULL;
    if ( !p_group->GetChildren().empty() ) {
        first_group = dynamic_cast<const CNamedGroup*>(p_group->GetChildren().front().GetPointer());
    }

    vector<CLayoutGroup*> groups;
    if (first_group) {
        NON_CONST_ITERATE (CSeqGlyph::TObjects, grp_iter, p_group->SetChildren()) {
            CNamedGroup* curr_group = dynamic_cast<CNamedGroup*>(grp_iter->GetPointer());
            _ASSERT(curr_group);
            groups.push_back(curr_group);
        }
        p_group->SetLayoutPolicy(m_GeneModelLayout);
    } else {
        groups.push_back(p_group);
    }

    NON_CONST_ITERATE (vector<CLayoutGroup*>, g_iter, groups) {
        CLayoutGroup* curr_group = *g_iter;
        if (m_Layout == eLayout_ExpandedByPos) {
            curr_group->SetLayoutPolicy(m_Simple);
            m_Simple->SetSortingType(CSimpleLayout::eSort_BySeqPos);
        } else {
            curr_group->SetLayoutPolicy(m_Column);
        }
    }
}


void CAlignmentTrack::x_AddAlignmentLayout(const CSGJobResult& result)
{
    //TIME_ME("x_AddAlignmentLayout()");
    SetMsg("");
    m_ObjNum = result.m_ObjectList.size();
    const CNamedGroup* first_group = NULL;
    if (m_ObjNum > 0) {
        first_group = dynamic_cast<const CNamedGroup*>(result.m_ObjectList.front().GetPointer());
    }

    if (m_AlnType == IAlnExplorer::fInvalid  &&  m_ObjNum > 0) {
        if (first_group) {
            m_AlnType = m_DS->GetAlignType(first_group->GetChildren().front());
        } else {
            m_AlnType = m_DS->GetAlignType(result.m_ObjectList.front());
        }
    }

    bool show_stat = (m_ShowAlnStat  ||
        (m_DS->HasCoverageGraph()  &&  m_ObjNum == 0))  &&
        m_Context->GetScale() <= m_StatConf->m_StatZoomLevel  &&
        // For cSRA data (with coverage graph), no need to check align type.
        (m_AlnType == IAlnExplorer::fDNA || m_DS->HasCoverageGraph());

    /// all alignment we want to calculate statistics
    CSGAlignmentDS::TAlnMgrVec stat_aligns;

    // Trying to find out where to add the newly loaded objects.
    // Depending on what is currently being shown, we might have different
    // object hierarchies:
    // 1. both pileup and alignments
    //     group--
    //           |-- pileup glyph
    //           |-- group--
    //                     |-- alignments
    //
    // 2. only smear bar or coverage graph
    //     group--
    //           |-- smear bar/coverage graph
    //
    // 3. only pileup
    //     group--
    //           |-- pileup glyph
    //
    // 4. only alignments
    //     group--
    //           |-- alignments
    //

    CLayoutGroup* p_group = NULL;
    if ( !show_stat ) {
        // If we are not going to show pileup (regardless of it was shown
        // or not), use the top level group as the parent group
        p_group = &SetGroup();
        m_StatGlyph.Reset();
    } else if (m_StatGlyph) {
        // we are showing the pipleup, and will keep showing it.
        if (SetGroup().GetChildrenNum() > 1) {
            // case #1
            _ASSERT(SetGroup().GetChildrenNum() == 2);
            CRef<CSeqGlyph> p_glyph = SetGroup().GetChild(1);
            p_group = dynamic_cast<CLayoutGroup*>(p_glyph.GetPointer());
            _ASSERT(p_group);
            if (m_ObjNum == 0) {
                SetGroup().Remove(p_group);
                p_group = NULL;
            }
        } else if (m_ObjNum > 0) {
            // case #3
            CRef<CLayoutGroup> group(new CLayoutGroup);
            p_group = group.GetPointer();
            SetGroup().PushBack(p_group);
        }
    } else {
        // We are not showing the pipleup, and will show it.
        // This falls to case #2 or #4
        p_group = &SetGroup();
    }

    if (p_group) {
        p_group->Clear();
    } else {
        x_UpdateLayout();
        if (show_stat) {
            m_DS->CalcAlnStat(stat_aligns, m_Context->GetVisSeqRange(),
                m_Context->GetScale(), eJob_AlignStat);
        }
        return;
    }

    p_group->Set(result.m_ObjectList);

    /// all alignments we want to project features from
    vector< CRef<CAlignGlyph> > aligns;

    bool project_feats = m_ShowAlignedSeqFeats  &&
        !m_Context->IsOverviewMode()  &&
        !m_ProjectedFeats.empty();


    vector<CLayoutGroup*> groups;
    if (first_group) {
        NON_CONST_ITERATE (CSeqGlyph::TObjects, grp_iter, p_group->SetChildren()) {
            CNamedGroup* curr_group = dynamic_cast<CNamedGroup*>(grp_iter->GetPointer());
            _ASSERT(curr_group);

            // initialize named group
            curr_group->SetTitleColor(m_gConfig->GetFGCommentColor());
            curr_group->SetBackgroundColor(m_gConfig->GetBGCommentColor());
            curr_group->SetTitleFont(m_gConfig->GetCommentFont().GetPointer());
            curr_group->SetIndent(GetIndent() + 2);
            curr_group->SetRepeatDist(m_gConfig->GetCommentRepeatDist());
            curr_group->SetRepeatTitle(m_gConfig->GetRepeatComment());
            curr_group->SetShowTitle(m_gConfig->GetShowComments());
            groups.push_back(curr_group);
        }
    } else {
        groups.push_back(p_group);
    }

    bool show_score = !m_Context->IsOverviewMode()  &&
        m_DS->GetEnableColoration() &&
        m_AlnType != IAlnExplorer::fInvalid;

    CSeqGlyph::TObjects scored_alns;
    m_ObjNum = 0;
    NON_CONST_ITERATE (vector<CLayoutGroup*>, g_iter, groups) {
        CSeqGlyph::TObjects& objs = (*g_iter)->SetChildren();
        m_ObjNum += objs.size();
        //if (show_score) {
        //    std::copy(objs.begin(), objs.end(), back_inserter(scored_alns));
        //}

        NON_CONST_ITERATE (CSeqGlyph::TObjects, iter, objs) {
            CSeqGlyph* tmp = *iter;
            tmp->SetRenderingContext(m_Context);
            EAlignGlyphType type = x_GetAlignGlyphType(tmp);

            switch (type)
            {
            case eAlign_Multi:
                {{
                    if ( !m_MultiAlignConf ) {
                        x_LoadMultiAlignSettings();
                    }
                    CAlignGlyph* aln = dynamic_cast<CAlignGlyph*>(tmp);
                    aln->SetConfig(m_MultiAlignConf);
                    if (show_stat) {
                        stat_aligns.push_back(CConstRef<IAlnGraphicDataSource>(&aln->GetAlignMgr()));
                    }
                    if (show_score) {
                        scored_alns.push_back(*iter);
                    }
                }}
                break;
            case eAlign_PW:
                {{
                    CAlignGlyph* aln = dynamic_cast<CAlignGlyph*>(tmp);
                    aln->SetConfig(m_PWAlignConf);
                    if (project_feats) {
                        // project the annotated features on the aligned
                        // sequence onto the top sequence
                        x_LoadAlignedSeqFeats(aligns, aln);
                    }
                    if (show_stat) {
                        stat_aligns.push_back(CConstRef<IAlnGraphicDataSource>(&aln->GetAlignMgr()));
                    }
                    if (show_score) {
                        if (aln->IsSimplified()) {
                            aln->SetShowScore();
                        } else {
                            scored_alns.push_back(*iter);
                        }
                    }
                }}
                break;
            case eAlign_MatePair:
                {{
                    if ( !m_MatePairConf ) {
                        x_LoadMPPWAlignSettings();
                        x_LoadMatePairSettings();
                    }
                    CMatePairGlyph* mp = dynamic_cast<CMatePairGlyph*>(tmp);
                    CMatePairGlyph::TAlignList& mp_aligns = mp->SetSeqAligns();
                    mp->SetConfig(m_MatePairConf);
                    NON_CONST_ITERATE(CMatePairGlyph::TAlignList, mp_iter, mp_aligns) {
                        (*mp_iter)->SetConfig(mp_aligns.size() == 1 ? m_PWAlignConf : m_MPPWAlignConf);
                        (*mp_iter)->SetRenderingContext(m_Context);
                        if (show_stat) {
                            stat_aligns.push_back(CConstRef<IAlnGraphicDataSource>(&(*mp_iter)->GetAlignMgr()));
                        }
                        if (show_score) {
                            if ((*mp_iter)->IsSimplified()) {
                                (*mp_iter)->SetShowScore();
                            } else {
                                scored_alns.push_back(CRef<CSeqGlyph>(mp_iter->GetPointer()));
                            }
                        }
                    }
                }}
                break;
            case eAlign_Smear:
                {{
                    if ( !m_AlignSmearConf ) {
                        x_LoadAlignSmearSettings();
                    }
                    CAlignSmearGlyph* as = dynamic_cast<CAlignSmearGlyph*>(tmp);
                    as->SetConfig(m_AlignSmearConf);
                }}
                break;
            case eAlign_Graph:
                {{
                    CHistogramGlyph* graph = dynamic_cast<CHistogramGlyph*>(tmp);
                    graph->SetDialogHost(dynamic_cast<IGlyphDialogHost*>(m_LTHost));
                    CRef<CHistParamsManager> hist_mgr = m_gConfig->GetHistParamsManager();
                    graph->SetConfigManager(hist_mgr);
                    SetMsg(", Coverage graph");
                }}
                break;
            default:
                return;
            }
        }
    }

    CAlignmentConfig::ECompactMode compact_mode = CAlignmentConfig::eNormal;
    if (m_CompactEnabled  &&  m_ObjNum > m_CompactThreshold  &&
        !m_Context->WillSeqLetterFit()) {
        compact_mode = CAlignmentConfig::eCompact;
    }
    
    if (compact_mode == CAlignmentConfig::eCompact  &&
        m_ObjNum > m_ExtremeCompactThreshold) {
        compact_mode = CAlignmentConfig::eExtremeCompact;
    }

    if (m_PWAlignConf) {
        m_PWAlignConf->m_CompactMode = compact_mode;
        m_PWAlignConf->m_ShowLabel =
            m_ShowLabel  &&  compact_mode == CAlignmentConfig::eNormal;
    }
    if (m_MPPWAlignConf) {
        m_MPPWAlignConf->m_CompactMode = compact_mode;
        m_MPPWAlignConf->m_ShowLabel =
            m_ShowLabel  &&  compact_mode == CAlignmentConfig::eNormal;
    }

    
    int vert_space = 2;
    float min_dist = 4.0;
    if (compact_mode == CAlignmentConfig::eCompact) {
        vert_space = 1;
        min_dist = 3.0;
    } else if (compact_mode == CAlignmentConfig::eExtremeCompact) {
        vert_space = 0;
        min_dist = 2.0;
    }

    m_Layered->SetVertSpace(vert_space);
    m_Column->SetVertSpace(vert_space);
    m_Column->SetMinDist(max(TSeqPos(1),
           TSeqPos(m_Context->ScreenToSeq(min_dist))));

    m_GeneModelLayout->SetVertSpace(2);

    if (show_stat) {
        // calculate alignment statistics
        // if stat_aligns is empty, it may be because alignments are packed as
        // smear bar or graph. So we still need to try
        m_DS->CalcAlnStat(stat_aligns, m_Context->GetVisSeqRange(),
            m_Context->GetScale(), eJob_AlignStat);
    }

    // projected features from the aligned sequences
    if ( !aligns.empty() ) {
        m_DS->LoadAlignFeats(m_Context->GetVisSeqRange(),
        m_Context->GetScale(), aligns, m_ProjectedFeats, eJob_AlignFeats);
    }

    // caclucate alignment score
    if ( !scored_alns.empty() ) {
        m_DS->CalculateAlignmentScore(scored_alns, eJob_AlignScore);
    }

    x_OnLayoutPolicyChanged();
    x_SetMsg();
    x_UpdateLayout();
}


void CAlignmentTrack::x_AddGraphLayout(const CSGJobResult& result)
{
    m_StatGlyph.Reset();
    if (result.m_ObjectList.empty()) {
        SetGroup().Clear();
        SetMsg(", no data");
    } else {
        _ASSERT(result.m_ObjectList.size() == 1);
        SetGroup().Set(result.m_ObjectList);
        CSeqGlyph* tmp = SetGroup().SetChildren().front().GetPointer();
        CHistogramGlyph* graph = dynamic_cast<CHistogramGlyph*>(tmp);
        if (graph) {
            graph->SetDialogHost(dynamic_cast<IGlyphDialogHost*>(m_LTHost));
            CRef<CHistParamsManager> hist_mgr = m_gConfig->GetHistParamsManager();
            graph->SetConfigManager(hist_mgr);
            SetMsg(", Coverage graph");
        }
    }

    x_UpdateLayout();
}


void CAlignmentTrack::x_AddAlignStatLayout(const CSGJobResult& result)
{
    x_SetFinishStatus();
    x_SetMsg();

    if (result.m_ObjectList.empty()) {
        if (m_StatGlyph) {
            // do clean up
            m_StatGlyph.Reset();
            if (SetGroup().GetChildrenNum() > 1) {
                _ASSERT(SetGroup().GetChildrenNum() == 2);
                CRef<CSeqGlyph> p_glyph = SetGroup().GetChild(1);
                CLayoutGroup* p_group = dynamic_cast<CLayoutGroup*>(p_glyph.GetPointer());
                _ASSERT(p_group);
                SetGroup().Set(p_group->GetChildren());
                x_OnLayoutPolicyChanged();
            } else {
                SetGroup().Clear();
            }
            x_UpdateLayout();
        }
        return;
    }
    
    CSeqGlyph::TObjects objs = result.m_ObjectList;
    CAlnStatGlyph* stat_glyph =
        dynamic_cast<CAlnStatGlyph*>(objs.front().GetPointer());
    if (stat_glyph) {
        bool need_update = false;
        stat_glyph->SetConfig(m_StatConf);
        stat_glyph->SetDialogHost(dynamic_cast<IGlyphDialogHost*>(m_LTHost));
        CLayoutGroup& p_group = SetGroup();
        if (m_StatGlyph) {
            p_group.Replace(stat_glyph, 0);
        } else {
            CSeqGlyph::TObjects aligns = p_group.GetChildren();
            p_group.Clear();
            p_group.PushBack(stat_glyph);
            CRef<CLayoutGroup> group(new CLayoutGroup);
            p_group.PushBack(group.GetPointer());
            group->Set(aligns);
            need_update = true;
        }

        m_StatGlyph.Reset(stat_glyph);
        if (need_update) {
            x_OnLayoutPolicyChanged();
        }
        x_UpdateLayout();
    }
}


void CAlignmentTrack::x_AddAlignFeatLayout(CBatchJobResult& results)
{
    NON_CONST_ITERATE (CBatchJobResult::TResults, a_iter, results.m_Results) {
        CLayoutGroup::TObjectList& objs = (*a_iter)->m_ObjectList;
        CAlignGlyph* align =
            dynamic_cast<CAlignGlyph*>((*a_iter)->m_Owner.GetPointer());
        if (objs.empty()  ||  !align) continue;

        x_AddAlignFeat_Recursive(objs);
        CLayoutGroup* parent_group =
            dynamic_cast<CLayoutGroup*>(align->SetParent());
        _ASSERT(parent_group);
        CLayoutGroup* group = new CLayoutGroup;
        parent_group->Insert(align, group);
        parent_group->Remove(align);
        group->PushBack(align);
        group->SetLayoutPolicy(m_GeneModelLayout);
        if (objs.size() > 1) {
            // separate objects according to feature types.
            // All gene model features will be grouped together.
            // Anything else will be grouped separately.
            typedef map<int, CLayoutGroup::TObjectList> TFeatGroups;
            TFeatGroups feat_groups;
            NON_CONST_ITERATE (CLayoutGroup::TObjectList, iter, objs) {
                CSeqGlyph* glyph = iter->GetPointer();
                // if the features are in a group (any CLayoutGroup-derived group)
                // we know it must be a gene group
                CLayoutGroup* c_group = dynamic_cast<CLayoutGroup*>(glyph);
                if (c_group) {
                    feat_groups[CSeqFeatData::eSubtype_gene].push_back(*iter);
                } else { // must be a feature
                    const CFeatGlyph* feat = dynamic_cast<const CFeatGlyph*>(glyph);
                    _ASSERT(feat);
                    CSeqFeatData::E_Choice type =
                        feat->GetFeature().GetData().Which();
                    CSeqFeatData::ESubtype s_type =
                        feat->GetFeature().GetData().GetSubtype();
                    if (CSGUtils::IsMainFeature(type, s_type)) {
                        feat_groups[CSeqFeatData::eSubtype_gene].push_back(*iter);
                    } else {
                        feat_groups[s_type].push_back(*iter);
                    }
                }
            }
            NON_CONST_ITERATE (TFeatGroups, iter, feat_groups) {
                if (iter->second.size() > 1) {
                    CLayoutGroup* sub_group = new CLayoutGroup;
                    group->PushBack(sub_group);
                    if ( !m_FeatGroupConf ) {
                        m_FeatGroupConf.Reset(new CBoundaryParams(
                            true, false, CRgbaColor(0.5, 0.5f, 0.5f, 0.5f),
                            CRgbaColor(1.0f, 1.0f, 1.0f, 1.0f), 1.0));
                    }
                    sub_group->SetConfig(m_FeatGroupConf);
                    sub_group->SetLayoutPolicy(m_Column);
                    sub_group->Set(iter->second);
                } else {
                    group->Append(iter->second);
                }
            }
        } else {
            group->Append(objs);
        }
    }
    x_SetMsg();

    // make sure we don't update layout twice for synchronized jobs
    if (m_DS->IsBackgroundJob()) {
        x_UpdateLayout();
    }
}


void CAlignmentTrack::x_AddAlignFeat_Recursive(CLayoutGroup::TObjectList& objs)
{
    NON_CONST_ITERATE (CLayoutGroup::TObjectList, iter, objs) {
        CSeqGlyph* glyph = iter->GetPointer();
        glyph->SetRenderingContext(m_Context);
        if (CGeneGroup* group = dynamic_cast<CGeneGroup*>(glyph)) {
            if ( !m_GeneGroupConf ) {
                m_GeneGroupConf.Reset(new CBoundaryParams(
                    true, false, CRgbaColor(0.0, 0.6f, 0.0f, 0.1f),
                    CRgbaColor(0.6f, 0.8f, 0.3f, 0.3f), 1.0));
            }
            group->SetConfig(m_GeneGroupConf);
            group->SetLayoutPolicy(m_GeneModelLayout);
            group->SetShowGene(true);
            x_AddAlignFeat_Recursive(group->SetChildren());
        } else if (CLayoutGroup* group = dynamic_cast<CLayoutGroup*>(glyph)) {
            // must be an exon group
            group->SetLayoutPolicy(m_Inline);
            x_AddAlignFeat_Recursive(group->SetChildren());
        } else {
            // must be a feature
            CFeatGlyph* feat = dynamic_cast<CFeatGlyph*>(glyph);
            _ASSERT(feat);
            CSeqFeatData::ESubtype subtype =
                feat->GetFeature().GetData().GetSubtype();
            feat->SetConfig(m_gConfig->GetFeatParams(subtype));
            feat->SetProjectedFeat(true);
        }
    }
}

void CAlignmentTrack::x_LoadAlignedSeqFeats(vector< CRef<CAlignGlyph> >& aligns,
                                            CAlignGlyph* aln)
{
    // anchored seq-loc
    const CSeq_loc& loc     = aln->GetLocation();
    TSeqRange range = m_Context->GetVisSeqRange();
    range.IntersectWith(loc.GetTotalRange());

    // don't project features when the size (on screen) is less than 4 pixels
    if (m_Context->SeqToScreen((TModelUnit)range.GetLength()) >= 4.0) {
        aligns.push_back(CRef<CAlignGlyph>(aln));
    }

    /*
    CScope& scope = m_Context->GetScope();
    try {
        const IAlnGraphicDataSource& aln_mgr = aln->GetAlignMgr();
        // aligned seq-id
        const CSeq_id& aligned_seq =
        aln_mgr.GetSeqId(aln_mgr.GetAnchor() == 0 ? 1 : 0);

        SConstScopedObject obj(aligned_seq, scope);
        CGeneModelFactory gene_model;

        // map visible range to product feature
        CSeq_loc_Mapper seq_range_mapper(
            aln->GetAlignmentObj(), aligned_seq, &scope);
        CRef<CSeq_loc> tmp_loc(new CSeq_loc());
        tmp_loc->SetInt().SetFrom(range.GetFrom());
        tmp_loc->SetInt().SetTo  (range.GetTo());
        tmp_loc->SetId(*loc.GetId());
        CSeq_loc::TRange vis_range =
            seq_range_mapper.Map(tmp_loc.GetObject())->GetTotalRange();

        ILayoutTrackFactory::SExtraParams params(-1, true);
        params.m_Range = vis_range;
        ILayoutTrackFactory::TTrackMap tracks =
            gene_model.CreateTracks(obj, m_DSContext, m_Context, params);
        if ( !tracks.empty() ) {
            CLayoutGroup* parent_group =
                dynamic_cast<CLayoutGroup*>(aln->SetParent());
            _ASSERT(parent_group);
            CLayoutGroup* group = new CLayoutGroup;
            parent_group->Insert(aln, group);
            parent_group->Remove(aln);
            group->PushBack(aln);
            group->SetLayoutPolicy(m_Simple);
            
            NON_CONST_ITERATE (ILayoutTrackFactory::TTrackMap, iter, tracks) {
                CDataTrack* track =
                    dynamic_cast<CDataTrack*>(iter->second.GetPointer());
                if (track) {
                    group->PushBack(track);
                    CRef<CCoordMapper_SeqLocMapper> mapper(
                        new CCoordMapper_SeqLocMapper(
                        aln->GetAlignmentObj(), *loc.GetId(), &scope));
                    track->GetDataSource()->SetCoordMapper(mapper);
                    track->SetHost(m_LTHost);
                    track->SetConfig(m_gConfig);
                    track->SetIndent(0);
                    track->SetTrackAttr(0);
                    track->LoadProfile("");

                    track->SetVisRange(vis_range);

                    if (CFeatureTrack* f_track =
                        dynamic_cast<CFeatureTrack*>(track)) {
                        f_track->SetLayout(CFeatureTrack::eLayout_AdaptiveInline);
                    }
                    track->Update(false);
                }
            }
        }

    } catch (CAnnotMapperException&) {
        /// ignore errors from location mapping
    }
    */
}


CAlignmentTrack::EAlignGlyphType
CAlignmentTrack::x_GetAlignGlyphType(const CSeqGlyph* glyph) const
{
    const CAlignGlyph* align = dynamic_cast<const CAlignGlyph*>(glyph);
    if (align) {
        if (align->GetAlignMgr().GetNumRows() == 2) {
            return eAlign_PW;
        } else {
            return eAlign_Multi;
        }
    }

    const CMatePairGlyph* mp = dynamic_cast<const CMatePairGlyph*>(glyph);
    if (mp) {
        return eAlign_MatePair;
    }

    const CAlignSmearGlyph* smear = dynamic_cast<const CAlignSmearGlyph*>(glyph);
    if (smear) {
        return eAlign_Smear;
    }

    const CHistogramGlyph* graph = dynamic_cast<const CHistogramGlyph*>(glyph);
    if (graph) {
        return eAlign_Graph;
    }

    return eAlign_Non;
}


void  CAlignmentTrack::x_SaveConfiguration(const string& preset_style) const
{
    CGuiRegistry& registry = CGuiRegistry::GetInstance();
    CGuiRegistry::TReadWriteView view =
        CSGConfigUtils::GetReadWriteView(registry, kBaseKey, preset_style, kDefProfile);
    view.Set("LinkMatePairAligns", m_DS->GetLinkMatePairs());
    view.Set("DNAScoringMethod", m_DS->GetDNAScoringMethod());
    view.Set("ProteinScoringMethod", m_DS->GetProteinScoringMethod());
    view.Set("EnableColoration", m_DS->GetEnableColoration());
    view.Set("ShowAlignedSeqFeats", m_ShowAlignedSeqFeats);
    view.Set("ShowLabel", m_ShowLabel);
    view.Set("ShowUnalignedTails", m_ShowUnalignedTails);
    view.Set("ShowUnalignedTailsForTrans2GenomicAln", m_ShowUnalignedTailsForTrans2GenomicAln);
    view.Set("Layout", LayoutValueToStr(m_Layout));

    view.Set("MultiAlignProfile", m_MultiAlignProfile);
    view.Set("PWAlignProfile", m_PWAlignProfile);
    view.Set("AlignSmearProfile", m_AlignSmearProfile);
    view.Set("MatePairProfile", m_MatePairProfile);

    x_SaveMultiAlignSettings();
    x_SavePWAlignSettings();
    x_SaveAlignSmearSettings();
    x_SaveMatePairSettings();
}


void CAlignmentTrack::x_LoadAlignmentSettings(const string& profile,
                                              CRef<CAlignmentConfig>& conf)
{
    if ( !conf ) {
        conf.Reset(new CAlignmentConfig);
    }

    CGuiRegistry& registry = CGuiRegistry::GetInstance();
    CGuiRegistry::TReadView view = CSGConfigUtils::GetReadView(
        registry, kAlignGlyphKey, profile, kDefProfile);
    //conf->m_ShowLabel = view.GetBool("ShowLabel", true);
    conf->m_ShowLabel = m_ShowLabel;
    conf->m_ShowUnalignedTails = m_ShowUnalignedTails;
    conf->m_ShowUnalignedTailsForTrans2GenomicAln =
        m_ShowUnalignedTailsForTrans2GenomicAln;

    view = CSGConfigUtils::GetLabelPosReadView(
        registry, kAlignGlyphKey, profile, m_gConfig->GetLabelPos(), kDefProfile);
    try {
        conf->m_LabelPos = LabelPosStrToValue(view.GetString("LabelPos"));
    } catch (CException& e) {
        LOG_POST(Warning << "CAlignmentTrack::x_LoadAlignmentSettings() " 
            << e.GetMsg());
    }

    // load color settings
    view = CSGConfigUtils::GetColorReadView(
        registry, kAlignGlyphKey, profile, m_gConfig->GetColorTheme(), kDefProfile);
    CSGConfigUtils::GetColor(view, "BG", conf->m_BG);
    CSGConfigUtils::GetColor(view, "FG", conf->m_FG);
    CSGConfigUtils::GetColor(view, "Label", conf->m_Label);
    CSGConfigUtils::GetColor(view, "Sequence", conf->m_Sequence);
    CSGConfigUtils::GetColor(view, "SeqMismatch", conf->m_SeqMismatch);

    // load size settings
    view = CSGConfigUtils::GetSizeReadView(
        registry, kAlignGlyphKey, profile, m_gConfig->GetSizeLevel(), kDefProfile);

    conf->m_BarHeight = view.GetInt("BarHeight");
    CSGConfigUtils::GetFont(view, "LabelFontFace", "LabelFontSize", conf->m_LabelFont);
    CSGConfigUtils::GetFont(view, "SeqFontFace", "SeqFontSize", conf->m_SeqFont);
}


void 
CAlignmentTrack::x_SaveAlignmentSettings(const string& profile,
                                         CConstRef<CAlignmentConfig> conf) const
{
    if ( !conf  ||  !conf->m_Dirty) {
        return;
    }

    CGuiRegistry& registry = CGuiRegistry::GetInstance();
    CGuiRegistry::TReadWriteView view = CSGConfigUtils::GetReadWriteView(
        registry, kAlignGlyphKey, profile, kDefProfile);
    view.Set("ShowLabel", conf->m_ShowLabel);
    view.Set("ShowUnalignedTails", conf->m_ShowUnalignedTails);
    view.Set("m_ShowUnalignedTailsForTrans2GenomicAln", conf->m_ShowUnalignedTailsForTrans2GenomicAln);
    
    // label position
    view = CSGConfigUtils::GetLabelPosRWView(
        registry, kAlignGlyphKey, profile, m_gConfig->GetLabelPos(), kDefProfile);
    view.Set("LabelPos", LabelPosValueToStr(conf->m_LabelPos));

    // load color settings
    view = CSGConfigUtils::GetColorRWView(
        registry, kAlignGlyphKey, profile, m_gConfig->GetColorTheme(), kDefProfile);
    CSGConfigUtils::SetColor(view, "BG", conf->m_BG);
    CSGConfigUtils::SetColor(view, "FG", conf->m_FG);
    CSGConfigUtils::SetColor(view, "Label", conf->m_Label);
    CSGConfigUtils::SetColor(view, "Sequence", conf->m_Sequence);
    CSGConfigUtils::SetColor(view, "SeqMismatch", conf->m_SeqMismatch);

    // load size settings
    view = CSGConfigUtils::GetSizeRWView(
        registry, kAlignGlyphKey, profile, m_gConfig->GetSizeLevel(), kDefProfile);

    view.Set("BarHeight", conf->m_BarHeight);
    CSGConfigUtils::SetFont(view, "LabelFontFace", "LabelFontSize", conf->m_LabelFont);
    CSGConfigUtils::SetFont(view, "SeqFontFace", "SeqFontSize", conf->m_SeqFont);

    conf->m_Dirty = false;
}


void CAlignmentTrack::x_LoadMatePairSettings()
{
    if ( !m_MatePairConf) {
        m_MatePairConf.Reset(new CMatePairConfig);
    }

    const string& profile = m_MatePairProfile;
    CMatePairConfig* conf = m_MatePairConf;

    CGuiRegistry& registry = CGuiRegistry::GetInstance();
    CGuiRegistry::TReadView view = CSGConfigUtils::GetReadView(
        registry, kMatePairKey, profile, kDefProfile);
    //conf->m_ShowLabel = view.GetBool("ShowLabel", true);
    conf->m_ShowLabel = m_ShowLabel;
    
    // load color settings
    view = CSGConfigUtils::GetColorReadView(
        registry, kMatePairKey, profile, m_gConfig->GetColorTheme(), kDefProfile);
    CSGConfigUtils::GetColor(view, "BG", conf->m_BG);
    CSGConfigUtils::GetColor(view, "FGDistance", conf->m_FGDistance);
    CSGConfigUtils::GetColor(view, "FGLink", conf->m_FGLink);
    CSGConfigUtils::GetColor(view, "FGNo", conf->m_FGNo);
    CSGConfigUtils::GetColor(view, "FGNonUnique", conf->m_FGNonUnique);
    CSGConfigUtils::GetColor(view, "FGOrientation", conf->m_FGOrientation);
    CSGConfigUtils::GetColor(view, "FGCoAlign", conf->m_FGCoAlign);
    CSGConfigUtils::GetColor(view, "FGContraAlign", conf->m_FGContraAlign);
    CSGConfigUtils::GetColor(view, "Label", conf->m_Label);
    CSGConfigUtils::GetColor(view, "SeqDistance", conf->m_SeqDistance);
    CSGConfigUtils::GetColor(view, "SeqMismatchDistance", conf->m_SeqMismatchDistance);
    CSGConfigUtils::GetColor(view, "SeqMismatchNo", conf->m_SeqMismatchNo);
    CSGConfigUtils::GetColor(view, "SeqMismatchNonUnique", conf->m_SeqMismatchNonUnique);
    CSGConfigUtils::GetColor(view, "SeqMismatchOrientation", conf->m_SeqMismatchOrientation);
    CSGConfigUtils::GetColor(view, "SeqNo", conf->m_SeqNo);
    CSGConfigUtils::GetColor(view, "SeqNonUnique", conf->m_SeqNonUnique);
    CSGConfigUtils::GetColor(view, "SeqOrientation", conf->m_SeqOrientation);

    // load size settings
    view = CSGConfigUtils::GetSizeReadView(
        registry, kMatePairKey, profile, m_gConfig->GetSizeLevel(), kDefProfile);
    CSGConfigUtils::GetFont(view, "LabelFontFace", "LabelFontSize", conf->m_LabelFont);
}


void CAlignmentTrack::x_SaveMatePairSettings() const
{
    if ( !m_MatePairConf  ||  !m_MatePairConf->m_Dirty) {
        return;
    }

    const string& profile = m_MatePairProfile;
    const CMatePairConfig* conf = m_MatePairConf;

    CGuiRegistry& registry = CGuiRegistry::GetInstance();
    CGuiRegistry::TReadWriteView view = CSGConfigUtils::GetReadWriteView(
        registry, kMatePairKey, profile, kDefProfile);
    view.Set("ShowLabel", conf->m_ShowLabel);
    
    // load color settings
    view = CSGConfigUtils::GetColorRWView(
        registry, kMatePairKey, profile, m_gConfig->GetColorTheme(), kDefProfile);
    CSGConfigUtils::SetColor(view, "BG", conf->m_BG);
    CSGConfigUtils::SetColor(view, "FGDistance", conf->m_FGDistance);
    CSGConfigUtils::SetColor(view, "FGLink", conf->m_FGLink);
    CSGConfigUtils::SetColor(view, "FGNo", conf->m_FGNo);
    CSGConfigUtils::SetColor(view, "FGNonUnique", conf->m_FGNonUnique);
    CSGConfigUtils::SetColor(view, "FGOrientation", conf->m_FGOrientation);
    CSGConfigUtils::SetColor(view, "Label", conf->m_Label);
    CSGConfigUtils::SetColor(view, "SeqDistance", conf->m_SeqDistance);
    CSGConfigUtils::SetColor(view, "SeqMismatchDistance", conf->m_SeqMismatchDistance);
    CSGConfigUtils::SetColor(view, "SeqMismatchNo", conf->m_SeqMismatchNo);
    CSGConfigUtils::SetColor(view, "SeqMismatchNonUnique", conf->m_SeqMismatchNonUnique);
    CSGConfigUtils::SetColor(view, "SeqMismatchOrientation", conf->m_SeqMismatchOrientation);
    CSGConfigUtils::SetColor(view, "SeqNo", conf->m_SeqNo);
    CSGConfigUtils::SetColor(view, "SeqNonUnique", conf->m_SeqNonUnique);
    CSGConfigUtils::SetColor(view, "SeqOrientation", conf->m_SeqOrientation);

    // load size settings
    view = CSGConfigUtils::GetSizeRWView(
        registry, kMatePairKey, profile, m_gConfig->GetSizeLevel(), kDefProfile);
    CSGConfigUtils::SetFont(view, "LabelFontFace", "LabelFontSize", conf->m_LabelFont);

    conf->m_Dirty = false;
}


void CAlignmentTrack::x_SetMsg()
{
    if (m_ObjNum > 0) {
        string msg = ", total ";
        msg += NStr::SizetToString(m_ObjNum, NStr::fWithCommas);
        msg += " object";
        msg += m_ObjNum > 1 ? "s shown" : " shown";
        SetMsg(msg);
    } else {
        SetMsg("");
    }
}


///////////////////////////////////////////////////////////////////////////////
/// CAlignmentTrackFactory
///////////////////////////////////////////////////////////////////////////////
ILayoutTrackFactory::TTrackMap
CAlignmentTrackFactory::CreateTracks(SConstScopedObject& object,
                                     ISGDataSourceContext* ds_context,
                                     CRenderingContext* r_cntx,
                                     const SExtraParams& params,
                                     const TAnnotMetaDataList& src_annots) const
{
    TAnnotNameTitleMap annots;

    if (params.m_SkipGenuineCheck  &&  !params.m_Annots.empty()) {
        ITERATE (SExtraParams::TAnnots, iter, params.m_Annots) {
            annots.insert(TAnnotNameTitleMap::value_type(*iter, ""));
        }
    } else {
        // collect non-NA tracks
        CIRef<ISGDataSource> pre_ds =
            ds_context->GetDS(typeid(CSGAlignmentDSType).name(), object);
        CSGAlignmentDS* aln_ds = dynamic_cast<CSGAlignmentDS*>(pre_ds.GetPointer());
        aln_ds->SetDepth(params.m_Level);
        aln_ds->SetAdaptive(params.m_Adaptive);
        SAnnotSelector sel(CSeqUtils::GetAnnotSelector(params.m_Annots));
        aln_ds->GetAnnotNames(sel, r_cntx->GetVisSeqRange(), annots);
        
        // collect NA tracks
        if ( !src_annots.empty() ) {
            GetMatchedAnnots(src_annots, params.m_Annots, "align", "", annots);
        }
    }
    
    // check if there are any coverage graph available
    vector<string> target_g_annots;
    ITERATE (TAnnotNameTitleMap, iter, annots) {
        if (CSeqUtils::NameTypeStrToValue(iter->first) == CSeqUtils::eAnnot_Other) {
            target_g_annots.push_back(iter->first);
        }
    }
        
    TAnnotNameTitleMap graph_annots;
    if ( !target_g_annots.empty() ) {
        CIRef<ISGDataSource> igraph_ds =
            ds_context->GetDS(typeid(CSGGraphDSType).name(), object);
        CSGGraphDS* graph_ds = dynamic_cast<CSGGraphDS*>(igraph_ds.GetPointer());
        graph_ds->SetDepth(params.m_Level);
        graph_ds->SetAdaptive(params.m_Adaptive);
        SAnnotSelector graph_sel(CSeqUtils::GetAnnotSelector(target_g_annots));
        graph_ds->GetAnnotNames(graph_sel, r_cntx->GetVisSeqRange(), graph_annots);
    }
    
    // create feature tracks
    TTrackMap tracks;
    ITERATE (TAnnotNameTitleMap, iter, annots) {
        CIRef<ISGDataSource> ds = ds_context->GetDS(
            typeid(CSGAlignmentDSType).name(), object);
        CSGAlignmentDS* aln_ds = dynamic_cast<CSGAlignmentDS*>(ds.GetPointer());
        aln_ds->SetDepth(params.m_Level);
        aln_ds->SetAdaptive(params.m_Adaptive);
        aln_ds->SetAnnotName(iter->first);
        aln_ds->SetSortBy(params.m_SortBy);
    
        bool has_coverage_graph = false;
        if ( !CSeqUtils::IsUnnamed(iter->first) ) { 
            has_coverage_graph = graph_annots.count(iter->first) > 0;
            aln_ds->SetHasCoverageGraph(has_coverage_graph);
        }

        CRef<CAlignmentTrack> aln_track(new CAlignmentTrack(aln_ds, r_cntx));
        // only do this for non-cSRA/BAM file
        aln_track->SetAlnType(aln_ds->InitAlignType());

        aln_track->SetAnnot(iter->first);
        if ( !iter->second.empty() ) {
            aln_track->SetTitle(iter->second);
        }
        aln_track->SetDSContext(ds_context);
        tracks[iter->first] = aln_track.GetPointer();
    }

    return tracks;
}

static const char* kDefaultScoringMethod = "Show Differences";

CRef<CTrackConfigSet>
CAlignmentTrackFactory::GetSettings(CSeqGraphicConfig* /*g_conf*/,
                                    const string& profile,
                                    const TKeyValuePairs& settings,
                                    const CTempTrackProxy* track_proxy) const
{
    CRef<CTrackConfigSet> config_set(new CTrackConfigSet);
    // create a track configure
    CRef<CTrackConfig> config(new CTrackConfig);
    config_set->Set().push_back(config);
    config->SetHelp() = GetThisTypeInfo().GetDescr();
    config->SetLegend_text("graphicalViewLegend.html#anchor_7");

    CGuiRegistry& registry = CGuiRegistry::GetInstance();
    CGuiRegistry::TReadView view =
        CSGConfigUtils::GetReadView(registry, kBaseKey, profile, kDefProfile);

    CLayoutTrack* track = const_cast<CLayoutTrack*>(track_proxy->GetTrack());
    CAlignmentTrack* aln_track = dynamic_cast<CAlignmentTrack*>(track);
    IAlnExplorer::EAlignType aln_type = IAlnExplorer::fInvalid;
    vector<string> scoring_methods;
    if (aln_track) {
        aln_type = aln_track->m_AlnType;
    }

    if (aln_type == IAlnExplorer::fProtein  ||
        aln_type == IAlnExplorer::fMixed) {
        scoring_methods.push_back("Column Quality score - Protein");
        scoring_methods.push_back("Frequency-Based Difference");
        scoring_methods.push_back(kDefaultScoringMethod);
        scoring_methods.push_back("Rasmol Amino Acid Colors");
        scoring_methods.push_back("Shapely Amino Acid Colors");
        scoring_methods.push_back("BLOSUM45");
        scoring_methods.push_back("BLOSUM62");
        scoring_methods.push_back("BLOSUM80");
        scoring_methods.push_back("Protein Quality Scoring with Coloring");
        scoring_methods.push_back("Hydropathy Scale");
        scoring_methods.push_back("Membrane preference");
        scoring_methods.push_back("Signal sequence");
        scoring_methods.push_back("Size");
    } else {
        scoring_methods.push_back("Column Quality score - DNA");
        scoring_methods.push_back("Frequency-Based Difference");
        scoring_methods.push_back(kDefaultScoringMethod);
        scoring_methods.push_back("Nucleic Acid Colors");
        if (profile == "BAM") {
            // HACK, HACK
            // It is a hack to indicate this is a BAM/cSRA track since
            // we don't have access to BAM/cSRA path at this stage.
            scoring_methods.push_back(CSGAlnQualityScore::GetScoreName());
        }
    }

    bool enable_color = view.GetBool("EnableColoration", true);
    string scoring_method = view.GetString("DNAScoringMethod", kDefaultScoringMethod);
    if (aln_type == IAlnExplorer::fProtein  ||  aln_type == IAlnExplorer::fMixed) {
        scoring_method = view.GetString("PorteinScoringMethod", kDefaultScoringMethod);
    }

    bool aligned_seqfeats = view.GetBool("ShowAlignedSeqFeats", false);
    bool label = view.GetBool("ShowLabel", true);
    bool link_mate_pair = view.GetBool("LinkMatePairAligns", false);
    //bool unaligned_tails = view.GetBool("ShowUnalignedTails", false);
    string rendering_style = view.GetString("Layout");
    int flag = view.GetInt("StatDisplay", 15);
    bool show_aln_stat = view.GetBool("ShowAlnStat", false);

    ITERATE (TKeyValuePairs, iter, settings) {
        try {
            if (NStr::EqualNocase(iter->first, "Color")) {
                if (NStr::EqualNocase(iter->second, "false")) {
                    enable_color = false;
                } else {
                    enable_color = true;
                    if ( !NStr::EqualNocase(iter->second, "true") ) {
                        // verify to see if it is a legitimate scoring method
                        ITERATE (vector<string>, sm_iter, scoring_methods) {
                            if (NStr::EqualNocase(iter->second, *sm_iter)) {
                                scoring_method = iter->second;
                                break;
                            }
                        }
                    }
                }
            } else if (NStr::EqualNocase(iter->first, "AlignedSeqFeats"))
                aligned_seqfeats = NStr::StringToBool(iter->second);
            else if (NStr::EqualNocase(iter->first, "Label"))
                label = NStr::StringToBool(iter->second);
            // else if (NStr::EqualNocase(iter->first, "UnalignedTails"))
            //     unaligned_tails = NStr::StringToBool(iter->second);
            else if (NStr::EqualNocase(iter->first, "Layout")) {
                // check whether the user-provided string is a valid layout string
                CAlignmentTrack::LayoutStrToValue(iter->second);
                rendering_style = iter->second;
            } else if (NStr::EqualNocase(iter->first, "StatDisplay")) {
                flag = NStr::StringToInt(iter->second);
            } else if (NStr::EqualNocase(iter->first, "ShowAlnStat")) {
                show_aln_stat = NStr::StringToBool(iter->second);
            } else if (NStr::EqualNocase(iter->first, "LinkMatePairAligns")) {
                link_mate_pair = NStr::StringToBool(iter->second);
            }

        } catch (CException&) {
            LOG_POST(Warning << "CAlignmentTrack::x_LoadSettings() invalid settings: " 
                     << iter->first << ":" << iter->second);
        }
    }


    // alignment rendering style settings
    CRef<CChoice> choice = CTrackConfigUtils::CreateChoice(
        "Layout", "Alignment Display", rendering_style,
        "Alignment rendering style");
    choice->SetValues().push_back(
         CTrackConfigUtils::CreateChoiceItem(
            CAlignmentTrack::LayoutValueToStr(CAlignmentTrack::eLayout_Smear),
            s_LayoutToDisplayName(CAlignmentTrack::eLayout_Smear),
            "Pack all alignments into a coverage graph or pileup graph",
            ""));
    choice->SetValues().push_back(
        CTrackConfigUtils::CreateChoiceItem(
            CAlignmentTrack::LayoutValueToStr(CAlignmentTrack::eLayout_AdaptiveSeq),
            s_LayoutToDisplayName(CAlignmentTrack::eLayout_AdaptiveSeq),
            "Show the individual alignments if at sequence level. Otherwise pack the alignments.",
            ""));
    choice->SetValues().push_back(
        CTrackConfigUtils::CreateChoiceItem(
            CAlignmentTrack::LayoutValueToStr(CAlignmentTrack::eLayout_Adaptive200),
            s_LayoutToDisplayName(CAlignmentTrack::eLayout_Adaptive200),
            "Show the individual alignments if less than 200 alignments. Otherwise pack the alignments.",
            ""));
    choice->SetValues().push_back(
        CTrackConfigUtils::CreateChoiceItem(
            CAlignmentTrack::LayoutValueToStr(CAlignmentTrack::eLayout_Adaptive1000),
            s_LayoutToDisplayName(CAlignmentTrack::eLayout_Adaptive1000),
            "Show the individual alignments if less than 1000 alignments. Otherwise pack the alignments.",
            ""));
    choice->SetValues().push_back(
        CTrackConfigUtils::CreateChoiceItem(
            CAlignmentTrack::LayoutValueToStr(CAlignmentTrack::eLayout_Adaptive5000),
            s_LayoutToDisplayName(CAlignmentTrack::eLayout_Adaptive5000),
            "Show the individual alignments if less than 5000 alignments. Otherwise pack the alignments.",
            ""));
    choice->SetValues().push_back(
        CTrackConfigUtils::CreateChoiceItem(
            CAlignmentTrack::LayoutValueToStr(CAlignmentTrack::eLayout_Adaptive20000),
            s_LayoutToDisplayName(CAlignmentTrack::eLayout_Adaptive20000),
            "Show the individual alignments if less than 20,000 alignments. Otherwise pack the alignments.",
            ""));
    choice->SetValues().push_back(
        CTrackConfigUtils::CreateChoiceItem(
            CAlignmentTrack::LayoutValueToStr(CAlignmentTrack::eLayout_Adaptive100000),
            s_LayoutToDisplayName(CAlignmentTrack::eLayout_Adaptive100000),
            "Show the individual alignments if less than 100,000 alignments. Otherwise pack the alignments.",
            ""));
    choice->SetValues().push_back(
        CTrackConfigUtils::CreateChoiceItem(
            CAlignmentTrack::LayoutValueToStr(CAlignmentTrack::eLayout_Adaptive250000),
            s_LayoutToDisplayName(CAlignmentTrack::eLayout_Adaptive250000),
            "Show the individual alignments if less than 250,000 alignments. Otherwise pack the alignments.",
            ""));
    choice->SetValues().push_back(
        CTrackConfigUtils::CreateChoiceItem(
            CAlignmentTrack::LayoutValueToStr(CAlignmentTrack::eLayout_ExpandedByPos),
            s_LayoutToDisplayName(CAlignmentTrack::eLayout_ExpandedByPos),
            "Show one alignment per row sorted by alignments' start location",
            ""));

    config->SetChoice_list().push_back(choice);

    // alignment statistics settings
    choice = CTrackConfigUtils::CreateChoice(
        "StatDisplay", "Pileup Display", NStr::IntToString(flag),
        "Alignment Pileup Display");

    /// For bam/csra track, show alignment statistics always.
    choice->SetValues().push_back(
        CTrackConfigUtils::CreateChoiceItem(
            "15",
            "Match/Mismatch graph (count)",
            "Show matches, mismatches, and gaps count as bar graph",
            ""));
    choice->SetValues().push_back(
        CTrackConfigUtils::CreateChoiceItem(
            "13",
            "Match/Mismatch graph (percentage)",
            "Show matches, mismatches, and gaps percentage as bar graph",
            ""));
    choice->SetValues().push_back(
        CTrackConfigUtils::CreateChoiceItem(
            "11",
            "ATGC graph (count)",
            "Show A, T, G, C, and gaps count as bar graph",
            ""));
    choice->SetValues().push_back(
        CTrackConfigUtils::CreateChoiceItem(
            "9",
            "ATGC graph (percentage)",
            "Show A, T, G, C, and gaps percentage as bar graph",
            ""));
    choice->SetValues().push_back(
        CTrackConfigUtils::CreateChoiceItem(
            "14",
            "Match/Mismatch table (count)",
            "Show matches, mismatches, and gaps count as table",
            ""));
    choice->SetValues().push_back(
        CTrackConfigUtils::CreateChoiceItem(
            "12",
            "Match/Mismatch table (percentage)",
            "Show matches, mismatches, and gaps percentage as table",
            ""));
    choice->SetValues().push_back(
        CTrackConfigUtils::CreateChoiceItem(
            "10",
            "ATGC table (count)",
            "Show A, T, G, C, and gaps count as table",
            ""));
    choice->SetValues().push_back(
        CTrackConfigUtils::CreateChoiceItem(
            "8",
            "ATGC table (percentage)",
            "Show A, T, G, C, and gaps percentage as table",
            ""));
    config->SetChoice_list().push_back(choice);


    if ( !enable_color ) {
        scoring_method = "false";
    }
    // setting for alignment socring method
    choice = CTrackConfigUtils::CreateChoice(
        "Color", "Score method", scoring_method,
        "Alignment score coloration method");

    ITERATE (vector<string>, iter, scoring_methods) {
        choice->SetValues().push_back(
            CTrackConfigUtils::CreateChoiceItem(
                *iter, *iter, *iter, ""));
    }
    choice->SetValues().push_back(
        CTrackConfigUtils::CreateChoiceItem(
            "false", "Disabled", "No score coloration", ""));
    config->SetChoice_list().push_back(choice);

    // setting for linking mate pair (on/off)
    config->SetCheck_boxes().push_back(
        CTrackConfigUtils::CreateCheckBox(
        "LinkMatePairAligns", "Link Mate Pairs",
        "Enable/disable linking mate reads", "", link_mate_pair));

    // setting for show pileup alongside with alignments (on/off)
    config->SetCheck_boxes().push_back(
        CTrackConfigUtils::CreateCheckBox(
            "ShowAlnStat","Show pileup", 
            "Enable/disable pileup display alongside with alignments", "",
            show_aln_stat));

    // setting for projecting features from aligned sequence
    config->SetCheck_boxes().push_back(
        CTrackConfigUtils::CreateCheckBox(
            "AlignedSeqFeats", "Project features",
            "Show features projected from the aligned sequences", "",
            aligned_seqfeats));

    // setting for label (on/off)
    config->SetCheck_boxes().push_back(
        CTrackConfigUtils::CreateCheckBox(
        "Label", "Show Labels", "Show/hide alignment labels", "", label));

    /*
    config->SetCheck_boxes().push_back(
        CTrackConfigUtils::CreateCheckBox(
        "UnalignedTails", "Show Unaligned Tails", 
        "Show/hide unaligned tails around an aligned sequences if there is any",
        "", unaligned_tails));
    */

    string sort_by = kEmptyStr;
    if (track_proxy)  {
         sort_by = track_proxy->GetSortBy();
    }

    choice = CTrackConfigUtils::CreateChoice(
        "sort_by", "Sort alignments by", sort_by,
        "Sort alignments according to a selected criterion.");

    choice->SetValues().push_back(
        CTrackConfigUtils::CreateChoiceItem(
        "", "No sorting", "Don't sort alignments", ""));

    IObjectSorter::SSorterDescriptor descr = CAlignStrandSorter::GetSorterDescr();

    choice->SetValues().push_back(
        CTrackConfigUtils::CreateChoiceItem(
        descr.m_Name, descr.m_DisplayName, descr.m_Descr, ""));

    config->SetChoice_list().push_back(choice);

    return config_set;
}


void CAlignmentTrackFactory::RegisterIconImages() const
{
    CLayoutTrack::RegisterIconImage("track_align_score", "track_align_score.png");
    CLayoutTrack::RegisterIconImage("track_stat", "track_stat.png");
}


string CAlignmentTrackFactory::GetExtensionIdentifier() const
{
    return CAlignmentTrack::m_TypeInfo.GetId();
}


string CAlignmentTrackFactory::GetExtensionLabel() const
{
    return CAlignmentTrack::m_TypeInfo.GetDescr();
}


END_NCBI_SCOPE
