/*  $Id: annot_conf.cpp 19013 2009-03-24 19:03:05Z dicuccio $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Robert G. Smith
 *
 */

/// @file annot_conf.cpp
///

#include <ncbi_pch.hpp>
#include <gui/widgets/seq_graphic/annot_conf.hpp>

#include <objects/seq/Annot_descr.hpp>
#include <objects/seq/Annotdesc.hpp>

#include <util/static_map.hpp>


BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

static const string kAnnotConfigTypeKey("SeqGraphicAlignAnnot");
static const string kSmearKey("smear");
static const string kBGColorKey("BGcolor");
static const string kShowLabelKey("showlabel");


/******
*       CAnnotReadConfig
*  for use by the renderer.
******/

CAnnotReadConfig::CAnnotReadConfig(const CSeq_annot& sa,
                                   CGuiRegistry* registry)
    : CSettingsSet(kAnnotConfigTypeKey, registry)
{
    // Get correct user object from sa.
    CConstRef<CUser_object> config_obj( CAnnotConfigUtils::GetConfigObject(sa) );
    // Merge it with our parent SettingsSet.
    MakeTempData(*config_obj);
}


CAnnotReadConfig::~CAnnotReadConfig()
{
    // pop the Seq-annot's config object off of our SettingsSet.
    CancelTempData();
}


bool CAnnotReadConfig::IsSmear() const
{
    return GetBool(kSmearKey, false);
}


bool CAnnotReadConfig::ShowLabel() const
{
    return GetBool(kShowLabelKey, true);
}


CRgbaColor CAnnotReadConfig::BGColor() const
{
    return GetColor(kBGColorKey);
}


/******
*  CAnnotWriteConfig
*  for use by the configuration dialog.
******/

CAnnotWriteConfig::CAnnotWriteConfig(CGuiRegistry* registry)
: CSettingsSet(kAnnotConfigTypeKey, registry)
{
}


CAnnotWriteConfig::~CAnnotWriteConfig()
{
    // pop the Seq-annot's config object off of our SettingsSet.
    if (IsAnnotSet()) {
        CancelTempData();
    }
}


bool CAnnotWriteConfig::IsAnnotConfigChanged() const
{
    if ( ! IsAnnotSet()) {
        return false;
    }
    CAnnotConfigSettings current_values;
    GetAnnotConfigSettings(current_values);
    return ! ( current_values == m_LastValues );
}


bool CAnnotWriteConfig::IsAnnotSet() const
{
    return m_SeqAnnot;
}

bool CAnnotWriteConfig::SetAnnot(CSeq_annot& sa)
{
    bool changed = false;
    if (IsAnnotSet()) {
        if (IsAnnotConfigChanged()) {
            SaveConfig();
            changed = true;
        }
        CancelTempData();
    }
    m_SeqAnnot.Reset(&sa);
    LoadConfig();
    return changed;
}


void CAnnotWriteConfig::LoadConfig()
{
    _ASSERT( IsAnnotSet() );
    if ( ! IsAnnotSet()) {
        return;
    }
    // Get correct user object from sa.
    CConstRef<CUser_object> config_obj( CAnnotConfigUtils::GetConfigObject(*m_SeqAnnot) );
    // Merge it with our parent SettingsSet.
    MakeTempData(*config_obj, true);
    // The values we just loaded are the ones to compare against.
    GetAnnotConfigSettings(m_LastValues);
}


void CAnnotWriteConfig::SaveConfig()
{
    _ASSERT(IsAnnotSet());
    if ( ! IsAnnotSet()) {
        return;
    }

    // get the modified user object.
    const CUser_object& uo = GetTempData();

    // add it back into the Seq-annot.
    CAnnotConfigUtils::SetConfigObject( *m_SeqAnnot, uo);

    // The values we currently have are the ones to compare against.
    GetAnnotConfigSettings(m_LastValues);
}


void CAnnotWriteConfig::GetAnnotConfigSettings(CAnnotConfigSettings& settings) const
{
    settings.m_Smear     = GetBool (kSmearKey, false);
    settings.m_ShowLabel = GetBool (kShowLabelKey, true);
    settings.m_BGColor   = GetColor(kBGColorKey);
}


void CAnnotWriteConfig::SetAnnotConfigSettings(const CAnnotConfigSettings& settings)
{
    if ( ! IsAnnotSet()) {
        return;
    }
    Set(kSmearKey,     settings.m_Smear);
    Set(kShowLabelKey, settings.m_ShowLabel);
    Set(kBGColorKey,   settings.m_BGColor);
}



/**********
    Functions to access configuration objects
    within CSeq_annot.
*********/
CConstRef<CUser_object> CAnnotConfigUtils::GetConfigObject(const CSeq_annot& sa)
{
    CConstRef<CUser_object> uo_ref;

    if (sa.CanGetDesc()) {
        ITERATE(CSeq_annot::TDesc::Tdata, ad_it, sa.GetDesc().Get()) {
            const CAnnotdesc& ad = **ad_it;
            if (ad.IsUser()) {
                const CUser_object& uo = ad.GetUser();
                if ( CRegistryFile::IsRegistry(uo) ) {
                    uo_ref.Reset(&uo);
                }
            }
        }
    }

    if (! uo_ref) {
        uo_ref = CRegistryFile::NewRegistryObject();
    }
    return uo_ref;
}


void CAnnotConfigUtils::SetConfigObject( CSeq_annot& sa, const CUser_object& uo)
{
    _ASSERT( CRegistryFile::IsRegistry(uo) );
    bool found = false;
    NON_CONST_ITERATE(CSeq_annot::TDesc::Tdata, ad_it, sa.SetDesc().Set()) {
        CAnnotdesc& ad = **ad_it;
        if (ad.IsUser()) {
            if ( CRegistryFile::IsRegistry(ad.GetUser()) ) {
                found = true;
                ad.SetUser().Assign(uo);
            }
        }
    }
    if (!found) {
        CRef<CAnnotdesc> ad_ref(new CAnnotdesc);
        ad_ref->SetUser().Assign(uo);
        sa.SetDesc().Set().push_back(ad_ref);
    }
}


END_NCBI_SCOPE
