/*  $Id: cds_glyph.cpp 27503 2013-02-26 18:16:55Z falkrb $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Liangshou Wu
 *
 * File Description:
 */

#include <ncbi_pch.hpp>
#include <gui/widgets/seq_graphic/cds_glyph.hpp>
#include <gui/widgets/seq_graphic/rendering_ctx.hpp>
#include <gui/opengl/glrender.hpp>
#include <objects/seqfeat/Cdregion.hpp>
#include <objmgr/util/sequence.hpp>
#include <util/sequtil/sequtil_manip.hpp>
#include <serial/iterator.hpp>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

static const int kVertSpace = 1;

CCdsGlyph::CCdsGlyph(const CMappedFeat& f)
    : CFeatGlyph(f)
{}

CCdsGlyph::CCdsGlyph(const CMappedFeat& f, const CSeq_loc& loc)
    : CFeatGlyph(f, loc)
{}


void CCdsGlyph::x_Draw() const
{   
    CGlRender& gl = GetGl();
    CFeatGlyph::x_Draw();

    if (m_Context->WillSeqLetterFit()) {
        bool flip_strand = m_Context->IsFlippedStrand();
        TModelUnit p_in_y = m_Height + m_Pos.Y();
        TModelUnit p_out_y = p_in_y;
        TModelUnit t_y = p_in_y;
        if (true/*IsSelected()*/) {
            TModelUnit trans_full = m_CdsConfig->m_TransFont.GetMetric(
                IGlFont::eMetric_FullCharHeight);
            t_y -= m_CdsConfig->m_ProdFont.TextHeight() + kVertSpace;
            p_in_y = t_y - trans_full;
        }

        // align sequence letter to bar center line vertically
        p_in_y -= (m_Config->GetBarHeight(false)
            - m_CdsConfig->m_ProdFont.TextHeight()) * 0.5;

        string prot_seq, trans_seq;
        TSeqPos offset = x_GetProtOffset(); // initial translation offset
        CRef<CSGSequenceDS> seq_ds = m_Context->GetSeqDS();
        const CSeqVector& vec = seq_ds->GetSeqVector();
        CSeqTranslator::Translate(GetFeature(),
                                  seq_ds->GetScope(),
                                  trans_seq,
                                  true, true);

        if (GetFeature().IsSetProduct()) {
            try {
                const CSeq_loc& product = GetFeature().GetProduct();
                CScope& scope = GetMappedFeature().GetAnnot().GetScope();
                CSeqVector prot_vec(product, scope,
                    CBioseq_Handle::eCoding_Iupac);
                prot_vec.GetSeqData (0, prot_vec.size(), prot_seq);
            }
            catch (CException&) {
                // ignore exceptions - these arise if the product doesn't resolve
            }
        } else {
            prot_seq = trans_seq;
        }

        TSeqPos seq_start = 0;
        size_t idx = 0;

        // TODO: should we use intervals instead of location ?
        for (CSeq_loc_CI lit(GetLocation()); lit; ++lit) {
            TSeqPos ifrom = lit.GetRange().GetFrom();
            TSeqPos ito   = lit.GetRange().GetTo();

            char basesIN[2], basesOUT[2];
            basesIN [1] = '\0';
            basesOUT[1] = '\0';

            bool neg = lit.GetStrand () == eNa_strand_minus;
            for (TSeqPos pos = neg ? ito : ifrom;
                (neg ? (pos - offset >= ifrom) : (pos + offset <= ito))  &&
                idx != prot_seq.length(); (neg ? pos -= 3 : pos += 3), ++idx)
            {
                bool mismatch = (prot_seq[idx] != trans_seq[idx]);
                seq_start = neg ? pos - offset : pos + offset;

                // skip out of visible range intervals
                if ( !m_Context->IsInVisibleRange(seq_start) )
                    continue;

                basesIN [0] = prot_seq[idx];
                basesOUT[0] = trans_seq[idx];

                if (mismatch) {
                    gl.ColorC(m_CdsConfig->m_SeqProtMismatch); // Mismatch
                } else {
                    gl.ColorC(m_CdsConfig->m_SeqProt);
                }

                // Draw protein sequence inside
                m_Context->TextOut(&m_CdsConfig->m_ProdFont, basesIN,
                    seq_start + 0.5, p_in_y, true);

                // draw translation sequence underneath the bar
                if (true/*IsSelected()*/) {
                    CRgbaColor col("dark blue");
                    gl.ColorC(col);
                    m_Context->TextOut(&m_CdsConfig->m_ProdFont, basesOUT,
                        seq_start + 0.5, p_out_y, true);

                    gl.ColorC(m_CdsConfig->m_SeqProtTrans);
                    try {
                        string main_seq, tmp_seq;
                        vec.GetSeqData(seq_start - 1, seq_start + 2, tmp_seq);

                        if (neg) {  // reverse sequence on negative strand
                            CSeqManip::Complement(tmp_seq, CSeqUtil::e_Iupacna,
                                0, tmp_seq.length(), main_seq);
                        } else {
                            main_seq = tmp_seq;
                        }

                        if (flip_strand) { // handle the flipped strand mode
                            reverse(main_seq.begin(), main_seq.end());
                        }

                        m_Context->TextOut(&m_CdsConfig->m_TransFont, main_seq.c_str(),
                            seq_start + 0.5, t_y, true);
                    }
                    catch (CException&) {
                        /// ignore exceptions - these are objmgr related
                        /// and indicate sequences that don't resolve
                    }
                }
            }

            // Use correct offset for the next interval
            int diff = neg ? seq_start - ifrom : ito - seq_start;
            if (diff == 2) {
                offset = 0;
            } else if (diff == 1) {
                offset = 1;
            } else if (diff == 0) {
                offset = 2;
            }
        } // for
    }
}


void CCdsGlyph::x_UpdateBoundingBox()
{
    CFeatGlyph::x_UpdateBoundingBox();
    if (m_Context->WillSeqLetterFit()  &&  true/*IsSelected()*/) {
        TModelUnit trans_full = m_CdsConfig->m_TransFont.GetMetric(
            IGlFont::eMetric_FullCharHeight);
        m_Height += trans_full;
        m_Height += m_CdsConfig->m_ProdFont.TextHeight() + kVertSpace;
    }
}


// This is exactly as it done in Sequin.
TSeqPos CCdsGlyph::x_GetProtOffset() const
{
    TSeqPos offset = 1; // translation offset
    const CCdregion& cdr = GetFeature().GetData().GetCdregion();
    if (cdr.IsSetFrame ()) {
        switch (cdr.GetFrame ()) {
        case CCdregion::eFrame_two :
            offset = 2;
            break;
        case CCdregion::eFrame_three :
            offset = 3;
            break;
        default :
            break;
        }
    }
    return offset;
}



END_NCBI_SCOPE
