/*  $Id: dbvar_track.cpp 27972 2013-05-03 20:34:20Z wuliangs $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Liangshou Wu
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>

#include <gui/widgets/seq_graphic/dbvar_track.hpp>

#include <gui/widgets/seq_graphic/seqgraphic_ds_manager.hpp>
#include <gui/widgets/seq_graphic/rendering_ctx.hpp>
#include <gui/widgets/seq_graphic/dbvar_feature_sorter.hpp>
#include <gui/widgets/seq_graphic/layout_track_proxy.hpp>

#include <gui/objutils/utils.hpp>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);


///////////////////////////////////////////////////////////////////////////////
///   CVarTrack
///////////////////////////////////////////////////////////////////////////////

CTrackTypeInfo CVarTrack::m_TypeInfo("dbvar_track",
                                     "structural variation track");

CVarTrack::CVarTrack(CSGFeatureDS* ds, CRenderingContext* r_cntx)
    : CFeatureTrack(ds, r_cntx)
{}


CVarTrack::~CVarTrack()
{}


const CTrackTypeInfo& CVarTrack::GetTypeInfo() const
{
    return m_TypeInfo;
}



///////////////////////////////////////////////////////////////////////////////
/// CVarTrackFactory
///////////////////////////////////////////////////////////////////////////////
ILayoutTrackFactory::TTrackMap
CVarTrackFactory::CreateTracks(SConstScopedObject& object,
                               ISGDataSourceContext* ds_context,
                               CRenderingContext* r_cntx,
                               const SExtraParams& params,
                               const TAnnotMetaDataList& src_annots) const
{
    TAnnotNameTitleMap annots;
    CSeqFeatData::ESubtype subtype = CSeqFeatData::eSubtype_variation_ref;

    if (params.m_SkipGenuineCheck  &&  !params.m_Annots.empty()) {
        ITERATE (SExtraParams::TAnnots, iter, params.m_Annots) {
            annots.insert(TAnnotNameTitleMap::value_type(*iter, ""));
        }
    } else {

        TSeqRange range = params.m_Range;
        if (range.Empty()) {
            range = r_cntx->GetVisSeqRange();
        }

        // collect non-NA tracks
        CIRef<ISGDataSource> pre_ds = ds_context->GetDS(
            typeid(CSGFeatureDSType).name(), object);
        CSGFeatureDS* seq_ds = dynamic_cast<CSGFeatureDS*>(pre_ds.GetPointer());
        seq_ds->SetDepth(params.m_Level);
        seq_ds->SetAdaptive(params.m_Adaptive);
        
        SAnnotSelector sel(CSeqUtils::GetAnnotSelector(params.m_Annots));
        sel.SetFeatSubtype(subtype);
        seq_ds->GetAnnotNames_var(sel, range, annots);
        
        // collect NA tracks
        if ( !src_annots.empty() ) {
            GetMatchedAnnots(src_annots, params.m_Annots, "ftable", "variation", annots);
        }
    }

    // create feature tracks
    TTrackMap tracks;
    ITERATE (TAnnotNameTitleMap, iter, annots) {
        CIRef<ISGDataSource> ds = ds_context->GetDS(
            typeid(CSGFeatureDSType).name(), object);
        CSGFeatureDS* seq_ds = dynamic_cast<CSGFeatureDS*>(ds.GetPointer());
        seq_ds->SetDepth(params.m_Level);
        seq_ds->SetAdaptive(params.m_Adaptive);
        // check the total levels of coverage graphs
        seq_ds->SetGraphLevels(iter->first);

        CRef<CVarTrack> feat_track(new CVarTrack(seq_ds, r_cntx));
        feat_track->SetAnnot(iter->first);
        feat_track->SetFeatSubtype(subtype);
        if ( !iter->second.empty() ) {
            feat_track->SetTitle(iter->second);
        }
        seq_ds->SetFilter(params.m_Filter);
        seq_ds->SetSortBy(params.m_SortBy);
        tracks[iter->first] = feat_track.GetPointer();
    }

    return tracks;
}


string CVarTrackFactory::GetExtensionIdentifier() const
{
    return CVarTrack::m_TypeInfo.GetId();
}


string CVarTrackFactory::GetExtensionLabel() const
{
    return CVarTrack::m_TypeInfo.GetDescr();
}


CRef<CTrackConfigSet>
CVarTrackFactory::GetSettings(CSeqGraphicConfig* /*g_conf*/,
                              const string& /*profile*/,
                              const TKeyValuePairs& settings,
                              const CTempTrackProxy* track_proxy) const
{
    CRef<CTrackConfigSet> config_set(new CTrackConfigSet);
    CRef<CTrackConfig> config(new CTrackConfig);
    config_set->Set().push_back(config);
    config->SetHelp() = GetThisTypeInfo().GetDescr();
    config->SetLegend_text("graphicalViewLegend.html#anchor_5");

    string rendering_str = CFeatureTrack::LinkedFeatValueToStr(CFeatureTrack::eFeat_Packed);

    ITERATE (TKeyValuePairs, iter, settings) {
        try {
            if (NStr::EqualNocase(iter->first, "Rendering")) {
                rendering_str = iter->second;
            }
        } catch (CException& e) {
            LOG_POST(Warning << "CVarTrackFactory::x_LoadSettings() "
                << e.GetMsg());
        }
    }

    string sort_by;
    if (track_proxy)  {
        sort_by = track_proxy->GetSortBy();
    }

    CRef<CChoice> choice = CTrackConfigUtils::CreateChoice(
        "Rendering", "Rendering options",
        rendering_str,
        "Rendering options for linked features.");

    choice->SetValues().push_back(
        CTrackConfigUtils::CreateChoiceItem(
        CFeatureTrack::LinkedFeatValueToStr(CFeatureTrack::eFeat_Default),
        "Show all",
        "All features, both parent and all child fetaures",
        "The black bar on top represents the variant region. Each bar blow represents \
        the supporting level variants used to define this region.<br>Red: A Deletion or Loss<br> \
        Green: A Gain or Insertion<br>Light Gray: An Inversion"));
    choice->SetValues().push_back(
        CTrackConfigUtils::CreateChoiceItem(
        CFeatureTrack::LinkedFeatValueToStr(CFeatureTrack::eFeat_Packed),
        "Show parent, Merge children",
        "Show parent on one line and all children merged below it.",
        "The black bar on top represents the variant region. The bar blow represents \
        all of the sample level variants merged onto a single line.<br>Red: A Deletion or Loss<br> \
        Green: A Gain or Insertion<br>Light Gray: An Inversion"));

    choice->SetValues().push_back(
        CTrackConfigUtils::CreateChoiceItem(
        CFeatureTrack::LinkedFeatValueToStr(CFeatureTrack::eFeat_ParentHidden),
        "Show children, not parent",
        "Show only the supporting variants (chidlren) not the variant region (parent)",
        "The bar represent supporting variants in this study.<br>Red: A Deletion or Loss<br>\
        Green: A Gain or Insertion<br>Light Gray: An Inversion"));

    choice->SetValues().push_back(
        CTrackConfigUtils::CreateChoiceItem(
        CFeatureTrack::LinkedFeatValueToStr(CFeatureTrack::eFeat_Hidden),
        "Show parent, not children ",
        "Show the variant region (parent) not the supporting variants (children)",
        "The black bar represents the variant region."));

    choice->SetValues().push_back(
        CTrackConfigUtils::CreateChoiceItem(
        "SingleLine",
        "Display on a single line",
        "Track is displayed, with all features collapsed on a single line.",
        "All the features are packed into one line."));
    choice->SetValues().push_back(
        CTrackConfigUtils::CreateChoiceItem(
        "Histogram",
        "Histogram",
        "Track is displayed, with all features displayed as a histogram.",
        "All the features are packed and shown as a Distribution histogram."));

    config->SetChoice_list().push_back(choice);

    CRef<CChoice> sortby_choice = CTrackConfigUtils::CreateChoice(
        "sort_by", "Sort variants by",
        sort_by,
        "Sort variants according to a selected criterion.");

    sortby_choice->SetValues().push_back(
        CTrackConfigUtils::CreateChoiceItem(
        "", "No sorting", "Don't sort variants", ""));

    IObjectSorter::SSorterDescriptor descr = CDbvarQualitySorter::GetSorterDescr();

    sortby_choice->SetValues().push_back(
        CTrackConfigUtils::CreateChoiceItem(
        descr.m_Name, descr.m_DisplayName, descr.m_Descr, ""));

    descr = CDbvarClinicalAsstSorter::GetSorterDescr();
    sortby_choice->SetValues().push_back(
        CTrackConfigUtils::CreateChoiceItem(
        descr.m_Name, descr.m_DisplayName, descr.m_Descr, ""));

    descr = CDbvarPilotSorter::GetSorterDescr();
    sortby_choice->SetValues().push_back(
        CTrackConfigUtils::CreateChoiceItem(
        descr.m_Name, descr.m_DisplayName, descr.m_Descr, ""));

    descr = CDbvarSamplesetTypeSorter::GetSorterDescr();
    sortby_choice->SetValues().push_back(
        CTrackConfigUtils::CreateChoiceItem(
        descr.m_Name, descr.m_DisplayName, descr.m_Descr, ""));

    descr = CDbvarValidationSorter::GetSorterDescr();
    sortby_choice->SetValues().push_back(
        CTrackConfigUtils::CreateChoiceItem(
        descr.m_Name, descr.m_DisplayName, descr.m_Descr, ""));

    config->SetChoice_list().push_back(sortby_choice);

    return config_set;
}


END_NCBI_SCOPE
