/*  $Id: feature_glyph.cpp 27939 2013-04-26 15:54:25Z wuliangs $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Mike DiCuccio, Liangshou Wu
 *
 * File Description:
 *    CFeatGlyph -- utility class to arrange CSeq_feat objects in hierarchical
 *                (tree) order.
 */

#include <ncbi_pch.hpp>
#include <gui/widgets/seq_graphic/feature_glyph.hpp>
#include <gui/widgets/seq_graphic/rendering_ctx.hpp>
#include <gui/widgets/seq_graphic/seqgraphic_utils.hpp>
#include <gui/widgets/gl/ruler.hpp>
#include <gui/objutils/obj_fingerprint.hpp>
#include <gui/objutils/utils.hpp>
#include <gui/objutils/gui_object_info_seq_feat.hpp>
#include <gui/opengl/gltexturefont.hpp>
#include <gui/opengl/glrender.hpp>

#include <objects/seqfeat/Variation_ref.hpp>
#include <objects/seqfeat/Variation_inst.hpp>
#include <objects/seqfeat/Delta_item.hpp>
#include <objects/seqfeat/Feat_id.hpp>
#include <objects/general/Dbtag.hpp>
#include <objects/general/Object_id.hpp>
#include <objects/general/User_object.hpp>
#include <objects/seqloc/Seq_interval.hpp>
#include <objects/seq/Seq_literal.hpp>
#include <objects/seqloc/Giimport_id.hpp>

#include <objmgr/util/sequence.hpp>
#include <objtools/alnmgr/aln_converters.hpp>
#include <math.h>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

/// feature ruler height.
static const int kRulerHeight = 16;

/// extra space for side labeling.
static const int kLabelSpace = 5;

/// vertical space between elements.
static const int kVertSpace = 2;

//
// simple functor for sorting CRange<> objects
//
template <class T>
struct SRangeSorter
{
    bool operator() (const CRange<T>& r0, const CRange<T>& r1) const
    {
        if (r0.GetFrom() < r1.GetFrom()) {
            return true;
        }
        if (r0.GetFrom() > r1.GetFrom()) {
            return false;
        }
        if (r0.GetTo() < r1.GetTo()) {
            return true;
        }
        return false;
    }
};


static bool s_IsSV(const CVariation_ref& var)
{
    if (var.IsSetId()  &&  var.GetId().GetTag().IsStr()) {
        const string& id = var.GetId().GetTag().GetStr();
        if (id.length() > 3  &&  id[1] == 's'  &&  id[2] == 'v') {
            return true;
        }
    }
    return false;
}


/* not used currently

static bool s_IsSSV(const CVariation_ref& var)
{
    if (var.IsSetId()  &&  var.GetId().GetTag().IsStr()) {
        const string& id = var.GetId().GetTag().GetStr();
        if (id.length() > 4  &&  id[1] == 's'  &&
            id[2] == 's'  &&  id[3] == 'v') {
            return true;
        }
    }
    return false;
}
*/


//
// CFeatGlyph::CFeatGlyph()
//
CFeatGlyph::CFeatGlyph(const CMappedFeat& f)
    : m_Feature(f)
    , m_Location(&f.GetLocation())
    , m_HideLabel(false)
    , m_ProjectedFeat(false)
{}

CFeatGlyph::CFeatGlyph(const CMappedFeat& f, const CSeq_loc& loc)
    : m_Feature(f)
    , m_Location(&loc)
    , m_HideLabel(false)
    , m_ProjectedFeat(false)
{}


bool CFeatGlyph::LessBySeqPos(const CSeqGlyph& obj) const
{
    const CFeatGlyph* feat = dynamic_cast<const CFeatGlyph*>(&obj);
    if (feat) {
        // two features - we sort in feature order
        try {
            SFeatByPos comp;
            return (comp(*this, *feat));
        } catch (CException&) {
            // guard against the case that two features have
            // differnt forms of seq-id.
        }
    }

    // otherwise, using default, compare by ranges
    return CSeqGlyph::LessBySeqPos(obj);
}


bool CFeatGlyph::NeedTooltip(const TModelPoint& /*p*/, string& /*tt*/, string& t_title) const
{
    return true;
}


void CFeatGlyph::GetTooltip(const TModelPoint& p, string& tt, string& t_title) const
{
    CScope& scope = GetMappedFeature().GetAnnot().GetScope();
    SConstScopedObject scoped_obj(GetMappedFeature().GetMappedFeature(), scope);
    CIRef<IGuiObjectInfo> gui_info(
        CreateObjectInterface<IGuiObjectInfo>(scoped_obj, NULL));

    if ( !gui_info ) return;

    CGuiObjectInfoSeq_feat* gui_info_feat =
        dynamic_cast<CGuiObjectInfoSeq_feat*>(gui_info.GetPointer());
    if (gui_info_feat) {
        gui_info_feat->SetLocation(*m_Location);
    }

    TSeqPos at_p = (TSeqPos)-1;
    if (p.X() >= 0) {
        at_p = (TSeqPos)p.X();
    }

    gui_info->GetToolTip(tt, t_title, at_p);

    string links;
    gui_info->GetLinks(links, false);
    if ( !links.empty() ) {
        tt += "<br>&nbsp";
        tt += links;
    }
}


TSeqRange CFeatGlyph::GetRange(void) const
{
    TSeqRange range;
    try { // watch out for mix loc with multiple seq-ids
        range = m_Location->GetTotalRange();
    } catch (CException&) {
        // get range from intervals
        ITERATE(TIntervals,  it,  m_Intervals) {
            range.CombineWith(*it);
        }
    }

    if (IsDbVar(GetFeature())) {
        if (GetLocation().IsInt()) {
            if (GetLocation().GetInt().IsSetFuzz_from()  &&
                GetLocation().GetInt().GetFuzz_from().IsRange()) {
                const CInt_fuzz::C_Range& f_range =
                    GetLocation().GetInt().GetFuzz_from().GetRange();
                range.SetFrom(f_range.GetMin());
            }
            if (GetLocation().GetInt().IsSetFuzz_to()  &&
                GetLocation().GetInt().GetFuzz_to().IsRange()) {
                const CInt_fuzz::C_Range& t_range =
                    GetLocation().GetInt().GetFuzz_to().GetRange();
                range.SetTo(t_range.GetMax());
            }
        }
    }

    return range;
}


void CFeatGlyph::GetHTMLActiveAreas(TAreaVector* p_areas) const
{
    CHTMLActiveArea area;
    CSeqGlyph::x_InitHTMLActiveArea(area);
    area.m_PositiveStrand = sequence::GetStrand(GetLocation()) != eNa_strand_minus;
    bool is_db_xref = GetFeature().IsSetDbxref();

    switch (GetFeature().GetData().GetSubtype()) {
        case CSeqFeatData::eSubtype_cdregion:
        case CSeqFeatData::eSubtype_mRNA:
            if (GetFeature().IsSetProduct()) {
                const CSeq_loc& product = GetFeature().GetProduct();
                CSeq_id_Handle shdl = sequence::GetId(*product.GetId(),
                    m_Context->GetScope(), sequence::eGetId_ForceGi);
                if (shdl) {
                    area.m_DB_ID =  shdl.AsString();
                }
            } else {
                try {
                    const CSeq_id& id = sequence::GetId(GetLocation(),
                        &m_Context->GetScope());
                    id.GetLabel(&area.m_DB_ID, CSeq_id::eContent);
                } catch (CException&) {
                    /// ignore it
                }
            }
            break;

        case CSeqFeatData::eSubtype_gene:
            if (is_db_xref) {
                CConstRef<CDbtag> tag = GetFeature().GetNamedDbxref("GeneID");
                if (tag.NotEmpty()) {
                    area.m_DB_Name = "gene";
                    area.m_DB_ID = tag->GetUrl();
                }
            }
            break;

        case CSeqFeatData::eSubtype_STS:
            if (is_db_xref) {
                CConstRef<CDbtag> tag = GetFeature().GetNamedDbxref("UniSTS");
                if (tag.NotEmpty()) {
                    area.m_DB_Name = "UniSTS";
                    area.m_DB_ID = tag->GetUrl();
                }
            }
            break;

        case CSeqFeatData::eSubtype_variation:
            if (is_db_xref) {
                CConstRef<CDbtag> tag = GetFeature().GetNamedDbxref("dbSNP");
                if (tag.NotEmpty()) {
                    area.m_DB_Name = "SNP";
                    area.m_DB_ID = tag->GetUrl();
                }
            }
            break;

        default:
            break;
    }


    // for features with editable flag set, we also report feature-id
    if (GetFeature().IsSetExts()) {
        const CSeq_feat::TExts& exts = GetFeature().GetExts();
        ITERATE (CSeq_feat::TExts, iter, exts) {
            if ( (*iter)->GetType().IsStr()  &&
                 (*iter)->GetType().GetStr() == "Editing"  &&
                 (*iter)->GetFieldRef("Editable")  &&
                 (*iter)->GetFieldRef("Editable")->GetData().GetBool()  &&
                 ( GetFeature().CanGetId()  ||
                   (GetFeature().CanGetIds()  &&
                    !GetFeature().GetIds().empty()) ) ) {
                area.m_Flags |= CHTMLActiveArea::fEditable;
                area.m_ID = x_GetFeatureId();
                if((*iter)->GetFieldRef("Ignorable")  &&
                   (*iter)->GetFieldRef("Ignorable")->GetData().GetBool()) {
                    area.m_Flags |= CHTMLActiveArea::fIgnorable;
                }
            }
        }
    }


    area.m_Signature = GetSignature();
    p_areas->push_back(area);
}


bool CFeatGlyph::IsClickable() const
{
    return true;
}


void CFeatGlyph::SetSelected(bool f)
{
    if (f != IsSelected()) {
        CSeqGlyph::SetSelected(f);
        // The object size may be changed for following two cases:
        //  - For feature with ruler enabled, we always show
        //    feature ruler when it is selected
        //  - For feature with top labeling, but somehow the label
        //    is forced to hide, we will show the label when it is
        //    selected.
        if ((m_Config->m_ShowFeatRuler)  ||
            (m_HideLabel  &&
            m_Config->m_LabelPos == CFeatureParams::ePos_Above)) {
            x_OnLayoutChanged();
        }
    }
}


void CFeatGlyph::SetHideLabel(bool b)
{
    m_HideLabel = b;
}


bool CFeatGlyph::HasSideLabel() const
{
    return !m_HideLabel  &&  m_Config->m_LabelPos == CFeatureParams::ePos_Side;
}


const objects::CSeq_loc& CFeatGlyph::GetLocation(void) const
{
    return *m_Location;
}


CConstRef<CObject> CFeatGlyph::GetObject(TSeqPos) const
{
    return CConstRef<CObject>(&m_Feature.GetOriginalFeature());
}


void  CFeatGlyph::GetObjects(vector<CConstRef<CObject> >& objs) const
{
    objs.push_back( CConstRef<CObject>(&m_Feature.GetOriginalFeature()) );
}


bool CFeatGlyph::HasObject(CConstRef<CObject> obj) const
{
    return &m_Feature.GetOriginalFeature() == obj.GetPointer();
}


string CFeatGlyph::GetSignature() const
{
    string sig = kEmptyStr;
    if (m_ProjectedFeat) {
        try {
            sig = CObjFingerprint::GetFeatSignature(
                GetFeature(), GetFeature().GetLocation(),
                &m_Context->GetScope(), m_Feature.GetAnnot());
        } catch (CException& e) {
            // failed to generated the signature.
            // it is likely the feature location contains multiple seq-ids.
            LOG_POST(Warning <<
                "CFeatGlyph::GetSignature() failed to generate signature: " <<
                e.GetMsg());
        }
    } else {
        sig = CObjFingerprint::GetFeatSignature(
            GetMappedFeature().GetMappedFeature(), GetLocation(),
            &m_Context->GetScope(), m_Feature.GetAnnot());
    }
    return sig;
}


const CFeatGlyph::TIntervals& CFeatGlyph::GetIntervals(void) const
{
    return m_Intervals;
}


TModelUnit CFeatGlyph::GetBarCenter() const
{
    TModelUnit y = m_Config->GetBarHeight(
        m_Context->IsOverviewMode()  &&  m_HideLabel) * 0.5;
    if (x_ShowLabel()  &&  m_Config->m_LabelPos == CFeatureParams::ePos_Above) {
        y += m_Config->m_LabelFont.TextHeight() + kVertSpace + 1;
    }

    // Ruler is always above a feature bar
    if (x_ShowRuler()) {
        y += kRulerHeight + kVertSpace;
    }

    return y;
}


void CFeatGlyph::GetLabel(string& label, CLabel::ELabelType type) const
{
    CScope& scope = GetMappedFeature().GetAnnot().GetScope();
    CLabel::GetLabel(GetFeature(), &label, type, &scope);
    if (type == CLabel::eContent) {
        NStr::ReplaceInPlace(label, "/standard_name=", "");
    }
}


static int s_GetLiteralLength(const CVariation_ref& var)
{
    int len = -1;
    if (var.GetData().IsInstance()  &&  var.GetData().GetInstance().IsSetDelta()) {
        const CVariation_inst::TDelta& delta = var.GetData().GetInstance().GetDelta();
        ITERATE (CVariation_inst::TDelta, iter, delta) {
            if ((*iter)->IsSetSeq()  &&  (*iter)->GetSeq().IsLiteral()) {
                len = (*iter)->GetSeq().GetLiteral().GetLength();
                break;
            }
        }
    }
    return len;
}


static int s_GetChildNum(const objects::CSeq_feat& feat)
{
    string num_str = feat.GetNamedQual("Child Count");
    if ( !num_str.empty() ) {
        try {
            return NStr::StringToInt(num_str);
        } catch (CException&) {
            // ingore it
        }
    }
    return -1;
}


size_t CFeatGlyph::GetCustomColorIdx(const CSeq_feat& feat)
{
    CCustomFeatureColor::EColorIdex idx = CCustomFeatureColor::eDefault;
    if (IsDbVar(feat)) {
        const CVariation_ref& var = feat.GetData().GetVariation();
        if (var.IsComplex()) {
            idx = CCustomFeatureColor::eComplex;
        } else if (var.IsInsertion()) {
            idx = CCustomFeatureColor::eInsertion;
        } else if (var.IsInversion()) {
            idx = CCustomFeatureColor::eInversion;
        } else if (var.IsEversion()) {
            idx = CCustomFeatureColor::eEverted;
        } else if (var.IsTranslocation()) {
            idx = CCustomFeatureColor::eTranschr;
        } else if (var.IsGain()) {
            idx = CCustomFeatureColor::eGain;
        } else if (var.IsLoss()  ||  var.IsDeletion()) {
            idx = CCustomFeatureColor::eLoss;
        } else if (var.IsCNV()) {
            idx = CCustomFeatureColor::eCNV;
        } else {
            idx = CCustomFeatureColor::eUnknown;
        }
    } else {
        const string& var_type = feat.GetNamedQual("Var_type");
        if ( !var_type.empty() ) {
            if (NStr::EqualNocase(var_type, "Deletion")  ||
                NStr::EqualNocase(var_type, "Loss")) {
                    idx = CCustomFeatureColor::eLoss;
            } else if (NStr::EqualNocase(var_type, "Insertion")  ||
                NStr::EqualNocase(var_type, "Duplication")) {
                    idx = CCustomFeatureColor::eInsertion;
            } else if (NStr::EqualNocase(var_type, "Gain")) {
                idx = CCustomFeatureColor::eGain;
            } else if (NStr::EqualNocase(var_type, "Inversion")) {
                idx = CCustomFeatureColor::eInversion;
            }
        } else {
            const string& identity = feat.GetNamedQual("identity");
            if ( !identity.empty() ) {
                try {
                    double identity_num = NStr::StringToDouble(identity);
                    if (identity_num > 99.0) {
                        idx = CCustomFeatureColor::eHighIdentity;
                    } else if (identity_num > 98.0) {
                        idx = CCustomFeatureColor::eMidIdentity;
                    } else if (identity_num > 90.0) {
                        idx = CCustomFeatureColor::eLowIdentity;
                    }
                } catch (CException&) {
                    /// ignore
                }
            }
        }
    }

    return (size_t)idx;
}


bool CFeatGlyph::IsDbVar(const objects::CSeq_feat& feat)
{
    if (feat.GetData().Which() == CSeqFeatData::e_Variation) {
        const CVariation_ref& sv = feat.GetData().GetVariation();
        if (sv.CanGetId()  &&  sv.GetId().GetDb() == "dbVar") {
            return true;
        }
    }
    return false;
}


void CFeatGlyph::x_Draw() const
{
    _ASSERT(m_Context);
    
    CGlRender& gl = GetGl();

    if (m_Height == 0  ||  m_Context->IntersectVisible(this).Empty()) {
        return;
    }

    // for debuging
    /*
    CScope& scope = GetMappedFeature().GetAnnot().GetScope();
    string sig = CObjFingerprint::GetFeatSignature(GetFeature(), &scope);
    int subtype;
    CSeq_id_Handle id;
    TSeqPos f, t;
    CObjFingerprint::EObjectType type;
    Uint4   fingerprint;
    CObjFingerprint::ParseSignature(sig, id, f, t, type, subtype, fingerprint);
    cout << ", From: " << f
         << ", To: " << t
         << ", Subtype: " << subtype << endl;
    */

    TModelUnit base = GetTop();
    // Draw a shading background to indicate exception
    if (CSeqUtils::IsException(GetFeature())) {
        gl.Color4f(0.3f, 0.21f, 0.07f, 0.25f);
        TModelUnit padding_y = 1.0;
        TModelUnit padding_x = m_Context->ScreenToSeq(padding_y);
        m_Context->DrawQuad(GetLeft() - padding_x, GetTop() - padding_y,
            GetRight() + padding_x, GetBottom() + padding_y);
    }

    bool label_first = x_LabelFirst();
    if (label_first) {
        x_DrawLabel(base);
    }

    if (x_ShowRuler()) {
        try {
            x_DrawRuler(base);
            base += kRulerHeight + kVertSpace;
        } catch (CException&) {
            // we may have problems on setting up the required alignment
            // mapping for features with mix location (multip seq-ids or
            // mix strands from one location)
        }
    }

    TModelUnit bar_height = x_GetBarHeight();
    base += bar_height * 0.5;   // bar center
    if (IsDbVar(GetFeature())) {
        x_DrawFeatureBar_sv(base);
    } else {
        x_DrawFeatureBar(base);
    }

    if ( !label_first ) {
        x_DrawLabel(base);
    }

    if (IsSelected()) {
        m_Context->DrawSelection(GetModelRect());
    }
}


void CFeatGlyph::x_UpdateBoundingBox()
{
    _ASSERT(m_Context);

    TSeqRange range = GetRange();
    m_Height = x_GetBarHeight();

    m_Width  = range.GetLength();
    m_Pos.m_X = range.GetFrom();

    // preserve space for undefined breakpoint indicator
    // for structural variants
    if (IsDbVar(GetFeature())) {
        TModelUnit tri_width = m_Context->ScreenToSeq(m_Height * 0.5);
        TModelUnit extra = 0.0;
        if (x_GetUndefinedBp_from(GetLocation()) != eBp_Unknown) {
            m_Pos.m_X -= tri_width;
            extra += tri_width;
        }
        if (x_GetUndefinedBp_to(GetLocation()) != eBp_Unknown) {
            extra += tri_width;
        }
        m_Width += extra;
    }

    // we preserve label space if it fits
    if (x_ShowLabel()) {
        const CGlTextureFont& font = m_Config->m_LabelFont;
        if (m_Config->m_LabelPos == CFeatureParams::ePos_Side) {
            string label;
            GetLabel(label, CLabel::eContent);
            TModelUnit text_w =
                min(font.TextWidth(label.c_str()), m_Context->GetMaxLabelWidth(font));
            text_w = m_Context->ScreenToSeq(text_w + kLabelSpace);
            m_Width += text_w;
            TModelUnit min_text_w = m_Context->GetMinLabelWidth();

            if (sequence::GetStrand(GetLocation()) == eNa_strand_minus) {
                CRef<CSGSequenceDS> seq_ds = m_Context->GetSeqDS();

                TModelRange vis_r = m_Context->IntersectVisible(this);
                if (m_Pos.m_X + m_Width > seq_ds->GetSequenceLength()  &&
                    m_Width > min_text_w  &&  vis_r.GetLength() >= min_text_w) {
                    // can't fit, move label above the bar
                    m_Height += font.TextHeight() + kVertSpace + 1;
                }
            } else {
                m_Pos.m_X -= text_w;
                TModelRange vis_r = m_Context->IntersectVisible(this);

                if (m_Pos.m_X < 0) {
                    if (vis_r.GetLength() >= min_text_w) {
                        // No enough space on the side, move label above the bar
                        m_Height += font.TextHeight() + kVertSpace + 1;
                    } else {
                        // Can't even show the minimal lenght of label
                        m_Width += m_Pos.m_X;
                        m_Pos.m_X = 0.0;
                    }
                }
            }
        } else if (m_Config->m_LabelPos == CFeatureParams::ePos_Above) {
            m_Height += font.TextHeight() + kVertSpace + 1;
        }
    }

    // Preserve space for feature ruler
    if (x_ShowRuler()) {
        m_Height += kRulerHeight + kVertSpace;
    }
}


void CFeatGlyph::x_DrawLabel(TModelUnit& base) const
{
    if ( !x_ShowLabel() ) return;

    CGlRender& gl = GetGl();

    TModelRange vis_r = m_Context->IntersectVisible(this);
    const CGlTextureFont& font = m_Config->m_LabelFont;
    TModelUnit font_h = font.TextHeight();

    if (m_Config->m_LabelPos == CFeatureParams::ePos_Side) {
        string fl_content;
        GetLabel(fl_content, CLabel::eContent);
        if (vis_r.GetLength() < m_Context->GetMinLabelWidth()) {
            return;
        }
        TModelUnit object_w = m_Context->SeqToScreen(vis_r.GetLength());

        TModelUnit label_w = font.TextWidth(fl_content.c_str());
        TModelUnit max_w = min(object_w, m_Context->GetMaxLabelWidth(font));
        if (label_w > max_w) {
            label_w = max_w;
            fl_content = font.Truncate(fl_content.c_str(), label_w);
        }
        label_w = m_Context->ScreenToSeq(label_w);
        TModelUnit label_x = 0.0;
        TModelUnit label_b = base;
        if (x_LabelFirst()) {
            label_b += font_h + 1;
            base += font_h + kVertSpace + 1;
        } else {
            label_b += font_h * 0.5;
        }
        if (sequence::GetStrand(GetLocation()) != eNa_strand_minus) {
            label_x = vis_r.GetFrom();
            if (vis_r.GetFrom() > GetLeft()) {
                gl.ColorC(m_Config->m_bgColor);
                m_Context->DrawBackground(
                    TModelRect(label_x, label_b + 1, label_x + label_w, label_b - font_h - 1),
                    0);
            }
        } else {
            label_x = vis_r.GetTo() - label_w;
            if (vis_r.GetTo() < GetRight()) {
                gl.ColorC(m_Config->m_bgColor);
                m_Context->DrawBackground(
                    TModelRect(label_x, label_b + 1, vis_r.GetTo(), label_b - font_h - 1),
                    0);
            }
        }
        gl.ColorC(IsSelected() ?
            m_Context->GetSelLabelColor() : m_Config->m_LabelColor);
        m_Context->TextOut(&font, fl_content.c_str(), label_x,
            label_b - 1.0, false, true);

    } else if (m_Context->WillLabelFit(vis_r)) {
        string fl_out;
        string fl_type;
        GetLabel(fl_type, CLabel::eType);
        TModelUnit lw_type =
            m_Context->ScreenToSeq(font.TextWidth(fl_type.c_str()));

        if (vis_r.GetLength() > lw_type * 4) { // 4 widths of eType label width
            GetLabel(fl_out, CLabel::eUserTypeAndContent);
        } else {
            GetLabel(fl_out, CLabel::eContent);
        }
        fl_out = font.Truncate(
            fl_out.c_str(), m_Context->SeqToScreen(vis_r.GetLength()));

        if (IsSelected()) {
            gl.ColorC(m_Context->GetSelLabelColor());
        }  else if (m_Config->m_LabelPos == CFeatureParams::ePos_Inside) {
            gl.ColorC(m_Config->m_fgColor.ContrastingColor());
        } else {
            gl.ColorC(m_Config->m_LabelColor);
        }

        TModelUnit xM = vis_r.GetFrom() + vis_r.GetLength() * 0.5;
        TModelUnit yM = base;
        if (m_Config->m_LabelPos == CFeatureParams::ePos_Above) {
            base += font_h;
            yM = base;
            base += kVertSpace + 1;
        } else {
            yM += font_h * 0.5;
        }

        m_Context->TextOut(&font, fl_out.c_str(), xM, yM, true, true);
    }
}


void CFeatGlyph::x_DrawRuler(TModelUnit& base) const
{
    bool horz = m_Context->IsHorizontal();
    bool flip_strand = m_Context->IsFlippedStrand();

    CGlRender& gl = GetGl();

    CRuler ruler_panel;
    ruler_panel.SetFont(CGlTextureFont::eFontFace_Helvetica_Bold, 8);
    ruler_panel.SetGeometryParam (CRuler::eMinorTickHeight, 1);
    ruler_panel.SetGeometryParam (CRuler::eMajorTickHeight, 3);
    ruler_panel.SetGeometryParam (CRuler::eLabelTickHeight, 3);
    ruler_panel.SetGeometryParam (CRuler::eOppMinorTickHeight, 0);
    ruler_panel.SetGeometryParam (CRuler::eOppMajorTickHeight, 0);
    ruler_panel.SetGeometryParam (CRuler::eOppLabelTickHeight, 2);

    ruler_panel.SetHorizontal(horz, horz ? CRuler::eTop : CRuler::eRight);
    ruler_panel.SetDisplayOptions(0); // don't show background
    //ruler_panel.SetColor(CRuler::eBackground, CRgbaColor("white"));

    CRef<CPairwiseAln> aln;
    if (GetFeature().IsSetProduct()) {
        const CSeq_loc& loc1 = GetLocation();
        const CSeq_loc& loc2 = GetFeature().GetProduct();
        // workaround for mix-loc with multiple seq-ids
        // or mixed strands.  Remove this when AM-25 is fixed
        if ( !loc1.GetId()  ||  !loc2.GetId()  ||
            loc1.GetStrand() == eNa_strand_other  ||
            loc2.GetStrand() == eNa_strand_other) {
            return;
        }

        TAlnSeqIdIRef id1(new CAlnSeqId(*loc1.GetId()));
        TAlnSeqIdIRef id2(new CAlnSeqId(*loc2.GetId()));

        // Sequence type is hardcoded, needs to be fixed:
        id1->SetBaseWidth(1); // location is on a nucleotide
        id2->SetBaseWidth(3); // product is on a protein

        // Do we really need to allow overlap? or data are incorrect?
        // Example feature: NP_057849.4 on NC_001802.1
        aln.Reset(new CPairwiseAln(id1, id2,
            CPairwiseAln::fAllowOverlap | CPairwiseAln::fKeepNormalized));
        ConvertSeqLocsToPairwiseAln(*aln, loc1, loc2,
            CAlnUserOptions::eBothDirections);
    } else {
        const CSeq_loc& loc1 = GetLocation();
        if ( !loc1.GetId()  ||  loc1.GetStrand() == eNa_strand_other) {
            return;
        }

        TAlnSeqIdIRef id1(new CAlnSeqId(*loc1.GetId()));

        CSeq_loc loc2;
        loc2.SetWhole();
        loc2.SetId(*loc1.GetId());

        // Sequence type is hardcoded, needs to be fixed:
        id1->SetBaseWidth(1); // location is on a nucleotide

        aln.Reset(new CPairwiseAln(id1, id1,
            CPairwiseAln::fAllowOverlap | CPairwiseAln::fKeepNormalized));
        ConvertSeqLocsToPairwiseAln(*aln, loc1, loc2, CAlnUserOptions::eBothDirections);
    }

    // revesre collention for filled strand
    CRef<CSGSequenceDS> seq_ds = m_Context->GetSeqDS();
    TSeqPos seq_length = seq_ds->GetSequenceLength();
    if (flip_strand) {
        ITERATE(CPairwiseAln, it, *aln)  {
            const CPairwiseAln::TAlnRng& r = *it;
            CPairwiseAln::TAlnRng* r1 = const_cast<CPairwiseAln::TAlnRng*>(&r);
            r1->SetFirstFrom(seq_length-r.GetFirstToOpen());
            r1->SetReversed( !r.IsReversed() );
        }
        aln->Sort();
    }
    ruler_panel.SetMapping(*aln);

    CGlPane* pane = m_Context->GetGlPane();
    gl.PushMatrix();
    gl.LoadIdentity();
    pane->Close();
    CGlPane RP(*pane);

    TModelUnit ruler_y = base - GetTop(), ruler_x = 0.0;
    x_Local2World(ruler_x, ruler_y);
    TModelUnit base_top = pane->ProjectY(ruler_y);
    TVPRect vpt = pane->GetViewport();
    vpt.SetTop((int)floor(base_top));
    vpt.SetBottom((int)ceil(base_top - kRulerHeight));
    RP.SetViewport(vpt);

    TModelRect rcV = pane->GetVisibleRect();
    if (flip_strand) {
        rcV.SetLeft(seq_length - rcV.Left());
        rcV.SetRight(seq_length - rcV.Right());
    }
    rcV.SetBottom(0);
    rcV.SetTop(kRulerHeight);
    RP.SetVisibleRect(rcV);

    ruler_panel.Render(RP);
    pane->OpenOrtho();
    gl.PopMatrix();
}


void CFeatGlyph::x_DrawFeatureBar(TModelUnit& base) const
{
    CGlRender& gl = GetGl();

    TModelUnit bar_height = m_Config->m_BarHeight;
    if (m_Context->IsOverviewMode()  &&  m_HideLabel) {
        bar_height = floor(bar_height * m_Config->m_OverviewFactor);
    }
    TModelUnit head_height = bar_height * m_Config->m_HeadHeight;
    TModelUnit tail_height = bar_height * m_Config->m_TailHeight;
    TModelUnit head_size = m_Context->ScreenToSeq(head_height);
    TModelUnit tail_size = m_Context->ScreenToSeq(tail_height);

    TModelUnit line_ym = base;
    TModelUnit line_y1 = line_ym - bar_height * 0.5f;
    TModelUnit line_y2 = line_ym + bar_height * 0.5f;

    bool neg_strand = (sequence::GetStrand(GetLocation()) == eNa_strand_minus);

    EFeatureParts head_style = eHead_No;
    EFeatureParts tail_style = eTail_No;

    CRgbaColor color;
    if ( !GetCustomColor(color) ) {
        color = m_Config->m_fgColor;
    }

    bool is_pseudo = CSeqUtils::IsPseudoFeature(GetFeature());
    if (is_pseudo) {
        color.SetAlpha(0.8f);
    }
    CRgbaColor color_lite = color;
    color_lite.Lighten(0.2f);

    // distance among strand indicators
    TModelUnit pix_size = m_Context->ScreenToSeq(1.0);
    TModelUnit apart = 100.0 * pix_size;    // 100 pixels on screen

    // Draw feature bar
    TSeqRange range = GetRange();
    TSeqPos from = range.GetFrom();
    TSeqPos to = range.GetTo();
    if (range.GetLength() > pix_size * 2) {  // At least 2 pixel
        TModelUnit prev_to_x = 0.0f;
        TModelUnit prev_from_x = 0.0f;

        bool first_pass = true;
        // draw exons and introns

        CSeq_loc_CI iter(GetLocation());
        for ( ;  iter;  ++iter) {
            CConstRef<CSeq_loc> curr_loc = iter.GetRangeAsSeq_loc();
            TSeqRange curr = iter.GetRange();
            TModelUnit f = curr.GetFrom();
            TModelUnit t = curr.GetTo();

            // adjust the size of exons to include style elements
            // such as head and tail pieces. Left (from)  or right (to) sides
            // are adjected according to strand
            // 1) adjust head
            bool fit      = head_size * 1.5 < fabs(t - f);
            bool adj_to   = !neg_strand  &&  t == to  &&  fit;
            bool adj_from = neg_strand  &&  f == from  &&  fit;

            if (adj_from || adj_to) {
                head_style = eNeedHead;
            }

            switch (m_Config->m_HeadStyle) {
            case CFeatureParams::eHead_Arrow:
            case CFeatureParams::eHead_NeckedTriangle:
                if (adj_to) {
                    t = to - head_size * 1.5;
                }
                if (adj_from) {
                    f = from + head_size * 1.5;
                }
                break;
            case CFeatureParams::eHead_Triangle:
                if (adj_to) {
                    t = to - head_size;
                }
                if (adj_from) {
                    f = from + head_size;
                }
                break;

            default:
                break;
            }

            // 2) adjust tail
            fit      = tail_size < fabs(t - f);
            adj_to   = neg_strand  &&  t == to  &&  fit;
            adj_from = !neg_strand  &&  f == from  &&  fit;

            if (adj_from || adj_to) {
                tail_style = eNeedTail;
            }

            switch (m_Config->m_TailStyle) {
            case CFeatureParams::eTail_Circle:
                if (adj_from) {
                    f = from + tail_size * 0.5f;
                }
                if (adj_to) {
                    t = to - tail_size * 0.5f;
                }
                break;
            case CFeatureParams::eTail_Square:
                if (adj_from) {
                    f = from + tail_size;
                }
                if (adj_to) {
                    t = to - tail_size;
                }
                break;
            case CFeatureParams::eTail_Fletched:
                if (adj_from) {
                    f = from + tail_size; //-1
                }
                if (adj_to) {
                    t = to - tail_size;
                }
                break;

            default:
                break;
            }

            // draw introns first
            TModelUnit from_x = neg_strand ? curr.GetTo() + 1 : curr.GetFrom();
            TModelUnit to_x = neg_strand ? curr.GetFrom() : curr.GetTo() + 1;

            if (!first_pass) {
                glPushAttrib(GL_LINE_BIT);
                switch (m_Config->m_LineStyle) {
                case CFeatureParams::eLine_Dashed:
                    gl.LineStipple(1, 0x0F0F);
                    gl.Enable(GL_LINE_STIPPLE);
                    break;
                case CFeatureParams::eLine_Dotted:
                    gl.LineStipple(1, 0x0202);
                    gl.Enable(GL_LINE_STIPPLE);
                    break;
                case CFeatureParams::eLine_DashDotted:
                    gl.LineStipple(1, 0x1C47);
                    gl.Enable(GL_LINE_STIPPLE);
                    break;
                case CFeatureParams::eLine_Solid:
                    break;
                }  // m_Config->m_LineStyle

                //TModelUnit off = (m_Config->m_LineThickness) * 0.5f;
                // In some cases, the intervals are not correctly ordered
                // for the feature. Maybe, it is a circular sequence, we
                // just can't handle it correctly.
                bool correctly_ordered = (neg_strand  ==  (curr.GetFrom() < prev_to_x));
                TModelUnit intron_f = correctly_ordered ? prev_to_x : prev_from_x;
                TModelUnit intron_t = correctly_ordered ? from_x : to_x;
                if (intron_f > intron_t) {
                    swap(intron_f, intron_t);
                }

                gl.LineWidth((float)(m_Config->m_LineWidth));
                gl.ColorC(color_lite);
                switch(m_Config->m_Connections) {
                case CFeatureParams::eCantedLine:
                    {{
                         TModelUnit middle_x =
                             intron_f + (intron_t - intron_f) * 0.5f;
                         m_Context->DrawLine(intron_f, line_ym, middle_x, line_y1);
                         m_Context->DrawLine(middle_x, line_y1, intron_t, line_ym);
                     }}
                    break;
                case CFeatureParams::eStraightLine:
                    m_Context->DrawLine(intron_f, line_ym, intron_t, line_ym);
                    break;
                case CFeatureParams::eFilledBox:
                    gl.Disable(GL_LINE_STIPPLE);
                    m_Context->Draw3DQuad(intron_f, line_y1, intron_t,
                        line_y2, color_lite);
                    break;
                }  // m_Config->m_Connections
                gl.Disable(GL_LINE_STIPPLE);
                gl.LineWidth(1.0f);
                glPopAttrib();

                if (m_Config->m_ShowStrandIndicator  &&
                    (!m_Context->IsOverviewMode()  ||  !m_HideLabel)) {
                    m_Context->DrawStrandIndicators(
                        TModelPoint(intron_f, line_y1),
                        intron_t - intron_f, apart, bar_height,
                        CRgbaColor(0.7f, 0.7f, 0.7f), neg_strand);
                }

            }  // connection lines (introns)

            if (is_pseudo) {
                CRgbaColor lcolor = color;
                lcolor.Lighten(0.4f);
                CRgbaColor dcolor = color;
                dcolor.Darken(0.7f);

                m_Context->DrawPseudoBar(f, line_y1, t + 1, line_y2, 
                    lcolor, dcolor);
            }

            // draw features as quads or hollow rectangles, as configured
            if(m_Config->m_BoxStyle == CFeatureParams::eBox_Filled) {
                m_Context->Draw3DQuad(f, line_y1, t, line_y2, color);
            } else {
				gl.ColorC(color);
				m_Context->DrawRect(f, line_y1, t+1, line_y2);
			}

            // draw strand indicators
            if (m_Config->m_ShowStrandIndicator  &&
                (!m_Context->IsOverviewMode()  ||  !m_HideLabel)) {

                bool exon_strand = (sequence::GetStrand(*curr_loc) == eNa_strand_minus);
                m_Context->DrawStrandIndicators(
                    TModelPoint(f, line_y1), fabs(t - f), apart, bar_height,
                    m_Config->m_fgColor.ContrastingColor(), exon_strand,
                    m_Config->m_LabelPos == CFeatureParams::ePos_Inside  &&
                    !m_HideLabel);
            }

            // draw partial loc indicators
            bool p_start = curr_loc->IsPartialStart(objects::eExtreme_Positional);
            bool p_stop = curr_loc->IsPartialStop(objects::eExtreme_Positional);
            if (p_start  ||  p_stop) {
                m_Context->DrawPartialBar(curr.GetFrom(), line_y1,
                    curr.GetToOpen(), line_y2, p_start, p_stop, true);
            }

            prev_to_x  = to_x;
            prev_from_x = from_x - 1;
            first_pass = false;
        }

        // Draw head
        if (head_style == eNeedHead) {
            gl.LineWidth((float)(m_Config->m_LineWidth + 0.1));

            switch (m_Config->m_HeadStyle) {
            case CFeatureParams::eHead_NeckedTriangle:
                m_Context->Draw3DTriangle(
                    TModelPoint(neg_strand?from:to+1, line_ym),
                    head_height, color, neg_strand, true); // with neck
                break;
            case CFeatureParams::eHead_Triangle:
                m_Context->Draw3DTriangle(
                    TModelPoint(neg_strand?from:to+1, line_ym),
                    head_height, color, neg_strand, false);
                break;
            case CFeatureParams::eHead_Arrow:
                gl.ColorC(color);
                m_Context->Draw3DArrow(
                    TModelPoint(neg_strand?from:to+1, line_ym),
                    head_height, neg_strand);
                break;

            default:
                break;
            }  // m_Config->m_HeadStyle
            gl.LineWidth(1.0f);
        }  // eNeedHead

        // Draw tail
        if (tail_style == eNeedTail) {
            gl.LineWidth((float) (m_Config->m_LineWidth + 0.1f));
            switch (m_Config->m_TailStyle) {
            case CFeatureParams::eTail_Square:
                m_Context->DrawSquare(
                    TModelPoint(neg_strand ? to + 1.0 : from, line_ym),
                    tail_height, color, neg_strand);
                break;
            case CFeatureParams::eTail_Circle:
                m_Context->DrawDisk(
                    TModelPoint(neg_strand ? to+1 : from, line_ym),
                    tail_height, color, neg_strand);
                break;
            case CFeatureParams::eTail_Fletched:
                gl.LineWidth(1.5f);
                m_Context->Draw3DFletch(
                    neg_strand ? to+1 : from, line_y1, line_y2,
                    tail_height, color, neg_strand);
                break;

            default:
                break;
            }  // m_Config->m_TailStyle
            gl.LineWidth(1.0f);
        }  // eNeedTail

        // draw partial feature indicators
        if (CSeqUtils::IsPartialFeature(GetFeature())) {
            m_Context->DrawPartialBar(from, line_y1, to + 1,
                line_y2, true, true, false);
        } else {
            bool p_start = CSeqUtils::IsPartialStart(GetFeature().GetLocation());
            bool p_stop = CSeqUtils::IsPartialStop(GetFeature().GetLocation());
            if (p_start || p_stop) {
                if (neg_strand) swap(p_start, p_stop);
                m_Context->DrawPartialBar(from, line_y1, to + 1,
                    line_y2, p_start, p_stop, true);
            }
        }
    } else {
        // bar is less then 2 pixel. Do not draw intervals.
        m_Context->Draw3DQuad(from, line_y1, to, line_y2, color, true);
    }
}


void CFeatGlyph::x_DrawFeatureBar_sv(TModelUnit& base) const
{
    CGlRender& gl = GetGl();

    TModelUnit bar_height = m_Config->m_BarHeight;
    if (m_Context->IsOverviewMode()  &&  m_HideLabel) {
        bar_height = floor(bar_height * m_Config->m_OverviewFactor);
    }
    TModelUnit line_ym = base;
    TModelUnit line_y1 = line_ym - bar_height * 0.5f;
    TModelUnit line_y2 = line_ym + bar_height * 0.5f;

    CRgbaColor color;
    if ( !GetCustomColor(color) ) {
        color = m_Config->m_fgColor;
    }
    CRgbaColor color_t = color;
    color_t.SetAlpha(0.35f);

    // Draw feature bar
    TSeqRange range = GetRange();
    TSeqPos f_out = range.GetFrom();
    TSeqPos t_out = range.GetTo();

    if (range.GetLength() > m_Context->ScreenToSeq(2.0)) {  // At least 2 pixel
        size_t idx = GetCustomColorIdx(GetFeature());
        int literal_len =
            s_GetLiteralLength(GetFeature().GetData().GetVariation());

        TModelUnit prev_to_x = 0.0f;
        TModelUnit prev_from_x = 0.0f;

        bool first_pass = true;
        for (CSeq_loc_CI iter(GetLocation());  iter;  ++iter) {
            CConstRef<CSeq_loc> loc = iter.GetRangeAsSeq_loc();
            range = iter.GetRange();
            f_out = range.GetFrom();
            t_out = range.GetTo();
            TSeqPos f_in = f_out;
            TSeqPos t_in = t_out;

            // initialize defined breakpoint ranges
            // inner start/stop, and outer start/stop
            if (loc->IsInt()) {
                if (loc->GetInt().IsSetFuzz_from()  &&
                    loc->GetInt().GetFuzz_from().IsRange()) {
                        const CInt_fuzz::C_Range& f_range =
                            loc->GetInt().GetFuzz_from().GetRange();
                        f_out = f_range.GetMin();
                        f_in = f_range.GetMax();
                }
                if (loc->GetInt().IsSetFuzz_to()  &&
                    loc->GetInt().GetFuzz_to().IsRange()) {
                        const CInt_fuzz::C_Range& t_range =
                            loc->GetInt().GetFuzz_to().GetRange();
                        t_out = t_range.GetMax();
                        t_in = t_range.GetMin();
                }
            }

            // draw the connecting line between intervals
            if ( !first_pass ) {
                // use a lighter color
                CRgbaColor color_lite = color;
                color_lite.Lighten(0.2f);
                TModelUnit intron_f = f_out > prev_to_x ? prev_to_x + 1 : t_out + 1;
                TModelUnit intron_t = f_out > prev_to_x ? f_out : prev_from_x;
                gl.LineWidth((float)(m_Config->m_LineWidth));
                gl.ColorC(color_lite);
                m_Context->DrawLine(intron_f, line_ym, intron_t, line_ym);
                gl.LineWidth(1.0f);
            }

            EUndefinedBpType undefined_bp_f = x_GetUndefinedBp_from(*loc);
            EUndefinedBpType undefined_bp_t = x_GetUndefinedBp_to(*loc);

            if (idx == CCustomFeatureColor::eLoss  &&  literal_len >= 0) {
                if (literal_len >= (int)(t_in - f_in)) {
                    literal_len = t_in - f_in - 2;
                }
                TModelUnit xm = (f_in + t_in) * 0.5;
                TModelUnit m = (f_in + t_in - literal_len) * 0.5;
                TModelUnit n = (f_in + t_in + literal_len) * 0.5;
                gl.ColorC(color_t);
                m_Context->DrawQuad(f_in, line_y1, m + 1.0, line_y2);
                m_Context->DrawQuad(n, line_y1, t_in + 1.0, line_y2);

                m += 1.0;
                TModelUnit half_w = m_Context->ScreenToSeq(3.0);
                if (literal_len < half_w * 2.0) {
                    half_w = literal_len * 0.5;
                }

                vector<TModelPoint> points;
                points.push_back(TModelPoint(m, line_ym));
                points.push_back(TModelPoint(xm - half_w, line_ym));
                points.push_back(TModelPoint(xm - half_w * 0.5, line_y1 - 1.0));
                points.push_back(TModelPoint(xm + half_w * 0.5, line_y2 + 1.0));
                points.push_back(TModelPoint(xm + half_w, line_ym));
                points.push_back(TModelPoint(n, line_ym));
                gl.Color3f(0.0f, 0.0f, 0.0f);
                m_Context->DrawLineStrip(points);

            } else if (idx == CCustomFeatureColor::eInsertion ||
                (idx == CCustomFeatureColor::eGain  &&  literal_len >= 0)) {

                gl.ColorC(color_t);
                m_Context->DrawQuad(f_in, line_y1, t_in + 1.0, line_y2);
                // draw insertion
                TModelUnit size_half = m_Context->ScreenToSeq(4.0);
                TModelUnit y_off = 1.0;
                TModelUnit x_off = 0.5;
                TModelUnit start = (f_in + t_in) * 0.5 + x_off;
                gl.Color3f(0.0f, 0.0f, 1.0f);
                m_Context->DrawLine(start, line_y1, start, line_y2);
                m_Context->DrawTriangle(start, line_ym,
                    start + size_half, line_y1 - y_off,
                    start - size_half, line_y1 - y_off);
                m_Context->DrawTriangle(start, line_ym,
                    start + size_half, line_y2 + y_off,
                    start - size_half, line_y2 + y_off);
            } else if (idx == CCustomFeatureColor::eTranschr) {
                m_Context->Draw3DQuad_HorzLines(f_in, line_y1, t_in + 1.0, line_y2, color, true);
            } else if (undefined_bp_f == eBp_Outer || undefined_bp_f == eBp_Outer) {
                gl.ColorC(color_t);
                m_Context->DrawQuad(f_in, line_y1, t_in + 1.0, line_y2);
            } else {
                m_Context->Draw3DQuad(f_in, line_y1, t_in, line_y2, color);
            }

            // render defined breakpoint ranges
            if (f_in != f_out) {
                gl.ColorC(color_t);
                m_Context->DrawQuad(f_out, line_y1, f_in, line_y2);
            }
            if (t_in != t_out) {
                gl.ColorC(color_t);
                m_Context->DrawQuad(t_in + 1, line_y1, t_out, line_y2);
            }

            // render undefined breakpoint ranges indicators
            TModelUnit tri_width = m_Context->ScreenToSeq(bar_height * 0.5);
            TModelUnit x_off = m_Context->ScreenToSeq(1.0);

            gl.ColorC(color_t);
            if (undefined_bp_f == eBp_Outer) {
                m_Context->DrawTriangle(f_in, line_ym, f_in - tri_width, line_y2,
                    f_in - tri_width, line_y1, true);
            } else if (undefined_bp_f == eBp_Inner) {
                m_Context->DrawTriangle(f_in - x_off, line_y1, f_in - x_off, line_y2,
                    f_in - x_off - tri_width, line_ym, true);
            }

            gl.ColorC(color_t);
            if (undefined_bp_t == eBp_Outer) {
                m_Context->DrawTriangle(t_in + 1.0, line_ym, t_in + 1.0 + tri_width, line_y2,
                    t_in + 1.0 + tri_width, line_y1, true);
            } else if (undefined_bp_t == eBp_Inner) {
                m_Context->DrawTriangle(t_in + x_off + 1.0, line_y1, t_in + x_off + 1.0, line_y2,
                    t_in + x_off + 1.0 + tri_width, line_ym, true);
            }

            first_pass = false;
            prev_to_x = t_out;
            prev_from_x = f_out;
        }

        // render Transchr ssv, only render it once per feature, not per interval
        if (idx == CCustomFeatureColor::eTranschr) {
            // use the total range for screen size comparison
            range = GetRange();
            if (m_Context->SeqToScreen(range.GetLength()) > bar_height * 2.0) {
                gl.Color3f(0.0f, 0.0f, 0.0f);
                TModelUnit x1 = t_out - m_Context->ScreenToSeq(bar_height) + 1.0;

                // first clear background
                gl.Disable(GL_BLEND);
                gl.Color3f(1.0f, 1.0f, 1.0f);
                m_Context->DrawQuad(x1, line_ym - bar_height, t_out + 1.0, line_ym + bar_height);

                // draw quad with lines
                m_Context->Draw3DQuad_HorzLines(
                    x1, line_ym - bar_height, t_out + 1.0, line_ym + bar_height, color, true);
            }
        }
    } else {
        // bar is less then 2 pixel. Do not draw intervals
        m_Context->Draw3DQuad(f_out, line_y1,
            t_out, line_y2, color, true);
    }
}


TModelUnit CFeatGlyph::x_GetBarHeight() const
{
    TModelUnit height = m_Config->GetBarHeight(
        m_Context->IsOverviewMode()  &&  m_HideLabel);

    if (IsDbVar(GetFeature())  &&
        m_Context->SeqToScreen(GetRange().GetLength()) > height * 2.0  &&
        GetCustomColorIdx(GetFeature()) == CCustomFeatureColor::eTranschr) {
            height += height;
    }

    return height;
}


CFeatGlyph::EUndefinedBpType
CFeatGlyph::x_GetUndefinedBp_from(const CSeq_loc& loc) const
{
    // currently only support location with type e_Int or e_Mix of e_Int.
    EUndefinedBpType type = eBp_Unknown;
    if (IsDbVar(GetFeature())  &&  GetFeature().GetData().GetVariation().IsInsertion()) {
        type = eBp_Outer;
    } else if (loc.IsInt()  &&
        loc.GetInt().IsSetFuzz_from()  &&
        loc.GetInt().GetFuzz_from().IsLim()) {
        CInt_fuzz::TLim lim = loc.GetInt().GetFuzz_from().GetLim();
        if (lim == CInt_fuzz::eLim_gt) {
            type = eBp_Outer;
        } else if (lim == CInt_fuzz::eLim_lt) {
            type = eBp_Inner;
        }
    } else if (loc.IsMix()) {
         const CSeq_loc* start_loc = loc.IsReverseStrand() ?
             loc.GetMix().Get().front() : loc.GetMix().Get().back();
         type = x_GetUndefinedBp_from(*start_loc);
    } else if (loc.IsPacked_int()) {
        const CSeq_interval* start_int = loc.IsReverseStrand() ?
             loc.GetPacked_int().Get().front() : loc.GetPacked_int().Get().back();
        if (start_int->IsSetFuzz_from()  && start_int->GetFuzz_from().IsLim()) {
            CInt_fuzz::TLim lim = start_int->GetFuzz_from().GetLim();
            if (lim == CInt_fuzz::eLim_gt) {
                type = eBp_Outer;
            } else if (lim == CInt_fuzz::eLim_lt) {
                type = eBp_Inner;
            }
        }
    }

    return type;
}


CFeatGlyph::EUndefinedBpType
CFeatGlyph::x_GetUndefinedBp_to(const CSeq_loc& loc) const
{
    // currently only support location with type e_Int or e_Mix of e_Int.
    EUndefinedBpType type = eBp_Unknown;
    if (IsDbVar(GetFeature())  &&  GetFeature().GetData().GetVariation().IsInsertion()) {
        type = eBp_Outer;
    } else if (loc.IsInt()  &&
        loc.GetInt().IsSetFuzz_to()  &&
        loc.GetInt().GetFuzz_to().IsLim()) {
        CInt_fuzz::TLim lim = loc.GetInt().GetFuzz_to().GetLim();
        if (lim == CInt_fuzz::eLim_gt) {
            type = eBp_Inner;
        } else if (lim == CInt_fuzz::eLim_lt) {
            type = eBp_Outer;
        }
    } else if (loc.IsMix()) {
         const CSeq_loc* stop_loc = loc.IsReverseStrand() ?
             loc.GetMix().Get().back() : loc.GetMix().Get().front();
         type = x_GetUndefinedBp_from(*stop_loc);
    } else if (loc.IsPacked_int()) {
         const CSeq_interval* stop_int = loc.IsReverseStrand() ?
             loc.GetPacked_int().Get().back() : loc.GetPacked_int().Get().front();
        if (stop_int->IsSetFuzz_to() && stop_int->GetFuzz_to().IsLim()) {
            CInt_fuzz::TLim lim = stop_int->GetFuzz_to().GetLim();
            if (lim == CInt_fuzz::eLim_gt) {
                type = eBp_Inner;
            } else if (lim == CInt_fuzz::eLim_lt) {
                type = eBp_Outer;
            }
        }
    }

    return type;
}


bool CFeatGlyph::x_ShowLabel() const
{
    bool shown = false;
    if (m_Config->m_LabelPos != CFeatureParams::ePos_NoLabel) {
        if (m_Config->m_LabelPos == CFeatureParams::ePos_Side) {
            shown = !m_HideLabel;
        } else { // can be either ePos_Above or ePos_Inside
            TModelRange r = m_Context->IntersectVisible(this);
            if (m_Context->WillLabelFit(r)) {
                shown = !m_HideLabel  ||  IsSelected();
            }
        }
    }
    return shown;
}


bool CFeatGlyph::x_LabelFirst() const
{
    bool first = true;
    if (m_Config->m_LabelPos != CFeatureParams::ePos_Above) {
        first = false;
        if (m_Config->m_LabelPos == CFeatureParams::ePos_Side  &&
            !m_HideLabel) {
                if (sequence::GetStrand(GetLocation()) == eNa_strand_minus) {
                    CRef<CSGSequenceDS> seq_ds = m_Context->GetSeqDS();
                    if (GetRight() > seq_ds->GetSequenceLength()) {
                        first = true;
                    }
                } else if (GetLeft() < 0) {
                    first = true;
                }
        }
    }
    return first;
}


bool CFeatGlyph::x_ShowRuler() const
{
    // Show ruler if a feature has ruler enabled and
    // is selected in details mode
    return  IsSelected()  &&  m_Config->m_ShowFeatRuler  &&
        !m_Context->IsOverviewMode();
}


string CFeatGlyph::x_GetFeatureId() const
{
    string id;
    const CSeq_feat& feat = GetFeature();
    const CSeq_feat::TId* feat_id = NULL;
    if (feat.IsSetId()) {
        feat_id = &feat.GetId();
    } else if (feat.IsSetIds()  &&  !feat.GetIds().empty()) {
        feat_id = feat.GetIds().front().GetPointer();
    }

    if (feat_id) {
        switch(feat_id->Which()) {
            case CFeat_id::e_Gibb:
                id = NStr::IntToString(feat_id->GetGibb());
                break;
            case CFeat_id::e_Giim:
                id = NStr::IntToString(feat_id->GetGiim().GetId());
                break;
            case CFeat_id::e_Local:
                if (feat_id->GetLocal().IsStr()) {
                    id = feat_id->GetLocal().GetStr();
                } else if (feat_id->GetLocal().IsId()) {
                    id = NStr::IntToString(feat_id->GetLocal().GetId());
                }
                break;
            case CFeat_id::e_General:
                id = feat_id->GetGeneral().GetDb() + ":";
                if (feat_id->GetGeneral().GetTag().IsStr()) {
                    id += feat_id->GetGeneral().GetTag().GetStr();
                } else if (feat_id->GetGeneral().GetTag().IsId()) {
                    id += NStr::IntToString(feat_id->GetGeneral().GetTag().GetId());
                }
                break;
            default:
                break;
        }
    }

    return id;
}


END_NCBI_SCOPE
