/*  $Id: feature_panel.cpp 27927 2013-04-25 19:24:57Z wuliangs $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Vlad Lebedev, Liangshou Wu
 *
 */


#include <ncbi_pch.hpp>
#include <gui/widgets/seq_graphic/feature_panel.hpp>
#include <gui/widgets/seq_graphic/seqgraphic_ds_manager.hpp>
#include <gui/widgets/seq_graphic/config_utils.hpp>
#include <gui/widgets/seq_graphic/rendering_ctx.hpp>
#include <gui/widgets/seq_graphic/feature_track.hpp>
#include <gui/widgets/seq_graphic/all_other_features_track.hpp>
#include <gui/widgets/seq_graphic/seqgraphic_graph_ds.hpp>
#include <gui/widgets/seq_graphic/rendering_ctx.hpp>
#include <gui/widgets/seq_graphic/seqgraphic_utils.hpp>
#include <gui/widgets/wx/wx_utils.hpp>
#include <gui/opengl/glutils.hpp>
#include <gui/opengl/glrender.hpp>
#include <gui/utils/system_path.hpp>
#include <gui/utils/extension_impl.hpp>

#include <objmgr/scope.hpp>
#include <objmgr/util/feature.hpp>
#include <objmgr/util/sequence.hpp>
#include <objtools/alnmgr/alnmix.hpp>
#include <corelib/ncbitime.hpp>
#include <serial/iterator.hpp>
#include <util/sequtil/sequtil_manip.hpp>

#include <algorithm>
#include <ctype.h>
#include <math.h>

#include <wx/menu.h>
#include <wx/utils.h>
#include <wx/string.h>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

/// cell sizes for CDS grid
static const int kMinCellSize = 30;
static const int kMaxCellSize = 72;
// CDS grid to follow codon boundaries
static const int kSteps[] = { 3, 6, 9 };

/// 5 extra pixels for better appearance.
static const TVPUnit kLayoutHeightExtra = 5;

/// at this point sequence will fit.
static const TModelUnit kPixPerBase4Sequence = 8.0;

/// maximal display level in multiple level layout.
static const int kMaxDisplayLevel = 3;

static const string kBaseKey("GBPlugins.SeqGraphicFeatPanel");

static const int kLevelTrackBaseID = 10000;

CTrackTypeInfo
CFeaturePanel::m_TypeInfo("feature_panel_track", "Feature Panel");


CFeaturePanel::CFeaturePanel(CRenderingContext* r_cntx, bool cgi_mode)
    : CTrackContainer(r_cntx)
    , m_SelFeatures(NULL)
    , m_SelCDSFeatures(NULL)
    , m_IsMultiLevel(false)
    , m_IconTexInitialized(false)
    , m_CgiMode(cgi_mode)
{
    SetConfigMgr(new CTrackConfigManager);
    x_InitTrackIcons();

    m_Simple->SetTopMargin(0);
    m_Simple->SetVertSpace(5);
    SetLayoutPolicy(m_Simple);

    m_Pos.Init(0.0, 0.0);

    // we want to make sure the direct children of feature panel have
    // zero indentation. So it must be -1 here.
    m_IndentLevel = -1;

    TTrackAttrFlags attr =
        fShowAlways | fFullTrack | fFrameVisible;
    SetTrackAttr(attr);

    // register track icons
    // we want to put help icon at the end, so deregister conten, and
    // add it back later.
    x_DeregisterIcon(eIcon_Content);
    x_RegisterIcon(SIconInfo(eIcon_Help,
        "Help", true, "track_help", "", "track_help_lite"));
    x_RegisterIcon(SIconInfo(eIcon_Content,
        "Content", true, "track_content", "", "track_content_lite"));
    x_RegisterIcon(SIconInfo(eIcon_Layout,
        "Layout style", true, "track_layout", "", "track_layout_lite"));
    x_RegisterIcon(SIconInfo(eIcon_Color,
        "Color", true, "track_color_theme", "", "track_color_theme_lite"));
    x_RegisterIcon(SIconInfo(eIcon_Size,
        "Size", true, "track_size", "", "track_size_lite"));
    x_RegisterIcon(SIconInfo(eIcon_Decoration, "Feature decoration",
        true, "track_decoration", "", "track_decoration_lite"));
    x_RegisterIcon(SIconInfo(eIcon_Label,
        "Label position", true, "track_label", "", "track_label_lite"));

    SetTitle(m_TypeInfo.GetDescr());

    // setup grid
    m_Gen.SetIntegerMode(true, true);
    m_Gen.EnableMinusOneBased(true, true);
    // don't update step automatically, the step will be updated
    // externally according to ruler's major ticks.
    m_Gen.SetAutoStepUpdate(false);
    m_Grid.EnableIntegerCentering(true);
}


CFeaturePanel::~CFeaturePanel()
{
    if (m_DS  &&  !m_CgiMode) {
        SaveTrackConfig();
    }
    DeleteAllJobs();
}


void CFeaturePanel::Update(bool layout_only)
{
    if (m_IsMultiLevel) {
        if (!layout_only) {
            if ( !m_DS->AllJobsFinished() ) {
                m_DS->DeleteAllJobs();
                SetTrackInitDone(true);
            }
            x_SetStartStatus();
        }

        // make sure we keep the track info up to date
        // E.g. the track titles get changed to more meaningful ones
        // when actual data get loaded.
        x_UpdateMsg();

        // we need to do something differently from CTrackContainer to
        // make sure that we will see data in layout tracks at level 1 and 2
        // before the tracks at level 3 get initialized which might be time
        // consumming.
        bool update_level_3 = true;
        TTrackProxies proxies = GetSubtrackProxies();
        proxies.sort(CTrackProxy::STPSorterByOrder::s_CompareCRefs);
        NON_CONST_ITERATE (TTrackProxies, iter, proxies) {
            if ( !(*iter)->GetShown() ) continue;

            CLayoutTrack* track = (*iter)->GetTrack();
            if ( !track) continue;

            const CTrackContainer* cont =
                dynamic_cast<const CTrackContainer*>(track);

            int level = -1;
            if ( cont ) {
                level = cont->GetAnnotLevel();
            }
            if (level < 0) {
                track->Update(layout_only);
            } else if (level < 2) {
                track->Update(layout_only);
                if ( !cont->IsTrackInitDone() ) {
                    update_level_3 = false;
                }
            } else if (update_level_3) {
                const TSeqRange& curr_range = m_Context->GetVisSeqRange();
                if (layout_only) {
                    if (curr_range != (*iter)->GetVisitedRange()) {
                        (*iter)->SetVisitedRange(m_Context->GetVisSeqRange());
                        track->Update(false);
                    } else {
                        track->Update(true);
                    }
                } else {
                    track->Update(false);
                    (*iter)->SetVisitedRange(curr_range);
                }                
            } else {
                track->Update(true);
            }
        }
        x_UpdateLayout();

        // check if we need to reinitialize the sub tracks.
        x_ReinitSubtracks();
    } else {
        CDataTrack::Update(layout_only);
    }

    if (m_CgiMode) {
        if ( !layout_only ) {
            int order_id = 0;
            TTrackProxies proxies;
            SetChildren().clear();
            x_MakeTopLevelTracks(GetSubtrackProxies(), proxies, order_id, "");
            SetSubtrackProxies(proxies);
            x_UpdateLayout();
        }
        SetMsg("");
    }
}


CRef<CSeqGlyph> CFeaturePanel::HitTest(const TModelPoint& p)
{
    CRef<CSeqGlyph> glyph;
     // this is a special case since feature panel is a special track
    if (x_HitTitleBar(p)) {
        glyph.Reset(this);
    } else {
        glyph = CGlyphContainer::HitTest(p);
    }
    return glyph;
}


void CFeaturePanel::SetInputObject(SConstScopedObject& obj)
{
    m_InputObj = obj;
    CIRef<ISGDataSource> ds =
        m_DSContext->GetDS(typeid(CFeaturePanelDSType).name(), obj);
    m_DS.Reset(dynamic_cast<CFeaturePanelDS*>(ds.GetPointer()));
    m_DS->SetJobListener(this);
    m_DS->SetDSContext(m_DSContext);
    m_DS->SetRenderingContext(m_Context);
}


void CFeaturePanel::ConfigureTracks()
{
    //TIME_ME("Configure track and load data");
    if (m_TrackSettings.empty()) {
        x_LoadSettings();
    }
    if (m_DS.NotNull()) {
        if ( !m_gConfig->GetViewerCtx().empty()  &&
             GetConfigMgr()->GetKnownAnnots(0).empty()) {
            // we might need to retrieve meta-data for NAAs
            SetTrackInitDone(false);
            m_DS->GetAnnotMetaData("Retrieve NA meta-data", m_gConfig->GetViewerCtx());
        } else {
            x_ConfigureTracks();
            Update(false);
        }
        x_OnLayoutChanged();
    }
}


void CFeaturePanel::OnDataChanged()
{
    GetConfigMgr()->SaveProfile_Recursive(this);
    GetConfigMgr()->LoadProfile_Recursive(this);
    x_ConfigureTracks();
    Update(false);
}


void CFeaturePanel::SetExternalGlyphs(const CSeqGlyph::TObjects& objs)
{
    if ( !m_ExtLayoutTrack ) {
        m_ExtLayoutTrack.Reset(new CExternalLayoutTrack(m_Context));
        m_ExtLayoutTrack->SetTitle("View reflection");
        SetTrack(m_ExtLayoutTrack, -100);
    }
    m_ExtLayoutTrack->SetObjects(objs);
    m_ExtLayoutTrack->Update(true);
    SetGroup().UpdateLayout();
}


void CFeaturePanel::LoadDefaultTracks()
{
    // remove the old track prifile from gui registry
    string path = kBaseKey + CGuiRegistry::kDecimalDot +
        CSGConfigUtils::TrackProfileKey() + CGuiRegistry::kDecimalDot;
    string profile = GetProfile();

    CGuiRegistry& registry = CGuiRegistry::GetInstance();
    CGuiRegistry::CReadWriteView view =
        registry.GetReadWriteView(path + profile + ".Tracks");
    CGuiRegistry::TKeys keys;
    view.GetTopKeys(keys);
    ITERATE (CGuiRegistry::TKeys, iter, keys) {
        view.DeleteField(iter->key);
    }

    // set the track profile name to the global default
    string default_profile = CSGConfigUtils::DefTrackProfile();
    SetProfile(default_profile);
    m_gConfig->SetTrackProfile(default_profile);
    m_gConfig->SetDirty(true);
    x_LoadSettings();
    m_gConfig->SyncUserSettingVersion();
    x_ConfigureTracks();
    Update(false);
}


void CFeaturePanel::SyncTrackSettings()
{
    if ( !m_gConfig->NeedToSyncUserSettings()  ||  !m_gConfig->AutoSync()) 
         return ;

    string def_profile = CSGConfigUtils::DefTrackProfile();

    if (m_gConfig->GetTrackProfile() != def_profile) {
        TTrackProxies def_proxies;
        string path = kBaseKey + CGuiRegistry::kDecimalDot +
            CSGConfigUtils::TrackProfileKey() +
            CGuiRegistry::kDecimalDot + def_profile;
        GetConfigMgr()->LoadTrackSettingsRecursive(def_proxies, path);

        typedef map<string, CRef<CTempTrackProxy> > TNamedTracks;
        TNamedTracks curr_proxies;
        ITERATE (TTrackProxies, iter, m_TrackSettings) {
            curr_proxies.insert(TNamedTracks::value_type((*iter)->GetName(), *iter));
        }

        ITERATE (TTrackProxies, iter, def_proxies) {
            TNamedTracks::iterator target_iter = curr_proxies.find((*iter)->GetName());
            if (target_iter != curr_proxies.end()) {
                // merge annotation names
                CTrackProxy& proxy = dynamic_cast<CTrackProxy&>(*target_iter->second);
                const CTrackProxy& src_proxy = dynamic_cast<const CTrackProxy&>(**iter);
                proxy.SetAnnots(src_proxy.GetAnnots());
            } else {
                m_TrackSettings.push_back(*iter);
            }
        }
        m_TrackSettings.sort(CTrackProxy::STPSorterByOrder::s_CompareCRefs);
    }

    // Otherwise, it is the first time user. No sync is needed,
    // but we do need to update the setting version
    m_gConfig->SyncUserSettingVersion();
}


void CFeaturePanel::SetTrackConfig(const TTrackSettingsSet& settings_set)
{
    GetConfigMgr()->SetTrackConfig(m_TrackSettings, settings_set);
}


void CFeaturePanel::SaveTrackConfig()
{
    GetConfigMgr()->SaveProfile_Recursive(this);
}


void CFeaturePanel::ShowTrack(const string& track_key)
{
    GetConfigMgr()->ShowTrack(m_TrackSettings, track_key);
}


const CTrackTypeInfo& CFeaturePanel::GetTypeInfo() const
{
    return m_TypeInfo;
}


void CFeaturePanel::OnTrackInitFinished(const CLayoutTrack* sender)
{
    if (!m_IsMultiLevel) return;

    const CTrackContainer* cont = dynamic_cast<const CTrackContainer*>(sender);

    if (!cont  ||  cont->GetAnnotLevel() == 2) return;
        
    bool update_level_3 = true;
    TTrackProxies proxies = GetSubtrackProxies();
    proxies.sort(CTrackProxy::STPSorterByOrder::s_CompareCRefs);
    NON_CONST_ITERATE (TTrackProxies, iter, proxies) {
        if ( !(*iter)->GetShown() ) continue;

        CTrackContainer* cont_track =
            dynamic_cast<CTrackContainer*>((*iter)->GetTrack());
        if ( !cont_track) continue;

        int level = cont_track->GetAnnotLevel();
        if (level < 2) {
            if ( !cont->IsTrackInitDone() ) {
               update_level_3 = false;
            }
        } else if (update_level_3) {
            const TSeqRange& curr_range = m_Context->GetVisSeqRange();
            if (curr_range != (*iter)->GetVisitedRange()) {
                (*iter)->SetVisitedRange(m_Context->GetVisSeqRange());
                cont_track->Update(false);
                x_UpdateBoundingBox();
            }
        }
    }
}


void CFeaturePanel::x_Draw() const
{
    CGlRender& gl = GetGl();

    CGlPane* pane = m_Context->GetGlPane();

    // draw grid
    if (m_gConfig->GetShowHorzGrid()  ||  m_gConfig->GetShowVertGrid()) {
        CRgbaColor v_color = m_gConfig->GetVerticalGridColor();
        CRgbaColor h_color = m_gConfig->GetHorizontalGridColor();
        if (x_NeedSpecialGrid()) {
            v_color.Lighten(0.2f);
            h_color.Lighten(0.2f);
        }
        {{
            CRegularGridRenderer* grid = const_cast<CRegularGridRenderer*>(&m_Grid);
            grid->SetHorzColor(h_color);
            grid->SetVertColor(v_color);
        }}
        m_Grid.Render(pane, pane, const_cast<CRegularGridGen*>(&m_Gen));
    }

    CGlPaneGuard GUARD(*pane, CGlPane::eOrtho);

    // make sure the icon texture objects have been generated.
    // don't initalize textures in cgi mode sinc we don't need them.
    if ( !m_IconTexInitialized  &&  !m_CgiMode) {
        // this call will check if the texture objects are truely invalid.
        m_IconTexInitialized = CLayoutTrack::InitIconTextures();
    }

    bool horz = m_Context->IsHorizontal();
    TModelUnit off_y = horz ? pane->GetOffsetY() : pane->GetOffsetX();
    gl.PushMatrix();
    gl.Translatef(0.0f, -off_y, 0.0f);

    // render speciality grid for selected CDSs
    x_RenderFeatureGrid();

    // Render all fair-lines for  selected features
    x_RenderHairLineSelections();

    gl.Enable(GL_BLEND);
    gl.BlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    gl.Enable(GL_LINE_SMOOTH);
    glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
    gl.LineWidth(1.0f);
    gl.PolygonMode(GL_FRONT_AND_BACK, GL_FILL);

    GetGroup().Draw();

    // Don't render icons/menu on pdf output
    if (gl.GetApi() != eRenderPDF)
        x_RenderTitleBar();

    if ( !IsTrackInitDone()  &&  !m_CgiMode ) {
        gl.Color4f(1.0f, 0.0f, 0.0f, 1.0f);
        //CGlBitmapFont font(CGlBitmapFont::eBitmap9x15);
        CGlTextureFont font(CGlTextureFont::eFontFace_Helvetica_Italic, 15);
        TModelRect rcm;
        x_GetTBRect(rcm);
        TModelUnit x = m_Context->GetVisibleFrom() +
            m_Context->ScreenToSeq(10.0);
        m_Context->AdjustToOrientation(x);
        m_Context->TextOut(&font, "Initializing tracks...",
            x, rcm.Bottom() - 10.0, false, false);
    }

    gl.PopMatrix();

    gl.Disable(GL_BLEND);
    gl.Disable(GL_LINE_SMOOTH);
}


bool CFeaturePanel::x_IsJobNeeded(CAppJobDispatcher::TJobID id) const
{
    if (m_DS->IsMetaDataJob(id)) {
        return true;
    }
    return m_DS->IsJobNeeded(id);
}


void CFeaturePanel::x_OnJobCompleted(CAppJobNotification& notify)
{
    CRef<CObject> res_obj = notify.GetResult();
    CAnnotMetaDataJobResult* result =
        dynamic_cast<CAnnotMetaDataJobResult*>(&*res_obj);
    if (result) {
        m_DS->ClearMetaDataJobId();
        if ( !result->m_Annots.empty() ) {
            GetConfigMgr()->SetKnownAnnots(result->m_Annots, 0);
        }
        x_ConfigureTracks();
        Update(false);
        x_OnLayoutChanged();
    } else {
        CTrackContainer::x_OnJobCompleted(notify);
    }
}


bool CFeaturePanel::x_Empty() const
{
    return GetChildren().empty();
}


void CFeaturePanel::x_LoadSettings(const string& preset_style,
                                   const TKeyValuePairs& /*settings*/)
{
    if (preset_style.empty()) {
        SetProfile(CSGConfigUtils::DefTrackProfile());
    } else {
        SetProfile(preset_style);
    }
    x_LoadSettings();
}


void CFeaturePanel::x_SaveSettings(const string& /*preset_style*/)
{
    x_SaveSettings();
}


void CFeaturePanel::x_GetTBRect(TModelRect& rect) const
{
    int icon_s = m_gConfig->GetIconSize();
    TModelUnit h = icon_s * 1.5;
    TModelUnit w = (m_Icons.size() * 1.5 + 0.5)* icon_s; 
    w = m_Context->ScreenToSeq(w) * 0.5;
    TModelUnit center_x =
        (m_Context->GetVisibleFrom() + m_Context->GetVisibleTo()) * 0.5;
    const CGlPane* pane = m_Context->GetGlPane();
    TModelUnit bottom = pane->GetVisibleRect().Bottom();
    if ( !m_Context->IsHorizontal() ) {
        bottom = pane->GetVisibleRect().Left();
    }
    rect.Init(center_x - w, bottom, center_x + w, bottom - h);
}


void CFeaturePanel::x_RenderTitleBar() const
{
    CGlRender& gl = GetGl();

    if (m_Icons.empty()  ||  m_CgiMode) {
        return;
    }

    CConstRef<CSeqGraphicConfig> config = x_GetGlobalConfig();

    // render track frame
    TModelRect rcm;
    x_GetTBRect(rcm);
    if (m_Attrs & fFrameVisible) {
        TModelUnit left = rcm.Left();
        TModelUnit right = rcm.Right();
        m_Context->AdjustToOrientation(left);
        m_Context->AdjustToOrientation(right);
        TModelRect rcm_adj(left, rcm.Bottom(), right, rcm.Top());
        TModelUnit w = m_Context->SeqToScreen(rcm_adj.Width());
        TModelUnit diff = 0; 
        if (w < 100) {
            diff = m_Context->ScreenToSeq(100 - w);
            w = m_Context->ScreenToSeq(200.0);
        }
        rcm_adj.Inflate(diff, 0.0);

        if (m_ShowFrame) {
            gl.Color4f(0.83f, 0.83f, 0.83f, 0.9f);
            m_Context->DrawQuad(rcm_adj);
            TModelUnit h = rcm.Bottom() - rcm.Top();
            m_Context->DrawDisk(TModelPoint(rcm_adj.Left(), rcm_adj.Bottom()),
                h, 90, 180);
            m_Context->DrawDisk(TModelPoint(rcm_adj.Right(), rcm_adj.Bottom()),
                h, 0, 90);
        }

        size_t icon_num = m_Icons.size();
        for (size_t idx = 0;  idx < icon_num;  ++idx) {
            x_RenderIcon(m_Icons[idx].m_Id, m_HighlightedIcon == (int)idx, !m_ShowFrame);
        }

        if (!m_ShowFrame) {
            gl.BlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
        }
    }
}


void CFeaturePanel::x_OnIconClicked(TIconID id)
{
    switch (id) {
    case eIcon_Help:
        x_OnHelpIconClicked();
        break;
    case eIcon_Layout:
        x_OnLayoutIconClicked();
        break;
    case eIcon_Color:
        x_OnColorIconClicked();
        break;
    case eIcon_Size:
        x_OnSizeIconClicked();
        break;
    case eIcon_Decoration:
        x_OnDecorationIconClicked();
        break;
    case eIcon_Label:
        x_OnLabelIconClicked();
        break;
    default:
        // use default handlers
        CTrackContainer::x_OnIconClicked(id);
    }
}


void CFeaturePanel::x_OnHelpIconClicked()
{
    typedef map <int, wxString> TMenuItems;
    TMenuItems items;
    int id_base = 10000;
    items[id_base] = wxT("Toggle Quick Help (L)");
    items[id_base + 1] = wxT("Interaction Guide");
    items[id_base + 2] = wxT("Graphical Legends");

    wxMenu menu;
    UseDefaultMarginWidth(menu);
    ITERATE (TMenuItems, iter, items) {
        menu.AppendCheckItem(iter->first, iter->second);
    }

    m_LTHost->LTH_PopupMenu(&menu);
    wxMenuItemList& item_list = menu.GetMenuItems();
    ITERATE (wxMenuItemList, iter, item_list) {
        if ((*iter)->IsChecked()) {
            switch ((*iter)->GetId() - id_base) {
                case 0:
                    m_LTHost->LTH_ToggleLegend();
                    break;
                case 1:
                    ::wxLaunchDefaultBrowser(wxT("http://www.ncbi.nlm.nih.gov/tools/gbench/tutorial1#Navigating_Graphical_View"));
                    break;
                case 2:
                    ::wxLaunchDefaultBrowser(wxT("http://www.ncbi.nlm.nih.gov/projects/sviewer/help/legends.pdf"));
                    break;
                default:
                    break;
            }
        }
    }
}


void CFeaturePanel::x_OnLayoutIconClicked()
{
    typedef map <int, wxString> TStyleMap;
    TStyleMap styles;
    int id_base = 10000;
    styles[id_base] = wxT("Default");
    styles[id_base + 1] = wxT("Multiple level");

    wxMenu menu;
    UseDefaultMarginWidth(menu);

    int id = 0;
    ITERATE (TStyleMap, iter, styles) {
        wxMenuItem* item = menu.AppendRadioItem(iter->first, iter->second);
        if (m_IsMultiLevel  &&  id == 1) {
            item->Check();
        } else if (id == 0) {
            item->Check();
        }
        ++id;
    }

    m_LTHost->LTH_PopupMenu(&menu);
    wxMenuItemList& item_list = menu.GetMenuItems();
    ITERATE (wxMenuItemList, iter, item_list) {
        if ((*iter)->IsChecked()) {
            id = (*iter)->GetId() - id_base;
            break;
        }
    }

    if ((id == 0  &&  m_IsMultiLevel)  ||  (id == 1  &&  !m_IsMultiLevel)) {
        GetConfigMgr()->SaveProfile_Recursive(this);
        GetConfigMgr()->LoadProfile_Recursive(this);
        m_IsMultiLevel = !m_IsMultiLevel;
        x_ConfigureTracks();
        Update(false);
        LOG_POST(Info << "Graphical view: switch to " <<
            (m_IsMultiLevel ? "multiple level layout" : "default layout mode"));
    }
}


void CFeaturePanel::x_OnColorIconClicked()
{
    // prohibit change settings before initialization gets finished
    if ( !IsTrackInitDone() ) return;

    typedef map <int, wxString> TMenuItems;
    TMenuItems items;
    int id_base = 10000;
    items[id_base] = wxT("Greyscale");
    items[id_base + 1] = wxT("Color");
    items[id_base + 2] = wxT("Highcontrast");

    wxMenu menu;
    UseDefaultMarginWidth(menu);
    ITERATE (TMenuItems, iter, items) {
        wxMenuItem* item = menu.AppendRadioItem(iter->first, iter->second);
        if (iter->second == ToWxString(m_gConfig->GetColorTheme())) {
            item->Check();
        }
    }

    m_LTHost->LTH_PopupMenu(&menu);
    wxMenuItemList& item_list = menu.GetMenuItems();
    ITERATE (wxMenuItemList, iter, item_list) {
        const string& new_color = ToStdString(items[(*iter)->GetId()]);
        if ((*iter)->IsChecked()  &&  new_color != m_gConfig->GetColorTheme()) {
            LOG_POST(Info << "Graphical view: switch color theme from " <<
                m_gConfig->GetColorTheme() << " to " << new_color);
            GetConfigMgr()->SaveProfile_Recursive(this);
            m_gConfig->SaveSettings(true);
            m_gConfig->SetColorTheme(new_color);
            m_gConfig->LoadSettings();
            if (m_DS->GetBioseqHandle().IsProtein()) {
                m_gConfig->IgnoreFeatureStrand();
            }
            GetConfigMgr()->LoadProfile_Recursive(this);
            m_gConfig->SetDirty(true);
            Update(false);
            break;
        }
    }
}


void CFeaturePanel::x_OnSizeIconClicked()
{
    // prohibit change settings before initialization gets finished
    if ( !IsTrackInitDone() ) return;

    typedef map <int, wxString> TMenuItems;
    TMenuItems items;
    int id_base = 10000;
    items[id_base] = wxT("Compact");
    items[id_base + 1] = wxT("Normal");
    items[id_base + 2] = wxT("Oversize");

    wxMenu menu;
    UseDefaultMarginWidth(menu);
    ITERATE (TMenuItems, iter, items) {
        wxMenuItem* item = menu.AppendRadioItem(iter->first, iter->second);
        if (iter->second == ToWxString(m_gConfig->GetSizeLevel())) {
            item->Check();
        }
    }

    m_LTHost->LTH_PopupMenu(&menu);
    wxMenuItemList& item_list = menu.GetMenuItems();
    ITERATE (wxMenuItemList, iter, item_list) {
        const string& new_size = ToStdString(items[(*iter)->GetId()]);
        if ((*iter)->IsChecked()  &&  new_size != m_gConfig->GetSizeLevel()) {
            LOG_POST(Info << "Graphical view: switch size level from " <<
                m_gConfig->GetSizeLevel() << " to " << new_size);
            GetConfigMgr()->SaveProfile_Recursive(this);
            m_gConfig->SaveSettings(true);
            m_gConfig->SetSizeLevel(new_size);
            m_gConfig->LoadSettings();
            if (m_DS->GetBioseqHandle().IsProtein()) {
                m_gConfig->IgnoreFeatureStrand();
            }
            GetConfigMgr()->LoadProfile_Recursive(this);
            m_gConfig->SetDirty(true);
            Update(false);
            break;
        }
    }
}


void CFeaturePanel::x_OnDecorationIconClicked()
{
    // prohibit change settings before initialization gets finished
    if ( !IsTrackInitDone() ) return;

    typedef map <int, wxString> TMenuItems;
    TMenuItems items;
    int id_base = 10000;
    items[id_base] = wxT("Default");
    items[id_base + 1] = wxT("Arrows");
    items[id_base + 2] = wxT("SquareAnchor");
    items[id_base + 3] = wxT("CircleAnchor");
    items[id_base + 4] = wxT("Fancy");

    wxMenu menu;
    UseDefaultMarginWidth(menu);
    ITERATE (TMenuItems, iter, items) {
        wxMenuItem* item = menu.AppendRadioItem(iter->first, iter->second);
        if (iter->second == ToWxString(m_gConfig->GetDecorateStyle())) {
            item->Check();
        }
    }

    m_LTHost->LTH_PopupMenu(&menu);
    wxMenuItemList& item_list = menu.GetMenuItems();
    ITERATE (wxMenuItemList, iter, item_list) {
        const string& new_decor = ToStdString(items[(*iter)->GetId()]);
        if ((*iter)->IsChecked()  &&  new_decor != m_gConfig->GetDecorateStyle()) {
            LOG_POST(Info << "Graphical view: switch feature decoration from " <<
                m_gConfig->GetDecorateStyle() << " to " << new_decor);
            x_SaveSettings();
            m_gConfig->SaveSettings(true);
            m_gConfig->SetDecorateStyle(new_decor);
            m_gConfig->LoadSettings();
            if (m_DS->GetBioseqHandle().IsProtein()) {
                m_gConfig->IgnoreFeatureStrand();
            }
            m_gConfig->SetDirty(true);
            Update(false);
            break;
        }
    }
}


void CFeaturePanel::x_OnLabelIconClicked()
{
    // prohibit change settings before initialization gets finished
    if ( !IsTrackInitDone() ) return;

    const string& old_label_pos = m_gConfig->GetLabelPos();
    typedef map <int, wxString> TMenuItems;
    TMenuItems items;
    int id_base = 10000;
    items[id_base] = wxT("Default");
    items[id_base + 1] = wxT("Top label");
    items[id_base + 2] = wxT("Side label");
    items[id_base + 3] = wxT("No label");

    wxMenu menu;
    UseDefaultMarginWidth(menu);
    ITERATE (TMenuItems, iter, items) {
        wxMenuItem* item = menu.AppendRadioItem(iter->first, iter->second);
        if (iter->second == ToWxString(old_label_pos)) {
            item->Check();
        }
    }

    m_LTHost->LTH_PopupMenu(&menu);
    wxMenuItemList& item_list = menu.GetMenuItems();
    ITERATE (wxMenuItemList, iter, item_list) {
        const string& new_label_pos = ToStdString(items[(*iter)->GetId()]);
        if ((*iter)->IsChecked()  &&  new_label_pos != old_label_pos) {
            LOG_POST(Info << "Graphical view: switch feature label position from " <<
                old_label_pos << " to " << new_label_pos);
            GetConfigMgr()->SaveProfile_Recursive(this);
            m_gConfig->SaveSettings(true);
            m_gConfig->SetLabelPos(new_label_pos);
            m_gConfig->LoadSettings();
            if (m_DS->GetBioseqHandle().IsProtein()) {
                m_gConfig->IgnoreFeatureStrand();
            }
            GetConfigMgr()->LoadProfile_Recursive(this);
            m_gConfig->SetDirty(true);
            Update(false);
            break;
        }
    }
}


void CFeaturePanel::x_ConfigureTracks()
{
    TIME_ME("Configure Tracks");

    CRef<CSeqGraphicConfig> config = m_gConfig;
    if (config) {
        // we don't need to show strand indicator for features on protein.
        // make sure we turn them off if this is a protein sequnce.
        if (m_DS->GetBioseqHandle().IsProtein()) {
            config->IgnoreFeatureStrand();
        }
    
        ClearTracks();

        if (m_IsMultiLevel) {
            // feature panel track will hold only globally unique tracks,
            // which have no annot level concept (that is why we use -2),
            // and the track containers with level 0, 1, and 2
            SetAnnotLevel(-2);
            SetAdaptive(false);

            /// configure those tracks with no level concept.
            x_ConfigureSubtracks_Recursive(this, m_TrackSettings, -2);

            // configure other level-dependent tracks
            // to avoid track order conflict with the level-independent
            // tracks, we start the level track order = 100000
            int id = kLevelTrackBaseID;
            for (int level = 0; level < kMaxDisplayLevel; ++level) {
                string track_name = "Level " + NStr::IntToString(level + 1);
                bool shown = (find(m_DisplayLevels.begin(),
                    m_DisplayLevels.end(), level) != m_DisplayLevels.end());
                CRef<CTrackContainer> cont =
                    x_CreateLevelTrack(level, id++, track_name, shown);
                x_ConfigureSubtracks_Recursive(cont, m_TrackSettings, level);
            }
        } else {
            // feature panel uses adaptive annot selector
            if (config->GetAnnotDepth() == CSeqGraphicConfig::eDepth_unknown) {
                SetAdaptive(true);
                int depth = 0;
                // first check if any feature fetch depth set explicitly
                if ( !CSGUtils::GetFeatFetchDepth(m_DS->GetBioseqHandle(),
                    m_DS->GetScope(), depth) ) {
                    // sequence-specific
                    if (CSGUtils::IsChromosome(m_DS->GetBioseqHandle(), m_DS->GetScope())) {
                        depth = 1;
                    } else if (CSGUtils::IsSegSet(m_DS->GetBioseqHandle(), m_DS->GetScope())) {
                        depth = -1;
                    }
                }
                SetAnnotLevel(depth);
            } else if (config->GetAnnotDepth() == CSeqGraphicConfig::eDepth_nolimit) {
                SetAdaptive(true);
                SetAnnotLevel(-1);
            } else {
                // use exact depth
                SetAdaptive(false);
                SetAnnotLevel((int)config->GetAnnotDepth());
            }
            x_ConfigureSubtracks_Recursive(this, m_TrackSettings, -1);
        }
    }
}


void CFeaturePanel::x_ConfigureSubtracks_Recursive(CTrackContainer* parent,
                                                   const TTrackProxies& settings,
                                                   int level)
{
    ITERATE (TTrackProxies, t_iter, settings) {
        const CTrackProxy* param =
            dynamic_cast<const CTrackProxy*>(t_iter->GetPointer());
        if (!param) {
            /// the track and its siblings are temporary tracks that
            /// will be created by its parent. We skip them here.
            break;
        }
        CTrackContainer* cont = x_ConfigureSubtracks(parent, param, level);
        if (cont) {
            x_ConfigureSubtracks_Recursive(cont, param->GetChildren(), level);
        }
    }
}


CTrackContainer*
CFeaturePanel::x_ConfigureSubtracks(CTrackContainer* parent,
                                    const CTrackProxy* param,
                                    int level)
{
    // returned track container.
    // set this value if the created track is a container track.
    // the returned container track will serve as the parent 
    // for recursive track loading.
    CTrackContainer* cont_track = NULL;
    CRef<CTrackProxy> t_proxy;

    const ILayoutTrackFactory* factory = 
        GetConfigMgr()->GetTrackFactory(param->GetKey());
    if ( !factory ) return cont_track;

    // this is a track that doesn't understand annotation level concept,
    // but we only want to show the tracks in a given annotation level (>=0)
    if (level >=0  &&  !factory->UnderstandLevel()) return cont_track;

    // this is a track that does understand annotation level concept,
    // but we only want to show the tracks with no annot concept.
    if (level == -2  &&  factory->UnderstandLevel()) return cont_track;

    int depth = level;
    if (GetAdaptive()) {
        depth = GetAnnotLevel();
    }

    // create track proxy for the track
    t_proxy.Reset(new CTrackProxy(*param));

    // create track immediately if no background initialization is required
    if ( !factory->NeedBackgroundInit()  &&
        (!m_CgiMode  ||  t_proxy->GetShown())) {
        ILayoutTrackFactory::SExtraParams extra_params(
            depth, GetAdaptive(), &t_proxy->GetAnnots(),
            t_proxy->GetSubkey(), t_proxy->GetFilter(), t_proxy->GetSortBy());
        extra_params.m_SkipGenuineCheck = GetSkipGenuineCheck();
        TTrackMap tracks = factory->CreateTracks(m_InputObj, m_DSContext,
            m_Context, extra_params, GetConfigMgr()->GetKnownAnnots(level));

        if ( !tracks.empty()) {
            parent->AddTracks(t_proxy, tracks);
            CLayoutTrack* track = tracks.begin()->second;
            if (CTrackContainer* c_track = dynamic_cast<CTrackContainer*>(track)) {
                c_track->SetSubtrackProxies(t_proxy->GetChildren());
                c_track->SetConfigMgr(GetConfigMgr());
            }

            if (CAllOtherFeaturesTrack* feats_track =
                dynamic_cast<CAllOtherFeaturesTrack*>(track)) {
                feats_track->SetProxy(t_proxy);
            }
        }
    } else {
        // add the track proxy to the parent track.
        parent->AddTrackProxy(t_proxy);
    }

    return cont_track;
}

static const string k_multi_level = "MultiLevel";
static const string k_display_level = "DisplayLevels";


void CFeaturePanel::x_LoadSettings()
{
    CConstRef<CSeqGraphicConfig> g_conf = x_GetGlobalConfig();
    m_Grid.SetHorzColor(g_conf->GetHorizontalGridColor());
    m_Grid.SetVertColor(g_conf->GetVerticalGridColor());
    m_Grid.SetShowHorzGrid(g_conf->GetShowHorzGrid());
    //m_Grid.SetShowVertGrid(g_conf->GetShowVertGrid());
    // no vertical grids for graphical sequence view
    m_Grid.SetShowVertGrid(false);

    string path = kBaseKey + CGuiRegistry::kDecimalDot + CSGConfigUtils::LayoutKey();

    // loading layout level settings
    CGuiRegistry& registry = CGuiRegistry::GetInstance();
    CGuiRegistry::TReadView view = CSGConfigUtils::GetReadView(registry,
        path, g_conf->GetLayout(), CSGConfigUtils::DefLayout());
    m_IsMultiLevel = view.GetBool(k_multi_level);
    m_DisplayLevels.clear();

    CConstRef<CUser_field> f = view.GetField(k_display_level);
    if (f) {
        if (f->GetData().Which() == CUser_field::TData::e_Ints) {
            m_DisplayLevels = f->GetData().GetInts();
        } else if (f->GetData().Which() == CUser_field::TData::e_Str) {
            string str = f->GetData().GetStr();
            NStr::TruncateSpacesInPlace(str);
            string::size_type sb = str.find_first_not_of("([");
            if (sb == string::npos) {
                sb = 0;
            }
            string::size_type se = str.find_last_not_of (")] ");
            if (se == string::npos) {
                se = str.size() - 1;
            }
            list<string> toks;
            NStr::Split(str.substr(sb, se - sb + 1), ", ", toks);
            try {
                ITERATE (list<string>, iter, toks) {
                    int level = NStr::StringToInt(*iter);
                    x_SetShowDisplayLevel(level, true);
                }
            } catch (CException& e) {
                LOG_POST(Error << "Feature panel: errors when parsing "
                    <<"display level settings, " << e.GetMsg());
            }
        } else {
            x_SetShowDisplayLevel(f->GetData().GetInt(), true);
        }
    }

    m_TrackSettings.clear();
    path = kBaseKey + CGuiRegistry::kDecimalDot + CSGConfigUtils::TrackProfileKey() +
        CGuiRegistry::kDecimalDot + GetProfile();

    GetConfigMgr()->ResetHistory();
    GetConfigMgr()->LoadTrackSettingsRecursive(m_TrackSettings, path);

    // check if we need to sync track settings
    SyncTrackSettings();
}


void CFeaturePanel::x_SaveSettings()
{
    CGuiRegistry& registry = CGuiRegistry::GetInstance();
    CConstRef<CSeqGraphicConfig> g_conf = x_GetGlobalConfig();
    if ( !g_conf ) return;

    // saving layout level settings
    string path = kBaseKey + CGuiRegistry::kDecimalDot + CSGConfigUtils::LayoutKey();
    CGuiRegistry::CReadWriteView view =
        CSGConfigUtils::GetReadWriteView(registry, path, g_conf->GetLayout(),
        CSGConfigUtils::DefLayout());
    view.Set(k_multi_level, m_IsMultiLevel);
    string levels = kEmptyStr;
    ITERATE (TDisplayLevels, iter, m_DisplayLevels) {
        levels += " " + NStr::IntToString(*iter);
    }
    CRef<CUser_field> field = view.SetField(k_display_level);
    field->SetData().SetStr(levels);

    if (m_IsMultiLevel) {
        const TTrackProxies& proxies = GetSubtrackProxies();
        for (int i = 2; i >= 0; --i) {
            ITERATE (TTrackProxies, iter, proxies) {
                const CTrackContainer* cont =
                    dynamic_cast<const CTrackContainer*>((*iter)->GetTrack());
                if (cont  &&  cont->GetAnnotLevel() == i) {
                    GetConfigMgr()->SyncSettings(m_TrackSettings,
                        cont->GetSubtrackProxies(), i);
                }
            }
        }
        GetConfigMgr()->SyncSettings(m_TrackSettings, GetSubtrackProxies(), -2);
    } else {
        GetConfigMgr()->SyncSettings(m_TrackSettings, GetSubtrackProxies(), -1);
    }

    path = kBaseKey + CGuiRegistry::kDecimalDot + CSGConfigUtils::TrackProfileKey() +
        CGuiRegistry::kDecimalDot;
    string profile = GetProfile();


    // clear original tracks
    view = registry.GetReadWriteView(path + profile + ".Tracks");
    CGuiRegistry::TKeys keys;
    view.GetTopKeys(keys);
    ITERATE (CGuiRegistry::TKeys, iter, keys) {
        view.DeleteField(iter->key);
    }

    // Use a different track profile name to differentiate it from the one
    // shipped with GBench so that users can restore/update the track 
    // profile to the original default one
    if ( !NStr::StartsWith(profile, CSGConfigUtils::UserSettingPrefix())) {
        // this is done only once when GBench creates GenomeWorkbench2 folder
        // and save user settings for the first time.
        profile = CSGConfigUtils::UserSettingPrefix() + profile;
        SetProfile(profile);
        m_gConfig->SetTrackProfile(profile);
        // track profile name changed. It needs to be saved.
        m_gConfig->SetDirty(true);
        m_gConfig->SaveSettings();
    }
    GetConfigMgr()->SaveTrackSettingsRecursive(m_TrackSettings, path + profile);
}


void CFeaturePanel::x_MakeTopLevelTracks(TTrackProxies& src_proxies,
                                         TTrackProxies& dst_proxies,
                                         int& order,
                                         const string& base_name)
{
    NON_CONST_ITERATE (TTrackProxies, iter, src_proxies) {
        CTempTrackProxy* t_proxy = iter->GetPointer();
        CLayoutTrack* track = t_proxy->GetTrack();
        if ( !track ) continue;
        CTrackContainer* c_track = dynamic_cast<CTrackContainer*>(track);
        if (c_track) {
            x_MakeTopLevelTracks(c_track->GetSubtrackProxies(),
                                 dst_proxies, order,
                                 (base_name.empty() ? "" : base_name + "---") + t_proxy->GetName());
        } else {
            t_proxy->SetOrder(order++);
            if (t_proxy->IsTempTrack()) {
                t_proxy->SetName(base_name + "---" + t_proxy->GetName());
            }
            dst_proxies.push_back(*iter);
            if (t_proxy->GetShown()) {
                track->SetIndent(m_IndentLevel + 1);
                Add(track);
            }
        }
    }
}


void CFeaturePanel::x_InitTrackIcons()
{
    static bool icon_image_registered = false;
    if ( !icon_image_registered  &&  !m_CgiMode) {
        CLayoutTrack::RegisterCommonIconImages();
        ITERATE (CTrackConfigManager::TFactoryMap, iter,
            GetConfigMgr()->GetTrackFactories()) {
            const IIconProvider* icon_track =
                dynamic_cast<const IIconProvider*>(iter->second.GetPointer());
            if (icon_track) {
                icon_track->RegisterIconImages();
            }
        }

        // register icon images for feature panel
        RegisterIconImage("track_color_theme", "track_color_theme.png");
        RegisterIconImage("track_size", "track_size.png");
        RegisterIconImage("track_decoration", "track_decoration.png");
        RegisterIconImage("track_label", "track_label.png");

        // lite versions
        RegisterIconImage("track_color_theme_lite", "track_color_theme_lite.png");
        RegisterIconImage("track_size_lite", "track_size_lite.png");
        RegisterIconImage("track_decoration_lite", "track_decoration_lite.png");
        RegisterIconImage("track_label_lite", "track_label_lite.png");
        icon_image_registered = true;
    }
}


void CFeaturePanel::x_RenderHairLineSelections() const
{
    if (!m_SelFeatures) {
        return;
    }
    // check hairline_options_dlg for values
    // 0. Show no hairlines with selections
    // 1. Show all hairlines with selections
    // 2. Show only hairlines that are not shared (i.e., places that are different)
    // 3. Show only hairlines that are shared by any two of the selections
    // 4. Show hairlines shared by *all* of the selections.

    int opt = m_gConfig->GetHairlineOption();
    if (opt == 0)
        return;

    // first, gather all From and To positions
    multiset <TSeqPos> all_pos;
    size_t total_glyph = 0;
    ITERATE (TSelectedGlyphs, iter, *m_SelFeatures) {
        CWeakRef<CSeqGlyph>::TRefType glyph = iter->Lock();
        if ( !glyph ) continue;

        const IObjectBasedGlyph* obj =
            dynamic_cast<const IObjectBasedGlyph*>(glyph.GetPointer());
        _ASSERT(obj);
        ITERATE (IObjectBasedGlyph::TIntervals, iter_i, obj->GetIntervals()) {
            const TSeqRange& curr = *iter_i;
            all_pos.insert(curr.GetFrom());
            all_pos.insert(curr.GetTo() + 1);
        }
        ++total_glyph;
    }  // done


    // finally, render hair-lines
    ITERATE (TSelectedGlyphs, iter, *m_SelFeatures) {
        CWeakRef<CSeqGlyph>::TRefType glyph = iter->Lock();
        if ( !glyph ) continue;

        TModelPoint top_left;
        glyph->GetPosInWorld(top_left);
        TModelUnit line_y = top_left.Y();
        const CFeatGlyph* feat =
            dynamic_cast<const CFeatGlyph*>(glyph.GetPointer());
        if (feat) {
            line_y += feat->GetBarCenter();
        }

        const IObjectBasedGlyph* obj =
            dynamic_cast<const IObjectBasedGlyph*>(glyph.GetPointer());
        ITERATE (IObjectBasedGlyph::TIntervals, iter_i, obj->GetIntervals()) {
            const TSeqRange& curr = *iter_i;
            x_DrawHairLine(opt, total_glyph, all_pos, curr.GetFrom(), line_y);
            x_DrawHairLine(opt, total_glyph, all_pos, curr.GetTo()+1, line_y);
        }
    }
}


void CFeaturePanel::x_DrawHairLine(int opt, size_t total, const multiset <TSeqPos>& all_pos,
        TSeqPos pos, TModelUnit line_y) const
{
    CGlRender& gl = GetGl();

    // use light color when no intersection, dark otherwise
    CRgbaColor c_light = m_gConfig->GetSelHairlineOneColor();
    CRgbaColor c_dark  = m_gConfig->GetSelHairlineManyColor();

    // if only one selected feature, we show all hairlines
    // This is for considering merged feature
    if (total == 1) {
        opt = 1;
    }

    bool need_line = true;
    CRgbaColor color;
    switch (opt) {
    case 1: // 1. Show all hairlines with selections
        color = all_pos.count(pos) > 1 ? c_dark : c_light;
        break;
    case 2: // 2. Show only hairlines that are not shared (i.e., places that are different)
        color = c_light;
        need_line = all_pos.count(pos) == 1;
        break;
    case 3: // 3. Show only hairlines that are shared by any two of the selections
        color = c_dark;
        need_line = all_pos.count(pos) >= 2;
        break;
    case 4: // 4. Show hairlines shared by *all* of the selections.
        color = c_dark;
        need_line = all_pos.count(pos) == total;
        break;
    }

    // render "From"
    if (need_line) {
        gl.ColorC(color);
        m_Context->DrawLine(pos, line_y, pos, GetTop());
        if (m_gConfig->GetShowDownwardHairlines()) {
            gl.ColorC(m_gConfig->GetSelHairlineDownColor());
            m_Context->DrawLine(pos, line_y, pos, GetBottom());
        }
    }
}


void CFeaturePanel::x_RenderFeatureGrid() const
{
    CGlRender& gl = GetGl();

    if ( !m_gConfig->GetShowHorzGrid()  ||  !x_NeedSpecialGrid()) return;

    CRgbaColor color = m_gConfig->GetVerticalGridColor();
    color.Darken(0.1f);
    gl.ColorC(color);

    ITERATE (TSelectedGlyphs, obj_iter, *m_SelCDSFeatures) {
        CWeakRef<CSeqGlyph>::TRefType glyph = obj_iter->Lock();
        const CFeatGlyph* feat =
            dynamic_cast<const CFeatGlyph*>(glyph.GetPointer());
        if ( !feat ) continue;

        TSeqRange range = feat->GetRange();

        // calculate grid spes for each CDS
        double length = range.GetTo() - range.GetFrom();
        // less than 3 bases, don't draw anything            
        if (length < 2) continue;

        double logRange = log10(length);
        logRange = ceil(logRange) - 1;
        double step = pow(10.0, logRange);

        // selecting base_step as step divided by 1, 2 or 5
        double base_step = step;
        double tick_num = length / base_step;
        int i=0;
        while(tick_num < 8) {
            base_step = step / kSteps[i++];
            tick_num = length / base_step;
        }

        double min_cell = m_Context->ScreenToSeq(kMinCellSize);
        double max_cell = m_Context->ScreenToSeq(kMaxCellSize);
        int grid_step = (int)x_CalcGridStep(base_step,  min_cell, max_cell);

        bool neg_strand = (sequence::GetStrand(feat->GetLocation()) == eNa_strand_minus);
        TModelPoint top_left;
        feat->GetPosInWorld(top_left);

        ITERATE (vector<TSeqRange>, iter, feat->GetIntervals()) {
            const TSeqRange& curr = *iter;
            TModelUnit f = curr.GetFrom();
            TModelUnit t = curr.GetTo();

            if (neg_strand) {
                int end = (int)floor(f - 1);
                for (i = (int)ceil(t); i >= end; i -= grid_step) {
                    m_Context->DrawLine(i + 1, GetBottom(), i + 1, GetTop());
                }
            } else {
                int end = (int)ceil(t + 1);
                for (i = (int)floor(f); i <= end; i += grid_step) {
                    m_Context->DrawLine(i, GetBottom(), i, GetTop());
                }
            }
        } // ITERATE
    } // ITERATE
}


double CFeaturePanel::x_CalcGridStep(double base_step,
                                     double min_step,
                                     double max_step) const
{
    if ((base_step >= min_step && base_step <= max_step) || (min_step == max_step))
        return base_step;
    else {
        // base_step has a form M * pow(10, P), where P is power
        double Power = log10(base_step);
        double P = ceil(Power) -1;
        double pow10P = pow(10.0, P);
        double M = base_step / pow10P;
        if (M >= 10) {
            M /= 10;
            pow10P *= 10;
        }

        int oldK = 1, K = 1, Index = 0;
        if(base_step < min_step) // reasing base_step to make it more then minPix
        {
            double minK = min_step / pow10P;
            while (K < minK)
            {
                if(Index <2)
                    K = oldK * kSteps[Index++];
                else {
                    K = oldK = oldK * 10;
                    Index = 0;
                }
            }
            base_step = pow10P * K;
        } else if (base_step > max_step) { // decreasing base_step to make it less then maxPix
            pow10P *= 10;
            double maxK = pow10P / max_step;
            while (K < maxK)
            {
                if(Index <2)
                    K = oldK * kSteps[Index++];
                else {
                    K = oldK = oldK * 10;
                    Index = 0;
                }
            }
            base_step = pow10P / K;
        }
        return base_step;
    }
}


bool CFeaturePanel::x_GetShowDisplayLevel(TLevel level) const
{
    return find(m_DisplayLevels.begin(), m_DisplayLevels.end(), level)
        != m_DisplayLevels.end();
}


void CFeaturePanel::x_SetShowDisplayLevel(TLevel level, bool show)
{
    TDisplayLevels::iterator iter =
        find(m_DisplayLevels.begin(), m_DisplayLevels.end(), level);
    if (show  &&  iter == m_DisplayLevels.end()) {
        m_DisplayLevels.push_back(level);
    } else if (!show  &&  iter != m_DisplayLevels.end()) {
        m_DisplayLevels.erase(iter);
    }
}


CRef<CTrackContainer> CFeaturePanel::x_CreateLevelTrack(int level,
                                                        int id,
                                                        const string& track_name,
                                                        bool shown)
{
    CRef<CTrackContainer> cont_track;

    const string& cont_key = CTrackContainerFactory::GetTypeInfo().GetId();
    const ILayoutTrackFactory* factory =
        GetConfigMgr()->GetTrackFactory(cont_key);
    if ( !factory ) return cont_track;

    // create track proxy for the track
    CRef<CTempTrackProxy> t_proxy(new CTempTrackProxy(id, track_name, shown));

    ILayoutTrackFactory::SExtraParams e_params(level, false, NULL);
    e_params.m_SkipGenuineCheck = GetSkipGenuineCheck();
    TTrackMap tracks = factory->CreateTracks(m_InputObj, m_DSContext,
        m_Context, e_params);
    
    CLayoutTrack* track = tracks.begin()->second;
    cont_track = dynamic_cast<CTrackContainer*>(track);
    cont_track->SetConfigMgr(GetConfigMgr());
    AddTracks(t_proxy, tracks);
    
    return cont_track;
}



END_NCBI_SCOPE
