/*  $Id: graph_track.cpp 27972 2013-05-03 20:34:20Z wuliangs $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Liangshou Wu
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>

#include <gui/widgets/seq_graphic/graph_track.hpp>
#include <gui/widgets/seq_graphic/seqgraphic_graph_ds.hpp>
#include <gui/widgets/seq_graphic/seqgraphic_job_result.hpp>
#include <gui/widgets/seq_graphic/seqgraphic_ds_manager.hpp>
#include <gui/widgets/seq_graphic/layout_track_proxy.hpp>
#include <gui/widgets/seq_graphic/comment_glyph.hpp>
#include <gui/widgets/seq_graphic/histogram_glyph.hpp>
#include <gui/widgets/seq_graphic/rendering_ctx.hpp>
#include <gui/widgets/wx/wx_utils.hpp>

#include <gui/objutils/utils.hpp>
#include <gui/objutils/label.hpp>

#include <wx/menu.h>


BEGIN_NCBI_SCOPE
USING_SCOPE(objects);


///////////////////////////////////////////////////////////////////////////////
///   CGraphTrack

static const string kTrackTitle = "Graphs";

CTrackTypeInfo CGraphTrack::m_TypeInfo("graph_track",
                                       "Various histograms related to the sequence.");

CGraphTrack::CGraphTrack(CSGGraphDS* ds, CRenderingContext* r_cntx)
    : CDataTrack(r_cntx)
    , m_DS(ds)
    , m_FixedScale(false)
{
    SetTitle(kTrackTitle);
    m_DS->SetJobListener(this);
    SetLayoutPolicy(m_Simple);

    x_RegisterIcon(SIconInfo(
        eIcon_Settings, "Settings", true, "track_settings"));

    // initialize annotation selector
    if (m_DS->IsSeqTable()) {
        m_Sel = CSeqUtils::GetAnnotSelector(CSeq_annot::C_Data::e_Seq_table);
    } else {
        m_Sel = CSeqUtils::GetAnnotSelector(CSeq_annot::C_Data::e_Graph);
    }
}


CGraphTrack::~CGraphTrack()
{
    if ( !m_DS->AllJobsFinished() ) {
        m_DS->DeleteAllJobs();
    }
}


const CTrackTypeInfo& CGraphTrack::GetTypeInfo() const
{
    return m_TypeInfo;
}


CRef<CSGGenBankDS> CGraphTrack::GetDataSource()
{
    return CRef<CSGGenBankDS>(m_DS.GetPointer());
}


CConstRef<CSGGenBankDS> CGraphTrack::GetDataSource() const
{
    return CConstRef<CSGGenBankDS>(m_DS.GetPointer());
}


void CGraphTrack::GetHTMLActiveAreas(TAreaVector* p_areas) const
{
    if (GetChildren().size() > 1) {
        // use the standard way
        CGlyphContainer::GetHTMLActiveAreas(p_areas);
    }
}


CHTMLActiveArea* CGraphTrack::InitHTMLActiveArea(TAreaVector* p_areas) const
{
    // for cases that there is only one graph, we initialize
    // it differently. 
    if (GetChildren().size() == 1) {
        TModelUnit tb_height = x_GetTBHeight();
        if (tb_height > 0.0) {
            // get child/graph's signature
            GetChildren().front()->GetHTMLActiveAreas(p_areas);
            // but use track range and size
            CHTMLActiveArea& area = p_areas->back();
            CSeqGlyph::x_InitHTMLActiveArea(area);
            TVPRect& bound = area.m_Bounds;
            bound.SetBottom(bound.Top() + 2 + (TVPUnit)tb_height);
            // add is_track flag
            area.m_Flags |= CHTMLActiveArea::fTrack; 
            return &area;
        }

        return NULL;
    }
    
    return CLayoutTrack::InitHTMLActiveArea(p_areas);
}


void CGraphTrack::SetAnnot(const string& annot)
{
    m_AnnotName = annot;
    m_Sel.ResetAnnotsNames();
    if (CSeqUtils::NameTypeStrToValue(m_AnnotName) ==
        CSeqUtils::eAnnot_Unnamed) {
        m_Sel.AddUnnamedAnnots();
    } else {
        m_Sel.AddNamedAnnots(m_AnnotName);
        if (NStr::StartsWith(m_AnnotName, "NA0")) {
            if (m_AnnotName.find(".") == string::npos) {
                m_AnnotName += ".1";
            }
            m_Sel.IncludeNamedAnnotAccession(m_AnnotName);
        }
    }
}


void CGraphTrack::x_LoadSettings(const string& /*preset_style*/,
                                 const TKeyValuePairs& settings)
{
    ITERATE (TKeyValuePairs, iter, settings) {
        try {
            if (NStr::EqualNocase(iter->first, "fixed_scale")) {
                m_FixedScale = NStr::StringToBool(iter->second);
            }
        } catch (std::exception&) {
            LOG_POST(Error << "CGraphTrack::x_LoadSettings: invalid settings "
                     << iter->second);
        }
    }
}


void CGraphTrack::x_SaveSettings(const string& /*preset_style*/)
{
    SetProfile("fixed_scale:" + NStr::BoolToString(m_FixedScale));
}


void CGraphTrack::x_OnIconClicked(TIconID id)
{
    switch (id) {
    case eIcon_Settings:
        x_OnSettingsIconClicked();
        break;
    default:
        // use default handlers
        CLayoutTrack::x_OnIconClicked(id);
    }
}


void CGraphTrack::x_UpdateData()
{
    m_DS->DeleteAllJobs();
    x_SetStartStatus();
    m_DS->LoadSeqGraphs(m_Sel, m_Context->GetVisSeqRange(),
        m_Context->GetScale(), m_FixedScale);
}


void CGraphTrack::x_OnJobCompleted(CAppJobNotification& notify)
{
    m_DS->ClearJobID(notify.GetJobID());
    CRef<CObject> res_obj = notify.GetResult();
    CSGJobResult* result = dynamic_cast<CSGJobResult*>(&*res_obj);
    if (!result) {
        ERR_POST("CGraphTrack::x_OnJobCompleted() notification for job \
                does not contain results.");
        return;
    }

    SetGroup().Clear();
    SetMsg("");
    size_t count = result->m_ObjectList.size();
    if (count > 0) {
        TObjects objs = result->m_ObjectList;
        vector<string> titles;
        NON_CONST_ITERATE (CSeqGlyph::TObjects, iter, objs) {
            CHistogramGlyph* hist =
               dynamic_cast<CHistogramGlyph*>(iter->GetPointer());
            hist->SetConfigManager(m_gConfig->GetHistParamsManager());
            string title = hist->GetTitle();
            title += ", Max:" + NStr::UIntToString((int)hist->GetAxisMax());
            title += " / Min:" + NStr::UIntToString((int)hist->GetAxisMin());
            titles.push_back(title);
        }
        if (count == 1) {
            SetTitle(titles[0]);
            SetObjects(objs);
        } else {
            string msg = ", " + NStr::SizetToString(count) + " graph";
            msg += (count != 1 ? "s shown" : " shown");
            SetMsg(msg);
            SetTitle(m_AnnotName);
            int i = 0;
            CRef<CCommentConfig> config(new CCommentConfig());
            config->m_LabelColor = CRgbaColor("blue");
            config->m_ShowConnection = false;
            NON_CONST_ITERATE (CSeqGlyph::TObjects, iter, objs) {
                CRef<CLayoutGroup> group(new CLayoutGroup);
                Add(group);
                group->SetLayoutPolicy(m_Simple);
                CRef<CCommentGlyph> label(
                    new CCommentGlyph(titles[i++], *iter));
                label->SetConfig(config);
                group->PushBack(label);
                group->PushBack(*iter);
            }
            SetLayoutPolicy(m_Layered);
        }
    }

    x_UpdateLayout();
}


void CGraphTrack::x_OnSettingsIconClicked()
{
    wxMenu menu;
    UseDefaultMarginWidth(menu);
    int id_base = 10000;
    wxMenuItem* item = menu.AppendCheckItem(id_base, wxT("Fixed scale"));
    if (m_FixedScale) {
        item->Check();
    }
    m_LTHost->LTH_PopupMenu(&menu);

    if (m_FixedScale != item->IsChecked()) {
        m_FixedScale = item->IsChecked();
        x_UpdateData();
    }
}


///////////////////////////////////////////////////////////////////////////////
/// CGraphTrackFactory methods
static const string kSeqTable = "seq-table";
static const string kSeqGraph = "seq-graph";

ILayoutTrackFactory::TTrackMap
CGraphTrackFactory::CreateTracks(SConstScopedObject& object,
                                 ISGDataSourceContext* ds_context,
                                 CRenderingContext* r_cntx,
                                 const SExtraParams& params,
                                 const TAnnotMetaDataList& src_annots) const
{
    // we need to separate tracks stored as seq-graph from seq-table
    // to initialize then differently.
    typedef map<string, TAnnotNameTitleMap> TAnnotGroup;
    TAnnotGroup annot_group;
    annot_group.insert(TAnnotGroup::value_type(kSeqTable, TAnnotNameTitleMap()));
    annot_group.insert(TAnnotGroup::value_type(kSeqGraph, TAnnotNameTitleMap()));

    if (params.m_SkipGenuineCheck  &&  !params.m_Annots.empty()) {
        // filter out the seq-table annots
        TAnnotNameTitleMap seqtable_annots;
        SAnnotSelector sel(CSeqUtils::GetAnnotSelector(params.m_Annots));
        sel.ExcludeNamedAnnots("SNP");
        CIRef<ISGDataSource> ds = ds_context->GetDS(
            typeid(CSGGraphDSType).name(), object);
        CSGGraphDS* graph_ds = dynamic_cast<CSGGraphDS*>(ds.GetPointer());
        graph_ds->SetDepth(params.m_Level);
        graph_ds->SetAdaptive(params.m_Adaptive);
        graph_ds->GetAnnotNames(sel, TSeqRange::GetWhole(), seqtable_annots, true);

        ITERATE (SExtraParams::TAnnots, iter, params.m_Annots) {
            if (seqtable_annots.count(*iter) > 0) {
                annot_group[kSeqTable].insert(TAnnotNameTitleMap::value_type(*iter, ""));
            } else {
                annot_group[kSeqGraph].insert(TAnnotNameTitleMap::value_type(*iter, ""));
            }
        }
    } else {
        // collect non-NA tracks
        CIRef<ISGDataSource> ds = ds_context->GetDS(
            typeid(CSGGraphDSType).name(), object);
        CSGGraphDS* graph_ds = dynamic_cast<CSGGraphDS*>(ds.GetPointer());
        graph_ds->SetDepth(params.m_Level);
        graph_ds->SetAdaptive(params.m_Adaptive);
        
        SAnnotSelector sel(CSeqUtils::GetAnnotSelector(params.m_Annots));
        sel.ExcludeNamedAnnots("SNP");
        graph_ds->GetAnnotNames(sel, r_cntx->GetVisSeqRange(), annot_group[kSeqGraph]);

        // collect NA tracks
        // For graph track we don't collect NA tracks if it is
        // not explicitly specified due to potential large number
        // of NA tracks.
        if ( !src_annots.empty()  &&  !params.m_Annots.empty() ) {
            GetMatchedAnnots(src_annots, params.m_Annots, "graph", "", annot_group[kSeqGraph]);
        }


        if (params.m_Annots.empty()) {
            sel = CSeqUtils::GetAnnotSelector(CSeq_annot::C_Data::e_Seq_table);
            sel.ExcludeNamedAnnots("SNP");
            graph_ds->GetSeqtableAnnots(sel, r_cntx->GetVisSeqRange(), annot_group[kSeqTable]);
        } else {
            sel = CSeqUtils::GetAnnotSelector(params.m_Annots);
            sel.ExcludeNamedAnnots("SNP");
            graph_ds->GetAnnotNames(sel, r_cntx->GetVisSeqRange(), annot_group[kSeqTable], true);
        }        
    }

    // create feature tracks
    TTrackMap tracks;
    ITERATE (TAnnotGroup, g_iter, annot_group) {
        const TAnnotNameTitleMap& annots = g_iter->second;
        ITERATE (TAnnotNameTitleMap, iter, annots) {
            CIRef<ISGDataSource> ds =
                ds_context->GetDS(typeid(CSGGraphDSType).name(), object);
            CSGGraphDS* graph_ds = dynamic_cast<CSGGraphDS*>(ds.GetPointer());
            graph_ds->SetDepth(params.m_Level);
            graph_ds->SetAdaptive(params.m_Adaptive);
            graph_ds->SetSeqTable(g_iter->first == kSeqTable);
            // Aggregated coverage graphs are not correctly computed
            // for graph data stored as seq-table currently. Disable it
            // for now.
            //graph_ds->SetGraphLevels(iter->first);

            CRef<CGraphTrack> track(new CGraphTrack(graph_ds, r_cntx));
            track->SetAnnot(iter->first);
            if ( !iter->second.empty() ) {
                track->SetTitle(iter->second);
            }
            tracks[iter->first] = track.GetPointer();
        }
    }

    return tracks;
}


string CGraphTrackFactory::GetExtensionIdentifier() const
{
    return CGraphTrack::m_TypeInfo.GetId();
}


string CGraphTrackFactory::GetExtensionLabel() const
{
    return CGraphTrack::m_TypeInfo.GetDescr();
}


CRef<CTrackConfigSet>
CGraphTrackFactory::GetSettings(CSeqGraphicConfig* /*g_conf*/,
                                const string& /*profile*/,
                                const TKeyValuePairs& settings,
                                const CTempTrackProxy* /*track_proxy*/) const
{
    CRef<CTrackConfigSet> config_set(new CTrackConfigSet);
    CRef<CTrackConfig> config(new CTrackConfig);
    config_set->Set().push_back(config);

    bool fixed_scale = false;
    ITERATE (TKeyValuePairs, iter, settings) {
        try {
            if (NStr::EqualNocase(iter->first, "fixed_scale")) {
                fixed_scale = NStr::StringToBool(iter->second);
            }
        } catch (std::exception&) {
            // ignore the errors
        }
    }


    config->SetCheck_boxes().push_back(
        CTrackConfigUtils::CreateCheckBox(
            "fixed_scale", "Fixed Scale", "The scale is fixed so the highest value shown \
encompasses the highest value in the data track (global) or the highest \
value shown encompasses the highest value in the range on screen (local).",
            "",
            fixed_scale));

    config->SetHelp() = GetThisTypeInfo().GetDescr();

    return config_set;
}


END_NCBI_SCOPE
