/*  $Id: histogram_glyph.cpp 27585 2013-03-08 20:58:05Z falkrb $
* ===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data, the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties, express or implied, including
*  warranties of performance, merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================
*
* Authors:  Vlad Lebedev, Liangshou Wu
*
* File Description:
*    CHistogramGlyph -- utility class for having
*                      feature density histograms his in graphical layouts.
*
*/

#include <ncbi_pch.hpp>
#include <gui/widgets/seq_graphic/histogram_glyph.hpp>
#include <gui/widgets/seq_graphic/rendering_ctx.hpp>
#include <gui/objutils/obj_fingerprint.hpp>
#include <gui/objutils/utils.hpp>
#include <gui/opengl/glrender.hpp>
#include <gui/opengl/glutils.hpp>
#include <gui/opengl/gltexturefont.hpp>
#include "wx_histogram_config_dlg.hpp"
#include <gui/widgets/wx/message_box.hpp>


BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

//
// CHistogramGlyph::CHistogramGlyph()
//
CHistogramGlyph::CHistogramGlyph(const TMap& map,
                                 CSeqFeatData::ESubtype subtype, const string& title)
    : m_DlgHost(NULL)
    , m_AnnotName(CSeqUtils::GetUnnamedAnnot())
    , m_Subtype(subtype)
    , m_ShowTitle(false)
{
    string desc = title;
    if (desc.empty()) {
        const CFeatList& feat_list = *CSeqFeatData::GetFeatList();
        desc = feat_list.GetDescription(
            CSeqFeatData::GetTypeFromSubtype(subtype), subtype);
    }
    m_Maps.insert(TMaps::value_type(desc, map));
    m_AxisMax = ceil(GetMax());
    m_AxisMin = floor(GetMin());
}


CHistogramGlyph::CHistogramGlyph(const TMap& map, const string& title)
    : m_DlgHost(NULL)
    , m_AnnotName(CSeqUtils::GetUnnamedAnnot())
    , m_Subtype(CSeqFeatData::eSubtype_any)
    , m_ShowTitle(false)
{
    m_Maps.insert(TMaps::value_type(title, map));
    m_AxisMax = ceil(GetMax());
    m_AxisMin = floor(GetMin());
}


CHistogramGlyph::CHistogramGlyph(const TMaps& maps,
                                 CSeqFeatData::ESubtype subtype)
    : m_DlgHost(NULL)
    , m_AnnotName(CSeqUtils::GetUnnamedAnnot())
    , m_Subtype(subtype)
    , m_ShowTitle(false)
{
    m_Maps = maps;
    m_AxisMax = ceil(GetMax());
    m_AxisMin = floor(GetMin());
}


CHistogramGlyph::CHistogramGlyph(const TMaps& maps, const string& name)
    : m_DlgHost(NULL)
    , m_AnnotName(name)
    , m_Subtype(CSeqFeatData::eSubtype_any)
    , m_ShowTitle(false)
{
    m_Maps = maps;
    m_AxisMax = ceil(GetMax());
    m_AxisMin = floor(GetMin());
}


bool CHistogramGlyph::OnLeftDblClick(const TModelPoint& /*p*/)
{
    string name;
    if (GetFeatSubtype() != CSeqFeatData::eSubtype_any) {
        const CFeatList& feats(*CSeqFeatData::GetFeatList());
        name = feats.GetStoragekey(GetFeatSubtype());
    } else {
        name = GetAnnotName();
    }
    
    NStr::ReplaceInPlace(name, ".", "_");
    if ( !m_ConfigMgr->IsTempSettings(name) ) {
        CHistConfigDlg dlg;
        dlg.SetConfig(x_GetConfig());
        dlg.SetConfigName(name);
        dlg.CreateX(NULL);
        if (m_DlgHost) {
            m_DlgHost->PreDialogShow();
        }
        if(dlg.ShowModal() == wxID_OK) {
            m_ConfigMgr->AddSettings(name, dlg.GetConfig());
            x_OnLayoutChanged();
        }
        if (m_DlgHost) {
            m_DlgHost->PostDialogShow();
        }
    }
    return true;
}


bool CHistogramGlyph::NeedTooltip(const TModelPoint& p, string& tt, string& t_title) const
{
    GetTooltip(p, tt, t_title);
    return true;
}


void CHistogramGlyph::GetTooltip(const TModelPoint& /*p*/, string& tt, string& /*t_title*/) const
{
    if (GetFeatSubtype() != CSeqFeatData::eSubtype_any) {
        tt = CSeqFeatData::GetFeatList()->GetDescription(GetFeatType(), GetFeatSubtype());
        tt += " distribution histogram";
        if (!CSeqUtils::IsUnnamed(GetTitle())) {
            tt += " (" + GetTitle() + ")";
        }
    } else { // it is graph
        if ( !GetDesc().empty() ) {
            tt = GetDesc();
            NStr::ReplaceInPlace(tt, "&#xA;", "\n");
            NStr::ReplaceInPlace(tt, "&#xa;", "\n");
        } else {
            tt = "Graph: " + GetTitle();
        }
    }
    tt += "\n\nDouble click on the histogram to change settings";
}


void CHistogramGlyph::GetHTMLActiveAreas(TAreaVector* p_areas) const
{
    CHTMLActiveArea area;
    CSeqGlyph::x_InitHTMLActiveArea(area);
    area.m_PositiveStrand = true;

    CSeq_loc loc;
    CSeq_interval& ival = loc.SetInt();
    CRef<CSGSequenceDS> seq_ds = m_Context->GetSeqDS();
    CConstRef<CSeq_id> seq_id = seq_ds->GetBioseqHandle().GetSeqId();
    ival.SetId().Assign(*seq_id);

    TMaps::const_iterator first_map = m_Maps.begin();
    int subtype = 0;
    CObjFingerprint::EObjectType type = CObjFingerprint::eUnknown;
    if (GetFeatSubtype() != CSeqFeatData::eSubtype_any) {
        area.m_Flags =
            CHTMLActiveArea::fNoSelection |
            CHTMLActiveArea::fTooltipEmbedded |
            CHTMLActiveArea::fNoPin;
        area.m_Descr =
            CSeqFeatData::GetFeatList()->GetDescription(GetFeatType(), GetFeatSubtype());
        area.m_Descr += " distribution histogram";
        if (CSeqUtils::IsUnnamed(GetTitle())) {
            area.m_Descr += " (" + GetTitle() + ")";
        }
        type = CObjFingerprint::eFeature;
        subtype = GetFeatSubtype();

        ival.SetFrom(first_map->second.GetStart());
        ival.SetTo(first_map->second.GetStop());
    } else {
        area.m_Flags = CHTMLActiveArea::fNoSelection;
        type = CObjFingerprint::eGraph;
        ival.SetFrom(0);
        ival.SetTo(seq_ds->GetSequenceLength() - 1);
    }

    area.m_Signature =
        CObjFingerprint::GetHistogramSignature(loc, type, &m_Context->GetScope(),
                                               GetTitle(), GetAnnotName(), subtype);
    if (type == CObjFingerprint::eFeature) {
        area.m_Signature = "fake|" + area.m_Signature;
    }

    p_areas->push_back(area);
}


bool CHistogramGlyph::IsClickable() const
{
    string name;
    if (GetFeatSubtype() != CSeqFeatData::eSubtype_any) {
        const CFeatList& feats(*CSeqFeatData::GetFeatList());
        name = feats.GetStoragekey(GetFeatSubtype());
    } else {
        name = GetAnnotName();
    }
    
    NStr::ReplaceInPlace(name, ".", "_");
    if (m_ConfigMgr->IsTempSettings(name)) {
        return false;
    }

    return true;
}


objects::CSeqFeatData::E_Choice CHistogramGlyph::GetFeatType() const
{
    return objects::CSeqFeatData::GetTypeFromSubtype(m_Subtype);
}


void CHistogramGlyph::x_Draw() const
{
    CGlRender& gl = GetGl();

    if (m_Maps.empty()) {
        return;
    }

    _ASSERT(m_Context);

    TModelRect rcm = GetModelRect();
    TModelUnit top = GetTop();
    TModelUnit bottom = top + GetHeight();

    CRef<CHistParams> config = x_GetConfig();
     if (config->m_DrawBg) {
        gl.ColorC(config->m_bgColor);
        m_Context->DrawQuad(rcm);
    }

    TSeqRange range = GetRange();
    TModelUnit two_pix_size = m_Context->ScreenToSeq(2.0);
    if (range.GetLength() < two_pix_size) {
        // bar is less then 2 pixel. Just draw a background
        gl.ColorC(config->m_fgColor);
        m_Context->DrawQuad(range.GetFrom(), top,
            range.GetFrom() + two_pix_size, bottom);
        return;
    }

    CGlAttrGuard AttrGuard(GL_LINE_BIT);
    gl.Disable(GL_LINE_SMOOTH);

    if (config->m_NeedRuler  &&
        (config->m_Type == CHistParams::eHistogram  ||
         config->m_Type == CHistParams::eLineGraph) ) {
        int tick_n = 2;
        TModelUnit dist_y = GetHeight() / tick_n;
        const CRgbaColor& r_c = config->m_RulerColor;
        gl.Color4f(r_c.GetRed(), r_c.GetGreen(), r_c.GetBlue(), 0.4f);
        m_Context->DrawLine(GetLeft(), top + dist_y, GetRight(), top + dist_y);
        gl.Color4f(r_c.GetRed(), r_c.GetGreen(), r_c.GetBlue(), 0.4f);
        m_Context->DrawLine(GetLeft(), top, GetRight(), top);

        int ruler_n = 4;
        TModelUnit dist_x = GetWidth() / ruler_n;
        TModelUnit min_dist = m_Context->ScreenToSeq(300.0);
        if (dist_x < min_dist) {
            dist_x = min_dist;
            ruler_n = (int)(GetWidth() / dist_x);
        }
        TModelUnit label_off = m_Context->ScreenToSeq(1.0);
        //CGlBitmapFont font(CGlBitmapFont::eHelvetica8);
        CGlTextureFont font(CGlTextureFont::eFontFace_Helvetica, 8);
        TModelUnit x = GetLeft() + dist_x * 0.1;
        for (;  x < GetRight();  x += dist_x) {
            gl.Color4f(r_c.GetRed(), r_c.GetGreen(), r_c.GetBlue(), 0.5f);
            m_Context->DrawLine(x, top, x, bottom);
            gl.ColorC(config->m_LabelColor);
            m_Context->TextOut(&font, NStr::DoubleToString(GetAxisMax()).c_str(),
                x + label_off, top + font.TextHeight() + 1, false);
            m_Context->TextOut(&font, NStr::DoubleToString(GetAxisMin()).c_str(),
                x + label_off,
                bottom - 1, false);
        }
    }

    float scale = GetAxisMax() - GetAxisMin();
    if (scale > 0.0) {
        // do pre-division
        scale = 1.0f / scale;
    }

    bool multi_map = m_Maps.size() > 1;
    ITERATE (TMaps, map_iter, m_Maps) {
        const TMap& densityMap = map_iter->second;
        if (config->m_Type == CHistParams::eSmearBar  ||
            config->m_Type == CHistParams::eMergedBar) {
            // draw smear bar
            CRgbaColor color_max = config->m_SmearColorMax;
            CRgbaColor color_min = config->m_SmearColorMin;
            float l_scale = densityMap.GetMax() - densityMap.GetMin();
            if (l_scale > 0.0) {
                // do pre-division
                l_scale = 1.0f / l_scale;
            }
            const CHistParams::TColorSet::const_iterator c_iter =
                config->m_Colors.find(map_iter->first);
            if (c_iter != config->m_Colors.end()) {
                color_max = c_iter->second;
                color_max.SetAlpha(0.9f);
                color_min = color_max;
                color_min.Lighten(0.5f);
                color_min.SetAlpha(0.9f);
            }
            TMap::runlen_iterator seg_it = densityMap.RunLenBegin();
            for (; seg_it; seg_it.Advance()) {
                TModelUnit  value = seg_it.GetValue();
                if (value == 0)
                    continue;

                TSeqPos f   = seg_it.GetSeqPosition();
                TSeqPos t   = seg_it.GetSeqRunEndPosition();

                float score = (value - GetAxisMin()) * l_scale;
                score = max(0.0f, score);
                score = min(1.0f, score);
                CRgbaColor color(CRgbaColor::Interpolate(color_max, color_min, score));
                gl.ColorC(color);

                // draw our segment as quad
                m_Context->DrawQuad(f, top, t, bottom);
            }
        } else if (config->m_Type == CHistParams::eHistogram) {
            // draw histogram
            if (multi_map  &&  config->m_Colors.count(map_iter->first) > 0) {
                CHistParams::TColorSet::const_iterator c_iter =
                    config->m_Colors.find(map_iter->first);
                gl.ColorC(c_iter->second);
            } else {
                gl.ColorC(config->m_fgColor);
            }
            gl.PushMatrix();
            gl.Translatef(0.0f, bottom, 0.0f);
            m_Context->DrawLine(rcm.Left(), 0.0, rcm.Right(), 0.0);
            gl.Scalef(1.0f, scale * GetHeight(), 1.0f);
            TMap::runlen_iterator seg_it = densityMap.RunLenBegin();
            for (; seg_it; seg_it.Advance()) {
                TModelUnit  value = seg_it.GetValue();
                if (value == 0)
                    continue;

                TSeqPos f   = seg_it.GetSeqPosition();
                TSeqPos t   = seg_it.GetSeqRunEndPosition();
                TModelUnit yy1 = value - GetAxisMin();

                // draw our segment as quad
                m_Context->DrawQuad(f, -yy1, t, 0.0);
            }
            gl.PopMatrix();
        } else { // draw line graph
            if (multi_map  &&  config->m_Colors.count(map_iter->first) > 0) {
                CHistParams::TColorSet::const_iterator c_iter =
                    config->m_Colors.find(map_iter->first);
                gl.ColorC(c_iter->second);
            } else {
                gl.ColorC(config->m_fgColor);
            }
            gl.PushMatrix();
            gl.Translatef(0.0f, bottom, 0.0f);
            m_Context->DrawLine(rcm.Left(), 0.0, rcm.Right(), 0.0);
            gl.Scalef(1.0f, - scale * GetHeight(), 1.0f);
            gl.Begin(GL_LINE_STRIP);
            TMap::runlen_iterator seg_it = densityMap.RunLenBegin();
            for (; seg_it; seg_it.Advance()) {
                TSeqPos f   = seg_it.GetSeqPosition();
                TSeqPos t   = seg_it.GetSeqRunEndPosition();;
                TModelUnit y_curr = seg_it.GetValue() - GetAxisMin();
                gl.Vertex2d(f - m_Context->GetOffset() + 0.5, y_curr);
                if (t != f) {
                    gl.Vertex2d(t - m_Context->GetOffset() + 0.5, y_curr);
                }
            }
            gl.End();
            gl.PopMatrix();
        }
    }

    //// draw the graph title
    //x_Color(CSeqGraphicColorConfig::eLabel_SeqGraph);
    //if (x_LabelWillFit(inrc, horz, flip_strand)) {
    //    TModelUnit widthP = horz ? fabs(inrc.Width()) / m_ScaleX : fabs(inrc.Height()) / m_ScaleY;
    //    string out = m_Font_Helv10.Truncate(title.c_str(), widthP);
    //    TModelUnit fs = m_Font_Helv10.TextHeight();
    //    TModelUnit xM = horz ? inrc.CenterPoint().X() : rcm.Right() + fs * (map_idx + 1);
    //    TModelUnit yM = horz ? rcm.Top() + fs * (map_idx + 1) : inrc.CenterPoint().Y();
    //    x_TextOut(&m_Font_Helv10, out.c_str(), horz, true, xM, yM);
    //}

    // Draw selection
    if (IsSelected()) {
        m_Context->DrawSelection(rcm);
    }
    gl.Enable(GL_LINE_SMOOTH);
}

void CHistogramGlyph::x_UpdateBoundingBox()
{
    CRef<CHistParams> config = x_GetConfig();
    m_Height = config->m_Height;
    TSeqRange range = GetRange();
    m_Width = range.GetLength();
    m_Pos.m_X = range.GetFrom();
}


END_NCBI_SCOPE
