/*  $Id: layout_group.cpp 27503 2013-02-26 18:16:55Z falkrb $
* ===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data, the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties, express or implied, including
*  warranties of performance, merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================
*
* Authors:  Liangshou Wu
*
*/

#include <ncbi_pch.hpp>
#include <gui/gui.hpp>
#include <gui/widgets/seq_graphic/layout_group.hpp>
#include <gui/widgets/seq_graphic/simple_layout_policy.hpp>
#include <gui/widgets/seq_graphic/rendering_ctx.hpp>
#include <gui/opengl/glrender.hpp>
#include <corelib/ncbi_limits.hpp>

BEGIN_NCBI_SCOPE

static const int kPadding = 2;

CLayoutGroup::CLayoutGroup()
{}

int CLayoutGroup::GetIndex(const CSeqGlyph* obj) const
{
    int i = 0;
    CConstRef<CSeqGlyph> child;
    TObjectList::const_iterator iter = m_Children.begin();
    while (iter != m_Children.end()) {
        if (obj == iter->GetPointer()) {
            break;
        }
        ++iter;
        ++i;
    }
    if (iter == m_Children.end()) {
        i = -1;
    }
    return i;    
}


/// Get the layout object at index 'idx'.
CConstRef<CSeqGlyph> CLayoutGroup::GetChild(int idx) const
{
    int i = 0;
    CConstRef<CSeqGlyph> child;
    TObjectList::const_iterator iter = m_Children.begin();
    while (iter != m_Children.end()) {
        if (i == idx) {
            // got what we want
            child.Reset(iter->GetPointer());
            break;
        }
        ++iter;
        ++i;
    }
    return child;
}


CRef<CSeqGlyph> CLayoutGroup::GetChild(int idx)
{
    int i = 0;
    CRef<CSeqGlyph> child;
    TObjectList::iterator iter = m_Children.begin();
    while (iter != m_Children.end()) {
        if (i == idx) {
            // got what we want
            child.Reset(iter->GetPointer());
            break;
        }
        ++iter;
        ++i;
    }
    return child;
}



/// Remove a layout object.
bool CLayoutGroup::Remove(CSeqGlyph* obj)
{
    NON_CONST_ITERATE (TObjectList, iter, m_Children) {
        if (obj == iter->GetPointer()) {
            m_Children.erase(iter);
            return true;
        }
    }
    return false;
}


bool CLayoutGroup::Replace(CSeqGlyph* obj, int idx)
{
    int i = 0;
    TObjectList::iterator iter = m_Children.begin();
    while (iter != m_Children.end()) {
        if (i == idx) {
            // got what we want
            (*iter).Reset(obj);
            obj->SetParent(this);
            if (m_Context) obj->SetRenderingContext(m_Context);
            return true;
        }
        ++iter;
        ++i;
    }
    return false;
}


/// Insert a layout object before the given index.
void CLayoutGroup::Insert(int at, CSeqGlyph* obj)
{
    int i = 0;
    CConstRef<CSeqGlyph> child;
    TObjectList::iterator iter = m_Children.begin();
    while (iter != m_Children.end()  &&  i != at) {
        ++iter;
        ++i;
    }

    m_Children.insert(iter, CRef<CSeqGlyph>(obj));
    obj->SetParent(this);
    if (m_Context) obj->SetRenderingContext(m_Context);
}

/// Insert a layout object before the given layout object.
void CLayoutGroup::Insert(const CSeqGlyph* at_obj, CSeqGlyph* obj)
{
    TObjectList::iterator iter = m_Children.begin();
    while (iter != m_Children.end()  &&  at_obj != iter->GetPointer()) {
        ++iter;
    }

    m_Children.insert(iter, CRef<CSeqGlyph>(obj));
    obj->SetParent(this);
    if (m_Context) obj->SetRenderingContext(m_Context);
}

/// Insert a layout object after the given layout object.
void CLayoutGroup::InsertAft(const CSeqGlyph* at_obj, CSeqGlyph* obj)
{
    obj->SetParent(this);
    if (m_Context) obj->SetRenderingContext(m_Context);

    NON_CONST_ITERATE (TObjectList, iter, m_Children) {
        if (at_obj == iter->GetPointer()) {
            m_Children.insert(++iter, CRef<CSeqGlyph>(obj));
            return;
        }
    }
    PushBack(obj);
}


void CLayoutGroup::Update(bool layout_only)
{
    NON_CONST_ITERATE (TObjectList, iter, m_Children) {
        (*iter)->Update(layout_only);
    }
    x_UpdateBoundingBox();
}


CRef<CSeqGlyph> CLayoutGroup::HitTest(const TModelPoint& p)
{
    CRef<CSeqGlyph> glyph;
    if (IsIn(p)) {
        TModelPoint pp(p);
        x_Parent2Local(pp);
        NON_CONST_ITERATE (TObjectList, iter, m_Children) {
            glyph = (*iter)->HitTest(pp);
            if (glyph) {
                break;
            }
        }
    }
    return glyph;
}


bool CLayoutGroup::Intersects(const TModelRect& rect,
                              TConstObjects& objs) const
{
    if (rect.Intersects(GetModelRect())) {
        TModelRect rc(rect);
        x_Parent2Local(rc);
        ITERATE (TObjectList, iter, m_Children) {
            (*iter)->Intersects(rc, objs);
        }
        return true;
    }
    return false;
}


void CLayoutGroup::GetHTMLActiveAreas(TAreaVector* p_areas) const
{
    try {
        ITERATE (TObjectList, iter, m_Children) {
            (*iter)->GetHTMLActiveAreas(p_areas);
        }
    } catch (CException& e) {
        LOG_POST(Error 
            << "CLayoutGroup::GetHTMLActiveAreas() failure on getting HTML active areas, error: "
            << e.GetMsg());
    }
}


bool CLayoutGroup::Accept(IGlyphVisitor* visitor)
{
    bool cont = visitor->Visit(this);
    TObjectList::iterator iter = m_Children.begin();
    while (cont  &&  iter != m_Children.end()) {
        cont = (*iter)->Accept(visitor);
        ++iter;
    }
    return cont;
}


void CLayoutGroup::x_Draw() const
{
    if (x_Visible_V()) {
        CGlRender& gl = GetGl();

        gl.PushMatrix();
        gl.Translated(0.0, GetTop(), 0.0f);
        x_DrawBoundary();
        ITERATE (TObjectList, iter, m_Children) {
            (*iter)->Draw();
        }
        gl.PopMatrix();
    }
}


void CLayoutGroup::x_UpdateBoundingBox()
{
    m_Height = 0.0;
    m_Width = 0.0;
    ILayoutPolicy::SBoundingBox bound;
    if (m_Policy) {
        size_t size = m_Children.size();
        m_Policy->BuildLayout(m_Children, bound);

        if (size != m_Children.size()) {
            // HACK! HACK! HACK!
            // The object hierachy is changed.  We need to do something
            // create a group to hold all the sub-groups
            TObjectList sub_groups = GetChildren();
            Clear();
            CLayoutGroup* group  = new CLayoutGroup;
            PushBack(group);
            group->Set(sub_groups);
            CSimpleLayout* simple = new CSimpleLayout;
            group->SetLayoutPolicy(simple);
            simple->SetVertSpace(0);
            simple->SetTopMargin(0);
            simple->BuildLayout(sub_groups, bound);
            group->SetLeft(bound.m_X);
            group->SetHeight(bound.m_Height);
            group->SetWidth(bound.m_Width);
            group->SetTop(0.0);
        }

        m_Height = bound.m_Height;
        m_Width = bound.m_Width;
        m_Pos.m_X = bound.m_X;
    }

    int padding_y = kPadding;
    if (m_Config) {
        padding_y = m_Config->GetPadding();
    }
    TModelUnit padding_x = m_Context->ScreenToSeq(padding_y);;
    m_Pos.m_X -= padding_x;
    m_Width += padding_x * 2;
    m_Height += padding_y;
}


void CLayoutGroup::x_DrawBoundary() const
{
    CGlRender& gl = GetGl();

    if (m_Config) {
        TModelRect rect(GetLeft(), 0.0, GetRight(), GetHeight());
        if (m_Config->ShowBackground()) {
            gl.ColorC(m_Config->GetBgColor());
            m_Context->DrawQuad(rect);
        }
        if (m_Config->ShowBoundary()) {
            gl.ColorC(m_Config->GetFgColor());
            gl.LineWidth(m_Config->GetLineWidth());
            m_Context->DrawRect(rect);
            gl.LineWidth(1.0f);
        }
        if (IsSelected()) {
            m_Context->DrawSelection(rect);
        }
    }
}


END_NCBI_SCOPE

