/*  $Id: layout_track_proxy.cpp 26453 2012-09-19 13:35:46Z wuliangs $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Liangshou Wu
 *
 */

/// @file
///

#include <ncbi_pch.hpp>
#include <gui/widgets/seq_graphic/layout_track_proxy.hpp>
#include <util/static_map.hpp>

BEGIN_NCBI_SCOPE


///////////////////////////////////////////////////////////////////////////////
/// CTempTrackProxy class implementation
///////////////////////////////////////////////////////////////////////////////

CTempTrackProxy::CTempTrackProxy(const CTempTrackProxy& param)
    : m_Order(param.GetOrder())
    , m_Name(param.GetName())
    , m_Id(param.GetId())
    , m_DispName(param.GetDisplayName())
    , m_Profile(param.GetTrackProfile())
    , m_Source(param.GetSource())
    , m_Category(param.GetCategory())
    , m_Subcategory(param.GetSubcategory())
    , m_SettingGroup(param.GetSettingGroup())
    , m_UseGroupDefault(param.GetUseGroupDefault())
    , m_Filter(param.m_Filter)
    , m_SortBy(param.m_SortBy)
    , m_Track(NULL)
    , m_Shown(param.GetShown())
    , m_Expanded(param.GetExpanded())
    , m_Realized(param.IsRealized())
    , m_Empty(false)
    , m_Comments(param.m_Comments)
    , m_Highlights(param.m_Highlights)
{
    ITERATE (TTrackProxies, iter, param.GetChildren()){
        m_Children.push_back((*iter)->Clone());
    }
}


CTempTrackProxy& CTempTrackProxy::operator=(const CTempTrackProxy& param)
{
    if (this == &param) return *this;

    m_Order = param.GetOrder();
    m_Name = param.GetName();
    m_Id = param.GetId();
    m_DispName = param.GetDisplayName();
    m_Profile = param.GetTrackProfile();
    m_Source = param.GetSource();
    m_Category = param.GetCategory();
    m_Subcategory = param.GetSubcategory();
    m_SettingGroup = param.GetSettingGroup();
    m_UseGroupDefault = param.GetUseGroupDefault();
    m_Filter = param.m_Filter;
    m_SortBy = param.m_SortBy;
    m_Track = NULL;
    m_Shown = param.GetShown();
    m_Expanded = param.GetExpanded();
    m_Realized = param.IsRealized();
    m_Empty = false;
    m_Comments = param.m_Comments;
    m_Highlights = param.m_Highlights;

    ITERATE (TTrackProxies, iter, param.GetChildren()){
        m_Children.push_back((*iter)->Clone());
    }

    return *this;
}


bool CTempTrackProxy::IsTempTrack() const
{
    return true;
}


CRef<CTempTrackProxy> CTempTrackProxy::Clone() const
{
    CRef<CTempTrackProxy> proxy(new CTempTrackProxy(*this));
    return proxy;
}


CRef<CTempTrackProxy> CTempTrackProxy::AddTempTrack(const string& name)
{
    CRef<CTempTrackProxy> track(new CTempTrackProxy(-1, name, true));
    int next_order = -1;
    ITERATE (TTrackProxies, iter, m_Children) {
        next_order = max(next_order, (*iter)->GetOrder());
    }
    track->SetOrder(++next_order);
    track->SetRealized(false);
    track->SetTrackProfile(GetTrackProfile());
    track->SetCategory(GetCategory());
    track->SetSubcategory(GetSubcategory());
    track->SetFilter(GetFilter());
    track->SetSortBy(GetSortBy());

    // special rule on NAA tracks:
    // For all initial NAA tracks, they are off by default.
    // They have to be turned on by user explicitly by any means.
    if (NStr::StartsWith(name, "NA0")) {
        track->SetShown(false);
    }

    m_Children.push_back(track);
    return track;
}

CRef<CTempTrackProxy> CTempTrackProxy::AddTempTrack(
        const string& name, const string& filter)
{
    CRef<CTempTrackProxy> track;
    if (filter.empty()) {
        track = AddTempTrack(name);
    } else {
        track = AddTempTrack(name + "-" + filter);
        track->SetFilter(filter);
    }
    return track;
}


CConstRef<CTempTrackProxy>
CTempTrackProxy::GetChildTempTrack(const string& name) const
{
    string name_modified(name);
    NStr::ReplaceInPlace(name_modified, ".", "_");
    ITERATE (CTrackProxy::TTrackProxies, iter, m_Children) {
        if ((*iter)->GetName() == name_modified) {
            return CConstRef<CTempTrackProxy>(iter->GetPointer());
        }
    }
    return CConstRef<CTempTrackProxy>();
}


CRef<CTempTrackProxy>
CTempTrackProxy::GetChildTempTrack(const string& name)
{
    string name_modified(name);
    NStr::ReplaceInPlace(name_modified, ".", "_");
    NON_CONST_ITERATE (CTrackProxy::TTrackProxies, iter, m_Children) {
        if ((*iter)->GetName() == name_modified) {
            // if the child doesn't have a profile, we
            // want to use its parent's profile.
            if ((*iter)->GetTrackProfile().empty()) {
                (*iter)->SetTrackProfile(GetTrackProfile());
            }
            if ((*iter)->GetFilter().empty()) {
                (*iter)->SetFilter(m_Filter);
            }
            if ((*iter)->GetSortBy().empty()) {
                (*iter)->SetSortBy(m_SortBy);
            }
            if ((*iter)->GetSource().empty()) {
                (*iter)->SetSource(name);
            }
            return *iter;
        }
    }
    CRef<CTempTrackProxy> track = AddTempTrack(name_modified);
    track->SetSource(name);
    return track;
}


CRef<CTempTrackProxy>
CTempTrackProxy::GetChildTempTrack(const string& name,
                                   const string& filter)
{
    CRef<CTempTrackProxy> track;
    if (filter.empty()) {
        track = GetChildTempTrack(name);
    } else {
        string name_modified(name + "-" + filter);
        NStr::ReplaceInPlace(name_modified, ".", "_");
        NON_CONST_ITERATE (CTrackProxy::TTrackProxies, iter, m_Children) {
            if ((*iter)->GetName() == name_modified) {
                // if the child doesn't have a profile, we
                // want to use its parent's profile.
                if ((*iter)->GetTrackProfile().empty()) {
                    (*iter)->SetTrackProfile(GetTrackProfile());
                }
                if ((*iter)->GetSource().empty()) {
                    (*iter)->SetSource(name);
                }
                track = *iter;
            }
        }
        if ( !track ) {
            track = AddTempTrack(name, filter);
            track->SetSource(name);
        }
    }
    return track;
}

///////////////////////////////////////////////////////////////////////////////
/// CTrackProxy class implementation
///////////////////////////////////////////////////////////////////////////////

CTrackProxy::CTrackProxy(const CTrackProxy& param)
    : CTempTrackProxy(param)
    , m_Key(param.m_Key)
    , m_Subkey(param.m_Subkey)
    , m_DB(param.m_DB)
    , m_Annots(param.m_Annots)
{}


CTrackProxy& CTrackProxy::operator=(const CTrackProxy& param)
{
    if (this == &param) return *this;

    CTempTrackProxy::operator=(param);
    m_Key = param.m_Key;
    m_Subkey = param.m_Subkey;
    m_DB = param.m_DB;
    m_Annots = param.m_Annots;

    return *this;
}


bool CTrackProxy::IsTempTrack() const
{
    return false;
}


CRef<CTempTrackProxy> CTrackProxy::Clone() const
{
    CRef<CTempTrackProxy> proxy(new CTrackProxy(*this));
    return proxy;
}


END_NCBI_SCOPE
