/*  $Id: mate_pair_glyph.cpp 27503 2013-02-26 18:16:55Z falkrb $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Vlad Lebedev, Liangshou Wu
 *
 * File Description:
 *   CMatePairGlyph -- utility class to layout mate pairs (a special type of
 *      pairwise alignments and hold a set of CAlignGlyph
 */

#include <ncbi_pch.hpp>
#include <gui/widgets/seq_graphic/mate_pair_glyph.hpp>
#include <gui/widgets/seq_graphic/rendering_ctx.hpp>
#include <gui/opengl/glrender.hpp>
#include <objects/seqloc/Seq_interval.hpp>
#include <objtools/alnmgr/alnvec.hpp>



BEGIN_NCBI_SCOPE
USING_SCOPE(objects);


CMatePairGlyph::CMatePairGlyph(const TAlignList& aligns)
    : m_SeqAligns(aligns)
    , m_LibraryId(eLibrary_NotSet)
    , m_ErrorType(eError_NotSet)
{
    if (m_SeqAligns.size() != 1) {
        m_ErrorType = eError_NoError;
    }

    const CSeq_id* id = NULL;
    TSeqRange range;
    ITERATE (TAlignList, iter, m_SeqAligns) {
        const CAlignGlyph& pw = **iter;
        TSeqRange curr_r = pw.GetLocation().GetTotalRange();
        range += curr_r;
        id = &pw.GetAlignMgr().GetSeqId(pw.GetAlignMgr().GetAnchor());

        const CSeq_align_Handle& align = pw.GetOrigAlignment();
        int type = 0;
        if (m_ErrorType != eError_NoError  &&
            align.GetSeq_align()->GetNamedScore("bad matepair code", type)) {
            m_ErrorType = static_cast<EErrorType>(type);
        }
        align.GetSeq_align()->GetNamedScore("matepair library", m_LibraryId);
        m_Intervals.push_back(curr_r);
    }

    if (m_LibraryId == eLibrary_NotSet) {
        m_LibraryId = eLibrary_NotFound;
    }

    if (m_ErrorType == eError_NotSet) {
        m_ErrorType = eError_NoError;
    }

    // if not set, set it based on pairs' orientation
    if (m_SeqAligns.size() == 2  &&  m_ErrorType == eError_NoError) {
        if (m_SeqAligns[0]->IsNegative() == m_SeqAligns[1]->IsNegative()) {
            m_ErrorType = eError_CoAlign;
        } else {
            m_ErrorType = eError_ContraAlign;
        }
    }

    if (id) {
        m_Location.Reset(new CSeq_loc());
        m_Location->SetInt().SetFrom(range.GetFrom());
        m_Location->SetInt().SetTo  (range.GetTo());
        m_Location->SetId(*id);
    }
}


bool CMatePairGlyph::NeedTooltip(const TModelPoint& /*p*/, string& /*tt*/, string& /*t_title*/) const
{
    return true;
}


void CMatePairGlyph::GetTooltip(const TModelPoint& /*p*/, string& tt, string& /*t_title*/) const
{
    x_GetTitle(&tt, CLabel::eDescription);
}


void CMatePairGlyph::GetHTMLActiveAreas(TAreaVector* p_areas) const
{
    CHTMLActiveArea area;
    CSeqGlyph::x_InitHTMLActiveArea(area);
    area.m_Signature = GetSignature();
    p_areas->push_back(area);
}


TSeqRange CMatePairGlyph::GetRange(void) const
{
    return m_Location->GetTotalRange();
}


bool CMatePairGlyph::IsClickable() const
{
    return true;
}


const objects::CSeq_loc& CMatePairGlyph::GetLocation(void) const
{
    return *m_Location;
}


CConstRef<CObject> CMatePairGlyph::GetObject(TSeqPos pos) const
{
    ITERATE (TAlignList, iter, m_SeqAligns) {
        CConstRef<CObject> obj = (*iter)->GetObject(pos);
        if (obj.NotEmpty()) {
            return obj;
        }
    }
    return CConstRef<CObject>();
}


void CMatePairGlyph::GetObjects(vector<CConstRef<CObject> >& objs) const
{
    ITERATE (TAlignList, iter, m_SeqAligns) {
        (*iter)->GetObjects(objs);
    }
}


bool CMatePairGlyph::HasObject(CConstRef<CObject> obj) const
{
    ITERATE(TAlignList, iter, m_SeqAligns) {
        if ( (*iter)->HasObject(obj) ) {
            m_ObjSel.push_back(obj);
            return true;
        }
    }
    return false;
}


string CMatePairGlyph::GetSignature() const
{
    // always return the signature of the first alignment
    ITERATE (TAlignList, iter, m_SeqAligns) {
        return (*iter)->GetSignature();
    }
    return string();
}


const CMatePairGlyph::TIntervals& CMatePairGlyph::GetIntervals(void) const
{
    return m_Intervals;
}


bool CMatePairGlyph::IsPairSelected(const CAlignGlyph* pw_aln) const
{
    ITERATE(vector<CConstRef<CObject> >, iter, m_ObjSel) {
        if (pw_aln->HasObject(*iter))
            return true;
    }
    return false;
}


void CMatePairGlyph::x_Draw() const
{
    CGlRender& gl = GetGl();

    // It's a vector of CRefs
    const TAlignList& pw_aligns = GetSeqAligns();
    TModelRect rcm = GetModelRect();
    TSeqRange range = GetRange();
    TModelRange inrc = m_Context->IntersectVisible(range);

    TModelUnit yy = rcm.Top();
    bool show_label = m_Config->m_ShowLabel  &&  m_Context->WillLabelFit(inrc);
    if (show_label) {
        yy = yy + m_Config->m_LabelFont.TextHeight() + 4;
    }

    TModelUnit bar_h = pw_aligns.front()->GetBarHeight();

    // Set Colors based on Error code for Mate Pairs
    CRgbaColor c_fg, c_seq, c_mis;
    switch (GetError()) {
    case eError_Orientation:
        c_fg = m_Config->m_FGOrientation;
        c_seq = m_Config->m_SeqOrientation;
        c_mis = m_Config->m_SeqMismatchOrientation;
        break;
    case eError_NonUnique:
        c_fg  = m_Config->m_FGNonUnique;
        c_seq = m_Config->m_SeqNonUnique;
        c_mis = m_Config->m_SeqMismatchNonUnique;
        break;
    case eError_Distance:
        c_fg  = m_Config->m_FGDistance;
        c_seq = m_Config->m_SeqDistance;
        c_mis = m_Config->m_SeqMismatchDistance;
        break;
    case eError_CoAlign:
        c_fg  = m_Config->m_FGCoAlign;
        c_seq = m_Config->m_SeqNo;
        c_mis = m_Config->m_SeqMismatchNo;
        break;
    case eError_ContraAlign:
        c_fg  = m_Config->m_FGContraAlign;
        c_seq = m_Config->m_SeqNo;
        c_mis = m_Config->m_SeqMismatchNo;
        break;
    default:
        c_fg  = m_Config->m_FGNo;
        c_seq = m_Config->m_SeqNo;
        c_mis = m_Config->m_SeqMismatchNo;
        break;
    }

    // Draw each Pairwise Alignment inside
    // If two mates overlap, move the second one below the first one
    TModelUnit y_off = 0.0;
    if (pw_aligns.size() == 2  &&
        pw_aligns[0]->GetRange().IntersectingWith(pw_aligns[1]->GetRange())) {
            y_off = bar_h;
    }
    TModelUnit yTop = yy;
    ITERATE(TAlignList, iter, pw_aligns) {
        TSeqRange l_range = (*iter)->GetRange();
        TModelRect rcm_pw(l_range.GetFrom(), yTop + bar_h, l_range.GetTo(), yTop);
        (*iter)->DrawPWAlignElem(rcm_pw, false, false, c_fg, c_seq, c_mis);
        yTop += y_off;
    }

    // draw link between mate pairs
    if (pw_aligns.size() == 2) {
        gl.ColorC(m_Config->m_FGLink);
        gl.LineWidth(1.0f);
        gl.Disable(GL_LINE_SMOOTH);

        TSeqRange r1 = pw_aligns[0]->GetRange();
        TSeqRange r2 = pw_aligns[1]->GetRange();
        TSeqRange inter_r = r1.IntersectionWith(r2);
        if (inter_r.Empty()) {
            TSeqPos from = r1.GetFrom() < r2.GetFrom() ? r1.GetTo() : r2.GetTo();
            TSeqPos to   = r1.GetFrom() < r2.GetFrom() ? r2.GetFrom() : r1.GetFrom();
            _ASSERT(from <= to);
            m_Context->DrawLine(from + 1, yy + bar_h * 0.5f - 1.0, to, yy + bar_h * 0.5f - 1.0);
        } else {
            m_Context->DrawLine(inter_r.GetFrom(), yy + bar_h, inter_r.GetToOpen(), yy + bar_h);
        }
    }

    /*
    // Highlight background based on Library ID
    if (m_ConfigSettings->GetShowAlignBg()  &&
        pair->GetLibraryId() > 0  &&
        m_ConfigSettings->GetShowPWAlignLabels()) {
        CRgbaColor color = m_ColorTable.GetColor(pair->GetLibraryId());
        gl.Color4f(color.GetRed(), color.GetGreen(), color.GetBlue(), 0.1f);
        x_DrawQuad(rcm, horz);
    }
    */

    // draw label
    if (show_label) {
        string title;
        x_GetTitle(&title, CLabel::eDefault);

        TModelUnit widthP  = m_Context->SeqToScreen(inrc.GetLength());
        title = m_Config->m_LabelFont.Truncate(title.c_str(), widthP);

        if (IsSelected()) {
            gl.ColorC(m_Context->GetSelLabelColor());
        } else {
            gl.ColorC(m_Config->m_Label);
        }

        TModelUnit xM = (inrc.GetFrom() + inrc.GetTo()) * 0.5;
        m_Context->TextOut(&m_Config->m_LabelFont, title.c_str(), xM, yy - 2, true);
    } // done with labels
 
    if (IsSelected()) {
        m_Context->DrawSelection(rcm);
    }
}


void CMatePairGlyph::x_UpdateBoundingBox()
{
    TAlignList::iterator iter = m_SeqAligns.begin();
    while (iter != m_SeqAligns.end()) {
        (*iter)->Update(true);
        ++iter;
    }

    iter = m_SeqAligns.begin();
    TModelRange range((*iter)->GetLeft(), (*iter)->GetRight());
    ++iter;
    if (iter != m_SeqAligns.end()) {
        range.CombineWith(TModelRange((*iter)->GetLeft(), (*iter)->GetRight()));
    }
    SetLeft(range.GetFrom());
    SetWidth(range.GetLength());

    m_Height = m_SeqAligns.front()->GetBarHeight();

    // in case two mate pair ends overlap, we put the second one
    // right under the first one.
    if (m_SeqAligns.size() == 2  &&
        m_SeqAligns[0]->GetRange().IntersectingWith(m_SeqAligns[1]->GetRange())) {
        m_Height += m_Height;
    }
    TModelRange inrc = m_Context->IntersectVisible(range);
    if (m_Config->m_ShowLabel  &&  m_Context->WillLabelFit(inrc)) {
        m_Height += m_Config->m_LabelFont.TextHeight() + 4;
    }    
}


void CMatePairGlyph::x_GetTitle(string* title, CLabel::ELabelType type) const
{
    // It's a vector of CRefs
    const TAlignList& pw_aligns = GetSeqAligns();

    *title = "Mate Pair: ";
    string tmp = "";

    ITERATE(TAlignList, iter, pw_aligns) {
        if (tmp.length() > 0) *title = *title + " / ";
        tmp.erase();
        (*iter)->GetTitle(&tmp, type);
        *title = *title + tmp;
    }
}

END_NCBI_SCOPE

