/*  $Id: segment_map_track.cpp 27877 2013-04-17 18:49:16Z wuliangs $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Liangshou Wu
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>

#include <gui/widgets/seq_graphic/segment_map_track.hpp>
#include <gui/widgets/seq_graphic/seqgraphic_job_result.hpp>
#include <gui/widgets/seq_graphic/seqgraphic_ds_manager.hpp>
#include <gui/widgets/seq_graphic/seqgraphic_utils.hpp>
#include <gui/widgets/seq_graphic/config_utils.hpp>
#include <gui/widgets/seq_graphic/layout_track_proxy.hpp>
#include <gui/widgets/seq_graphic/rendering_ctx.hpp>
#include <gui/widgets/seq_graphic/seqgraphic_utils.hpp>
#include <gui/widgets/seq_graphic/comment_glyph.hpp>
#include <gui/widgets/wx/wx_utils.hpp>

#include <gui/utils/rgba_color.hpp>
#include <gui/opengl/gltexturefont.hpp>
#include <gui/objutils/utils.hpp>
#include <gui/objutils/label.hpp>
#include <gui/objutils/registry.hpp>


#include <objmgr/util/feature.hpp>
#include <objmgr/util/sequence.hpp>
#include <objmgr/annot_ci.hpp>

#include <objects/seqfeat/SeqFeatData.hpp>

#include <wx/menu.h>


BEGIN_NCBI_SCOPE
USING_SCOPE(objects);
    
///////////////////////////////////////////////////////////////////////////////
///   CSegmentMapTrack
///////////////////////////////////////////////////////////////////////////////

static const size_t kMaxSegmentMapRows = 6;
static const string kDefTrackTitle = "Component map";
static const string kDefProfile = "Default";
static const string kBaseKey = "GBPlugins.SeqGraphicComponentMap";
static const int kSegmentCutoff = 500;
static const size_t kCompactThreshold = 80;

typedef SStaticPair<const char*, CSegmentMapTrack::ESegmentLevel> TLevelStr;
static const TLevelStr s_LevelStrs[] = {
    { "Adaptive",       CSegmentMapTrack::eAdaptive },
    { "Component Map",  CSegmentMapTrack::eComponent },
    { "Contig Map",     CSegmentMapTrack::eContig },
};


typedef CStaticArrayMap<string, CSegmentMapTrack::ESegmentLevel> TLevelMap;
DEFINE_STATIC_ARRAY_MAP(TLevelMap, sm_LevelMap, s_LevelStrs);


CSegmentMapTrack::ESegmentLevel
CSegmentMapTrack::LevelStrToValue(const string& level)
{
    TLevelMap::const_iterator iter = sm_LevelMap.find(level);
    if (iter != sm_LevelMap.end()) {
        return iter->second;
    }
    NCBI_THROW(CException, eInvalid, "Invalid level string: " + level);

}


const string& CSegmentMapTrack::LevelValueToStr(ESegmentLevel level)
{
    TLevelMap::const_iterator iter;
    for (iter = sm_LevelMap.begin();  iter != sm_LevelMap.end();  ++iter) {
        if (iter->second == level) {
            return iter->first;
        }
    }
    return kEmptyStr;
}


typedef SStaticPair<const char*, CSegmentConfig::ELabelPosition> TLabelPosStr;
static const TLabelPosStr s_LabelPosStrs[] = {
    { "above",    CSegmentConfig::ePos_Above },
    { "inside",   CSegmentConfig::ePos_Inside },
    { "no label", CSegmentConfig::ePos_NoLabel },
    { "side",     CSegmentConfig::ePos_Side },
};

typedef CStaticArrayMap<string, CSegmentConfig::ELabelPosition> TLabelPosMap;
DEFINE_STATIC_ARRAY_MAP(TLabelPosMap, sm_LabelPosMap, s_LabelPosStrs);


CSegmentConfig::ELabelPosition
CSegmentMapTrack::LabelPosStrToValue(const string& pos)
{
    TLabelPosMap::const_iterator iter = sm_LabelPosMap.find(pos);
    if (iter != sm_LabelPosMap.end()) {
        return iter->second;
    }
    NCBI_THROW(CException, eInvalid, "Invalid label position string: " + pos);
}


const string&
CSegmentMapTrack::LabelPosValueToStr(CSegmentConfig::ELabelPosition pos)
{
    TLabelPosMap::const_iterator iter;
    for (iter = sm_LabelPosMap.begin();  iter != sm_LabelPosMap.end();  ++iter) {
        if (iter->second == pos) {
            return iter->first;
        }
    }
    return kEmptyStr;
}


CTrackTypeInfo CSegmentMapTrack::m_TypeInfo("segment_map_track",
                                            "Graphical View Segment Map Track");

CSegmentMapTrack::CSegmentMapTrack(CSGSegmentMapDS* ds,
                                   CRenderingContext* r_cntx,
                                   int total_level)
    : CDataTrack(r_cntx)
    , m_DS(ds)
    , m_Depth(m_DS->GetDepth())
    , m_TotalLevel(total_level)
    , m_ResolveSequences(true)
    , m_UseCutoff(false)
{
    m_DS->SetJobListener(this);
    SetLayoutPolicy(m_Layered);
    if (m_Depth == -1  &&  total_level > 1) {
        x_RegisterIcon(SIconInfo(eIcon_Content, "Component level",
            true, "track_content"));
    }
}


CSegmentMapTrack::~CSegmentMapTrack()
{
    if ( !m_DS->AllJobsFinished() ) {
        m_DS->DeleteAllJobs();
    }
}


const CTrackTypeInfo& CSegmentMapTrack::GetTypeInfo() const
{
    return m_TypeInfo;
}


string CSegmentMapTrack::GetFullTitle() const
{
    if (GetTitle().empty()) {
        return kDefTrackTitle;
    }
    return GetTitle();
}


void  CSegmentMapTrack::x_LoadSettings(const string& preset_style,
                                       const TKeyValuePairs& settings)
{
    if ( !m_Config ) {
        m_Config.Reset(new CSegmentConfig);
    }

    if (preset_style.empty()) {
        SetProfile(kDefProfile);
    } else {
        SetProfile(preset_style);
    }

    CGuiRegistry& registry = CGuiRegistry::GetInstance();
    CGuiRegistry::TReadView view;
    view = CSGConfigUtils::GetReadView(
        registry, kBaseKey, GetProfile(), kDefProfile);
    m_ResolveSequences = view.GetBool("ResolveSequences", true);
    m_UseCutoff = view.GetBool("UseCutoff", false);
    string depth_str = view.GetString("Depth");

    if (m_DS->GetDepth() != -1) {
        // We are in in multiple level layout. We honor layout level
        // over the the track level setting.
        m_Depth = m_DS->GetDepth();
    } else {
        ITERATE (CSGConfigUtils::TKeyValuePairs, iter, settings) {
            try {
                if (NStr::EqualNocase(iter->first, "Depth")) {
                    depth_str = iter->second;
                    /// HACK: for backward compatibility, used to be -1, 0 and 1
                    /// now use the expressive strings
                    if (depth_str.length() == 1) { // old settings
                        depth_str = LevelValueToStr((ESegmentLevel)NStr::StringToInt(depth_str));
                    }
                } else if (NStr::EqualNocase(iter->first, "ResolveSequences")) {
                    m_ResolveSequences = NStr::StringToBool(iter->second);
                } else if (NStr::EqualNocase(iter->first, "UseCutoff")) {
                    m_UseCutoff = NStr::StringToBool(iter->second);
                }
            } catch (CException&) {
                LOG_POST(Warning << "CSegmentMap::LoadSettings() - invalid setting: " 
                    << iter->first << ":" << iter->second);
            }
        }

        int depth = -1;
        try {
            depth = LevelStrToValue(depth_str);
        } catch (CException& e) {
            LOG_POST(Warning << e.GetMsg());
        }
        
        x_SetDepth(depth);
    }

    // label position
    view = CSGConfigUtils::GetLabelPosReadView(
        registry, kBaseKey, GetProfile(), m_gConfig->GetLabelPos(), kDefProfile);
    try {
        m_Config->m_LabelPos = LabelPosStrToValue(view.GetString("LabelPos"));
    } catch (CException& e) {
        LOG_POST(Warning << e.GetMsg());
    }

    // sizes
    view = CSGConfigUtils::GetSizeReadView(
        registry, kBaseKey, GetProfile(), m_gConfig->GetSizeLevel(), kDefProfile);
    m_Config->m_BarHeight = view.GetInt("BarHeight", 10);
    CSGConfigUtils::GetFont(
        view, "LabelFontFace", "LabelFontSize", m_Config->m_LabelFont);

    // colors
    view = CSGConfigUtils::GetColorReadView(
        registry, kBaseKey, GetProfile(), m_gConfig->GetColorTheme(), kDefProfile);
    CSGConfigUtils::GetColor(view, "BG",        m_Config->m_BG);
    CSGConfigUtils::GetColor(view, "FGExc",     m_Config->m_FGExc);
    CSGConfigUtils::GetColor(view, "FGInc_F",   m_Config->m_FGInc_F);
    CSGConfigUtils::GetColor(view, "FGInc_D",   m_Config->m_FGInc_D);
    CSGConfigUtils::GetColor(view, "FGInc_W",   m_Config->m_FGInc_W);
    CSGConfigUtils::GetColor(view, "FGInc_U",   m_Config->m_FGInc_U);
    CSGConfigUtils::GetColor(view, "FGGap",     m_Config->m_FGGap);
    CSGConfigUtils::GetColor(view, "Label",     m_Config->m_Label);
    CSGConfigUtils::GetColor(view, "SelLabel",     m_Config->m_SelLabel);

    m_Layered->SetVertSpace(m_gConfig->GetObjectSpace());
}


void  CSegmentMapTrack::x_SaveSettings(const string& preset_style)
{
    TKeyValuePairs settings;

    if ( !preset_style.empty() ) {
        settings["profile"] = preset_style;
    }

    int depth = x_ToUniversalLevel(m_Depth);
    settings["Depth"] = LevelValueToStr((ESegmentLevel)depth);
    SetProfile(CSGConfigUtils::ComposeProfileString(settings));
}


void  CSegmentMapTrack::x_SaveConfiguration(const string& /*preset_style*/) const
{
    const string& profile = GetProfile();
    CGuiRegistry& registry = CGuiRegistry::GetInstance();
    CGuiRegistry::TReadWriteView view =
        CSGConfigUtils::GetReadWriteView(registry, kBaseKey, profile, kDefProfile);
    int depth = x_ToUniversalLevel(m_Depth);
    view.Set("Depth", LevelValueToStr((ESegmentLevel)depth));

    // it is very unlikely we will change these
    //if (m_Config) {
    if (false) {
        // label position
        view = CSGConfigUtils::GetLabelPosRWView(
            registry, kBaseKey, profile, m_gConfig->GetLabelPos(), kDefProfile);
        view.Set("LabelPos", LabelPosValueToStr(m_Config->m_LabelPos));

        // save size settings
        view = CSGConfigUtils::GetSizeRWView(
            registry, kBaseKey, profile, m_gConfig->GetSizeLevel(), kDefProfile);
        view.Set("BarHeight", m_Config->m_BarHeight);
        CSGConfigUtils::SetFont(
            view, "LabelFontFace", "LabelFontSize", m_Config->m_LabelFont);

        // save color settings
        view = CSGConfigUtils::GetColorRWView(
            registry, kBaseKey, profile, m_gConfig->GetColorTheme(), kDefProfile);
        CSGConfigUtils::SetColor(view, "BG",        m_Config->m_BG);
        CSGConfigUtils::SetColor(view, "FGExc",     m_Config->m_FGExc);
        CSGConfigUtils::SetColor(view, "FGInc_F",   m_Config->m_FGInc_F);
        CSGConfigUtils::SetColor(view, "FGInc_D",   m_Config->m_FGInc_D);
        CSGConfigUtils::SetColor(view, "FGInc_W",   m_Config->m_FGInc_W);
        CSGConfigUtils::SetColor(view, "FGInc_U",   m_Config->m_FGInc_U);
        CSGConfigUtils::SetColor(view, "FGGap",     m_Config->m_FGGap);
        CSGConfigUtils::SetColor(view, "Label",     m_Config->m_Label);
        CSGConfigUtils::SetColor(view, "SelLabel",  m_Config->m_SelLabel);
    }
}


void CSegmentMapTrack::x_OnIconClicked(TIconID id)
{
    if (id == eIcon_Content) {
        int depth = x_ToUniversalLevel(m_Depth);
        wxMenu menu;
        UseDefaultMarginWidth(menu);
        int menu_id = 10000;
        for (int level_idx = -1; level_idx < 2; ++level_idx) {
            wxString str = ToWxString(LevelValueToStr((ESegmentLevel)level_idx));
            wxMenuItem* item = menu.AppendRadioItem(menu_id++, str);
            if (depth == level_idx) {
                item->Check();
            }
        }

        m_LTHost->LTH_PopupMenu(&menu);
        wxMenuItemList& item_list = menu.GetMenuItems();
        int level = -1;
        ITERATE (wxMenuItemList, iter, item_list) {
            if ( (*iter)->IsChecked()  &&  depth != level) {
                x_SetDepth(level);
                x_UpdateData();
                break;
            }
            level += 1;
        }
    } else {
        // use default handlers
        CLayoutTrack::x_OnIconClicked(id);
    }
    x_OnLayoutChanged();
}


void CSegmentMapTrack::x_UpdateData()
{
    m_DS->DeleteAllJobs();
    x_SetStartStatus();
    int cutoff = m_UseCutoff ? kSegmentCutoff : -1;
    m_DS->LoadSegmentMap(m_Context->GetVisSeqRange(), cutoff, eSegmentMap);
}


void CSegmentMapTrack::x_OnJobCompleted(CAppJobNotification& notify)
{
    m_DS->ClearJobID(notify.GetJobID());
    CRef<CObject> res_obj = notify.GetResult();
    CJobResultBase* result = dynamic_cast<CJobResultBase*>(&*res_obj);
    if ( !result ) {
        LOG_POST(Error << "CSegmentMapTrack::x_OnJobCompleted() notification for job \
                does not contain results.");
        return;
    }

    EJobType type = (EJobType)result->m_Token;
    switch (type) {
    case eSegmentMap:
        x_AddSegmentMapLayout(dynamic_cast<CSGJobResult&>(*res_obj));
        break;
    case eSegmentMapSeqID:
        {{
            int num_partial = 0;
            int num_finished = 0;
            int num_wgs = 0;
            int num_unknown = 0;
            ITERATE (CSeqGlyph::TObjects, iter, GetChildren()) {
                CSeqGlyph* obj = const_cast<CSeqGlyph*>(iter->GetPointer());
                CSegmentGlyph* seq_map = dynamic_cast<CSegmentGlyph*>(obj);
                CSegmentConfig::ESequenceQuality qual = seq_map->GetSeqQuality();
                switch (qual) {
                    case CSegmentConfig::eDraft:
                        num_partial++;
                        break;
                    case CSegmentConfig::eFinished:
                        num_finished++;
                        break;
                    case CSegmentConfig::eWgs:
                        num_wgs++;
                        break;
                    default:
                        num_unknown++;
                        break;
                }
            }
            string msg = "";
            if (num_partial > 0 ) msg += ", partial: " + NStr::IntToString(num_partial);
            if (num_finished > 0 ) msg += ", finished: " + NStr::IntToString(num_finished);
            if (num_wgs > 0 ) msg += ", WGS: " + NStr::IntToString(num_wgs);
            if (num_unknown > 0 ) msg += ", other segments: " + NStr::IntToString(num_unknown);
            SetMsg(msg);
        }}
        break;
    default:
        break;
    }

    x_OnLayoutChanged();
}


void CSegmentMapTrack::x_AddSegmentMapLayout(const CSGJobResult& result)
{
    SetGroup().Clear();
    size_t count = result.m_ObjectList.size();
    if (count == 0) {
        SetMsg("");
        x_UpdateLayout();
        return;
    }

    SetObjects(result.m_ObjectList);
    CCommentGlyph* comment =
        dynamic_cast<CCommentGlyph*>(SetChildren().front().GetPointer());
    if (comment) {
        CRef<CCommentConfig> c_config(new CCommentConfig);
        c_config->m_ShowBoundary = false;
        c_config->m_ShowConnection = false;
        c_config->m_Centered = true;
        c_config->m_LabelColor.Set(1.0f, 0.0f, 0.0f);
        c_config->m_LineColor.Set(1.0f, 1.0f, 1.0f);
        c_config->m_Font.SetFontFace(CGlTextureFont::eFontFace_Helvetica);
        c_config->m_Font.SetFontSize(12);
        comment->SetConfig(c_config);
        SetMsg("");
    } else {
        bool drop_label = false;
        if (result.m_ObjectList.size() > kCompactThreshold  &&
            m_Config->m_LabelPos == CSegmentConfig::ePos_Side) {
            drop_label = true;
        }
        NON_CONST_ITERATE (CSeqGlyph::TObjects, iter, SetChildren()) {
            CSegmentGlyph* seg = dynamic_cast<CSegmentGlyph*>(iter->GetPointer());
            seg->SetConfig(m_Config);
            if (!GetHighlights().empty()  &&  x_NeedHighlight(seg)) {
                seg->SetHighlighted(true);
            } else {
                seg->SetHideLabel(drop_label);
            }
        }
        string msg = ", " + NStr::SizetToString(count) + " component";
        msg += (count != 1 ? "s shown" : " shown");
        SetMsg(msg);
    }

    // Compress Segment Map if there are more than
    // kMaxSegmentMapRows rows in the layout
    m_Layered->SetMaxRow(kMaxSegmentMapRows);

    // launch a new job to resolve segment map seq_ids
    if (!comment  &&  (count < 10  ||  m_ResolveSequences)) {
        x_SetStatus(", Resolving component seq-ids...", 0);
        m_DS->LoadSegmentMapSeqIDs(SetChildren(), eSegmentMapSeqID);
    }

    x_UpdateLayout();
}


bool CSegmentMapTrack::x_NeedHighlight(const CSegmentGlyph* seg) const
{
    if (seg) {
        string label;
        seg->GetLabel(label, CLabel::eContent);
        NStr::ToLower(label);
        if (GetHighlights().count(label) > 0) {
            return true;
        }
    }
    return false;
}


int CSegmentMapTrack::x_ToUniversalLevel(int depth) const
{
    if (depth != -1  &&  m_TotalLevel == 1) {
        if (depth == -2) {
            depth  = 0;
        } else {
            depth += 1;
        } 
    }
    return depth;
}


void CSegmentMapTrack::x_SetDepth(int level)
{
    if (level > 1) level = 1;

    if (level == -1  ||  m_TotalLevel > 1) {
        m_Depth = level;
        m_DS->SetDepth(level);
    } else {
        if (level == 0) {
            // scaffold level, there is no scaffold map for this sequence.
            // Add level icon to allow users switch to different level.
            x_RegisterIcon(SIconInfo(eIcon_Content, "Component level",
                true, "track_content"));
            m_Depth = -2;
        } else {
            // for non-chromosome sequences, there is only one level segment map
            // which is compoment map (level 0 for non-chromosome sequences)
            m_Depth = 0;
        }
        m_DS->SetDepth(m_Depth);
    }
}


///////////////////////////////////////////////////////////////////////////////
/// CSegmentMapTrackFactory methods
///////////////////////////////////////////////////////////////////////////////
ILayoutTrackFactory::TTrackMap
CSegmentMapTrackFactory::CreateTracks(SConstScopedObject& object,
                                      ISGDataSourceContext* ds_context,
                                      CRenderingContext* r_cntx,
                                      const SExtraParams& params,
                                      const TAnnotMetaDataList& /*src_annots*/) const
{
    TTrackMap tracks;
    int level = params.m_Level;
    if (params.m_Adaptive) {
        level = -1;
    }
    // For multi-level layout, we move component map one level down.
    // So we will show level 0 components on level 1, and so on.
    // Therefore, there will be no component map shown on level 0.
    else {
        level -= 1;
        if (level < 0) {
            return tracks;
        }
    }

    CIRef<ISGDataSource> ds = ds_context->GetDS(
        typeid(CSGSegmentMapDSType).name(), object);
    CSGSegmentMapDS* segment_ds = dynamic_cast<CSGSegmentMapDS*>(ds.GetPointer());

    segment_ds->SetDepth(level);
    TSeqRange range = params.m_Range;
    if (range.Empty()) {
        range = r_cntx->GetVisSeqRange();
    }
    
    if (range.IsWhole()) {
        CRef<CSGSequenceDS> seq_ds = r_cntx->GetSeqDS();
        range.Set(0, seq_ds->GetSequenceLength() - 1);
    }

    if (params.m_SkipGenuineCheck  ||  segment_ds->HasSegmentMap(level, range)) {
        int total_level = level + 1;
        if (total_level < 2) {
            if (CSGUtils::IsChromosome(segment_ds->GetBioseqHandle(), segment_ds->GetScope())) {
                total_level = 2;
            } else {
                total_level = segment_ds->GetSegmentMapLevels(range);
                if (total_level < 1) {
                    total_level = 1;
                }
            }
        }
        CRef<CSegmentMapTrack> track(new CSegmentMapTrack(segment_ds, r_cntx, total_level));
        tracks[kDefTrackTitle] = track.GetPointer();
    }
    return tracks;
}


string CSegmentMapTrackFactory::GetExtensionIdentifier() const
{
    return CSegmentMapTrack::m_TypeInfo.GetId();
}


string CSegmentMapTrackFactory::GetExtensionLabel() const
{
    return CSegmentMapTrack::m_TypeInfo.GetDescr();
}


CRef<CTrackConfigSet>
CSegmentMapTrackFactory::GetSettings(CSeqGraphicConfig* /*g_conf*/,
                                     const string& /*profile*/,
                                     const TKeyValuePairs& settings,
                                     const CTempTrackProxy* track_proxy) const
{
    CRef<CTrackConfigSet> config_set(new CTrackConfigSet);
    // const CSegmentMapTrack* track =
    //     dynamic_cast<const CSegmentMapTrack*>(track_proxy->GetTrack());
    // if (track  &&  track->Empty()) {
    //     return config_set;
    // }

    CRef<CTrackConfig> config(new CTrackConfig);
    config_set->Set().push_back(config);
    config->SetHelp() = "GenBank sequences used to construct chromosome";
    config->SetLegend_text("graphicalViewLegend.html#anchor_9");

    // bool use_cutoff = true;
    // bool resolve_ids = false;
    CSegmentMapTrack::ESegmentLevel level = CSegmentMapTrack::eAdaptive;
    ITERATE (CSGConfigUtils::TKeyValuePairs, iter, settings) {
        try {
            if (NStr::EqualNocase(iter->first, "Depth")) {
                level = CSegmentMapTrack::LevelStrToValue(iter->second);
            // } else if (NStr::EqualNocase(iter->first, "ResolveSequences")) {
            //     resolve_ids = NStr::StringToBool(iter->second);
            // } else if (NStr::EqualNocase(iter->first, "UseCutoff")) {
            //     use_cutoff = NStr::StringToBool(iter->second);
            }
        } catch (CException&) {
            LOG_POST(Warning << "CSegmentMap::GetSettings() - invalid setting: " 
                << iter->first << ":" << iter->second);
        }
    }

    /*
    if (level == CSegmentMapTrack::eComponent) {
//         config->AddCheckBox("Auto hide", "UseCutoff", "If checked, component "
// "map will be shown only when there are no more than 500 components in the given "
// "range.", use_cutoff);
//         config->AddCheckBox("Color segments", "ResolveSequences", "Color the "
// "segments based on the color code. By default, the color coding is disabled (for "
// "performance consideration).", resolve_ids);

        config->SetHidden_settings().push_back(
            CTrackConfigUtils::CreateHiddenSetting(
                "Depth",
                CSegmentMapTrack::LevelValueToStr(CSegmentMapTrack::eComponent)));
            
        config->SetHidden_settings().push_back(
            CTrackConfigUtils::CreateHiddenSetting("ResolveSequences", "false"));
        config->SetHelp() = "The GenBank sequences are shown in the order that \
they contribute to the assembly. Colors are as follows:<br> Blue:finished sequence<br>\
Orange:Draft sequence<br>Green:WGS<br>Gray:Other<br>Black:Gap<br>\
Beige:Sequence overlap not used to generate chromosome sequence.";

    } else if (level == CSegmentMapTrack::eContig) {
        config->SetHidden_settings().push_back(
            CTrackConfigUtils::CreateHiddenSetting(
                "Depth",
                CSegmentMapTrack::LevelValueToStr(CSegmentMapTrack::eContig)));
        config->SetHelp() = "Coordinated, mapped set of overlapping DNA segments \
derived from the genetic source and representing a complete molecule segment.";

    } else { // adaptive
        CRef<CChoice> choice = CTrackConfigUtils::CreateChoice(
            "Depth", "Segment Level",
            CSegmentMapTrack::LevelValueToStr((CSegmentMapTrack::ESegmentLevel)-1),
            "Choose which segment level to show");

        choice->SetValues().push_back(
            CTrackConfigUtils::CreateChoiceItem(
                CSegmentMapTrack::LevelValueToStr(CSegmentMapTrack::eAdaptive),
                CSegmentMapTrack::LevelValueToStr(CSegmentMapTrack::eAdaptive),
                "Adaptively show different level of segment based on zoom level",
                ""));

        choice->SetValues().push_back(
            CTrackConfigUtils::CreateChoiceItem(
                CSegmentMapTrack::LevelValueToStr(CSegmentMapTrack::eContig),
                CSegmentMapTrack::LevelValueToStr(CSegmentMapTrack::eContig),
                "Show contig map",
                ""));
        choice->SetValues().push_back(
            CTrackConfigUtils::CreateChoiceItem(
                CSegmentMapTrack::LevelValueToStr(CSegmentMapTrack::eComponent),
                CSegmentMapTrack::LevelValueToStr(CSegmentMapTrack::eComponent),
                "Show component map",
                "")); 
        
        config->SetChoice_list().push_back(choice);
        config->SetHelp() = "Segment map (either contig or components map)";
    }
    */

    CRef<CChoice> choice = CTrackConfigUtils::CreateChoice(
        "Depth", "Segment Level",
        CSegmentMapTrack::LevelValueToStr(level),
        "Choose which segment level to show");

    choice->SetValues().push_back(
        CTrackConfigUtils::CreateChoiceItem(
        CSegmentMapTrack::LevelValueToStr(CSegmentMapTrack::eContig),
        CSegmentMapTrack::LevelValueToStr(CSegmentMapTrack::eContig),
        "Show contig map",
        ""));
    choice->SetValues().push_back(
        CTrackConfigUtils::CreateChoiceItem(
        CSegmentMapTrack::LevelValueToStr(CSegmentMapTrack::eComponent),
        CSegmentMapTrack::LevelValueToStr(CSegmentMapTrack::eComponent),
        "Show component map",
        "")); 

    config->SetChoice_list().push_back(choice);
    config->SetHelp() = "Segment map (either contig or components map)";

    return config_set;
}


END_NCBI_SCOPE
