/*  $Id $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Robert G. Smith
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>
#include <gui/widgets/seq_graphic/seqgraphic_conf_panel.hpp>

#include <gui/widgets/seq_graphic/seqgraphic_preview.hpp>
#include <gui/widgets/seq_graphic/seqgraphic_conf.hpp>
#include <gui/widgets/seq_graphic/seqgraphic_ds.hpp>

#include <gui/utils/string_utils.hpp>

#include <objmgr/object_manager.hpp>
#include <objects/seqset/Seq_entry.hpp>
#include <objmgr/scope.hpp>
#include <serial/serial.hpp>
#include <serial/objistr.hpp>


BEGIN_NCBI_SCOPE

CSeqGraphicConfigPanel::CSeqGraphicConfigPanel()
{
    // Set our direction.
    // can't change after making the widgets.
    SetLayoutDirection(CThemeConfigPanel2::eLayoutDirection_Vertical);
}


CSeqGraphicConfigPanel::~CSeqGraphicConfigPanel()
{
}


void CSeqGraphicConfigPanel::MakeWidgets()
{
    CThemeConfigPanel2::MakeWidgets();


    _ASSERT(m_PreviewConfig.IsNull()); // Should only be called once in the life time of this class.

    // Do all the plumbing to support the preview:

    // create our own CSeqGraphicConfig.
    m_PreviewConfig.Reset(new CSeqGraphicConfig());

    // create a CThemePreviewMediator which contains this panel and our private config.
      // when any widget here changes we tell this to save the changes
      // and notify the preview widget.
    m_PreviewMediator.Reset(new CThemePreviewMediator(*m_PreviewConfig, *this));

    // Size and put the preview widget in the parent's preview group.
    // It got created by the parent above.
    Fl_Group* preview_group = GetPreviewLocation();

    // create a preview widget.
    const int inset = 2;
    m_PreviewWidget = new CSeqGraphicPreview(
        preview_group->x() + inset,
        preview_group->y() + inset,
        preview_group->w() - 2 * inset,
        preview_group->h() - 2 * inset);

    // Set its config to our private config.
    m_PreviewWidget->SetConfig(m_PreviewConfig);

    // give it the special Data source.

    m_PreviewWidget->SetDataSource(& x_GetFakeDataSource());

    preview_group->add(m_PreviewWidget);
    preview_group->resizable(m_PreviewWidget);
}



void CSeqGraphicConfigPanel::UpdatePreview()
{
    // Save the dialog's values so we can display them.
    m_PreviewMediator->Save(kEmptyStr);

    // Tell the preview that its configuration has changed.
    m_PreviewWidget->UpdateConfig();

}

// file defines s_PreviewData
#include "preview_data.hpp"
const string kPreviewId("lcl|fake_nuc");

CSeqGraphicDataSource& CSeqGraphicConfigPanel::x_GetFakeDataSource()
{
    USING_SCOPE(objects);

    if (! m_PreviewDS) {
        // read in data from a static memory into a seq entry.
        string preview_data_str = CStringUtil::BuildString(s_PreviewData);
        CRef<CSeq_entry> preview_TSE(new CSeq_entry);

        CNcbiIstrstream is(preview_data_str.c_str(), strlen(preview_data_str.c_str()));
        auto_ptr<CObjectIStream>
            pd_in(CObjectIStream::Open(eSerial_AsnText, is));
        *pd_in >> *preview_TSE;
        _TRACE("CSettings::x_GetFakeDataSource(): "
               "text preview data source read from string: \""
               << preview_data_str.substr(0,20) << " ... \"" );

        CRef<CScope> preview_scope(new CScope(*CObjectManager::GetInstance()) );
        preview_scope->AddTopLevelSeqEntry(*preview_TSE);

        CSeq_id preview_id(kPreviewId);

        // create data source from the above.
        m_PreviewDS.Reset(new CSeqGraphicDataSource(*preview_scope, preview_id));
    }

    return *m_PreviewDS;
}


END_NCBI_SCOPE
