/*  $Id: seqgraphic_graph_ds.cpp 27972 2013-05-03 20:34:20Z wuliangs $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Liangshou Wu
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>

#include <gui/widgets/seq_graphic/seqgraphic_graph_ds.hpp>
#include <gui/widgets/seq_graphic/seqgraphic_job_result.hpp>

#include <gui/widgets/seq_graphic/seqgraphic_utils.hpp>
#include <gui/widgets/seq_graphic/histogram_glyph.hpp>
#include <gui/objutils/utils.hpp>

// for graph job
#include <objmgr/seq_table_ci.hpp>
#include <objmgr/graph_ci.hpp>
#include <objmgr/table_field.hpp>
#include <objects/general/User_object.hpp>
#include <objects/general/User_field.hpp>
#include <objects/seq/Seq_annot.hpp>
#include <objects/seq/Annot_descr.hpp>
#include <objects/seq/Annotdesc.hpp>


BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

///////////////////////////////////////////////////////////////////////////////
/// CSGGraphDS
CSGGraphDS::CSGGraphDS(CScope& scope, const CSeq_id& id)
    : CSGGenBankDS(scope, id)
    , m_SeqTable(false)
{}


void CSGGraphDS::LoadSeqGraphs(objects::SAnnotSelector& sel,
                               const TSeqRange& range, float window,
                               bool fixed_scale)
{
    CSeqUtils::SetResolveDepth(sel, m_Adaptive, m_Depth);
    CRef<CSGSeqGraphJob> job(new CSGSeqGraphJob("Graph", m_Handle,
        sel, range, window, fixed_scale, m_SeqTable));
    job->SetGraphLevel(GetNearestLevel(window));
    x_LaunchJob(*job);
}


void CSGGraphDS::GetAnnotNames(objects::SAnnotSelector& sel,
                               const TSeqRange& range,
                               TAnnotNameTitleMap& names, bool seq_table) const
{
    CSeqUtils::SetResolveDepth(sel, m_Adaptive, m_Depth);
    CSGSeqGraphJob::GetAnnotNames(m_Handle, range, sel, names, seq_table);
}

void CSGGraphDS::GetSeqtableAnnots(objects::SAnnotSelector& sel,
                                   const TSeqRange& range,
                                   TAnnotNameTitleMap& names) const
{
    CSeqUtils::SetResolveDepth(sel, m_Adaptive, m_Depth);
    CSGSeqGraphJob::GetSeqtableAnnots(m_Handle, range, sel, names);
}


///////////////////////////////////////////////////////////////////////////////
/// CSGGraphDSType

ISGDataSource* CSGGraphDSType::CreateDS(SConstScopedObject& object) const
{
    const CSeq_id& id = dynamic_cast<const CSeq_id&>(object.object.GetObject());
    return new CSGGraphDS(object.scope.GetObject(), id);
}


string CSGGraphDSType::GetExtensionIdentifier() const
{
    static string sid("seqgraphic_graph_ds_type");
    return sid;
}


string CSGGraphDSType::GetExtensionLabel() const
{
    static string slabel("Graphical View Graph Data Source Type");
    return slabel;
}


bool CSGGraphDSType::IsSharable() const
{
    return false;
}


///////////////////////////////////////////////////////////////////////////////
/// CSGSeqGraphJob

void CSGSeqGraphJob::GetAnnotNames(const objects::CBioseq_Handle& handle,
                                     const TSeqRange& range,
                                     objects::SAnnotSelector& sel,
                                     TAnnotNameTitleMap& names, bool seq_table)
{
    if (seq_table) { /// search seq-table
        CAnnotTypes_CI annot_it(CSeq_annot::C_Data::e_Seq_table, handle,
            TSeqRange::GetWhole(), eNa_strand_unknown, &sel);
        ITERATE (CAnnotTypes_CI::TAnnotNames, iter, annot_it.GetAnnotNames()) {
            if (iter->IsNamed()) {
                if (iter->GetName().find("@@") == string::npos) {
                    names.insert(TAnnotNameTitleMap::value_type(iter->GetName(), ""));
                }
            } else {
                names.insert(TAnnotNameTitleMap::value_type(
                    CSeqUtils::GetUnnamedAnnot(), ""));
            }
        }
    } else {
        CGraph_CI graph_iter(handle, range, sel);
        ITERATE (CGraph_CI::TAnnotNames, iter, graph_iter.GetAnnotNames()) {
            if (iter->IsNamed()) {
                // IMPORTANT: exclude any alignment pileup graph
                // That is used for BAM/cSRA pileup display in alignment track
                if (iter->GetName().find("@@") == string::npos  &&
                    iter->GetName().find("pileup graphs") == string::npos) {
                    names.insert(TAnnotNameTitleMap::value_type(iter->GetName(), ""));
                }
            } else {
                names.insert(TAnnotNameTitleMap::value_type(
                    CSeqUtils::GetUnnamedAnnot(), ""));
            }
        }
    }
}


void CSGSeqGraphJob::GetSeqtableAnnots(const objects::CBioseq_Handle& handle,
                                       const TSeqRange& range,
                                       objects::SAnnotSelector& sel,
                                       TAnnotNameTitleMap& names)
{
    CConstRef<CSeq_loc> loc =
        handle.GetRangeSeq_loc(range.GetFrom(), range.GetTo());
    CAnnot_CI it(handle.GetScope(), *loc, sel);

    for (; it; ++it) {
        CSeq_annot_Handle annot = *it;

        if (annot.Seq_annot_CanGetDesc()) {
            const CSeq_annot::TDesc::Tdata& descs = annot.Seq_annot_GetDesc().Get();
            CSeq_annot::TDesc::Tdata::const_iterator d_iter = descs.begin();
            for (; d_iter != descs.end(); ++d_iter) {
                if ((*d_iter)->IsUser()) {
                    CConstRef<CUser_field> type_f = (*d_iter)->GetUser().GetFieldRef("track type");
                    if (type_f  &&  type_f->GetData().IsStr()  &&
                        type_f->GetData().GetStr() == "graph") {
                        if (annot.IsNamed()) {
                            if (it->GetName().find("@@") == string::npos) {
                                names.insert(TAnnotNameTitleMap::value_type(it->GetName(), ""));
                            }
                        } else {
                            names.insert(TAnnotNameTitleMap::value_type(
                                CSeqUtils::GetUnnamedAnnot(), ""));
                        }
                    }
                }
            }
        }
    }
}


IAppJob::EJobState CSGSeqGraphJob::x_Execute()
{
    if (GetGraphLevel() > 0) {
        return x_LoadGraphFromCoverageGraph();
    }

    if (m_SeqTable) {
        return x_LoadSeqTable();
    } 

    return x_LoadGraph();
}


enum EValueColType {
    eVal_Byte,
    eVal_Int,
    eVal_Real,
    eVal_Invalid
};


class CPosMapper : public CObject
{
public:
    CPosMapper(TSeqPos src_f, TSeqPos src_t, TSeqPos target_f, bool reversed)
        : m_SrcFrom(src_f)
        , m_SrcTo(src_t)
        , m_TargetFrom(target_f)
        , m_Reversed(reversed)
    {}

    void Map(int& from, int& to)
    {
        if (m_Reversed) {
            from = m_SrcTo + m_TargetFrom - from;
            to = m_SrcTo + m_TargetFrom - to;
            swap(from, to);
        } else {
            from += m_TargetFrom - m_SrcFrom;
            to += m_TargetFrom - m_SrcFrom;
        }
    }

private:
    TSeqPos m_SrcFrom;
    TSeqPos m_SrcTo;
    TSeqPos m_TargetFrom;
    bool    m_Reversed;
};


IAppJob::EJobState CSGSeqGraphJob::x_LoadSeqTable()
{
    CSGJobResult* result = new CSGJobResult();
    m_Result.Reset(result);

    CSeq_table_CI it(m_Handle,  m_Range, m_Sel);

    if ( !it ) return eCompleted;

    string name = CSeqUtils::GetAnnotName(it.GetAnnot());

    CTableFieldHandle<int> col_pos(CSeqTable_column_info::eField_id_location_from);
    CTableFieldHandle<int> col_span("span");
    CTableFieldHandle<double> col_val_real("values");
    CTableFieldHandle<int> col_val_int("values");

    typedef map<string, SGraphObject> TDMaps;
    TDMaps d_maps;

    for (; it; ++it) {
        CSeq_annot_Handle annot = it.GetAnnot();
        CRef<CPosMapper> mapper;
        if (it.IsMapped()) {
            const CSeq_loc& loc1 = it.GetOriginalLocation();
            const CSeq_loc& loc2 = it.GetMappedLocation();
            TSeqRange range = loc1.GetTotalRange();
            mapper.Reset(new CPosMapper(range.GetFrom(), range.GetTo(),
                loc2.GetTotalRange().GetFrom(), loc1.GetStrand() != loc2.GetStrand()));
        }

        // check if the given seq-table indeed contains graph data
        bool valid_seqtable = false;
        if (annot.Seq_annot_CanGetDesc()) {
            const CSeq_annot::TDesc::Tdata& descs =
                annot.Seq_annot_GetDesc().Get();
            CSeq_annot::TDesc::Tdata::const_iterator d_iter = descs.begin();
            for (; d_iter != descs.end(); ++d_iter) {
                if ((*d_iter)->IsUser()) {
                    CConstRef<CUser_field> type_f =
                        (*d_iter)->GetUser().GetFieldRef("track type");
                    if (type_f  &&  type_f->GetData().IsStr()  &&
                        type_f->GetData().GetStr() == "graph") {
                        valid_seqtable = true;
                        break;
                    }
                }
            }
        }
        if ( !valid_seqtable ) continue;

        // Separate and show graph in a separate track based on seq-annot::desc::title
        // CSeqUtils::GetAnnotName(annot) returns title if title is set.
        string g_title = CSeqUtils::GetAnnotName(annot);
        string g_name = CSeqUtils::GetUnnamedAnnot();
        if (annot.IsNamed()) {
            g_name = annot.GetName();
        }
        TDMaps::iterator iter = d_maps.find(g_title);
        if (iter == d_maps.end()) {
            string g_desc = CSeqUtils::GetAnnotComment(annot);
            iter = d_maps.insert(
                TDMaps::value_type(
                    g_title,
                    SGraphObject(
                        m_Range.GetFrom(),
                        m_Range.GetTo(),
                        (float)m_Window,
                        new max_func<CHistogramGlyph::TDataType>,
                        g_desc, g_name))).first;
        }

        SGraphObject& g_obj = iter->second;
        CHistogramGlyph::TMap* max_data = &g_obj.m_Map;

        size_t rows = annot.GetSeq_tableNumRows();

        // There is no way to tell the value column is
        // stored as BYTE, INTEGER or REAL. Only way to
        // find out is try and see.
        EValueColType val_type = eVal_Invalid;
        vector<char> byte_vals;
        {
            double val_r = 0.0;
            int val_i = 0;
            try {
                if (col_val_real.TryGet(annot, 0, val_r)) {
                    val_type = eVal_Real;
                }
            } catch (CException&) {
            }

            if (val_type == eVal_Invalid) {
                try {
                    if (col_val_int.TryGet(annot, 0, val_i)) {
                        val_type = eVal_Int;
                    }
                } catch (CException&) {
                }
            }
            if (val_type == eVal_Invalid) {
                try {
                    if (col_val_int.TryGet(annot, 0, byte_vals)) {
                        val_type = eVal_Byte;
                    }
                } catch (CException&) {
                }
            }
        }

        if (val_type == eVal_Invalid) {
            LOG_POST(Error << "Can't load the value column");
            continue;
        }

        TSeqPos from = m_Range.GetFrom();
        TSeqPos to = m_Range.GetTo();

        int span;
        int curr_from, curr_to;
        col_pos.TryGet(annot, rows - 1, curr_from);
        col_span.TryGet(annot, rows - 1, span);
        curr_to = curr_from + span - 1;
        col_pos.TryGet(annot, 0, curr_from);
        if (mapper) {
            mapper->Map(curr_from, curr_to);
        }

        if ((TSeqPos)curr_to < from  ||  (TSeqPos)curr_from > to) {
            continue;
        }


        // TODO: performance optimization
        // Try to see if it is possible to do early termination.
        // find the start row
        //int row = 0;
        //int r_start = 0;
        //int r_end = rows - 1;
        //do {
        //    row = (r_start + r_end) / 2;
        //    col_pos.TryGet(annot, row, pos);
        //    if ((TSeqPos)pos < from )   r_start = row;
        //    else                        r_end = row;
        //} while ((TSeqPos)pos != from  &&  r_start < r_end - 1);

        for (int row = 0;  (size_t)row < rows;  ++row) {
            if ( col_pos.TryGet(annot, row, curr_from) ) {
                span = 1;
                col_span.TryGet(annot, row, span);
                curr_to = curr_from + span - 1;
                if (mapper) {
                    mapper->Map(curr_from, curr_to);
                }

                double value = 0.0;
                if (val_type == eVal_Real) {
                    col_val_real.TryGet(annot, row, value);
                } else if (val_type == eVal_Int) {
                    int val_i = 0;
                    col_val_int.TryGet(annot, row, val_i);
                    value = (double)val_i;
                } else {
                    _ASSERT(byte_vals.size() > (size_t)row);
                    value = (double)byte_vals[row];
                }

                max_data->AddRange(TSeqRange(curr_from, curr_to), (float)value);
            }
        }
    }

    CSeqGlyph::TObjects tmp;
    NON_CONST_ITERATE (TDMaps, iter, d_maps) {
        SGraphObject& g_obj = iter->second;
        CHistogramGlyph *hist = new CHistogramGlyph(g_obj.m_Map, iter->first);
        CRef<CSeqGlyph> fref(hist);
        hist->SetDesc(g_obj.m_Desc);
        hist->SetAnnotName(g_obj.m_AnnotName);
        if (m_FixedScale) {
            hist->SetAxisMax(g_obj.m_Max);
            hist->SetAxisMin(g_obj.m_Min);
        }
        tmp.push_back(fref);
    }

    result->m_ObjectList.swap(tmp);
    result->m_Token = m_Token;

    return eCompleted;
}


IAppJob::EJobState CSGSeqGraphJob::x_LoadGraph()
{
    CSGJobResult* result = new CSGJobResult();
    m_Result.Reset(result);
    SetTaskName("Loading graphs...");
    CSeqGlyph::TObjects glyphs;
    EJobState state =
        x_CreateHistFromGraph(glyphs, m_Sel, m_Window, m_FixedScale);
    if (state == eCompleted  &&  !glyphs.empty()) {
        result->m_ObjectList.swap(glyphs);
    }
    result->m_Token = m_Token;
    return state;
}


IAppJob::EJobState CSGSeqGraphJob::x_LoadGraphFromCoverageGraph()
{
    CSGJobResult* result = new CSGJobResult();
    m_Result.Reset(result);
    SetTaskName("Loading graphs...");

    CSeqGlyph::TObjects glyphs;
    EJobState state =
        x_LoadCoverageGraph(glyphs, GetGraphLevel(), m_Window, "", m_FixedScale);
    if (state == eCompleted  &&  !glyphs.empty()) {
        result->m_ObjectList.swap(glyphs);
    }
    return state;
}


END_NCBI_SCOPE
