/*  $Id: seqgraphic_job.cpp 27972 2013-05-03 20:34:20Z wuliangs $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Informationion
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Vlad Lebedev, Liangshou Wu, Victor Joukov
 *
 * File Description:
 *
 */


#include <ncbi_pch.hpp>
#include <gui/widgets/seq_graphic/seqgraphic_job.hpp>
#include <gui/widgets/seq_graphic/seqgraphic_job_result.hpp>
#include <objmgr/graph_ci.hpp>
#include <objects/seqres/Byte_graph.hpp>
#include <objects/seqres/Int_graph.hpp>
#include <objects/seqres/Real_graph.hpp>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

///////////////////////////////////////////////////////////////////////////////
/// CSeqGraphicJob
///////////////////////////////////////////////////////////////////////////////

CSeqGraphicJob::CSeqGraphicJob(const string& desc)
    : m_Desc(desc)
    , m_Token(kInvalidJobToken)
    , m_TaskTotal(0)
    , m_TaskCompleted(0)
{
    m_Error.Reset(new CAppJobError("Job failed"));
    SetTaskName("Loading...");
}



IAppJob::EJobState CSeqGraphicJob::Run()
{
    m_Result.Reset();
    return x_Execute();
}


CConstIRef<IAppJobProgress> CSeqGraphicJob::GetProgress()
{
    CFastMutexGuard lock(m_Mutex);
    float progress = 0.0;
    if (m_TaskTotal != 0) {
        progress = (float)m_TaskCompleted/m_TaskTotal;
    }
    return CConstIRef<IAppJobProgress>
        (new CAppJobProgress(progress, m_TaskName));
}


CRef<CObject> CSeqGraphicJob::GetResult()
{
    return m_Result;
}


CConstIRef<IAppJobError> CSeqGraphicJob::GetError()
{
    return CConstIRef<IAppJobError>(m_Error.GetPointer());
}


string CSeqGraphicJob::GetDescr() const
{
    return m_Desc;
}


void CSeqGraphicJob::SetTaskName(const string& name)
{
    CFastMutexGuard lock(m_Mutex);
    m_TaskName = name;
}


void CSeqGraphicJob::SetTaskCompleted(int completed)
{
    CFastMutexGuard lock(m_Mutex);
    m_TaskCompleted = completed;
}


void CSeqGraphicJob::AddTaskCompleted(int delta)
{
    CFastMutexGuard lock(m_Mutex);
    m_TaskCompleted += delta;
}


void CSeqGraphicJob::SetTaskTotal(int total)
{
    CFastMutexGuard lock(m_Mutex);
    m_TaskTotal = total;
}


bool CSeqGraphicJob::StopRequested() const
{
    return IsCanceled();
}



///////////////////////////////////////////////////////////////////////////////
/// CSGAnnotJob
///////////////////////////////////////////////////////////////////////////////

CSGAnnotJob::CSGAnnotJob(const string& desc, CBioseq_Handle handle,
                         const SAnnotSelector& sel, const TSeqRange& range)
    : CSeqGraphicJob(desc)
    , m_Handle(handle)
    , m_Sel(sel)
    , m_Range(range)
    , m_GraphLevel(0)
{}


IAppJob::EJobState
CSGAnnotJob::x_LoadCoverageGraph(CSeqGlyph::TObjects& glyphs,
                                 int level, TModelUnit window,
                                 const string& title, bool fixed_scale)
{
    EJobState state = eFailed;
    SetTaskName("Loading coverage graph...");
    SAnnotSelector sel = m_Sel;
    sel.SetAnnotType(CSeq_annot::C_Data::e_Graph);
    sel.ResetAnnotsNames();

    ITERATE (SAnnotSelector::TNamedAnnotAccessions,
        iter, sel.GetNamedAnnotAccessions()) {
        sel.IncludeNamedAnnotAccession(iter->first, level);
        sel.AddNamedAnnots(CombineWithZoomLevel(iter->first, level));
    }

    state = x_CreateHistFromGraph(glyphs, sel, window, fixed_scale);
    return state;
}


IAppJob::EJobState
CSGAnnotJob::x_CreateHistFromGraph(CSeqGlyph::TObjects& glyphs,
                                   const SAnnotSelector& sel,
                                   TModelUnit window,
                                   bool fixed_scale)
{
    CConstRef<CSeq_loc> loc(
        m_Handle.GetRangeSeq_loc(m_Range.GetFrom(), m_Range.GetTo()) );
    CGraph_CI graph_iter(m_Handle.GetScope(), *loc, sel);

    // We want to merge graphs with the same graph name.
    // Graphs will have the same name if they have the same annotation
    // name and graph title .
    typedef map<string, SGraphObject> TDMaps;
    TDMaps d_maps;
    for (;  graph_iter;  ++graph_iter) {
        if (IsCanceled()) {
            return eCanceled;
        }
        const CMappedGraph& gr = *graph_iter;
        string g_title = x_GetGraphName(gr);
        string g_name = CSeqUtils::GetUnnamedAnnot();
        if (gr.GetAnnot().IsNamed()) {
            g_name = gr.GetAnnot().GetName();
        }
        TDMaps::iterator iter = d_maps.find(g_title);
        if (iter == d_maps.end()) {
            string g_desc = CSeqUtils::GetAnnotComment(gr.GetAnnot());
            iter = d_maps.insert(
                TDMaps::value_type(
                    g_title,
                    SGraphObject(
                        m_Range.GetFrom(),
                        m_Range.GetTo(),
                        (float)window,
                        new max_func<CHistogramGlyph::TDataType>,
                        g_desc, g_name))).first;
        }
        SGraphObject& g_obj = iter->second;
        switch (gr.GetGraph().Which()) {
        case CSeq_graph::TGraph::e_Byte:
        {{
            const CByte_graph& graph = gr.GetGraph().GetByte();
            x_AddGraphToMap(g_obj.m_Map, gr, graph.GetValues(),
                graph.GetMax(), graph.GetMin(), g_obj.m_Max,
                g_obj.m_Min, fixed_scale);
            break;
        }}
        case CSeq_graph::TGraph::e_Int:
        {{
            const CInt_graph& graph = gr.GetGraph().GetInt();
            x_AddGraphToMap(g_obj.m_Map, gr, graph.GetValues(),
                graph.GetMax(), graph.GetMin(), g_obj.m_Max,
                g_obj.m_Min, fixed_scale);
            break;
        }}
        case CSeq_graph::TGraph::e_Real:
        {{
            const CReal_graph& graph = gr.GetGraph().GetReal();
            x_AddGraphToMap(g_obj.m_Map, gr, graph.GetValues(),
                graph.GetMax(), graph.GetMin(), g_obj.m_Max,
                g_obj.m_Min, fixed_scale);
            break;
        }}
        default:
            break;
        }
    }

    NON_CONST_ITERATE (TDMaps, iter, d_maps) {
        SGraphObject& g_obj = iter->second;
        CHistogramGlyph *hist = new CHistogramGlyph(g_obj.m_Map, iter->first);
        CRef<CSeqGlyph> fref(hist);
        hist->SetDesc(g_obj.m_Desc);
        hist->SetAnnotName(g_obj.m_AnnotName);
        if (fixed_scale) {
            hist->SetAxisMax(g_obj.m_Max);
            hist->SetAxisMin(g_obj.m_Min);
        }
        glyphs.push_back(fref);
    }

    return eCompleted;
}


template <typename ValType>
void CSGAnnotJob::x_AddGraphToMap(CHistogramGlyph::TMap& the_map,
                                  const CMappedGraph& gr,
                                  const vector<ValType>& vec,
                                  double l_max, double l_min,
                                  double& g_max, double& g_min,
                                  bool fixed_scale)
{
    if (vec.empty()) return;

    typedef CHistogramGlyph::TDataType TDataType;
    TSeqRange r  = gr.GetLoc().GetTotalRange();
    TSeqRange map_r = r.IntersectionWith(m_Range);
    TSeqPos pos  = r.GetFrom();
    double a = gr.IsSetA() ? gr.GetA() : 1.0;
    double b = gr.IsSetB() ? gr.GetB() : 0.0;

    TSeqPos comp = gr.IsSetComp() ? gr.GetComp() : 0;
    if (comp == 0) {
        comp = r.GetLength() / vec.size();
    }
    if (comp == 0) {
        return;
    }
    // skip the uninterested range

    typename vector<ValType>::const_iterator iter_gr = vec.begin();
    int offset = (map_r.GetFrom() - pos) / comp;
    iter_gr += offset;
    pos += offset * comp;

    // if the data type is char, we need to convert them into unsigned char
    bool unsigned_char = false;
    if (sizeof(ValType) == 1) unsigned_char = true;
    TDataType val;
    while (iter_gr != vec.end()  &&  pos < map_r.GetToOpen()) {
        TSeqRange range = TSeqRange(pos, pos + comp - 1);
        if (unsigned_char) {
            val = (TDataType)((unsigned char)(*iter_gr) * a + b);
        } else {
            val = (TDataType)((*iter_gr) * a + b);
        }
        the_map.AddRange(range, val);
        pos += comp;
        ++iter_gr;
    }

    if (fixed_scale) {
        g_max = max(l_max * a + b, g_max);
        g_min = min(l_min * a + b, g_min);
    }
}


string CSGAnnotJob::x_GetGraphName(const CMappedGraph& gr) const
{
    string name = kEmptyStr;

    /// We should favor title over annotation name.
    if (gr.IsSetTitle()) {
        name = gr.GetTitle();
    } else {
        name = CSeqUtils::GetAnnotName(gr.GetAnnot());
    }

    return name;
}


END_NCBI_SCOPE
