/*  $Id: track_config_manager.cpp 27773 2013-03-29 19:07:24Z wuliangs $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Liangshou Wu
 *
 */


#include <ncbi_pch.hpp>
#include <gui/widgets/seq_graphic/track_config_manager.hpp>
#include <gui/widgets/seq_graphic/seqgraphic_conf.hpp>

#include <gui/widgets/seq_graphic/config_utils.hpp>
#include <gui/widgets/seq_graphic/seqgraphic_utils.hpp>

#include <gui/widgets/seq_graphic/sequence_track.hpp>
#include <gui/widgets/seq_graphic/gene_model_track.hpp>
#include <gui/widgets/seq_graphic/feature_track.hpp>
#include <gui/widgets/seq_graphic/segment_map_track.hpp>
#include <gui/widgets/seq_graphic/alignment_track.hpp>
#include <gui/widgets/seq_graphic/all_other_features_track.hpp>
#include <gui/widgets/seq_graphic/six_frames_trans_track.hpp>
#include <gui/widgets/seq_graphic/graph_track.hpp>
#include <gui/widgets/seq_graphic/seqgraphic_graph_ds.hpp>
#include <gui/widgets/seq_graphic/epigenomics_track.hpp>
#include <gui/widgets/seq_graphic/dbvar_track.hpp>
#include <gui/widgets/seq_graphic/track_container_track.hpp>
#include <gui/widgets/seq_graphic/seqtable_graph_track.hpp>

#include <gui/widgets/seq_graphic/track_container_track.hpp>

#include <gui/utils/extension_impl.hpp>
#include <gui/objutils/utils.hpp>

#include <limits.h>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

#define MAX_LEVEL numeric_limits<int>::max()


CTrackConfigManager::CTrackConfigManager()
{
    x_InitTrackManager();
    m_KnownAnnots[MAX_LEVEL] = TAnnotMetaDataList();
}


void CTrackConfigManager::RegisterSGTracks()
{
    // register layout track factories
    CExtensionDeclaration(EXT_POINT__SEQGRAPHIC_LAYOUT_TRACK_FACTORY,
        new CSequenceTrackFactory());
    CExtensionDeclaration(EXT_POINT__SEQGRAPHIC_LAYOUT_TRACK_FACTORY,
        new CGeneModelFactory());
    CExtensionDeclaration(EXT_POINT__SEQGRAPHIC_LAYOUT_TRACK_FACTORY,
        new CFeatureTrackFactory());
    CExtensionDeclaration(EXT_POINT__SEQGRAPHIC_LAYOUT_TRACK_FACTORY,
        new CSegmentMapTrackFactory());
    CExtensionDeclaration(EXT_POINT__SEQGRAPHIC_LAYOUT_TRACK_FACTORY,
        new CAlignmentTrackFactory());
    CExtensionDeclaration(EXT_POINT__SEQGRAPHIC_LAYOUT_TRACK_FACTORY,
        new CGraphTrackFactory());
    CExtensionDeclaration(EXT_POINT__SEQGRAPHIC_LAYOUT_TRACK_FACTORY,
        new CTrackContainerFactory());
    CExtensionDeclaration(EXT_POINT__SEQGRAPHIC_LAYOUT_TRACK_FACTORY,
        new CAllOtherFeaturesTrackFactory());
    CExtensionDeclaration(EXT_POINT__SEQGRAPHIC_LAYOUT_TRACK_FACTORY,
        new CSixFramesTransTrackFactory());
    CExtensionDeclaration(EXT_POINT__SEQGRAPHIC_LAYOUT_TRACK_FACTORY,
        new CEpigenomicsTrackFactory());
    CExtensionDeclaration(EXT_POINT__SEQGRAPHIC_LAYOUT_TRACK_FACTORY,
        new CVarTrackFactory());
    CExtensionDeclaration(EXT_POINT__SEQGRAPHIC_LAYOUT_TRACK_FACTORY,
        new CSeqTableGraphTrackFactory());

    // register data source extensions
    CExtensionDeclaration(EXT_POINT__SEQGRAPHIC_DATA_SOURCE_TYPE,
        new CSGSequenceDSType());
    CExtensionDeclaration(EXT_POINT__SEQGRAPHIC_DATA_SOURCE_TYPE,
        new CSGSegmentMapDSType());
    CExtensionDeclaration(EXT_POINT__SEQGRAPHIC_DATA_SOURCE_TYPE,
        new CSGFeatureDSType());
    CExtensionDeclaration(EXT_POINT__SEQGRAPHIC_DATA_SOURCE_TYPE,
        new CSGAlignmentDSType());
    CExtensionDeclaration(EXT_POINT__SEQGRAPHIC_DATA_SOURCE_TYPE,
        new CSGGraphDSType());
    CExtensionDeclaration(EXT_POINT__SEQGRAPHIC_DATA_SOURCE_TYPE,
        new CFeaturePanelDSType());
    CExtensionDeclaration(EXT_POINT__SEQGRAPHIC_DATA_SOURCE_TYPE,
        new CSFTransDSType());
    CExtensionDeclaration(EXT_POINT__SEQGRAPHIC_DATA_SOURCE_TYPE,
        new CEpigenomicsDSType());
}


static string s_UpperFirstChar(const string& str)
{
    string new_str(str);
    if ( !new_str.empty() ) {
        new_str[0] = toupper(str[0]);
    }
    return new_str;
}


const ILayoutTrackFactory*
CTrackConfigManager::GetTrackFactory(const string& key) const
{
    TFactoryMap::const_iterator f_iter = m_Factories.find(key);
    if (f_iter != m_Factories.end()) {
        return f_iter->second;
    }
    
    return NULL;
}


const CTrackConfigManager::TAnnotMetaDataList&
CTrackConfigManager::GetKnownAnnots(int level) const
{
    if (level == -1) level = 0;
    TKnownAnnotMap::const_iterator iter = m_KnownAnnots.find(level);
    if (iter == m_KnownAnnots.end()) {
        level = MAX_LEVEL;
        iter = m_KnownAnnots.find(level);
        _ASSERT(iter != m_KnownAnnots.end());
    }
    
    return iter->second;
}


void CTrackConfigManager::SetNAAs(const CTrackConfigManager::TNAAs& naas)
{
    m_NAAs = naas;

    // now we remove the NAAs for those have been
    // used by any particular tracks.
    TNAAs::iterator iter = m_NAAs.begin();
    while (iter != m_NAAs.end()) {
        bool found = false;
        ITERATE (TUsedAnnotMap, t_iter, m_UsedAnnots) {
            TUsedAnnots::const_iterator a_iter = t_iter->second.find(*iter);
            if (a_iter != t_iter->second.end()) {
                iter = m_NAAs.erase(iter);
                found = true;
                break;
            }
        }

        if ( !found ) {
            ++iter;
        }
    }

    // add to gui registry
    CGuiRegistry& reg = CGuiRegistry::GetInstance();
    CGuiRegistry::TReadWriteView view =
        reg.GetReadWriteView("GBENCH.Utils.NamedAnnots");
    int idx = 0;
    ITERATE (TNAAs, na_iter, m_NAAs) {
        view.Set(NStr::IntToString(++idx), *na_iter);
    }

}


void CTrackConfigManager::ResetHistory()
{
    m_UsedAnnots.clear();
    x_InitOtherFeatList();
}


void CTrackConfigManager::NeedToShow(const CTempTrackProxy* p_proxy,
                                     CRef<CTempTrackProxy> t_proxy,
                                     const string& track_name) const
{
    if ( !t_proxy->GetShown()  &&  CSeqUtils::IsNAA(track_name) ) {
        if (!m_NAAs.empty()  &&
            (m_NAAs.front() == "NA*"  ||
             find(m_NAAs.begin(), m_NAAs.end(), track_name) !=
             m_NAAs.end())) {
            t_proxy->SetShown(true);
        } else {
            const CTrackProxy* temp_proxy =
                dynamic_cast<const CTrackProxy*>(p_proxy);
            if (temp_proxy) {
                const CTrackProxy::TAnnots& annots = temp_proxy->GetAnnots();
                if (find(annots.begin(), annots.end(), track_name) != annots.end()  ||
                    (!annots.empty()  &&  annots.front() == "NA*")) {
                    t_proxy->SetShown(true);
                }
            }
        }
    }
}


bool CTrackConfigManager::OnShownList(const CTempTrackProxy* p_proxy,
                                      const CTempTrackProxy* /*t_proxy*/,
                                      const string& track_name) const
{
    bool is_naa = CSeqUtils::IsNAA(track_name);

    if ( !p_proxy->GetShown() ) {
        if (is_naa) {
            if (!m_NAAs.empty()  &&
                (m_NAAs.front() == "NA*"  ||
                 find(m_NAAs.begin(), m_NAAs.end(), track_name) !=
                 m_NAAs.end())) {
                return true;
            }
        }
    } else {
        if (is_naa) {
            const CTrackProxy* temp_proxy =
                dynamic_cast<const CTrackProxy*>(p_proxy);
            if (temp_proxy) {
                const CTrackProxy::TAnnots& annots = temp_proxy->GetAnnots();
                if (find(annots.begin(), annots.end(), track_name) != annots.end()  ||
                    (!annots.empty()  &&  annots.front() == "NA*")) {
                    return true;
                }
            }
        } else {
            return true;
        }
    }
    
    return false;
}


void CTrackConfigManager::AdjustTrackOrder(CTempTrackProxy* t_proxy) const
{
    // sort the tracks according to the order of both 
    // annotation list in the track proxy and NAAs.
    // if there is any conflict, the former order is prefered.
    // we only sort the tracks under some special cases,
    // mainly for cgi mode, to avoid frustrating users.
    // annot list contain more than one annot and at least
    // one of them is NAA, or NAAs is not empty.
    CTrackContainer* con_track =
        dynamic_cast<CTrackContainer*>(t_proxy->GetTrack());
    if ( !con_track  || con_track->GetSubtrackProxies().size() < 2 ) {
        return; // no need to sort
    }

    typedef pair<string, bool> TAnnotStatus;
    list<TAnnotStatus> all_tracks;

    // honor the track order in track's annot list over m_NAAs list
    const CTrackProxy* full_track = dynamic_cast<const CTrackProxy*>(t_proxy);
    if (full_track) {
        ITERATE (CTrackProxy::TAnnots, iter, full_track->GetAnnots()) {
            string name = *iter;
            CSeqUtils::TAnnotNameType type =
                CSeqUtils::NameTypeStrToValue(name);
            if (type == CSeqUtils::eAnnot_Other  ||
                type == CSeqUtils::eAnnot_Unnamed) {
                NStr::ReplaceInPlace(name, ".", "_");
                all_tracks.push_back(TAnnotStatus(name, false));
            }
        }
    }

    // Now, consider track order in m_NAAs list
    ITERATE (TNAAs, iter, m_NAAs) {
        string name = *iter;
        NStr::ReplaceInPlace(name, ".", "_");
        all_tracks.push_back(TAnnotStatus(name, false));
        // NOTE: m_NAAs should not contain any NAA explicitly listed in 
        // track annots list.  We have removed them from m_NAAs.
    }

    if ( !all_tracks.empty() ) {
        // If no specific order has been specified for Unnamed annotatioin,
        // we always put it as the first track
        list<TAnnotStatus>::iterator t_iter = all_tracks.begin();
        while (t_iter != all_tracks.end()  &&  !CSeqUtils::IsUnnamed(t_iter->first)) {
            ++t_iter;
        }
        if (t_iter == all_tracks.end()) {
            all_tracks.push_front(TAnnotStatus(CSeqUtils::GetUnnamedAnnot(), false));
        }

        vector<int> old_orders;

        // mark the tracks that are in track container and save there old orders
        ITERATE (TTrackProxies, iter, con_track->GetSubtrackProxies()) {
            const string& name = (*iter)->GetName();
            NON_CONST_ITERATE (list<TAnnotStatus>, tt_iter, all_tracks) {
                if (tt_iter->first == name) {
                    old_orders.push_back((*iter)->GetOrder());
                    tt_iter->second = true;
                    break;
                }
            }
        }

        // remove the tracks that are not in track container from all_tracks
        t_iter = all_tracks.begin();
        while (t_iter != all_tracks.end()) {
            if ( !t_iter->second ) {
                t_iter = all_tracks.erase(t_iter);
            } else {
                ++t_iter;
            }
        }

        if (all_tracks.size() > old_orders.size()) {
            all_tracks.pop_front();
        }

        NON_CONST_ITERATE (TTrackProxies, iter, con_track->GetSubtrackProxies()) {
            int idx = 0;
            t_iter = all_tracks.begin();
            while (t_iter != all_tracks.end()  &&
                t_iter->first != (*iter)->GetName()) {
                    ++idx;
                    ++t_iter;
            }
            if (t_iter != all_tracks.end()) {
                (*iter)->SetOrder(old_orders[idx]);
            }
        }
        con_track->GetSubtrackProxies().sort(CTrackProxy::STPSorterByOrder::s_CompareCRefs);
        con_track->SortChildTracksByOrder();
    }
}


/// help struct for storing track settings passed from url.
struct STrackSettings
{
    typedef vector<STrackSettings> TTrackConfig;
    typedef CSGConfigUtils::TKeyValuePairs   TProfile;

    STrackSettings()
        : m_Key(kEmptyStr)
        , m_Subkey(kEmptyStr)
        , m_Name(kEmptyStr)
        , m_Id(kEmptyStr)
        , m_DB(kEmptyStr)
        , m_Category(kEmptyStr)
        , m_Subcategory(kEmptyStr)
        , m_SettingGroup(kEmptyStr)
        , m_UseGroupDefault(false)
        , m_Filter(kEmptyStr)
        , m_SortBy(kEmptyStr)
        , m_DispName(kEmptyStr)
        , m_Order(-1)
        , m_Shown(true)
        {}

    static void ParseSettings(
        const CTrackConfigManager::TTrackSettingsSet& settings_set,
        TTrackConfig& config)
        {
            ITERATE (CTrackConfigManager::TTrackSettingsSet, iter, settings_set) {
                const CTrackConfigManager::TTrackSettings& track = *iter;
                CTrackConfigManager::TTrackSettings::const_iterator s_iter = track.find("key");
                if (s_iter == iter->end()) continue;
        
                STrackSettings settings;
                s_iter = track.begin();
                while (s_iter != track.end()) {
                    const string& k_str = s_iter->first;
                    if (NStr::EqualNocase(k_str, "key")) {
                        settings.m_Key = s_iter->second;
                    } else if (NStr::EqualNocase(k_str, "subkey")) {
                        settings.m_Subkey = s_iter->second;
                    } else if (NStr::EqualNocase(k_str, "name")) {
                        settings.m_Name = CSeqUtils::RemoveEscape(s_iter->second);
                    } else if (NStr::EqualNocase(k_str, "id")) {
                        settings.m_Id = CSeqUtils::RemoveEscape(s_iter->second);
                    } else if (NStr::EqualNocase(k_str, "dbname")) {
                        settings.m_DB = s_iter->second;
                    } else if (NStr::EqualNocase(k_str, "order")) {
                        try {
                            settings.m_Order = NStr::StringToInt(s_iter->second);
                        } catch (CException&) {
                            // ignore
                        }
                    } else if (NStr::EqualNocase(k_str, "shown")) {
                        try {
                            settings.m_Shown = NStr::StringToBool(s_iter->second);
                        } catch (CException&) {
                            // ignore
                        }
                    } else if (NStr::EqualNocase(k_str, "category")) {
                        settings.m_Category = s_iter->second;
                    } else if (NStr::EqualNocase(k_str, "subcategory")) {
                        settings.m_Subcategory = s_iter->second;
                    } else if (NStr::EqualNocase(k_str, "setting_group")) {
                        vector<string> tokens;
                        CSeqUtils::TokenizeWithEscape(s_iter->second, "|", tokens, true);
                        if (tokens.size() > 0) {
                            settings.m_SettingGroup = NStr::TruncateSpaces(tokens[0]);
                            if (tokens.size() > 1) {
                                try {
                                    settings.m_UseGroupDefault =
                                        NStr::StringToBool(NStr::TruncateSpaces(tokens[1]));
                                } catch (CException&) {
                                    // ignore it
                                }
                            }
                        }
                    } else if (NStr::EqualNocase(k_str, "filter")) {
                        settings.m_Filter = s_iter->second;
                    } else if (NStr::EqualNocase(k_str, "sort_by")) {
                        settings.m_SortBy = s_iter->second;
                    } else if (NStr::EqualNocase(k_str, "display_name")) {
                        settings.m_DispName = CSeqUtils::RemoveEscape(s_iter->second);
                    } else if (NStr::EqualNocase(k_str, "annots")) {
                        CSeqUtils::TokenizeWithEscape(s_iter->second, "|", settings.m_Annots, true);
                    } else if (NStr::EqualNocase(k_str, "comments")) {
                        settings.m_Comments = s_iter->second;
                    } else if (NStr::EqualNocase(k_str, "highlights")) {
                        settings.m_Highlights = s_iter->second;
                    } else {
                        settings.m_Profile[s_iter->first] = s_iter->second;
                    }
                    ++s_iter;
                }
                config.push_back(settings);
            }
        }
   
    static string CreateProfileString(const TProfile& profile) 
        {
            string profile_str = kEmptyStr;
            bool first = true;
            ITERATE (TProfile, iter, profile) {
                if (first) first = false;
                else  profile_str += ",";
                profile_str += iter->first + ":" + iter->second;
            }
            return profile_str;
        }

    string m_Key;
    string m_Subkey;
    string m_Name;
    string m_Id;
    string m_DB;
    string m_Category;
    string m_Subcategory;
    string m_SettingGroup;
    bool   m_UseGroupDefault;
    string m_Filter;
    string m_SortBy;
    string m_DispName;
    int    m_Order;
    bool   m_Shown;

    TProfile m_Profile;
    CTrackProxy::TAnnots m_Annots;
    string m_Comments;
    string m_Highlights;
};


/// base help functor to search a track proxy using a key value.
class STPFinder 
{
public:
    STPFinder(const string& key) : m_Key(key) {}
    virtual ~STPFinder() {}
    virtual bool operator() (const CRef<CTempTrackProxy>& t) = 0;

protected:
    string m_Key;
};

/// help functor to search a track proxy by track key.
class STPFinderByKey : public STPFinder
{
public:
    STPFinderByKey(const string& key) : STPFinder(key) {}

    virtual bool operator() (const CRef<CTempTrackProxy>& t)
        {
            const CTrackProxy* track =
                dynamic_cast<const CTrackProxy*>(t.GetPointer());
            if (!track) return false;
            return m_Key == track->GetKey();
        }
};


/// help functor to search a track proxy by track subkey.
class STPFinderBySubkey : public STPFinder
{
public:
    STPFinderBySubkey(const string& key) : STPFinder(key) {}

    virtual bool operator() (const CRef<CTempTrackProxy>& t)
        {
            const CTrackProxy* track =
                dynamic_cast<const CTrackProxy*>(t.GetPointer());
            if (!track) return false;
            return m_Key == track->GetSubkey();
        }
};


/// help functor to search a track proxy by track name
class STPFinderByName : public STPFinder
{
public:
    STPFinderByName(const string& key) : STPFinder(key) {}

    bool operator() (const CRef<CTempTrackProxy>& t)
        {
            return m_Key == t->GetName();
        }
};


void CTrackConfigManager::AlignTrackConfig(TTrackProxies& t_settings,
    const TTrackSettingsSet& in_settings)
{
    // in_settings: the input track list (and settings) that we want
    //              to apply to the existing track list
    // t_settings:  an existing track list which we want to update

    // what we want to acheive:
    //   1. only show the tracks presented in the in_settings
    //   2. keep the original settings not presented in the input track
    //      settings for a given track
    //   3. insert the tracks not present in the original track list
    //   4. show them in the order presented in the input list
    //
    // what we will do to acheive above:
    //   1. Reset the track on/off state
    //   2. Update the track settings in the original track list that
    //      also present in the input track list. The potential updated
    //      settings include track on/off state, annot list, and
    //      track-specific settings.  Track order is not updated at
    //      this point.
    //      Put the tracks that present in both the input track list and
    //      the original track list in a new list acording to the track 
    //      order in the input track list. For track that presents only
    //      in the input track list, create a track proxy for them and
    //      add them into the bottom of the oringal track list, and put 
    //      it in the newly created track list accordingly.
    //   3. Go through the updated original track list, create a order list
    //      for the tracks that are set to on.
    //   4. Go through the newly created track list, update the track
    //      order ids using the stored order list.
    //   5. Sort the tracks.

    // first, turn all track off
    NON_CONST_ITERATE (TTrackProxies, iter, t_settings) {
        (*iter)->SetShown(false);
    }

    STrackSettings::TTrackConfig config;
    STrackSettings::ParseSettings(in_settings, config);
    t_settings.sort(CTrackProxy::STPSorterByOrder::s_CompareCRefs);
    int last_id = t_settings.back()->GetOrder();
    
    // second, update the track settings
    TTrackProxies show_list;
    ITERATE (STrackSettings::TTrackConfig, iter, config) {
        TTrackProxies::iterator t_iter = t_settings.end();
        if ( !iter->m_Name.empty() ) { 
            STPFinderByName finder(iter->m_Name);
            t_iter = find_if(t_settings.begin(), t_settings.end(), finder);
        } else if ( !iter->m_Subkey.empty() ) {
            STPFinderBySubkey finder(iter->m_Subkey);
            t_iter = find_if(t_settings.begin(), t_settings.end(), finder);
        } else {
            STPFinderByKey finder(iter->m_Key);
            t_iter = find_if(t_settings.begin(), t_settings.end(), finder);
            // The target track is not supposed to have subkey.
            // Make sure the searched track if found has no subkey.
            if (t_iter != t_settings.end()) {
                const CTrackProxy* this_track = dynamic_cast<const CTrackProxy*>(t_iter->GetPointer());
                _ASSERT(this_track);
                if ( !this_track->GetSubkey().empty() ) {
                    // The found track doesn't actually match the target.
                    t_iter = t_settings.end();
                }
            }
        }

        CRef<CTrackProxy> track;
        if (t_iter != t_settings.end()) {
            track.Reset(dynamic_cast<CTrackProxy*>(t_iter->GetPointer()));
        } else {
            track.Reset(new CTrackProxy());
            t_settings.push_back(CRef<CTempTrackProxy>(track));
            if (iter->m_Order >= 0) {
                track->SetOrder(iter->m_Order);
            } else {
                track->SetOrder(++last_id);
            }
            track->SetKey(iter->m_Key);
            if ( !iter->m_Name.empty() ) {
                track->SetName(iter->m_Name);
            }
            if ( !iter->m_Subkey.empty() ) {
                track->SetSubkey(iter->m_Subkey);
                // assign a name if in case it is not named.
                if (track->GetName().empty()) {
                    track->SetName(iter->m_Subkey);
                }

                // udpate other feature list
                CFeatListItem item;
                CSeqFeatData::GetFeatList()->GetItemByKey(iter->m_Subkey, item);
                TSubtypeList::iterator feat_iter = find(m_OtherFeats.begin(),
                                                        m_OtherFeats.end(), item.GetSubtype());
                if (feat_iter != m_OtherFeats.end()) {
                    m_OtherFeats.erase(feat_iter);
                }

            }
        }
        if (iter->m_Shown) {
            show_list.push_back(CRef<CTempTrackProxy>(track));
            track->SetShown(true);
        } else {
            track->SetShown(false);
        }
        if ( !iter->m_Profile.empty() ) {
            track->SetTrackProfile(STrackSettings::CreateProfileString(iter->m_Profile));
        }
        if ( !iter->m_Annots.empty() ) {
            track->SetAnnots(iter->m_Annots);
        }
        if ( !iter->m_Category.empty() ) {
            track->SetCategory(iter->m_Category);
        }
        if ( !iter->m_Subcategory.empty() ) {
            track->SetSubcategory(iter->m_Subcategory);
        }
        if ( !iter->m_SettingGroup.empty() ) {
            track->SetSettingGroup(iter->m_SettingGroup);
        }
        track->SetUseGroupDefault(iter->m_UseGroupDefault);

        if ( !iter->m_Filter.empty() ) {
            track->SetFilter(iter->m_Filter);
        }
        if ( !iter->m_SortBy.empty() ) {
            track->SetSortBy(iter->m_SortBy);
        }
        if ( !iter->m_DispName.empty() ) {
            track->SetDisplayName(iter->m_DispName);
        }
        if ( !iter->m_Id.empty() ) {
            track->SetId(iter->m_Id);
        }
        if ( !iter->m_DB.empty() ) {
            track->SetDB(iter->m_DB);
        }

        track->SetComments(iter->m_Comments);
        track->SetHighlights(iter->m_Highlights);
    }

    vector<int> order_list;
    // we also want to refill the annotations used since the track
    // config may be changed.
    m_UsedAnnots.clear();
    NON_CONST_ITERATE (TTrackProxies, iter, t_settings) {
        CTrackProxy* track = dynamic_cast<CTrackProxy*>(iter->GetPointer());
        if (track->GetShown()) {
            order_list.push_back(track->GetOrder());
        }
        ITERATE(CTrackProxy::TAnnots, a_iter, track->GetAnnots()) {
            string track_key = track->GetKey() + track->GetSubkey() +
                track->GetFilter() + track->GetSortBy();
            m_UsedAnnots[track_key].insert(*a_iter);
        }
    }

    // make sure track order in both the input tracks and original track
    // settings are preserved in favor of input tracks order if there is a conflict.
    _ASSERT(order_list.size() == show_list.size());
    size_t idx = 0;
    NON_CONST_ITERATE (TTrackProxies, iter, show_list) {
        (*iter)->SetOrder(order_list[idx++]);
    }

    // sort the tracks
    t_settings.sort(CTrackProxy::STPSorterByOrder::s_CompareCRefs);
}


void CTrackConfigManager::UpdateTrackConfig(TTrackProxies& t_settings,
    const TTrackSettingsSet& in_settings)
{
    STrackSettings::TTrackConfig config;
    STrackSettings::ParseSettings(in_settings, config);

    ITERATE (STrackSettings::TTrackConfig, iter, config) {
        if (iter->m_Annots.empty()  &&  iter->m_Profile.empty()) continue;

        bool apply_to_many = iter->m_Name.empty();
        NON_CONST_ITERATE (TTrackProxies, t_iter, t_settings) {
            CTrackProxy* track =
                dynamic_cast<CTrackProxy*>(t_iter->GetPointer());
            if (!track) continue;
            
            if ( (!iter->m_Name.empty()  &&  iter->m_Name == track->GetName())  ||
                 (!iter->m_Subkey.empty()  &&  iter->m_Subkey == track->GetSubkey())  ||
                 (iter->m_Subkey.empty()  &&  iter->m_Key == track->GetKey()) ) {

                if ( !iter->m_Annots.empty() ) {
                    track->SetAnnots(iter->m_Annots);
                }
                if ( !iter->m_Category.empty() ) {
                    track->SetCategory(iter->m_Category);
                }
                if ( !iter->m_Subcategory.empty() ) {
                    track->SetSubcategory(iter->m_Subcategory);
                }
                if ( !iter->m_SettingGroup.empty() ) {
                    track->SetSettingGroup(iter->m_SettingGroup);
                }
                track->SetUseGroupDefault(iter->m_UseGroupDefault);

                if ( !iter->m_Filter.empty() ) {
                    track->SetFilter(iter->m_Filter);
                }
                if ( !iter->m_SortBy.empty() ) {
                    track->SetSortBy(iter->m_SortBy);
                }
                if ( !iter->m_DispName.empty() ) {
                    track->SetDisplayName(iter->m_DispName);
                }
                if ( !iter->m_Id.empty() ) {
                    track->SetId(iter->m_Id);
                }
                if ( !iter->m_DB.empty() ) {
                    track->SetDB(iter->m_DB);
                }

                track->SetComments(iter->m_Comments);
                track->SetHighlights(iter->m_Highlights);

                if ( track->GetTrackProfile().empty() ) {
                    track->SetTrackProfile(STrackSettings::CreateProfileString(iter->m_Profile));
                } else if ( !iter->m_Profile.empty() ) {
                    typedef CSGConfigUtils::TKeyValuePairs   TProfile;
                    TProfile old_profile;
                    CSGConfigUtils::ParseProfileString(track->GetTrackProfile(), old_profile);
                    if (old_profile.empty()) {
                        // it is the preset style with now 'key'
                        old_profile["profile"] = track->GetTrackProfile();
                    }
                    ITERATE (TProfile, kv_iter, iter->m_Profile) {
                        if (old_profile.count(kv_iter->first) == 0) {
                            old_profile[kv_iter->first] = kv_iter->second;
                        }
                    }
                    track->SetTrackProfile(STrackSettings::CreateProfileString(old_profile));
                }

                if ( !apply_to_many ) break;
            }
        }
    }
}


void CTrackConfigManager::MergeTrackConfig(TTrackProxies& t_settings,
                                           const TTrackSettingsSet& in_settings,
                                           EAnnotMerge merge)
{
    STrackSettings::TTrackConfig config;
    STrackSettings::ParseSettings(in_settings, config);
    t_settings.sort(CTrackProxy::STPSorterByOrder::s_CompareCRefs);
    int last_id = t_settings.back()->GetOrder();

    ITERATE (STrackSettings::TTrackConfig, iter, config) {
        TTrackProxies::iterator t_iter = t_settings.end();
        if ( !iter->m_Name.empty() ) { 
            STPFinderByName finder(iter->m_Name);
            t_iter = find_if(t_settings.begin(), t_settings.end(), finder);
        } else if ( !iter->m_Subkey.empty() ) {
            STPFinderBySubkey finder(iter->m_Subkey);
            t_iter = find_if(t_settings.begin(), t_settings.end(), finder);
        } else {
            STPFinderByKey finder(iter->m_Key);
            t_iter = find_if(t_settings.begin(), t_settings.end(), finder);
            // The target track is not supposed to have subkey.
            // Make sure the searched track if found has no subkey.
            if (t_iter != t_settings.end()) {
                const CTrackProxy* this_track = dynamic_cast<const CTrackProxy*>(t_iter->GetPointer());
                _ASSERT(this_track);
                if ( !this_track->GetSubkey().empty() ) {
                    // The found track doesn't actually match the target.
                    t_iter = t_settings.end();
                }
            }
        }

        CRef<CTrackProxy> track;
        if (t_iter != t_settings.end()) {
            track.Reset(dynamic_cast<CTrackProxy*>(t_iter->GetPointer()));
        } else {
            track.Reset(new CTrackProxy());
            t_settings.push_back(CRef<CTempTrackProxy>(track));
            if (iter->m_Order >= 0) {
                track->SetOrder(iter->m_Order);
            } else {
                track->SetOrder(++last_id);
            }

            // udpate other feature list
            if ( !iter->m_Subkey.empty() ) {
                CFeatListItem item;
                CSeqFeatData::GetFeatList()->GetItemByKey(iter->m_Subkey, item);
                TSubtypeList::iterator feat_iter = find(m_OtherFeats.begin(),
                    m_OtherFeats.end(), item.GetSubtype());
                if (feat_iter != m_OtherFeats.end()) {
                    m_OtherFeats.erase(feat_iter);
                }
            }
        }
        bool was_on = track->GetShown();
        track->SetShown(iter->m_Shown);
        track->SetKey(iter->m_Key);
        if ( !iter->m_Profile.empty() ) {
            track->SetTrackProfile(STrackSettings::CreateProfileString(iter->m_Profile));
        }
        if ( !iter->m_Name.empty() ) {
            track->SetName(iter->m_Name);
        }

        if ( !iter->m_Annots.empty() ) {
            ITERATE(CTrackProxy::TAnnots, a_iter, iter->m_Annots) {
                string track_key = iter->m_Key + iter->m_Subkey +
                    iter->m_Filter + iter->m_SortBy;
                m_UsedAnnots[track_key].insert(*a_iter);
            }
            if (merge == eAnnot_Merge) {
                if ( !was_on ) {
                    track->SetAnnots(iter->m_Annots);
                } else if (!track->GetAnnots().empty() ) {
                    CTrackProxy::TAnnots annots = track->GetAnnots();
                    copy(iter->m_Annots.begin(), iter->m_Annots.end(), back_inserter(annots));
                    track->SetAnnots(annots);
                } else {
                    track->SetAnnots(iter->m_Annots);
                }
            } else if (merge == eAnnot_Replace) {
                track->SetAnnots(iter->m_Annots);
            }
        }

        if ( !iter->m_Category.empty() ) {
            track->SetCategory(iter->m_Category);
        }
        if ( !iter->m_Subcategory.empty() ) {
            track->SetSubcategory(iter->m_Subcategory);
        }
        if ( !iter->m_SettingGroup.empty() ) {
            track->SetSettingGroup(iter->m_SettingGroup);
        }
        track->SetUseGroupDefault(iter->m_UseGroupDefault);

        if ( !iter->m_Filter.empty() ) {
            track->SetFilter(iter->m_Filter);
        }
        if ( !iter->m_SortBy.empty() ) {
            track->SetSortBy(iter->m_SortBy);
        }
        if ( !iter->m_DispName.empty() ) {
            track->SetDisplayName(iter->m_DispName);
        }
        if ( !iter->m_Subkey.empty() ) {
            track->SetSubkey(iter->m_Subkey);
            if (track->GetName().empty()) track->SetName(iter->m_Subkey);
        }
        if ( !iter->m_Id.empty() ) {
            track->SetId(iter->m_Id);
        }
        if ( !iter->m_DB.empty() ) {
            track->SetDB(iter->m_DB);
        }

        track->SetComments(iter->m_Comments);
        track->SetHighlights(iter->m_Highlights);
    }
}


void CTrackConfigManager::SetTrackConfig(TTrackProxies& t_settings,
                                         const TTrackSettingsSet& in_settings)
{
    // what we want to acheive:
    //   1. only show the tracks presented
    //   2. keep the original settings not presented in the track
    //      settings for a given track
    //   3. show them in the order presented in the input list
    //
    // what we will do to acheive above:
    //   1. Reset the track order ids with a large number (max_id)
    //   2. Replace the track settings for a given track (in order)
    //      from the input in_settings.  The track order will
    //      be the sequence order number in the list
    //   3. Sort the tracks by order
    //   4. Remove all tracks whose track order is equal to the max_order


    // we also want to refill the annotations used since the track
    // conofig will be changed.
    ResetHistory();

    const int max_order = INT_MAX;

    // reset track order
    NON_CONST_ITERATE (TTrackProxies, iter, t_settings) {
        (*iter)->SetOrder(max_order);
    }

    // replace the track settings
    int order = 0;
    STrackSettings::TTrackConfig config;
    STrackSettings::ParseSettings(in_settings, config);
    ITERATE (STrackSettings::TTrackConfig, iter, config) {
        TTrackProxies::iterator t_iter = t_settings.end();
        if ( !iter->m_Name.empty() ) { 
            STPFinderByName finder(iter->m_Name);
            t_iter = find_if(t_settings.begin(), t_settings.end(), finder);
        } else if ( !iter->m_Subkey.empty() ) {
            STPFinderBySubkey finder(iter->m_Subkey);
            t_iter = find_if(t_settings.begin(), t_settings.end(), finder);

            // udpate other feature list
            CFeatListItem item;
            CSeqFeatData::GetFeatList()->GetItemByKey(iter->m_Subkey, item);
            TSubtypeList::iterator feat_iter = find(m_OtherFeats.begin(),
                m_OtherFeats.end(), item.GetSubtype());
            if (feat_iter != m_OtherFeats.end()) {
                m_OtherFeats.erase(feat_iter);
            }
        } else {
            STPFinderByKey finder(iter->m_Key);
            t_iter = find_if(t_settings.begin(), t_settings.end(), finder);
            // The target track is not supposed to have subkey.
            // Make sure the searched track if found has no subkey.
            if (t_iter != t_settings.end()) {
                const CTrackProxy* this_track = dynamic_cast<const CTrackProxy*>(t_iter->GetPointer());
                _ASSERT(this_track);
                if ( !this_track->GetSubkey().empty() ) {
                    // The found track doesn't actually match the target.
                    t_iter = t_settings.end();
                }
            }
        }

        CRef<CTrackProxy> track;
        if (t_iter != t_settings.end()) {
            track.Reset(dynamic_cast<CTrackProxy*>(t_iter->GetPointer()));
        } else {
            track.Reset(new CTrackProxy());
            t_settings.push_back(CRef<CTempTrackProxy>(track));
        }

        track->SetShown(iter->m_Shown);
        track->SetOrder(order++);
        track->SetKey(iter->m_Key);
        if ( !iter->m_Profile.empty() ) {
            track->SetTrackProfile(STrackSettings::CreateProfileString(iter->m_Profile));
        }
        if ( !iter->m_Name.empty() ) {
            track->SetName(iter->m_Name);
        }
        if ( !iter->m_Annots.empty() ) {
            track->SetAnnots(iter->m_Annots);
        }
        if ( !iter->m_Subkey.empty() ) {
            track->SetSubkey(iter->m_Subkey);
            if (track->GetName().empty()) track->SetName(iter->m_Subkey);
        }

        if ( !iter->m_Category.empty() ) {
            track->SetCategory(iter->m_Category);
        }
        if ( !iter->m_Subcategory.empty() ) {
            track->SetSubcategory(iter->m_Subcategory);
        }
        if ( !iter->m_SettingGroup.empty() ) {
            track->SetSettingGroup(iter->m_SettingGroup);
        }
        track->SetUseGroupDefault(iter->m_UseGroupDefault);

        if ( !iter->m_Filter.empty() ) {
            track->SetFilter(iter->m_Filter);
        }
        if ( !iter->m_SortBy.empty() ) {
            track->SetSortBy(iter->m_SortBy);
        }
        if ( !iter->m_DispName.empty() ) {
            track->SetDisplayName(iter->m_DispName);
        }
        if ( !iter->m_Id.empty() ) {
            track->SetId(iter->m_Id);
        }
        if ( !iter->m_DB.empty() ) {
            track->SetDB(iter->m_DB);
        }

        track->SetComments(iter->m_Comments);
        track->SetHighlights(iter->m_Highlights);

        // reinitialize the used annot list
        ITERATE(CTrackProxy::TAnnots, a_iter, track->GetAnnots()) {
            string track_key = track->GetKey() + track->GetSubkey() +
                track->GetFilter() + track->GetSortBy();
            m_UsedAnnots[track_key].insert(*a_iter);
        }
    }

    // sort the tracks
    t_settings.sort(CTrackProxy::STPSorterByOrder::s_CompareCRefs);

    // remove the rest of tracks
    TTrackProxies::iterator t_iter = t_settings.begin();
    while (t_iter != t_settings.end()  &&  (*t_iter)->GetOrder() < max_order) {
        ++t_iter;
    }
    t_settings.erase(t_iter, t_settings.end());
}


CRef<CTrackConfig> CTrackConfigManager::GetTrackSettings(
    const TTrackSettings& settings, CSeqGraphicConfig& g_config)
{
    CRef<CTrackConfig> config;
    
    TTrackSettings::const_iterator iter = settings.find("key");
    if (iter == settings.end()) return config;
    const string& key = iter->second;

    iter = settings.find("name");
    if (iter == settings.end()) return config;
    string name = CSeqUtils::RemoveEscape(iter->second);

    TFactoryMap::const_iterator f_iter = m_Factories.find(key);
    if (f_iter == m_Factories.end()) return config;
    
    CRef<CTrackConfigSet> configs;
    const ITrackConfigurable* factory =
        dynamic_cast<const ITrackConfigurable*>(f_iter->second.GetPointer());
    if (factory) {
        configs = factory->GetSettings(&g_config, "Default", settings, NULL);
        if ( !configs->Get().empty() ) {
            /// there must be only one track. Otherwise something is wrong
            _ASSERT(configs->Get().size() == 1);
            config = configs->Set().front();
        }
    } else {
        config.Reset(new CTrackConfig);
        config->SetHelp() = f_iter->second->GetThisTypeInfo().GetDescr();
    }
    
    if (config) {
        config->SetName() = name;
        config->SetKey() = key;

        iter = settings.find("display_name");
        if (iter != settings.end()) {
            config->SetDisplay_name() = CSeqUtils::RemoveEscape(iter->second);
        }
        iter = settings.find("category");
        if (iter != settings.end()) {
            CRef<CCategory> cat =
                CTrackConfigUtils::CreateCategory(iter->second, iter->second, "", 0);
            config->SetCategory(*cat);
        }
        iter = settings.find("subkey");
        if (iter != settings.end()) {
            config->SetSubkey() = iter->second;
        }

        iter = settings.find("setting_group");
        if (iter != settings.end()) {
            config->SetSetting_group() = iter->second;
        }

        iter = settings.find("id");
        if (iter != settings.end()  &&  !iter->second.empty()) {
            config->SetId(iter->second);
        }

        iter = settings.find("dbname");
        if (iter != settings.end()  &&  !iter->second.empty()) {
            config->SetDbname(iter->second);
        }
    }

    return config;
}


CRef<CTrackConfigSet>
CTrackConfigManager::GetTrackSettings(TTrackProxies& t_settings,
                                      CSeqGraphicConfig& g_config)
{
    t_settings.sort(CTrackProxy::STPSorterByOrder::s_CompareCRefs);
    CRef<CTrackConfigSet> config_set(new CTrackConfigSet);
    CRef<CCategoryConfig> cat_config = g_config.GetCategoryConfig();

    NON_CONST_ITERATE (TTrackProxies, t_iter, t_settings) {
        CTempTrackProxy* temp_proxy = t_iter->GetPointer();
        CTrackProxy* proxy = dynamic_cast<CTrackProxy*>(temp_proxy);
        CRef<CLayoutTrack> track(temp_proxy->GetTrack());
        // disconnect the track from its track proxy to use all settings
        // from the track proxy instead of the track
        temp_proxy->SetTrack(NULL);

        if ( !proxy  &&  !track ) continue;

        const string& key =
            proxy ? proxy->GetKey() : track->GetTypeInfo().GetId();
        TFactoryMap::const_iterator f_iter = m_Factories.find(key);
        if (f_iter == m_Factories.end()) continue;

        CRef<CTrackConfigSet> configs;
        // extract track profile and settings
        TKeyValuePairs settings;
        string profile = temp_proxy->GetTrackProfile();
        CSGConfigUtils::ParseProfileString(profile, settings);

        TKeyValuePairs::const_iterator iter = settings.find("profile");
        if (iter != settings.end()) {
            profile = iter->second;
        } else if ( !settings.empty()  ||  profile.empty()) {
            profile = "Default";
        }

        // link the track with proxy again in case it is needed for 
        // retrieving track settings
        temp_proxy->SetTrack(track.GetPointer());
        const ITrackConfigurable* factory =
            dynamic_cast<const ITrackConfigurable*>(f_iter->second.GetPointer());
        if (factory) {
            configs = factory->GetSettings(&g_config, profile, settings, temp_proxy);
        } else {
            CRef<CTrackConfig> config(new CTrackConfig);
            configs->Set().push_back(config);
            config->SetHelp() = f_iter->second->GetThisTypeInfo().GetDescr();
        }

        if (configs->Get().empty()) continue;

        config_set->Set().insert(config_set->Set().end(),
                                 configs->Get().begin(), configs->Get().end());
        CTrackConfig* config = configs->Set().front();
        if ( !temp_proxy->GetId().empty() ) {
            config->SetId(temp_proxy->GetId());
        }
        if (proxy  &&  !proxy->GetDB().empty() ) {
            config->SetDbname(proxy->GetDB());
        }

        if (config->SetKey().empty()) {
            // this can happen only when the config created is for
            // the proxy itself, hence, there will be only one config.
            _ASSERT(configs->Get().size() == 1);

            config->SetName() = temp_proxy->GetName();
            config->SetDisplay_name() =
                temp_proxy->GetDisplayName().empty() ? 
                track->GetFullTitle() : temp_proxy->GetDisplayName();
            config->SetKey() = key;
            config->SetOrder() = temp_proxy->GetOrder();
            if (proxy) {
                config->SetSubkey() = proxy->GetSubkey();
            } else {
                CFeatureTrack* f_track =
                    dynamic_cast<CFeatureTrack*>(track.GetPointer());
                if (f_track) {
                    const CFeatList& feats(*CSeqFeatData::GetFeatList());
                    config->SetSubkey() =
                        feats.GetStoragekey(f_track->GetFeatSubtype());
                }
            }

            config->SetFilter() = temp_proxy->GetFilter();
            config->SetSort_by() = temp_proxy->GetSortBy();
            config->SetShown() = temp_proxy->GetShown();
            CRef<CCategory> cat;
            CRef<CCategory> sub_cat;
            if (cat_config) {
                cat = cat_config->GetCategory(temp_proxy->GetCategory());
                sub_cat = cat_config->GetSubcategory(temp_proxy->GetCategory(),
                    temp_proxy->GetSubcategory());
            }
            if ( !cat ) {
                cat = CTrackConfigUtils::CreateCategory(
                    temp_proxy->GetCategory(), temp_proxy->GetCategory(), "", 0);
            }
            config->SetCategory(*cat);

            if ( !sub_cat ) {
                sub_cat = CTrackConfigUtils::CreateCategory(
                    temp_proxy->GetSubcategory(), temp_proxy->GetSubcategory(), "", 0);
            }
            config->SetSubcategory(*sub_cat);
            
            if ( !temp_proxy->GetSettingGroup().empty() ) {
                config->SetSetting_group() = temp_proxy->GetSettingGroup();
            }

            if (track) {
                if (temp_proxy->GetChildren().size() > 1) {
                    ITERATE (CTempTrackProxy::TTrackProxies, c_iter, temp_proxy->GetChildren()) {
                        if ((*c_iter)->GetTrack() != NULL) {
                            config->SetAnnots().push_back((*c_iter)->GetSource());
                        }
                    }
                } else {
                    string annot_name;
                    if (proxy  &&  proxy->GetAnnots().size() == 1) {
                        annot_name = *proxy->GetAnnots().begin();
                    } else {
                        annot_name = temp_proxy->GetSource();
                    }

                    if ( !annot_name.empty() ) {
                        config->SetAnnots().push_back(annot_name);
                    }
                }
            } else {
                config->SetAnnots().clear();
                std::copy(proxy->GetAnnots().begin(), proxy->GetAnnots().end(),
                          back_inserter(config->SetAnnots()));
            }
        }
    }
    
    return config_set;
}


CRef<CTrackConfig>
CTrackConfigManager::CreateTrackConfig(CTempTrackProxy* temp_proxy,
                                       CSeqGraphicConfig& g_config,
                                       const string& title_base) const
{
    CRef<CCategoryConfig> cat_config = g_config.GetCategoryConfig();
    
    const CTrackProxy* proxy = dynamic_cast<const CTrackProxy*>(temp_proxy);
    CRef<CLayoutTrack> track(temp_proxy->GetTrack());
    _ASSERT(track);
    
    // disconnect the track from its track proxy to use all settings
    // from the track proxy instead of the track
    temp_proxy->SetTrack(NULL);
    
    const string& key = track->GetTypeInfo().GetId();
    const ILayoutTrackFactory* factory = GetTrackFactory(key);
    _ASSERT(factory);

    // extract track profile and settings
    TKeyValuePairs settings;
    string profile = temp_proxy->GetTrackProfile();
    CSGConfigUtils::ParseProfileString(profile, settings);
    
    TKeyValuePairs::const_iterator p_iter = settings.find("profile");
    if (p_iter != settings.end()) {
        profile = p_iter->second;
    } else if ( !settings.empty()  ||  profile.empty()) {
        profile = "Default";
    }
    
    const string& annot_name = temp_proxy->GetSource();

    // get annotation meta-data if exists
    const CAnnotMetaData* meta_data = NULL;
    {{
            const TAnnotMetaDataList& md_list = GetKnownAnnots(0);
            CRef<CAnnotMetaData> dummy(new CAnnotMetaData);
            dummy->m_Name = annot_name;
            TAnnotMetaDataList::const_iterator iter = md_list.find(dummy);
            if (iter != md_list.end()) {
                meta_data = iter->GetPointer();
            }
    }}

    // link the track with proxy again in case it is needed for 
    // retrieving track settings
    temp_proxy->SetTrack(track.GetPointer());

    // create track-specific settings
    CRef<CTrackConfig> config;
    const ITrackConfigurable* conf_factory =
        dynamic_cast<const ITrackConfigurable*>(factory);
    if (conf_factory) {
        CRef<CTrackConfigSet> configs = 
            conf_factory->GetSettings(&g_config, profile, settings, temp_proxy);
        if ( !configs->Get().empty() ) {
            config = configs->Set().front();
        }
    } else {
        config.Reset(new CTrackConfig);
        config->SetHelp() = factory->GetThisTypeInfo().GetDescr();
    }
    
    if ( !config ) {
        return config;
    }

    // add 'comments' settings if any
    if ( !temp_proxy->GetComments().empty() ) {
        typedef vector<string> TCommentPos;
        TCommentPos cmt_pos;
        CSeqUtils::TokenizeWithEscape(temp_proxy->GetComments(), "|", cmt_pos, true);
        TCommentPos::const_iterator iter = cmt_pos.begin();
        while (iter != cmt_pos.end()) {
            const string& label = *iter;
            if (++iter != cmt_pos.end()) {
                config->SetComments().push_back(
                    CTrackConfigUtils::CreateComment(label, *iter));
                ++iter;
            }
        }
    }

    // add 'highlights' settings if any
    if ( !temp_proxy->GetHighlights().empty() ) {
        typedef vector<string> THighlights;
        THighlights hls;
        CSeqUtils::TokenizeWithEscape(temp_proxy->GetHighlights(), "|", hls, true);
        ITERATE (THighlights, h_iter, hls) {
            string hl = NStr::TruncateSpaces(*h_iter);
            if ( !hl.empty() ) {
                config->SetHighlights().push_back(hl);
            }
        }
    }

    // use annotation description for track help if exists
    if (meta_data  &&  !meta_data->m_Descr.empty()) {
        config->SetHelp() = NStr::Replace(meta_data->m_Descr, "\n", "<br/>");
    }

    // initialize track name and display name (title)
    config->SetName() = temp_proxy->GetName();
    if (!temp_proxy->GetDisplayName().empty()) {
        config->SetDisplay_name() = temp_proxy->GetDisplayName();
    } else if (meta_data  &&  !meta_data->m_Title.empty()) {
        config->SetDisplay_name() = meta_data->m_Title;
    } else if (CSeqUtils::IsUnnamed(annot_name)) {
        config->SetDisplay_name() = title_base;
    } else if (annot_name.find(title_base) != string::npos) {
        config->SetDisplay_name() = annot_name;
    } else if (title_base.find(annot_name) == string::npos) {
        config->SetDisplay_name() = title_base + " - " + annot_name;
    } else {
        config->SetDisplay_name() = title_base;
    }

    // initialize track key (track type) and subkey (subtype)
    config->SetKey() = key;
    config->SetOrder() = temp_proxy->GetOrder();
    string subkey = "";
    if (proxy) {
        if ( !proxy->GetSubkey().empty() ) {
            subkey = config->SetSubkey() = proxy->GetSubkey();
        }
    } else {
        const CFeatureTrack* f_track =
            dynamic_cast<const CFeatureTrack*>(track.GetPointer());
        if (f_track) {
            const CFeatList& feats(*CSeqFeatData::GetFeatList());
            config->SetSubkey() =
                feats.GetStoragekey(f_track->GetFeatSubtype());
        }
    }

    // initialize track category and subcategory
    // We set category and subcat in the following priority order:
    //  - setting from track proxy
    //  - xClass from meta-data
    //  - track type and subtype
    string cat_name = temp_proxy->GetCategory();
    string sub_cat_name = temp_proxy->GetSubcategory();
    if (cat_name.empty()) {
        if (meta_data  &&  !meta_data->m_xClass.empty()) {
            cat_name = meta_data->m_xClass;
            //if ( !meta_data->m_Provider.empty() ) {
            //    sub_cat_name = meta_data->m_Provider;
            //}
        } else {
            if (key == CGraphTrackFactory::GetTypeInfo().GetId()) {
                cat_name = "Graphs";
            } else if (key == CFeatureTrackFactory::GetTypeInfo().GetId()) {
                if (subkey.find("RNA") != string::npos  ||
                    subkey.find("exon") != string::npos) {
                    cat_name = "Genes";
                } else {
                    cat_name  = "Features";
                    sub_cat_name = title_base;
                }
            } else if (key == CGeneModelFactory::GetTypeInfo().GetId()) {
                cat_name = "Genes";
            } else if (key == CAlignmentTrackFactory::GetTypeInfo().GetId()) {
                cat_name = "Alignments";
            } else {
                cat_name = "Others";
            }
        }
    }


    CRef<CCategory> cat;
    CRef<CCategory> sub_cat;
    if (cat_config) {
        cat = cat_config->GetCategory(cat_name);
        sub_cat = cat_config->GetSubcategory(cat_name, sub_cat_name);
    }
    if ( !cat ) {
        cat = CTrackConfigUtils::CreateCategory(cat_name, cat_name, "", 10000);
    }
    config->SetCategory(*cat);

    if ( !sub_cat  &&  !sub_cat_name.empty() ) {
        sub_cat = CTrackConfigUtils::CreateCategory(sub_cat_name, sub_cat_name, "", 100);
    }

    if (sub_cat) {
        config->SetSubcategory(*sub_cat);
    }

    if ( !temp_proxy->GetSettingGroup().empty() ) {
        config->SetSetting_group() = temp_proxy->GetSettingGroup();
    }

    // fill in other track information.
    if ( !temp_proxy->GetFilter().empty() ) {
        config->SetFilter() = temp_proxy->GetFilter();
    }
    if ( !temp_proxy->GetSortBy().empty() ) {
        config->SetSort_by() = temp_proxy->GetSortBy();
    }
    if ( !temp_proxy->GetId().empty() ) {
        config->SetId(temp_proxy->GetId());
    }
    if (proxy  &&  !proxy->GetDB().empty() ) {
        config->SetDbname(proxy->GetDB());
    }

    config->SetShown() = temp_proxy->GetShown();
    config->SetAnnots().push_back(annot_name);
    
    return config;
}


void CTrackConfigManager::ShowTrack(TTrackProxies& t_settings,
                              const string& track_key)
{
    NON_CONST_ITERATE (TTrackProxies, t_iter, t_settings) {
        CTrackProxy* track = dynamic_cast<CTrackProxy*>(t_iter->GetPointer());
        if (track  &&  track_key == track->GetKey()) {
            track->SetShown(true);
            break;
        }
    }
}


void CTrackConfigManager::ShowAllTracks(TTrackProxies& t_settings)
{
    NON_CONST_ITERATE (TTrackProxies, t_iter, t_settings) {
        CTrackProxy* track = dynamic_cast<CTrackProxy*>(t_iter->GetPointer());
        if (track) {
            track->SetShown(true);
        }
    }
}


static const string k_order = "Order";
static const string k_category = "Category";
static const string k_subcategory = "Subcategory";
static const string k_setting_group = "SettingGroup";
static const string k_shown = "Shown";
static const string k_expanded = "Expanded";
static const string k_track_profile = "TrackProfile";
static const string k_key = "Key";
static const string k_subkey = "Subkey";
static const string k_filter = "Filter";
static const string k_sortby = "SortBy";
static const string k_display_name = "DisplayName";
static const string k_annots = "Annots";
static const string k_annot_separation = "AnnotSeparation";
static const string k_id = "Id";
static const string k_db = "dbname";


void CTrackConfigManager::SyncSettings(TTrackProxies& dst_proxies,
                                       const TTrackProxies& src_proxies,
                                       int level)
{
    // In multiple leve layout, we don't need to save settings on 
    // track container for level 0 - 2. So it will be partial
    bool partial = (level == -2);

    // sync settings
    ITERATE (TTrackProxies, iter, src_proxies) {
        const CTempTrackProxy* src_proxy = *iter;

        if ( !src_proxy->IsRealized() ) continue;

        if (partial  &&  src_proxy->GetTrack()) {
            const CTrackContainer* cont =
                dynamic_cast<const CTrackContainer*>(src_proxy->GetTrack());
            if (cont  &&  cont->GetAnnotLevel() != -2) {
                continue;
            }
        }

        TTrackProxies::iterator dst_iter = dst_proxies.begin();
        while (dst_iter != dst_proxies.end()  &&
            (*dst_iter)->GetName() != src_proxy->GetName()) {
            ++dst_iter;
        }
        if (dst_iter != dst_proxies.end()) {
            // ignore the proxy with no real track created
            if ( !src_proxy->GetTrack() ) continue;

            CTempTrackProxy* dst_proxy = *dst_iter;
            dst_proxy->SetOrder(src_proxy->GetOrder());
            dst_proxy->SetShown(src_proxy->GetShown());
            dst_proxy->SetExpanded(src_proxy->GetExpanded());
            dst_proxy->SetTrackProfile(src_proxy->GetTrackProfile());
            dst_proxy->SetSource(src_proxy->GetSource());
            dst_proxy->SetFilter(src_proxy->GetFilter());
            dst_proxy->SetSortBy(src_proxy->GetSortBy());
            dst_proxy->SetId(src_proxy->GetId());
            dst_proxy->SetDisplayName(src_proxy->GetDisplayName());
            if (src_proxy->GetTrack()) {
                const CTrackContainer* cont =
                    dynamic_cast<const CTrackContainer*>(src_proxy->GetTrack());
                if (cont  &&  !cont->GetSubtrackProxies().empty()) {
                    SyncSettings(dst_proxy->GetChildren(),
                        cont->GetSubtrackProxies(), level);
                }
            }

            if (src_proxy->IsEmpty()) { // remove a track
                dst_proxies.erase(dst_iter);
            }
        } else if ( !src_proxy->IsEmpty() ) {
            dst_proxies.push_back(src_proxy->Clone());
            if (src_proxy->GetTrack()) {
                const CTrackContainer* cont =
                    dynamic_cast<const CTrackContainer*>(src_proxy->GetTrack());
                if (cont  &&  !cont->GetSubtrackProxies().empty()) {
                    CTempTrackProxy* dst_proxy = dst_proxies.back();
                    SyncSettings(dst_proxy->GetChildren(),
                        cont->GetSubtrackProxies(), level);
                }
            }
        }
    }
    // sort the proxies according to track order
    dst_proxies.sort(CTrackProxy::STPSorterByOrder::s_CompareCRefs);
}


void CTrackConfigManager::LoadTrackSettingsRecursive(TTrackProxies& proxies,
                                                     const string& path)
{
    string key = path + ".Tracks";
    CGuiRegistry& registry = CGuiRegistry::GetInstance();
    CGuiRegistry::TReadView view = registry.GetReadView(key);
    CGuiRegistry::TKeys keys;
    view.GetTopKeys(keys);
    ITERATE (CGuiRegistry::TKeys, iter, keys) {
        /// HACK, ingore the old settings that doesn't exist in the next version.
        /// this code will be gone latter.
        if (iter->key == "All other features") continue;

        string base_key = key + CGuiRegistry::kDecimalDot + iter->key;
        // load settings for this track
        CRef<CTrackProxy> track(new CTrackProxy);
        CGuiRegistry::TReadView sub_view = registry.GetReadView(base_key);
        track->SetKey(sub_view.GetString(k_key));
        track->SetSubkey(sub_view.GetString(k_subkey, kEmptyStr));
        track->SetFilter(sub_view.GetString(k_filter, kEmptyStr));
        track->SetSortBy(sub_view.GetString(k_sortby, kEmptyStr));
        track->SetDisplayName(sub_view.GetString(k_display_name, kEmptyStr));
        track->SetOrder(sub_view.GetInt(k_order, -1));
        track->SetId(sub_view.GetString(k_id, ""));
        track->SetDB(sub_view.GetString(k_db, ""));
        track->SetCategory(sub_view.GetString(k_category, kEmptyStr));
        track->SetSubcategory(sub_view.GetString(k_subcategory, kEmptyStr));
        {
            string setting_g_str = sub_view.GetString(k_setting_group, kEmptyStr);
            vector<string> tokens;
            CSeqUtils::TokenizeWithEscape(setting_g_str, "|", tokens, true);
            if (tokens.size() > 0) {
                track->SetSettingGroup(NStr::TruncateSpaces(tokens[0]));
                if (tokens.size() > 1) {
                    try {
                        track->SetUseGroupDefault(NStr::StringToBool(NStr::TruncateSpaces(tokens[1])));
                    } catch (CException&) {
                        // ignore it
                    }
                }
            }
        }

        track->SetShown(sub_view.GetBool(k_shown));
        track->SetExpanded(sub_view.GetBool(k_expanded, true));
        track->SetTrackProfile(sub_view.GetString(k_track_profile));
        string annot_str = sub_view.GetString(k_annots, kEmptyStr);
        if (annot_str.empty()) {
            annot_str = sub_view.GetString(k_annot_separation, kEmptyStr);
        }
        CTrackProxy::TAnnots annots;
        NStr::Tokenize(annot_str, ",|", annots);
        track->SetAnnots(annots);
        ITERATE(CTrackProxy::TAnnots, a_iter, annots) {
            string track_key = track->GetKey() +
                track->GetSubkey() + track->GetFilter() + track->GetSortBy();
            m_UsedAnnots[track_key].insert(*a_iter);
        }
        track->SetName(iter->key);
        proxies.push_back(CRef<CTempTrackProxy>(track.GetPointer()));

        // load children temporary tracks
        x_LoadTempTrackRecursive(track->GetChildren(), base_key);
        // load children tracks
        LoadTrackSettingsRecursive(track->GetChildren(), base_key);

        // remove the feature subtype from all other feautres list
        // if it presents in a separate track.
        if (track->GetKey() == CFeatureTrackFactory::GetTypeInfo().GetId()) {
            CFeatListItem item;
            CSeqFeatData::GetFeatList()->GetItemByKey(track->GetSubkey(), item);
            TSubtypeList::iterator o_iter = find(m_OtherFeats.begin(),
                m_OtherFeats.end(), item.GetSubtype());
            if (o_iter != m_OtherFeats.end()) {
                m_OtherFeats.erase(o_iter);
            }
        }
    }

    proxies.sort(CTrackProxy::STPSorterByOrder::s_CompareCRefs);
}


void CTrackConfigManager::x_LoadTempTrackRecursive(
    TTrackProxies& proxies, const string& path)
{
    CGuiRegistry& registry = CGuiRegistry::GetInstance();
    string key = path + CGuiRegistry::kDecimalDot + "TempTracks";
    CGuiRegistry::TReadView view = registry.GetReadView(key);
    CGuiRegistry::TKeys keys;
    view.GetTopKeys(keys);
    ITERATE (CGuiRegistry::TKeys, iter, keys) {
        string sub_key = key + CGuiRegistry::kDecimalDot + iter->key;
        CGuiRegistry::TReadView sub_view = registry.GetReadView(sub_key);
        CRef<CTempTrackProxy> track(new CTempTrackProxy);
        track->SetOrder(sub_view.GetInt(k_order, -1));
        track->SetCategory(sub_view.GetString(k_category, kEmptyStr));
        track->SetSubcategory(sub_view.GetString(k_subcategory, kEmptyStr));
        {
            string setting_g_str = sub_view.GetString(k_setting_group, kEmptyStr);
            vector<string> tokens;
            CSeqUtils::TokenizeWithEscape(setting_g_str, "|", tokens, true);
            if (tokens.size() > 0) {
                track->SetSettingGroup(NStr::TruncateSpaces(tokens[0]));
                if (tokens.size() > 1) {
                    try {
                        track->SetUseGroupDefault(NStr::StringToBool(NStr::TruncateSpaces(tokens[1])));
                    } catch (CException&) {
                        // ignore it
                    }
                }
            }
        }

        track->SetShown(sub_view.GetBool(k_shown, true));
        track->SetExpanded(sub_view.GetBool(k_expanded, true));
        track->SetName(iter->key);
        track->SetDisplayName(sub_view.GetString(k_display_name, kEmptyStr));
        track->SetTrackProfile(sub_view.GetString(k_track_profile, kEmptyStr));
        proxies.push_back(track);
        x_LoadTempTrackRecursive(track->GetChildren(), sub_key);
    }
}


void CTrackConfigManager::SaveTrackSettingsRecursive(const TTrackProxies& proxies,
                                                     const string& path) const
{
    CGuiRegistry& registry = CGuiRegistry::GetInstance();
    TTrackProxies settings = proxies;
    settings.sort(CTrackProxy::STPSorterByOrder::s_CompareCRefs);
    int off = 0;
    int pre_id = -1;

    ITERATE (TTrackProxies, iter, settings) {
        const CTempTrackProxy* params = *iter;
        const CTrackProxy* full_params =
            dynamic_cast<const CTrackProxy*>(iter->GetPointer());

        // for those unrealized trackd, we ignore them
        if ( !params->IsRealized() ) {
            --off;
            continue;
        }

        string key = path + ".TempTracks";
        if (full_params) {
            key = path + ".Tracks";
        }

        string sub_key = key + CGuiRegistry::kDecimalDot + params->GetName();
        CGuiRegistry::TReadWriteView view = registry.GetReadWriteView(sub_key);
        if (pre_id == params->GetOrder()) {
            ++off;
        }

        view.Set(k_order, params->GetOrder() + off);
        view.Set(k_category, params->GetCategory());
        view.Set(k_subcategory, params->GetSubcategory());
        if ( !params->GetSettingGroup().empty() ) {
            view.Set(k_setting_group, params->GetSettingGroup());
        }
        view.Set(k_shown, params->GetShown());

        // save collapse state only, by default, a track is expanded.
        if ( !params->GetExpanded() ) {
            view.Set(k_expanded, params->GetExpanded());
        }
        if ( !params->GetDisplayName().empty() ) {
            view.Set(k_display_name, params->GetDisplayName());
        }

        if ( !params->GetTrackProfile().empty() ) {
            view.Set(k_track_profile, params->GetTrackProfile());

            // consider the case where there is only one subtrack so 
            // that the current track profile should actually be saved
            // to its child track
            if ( !params->GetSource().empty()  &&
                !params->GetTrackProfile().empty()) {
                    CConstRef<CTempTrackProxy> c_params =
                        params->GetChildTempTrack(params->GetSource());
                    if (c_params) {
                        CTempTrackProxy* t_params =
                            const_cast<CTempTrackProxy*>(c_params.GetPointer());
                        t_params->SetTrackProfile(params->GetTrackProfile());
                    }
            }
        }

        if (full_params) {
            view.Set(k_key, full_params->GetKey());
            if ( !full_params->GetSubkey().empty() ) {
                view.Set(k_subkey, full_params->GetSubkey());
            }
            if ( !full_params->GetFilter().empty() ) {
                view.Set(k_filter, full_params->GetFilter());
            }
            if ( !full_params->GetSortBy().empty() ) {
                view.Set(k_sortby, full_params->GetSortBy());
            }

            CTrackProxy::TAnnots annots = full_params->GetAnnots();
            CTrackProxy::TAnnots::iterator a_iter =
                find(annots.begin(), annots.end(),
                CSeqUtils::NameTypeValueToStr(CSeqUtils::eAnnot_All));
            if (a_iter != annots.end()) {
                annots.erase(a_iter);
            }
            if ( !annots.empty() ) {
                string annots_str = kEmptyStr;
                bool first_one = true;
                ITERATE (CTrackProxy::TAnnots, aa_iter, annots) {
                    if (first_one) {
                        annots_str = *aa_iter;
                        first_one = false;
                    } else {
                        annots_str += "|" + *aa_iter;
                    }
                }
                view.Set(k_annots, annots_str); 
            }
        }

        if ( !params->GetChildren().empty() ) {
            SaveTrackSettingsRecursive(params->GetChildren(), sub_key);
        }

        pre_id = params->GetOrder();
    }
}


void CTrackConfigManager::LoadProfile_Recursive(CLayoutTrack* track)
{
    track->LoadProfile(track->GetProfile());
    CTrackContainer* cont_track = dynamic_cast<CTrackContainer*>(track);
    if (cont_track) {
        NON_CONST_ITERATE (TTrackProxies, iter, cont_track->GetSubtrackProxies()) {
            if ((*iter)->GetTrack()) {
                LoadProfile_Recursive((*iter)->GetTrack());
                (*iter)->GetTrack()->SetComments((*iter)->GetComments());
                (*iter)->GetTrack()->SetHighlights((*iter)->GetHighlights());
            }
        }
    }
}


void CTrackConfigManager::SaveProfile_Recursive(CLayoutTrack* track) const
{
    CTrackContainer* cont_track = dynamic_cast<CTrackContainer*>(track);
    if (cont_track) {
        NON_CONST_ITERATE (TTrackProxies, iter, cont_track->GetSubtrackProxies()) {
            if ((*iter)->GetTrack()) {
                SaveProfile_Recursive((*iter)->GetTrack());
            }
        }
    }
    track->SaveProfile();
}


void CTrackConfigManager::x_InitTrackManager()
{
    typedef vector< CIRef<ILayoutTrackFactory> >  TTrackFactoryVec;
    TTrackFactoryVec factories;
    GetExtensionAsInterface(
        EXT_POINT__SEQGRAPHIC_LAYOUT_TRACK_FACTORY, factories);

    ITERATE (TTrackFactoryVec, iter, factories) {
        m_Factories[(*iter)->GetThisTypeInfo().GetId()] = *iter;
    }
}


void CTrackConfigManager::x_InitOtherFeatList()
{
    m_OtherFeats.clear();
    const CFeatList& feats(*CSeqFeatData::GetFeatList());
    ITERATE(CFeatList, iter, feats) {
        int subtype = iter->GetSubtype();
        int type = iter->GetType();
        if (subtype == CSeqFeatData::eSubtype_any  ||
            subtype == CSeqFeatData::eSubtype_variation  ||
            subtype == CSeqFeatData::eSubtype_variation_ref  ||
            CSGUtils::IsMainFeature(type, subtype) ) {
            continue;
        }
        m_OtherFeats.push_back(subtype);
    }
}


END_NCBI_SCOPE
