/*  $Id: track_container_track.cpp 28147 2013-05-16 18:05:09Z wuliangs $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Liangshou Wu
 *
 */

#include <ncbi_pch.hpp>
#include <gui/widgets/seq_graphic/track_container_track.hpp>
#include <gui/widgets/seq_graphic/layout_track_proxy.hpp>
#include <gui/widgets/seq_graphic/rendering_ctx.hpp>
#include <gui/widgets/wx/wx_utils.hpp>
#include <gui/objutils/reg_settings.hpp>
#include <gui/objutils/utils.hpp>
#include <wx/event.h>
#include <wx/menu.h>


BEGIN_NCBI_SCOPE

static const int kBaseID = 10000;
static const int kMaxID = 1000000;
static const size_t kMaxNameLen = 50;

CTrackTypeInfo
CTrackContainer::m_TypeInfo("track_container", "Track Container Track");

class CTrackContainerEvtHandler : public wxEvtHandler
{
public:
    CTrackContainerEvtHandler(CTrackContainer* track)
        : m_Track( track )
    {}

    void OnContextMenu( wxContextMenuEvent& anEvent );

private:
    /// @name event handlers.
    /// @{
    void x_OnSelectAll(wxCommandEvent& /*event*/)
    { m_Track->OnSelectAll(); }
    void x_OnDeselectAll(wxCommandEvent& /*event*/)
    { m_Track->OnDeselectAll(); }
    void x_OnRevertSelection(wxCommandEvent& /*event*/)
    { m_Track->OnRevertSelection(); }
    void x_OnToggleTrack(wxCommandEvent& event)
    {
        int track_order = event.GetId() - kBaseID;
        m_Track->OnToggleTrack(track_order);
    }
    /// @}

private:
    CTrackContainer* m_Track;

    DECLARE_EVENT_TABLE()
};


BEGIN_EVENT_TABLE(CTrackContainerEvtHandler, wxEvtHandler)
    EVT_MENU(kBaseID - 3, CTrackContainerEvtHandler::x_OnSelectAll)
    EVT_MENU(kBaseID - 2, CTrackContainerEvtHandler::x_OnDeselectAll)
    EVT_MENU(kBaseID - 1, CTrackContainerEvtHandler::x_OnRevertSelection)
    EVT_MENU_RANGE(kBaseID, kBaseID + kMaxID, CTrackContainerEvtHandler::x_OnToggleTrack)
END_EVENT_TABLE()


CTrackContainer::CTrackContainer(CRenderingContext* r_cntx, CFeaturePanelDS* ds)
    : CDataTrack(r_cntx)
    , m_DS(ds)
    , m_Handler(new CTrackContainerEvtHandler(this))
    , m_AnnotLevel(-1)
    , m_Adaptive(true)
    , m_TrackInitFinished(true)
    , m_SkipGenuineCheck(false)
{
    x_RegisterIcon(SIconInfo(eIcon_Content, "Content", true, "track_content"));
    if (m_DS) {
       m_DS->SetJobListener(this);
    }
}


CTrackContainer::~CTrackContainer()
{
    if (m_DS) {
        DeleteAllJobs();
    }
    delete m_Handler;
}


void CTrackContainer::GetHTMLActiveAreas(TAreaVector* p_areas) const
{
    // Get html active areas for child tracks
    ITERATE (TTrackProxies, iter, m_TrackProxies) {
        const CLayoutTrack* track = (*iter)->GetTrack();
        if (track  &&  track->IsOn()) {
            CHTMLActiveArea* area = track->InitHTMLActiveArea(p_areas);
            if (area) {
                string t_name = (*iter)->GetName();
                if (t_name.find("---") == 0) {
                    // temporary track
                    t_name = t_name.substr(3);
                }

                if (area->m_Signature.empty()) {
                    area->m_Signature = NStr::JsonEncode(t_name);
                } else {
                    // support multiple signatures including
                    // underlying object signature and track signature.
                    area->m_Signature = NStr::JsonEncode(t_name) + ";" + area->m_Signature;
                }
            }
        }
    }

    // Collect html active areas for non-track glyphs.
    GetGroup().GetHTMLActiveAreas(p_areas);
}


void CTrackContainer::DeleteAllJobs()
{
    if (m_DS) {
        m_DS->DeleteMetaDataJob();
    }
    CDataTrack::DeleteAllJobs();
    NON_CONST_ITERATE (TTrackProxies, iter, GetSubtrackProxies()) {
        if ((*iter)->GetTrack()) {
            CDataTrack* data_track =
                dynamic_cast<CDataTrack*>((*iter)->GetTrack());
            if (data_track) {
                data_track->DeleteAllJobs();
            }
        }
    }
    SetTrackInitDone(true);
}


bool CTrackContainer::AllJobsFinished() const
{
    bool done = IsTrackInitDone();
    if (done  &&  m_DS) {
        done = m_DS->AllJobsFinished();
    }

    TTrackProxies::const_iterator iter = GetSubtrackProxies().begin();
    while (done  &&  iter != GetSubtrackProxies().end()) {
        if ((*iter)->GetTrack()  &&  (*iter)->GetShown()) {
            const CDataTrack* data_track =
                dynamic_cast<const CDataTrack*>((*iter)->GetTrack());
            if (data_track) {
                done = data_track->AllJobsFinished();
            } else {
                const CTrackContainer* track_cont =
                    dynamic_cast<const CTrackContainer*>((*iter)->GetTrack());
                if (track_cont) {
                    done = track_cont->AllJobsFinished();
                }
            }
        }
        ++iter;
    }
    return done;
}


/// help functor for sorting CSeqGlyphs (known as CLayoutTrack) by order.
static bool s_TrackByOrder(const CRef<CSeqGlyph>& t1,
                        const CRef<CSeqGlyph>& t2)
{
    return dynamic_cast<const CLayoutTrack&>(*t1).GetOrder() <
        dynamic_cast<const CLayoutTrack&>(*t2).GetOrder();
}


void CTrackContainer::AddTrackProxy(CTempTrackProxy* proxy)
{
    _ASSERT(proxy->GetOrder() >= 0);

    // find out where we need to insert into the complete track list
    TTrackProxies::iterator iter = m_TrackProxies.begin();
    while (iter != m_TrackProxies.end()  &&  (*iter)->GetOrder() != proxy->GetOrder()) {
        ++iter;
    }
    if (iter == m_TrackProxies.end()) {
        m_TrackProxies.push_back(CRef<CTempTrackProxy>(proxy));
        m_TrackProxies.sort(CTrackProxy::STPSorterByOrder::s_CompareCRefs);
    }

    // add the track into the visible group if shown.
    CLayoutTrack* track = proxy->GetTrack();
    if (track) {
        // initialize the track
        track->SetHost(m_LTHost);
        track->SetConfig(m_gConfig);
        track->SetIndent(m_IndentLevel + 1);
        track->SetParentTrack(this);
        if (track->IsOn()) {
            Add(track);
            SortChildTracksByOrder();
        }
    }

    x_UpdateMsg();
}


CTrackProxy* CTrackContainer::AddNewTrack(const string& track_key,
                                          const CTrackProxy::TAnnots& annots,
                                          const string& disp_name,
                                          const string& name,
                                          const string& filter,
                                          const string& sub_key)
{
    int order = 0;
    if ( !m_TrackProxies.empty() ) {
        order = m_TrackProxies.back()->GetOrder() + 1;
    }
    string track_name = name;
    if (track_name.empty()) {
        track_name = track_key;
        ITERATE (CTrackProxy::TAnnots, iter, annots) {
            track_name += NStr::Replace(*iter, ".", "_");
        }
        int post_fix = 0;
        while (IsTrackNameUsed(track_name + NStr::IntToString(post_fix))) {
            ++post_fix;
        }
        track_name += NStr::IntToString(post_fix);
    }

    CTrackProxy* track = new CTrackProxy(order, track_name, true, track_key, sub_key);
    CRef<CTempTrackProxy> proxy(track);
    track->SetAnnots(annots);
    track->SetFilter(filter);
    track->SetDisplayName(disp_name);
    m_TrackProxies.push_back(proxy);
    x_ReinitTrack(track);
    return track;
}

static string s_GetBestTrackTitle(const CTempTrackProxy* proxy,
                                  const CLayoutTrack* track)
{
    if (proxy  &&  !proxy->GetDisplayName().empty() ) {
        return proxy->GetDisplayName();
    }
    if (track  &&  !track->GetTitle().empty() ) {
        return track->GetTitle();
    }

    if (proxy) {
        return proxy->GetName();
    } 

    return "";
}

void CTrackContainer::AddTracks(CTempTrackProxy* t_proxy, TTrackMap& tracks)
{
    CRef<CLayoutTrack> original_track(t_proxy->GetTrack());
    if (original_track) {
        CTrackContainer* con_track =
            dynamic_cast<CTrackContainer*>(original_track.GetPointer());
        if (con_track) { // we already have a container track to hold the subtracks
            NON_CONST_ITERATE (TTrackMap, iter, tracks) {
                CLayoutTrack* track = iter->second;

                CRef<CTempTrackProxy> proxy =
                    t_proxy->GetChildTempTrack(iter->first);
                string annot_name = ILayoutTrackFactory::ExtractAnnotName(iter->first);
                string filter_name = ILayoutTrackFactory::ExtractFilterName(iter->first);
                if ( !filter_name.empty() ) {
                    proxy->SetFilter(filter_name);
                }

                if (GetConfigMgr()) {
                    GetConfigMgr()->NeedToShow(t_proxy, proxy, annot_name);
                }

                if ( !proxy->GetTrack() ) {
                    if ( !proxy->GetDisplayName().empty() ) {
                        track->SetTitle(proxy->GetDisplayName());
                    }
                    x_AddSubtrack(con_track, track, proxy);
                    track->Update(false);
                }
            }
        } else { // there was only one track originally
            string data_source =
                ILayoutTrackFactory::MergeAnnotAndFilter(
                    t_proxy->GetSource(), 
                    t_proxy->GetFilter());

            _ASSERT( !data_source.empty() );
            TTrackMap::iterator t_iter = tracks.find(data_source);
            if (t_iter != tracks.end()) {
                original_track->SetTitle(t_iter->second->GetTitle());
                tracks.erase(t_iter);
            }

            if ( !tracks.empty() ) { // now there are at least two subtracks
                // first, remove the original track from its parent.
                bool was_on = original_track->IsOn();
                RemoveTrack(original_track->GetOrder());
                if (was_on) t_proxy->SetShown(true);

                CRef<CTempTrackProxy> proxy =
                    t_proxy->GetChildTempTrack(data_source);
                // copy over the potential modified profile
                proxy->SetTrackProfile(t_proxy->GetTrackProfile());

                // we will always show the original track no matter what
                proxy->SetShown(true);

                // second, create a track container for holding the subtracks.
                CRef<CTrackContainer> cont(new CTrackContainer(m_Context));
                cont->SetAnnotLevel(GetAnnotLevel());
                cont->SetTitle(s_GetBestTrackTitle(t_proxy, NULL));
                x_AddSubtrack(this, cont, t_proxy);

                // third, add the only original track to the track container.
                if ( !proxy->GetDisplayName().empty() ) {
                    original_track->SetTitle(proxy->GetDisplayName());
                }
                x_AddSubtrack(cont, original_track, proxy);

                // last add the newly created subtrack.
                NON_CONST_ITERATE (TTrackMap, iter, tracks) {
                    CLayoutTrack* sub_track = iter->second;
                    CRef<CTempTrackProxy> c_proxy =
                        t_proxy->GetChildTempTrack(iter->first);
                    string annot_name = ILayoutTrackFactory::ExtractAnnotName(iter->first);
                    string filter_name = ILayoutTrackFactory::ExtractFilterName(iter->first);

                    if (c_proxy->GetDisplayName().empty()  &&  !sub_track->GetTitle().empty()) {
                        c_proxy->SetDisplayName(sub_track->GetTitle());
                    }                    

                    if ( !filter_name.empty() ) {
                        c_proxy->SetFilter(filter_name);
                    }
                    if (GetConfigMgr()) {
                        GetConfigMgr()->NeedToShow(t_proxy, c_proxy, annot_name);
                    }

                    if ( !c_proxy->GetDisplayName().empty() ) {
                        sub_track->SetTitle(c_proxy->GetDisplayName());
                    }
                    x_AddSubtrack(cont, sub_track, c_proxy);
                    sub_track->Update(false);
                }
            }
        }
    } else {
        if (tracks.size() > 1) { // we need a container track
            CRef<CTrackContainer> cont(new CTrackContainer(m_Context));
            cont->SetAnnotLevel(GetAnnotLevel());
            cont->SetTitle(s_GetBestTrackTitle(t_proxy, NULL));
            x_AddSubtrack(this, cont, t_proxy);
            NON_CONST_ITERATE (TTrackMap, iter, tracks) {
                CLayoutTrack* track = iter->second;
                string annot_name = ILayoutTrackFactory::ExtractAnnotName(iter->first);
                string filter_name = ILayoutTrackFactory::ExtractFilterName(iter->first);

                CRef<CTempTrackProxy> proxy = t_proxy->GetChildTempTrack(iter->first);
                if (proxy->GetDisplayName().empty()  &&  !track->GetTitle().empty()) {
                    proxy->SetDisplayName(track->GetTitle());
                }                    

                proxy->SetSource(annot_name);
                if ( !filter_name.empty() ) {
                    proxy->SetFilter(filter_name);
                }

                if (GetConfigMgr()) {
                    GetConfigMgr()->NeedToShow(t_proxy, proxy, annot_name);
                }

                if ( !proxy->GetDisplayName().empty() ) {
                    track->SetTitle(proxy->GetDisplayName());
                }
                x_AddSubtrack(cont, track, proxy);
            }
            cont->Update(false);
        } else {
            CLayoutTrack* track = tracks.begin()->second;
            track->SetTitle(s_GetBestTrackTitle(t_proxy, track));

            string annot_name = ILayoutTrackFactory::ExtractAnnotName(tracks.begin()->first);
            t_proxy->SetSource(annot_name);

            string filter_name = ILayoutTrackFactory::ExtractFilterName(tracks.begin()->first);
            if ( !filter_name.empty() ) {
                t_proxy->SetFilter(filter_name);
            }
            // make sure we use the correct profile
            // it is possible we are looking at a range within which 
            // there is only one track.  There might be more than one
            // tracks when zoomming out.  So we are using the parent
            // tracks's profile, not the track's own profile.
            CRef<CTempTrackProxy> proxy = t_proxy->GetChildTempTrack(tracks.begin()->first);
            t_proxy->SetTrackProfile(proxy->GetTrackProfile());
            x_AddSubtrack(this, track, t_proxy);
            track->Update(false);
        }
    }

    m_ConfigMgr->AdjustTrackOrder(t_proxy);
}


void CTrackContainer::SetTrack(CLayoutTrack* track, int order, bool shown)
{
    // initialize the track
    track->SetHost(m_LTHost);
    track->SetConfig(m_gConfig);
    track->SetIndent(m_IndentLevel + 1);
    track->SetOrder(order);
    track->SetParentTrack(this);

    // find out where we need to insert into the complete track list
    TTrackProxies::iterator iter = m_TrackProxies.begin();
    while (iter != m_TrackProxies.end()  &&  (*iter)->GetOrder() != order) {
        ++iter;
    }

    if (iter != m_TrackProxies.end()) {
        if ((*iter)->GetTrack()) {
            SetGroup().Remove((*iter)->GetTrack());
        }
        (*iter)->SetTrack(track);
    } else if (track->IsMovable()) {
        /// we shouldn't do this
        _ASSERT(false);
        return;
    }

    // add the track into the visible group if shown.
    track->SetShow(shown);
    if (shown) {
        Add(track);
        SortChildTracksByOrder();
    }
    
    x_UpdateMsg();
}


void CTrackContainer::CloneTrack(int order, const string& track_title)
{
    // find the source track that we want to duplicate
    TTrackProxies::iterator iter = m_TrackProxies.begin();
    while (iter != m_TrackProxies.end()  &&  (*iter)->GetOrder() != order) {
        ++iter;
    }
    if (iter == m_TrackProxies.end()) return;
    CRef<CTempTrackProxy> src_proxy = *iter;

    // depending on if the souce track is explicitly specified in
    // configure file or it is just a derived sub-track of other such
    // track, we will insert the new track at a different place.
    CTrackContainer* p_track = this;
    if (src_proxy->IsTempTrack()) {
        // So we need to find out where it will be added and who its
        // parent/grandparent is.
        while (p_track->GetParentTrack()) {
            order = p_track->GetOrder();
            p_track = dynamic_cast<CTrackContainer*>(p_track->GetParentTrack());
            _ASSERT(p_track);
            TTrackProxies& proxies = p_track->GetSubtrackProxies();
            iter = proxies.begin();
            while (iter != proxies.end()  &&  (*iter)->GetOrder() != order) {
                ++iter;
            }
            if (iter == proxies.end()) return;

            if ( !(*iter)->IsTempTrack() )  {
                break;
            }
        }
    }

    // generate a unique track name for the new track
    string track_name = src_proxy->GetName() + " clone";
    int post_fix = 0;
    while (IsTrackNameUsed(track_name + NStr::IntToString(post_fix))) {
        ++post_fix;
    }
    track_name += NStr::IntToString(post_fix);

    // create proxy for the new track
    CRef<CTempTrackProxy> new_proxy = (*iter)->Clone();
    CTrackProxy* full_proxy = dynamic_cast<CTrackProxy*>(new_proxy.GetPointer());

    if (src_proxy->IsTempTrack()) {
        // For derived track, track key(, subkey), and annotation
        // are not available in src_proxy. We need to extract them
        // from somewhere else.
        const CLayoutTrack* src_track = src_proxy->GetTrack();
        _ASSERT(src_track);
        string track_key = src_track->GetTypeInfo().GetId();
        full_proxy->SetKey(track_key);
        CTrackProxy::TAnnots annots;
        annots.push_back(src_proxy->GetSource());
        full_proxy->SetAnnots(annots);
        const ILayoutTrackFactory* factory =
            p_track->GetConfigMgr()->GetTrackFactory(track_key);
        factory->CloneTrack(src_track, full_proxy);
    }

    // initialize other info
    new_proxy->SetName(track_name);
    new_proxy->SetDisplayName(track_title);
    // expand the cloned track automatically
    new_proxy->SetExpanded(true);

    new_proxy->SetOrder(++order);
    ++iter;

    // update orders to make sure the newly added track appear right below
    // the source track.
    iter = p_track->GetSubtrackProxies().insert(iter, new_proxy);
    ++order;
    while (iter != p_track->GetSubtrackProxies().end()  &&  order > (*iter)->GetOrder()) {
        (*iter)->SetOrder(order);
        ++order;
        ++iter;
    }

    p_track->x_ReinitTrack(full_proxy);
}


void CTrackContainer::RenameChildTrack(int order, const string& new_name)
{
    TTrackProxies::iterator iter = m_TrackProxies.begin();
    while (iter != m_TrackProxies.end()  &&  (*iter)->GetOrder() != order) {
        ++iter;
    }
    if (iter != m_TrackProxies.end()) {
        (*iter)->SetDisplayName(new_name);
        (*iter)->GetTrack()->SetTitle(new_name);
    }
}


void CTrackContainer::RemoveTrack(int order, bool immediately)
{
    TTrackProxies::iterator iter = m_TrackProxies.begin();
    while (iter != m_TrackProxies.end()  &&  (*iter)->GetOrder() != order) {
        ++iter;
    }
    if (iter != m_TrackProxies.end()) {
        CDataTrack* data_track = dynamic_cast<CDataTrack*>((*iter)->GetTrack());
        if (data_track) {
            data_track->DeleteAllJobs();
        }
        (*iter)->GetTrack()->SaveProfile();
        SetGroup().Remove((*iter)->GetTrack());
        (*iter)->SetShown(false);
        if (immediately) {
            m_TrackProxies.erase(iter);
        } else {
            // mark it as empty track in order to remove it
            // completely from GUI registry.
            (*iter)->SetEmpty(true);
            x_UpdateLayout();
        }
    }
}


void CTrackContainer::CloseTrack(int order)
{
    TTrackProxies::iterator iter = m_TrackProxies.begin();
    while (iter != m_TrackProxies.end()  &&  (*iter)->GetOrder() != order) {
        ++iter;
    }
    if (iter != m_TrackProxies.end()) {
        CDataTrack* data_track = dynamic_cast<CDataTrack*>((*iter)->GetTrack());
        if (data_track) {
            data_track->DeleteAllJobs();
        }
        SetGroup().Remove((*iter)->GetTrack());
        (*iter)->SetShown(false);
        x_UpdateMsg();
        SetGroup().UpdateLayout();
    }
}


void CTrackContainer::MoveUp(int order)
{
    // get the track order for the track above the moved track
    CLayoutGroup::TObjectList::iterator iter = SetChildren().begin();
    CLayoutGroup::TObjectList::iterator p_iter = iter;
    while (iter != SetChildren().end()  &&
        dynamic_cast<const CLayoutTrack&>(**iter).GetOrder() != order) {
        p_iter = iter;
        ++iter;
    }
    if (iter == SetChildren().end()  ||  iter == p_iter) return;
    int pre_order = dynamic_cast<const CLayoutTrack&>(**p_iter).GetOrder();

    // get track proxy for the moved track (with track_order = order)
    TTrackProxies::iterator proxy_iter = m_TrackProxies.begin();
    while (proxy_iter != m_TrackProxies.end()  &&
        (*proxy_iter)->GetOrder() != order) {
        ++proxy_iter;
    }

    // get track proxy for the previous track  (with track_order = pre_order)
    TTrackProxies::iterator pre_proxy_iter = m_TrackProxies.begin();
    while (pre_proxy_iter != m_TrackProxies.end()  &&
        (*pre_proxy_iter)->GetOrder() != pre_order) {
        ++pre_proxy_iter;
    }

    _ASSERT(proxy_iter != m_TrackProxies.end()  &&
        pre_proxy_iter != m_TrackProxies.end());

    // swap track orders
    (*pre_proxy_iter)->SetOrder(order);
    (*proxy_iter)->SetOrder(pre_order);

    // sort both track proxies and tracks
    m_TrackProxies.sort(CTrackProxy::STPSorterByOrder::s_CompareCRefs);
    SortChildTracksByOrder();
}


void CTrackContainer::MoveDown(int order)
{
    // get the track order for the track below the moved track
    CLayoutGroup::TObjectList::reverse_iterator iter = SetChildren().rbegin();
    CLayoutGroup::TObjectList::reverse_iterator n_iter = iter;
    while (iter != SetChildren().rend()  &&
        dynamic_cast<const CLayoutTrack&>(**iter).GetOrder() != order) {
        n_iter = iter;
        ++iter;
    }
    if (iter == SetChildren().rend()  ||  iter == n_iter) return;
    int n_order = dynamic_cast<const CLayoutTrack&>(**n_iter).GetOrder();

    // get track proxy for the moved track (with track_order = order)
    TTrackProxies::reverse_iterator proxy_iter = m_TrackProxies.rbegin();
    while (proxy_iter != m_TrackProxies.rend()  &&
        (*proxy_iter)->GetOrder() != order) {
        ++proxy_iter;
    }

    // get track proxy for the next track  (with track_order = pre_order)
    TTrackProxies::reverse_iterator n_proxy_iter = m_TrackProxies.rbegin();
    while (n_proxy_iter != m_TrackProxies.rend()  &&
        (*n_proxy_iter)->GetOrder() != n_order) {
        ++n_proxy_iter;
    }

    _ASSERT(proxy_iter != m_TrackProxies.rend()  &&
        n_proxy_iter != m_TrackProxies.rend());

    // swap track orders
    (*n_proxy_iter)->SetOrder(order);
    (*proxy_iter)->SetOrder(n_order);

    // sort both track proxies and tracks
    m_TrackProxies.sort(CTrackProxy::STPSorterByOrder::s_CompareCRefs);
    SortChildTracksByOrder();
}


void CTrackContainer::MoveTrackToTop(int order)
{
    // get track proxy for the moved track (with track_order = order)
    TTrackProxies::reverse_iterator proxy_iter = m_TrackProxies.rbegin();
    while (proxy_iter != m_TrackProxies.rend()  &&
        (*proxy_iter)->GetOrder() != order) {
        ++proxy_iter;
    }

    TTrackProxies::reverse_iterator next_track(proxy_iter);
    if (proxy_iter != m_TrackProxies.rend()) ++next_track;
    while (next_track != m_TrackProxies.rend()) {
        (*proxy_iter)->SetOrder((*next_track)->GetOrder());
        (*next_track)->SetOrder(order);
        order = (*proxy_iter)->GetOrder();
        ++next_track;
    }
    // sort both track proxies and tracks
    m_TrackProxies.sort(CTrackProxy::STPSorterByOrder::s_CompareCRefs);
    SortChildTracksByOrder();
    x_UpdateLayout();
}


void CTrackContainer::MoveTrackToBottom(int order)
{
    // get track proxy for the moved track (with track_order = order)
    TTrackProxies::iterator proxy_iter = m_TrackProxies.begin();
    while (proxy_iter != m_TrackProxies.end()  &&
        (*proxy_iter)->GetOrder() != order) {
        ++proxy_iter;
    }

    TTrackProxies::iterator next_track(proxy_iter);
    if (proxy_iter != m_TrackProxies.end()) ++next_track;
    while (next_track != m_TrackProxies.end()) {
        (*proxy_iter)->SetOrder((*next_track)->GetOrder());
        (*next_track)->SetOrder(order);
        order = (*proxy_iter)->GetOrder();
        ++next_track;
    }
    // sort both track proxies and tracks
    m_TrackProxies.sort(CTrackProxy::STPSorterByOrder::s_CompareCRefs);
    SortChildTracksByOrder();
    x_UpdateLayout();
}


bool CTrackContainer::HasSubtracks() const
{
    bool has_any = !GetChildren().empty();
    TTrackProxies::const_iterator iter = m_TrackProxies.begin();
    while(!has_any  &&  iter != m_TrackProxies.end()) {
        has_any = ((*iter)->GetTrack() != NULL);
        ++iter;
    }
    return has_any;
}


bool CTrackContainer::NoSubtrackEver() const
{
    bool no_track = GetChildren().empty();
    CRef<CSGSequenceDS> seq_ds = m_Context->GetSeqDS();
    int seq_len = seq_ds->GetSequenceLength();
    int off = (int)(seq_len * 0.0001);
    TSeqRange total_range(off, seq_len - off);
    TTrackProxies::const_iterator iter = m_TrackProxies.begin();
    while(no_track  &&  iter != m_TrackProxies.end()) {
        no_track =
            (*iter)->IsEmpty()  ||
            ((*iter)->GetTrack() == NULL  &&
            dynamic_cast<const CTrackProxy*>(iter->GetPointer())  &&
            (*iter)->GetVisitedRange().IntersectionWith(total_range) == total_range);
        ++iter;
    }
    return no_track;
}


void CTrackContainer::SaveTrackStates(int order)
{
    TTrackProxies::iterator iter = m_TrackProxies.begin();
    while (iter != m_TrackProxies.end()  &&  (*iter)->GetOrder() != order) {
        ++iter;
    }
    if (iter != m_TrackProxies.end()) {
        const CLayoutTrack* track = (*iter)->GetTrack();
        if (track) {
            (*iter)->SetExpanded(track->IsExpanded());
            (*iter)->SetShown(track->IsOn());
        }
    }
}


void CTrackContainer::ClearTracks()
{
    m_Height = 0;
    // canceling all unfinished jobs if any
    DeleteAllJobs();
    SetGroup().Clear();
    m_TrackProxies.clear();
}


bool CTrackContainer::IsTrackNameUsed(const string& name) const
{
    ITERATE (TTrackProxies, iter, m_TrackProxies) {
        if ((*iter)->GetName() == name) {
            return true;
        }
    }
    return false;
}


CTrackContainer*
CTrackContainer::GetChildContainerTrack(const string& name)
{
    NON_CONST_ITERATE (TTrackProxies, iter, m_TrackProxies) {
        CTrackContainer* track =
            dynamic_cast<CTrackContainer*>(iter->GetPointer());
        if (track  &&  track->GetTitle() == name) {
            return track;
        }
    }
    return NULL;
}


const CTrackTypeInfo& CTrackContainer::GetTypeInfo() const
{
    return m_TypeInfo;
}


CRef<CSGGenBankDS>  CTrackContainer::GetDataSource()
{
    return CRef<CSGGenBankDS>(m_DS.GetPointer());
}


CConstRef<CSGGenBankDS>  CTrackContainer::GetDataSource() const
{
    return CConstRef<CSGGenBankDS>(m_DS.GetPointer());
}


void CTrackContainer::ExpandAll(bool expand)
{
    NON_CONST_ITERATE (TTrackProxies, iter, m_TrackProxies) {
        CLayoutTrack* track = (*iter)->GetTrack();
        if (track) {
            CTrackContainer* cont = dynamic_cast<CTrackContainer*>(track);
            if (cont) {
                cont->ExpandAll(expand);
            }
            track->Expand(expand);
        }
    }
}


void CTrackContainer::ShowAll(bool shown, bool recursive)
{
    NON_CONST_ITERATE (TTrackProxies, iter, m_TrackProxies) {
        CTempTrackProxy* proxy = *iter;
        CLayoutTrack* track = proxy->GetTrack();
        if (track) {
            if (shown != proxy->GetShown()) {
                x_ShowTrack(track, shown);
            }
            if (recursive) {
                // If the track is a track container track, do the same
                // for its children.
                CTrackContainer* c_track =
                    dynamic_cast<CTrackContainer*>(iter->GetPointer());
                if (c_track) {
                    c_track->ShowAll(shown, recursive);
                }
            }
        } else {
            CTrackProxy* t_proxy = dynamic_cast<CTrackProxy*>(proxy);
            if (t_proxy  &&  shown != t_proxy->GetShown()) {
                t_proxy->SetShown(shown);
                if (shown) x_ReinitTrack(t_proxy);
            }
            if (recursive) {
                NON_CONST_ITERATE(TTrackProxies, t_iter, proxy->GetChildren()) {
                    t_proxy->SetShown(shown);
                }
            }
        }
    }
}


void CTrackContainer::SortChildTracksByOrder()
{
    SetChildren().sort(s_TrackByOrder);
}


void CTrackContainer::RemoveEmptyTracks()
{
    TTrackProxies::iterator iter = m_TrackProxies.begin();
    CRef<CSGSequenceDS> seq_ds = m_Context->GetSeqDS();
    int seq_len = seq_ds->GetSequenceLength();

    // when users 
    int off = min((int)(seq_len * 0.0001), 100);
    TSeqRange total_range(off, seq_len - off);
    while (iter != m_TrackProxies.end()) {
        CTempTrackProxy* proxy = *iter;
        // HACK, HACK
        // we don't want to remove empty track_container
        // if it is not temporary track.
        const CTrackProxy* full_proxy =
            dynamic_cast<const CTrackProxy*>(proxy);
        if (full_proxy  &&  full_proxy->GetKey() ==
            CTrackContainerFactory::GetTypeInfo().GetId()) {
                ++iter;
                continue;
        }

        CLayoutTrack* track = proxy->GetTrack();

        // TODO: need to figure out a way to check if a track is 
        // indeed an empty track. Currently, we only remove the
        // empty track containers.
        CTrackContainer* c_track = dynamic_cast<CTrackContainer*>(track);
        if (c_track) {
            if (c_track->NoSubtrackEver()) {
                if (c_track->IsOn()) {
                    SetGroup().Remove(c_track);
                    SetGroup().UpdateBoundingBox();
                    proxy->SetShown(false);
                }
                // don't remove the track now.
                // this function might be called by the the
                // empty track.
                // Note: don't remove the track if a track is not a 
                // temporary track.
                proxy->SetEmpty(!full_proxy);
            }
            ++iter;
        } else if (track == NULL  &&  full_proxy  &&
            proxy->GetVisitedRange().IntersectionWith(total_range) == total_range) {
                iter = m_TrackProxies.erase(iter);
        } else {
            ++iter;
        }
    }

    if (NoSubtrackEver()) {
        CTrackContainer* parent =
            dynamic_cast<CTrackContainer*>(GetParentTrack());
        if (parent) {
            parent->RemoveEmptyTracks();
        }
    }
}


void CTrackContainer::OnRevertSelection()
{
    TTrackProxies::iterator t_iter = m_TrackProxies.begin();
    while (t_iter != m_TrackProxies.end()) {
        CRef<CTempTrackProxy> proxy = *t_iter;
        if (proxy->GetShown()) {
            if (proxy->GetTrack()) {
                x_ShowTrack(proxy->GetTrack(), false);
            } else {
                proxy->SetShown(false);
            }
        } else {
            if (proxy->GetTrack()) {
                x_ShowTrack(proxy->GetTrack(), true);
            } else {
                CTrackProxy* t_proxy =
                    dynamic_cast<CTrackProxy*>(proxy.GetPointer());
                if (t_proxy) {
                    t_proxy->SetShown(true);
                    x_ReinitTrack(t_proxy);
                }
            }
        }
        ++t_iter;
    }
}


void CTrackContainer::OnToggleTrack(int order)
{
    TTrackProxies::iterator t_iter = m_TrackProxies.begin();
    while (t_iter != m_TrackProxies.end()) {
        CRef<CTempTrackProxy> proxy = *t_iter;
        if (proxy->GetOrder() == order) {
            if (proxy->GetTrack()) {
                x_ShowTrack(proxy->GetTrack(), !proxy->GetShown());
            } else {
                CTrackProxy* t_proxy = dynamic_cast<CTrackProxy*>(proxy.GetPointer());
                if (t_proxy) {
                    t_proxy->SetShown(!proxy->GetShown());
                    if (proxy->GetShown()) {
                        x_ReinitTrack(t_proxy);
                    }
                }
            }
            break;
        }
        ++t_iter;
    }
}


bool CTrackContainer::x_Empty() const
{
    ITERATE (TTrackProxies, iter, m_TrackProxies) {
        if ((*iter)->GetTrack() != NULL  &&  (*iter)->GetTrack()->IsOn()) {
            return false;
        }
    }
    return true;
}


void CTrackContainer::x_UpdateData()
{
    if (m_DS) {
        if ( !m_DS->AllJobsFinished() ) {
            m_DS->DeleteAllJobs();
            SetTrackInitDone(true);
        }
        x_SetStartStatus();
    }

    // make sure we keep the track info up to date
    // E.g. the track titles get changed to more meaningful ones
    // when actual data get loaded.
    x_UpdateMsg();
    CGlyphContainer::Update(false);

    // check if we need to reinitialize the sub tracks.
    if (GetConfigMgr()) {
        x_ReinitSubtracks();
    }
}


void CTrackContainer::x_OnJobCompleted(CAppJobNotification& notify)
{
    m_DS->ClearJobID(notify.GetJobID());
    CRef<CObject> res_obj = notify.GetResult();
    CTrackInitJobResult* result = dynamic_cast<CTrackInitJobResult*>(&*res_obj);
    if (!result) {
        LOG_POST(Error << "CFeaturePanel::x_OnJobCompleted() notification \
                          for job does not contain results.");
        return;
    }

    TTrackMap& tracks = result->m_Tracks;
    //CTrackContainer* parent = result.m_ParentTrack;
    CTempTrackProxy*  t_proxy = result->m_TrackParams;

    // remove the tracks with annotation name used by other tracks
    CTrackProxy* full_track = dynamic_cast<CTrackProxy*>(t_proxy);
    TTrackMap::iterator track_iter = tracks.begin();
    while (full_track  &&  track_iter != tracks.end()) {
        string annot_name = ILayoutTrackFactory::ExtractAnnotName(track_iter->first);
        const CTrackProxy::TAnnots& annots = full_track->GetAnnots();
        if (find(annots.begin(), annots.end(), annot_name) != annots.end()) {
            ++track_iter;
            continue;
        }

        string track_key = full_track->GetKey() +
            full_track->GetSubkey() + full_track->GetFilter();
        if (m_ConfigMgr->IsAnnotUsed(track_key, annot_name)) {
            tracks.erase(track_iter++);
        } else {
            ++track_iter;
        }
    }

    if ( !tracks.empty() ) {
        if (CTrackContainer* c_track =
            dynamic_cast<CTrackContainer*>(tracks.begin()->second.GetPointer())) {
            c_track->SetConfigMgr(m_ConfigMgr);
            c_track->SetSubtrackProxies(t_proxy->GetChildren());
        }
    }        

    // update the visited range by including the previous visited range
    // if both intersect to each other
    TSeqRange vis_range = m_Context->GetVisSeqRange();
    if (vis_range.IntersectingWith(t_proxy->GetVisitedRange())) {
        vis_range.CombineWith(t_proxy->GetVisitedRange());
    }
    t_proxy->SetVisitedRange(vis_range);
    t_proxy->SetRealized(true);

    bool need_update = false;
    if ( !tracks.empty() ) {
        need_update = true;
        AddTracks(t_proxy, tracks);
        SetMsg(", Initializing subtracks..., Finish initializing "
            + t_proxy->GetName());
    }

    if (!m_DS->AllJobsFinished()  &&  need_update) {
        x_UpdateLayout();
    }
}


void CTrackContainer::x_OnIconClicked(TIconID id)
{
    if (id == eIcon_Content) {
        bool all_on = true;
        bool all_off = true;
        TTrackProxies::const_iterator iter = m_TrackProxies.begin();
        while ((all_on  ||  all_off)  &&  iter != m_TrackProxies.end()) {
            if ((*iter)->GetShown()) {
                all_off = false;
            } else {
                all_on = false;
            }
            ++iter;
        }

        wxMenu menu;
        wxMenuItem* item = menu.Append(kBaseID - 3, wxT("Select all"));
        if (all_on) item->Enable(false);
        item = menu.Append(kBaseID - 2, wxT("Deselect all"));
        if (all_off) item->Enable(false);
        menu.Append(kBaseID - 1, wxT("Revert selections"));
        menu.AppendSeparator();

        int track_order = 0;
        iter = m_TrackProxies.begin();
        while (iter != m_TrackProxies.end()) {
            const CTempTrackProxy* proxy = *iter;

            // we will hide the tracks which are either empty
            // (no data in the whole renage), or temporary and 
            // no data so far within the visited ranges.
            if (proxy->IsEmpty()  ||
                ( !dynamic_cast<const CTrackProxy*>(proxy)  &&
                !proxy->GetTrack() )) {
                    ++iter;
                    continue;
            }

            string item_name = proxy->GetDisplayName();
            if (item_name.empty()) {
                item_name = proxy->GetName();
            }
            string help_text = item_name;
            if (proxy->GetTrack()) {
                if (proxy->GetShown()  &&  !proxy->GetTrack()->GetFullTitle().empty() ) {
                    help_text = proxy->GetTrack()->GetFullTitle();
                    if (help_text.length() < kMaxNameLen) {
                        item_name = help_text;
                    } else {
                        item_name = help_text.substr(0, kMaxNameLen);
                        item_name += "...";
                    }
                }
                const CTrackContainer* cont =
                    dynamic_cast<const CTrackContainer*>(proxy->GetTrack());
                if (cont  &&  !cont->HasSubtracks()) {
                    if (cont->IsOn()) {
                        item_name += " (no data found so far)";
                    } else {
                        item_name += " (potential)";
                    }
                }
            } else if (proxy->GetShown()) {
                item_name += " (no data found so far)";
            } else {
                item_name += " (potential)";
            }

            track_order = proxy->GetOrder();
            item = menu.AppendCheckItem(
                kBaseID + track_order, ToWxString(item_name), ToWxString(help_text));
            if (proxy->GetShown()) {
                item->Check();
            }
            if (proxy->GetTrack()  && !proxy->GetTrack()->IsCollapsible() ) {
                item->Enable(false);
            }
            ++iter;
        }

        m_LTHost->LTH_PushEventHandler(m_Handler);
        m_LTHost->LTH_PopupMenu(&menu);
        m_LTHost->LTH_PopEventHandler();

    } else {
        // use default handlers
        CLayoutTrack::x_OnIconClicked(id);
    }
    x_OnLayoutChanged();
}


void CTrackContainer::x_OnAllJobsFinished()
{
    SetTrackInitDone(true);
    ITrackInitObserver* observer =
        dynamic_cast<ITrackInitObserver*>(GetParentTrack());
    if (observer) {
        observer->OnTrackInitFinished(this);
    }
    if (CSGDataSource::IsBackgroundJob()) {
        RemoveEmptyTracks();
        x_UpdateMsg();
    }
    x_UpdateLayout();
}


void CTrackContainer::x_ReinitSubtracks()
{
    NON_CONST_ITERATE (TTrackProxies, t_iter, m_TrackProxies) {
        CTrackProxy* t_proxy = dynamic_cast<CTrackProxy*>(t_iter->GetPointer());
        if (!t_proxy) {
            break;
        }
        if (t_proxy->GetShown()) {
            x_ReinitTrack(t_proxy);
        }
    }
}


void CTrackContainer::x_ReinitTrack(CTrackProxy* t_proxy)
{
    const TSeqRange& curr_range = m_Context->GetVisSeqRange();
    // visited range covers the current visible range,
    // no need to reinitialize the track.
    if (t_proxy->GetVisitedRange().IntersectionWith(curr_range) ==  curr_range) {
        return;
    }

    const ILayoutTrackFactory* factory =
        m_ConfigMgr->GetTrackFactory(t_proxy->GetKey());

    // those tracks without background initialization have
    // been initialized already somewhere else.
    if (factory->NeedBackgroundInit()  ||  !t_proxy->GetTrack()) {
        ILayoutTrackFactory::SExtraParams params(m_AnnotLevel,
            m_Adaptive, &t_proxy->GetAnnots(),
            t_proxy->GetSubkey(), t_proxy->GetFilter(), t_proxy->GetSortBy());
        params.m_SkipGenuineCheck = GetSkipGenuineCheck();

        // create tracks for annotation annotated on the sequence directly.
        int level = GetAnnotLevel();
        if (GetAdaptive()) {
            level = -1;
        }

        m_DS->CreateTrack("creating " + t_proxy->GetName() + " track",
            factory, this, t_proxy, params,
            m_ConfigMgr->GetKnownAnnots(level));
        
        // make sure we only do this for background jobs
        if (CSGDataSource::IsBackgroundJob()) {
            SetTrackInitDone(false);
            SetMsg(", Initializing subtracks...");
        }
    }
}


void CTrackContainer::x_UpdateMsg()
{
    size_t on_num = GetChildren().size();
    size_t total = 0;
    ITERATE  (TTrackProxies, iter, m_TrackProxies) {
        if ((*iter)->GetTrack()) ++total;
    }

    string msg = ", Subtracks: " + NStr::SizetToString(on_num) + " on, ";
    msg += NStr::SizetToString(total - on_num) + " off";
    SetMsg(msg);
}


void CTrackContainer::x_AddSubtrack(CTrackContainer* parent,
                                    CLayoutTrack* track,
                                    CTempTrackProxy* proxy)
{
    // connect the real track to the proxy
    proxy->SetTrack(track);
    parent->AddTrackProxy(proxy);

    // load the profile
    track->LoadProfile(proxy->GetTrackProfile());
    track->SetComments(proxy->GetComments());
    track->SetHighlights(proxy->GetHighlights());
}


void CTrackContainer::x_ShowTrack(CLayoutTrack* track, bool on)
{
    bool was_on = track->IsOn();
    if (on != was_on) {
        if (was_on) {
            track->CloseMe();
        } else {
            // add the track back to visible list
            Add(track);
            SortChildTracksByOrder();
            track->ShowMe();
        }
        x_UpdateMsg();
    }
}


///////////////////////////////////////////////////////////////////////////////
/// CTrackContainerFactory
///////////////////////////////////////////////////////////////////////////////
ILayoutTrackFactory::TTrackMap
CTrackContainerFactory::CreateTracks(SConstScopedObject& object,
                                     ISGDataSourceContext* ds_context,
                                     CRenderingContext* r_cntx,
                                     const SExtraParams& params,
                                     const TAnnotMetaDataList& /*src_annots*/) const
{
    TTrackMap tracks;

    CIRef<ISGDataSource> ds =
        ds_context->GetDS(typeid(CFeaturePanelDSType).name(), object);
    CFeaturePanelDS* fp_ds = dynamic_cast<CFeaturePanelDS*>(ds.GetPointer());
    fp_ds->SetDSContext(ds_context);
    fp_ds->SetRenderingContext(r_cntx);
    CTrackContainer* track = new CTrackContainer(r_cntx, fp_ds);
    tracks["Container track"] = CRef<CLayoutTrack>(track);
    track->SetAnnotLevel(params.m_Level);
    track->SetAdaptive(params.m_Adaptive);
    track->SetSkipGenuineCheck(params.m_SkipGenuineCheck);
    return tracks;
}

string CTrackContainerFactory::GetExtensionIdentifier() const
{
    return CTrackContainer::m_TypeInfo.GetId();
}

string CTrackContainerFactory::GetExtensionLabel() const
{
    return CTrackContainer::m_TypeInfo.GetDescr();
}

END_NCBI_SCOPE
