/*  $Id: seq_text_ds.cpp 27374 2013-02-04 19:04:34Z bollin $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Colleen Bollin (adapted from a file by Andrey Yazhuk)
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>

#include <gui/widgets/seq_text/seq_text_ds.hpp>

#include <gui/objutils/label.hpp>  // TODO

#include <objects/seq/MolInfo.hpp>

#include <objmgr/bioseq_ci.hpp>
#include <objmgr/seq_vector.hpp>
#include <objmgr/util/sequence.hpp>
#include <objmgr/util/seq_loc_util.hpp>
#include <objmgr/feat_ci.hpp>
#include <objmgr/seqdesc_ci.hpp>

#include <objmgr/util/feature.hpp>

#include <gui/objutils/utils.hpp>
#include <objects/seqfeat/Gb_qual.hpp>

BEGIN_NCBI_SCOPE
USING_SCOPE(ncbi::objects);


CSeqTextDataSource::CSeqTextDataSource(CSeq_entry& sep,
                     CScope& scope)
{
    m_Sep = &sep;
    m_Scope = &scope;

    CSeq_entry_Handle handle = scope.GetSeq_entryHandle(sep);
    CBioseq_CI it(handle);

    m_SubjectHandle = (*it);
    m_ID = m_SubjectHandle.GetSeqId();
    CSeq_id_Handle idh = sequence::GetId(m_SubjectHandle, sequence::eGetId_Best);
    if (idh) {
        m_ID = idh.GetSeqId();
    }
    m_Loc.Reset ();

    CSeq_loc tmp(const_cast<CSeq_id&>(*CBioseq_Handle(*it).
                 GetSeqId()),
                 0,CBioseq_Handle(*it).GetBioseqLength() - 1);
    m_Loc.Add(tmp);

}


CSeqTextDataSource::CSeqTextDataSource(CBioseq_Handle handle,
                     CScope& scope)
{
    m_Scope = &scope;

    m_SubjectHandle = handle;

    m_ID = handle.GetSeqId();
    CSeq_id_Handle idh = sequence::GetId(handle, sequence::eGetId_Best);
    if (idh) {
        m_ID = idh.GetSeqId();
    }

    m_Loc.Reset ();
    CSeq_loc tmp ( const_cast<CSeq_id&>(*CBioseq_Handle(handle).
                   GetSeqId()),
                   0,CBioseq_Handle(handle).GetBioseqLength() - 1);
    m_Loc.Add (tmp);

}

CSeqTextDataSource::CSeqTextDataSource(CSeq_loc &loc, CScope& scope)
{
    m_Scope = &scope;
    m_SubjectHandle = scope.GetBioseqHandle(*loc.GetId());

    if (!m_SubjectHandle)
        NCBI_THROW(CException, eUnknown, "failed to retrieve sequence");

    m_ID = m_SubjectHandle.GetSeqId();
    CSeq_id_Handle idh = sequence::GetId(m_SubjectHandle, sequence::eGetId_Best);
    if (idh) {
        m_ID = idh.GetSeqId();
    }

    m_Loc.Reset();
    m_Loc.Add (loc);

}


CSeqTextDataSource::~CSeqTextDataSource()
{
}


CScope&     CSeqTextDataSource::GetScope()
{
    return m_Scope.GetObject();
}


TSeqPos CSeqTextDataSource::SequencePosToSourcePos (TSeqPos sequence_pos, bool *found_in_source)
{
    TSeqPos offset = 0;
    bool    found = false;
    bool    in_source = false;
    TSeqPos source_pos = 0;

    for(CSeq_loc_CI loc_it(m_Loc); loc_it && ! found; ++loc_it) {
        CSeq_loc_CI::TRange this_range = loc_it.GetRange();
        TSeqPos seq_start = this_range.GetFrom();
        TSeqPos seq_stop = this_range.GetTo();

        if (sequence_pos >= seq_start && sequence_pos <= seq_stop) {
            source_pos = offset + sequence_pos - seq_start;
            found = true;
            in_source = true;
        } else if (sequence_pos < seq_start) {
            source_pos = offset;
            found = true;
        }
        offset += seq_stop - seq_start + 1;
    }
    if (!found)
    {
        source_pos = offset - 1;
    }
    if (found_in_source != NULL) {
        *found_in_source = in_source;
    }
    return source_pos;
}


TSeqPos CSeqTextDataSource::SourcePosToSequencePos (TSeqPos source_pos)
{
    TSeqPos offset = 0;
    TSeqPos sequence_pos = 0;
    bool    found = false;

    if (!this) return 0;

    for(CSeq_loc_CI loc_it(m_Loc); loc_it && !found; ++loc_it) {
        CSeq_loc_CI::TRange this_range = loc_it.GetRange();
        TSeqPos seq_start = this_range.GetFrom();
        TSeqPos seq_stop = this_range.GetTo();

        if (source_pos >= offset && source_pos < offset + seq_stop - seq_start + 1) {
            sequence_pos = seq_start + source_pos - offset;
            found = true;
        }
        offset += seq_stop - seq_start + 1;
    }
    return sequence_pos;
}


void CSeqTextDataSource::GetSeqString (TSeqPos start, TSeqPos stop, string &buffer)
{
    string  seq_piece;
    TSeqPos offset = 0;

    // initialize buffer to empty
    buffer.erase();

    for(CSeq_loc_CI loc_it(m_Loc); loc_it; ++loc_it) {
        CSeq_loc_CI::TRange this_range = loc_it.GetRange();
        TSeqPos seq_start = this_range.GetFrom();
        TSeqPos seq_stop = this_range.GetTo();

        if (offset + seq_stop - seq_start > start && offset  <= stop)
        {
            CBioseq_Handle handle = GetScope().GetBioseqHandle(*loc_it.GetEmbeddingSeq_loc().GetId());
            CSeqVector vect = handle.GetSeqVector (CBioseq_Handle::eCoding_Iupac);
            TSeqPos piece_start = seq_start + start - offset;
            TSeqPos piece_stop = min (seq_stop + 1, seq_start + stop - offset);
            vect.GetSeqData (piece_start, piece_stop, seq_piece);
            buffer.append (seq_piece);
            start = offset + seq_stop - seq_start + 1;
        }
        offset += seq_stop - seq_start + 1;
    }
}


void InvertCase(string& seq)
{
    for (unsigned int i = 0; i < seq.length(); i++) {
        if (islower (seq[i])) {
            seq[i] = toupper(seq[i]);
        } else {
            seq[i] = tolower(seq[i]);
        }
    }
}


void CSeqTextDataSource::GetSeqData (TSeqPos start, TSeqPos stop, string& buffer, SAnnotSelector *feat_sel, bool showFeatAsLower)
{
    GetSeqString (start, stop, buffer);

    if (feat_sel)
    {
        TSeqPos offset = 0;
        for (CSeq_loc_CI seq_loc_it (m_Loc); seq_loc_it; ++ seq_loc_it) {
            CSeq_loc_CI::TRange seq_range = seq_loc_it.GetRange();
            TSeqPos seq_start = seq_range.GetFrom();
            TSeqPos seq_stop = seq_range.GetTo();
            TSeqPos adjusted_seq_stop = SequencePosToSourcePos (seq_stop);
            if (offset + seq_stop - seq_start >= start && offset < stop) {
                TSeqPos range_start = seq_start + max (start - offset, (TSeqPos) 0);
                TSeqPos range_stop = seq_start + min (stop - offset, seq_stop - seq_start + 1);
                TSeqRange range (range_start, range_stop);

                CBioseq_Handle handle = GetScope().GetBioseqHandle(*seq_loc_it.GetEmbeddingSeq_loc().GetId());
                for (CFeat_CI feat(handle, range, *feat_sel); feat; ++feat) {
                    const CSeq_loc& loc = feat->GetLocation();

                    for(CSeq_loc_CI loc_it(loc); loc_it; ++loc_it) {
                        CSeq_loc_CI::TRange feat_range = loc_it.GetRange();
                        TSeqPos feat_start = feat_range.GetFrom();
                        TSeqPos feat_stop = feat_range.GetTo();
                        TSeqPos adjusted_feat_start = SequencePosToSourcePos (feat_start);
                        if (feat_start != SourcePosToSequencePos (adjusted_feat_start)
                            && adjusted_feat_start == adjusted_seq_stop) {
                            // if the start of the feature is not in our source
                            // don't draw it
                            continue;
                        }

                        feat_start = max (adjusted_feat_start, start);
                        feat_stop = SequencePosToSourcePos (feat_stop);

                        while (feat_start <= feat_stop && feat_start - start < buffer.length())
                        {
                            buffer [feat_start - start] = tolower(buffer [feat_start - start]);
                            feat_start ++;
                        }
                    }
                }
            }
            offset += seq_stop - seq_start + 1;
        }
    }
    if (!showFeatAsLower) {
        InvertCase(buffer);
    }
}


int CSeqTextDataSource::ChooseBetterSubtype (int subtype1, int subtype2)
{
    if (subtype1 == CSeqFeatData::eSubtype_cdregion || subtype2 == CSeqFeatData::eSubtype_cdregion) {
        return CSeqFeatData::eSubtype_cdregion;
    } else if (subtype1 == CSeqFeatData::eSubtype_mRNA || subtype2 == CSeqFeatData::eSubtype_mRNA) {
        return CSeqFeatData::eSubtype_mRNA;
    } else if (subtype1 == CSeqFeatData::eSubtype_gene || subtype2 == CSeqFeatData::eSubtype_gene) {
        return CSeqFeatData::eSubtype_gene;
    } else if (subtype1 == CSeqFeatData::eSubtype_bad) {
        return subtype2;
    } else if (subtype2 == CSeqFeatData::eSubtype_bad) {
        return subtype1;
    } else if (subtype1 < subtype2) {
        return subtype1;
    } else {
        return subtype2;
    }
}


bool CSeqTextDataSource::x_IsSpliceSite(string splice_buffer, bool before_loc, bool is_minus)
{
    bool is_splice = false;

    if (splice_buffer.length() < 2) {
        is_splice = false;
    } else if (before_loc) {
        if (is_minus) {
            if ((splice_buffer[0] == 'A' && (splice_buffer[1] == 'C' || splice_buffer[1] == 'G'))
                || (splice_buffer[0] == 'G' && splice_buffer[1] == 'C')) {
                is_splice = true;
            } else {
                is_splice = false;
            }
        } else {
            if (splice_buffer[0] == 'A' && (splice_buffer[1] == 'G' || splice_buffer[1] == 'C')) {
                is_splice = true;
            } else {
                is_splice = false;
            }
        }
    } else {
        if (is_minus) {
            if ((splice_buffer[0] == 'C' || splice_buffer[0] == 'G') && splice_buffer[1] == 'T') {
                is_splice = true;
            } else {
                is_splice = false;
            }
        } else {
            if ((splice_buffer[0] == 'G' && (splice_buffer[1] == 'T' || splice_buffer[1] == 'C'))
                || (splice_buffer[0] == 'C' && splice_buffer[1] == 'T')) {
                is_splice = true;
            } else {
                is_splice = false;
            }
        }
    }
    return is_splice;
}


void CSeqTextDataSource::LookForSpliceJunctions (const CMappedFeat& feat, CSeqTextDefs::TSpliceSiteVector &splice_sites, TSeqPos start)
{
    const CSeq_loc& loc = feat.GetLocation();
    string          splice_buffer;
    bool            is_start = true;
    TSeqPos         unset_pos[2];
    bool            need_unset[2];

    need_unset[0] = false;
    need_unset[1] = false;
    unset_pos[0] = 0;
    unset_pos[1] = 0;
    for(CSeq_loc_CI loc_it(loc); loc_it; ++loc_it) {
        CSeq_loc_CI::TRange this_range = loc_it.GetRange();
        TSeqPos feat_start = this_range.GetFrom();
        TSeqPos feat_stop = this_range.GetTo();

        bool is_minus = false;

        if (loc_it.GetStrand() == eNa_strand_minus) {
            is_minus = true;
        }

        CBioseq_Handle handle = GetScope().GetBioseqHandle(*loc_it.GetEmbeddingSeq_loc().GetId());
        CSeqVector vect = handle.GetSeqVector (CBioseq_Handle::eCoding_Iupac);
        if (feat_start > 2  &&  ! is_start) {
            // check the two positions before the start of this location
            // to see if they contain A and G
            vect.GetSeqData (feat_start - 2, feat_start, splice_buffer);
            if (x_IsSpliceSite(splice_buffer, true, is_minus)) {
                TSeqPos source_pos;

                source_pos = SequencePosToSourcePos (feat_start - 2);
                if (SourcePosToSequencePos (source_pos) == feat_start - 2
                    && source_pos >= start
                    && source_pos - start < splice_sites.size()) {
                    splice_sites[source_pos - start] = true;
                }
                source_pos = SequencePosToSourcePos (feat_start - 1);
                if (SourcePosToSequencePos (source_pos) == feat_start - 1
                    && source_pos >= start
                    && source_pos - start < splice_sites.size()) {
                    splice_sites[source_pos - start] = true;
                }
            }
        }
        need_unset [0] = false;
        need_unset [1] = false;
        if (feat_stop <= handle.GetBioseqLength() - 2) {
            // check the two positions after the start of this location
            // to see if they contain G and T
            vect.GetSeqData (feat_stop + 1, feat_stop + 3, splice_buffer);
            if (x_IsSpliceSite(splice_buffer, false, is_minus)) {
                TSeqPos source_pos;

                source_pos = SequencePosToSourcePos (feat_stop + 1);
                if (SourcePosToSequencePos (source_pos) == feat_stop + 1
                    && source_pos >= start
                    && source_pos - start < splice_sites.size()
                    && !splice_sites[source_pos - start]) {
                    splice_sites[source_pos - start] = true;
                    need_unset [0] = true;
                    unset_pos [0] = source_pos - start;
                }
                source_pos = SequencePosToSourcePos (feat_stop + 2);
                if (SourcePosToSequencePos (source_pos) == feat_stop + 2
                    && source_pos >= start
                    && source_pos - start < splice_sites.size()
                    && ! splice_sites[source_pos - start]) {
                    splice_sites [source_pos - start] = true;
                    need_unset [1] = true;
                    unset_pos [1] = source_pos - start;
                }
            }
        }
        is_start = false;
    }
    /* unset splice site after last feature interval */
    if (need_unset [0]) {
        splice_sites[unset_pos[0]] = false;
    }
    if (need_unset[1]) {
        splice_sites[unset_pos[1]] = false;
    }

}


void
CSeqTextDataSource::x_AddVariationsFromFeature
(const CSeq_feat &feat,
 const CSeq_loc  &loc,
 CBioseq_Handle handle,
 CSeqTextDefs::TVariationGraphVector &variations)
{

    TSeqPos feat_left = loc.GetStart(eExtreme_Positional);
    TSeqPos feat_right = loc.GetStop(eExtreme_Positional);
    bool    replace_found = false;

    // get contents of sequence at this location, so we know which /replace
    // represents the original text (and we can display the other instead)
    string  seq_piece;
    CSeqVector vect = handle.GetSeqVector (CBioseq_Handle::eCoding_Iupac);
    vect.GetSeqData (feat_left, feat_right + 1, seq_piece);

    // needed sequence coordinates for getting actual sequence text
    // store source coordinates in variation holder
    feat_left = SequencePosToSourcePos(feat_left);
    feat_right = SequencePosToSourcePos(feat_right);

    if (feat.CanGetQual()) {
        ITERATE( CSeq_feat::TQual, it, feat.GetQual()) {
            if (!NStr::CompareNocase((*it)->GetQual(),"replace")
                && NStr::CompareNocase((*it)->GetVal(),seq_piece))
            {
                variations.push_back (CSeqTextVariationGraph(feat_left, feat_right, (*it)->GetVal()));
                replace_found = true;
            }
        }
    }
    if (!replace_found) {
        variations.push_back (CSeqTextVariationGraph(feat_left, feat_right, "-"));
    }
}


void
CSeqTextDataSource::GetFeatureData
(TSeqPos start_offset,
 TSeqPos stop_offset,
 CSeqTextConfig *cfg,
 ISeqTextGeometry* pParent,
 CSeqTextDefs::TSubtypeVector &subtypes,
 CSeqTextDefs::TSpliceSiteVector &splice_sites,
 CSeqTextDefs::TVariationGraphVector &variations)
{
    unsigned int i;

    subtypes.clear();
    splice_sites.clear();

    variations.clear();

    // seq_data_offset contains the lengths of the intervals of source data
    // that have already passed.  start_offset and stop_offset should be
    // interpreted relative to these values.
    TSeqPos seq_data_offset = 0;
    // output_range_offset contains the lengths of output data that have already
    // been filled in.  data for a range should be filled in based on the
    // distance from the start of the range plus the output_range_offset
    //TSeqPos output_range_offset = 0;

    // this fills in the output range
    for (i = 0; i < stop_offset - start_offset + 2; i++) {
        subtypes.push_back (CSeqFeatData::eSubtype_bad);
        splice_sites.push_back (false);
    }

    // no configuration - leave the feature data blank
    if (cfg == NULL) {
        return;
    }

    // don't look for features after the end of the data
    stop_offset = min (stop_offset, GetDataLen() - 1);

    // examine each interval in the source data
    for (CSeq_loc_CI seq_loc_it (m_Loc); seq_loc_it; ++ seq_loc_it) {
        CSeq_loc_CI::TRange seq_range = seq_loc_it.GetRange();
        TSeqPos seq_start = seq_range.GetFrom();
        TSeqPos seq_stop = seq_range.GetTo();
        // if this interval contains locations between start_offset and stop_offset, collect feature data
        if (seq_data_offset + seq_stop - seq_start >= start_offset && seq_data_offset < stop_offset + 1) {
            TSeqPos feat_range_start = seq_start;
            // if this interval starts after the beginning of the output range, adjust the start
            // of where we look for features
            if (start_offset > seq_data_offset) {
                feat_range_start = SourcePosToSequencePos (start_offset);
            }
            // if this interval ends after the end of the output range, adjust the end of where
            // we look for features
            TSeqPos feat_range_stop = min (seq_stop, SourcePosToSequencePos(stop_offset));

            // this is the range in which we will look for features
            TSeqRange range (feat_range_start, feat_range_stop);

            CBioseq_Handle handle = GetScope().GetBioseqHandle(*seq_loc_it.GetEmbeddingSeq_loc().GetId());

            objects::SAnnotSelector sel = CSeqUtils::GetAnnotSelector();

            // iterate through the features located in the range of interest
            CFeat_CI feat_ci(handle, range, sel);
            while (feat_ci)
            {
                // feature subtypes
                int subtype = feat_ci->GetFeatSubtype();
                if (cfg->GetShow (subtype) && cfg->GetFeatureColorationChoice() == CSeqTextPaneConfig::eAll) {
                    if (pParent) {
                        pParent->STG_SetSubtypesForFeature (subtypes, feat_ci->GetLocation(), feat_ci->GetFeatSubtype(), start_offset, stop_offset);
                    }
                }

                if (subtype == CSeqFeatData::eSubtype_cdregion) {
                    LookForSpliceJunctions (*feat_ci, splice_sites, start_offset);
                }

                if (subtype == CSeqFeatData::eSubtype_variation) {
                    x_AddVariationsFromFeature (feat_ci->GetOriginalFeature(), feat_ci->GetLocation(), handle, variations);
                }

                ++feat_ci;
            }
        }
        // the offset of source data is increased by the length of this interval
        seq_data_offset += seq_stop - seq_start + 1;
    }
}


bool CSeqTextDataSource::IsmRNASequence ()
{
    bool is_mRNA = true;

    for (CSeq_loc_CI seq_loc_it (m_Loc); seq_loc_it && is_mRNA; ++ seq_loc_it) {
        CBioseq_Handle handle = GetScope().GetBioseqHandle(*seq_loc_it.GetEmbeddingSeq_loc().GetId());
        if (!handle.CanGetInst_Mol() || handle.GetInst_Mol() != CSeq_inst::eMol_rna) {
            is_mRNA = false;
        } else {
            CSeqdesc_CI di (handle, CSeqdesc::e_Molinfo);
            if (!di || di->GetMolinfo().GetBiomol() != CMolInfo::eBiomol_mRNA) {
                is_mRNA = false;
            }
        }
    }
    return is_mRNA;
}


void
CSeqTextDataSource::GetSubtypesForAlternatingExons
(TSeqPos start_offset,
 TSeqPos stop_offset,
 ISeqTextGeometry* pParent,
 CSeqTextDefs::TSubtypeVector &subtypes)
{
    unsigned int i;
    bool         even = false;

    // only do this when sequence is mRNA

    if (!IsmRNASequence ()) return;

    subtypes.clear();
    if (!pParent) return;

    // seq_data_offset contains the lengths of the intervals of source data
    // that have already passed.  start_offset and stop_offset should be
    // interpreted relative to these values.
    TSeqPos seq_data_offset = 0;
    // output_range_offset contains the lengths of output data that have already
    // been filled in.  data for a range should be filled in based on the
    // distance from the start of the range plus the output_range_offset
    //TSeqPos output_range_offset = 0;

    // this fills in the output range
    for (i = 0; i < stop_offset - start_offset + 2; i++) {
        subtypes.push_back (CSeqFeatData::eSubtype_bad);
    }



    // don't look for features after the end of the data
    stop_offset = min (stop_offset, GetDataLen() - 1);

    objects::SAnnotSelector sel(CSeqFeatData::eSubtype_exon);

    // examine each interval in the source data
    for (CSeq_loc_CI seq_loc_it (m_Loc); seq_loc_it; ++ seq_loc_it) {
        CSeq_loc_CI::TRange seq_range = seq_loc_it.GetRange();
        TSeqPos seq_start = seq_range.GetFrom();
        TSeqPos seq_stop = seq_range.GetTo();

        // we need to count the exons that appear before seq_data_offset, so that
        // we can alternate the colors properly
        if (seq_data_offset < start_offset) {
            TSeqPos feat_range_start = seq_start;
            TSeqPos feat_range_stop = min (seq_stop, SourcePosToSequencePos(start_offset - 1));
            // this is the range in which we will look for features
            TSeqRange range (feat_range_start, feat_range_stop);

            CBioseq_Handle handle = GetScope().GetBioseqHandle(*seq_loc_it.GetEmbeddingSeq_loc().GetId());

            // iterate through the features located in the range of interest
            CFeat_CI feat_ci(handle, range, sel);
            while (feat_ci) {
                // only count if feature ends before start of display
                CSeq_loc_CI::TRange feat_range = feat_ci->GetLocation().GetTotalRange();
                if (feat_range.GetTo() < SourcePosToSequencePos (start_offset))
                {
                    even = !even;
                }
                ++feat_ci;
            }
        }

        // if this interval contains locations between start_offset and stop_offset, collect feature data
        if (seq_data_offset + seq_stop - seq_start >= start_offset && seq_data_offset < stop_offset + 1) {
            TSeqPos feat_range_start = seq_start;
            // if this interval starts after the beginning of the output range, adjust the start
            // of where we look for features
            if (start_offset > seq_data_offset) {
                feat_range_start = SourcePosToSequencePos (start_offset);
            }
            // if this interval ends after the end of the output range, adjust the end of where
            // we look for features
            TSeqPos feat_range_stop = min (seq_stop, SourcePosToSequencePos(stop_offset));

            // this is the range in which we will look for features
            TSeqRange range (feat_range_start, feat_range_stop);

            CBioseq_Handle handle = GetScope().GetBioseqHandle(*seq_loc_it.GetEmbeddingSeq_loc().GetId());

            // iterate through the features located in the range of interest
            CFeat_CI feat_ci(handle, range, sel);
            while (feat_ci)
            {
                // only color odd exons
                if (!even) {
                    pParent->STG_SetSubtypesForFeature (subtypes, feat_ci->GetLocation(), feat_ci->GetFeatSubtype(), start_offset, stop_offset);
                }
                even = !even;
                ++feat_ci;
            }
        }
        // the offset of source data is increased by the length of this interval
        seq_data_offset += seq_stop - seq_start + 1;
    }
}


void CSeqTextDataSource::GetVariations (TSeqPos start_offset, TSeqPos stop_offset, CSeqTextDefs::TVariationGraphVector &variations)
{
    variations.clear();

    // seq_data_offset contains the lengths of the intervals of source data
    // that have already passed.  start_offset and stop_offset should be
    // interpreted relative to these values.
    TSeqPos seq_data_offset = 0;
    // output_range_offset contains the lengths of output data that have already
    // been filled in.  data for a range should be filled in based on the
    // distance from the start of the range plus the output_range_offset
    //TSeqPos output_range_offset = 0;

    // don't look for features after the end of the data
    stop_offset = min (stop_offset, GetDataLen() - 1);

    // examine each interval in the source data
    for (CSeq_loc_CI seq_loc_it (m_Loc); seq_loc_it; ++ seq_loc_it) {
        CSeq_loc_CI::TRange seq_range = seq_loc_it.GetRange();
        TSeqPos seq_start = seq_range.GetFrom();
        TSeqPos seq_stop = seq_range.GetTo();
        // if this interval contains locations between start_offset and stop_offset, collect feature data
        if (seq_data_offset + seq_stop - seq_start >= start_offset && seq_data_offset < stop_offset + 1) {
            TSeqPos feat_range_start = seq_start;

            // if this interval starts after the beginning of the output range, adjust the start
            // of where we look for features
            if (start_offset > seq_data_offset) {
                feat_range_start = SourcePosToSequencePos (start_offset);
            }
            // if this interval ends after the end of the output range, adjust the end of where
            // we look for features
            TSeqPos feat_range_stop = min (seq_stop, SourcePosToSequencePos(stop_offset));

            // this is the range in which we will look for features
            TSeqRange range (feat_range_start, feat_range_stop);


            CBioseq_Handle handle = GetScope().GetBioseqHandle(*seq_loc_it.GetEmbeddingSeq_loc().GetId());

            // we only want variations
            objects::SAnnotSelector sel = CSeqUtils::GetAnnotSelector();
            sel.SetFeatSubtype(CSeqFeatData::eSubtype_variation);

            // iterate through the features located in the range of interest
            CFeat_CI feat_ci(handle, range, sel);
            while (feat_ci)
            {
                x_AddVariationsFromFeature (feat_ci->GetOriginalFeature(), feat_ci->GetLocation(), handle, variations);
                ++feat_ci;
            }
       }
        // the offset of source data is increased by the length of this interval
        seq_data_offset += seq_stop - seq_start + 1;
    }
}


void CSeqTextDataSource::RenderFeatureExtras (ISeqTextGeometry* pParent, CGlPane &pane, TSeqPos start_offset, TSeqPos stop_offset)
{
    if (pParent == NULL) return;

    // open pane for drawing
    pane.OpenOrtho();

    // seq_data_offset contains the lengths of the intervals of source data
    // that have already passed.  start_offset and stop_offset should be
    // interpreted relative to these values.
    TSeqPos seq_data_offset = 0;
    // output_range_offset contains the lengths of output data that have already
    // been filled in.

    // don't look for features after the end of the data
    stop_offset = min (stop_offset, GetDataLen() - 1);

    // examine each interval in the source data
    for (CSeq_loc_CI seq_loc_it (m_Loc); seq_loc_it; ++ seq_loc_it) {
        CSeq_loc_CI::TRange seq_range = seq_loc_it.GetRange();
        TSeqPos seq_start = seq_range.GetFrom();
        TSeqPos seq_stop = seq_range.GetTo();
        // if this interval contains locations between start_offset and stop_offset, draw feature extras
        if (seq_data_offset + seq_stop - seq_start >= start_offset && seq_data_offset < stop_offset + 1) {
            TSeqPos feat_range_start = seq_start;
            // if this interval starts after the beginning of the output range, adjust the start
            // of where we look for features
            if (start_offset > seq_data_offset) {
                feat_range_start = SourcePosToSequencePos (start_offset);
            }
            // if this interval ends after the end of the output range, adjust the end of where
            // we look for features
            TSeqPos feat_range_stop = min (seq_stop, SourcePosToSequencePos(stop_offset));

            // this is the range in which we will look for features
            TSeqRange range (feat_range_start, feat_range_stop);

            CBioseq_Handle handle = GetScope().GetBioseqHandle(*seq_loc_it.GetEmbeddingSeq_loc().GetId());

            objects::SAnnotSelector sel = CSeqUtils::GetAnnotSelector();

            // iterate through the features located in the range of interest
            CFeat_CI feat_ci(handle, range, sel);
            while (feat_ci)
            {
                pParent->STG_RenderFeatureExtras(*feat_ci);
                ++feat_ci;
            }
        }
    }
    pane.Close();
}


void
CSeqTextDataSource::GetIntervalBreaks
(TSeqPos start_offset,
 TSeqPos stop_offset,
 CSeqTextDefs::TSeqPosVector &breaks)
{
    bool    is_first = true;
    TSeqPos offset = 0;

    breaks.clear();


    for (CSeq_loc_CI seq_loc_it (m_Loc); seq_loc_it && offset < stop_offset; ++ seq_loc_it) {
        if (is_first) {
            is_first = false;
        }
        else if (offset >= start_offset)
        {
            breaks.push_back (offset);
        }
        CSeq_loc_CI::TRange seq_range = seq_loc_it.GetRange();
        TSeqPos seq_start = seq_range.GetFrom();
        TSeqPos seq_stop = seq_range.GetTo();
        offset += seq_stop - seq_start + 1;
    }
}


TSeqPos CSeqTextDataSource::GetDataLen ()
{
    if (!this) {
        return 0;
    } else {
        return sequence::GetLength(m_Loc, m_Scope);
    }
}

string CSeqTextDataSource::GetTitle ()
{
    string s = "Sequence : ";
    auto_ptr<sequence::CDeflineGenerator> gen(new sequence::CDeflineGenerator());
    s += gen->GenerateDefline(m_SubjectHandle);
    return s;
}


CSeqTextDataSource::TIdRef  CSeqTextDataSource::GetId() const
{
    return m_ID;
}


CFeat_CI * CSeqTextDataSource::GetFeaturesAtPosition(TSeqPos source_pos, SAnnotSelector *feat_sel)
{
    if (!this) return NULL;

    const CSeq_id *seq_id = m_Loc.GetId();
    if (!seq_id) return NULL;

    if (source_pos > GetDataLen() - 1) {
        return NULL;
    }
    TSeqPos seq_pos = SourcePosToSequencePos(source_pos);

    CBioseq_Handle handle = GetScope().GetBioseqHandle(*seq_id);
    TSeqRange range (seq_pos, seq_pos);
    if (feat_sel) {
        return new CFeat_CI(handle, range, *feat_sel);
    }

    objects::SAnnotSelector sel = CSeqUtils::GetAnnotSelector();
    return new CFeat_CI(handle, range, sel);
}

string CSeqTextDataSource::GetToolTipForSourcePos (TSeqPos source_pos, SAnnotSelector *feat_sel)
{
    string  tooltip_text = "";
    CFeat_CI * pfeat_it = GetFeaturesAtPosition(source_pos, feat_sel);

    if (pfeat_it) {
        CFeat_CI feat = *pfeat_it;
        while (feat) {
            string feat_title  = "";
            CLabel::GetLabel(feat->GetOriginalFeature(), &feat_title,
                             CLabel::eUserTypeAndContent, m_Scope);
            if (!feat_title.empty()) {
                if (!tooltip_text.empty()) {
                    tooltip_text.append ("\n");
                }
                tooltip_text.append (feat_title);
            }
            ++feat;
        }
        delete pfeat_it;
    }
    return tooltip_text;
}


const CSeq_loc* CSeqTextDataSource::GetLoc() const
{
    return &m_Loc;
}

int CSeqTextDataSource::FindSequenceFragment (const string& fragment, TSeqPos start_search)
{
    TSeqPos data_len = GetDataLen();
    TSeqPos search_buffer_len;
    TSeqPos end_search;
    string  search_buffer;

    if (fragment.length() > data_len - start_search) {
        return -1;
    }

    search_buffer_len = max ((int) 1000, (int)(3 * fragment.length()));
    search_buffer_len = min (search_buffer_len, data_len - start_search);

    while (start_search < data_len - fragment.length() + 1) {
        end_search = start_search + search_buffer_len;
        GetSeqString (start_search, end_search, search_buffer);
        string::size_type pos = NStr::FindNoCase(search_buffer, fragment);
        if (pos != string::npos) {
            return pos + start_search;
        }
        start_search = end_search - fragment.length() + 1;
    }
    return -1;
}


void CSeqTextDataSource::FindSequenceFragmentList (const string& fragment,
                                                   CSeqTextDefs::TSeqPosVector &locations)
{
    TSeqPos data_len = GetDataLen();
    TSeqPos search_buffer_len;
    TSeqPos start_search = 0, end_search;
    string  search_buffer;

    locations.clear();

    if (fragment.length() > data_len - start_search) {
        return;
    }

    search_buffer_len = max ((int) 1000, (int)(3 * fragment.length()));
    search_buffer_len = min (search_buffer_len, data_len - start_search);

    while (start_search < data_len - fragment.length() + 1) {
        end_search = start_search + search_buffer_len;
        GetSeqString (start_search, end_search, search_buffer);
        string::size_type pos = NStr::FindNoCase(search_buffer, fragment);
        while (pos != string::npos) {
            locations.push_back (pos + start_search);
            if (end_search > pos + fragment.length()) {
                pos = NStr::FindNoCase(search_buffer, fragment, pos + 1);
            } else {
                pos = string::npos;
            }
        }
        start_search = end_search - fragment.length() + 1;
    }
}


END_NCBI_SCOPE
