/*  $Id: draw_text_stream.cpp 23892 2011-06-17 15:13:45Z katargir $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Roman Katargin
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>

#include <gui/widgets/text_widget/draw_text_stream.hpp>
#include <gui/widgets/wx/wx_utils.hpp>

BEGIN_NCBI_SCOPE

CDrawTextOStream::CDrawTextOStream(
    wxDC& dc, int x, int y,
    size_t rowHeight, size_t indentStep,
    int selStartRow, int selEndRow,
    int selStartCol, int selEndCol) :
        CBaseTextOStream(dc, x, y, rowHeight, indentStep),
            m_SelStartRow(selStartRow), m_SelStartCol(selStartCol),
            m_SelEndRow(selEndRow), m_SelEndCol(selEndCol)
{
}

CStyledTextOStream& CDrawTextOStream::operator<< (const string& text_in)
{
    wxString text = ToWxString(text_in);
    int length = text.length();
    if (length == 0)
        return *this;

    int selStart = length, selEnd = length;

    if (m_Row >= m_SelStartRow && m_Row <= m_SelEndRow) {
        selStart = (m_Row == m_SelStartRow) ?
                    max((int)(m_SelStartCol-m_CurrentLineLength),0) : 0;

        if (m_Row == m_SelEndRow)
            selEnd = max((int)(m_SelEndCol-m_CurrentLineLength),0);

        //if (selEnd < selStart) swap(selStart, selEnd);

        if (selEnd < 0 || selStart >= length || selStart == selEnd)
            selStart = selEnd = length;
    }

    wxArrayInt widths;
    m_DC.GetPartialTextExtents(text, widths);

    int y = m_Row*(int)m_RowHeight;
    if (selStart > 0) {
        wxString substr = text.substr(0, selStart);
        m_DC.DrawText(substr, m_x, y);
    }

    if (selStart < length) {
        m_DC.SetBackgroundMode(wxSOLID);
        m_DC.SetTextForeground(wxColor(255,255,255));
        m_DC.SetTextBackground(wxColor(49,106,197));

        wxString substr = text.substr(selStart, selEnd-selStart);
        m_DC.DrawText(substr, m_x+(selStart>0 ? widths[selStart-1]:0), y);

        if (m_CurrentStyle != 0) {
            m_DC.SetBackgroundMode(m_CurrentStyle->GetBackgroundMode());
            m_DC.SetTextForeground(m_CurrentStyle->GetTextForeground());
            m_DC.SetTextBackground(m_CurrentStyle->GetTextBackground());
        }
        else {
            m_DC.SetBackgroundMode(m_BackgroundMode);
            m_DC.SetTextForeground(m_TextForeground);
            m_DC.SetTextBackground(m_TextBackground);
        }
    }

    if (selEnd < length) {
        wxString substr = text.substr(selEnd, length-selEnd);
        m_DC.DrawText(substr, m_x+widths[selEnd-1], y);
    }

    x_UpdatePos(length, widths[length-1]);

    if (x_InNB()) x_UpdateNBRect(wxSize(widths[length-1], m_RowHeight));

    return *this;
}

void CDrawTextOStream::Indent (size_t steps)
{
    if (steps == 0)
        return;

    int selStart = steps, selEnd = steps;

    if (m_Row >= m_SelStartRow && m_Row <= m_SelEndRow) {
        selStart = (m_Row == m_SelStartRow) ?
                    max((int)(m_SelStartCol-m_CurrentLineLength),0) : 0;

        if (m_Row == m_SelEndRow)
            selEnd = max((int)(m_SelEndCol-m_CurrentLineLength),0);

        //if (selEnd < selStart) swap(selStart, selEnd);

        if (selEnd < 0 || selStart >= (int)steps || selStart == selEnd)
            selStart = selEnd = steps;
    }

    if (selStart < (int)steps) {
        wxBrush saveBrush = m_DC.GetBrush();
        wxPen savePen = m_DC.GetPen();

        wxColor selClr(49,106,197);
        m_DC.SetBrush(wxBrush(selClr));
        m_DC.SetPen(wxPen(selClr, 1));
        m_DC.DrawRectangle(m_x + selStart*m_IndentStep, m_Row*(int)m_RowHeight,
                           (selEnd - selStart)*m_IndentStep, m_RowHeight);

        m_DC.SetBrush(saveBrush);
        m_DC.SetPen(savePen);
    }

    size_t width = m_IndentStep*steps;
    x_UpdatePos(steps, width);
    if (x_InNB()) x_UpdateNBRect(wxSize(width, m_RowHeight));
}

void CDrawTextOStream::x_OnNBEnd()
{
    wxPen savePen = m_DC.GetPen();
    m_DC.SetPen(wxPen(wxColor(128,128,128),1));

    m_DC.DrawLine(m_NBRect.x + 2, m_NBRect.y, m_NBRect.x + m_NBRect.width - 2, m_NBRect.y);
    m_DC.DrawLine(m_NBRect.x + m_NBRect.width - 2, m_NBRect.y + 1,
                  m_NBRect.x + m_NBRect.width - 2, m_NBRect.y + m_NBRect.height - 1);
    m_DC.DrawLine(m_NBRect.x + 2, m_NBRect.y + m_NBRect.height - 1,
                  m_NBRect.x + m_NBRect.width - 2, m_NBRect.y + m_NBRect.height - 1);
    m_DC.DrawLine(m_NBRect.x + 1, m_NBRect.y + 1,
                  m_NBRect.x + 1, m_NBRect.y + m_NBRect.height - 1);

    m_DC.SetPen(savePen);
}

END_NCBI_SCOPE
