/*  $Id: text_item_panel.cpp 27922 2013-04-25 16:51:42Z katargir $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Roman Katargin
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>

#include <gui/utils/view_event.hpp>
#include <gui/utils/event_handler.hpp>

#include <gui/widgets/wx/popup_event.hpp>
#include <gui/widgets/wx/wx_utils.hpp>
#include <gui/widgets/wx/ui_command.hpp>

#include <gui/widgets/text_widget/text_item_panel.hpp>
#include <gui/widgets/text_widget/text_item.hpp>
#include <gui/widgets/text_widget/composite_text_item.hpp>
#include <gui/widgets/text_widget/text_event.hpp>
#include <gui/widgets/text_widget/calc_caretpos_stream.hpp>
#include <gui/widgets/text_widget/text_line_iterator.hpp>
#include <gui/widgets/text_widget/expand_item.hpp>

#include <gui/utils/menu_item.hpp>

#include <wx/sizer.h>
#include <wx/menu.h>
#include <wx/dcclient.h>
#include <wx/dcbuffer.h>
#include <wx/settings.h>
#include <wx/caret.h>

#include <wx/sstream.h>
#include <wx/txtstrm.h>
#include <wx/clipbrd.h>
#include <wx/dataobj.h>

BEGIN_NCBI_SCOPE

void CTextPanelContext::SetPanel(CTextItemPanel* panel)
{
    if (m_Panel) {
        m_Panel->PopEventHandler();
    }

    m_Panel = panel;

    if (m_Panel) {
        m_Panel->PushEventHandler(this);
    }
}

void CTextPanelContext::SelectItem(ITextItem* item, bool controlDown, bool WXUNUSED(shiftDown))
{
    if (controlDown) {
        set<ITextItem*>::iterator it = m_SelectedItems.find(item);
        if (it != m_SelectedItems.end())
            m_SelectedItems.erase(it);
        else
            m_SelectedItems.insert(item);
    }
    else {
        m_SelectedItems.clear();
        m_SelectedItems.insert(item);
    }

    m_Panel->Refresh();

    CEvent evt(CEvent::eEvent_Message, CViewEvent::eWidgetSelectionChanged);
    Send(&evt, ePool_Parent);
}

void CTextPanelContext::EditItem(ITextItem* WXUNUSED(item), bool WXUNUSED(controlDown), bool WXUNUSED(shiftDown))
{
}

bool CTextPanelContext::IsItemSelected(ITextItem* item)
{
    return (m_SelectedItems.find(item) != m_SelectedItems.end());
}

void CTextPanelContext::GetSelectedObjects (vector< CConstRef<CObject> >& objs)
{
    ITERATE (set<ITextItem*>, it, m_SelectedItems) {
        CConstRef<CObject> object = (*it)->GetAssosiatedObject();
        if (object) objs.push_back(CConstRef<CObject>(object));
    }
}

void CTextPanelContext::GetSelectedObjects (TConstScopedObjects& objects)
{
    ITERATE (set<ITextItem*>, it, m_SelectedItems) {
        CConstRef<CObject> object = (*it)->GetAssosiatedObject();
        if (object) {
            SConstScopedObject scopes_object;
            scopes_object.object = object;
            objects::CScope* scope = (*it)->GetAssosiatedScope();
            if (scope)
                scopes_object.scope.Reset(scope);
            objects.push_back(scopes_object);
        }
    }
}

void CTextPanelContext::ClearMouseMove()
{
    m_Hier.clear();
    m_MouseOverItem = 0;
}

void CTextPanelContext::AddHier(ITextItem* item)
{
    m_Hier.push_back(item);
}

void CTextPanelContext::MouseOverItem(ITextItem* item)
{
    m_MouseOverItem = item;
}

void CTextPanelContext::TriggerTooltip()
{
    if (m_Panel)
        m_Panel->TriggerTooltip();
}

void CTextPanelContext::SetSelectedObjects (vector< CConstRef<CObject> >& objs)
{
    m_SelectedItems.clear();

    set<const CObject*> objSet;
    ITERATE (vector< CConstRef<CObject> >, it, objs)
        objSet.insert(it->GetPointer());

    if (objSet.empty())
        return;

    vector<ITextItem*> leaves;
    m_Panel->GetItem()->GatherLeaves(leaves, vector<int>(), vector<int>());
    ITERATE (vector<ITextItem*>, it, leaves) {
        if (objSet.find((*it)->GetAssosiatedObject()) != objSet.end())
            m_SelectedItems.insert(*it);
    }

    m_Panel->Refresh();
}

BEGIN_EVENT_TABLE( CTextItemPanel, wxScrolledWindow )
    EVT_MOUSE_EVENTS( CTextItemPanel::OnMouseEvent )
    EVT_PAINT( CTextItemPanel::OnPaint )
    EVT_ERASE_BACKGROUND( CTextItemPanel::OnEraseBackground )
    EVT_SET_FOCUS(CTextItemPanel::OnSetFocus)
    EVT_KILL_FOCUS(CTextItemPanel::OnKillFocus)
    EVT_TIMER(-1, CTextItemPanel::OnTimer)
    EVT_CONTEXT_MENU( CTextItemPanel::OnContextMenu )
    EVT_MENU( wxID_COPY, CTextItemPanel::OnCopy )
    EVT_UPDATE_UI( wxID_COPY, CTextItemPanel::OnUpdateCopy )
    EVT_MENU( wxID_SELECTALL, CTextItemPanel::OnSelectAll )
END_EVENT_TABLE()

CTextItemPanel::CTextItemPanel(wxWindow* parent, wxWindowID id,
                               const wxPoint& pos, const wxSize& size, long style)
    : m_Selecting(false), m_Timer(this), m_TooltipWindow(0), m_MousePos(-1,-1),
      m_BlockEvents(false)
{
    Create(parent, id, pos, size, style);
    Init();
}

CTextItemPanel::~CTextItemPanel()
{
    if (m_Context.get())
        m_Context->SetPanel(0);
}

bool CTextItemPanel::ProcessEvent(wxEvent& event)
{
    if (m_BlockEvents)
        return false;

    return wxScrolledWindow::ProcessEvent(event);
}

void CTextItemPanel::Init()
{
#if (wxMAJOR_VERSION == 2 && wxMINOR_VERSION < 9)
#   ifdef __WXMAC__    
        SetBackgroundStyle(wxBG_STYLE_COLOUR);
        SetBackgroundColour(*wxWHITE);
#   else
        SetBackgroundStyle(wxBG_STYLE_SYSTEM);
#   endif
#else
    SetBackgroundStyle(wxBG_STYLE_PAINT);
#endif

    wxClientDC dc(this);
    InitDC(dc);
    SetScrollRate(dc.GetCharWidth(),dc.GetCharHeight());

    wxCaret *caret = new wxCaret(this, 1, dc.GetCharHeight());
    SetCaret(caret);
}

bool CTextItemPanel::IsTextSelected() const
{
    return !m_TextSelection.IsEmpty();
}

void CTextItemPanel::SelStart(int row, int col, bool shift)
{
    CTextPosition pos(row, col);

    if (shift)
        m_TextSelection = CTextBlock(m_TextSelection.GetStart(), pos);
    else
        m_TextSelection = CTextBlock(pos, pos);

    m_Selecting = true;
    CaptureMouse();
    x_TextPosChanged();
}

void CTextItemPanel::SelMove(int row, int col)
{
    m_TextSelection = CTextBlock(m_TextSelection.GetStart(), CTextPosition(row, col));
    x_TextPosChanged();
}

void CTextItemPanel::SelEnd(int row, int col)
{
    m_TextSelection = CTextBlock(m_TextSelection.GetStart(), CTextPosition(row, col));
    m_Selecting = false;
    ReleaseMouse();
    x_TextPosChanged();
}

void CTextItemPanel::SetSelection(const CTextBlock& selection)
{
    m_TextSelection = selection;
    x_TextPosChanged();
}

void CTextItemPanel::x_TextPosChanged()
{
    Refresh();

    if (FindFocus() != this) return;

    CTextWidgetEvent evt(CEvent::eEvent_Message, CTextWidgetEvent::ePositionChanged);
    CTextWidgetEvent::TAttachment* attachment = new CTextWidgetEvent::TAttachment();
    attachment->SetData(m_TextSelection);
    evt.SetAttachment(attachment, CEvent::eDelete);
    Send(&evt, ePool_Parent);
}

bool CTextItemPanel::Layout()
{
    if (m_Item.get()) {
        SetVirtualSize(m_Item->GetSize());
    }

    return true;
}

bool CTextItemPanel::SetFont(const wxFont& font)
{
    bool retVal = wxScrolledWindow::SetFont(font);

    wxClientDC dc(this);
    InitDC(dc);

    RecalSize();
    Layout();
    Refresh();

    SetScrollRate(dc.GetCharWidth(),dc.GetCharHeight());

    wxCaret* caret = GetCaret();
    if (caret)
        caret->SetSize(1, dc.GetCharHeight());

    return retVal;
}

void CTextItemPanel::SetData(ITextItem* item, CTextPanelContext* context)
{
    if (m_Context.get())
        m_Context->SetPanel(0);

    m_TextSelection = CTextBlock();

    m_Item.reset(item);
    m_Context.reset(context);

    if (m_Item.get() != 0) {
        m_Context->SetPanel(this);
        RecalSize();
        m_Context->AddListener(this, CEventHandler::ePool_Parent);
    }
    else
        SetVirtualSize(wxSize(0,0));

    m_BlockEvents = false;
}

void CTextItemPanel::RecalSize()
{
    if (m_Item.get() == 0)
        return;

    wxClientDC dc(this);
    InitDC(dc);
    m_Context->m_CharWidth = dc.GetCharWidth();
    m_Context->m_CharHeight = dc.GetCharHeight();
    wxCoord w, h;
    dc.GetTextExtent(wxT(" "), &w, &h);
    m_Context->m_SpaceWidth = w;

    m_Context->ResetTextSizeCache();
    m_Item->CalcSize(dc, m_Context.get());
    m_Context->ResetTextSizeCache();
}

void CTextItemPanel::AddItems(const vector<ITextItem*>& items)
{
    CCompositeTextItem* composite = dynamic_cast<CCompositeTextItem*>(m_Item.get());
    if (!composite)
        return;

    {
        wxBusyCursor wait;
        wxClientDC dc(this);
        InitDC(dc);

        m_Context->ResetTextSizeCache();

        ITERATE(vector<ITextItem*>, it, items) {
            (*it)->CalcSize(dc, m_Context.get());
            composite->AddItem(*it);
        }

        m_Context->ResetTextSizeCache();
    }

    RecalSize();
    Layout();
    Refresh();
}

void CTextItemPanel::InsertItems(const CCompositeTextItem::TInsertBlockVec& insertBlocks)
{
    CCompositeTextItem* rootContainer = dynamic_cast<CCompositeTextItem*>(m_Item.get());
    if (!rootContainer)
        return;

    {
        wxBusyCursor wait;
        wxClientDC dc(this);
        InitDC(dc);

        m_Context->ResetTextSizeCache();

        ITERATE(CCompositeTextItem::TInsertBlockVec, it, insertBlocks) {
            const vector<ITextItem*>& items = it->m_Items;
            if (!items.empty()) {
                CCompositeTextItem* container = it->m_Container ? it->m_Container : rootContainer;
                int index = it->m_Index < 0 ? container->GetItemCount() + it->m_Index + 1 : it->m_Index;
                ITERATE(vector<ITextItem*>, itItem, items)
                    (*itItem)->CalcSize(dc, m_Context.get());
                container->InsertItems(index, &items[0], items.size());
            }
        }
        
        m_Context->ResetTextSizeCache();
    }

    Layout();
    Refresh();
}

void CTextItemPanel::RemoveItem (size_t index)
{
    CCompositeTextItem* composite = dynamic_cast<CCompositeTextItem*>(m_Item.get());
    if (!composite)
        return;

    composite->RemoveItem (index);

    Layout();
    Refresh();
}

void CTextItemPanel::OnEraseBackground(wxEraseEvent& WXUNUSED(event))
{
}

void CTextItemPanel::OnPaint(wxPaintEvent& WXUNUSED(event))
{
    wxBufferedPaintDC dc(this);
    InitDC(dc);

// Fill background
    wxRect updateRect;
    GetUpdateRegion().GetBox(updateRect.x, updateRect.y, updateRect.width, updateRect.height);
    dc.DrawRectangle(updateRect);

    if (m_Item.get() == 0)
        return;

    CalcUnscrolledPosition(updateRect.x,updateRect.y,&updateRect.x, &updateRect.y);
    wxPoint origin;
    CalcScrolledPosition(0,0,&origin.x, &origin.y);
    dc.SetDeviceOrigin(origin.x, origin.y);
    try {
        m_Item->Draw(dc, updateRect, m_Context.get());
    }
    catch (const exception&) {
    }
}

void CTextItemPanel::InitDC(wxDC& dc)
{
    wxColor bkgdColor = GetBackgroundColour();
    if (!bkgdColor.Ok())
        bkgdColor = wxSystemSettings::GetColour(wxSYS_COLOUR_WINDOW);

    dc.SetBrush(wxBrush(bkgdColor));
    dc.SetPen(wxPen(bkgdColor,1));
    dc.SetTextBackground(bkgdColor);
    dc.SetFont(GetFont());
}

void CTextItemPanel::OnMouseEvent(wxMouseEvent& event)
{
    if (m_TooltipWindow && event.GetEventType() != wxEVT_LEAVE_WINDOW) {
        m_TooltipWindow->Destroy();
        m_TooltipWindow = 0;
    }

    m_MousePos = event.GetPosition();

    if (m_Item.get() != 0 && m_Context.get() != 0) {
        wxMouseEvent ev(event);
        CalcUnscrolledPosition(event.m_x, event.m_y, &ev.m_x, &ev.m_y);
        m_Context->ClearMouseMove();
        m_Item->MouseEvent(ev, *m_Context);

        if (event.Moving())
            TriggerTooltip();
        //wxRect rect = m_Item->GetRect();
        //if (rect.Contains(ev.m_x, ev.m_y))
            //m_Item->ProcessEvent(ev);
        //else if (event.GetEventType() == wxEVT_MOTION)
            //SetCursor(wxCursor(wxCURSOR_ARROW));
    }

    event.Skip();
}

void CTextItemPanel::SetCaretPos(int x, int y)
{
    CalcScrolledPosition(x,y,&x,&y);
    GetCaret()->Move(x, y);
}

void CTextItemPanel::UpdateCaretPos()
{
    if (m_Item.get() == 0)
        return;

    CTextPosition caretPos = m_TextSelection.GetEnd();

    CTextItem* textItem = dynamic_cast<CTextItem*>(m_Item->GetItemByLine(caretPos.GetRow(), false));
    if (textItem == 0)
        return;

    wxClientDC dc(this);
    InitDC(dc);

    caretPos.ShiftRow(-textItem->GetStartLine());
    textItem->TextPosToCollapsed(caretPos, m_Context.get());
    CCalcCaretPosOStream ostr(dc, m_Context->GetLeftMargin(), 0,
                              dc.GetCharHeight(), m_Context->GetIndentWidth(),
                              caretPos.GetCol(), caretPos.GetRow());

    textItem->RenderText(ostr, m_Context.get());
    int x, y;
    ostr.GetCaretPos(x, y);
    wxPoint itemPos = textItem->GetPosition();
    CalcScrolledPosition(x + itemPos.x, y + itemPos.y, &x, &y);
    GetCaret()->Move(x, y);
}

static
WX_DEFINE_MENU(kPopupMenu)
    WX_MENU_SEPARATOR_L("Top Actions")
    WX_MENU_SEPARATOR_L("Text Commands")
    WX_MENU_ITEM(wxID_COPY)
    WX_MENU_ITEM(wxID_SELECTALL)
WX_END_MENU()

void CTextItemPanel::OnSetFocus(wxFocusEvent& WXUNUSED(event))
{
    GetCaret()->Show();

    CTextWidgetEvent evt(CEvent::eEvent_Message, CTextWidgetEvent::eGotFocus);
    Send(&evt, ePool_Parent);
}

void CTextItemPanel::OnKillFocus(wxFocusEvent& WXUNUSED(event))
{
    GetCaret()->Hide();

    CTextWidgetEvent evt(CEvent::eEvent_Message, CTextWidgetEvent::eLostFocus);
    Send(&evt, ePool_Parent);
}

void CTextItemPanel::OnContextMenu(wxContextMenuEvent& event)
{
    if (m_Item.get() == 0)
        return;

    wxPoint point = event.GetPosition();
    // If from keyboard
    if (point.x == -1 && point.y == -1) {
        wxSize size = GetSize();
        point.x = size.x / 2;
        point.y = size.y / 2;
    }
    else {
        point = ScreenToClient(point);
    }

    CUICommandRegistry& cmd_reg = CUICommandRegistry::GetInstance();
    wxMenu* menu = cmd_reg.CreateMenu(kPopupMenu);

    if (m_Context.get()) {
        auto_ptr<wxMenu> contextMenu(m_Context->CreateMenu());
        if (contextMenu.get()) {
            Merge(*menu, *contextMenu);
        }
    }

    // send "Decorate" event to listeners
    CPopupMenuEvent evt(this, menu);
    Send(&evt, eDispatch_Default, ePool_Parent);

    menu = evt.GetMenu();

    // obtain the modified menu and clean it empty sections
    CleanupSeparators(*menu);

    // connect contributed wxEvtHandlers so that they can handle commands
    vector<wxEvtHandler*>& handlers = evt.GetHandlers();
    for( size_t  i = 0;  i < handlers.size();  i++  )   {
        PushEventHandler(handlers[i]);
    }

    // show the menu
    PopupMenu(menu, point);
    delete menu;

    // disconnect and destroy the handlers
    for( size_t  i = 0;  i < handlers.size();  i++  )   {
        wxEvtHandler* h = PopEventHandler();
        _ASSERT(h == handlers[i]);
        delete h;
    }
}

void CTextItemPanel::OnCopy(wxCommandEvent& WXUNUSED(event))
{
    if (m_Item.get() == 0 || !IsTextSelected())
        return;

    if (wxTheClipboard->Open()) {
        wxStringOutputStream stream;
        wxTextOutputStream os(stream);
        m_Item->GetText(os, m_TextSelection.Normalized(), m_Context.get());
        wxTheClipboard->SetData( new wxTextDataObject(stream.GetString()));
        wxTheClipboard->Close();
    }
}

void CTextItemPanel::OnUpdateCopy(wxUpdateUIEvent& event)
{
    event.Enable(IsTextSelected());
}

void CTextItemPanel::OnSelectAll(wxCommandEvent& WXUNUSED(event))
{
    if (m_Item.get() == 0)
        return;

    SetSelection(CTextBlock(0, m_Item->GetLineCount()));
}

void CTextItemPanel::ScrollToBottom()
{
    int w, h, cw, ch;
    GetClientSize(&cw, &ch);
    GetVirtualSize(&w, &h);
    if (h <= ch) return;

    int xUnit, yUnit;
    GetScrollPixelsPerUnit(&xUnit, &yUnit);
    Scroll(0, (h - ch)/yUnit);
/*
    GetViewStart(&x, &y);
    CalcUnscrolledPosition(x, y, &x, &y);

    CalcScrolledPosition(x, h - ch, &x, &y);
    Scroll(x, y);
 */
}

bool CTextItemPanel::FindText(const wxString& findText, bool forward)
{
    string text = ToStdString(findText);
    CTextBlock selNorm = GetTextSelection().Normalized();
    CTextPosition cursorPos = forward ? selNorm.GetEnd() : selNorm.GetStart();

    for (CTextLineIterator it(this, cursorPos, forward); it; ++it) {
        string line = *it;
        if (line.length() >= text.length()) {
            size_t pos = line.find(text);
            if (pos != string::npos) {
                CTextPosition start = it.GetPosition();
                start.ShiftCol((int)pos);
                CTextPosition end(start);
                end.ShiftCol((int)text.length());
                MakePosVisible(start);
                SetSelection(CTextBlock(start, end));
                return true;
            }
        }
    }
    return false;
}

void CTextItemPanel::MakePosVisible(const CTextPosition& pos)
{
    CTextItem* posItem = dynamic_cast<CTextItem*>(m_Item->GetItemByLine(pos.GetRow(), true));
    if (posItem == 0)
        return;

    for (ITextItem* item = posItem; item;) {
        IContainer* container = 0;
        int index = 0;
        item->GetContainer(container, index);
        CExpandItem* expandItem = dynamic_cast<CExpandItem*>(container);
        if (expandItem && !expandItem->IsExpanded())
            expandItem->Toggle(m_Context.get());
        item = dynamic_cast<ITextItem*>(container);
    }

    int w, h, cw, ch;
    GetClientSize(&cw, &ch);
    GetVirtualSize(&w, &h);
    if (h <= ch) return;

    wxPoint itemPos = posItem->GetPosition();

    wxClientDC dc(this);
    InitDC(dc);

    CCalcCaretPosOStream ostr(dc, m_Context->GetLeftMargin(), 0,
                              dc.GetCharHeight(), m_Context->GetIndentWidth(),
                              pos.GetCol(), pos.GetRow() - posItem->GetStartLine());

    posItem->RenderText(ostr, m_Context.get());
    int x, y;
    ostr.GetCaretPos(x, y);
    int yN = y + itemPos.y;
    CalcScrolledPosition(x + itemPos.x, y + itemPos.y, &x, &y);

    int xUnit, yUnit;
    GetScrollPixelsPerUnit(&xUnit, &yUnit);

    //if (y + ch > h)
        //y = h - ch;
    if (y < 0)
        Scroll(0, yN/yUnit);
    else if (y + yUnit > ch)
        Scroll(0, (yN + 2*yUnit - 1 - ch)/yUnit);
}


void CTextItemPanel::TriggerTooltip()
{
    if (m_TooltipWindow) {
        m_TooltipWindow->Destroy();
        m_TooltipWindow = 0;
    }

    m_Timer.Stop();
    m_Timer.Start(500, wxTIMER_ONE_SHOT);
}

void CTextItemPanel::OnTimer(wxTimerEvent& WXUNUSED(event))
{
    if (m_Context.get() == 0)
        return;

    if (m_Timer.IsRunning())
        return;

    m_TooltipWindow = m_Context->CreateTooltipWindow();
    if (m_TooltipWindow) {
        wxSize tooltipSize = m_TooltipWindow->GetSize();
        wxSize clientSize = GetClientSize();

        wxPoint pos = m_MousePos;

        if (pos.y + tooltipSize.GetHeight() > clientSize.GetHeight() &&
            pos.y - tooltipSize.GetHeight() >= 0) {
            pos.y -= tooltipSize.GetHeight();
        }

        if (pos.x + tooltipSize.GetWidth() > clientSize.GetWidth()) {
            pos.x = max(0, clientSize.GetWidth() - tooltipSize.GetWidth());
        }

        m_TooltipWindow->Move(pos.x, pos.y);
        m_TooltipWindow->Show();
    }
}


END_NCBI_SCOPE
