/*  $Id: async_call.cpp 26643 2012-10-18 17:54:05Z katargir $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Roman  Katargin
 *
 */

#include <ncbi_pch.hpp>

#include <corelib/ncbistd.hpp>

#include <gui/utils/app_job_dispatcher.hpp>
#include <gui/widgets/wx/async_call.hpp>
#include <gui/widgets/wx/ind_progress_bar.hpp>

#include <wx/utils.h>
#include <wx/dialog.h>
#include <wx/timer.h>
#include <wx/sizer.h>
#include <wx/stattext.h>
#include <wx/button.h>
#include <wx/thread.h>
#include <wx/app.h>

BEGIN_NCBI_SCOPE

class CAsyncCallDlg : public wxDialog
{
public:
    CAsyncCallDlg(wxWindow* parent, const wxString& text, int jobId, bool CanCancel);
    bool Create(wxWindow* parent, bool CanCancel);
    void CreateControls(bool CanCancel);

    DECLARE_EVENT_TABLE()

    void OnTimer(wxTimerEvent& event);
    void OnCancelClick(wxCommandEvent& event);

private:
    wxString m_Text;
    int m_JobId;
    wxTimer m_Timer;

    enum {
        ID_STATIC_TEXT = 10001,
        ID_BUTTON_CANCEL = 10002
    };
};

class CExecJob : public CObject, public IAppJob
{
public:
    CExecJob (CAsyncCall::Call* call)
      : m_Call(call), m_Ex1(NULL), m_Ex2(NULL) {}
    ~CExecJob();
    virtual EJobState   Run();
    virtual CConstIRef<IAppJobProgress> GetProgress() { return CConstIRef<IAppJobProgress>(); }
    virtual CRef<CObject>   GetResult() { return CRef<CObject>(); }
    virtual CConstIRef<IAppJobError>    GetError() { return CConstIRef<IAppJobError>(); }
    virtual string    GetDescr() const { return NcbiEmptyString; }
    virtual void RequestCancel() { m_Call->RequestCancel(); }
    virtual bool IsCanceled() const { return m_Call->IsCanceled(); }

    void Finish();

private:
    CAsyncCall::Call* m_Call;

    CException* m_Ex1;
    std::exception* m_Ex2;
};

IAppJob::EJobState CExecJob::Run()
{
    try {
        m_Call->Execute();
        return eCompleted;
    } catch( const CException& e ) {
        m_Ex1 = new CException(e);
    } catch( const std::exception& e ) {
        m_Ex2 = new std::exception(e);
    }
    return eFailed;
}

CExecJob::~CExecJob()
{
    delete m_Ex1, delete m_Ex2;
}

void CExecJob::Finish()
{
    if (m_Ex1) throw *m_Ex1;
    if (m_Ex2) throw *m_Ex2;
}


CAsyncCall::CAsyncCall(Call& call)
: m_Call(call), m_JobId(CAppJobDispatcher::eInvalidJobID)
{
}

void CAsyncCall::Execute()
{
    if( !::wxIsMainThread() ){
        m_Call.Execute();
        return;
    }

    CRef<CExecJob> job = CRef<CExecJob>( new CExecJob( &m_Call ) );

    CAppJobDispatcher& disp = CAppJobDispatcher::GetInstance();
    m_JobId = disp.StartJob( *job, "ObjManagerEngineAsyncCall" );
    if( m_JobId == CAppJobDispatcher::eInvalidJobID ){
        return;
    }

    for( unsigned long misec = 0; misec < 1000; misec=misec*2+1 ){
        ::wxMilliSleep( misec ); // in case of 0 it works like Yield()

        IAppJob::EJobState state = disp.GetJobState( m_JobId );
        if( state != IAppJob::eRunning && state != IAppJob::eSuspended ){
            //LOG_POST( Trace << "UI waiting job is finished by " << misec << " ms step." );
            job->Finish();
            return;
        }
    }

    wxWindow* p = wxWindow::FindFocus();
    while( p && !p->IsTopLevel() ){
        p = p->GetParent();
    }
    wxWindow* pTopWnd = p;

    if( pTopWnd == NULL ){
        pTopWnd = wxTheApp->GetTopWindow();
    }

    CAsyncCallDlg( pTopWnd, m_Call.GetDescr(), m_JobId, m_Call.CanCancel() ).ShowModal();
    job->Finish();
}

void CAsyncCall::Execute(Call& call)
{
    CAsyncCall asyncCall(call);
    asyncCall.Execute();
}

void CAsyncCall::ExecuteYield()
{
    if( !::wxIsMainThread() ){
        m_Call.Execute();
        return;
    }

    CRef<CExecJob> job = CRef<CExecJob>( new CExecJob( &m_Call ) );

    CAppJobDispatcher& disp = CAppJobDispatcher::GetInstance();
    m_JobId = disp.StartJob( *job, "ObjManagerEngineAsyncCall" );
    if( m_JobId == CAppJobDispatcher::eInvalidJobID ){
        return;
    }

    IAppJob::EJobState state = disp.GetJobState( m_JobId );

    while (state == IAppJob::eRunning || state == IAppJob::eSuspended) {
        ::wxMilliSleep( 100 );
      
        // This gives the opportunity to interact with a cancel button.
        // The user of this class should disable any interface elements
        // that allow the user to modify/access data in a way that conflicts
        // with the runnnig job.
        wxTheApp->Yield(true);

        state = disp.GetJobState( m_JobId );
        if( state != IAppJob::eRunning && state != IAppJob::eSuspended ){
            //LOG_POST( Trace << "UI waiting job is finished by " << misec << " ms step." );
            job->Finish();
        }
    }
}

CAsyncCall::~CAsyncCall()
{
    if (m_JobId != CAppJobDispatcher::eInvalidJobID) {
        CAppJobDispatcher::GetInstance().DeleteJob(m_JobId);
    }
}

BEGIN_EVENT_TABLE( CAsyncCallDlg, wxDialog )
    EVT_TIMER(-1, CAsyncCallDlg::OnTimer)
    EVT_BUTTON( ID_BUTTON_CANCEL, CAsyncCallDlg::OnCancelClick)
END_EVENT_TABLE()

CAsyncCallDlg::CAsyncCallDlg(wxWindow* parent, const wxString& text, int jobId, bool CanCancel)
: m_Text(text), m_JobId(jobId), m_Timer(this)
{
    Create(parent, CanCancel);
}

bool CAsyncCallDlg::Create(wxWindow* parent, bool CanCancel)
{
    wxDialog::Create( parent, wxID_ANY, wxEmptyString, wxDefaultPosition, wxDefaultSize, 0 );

    SetBackgroundStyle(wxBG_STYLE_COLOUR);
    SetBackgroundColour(wxColor(wxT("#FFE793")));
    //SetForegroundColour(wxColor(wxT("#A68513")));

    CreateControls(CanCancel);

    if (GetSizer())
    {
        GetSizer()->SetSizeHints(this);
    }

    //Centre();

    wxRect parentRect = parent->GetScreenRect();
    wxRect rect = GetScreenRect();
    Move(parentRect.GetRight() - rect.GetWidth() - 2, parentRect.GetBottom() - rect.GetHeight() - 2);

    return true;
}

void CAsyncCallDlg::CreateControls(bool CanCancel)
{
    CAsyncCallDlg* itemDialog1 = this;
    wxBoxSizer* itemBoxSizer1 = new wxBoxSizer(wxVERTICAL);
    itemDialog1->SetSizer(itemBoxSizer1);
    wxStaticText* itemStaticText3 = new wxStaticText( itemDialog1, ID_STATIC_TEXT, m_Text, wxDefaultPosition, wxDefaultSize, 0 );
    wxFont font = itemStaticText3->GetFont();
    font.SetWeight(wxFONTWEIGHT_BOLD);
    itemStaticText3->SetFont(font);
    itemStaticText3->SetForegroundColour(*wxBLACK);
    itemBoxSizer1->Add(itemStaticText3, 0, wxALIGN_LEFT|wxTOP|wxLEFT|wxRIGHT, 5);

    wxBoxSizer* itemBoxSizer2 = new wxBoxSizer(wxHORIZONTAL);
    itemBoxSizer1->Add(itemBoxSizer2);

    CIndProgressBar* progressBar = new CIndProgressBar(itemDialog1, wxID_ANY, wxDefaultPosition, (CanCancel ? 160 : 240));
    itemBoxSizer2->Add(progressBar, 0, wxALIGN_LEFT|wxALIGN_CENTER_VERTICAL|wxALL, 5);

    if (CanCancel) {
        wxButton* itemButton1 = new wxButton( itemDialog1, ID_BUTTON_CANCEL, _("Cancel"), wxDefaultPosition, wxDefaultSize );
        itemButton1->SetFocus();
        itemBoxSizer2->Add(itemButton1, 0, wxALIGN_LEFT|wxALL, 5);
   } 

    m_Timer.Start(300);
}

void CAsyncCallDlg::OnTimer(wxTimerEvent&)
{
    CAppJobDispatcher& disp = CAppJobDispatcher::GetInstance();
    IAppJob::EJobState state = disp.GetJobState(m_JobId);
    if (state != IAppJob::eRunning && state != IAppJob::eSuspended) {
        m_Timer.Stop();
        EndModal(wxID_CANCEL);
    }
}

void CAsyncCallDlg::OnCancelClick(wxCommandEvent&)
{
    CAppJobDispatcher& disp = CAppJobDispatcher::GetInstance();
    disp.CancelJob(m_JobId);
    FindWindow(ID_BUTTON_CANCEL)->Enable(false);
}

END_NCBI_SCOPE
