/*  $Id: dock_layout.cpp 24019 2011-07-08 13:45:08Z wuliangs $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Andrey Yazhuk
 *
 * File Description:
 *
 */
#include <ncbi_pch.hpp>

#include <gui/widgets/wx/dock_layout.hpp>
#include <gui/widgets/wx/wx_utils.hpp>
#include <gui/widgets/wx/splitter.hpp>
#include <gui/widgets/wx/dock_manager.hpp>

#ifdef _DEBUG
    #include <gui/widgets/wx/dock_notebook.hpp>
#endif


BEGIN_NCBI_SCOPE

///////////////////////////////////////////////////////////////////////////////
/// CDockLayoutTree::CNode

CDockLayoutTree::CNode::CNode()
:   m_Type(eClient),
    m_Hidden(false),
    m_Window(NULL)
{
}


CDockLayoutTree::CNode::CNode(const CNode& orig)
:   m_Type(orig.m_Type),
    m_Hidden(orig.m_Hidden),
    m_Clone(NULL),
    m_Window(orig.m_Window)
{
}


CDockLayoutTree::CNode::CNode(ENodeType type, wxWindow* window, bool hidden)
:   m_Type(type),
    m_Hidden(hidden),
    m_Clone(NULL),
    m_Window(window)
{
}


CDockLayoutTree::CNode::~CNode()
{
}


CRef<CDockLayoutTree::CNode> CDockLayoutTree::CNode::GetTheOnlyChild()
{
    if (m_Children.size() == 1) {
        return m_Children[0];
    } else {
        _ASSERT(false);
        return CRef<CNode>();
    }
}


CRef<CDockLayoutTree::CNode> CDockLayoutTree::CNode::GetVisibleParent()
{
    CNode* p = m_Parent.GetPointer();
    while(p) {
        if (p->IsVisible())
            break;
        p = p->m_Parent;
    }
    return CRef<CNode>(p);
}


CRef<CDockLayoutTree::CNode> CDockLayoutTree::CNode::GetTopHiddenParent()
{
    CNode* p = m_Parent.GetPointer();
    CNode* h_p = NULL;
    while(p) {
        if (p->IsVisible()) {
            break;
        }
        if (p->IsHidden()) {
            h_p = p; // remeber hidden parent
        }
        p = p->m_Parent;
    }
    return CRef<CNode>(h_p);
}


// retruns a node that represents the topmost parent node in full tree
// that has the node as its only visible chils
CRef<CDockLayoutTree::CNode> CDockLayoutTree::CNode::GetBranchTop()
{
    CRef<CNode> top_h_p = GetTopHiddenParent();
    return top_h_p ? top_h_p : CRef<CNode>(this);
}


void CDockLayoutTree::CNode::SetWindow(wxWindow* window)
{
#ifdef _DEBUG
    if (window) {
        //LOG_POST("CNode::SetWindow() " << window << "  " << typeid(*window).name());

        bool splitter = (dynamic_cast<CSplitter*>(window) != NULL);
        bool tab = (dynamic_cast<CDockNotebook*>(window) != NULL);

        switch(m_Type) {
        case eTab:
            _ASSERT(tab);
            break;
        case eHorzSplitter:
        case eVertSplitter:
            _ASSERT(splitter);
            break;
        default:
            _ASSERT(! tab  &&  ! splitter);
        }
    }
#endif

    m_Window = window;
}


void CDockLayoutTree::CNode::AddChild(CNode& node)
{
    m_Children.push_back(CRef<CNode>(&node));
    node.m_Parent.Reset(this);
}


void CDockLayoutTree::CNode::InsertChild(size_t index, CNode& node)
{
    _ASSERT(index <= m_Children.size());
    m_Children.insert(m_Children.begin() + index, CRef<CNode>(&node));
    node.m_Parent.Reset(this);
}


void CDockLayoutTree::CNode::RemoveChild(CNode& node)
{
    CRef<CNode> ref(&node);
    TNodeVector::iterator it = std::find(m_Children.begin(), m_Children.end(), ref);

    _ASSERT(it != m_Children.end());

    m_Children.erase(it);
    node.m_Parent.Reset();
}


void CDockLayoutTree::CNode::RemoveChild(size_t index)
{
    _ASSERT(index < m_Children.size());

    TNodeVector::iterator it = m_Children.begin() + index;
    (*it)->m_Parent.Reset();
    m_Children.erase(it);
}


void CDockLayoutTree::CNode::RemoveAllChildren()
{
    NON_CONST_ITERATE(TNodeVector, it, m_Children) {
        CNode& node = **it;
        node.m_Parent.Reset();
    }
    m_Children.clear();
}


size_t CDockLayoutTree::CNode::ReplaceChild(CNode& old_node, CNode& new_node)
{
    CRef<CNode> ref(&old_node);
    TNodeVector::iterator it = std::find(m_Children.begin(), m_Children.end(), ref);

    _ASSERT(it != m_Children.end());

    it->Reset(&new_node);

    old_node.m_Parent.Reset();
    new_node.m_Parent.Reset(this);
    return it - m_Children.begin();
}


void CDockLayoutTree::CNode::HideChild(wxWindow& window)
{
    for( size_t i = 0;  i < m_Children.size();  i++ ) {
        CNode& node = *m_Children[0];
        if (node.m_Window == &window) {
            node.m_Hidden = true;
            return;
        }
    }
    _ASSERT(false);
}


// rearrange m_Children according to the given order of windows
void CDockLayoutTree::CNode::EnforceChildOrder(const vector<wxWindow*>& windows)
{
    _ASSERT(windows.size() <= m_Children.size());
    // here we accumulate Nodes in the right order
    TNodeVector sel;

    for( size_t i = 0;  i < windows.size(); i++ ) {
        wxWindow* win = windows[i];
        // find a child for this window
        for( size_t j = 0;  j < m_Children.size(); j++ ) {
            CRef<CNode>& child = m_Children[j];
            if (child->GetWindow() == win) {
                sel.push_back(child);
                m_Children.erase(m_Children.begin() + j);
                break;
            }
        }
    }
    _ASSERT(sel.size() == windows.size());

    for( size_t i = 0;  i < sel.size(); i++ ) {
        CRef<CNode>& child = sel[i];
        m_Children.insert(m_Children.begin() + i, child);
    }
}


void CDockLayoutTree::CNode::Link(CNode& clone)
{
    m_Clone.Reset(&clone);
    clone.m_Clone.Reset(this);
}


void CDockLayoutTree::CNode::Unlink()
{
    if (m_Clone) {
        m_Clone->m_Clone.Reset();
        m_Clone.Reset();
    }
}


int CDockLayoutTree::CNode::GetIndex(CNode& node) const
{
    CRef<CNode> ref(&node);
    TNodeVector::const_iterator it = std::find(m_Children.begin(), m_Children.end(), ref);
    return (it == m_Children.end()) ? -1 : (it - m_Children.begin());
}


// find index of a child node that has the given node as one of its children
// / grandchildren (on any level). In other words find a child node that is
// the root of a branch containing the given node.
int CDockLayoutTree::CNode::GetBranchIndex(CNode& grandchild_node) const
{
    CNode* p = &grandchild_node;
    while(p) {
        if (p->m_Parent == this) {
            // this is our child
            int index = GetIndex(*p);
            _ASSERT(index >= 0);
            return index;
        }
        p = p->GetParent();
    }
    return -1; // not our grandchild
}

bool CDockLayoutTree::CNode::HasOnlyOneChild() const
{
    return m_Children.size() == 1;
}


void CDockLayoutTree::CNode::GetVisibleChildren(TNodeVector& children)
{
    for( size_t i = 0;  i < m_Children.size();  i++ ) {
        CRef<CNode>& child_node = m_Children[i];
        if (child_node->IsVisible()) {
            children.push_back(child_node);
        }
    }
}

static string kWindowLayoutHdr = "Window Layout: ";

void CDockLayoutTree::CNode::ValidateSplitterSizes()
{
    if (m_Type != eHorzSplitter && m_Type != eVertSplitter)
        return;

    if (m_SplitterSizes.size() == 0)
        return;

    if (m_SplitterSizes.size() != 2) {
        LOG_POST(Error << kWindowLayoutHdr << "Invalid dimension of splitter sizes: " << m_SplitterSizes.size());
        m_SplitterSizes.resize(2);
        m_SplitterSizes[0] = m_SplitterSizes[1] = 100;
        return;
    }

    if (m_SplitterSizes[0] <= 0 || m_SplitterSizes[1] <= 0) {
        LOG_POST(Error << kWindowLayoutHdr << "Invalid splitter sizes: ("
            << m_SplitterSizes[0] << ", " << m_SplitterSizes[1] << ")");

        m_SplitterSizes[0] = m_SplitterSizes[1] = 100;
    }
}

void CDockLayoutTree::CNode::LogPost(const string& offset) const
{
    if (m_Clone) {
        _ASSERT(m_Clone->m_Clone  == this);
    }
    static string   types[4] = { "(client)              ",
                                 "(horizontal splitter) ",
                                 "(vertical splitter)   ",
                                 "(tab)                 "  };

    string rect;
    if (GetWindow()) {
        wxRect rc = ncbi::GetScreenRect(*GetWindow());
        rect = ToString(rc);
    }

    LOG_POST(Info << offset
             << "Node " << this << " " << types[m_Type]
             << ", parent " << m_Parent
             << (m_Hidden ? " HIDDEN" : " VISIBLE")
             << ", win " <<  m_Window
             << ", hash " << m_Fingerprint.AsPersistentString()
             //<< "  " << rect
             //<< " Last Size " << m_LastSize.x << ", " << m_LastSize.y
            );
}


///////////////////////////////////////////////////////////////////////////////
/// CDockLayoutTree
CDockLayoutTree::CDockLayoutTree(CNode* root, CNode* central_pane)
:   m_Root(root),
    m_CentralPane(central_pane)
{
    if (m_Root) {
        x_AddNodesToMap(*m_Root);
    }
}


CDockLayoutTree::~CDockLayoutTree()
{
    DisconnectAllNodes();
}


void CDockLayoutTree::DisconnectAllNodes()
{
    if (m_Root) {
        x_DisconnectNodes(*m_Root);
        m_Root.Reset();
    }
}


// disconnect all nodes from each other, this most likely will result in nodes
// being deleted (unless someone hold a reference to a node)
void CDockLayoutTree::x_DisconnectNodes(CNode& node)
{
    for( size_t i = 0;  i < node.m_Children.size(); i++) {
        CNode& child = *node.m_Children[i];
        x_DisconnectNodes(child);
    }
    node.RemoveAllChildren();
    node.Unlink();
}


void CDockLayoutTree::HideAll()
{
    if (m_Root) {
        x_HideNodes(*m_Root);
    }
}


CDockLayoutTree::CNode* CDockLayoutTree::FindNodeByWindow(wxWindow* window)
{
    TWindowToNode::iterator it = m_WindowToNode.find(window);
    return (it == m_WindowToNode.end()) ? NULL : it->second;
}


CDockLayoutTree::CNode* CDockLayoutTree::FindByFingerprint(const TFingerprint& s)
{
    if (m_Root) {
        return x_FindByFingerprint(*m_Root, s);
    } else {
        return NULL;
    }
}


CDockLayoutTree::CNode*
    CDockLayoutTree::x_FindByFingerprint(CNode& node, const TFingerprint& s)
{
    if (node.GetFingerprint() == s) {
        return &node;
    } else if (node.IsContainer()) {
        for( size_t i = 0;  i < node.m_Children.size();  i++ ) {
            CNode& child = *node.m_Children[i];
            CNode* res = x_FindByFingerprint(child, s);
            if (res) {
                return res;
            }
        }
    }
    return NULL;
}

void CDockLayoutTree::GetAllWindows(vector<wxWindow*>& windows)
{
    NON_CONST_ITERATE(TWindowToNode, it, m_WindowToNode) {
        windows.push_back(it->first);
    }
}


void CDockLayoutTree::LogPost() const
{
    if (m_Root) {
        LogPost(*m_Root, "    ");
    } else {
        LOG_POST(Info << "    Root is NULL");
    }

    LOG_POST(Info << "\n    Central Pane " << m_CentralPane.GetPointer());

    if (m_CentralPane) {
        m_CentralPane->LogPost("     ");
    }
}


void CDockLayoutTree::LogPost(const CNode& node, const string& offset) const
{
    node.LogPost(offset);

    if ( ! node.m_Children.empty()) {
        string off = offset + "\t";

        for( size_t i = 0;  i < node.m_Children.size();  i++) {
            const CNode* child = node.m_Children[i];
            LogPost(*child, off);
        }
    }
}


void CDockLayoutTree::Validate(const CNode& node, bool vis_tree) const
{
    // must have a parent or be the root of the tree
    _ASSERT(node.GetParent()  ||  m_Root.GetPointer() == &node);

    // visible nodes must have windows
    _ASSERT(node.IsVisible()  ==  (node.GetWindow()  != NULL));

    if (node.IsVisible()) {
        // visible nodes must have clones
        _ASSERT(node.m_Clone  &&  node.m_Clone->m_Clone == &node);
    } else {
        _ASSERT( ! vis_tree  &&  node.m_Clone == NULL);
    }

    if (node.IsContainer()) {
        // must be root tab or non empty
        _ASSERT(m_Root.GetPointer() == &node  ||  node.m_Children.size());

        for( size_t i = 0;  i < node.m_Children.size();  i++) {
            _ASSERT(node.m_Children[i]->m_Parent == &node); // parent-child consistency
        }
    }
}


void CDockLayoutTree::x_AddNodesToMap(CNode& node)
{
    if (node.GetWindow()) {
        m_WindowToNode[node.GetWindow()] = &node;
    }
    if (node.IsContainer()) {
        for( size_t i = 0;  i < node.m_Children.size();  i++ ) {
            CNode& child_node = *node.m_Children[i];
            x_AddNodesToMap(child_node);
        }
    }
}


void CDockLayoutTree::x_HideNodes(CNode& node)
{
    node.SetHidden(true);

    if (node.IsContainer()) {
        for( size_t i = 0;  i < node.m_Children.size();  i++ ) {
            CNode& child_node = *node.m_Children[i];
            x_HideNodes(child_node);
        }
    }
}


END_NCBI_SCOPE
