/*  $Id: glcanvas.cpp 27992 2013-05-06 20:30:27Z rafanovi $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Roman Katargin
 *
 * File Description:
 */

#include <ncbi_pch.hpp>

#include <gui/widgets/wx/glcanvas.hpp>

#include <gui/opengl.h>
#include <gui/opengl/glutils.hpp>
#include <gui/widgets/wx/message_box.hpp>

#include <wx/dcclient.h>



BEGIN_NCBI_SCOPE

/// Not having all programs share the same window attributes has caused
/// crashing on Linux, so each window will use these attributes (this should
/// be the superset of attributes needed by program windows).
int g_attriblist[8] = {WX_GL_RGBA, WX_GL_DOUBLEBUFFER, WX_GL_STENCIL_SIZE, 8, 0 };

#ifndef __WXMAC_CARBON__

class CGLContextCache
{
public:
    static CGLContextCache& GetInstance();
    void AddRef();
    void Release();
    wxGLContext* GetContext(wxGLCanvas* win);
    GLEWContext* GetGlewContext() { return &m_GLEWContext; }

private:
    CGLContextCache() : m_GLContext(0), m_RefCnt(0) {}
    ~CGLContextCache()
    { 
        // We could try to delete the context here but, at least on windows,
        // by the time this dtor is called, the memory is already cleand up
        // so an exception is thrown when we try to delete the (already deleted)
        // wxGLContext
    }

private:
    wxGLContext* m_GLContext;
    int m_RefCnt;
    GLEWContext  m_GLEWContext;
};


CGLContextCache& CGLContextCache::GetInstance()
{
    // Declare the static instance (singleton)
    static CGLContextCache instance;

    // push the address of the glew context to the static instance of CGLGlewContext
    // since that is in the base opengl library which needs the (glew) context, but
    // does not have any dependencies on wxWidgets.
    if (CGLGlewContext::GetInstance().GetGlewContext() == NULL)
        CGLGlewContext::GetInstance().SetGlewContext(&instance.m_GLEWContext);

    return instance;
}

void CGLContextCache::AddRef()
{
    ++m_RefCnt;
}

void CGLContextCache::Release()
{
    --m_RefCnt;
    if (m_RefCnt == 0) {
        // Keep context - we share one context for all windows and will
        // reuse it when the next window opens.  Also, some classes (e.g. 
        // CGlTextureFont) may save gl objects like texture maps (fonts) 
        // between window instances so context must stay alive.
    }
}

wxGLContext* CGLContextCache::GetContext(wxGLCanvas* win)
{
    if (m_GLContext == 0) {
        m_GLContext = new wxGLContext(win);

        m_GLContext->SetCurrent(*win);
    
        GLenum err = glewContextInit(&m_GLEWContext);
        if (GLEW_OK != err)
        {
            // Problem: Error creating glew context
            _TRACE("Error creating glew context");
        } 
        else {
            err = glewInit();      
            if (GLEW_OK != err)
            {
                // Problem: glewInit failed, something is seriously wrong.
                //   No extenstions will be available.
                _TRACE("Error loading opengl extensions");
            }
        }
    }
    return m_GLContext;
}

#else

END_NCBI_SCOPE

GLEWContext*   NCBI_NS_NCBI::CGLCanvas::m_CurrentGLEWContext = NULL;

// GLEWContext*  glewGetContext() 
// { 
//     return  NCBI_NS_NCBI::CGLCanvas::GetGlewContext();
// }

BEGIN_NCBI_SCOPE

#endif


BEGIN_EVENT_TABLE(CGLCanvas, wxGLCanvas)
    EVT_PAINT(CGLCanvas::OnPaint)
    EVT_ERASE_BACKGROUND(CGLCanvas::OnEraseBackground)
END_EVENT_TABLE()


#ifdef __WXMAC_CARBON__
CGLCanvas::CGLCanvas(wxWindow* parent, wxWindowID id,
                     const wxPoint& pos, const wxSize& size, long style) :   
    wxGLCanvas(parent, id, pos, size, style, wxT("GLCanvas"), g_attriblist, wxNullPalette),
    m_DoNotUpdate(false)
{
    SetBackgroundStyle(wxBG_STYLE_CUSTOM);

    // Check that we have an opengl level of at least 1.4 and display
    // a warning message if we don't
/*
    static bool opengl_level_checked = false;
    if (!opengl_level_checked) {
        opengl_level_checked = true;
        x_CheckGlVersion();
    }
*/
}

CGLCanvas::~CGLCanvas()
{
}
#else
CGLCanvas::CGLCanvas(wxWindow* parent, wxWindowID id,
                     const wxPoint& pos, const wxSize& size, long style) :

    wxGLCanvas(parent, id, g_attriblist, pos, size, style),
    m_DoNotUpdate(false)
{
    CGLContextCache::GetInstance().AddRef();
    SetBackgroundStyle(wxBG_STYLE_CUSTOM);
    
    // Check that we have an opengl level of at least 1.4 and display
    // a warning message if we don't
/*
    static bool opengl_level_checked = false;
    if (!opengl_level_checked) {
        opengl_level_checked = true;
        x_CheckGlVersion();
    }
*/
}

CGLCanvas::~CGLCanvas()
{
    CGLContextCache::GetInstance().Release();
}
#endif

/// This is not used in the main app.  Some demo programs use it
/// when compiled for the mac.
bool CGLCanvas::Create(wxWindow* parent, wxWindowID id,
        const wxPoint& pos, const wxSize& size,
        long style, const wxString& name)
{
#if (wxMINOR_VERSION > 8)
    return wxGLCanvas::Create(parent, id, pos, size, style, name);
#else
#if defined __WXGTK__
    return wxGLCanvas::Create(parent, NULL, NULL, id, pos, size, style, name);
#elif defined __WXMAC_CARBON__
    return wxGLCanvas::Create(parent, NULL, id, pos, size, style, name, 0, wxNullPalette);
#else
    return wxGLCanvas::Create(parent, id, pos, size, style, name);
#endif
#endif
}

#ifdef __WXMAC_CARBON__
GLEWContext* CGLCanvas::GetGlewContext()
{
    return m_CurrentGLEWContext;
}
#endif

// Disable deprecated warning for AGLDrawable in wx code (MAC specific)
#ifdef __WXMAC_CARBON__
#pragma GCC diagnostic ignored "-Wdeprecated-declarations"
#endif

void CGLCanvas::x_CheckGlVersion()
{
    x_SetupGLContext();
    string version = (const char*)glGetString(GL_VERSION);
    vector<string> arr;

    if (version != "")
        NStr::Tokenize(version, ". \t", arr);

    bool version_passed = false;

    // Check major version
    int major_version = 0;
    if (arr.size() >= 1) {       
        try {
            major_version = NStr::StringToInt(arr[0]);
            if (major_version > 1) {
                version_passed = true;
            }
        }
        catch (CException& ) {
            //didn't parse as an int...
        }
    }

    // If major version is only 1, check minor version:
    if (arr.size() > 1 && major_version == 1) {
        try {
            int minor_version = NStr::StringToInt(arr[1]);
            if (minor_version > 4) {
                version_passed = true;
            }
        }
        catch (CException&) {
            //minor version didn't parse as an int
        }
    }

    string user_warning;
    if (!version_passed) {
        user_warning = "Warning - Your computer's graphics do not meet the\n";
        user_warning += "minimum level required by Genome Workbench. Visual or \n";
        user_warning += "stability issues may occur.\n\n";
        user_warning += "OpenGL version 1.5 or better is needed. Your graphics are:\n\n";
        user_warning += version;
        user_warning += "\n";
        user_warning += (const char*)glGetString(GL_RENDERER);

        NcbiWarningBox(user_warning, "Graphics Warning");
    }
}

void CGLCanvas::x_SetupGLContext()
{
#ifdef __WXMAC_CARBON__
    AGLDrawable drawable =
        (AGLDrawable) GetWindowPort(WindowRef(this->MacGetTopLevelWindowRef()));

    if (m_glContext->m_drawable != drawable) {
        m_glContext->m_drawable = drawable;
        aglSetDrawable(m_glContext->m_glContext, drawable);
    }
    m_glContext->SetCurrent();


    /// For GLEW support (opengl extensions), create a new GLEWContext if needed,
    /// and set the current context (whether or not a new one was created).
        
    m_CurrentGLEWContext = NULL;
    std::vector<pair<wxGLContext*, GLEWContext*> >::iterator iter;
    for (iter=m_Contexts.begin(); iter!=m_Contexts.end(); ++iter) {
        if ((*iter).first == m_glContext) {
            m_CurrentGLEWContext = (*iter).second;
        }
    }

    if (m_CurrentGLEWContext == NULL) {       
        GLenum  err = glewContextInit(&m_GLEWContext);
        if (GLEW_OK != err)
        {
            // Problem: Error creating glew context. Won't be able to get
            // to extensions
            _TRACE("Error creating glew context");
        } 
        m_CurrentGLEWContext = &m_GLEWContext;
        m_Contexts.push_back(std::pair<wxGLContext*, GLEWContext*>
                                 (m_glContext, m_CurrentGLEWContext));
    
        err = glewInit();    
        if (GLEW_OK != err)
        {
            // Problem: glewInit failed, something is seriously wrong.
            //   No extenstions will be available.
            _TRACE("Error loading opengl extensions");
        } 
    }

#else
    wxGLContext* context = CGLContextCache::GetInstance().GetContext(this);
    context->SetCurrent(*this);
#endif

    glDisable(GL_DEPTH_TEST);
    glDisable(GL_LIGHTING);

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
}

#ifdef __WXMAC_CARBON__
#pragma GCC diagnostic warning "-Wdeprecated-declarations"
#endif

void CGLCanvas::OnPaint( wxPaintEvent& WXUNUSED(event) )
{
    //LOG_POST("CGLCanvas::OnPaint()");

    // This is a dummy, to avoid an endless succession of paint messages.
    // OnPaint handlers must always create a wxPaintDC.
    wxPaintDC dc(this);

    // The hack is used for tooltip window
    // SwapBuffers ovewrites child windows on some systems
    // Ubuntu with Intel videocard (netbook).
    if (m_DoNotUpdate) {
        m_DoNotUpdate = false;
        return;
    }

    x_SetupGLContext();
    x_Render();

    // For some implementations of graphics card drivers,
    // the update of the stored image underneath the mouse
    // cursor is not sychronized with the udpate of OpenGL
    // window.  Here we force to update the stored image
    // after the frame butter gets updated.
    // Currently, This works on Windows only.
#if defined __WXMSW__
    ShowCursor(false);
#endif

    SwapBuffers();

#if defined __WXMSW__
    ShowCursor(true);
#endif
}

void CGLCanvas::x_Render()
{
    wxColour clr = GetBackgroundColour();
    glClearColor(clr.Red()/255.f, clr.Green()/255.f, clr.Blue()/255.f, 0);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
}

void CGLCanvas::OnEraseBackground( wxEraseEvent& event )
{
}


END_NCBI_SCOPE