/*  $Id: table_utils.cpp 26230 2012-08-07 21:04:16Z voronov $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors: Andrey Yazhuk
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>

#include <gui/widgets/wx/table_utils.hpp>
#include <gui/widgets/wx/wx_utils.hpp>

#include <gui/widgets/wx/csv_export_dlg.hpp>
#include <gui/widgets/wx/csv_exporter.hpp>

BEGIN_NCBI_SCOPE

// We save the following information
/// - a set of visible and invisible columns
/// - columns widths
/// - sort order

static const char* kColNumTag  = "Count";
static const char* kColNumVisTag  = "Vis-Count";
static const char* kColNumHidTag  = "Hid-Count";
static const char* kColNumModTag  = "Mod-Count";
static const char* kColumnTag   = "Column";
static const char* kColModelName  = "Name";
static const char* kColVisOrder = "Vis-Order";
static const char* kColModIndex = "Mod-Index";
static const char* kColumnWidth = "Width";
static const char* kColumnShown = "Shown";
static const char* kSortedGlobal = "Sorted-Global";
static const char* kSortedColumn = "Sorted-Col";
static const char* kSortedOrder = "Sorted-Asc";


static string GetColumnSectionName( int ix )
{
    return kColumnTag + string( "_" ) + NStr::IntToString( ix );
}

void SaveTableSettings( const CwxTableListCtrl& table, CGuiRegistry::TReadWriteView view )
{
    string error;
    try {
        IwxTableModel* model = table.GetModel();
        if( model == NULL ){
            //_ASSERT( false );

            NCBI_THROW( CException, eUnknown, "Table model is not present" );
        }

        int total_num = table.GetColumnCount();
        view.Set( kColNumTag, total_num );

        int num_vis = table.GetNumColumnsVisible();
        int num_hid = table.GetNumColumnsHidden();
        int num_mod = table.GetModel()->GetNumColumns();

        _ASSERT( total_num == num_vis + num_hid );

        view.Set( kColNumVisTag, num_vis );
        view.Set( kColNumHidTag, num_hid );
        view.Set( kColNumModTag, num_mod );

        if( !table.GetSortedCols().empty() ){
            CwxTableListCtrl::ESortedState sorted_global = table.GetSortedState();
            if( sorted_global != CwxTableListCtrl::eNoSort ){
                view.Set( kSortedGlobal, sorted_global == CwxTableListCtrl::eAscSort );
            } else {
                view.DeleteField( kSortedGlobal );
            }

            vector<int> sorted_cols;
            vector<int> sorted_orders;

            ITERATE( vector<CwxTableListCtrl::TSortedCol>, coltr, table.GetSortedCols() ){
                sorted_cols.push_back( coltr->first );
                sorted_orders.push_back( (int) coltr->second );
            }

            view.Set( kSortedColumn, sorted_cols );
            view.Set( kSortedOrder, sorted_orders );

        } else {
            view.DeleteField( kSortedGlobal );
            view.DeleteField( kSortedColumn );
            view.DeleteField( kSortedOrder );
        }

        wxArrayInt order_array = table.GetColumnsOrder();

        for( int ix = 0; ix < total_num; ix++ ){
            const int mod_ix = table.ColumnIndexToModel( ix );
            const bool shown = table.IsColumnVisibleByIndex( ix );

            string name_str = ToStdString(table.GetColumnNameByModel( mod_ix ));
            _ASSERT( !name_str.empty() );

            int width = shown ? table.GetColumnWidth( ix ) : table.GetColumn( ix ).m_Width;
            _ASSERT( width > 0 );

            string section = GetColumnSectionName( ix );

            view.Set( CGuiRegistryUtil::MakeKey(section, kColModelName), name_str );
            view.Set( CGuiRegistryUtil::MakeKey(section, kColModIndex), mod_ix );
            view.Set( CGuiRegistryUtil::MakeKey(section, kColVisOrder), order_array[ix] );
            view.Set( CGuiRegistryUtil::MakeKey(section, kColumnShown), shown );
            view.Set( CGuiRegistryUtil::MakeKey(section, kColumnWidth), width );
        }

        return;

    } catch( CException& e ){
        error = e.GetMsg();

    } catch( std::exception& e ){
        error = e.what();

    } 

    if( !error.empty() ){
        ERR_POST( "Error saving table settings " << error );
    }
}


void LoadTableSettings(
    CwxTableListCtrl& table, const CGuiRegistry::TReadView& view, bool byName
){
    string error;
    try {
        IwxTableModel* model = table.GetModel();
        if( model == NULL ){
            // _ASSERT( false );
            NCBI_THROW( CException, eUnknown, "Table model is not present" );
        }

        // read configuration
        int total_num = view.GetInt( kColNumTag, -1 );
        if( total_num == -1 ){
            //NCBI_THROW( CException, eUnknown, "No settings for the table" );
			return;

        } else if( total_num == 0 ){
            NCBI_THROW( CException, eUnknown, "Wrong settings for the table" );
        }

        int num_vis = view.GetInt( kColNumVisTag, -1 );
        int num_hid = view.GetInt( kColNumHidTag, -1 );
        //int num_mod = view.GetInt( kColNumModTag, -1 );

        if( total_num != num_vis + num_hid ){
            // file is inconsistent
            NCBI_THROW( CException, eUnknown, "Inconsistent info - column counts are wrong" );
        }



        vector<bool> mod_vec( table.GetModel()->GetNumColumns(), false );

        wxArrayInt order_array;
        order_array.SetCount( total_num, -1 );

        for( int ix = 0; ix < total_num; ix++ ){
            string section = GetColumnSectionName( ix );

            string name_str = view.GetString( CGuiRegistryUtil::MakeKey(section, kColModelName) );
            int mod_ix = view.GetInt( CGuiRegistryUtil::MakeKey(section, kColModIndex) );
            int ord_ix = view.GetInt( CGuiRegistryUtil::MakeKey(section, kColVisOrder) );
            bool shown = view.GetBool( CGuiRegistryUtil::MakeKey(section, string( kColumnShown )) );
            int width = view.GetInt( CGuiRegistryUtil::MakeKey(section, kColumnWidth) );

            // Match columns by name first, then by index
            // Find column in visibles first, then in hidden, and vice versa

            int new_mod = table.GetModelIndexByName( ToWxString(name_str) );
            if( new_mod == -1 ){
                LOG_POST( Warning << "No model column with such a name " << name_str );

                if( byName
                    || mod_ix >= table.GetModel()->GetNumColumns() 
                ){
                    continue;
                }

            } else if( new_mod != mod_ix ){
                LOG_POST( Warning << "Model index is different for the name " << name_str );

                mod_ix = new_mod;
            } 
            mod_vec[mod_ix] = true;

            int cur_ix = table.ColumnModelToIndex( mod_ix );
            if( cur_ix == -1 ){
                cur_ix = table.GetColumnCount();
                table.AddColumn( mod_ix );
            }

            table.GetColumn( cur_ix ).m_Visible = true;
            table.SetColumnWidth( cur_ix, width );

            if( !shown ){
                table.RemoveColumn( cur_ix, true );
            }

            if( !byName ){
                if( ix < table.GetColumnCount() ){
                    if( cur_ix != ix ){
                        table.MoveColumn( cur_ix, ix );
                        cur_ix = ix;
                    }
                }

                //??? that's strange
                order_array.SetCount( cur_ix+1, -1 );
                order_array[cur_ix] = ord_ix;
            }
        }

        if( !byName ){
            for( int mod_ix = 0; mod_ix < table.GetModel()->GetNumColumns(); mod_ix++ ){
                if( mod_vec[mod_ix] ){ continue; }

                int real_cur_ix = table.ColumnModelToIndex( mod_ix );
                if( real_cur_ix == -1 ){ continue; }

                int cur_ix;

                if( mod_ix == 0 ){
                    cur_ix = 0;
                
                } else {
                    cur_ix = table.ColumnModelToIndex( mod_ix -1 ) +1;
                }

                if( cur_ix == real_cur_ix ){ continue; }

                table.MoveColumn( real_cur_ix, cur_ix );

                //??? that's strange
                //order_array.SetCount( cur_ix+1, -1 );
                //order_array[cur_ix] = ord_ix;
            }
        }

        if( !byName ){
            table.SetColumnsOrder( order_array );
        }

        if( view.HasField( kSortedGlobal ) ){
            table.ResetSorting();

            bool sorted_global = view.GetBool( kSortedGlobal, true );
            table.SetSortedState(
                sorted_global ? CwxTableListCtrl::eAscSort : CwxTableListCtrl::eDescSort
            );

            vector<int> sorted_cols;
            vector<int> sorted_orders;
            
            view.GetIntVec( kSortedColumn, sorted_cols );
            view.GetIntVec( kSortedOrder, sorted_orders );

            size_t cols_num = sorted_cols.size();
            if( sorted_orders.size() < cols_num ){
                cols_num = sorted_orders.size();
            }

            for( size_t ix = 0; ix < cols_num; ix++ ){
                int sort_ix = sorted_cols[ix];
                if( sort_ix >= 0 && sort_ix < table.GetColumnCount() ){
                    table.AddSorterIndex( 
                        sort_ix, (CwxTableListCtrl::ESortedState)sorted_orders[ix] 
                    );
                }
            }

            table.SetSorterByColumns();

            table.ApplySorting( false );
        }

        table.Refresh();

    } catch( CException& e ){
        error = e.GetMsg();

    } catch( std::exception& e ){
        error = e.what();

    } 

    if( !error.empty() ){
        ERR_POST( "Error loading table settings: " << error );
    }
}

void MakeSelectedVisible( CwxTableListCtrl& table )
{
    int sel_ix_min = table.GetFirstSelected();
    //int sel_ix_max = table.GetNextItem( table.GetItemCount(), wxLIST_NEXT_ABOVE, wxLIST_STATE_SELECTED );
    int sel_ix_max = -1;

    for( int ix = sel_ix_min; ix != -1; ix = table.GetNextSelected( ix ) ){
        sel_ix_max = ix;
    }

    if( sel_ix_max != -1 ){
        table.EnsureVisible( sel_ix_max );
    }
    if( sel_ix_min != -1 ){
        table.EnsureVisible( sel_ix_min );
    }
}

void ExportTableControlToCSV( CwxTableListCtrl& table )
{
    CwxCSVExportDlg dlg(&table);
    dlg.SetRegistryPath("Dialogs.ExportTableControlToCSV");

    if( dlg.ShowModal() == wxID_OK) {
        wxString fileName = dlg.GetFileName();
        if (fileName.empty() )
            return;

        auto_ptr<CNcbiOstream> os;
        os.reset( new CNcbiOfstream( fileName.fn_str(), IOS_BASE::out ) );

        if( os.get() == NULL ){
            NCBI_THROW(
                CException, eUnknown,
                "File is not accessible"
            );
        }

        CCSVExporter exporter(*os, ',', '"');

        if( dlg.GetWithHeaders() ){
            for( int i = 0; i < table.GetNumColumnsVisible(); i++ ){
                wxString value = table.GetColumnNameVisible( i );
                exporter.Field(value);
            }
            exporter.NewRow();
        }

        int flags = dlg.GetSelectedOnly() ? wxLIST_STATE_SELECTED : wxLIST_STATE_DONTCARE;
        int item = -1;
        while( (item = table.GetNextItem( item, wxLIST_NEXT_ALL, flags )) != -1 ){
            for( int i = 0; i < table.GetNumColumnsVisible(); i++ ){
                wxString value;
                table.GetValueVisibleAt( item, i ).Convert( &value );
                exporter.Field(value);
            }
            exporter.NewRow();
        }
    }    
}


/*
///////////////////////////////////////////////////////////////////////////////
/// CTableColumnDialog
enum    ECommands   {
    eCmdUp = 10000,
    eCmdDown,
    eCmdOk,
    eCmdCancel
};

class CTableColumnDialog : public CDialog
{
public:
    typedef vector< pair<string, bool> >  TColumns;

    CTableColumnDialog(const TColumns& columns, const string& title);

    void    GetColumns(TColumns& columns);

    void    OnUp();
    void    OnDown();
    void    OnOk();
    void    OnCancel();

    static void Callback(Fl_Widget* w, void* ptr);

protected:
    DECLARE_EVENT_MAP();

    void    x_UpdateButtonsState();
    void    x_OnSelectionChanged();

protected:
    string  m_Title;

    CGridLayout m_BtnLayout;

    CCheckBrowser*  m_Browser;

    CButton*    m_UpBtn;
    CButton*    m_DownBtn;
    CButton*    m_OkBtn;
    CButton*    m_CancelBtn;
};


BEGIN_EVENT_MAP(CTableColumnDialog, CCommandTarget)
    ON_COMMAND(eCmdUp, &CTableColumnDialog::OnUp)
    ON_COMMAND(eCmdDown, &CTableColumnDialog::OnDown)
    ON_COMMAND(eCmdOk, &CTableColumnDialog::OnOk)
    ON_COMMAND(eCmdCancel, &CTableColumnDialog::OnCancel)
    //ON_UPDATE_COMMAND_UI(eCmdUp, &CTableColumnDialog::OnUpdateUp)
    //ON_UPDATE_COMMAND_UI(eCmdDown, &CTableColumnDialog::OnUpdateDown)
END_EVENT_MAP()


class CTableColumnGroup : public CGroup
{
public:
    CTableColumnGroup(int x, int y, int w, int h)
        : CGroup(x, y, w, h)  {}
    virtual void resize(int x, int y, int w, int h)
    {
        Fl_Widget::resize(x, y, w, h);
        m_Layout.Resize(10, 10, w - 20, h - 30);
    }
    CGridLayout m_Layout;
};


CTableColumnDialog::CTableColumnDialog(const TColumns& columns, const string& title)
{
    m_Title = title.empty() ? string("Customize Table Columns") : title;

    static const int kWinW = 400;
    static const int kWinH = 300;
    m_Window.reset(new Fl_Window(kWinW, kWinH, m_Title.c_str()));

    CTableColumnGroup* grp = new CTableColumnGroup(0, 0, kWinW, kWinH);
    m_Window->add_resizable(*grp);

    //create controls
    grp->m_Layout.Create(2, 1);
    grp->m_Layout.SetResizable(0, 0);
    grp->m_Layout.SetOffset(2, 4);

    m_Browser = new CCheckBrowser();
    m_Browser->callback(Callback, this);
    grp->m_Layout.Insert(m_Browser, 0, 0, 0, 100, 100);

    m_BtnLayout.Create(1, 5);
    m_BtnLayout.SetResizable(-1, 2);
    m_BtnLayout.SetOffset(0, 2);
    grp->m_Layout.Insert(&m_BtnLayout, 1, 0);

    m_UpBtn = new CButton(eCmdUp, this, "Move Up");
    m_BtnLayout.Insert(m_UpBtn, 0, 0, 0, CButton::eDefaultWidth, CButton::eDefaultHeight);

    m_DownBtn = new CButton(eCmdDown, this, "Move Down");
    m_BtnLayout.Insert(m_DownBtn, 0, 1, 0, CButton::eDefaultWidth, CButton::eDefaultHeight);

    m_OkBtn = new CButton(eCmdOk, this, "Ok");
    m_BtnLayout.Insert(m_OkBtn, 0, 3, 0, CButton::eDefaultWidth, CButton::eDefaultHeight);

    m_CancelBtn = new CButton(eCmdCancel, this, "Cancel");
    m_BtnLayout.Insert(m_CancelBtn, 0, 4, 0, CButton::eDefaultWidth, CButton::eDefaultHeight);

    grp->end();
    grp->resize(0, 0, kWinW, kWinH);

    m_Window->end();

    // update data
    for( size_t i = 0;  i < columns.size();  i++ )  {
        const pair<string, bool>& p = columns[i];
        m_Browser->Add(p.first, p.second);
    }

    m_Browser->value(1);
    x_UpdateButtonsState();
}


void CTableColumnDialog::GetColumns(TColumns& vis_columns)
{
    int n = m_Browser->size();
    vis_columns.resize(n);
    for( int i = 0;  i < n;  i++ )  {
        vis_columns[i].first = m_Browser->text(i + 1);
        vis_columns[i].second = m_Browser->GetCheck(i + 1);
    }
}


void CTableColumnDialog::OnUp()
{
    int sel_index = m_Browser->value();
    _ASSERT(sel_index > 1  &&  sel_index <= m_Browser->size());

    m_Browser->Swap(sel_index, sel_index - 1);
    m_Browser->redraw();

    x_UpdateButtonsState();
}


void CTableColumnDialog::OnDown()
{
    int sel_index = m_Browser->value();
    _ASSERT(sel_index >= 1  &&  sel_index < m_Browser->size());

    m_Browser->Swap(sel_index, sel_index + 1);
    m_Browser->redraw();

    x_UpdateButtonsState();
}


void CTableColumnDialog::x_UpdateButtonsState()
{
    int sel_index = m_Browser->value();
    int n = m_Browser->size();
    m_UpBtn->Activate(sel_index != 1);
    m_DownBtn->Activate(sel_index != n);
}


void CTableColumnDialog::OnOk()
{
    x_OnOK();
}


void CTableColumnDialog::OnCancel()
{
    x_OnCancel();
}


void CTableColumnDialog::Callback(Fl_Widget* w, void* ptr)
{
    ((CTableColumnDialog*)ptr)->x_UpdateButtonsState();
}


///////////////////////////////////////////////////////////////////////////////
///
bool    ShowColumnDialog(CwxTableListCtrl& table, const string& title)
{
    typedef pair<string, bool> TPair;
    vector<TPair> columns;

    columns.push_back(TPair("First", true));
    columns.push_back(TPair("Second", false));
    columns.push_back(TPair("Fird", true));
    columns.push_back(TPair("Forth", true));
    columns.push_back(TPair("Fith", false));
    columns.push_back(TPair("Sizth", false));

    CTableColumnDialog dlg(columns, title);
    dlg.CenterOnActive();
    dlg.ShowModal();

    dlg.GetColumns(columns);

    for(  size_t i = 0;  i < columns.size();  i++ ) {
        LOG_POST(columns[i].first << "  " << columns[i].second);
    }

    return true;
}
*/


END_NCBI_SCOPE
