/*  $Id: tax_id_helper.cpp 26501 2012-09-27 15:23:11Z kuznets $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Roman Katargin
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>

#include <gui/widgets/wx/async_call.hpp>
#include <gui/widgets/wx/wx_utils.hpp>
#include <gui/widgets/wx/tax_id_helper.hpp>
#include <gui/widgets/wx/sys_path.hpp>

#include <gui/objutils/taxon_cache.hpp>

#include <wx/ctrlsub.h>

BEGIN_NCBI_SCOPE

/// Functor class for background labels preparation
/// @internal
class CGetTaxIdLabels : public CAsyncCall::Call
{
public:
    CGetTaxIdLabels(const vector<int>& taxIds, vector<wxString>& labels) :
      Call(wxT("Getting taxon labels...")), m_TaxIds(taxIds), m_Labels(labels) {}
    virtual void Execute();
private:
    const vector<int>& m_TaxIds;
    vector<wxString>& m_Labels;
};

void CGetTaxIdLabels::Execute()
{
    wxString tax_cache_path = CSysPath::ResolvePath( wxT("<home>/tax_name_cache.txt") );
    string s(tax_cache_path.ToStdString());
    CTaxonCache& cache = CTaxonCache::GetInstance(s);
    cache.Initialize(m_TaxIds);
    ITERATE(vector<int>, it, m_TaxIds) {
        wxString label = wxString::Format(wxT("%6d"), *it);
        string name = cache.GetLabel(*it);

        if (!name.empty()) {
            label += wxT(" ") + ToWxString(name);
        }

        m_Labels.push_back(label);
    }
    
    // Close connection to taxon service. Action optional, but useful for better
    // resource utilization
    //
    cache.ResetConnection(); 
}

static bool TaxIdComp(const int& i1, const int& i2)
{
    if (i1 == 9606)
        return true;
    if (i2 == 9606)
        return false;

    if (i1 == 10090)
        return (i2 != 9660);

    if (i2 == 10090)
        return (i2 == 9660);

    return i1 < i2;
}

void CTaxIdHelper::FillTaxIds(wxControlWithItems& ctrl, const vector<int>& ids, int selectTaxId)
{
    ctrl.Clear();

    ctrl.Append(wxT("none"), (void*)0);

    if (ids.empty()) {
        ctrl.SetSelection(0);
        return;
    }

    vector<int> sorted = ids;
    sort(sorted.begin(), sorted.end(), TaxIdComp);

    vector<wxString> strs;
    CGetTaxIdLabels getLabels(sorted, strs);
    CAsyncCall::Execute(getLabels);

    int selectIndex = 0;
    for (unsigned int i = 0; i < strs.size(); ++i) {
        if (sorted[i] == selectTaxId)
            selectIndex = i;
        ctrl.Append(strs[i], (void*)sorted[i]);
    }
    ctrl.SetSelection(selectIndex);
}

END_NCBI_SCOPE
