/*===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data, the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties, express or implied, including
*  warranties of performance, merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================
*
*/

/**
* Unit tests for TraceOS project
*/

#include <sysalloc.h>

#include <stdexcept>

#include <ktst/unit_test.hpp>

#include <common/test_assert.h>

extern "C"
{
    #include "../../../services/open-server/tros_srautils.c"
    #undef Main
    #undef FAIL
    #undef Args
}

using namespace std;

TEST_SUITE(TraceOS_Suite);

class SraFixture
{
public:
    SraFixture() 
    {
        if (SRAMgrMakeRead(&mgr) != 0)
            throw logic_error("SraFixture: SRAMgrMakeRead failed");
        memset(&data, 0, sizeof(data));
    }
    ~SraFixture() 
    {
        SraUserDataReset(&data);
        if (SRAMgrRelease(mgr) != 0)
            throw logic_error("~SraFixture: SRAMgrRelease failed");
    }

    const SRAMgr* mgr;
    SraUserData data;
};

FIXTURE_TEST_CASE(SetRunAccession_Open_Reset, SraFixture)
{
    string acc = "SRR000123";
    char* errStr;
    REQUIRE_RC(SraUserDataSetRunAccession(mgr, &data, acc.c_str(), &errStr));
    
    REQUIRE_EQ(string(data.acc), acc);
    // etc...
    
    REQUIRE_RC(SraUserDataOpen(&data, 1, 0, 0));
}

FIXTURE_TEST_CASE(SetRunAccession_Open_Delete, SraFixture)
{
    SraUserData* dataPtr = (SraUserData*)malloc(sizeof(SraUserData));
    memset(dataPtr, 0, sizeof(SraUserData));
    
    string acc = "SRR000123";
    char* errStr;
    REQUIRE_RC(SraUserDataSetRunAccession(mgr, dataPtr, acc.c_str(), &errStr));
    REQUIRE_RC(SraUserDataOpen(dataPtr, 1, 0, 0));
    SraUserDataReset(dataPtr);
    REQUIRE_RC(SraUserDataOpen(dataPtr, 2, 0, 0));
    
    SraUserDataDelete(dataPtr);
}

FIXTURE_TEST_CASE(GetFileList, SraFixture)
{
    const int MaxFiles = 1024;
    SraFiles files[MaxFiles];
    int first = MaxFiles;
    REQUIRE_RC(SraGetFileList("/netmnt/traces04", "sra4/SRR/000000", 0, files, &first, false, false));
    REQUIRE_GT(MaxFiles, first);
}

//////////////////////////////////////// Mock implementation of OpenServer API
static int g_spp;
static UserThreadDataPtr threadData;

#define REQ_EQ(act, exp) if (act != exp) { cout << __FILE__ << "(" << __LINE__ << "): Actual \"" << act << "\" expected \"" << exp << "\"" << endl; return CS_FAIL; }

CS_RETCODE srv_thread_props (SRV_PROC *p_spp,CS_INT  cmd,CS_INT  property,CS_VOID *buf,CS_INT  buflen,CS_INT  *outlen)
{
    REQ_EQ(p_spp, &g_spp);
    REQ_EQ(cmd, CS_GET);
    REQ_EQ(property, SRV_T_USERDATA);
    if (buf == NULL)                    return CS_FAIL;
    if (buflen < (CS_INT)sizeof(threadData)) return CS_FAIL;
    
    if (outlen != NULL)
        *outlen = sizeof(threadData);
    
    memcpy(buf, &threadData, sizeof(threadData));
    
    return CS_SUCCEED;
}
CS_RETCODE srv_descfmt (SRV_PROC *p_spp,CS_INT cmd,CS_INT type,CS_INT item,CS_DATAFMT *rmtfmtp)
{
    REQ_EQ(p_spp, &g_spp);
    REQ_EQ(cmd, CS_SET);
    REQ_EQ(type, SRV_ROWDATA);
    
    return CS_SUCCEED;
}
CS_RETCODE srv_bind (SRV_PROC *p_spp,CS_INT cmd,CS_INT type,CS_INT item,CS_DATAFMT *lclfmtp,CS_BYTE *varaddrp,CS_INT *varlenp,CS_SMALLINT *indp)
{
    REQ_EQ(p_spp, &g_spp);
    REQ_EQ(cmd, CS_SET);
    REQ_EQ(type, SRV_ROWDATA);
    
    const SraUserData* ud = (const SraUserData*)threadData->userdata;
    if (ud == NULL) return CS_FAIL;
    
    string val = string((const char*)varaddrp, *varlenp);

        // TODO checck indp for NULL/GOOD depending on name
    if (string(lclfmtp->name) == string("run_acc"))
        REQ_EQ(val, string(ud->acc));
    if (string(lclfmtp->name) == string("platform"))
        REQ_EQ(val, string(ud->platform_str));
    
    return CS_SUCCEED;
}
CS_RETCODE srv_xferdata (SRV_PROC *p_spp,CS_INT cmd,CS_INT type)
{
    REQ_EQ(p_spp, &g_spp);
    REQ_EQ(cmd, CS_SET);
    REQ_EQ(type, SRV_ROWDATA);
    
    return CS_SUCCEED;
}
CS_RETCODE srv_senddone (SRV_PROC *p_spp,CS_INT status,CS_INT transtate,CS_INT count)
{
    REQ_EQ(p_spp, &g_spp);
    if (status != (SRV_DONE_MORE|SRV_DONE_COUNT))   return CS_FAIL;
    if (transtate != CS_TRAN_UNDEFINED)             return CS_FAIL;
    if (count != 1)                                 return CS_FAIL;
    
    return CS_SUCCEED;
}
    
// CS_RETCODE SraUserSendMetadata(SRV_PROC *spp);
FIXTURE_TEST_CASE(SendMetadata, SraFixture)
{
    threadData = new UserThreadData;
    threadData->userdata = & data;
    char* errStr;
    REQUIRE_RC(SraUserDataSetRunAccession(mgr, & data, "SRR000123", &errStr));
    REQUIRE_EQ(SraUserSendMetadata(&g_spp), CS_SUCCEED);
    delete threadData;
    threadData = 0;
}

// CS_RETCODE SraUserSpotDataPrepare(SRV_PROC *spp,SraUserData *sd,int format_mask,int read_mask);
// CS_RETCODE SraUserSpotDataSend(SRV_PROC *spp,SraUserData *sd,int format_mask,int read_mask,int do_clip);
// CS_RETCODE SraUserSignalSend(SRV_PROC *spp,SraUserData *sd);
// CS_RETCODE SraUserSequenceSend(SRV_PROC *spp,SraUserData *sd);
// CS_RETCODE Sra_send_asnprop(SRV_PROC *spp);

//////////////////////////////////////////// Main
extern "C"
{

#include <kapp/args.h>
#include <kfg/config.h>

ver_t CC KAppVersion ( void )
{
    return 0x1000000;
}
rc_t CC UsageSummary (const char * progname)
{
    return 0;
}

rc_t CC Usage ( const Args * args )
{
    return 0;
}

const char UsageDefaultName[] = "test-open-server";

rc_t CC KMain ( int argc, char *argv [] )
{
    KConfigDisableUserSettings();
    rc_t rc=TraceOS_Suite(argc, argv);
    return rc;
}

}
