/*===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data, the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties, express or implied, including
*  warranties of performance, merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================
*
*/

/**
* Unit tests for SRA interfaces
*/

#include <sysalloc.h>

#include <cstring>

#include <stdexcept>

#include <ktst/unit_test.hpp>

#include <sra/sradb.h>
#include <sra/sradb-priv.h>
#include "../../libs/sra/sra-priv.h"

#include <common/test_assert.h>

using namespace std;

TEST_SUITE(SraTestSuite);

class SraFixture
{
public:
    SraFixture() 
    {
        if (SRAMgrMakeRead(&mgr) != 0)
            throw logic_error("SraFixture: SRAMgrMakeRead failed");
    }
    ~SraFixture() 
    {
        if (SRAMgrRelease(mgr) != 0)
            throw logic_error("~SraFixture: SRAMgrRelease failed");
    }

    const SRAMgr* mgr;
};

//////////////////// SRACacheElement
class SRACacheElementFixture : public SraFixture
{
public: 
    SRACacheElementFixture()
    : elem(0)
    {
    }
    ~SRACacheElementFixture()
    {
        if (elem && SRACacheElementDestroy(elem) != 0)
            throw logic_error("~SRACacheElementFixture: SRACacheElementDestroy failed");
    }
    
    SRACacheIndex index;
    SRACacheElement* elem;
};

FIXTURE_TEST_CASE(SRACacheElement_Basic, SRACacheElementFixture)
{
    SRATable* table;
    REQUIRE_RC(SRAMgrOpenTableRead(mgr, (const SRATable**)&table, "SRR000123"));

    SRACacheMetrics metrics = {1, 2, 3};
    REQUIRE_RC(SRACacheElementMake(&elem, table, &index, 100, &metrics));
    
    REQUIRE_EQ(table, elem->object);
    REQUIRE_EQ(&index, elem->index);
    REQUIRE_EQ((uint32_t)100, elem->key);
    REQUIRE_EQ(memcmp(&metrics, &elem->metrics, sizeof(metrics)), 0);
    
    REQUIRE_RC(SRACacheElementDestroy(elem));
    elem = NULL;
    
    REQUIRE_RC(SRATableRelease(table));
}

FIXTURE_TEST_CASE(SRACacheElement_Errors, SRACacheElementFixture)
{
    SRATable* table;
    REQUIRE_RC(SRAMgrOpenTableRead(mgr, (const SRATable**)&table, "SRR000123"));
    SRACacheElement* elem;
    SRACacheIndex index;
    SRACacheMetrics metrics = {1, 2, 3};
    
    REQUIRE_RC_FAIL(SRACacheElementMake(NULL,  table, &index, 100, &metrics));
    REQUIRE_RC_FAIL(SRACacheElementMake(&elem, NULL,  &index, 100, &metrics));
    REQUIRE_RC_FAIL(SRACacheElementMake(&elem, table, NULL,   100, &metrics));
    REQUIRE_RC_FAIL(SRACacheElementMake(&elem, table, &index, 100, NULL));
    
    REQUIRE_RC(SRATableRelease(table));
}

//////////////////// SRACacheIndex
FIXTURE_TEST_CASE(SRACacheIndex_Basic, SRACacheElementFixture)
{
    SRACacheIndex* index;
    String prefix;
    CONST_STRING(&prefix, "pref");
    
    REQUIRE_RC(SRACacheIndexMake(&index, &prefix));
    REQUIRE_NOT_NULL(index);
    REQUIRE_EQ(StringCompare(&prefix, index->prefix), 0);
    REQUIRE_NOT_NULL(index->body);
    REQUIRE_RC(SRACacheIndexDestroy(index));
}
FIXTURE_TEST_CASE(SRACacheIndex_Errors, SRACacheElementFixture)
{
    SRACacheIndex* index;
    String prefix;
    CONST_STRING(&prefix, "pref");
    
    REQUIRE_RC_FAIL(SRACacheIndexMake(NULL, &prefix));
    REQUIRE_RC_FAIL(SRACacheIndexMake(&index, NULL));
    REQUIRE_RC_FAIL(SRACacheIndexDestroy(NULL));
}

//////////////////// SRACache
class SRACacheFixture : public SraFixture
{
public: 
    SRACacheFixture()
    : cache(0), table(0)
    {
        if (SRACacheInit(&cache) != 0)
            throw logic_error("SRACacheFixture: SRACacheInit failed");
        ClearMetrics();
    }
    ~SRACacheFixture()
    {
        if (table && SRATableRelease(table) != 0)
            throw logic_error("~SRACacheFixture: SRATableRelease failed");
        if (cache && SRACacheWhack(cache) != 0)
            throw logic_error("~SRACacheFixture: SRACacheWhack failed");
    } 
    
    void ClearMetrics()
    {
        memset(&metrics, 0, sizeof(metrics));
    }
    // false - not added
    bool AddTable(const char* acc)
    {
        if (table && SRATableRelease(table) != 0)
            throw logic_error("AddTable: SRATableRelease failed");
        if (SRAMgrOpenAltTableRead(mgr, (const SRATable**)&table, NULL, acc) != 0)
            throw logic_error("AddTable: SRAMgrOpenAltTableRead failed");
        return SRACacheAddTable(cache, acc, table) == 0;
    }
    
    SRACache* cache;
    SRACacheMetrics metrics;
    SRATable* table;
};

FIXTURE_TEST_CASE(SRACache_Basic, SRACacheFixture)
{
    {
        REQUIRE_RC(SRACacheGetSoftThreshold(cache, &metrics));
        REQUIRE_EQ(metrics.elements,    SRACacheThresholdSoftElementsDefault);
        REQUIRE_EQ(metrics.bytes,       SRACacheThresholdSoftBytesDefault);
        REQUIRE_EQ(metrics.threads,     SRACacheThresholdSoftThreadsDefault);
        REQUIRE_EQ(metrics.fds,         SRACacheThresholdSoftFdsDefault);
        ClearMetrics();
    }
    {
        REQUIRE_RC(SRACacheGetHardThreshold(cache, &metrics));  
        REQUIRE_EQ(metrics.elements,    SRACacheThresholdHardElementsDefault);
        REQUIRE_EQ(metrics.bytes,       SRACacheThresholdHardBytesDefault);
        REQUIRE_EQ(metrics.threads,     SRACacheThresholdHardThreadsDefault);
        REQUIRE_EQ(metrics.fds,         SRACacheThresholdHardFdsDefault);
        ClearMetrics();
    }
}

FIXTURE_TEST_CASE(SRACache_SetThresholds, SRACacheFixture)
{
    {
        SRACacheMetrics thr={1, 2, 3, 4};
        REQUIRE_RC(SRACacheSetSoftThreshold(cache, &thr));
        REQUIRE_RC(SRACacheGetSoftThreshold(cache, &metrics));
        REQUIRE_EQ(metrics.elements,    thr.elements);
        REQUIRE_EQ(metrics.bytes,       thr.bytes);
        REQUIRE_EQ(metrics.threads,     thr.threads);
        REQUIRE_EQ(metrics.fds,         thr.fds);
        ClearMetrics();
    }
    {
        SRACacheMetrics thr={5, 6, 7, 8};
        REQUIRE_RC(SRACacheSetHardThreshold(cache, &thr));
        REQUIRE_RC(SRACacheGetHardThreshold(cache, &metrics));
        REQUIRE_EQ(metrics.elements,    thr.elements);
        REQUIRE_EQ(metrics.bytes,       thr.bytes);
        REQUIRE_EQ(metrics.threads,     thr.threads);
        REQUIRE_EQ(metrics.fds,         thr.fds);
        ClearMetrics();
    }
}

FIXTURE_TEST_CASE(SRACache_Errors, SRACacheFixture)
{
    REQUIRE_RC_FAIL(SRACacheInit(NULL));
    
    SRACacheMetrics thr={0, 0, 0};
    REQUIRE_RC_FAIL(SRACacheGetSoftThreshold(NULL, &thr));
    REQUIRE_RC_FAIL(SRACacheGetSoftThreshold(cache, NULL));
    REQUIRE_RC_FAIL(SRACacheGetHardThreshold(NULL, &thr));
    REQUIRE_RC_FAIL(SRACacheGetHardThreshold(cache, NULL));

    REQUIRE_RC_FAIL(SRACacheGetUsage(NULL, &metrics));
    REQUIRE_RC_FAIL(SRACacheGetUsage(cache, NULL));

    REQUIRE_RC_FAIL(SRACacheWhack(NULL));
}

// SRACacheAddTable
FIXTURE_TEST_CASE(SRACacheAddTable_Once, SRACacheFixture)
{
    REQUIRE(AddTable("SRR000123"));
    
    // validate cache->indexes
    const SRACacheIndex* index = (SRACacheIndex*)BSTreeFirst(&cache->indexes);
    REQUIRE_NOT_NULL(index);
    REQUIRE_EQ(index, (const SRACacheIndex*)BSTreeLast(&cache->indexes)); // the only entry
    
    String p;
    CONST_STRING(&p, "SRR");
    REQUIRE_EQ(StringCompare(index->prefix, &p), 0);

    const SRACacheElement* value;
    REQUIRE_RC(KVectorGetPtr(index->body, 123, (void**)&value));    
    
    REQUIRE_EQ(value->object, table);
    REQUIRE_EQ(value->index, (SRACacheIndex*)index);
    REQUIRE_EQ(value->key, 123u);
    REQUIRE_EQ(value->metrics.elements, 1U);
//TBD    REQUIRE_NE(metrics.bytes,       0LU);
//TBD    REQUIRE_NE(metrics.threads,     0U);
//TBD    REQUIRE_NE(metrics.fds,         0U);
    
    // validate cache->lru
    REQUIRE_NOT_NULL(cache->lru.head);
    REQUIRE_EQ(cache->lru.head, cache->lru.tail);
    REQUIRE_EQ((const SRACacheElement*)cache->lru.head, value);

    // add a different table
    REQUIRE(AddTable("SRR000122"));
    // this one is already in, should be rejected
    REQUIRE( ! AddTable("SRR000123") );
} 

// SRACacheGetUsage
FIXTURE_TEST_CASE(SRACache_GetUsageEmpty, SRACacheFixture)
{
    REQUIRE_RC(SRACacheGetUsage(cache, &metrics));
    REQUIRE_EQ(metrics.elements,    0U);
    REQUIRE_EQ(metrics.bytes,       (uint64_t)0);
    REQUIRE_EQ(metrics.threads,     0U);
    REQUIRE_EQ(metrics.fds,         0U);
}
FIXTURE_TEST_CASE(SRACache_GetUsageOneTable, SRACacheFixture)
{
    REQUIRE(AddTable("SRR000123"));
    
    REQUIRE_RC(SRACacheGetUsage(cache, &metrics));
    REQUIRE_EQ(metrics.elements,    1U);
//TBD    REQUIRE_NE(metrics.bytes,       0LU);
//TBD    REQUIRE_NE(metrics.threads,     0U);
//TBD    REQUIRE_NE(metrics.fds,         0U);
}
FIXTURE_TEST_CASE(SRACache_GetUsageMultipleTables, SRACacheFixture)
{
    REQUIRE(AddTable("SRR000120"));
    REQUIRE(AddTable("SRR000001"));
    REQUIRE(AddTable("SRR000122"));
    
    REQUIRE_RC(SRACacheGetUsage(cache, &metrics));
    REQUIRE_EQ(metrics.elements,    3U);
//TBD    REQUIRE_NE(metrics.bytes,       0LU);
//TBD    REQUIRE_NE(metrics.threads,     0U);
//TBD    REQUIRE_NE(metrics.fds,         0U);
}

// SRACacheGetTable
FIXTURE_TEST_CASE(SRACacheGetTable_Empty, SRACacheFixture)
{
    REQUIRE_RC(SRACacheGetTable(cache, "SRR000121", (const SRATable**)&table));
    REQUIRE_NULL(table);
}

FIXTURE_TEST_CASE(SRACacheGetTable_Single, SRACacheFixture)
{
    REQUIRE(AddTable("SRR000121"));
    const SRATable* obj;
    REQUIRE_RC_FAIL(SRACacheGetTable(cache, "SRR000121", &obj)); // busy
    REQUIRE_RC(SRATableRelease(table)); // now, available and still in the cache
    REQUIRE_RC(SRACacheGetTable(cache, "SRR000121", &obj)); 
    REQUIRE_NOT_NULL(obj);
    REQUIRE_EQ((const SRATable*)table, obj);
}

FIXTURE_TEST_CASE(SRACacheGetTable_Multiple, SRACacheFixture)
{
    REQUIRE(AddTable("SRR000121"));
    REQUIRE(AddTable("SRR000122"));
    const SRATable* obj1 = table;
    table = 0;
    REQUIRE(AddTable("SRR000123"));
    REQUIRE(AddTable("SRR000124"));

    const SRATable* obj2 = 0;
    REQUIRE_RC_FAIL(SRACacheGetTable(cache, "SRR000122", &obj2)); // busy
    REQUIRE_RC(SRATableRelease(obj1)); // now, available and still in the cache
    REQUIRE_RC(SRACacheGetTable(cache, "SRR000122", &obj2)); 
    REQUIRE_NOT_NULL(obj2);
    REQUIRE_EQ(obj1, obj2);
    REQUIRE_RC(SRATableRelease(obj2));
}

// SRACacheFlush
FIXTURE_TEST_CASE(SRACacheFlush_Empty, SRACacheFixture)
{
    REQUIRE_RC(SRACacheFlush(cache));
    REQUIRE_RC(SRACacheGetUsage(cache, &metrics));
    REQUIRE_EQ(metrics.elements, 0U);
}
FIXTURE_TEST_CASE(SRACacheFlush_OverThreshold, SRACacheFixture)
{   // multiple SRACacheAddTable, flush correct # of elements, correct SRACacheGet's
    metrics.elements = 2;
    REQUIRE_RC(SRACacheSetSoftThreshold(cache, &metrics));
    
    REQUIRE(AddTable("SRR000121")); 
    REQUIRE(AddTable("SRR000122")); 
    REQUIRE(AddTable("SRR000123")); 
    REQUIRE(AddTable("SRR000124")); 
    REQUIRE_RC(SRATableRelease(table));
    table = 0;
    
    REQUIRE_RC(SRACacheFlush(cache));
    
    REQUIRE_RC(SRACacheGetUsage(cache, &metrics));
    REQUIRE_EQ(metrics.elements, 1U);
    const SRATable* obj;
    REQUIRE_RC(SRACacheGetTable(cache, "SRR000121", &obj));
    REQUIRE_NULL(obj);
    REQUIRE_RC(SRACacheGetTable(cache, "SRR000122", &obj));
    REQUIRE_NULL(obj);
    REQUIRE_RC(SRACacheGetTable(cache, "SRR000123", &obj));
    REQUIRE_NULL(obj);
    REQUIRE_RC(SRACacheGetTable(cache, "SRR000124", &obj));
    REQUIRE_NOT_NULL(obj);
    REQUIRE_RC(SRATableRelease(obj));
}

////////////////////////////////////////// tests for public API //////////////////////////////////////////

//////////////////// SRAMgrOpenTableRead
//      if after addition metrics cross hard threshold, performs in-thread flush
//      else if metrics cross soft threshold, signals BG thread to wake up 
FIXTURE_TEST_CASE(SRAMgrOpenTableRead_CacheBasic, SraFixture)
{
    const SRATable* table;
    REQUIRE_RC(SRAMgrOpenTableRead(mgr, &table, "SRR000124"));
    REQUIRE_NOT_NULL(mgr->cache);

    SRACacheMetrics metrics;
    REQUIRE_RC(SRACacheGetUsage(mgr->cache, &metrics));
    REQUIRE_EQ(metrics.elements, 1u);
    
    REQUIRE_RC(SRATableRelease(table));
}

FIXTURE_TEST_CASE(SRAMgrOpenTableRead_CacheMultiple, SraFixture)
{   // multiple SRAMgrOpenTableRead for diff accessions, correct # of elements, 
    const SRATable* table;
    REQUIRE_RC(SRAMgrOpenTableRead(mgr, &table, "SRR000124")); 
    REQUIRE_RC(SRATableRelease(table));
    REQUIRE_RC(SRAMgrOpenTableRead(mgr, &table, "SRR000124")); 
    REQUIRE_RC(SRATableRelease(table));
    REQUIRE_RC(SRAMgrOpenTableRead(mgr, &table, "SRR000123")); 
    REQUIRE_RC(SRATableRelease(table));
    REQUIRE_RC(SRAMgrOpenTableRead(mgr, &table, "SRR000001")); 
    REQUIRE_RC(SRATableRelease(table));
    table = NULL;

    SRACacheMetrics metrics;
    REQUIRE_RC(SRACacheGetUsage(mgr->cache, &metrics));
    REQUIRE_EQ(metrics.elements, 3u);
}

FIXTURE_TEST_CASE(SRAMgrOpenTableRead_CacheFlushBG, SraFixture)
{   // multiple SRAMgrOpenTableRead for diff accessions, enough to exceed the soft threshold, SRAMgrRunBGTasks, correct # of elements after flush
    SRACacheMetrics metrics;
    metrics.elements = 2;
    REQUIRE_RC(SRACacheSetSoftThreshold(mgr->cache, &metrics));

    const SRATable* table;
    REQUIRE_RC(SRAMgrOpenTableRead(mgr, &table, "SRR000120")); REQUIRE_RC(SRATableRelease(table));
    REQUIRE_RC(SRAMgrOpenTableRead(mgr, &table, "SRR000121")); REQUIRE_RC(SRATableRelease(table));
    REQUIRE_RC(SRAMgrOpenTableRead(mgr, &table, "SRR000122")); REQUIRE_RC(SRATableRelease(table));
    REQUIRE_RC(SRAMgrOpenTableRead(mgr, &table, "SRR000123")); REQUIRE_RC(SRATableRelease(table));
    table = NULL;

    REQUIRE_RC(SRAMgrRunBGTasks(mgr));
    
    REQUIRE_RC(SRACacheGetUsage(mgr->cache, &metrics));
    REQUIRE_EQ(metrics.elements, 1u);
}

FIXTURE_TEST_CASE(SRAMgrOpenTableRead_CacheFlushFG, SraFixture)
{   // multiple SRAMgrOpenTableRead for diff accessions, enough to exceed the hard threshold, correct # of elements after flush
    SRACacheMetrics metrics;
    metrics.elements = 2;
    REQUIRE_RC(SRACacheSetSoftThreshold(mgr->cache, &metrics));
    metrics.elements = 4;
    REQUIRE_RC(SRACacheSetHardThreshold(mgr->cache, &metrics));

    const SRATable* table;
    REQUIRE_RC(SRAMgrOpenTableRead(mgr, &table, "SRR000120")); REQUIRE_RC(SRATableRelease(table));
    REQUIRE_RC(SRAMgrOpenTableRead(mgr, &table, "SRR000121")); REQUIRE_RC(SRATableRelease(table));
    REQUIRE_RC(SRAMgrOpenTableRead(mgr, &table, "SRR000122")); REQUIRE_RC(SRATableRelease(table));
    REQUIRE_RC(SRAMgrOpenTableRead(mgr, &table, "SRR000123")); // this should kick off a flush
    REQUIRE_RC(SRATableRelease(table));
    table = NULL;

    REQUIRE_RC(SRACacheGetUsage(mgr->cache, &metrics));
    REQUIRE_EQ(metrics.elements, 1u);
}

FIXTURE_TEST_CASE(SRAMgrOpenTableRead_CacheBusy, SraFixture)
{   // a cache object cannot be reused until it has been released 
    const SRATable* table1;
    REQUIRE_RC(SRAMgrOpenTableRead(mgr, &table1, "SRR000124")); 
    {
        const SRATable* table2;
        REQUIRE_RC(SRAMgrOpenTableRead(mgr, &table2, "SRR000124")); 
        REQUIRE_NE(table1, table2);
        REQUIRE_RC(SRATableRelease(table2));
    }
    REQUIRE_RC(SRATableRelease(table1));
    
    // table1 stays in the cache ready to be reused
    int n;
    REQUIRE_EQ((n = atomic32_read(&table1->refcount)), 1);

    {
        const SRATable* table3;
        REQUIRE_RC(SRAMgrOpenTableRead(mgr, &table3, "SRR000124")); 
        // table1 is reused
        REQUIRE_EQ(table1, table3); 
        REQUIRE_EQ((n = atomic32_read(&table1->refcount)), 2);
        REQUIRE_RC(SRATableRelease(table3));
    }
    REQUIRE_EQ((n = atomic32_read(&table1->refcount)), 1);
}


//////////////////////////////////////////// Main
extern "C"
{

#include <kapp/args.h>
#include <kfg/config.h>

ver_t CC KAppVersion ( void )
{
    return 0x1000000;
}
rc_t CC UsageSummary (const char * progname)
{
    return 0;
}

rc_t CC Usage ( const Args * args )
{
    return 0;
}
const char UsageDefaultName[] = "test-sra";

rc_t CC KMain ( int argc, char *argv [] )
{
    KConfigDisableUserSettings();
    rc_t rc=SraTestSuite(argc, argv);
    return rc;
}

}
