#ifndef PKG_ALIGNMENT___NET_BLAST_UI_DATA_SOURCE__HPP
#define PKG_ALIGNMENT___NET_BLAST_UI_DATA_SOURCE__HPP

/*  $Id: net_blast_ui_data_source.hpp 26573 2012-10-09 19:02:34Z kuznets $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Andrey Yazhuk
 *
 * File Description:
 *
 */

#include <corelib/ncbistl.hpp>
#include <corelib/ncbiobj.hpp>
#include <corelib/hash_map.hpp>

#include <gui/gui_export.h>

#include <gui/packages/pkg_alignment/net_blast_job_descr.hpp>

#include <gui/core/view_extensions.hpp>
#include <gui/core/app_explorer_item.hpp>

#include <gui/utils/app_job_dispatcher.hpp>

#include <gui/framework/ui_data_source.hpp>
#include <gui/framework/observable.hpp>
#include <gui/framework/service.hpp>

#include <objects/blast/Blast4_get_databases_reply.hpp>
#include <objects/general/User_field.hpp>

#include <gui/utils/extension.hpp>
#include <gui/utils/ui_object.hpp>

#include <gui/objutils/reg_settings.hpp>


BEGIN_NCBI_SCOPE

/** @addtogroup GUI_PKG_ALIGNMENT
 *
 * @{
 */

class CNetBLASTUIDataSourceType;
class CBlastSearchTask;

/// commands introduced by Net BLAST data source
enum ENetBlastCommands  {
    eCmdRetriveBlastJobs = 12200,
    eCmdMonitorBlastJobs,
    eCmdDeleteBlastJobs,
    eCmdExploreNetBLAST,
    eCmdLoadRIDs,
    eCmdBlastWinMask
};


#define NET_BLAST_DM_TOOL_NAME "NCBI Net BLAST"

///////////////////////////////////////////////////////////////////////////////
/// CNetBLASTUIDataSource
class  CNetBLASTUIDataSource :
    public CObjectEx,
    public CObservable,
    public IExtension,
    public IServiceLocatorConsumer,
    public IUIDataSource,
    public IObjectCmdContributor,
    public IExplorerItemCmdContributor,
    public IRegSettings
{
    // these classes work in collaboration with the Data Source, in a sense,
    // they are functional aspects of the Data Source packaged as separate classes
    friend class CNetBlastJobDescriptor;
    friend class CNetBlastDSEvtHandler;
public:
    typedef map<string, CRef<objects::CBlast4_database_info> > TDbMap;
    typedef vector< CRef<CNetBlastJobDescriptor> >  TJobDescrVec;
    typedef map<int, string> TTaxIdLabelMap;

    CNetBLASTUIDataSource(CNetBLASTUIDataSourceType& type);
    virtual ~CNetBLASTUIDataSource();

    void RequestCancel() { m_StopRequested.Add(1); }

    /// @name IExtension implementation
    /// @{
    virtual string  GetExtensionIdentifier() const;
    virtual string  GetExtensionLabel() const;
    /// @}

    /// @name IServiceLocatorConsumer implementation
    /// @{
    virtual void    SetServiceLocator(IServiceLocator* locator);
    /// @}
    
    /// @name IRegSettings interface implementation
    /// @{
    virtual void    SetRegistryPath(const string& path);
    virtual void    SaveSettings() const;
    virtual void    LoadSettings();
    /// @}

    /// @name IUIDataSource interface implementation
    /// @{
    virtual IUIDataSourceType&  GetType() const;
    virtual const IUIObject&    GetDescr();
    virtual bool    IsOpen();
    virtual bool    Open();
    virtual bool    Close();
    virtual void    EditProperties();
    virtual IUIToolManager*    GetLoadManager();
    /// @}

    /// @name IObjectCmdContributor implementation
    /// @{
    virtual IObjectCmdContributor::TContribution
                GetMenu(TConstScopedObjects& objects);
    /// @}

    /// @name IExplorerItemCmdContributor implementation
    /// @{
    virtual IExplorerItemCmdContributor::TContribution
                GetMenu(TItemRefVector& items, CAppExplorerService& app_service);
    /// @}

    /// @name Interface for accessing Net BLAST services and objects
    /// @{

    /// Wait for background loading finishes, before making datasource fully
    /// available
    void WaitReady();

    /// returns a map of all available BLAST databases
    const TDbMap&   GetDbMap(bool nuc_db) const;

    string  GetDefaultDB(bool nuc_db) const;
    void    GetDefaultMRU_DBs(bool nuc_db, list<string>& mru_dbs);

    CRef<CNetBlastJobDescriptor>    CreateJobDescriptor(const CBLASTParams& params);
    CRef<CNetBlastJobDescriptor>    CreateJobDescriptor(const string& rid);

    /// returns Job Descriptors not marked as Deleted
    void    GetJobDescriptors(TJobDescrVec& descriptors);

    CRef<CNetBlastJobDescriptor>    FindJobDescriptor(const string& rid);

    void    StartRetrievingTask(TJobDescrVec& descriptors);
    void    StartMonitoringTask(TJobDescrVec& descriptors);

    void    DeleteJobDescriptors(TJobDescrVec& descriptors);

    void    OnExplore();
    void    OnLoadRIDs();

    /// @}

protected:
    typedef list< CConstRef<objects::CUser_field> > TUserFields;

    void    x_InitDatabases();
    void    x_ClearDatabases();

    bool    x_NeedToRefreshCachedDBs();
//    void    x_ReadCachedDBs();
    void    x_CategorizeDBs();
    void    x_GetDBTree(TUserFields& fields);
    void    x_AddCategorizedDBs(const TDbMap& valid_db_names, set<string>& categorized_dbs);

    void    x_SaveJobDescriptors();
    void    x_LoadJobDescriptors();

    /// callback - invoked by CNetBlastJobDescriptor
    void    x_OnJobDescrChanged(CNetBlastJobDescriptor& descr);

    void    x_StartTask(CBlastSearchTask& task);
    void    x_AutoStartMonitoring();

    void    x_SetDataBases(CRef<objects::CBlast4_get_databases_ex_reply> dbs);
    void    x_SetDBLoadJobId(CAppJobDispatcher::TJobID job_id);
    void    x_WaitForDBLoadJob();

protected:
    friend class CNetBlastLoadDBListJob;
protected:
    /// Indicated that Job execution shall be canceled
    CAtomicCounter_WithAutoInit   m_StopRequested;

    CUIObject   m_Descr;
    mutable CRef<CNetBLASTUIDataSourceType> m_Type;

    IServiceLocator*    m_SrvLocator;
    string m_RegPath;

    bool    m_Open;

    CRef<objects::CBlast4_get_databases_ex_reply>   m_Databases;
    CAppJobDispatcher::TJobID                    m_DBLoadJobId;
    TDbMap                                       m_NucDbMap;
    TDbMap                                       m_ProtDbMap;
    mutable CFastMutex                           m_DBLoadMutex;

    string                                       m_Nuc_DefDB;
    string                                       m_Prot_DefDB;
    TJobDescrVec                                 m_JobDescrs;
    mutable CFastMutex                           m_JobDescrMutex;
    vector<string>                               m_Nuc_DefMRU_DBs;
    vector<string>                               m_Prot_DefMRU_DBs;

    TTaxIdLabelMap                               m_TaxId2Label;
};


///////////////////////////////////////////////////////////////////////////////
/// CNetBLASTUIDataSourceType

class  CNetBLASTUIDataSourceType :
    public CObject,
    public IUIDataSourceType,
    public IExtension
{
public:
    CNetBLASTUIDataSourceType();

    /// @name IUIDataSourceType interface implementation
    /// @{
    virtual const IUIObject&    GetDescr();
    virtual IUIDataSource*  CreateDataSource();
    virtual bool    AutoCreateDefaultDataSource();
    /// @}

    /// @name IExtension interface implementation
    /// @{
    virtual string  GetExtensionIdentifier() const;
    virtual string  GetExtensionLabel() const;
    /// @}
protected:
    CUIObject   m_Descr;
};


/* @} */

END_NCBI_SCOPE;

#endif  // PKG_ALIGNMENT___NET_BLAST_UI_DATA_SOURCE__HPP

