#ifndef GUI_PACKAGES_SNP_GAP_TRACK__GAP_TRACK__HPP
#define GUI_PACKAGES_SNP_GAP_TRACK__GAP_TRACK__HPP

/*  $Id: gap_track.hpp 27815 2013-04-08 15:10:18Z rudnev $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Melvin Quintos
 *
 * File Description:
 *    This file contains class declarations for CGapTrack and CGapTrackFactory
 *
 */

#include <corelib/ncbiobj.hpp>
#include <gui/utils/extension.hpp>
#include <gui/widgets/seq_graphic/data_track.hpp>
#include <gui/widgets/seq_graphic/seqgraphic_job_result.hpp>
#include <gui/packages/pkg_snp/gap/gap_ds.hpp>
#include <gui/widgets/seq_graphic/layout_track.hpp>

#include <objects/seqloc/Seq_interval.hpp>
#include <objmgr/annot_selector.hpp>


BEGIN_NCBI_SCOPE

///////////////////////////////////////////////////////////////////////////////
//   CGapTrack
///////////////////////////////////////////////////////////////////////////////
class CGapTrack :
    public CDataTrack
{
    friend class CGapTrackFactory;

public:
    CGapTrack(CSGGapDS* ds, CRenderingContext* r_cntx);
    virtual ~CGapTrack();

    /// @name CDataTrack reimplemented methods.
    /// @{
    virtual CRef<CSGGenBankDS>  GetDataSource();
    virtual CConstRef<CSGGenBankDS>  GetDataSource() const;
    /// @}

    virtual void GetHTMLActiveAreas(TAreaVector* p_areas) const;
    virtual CHTMLActiveArea* InitHTMLActiveArea(TAreaVector* p_areas) const;

    /// @name CLayoutTrack reimplemented methods.
    /// @{
    virtual const CTrackTypeInfo&  GetTypeInfo() const;
    virtual string GetFullTitle() const;
    /// @}

    void SetAnnot(const string& name);

protected:
    virtual void x_LoadSettings(const string& preset_style,
                                const TKeyValuePairs& settings);

    virtual void x_OnIconClicked(TIconID id);

    /// @name CDataTrack pure virtual interfaces
    /// @{
    virtual void x_UpdateData();
    virtual void x_OnJobCompleted(CAppJobNotification& notify);
    /// @}

    void x_CreateLayout(const SGapJobResult& result);
    void x_UpdateLayout();

private:
    /// @name prohibited copy constructor and assignment operator.
    /// @{
    CGapTrack(const CGapTrack&);
    CGapTrack& operator=(const CGapTrack&);
    /// @}

    bool x_IsOverviewMode() const;

private:

    enum ELayout {
        eLayout_Adaptive = 260,  ///< All SNPs drawn in one line
        eLayout_Density,         ///< Always show density
        eLayout_Features,        ///< Show features whenever possible
        eLayout_Labels           ///< Show labels and cascade features (CLayeredLayoutPolicy)
    };

    int                     m_eLayout;

    string                  m_AnnotName;
    CRef<CSGGapDS>          m_DS;
    objects::SAnnotSelector m_FeatSel;

    SIconInfo               m_IconLayout;
    CRef<SGapJobResult>     m_JobResult;

    static CTrackTypeInfo   m_TypeInfo;


};

///////////////////////////////////////////////////////////////////////////////
//   CGapTrackFactory
///////////////////////////////////////////////////////////////////////////////
class NCBI_GUIWIDGETS_SNP_EXPORT CGapTrackFactory :
    public CObject,
    public ILayoutTrackFactory,
    public ITrackConfigurable,
    public IExtension
{
public:
    CGapTrackFactory(){}

    static const CTrackTypeInfo& GetTypeInfo()
        { return CGapTrack::m_TypeInfo; }

    /// @name ILayoutTrackFactory interface implementation
    /// @{
    virtual TTrackMap CreateTracks(
        SConstScopedObject& object,
        ISGDataSourceContext* ds_context,
        CRenderingContext* r_cntx,
        const SExtraParams& params = SExtraParams(),
        const TAnnotMetaDataList& src_annots = TAnnotMetaDataList()) const;

    virtual bool UnderstandLevel() const
    { return true; }

    virtual bool NeedBackgroundInit() const
    { return true; }

    virtual const CTrackTypeInfo&  GetThisTypeInfo() const
        { return GetTypeInfo(); }
    /// @}

    /// @name ITrackConfigurable reimplemented methods.
    /// @{
    CRef<objects::CTrackConfigSet>
    GetSettings(CSeqGraphicConfig* g_conf,
              const string& profile,
              const TKeyValuePairs& settings,
              const CTempTrackProxy* track_proxy) const;
    /// @}

    /// @name IExtension interface implementation
    /// @{
    virtual string GetExtensionIdentifier() const;
    virtual string GetExtensionLabel() const;
    /// @}
};

///////////////////////////////////////////////////////////////////////////////
///   CGapTrack inline methods
///////////////////////////////////////////////////////////////////////////////
inline /*virtual*/
CRef<CSGGenBankDS> CGapTrack::GetDataSource()
{
    return CRef<CSGGenBankDS>(m_DS.GetPointer());
}

inline /*virtual*/
CConstRef<CSGGenBankDS> CGapTrack::GetDataSource() const
{
    return CConstRef<CSGGenBankDS>(m_DS.GetPointer());
}


END_NCBI_SCOPE

#endif  // GUI_PACKAGES_SNP_GAP_TRACK__GAP_TRACK__HPP
