#ifndef GUI_PACKAGES_SNP_GAP_TRACK__GWAS_GLYPH__HPP
#define GUI_PACKAGES_SNP_GAP_TRACK__GWAS_GLYPH__HPP

/*  $Id: gwas_glyph.hpp 27829 2013-04-09 21:26:28Z rudnev $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Melvin Quintos
 *
 * File Description:
 *    CGwasGlyph -- Glyph to represent gwas track
 *
 */


#include <corelib/ncbiobj.hpp>
#include <objects/seqloc/Seq_interval.hpp>
#include <objmgr/scope.hpp>
#include <gui/utils/rgba_color.hpp>
#include <gui/objutils/density_map.hpp>
#include <gui/widgets/seq_graphic/seq_glyph.hpp>

BEGIN_NCBI_SCOPE

class NCBI_GUIWIDGETS_SNP_EXPORT CGwasGlyph : public CSeqGlyph
{
////////////////////////////////////////////
// Public Structs/Classes/Typedefs
////////////////////////////////////////////
public:
    struct SValue {
        int             count;
        float           pvalue; // -log10(raw pvalue)
        unsigned int    rsid;

        SValue(int x) { count = x; pvalue = 0; rsid = 0; }
        SValue()      { count = 0; pvalue = 0; rsid = 0; }

        bool operator < (const SValue &rhs) const
        {
            return (pvalue < rhs.pvalue);
        }
    };

    struct SBin : public CObject {
        SValue    val;
        TSeqRange range;
        string    signature;  // object signature.  See obj_fingerprint.hpp
    };

    friend SValue operator+(const SValue& val1, const SValue& val2);

    struct accum_study : public binary_functor<SValue> {
        virtual SValue operator() (const SValue& x,
                                     const SValue& y) const
        {
            return x + y;
        }

        virtual binary_functor<SValue>* clone() const
        {
            return new accum_study;
        }
    };

    typedef CDensityMap<SValue>   TDensityMap;

//////////////////////////////////////////////
// Public Methods
//////////////////////////////////////////////
public:

    /// @name Static Methods
    /// @{
    // if the given title is empty, will generate a special signature for the track caption
    static string GenerateBinSignature(const string& title,
                                       const string& annot_name,
                                       int gi,
                                       TSeqRange range);
    /// @}

    /// @name ctors
    /// @{

    CGwasGlyph(const TDensityMap&       values,
               const objects::CSeq_loc& loc,
               objects::CScope *        scope);
    /// @}

    /// @name CSeqGlyph virtual methods.
    /// @{
    virtual void GetHTMLActiveAreas(TAreaVector* p_areas) const;
    virtual bool OnLeftDblClick(const TModelPoint& p);
    virtual bool NeedTooltip(const TModelPoint& p, string& tt, string& t_title) const;
    virtual void GetTooltip(const TModelPoint& p, string& tt, string& t_title) const;
    virtual bool IsClickable() const;

    virtual TSeqRange GetRange(void) const;
    ///@}

    /// @name Non-Inherited Getters
    /// @{

    /// Access the first density map.
    const TDensityMap& GetDensityMap(void) const;

    // Get range for a particular bin
    TSeqRange GetBinRangeAt(TSeqPos pos) const;

    ///@}

    void  SetTitle(const string& title);
    const string& GetTitle() const;

    void SetAnnotName(const string& name);
    const string& GetAnnotName() const;


//////////////////////////////////////////////
// Protected Methods
//////////////////////////////////////////////
protected:
    /// @name CSeqGlyph virtual methods.
    /// @{
    virtual void x_Draw() const;
    virtual void x_UpdateBoundingBox();
    ///@}


//////////////////////////////////////////////
// Private Methods
//////////////////////////////////////////////
private:
    // This populates 'm_listObjs'.  It is labeled 'const'
    //   but modifies the mutable m_listObjs
    void        x_BuildObjectList() const;
    TSeqRange   x_GetBinRange(unsigned int bin) const;

private:
    string                          m_Title;
    string                          m_AnnotName;
    CConstRef<objects::CSeq_loc>    m_Location;
    TDensityMap                     m_Map;

    mutable bool                    m_bHasObjectList;

    mutable vector<CConstRef<CObject> >     m_listObjs;
    mutable CRef<objects::CScope>           m_Scope;
};

///////////////////////////////////////////////////////////////////////////////
/// CHistogramGlyph inline method implementation.
inline
const CGwasGlyph::TDensityMap& CGwasGlyph::GetDensityMap(void) const
{
    return m_Map;
}

inline
void CGwasGlyph::SetTitle(const string& name)
{ m_Title = name; }

inline
const string& CGwasGlyph::GetTitle() const
{ return m_Title; }


inline
void CGwasGlyph::SetAnnotName(const string& name)
{ m_AnnotName = name; }

inline
const string& CGwasGlyph::GetAnnotName() const
{ return m_AnnotName; }


inline
CGwasGlyph::SValue operator+(const CGwasGlyph::SValue& val1,
                             const CGwasGlyph::SValue& val2)
{
    CGwasGlyph::SValue t;
    t.count  = val1.count + val2.count;
    t.pvalue = std::max(val1.pvalue, val2.pvalue);
    t.rsid   = val1.pvalue > val2.pvalue ? val1.rsid : val2.rsid;
    return t;
}

END_NCBI_SCOPE

/* @} */

#endif  // GUI_PACKAGES_SNP_GAP_TRACK__GWAS_GLYPH__HPP
