#ifndef GUI_WIDGETS_SNP___FILTER2_HPP
#define GUI_WIDGETS_SNP___FILTER2_HPP

/*  $Id: filter2.hpp 18224 2008-11-04 17:28:57Z dicuccio $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Melvin Quintos
 *
 * File Description:
 *
 */

#include <util/xregexp/regexp.hpp>

BEGIN_NCBI_SCOPE


class CFilter2
{
public:
    virtual ~CFilter2() {}
    enum EWhere {
        eEqual, eNotEqual, eGreaterThan, eLessThan,
        eGreaterThanOrEqual, eLessThanOrEqual,
        eLike, eCustom };

};

template <typename T>
class CFilterT : virtual public CFilter2
{
public:
    typedef T class_type;

public:
    virtual ~CFilterT() {}
    virtual bool Passes(const T& v) const =0;
};

template< typename T>
class CFilterDefault : public CFilterT<T>
{
public:
    CFilterDefault(const T& v, CFilter2::EWhere e) : m_Value(v), m_eWhere(e) {}
    virtual bool Passes(const T &v) const {
        switch(m_eWhere) {
        case CFilter2::eEqual:                return  v == m_Value;
        case CFilter2::eNotEqual:             return  v != m_Value;
        case CFilter2::eGreaterThan:          return  v > m_Value;
        case CFilter2::eLessThan:             return  v < m_Value;
        case CFilter2::eGreaterThanOrEqual:   return  v >= m_Value;
        case CFilter2::eLessThanOrEqual:      return  v <= m_Value;
        default:                              return true;
        }
        return true;
    }
private:
    T                m_Value;
    CFilter2::EWhere  m_eWhere;
};

template <typename T>
class CFilterNull : public CFilterT<T>
{
public:
    CFilterNull() {}
    virtual bool Passes(const T &v) const { return false; }
};

// Filter using regular expression
class CFilterRegEx : public CFilterT<string> {
public:
    CFilterRegEx(const string &regex) : m_Regex(regex)
    {
    }

    virtual bool Passes(const string & v) const{
        return m_Regex.IsMatch(v);
    }
private:
    mutable CRegexp m_Regex;
};
class CFilterSet2
{
public:
    virtual ~CFilterSet2();

    template <typename T>
    void Add(const string &name, CFilter2::EWhere e, T value);

    template <typename T>
    void AddCustom(const string &name, CFilterT<T> *);

    template <typename T>
    CFilterT<T> & GetFilter(string name);

    bool HasFilter(string name) { return m_mapFilters.find(name) != m_mapFilters.end(); }


private:
    map<string, CFilter2*> m_mapFilters;
};

inline CFilterSet2::~CFilterSet2()
{
    typedef map<string, CFilter2*> TMapFilters;
    TMapFilters::iterator iter;
    iter = m_mapFilters.begin();
    for( ; iter != m_mapFilters.end(); ++iter) {
        delete iter->second;
    }
}

template <typename T>
inline
CFilterT<T>& CFilterSet2::GetFilter(string name)
{
    if( m_mapFilters.find(name) == m_mapFilters.end() ) {
        m_mapFilters[name] = new CFilterNull<T>();
    }

    CFilterT<T> *filter = dynamic_cast<CFilterT<T> *>(m_mapFilters[name]);

    _ASSERT(filter != 0); // wrong type queried
    if(filter==0) throw std::bad_cast("mismatch of filter's template type and argument type");

    return *filter;
}

template <typename T>
inline
void CFilterSet2::AddCustom(const string &name, CFilterT<T> *customFilter)
{
    if( m_mapFilters.find(name) != m_mapFilters.end() )
        delete m_mapFilters[name];
    m_mapFilters[name] = customFilter;
}

template<>
inline
void CFilterSet2::Add<string>(const string &name, CFilter2::EWhere e, string value)
{
    if( m_mapFilters.find(name) != m_mapFilters.end() )
        delete m_mapFilters[name];

    CFilter2* filter = 0;
    if( e==CFilter2::eLike ) // create regular expression
        filter = new CFilterRegEx(value);
    else                    // otherwise use default comparisons
        filter = new CFilterDefault<string>(value, e);;

    m_mapFilters[name] = filter;

}

template <typename T>
inline
void CFilterSet2::Add(const string &name, CFilter2::EWhere e, T value)
{
    if( m_mapFilters.find(name) != m_mapFilters.end() )
        delete m_mapFilters[name];

    _ASSERT( e != CFilter2::eLike ); // eLike must use string specialization
    if(e == CFilter2::eLike)
        throw std::runtime_error("cannot use CFilter2::eLike with non-string value");

    m_mapFilters[name] = new CFilterDefault<T>(value, e);
}

END_NCBI_SCOPE

#endif // GUI_WIDGETS_SNP___FILTER2_HPP

