/*  $Id: table_data_snp.cpp 27969 2013-05-01 21:12:34Z rudnev $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors: Dmitry Rudnev
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>

#include <gui/objutils/label.hpp>
#include <gui/objutils/table_data_snp.hpp>
#include <gui/objutils/utils.hpp>
#include <gui/utils/app_job.hpp>
#include <gui/utils/app_job_impl.hpp>

#include <objmgr/util/create_defline.hpp>
#include <objmgr/util/sequence.hpp>

#include <objects/seq/Seq_annot.hpp>
#include <objects/seq/Seq_literal.hpp>
#include <objects/seq/Seq_data.hpp>
#include <objects/general/Dbtag.hpp>
#include <objects/general/Object_id.hpp>
#include <objects/general/User_object.hpp>
#include <objects/general/User_field.hpp>
#include <objects/seqfeat/Variation_inst.hpp>
#include <objects/seqfeat/Delta_item.hpp>
#include <objtools/snputil/snp_utils.hpp>
#include <objects/variation/VariantPlacement.hpp>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);


class CTableDataSNPJob : public CJobCancelable
{
public:
    CTableDataSNPJob( CScope& scope, const CSeq_loc& loc);
    virtual ~CTableDataSNPJob() {}

    /// @name IAppJob implementation
    /// @{
    virtual EJobState                   Run();
    virtual CConstIRef<IAppJobProgress> GetProgress();
    virtual CRef<CObject>               GetResult();
    virtual CConstIRef<IAppJobError>    GetError();
    virtual string                      GetDescr() const;
    /// @}

private:
    void x_FetchAll();

private:
    CRef<CScope>            m_Scope;
    CConstRef<CSeq_loc>     m_Loc;
    CRef<CObject> m_listSnps;

    size_t  m_Total;
    size_t  m_Completed;
    CRef<CAppJobError>      m_Error;
};


CTableDataSNPJob::CTableDataSNPJob(CScope& scope, const CSeq_loc& loc)
:   m_Scope(&scope),
    m_Loc(&loc),
	m_Total(100),
    m_Completed(0)
{
    m_listSnps.Reset(new CObjectFor<CTableDataSNP::TSnpDataList>());
}

CConstIRef<IAppJobProgress> CTableDataSNPJob::GetProgress()
{
    CRef<CAppJobProgress> p( new CAppJobProgress() );
    p->SetNormDone( float(m_Completed) / float(m_Total) );
    return CConstIRef<IAppJobProgress>( p.GetPointer() );
}

CRef<CObject> CTableDataSNPJob::GetResult()
{
    return m_listSnps;
}


CConstIRef<IAppJobError> CTableDataSNPJob::GetError()
{
    return CConstIRef<IAppJobError>( m_Error.GetPointer() );
}


string CTableDataSNPJob::GetDescr() const
{
    return "SNP table data loader job";
}


IAppJob::EJobState CTableDataSNPJob::Run()
{
    x_FetchAll();

    return IsCanceled() ? eCanceled : eCompleted;
}


void CTableDataSNPJob::x_FetchAll()
{
    SAnnotSelector      sel;       // annotation selector
    size_t              size;      // number of features returned (not all of them will be used though)

    CTableDataSNP::TSnpDataList& snp_list(dynamic_cast<CObjectFor<CTableDataSNP::TSnpDataList>*>(m_listSnps.GetPointer())->GetData());

    // Prepare Annotation Selection to find the SNPs
    sel = CSeqUtils::GetAnnotSelector(CSeqFeatData::eSubtype_variation);
    sel.AddNamedAnnots("SNP")
        .SetExcludeExternal(false)
        .SetMaxSize(CTableDataSNP::c_MaxRows);

    // Fetch the SNPs
    CFeat_CI feat_it(*m_Scope, *m_Loc, sel);

    // Determine the amount of features returned
    size = feat_it.GetSize();

    // make sure there is data
    if (size < 1)
        return;
    m_Total = size;

    // Prepare the list, and populate it
    snp_list.reserve(size);
    
    for (; feat_it; ++feat_it) {
        ++m_Completed;
        const CSeq_feat& or_feat(feat_it->GetOriginalFeature());

        // Grab the database info
        CConstRef<CDbtag> tag;
        CRef<CVariation> pVariation(new CVariation());
        if(!NSNPVariationHelper::ConvertFeat(*pVariation, or_feat))
            continue;

        // Retrieve gene associated with SNP
        tag.Reset();  // reuse tag reference.
        CConstRef<CSeq_feat> gene_feat;
        gene_feat = sequence::GetBestOverlapForSNP(or_feat, 
                        CSeqFeatData::eSubtype_gene,
                        *m_Scope);
        if (gene_feat) {
            tag = gene_feat->GetNamedDbxref("GeneID");
            if (!tag) {
                tag = gene_feat->GetNamedDbxref("LocusLink");
            }
        }
        if (tag) {
            pVariation->SetOther_ids().push_back(CRef<CDbtag>(const_cast<CDbtag*>(tag.GetPointer())));
        }

        snp_list.push_back(pVariation);
    }
}


BEGIN_EVENT_MAP( CTableDataSNP, CEventHandler )
    ON_MESSAGE(CAppJobNotification, CAppJobNotification::eStateChanged, &CTableDataSNP::OnJobNotification)
END_EVENT_MAP()

// maximal amount of SNP we should realistically see in a SNP table
const size_t CTableDataSNP::c_MaxRows(100000);

CTableDataSNP::CTableDataSNP(TConstScopedObjects& objects)
:
m_DataLoadStatus(eNotReady)
{
    _ASSERT( !objects.empty() );
    m_Scope = objects.front().scope;
    m_OrigLoc.Reset(dynamic_cast<const CSeq_loc*>(objects[0].object.GetPointer()));

    if (m_OrigLoc) {
        x_LoadData();
    }
}

CTableDataSNP::~CTableDataSNP()
{
    if(m_DataLoadStatus == eNotReady) {
        CAppJobDispatcher& disp = CAppJobDispatcher::GetInstance();
        disp.DeleteJob(m_hJob);
    }
}



void CTableDataSNP::x_LoadData()
{
    CRef<CTableDataSNPJob> job(new CTableDataSNPJob(*m_Scope, *m_OrigLoc));
    CAppJobDispatcher& disp = CAppJobDispatcher::GetInstance();
    try {
        m_hJob = disp.StartJob(*job, "ObjManagerEngine", *this);
    } catch( CAppJobException& e ){
        m_DataLoadStatus = eFailed;
        LOG_POST( Error
            << "CTableDataSNP failed to start job: "
            << e.GetMsg()
        );
        LOG_POST(e.ReportAll());
    }
}


void CTableDataSNP::OnJobNotification(ncbi::CEvent *evt)
{
    CAppJobNotification* notif = dynamic_cast<CAppJobNotification*>( evt );

    _ASSERT(notif);

    int                 job_id  = notif->GetJobID();
    IAppJob::EJobState  state   = notif->GetState();

    if (job_id != m_hJob) return;

    switch (state) {
        case IAppJob::eCompleted:
        {
            CAppJobDispatcher& disp(CAppJobDispatcher::GetInstance());
            m_listSnps.Reset(dynamic_cast<CObjectFor<CTableDataSNP::TSnpDataList>*>(disp.GetJobResult(job_id).GetPointer()));
            m_DataLoadStatus = eReady;

            // send a message to the view that data has changed (i.e. ready)
            CTableDataEvent evt(CEvent::eEvent_Message, CTableDataEvent::eDataChanged);
            CTableDataEvent::TDataAttachment* attachment = new CTableDataEvent::TDataAttachment();
            CIRef<ITableData> table_data;
            table_data.Reset(this);
            attachment->SetData(table_data);
            evt.SetAttachment(attachment, CEvent::eRelease);
            Send(&evt, ePool_Parent);
            break;
        }
        case IAppJob::eCanceled:
        case IAppJob::eFailed:
            m_DataLoadStatus = eFailed;
            break;
        default:
            break;
    }
}




// number of entries and order must correspond with CTableDataSNP::EColContents
static const string s_ColNames[] = {
    "RS ID",
    "Alleles",
    "Sequence",
    "Location",
    "FXN Class",
    "Weight",
    // "Pop Freq", -- dropped as asked in SNP-4726
    "Variation Class",
    "Resource Link",
    // "Mapping", -- dropped as asked in SNP-4726
    "Frequency Validation",
    // "Haplotype Set", -- dropped as asked in SNP-4726
    "Quality Check",
    "Phenotype",
    "1000 Genomes"
};




ITableData::ColumnType CTableDataSNP::GetColumnType(size_t col) const
{
    if(m_DataLoadStatus != eReady) {
        return ITableData::kString;
    }
    switch(col){
        case eWeight:
        case ePosition:
            return ITableData::kInt;

        case eRsId:
            return ITableData::kInt;

        case eResourceLink:
        case ePhenotype:
//        case ePopFreq: -- dropped as asked in SNP-4726
        case eFXNClass:
            return ITableData::kString;

        default:
            return ITableData::kString;
    }
    return ITableData::kNone;
}


bool CTableDataSNP::IsHtmlColumn(size_t col) const
{
    if(m_DataLoadStatus != eReady) {
        return false;
    }
    switch(col){
        case eRsId:
        case eResourceLink:
        case ePhenotype:
        case eFXNClass:
            return true;
        default:
            return false;
    }
}

string CTableDataSNP::GetColumnLabel(size_t col) const
{
    if(m_DataLoadStatus != eReady && col == 0) {
        return "Status";
    }
    if (col < CTableDataSNP::eMaxColNum)
        return s_ColNames[col];

    return kEmptyStr;
}


size_t CTableDataSNP::GetRowsCount() const
{
    return m_DataLoadStatus == eReady ? m_listSnps->GetData().size() : 1;
}


size_t CTableDataSNP::GetColsCount() const
{
    return (size_t)CTableDataSNP::eMaxColNum;
}


string CTableDataSNP::GetStringValue(size_t row, size_t col) const
{
    if(m_DataLoadStatus == eNotReady) {
        return col == 0 ? "Loading..." : "";
    }
    if(m_DataLoadStatus == eFailed) {
        return col == 0 ? "Load failed!" : "";
    }
    if(row+1 == c_MaxRows) {
        return "Warning: Only the first " + NStr::NumericToString(c_MaxRows) + " rows can be shown";
    }
    list<string> values;
    string sDelimiter(", ");
    if (row < GetRowsCount()) {
        CConstRef<CVariation> pVariation(x_GetRowData(row));
        bool isProp(pVariation->CanGetVariant_prop());
        CSnpBitfield bf;
        if(pVariation->CanGetExt()) {
            const CVariation::TExt& Ext(pVariation->GetExt());
            ITERATE(CVariation::TExt, iExt, Ext) {
                if((*iExt)->CanGetClass() &&
                    (*iExt)->GetClass() == SNP_VAR_EXT_CLASS &&
                    (*iExt)->GetFieldRef(SNP_VAR_EXT_BITFIELD).NotEmpty() &&
                    (*iExt)->GetFieldRef(SNP_VAR_EXT_BITFIELD)->CanGetData() &&
                    (*iExt)->GetFieldRef(SNP_VAR_EXT_BITFIELD)->GetData().IsOs()) {
                        bf = (*iExt)->GetFieldRef(SNP_VAR_EXT_BITFIELD)->GetData().GetOs();
                }
            }
        }
        bool isBitfield(bf.GetVersion() > 0);

        switch(col){
            case ePosition:
            case eWeight:
                values.push_back(NStr::NumericToString(GetIntValue(row, col)));
                break;

            case eRsId:
            {
                string sValue;
                if(pVariation->GetId().CanGetTag()) {
                    CNcbiOstrstream ostr;
                    pVariation->GetId().GetTag().AsString(ostr);
                    sValue = CNcbiOstrstreamToString(ostr);
                } else
		            pVariation->GetId().GetLabel(&sValue);

                values.push_back(sValue);
                break;
            }
            case eResourceLink:
            {
                list<string> Texts;
                list<string> URLs;
                if(isProp) {
                    NSNPVariationHelper::VariantPropAsStrings(Texts, 
                                                                pVariation->GetVariant_prop(),
                                                                NSNPVariationHelper::eSNPPropName_ResourceLink);
                    NSNPVariationHelper::VariantPropAsStrings(URLs, 
                                                                pVariation->GetVariant_prop(),
                                                                NSNPVariationHelper::eSNPPropName_ResourceLinkURL);
                }
                /* the entities that do not have a link are disabled per SNP-4726
                //!! no representation in CVariantProperties for eHasPubmedArticle, so use bitfield
                if(isBitfield && bf.IsTrue(CSnpBitfield::eHasPubmedArticle)) {
                    Texts.push_back("PubMed Central article");
                }
                //!! no representation in CVariantProperties for eHasShortReadArchive, so use bitfield
                if(isBitfield && bf.IsTrue(CSnpBitfield::eHasShortReadArchive)) {
                    Texts.push_back("Short Read Archive");
                }
                //!! no representation in CVariantProperties for eHasOrganism, so use bitfield
                if(isBitfield && bf.IsTrue(CSnpBitfield::eHasOrganism)) {
                    Texts.push_back("OrganismDBLink");
                }
                //!! no representation in CVariantProperties for eFromMgcClone, so use bitfield
                if(isBitfield && bf.IsTrue(CSnpBitfield::eFromMgcClone)) {
                    Texts.push_back("From MGC clone");
                }
                //!! no representation in CVariantProperties for eHasTrace, so use bitfield
                if(isBitfield && bf.IsTrue(CSnpBitfield::eHasTrace)) {
                    Texts.push_back("Trace Archive");
                }
                //!! no representation in CVariantProperties for eHasAssembly, so use bitfield
                if(isBitfield && bf.IsTrue(CSnpBitfield::eHasAssembly)) {
                    Texts.push_back("Assembly Archive");
                }
                //!! no representation in CVariantProperties for eHasGEO, so use bitfield
                if(isBitfield && bf.IsTrue(CSnpBitfield::eHasGEO)) {
                    Texts.push_back("Entrez GEO");
                }
                //!! no representation in CVariantProperties for eHasProbeDB, so use bitfield
                if(isBitfield && bf.IsTrue(CSnpBitfield::eHasProbeDB)) {
                    Texts.push_back("ProbeDB");
                }
                //!! no representation in CVariantProperties for eHasEntrez, so use bitfield
                if(isBitfield && bf.IsTrue(CSnpBitfield::eHasEntrez)) {
                    Texts.push_back("Entrez Gene");
                }
                //!! no representation in CVariantProperties for eHasSTS, so use bitfield
                if(isBitfield && bf.IsTrue(CSnpBitfield::eHasSTS)) {
                    Texts.push_back("Entrez STS");
                }
                */
                // scan and skip values that do not have an URL per SNP-4726
                list<string>::const_iterator iTexts(Texts.begin());
                ITERATE(list<string>, iURLs, URLs) {
                    if(!iURLs->empty()) {
                        values.push_back(*iTexts);
                    }
                    ++iTexts;
                }
                break;
            }
/*            case ePopFreq:  -- dropped as asked in SNP-4726
                if(isProp && 
                    pVariation->GetVariant_prop().CanGetGenotype() &&
                    pVariation->GetVariant_prop().GetGenotype() & CVariantProperties::eGenotype_has_genotypes) {
                    values.push_back("GTY");
                }
                break;
*/
            case eAlleles:
                // for now, just extract allele strings out of the deltas and catenate them with '/'
                // if a delta is empty, show it as "-"
                sDelimiter = "/";
                NSNPVariationHelper::GetDeltas<CVariation>(values, pVariation);
				// testing the filter for GB-2588:
				// values.push_back(">5pct");
                break;

            case eSequence:
                if(pVariation->CanGetPlacements()) {
                    const CVariation::TPlacements& VarPlacements(pVariation->GetPlacements());
                    if(!VarPlacements.empty()) {
                        CRef<CVariantPlacement> pPlacement(VarPlacements.front());
                        if(pPlacement->CanGetLoc()) {
                            CSeq_id_Handle idh = sequence::GetIdHandle(pPlacement->GetLoc(), const_cast<CScope*>(m_Scope.GetPointer()));
                            idh = sequence::GetId(idh, *const_cast<CScope*>(m_Scope.GetPointer()), sequence::eGetId_Best);
                            string sLabel;
                            idh.GetSeqId()->GetLabel(&sLabel, CSeq_id::eContent);
                            values.push_back(sLabel);
                        }
                    }
                }
                break;

            case eFXNClass:
                if(isProp) {
                    list<string> GeneLocations;
                    NSNPVariationHelper::VariantPropAsStrings(GeneLocations,
                                                                pVariation->GetVariant_prop(),
                                                                NSNPVariationHelper::eSNPPropName_GeneLocation);
                    list<string> Effects;
                    NSNPVariationHelper::VariantPropAsStrings(Effects,
                                                                pVariation->GetVariant_prop(),
                                                                NSNPVariationHelper::eSNPPropName_Effect);
                    values.swap(GeneLocations);
                    values.insert(values.end(), Effects.begin(), Effects.end());
                }
                break;

            case eVariationClass:
                //!! kludgy thing, use bitfield to recover a lot of variation classes
                if(pVariation->CanGetData() && pVariation->GetData().IsInstance()) {
                    switch(pVariation->GetData().GetInstance().GetType()) {
                        case CVariation_inst::eType_snv:
                            values.push_back("SNP");
                            break;
                        case CVariation_inst::eType_mnp:
                            values.push_back("MNP");
                            break;
                        case CVariation_inst::eType_delins:
                            values.push_back("INDEL");
                            break;
                        case CVariation_inst::eType_microsatellite:
                            values.push_back("MSAT");
                            break;
                        default:
                            break;
                    }
                } else {
                    //!! a catch-all for all variation classes not properly captured in CVariation
                }
                //!! when in doubt, try to get data from the original bitfield, while we use this kludge
                if(isBitfield && values.empty())
                    values.push_back(bf.GetVariationClassString());
                break;
            
/*            case eMapping: -- dropped as asked in SNP-4726
                if(isProp)
                    NSNPVariationHelper::VariantPropAsStrings(values, 
                                                                pVariation->GetVariant_prop(),
                                                                NSNPVariationHelper::eSNPPropName_Mapping);
                break;
*/
            case eFreqValidation:
                if(isProp)
                    NSNPVariationHelper::VariantPropAsStrings(values,
                                                                pVariation->GetVariant_prop(),
                                                                NSNPVariationHelper::eSNPPropName_FreqValidation);
                //!! no representation in CVariantProperties for eGMAF0dot01, so use bitfield
                if(isBitfield && bf.IsTrue(CSnpBitfield::eGMAF0dot01)) {
                    values.push_back("GMAF >= 0.01");
                }
                break;

/*            case eInHaplotypeSet: -- dropped as asked in SNP-4726
                values.push_back((isProp && 
                                pVariation->GetVariant_prop().CanGetGenotype() &&
                                pVariation->GetVariant_prop().GetGenotype() & CVariantProperties::eGenotype_in_haplotype_set)
                                   ?
                                   "Y" : "N");
                break;
*/
            case eQualityCheck:
                if(isProp)
                    NSNPVariationHelper::VariantPropAsStrings(values,
                                                                pVariation->GetVariant_prop(),
                                                                NSNPVariationHelper::eSNPPropName_QualityCheck);
                //!! no representation in CVariantProperties for eIsSomatic, so use bitfield
                if(isBitfield && bf.IsTrue(CSnpBitfield::eIsSomatic)) {
                    values.push_back("Somatic, not germline");
                }
                //!! no representation in CVariantProperties for eIsSuspect, so use bitfield
                if(isBitfield && bf.IsTrue(CSnpBitfield::eIsSuspect)) {
                    values.push_back("Suspect");
                }
                break;

            case ePhenotype:
                //!! no representation in CVariantProperties for eHasMeSH, so use bitfield
                if(isBitfield && bf.IsTrue(CSnpBitfield::eHasMeSH)) {
                    values.push_back("MeSH");
                }
                //!! no representation in CVariantProperties for eHasClinicalAssay, so use bitfield
                if(isBitfield && bf.IsTrue(CSnpBitfield::eHasClinicalAssay)) {
                    values.push_back("Clinical diagnostic assay");
                }
                //!! no representation in CVariantProperties for eHasTranscriptionFactor, so use bitfield
                if(isBitfield && bf.IsTrue(CSnpBitfield::eHasTranscriptionFactor)) {
                    values.push_back("Transcription factor");
                }
                //!! no representation in CVariantProperties for eHasDiseaseInfo, so use bitfield
                if(isBitfield && bf.IsTrue(CSnpBitfield::eHasDiseaseInfo)) {
                    values.push_back("From LSDB");
                }
                //!! no representation in CVariantProperties for eHasLowdbGaPPValue, so use bitfield
                if(isBitfield && bf.IsTrue(CSnpBitfield::eHasLowdbGaPPValue)) {
                    values.push_back("p-Value <= 0.001 in dbGaP");
                }
                //!! no representation in CVariantProperties for eHasLodScore, so use bitfield
                if(isBitfield && bf.IsTrue(CSnpBitfield::eHasLodScore)) {
                    values.push_back("LOD score >= 2.0 in dbGaP");
                }
                //!! no representation in CVariantProperties for eHasMicroattribution, so use bitfield
                if(isBitfield && bf.IsTrue(CSnpBitfield::eHasMicroattribution)) {
                    values.push_back("Microattribution");
                }
                //!! no representation in CVariantProperties for eHasOMIM_OMIA, so use bitfield
                if(isBitfield && bf.IsTrue(CSnpBitfield::eHasOMIM_OMIA)) {
                    values.push_back("OMIM/OMIA");
                }
                break;
            
            case e1000Genomes:
                //!! no representation in CVariantProperties for eTGPValidated, so use bitfield
                if(isBitfield && bf.IsTrue(CSnpBitfield::eTGPValidated)) {
                    values.push_back("1000G validated");
                }
                //!! no representation in CVariantProperties for eTGPPhase1, so use bitfield
                if(isBitfield && bf.IsTrue(CSnpBitfield::eTGPPhase1)) {
                    values.push_back("1000G Phase 1");
                }
                //!! no representation in CVariantProperties for eTGPPilot, so use bitfield
                if(isBitfield && bf.IsTrue(CSnpBitfield::eTGPPilot)) {
                    values.push_back("1000G pilot");
                }
                //!! no representation in CVariantProperties for eTGPOnlyNotExclusive, so use bitfield
                if(isBitfield && bf.IsTrue(CSnpBitfield::eTGPOnlyNotExclusive)) {
                    values.push_back("Has 1000G submission");
                }
                //!! no representation in CVariantProperties for eTGPNoneNotExclusive, so use bitfield
                if(isBitfield && bf.IsTrue(CSnpBitfield::eTGPNoneNotExclusive)) {
                    values.push_back("Not present in 1000G");
                }
                break;
            
            default:
                break;
        }
    }
    return NStr::Join(values, sDelimiter);
}


string CTableDataSNP::GetHtmlValue(size_t row, size_t col) const
{
    list<string> values;
    string sDelimiter(", ");
    if (row < GetRowsCount()) {
        CConstRef<CVariation> pVariation(x_GetRowData(row));
        bool isProp(pVariation->CanGetVariant_prop());
        CSnpBitfield bf;
        if(pVariation->CanGetExt()) {
            const CVariation::TExt& Ext(pVariation->GetExt());
            ITERATE(CVariation::TExt, iExt, Ext) {
                if((*iExt)->CanGetClass() &&
                    (*iExt)->GetClass() == SNP_VAR_EXT_CLASS &&
                    (*iExt)->GetFieldRef(SNP_VAR_EXT_BITFIELD).NotEmpty() &&
                    (*iExt)->GetFieldRef(SNP_VAR_EXT_BITFIELD)->CanGetData() &&
                    (*iExt)->GetFieldRef(SNP_VAR_EXT_BITFIELD)->GetData().IsOs()) {
                        bf = (*iExt)->GetFieldRef(SNP_VAR_EXT_BITFIELD)->GetData().GetOs();
                }
            }
        }
        bool isBitfield(bf.GetVersion() > 0);

        switch(col){
            case eRsId:
            {
                string sValue;
                string sURL(pVariation->GetId().GetUrl());
                if(pVariation->GetId().CanGetTag()) {
                    CNcbiOstrstream ostr;
                    pVariation->GetId().GetTag().AsString(ostr);
                    sValue = CNcbiOstrstreamToString(ostr);
                } else
		            pVariation->GetId().GetLabel(&sValue);

                if(sURL.empty())
                    values.push_back(sValue);
                else
                    values.push_back("<a href=\"" + sURL + "\">" + sValue + "</a>");
                break;
            }
            case eResourceLink:
            {
                list<string> Texts;
                list<string> URLs;
                sDelimiter = "<br>";
                if(isProp) {
                    NSNPVariationHelper::VariantPropAsStrings(Texts, 
                                                                pVariation->GetVariant_prop(),
                                                                NSNPVariationHelper::eSNPPropName_ResourceLink);
                    NSNPVariationHelper::VariantPropAsStrings(URLs, 
                                                                pVariation->GetVariant_prop(),
                                                                NSNPVariationHelper::eSNPPropName_ResourceLinkURL);
                }
                //!! no representation in CVariantProperties for eHasPubmedArticle, so use bitfield
                if(isBitfield && bf.IsTrue(CSnpBitfield::eHasPubmedArticle)) {
                    Texts.push_back("PubMed Central article");
                    URLs.push_back("");
                }
                //!! no representation in CVariantProperties for eHasShortReadArchive, so use bitfield
                if(isBitfield && bf.IsTrue(CSnpBitfield::eHasShortReadArchive)) {
                    Texts.push_back("Short Read Archive");
                    URLs.push_back("");
                }
                //!! no representation in CVariantProperties for eHasOrganism, so use bitfield
                if(isBitfield && bf.IsTrue(CSnpBitfield::eHasOrganism)) {
                    Texts.push_back("OrganismDBLink");
                    URLs.push_back("");
                }
                //!! no representation in CVariantProperties for eFromMgcClone, so use bitfield
                if(isBitfield && bf.IsTrue(CSnpBitfield::eFromMgcClone)) {
                    Texts.push_back("From MGC clone");
                    URLs.push_back("");
                }
                //!! no representation in CVariantProperties for eHasTrace, so use bitfield
                if(isBitfield && bf.IsTrue(CSnpBitfield::eHasTrace)) {
                    Texts.push_back("Trace Archive");
                    URLs.push_back("");
                }
                //!! no representation in CVariantProperties for eHasAssembly, so use bitfield
                if(isBitfield && bf.IsTrue(CSnpBitfield::eHasAssembly)) {
                    Texts.push_back("Assembly Archive");
                    URLs.push_back("");
                }
                //!! no representation in CVariantProperties for eHasGEO, so use bitfield
                if(isBitfield && bf.IsTrue(CSnpBitfield::eHasGEO)) {
                    Texts.push_back("Entrez GEO");
                    URLs.push_back("");
                }
                //!! no representation in CVariantProperties for eHasProbeDB, so use bitfield
                if(isBitfield && bf.IsTrue(CSnpBitfield::eHasProbeDB)) {
                    Texts.push_back("ProbeDB");
                    URLs.push_back("");
                }
                //!! no representation in CVariantProperties for eHasEntrez, so use bitfield
                if(isBitfield && bf.IsTrue(CSnpBitfield::eHasEntrez)) {
                    Texts.push_back("Entrez Gene");
                    URLs.push_back("");
                }
                //!! no representation in CVariantProperties for eHasSTS, so use bitfield
                if(isBitfield && bf.IsTrue(CSnpBitfield::eHasSTS)) {
                    Texts.push_back("Entrez STS");
                    URLs.push_back("");
                }
                // scan and replace the URL subst fields
                // use all existing sResourceLink_ values to cover all cases
                // from those and texts, form full URLs
                list<string>::const_iterator iTexts(Texts.begin());
                ITERATE(list<string>, iURLs, URLs) {
                    string sFullURL(NStr::Replace(*iURLs, 
                                                  NSNPVariationHelper::sResourceLink_RsID, 
                                                  NStr::LongToString(GetIntValue(row, eRsId))));

                    if(sFullURL.empty()) {
                        // values.push_back(*iTexts);   -- values w/o an actual URL are disabled per SNP-4726
                    } else
                        values.push_back("<a href=\"" + sFullURL + "\">" + *iTexts + "</a>");
                    ++iTexts;
                }
                break;
            }
/*            case ePopFreq: -- dropped as asked in SNP-4726
                if(isProp && 
                    pVariation->GetVariant_prop().CanGetGenotype() &&
                    pVariation->GetVariant_prop().GetGenotype() & CVariantProperties::eGenotype_has_genotypes) {
                    values.push_back("<a href=\""
                                     "http://www.ncbi.nlm.nih.gov/sites/entrez?cmd=Retrieve&db=Snp&dopt=GEN&list_uids=" + 
                                     NStr::LongToString(GetIntValue(row, eRsId)) +
                                     "\">GTY</a>");
                }
                break;
*/
            case eFXNClass:
                if(isProp) {
                    list<string> GeneLocations;
                    NSNPVariationHelper::VariantPropAsStrings(GeneLocations,
                                                                pVariation->GetVariant_prop(),
                                                                NSNPVariationHelper::eSNPPropName_GeneLocation);
                    list<string> Effects;
                    NSNPVariationHelper::VariantPropAsStrings(Effects,
                                                                pVariation->GetVariant_prop(),
                                                                NSNPVariationHelper::eSNPPropName_Effect);
                    string sURL(!Effects.empty() 
                                ? 
                                    pVariation->GetId().GetUrl() + "#locus"
                                :
                                    "http://www.ncbi.nlm.nih.gov/sites/entrez?"
                                    "Db=gene&DbFrom=snp&Cmd=Link&LinkName=snp_gene&LinkReadableName=Gene&"
                                    "IdsFromResult=" +
                                    NStr::LongToString(GetIntValue(row, eRsId)));
                    ITERATE(list<string>, iGeneLocations, GeneLocations) {
                        values.push_back("<a href=\"" + sURL + "\">" + *iGeneLocations + "</a>");
                    }
                    ITERATE(list<string>, iEffects, Effects) {
                        values.push_back("<a href=\"" + sURL + "\">" + *iEffects + "</a>");
                    }
                }
                break;

            case ePhenotype:
                //!! no representation in CVariantProperties for eHasMeSH, so use bitfield
                if(isBitfield && bf.IsTrue(CSnpBitfield::eHasMeSH)) {
                    values.push_back("MeSH");
                }
                //!! no representation in CVariantProperties for eHasClinicalAssay, so use bitfield
                if(isBitfield && bf.IsTrue(CSnpBitfield::eHasClinicalAssay)) {
                    values.push_back("Clinical diagnostic assay");
                }
                //!! no representation in CVariantProperties for eHasTranscriptionFactor, so use bitfield
                if(isBitfield && bf.IsTrue(CSnpBitfield::eHasTranscriptionFactor)) {
                    values.push_back("Transcription factor");
                }
                //!! no representation in CVariantProperties for eHasDiseaseInfo, so use bitfield
                if(isBitfield && bf.IsTrue(CSnpBitfield::eHasDiseaseInfo)) {
                    bool isLinkFound(false);
                    if(pVariation->CanGetOther_ids()) {
                        const CVariation::TOther_ids& OtherIDs(pVariation->GetOther_ids());
                        ITERATE(CVariation::TOther_ids, iOtherIDs, OtherIDs) {
                            if((*iOtherIDs)->CanGetDb() &&
                                ((*iOtherIDs)->GetDb() == "GeneID" ||
                                (*iOtherIDs)->GetDb() == "LocusLink") &&
                                (*iOtherIDs)->CanGetTag()) {
                                    values.push_back("<a href=\""
                                                     "http://www.ncbi.nlm.nih.gov/sites/varvu?gene=" + 
                                                     NStr::NumericToString((*iOtherIDs)->GetTag().GetId()) +
                                                     "\">From LSDB</a>");
                                    isLinkFound = true;
                            }
                        }
                    }
                    // even if no URL links were found, we need to signify LSDB prsence at least once
                    if(!isLinkFound)
                        values.push_back("From LSDB");
                }
                //!! no representation in CVariantProperties for eHasLowdbGaPPValue, so use bitfield
                if(isBitfield && bf.IsTrue(CSnpBitfield::eHasLowdbGaPPValue)) {
                    values.push_back("p-Value <= 0.001 in dbGaP");
                }
                //!! no representation in CVariantProperties for eHasLodScore, so use bitfield
                if(isBitfield && bf.IsTrue(CSnpBitfield::eHasLodScore)) {
                    values.push_back("LOD score >= 2.0 in dbGaP");
                }
                //!! no representation in CVariantProperties for eHasMicroattribution, so use bitfield
                if(isBitfield && bf.IsTrue(CSnpBitfield::eHasMicroattribution)) {
                    values.push_back("Microattribution");
                }
                //!! no representation in CVariantProperties for eHasOMIM_OMIA, so use bitfield
                if(isBitfield && bf.IsTrue(CSnpBitfield::eHasOMIM_OMIA)) {
                    values.push_back("<a href=\""
                                     "http://www.ncbi.nlm.nih.gov/SNP/snp_redirect.cgi?snp_id=" + 
                                     NStr::LongToString(GetIntValue(row, eRsId)) +
                                     "\">OMIM/OMIA</a>");
                }
                break;
            
            default:
                break;
        }
    }
    return NStr::Join(values, sDelimiter);
}




long CTableDataSNP::GetIntValue(size_t row, size_t col) const
{
    long value(0);
    if (row < GetRowsCount()) {
        CConstRef<CVariation> pVariation(x_GetRowData(row));

        CSnpBitfield bf;
        if(pVariation->CanGetExt()) {
            const CVariation::TExt& Ext(pVariation->GetExt());
            ITERATE(CVariation::TExt, iExt, Ext) {
                if((*iExt)->CanGetClass() &&
                    (*iExt)->GetClass() == SNP_VAR_EXT_CLASS &&
                    (*iExt)->GetFieldRef(SNP_VAR_EXT_BITFIELD).NotEmpty() &&
                    (*iExt)->GetFieldRef(SNP_VAR_EXT_BITFIELD)->CanGetData() &&
                    (*iExt)->GetFieldRef(SNP_VAR_EXT_BITFIELD)->GetData().IsOs()) {
                        bf = (*iExt)->GetFieldRef(SNP_VAR_EXT_BITFIELD)->GetData().GetOs();
                }
            }
        }
        bool isBitfield(bf.GetVersion() > 0);
        switch(col){
            case eRsId:
            {
                string sValue;
                if(pVariation->GetId().CanGetTag()) {
                    CNcbiOstrstream ostr;
                    pVariation->GetId().GetTag().AsString(ostr);
                    sValue = CNcbiOstrstreamToString(ostr);
                } else
		            pVariation->GetId().GetLabel(&sValue);

                value = NStr::StringToLong(sValue, NStr::fConvErr_NoThrow);
                break;
            }
            case ePosition:
                if(pVariation->CanGetPlacements()) {
                    const CVariation::TPlacements& VarPlacements(pVariation->GetPlacements());
                    if(!VarPlacements.empty()) {
                        CRef<CVariantPlacement> pPlacement(VarPlacements.front());
                        if(pPlacement->CanGetLoc())
                            value = pPlacement->GetLoc().GetTotalRange().GetFrom() + 1;
                    }
                }
                break;
            case eWeight:
                //!! weight conversion from bitfield to CVariantProperties has been disabled (SNP-5729),
				//!! so use bitfield
                if(isBitfield) {
					value = bf.GetWeight();
                }
/*                if(pVariation->CanGetVariant_prop() &&
                    pVariation->GetVariant_prop().CanGetMap_weight()) {
                    value = pVariation->GetVariant_prop().GetMap_weight();
				}
*/              break;
        }
    }
    return value;
}


double CTableDataSNP::GetRealValue(size_t row, size_t col) const
{
    double value = 0.0;
    return value;
}


SConstScopedObject CTableDataSNP::GetObjectValue(size_t row, size_t col) const
{
    return GetRowObject(row);
}


SConstScopedObject CTableDataSNP::GetRowObject(size_t row) const
{
    SConstScopedObject value;
    if(m_DataLoadStatus == eReady && row < GetRowsCount()) {
        value.scope = m_Scope;
        value.object = x_GetRowData(row);
    }
    return value;
}


CConstRef<CVariation> CTableDataSNP::x_GetRowData(size_t row) const
{
    return row < m_listSnps->GetData().size() ? m_listSnps->GetData()[row] : CConstRef<CVariation>();
}


END_NCBI_SCOPE

