/*  $Id: blast_search_options_panel.cpp 27233 2013-01-15 20:49:59Z kuznets $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors: Andrey Yazhuk
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>

#include <gui/packages/pkg_alignment/blast_search_options_panel.hpp>

#include <gui/packages/pkg_alignment/blast_search_params.hpp>
#include <gui/packages/pkg_alignment/blast_db_dialog.hpp>
#include <gui/packages/pkg_alignment/net_blast_ui_data_source.hpp>

#include <gui/widgets/object_list/object_list_widget.hpp>

#include <gui/widgets/wx/wx_utils.hpp>
#include <gui/widgets/wx/table_utils.hpp>

#include <wx/sizer.h>
#include <wx/image.h>
#include <wx/stattext.h>
#include <wx/button.h>
#include <wx/combobox.h>
#include <wx/radiobut.h>
#include <wx/statbox.h>

////@begin XPM images
////@end XPM images


BEGIN_NCBI_SCOPE


IMPLEMENT_DYNAMIC_CLASS( CBLASTSearchOptionsPanel, wxPanel )

BEGIN_EVENT_TABLE( CBLASTSearchOptionsPanel, wxPanel )

////@begin CBLASTSearchOptionsPanel event table entries
    EVT_RADIOBUTTON( ID_NUC_RADIO, CBLASTSearchOptionsPanel::OnNucRadioSelected )

    EVT_RADIOBUTTON( ID_PROT_RADIO, CBLASTSearchOptionsPanel::OnProtRadioSelected )

    EVT_COMBOBOX( ID_PROGRAM_COMBO, CBLASTSearchOptionsPanel::OnProgramComboSelected )

    EVT_BUTTON( ID_DB_BTN, CBLASTSearchOptionsPanel::OnBrowseDatabases )

////@end CBLASTSearchOptionsPanel event table entries

END_EVENT_TABLE()


CBLASTSearchOptionsPanel::CBLASTSearchOptionsPanel()
    : m_DataSource(NULL)
    , m_Remote(false)
    , m_NucObjects(NULL)
    , m_ProtObjects(NULL)
    , m_Params(NULL)
    , m_QueryListDirty(false)
    , m_SubjectListDirty(false)
{
    Init();
}


CBLASTSearchOptionsPanel::CBLASTSearchOptionsPanel( wxWindow* parent, wxWindowID id, const wxPoint& pos, const wxSize& size, long style )
{
    Init();
    Create(parent, id, pos, size, style);
}


bool CBLASTSearchOptionsPanel::Create( wxWindow* parent, wxWindowID id, const wxPoint& pos, const wxSize& size, long style )
{
    wxPanel::Create( parent, id, pos, size, style );

    Hide();

    CreateControls();
    if (GetSizer())
    {
        GetSizer()->SetSizeHints(this);
    }

    return true;
}


CBLASTSearchOptionsPanel::~CBLASTSearchOptionsPanel()
{
////@begin CBLASTSearchOptionsPanel destruction
////@end CBLASTSearchOptionsPanel destruction
}


void CBLASTSearchOptionsPanel::Setup(CNetBLASTUIDataSource& ds,
                                     const string& tool_name,
                                     bool remote)
{
    m_DataSource = &ds;
    m_ToolName = tool_name;
    m_Remote = remote;

    m_DataSource->WaitReady();
}


void CBLASTSearchOptionsPanel::SetParams(CBLASTParams* params,
                                         TConstScopedObjects* nuc_objects,
                                         TConstScopedObjects* prot_objects)
{
    m_Params = params;

    _ASSERT(nuc_objects  &&  prot_objects);

    m_NucObjects = nuc_objects;
    m_ProtObjects = prot_objects;

}


void CBLASTSearchOptionsPanel::Init()
{
////@begin CBLASTSearchOptionsPanel member initialisation
    m_NucRadio = NULL;
    m_ProtRadio = NULL;
    m_QueryListCtrl = NULL;
    m_ProgramCombo = NULL;
    m_DatabaseStatic = NULL;
    m_DatabaseCombo = NULL;
    m_BrowseDBBtn = NULL;
    m_EntrezStatic = NULL;
    m_QueryCombo = NULL;
    m_SubjPanel = NULL;
    m_SubjSizer = NULL;
    m_SubjectListCtrl = NULL;
////@end CBLASTSearchOptionsPanel member initialisation

    m_DataSource = NULL;
    m_ToolName.clear();

    m_NucObjects = NULL;
    m_ProtObjects = NULL;

    m_Params = NULL;

    m_QueryListDirty = true;
    m_SubjectListDirty = true;
}


void CBLASTSearchOptionsPanel::CreateControls()
{
////@begin CBLASTSearchOptionsPanel content construction
    CBLASTSearchOptionsPanel* itemPanel1 = this;

    wxBoxSizer* itemBoxSizer2 = new wxBoxSizer(wxVERTICAL);
    itemPanel1->SetSizer(itemBoxSizer2);

    wxStaticBox* itemStaticBoxSizer3Static = new wxStaticBox(itemPanel1, wxID_ANY, _("Query Sequences"));
    wxStaticBoxSizer* itemStaticBoxSizer3 = new wxStaticBoxSizer(itemStaticBoxSizer3Static, wxVERTICAL);
    itemBoxSizer2->Add(itemStaticBoxSizer3, 1, wxGROW|wxALL, 5);

    wxBoxSizer* itemBoxSizer4 = new wxBoxSizer(wxHORIZONTAL);
    itemStaticBoxSizer3->Add(itemBoxSizer4, 0, wxGROW, 5);

    m_NucRadio = new wxRadioButton( itemPanel1, ID_NUC_RADIO, _("Nucleotides"), wxDefaultPosition, wxDefaultSize, wxRB_GROUP );
    m_NucRadio->SetValue(true);
    itemBoxSizer4->Add(m_NucRadio, 0, wxALIGN_CENTER_VERTICAL|wxALL, 5);

    m_ProtRadio = new wxRadioButton( itemPanel1, ID_PROT_RADIO, _("Proteins"), wxDefaultPosition, wxDefaultSize, 0 );
    m_ProtRadio->SetValue(false);
    itemBoxSizer4->Add(m_ProtRadio, 0, wxALIGN_CENTER_VERTICAL|wxALL, 5);

    itemBoxSizer4->Add(7, 8, 1, wxALIGN_CENTER_VERTICAL, 5);

    wxButton* itemButton8 = new wxButton( itemPanel1, ID_BUTTON2, _("Add"), wxDefaultPosition, wxDefaultSize, 0 );
    itemButton8->Enable(false);
    itemBoxSizer4->Add(itemButton8, 0, wxALIGN_CENTER_VERTICAL|wxALL, 5);

    m_QueryListCtrl = new CObjectListWidget( itemPanel1, ID_LIST, wxDefaultPosition, itemPanel1->ConvertDialogToPixels(wxSize(350, 75)), wxLC_REPORT );
    itemStaticBoxSizer3->Add(m_QueryListCtrl, 1, wxGROW|wxALL, 5);

    wxFlexGridSizer* itemFlexGridSizer10 = new wxFlexGridSizer(3, 2, 0, 0);
    itemFlexGridSizer10->AddGrowableCol(1);
    itemBoxSizer2->Add(itemFlexGridSizer10, 0, wxGROW|wxALL, 5);

    wxStaticText* itemStaticText11 = new wxStaticText( itemPanel1, wxID_STATIC, _("Program:"), wxDefaultPosition, wxDefaultSize, 0 );
    itemFlexGridSizer10->Add(itemStaticText11, 0, wxALIGN_RIGHT|wxALIGN_CENTER_VERTICAL|wxALL, 5);

    wxArrayString m_ProgramComboStrings;
    m_ProgramCombo = new wxComboBox( itemPanel1, ID_PROGRAM_COMBO, wxEmptyString, wxDefaultPosition, wxDefaultSize, m_ProgramComboStrings, wxCB_READONLY );
    itemFlexGridSizer10->Add(m_ProgramCombo, 0, wxGROW|wxALIGN_CENTER_VERTICAL|wxALL, 5);

    m_DatabaseStatic = new wxStaticText( itemPanel1, wxID_STATIC, _("Database:"), wxDefaultPosition, wxDefaultSize, 0 );
    itemFlexGridSizer10->Add(m_DatabaseStatic, 0, wxALIGN_RIGHT|wxALIGN_CENTER_VERTICAL|wxALL, 5);

    wxBoxSizer* itemBoxSizer14 = new wxBoxSizer(wxHORIZONTAL);
    itemFlexGridSizer10->Add(itemBoxSizer14, 0, wxGROW|wxALIGN_CENTER_VERTICAL, 5);

    wxArrayString m_DatabaseComboStrings;
    m_DatabaseCombo = new wxComboBox( itemPanel1, ID_DB_COMBO, wxEmptyString, wxDefaultPosition, wxDefaultSize, m_DatabaseComboStrings, wxCB_READONLY );
    itemBoxSizer14->Add(m_DatabaseCombo, 1, wxALIGN_CENTER_VERTICAL|wxALL, 5);

    m_BrowseDBBtn = new wxButton( itemPanel1, ID_DB_BTN, _("..."), wxDefaultPosition, wxSize(itemPanel1->ConvertDialogToPixels(wxSize(30, -1)).x, -1), 0 );
    if (CBLASTSearchOptionsPanel::ShowToolTips())
        m_BrowseDBBtn->SetToolTip(_("More Databases"));
    itemBoxSizer14->Add(m_BrowseDBBtn, 0, wxALIGN_CENTER_VERTICAL|wxALL|wxFIXED_MINSIZE, 5);

    m_EntrezStatic = new wxStaticText( itemPanel1, wxID_STATIC, _("Entrez Query:"), wxDefaultPosition, wxDefaultSize, 0 );
    itemFlexGridSizer10->Add(m_EntrezStatic, 0, wxALIGN_RIGHT|wxALIGN_CENTER_VERTICAL|wxALL, 5);

    wxArrayString m_QueryComboStrings;
    m_QueryCombo = new wxComboBox( itemPanel1, ID_ENTREZ_COMBO, wxEmptyString, wxDefaultPosition, wxDefaultSize, m_QueryComboStrings, wxCB_DROPDOWN );
    itemFlexGridSizer10->Add(m_QueryCombo, 0, wxGROW|wxALIGN_CENTER_VERTICAL|wxALL, 5);

    m_SubjPanel = new wxPanel( itemPanel1, ID_SUBJ_PANEL, wxDefaultPosition, wxDefaultSize, wxTAB_TRAVERSAL );
    itemBoxSizer2->Add(m_SubjPanel, 1, wxGROW, 5);

    m_SubjSizer = new wxStaticBox(m_SubjPanel, wxID_ANY, _("Subject Sequences"));
    wxStaticBoxSizer* itemStaticBoxSizer20 = new wxStaticBoxSizer(m_SubjSizer, wxVERTICAL);
    m_SubjPanel->SetSizer(itemStaticBoxSizer20);

    wxBoxSizer* itemBoxSizer21 = new wxBoxSizer(wxHORIZONTAL);
    itemStaticBoxSizer20->Add(itemBoxSizer21, 0, wxGROW, 5);

    itemBoxSizer21->Add(7, 8, 1, wxALIGN_CENTER_VERTICAL, 5);

    wxButton* itemButton23 = new wxButton( m_SubjPanel, ID_SUBJ_ADD, _("Add"), wxDefaultPosition, wxDefaultSize, 0 );
    itemButton23->Enable(false);
    itemBoxSizer21->Add(itemButton23, 0, wxALIGN_CENTER_VERTICAL|wxALL, 5);

    m_SubjectListCtrl = new CObjectListWidget( m_SubjPanel, ID_SUBJ_LIST, wxDefaultPosition, m_SubjPanel->ConvertDialogToPixels(wxSize(350, 75)), wxLC_REPORT );
    itemStaticBoxSizer20->Add(m_SubjectListCtrl, 1, wxGROW|wxALL, 5);

    // Set validators
    m_ProgramCombo->SetValidator( wxGenericValidator(& m_Program) );
    m_DatabaseCombo->SetValidator( wxGenericValidator(& m_Database) );
    m_QueryCombo->SetValidator( wxGenericValidator(& m_Query) );
////@end CBLASTSearchOptionsPanel content construction

    if(m_Remote)    {
        m_SubjPanel->Hide();
    } else {
        m_DatabaseStatic->Hide();
        m_DatabaseCombo->Hide();

        m_EntrezStatic->Hide();
        m_BrowseDBBtn->Hide();
        m_QueryCombo->Hide();
    }
}


bool CBLASTSearchOptionsPanel::ShowToolTips()
{
    return true;
}


wxBitmap CBLASTSearchOptionsPanel::GetBitmapResource( const wxString& name )
{
    // Bitmap retrieval
////@begin CBLASTSearchOptionsPanel bitmap retrieval
    wxUnusedVar(name);
    return wxNullBitmap;
////@end CBLASTSearchOptionsPanel bitmap retrieval
}


wxIcon CBLASTSearchOptionsPanel::GetIconResource( const wxString& name )
{
    // Icon retrieval
////@begin CBLASTSearchOptionsPanel icon retrieval
    wxUnusedVar(name);
    return wxNullIcon;
////@end CBLASTSearchOptionsPanel icon retrieval
}

void CBLASTSearchOptionsPanel::SetRegistryPath(const string& path)
{
    m_RegPath = path; // store for later use
}


static const char* kTableTag = "Table";

void CBLASTSearchOptionsPanel::SaveSettings() const
{
    if( ! m_RegPath.empty())   {
        CGuiRegistry& gui_reg = CGuiRegistry::GetInstance();
        string reg_path = CGuiRegistryUtil::MakeKey(m_RegPath, kTableTag);

        CGuiRegistry::TReadWriteView query_table_view = gui_reg.GetReadWriteView( 
            reg_path + ".QueryTable" 
        );
        ncbi::SaveTableSettings( *m_QueryListCtrl, query_table_view );

        CGuiRegistry::TReadWriteView subject_table_view = gui_reg.GetReadWriteView( 
            reg_path + ".SubjectTable" 
        );
        ncbi::SaveTableSettings( *m_SubjectListCtrl, subject_table_view );
    }
}


void CBLASTSearchOptionsPanel::LoadSettings()
{
    if( ! m_RegPath.empty())   {
        CGuiRegistry& gui_reg = CGuiRegistry::GetInstance();
        string reg_path = CGuiRegistryUtil::MakeKey(m_RegPath, kTableTag);

        CGuiRegistry::TReadView query_table_view = gui_reg.GetReadView(
            reg_path + ".QueryTable" 
        );
        ncbi::LoadTableSettings( *m_QueryListCtrl, query_table_view );

        CGuiRegistry::TReadView subject_table_view = gui_reg.GetReadView(
            reg_path + ".SubjectTable" 
        );
        ncbi::LoadTableSettings( *m_SubjectListCtrl, subject_table_view );
    }
}


void sMRUListToCombo(CMRUList<string>&                    list, 
                     wxComboBox&                          combo, 
                     const CNetBLASTUIDataSource::TDbMap& map)
{

    wxArrayString arr;
    ITERATE(CMRUList<string>::TItems, it, list.GetItems())    {
        const string& db_name = *it;
        if (map.find(db_name) != map.end()) {
            arr.Add(ToWxString(db_name));
            LOG_POST(Info << "\t" << db_name);
        }
    }
    combo.Clear();
    combo.Append(arr);
}


bool CBLASTSearchOptionsPanel::TransferDataToWindow()
{
    if(m_QueryListDirty)    {
        // update program list
        m_ProgramLabels.Clear();

        vector<string> labels;
        CBLASTParams::GetProgramLabels(m_Params->IsNucInput(), labels);
        ToArrayString(labels, m_ProgramLabels);

        m_ProgramCombo->Clear();
        m_ProgramCombo->Append(m_ProgramLabels);

        TConstScopedObjects& objects = m_Params->IsNucInput() ? *m_NucObjects : *m_ProtObjects;
        m_QueryListCtrl->SetObjects(objects);
        if (m_QueryListCtrl->GetItemCount() <= 100) {
            m_QueryListCtrl->SelectAll();
        }
        else if (m_QueryListCtrl->GetItemCount() > 0) {
            m_QueryListCtrl->Select(0);
        }
        m_QueryListDirty = false;
    }

    if(! m_Remote  &&  m_SubjectListDirty)    {
        TConstScopedObjects& objects = m_Params->IsSubjNucInput() ? *m_NucObjects : *m_ProtObjects;
        m_SubjectListCtrl->SetObjects(objects);
        if (m_SubjectListCtrl->GetItemCount() <= 100) {
            m_SubjectListCtrl->SelectAll();
        }
        else if (m_SubjectListCtrl->GetItemCount() > 0) {
            m_SubjectListCtrl->Select(0);
        }
        m_SubjectListDirty = false;
    }

    m_NucRadio->SetValue(m_Params->IsNucInput());
    m_ProtRadio->SetValue( ! m_Params->IsNucInput());

    m_Program = ToWxString(CBLASTParams::GetProgramLabel
                           (m_Params->GetCurrProgram()));

    if(m_Remote)    {
        m_Database = ToWxString(m_Params->GetCurrDatabase());

        bool nuc_db = m_Params->IsNucInput();
        const CNetBLASTUIDataSource::TDbMap& map = m_DataSource->GetDbMap(nuc_db);

        sMRUListToCombo(m_Params->GetCurrMRUDatabases(), *m_DatabaseCombo, map);

        m_Query = ToWxString(m_Params->GetEntrezQuery());
        //TODO MRU queries from extension point
        m_QueryCombo->Clear();
        m_QueryCombo->Append(m_Query.c_str());
    }

    bool res = wxPanel::TransferDataToWindow();
    return res;
}


bool CBLASTSearchOptionsPanel::TransferDataFromWindow()
{
    if(wxPanel::TransferDataFromWindow())   {
        if(m_Remote)    {
            m_Params->GetEntrezQuery() = ToStdString(m_Query);

            m_Params->GetCurrDatabase() = ToStdString(m_Database);
        } else {
            m_Params->GetSubjSeqLocs().clear();
            m_SubjectListCtrl->GetSelection(m_Params->GetSubjSeqLocs());
        }

        m_Params->GetCurrProgram() =
            CBLASTParams::GetProgramByLabel(ToStdString(m_Program));

        m_Params->GetSeqLocs().clear();
        m_QueryListCtrl->GetSelection(m_Params->GetSeqLocs());

        m_Params->IsNucInput() = m_NucRadio->GetValue();

        return true;
    }
    return false;
}


void CBLASTSearchOptionsPanel::OnProtRadioSelected(wxCommandEvent& event)
{
    x_OnInputTypeChanged();
}


void CBLASTSearchOptionsPanel::OnNucRadioSelected(wxCommandEvent& event)
{
    x_OnInputTypeChanged();
}


void CBLASTSearchOptionsPanel::x_OnInputTypeChanged()
{
    m_Params->IsNucInput() = m_NucRadio->GetValue();
    m_QueryListDirty = true;
    m_SubjectListDirty = true;

    TransferDataToWindow();
}


void CBLASTSearchOptionsPanel::OnProgramComboSelected( wxCommandEvent& event )
{
    bool nuc_subj = m_Params->IsSubjNucInput();

    TransferDataFromWindow();

    m_SubjectListDirty = (m_Params->IsSubjNucInput() != nuc_subj);

    TransferDataToWindow();
}


static const char* kDBBrowserSection = ".DB Browser";

void CBLASTSearchOptionsPanel::OnBrowseDatabases( wxCommandEvent& event )
{
    vector<string> databases;
    databases.push_back(ToStdString(m_Database));

    CBLAST_DB_Dialog dlg(this);
    dlg.SetRegistryPath(m_RegPath + kDBBrowserSection);

    dlg.SetToolName(m_ToolName);
    dlg.SelectDatabases(databases);

    blast::EProgram program = m_Params->GetCurrProgram();
    bool nuc_db = CBLASTParams::IsNucDatabase(program);
    const CNetBLASTUIDataSource::TDbMap& map = m_DataSource->GetDbMap(nuc_db);
    dlg.SetDBMap(map);

    if(dlg.ShowModal() == wxID_OK)   {
        // User selected a database
        databases.clear();
        dlg.GetSelectedDatabases(databases);

        _ASSERT(databases.size() == 1);

        string db = databases[0];
        m_Params->GetCurrDatabase() = db;
        m_Params->GetCurrMRUDatabases().AddItem(db);

        TransferDataToWindow();
    }
}


END_NCBI_SCOPE
