/*  $Id: blast_search_tool_manager.cpp 26399 2012-09-11 22:01:16Z voronov $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Andrey Yazhuk
 *
 */

#include <ncbi_pch.hpp>

#include <gui/packages/pkg_alignment/blast_search_tool_manager.hpp>

#include <gui/packages/pkg_alignment/blast_search_options_panel.hpp>
#include <gui/packages/pkg_alignment/blast_search_params_panel.hpp>
#include <gui/packages/pkg_alignment/blast_search_task.hpp>

#include <gui/framework/app_job_task.hpp>

#include <gui/widgets/wx/message_box.hpp>

#include <gui/objutils/registry.hpp>

#include <gui/utils/extension_impl.hpp>

#include <objects/seqloc/Seq_loc.hpp>


#include <wx/panel.h>
#include <wx/arrstr.h>


BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

///////////////////////////////////////////////////////////////////////////////
/// CBLASTSearchToolManager
CBLASTSearchToolManager::CBLASTSearchToolManager()
{
    m_Descriptor.Init("BLAST Search", "",
                      "Execute BLAST searches using NCBI's BLAST service",
                      "Submit a sequence for sequence similarity comparison "
                      "using NCBI's BLAST service.");
}


string CBLASTSearchToolManager::GetExtensionIdentifier() const
{
    return "blast_search_tool_manager";
}


string CBLASTSearchToolManager::GetExtensionLabel() const
{
    return "BLAST Search Tool";
}


// validate parameters enetred on the Options page
bool CBLASTSearchToolManager::x_ValidateOptions()
{
    string err;
    if(m_Params.GetSeqLocs().empty())  {
        err = "Please select at least one sequence!";
    } else if(m_Params.GetCurrDatabase().empty())  {
        err = "Please select a database!";
    }
    if( ! err.empty()) {
        NcbiErrorBox(err, m_Descriptor.GetLabel());
        return false;
    }
    return true;
}


IAppTask* CBLASTSearchToolManager::GetTask()
{
    CRef<CNetBLASTUIDataSource> blast_ds = GetBLASTDataSource();

    CRef<CBlastSearchTask> task(
        new CBlastSearchTask(m_SrvLocator, *blast_ds, m_Descriptor.GetLabel())
    );
    task->Init_SubmitSearch(m_Params);

    // Log the parameters
    m_Params.PrintParams("CBLASTSearchToolManager::");

    if( m_ProjectSelPanel ){
        m_ProjectSelPanel->GetParams( m_ProjectParams );
    } else {
        m_ProjectParams.m_CreateFolder = false;
        m_ProjectParams.m_FolderName = "";

        CProjectService* prj_srv = m_SrvLocator->GetServiceByType<CProjectService>();

        vector<objects::CGBProjectHandle::TId> projects;
        prj_srv->GetObjProjects( m_Params.GetSeqLocs(), projects );
        if( projects.empty() ){
            m_ProjectParams.m_ProjectMode = SProjectSelectorParams::eCreateOneProject;
        } else {
            m_ProjectParams.m_ProjectMode = SProjectSelectorParams::eAddToExistingProject;
            m_ProjectParams.m_SelectedProjectId = projects.front();
        }
    }

    CDataLoadingOptions options;
    m_ProjectParams.ToLoadingOptions(options);
    task->SetLoadingOptions(options);

    return task.Release();
}

IAppTask* CBLASTSearchToolManager::QuickLaunch()
{
    x_SelectCompatibleInputObjects();

    m_Params.IsNucInput() = x_IsNucInput();

    CRef<CNetBLASTUIDataSource> blast_ds = GetBLASTDataSource();

    if( m_Remote ){
        /// make sure the default Database is not empty
        x_InitDBsInParams( *blast_ds, true );
        x_InitDBsInParams( *blast_ds, false );
    }
    // Create Options panel and initialize it

    if( m_Params.IsNucInput() ){
        m_Params.GetSeqLocs() = m_NucObjects;
    
    } else {
        m_Params.GetSeqLocs() = m_ProtObjects;
        if( m_UseNucForMixInput ){
            copy(m_NucObjects.begin(), m_NucObjects.end(), back_inserter(m_Params.GetSeqLocs()));
        }
    }

    SProjectSelectorParams project_params;
    project_params.m_CreateFolder = false;
    project_params.m_FolderName = "";

    CProjectService* prj_srv = m_SrvLocator->GetServiceByType<CProjectService>();

    vector<objects::CGBProjectHandle::TId> projects;
    prj_srv->GetObjProjects( m_Params.GetSeqLocs(), projects );
    if( projects.empty() ){
        project_params.m_ProjectMode = SProjectSelectorParams::eCreateOneProject;
    } else {
        project_params.m_ProjectMode = SProjectSelectorParams::eAddToExistingProject;
        project_params.m_SelectedProjectId = projects.front();
    }

    //m_ProjectSelPanel->SetParams( project_params );

    return GetTask();
}


END_NCBI_SCOPE
