/*  $Id: blast_tool_manager_base.cpp 26938 2012-11-28 15:46:40Z katargir $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Andrey Yazhuk
 *
 */

#include <ncbi_pch.hpp>

#include <gui/packages/pkg_alignment/blast_tool_manager_base.hpp>

#include <gui/packages/pkg_alignment/blast_search_options_panel.hpp>
#include <gui/packages/pkg_alignment/blast_search_params_panel.hpp>
#include <gui/packages/pkg_alignment/blast_search_task.hpp>

#include <gui/core/ui_data_source_service.hpp>

#include <gui/widgets/wx/message_box.hpp>

#include <gui/objutils/registry.hpp>

#include <objects/seqloc/Seq_loc.hpp>

#include <gui/core/async_obj_convert.hpp>

#include <wx/panel.h>
#include <wx/arrstr.h>


BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

///////////////////////////////////////////////////////////////////////////////
/// CBLASTToolManagerBase
CBLASTToolManagerBase::CBLASTToolManagerBase()
:   m_Descriptor("", ""),
    m_Remote(true),
    m_SrvLocator(NULL),
    m_ParentWindow(NULL),
    m_UseNucForMixInput(true),
    m_State(eInvalid),
    m_OptionsPanel(NULL),
    m_ParamsPanel(NULL),
    m_ProjectSelPanel(NULL)
{
}


void CBLASTToolManagerBase::SetServiceLocator(IServiceLocator* srv_locator)
{
    m_SrvLocator = srv_locator;
}


void CBLASTToolManagerBase::SetParentWindow(wxWindow* parent)
{
    m_ParentWindow = parent;
}


const IUIObject& CBLASTToolManagerBase::GetDescriptor() const
{
    return m_Descriptor;
}


void CBLASTToolManagerBase::InitUI()
{
    // initially tool panels are not displayed and so the first
    // step of the Run Tool dialog corresponds to invalid state
    m_State = eInvalid;
    m_InputObjects.clear();

    m_OptionsPanel = NULL;
    m_ParamsPanel = NULL;
    m_ProjectSelPanel = NULL;
}


void CBLASTToolManagerBase::CleanUI()
{
    m_State = eInvalid;
    m_InputObjects.clear();

    //TODO Current panel lifetime management
    m_OptionsPanel = NULL;
    m_ParamsPanel = NULL;
    m_ProjectSelPanel = NULL;
}


wxPanel* CBLASTToolManagerBase::GetMaxPanel()
{
    CBLASTSearchOptionsPanel* optionsPanel = new CBLASTSearchOptionsPanel();
    optionsPanel->SetRemote(m_Remote);
    optionsPanel->Create(m_ParentWindow);
    return optionsPanel;
}

wxPanel* CBLASTToolManagerBase::GetCurrentPanel()
{
    switch(m_State) {
    case eOptions:
        return m_OptionsPanel;
    case eParams:
        return m_ParamsPanel;
    case eSelectProject:
        return m_ProjectSelPanel;
    default:
        return NULL;
    }
}


void CBLASTToolManagerBase::x_CreateOptionsPanelIfNeeded()
{
    if(m_OptionsPanel == NULL)   {
        // preapare input data
        x_SelectCompatibleInputObjects();

        m_Params.IsNucInput() = x_IsNucInput();
        CRef<CNetBLASTUIDataSource> blast_ds = GetBLASTDataSource();

        if(m_Remote)    {
            /// make sure the default Database is not empty
            x_InitDBsInParams(*blast_ds, true);
            x_InitDBsInParams(*blast_ds, false);
        }
        // Create Options panel and initialize it
        m_OptionsPanel = new CBLASTSearchOptionsPanel();
        m_OptionsPanel->Setup(*blast_ds, m_Descriptor.GetLabel(), m_Remote);
        m_OptionsPanel->SetParams(&m_Params, &m_NucObjects, &m_ProtObjects);

        m_OptionsPanel->Create(m_ParentWindow);

        m_OptionsPanel->SetRegistryPath(m_RegPath + ".OptionsPanel");
        m_OptionsPanel->LoadSettings();
    }
}


/// validate DB and MRU DB fields in BLAST Params and initializes them
/// if needed
void CBLASTToolManagerBase::x_InitDBsInParams(CNetBLASTUIDataSource& ds, bool nuc_db)
{
    typedef CBLASTParams::TMRUDatabases::TItems TItems;

    string& db = m_Params.GetDatabase(nuc_db);
    CBLASTParams::TMRUDatabases& mru_dbs = m_Params.GetMRUDatabases(nuc_db);
    const TItems& items = mru_dbs.GetItems();

    if(mru_dbs.GetItems().empty())  {
        // invalid MRU list - initialize from data source
        TItems new_items;
        ds.GetDefaultMRU_DBs(nuc_db, new_items);
        mru_dbs.SetItems(new_items);
    }
    _ASSERT( ! mru_dbs.GetItems().empty());

    TItems::const_iterator it = std::find(items.begin(), items.end(), db);
    if(it == items.end())   {
        // DB is not in the list, we msut fix it
        const CNetBLASTUIDataSource::TDbMap& db_map = ds.GetDbMap(nuc_db);
        if(db_map.find(db) != db_map.end())    {
            mru_dbs.AddItem(db); // valid DB name - add to the MRU
        } else {
            m_Params.GetDatabase(nuc_db) = *mru_dbs.GetItems().begin();
        }
    }
}


CRef<CNetBLASTUIDataSource> CBLASTToolManagerBase::GetBLASTDataSource()
{
    CIRef<CUIDataSourceService> ds_srv =
            m_SrvLocator->GetServiceByType<CUIDataSourceService>();

    CIRef<IUIDataSource> ds = ds_srv->GetUIDataSourceByLabel("NCBI Net BLAST");
    CNetBLASTUIDataSource* blast_ds = dynamic_cast<CNetBLASTUIDataSource*>(ds.GetPointer());
    return CRef<CNetBLASTUIDataSource>(blast_ds);
}


void CBLASTToolManagerBase::x_CreateParamsPanelIfNeeded()
{
    if( m_ParamsPanel == NULL ){
        m_ParamsPanel = new CBLASTSearchParamsPanel( m_ParentWindow );
        m_ParamsPanel->SetParams( &m_Params, ! m_Remote );
    }
}


void CBLASTToolManagerBase::x_CreateProjectPanelIfNeeded()
{
    if(m_ProjectSelPanel == NULL)   {
        CIRef<CProjectService> srv = m_SrvLocator->GetServiceByType<CProjectService>();

        m_ProjectParams.SelectProjectByObjects(m_Params.GetSeqLocs(), &*srv);

        m_ProjectSelPanel = new CProjectSelectorPanel(m_ParentWindow);

        m_ProjectSelPanel->SetProjectService(srv);
        m_ProjectSelPanel->SetParams(m_ProjectParams);
    }
}


bool CBLASTToolManagerBase::CanDo(EAction action)
{
    switch(m_State) {
    case eInvalid:
        return action == eNext;
    case eOptions:
        return action == eNext;
    case eParams:
    case eSelectProject:
        return action == eNext ||  action == eBack;
    case eCompleted:
        return false; // nothing left to do
    default:
        _ASSERT(false);
        return false;
    }
}


bool CBLASTToolManagerBase::IsFinalState()
{
    return m_State == eSelectProject;
}


bool CBLASTToolManagerBase::IsCompletedState()
{
    return m_State == eCompleted;
}


bool CBLASTToolManagerBase::DoTransition(EAction action)
{
    bool next = action == eNext;

    if(m_State == eInvalid  &&  next)    {
        m_State = eOptions;
        x_CreateOptionsPanelIfNeeded();
        m_OptionsPanel->TransferDataToWindow();
        return true;
    } else if(m_State == eOptions) {
        if(next)    {
            if(m_OptionsPanel->TransferDataFromWindow()  &&  x_ValidateOptions())    {
                m_State = eParams;
                x_CreateParamsPanelIfNeeded();
                m_ParamsPanel->TransferDataToWindow();
                return true;
            } else return false;
        } else {
            m_State = eInvalid;
            return true;
        }
    } else if(m_State == eParams) {
        if(next)    {
            if(m_ParamsPanel->TransferDataFromWindow()) {
                try{
                    LOG_POST(Trace << "Validationg BLAST Search params...");
                    x_ValidateParams();

                    x_CreateProjectPanelIfNeeded();
                    m_ProjectSelPanel->TransferDataToWindow();
                    m_State = eSelectProject;
                    return true;
                } catch (CException& e) {
                    ERR_POST(e.ReportAll());
                    NcbiErrorBox(e.GetMsg());
                }
            }
        } else {
            if(m_ParamsPanel->TransferDataFromWindow()) {
                m_State = eOptions;
                m_OptionsPanel->TransferDataToWindow();
                return true;
            }
        }
        return false;
    } else if(m_State == eSelectProject) {
        if(next)    {
            if(m_ProjectSelPanel->TransferDataFromWindow()) {
                m_State = eCompleted;
                return true;
            }
        } else {
            m_State = eParams;
            m_ParamsPanel->TransferDataToWindow();
            return true;
        }
        return false;
    }
    _ASSERT(false);
    return false;
}


/// validate parameters after Params page
bool CBLASTToolManagerBase::x_ValidateParams()
{
    CRef<blast::CBlastOptionsHandle> handle = m_Params.ToBlastOptions();
    return handle->SetOptions().Validate();
}


string CBLASTToolManagerBase::GetCategory()
{
    return "Alignment Creation";
}


void CBLASTToolManagerBase::SetInputObjects(const TConstScopedObjects& objects)
{
    m_InputObjects = objects;
}


void CBLASTToolManagerBase::x_SelectCompatibleInputObjects()
{
    m_NucObjects.clear();
    m_ProtObjects.clear();

    TConstScopedObjects objects;
    x_ConvertInputObjects(CSeq_loc::GetTypeInfo(), objects);

    wxBusyCursor wait;

    NON_CONST_ITERATE(TConstScopedObjects, it, objects) {
        const CSeq_loc* seq_loc = dynamic_cast<const CSeq_loc*>(it->object.GetPointer());
        if (!seq_loc  ||  !seq_loc->GetId())
            continue;

        CBioseq_Handle handle = it->scope->GetBioseqHandle(*seq_loc->GetId());
        if (!handle)
            continue;

        TConstScopedObjects* target = handle.IsNa() ? &m_NucObjects : &m_ProtObjects;
        target->push_back(*it);
    }
}

void CBLASTToolManagerBase::x_ConvertInputObjects(
    const CTypeInfo* typeInfo, TConstScopedObjects& results)
{
    CAsyncObjConvert2 convertor(typeInfo, m_InputObjects, results);
    CAsyncCall::Execute(convertor);
}

/// defines whether Nucleotide or Protein options shall be selected by default
bool CBLASTToolManagerBase::x_IsNucInput() const
{
    if(m_ProtObjects.empty())   {
        return true; // no proteins - select Nucleotides
    } else {
        // have proteins - use them if not nucleotides, otherwise
        // use m_UseNucForMixInput setting
        return m_NucObjects.empty() ? false : m_UseNucForMixInput;
    }
}


void CBLASTToolManagerBase::SetRegistryPath(const string& path)
{
    m_RegPath = path; // store for later use
}

static const char* kParamsTag = "Params";
static const char* kUseNucForMixInput = "UseNucForMixInput";
static const char* kLocalWinMaskDirPath = "Workbench.Services.DefaultUIDS.LocalMaskPath";

void CBLASTToolManagerBase::SaveSettings() const
{
    if( ! m_RegPath.empty())   {
        CGuiRegistry& gui_reg = CGuiRegistry::GetInstance();
        CGuiRegistry::TReadWriteView view = gui_reg.GetReadWriteView(m_RegPath);

        if( ! m_ProtObjects.empty()  &&  ! m_NucObjects.empty())    {
            // update option
            m_UseNucForMixInput = m_Params.IsNucInput();
        }
        view.Set(kUseNucForMixInput, m_UseNucForMixInput);

        /// save BLAST parameters
        CRef<CUser_object> param_obj(m_Params.ToUserObject());
        CRef<CUser_field> field =
            view.SetField(CGuiRegistryUtil::MakeKey(kParamsTag, kParamsTag));
        field->SetData().SetObject(*param_obj);

        if(m_OptionsPanel)  {
            m_OptionsPanel->SaveSettings();
        }
    }
}


void CBLASTToolManagerBase::LoadSettings()
{
    if( ! m_RegPath.empty())   {
        CGuiRegistry& gui_reg = CGuiRegistry::GetInstance();
        CGuiRegistry::TReadView view = gui_reg.GetReadView(m_RegPath);

        m_UseNucForMixInput = view.GetBool(kUseNucForMixInput, m_UseNucForMixInput);

        /// load BLAST parameters
        string key = CGuiRegistryUtil::MakeKey(kParamsTag, kParamsTag);
        if(view.HasField(key))   {
            const CUser_object& param_obj = view.GetField(key)->GetData().GetObject();
            m_Params.FromUserObject(param_obj);
        }

        m_LocalWinMaskDirPath = gui_reg.GetString( kLocalWinMaskDirPath );
    }
}


END_NCBI_SCOPE
