/*  $Id: create_gene_model_tool_manager.cpp 28116 2013-05-10 19:35:00Z voronov $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Mike DiCuccio
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>
#include <gui/packages/pkg_alignment/create_gene_model_tool_manager.hpp>
#include <gui/widgets/wx/message_box.hpp>

#include <gui/objutils/label.hpp>
#include <gui/objects/ProjectItem.hpp>

#include <algo/sequence/align_group.hpp>
#include <objects/seq/Annot_descr.hpp>
#include <objects/seq/Annotdesc.hpp>
#include <objects/seqalign/Seq_align.hpp>
#include <algo/sequence/gene_model.hpp>
#include <gui/objutils/label.hpp>
#include <objmgr/util/sequence.hpp>
#include <objects/general/Dbtag.hpp>
#include <objects/general/Object_id.hpp>


BEGIN_NCBI_SCOPE
USING_SCOPE(objects);


CCreateGeneModelToolManager::CCreateGeneModelToolManager()
:   CAlgoToolManagerBase("Create Gene Model",
                         "",
                         "Create a gene model from an alignment",
                         "Create a new gene model based on one or more "
                         "alignments of related transcripts to a genome",
                         "CREATE_GENE_MODEL",
                         "Alignment Creation"),
    m_ParamsPanel(NULL)
{
}


string CCreateGeneModelToolManager::GetExtensionIdentifier() const
{
    return "create_gene_model_tool_manager";
}


string CCreateGeneModelToolManager::GetExtensionLabel() const
{
    return "Create Gene Model Tool";
}


void CCreateGeneModelToolManager::InitUI()
{
    CAlgoToolManagerBase::InitUI();

    m_ParamsPanel = NULL;
}


void CCreateGeneModelToolManager::CleanUI()
{
    m_ParamsPanel = NULL;

    CAlgoToolManagerBase::CleanUI();
}


void CCreateGeneModelToolManager::x_CreateParamsPanelIfNeeded()
{
    if(m_ParamsPanel == NULL)   {
        x_SelectCompatibleInputObjects();

        m_ParamsPanel = new CCreateGeneModelParamsPanel();
        m_ParamsPanel->Hide(); // to reduce flicker
        m_ParamsPanel->Create(m_ParentWindow);
        m_ParamsPanel->SetParams(&m_Params, &m_Alignments);

        m_ParamsPanel->SetRegistryPath(m_RegPath + ".ParamsPanel");
        m_ParamsPanel->LoadSettings();
    }
}


bool CCreateGeneModelToolManager::x_ValidateParams()
{
    string err;
    if(m_Params.m_Alignments.size() == 0)  {
        err = "Please select at least one set of alignments!";
    }
    if( ! err.empty()) {
        NcbiErrorBox(err, m_Descriptor.GetLabel());
        return false;
    }
    return true;
}


/// select only Seq-aligns
void CCreateGeneModelToolManager::x_SelectCompatibleInputObjects()
{
    m_Params.m_Alignments.clear();
    x_ConvertInputObjects(CSeq_align::GetTypeInfo(), m_Alignments);
}


CAlgoToolManagerParamsPanel* CCreateGeneModelToolManager::x_GetParamsPanel()
{
    return m_ParamsPanel;
}


IRegSettings* CCreateGeneModelToolManager::x_GetParamsAsRegSetting()
{
    return &m_Params;
}


CDataLoadingAppJob* CCreateGeneModelToolManager::x_CreateLoadingJob()
{
    CIRef<CProjectService> srv = m_SrvLocator->GetServiceByType<CProjectService>();
    CCreateGeneModelJob* job = new CCreateGeneModelJob(srv.GetPointer(), m_Params);
    return job;
}




///////////////////////////////////////////////////////////////////////////////
/// CCreateGeneModelJob

CCreateGeneModelJob::CCreateGeneModelJob(CProjectService* service,
                                         const SCreateGeneModelParams& params)
:   CDataLoadingAppJob(service),
    m_Params(params)
{
    CFastMutexGuard lock(m_Mutex);

    m_Descr = "Creating gene models"; //TODO
}


void CCreateGeneModelJob::x_CreateProjectItems()
{
    TConstScopedObjects& aligns = m_Params.m_Alignments;

    ///
    /// assure we're all in one scope
    ///
    CRef<CScope> scope;
    {{
        ITERATE( TConstScopedObjects, iter, aligns ){
            if( !scope ){
                scope.Reset( const_cast<CScope*>(&*iter->scope) );
            } else if( scope != &*iter->scope ){
                scope.Reset();
                NCBI_THROW(
                    CException, eUnknown,
                    "All alignments must be within the same project"
                );
            }
        }
    }}

    ///
    /// meat goes here
    ///

    CFeatureGenerator::TFeatureGeneratorFlags opts = 0;
    if( m_Params.m_CreateGene ){
        opts |= CFeatureGenerator::fCreateGene;
    }
    if( m_Params.m_CreateMrna ){
        opts |= CFeatureGenerator::fCreateMrna;
    }
    if( m_Params.m_CreateCds ){
        opts |= CFeatureGenerator::fCreateCdregion;
    }
    if( m_Params.m_PropagateNcrnaFeats ){
        opts |= CFeatureGenerator::fPropagateNcrnaFeats;
    }
    if( m_Params.m_TranscribeMrna ){
        opts |= CFeatureGenerator::fForceTranscribeMrna;
    }
    if( m_Params.m_TranslateCds ){
        opts |= CFeatureGenerator::fForceTranslateCds;
    }

    bool f_group_by_gene_id  = m_Params.m_GroupByGeneId;
    f_group_by_gene_id &= (opts & CGeneModel::fCreateGene) != 0;

    size_t count_entries = 0;

    list< CRef<CSeq_annot> > annots;
    CRef<CSeq_entry> entry;
    ITERATE( TConstScopedObjects, iter, aligns ){
        const CSeq_align& align = dynamic_cast<const CSeq_align&>(*iter->object);
        //try 
        {
            CRef<CSeq_annot> annot( new CSeq_annot() );
            CRef<CBioseq_set> translated_proteins( new CBioseq_set() );
            CFeatureGenerator FeatureGenerator(*scope);
            FeatureGenerator.SetFlags(opts | CFeatureGenerator::fGenerateLocalIds);
            FeatureGenerator.SetAllowedUnaligned(CFeatureGenerator::kDefaultAllowedUnaligned);
            CConstRef<CSeq_align> clean_align = FeatureGenerator.CleanAlignment(align);
            FeatureGenerator.ConvertAlignToAnnot(*clean_align, *annot, *translated_proteins);

            if( annot->IsSetData() ){
                annots.push_back( annot );
                if( translated_proteins->IsSetSeq_set() ){
                    if( !entry ){
                        entry.Reset( new CSeq_entry() );
                    }
                    CRef<CSeq_entry> proteins( new CSeq_entry() );
                    proteins->SetSet( *translated_proteins );
                    entry->SetSet().SetSeq_set().push_back( proteins );
                    ++count_entries;
                }
            }
//        } catch( CException& e ){
//            LOG_POST( Error << "CreateGeneModel-from-align tool: " << e.GetMsg() );
        }
    }

    if( annots.size() ){
        ///
        /// first, separate by seq-id
        /// we plan to create a single annotation for each placed sequence
        ///
        typedef map<CSeq_id_Handle, list< CRef<CSeq_feat> > > TFeatMap;
        TFeatMap feats;

        NON_CONST_ITERATE (list< CRef<CSeq_annot> >, iter, annots) {
            NON_CONST_ITERATE (CSeq_annot::TData::TFtable, i, (*iter)->SetData().SetFtable()) {
                CSeq_feat& feat = **i;
                CSeq_id_Handle idh = sequence::GetIdHandle(feat.GetLocation(), scope);
                if (idh) { // check for an empty handle (e.g., when the location contains more than one id)
                    feats[idh].push_back(CRef<CSeq_feat>(&feat));
                }
            }
        }
        annots.clear();

        CTime time(CTime::eCurrent);

        NON_CONST_ITERATE( TFeatMap, iter, feats ){
            list< CRef<CSeq_feat> >::const_iterator feat_it = iter->second.begin();
            if( feat_it == iter->second.end() ){
                continue;
            }

            TSeqRange range = (*feat_it)->GetLocation().GetTotalRange();
            for( ++feat_it;  feat_it != iter->second.end();  ++feat_it ){
                range += (*feat_it)->GetLocation().GetTotalRange();
            }

            CSeq_id_Handle idh = iter->first;
            idh = sequence::GetId( idh, *scope, sequence::eGetId_Best );

            string label;
            idh.GetSeqId()->GetLabel( &label, CSeq_id::eContent );

            label += " (";
            label += NStr::IntToString( range.GetFrom() + 1, NStr::fWithCommas );
            label += "..";
            label += NStr::IntToString( range.GetTo() + 1, NStr::fWithCommas );
            label += ") ";
            label += "Gene Models from Alignments";

            CRef<CSeq_annot> annot( new CSeq_annot() );
            annot->SetNameDesc( "Gene Models from Alignments" );
            annot->SetTitleDesc( label );
            annot->SetCreateDate( time );

            annot->SetData().SetFtable().insert(
                annot->SetData().SetFtable().end(),
                iter->second.begin(), iter->second.end()
            );

            ///
            /// make sure we deal with single gene requirements
            ///
            if( f_group_by_gene_id ){
                ///
                /// now, find the gene for each list of features
                /// when we're done, create an annotation for each
                ///
                typedef map<int, list< CRef<CSeq_feat> > > TGeneFeats;
                TGeneFeats feats_by_gene;
                ITERATE( CSeq_annot::TData::TFtable, it, annot->GetData().GetFtable() ){
                    int gene_id = 0;
                    CConstRef<CDbtag> tag = (*it)->GetNamedDbxref("GeneID");
                    if( !tag ){
                        tag = (*it)->GetNamedDbxref("LocusLink");
                    }
                    if( tag ){
                        gene_id = tag->GetTag().GetId();
                    }
                    feats_by_gene[gene_id].push_back(*it);
                }

                list< CRef<CSeq_feat> > feats;
                NON_CONST_ITERATE( TGeneFeats, it, feats_by_gene ){
                    if( !it->first ){
                        /// failed to find gene id - pass through
                        feats.insert( feats.end(), it->second.begin(), it->second.end() );

                    } else {
                        /// scan for gene feature
                        CRef<CSeq_feat> gene;
                        TSeqRange longest_range;
                        list< CRef<CSeq_feat> >::iterator i = it->second.begin();
                        while( i != it->second.end() ){
                            CSeq_feat& feat = **i;
                            if( feat.GetData().GetSubtype() != CSeqFeatData::eSubtype_gene ){
                                ++i;
                            } else {
                                if( !gene ){
                                    gene = *i;
                                    longest_range = gene->GetLocation().GetTotalRange();
                                } else {
                                    TSeqRange range = (*i)->GetLocation().GetTotalRange();
                                    longest_range += range;
                                }
                                i = it->second.erase(i);
                            }
                        }

                        if( gene ){
                            gene->SetLocation().SetInt().SetFrom( longest_range.GetFrom() );
                            gene->SetLocation().SetInt().SetTo( longest_range.GetTo() );
                            feats.push_back( gene );
                        }

                        feats.insert( feats.end(), it->second.begin(), it->second.end() );
                    }
                }

                annot->SetData().SetFtable().swap( feats );
            }

            annots.push_back( annot );

            if( m_Params.m_MergeResults ){
                CBioseq_Handle handle = scope->GetBioseqHandle( idh );
                if( handle ){
                    CSeq_entry_Handle seh = handle.GetTopLevelEntry();
                    CSeq_entry_EditHandle main_entry = seh.GetEditHandle();

                    try {
                        main_entry.AttachAnnot( *annot );
                    
                    } catch( CException& aex ){
                        //try 
                        {
                            CSeq_annot_Handle annot_handle = scope->GetSeq_annotHandle( *annot );
                            CSeq_annot_EditHandle annot_ehandle = annot_handle.GetEditHandle();
                            main_entry.TakeAnnot( annot_ehandle );
/*
                        } catch( CException& e ){
                            LOG_POST( Error 
                                << "TakeAnnot() is failed as well as AttachAnnot(): "
                                << e.GetMsg() << " after " << aex.GetMsg() 
                            );
*/
                        }
                    }
                }
            }
        }


        ///
        /// add as appropriate
        ///
        if( !m_Params.m_MergeResults ){
            NON_CONST_ITERATE( list< CRef<CSeq_annot> >, iter, annots ){
                /// now create a Project Item for the data
                CRef<CProjectItem> item(new CProjectItem());
                item->SetItem().SetAnnot(**iter);

                string name;
                CLabel::GetLabel(**iter, &name, CLabel::eDefault, scope.GetPointer());
                item->SetLabel(name);

                m_Items.push_back(item);
            }

        } else {
            // We do not need any message like "No items created".
            x_SetTextResult( "" );
        }
    }
}

END_NCBI_SCOPE

