/*  $Id: fasta_export_job.cpp 25006 2012-01-09 16:24:23Z katargir $
* ===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data, the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties, express or implied, including
*  warranties of performance, merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================
*
* Authors: Roman Katargin
*
*/

#include <ncbi_pch.hpp>

#include <corelib/ncbifile.hpp>

#include <objects/seqloc/Seq_loc.hpp>
#include <objects/seqset/Seq_entry.hpp>
#include <objects/seq/Seq_descr.hpp>

#include <objmgr/util/sequence.hpp>
#include <objmgr/bioseq_ci.hpp>
#include <objmgr/seq_vector.hpp>

#include <gui/objutils/label.hpp>
#include <gui/objutils/utils.hpp>
#include <gui/objutils/obj_convert.hpp>

#include <gui/packages/pkg_sequence/fasta_export_job.hpp>

BEGIN_NCBI_SCOPE
USING_SCOPE(ncbi::objects);

CFastaExportJob::CFastaExportJob(const CFastaExportParams& params)
: CAppJob("FASTA Export"), m_Params(params)
{
}

IAppJob::EJobState CFastaExportJob::Run()
{
    bool master = m_Params.GetSaveMasterRecord();
    wxString fname = m_Params.GetFileName();
    string err_msg;

    try {
        CNcbiOfstream ostr(fname.fn_str());
        CFastaOstream fasta_ostr(ostr);
        fasta_ostr.SetFlag(CFastaOstream::fAssembleParts);
        fasta_ostr.SetFlag(CFastaOstream::fInstantiateGaps);

        // Do not dump duplicate seq_ids
        set<CSeq_id_Handle> idh_set;

        TConstScopedObjects objects = m_Params.GetObjects();
        ITERATE(TConstScopedObjects, it, objects) {
            const CSeq_loc* loc = dynamic_cast<const CSeq_loc*>(it->object.GetPointer());
            if (0 == loc)
                continue;

            CScope& scope = const_cast<CScope&>(*(it->scope));

            if (master) {
                CObjectConverter::TObjList objs;
                CObjectConverter::Convert(scope, *loc,
                                          CSeq_entry::GetTypeInfo(), objs);
                ITERATE (CObjectConverter::TObjList, iter, objs) {
                    const CSeq_entry& entry =
                        dynamic_cast<const CSeq_entry&>(iter->GetObject());
                    CBioseq_CI bioseq_iter(scope, entry);
                    for ( ;  bioseq_iter;  ++bioseq_iter) {
                        fasta_ostr.Write(*bioseq_iter);
                    }
                }
            } else if ( !loc->IsWhole() ) {

                CBioseq_Handle handle = scope.GetBioseqHandle(*loc);
                fasta_ostr.Write(handle, loc);
/*
                /// build a fake bioseq to hold the constructed sequence
                CBioseq bioseq;

                CRef<CSeq_id> id(new CSeq_id());
                id->Assign(sequence::GetId(*loc, &scope));
                bioseq.SetId().push_back(id);

                string title("sequence constructed from ");
                CLabel::GetLabel(*loc, &title, CLabel::eDefault, &scope);
                CRef<CSeqdesc> desc(new CSeqdesc());
                desc->SetTitle(title);
                bioseq.SetDescr().Set().push_back(desc);

                CBioseq_Handle handle = scope.GetBioseqHandle(*loc);
                CSeqVector vec(*loc, scope, CBioseq_Handle::eCoding_Iupac);
                string seq;
                vec.GetSeqData(0, vec.size(), seq);
                if (vec.IsProtein()) {
                    bioseq.SetInst().SetSeq_data().SetIupacaa().Set(seq);
                    bioseq.SetInst().SetMol(CSeq_inst::eMol_aa);
                } else {
                    bioseq.SetInst().SetSeq_data().SetIupacna().Set(seq);
                    bioseq.SetInst().SetMol(CSeq_inst::eMol_na);
                }
                bioseq.SetInst().SetRepr(CSeq_inst::eRepr_raw);
                bioseq.SetInst().SetLength(seq.length());

                fasta_ostr.Write(bioseq);
*/
            } else {
                const CSeq_id* seq_id = loc->GetId();
                if (seq_id) {
                    CSeq_id_Handle idh = CSeq_id_Handle ::GetHandle(*seq_id);
                    if (idh_set.find(idh) != idh_set.end())
                        continue;
                    idh_set.insert(idh);
                }

                CBioseq_Handle handle = scope.GetBioseqHandle(*loc);
                fasta_ostr.Write(handle);
            }
        }
    }
    catch (CException& e) {
        err_msg = "Failed to save file:\n";
        err_msg += e.GetMsg();
    }

    if (err_msg.empty())  {
        LOG_POST(Info << "CFastaExportJob::Run()  Finished  " << m_Descr);
        return eCompleted;
    } else {
        m_Error.Reset(new CAppJobError(err_msg));
        return eFailed;
    }
}

END_NCBI_SCOPE

