/*  $Id: snp_search_tool.cpp 27635 2013-03-15 19:33:00Z katargir@NCBI.NLM.NIH.GOV $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Andrey Yazhuk, modified for SNP by Dmitry Rudnev
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>

#include <gui/packages/pkg_sequence/snp_search_tool.hpp>
#include <gui/core/data_mining_service.hpp>

#include <gui/objutils/label.hpp>

#include <gui/widgets/wx/ui_tool_registry.hpp>

#include <gui/utils/extension_impl.hpp>
#include <gui/utils/fetch_url.hpp>

#include <objmgr/object_manager.hpp>

#include <wx/sizer.h>
#include <wx/stattext.h>
#include <wx/choice.h>
#include <wx/srchctrl.h>

#include <gui/widgets/wx/wx_utils.hpp>

#include <misc/xmlwrapp/xmlwrapp.hpp>
#include <misc/xmlwrapp/event_parser.hpp>
#include <connect/ncbi_conn_stream.hpp>

#include <objects/general/Dbtag.hpp>
#include <objects/general/Object_id.hpp>

#include <objects/variation/Variation.hpp>
#include <objects/variation/VariantPlacement.hpp>
#include <objects/seqfeat/VariantProperties.hpp>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);


//!! known problems:
// - create feature does not work

#define ID_COMBOBOX 11003 //!! ?? which IDs to use, is there any significance in these particular values?
#define ID_TEXT     11414

typedef list<CRef<CVariation> > TSNPSearchResultList;


class CSNPWebService
{
public:
	static void Query(const string& sTerm, TSNPSearchResultList& ResultList);
};

void CSNPWebService::Query(const std::string& sTerm, TSNPSearchResultList& ResultList)
{
    string sSNPSearchURL("http://www.ncbi.nlm.nih.gov/projects/SNP/beVarSearch.cgi?format=xml&report=varloc&id=");

	sSNPSearchURL += sTerm;
	
	string sSearchResult;
	CFetchURL::Fetch(sSNPSearchURL, sSearchResult);

	xml::document ResultXml(sSearchResult.c_str(), sSearchResult.length(), NULL);
    xml::node& ResultRoot(ResultXml.get_root_node());
	// get the SNP id common for all results
	xml::xpath_expression ParamIdExpr("/result_set/query/param_id");
	const xml::node_set ParamIdSet(ResultRoot.run_xpath_query(ParamIdExpr));

	// currently expect one and only one ParamId
	NCBI_ASSERT(ParamIdSet.size() == 1, "Unexpected number of <param_id> in SNP search result!");

	const xml::attributes& ParamIdAttrs(ParamIdSet.begin()->get_attributes());
	string sClin, sComment, sResultID;
	ITERATE(xml::attributes, iParamIdAttrs, ParamIdAttrs) {
		string sAttrName(iParamIdAttrs->get_name());
		if(sAttrName == "isClin")
			sClin = iParamIdAttrs->get_value();
		if(sAttrName == "comment")
			sComment = iParamIdAttrs->get_value();
		if(sAttrName == "result_id")
			sResultID = iParamIdAttrs->get_value();
	}
	
	// parse the individual results which are expected to have the following structure:
	// <result_set><query><placement_set><placement placement_type= to= from= assembly_name= accession= /><placement .../> etc
	xml::xpath_expression PlacementExpr("/result_set/query/placement_set/placement");
	const xml::node_set PlacementSet(ResultRoot.run_xpath_query(PlacementExpr));

	ITERATE(xml::node_set, iPlacementSet, PlacementSet) {
		const xml::attributes& PlacementAttrs(iPlacementSet->get_attributes());
		string sAccession, sFrom, sTo, sAssemblyName, sGeneSymbol;

		ITERATE(xml::attributes, iPlacementAttrs, PlacementAttrs) {
			string sAttrName(iPlacementAttrs->get_name());
			if(sAttrName == "to")
				sTo = iPlacementAttrs->get_value();
			if(sAttrName == "from")
				sFrom = iPlacementAttrs->get_value();
			if(sAttrName == "assembly_name")
				sAssemblyName = iPlacementAttrs->get_value();
			if(sAttrName == "accession")
				sAccession = iPlacementAttrs->get_value();
			if(sAttrName == "gene_symbol")
				sGeneSymbol = iPlacementAttrs->get_value();
		}
        CRef<CSeq_id>  pId(new CSeq_id(sAccession));
        CRef<CSeq_loc> pLocation(new CSeq_loc(*pId, 
											  NStr::StringToNumeric(sFrom),
											  NStr::StringToNumeric(sTo)));

		CRef<CVariantPlacement> pPlacement(new CVariantPlacement);
        pPlacement->SetLoc(*pLocation);
        //!! review this, try to obtain actual molecule type from pId
        pPlacement->SetMol(CVariantPlacement::eMol_genomic);
		pPlacement->SetAssembly().SetDb("");
		pPlacement->SetAssembly().SetTag().SetStr(sAssemblyName);
		CRef<CVariation> pVariation(new CVariation);
        pVariation->SetPlacements().push_back(pPlacement);
		pVariation->SetData().SetNote("SNP search result");
		pVariation->SetDescription("SNP search result");
		pVariation->SetId().SetDb("dbSNP");
		pVariation->SetId().SetTag().SetStr(sResultID);
		if(sClin == "Y")
			pVariation->SetVariant_prop().SetResource_link(CVariantProperties::eResource_link_clinical);
		if(!sComment.empty())
			pVariation->SetDescription(sComment);

		// there is no direct place to put gene_symbol, so it goes to user object
		if(!sGeneSymbol.empty()) {
			CRef<CUser_object> pGeneSymbol(new CUser_object());
			pGeneSymbol->AddField("gene_symbol", sGeneSymbol);
			pVariation->SetExt().push_back(pGeneSymbol);
		}


		ResultList.push_back(pVariation);
	}
}



///////////////////////////////////////////////////////////////////////////////
/// CSNPTableModel


class CSNPTableModel : public CObjectListTableModel
{
public:
    CSNPTableModel( );

    virtual wxVariant GetExtraValueAt( int row, int col ) const;
};


///////////////////////////////////////////////////////////////////////////////
/// IDMSearchTool
CSNPSearchTool::CSNPSearchTool()
{
}


IUITool* CSNPSearchTool::Clone() const
{
    return new CSNPSearchTool();
}


string CSNPSearchTool::GetName() const
{
    static string s_name("Search dbSNP");
    return s_name;
}


string CSNPSearchTool::GetDescription() const
{
    return "Search for rs or ss accessions from dbSNP";
}


CIRef<IDMSearchForm> CSNPSearchTool::CreateSearchForm()
{
    CIRef<IDMSearchForm> form(new CSNPSearchForm(*this));
    return form;
}


bool CSNPSearchTool::IsCompatible(IDataMiningContext* context)
{
    return false;
}


CRef<CSearchJobBase> CSNPSearchTool::x_CreateJob(IDMSearchQuery& query)
{
    CRef<CSearchJobBase> job;
    CSNPSearchQuery* e_query = dynamic_cast<CSNPSearchQuery*>(&query);
    if(e_query) {
        job.Reset(new CSNPSearchJob(*e_query));
    }
    return job;
}

string  CSNPSearchTool::GetExtensionIdentifier() const
{
    return "search_tool::SNP_search_tool";
}

string  CSNPSearchTool::GetExtensionLabel() const
{
    return "Datamining Tool - dbSNP Search";
}


///////////////////////////////////////////////////////////////////////////////
/// CSNPSearchTool
CSNPSearchQuery::CSNPSearchQuery(const string& terms)
:   m_Terms(terms)
{
}


///////////////////////////////////////////////////////////////////////////////
/// CSNPSearchForm

CSNPSearchForm::CSNPSearchForm(CSNPSearchTool& tool)
:   m_Tool(&tool)
{
}


void CSNPSearchForm::Create()
{
}


void CSNPSearchForm::Init()
{
    CSearchFormBase::Init();
}


void CSNPSearchForm::x_LoadSettings(CGuiRegistry::TReadView& view)
{
    CSearchFormBase::x_LoadSettings(view);
}


void CSNPSearchForm::x_SaveSettings(CGuiRegistry::TReadWriteView& view) const
{
    CSearchFormBase::x_SaveSettings(view);
}


wxSizer* CSNPSearchForm::GetWidget(wxWindow * parent)
{
    if ( !m_Sizer) {        
        wxFlexGridSizer * sz = new wxFlexGridSizer(1, 1, 0, 0);     
        sz->AddGrowableCol(0);
        m_Sizer = sz;
        m_Text = new CSearchControl(parent, ID_TEXT, wxT(""),
                                    wxDefaultPosition, wxDefaultSize,
                                    wxTE_PROCESS_ENTER );                       
        m_Sizer->Add(m_Text,1, wxGROW|wxALIGN_CENTER_VERTICAL|wxALL, 5);        
    }
    return m_Sizer;
}

void  CSNPSearchForm::Update()
{
}

void CSNPSearchForm::UpdateContexts()
{
    CSearchFormBase::UpdateContexts();
    // context - independent
    if (m_Controller) m_Controller->OnSearchEnabled(true);
}


CIRef<IDMSearchQuery> CSNPSearchForm::ConstructQuery()
{
    string q_s = ToStdString(m_Text->GetValue());
    CIRef<IDMSearchQuery> ref(new CSNPSearchQuery(q_s));
    return ref;
}


///////////////////////////////////////////////////////////////////////////////
/// CSNPSearchJob

CSNPSearchJob::CSNPSearchJob(CSNPSearchQuery& query)
:   m_Query(&query)
{
    m_Descr = "Query: " + m_Query->GetTerms();
}


bool CSNPSearchJob::x_ValidateParams()
{
    if(m_Query->GetTerms().empty()) {
        m_Error = new CAppJobError("Invalid input parameters - no search terms specified.");
        return false;
    }
    return true;
}


IAppJob::EJobState CSNPSearchJob::x_DoSearch()
{
    // prepare search params and search
    string terms = m_Query->GetTerms();
    size_t total_uids = 0;

    TSNPSearchResultList SNPSearchResultList;
    
	CSNPWebService::Query(terms, SNPSearchResultList);
    
    if(!IsCanceled())    {
		CRef<CObjectManager> om(CObjectManager::GetInstance());
		CRef<CScope> scope(new CScope(*om));
		scope->AddDefaults();

        // process results
        CMutexGuard Guard(m_Mutex);

        CObjectList * obj_list = m_TempResult->GetObjectList();
        NON_CONST_ITERATE(TSNPSearchResultList, iSNPSearchResultList, SNPSearchResultList)    {
			CObject* pSearchResultAsObject(iSNPSearchResultList->GetPointer());
            int NewRow(obj_list->AddRow(pSearchResultAsObject, scope.GetPointer()));

			// CVariation used as a search result can have one and only one placement
			NCBI_ASSERT((*iSNPSearchResultList)->CanGetPlacements(), "Unexpected absence of placements in SNP Search Result!");
			const CVariation::TPlacements& Placements((*iSNPSearchResultList)->GetPlacements());
			NCBI_ASSERT(Placements.size(), "Unexpected number of placements in SNP Search Result!");
			CRef<CVariantPlacement> pPlacement(Placements.front());
			int iCol(-1);
			obj_list->SetInteger(++iCol, NewRow, pPlacement->GetLoc().GetStart(eExtreme_Positional) + 1);
			obj_list->SetInteger(++iCol, NewRow, pPlacement->GetLoc().GetStop(eExtreme_Positional) + 1);
			obj_list->SetString(++iCol, NewRow, pPlacement->GetAssembly().GetTag().GetStr());
			
			// the next column is gene_symbol which is encoded in ext and possibly not present
			++iCol;
			if((*iSNPSearchResultList)->CanGetExt()) {
				const CVariation::TExt& Ext((*iSNPSearchResultList)->GetExt());
				ITERATE(CVariation::TExt, iExt, Ext) {
					if((*iExt)->HasField("gene_symbol")) {
						const CUser_field& UserField((*iExt)->GetField("gene_symbol"));
						// must be a string or we cry
						NCBI_ASSERT(UserField.CanGetData() && UserField.GetData().IsStr(), "Wrong type of user field for gene_symbol");
						obj_list->SetString(iCol, NewRow, UserField.GetData().GetStr());
						break;
					}
				}
			}

            if(IsCanceled()) {
                return eCanceled;
            }
        }
		
		return eCompleted;
    }
    return eCanceled;
}

void CSNPSearchJob::x_SetupColumns( CObjectList& obj_list )
{
	obj_list.ClearObjectLabels();
	obj_list.AddObjectLabel("Accession", CLabel::eContent );

    obj_list.AddColumn( CObjectList::eInteger, "Start" );
    obj_list.AddColumn( CObjectList::eInteger, "Stop" );
    obj_list.AddColumn( CObjectList::eString, "Assembly Name" );
    obj_list.AddColumn( CObjectList::eString, "Gene symbol" );
}


CObjectListTableModel* CSNPSearchJob::x_GetNewOLTModel() const
{
    return new CSNPTableModel();
}


///////////////////////////////////////////////////////////////////////////////
/// CSNPTableModel


CSNPTableModel::CSNPTableModel()
{
}


wxVariant CSNPTableModel::GetExtraValueAt( int row, int col ) const
{
    const CObject* obj = m_ObjectList->GetObject( row );
    const CVariation* pSNPSearchResult = dynamic_cast<const CVariation*>(obj);
	if(pSNPSearchResult) {
		// CVariation used as a search result can have one and only one placement
		NCBI_ASSERT(pSNPSearchResult->CanGetPlacements(), "Unexpected absence of placements in SNP Search Result!");
		const CVariation::TPlacements& Placements(pSNPSearchResult->GetPlacements());
		NCBI_ASSERT(Placements.size(), "Unexpected number of placements in SNP Search Result!");
		CRef<CVariantPlacement> pPlacement(Placements.front());
		const CSeq_id* pId(pPlacement->GetLoc().GetId());
		NCBI_ASSERT(pId, "Unexpected no or multiple IDs per SNP Search Result!");

		return ToWxString(pId->GetSeqIdString(true));
	}
    return wxT("");  
}

END_NCBI_SCOPE
