/*  $Id: bins_ds.cpp 27815 2013-04-08 15:10:18Z rudnev $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Melvin Quintos, Dmitry Rudnev
 *
 * Description:  This file defines CBinsJob, CBinsDS, and CBinsDSType
 *
 */

#include <ncbi_pch.hpp>
#include <gui/packages/pkg_snp/bins/bins_ds.hpp>
#include <gui/packages/pkg_snp/bins/bins_glyph.hpp>
#include <gui/packages/pkg_snp/utils/utils.hpp>

#include <gui/objutils/utils.hpp>
#include <gui/utils/app_job_dispatcher.hpp>
#include <gui/widgets/seq_graphic/histogram_glyph.hpp>

#include <objmgr/graph_ci.hpp>
#include <objmgr/seq_loc_mapper.hpp>
#include <objmgr/seq_map_switch.hpp>
#include <objmgr/seqdesc_ci.hpp>
#include <objmgr/table_field.hpp>
#include <objmgr/util/feature.hpp>
#include <objmgr/util/sequence.hpp>
#include <objmgr/util/sequence.hpp>

#include <objects/seq/Annot_descr.hpp>
#include <objects/seq/Annotdesc.hpp>

#include <objects/seqres/Byte_graph.hpp>
#include <cmath>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

///////////////////////////////////////////////////////////////////////////////
///   CBinsJob
///////////////////////////////////////////////////////////////////////////////
class CBinsJob : public CSGAnnotJob
{
public:
    struct SParams {
        bool        isSeqView;
        bool        bAdaptiveSelector;
        TModelUnit  scale;
        int         binWidth;
        int         depth;
        string      name;
    };

    CBinsJob(const CBioseq_Handle& handle, const TSeqRange& range,
        const SParams& params)
        : CSGAnnotJob("SNP", handle, SAnnotSelector(), range)
        , m_Params(params) {}

    virtual EJobState   x_Execute();

private:
    void    x_FetchGraphs(const CSeq_loc &, CSeqGlyph::TObjects*);
    void    x_FetchFtable(const CSeq_loc &, CSeqGlyph::TObjects*);

    // will try to load and create bins and return true;
    // if however the amount of positions would be unreasonably (>500) big, will return
    // false and glyphs will be undefined
    bool x_PreFetchFtable(const SAnnotSelector& sel,
							  const objects::CSeq_loc &loc,
							  CSeqGlyph::TObjects* glyphs);
    void x_ReadAnnotDesc(const CSeq_annot_Handle& handle, string* title, string* comment);

private:
    SParams                     m_Params;
    SBinsJobResult::EDataType   m_eDataType;
};

IAppJob::EJobState CBinsJob::x_Execute()
{
    EJobState state = eCompleted;

    try {
        CSeqGlyph::TObjects glyphs;

        CRef<CSeq_loc> loc = m_Handle.GetRangeSeq_loc(m_Range.GetFrom(), m_Range.GetTo());

        x_FetchFtable(*loc, &glyphs);

        SBinsJobResult *result = new SBinsJobResult();
        m_Result.Reset(result);

        result->listObjs.swap(glyphs);
        result->eDataType = m_eDataType;
    }
    catch (CException& ex) {
        m_Error.Reset(new CAppJobError(ex.GetMsg()));
        state = eFailed;
    }
    catch (std::exception& ex) {
        m_Error.Reset(new CAppJobError(ex.what()));
        state = eFailed;
    }

    return state;
}

void CBinsJob::x_ReadAnnotDesc(const CSeq_annot_Handle& handle, string* title, string* comment)
{
    if (handle.Seq_annot_CanGetDesc()) {
        // Extract the Annotations properties
        ITERATE( CAnnot_descr::Tdata, it, handle.Seq_annot_GetDesc().Get()) {
            const CAnnotdesc &desc = **it;

            if (desc.IsComment()) {
                *comment = desc.GetComment();
            }
            else if (desc.IsTitle()) {
                *title = desc.GetTitle();
            }
        }
    }

}

void CBinsJob::x_FetchFtable(const CSeq_loc &loc, CSeqGlyph::TObjects *objs)
{
    SAnnotSelector sel =
        CSeqUtils::GetAnnotSelector(CSeq_annot::TData::e_Seq_table);
    sel.IncludeNamedAnnotAccession(m_Params.name);
    sel.AddNamedAnnots(m_Params.name);
    CSeqUtils::SetResolveDepth(sel, m_Params.bAdaptiveSelector, m_Params.depth);

    CSeqGlyph::TObjects tmpList;
    bool isBinsLoaded(x_PreFetchFtable(sel, loc, &tmpList));

    if(isBinsLoaded) {
	    objs->swap(tmpList);
        m_eDataType = SBinsJobResult::eBins;
    }
    else {
        x_FetchGraphs(loc, objs);
    }
}

bool CBinsJob::x_PreFetchFtable(const SAnnotSelector& sel,
                                const CSeq_loc &loc,
                                CSeqGlyph::TObjects *objs)
{
    try {

    CTableFieldHandle<int>      col_pos("pos");
    CTableFieldHandle<double>   col_val("pvalue");
    CTableFieldHandle<string>   col_trait("trait");
    CTableFieldHandle<string>   col_pmids("pmids");
    CTableFieldHandle<string>   col_rgenes("reportedGenes");
    CTableFieldHandle<string>   col_mgenes("mappedGenes");
    CTableFieldHandle<int>      col_snpid("snpId");
    CTableFieldHandle<int>      col_type("trackType");
    CTableFieldHandle<int>      col_clinsigid("clinSigID");

    int     pos, snpid, type, ClinSigID;
    double  pvalue;
    string  trait, pmids, rgenes, mgenes;
    string  title, comment;
    string  dbgaptext;
    int     pos_start, pos_end;

    bool    hasResult   = false; // data has at least 1 non-zero bin
    double  scale       = fabs(m_Params.scale);
    const int binWidth  = m_Params.binWidth;

    double window = scale*binWidth;

    if (window < 1 || m_Params.isSeqView) {
        window = 1;
    }

    CBinsGlyph::TDensityMap   dmap(
        m_Range.GetFrom(),
        m_Range.GetTo(),
        window,
        new CBinsGlyph::accum_study());

    // Grab annotations and start processing data
    CAnnot_CI iter(m_Handle.GetScope(), loc, sel);

    //!! indicate internal error here
    if (iter.size()!=1) return false;

    CSeq_annot_Handle annot = *iter;
    x_ReadAnnotDesc(annot, &title, &comment);
    NSnpUtils::FindRange(annot, (int)m_Range.GetFrom(), (int)m_Range.GetTo(), &pos_start, &pos_end);

    // Read the type information.
    //!! indicate internal error here
    if (!col_type.TryGet(annot, 0, type)) return false;

    // if there are too many rows, abort the load and indicate that a graph should be loaded instead
    if(pos_end - pos_start > 500)
        return false;

    // Read all the rows and columns
    for ( int row = pos_start; row < pos_end; ++row ) {
        if ( col_pos.TryGet(annot, row, pos)       &&
             col_val.TryGet(annot, row, pvalue)    &&
             col_trait.TryGet(annot, row, trait)   &&
             col_pmids.TryGet(annot, row, pmids)   &&
             col_rgenes.TryGet(annot, row, rgenes) &&
             col_mgenes.TryGet(annot, row, mgenes) &&
             col_snpid.TryGet(annot, row, snpid)) {

            TSeqRange    range  = TSeqRange(pos, pos);

            CRef<CBinsGlyph::SSnp> snp(new CBinsGlyph::SSnp);
            snp->pvalue = -log10(pvalue);
            snp->traits = trait;
            snp->pmids  = pmids;
            snp->rgenes = rgenes;
            snp->mgenes = mgenes;
            snp->snpid  = (unsigned int)snpid;

            // clinSigID is optional
            try {
            snp->ClinSigID = col_clinsigid.TryGet(annot, row, ClinSigID) ? ClinSigID : -1;
            } catch(...) {
                snp->ClinSigID = -1;
            }


            CBinsGlyph::SBinMapped bin;
            bin.obj.Reset(new CBinsGlyph::SBin);
            bin.obj->count = 1;
            bin.obj->range = range;
            bin.obj->snps.push_back(snp);
            bin.obj->sigSnp = snp;

            dmap.AddRange(range, bin);
            hasResult = true;
        }
    }

    if (hasResult) {
        size_t binCount(0);
        NON_CONST_ITERATE(CBinsGlyph::TDensityMap, iBin, dmap) {
            iBin->SequentialNumber = binCount++;
        }
        CRef<CBinsGlyph>  bins;
        bins.Reset(new CBinsGlyph(dmap, loc, &m_Handle.GetScope()));
        bins->SetTitle(title);
        bins->SetAnnotName(m_Params.name);
        bins->SetType((CBinsGlyph::EType)type);
        CRef<CSeqGlyph> fref(bins.GetPointer());
        objs->push_back(fref);
    }

    } catch (CException& e) {
        m_Error.Reset(new CAppJobError(e.GetMsg()));
        //!! abort properly
        //state = eFailed;
    } catch (std::exception&) {
        m_Error.Reset(new CAppJobError("Exception when loading Named Annotations "));
        //!! abort properly
        //state = eFailed;
    }
    return true;
}

void CBinsJob::x_FetchGraphs(const CSeq_loc& loc, CSeqGlyph::TObjects *glyphs)
{
    SAnnotSelector sel;
    sel = CSeqUtils::GetAnnotSelector(CSeq_annot::TData::e_Seq_table);
    sel.IncludeNamedAnnotAccession(m_Params.name);
    sel.AddNamedAnnots(m_Params.name);
    CSeqUtils::SetResolveDepth(sel, m_Params.bAdaptiveSelector, m_Params.depth);
    CSeqGlyph::TObjects tmpList;

    CTableFieldHandle<int>      col_pos("pos");

    int     pos;
    int     pos_start, pos_end;

    CHistogramGlyph::TMap dmap(m_Range.GetFrom(), m_Range.GetTo(), m_Params.scale);

    // Grab annotations and start processing data
    CAnnot_CI iter(m_Handle.GetScope(), loc, sel);

    //!! indicate internal error here
    if (iter.size()!=1) return;

    CSeq_annot_Handle annot = *iter;
    string  title, comment;
    x_ReadAnnotDesc(annot, &title, &comment);
    NSnpUtils::FindRange(annot, (int)m_Range.GetFrom(), (int)m_Range.GetTo(), &pos_start, &pos_end);

    // Read all the rows and columns
    for ( int row = pos_start; row < pos_end; ++row ) {
        if ( col_pos.TryGet(annot, row, pos)) {

            TSeqRange    range  = TSeqRange(pos, pos);
            dmap.AddRange(range);
        }
    }
    CHistogramGlyph* hist = new CHistogramGlyph(dmap, CSeqFeatData::eSubtype_variation, title);
    hist->SetAnnotName(m_Params.name);
    glyphs->push_back(CRef<CSeqGlyph>(hist));

    m_eDataType = SBinsJobResult::eHistFromFeats;
}



///////////////////////////////////////////////////////////////////////////////
/// CBinsDS
///////////////////////////////////////////////////////////////////////////////
CBinsDS::CBinsDS(CScope& scope, const CSeq_id& id)
    : CSGGenBankDS(scope, id)
{}

void CBinsDS::GetTrackNames(SConstScopedObject& object,
                            TAnnotNameTitleMap& names,
                            const TSeqRange& range,
                            const ILayoutTrackFactory::SExtraParams& params)
{
    const CSeq_id& id = dynamic_cast<const CSeq_id&>(object.object.GetObject());
    CBioseq_Handle handle = object.scope->GetBioseqHandle(id);

    // Find data
    SAnnotSelector sel = CSeqUtils::GetAnnotSelector(params.m_Annots);
    CSeqUtils::SetResolveDepth(sel, params.m_Adaptive, params.m_Level);
    sel.SetAnnotType(CSeq_annot::TData::e_Seq_table);

    CAnnotTypes_CI annot_it(CSeq_annot::C_Data::e_Seq_table, handle,
        range, eNa_strand_unknown, &sel);
    ITERATE (CAnnotTypes_CI::TAnnotNames, iter, annot_it.GetAnnotNames()) {
        if (iter->IsNamed()) {
            names.insert(TAnnotNameTitleMap::value_type(iter->GetName(), ""));
        } else {
            names.insert(TAnnotNameTitleMap::value_type("Unnamed", ""));
        }
    }
}


void CBinsDS::SetName(const string& name)
{
    m_Name = name;
}

void CBinsDS::LoadData(const TSeqRange& range,
                        TModelUnit scale,
                        bool isSeqView)
{
    CRef<CBinsJob> job;

    CBinsJob::SParams p;
    p.isSeqView   = isSeqView;
    p.bAdaptiveSelector = m_Adaptive;
    p.scale       = scale;
    p.binWidth    = 14; // TODO Should read from global config file.
    p.depth       = m_Depth;
    p.name        = m_Name;

    job.Reset(new CBinsJob(m_Handle, range, p));

    x_LaunchJob(*job);
}

///////////////////////////////////////////////////////////////////////////////
/// CBinsDSType
///////////////////////////////////////////////////////////////////////////////

ISGDataSource*
CBinsDSType::CreateDS(SConstScopedObject& object) const
{
    const CSeq_id& id = dynamic_cast<const CSeq_id&>(object.object.GetObject());
    return new CBinsDS(object.scope.GetObject(), id);
}


string CBinsDSType::GetExtensionIdentifier() const
{
    static string sid("snp_bins_ds_type");
    return sid;
}

string CBinsDSType::GetExtensionLabel() const
{
    static string slabel("SNP Bin Track Data Source Type");
    return slabel;
}

bool CBinsDSType::IsSharable() const
{
    return false;
}


END_NCBI_SCOPE
