/*  $Id: bins_glyph.cpp 27815 2013-04-08 15:10:18Z rudnev $
* ===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data, the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties, express or implied, including
*  warranties of performance, merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================
*
* Authors:  Melvin Quintos
*
* File Description:
*    CBinsGlyph -- Glyph to represent the bins track
*
*/

#include <ncbi_pch.hpp>
#include <gui/packages/pkg_snp/bins/bins_glyph.hpp>
#include <gui/widgets/seq_graphic/rendering_ctx.hpp>
#include <gui/widgets/wx/message_box.hpp>
#include <gui/objutils/obj_fingerprint.hpp>
#include <objmgr/util/sequence.hpp>
#include <gui/opengl/gltexturefont.hpp>
#include <gui/opengl/glrender.hpp>
#include <util/checksum.hpp>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

// XXX Remove this later and place it in a configuration file
static CRgbaColor g_ColorMapping(double value)
{
    static float colors[][6] = {// color       pvalue
        {.75, 1.0, 1.0, 1.0},   // teal        < 2
        {.25, 1.0, 1.0, 1.0},   // sky blue    [2-3)
        {0.0, .50, 1.0, 1.0},   // blue        [3-4)
        {0.0, 1.0, 0.0, 1.0},   // green       [4-5)
        {1.0, 1.0, 0.0, 1.0},   // yellow      [5-6)
        {1.0, .50, 0.0, 1.0},   // orange      [6-7)
        {1.0, 0.0, 0.0, 1.0}    // red         >= 7
    };

    int v = (int)floor(value);
    float *c = 0;

    if( v < 2)
        c = colors[0];
    else if (v < 7)
        c = colors[v-1];
    else
        c = colors[6];

    return CRgbaColor( c[0], c[1], c[2], c[3] );
}

// coloring for clinical significance
static CRgbaColor g_ColorMappingClinSig(int ClinSigID)
{
/* taken from [human_9606].[dbo].[ClinSigCode] on 05/26/2011

code	abbrev	descrip
0	NULL	unknown
1	NULL	untested
2	NULL	non-pathogenic
3	NULL	probable-non-pathogenic
4	NULL	probable-pathogenic
5	NULL	pathogenic
6	NULL	drug-response
7	NULL	histocompatibility
255	NULL	other
*/
    //!! unfortunately no enums for ClinSigID yet
    switch(ClinSigID)
    {
    case 4: return CRgbaColor(0.57f, 0.43f, 0.85f, 1.0f); // light purple
    case 5: return CRgbaColor(0.5f, 0.0f, 0.5f, 1.0f); // purple
    }
    return CRgbaColor(.75f, 1.0f, .75f);    // light green
}

// coloring for clinical significance text
static CRgbaColor g_ColorMappingTextClinSig(int ClinSigID)
{
    //!! unfortunately no enums for ClinSigID yet
    switch(ClinSigID)
    {
    case 4:
    case 5: return CRgbaColor(1.0f, 1.0f, 1.0f, 1.0f); // white
    }
    return CRgbaColor(.0f, .0f, 0.0f, 1.0f);    // black
}

string CBinsGlyph::GenerateBinSignature(const string& title,
                                        const string& annot_name,
                                        int gi,
                                        TSeqRange range)
{
    char buff[128];
    Uint2 type;

    // create type
    type = (CObjFingerprint::eSnpTrack) << 8;
    type |= 0x02;  // Need to enumerate these types.   2 is for SNP Bins Format

    // create checksum
    CChecksum cs(CChecksum::eCRC32);
    Uint4 checksum;
    cs.AddLine(title);
    checksum = cs.GetChecksum();

    CChecksum cs_name(CChecksum::eCRC32);
    Uint4 checksum_name;
    cs_name.AddLine(annot_name);
    checksum_name = cs_name.GetChecksum();

    sprintf(buff, "%d-%08x-%08x-%04x-%08x-%08x",
            gi, range.GetFrom(), range.GetTo(),
            type, checksum, checksum_name);

    return string(buff);
}

//
// CBinsGlyph::CBinsGlyph()
//
CBinsGlyph::CBinsGlyph(
        const TDensityMap& mapValues,
        const CSeq_loc& loc,
        objects::CScope * scope)
    : m_Location(&loc)
    , m_Map(mapValues)
    , m_Type(eGAP)
    , m_bHasObjectList(false)
    , m_Scope(scope)
{
}

bool CBinsGlyph::OnLeftDblClick(const TModelPoint& /*p*/)
{
    string name;
    // TODO
    return true;
}


void CBinsGlyph::GetHTMLActiveAreas(CSeqGlyph::TAreaVector* p_areas) const
{
    bool isFlipped = m_Context->IsFlippedStrand();

    if (m_bHasObjectList == false) {
        x_BuildNonEmptyBinList();
    }
    double              window  = m_Map.GetWindow();
    const TSeqPos       start   = m_Map.GetStart();

    ITERATE(TMappedBinList, iNonEmptyBins, m_NonEmptyBins) {

        TSeqRange range = (*iNonEmptyBins)->obj->range;
        // the same formula as in x_Draw so the generated areas for drawing and javascript are as close as possible
        TVPUnit x1 = GetRenderingContext()->SeqToScreenX(start + (*iNonEmptyBins)->SequentialNumber*window);
        TVPUnit x2 = GetRenderingContext()->SeqToScreenX(start + ((*iNonEmptyBins)->SequentialNumber+1)*window);

        if (isFlipped) {
            x1 = -x1;
            x2 = -x2;
        }

        CHTMLActiveArea area;
        CSeqGlyph::x_InitHTMLActiveArea(area);
        area.m_Bounds.SetHorz(x1,x2);
        area.m_Flags = CHTMLActiveArea::fNoSelection | CHTMLActiveArea::fNoPin;
        area.m_SeqRange = range;
        area.m_Signature = (*iNonEmptyBins)->obj->signature;
        p_areas->push_back(area);
    }
}


bool CBinsGlyph::NeedTooltip(const TModelPoint& p, string& tt, string& t_title) const
{
    GetTooltip(p, tt, t_title);
    return !tt.empty();
}


void CBinsGlyph::GetTooltip(const TModelPoint& pos, string& tt, string& /*t_title*/) const
{
    // Determine bin number
    // bin number is floor (pos/window)
    float           window  = m_Map.GetWindow();
    TSeqPos         start   = m_Map.GetStart();
    unsigned int    bin     = (unsigned int) floor( (pos.X()-start)/window );

    // Make sure 'bin' falls within [0, m_Map.GetBins)
    bin = max<int>(0, bin);
    bin = min<int>(bin, m_Map.GetBins()-1);

    if (m_Map[bin].obj.Empty()) return;

    switch (m_Type) {
        case eGAP:      x_GetTooltipGAP(m_Map[bin].obj.GetObject(), tt);    break;
        case eGCAT:     x_GetTooltipGCAT(m_Map[bin].obj.GetObject(), tt);   break;
        case eCLIN:
        case eIND:      x_GetTooltipCLIN(m_Map[bin].obj.GetObject(), tt);   break;
        case eCITED:    x_GetTooltipCITED(m_Map[bin].obj.GetObject(), tt);  break;

        default:
            break;
    }
}


bool CBinsGlyph::IsClickable() const
{
    // TODO
    return false;
}

TSeqRange CBinsGlyph::x_GetBinRange(unsigned int bin) const
{
    // determine selected bin
    // bin number is floor (pos/window)
    double window  = m_Map.GetWindow();
    TSeqPos start = m_Map.GetStart();

    TSeqPos startPos = (TSeqPos) (
                         start              // start position
                       + bin*window
                       );

    TSeqPos endPos   =  (TSeqPos) (window != 1 ?
                             start                  // start position
                             + (bin+1)*window - 1
                           :
                             startPos
                       );

    TSeqRange range(startPos, endPos);
    return range;
}

void CBinsGlyph::x_BuildNonEmptyBinList() const
{
    // for every non-zero bin, generate the bin's signature and range
    if(!m_bHasObjectList) {
        int gi = sequence::GetGiForId(*(m_Location->GetId()), *m_Scope);
        TDensityMap::const_iterator iter;

        NON_CONST_ITERATE(TDensityMap, iBin, m_Map) {
            if(iBin->obj.NotEmpty()) {
                iBin->obj->signature = GenerateBinSignature(GetTitle(),
                                                      GetAnnotName(),
                                                      gi,
                                                      iBin->obj->range);
                m_NonEmptyBins.push_back(&(*iBin));
            }
        }
        m_bHasObjectList = true;
    }
}

void CBinsGlyph::x_Draw() const
{
    CGlRender& gl = GetGl();
    const bool isFlipped = m_Context->IsFlippedStrand();
    const int binHeight = 14; //m_ConfigSettings->GetGwasBinHeight();
    const int binWidth  = 14; //m_ConfigSettings->GetGwasBinWidth();

    TModelUnit top = GetTop();

    TModelUnit line_y1 = 1 + top;
    TModelUnit line_y2 = line_y1 + binHeight;

    // SETUP COMMON attributes
    double              window  = m_Map.GetWindow();
    const TSeqPos       start   = m_Map.GetStart();
    const size_t        numBins = m_Map.GetBins();
    //const TModelUnit    offset  = m_Context->GetOffset();

    // Y values are the same
    TModelUnit y1 = line_y1;
    TModelUnit y2 = line_y2;

    CRgbaColor color;

    TModelUnit x1 = 0.0;
    TModelUnit x2 = 0.0;

    // Render the colored bins
    for(size_t bin = 0;  bin < numBins; bin++)
    {
        if (m_Map[bin].obj.NotEmpty()) {
            TModelUnit value = m_Map[bin].obj->sigSnp->pvalue;

            switch(m_Type) {
                case eCITED:
                    color.Set(.75f, 1.0f, 1.0f);
                    break;
                case eCLIN:
                    color = g_ColorMappingClinSig(m_Map[bin].obj->sigSnp->ClinSigID);
                    break;
                case eIND:
                    color.Set(.75f, .75f, .75f);
                    break;
                case eGAP:
                default:
                    color = g_ColorMapping(value);
                    break;
            }

            // draw the stuff
            x1 = start + bin*(window);
            x2 = start + (bin+1)*(window);

            // Draw filled box
            gl.ColorC(color);
            m_Context->DrawQuad(x1, y1, x2, y2);

            gl.Color4f(.76f,.76f,.76f, 1.0f);
            m_Context->DrawRect(x1, y1, x2, y2);
        }
    }

    //CRef<CGlBitmapFont> font( new CGlBitmapFont(CGlBitmapFont::eHelvetica8) ); //m_ConfigSettings->GetGwasFont();
    CGlTextureFont font(CGlTextureFont::eFontFace_Helvetica_Bold, 8);
    CGlTextureFont font_plus(CGlTextureFont::eFontFace_Helvetica_Bold, 10);
    string text;

    for(size_t bin = 0;  bin < numBins; bin++) {
        if(m_Map[bin].obj.NotEmpty()) {
            switch(m_Type) {
                case eCLIN:
                    color = g_ColorMappingTextClinSig(m_Map[bin].obj->sigSnp->ClinSigID);
                    break;
                default:
                    color.Set(.0f, .0f, 0.0f, 1.0f);
                    break;
            }
            gl.ColorC(color);
            int count = m_Map[bin].obj->count;
            x1   = ((bin)*(window) + start);
            if( count > 9)
                text = "+";
            else
                NStr::IntToString(text, count);

            int dx = (isFlipped) ? 2 : -2;

            TModelUnit w = ((binWidth + dx)/2)*(m_Context->GetScale());
            TModelUnit h = font.TextHeight();

            if (text == "+") {
                m_Context->TextOut(&font_plus, text.c_str(),
                    x1+w, y2 - (binHeight - h) * 0.5, true);
            }
            else {
                m_Context->TextOut(&font, text.c_str(),
                    x1+w, y2 - (binHeight - h) * 0.5, true);
            }
        }
    }
    gl.Color4f(0.0f, 0.0f,0.0f, 1.0f);
}

void CBinsGlyph::x_UpdateBoundingBox()
{
    m_Height = 14;
    TSeqRange range = GetRange();
    m_Width = range.GetLength();
    m_Pos.m_X = range.GetFrom();
}

TSeqRange CBinsGlyph::GetRange(void) const
{
    return m_Location->GetTotalRange();
}

END_NCBI_SCOPE
