
#include <ncbi_pch.hpp>

#include "snp_filter_job.hpp"

#include <objects/seqres/Seq_graph.hpp>
#include <objects/seqres/Int_graph.hpp>
#include <objects/seqres/Byte_graph.hpp>
#include <objects/seqloc/Seq_loc.hpp>
#include <objects/general/User_object.hpp>

#include <objmgr/annot_selector.hpp>
#include <objmgr/feat_ci.hpp>

#include <gui/objutils/utils.hpp>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

CSnpFilterJob::CSnpFilterJob( CScope& scope, CSeq_loc& loc, const SSnpFilter &filter )
    : m_Scope(&scope)
    , m_Loc(&loc)
    , m_Total(100)
    , m_Completed(0)
{
    m_Filter.Reset(new SSnpFilter(filter));
}



CConstIRef<IAppJobProgress> CSnpFilterJob::GetProgress()
{
    CRef<CAppJobProgress> p( new CAppJobProgress() );
    p->SetNormDone( float(m_Completed) / float(m_Total) );
    return CConstIRef<IAppJobProgress>( p.GetPointer() );
}


CRef<CObject> CSnpFilterJob::GetResult()
{
    return CRef<CObject>( m_Result.GetPointer() );
}


CConstIRef<IAppJobError> CSnpFilterJob::GetError()
{
    return CConstIRef<IAppJobError>( m_Error.GetPointer() );
}


string CSnpFilterJob::GetDescr() const
{
    return string( "CSnpFilterJob" );
}


IAppJob::EJobState CSnpFilterJob::Run()
{
    const int COMPRESSION = 100;
    unsigned long   size  = 0;
    m_Error.Reset( NULL );

    CSeq_loc::TRange range;
    range = m_Loc->GetTotalRange();
    size  = range.GetLength();

    size = (size / COMPRESSION) + 1;  // number of values needed for compressed data

    CRef<CSeq_loc> loc = m_Loc;
    // Create graph
    CSeq_graph *graph  = new CSeq_graph();
    graph->SetLoc(*loc);
    graph->SetNumval(size);
    graph->SetGraph().SetByte().SetMin(0);
    graph->SetGraph().SetByte().SetMax(COMPRESSION); // max value to represent in graph
    graph->SetGraph().SetByte().SetAxis(0);
    graph->SetGraph().SetByte().SetValues().resize(size);
    graph->SetComp(COMPRESSION);
    m_Result.Reset(graph);

    CByte_graph::TValues& values  = graph->SetGraph().SetByte().SetValues();

    std::fill(values.begin(), values.end(), 0);

    // Fetch the data
    SAnnotSelector sel =
        CSeqUtils::GetAnnotSelector(CSeqFeatData::eSubtype_variation);

    sel.AddNamedAnnots("SNP")
       .SetExcludeExternal(false);

    CFeat_CI feat_it(*m_Scope, *m_Loc, sel);
    m_Total         = feat_it.GetSize();
    m_Completed     = 0;
    TSeqPos offset  = range.GetFrom();

    for (; feat_it && !IsCanceled(); ++feat_it) {
        const CSeq_feat &or_feat = feat_it->GetOriginalFeature();
        TSeqPos pos              = feat_it->GetLocation().GetTotalRange().GetFrom();
        const CUser_object &user = or_feat.GetExt();

        // Retrieve bitfield data
        CConstRef<CUser_field> field =
            user.GetFieldRef("dbSnpQAdata.QualityCodes");

        if (field) {
            // Filter the data
            CSnpBitfield bitfield;
            bitfield = field->GetData().GetOs();
            if (m_Filter->Passes(bitfield)) {
                // count the snp
                unsigned long cpos  = (pos-offset)/COMPRESSION;
                values[cpos]++;
            }
        }

        m_Completed++;
    }

    return IsCanceled() ? eCanceled : eCompleted;
}

END_NCBI_SCOPE

