/*  $Id: snp_filter_ui.cpp 25491 2012-03-27 17:52:21Z kuznets $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data,  the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties,  express or implied,  including
 *  warranties of performance,  merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Melvin Quintos
 */


#include <ncbi_pch.hpp>

////@begin includes
////@end includes

#include <gui/objutils/registry.hpp>
#include <gui/widgets/wx/message_box.hpp>
#include <gui/widgets/wx/wx_utils.hpp>

#include <gui/widgets/snp/filter/snp_filter_ui.hpp>
#include "rename_dlg.hpp"

#include <wx/sizer.h>
#include <wx/bitmap.h>
#include <wx/icon.h>
#include <wx/filedlg.h>

////@begin XPM images
////@end XPM images

BEGIN_NCBI_SCOPE

///////////////////////////////////////////////////////////////////////////////
// File-Scope Functions
///////////////////////////////////////////////////////////////////////////////

// Validation text
static const char *kValidCharactersComment = \
    "Valid characters are 0-9 a-z A-Z ',' '=' '(' and ')'";

// validator (must match text in kValidCharactersComment
static bool   valid_name(const string &name)
{
    CRegexp regex("^[\\w\\s(),=]+$");
    return regex.IsMatch(name);
}

static string find_unique_name(string name, const CSnpFilterUI::TMapFilters & filters)
{
    while (filters.find(name)!=filters.end()) {
        // (Check to see if there is '(number)'
        CRegexp regex("\\((\\d+)\\)$");  //   /\((\d+)\)$/  find
        if (regex.IsMatch(name)) {
            string  strNum = regex.GetSub(name,1);
            int     nNum   = NStr::StringToInt(strNum);
            string  newNum = NStr::IntToString(nNum+1);

            name = name.substr(0, name.length() - strNum.length()-2);
            name += "(" + newNum + ")";
        }
        else {
            name += " (1)";
        }
    }

    return name;
}


/*!
 * CSnpFilterUI type definition
 */

IMPLEMENT_DYNAMIC_CLASS( CSnpFilterUI, wxDialog )


/*!
 * CSnpFilterUI event table definition
 */

BEGIN_EVENT_TABLE( CSnpFilterUI, wxDialog )

////@begin CSnpFilterUI event table entries
    EVT_LISTBOX( ID_LISTBOX, CSnpFilterUI::OnListboxSelected )

    EVT_BUTTON( ID_BUTTON_ADD, CSnpFilterUI::OnButtonAddClick )

    EVT_BUTTON( ID_BUTTON_DELETE, CSnpFilterUI::OnButtonDeleteClick )

    EVT_BUTTON( ID_BUTTON_RENAME, CSnpFilterUI::OnButtonRenameClick )

    EVT_BUTTON( ID_BUTTON_COPY, CSnpFilterUI::OnButtonCopyClick )

    EVT_BUTTON( ID_BUTTON_SAVE, CSnpFilterUI::OnButtonSaveClick )

    EVT_BUTTON( ID_BUTTON_IMPORT, CSnpFilterUI::OnButtonImportClick )

    EVT_BUTTON( ID_BUTTON_EXPORT, CSnpFilterUI::OnButtonExportClick )

    EVT_BUTTON( wxID_OK, CSnpFilterUI::OnOkClick )

////@end CSnpFilterUI event table entries

END_EVENT_TABLE()


/*!
 * CSnpFilterUI constructors
 */

CSnpFilterUI::CSnpFilterUI()
{
    Init();
}

CSnpFilterUI::CSnpFilterUI( wxWindow* parent, wxWindowID id, const wxString& caption, const wxPoint& pos, const wxSize& size, long style )
{
    Init();
    CreateX(parent, id, caption, pos, size, style);
}

/*!
 * CSnpFilterUI creator
 */

bool CSnpFilterUI::CreateX( wxWindow* parent, wxWindowID id, const wxString& caption, const wxPoint& pos, const wxSize& size, long style )
{
////@begin CSnpFilterUI creation
    SetExtraStyle(wxWS_EX_BLOCK_EVENTS);
    wxDialog::Create( parent, id, caption, pos, size, style );

    CreateControls();
    if (GetSizer())
    {
        GetSizer()->SetSizeHints(this);
    }
    Centre();
////@end CSnpFilterUI creation

    x_LoadFromRegistry();
    return true;
}

void CSnpFilterUI::SetSelection(const string& name)
{    
    m_nSelectedRow = m_Browser->FindString(ToWxString(name));

    if (m_nSelectedRow != wxNOT_FOUND) {
        m_Browser->Select(m_nSelectedRow);
        x_OnBrowserChange();
    }
}

/*!
 * CSnpFilterUI destructor
 */

CSnpFilterUI::~CSnpFilterUI()
{
////@begin CSnpFilterUI destruction
////@end CSnpFilterUI destruction

    Disconnect( wxID_ANY,
        wxEVT_COMMAND_CHECKBOX_CLICKED,
        wxCommandEventHandler(CSnpFilterUI::OnPropertyChanged));    // Listen for any changes to properties
    Disconnect( wxID_ANY,
        wxEVT_COMMAND_RADIOBUTTON_SELECTED,
        wxCommandEventHandler(CSnpFilterUI::OnPropertyChanged));    // Listen for any changes to properties    
}


/*!
 * Member initialisation
 */

void CSnpFilterUI::Init()
{
////@begin CSnpFilterUI member initialisation
    m_Browser = NULL;
    m_btnDelete = NULL;
    m_btnRename = NULL;
    m_btnCopy = NULL;
    m_FilterPane = NULL;
    m_Tabs = NULL;
    m_btnSave = NULL;
////@end CSnpFilterUI member initialisation
    m_nSelectedRow = wxNOT_FOUND;
    m_bNeedsSave = false;
}

/* TODO: redo
const CSnpFilterUI::TMapConstFilters& CSnpFilterUI::GetFilters() const
{
    return sm_cmapFilters;
}
*/

/*!
 * Control creation for CSnpFilterUI
 */

void CSnpFilterUI::CreateControls()
{
////@begin CSnpFilterUI content construction
    CSnpFilterUI* itemDialog1 = this;

    wxBoxSizer* itemBoxSizer2 = new wxBoxSizer(wxVERTICAL);
    itemDialog1->SetSizer(itemBoxSizer2);

    wxBoxSizer* itemBoxSizer3 = new wxBoxSizer(wxHORIZONTAL);
    itemBoxSizer2->Add(itemBoxSizer3, 1, wxGROW|wxALL, 5);

    wxStaticBox* itemStaticBoxSizer4Static = new wxStaticBox(itemDialog1, wxID_ANY, wxT("Filters"));
    wxStaticBoxSizer* itemStaticBoxSizer4 = new wxStaticBoxSizer(itemStaticBoxSizer4Static, wxVERTICAL);
    itemBoxSizer3->Add(itemStaticBoxSizer4, 0, wxGROW|wxALL, 5);

    wxArrayString m_BrowserStrings;
    m_Browser = new wxListBox( itemDialog1, ID_LISTBOX, wxDefaultPosition, wxSize(235, 275), m_BrowserStrings, wxLB_SINGLE );
    itemStaticBoxSizer4->Add(m_Browser, 1, wxALIGN_CENTER_HORIZONTAL|wxALL, 5);

    wxGridSizer* itemGridSizer6 = new wxGridSizer(2, 3, 0, 0);
    itemStaticBoxSizer4->Add(itemGridSizer6, 0, wxALIGN_CENTER_HORIZONTAL|wxALL, 5);

    wxButton* itemButton7 = new wxButton( itemDialog1, ID_BUTTON_ADD, wxT("Add"), wxDefaultPosition, wxDefaultSize, 0 );
    itemGridSizer6->Add(itemButton7, 0, wxALIGN_CENTER_HORIZONTAL|wxALIGN_CENTER_VERTICAL|wxALL, 2);

    m_btnDelete = new wxButton( itemDialog1, ID_BUTTON_DELETE, wxT("Delete"), wxDefaultPosition, wxDefaultSize, 0 );
    itemGridSizer6->Add(m_btnDelete, 0, wxALIGN_CENTER_HORIZONTAL|wxALIGN_CENTER_VERTICAL|wxALL, 2);

    m_btnRename = new wxButton( itemDialog1, ID_BUTTON_RENAME, wxT("Rename"), wxDefaultPosition, wxDefaultSize, 0 );
    itemGridSizer6->Add(m_btnRename, 0, wxALIGN_CENTER_HORIZONTAL|wxALIGN_CENTER_VERTICAL|wxALL, 2);

    m_btnCopy = new wxButton( itemDialog1, ID_BUTTON_COPY, wxT("Copy"), wxDefaultPosition, wxDefaultSize, 0 );
    itemGridSizer6->Add(m_btnCopy, 0, wxALIGN_CENTER_HORIZONTAL|wxALIGN_CENTER_VERTICAL|wxALL, 2);

    m_FilterPane = new wxStaticBox(itemDialog1, wxID_ANY, wxT("<no filter selected>"));
    wxStaticBoxSizer* itemStaticBoxSizer11 = new wxStaticBoxSizer(m_FilterPane, wxVERTICAL);
    itemBoxSizer3->Add(itemStaticBoxSizer11, 1, wxGROW|wxALL, 5);

    m_Tabs = new wxNotebook( itemDialog1, ID_NOTEBOOK, wxDefaultPosition, wxSize(450, 306), wxBK_DEFAULT );

    itemStaticBoxSizer11->Add(m_Tabs, 1, wxGROW|wxALL, 5);

    wxBoxSizer* itemBoxSizer13 = new wxBoxSizer(wxHORIZONTAL);
    itemStaticBoxSizer11->Add(itemBoxSizer13, 0, wxGROW, 5);

    itemBoxSizer13->Add(5, 5, 1, wxALIGN_CENTER_VERTICAL|wxALL, 5);

    m_btnSave = new wxButton( itemDialog1, ID_BUTTON_SAVE, wxT("Save"), wxDefaultPosition, wxDefaultSize, 0 );
    itemBoxSizer13->Add(m_btnSave, 0, wxALIGN_CENTER_VERTICAL|wxALL, 5);

    wxBoxSizer* itemBoxSizer16 = new wxBoxSizer(wxHORIZONTAL);
    itemBoxSizer2->Add(itemBoxSizer16, 0, wxGROW, 5);

    wxBoxSizer* itemBoxSizer17 = new wxBoxSizer(wxHORIZONTAL);
    itemBoxSizer16->Add(itemBoxSizer17, 0, wxALIGN_CENTER_VERTICAL|wxLEFT|wxRIGHT, 5);

    wxButton* itemButton18 = new wxButton( itemDialog1, ID_BUTTON_IMPORT, wxT("Import"), wxDefaultPosition, wxDefaultSize, 0 );
    itemBoxSizer17->Add(itemButton18, 0, wxALIGN_CENTER_VERTICAL|wxALL, 3);

    wxButton* itemButton19 = new wxButton( itemDialog1, ID_BUTTON_EXPORT, wxT("Export"), wxDefaultPosition, wxDefaultSize, 0 );
    itemBoxSizer17->Add(itemButton19, 0, wxALIGN_CENTER_VERTICAL|wxALL, 3);

    wxBoxSizer* itemBoxSizer20 = new wxBoxSizer(wxHORIZONTAL);
    itemBoxSizer16->Add(itemBoxSizer20, 1, wxALIGN_CENTER_VERTICAL|wxLEFT|wxRIGHT, 5);

    itemBoxSizer20->Add(0, 5, 1, wxALIGN_CENTER_VERTICAL|wxALL, 0);

    wxButton* itemButton22 = new wxButton( itemDialog1, wxID_OK, wxT("&OK"), wxDefaultPosition, wxDefaultSize, 0 );
    itemButton22->SetDefault();
    itemBoxSizer20->Add(itemButton22, 0, wxALIGN_CENTER_VERTICAL|wxALL, 5);

////@end CSnpFilterUI content construction

    x_SetupTabs();

    Connect( wxID_ANY,
        wxEVT_COMMAND_CHECKBOX_CLICKED,
        wxCommandEventHandler(CSnpFilterUI::OnPropertyChanged));    // Listen for any changes to properties
    Connect( wxID_ANY,
        wxEVT_COMMAND_RADIOBUTTON_SELECTED,
        wxCommandEventHandler(CSnpFilterUI::OnPropertyChanged));    // Listen for any changes to properties

    x_EnableCopyDelete(false);
    x_EnableSave(false);
}


/*!
 * Should we show tooltips?
 */
/*
bool CSnpFilterUI::ShowToolTips()
{
    return true;
}
*/

/*!
 * Get bitmap resources
 */

wxBitmap CSnpFilterUI::GetBitmapResource( const wxString& name )
{
    // Bitmap retrieval
////@begin CSnpFilterUI bitmap retrieval
    wxUnusedVar(name);
    return wxNullBitmap;
////@end CSnpFilterUI bitmap retrieval
}

/*!
 * Get icon resources
 */

wxIcon CSnpFilterUI::GetIconResource( const wxString& name )
{
    // Icon retrieval
////@begin CSnpFilterUI icon retrieval
    wxUnusedVar(name);
    return wxNullIcon;
////@end CSnpFilterUI icon retrieval
}

/*!
 * wxEVT_COMMAND_BUTTON_CLICKED event handler for ID_BUTTON_IMPORT
 */

void CSnpFilterUI::OnButtonImportClick( wxCommandEvent& /*event*/ )
{
    x_CheckChanges();

    // Open file
    // for every line in the file
    //   split the name and serialized filter
    //   create a new filter and serialize from
    //   add to browser and filter map
    // close file


    SWFileDlgData data(wxT("Import Filters"), wxT("SNP Filters (*.ftr)|*.ftr"), wxFD_OPEN);

    if (NcbiFileBrowser(data) == wxID_OK) {
        wxString filename = data.GetFilename();
        ifstream inFile;
        string   line, name, strFilter;

        // clear old entries
        x_Clear();

        // open file
        inFile.open(filename.fn_str());

        while ( !inFile.eof() ) {
            getline(inFile, line);
            NStr::TruncateSpacesInPlace(line);

            if (line.empty())
                continue;

            SSnpFilter filter;
            filter.SerializeFrom(line);

            // avoid name conflicts
            name = find_unique_name(filter.name, m_mapFilters);
            filter.name = name;

            x_AddFilter(name, new SSnpFilter(filter));
            x_AddSorted(name);
        }

        inFile.close();
    }
}

void CSnpFilterUI::OnPropertyChanged( wxCommandEvent & /*event*/)
{
    m_bNeedsSave = true;
    x_EnableSave(true);
}

/*!
 * wxEVT_COMMAND_BUTTON_CLICKED event handler for ID_BUTTON_ADD
 */

void CSnpFilterUI::OnButtonAddClick( wxCommandEvent& event )
{
    x_CheckChanges();

    SSnpFilter* filter;
    filter = new SSnpFilter();
    filter->name = find_unique_name("new filter", m_mapFilters);

    x_AddFilter(filter->name, filter);  // add to internal maps
    x_AddSorted(filter->name);          // add to listbox
    x_OnBrowserChange();

    OnButtonRenameClick(event);
}

void CSnpFilterUI::x_AddSorted(const string &text)
{
    // check each line and insert into correct spot
    unsigned int i=0;
    for (; i < m_Browser->GetCount(); i++) {
        string line = ToStdString(m_Browser->GetString(i));
        if ( NStr::CompareNocase(text, line) < 0)
            break;
    }

    m_Browser->Insert(ToWxString(text), i);
    m_Browser->Select(i);
    m_nSelectedRow = i;
}


void CSnpFilterUI::x_LoadFromRegistry()
{
    m_mapFilters.clear();

    // Read filters from registry
    CGuiRegistry& reg = CGuiRegistry::GetInstance();
    CGuiRegistry::TReadWriteView view = reg.GetReadWriteView("GBENCH.Plugins.View.Snp");

    typedef list<string> TListStrings;
    TListStrings filters;
    view.GetStringList("filters", filters);

    CRef<SSnpFilter> filter; // a temporary filter object
    
    ITERATE(TListStrings, iter, filters) {
        filter.Reset(new SSnpFilter(*iter));
        m_mapFilters[filter->name] = filter;
        m_Browser->Append(ToWxString(filter->name));
    }
}

/*!
 * wxEVT_COMMAND_BUTTON_CLICKED event handler for ID_BUTTON_EXPORT
 */

void CSnpFilterUI::OnButtonExportClick( wxCommandEvent& /*event*/ )
{
    x_CheckChanges();

    // for every filter in map
    //   call serialize to
    //   write a new line to the file
    // close file

    SWFileDlgData data(wxT("Export Filters"), wxT("SNP Filters (*.ftr)|*.ftr"), wxFD_SAVE);

    if (NcbiFileBrowser(data) == wxID_OK) {
        wxString filename = data.GetFilename();
        ofstream oFile;
        oFile.open(filename.fn_str());

        NON_CONST_ITERATE(TMapFilters, iter, m_mapFilters) {
            string strFilter;
            SSnpFilter &filter  = *iter->second;
            filter.SerializeTo(strFilter);
            oFile << strFilter << endl;
        }
        oFile.close();
    }
}

void CSnpFilterUI::x_AddFilter(const string& name, SSnpFilter* filter)
{
    m_mapFilters[name].Reset(filter);
}

void CSnpFilterUI::x_RemoveFilter(const string& name)
{
    m_mapFilters.erase(name);
}

void CSnpFilterUI::x_ResetFilterTab()
{
    m_GrpChecksLinks     ->Reset();
    m_GrpChecksMapping   ->Reset();
    m_GrpChecksAlleleFreq->Reset();
    m_GrpChecksGaP       ->Reset();
    m_GrpChecksQuality   ->Reset();
    m_GrpChecksGeneFxn   ->Reset();
    m_GrpRadiosWeight    ->Reset();
    m_GrpRadiosAlleleFreq->Reset();
    m_GrpRadiosVarClass  ->Reset();
}


void CSnpFilterUI::x_OnBrowserChange()
{
    x_CheckChanges();
    m_nSelectedRow = m_Browser->GetSelection();

    if (m_nSelectedRow!=wxNOT_FOUND) {
        x_EnableCopyDelete(true);
        x_LoadFilter();
    }
    else {
        x_EnableCopyDelete(false);
    }
}

void CSnpFilterUI::x_CheckChanges()
{
    if (m_bNeedsSave) {
        _ASSERT(m_nSelectedRow!=wxNOT_FOUND);

        // Get
        string text = ToStdString(m_Browser->GetString(m_nSelectedRow));
        text        = "Save changes to \"" + text + "\" ?";

        EDialogReturnValue ret;

        ret = NcbiMessageBox(text.c_str(),
                        eDialog_YesNo | eDialog_Modal | eDialog_Centered,
                        eIcon_Question,
                        "Filter was modified");

        if (ret == eYes) {
            wxCommandEvent evt;
            OnButtonSaveClick(evt);
        }
        else {
            m_bNeedsSave = false;
        }

        x_EnableSave(false);
    }
}

void CSnpFilterUI::x_EnableCopyDelete(bool enable)
{
    if (enable) {
        m_Tabs->Enable();
        m_btnRename->Enable();
        m_btnDelete->Enable();
        m_btnCopy->Enable();
    }
    else {
        m_Tabs->Disable();
        m_btnRename->Disable();
        m_btnDelete->Disable();
        m_btnCopy->Disable();
    }
}

void CSnpFilterUI::x_EnableSave(bool enable)
{
    if (enable) {
        m_btnSave->Enable();
    }
    else {
        m_btnSave->Disable();
    }
}

/*!
 * wxEVT_COMMAND_LISTBOX_SELECTED event handler for ID_LISTBOX
 */

void CSnpFilterUI::OnListboxSelected( wxCommandEvent& /*event*/ )
{
    x_OnBrowserChange();
}

/*!
 * wxEVT_COMMAND_BUTTON_CLICKED event handler for ID_BUTTON_DELETE
 */

void CSnpFilterUI::OnButtonDeleteClick( wxCommandEvent& /*event*/ )
{
    if (m_nSelectedRow!=wxNOT_FOUND) {
        string n = ToStdString(m_Browser->GetString(m_nSelectedRow));

        x_RemoveFilter(n);
        m_Browser->Delete(m_nSelectedRow);
        m_FilterPane->SetLabel(wxT("<no filter selected>"));

        x_EnableCopyDelete(false);
        m_bNeedsSave = false;
    }
}


/*!
 * wxEVT_COMMAND_BUTTON_CLICKED event handler for ID_BUTTON_RENAME
 */

void CSnpFilterUI::OnButtonRenameClick( wxCommandEvent& /*event*/ )
{
    x_CheckChanges();

    // Get text of row
    if (m_nSelectedRow!=wxNOT_FOUND) {
        string name         = ToStdString(m_Browser->GetString(m_nSelectedRow));
        string oldname      = name;
        bool   is_new_name  = false;

        CRenameDlg dlg;
        dlg.CreateX(ToWxString(name), this);

        while ( is_new_name == false && dlg.ShowModal() == wxID_OK ) {
            name = ToStdString(dlg.Text->GetValue());

            // check if name is valid
            if (valid_name(name) == false) {
                NcbiMessageBox(kValidCharactersComment,
                                eDialog_Ok | eDialog_Modal | eDialog_Centered,
                                eIcon_Exclamation,
                                "Invalid name");
                continue;
            }

            // check if unique
            if ( m_mapFilters.find(name) == m_mapFilters.end()) {
                is_new_name = true;

                // change the name
                SSnpFilter &filter = *m_mapFilters[oldname];
                filter.name = name;

                x_AddFilter(name, &filter);
                x_RemoveFilter(oldname);
                m_Browser->Delete(m_nSelectedRow);
                x_AddSorted(name);
                x_LoadFilter();
            }
            else {
                // name is not unique.  ask question again
                NcbiMessageBox("Name \"" + name + "\" already in use");
                continue;
            }
        }
    }
}


/*!
 * wxEVT_COMMAND_BUTTON_CLICKED event handler for ID_BUTTON_COPY
 */

void CSnpFilterUI::OnButtonCopyClick( wxCommandEvent& /*event*/ )
{
    x_CheckChanges();

    if (m_nSelectedRow!=wxNOT_FOUND) {
        string name         = ToStdString(m_Browser->GetString(m_nSelectedRow));
        const SSnpFilter &f = *m_mapFilters[name];

        SSnpFilter* new_filter;
        new_filter          = new SSnpFilter(f);
        new_filter->name    = find_unique_name("Copy of "+name, m_mapFilters);

        x_AddFilter(new_filter->name, new_filter);
        x_AddSorted(new_filter->name);
        x_OnBrowserChange();
    }
}


/*!
 * wxEVT_COMMAND_BUTTON_CLICKED event handler for wxID_OK
 */

void CSnpFilterUI::OnOkClick( wxCommandEvent& event )
{
    x_CheckChanges();

    CGuiRegistry& reg = CGuiRegistry::GetInstance();
    CGuiRegistry::TReadWriteView view = reg.GetReadWriteView("GBENCH.Plugins.View.Snp");

    list<string>  filterList;
    ITERATE(TMapFilters, iter, m_mapFilters) {
        string      name    = iter->first;
        const SSnpFilter &filter  = *iter->second;

        string strFilter;
        filter.SerializeTo(strFilter);

        filterList.push_back(strFilter);
    }

    view.Set("filters", filterList);

    event.Skip(); // call the rest of the handling functions
}

int CSnpFilterUI::GetSelectedRow() const
{
    return m_Browser->GetSelection();
}

/*!
 * wxEVT_COMMAND_BUTTON_CLICKED event handler for ID_BUTTON_SAVE
 */

void CSnpFilterUI::OnButtonSaveClick( wxCommandEvent& /*event*/ )
{
    x_OnBtnSave();
}

END_NCBI_SCOPE



