/*  $Id: gap_ds.cpp 26643 2012-10-18 17:54:05Z katargir $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Melvin Quintos, Dmitry Rudnev
 *
 * Description:  This file defines CGapJob, CSGGapDS, and CSGGapDSType
 *
 */

#include <ncbi_pch.hpp>
#include <gui/packages/pkg_snp/gap/gap_ds.hpp>
#include <gui/packages/pkg_snp/gap/gwas_glyph.hpp>
#include <gui/packages/pkg_snp/bins/bins_glyph.hpp>
#include <gui/packages/pkg_snp/utils/utils.hpp>

#include <gui/objutils/utils.hpp>
#include <gui/utils/app_job_dispatcher.hpp>
#include <gui/widgets/seq_graphic/annot_conf.hpp>

#include <objmgr/graph_ci.hpp>
#include <objmgr/seq_loc_mapper.hpp>
#include <objmgr/seq_map_switch.hpp>
#include <objmgr/seqdesc_ci.hpp>
#include <objmgr/table_field.hpp>
#include <objmgr/util/feature.hpp>
#include <objmgr/util/sequence.hpp>
#include <objmgr/util/sequence.hpp>

#include <objects/general/Object_id.hpp>
#include <objects/seq/Annot_descr.hpp>
#include <objects/seq/Annotdesc.hpp>
#include <objects/seqres/Byte_graph.hpp>
#include <objects/seqtable/Seq_table.hpp>
#include <objects/seqtable/SeqTable_column.hpp>

#include <cmath>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

///////////////////////////////////////////////////////////////////////////////
///   CGapJob
///////////////////////////////////////////////////////////////////////////////
class CGapJob : public CSGAnnotJob
{
public:
    struct SParams {
        bool        bPreferFeat;  ///< Only return feat glyphs when not overview
        bool        isSeqView;
        bool        bAdaptiveSelector;
        TModelUnit  scale;
        int         binWidth;
        int         depth;
        CSGGapDS::SAnnotInfo annotinfo;
        string      annot;
    };

    CGapJob(const CBioseq_Handle& handle, const TSeqRange& range, 
            const SParams& params)
        : CSGAnnotJob("SNP", handle, SAnnotSelector(), range)
        , m_Params(params) {}

    virtual EJobState   x_Execute();

private:
    //bool    x_IsOverview() const; 
    void x_Load(CSeqGlyph::TObjects*);
    // returns false if loading failed/GWAS table was not found
    void x_LoadGWAS(CSeqGlyph::TObjects*);
    void x_LoadCHPA(CSeqGlyph::TObjects*, const CSeq_graph& gr_chpa);

private:
    SParams                     m_Params;
};

IAppJob::EJobState CGapJob::x_Execute()
{
    EJobState state = eCompleted;

    try {    
        CSeqGlyph::TObjects glyphs;
        
        x_Load(&glyphs);

        SGapJobResult *result = new SGapJobResult();
        m_Result.Reset(result);  

        result->listObjs.swap(glyphs);
    } catch (CException& ex) {
        m_Error.Reset(new CAppJobError(ex.GetMsg()));
        state = eFailed;
    } catch (std::exception& ex) {
        m_Error.Reset(new CAppJobError(ex.what()));
        state = eFailed;
    }

    return state;
}

void CGapJob::x_Load(CSeqGlyph::TObjects *objs)
{
    
    try {  
        bool    isCHPAFound(false);
        bool    isGWASFound(false);

        CConstRef<CSeq_graph> gr_chpa;
    
        //  CHECK CHPA graphs
        if(m_Params.annotinfo.m_Type == CSGGapDS::eGraph) {            
            SAnnotSelector sel;
            
            sel.SetAnnotType(CSeq_annot::TData::e_Graph);
            sel.ExcludeNamedAnnots("SNP");
            sel.IncludedAnnotName("GWAS");
            
            CRef<CSeq_loc> loc = m_Handle.GetRangeSeq_loc(m_Range.GetFrom(), m_Range.GetTo());
        
            CGraph_CI graph_iter(m_Handle.GetScope(), *loc, sel);

            for (;  graph_iter ;  ++graph_iter) {
                if (IsCanceled())
                    return;

                const CMappedGraph& graph = *graph_iter;
                
                if (  (graph.GetAnnot().IsNamed()) && 
                      (graph.GetAnnot().GetName() == m_Params.annot)  ) {
                    gr_chpa.Reset(&graph.GetOriginalGraph());                
                    isCHPAFound = true;
                    break;
                }
            }
            if(isCHPAFound)
                x_LoadCHPA(objs, *gr_chpa);
        } else {
            isGWASFound = true;
            x_LoadGWAS(objs);
        }

        if ( !isGWASFound  &&  !isCHPAFound )  {
            throw "Did not find expected data";
        }
    } catch (CException& e) {
        m_Error.Reset(new CAppJobError(e.GetMsg()));
        //state = eFailed;
    } catch (std::exception&) {
        m_Error.Reset(new CAppJobError("Exception when loading Named Annotations"));
        //state = eFailed;
    }
}

void CGapJob::x_LoadGWAS(CSeqGlyph::TObjects *objs)
{
    double  scale       = fabs(m_Params.scale);        
    const int binWidth  = m_Params.binWidth;

    CRef<CSeq_loc> loc = m_Handle.GetRangeSeq_loc(m_Range.GetFrom(), m_Range.GetTo());   

    float window = (float)scale*binWidth;

    if (window < 1 || m_Params.isSeqView) {
        window = 1;
    }

    // TODO.  Do different logic based on the type of data  (CHPA or Other)
    CGwasGlyph::TDensityMap   dmap(
        m_Range.GetFrom(), 
        m_Range.GetTo(),
        window,
        new CGwasGlyph::accum_study());
    
    // fetch the seq-table and populate the density map
    // Grab seq-graph data
    SAnnotSelector sel =
        CSeqUtils::GetAnnotSelector(CSeq_annot::TData::e_Seq_table);
    if (CSeqUtils::IsNAA(m_Params.annot)) {
        sel.IncludeNamedAnnotAccession(m_Params.annot);
    }
    sel.AddNamedAnnots(m_Params.annot);
    CSeqUtils::SetResolveDepth(sel, m_Params.bAdaptiveSelector, m_Params.depth);

    // grab the columns for position and value
    CTableFieldHandle<int>      col_pos("pos");
    CTableFieldHandle<double>   col_val("pvalue");
    CTableFieldHandle<string>   col_type("trackType");
    int     pos;
    double  pvalue;
    int     pos_start, pos_end;

    CAnnot_CI it(m_Handle.GetScope(), *loc, sel);

    bool isFound(false);
    /* //!! this check is nice, but would miss the older GWAS tracks that did not have this column
    for (; it; ++it) {
        CSeq_annot_Handle annot = *it;

        try {
            if (col_type.TryGet(annot, 0, type) && type == CBinsGlyph::eGCAT) { */
                isFound = true;
    /*            break;
            }
        }
        catch (std::exception&) {
            continue;
        }
    }*/

    if(!isFound) return;
    if (!it.empty()) {
        CSeq_annot_Handle annot = *it;
        NSnpUtils::FindRange(annot, (int)m_Range.GetFrom(), (int)m_Range.GetTo(), &pos_start, &pos_end);

        // try and get position
        for ( int row = pos_start; row < pos_end; ++row ) {
            if ( col_pos.TryGet(annot, row, pos) &&
                 col_val.TryGet(annot, row, pvalue) ) {

                TSeqRange    range  = TSeqRange(pos, pos);

                CGwasGlyph::SValue  s;
                s.count  = 1;
                s.pvalue = (float) -log10(pvalue);
                s.rsid   = 0;

                dmap.AddRange(range, s);
            }
        }
    }
    CRef<CGwasGlyph>  gwas;
    gwas.Reset(new CGwasGlyph(dmap, *loc, &m_Handle.GetScope()));
    gwas->SetTitle(m_Params.annotinfo.m_Title);
    gwas->SetAnnotName(m_Params.annot);
    CRef<CSeqGlyph> fref(gwas.GetPointer());
    objs->push_back(fref);
}

void CGapJob::x_LoadCHPA(CSeqGlyph::TObjects* objs, const CSeq_graph& gr)
{    
    bool    fetchTable  = false;
    bool    hasResult = false;

    double  scale       = fabs(m_Params.scale);        
    const int binWidth  = m_Params.binWidth;

    if (gr.GetComp() > scale) {
        fetchTable = true;
    }

    // create new densitymap
    float window = (float)scale*binWidth;

    if (window < 1) {
        window = 1;
    }

    CGwasGlyph::TDensityMap   dmap(
        m_Range.GetFrom(), 
        m_Range.GetTo(),
        window,
        new CGwasGlyph::accum_study());

    // If we don't fetch the table, then use the graph data to create the density map
    if (fetchTable == false) {
        TSeqRange r    = gr.GetLoc().GetTotalRange();
        TSeqPos   comp = gr.GetComp();
        TSeqPos   pos  = r.GetFrom();

        CGwasGlyph::SValue s;
        ITERATE (CByte_graph::TValues, iter_gr, gr.GetGraph().GetByte().GetValues()) {
            TSeqRange    range  = TSeqRange(pos, pos + comp);
            char         cvalue = char(*iter_gr);
            unsigned int pvalue = cvalue & 0x0F;
            int          count  = (((unsigned int)cvalue) >> 4) & 0x0F;

            s.count  = count;
            s.pvalue = (float)pvalue;
            s.rsid   = 0;

            dmap.AddRange(range, s);

            pos += comp;
            hasResult = true;
        }
    } 
    else {
        // Grab seq-ftable data        
        CRef<CSeq_loc> loc = m_Handle.GetRangeSeq_loc(m_Range.GetFrom(), m_Range.GetTo());

        SAnnotSelector sel;

        sel = CSeqUtils::GetAnnotSelector(CSeqFeatData::eSubtype_user);        
        sel.AddNamedAnnots(m_Params.annot)
           .SetExcludeExternal(false)
           .SetExactDepth(true)
           .SetResolveDepth(0);
        
        CFeat_CI study_it(m_Handle.GetScope(), *loc, sel);
        
        bool isFirst = true;
        string title;
        for ( ; study_it; ++study_it) {
            const CSeq_feat& or_feat = study_it->GetOriginalFeature();
            const CUser_object &user = or_feat.GetData().GetUser();

            if (isFirst) {
                // Extract the Annotations properties
                ITERATE( CAnnot_descr::Tdata, it, study_it->GetAnnot().GetCompleteSeq_annot()->GetDesc().Get()) {
                    const CAnnotdesc &desc = **it;
                    if (desc.IsUser() && desc.GetUser().GetType().IsStr() &&
                        desc.IsUser() && desc.GetUser().GetType().GetStr() == "AssnDescr") {
                        title = desc.GetUser().GetField("title").GetData().GetStr();
                        break;
                    }
                }
                isFirst = false;
            }

            string name   = title;
            double pvalue = user.GetField("value").GetData().GetReal();
            string  id    = user.GetField("id").GetData().GetStr().substr(2);
            int rsid      = NStr::StringToInt(id);

            
            CGwasGlyph::SValue s;
            s.count  = 1;
            s.pvalue = (float) -log10(pvalue);
            s.rsid   = rsid;

            TSeqPos     pos = or_feat.GetLocation().GetPnt().GetPoint();
            TSeqRange range = TSeqRange(pos, pos);

            dmap.AddRange(range, s);
            hasResult = true;
        }
    }
    
    if (hasResult) {       
        CRef<CSeq_loc> loc = m_Handle.GetRangeSeq_loc(m_Range.GetFrom(), m_Range.GetTo());
     
        CRef<CGwasGlyph>  gwas;
        gwas.Reset(new CGwasGlyph(dmap, *loc, &m_Handle.GetScope()));
        gwas->SetTitle(m_Params.annotinfo.m_Title);
        gwas->SetAnnotName(m_Params.annot);
        CRef<CSeqGlyph> fref(gwas.GetPointer());
        objs->push_back(fref);
    }

}


///////////////////////////////////////////////////////////////////////////////
/// CSGGapDS
///////////////////////////////////////////////////////////////////////////////
CSGGapDS::CSGGapDS(CScope& scope, const CSeq_id& id)
    : CSGGenBankDS(scope, id)
{}

void CSGGapDS::GetTrackNames(SConstScopedObject& object,
                             TAnnotInfoMap& annots,
                             const TSeqRange& range,
                             const ILayoutTrackFactory::SExtraParams& params)
{    
    const CSeq_id& id = dynamic_cast<const CSeq_id&>(object.object.GetObject());
    CBioseq_Handle handle = object.scope->GetBioseqHandle(id);
    CConstRef<CSeq_loc> loc(handle.GetRangeSeq_loc(range.GetFrom(), range.GetTo()));   

    // Find NHGRI Catalog
    {{
        ITERATE(ILayoutTrackFactory::SExtraParams::TAnnots, iAnnots, params.m_Annots) {
            SAnnotSelector sel =
                CSeqUtils::GetAnnotSelector(CSeq_annot::TData::e_Seq_table);
            if (CSeqUtils::IsNAA(*iAnnots)) {
                sel.IncludeNamedAnnotAccession(*iAnnots);
            }
            sel.AddNamedAnnots(*iAnnots);
            CSeqUtils::SetResolveDepth(sel, params.m_Adaptive, params.m_Level);

            CAnnot_CI table_iter(handle.GetScope(), *loc, sel);
            for(; table_iter ; ++table_iter) {
                CConstRef<CSeq_annot> seq_annot(CSeq_annot_Handle(*table_iter).GetCompleteSeq_annot());
                string sTitle;
                if(seq_annot->CanGetDesc()) {
                    const CAnnot_descr& descr(seq_annot->GetDesc());

                    ITERATE(CAnnot_descr::Tdata, iDesc, descr.Get())
                    {
                        if((*iDesc)->IsTitle()) {
                            sTitle = (*iDesc)->GetTitle();
                            break;
                        }
                    }
                }
                // sanity checks: a suitable table must have a pvalue column and also track type of eGCAT
                CSeq_table::TColumns columns(seq_annot->GetData().GetSeq_table().GetColumns());
                bool hasPvalue(false);
                bool isGWAS(false);

                ITERATE(CSeq_table::TColumns, iColumns, columns) {
                    if((*iColumns)->GetHeader().GetField_name() == "pvalue")
                        hasPvalue = true;
                    
                    /* //!! this check is nice, but would miss the older GWAS tracks that did not have this column
                    if((*iColumns)->GetHeader().GetField_name() == "trackType") {
                        CTableFieldHandle<int> col_type("trackType");
                        int type;
                        if (col_type.TryGet(*table_iter, 0, type) && type == CBinsGlyph::eGCAT) */
                            isGWAS = true;
                    // }
                }
                if(hasPvalue && isGWAS)
                    annots[*iAnnots] = SAnnotInfo(eSeqTable, sTitle);
            }
        }
    }}

    // Find the other GWAS Tracks that can be loaded from CHPA files
    //   Only need to check Top Sequence
    {{
    SAnnotSelector sel;
    
    sel.SetAnnotType(CSeq_annot::TData::e_Graph);
    sel.ExcludeNamedAnnots("SNP");
    //sel.IncludedAnnotName("GWAS_");
    sel.SetCollectNames();
    
    CGraph_CI graph_iter(handle, sel);

    ITERATE (CGraph_CI::TAnnotNames, iter, graph_iter.GetAnnotNames()) {
        /// From Lianshou Wu:  DIMA, please fix this.  The fix I made is just an emergency fix.
        /// Problem #1: It crashes inside GetOriginalGraph()
        ///             To reproduce, you could try loading sequence gnl|TRACE_ASSM|992.
        ///             You might not be able to get the graph here since the selector 
        ///             is set to collecting names.
        /// Problem #2: The same title gets mapped to different annotations
        /*
        if (iter->IsNamed()) {
            annots[iter->GetName()] = TAnnotInfo(eGraph, graph_iter->GetOriginalGraph().GetTitle());
        } else {
            annots["Unnamed"] = SAnnotInfo(eGraph, graph_iter->GetOriginalGraph().GetTitle());
        }
        */
        if (iter->IsNamed()  &&  NStr::StartsWith(iter->GetName(), "GWAS_")) {
            annots[iter->GetName()] = SAnnotInfo(eGraph, iter->GetName());
        }
    }

    }}
}


void CSGGapDS::SetAnnotInfo(const TAnnotInfo& annotinfo) 
{
    m_AnnotInfo = annotinfo;
}

void CSGGapDS::SetAnnot(const string& annot) 
{
    m_Annot = annot;
}

void CSGGapDS::LoadData(const TSeqRange& range, 
                        TModelUnit scale, 
                        bool preferFeats,
                        bool isSeqView)
{
    CRef<CGapJob> job;
    
    CGapJob::SParams p;

    p.bPreferFeat = preferFeats;
    p.isSeqView   = isSeqView;
    p.bAdaptiveSelector = m_Adaptive;
    p.scale       = scale;
    p.binWidth    = 14; // TODO Should read from global config file.
    p.annotinfo   = m_AnnotInfo;
    p.depth       = m_Depth;
    p.annot       = m_Annot;

    job.Reset(new CGapJob(m_Handle, range, p));

    x_LaunchJob(*job);
}

///////////////////////////////////////////////////////////////////////////////
/// CSGGapDSType
///////////////////////////////////////////////////////////////////////////////

ISGDataSource*
CSGGapDSType::CreateDS(SConstScopedObject& object) const
{
    const CSeq_id& id = dynamic_cast<const CSeq_id&>(object.object.GetObject());
    return new CSGGapDS(object.scope.GetObject(), id);
}


string CSGGapDSType::GetExtensionIdentifier() const
{
    static string sid("seqgraphic_gap_ds_type");
    return sid;
}

string CSGGapDSType::GetExtensionLabel() const
{
    static string slabel("Graphical View SNP Data Source Type");
    return slabel;
}

bool CSGGapDSType::IsSharable() const
{
    return false;
}


END_NCBI_SCOPE
