/*  $Id: gap_track.cpp 27877 2013-04-17 18:49:16Z wuliangs $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Melvin Quintos
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>

#include <gui/packages/pkg_snp/gap/gap_track.hpp>
#include <gui/packages/pkg_snp/gap/gwas_glyph.hpp>
#include <gui/widgets/seq_graphic/seqgraphic_utils.hpp>
#include <gui/widgets/seq_graphic/simple_layout_policy.hpp>
#include <gui/widgets/seq_graphic/seqgraphic_ds_manager.hpp>
#include <gui/widgets/seq_graphic/layered_layout_policy.hpp>
#include <gui/widgets/seq_graphic/rendering_ctx.hpp>
#include <gui/widgets/seq_graphic/histogram_glyph.hpp>

#include <gui/utils/rgba_color.hpp>
#include <gui/objutils/utils.hpp>
#include <gui/objutils/label.hpp>

#include <objmgr/util/feature.hpp>
#include <objmgr/util/sequence.hpp>
#include <objmgr/annot_ci.hpp>

#include <objects/seqfeat/SeqFeatData.hpp>

#include <wx/menu.h>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

///////////////////////////////////////////////////////////////////////////////
///   CGapTrack
///////////////////////////////////////////////////////////////////////////////

CTrackTypeInfo CGapTrack::m_TypeInfo("GWAS_track",
                                     "SNP in NHGRI Catalog of Published \
Genome-Wide Association Studies. Catalog is manually curated from publications. ");

CGapTrack::CGapTrack(CSGGapDS* ds, CRenderingContext* r_cntx)
    : CDataTrack(r_cntx)
    , m_DS(ds)
    , m_IconLayout(eIcon_Layout, "Layout", true, "track_layout")
{
    m_DS->SetJobListener(this);
    m_FeatSel = CSeqUtils::GetAnnotSelector();
}


CGapTrack::~CGapTrack()
{
    if ( !m_DS->AllJobsFinished() ) {
        m_DS->DeleteAllJobs();
    }
}

/// @name CLayoutTrack reimplemented methods.
/// @{
const CTrackTypeInfo& CGapTrack::GetTypeInfo() const
{
    return m_TypeInfo;
}

string CGapTrack::GetFullTitle() const
{
    const string& title = GetTitle();
    if (title.empty()) {
        return m_AnnotName;
    }
    return title;
}


void CGapTrack::SetAnnot(const string& name)
{
    m_AnnotName = name;
}

void CGapTrack::GetHTMLActiveAreas(TAreaVector* p_areas) const
{
    if (GetChildren().size() > 1) {
        // use the standard way
        CGlyphContainer::GetHTMLActiveAreas(p_areas);
    }
}


CHTMLActiveArea* CGapTrack::InitHTMLActiveArea(TAreaVector* p_areas) const
{
    if (GetChildren().size() == 1) {
        TModelUnit tb_height = x_GetTBHeight();
        if (tb_height > 0.0) {
            // generate areas for bins
            GetChildren().front()->GetHTMLActiveAreas(p_areas);

            // generate a separate area for the track caption
            CHTMLActiveArea area;
            CSeqGlyph::x_InitHTMLActiveArea(area);

            // adjust the bounds height to that of the caption only
            TVPRect& bound(area.m_Bounds);
            bound.SetBottom(bound.Top() + 2 + (TVPUnit)tb_height);

            // find out the gi
            int gi(sequence::GetGiForId(*(sequence::GetId(m_DS->GetBioseqHandle(), sequence::eGetId_ForceGi).GetSeqId()),
                                        m_DS->GetBioseqHandle().GetScope()));

            // create the signature
            string sSignature(CGwasGlyph::GenerateBinSignature("",
                                                               m_AnnotName,
                                                               gi,
                                                               area.m_SeqRange));
            area.m_Signature = sSignature;
            area.m_Flags = CHTMLActiveArea::fNoSelection | CHTMLActiveArea::fNoPin | CHTMLActiveArea::fTrack;
            p_areas->push_back(area);
            return &(p_areas->back());
        }

        return NULL;
    }

    return CLayoutTrack::InitHTMLActiveArea(p_areas);
}


void CGapTrack::x_OnIconClicked(TIconID id)
{
    bool bIsLayoutOnlyChange = true;

    if (id == eIcon_Layout) {
        wxMenu menu;

        //menu.AppendRadioItem(eLayout_Default,   wxT("Default"));
        menu.AppendRadioItem(eLayout_Labels,    wxT("Labels"));

        menu.Check(m_eLayout, true);

        m_LTHost->LTH_PopupMenu(&menu);

        wxMenuItemList& item_list = menu.GetMenuItems();
        ITERATE (wxMenuItemList, iter, item_list) {
            ELayout new_layout = (ELayout)((*iter)->GetId());
            if ((*iter)->IsChecked()  &&  id != m_eLayout) {

                // Reload data if switching back or forth to Labels
                if ( (m_eLayout ==eLayout_Labels) ||
                     (new_layout==eLayout_Labels) )
                {
                    bIsLayoutOnlyChange = false;
                }

                m_eLayout = new_layout;
                break; // stop processing loop
            }
        }
    } else {
        // use default handlers
        CLayoutTrack::x_OnIconClicked(id);
    }

    Update(bIsLayoutOnlyChange);
    x_OnLayoutChanged();
}
/// @}

void CGapTrack::x_LoadSettings(const string& /*preset_style*/,
                               const TKeyValuePairs& /*settings*/)
{
    m_Layered->SetVertSpace(m_gConfig->GetObjectSpace());
}

/// @}

bool CGapTrack::x_IsOverviewMode() const
{
    static double kNPP_Cutoff= 50;  // Nucliotides per pixel cutoff for SNP Overview
    return (m_Context->GetScale() > kNPP_Cutoff);
}

/// @name CDataTrack pure virtual interfaces
/// @{

void CGapTrack::x_UpdateData()
{
    bool bShowLabels = (m_eLayout == eLayout_Labels);
    m_DS->DeleteAllJobs();
    x_SetStartStatus();

    m_DS->LoadData(m_Context->GetVisSeqRange(),
                   m_Context->GetScale(),
                   bShowLabels,   // fetch features if we show labels.
                   m_Context->WillSeqLetterFit());
}


void CGapTrack::x_OnJobCompleted(CAppJobNotification& notify)
{
    m_DS->ClearJobID(notify.GetJobID());
    CRef<CObject> res_obj = notify.GetResult();
    SGapJobResult* result = dynamic_cast<SGapJobResult*>(&*res_obj);
    if (!result) {
        ERR_POST("CGapTrack::x_OnJobCompleted() notification for job \
                does not contain results.");
        return;
    }

    // Depending on view, setup icon
    if (x_IsOverviewMode())
        x_DeregisterIcon(m_IconLayout.m_Id);
    else
        x_RegisterIcon(m_IconLayout);

    SetMsg("");
    x_CreateLayout(*result);
}
/// @}

void CGapTrack::x_CreateLayout(const SGapJobResult& result)
{
    const CSeqGlyph::TObjects& objs = result.listObjs;

    m_Group.Clear();

    if (objs.empty()) {
        Update(true);
        x_OnLayoutChanged();
        return;
    }

    ITERATE(CSeqGlyph::TObjects, it, objs) {
        CSeqGlyph *glyph = const_cast<CSeqGlyph*> (it->GetPointer());
        // TODO configure the glyph here
        Add(glyph);
    }

    Update(true);
    x_OnLayoutChanged();
}



///////////////////////////////////////////////////////////////////////////////
/// CGapTrackFactory
///////////////////////////////////////////////////////////////////////////////
ILayoutTrackFactory::TTrackMap
CGapTrackFactory::CreateTracks(SConstScopedObject& object,
                                  ISGDataSourceContext* ds_context,
                                  CRenderingContext* r_cntx,
                                  const SExtraParams& params,
                                  const TAnnotMetaDataList& src_annots) const
{
    TTrackMap tracks;
    CSGGapDS::TAnnotInfoMap annots;
    CSGGapDS::GetTrackNames(object, annots, r_cntx->GetVisSeqRange(), params);

    // collect NA tracks
    if ( !src_annots.empty() ) {
        TAnnotNameTitleMap t_annots;
        // TODO: please fill in the track type string. If there are multiple ones
        // please call this multiple times using the corresponding type name.
        GetMatchedAnnots(src_annots, params.m_Annots, "seq-table", "", t_annots);
        ITERATE (TAnnotNameTitleMap, iter, t_annots) {
            annots[iter->first] = CSGGapDS::SAnnotInfo(CSGGapDS::eSeqTable, iter->second);
        }
    }

    // create feature tracks
    ITERATE (CSGGapDS::TAnnotInfoMap, iter, annots) {
        // Create a new DS object for each track
        CIRef<ISGDataSource> pre_ds =
            ds_context->GetDS(typeid(CSGGapDSType).name(), object);
        CSGGapDS* seq_ds = dynamic_cast<CSGGapDS*>(pre_ds.GetPointer());
        seq_ds->SetDepth(params.m_Level);
        seq_ds->SetAdaptive(params.m_Adaptive);
        seq_ds->SetAnnotInfo(iter->second);
        seq_ds->SetAnnot(iter->first);

        CRef<CGapTrack> track(new CGapTrack(seq_ds, r_cntx));
        if ( !iter->second.m_Title.empty() ) {
            track->SetTitle(iter->second.m_Title);
        }
        tracks[iter->first] = track.GetPointer();

        track->SetAnnot(iter->first);
    }

    return tracks;
}


string CGapTrackFactory::GetExtensionIdentifier() const
{
    return CGapTrack::m_TypeInfo.GetId();
}


string CGapTrackFactory::GetExtensionLabel() const
{
    return CGapTrack::m_TypeInfo.GetDescr();
}


CRef<CTrackConfigSet>
CGapTrackFactory::GetSettings(CSeqGraphicConfig* /*g_conf*/,
                               const string& /*profile*/,
                               const TKeyValuePairs& /*settings*/,
                               const CTempTrackProxy* /*track_proxy*/) const
{
    CRef<CTrackConfigSet> config_set(new CTrackConfigSet);
    CRef<CTrackConfig> config(new CTrackConfig);
    config_set->Set().push_back(config);
    config->SetHelp() = GetThisTypeInfo().GetDescr();
    config->SetLegend_text("graphicalViewLegend.html#anchor_4.4");
    return config_set;
}


END_NCBI_SCOPE
