/*  $Id: gwas_glyph.cpp 27829 2013-04-09 21:26:28Z rudnev $
* ===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data, the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties, express or implied, including
*  warranties of performance, merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================
*
* Authors:  Vlad Lebedev, Liangshou Wu
*
* File Description:
*    CGwasGlyph -- utility class for having
*                      feature density histograms his in graphical layouts.
*
*/

#include <ncbi_pch.hpp>
#include <gui/packages/pkg_snp/gap/gwas_glyph.hpp>
#include <gui/widgets/seq_graphic/rendering_ctx.hpp>
#include <gui/widgets/wx/message_box.hpp>
#include <gui/objutils/obj_fingerprint.hpp>
#include <gui/opengl/gltexturefont.hpp>
#include <gui/opengl/glrender.hpp>
#include <objmgr/util/sequence.hpp>
#include <util/checksum.hpp>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

// XXX Remove this later
static CRgbaColor g_ColorMapping(double value)
{
    static float colors[][6] = {// color       pvalue
        {.75, 1.0, 1.0, 1.0},   // teal        < 2
        {.25, 1.0, 1.0, 1.0},   // sky blue    [2-3)
        {0.0, .50, 1.0, 1.0},   // blue        [3-4)
        {0.0, 1.0, 0.0, 1.0},   // green       [4-5)
        {1.0, 1.0, 0.0, 1.0},   // yellow      [5-6)
        {1.0, .50, 0.0, 1.0},   // orange      [6-7)
        {1.0, 0.0, 0.0, 1.0}    // red         >= 7
    };

    int v = (int)floor(value);
    float *c = 0;

    if( v < 2)
        c = colors[0];
    else if (v < 7)
        c = colors[v-1];
    else
        c = colors[6];

    return CRgbaColor( c[0], c[1], c[2], c[3] );
}

//
// CGwasGlyph::CGwasGlyph()
//
CGwasGlyph::CGwasGlyph(
        const TDensityMap& mapValues,
        const CSeq_loc& loc,
        objects::CScope * scope)
    : m_Location(&loc)
    , m_Map(mapValues)
    , m_bHasObjectList(false)
    , m_Scope(scope)
{
}

bool CGwasGlyph::OnLeftDblClick(const TModelPoint& /*p*/)
{
    string name;
    // TODO
    return true;
}


bool CGwasGlyph::NeedTooltip(const TModelPoint& p, string& tt, string& t_title) const
{
    GetTooltip(p, tt, t_title);
    return true;
}


void CGwasGlyph::GetTooltip(const TModelPoint& pos, string& tt, string& /*t_title*/) const
{
    // Determine bin number
    // bin number is floor (pos/window)
    float           window  = m_Map.GetWindow();
    TSeqPos         start   = m_Map.GetStart();
    unsigned int    bin     = (unsigned int) floor( (pos.X()-start)/window );

    // Make sure 'bin' falls within [0, m_MapStudies.GetBins)
    bin = max<int>(0, bin);
    bin = min<int>(bin, m_Map.GetBins()-1);

    // Calculate value
    float value = m_Map[bin].pvalue;

    if(value > 15)
        tt += "-log10(pval): >15";
    else {
        tt += "-log10(pval): ";
        tt += NStr::DoubleToString(value);
    }

    // Calculate the count
    int counts = m_Map[bin].count;

    if(counts > 15) {
        tt += "\nSNP Count: >15";
    }
    else {
        tt += "\nSNP Count: ";
        tt += NStr::IntToString(counts);
    }

    // display rsid
    int rsid = m_Map[bin].rsid;
    if(rsid != 0) {
        tt += "\nSignificant: rs ";
        tt += NStr::IntToString(rsid);
    }
}


bool CGwasGlyph::IsClickable() const
{
    // TODO
    return false;
}

string CGwasGlyph::GenerateBinSignature(const string& title,
                                        const string& annot_name,
                                        int gi,
                                        TSeqRange range)
{
    char buff[128];
    Uint2 type;

    // create type
    type = (CObjFingerprint::eSnpTrack) << 8;
    type |= 0x01;  //!! Need to enumerate these types.   1 is for Gwas/Gap Track

    // create checksum
    CChecksum cs(CChecksum::eCRC32);
    Uint4 checksum(0);
    if(!title.empty()) {
        cs.AddLine(title);
        checksum = cs.GetChecksum();
    }
    CChecksum cs_name(CChecksum::eCRC32);
    Uint4 checksum_name;
    cs_name.AddLine(annot_name);
    checksum_name = cs_name.GetChecksum();

    sprintf(buff, "gi|%d-%08x-%08x-%04x-%08x-%08x",
            gi, range.GetFrom(), range.GetTo(),
            type, checksum, checksum_name);

    return string(buff);
}

TSeqRange CGwasGlyph::x_GetBinRange(unsigned int bin) const
{
    // determine selected bin
    // bin number is floor (pos/window)
    double window  = m_Map.GetWindow();
    TSeqPos start = m_Map.GetStart();

    TSeqPos startPos = (TSeqPos) (
                         start              // start position
                       + bin*window + .5f  // + rounded offset
                       );


    TSeqPos endPos   =  (TSeqPos) (
                         start                  // start position
                       + ((bin+1)*window + .5f) // + rounded bin position
                       - 1                      // subtract one less bp  [start, end)
                       );

    TSeqRange range(startPos, endPos);
    return range;
}

void CGwasGlyph::x_BuildObjectList() const
{
    // for every non-zero bin, generate the bin's signature and range
    if (m_bHasObjectList==false) {
        int gi = sequence::GetGiForId(*(m_Location->GetId()), *m_Scope);
        TDensityMap::const_iterator iter;
        unsigned int binNum = 0;

        for (iter = m_Map.begin(); iter != m_Map.end(); ++iter, ++binNum) {
            const SValue& mbin = *iter;
            if (mbin.count> 0) {
                SBin* bin = new SBin;
                CRef<CObject> obj(bin);
                bin->val = mbin;
                bin->range     = x_GetBinRange(binNum);
                bin->signature = GenerateBinSignature(GetTitle(),
                                                      GetAnnotName(),
                                                      gi,
                                                      bin->range);

                m_listObjs.push_back(obj);
            }
        }
        m_bHasObjectList = true;
    }
}

void CGwasGlyph::GetHTMLActiveAreas(CSeqGlyph::TAreaVector* p_areas) const
{
    bool isFlipped = m_Context->IsFlippedStrand();

    if (m_bHasObjectList == false) {
        x_BuildObjectList();
    }

    ITERATE(vector<CConstRef<CObject> >, iter, m_listObjs) {
        const SBin* bin =
            dynamic_cast<const SBin*> ((*iter).GetPointer());

        TSeqRange range = bin->range;
        TVPUnit x1 = GetRenderingContext()->SeqToScreenX(range.GetFrom());
        TVPUnit x2 = GetRenderingContext()->SeqToScreenX(range.GetToOpen());

        if (isFlipped) {
            x1 = -x1;
            x2 = -x2;
        }

        CHTMLActiveArea area;
        CSeqGlyph::x_InitHTMLActiveArea(area);
        area.m_Bounds.SetHorz(x1,x2);
        area.m_Flags = CHTMLActiveArea::fNoSelection | CHTMLActiveArea::fNoPin;
        area.m_SeqRange  = bin->range;
        area.m_Signature = bin->signature;
        p_areas->push_back(area);
    }
}

void CGwasGlyph::x_Draw() const
{
    CGlRender& gl = GetGl();
    const bool isFlipped = m_Context->IsFlippedStrand();
    const int binHeight = 14; //m_ConfigSettings->GetGwasBinHeight();
    const int binWidth  = 14; //m_ConfigSettings->GetGwasBinWidth();

    TModelUnit top = GetTop();

    TModelUnit line_y1 = 1 + top;
    TModelUnit line_y2 = line_y1 + binHeight;

    // SETUP COMMON attributes
    double              window  = m_Map.GetWindow();
    const TSeqPos       start   = m_Map.GetStart();
    const size_t        numBins = m_Map.GetBins();

    // Y values are the same
    TModelUnit y1 = line_y1;
    TModelUnit y2 = line_y2;

    CRgbaColor color;

    // Render a white background for the track
    TModelUnit x1   = 0.0;
    TModelUnit x2   = 0.0;

    // Render the colored bins
    for(size_t bin = 0;  bin < numBins; bin++)
    {
        TModelUnit value = m_Map[bin].pvalue;

        if (value != 0) {
            // Based on pvalue, create the color
            // if the value was not compressed take -log10 of value
            color = g_ColorMapping(value);

            // draw the stuff
            x1 = (start + bin*(window));
            x2 = (window!=1) ? (start + (bin+1)*(window)) : x1+1;

            // Draw filled box
            gl.ColorC(color);
            m_Context->DrawQuad(x1, y1, x2, y2);

            gl.Color4f(.76f,.76f,.76f, 1.0f);
            m_Context->DrawRect(x1, y1, x2, y2);
         }
    }

    //CRef<CGlBitmapFont> font( new CGlBitmapFont(CGlBitmapFont::eHelvetica8) ); //m_ConfigSettings->GetGwasFont();
    CGlTextureFont font(CGlTextureFont::eFontFace_Helvetica_Bold, 8);
    // Use larger font here for "+" sign since its smaller than other (numeric) characters
    CGlTextureFont font_plus(CGlTextureFont::eFontFace_Helvetica_Bold, 10);
    string text;

    gl.Color4f(0.0f, 0.0f,0.0f, 1.0f);

    for(size_t bin = 0;  bin < numBins; bin++) {
        int count = m_Map[bin].count;
        if(count != 0) {
            x1   = ((bin)*(window) + start);
            if( count > 9)
                text = "+";
            else
                NStr::IntToString(text, count);

            int dx = (isFlipped) ? 2 : -2;

            TModelUnit w = ((binWidth + dx)/2)*(m_Context->GetScale());
            TModelUnit h = font.TextHeight();

            if (text == "+") {
                m_Context->TextOut(&font_plus, text.c_str(),
                    x1+w, y2 - (binHeight - h) * 0.5, true);
            }
            else
            {
                m_Context->TextOut(&font, text.c_str(),
                    x1+w, y2 - (binHeight - h) * 0.5, true);
            }
        }
    }
}

void CGwasGlyph::x_UpdateBoundingBox()
{
/*
    const CHistParams* config = x_GetConfig();
    m_Height = config->m_Height;
    TSeqRange range = GetRange();
    m_Width = range.GetLength();
    m_Pos.m_X = range.GetFrom();
*/
    m_Height = 14;
    TSeqRange range = GetRange();
    m_Width = range.GetLength();
    m_Pos.m_X = range.GetFrom();
}

TSeqRange CGwasGlyph::GetRange(void) const
{
    return m_Location->GetTotalRange();
}

END_NCBI_SCOPE
