/*  $Id: analysis_file.cpp 23938 2011-06-22 18:50:32Z wuliangs $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Melvin Quintos
 *
 * File Description:
 *    Loader for association study files.  See study_loader.hpp
 *
 */

#include <ncbi_pch.hpp>

#include <util/xregexp/regexp.hpp>
#include <fstream>

#include <gui/objutils/registry.hpp>
#include "analysis_file.hpp"

using namespace std;

BEGIN_NCBI_SCOPE

static const string kMarkerColumn    = "marker accession";
static const string kColumnDelimiter = "\t";

CAnalysisFile::CAnalysisFile()
: m_nIdPos(0), m_nValuePos(0), m_pMapColumnAliases(0)
{
}

const string& CAnalysisFile::GetColorMap() const
{
    return m_strColorMap;
}

const string& CAnalysisFile::GetRankFn() const
{
    return m_strRankFn;
}

string CAnalysisFile::GetHeaderInfo(const string &name) const
{
    string str = kEmptyStr;
    TMapStringString::const_iterator it = m_mapKeywords.find(name);
    if (it != m_mapKeywords.end()) {
        str = it->second;
    }
    return str;
}

const CAnalysisFile::TVecDataRows&  CAnalysisFile::GetRows() const
{
    return m_DataRows;
}

const CAnalysisFile::TListString&   CAnalysisFile::GetColumnNames() const
{
    return m_listColumnNames;
}

int CAnalysisFile::GetColumnPos(const string& name) const
{
    // search for exact column name
    TMapNamePos::const_iterator itPos = m_mapColumnPos.find(name);
    if (itPos != m_mapColumnPos.end())    {
        return itPos->second;
    }

    // if it fails, search for it by alias
    if (m_pMapColumnAliases) {
        TMMapStringString::const_iterator it2 = m_pMapColumnAliases->find(name);
        if (it2 != m_pMapColumnAliases->end()) {
            do {
                const string& alias = it2->second;
                itPos = m_mapColumnPos.find(alias);
                if (itPos != m_mapColumnPos.end())    {
                    return itPos->second;
                }
                it2++;
            } while( it2 != m_pMapColumnAliases->upper_bound(name));
        }
    }

    return -1;  // otherwise no column exists
}

bool CAnalysisFile::Load(const wxString& filename, const TMMapStringString *pMapAliases, bool bHeadersOnly)
{
    bool    ret                 = false;
    bool    bFoundColumnHeaders = false;

    size_t      strPos;
    CRegexp     keywords("^\\s*#(.+?):\\s*(.+)$");  // find keywords
    CRegexp     dataline("^\\s*[^#\\s]\\w");          // find a non blank, non-comment line

    // store alias mapping
    x_AddColumnAliases(pMapAliases);

    try {
        CNcbiIfstream file(filename.fn_str());

        if (!file.is_open())
            return false;

        // read every line
        while (!file.eof()) {
            string line;
            getline(file, line);

            // look for lines with a comment
            bool isComment = (strPos=line.find("#") != string::npos);

            if (isComment) {
                // look for keywords
                if ( keywords.IsMatch(line) ) {
                    string keyword = keywords.GetSub(line, 1);
                    string value   = keywords.GetSub(line, 2);
                    NStr::TruncateSpacesInPlace(keyword);
                    NStr::TruncateSpacesInPlace(value);

                    m_mapKeywords[keyword] = value;
                    continue;
                }
                // otherwise modify line to remove comment
                else {
                    line = line.substr(0,strPos-1);
                }
            }

            if ( dataline.IsMatch(line) ) {                                
                if (!bFoundColumnHeaders) {
                    x_ReadColumnHeaders(line);
                    bFoundColumnHeaders = true;

                    // exit if we only read the header section
                    if (bHeadersOnly) 
                        break;
                }
                else {
                    x_ReadDataLine(line);
                }
            }
        } // end while(!file.eof())

        if (bFoundColumnHeaders) {
            ret = true;
        }
    }    
    catch (const string error) {
        m_strError = error;
    }
    catch (const char *error) {
        m_strError = error;
    }

    return ret;
}

const string& CAnalysisFile::GetError() const 
{
    return m_strError;
}

void CAnalysisFile::x_AddColumnAliases(const TMMapStringString *map)
{
    m_setValidColumns.clear();
    m_setValidColumns.insert(kMarkerColumn);

    // assume that chrom and chrom_pos columns are valid columns
    m_setValidColumns.insert("chrom");
    m_setValidColumns.insert("chrom_pos");

    // Add the aliases for chrom and chrom_pos
    if (map) {
        m_pMapColumnAliases = map;
        CAnalysisFile::TMMapStringString::const_iterator it;

        it = map->find("chrom");
        if (it != map->end()) {
            do {
                m_setValidColumns.insert(it->second);
                it++;
            } while (it != map->upper_bound("chrom"));
        }

        it = map->find("chrom_pos");
        if (it != map->end()) {
            do {
                m_setValidColumns.insert(it->second);
                it++;
            } while (it != map->upper_bound("chrom_pos"));
        }
    }
}

void CAnalysisFile::x_ReadDataLine(const string& line)
{

    TVecString tokens;

    NStr::Tokenize(line, kColumnDelimiter, tokens);

    // grab the id and value columns
    string id  = tokens.at(m_nIdPos);
    double val = NStr::StringToDouble(tokens.at(m_nValuePos));

    m_DataRows.push_back( CRef<SDataRow>(new SDataRow()) );
    SDataRow& row = *m_DataRows.back();
    row.id      = id;
    row.value   = val;

    ITERATE(TListInt, it, m_listColumnPos) {
        row.column_data.push_back(tokens.at(*it));
    }
}

// throws an exception if either ID or VALUE columns not found
void CAnalysisFile::x_ReadColumnHeaders(const string& line)
{
    // At this point, header of analysis file has been read

    // read configuration for the given track type
    // This will initialize various member variables
    string track_type   = GetHeaderInfo("Track Type");
    string access_type  = GetHeaderInfo("Access Type");

    x_ReadRegistryInfo( track_type, access_type );

    const string& storeColumns = m_strStoreCols;
    const string& valueColumn  = m_strValueCol;

    //  setup the list of valid columns that will be read from file
    if (valueColumn != kEmptyStr)
        m_setValidColumns.insert(valueColumn);

    if (storeColumns != kEmptyStr) {
        // tokenize the visual columns and add each token into setValidColumns
        TVecString col_tokens;
        string delim = ",";
        NStr::Tokenize(storeColumns, delim, col_tokens);
        NON_CONST_ITERATE(TVecString, it, col_tokens) {
            string& col_name = *it;
            NStr::TruncateSpacesInPlace(col_name);
            if (!col_name.empty())
                m_setValidColumns.insert(*it);
        }
    }

    //  Now read the column header line and save the position of the column
    //   in the map
    TVecString tokens;

    NStr::Tokenize(line, kColumnDelimiter, tokens);

    int numCol = (int) tokens.size();
    for (int i=0, nCol=0; i < numCol; i++) {
        string& columnName = tokens[i];
        NStr::TruncateSpacesInPlace(columnName);

        // if column name is a valid column to store, store its position
        TSetString::iterator itColName = m_setValidColumns.find(columnName);
        if (itColName != m_setValidColumns.end()) {
            m_mapFileColumnPos[columnName] = i;

            m_listColumnNames.push_back(columnName);
            m_mapColumnPos[columnName] = nCol++;
        }
    }

    // make sure the id and value columns exist
    if (m_mapFileColumnPos.find(kMarkerColumn) == m_mapFileColumnPos.end() ||
        m_mapFileColumnPos.find(valueColumn)   == m_mapFileColumnPos.end())
    {
        string error =  "Could not find proper columns related to\n"
                        "Access Type: '" + access_type + "'\n"
                        "Track Type:  '" + track_type + "'\n\n"
                        "Expecting columns '" + kMarkerColumn + "' and '" + valueColumn + "'";
        throw error;
    }

    m_nIdPos  = m_mapFileColumnPos[kMarkerColumn];
    m_nValuePos = m_mapFileColumnPos[valueColumn];

    // Create the order which is used to read the columns in
    ITERATE(TListString, it, m_listColumnNames) {
        int col_pos = m_mapFileColumnPos[*it];
        m_listColumnPos.push_back(col_pos);
    }
}

void CAnalysisFile::x_ReadRegistryInfo(const string &track_name, const string &access_type)
{
    const string  snp_section = "GBENCH.Plugins.View.Snp.TrackTypes.";
    string default_value, default_fn, default_colors, default_stores;
    string track_type = track_name + "_" + access_type;
    
    // Read track types from registry
    CGuiRegistry& reg = CGuiRegistry::GetInstance();
    CGuiRegistry::TReadView view = reg.GetReadView(snp_section + track_name + access_type);

    x_GetDefaults( track_type, &default_value, &default_fn, &default_colors, &default_stores );

    m_strRankFn     = view.GetString("rank function", default_fn);
    m_strColorMap   = view.GetString("color map",     default_colors);
    m_strValueCol   = view.GetString("value column",  default_value);
    m_strStoreCols  = view.GetString("store columns", default_stores);
}

void CAnalysisFile::x_GetDefaults(const string &name, string *value, string *fn, string *colors, string* /*stores*/)
{    
    if        (NStr::CompareNocase(name, "Association_Public")==0) {
        *value  = "Plot data";
        *fn     = "";
        *colors = "0-1  (192, 255, 255);2-2  ( 64, 255, 255);3-3  (  0, 128, 255);"
                  "4-4  (  0, 255,   0);5-5  (255, 255,   0);6-6  (255, 128,   0);"
                  "7-15 (255,   0,   0);";
    } else if (NStr::CompareNocase(name, "Association_Controlled")==0) {
        *value  = "P-value";
        *fn     = "-log(x)";
        *colors = "0-1  (192, 255, 255);2-2  ( 64, 255, 255);3-3  (  0, 128, 255);"
                  "4-4  (  0, 255,   0);5-5  (255, 255,   0);6-6  (255, 128,   0);"
                  "7-15 (255,   0,   0);";

    } else if (NStr::CompareNocase(name,"Linkage")==0) {
        *value  = "LOD";
        *fn     = "";
        *colors = "0-15 (255,0,0)";
    }
    else {
        *value  = "value";
        *fn     = "";
        *colors = "0-15 (255,0,0)";
    }
}

END_NCBI_SCOPE

