#ifndef GUI_PLUGINS_DOC_ASSN_STUDY___ANALYSIS_FILE___HPP
#define GUI_PLUGINS_DOC_ASSN_STUDY___ANALYSIS_FILE___HPP

/*  $Id: analysis_file.hpp 22654 2010-12-08 20:13:39Z katargir $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Melvin Quintos
 *
 * File Description:
 *      CAnalysisFile loads and holds data for analysis files.
 *
 */

#include <corelib/ncbiobj.hpp>

#include <set>

#include <wx/string.h>

BEGIN_NCBI_SCOPE

class CAnalysisFile
{
public:

// case insensitve searches
struct less_icase  {
    bool operator() (const string& x, const string& y) const {
        return (NStr::CompareNocase(x,y) < 0);
    }
};

typedef vector<string>                  TVecString;
typedef list<string>                    TListString;
typedef map<string, int, less_icase>    TMapNamePos;
typedef map<string, string, less_icase> TMapStringString;
typedef multimap<string, string, less_icase> TMMapStringString;

struct SDataRow : public CObject
{
    string      id;
    double      value;
    TVecString  column_data;
};

typedef vector< CRef<SDataRow> >    TVecDataRows;

public:
    CAnalysisFile();

    bool    Load(const wxString& fn, const TMMapStringString* col_aliases=0, bool headersOnly = false);

    // Returns the 0-based position of a given column.  Returns -1 if column doesn't exist.
    //  Column names are case sensitive
    int     GetColumnPos(const string &name) const;
    string  GetHeaderInfo(const string& name) const;
    const string& GetColorMap() const;
    const string& GetRankFn()   const;
    const string& GetError()    const;  // Gets error when loading fails


    const TVecDataRows&  GetRows() const;
    const TListString&   GetColumnNames() const;

private:
    typedef set<string, less_icase> TSetString;
    typedef list<int>   TListInt;

private:
    void x_AddColumnAliases(const TMMapStringString*);
    void x_ReadDataLine(const string& line);
    void x_ReadColumnHeaders(const string& line);
    void x_ReadRegistryInfo(const string &track_name, const string &access_type);
    void x_GetDefaults(const string &name, string *value, string *fn, string *colors, string *stores);

private:
    int                 m_nIdPos;
    int                 m_nValuePos;
    TListInt            m_listColumnPos;    // List of stored column position as they appear in file

    string              m_strError;         // error on load
    string              m_strRankFn;
    string              m_strColorMap;
    string              m_strValueCol;
    string              m_strStoreCols;    // extra columns to be stored in memory when file is read

    TVecDataRows        m_DataRows;
    TListString         m_listColumnNames;  // List of stored column names in order as they appear in file
    TMapNamePos         m_mapColumnPos;     // Map of column name to its position in SDataRow::column_data
    TMapNamePos         m_mapFileColumnPos; // Map of column name to position in file
    TMapStringString    m_mapKeywords;      // Map of keywords found in file's header
    TSetString          m_setValidColumns;  // Names of columns to be read from file

    const TMMapStringString* m_pMapColumnAliases; // Map of aliases for column names (e.g. chr -> chrom)

};

END_NCBI_SCOPE



#endif // GUI_PLUGINS_DOC_ASSN_STUDY___ANALYSIS_FILE___HPP

