/*  $Id: analysis_loader.cpp 24508 2011-10-05 19:32:18Z wuliangs $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Melvin Quintos
 *
 * File Description:
 *    Loader for association study files.  See analysis_loader.hpp
 *
 */

#include <ncbi_pch.hpp>
#include <objmgr/scope.hpp>

#include <objects/seq/Annot_descr.hpp>
#include <objects/seq/Annotdesc.hpp>
#include <objects/seqres/Seq_graph.hpp>
#include <objects/seqres/Byte_graph.hpp>
#include <objects/seqloc/Seq_loc.hpp>
#include <objects/seqfeat/Seq_feat.hpp>
#include <objects/seqloc/Seq_point.hpp>
#include <objects/general/User_object.hpp>
#include <objects/general/Object_id.hpp>
#include <objmgr/seq_vector.hpp>
#include <util/xregexp/regexp.hpp>

#include <math.h>
#include <sstream>

#include "analysis_file.hpp"
#include "analysis_loader.hpp"

#include <gui/widgets/wx/wx_utils.hpp>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

static const string kAssnStudyId = "assn_study_val"; // identifier for user object features
static const string kAssnAnnotId = "GWAS_";           // unique annotation identifier
static const string kDescrAnnotId = "AssnDescr";     // object-id for Analysis annotations

static const string kArrayBuild36[] =
{ "", "NC_000001.9", "NC_000002.10", "NC_000003.10", "NC_000004.10", "NC_000005.8",
  "NC_000006.10", "NC_000007.12", "NC_000008.9", "NC_000009.10", "NC_000010.9",
  "NC_000011.8", "NC_000012.10", "NC_000013.9", "NC_000014.7", "NC_000015.8",
  "NC_000016.8", "NC_000017.9", "NC_000018.8", "NC_000019.8", "NC_000020.9",
  "NC_000021.7", "NC_000022.9", "NC_000023.9", "NC_000024.8" };

static const string kArrayBuild37[] =
{ "", "NC_000001.10", "NC_000002.11", "NC_000003.11", "NC_000004.11", "NC_000005.9",
  "NC_000006.11", "NC_000007.13", "NC_000008.10", "NC_000009.11", "NC_000010.10",
  "NC_000011.9", "NC_000012.11", "NC_000013.10", "NC_000014.8", "NC_000015.9",
  "NC_000016.9", "NC_000017.10", "NC_000018.9", "NC_000019.9", "NC_000020.10",
  "NC_000021.8", "NC_000022.10", "NC_000023.10", "NC_000024.9"};

//////////////////////////////////////////////////////////////////////////////
// Static Functions
//////////////////////////////////////////////////////////////////////////////

// convert chr 1-22,x,y to NC_00000[01-24].  Throws exception otherwise
string s_GetChrAccession(const string& chr, const string& build)
{
    static CRegexp rxBuild("(\\d+)(\\.\\d+)?$");  // Match something like 36 or 36.3 or Hs36.3
    string acc, buildNum;
    const string *pArrayBuild = 0;

    if (rxBuild.IsMatch(build)) {
        buildNum = rxBuild.GetSub(build,1);
    }

    // Determine chr accession set from build
    if (buildNum == "36") {
        pArrayBuild=kArrayBuild36;
    }
    else if (buildNum == "37") {
        pArrayBuild=kArrayBuild37;
    }
    else {
        throw "Need Build Info!";
    }

    // Try to retrieve chr accession.version
    try {
        int chrN = NStr::StringToInt(chr);
        if (chrN > 0  &&  chrN < 23)
            acc = pArrayBuild[chrN];
        else            
            throw "Unexpected chromosome number!";
    }
    catch (CStringException &) {
        // Converting the number failed.
        // See if it's the 'x' or 'y' chromosome
        if (NStr::CompareNocase(acc.c_str(), "x") == 0) {
            acc = pArrayBuild[23];
        }
        else if (NStr::CompareNocase(acc.c_str(), "y") == 0) {
            acc = pArrayBuild[24];
        }
        else {
            // There's no hope, just bail!
            throw "Need Chromosome value!";
        }
    }
    return acc;
}

//////////////////////////////////////////////////////////////////////////////
// Public Methods
//////////////////////////////////////////////////////////////////////////////

CAnalysisFileLoader::CAnalysisFileLoader(int comp)
    : m_bUseRankFunc(false),
      m_nCompression(comp),
      m_pParams(0)
{
    // Create Scope
    m_ObjMgr = CObjectManager::GetInstance();

    m_Scope.Reset(new CScope(*m_ObjMgr));
    m_Scope->AddDefaults();
}

void CAnalysisFileLoader::Load(const wxString& filename)
{
    TMapMapValues    theStudy;

    if(filename.empty()) return;

    try {
        // TODO load an alias map.  For now, just make one up
        typedef CAnalysisFile::TMMapStringString TAliases;
        TAliases aliases;
        aliases.insert(TAliases::value_type("chrom",     "chr"));
        aliases.insert(TAliases::value_type("chrom",     "chr id"));
        aliases.insert(TAliases::value_type("chrom_pos", "position"));
        aliases.insert(TAliases::value_type("chrom_pos", "chr position"));
        aliases.insert(TAliases::value_type("chrom_pos", "chr Position"));

        m_AnaFile.reset(new CAnalysisFile());
        if (m_AnaFile->Load(filename, &aliases)) {
            // Read the data.  Store data in the study
            x_ReadData(&theStudy);
            // Create seq-annots for this study
            x_CreateSeqAnnots(theStudy);
        }
    }
    catch (std::exception&)
    {}
}

void CAnalysisFileLoader::Load(const wxString& filename, const CGwasLoadParams &p)
{
    m_pParams = &p;
    Load(filename);
}

void CAnalysisFileLoader::SetCompression(int comp)
{
    m_nCompression = max(1, comp); // must be larger than 1
}

void CAnalysisFileLoader::GetAnnots(CAnalysisFileLoader::TListAnnots *list)
{
    list->clear();
    ITERATE(TListAnnots, iter, m_listAnnots) {
        list->push_back(*iter);
    }
}

//////////////////////////////////////////////////////////////////////////////
// Private Methods
//////////////////////////////////////////////////////////////////////////////
string CAnalysisFileLoader::x_GetLabel() const
{   
    string studyName;    
    
    if (m_pParams) 
        studyName = ToStdString(m_pParams->GetTitle());
    else 
        studyName = m_AnaFile->GetHeaderInfo("name"); 

    return studyName;
}

// Read the Analysis file object.  Store results in 'mapValues'
void CAnalysisFileLoader::x_ReadData(TMapMapValues * mapValues)
{
    TMapMapValues &mapMapValues = *mapValues;

    /*
        TODO:
        The current strategy is to load a marker's chromosome and position
        from the data columns.  Future versions (when external database is ready)
        will use the marker's id to lookup the chromosome and position externally
        via file, or web service

        Current implementation will look for columns labelled 'chr' and 'chr_pos'
    */

    int chrCol    = m_AnaFile->GetColumnPos("chrom");
    int chrPosCol = m_AnaFile->GetColumnPos("chrom_pos");

    // make sure columns exist.  If they don't, just return
    if (chrCol == -1 || chrPosCol== -1)
        return;

    // For all data rows, extract the id, value, chr, and chr_pos
    ITERATE(CAnalysisFile::TVecDataRows, it, m_AnaFile->GetRows()) {
        const CAnalysisFile::SDataRow& row = **it;

        try {

        SValue v;
        v.value = row.value;
        v.id    = row.id;

        const string& chr   = row.column_data.at(chrCol);
        unsigned int pos    = NStr::StringToUInt(row.column_data.at(chrPosCol));

        pos -= 1; // Analysis files are 1-based.

        // store in the map of values by chromosome and position
        mapMapValues[chr][pos] = v;

        } catch (std::exception&) { /* ignore parsing errors, continue to next row */ }
    }

    // Attempt to read the rank function
    try {
        mu::string_type rank_func(m_AnaFile->GetRankFn().c_str());
        m_RankFunc.SetExpr(rank_func);
        m_bUseRankFunc = true;
    }
    catch(std::exception&) {
        m_bUseRankFunc = false;
    }
}

void CAnalysisFileLoader::x_CreateSeqAnnots(const TMapMapValues & study)
{
    string build = m_AnaFile->GetHeaderInfo("Human genome build");

    /* For each chromosome, create two seq-annots.  The first containing a seq-graph,
        and the other containing a seq-feat table */
    ITERATE(TMapMapValues, it, study) {
        const string& chr           = it->first;  // chromosome
        const TMapValues& mapValue  = it->second; // value by position

        // translate the chromosome to a chromosome accession
        string chrAcc = s_GetChrAccession(chr, build);

        CRef<CSeq_annot> annot_graph;
        CRef<CSeq_annot> annot_feat;
        annot_graph = x_CreateSeqGraph(chrAcc, mapValue);
        annot_feat  = x_CreateSeqFtable(chrAcc, mapValue);

        x_AddAnnotDescr(&*annot_graph);
        x_AddAnnotDescr(&*annot_feat);

        m_listAnnots.push_back(annot_graph);
        m_listAnnots.push_back(annot_feat);
    }
}

/* TODO:
        add the following to the annotation's description:
        Color map string
        base url
        description
*/
void CAnalysisFileLoader::x_AddAnnotDescr(CSeq_annot *annot)
{
    CRef<CAnnotdesc>    descr(new CAnnotdesc());
    CRef<CUser_object>  user(new CUser_object());

    string color_map = m_AnaFile->GetColorMap();
    string rank_func = m_AnaFile->GetRankFn();
    string base_url  = m_AnaFile->GetHeaderInfo("base url");
    string title     = x_GetLabel();
    
    user->AddField("title", title);
    user->AddField("base_url", base_url);
    user->AddField("rank_func",rank_func);
    user->AddField("color_map",color_map);
    user->SetType().SetStr(kDescrAnnotId);

    descr->SetUser(*user);

    annot->SetDesc().Set().push_back(descr);

}

CRef<CSeq_annot> CAnalysisFileLoader::x_CreateSeqGraph(const string& accName, const TMapValues& mapValue)
{
    // Get handle to accession
    CBioseq_Handle  handle;
    CSeq_id         id;

    id.Set(accName);
    handle = m_Scope->GetBioseqHandle(id);

    if ( !handle )
        throw "No bioseq for accession";

    // Calculate graph representation across entire accession length
    unsigned long size = handle.GetSeqVector(CBioseq_Handle::eCoding_Iupac).size();
    CRef<CSeq_loc> loc = handle.GetRangeSeq_loc(0,size);
    size = ( (size-1) / m_nCompression) + 1;  // number of values needed for compressed data

    // create seq-graph
    CRef<CSeq_graph> graph(new CSeq_graph());
    string studyName   = x_GetLabel();

    graph->SetTitle(studyName);
    graph->SetLoc(*loc);
    graph->SetNumval(size);
    graph->SetGraph().SetByte().SetMin(0);
    graph->SetGraph().SetByte().SetMax(10); // max value to represent in graph
    graph->SetGraph().SetByte().SetAxis(0);
    graph->SetGraph().SetByte().SetValues().resize(size);
    graph->SetComp(m_nCompression);

    // Fill the graph values with the Value by position map
    CByte_graph::TValues& values  = graph->SetGraph().SetByte().SetValues();

    ITERATE(TMapValues, valIter, mapValue) {
        // assign the values
        unsigned int pos   = valIter->first;   // position

        // use compression
        unsigned int cpos  = pos/m_nCompression;
        values[cpos]        = x_MaxValueAndCount(mapValue,
                                                    pos,
                                                    m_nCompression); // use compression
    }

    // create annot for graph
    CRef<CSeq_annot> annot(new CSeq_annot());
    annot->SetNameDesc(kAssnAnnotId+studyName);
    annot->SetName(accName+";"+studyName);
    annot->SetData().SetGraph().push_back(graph);

    return annot;
}

// throws exception if no bioseqhandle found
CRef<CSeq_annot> CAnalysisFileLoader::x_CreateSeqFtable(
    const string& accName,
    const TMapValues&  mapValues )
{
    CRef<CSeq_id> id;
    CBioseq_Handle handle;
    string studyName;

    studyName = x_GetLabel();

    id.Reset(new CSeq_id());
    id->Set(accName);

    handle = m_Scope->GetBioseqHandle(*id);

    if ( !handle )
        throw "No bioseq";

    // create annotation
    CRef<CSeq_annot> annot(new CSeq_annot());
    annot->SetNameDesc(kAssnAnnotId+studyName);
    annot->SetName(accName+":"+studyName);

    ITERATE(TMapValues, pval, mapValues) {
        unsigned long   pos     = pval->first;        //
        double          value   = pval->second.value;  // value
        string          rsid    = pval->second.id;

        // Create seq-loc
        CSeq_loc *  loc = new CSeq_loc();
        loc->SetPnt().SetPoint(pos);
        loc->SetPnt().SetId(*id);

        // create seq-feat
        CRef<CSeq_feat> feat(new CSeq_feat());
        feat->SetLocation(*loc);
        feat->SetData().SetUser().AddField("value",value);
        feat->SetData().SetUser().AddField("id", rsid);
        feat->SetData().SetUser().SetType().SetStr(kAssnStudyId);

        // TODO - Add another user object for the column_data

        // add feature to annotation
        annot->SetData().SetFtable().push_back(feat);
    }

    return annot;
}

// translations for graph values
char CAnalysisFileLoader::x_Translate(double pval)
{
    // take -log10
    double val;

    if (m_bUseRankFunc) {
        try {
            m_RankFunc.DefineVar("x", &pval);
            val = m_RankFunc.Eval();
        }
        catch (std::exception&) {
            val = pval;
            m_bUseRankFunc=false;
        }
    }
    else {
        val = pval;
    }

    // bound val from [1-100]
    // and round to nearest integer
    val = max<double>(1, val);
    val = min<double>(val, 100);

    return (char) val;
}

// Find the maximum value in the region
// This function calculates the bin's rank (color) and count
char  CAnalysisFileLoader::x_MaxValueAndCount(const TMapValues &pval,
                                       unsigned long    pos,
                                       int              comp)
{
    unsigned long start = pos - (pos % comp);
    unsigned long end   = start + comp;
    int max = 0;
    int count = 0;

    TMapValues::const_iterator it      = pval.lower_bound(start);
    TMapValues::const_iterator it_end  = pval.upper_bound(end);
    for (; it != it_end; ++it) {
        count ++;
        unsigned char p_val = x_Translate(it->second.value);
        if (p_val > max) {
            max = p_val;
        }
    }
    if(count > 15)
        count = 15;
    if(max > 15)
        max = 15;
    char c = (count << 4);
    c |= 0x0F & max;

    return c;
}

END_NCBI_SCOPE

