/*  $Id: gwas_track_load_manager.cpp 26938 2012-11-28 15:46:40Z katargir $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Melvin Quintos
 *
 */

#include <ncbi_pch.hpp>

#include <wx/msgdlg.h>

#include <corelib/ncbifile.hpp>

#include "gwas_load_job.hpp"
#include "gwas_track_load_manager.hpp"
#include "gwas_track_params_panel.hpp"
#include "analysis_file.hpp"

#include <gui/widgets/wx/wx_utils.hpp>

#include <wx/filename.h>

BEGIN_NCBI_SCOPE

///////////////////////////////////////////////////////////////////////////////
/// CGwasTrackLoadManager
CGwasTrackLoadManager::CGwasTrackLoadManager()
:   m_Descr("Analysis Track files (*.chpa)", ""),
    m_SrvLocator(NULL),
    m_ParentWindow(NULL),
    m_State(eInvalid),
    m_ParamsPanel(NULL)
{
}


void CGwasTrackLoadManager::SetServiceLocator(IServiceLocator* srv_locator)
{
    m_SrvLocator = srv_locator;
}


void CGwasTrackLoadManager::SetParentWindow(wxWindow* parent)
{
    m_ParentWindow = parent;
}


const IUIObject& CGwasTrackLoadManager::GetDescriptor() const
{
    return m_Descr;
}


void CGwasTrackLoadManager::InitUI()
{
    m_State = eParams;
}


void CGwasTrackLoadManager::CleanUI()
{
    m_State = eInvalid;
    m_ParamsPanel = NULL; // window is destroyed by the system
}


wxPanel* CGwasTrackLoadManager::GetCurrentPanel()
{
    if(m_State == eParams)   {
        if(m_ParamsPanel == NULL)   {
            m_ParamsPanel = new CGwasTrackParamsPanel(m_ParentWindow);
            m_ParamsPanel->SetData(m_Params);
        }

        m_ParamsPanel->TransferDataToWindow();
        return m_ParamsPanel;
    }
    return NULL;
}


bool CGwasTrackLoadManager::CanDo(EAction action)
{
    switch(m_State) {
    case eParams:
        return action == eNext;
    case eCompleted:
        return action == eBack;
    default:
        _ASSERT(false);
        return false;
    }
}


bool CGwasTrackLoadManager::IsFinalState()
{
    return m_State == eParams;
}


bool CGwasTrackLoadManager::IsCompletedState()
{
    return m_State == eCompleted; 
}


bool CGwasTrackLoadManager::DoTransition(EAction action)
{
    if(m_State == eParams  &&  action == eNext)    {
        if(m_ParamsPanel->TransferDataFromWindow()) {
            //TODO validate
            m_Params = m_ParamsPanel->GetData();
            m_State = eCompleted;
            return true;
        }
        return false;
    } else if(m_State == eCompleted  &&  action == eBack)   {
        m_State = eParams;
        return true;
    }
    _ASSERT(false);
    return false;
}


IAppTask* CGwasTrackLoadManager::GetTask()
{
    if (m_FileNames.empty()) return 0;
    CIRef<CProjectService> srv = m_SrvLocator->GetServiceByType<CProjectService>();
    CGwasLoadJob* job = new CGwasLoadJob(srv.GetPointer(), m_Params, m_FileNames);
    return new CDataLoadingAppTask(*job);
}

wxString CGwasTrackLoadManager::GetFormatWildcard()
{
    return wxT("Analysis File (*.chpa)|*.chpa");
}


bool CGwasTrackLoadManager::ValidateFilenames(const vector<wxString>& filenames)
{
    /*
        NOTE: This seems to be the only place to verify the file selected is
        of the correct format.  If the file is properly validated, then
        the next panel will display parameters related to the file.
    */
    bool isValid = false;

    if (filenames.size() != 1) {
        wxMessageBox(wxT("Please select only 1 file"), wxT("Information"),
                     wxICON_INFORMATION);
    
    }
    else {
        const wxString& fn = filenames.front();
        wxString ext;
        wxFileName::SplitPath(fn, 0, 0, &ext);
        if (ext == wxT("chpa")) {
            // The file extension is right. Check the file headers
            isValid = x_IsValid(fn);
        }
        else {
            wxMessageBox(wxT("File is not CHPA"), wxT("Information"),
                     wxICON_INFORMATION);
        }
    }

    return isValid;
}


void CGwasTrackLoadManager::SetFilenames(const vector<wxString>& filenames)
{
    // Why is this function polymorphic?
    m_FileNames = filenames;
}

void CGwasTrackLoadManager::GetFilenames(vector<wxString>& filenames) const
{
    filenames = m_FileNames;
}

bool CGwasTrackLoadManager::IsInitialState()
{
    return m_State == eParams;
}


bool CGwasTrackLoadManager::RecognizeFormat(const wxString& filename)
{
    wxString ext;
    wxFileName::SplitPath(filename, 0, 0, &ext);
    return ext == wxT("chpa");
}


bool CGwasTrackLoadManager::RecognizeFormat(CFormatGuess::EFormat /*guess*/)
{
    return false;
}

string CGwasTrackLoadManager::GetExtensionIdentifier() const
{
    static string sid("gwas_track_load_manager");
    return sid;
}


string CGwasTrackLoadManager::GetExtensionLabel() const
{
    static string slabel("GWAS Track Load Manager");
    return slabel;
}

void CGwasTrackLoadManager::SetRegistryPath(const string& path)
{
    m_RegPath = path; // store for later use
    //TODO
    // m_Params.SetRegistryPath(m_RegPath + ".FASTAParams");
}

void CGwasTrackLoadManager::SaveSettings() const
{
    //m_Params.SaveSettings();
}


void CGwasTrackLoadManager::LoadSettings()
{
    //m_Params.LoadSettings();
}

bool CGwasTrackLoadManager::x_IsValid(const wxString& fn)
{
    bool isValid = false;
    CAnalysisFile file;
    string title, track_type;
    CGwasLoadParams::ETrackType eType;
   
    eType = CGwasLoadParams::eAssociation; // default value
 
    // Load the file's headers to extract key information
    if (file.Load(fn, 0, true)) {
        title       = file.GetHeaderInfo("name");
        track_type  = file.GetHeaderInfo("track type");

        if (NStr::CompareNocase(track_type.c_str(), "association")==0) {
            eType = CGwasLoadParams::eAssociation;
        } else if (NStr::CompareNocase(track_type.c_str(), "linkage")==0) {
            eType = CGwasLoadParams::eLinkage;
        }
        else {
            wxMessageBox(wxT("Did not recognize 'Track Type'\n")
                            wxT("File may not load correctly\n\n")
                            wxT("Set '# Track Type' in file's header to fix warning"), wxT("Warning"),
                            wxICON_WARNING);
        }
        isValid = true;
    }
    else 
    {
        // An error occured
        string error = "Could not load file.  \n\n" + file.GetError();

        wxMessageBox(ToWxString(error), wxT("Information"),
                     wxICON_INFORMATION);
    }

    if (isValid) {
        m_Params.SetTitle(ToWxString(title));
        m_Params.SetETrackType(eType);
    }

    return isValid;
}

END_NCBI_SCOPE
