/*  $Id: model_snp_table.cpp 25502 2012-03-27 21:23:47Z kuznets $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Melvin Quintos
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>

#include <gui/packages/pkg_snp/table/model_snp_table.hpp>
#include <gui/packages/pkg_snp/job_update.hpp>


#include <gui/objutils/utils.hpp>
#include <gui/objutils/label.hpp>
#include <gui/objutils/obj_event.hpp>
#include <gui/utils/app_job.hpp>
#include <gui/utils/app_job_impl.hpp>
#include <gui/utils/event_translator.hpp>
#include <gui/widgets/wx/wx_utils.hpp>
#include <objects/seqfeat/Gb_qual.hpp>
#include <objects/seqloc/Seq_bond.hpp>
#include <objects/seqloc/Seq_interval.hpp>
#include <objects/seqloc/Seq_loc.hpp>
#include <objects/seqloc/Seq_loc_equiv.hpp>
#include <objects/seqloc/Seq_point.hpp>
#include <objmgr/feat_ci.hpp>
#include <objmgr/util/feature.hpp>
#include <objmgr/util/sequence.hpp>

#include "snp_table_job.hpp"

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

BEGIN_EVENT_MAP( CSnpTableModel, CEventHandler )
    ON_MESSAGE( CAppJobNotification, CAppJobNotification::eStateChanged, &CSnpTableModel::OnJobNotification )
END_EVENT_MAP()

CSnpTableModel::CSnpTableModel()
    : m_nUnfilteredSnps(0), m_hJob(0), m_numRows(0)
{
    m_listSnps.Reset( new TSnpPtrListObject() );
    m_Filter.Reset( new SSnpFilter() );
}

void CSnpTableModel::ApplyFilter(const SSnpFilter& filter) 
{
    m_Filter.Reset(new SSnpFilter(filter));
    x_StartFilterJob();
}

void CSnpTableModel::LoadData(const CSeq_loc& loc, CScope* scope, bool onlySpecialsnps) {

    m_bShowOnlySpecial = onlySpecialsnps;
    m_Loc.Reset(&loc);

    if (scope != 0) {
        m_Scope.Reset(scope);
    }
    if ( !m_Scope.Empty() && !m_Loc.Empty() ) {
        x_StartLoadJob();
    }
}

void CSnpTableModel::OnJobNotification(ncbi::CEvent *evt)
{
    CAppJobNotification* notif = dynamic_cast<CAppJobNotification*>( evt );

    _ASSERT(notif);

    int                 job_id  = notif->GetJobID();
    IAppJob::EJobState  state   = notif->GetState();

    if (job_id != m_hJob) return;

    switch (state) {
    case IAppJob::eCompleted:
        {
        CAppJobDispatcher& disp = CAppJobDispatcher::GetInstance();
        CRef<CObject> res = disp.GetJobResult(job_id);
        CSnpTableJob::SResult &result = (CSnpTableJob::SResult&) res.GetObject();

        if (m_bAppliedFilter) {
            // Filter has been applied.  Go and update the table data
            _ASSERT(result.eJobType == CSnpTableJob::eJobFilter);
            m_listSnps.Reset(result.listSnpsFiltered.GetPointer());

            m_numRows = m_listSnps->GetData().size();

            CJobUpdate  job_notice(this, job_id, IAppJob::eCompleted, "Done!");
            x_Update(job_notice);
        }
        else {
            // have not applied filter.  The job that just finished was for fetching snps
            _ASSERT(result.eJobType == CSnpTableJob::eJobFetch);

            m_listUnfilteredSnps.Reset(result.listSnpsFetched.GetPointer());
            m_numRows = m_nUnfilteredSnps = m_listUnfilteredSnps->GetData().size();

            // Add all original items to be filtered
            m_listSnps->GetData().clear();
            TSnpList  &theList    = m_listUnfilteredSnps->GetData();
            TSnpListIter iter = theList.begin();
            for( ; iter != theList.end(); ++iter)
                m_listSnps->GetData().push_back(&(*iter));

            // Inform views that unfiltered data is ready
            CJobUpdate  job_notice(this, job_id, IAppJob::eCompleted, "Done Preparing SNPs ...");
            x_Update(job_notice);

            // Start filtering job
            x_StartFilterJob();
        }
        }
        break;
    case IAppJob::eCanceled:
    case IAppJob::eFailed:
        {
        CJobUpdate  job_notice(this, job_id, state, "");
        x_Update(job_notice);
        }
        break;
    default:
        break;
    } // end switch
}

/*
void CSnpTableModel::x_DisseminateJobUpdate( 
    int job_id, IAppJob::EJobState state, const string& msg
){
    CJobUpdate( )

}
*/

void CSnpTableModel::x_Update( CJobUpdate &notice )
{
    if( notice.GetJobState() == IAppJob::eCompleted ){
        x_FireDataChanged();
    }

    x_UpdateObserversNow( notice );
}

void CSnpTableModel::x_StartLoadJob()
{
    CRef<CSnpTableJob> job( new CSnpTableJob( *m_Scope, *m_Loc, m_bShowOnlySpecial) );
    CAppJobDispatcher& disp = CAppJobDispatcher::GetInstance();

    // reset filter flag
    m_bAppliedFilter = false;

    try {
        m_hJob = disp.StartJob( *job, "ObjManagerEngine", *this);
        CJobUpdate  job_notice(this, m_hJob, IAppJob::eRunning, "Preparing SNPs ...");
        x_Update(job_notice);
    } catch( CAppJobException& e ){
        LOG_POST( Error
            << "CSnpTableModel::x_StartLoadJob(): Failed to start job: "
            << e.GetMsg()
        );
        LOG_POST(e.ReportAll());
    }

}

void CSnpTableModel::x_StartFilterJob()
{
    CRef<CSnpTableJob> job( new CSnpTableJob(m_listUnfilteredSnps.GetPointer(), *m_Filter) );
    CAppJobDispatcher& disp = CAppJobDispatcher::GetInstance();

    // set filter flag
    m_bAppliedFilter = true;

    try {
        
        m_hJob = disp.StartJob( *job, "ObjManagerEngine", *this);

        CJobUpdate  job_notice(this, m_hJob, IAppJob::eRunning, "Filtering ...");
        x_Update(job_notice);

    } catch( CAppJobException& e ){
        LOG_POST( Error
            << "CSnpTableModel::x_StartFilterJob(): Failed to start job: "
            << e.GetMsg()
        );
        LOG_POST(e.ReportAll());
    }
}

////////////////////////////////////////////////////
// Part of ITableModel implementation

int CSnpTableModel::GetNumRows() const
{
    return m_numRows;
}

#include "snp_table_col_names.hpp"

int CSnpTableModel::GetNumColumns() const
{
    return  NUM_COLUMNS;
}

int CSnpTableModel::GetNumRowsUnfiltered() const
{
    return  m_nUnfilteredSnps;
}


wxString CSnpTableModel::GetColumnName(int aColIx) const
{
    if(aColIx < NUM_USER_COLUMNS)
    {
        return ToWxString(USER_COL_NAMES[aColIx]);
    }
    else {
        // shift over to bit columns
        aColIx -= NUM_USER_COLUMNS;

        if(aColIx < NUM_BIT_COLUMNS) {
            return ToWxString(BIT_COL_NAMES[aColIx]);
        } else {
            return wxT("UNKNOWN");
        }
    }
}

wxString CSnpTableModel::GetColumnType( int aColIx ) const
{
    switch( aColIx ){
        case eSnp3d:
        case eGtyAvailable:
        case eGene:
        case eOmim:
        case eDiseaseInfo:
        case eBitString:
        case eVariationClass:
        case eSequence:
        case eStudyName:
        case eRsId:
        case eAlleles:
            return wxT("string");

        case eWeight:
        case ePosition:
            return wxT("int");

        case eStudyValue:
            return wxT("double");

        default:
            return wxT("char");
    }
}

const SSnpData & CSnpTableModel::GetSnpAt(long pos) const
{
    const TSnpPtrList &listSnps = m_listSnps->GetData();
    return *listSnps[pos];
}

wxVariant CSnpTableModel::GetValueAt(int row, int col) const
{
    string value;

    const TSnpPtrList & listSnps = m_listSnps->GetData();
    if(col < NUM_USER_COLUMNS)   {
        const SSnpData& data = *listSnps[row];

        const static string snp3dUrl  = "http://www.ncbi.nlm.nih.gov/SNP/snp3D.cgi?rsnum=";
        const static string geneUrl   = "http://www.ncbi.nlm.nih.gov/sites/entrez?"
                                        "Db=gene&DbFrom=snp&Cmd=Link&LinkName=snp_gene&LinkReadableName=Gene&"
                                        "IdsFromResult=";
        const static string gtyUrl    = "http://www.ncbi.nlm.nih.gov/sites/entrez?cmd=Retrieve&db=Snp&dopt=GEN&list_uids=";
        const static string OmimUrl   = "http://www.ncbi.nlm.nih.gov/SNP/snp_redirect.cgi?snp_id=";
        const static string diseaseUrl= "http://www.ncbi.nlm.nih.gov/sites/varvu?gene=";

        switch(col) {
        default:
        case eRsId:
            try {
                value = data.rsid;

                // Convert any SNPs marked with analysis values with a '* '
           //     if (m_bShowOnlySpecial) {
                    if (value.at(0) == '-') {
                        value.replace(0,1, "* ");
                    }
           //     }

            }catch (std::exception&) { return 0; }

            break;

        case eAlleles:
            if (data.orig_feat->CanGetQual()) {
                ITERATE (CSeq_feat::TQual, it, data.orig_feat->GetQual()) {
                    const CGb_qual& qual = **it;
                    if (qual.GetQual() == "replace") {
                        value += qual.GetVal() + "/";
                    }
                }
            }
            value.erase( value.find_last_of('/') );
            break;
        case eGtyAvailable:
            if(data.bitfield.IsTrue(CSnpBitfield::eHasGenotype)) {
                value = "GTY";
            }
            break;

        case eOmim:
            if(data.bitfield.IsTrue(CSnpBitfield::eHasOMIM_OMIA)) {
                value = "OMIM";
            }
            break;

        case eDiseaseInfo:
            if(data.bitfield.IsTrue(CSnpBitfield::eHasDiseaseInfo)) {
                value = "Disease";
            }
            break;

        case eSnp3d:
            if(data.bitfield.IsTrue(CSnpBitfield::eHasSnp3D)) {
                value = "Snp3D";
            }
            break;

        case eGene:
            {
                CSnpBitfield::EFunctionClass f = data.bitfield.GetFunctionClass();
                if (f != CSnpBitfield::eUnknownFxn) {
                    value = data.bitfield.GetGenePropertyString();
                }
            }
            break;

        case eWeight:
            return data.bitfield.GetWeight();
        case ePosition:
            {{
                int pos = 0;
                if (data.loc) {
                    pos = data.loc->GetTotalRange().GetFrom() + 1;
                }
                return pos;
            }}
        case eSequence:
            if (data.loc) {
                CRef<CScope> scope = m_Scope;
                CSeq_id_Handle idh =
                    sequence::GetIdHandle(*data.loc, scope);
                idh = sequence::GetId(idh, *scope,
                                      sequence::eGetId_Best);
                idh.GetSeqId()->GetLabel(&value, CSeq_id::eContent);
            }
            break;
        case eBitString:
            value = data.bitfield.GetString();
            break;
        case eVariationClass:
            value =  data.bitfield.GetVariationClassString();
            break;
        case eStudyName:
            value = data.studyName;
            break;

        case eStudyValue:
            return data.studyValue;
        }
    }
    else {

        col -= NUM_USER_COLUMNS;
        if(col < NUM_BIT_COLUMNS)
        {
            CSnpBitfield::EProperty p = BIT_TYPE[col];

            bool bSet;
            bSet = listSnps[row]->bitfield.IsTrue(p);
            return (bSet == true) ? 'Y' : 'N';
        }
        else
        {
            return '?';
        }
    }

    return ToWxString(value);
}

END_NCBI_SCOPE
