/*  $Id: snp_table_job.cpp 27035 2012-12-11 19:21:58Z rudnev $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Melvin Quintos
 *
 *  File Description:
 *
 */

#include <ncbi_pch.hpp>

#include "snp_table_job.hpp"

#include <objects/seqres/Seq_graph.hpp>
#include <objects/seqres/Int_graph.hpp>
#include <objects/seqres/Byte_graph.hpp>
#include <objects/seqloc/Seq_loc.hpp>
#include <objects/general/User_object.hpp>

#include <objmgr/annot_selector.hpp>
#include <objmgr/feat_ci.hpp>

#include <gui/objutils/utils.hpp>

#include <gui/widgets/snp/snp_data.hpp>

#include <objects/general/Dbtag.hpp>
#include <objects/general/Object_id.hpp>
#include <objects/seq/Annot_descr.hpp>
#include <objects/seq/Annotdesc.hpp>
#include <objmgr/util/sequence.hpp>

BEGIN_NCBI_SCOPE
BEGIN_objects_SCOPE

CSnpTableJob::CSnpTableJob( CScope& scope, const CSeq_loc& loc, bool fetchStudies )
    : m_bFetchOnlySpecial(fetchStudies)
    , m_Total(100)
    , m_Completed(0)
    , m_Scope(&scope)
    , m_Loc(&loc)
{
    m_eType = eJobFetch;

    m_Result.Reset( new SResult() );

    SResult & res = (SResult&)m_Result.GetObject();
    res.eJobType = eJobFetch;
    res.listSnpsFetched.Reset( new CObjectFor<TSnpList>());
}

CSnpTableJob::CSnpTableJob( TSnpListObject * unfiltered_snps, const SSnpFilter &filter )
    : m_bFetchOnlySpecial(false)
    , m_Total(100)
    , m_Completed(0)
{
    m_eType = eJobFilter;

    m_Result.Reset( new SResult() );

    SResult & res = (SResult&)m_Result.GetObject();
    res.eJobType = eJobFilter;
    res.listSnpsFiltered.Reset( new CObjectFor<TSnpPtrList>());
    res.listSnpsFetched.Reset( unfiltered_snps );

    // make a deep copy of the filter
    string str;
    filter.SerializeTo(str);
    m_Filter.Reset( new SSnpFilter(str) );
}




CConstIRef<IAppJobProgress> CSnpTableJob::GetProgress()
{
    CRef<CAppJobProgress> p( new CAppJobProgress() );
    p->SetNormDone( float(m_Completed) / float(m_Total) );
    return CConstIRef<IAppJobProgress>( p.GetPointer() );
}


CRef<CObject> CSnpTableJob::GetResult()
{
    return m_Result;
}


CConstIRef<IAppJobError> CSnpTableJob::GetError()
{
    return CConstIRef<IAppJobError>( m_Error.GetPointer() );
}


string CSnpTableJob::GetDescr() const
{
    return string( "CSnpTableJob" );
}


IAppJob::EJobState CSnpTableJob::Run()
{
    switch (m_eType) {
        case eJobFetch:
            {{
            if (m_bFetchOnlySpecial)
                x_FetchOnlySpecial();
            else
                x_FetchAll();
            }}
            break;

        case eJobFilter:
            x_ApplyFilter();
            break;
    }

    return IsCanceled() ? eCanceled : eCompleted;
}


void CSnpTableJob::x_FetchOnlySpecial()
{
    //
    // Find the studies
    //
    SAnnotSelector sel;

    sel = CSeqUtils::GetAnnotSelector(CSeqFeatData::eSubtype_user);
    sel.AddNamedAnnots("GWAS")
       .SetExcludeExternal(false);

    SResult&    res      = (SResult&) m_Result.GetObject();
    TSnpList&   snp_list = res.listSnpsFetched->GetData();

    CFeat_CI study_it(*m_Scope, *m_Loc, sel);

    //
    // Find the SNPs
    //
    sel = CSeqUtils::GetAnnotSelector(CSeqFeatData::eSubtype_variation);
    sel.AddNamedAnnots("SNP")
       .SetExcludeExternal(false);

    CFeat_CI snp_it(*m_Scope, *m_Loc, sel);

    size_t size = snp_it.GetSize() + study_it.GetSize();

    if (IsCanceled() || size < 1) return;

    m_Total = size;

    // Generally fewer study values than SNPs for a given range
    // Build a map of study values based on RSID
    typedef list<pair<string, double> > TListStudies;  // list of 'study name' and value
    typedef map<int, TListStudies >     TMapStudies;   // map of rsid and list of studies

    TMapStudies mapStudies;
    for ( ; study_it && !IsCanceled(); ++study_it) {
        m_Completed++;
        const CSeq_feat& or_feat = study_it->GetOriginalFeature();
        const CUser_object &user = or_feat.GetData().GetUser();

        // Extract the Annotations properties
        string title;
        ITERATE( CAnnot_descr::Tdata, it, study_it->GetAnnot().GetCompleteSeq_annot()->GetDesc().Get()) {
            const CAnnotdesc &desc = **it;
            if (desc.IsUser() && desc.GetUser().GetType().IsStr() &&
                desc.IsUser() && desc.GetUser().GetType().GetStr() == "AssnDescr") {
                title = desc.GetUser().GetField("title").GetData().GetStr();
                break;
            }
        }

        string name   = title;
        double pvalue = user.GetField("value").GetData().GetReal();
        string  id    = user.GetField("id").GetData().GetStr().substr(2);
        int rsid      = NStr::StringToInt(id);

        pair<string, double> item(name, pvalue);
        mapStudies[rsid].push_back(item);
    }

    if (!IsCanceled()) {
        snp_list.clear();
        snp_list.reserve(study_it.GetSize());
    }

    // Iterate through the list of SNPs, ignore unknown rsids,
    // and populate table
    set<unsigned int> setRsIds;  // keep track of rsids in case of duplicates

    for ( ; snp_it &&  !IsCanceled(); ++snp_it) {
        m_Completed++;
        const CSeq_feat &or_feat = snp_it->GetOriginalFeature();

        // Retrieve rsid and URL
        CConstRef<CDbtag> tag = or_feat.GetNamedDbxref("dbSNP");
        if (!tag) continue;

        int rsid = tag->GetTag().GetId();

        // check if we have duplicate rsids
        if (setRsIds.find(rsid)!=setRsIds.end()) {
            continue;
        }
        else {
            setRsIds.insert(rsid);
        }


        // See if RSID has a study value
        // if so, add to the row
        TMapStudies::iterator iter = mapStudies.find(rsid);
        if (iter != mapStudies.end() ) {

            // Create snpData object
            SSnpData temp_snp;

            temp_snp.rsid = NStr::IntToString(rsid);
            temp_snp.url  = tag->GetUrl();
            temp_snp.orig_feat.Reset(&or_feat);
            temp_snp.loc.Reset(&snp_it->GetLocation());

            // Retrieve name of gene associated with SNP
            tag.Reset();  // reuse tag reference.
            CConstRef<CSeq_feat> gene_feat;
            gene_feat = sequence::GetOverlappingGene(or_feat.GetLocation(), *m_Scope);
            int gene_id = 0;
            if (gene_feat) {
                tag = gene_feat->GetNamedDbxref("GeneID");
                if ( !tag ) {
                    tag = gene_feat->GetNamedDbxref("LocusLink");
                }
            }
            if (tag) {
                gene_id = tag->GetTag().GetId();
            }
            temp_snp.gene_id = NStr::Int8ToString(gene_id);

            // Retrieve bitfield data
            if (or_feat.IsSetExt()) {
                const CUser_object &user = or_feat.GetExt();
                CConstRef<CUser_field> field =
                    user.GetFieldRef("QualityCodes");
                if (field) {
                    string s;
                    try {
                        temp_snp.bitfield = field->GetData().GetOs();
                    }
                    catch (std::length_error &e) {
                        LOG_POST(Error << e.what());
                        continue;
                    }
                }
            }

            // for each item in the list
            TListStudies &studies = iter->second;
            TListStudies::iterator sv_it = studies.begin();
            for ( ; sv_it != studies.end(); ++sv_it) {
                temp_snp.studyName  = sv_it->first;
                temp_snp.studyValue = sv_it->second;
                snp_list.push_back(temp_snp);
            }
        } // end if iter != mapStudies.end()
    }// end for  ; snp_it; ++snp_it
}


void CSnpTableJob::x_FetchAll()
{
    set<unsigned int>   setRsIds;  // store uniqure rsids
    SAnnotSelector      sel;       // annotation selector
    size_t              size;      // number of features returned

    // Create the new result object
    // and the SNP List
    SResult & res = (SResult&)m_Result.GetObject();
    TSnpList &snp_list = res.listSnpsFetched->GetData();

    //
    // Get All the Studies
    //
    sel = CSeqUtils::GetAnnotSelector(CSeqFeatData::eSubtype_user);
    sel.AddNamedAnnots("GWAS")
        .SetExcludeExternal(false).SetExactDepth(true).SetResolveDepth(0);

    CFeat_CI study_it(*m_Scope, *m_Loc, sel);

    if(IsCanceled()) return;  // check cancel request

    // Build a map of study values based on RSID
    typedef set<int>     TSetStudies;   // set of rsid with studies

    TSetStudies setStudies;
    for ( ; study_it && !IsCanceled(); ++study_it) {
        const CSeq_feat& or_feat = study_it->GetOriginalFeature();
        const CUser_object &user = or_feat.GetData().GetUser();

        string  id    = user.GetField("id").GetData().GetStr().substr(2);
        int rsid      = NStr::StringToInt(id);

        setStudies.insert(rsid);
    }

    // Prepare Annotation Selection to find the SNPs
    sel = CSeqUtils::GetAnnotSelector(CSeqFeatData::eSubtype_variation);
    sel.AddNamedAnnots("SNP")
        .SetExcludeExternal(false)
        .SetMaxSize(100000);  // TODO, make user adjustable?

    // Fetch the SNPs
    CFeat_CI feat_it(*m_Scope, *m_Loc, sel);

    // Determine the amount of features returned
    size = feat_it.GetSize();

    // make sure the data is correct
    if (size < 1)
        return;  // just leave now

    m_Total = size;

    // Prepare the list, and populate it
    snp_list.reserve(m_Total);
    m_Completed = 0;

    for (; feat_it && !IsCanceled(); ++feat_it) {
        m_Completed++;

        const CSeq_feat &       or_feat = feat_it->GetOriginalFeature();

        // Create a new row of data
        snp_list.push_back(SSnpData());
        SSnpData& snp = snp_list.back();

        // Grab the database info
        CConstRef<CDbtag> tag = or_feat.GetNamedDbxref("dbSNP");
        if (!tag) {
            snp_list.pop_back();
            continue;
        }

        // store data
        snp.orig_feat.Reset(&or_feat);
        snp.loc.Reset(&feat_it->GetLocation());
        snp.url  = tag->GetUrl();
        snp.studyValue = 0;

        int rsid = tag->GetTag().GetId();

        // Use a negative rsid to indicate that it's special (e.g. has study values)
        if (setStudies.find(rsid) != setStudies.end()) {
            rsid = -rsid;
        }

        // check for duplicates
        if (setRsIds.find(rsid)!=setRsIds.end()) {
            snp_list.pop_back();
            continue;
        }

        snp.rsid = NStr::IntToString(rsid);

        //
        // Retrieve name of gene associated with SNP
        //
        tag.Reset();  // reuse tag reference.
        CConstRef<CSeq_feat> gene_feat;
        gene_feat = sequence::GetBestOverlapForSNP(or_feat, 
                        CSeqFeatData::eSubtype_gene,
                        *m_Scope);
        int gene_id = 0;
        if (gene_feat) {
            tag = gene_feat->GetNamedDbxref("GeneID");
            if ( !tag ) {
                tag = gene_feat->GetNamedDbxref("LocusLink");
            }
        }
        if (tag) {
            gene_id = tag->GetTag().GetId();
        }

        snp.gene_id = NStr::IntToString(gene_id);

        // Retrieve bitfield data
        if (or_feat.IsSetExt()) {
            const CUser_object &    user    = or_feat.GetExt();
            CConstRef<CUser_field> field =
                user.GetFieldRef("QualityCodes");
            if (field) {
                try {
                    snp.bitfield = field->GetData().GetOs();
                }
                catch (std::length_error &e) {
                    LOG_POST(Error << e.what());
                    snp_list.pop_back();
                    continue;
                }
            }
        }

        setRsIds.insert(rsid); // save this rsid to avoid duplicates
    }
}

void CSnpTableJob::x_ApplyFilter()
{
    SResult         &res        = (SResult&)m_Result.GetObject();

    TSnpList        &theList    = res.listSnpsFetched->GetData();
    TSnpPtrList     &filtered   = res.listSnpsFiltered->GetData();

    size_t size = theList.size();

    if (size < 1)
        return;

    m_Total = size;

    // Add all original items to be filtered
    filtered.clear();
    TSnpListIter iter = theList.begin();
    for( ; iter != theList.end() && !IsCanceled(); ++iter)
        filtered.push_back(&(*iter));

    // Start filtering
    {{
    TSnpPtrListIter f_iter = filtered.begin();
    for( ; f_iter != filtered.end() && !IsCanceled(); ) {
        m_Completed++;
        const SSnpData &data = *(*f_iter);

        // Remove items that don't pass filter
        if (m_Filter->Passes(data.bitfield) == false) {
            f_iter = filtered.erase(f_iter);
        }
        else {
            ++f_iter;
        }
    }
    }}
}

END_objects_SCOPE
END_NCBI_SCOPE
