/*  $Id: view_snp_table.cpp 27556 2013-03-05 16:04:52Z katargir $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Melvin Quintos
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>

#include <gui/packages/pkg_snp/table/view_snp_table.hpp>
#include <gui/packages/pkg_snp/table/widget_snp_table.hpp>
#include <gui/packages/pkg_snp/table/model_snp_table.hpp>

#include <gui/framework/workbench.hpp>
#include <gui/core/project_service.hpp>
#include <gui/objutils/visible_range.hpp>

#include <gui/objutils/utils.hpp>
#include <gui/objutils/label.hpp>
#include <gui/objutils/user_type.hpp>

#include <corelib/ncbistre.hpp>

#include <serial/typeinfo.hpp>

#include <gui/types.hpp>

#include <gui/widgets/wx/table_utils.hpp>
#include <gui/widgets/wx/fileartprov.hpp>

#include <gui/widgets/wx/popup_event.hpp>

#include <objects/seqloc/Seq_interval.hpp>
#include <objmgr/util/sequence.hpp>

#include <wx/menu.h>
#include <wx/msgdlg.h>


BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

///////////////////////////////////////////////////////////////////////////////
//  CSnpTableView
///////////////////////////////////////////////////////////////////////////////
CSnpTableView::CSnpTableView()
    : m_bShowOnlySpecial(false)
    , m_tableWidget( NULL )
    , m_DataSource( new CSnpTableModel() )
{
    m_TypeDescr.reset( new CViewTypeDescriptor(
        "SNPs", // type name
        "snp_table_view", // icon alias TODO
        "Show tabular information about features",
        "The SNP Table View provides a tabular break-down of SNPs for a sequence",
        "SNP_TABLE_VIEW", // help ID
        "Sequence",   // category
        false       // not a singleton
    ));
}

void CSnpTableView::x_SetVisibleRange(const CSeq_loc& /*loc*/)
{
    //m_tableWidget->SetVisibleRange(loc);
}

// Ask the user if this view will show special SNPs
// This should be called AFTER x_AttachToProject is called
void CSnpTableView::x_DoDialogShowSpecial() 
{   
    string          label;
    SAnnotSelector  sel;        
    CScope* scope;
    
    scope = x_GetScope();
    m_OrigLoc->GetLabel(&label);

    // NOTE - This will try to find Analysis data
    //   by fetching the features.  This may not be the most efficient way
    //   and may need another solution if too slow.
    sel = CSeqUtils::GetAnnotSelector(CSeqFeatData::eSubtype_user);
    sel.AddNamedAnnots("GWAS").SetExcludeExternal(false);        

    CFeat_CI study_it(*scope, *m_OrigLoc, sel);

    if (study_it.GetSize() > 0) {    
        
        // Analysis data exists within the range
        // ask user for action
        string msg = "Show only special SNPs in region?\n" + label;
        int answer = wxMessageBox(wxT("Show only special SNPs?"), 
                               wxT("Question"), 
                               wxICON_QUESTION | wxYES_NO);
        if (answer == wxYES) {
            m_bShowOnlySpecial = true;
        }
    }

    // Change window title

    if (m_bShowOnlySpecial) {                        
        m_TypeDescr->SetLabel(" Special SNPs @ " + label);
    }
    else {            
        m_TypeDescr->SetLabel(" All SNPs @ " + label);
    }
    x_UpdateContentLabel();
}

void CSnpTableView::x_Update()
{
    if(m_tableWidget  &&  x_HasProject())    {
        CScope* scope = x_GetScope();

        m_DataSource->LoadData(*m_OrigLoc, scope, m_bShowOnlySpecial);
        m_tableWidget->ShowAnalysisValues(m_bShowOnlySpecial);
    }
}

//rebuild index  that is used for selection handling
void CSnpTableView::x_RebuildObjectIndex()
{
}

///////////////////////////////////////////////////////////////////////////////
// CProjectViewBase
///////////////////////////////////////////////////////////////////////////////
// override function to suppress Navigation toolbar
void CSnpTableView::GetCompatibleToolBars(vector<string>& names)
{
    names.push_back(CProjectViewToolBatFactory::sm_SelTBName);
}

void CSnpTableView::GetVisibleRanges(CVisibleRange& /*vrange*/) const
{
}

void CSnpTableView::OnVisibleRangeChanged(const CVisibleRange& vrange,
                                          IVisibleRangeClient* /*source*/)
{
    if( vrange.Clamp( const_cast<CSeq_loc&>(*m_OrigLoc) ) ){
        x_SetVisibleRange( *m_OrigLoc );
    }
}

///////////////////////////////////////////////////////////////////////////////
// IProjectView
///////////////////////////////////////////////////////////////////////////////
bool CSnpTableView::InitView(TConstScopedObjects& objects, const objects::CUser_object*)
{
    CScope* scope = NULL;
    const CObject* object = NULL;

    if (objects.size() == 1) {
        object = objects[0].object.GetPointer();
        scope = objects[0].scope.GetPointer();
    }

    m_OrigLoc.Reset(dynamic_cast<const CSeq_loc*>(object));

    if (m_OrigLoc)   {
        CIRef<CProjectService> srv = m_Workbench->GetServiceByType<CProjectService>();
        _ASSERT(srv);

        CRef<CWorkspaceConstPrx> workspace = srv->GetWorkspaceConst();
        CRef<CProjectPrx> project = workspace->FindProjectByScope(scope);

        _ASSERT(project);

        x_AttachToProject(*project);
        x_DoDialogShowSpecial();
        OnProjectChanged();
        return true;
    } else {
        // cannot represent the data
        x_ReportInvalidInputData(objects);
        return false;
    }
}

void CSnpTableView::OnProjectChanged()
{
    if (x_HasProject()) {
        x_Update();
    }
    x_UpdateContentLabel();
}

///////////////////////////////////////////////////////////////////////////////
// IWMClient
///////////////////////////////////////////////////////////////////////////////
wxWindow* CSnpTableView::GetWindow()
{
    _ASSERT(m_tableWidget);
    return m_tableWidget;
}

///////////////////////////////////////////////////////////////////////////////
// IView
///////////////////////////////////////////////////////////////////////////////
const CViewTypeDescriptor& CSnpTableView::GetTypeDescriptor() const
{
    return *m_TypeDescr;
}

void CSnpTableView::CreateViewWindow(wxWindow* parent)
{
    _ASSERT( !m_tableWidget );

    m_tableWidget = new CSnpTableWidget( parent );
    m_tableWidget->SetDataSource( m_DataSource );

    AddListener( m_tableWidget, ePool_Child );
    m_tableWidget->AddListener( this, ePool_Parent );
}

void CSnpTableView::DestroyViewWindow()
{
    if(m_tableWidget)   {
        m_tableWidget->RemoveModel();
        m_tableWidget->Destroy();
        m_tableWidget = NULL;
    }
    //TODO delete m_DataSource ?
}

///////////////////////////////////////////////////////////////////////////////
// ISelectionClient
///////////////////////////////////////////////////////////////////////////////
void CSnpTableView::GetSelection(TConstScopedObjects& objs) const
{
    if (m_tableWidget  &&  x_HasProject())    {
        CScope* scope = x_GetScope();

        // add Object selections
        TConstObjects sel_objs;
        m_tableWidget->GetSelection(sel_objs);

        ITERATE(TConstObjects, it, sel_objs)    {
            objs.push_back(SConstScopedObject(**it, *scope));
        }
    }
}


void CSnpTableView::GetSelection(CSelectionEvent& evt) const
{
    if (x_HasProject()) {
        // Object Selection
        TConstObjects sel_objs;
        m_tableWidget->GetSelection(sel_objs);
        evt.AddObjectSelection(sel_objs);
    }
}

void CSnpTableView::x_InitMenuResources()
{
    static bool initialized = false;

    // this code should execute only once
    if( ! initialized ){
        // register commands
        //CUICommandRegistry& cmd_reg = m_Workbench->GetUICommandRegistry();
        //cmd_reg->RegisterCommand( ... );

        // register icons
        //wxFileArtProvider& provider = m_Workbench->GetFileArtProvider();
        //provider.RegisterFileAlias( ... );

        initialized = true;
    }
}

void CSnpTableView::x_CreateMenuBarMenu()
{
    x_InitMenuResources();

    m_MenuBarMenu.reset( new wxMenu() );
}





///////////////////////////////////////////////////////////////////////////////
// IRegSettings
///////////////////////////////////////////////////////////////////////////////
void CSnpTableView::SetRegistryPath(const string& reg_path)
{
    m_RegPath = reg_path;
}

void CSnpTableView::LoadSettings()
{
    /// restore our table's settings
    CGuiRegistry& reg = CGuiRegistry::GetInstance();
    CGuiRegistry::TReadView view = reg.GetReadView( m_RegPath );
    LoadTableSettings( *m_tableWidget, view );
}


void CSnpTableView::SaveSettings() const
{
    /// save our table's settings
    CGuiRegistry& reg = CGuiRegistry::GetInstance();
    CGuiRegistry::TReadWriteView view = reg.GetReadWriteView( m_RegPath );
    SaveTableSettings( *m_tableWidget, view );
}




const CObject* CSnpTableView::x_GetOrigObject() const
{
    return m_OrigLoc.GetPointer();
}

// handles incoming selection broadcast (overriding CView virtual function)
void CSnpTableView::x_OnSetSelection( CSelectionEvent& /*evt*/ )
{
    //m_StatusBar->SetMessage( "No features." );

    /* MRQ TODO
    if( evt.HasObjectSelection() ){ // set Object Selection
        m_tableWidget->ClearSelection();

        TConstObjects sel_objs;
        evt.GetAllObjects(sel_objs);

        CScope& scope = *m_DataSource->GetScope();
        CObjectIndex::TResults results;

        ITERATE( TConstObjects, it_obj, sel_objs ){
            const CObject& obj = **it_obj;
            results.clear();
            m_ObjectIndex.GetMatches( obj, scope, results );

            ITERATE( CObjectIndex::TResults, it_r, results ){
                const CSelFeatHandle* h = reinterpret_cast<const CSelFeatHandle*>(*it_r);
                m_tableWidget->SelectFeature(h->GetFeature());
            }
        }

        //m_tableWidget->ScrollSelectionToVisible();

        x_UpdateStatusMessage(); // prepare our status bar message
    }
    */
}

END_NCBI_SCOPE
