/*  $Id: widget_snp_table.cpp 23953 2011-06-23 19:29:53Z wuliangs $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Melvin Quintos
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>

#include <gui/packages/pkg_snp/table/widget_snp_table.hpp>
#include <gui/packages/pkg_snp/table/model_snp_table.hpp>
#include <gui/widgets/snp/filter/snp_filter_list_dlg.hpp> // TODO move files

#include <gui/objutils/utils.hpp>
#include <gui/objutils/label.hpp>
#include <gui/objutils/obj_event.hpp>
#include <gui/utils/app_popup.hpp>
#include <gui/widgets/wx/table_utils.hpp>
#include <gui/widgets/wx/ui_command.hpp>
#include <gui/widgets/wx/popup_event.hpp>
#include <gui/widgets/wx/wx_utils.hpp>
#include <objects/seqloc/Seq_bond.hpp>
#include <objects/seqloc/Seq_interval.hpp>
#include <objects/seqloc/Seq_loc.hpp>
#include <objects/seqloc/Seq_loc_equiv.hpp>
#include <objects/seqloc/Seq_point.hpp>
#include <objmgr/feat_ci.hpp>
#include <objmgr/util/feature.hpp>
#include <objmgr/util/sequence.hpp>
#include <wx/menu.h>

#include "snp_table_job.hpp"

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

/// Definitions for LinkoutCommands
///  Only used by CSnpTableWidget
enum ELinkOutCommands {
    eCmdLinkOMIM = eBaseCmdLast + 6500,
    eCmdLinkRefSNP,
    eCmdLinkSnp3D,
    eCmdLinkGene,
    eCmdLinkGTY,
    eCmdLinkDisease,
    eCmdExportCSV,
    eCmdSelectFilter
};

BEGIN_EVENT_TABLE(CSnpTableWidget, CwxTableListCtrl)
    EVT_CONTEXT_MENU(CSnpTableWidget::OnContextMenu)
    EVT_MENU(eCmdLinkRefSNP, CSnpTableWidget::OnEventLinkout)
    EVT_MENU(eCmdLinkOMIM, CSnpTableWidget::OnEventLinkout)
    EVT_MENU(eCmdLinkGTY, CSnpTableWidget::OnEventLinkout)
    EVT_MENU(eCmdLinkGene, CSnpTableWidget::OnEventLinkout)
    EVT_MENU(eCmdLinkSnp3D, CSnpTableWidget::OnEventLinkout)
    EVT_MENU(eCmdLinkDisease, CSnpTableWidget::OnEventLinkout)
    EVT_MENU(eCmdExportCSV, CSnpTableWidget::OnExportCSV)
    EVT_MENU(eCmdSelectFilter, CSnpTableWidget::OnSelectFilter)
END_EVENT_TABLE()

static
WX_DEFINE_MENU(sPopupMenu)
    WX_MENU_SEPARATOR_L("Top SNP Actions")
    WX_MENU_SEPARATOR()
    WX_MENU_ITEM(eCmdLinkRefSNP)
    WX_MENU_ITEM(eCmdLinkGene)
    WX_MENU_ITEM(eCmdLinkGTY)
    WX_MENU_ITEM(eCmdLinkOMIM)
    WX_MENU_ITEM(eCmdLinkSnp3D)
    WX_MENU_ITEM(eCmdLinkDisease)
    WX_MENU_SEPARATOR()
    WX_MENU_ITEM(eCmdExportCSV)
    WX_MENU_ITEM(eCmdSelectFilter)
WX_END_MENU()

static const char LINK_GENE[]       = "Gene Function Link";
static const char LINK_OMIM[]       = "OMIM Link";
static const char LINK_REFSNP[]     = "RefSNP Link";
static const char LINK_SNP3D[]      = "SNP3D Link";
static const char LINK_GTY[]        = "Genotype Link";
static const char LINK_DISEASE[]    = "Disease Link";
static const char EXPORT_CSV[]      = "Export to CSV";
static const char SELECT_FILTER[]   = "Filter Results";

void CSnpTableWidget::RegisterCommands(CUICommandRegistry& cmd_reg, 
                                       wxFileArtProvider& /*provider*/)
{    
    static bool initialized = false;

    if (!initialized) {
        cmd_reg.RegisterCommand(eCmdLinkGene,
            LINK_GENE, LINK_GENE, "", "");
        cmd_reg.RegisterCommand(eCmdLinkOMIM,
            LINK_OMIM, LINK_OMIM, "", "");
        cmd_reg.RegisterCommand(eCmdLinkRefSNP,
            LINK_REFSNP, LINK_REFSNP, "", "");
        cmd_reg.RegisterCommand(eCmdLinkSnp3D,
            LINK_SNP3D, LINK_SNP3D, "", "");
        cmd_reg.RegisterCommand(eCmdLinkGTY,
            LINK_GTY, LINK_GTY, "", "");
        cmd_reg.RegisterCommand(eCmdLinkDisease,
            LINK_DISEASE, LINK_DISEASE, "", "");
        cmd_reg.RegisterCommand(eCmdExportCSV,
            EXPORT_CSV, EXPORT_CSV, "", "");
        cmd_reg.RegisterCommand(eCmdSelectFilter,
            SELECT_FILTER, SELECT_FILTER, "", "");
    
        initialized = true;
    }
}

/////////////////////////////////////////////////////////////////////////////
/// CSnpTableWidget

CSnpTableWidget::CSnpTableWidget(
    wxWindow* parent, wxWindowID id, const wxPoint& pos, const wxSize& size, long style
)
    : CwxTableListCtrl( parent, id, pos, size, style )
{
    m_SnpTableModel.Reset( new CSnpTableModel() );
    SetModel( m_SnpTableModel );
}

CSnpTableWidget::~CSnpTableWidget()
{
}

void CSnpTableWidget::OnSelectFilter( wxCommandEvent& /*event*/)
{
    // Bring up the CSnpFilterListDialog
    // Get return data
    // If SNPs were modified, broadcast event
    // Do an update event using the name of the filter

    CSnpFilterListDialog dlg;
    
    // Create the dialog showing the 'Edit' button
    dlg.CreateX(true, this);

    if (dlg.ShowModal() == wxID_OK) {
        SSnpFilter filter;
        dlg.GetFilter(&filter);
        // TODO if (dlg.HasModifiedFilters()) x_BroadCastUpdate();
        m_SnpTableModel->ApplyFilter(filter);

    }    
}

void CSnpTableWidget::OnExportCSV( wxCommandEvent& /*event*/ )
{
    ExportTableControlToCSV(*this);
}

void CSnpTableWidget::OnEventLinkout( wxCommandEvent& event )
{    
    const static string snp3dUrl  = "http://www.ncbi.nlm.nih.gov/SNP/snp3D.cgi?rsnum=";
    const static string geneUrl   = "http://www.ncbi.nlm.nih.gov/sites/entrez?"
                                    "Db=gene&DbFrom=snp&Cmd=Link&LinkName=snp_gene&LinkReadableName=Gene&"
                                    "IdsFromResult=";
    const static string gtyUrl    = "http://www.ncbi.nlm.nih.gov/sites/entrez?cmd=Retrieve&db=Snp&dopt=GEN&list_uids=";
    const static string OmimUrl   = "http://www.ncbi.nlm.nih.gov/SNP/snp_redirect.cgi?snp_id=";
    const static string diseaseUrl= "http://www.ncbi.nlm.nih.gov/sites/varvu?gene=";
    
    long sel = GetFirstSelected();
    
    if (sel < 0) 
        return;
    sel = RowVisibleToData(sel);

    string url;
    const SSnpData& snp = m_SnpTableModel->GetSnpAt(sel);
    
    switch (event.GetId()) {
        case eCmdLinkRefSNP:    url = snp.url;                  break;
        case eCmdLinkOMIM:      url = OmimUrl + snp.rsid;       break;
        case eCmdLinkGTY:       url = gtyUrl + snp.rsid;        break;
        case eCmdLinkDisease:   url = diseaseUrl + snp.gene_id; break;
        case eCmdLinkSnp3D:     url = snp3dUrl + snp.rsid;      break;
        case eCmdLinkGene:      
            {   
                CSnpBitfield::EFunctionClass f = snp.bitfield.GetFunctionClass();                
                if (f == CSnpBitfield::eMultipleFxn) {
                    url = snp.url + "#locus";
                }
                else {
                    url = geneUrl + snp.rsid;
                }
            } break;
    }
    
    CAppPopup::PopupURL(url);
}

void CSnpTableWidget::OnContextMenu( wxContextMenuEvent& /*event*/ )
{
    CUICommandRegistry& cmd_reg = CUICommandRegistry::GetInstance();
    wxMenu* menu = cmd_reg.CreateMenu( sPopupMenu );

    x_PrepareMenu(menu);
    PopupMenu(menu);
    delete menu;
}


wxListItemAttr* CSnpTableWidget::OnGetItemAttr( long aRow ) const
{
    // if row is in set of rows to be colored
    //   set static attribute to a background color
    //  
    static wxListItemAttr s_attr;
        
    wxString str = GetValueVisibleAt(aRow, 0).GetString();
    
    if        (m_bShowStudies && str.at(0) == '*') {
        s_attr.SetBackgroundColour( wxColour(150,229,255));
    } else if ( aRow%2 != 0 ){        
        s_attr.SetBackgroundColour( wxColour(255,255,232));
    } else {
        s_attr.SetBackgroundColour( wxColour(255,255,182));
    }

    return &s_attr;
}


void CSnpTableWidget::x_PrepareMenu(wxMenu *menu)
{
    // Retrieve selected row
    int id;
    int sel;
    bool hasRefSNP, hasSnp3D, hasGty, hasOMIM, hasDisease, hasGeneFxn;
    wxMenuItem *item;

    // initialize to false
    hasRefSNP = hasSnp3D = hasGty = hasOMIM = hasDisease = hasGeneFxn = false;

    sel = GetFirstSelected();
    
    if (sel >= 0) {    
        sel = RowVisibleToData(sel);

        const SSnpData& snp = m_SnpTableModel->GetSnpAt(sel);
        
        hasRefSNP   = true; // All stored SNPs in ID have a refSNP page
        hasOMIM     = snp.bitfield.IsTrue(CSnpBitfield::eHasOMIM_OMIA);
        hasSnp3D    = snp.bitfield.IsTrue(CSnpBitfield::eHasSnp3D);
        hasGty      = snp.bitfield.IsTrue(CSnpBitfield::eHasGenotype);
        hasDisease  = snp.bitfield.IsTrue(CSnpBitfield::eHasDiseaseInfo);
        hasGeneFxn  = !snp.bitfield.IsTrue(CSnpBitfield::eUnknownFxn);
    }

    id      = menu->FindItem(ToWxString(LINK_REFSNP));
    item    = menu->FindItem(id);
    item->Enable(hasRefSNP);

    id      = menu->FindItem(ToWxString(LINK_GENE));
    item    = menu->FindItem(id);
    item->Enable(hasGeneFxn);
    
    id      = menu->FindItem(ToWxString(LINK_OMIM));
    item    = menu->FindItem(id);
    item->Enable(hasOMIM);
    
    id      = menu->FindItem(ToWxString(LINK_SNP3D));
    item    = menu->FindItem(id);
    item->Enable(hasSnp3D);
    
    id      = menu->FindItem(ToWxString(LINK_GTY));
    item    = menu->FindItem(id);
    item->Enable(hasGty);
    
    id      = menu->FindItem(ToWxString(LINK_DISEASE));
    item    = menu->FindItem(id);
    item->Enable(hasDisease);
    
    CleanupSeparators(*menu);

}

void CSnpTableWidget::ShowAnalysisValues(bool show)
{
    if (show) {
        ShowColumn(1);
        ShowColumn(2);
    }
    else {
        HideColumn(1);
        HideColumn(2);
    }
}

void CSnpTableWidget::SetDataSource(CSnpTableModel* new_ds)
{
    if (new_ds != m_SnpTableModel.GetPointer())    {
        RemoveModel();

        m_SnpTableModel.Reset(new_ds);
        m_SnpTableModel->AddTMListener(*this);

        SetModel(new_ds);
    }
}



void CSnpTableWidget::GetSelection(TConstObjects& /*objs*/) const
{
    /*
    wxArrayInt objIxs = GetDataRowsSelected();

    for( int i = 0; i < (int)objIxs.size(); ++i ){
        const CSeq_feat& feat = m_DataSource->GetOriginalFeature( objIxs[i] );
        objs.push_back( CConstRef<CObject>( &feat ) );
    }
    */
}


/*
void CSnpTableWidget::SaveSettings(CGuiRegistry::TReadWriteView& view) const
{
    ncbi::SaveTableSettings(*m_Table, view);
}


void CSnpTableWidget::LoadSettings(CGuiRegistry::TReadView& view)
{
    ncbi::LoadTableSettings(*m_Table, view);
}
*/
END_NCBI_SCOPE
