/*  $Id: fasta_retrieve_job.cpp 26643 2012-10-18 17:54:05Z katargir $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Authors:  Roman Katargin
 *
 * File Description:
 *
 */

#include <ncbi_pch.hpp>

#include <objmgr/bioseq_handle.hpp>
#include <objmgr/bioseq_ci.hpp>
#include <objmgr/util/sequence.hpp>

#include <gui/widgets/text_widget/expand_item.hpp>
#include <gui/widgets/text_widget/composite_text_item.hpp>
#include <gui/widgets/text_widget/text_item_panel.hpp>

#include <gui/widgets/seq/fasta_view_params.hpp>
#include <gui/widgets/seq/fasta_title_item.hpp>
#include <gui/widgets/seq/fasta_seqhdr_item.hpp>
#include <gui/widgets/seq/fasta_seqblock_item.hpp>

#include <gui/widgets/seq/fasta_retrieve_job.hpp>

BEGIN_NCBI_SCOPE
USING_SCOPE(objects);

CFastaRetrieveJob::CFastaRetrieveJob(CSeq_entry_Handle& h) : m_Handle(h)
{
}

CFastaRetrieveJob::~CFastaRetrieveJob()
{
}

IAppJob::EJobState CFastaRetrieveJob::x_Run()
{
    if (!m_Handle) return eCompleted;

    for (CBioseq_CI it(m_Handle);  it;  ++it) {
        if (IsCanceled())
            break;
        CFastaTitleItem* titleItem = new CFastaTitleItem(*it);
        CCompositeTextItem* expanded = new CCompositeTextItem();
        expanded->AddItem(new CFastaTitleItem(*it));
        TSeqPos seqLength = it->GetBioseqLength();
        size_t blockSize = kFastaSeqLine*kFastaSeqBlockSize;
        size_t blockCount = (seqLength + blockSize - 1)/blockSize;

        for (size_t i = 0; i < blockCount; ++i) {
            if (IsCanceled())
                break;
            size_t from = i*blockSize;
            size_t count = (i + 1 == blockCount) ? seqLength - from : blockSize;
            CFastaSeqHdrItem* seqCollapsed = new CFastaSeqHdrItem(from, count, *titleItem);
            CFastaSeqBlockItem* seqExpanded = new CFastaSeqBlockItem(*it, from, count);
            expanded->AddItem(new CExpandItem(seqCollapsed, seqExpanded, (blockCount == 1)));
        }

        {{
            CFastMutexGuard guard(m_Mutex);
            m_Items.push_back(new CExpandItem(titleItem, expanded));
        }}
    }

    CFastMutexGuard guard(m_Mutex);
    if (!m_Items.empty()) {
        m_Result.Reset(new CTextViewRetrieveResult());

        CCompositeTextItem::TInsertBlockVec newItems;
        newItems.push_back(CCompositeTextItem::TInsertBlock(m_Items, -2));
        m_Result->SetItems(newItems);
        m_Items.clear();
    }

    return eCompleted;
}

CConstIRef<IAppJobProgress> CFastaRetrieveJob::GetProgress()
{
    CFastMutexGuard guard(m_Mutex);

    CTextViewRetrieveResult* result = new CTextViewRetrieveResult();

    CCompositeTextItem::TInsertBlockVec newItems;
    newItems.push_back(CCompositeTextItem::TInsertBlock(m_Items, -2));
    result->SetItems(newItems);
    m_Items.clear();

    CTextViewRetrieveProgress* progress = new CTextViewRetrieveProgress(result);
    return CConstIRef<IAppJobProgress>(progress);
}

CFastaRetrieveLocJob::CFastaRetrieveLocJob(const objects::CSeq_loc& loc, objects::CScope& scope)
    : m_Loc(&loc), m_Scope(&scope)
{
}

CFastaRetrieveLocJob::~CFastaRetrieveLocJob()
{
}

IAppJob::EJobState CFastaRetrieveLocJob::x_Run()
{
    if (!m_Loc) return eCompleted;

    for( CSeq_loc_CI it(*m_Loc);  it;  ++it) {
        if (IsCanceled())
            break;

        CBioseq_Handle handle = m_Scope->GetBioseqHandle(it.GetSeq_id());
        if (!handle)
            continue;

        CConstRef<CSeq_loc> loc = it.GetRangeAsSeq_loc();
        CFastaTitleItem* titleItem = new CFastaTitleItem(handle, loc);
        CCompositeTextItem* expanded = new CCompositeTextItem();
        expanded->AddItem(new CFastaTitleItem(handle, loc));

        CSeq_loc_CI::TRange range = it.GetRange();
        TSeqPos seqLength = range.GetLength();
        size_t blockSize = kFastaSeqLine*kFastaSeqBlockSize;
        size_t blockCount = (seqLength + blockSize - 1)/blockSize;

        for (size_t i = 0; i < blockCount; ++i) {
            if (IsCanceled())
                break;
            size_t from = range.GetFrom() + i*blockSize;
            size_t count = (i + 1 == blockCount) ? seqLength - i*blockSize : blockSize;
            CFastaSeqHdrItem* seqCollapsed = new CFastaSeqHdrItem(from, count, *titleItem);
            CFastaSeqBlockItem* seqExpanded = new CFastaSeqBlockItem(handle, from, count);
            expanded->AddItem(new CExpandItem(seqCollapsed, seqExpanded, (blockCount == 1)));
        }

        {{
            CFastMutexGuard guard(m_Mutex);
            m_Items.push_back(new CExpandItem(titleItem, expanded));
        }}
    }

    CFastMutexGuard guard(m_Mutex);
    if (!m_Items.empty()) {
        m_Result.Reset(new CTextViewRetrieveResult());

        CCompositeTextItem::TInsertBlockVec newItems;
        newItems.push_back(CCompositeTextItem::TInsertBlock(m_Items, -2));
        m_Result->SetItems(newItems);
        m_Items.clear();
    }

    return eCompleted;
}

CConstIRef<IAppJobProgress> CFastaRetrieveLocJob::GetProgress()
{
    CFastMutexGuard guard(m_Mutex);

    CTextViewRetrieveResult* result = new CTextViewRetrieveResult();

    CCompositeTextItem::TInsertBlockVec newItems;
    newItems.push_back(CCompositeTextItem::TInsertBlock(m_Items, -2));
    result->SetItems(newItems);
    m_Items.clear();
    CTextViewRetrieveProgress* progress = new CTextViewRetrieveProgress(result);
    return CConstIRef<IAppJobProgress>(progress);
}


END_NCBI_SCOPE
